note

	description:

		"[
        String interface to managed C memory

        This is a simple wrapper around a MANAGED_POINTER.
       ]"

	author: "David Schwartz"
	last_editor: "$Author: c4 $"
	date: "$Date: 2014/02/14 $"
	revision: "$Revision: #1 $"
	path: "$File: //rose/source/library/external/src/rose_bytes.e $"

class	ROSE_BYTES

create

	make,
	make_empty,
	make_from_pointer

feature {NONE} -- Initialization

	make (a_string: STRING)
			-- Make as copy of `a_string'.
		require
			a_string_not_void: a_string /= Void
		local
			i, n: INTEGER
		do
			if a_string.count = 0 then
				n := a_string.capacity
			else
				n := a_string.count
			end
			create item.make (n + 1)
			from
				i := 0
			until
				i > n
			loop
				item.put_natural_8 (a_string.area.item (i).code.to_natural_8, i)
				i := i + 1
			end
		end

	make_empty (a_length: INTEGER)
			-- Make an empty C string of `a_length' characters.
			-- C memory area is not initialized.
		require
			positive_length: a_length >= 0
		do
			create item.make (a_length)
		end

	make_from_pointer (a_pointer: POINTER; a_length: INTEGER)
			-- Initialize from `a_length' bytes at `a_pointer'.
		require
			a_pointer /= default_pointer and then a_length >= 0
		do
			create item.make_from_pointer (a_pointer, a_length)
		end

feature -- Access

	count: INTEGER
			-- Number of 8-bit characters in `Current'
		do
			Result := item.count - 1
		end
	
	string_at (a_start, a_end: INTEGER): STRING
			-- Substring from `a_start' to `a_end' inclusive
		require
			a_start >= 1 and then a_end <= count
		local
			i, nb: INTEGER
		do
			from
				i := a_start - 1
				nb := a_end
				create Result.make (nb - i)
			until
				i >= nb
			loop
				Result.extend (item.read_natural_8 (i).to_character_8)
				i := i + 1
			end
		ensure
			result_not_void: Result /= Void
		end

	string: STRING
			-- `Current' converted to STRING_8
		do
			Result := string_at (1, count)
		end

	integer_at (i: INTEGER): INTEGER
			-- 32-bit integer at position `i'
		require
			i_strictly_positive: i > 0
			valid_position: (i - 1 + {PLATFORM}.integer_32_bytes) <= count
		do
			Result := item.read_integer_32 (i - 1)
		end

	integer_8_at (i: INTEGER): INTEGER_8
			-- 8-bit integer at position `i'
		require
			i_strictly_positive: i > 0
			valid_position: (i - 1 + {PLATFORM}.integer_8_bytes) <= count
		do
			Result := item.read_integer_8 (i - 1)
		end

	integer_16_at (i: INTEGER): INTEGER_16
			-- 16-bit integer at position `i'
		require
			i_strictly_positive: i > 0
			valid_position: (i - 1 + {PLATFORM}.integer_16_bytes) <= count
		do
			Result := item.read_integer_16 (i - 1)
		end

	real_at (i: INTEGER): REAL
			-- 32-bit floating-point value at position `i'
		require
			i_strictly_positive: i > 0
			valid_position: (i - 1 + {PLATFORM}.real_32_bytes) <= count			
		do
			Result := item.read_real_32 (i - 1)
		end

	double_at (i: INTEGER): DOUBLE
			-- 64-bit floating-point value at position `i'
		require
			i_strictly_positive: i > 0
			valid_position: (i - 1 + {PLATFORM}.real_64_bytes) <= count	
		do
			Result := item.read_real_64 (i - 1)
		end

	character_at (i: INTEGER): CHARACTER
			-- 8-bit character at position `i'
		require
			i_strictly_positive: i > 0
			valid_position: (i - 1 + {PLATFORM}.character_8_bytes) <= count	
		do
			Result := item.read_character (i - 1)
		end

feature -- Basic operations

	set_count (a_count: INTEGER)
			-- Set `count' to `a_count'.
		require
			a_count_nonnegative: a_count >= 0
		do
			item.resize (a_count)
		end

	put_integer (a_value, i: INTEGER)
			-- Insert `a_value' at `i'.
		require
			i_strictly_positive: i > 0
			valid_position: (i - 1 + {PLATFORM}.integer_32_bytes) <= count
		do
			item.put_integer_32 (a_value, i - 1)
		ensure
			inserted: a_value = item.read_integer_32 (i - 1)
		end

	put_integer_8 (a_value: INTEGER_8; i: INTEGER)
			-- Insert `a_value' at `i'.
		require
			i_strictly_positive: i > 0
			valid_position: (i - 1 + {PLATFORM}.integer_8_bytes) <= count
		do
			item.put_integer_8 (a_value, i)
		ensure
			inserted: a_value = item.read_integer_8 (i - 1)
		end

	put_integer_16 (a_value: INTEGER_16; i: INTEGER)
			-- Insert `a_value' at `i'.
		require
			i_strictly_positive: i > 0
			valid_position: (i - 1 + {PLATFORM}.integer_16_bytes) <= count	
		do
			item.put_integer_16 (a_value, i - 1)
		ensure
			inserted: a_value = item.read_integer_16 (i - 1)			
		end

	put_real (a_value: REAL; i: INTEGER)
			-- Insert `a_value' at `i'.
		require
			i_strictly_positive: i > 0
			valid_position: (i - 1 + {PLATFORM}.real_32_bytes) <= count
		do
			item.put_real_32 (a_value, i - 1)
		ensure
			inserted: (a_value = a_value) implies (a_value = item.read_real_32 (i - 1))
		end

	put_double (a_value: DOUBLE; i: INTEGER)
			-- Insert `a_value' at `i'.
		require
			i_strictly_positive: i > 0
			valid_position: (i - 1 + {PLATFORM}.real_64_bytes) <= count
		do
			item.put_real_64 (a_value, i - 1)
		ensure
			inserted: (a_value = a_value) implies (a_value = item.read_real_64 (i - 1))		
		end

	put_character (a_value: CHARACTER; i: INTEGER)
			-- Insert `a_value' at `i'.
		require
			i_strictly_positive: i > 0
			valid_position: (i - 1 + {PLATFORM}.character_8_bytes) <= count	
		do
			item.put_character (a_value, i - 1)
		ensure
			inserted: a_value = item.read_character (i - 1)
		end

	pointer_to (a_offset: INTEGER): POINTER
			-- Pointer to character at position `a_offset'
		require
			a_offset_strictly_positive: a_offset > 0
			valid_position: (a_offset - 1 + {PLATFORM}.character_8_bytes) <= count
		do
			Result := item.item + (a_offset - 1)
		end

feature {NONE} -- Implementation

	item: MANAGED_POINTER
			-- Management of C-memory
	
invariant

	item_attached: item /= Void

end
