note

	description: "Fixed size block with bytes as integers"

	author: "Franck Arnaud (from pre-AXA Rosenberg days)"
	last_editor: "$Author: c4 $"
	date: "$Date: 2014/02/10 $"
	revision: "$Revision: #1 $"
	path: "$File: //rose/source/third_party/crypto/src/crypt_bytes.e $"
	project: "http://www.nenie.org/eiffel/"
	license: "http://www.opensource.org/licenses/mit-license.html"


deferred class CRYPT_BYTES

inherit

	CRYPT_BIT32_PRIVATE_ROUTINES
		export
			{ANY} is_byte
		end
	
	KL_IMPORTED_STRING_ROUTINES

	KL_IMPORTED_INTEGER_ROUTINES
		export {NONE} all end
		
feature -- Bounds

	lower: INTEGER = 1
			-- One
			
	upper: INTEGER
			-- Upper bound of block (inclusive)
		do
			Result := count
		end
		
	count: INTEGER
			-- Number of bytes
		deferred
		end
		
 feature -- Definition
		
	item (i: INTEGER): NATURAL_8
			-- Byte `i'
		require
			lower: i >= lower
			upper: i <= upper
		deferred
		ensure
			byte: is_byte (Result)
		end
		
	put (a_byte: NATURAL_8; i: INTEGER)
			-- Put `a_byte' at `i'.
		require
			lower: i >= lower
			upper: i <= upper
			byte: is_byte (a_byte)
		deferred
		ensure
			set: a_byte = item (i)
		end

feature -- Copy

	copy_bytes (a: CRYPT_BYTES)
			-- Copy `a' to `Current'.
		require
			a_attached: a /= Void
			size: a.count = count
		local
			i: INTEGER
		do
			from
				i := 1
			until
				i > count
			loop
				put (a.item (i), i)
				i := i + 1
			end
		ensure
			same: same_bytes (a)
		end
		
	same_bytes (a: CRYPT_BYTES): BOOLEAN
			-- Is `a' equal to `Current'?
		require
			a_attached: a /= Void
		local
			i: INTEGER
		do
			Result := (a = Current)
			if (not Result) and (a.count = count) then
				from
					i := 1
					Result := True
				until
					(i > count) or (not Result)
				loop
					Result := (a.item (i) = item (i))
					i := i + 1
				end
			end
		end
		
	copy_bytes_at (a: CRYPT_BYTES; a_from: INTEGER)
			-- Copy bytes from within larger block `a', starting at positon`a_from' to `Current'.
		require
			a_attached: a /= Void
			inside_min: a_from > 0
			inside_max: a_from + count - 1 <= a.count
		local
			i: INTEGER
		do
			from
				i := 1
			until
				i > count
			loop
				put (a.item (i + a_from - 1), i)
				i := i + 1
			variant
				count - i + 1				
			end
		end

feature -- Operations

	xor_bytes (a: CRYPT_BYTES)
			-- `Current' := `Current' xor `a', bitwise
			-- Partial if `a' shorter than `Current'.
		require
			a_attached: a /= Void
			size: a.count <= count
		local
			i: INTEGER
		do
			from
				i := a.count
			until
				i = 0
			loop
				put (byte_xor (a.item (i), item (i)), i)
				i := i - 1
			variant
				i
			end
		end
		
feature -- First/last

	first: NATURAL_8
			-- First byte
		do
			Result := item (1)
		ensure
			definition: Result = item (1)
		end
		
	last: NATURAL_8
			-- Last byte
		do
			Result := item (count)
		ensure
			definition: Result = item (count)
		end
				
feature -- Hexadecimal

	hex_string: STRING
			-- Hexadecimal representation of `Current'
		local
			j: INTEGER
			a_byte: NATURAL_8
		do
			from
				create Result.make (0)
				j := 0
			until
				j = count
			loop
				a_byte := item (lower+j)
				Result.append_character (int_to_hex((a_byte//16)\\16))
				Result.append_character (int_to_hex((a_byte\\16)))
				j := j + 1
			end
		ensure
			hex_string_attached: Result /= Void
		end
		
	from_hex (a: STRING)
			-- Initialise from hexadecimal string `a'.
		require
			a_attached: a /= Void
			size: a.count = (count * 2)
			hex: STRING_.is_hexadecimal (a)
		local
			i, j: INTEGER
			a_byte: NATURAL_8
		do
			from
				i := 1
				j := 1
			until
				i > count
			loop
				a_byte := hex_to_int (a.item (j)) * 16
				j := j + 1
				a_byte := a_byte + hex_to_int (a.item (j))
				j := j + 1
				
				put (a_byte, i)
				i := i + 1
			end
		ensure
			reverse: hex_string.is_equal (a.as_lower)
		end

feature -- String as byte block

	from_string (a: STRING)
			-- Initialise from `a' assuming 1 byte per character.
		require
			a_attached: a /= Void
			size: a.count = count
			max_char: across (1 |..| count) as i_string all a.item_code (i_string.item) <= 255 end
		local
			i: INTEGER
		do
			from
				i := 1
			until
				i > count
			loop
				put (a.item_code (i).as_natural_8, i)
				i := i + 1
			end
		end
		
feature -- Array

	from_character_array (a_array: like as_character_array)
			-- Set from character array.
		require
			a_array_attached: a_array /= Void
			count: a_array.count = count
		local
			i: INTEGER
		do
			from
				i := a_array.lower
			until
				i > a_array.upper
			loop
				put (a_array.item (i).code.as_natural_8, i - a_array.lower + 1) 
				i := i + 1
			variant
				a_array.upper - i + 1
			end
		ensure
			items: count > 0 implies across (1 |..| count) as j all item (j.item) = a_array.item (j.item).code end
		end
		
	as_character_array: ARRAY [CHARACTER]
			-- `Current' as array of characters
		local
			i: INTEGER
		do
			create Result.make (1, count)
			from
				i := 1
			until
				i > count
			loop
				Result.put (INTEGER_.to_character (item (i)), i)
				i := i + 1
			variant
				count - i + 1
			end
		ensure
			as_character_array_attached: Result /= Void
			result_count: Result.count = count
			items: Result.count > 0 implies across (1 |..| count) as j all Result.item (j.item).code = item (j.item) end
		end

feature -- Convenience
 
 	sub (a_min, a_max: INTEGER): CRYPT_BYTES_ARRAY
			-- New array consisting of items at indexes
			-- in `a_min .. a_max'
		require
			a_min_large_enough: 1 <= a_min
			a_max_small_enough: a_max <= count
			valid_bounds: a_min <= a_max + 1
		local
			cnt: INTEGER
			i: INTEGER
		do
			cnt := a_max - a_min + 1
			create Result.make (cnt)
			from
				i := 1
			until
				i > cnt
			loop
				Result.put (item (i + a_min - 1), i)
				i := i + 1
			end
		ensure
			subarray_not_void: Result /= Void
			count_set: Result.count = a_max - a_min + 1
			limits_set: Result.count > 0 implies 
					(Result.item (1) = item (a_min) 
					and Result.item (Result.count) = item (a_max))
		end

end
