note

	description:

		"Concurrent linked queue using conditional variables."

	author: "Sami Ben-Grine"
	last_editor: "$Author: c4 $"
	date: "$Date: 2014/05/13 $"
	revision: "$Revision: #1 $"
	path: "$File: //rose/source/library/containers_base/src/concurrent/rose_concurrent_linked_queue.e $"

class	ROSE_CONCURRENT_LINKED_QUEUE [G]

create
	
	make

feature {NONE} -- Initialization

	make
			-- create a new empty queue.
		do
			first := Void
			last := Void
			create monitor.make
			create notempty.make
			create notfull.make
		end

feature -- Basic Operations

	put, extend (a_value: attached G)
			-- Put a value at tail. Signals waiting threads
			-- if queue was empty.
		require
			a_value_attached: a_value /= Void
		local
			l_node: like last
		do
			create l_node.make (a_value)
			monitor.lock
			if is_full then
				from
				until
					not is_full or else is_closed
				loop
					notfull.wait (monitor)
				end
			end
			if not is_closed then
				if is_empty then
					first := l_node
					last := l_node
				else
					check attached last as l_last then
							-- postcondition "definition" of `is_empty' and invariant "first_implies_last" 
						l_last.put_right (l_node)
						last := l_node
					end
				end
				count := count + 1
			end
			check empty_list_check: (first = Void implies last = Void) and (last = Void implies first = Void) end
			notempty.broadcast
			monitor.unlock
		end
	
feature -- Removal
	
	get: detachable G
			-- Get element at first position and return its value.
			-- if the queue is empty, blocks until there is
			-- something in the queue.
			-- Also removes that item from the queue, violating CQS.
		do
			monitor.lock
			if is_empty then
				-- the queue was empty, wait for signal from put
				from
				until
					not is_empty or else is_closed
				loop
					notempty.wait (monitor)
				end
			end
			if not is_closed then
				if attached first as l_first then
					Result := l_first.item
					first := l_first.right
					if first = Void then
						last := Void
					end
					count := count - 1
				end
			end
			check
				empty_list_check: (first = Void implies last = Void) and
					(last = Void implies first = Void)
			end
			notfull.broadcast
			monitor.unlock
		end

	get_with_timeout (a_timeout: INTEGER): TUPLE [valid: BOOLEAN; value: detachable G]
			-- Get element at first position and return its value.
			-- if the queue is empty, blocks at most `a_timeout' ms until there is
			-- something in the queue.
			-- Returns a tuple with field `valid' that indicates whether the Tuple field `value'
			-- contains a valid result.
			-- Also removes that item from the queue, violating CQS.
		local
			l_result: detachable G
			l_condition: BOOLEAN
		do
			monitor.lock
			if is_empty then
				-- the queue was empty, wait for signal from put
				l_condition := notempty.wait_with_timeout (monitor, a_timeout)
			else
				l_condition := True
			end
			create Result
			if l_condition and then (not is_closed and not is_empty) then
				check attached first as l_first then
						-- postcondition "definition" of `is_empty'
					Result.valid := True
					l_result := l_first.item
					first := l_first.right
					if first = Void then
						last := Void
					end
					count := count - 1
				end
			end
			Result.value := l_result
			check
				empty_list_check: (first = Void implies last = Void) and
					(last = Void implies first = Void)
			end
			notfull.broadcast
			monitor.unlock
		end

	get_non_blocking: detachable G
			-- Get the next object from queue.
			-- returns Void if no next object.
			-- Also removes that item from the queue, violating CQS.
		local
			l_default: detachable G
		do
			monitor.lock
			if is_empty then
				-- the queue is empty, return default_value
				Result := l_default
			else
				check attached first as l_first then
						-- postcondition "definition" of `is_empty'
					if not is_closed then
						Result := l_first.item
						first := l_first.right
						if first = Void then
							last := Void
						end
						count := count - 1
					end
				end
			end
			check
				empty_list_check: (first = Void implies last = Void) and
					(last = Void implies first = Void)
			end
			notfull.broadcast
			monitor.unlock
		end

	wipe_out
			-- Remove all items from queue. The events
			-- are discarded.
		do
			monitor.lock
			first := Void
			last := Void
			notfull.broadcast
			monitor.unlock
		end
	
feature -- Setting
	
	set_max_size (size: INTEGER)
			-- sets `max_size'.
		require
			size_valid: size >= 0
		do
			max_size := size
		end
	
feature -- Basic operations
	
	close
			-- set is_closed.
		do
			is_closed := True
		end

	open
			-- sets is_closed.
		do
			is_closed := False
		end


feature -- Status Report

	max_size: INTEGER
			-- the maximum size of the buffer.
			-- if this value is zero, there is no size-limit.
			-- the putter will wait, until some space has
			-- become available.

	count: INTEGER
			-- number of elements contained

	is_empty: BOOLEAN
			-- is the queue empty?
		do
			Result := first = Void
		ensure
			definition: Result = (first = Void)
		end

	is_full: BOOLEAN
			-- is the queue full?
		do
			Result := max_size /= 0 and then count >= max_size
		end

	is_closed: BOOLEAN
			-- is the queue closed?
			-- (suspends all activities)

	notfull, notempty: CONDITION_VARIABLE
			-- Condition for buffer not empty or not full.

feature {NONE} -- Implementation

	first: detachable DS_LINKABLE [attached G]
			-- Head of queue

	last: detachable DS_LINKABLE [attached G]
			-- Tail of queue

	monitor: MUTEX
			-- Mutex on both sides of the queue (head and tail)

invariant

	max_size_non_negative: max_size >= 0
	monitor_attached: monitor /= Void
	notempty_attached: notempty /= Void
	notfull_attached: notfull /= Void
	first_implies_last: (first /= Void) implies (last /= Void)
	
end
