note

	description:

		"A queue implemented with a combination of LINEAR, a dlist and a hash table."

	author: "Mark Howard"
	last_editor: "$Author: c4 $"
	date: "$Date: 2014/05/13 $"
	revision: "$Revision: #1 $"
	path: "$File: //rose/source/library/containers_base/src/queue/rose_hashed_queue.e $"

class ROSE_HASHED_QUEUE [G -> HASHABLE]

inherit

	ROSE_CELL_CONTAINER [G]
		export
			{NONE} cell
		redefine
			has,
			do_all, do_if, there_exists, for_all
		end

create

	make

feature {NONE} -- Initialization

	make
			-- Initializes the queue
		do
			create table.make (0)
		end

feature -- Access

	has (v: like item): BOOLEAN
			-- Does structure include an occurrence of `v'?
		do
			Result := table.has (v)
		end

	oldest: detachable G
			-- Oldest item in the queue
		do
			if attached oldest_element as l_oldest then
				Result := l_oldest.item
			end
		end

	newest: detachable G
			-- Newest item in the queue
		do
			if attached newest_element as l_newest then
				Result := l_newest.item
			end
		end

	index: INTEGER
			-- Location in queue while traversing it (Should this be public?)

	item: G
			-- Current element
		do
			check attached current_cell as l_cell then
					-- precondition "not_off", and postcondition "implementation" of `after'
					-- and postcondition "implementation" of `count' plus the attached status of
					-- items in `table' plus a bit more logic ensure this
				Result := l_cell.item
			end
		end

	count: INTEGER
			-- Number of items in the table
			-- (which should be the same as the count of the dlist.)
		do
			Result := table.count
		ensure
			implementation: Result = (table.count)
		end

	is_empty: BOOLEAN
			-- True if the queue is empty
		do
			Result := newest_element = Void
		ensure then
			definition: Result = (newest_element = Void)
		end

feature -- Element change

	extend (a_item: G)
			-- Add an item to the queue
			-- as well as the table.
		require
			not_has: not has (a_item)
		local
			l_cell: attached like cell
			l_reconfiguring: BOOLEAN
		do
			l_reconfiguring := reconfiguring
			reconfiguring := True
			l_cell := new_cell (a_item)
			table.extend (l_cell, a_item)
			l_cell.set_right (newest_element)
			if attached newest_element as l_newest then
				l_newest.set_left (l_cell)
			end
			set_newest (l_cell)
			if oldest_element = Void then
				oldest_element := l_cell
				start
			end
			reconfiguring := l_reconfiguring
		ensure
			added: has (a_item)
			same_reconfiguring: reconfiguring = old reconfiguring
		end

	renew (a_item: G)
			-- Make `a_item' the newest element in the queue,
			-- if it isn't already.
			-- If it's not present, add it.
		local
			l_potential_new_last: like cell
			l_reconfiguring: BOOLEAN
		do
			if a_item /~ newest then
				if attached cell (a_item) as l_cell then
					l_reconfiguring := reconfiguring
					reconfiguring := True
					l_potential_new_last := l_cell.left
					check attached l_cell.left as l_left then
							-- this routine and `extend' ensure only `newest_cell' has a Void `left'
						l_left.set_right (l_cell.right)
					end
					if attached l_cell.right as l_right then
						l_right.set_left (l_cell.left)
					end
					l_cell.set_right (newest_element)
					if attached newest_element as l_newest then
						l_newest.set_left (l_cell)
					end
					set_newest (l_cell)
					if oldest_element = l_cell then
						if current_cell = oldest_element then
							current_cell := l_potential_new_last
						end
						set_oldest (l_potential_new_last)
					end
					reconfiguring := l_reconfiguring
				else
					extend (a_item)
				end
			end
		ensure
			a_item_present: has (a_item)
			same_reconfiguring: reconfiguring = old reconfiguring
		end
	
feature -- Removal
	
	pop
			-- Remove the oldest (first inserted) item
		local
			l_cell: like cell
			l_was_at_end, l_reconfiguring: BOOLEAN
		do
			l_reconfiguring := reconfiguring
			reconfiguring := True
			if not is_empty then
				l_was_at_end := current_cell = oldest_element
				check attached oldest_element as l_oldest and attached oldest as l_oldest_item then
						-- invariant "older_and_more_recent_defined_or_none" and
						-- postcondition "definition" of `is_empty'
					l_cell := l_oldest.left
					table.remove (l_oldest_item)
					l_oldest.detach
				end
				set_oldest (l_cell)
				if oldest_element = Void then
					newest_element := Void
				end
				if l_was_at_end or after then
					finish
				end
			end
			reconfiguring := l_reconfiguring
		ensure
			one_less_item: (count = old count - 1) or (old count = 0 and count = old count)
			same_reconfiguring: reconfiguring = old reconfiguring
		end

feature -- Status report

	after: BOOLEAN
			-- Off the end
		do
			Result := index > count
		ensure then
			implementation: Result = (index > count)
		end

feature -- Cursor movement

	start
			-- The very beginning (ie the newest entry)
		do
			current_cell := newest_element
			index := 1
		end

	forth
			-- Progress forward
		do
			check attached current_cell as l_current then
					-- precondition "not_after", and postcondition "implementation" of `after'
					-- and postcondition "implementation" of `count' plus the attached status of
					-- items in `table' plus a bit more logic ensure this
				current_cell := l_current.right
			end
			index := index + 1
		end

	finish
			-- Move to the oldest_element entry
		do
			current_cell := oldest_element
			index := count
		end

feature -- Iteration

	do_all (a_action: PROCEDURE [ANY, TUPLE [G]])
			-- <Precursor>
		local
			l_index: INTEGER
		do
			l_index := index
			Precursor (a_action)
			go_i_th (l_index)
		ensure then
			unchanged_position: index = old index
		end

	do_if (a_action: PROCEDURE [ANY, TUPLE [G]]; a_test: FUNCTION [ANY, TUPLE [G], BOOLEAN])
			-- <Precursor>
		local
			l_index: INTEGER
		do
			l_index := index
			Precursor (a_action, a_test)
			go_i_th (l_index)
		ensure then
			unchanged_position: index = old index
		end

	there_exists (a_test: FUNCTION [ANY, TUPLE [G], BOOLEAN]): BOOLEAN
			-- <Precursor>
		local
			l_index: INTEGER
		do
			l_index := index
			Result := Precursor (a_test)
			go_i_th (l_index)
		ensure then
			unchanged_position: index = old index
		end

	for_all (a_test: FUNCTION [ANY, TUPLE [G], BOOLEAN]): BOOLEAN
			-- <Precursor>
		local
			l_index: INTEGER
		do
			l_index := index
			Result := Precursor (a_test)
			go_i_th (l_index)
		ensure then
			unchanged_position: index = old index
		end

feature {NONE} -- Implementation

	current_cell: like cell
			-- Cell containing current element

	table: HASH_TABLE [attached like cell, G]
			-- Items of the hashed queue

	newest_element: like cell
			-- Newest added item

	oldest_element: like cell
			-- Oldest added item

	new_cell (a_item: G): attached like cell
			-- New cell containing `a_item'
		do
			create Result.make (Current, a_item)
		ensure
			new_cell_not_void: Result /= Void
			correct_container: Result.container = Current
			correct_item: Result.item = a_item
			detached: Result.left = Void and Result.right = Void
		end

	set_newest (a_cell: like cell)
			-- Set the newest element to `a_cell'
		do
			newest_element := a_cell
			if attached newest_element as l_newest then
				l_newest.set_left (Void)
			end
		ensure
			newest_aliased: newest_element = a_cell
		end

	set_oldest (a_cell: like cell)
			-- Set the oldest element to `a_cell'
		do
			oldest_element := a_cell
			if attached oldest_element as l_oldest then
				l_oldest.set_right (Void)
			end
		ensure
			oldest_aliased: oldest_element = a_cell
		end

	cell (a_item: G): detachable ROSE_LINK_CELL [G]
			-- The hash table provides quick lookup of random elements
			-- within the queue
		do
			Result := table.item (a_item)
		end

	go_i_th (a_index: INTEGER)
			-- Move to location `a_index' in the queue
		require
			a_index_large_enough: a_index >= 0
			a_index_small_enough: a_index <= connected_count
		do
			from
				start
			until
				index >= a_index
			loop
				forth
			end
		ensure
			correct_index: index = a_index
		end

	reconfiguring: BOOLEAN
			-- Flag to indicate that we are in the process of reordering things.
			-- So certain contracts do not hold

	connected_count: INTEGER
			-- Number of items in the dlist
		local
			l_item: like cell
		do
			from
				l_item := newest_element
			until
				l_item = Void
			loop
				Result := Result + 1
				check attached l_item as l_item_2 then
						-- until
					l_item := l_item_2.right
				end
			end
		ensure
			positive_connected_count: Result >= 0
			unchanged_position: index = old index
		end

invariant

	no_orphans: is_empty = (count = 0)
	table_not_void: table /= Void
	no_older_more_recent_for_empty_queue: is_empty = (newest_element = Void and oldest_element = Void)
	older_and_more_recent_defined_or_none: (newest_element = Void) = (oldest_element = Void)
	first_connected: (not reconfiguring) implies ((attached newest_element as l_newest and then l_newest.right = Void) implies count = 1)
	last_connected: (not reconfiguring) implies ((attached oldest_element as l_oldest and then l_oldest.left = Void) implies count = 1)
	ends_are_ends: (not is_empty) implies (
		(attached newest_element as l_newest and then l_newest.left = Void) and
		(attached oldest_element as l_oldest and then l_oldest.right = Void))
	all_connected: (not reconfiguring) implies connected_count = count
	
end
