note

	description:

		"Array which is extendable and linear"

	keywords: "array,linear,extendable,iterable"
	author: "Mark Howard"
	last_editor: "$Author: c4 $"
	date: "$Date: 2014/05/13 $"
	revision: "$Revision: #1 $"
	path: "$File: //rose/source/library/containers_core/src/array/extendable_bounded_array.e $"

class	EXTENDABLE_BOUNDED_ARRAY [G]

inherit
	
	RP_SERIALIZABLE

	ROSE_LINEAR [G]
		rename
			item as current_item
		undefine
			copy, off, is_equal, has, occurrences,
			do_all, do_if, for_all, there_exists
		end

	ARRAY [G]
		rename
			subarray as array_subarray,
			make_from_array as array_make_from_array
		undefine
			linear_representation, prunable, prune, extendible
		redefine
			make, make_empty, extend, prune_all, full, wipe_out, has, occurrences,
			is_inserted
		end

create

	make,
	make_empty,
	make_from_array,
	make_from_linear,
	make_from_array_reversed,
	make_with_default_value

feature -- Initialization

	make (a_lower, a_upper: INTEGER)
			-- Allocate an array with `a_lower' and `a_upper' bounds.
		do
			make_filled (({G}).default, a_lower, a_upper)
			index := a_lower
			compare_objects
		end

	make_empty
			-- Allocate an empty array.
		do
			make (1, 0)
		end

	make_from_array (a_array: ARRAY[like item])
			-- Allocate an array initially populated with all elements from `a_array'.
		require
			array_exists: a_array /= Void
		local
			i: INTEGER
		do
			make (1,a_array.count)
			from
				i := a_array.lower
			until
				i > a_array.upper
			loop
				put (a_array.item (i),i-a_array.lower + 1)
				i := i + 1
			end
		ensure
			same_size: count = a_array.count
		end

	make_with_default_value (a_lower, a_upper: INTEGER; a_default_value: like item)
			-- Make filled with 'a_default_value'.
		local
			i: INTEGER
		do
			make (a_lower, a_upper)
			fill_with (a_default_value)
		end

feature {NONE} -- Initialization

	make_from_linear (a_linear: LINEAR [like item])
			-- Allocate an array initially populated with all elements from `a_linear'.
		require
			a_linear_attached: a_linear /= Void
		do
			make_empty
			a_linear.do_all (agent extend)
		end

	make_from_array_reversed (a_array: ARRAY [like item])
			-- Allocate an array initially populated with all elements from `a_array' in reverse order.
		require
			a_array_attached: a_array /= Void
		local
			i: INTEGER
		do
			make (1, a_array.count)
			from
				i := a_array.upper
			until
				i < a_array.lower
			loop
				put (a_array.item (i), a_array.upper - i + 1)
				i := i - 1
			end
		end

feature -- Access

	safe_item (i: INTEGER): detachable like item
			-- `i'-th item if valid index `i', Void (or default value) otherwise
		do
			if valid_index (i) then
				Result := item (i)
			end
		end

	current_item: like item
			-- <Precursor>
		do
			Result := area.item (index - lower)
		end

	previous_item: detachable like item
			-- Item just previous to current item
		require
			index_greater_than_lower: index > lower
		do
			Result := safe_item (index - 1)
		end

	next_item: detachable like item
			-- Item just after current item
		require
			index_less_than_upper: index < upper
		do
			Result := item (index + 1)
		end

	first: detachable like item
			-- Item at first position, if any
		do
			Result := safe_item (lower)
		end

	last: detachable like item
			-- Item at last position, if any
		do
			if count > 0 then
				Result := item (upper)
			end
		end

	index: INTEGER
			-- Current index of `item', may be invalid

	found_index: INTEGER
			-- Index of item eventually found by last call to `has' (this is a side effect)

	new_index_of (a_item: like item): INTEGER
			-- Index of first occurence of `a_item' in this array
		local
			i: INTEGER
		do
			Result := lower - 1
			from
				i := lower
			until
				Result >= lower or else i > upper
			loop
				if a_item ~ item (i) then
					Result := i
				end
				i := i + 1
			end
		ensure
			new_index_of_at_most_upper: Result <= upper
			new_index_at_least_lower_minus_one: Result >= lower - 1
		end


feature -- Status report

	prunable: BOOLEAN
			-- May items be removed? (Answer: yes.)
		do
			Result := True
		end

	extendible: BOOLEAN
			-- May items be added?
		do
			Result := True
		end

	full: BOOLEAN
			-- Is structure filled to capacity?
		do
			Result := (count = capacity)
		end

	at_first: BOOLEAN
			-- Is the cursor at the first item?
		do
			Result := index = lower
		end

	at_last: BOOLEAN
			-- Is the cursor at the last item?
		do
			Result := index = upper
		end

	off: BOOLEAN
			-- Is there no current item?
		do
			Result := before or after
		end

	after: BOOLEAN
			-- Is there no valid position to the right of current one?
		do
			Result := index > upper
		end

	before: BOOLEAN
			-- Is there no valid position to the left of current one?
		do
			Result := index < lower
		end
	is_inserted(a_value : like item) : BOOLEAN
			-- Faster version of is_inserted for postconditions
		do
			if not is_empty then
				Result := a_value = last or else Precursor(a_value)
			end
		end

	has (a_value: like item): BOOLEAN
			-- Does `a_value' appear in this array?
 			-- (Reference or object equality, based on `object_comparison'.)
			-- Sets `found_index'.
		local
			n, l_found: INTEGER
		do
			n := upper
			if object_comparison then
				from
					l_found := lower
				until
					l_found > n or else (a_value ~ item (l_found))
				loop
					l_found := l_found + 1
				end
			else
				from
					l_found := lower
				until
					l_found > n or else (item (l_found) = a_value)
				loop
					l_found := l_found + 1
				end
			end
			Result := not (l_found > n)
			found_index := l_found
		end

	occurrences (v: like item): INTEGER
			-- Number of times `v' appears in structure
		local
			i: INTEGER
		do
			if object_comparison then
				from
					i := lower
				until
					i > upper
				loop
					if v ~ item (i) then
						Result := Result + 1
					end
					i := i + 1
				end
			else
				from
					i := lower
				until
					i > upper
				loop
					if item (i) = v then
						Result := Result + 1
					end
					i := i + 1
				end
			end
		end

feature -- Cursor movement

	move (i: INTEGER)
			-- Move cursor `i' positions, does not go outside array bounds.
		do
			index := index + i
			if index > upper then
				index := upper
			elseif index < lower then
				index := lower
			end
		ensure
			within_bounds: not is_empty implies valid_index (index)
		end

	start
			-- Move cursor to first position if any.
		do
			index := lower
		ensure then
			after_when_empty: is_empty implies after
		end

	finish
			-- Move cursor to last position if any.
		do
			index := upper
		ensure then
			before_when_empty: is_empty implies before
		end

	forth
			-- Move cursor one position forward.
		do
			index := index + 1
		end

	back
			-- Move cursor one position backward.
		do
			index := index - 1
		end

	go_i_th (i: INTEGER)
			-- Move cursor to `i'-th position.
		do
			index := i
		end

	start_forth
			-- Move cursor to first position if any, then if not off, move one position forth.
		do
			start
			if not off then forth end
		end

	finish_back
			-- Move cursor to last position if any, then if not off, move one position back.
		do
			finish
			if not off then back end
		end

	swap (i: INTEGER)
			-- Exchange item at `i'-th position with item
			-- at `index' position.
		require
			i_valid: valid_index (i)
			not_off: not off
		local
			l_item: like item
		do
			l_item := current_item
			replace (item (i))
			put (l_item, i)
		end

	put_front (a_value: like item)
			-- Add `a_value' to the beginning.
			-- Do not move cursor.
		do
			if is_empty then
				extend (a_value)
			else
				insert (a_value,lower)
			end
		end

	extend (a_value: like item)
			-- Add `a_value' to end.
			-- Do not move cursor.
		do
			force (a_value, upper + 1)
		end

	put_left (a_value: like item)
			-- Add `a_value' to the left of current position.
			-- Do not move cursor.
		do
			if after or is_empty then
				extend (a_value)
				index := index + 1
			else
				insert (a_value, index)
			end
		end

	put_right (a_value: like item)
			-- Add `a_value' to the right of current position.
			-- Do not move cursor.
		do
			if index = count then
				extend (a_value)
			else
				insert (a_value, index + 1)
			end
		end

	replace (a_value: like item)
			-- Replace current item by `a_value'.
		require
			not_off: not off
		do
			put (a_value, index)
		end

	append_linear (a_linear: LINEAR [like item])
			-- Append contents of 'a_linear' to this array.
		require
			a_linear_not_void: a_linear /= Void
		local
			n: INTEGER
		do
			if a_linear = Current then
				n := upper
				do_if (agent extend_ignoring_index, agent index_not_greater (?, ?, n))
			else
				a_linear.do_all (agent extend)
			end
		end

	append_array (a_array: ARRAY [like item])
			-- Append contents of 'a_array' to this array.
		local
			i: INTEGER
		do
			if a_array /= Void then
				a_array.do_all (agent extend)
			end
		end

	apply (a_agent: PROCEDURE[ANY,TUPLE[like item]])
			--version of do_all which does not affect the state, so safe for recursion
		local
			 i: INTEGER
			 t: TUPLE [like item]
		do
			create t
			from i := lower until i > upper loop
				t.put (item (i),1)
				a_agent.call (t)
				i := i + 1
			end
		end

	restrict (a_lower, a_upper: INTEGER)
			-- Restrict bounds of this array to `a_lower' and `a_upper'
		do
			conservative_resize (a_lower,a_upper)
			lower := a_lower
			upper := a_upper
		end

feature -- Transformations

	permute_randomly
			-- Randomly permutes elements of current array.
		local
			l_time_seed: ROSE_DATE
		do
			create l_time_seed.make_now
			permute_randomly_with_seed (l_time_seed.time.hhmmssfff)
		ensure
			same_count: count = old count
			same_lower: lower = old lower
			same_upper: upper = old upper
		end

	permute_randomly_with_seed (a_seed: INTEGER)
			-- Randomly permutes elements of current array.  Caller can set `a_seed' for replications of random permutations.
			-- Uses in-place "Knuth shuffle".
		require
			non_negative_seed: a_seed >= 0
		local
			l_element: G
			l_index, l_random_index: INTEGER
			l_random: RANGED_RANDOM
		do
			create l_random.make
			l_random.set_seed (a_seed)
			from l_index := upper
			until l_index = lower
			loop
				l_random_index := l_random.next_item_in_range (lower, l_index)
				l_element := Current[l_index]
				Current[l_index] := Current[l_random_index]
				Current[l_random_index] := l_element
				l_index := l_index - 1
			end
		ensure
			same_count: count = old count
			same_lower: lower = old lower
			same_upper: upper = old upper
		end

feature -- Removal

	prune (a_value: like item)
			-- Remove first occurrence of `a_value', if any,
			-- after cursor position.
			-- Move cursor to right neighbor
			-- (or `after' if no right neighbor or `a_value' does not occur)
		do
			if before then
				index := lower
			end
			if object_comparison then
				from
				until
					after or else (a_value ~ current_item)
				loop
					forth
				end
			else
				from
				until
					after or else current_item = a_value
				loop
					forth
				end
			end
			if not after then remove end
		end

	remove
			-- Remove current item.
			-- Move cursor to right neighbor
			-- (or `after' if no right neighbor)
		local
			i, j: INTEGER
			l_default: detachable G
		do
			if not off then
				from
					i := index
				until
					i > upper
				loop
					j := i + 1
					if j > upper then
						if ({G}).has_default then
							put (({G}).default, i)
						else
							check attached l_default as l_default_2 then
									-- since it is not expanded or detachable, it must be self-initializing
								put (l_default_2, i)
							end
						end
					else
						put (item (j), i)
					end
					i := j
				end
				if ({G}).has_default then
					put (({G}).default, upper)
				else
					check attached l_default as l_default_2 then
							-- since it is not expanded or detachable, it must be self-initializing
						put (l_default_2, upper)
					end
				end	
				upper := upper - 1
			end
		end

	prune_all (a_value: like item)
			-- Remove all occurrences of `a_value'.
			-- (Reference or object equality,
			-- based on `object_comparison'.)
			-- Leave cursor `after'.
		local
			i: INTEGER
			l_item: like item
			l_default: detachable G
		do
			if object_comparison then
				from
					start
				until
					after or else a_value ~ current_item
				loop
					forth
				end
				from
					if not after then
						i := index
						forth
					end
				until
					after
				loop
					l_item := current_item
					if l_item /~ a_value then
						put (l_item, i)
						i := i + 1
					end
					forth
				end
			else
				from
					start
				until
					after or else (current_item = a_value)
				loop
					forth
				end
				from
					if not after then
						i := index
						forth
					end
				until
					after
				loop
					l_item := current_item
					if l_item /= a_value then
						put (l_item, i)
						i := i + 1
					end
					forth
				end
			end
			if i >= lower then
				from
					index := i
				until
					i > upper
				loop
					if ({G}).has_default then
						put (({G}).default, i)
					else
						check attached l_default as l_default_2 then
								-- since it is not expanded or detachable, it must be self-initializing
							put (l_default_2, i)
						end
					end
					i := i + 1
				end
				upper := index - 1
			end
		ensure then
			is_after: after
		end

	remove_left
			-- Remove item to the left of cursor position.
			-- Do not move cursor.
		do
			back
			remove
		end

	remove_right
			-- Remove item to the right of cursor position
			-- Do not move cursor
		do
			forth
			remove
			back
		end

	empty_out,wipe_out
			-- Remove all items.
		do
			discard_items
			index := lower
			upper := lower - 1
		end

	insert (a_value: like item; pos: INTEGER)
			-- Add `a_value' at `pos', moving subsequent items
			-- to the right.
		require
			index_small_enough: pos <= upper
			index_large_enough: pos >= lower
		local
			i,j: INTEGER
		do
			if index >= pos then
				index := index + 1
			end
			if attached last as l_last then
				force (l_last, upper + 1)
			end
			from
				i := upper - 1
			until
				i <= pos
			loop
				j := i - 1
				put (item (j), i)
				i := j
			end
			put (a_value, pos)
		ensure
			new_count: count = old count + 1
			insertion_done: item (pos) = a_value
		end

	valid_cursor (a_cursor : like cursor) : BOOLEAN
			--is a_cursor valid?
		do
			Result := a_cursor.index >= lower - 1 and then
			a_cursor.index <= upper + 1
		end
	
	extend_ignoring_index (a_value: G; a_index: INTEGER)
			-- Append `a_value' to `Current'.
			-- `a_index' is required by interface.
		do
			extend (a_value)
		ensure
			item_inserted: is_inserted (a_value)
		end
		
	index_not_greater (a_dummy: G; a_index, a_limit: INTEGER): BOOLEAN
			-- Is `a_index' <= `a_limit'?
			-- `a_dummy' is required by interface.
		do
			Result := a_index <= a_limit
		ensure
			definition: Result = (a_index <= a_limit)
		end
			
end
