note

	description:

		"Single bit representation of characteristic function"

	author: "Mark Howard"
	last_editor: "$Author: c4 $"
	date: "$Date: 2014/05/14 $"
	revision: "$Revision: #2 $"
	path: "$File: //rose/source/library/containers_core/src/bitmap/rose_bitmap_base.e $"

class
	ROSE_BITMAP_BASE

inherit
	RP_SERIALIZABLE
	ARRAY [INTEGER_8]
		rename
			make as array_make,
			make_empty as array_make_empty,
			put as array_put,
			item as array_item,
			count as array_count
		redefine
			clear_all
		end

create
	make,
	make_empty

feature {NONE} --Initialization

	make (a_count: INTEGER)
			-- Create a bitmap with all bits True
		require
			valid_count: a_count > 0
		do
			count := a_count
			make_filled (constant_255, 1, byte_count)
			cardinality := count
		end

	make_empty (a_count: INTEGER)
			-- Create a bitmap with all bits False
		require
			valid_count: a_count > 0
		do
			count := a_count
			make_filled (0, 1, byte_count)
		end

feature -- Element modification

	put, set (b: BOOLEAN; i: INTEGER)
			-- Set item at `i' to the value `b'
		require
			valid_index: i > 0 and i <= count
		local
			l_cell: INTEGER
			l_offset: INTEGER
			l_item: INTEGER_8
		do
			l_cell := cell (i)
			l_offset := offset (i)
			l_item := array_item (l_cell)
			if b then
				if not l_item.bit_test (l_offset) then
					l_item := l_item.set_bit (b, l_offset)
					array_put (l_item, l_cell)
					cardinality := cardinality + 1
				end
			else
				if l_item.bit_test (l_offset) then
					l_item := l_item.set_bit (b, l_offset)
					array_put (l_item, l_cell)
					cardinality := cardinality - 1
				end
			end
		ensure
			properly_set: get (i) = b
		end

	set_from_linear (a_indices: LINEAR [INTEGER])
			-- Set with 'a_indices'
		require
			indices_not_void: a_indices /= Void
--			indices_large_enough: a_indices.for_all (agent (0).is_less)
--			indices_small_enough: a_indices.for_all (agent count.is_greater_equal)
-- not supported by compiler yet
		do
			clear_all
			from
				a_indices.start
			until
				a_indices.off
			loop
				set (True, a_indices.item)
				a_indices.forth
			end
		ensure
			all_set: a_indices.for_all (agent get)
		end

feature -- Access

	item, get (i: INTEGER): BOOLEAN
			-- Value of item at index `i'
		do
			Result := i >= 1 and then i <= count and then array_item (cell (i)).bit_test (offset (i))
		end

	count_from_start_to_end(a_start, a_end: INTEGER): INTEGER
			-- Efficient count of "on" bits from `a_start' to `a_end'
	local
		i: INTEGER
		l_cell_pos: INTEGER
		l_last_cell_pos: INTEGER
		l_cell_data: INTEGER_8
		l_last_i: INTEGER
		do
			if a_start <= 1 and a_end >= count then
				Result := cardinality
			else
				-- Can't start before 1 or end after 'count'.
				i := a_start.max(1)
				l_last_i := a_end.min(count)
				
				-- Count the "on" items in the first cell.
				from
					l_cell_pos := cell(i)
				until
					i > l_last_i or else cell(i) /= l_cell_pos
				loop
					if item(i)
					 then
						Result := Result + 1
					end
					
					i := i + 1
				end
				
				-- Count the "on" items in the cells between first and last.
				from
					l_cell_pos := cell(i)
					l_last_cell_pos := cell(l_last_i)
				until
					l_cell_pos >= l_last_cell_pos
				loop
					l_cell_data := array_item(l_cell_pos)
					if l_cell_data /= 0
					 then
						if l_cell_data.bit_test(0) then Result := Result + 1 end
						if l_cell_data.bit_test(1) then Result := Result + 1 end
						if l_cell_data.bit_test(2) then Result := Result + 1 end
						if l_cell_data.bit_test(3) then Result := Result + 1 end
						if l_cell_data.bit_test(4) then Result := Result + 1 end
						if l_cell_data.bit_test(5) then Result := Result + 1 end
						if l_cell_data.bit_test(6) then Result := Result + 1 end
						if l_cell_data.bit_test(7) then Result := Result + 1 end
					end
					
					l_cell_pos := l_cell_pos + 1
					i := i + 8		-- This won't be used until the next loop.
				end
				
				-- Count the "on" items in the last cell.
				from
					-- i is already setup.
				until
					i > l_last_i
				loop
					if item(i)
					 then
						Result := Result + 1
					end
					
					i := i + 1
				end
			end
		end

	ith_set_after (a_start, a_count: INTEGER): INTEGER
			-- Index of the a_count_th set bit after a_start (not including a_start);
			-- Returns -1 otherwise.
		require
			valid_indexes: a_start >= 0 and a_start + a_count <= count
			at_least_one_count: a_count /= 0
		local
			i, l_count: INTEGER
		do
			from
				i := a_start
			until
				i >= count or else l_count = a_count
			loop
				i := i + 1
				if get (i) then
					l_count := l_count + 1
				end
			end
			if l_count = a_count then
				Result := i
			else
				Result := -1
			end
		end

	ith_set_before (a_start, a_count: INTEGER): INTEGER
			--returns the index of the a_count_th set bit before a_start (not including a_start)
		require
			valid_indexes: a_start > a_count and a_start <= count + 1
			at_least_one_count: a_count /= 0
		local
			i, l_count: INTEGER
		do
			from
				i := a_start
			until
				i <= 1 or else l_count = a_count
			loop
				i := i - 1
				if get (i) then
					l_count := l_count + 1
				end
			end
			if l_count = a_count then
				Result := i
			else
				Result := -1
			end
		end

	cardinality: INTEGER
			-- Number of items set to True

	count: INTEGER
			-- Size of the bitmap, ie how many items can we store

feature -- Basic operations

	set_all
			-- Set all items to True
		do
			fill_with (constant_255)
			cardinality := count
		end

	clear_all
			-- Set all items to False
		do
			Precursor
			cardinality := 0
		end

	complement
			-- Puts the complement of each item
		local
			l_item: INTEGER_8
			i: INTEGER
		do
			from
				i := 1
			until
				i > array_count
			loop
				l_item := array_item (i).bit_not
				array_put (l_item, i)
				i := i + 1
			end
			cardinality := count - cardinality
		end

	or_with (a_bitmap: ROSE_BITMAP_BASE)
			-- Adds `a_bitmap' to Current with `or' on each item
		require
			other_not_void: a_bitmap /= Void
		local
			l_count, i, j: INTEGER
			l_other: INTEGER_8
		do
			l_count := array_count.min (a_bitmap.array_count)
			from
				j := 1
			until
				j > l_count - 1
			loop
				l_other := a_bitmap.array_item (j)
				if l_other /= 0 and then array_item (j) /= l_other then
					array_put (array_item (j) | l_other, j)
				end
				j := j + 1
			end
			l_count := count.min (a_bitmap.count)
			from
				i := (j - 1) * 8 + 1
			until
				i > l_count
			loop
				set (get (i) or a_bitmap.get (i), i)
				i := i + 1
			end
			set_cardinality
		end

	or_with_smaller (a_bitmap: ROSE_BITMAP_BASE)
			-- Like 'or_with' except that the result only goes as high as the smaller of the two bitmaps.
			-- This will be used in ZONE:market_dates to 'or' together two calendars in a way that doesn't
			-- allow one calendar to dominate the total length of the result.
		require
			other_not_void: a_bitmap /= Void
		local
			l_hi_count, l_lo_count, i: INTEGER
			l_last_1, l_last_2: INTEGER
		do
			l_last_1 := ith_set_before (count + 1, 1)
			l_last_2 := a_bitmap.ith_set_before (a_bitmap.count + 1, 1)
			if l_last_1 = 1 then
					-- Current bitmap is empty.  Use a fake 'l_last_1'.
				l_last_1 := l_last_2
			end
			l_lo_count := l_last_1.min (l_last_2)
			l_hi_count := l_last_1.max (l_last_2)
			from
				i := l_lo_count + 1
			until
				i > l_hi_count
			loop
				set (False, i)
				i := i + 1
			end
			from
				i := 1
			until
				i > l_lo_count
			loop
				set (get (i) or a_bitmap.get (i), i)
				i := i + 1
			end
		end

	and_with (a_bitmap: ROSE_BITMAP_BASE)
			-- Adds `a_bitmap' to Current with `and' on each item
		require
			a_bitmap_not_void: a_bitmap /= Void
		local
			l_count, i: INTEGER
		do
			l_count := count.min (a_bitmap.count)
			from
				i := 1
			until
				i > l_count
			loop
				set (get (i) and a_bitmap.get (i), i)
				i := i + 1
			end
		end

feature {NONE} --Implementation

	set_cardinality
			--compute cardinality (total number of bits set)
		local
			i, j: INTEGER
			l_byte: INTEGER_8
			l_count: INTEGER
		do
			cardinality := 0
			from
				j := 1
				l_count := byte_count
			until
				j > l_count -1
			loop
				l_byte := array_item (j)
				if l_byte /= 0 then
					from
						i := 0
					until
						i > 7
					loop
						if l_byte.bit_test (i) then
							cardinality := cardinality + 1
						end
						i := i + 1
					end
				end
				j := j + 1
			end
			l_count := count
			l_byte := array_item (j)
			if l_byte /= 0 then
				from
					i := (j - 1) * 8 + 1
				until
					i > l_count
				loop
					if l_byte.bit_test (offset (i)) then
						cardinality := cardinality + 1
					end
					i := i + 1
				end
			end
		end

	cell (i: INTEGER): INTEGER
			--Cell for i
		do
			Result := ((i - 1) // 8) + 1
		end

	offset (i: INTEGER): INTEGER
			--Offset in cell
		do
			Result := i \\ 8
		end

	byte (i: INTEGER): INTEGER_8
			--byte containing bit
		do
			Result := array_item (cell (i))
		end

	byte_count: INTEGER
			-- Returns the true size in bytes rather than in bits.
		do
			Result := (count + 7) // 8
		end

feature {NONE} -- Constants

	constant_255: INTEGER_8 = 0xFF

end
