note

	description:

		"[
        Hash table with facilities for iterating, merging and
         allows to use TUPLEs as keys correctly
       ]"

	storable_version: ""
	author: "Michael Gacsaly"
	last_editor: "$Author: c4 $"
	date: "$Date: 2014/05/13 $"
	revision: "$Revision: #1 $"
	path: "$File: //rose/source/library/containers_core/src/table/rose_hash_table.e $"

class	ROSE_HASH_TABLE [G, H -> detachable HASHABLE]

inherit
	
	RP_SERIALIZABLE

	HASH_TABLE [G, H]
		redefine
			has, item, put, force, extend, replace_key, remove
		end

create

	make

feature -- Access

	has (key: H): BOOLEAN
			-- Is there an item in the table with key `key'?
		do
			fix_key (key)
			Result := Precursor (key)
		end

	item alias "[]" (key: H): G assign force
			-- Item associated with `key', if present
			-- otherwise default value of type `G'
		do
			fix_key (key)
			if attached Precursor (key) as l_result then
				Result := l_result
			else
				check
					not_self_initializing: ({G}).has_default
						-- as it failed the if-clause
				end
				Result := ({G}).default
			end
		end

	union (a_other: HASH_TABLE [G, H]; a_combiner: FUNCTION [ANY, TUPLE [G, G], G]): like Current
			-- Union of `a_other' and `Current' Using `a_combiner' to combine items present in both.
		require
			other_ok: a_other /= Void
			combiner_ok: a_combiner /= Void
		do
			create Result.make (0)
			Result.combine (Current, a_combiner)
			Result.combine (a_other, a_combiner)
		end

	current_items: ARRAY [G]
			-- New array containing actually used items, from 1 to `count'
		local
			j: INTEGER
			l_old_iteration_position: INTEGER
		do
			l_old_iteration_position := iteration_position
			from
				create Result.make (1, count)
				start
			until
				off
			loop
				j := j + 1
				Result.put (item_for_iteration, j)
				forth
			end
			iteration_position := l_old_iteration_position
		ensure
			current_items_attached: Result /= Void
			good_count: Result.count = count
 		end

feature -- Element change

	put (new: G; key: H)
			-- Insert `new' with `key' if there is no other item
			-- associated with the same key.
		do
			fix_key (key)
			Precursor (new, key)
		end

	force (new: G; key: H)
			-- Update table so that `new' will be the item associated
			-- with `key'.
		do
			fix_key (key)
			Precursor (new, key)
		end

	extend (new: G; key: H)
			-- Assuming there is no item of key `key',
			-- insert `new' with `key'.
		do
			fix_key (key)
			Precursor (new, key)
		end

	replace_key (new_key: H; old_key: H)
			-- If there is an item of key `old_key' and no item of key
			-- `new_key', replace the former's key by `new_key',
		do
			fix_key (new_key)
			fix_key (old_key)
			Precursor (new_key, old_key)
		end
	
feature -- Removal
	
	remove (key: H)
			-- Remove item associated with `key', if present.
		do
			fix_key (key)
			Precursor (key)
		end
	
feature -- Iteration
	
	do_all (a_agent: PROCEDURE [ANY, TUPLE [G]])
			-- Call `a_agent' on all items of this hash table.
		require
			agent_ok: a_agent /= Void
		do
			from
				start
			until
				off
			loop
				a_agent.call ([item_for_iteration])
				forth
			end
		end
	
feature -- Basic operations
	
	combine (a_other: HASH_TABLE [G, H]; a_combiner: FUNCTION [ANY, TUPLE [G, G], G])
			-- Combine items form `a_other' to `Current' using `a_combiner'
		require
			other_ok: a_other /= Void
			combiner_ok: a_combiner /= Void
		local
			l_key: H
			l_hash: like Current
		do
			from
				l_hash := twin
				l_hash.start
			until
				l_hash.off
			loop
				l_key := l_hash.key_for_iteration
				if a_other.has (l_key) and then attached a_other.item (l_key) as l_item then
					force (a_combiner.item ([l_hash.item_for_iteration, l_item]), l_key)
				elseif ({G}).has_default then
					force (a_combiner.item ([l_hash.item_for_iteration, ({G}).default]), l_key)
				end
				l_hash.forth
			end
			from
				a_other.start
			until
				a_other.off
			loop
				l_key := a_other.key_for_iteration
				if not has (l_key) then
					if ({G}).has_default then
						force (a_combiner.item ([({G}).default, a_other.item_for_iteration]), l_key)
					end
				end
				a_other.forth
			end
		end

feature {NONE} -- Implementation

	fix_key (a_key: H)
			-- Fix `a_key' if it is a TUPLE
		do
			if attached {TUPLE} a_key as l_tuple and then not l_tuple.object_comparison then
				l_tuple.compare_objects
			end
		end

end
