note

	description:

		"Looks like a ROSE_LINEAR_ARRAY using either a proxy or a single element."

	author: "Randy John"
	last_editor: "$Author: c4 $"
	date: "$Date: 2014/08/06 $"
	revision: "$Revision: #7 $"
	path: "$File: //rose/source/library/containers_core/src/array/rose_linear_array_or_one.e $"

class	ROSE_LINEAR_ARRAY_OR_ONE [G]

--
--	Use this class when you need a lot of single-element arrays.  This class
--	will only build an actual array if the number of elements exceeds one.
--
--	To modify:  If a feature is missing, look up the corresponding feature in
--	ROSE_LINEAR_ARRAY.  You can steal the implementation.  If the feature
--	actually comes from ROSE_LINEAR then make then change in ROSE_LINEAR_OR_ONE.
--
--	Note that with only a couple of changes, you can turn this class back into
--	a ROSE_LINEAR_ARRAY.  Just remove the inheritance of ROSE_LINEAR_OR_ONE and
--	uncomment the inheritance of ROSE_LINEAR_ARRAY.  Also, comment out ALL of
--	the features.
--

inherit

	ROSE_LINEAR_OR_ONE [G]
		rename
			linear_ as array_,
			item as current_item
		redefine
			array_,
			off,
			do_all
		end

create

	make,
	make_empty,
	make_from_array,
	make_from_linear

feature {NONE} -- Initialization

	make (a_lower, a_upper: INTEGER)
			-- Allocate array with `a_lower' and `a_upper' bounds.
		do
			if
				a_upper > a_lower
			then
				create array_.make (a_lower, a_upper)
			else
				index_ := 1
				lower_ := a_lower
				if a_upper = a_lower then
					has_object_ := True
				end
			end
		end

	make_empty
			-- Allocate empty array.
		do
			make (1,0)
		end

	make_from_array (a_array: ARRAY[G])
			-- Allocate array with all elements from `a_array'.
		local
			i: INTEGER
		do
			make (1, a_array.count)
			from
				i := a_array.lower
			until
				i > a_array.upper
			loop
				put (a_array.item (i), i - a_array.lower + 1)
				i := i + 1
			end
		end

	make_from_linear (a_linear: LINEAR [G])
			-- Allocate array with all elements from `a_linear'.
		do
			make_empty
			from
				a_linear.start
			until
				a_linear.off
			loop
				extend (a_linear.item)
				a_linear.forth
			end
		end

feature -- Access

	index: INTEGER
			-- Index of `item'
		do
			if attached array_ as l_array then
				Result := l_array.index
			else
				Result := index_
			end
		end

feature -- Status report

	after: BOOLEAN
			-- Is there no valid position to the right of current one?
		do
			if attached array_ as l_array then
				Result := l_array.after
			else
				Result := index_ > upper
			end
		end

	before: BOOLEAN
			-- Are we before the first item?
		do
			if attached array_ as l_array then
				Result := l_array.before
			else
				Result := index_<upper
			end
		end

feature -- Traversal

	forth
		-- Move cursor one position forward.
	do
		if attached array_ as l_array then
			l_array.forth
		else
			index_ := index_ + 1
		end
	end

	back
		-- Move cursor one position forward.
	do
		if attached array_ as l_array then
			l_array.back
		else
			index_ := index_ - 1
		end
	end

	go_i_th (i: INTEGER)
		-- Move cursor to `i'-th position.
	do
		if attached array_ as l_array then
			l_array.go_i_th (i)
		else
			index_ := i
		end
	end

feature -- Measurement

	count, capacity: INTEGER
		-- Number of available indices
		do
			if attached array_ as l_array then
				Result := l_array.count
			else
				if has_object_	then
					Result := 1
				else
					Result := 0
				end
			end
		ensure then
			consistent_with_bounds: Result = upper - lower + 1
		end

	is_empty: BOOLEAN
		-- Is structure empty?
		do
			Result := (count = 0)
		end

feature -- Access

	current_item: G
			-- Item at current position
		do
			if attached array_ as l_array	then
				Result := l_array.current_item
			else
				if has_object_	then
					if attached object_ as l_object then
						Result := l_object
					else
						check ({G}).has_default then
								-- since it is not attached, it must be detachable
							Result := ({G}).default
						end
					end
				else
					check	precondition_not_met: False then
							-- Empty, so `off'
					end
				end
			end
		end

feature -- Traversal

	start
		-- Move cursor to first position if any.
	do
		if attached array_ as l_array then
			l_array.start
		else
			index_ := lower
		end
	ensure then
		after_when_empty: is_empty implies after
	end

feature -- Access

	valid_index (i: INTEGER): BOOLEAN
		-- Is `i' within the bounds of the array?
		do
			if attached array_ as l_array then
				Result := l_array.valid_index(i)
			else
				Result := (lower <= i) and then (i <= upper)
			end
		end

	valid_cursor (a_cursor: like cursor): BOOLEAN
			-- <Precursor>
		do
			Result := a_cursor.index >= lower - 1 and then
				a_cursor.index <= count + 1
		end

	item, infix "@" (i: INTEGER): detachable G
			-- Entry at index `i', if in index interval
		do
			if attached array_ as l_array then
				Result := l_array.item(i)
			else
				if has_object_ and valid_index(i) then
					Result := object_
				end
			end
		end

	lower: INTEGER
			-- Index of first item
		do
			if attached array_ as l_array then
				Result := l_array.lower
			else
				Result := lower_
			end
		end

	upper: INTEGER
			-- Index of last item
		do
			if attached array_ as l_array then
				Result := l_array.upper
			else
				if has_object_
				then
					Result := lower_
				else
					Result := lower_ - 1
				end
			end
		end

feature -- Element change

	put (v: G; i: INTEGER)
			--	Replace `i' th element with `v'.
		require
			valid_index: valid_index (i)
		do
			if attached array_ as l_array then
				l_array.put (v, i)
			else
				has_object_ := True
				object_ := v
			end
		end

	extend (v: G)
		-- Add `v' to end.
		-- Do not move cursor.
		do
			force (v, upper + 1)
		end

	force (v: G; i: INTEGER)
		-- Assign item `v' to `i'-th entry.
		-- Always applicable: resize the array if `i' falls out of
		-- currently defined bounds; preserve existing items.
		local
			l_lower, l_upper: INTEGER
			l_array_2: attached like array_
		do
			if attached array_ as l_array then
				l_array.force (v, i)
				lower_ := l_array.lower
			else
				l_lower := i.min (lower)
				l_upper := i.max (upper)
				if l_upper > l_lower then
					lower_ := l_lower
					create l_array_2.make (l_lower, l_upper)
					array_ := l_array_2
					l_array_2.put (v, i)
					if has_object_ then
						has_object_ := False
						if attached object_ as l_object then
							l_array_2.put (l_object, l_lower)
						else
							check ({G}).has_default end
							l_array_2.put (({G}).default, l_lower)							
						end
						if ({G}).has_default then
							object_ := ({G}).default
						end
					end
				else
					lower_ := l_lower
					has_object_ := True
					object_ := v
				end
			end
		ensure
			inserted: item (i) = v
			higher_count: count >= old count
		end

feature -- Traversal

	do_all (action: PROCEDURE [ANY, TUPLE [G]])
			-- <Precursor>
		do
			if attached array_ as l_array then
				l_array.do_all (action)
			elseif attached object_ as l_object then
				action.call ([l_object])
			elseif count > 0 then
					check ({G}).has_default then
							-- since it is not attached, it must be detachable
						action.call ([({G}).default])
					end
			end
		end

	finish
			-- Move cursor to last position if any.
		do
			if attached array_ as l_array then
				l_array.finish
			else
				index_ := upper
			end
		ensure then
			before_when_empty: is_empty implies before
		end

feature -- Sort

	sort_by (a_comparison: FUNCTION [ANY, TUPLE [G, G], BOOLEAN])
			-- Sort elements according to `a_comparison'.
		require
			a_comparison_attached: a_comparison /= Void
		do
			if attached array_ as l_array then
				l_array.sort_by (a_comparison)
			else
				-- No need to sort since there aren't even two elements.
			end
		end

feature -- Access

	first: detachable G
			-- Item at first position
		do
			Result := item (lower)
		end

	last: like first
			-- Item at last position if any
		do
			if count > 0 then
				Result := item (upper)
			end
		end

feature -- Element change

	append_array (a_array: detachable like Current)
			-- Add elements of `a_array' to end of `Current'.
		local
			i: INTEGER
		do
			if a_array /= Void then
				from
					i := a_array.lower
				until
					i > a_array.upper
				loop
					if attached a_array.item (i) as l_item then
						extend (l_item)
					elseif ({G}).has_default then
						extend (({G}).default)
					else
						check
							cant_happen_as_G_must_be_detachable: False
						end
					end
					i := i + 1
				end
			end
		end

feature -- Status report

	off: BOOLEAN
			-- <Precursor>
	do
		Result := before or after
	end

feature -- Removal

	remove
			-- <Precursor>
		do
			if has_object_ then
				has_object_ := False
			else
				if attached array_ as l_array and then l_array.count > 1 then
					l_array.remove
					if l_array.count = 1 then
						object_ := l_array.first
						has_object_:= True
						array_ := Void
					end
				end
			end
		end

	wipe_out
			-- Remove all items.
		local
			l_default: detachable G
		do
			if array_ /= Void then
				array_ := Void
			end
			if has_object_ then
				if ({G}).has_default then
					object_ := ({G}).default
				else
					check attached l_default as l_default_2 then
							-- self-initializing type
						object_ := l_default_2
					end
				end
				has_object_ := False
			end
			make_empty
		end

feature {NONE} -- Implementation

	array_: detachable ROSE_LINEAR_ARRAY [G]
			-- Proxied array

	index_: INTEGER
			-- Position of `current_item'

	lower_: INTEGER
			-- Position of first element (if any)
	
end
