note
	
	description: "Critical section wrapper"

	author: "Darin Peshev"
	last_editor: "$Author: c4 $"
	date: "$Date: 2014/05/13 $"
	revision: "$Revision: #1 $"
	path: "$File: //rose/source/library/kernel_infrastructure_base/src/windows/critical_section.e $"
	documentation: "http://msdn2.microsoft.com/en-us/library/ms682530.aspx"
	keywords: "synchronization,thread,process,mutex,interprocess"

class CRITICAL_SECTION

inherit
	
	DISPOSABLE

create

	make

feature {NONE} -- Initialization

	make
			-- Create critical section.
		do
			ptr := ptr.memory_alloc (size_of_critical_section)
			if not initialize_critical_section_and_spin_count (ptr, 4000) then
				ptr := default_pointer
			end
		end

feature -- Access

	is_set: BOOLEAN
			-- True if the critical section is initialized sucessfully
		do
			Result := ptr /= default_pointer
		end

feature -- Basic operations

	lock
			-- Waits for ownership of the specified critical section object.
		require
			is_set: is_set
		do
			enter_critical_section (ptr)
		end

	try_lock: BOOLEAN
			-- Attempts to enter a critical section without blocking.
			-- The Result is True if the entry was sucessful and False otherwise
		require
			is_set: is_set
		do
			Result := try_enter_critical_section (ptr)
		end

	unlock
			-- Releases ownership of the specified critical section object.
		require
			is_set: is_set
		do
			leave_critical_section (ptr)
		end

	dispose
			-- release allocated memory
		do
			if ptr /= default_pointer then
				delete_critical_section (ptr)
				ptr.memory_free
				ptr := default_pointer
			end
		end

feature {NONE} -- Implementation

	ptr: POINTER
			-- Pointer to underlying critical section

	size_of_critical_section: INTEGER
			-- size in bytes of a CRITICAL_SECTION structure
		external
			"C macro use <windows.h>"
		alias
			"sizeof(CRITICAL_SECTION)"
		end

	initialize_critical_section_and_spin_count (a_ptr: POINTER; a_spin_count: INTEGER): BOOLEAN
			-- Initializes a critical section object and sets the spin count for the critical section.
		external
			"C blocking signature (LPCRITICAL_SECTION, DWORD):EIF_BOOLEAN use <windows.h>"
		alias
			"InitializeCriticalSectionAndSpinCount"
		end

	delete_critical_section (a_ptr: POINTER)
			-- Releases all resources used by an unowned critical section object.
		external
			"C blocking signature (LPCRITICAL_SECTION) use <windows.h>"
		alias
			"DeleteCriticalSection"
		end

	enter_critical_section (a_ptr: POINTER)
			-- Waits for ownership of the specified critical section object.
		external
			"C blocking signature (LPCRITICAL_SECTION) use <windows.h>"
		alias
			"EnterCriticalSection"
		end

	leave_critical_section (a_ptr: POINTER)
			-- Releases ownership of the specified critical section object.
		external
			"C blocking signature (LPCRITICAL_SECTION) use <windows.h>"
		alias
			"LeaveCriticalSection"
		end

	try_enter_critical_section (a_ptr: POINTER): BOOLEAN
			-- Attempts to enter a critical section without blocking.
		external
			"C blocking signature (LPCRITICAL_SECTION): EIF_BOOLEAN use <windows.h>"
		alias
			"TryEnterCriticalSection"
		end

end
