note

	description:

		"Reviewed mutex class for ELOG, tries to help with debugging with critical sections"

	library: "elogger"
	copyright: "Copyright (c) 2006, AXA Rosenberg, and others (see copyright.txt)"
	license: "Eiffel Forum License v2 (see forum.txt)"
	author: "Ulrich Weiss, and others"
	last_editor: "$Author: c4 $"
	date: "$Date: 2014/05/13 $"
	revision: "$Revision: #1 $"
	path: "$File: //rose/source/library/kernel_infrastructure_base/src/elog_mutex.e $"

class ELOG_MUTEX

create
	make
	
convert
	to_mutex: {MUTEX}
	
feature {NONE} -- Initialization
	
	make
			-- Initialize `Current' to be noisy.
		do
			if {PLATFORM}.is_thread_capable then
				create mutex.make
				create thread_control
			end
		end

feature -- Access

	to_mutex: MUTEX
			-- Mutex associated to this ELOG_MUTEX
		do
			if attached mutex as l_mutex then
				Result := l_mutex
			else
				create Result.make
			end
		end

	is_owned_by_current_thread: BOOLEAN
			-- Does the current thread own this mutex?
		do
			if mutex /= Void then
				Result := owning_thread = get_current_id
			else
				Result := lock_count > 0
			end
		end

	lock_count: INTEGER
			-- How many times `owning_thread' has locked this mutex

	owning_thread: POINTER
			-- The thread that has this mutex

feature -- Basic operations

	lock
			-- Lock mutex, waiting if necessary until that becomes possible
		local
			l_id: like get_current_id
		do
			if attached mutex as l_mutex then
				l_mutex.lock
				l_id := get_current_id
				if owning_thread = l_id then
					-- Do nothing
				elseif owning_thread = default_pointer then
					owning_thread := l_id
				else
					check bad_owner: False end
				end
			end
			lock_count := lock_count + 1
			debug ("mutex")
				print ("Mutex acquired: " + owning_thread.out + " : count=" + lock_count.out + "%N")
			end
		ensure
			is_owned_by_current_thread: is_owned_by_current_thread
		end

	try_lock: BOOLEAN
			-- Has client been successful in locking mutex without waiting?
		local
			l_id: like get_current_id
		do
			if attached mutex as l_mutex then
				Result := l_mutex.try_lock
				if Result then
					l_id := get_current_id
					if owning_thread = l_id then
						-- Do nothing
					elseif owning_thread = default_pointer then
						owning_thread := l_id
					else
						check bad_owner: False end
					end
					lock_count := lock_count + 1
				end
			else
				Result := True
				lock_count := lock_count + 1
			end
			debug ("mutex")
				if Result then
					print ("Mutex acquired: " + owning_thread.out + " : count=" + lock_count.out + "%N")
				else
					print ("Mutex acquisition failed (try_lock)%N")
				end
			end
		ensure
			locked_means_owned: Result implies is_owned_by_current_thread
		end

	unlock
			-- Unlock mutex.
		require
			lock_acquired: lock_count > 0 and then is_owned_by_current_thread
		do
			lock_count := lock_count - 1
			if attached mutex as l_mutex then
				if lock_count = 0 then
					owning_thread := default_pointer
				end
				l_mutex.unlock
			end
			debug ("mutex")
				print ("Mutex released: " + owning_thread.out + " : count=" + lock_count.out + "%N")
			end
		--ensure
			-- These can't be guaranteed once the mutex is released
			--when_unlocked_owning_thread_unset: lock_count = 0 implies owning_thread = default_pointer
			--when_locked_owning_thread_is_current_thread: lock_count /= 0 implies owning_thread = get_current_id
		end

feature {NONE} -- Implementation

	mutex: detachable MUTEX
			-- Critical section used for effective synchronization

	thread_control: detachable ROSE_THREAD_CONTROL
			-- access to thread control when we are indeed multi-threaded

	get_current_id: POINTER
			-- the current thread id
		require
			thread_control_void: thread_control /= Void
		do
			check attached thread_control as l_control then
					-- precondition
				Result := l_control.get_current_id
			end
		end
		
invariant
	
	lock_count_not_negative: lock_count >= 0

end
