note

	description:

		"Ancestor for various date types"

	author: "Mark Howard"
	last_editor: "$Author: c4 $"
	date: "$Date: 2014/10/20 $"
	revision: "$Revision: #3 $"
	path: "$File: //rose/source/library/datetime/src/rose_chronological.e $"

class	ROSE_CHRONOLOGICAL

inherit

	DATE
		redefine
			is_less,
			infix "-",
			make,
			month_add,
			origin,
			out
		end

	HASHABLE
		undefine
			is_equal,
			out
		end

	ROSE_DATE_INTERVAL
		undefine
			is_equal,
			out
		end

create
	
	make,
	make_month_day_year,
	make_day_month_year,
	make_now,
	make_by_days,
	make_from_string_default,
	make_ccyymmdd,
	make_ccyymmdd_forgiving,
	make_from_rose_julian_day,
	make_w_string,
	make_from_julian_month,
	make_from_julian_quarter,
	make_unavailable,
	make_undetermined,
	make_today,
	make_yymmdd,
	make_mmddccyy,
	make_w_delimited_mmddyy_string,
	make_w_delimited_dd_mmm_yy_string,
	make_w_delimited_dd_mmm_ccyy_string

feature -- Initialization

	make (a_year, a_month, a_day: INTEGER)
			-- <Precursor>
		do
			if is_correct_date (a_year, a_month, a_day) then
				Precursor (a_year, a_month, a_day)
			else
				make_unavailable
			end
		end

	make_ccyymmdd_forgiving (a_ccyymmdd: INTEGER)
			-- Create from `a_ccyymmdd'.
			-- Intended to be 8 digits, but also works for 6 digits (via a heuristic).
			-- May result in `is_undetermined', especially with other numbers of digits.
		local
			yyyy: INTEGER
		do
			if a_ccyymmdd = 19999999
			or else a_ccyymmdd >= 99999999
			or else a_ccyymmdd <= 999999 then
				make_undetermined
			else
				yyyy := a_ccyymmdd // 10000
				if yyyy < 100 then    --| a 6 digit number
					if yyyy < 50 then --| infers a Y2K date
						yyyy := yyyy + 2000
					else              --| infers a Y1K date
						yyyy := yyyy + 1900
					end
				end
				make (yyyy, (a_ccyymmdd \\ 10000) // 100, a_ccyymmdd \\ 100)
			end
		end

	make_ccyymmdd (a_ccyymmdd: INTEGER)
			-- Create from `a_ccyymmdd'.
			-- Intended to be 8 digits, but also works for 6 digits (via a heuristic).
			-- May result in `is_undetermined', especially with other numbers of digits.
			-- Differs from `make_ccyymmdd_forgiving' only in descendant `ROSE_DATE', where
			--  it ensures persecs is zero. Therefore no difference when called as a creation procedure.
		do
			make_ccyymmdd_forgiving (a_ccyymmdd)
		end

	make_from_rose_julian_day (a_julian: INTEGER)
			-- Create date from `a_julian'.
		do
			make_by_days (a_julian + julian_start - 1)
		ensure
			days_set: days = (a_julian + julian_start - 1)
		end

	make_w_string (a_ccyymmdd: STRING)
			-- Create date from `a_ccyymmdd' as string.
		require
			valid_parameter: a_ccyymmdd /= Void and then a_ccyymmdd.is_integer
		do
			make_ccyymmdd (a_ccyymmdd.to_integer)
		end

	make_from_julian_month (i: INTEGER)
			-- Create by interpreting `i' as number of months since January 1950.
			-- Nothing to do with the Julian calender, despite the name.
			-- Represents last day of such month.
		do
			make (1950 + (i-1)//12, (i-1)\\12 + 1, 1)
			set_to_end_of_month
		end

	make_from_julian_quarter(a_julian: INTEGER)
			-- Create by interpreting `a_julian' as number of quarters since January 1950.
			-- Nothing to do with the Julian calender, despite the name.
			-- Represents last day of such quarter.		
		do
			make_from_julian_month (a_julian * 3)
		end

	make_unavailable
			-- Create as January 1st 1800.
		do
			make (1800, 1, 1)
		ensure
			unavailable: is_unavailable
		end

	make_undetermined
			-- Create as January 1st 9999.
		do
			make (9999, 1, 1)
		ensure
			undetermined: is_undetermined
		end

	make_today
			-- Set `Current' to today's date.
		do
			make_now
		end

	make_yymmdd (a_yymmdd: INTEGER)
			-- Create as a date prior to 2050. Infer century accordingly.
		do
			if a_yymmdd < 500000 then
				make_ccyymmdd (20000000 + a_yymmdd)
			else
				make_ccyymmdd (19000000 + a_yymmdd)
			end
		end

	make_mmddccyy (a_mmddccyy: INTEGER)
			-- Create from integer of the form MMDDCCYY.
		local
			l_ccyymmdd: INTEGER
		do
				--| Get CCYY from a_mmddccyy
			l_ccyymmdd := (a_mmddccyy \\ 10000) * 10000
				--| Add MMDD from a_mmddccyy
			l_ccyymmdd := l_ccyymmdd + (a_mmddccyy // 10000)
			make_ccyymmdd (l_ccyymmdd)
		end

	make_w_delimited_mmddyy_string (a_time_as_string: STRING)
			-- Create time from delimited string of form MM/DD/YY or M/DD/YY or M/D/YY
			-- with possible trailing blank characters.
			-- Mutates `a_time_as_string'.
		require
			a_time_as_string_not_void: a_time_as_string /= Void
		local
			l_yymmdd_string: STRING
			l_substring: STRING
			l_first_delimiter_position: INTEGER
			l_second_delimiter_position: INTEGER
			l_ccyymmdd: INTEGER
		do
			a_time_as_string.prune_all_trailing (' ')
			l_first_delimiter_position := a_time_as_string.index_of ('/', 1)
			l_second_delimiter_position := a_time_as_string.index_of ('/', l_first_delimiter_position+1)
			l_substring := a_time_as_string.substring (1,l_first_delimiter_position-1)
			if l_substring.count = 1 then
				l_substring.insert_string ("0", 1)
			end
			l_yymmdd_string := l_substring.twin
			l_substring := a_time_as_string.substring (l_first_delimiter_position+1, l_second_delimiter_position-1)
			if l_substring.count = 1 then
				l_substring.insert_string ("0", 1)
			end
			l_yymmdd_string.append (l_substring)
			l_substring := a_time_as_string.substring (l_second_delimiter_position+1, a_time_as_string.count)
			l_yymmdd_string.insert_string (l_substring, 1)
			if l_yymmdd_string.is_integer then
				l_ccyymmdd := l_yymmdd_string.to_integer
				if l_ccyymmdd < 500000 then	--| assumes 01/01/49 is 01/01/2049
					l_ccyymmdd := l_ccyymmdd + 20000000
				else						--| assumes 01/01/50 is 01/01/1950
					l_ccyymmdd := l_ccyymmdd + 19000000
				end
				make_ccyymmdd (l_ccyymmdd)
			end
		end

	make_w_delimited_mmddccyy_string (a_time_as_string: STRING)
			-- Create time from delimited string of form MM/DD/CCYY or M/DD/CCYY or M/D/CCYY
			-- with possible trailing blank characters.
			-- Mutates `a_time_as_string'.
		require
			a_time_as_string_not_void: a_time_as_string /= Void
		local
			l_yymmdd_string: STRING
			l_substring: STRING
			l_first_delimiter_position: INTEGER
			l_second_delimiter_position: INTEGER
			l_ccyymmdd: INTEGER
		do
			a_time_as_string.prune_all_trailing (' ')
			l_first_delimiter_position := a_time_as_string.index_of ('/', 1)
			l_second_delimiter_position := a_time_as_string.index_of ('/', l_first_delimiter_position+1)
			l_substring := a_time_as_string.substring (1,l_first_delimiter_position-1)
			if l_substring.count = 1 then
				l_substring.insert_string ("0", 1)
			end
			l_yymmdd_string := l_substring.twin
			l_substring := a_time_as_string.substring (l_first_delimiter_position+1, l_second_delimiter_position-1)
			if l_substring.count = 1 then
				l_substring.insert_string ("0", 1)
			end
			l_yymmdd_string.append (l_substring)
			l_substring := a_time_as_string.substring (l_second_delimiter_position+1, a_time_as_string.count)
			l_yymmdd_string.insert_string (l_substring, 1)
			if l_yymmdd_string.is_integer then
				l_ccyymmdd := l_yymmdd_string.to_integer
				make_ccyymmdd (l_ccyymmdd)
			end
		end

	make_w_delimited_dd_mmm_yy_string (a_date_as_delimited_string: STRING)
			-- Create time from delimited string of form as 01-Jan-99, where Jan is case-insensitive
			-- with possible trailing blank characters.
			-- Mutates `a_date_as_delimited_string'.
		require
			a_date_as_delimited_string_not_void: a_date_as_delimited_string /= Void
		local
			l_substring: STRING
			l_first_delimiter_position: INTEGER
			l_second_delimiter_position: INTEGER
			l_ccyymmdd: INTEGER
			l_dd_string, l_mmdd_string: STRING
		do
			a_date_as_delimited_string.prune_all_trailing (' ')
			l_first_delimiter_position := a_date_as_delimited_string.index_of ('-', 1)
			l_second_delimiter_position := a_date_as_delimited_string.index_of ('-', l_first_delimiter_position+1)
			l_substring := a_date_as_delimited_string.substring (1, l_first_delimiter_position-1)
			if l_substring.count = 1 then
				l_substring.insert_string ("0", 1)
			end
			l_dd_string:= l_substring.twin
			l_substring := a_date_as_delimited_string.substring (l_first_delimiter_position+1, l_second_delimiter_position-1)
			--
			if l_substring /= Void then
				l_substring.to_lower
				if     l_substring.is_equal ("jan") then l_substring := "01"
				elseif l_substring.is_equal ("feb") then l_substring := "02"
				elseif l_substring.is_equal ("mar") then l_substring := "03"
				elseif l_substring.is_equal ("apr") then l_substring := "04"
				elseif l_substring.is_equal ("may") then l_substring := "05"
				elseif l_substring.is_equal ("jun") then l_substring := "06"
				elseif l_substring.is_equal ("jul") then l_substring := "07"
				elseif l_substring.is_equal ("aug") then l_substring := "08"
				elseif l_substring.is_equal ("sep") then l_substring := "09"
				elseif l_substring.is_equal ("oct") then l_substring := "10"
				elseif l_substring.is_equal ("nov") then l_substring := "11"
				elseif l_substring.is_equal ("dec") then l_substring := "12"
				else
					do_nothing
				end
			end
			l_mmdd_string := l_substring.twin
			if l_mmdd_string/= Void then
				l_mmdd_string.append (l_dd_string)
			end
			l_substring := a_date_as_delimited_string.substring (l_second_delimiter_position+1, a_date_as_delimited_string.count)
			l_mmdd_string.insert_string (l_substring, 1)
			if l_mmdd_string.is_integer then
				l_ccyymmdd := l_mmdd_string.to_integer
				if (l_ccyymmdd < 500000) then -- assumes 01/01/49 is 01/01/2049
					l_ccyymmdd := l_ccyymmdd + 20000000
				else                          -- assumes 01/01/50 is 01/01/1950
					l_ccyymmdd := l_ccyymmdd + 19000000
				end
				make_ccyymmdd ( l_ccyymmdd )
			end
		end

	make_w_delimited_dd_mmm_ccyy_string (a_date_as_delimited_string: STRING)
			-- Create time from delimited string of form as 01-Jan-99, where Jan is case-insensitive
			-- with possible trailing blank characters.
			-- Mutates `a_date_as_delimited_string'.
		require
			a_date_as_delimited_string_not_void: a_date_as_delimited_string /= Void
		local
			l_substring: STRING
			l_first_delimiter_position: INTEGER
			l_second_delimiter_position: INTEGER
			l_ccyymmdd: INTEGER
			l_dd_string, l_mmdd_string: STRING
		do
			a_date_as_delimited_string.prune_all_trailing (' ')
			l_first_delimiter_position := a_date_as_delimited_string.index_of ('-', 1)
			l_second_delimiter_position := a_date_as_delimited_string.index_of ('-', l_first_delimiter_position+1)
			l_substring := a_date_as_delimited_string.substring (1, l_first_delimiter_position-1)
			if l_substring.count = 1 then
				l_substring.insert_string ("0", 1)
			end
			l_dd_string:= l_substring.twin
			l_substring := a_date_as_delimited_string.substring (l_first_delimiter_position+1, l_second_delimiter_position-1)
			--
			if l_substring /= Void then
				l_substring.to_lower
				if     l_substring.is_equal ("jan") then l_substring := "01"
				elseif l_substring.is_equal ("feb") then l_substring := "02"
				elseif l_substring.is_equal ("mar") then l_substring := "03"
				elseif l_substring.is_equal ("apr") then l_substring := "04"
				elseif l_substring.is_equal ("may") then l_substring := "05"
				elseif l_substring.is_equal ("jun") then l_substring := "06"
				elseif l_substring.is_equal ("jul") then l_substring := "07"
				elseif l_substring.is_equal ("aug") then l_substring := "08"
				elseif l_substring.is_equal ("sep") then l_substring := "09"
				elseif l_substring.is_equal ("oct") then l_substring := "10"
				elseif l_substring.is_equal ("nov") then l_substring := "11"
				elseif l_substring.is_equal ("dec") then l_substring := "12"
				else
					do_nothing
				end
			end
			l_mmdd_string := l_substring.twin
			if l_mmdd_string/= Void then
				l_mmdd_string.append (l_dd_string)
			end
			l_substring := a_date_as_delimited_string.substring (l_second_delimiter_position+1, a_date_as_delimited_string.count)
			l_mmdd_string.insert_string (l_substring, 1)
			if l_mmdd_string.is_integer then
				l_ccyymmdd := l_mmdd_string.to_integer
				make_ccyymmdd ( l_ccyymmdd )
			end
		end

feature -- Access

	quarter: INTEGER
			-- Quarter of year in which `Current' falls;
			-- (1=1stJan - 31st Mar; 2= 1st Apr - 30th Jun; 3=1st Jul - 30th Sep; 4=1st Oct - 31st Dec.)
		do
			Result := (month + 2) // 3
		ensure
			one_to_four: Result >= 1 and Result <= 4
		end

	ccyymmdd: INTEGER
			-- Date encoded as integer such that its printed form is intelligible (YYYYMMDD)
		do
			Result := (year * 10000) + (month * 100) + day
		end

	ccyymm: INTEGER
			-- Date encoded as integer such that its printed form is intelligible (YYYYMM)
		do
			Result := (year * 100) + month
		end

	yymmdd: INTEGER
			-- Date encoded as integer such that its printed form is intelligible (YYMMDD);
			-- Century information is lost.
		do
			Result := ccyymmdd \\ 1000000
		end

	yymm: INTEGER
			-- Date encoded as integer such that its printed form is intelligible (YYMM);
			-- Century and day information is lost.
		do
			Result := yymmdd // 100
		end

	mmdd: INTEGER
			-- Date encoded as integer such that its printed form is intelligible (MMDD);
			-- Year information is lost.		
		do
			result:=yymmdd\\10000
		end

	ccyyq: INTEGER
			-- Date encoded as integer such that its printed form is intelligible (CCYYQ);
			-- Month information is lost.		
		do
			Result := century_year * 10 + quarter
		end

	day_of_year: INTEGER
			-- Number of days since December 31st in previous year
		do
			Result := year_day
		ensure
			result_large_enough: Result >= 1
			result_small_enough: Result <= days_in_year			
		end

	day_of_week_name: STRING
			-- English name for the day of the week
		do
			Result := day_names.item (day_of_the_week)
			if Result /= Void then
				Result := Result.twin
			end
		ensure
			day_of_week_name_attached: Result /= Void
		end

	previous_date: like Current
			-- Day before `Current'
		do
			Result := plus_n_date (-1)
		ensure
			previous_date_attached: Result /= Void
			previous_date: Result.days = (days - 1)
		end

	previous_weekday: like Current
			-- Most recent week-day that is earlier than `Current'
		do
			Result := plus_n_date (-1)
			if Result.is_weekend then
				Result := plus_n_date (-2)
				if Result.is_weekend then
					Result := plus_n_date (-3)
				end
			end
		ensure
			previous_weekday_attached: Result /= Void
			one_to_three_days_earlier: Result.days < days and Result.days >= (days - 3)
			not_weekend: not Result.is_weekend
		end

	current_or_previous_weekday: like Current
			-- `Current' if `is_weekday', otherwise `previous_weekday'
		do
			if is_weekday then
				Result := twin
			else
				Result := previous_weekday
			end
		ensure
			current_or_previous_weekday_attached: Result /= Void
			new_object: Result /= Current
			current_if_weekday: is_weekday implies (Result ~ Current)
			one_or_two_days_earlier: not is_weekday implies (Result.days < days and Result.days >= (days - 2))
			not_weekend: not Result.is_weekend
		end

	next_date: like Current
			-- Day after `Current'
		do
			Result := plus_n_date (1)
		ensure
			next_date_attached: Result /= Void
			next_date: Result.days = (days + 1)
		end

	next_weekday: like Current
			-- Earliest week-day that is later than `Current'
		do
			Result := plus_n_date (1)
			if Result.is_weekend then
				Result := plus_n_date (2)
				if Result.is_weekend then
					Result := plus_n_date (3)
				end
			end
		ensure
			next_weekday_attached: Result /= Void
			one_to_three_days_later: Result.days > days and Result.days <= (days + 3)
		end

	plus_n_date (n: INTEGER): like Current
			-- Date that is `n' days later than `Current'
		do
			Result := twin
			Result.day_add (n)
		ensure
			plus_n_date_attached: Result /= Void
			new_object: Result /= Current
			valid_diff: Result.days - days = n
		end

	plus_n_month (n: INTEGER): like Current
			-- Date that is `n' months later than `Current'
		do
			Result := twin
			Result.month_add (n)
		ensure
			plus_n_month_attached: Result /= Void
			new_object: Result /= Current
		end

	julian_quarter: INTEGER
			-- `Current' expressed as number of quarters since winter 1949;
			-- Despite the name, nothing to do with the Julian calendar.
		do
			Result := (julian_month + 2) // 3
		end

	julian_month: INTEGER
			-- `Current' expressed as number of months since December 1949;
			-- Despite the name, nothing to do with the Julian calendar.
		do
			Result := (year - 1950) * 12 + month
		end

	previous_julian_month: INTEGER
			-- `Current' expressed as number of months since November 1949;
			-- Despite the name, nothing to do with the Julian calendar.
		do
			Result := julian_month - 1
		end

	rose_julian_day: INTEGER
			-- Rosenberg definition of "julian" day: 1/1/1950 is Rosenberg "julian" day 1;
			-- Despite the name, nothing to do with the Julian calendar.
		do
			Result := days - julian_start + 1
		end

	rose_julian_week: INTEGER
			-- Rosenberg julian week: number of weeks since 1/1/1950;
			-- Despite the name, nothing to do with the Julian calendar.
		do
			Result := 1 + ((rose_julian_day) // 7)
		end

	rose_julian_week_start: ROSE_DATE
			-- January 1st 1950;
			-- Despite the name, nothing to do with the Julian calendar, nor with weeks.
		once
			create Result.make_ccyymmdd (19500101)
		ensure
			rose_julian_week_start_attached: Result /= Void
		end

	week_begining: ROSE_DATE
			-- First day of week containing today;
			-- First day of week = Saturday for this routine.
		do
			Result := rose_week.first_date
		ensure
			week_begining_attached: Result /= Void
			new_object: Result /= Current
		end

	month_beginning: ROSE_DATE
			-- First day of `Current's month
		do
			create Result.make (year, month, 1)
		ensure
			month_beginning_attached: Result /= Void
			new_object: Result /= Current
		end

	quarter_beginning: ROSE_DATE
			-- First day of `Current's quarter-year
		do
			create Result.make (year,quarter * 3 - 2,1)
		ensure
			quarter_beginning_attached: Result /= Void
			new_object: Result /= Current
		end

	year_beginning: ROSE_DATE
			-- First day of `Current's year
		do
			create Result.make (year,1,1)
		ensure
			year_beginning_attached: Result /= Void
			new_object: Result /= Current
		end

	year_end: ROSE_DATE
			-- Last day of `Current's year
		do
			create Result.make (year,12,31)
		ensure
			year_end_attached: Result /= Void
			new_object: Result /= Current			
		end

	month_end: ROSE_DATE
			-- Last day of `Current's month
		do
			create Result.make (year,month,1)
			Result.set_to_end_of_month
		ensure
			month_end_attached: Result /= Void
			new_object: Result /= Current
		end

	nearest_month_end: ROSE_DATE
			-- Last day of `Current's month if `day' is greater than 10th;
			-- Else prevous month
		do
			create Result.make_by_date (Current)
			if day >= 11 then
				Result.set_to_end_of_month
			else
				Result.month_back
				Result.set_to_end_of_month
			end
		ensure
			nearest_month_end_attached: Result /= Void
			new_object: Result /= Current
		end

	quarter_end: ROSE_DATE
			-- Last day of `Current's quarter-year
		do
			create Result.make (year, quarter * 3, 1)
			Result.set_to_end_of_month
		ensure
			quarter_end_attached: Result /= Void
			new_object: Result /= Current
		end

	previous_week_end: ROSE_DATE
			-- Last day of previous week
		do
			Result := week_begining.previous_date
		ensure
			previous_week_end_attached: Result /= Void
			new_object: Result /= Current
		end

	previous_month_end: ROSE_DATE
			-- Last day of previous month
		do
			Result:= month_beginning.previous_date
		ensure
			previous_month_end_attached: Result /= Void
			new_object: Result /= Current		
		end

	previous_quarter_end: ROSE_DATE
			-- Last day of previous quarter-year
		do
			Result:= quarter_beginning.previous_date
		ensure
			previous_quarter_end_attached: Result /= Void
			new_object: Result /= Current
		end

	previous_year_end: ROSE_DATE
			-- Last day of previous year
		do
			Result:= year_beginning.previous_date
		ensure
			previous_year_end_attached: Result /= Void
			new_object: Result /= Current
		end

	closest_month_end (a_day_cutoff: INTEGER): ROSE_DATE
			-- Closest month end;
			-- If `day' > `a_day_cutoff', this will be last day of current month;
			-- Otherwise last day of previous month.
		do
			if day <= a_day_cutoff then
				Result := previous_month_end
			else
				Result := month_end
			end
		ensure
			closest_month_end_attached: Result /= Void
			new_object: Result /= Current
		end

	is_end_of_quarter: BOOLEAN
			-- Does `Current' represent midnight on the last day of a quarter (31st March, 30th June, 30th September or 31st December)?
		do
			Result := same_date_time (quarter_end)
		end

	is_end_of_year: BOOLEAN
			-- Does `Current' represent midnight on the last day of a year?
		do
			Result := same_date_time (year_end)
		end

	next_quarter_beginning: ROSE_DATE
			-- First day of next quarter (1st January, 1st April, 1st July or 1st October)
		do
			Result := quarter_end
			Result.day_add (1)
		ensure
			next_quarter_beginning_attached: Result /= Void
			new_object: Result /= Current
		end

	rose_month: ROSE_MONTH
			-- Last day of month
		do
			create Result.make (year, month)
		ensure
			rose_month_attached: Result /= Void
			new_object: Result /= Current
		end

	quarter_ending_month: INTEGER
			-- `Current' expressed as number of months since January 1950
		do
			Result := julian_quarter * 3 - 1
		end

	quarter_beginning_month: INTEGER
			-- `Current' expressed as number of months since February 1950
		do
			Result := julian_quarter * 3 - 2
		end

	most_recent_month: ROSE_MONTH
			-- Latest completed month that is not later than `Current'
		do
			if day = days_in_month then
				create Result.make (year, month)
			elseif month > 1 then
				create Result.make (year, month - 1)
			else
				create Result.make (year - 1, 12)
			end
		ensure
			most_recent_month_attached: Result /= Void
			new_object: Result /= Current
		end

	rose_week: ROSE_WEEK
			-- Week of `Current';
			-- For this function, a week starts on Saturday.
		do
			create Result.make_from_rose_julian_week (rose_julian_week)
		ensure
			rose_week_attached: Result /= Void
			new_object: Result /= Current
		end

	century_year: INTEGER
			-- Synonym for `year'
		do
			Result := year
		ensure
			synonym_for_year: Result = year
		end

	rose_date: ROSE_DATE
			-- Day of `Current'
		do
			create Result.make_day_month_year (day, month, year)
		ensure
			rose_date_attached: Result /= Void
			new_object: Result /= Current		
		end

	first: ROSE_DATE
			-- First sub-unit of `Current';
			-- The meaning of a sub-unit depends upon the descendant, but it is at least as large as a day.
		do
			create Result.make_day_month_year (day, month, year)
		ensure then
			first_attached: Result /= Void
			new_object: Result /= Current
		end
	
	last: ROSE_DATE
			-- Last sub-unit of `Current';
			-- The meaning of a sub-unit depends upon the descendant, but it is at least as large as a day.
			-- The meaning is not necessarily the same as for `first' in any given descendant.
		do
			create Result.make_day_month_year (day, month, year)
		ensure then
			last_attached: Result /= Void
			new_object: Result /= Current			
		end

	origin: ROSE_CHRONOLOGICAL
			-- Origin date
		once
			create Result.make (1600, 1, 1)
		ensure then
			new_object: Result /= Current
			year_1600: Result.year = 1600
			month_january: Result.month = January
			first_of_month: Result.day = 1
		end

	unavailable_date: ROSE_DATE
			-- Shared object representing first day of latest year considered as too early for information
			--  of any business object to be available;
			-- Note that this is 200 years later than `origin'.
		once
			create Result.make_unavailable
		ensure
			unavailable_date_attached: Result /= Void
			unavialable: Result.is_unavailable
		end

	undetermined_date: ROSE_DATE
			-- Shared object representing a date too far into the future to be distinguished from infinity
		once
			create Result.make_undetermined
		ensure
			undetermined_date_attached: Result /= Void
			undetermined: Result.is_undetermined
		end

	hash_code: INTEGER
			-- <Precursor>
		do
			Result := compact_date.hash_code
		end

	trailing_day_sequence (a_count: INTEGER): ARRAY [like Current]
			-- Last `a_count' days prior to `Current' in ascending order
		require
			a_count_greater_than_zero : a_count > 0
		local
			i: INTEGER
			l_date: like Current
		do
			create Result.make (1, a_count)
			from i := 1 until i > a_count loop
				l_date := plus_n_date (i - a_count - 1)
				Result.put (l_date, i)
				i := i + 1
			end
		ensure
			trailing_day_sequence_attached: Result /= Void
			correct_count: Result.count = a_count
			one_indexed: Result.lower = 1
			all_dates_new_objects: across Result as i_date all attached i_date.item as l_date_2 and then l_date_2 /= Current end
		end

	dates_to (a_date: like Current): ARRAY [like Current]
			-- Inclusive sequence of days from `Current' to `a_date'
		require
			a_date_attached: a_date /= Void
			current_less_than_a_date : Current <= a_date
		local
			l_date: like Current
			i, l_count: INTEGER
		do
			l_count := 1 + a_date.days - days
			create Result.make (1, l_count)
			from i := 1 until i > l_count loop
				l_date := plus_n_date (i - 1)
				Result.put (l_date, i)
				i := i + 1
			end
		ensure
			dates_to_attached: Result /= Void
			correct_count: Result.count = (1 + a_date.days - days)
			one_indexed: Result.lower = 1
			all_dates_new_objects: across Result as i_date all attached i_date.item as l_date_2 and then l_date_2 /= Current end
			first_date_is_current: same_date (Result [1])
			last_date_is_a_date: a_date.same_date (Result [Result.count]) 
		end

	dates_from (a_date: like Current): ARRAY [like Current]
			-- Inclusive sequence of days from `a_date' to `Current'
		require
			a_date_attached: a_date /= Void
			current_greater_than_a_date: Current >= a_date
		local
			l_date: like Current
			i, l_count: INTEGER
		do
			l_count := 1 + days - a_date.days
			create Result.make (1, l_count)
			from i := 1 until i > l_count loop
				l_date := a_date.plus_n_date (i - 1)
				Result.put (l_date, i)
				i := i + 1
			end
		ensure
			dates_from_attached: Result /= Void
			correct_count: Result.count = (1 + days - a_date.days)
			one_indexed: Result.lower = 1
			all_dates_new_objects: across Result as i_date all attached i_date.item as l_date_2 and then l_date_2 /= Current end
			first_date_is_a_date: a_date.same_date (Result [1])
			last_date_is_current: same_date (Result [Result.count]) 
		end
	
	dates_from_descending (a_date: like Current): ARRAY [like Current]
			-- Inclusive sequence of days from `a_date' to `Current'
		require
			a_date_attached: a_date /= Void
			current_less_than_a_date: Current <= a_date
		local
			l_date: like Current
			i, l_count: INTEGER
		do
			l_count := 1 + a_date.days - days
			create Result.make (1, l_count)
			from
				i := 1
			until
				i > l_count
			loop
				l_date := a_date.plus_n_date (1 - i)
				Result.put (l_date, i)
				i := i + 1
			end
		ensure
			dates_from_attached: Result /= Void
			correct_count: Result.count = (1 + a_date.days - days)
			one_indexed: Result.lower = 1
			all_dates_new_objects: across Result as i_date all attached i_date.item as l_date_2 and then l_date_2 /= Current end
			first_date_is_a_date: a_date.same_date (Result [1])
			last_date_is_current: same_date (Result [Result.count]) 
		end

feature -- Element change

	set_to_end_of_month
			-- Set `Current' to end of `Current's month.
		do
			set_day (days_in_month)
		ensure
			day_set: day = days_in_month
		end

	add_century
			-- Add century with 20501231 being the 20th century cutoff.
			-- Does nothing if `year' > 99.
		do
			if year <= 50 then
				set_year (2000 + year)
			elseif year < 100 then
				set_year (1900 + year)
			end
		ensure
			century_included: year >= 100
		end

feature -- Status report

	is_available: BOOLEAN
			-- Is `Current' considered to be a date at which business objects might be available?
		do
			Result := year > 1800
		end

	is_undetermined: BOOLEAN
			-- Is `Current' too far into the future to be distinguished from infinity?
		do
			Result := year = 9999
		end

	frozen is_unavailable: BOOLEAN
			-- Is `Current' considered to be a date at which business objects can not be available?
		do
			Result := not is_available
		ensure
			definition: Result = not is_available
		end

	frozen is_weekday: BOOLEAN
			-- Is current date a weekday?
		do
			Result := not is_weekend
		ensure
			definition: Result = not is_weekend
		end

	is_weekend: BOOLEAN
			-- Is `Current' a Saturday or Sunday?
		do
			Result := day_of_the_week = Saturday or day_of_the_week = Sunday
		ensure
			definition: Result = (day_of_the_week = Saturday or day_of_the_week = Sunday)
		end

	is_end_of_month: BOOLEAN
			-- Is current date the last day of its month?
			-- (Do not take time in the day into account.)
		do
			Result := (day = days_in_i_th_month (month, year))
		end

	is_last_weekday_of_month: BOOLEAN
			-- Is current date the last weekday of its month?
			-- (Do not take time in the day into account.)
		local
			l_date: ROSE_DATE
		do
			l_date := month_end
			from until l_date.is_weekday loop
				l_date.add_n_days (-1)
			end
			Result := day = l_date.day
		end

	is_valid_date (a_year: INTEGER; a_month: INTEGER; a_day: INTEGER): BOOLEAN
			-- Is date with `a_year', `a_month', and `a_day' a valid date?
		do
			Result := a_month >= 1 and then a_month <= 12
				and then a_day >= 1 and then a_day <= days_in_i_th_month (a_month,a_year)
		end

	is_mm_dd_yy_valid (a_string: STRING): BOOLEAN
			-- Is `a_string' of the form "mm/dd/yy"?
			-- (Here, '/' may be any character at all - it is not checked.)
		require
			a_string_not_void : a_string /= Void
		local
			s: STRING
		do
			s := a_string.twin
			s.left_adjust
			Result := s /= Void and then s.count >= 8
				and then s.substring (1,2).is_integer
				and then s.substring (4,5).is_integer
				and then s.substring (7,8).is_integer
		end

	is_ccyymmdd_valid (a_string: STRING): BOOLEAN
			-- Is `a_string' of the form ccyymmdd?
		require
			valid_string: a_string /= Void
		local
			l_ccyymmdd : INTEGER
		do
			if a_string.is_integer then
				l_ccyymmdd := a_string.to_integer
				Result := l_ccyymmdd >= 10000000 and then l_ccyymmdd <= 99999999
			end
		end

	is_valid_ccyymmdd_date (a_string: STRING): BOOLEAN
			-- Is `a_string' of the form ccyymmdd and also a valid date?
		require
			a_string_not_void : a_string /= Void
		do
			Result := is_ccyymmdd_valid (a_string)
			if Result then
				Result := is_valid_date (a_string.substring (1, 4).to_integer, a_string.substring (5, 6).to_integer, a_string.substring (7, 8).to_integer)
			end
		end

feature -- Comparison

	is_equal_in_sec (a_other: like Current): BOOLEAN
			-- Is `Current' equal to `a_other' up to the seconds part? (ie, disregarding the finer-than-a-second part)
		require
			a_other_not_void: a_other /= Void
		do
			Result := (ordered_compact_date = a_other.ordered_compact_date) and
				((persecs // 10000) = (a_other.persecs // 10000))
		end

	is_less alias "<" (a_other: like Current): BOOLEAN
			-- <Precursor>
		do
			Result := Precursor (a_other)
			if not Result and then ordered_compact_date = a_other.ordered_compact_date then
				check attached a_other as l_other then
						 -- precondition clause other_exists
					Result := persecs < l_other.persecs
				end
			end
		end

	same_date (a_other: ROSE_CHRONOLOGICAL): BOOLEAN
			-- Are `Current' and `a_other' the same date?
			-- (Do not take time in the day into account.)
		require
			a_other_not_void: a_other /= Void
		do
			Result := ordered_compact_date = a_other.ordered_compact_date
		end

	same_date_time (a_other: ROSE_CHRONOLOGICAL): BOOLEAN
			-- Are `Current' and `a_other' the same date and time?
		require
			a_other_not_void: a_other /= Void
		do
			Result := ordered_compact_date = a_other.ordered_compact_date and then persecs = a_other.persecs
		end

feature -- Basic operations

	infix "-" (a_other: ROSE_CHRONOLOGICAL): INTERVAL [ROSE_CHRONOLOGICAL]
			-- Interval between current object and `other'
		do
			create Result.make (a_other, Current)
		end

	diff_in_days (a_other: like Current): INTEGER
			-- Difference with `a_other' date in days
		require
			valid_other: a_other /= Void
		do
			Result := days - a_other.days
		end

	add_n_days (n: INTEGER)
			-- Add `n' days to `Current'.
		do
			day_add (n)
		ensure
			days_set: days = old days + n			
		end

	add_n_weeks (n: INTEGER)
			-- Add `n' weeks to `Current'.
		do
			week_add (n)
		ensure
			days_set: days = old days + (7 * n)				
		end

	add_n_months (n: INTEGER)
			-- Add `n' months to `Current'.
			-- Can move `days' backward.
		do
			month_add (n)
		end

	add_n_years (n: INTEGER)
			-- Add `n' years to the current date.
			-- May cut the 29th february.
		do
			year_add (n)
		ensure
			year_set: year = old year + n
		end

	advance_to_friday
			-- Stay if on friday, else move forward to next friday.
		do
			from
			until
				day_of_the_week = Friday
			loop
				add_n_days (1)
			end
		ensure
			friday: day_of_the_week = Friday
		end


	week_add (w: INTEGER)
			-- Add `w' weeks to `Current'.
		do
			add_n_days (w * 7)
		ensure
			days_set: days = old days + (7 * w)
		end

	month_add (m: INTEGER)
			-- Add `m' months to the current date. Can move days backward.
		local
			l_day, l_new_year, l_new_month, l_days_in_new_month: INTEGER
			l_end_of_month: BOOLEAN
		do
			l_day := day
			l_end_of_month := l_day = days_in_i_th_month (month, year)
			l_new_year := year + div (month + m - 1, Months_in_year)
			if l_day <= days_in_i_th_month (month, l_new_year) then
				set_year (l_new_year)
			else
				set_day (1)
				set_year (l_new_year)
			end
			l_new_month := mod (month + m - 1, Months_in_year) + 1
			l_days_in_new_month := days_in_i_th_month (l_new_month, year)
			set_day (1)
			set_month (l_new_month)
			if l_day > l_days_in_new_month or else l_end_of_month then
				set_day (l_days_in_new_month)
			else
				set_day (l_day)
			end
		end

feature -- Output

	out: STRING
			-- <Precursor>
		do
			Result := ccyymmdd.out
		end

feature -- Constants

	frozen Sunday: INTEGER = 1
			-- Index into `day_names'
	
	frozen Monday: INTEGER = 2
			-- Index into `day_names'

	frozen Tuesday: INTEGER = 3
			-- Index into `day_names'

	frozen Wednesday: INTEGER = 4
			-- Index into `day_names'

	frozen Thursday: INTEGER = 5
			-- Index into `day_names'

	frozen Friday: INTEGER = 6
			-- Index into `day_names'

	frozen Saturday: INTEGER = 7
			-- Index into `day_names'

	frozen January: INTEGER = 1
			-- Index into `month_names'
	
	frozen February: INTEGER = 2
			-- Index into `month_names'

	frozen March: INTEGER = 3
			-- Index into `month_names'

	frozen April: INTEGER = 4
			-- Index into `month_names'

	frozen May: INTEGER = 5
			-- Index into `month_names'

	frozen June: INTEGER = 6
			-- Index into `month_names'

	frozen July: INTEGER = 7
			-- Index into `month_names'

	frozen August: INTEGER = 8
			-- Index into `month_names'

	frozen September: INTEGER = 9
			-- Index into `month_names'

	frozen October: INTEGER = 10
			-- Index into `month_names'

	frozen November: INTEGER = 11
			-- Index into `month_names'

	frozen December: INTEGER = 12
			-- Index into `month_names'

	frozen day_names: ARRAY [STRING]
			-- English names for days of the week, indexed by day number
		once
			create Result.make (sunday, Saturday)
			Result.put ("Sunday", sunday)
			Result.put ("Monday", monday)
			Result.put ("Tuesday", tuesday)
			Result.put ("Wednesday", wednesday)
			Result.put ("Thursday", thursday)
			Result.put ("Friday", friday)
			Result.put ("Saturday", saturday)
		ensure
			day_names_attached: Result /= Void
			starts_on_sunday: Result.lower = Sunday
			ends_on_saturday: Result.upper = Saturday
		end

	frozen month_names: ARRAY [STRING]
			-- English names for months, indexed by month number
		once
			create Result.make (january, december)
			Result.put ("January", january)
			Result.put ("February", february)
			Result.put ("March", march)
			Result.put ("April", april)
			Result.put ("May", may)
			Result.put ("June", june)
			Result.put ("July", july)
			Result.put ("August", august)
			Result.put ("September", september)
			Result.put ("October", october)
			Result.put ("November", november)
			Result.put ("December", december)
		ensure
			month_names_attached: Result /= Void
			starts_in_january: Result.lower = January
			ends_in_december: Result.upper = December
		end

feature {ROSE_CHRONOLOGICAL} -- Implementation

	persecs: INTEGER
			-- 10000ths of seconds elapsed since midnight
		do
			Result := 0
		end

feature {NONE} -- Implementation

	julian_start: INTEGER = 127835
			-- Number of days from `origin' to Rosenberg epoch (1950)

	centuryless_time_cutoff: INTEGER
			-- Centuryless times below this are assumed to be in the twentyfirst century, those
			-- above it are in the twentieth century.
			-- Should be:
			-- 25 for ROSE_YEAR
			-- 2512 for ROSE_MONTH
			-- 251231 for ROSE_DATE
		do
			Result := centuryless_time_cutoff_internal // cutoff_time_divisor
		end

	Centuryless_time_cutoff_internal: INTEGER = 251231
			-- Numerator for `centuryless_time_cutoff'

	cutoff_time_divisor: INTEGER
			-- Amount to divide `centuryless_time_cutoff_internal' by to get the `centuryless_time_cutoff'
		do
			Result := 1000000 // century_multiplier
		end

	century_multiplier: INTEGER
			-- Synonym for `Date_century_multiplier'
		once
			Result := date_century_multiplier
		end

	Date_century_multiplier: INTEGER = 1000000
			-- Denominator for  `cutoff_time_divisor'

	twentyfirst_century: INTEGER
			-- Amount to add to a centuryless time to make it part of the twentyfirst century.
		do
			Result := twentyfirst_cc * century_multiplier
		end

	Twentieth_cc: INTEGER = 19
			-- Numerator for `twentieth_century'
	
	Twentyfirst_cc: INTEGER = 20
			-- Numerator for `twentyfirst_century'

	twentieth_century: INTEGER
			-- Amount to add to a centuryless time to make it part of the twentieth century
		do
			Result := twentieth_cc * century_multiplier
		end
	
invariant
	
	sensible_interval: first <= last

end
