note

	description:

		"Tree of partitions of a set of objects, partitionned accordingly to a scale of criteria"

	author: "Zoran Simic"
	last_editor: "$Author: c4 $"
	date: "$Date: 2015/04/29 $"
	revision: "$Revision: #2 $"
	path: "$File: //rose/source/library/quotient/src/rose_quotient.e $"

class	ROSE_QUOTIENT [G]

create

	make

create {ROSE_QUOTIENT}

make_child

feature {NONE} -- Initialization

	make (a_definition: ROSE_QUOTIENT_DEFINITION [G])
			-- Create a quotient associated with `a_definition'
		require
			valid_definition: a_definition /= Void
		do
			definition := a_definition
			create invariants.make (Current, Void)
			wipe_out
		ensure
			is_empty: is_empty
		end

	make_child (a_parent: like Current)
			-- Make as a child of `a_parent'.
		require
			valid_parent: a_parent /= Void
			parent_scaled: a_parent.scale /= Void
		do
			definition := a_parent.definition
			create invariants.make (Current, Void) --| bodge for void-safety checking
			check attached a_parent.scale as l_scale then
					-- precondition "parent_scaled"
				scale := l_scale.following
				parent := a_parent
				create invariants.make (Current, l_scale.first)
			end
			wipe_out
		ensure
			empty: is_empty
		end

feature -- Access

	definition: ROSE_QUOTIENT_DEFINITION [G]
			-- Associated definition

	scale: detachable ROSE_QUOTIENT_SCALE [G]
			-- Scale (criteria sequence) in which children are partitioned and sorted

	parent: detachable like Current
			-- Parent node

	invariants: ROSE_QUOTIENT_INVARIANT [G]
			-- Invariants in this quotient, by criteria;
			-- Invariants are the set of criteria common to all objects

	children: detachable ROSE_LINEAR_ARRAY [like Current]
			-- Children partitioned accordingly to `criteria'
			-- (Not Void only if `scale' is not Void)

	objects: detachable ROSE_LINEAR_ARRAY [G]
			-- Objects in this node
			-- (Not Void only if `scale' is Void)

	all_objects: detachable ROSE_LINEAR_ARRAY [G]
			-- All objects descending from this node, sorted by `scale'
		do
			if objects /= Void then
				Result := objects
			elseif children /= Void then
				create Result.make_empty
				fill_in_objects (Result, Current)
			end
		end

	depth: INTEGER
			-- Depth of this tree

	new_quotient (a_scale: detachable ROSE_QUOTIENT_SCALE [G]): like Current
			-- New quotient organized accordingly to `a_scale' of criteria
			-- and containing all the objects of Current
		do
			create Result.make (definition)
			Result.set_scale (a_scale)
			Result.extend (all_objects)
		ensure
			new_quotient_valid: Result /= Void and then Result.is_root
		end

	root: like Current
			-- Root quotient of this quotient tree
		do
			if attached parent as l_parent then
				Result := l_parent.root
			else
				Result := Current
			end
		ensure
			root_attached: Result /= Void
		end

feature -- Measure

	leave_count: INTEGER
			-- Number of leaves in this tree
		local
			i: INTEGER
		do
			if attached children as l_children then
				from
					i := l_children.lower
				until
					i > l_children.upper
				loop
					Result := Result + l_children.item (i).leave_count
					i := i + 1
				end
			else
				Result := 1
			end
		end

	object_count: INTEGER
			-- Number of objects descending from this node
		local
			i: INTEGER
		do
			if attached objects as l_objects then
				Result := l_objects.count
			elseif attached children as l_children then
				from
					i := l_children.lower
				until
					i > l_children.upper
				loop
					Result := Result + l_children.item (i).object_count
					i := i + 1
				end
			end
		end

feature -- Status report

	is_root: BOOLEAN
			-- Is this node a root?
		do
			Result := parent = Void
		ensure
			no_parent: Result = (parent = Void)
		end

	is_leave: BOOLEAN
			-- Is this node a leaf?
		do
			Result := children = Void
		ensure
			no_children: Result = (children = Void)
		end

	is_empty: BOOLEAN
			-- Is this node empty (no objects)?
		do
			Result := object_count = 0
		end

feature -- Basic operations

	set_scale (a_scale: detachable ROSE_QUOTIENT_SCALE [G])
			-- Set `scale' for this quotient, this resets this quotient
			-- Last criteria in `scale' is used for sorting only,
			-- all the others are used for partitioning and sorting
		require
			root_of_tree: is_root
		do
			scale := a_scale
			wipe_out
		ensure
			is_empty: is_empty
			is_root: is_root
		end

	add (a_object: G)
			-- Add `a_object' to this tree, at the correct place.
		require
			a_object_attached: a_object /= Void
		local
			l_quotient: detachable like Current
			i, j: INTEGER
		do
			invariants.synchronize (a_object)
			if attached objects as l_objects then
					l_objects.extend (a_object)
			elseif attached children as l_children then
				from
					i := l_children.lower
				until
					l_quotient /= Void or else i > l_children.upper
				loop
					j := l_children.item (i).invariants.compared (a_object)
					if j = 0 then
						l_quotient := l_children.item (i)
					elseif j < 0 then
						create l_quotient.make_child (Current)
						l_children.insert (l_quotient, i)
					end
					i := i + 1
				end
				if l_quotient = Void then
					create l_quotient.make_child (Current)
					l_children.extend (l_quotient)
				end
				l_quotient.add (a_object)
			end
		end

	extend (a_linear: detachable LINEAR [G])
			-- Extend this node with all items from `a_linear'.
		do
			if a_linear /= Void then
				from
					a_linear.start
				until
					a_linear.off
				loop
					if attached a_linear.item as l_item then
						add (l_item)
					end
					a_linear.forth
				end
			end
		end

	extend_array (a_array: ARRAY [G])
			-- Extend this node with all items from `a_array'.
		local
			i: INTEGER
		do
			if a_array /= Void then
				from
					i := a_array.lower
				until
					i > a_array.upper
				loop
					if attached a_array.item (i) as l_item then
						add (l_item)
					end
					i := i + 1
				end
			end
		end

	extend_hash (a_hash: HASH_TABLE [G, HASHABLE])
			-- Extend this node with all items from 'a_hash'.
		do
			if a_hash /= Void then
				from
					a_hash.start
				until
					a_hash.off
				loop
					if a_hash.item_for_iteration /= Void then
						add (a_hash.item_for_iteration)
					end
					a_hash.forth
				end
			end
		end
	
feature -- Removal
	
	wipe_out
			-- Clear all items from this quotient.
		do
			children := Void
			objects := Void
			if not attached scale as l_scale or else l_scale.is_atomic then
				create objects.make_empty
				depth := 0
			else
				create children.make_empty
				depth := l_scale.depth
			end
			invariants.wipe_out
		end

feature {ROSE_QUOTIENT} -- Implementation

	fill_in_objects (a_array: ROSE_LINEAR_ARRAY [G]; a_node: like Current)
			-- Fill-in `a_array' with all objects from `a_node'.
		require
			valid_array: a_array /= Void
			valid_node: a_node /= Void
		local
			i: INTEGER
		do
			if attached a_node.objects as l_objects then
				from
					i := l_objects.lower
				until
					i > l_objects.upper
				loop
					a_array.extend (l_objects.item (i))
					i := i + 1
				end
			elseif attached a_node.children as l_children then
				from
					i := l_children.lower
				until
					i > l_children.upper
				loop
					fill_in_objects (a_array, l_children.item (i))
					i := i + 1
				end
			end
		end

invariant

	node_constraint: (children = Void) = (scale = Void)
	leaves_constraint: (objects /= Void) = (scale = Void)
	invariants_attached: invariants /= Void
	definition_attached: definition /= Void

end
