note

	description:

		"Composed partitioning criteria (sequence of atomic criterias)"

	author: "Zoran Simic"
	last_editor: "$Author: c4 $"
	date: "$Date: 2014/05/13 $"
	revision: "$Revision: #1 $"
	path: "$File: //rose/source/library/quotient/src/rose_quotient_composed_criteria.e $"

class	ROSE_QUOTIENT_COMPOSED_CRITERIA [G]

inherit

	ROSE_QUOTIENT_CRITERIA [G]

create

	make, make_empty

feature {NONE} -- Initialization

	make (a_criteria: ARRAY [ROSE_QUOTIENT_ATOMIC_CRITERIA [G]])
			-- Create a composed criteria from `a_criteria' of criteria.
		require
			valid_criteria: a_criteria /= Void and then not a_criteria.is_empty
		do
			criteria := a_criteria
		ensure
			criteria_set: criteria = a_criteria
		end

	make_empty
			-- Create an empty composed criteria.
		do
			create criteria.make (1, 0)
		end

feature -- Access

	criteria: ARRAY [ROSE_QUOTIENT_ATOMIC_CRITERIA [G]]
			-- Sequence of composed atomic criteria

	property (a_object: detachable G): detachable COMPARABLE
			-- First property for `a_object'
		do
			if not criteria.is_empty then
				Result := criteria.item (criteria.lower).property (a_object)
			end
		end

	properties (a_object: detachable G): ARRAY [detachable COMPARABLE]
			-- All properties in this criteria for `a_object'
		local
			i: INTEGER
		do
			from
				create Result.make (criteria.lower, criteria.upper)
				i := criteria.lower
			until
				i > criteria.upper
			loop
				Result.put (criteria.item (i).property (a_object), i)
				i := i + 1
			end
		end

	compared (a, b: G): INTEGER
			-- Compare `a' and `b', return -1 if a < b, 0 if equal, 1 if a > b
		local
			i: INTEGER
		do
			from
				i := criteria.lower
			until
				Result /= 0 or else i > criteria.upper
			loop
				Result := criteria.item (i).compared (a, b)
				i := i + 1
			end
		end

	invariant_compared (a_object: detachable G; a_invariant: ROSE_QUOTIENT_INVARIANT [G]): INTEGER
			-- Compare `a_object' to `a_invariant' accordingly to this criteria
			-- Return -1 if a_object < Current, 0 if equal, 1 otherwise
		local
			i: INTEGER
		do
			from
				i := criteria.lower
			until
				Result /= 0 or else i > criteria.upper
			loop
				Result := criteria.item (i).invariant_compared (a_object, a_invariant)
				i := i + 1
			end
		end

	names: ARRAY [STRING]
			-- Atomic criteria names composing this criteria
		local
			i: INTEGER
		do
			from
				create Result.make (criteria.lower, criteria.upper)
				i := criteria.lower
			until
				i > criteria.upper
			loop
				Result.put (criteria.item (i).name, i)
				i := i + 1
			end
		end

	has (a_criteria: ROSE_QUOTIENT_ATOMIC_CRITERIA [G]): BOOLEAN
			-- Is `a_criteria' part of this criteria?
		local
			i: INTEGER
		do
			from
				i := criteria.lower
			until
				Result or else i > criteria.upper
			loop
				Result := criteria.item (i).has (a_criteria)
				i := i + 1
			end
		end

	includes (a_other: ROSE_QUOTIENT_CRITERIA [G]): BOOLEAN
			-- Does this criteria include `a_other'?
		do
			if attached {ROSE_QUOTIENT_ATOMIC_CRITERIA [G]} a_other as l_atomic then
				Result := has (l_atomic)
			elseif attached {ROSE_QUOTIENT_COMPOSED_CRITERIA [G]} a_other as l_composed then
				Result := across l_composed.criteria as i_composed some has (i_composed.item) end
			end
		end

	is_empty: BOOLEAN
			-- Is this criteria empty?
		do
			Result := criteria.is_empty
		end

	string_representation (a_object: detachable G; a_separator: STRING): STRING
			-- Rpresentation of properties of `a_object' via this criteria;
			-- `a_separator' is used to separated multiple values
		local
			i: INTEGER
		do
			from
				create Result.make (criteria.count * 10)
				i := criteria.lower
			until
				i > criteria.upper
			loop
				Result.append (criteria.item (i).string_representation (a_object, a_separator))
				i := i + 1
				if i < criteria.upper then
					Result.append (a_separator)
				end
			end
		end

	out, debug_output: STRING
			-- New string containing terse printable representation of current object
		local
			i: INTEGER
		do
			from
				create Result.make (criteria.count*10)
				i := criteria.lower
			until
				i > criteria.upper
			loop
				Result.append (criteria.item (i).out)
				i := i + 1
				if i < criteria.upper then
					Result.append (", ")
				end
			end
		end

feature -- Basic operations

	extend (a_criteria: ROSE_QUOTIENT_ATOMIC_CRITERIA [G])
			-- Extend this order with `a_criteria'.
		require
			valid_criteria: not has (a_criteria)
		do
			criteria.force (a_criteria, criteria.upper + 1)
		end
	
invariant

	criteria_attached: criteria /= Void

end
