note

	description:

		"Definition of quotient criterias"

	author: "Zoran Simic"
	last_editor: "$Author: c4 $"
	date: "$Date: 2015/04/21 $"
	revision: "$Revision: #2 $"
	path: "$File: //rose/source/library/quotient/src/rose_quotient_definition.e $"

class	ROSE_QUOTIENT_DEFINITION [G]

create

	make_empty

feature {NONE} -- Initialization

	make_empty (a_id: STRING)
			-- Create an empty quotient definition (with no criteria).
		do
			id := a_id
			create all_criterias.make (1, 0)
			create by_name.make (0)
			create complete_criteria.make_empty
		end

feature -- Access

	id: STRING
			-- Quotient definition id

	all_criterias: ARRAY [ROSE_QUOTIENT_ATOMIC_CRITERIA [G]]
			-- All criteria composing this definition, in order they were added

	by_name: HASH_TABLE [ROSE_QUOTIENT_ATOMIC_CRITERIA [G], STRING]
			-- Criterias hashed by their symbolic name

	complete_criteria: ROSE_QUOTIENT_COMPOSED_CRITERIA [G]
			-- Sample criteria containing all possible atomic criteria, in order they were added

	count: INTEGER
			-- Number of criteria in this definition
		do
			Result := all_criterias.count
		end

	atomic (a_name: STRING): detachable ROSE_QUOTIENT_ATOMIC_CRITERIA [G]
			-- Criteria corresponding to `a_name', if any
		do
			if attached name_key (a_name) as l_key then
				if l_key.is_integer then
					Result := all_criterias.item (l_key.to_integer)
				else
					Result := by_name.item (l_key)
				end
			end
		end

	composed (a_criterias: STRING): detachable ROSE_QUOTIENT_COMPOSED_CRITERIA [G]
			-- Composed criteria with all `a_criterias'
			-- `a_criterias' contains a comma-separated list of atomic criteria
		local
			l_splitter: ST_SPLITTER
			l_list: DS_LIST [STRING]
			l_criteria_list: ARRAY [ROSE_QUOTIENT_ATOMIC_CRITERIA [G]]
			l_criteria: detachable ROSE_QUOTIENT_ATOMIC_CRITERIA [G]
		do
			create l_splitter.make_with_separators (",")
			l_list := l_splitter.split (a_criterias)
			if l_list /= Void then
				create l_criteria_list.make (1, 0)
				from
					l_list.start
				until
					l_list.off
				loop
					l_criteria := atomic (l_list.item_for_iteration)
					if l_criteria /= Void then
						l_criteria_list.force (l_criteria, l_criteria_list.upper + 1)
					end
					l_list.forth
				end
				if not l_criteria_list.is_empty then
					create Result.make (l_criteria_list)
				end
			end
		end

	scale (a_scale_representation: detachable STRING): detachable ROSE_QUOTIENT_SCALE [G]
			-- Scale represented by `a_scale_representation'
			-- `a_scale_representation' contains a semi-colon separated list of composed criteria
		local
			l_splitter: ST_SPLITTER
			l_list: DS_LIST [STRING]
		do
			if a_scale_representation /= Void then
				create l_splitter.make_with_separators (";")
				create Result.make_atomic (Void)
				from
					l_list := l_splitter.split (a_scale_representation)
					l_list.start
				until
					l_list.after
				loop
					if attached composed (l_list.item_for_iteration) as l_item then
						Result.add (l_item)
					end
					l_list.forth
				end
			end
		end

	property_by_name (a_object: detachable G; a_name: STRING): detachable COMPARABLE
			-- Property of critereon named `a_name' denoted by `a_object'
		require
			a_name_attached: a_name /= Void
		do
			if attached by_name.item (a_name) as l_atomic then
				Result := l_atomic.property (a_object)
			end
		end

feature -- Basic operations

	add (a_name: STRING; a_getter: FUNCTION [ANY, TUPLE [G], COMPARABLE])
			-- Add criteria with `a_name' and bound to `a_getter'.
		require
			a_name_attached: a_name /= Void
			a_getter_attached: a_getter /= Void
		local
			l_index: INTEGER
			l_criteria: ROSE_QUOTIENT_ATOMIC_CRITERIA [G]
		do
			l_index := all_criterias.upper + 1
			create l_criteria.make (l_index, a_name, a_getter)
			all_criterias.force (l_criteria, l_index)
			by_name.put (l_criteria, l_criteria.name)
			complete_criteria.extend (l_criteria)
		end

feature {NONE} -- Implementation

	name_key (a_name: detachable STRING): detachable STRING
			-- Lower-cased, trimmed, version of `a_name'
		do
			if a_name /= Void then
				Result := a_name.twin
				Result.left_adjust
				Result.right_adjust
				Result.to_lower
			end
		end
	
invariant

	all_criteria_attached: all_criterias /= Void
	by_name_attached: by_name /= Void
	complete_criteria_attached: complete_criteria /= Void
	
end
