note

	description:
	"[
		Common ancestor for all matrix classes without commitment to storage.
		Since we don't know whether the object can be modified, this class
		 provides no interface for modification.
	]"
	author: "Pierre Nallet"
	last_editor: "$Author: mb $"
	date: "$Date: 2014/05/07 $"
	revision: "$Revision: #106 $"
	path: "$File: //rose/source/kernel/infrastructure/math/abstract_readable_matrix.e $"

deferred class ABSTRACT_READABLE_MATRIX

inherit

	ROSE_IEEE_UNDEFINED

	DOUBLE_MATH

	APPROXIMATE_EQUALITY
		export {NONE} all end

	IEEE_MATH
		export {NONE} all end

	LABELED_2D

feature -- Dimension

	rows_interval: INTEGER_INTERVAL is
			-- Interval [1, height]
			-- (can be used for agent-based iteration)
		do
			create Result.make (1, height)
		ensure
			not_void: Result /= Void
			definition: Result.lower = 1 and Result.upper = height
		end

	columns_interval: INTEGER_INTERVAL is
			-- Interval [1, width]
			-- (can be used for agent-based iteration)
		do
			create Result.make (1, width)
		ensure
			not_void: Result /= Void
			definition: Result.lower = 1 and Result.upper = width
		end


	row_count : INTEGER is
			--Alias for height, should be removed
		do
			Result := height
		ensure
			row_count_matches_height: Result = height
		end

	column_count : INTEGER is
			-- Alias for width, should be removed
		do
			Result := width
		ensure
			column_count_matches_width: Result = width
		end

feature -- Status

	is_square: BOOLEAN is
			-- Is the matrix square?
		do
			Result := height = width
		ensure
			height_matches_width: Result = (height = width)
		end

	is_symmetric: BOOLEAN is
			-- Is `Current' a good approximation to a symmetric matrix?
		local
			i, j: INTEGER
		do
			if is_square then
				Result := True
				from i := 1 until i > height or not Result loop
					from j := 1 until j >= i or not Result loop
						Result := Result and approx_equal (Current [i, j], Current [j, i])
						j := j + 1
					end
					i := i + 1
				end
			end
		ensure
			symmetry_implies_square: Result implies is_square
		end

	is_identity: BOOLEAN is
			-- Is `Current' a good approximation to an identity matrix?
		local
			i, j: INTEGER
		do
			Result := is_square
			from i := 1 until i > height or not Result loop
				from j := 1 until j > width or not Result loop
					Result := Result and is_identity_item (i, j)
					j := j + 1
				end
				i := i + 1
			end
		ensure
			identity_implies_symmetry: Result implies is_symmetric
		end

feature {NONE} -- Implementation

	is_identity_item (i, j: INTEGER): BOOLEAN is
			-- Is item equal to zero or on diagonal and equal to 1, approximately?
		require
			valid_i: i >= 1 and i <= height
			valid_j: j >= 1 and j <= width
		do
			Result := (i = j implies approx_equal (Current [i, j], 1.0))
					and (i /= j implies approx_equal (Current [i, j] + 1.0, 1.0))
						-- comparing to 1+x with 1 instead of x with 0
						-- to force 1.0 to define the magnitude of zero
						-- (approx_equal (0, x) is False for all x)
		end

feature -- Status

	valid_row_index(a_ix : INTEGER) : BOOLEAN is
			-- Is `a_ix' a valid row index?
		do
			Result := a_ix >= 1 and a_ix <= height
		ensure
			index_between_1_and_height: Result = (a_ix >= 1 and then a_ix <= height)
		end

	valid_column_index(a_ix : INTEGER) : BOOLEAN is
			-- Is `a_ix' a valid column index?
		do
			Result := a_ix >= 1 and a_ix <= width
		ensure
			index_between_1_and_width: Result = (a_ix >= 1 and then a_ix <= width)
		end

	valid_row_flags : ROSE_BOOLEAN_ARRAY is
			-- Flags indicating whether each row is `is_all_defined'
		local
			i : INTEGER
		do
			create Result.make(1, height)
			from i := 1 until i > height loop
				if row(i).is_all_defined then
					Result[i] := True
				else
					Result[i] := False
				end
				i := i + 1
			end
		ensure
			not_void: Result /= Void
		end

	valid_column_flags : ROSE_BOOLEAN_ARRAY is
			-- Flags indicating whether each column `is_all_defined'
		local
			i : INTEGER
		do
			create Result.make(1, width)
			from i := 1 until i > width loop
				if column(i).is_all_defined then
					Result[i] := True
				else
					Result[i] := False
				end
				i := i + 1
			end
		ensure
			not_void: Result /= Void
		end

	valid_row_indice : ROSE_LINEAR_ARRAY[INTEGER] is
			-- Indices of rows for which `is_all_defined' applies.
		local
			i : INTEGER
		do
			create Result.make_empty
			from i := 1 until i > height loop
				if row(i).is_all_defined then
					Result.extend(i)
				end
				i := i + 1
			end
		ensure
			not_void: Result /= Void
			--all_rows_defined: Result.for_all (agent (i: INTEGER): BOOLEAN is do Result := row(i).is_all_defined end)
		end

	non_null_row_indice : ROSE_LINEAR_ARRAY[INTEGER] is
			-- Indices of non null rows
			-- (as defined by `is_null_row')
		local
			i : INTEGER
		do
			create Result.make_empty
			from i := 1 until i > height loop
				if not is_null_row(i) then
					Result.extend(i)
				end
				i := i + 1
			end
		ensure
			not_void: Result /= Void
			--all_rows_not_null: Result.for_all (agent (i: INTEGER): BOOLEAN is do Result := not is_null_row (i) end)
		end

	non_constant_row_indice : ROSE_LINEAR_ARRAY[INTEGER] is
			-- Indices of non-constant rows
			-- (as defined by `is_constant_row')
		local
			i : INTEGER
		do
			create Result.make_empty
			from i := 1 until i > height loop
				if not is_constant_row (i) then
					Result.extend(i)
				end
				i := i + 1
			end
		ensure
			not_void: Result /= Void
			all_indices: Result.for_all (agent valid_row_index)
			--all_rows_not_constant: Result.for_all (agent (i: INTEGER): BOOLEAN is do Result := not is_constant_row (i) end)
		end

	is_null_row (i: INTEGER): BOOLEAN is
		-- Does this row only contain undefined values?
		local
			l_ix: INTEGER
			l_double: DOUBLE
		do
			if valid_row_index (i) then
				Result := True
				from l_ix := 1 until l_ix > width or Result = False loop
					l_double := Current [i, l_ix]
					if is_defined_double(l_double) then
						Result := False
					end
					l_ix := l_ix + 1
				end
			else
				Result := False
			end
		ensure
			valid_row_index: Result implies valid_row_index (i)
		end

	is_null_column(i : INTEGER) : BOOLEAN is
		--does this column only contain undefined values?
		local
			l_ix : INTEGER
			l_double : DOUBLE
		do
			if valid_column_index(i) then
				Result := True
				from l_ix := 1 until l_ix > height or Result = False loop
					l_double := Current[l_ix, i]
					if is_defined_double(l_double) then
						Result := False
					end
					l_ix := l_ix + 1
				end
			else
				Result := False
			end
		end

	is_null : BOOLEAN is
		--does this matrix only contain undefined values?
		local
			l_ix : INTEGER
		do
			Result := True
			from l_ix := 1 until l_ix > height or (not Result) loop
				Result := Result and is_null_row(l_ix)
				l_ix := l_ix + 1
			end
		end

	is_constant_row (a_row: INTEGER): BOOLEAN is
			-- Are all values in this row (approximate) equal?
		require
			a_row_index: valid_row_index (a_row)
		local
			i: INTEGER
			a_first: DOUBLE
		do
			Result := True
			if width > 1 then
				from
					i := 2
					a_first := Current [a_row, 1]
				until
					i > width or not Result
				loop
					Result := approx_equal (a_first, Current [a_row, i])
					i := i + 1
				end
			end
		ensure
			zero_or_one_rows: width < 2 implies Result
		end

	is_constant_column (a_column: INTEGER): BOOLEAN is
			-- Are all values in this column (approximate) equal?
		require
			a_column_index: valid_column_index (a_column)
		local
			i: INTEGER
			a_first: DOUBLE
		do
			Result := True
			if height > 1 then
				from
					i := 2
					a_first := Current [1, a_column]
				until
					i > height or not Result
				loop
					Result := approx_equal (a_first, Current [i, a_column])
					i := i + 1
				end
			end
		ensure
			zero_or_one_columns: height < 2 implies Result
		end

	is_defined (a_is_defined:FUNCTION [ANY, TUPLE [ABSTRACT_READABLE_MATRIX, INTEGER, INTEGER], BOOLEAN];i,j : INTEGER) : BOOLEAN is
		--is the i,j th element defined according to a_is_defined?
		do
			Result := a_is_defined.item([Current,i,j])
		ensure
			correctly_defined: Result = a_is_defined.item([Current,i,j])
		end

feature -- Indices

	first_valid_row_index: INTEGER is
		-- Index of first non-null row, or `height' + 1 if there is no non-null row
		do
			from
				Result := 1
			variant
				increasing_index: height + 1 - Result
			until
				Result > height or not is_null_row (Result)
			loop
				Result := Result + 1
			end
		ensure
			definition: valid_row_index (Result) xor Result = height + 1
		end

	last_valid_row_index: INTEGER is
			-- Index of last non-null row, or 0 if there is no non-null row
		do
			from
				Result := height
			variant
				decreasing_result: Result
			until
				Result < 1 or not is_null_row (Result)
			loop
				Result := Result - 1
			end
		ensure
			definition: valid_row_index (Result) xor Result = 0
		end

	indices_of_row_max : INTEGER_VECTOR is
			-- For each row, index of maximum element
		local
			i : INTEGER
		do
			create Result.make(1, height)
			from i := 1 until i > height loop
				Result.put(row(i).index_of_max, i)
				i := i + 1
			end
		ensure
			not_void: Result /= Void
		end

	indices_of_row_min : INTEGER_VECTOR is
			-- For each row, index of minimum element
		local
			i : INTEGER
		do
			create Result.make(1, height)
			from i := 1 until i > height loop
				Result.put(row(i).index_of_min, i)
				i := i + 1
			end
		ensure
			not_void: Result /= Void
		end

feature -- Scalar queries

	max_value: DOUBLE is
			-- Maximum value
		local
			i, j: INTEGER
			l_item: DOUBLE
		do
			Result := Current[1,1]
			from i := 1 until i > height loop
				from j := 1 until j > width loop
					l_item := Current[i,j]
					if l_item > Result then
						Result := l_item
					end
					j := j + 1
				end
				i := i + 1
			end
		end

	min_value: DOUBLE is
			-- Minimum value
		local
			i, j: INTEGER
			l_item: DOUBLE
		do
			Result := Current [1,1]
			from i := 1 until i > height loop
				from j := 1 until j > width loop
					l_item := Current [i,j]
					if l_item < Result then
						Result := l_item
					end
					j := j + 1
				end
				i := i + 1
			end
		end

	max_abs_value: DOUBLE is
			-- Maximum absolute value
		local
			i, j: INTEGER
			l_item: DOUBLE
		do
			Result := item (1,1).abs
			from i := 1 until i > height loop
				from j := 1 until j > width loop
					l_item := item (i,j).abs
					if l_item > Result then
						Result := l_item
					end
					j := j + 1
				end
				i := i + 1
			end
		ensure
			non_negative: Result >= 0
		end

	min_abs_value: DOUBLE is
			-- Minimum absolute value
		local
			i, j: INTEGER
			l_item: DOUBLE
		do
			Result := item (1,1).abs
			from i := 1 until i > height loop
				from j := 1 until j > width loop
					l_item := item (i,j).abs
					if l_item < Result then
						Result := l_item
					end
					j := j + 1
				end
				i := i + 1
			end
		ensure
			non_negative: Result >= 0
		end

	trace : DOUBLE is
			-- Sum of diagonal items
		require
			square : is_square
		local
			i : INTEGER
		do
			if is_square then
				from i := 1 until i > width loop
					Result := Result + Current[i, i]
					i := i + 1
				end
			end
		end

	column_sum (j : INTEGER) : DOUBLE is
			-- Sum of column items
		local
			i : INTEGER
		do
			from i := 1 until i > height loop
				Result := Result + Current[i,j]
				i := i + 1
			end
		end

	column_minimum (a_col: INTEGER): DOUBLE is
			-- Minimum value in column `a_col'
		require
			a_col_valid: valid_column_index (a_col)
		local
			i: INTEGER
		do
			from i := 1 until i > height loop
				if i = 1 then
					Result := Current [i, a_col]
				else
					Result := Result.min (Current [i, a_col])
				end
				i := i + 1
			end
		end

	column_maximum (a_col: INTEGER): DOUBLE is
			-- Maximum value in column `a_col'
		require
			a_col_valid: valid_column_index (a_col)
		local
			i: INTEGER
		do
			from i := 1 until i > height loop
				if i = 1 then
					Result := Current [i, a_col]
				else
					Result := Result.max (Current [ i, a_col])
				end
				i := i + 1
			end
		end

	row_minimum (a_row: INTEGER): DOUBLE is
			-- Minimum value in row `a_row'
		require
			a_row_valid: valid_row_index (a_row)
		local
			j: INTEGER
		do
			from j := 1 until j > width loop
				if j = 1 then
					Result := Current [a_row, j]
				else
					Result := Result.min (Current [a_row, j])
				end
				j := j + 1
			end
		end

	row_maximum (a_row: INTEGER): DOUBLE is
			-- Maximum value in row `a_row'
		require
			a_row_valid: valid_row_index (a_row)
		local
			j: INTEGER
		do
			from j := 1 until j > width loop
				if j = 1 then
					Result := Current [a_row, j]
				else
					Result := Result.max (Current [a_row, j])
				end
				j := j + 1
			end
		end

	row_sum (i : INTEGER) : DOUBLE is
			-- Sum of row items
		local
			j : INTEGER
		do
			from j := 1 until j > width loop
				Result := Result + Current[i,j]
				j := j + 1
			end
		end

	vector_norm (a_vector: ABSTRACT_READABLE_VECTOR): DOUBLE is
			-- `a_vector' * transpose @ `a_vector'
		require
			is_square: a_vector /= Void and then a_vector.count = width
		local
			l_vector: VECTOR
		do
			l_vector := transpose_at (a_vector.as_vector)
			Result := l_vector.cross(a_vector).sum
		ensure
			definition: Result = a_vector * transpose_at (a_vector.as_vector)
		end

	aggregate (a_agent : FUNCTION[ANY,TUPLE[DOUBLE,DOUBLE],DOUBLE]) : DOUBLE is
			-- For each element (by row and column), apply `a_agent', starting with 0.0. `a_agent' takes
			-- the intermediate result as first argument, and the MATRIX item as second argument. The
			-- result is the return value after the last application.
			-- Example: f(x, y) = x + y will yield the sum of all elements in the MATRIX.
		local
			i,j : INTEGER
		do
			from i := 1 until i > height loop
				from j := 1 until j > width loop
					if i = 1 and j = 1 then
						Result := item(i,j)
					else
						Result := a_agent.item([Result,item(i,j)])
					end
					j := j + 1
				end
				i := i + 1
			end
		end

feature -- Derived matrices

	to_matrix : MATRIX is
			-- Writable version of `Current', always a copy
		local
			i, j : INTEGER
		do
			create Result.make(height, width)
			from i := 1 until i > height loop
				from j := 1 until j > width loop
				--	Result.put(item(i,j), i, j)
					Result[i, j] := Current[i, j]
					j := j + 1
				end
				i := i + 1
			end
			copy_names(Result)
		ensure
			not_void: Result /= Void
			different : Result /= Current
			same: is_same (Result)
		end

	as_matrix : MATRIX is
			-- Writable version of `Current', might not be a copy
			-- This will be redefined in MATRIX to just return `Current', which will then make it different to `to_matrix'.
		do
			Result := to_matrix
		end

	to_coordinated_matrix(a_is_flip : BOOLEAN) : MATRIX is
			-- convert to a three column matrix with (i, j, value) for each row
		local
			i, j, l_row : INTEGER
		do
			create Result.make(height * width, 3)
			from i := 1 until i > height loop
				from j := 1 until j > width loop
					l_row := l_row + 1
					if a_is_flip then
						Result[l_row, 2] := height - i + 1
					else
						Result[l_row, 2] := i
					end
					Result[l_row, 1] := j
					Result[l_row, 3] := Current[i, j]
					j := j + 1
				end
				i := i + 1
			end
		end

	to_correlation: SYMMETRIC_MATRIX is
		-- require semi positive definite matrix (like covariance matrix)
		require
			symmetric: height = width
		local
			i, j: INTEGER
			l_ij, l_ii, l_jj: DOUBLE
		do
			create Result.make (width)
			from i := 1 until i > width loop
				from j := 1 until j > i loop
				--	Result.put (item (i, j) / (sqrt (item (i, i) * item (j, j))), i, j)
					l_ii := Current[i, i]
					l_jj := Current[j, j]
					l_ij := Current[i, j]
					if is_defined_double (l_ii) and then is_defined_double (l_jj) and then is_defined_double (l_ij) and then l_ii > 0. and then l_jj > 0. then
						Result[i, j] := l_ij/sqrt (l_ii * l_jj)
					else
						Result[i, j] := undefined_double
					end
					j := j + 1
				end
			i := i + 1
			end
			if column_names/=void then
				Result.set_column_names(column_names)
				Result.set_row_names(column_names)
			end
		ensure
			not_void: Result /= Void
		end

	extend_matrix (a_matrix : ABSTRACT_READABLE_MATRIX): MATRIX is
			-- Extend the width of a matrix with new columns
			require
				other_not_void : a_matrix /= Void
				same_height: height = a_matrix.height
			local
				i: INTEGER --,k:INTEGER
				l_v1,l_v2:VECTOR
				l_column_names : like column_names
			do
				create result.make(height,width+a_matrix.width)
				from i:=1 until i>height loop
					l_v1:=row(i)
					l_v2:=a_matrix.row(i)
					l_v1.extend_vector(l_v2)
					result.put_row_vector(l_v1,i)
					i:=i+1
				end
				Result.set_row_names(row_names)
				if  column_names /= Void and then a_matrix.column_names /= Void then
					l_column_names := clone(column_names)
					l_column_names.append_array(a_matrix.column_names)
					Result.set_column_names(l_column_names)
				end
			end

	extend_column_vectors(a_vectors : ARRAY[ABSTRACT_READABLE_VECTOR]) : MATRIX is
			-- extend column vectors
		local
			i, l_lower : INTEGER
			l_column_names : like column_names
		do
			create Result.make(height, width + a_vectors.count)
			Result.put_matrix(Current, 1, 1)
			l_lower := a_vectors.lower
			from i := 1 until i > a_vectors.count loop
				Result.put_column_vector(a_vectors.item(l_lower + i - 1), width + i)
				i := i + 1
			end
			Result.set_row_names(row_names)
			if  column_names /= Void then
				l_column_names := clone(column_names)
				from i := 1 until i > a_vectors.count loop
					if a_vectors.item(i).name /= Void then
						l_column_names.extend(a_vectors.item(l_lower + i - 1).name)
					else
						l_column_names.extend("column_" + (width + i).out)
					end
					i := i + 1
				end
				Result.set_column_names(l_column_names)
			end
		end

	extend_row_vectors (a_vectors: ARRAY [VECTOR]): MATRIX
			-- Copy of `Current' with `a_vectors' added at the bottom.
		require
			a_vectors_attached: a_vectors /= Void
			all_vectors_attached: across a_vectors as i_vector all i_vector.item /= Void end
		local
			i, l_lower : INTEGER
			l_row_names : like row_names
		do
			create Result.make(height + a_vectors.count, width)
			Result.put_matrix(Current, 1, 1)
			l_lower := a_vectors.lower
			from i := 1 until i > a_vectors.count loop
				Result.put_row_vector(a_vectors.item(l_lower + i - 1), height + i)
				i := i + 1
			end
			Result.set_column_names(column_names)
			if  row_names /= Void then
				l_row_names := clone(row_names)
				from i := 1 until i > a_vectors.count loop
					if a_vectors.item(i).name /= Void then
						l_row_names.extend(a_vectors.item(l_lower + i - 1).name)
					else
						l_row_names.extend("column_" + (width + i).out)
					end
					i := i + 1
				end
				Result.set_row_names(l_row_names)
			end
		end

	sorted_by_column (a_col_index:INTEGER; a_descending :BOOLEAN): MATRIX
			-- Matrix like `Current' but sorted by column `a_col_index' with order determined by `a_descending'
		local
			l_order: ROSE_LINEAR_ARRAY [INTEGER]
			l_column: VECTOR
			i, k: INTEGER
			l_row_names: ROSE_LINEAR_ARRAY [STRING]
		do
			l_column := column (a_col_index)
			if a_descending then
				l_order := l_column.descending_order
			else
				l_order := l_column.ascending_order
			end
			if row_names /= Void then
				create l_row_names.make (1, row_names.count)
			end
			create Result.make (height, width)
			from
				i := 1
			until
				i > height
			loop
				k := l_order.item (i)
				Result.put_row_vector (row (k), i)
				if l_row_names /= Void then
					l_row_names.put (row_names.item (k), i)
				end
				i := i + 1
			end
			Result.set_column_names (column_names)
			Result.set_row_names (l_row_names)
			if name /= Void then
				Result.set_name (name)
			end
		end

	pairwise_column_diff_matrix : MATRIX is
			-- pairwise column diff matrix
		require
			even_number_of_columns : width \\ 2 = 0
		local
			i, l_num_pair : INTEGER
			l_vec1, l_vec2, l_change_vector, l_ratio_vector : VECTOR
			l_column_names : ROSE_LINEAR_ARRAY[STRING]
		do
			l_num_pair := width // 2
			create Result.make(height, l_num_pair * 4)
			create l_column_names.make(1, l_num_pair * 4)
			from i := 1 until i > l_num_pair loop
				l_vec1 := column(2 * i - 1)
				l_vec2 := column(2 * i)
				l_change_vector := (l_vec1 - l_vec2).abs
				l_ratio_vector := l_vec1.diff_ratio_vector (l_vec2)
				Result.put_column_vector(l_vec1, 4 * i - 3)
				Result.put_column_vector(l_vec2, 4 * i - 2)
				Result.put_column_vector(l_change_vector, 4 * i - 1)
				Result.put_column_vector(l_ratio_vector, 4 * i)

				if column_names /= Void then
					l_column_names.put(column_names.item(2 * i - 1) , 4 * i - 3)
					l_column_names.put(column_names.item(2 * i) , 4 * i - 2)
				else
					l_column_names.put("1-" + "Var_" + i.out, 4 * i - 3)
					l_column_names.put("2-" + "Var_" + i.out , 4 * i - 2)
				end
				l_column_names.put("Abs_Change", 4 * i - 1)
				l_column_names.put("Check_Ratio" , 4 * i)
				i := i + 1
			end
			Result.set_name(name)
			Result.set_row_names(row_names)
			Result.set_column_names(l_column_names)
		end

	sub_matrix (r: ARRAY[INTEGER]; j0, j1: INTEGER): MATRIX is
			-- Submatrix composed of columns within [j0, j1] and rows in `r'
		require
			r_not_void: r /= Void
			valid_columns: valid_column_index (j0) and valid_column_index (j1)
		local
			i, j: INTEGER
		do
			create Result.make (r.count, j1 - j0 + 1)
			from i := 1 until i > r.count loop
				from j := j0 until j > j1 loop
					if valid_row_index (r.item(i)) then
					--	Result.put (Current [r [i], j], i, j - j0 + 1)
						Result [i, j - j0 + 1] := Current [r [i], j]
					end
					j := j + 1
				end
				i := i + 1
			end
		ensure
			not_void: Result /= Void
		end

	sub_matrix_c (i0, i1: INTEGER; c: ARRAY[INTEGER]): MATRIX is
			-- Submatrix composed of rows within [i0, i1] and columns in `c'
		require
			c_not_void: c /= Void
			valid_rows: valid_row_index (i0) and valid_row_index (i1)
		local
			i, j: INTEGER
		do
			create Result.make (i1 - i0 + 1, c.count)
			from j := 1 until j > c.count loop
				from i := i0 until i > i1 loop
					if valid_column_index (c[j + c.lower - 1 ]) then
						Result [i - i0 + 1, j] := Current [i, c [j + c.lower - 1]]
					end
					i := i + 1
				end
				j := j + 1
			end
		ensure
			not_void: Result /= Void
		end

	sub_matrix_2 (i0, i1, j0, j1: INTEGER): MATRIX is
			-- Submatrix composed of rows within [i0,i1] and columns within [j0,j1]
		local
			i, j: INTEGER
		do
			create Result.make (i1 - i0 + 1, j1 - j0 + 1)
			from i := i0 until i > i1 loop
				from j := j0 until j > j1 loop
				--	Result.put (item (i, j), i - i0 + 1, j - j0 + 1)
					Result[i - i0 + 1, j - j0 + 1] := Current [i, j]
				j := j + 1
				end
			i := i + 1
			end
		ensure
			not_void: Result /= Void
		end

	new_matrix_anchor: MATRIX is
			-- Anchor type for newly created matrices
		 do end

	sub_matrix_3_obsolete(r,c : ARRAY[INTEGER]) : like new_matrix_anchor is
			-- Sub matrix
		local
			i, j,k,m: INTEGER
		do
			create Result.make (r.count, c.count)
			from i := 1 until i > r.count loop
				from j := 1 until j > c.count loop
					k:=r.item (i)
					m:=c.item(j)
					if valid_row_index(k) and valid_column_index(m) then
					--	Result.put (item (k, m), i, j)
						Result[i, j] := Current[k, m]
					else
--						Result.put(undefined_double,i,j)
						Result[i, j] := undefined_double
		end
				j := j + 1
				end
			i := i + 1
			end
		ensure
			not_void: Result /= Void
		end

	selected_columns(c: ARRAY[INTEGER]): MATRIX is
		--matrix comprised of the columns in c
		local
			l_row_index:ROSE_LINEAR_ARRAY[INTEGER]
			i: INTEGER --, j: INTEGER
		do
			create l_row_index.make(1,height)
			from i:=1 until i> height loop
				l_row_index.put(i,i)
				i:=i+1
			end
			--Result:=sub_matrix_3(l_row_index,c)
			create Result.make_sub_matrix_3 (Current, l_row_index, c)
		  	Result.set_row_names(row_names)
		  	if column_names /= Void then
		  		Result.set_column_names(column_names.selected_items(c))
		  	end
		end

	selected_rows (l_rows : ARRAY[INTEGER]): MATRIX is
	  	-- selected rows
	  local
		 l_height,i: INTEGER --,j : INTEGER
	  do
		  	l_height:=l_rows.count
		  	create Result.make(l_height,width)
		  	from i:=1 until i> l_height loop
				Result.put_row_vector(row(l_rows.item(i)),i)
		  		i:=i+1
		  	end
		  	Result.set_column_names(column_names)
		  	if row_names /= Void then
		  		Result.set_row_names(row_names.selected_items(l_rows))
		  	end
	  end

	 top_rows(a_rows:INTEGER): MATRIX is
		 	-- top rows
		 local
		 	l_array:ARRAY[INTEGER]
		 	i:INTEGER
		 do
		 	create l_array.make(1,a_rows)
		 	from i:=1 until i>a_rows loop
		 		l_array.put(i,i)
		 		i:=i+1
		 	end
		 	result:=selected_rows(l_array)
		 end

	bottom_rows(a_rows:INTEGER): MATRIX is
		 	-- bottom rows
		 local
		 	l_array:ARRAY[INTEGER]
		 	i:INTEGER
		 do
		 	create l_array.make(1,a_rows)
		 	from i:=1 until i>a_rows loop
		 		l_array.put(height-a_rows+i,i)
		 		i:=i+1
		 	end
		 	result:=selected_rows(l_array)
		 end

	matrix_with_valid_rows : MATRIX is
			-- matrix with valid rows
		local
			l_ixs : ROSE_LINEAR_ARRAY[INTEGER]
		do
			l_ixs := valid_row_indice
			Result := selected_rows(l_ixs)
		end

	itemwise_inverse: MATRIX is
			-- item by item inverse such that an itemwise_product of A and A_itemwise_inverse would
			-- be a marix of ones.
			-- can be no zero elements
		local
			i, j: INTEGER
		do
			create Result.make(height, width)
			from i:= 1 until i> height loop
				from j:= 1 until j> width loop
				--	Result.put(1/ Current[i, j], i, j)
					Result [i, j] := 1 / Current [i, j]
					j:= j+ 1
					end
				i:= i+ 1
			end
		end

	normalized (a_is_defined:FUNCTION [ANY, TUPLE [ABSTRACT_READABLE_MATRIX, INTEGER, INTEGER], BOOLEAN]): MATRIX is
			-- normalized
		local
			m: VECTOR
			l_std: VECTOR
			i, j: INTEGER
		do
			m := column_mean_with_unavailables (void, a_is_defined)
			l_std := column_standard_deviation_with_unavailables (void,a_is_defined)
			create Result.make (height, width)
			from i := 1 until i > height loop
				from j := 1 until j > width loop
					if (a_is_defined.item ([Current, i, j])) then
					--	Result.put ((item (i, j) - m.item(j)) / l_std.item (j), i, j)
						Result[i, j] := (Current [i, j] - m[j]) / l_std[j]
					else
--						Result.put (item (i, j), i, j)
						Result[i, j] := Current[i, j]
					end
					j := j + 1
				end
				i := i + 1
			end
			copy_names(Result)
		end

	squared_norm : MATRIX is
			-- Squared normalized version of `Current'
		do
	 		Result := operations.squared_norm (Current)
		end

	covariance (a_weights: VECTOR; a_means: VECTOR) : SYMMETRIC_MATRIX is
		-- variance/covariance of column data series
		local
			d: DOUBLE
			i, j, k, l_count: INTEGER
			l_weight_sum, l_weight , l_item_i, l_item_j: DOUBLE
			l_means,l_weights : ABSTRACT_READABLE_VECTOR
		do
			l_weights := safe_row_weights(a_weights)
--			l_weight_sum := l_weights.sum
			if (a_means /= Void) then
				l_means := a_means
			else
				l_means := column_mean (a_weights)
			end

			create Result.make (width)
			from i := 1 until i > width loop
				from j := 1 until j > i loop
					d := 0
					l_weight_sum := 0.0
					l_count := 0
					from k := 1 until k > height loop
						l_weight := l_weights [k]
						l_item_i := Current [k, i]
						l_item_j := Current [k, j]
						if is_defined_double (l_item_i) and then is_defined_double (l_item_j) then
							d := d + (l_item_i - l_means[i]) * (l_item_j - l_means [j]) * l_weight
							l_weight_sum := l_weight_sum + l_weight
							l_count := l_count + 1
						end
						k := k + 1
					end
				--	Result.put (d / (l_weight_sum - 1), i, j)
					if l_weight_sum /= 0.0 and then l_count > 1 then
						Result [i, j] := d / (l_weight_sum  - 1)
					end
					j := j + 1
				end
				i := i + 1
			end
			if column_names/=void then
				Result.set_name("Covariance Matrix")
				Result.set_column_names(column_names)
				Result.set_row_names(column_names)
			end
		end

	correlation: SYMMETRIC_MATRIX is
			-- equavelent to only considering defined double when calculate correlation
		local
			l_non_null: MATRIX
			l_cov: SYMMETRIC_MATRIX
		do
			l_non_null := null_replaced
			l_cov := l_non_null.covariance(Void, Void)
			Result := l_cov.to_correlation
			if column_names/=void then
				Result.set_name("Correlation")
				Result.set_column_names(column_names)
				Result.set_row_names(column_names)
			end
		end

	correlation_matrix : SYMMETRIC_MATRIX is
		-- caution: the correlation are not guaranteed to be less than 1.
		local
			l_cov : SYMMETRIC_MATRIX
		do
			l_cov := covariance(Void, Void)
			Result := l_cov.to_correlation
			if column_names/=void then
				Result.set_name("Correlation Matrix")
				Result.set_column_names(column_names)
				Result.set_row_names(column_names)
			end
		end

	column_acfs(a_max_lag:INTEGER) : MATRIX is
			--column time series stats	
			local
				l_vector : VECTOR
				i:INTEGER
				l_names:ROSE_LINEAR_ARRAY[STRING]
			do
				create result.make(a_max_lag,width)
				create l_names.make_empty
				from i:=1 until i>width loop
					l_vector := column(i)
					result.put_column_vector(l_vector.acf(a_max_lag),i)
					i:=i+1
				end
				from i:=1 until i>a_max_lag loop
					l_names.extend("lag_"+i.out)
					i:=i+1
				end
				result.set_column_names(column_names)
				result.set_row_names(l_names)
				result.set_name("column ACFs")
			end

	column_pacfs(a_max_lag:INTEGER) : MATRIX is
			--column time series stats
			local
				l_vector : VECTOR
				i:INTEGER
				l_names:ROSE_LINEAR_ARRAY[STRING]
			do
				create result.make(a_max_lag,width)
				create l_names.make_empty
				from i:=1 until i>width loop
					l_vector := column(i)
					result.put_column_vector(l_vector.pacf(a_max_lag),i)
					i:=i+1
				end
				from i:=1 until i>a_max_lag loop
					l_names.extend("lag_"+i.out)
					i:=i+1
				end
				result.set_column_names(column_names)
				result.set_row_names(l_names)
				result.set_name("column PACFs")
			end

	null_replaced : MATRIX is
			-- replace null value by zero
		local
			l_v: DOUBLE
			i,j: INTEGER
		do
			create Result.make (height, width)
			from i := 1 until i > height loop
				from j := 1 until j > width loop
					l_v := Current [i, j]
					if is_defined_double (l_v) then
						Result [i, j] := l_v
					else
						Result [i, j] := 0.
					end
					j := j + 1
				end
				i := i + 1
			end
		end

	filtered_rows_as_matrix (a_valid_function: FUNCTION [ANY, TUPLE [INTEGER, ABSTRACT_READABLE_MATRIX], BOOLEAN]): MATRIX is
			-- `Current' filtered by `a_valid_function'
		require
			available_function: a_valid_function /= Void
		local
			i: INTEGER
			l_valid_rows: ROSE_LINEAR_ARRAY [VECTOR]
			l_valid_names: ROSE_LINEAR_ARRAY [STRING]
		do
			create l_valid_rows.make_empty
			create l_valid_names.make_empty
			from
				i := 1
			until
				i > height
			loop
				if a_valid_function.item ([i, Current]) then
					l_valid_rows.extend (row (i))
					l_valid_names.extend (row_names [i])
				end
				i := i + 1
			end
			if not l_valid_rows.is_empty then
				create Result.make_from_array_of_vector (l_valid_rows)
				Result.set_row_names (l_valid_names)
			else
				create Result.make (0, column_names.count)
			end
			Result.set_column_names (column_names.twin)
		ensure
			result_valid: attached Result
			same_width: Result.width = width
			subset_height: Result.height <= height
			same_column_names: Result.column_names ~ column_names
			subset_row_names: (create {HASHED_SET[STRING]}.make_by_array (row_names)).is_superset(create {HASHED_SET[STRING]}.make_by_array (Result.row_names))
		end

feature --derived vectors

	to_vector : VECTOR is
		-- concatenate the columns into a vector
		local
			i, j, l_count : INTEGER
			l_names:ROSE_LINEAR_ARRAY[STRING]
			l_col_name, l_row_name: STRING
		do
			create Result.make(height * width)
			from i := 1 until i > width loop
				from j := 1 until j > height loop
					l_count := l_count + 1
					Result.put(Current[j, i], l_count)
					j := j + 1
				end
				i := i + 1
			end
			if column_names/=void and then row_names/=void then
				create l_names.make(1,width*height)
				l_count:=0
				from i := 1 until i > width loop
					from j := 1 until j > height loop
						l_count := l_count + 1
						if column_names.count >= i then
							l_col_name := column_names.item (i)
						else
							l_col_name := " "
						end
						if row_names.count >= j then
							l_row_name := row_names.item (j)
						else
							l_row_name := " "
						end
						l_names.put(l_col_name + "_" + l_row_name, l_count)
						j := j + 1
					end
					i := i + 1
				end
				Result.set_names (l_names)
			end
		ensure
			not_void: Result /= Void
			size: Result.count = height * width
		end

	sub_row(a_row : INTEGER; a_subset : ARRAY[INTEGER]) : VECTOR is
		local
			i, l_ix : INTEGER
		do
			create Result.make(a_subset.count)
			from i := a_subset.lower until i > a_subset.upper loop
				l_ix := a_subset.item(i)
				if valid_column_index (l_ix) then
					Result.put(Current[a_row, l_ix], i - a_subset.lower + 1)
				end
				i := i + 1
			end
		ensure
			not_void: Result /= Void
		end

	sub_column(a_col : INTEGER; a_subset : ARRAY[INTEGER]) : VECTOR is
		local
			i, l_ix : INTEGER
		do
			create Result.make(a_subset.count)
			from i := a_subset.lower until i > a_subset.upper loop
				l_ix := a_subset.item(i)
				if valid_row_index (l_ix) then
					Result.put(Current[l_ix, a_col], i - a_subset.lower + 1)
				end
				i := i + 1
			end
		ensure
			not_void: Result /= Void
		end

	row (i: INTEGER): VECTOR is
			-- I-th row as a vector
		local
			j: INTEGER
		do
			create Result.make (width)
			from j := 1 until j > width loop
				Result.put (Current [i, j], j)
				j := j + 1
			end
			Result.set_names (column_names)
			if row_names /= Void and then row_names.valid_index (i) then
				Result.set_name (row_names [i])
			end
		end

	rows_array : ROSE_LINEAR_ARRAY[VECTOR] is
			-- array of rows
		local
			i:INTEGER
			l_row:VECTOR
		do
			create result.make(1,height)
			from i:=1 until i>height loop
				l_row:=row(i)
				l_row.set_names (column_names)
				if row_names/=void then
					l_row.set_name (row_names.item(i))
				end
				result.put(l_row,i)
				i:=i+1
			end
		end

	column (j: INTEGER): VECTOR is
			-- J-th column as a vector
		local
			i: INTEGER
		do
			create Result.make (height)
			from i := 1 until i > height loop
				Result.put (Current [i, j], i)
				i := i + 1
			end
			if column_names/=void then
				result.set_name (column_names.item (j))
			end
			Result.set_names (row_names)
		ensure then
			correct_column: column_matches (Result, j)
		end

	column_sums : VECTOR is
			-- Sum of each column
		local
			j : INTEGER
		do
			create Result.make(width)
			from j := 1 until j > width loop
				Result.put(column_sum(j), j)
				j := j + 1
			end
		ensure
			not_void: Result /= Void
			correct_size: Result.count = width
		end

	column_minima: VECTOR is
			-- Minimum value for each column
		local
			j: INTEGER
		do
			create Result.make (width)
			from j := 1 until j > width loop
				Result.put( column_minimum (j), j)
				j := j + 1
			end
		ensure
			column_minima_not_void: Result /= Void
			correct_size: Result.count = width
		end

	column_maxima: VECTOR is
			-- Maximum value for each column
		local
			j: INTEGER
		do
			create Result.make (width)
			from j := 1 until j > width loop
				Result.put (column_maximum (j), j)
				j := j + 1
			end
		ensure
			column_maxima_not_void: Result /= Void
			correct_size: Result.count = width
		end

	row_minima: VECTOR is
			-- Minimum value for each row
		local
			i: INTEGER
		do
			create Result.make (height)
			from i := 1 until i > height loop
				Result.put (row_minimum (i), i)
				i := i + 1
			end
		ensure
			row_minima_not_void: Result /= Void
			correct_size: Result.count = height
		end

	row_maxima: VECTOR is
			-- Maximum value for each row
		local
			i: INTEGER
		do
			create Result.make (height)
			from i := 1 until i > height loop
				Result.put (row_maximum (i), i)
				i := i + 1
			end
		ensure
			row_maxima_not_void: Result /= Void
			correct_size: Result.count = height
		end

	row_averages : VECTOR is
			-- A Vector containing the averages of rows
		local
			l_sum : DOUBLE
			i,j : INTEGER
		do
			create Result.make(height)
			from i := 1 until i > height loop
				l_sum := 0
				from j := 1 until j > width loop
					l_sum := l_sum + Current[i,j]
					j := j + 1
				end
				Result.put(l_sum / width,i)
				i := i + 1
			end
		end

	row_sums : VECTOR is
			-- vector of the sums of the rows
		local
			i : INTEGER
		do
			create Result.make(height)
			from i := 1 until i > height loop
				Result.put(row_sum(i), i)
				i := i + 1
			end
		ensure
			not_void: Result /= Void
			correct_size: Result.count = height
		end

	sub_row_mean(a_subset : ARRAY[INTEGER]; a_weights : ABSTRACT_READABLE_VECTOR) : VECTOR is
		require
			same_dim : a_subset.count = a_weights.count
		local
			j : INTEGER
			l_sub_vector : VECTOR
		do
			create Result.make (height)
			from j := 1 until j > height loop
				l_sub_vector := sub_row(j, a_subset)
				Result.put(l_sub_vector.weighted_mean(a_weights),j)
				j := j + 1
			end
		end

	row_mean (a_weights: ABSTRACT_READABLE_VECTOR): VECTOR is
		local
			j : INTEGER
		do
			create Result.make (height)
			from j := 1 until j > height loop
				Result.put(row(j).weighted_mean(a_weights),j)
				j := j + 1
			end
		end

	row_mean_with_matrix_weights (a_weight_matrix: ABSTRACT_READABLE_MATRIX): VECTOR is
		local
			j : INTEGER
		do
			create Result.make (height)
			from j := 1 until j > height loop
				Result.put(row(j).weighted_mean(a_weight_matrix.row(j)),j)
				j := j + 1
			end
		end

	column_mean (a_weights: ABSTRACT_READABLE_VECTOR): VECTOR is
		local
			j : INTEGER
		do
			create Result.make (width)
			from j := 1 until j > width loop
				Result.put(column(j).weighted_mean(a_weights),j)
				j := j + 1
			end
		end

	column_mean_with_unavailables (a_weights: ABSTRACT_READABLE_VECTOR; a_is_defined: FUNCTION [ANY, TUPLE [ABSTRACT_READABLE_MATRIX, INTEGER, INTEGER], BOOLEAN]): VECTOR is
			-- Column mean ignoring unavailable values.
		local
			j : INTEGER
		do
			create Result.make(width)
			from j := 1 until j > width loop
				Result.put(column(j).weighted_mean_with_unavailables(a_weights,agent is_defined(a_is_defined,?,j)),j)
				j := j + 1
			end
		end

	column_valid_variable_flags (a_weights: ABSTRACT_READABLE_VECTOR; a_is_defined: FUNCTION [ANY, TUPLE [ABSTRACT_READABLE_MATRIX, INTEGER, INTEGER], BOOLEAN]; a_min_wt_with_valid_values: DOUBLE): VECTOR is
			-- Vector of 1/0 flags for each column. 1 if the sum of weights with available values is greater than a_min_wt_with_valid_values, else 0.
		require
			args_not_void: a_is_defined /= Void
		local
			j : INTEGER
		do
			create Result.make(width)
			from j := 1 until j > width loop
				Result.put(column(j).has_sufficient_weight_with_valid_values(a_weights,agent is_defined(a_is_defined,?,j), a_min_wt_with_valid_values),j)
				j := j + 1
			end
		ensure
			result_not_void: Result /= Void
		end

	column_weighted_average (a_data,a_weights: ABSTRACT_READABLE_VECTOR): VECTOR is
	--the columns are used as weights
		local
			j : INTEGER
		do
			create Result.make (width)
			from j := 1 until j > width loop
				if  a_weights=void then
					Result.put(a_data.weighted_mean(column(j)),j)
				else
					Result.put(a_data.weighted_mean(column(j).cross(a_weights)),j)
				end
				j := j + 1
			end
		end
	column_standard_deviation (a_weights: ABSTRACT_READABLE_VECTOR): VECTOR is
			--Vector of column weighted standard deviations
		local
			j: INTEGER
			l_weights : ABSTRACT_READABLE_VECTOR
		do
			l_weights := safe_row_weights(a_weights)
			create Result.make (width)
			from j := 1 until j > width loop
				Result.put (column (j).weighted_standard_deviation(l_weights), j)
				j := j + 1
			end
		end

	column_standard_deviation_with_unavailables (a_weights: VECTOR; a_is_defined:FUNCTION [ANY, TUPLE [ABSTRACT_READABLE_MATRIX, INTEGER, INTEGER], BOOLEAN]): VECTOR is
		local
			l_weights : ABSTRACT_READABLE_VECTOR
			j : INTEGER
		do
			if (a_is_defined = Void) then
				Result := column_standard_deviation (a_weights)
			else
				l_weights := safe_row_weights(a_weights)
				create Result.make (width)
				from j := 1 until j > width loop
					Result.put (column(j).weighted_standard_deviation_with_unavailables(a_weights,agent is_defined(a_is_defined,?,j)), j)
					j := j + 1
				end
			end
		end

	safe_row_weights (a_weights : ABSTRACT_READABLE_VECTOR) : ABSTRACT_READABLE_VECTOR is
		do
			if a_weights /= void then
				Result := a_weights
			else
				create {UNIT_VECTOR} Result.make(height)
			end
		end

	diagonal : VECTOR is
			-- diagonal of the matrix
		require
			square : width = height
		local
			i : INTEGER
		do
			create Result.make(width)
			from i := 1 until i > width loop
				Result.put(Current[i,i],i)
				i := i + 1
			end
		end
feature -- Derived arrays

	to_columnwise_array : ROSE_LINEAR_ARRAY[ROSE_LINEAR_ARRAY[DOUBLE]] is
			-- Array of column-arrays
		local
			i : INTEGER
		do
			create Result.make(1,width)
			from i := 1 until i > width loop
				Result.put(column(i).to_array,i)
				i := i + 1
			end
		end

	to_rowwise_array : ROSE_LINEAR_ARRAY[ROSE_LINEAR_ARRAY[DOUBLE]] is
			-- Array of row-arrays
		local
			i : INTEGER
		do
			create Result.make(1,height)
			from i := 1 until i > height loop
				Result.put(row(i).to_array,i)
				i := i + 1
			end
		end

	to_row_vectors: ROSE_LINEAR_ARRAY [VECTOR] is
			-- `Current' as an array of row vectors
		local
			i: INTEGER
		do
			create Result.make (1, height)
			from
				i := 1
			until
				i > height
			loop
				Result.put (row (i), i)
				i := i + 1
			end
		ensure
			rows_not_void: Result /= Void
			correct_count: Result.count = height
		end

	to_column_vectors: ROSE_LINEAR_ARRAY [VECTOR] is
			-- `Current' as an array of column vectors
		local
			i: INTEGER
		do
			create Result.make (1, width)
			from
				i := 1
			until
				i > width
			loop
				Result.put (column (i), i)
				i := i + 1
			end
		ensure
			rows_not_void: Result /= Void
			correct_count: Result.count = width
		end

	row_array(i: INTEGER) : ROSE_LINEAR_ARRAY[DOUBLE] is
		local
			l_ix : INTEGER
			l_double : DOUBLE
		do
			if valid_row_index(i) then
				create Result.make(1, width)
				from l_ix := 1 until l_ix > width loop
					l_double := Current [i, l_ix]
					Result.put(l_double, l_ix)
					l_ix := l_ix + 1
				end
			else
				Result := Void
			end
		end

	column_array(i: INTEGER) : ROSE_LINEAR_ARRAY[DOUBLE] is
		local
			l_ix : INTEGER
			l_double : DOUBLE
		do
			if valid_column_index(i) then
				create Result.make(1, height)
				from l_ix := 1 until l_ix > height loop
					l_double := Current [l_ix, i]
					Result.put(l_double, l_ix)
					l_ix := l_ix + 1
				end
			else
				Result := Void
			end
		end
	indices_of_nonzero_column : ROSE_LINEAR_ARRAY[INTEGER] is
		local
			i, j : INTEGER
		do
			create Result.make_empty
			from i := 1 until i > width loop
				from j := 1 until j > height or else Current[j,i] /= 0 loop
					j := j + 1
				end
				if j <= height then
					Result.extend(i)
				end
				i := i + 1
			end
		end

	transformed_matrix (a_function: FUNCTION [ANY, TUPLE [DOUBLE], DOUBLE]): MATRIX is
			-- Result of applying `a_function' to every item of a matrix
		require
			a_function_not_void: a_function /= Void
		local
			l_finished, l_finished_inner: BOOLEAN
			i, j: INTEGER
		do
			create Result.make (height, width)
			from
				i := 1
			invariant
				row_indexable: l_finished or valid_row_index (i)
				target_row_indexable: l_finished or Result.valid_row_index (i)
			variant
				increasing_row_index: height + 1 - i
			until l_finished loop
				from
					j := 1
					l_finished_inner := False
				invariant
					column_indexable: l_finished_inner or valid_column_index (j)
					target_column_indexable: l_finished_inner or Result.valid_column_index (j)
				variant
					increasing_column_index: width + 1 - j
				until l_finished_inner loop
					Result [i, j] := a_function.item ([Current [i, j]])
					j := j + 1
					if j > width then
						l_finished_inner := True
					end
				end
				i := i + 1
				if i > height then
					l_finished := True
				end
			end
		ensure
			transformed_matrix_not_void: Result /= Void
			correct_width: Result.width = width
			correct_height: Result.height = height
			--transformed: Result.for_all_with_index (agent is_transformed_item (a_function, ?, ?, ?))
		end

feature -- Iteration

	columns_transformed_to_vector (a_agent: FUNCTION [ANY, TUPLE [VECTOR], DOUBLE]): VECTOR is
			-- Results of flattening each column by `a_agent'
		require
			a_agent_not_void: a_agent /= Void
		local
			i: INTEGER
		do
			create Result.make (width)
			from
				i := 1
			until
				i > width
			loop
				Result [i] := a_agent.item ([column (i)])
				i := i + 1
			end
		ensure
			result_not_void: Result /= Void
			correct_count: Result.count = width
		end

	do_all_with_indices (a_action: PROCEDURE [ANY, TUPLE [DOUBLE, INTEGER, INTEGER]]) is
			-- Apply `a_action' to all item values, passing row and column indices.
			-- Iteration order is all cells in first row in increasing column index, followed by second row etc.
		require
			a_action_not_void: a_action /= Void
		local
			l_args: TUPLE [DOUBLE, INTEGER, INTEGER]
			l_row, l_column: INTEGER
		do
			from
				create l_args
				l_row := 1
			until
				l_row > height
			loop
				from
					l_column := 1
				until
					l_column > width
				loop
					l_args.put_double (Current [l_row, l_column], 1)
					l_args.put_integer (l_row, 2)
					l_args.put_integer (l_column, 3)
					a_action.call (l_args)
					l_column := l_column + 1
				end
				l_row := l_row + 1
			end
		end

feature -- Comparison

	max_diff (a_other: like Current): DOUBLE is
		require
			a_other_not_void: a_other /= void
		local
			l_diff : MATRIX
			i : INTEGER
		do
			l_diff := Current - a_other
			from i := 1 until i > l_diff.width loop
				Result := Result.max(l_diff.column(i).abs.max)
				i := i + 1
			end
		end

	is_equivalent (a_other: ABSTRACT_READABLE_MATRIX): BOOLEAN is
		require
			a_other_not_void: a_other /= void
		local
			i,j : INTEGER
			l_difference : DOUBLE
		do
			Result := a_other = Current
			if not Result then
				Result := height = a_other.height and width = a_other.width
				from i := 1 until i > height or not Result loop
					from j := 1 until j > width loop
						l_difference := Current [i,j] - a_other[i,j]
						if is_defined_double(l_difference) then
							Result := l_difference.abs < epsilon
						else
							Result := is_defined_double(Current[i,j]) = is_defined_double(a_other[i,j])
						end
						j := j + 1
					end
					i := i + 1
				end
			end
		end

	contrast_matrix(a_other : ABSTRACT_READABLE_MATRIX) : MATRIX is
		require
			same_size : a_other.height = height and a_other.width = a_other.width
		local
			i : INTEGER
			l_vec1, l_vec2 : VECTOR
			l_column_names : ROSE_LINEAR_ARRAY[STRING]
		do
			create Result.make(height, width * 2)
			create l_column_names.make(1, width * 2)
			from i := 1 until i > width loop
				l_vec1 := column(i)
				l_vec2 := a_other.column(i)
				Result.put_column_vector(l_vec1, 2 * i - 1)
				Result.put_column_vector(l_vec2, 2 * i)

				if column_names /= Void and then column_names.valid_index(i) then
					l_column_names.put("1-" + column_names.item(i), 2 * i -1)
				else
					l_column_names.put("1-" + "Var_" + i.out, 2 * i -1)
				end
				if a_other.column_names /= Void and then a_other.column_names.valid_index(i) then
					l_column_names.put("2-" + a_other.column_names.item(i), 2 * i)
				else
					l_column_names.put("2-" + "Var_" + i.out, 2 * i)
				end
				i := i + 1
			end
			Result.set_row_names(row_names)
			Result.set_column_names(l_column_names)
		end

	is_pseudo_inverse (a_other : MATRIX) : BOOLEAN is
		do
			Result := is_equivalent(as_matrix.product (a_other).product (Current)) and
				a_other.is_equivalent(a_other * product(a_other))
		end


feature --operators

	product,infix "*" (a_other : ABSTRACT_READABLE_MATRIX) : MATRIX is
		require
			dimensions_match : width = a_other.height
		local
			i,j,k : INTEGER
			l_dot_product : DOUBLE
		do
			create Result.make(height,a_other.width)
			from i := 1 until i > height loop
				from j := 1 until j > a_other.width loop
					l_dot_product := 0
					from k := 1 until k > a_other.height loop
						l_dot_product := l_dot_product + Current[i,k] * a_other [k,j]
						k := k + 1
					end
				--	Result.put(l_dot_product,i,j)
					Result [i, j] := l_dot_product
					j := j + 1
				end
				i := i + 1
			end
		end

	infix "+" (a_other : ABSTRACT_READABLE_MATRIX) : MATRIX is
		require
			dimensions_match : width = a_other.width and height = a_other.height
		local
			i,j : INTEGER
		do
			create Result.make(height,a_other.width)
			from i := 1 until i > height loop
				from j := 1 until j > a_other.width loop
				--	Result.put(item(i,j) + a_other.item(i,j),i,j)
					Result [i, j] := Current[i, j] + a_other[i, j]
					j := j + 1
				end
				i := i + 1
			end
		end

	infix "-", difference (a_other : ABSTRACT_READABLE_MATRIX) : MATRIX is
		require
			dimensions_match : width = a_other.width and height = a_other.height
		local
			i,j : INTEGER
		do
			create Result.make(height,a_other.width)
			from i := 1 until i > height loop
				from j := 1 until j > a_other.width loop
				--	Result.put(item(i,j) - a_other.item(i,j),i,j)
					Result [i, j] := Current [i, j] - a_other[i, j]
					j := j + 1
				end
				i := i + 1
			end
		end

	vector_product,infix "@" (a_vector : ABSTRACT_READABLE_VECTOR) : VECTOR is
			-- `current' times `a_vector' where undefineds are treated as 0s
		require
			dimensions_match : a_vector.count = width
		local
			l_data : SPECIAL[DOUBLE]
			i,j : INTEGER
			l_sum : DOUBLE
		do
			create Result.make(height)
			l_data := Result.data
			from i := 1 until i > height loop
				l_sum := 0
				from j := 1 until j > width loop
					if is_defined_double(Current[i, j]) and is_defined_double(a_vector [j]) then
						l_sum := l_sum + Current [i, j] * a_vector [j]
					end
					j := j + 1
				end
				l_data.put(l_sum,i)
				i := i + 1
			end
		end

	fast_vector_product (a_vector : VECTOR) : VECTOR is
			-- faster implementation of `current' times `a_vector' where `current' and `a_vector' are assumed to be all defined
		require
			dimensions_match : a_vector.count = width
			a_vector_all_defined : a_vector.all_defined
			is_not_null : not is_null
		local
			l_data : SPECIAL[DOUBLE]
			i,j : INTEGER
			l_sum : DOUBLE
		do
			create Result.make(height)
			l_data := Result.data
			from i := 1 until i > height loop
				l_sum := 0
				from j := 1 until j > width loop
					l_sum := l_sum + Current [i, j] * a_vector [j]
					j := j + 1
				end
				l_data.put(l_sum,i)
				i := i + 1
			end
			-- this can slow things down, but is safer
			if not Result.all_defined then
				Result := vector_product(a_vector)
			end
		end

	infix "|" (a_other : ABSTRACT_READABLE_MATRIX) : HORIZONTALLY_COMBINED_MATRIX is
	 	do
	 		create Result.make(Current,a_other)
	 	end

	itemwise_product(a_other: ABSTRACT_READABLE_MATRIX):MATRIX is
			require
			dimensions_match : width = a_other.width and height = a_other.height
		local
			i,j : INTEGER
		do
			create Result.make(height,a_other.width)
			from i := 1 until i > height loop
				from j := 1 until j > a_other.width loop
				--	Result.put(item(i,j)*a_other.item(i,j),i,j)
					Result [i, j] := Current [i, j] * a_other [i, j]
					j := j + 1
				end
				i := i + 1
			end
		end

	partial_product (a_vector : VECTOR) : VECTOR is
		--compute image of initial segment of a_vector
		require
			dimensions_match : a_vector.count >= width
		local
			l_data : SPECIAL[DOUBLE]
			i,j : INTEGER
			l_sum : DOUBLE
		do
			create Result.make(height)
			l_data := Result.data
			from i := 1 until i > height loop
				l_sum := 0
				from j := 1 until j > width loop
					if is_defined_double(Current[i, j]) and is_defined_double(a_vector [j]) then
						l_sum := l_sum + Current [i, j] * a_vector [j]
					end
					j := j + 1
				end
				l_data[i] := l_sum
				i := i + 1
			end
		end


feature --transpose
	transpose, transposed : ABSTRACT_MATRIX is
		local
			i,j : INTEGER
		do
			create {MATRIX}Result.make(width, height)
			from i := 1 until i > height loop
				from j := 1 until j > width loop
				--	Result.put(item(i,j),j,i)
					Result [j, i] := Current [i, j]
					j := j + 1
				end
				i := i + 1
			end
			Result.set_name(name)
			Result.set_column_names(row_names)
			Result.set_row_names(column_names)
		ensure
			not_void: Result /= Void
		end

	transpose_at (a_vector : ABSTRACT_READABLE_VECTOR) : VECTOR is
		require
			dimensions_match : a_vector.count = height
		local
			l_data : SPECIAL[DOUBLE]
			i,j : INTEGER
			l_sum : DOUBLE
		do
			create Result.make(width)
			l_data := Result.data
			from i := 1 until i > width loop
				l_sum := 0
				from j := 1 until j > height loop
					l_sum := l_sum + Current[j, i] * a_vector.item(j)
					j := j + 1
				end
				l_data.put(l_sum,i)
				i := i + 1
			end
		end

	transpose_product (a_matrix: ABSTRACT_READABLE_MATRIX): MATRIX is
			-- returns Current (transposed) * a_matrix
		require
			same_height: height = a_matrix.height
		local
			i, j, k: INTEGER
			d: DOUBLE
		do
			create Result.make (width, a_matrix.width)
			from i := 1 until i > Result.height loop
				from j := 1 until j > Result.width loop
					d := 0
					from k := 1 until k > height loop
						d := d + Current [k, i] * a_matrix[k, j]
					k := k + 1
					end
				--	Result.put (d, i, j)
					Result [i, j] := d
				j := j + 1
				end
			i := i + 1
			end
		end

	product_transpose (a_matrix: ABSTRACT_READABLE_MATRIX): MATRIX is
			-- returns Current * a_matrix (transposed)
		require
			width = a_matrix.width
		local
			i, j, k: INTEGER
			d: DOUBLE
		do
			create Result.make (height, a_matrix.height)
			from i := 1 until i > Result.height loop
				from j := 1 until j > Result.width loop
					d := 0
					from k := 1 until k > width loop
						d := d + Current [i, k] * a_matrix[j, k]
					k := k + 1
					end
				--	Result.put (d, i, j)
					Result [i, j] := d
				j := j + 1
				end
			i := i + 1
			end
		end


	xtx : SYMMETRIC_MATRIX is
			-- X.transpose times X
		do
			create Result.make_xtx(Current)
			Result.set_column_names(column_names)
			Result.set_row_names(column_names)
		ensure
			valid_result: Result /= Void
		end

	xWxt(a_w : ABSTRACT_MATRIX) : SYMMETRIC_MATRIX is
		require
			size : a_w.height = width
		local
			l_matrix : MATRIX
		do
			l_matrix := as_matrix.product (a_w)
			l_matrix := l_matrix.product_transpose(Current)
			create Result.make_from_triangle(l_matrix, True)
		end

	xtx_with_contributors: XTX_WITH_CONTRIBUTORS is
			-- keeps track of largest contributors to diagonal
			-- pairs of row, incremental x'x value store for the top 3
			do
				Create Result.make_xtx(Current)
			end

feature --inverse
	inverse : ABSTRACT_MATRIX is
		--return the inverse of Current
		require
			square : width = height
		do
			create {MATRIX}Result.make_as_inverse(Current)
		ensure
			is_inverse : (Result * Current).is_identity
		end

	is_invertible: BOOLEAN
			-- is the matrix invertible
		local
			m: MATRIX
			lu: LU_DECOMPOSITION
		do
			create m.make_identity (width)
			create lu.make (Current)
			m := lu.solve (m)
			Result := lu.is_valid
		end


	pseudo_inverse : ABSTRACT_MATRIX is
		--return a psuedo inverse
		require
			width_not_greater_than_height : width <= height
		do
			Result := svd.pseudo_inverse
		ensure
			is_psuedo_inverse : Result.is_pseudo_inverse(as_matrix)
		end

feature --decompositions
	svd : SINGULAR_VALUE_DECOMPOSITION is
		do
			create Result.make(Current)
		end

	lud : LU_DECOMPOSITION is
		do
			create Result.make(Current)
		end


feature--names
feature -- ARRAY2 representation

	excel_output (a_row_label, a_column_label: ARRAY[ANY]; a_first: ANY): ARRAY2 [ANY] is
			-- Object formatted for output to excel;
			-- Adds 10 rows and 10 columns to matrix dimensions.
			require
			 a_row_label_valid: a_row_label = Void or a_row_label.count <= height
			 a_column_label_valid: a_column_label = Void or a_column_label.count <=width
			local
				i,j:INTEGER
			do
			create result.make(height+10,width+10)
			if a_first/=void then
				result.put(a_first,1,1)
			end
			if a_row_label/=void  then
				from i:=1 until i>a_row_label.count loop
				result.put(a_row_label.item(i).out,i+1,1)
				i:=i+1
				end
			end
			if a_column_label/=void  then
				from i:=1 until i>a_column_label.count loop
				result.put(a_column_label.item(i).out,1,i+1)
				i:=i+1
				end
			end
			from i:=1 until i>width	loop
				from j:=1 until j>height	loop
					result.put(Current[j,i],j+1,i+1)
				j:=j+1
				end
				i:=i+1
			end
		end

	excel_output_extend_one(a_row_label,a_column_label:ARRAY[ANY];a_first:ANY):ARRAY2[ANY] is
		--Creates object formatted to output to excel.  Adds only one row and column to matrix dimensions.
			require
			 a_row_label_valid: a_row_label = Void or  a_row_label.count <= height
			 a_column_label_valid: a_column_label = Void or a_column_label.count <=width
			local
				i,j:INTEGER
			do
			create result.make(height+1,width+1)
			if a_first/=void then
				result.put(a_first,1,1)
			end
			if a_row_label/=void  then
				from i:=1 until i>a_row_label.count loop
				result.put(a_row_label.item(i).out,i+1,1)
				i:=i+1
				end
			end
			if a_column_label/=void  then
				from i:=1 until i>a_column_label.count loop
				result.put(a_column_label.item(i).out,1,i+1)
				i:=i+1
				end
			end
			from i:=1 until i>width	loop
				from j:=1 until j>height	loop
					result.put(Current[j,i],j+1,i+1)
				j:=j+1
				end
				i:=i+1
			end
		end

	combined_output(a_row_label:ARRAY[ANY];a_column_label:ARRAY[ANY];a_other:ARRAY2[ANY]):ARRAY2[ANY]	is
		--formated for excel spreadsheet
		require
			 a_row_label =void or  a_row_label.count <= height
			 a_column_label=void or a_column_label.count <=width
			 a_other=void or a_other.height<= height+1
		local
			i,j:INTEGER
			l_item:ANY
			l_combine:BOOLEAN
		do
			if  a_other=void then
				create result.make(height+1,width+1)
			else
				l_combine:=true
				create result.make(height+1,width+1+a_other.width)
			end
			if a_row_label/=void  then
				from i:=1 until i>a_row_label.count loop
				l_item:=a_row_label.item(i)
				if l_item/=void then
					result.put(a_row_label.item(i).out,i+1,1)
				end
				i:=i+1
				end
			end
			if a_column_label/=void  then
				from i:=1 until i>a_column_label.count loop
				l_item:=a_column_label.item(i)
				if l_item/=void then
					result.put(a_column_label.item(i).out,1,i+1)
				end
				i:=i+1
				end
			end
			from i:=1 until i>width	loop
				from j:=1 until j>height	loop
					Result.put(Current[j,i],j+1,i+1)
				j:=j+1
				end
				i:=i+1
			end
			if l_combine then
				from i:=1 until i>a_other.width loop
					from j:=1 until j>a_other.height loop
							result.put(a_other.item(j,i),j,i+width+1)
						j:=j+1
					end
					i:=i+1
				end
			end
		end
	output:ARRAY2[ANY] is
			-- array2 with labels
		do
			result:=combined_output (row_names,column_names,void)
		end

	raw_output : ROSE_ARRAY2[ANY] is
			-- array2 without label
		local
			i, j : INTEGER
		do
			create Result.make(height, width)
			from i := 1 until i > height loop
				from j := 1 until j > width loop
					Result.put(Current[i, j], i, j)
					j := j + 1
				end
				i := i + 1
			end
		end

	dump_data (a_path: STRING): BOOLEAN
			-- Dump relevant input data to `a_path' for analysis
			-- Can be called in the debugger watch window
		require
			valid_path: a_path /= Void and then not a_path.is_empty
		local
			l_file: ROSE_TEXT_FILE
			i, j: INTEGER
		do
			create l_file.make_open_write (a_path)
			from
				i := 1
			until
				i > height
			loop
				from
					j := 1
				until
					j > width
				loop
					l_file.put_string (item (i, j).out)
					if j < width then
						l_file.put_string (",")
					end
					j := j + 1
				end
				l_file.put_string ("%N")
				i := i + 1
			end
			l_file.flush
			l_file.close
			Result := True
		end

feature -- String representation

	string : STRING is
			-- string
		local
			i,j : INTEGER
		do
			Result := ""
			from i := 1 until i > height loop
				from j := 1 until j > width loop
					Result := Result + format.formatted(Current [i,j]+ format_precision) + " "
					j := j + 1
				end
				Result := Result + "%N"
				i := i + 1
			end
		end
	string_representation : STRING is
			-- string representation
		local
			i, j: INTEGER
		do
			create Result.make (123)
			from i := 1 until i > height loop
				from j := 1 until j > width loop
					Result.append (item(i, j).out + " ")
					j := j + 1
				end
				Result.append ("%N")
				i := i + 1
			end
		end

	display is
			-- display
		do
			print(string + "%N")
		end

	verbose_display is
			-- display
		do
			print(string_representation + "%N")
		end

feature -- Comparison


	is_same (a_other: ABSTRACT_READABLE_MATRIX): BOOLEAN is
			-- Is `a_other' numerically equal to `Current'?
		require
			a_other_not_void: a_other /= Void
		local
			i, j: INTEGER
		do
			Result := height = a_other.height and width = a_other.width
			from i := 1 until i > height or not Result loop
				from j := 1 until j > width or not Result loop
					Result := Current [i, j] = a_other [i, j]
					j := j + 1
				end
				i := i + 1
			end
		ensure
			same_dimensions: Result implies a_other.width = width	and a_other.height = height
		end

feature {UPDATEABLE_DECOMPOSITION} --internal

	format : FORMAT_DOUBLE is
			-- Item formatter
		once
			create Result.make(9,4)
		end

	operations : MATRIX_OPERATIONS[MATRIX] is
			-- Matrix operations
		once
			create Result
		end
	anchor : ABSTRACT_MATRIX is --anchor type
	do end
	epsilon : DOUBLE is 0.0000000000001 --tolerance
	format_precision : DOUBLE is .00000001
feature --inappropriate

	column_as_industry_weighted_medians(a_data:VECTOR):VECTOR is
		--this was a way to calculate industry weights
		local
			i:INTEGER
		do
			create result.make (width)
			from i:=1 until i>width loop
				result.put (a_data.industry_median (column(i)), i)
				i:=i+1
			end
		end
	normalized_as_descriptor(a_weights : ABSTRACT_READABLE_VECTOR) : MATRIX is
		-- Using cap weighted mean and equal weighted Stddev
		local
			m: VECTOR
			l_std: VECTOR
			i, j: INTEGER
		do
			m := column_mean (a_weights)
			l_std := column_standard_deviation (Void)
			create Result.make (height, width)
			from i := 1 until i > height loop
				from j := 1 until j > width loop
					if is_defined_double (Current [i,j]) then
						Result [i, j] := (Current[i, j] - m[j]) / l_std[j]
					else
						Result [i, j] := Current [i, j]
					end
					j := j + 1
				end
				i := i + 1
			end
			copy_names(Result)
		end

invariant
	sensible_size : height >= 0 and width >= 0
end
