note

	description:

		"Readable vectors with no commitment to a specific implementation"

	keywords: "vector, array, matrix"
	author: "Mark Howard"
	last_editor: "$Author: rj $"
	date: "$Date: 2014/12/12 $"
	revision: "$Revision: #108 $"
	path: "$File: //rose/source/kernel/infrastructure/math/abstract_readable_vector.e $"

deferred class	ABSTRACT_READABLE_VECTOR

inherit

	DOUBLE_MATH
		export
			{NONE} all
		undefine
			is_equal
		end

	IEEE_MATH
		export
			{NONE} all
		undefine
			is_equal
		end

	ROSE_LEGACY_UNDEFINED
		export
			{NONE} all
			{ANY} is_defined_double
		undefine
			is_equal
		end

	MISMATCH_CORRECTOR
			-- Remove inheritance to this once all storables have been regenerated
			-- (ROSE_STORABLE removed from ancestors on 04/30/2004)
		redefine
			correct_mismatch, is_equal
		end

	APPROXIMATE_EQUALITY
		export
			{NONE} all
			{ANY} is_missing
		undefine
			is_equal
		end

feature -- Access

	item alias "[]" (i: INTEGER): DOUBLE is
			-- Element at position `i'
		require
			index_in_range: valid_index (i)
		deferred
		end

	lower: INTEGER is 1
			-- Index of first element in `Current'

	upper: INTEGER is
			-- Index of last element in `Current'
		do
			Result := count
		ensure
			definition: Result = count
		end

	last: DOUBLE is
			-- Last value
		require
			not_empty: count > 0
		do
			Result := item (count)
		end

	first: DOUBLE is
			-- First value
		require
			not_empty: count > 0
		do
			Result := item (1)
		end

	name: STRING
			-- Name of `Current'

	names: ROSE_LINEAR_ARRAY [STRING]
			-- Name of each element in `Current'

	has_names: BOOLEAN is
			-- Is `names' present and populated?
		do
			Result := names /= Void and then not names.has (Void)
		ensure
			definition: Result = (names /= Void and then not names.has (Void))
		end

	name_for_item (i: INTEGER): STRING is
			-- Name of element `i' in `Current'
		require
			index_in_range: valid_index (i)
		do
			if names /= Void then
				Result := names [i]
			end
		ensure
			result_not_void: names /= Void implies Result /= Void
		end

	item_by_name (a_name:STRING):DOUBLE is
			-- Synonym of `value_at'
		obsolete
			"Use value_at"
		require
			has_names: has_names
			a_name_not_void: a_name /= Void
		do
			Result := value_at (a_name)
		end

	index_by_name (a_name:STRING): INTEGER is
			-- Synonym of `index_of'
		obsolete
			"Use index_of"
		require
			has_names: has_names
			a_name_not_void: a_name /= Void
		do
			Result := name_table.item (a_name)
		end

	index_of (a_name: STRING): INTEGER is
			-- Index of item by name
		require
			has_names: has_names
			a_name_not_void: a_name /= Void
		do
			Result := name_table.item (a_name)
		end

	value_at (a_name: STRING): DOUBLE is
			-- Item by name
		require
			has_names: has_names
			a_name_not_void: a_name /= Void
		local
			k:INTEGER
		do
			k := index_of (a_name)
			if k > 0 then
				Result := item (k)
			else
				Result := undefined_double
			end
		end

	divided_by (a_vector: ABSTRACT_READABLE_VECTOR): VECTOR is
			-- Element level division
			-- (divide by zero produces 0 instead of NaN)
		require
			same_size: a_vector /= Void and then a_vector.count = count
		local
			i: INTEGER
		do
			create Result.make (count)
			from i := 1 until i > count loop
				if a_vector.item (i) /= 0.0 then
					Result.put (item (i) / a_vector.item (i), i)
				end
				i := i + 1
			end
		end

	element_by_element_multiply (a_vector: ABSTRACT_READABLE_VECTOR): VECTOR is
			-- Multiply elements
			-- (as opposed to vector multiplication)
		require
			same_size: a_vector /= Void and then a_vector.count = count
		local
			i: INTEGER
		do
			create Result.make (count)
			from i := 1 until i > count loop
				if a_vector.item (i) /= 0.0 then
					Result.put (item (i) * a_vector.item (i), i)
				end
				i := i + 1
			end
		ensure
			element_by_element_multiply_not_void: Result /= Void
			same_count: Result.count = count
		end

	ln: VECTOR is
			-- Natural logarithms of elements;
		local
			i: INTEGER
			l_neg: BOOLEAN
		do
			create Result.make (count)
			from
				i:= 1
			until
				i > count or l_neg
			loop
				if item (i) <= 0.0 then
					Result[i] := constants.missing_double
				else
					Result[i] := log (item (i))
				end
				i := i + 1
			end
		ensure
			same_count: Result /= Void and then Result.count = count
		end

	inverse: VECTOR is
			-- Element by element inverse (1/x);
			-- 0.0 maps to 0.0.
		local
			i: INTEGER
		do
			create Result.make (count)
			from
				i := 1
			until
				i > count
			loop
				if item (i) /= 0.0 then
					Result.put (1.0 / item (i), i)
				end
				i := i + 1
			end
		ensure
			inverse_not_void: Result /= Void
			same_count: Result.count = count
		end

	subset_by_names(a_names: ARRAY [STRING]): VECTOR is
			-- Subvector by named index
		require
			has_names: has_names
			a_names_not_void: a_names /= Void
			a_names_not_has_void: not a_names.has (Void)
		local
			i: INTEGER
			l_names: like names
		do
			create l_names.make_from_array (a_names)
			create Result.make_with_names (l_names)
			from
				i := 1
			until
				i > a_names.count
			loop
				Result.put (value_at (l_names.item (i)), i)
				i := i + 1
			end
		ensure
			subset_by_names_not_void: Result /= Void
			correct_count: Result.count = a_names.count
			subset_by_names_has_names: Result.has_names
			correct_values: True -- Result.for_all_with_index (...)
		end

	name_table: ROSE_HASH_TABLE [INTEGER, STRING] is
			-- `names' as a hash table
			-- Map of names to indices
		require
			has_names: has_names
		local
			i: INTEGER
		do
			Result := name_table_saved
			if Result = Void then
				create Result.make (names.count)
				from
					i := 1
				until
					i > names.count
				loop
					Result.put (i, names.item (i))
					i := i + 1
				end
				name_table_saved := Result
			end
		ensure
			result_not_void: Result /= Void
			same_count: Result.count = names.count
			table_cached: Result = name_table_saved
		end

	max: DOUBLE is
			-- Maximum defined value in `Current'
		local
			i: INTEGER
		do
			Result := negative_infinity
			from i := 1 until i > count loop
				if not is_missing (Current [i]) and then Current [i] > Result then
					Result := Current [i]
				end
				i := i + 1
			end
		ensure
			defined_result: not is_missing (Result)
			maximum_defined_double: subset (defined_item_indices).for_all (agent not_less_than (Result, ?))
		end

	min: DOUBLE is
			-- Minimum defined value in `Current'
		local
			i: INTEGER
		do
			Result := positive_infinity
			from i := 1 until i > count loop
				if not is_missing (Current [i]) and then Current [i] < Result then
					Result := Current [i]
				end
				i := i + 1
			end
		ensure
			defined_result: not is_missing (Result)
			minimum_defined_double: subset (defined_item_indices).for_all (agent not_greater_than (Result, ?))
		end

	min_max: PAIR [DOUBLE, DOUBLE] is
			-- Minimum and maximum defined values in `Current'
		do
			create Result.make (min, max)
		ensure
			result_not_void: Result /= Void
			defined_minimum: not is_missing (Result.first)
			defined_maximum: not is_missing (Result.second)
			minimum_not_greater_than_maximum: count > 0 implies Result.first <= Result.second
		end

	index_of_max: INTEGER is
			-- Position of largest non missing value in `Current'
		local
			i: INTEGER
		do
			if count > 0 then
				Result := 1
			end
			from i := 1 until i > count loop
				if less_than (Current [Result], Current [i]) then
					Result := i
				end
				i := i + 1
			end
		ensure
			result_for_empty_vector: count = 0 implies Result = 0
			strictly_positive_result_for_non_empty_vector: count > 0 implies Result > 0
			result_in_range: Result <= upper
			defined_at_resulting_index: Result > 1 implies not is_missing (Current [Result])
			maximum_value_at_resulting_index: Result > 1 and then not is_missing (Current [Result]) implies max = (Current [Result])
		end

	index_of_min: INTEGER is
			-- Position of smallest value in `Current'
		local
			i: INTEGER
		do
			if count > 0 then
				Result := 1
			end
			from i := 1 until i > count loop
				if Current [i] < Current [Result] then
					Result := i
				end
				i := i + 1
			end
		ensure
			result_for_empty_vector: count = 0 implies Result = 0
			strictly_positive_result_for_non_empty_vector: count > 0 implies Result > 0
			result_in_range: Result <= upper
			maybe_defined_at_resulting_index: True
		end

	min_other (a_other: ABSTRACT_READABLE_VECTOR): VECTOR is
			-- Minima of corresponding elements;
			-- Semantics depend upon those of `{DOUBLE}.min'.
		require
			other_vector_not_void: a_other /= Void
		local
			i, k: INTEGER
		do
			k := count.min (a_other.count)
			create Result.make (k)
			from i := 1 until i > k loop
				Result [i] := (Current [i]).min (a_other [i])
				i := i + 1
			end
		ensure
			result_not_void: Result /= Void
			new_object: as_readonly_vector /= Result
			lowest_common_count: Result.count = count.min (a_other.count)
			minima: True
		end

	max_other (a_other: ABSTRACT_READABLE_VECTOR): VECTOR is
			-- Maxima of corresponding elements;
			-- Semantics depend upon those of `{DOUBLE}.max'.
		require
			other_vector_not_void: a_other /= Void
		local
			i, k: INTEGER
		do
			k := count.min (a_other.count)
			create Result.make (k)
			from i := 1 until i > k loop
				Result [i] := (Current [i]).max (a_other [i])
				i := i + 1
			end
		ensure
			result_not_void: Result /= Void
			new_object: as_readonly_vector /= Result
			lowest_common_count: Result.count = count.min (a_other.count)
			maxima: True
		end

	first_non_zero_item_index : INTEGER is
			-- index of first non zero item
		local
			i : INTEGER
		do
			from i := 1 until i > count or Result > 0 loop
				if Current[i] /= 0 then
					Result := i
				end
				i := i + 1
			end
		end

	first_defined_positive_item_index : INTEGER is
			-- index of first defined positive item
		local
			i : INTEGER
		do
			from i := 1 until i > count or Result > 0 loop
				if is_defined_double (Current[i]) and then Current[i] > 0 then
					Result := i
				end
				i := i + 1
			end
		end

	matching_item_indices (a_test: FUNCTION [ANY, TUPLE [DOUBLE], BOOLEAN]): ROSE_LINEAR_ARRAY [INTEGER] is
			-- Indices of elements that pass `a_test';
		require
			a_test_not_void: a_test /= Void
		local
			i: INTEGER
		do
			create Result.make_empty
			from i := 1 until i > count loop
				if a_test.item ([Current [i]]) then
					Result.extend (i)
				end
				i := i + 1
			end
		ensure
			result_not_void: Result /= Void
			maximum_count: Result.count <= count
			all_indices_valid: Result.for_all (agent valid_index)
		end

	industry_median (a_industry_weight: like Current): DOUBLE is
			-- Median of elements corresponding to non-zero elements of `a_industry_weight'
		require
			industry_weights_not_void: a_industry_weight /= Void
		do
			Result := subset (a_industry_weight.matching_item_indices (agent is_non_zero)).median
		end

	non_zero_subset:VECTOR is
		--vector of nonzero values
		do
			result:=subset (non_zero_indices)
		end

	non_zero_indices:ROSE_LINEAR_ARRAY[INTEGER] is
		--array of indices where the vector is nonzero
		do
			Result := matching_item_indices (agent is_non_zero)
		ensure
			nonvoid : Result /= void
		end

	defined_item_indices:ROSE_LINEAR_ARRAY[INTEGER] is
		--array of indices where the vector is nonzero
		do
			Result := matching_item_indices (agent is_defined)
		ensure
			nonvoid : Result /= void
		end

	lagged_matrix (a_lag: INTEGER): MATRIX is
			-- Lagged matrix made of the vector: each of width a_lag, starting from item [a_lag]; assuming the Current[1] is the earliest observation
		require
			a_lag_strictly_positive: a_lag > 0
			a_lag_smaller_than_count: a_lag < count
		local
			l_matrix: MATRIX
			l_width, l_height, l_idx, l_lag: INTEGER
		do
			l_width := a_lag
			l_height := count - a_lag
			create l_matrix.make (l_height, l_width)
			from l_idx := 1 until l_idx > l_height loop
				from l_lag := 1 until l_lag > a_lag loop
					l_matrix[l_idx, l_lag] := Current [l_idx  + a_lag - l_lag]
					l_lag := l_lag + 1
				end
				l_idx := l_idx + 1
			end
			Result := l_matrix
		ensure
			lagged_matrix_not_void: Result /= Void
			correct_width: Result.width = a_lag
			correct_height: Result.height = count - a_lag
		end

	lagged_matrix_with_latest_row (a_lag: INTEGER): MATRIX is
			-- Lagged matrix made of the vector: each of width a_lag, starting from item [a_lag]; assuming the Current[1] is the earliest observation
			-- including the latest row
		require
			a_lag_strictly_positive: a_lag > 0
			a_lag_smaller_than_count: a_lag < count
		local
			l_matrix: MATRIX
			l_width, l_height, l_idx, l_lag: INTEGER
		do
			l_width := a_lag
			l_height := count - a_lag + 1
			create l_matrix.make (l_height, l_width)
			from l_idx := 1 until l_idx > l_height loop
				from l_lag := 1 until l_lag > a_lag loop
					l_matrix[l_idx, l_lag] := Current [l_idx  + a_lag - l_lag]
					l_lag := l_lag + 1
				end
				l_idx := l_idx + 1
			end
			Result := l_matrix
		ensure
			lagged_matrix_with_latest_row_not_void: Result /= Void
			correct_width: Result.width = a_lag
			correct_height: Result.height = count + 1 - a_lag
		end

	magnitude : DOUBLE is
			-- Magnitude of `Current'
		local
			i : INTEGER
			l_sum : DOUBLE
		do
			from i := 1 until i > count loop
				l_sum := l_sum + (item (i) * item (i))
				i := i + 1
			end
			Result := sqrt (l_sum)
		ensure
			magnitude_non_negative: is_defined(Result) implies Result >= 0.0
		end

feature -- Access

	positive_indices: ROSE_LINEAR_ARRAY [INTEGER] is
			-- Indices of elements that are defined and greater than zero
		do
			Result := matching_item_indices (agent is_finite_and_greater_than (?, 0))
		ensure
			result_not_void: Result /= Void
			maximum_count: Result.count <= count
			all_indices_valid: Result.for_all (agent valid_index)
		end

	negative_indices: ROSE_LINEAR_ARRAY [INTEGER] is
			-- Indices of elements that are defined and less than zero
		do
			Result := matching_item_indices (agent is_finite_and_less_than (?, 0))
		ensure
			result_not_void: Result /= Void
			maximum_count: Result.count <= count
			all_indices_valid: Result.for_all (agent valid_index)
		end

	index_set: INTEGER_INTERVAL is
			-- Index interval for iteration
		do
			create Result.make (1, count)
		ensure
			result_not_void: Result /= Void
			same_bounds: (Result.lower = lower) and (Result.upper = upper)
			all_indices_valid: Result.for_all (agent valid_index)
		end

	abs_values_sum: DOUBLE is
			-- Sum of absolute values of items
		local
			i: INTEGER
		do
			from
				i := 1
			until i > count
			loop
				Result := Result + item (i).abs
				i := i + 1
			end
		ensure
			undefined_result: not all_defined implies not is_defined_double (Result)
			abs_values_sum_non_negative: all_defined implies Result >= 0.0
		end

	find_index (a_d: DOUBLE; a_start, a_end: INTEGER): INTEGER is
			-- Index of nearest value in `Current' to `a_d' between positions `a_start' and `a_end' inclusive;
			-- `Current' is assumed to be sorted in ascending order.
		require
			a_start_valid_index: valid_index (a_start)
			a_end_valid_index: valid_index (a_end)
			a_start_before_a_end: a_start < a_end
		local
			i: INTEGER
			l_stop: BOOLEAN
		do
			if is_missing(a_d) or else a_d < item (a_start) then
				Result := a_start
			elseif a_d > item (a_end) then
				Result := a_end
			else
				from
					i := a_start
				invariant
					i_plus_one_valid_index: l_stop or else valid_index (i)
				variant
					increasing_index: a_end - i + 1
				until
					i > a_end - 1 or l_stop
				loop
					if item (i) <= a_d and item (i + 1) >= a_d then
						if (item (i) - a_d).abs <= (item (i + 1) - a_d).abs then
							Result := i
						else
							Result := i + 1
						end
						l_stop := True
					end
					i := i + 1
				end
			end
		ensure
			a_start_lowest_result: Result >= a_start
			a_end_highest_possible_result: Result <= a_end
			valid_result: valid_index (Result)
		end

	part_index (a_value: DOUBLE): INTEGER is
			-- Index of next greater value in `Current' to `a_value';
			-- `Current' is assumed to be sorted in ascending order.
		require
			not_empty: count > 0
		local
			i: INTEGER
			l_stop: BOOLEAN
		do
			if is_missing (a_value) then
				Result := count + 1
			elseif a_value < item (1) then
				Result := 1
			elseif a_value >= item (count) then
				Result := count + 1
			else
				from
					i := 1
				until
					(i > count - 1) or l_stop
				loop
					if item (i) <= a_value and item (i + 1) > a_value then
						Result := i + 1
						l_stop := True
					end
					i := i + 1
				end
			end
		ensure
			not_found_low: Result >= 1
			not_found_high: Result <= count + 1
		end

	part_index_by (a_value: DOUBLE; a_desc: BOOLEAN): INTEGER is
			-- Index of nearest item in `Current' to `a_value';
			-- `Current' is assumed to be sorted in `a_desc' order.
		local
			i: INTEGER
			l_stop: BOOLEAN
		do
			-- N.B. these first two clauses are inefficient, due to the assumption of sorted data.
			if is_missing(a_value) then
				Result := 0
			elseif a_value <= min then
				Result := index_of_min
			elseif a_value >= max then
				Result := index_of_max
			else
				from
					i := 1
				until
					(i > count - 1) or l_stop
				loop
					if a_desc then
						if item (i) >= a_value and item (i + 1) < a_value then
							if (item (i) - a_value).abs <= (item (i + 1) - a_value).abs then
								Result := i
							else
								Result := i + 1
							end
							l_stop := True
						end
					else
						if item (i) <= a_value and item (i + 1) > a_value then
							if (item (i) - a_value).abs <= (item (i + 1) - a_value).abs then
								Result := i
							else
								Result := i + 1
							end
							l_stop := True
						end
					end
					i := i + 1
				end
			end
		ensure
			part_index_by_not_negative: Result >= 0
			maximum_index: Result <= count
		end

	sector_index (a_value: DOUBLE): INTEGER is
			-- Index of next lower value in `Current' to `a_value';
			-- `Current' is assumed to be sorted in decending order.
		require
			not_empty: count > 0
		local
			i: INTEGER
			l_stop: BOOLEAN
		do
			if is_missing (a_value) then
				Result := count + 1
			elseif a_value >= item (1) then
				Result := 1
			elseif a_value < item (count) then
				Result := count + 1
			else
				from
					i := 1
				until
					(i > count -1) or l_stop
				loop
					if item (i) > a_value and item (i + 1) <= a_value then
						Result := i + 1
						l_stop := True
					end
					i := i + 1
				end
			end
		ensure
			not_found_low: Result <= count + 1
			not_found_high: Result >= 1
		end


	find_indice (a_vector: VECTOR): ROSE_LINEAR_ARRAY [INTEGER] is
			-- Indices of all items in `a_vector', according to `find_index';
			-- An additional index of 1 is provided at the beginning of the result.
			-- Yet another additional index of `count' is provided at the end of the result.
			-- If `Current' is not in ascending order, then the results will be meaningless.
			-- Otherwise, if `a_vector' is also sorted in ascending order, then the result will be
			--  (not necessarily strictly) monotonically increasing.
		require
			a_vector_not_void: a_vector /= Void
		local
			i, k: INTEGER
		do
			create Result.make (1, a_vector.count + 2)
			Result.put (1, 1)
			k := 1
			from
				i := 1
			until
				i > a_vector.count
			loop
				k := find_index (a_vector.item (i), k , count)
				Result.put (k, i + 1)
				i := i + 1
			end
			Result.put (count, a_vector.count + 2)
		ensure
			find_indice_not_void: Result /= Void
			count: Result.count = a_vector.count + 2
			first_result: Result [1] = 1
			last_result: Result [Result.count] = count
		end

	value_indice (a_value: DOUBLE): ROSE_LINEAR_ARRAY [INTEGER] is
			-- Indices of elements exactly matching `a_value';
			-- Duplicate of `indices_of_value'.
		local
			i: INTEGER
		do
			create Result.make_empty
			from
				i := 1
			until
				i > count
			loop
				if item (i) = a_value then
					Result.extend (i)
				end
				i := i + 1
			end
		ensure
			value_indice_not_void: Result /= Void
			maximum_count: Result.count <= count
			all_indices_valid: Result.for_all (agent valid_index)
		end

	count_by_value (a_value: DOUBLE): INTEGER is
			-- Number of elements matching exactly 'a_value'
		local
			i: INTEGER
			l_value_is_missing: BOOLEAN
		do
			from
				l_value_is_missing := is_missing(a_value)
				i := 1
			until
				i > count
			loop
				-- The following 'if' blob will increment Result when it finds 'a_value' in this vector - even if 'a_value' is NaN.
				if l_value_is_missing then
					if is_missing (item(i)) then
						Result := Result + 1		-- They were both NaN.
					end
				elseif not is_missing (item(i)) then
					if item (i) = a_value then
						Result := Result + 1		-- They were both the same non-NaN value.
					end
				end
				i := i + 1
			end
		ensure
			non_negative: Result >= 0
			maximum_result: Result <= count
		end

	distance (a_other : ABSTRACT_READABLE_VECTOR) : DOUBLE is
			-- Distance between `Current' and `a_other'
		require
			a_other_not_void : a_other /= Void
			same_size : count = a_other.count
		local
			i : INTEGER
			l_value : DOUBLE
			l_sum : DOUBLE
		do
			from
				i := 1
			until
				i > a_other.count
			loop
				l_value := item (i) - a_other [i]
				l_sum := l_sum + (l_value * l_value)
				i := i + 1
			end
			Result := sqrt (l_sum)
		ensure
			distance_non_negative: is_defined_double(Result) implies Result >= 0
		end

feature -- Agents

	is_less_than (a_double, a_cutoff: DOUBLE): BOOLEAN is
			-- Is `a_double' defined and less than `a_cutoff'?
		do
			Result := not(is_missing(a_double) or else is_missing(a_cutoff)) and then a_double < a_cutoff
		end

	is_greater_than (a_double, a_cutoff : DOUBLE) : BOOLEAN is
			-- Is `a_double' defined and greater than `a_cutoff'?
		do
			Result := not(is_missing(a_double) or else is_missing(a_cutoff)) and then a_double > a_cutoff
		end

	is_finite_and_less_than(a_double,a_cutoff : DOUBLE) : BOOLEAN
			--finite and less than
		do
			Result := is_defined (a_double) and then is_less_than(a_double,a_cutoff)
		end

	is_finite_and_greater_than(a_double,a_cutoff : DOUBLE) : BOOLEAN
			--finite and greater than
		do
			Result := is_defined (a_double) and then is_greater_than(a_double,a_cutoff)
		end

	indices_of_value (a_value: DOUBLE): ROSE_LINEAR_ARRAY [INTEGER] is
			-- Indices of elements that match `a_value';
			-- Duplicate of `value_indice', but better named.
		do
			Result := matching_item_indices (agent is_equal_value (?, a_value))
		ensure
			result_not_void: Result /= Void
			maximum_count: Result.count <= count
			all_indices_valid: Result.for_all (agent valid_index)
		end

	is_equal_value (a_double, a_target : DOUBLE) : BOOLEAN is
			-- Is `a_double' equal to `a_target'?
		do
			Result :=  (is_missing(a_double) and then is_missing(a_target)) or else a_double = a_target
		end

	is_undefined_double (a_double : DOUBLE) : BOOLEAN is
			-- Is `a_double' undefined?
		do
			Result := not is_defined_double (a_double)
		end

feature -- Measurement

	count: INTEGER
			-- Number of items in `Current'

	matching_item_count (a_test: FUNCTION [ANY, TUPLE [DOUBLE], BOOLEAN]): INTEGER is
			-- Number of elements that pass `a_test';
		require
			a_test_not_void: a_test /= Void
		local
			i: INTEGER
		do
			from i := 1 until i > count loop
				if a_test.item ([Current [i]]) then
					Result := Result + 1
				end
				i := i + 1
			end
		ensure
			non_negative_result: Result >= 0
			upper_bound_is_count: Result <= count
		end

	defined_item_count: INTEGER is
			-- Number of elements with a normal value;
			-- Abnormal values are NaNs, Infinities and legacy undefined doubles (-999.).
		do
			Result := matching_item_count (agent is_defined_double)
		ensure
			non_negative_result: Result >= 0
			upper_bound_is_count: Result <= count
		end

	all_defined: BOOLEAN is
			-- Are all elements defined?
		do
			Result := defined_item_count = count
		ensure
			definition: Result = (defined_item_count = count)
		end

feature -- Comparison

	is_equal (a_other: like Current): BOOLEAN is
			-- Is `a_other' attached to an object considered equal to current object?
		do
			Result := same_vector (a_other)
		end

	matching_item_flags (a_test: FUNCTION [ANY, TUPLE [DOUBLE], BOOLEAN]): ROSE_BOOLEAN_ARRAY is
			-- Test results for elements against `a_test'
		require
			a_test_not_void: a_test /= Void
		local
			i: INTEGER
		do
			create Result.make (lower, upper)
			from i := 1 until i > count loop
				Result [i] := a_test.item([Current [i]])
				i := i + 1
			end
		ensure
			result_not_void: Result /= Void
			same_count: Result.count = count
			matching_items: True -- Result.for_all_with_index (agent true_implies_a_test.item) -TODO: for_all_with_index needed in ROSE_LINEAR
			non_matching_items: True -- Result.for_all_with_index (agent false_implies_not_a_test.item) -TODO: for_all_with_index needed in ROSE_LINEAR
		end

	same_vector (a_other: ABSTRACT_READABLE_VECTOR): BOOLEAN is
			-- Does `Current' have same values as `a_other'?
			-- NaNs compare equal
		require
			a_other_not_void: a_other /= Void
		do
			if a_other.count = count then
				Result := for_all_with_index (agent same_double_as_other (?, ?, a_other))
			end
		ensure
			same_count: Result implies a_other.count = count
			all_values_equal: Result implies for_all_with_index (agent same_double_as_other (?, ?, a_other))
		end

feature -- Status report

	valid_index (i: INTEGER): BOOLEAN is
			-- Is `i' a valid position within `Current'?
		do
			Result := i >= lower and i <= upper
		ensure
			definition: Result = (i >= lower and i <= upper)
		end

	undefined_item_flags: ROSE_BOOLEAN_ARRAY is
			-- Flags show which elements abnormal values
			-- Abnormal values are NaNs, Infinities and legacy undefined doubles (-999.)
		do
			Result := matching_item_flags (agent is_undefined_double)
		ensure
			result_not_void: Result /= Void
			same_count: Result.count = count
		end

	defined_item_flags: ROSE_BOOLEAN_ARRAY is
			-- Flags show which elements have normal values
			-- Abnormal values are NaNs, Infinities and legacy undefined doubles (-999.)
		do
			Result := matching_item_flags (agent is_defined_double)
		ensure
			result_not_void: Result /= Void
			same_count: Result.count = count
			defined_values: Result.for_all_with_index (agent is_correct_defined_status)
		end

	positive_item_flags: ROSE_BOOLEAN_ARRAY is
			-- Flags show which elements have positive normal values
			-- Abnormal values are NaNs, Infinities and legacy undefined doubles (-999.)
		do
			Result := matching_item_flags (agent is_finite_and_greater_than (?, 0))
		ensure
			result_not_void: Result /= Void
			same_count: Result.count = count
		end

	negative_item_flags: ROSE_BOOLEAN_ARRAY is
			-- Flags show which elements have negative normal values
			-- Abnormal values are NaNs, Infinities and legacy undefined doubles (-999.)
		do
			Result := matching_item_flags (agent is_finite_and_less_than (?, 0))
		ensure
			result_not_void: Result /= Void
			same_count: Result.count = count
		end

	non_negative: BOOLEAN is
			-- Are all elements non-negative?
		local
			i: INTEGER
		do
			Result := True
			from i := 1 until (not Result) or (i > count) loop
				Result := Result and (Current [i] > 0 or Current [i] = 0) -- <= does not handle unavailables right
					i := i + 1
			end
		ensure
			all_non_negative: Result implies for_all (agent is_non_negative)
		end

	has_positive: BOOLEAN is
			-- Does `Current' contain at least one value greater than zero?
		local
			i: INTEGER
		do
			from
				i := 1
			until
				Result or else i > count
			loop
				if item(i) > 0.0 then
					Result := True
				end
				i := i + 1
			end
		end

	has_at_least_n_defined (n: INTEGER): BOOLEAN is
			-- Are there at least 'n' defined elements?
		require
			strictly_positive_n: n > 0
		local
			i, l_count: INTEGER
		do
			from
				i := 1
			until
				i > count or Result or count - i + l_count < n - 1
				-- Break loop when there aren't enough items left to
				-- make this true
			loop
				if is_defined_double (item (i)) then
					l_count := l_count + 1
					Result := l_count >= n
				end
				i := i + 1
			end
		end

	is_all_defined: BOOLEAN is
			-- Are all elements in this vector defined?
		local
			i: INTEGER
		do
			Result := True
			from
				i := 1
			until
				i > count or not Result
			loop
				Result := is_defined_double (item (i))
				i := i + 1
			end
		ensure
			empty_true: count = 0 implies Result
		end

	is_null: BOOLEAN is
			-- Are all elements in this vector undefined?
		local
			i: INTEGER
		do
			Result := True
			from
				i := 1
			until
				i > count or not Result
			loop
				Result := not is_defined_double (item (i))
				i := i + 1
			end
		ensure
			empty_true: count = 0 implies Result
		end

	is_above_or_equal (a_other: ABSTRACT_READABLE_VECTOR): BOOLEAN is
			-- Are all items from `Current' >= to positionally corresponding items from `a_other'?
		require
			a_other_not_void: a_other /= Void
			a_other_count_sufficient: a_other.count >= count
		local
			i: INTEGER
		do
			Result := True
			from
				i := 1
			until
				i > count or not Result
			loop
				Result := Result and then (not (a_other.item(i) > item (i)) )
				i := i + 1
			end
		ensure
			empty_true: count = 0 implies Result
		end

feature -- Status setting

	set_count (a_count: INTEGER) is
			-- Set `count' to `a_count'.
		require
			non_negative_count: a_count >= 0
			restrict_only: a_count <= count
		do
			count := a_count
		ensure
			count_set: count = a_count
		end

	set_name (a_name: STRING) is
			-- Set `name'.
		do
			name := a_name
		ensure
			name_set: name = a_name
		end

	set_names (a_names: ARRAY [STRING]) is
			-- Set `names'.
		do
			if a_names /= Void then
				names ?= a_names
				if names = Void then
					create names.make_from_array (a_names)
				end
				name_table_saved := Void
			end
		end

feature -- Iteration

	do_all (a_action: PROCEDURE [ANY, TUPLE [DOUBLE]]) is
			-- Apply `a_action' to every element of `Current'.
		require
			a_action_not_void: a_action /= Void
		local
			l_arg: TUPLE [DOUBLE]
			i: INTEGER
		do
			from
				create l_arg
				i := lower
			until
				i > upper
			loop
				l_arg.put_double (Current [i], 1)
				a_action.call (l_arg)
				i := i + 1
			end
		end

	do_all_with_index (a_action: PROCEDURE [ANY, TUPLE [DOUBLE, INTEGER]]) is
			-- Apply `a_action' to every element of `Current'.
			-- `a_action' receives element and it's index.
		require
			a_action_not_void: a_action /= Void
		local
			l_args: TUPLE [DOUBLE, INTEGER]
			i: INTEGER
		do
			from
				create l_args
				i := lower
			until
				i > upper
			loop
				l_args.put_double (Current [i], 1)
				l_args.put_integer (i, 2)
				a_action.call (l_args)
				i := i + 1
			end
		end

	do_if (a_action: PROCEDURE [ANY, TUPLE [DOUBLE]]; a_test: FUNCTION [ANY, TUPLE [DOUBLE], BOOLEAN]) is
			-- Apply `a_action' to every element of `Current' that satisfies `a_test'.
		require
			a_action_not_void: a_action /= Void
		local
			l_arg: TUPLE [DOUBLE]
			i: INTEGER
		do
			from
				create l_arg
				i := lower
			until
				i > upper
			loop
				l_arg.put_double (Current [i], 1)
				if a_test.item (l_arg) then
					a_action.call (l_arg)
				end
				i := i + 1
			end
		end

	do_if_with_index (a_action: PROCEDURE [ANY, TUPLE [DOUBLE, INTEGER]]; a_test: FUNCTION [ANY, TUPLE [DOUBLE, INTEGER], BOOLEAN]) is
			-- Apply `a_action' to every element of `Current' that satisfies `a_test'.
			-- `a_action' and `a_test' receive element and its index.
		require
			a_action_not_void: a_action /= Void
			a_test_not_void: a_test /= Void
		local
			l_args: TUPLE [DOUBLE, INTEGER]
			i: INTEGER
		do
			from
				create l_args
				i := lower
			until
				i > upper
			loop
				l_args.put_double (Current [i], 1)
				l_args.put_integer (i, 2)
				if a_test.item (l_args) then
					a_action.call (l_args)
				end
				i := i + 1
			end
		end

	for_all (a_test: FUNCTION [ANY, TUPLE [DOUBLE], BOOLEAN]): BOOLEAN is
			-- Is `a_test' true for all elements?
		require
			a_test_not_void: a_test /= Void
		local
			l_arg: TUPLE [DOUBLE]
			i: INTEGER
		do
			from
				create l_arg
				i := lower
				Result := True
			until
				i > upper or not Result
			loop
				l_arg.put (Current [i], 1)
				Result := a_test.item (l_arg)
				i := i + 1
			end
		end

	for_all_with_index (a_test: FUNCTION [ANY, TUPLE [DOUBLE, INTEGER], BOOLEAN]): BOOLEAN is
			-- Is `a_test' true for all elements?
			-- `a_test' receives element and it's index.
		require
			a_test_not_void: a_test /= Void
		local
			l_args: TUPLE [DOUBLE, INTEGER]
			i: INTEGER
		do
			from
				create l_args
				i := lower
				Result := True
			until
				i > upper or not Result
			loop
				l_args.put (Current [i], 1)
				l_args.put_integer (i, 2)
				Result := a_test.item (l_args)
				i := i + 1
			end
		end

	accumulation (a_initial: DOUBLE; a_function: FUNCTION [ANY, TUPLE [DOUBLE, DOUBLE], DOUBLE]): DOUBLE is
			-- Value accumulated by successive application of `a_function' to partial result and each element;
			-- Partial result is seeded by `a_initial'.
			-- E.g. `a_initial' = 0 and `a_function' = `+' yields sum of all elements.
		require
			accumulator_not_void: a_function /= Void
		local
			i: INTEGER
		do
			Result := a_initial
			from i := 1 until i > count loop
				Result := a_function.item ([Result, Current [i]])
				i := i + 1
			end
		ensure
			empty_result: (count = 0 and not isnan (a_initial)) implies Result = a_initial
			singleton_result: (count = 1 and not isnan (a_initial) and all_defined) implies Result = a_function.item ([a_initial, item (1)])
			recursive_result: (count > 1 and not isnan (a_initial) and all_defined) implies
				Result = sub_vector (2, count).accumulation (a_function.item ([a_initial, item (1)]), a_function)
		end

	pairwise_accumulation (a_initial: DOUBLE; a_accumulating_function, a_pairwise_function: FUNCTION [ANY, TUPLE [DOUBLE, DOUBLE], DOUBLE];
		a_other: ABSTRACT_READABLE_VECTOR): DOUBLE is
			-- Value accumulated by successive application of `a_accumulating_function,' to partial result
			-- and result of `a_pairwise_function' applied to each element and it's corresponing element in `a_other';
			-- Partial result is seeded by `a_initial'.
			-- E.g. `a_initial' = 0 and `a_accumulating_function,' = `+' and `a_pairwise_function' = `*'
			--  yields sum of products of all corresponding elements.
			-- All undefined doubles are ignored
		require
			accumulator_not_void: a_accumulating_function /= Void
			pairwise_function_not_void: a_pairwise_function /= Void
			other_vector_not_void: a_other /= Void
			consistent_size: a_other.count >= count
		local
			i: INTEGER
			l_value: DOUBLE
		do
			if is_defined_double (a_initial) then
				Result := a_initial
			end
			from i := 1 until i > count loop
				l_value := a_accumulating_function.item ([Result, a_pairwise_function.item ([Current [i], a_other [i]])])
				if is_defined_double (l_value) then
					Result := l_value
				end
				i := i + 1
			end
		end

	pairwise_do_all (a_pairwise_action: PROCEDURE [ANY, TUPLE [DOUBLE, DOUBLE]]; a_other: ABSTRACT_READABLE_VECTOR) is
			-- Apply action to items at same position from a pair
			-- of vectors of consistent sizes. The first argument of the
			-- action if from `Current' and the second from the same position
			-- in `a_other'.
		require
			action_not_void: a_pairwise_action /= Void
			other_vector_not_void: a_other /= Void
			consistent_size: a_other.count >= count
		local
			l_args: TUPLE [DOUBLE, DOUBLE]
			i: INTEGER
		do
			from
				create l_args
				i := 1
			until
				i > count
			loop
				l_args.put (Current [i], 1)
				l_args.put (a_other [i], 2)
				a_pairwise_action.call (l_args)
				i := i + 1
			end
		end

feature -- Transformed vectors

	unitized_vector : VECTOR is
			-- Vector with same direction as `Current' and magnitude 1
		local
			i : INTEGER
			l_sum : DOUBLE
		do
			create Result.make (count)
			from i := 1 until i > count loop
				l_sum := l_sum + (Current [i]^2)
				i := i + 1
			end
			l_sum := sqrt (l_sum)
			from  i := 1 until i > count loop
				Result [i] := Current [i] / l_sum
				i := i + 1
			end
		ensure
			unitized_vector_not_void: Result /= Void
			same_count: Result.count = count
			magnitude_1: approx_equal (1.0, Result.magnitude)
		end

	diff_ratio_vector (a_other: ABSTRACT_READABLE_VECTOR): VECTOR is
			-- Ratios of differences to sums of corresponding elements
		require
			other_vector_not_void: a_other /= Void
			same_size: count = a_other.count
		local
			i: INTEGER
		do
			create Result.make(count)
			from i := 1 until i > count loop
				Result [i] := check_ratio.fit (Current [i], a_other [i])
				i := i + 1
			end
		ensure
			result_not_void: Result /= Void
			new_object: as_readonly_vector /= Result and Current /= a_other
			same_count: Result.count = count
			elementwise_difference_sum_ratios: True -- Result.for_all_with_index (agent cant_be_bothered_to_write_it)
		end

	signed_diff_ratio_vector (a_other: ABSTRACT_READABLE_VECTOR): VECTOR is
			-- Ratios of signed differences to averages of corresponding elements
		require
			other_vector_not_void: a_other /= Void
			same_size: count = a_other.count
		local
			i: INTEGER
		do
			create Result.make(count)
			from i := 1 until i > count loop
				Result [i] := check_ratio.signed_diff (Current [i], a_other [i])
				i := i + 1
			end
		ensure
			result_not_void: Result /= Void
			new_object: as_readonly_vector /= Result and Current /= a_other
			same_count: Result.count = count
			elementwise_difference_sum_ratios: True -- Result.for_all_with_index (agent cant_be_bothered_to_write_it)
		end

	minus (a_margin: DOUBLE): VECTOR is
			-- New vector with all elements reduced by `a_margin'
		local
			i: INTEGER
		do
			create Result.make (count)
			from i := 1 until i > count loop
				Result [i] := Current [i] - a_margin
				i:=i+1
			end
		ensure
			result_not_void: Result /= Void
			new_object: as_readonly_vector /= Result
			same_count: Result.count = count
			elements_reduced_by_a_margin: Result.for_all_with_index (agent is_reduced_by (?, ?, a_margin))
		end

	square_roots: VECTOR is
			-- Square roots of elements of `Current'
		require
			all_elements_non_negative: non_negative
		local
			i: INTEGER
		do
			create Result.make (count)
			from i := 1 until i > count loop
				Result [i] := sqrt (Current [i])
				i := i + 1
			end
		ensure
			result_not_void: Result /= Void
			new_object: as_readonly_vector /= Result
			same_count: Result.count = count
			definition: Result.for_all_with_index (agent is_square_root_of)
		end

	squares: VECTOR is
			-- Squares of elements of `Current'
		local
			i: INTEGER
		do
			create Result.make(count)
			from i := 1 until i > count loop
				Result [i] := Current [i] * Current [i]
				i := i + 1
			end
		ensure
			result_not_void: Result /= Void
			same_count: Result.count = count
			definition: Result.for_all_with_index (agent is_square_of)
		end

	power(a_pow : DOUBLE) : VECTOR is
			-- integer power
		local
			i : INTEGER
			l_value : DOUBLE
		do
			create Result.make(count)
			from i := 1 until i > count loop
				l_value := (Current[i]) ^ a_pow
				Result[i] := l_value
				i := i + 1
			end
		ensure
			result_not_void: Result /= Void
			same_count: Result.count = count
		end

	abs: VECTOR is
			-- Absolute values of elements of `Current'
		local
			i: INTEGER
		do
			create Result.make(count)
			from i := 1 until i > count loop
				if is_defined_double (Current [i]) then
					Result [i] := dabs (Current [i])
				else
					Result [i] := constants.missing_double
				end
				i := i + 1
			end
			Result.set_names (names)
		ensure
			result_not_void: Result /= Void
			same_count: Result.count = count
			definition: Result.for_all_with_index (agent is_abs_of)
		end

	normalized: attached VECTOR is
			-- `Current' normalized so that elements sum to 1.0
		require
			strictly_positive_sum: count > 0 implies sum > 0
		local
			i: INTEGER
			l_sum: DOUBLE
		do
			l_sum := sum
			create Result.make (count)
			from i := 1 until i > count loop
				Result [i] := Current [i] / l_sum
				i := i + 1
			end
			if names/=void then
				result.set_names(names)
			end
		ensure
			result_not_void: Result /= Void
			same_count: Result.count = count
			sum_is_1_0: approx_equal (Result.sum, 1.0)
		end

	normalized_with_exponential: attached VECTOR is
			-- `Current' normalized with exponential so that elements sum to 1.0 and are positive.
		local
			i: INTEGER
			l_exp, l_sum: DOUBLE
		do
			create Result.make (count)
			from i := 1 until i > count loop
				l_exp := exp (Current[i])
				l_sum := l_sum + l_exp
				Result [i] := l_exp
				i := i + 1
			end
			from i := 1 until i > count loop
				Result [i] := Result[i] / l_sum
				i := i + 1
			end
		ensure
			result_not_void: Result /= Void
			same_count: Result.count = count
			sum_is_1_0: approx_equal (Result.sum, 1.0)
			result_non_negative: Result.non_negative
		end

	normalized_to_nearly_one (a_residual: DOUBLE): VECTOR is
			-- `Current' normalized to sum to 1.0 - `a_residual'
		require
			a_residual_non_zero: a_residual /= 0.0
		local
			l_factor : DOUBLE
		do
			l_factor := scaled (1.0 / (1.0 - a_residual)).sum
			if l_factor = 0.0 then
				create Result.make_constant (undefined_double, count)
			else
				Result := scaled (1.0 / l_factor)
			end
		ensure
			result_not_void: Result /= Void
			same_count: Result.count = count
			all_undefined_or_sum_is_nearly_1_0: Result.for_all (agent is_undefined_double) or approx_equal (Result.sum, 1.0 - a_residual)
		end

	normalized_with_residual (a_residual: DOUBLE): VECTOR is
			-- `Current' normalized so that elements sum to 1.0 + `a_residual'
		require
			a_residual_ok: a_residual /= - sum
		local
			i : INTEGER
			l_sum: DOUBLE
		do
			l_sum := sum + a_residual
			create Result.make (count)
			from i := 1 until i > count loop
				Result [i] := Current [i] / l_sum
				i := i + 1
			end
		ensure
			result_not_void: Result /= Void
			same_count: Result.count = count
		end

	normalized_square_roots: VECTOR is
			-- Square roots of (elements of `Current' normalized to sum to 1.0)
		require
			non_negative: non_negative
		local
			i: INTEGER
			l_sum: DOUBLE
		do
			l_sum := sum
			create Result.make (count)
			from i := 1 until i > count loop
				Result [i] := sqrt (Current [i] / l_sum)
				i := i + 1
			end
		ensure
			result_not_void: Result /= Void
			same_count: Result.count = count
		end

	scaled (a_weight: DOUBLE): attached VECTOR is
			-- `Current' scaled by `a_weight'
		local
			i: INTEGER
		do
			create Result.make(count)
			from i := 1 until i > count loop
				Result [i] := a_weight * Current [i]
				i := i + 1
			end
		ensure
			result_not_void: Result /= Void
			same_count: Result.count = count
			definition: Result.for_all_with_index (agent is_scaled_by (?, a_weight, ?))
		end

	transformed_vector (a_function: FUNCTION [ANY, TUPLE [DOUBLE], DOUBLE]): VECTOR is
			-- `Current' with each element transformed by `a_function'
		require
			transforming_function_not_void: a_function /= Void
		local
			i: INTEGER
		do
			create Result.make (count)
			from i := 1 until i > count loop
				Result [i] := a_function.item ([Current [i]])
				i := i + 1
			end
		ensure
			result_not_void: Result /= Void
			same_count: Result.count = count
			each_element_transformed: Result.for_all_with_index (agent is_transformed_by (?, a_function, ?))
		end

	transformed (a_function: FUNCTION [ANY, TUPLE [DOUBLE, INTEGER], DOUBLE]): VECTOR is
			-- `Current' with each element transformed by `a_function';
			-- Index of current item is passed as an additional argument.
		require
			transforming_function_not_void: a_function /= Void
		local
			i: INTEGER
		do
			create Result.make (count)
			from i := 1 until i > count loop
				Result [i] := a_function.item ([Current [i], i])
				i := i + 1
			end
		ensure
			result_not_void: Result /= Void
			same_count: Result.count = count
			each_element_transformed: -- True
		end

	xtx: SYMMETRIC_MATRIX is
			--	Dot product of `Current' as a column vector with `Current'
		do
			Result := to_row_matrix.xtx
		ensure
			xtx_not_void: Result /= Void
			same_width: Result.width = count
		end

	xtx_with_matrix (a_matrix: ABSTRACT_READABLE_MATRIX): DOUBLE is
			-- `Current' * a_matrix * `Current'.`transposed'
		require
			a_matrix_not_void: a_matrix /= Void
			height_matches_width : a_matrix.height = a_matrix.width
			width_matches_count : a_matrix.width = count
		do
			Result := (to_row_matrix.product (a_matrix) * to_column_matrix) [1, 1]
		end

	accumulation_with_partials (a_initial: DOUBLE; a_function: FUNCTION [ANY, TUPLE [DOUBLE, DOUBLE], DOUBLE]): VECTOR is
			-- Value accumulated by successive application of `a_function' to partial result and each element;
			-- Partial results are recorded in elements 1 to (count - 1)
			-- Partial result is seeded by `a_initial'.
			-- E.g. `a_initial' = 0 and `a_function' = `+' yields sum of all elements.
		require
			accumulator_not_void: a_function /= Void
			not_empty: count > 0
		do
			create Result.make (count)
			do_all_with_index (agent store_partial (Result, a_initial, ?, ?, a_function))
		ensure
			accumulation_with_partials_not_void: Result /= Void
			same_count: Result.count = count
			singleton_result: (count >= 1 and not isnan (a_initial) and all_defined) implies
				Result [1] = accumulation_with_partials (a_initial, a_function) [1]
			recursive_result: (count > 1 and not isnan (a_initial) and all_defined) implies
				Result.sub_vector (2, count).same_vector (sub_vector (2, count).accumulation_with_partials (Result [1], a_function))
		end

	bounded (a_lower, a_upper: ABSTRACT_READABLE_VECTOR): VECTOR is
			-- Copy of `Current' where all elements are bounded by the interval (`a_lower', `a_upper')
		require
			a_lower_not_void: a_lower /= Void
			a_lower_large_enough: a_lower.count >= count
			a_upper_not_void: a_upper /= Void
			a_upper_large_enough: a_upper.count >= count
		do
			Result := transformed (agent bounded_element (?, ?, a_lower, a_upper))
		ensure
			result_not_void: Result /= Void
			same_count: Result.count = count
		end

	winsorized (a_winsor: ABSTRACT_READABLE_VECTOR): VECTOR is
			-- Copy of `Current' where all elements are bounded by the interval (-`a_winsor', +`a_winsor')
		require
			a_winsor_not_void: a_winsor /= Void
			a_winsor_large_enough: a_winsor.count >= count
		do
			Result := bounded (a_winsor.abs.scaled (-1), a_winsor.abs)
		ensure
			result_not_void: Result /= Void
			same_count: Result.count = count
		end

	uniformly_winsorized (a_winsor: DOUBLE): VECTOR is
			-- Copy of `Current' where all elements are bounded by the interval (-`a_winsor', +`a_winsor')
		require
			positive_winsor: a_winsor > 0
		do
			Result := winsorized ((create {UNIT_VECTOR}.make (count)).scaled (a_winsor))
		ensure
			result_not_void: Result /= Void
			same_count: Result.count = count
		end

feature {NONE} -- Transformed vectors

	store_partial (a_destination: VECTOR; a_initial, a_value: DOUBLE; a_index: INTEGER;
		a_function: FUNCTION [ANY, TUPLE [DOUBLE, DOUBLE], DOUBLE]) is
			-- Store partial result of `a_function' in `a_destination [a_index]'.
		require
			a_destination_not_void: a_destination /= Void
			same_count: a_destination.count = count
			a_index_large_enough: a_index > 0
			a_index_small_enough: a_index <= count
			a_function_not_void: a_function /= Void
		local
			l_initial: DOUBLE
		do
			if a_index = 1 then
				l_initial := a_initial
			else
				l_initial := a_destination [a_index - 1]
			end
			a_destination [a_index] := a_function.item ([l_initial, a_value])
		ensure
			initial_case: (a_index = 1 and not isnan (a_initial) and all_defined) implies
				(isnan (a_destination [a_index]) or a_destination [a_index] =  a_function.item ([a_initial, a_value]))
			non_initial_case: (a_index > 1 and not isnan (a_initial) and all_defined) implies
				(isnan (a_destination [a_index]) or a_destination [a_index] =  a_function.item ([a_destination [a_index - 1], a_value]))
		end

feature -- Conversion

	to_column_matrix: MATRIX is
			--	`Current' as a single-column matrix
		local
			i: INTEGER
		do
			create Result.make (count, 1)
			from i := 1 until i > count loop
				Result [i, 1] := Current [i]
				i := i + 1
			end
			if names /= Void then
				Result.set_row_names (names)
			end
			if name /= Void then
				Result.set_column_names (<<name>>)
			end
		ensure
			result_not_void: Result /= Void
			correct_row_count: Result.height = count
			one_column: Result.width = 1
			values_set: True -- TODO
			same_row_names: names /= Void implies True -- TODO
			same_column_name: name /= Void implies
				(Result.column_names /= Void and then (Result.column_names.count = 1
				and name.same_string (Result.column_names [1])))
		end

	to_row_matrix: MATRIX is
			--	`Current' as a new single-row matrix
		local
			i: INTEGER
		do
			create Result.make (1, count)
			from i := 1 until i > count loop
				Result [1, i] := Current [i]
				i := i + 1
			end
			if names /= Void then
				Result.set_column_names (names)
			end
			if name /= Void then
				Result.set_row_names (<<name>>)
			end
		ensure
			result_not_void: Result /= Void
			correct_column_count: Result.width = count
			one_row: Result.height = 1
			values_set: True -- TODO
			same_column_names: names /= Void implies True -- TODO
			same_row_name: name /= Void implies
				(Result.row_names /= Void and then (Result.row_names.count = 1
				and name.same_string (Result.row_names [1])))
		end

	to_diagonal_matrix: DIAGONAL_MATRIX is
			--	`Current' as a new diagonal matrix
		local
			i: INTEGER
		do
			create Result.make (count)
			from i := 1 until i > count loop
				Result [i, i] := Current [i]
				i := i + 1
			end
		ensure
			result_not_void: Result /= Void
			correct_column_count: Result.width = count
			correct_row_count: Result.height = count
			values_set: True -- TODO
		end

	to_matrix (a_height, a_width: INTEGER): MATRIX is
			-- `Current' as an `a_height' by `a_width' matrix
		require
			correct_length: count = a_height * a_width
		local
			i, j, k: INTEGER
			l_finished, l_finished_inner: BOOLEAN
		do
			from
				create Result.make (a_height, a_width)
				j := 1
			invariant
				result_not_void: Result /= Void
				column_addressable: l_finished xor Result.valid_column_index (j)
			variant
				increasing_column_index: a_width + 1 - j
			until
				l_finished
			loop
				from
					l_finished_inner := False
					i := 1
				invariant
					row_addressable: l_finished_inner xor Result.valid_row_index (i)
					indexable: k = 0 xor valid_index (k)
				variant
					increasing_row_index: a_height + 1 - i
				until
					l_finished_inner
				loop
					k := (j - 1) * a_height + i
					Result [i, j] := Current [k]
					i := i + 1
					l_finished_inner := i > a_height
				end
				j := j + 1
				l_finished := j > a_width
			end
		ensure
			to_matrix_not_void: Result /= Void
			correct_width: Result.width = a_width
			correct_height: Result.height = a_height
			round_trip: Result.to_vector.same_vector (Current)
			no_column_names: Result.column_names = Void
			no_row_names: Result.row_names = Void
			unnamed_result: Result.name = Void
		end

	to_array: attached ROSE_LINEAR_ARRAY [DOUBLE] is
			--	`Current' as a new array
		local
			i : INTEGER
		do
			create Result.make (1, count)
			from i := 1 until i > count loop
				Result [i] := Current [i]
				i := i + 1
			end
		ensure
			result_not_void: Result /= Void
			same_count: Result.count = count
			values_set: True -- TODO
		end

	as_readonly_vector: ABSTRACT_READABLE_VECTOR is
			-- Read-only version of `Current'
		do
			Result := Current
		ensure
			result_not_void: Result /= Void
			same_object: Result = Current
		end

	to_vector: VECTOR is
			--	`Current' as a new VECTOR
		local
			i: INTEGER
		do
			Create Result.make (count)
			from i:= 1 until i > count loop
				Result [i] := Current [i]
				i := i + 1
			end
		ensure
			result_not_void: Result /= Void
--not legal in 5.6.1218			new_object: as_readonly_vector /= Result
			same_count: Result.count = count
			values_set: Result.for_all_with_index (agent equal_at (?, ?, 0))
			different : Result /= Current
		end

	to_reversed_vector: VECTOR is
			-- Current as a new vector but in reversed order
		do
			create Result.make_from_array (Current.to_array.reversed)
		ensure
			to_reversed_vector_not_void: Result /= Void
			same_count: Result.count = count
			reversed: for_all_with_index (agent same_double_as_other_reversed (?, ?, Result))
		end

	as_vector : VECTOR is
			--VECTOR version of Current, possibly Current
		do
			Result := to_vector
		ensure
			as_vector_not_void: Result /= Void
			same_vector: Result.same_vector (Current)
		end

	sorted (a_descending: BOOLEAN): VECTOR is
			-- Items of `Current' sorted in `a_descending' order;
			-- Item names are preserved.
		local
			l_order: like ascending_order
			i: INTEGER
			l_names: like names
		do
			create Result.make (count)
			if a_descending then
				l_order := descending_order
			else
				l_order := ascending_order
			end
			if names /= Void then
				create l_names.make (1, count)
			end
			from
				i := 1
			until
				i > count
			loop
				Result.put (item (l_order.item (i)), i)
				if names /= Void then
					l_names.put (names.item (l_order.item (i)), i)
				end
				i := i + 1
			end
			if names /= Void then
				Result.set_names (l_names)
			end
		ensure
			sorted_not_void: Result /= Void
			no_name: Result.name = Void
			same_count: Result.count = count
			same_names: names /= Void implies Result.names /= Void
			ascending_order: True -- lack of stability means the following is good in spirit, but not
			-- necessarily in practice: a_descending = False implies
				-- Result.ascending_order.same_items ((1 |..| count).as_array)
			descending_order: True -- See above: a_descending = True implies
				-- Result.descending_order.same_items ((1 |..| count).as_array)
		end

	sorted_ascending: VECTOR is
			-- Items of `Current' sorted in ascending order;
			-- Item names are preserved.
		do
			Result := sorted (False)
		ensure
			sorted_ascending_not_void: Result /= Void
			no_name: Result.name = Void
			same_count: Result.count = count
			same_names: names /= Void implies Result.names /= Void
			ascending_order: True -- See `sorted': Result.ascending_order.same_items ((1 |..| count).as_array)
		end

	sorted_descending: VECTOR is
			-- Items of `Current' sorted in descending order;
			-- Item names are preserved.
		do
			Result := sorted (True)
		ensure
			sorted_descending_not_void: Result /= Void
			no_name: Result.name = Void
			same_count: Result.count = count
			same_names: names /= Void implies Result.names /= Void
			descending_order: True -- See `sorted': Result.descending_order.same_items ((1 |..| count).as_array)			
		end

	weighted_ranking (a_weights: VECTOR): VECTOR is
			-- This is to compute ranking percentile from 0 to 1
			-- a_weight must be positive or 0
		local
			l_order: ROSE_LINEAR_ARRAY [INTEGER]
			l_weights, l_ordered_weights, l_cum_weights: VECTOR
			i, j: INTEGER
		do
			l_order := ascending_order
			if a_weights = Void then
				create l_weights.make_identity (count)
			end
			create l_ordered_weights.make (count)
			from
				i := 1
			until
				i > count
			loop
				l_ordered_weights.put (l_weights.item(i), l_order.item (i))
				i := i + 1
			end
			l_ordered_weights := l_ordered_weights.normalized
			l_cum_weights := l_ordered_weights.cumulative_sum
			create Result.make (count)
			from
				i := 1
			until
				i > count
			loop
				j := l_order.item (i)
				Result.put (l_cum_weights.item (j) - l_ordered_weights.item (j) / 2, i)
				i := i + 1
			end
		ensure
			weighted_ranking_not_void: Result /= Void
			same_count: Result.count = count
		end

	subset (a_indices: ARRAY [INTEGER]): VECTOR is
			-- Vector made up of elements whose position in `Current' is in `a_indices'
		require
			indices_not_void: a_indices /= Void
			all_indices_in_range: a_indices.for_all (agent valid_index)
			-- The following commented-out assertion-sketch is necesary for the name `subset' to be meaningfull,
			--  but it can't be proven wrong by a post-condition (`Current' would need to be a set for a start.
			--  also, VECTOR implies order), so isn't of practical use, even if we could write it,
			--  so instead, we need a better name:
			-- no_duplicate_indices: need an inline agent along the lines of: a_indices.for_all (agent (occurrences (?) = 1))
		local
			i, k: INTEGER
		do
			k := a_indices.count
			create Result.make (k)
			from i := 1 until i > k loop
				Result [i] := Current [a_indices [i]]
				i := i + 1
			end
			if names/=void then
				result.set_names(names.subset (a_indices))
			end
		ensure
			result_not_void: Result /= Void
			new_object: as_readonly_vector /= Result
			correct_count: Result.count = a_indices.count
		end

	subset3 (a_indices: ARRAY [INTEGER]): VECTOR is
			-- Vector made up of elements whose position in `Current' is in `a_indices'
			-- Comparing to subset3, subset3 has a loose requirement. If a indices[i] is invalid, just put zero in the new vector at position i
		require
			indices_not_void: a_indices /= Void
			-- The following commented-out assertion-sketch is necesary for the name `subset' to be meaningfull,
			--  but it can't be proven wrong by a post-condition (`Current' would need to be a set for a start.
			--  also, VECTOR implies order), so isn't of practical use, even if we could write it,
			--  so instead, we need a better name:
			-- no_duplicate_indices: need an inline agent along the lines of: a_indices.for_all (agent (occurrences (?) = 1))
		local
			i, k, l_index: INTEGER
		do
			k := a_indices.count
			create Result.make (k)
			from i := 1 until i > k loop
				l_index := a_indices [i]
				if l_index >= lower and then l_index <= upper then
					Result [i] := Current [l_index]
				end
				i := i + 1
			end
			if names/=void then
				result.set_names(names.subset_loose (a_indices))
			end
		ensure
			result_not_void: Result /= Void
			new_object: as_readonly_vector /= Result
			correct_count: Result.count = a_indices.count
		end

	positive_vector: VECTOR is
			-- Copy of `Current' where all elements are defined and >= 0
		local
			i: INTEGER
		do
			create Result.make (count)
			from i := 1 until i > count loop
				if is_defined_double (Current [i]) and Current [i] > 0 then
					Result [i] := Current [i]
				end
				i := i + 1
			end
		ensure
			result_not_void: Result /= Void
			new_object: as_readonly_vector /= Result
			same_count: Result.count = count
			definition: for_all_with_index (agent is_same_positive (?, ?, Result))
		end

	negative_vector: VECTOR is
			-- Copy of `Current' where all elements are defined and <= 0
		local
			i: INTEGER
		do
			create Result.make (count)
			from i := 1 until i > count loop
				if is_defined_double (Current [i]) and Current [i] < 0 then
					Result [i] := Current [i]
				end
				i := i + 1
			end
		ensure
			result_not_void: Result /= Void
			new_object: as_readonly_vector /= Result
			same_count: Result.count = count
			definition: for_all_with_index (agent is_same_negative (?, ?, Result))
		end

	ascending_order: ROSE_LINEAR_ARRAY [INTEGER] is
			-- Indices of elements of `Current' to specify an ascending order;
			-- N.B. this order is not necessarily stable. I.e. a vector that is already
			--  sorted in ascending order will not necessarily produce the sequnce
			--  1 |..| count from this call.
		local
			i: INTEGER
		do
			create Result.make (1, count)
			from i := 1 until i > count loop
				Result [i] := i
				i := i + 1
			end
			Result.sort_by (agent are_elements_ascending)
		ensure
			result_not_void: Result /= Void
			same_count: Result.count = count
			all_indices_valid: Result.for_all (agent valid_index)
			ascending_order_specification: Result.for_all_with_index (agent is_next_element_greater (?, ?, Result))
		end

	descending_order: ROSE_LINEAR_ARRAY [INTEGER] is
			-- Indices of elements of `Current' to specify a descending order;
			-- N.B. this order is not necessarily stable. I.e. a vector that is already
			--  sorted in descending order will not necessarily produce the sequnce
			--  1 |..| count from this call.
		local
			i: INTEGER
		do
			create Result.make (1, count)
			from i := 1 until i > count loop
				Result [i] := i
				i := i + 1
			end
			Result.sort_by (agent are_elements_descending)
		ensure
			result_not_void: Result /= Void
			same_count: Result.count = count
			descending_order_specification: Result.for_all_with_index (agent is_next_element_less (?, ?, Result))
		end

	ascending_ranks : INTEGER_VECTOR is
			-- Integer array indicating each element's rank: rank 1 is min and rank count is max
		local
			i : INTEGER
			l_order : ROSE_LINEAR_ARRAY[INTEGER]
		do
			l_order := ascending_order
			create Result.make(1, count)
			from i := 1 until i > count loop
				Result[l_order[i]] := i
				i := i + 1
			end
		ensure
			ascending_ranks_not_void: Result /= Void
			same_count: Result.count = count
			definition: (1 |..| count).for_all (agent are_adjacent_ascending_ranks_consistent (?, Result))
		end

	descending_ranks : INTEGER_VECTOR is
			-- integer array indicating each element's rank: rank 1 is max and rank count is min
		local
			i : INTEGER
			l_order : ROSE_LINEAR_ARRAY[INTEGER]
		do
			l_order := descending_order
			create Result.make(1, count)
			from i := 1 until i > count loop
				Result[l_order[i]] := i
				i := i + 1
			end
		ensure
			descending_ranks_not_void: Result /= Void
			same_count: Result.count = count
			definition: (1 |..| count).for_all (agent are_adjacent_descending_ranks_consistent (?, Result))
		end
feature -- Duplication

	sub_vector (a_start, a_end: INTEGER): VECTOR is
			-- Slice of `Current' from `a_start' to `a_end'
		require
			valid_start_index: valid_index (a_start)
			valid_end_index: valid_index (a_end)
			valid_range: a_end >= a_start
		local
			i, k: INTEGER
			l_end: INTEGER
		do
			k := a_end - a_start + 1
			create Result.make (k)
			from i := 1 until i > k loop
				Result [i] := Current [a_start + i - 1]
				i := i + 1
			end
			if names /= Void and then names.valid_index (a_start) then
				from
					l_end := a_end
				until
					names.valid_index (l_end)
				loop
					l_end := l_end - 1
				end
				Result.set_names (names.subarray (a_start, l_end))
			end
		ensure
			result_not_void: Result /= Void
			new_object: as_readonly_vector /= Result
			correct_count: Result.count = a_end - a_start + 1
			correct_result: Result.for_all_with_index (agent equal_at (?, ?, a_start - 1))
		end

	sub_vector_by_flags (a_flags: ARRAY [BOOLEAN]): VECTOR is
			-- Sub-vector of `Current' where elements are included if corresponding element in `a_flag' is `True'
		require
			a_flags_not_void: a_flags /= Void
			not_too_many_flags: a_flags.count <= count
		local
			i, l_count: INTEGER
		do
			from i := a_flags.lower until i > a_flags.upper loop
				if a_flags [i] then
					l_count := l_count + 1
				end
				i := i + 1
			end
			create Result.make (l_count)
			l_count := 0
			from i := a_flags.lower until i > a_flags.upper loop
				if a_flags [i] then
					l_count := l_count + 1
					Result [l_count] := Current [i - a_flags.lower + 1]
				end
				i := i + 1
			end
		ensure
			result_not_void: Result /= Void
			new_object: as_readonly_vector /= Result
		end

	matching_items (a_test: FUNCTION [ANY, TUPLE [DOUBLE], BOOLEAN]): VECTOR is
			-- Elements that pass `a_test'
		require
			a_test_not_void: a_test /= Void
		do
			Result := subset (matching_item_indices (a_test))
		ensure
			result_not_void: Result /= Void
			new_object: as_readonly_vector /= Result
			maximum_count: Result.count <= count
			all_values_pass_test: Result.for_all (a_test)
		end


feature -- Basic operations

	infix "+" (a_other: ABSTRACT_READABLE_VECTOR): VECTOR is
			-- Sums of corresponding elements
		require
			other_vector_not_void: a_other /= Void
		local
			i: INTEGER
		do
			create Result.make (count)
			from i := 1 until i > count loop
				if is_defined_double (Current [i]) and is_defined_double (a_other [i]) then
					Result [i] := Current [i] + a_other [i]
				else
					Result [i] := constants.missing_double
				end
				i := i + 1
			end
		ensure
			result_not_void: Result /= Void
			new_object: as_readonly_vector /= Result and a_other /= Result
			same_count: Result.count = count
			elementwise_sums: Result.for_all_with_index (agent is_correct_sum (?, ?, a_other))
		end

	infix "-" (a_other: ABSTRACT_READABLE_VECTOR): VECTOR is
			-- Differences of corresponding elements
		require
			other_vector_not_void: a_other /= Void
			same_size: count = a_other.count
		local
			i: INTEGER
		do
			create Result.make (count)
			from i := 1 until i > count loop
				if is_defined_double (Current [i]) and is_defined_double (a_other [i]) then
				Result [i] := Current [i] - a_other [i]
				else
					Result [i] := constants.missing_double
				end
				i := i + 1
			end
		ensure
			result_not_void: Result /= Void
			new_object: as_readonly_vector /= Result and a_other /= Result
			same_count: Result.count = count
			elementwise_differences: Result.for_all_with_index (agent is_correct_difference (?, ?, a_other))
		end

	infix "*" (a_other: ABSTRACT_READABLE_VECTOR): DOUBLE is
			-- Sum of products of corresponding elements;
			-- See also `{VECTOR/SPARSE_VECTOR}.dot_product'.
		require
			other_vector_not_void: a_other /= Void
			same_size: count = a_other.count
		local
			i: INTEGER
		do
			from
				i := 1
			until
				i > count
			loop
				if is_defined_double (Current [i]) and is_defined_double (a_other [i]) then
					Result := Result + Current [i] * a_other [i]
				end
				i := i + 1
			end
		ensure
			not_NaN: Result = Result
			sum_of_products: (is_all_defined and a_other.is_all_defined) implies
				Result = pairwise_accumulation (0.0, agent doubles_sum, agent doubles_product, a_other)
		end

	cross (a_other: ABSTRACT_READABLE_VECTOR): VECTOR is
			-- Products of corresponding elements
		require
			a_other_not_void: a_other /= Void
			a_other_not_shorter: a_other.count >= count
		local
			i: INTEGER
		do
			create Result.make (count)
			from i := 1 until i > count loop
				if is_defined_double (Current [i]) and is_defined_double (a_other [i]) then
					Result [i] := Current [i] * a_other [i]
				else
					Result [i] := constants.missing_double
				end
				i := i + 1
			end
		ensure
			result_not_void: Result /= Void
			new_object: as_readonly_vector /= Result and a_other /= Result
			same_count: Result.count = count
			elementwise_products: Result.for_all_with_index (agent is_correct_product (?, ?, a_other))
		end

	element_product: DOUBLE is
			-- Product of all elements
		local
			i: INTEGER
		do
			Result := 1
			from i := 1 until i > count loop
				Result := Result * Current [i]
				i := i + 1
			end
		ensure
			element_product: Result = accumulation (1.0, agent doubles_product)
		end

	squared_norm: DOUBLE is
			-- Sum of squares of `Current'
		do
			Result := Current * Current
		ensure
			conditionally_defined_result: for_all (agent is_defined_double) implies is_defined_double (Result)
		end

	root_mean_square : DOUBLE is
			-- root mean square
		do
			Result := squared_norm / count
			Result := sqrt (Result)
		end

	sub_sum (a_start, a_end: INTEGER): DOUBLE is
			-- Sum of all defined elements from `a_start' to `a_end' inclusive
		require
			valid_start_index: valid_index (a_start)
			valid_end_index: valid_index (a_end)
		local
			i: INTEGER
		do
			from i := a_start until i > a_end loop
				if is_defined_double (Current [i]) then
					Result := Result + Current [i]
				end
				i := i + 1
			end
		ensure
			sub_sum: True -- very difficult
		end

	sum: DOUBLE is
			-- Sum of all defined elements
		do
			if count > 0 then
				Result := sub_sum (1, count)
			end
		ensure
			empty_result: count = 0 implies Result = 0
			non_empty_result: count > 0 implies Result = sub_sum (1, count)
		end

	positive_sum: DOUBLE is
			-- Sum of all defined positive elements
		local
			i: INTEGER
		do
			from i := 1 until i > count loop
				if is_defined_double (Current [i]) and Current [i] > 0 then
					Result := Result + Current [i]
				end
				i := i + 1
			end
		ensure
			non_negative_result: Result >= 0
		end

feature -- Statistics

	average: DOUBLE is
			-- Arithmetic mean of all elements;
			-- Undefined elements are treated as 0.0
		require
			not_empty: count > 0
		do
			Result := sum / count
		ensure
			definition: Result = sum / count
		end

	geometric_mean : DOUBLE
			-- nth root of the product of the values where n is the count
		do
			if count > 0 then
				Result := exp(ln.sum / count)
			end
		end

	average_or_undefined: DOUBLE
			-- median, returns undefined_double if no elements
		do
			if count = 0 then
				Result := undefined_double
			else
				Result := average
			end
		ensure
			definition_1: count > 0 implies Result = average
			definition_1: count = 0 implies not is_defined_double (Result)
		end

	weighted_mean (a_weights: ABSTRACT_READABLE_VECTOR): DOUBLE is
			-- Arithmetic mean of all elements weighted by `a_weights';
			-- Elements are ignored if undefined, or if their weight is undefined.
			-- See `average' for unweighted mean.
		require
			count_matches: a_weights /= Void implies a_weights.count = count
		local
			l_weights: ABSTRACT_READABLE_VECTOR
			l_total_weight, l_item, l_weight: DOUBLE
			i: INTEGER
		do
			if count > 0 then
				if a_weights = Void then
					create {UNIT_VECTOR} l_weights.make (count)
				else
					l_weights := a_weights
				end
				from i := 1 until i > count loop
					l_item := Current [i]
					l_weight := l_weights [i]
					if is_defined_double (l_item) and is_defined_double (l_weight) then
						Result := Result + l_item * l_weight
						l_total_weight := l_total_weight + l_weight
					end
					i := i + 1
				end
				if l_total_weight /= 0.0 then
					Result := Result / l_total_weight
				end
			end
		ensure
			defined_result: is_defined_double (Result)
		end

	weighted_mean_with_unavailables (a_weights: ABSTRACT_READABLE_VECTOR; a_defined: FUNCTION [ANY, TUPLE [INTEGER], BOOLEAN]): DOUBLE is
			-- Arithmetic mean of all elements weighted by `a_weights';
			-- Elements or weights are ignored if undefined according to `a_defined'.
		require
			count_matches: a_weights /= Void implies a_weights.count = count
		local
			l_weights: ABSTRACT_READABLE_VECTOR
			l_total_weight, l_weight: DOUBLE
			i: INTEGER
		do
			if a_defined = Void then
				Result := weighted_mean (a_weights)
			else
				if a_weights = Void then
					create {UNIT_VECTOR} l_weights.make (count)
				else
					l_weights := a_weights
				end
				from i := 1 until i > count loop
					if a_defined.item ([i]) then
						l_weight := l_weights [i]
						Result := Result + Current [i] * l_weight
						l_total_weight := l_total_weight + l_weight
					end
					i := i + 1
				end
				if l_total_weight /= 0.0 then
					Result := Result / l_total_weight
				end
			end
		ensure
			result_may_not_be_defined: True
		end

	has_sufficient_weight_with_valid_values (a_weights: ABSTRACT_READABLE_VECTOR; a_defined: FUNCTION [ANY, TUPLE [INTEGER], BOOLEAN]; a_min_wt_with_valid_values: DOUBLE): DOUBLE is
			-- Returns a 1 if the weight of valid values is greater than a_min_wt_with_valid_values, else 0
		require
			count_matches: a_weights /= Void implies a_weights.count = count
			args_not_void: a_defined /= Void
		local
			l_weights: ABSTRACT_READABLE_VECTOR
			l_total_weight, l_weight, l_sum_weight: DOUBLE
			i: INTEGER
		do
			if a_weights = Void then
				create {UNIT_VECTOR} l_weights.make (count)
			else
				l_weights := a_weights
			end
			from i := 1 until i > count loop
				l_sum_weight := l_sum_weight + l_weights[i]
				if a_defined.item ([i]) then
					l_weight := l_weights [i]
					l_total_weight := l_total_weight + l_weight
				end
				i := i + 1
			end
			if l_total_weight /= 0.0 and then l_total_weight/l_sum_weight > a_min_wt_with_valid_values then
				Result := 1
			end
		end

	median_or_undefined: DOUBLE
			-- median, returns undefined_double if no elements
		do
			if count = 0 then
				Result := undefined_double
			else
				Result := median
			end
		ensure
			definition_1: count > 0 implies Result = median
			definition_1: count = 0 implies not is_defined_double (Result)
		end

	median: DOUBLE is
			-- Median of elements of `Current'
		require
			non_empty_vector: count > 0
		local
			l_med: INTEGER
			l_ascending_order: like ascending_order
		do
			l_med := count // 2
			if count = 1 then
				Result := Current [1]
			else
				l_ascending_order := ascending_order
				if count - l_med * 2 = 0 then
					Result := (Current [l_ascending_order [l_med]] + Current [l_ascending_order [l_med + 1]]) / 2.0
				else
					Result := Current [l_ascending_order [l_med + 1]]
				end
			end
		ensure
			middle_of_odd_count: count \\ 2 /= 0 implies ieee_is_eq(Result, Current [ascending_order [count // 2 + 1]])
			average_of_even_count: count > 0 and count \\ 2 = 0 implies
				ieee_is_eq(Result,(Current [ascending_order [count // 2]] + Current [ascending_order [count // 2 + 1]]) / 2.0)
		end

	standard_deviation_from_mean (a_mean: DOUBLE): DOUBLE is
			-- Standard deviation from `a_mean' of elements of `Current'
		local
			i: INTEGER
			l_value, l_total: DOUBLE
		do
			from i := 1 until i > count loop
				l_value := Current [i] - a_mean
				l_total := l_total  + l_value * l_value
				i := i + 1
			end
			if count > 1 then
				Result := sqrt (l_total / (count - 1))
			else
				Result := sqrt (l_total)
			end
		ensure
			non_negative_result: Result >= 0
		end

	weighted_standard_deviation (a_weights: ABSTRACT_READABLE_VECTOR): DOUBLE is
			-- Standard deviation of elements of `Current' weighted by `a_weights';
			-- Elements are ignored if undefined, or if their weight is undefined.
			-- N.B. This uses a different divisor to SAS, so you will get different results.
		require
			count_matches: a_weights /= Void implies a_weights.count = count
		local
			d: DOUBLE
			d0: DOUBLE
			i, l_count: INTEGER
			l_weight_sum, l_item, l_weight: DOUBLE
			l_weight_mean, l_weighted_mean: DOUBLE
			l_weights: ABSTRACT_READABLE_VECTOR
		do
			if count > 0 then
				if a_weights = Void then
					create {UNIT_VECTOR} l_weights.make (count)
				else
					l_weights := a_weights
				end
				l_weighted_mean := weighted_mean (l_weights)
				from i := 1 until i > count loop
					l_item := Current [i]
					l_weight := l_weights [i]
					if is_defined_double (l_item) and is_defined_double (l_weight) then
						d0 := l_item - l_weighted_mean
						d := d + d0 * d0 * l_weight
						l_weight_sum := l_weight_sum + l_weight
						l_count := l_count + 1
					end
					i := i + 1
				end
				if l_count > 1 then
					l_weight_mean := l_weight_sum / l_count
					Result := sqrt (d / (l_weight_sum - l_weight_mean))
				end
			end
		end

	weighted_standard_deviation_with_unavailables (a_weights: VECTOR; a_defined: FUNCTION [ANY, TUPLE [INTEGER], BOOLEAN]): DOUBLE is
			-- Standard deviation of elements of `Current' weighted by `a_weights';
			-- Elements are ignored if undefined according to `a_defined'.
		require
			count_matches: a_weights /= Void implies a_weights.count = count
		local
			d: DOUBLE
			d0: DOUBLE
			i: INTEGER
			l_weight_sum : DOUBLE
			l_weight_mean,l_weighted_mean : DOUBLE
			l_weights : ABSTRACT_READABLE_VECTOR
			l_count : INTEGER
		do
			if a_defined = Void then
				Result := weighted_standard_deviation (a_weights)
			else
				if a_weights = void then
					create {UNIT_VECTOR} l_weights.make (count)
				else
					l_weights := a_weights
				end
				l_weighted_mean := weighted_mean_with_unavailables(l_weights, a_defined)
				d := 0
				from i := 1 until i > count loop
					if a_defined.item([i]) then
						d0 := item (i) - l_weighted_mean
						d := d + d0* d0 * l_weights.item (i)
						l_weight_sum := l_weight_sum + l_weights.item(i)
						l_count := l_count + 1
					end
					i := i + 1
				end
				if l_count > 1 then
					l_weight_mean := l_weight_sum / l_count
					Result := sqrt (d / (l_weight_sum - l_weight_mean))
				end
			end
		end

	fast_rank_correlation(a_other : ABSTRACT_READABLE_VECTOR) : DOUBLE is
			-- fast computation: cannot have ties
		require
			same_size : count = a_other.count
		local
			i, l_d: INTEGER
			l_sum, l_denom: DOUBLE
			l_ranks, l_other_ranks: INTEGER_VECTOR
		do
			if count <= 1 then
				Result := 1.
			else
				l_ranks := ascending_ranks
				l_other_ranks := a_other.ascending_ranks
				from i := 1 until i > count loop
					l_d := l_ranks[i] - l_other_ranks[i]
					l_sum := l_sum + l_d * l_d
					i := i + 1
				end
				l_denom := count * count - 1
				l_denom := l_denom * count
				Result := 1.0 - 6.0 * l_sum / l_denom
			end
		end

	rank_correlation(a_other : ABSTRACT_READABLE_VECTOR) : DOUBLE is
			-- rank correlation
		require
			same_size : count = a_other.count
		local
			l_ranks, l_other_ranks : INTEGER_VECTOR
			l_vec1, l_vec2 : VECTOR
		do
			if count <= 1 then
				Result := 1.
			else
				l_ranks := ascending_ranks
				l_other_ranks := a_other.ascending_ranks

				l_vec1 := l_ranks.to_vector
				l_vec2 := l_other_ranks.to_vector
				Result := l_vec1.correlation (l_vec2)
			end
		end

	cum_return: DOUBLE is
			-- Cumulative return
			-- (product of (1+x) for all x) - 1
		local
			i: INTEGER
			l_cumret: DOUBLE
		do
			l_cumret := 1.0
			from
				i := 1
			until
				i > count
			loop
				if is_defined_double (item(i)) then
					l_cumret := l_cumret * (1.0 + item (i))
				end
				i := i + 1
			end
			Result := l_cumret - 1.0
		ensure
			cum_return_defined: is_defined_double (Result)
		end

	cum_returns: VECTOR is
			-- Cumulative returns;
			-- Result [i] = subset (1,i).cum_return (not true in the face of undefined items)
		local
			i: INTEGER
			l_cumret: DOUBLE
		do
			create Result.make (count)
			l_cumret := 1.0
			from
				i := 1
			until
				i > count
			loop
				l_cumret := l_cumret * (1.0 + item (i))
				Result.put (l_cumret - 1.0, i)
				i := i + 1
			end
		ensure
			cum_returns_not_void: Result /= Void
			same_count: Result.count = count
			last_item_if_all_defined: is_all_defined implies Result [count] = cum_return
			all_but_last_item_if_all_defined: (is_all_defined and count > 1) implies
				Result.subset ((1 |..| (count - 1)).as_array).same_vector (
					   (subset ((1 |..| (count - 1)).as_array).cum_returns))
		end

	cumulative_sum: VECTOR
			-- Cumulative sum of entries in `Current'
			-- Result [i] = subset (1, i).sum (not true in the face of undefined items)
		local
			i: INTEGER
			l_cum_sum: DOUBLE
		do
			create Result.make (count)
			l_cum_sum := 0.0
			from
				i := 1
			until
				i > count
			loop
				l_cum_sum := l_cum_sum + item (i)
				Result [i] := l_cum_sum
				i := i + 1
			end
		ensure
			cumulative_sum_not_void: Result /= Void
			same_count: Result.count = count
			last_item_if_all_defined: is_all_defined implies Result [count] = sum
			all_but_last_item_if_all_defined: (is_all_defined and count > 1) implies
				Result.subset ((1 |..| (count - 1)).as_array).same_vector (
					(subset ((1 |..| (count - 1)).as_array).cumulative_sum))
		end

	trailing_returns (a_window: INTEGER): VECTOR is
			-- `a_window'-period returns for last `a_window' entries of `Current'
		require
			valid_window: a_window > 0 and a_window <= count
		local
			i: INTEGER
			l_value: DOUBLE
			l_cum_returns: VECTOR
		do
			create Result.make (count - a_window + 1)
			l_cum_returns := cum_returns
			l_cum_returns.add_scalar (1.0)
			from
				i := a_window
			until
				i > count
			loop
				if i = a_window then
					l_value := l_cum_returns [i] - 1.0
				else
					l_value := (l_cum_returns [i] / l_cum_returns [i - a_window]) - 1.0
				end
				Result [i - a_window + 1] := l_value
				i := i + 1
			end
		ensure
			trailing_returns_not_void: Result /= Void
			correct_count: Result.count = (count - a_window + 1)
			first_value: is_all_defined implies Result [1] = (cum_returns [a_window])
			other_values_might_be_NaN: True -- (even when `is_all_defined', as there might be a division by zero)
		end

	trailing_returns_with_partial_values (a_window: INTEGER): VECTOR is
			-- `a_window'-period returns for last `a_window' entries of `Current';
			-- Single-period returns for earlier entries.
		require
			valid_window : a_window > 0 and a_window <= count
		local
			i: INTEGER
			l_value: DOUBLE
			l_cum_returns: VECTOR
		do
			create Result.make (count)
			l_cum_returns := cum_returns
			l_cum_returns.add_scalar (1.0)
			from
				i := 1
			until
				i > count
			loop
				if i <= a_window then
					l_value := (l_cum_returns [i] / l_cum_returns [1]) - 1.0
				else
					l_value := (l_cum_returns [i] / l_cum_returns [i - a_window]) - 1.0
				end
				Result [i] := l_value
				i := i + 1
			end
		ensure
			trailing_returns_with_partial_values_not_void: Result /= Void
			same_count: Result.count = count
			first_value: is_all_defined implies Result [1] = 0.0
		end

	correlation (a_other: like Current): DOUBLE is
			-- Correlation between two series
		require
			a_other_not_void: a_other /= Void
			equal_size: a_other.count = count
		local
			l_matrix: MATRIX
		do
			create l_matrix.make (count, 2)
			l_matrix.put_column_vector (Current, 1)
			l_matrix.put_column_vector (a_other, 2)
			Result := l_matrix.correlation_matrix [1, 2]
		end

	element_by_element_max_ratio (a_other: ABSTRACT_READABLE_VECTOR): VECTOR is
			-- For each element pair, get max item divided by min item
		require
			a_other_not_void: a_other /= Void
			counts_match : a_other.count = count
		local
			i: INTEGER
			l_numerator: DOUBLE
			l_denominator: DOUBLE
			l_ratio: DOUBLE
		do
			create Result.make (count)
			from
				i := 1
			until
				i > count
			loop
				if item (i) > a_other.item (i) then
					l_numerator := item (i)
					l_denominator := a_other.item (i)
				else
					l_numerator := a_other.item (i)
					l_denominator := item (i)
				end
				if l_denominator /= 0.0 then
					l_ratio := l_numerator / l_denominator
				else
					l_ratio := undefined_double
				end
				Result.put (l_ratio, i)
				i := i + 1
			end
		ensure
			element_by_element_max_ratio_not_void: Result /= Void
			same_count: Result.count = count
		end

	counts_by_range_left_inclusive (a_ranges: ARRAY [DOUBLE]): ROSE_LINEAR_ARRAY [TUPLE [DOUBLE, DOUBLE, INTEGER]] is
			-- Value counts in each bucket defined by `a_ranges';
			-- e.g., the frequences in a histogram, defined by [a, b),
			-- `a_ranges', in ascending order, defines the left and right boundaries of a range/interval
		require
			a_ranges_not_void: a_ranges /= Void
		local
			i, j: INTEGER
			l_counts: ARRAY [INTEGER]
			l_range_count: INTEGER
			l_item_was_processed: BOOLEAN
		do
			l_range_count := a_ranges.count + 1
			create l_counts.make (1, l_range_count)
			from
				i := 1
			until
				i > count
			loop
				l_item_was_processed := False
				from
					j := 1
				until
					j > a_ranges.count or l_item_was_processed = True
				loop
					if item (i) < a_ranges.item (j) then
						l_counts.force (l_counts.item (j) + 1, j)
						l_item_was_processed := True
					elseif item (i) >= a_ranges.item (a_ranges.count) then
						l_counts.force (l_counts.item (l_range_count) + 1, l_range_count)
						l_item_was_processed := True
					else
						-- Do nothing
					end
					j := j + 1
				end
				i := i + 1
			end
			Result := counts_merged_w_ranges (a_ranges, l_counts)
		ensure
			not_void: Result /= Void
			count_one_more_than_a_ranges: Result.count = a_ranges.count + 1
		end

	counts_by_range_right_inclusive (a_ranges: ARRAY [DOUBLE]): ROSE_LINEAR_ARRAY [TUPLE [DOUBLE, DOUBLE, INTEGER]] is
			-- Value counts in each bucket defined by `a_ranges';
			-- e.g., the frequences in a histogram, defined by (a, b]
			-- `a_ranges', in ascending order, defines the left and right boundaries of a range/interval
		require
			a_ranges_not_void: a_ranges /= Void
		local
			i, j: INTEGER
			l_counts: ARRAY [INTEGER]
			l_range_count: INTEGER
			l_item_was_processed: BOOLEAN
		do
			l_range_count := a_ranges.count + 1
			create l_counts.make (1, l_range_count)
			from
				i := 1
			until
				i > count
			loop
				l_item_was_processed := False
				from
					j := 1
				until
					j > a_ranges.count or l_item_was_processed = True
				loop
					if item (i) <= a_ranges.item (j) then
						l_counts.force (l_counts.item (j) + 1, j)
						l_item_was_processed := True
					elseif item (i) > a_ranges.item (a_ranges.count) then
						l_counts.force (l_counts.item (l_range_count) + 1, l_range_count)
						l_item_was_processed := True
					else
						-- Do nothing
					end
					j := j + 1
				end
				i := i + 1
			end
			Result := counts_merged_w_ranges (a_ranges, l_counts)
		ensure
			not_void: Result /= Void
			count_one_more_than_a_ranges: Result.count = a_ranges.count + 1
		end

feature {NONE} -- Implementation

	counts_merged_w_ranges (a_ranges: ARRAY [DOUBLE]; a_counts: ARRAY [INTEGER]): ROSE_LINEAR_ARRAY [TUPLE [DOUBLE, DOUBLE, INTEGER]] is
			-- Implementation function for counts_by_range_*, creating ranges from boundaries and adding the resulting count from `a_counts'
		require
			a_ranges_not_void: a_ranges /= Void
			a_counts_not_void: a_counts /= Void
			a_counts_one_longer: a_counts.count = (a_ranges.count + 1)
		local
			i, l_range_count: INTEGER
		do
			create Result.make_empty
			l_range_count := a_ranges.count + 1
			from
				i := 1
			until
				i > l_range_count
			loop
				if i = 1 then
					Result.extend ([negative_infinity, a_ranges.item (i), a_counts.item (i)])
				elseif i = l_range_count then
					Result.extend ([a_ranges.item (i - 1), positive_infinity, a_counts.item (i)])
				else
					Result.extend ([a_ranges.item (i - 1), a_ranges.item (i), a_counts.item (i)])
				end
				i := i + 1
			end
		ensure
			not_void: Result /= Void
			count_one_more_than_a_ranges: Result.count = a_ranges.count + 1
		end

feature -- Time series analysis

	acf (a_lag: INTEGER): VECTOR is
			-- Auto-correlations at `a_lag'
		require
			valid_lag: a_lag > 0 and a_lag < count
		local
			i: INTEGER
			l_vec: VECTOR
			l_mean: DOUBLE
			l_gamma_0: DOUBLE
		do

			-- Very simple but robust algorithm confirmed with SPLUS;
			-- slight different from NAG (AUTO_CORRELATION)

			create Result.make (a_lag)
			l_mean := weighted_mean (Void)
			l_vec := minus (l_mean)
			l_gamma_0 := l_vec * l_vec / count
			from i := 1 until i > a_lag loop
				Result [i] := (l_vec.sub_vector (1, count - i) * l_vec.sub_vector (i + 1, count)) / count
				i := i + 1
			end
			Result := Result.scaled (1. / l_gamma_0)
		ensure
			result_not_void: Result /= Void
			correct_count: Result.count = a_lag
		end

	pacf (a_lag: INTEGER): VECTOR is
			-- Partial auto-correlations at `a_lag';
			-- This routine is sensitive to rounding error and should not be used
			--  if the parameters are near the nonstationarity boundary. (TODO: expressable as pre-condition?)
			-- May use `pacf_ols' instead.
		require
			valid_lag: a_lag > 0 and a_lag < count
		local
			i, j: INTEGER
			l_den, l_nom, l_phi: DOUBLE
			l_acf: VECTOR
			l_phi_matrix: MATRIX
		do

			-- Using Yule-Walker equations and Durbin recursion;
			-- confirmed with SPLUS; slight different from NAG (AUTO_CORRELATION)

			l_acf := acf (a_lag)
			create l_phi_matrix.make (a_lag, a_lag)

			from i := 1 until i > a_lag loop
				if i = 1 then
					l_phi_matrix [1, 1] := l_acf [1]
				else
					l_nom := l_acf [i]
					l_den := 1.
					from j := 1 until j > i - 1 loop
						l_nom := l_nom - l_phi_matrix [i-1, j] * l_acf [i - j]
						l_den := l_den - l_phi_matrix [i-1, j] * l_acf [j]
						j := j + 1
					end
					l_phi := l_nom / l_den
					l_phi_matrix [i, i] := l_phi
					from j := 1 until j > i -1 loop
						l_phi := l_phi_matrix [i-1, j] - l_phi_matrix [i, i] * l_phi_matrix [i - 1, i - j]
						l_phi_matrix [i, j] := l_phi
						j := j + 1
					end
				end
				i := i + 1
			end

			Result := l_phi_matrix.diagonal
		ensure
			result_not_void: Result /= Void
			correct_count: Result.count = a_lag
		end


	pacf_ols (a_lag: INTEGER): VECTOR is
			-- Partial auto-correlations at `a_lag' estimated via ordinary least squares
		require
			valid_lag: a_lag > 0 and a_lag < count
		local
			i: INTEGER
			l_vec: VECTOR
		do
			create Result.make (a_lag)
			from i := 1 until i > a_lag loop
				l_vec := ar_regression (i)
				Result [i] := l_vec.last -- TODO: check that this is ok - requires not_empty
				i := i + 1
			end
		ensure
			result_not_void: Result /= Void
			correct_count: Result.count = a_lag
		end

	ar_regression (a_lag: INTEGER): VECTOR is
			-- Auto-regressive regression
		require
			valid_lag: a_lag > 0 and a_lag < count
		local
			l_vec: VECTOR
			i: INTEGER
			l_y, l_xty: VECTOR
			l_x: MATRIX
			l_xtx, l_xtx_inverse: SYMMETRIC_MATRIX
		do
			l_vec := minus (weighted_mean (Void))

			create l_x.make (count - a_lag, a_lag)
			l_y := l_vec.sub_vector (a_lag + 1, count)

			from i := 1 until i > a_lag loop
				l_x.put_column_sub_vector (l_vec, i, i + 1)
				i := i + 1
			end

			l_xty := l_x.transpose_at (l_y)
			create l_xtx.make_xtx (l_x)
			l_xtx_inverse := l_xtx.pseudo_inverse

			Result := l_xtx_inverse @ l_xty
		ensure
			result_not_void: Result /= Void
			correct_count: True -- TODO: Result.count = ?
		end

	growth_rate (a_minimum_number_of_observation: INTEGER; a_is_percentage: BOOLEAN): VECTOR is
			-- Growth rate
		require
			at_least_three_observations: a_minimum_number_of_observation > 2
		local
			l_grate, l_ave, l_sigma, l_tstat, l_cv, l_cv2: DOUBLE
			i, l_obs: INTEGER
			l_factor, l_av, l_avabs, l_sumt,l_tbar: DOUBLE
			l_xx,l_yy,l_xy,l_trend,l_exss,l_rss,l_rms, l_atstat: DOUBLE
			l_tvar: VECTOR
		do
			-- Implemented based on subroutine growth_rate() in monthlydetail_descriptor_subs.for by JY
			if count >= a_minimum_number_of_observation then
				--- set weight
				if a_is_percentage then
					l_factor := 100.0
				else
					l_factor := 1.0
				end
				create l_tvar.make_sequence (count, 1, -1)
				from i := 1	until i > count	loop
					if is_defined_double (Current [i]) then
						l_obs := l_obs + 1
						l_av := l_av + Current [i]
						l_sumt := l_sumt + l_tvar [i]
						l_avabs := l_avabs + (Current [i]).abs
					end
					i := i + 1
				end
			end
			if (l_obs >= a_minimum_number_of_observation) then
				l_ave := l_av / l_obs
				l_avabs := l_avabs / l_obs
				l_tbar := l_sumt / l_obs
				from i := 1 until i > count loop
					if is_defined_double (Current [i]) then
						l_xx := l_xx + (l_tvar [i] - l_tbar) * (l_tvar [i] - l_tbar)
							l_xy := l_xy + (l_tvar [i] - l_tbar) * (Current [i] - l_ave)
								l_yy := l_yy + (Current [i] - l_ave) * (Current [i] - l_ave)
					end
					i := i + 1
				end
				l_sigma := l_yy / (l_obs - 1.)
				if (l_sigma > 0.) then
					l_sigma := sqrt (l_sigma)
				end
				l_trend := l_xy / l_xx
				if (l_ave > (4.0 / 3.0) * l_trend.abs) then
					l_grate := l_factor * l_trend / l_ave
				elseif (l_avabs > 0.) then
					l_grate := l_factor * l_trend / l_avabs
				else
					l_grate := 0.
				end
				l_exss := l_trend * l_trend * l_xx

                if not ensure_clean_zero_in_growth_rate_calculation then
                	l_rss := l_yy - l_exss
                else
                	-- Clean 0 is needed for example by EG_T, 143u.
				    if l_yy >= 10^-6 then -- no need to take the absolute value of l_yy because it's > 0 always.
    			    	l_rss := l_yy - l_exss
					    if l_rss.abs <= 10^-11 then -- fast test to see if l_rss should be forced to a clean 0.
							l_rss := 0.0
					    end
				   	else
					    if (l_yy - l_exss).abs <= (10^-3) * ((1.0 / 2) * (l_yy + l_exss)).max (1.0) then -- more expensive test than the one above. l_yy and l_exss are both > 0.
 					    	l_rss := 0.0
					    else
							l_rss := l_yy - l_exss
					    end
				    end
                end

				l_rms := l_rss / ( l_obs - 2.)
				if (l_avabs > 0.) then
					l_cv := l_sigma / l_avabs
					if l_obs >= 5 and l_rms > 0. then
						l_cv2 := sqrt (l_rms) / l_avabs -- use this with 5 observations or longer
					else
						l_cv2 := l_cv
					end
				else
					l_cv := 0.
					l_cv2 := 0.
				end
				if (l_rms <= 0.) then
					l_atstat := 3.0
				else
					l_atstat := (l_exss / l_rms) ^ (0.5)
				end
				if not treat_zero_as_having_sign_of_positive_one then
					l_tstat := l_atstat * l_trend.sign -- l_trend = 0 => l_trend.sign = 0, but in FORTRAN, l_trend = 0 => sign (l_trend) = 1. See line containing text TSTAT=SIGN(ATSTAT,TREND) in file b_mondat_src:MONTHLYDETAIL_DESCRIPTOR_SUBS.FOR
				else
					if l_trend = 0 then
						l_tstat	:= l_atstat.abs
					else
						l_tstat := l_atstat * l_trend.sign
					end
				end
				create Result.make_from_array (<<l_grate, l_ave, l_sigma, l_tstat, l_cv, l_cv2>>)
			else
				create Result.make_from_array (<<constants.missing_double, constants.missing_double,
				constants.missing_double, constants.missing_double,
				constants.missing_double, constants.missing_double>>)
			end
			Result.set_names (k_growth_names)
		ensure
			result_not_void: Result /= Void
			correct_count: Result.count = 6
		end

   	ensure_clean_zero_in_growth_rate_calculation: BOOLEAN is True
   			-- Should `small error' monitoring be used in residual calculations in `growth_rate'?

   	treat_zero_as_having_sign_of_positive_one: BOOLEAN is True
   			-- Should zero entry be treated as having a positive sign?

feature -- Output

	display is
			-- Print each element on a separate line to standard output.
		local
			i: INTEGER
		do
			from i := 1 until i > count loop
				print((Current [i]).out + "%N")
				i := i + 1
			end
		end

	comma_separated_display_string: STRING is
			-- Comma separated string of elements
		local
			i: INTEGER
		do
			Result := ""
			from i := 1 until i > count loop
				Result.append ((Current [i]).out)
				if i /= count then
					Result.append (",")
				end
				i := i + 1
			end
		ensure
			valid_result: Result /= Void
		end

	as_formatted_string (a_formatter: FORMAT_DOUBLE; a_separator: STRING ) : STRING
			-- Current as a string with items formatted using `a_formatter' separated by `a_separator'
		require
			formatter_not_void: a_formatter /= Void
			separator_not_void_nor_empty: a_separator /= Void and then not a_separator.is_empty
		do
			create Result.make (100) -- just a wild guess, it could be any reasonable value for string capacity
			do_all (agent append_formatted_separated_item (Result, ?, a_formatter, a_separator))
		ensure
			result_not_void: Result /= Void -- but it could be empty if Current is empty
		end

feature {NONE} -- Output

	append_formatted_separated_item (a_cumulator: STRING; a_item: DOUBLE; a_formatter: FORMAT_DOUBLE; a_separator: STRING) is
			-- append formatted item to `a_cumulator' prepended by `a_separator' if `a_cumulator' is not empty
		require
			cumulator_not_void: a_cumulator /= Void
			formatter_not_void: a_formatter /= Void
			separator_not_void: a_separator /= Void -- the agent doesn't care if it's empty, but an stupid output will be produced
		do
			if not a_cumulator.is_empty then
				a_cumulator.append (a_separator)
			end
			a_cumulator.append (a_formatter.formatted (a_item))
		ensure
			cumulator_longer: a_cumulator.count > old a_cumulator.count
		end

feature -- Obsolete


	cum_vector: VECTOR is
			-- Duplicate of `cumulative_sum'
		obsolete
		   "Use `cumulative_sum' instead"
		do
			Result := cumulative_sum
		end

	postive_sum: DOUBLE is
			-- Sum of all defined positive elements
		obsolete
		   "Use `positive_sum' instead"
		do
			Result := positive_sum
		ensure
			non_negative_result: Result >= 0
		end

	postive_vector: VECTOR is
			-- Copy of `Current' where all elements are defined and >= 0
		obsolete
		   "Use `positive_vector' instead"
		do
			Result := positive_vector
		ensure
			result_not_void: Result /= Void
			new_object: as_readonly_vector /= Result
			same_count: Result.count = count
			definition: True
		end

	condition_count (a_test: FUNCTION [ANY, TUPLE [DOUBLE], BOOLEAN]): INTEGER is
			-- Number of elements that pass `a_test'
		obsolete
		   "Use `matching_item_count' instead"
		require
			a_test_not_void: a_test /= Void
		do
			Result := matching_item_count (a_test)
		ensure
			non_negative_result: Result >= 0
			maximum_result_is_count: Result <= count
		end

	valid_count: INTEGER is
			-- Number of elements with a normal value;
			-- Abnormal values are NaNs, Infinities and legacy undefined doubles (-999.)
		obsolete
		   "Use `defined_item_count' instead"
		do
			Result := defined_item_count
		ensure
			non_negative_result: Result >= 0
			maximum_result_is_count: Result <= count
		end

	condition_flags (a_test: FUNCTION [ANY, TUPLE [DOUBLE], BOOLEAN]): ROSE_BOOLEAN_ARRAY is
			-- Test results for elements against `a_test'
		obsolete
		   "Use `matching_item_flags' instead"
		require
			a_test_not_void: a_test /= Void
		do
			Result := matching_item_flags (a_test)
		ensure
			result_not_void: Result /= Void
			same_count: Result.count = count
		end

	valid_flags: ROSE_LINEAR_ARRAY [BOOLEAN] is
			-- Flags show which elements have normal values
			-- Abnormal values are NaNs, Infinities and legacy undefined doubles (-999.)
		obsolete
		   "Use `defined_item_flags' instead"
		do
			Result := defined_item_flags
		ensure
			result_not_void: Result /= Void
			same_count: Result.count = count
			defined_values: True
		end

	condition_indices (a_test: FUNCTION [ANY, TUPLE [DOUBLE], BOOLEAN]): ROSE_LINEAR_ARRAY [INTEGER] is
			-- Indices of elements that pass `a_test'
		obsolete
		   "Use `matching_item_indices' instead"
		require
			a_test_not_void: a_test /= Void
		do
			Result := matching_item_indices (a_test)
		ensure
			result_not_void: Result /= Void
			maximum_count: Result.count <= count
			all_indices_valid: Result.for_all (agent valid_index)
		end

	condition_values (a_test: FUNCTION [ANY, TUPLE [DOUBLE], BOOLEAN]): VECTOR is
			-- Elements that pass `a_test'
		obsolete
		   "Use `matching_items' instead"
		require
			a_test_not_void: a_test /= Void
		do
			Result := matching_items (a_test)
		ensure
			result_not_void: Result /= Void
			maximum_count: Result.count <= count
			all_values_pass_test: True
		end

	subset_by_flags (a_flags: ARRAY [BOOLEAN]): VECTOR is
			-- Sub-vector of `Current' where elements are included if corresponding element in `a_flag' is `True'
		obsolete
		   "Use `sub_vector_by_flags' instead"
		require
			a_flags_not_void: a_flags /= Void
			not_too_many_flags: a_flags.count <= count
		do
			Result := sub_vector_by_flags (a_flags)
		ensure
			result_not_void: Result /= Void
			new_object: as_readonly_vector /= Result
		end

	order: ROSE_LINEAR_ARRAY [INTEGER] is
			-- Indices of elements of `Current' to specify an ascending order
		obsolete
		   "Use `ascending_order' instead"
		do
			Result := ascending_order
		ensure
			result_not_void: Result /= Void
			same_count: Result.count = count
			ascending_order_specification: True
		end

			order_desc: ROSE_LINEAR_ARRAY [INTEGER] is
				-- Indices of elements of `Current' to specify a descending order
		obsolete
		   "Use `descending_order' instead"
		do
			Result := descending_order
		ensure
			result_not_void: Result /= Void
			same_count: Result.count = count
			descending_order_specification: True
		end

	square_root: VECTOR is
			-- Square roots of elements of `Current'
		obsolete
		   "Use `square_roots' instaed"
		require
			all_elements_non_negative: non_negative
		do
			Result := square_roots
		ensure
			result_not_void: Result /= Void
			same_count: Result.count = count
		end

	square: VECTOR is
			-- Squares of elements of `Current'
		obsolete
		   "Use `squares' instead"
		do
			Result := squares
		ensure
			result_not_void: Result /= Void
			same_count: Result.count = count
		end

	normalized_square_root: VECTOR is
			-- Square roots of elements of `Current' normalized so that ??? TODO
		obsolete
		   "Use `normalized_square_roots' instead"
		require
			non_negative: non_negative
		do
			Result := normalized_square_roots
		ensure
			result_not_void: Result /= Void
			same_count: Result.count = count
		end

feature -- Contract support

	is_correct_defined_status (a_status: BOOLEAN; a_index: INTEGER): BOOLEAN is
			-- Is `a_status' the correct defined-ness of element at `a_index'?
		require
			valid_index: valid_index (a_index)
		do
			Result := a_status = is_defined_double (Current [a_index])
		end

	is_non_zero (a_double: DOUBLE): BOOLEAN is
			-- Is `a_double' non-zero?
		do
			Result := not is_missing(a_double) and then a_double /= 0.0
		ensure
			definition: Result = (not is_missing(a_double) and then a_double /= 0.0)
		end

	are_elements_ascending (i, j: INTEGER): BOOLEAN is
			-- Is element at `i' less than element at `j'?
		do
			Result := less_than (Current [i], Current [j])
		ensure
			definition: Result implies less_than (Current [i], Current [j])
		end

	are_elements_descending (i, j: INTEGER): BOOLEAN is
			-- Is element at `j' less than element at `i'?
		do
			Result := less_than (Current [j], Current [i])
		ensure
			definition: Result implies less_than (Current [j], Current [i])
		end

	is_next_element_greater (a_value, a_index: INTEGER; a_order: ROSE_LINEAR_ARRAY [INTEGER]): BOOLEAN is
			-- Is `Current [a_order [a_index + 1]]' greater than `Current [a_value]'?
			-- `True' if `a_index' = `a_order.count'.
			-- Undefined elements are ignored (assumed `True').
		require
			a_order_not_void: a_order /= Void
			a_index_valid: a_order.valid_index (a_index)
			all_indices_valid: a_order.for_all (agent valid_index)
		do
			if not a_order.valid_index (a_index + 1) then
				Result := True
			elseif is_defined_double (Current [a_value]) and is_defined_double (Current [a_order [a_index + 1]]) then
				Result := not_greater_than (Current [a_value], Current [a_order [a_index + 1]])
			else
				Result := True
			end
		end

	is_next_element_less  (a_value, a_index: INTEGER; a_order: ROSE_LINEAR_ARRAY [INTEGER]): BOOLEAN is
			-- Is `Current [a_order [a_index + 1]]' less than `Current [a_value]'?
			-- `True' if `a_index' = `a_order.count'.
			-- Results on undefined elements are treated as `True'.
		require
			a_order_not_void: a_order /= Void
			a_index_valid: a_order.valid_index (a_index)
			all_indices_valid: a_order.for_all (agent valid_index)
		do
			if not a_order.valid_index (a_index + 1) then
				Result := True
			elseif is_defined_double (Current [a_value]) and is_defined_double (Current [a_order [a_index + 1]]) then
				Result := not_less_than (Current [a_value], Current [a_order [a_index + 1]])
			else
				Result := True
			end
		end

	are_adjacent_ascending_ranks_consistent (a_index: INTEGER; a_ranks: INTEGER_VECTOR): BOOLEAN is
			-- Is `Current [a_index + 1]' in same rank relation to `Current [a_index]'
			--  as `a_ranks [a_index + 1]' is to `a_ranks [a_index]'?
			-- `True' if `a_index' = `count'.
			-- Results on undefined elements are treated as `True'.
		require
			a_index_valid: valid_index (a_index)
			a_ranks_not_void: a_ranks /= Void
			a_ranks_correct_count: a_ranks.count = count
					-- N.B. we assume the missing invariant from INTEGER_VECTOR: `lower' = 1
		do
			if not valid_index (a_index + 1) then
				Result := True
			elseif is_defined_double (Current [a_index]) and is_defined_double (Current [a_index + 1]) then
				Result := (not_greater_than (Current [a_index], Current [a_index + 1]) =
					not_greater_than (a_ranks [a_index], a_ranks [a_index + 1]))
			else
				Result := True
			end
		end

	are_adjacent_descending_ranks_consistent (a_index: INTEGER; a_ranks: INTEGER_VECTOR): BOOLEAN is
			-- Is `Current [a_index + 1]' in same rank relation to `Current [a_index]'
			--  as `a_ranks [a_index + 1]' is to `a_ranks [a_index]'?
			-- `True' if `a_index' = `count'.
			-- Results on undefined elements are treated as `True'.
		require
			a_index_valid: valid_index (a_index)
			a_ranks_not_void: a_ranks /= Void
			a_ranks_correct_count: a_ranks.count = count
					-- N.B. we assume the missing invariant from INTEGER_VECTOR: `lower' = 1
		do
			if not valid_index (a_index + 1) then
				Result := True
			elseif is_defined_double (Current [a_index]) and is_defined_double (Current [a_index + 1]) then
				Result := (not_less_than (Current [a_index], Current [a_index + 1]) =
					not_less_than (a_ranks [a_index + 1], a_ranks [a_index]))
			else
				Result := True
			end
		end

	not_less_than (a_first, a_second: DOUBLE): BOOLEAN is
			-- Is `a_first' >= `a_second'?
		require
			both_arguments_defined: not is_missing (a_first) and then not is_missing (a_second)
		do
			Result := a_first >= a_second
		ensure
			definition: Result = (a_first >= a_second)
		end

	not_greater_than (a_first, a_second: DOUBLE): BOOLEAN is
			-- Is `a_first' <= `a_second'?
		require
			both_arguments_defined: not is_missing (a_first) and then not is_missing (a_second)
		do
			Result := a_first <= a_second
		ensure
			definition: Result = (a_first <= a_second)
		end

	equal_strings (a_names: ARRAY [STRING]; a_name: STRING; a_index: INTEGER): BOOLEAN is
			-- Is `a_name' `equal' to `a_names [a_index]'?
		require
			a_names_not_void: a_names /= Void
			a_index_valid: a_names.valid_index (a_index + a_names.lower - 1)
		do
			Result := equal (a_name, a_names [a_index + a_names.lower - 1])
		ensure
			definition: Result = equal (a_name, a_names [a_index + a_names.lower - 1])
		end

feature {NONE} -- Implementation

	name_table_saved: like name_table
			-- Cached result of `name_table'

	bounded_element (a_value: DOUBLE; a_index: INTEGER; a_lower, a_upper: ABSTRACT_READABLE_VECTOR): DOUBLE is
			-- Restriction of `a_value' to interval (`a_lower [a_index]', `a_upper [a_index]')
		require
			positive_index: a_index > 0
			a_index_small_enough: a_index <= count
			a_lower_not_void: a_lower /= Void
			a_lower_large_enough: a_lower.count >= a_index
			a_upper_not_void: a_upper /= Void
			a_upper_large_enough: a_upper.count >= a_index
		do
			Result := a_value.max (a_lower [a_index]).min (a_upper [a_index])
		ensure
			definition: Result = a_value.max (a_lower [a_index]).min (a_upper [a_index])
		end

feature -- Indexing

	k_growth_names: ROSE_LINEAR_ARRAY [STRING] is
			-- Names of growth rate variables
		once
			create Result.make_from_array (<<"grate", "ave", "sigma", "tstat", "cv", "cv2">>)
		ensure
			result_not_void: Result /= Void
			correct_count: Result.count = 6
		end

	k_growth_grate_index: INTEGER is 1
			-- Index for 'grate'
	k_growth_ave_index: INTEGER is 2
			-- Index for 'ave'
	k_growth_sigma_index: INTEGER is 3
			-- Index for 'sigma'
	k_growth_tstat_index: INTEGER is 4
			-- Index for 'tstat'
	k_growth_cv_index: INTEGER is 5
			-- Index for 'cv'
	k_growth_cv2_index: INTEGER is 6
			-- Index for 'cv2'

feature -- Query

	is_defined (a_value: DOUBLE): BOOLEAN is
			-- Is `a_value' defined?
			-- Descendants are expected to redefine this routine.
		do
			Result := is_defined_double (a_value)
		end

	less_than (a_first,a_second : DOUBLE) : BOOLEAN
			--extended_less from constants
		do
			Result := extended_less(a_first,a_second)
		end

	is_same_positive (a_original: DOUBLE; a_index: INTEGER; a_other: VECTOR): BOOLEAN is
			-- Is `a_other' [`a_index'] equal to `a_original' or zero when latter is negative or undefined?
		require
			valid_index: valid_index (a_index)
			other_vector_not_void: a_other /= Void
		do
			if is_defined_double (a_original) and then a_original > 0 then
				Result := a_other [a_index] = a_original
			else
				Result := a_other [a_index] = 0
			end
		end

	is_same_negative (a_original: DOUBLE; a_index: INTEGER; a_other: VECTOR): BOOLEAN is
			-- Is `a_other' [`a_index'] equal to `a_original' or zero when latter is positive or undefined?
		require
			valid_index: valid_index (a_index)
			other_vector_not_void: a_other /= Void
		do
			if is_defined_double (a_original) and then a_original < 0 then
				Result := a_other [a_index] = a_original
			else
				Result := a_other [a_index] = 0
			end
		end

	equal_at (a_number: DOUBLE; a_index, a_offset: INTEGER): BOOLEAN is
			-- Is `a_number' equal to Current [`a_index' + `a_offset']?
		require
			valid_index: valid_index (a_index + a_offset)
		do
			Result := (a_number = a_number) implies (a_number = Current [a_index + a_offset])
		end

	is_correct_sum (a_sum: DOUBLE; a_index: INTEGER; a_other: ABSTRACT_READABLE_VECTOR): BOOLEAN is
			-- Is `a_sum' sum of elements in `Current' and `a_other' at `a_index'?
		require
			valid_index: valid_index (a_index)
			other_vector_not_void: a_other /= Void
		do
			Result := ieee_is_eq(a_sum, Current [a_index] + a_other [a_index])
		end

	is_correct_difference (a_difference: DOUBLE; a_index: INTEGER; a_other: ABSTRACT_READABLE_VECTOR): BOOLEAN is
			-- Is `a_difference' differnece of elements in `Current' from `a_other' at `a_index'?
		require
			valid_index: valid_index (a_index)
			other_vector_not_void: a_other /= Void
		do
			Result := ieee_is_eq(a_difference, Current [a_index] - a_other [a_index])
		end

	is_correct_product (a_product: DOUBLE; a_index: INTEGER; a_other: ABSTRACT_READABLE_VECTOR): BOOLEAN is
			-- Is `a_product' product of elements in `Current' with `a_other' at `a_index'?
		require
			valid_index: valid_index (a_index)
			other_vector_not_void: a_other /= Void
		do
			Result := ieee_is_eq(a_product, Current [a_index] * a_other [a_index])
		end

	doubles_sum (a_first, a_second: DOUBLE): DOUBLE is
			-- Sum of `a_first' and `a_second'
		do
			Result := a_first + a_second
		end

	doubles_product (a_first, a_second: DOUBLE): DOUBLE is
			-- Product of `a_first' and `a_second'
		do
			Result := a_first * a_second
		end

	is_reduced_by (a_number: DOUBLE; a_index: INTEGER; a_margin: DOUBLE): BOOLEAN is
			-- Is `a_number' = `Current [a_index]' - `a_margin'?
		require
			valid_index: valid_index (a_index)
		do
			Result := approx_equal ((Current [a_index] - a_margin), a_number)
		end

	is_square_root_of (a_number: DOUBLE; a_index: INTEGER): BOOLEAN is
			-- Is `a_number' square root of `Current [a_index]'?
		require
			valid_index: valid_index (a_index)
		do
			Result := approx_equal (sqrt (Current [a_index]), a_number)
		end

	is_square_of (a_number: DOUBLE; a_index: INTEGER): BOOLEAN is
			-- Is `a_number' square of `Current [a_index]'?
		require
			valid_index: valid_index (a_index)
		do
			Result := approx_equal (Current [a_index] * Current [a_index], a_number)
		end

	is_abs_of (a_number: DOUBLE; a_index: INTEGER): BOOLEAN is
			-- Is `a_number' absolute value of `Current [a_index]'?
		require
			valid_index: valid_index (a_index)
		do
			Result := approx_equal (dabs (Current [a_index]), a_number)
		end

	is_non_negative (a_number: DOUBLE): BOOLEAN is
			-- Is `a_number' non_negative?
		do
			Result := a_number = 0 or else a_number > 0  -- <= does not handle unavailables right
		end

	is_scaled_by (a_result, a_weight: DOUBLE; a_index: INTEGER): BOOLEAN is
			-- Is `a_result' = `Current [a_index]' * `a_weight'?
		require
			valid_index: valid_index (a_index)
		do
			if is_defined_double ((Current [a_index]) * a_weight) then
				Result := approx_equal (a_result, (Current [a_index]) * a_weight)
			else
				Result := not is_defined_double (a_result)
			end
		end

	is_transformed_by (a_result: DOUBLE; a_function: FUNCTION [ANY, TUPLE [DOUBLE], DOUBLE]; a_index: INTEGER): BOOLEAN is
			-- Is `a_result' = `Current [a_index]' transformed by `a_function'?
		require
			valid_index: valid_index (a_index)
		do
			Result := approx_equal (a_result, a_function.item ([Current [a_index]]))
		end

	rose_constants: ROSE_CONSTANT_VALUES is
			-- Symbolic names for numbers
		once
			create Result
		ensure
			result_not_void: Result /= Void
		end

	check_ratio : CHECK_RATIO is
			-- TODO
		once
			create Result.make (1)
		ensure
			result_not_void: Result /= Void
		end

	same_double_as_other (a_value: DOUBLE; a_index: INTEGER; a_other: ABSTRACT_READABLE_VECTOR): BOOLEAN is
			-- Is `a_value' same as `a_other [a_index]', treating NaNs as equal?
		require
			a_other_not_void: a_other /= Void
			a_index_small_enough: a_index <= a_other.count
			a_index_large_enough: a_index >= 1
		do
			if isnan (a_value) then
				Result := isnan (a_other [a_index])
			else
				Result := (a_value = a_other [a_index])
			end
		ensure
			nan_definition: isnan (a_value) implies Result = isnan (a_other [a_index])
			ordinary_definition: not isnan (a_value) implies Result = (a_value = a_other [a_index])
		end

	same_double_as_other_reversed (a_value: DOUBLE; a_index: INTEGER; a_other: ABSTRACT_READABLE_VECTOR): BOOLEAN is
			-- Is `a_value' same as `a_other [count + 1 - a_index]', treating NaNs as equal?
		require
			a_other_not_void: a_other /= Void
			a_index_valid: a_other.valid_index (count + 1 - a_index)
		do
			if isnan (a_value) then
				Result := isnan (a_other [count + 1 - a_index])
			else
				Result := (a_value = a_other [count + 1 - a_index])
			end
		ensure
			nan_definition: isnan (a_value) implies Result = isnan (a_other [count + 1 - a_index])
			ordinary_definition: not isnan (a_value) implies Result = (a_value = a_other [count + 1 - a_index])
		end

feature {NONE} -- Correction

	correct_mismatch is
			-- Attempt to correct object mismatch using `mismatch_information'.
		local
			l_mismatch: MISMATCH_INFORMATION
		do
			l_mismatch := mismatch_information
		end

invariant

	non_negative_count: count >= 0
	consistent_count: count = upper

end
