note

	description:
	"[
		Object representing the notion of a source for exchange rates for AXA Rosenberg.
		This takes an integrated view of IDC, EXTEL, and MSCI exchange rates depending on historical data quality and availability.
		Country specific EURO exchange rates are also handled within this class.
	]"
	author: "Mark Howard"
	last_editor: "$Author: rj $"
	date: "$Date: 2014/12/12 $"
	revision: "$Revision: #25 $"
	path: "$File: //rose/source/kernel/market/axa_rosenberg_default_market.e $"
	storable_version: ""

class AXA_ROSENBERG_DEFAULT_MARKET

inherit

	RP_SERIALIZABLE

	MARKET

	ROSE_LEGACY_UNDEFINED
		export {NONE} all	end

	ROSE_AWARE

	ROSE_CATEGORY_IDS
		export {NONE} all	end

feature -- Access

	exchange_rate_asof (a_currency_1, a_currency_2: CURRENCY; a_date: ROSE_DATE; a_asof: detachable ROSE_DATE): DOUBLE
			-- Integrated AXA Rosenberg exchange rate from `a_currency_2' to `a_currency_1' at `a_date' as of `a_asof';
			-- If `a_asof' is `Void' then our latest data is used.
		local
			l_double_1, l_double_2, l_gbp_double: DOUBLE
			l_curr_1_in_gbp, l_curr_2_in_gbp: DOUBLE
			l_date_iter: ROSE_DATE
			l_currency_1, l_currency_2: CURRENCY
			l_loop_count: INTEGER
		do
			if a_currency_1 = a_currency_2 or else a_currency_1.same_currency (a_currency_2) then
				Result := 1
			-- Euro conversions with pre-euro currencies ignore `a_asof':
			elseif (a_date >= first_euro_12 and then ((a_currency_1.same_currency (eur) and eur_currencies_19981231.has (a_currency_2)) or (a_currency_2.same_currency (eur) and eur_currencies_19981231.has (a_currency_1)))) then
				-- Euro created on first_euro_12 date and we are looking at a euro conversion for one of these currencies
				if a_currency_1.same_currency (eur) then
					Result := 1 / eur_currencies_19981231.item (a_currency_2)
				elseif a_currency_2.same_currency (eur) then
					Result := eur_currencies_19981231.item (a_currency_1)
				end
			elseif (a_date >= first_euro_grd and then ((a_currency_1.same_currency (eur) and eur_currencies_20001231.has (a_currency_2)) or (a_currency_2.same_currency (eur) and eur_currencies_20001231.has (a_currency_1)))) then
				-- Addition of GRD to euro zone
				if a_currency_1.same_currency (eur) then
					Result := 1 / eur_currencies_20001231.item (a_currency_2)
				elseif a_currency_2.same_currency (eur) then
					Result := eur_currencies_20001231.item (a_currency_1)
				end
			elseif (a_date >= first_euro_skk and then ((a_currency_1.same_currency (eur) and eur_currencies_20090101.has (a_currency_2)) or (a_currency_2.same_currency (eur) and eur_currencies_20090101.has (a_currency_1)))) then
				-- Addition of SKK to euro zone
				if a_currency_1.same_currency (eur) then
					Result := 1 / eur_currencies_20090101.item (a_currency_2)
				elseif a_currency_2.same_currency (eur) then
					Result := eur_currencies_20090101.item (a_currency_1)
				end
			else
				Result := Undefined_double
				from
					l_date_iter := a_date.twin
				until
					l_date_iter.is_weekday
				loop
					l_date_iter.add_n_days (-1)
				end
				from
					l_loop_count := max_lookback -- Should this be very large when `a_asof' is not `Void'?
				variant
					l_loop_count
				until
					l_loop_count = 0 or is_defined_double (Result)
				loop
					l_currency_1 := reported_currency (a_currency_1, l_date_iter) -- should these be sensitive to `a_asof'?
					l_currency_2 := reported_currency (a_currency_2, l_date_iter)
					if l_date_iter < extel.first_extel then
						Result := idc.exchange_rate_asof (l_currency_1, l_currency_2, l_date_iter, a_asof)
					else
						Result := extel.exchange_rate_asof (l_currency_1, l_currency_2, l_date_iter, a_asof)
					end
					if not is_defined_double (Result) then
						if l_date_iter < extel.first_extel then
							Result := extel.exchange_rate_asof (l_currency_1, l_currency_2, l_date_iter, a_asof)
						elseif l_date_iter < extel.first_euro then
							l_gbp_double := idc.exchange_rate_asof (gbp, usd, l_date_iter, a_asof)
							if is_defined_double (l_gbp_double) then
								l_curr_1_in_gbp := extel.exchange_rate_asof (l_currency_1, gbp, l_date_iter, a_asof)
								if is_defined_double (l_curr_1_in_gbp) then
									l_double_1 := l_curr_1_in_gbp
								else
									l_double_1 := idc.exchange_rate_asof (l_currency_1, usd, l_date_iter, a_asof)
									if is_defined_double (l_double_1) then
										l_double_1 := l_double_1 / l_gbp_double
									end
								end
								l_curr_2_in_gbp := extel.exchange_rate_asof (l_currency_2, gbp, l_date_iter, a_asof)
								if is_defined_double (l_curr_2_in_gbp) then
									l_double_2 := l_curr_2_in_gbp
								else
									l_double_2 := idc.exchange_rate_asof (l_currency_2, usd, l_date_iter, a_asof)
									if is_defined_double (l_double_2) then
										l_double_2 := l_double_2 / l_gbp_double
									end
								end
								if is_defined_double (l_double_1) and is_defined_double (l_double_2) then
									Result := l_double_1 / l_double_2
								end
							end
						end
					end
					if not is_defined_double (Result) then
						Result := idc.exchange_rate_asof (l_currency_1, l_currency_2, l_date_iter, a_asof)
					end
					if not is_defined_double (Result) then
						if l_date_iter >= extel.first_extel and l_date_iter < extel.first_euro then
							Result := msci.exchange_rate_asof (l_currency_1, l_currency_2, l_date_iter, a_asof)
						end
						if not is_defined_double (Result) then
							l_date_iter.day_back
						end
					end
					l_loop_count := l_loop_count - 1
				end
				if
					not is_defined_double (Result) and a_asof = Void and then
					(a_date.is_end_of_month or else a_date.month_end.previous_weekday <= a_date)
				then
					Result := gv.exchange_rate_asof (l_currency_1, l_currency_2, a_date.month_end, Void)
					if not is_defined_double (Result) then
						Result := undefined_double
					end
				end
			end
		end

feature -- Basic operations

	reset_xrate_cache is
			-- Clear-out `xrate_cache'.
		do
			extel.reset_xrate_cache
			idc.reset_xrate_cache
			msci.reset_xrate_cache
		end

feature {NONE} -- Various markets

	max_lookback: INTEGER = 30
			-- Lookback up to 30 days (~1 month) to find an exchange rate

	idc: IDC_CURRENCY_MARKET
			-- idc market
		once
			create Result
		end

	extel: EXTEL_CURRENCY_MARKET
			-- extel market
		once
			create Result
		end

	msci: MSCI_CURRENCY_MARKET
			-- msci market
		once
			create Result
		end

	gv: GV_CURRENCY_EXCHANGE_MARKET
			-- GV Market
		once
			create Result
		end

feature {NONE} -- Implementation

	usd: CURRENCY
			-- US Dollar
		once
			Result := context.usd
		end

	gbp: CURRENCY
			-- GB Pound
		once
			Result := context.gbp
		end

	jpy: CURRENCY
			-- Japanese Yen
		once
			Result := context.jpy
		end

	eur: CURRENCY
			-- Currency of the European Union
		once
			Result := context.eur
		end

	luf: CURRENCY
			-- Luxembourg Franc
		once
			Result := context.currency_by_iso_code ("LUF", Void)
		end

	lul: CURRENCY
			-- Luxembourg Reporting
		once
			Result := context.currency_by_iso_code ("LUL", Void)
		end

	zal: CURRENCY
			-- South Africa Rand Financial
		once
			Result := context.currency_by_iso_code ("ZAL", Void)
		end

	zar: CURRENCY
			-- South African Rand
		once
			Result := context.currency_by_iso_code ("ZAR", Void)
		end

	bef: CURRENCY
			-- Belgian Franc
		once
			Result := context.currency_by_iso_code ("BEF", Void)
		end

	bec: CURRENCY
			-- Convertible Belgian Franc
		once
			Result := context.currency_by_iso_code ("BEC", Void)
		end

feature {NONE} -- EUR special treatment

	eur_currencies_19981231: HASH_TABLE [DOUBLE, CURRENCY]
			-- eur currencies switched 19981231
		once
			create Result.make (12)
			Result.extend (13.7603, context.currency_by_iso_code ("ATS", Void))
			Result.extend (40.3399, context.currency_by_iso_code ("BEF", Void))
			Result.extend (1.95583, context.currency_by_iso_code ("DEM", Void))
			Result.extend (166.386, context.currency_by_iso_code ("ESP", Void))
			Result.extend (5.94573, context.currency_by_iso_code ("FIM", Void))
			Result.extend (6.55957, context.currency_by_iso_code ("FRF", Void))
			Result.extend (0.787564, context.currency_by_iso_code ("IEP", Void))
			Result.extend (1936.27, context.currency_by_iso_code ("ITL", Void))
			Result.extend (40.3399, context.currency_by_iso_code ("LUF", Void))
			Result.extend (40.3399, context.currency_by_iso_code ("LUL", Void))
			Result.extend (2.20371, context.currency_by_iso_code ("NLG", Void))
			Result.extend (200.482, context.currency_by_iso_code ("PTE", Void))
		ensure
			attached_result: Result /= Void
			non_void_elements: not Result.has (Void)
			non_zero_content: not Result.linear_representation.has (0.0)
		end

	first_euro_12: ROSE_DATE
			-- first eur date for the block of 12
		once
			create Result.make_ccyymmdd (19981231)
		end

	eur_currencies_20001231: HASH_TABLE [DOUBLE, CURRENCY]
			-- eur currencies switched 20001231
		once
			create Result.make (1)
			Result.extend (340.750, context.currency_by_iso_code ("GRD", Void))
		end

	first_euro_grd: ROSE_DATE
			-- first eur date for grd
		once
			create Result.make_ccyymmdd (20001231)
		end

	eur_currencies_20090101: HASH_TABLE [DOUBLE, CURRENCY]
			-- eur currencies switched 20090101
		once
			create Result.make (1)
			Result.extend (30.1260, context.currency_by_iso_code ("SKK", Void))
		end

	first_euro_skk: ROSE_DATE
			-- euro effective date for SKK
		once
			create Result.make_ccyymmdd (20090101)
		end

feature {NONE} -- Reporting Currency Handling

	first_bec: ROSE_DATE
			-- first bec date
		once
			create Result.make_ccyymmdd (19810102)
		end

	last_bec: ROSE_DATE
			-- last bec date
		once
			create Result.make_ccyymmdd (19840131)
		end

	first_zal: ROSE_DATE
			-- first zal date
		once
			create Result.make_ccyymmdd (19850901)
		end

	last_zal: ROSE_DATE
			-- last zal date
		once
			create Result.make_ccyymmdd (19950313)
		end

	reported_currency (a_currency: CURRENCY; a_date: ROSE_DATE): CURRENCY
			-- Reporting currency for 'a_currency' on 'a_date'
			-- LUL and LUF are assigned to BEF
			-- BEC and ZAL are only valid for a fixed date range		
		require
			valid_currency: a_currency /= Void
			valid_date: a_date /= Void
		do
			if a_currency = usd or else a_currency = gbp or else a_currency = jpy then
				-- For efficiency, this handles the vast majority of queries.
				Result := a_currency
			elseif a_currency.same_currency (luf) or a_currency.same_currency (lul) then
				Result := bef
			elseif a_currency.same_currency (bec) then
				if a_date < first_bec or a_date > last_bec then
					Result := bef
				else
					Result := bec
				end
			elseif a_currency.same_currency (zal) then
				if a_date < first_zal or a_date > last_zal then
					Result := zar
				else
					Result := zal
				end
			else
				Result := a_currency
			end
		ensure
			attached_result: Result /= Void
		end

end


