indexing

	description: "A low-level string class to convert Eiffel strings to Win32 unicode strings."

	legal: "See notice at end of class."
	status: "See notice at end of class."
	date: "$Date: 2009/10/21 $"
	revision: "$Revision: #5 $"

class	C_STRING_32

inherit

	ANY

	ROSE_STRING_ROUTINES
		export {NONE} all end

create

	make,
	make_empty,
	make_by_pointer,
	make_by_pointer_and_count,
	make_shared_from_pointer,
	make_shared_from_pointer_and_count

feature --{NONE} -- Initialization

	make (a_string: STRING_GENERAL) is
			-- Make a C string from `a_string'.
			-- Embedded NULLs are lost.
		require
			a_string_not_void: a_string /= Void
		do
			make_empty (a_string.count)
			set_string (a_string)
		end

	make_empty (a_length: INTEGER) is
			-- Make an empty C string of `a_length' characters.
			-- All characters are assumed to be supplementary, for safety (i.e. memory is usually over-allocated).
			-- C memory area is not initialized.
		require
			a_length_positive: a_length >= 0
		do
			create managed_data.make ((2 * a_length + 1) * tchar_size)
			count := 0
		end

	make_by_pointer (a_ptr: POINTER) is
			-- Make a copy of string pointed by `a_ptr'.
			-- WARNING! `count' will be set too high if `a_ptr' contains surrogates.
			-- Restoring to a STRING_32 via `string' will reproduce the correct `count'.
		require
			a_ptr_not_null: a_ptr /= default_pointer
		do
			make_by_pointer_and_count (a_ptr, buffer_length (a_ptr))
		end

	make_by_pointer_and_count (a_ptr: POINTER; a_length: INTEGER) is
			-- Make a copy of first `a_length' bytes of string pointed by `a_ptr'.
			-- WARNING! `count' will be set too high if `a_ptr' contains surrogates.
			-- Restoring to a STRING_32 via `string' will reproduce the correct `count'.
		require
			a_ptr_not_null: a_ptr /= default_pointer
			a_length_non_negative: a_length >= 0
		do
			count := a_length // tchar_size
			create managed_data.make (a_length)
			managed_data.item.memory_copy (a_ptr, a_length)
		end

	make_shared_from_pointer (a_ptr: POINTER) is
			-- New instance sharing `a_ptr'.
		require
			a_ptr_not_null: a_ptr /= default_pointer
		do
			make_shared_from_pointer_and_count (a_ptr, buffer_length (a_ptr))
		end

	make_shared_from_pointer_and_count (a_ptr: POINTER; a_length: INTEGER) is
			-- New instance sharing `a_ptr' of `a_length' bytes.
		require
			a_ptr_not_null: a_ptr /= default_pointer
			a_length_non_negative: a_length >= 0
		do
			count := a_length // tchar_size
			create managed_data.share_from_pointer (a_ptr, a_length)
		end

feature -- Access

	string: STRING_32 is
			-- Eiffel string, ignoring `count'. Reads until a null character is being read.
			-- Assumes that `managed_data' points to valid UTF16LE data.
		local
			i, l_count: INTEGER
			l_data: like managed_data
			l_most, l_least, l_low_least, l_low_most: NATURAL_8
			l_finished: BOOLEAN
		do
			l_data := managed_data
			l_count := l_data.count
			create Result.make (l_count)
			from
				i := 1
			variant
				l_count - i + 1
			until
				l_finished or i > l_count
			loop
				l_least := l_data.read_natural_8_le (i - 1)
				l_most := l_data.read_natural_8_le (i)
				if l_least = 0 and l_most = 0 then
					l_finished := True
				elseif utf16.is_surrogate (l_most) then
					check
						valid_implies_high: utf16.is_high_surrogate (l_most)
						valid_implies_size: i + 2 < l_count
						-- Unsound checks - validity is only assumed, not inferred
					end
					i := i + 2
					l_low_least := l_data.read_natural_8_le (i - 1)
					l_low_most := l_data.read_natural_8_le (i)
					check
						valid_implies_low: utf16.is_low_surrogate (l_low_most)
						-- Unsound check - validity is only assumed, not inferred
					end
					Result.append_character (utf16.surrogate_from_bytes (l_most, l_least, l_low_most, l_low_least).to_character_32)
				else
					Result.append_character ((l_most * 256 + l_least).to_character_32)
				end
				i := i + 2
			end
		ensure
			string_not_void: Result /= Void
		end

	item: POINTER is
			-- Get pointer to allocated area.
		do
			Result := managed_data.item
		ensure
			item_not_null: Result /= default_pointer
		end

	managed_data: MANAGED_POINTER
			-- Hold data of Current.

feature -- Measurement

	count: INTEGER
			-- Number of characters in Current.

feature -- Element change

	set_string (a_string: STRING_GENERAL) is
			-- Set `string' with `a_string'.
		require
			a_string_not_void: a_string /= Void
		local
			l_bytes: STRING_8
			i: INTEGER
			l_new_size: INTEGER
		do
			l_bytes := to_utf16_le (a_string)
			count := a_string.count
			l_new_size := l_bytes.count + tchar_size	-- Extra character for terminating NULL char
			if managed_data.count < l_new_size  then
				managed_data.resize (l_new_size)
			end
			from
				i := 0
			variant
				l_new_size - i
			until
				i = l_new_size - 2
			loop
				managed_data.put_natural_8 (l_bytes.item_code (i + 1).to_natural_8, i)
				i := i +  1
			end
			managed_data.put_natural_16 (0, l_new_size - tchar_size)
		end

feature {NONE} -- Implementation

	buffer_length (ptr: POINTER): INTEGER is
			-- Number of bytes to hold `ptr'.
		do
			Result := c_strlen (ptr) * tchar_size
		end

	tchar_size: INTEGER is
			-- Number of bytes occupied by a TCHAR.
		external
			"C inline use <tchar.h>"
		alias
			"sizeof(TCHAR)"
		end

	c_strlen (ptr: POINTER): INTEGER is
			-- Number of UCS-2 characters in `ptr' (maybe more than number of Unicode characters).
		external
			"C inline use <tchar.h>"
		alias
			"_tcslen ((wchar_t *) $ptr)"
		end

invariant

	managed_data_not_void: managed_data /= Void
	count_not_negative: count >= 0
	UCS_2_units: tchar_size = 2

indexing
	copyright:	"Copyright (c) 1984-2006, Eiffel Software and others"
	license:	"Eiffel Forum License v2 (see http://www.eiffel.com/licensing/forum.txt)"
	source: "[
			 Eiffel Software
			 356 Storke Road, Goleta, CA 93117 USA
			 Telephone 805-685-1006, Fax 805-685-6869
			 Website http://www.eiffel.com
			 Customer support http://support.eiffel.com
		]"

end
