indexing
	description: "64-bit value representing the number of 100-nanosecond intervals since January 1, 1601 (UTC)"
	author: "Zoran Simic"
	date: "$Date: 2009/09/11 $"
	revision: "$Revision: #5 $"
	archive: "$File: //rose/source/kernel/infrastructure/windows/c_struct_filetime.e $"
	original: "[
		typedef struct _FILETIME {
			DWORD dwLowDateTime;
			DWORD dwHighDateTime;
		} FILETIME, *PFILETIME;
	]"

class
	C_STRUCT_FILETIME

inherit
	C_STRUCT

create
	make, set_ticks, set_from_systemtime, make_now_utc

convert
	item: {POINTER},
	ticks: {NATURAL_64},
	date: {ROSE_DATE},
	set_ticks ({NATURAL_64}),
	set_from_systemtime ({C_STRUCT_SYSTEMTIME})

feature -- Initialization

	make_now_utc is
			-- Current system date and time, expressed in Coordinated Universal Time (UTC)
		do
			make
			GetSystemTimeAsFileTime (item)
		end

	set_ticks (a_value: NATURAL_64) is
			-- Set 'ticks', unit is 100 nanoseconds since epoch.
		do
			make
			pointer.put_natural_64 (a_value, 0)
		end

	set_from_systemtime (a_systemtime: C_STRUCT_SYSTEMTIME) is
			-- Set from 'a_systemtime'
		require
			time_ok: a_systemtime /= Void
		do
			make
			if SystemTimeToFileTime (a_systemtime.item, item) then
			end
		end

feature -- Access

	struct_size: INTEGER is
			-- Size of this external C structure
		external
			"C blocking inline use <windows.h>"
		alias
			"sizeof(FILETIME)"
		end

	ticks_per_second: NATURAL_64 is 10_000_000
			-- Ticks' unit is 100 nanoseconds, so 10^7 ticks in one second

	ticks: NATURAL_64 is
			-- Represented as ticks, units is 100 nanoseconds since epoch
		require
			pointer_ok: pointer /= Void
		do
			Result := pointer.read_natural_64 (0)
		end

	dwLowDateTime: INTEGER is
			-- Low-order part of the file time.
		require
			pointer_ok: pointer /= Void
		do
			Result := c_get_dwLowDateTime (pointer.item)
		end

	dwHighDateTime: INTEGER is
			-- High-order part of the file time.
		require
			pointer_ok: pointer /= Void
		do
			Result := c_get_dwHighDateTime (pointer.item)
		end

	systemtime: C_STRUCT_SYSTEMTIME is
			-- Equivalent SYSTEMTIME structure
		require
			pointer_ok: pointer /= Void
		do
			create Result.set_from_filetime (Current)
		ensure
			result_ok: Result /= Void
		end

	date: ROSE_DATE is
			-- UTC date contained in this FILETIME structure
		require
			pointer_ok: pointer /= Void
		do
			Result := systemtime.date
		end

	localized_date: ROSE_DATE is
			-- Localized date contained in this FILETIME structure (which itself is UTC)
		require
			pointer_ok: pointer /= Void
		local
			l_local: C_STRUCT_FILETIME
		do
			create l_local.make
			if FileTimeToLocalFileTime (item, l_local.item) then
				Result := l_local.date
			end
		end

	seconds: INTEGER_64 is
			-- Number of seconds since epoch (1/1/1601)
		require
			pointer_ok: pointer /= Void
		do
			Result := (ticks // ticks_per_second).as_integer_64
		end

	infix "-" (a_other: like Current): NATURAL_64 is
			-- Result of subtracting `other'
		do
			if a_other /= Void then
				Result := ticks - a_other.ticks
			else
				Result := ticks
			end
		end

feature -- Basic operations

	set_dwLowDateTime (a_value: INTEGER) is
			-- Set 'dwLowDateTime'.
		require
			pointer_ok: pointer /= Void
		do
			c_set_dwLowDateTime (pointer.item, a_value)
		end

	set_dwHighDateTime (a_value: INTEGER) is
			-- Set 'dwHighDateTime'.
		require
			pointer_ok: pointer /= Void
		do
			c_set_dwHighDateTime (pointer.item, a_value)
		end

feature {NONE} -- Implementation

	GetSystemTimeAsFileTime (lpSystemTimeAsFileTime: POINTER) is
			-- Retrieves the current system date and time.
			-- The information is in Coordinated Universal Time (UTC) format
		external
			"C blocking signature (LPFILETIME) use <windows.h>"
		alias
			"GetSystemTimeAsFileTime"
		end

	SystemTimeToFileTime (lpSystemTime, lpFileTime: POINTER): BOOLEAN is
			-- Retrieves the current system date and time.
			-- The information is in Coordinated Universal Time (UTC) format
		external
			"C blocking signature (SYSTEMTIME*, LPFILETIME): EIF_BOOLEAN use <windows.h>"
		alias
			"SystemTimeToFileTime"
		end

	FileTimeToLocalFileTime (lpFileTime, lpLocalFileTime: POINTER): BOOLEAN is
			-- Converts a file time to a local file time
		external
			"C blocking signature (FILETIME*, LPFILETIME): EIF_BOOLEAN use <windows.h>"
		alias
			"FileTimeToLocalFileTime"
		end

	c_get_dwLowDateTime (a_pointer: POINTER): INTEGER
			-- Access to dwLowDateTime structure member
		external
			"C blocking inline use <windows.h>"
		alias
			"((FILETIME *) $a_pointer)->dwLowDateTime"
		end

	c_get_dwHighDateTime (a_pointer: POINTER): INTEGER
			-- Access to dwHighDateTime structure member
		external
			"C blocking inline use <windows.h>"
		alias
			"((FILETIME *) $a_pointer)->dwHighDateTime"
		end

	c_set_dwLowDateTime (a_pointer: POINTER; a_value: INTEGER)
			-- Set 'dwLowDateTime' member of this structure with 'a_value'
		external
			"C blocking inline use <windows.h>"
		alias
			"((FILETIME *) $a_pointer)->dwLowDateTime = (DWORD)$a_value"
		end

	c_set_dwHighDateTime (a_pointer: POINTER; a_value: INTEGER)
			-- Set 'dwHighDateTime' member of this structure with 'a_value'
		external
			"C blocking inline use <windows.h>"
		alias
			"((FILETIME *) $a_pointer)->dwHighDateTime = (DWORD)$a_value"
		end

end
