indexing
	description: "Contains the memory statistics for a process"
	author: "Zoran Simic"
	date: "$Date: 2009/09/11 $"
	revision: "$Revision: #5 $"
	archive: "$File: //rose/source/kernel/infrastructure/windows/c_struct_process_memory_counters.e $"
	original: "[
		typedef struct _PROCESS_MEMORY_COUNTERS {
			DWORD cb;
			DWORD PageFaultCount;
			SIZE_T PeakWorkingSetSize;
			SIZE_T WorkingSetSize;
			SIZE_T QuotaPeakPagedPoolUsage;
			SIZE_T QuotaPagedPoolUsage;
			SIZE_T QuotaPeakNonPagedPoolUsage;
			SIZE_T QuotaNonPagedPoolUsage;
			SIZE_T PagefileUsage;
			SIZE_T PeakPagefileUsage;
		} PROCESS_MEMORY_COUNTERS, *PPROCESS_MEMORY_COUNTERS;
	]"

class
	C_STRUCT_PROCESS_MEMORY_COUNTERS

inherit
	C_STRUCT
		rename
			make as make_struct
		end

create
	make

convert
	item: {POINTER}

feature {NONE} -- Initialization

	make (a_parent: like parent) is
			-- Create bound to 'a_parent' process info
		require
			parent_ok: a_parent /= Void
		do
			parent := a_parent
			make_struct
			c_set_cb (pointer.item, struct_size)
		end

feature -- Access

	struct_size: INTEGER is
			-- Size of this external C structure
		external
			"C blocking inline use <psapi.h>"
		alias
			"sizeof(PROCESS_MEMORY_COUNTERS)"
		end

	parent: WINDOWS_PROCESS
			-- Parent process for which these memory counters apply

	cb: INTEGER is
			-- Size of this structure, in bytes.
		require
			pointer_ok: pointer /= Void
		do
			Result := c_get_cb (pointer.item)
		end

	PageFaultCount: INTEGER is
			-- Number of page faults.
		require
			pointer_ok: pointer /= Void
		do
			Result := c_get_PageFaultCount (pointer.item)
		end

	PeakWorkingSetSize: NATURAL_64 is
			-- Peak working set size, in bytes.
		require
			pointer_ok: pointer /= Void
		do
			Result := c_get_PeakWorkingSetSize (pointer.item)
		end

	WorkingSetSize: NATURAL_64 is
			-- Current working set size, in bytes.
		require
			pointer_ok: pointer /= Void
		do
			Result := c_get_WorkingSetSize (pointer.item)
		end

	QuotaPeakPagedPoolUsage: NATURAL_64 is
			-- Peak paged pool usage, in bytes.
		require
			pointer_ok: pointer /= Void
		do
			Result := c_get_QuotaPeakPagedPoolUsage (pointer.item)
		end

	QuotaPagedPoolUsage: NATURAL_64 is
			-- Current paged pool usage, in bytes.
		require
			pointer_ok: pointer /= Void
		do
			Result := c_get_QuotaPagedPoolUsage (pointer.item)
		end

	QuotaPeakNonPagedPoolUsage: NATURAL_64 is
			-- Peak nonpaged pool usage, in bytes.
		require
			pointer_ok: pointer /= Void
		do
			Result := c_get_QuotaPeakNonPagedPoolUsage (pointer.item)
		end

	QuotaNonPagedPoolUsage: NATURAL_64 is
			-- Current nonpaged pool usage, in bytes.
		require
			pointer_ok: pointer /= Void
		do
			Result := c_get_QuotaNonPagedPoolUsage (pointer.item)
		end

	PagefileUsage: NATURAL_64 is
			-- Current space allocated for the pagefile, in bytes. Those pages may or may not be in memory.
		require
			pointer_ok: pointer /= Void
		do
			Result := c_get_PagefileUsage (pointer.item)
		end

	PeakPagefileUsage: NATURAL_64 is
			-- Peak space allocated for the pagefile, in bytes.
		require
			pointer_ok: pointer /= Void
		do
			Result := c_get_PeakPagefileUsage (pointer.item)
		end

feature -- Basic operations

	refresh is
			-- Refresh the data of this structure from OS.
		require
			pointer_ok: pointer /= Void
		do
			GetProcessMemoryInfo (parent.handle, item, cb)
		end

feature {NONE} -- Implementation

	GetProcessMemoryInfo (hProcess, ppsmemCounters: POINTER; a_length: INTEGER) is
			-- Retrieves information about the memory usage of the specified process.
		external
			"C blocking signature (HANDLE, PPROCESS_MEMORY_COUNTERS, DWORD) use <psapi.h>"
		alias
			"GetProcessMemoryInfo"
		end

	c_set_cb (a_pointer: POINTER; a_value: INTEGER)
			-- Set 'cb' member of this structure with 'a_value'
		external
			"C blocking inline use <windows.h>"
		alias
			"((PROCESS_MEMORY_COUNTERS *) $a_pointer)->cb = (DWORD)$a_value"
		end

	c_get_cb (a_pointer: POINTER): INTEGER
			-- Access to cb structure member
		external
			"C blocking inline use <windows.h>"
		alias
			"((PROCESS_MEMORY_COUNTERS *) $a_pointer)->cb"
		end

	c_get_PageFaultCount (a_pointer: POINTER): INTEGER
			-- Access to PageFaultCount structure member
		external
			"C blocking inline use <windows.h>"
		alias
			"((PROCESS_MEMORY_COUNTERS *) $a_pointer)->PageFaultCount"
		end

	c_get_PeakWorkingSetSize (a_pointer: POINTER): NATURAL_64
			-- Access to PeakWorkingSetSize structure member
		external
			"C blocking inline use <windows.h>"
		alias
			"((PROCESS_MEMORY_COUNTERS *) $a_pointer)->PeakWorkingSetSize"
		end

	c_get_WorkingSetSize (a_pointer: POINTER): NATURAL_64
			-- Access to WorkingSetSize structure member
		external
			"C blocking inline use <windows.h>"
		alias
			"((PROCESS_MEMORY_COUNTERS *) $a_pointer)->WorkingSetSize"
		end

	c_get_QuotaPeakPagedPoolUsage (a_pointer: POINTER): NATURAL_64
			-- Access to QuotaPeakPagedPoolUsage structure member
		external
			"C blocking inline use <windows.h>"
		alias
			"((PROCESS_MEMORY_COUNTERS *) $a_pointer)->QuotaPeakPagedPoolUsage"
		end

	c_get_QuotaPagedPoolUsage (a_pointer: POINTER): NATURAL_64
			-- Access to QuotaPagedPoolUsage structure member
		external
			"C blocking inline use <windows.h>"
		alias
			"((PROCESS_MEMORY_COUNTERS *) $a_pointer)->QuotaPagedPoolUsage"
		end

	c_get_QuotaPeakNonPagedPoolUsage (a_pointer: POINTER): NATURAL_64
			-- Access to QuotaPeakNonPagedPoolUsage structure member
		external
			"C blocking inline use <windows.h>"
		alias
			"((PROCESS_MEMORY_COUNTERS *) $a_pointer)->QuotaPeakNonPagedPoolUsage"
		end

	c_get_QuotaNonPagedPoolUsage (a_pointer: POINTER): NATURAL_64
			-- Access to QuotaNonPagedPoolUsage structure member
		external
			"C blocking inline use <windows.h>"
		alias
			"((PROCESS_MEMORY_COUNTERS *) $a_pointer)->QuotaNonPagedPoolUsage"
		end

	c_get_PagefileUsage (a_pointer: POINTER): NATURAL_64
			-- Access to PagefileUsage structure member
		external
			"C blocking inline use <windows.h>"
		alias
			"((PROCESS_MEMORY_COUNTERS *) $a_pointer)->PagefileUsage"
		end

	c_get_PeakPagefileUsage (a_pointer: POINTER): NATURAL_64
			-- Access to PeakPagefileUsage structure member
		external
			"C blocking inline use <windows.h>"
		alias
			"((PROCESS_MEMORY_COUNTERS *) $a_pointer)->PeakPagefileUsage"
		end

end
