indexing
	description: "The SYSTEMTIME structure represents a date and time using individual members for the month, day, year, weekday, hour, minute, second, and millisecond"
	author: "Zoran Simic"
	date: "$Date: 2009/09/11 $"
	revision: "$Revision: #4 $"
	archive: "$File: //rose/source/kernel/infrastructure/windows/c_struct_systemtime.e $"
	original: "[
		typedef struct _SYSTEMTIME {
			WORD wYear;
			WORD wMonth;
			WORD wDayOfWeek;
			WORD wDay;
			WORD wHour;
			WORD wMinute;
			WORD wSecond;
			WORD wMilliseconds;
		} SYSTEMTIME, *PSYSTEMTIME;
	]"

class
	C_STRUCT_SYSTEMTIME

inherit
	C_STRUCT

create
	make, set_from_date, set_from_filetime, make_now_utc, make_local

convert
	item: {POINTER},
	date: {ROSE_DATE},
	set_from_date ({ROSE_DATE}),
	set_from_filetime ({C_STRUCT_FILETIME})

feature -- Initialization

	make_now_utc is
			-- Current system date and time, expressed in Coordinated Universal Time (UTC)
		do
			make
			GetSystemTime (item)
		end

	make_local is
			-- Current local date and time
		do
			make
			GetLocalTime (item)
		end

feature -- Access

	struct_size: INTEGER is
			-- Size of this external C structure
		external
			"C blocking inline use <windows.h>"
		alias
			"sizeof(SYSTEMTIME)"
		end

	wYear: INTEGER_16 is
			-- The year (1601 - 30827).
		require
			pointer_ok: pointer /= Void
		do
			Result := c_wYear (pointer.item)
		end

	wMonth: INTEGER_16 is
			-- The month.(January = 1 .. December = 12)
		require
			pointer_ok: pointer /= Void
		do
			Result := c_wMonth (pointer.item)
		end

	wDayOfWeek: INTEGER_16 is
			-- The day of the week. (Sunday = 0 .. Saturday = 6)
		require
			pointer_ok: pointer /= Void
		do
			Result := c_wDayOfWeek (pointer.item)
		end

	wDay: INTEGER_16 is
			-- The day of the month (1-31).
		require
			pointer_ok: pointer /= Void
		do
			Result := c_wDay (pointer.item)
		end

	wHour: INTEGER_16 is
			-- The hour (0-23).
		require
			pointer_ok: pointer /= Void
		do
			Result := c_wHour (pointer.item)
		end

	wMinute: INTEGER_16 is
			-- The minute (0-59).
		require
			pointer_ok: pointer /= Void
		do
			Result := c_wMinute (pointer.item)
		end

	wSecond: INTEGER_16 is
			-- The second (0-59).
		require
			pointer_ok: pointer /= Void
		do
			Result := c_wSecond (pointer.item)
		end

	wMilliseconds: INTEGER_16 is
			-- The millisecond (0-999).
		require
			pointer_ok: pointer /= Void
		do
			Result := c_wMilliseconds (pointer.item)
		end

	date: ROSE_DATE is
			-- UTC ROSE_DATE corresponding to this SYSTEMTIME structure
		require
			pointer_ok: pointer /= Void
		do
			if wYear > 1601 or else wMonth > 1 or else wDay > 1 then
				create Result.make_fine (wYear, wMonth, wDay, wHour, wMinute, wSecond + wMilliseconds / 1000)
			end
		end

feature -- Basic operations

	set_wYear (a_value: INTEGER_16) is
			-- Set year (1601 - 30827).
		require
			year_ok: a_value >= 1601 and then a_value <= 30827
		do
			c_set_wYear (pointer.item, a_value)
		end

	set_wMonth (a_value: INTEGER_16) is
			-- Set month.(January = 1 .. December = 12)
		require
			month_ok: a_value >= 1 and then a_value <= 12
		do
			c_set_wMonth (pointer.item, a_value)
		end

	set_wDayOfWeek (a_value: INTEGER_16) is
			-- Set day of the week. (Sunday = 0 .. Saturday = 6)
		require
			day_ok: a_value >= 0 and then a_value <= 6
		do
			c_set_wDayOfWeek (pointer.item, a_value)
		end

	set_wDay (a_value: INTEGER_16) is
			-- Set day of the month (1-31).
		require
			day_ok: a_value >= 1 and then a_value <= 31
		do
			c_set_wDay (pointer.item, a_value)
		end

	set_wHour (a_value: INTEGER_16) is
			-- Set hour (0-23).
		require
			hour_ok: a_value >= 0 and then a_value <= 23
		do
			c_set_wHour (pointer.item, a_value)
		end

	set_wMinute (a_value: INTEGER_16) is
			-- Set minute (0-59).
		require
			minute_ok: a_value >= 0 and then a_value <= 59
		do
			c_set_wMinute (pointer.item, a_value)
		end

	set_wSecond (a_value: INTEGER_16) is
			-- Set second (0-59).
		require
			second_ok: a_value >= 0 and then a_value <= 59
		do
			c_set_wSecond (pointer.item, a_value)
		end

	set_wMilliseconds (a_value: INTEGER_16) is
			-- Set millisecond (0-999).
		require
			ms_ok: a_value >= 0 and then a_value <= 999
		do
			c_set_wMilliseconds (pointer.item, a_value)
		end

	set_from_date (a_date: ROSE_DATE) is
			-- Set from UTC 'a_date'
		local
			l_time: ROSE_TIME
		do
			make
			if a_date /= Void then
				l_time := a_date.time
				set_wYear (a_date.year.to_integer_16)
				set_wMonth (a_date.month.to_integer_16)
				set_wDayOfWeek ((a_date.day_of_the_week - 1).to_integer_16)
				set_wDay (a_date.day.to_integer_16)
				set_wHour (l_time.hour.to_integer_16)
				set_wMinute (l_time.minute.to_integer_16)
				set_wSecond (l_time.second.to_integer_16)
				set_wMilliseconds (l_time.fractional_second.truncated_to_integer.to_integer_16)
			else
				set_wYear (1601)
				set_wMonth (1)
				set_wDayOfWeek (0)
				set_wDay (1)
				set_wHour (0)
				set_wMinute (0)
				set_wSecond (0)
				set_wMilliseconds (0)
			end
		end

	set_from_filetime (a_filetime: C_STRUCT_FILETIME) is
			-- Set from 'a_filetime'
		require
			time_ok: a_filetime /= Void
		do
			make
			if FileTimeToSystemTime (a_filetime.item, item) then
			end
		end

	set_system_time_with_current: BOOLEAN is
			-- Sets the current system time and date with values in this object
			-- The system time is expressed in Coordinated Universal Time (UTC)
			-- Returns True if time was set succesfully
			-- False otherwise
		do
			Result := SetSystemTime (item)
		end

feature {NONE} -- Implementation

	FileTimeToSystemTime (lpFileTime, lpSystemTime: POINTER): BOOLEAN is
			-- The FileTimeToSystemTime function converts a file time to system time format.
		external
			"C blocking signature (FILETIME*, LPSYSTEMTIME): EIF_BOOLEAN use <windows.h>"
		alias
			"FileTimeToSystemTime"
		end

	GetSystemTime (lpSystemTime: POINTER) is
			-- Retrieves the current system date and time.
			-- The system time is expressed in Coordinated Universal Time (UTC)
		external
			"C blocking signature (LPSYSTEMTIME) use <windows.h>"
		alias
			"GetSystemTime"
		end

	GetLocalTime (lpSystemTime: POINTER) is
			-- Retrieves the current local date and time
		external
			"C blocking signature (LPSYSTEMTIME) use <windows.h>"
		alias
			"GetLocalTime"
		end

	SetSystemTime (lpSystemTime: POINTER): BOOLEAN is
			-- Sets the current system time and date.
			-- The system time is expressed in Coordinated Universal Time (UTC)
		external
			"C blocking signature (LPSYSTEMTIME): EIF_BOOLEAN use <windows.h>"
		alias
			"SetSystemTime"
		end

	c_wYear (a_pointer: POINTER): INTEGER_16
			-- Access to wYear structure member
		external
			"C blocking inline use <windows.h>"
		alias
			"((SYSTEMTIME *) $a_pointer)->wYear"
		end

	c_wMonth (a_pointer: POINTER): INTEGER_16
			-- Access to wMonth structure member
		external
			"C blocking inline use <windows.h>"
		alias
			"((SYSTEMTIME *) $a_pointer)->wMonth"
		end

	c_wDayOfWeek (a_pointer: POINTER): INTEGER_16
			-- Access to wDayOfWeek structure member
		external
			"C blocking inline use <windows.h>"
		alias
			"((SYSTEMTIME *) $a_pointer)->wDayOfWeek"
		end

	c_wDay (a_pointer: POINTER): INTEGER_16
			-- Access to wDay structure member
		external
			"C blocking inline use <windows.h>"
		alias
			"((SYSTEMTIME *) $a_pointer)->wDay"
		end

	c_wHour (a_pointer: POINTER): INTEGER_16
			-- Access to wHour structure member
		external
			"C blocking inline use <windows.h>"
		alias
			"((SYSTEMTIME *) $a_pointer)->wHour"
		end

	c_wMinute (a_pointer: POINTER): INTEGER_16
			-- Access to wMinute structure member
		external
			"C blocking inline use <windows.h>"
		alias
			"((SYSTEMTIME *) $a_pointer)->wMinute"
		end

	c_wSecond (a_pointer: POINTER): INTEGER_16
			-- Access to wSecond structure member
		external
			"C blocking inline use <windows.h>"
		alias
			"((SYSTEMTIME *) $a_pointer)->wSecond"
		end

	c_wMilliseconds (a_pointer: POINTER): INTEGER_16
			-- Access to wMilliseconds structure member
		external
			"C blocking inline use <windows.h>"
		alias
			"((SYSTEMTIME *) $a_pointer)->wMilliseconds"
		end

	c_set_wYear (a_pointer: POINTER; a_value: INTEGER_16)
			-- Set 'wYear' member of this structure with 'a_value'
		external
			"C blocking inline use <windows.h>"
		alias
			"((SYSTEMTIME *) $a_pointer)->wYear = (WORD)$a_value"
		end

	c_set_wMonth (a_pointer: POINTER; a_value: INTEGER_16)
			-- Set 'wMonth' member of this structure with 'a_value'
		external
			"C blocking inline use <windows.h>"
		alias
			"((SYSTEMTIME *) $a_pointer)->wMonth = (WORD)$a_value"
		end

	c_set_wDayOfWeek (a_pointer: POINTER; a_value: INTEGER_16)
			-- Set 'wDayOfWeek' member of this structure with 'a_value'
		external
			"C blocking inline use <windows.h>"
		alias
			"((SYSTEMTIME *) $a_pointer)->wDayOfWeek = (WORD)$a_value"
		end

	c_set_wDay (a_pointer: POINTER; a_value: INTEGER_16)
			-- Set 'wDay' member of this structure with 'a_value'
		external
			"C blocking inline use <windows.h>"
		alias
			"((SYSTEMTIME *) $a_pointer)->wDay = (WORD)$a_value"
		end

	c_set_wHour (a_pointer: POINTER; a_value: INTEGER_16)
			-- Set 'wHour' member of this structure with 'a_value'
		external
			"C blocking inline use <windows.h>"
		alias
			"((SYSTEMTIME *) $a_pointer)->wHour = (WORD)$a_value"
		end

	c_set_wMinute (a_pointer: POINTER; a_value: INTEGER_16)
			-- Set 'wMinute' member of this structure with 'a_value'
		external
			"C blocking inline use <windows.h>"
		alias
			"((SYSTEMTIME *) $a_pointer)->wMinute = (WORD)$a_value"
		end

	c_set_wSecond (a_pointer: POINTER; a_value: INTEGER_16)
			-- Set 'wSecond' member of this structure with 'a_value'
		external
			"C blocking inline use <windows.h>"
		alias
			"((SYSTEMTIME *) $a_pointer)->wSecond = (WORD)$a_value"
		end

	c_set_wMilliseconds (a_pointer: POINTER; a_value: INTEGER_16)
			-- Set 'wMilliseconds' member of this structure with 'a_value'
		external
			"C blocking inline use <windows.h>"
		alias
			"((SYSTEMTIME *) $a_pointer)->wMilliseconds = (WORD)$a_value"
		end

end
