note

	description:

		"Handler of category of entities from a supplier context: provides strongly typed entities, cacheing, peers."

	author: "Mark Howard"
	last_editor: "$Author: er $"
	date: "$Date: 2014/02/26 $"
	revision: "$Revision: #88 $"
	path: "$File: //rose/source/kernel/entities/reference/category_handler.e $"

class CATEGORY_HANDLER [reference H -> SUPPLIER_CONTEXT, reference G -> ENTITY create make, make_anonymous end]

inherit

	ROSE_ID_USER
		export
			{NONE} all
		end
	ROSE_CONSTANTS
		export
			{NONE} all
		end
	MOVIE_AWARE
		export
			{NONE} all
		end
	DEBUG_OUTPUT
		export
			{NONE} all
		end

	EXCEPTIONS
		export
			{NONE} all
		end

	PES_AWARE_CONTEXT

	UID_AWARE

create
	make

feature {NONE} -- Initialization

	make (a_context: H; a_category_id: INTEGER) is
			-- Create bound to `a_context' and `a_category_id'.
		require
			context_not_void: a_context /= Void
			category_id_sensible: a_category_id > 0
		do
			context := a_context
			category_id := a_category_id
			create ur_entity.make_anonymous (a_context, category_id)
			name := ids.category_name (category_id)
			if name = Void and then movie /= Void and then movie.playback_movie then
				-- This happens to a movie when a new category is added.
				-- We will skip it on the assumption that if the category didn't exist when
				-- the movie was made then it won't be needed now.
				-- The fact that 'identifiers' is Void will be used by ENTITY_ACCESSORS to
				-- show that this is a new category that it not in the movie.
				name := "Unknown_" + a_category_id.out
			else
				if name = Void then
					name := "Unknown_" + a_category_id.out
				end
				identifiers := context.category_identifiers (category_id)
			end
			full_name := "[" + a_context.name + "]" + name
		ensure
			context_not_void: context /= Void
			category_id_sensible: category_id > 0
			ur_entity_not_void: ur_entity /= Void
			ur_entity_matches_category_id: ur_entity.category_id = category_id
--			ur_entity_matches_context: ur_entity.context = context
			identifiers_not_void: identifiers /= Void
			name_not_void: name /= Void
		end

feature -- Access

	category_id: INTEGER
			-- Associated category id

	context: H
			-- Associated context

	name: STRING
			-- Category name, within 'context'

	full_name: STRING
			-- Full category name, of the form "[SUPPLIER]CATEGORY"

	ur_entity: G
			-- Generic entity type, represents the type returned by this category handler

	identifiers: HASH_TABLE [ENTITY_IDENTIFIER, INTEGER]
			-- Identifiers defined in this category

	is_bound_to_rose: BOOLEAN is
			-- Is this context bound to the rose-best context?
		do
			Result := context.has_rose_handles
		end

	debug_output: STRING is
			-- String that should be displayed in debugger to represent `Current'.
		do
			Result := full_name
		end
feature -- Entities

	entity_from_uid (a_uid : INTEGER_64) : like ur_entity is
			-- Entity corresponding to `a_uid';
			-- Since we know the context and category we just use the handle from the uid.
		require
			right_category_context : context.supplier_id = a_uid.bit_and (0xFFFF000000000000) |>> 48 and then category_id = a_uid.bit_and (0x0000FFFF00000000) |>> 32
		do
			Result := entity(id_handler.uid_handle(a_uid))
		end

	entity (a_handle: INTEGER): like ur_entity is
			-- Entity with `a_handle'
		local
			l_record: ENTITY_ID_BASE
			l_records: ROSE_LINEAR_ARRAY [ENTITY_ID_BASE]
		do
			Result := table.item (a_handle)
			if Result = Void then
				if a_handle = 0 then
					Result := ur_entity
				else
					l_record := entity_id_table.item (a_handle)
					if l_record = Void then
						l_records := context.category_entity_records (Current, a_handle)
						if l_records.count > 0 then
							l_record := l_records.item (1)
						end
					end
					if l_record /= Void then
						add_entity (l_record)
					end
					Result := table.item (a_handle)
				end
			end
		end

	any_entity_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like ur_entity is
			-- Entity for a given identifier
		do
			Result := entity_from_identifier (a_attribute.key, a_date, context.identifier_from_rose (a_attribute.id))
		end


	entity_from_table (a_handle: INTEGER): like ur_entity is
			-- Entity with 'a_handle' if in table
		do
			Result := table.item (a_handle)
		end

	history (a_identifier: ENTITY_IDENTIFIER; a_key: STRING): ROSE_ENTITY_HISTORY [like ur_entity] is
			-- History for a given identifier
		require
			identifier_not_void: a_identifier /= Void
			key_not_void: a_key /= Void
		local
			l_records: ROSE_LINEAR_ARRAY [ENTITY_ATTRIBUTE_RECORD]
			l_start_date, l_end_date: ROSE_DATE
			l_interval: ENTITY_INTERVAL [like ur_entity]
		do
			create Result.make_empty
			l_records := context.entity_records (a_key, Current, a_identifier)
			if l_records /= Void and then l_records.count > 0 then
				from l_records.start until l_records.off loop
					l_start_date := l_records.current_item.start_date
					l_end_date := l_records.current_item.end_date
					create l_interval.make (entity (l_records.current_item.entity_id), l_start_date, l_end_date, 1)
					Result.extend (l_interval)
					l_records.forth
				end
			end
		ensure
			result_not_void: Result /= Void
		end

	identity (a_entity: ENTITY): like ur_entity is
			--entity from this context with the same entity_id as a_entity
		require
			entity_not_void: a_entity /= Void
			category_matches: a_entity.category_id = category_id
		do
			if a_entity.category_id = category_id then
				Result := entity (a_entity.handle)
			end
		ensure
			result_not_void: a_entity.category_id = category_id implies Result /= Void
		end

	all_entities: ROSE_LINEAR_ARRAY [like ur_entity] is
			-- All entities in this category (and context)
		do
			if not all_entities_retrieved then
				load_all_entities
				if entity_id_table_saved /= Void then
					from
						entity_id_table_saved.start
					until
						entity_id_table_saved.off
					loop
						add_entity (entity_id_table_saved.item_for_iteration)
						entity_id_table_saved.forth
					end
				end
				all_entities_retrieved := True
				create all_entities_saved.make_from_linear (table.linear_representation)
			end
			Result := all_entities_saved
		ensure
			all_entities_retrieved: all_entities_retrieved
			table_not_void: table /= Void
		end

	entity_from_identifier (a_key: STRING; a_date: ROSE_DATE; a_identifier: ENTITY_IDENTIFIER): like ur_entity is
			-- Get the first matching entity whose attribute from `a_identifier' matches `a_key'
		require
			same_context: context = a_identifier.context
		do
			Result := entity_list_from_identifier (a_key, a_date, a_identifier).safe_item (1)
		end

	entity_from_identifier_case_sensitive (a_key: STRING; a_date: ROSE_DATE; a_identifier: ENTITY_IDENTIFIER): like ur_entity is
			-- Get the first matching entity whose attribute from `a_identifier' matches `a_key' case_sensitively.
			-- Note that case_sensitive matching on identifiers is usually not required. However, in the case of rics, it is (e.g. TA.TO versus Ta.TO).
		require
			same_context: context = a_identifier.context
		local
			l_entities : ROSE_LINEAR_ARRAY [like ur_entity]
			l_entity : like ur_entity
		do
			l_entities := entity_list_from_identifier (a_key, a_date, a_identifier)
			from l_entities.start until l_entities.off or Result /= Void loop
				l_entity := l_entities.current_item
				if l_entity.attribute_at (context.rose_attribute_from_local [a_identifier.handle], a_date) ~ a_key then
					Result := l_entity
				end
				l_entities.forth
			end
		end

	most_recent_entity_from_identifer (a_key: STRING; a_date: ROSE_DATE; a_identifier: ENTITY_IDENTIFIER): ENTITY_INTERVAL [like ur_entity] is
		local
			l_date, l_end_date, l_start_date: ROSE_DATE
			l_records: ROSE_LINEAR_ARRAY [ENTITY_ATTRIBUTE_RECORD]
			l_entity: like ur_entity
		do
			l_records := context.entity_records (a_key, Current, a_identifier)
			if l_records.count > 0 then
				if a_date = Void then
					l_entity := entity (l_records.item (1).entity_id)
					l_start_date := l_records.item (1).start_date
					l_end_date := l_records.item (1).end_date
				else
					l_date := valid_date (a_date)
					from
						l_records.start
					until
						l_entity /= Void or else l_records.off
					loop
						l_start_date := l_records.current_item.start_date
						l_end_date := l_records.current_item.end_date
						if l_start_date <= l_date then
							l_entity := entity (l_records.current_item.entity_id)
						end
						l_records.forth
					end
				end
				if l_entity /= Void then
					create Result.make (l_entity, l_start_date, l_end_date, 1)
				end
			end
		end

	entity_list_from_identifier (a_key: STRING; a_date: ROSE_DATE; a_identifier: ENTITY_IDENTIFIER): ROSE_LINEAR_ARRAY [like ur_entity] is
			-- List of entities given an identifier
		local
			l_date, l_end_date, l_start_date: ROSE_DATE
			l_records: ROSE_LINEAR_ARRAY [ENTITY_ATTRIBUTE_RECORD]
			l_entity: like ur_entity
			l_query: PES_QUERY_ENTITY_LIST
			l_result: PES_RESULT_ENTITY_LIST
			l_retried: BOOLEAN
		do
			if use_pes and not l_retried then
				create l_query.make (category_id, context.context_index, a_key, a_date, a_identifier.handle)
				l_result ?= pes_api.send_general_request (l_query).get
				check
					result_not_void: l_result /= Void
				end
				if l_result.remote_exception then
					raise (l_result.tag_name)
				end
				Result ?= l_result.entity_list
				check
					result_not_void: Result /= Void
				end
			else
				create Result.make_empty
				l_records := context.entity_records (a_key, Current, a_identifier)
				l_date := valid_date (a_date)
				if l_records.count > 0 then
					from
						l_records.start
					until
						l_records.off
					loop
						l_start_date := l_records.current_item.start_date
						l_end_date := l_records.current_item.end_date
						if a_date = Void or else (l_start_date <= l_date and then l_end_date >= l_date) then
							l_entity := entity (l_records.current_item.entity_id)
							if l_entity /= Void then
								Result.extend (l_entity)
							end
						end
						l_records.forth
					end
				end
			end
		rescue
			if not l_retried then
				l_retried := True
				retry
			end
		end

	entity_history_from_identifier (a_key: STRING; a_identifier: ENTITY_IDENTIFIER): ROSE_ENTITY_HISTORY [like ur_entity] is
		local
			l_end_date, l_start_date: ROSE_DATE
			l_records: ROSE_LINEAR_ARRAY [ENTITY_ATTRIBUTE_RECORD]
			l_interval: ENTITY_INTERVAL [like ur_entity]
		do
			create Result.make_empty
			l_records := context.entity_records (a_key, Current, a_identifier)
			if l_records /= Void and then l_records.count > 0 then
				from
					l_records.start
				until
					l_records.off
				loop
					l_start_date := l_records.current_item.start_date
					l_end_date := l_records.current_item.end_date
					create l_interval.make (entity (l_records.current_item.entity_id), l_start_date, l_end_date, 1)
					Result.extend (l_interval)
					l_records.forth
				end
			end
		end

feature -- Relationships

	active_relationships: ROSE_LINEAR_ARRAY [RELATIONSHIP_HANDLER [like ur_entity, ENTITY]] is
		do
			if active_relationships_saved = Void then
				active_relationships_saved := new_active_relationships
			end
			Result := active_relationships_saved
		ensure
			nonvoid : Result /= Void
		end

	active_corelationships: ROSE_LINEAR_ARRAY [RELATIONSHIP_HANDLER [ENTITY, like ur_entity]] is
		do
			if active_corelationships_saved = Void then
				active_corelationships_saved := new_active_corelationships
			end
			Result := active_corelationships_saved
		ensure
			nonvoid : Result /= Void
		end

feature -- Transformations

	transformation (a_entities: ARRAY [ENTITY]): ROSE_LINEAR_ARRAY [like ur_entity] is
			-- All versions of these entities from a_other_context
		local
			i: INTEGER
			l_entity: ENTITY
			l_array: ROSE_LINEAR_ARRAY [like ur_entity]
		do
			if a_entities /= Void and then a_entities.count > 0 then
				create Result.make_empty
				from i := 1 until i > a_entities.count loop
					l_entity := a_entities.item (i)
					if is_valid_same_type (l_entity) then
						l_array := all_peers (l_entity)
						if l_array /= Void then
							Result.append_array (l_array)
						end
					end
					i := i + 1
				end
			end
		ensure
			entities_non_void_implies_result_non_void : a_entities /= Void implies Result /= Void
		end

	transformation_at_date (a_entities: ARRAY [ENTITY]; a_date: ROSE_DATE): ROSE_LINEAR_ARRAY [like ur_entity] is
			-- All versions of these entities from a_other_context as of a date
		local
			i: INTEGER
			l_entity: ENTITY
			l_array: ROSE_LINEAR_ARRAY [like ur_entity]
		do
			if a_entities /= Void and then a_entities.count > 0 then
				create Result.make_empty
				from i := 1 until i > a_entities.count loop
					l_entity := a_entities.item (i)
					if is_valid_same_type (l_entity) then
						l_array := peers (l_entity, a_date)
						if l_array /= Void then
							Result.append_array (l_array)
						end
					end
					i := i + 1
				end
			end
		ensure
			result_not_void : (a_entities /= Void and then not a_entities.is_empty) implies Result /= Void
		end

feature -- Peers

	peer(a_entity : ENTITY;a_date : ROSE_DATE) : like ur_entity
			--Entity from this category handler which is mapped (by integration) to a_entity on a_date, if any.  Ur_entity if none, arbitrary witness if more than one.
		require
			valid_entity: a_entity /= Void
			a_date_not_void: a_date /= Void
			categories_match: a_entity.category_id = category_id
		do
			Result := peer_from_uid(a_entity.uid,a_date)
		end
	peer_from_uid (a_uid : INTEGER_64; a_date: ROSE_DATE): like ur_entity is
			--Entity from this category handler which is mapped (by integration) to the entity with uid a_uid on a_date, if any.  Ur_entity if none, arbitrary witness if more than one.
		require
			a_date_not_void: a_date /= Void
			categories_match: id_handler.uid_category_id (a_uid) = category_id
		local
			l_supplier_id,l_handle: INTEGER
			l_other_handler : ID_HANDLER
		do
			l_supplier_id := id_handler.uid_supplier_id (a_uid)
			l_other_handler := id_handler_from_id (l_supplier_id,category_id)
			l_handle := id_handler.uid_handle (a_uid)
			if l_supplier_id <= 0 or else l_handle <= 0 then
				Result := ur_entity
			else
				Result := entity(id_handler.peer_handle(a_uid,a_date))
				if Result = Void then
					Result := ur_entity
				end
			end
		ensure
			peer_not_void_but_may_be_ur_entity: Result /= Void
		end

	most_recent_peer (a_entity: ENTITY; a_date: ROSE_DATE): like ur_entity is
			-- Last peer of 'a_entity' in 'a_other_context' before or at 'a_date'
			-- Result represents same entity (same category) in 'a_other_context'.
			-- Void if no peer,
		require
			valid_entity: a_entity /= Void and then a_entity.handle /= 0
			valid_date: a_date /= Void
			categories_match: a_entity.category_id = category_id
		local
			l_mappings: ROSE_INTERVAL_HISTORY [INTEGER, ROSE_CLOSED_OPEN_INTERVAL [INTEGER]]
			l_interval: ROSE_CLOSED_OPEN_INTERVAL [INTEGER]
			i: INTEGER
		do
			l_mappings := peer_handler (a_entity.context).mappings_history (a_entity.handle, Current)
			if l_mappings /= Void then
				l_interval := l_mappings.most_recent (a_date)
				if l_interval /= Void then
					i := l_interval.item
					if i /= 0 then
						Result := entity (i)
					end
				end
			end
		end

	peers (a_entity: ENTITY; a_date: ROSE_DATE): ROSE_LINEAR_ARRAY [like ur_entity] is
			-- Peers of 'a_entity' in 'a_other_context' at 'a_date'
			-- Result represents same entities (same category) in 'a_other_context'.
			-- Void if no peers
		require
			valid_entity: a_entity /= Void and then a_entity.handle /= 0
			category_ids_match: a_entity.category_id = category_id
		local
			l_mappings: ROSE_INTERVAL_HISTORY [INTEGER, ROSE_CLOSED_OPEN_INTERVAL [INTEGER]]
			l_interval: ROSE_CLOSED_OPEN_INTERVAL [INTEGER]
		do
			l_mappings := peer_handler (a_entity.context).mappings_history (a_entity.handle, Current)
			if l_mappings /= Void then
				from
					create Result.make_empty
					l_mappings.start
				until
					l_mappings.off
				loop
					l_interval := l_mappings.current_item
					if a_date = Void or else l_interval.brackets (a_date) then
						Result.extend (entity (l_interval.item))
					end
					l_mappings.forth
				end
			end
		end

	all_peers (a_entity: ENTITY): ROSE_LINEAR_ARRAY [like ur_entity] is
			-- All Peers of 'a_entity' in 'a_other_context'
			-- Result represents same entities (same category) in 'a_other_context'.
			-- Void if no peers
		require
			valid_entity: a_entity /= Void and then a_entity.handle /= 0
			category_ids_match: a_entity.category_id = category_id
		local
			l_mappings: ROSE_INTERVAL_HISTORY [INTEGER, ROSE_CLOSED_OPEN_INTERVAL [INTEGER]]
			l_interval: ROSE_CLOSED_OPEN_INTERVAL [INTEGER]
		do
			l_mappings := peer_handler (a_entity.context).mappings_history (a_entity.handle, Current)
			if l_mappings /= Void then
				from
					create Result.make_empty
					l_mappings.start
				until
					l_mappings.off
				loop
					l_interval := l_mappings.current_item
					Result.extend (entity (l_interval.item))
					l_mappings.forth
				end
			end
		end

	peer_handler (a_context: SUPPLIER_CONTEXT): CATEGORY_HANDLER [SUPPLIER_CONTEXT, ENTITY] is
			-- Category handler for this id in a_context
		do
			Result := a_context.handler (category_id)
		end

	peers_history_uncached (a_entity: ENTITY): ROSE_ENTITY_HISTORY [like ur_entity] is
			-- Peers of 'a_entity' in 'a_other_context' with date intervals
			-- Result represents same entities (same category) in 'a_other_context'.
			-- Void if no peers
		require
			valid_entity: a_entity /= Void and then a_entity.handle /= 0
		local
			l_mappings: ROSE_INTERVAL_HISTORY [INTEGER, ROSE_CLOSED_OPEN_INTERVAL [INTEGER]]
			l_interval: ROSE_CLOSED_OPEN_INTERVAL [INTEGER]
		do
			l_mappings := peer_handler (a_entity.context).mappings_history_uncached (a_entity.handle, Current)
			if l_mappings /= Void then
				from
					create Result.make_empty
					l_mappings.start
				until
					l_mappings.off
				loop
					l_interval := l_mappings.current_item
					Result.extend (create {ENTITY_INTERVAL [like ur_entity]} .make (
						entity (l_interval.item),
						l_interval.start_date,
						l_interval.end_date,
						1)
					)
					l_mappings.forth
				end
			end
		end

	peers_history (a_entity: ENTITY): ROSE_ENTITY_HISTORY [like ur_entity] is
			-- Peers of 'a_entity' in 'a_other_context' with date intervals
			-- Result represents same entities (same category) in 'a_other_context'.
			-- Void if no peers
		require
			valid_entity: a_entity /= Void and then a_entity.handle /= 0
		local
			l_mappings: ROSE_INTERVAL_HISTORY [INTEGER, ROSE_CLOSED_OPEN_INTERVAL [INTEGER]]
			l_interval: ROSE_CLOSED_OPEN_INTERVAL [INTEGER]
		do
			l_mappings := peer_handler (a_entity.context).mappings_history (a_entity.handle, Current)
			if l_mappings /= Void then
				from
					create Result.make_empty
					l_mappings.start
				until
					l_mappings.off
				loop
					l_interval := l_mappings.current_item
					Result.extend (create {ENTITY_INTERVAL [like ur_entity]} .make (
						entity (l_interval.item),
						l_interval.start_date,
						l_interval.end_date,
						1)
					)
					l_mappings.forth
				end
			end
		end

	peers_history_interval_relaxed (a_entity: ENTITY): ROSE_ENTITY_HISTORY [like ur_entity] is
			-- This solves an extremely common problem of having start date being late in the mapping intervals.
			-- This routine will relax the start date of the first interval if no conflict is found in the reverse entity map.
		require
			valid_entity: a_entity /= Void and then a_entity.handle /= 0
		local
			l_history: ROSE_ENTITY_HISTORY [like ur_entity]
			l_interval: ENTITY_INTERVAL [like ur_entity]
			l_rose_history: ROSE_ENTITY_HISTORY [ENTITY]
			l_ignore_start_date: BOOLEAN
		do
			l_history := peers_history (a_entity)
			if l_history /= Void then
				from
					create Result.make_empty
					l_history.sort_by_start_date
					l_history.start
				until
					l_history.off
				loop
					l_ignore_start_date := False
					l_interval := l_history.current_item
					if l_interval /= void and then l_interval.entity /= void then
						if l_history.index = l_history.lower then
							l_rose_history := a_entity.context.company_accessor.peers_history (l_interval.entity)
							if l_rose_history.count < 2 then
								l_ignore_start_date := True
							end
						end
						if l_ignore_start_date then
							Result.extend (create {ENTITY_INTERVAL [like ur_entity]}.make (
								l_interval.entity,
								constants.minimum_date_time,
								l_interval.end_date,
								1)
							)
						else
							Result.extend (l_interval)
						end
					end
					l_history.forth
				end
			end
		end

	mappings_history_uncached (a_entity_id: INTEGER; a_other_handler: CATEGORY_HANDLER [SUPPLIER_CONTEXT, ENTITY]): ROSE_INTERVAL_HISTORY [INTEGER, ROSE_CLOSED_OPEN_INTERVAL [INTEGER]] is
		local
			l_rose_peers: ROSE_INTERVAL_HISTORY [INTEGER, ROSE_CLOSED_OPEN_INTERVAL [INTEGER]]
			i: INTEGER
		do
			if a_entity_id = 0 or else a_other_handler = Void then
				--| Do nothing
			elseif is_bound_to_rose and a_other_handler.is_bound_to_rose then
				create Result.make_empty
				Result.extend (create {ROSE_CLOSED_OPEN_INTERVAL [INTEGER]}.make (a_entity_id, constants.minimum_date_time, constants.maximum_date_time, 0.0))
			elseif is_bound_to_rose then
				create Result.make_empty
				id_handler.mapping_to_supplier_records (context.has_rose_bg_handles, category_id, a_other_handler.context.supplier_id, a_entity_id).apply (agent id_handler.add_mapping_uncached (Result, ?))
			elseif a_other_handler.is_bound_to_rose then
				create Result.make_empty
				id_handler.mapping_to_rose_records (a_other_handler.context.has_rose_bg_handles, category_id, context.supplier_id, a_entity_id).apply (agent id_handler.add_mapping_uncached (Result, ?))
			else
				create l_rose_peers.make_empty
				id_handler.mapping_to_rose_records (True, category_id, context.supplier_id, a_entity_id).apply (agent id_handler.add_mapping_uncached (l_rose_peers, ?))
				from
					create Result.make_empty
					i := l_rose_peers.lower
				until
					i > l_rose_peers.upper
				loop
					if l_rose_peers.item (i) /= Void then
						id_handler.mapping_to_supplier_records  (True, category_id, a_other_handler.context.supplier_id, l_rose_peers.item (i).item).apply (agent id_handler.add_mapping_uncached (Result, ?))
					end
					i := i + 1
				end
			end
		end


	mappings_history(a_entity_id: INTEGER; a_other_category_handler : CATEGORY_HANDLER[SUPPLIER_CONTEXT,ENTITY]): ROSE_INTERVAL_HISTORY [INTEGER, ROSE_CLOSED_OPEN_INTERVAL [INTEGER]]
			--entity intervals mapped to a_entity_id in a_other_category_handler
		require
			category_handler_exists : a_other_category_handler /= Void
		do
			Result := id_handler.mappings_history_from_id(a_entity_id,a_other_category_handler.category_id,a_other_category_handler.context.supplier_id)
		end
	id_handler : ID_HANDLER
			--  UID manager from context
		do
			Result := context.id_handler(category_id)
		ensure
			non_void : Result /= Void
		end
feature {SUPPLIER_CONTEXT} -- Contract support	

	same_category (a_id: ENTITY_ID_BASE): BOOLEAN is
			-- Is `a_id.category_id' equal to `category_id'?
		require
			a_id_not_void: a_id /= Void
		do
			Result :=  (a_id.category_id = category_id)
		ensure
			definition: Result = (a_id.category_id = category_id)
		end

feature -- Load from PES

	load_from_pes_result (a_result: PES_RESULT_LOAD_CATEGORY) is
			-- Load lookup caches from `a_result'.
		require
			a_result_not_void: a_result /= Void
			no_remote_exception: not a_result.remote_exception
		do
			if a_result.entity_id_table_saved /= Void then
				entity_id_table_saved := a_result.entity_id_table_saved
			end
			if attached {like table_saved} a_result.table_saved as l_table_saved then
					-- we can't actually use generics because there would be no way to specify
					-- the result type of the query within send_load_category_request, with enough specificity
					-- to avoid the check.
				table_saved := l_table_saved
			end
			if a_result.identifiers /= Void then
				identifiers := a_result.identifiers
			end
		ensure
			entity_id_table_saved_not_void: a_result.entity_id_table_saved /= Void implies
				entity_id_table_saved = a_result.entity_id_table_saved
			table_saved_not_void: a_result.table_saved /= Void implies table_saved /= Void
			identifiers_not_void: a_result.identifiers /= Void implies identifiers = a_result.identifiers
		end

feature {SUPPLIER_CONTEXT} -- Load PES

	load_entities_fast (a_entity_id_base: ROSE_LINEAR_ARRAY [ENTITY_ID_BASE]) is
			-- Put all entity_ids and entities from `a_entity_id_base' which correspond to `Current' into the hash tables.
			-- `a_i' should be the position in the array where the records which correspond to `Current' start.
			-- `a_entity_id_base' should be sorted in order to category_id.
		require
			valid_entity_id: a_entity_id_base /= Void
			no_void_entity_id: not a_entity_id_base.has (Void)
			correct_category: a_entity_id_base.for_all (agent same_category)
		local
			i, l_entity_count: INTEGER
			l_entity: like ur_entity
		do
			l_entity_count := context.count_entity_from_category (Current)
			table.accommodate (l_entity_count)
			entity_id_table.accommodate (l_entity_count)
			from
				i := a_entity_id_base.lower
				table_saved := Void
				entity_id_table_saved := Void
			until
				i > a_entity_id_base.upper
			loop
				entity_id_table.put (a_entity_id_base.item(i), a_entity_id_base.item(i).entity_id)
				create l_entity.make (context, category_id, a_entity_id_base.item(i).entity_id, a_entity_id_base.item(i).start_date, a_entity_id_base.item(i).end_date)
				table.put (l_entity, l_entity.handle)
				i := i + 1
			end

			all_entities_loaded := True
		ensure
			all_entities_loaded: all_entities_loaded
		end

	update_entity (a_record: ENTITY_ID_BASE) is
			-- update this entity in the cache
		local
			l_entity: like ur_entity
		do
			entity_id_table.force (a_record, a_record.entity_id)
			create l_entity.make (context, category_id, a_record.entity_id, a_record.start_date, a_record.end_date)
			table.force (l_entity, l_entity.handle)
		end

feature -- Basic operations

	 cache_attributes (a_identifiers: ROSE_LINEAR_ARRAY [ENTITY_IDENTIFIER]) is
			-- Preload attributes for the listed identifiers.
		do
			if a_identifiers = Void or else a_identifiers.count = 0 then
				cache_all_attributes
			else
				a_identifiers.apply (agent {ENTITY_IDENTIFIER}.load_key_history_records_by_category (category_id))
			end
		end


	cache_common_attributes is
			-- Cache only those attributes that will be needed when making the new entities.
			-- Called by ROSE_ATTRIBUTE_TABLE:add_item when a lot of entities will be made.
			-- Note that this could be changed to handle any ENTITY but will require changes to 35 classes.
		do
			cache_attributes(context.common_attributes(ur_entity))
		end

	cache_all_attributes is
			-- Preload attributes for all of the identifiers for this category.
		do
			from
				identifiers.start
			until
				identifiers.off
			loop
				identifiers.item_for_iteration.load_key_history_records_by_category (category_id)
				identifiers.forth
			end
		end

	clear_cache is
			-- clear cache
		do
			all_entities_retrieved := false
			all_entities_loaded := false
			if table_saved /= void then
				table_saved.clear_all
			end
			if entity_id_table_saved /= void then
				entity_id_table_saved.clear_all
			end

			active_relationships_saved := void

			all_entities_saved := void

			corelationship_ids_saved := void

			id_handler.clear_cache
			relationship_ids_saved := void
		end


	load_all_entities is
			-- load all entity_id s
		require
			context_not_void: context /= Void
		local
			l_records: ROSE_LINEAR_ARRAY [ENTITY_ID_BASE]
		do
			if not all_entities_loaded then
				l_records := context.all_entity_records (Current)
				l_records.do_all (agent add_entity_id)
				all_entities_loaded := True
			end
		ensure
			all_entities_loaded: all_entities_loaded
		end

	cache_entity (a_entity: ENTITY) is
			-- Add `a_entity' to the cache
			-- If it is of the correct type
		require
			entity_not_void: a_entity /= Void
		local
			l_entity: G
		do
			l_entity ?= a_entity
			if l_entity /= Void then
				table.put (l_entity, l_entity.handle)
			end
		ensure
			cached: a_entity.same_type (ur_entity) implies table.has (a_entity.handle)
		end

feature {CATEGORY_HANDLER} -- Implementation for peers


feature {NONE} -- Implementation

	pes_context_id: INTEGER is
			-- Context ID for PES
		do
			Result := context.context_index
		end

	all_entities_saved: ROSE_LINEAR_ARRAY [like ur_entity]			-- See 'all_entities'.

	entity_id_table_saved: HASH_TABLE [ENTITY_ID_BASE, INTEGER]
			-- CAche for `entity_id_table'

	entity_id_table: like entity_id_table_saved is
		do
			if entity_id_table_saved = Void then
				create entity_id_table_saved.make (1000)
			end
			Result := entity_id_table_saved
		ensure
			result_not_void : Result /= Void
		end

	table: like table_saved is
		do
			if table_saved = Void then
				create table_saved.make (500)
			end
			Result := table_saved
		ensure
			result_not_void : Result /= Void
		end

	table_saved: HASH_TABLE [like ur_entity, INTEGER]

	add_entity (a_record: ENTITY_ID_BASE) is
			--add the entity for this record if not already there
		local
			l_entity: like ur_entity
		do
			if not table.has (a_record.entity_id) then
				add_entity_id (a_record.twin)	-- Is this 'twin' really necessary?
				create l_entity.make (context, category_id, a_record.entity_id, a_record.start_date, a_record.end_date)
				table.put (l_entity, l_entity.handle)
			end
		end

	add_entity_id (a_record: ENTITY_ID_BASE) is
			--add entity_id
		do
			if not entity_id_table.has (a_record.entity_id) then
				-- Here, we used to 'twin' 'a_record'.  This produced a garbage collection from hell
				-- when called from 'load_all_entities' for listings in rose_entity_background.  There
				-- are over 600,000 listings there and since we didn't keep the passed records
				-- (we 'twin'ed them) we produced 600,000 pieces of garbage.  Thus begins a tale that
				-- is much too long to relate here.  If you really want to 'twin' the record
				-- then please do so in the caller.  RJ
				entity_id_table.put (a_record, a_record.entity_id)
			end
		end

	all_entities_retrieved: BOOLEAN
		-- Indicates whether `all_entities_saved' has been loaded

	all_entities_loaded: BOOLEAN
		-- Indicates whether `entity_id_table_saved' has been loaded

	new_active_relationships: like active_relationships is
		local
			l_relationship: RELATIONSHIP_HANDLER [like ur_entity, ENTITY]
			i: INTEGER
		do
			create Result.make_empty
			from
				i := 1
			until
				i > relationship_ids.count
			loop
				l_relationship ?= context.relationship (relationship_ids.item (i))
				if l_relationship /= Void then
					Result.extend (l_relationship)
				end
				i := i + 1
			end
		end

	new_active_corelationships: like active_corelationships is
		local
			l_relationship: RELATIONSHIP_HANDLER [ENTITY, like ur_entity]
			i: INTEGER
		do
			create Result.make_empty
			from i := 1 until i > corelationship_ids.count loop
				l_relationship ?= context.relationship (corelationship_ids.item (i))
				if l_relationship /= Void then
					Result.extend (l_relationship)
				end
				i := i + 1
			end
		end

	relationship_ids: ROSE_LINEAR_ARRAY [INTEGER] is
		local
			l_handlers: ROSE_LINEAR_ARRAY [RELATIONSHIP_HANDLER [ENTITY, ENTITY]]
		do
			if relationship_ids_saved = Void then
				create relationship_ids_saved.make_empty
				l_handlers := context.relationships
				from l_handlers.start until l_handlers.after loop
					if l_handlers.current_item /= Void and l_handlers.current_item.category_1_id = category_id then
						relationship_ids_saved.extend (l_handlers.current_item.relationship_id)
					end
					l_handlers.forth
				end
			end
			Result := relationship_ids_saved
		end

	corelationship_ids: ROSE_LINEAR_ARRAY [INTEGER] is
		local
			l_handlers: ROSE_LINEAR_ARRAY [RELATIONSHIP_HANDLER [ENTITY, ENTITY]]
		do
			if corelationship_ids_saved = Void then
				create corelationship_ids_saved.make_empty
				l_handlers := context.relationships
				from l_handlers.start until l_handlers.after loop
					if l_handlers.current_item /= Void and then l_handlers.current_item.category_2_id = category_id then
						corelationship_ids_saved.extend (l_handlers.current_item.relationship_id)
					end
					l_handlers.forth
				end
			end
			Result := corelationship_ids_saved
		end

	relationship_ids_saved: ROSE_LINEAR_ARRAY [INTEGER]
	corelationship_ids_saved: ROSE_LINEAR_ARRAY [INTEGER]
	active_relationships_saved: ROSE_LINEAR_ARRAY [RELATIONSHIP_HANDLER [like ur_entity, ENTITY]]
	active_corelationships_saved: ROSE_LINEAR_ARRAY [RELATIONSHIP_HANDLER [ENTITY, like ur_entity]]

	valid_date (a_date: ROSE_DATE): ROSE_DATE is
			-- Returns a today is a date is void
		do
			if a_date /= Void then
				Result := a_date
			else
				create Result.make_today
			end
		end

	is_valid_same_type (a_entity: ENTITY): BOOLEAN is
			-- Is `a_entity' valid and same category as `Current'?
		do
			Result := a_entity /= Void and then a_entity.is_valid and then a_entity.category_id = category_id
		end


feature	{PES_QUERY_LOAD_CATEGORY} -- PES Result

	get_table_saved: like table_saved is
			-- export table_saved
		do
			Result := table_saved
		end

	get_entity_id_table_saved: like entity_id_table_saved is
			-- export entity_id_table_saved
		do
			Result := entity_id_table_saved
		end
invariant
	context_exists : context /= Void
end
