indexing
	description: "[
				  This class is the server component for the clustered tagged item database.
				  It is a descendent of the signature tagged server and has implementation for clustered databases.
				  ]"
	author: "Randy John"
	last_editor: "$Author: rj $"
	date: "$Date: 2014/09/11 $"
	revision: "$Revision: #72 $"
	path: "$File: //rose/source/kernel/persistence/databases/signature_tagged/clustered_tagged_server.e $"

class
	CLUSTERED_TAGGED_SERVER[G -> COMPARABLE,
						    H -> COMPARABLE,
						    reference I -> CLUSTERED_SIGNATURE[G, H, ENTITY] create default_create end]

inherit
	SIGNATURE_TAGGED_SERVER[I]
		redefine
			path,
			set_path,
			signature_id,
			select_signature_ids,
			signature_selection,
			signature_insert,
			signature_insert_with_signature_data_process_id,
			virtual_record_set,
			set_signature_item_server,
			paths_touched_since,
			set_ad_hoc_query,
			set_cache_from_results,
			submit_signature_touches
		end
	MEMORY
		export
			{NONE} all
		end

create
	make

feature {SIGNATURE_TAGGED_ITEM_DATABASE} -- Implementation

	set_signature_item_server (a_server: SQL_DATA_SERVER) is
			-- <Precursor>
		do
			Precursor (a_server)
			virtual_record_set.set_path_tree (path_tree)
		end

	set_ad_hoc_query (a_value: BOOLEAN) is
			-- If 'a_value' is True, queries will process only at the record level and will not retrieve by cluster
		do
			ad_hoc_query := a_value
			virtual_record_set.set_ad_hoc_query (a_value)
		end

	submit_signature_touches (a_signature_touches: ROSE_LINEAR_ARRAY[I]) is
			-- Update the signature record with the fact that a data item for that record has been touched.
		do
			if a_signature_touches /= Void and then not a_signature_touches.is_empty then
				Precursor (a_signature_touches)
			end
		end

	set_cache_from_results (a_virtual_record: ROSE_VIRTUAL_RECORD[I]; a_client: SIGNATURE_TAGGED_ITEM_DATABASE[I]) is
			-- <Precursor>
		do
			if a_virtual_record = Void and then query_signatures /= Void and not query_signatures.is_empty then
				-- Use the query signature to help find the correct path to store the empty result set
				Precursor (create {ROSE_VIRTUAL_RECORD[I]}.make (query_signatures.first),a_client)
			else
				Precursor (a_virtual_record,a_client)
			end
		end

feature {CLUSTERED_TAGGED_ITEM_DATABASE} -- Implementation

	set_query_signatures (a_signatures: ROSE_LINEAR_ARRAY[I]) is
			-- Set query_signatures to 'a_signatures'
		do
			query_signatures := a_signatures
		end

	latest_path_timestamp (a_path: INTEGER): ROSE_DATE is
			-- When was the last time any record in the 'a_path' cluster was updated
		local
			l_retry: BOOLEAN
		do
			if not l_retry then
				if latest_path_timestamp_predicate = Void then
					create latest_path_timestamp_predicate.make (signature_item_server,"select max (timestamp) as timestamp from " +
																 data_name_prefix + "data_signatures (readuncommitted) where path = ?")
					if not latest_path_timestamp_predicate.success then
						raise ("failed to create latest_path_timestamp_predicate")
					end

					create latest_versions_timestamp_predicate.make (signature_item_server,"select max (a.timestamp) as timestamp from " +
																	 data_name_prefix + "data_signature_versions a (readuncommitted)," +
																	 data_name_prefix + "data_signatures b WITH (INDEX (path_clustered_index) readuncommitted) " +
																	 "where path = ? and a.signature_id = b.signature_id option (loop join)")
					if not latest_versions_timestamp_predicate.success then
						raise ("failed to create latest_versions_timestamp_predicate")
					end
				end

				latest_path_timestamp_predicate.retrieve (Void,<<a_path>>)
				if not latest_path_timestamp_predicate.success then
					raise ("failed to retrieve latest_path_timestamp_predicate for path " + a_path.out)
				end
				if latest_path_timestamp_predicate.array.first /= Void then
					Result := latest_path_timestamp_predicate.array.first
				end

				latest_versions_timestamp_predicate.retrieve (Void,<<a_path>>)
				if not latest_versions_timestamp_predicate.success then
					raise ("failed to retrieve latest_versions_timestamp_predicate for path " + a_path.out)
				end
				if latest_versions_timestamp_predicate.array.first /= Void then
					if Result /= Void then
						Result := Result.max (latest_versions_timestamp_predicate.array.first)
					else
						Result := latest_versions_timestamp_predicate.array.first
					end
				end

				if Result = Void then
					create Result.make_min_date
				end
			end
		rescue
			output (Global.interpreted ("($now) Exception in latest_path_timestamp called with $1%N"),<<a_path>>)
			output(exception_trace + "%N", Void)
			l_retry := True
			latest_path_timestamp_predicate := Void
			latest_versions_timestamp_predicate := Void
			create Result.make_max_date			-- This will force the caller to retrieve from the database.
			retry
		end

	paths_touched_since (a_process_id: INTEGER): ROSE_LINEAR_ARRAY[INTEGER] is
			-- Collection of path indexes that were updated since 'a_process_id'
		do
			if saved_paths_touched_since_predicate = Void then
				create saved_paths_touched_since_predicate.make_strict (signature_item_server,"select distinct path from " + data_name_prefix +
																		"data_signatures (readuncommitted) where last_process_id > ?")
			end
			saved_paths_touched_since_predicate.retrieve (Void,<<a_process_id>>)
			Result := saved_paths_touched_since_predicate.array
		end

	paths_touched_since_time (a_time_ut: ROSE_DATE): ROSE_LINEAR_ARRAY[INTEGER] is
			-- Collection of path indexes that were updated since 'a_time_ut'.
		require
			meaningful_time: a_time_ut /= Void
		local
			l_statement: STRING
			l_paths_touched_since_time_predicate: INTEGER_PREDICATE
		do
			-- The following statement could be prepared once and have arguments fed to it at run-time.  I
			-- tried that but the execution plan was lousy and the query took 6 seconds.  This takes no time at all.
			l_statement := Global.formatted
							(
								"[
								select path
								from
									$1 (readuncommitted)
								where
									timestamp>$3
								group by path
								union
								select path
								from
									$1 a (readuncommitted),
									$2 b (readuncommitted)
								where
									a.signature_id=b.signature_id and
									b.timestamp>$3
								group by path
								]",
								<<
									data_name_prefix + "data_signatures",
									data_name_prefix + "data_signature_versions",
									"'" + a_time_ut.debug_output + "'"
								>>
							)
			create l_paths_touched_since_time_predicate.make_strict(signature_item_server, l_statement)
			l_paths_touched_since_time_predicate.retrieve (Void, Void)
			Result := l_paths_touched_since_time_predicate.array
		end

feature {NONE} -- Path timing query predicates

	latest_path_timestamp_predicate: DATE_PREDICATE
	latest_versions_timestamp_predicate: DATE_PREDICATE
	saved_paths_touched_since_predicate: INTEGER_PREDICATE
	saved_paths_touched_since_time_predicate: INTEGER_PREDICATE

feature {CLUSTERED_TAGGED_ITEM_DATABASE} -- Implementation

	path (a_signature: I): INTEGER is
			-- The path cluster id for a_signature
		do
			Result := path_leaf (a_signature).path
		end

	path_leaf (a_signature: I): PARTITION_TREE_RECORD[G, H] is
			-- The leaf node containing a_signature
		do
			if not ad_hoc_query and then a_signature /= Void then
				Result := path_tree.leaf_record (a_signature.key_0, a_signature.key_1)
			else
				create Result
			end
		end

	set_path (a_record: ANY; a_signature: I) is
			-- Set the path of a_record to the path for a_signature
		local
			l_record: DATA_RECORD[ANY]
			l_last_leaf: PARTITION_TREE_RECORD[G, H]
		do
			l_last_leaf := path_tree.leaf_record (a_signature.key_0, a_signature.key_1)
			if a_record /= Void then
				l_record ?= a_record
				if l_record /= Void then
					l_record.set_path (l_last_leaf.path)
				end
			end
			if a_signature /= Void then
				a_signature.set_path (l_last_leaf.path)
			end
		end

	signature_id (a_signature: I; a_can_read_old_signature, a_can_write_new_signature: BOOLEAN): INTEGER
			-- The signature id for the unique key values in a_signature
			-- Create a signature id if one doesn't already exist
			-- See SIGNATURE_TAGGED_SERVER.signature_id for documentation about the 'a_can_*' arguments.
		local
			l_integer_ref: INTEGER_REF
		do
			if a_can_read_old_signature
			then
				signature_search.retrieve (a_signature,Void)
				if not signature_search.success then
					raise ("signature_search failed.")
				end
			else
				signature_search.set_container(empty_array)
			end

			if signature_search.array.count = 1 then
				l_integer_ref ?= signature_search.array.item (1).item (1)
				if l_integer_ref.item /= 0 then
					Result := l_integer_ref.item
					last_signature_created := False
				end
			elseif signature_search.array.count = 0 then
				a_signature.set_path (path_tree.leaf_record (a_signature.key_0, a_signature.key_1).path)
				-- An insert into the data_signature table follows.  In the past a common error was
				-- to forget to set the identity_id.

				if a_can_write_new_signature
				then
					if table_structure.signature_data_process_id_available then
						signature_insert_with_signature_data_process_id.retrieve (a_signature,<<process_id,process_id,process_id,process_id,a_signature.path>>)
						if not signature_insert_with_signature_data_process_id.success then
							raise ("Predicate failed - signature_insert_with_signature_data_process_id")
						end
						a_signature.set_signature_data_process_id (process_id)
						Result := signature_insert_with_signature_data_process_id.array[1]
					else
						signature_insert.retrieve (a_signature, <<process_id, process_id, process_id, a_signature.path>>)
						if not signature_insert.success then
							raise ("Predicate failed - signature_insert")
						end
						Result := signature_insert.array[1]
					end

					last_signature_created := True
					signatures_created := signatures_created + 1
				end
			else
				raise ("duplicate signature, check " + data_name_prefix + "data_signatures")
			end
		end

	path_tree_cache_key: STRING is
			-- String key used to help cache path_tree retrieval
		do
			Result := data_name_prefix + signature_item_server.node + signature_item_server.database
		ensure
			valid_result: Result /= Void
		end

	path_tree: PARTITION_TREE[G,H] is
			-- Precomputed binary tree used to find cluster id for a signature retrieved from DB.
		local
			l_object_predicate: DATA_SERVER_STREAM_COMMAND
			l_storable_predicate: EXECUTION_PREDICATE
			l_stream: ROSE_STREAM
			l_integer_predicate: INTEGER_PREDICATE
			l_object_size: INTEGER
			l_data_server: SQL_DATA_SERVER
			l_retry: BOOLEAN
			l_storable: BOOLEAN
			l_collecting: BOOLEAN
			--			l_dump_file: ROSE_FILE
			l_eiffel_version: STRING
			l_factory: PARTITION_TREE_FACTORY [G, H]
		do
			if path_tree_saved = Void then
				path_tree_cache.search (path_tree_cache_key)
				if path_tree_cache.found then
					path_tree_saved ?= path_tree_cache.found_item
				end
			end
			if path_tree_saved = Void then
				l_collecting := collecting
				collection_off
				-- FIX ME:
				-- This is ugly but until we can store objects knowing what version was used we are only going to
				-- allow these particular applications to get the header from the database.
				if not l_retry and
					(
					 False
					 --l_env.command_line.argument (0).has_substring ("data_platform") or
					 --exenv.command_line.argument (0).as_lower.has_substring ("served_market_data")
					 -- per Mauricio, 8/19/2010: -- or exenv.command_line.argument (0).as_lower.has_substring ("ptrade_")--
					 -- or l_env.command_line.argument (0).has_substring ("vms_data_replication")) then
					)
				then
					l_storable := True
					l_eiffel_version := Global.app.info.build_estudio
					--print (Global.app.info.out)
					if l_eiffel_version /= Void then
						create l_integer_predicate.make_strict (signature_item_server,
							"select max (datalength (storable)) as leng from " + data_name_prefix + "path_tree_object_with_version where eiffel_version=?")
						l_integer_predicate.retrieve (Void, <<l_eiffel_version>>)
						if l_integer_predicate.success then
							l_object_size := l_integer_predicate.array.first.max (0)
							l_object_predicate := signature_item_server.stream_command (
							"set textsize " + l_object_size.out + "%Nselect storable from " + data_name_prefix + "path_tree_object_with_version where eiffel_version='" + l_eiffel_version + "'")
							l_object_predicate.retrieve (Void)
						end
					else
						warn ("path_tree","WARNING!  I can't figure out what version of Eiffel was used.%N",Void)
						warn ("path_tree","I'll have to use 'path_tree_object' instead of 'path_tree_object_with_version'.%N",Void)
						create l_integer_predicate.make (signature_item_server,
							"select max (datalength (storable)) as leng from " + data_name_prefix + "path_tree_object")
						l_integer_predicate.retrieve (Void,Void)
						if l_integer_predicate.success then
							l_object_size := l_integer_predicate.array.first.max (0)
							l_object_predicate := signature_item_server.stream_command (
							"set textsize " + l_object_size.out + "%Nselect storable from " + data_name_prefix + "path_tree_object")
							l_object_predicate.retrieve (Void)
						end
					end
				end
				if not l_storable or l_object_predicate.query_result.count = 0 or l_retry then
					create l_factory.make (table_structure)
					path_tree_saved := l_factory.retrieved_path_tree
					if l_storable then
						inform ("path_tree", "Saving the $1path_tree storable to the database.%N",<<data_name_prefix>>)
						l_data_server := servers.new_server_from_repository (signature_item_server.logical)
						if l_eiffel_version /= Void then
							l_data_server.sql_execute_immediate ("delete from " + data_name_prefix + "path_tree_object_with_version where eiffel_version='" + l_eiffel_version + "'")
							create l_storable_predicate.make (l_data_server,"insert into " + data_name_prefix + "path_tree_object_with_version (eiffel_version, storable) values ('" + l_eiffel_version + "', ?)")
						else
							l_data_server.sql_execute_immediate ("delete from " + data_name_prefix + "path_tree_object")
							create l_storable_predicate.make (l_data_server,"insert into " + data_name_prefix + "path_tree_object (storable) values (?)")
						end
						create l_stream
						l_stream.make_with_size (40_000_000)
						l_stream.independent_store (path_tree_saved)
						-- This fixes a bug in 5.4.0907.
						-- 'buffer_size' is bigger and object_buffer_size but if that's what they wanted last time...
						--l_object_size := l_stream.buffer_size
						--create l_stream
						--l_stream.make_with_size (l_object_size)
						--l_stream.independent_store (path_tree_saved)
						-- For testing only.
						--						path_tree_saved ?= l_stream.retrieved
						--						create l_dump_file.make_open_write ("k.dat")
						--						l_dump_file.put_string (l_stream.string_representation)
						--						l_dump_file.independent_store (path_tree_saved)
						--						l_dump_file.close
						l_storable_predicate.execute (Void, <<l_stream>>)
						l_data_server.commit
						--print ("Done.%N")
					end
				else
					if save_path_tree_as_string then
						path_tree_as_string := l_object_predicate.query_result
					end
					-- For testing only.
					--					create l_dump_file.make_open_read ("k.dat")
					--					path_tree_saved ?=  l_dump_file.retrieved
					--					l_dump_file.close
					--print ("Attempting to decode the path_tree storable.%N")
					create l_stream
					path_tree_saved ?= l_stream.retrieve_from_string (l_object_predicate.query_result)
					--print ("Success.%N")
				end
				virtual_record_set.set_path_tree (path_tree_saved)
				path_tree_cache.force (path_tree_saved, path_tree_cache_key)
				if l_collecting then
					collection_on
				end
			end
			Result := path_tree_saved
			--/-
		rescue
			if l_collecting	then
				collection_on
			else
				collection_off
			end
			if not l_retry then
				print ("CLUSTERED_TAGGED_SERVER:path_tree has suffered a failure.  Will retry once.%N")
				l_retry := True
				retry
			end
		end

	order_by_path (a, b: I): BOOLEAN is
			-- Sorting agent for signature selection
		do
			Result := path (a) < path (b)
		end

	reset_path_tree is
			-- Set the path_tree to Void in order to force it to be refreshed from source
		do
			path_tree_saved := Void
			path_tree_cache.remove (data_name_prefix + signature_item_server.logical)
		end

	data_axis_max_bounds_predicate (a_axis: INTEGER; a_range: INTEGER): COLLECTING_PREDICATE[PARTITION_TREE_RECORD[G,H]] is
			-- A predicate to retrieve the bounds of maximum 'a_axis' going back 'a_range'
		require
			valid_axis: a_axis = 0 or a_axis = 1
			meaningful_range: a_range > 0
		local
			l_signature: I
		do
			create l_signature
			if a_axis = 0 then
				create Result.make_strict (signature_item_server, "select " +
					"max (" + l_signature.key_0_name + ") as max_" + a_axis.out + ", " +
					"max (" + l_signature.key_0_name + ") - " + a_range.out + " as min_" + a_axis.out +
					" from " + data_name_prefix + "data_signatures")
			else
				create Result.make_strict (signature_item_server, "select " +
					"max (" + l_signature.key_1_name + ") as max_" + a_axis.out + ", " +
					"max (" + l_signature.key_1_name + ") - " + a_range.out + " as min_" + a_axis.out +
					" from " + data_name_prefix + "data_signatures")
			end
		end

	data_axis_min_bounds_predicate (a_axis: INTEGER; a_range: INTEGER): COLLECTING_PREDICATE[PARTITION_TREE_RECORD[G,H]] is
			-- A predicate to retrieve the bounds of minumum 'a_axis' going forward 'a_range'
		require
			valid_axis: a_axis = 0 or a_axis = 1
			meaningful_range: a_range > 0
		local
			l_signature: I
		do
			create l_signature
			if a_axis = 0 then
				create Result.make_strict (signature_item_server, "select " +
					"min (" + l_signature.key_0_name + ") as min_" + a_axis.out + ", " +
					"min (" + l_signature.key_0_name + ") + " + a_range.out + " as max_" + a_axis.out +
					" from " + data_name_prefix + "data_signatures")
			else
				create Result.make_strict (signature_item_server, "select " +
					"min (" + l_signature.key_1_name + ") as min_" + a_axis.out + ", " +
					"min (" + l_signature.key_1_name + ") + " + a_range.out + " as max_" + a_axis.out +
					" from " + data_name_prefix + "data_signatures")
			end
		end

	set_save_path_tree_as_string is
			-- set 'save_path_tree_as_string' to True
		do
			save_path_tree_as_string := True
		end

	clear_path_tree_as_string
			-- Removes saved versions of the path tree.  Potentially large memory savings.
		do
			path_tree_as_string := Void
			save_path_tree_as_string := False
		end

	path_tree_saved: PARTITION_TREE[G,H]
			-- The path tree object
	path_tree_as_string: STRING
			-- The serialized path tree object as a string

	virtual_record_set: CLUSTERED_TAGGED_VIRTUAL_RECORD_SET[G,H,I]

feature {PERSISTENT_CLUSTERED_TAGGED_SERVER_COMMAND} -- Implementation

	query_signatures: ROSE_LINEAR_ARRAY[I]
			-- Collection of signatures used in the last query

feature {NONE} -- File Based Cacheing

	path_cacheing_enabled: BOOLEAN is
			-- is path cacheing enabled?
		local
			l_predicate: ANONYMOUS_PREDICATE
		do
			if path_cacheing_enabled_saved = Void then
				if (exenv.command_line.argument (0).has_substring ("monthly_data_pilot") or
					exenv.command_line.argument (0).has_substring ("fundamental_variable_server")) then
					create l_predicate.make_using_server (signature_item_server, "exec sp_columns @table_name = '" + data_name_prefix + "path_cache'")
					l_predicate.retrieve (Void, Void)
					if l_predicate.array.count > 0 then
						create path_cacheing_enabled_saved.put (True)
					else
						create path_cacheing_enabled_saved.put (False)
					end
				else
					create path_cacheing_enabled_saved.put (False)
				end
			end
			Result := path_cacheing_enabled_saved.item
		end

	path_cacheing_enabled_saved: CELL[BOOLEAN]

feature {NONE} -- Implementation

	save_path_tree_as_string: BOOLEAN
			-- Should we persiste the path tree in the database?

	simple_signature_selection: SIMPLE_PREDICATE[I] is
			-- Predicate to select signatures by path
		do
			if simple_signature_selection_saved = Void then
				create simple_signature_selection_saved.make (signature_item_server,
				"select * from " + data_name_prefix + "data_signatures " +
					"where path = ? order by signature_id")
				if not simple_signature_selection_saved.success then
					raise ("Failed to create simple_signature_selection")
				end
			end
			Result := simple_signature_selection_saved
		end
	simple_signature_selection_saved: SIMPLE_PREDICATE[I]

	signature_insert: INTEGER_PREDICATE
			-- Insert into the data signatures table
		local
			l_statement: STRING
		do
			if signature_insert_saved = Void then
				l_statement := new_signature_insert_statement(False, True)
				create signature_insert_saved.make(signature_item_server, l_statement)
				if not signature_insert_saved.success then
					raise ("Failed to create signature_insert_saved")
				end
			end
			Result := signature_insert_saved
		end

	signature_insert_with_signature_data_process_id: INTEGER_PREDICATE
			-- Insert into the data signatures table when the signature may contain data
		local
			l_statement: STRING
		do
			if signature_insert_with_signature_data_process_id_saved = Void then
				l_statement := new_signature_insert_statement(True, True)
				create signature_insert_with_signature_data_process_id_saved.make(signature_item_server, l_statement)
				if not signature_insert_with_signature_data_process_id_saved.success then
					raise ("Failed to create signature_insert_with_signature_data_process_id_saved")
				end
			end
			Result := signature_insert_with_signature_data_process_id_saved
		end

	signature_selection: ANONYMOUS_PREDICATE is
			-- Predicate to retrieve the signatures by path
		local
			l_statement: STRING
			l_signature: I
		do
			if ad_hoc_query then
				Result := Precursor
			else
				create l_signature.default_create
				l_statement := "select "
				if l_signature.signatures_fields_clause.count > 0 then
					l_statement := l_statement + l_signature.signature_and_data_fields_clause + ","
				end
				if table_structure.signature_data_process_id_available then
					l_statement := l_statement + data_name_prefix + "data_signatures.signature_data_process_id,"
				end
				l_statement := l_statement +
					data_name_prefix + "data_signatures.signature_id," +
					data_name_prefix + "data_signatures.last_process_id," +
					data_name_prefix + "data_signatures.first_process_id, " +
					data_name_prefix + "data_signatures.continuing_process_id " +
					"from " + data_name_prefix + "data_signatures " +
					"where path = ? order by signature_id"
				Result := signature_selection_predicate (l_statement, <<"INTEGER">>)
			end
		end

	last_leaf: PARTITION_TREE_RECORD[G,H]
			-- The last leaf retrieved from the path tree, used for query optimization

feature {SIGNATURE_TAGGED_ITEM_DATABASE} -- Implementation

	select_signature_ids (a_columns: COLUMN_DESCRIPTORS[I]) is
			-- Set signature_ids to be retrieved
		local
			l_last_path: INTEGER
		do
			if ad_hoc_query then
				Precursor (a_columns)
			else
				if query_signatures.count = 1 then
					l_last_path := path (query_signatures.first)
					simple_signature_selection.set_container (cube_signature_result)
					simple_signature_selection.retrieve (Void,<<l_last_path>>)
					if not simple_signature_selection.success then
						raise ("simple_signature_selection 1 failed.")
					end
					if sig_tagged_trace_requested then
						sig_tagged_trace_file.put_string (database_name + ":" + table_name + ":CLUSTERED_TAGGED_SERVER:select_signature_ids:1:path=" + l_last_path.out + "%N")
					end
				else
					query_signatures.sort_by (agent order_by_path)
					from query_signatures.start until query_signatures.off loop
						if path (query_signatures.current_item) /= l_last_path then
							--set_results_from_cache (query_signatures.current_item)
							simple_signature_selection.set_container (cube_signature_result)
							simple_signature_selection.retrieve (Void,<<path (query_signatures.first)>>)
							if not simple_signature_selection.success then
								raise ("simple_signature_selection 2 failed.")
							end
							l_last_path := path (query_signatures.current_item)
						end
						query_signatures.forth
					end
				end
			end
		end

feature {NONE} -- Helpers

	path_tree_cache: HASH_TABLE[ANY,STRING] is
			-- Global cache of path_tree objects keyed by data_name_prefix + data_server.logical
		once
			create Result.make (10)
		end

	exenv: EXECUTION_ENVIRONMENT is
			-- Access to the execution environment for program args
		once
			create Result
		end

	field_descriptors: FIELD_DESCRIPTORS is
			-- Facility to inspect object attribute values
		once
			create Result
		end

end
