note

	description:

		"Objects that provide a dialog for selecting color-sequences or ROSE_PALETTE specified by ColorBrewer"

	author: "Colin Adams"
	last_editor: "$Author: c4 $"
	documentation: "http://colorbrewer2.org/"
	date: "$Date: 2013/06/12 $"
	revision: "$Revision: #1 $"
	path: "$File: //rose/source/kernel/infrastructure/widgets/charts/color_brewer_widget.e $"

class COLOR_BREWER_WIDGET

inherit

	COLOR_BREWER

	ROSE_SHARED_GLOBAL_GUI
		export {NONE} all end

feature {NONE} -- Initialization

	initialize_dialog is
			-- Initialize `dialog', if non-Void.
		local
			l_main_box, l_box: EV_BOX
			l_widget: EV_WIDGET
		do
			if dialog = Void then
				color_count := 12
				selected_nature := Qualitative_data
				dialog := gui.factory.new_dialog ("Palette selector", "palette.ico", 0, 0)
				l_main_box := gui.factory.new_vbox (dialog, -4, 2)
				l_box := gui.factory.new_hbox (l_main_box, 0, 2)
				nature_selector := gui.factory.new_combo (l_box, "Nature of data: ",
					<<sequential_data_label, diverging_data_label, qualitative_data_label>>,
					qualitative_data_label, agent set_nature, 100)
				nature_selector.return_actions.extend (agent set_nature)
				size_selector := gui.factory.new_combo (l_box, "Size: ", <<3, 4, 5, 6, 7, 8, 9, 10, 11, 12>>, color_count, agent set_size, 50)
				size_selector.return_actions.extend (agent set_size)
				radio_box := gui.factory.new_vbox (l_box, -4, 2)
				colors_box := gui.factory.new_hbox (l_main_box, 0, -2)
				fill_radio_box
				l_box := gui.factory.new_hbox (l_main_box, 0, 2)
				l_widget := gui.factory.new_label (l_box, Void, 0)
				l_widget := gui.factory.new_button (l_box, "OK", agent notify_on_ok, 60)
				l_widget := gui.factory.new_button (l_box, "Cancel", agent notify_on_cancel, 60)
			end
		ensure
			dialog_not_void: dialog /= Void
		end

feature -- Access

	color_count: NATURAL_8
			-- Number of colors in `selected_sequence'

	selected_sequence: ARRAY [EV_COLOR]
			-- Color-sequence selected by user (or defaulted)
	
	selected_nature: NATURAL_8
			-- Nature of data, according to user

	palette: ROSE_PALETTE is
			-- Palette from `selected_sequence'
		require
			selected_sequence_not_void: selected_sequence /= Void
		do
			create Result.make_from_array (selected_sequence, True, True)
		ensure
			palette_not_void: palette /= Void
		end
	
feature -- Basic operations

	show_modal (a_widget: EV_WIDGET) is
			-- Show `dialog' relative to window containing `a_widget'.
		require
			a_widget_not_void: a_widget /= Void
		local
			l_window: EV_WINDOW
		do
			l_window := gui.parent_window (a_widget)
			if l_window /= Void then
				initialize_dialog
				dialog.show_modal_to_window (l_window)
			end
		end

feature {NONE} -- Widget
	
	dialog: EV_DIALOG
			-- Dialog holding the palette chooser

	nature_selector: ROSE_COMBO_BOX
			-- Selector for nature of data to be presented
	
	size_selector: ROSE_COMBO_BOX
			-- Selector for number of colors to use
	
	radio_box: EV_VERTICAL_BOX
			-- Holder for radio buttons
	
	colors_box: EV_BOX
		-- Holder displaying colors of the palette
		
	sequential_data_label: STRING is "sequential"
			-- Value shown in `nature_selector'
	
	diverging_data_label: STRING is "diverging"
			-- Value shown in `nature_selector'
	
	qualitative_data_label: STRING is "qualitiative"
			-- Value shown in `nature_selector'
	
	set_default is
			-- Set default value of `selected_sequence' according to `selected_nature'.
		do
			inspect
				selected_nature
			when Sequential_data then
				selected_sequence := greens_9
			when Diverging_data then
				selected_sequence := spectral_11
			when Qualitative_data then
				selected_sequence := set3_12
			end
		end

	fill_radio_box is
			-- Fill `radio_box' with radio buttons denoting valid schemes for `selected_nature' and `color_count'.
		require
			radio_box_not_void: radio_box /= Void
		local
			l_choices: ARRAY [TUPLE [sequence_name: STRING; sequence_array: ARRAY [EV_COLOR]; is_default_choice: BOOLEAN]]
			l_tuple: TUPLE [sequence_name: STRING; sequence_array: ARRAY [EV_COLOR]; is_default_choice: BOOLEAN]
			l_button: EV_RADIO_BUTTON
		do
			l_choices := available_sequences (selected_nature, color_count)
			if not l_choices.is_empty then
				selected_sequence := Void
				radio_box.wipe_out
				across l_choices as i_choice loop
					l_tuple := i_choice.item
					l_button := gui.factory.new_radio_button (radio_box, l_tuple.is_default_choice, l_tuple.sequence_name,
						agent set_sequence (l_tuple.sequence_array), 0)
					if l_tuple.is_default_choice then
						set_sequence (l_tuple.sequence_array)
					end
				end
			else
				error.warn ("Too many colors requested. Selected sequence unchanged.", Void)
				error.flush
			end
		end
	
	set_sequence (a_sequence: ARRAY [EV_COLOR]) is
			-- Set `selected_sequence' to `a_sequence'.
		require
			a_sequence_not_void: a_sequence /= Void
			no_void_color: not a_sequence.has (Void)
			colors_box_not_void: colors_box /= Void
		do
			if a_sequence.count = color_count then
				selected_sequence := a_sequence
				fill_colors_box
			else
				do_nothing
					-- this happens when wiping out `radio_box'
			end
		ensure
			selected_sequence_aliased: selected_sequence = a_sequence
		end
	
	fill_colors_box is
			-- Fill `colors_box' with `selected_sequence'.
		require
			colors_box_not_void: colors_box /= Void
			selected_sequence_not_void: selected_sequence /= Void
		local
			l_view: ROSE_COLOR_VIEW
		do
			colors_box.wipe_out
			across selected_sequence as i_sequence loop
				create l_view.make (colors_box, i_sequence.item)
				l_view.set_index (i_sequence.cursor_index)
				l_view.set_modifiable (False)
				l_view.set_pebble (l_view)
				l_view.set_name ("Color " + i_sequence.item.out)
			end
		ensure
			correct_count: colors_box.count = color_count
		end

	set_nature is
			-- Set `selected_nature' according to `nature_selector'.
			-- Also, populate `size_selector'.
		local
			l_text: STRING
			l_count: INTEGER
			l_strings: ROSE_LINEAR_ARRAY [STRING]
		do
			if nature_selector /= Void and then nature_selector.has_selection and size_selector /= Void then
				l_text := nature_selector.selected_text.as_string_8
				if sequential_data_label.same_string (l_text) then
					l_count := 9
					selected_nature := Sequential_data
				elseif diverging_data_label.same_string (l_text) then
					l_count := 11
					selected_nature := Diverging_data
				elseif qualitative_data_label.same_string (l_text) then
					l_count := 12
					selected_nature := Qualitative_data
				end
				if l_count > 0 then
					size_selector.wipe_out
					create l_strings.make_empty
					across (3 |..| l_count) as i_interval loop
						l_strings.extend (i_interval.item.out)
					end
					size_selector.set_strings (l_strings)
				end
			end
		end
	
	set_size is
			-- Set `color_count' to selection from `size_selector'.
			-- Also fill `radio_box'.
		local
			l_text: STRING
		do
			if size_selector /= Void and then size_selector.has_selection and radio_box /= Void then
				l_text := size_selector.selected_text.as_string_8
				if l_text.is_integer then
					color_count := l_text.to_integer.as_natural_8
					fill_radio_box
				end
			end
		end

feature {NONE} -- Button agents

	notify_on_ok is
			-- Take action after OK button is pressed.
		do
			dialog.hide
		ensure
			no_palette_available: selected_sequence = Void
		end

	notify_on_cancel is
			-- Take action after Cancel button is pressed.
		do
			selected_sequence := Void
			dialog.hide
		ensure
			no_palette_available: selected_sequence = Void
		end

invariant

	consistent_color_count: selected_sequence /= Void implies selected_sequence.count = color_count
	no_void_color: selected_sequence /= Void implies not selected_sequence.has (Void)
		
end

