indexing
	description: "Net of fee performance for composites"
	author: "Dan Wehn"
	date: "$Date: 2014/02/26 $"
	archive: "$File: //rose/source/kernel/expert_system/aflows/composite_net_of_fee_returns.e $"

class
	COMPOSITE_NET_OF_FEE_RETURNS
inherit
	ROSE_SHARED_GLOBAL


feature -- Basic operations

	monthly_nof_returns(a_composite: RP_PRODUCT; a_currency: RP_CURRENCY): DS_HASH_TABLE[RP_DECIMAL, ROSE_DATE] is
			-- Entire set of nof monthly returns for 'a_composite' in 'a_currency'
		require
			composite_not_void: a_composite /= Void
			currency_not_voif: a_currency /= Void
		local
			l_perf_system: RP_PERFORMANCE_SYSTEM
		do
			l_perf_system := manager.performance_system
			create Result.make_default
			if a_composite.perf_fee_net_alpha.is_empty then
				Result := nof_perf_with_base_fee(a_composite, a_currency)
			else
				Result := nof_perf_with_performance_fee(a_composite, a_currency)
			end
		ensure
			non_void_result: Result /= Void
		end


feature {NONE} -- Implementation

	date_check(a_table: DS_HASH_TABLE[RP_DECIMAL, ROSE_DATE]; a_date: ROSE_DATE): BOOLEAN is
			-- check to see if 'a_table' has 'a_date'
		do
			Result := a_table.has (a_date)
		end


	date_range_cache(a_returns: DS_HASH_TABLE[RP_DECIMAL, ROSE_DATE]; a_composite: RP_PRODUCT): DS_HASH_TABLE[ROSE_DATE, ROSE_DATE] is
			-- mapped begin-end dates for full composite monthly date range
			-- the begin date is either the inception date or the previous December
		require
			args_not_void: a_returns /= Void and a_composite /= Void
		local
			l_dates: ROSE_LINEAR_ARRAY[ROSE_DATE]
		do
			create Result.make_default
			l_dates := get_sorted_dates (a_returns)
			l_dates.do_all (agent fill_date_cache(l_dates, Result, a_composite, ?))
		ensure
			non_void_result: Result /= Void
		end

	fill_cums(a_result_table: DS_HASH_TABLE[RP_DECIMAL, ROSE_DATE]; a_return: RP_DECIMAL; a_date: ROSE_DATE) is
			-- fill 'a_result_table' with cum 'a_return'
		require
			result_table_not_void: a_result_table /= Void
			return_not_void: a_return /= Void
			valid_date: a_date /= Void
		local
			l_cum: RP_DECIMAL
		do
			l_cum := a_result_table.last * (1 + a_return/100)
			a_result_table.force_last (l_cum, a_date)
		end

	fill_dates(a_list: DS_LINKED_LIST[ROSE_DATE]; a_value: RP_DECIMAL; a_date: ROSE_DATE) is
			-- fill 'a_array' with 'a_date'
		require
			list_not_void: a_list /= Void
			date_not_void: a_date /= Void
		do
			a_list.force_last (a_date)
		end

	fill_date_cache(a_date_collection: ROSE_LINEAR_ARRAY[ROSE_DATE]; a_cache: DS_HASH_TABLE[ROSE_DATE, ROSE_DATE]; a_composite: RP_PRODUCT; a_date: ROSE_DATE) is
			-- fill cache
		do
			if a_date_collection.first > get_begin_date (a_date, a_composite.investment_region) then
				a_cache.force_last (a_date_collection.first, a_date)
			else
				a_cache.force_last (get_begin_date (a_date, a_composite.investment_region), a_date)
			end
		end


	fill_net_monthly_perf_with_base_fees(a_composite: RP_PRODUCT; a_return_cache: DS_HASH_TABLE[RP_DECIMAL, ROSE_DATE]; a_perf: RP_DECIMAL; a_date: ROSE_DATE ) is
			-- return net composite monthly performace using base fees
		require
			cache_not_void: a_return_cache /= Void
			return_not_void: a_perf /= Void
			date_not_void: a_date /= Void
		local
			l_percent_base_fee: RP_DECIMAL
		do

			if not a_composite.base_fee.is_empty then
				l_percent_base_fee := (a_composite.base_fee.to_double / 100) / 12
				a_return_cache.force_last (a_perf - l_percent_base_fee, a_date)
			else
				(create {EXCEPTIONS}).raise (a_composite.code + " does not have a base fee in GCDB! " )
			end
		end


	fill_perf_fee_nof_returns(a_net, a_bog, a_cum_result, a_result: DS_HASH_TABLE[RP_DECIMAL, ROSE_DATE];
							a_composite: RP_PRODUCT;
							a_date: ROSE_DATE) is
			-- fill cum net comp returns with perf fees included
		require
			non_void_args: a_net /= Void and a_bog /= Void and a_date /= Void and a_result /= Void
			tables_have_date: date_check(a_net, a_date) and date_check(a_bog, a_date)
		local
			l_net_perf_with_fee, l_numerator, l_denominator: RP_DECIMAL
			l_result, l_last_month_cum_return: RP_DECIMAL
			l_perf_fee_adjustment: RP_DECIMAL
			l_alpha_period_begin: ROSE_DATE
			l_date_cache: DS_HASH_TABLE[ROSE_DATE, ROSE_DATE]
		do
			if a_date = a_net.keys.first then
				a_result.force_last (a_net.first, a_date)
				a_cum_result.force_last (a_net.first, a_date)
			else
				a_net.search (a_date)
				if a_net.found then
					l_numerator := a_net.found_item
				else
					(create {EXCEPTIONS}).raise (
						"No net return found for: " + a_date.out)
				end
				if nof_alpha (a_net.cloned_object, a_bog.cloned_object, a_date, a_composite) < 0 then
					l_perf_fee_adjustment := 0
				else
					l_perf_fee_adjustment := nof_alpha (a_net.cloned_object, a_bog.cloned_object, a_date, a_composite)
				end
				l_last_month_cum_return := a_cum_result.last
				l_date_cache := date_range_cache (a_net, a_composite)
				l_date_cache.search (a_date)
				if l_date_cache.found then
					l_alpha_period_begin := l_date_cache.found_item
				else
					(create {EXCEPTIONS}).raise ("No beginning date found for end date: " + a_date.out)
				end
				a_net.search (l_alpha_period_begin)
				if a_net.found and then not a_net.found_item.is_zero then
					l_denominator := a_net.found_item
					a_cum_result.search (l_alpha_period_begin)
					if a_cum_result.found then
						l_net_perf_with_fee := ((l_numerator / l_denominator) - l_perf_fee_adjustment) * a_cum_result.found_item
						if not l_last_month_cum_return.is_zero then
							l_result := ((l_net_perf_with_fee / l_last_month_cum_return) - 1) * 100
							a_result.force_last (l_result, a_date)
						else
							(create {EXCEPTIONS}).raise (
								"Bad beginning cumulative return")
						end
						a_cum_result.force_last (l_net_perf_with_fee, a_date)
					else
						(create {EXCEPTIONS}).raise (
							"Bad cum net return for: " + l_alpha_period_begin.out)
					end
				else
					(create {EXCEPTIONS}).raise (
						"Bad net return for: " + l_alpha_period_begin.out)
				end

			end

		end

	get_begin_date(a_date: ROSE_DATE; a_zone: RP_ZONE): ROSE_DATE is
			-- return last end-of-december date before 'a_date'
		local
			l_calendar: RP_CALENDAR
			l_beginning_of_year: ROSE_DATE
		do
			create l_beginning_of_year.make (a_date.year, 1, 1)
			l_calendar := manager.legacy.performance_system.calendar (a_zone)
			Result := l_calendar.ith_work_day_before (l_beginning_of_year, 1)
		end


	get_sorted_dates(a_table: DS_HASH_TABLE[RP_DECIMAL, ROSE_DATE]): ROSE_LINEAR_ARRAY[ROSE_DATE] is
			-- returns sorted date array from 'a_table'
		require
			non_void_args: a_table /= Void
		local
			l_date_list: DS_LINKED_LIST[ROSE_DATE]
		do
			create l_date_list.make_default
			a_table.do_all_with_key (agent fill_dates(l_date_list, ?, ?))
			Result := l_date_list.to_array
			Result.sort_by (agent sort_date_ascending)
		ensure
			non_void_result: Result /= Void
		end


	manager: RP_AXAROSENBERG is
			-- instance of RP_AXAROSENBERG
		once
			create Result
		end


	nof_perf_with_performance_fee(a_composite: RP_PRODUCT; a_currency: RP_CURRENCY): DS_HASH_TABLE[RP_DECIMAL, ROSE_DATE] is
			-- get composite net of fee performance using performance fees
		require
			composite_not_void: a_composite /= Void
			currency_not_void: a_currency /= Void
		local
			l_cum_net_comp_returns, l_cum_bogie_returns, l_cum_nof_with_perf_fee: DS_HASH_TABLE[RP_DECIMAL, ROSE_DATE]
		do

			create Result.make_default
			create l_cum_nof_with_perf_fee.make_default
			l_cum_net_comp_returns := to_cumulative (nof_perf_with_base_fee (a_composite, a_currency))
			l_cum_bogie_returns := to_cumulative (manager.performance_system.official_perf_composite_benchmark_returns (a_composite, a_currency))
			l_cum_net_comp_returns.keys.do_all (
							agent fill_perf_fee_nof_returns(l_cum_net_comp_returns,
															l_cum_bogie_returns,
															l_cum_nof_with_perf_fee,
															Result,
															a_composite,
															?))
		end


	nof_alpha(a_nof_returns, a_bogie_returns: DS_HASH_TABLE[RP_DECIMAL, ROSE_DATE]; a_date: ROSE_DATE; a_composite: RP_PRODUCT): RP_DECIMAL is
			-- nof alpha (with performance fee) on 'a_date'
		local
			l_end_net_cum, l_begin_net_cum, l_end_bogie_cum, l_begin_bogie_cum: RP_DECIMAL
			l_begin_date: ROSE_DATE
			l_fee: RP_DECIMAL
			l_dates: ROSE_LINEAR_ARRAY[ROSE_DATE]
		do
			l_dates := get_sorted_dates (a_nof_returns)
			if l_dates.first > get_begin_date(a_date, a_composite.investment_region) then
				l_begin_date := l_dates.first
			else
				l_begin_date := get_begin_date(a_date, a_composite.investment_region)
			end
			create l_fee.make_from_string (a_composite.perf_fee_net_alpha)
			a_nof_returns.search (a_date)
			if a_nof_returns.found then
				l_end_net_cum := a_nof_returns.found_item
				a_bogie_returns.search (a_date)
				if a_bogie_returns.found then
					l_end_bogie_cum := a_bogie_returns.found_item
					a_nof_returns.search (l_begin_date)
					if a_nof_returns.found then
						l_begin_net_cum := a_nof_returns.found_item
						a_bogie_returns.search (l_begin_date)
						if a_bogie_returns.found then
							l_begin_bogie_cum := a_bogie_returns.found_item
							Result := l_fee * ((l_end_net_cum / l_begin_net_cum) - (l_end_bogie_cum / l_begin_bogie_cum))
						else
							output("No benchmark returns for '$1'" + "%N", <<l_begin_date>>)
							Result := -999
						end
					else
						output("No nof returns for '$1'" + "%N", <<l_begin_date>>)
						Result := -999
					end
				else
					output("No benchmark returns for '$1'" + "%N", <<a_date>>)
					Result := -999
				end
			else
				output("No nof returns for '$1'" + "%N", <<a_date>>)
				Result := -999
			end
		ensure
			non_void_result: Result /= Void
		end


	nof_perf_with_base_fee(a_comp: RP_PRODUCT; a_currency: RP_CURRENCY): DS_HASH_TABLE[RP_DECIMAL, ROSE_DATE] is
			-- table of dated nof returns for 'a_composite' in 'a_currency' using base fees only
		require
			comp_not_void: a_comp /= Void
			currency_not_void: a_currency /= Void
		do
			create Result.make_default
			manager.performance_system.official_perf_composite_returns (a_comp, a_currency).do_all_with_key (agent fill_net_monthly_perf_with_base_fees(a_comp, Result, ?, ?))
		ensure
			non_void_result: Result /= Void
		end



	search_return(a_table: DS_HASH_TABLE[RP_DECIMAL, ROSE_DATE]; a_date: ROSE_DATE): BOOLEAN is
			-- search 'a_table' for 'a_date'
		require
			args_not_void: a_table /= Void and then a_date /= Void
		do
			a_table.search (a_date)
			if not a_table.found then
				output ("No return value found on " + a_date.out, Void)
			else
				Result := True
			end
		end


	sort_date_ascending(a_date1, a_date2: ROSE_DATE): BOOLEAN is
			-- sorting agent for ROSE_DATES
		do
			Result := (a_date1 /= Void and a_date2 /= Void) and then a_date1 < a_date2
		end


	to_cumulative(a_returns: DS_HASH_TABLE[RP_DECIMAL, ROSE_DATE]): DS_HASH_TABLE[RP_DECIMAL, ROSE_DATE] is
			-- make table of dated cumulative values from a dated set of percentage values
		require
			table_not_void: a_returns /= Void
			table_not_empty: not a_returns.is_empty
		do
			create Result.make_default
			Result.force_last (100, a_returns.keys.first.previous_month_end)
			a_returns.do_all_with_key (agent fill_cums(Result, ?, ?))
		end


end
