indexing
	description: "[
		Credit Default Swap.
		Extremely simplistic implementation for prototypical exploration.
	]"
	author: "Yves Chauvin"
	last_editor: "$Author: yc $"
	date: "$Date: 2009/01/21 $"
	revision: "$Revision: #1 $"
	archive: "$Archive: $"

class
	CREDIT_DEFAULT_SWAP

inherit
	ROSE_MATH_FUNCTIONS

feature -- Settings

	set_credit_spread (a_value: DOUBLE) is
			-- setting cretid spread
		require
			positive_spread: a_value >= 0
		do
			credit_spread := a_value -- actually of underlying instrument
			set_default_probability
		end

	set_interest_rate (a_rate: DOUBLE) is
			-- setting interest rate
		require
			positive_rate: a_rate >= 0
		do
			interest_rate := a_rate
		end

	set_nominal_value (a_value: DOUBLE) is
			-- setting nominal value of underlying instrument
		require
			positive_nominal: a_value >= 0
		do
			nominal_value := a_value
		end

	set_term (a_term: INTEGER) is
			-- sets the term of the contract and therefore of the underlying default distribution
		require
			positive_term: a_term >= 0
		do
			term := a_term
			create default_probability_distribution
			default_probability_distribution.set_order (number_of_steps + 1) -- the distribution is winsorized
		end

feature -- Access

	credit_spread: DOUBLE
			-- credit spread for underlying instrument
	term: INTEGER
			-- in years
	interest_rate: DOUBLE
			-- per year
	nominal_value: DOUBLE
			-- nominal value
	premium_payments: ROSE_LINEAR_ARRAY[CREDIT_DEFAULT_SWAP_PAYMENT]
			-- premium payments
	default_payments: ROSE_LINEAR_ARRAY[CREDIT_DEFAULT_SWAP_PAYMENT]
			-- default payment
	aggregated_premium_payments: ROSE_LINEAR_ARRAY[CREDIT_DEFAULT_SWAP_PAYMENT]
			-- aggregated premium payments (virtual contract)

	expected_default_payment: DOUBLE
			-- expected value of default payment

feature -- Basic operations

	aggregate_premium_payments is
			-- aggregate premiums
		local
			i_aggregate: INTEGER
			l_aggregate: CREDIT_DEFAULT_SWAP_PAYMENT
		do
			from i_aggregate := 2 until i_aggregate > aggregated_premium_payments.count	loop
				l_aggregate := aggregated_premium_payments.item (i_aggregate)
				l_aggregate.set_value (aggregated_premium_payments.item (i_aggregate - 1).present_value + premium_payments.item (i_aggregate).present_value)
				i_aggregate := i_aggregate + 1
			end
		end

	cds_system: RP_CDS_SYSTEM is
			-- CDS system
		once
			create Result
		end

	name: STRING is
			-- CDS name
		once
			Result := "CDS 1"
		end

	initialize is
			-- initialization
		local
			i_step: INTEGER
			l_premium, l_default, l_aggregate: CREDIT_DEFAULT_SWAP_PAYMENT
		do
			create premium_payments.make_empty
			create default_payments.make_empty
			create aggregated_premium_payments.make_empty
			from i_step := 1 until i_step > number_of_steps loop
				create l_premium.make (cds_system, name, create {ROSE_DATE}.make_today) -- argument time to fix
				l_premium.set_step (i_step)
				premium_payments.extend (l_premium)
				create l_default.make (cds_system, name, create {ROSE_DATE}.make_today)
				l_default.set_step (i_step)
				default_payments.extend (l_default)
				create l_aggregate.make (cds_system, name, create {ROSE_DATE}.make_today)
				l_aggregate.set_step (i_step)
				aggregated_premium_payments.extend (l_aggregate)
				i_step := i_step + 1
			end
			premium_payments.do_all (agent initialize_payment)
			default_payments.do_all (agent initialize_payment)
			aggregated_premium_payments.do_all (agent initialize_payment)
		end

	number_of_steps: INTEGER is
			-- number of premiums
		do
			Result := term * 4 -- quarterly payments
		end

	compute_premium_payments is
			-- computes premium payments from CDS parameters
		do
			premium_payments.do_all (agent compute_premium_payment)
		end

	compute_default_payments is
			-- computes default payments from CDS parameters
		do
			default_payments.do_all (agent compute_default_payment)
		end

	compute_aggregated_payments is
			-- computed aggregate premium payments
		do
			aggregated_premium_payments.do_all (agent compute_aggregated_premium_payment)
			aggregate_premium_payments
		end

	annuity: DOUBLE is
			-- quarterly premium payments
		do
			Result := nominal_value / (number_of_steps)
		end

	expected_present_value_net: DOUBLE is
			-- expected value of net
		local
			i_period: INTEGER
			l_probability_default: DOUBLE
			l_present_cost: DOUBLE
		do
			from i_period := 1 until i_period > 4 * term loop
				l_probability_default := default_probability_distribution.probability_mass_function (i_period)
				l_present_cost := present_value_net (i_period)
				Result := Result + l_present_cost * l_probability_default
				i_period := i_period + 1
			end
			l_probability_default := default_probability_distribution.probability_mass_function (i_period)
			l_present_cost := - present_value_annuities (i_period - 1) -- no default within term: no default payment, but total annuities
			Result := Result + l_present_cost * l_probability_default
		end

	expected_present_value_cost: DOUBLE is
			-- expected present value of cost
		local
			i_period: INTEGER
			l_probability_default: DOUBLE
			l_present_cost: DOUBLE
		do
			from i_period := 1 until i_period > 4 * term loop
				l_probability_default := default_probability_distribution.probability_mass_function (i_period)
				l_present_cost := present_value_annuities (i_period)
				Result := Result + l_present_cost * l_probability_default
				i_period := i_period + 1
			end
			l_probability_default := default_probability_distribution.probability_mass_function (i_period)
			l_present_cost := present_value_annuities (i_period - 1)
			Result := Result + l_present_cost * l_probability_default
		end

	expected_present_value_default_payment_old: DOUBLE is
			-- expected present value of premium payments
		local
			i_period: INTEGER
			l_probability_default: DOUBLE
			l_present_payment: DOUBLE
		do
			from i_period := 1 until i_period > 4 * term loop
				l_probability_default := default_probability_distribution.probability_mass_function (i_period)
				l_present_payment := present_value_default_payment (i_period)
				Result := Result + l_present_payment * l_probability_default
				i_period := i_period + 1
			end
		end

	expected_present_value_default_payment: DOUBLE is
			-- computes aggregated default payment
		do
			default_payments.do_all (agent sum_expected_default_payment)
		end

	present_value_net (a_number_of_annuities: INTEGER): DOUBLE
			-- present value of payment - present value of annuities
		do
			Result := present_value_default_payment (a_number_of_annuities) - present_value_annuities (a_number_of_annuities)
		end

	default_payment_number: DOUBLE is
			-- payment of CDS seller to buyer in case of default
		do
			Result := nominal_value
		end

	premium: DOUBLE is
			-- assuming no profit: expected cost = expected payment
		local
			l_neutral_factor: DOUBLE
		do
			l_neutral_factor := expected_present_value_default_payment / expected_present_value_cost
			Result := nominal_value * l_neutral_factor / number_of_steps
		end

	present_value_annuities (a_number_of_annuities: INTEGER) : DOUBLE is
			-- present value annuities
		require
			a_number_of_annuities_pos: a_number_of_annuities >= 0
		do
			Result := present_value_factor_annuities (a_number_of_annuities) * annuity
		end

	present_value_factor (a_number_of_annuities: INTEGER): DOUBLE is
			-- present value factor of a single amount
		require
			a_number_of_annuities_pos: a_number_of_annuities >= 0
		do
			Result := 1 / (1 + (interest_rate/4))^(a_number_of_annuities)
		end

	present_value_factor_annuities (a_number_of_annuities: INTEGER): DOUBLE is
			-- present value factor of annuities
		require
			a_number_of_annuities_pos: a_number_of_annuities >= 0
		do
			if interest_rate = 0 then
				Result := a_number_of_annuities
			else
				Result := 1 / (interest_rate/4) * (1 - 1 / (1 + (interest_rate/4))^(a_number_of_annuities))
			end
		end

	present_value_default_payment (a_number_of_annuities: INTEGER): DOUBLE Is
			-- present value of payment in case of failure
		require
			a_number_of_annuities_pos: a_number_of_annuities >= 0
		do
			Result := present_value_factor (a_number_of_annuities) * default_payment_number
		end

feature {CREDIT_DEFAULT_SWAP_PAYMENT} -- probabilistic parameters

	default_probability_distribution: ROSE_PR_DISTRIBUTION_GEOMETRIC_WINSORIZED
			-- probability distribution of default

feature {NONE} -- Implementation

	set_default_probability is
			-- computes default probability from credit spread
		local
			l_default: DOUBLE
		do
			l_default := 1 - euler ^ (- credit_spread) -- actually spread of underlying obligation
			default_probability_distribution.set_probability (l_default)
		end

	compute_default_payment (a_default: CREDIT_DEFAULT_SWAP_PAYMENT) is
			-- computes default payment
		do
			a_default.set_value (nominal_value)
			a_default.set_probability (default_probability_distribution.probability_mass_function (a_default.step))
			a_default.compute
		end

	compute_aggregated_premium_payment (a_premium: CREDIT_DEFAULT_SWAP_PAYMENT) is
			-- computed aggregaged premium
		do
			a_premium.set_probability (default_probability_distribution.probability_mass_function (a_premium.step))
			a_premium.compute
		end

	compute_premium_payment (a_premium: CREDIT_DEFAULT_SWAP_PAYMENT) is
			-- computes premium
		do
			a_premium.set_value (nominal_value / (number_of_steps))
			a_premium.set_probability (default_probability_distribution.probability_mass_function (a_premium.step))
			a_premium.compute
		end

	sum_expected_default_payment (a_default: CREDIT_DEFAULT_SWAP_PAYMENT) is
			-- aggregates expected default payment
		do
			expected_default_payment := expected_default_payment + a_default.present_expected_value
		end

	initialize_payment (a_premium: CREDIT_DEFAULT_SWAP_PAYMENT) is
			-- initializes premium
		do
			a_premium.set_cds (Current)
		end


end
