note

	description:
	"[
		Object model for CURRENCY.  Exchange rates and returns are provided
		through this object as well as access to the major currencies used
		in the expert system.
	]"
	author: "Mark Howard"
	last_editor: "$Author: c4 $"
	date: "$Date: 2015/02/24 $"
	revision: "$Revision: #115 $"
	path: "$File: //rose/source/kernel/investment_entities/currency.e $"
	storable_version: ""

class CURRENCY

inherit

	RP_SERIALIZABLE

	INVESTMENT_ENTITY
		rename
			iso as investment_entity_iso
		undefine
			is_equal,
			name_identifier_key
		redefine
			make,
			debug_output,
			ur_targets
		end

	PERSISTENT_CURRENCY
		undefine
			context,
			make,
			make_context,
			debug_output,
			ur_targets,
			name_identifier_key,
			market_exchange_rate
		redefine
			default_market
		end

	DATABASE_SERVER
		undefine
			is_equal
		end

	DOUBLE_MATH
		undefine
			is_equal
		end

	ROSE_SHARED_GLOBAL
		undefine
			is_equal
		end

create

	make,
	make_anonymous

feature {NONE} -- Creation

	make (a_context: like context; a_category_id: INTEGER; a_handle: INTEGER; a_start_date, a_end_date: ROSE_DATE) is
			-- Make currency object in a context with a category, handle, start, and end date.
		do
			Precursor {INVESTMENT_ENTITY} (a_context, a_category_id, a_handle, a_start_date, a_end_date)
			iso := iso_code_key (a_end_date)
			name := name_at (a_end_date)
			create default_market
		end

feature -- Basic operations

	register is
			-- Register this entity.
		do
		end

	make_from_entities (a_currency: ENTITIES_CURRENCY) is
			-- make from entities
		require
			valid_currency: a_currency /= Void
		do
			if a_currency.iso_code /= Void then
				iso := a_currency.iso_code.twin
			end
			name := a_currency.name.twin
				--make(a_currency.rose_id)
			index := a_currency.old_positional_id
			entities_currency := a_currency.twin
		end

feature -- Access

	entities_currency: ENTITIES_CURRENCY

	index: INTEGER

	name: STRING

	monthly_return (a_currency: like current; a_month: ROSE_MONTH): DOUBLE is
			-- Return for `a_currency' / Current for `a_month'
		require
			valid_currency: a_currency /= Void
			valid_month: a_month /= Void
		do
			Result := exchange_return (a_currency, a_month.previous_month, a_month.last)
		end

	is_known_currency: BOOLEAN is
			-- does currency object qualify as a valid currency
		local
			l_iso: STRING
		do
			if iso /= Void then
				l_iso := iso
				l_iso.prune_all (' ')
			end
			Result := handle /= 0 and then iso /= Void and then not l_iso.is_empty and then name /= Void and then not name.as_lower.is_equal ("unknown")
		end

	forward_exchange_rate (a_currency: CURRENCY; a_as_of_date: ROSE_DATE; a_maturity_date: ROSE_DATE): FORWARD_EXCHANGE_RATE is
			-- Forward exchange rate of `a_currency' to `Current'
		local
			l_wm_spot: DOUBLE
			l_bloomberg_spot_record, l_bloomberg_forward_record: BLOOMBERG_FORWARD_RECORD
			l_bloomberg_spot, l_bloomberg_forward: DOUBLE
			l_bloomberg_direct_points, l_forward_rate: DOUBLE
			l_final_iso: detachable STRING
		do
			create Result.make
			if iso ~ "USD" then
					-- now only consider USD based currency
				l_wm_spot := exchange_rate (a_currency, a_as_of_date)
				l_bloomberg_spot_record := bloomberg_currency_data_access.get_forward (a_currency.iso, a_as_of_date, a_as_of_date)
				l_bloomberg_forward_record := bloomberg_currency_data_access.get_forward (a_currency.iso, a_as_of_date, a_maturity_date)
				l_bloomberg_spot := l_bloomberg_spot_record.forward_rate
				l_bloomberg_forward := l_bloomberg_forward_record.forward_rate
				if l_bloomberg_spot = undefined_integer or l_bloomberg_forward = undefined_integer then
					-- in valid result, do nothing
					Result.set_forward_rate (undefined_integer)
					Result.set_valid_flag (FALSE)
				else
					if a_currency.iso ~ "AUD" or a_currency.iso ~ "GBP" or a_currency.iso ~ "EUR" then
						l_bloomberg_direct_points := l_bloomberg_forward - l_bloomberg_spot
					else
						l_bloomberg_direct_points := (1.0 / l_bloomberg_forward) - (1.0 / l_bloomberg_spot)
					end
					l_forward_rate := l_wm_spot + l_bloomberg_direct_points
					Result.set_forward_rate (l_forward_rate)
					Result.set_valid_flag (True)
					Result.set_forward_data_date (l_bloomberg_forward_record.data_date)
				end
			else
				if attached a_currency.iso as l_other_iso and attached iso as l_iso then
					l_final_iso := l_other_iso + l_iso
				end
				l_wm_spot := exchange_rate (a_currency, a_as_of_date)
				l_bloomberg_spot_record := bloomberg_currency_data_access.get_forward (l_final_iso, a_as_of_date, a_as_of_date)
				l_bloomberg_forward_record := bloomberg_currency_data_access.get_forward (l_final_iso, a_as_of_date, a_maturity_date)
				l_bloomberg_spot := l_bloomberg_spot_record.forward_rate
				l_bloomberg_forward := l_bloomberg_forward_record.forward_rate
				if l_bloomberg_spot = undefined_integer or l_bloomberg_forward = undefined_integer then
					Result.set_forward_rate (undefined_integer)
					Result.set_valid_flag (False)
						--	Result.set_forward_data_date( l_bloomberg_forward_record.data_date )					
				else
					l_bloomberg_direct_points := l_bloomberg_forward - l_bloomberg_spot
					l_forward_rate := l_wm_spot + l_bloomberg_direct_points
					Result.set_forward_rate (l_forward_rate)
					Result.set_valid_flag (True)
					Result.set_forward_data_date (l_bloomberg_forward_record.data_date)
				end
			end
		end

feature -- Elemenet modification

	set_name (a_string: STRING) is
			-- set name
		do
			name := a_string
		end

	set_iso_code (a_iso: STRING) is
			-- set iso code
		do
			iso := a_iso
		end

	set_index (a_index: INTEGER) is
		do
			index := a_index
		end

feature -- used for industry aggregates

	exchange_rate_array (a_currency: CURRENCY; a_dates: ARRAY [ROSE_DATE]): ROSE_LINEAR_ARRAY [DOUBLE] is
			--Exchange rates for a_currency keyed by a_dates
		local
			i: INTEGER
			l_double: DOUBLE
			l_date: ROSE_DATE
		do
			if a_currency /= Void and then (a_dates /= Void and then a_dates.count > 0) then
				create Result.make (1, a_dates.count)
				from
					i := 1
				until
					i > a_dates.count
				loop
					l_date := a_dates.item (i)
					if l_date /= Void then
						l_double := exchange_rate (a_currency, l_date)
						Result.put (l_double, i)
					end
					i := i + 1
				end
			end
		end

	quarterly_usd_rates: ROSE_MULTI_KEYED_TABLE [REAL] is
			-- quarterly usd rates
		local
			l_path, l_filename: STRING
			l_raw_file: ROSE_RAW_FILE
			l_pointer: POINTER
			l_count: INTEGER
			i, j: INTEGER
			l_rates: QUARTERLY_EXCHANGE_RATES
			l_rate_array: ARRAY [REAL]
		once
			create Result.make (1000)
			l_path := global.app.storable_path ("model\valdata\")
			l_filename := config.value_default ("/valdata/ia_exchange_rates", l_path + "ia_exchange_rates.data")
				--			l_filename := config.value_default ("/valdata/ia_exchange_rates", "\\axaros.com\shares\Eiffel\Storables\model\valdata\ia_exchange_rates.data")
				--			l_filename := config.value_default ("/valdata/ia_exchange_rates", "\\axaros.com\shares\Pub_nt\sections\ia_exchange_rates.data")
				--		  	create l_raw_file.make_open_read("\\pinegv\prod_c\valdata_w\d\ia_exchange_rates.data")
				--		  	create l_raw_file.make_open_read("\\axaros.com\shares\Pub_nt\sections\ia_exchange_rates.data")
			create l_raw_file.make_open_read (l_filename)
			create l_rates
			l_count := l_raw_file.count
			l_pointer := l_pointer.memory_alloc (l_count)
			l_raw_file.read_data (l_pointer, l_count)
			l_rates.set (l_pointer)
			l_rates.read
			from
				i := l_rates.rates.lower
			until
				i > l_rates.rates.upper
			loop
				l_rate_array := l_rates.rates.item (i).xrate
				from
					j := l_rate_array.lower
				until
					j > l_rate_array.upper
				loop
					if is_defined_double (l_rate_array.item (j)) then
						Result.put (l_rate_array.item (j), [i, j])
					end
					j := j + 1
				end
				i := i + 1
			end
			l_pointer.memory_free
		end

	quarterly_usd_rate (a_country: COUNTRY; a_quarter: ROSE_DATE): DOUBLE is
		do
			Quarterly_usd_rates.search ([a_quarter.julian_quarter, a_country.old_positional_id])
			if quarterly_usd_rates.found then
				Result := quarterly_usd_rates.found_item
			end
		end

feature -- definition of the target

	ur_targets: ROSE_LINEAR_ARRAY [ROSE_TARGET [ENTITY]] is
		once
			create Result.make_empty
			Result.extend (create {FOREX_SIGNATURE})
		end

feature -- legacy access

	fx_t_at (a_date: ROSE_DATE): PAIR [FX_T, BOOLEAN]
		do
			-- All callers are broken anyway, so body emptied to remove call on atlas
		end

	interest_rate (a_date: ROSE_DATE; a_months: INTEGER): DOUBLE is
			-- Interest rate bid for this currency's country on 'a_date'
		require
			valid_months: a_months > 0 and a_months <= 12
			valid_date: a_date /= Void
		local
			l_fx_t_pair: like fx_t_at
			l_fx_t: FX_T
		do
			l_fx_t_pair := fx_t_at (a_date)
			if l_fx_t_pair.second then
				l_fx_t := l_fx_t_pair.first
				inspect a_months
				when 1 then
					Result := l_fx_t.euro_1m.bid
				when 2 then
					Result := l_fx_t.euro_2m.bid
				when 3 then
					Result := l_fx_t.euro_3m.bid
				when 6 then
					Result := l_fx_t.euro_6m.bid
				when 12 then
					Result := l_fx_t.euro_12m.bid
				else
				end
			else
				Result := Undefined_Double
			end
		end

	forward_usd_rate (a_date: ROSE_DATE; a_months: INTEGER): DOUBLE is
			-- Forward USD rate on 'a_date'
		require
			valid_months: a_months > 0 and a_months <= 12
			valid_date: a_date /= Void
		local
			l_fx_t_pair: like fx_t_at
			l_fx_t: FX_T
		do
			l_fx_t_pair := fx_t_at (a_date)
			if l_fx_t_pair.second then
				l_fx_t := l_fx_t_pair.first
				inspect a_months
				when 1 then
					Result := l_fx_t.fwd_1m.ask
				when 2 then
					Result := l_fx_t.fwd_2m.ask
				when 3 then
					Result := l_fx_t.fwd_3m.ask
				when 6 then
					Result := l_fx_t.fwd_6m.ask
				when 12 then
					Result := l_fx_t.fwd_12m.ask
				else
				end
			else
				Result := Undefined_Double
			end
		end

	forward_premium_1m (a_date: ROSE_DATE): DOUBLE is
			-- relative to USD
		local
			l_fx_t_pair: like fx_t_at
			l_fx_t: FX_T
		do
			l_fx_t_pair := fx_t_at (a_date)
			if l_fx_t_pair.second then
				l_fx_t := l_fx_t_pair.first
				Result := (l_fx_t.fwd_1m.bid + l_fx_t.fwd_1m.ask) / (l_fx_t.spot.bid + l_fx_t.spot.ask) - 1.0
			else
				Result := undefined_double
			end
		end

	forward_premium_1m_at (a_other: like current; a_date: ROSE_DATE): DOUBLE 
			--forward_premium_1m / a_other.forward_premium_1m
		local
			l_this, l_other: DOUBLE
		do
			l_this := forward_premium_1m (a_date)
			l_other := a_other.forward_premium_1m (a_date)
			if is_defined_double (l_this) and is_defined_double (l_other) then
				Result := (1.0 + l_this) / (1.0 + l_other) - 1.0
			else
				Result := undefined_double
			end
		end

	last_fx_t_date (a_date: ROSE_MONTH): ROSE_DATE is
		require
			valid_date: a_date /= Void
		local
			l_date: ROSE_DATE
			i: INTEGER
			l_success: BOOLEAN
			l_fx_t_pair: like fx_t_at
		do
			l_date := a_date.twin.last
			from
				i := 1
			until
				l_success or i > 5
			loop
				l_fx_t_pair := usd.fx_t_at (l_date)
				l_success := l_fx_t_pair.second
				Result := l_date.twin
				l_date := l_date.previous_date
				i := i + 1
			end
		end

	bloomberg_currency_data_access: BLOOMBERG_CURRENCY_ACCESS is
		once
			create Result
		end

feature -- representation

	debug_output: STRING is
			-- String that should be displayed in debugger to represent `Current'.
		do
			if attached iso as l_iso then
				Result := l_iso
			else
				Result := "<no iso code>"
			end
		end

feature -- definition of FX rates

	market_exchange_rate (a_market: MARKET; a_currency: PERSISTENT_CURRENCY; a_date: ROSE_DATE): DOUBLE
			-- default call to get FX rates
		local
			l_market: MARKET
		do
			if default_market = Void then
				create default_market
			end
			l_market := a_market
			if l_market = Void then
				l_market := default_market
			end
			if a_currency /= Void and a_date /= Void then
				Result := l_market.exchange_rate (Current, a_currency, a_date)
			else
				Result := undefined_double
			end
		end

feature -- Constants

	first_euro: ROSE_DATE is
		once
			create Result.make (1999, 1, 4)
		end

	first_extel: ROSE_DATE
			-- Earliest date for which Extel provide exchange rates
		once
			create Result.make (1995, 8, 1)
		ensure
			first_extel_attached: Result /= Void
		end

	constant_rate: EXCHANGE_RATE is
			-- 1 to 1 exchange rate
		once
			create Result.make (Void, Void, 1, Void)
		end

	name_identifier_key: STRING is
			-- Name id key
		once
			Result := "name"
		end

	usd: CURRENCY is
			-- currency object corresponding to US Dollar
		do
			Result := context.usd
		end

	jpy: CURRENCY is
			-- currency object corresponding to Japanese Yen
		do
			Result := context.jpy
		end

	dem: CURRENCY is
			-- currency object corresponding to Deutsche Mark
		once
			Result := context.currency_by_iso_code ("DEM", Void)
		end

	cad: CURRENCY is
			-- currency object corresponding to Canadian Dollar
		do
			Result := context.cad
		end

	gbp: CURRENCY is
			-- currency object corresponding to British Pound
		do
			Result := context.gbp
		end

	euro: CURRENCY is
			-- currency object corresponding to Euro
		do
			Result := context.eur
		end

feature {RCMD_ABC_GLOBAL_RI_STANDARDIZATION} -- Implementation

	default_market: AXA_ROSENBERG_DEFAULT_MARKET
			-- Default market which exchanges this currency.

end
