note

	description:

		"Representation of a record component, for example a substructure in C, or a column in Sql"

	author: "Mark Howard"
	last_editor: "$Author: c4 $"
	date: "$Date: 2014/05/13 $"
	revision: "$Revision: #31 $"
	path: "$File: //rose/source/kernel/infrastructure/persistence/data_component.e $"

deferred class

	DATA_COMPONENT

inherit

	DISPLAYABLE
		undefine
			display
		end
	ROSE_CONSTANT_VALUES
	C_POINTER
		rename
			c_data as pointer
		end
	DATA_DESCRIPTOR

	ROSE_SHARED_GLOBAL
		rename
			trace as global_trace
		end

feature -- Interpreter

	default_interpreter: STRING is do end
			-- Default interpreter type
	interpreter_type: STRING
			-- Type of interpreter

	interpreter_name: STRING is
			-- The name of the interpreter
		do
			if interpreter_type /= Void and not interpreter_type.is_empty then
				Result := interpreter_type.as_lower
				Result.append("_")
			elseif default_interpreter /= Void then
				Result := default_interpreter.twin
			end
		end

  	external_record_text_with_parents (a_parent_classes: ARRAY [STRING]; a_comment: STRING): STRING is
		-- text describing the code generation details and including the "inherit" clause for a generated class
  		do
  		end

feature -- Access

	component_closure: LINKED_LIST [DATA_COMPONENT] is
			-- flattened list of components
		deferred
		end

	value: ANY is
			-- Value of this field as an ANY
		deferred
		end

	get_field (an_index: INTEGER; a_receptacle: ANY) is
			-- Get the field and place it into a_receptable at field index an_index
		deferred
		end

	read (a_record: ANY) is
			-- Read from the field into a_record
		deferred
		end

	read_field (an_index: INTEGER; a_record: ANY) is
			-- Read from the field into a_record attribute at an_index
		deferred
		end

feature -- Basic Operations

	is_valid: BOOLEAN is
			-- Is this field currently accessible?
		do
			Result := True
		end

	repeat_count: INTEGER is
			-- The count of the array if this field is an array, otherwise 1
		do
			Result := (max_index - min_index + 1)
		end

	is_array: BOOLEAN is
			-- Is this field an array?
		do
			Result := has_bounds or uses_symbolic_indexes
		end

	pointer: POINTER
			-- The pointer attached to this field if this field is a pointer
	is_index: BOOLEAN
			-- Is this field an index?
	stored_length: INTEGER
			-- The length stored as a saved integer
	index: INTEGER
			-- Index of this field in the structure
	offset: INTEGER
			-- Offset of this field in the structure
	symbolic_offset: STRING
	repeat_index: INTEGER
	min_index: INTEGER
			-- Minimum index of this field if it is an array
	min_index_name: STRING
			-- Minimum index name of this field if it is an array that uses symbolic indexes
	max_index: INTEGER
			-- Maximum index of this field if it is an array
	max_index_name: STRING
			-- Maximum index name of this field if it is an array that uses symbolic indexes
	type_string: STRING
			-- Field's type as a string
	has_bounds: BOOLEAN
			-- Does this field have bounds, ie. is it an array with min and max indexes
	length: INTEGER is
			-- Field length
		deferred
		end

feature -- Modification

	set_value (a_value: ANY) is
			-- Set this field to a_value
		deferred
		end

	set_is_index is
			-- Set this field to be an index
		do
			is_index := True
		ensure
			is_index: is_index
		end

	set_type_string (a_string: STRING) is
			-- set 'type_string' to 'a_string'
		require
			string_exists: a_string /= Void
		do
			type_string := a_string.as_upper
			type_string.prune_all (' ')
		ensure
			type_string_set: type_string /= Void
		end

	set_offset (an_offset: INTEGER) is
			-- set 'offset' to 'an_offset'
		do
			offset := an_offset
		ensure
			offset_set: offset = an_offset
		end

	set_symbolic_offset (a_symbolic_offset: STRING) is
			-- set 'symbolic_offset' to 'a_symbolic_offset'
		do
			symbolic_offset := a_symbolic_offset
		ensure
			symbolic_offset_set: symbolic_offset = a_symbolic_offset
		end

	from_dictionary (a_dictionary_record: RAW_DICTIONARY_RECORD) is
			-- Set attributes of this field from a_dictionary_record
		require
			valid_record: a_dictionary_record /= Void
			valid_attribute: a_dictionary_record.attribute_name /= Void
			valid_type_name: a_dictionary_record.type_name /= Void
		do
			name := a_dictionary_record.attribute_name.as_lower
			name.prune_all (' ')
			type_string := a_dictionary_record.type_name.as_upper
			type_string.prune_all (' ')
			stored_length := a_dictionary_record.length
			index := a_dictionary_record.iseq
			if a_dictionary_record.interpreter /= Void then
				interpreter_type := a_dictionary_record.interpreter.twin
			end
			if a_dictionary_record.minim /= Void and then
				a_dictionary_record.minim.is_integer
			 then
				min_index := a_dictionary_record.minim.to_integer
				min_index_name := Void
				has_bounds := true
			else
				min_index := 0
				if a_dictionary_record.minim /= Void then
					min_index_name := a_dictionary_record.minim
					min_index_name.prune_all (' ')
					if min_index_name.is_empty then
						min_index_name := Void
					end
				end
			end
			if a_dictionary_record.maxim /= Void and then
				a_dictionary_record.maxim.is_integer
			 then
				max_index := a_dictionary_record.maxim.to_integer
				max_index_name := Void
				has_bounds := true
			else
				max_index := 0
				if a_dictionary_record.maxim /= Void then
					max_index_name := a_dictionary_record.maxim
					max_index_name.prune_all(' ')
					if max_index_name.is_empty then
						max_index_name := Void
					end
				end
			end
		end

feature -- Fields as output strings

	min_index_out: STRING is
			-- Min index as a string
		do
			if min_index_name = Void then
				Result := min_index.out
			else
				Result := min_index_name
			end
		ensure
			exists: Result /= Void
		end

	max_index_out: STRING is
			-- Max index as a string
		do
			if max_index_name = Void then
				Result := max_index.out
			else
				Result := max_index_name
			end
		ensure
			exists: Result /= Void
		end

	length_out: STRING is
			-- Length of the field as a string
		do
			Result := length.out
		ensure
			exists: Result /= Void
		end

	repeat_count_out: STRING is
			-- repeat_count as a string
		local
			l_min_index_name: STRING
			l_max_index_name: STRING
		do
			if uses_symbolic_indexes then
				if min_index_name /= Void then
					l_min_index_name := min_index_name
				else
					l_min_index_name := min_index.out
				end

				if max_index_name/=Void then
					l_max_index_name := max_index_name
				else
					l_max_index_name := max_index.out
				end

				Result := "(" + l_max_index_name + "-(" + l_min_index_name + ")+1)"
			else
				Result := repeat_count.out
			end
		ensure
			exists: Result /= Void
		end

	offset_out: STRING is
			-- Offset as a string
		do
			if symbolic_offset = Void then
				Result := offset.out
			else
				Result := symbolic_offset
			end
		ensure
			exists: Result /= Void
		end

	array_item_string: STRING is
			-- see field_type_string
		do
			Result := field_type_string
		ensure
			exists: Result /= Void
		end

	uses_symbolic_indexes: BOOLEAN is
			-- Does this field use symbolic indexes?
		do
			Result := min_index_name /= Void or max_index_name /= Void
		end

	attach_to_pointer (a_pointer: POINTER) is
			-- Attach this field to a pointer
		deferred
		end

feature -- Code generation

  	make_one_text: STRING is
  			-- code text to instantiate an object
  		do
  			Result := ""
		ensure
			valid_result: Result /= Void
  		end

	class_declaration: STRING is
			-- class declaration
		local
			l_class_name : STRING
			l_name : STRING
		do
			l_name := name.as_lower
			l_name.prune_all (' ')
			l_class_name := declaration_string
			if is_array then
				 l_class_name.append ("_RECORD_ARRAY")
			end
			Result := "%T" + l_name + " : " + l_class_name
		end

	declaration_string: STRING is
  			-- string for type declaration
  		do
  			if type_string.is_equal("CHARACTER") and length > 1 then
  				Result := "STRING"
  			else
  				Result := type_string
  			end
		ensure
			valid_result: Result /= Void
  		end

	read_text: STRING is
			-- unknown
		do
			Result := ""
		ensure
			valid_result: Result /= Void
		end

	deferred_set_text: STRING is
			-- The 'set_XXX' routine - deferred.
		do
			Result :=	"	set_" + name + "(a_other : " + declaration_string + ") is deferred end%N"
		end

	deferred_ith_text: STRING is
			-- text for deferred ith getter
		do
			Result := "	ith_" + name + "(i: INTEGER): " + declaration_string + " is deferred end%N"
		end

	deferred_set_ith_text: STRING is
			-- text for deferred ith setter
		do
			Result := "	set_ith_" + name + "(a_other: " + declaration_string + "i: INTEGER) is deferred end%N"
		end

	make_me_text: STRING is
			-- text to create the object
		do
			if is_array then
				Result := "		create " + name + ".make(" + min_index_out + "," + max_index_out + ")%N"
			else
				Result := make_one_text.twin
			end
		end

	array_set_text: STRING is
			-- text for an array setter
		do
			Result :=
				"			if " + name + " = Void then !!" + name + ".make(" + min_index_out + "," + max_index_out + ") end%N" +
				"			from i := " + min_index_out + " until i > " + max_index_out + "%N" +
				"			loop%N" +
				"				set_ith_" + name + "(a_other.ith_" + name + "(i),i)%N" +
				"				i := i+1%N" +
				"			end%N"
		end

	field_set_text : STRING is
			-- test for a field setter
		do
			Result :=	"			set_" + name + "(a_other." + name + ")%N"
		end

	deferred_field_text: STRING is
			-- text for a deferred field
		do
			Result := "	" + name + " : " + declaration_string + " is deferred end%N"
		end

feature -- Deferred code generation routines

	record_set_ith_text: STRING is deferred end
	record_ith_text: STRING is deferred end
	record_set_text: STRING is deferred end
	record_field_text: STRING is deferred end
	navigator_set_ith_text: STRING is deferred end
	navigator_ith_text: STRING is deferred end
	navigator_set_text: STRING is deferred end
	navigator_field_text: STRING is deferred end
	text_of_set_from (a_source: STRING): STRING is deferred end
	external_record_text: STRING is deferred end
	needs_make: BOOLEAN is deferred end
	display_specific_text: STRING is deferred end
	read_specific_text: STRING is deferred end
	write_specific_text: STRING is deferred end
	feature_body: STRING is deferred end
	
	array_type: STRING
			-- Name of Eiffel array class to be used in code generation
		deferred
		end

	field_type_string: STRING is
			-- Field type string
		deferred
		end

feature -- Display

	display is
			-- Display this component's value
		local
			l_string: STRING
			l_retry: BOOLEAN
			l_value: ANY
			l_message: STRING
			l_other_message: STRING
		do
			if not l_retry then
				-- Please don't join these print statements.This prevents crashes when
				-- a value is Void (that happened to me).
				l_message :=  "                " + padded_string(index.out, 3, True) + " -> "
				l_message.append (padded_string (name, 24, False))
				l_message.append (padded_string (type_string, 16, False))
				if length > 0 then
					l_message.append (" (")
					l_message.append (length_out)
					l_message.append (") ")
				end
				if is_array then
					l_message.append ("[")
					l_message.append (min_index_out)
					l_message.append (" : ")
					l_message.append (max_index_out)
					l_message.append ("] ")
				end
				if not is_valid then
					l_message.append ("<not valid>")
				elseif is_null then
					l_message.append ("<NULL>")
				else
					l_other_message := other_display
					if l_other_message/=Void
					then
						l_message.append(l_other_message)
					else
						l_value := value
						l_string ?= l_value
						if l_string /= Void and l_string.count > 200 then
							l_message.append (l_string.substring (1, 200))
						elseif l_value /= Void then
							l_message.append (l_value.out)
						end
					end
				end
				l_message.append ("%N")
				print (l_message)
			else
				print ("%NFailure during display:DATA_COMPONENT!!!!!!!!!%N")
			end
		rescue
			l_retry := True
			retry
		end

	other_display: STRING
	-- An alternative to 'display'.  Must be defined by the descendant to be of any use.
	do
	end

feature {NONE} -- Implementation

	padded_string (a_string: STRING; a_size: INTEGER; a_prepend: BOOLEAN): STRING is
			-- a_string either truncated to a_size or padded out with spaces to a_size
		require
			valid_size: a_size > 0
		do
			if a_string = Void then
				create Result.make_filled (' ',a_size)
			elseif a_string.count > a_size then
				Result := a_string.substring (1, a_size)
			else
				Result := a_string.twin
				from until Result.count >= a_size loop
					if a_prepend then
						Result.prepend (" ")
					else
						Result.append (" ")
					end
				end
			end
		end

	get_today: STRING is
			-- now value as a string
		local
			l_now: ROSE_DATE
		do
			create l_now.make_now
			Result := l_now.formatted_out("[0]mm/[0]dd/yyyy [0]hh:[0]mi:[0]ss")
		end

end
