note
	description: "Abstraction of a data server"
	last_updater: "$Author: rj $"
	author: "Pierre Nallet"
	last_editor: "$Author: rj $"
	date: "$Date: 2015/03/25 $"
	revision: "$Revision: #43 $"
	path: "$File: //rose/source/kernel/infrastructure/persistence/data_server.e $"

deferred class
	DATA_SERVER

inherit
	TRACEABLE
		rename
			status as t_status
		undefine
			clear_error
		redefine
			trace
		end
	DEBUG_OUTPUT
	PM_METRIC_SOURCE
	MOVIE_CAMERA_AWARE
		export
			{NONE} all
			{ANY} playback_movie, record_movie
		end
	ROSE_SHARED_GLOBAL
		rename
			trace as shared_trace,
			output as shared_output
		end

feature {NONE} -- Initialization

	make (a_node, a_name, a_passwd, a_database, a_tag: STRING) is
			-- Create the server
		deferred
		end

feature -- Access

	definition: ROSE_DATABASE_DEFINITION
			-- Definition used to connect to this data server

	logical: STRING is
			-- Logical name for the data-server
		do
			if definition /= Void then
				Result := definition.logical
			end
		end

	node: STRING
			-- Actual machine hosting the server

	name: STRING
			-- ?

	database: STRING
			-- Actual database name attached

	tag: STRING
			-- ?

	new_executor (a_string: STRING; a_parameter_template: ARRAY [STRING]): EXECUTOR is
			-- Create a new executor object on this data_server
		deferred
		end

	new_selector (a_string: STRING; a_parameter_template: ARRAY [STRING]): SELECTOR is
			-- Create a new selector object on this data_server
		deferred
		end

	is_still_valid: BOOLEAN
			-- Is connection to this data server still valid?
			-- And, is it still connected to the database that it claims to be connected to?
		deferred
		end

	property (a_key: STRING): STRING is
			-- Return value of property with 'a_key', if any
		deferred
		end

	last_error_text: STRING is
			-- Returns the text value of the last error.
		deferred
		end

	rows_affected: INTEGER is
			-- Number of rows affected by last query on this server object
		deferred
		end

	fast_loader (a_table_name: STRING; a_template: ANY): DATA_SERVER_FAST_LOAD is
			-- A new fast loader instance for this database on a_table_name
		require
			valid_table: a_table_name /= Void
		do
		end

	fast_loader_with_options (a_table_name: STRING; a_template: ANY; a_options: STRING): DATA_SERVER_FAST_LOAD is
			-- A new fast loader instance for this database on a_table_name using options a_options
		require
			valid_table: a_table_name /= Void
		do
		end

	fast_loader_keep_identity (a_table_name: STRING; a_template: ANY): DATA_SERVER_FAST_LOAD is
			-- A new fast loader instance for this database on a_table_name keeping identity values in inserted objects
		require
			valid_table: a_table_name /= Void
		do
		end

	stream_command (a_predicate_statement: STRING): DATA_SERVER_STREAM_COMMAND is
			-- A new stream command on this database using a_predicate_statement
		require
			valid_predicate_statement: a_predicate_statement /= Void
		do
		end

	simple_command (a_predicate_statement: STRING): DATA_SERVER_SIMPLE_COMMAND is
			-- A new simple command on this database using a_predicate_statement
		require
			valid_predicate_statement: a_predicate_statement /= Void
		do
		end

	debug_output: STRING
			-- String that should be displayed in debugger to represent `Current'.
		do
			create Result.make_empty
			Result.append_string (node)
			if database /= Void and then not database.is_empty then
				if Result.count /= 0 then
					Result.append_character ('/')
				end
				Result.append_string (database)
			end
		end

feature -- Status report

	make_succeeded: BOOLEAN
			-- Was the data server created successfully?

	reserved (a_string: STRING): BOOLEAN is
			-- Is a_string a reserved word for this data_server provider?
		deferred
		end

	duplicate_detected: BOOLEAN is
			-- Did the last execute result in a duplicate insert
		deferred
		end

	lock_detected: BOOLEAN is
			-- Did the last execute result in a lock conflict
		deferred
		end

	is_read_only: BOOLEAN is
			-- Is this data server read-only?
		require
			read_write_status_set: is_really_read_only or is_really_read_write
		do
			Result := is_really_read_only
		end

	is_read_write: BOOLEAN is
			-- Is the server read write?
		require
			read_write_status_set: is_really_read_only or is_really_read_write
		do
			Result := is_really_read_write
		end

	is_really_read_only: BOOLEAN
	is_really_read_write: BOOLEAN

	is_storable_date (a_date: ROSE_DATE): BOOLEAN is
			-- write a feature that checks to see if the given date is within the date range of a dataserver
		deferred
		end

	is_production: BOOLEAN is
			-- Is this data server flagged as production through the data server property
		local
			l_stored_mode, l_mode: STRING
		do
			if node /= Void and then database /= Void then
				l_mode := "production " + node + "." + database
				l_stored_mode := property ("mode")
				if l_stored_mode /= Void then
					Result := equal (l_stored_mode.as_lower, l_mode.as_lower)
				end
			end
		end

	status: INTEGER is
			-- Status code
		deferred
		end

feature -- Basic operations

	reconnect
			-- Reconnect to this data server
		deferred
		end

	set_definition (a_definition: ROSE_DATABASE_DEFINITION) is
			-- Override this server's definition
		require
			valid_definition: a_definition /= Void
		do
			definition:= a_definition
		end

	set_defintion_node_and_database(a_definition: ROSE_DATABASE_DEFINITION)
	-- Override the definition, node and database.
	require
		definition_not_void: a_definition/=Void
	do
		set_definition(a_definition)
		node := a_definition.server
		database := a_definition.path
	end

	set_error_handler (a_handler: ANY) is
			-- Set the error_handler to be used by this data_server
		deferred
		end

	trace is
			-- Print identifying information for this server
		local
			l_message: STRING
		do
			Precursor
			if node /= Void and database /= Void then
				l_message := "     Timestamp      : " + Global.interpreted ("$now") + "%N"
				l_message.append ("     Database       : " + node + "." + database + "%N")
				inform ("trace",l_message,Void)
				l_message.prepend (generating_type + ".trace%N")
				shared_output (l_message, Void)
			end
		end

	disconnect is
			-- Disconnect the underlying database connection.
		do
		end

feature -- Commit

	begin_transaction is
			-- Starts a new transaction
		deferred
		end

	commit is
			-- Commit changes to the server
		deferred
		end

	rollback is
			-- Rollback changes to the server
		deferred
		end

feature -- Read / Write

	read_only is
			-- Set the server to read only access
		do
			is_really_read_only := True
			is_really_read_write := False
		end

	read_only_with_options (a_options: STRING) is
			-- Set the server to read only access using options from a_options
		do
			read_only
		end

	read_write is
			-- Set the server to read write access
		do
			is_really_read_write := True
			is_really_read_only := False
		end

	read_write_with_options (a_options: STRING) is
			-- Set the server to read write access using options from a_options
		do
			read_write
		end

end










