indexing
	description: "[
		A date and time picker (DTP) control
		provides a simple and intuitive interface through
		which to exchange date and time information with a
		user. For example, with a DTP control you can ask
		the user to enter a date and then retrieve his or
		her selection with ease.
		This control requires version 4.70 and later of
		Comctl32.dll.  Ensure that WEL_COMMON_CONTROLS_DLL has
		been instantiate before creating an instance of this
		class.
	]"
	author: "David Jenkins"
	date: "$Date: 2008/12/01 $"
	revision: "$Revision: #6 $"

class
	DATE_TIME_PICKER_CONTROL

inherit
	WEL_CONTROL
		redefine
			process_notification_info,
			make_by_id
		end

	WEL_ICC_CONSTANTS
		export
			{NONE} all
		end

	DTS_CONSTANTS
		export
			{ANY} valid_dts_value
			{NONE} all
		end

	DTN_CONSTANTS
		export
			{NONE} all
		end

	GDTR_CONSTANTS
		export
			{NONE} all
		end

	GDT_CONSTANTS
		export
			{NONE} all
			{ANY} valid_gdt_value, gdt_none
		end

	MCSC_CONSTANTS
		export
			{NONE} all
			{ANY} valid_mcsc_value
		end

	DTM_CONSTANTS
		export
			{NONE} all
		end

create
	make,
	make_with_style,
	make_by_id

feature {NONE} -- Initialization

	make (a_parent: WEL_WINDOW; a_x, a_y, a_width, a_height, an_id: INTEGER) is
			-- Creation procedure.
		require
			a_parent_not_void: a_parent /= Void
			parent_exists: a_parent.exists
		do
			make_with_style (a_parent, a_x, a_y, a_width, a_height, an_id, Dts_shortdateformat)
		ensure
			exists: exists
			parent_set: parent = a_parent
			id_set: id = an_id
		end

	make_with_style (a_parent: WEL_WINDOW; a_x, a_y, a_width, a_height, an_id, a_style: INTEGER) is
		require
			a_parent_exists: a_parent /= Void
			valid_style: valid_dts_value (a_style)
		do
			register_class
			internal_window_make (a_parent, class_name,
				default_style + a_style, a_x, a_y, a_width, a_height,
				an_id, default_pointer);
			id := an_id
		ensure
			exists: exists
			parent_set: parent = a_parent
			id_set: id = an_id
		end

	make_by_id (a_parent: WEL_DIALOG; an_id: INTEGER) is
			-- Ensure that window class is registered
		do
			register_class
			Precursor (a_parent,an_id)
		end

	register_class is
			-- Ensure that Windows knows about this class
		local
			init_commctrl: WEL_INIT_COMMCTRL_EX
		once
			create init_commctrl.make_with_flags (Icc_date_classes)
			is_editable := True
		end

feature -- Access

	is_editable: BOOLEAN
			-- Is this date-field currently editable?

	month_cal_color (opt: INTEGER): WEL_COLOR_REF is
			-- Retrieves the color for a given portion of the
			-- month calendar within a date and time picker (DTP)
			-- control.
		require
			valid_option: valid_mcsc_value (opt)
		local
			c_ref: INTEGER
		do
			c_ref := {WEL_API}.send_message_result_integer (item, Dtm_getmccolor, to_wparam (opt), default_pointer)
			create Result.make_by_color (c_ref)
		end

	month_cal_font: WEL_FONT is
			-- Retrieves the font that the date and time picker (DTP) control's child month calendar control is currently using.
		local
			res: POINTER
		do
			res := {WEL_API}.send_message_result (item, Dtm_getmcfont, default_pointer, default_pointer)
			create Result.make_by_pointer (res)
		end

	min_date: DATE is
			-- The minimum date to which the control can be set.
			-- Returns Void if no minimum
		local
			syst_pair: WEL_SYSTEM_TIME_PAIR
			flag: INTEGER
		do
			create syst_pair.make
			flag := {WEL_API}.send_message_result_integer (item, Dtm_getrange, default_pointer, syst_pair.item)
			if flag & Gdtr_min /= 0 then
				Result := syst_pair.low_system_time.as_date
			end
		end

	max_date: DATE is
			-- The maximum date to which the control can be set.
			-- Returns Void if no maxiumum.
		local
			syst_pair: WEL_SYSTEM_TIME_PAIR
			flag: INTEGER
		do
			create syst_pair.make
			flag := {WEL_API}.send_message_result_integer (item, Dtm_getrange, default_pointer, syst_pair.item)
			if flag & Gdtr_max /= 0 then
				Result := syst_pair.high_system_time.as_date
			end
		end

	system_time: DATE_TIME is
			-- Retrieves the currently selected time
		require
			system_time_available: system_time_available
		local
			syst:MC_SYSTEM_TIME
		do
			create syst.make
			{WEL_API}.send_message (item, Dtm_getsystemtime, default_pointer, syst.item)
			Result := syst.as_date_time
		end

	system_time_available: BOOLEAN is
			-- Can system time be retrieved?
		local
			dummy:WEL_SYSTEM_TIME
		do
			create dummy.make
			Result := style & Dts_shownone = 0 or else
				{WEL_API}.send_message_result_integer (item, Dtm_getsystemtime, default_pointer, dummy.item) = Gdt_valid
		end

feature -- Basic operations

	enable_edit (a_on: BOOLEAN) is
		do
			if is_editable /= a_on then
				is_editable := a_on
			end
		end

	set_format (fs: STRING_GENERAL) is
			-- Sets the display based on the format string fs
			-- Calling with fs = Void resets control to default
			-- display.
		local
			l_wel_string: WEL_STRING
		do
			if fs = Void then
				{WEL_API}.send_message (item, Dtm_setformat, default_pointer, default_pointer)
			else
				create l_wel_string.make (fs)
				{WEL_API}.send_message (item, Dtm_setformat, default_pointer, l_wel_string.item)
			end
		end

	set_month_cal_color (i_color: INTEGER; clr: WEL_COLOR_REF) is
			-- Sets the color for a given region of the child
			-- month calendar control.
		require
			valid_region: valid_mcsc_value (i_color)
			clr_exists: clr /= Void
		do
			{WEL_API}.send_message (item, Dtm_setmccolor, to_wparam (i_color), to_lparam (clr.item))
		ensure
			color_set: month_cal_color (i_color).is_equal (clr)
		end

	set_mc_font (f: WEL_FONT; redraw_immed: BOOLEAN) is
			-- Set the Month Calendar font to f.
			-- Will immediately redraw depending on redraw_immed
		require
			font_exists: f /= Void
		local
			f_redraw:INTEGER
		do
			if redraw_immed then
				f_redraw := 1
			else
				f_redraw := 0
			end
			{WEL_API}.send_message (item, Dtm_setmcfont, f.item, to_lparam (f_redraw))
		end

	set_range (a_min, a_max: DATE) is
			-- Sets range of allowable dates
		local
			syst_pair: WEL_SYSTEM_TIME_PAIR
			l_min, l_max, d: DATE
		do
			if a_min = Void then
				create l_min.make (1, 1, 1)
			else
				create l_min.make (a_min.year, a_min.month, a_min.day)
			end
			if a_max = Void then
				create l_max.make (9999, 1, 1)
			else
				create l_max.make (a_max.year, a_max.month, a_max.day)
			end
			d := l_min.max (l_max)
			l_min := l_min.min (l_max)
			l_max := d
			create syst_pair.make
			syst_pair.low_system_time.set_date (l_min)
			syst_pair.high_system_time.set_date (l_max)
			{WEL_API}.send_message (item, Dtm_setrange, to_wparam (Gdtr_min + Gdtr_max), syst_pair.item)
		end

	remove_range is
		do
			set_range (Void, Void)
		end

	set_min_date (a_date: DATE) is
			-- Sets the minimum allowable date that can be entered, Void clears an eventually set minimum
		local
			syst_pair: WEL_SYSTEM_TIME_PAIR
		do
			if a_date /= Void then
				create syst_pair.make
				syst_pair.low_system_time.set_date (a_date)
				{WEL_API}.send_message (item, Dtm_setrange, to_wparam (Gdtr_min), syst_pair.item)
			else
				{WEL_API}.send_message (item, Dtm_setrange, to_wparam (Gdtr_min), default_pointer)
			end
		end

	set_max_date (a_date: DATE) is
			-- Sets the minimum allowable date that can be entered, Void clears an eventually set maximum
		local
			syst_pair: WEL_SYSTEM_TIME_PAIR
		do
			if a_date /= Void then
				create syst_pair.make
				syst_pair.high_system_time.set_date (a_date)
				{WEL_API}.send_message (item, Dtm_setrange, to_wparam (Gdtr_max), syst_pair.item)
			else
				{WEL_API}.send_message (item, Dtm_setrange, to_wparam (Gdtr_max), default_pointer)
			end
		end
	set_system_time (d_t:DATE_TIME;opt:INTEGER) is
			-- Sets the time
		require
			time_exists: style & Gdt_none = 0 implies d_t /= Void
			opt_valid: valid_gdt_value (opt)
		local
			syst: MC_SYSTEM_TIME
		do
			create syst.make_by_date_time (d_t)
			{WEL_API}.send_message (item, Dtm_setsystemtime, to_wparam (opt), syst.item)
		end

feature -- Notifications

	on_dtn_closeup (info: WEL_NMHDR) is
		require
			exists: exists
		do
		end

	on_dtn_datetimechange (info: WEL_NM_DATE_TIME_CHANGE) is
		require
			exists: exists
		do
		end

	on_dtn_dropdown (info: WEL_NMHDR) is
		require
			exists: exists
		do
		end

	on_dtn_format (info: WEL_NM_DATE_TIME_FORMAT) is
		require
			exists: exists
		do
		end

	on_dtn_format_query (info: WEL_NM_DATE_TIME_FORMAT_QUERY) is
		require
			exists: exists
		do
		end

	on_dtn_userstring (info: WEL_NM_DATE_TIME_STRING) is
			-- Sent by a date and time picker (DTP) control when
			-- a user finishes editing a string in the control.
			-- This notification message is only sent by DTP
			-- controls that are set to the DTS_APPCANPARSE style.
		require
			exists: exists
		do
		end

	on_dtn_wmkeydown (info: WEL_NM_DATE_TIME_WMKEYDOWN) is
			-- Sent by a date and time picker (DTP) control when
			-- the user types in a callback field.
		require
			exists: exists
		do
		end

feature {WEL_COMPOSITE_WINDOW} -- Implementation

	process_notification_info (notification_info: WEL_NMHDR) is
			-- Process a `notification_code' sent by Windows
			-- through the Wm_notify message
		local
			code: INTEGER;
			dtchg_info: WEL_NM_DATE_TIME_CHANGE
			dtfmt_info: WEL_NM_DATE_TIME_FORMAT
			dtfmtq_info: WEL_NM_DATE_TIME_FORMAT_QUERY
			dtstr_info: WEL_NM_DATE_TIME_STRING
			dtwmkd_info: WEL_NM_DATE_TIME_WMKEYDOWN
		do
			code := notification_info.code
			if code = dtn_closeup then
				on_dtn_closeup (notification_info)
			elseif code = dtn_datetimechange then
				create dtchg_info.make_by_nmhdr (notification_info)
				on_dtn_datetimechange (dtchg_info)
			elseif code = dtn_dropdown then
				on_dtn_dropdown (notification_info)
			elseif code = dtn_format then
				create dtfmt_info.make_by_nmhdr (notification_info)
				on_dtn_format (dtfmt_info)
			elseif code = dtn_formatquery then
				create dtfmtq_info.make_by_nmhdr (notification_info)
				on_dtn_format_query (dtfmtq_info)
			elseif code = dtn_userstring then
				create dtstr_info.make_by_nmhdr (notification_info)
				on_dtn_userstring (dtstr_info)
			elseif code = dtn_wmkeydown then
				create dtwmkd_info.make_by_nmhdr (notification_info)
				on_dtn_wmkeydown  (dtwmkd_info)
			end
		end

feature {NONE} -- Implementation

	default_style: INTEGER is
		once
			Result := Ws_border + Ws_child + Ws_visible
		end

	class_name: STRING_32 is
		local
			c_string: C_STRING_32
		once
			create c_string.make_by_pointer (cwin_datetimepick_class)
			Result := c_string.string
		end

feature {NONE} -- Externals

	cwin_datetimepick_class: POINTER is
		external
			"C[macro %"cctrl.h%"]:EIF_POINTER"
		alias
			"DATETIMEPICK_CLASS"
		end;

end
