note

	description:

		"Diagonal matrix - with specialized arithmetic to improve performance"

	author: "Michael Zhang"
	last_editor: "$Author: d3 $"
	date: "$Date: 2014/01/09 $"
	revision: "$Revision: #18 $"
	path: "$File: //rose/source/kernel/infrastructure/math/diagonal_matrix.e $"
	storable_version: ""

class DIAGONAL_MATRIX

inherit

	SYMMETRIC_MATRIX
		rename
			make_from_array as sm_make_from_area,
			test as sm_test
		redefine
			cloned,
			is_identity,
			is_same,
			put,
			item,
			make,
			make_identity,
			inverse,
			transpose_at,
			make_from_matrix
		end

create

	make,
	make_from_vector,
	make_from_matrix,
	make_identity,
	make_from_array,
	test

create {DIAGONAL_MATRIX}

	matrix_make,
	make_xtx

feature {NONE} -- Initialization

	make (a_size: INTEGER)
			-- Make zeroed diagonal matrix of (equal) dimensions `a_size'.
		do
			height := a_size
			width := a_size
			create data.make_filled (Void, 2)
			data.put (create {SPECIAL [DOUBLE]}.make_filled (0.0, width + 1), 1)
		ensure then
			height_set: height = a_size
			width_set: width = a_size
		end

	make_from_vector (a_vector: ABSTRACT_READABLE_VECTOR)
			-- Make with `a_vector'.
		require
			valid_vector: a_vector /= Void
		local
			i: INTEGER
		do
			make (a_vector.count)
			from
				i := 1
			until
				i > width
			loop
				put (a_vector.item (i), i, i)
				i := i + 1
			end
		end

	make_from_array (a_array: ARRAY [DOUBLE])
			--	Make with `a_array'.
		require
			valid_array: a_array /= Void
		local
			l_vector: VECTOR
		do
			create l_vector.make_from_array (a_array)
			make_from_vector (l_vector)
		end

	make_from_matrix (a_matrix: MATRIX)
			-- Make with diagonal entries of `a_matrix'.
		local
			i: INTEGER
		do
			make (a_matrix.height)
			from
				i := 1
			until
				i > height
			loop
				put (a_matrix [i, i], i, i)
				i := i + 1
			end
		end

	make_identity (a_size: INTEGER)
			-- Make as identity matrix of dimension `a_size'.
		local
			i: INTEGER
		do
			make (a_size)
			from
				i := 1
			until
				i > width
			loop
				put (1.0, i, i)
				i := i + 1
			end
		end

feature -- Access

	item alias "[]" (i, j: INTEGER): DOUBLE
			-- Matrix entry at intersection of row `i' and column `j'
		assign
			put
		do
			if	i = j then
				Result := data.item (1).item (i)
			else
				Result := 0.0
			end
		end

	cloned: like Current
			-- Copy of matrix optimised for non-zero entries only being on diagonal
		local
			i: INTEGER
		do
			create Result.make (height)
			from
				i := 1
			until
				i > width
			loop
				Result [i, i] := Current [i, i]
				i := i + 1
			end
		end

	inverse: like Current
			-- Inverse of diaogonal matrix with exception capture if zero on diagonal
		local
			i: INTEGER
			l_fail: BOOLEAN
		do
			create Result.make (width)
			from i := 1 until i > width or l_fail loop
				if Current [i, i] = 0.0 then
					l_fail := True
					raise ("Diagonal matrix element " + i.out + " is zero; Matrix cannot be inverted.")
				else
					Result [i, i] := 1.0 / Current [i, i]
				end
				i := i + 1
			end
		end

	approximate_inverse: like Current
			-- Inverse of diagonal matrix with zero entries mapped to zero cf. pseudo_inverse
		local
			i: INTEGER
			l_fail: BOOLEAN
		do
			create Result.make (width)
			from i := 1 until i > width or l_fail loop
				if Current [i, i] = 0.0 then
					Result [i, i] := 0.0
				else
					Result [i, i] := 1.0 / Current [i, i]
				end
				i := i + 1
			end
		end


feature -- Status

	is_identity: BOOLEAN
			-- Is `Current' close to an identity matrix?
		local
			i: INTEGER
		do
			Result := height = width
			from
				i := 1
			until
				i > height or not Result
			loop
				Result := Result and is_identity_item (i, i)
				i := i + 1
			end
		end

	is_same (a_other: ABSTRACT_READABLE_MATRIX): BOOLEAN is
			-- Is `a_other' numerically equal to `Current'?
		local
			i: INTEGER
		do
			if attached {DIAGONAL_MATRIX} a_other as l_other then
				Result := height = a_other.height and width = a_other.width
				from
					i := 1
				until
					i > height or not Result
				loop
					Result := Current [i, i] = a_other [i, i]
					i := i + 1
				end
			else
				Result := Precursor (a_other)
			end
		end

feature -- Arithmetic operations

	right_product (a_matrix: ABSTRACT_READABLE_MATRIX): MATRIX
			-- Product of `Current' with `a_matrix' placed on right,
			-- that is, D * M
		require
			valid_matrix: a_matrix /= Void
			valid_dim: a_matrix.height = width
		local
			i, j: INTEGER
			l_value, l_value_d: DOUBLE
		do
			create Result.make (a_matrix.height, a_matrix.width)
			from
				i := 1
			until
				i > a_matrix.height
			loop
				l_value_d := Current [i, i]
				from
				 	j := 1
				 until j > a_matrix.width
				 loop
					l_value := l_value_d * a_matrix [i, j]
					Result [i, j] := l_value
					j := j + 1
				end
				i := i + 1
			end
		ensure
			product_not_vold: Result /= Void
		end

	left_product (a_matrix: ABSTRACT_READABLE_MATRIX): MATRIX
			-- Product of `Current' with `a_matrix' placed on left,
			-- that is, M * D
		require
			valid_matrix: a_matrix /= Void
			valid_dim: a_matrix.width = height
		local
			i, j: INTEGER
			l_value, l_value_d: DOUBLE
		do
			create Result.make (a_matrix.height, a_matrix.width)
			from
				i := 1
			until
				i > a_matrix.width
			loop
				l_value_d := Current [i, i]
				from
					j := 1
				until
					j > a_matrix.height
				loop
					l_value := l_value_d * a_matrix [j, i]
					Result [j, i] := l_value
					j := j + 1
				end
				i := i + 1
			end
		ensure
			product_not_vold: Result /= Void
		end

	matrix_add (a_matrix: ABSTRACT_MATRIX): MATRIX
			-- Sum of `Current' with `a_matrix'
		require
			matrix_valid: a_matrix /= Void
			valid_width: a_matrix.width = width
			valid_height: a_matrix.height = height
		local
			i: INTEGER
		do
			create Result.duplicate (a_matrix)
			from
				i := 1
			until
				i > width
			loop
				Result [i, i] := Result [i, i] + Current [i, i]
				i := i + 1
			end
		ensure
			sum_not_vold: Result /= Void
		end

	transpose_at (a_vector: ABSTRACT_READABLE_VECTOR): VECTOR
			-- Vector product v^t D
		local
			i: INTEGER
		do
			create Result.make (width)
			from
				i := 1
			until
				i > width
			loop
				Result [i] := a_vector [i] * Current [i, i]
				i := i + 1
			end
		end

	xtDx (a_x: ABSTRACT_READABLE_MATRIX): SYMMETRIC_MATRIX
			-- Product of `a_x' transpose and `a_x' either side of `Current'
		require
			matrix_not_void: a_x /= Void
			same_size: a_x.height = height
		do
			create Result.make_xtDx (a_x, diagonal)
		end

	vtDv (a_v: ABSTRACT_READABLE_VECTOR): DOUBLE
			-- Norm generated from vector product of `a_v' either side of `Current'
		require
			vector_not_void: a_v /= Void
			same_size: a_v.count = width
		do
			Result := diagonal * a_v.squares
		end

feature {NONE} -- Implementation

	put (a_value: DOUBLE; i, j: INTEGER)
			-- Item `put' that ignores off-diagonal information
		do
			if i = j then
				data.item (1).put (a_value, i)
			end
		end

feature -- Test

	test is
			-- Test
		local
			l_dmatrix: like Current
			l_matrix, l_tm: MATRIX
		do
			make_identity(10)
			create l_dmatrix.make_from_array (<<1., 2.>>)
			create l_matrix.make_from_array (<< <<1., 2.>>, <<3., 4.>> >>)
			l_tm := l_dmatrix * l_matrix
		end

end
