note

	description:

		"ROSE_TIME_SERIES specialized to DOUBLE"

	author: "Michael Zhang"
	last_editor: "$Author: mb $"
	date: "$Date: 2011/08/24 $"
	revision: "$Revision: #21 $"
	path: "$File: //rose/source/kernel/math/double_time_series.e $"

class DOUBLE_TIME_SERIES

inherit

	ROSE_TIME_SERIES [DOUBLE]
		redefine
			make_from_array
		end

create

	make,
	make_empty,
	make_from_vector,
	make_from_array,
	make_with_dates,
	make_with_agent

feature {NONE} -- Irrelevant

	entity: ENTITY
			-- Unused and unwanted attribute

	variable_name: STRING
			-- Unused and unwanted attribute

feature {NONE} -- Initialization

	make_from_vector (a_vector: ABSTRACT_READABLE_VECTOR) is
			-- This is just to facilitate the use of value only features without dates.
			-- Which makes it useless for a time series!
		obsolete
			"Use `{ABSTRACT_READABLE_VECTOR}.to_array'."
		require
			a_vector_not_void: a_vector /= Void
		do
			make_from_array (a_vector.to_array)
			create dates.make_empty
		end

	make_from_array (a_array: ARRAY [DOUBLE]) is
			-- <Precursor>
		obsolete
			"Use `{ROSE_LINEAR_ARRAY}.make_from_array'."
		do
			Precursor (a_array)
			create dates.make_empty
		end

feature -- Conversion

	to_vector: VECTOR is
			-- Data items as a vector
		do
			create Result.make_from_array (Current)
		ensure
			to_vector_not_void: Result /= Void
			same_count: Result.count = count
		end

	to_row_vector: MATRIX is
			-- Data items as a single-row matrix
		local
			i, l_count: INTEGER
		do
			create Result.make (1, count)
			from
				i := 1
				l_count := count
			until
				i > l_count
			loop
				Result.put (item (i), 1, i)
				i := i + 1
			end
		ensure
			to_row_vector_not_void: Result /= Void
			single_row: Result.height = 1
			correct_column_count: Result.width = count
		end

	to_column_vector: MATRIX is
			-- Data items as a single-column matrix
		local
			i, l_count: INTEGER
		do
			create Result.make (count, 1)
			from
				i := 1
				l_count := count
			until
				i > l_count
			loop
				Result.put (item (i), i, 1)
				i := i + 1
			end
		ensure
			to_column_vector_not_void: Result /= Void
			single_column: Result.width = 1
			correct_row_count: Result.height = count
		end

feature -- Operators

	infix "-" (a_array: ARRAY [DOUBLE]): like Current is
			-- Copy of `Current' with defined items decremented by corresponding defined entries in `a_array'
		require
			a_array_not_void: a_array /= Void
			same_size: a_array.count = count
		local
			i, l_count: INTEGER
			l_double, l_item_1, l_item_2: DOUBLE
		do
			Result := cloned
			from
				i := 1
				l_count := count
			until
				i > l_count
			loop
				l_item_1 := item (i)
				l_item_2 := a_array.item (i)
				if is_defined_double (l_item_1) and is_defined_double (l_item_2) then
					l_double := l_item_1 - l_item_2
				else
					l_double := constants.missing_double
				end
				Result.put (l_double, i)
				i := i + 1
			end
		ensure
			difference_not_void: Result /= Void
			same_count: Result.count = count
			same_dates: Result.dates.same_items (dates)
			same_count_for_dates: Result.dates.count = dates.count
		end

	infix "/" (a_array: ARRAY [DOUBLE]): like Current is
			-- Copy of `Current' with defined items divided by corresponding defined entries in `a_array'
		require
			a_array_not_void: a_array /= Void
			same_size: a_array.count = count
		local
			i, l_count: INTEGER
			l_double, l_item_1, l_item_2: DOUBLE
		do
			Result := cloned
			from
				i := 1
				l_count := count
			until
				i > l_count
			loop
				l_item_1 := item (i)
				l_item_2 := a_array.item (i)
				if is_defined_double (l_item_1) and (is_defined_double (l_item_2) and then l_item_2 /= 0) then
					l_double := l_item_1 / l_item_2
				else
					l_double := constants.missing_double
				end
				Result.put (l_double, i)
				i := i + 1
			end
		ensure
			divided_series_not_void: Result /= Void
			same_count: Result.count = count
			same_dates: Result.dates.same_items (dates)
			same_count_for_dates: Result.dates.count = dates.count
		end

feature -- Access

	diff (a_n: INTEGER): like Current is
			-- Difference between initial and final sub-series of `Current'
		require
			a_n_strictly_positive: a_n > 0
			a_n_small_enough: a_n < count
			same_dates_count: count = dates.count
		do
			Result := subindex_series (a_n + 1, count) - lag_series (a_n)
		ensure
			diff_not_void: Result /= Void
			correct_count: Result.count = (count - a_n)
			correct_dates_count: Result.dates.count = Result.count
		end

	return_series (a_n: INTEGER): like Current is
			-- Ratio of difference elements of initial sub-series and final sub-series
			--  to final sub-series
		require
			fixed_a_n: a_n = (count - a_n)
		do
			Result := diff (a_n) / lag_series (a_n)
		ensure
			return_series_not_void: Result /= Void
			correct_count: Result.count = (count - a_n)
			correct_dates_count: Result.dates.count = Result.count
		end

	sum: DOUBLE is
			-- Sum of defined data elements of `Current'
		local
			i, l_count: INTEGER
		do
			from
				i := 1
				l_count := count
			until
				i > l_count
			loop
				if is_defined_double (item (i)) then
					Result := Result + item (i)
				end
				i := i + 1
			end
		ensure
			defined_sum: not constants.ieee.isnan (Result)
		end

	cumsum_series: like Current is
			-- Time series of partial sums of `Current'
		local
			i, l_count: INTEGER
			l_sum: DOUBLE
		do
			Result := cloned
			from
				i := 1
				l_count := count
			until
				i > l_count
			loop
				l_sum := l_sum + item (i)
				Result.put (l_sum, i)
				i := i + 1
			end
		ensure
			cumsum_series_not_void: Result /= Void
			same_count: Result.count = count
			same_dates: Result.dates.same_items (dates)
			same_count_for_dates: Result.dates.count = dates.count
		end

	cumret_series: like Current is
			-- Time series of cumulative returns from `Current'
		local
			i, l_count: INTEGER
			l_cumret: DOUBLE
		do
			Result := cloned
			from
				i := 1
				l_count := count
				l_cumret := 1.0
			until
				i > l_count
			loop
				l_cumret := l_cumret * (1.0 + item (i))
				Result.put (l_cumret - 1, i)
				i := i + 1
			end
		ensure
			cumret_series_not_void: Result /= Void
			same_count: Result.count = count
			same_dates: Result.dates.same_items (dates)
			same_count_for_dates: Result.dates.count = dates.count
		end

	interpolated_item_at (a_date: ROSE_DATE): DOUBLE is
			-- Linear interpolated value if value for `a_date' doesn't exist;
			-- Interpolation not done for start/end dates.
		require
			a_date_not_void: a_date /= Void
			ascending_order: is_sorted_ascending
		local
			l_val1, l_val2: DOUBLE
			l_days, l_days_interpolated: INTEGER
			l_date1, l_date2: ROSE_DATE
		do
			if a_date >= dates.first and then a_date <= dates.last then
				if has_date (a_date) then
					Result := item_at (a_date)
				else
					l_val1 := item_before (a_date)
					l_val2 := item_after (a_date)
					l_date1 := date_before (a_date)
					l_date2 := date_after (a_date)
					l_days := l_date2.diff_in_days (l_date1)
					l_days_interpolated := a_date.diff_in_days (l_date1)
					Result := l_val1 + (l_val2 - l_val1) * l_days_interpolated / l_days
				end
			end
		end

	subvector_by_date_range (a_start_date, a_end_date: ROSE_DATE): VECTOR is
			-- The vector of data between `a_start_date' and `a_end_date'
		require
			a_start_date_not_void: a_start_date /= Void
			a_end_date_not_void: a_end_date /= Void
			valid_range: a_start_date <= a_end_date
			ascending_order: is_sorted_ascending
		local
			l_start_ix, l_end_ix: INTEGER_32
		do
			l_start_ix := index_at (a_start_date)
			if l_start_ix = 0 then
				l_start_ix := index_after (a_start_date)
			end
			l_end_ix := index_at (a_end_date)
			if l_end_ix = 0 then
				l_end_ix := index_before (a_end_date)
			end
			if (l_start_ix > 0 and l_end_ix > 0) and then l_start_ix <= l_end_ix then
				create Result.make_from_array (subarray (l_start_ix, l_end_ix))
			end
		end

feature -- Summary statistics

	ljung_box (a_lag: INTEGER): DOUBLE is
			-- Ljung-Box statistic (for testing autocorrelations different from zero)
		require
			valid_lag: a_lag > 0 and a_lag < count
		local
			l_acf: VECTOR
			i: INTEGER
		do
			l_acf := acf (a_lag)
			from
				i := 1
			until
				i > a_lag
			loop
				Result := Result + l_acf.item (i) * l_acf.item (i) / (count - i)
				i := i + 1
			end
			Result := Result / (count * (count + 2))
		end

	acf (a_lag: INTEGER): VECTOR is
			-- Auto-correlations at `a_lag'
		require
			valid_lag: a_lag > 0 and a_lag < count
		local
			l_vector: VECTOR
		do
			l_vector := to_vector
			Result := l_vector.acf (a_lag)
		ensure
			acf_not_void: Result /= Void
			correct_count: Result.count = a_lag
		end

	moments: ROSE_MOMENTS is
			-- Unweighted moments
		do
			create Result.make (Current, Void)
		ensure
			moments_not_void: Result /= Void
		end

	weighted_moments (a_weights: ARRAY [DOUBLE]): ROSE_MOMENTS is
			-- Moments of data items `Current' weighted by `a_weights'
		require
			valid_weights: a_weights /= Void and then a_weights.count = count
		do
			create Result.make (Current, a_weights)
		ensure
			weighted_moments_not_void: Result /= Void
		end

feature -- Status report

	has_reversal (a_date: ROSE_DATE; a_positive_cut, a_negative_cut: DOUBLE): BOOLEAN is
			-- Can `Current' be split at `a_date' such that either:
			-- (1) the vector before (not in a temporal sense) `a_date' has a minimum datum
			--  at or below `a_negative_cut' and the vector starting from (not in a temporal sense)
			--  `a_date' has a maximum datum greater then `a_positive_cut', or
			-- (2) the vector before (not in a temporal sense) `a_date' has a maximum datum
			--  greater then `a_positive_cut' and the vector starting from (not in a temporal sense)
			--  `a_date' has a minimum datum at or below `a_negative_cut'
		require
			valid_cut: a_positive_cut > 0 and a_negative_cut < 0
		local
			l_left_vector, l_right_vector: VECTOR
			l_left_min_max, l_right_min_max: PAIR [DOUBLE, DOUBLE]
			l_ix: INTEGER
		do
			if has_date (a_date) then
				l_ix := index_at (a_date)
				if l_ix > 1 then
					l_left_vector := subarray (1, l_ix - 1).to_vector
					l_right_vector := subarray (l_ix, count).to_vector
					l_left_vector.replace_null
					l_right_vector.replace_null
					l_left_min_max := l_left_vector.min_max
					l_right_min_max := l_right_vector.min_max
					Result := ((l_left_min_max.first <= a_negative_cut) and (l_right_min_max.second > a_positive_cut)) or
						((l_left_min_max.second > a_positive_cut) and (l_right_min_max.first <= a_negative_cut))
				end
			end
		end

	has_null: BOOLEAN is
			-- Is any item not defined?
		local
			i: INTEGER
		do
			from
				i := 1
			until
				i > upper or Result
			loop
				Result := not is_defined_double (item (i))
				i := i + 1
			end
		end

	valid: BOOLEAN is
			-- Are all items of `Current' defined?
		do
			Result := not has_null
		ensure
			definition: Result = (not has_null)
		end

feature -- Basic operations

	truncate_at (a_min, a_max: DOUBLE) is
			-- Bound all defined elements within `a_min' and `a_max'.
		local
			i: INTEGER
			l_v: DOUBLE
		do
			from
				i := 1
			until
				i > upper
			loop
				l_v := item (i)
				if is_defined_double (l_v) then
					put (l_v.max (a_min).min (a_max), i)
				end
				i := i + 1
			end
		end

	squared is
			-- Square all defined items.
		local
			i: INTEGER
			l_v: DOUBLE
		do
			from
				i := 1
			until
				i > upper
			loop
				l_v := item (i)
				if is_defined_double (l_v) then
					put (l_v * l_v, i)
				end
				i := i + 1
			end
		end

end
