indexing
	description: "[
 		A questionable mechanism to get one's hands on a
semblance of the eiffel call stack. This should not be used under
normal circumstances, as it is hackish and very slow. If it is
used, it should only be inside a debug() block. At some point it
could be modified to have more direct access to the real stack.
		]"
	author: "Theo Lengyel"
	date: "$Date: 2010/06/14 $"
	revision: "$Revision: #2 $"
	archive: "$File: //rose/source/kernel/infrastructure/logging/eiffel_call_stack.e $"
	last_editor: "$Author: tg $"
	
class EIFFEL_CALL_STACK


feature -- Access
	
	call_stack: ARRAY[STRING] is
			-- create the call stack and remove the frames that were created within this class
		local
			l_exception: EIFFEL_CALL_STACK_EXCEPTION
		do
			if l_exception = Void then
				create l_exception
				l_exception.set_message( "eiffel_call_stack" )
				l_exception.raise
			else
				Result := adjusted_call_stack( parse_trace( l_exception ) )
			end
		ensure
			result_not_void: Result /= Void
		rescue
			if l_exception = exception_manager.last_exception then
				retry
			end
		end

feature {NONE} -- Implementation

	adjusted_call_stack ( a_stack: like call_stack ): like call_stack is
			-- the call stack we have produced, excluding the frames introduced by the calls within this class
		do
			Result := a_stack.subarray( a_stack.lower + call_stack_offset, a_stack.upper )
		end
	
	call_stack_offset: INTEGER is 5
			-- from the client perspective, the number of stack frames produced after the call to `call_stack'
	
	exception_manager: EXCEPTION_MANAGER is
			-- helper to determine if the exception was our helper exception
		once
			create Result
		end
	
	parse_trace( a_exception: DEVELOPER_EXCEPTION ): ARRAY[STRING] is
			-- convert the STRING stack trace into an array of frames
			-- FIXME -- this could probably be better optimized
			-- FIXME -- also, this is highly dependant on the format of the stack trace string, especially the "frame_separator"
		local
			l_lines: LIST[STRING]
			l_frame: ARRAY[STRING]
			l_line: STRING
			l_depth: INTEGER
			l_trace: STRING
			l_caller: STRING
		do
			l_depth := -1
			create Result.make(1,0)
			l_trace := a_exception.exception_trace
			l_lines := l_trace.split('%N')
			from
				l_lines.start
			until
				l_lines.off
			loop
				l_line := l_lines.item.string
				if is_frame_separator( l_line ) then
					l_depth := l_depth + 1
					if l_depth > 0 then
						if l_frame = Void then
							create l_frame.make(1,0)
						else
							l_caller := caller_from_frame( l_frame )
							Result.force( l_caller, Result.count + 1 )
							create l_frame.make(1,0)
						end
					end
				else
					if l_depth > 0 then
						l_frame.force( l_line, l_frame.count + 1 )
					end
				end
				l_lines.forth
			end
		end

	caller_from_frame ( a_frame: ARRAY[STRING] ): STRING is
			-- extract the class and feature from a flattened stack frame
		local
			l_tokens: ARRAY[STRING]
			l_class: STRING
			l_feature: STRING
			l_instruction: STRING
		do
			l_tokens := tokenize_frame( a_frame )
			l_class := l_tokens.item(1)
			l_feature := l_tokens.item(2)
			l_instruction := l_tokens.item(3)
			Result := l_class + "." + l_feature + ": " + l_instruction
		ensure
			result_not_void: Result /= Void
		end
	
	tokenize_frame ( a_frame: ARRAY[STRING] ): ARRAY[STRING] is
			-- take a stack frame that is separated into x number of lines and flatten it into a single array of tokens
		local
			l_line: STRING
			l_tokens: LIST[STRING]
			l_index: INTEGER
			l_result_index: INTEGER
		do
			from
				l_index := a_frame.lower
				create Result.make(1,0)
				l_result_index := 1
			until
				l_index > a_frame.upper
			loop
				l_line := a_frame.item(l_index).string
				l_line.prune_all_leading(' ')
				l_line.prune_all_trailing(' ')
				l_tokens := l_line.split(' ')
				l_tokens.compare_objects
				l_tokens.prune_all( Void )
				l_tokens.prune_all( empty_string )
				from
					l_tokens.start
				until
					l_tokens.off
				loop
					if not l_tokens.item_for_iteration.is_equal(empty_string) then
						Result.force(l_tokens.item_for_iteration, l_result_index )
						l_result_index := l_result_index + 1
					end
					l_tokens.forth
				end
				l_index := l_index + 1
			end
		ensure
			at_least_three_items: Result.count >= 3
		end

	is_frame_separator ( a_line: STRING ): BOOLEAN is
			-- is this line a frame_separator
		require
			a_line_not_void: a_line /= Void
		do
			Result := a_line.is_equal( frame_separator )
		end

	empty_string: STRING is ""
	frame_separator: STRING is "-------------------------------------------------------------------------------"
			-- frames on the stack are separated by this string
end	

-- Here is a sample call stack string:

--		-------------------------------------------------------------------------------
--		Class / Object      Routine                Nature of exception           Effect
--		-------------------------------------------------------------------------------
--		ISE_EXCEPTION_MANAGER
--		                    developer_raise        elog_severity:INFO:INFORMATION:
--		<0000000015D44000>                         Developer exception.          Fail
--		-------------------------------------------------------------------------------
--		ISE_EXCEPTION_MANAGER
--		                    developer_raise
--		<0000000015D44000>                         Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		ISE_EXCEPTION_MANAGER
--		                    raise @11
--		<0000000015D44000>                         Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		DEVELOPER_EXCEPTION raise @2
--		<00000000101D4E6C>  (From EXCEPTION)       Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		ELOG_LOGGER         forced_log @6
--		<0000000015D43220>                         Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		ELOG_LOGGER         log_information @3
--		<0000000015D43220>                         Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		ELOG_LOGGER         info @2
--		<0000000015D43220>                         Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		ACCESS_CONTROL_AGENT
--		                    force_double @1
--		<000000001479F548>  (From GENERIC_ACCESS_CONTROL_AGENT)
--		                                           Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		VALU                generated_double @9
--		<0000000015598810>  (From ROSE_EXPRESSION) Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		VALU                generate @3
--		<0000000015598810>  (From ROSE_EXPRESSION) Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		RCMD_VARIABLE_CALCULATION_VALREG
--		                    compute_all_valreg_variable @23
--		<000000001599AFE0>                         Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		RCMD_VARIABLE_CALCULATION_VALREG
--		                    do_execute @27
--		<000000001599AFE0>                         Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		RCMD_VARIABLE_CALCULATION_VALREG
--		                    protected_execute @3
--		<000000001599AFE0>  (From ROSE_COMMAND_PROCESS)
--		                                           Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		RCMD_VARIABLE_CALCULATION_VALREG
--		                    execute_prepared @10
--		<000000001599AFE0>  (From ROSE_COMMAND)    Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		RCMD_MACRO          do_execute @4
--		<000000001599AD04>                         Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		RCMD_MACRO          protected_execute @3
--		<000000001599AD04>  (From ROSE_COMMAND)    Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		RCMD_MACRO          execute_prepared @10
--		<000000001599AD04>  (From ROSE_COMMAND)    Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		RCMD_EXECUTE        execute_line @13
--		<000000001599A764>                         Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		RCMD_EXECUTE        procede_input @12
--		<000000001599A764>                         Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		RCMD_EXECUTE        do_execute @15
--		<000000001599A764>                         Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		RCMD_EXECUTE        protected_execute @3
--		<000000001599A764>  (From ROSE_COMMAND)    Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		RCMD_EXECUTE        execute_prepared @10
--		<000000001599A764>  (From ROSE_COMMAND)    Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		RCMD_MACRO          do_execute @4
--		<000000001599A488>                         Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		RCMD_MACRO          protected_execute @3
--		<000000001599A488>  (From ROSE_COMMAND)    Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		RCMD_MACRO          execute_prepared @10
--		<000000001599A488>  (From ROSE_COMMAND)    Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		ROSE_COMMAND_MANAGER
--		                    execute @3
--		<0000000015992574>                         Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		RCMD_ALIAS_MACRO    do_execute @2
--		<00000000159913B0>                         Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		RCMD_ALIAS_MACRO    protected_execute @3
--		<00000000159913B0>  (From ROSE_COMMAND)    Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		RCMD_ALIAS_MACRO    execute_prepared @10
--		<00000000159913B0>  (From ROSE_COMMAND)    Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		RCMD_MACRO          do_execute @4
--		<000000001599108C>                         Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		RCMD_MACRO          protected_execute @3
--		<000000001599108C>  (From ROSE_COMMAND)    Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		RCMD_MACRO          execute_prepared @10
--		<000000001599108C>  (From ROSE_COMMAND)    Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		ROSE_PILOT          run_pilot @2
--		<0000000015990054>  (From ROSE_PILOT_BASE) Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		ROSE_PILOT          run_application @15
--		<0000000015990054>  (From ROSE_PILOT_BASE) Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		ROSE_PILOT          run @1
--		<0000000015990054>  (From ROSE_PILOT_BASE) Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		VARIABLE_CALCULATION_PILOT
--		                    make @3
--		<0000000011E92478>                         Routine failure.              Fail
--		-------------------------------------------------------------------------------
--		VARIABLE_CALCULATION_PILOT
--		                    root's creation
--		<0000000011E92478>                         Routine failure.              Exit
--
