indexing
	description: "Helper class for manipulating/traversing directories"
	author: "Zoran Simic"
	date: "$Date: 2011/09/20 $"
	revision: "$Revision: #22 $"
	last_editor: "$Author: bn $"
	path: "$File: //rose/source/kernel/infrastructure/framework/support/el_directory.e $"

class
	EL_DIRECTORY

inherit
	EL_FILEDIR
		redefine
			make, reset, size
		end

create
	make, make_respect_case

feature {NONE} -- Initialization

	make (a_path: STRING) is
			-- Associate to 'a_path' (by default all paths are lowercased).
		do
			path := global.convention.clean_directory_path (canonical_form (a_path))
			if path.count > 0 and then path.item (path.count) = '\' then
				basename := global.formatter.file_name (path.substring (1, path.count - 1))
			else
				basename := global.formatter.file_name (path)
			end
		end

feature -- Access

	is_writable: BOOLEAN is
			-- Is this file writeable by current process?
		local
			l_dir: DIRECTORY
			retried: BOOLEAN
		do
			if not retried then
				create l_dir.make (path)
				Result := l_dir.exists and then l_dir.is_writable
			end
		rescue
			if not retried then
				retried := True
				retry
			end
		end

	subdirs: ROSE_LINEAR_ARRAY [like Current] is
			-- Sub-directories
		do
			auto_refresh
			Result := subdirs_i
		end

	files: ROSE_LINEAR_ARRAY [EL_FILE] is
			-- Files in this folder
		do
			auto_refresh
			Result := files_i
		end

	subdir (a_basename: STRING): like Current is
			-- File within this directory with 'a_basename', if any
		do
			if a_basename /= Void then
				auto_refresh
				Result := find_subdir (canonical_form (a_basename))
			end
		end

	file (a_basename: STRING): EL_FILE is
			-- File within this directory with 'a_basename', if any
		do
			if a_basename /= Void then
				auto_refresh
				Result := find_file (canonical_form (a_basename))
			end
		end

	files_with_extension (a_extension: STRING): ROSE_LINEAR_ARRAY [EL_FILE] is
			-- All files with 'a_extension' in this directory
		require
			extension_ok: a_extension /= Void and then not a_extension.is_empty
			starts_with_dot: a_extension.item (1) = '.'
		local
			i: INTEGER
			l_ext: STRING
		do
			create Result.make_empty
			auto_refresh
			if files_i /= Void then
				l_ext := a_extension.as_lower
				from
					i := files_i.lower
				until
					i > files_i.upper
				loop
					if l_ext.is_equal (files_i.item (i).extension.as_lower) then
						Result.extend (files_i.item (i))
					end
					i := i + 1
				end
			end
		ensure
			result_ok: Result /= Void
		end

	size: INTEGER_64 is
			-- Size of this directory, including subdirectories
		local
			l_info: UNIX_FILE_INFO
		do
			create l_info.make
			Result := get_recursive_size (l_info, path)
		end

feature -- Basic operations

	reset is
			-- Reset this object's state.
		do
			Precursor
			subdirs_h := Void
			files_h := Void
			subdirs_i := Void
			files_i := Void
		end

	copy_from (a_other: like Current; a_callback: FUNCTION [ANY, TUPLE [STRING, INTEGER, INTEGER], BOOLEAN]): ROSE_STATUS is
			-- Copy contents of 'a_other' directory into current's
		require
			other_ok: a_other /= Void
		local
			l_report: like progress_report
		do
			create Result
			a_other.refresh
			if not a_other.exists then
				Result.errorf ("Can't copy contents of non existant directory '$1'", <<a_other.path>>)
			else
				Result.append (create_if_needed)
				if not exists then
					Result.errorf ("Could not create directory '$1'", <<path>>)
				else
					l_report := progress_report (a_callback)
					l_report.dirs.extend (path)
					l_report.recursive_gather (a_other.path)
					if l_report.cancel then
						Result.warnf ("Did not copy directory '$1' to '$2' (cancelled after scanning $3 items)", <<a_other.path, path, l_report.total>>)
					else
						recursive_copy (a_other, l_report)
						if l_report.cancel and then l_report.done /= l_report.total then
							Result.warnf ("Did not copy directory '$1' to '$2' (cancelled after copying $3/$4 items)", <<a_other.path, path, l_report.done, l_report.total>>)
						end
						reset
						Result.append (l_report.status)
						if Result.is_acceptable then
							if l_report.files /= Void and then l_report.dirs /= Void then
								Result.outputf ("Copied $1 files and $2 folders from $3 to $4", <<l_report.files.count, l_report.dirs.count, a_other.path, path>>)
							else
								Result.outputf ("Copied $1 items from $2 to $3", <<l_report.done, a_other.path, path>>)
							end
						end
					end
				end
			end
		end

	create_if_needed: ROSE_STATUS is
			-- Create directory with 'a_path' if not already created
		do
			create Result
			refresh
			if not exists then
				global.formatter.recursive_create_directory (path)
				refresh
				if exists then
					Result.outputf ("Created '$1'", <<path>>)
				else
					Result.errorf ("Could not create '$1'", <<path>>)
				end
			end
		ensure
			result_ok: Result /= Void
		end

	delete (a_callback: FUNCTION [ANY, TUPLE [STRING, INTEGER, INTEGER], BOOLEAN]; a_retries: INTEGER): ROSE_STATUS is
			-- Delete this directory, return a status report (will report eventual errors)
			-- 'a_callback' is called for each directory deleted, the 2 integers inicate
			-- the number of files/dirs deleted so far, and the total respectively
			-- 'a_retries' indicates how many times to retry deleting files that can't succesfully be deleted right away
		local
			l_report: like progress_report
		do
			create Result
			if exists then
				l_report := progress_report (a_callback)
				l_report.dirs.extend (path)
				l_report.recursive_gather (path)
				if l_report.cancel then
					Result.warnf ("Did not delete directory '$1' (cancelled after scanning $2 items)", <<path, l_report.total>>)
				else
					l_report.recursive_delete (a_retries)
					if l_report.cancel and then l_report.done /= l_report.total then
						Result.warnf ("Did not delete directory '$1' (cancelled after deleting $2/$3 items)", <<path, l_report.done, l_report.total>>)
					end
					reset
					Result.append (l_report.status)
				end
				if Result.is_ok then
					Result.outputf ("Deleted directory '$1' ($2 items)", <<path, l_report.total>>)
				end
			else
				Result.errorf ("Can not delete '$1', directory does not exist", <<path>>)
			end
		end

	delete_all_contents (a_callback: FUNCTION [ANY, TUPLE [STRING, INTEGER, INTEGER], BOOLEAN]; a_retries: INTEGER): ROSE_STATUS is
			-- Delete all contents of this directory (without deleting the directory itself)
		local
			l_report: like progress_report
		do
			create Result
			if exists then
				l_report := progress_report (a_callback)
				l_report.recursive_gather (path)
				if l_report.cancel then
					Result.warnf ("Did not delete directory '$1' (cancelled after scanning $2 items)", <<path, l_report.total>>)
				else
					l_report.recursive_delete (a_retries)
					if l_report.cancel and then l_report.done /= l_report.total then
						Result.warnf ("Did not delete directory '$1' (cancelled after deleting $2/$3 items)", <<path, l_report.done, l_report.total>>)
					end
					reset
					Result.append (l_report.status)
				end
				if Result.is_ok and then l_report.total > 0 then
					Result.outputf ("Deleted all contents of directory '$1' ($2 items)", <<path, l_report.total>>)
				end
			else
				Result.errorf ("Can not delete '$1', directory does not exist", <<path>>)
			end
		end

	change_name (a_new_name: STRING): !ROSE_STATUS is
			-- Rename this directory to 'a_new_name'
		require
			name_ok: a_new_name /= Void
		local
			l_dir: DIRECTORY
			retried: INTEGER
		do
			if retried > 0 then
				Result.errorf ("Could not rename $1 to $2", <<path, a_new_name>>)
			else
				create Result
				create l_dir.make (path)
				if l_dir.exists then
					l_dir.change_name (a_new_name)
					Result.outputf ("Renamed '$1' to '$2'", <<path, a_new_name>>)
				else
					Result.errorf ("Could not rename $1 to $2: file does not exist", <<path, a_new_name>>)
				end
			end
		rescue
			retried := retried + 1
			retry
		end

feature {EL_DIRECTORY} -- Implementation

	recursive_copy (a_other: like Current; a_report: like progress_report) is
			-- Delete all items in 'a_report.dirs' and 'a_report.files', call 'a_report.callback' for progress feedback
		local
			l_dirs: ROSE_LINEAR_ARRAY [like Current]
			l_files: ROSE_LINEAR_ARRAY [EL_FILE]
			i: INTEGER
		do
			if not a_report.cancel then
				l_dirs := a_other.subdirs
				if l_dirs /= Void then
					from
						i := l_dirs.lower
					until
						a_report.cancel or else i > l_dirs.upper
					loop
						copy_dir (a_report, l_dirs.item (i))
						i := i + 1
					end
				end
			end
			if not a_report.cancel then
				l_files := a_other.files
				if l_files /= Void then
					from
						i := l_files.lower
					until
						a_report.cancel or else i > l_files.upper
					loop
						copy_file (a_report, l_files.item (i))
						i := i + 1
					end
				end
			end
		end

feature {NONE} -- Implementation

	subdirs_h: HASH_TABLE [like Current, STRING]	-- Subdirs hash
	files_h: HASH_TABLE [EL_FILE, STRING]			-- Files hash
	subdirs_i: ROSE_LINEAR_ARRAY [like Current]		-- Subdirs array
	files_i: ROSE_LINEAR_ARRAY [EL_FILE]			-- Files array
	callback_threshold: INTEGER is 50				-- Threshold for calling back 'callback'

	does_exist: BOOLEAN is
			-- Does this item exist on disk?
		do
			Result := file_system.directory_exists (path)
		end

	scan (a_info: UNIX_FILE_INFO) is
			-- Scan this directory for contents.
		local
			l_new_subdirs_h: like subdirs_h
			l_new_files_h: like files_h
			l_new_subdirs_i: like subdirs_i
			l_new_files_i: like files_i
			l_entry, l_path: STRING
			l_this_dir: DIRECTORY
			l_dir: like Current
			l_file: EL_FILE
			retried: BOOLEAN
		do
			if not retried then
				from
					create l_new_subdirs_h.make (0)
					create l_new_files_h.make (0)
					create l_new_subdirs_i.make_empty
					create l_new_files_i.make_empty
					create l_this_dir.make (path)
					l_this_dir.open_read
					l_this_dir.start
					l_this_dir.readentry
					l_entry := l_this_dir.lastentry
				until
					l_entry = Void
				loop
					l_entry := canonical_form (l_entry)
					if is_valid_directory_entry (l_entry) then
						l_path := file_system.pathname (path, l_entry)
						a_info.update (l_path)
						if a_info.is_directory then
							l_dir := find_subdir (l_entry)
							if l_dir = Void then
								create l_dir.make_respect_case (a_info.file_name, is_case_respected)
							end
							l_new_subdirs_h.put (l_dir, l_entry)
							l_new_subdirs_i.extend (l_dir)
						else
							l_file := find_file (l_entry)
							if l_file = Void then
								create l_file.make_respect_case (a_info.file_name, is_case_respected)
							end
							l_new_files_h.put (l_file, l_entry)
							l_new_files_i.extend (l_file)
						end
					end
					l_this_dir.readentry
					l_entry := l_this_dir.lastentry
				end
				l_this_dir.close
				subdirs_h := l_new_subdirs_h
				files_h := l_new_files_h
				subdirs_i := l_new_subdirs_i
				files_i := l_new_files_i
			end
		rescue
			if not retried then
				retried := True
				retry
			end
		end

	find_subdir (a_basename: STRING): like Current is
			-- Subdir with 'a_basename', if any
		require
			basename_ok: a_basename /= Void
		do
			if subdirs_h /= Void then
				Result := subdirs_h.item (a_basename)
			end
		end

	find_file (a_basename: STRING): EL_FILE is
			-- File with 'a_basename', if any
		require
			basename_ok: a_basename /= Void
		do
			if files_h /= Void then
				Result := files_h.item (a_basename)
			end
		end

	get_recursive_size (a_info: UNIX_FILE_INFO; a_path: STRING): INTEGER_64 is
			-- Compute the size of the directory at 'a_path'
		require
			info_ok: a_info /= Void
			path_ok: a_path /= Void
		local
			l_dir: DIRECTORY
			l_entry, l_path: STRING
			retried: BOOLEAN
		do
			if not retried then
				from
					create l_dir.make (a_path)
					l_dir.open_read
					l_dir.start
					l_dir.readentry
					l_entry := l_dir.lastentry
				until
					l_entry = Void
				loop
					if is_valid_directory_entry (l_entry) then
						l_path := file_system.pathname (a_path, l_entry)
						a_info.update (l_path)
						if a_info.is_directory then
							Result := Result + get_recursive_size (a_info, l_path)
						else
							Result := Result + a_info.size
						end
					end
					l_dir.readentry
					l_entry := l_dir.lastentry
				end
				l_dir.close
			end
		ensure
			result_ok: Result >= 0
		rescue
			if not retried then
				retried := True
				retry
			end
		end

	progress_report (a_callback: FUNCTION [ANY, TUPLE [STRING, INTEGER, INTEGER], BOOLEAN]): ROSE_DIRECTORY_SCAN_REPORT is
			-- Progress report
		do
			create Result.make_with_callback (a_callback)
		end

	copy_dir (a_report: like progress_report; a_other: like Current) is
			-- Copy 'a_other' to this directory.
		require
			report_ok: a_report /= Void
			other_ok: a_other /= Void
		local
			l_dir: like Current
		do
			a_report.increment_done (1, a_other.basename)
			create l_dir.make_respect_case (global.formatter.composed_path (<<path, a_other.basename>>), is_case_respected)
			l_dir.create_if_needed.do_nothing
			if l_dir.exists then
				l_dir.recursive_copy (a_other, a_report)
			else
				a_report.status.errorf ("Could not create directory '$1'", <<l_dir>>)
			end
		end

	copy_file (a_report: like progress_report; a_file: EL_FILE) is
			-- Copy 'a_file' to this directory.
		require
			report_ok: a_report /= Void
			file_ok: a_file /= Void
		do
			a_report.increment_done (1, a_file.basename)
			global.formatter.file_system.copy_file (a_file.path, global.formatter.composed_path (<<path, a_file.basename>>))
		end

	is_valid_directory_entry (a_entry: STRING): BOOLEAN is
			-- Is 'a_entry' to be further examined?
		do
			if a_entry.count > 0 then
				if a_entry.item (1) = '.' then
					if a_entry.count = 2 then
						Result := a_entry.item (2) /= '.'
					else
						Result := a_entry.count > 2
					end
				else
					Result := True
				end
			end
		end

end
