note

	description:

		"Helper class for manipulating files"

	author: "Zoran Simic"
	last_editor: "$Author: an $"
	date: "$Date: 2014/09/11 $"
	revision: "$Revision: #20 $"
	path: "$File: //rose/source/kernel/infrastructure/framework/support/el_file.e $"

class
	EL_FILE

inherit
	EL_FILEDIR
		rename
			delete as delete_with_callback
		redefine
			make
		end

	ROSE_IO_CONVENTION
		export
			{NONE} all
		undefine
			is_equal
		end

create
	make, make_respect_case

feature {NONE} -- Initialization

	make (a_path: STRING) is
			-- Make associated to 'a_path' (by default all paths are lowercased).
		do
			Precursor (a_path)
			name := basename.twin
			extension := file_system.extension (basename)
			if extension.count > 0 then
				name.keep_head (name.count - extension.count)
			end
		end

feature -- Access

	name: STRING
			-- Filename without full path nor extension, example: "file"

	extension: STRING
			-- Extension of this file (include the leading '.'), example: ".ext"

	is_writable: BOOLEAN is
			-- Is this file writeable by current process?
		local
			l_file: RAW_FILE
			retried: BOOLEAN
		do
			if not retried then
				create l_file.make (path)
				Result := (l_file.exists and then l_file.is_writable) or else l_file.is_creatable
			end
		rescue
			if not retried then
				retried := True
				retry
			end
		end

	contents: STRING is
			-- Contents of this file as a string
		local
			retried: BOOLEAN
			l_file: PLAIN_TEXT_FILE
		do
			if not retried and then exists then
				create l_file.make (path)
				if l_file.exists and then l_file.is_readable then
					from
						create Result.make (4096)
						l_file.open_read
					until
						l_file.end_of_file
					loop
						l_file.read_line
						Result.append_string (l_file.last_string)
						Result.append_character ('%N')
					end
					l_file.close
				end
			else
				Result := Void
			end
		rescue
			if not retried then
				retried := True
				retry
			end
		end

	size: INTEGER_64 is
			-- Size of this file, in bytes
		local
			l_info: UNIX_FILE_INFO
			retried: BOOLEAN
		do
			if not retried and then does_exist then
				create l_info.make
				l_info.update (path)
				Result := l_info.size
			end
		rescue
			if not retried then
				retried := True
				retry
			end
		end

feature -- Basic operations

	set_modified (a_timestamp: INTEGER) is
			-- Set last modification date with `a_timestamp'.
		local
			l_file: RAW_FILE
		do
			create l_file.make (path)
			l_file.set_date (a_timestamp)
			l_file.stamp (a_timestamp)
		end

	set_modified_date_from_date (a_date : ROSE_DATE) is
			-- Set last modification date with `a_date'
		require
			a_date_not_void : a_date /= Void
		do
			set_modified (a_date.unix_seconds.to_integer)
		end

	set_writable is
			-- Make this file writable
		local
			l_file: RAW_FILE
			retried: BOOLEAN
		do
			if not retried then
				create l_file.make (path)
				if l_file.exists then
					l_file.change_mode (219)		-- 0c333
				end
			end
		rescue
			if not retried then
				retried := True
				retry
			end
		end

	save (a_contents: STRING): !ROSE_STATUS is
			-- Save 'a_contents' to this file.
			-- Returns Void if saved OK
			-- Returns error message if could not save.
		require
			contents_ok: a_contents /= Void
		local
			rescued: STRING
			l_file: PLAIN_TEXT_FILE
		do
			if rescued /= Void then
				Result.errorf ("Crashed while saving file '$1':%N$2", <<path, rescued>>)
			else
				create Result
				create l_file.make (path)
				if l_file.exists and then not l_file.is_writable then
					Result.errorf ("File '$1' is not writeable%NPlease make sure that you have priviliges to write it, and that it is not set as read-only", <<path>>)
				elseif not l_file.is_creatable then
					Result.errorf ("Can not create file '$1'%NPlease make sure that the path is correct and/or you have privileges to write to that folder", <<path>>)
				else
					l_file.open_write
					l_file.put_string (a_contents)
					l_file.close
					Result.outputf ("Saved $1 bytes to '$2'", <<a_contents.count, path>>)
				end
			end
		rescue
			if rescued = Void then
				rescued := last_exception_explanation
				retry
			end
		end

	delete (a_retries: INTEGER): ROSE_STATUS is
			-- Delete this file, return a status report (will report eventual errors)
		local
			retried: INTEGER
		do
			if retried <= a_retries then
				create Result
				tmp_file.make (path)
				if tmp_file.exists then
					tmp_file.change_mode (219)		-- 0c333
					tmp_file.delete
					modified_i := 0
				end
			end
			if retried > a_retries or else exists then
				Result.errorf ("Could not delete '$1', please check that this file is deletable by you", <<path>>)
			end
		rescue
			if retried <= a_retries then
				retried := retried + 1
				retry
			end
		end

	copy_to (a_destination: EL_FILE): ROSE_STATUS is
			-- Copy contents of this file to 'a_destination' file, if possible
		require
			destination_ok: a_destination /= Void
		local
			l_c_source, l_c_dest: C_STRING
			l_codes: WINDOWS_SYSTEM_CODES
			l_code: INTEGER
		do
			create Result
			if not exists then
				Result.errorf ("Can not copy '$1' to '$2': source does not exist", <<path, a_destination.path>>)
			elseif not is_readable then
				Result.errorf ("Can not copy '$1' to '$2': source is not readable", <<path, a_destination.path>>)
			elseif a_destination.exists and then not a_destination.is_writable then
				Result.errorf ("Can not copy '$1' to '$2': destination is not writable", <<path, a_destination.path>>)
			elseif not a_destination.exists and then not a_destination.parent.is_writable then
				Result.errorf ("Can not copy '$1' to '$2': destination folder is not writable", <<path, a_destination.path>>)
			else
				create l_c_source.make (path)
				create l_c_dest.make (a_destination.path)
				if c_copy_file (l_c_source.item, l_c_dest.item) then
					Result.outputf ("Copied '$1' to '$2'", <<path, a_destination.path>>)
				else
					create l_codes
					l_code := l_codes.GetLastError
					if l_code = 0 then
						Result.errorf ("Can not copy '$1' to '$2'", <<path, a_destination.path>>)
					else
						Result.errorf ("Can not copy '$1' to '$2': [$3] $4", <<path, a_destination.path, l_code, l_codes.meaning (l_code)>>)
					end
				end
			end
		ensure
			result_ok: Result /= Void
		rescue
			Result.errorf ("Can not copy '$1' to '$2': operation crashed", <<path, a_destination.path>>)
		end

	archive_with_timestamp (a_default_archive_subfolder, a_new_name: detachable STRING): ROSE_STATUS
			-- Archive current file with a timestamped name.
			-- Move this file to an archive folder as defined below:
			--  . if the current file is in a directory tree structure containing
			--    a file 'site.cfg' which has an entry for "/path/deployment_archive", then
			--    use this archive path;
			--  . otherwise use the relative path `a_default_archive_subfolder' if not void;
			--  . or leave this file in its current folder.
			-- The archive path can be:
			--  . an absolute pathname of the form "\\axaros.com\shares\archive\production",
			--    in which case the file will be copied to the corresponding tree structure
			--    using the same relative path as the one leading to the directory containing
			--    the file 'site.cfg';
			--  . a relative pathname of the form "archive", in which case the file will be
			--    copied to this subfolder of the directory containing this file.
			-- 'a_new_name' (if provided) is modified to contain the new archive path.
			--
			-- Note: Rename with timestamp and then move to archive folder, otherwise renaming
			-- and moving in a single operation may fail if the file is currently being used.
		local
			l_path_to_site: detachable STRING
			l_config: ROSE_CONFIGURATION
			l_folder: detachable STRING
			l_dirname: STRING
			l_site_dirname: STRING
			l_relative_path: STRING
		do
			l_path_to_site := deliveries.path_to_site_configuration (path)
			if l_path_to_site /= Void then
				create l_config.make_default
				Global.app.load_config (l_config, l_path_to_site)
				l_folder := l_config.value (configuration_path_deployment_archive)
			end
			if l_folder /= Void then
				l_dirname := file_system.absolute_parent_directory (path)
				l_site_dirname := file_system.absolute_parent_directory (l_path_to_site)
				if l_dirname.starts_with (l_site_dirname) then
					l_relative_path := Global.formatter.relative_path (l_site_dirname, l_dirname)
					if not l_relative_path.is_empty then
						l_folder := file_system.pathname (l_folder, l_relative_path)
					end
				end
				Result := archive_with_timestamp_to_folder (l_folder, a_new_name)
			else
				Result := archive_with_timestamp_to_subfolder (a_default_archive_subfolder, a_new_name)
			end
		ensure
			status_not_void: Result /= Void
		end

	archive_with_timestamp_to_subfolder (a_subfolder, a_new_name: detachable STRING): ROSE_STATUS
			-- Archive this file with a timestamped name, and move it to 'a_subfolder' if specified.
			-- 'a_new_name' is modified to contain the new archive path (if provided).
			--
			-- Note: Rename with timestamp and then move to archive folder, otherwise renaming
			-- and moving in a single operation may fail if the file is currently being used.
		local
			l_folder: detachable STRING
		do
			if a_subfolder /= Void then
				l_folder := Global.formatter.composed_path (<<parent.path, a_subfolder>>)
			end
			Result := archive_with_timestamp_to_folder (l_folder, a_new_name)
		ensure
			status_not_void: Result /= Void
		end

	archive_with_timestamp_to_folder (a_folder, a_new_name: detachable STRING): ROSE_STATUS
			-- Archive this file with a timestamped name, and move it to 'a_folder' if specified.
			-- 'a_new_name' is modified to contain the new archive path (if provided).
			--
			-- Note: Rename with timestamp and then move to archive folder, otherwise renaming
			-- and moving in a single operation may fail if the file is currently being used.
		local
			l_archive: EL_DIRECTORY
			l_file: RAW_FILE
			l_timestamp_name: STRING
			l_archive_name: STRING
			l_renamed_name: STRING
			l_date: ROSE_DATE
			l_retried: BOOLEAN
		do
			if not l_retried then
				create Result
				create l_date.make_now
				create l_timestamp_name.make_empty
				l_timestamp_name.append_string (name)
				l_timestamp_name.append_character ('_')
				l_timestamp_name.append_string (l_date.ccyymmdd.out)
				l_timestamp_name.append_character ('_')
				l_timestamp_name.append_string (l_date.time.hhmmss)
				l_timestamp_name.append_string (extension)
				l_renamed_name := global.formatter.composed_path (<<parent.path, l_timestamp_name>>)
				if a_folder /= Void then
					l_archive_name := global.formatter.composed_path (<<a_folder, l_timestamp_name>>)
				else
					l_archive_name := l_renamed_name
				end
				create l_file.make (path)
				if l_file.exists then
					if a_folder /= Void then
						create l_archive.make_respect_case (a_folder, is_case_respected)
						Result.append (l_archive.create_if_needed)
						if not l_archive.exists then
							Result.errorf ("Cannot move '$1' to '$2': cannot create archive folder '$3'", <<path, l_archive_name, l_archive.path>>)
						elseif not l_archive.is_writable then
							Result.errorf ("Cannot move '$1' to '$2': archive folder '$3' is not writable", <<path, l_archive_name, l_archive.path>>)
						end
					elseif not parent.is_writable then
						Result.errorf ("Cannot move '$1' to '$2': archive folder '$3' is not writable", <<path, l_archive_name, parent.path>>)
					end
					if Result.is_acceptable then
						if parent.is_writable then
								-- Rename with timestamp and then move to archive folder, otherwise renaming
								-- and moving in a single operation may fail if the file is currently being used.
							l_file.change_name (l_renamed_name)
							if a_folder /= Void then
								create l_file.make (l_renamed_name)
								if l_file.exists then
									l_file.change_name (l_archive_name)
								else
									Result.errorf ("Cannot move '$1' to '$2': renaming to '$3' failed", <<path, l_archive_name, l_renamed_name>>)
								end
							end
						else
								-- Rename and move in a single operation.
							l_file.change_name (l_archive_name)
						end
						if Result.is_acceptable then
							Result.outputf ("Renamed '$1' to '$2'", <<path, l_archive_name>>)
							if a_new_name /= Void then
								a_new_name.copy (l_archive_name)
							end
						end
					end
				else
					Result.errorf ("File '$1' does not exist", <<path>>)
				end
			end
			if l_retried or else not Result.is_acceptable then
				Result.errorf ("Could not rename '$1' to '$2'", <<path, l_archive_name>>)
			end
		ensure
			status_not_void: Result /= Void
		rescue
			if not l_retried then
				l_retried := True
				retry
			end
		end

	change_name (a_new_name: STRING): ROSE_STATUS is
			-- Rename this file to 'a_new_name'
		require
			name_ok: a_new_name /= Void
		local
			l_file: RAW_FILE
			retried: INTEGER
		do
			if retried > 0 then
				Result.errorf ("Could not rename $1 to $2", <<path, a_new_name>>)
			else
				create Result
				create l_file.make (path)
				if l_file.exists then
					l_file.change_name (a_new_name)
					Result.outputf ("Renamed '$1' to '$2'", <<path, a_new_name>>)
				else
					Result.errorf ("Could not rename $1 to $2: file does not exist", <<path, a_new_name>>)
				end
			end
		rescue
			retried := retried + 1
			retry
		end

feature {NONE} -- Implementation

	tmp_file: RAW_FILE is
			-- Temporary file used to avoid creating many temp file objects
		once
			create Result.make ("temp")
		end

	scan (a_info: UNIX_FILE_INFO) is
			-- Scan contents
		do
		end

	does_exist: BOOLEAN is
			-- Does this item exist on disk?
		do
			tmp_file.make (path)
			Result := tmp_file.exists
		end

	delete_with_callback (a_callback: FUNCTION [ANY, TUPLE [STRING, INTEGER, INTEGER], BOOLEAN]; a_retries: INTEGER): ROSE_STATUS is
			-- Delete this file, return a status report (will report eventual errors)
			-- 'a_callback' is imposed by common interface file/directory
			-- 'a_retries' indicates how many times to retry deleting files that can't succesfully be deleted right away
		do
			Result := delete (a_retries)
		end

	c_copy_file (a_source, a_destination: POINTER): BOOLEAN is
			-- Copies an existing file to a new file
		external
			"C blocking inline use %"eif_eiffel.h%""
		alias
			"[
#ifdef EIF_WINDOWS
		return (EIF_BOOLEAN)CopyFileA((LPCSTR)$a_source, (LPCSTR)$a_destination, FALSE);
#else
		// add support for other OS-es here
#endif
			]"
		end

end
