indexing
	description: "Common ancestor to EL_FILE and EL_DIRECTORY"
	author: "Zoran Simic"
	date: "$Date: 2010/08/30 $"
	revision: "$Revision: #11 $"

deferred class
	EL_FILEDIR

inherit
	HASHABLE
		redefine
			is_equal
		end
	ROSE_SHARED_GLOBAL
		undefine
			is_equal
		end
	KL_SHARED_FILE_SYSTEM
		undefine
			is_equal
		end
	DEBUG_OUTPUT
		undefine
			is_equal
		end

feature {NONE} -- Initialization

	make (a_path: STRING) is
			-- Associate to 'a_path' (by default all paths are lowercased).
		require
			path_ok: a_path /= Void and then not a_path.is_empty
		do
			path := canonical_form (a_path)
			basename := file_system.basename (path)
		end

	make_respect_case (a_path: STRING; a_respect_case: BOOLEAN)
			-- Make with 'a_path', but do not change its casing (by default all paths are lowercased).
		do
			is_case_respected := a_respect_case
			make (a_path)
		end

	is_case_respected: BOOLEAN
			-- Should the casing of file names/paths be respected?

	canonical_form (a_path: STRING): STRING
			-- Canonical form for 'a_path'
		do
			if not is_case_respected then
				Result := a_path.as_lower
			else
				Result := a_path
			end
		end

feature -- Access

	parent: EL_DIRECTORY is
			-- Directory containing this file
		do
			create Result.make_respect_case (file_system.absolute_parent_directory (path), is_case_respected)
		ensure
			result_ok: Result /= Void
		end

	path: STRING
			-- Full path to this file, example: "C:\path\file.ext"

	basename: STRING
			-- Filename without full path, example: "file.ext"

	hash_code: INTEGER is
			-- Hash code value
		do
			Result := path.hash_code
		end

	is_equal (a_other: like Current): BOOLEAN is
			-- Is `a_other' attached to an object considered equal to current object?
		do
			Result := path.is_equal (a_other.path)
		end

	exists: BOOLEAN is
			-- Does this file exist?
		do
			Result := does_exist
		end

	is_readable: BOOLEAN is
			-- Is this file readable by current process?
		local
			l_tmp_info: UNIX_FILE_INFO
			retried: BOOLEAN
		do
			if not retried then
				if does_exist then
					create l_tmp_info.make
					l_tmp_info.update (path)
					Result := l_tmp_info.is_readable
				end
			end
		rescue
			if not retried then
				retried := True
				retry
			end
		end

	is_writable: BOOLEAN is
			-- Is this file writeable by current process?
		deferred
		end

	created: ROSE_DATE is
			-- Date when directory created
		do
			auto_refresh
			if created_i > 0 then
				create Result.make_by_unix_time (created_i)
			end
		ensure
			result_ok: exists implies Result /= Void
		end

	modified: ROSE_DATE is
			-- UTC date when file or directory was last modified
		do
			auto_refresh
			if modified_i > 0 then
				create Result.make_by_unix_time (modified_i)
			end
		ensure
			result_ok: exists implies Result /= Void
		end

	modified_i: INTEGER
			-- Timestamp of last modification of this file or directory

	size: INTEGER_64 is
			-- Size of this file or directory, in bytes
		deferred
		end

	size_kilobytes: INTEGER is
			-- Size of this file or directory, in kilobytes
		do
			Result := (size // 1024).as_integer_32
		end

	is_unc: BOOLEAN is
			-- Is this folder pointing to a network share?
		do
			Result := global.formatter.is_unc (path)
		end

	is_absolute_path: BOOLEAN is
			-- Is this folder pointing to a network share?
		do
			Result := global.formatter.is_absolute_path (path)
		end

	debug_output: STRING is
			-- String that should be displayed in debugger to represent `Current'.
		do
			Result := path
		end

feature -- Basic operations

	refresh is
			-- Refresh state of this object accordingly to actual file.
		local
			l_tmp_info: UNIX_FILE_INFO
			i: INTEGER
			retried: BOOLEAN
		do
			if retried then
				reset
				exists_i := does_exist
			else
				if modified_i = 0 or else global.app.is_running_in_estudio then
					exists_i := does_exist
				end
				if exists_i then
					create l_tmp_info.make
					l_tmp_info.update (path)
					i := l_tmp_info.date
					if modified_i /= i then
						modified_i := i
						created_i := l_tmp_info.change_date
						scan (l_tmp_info)
					end
				elseif modified_i /= 0 then
					reset
				end
			end
		rescue
			if not retried then
				retried := True
				retry
			end
		end

	force_refresh is
			-- Refresh state of this object accordingly to actual file.
		do
			reset
			refresh
		end

	reset is
			-- Reset
		do
			modified_i := 0
			created_i := 0
		end

	delete (a_callback: FUNCTION [ANY, TUPLE [STRING, INTEGER, INTEGER], BOOLEAN]; a_retries: INTEGER): ROSE_STATUS is
			-- Delete this directory, return a status report (will report eventual errors)
			-- 'a_callback' is called for each directory deleted, the 2 integers inicate
			-- the number of files/dirs deleted so far, and the total respectively
			-- 'a_retries' indicates how many times to retry deleting files that can't succesfully be deleted right away
		require
			retries_ok: a_retries >= 0
		deferred
		ensure
			result_ok: Result /= Void
		end

feature {NONE} -- Implementation

	exists_i: BOOLEAN	-- Does this file or directory exist?
	created_i: INTEGER	-- Timestamp when file or directory created

	auto_refresh is
			-- Call refresh if needed.
		do
			if modified_i = 0 then
				refresh
			end
		end

	scan (a_info: UNIX_FILE_INFO) is
			-- Scan contents
		require
			info_ok: a_info /= Void
		deferred
		end

	does_exist: BOOLEAN is
			-- Does this item exist on disk?
		deferred
		end

invariant

	path_ok: path /= Void

end
