note

	description:

		"Deferred appender class. This class defines the target of the logs."

	library: "elogger"
	copyright: "Copyright (c) 2006, AXA Rosenberg, and others (see copyright.txt)"
	license: "Eiffel Forum License v2 (see forum.txt)"
	author: "Ulrich Weiss, and others"
	last_editor: "$Author: e2 $"
	date: "$Date: 2014/07/04 $"
	revision: "$Revision: #15 $"
	path: "$File: //rose/source/kernel/infrastructure/logging/elog_appender.e $"

deferred class ELOG_APPENDER

inherit

	ANY

	ELOG_FILTER_CONSTANTS
		export
			{NONE} all
		end

	ELOG_SHARED_INTERNAL_LOGGER
		export
			{NONE} all
		end

	MEMORY
		export
			{NONE} all
		undefine
			dispose
		end

	-- In 6.3, switch to EXECUTION_ENVIRONMENT to avoid warnings
	-- EXECUTION_ENVIRONMENT
--	THREAD_CONTROL
--		export
--			{NONE} all
--		end

	ELOG_SHARED_SYSTEM_CLOCK
		export
			{ANY} set_use_utc
			{NONE} all
		end

feature {NONE} -- Initialisation

	make (new_name: STRING; synchronous: BOOLEAN)
			-- Create new appender with name 'new_name'
			-- Note: The appender is open by default,
			-- is_blocking is set to True and
			-- the layout is set to ELOG_SIMPLE_LAYOUT.
		require
			new_name_exists: new_name /= Void
		do
			name := new_name

			create monitor.make
			create append_monitor.make
			create {ELOG_SIMPLE_LAYOUT} layout
			is_synchronous := True
			do_flush := True

			backup_appender_timeout := Default_backup_appender_timeout_min
			set_backup_appender_timeout_max (Default_backup_appender_timeout_max)
			set_backup_appender_timeout_min (Default_backup_appender_timeout_min)

			if not synchronous then
				-- the appender is set to asynchronous
				is_synchronous := False
				buffer_size := default_buffer_size
				initialize_asynchronrous_event_pool
			end
		end

	make_open (new_name: STRING; synchronous: BOOLEAN)
			-- Initialize `Current'.
		require
			new_name_exists: new_name /= Void
		do
			make(new_name, synchronous)
			is_open := True
		ensure
			open: is_open
		end

feature -- Status Report

	name: STRING
			-- Name of this appender that uniquely
			-- identifies it.

	layout: ELOG_LAYOUT
			-- Layout used to format events for this appender. May be Void
			-- if no layout is used.

	is_open: BOOLEAN
			-- Is the appender open for appending?

	backup_appender: ELOG_APPENDER
			-- if something goes wrong, use this appender.

	is_synchronous: BOOLEAN
			-- Are the events beeing logged synchronously?

	buffer_size: INTEGER
			-- The maximum size of 'event_pool' in number
			-- of elements. If this value is zero, there
			-- no limit to the buffer size (except the memory
			-- available).

	is_blocking: BOOLEAN
			-- is the buffer blocking if it is full?
			-- if it isn't, the events are summarized
			-- and discarded.

	do_flush: BOOLEAN
			-- Should we flush each time we append an event?

	use_backup_appender: BOOLEAN
			-- use backup_appender instead

	backup_appender_timeout: INTEGER
			-- timeout after which we retry this appender again after
			-- a failure. this value starts from Start_backup_appender_timeout,
			-- and doubles with each time the appender did not work.
			-- e.g. 100ms, 200ms, 400ms, 800ms ... etc

	backup_appender_timeout_min: INTEGER
			-- The start of the timeout counter, e.g. 100 ms.
			-- see 'backup_appender_timeout' for more information.

	backup_appender_timeout_max: INTEGER
			-- The maximum time the timeout grows to. If this value is 0,
			-- then the 'backup_appender_timeout' will grow unboundedly.
			-- (until INTEGER.max is reached, which is 24 days)

feature -- Status Setting

	set_layout (new_layout: ELOG_LAYOUT)
			-- Set the layout that this appender should use.
		require
			layout_exists: new_layout /= Void
		do
			-- twinning the layout ensures, that format will not be called on
			-- two different appenders at the same time on the same layout.
			layout := new_layout.twin
		end

	close
			-- Release of resources.
		do
			if is_open then
				if is_synchronous then
					is_open := False
				else
					if is_blocking then
						close_wait
					else
						close_dispose
					end
				end

				internal_close
			end

			if backup_appender /= Void then
				backup_appender.close
			end
		end

	set_name (new_name: STRING)
			-- Set the name of this appender
		require
			name_exists: new_name /= Void
		do
			name := new_name
		ensure
			name_set: name = new_name
		end

	set_filter (a_filter: ELOG_FILTER)
			-- Set 'filter' to 'a_filter'. See ELOG_AND_FILTER or
			-- ELOG_OR_FILTER on how to set multiple filters.
			-- If a_filter is void, no filter will be used.
		do
			filter := a_filter
		ensure
			filter_set: filter = a_filter
		end

	set_backup_appender (a_appender: ELOG_APPENDER; pass_settings: BOOLEAN)
			-- sets 'backup_appender'
			-- if pass_settings is true, all the settings
			-- will be passed on.
		do
			backup_appender := a_appender
			if a_appender /= Void and pass_settings then
				a_appender.set_filter (filter)
				a_appender.set_layout (layout)
				a_appender.set_flush (do_flush)
			end
		ensure
			no_cycle: not check_for_backup_cycle
			backup_appender_set: backup_appender = a_appender
		end

	set_backup_appender_timeout_min (a_backup_appender_timeout: INTEGER)
			-- sets 'backup_appender_timeout' in milliseconds.
		require
			a_backup_appender_valid: a_backup_appender_timeout > 0 and
				a_backup_appender_timeout <= backup_appender_timeout_max
		do
			backup_appender_timeout_min := a_backup_appender_timeout
		ensure
			backup_appender_timeout_min_set: backup_appender_timeout_min = a_backup_appender_timeout
		end

	set_backup_appender_timeout_max (a_backup_appender_timeout_max: INTEGER)
			-- sets 'backup_appender_timeout_max' in milliseconds.
		require
			a_backup_appender_valid: a_backup_appender_timeout_max >
				backup_appender_timeout
		do
			backup_appender_timeout_max := a_backup_appender_timeout_max
		ensure
			backup_appender_timeout_max_set: backup_appender_timeout_max = a_backup_appender_timeout_max
		end

	set_synchronous_wait
			-- sets 'is_synchronous' to True.
			-- If the buffer still contains elements, we wait
			-- here until the pool is empty.
		do
			monitor.lock
			append_monitor.lock
			if not is_synchronous then
				wait (True)
				is_synchronous := True
			end
			append_monitor.unlock
			monitor.unlock
		end

	set_synchronous_force
			-- sets 'is_synchronous' to True.
			-- If the buffer still contains elements,
			-- it will be emptied and the events discarded.
		do
			monitor.lock
			append_monitor.lock
			if not is_synchronous then
				is_synchronous := True
				event_pool.wipe_out
			end
			append_monitor.unlock
			monitor.unlock
		end

	set_asynchronous
			-- sets 'is_synchronous' to False.
		do
			monitor.lock
			append_monitor.lock
			if is_synchronous then
				is_synchronous := False
				if event_pool = Void then
					initialize_asynchronrous_event_pool
				end
			end
			append_monitor.unlock
			monitor.unlock
		end

	set_blocking (a_flag: BOOLEAN)
			-- sets 'is_blocking'
		do
			is_blocking := a_flag
		ensure
			is_blocking_set: is_blocking = a_flag
		end

	set_buffer_size (a_size: INTEGER)
			-- sets the maximal buffer size to 'a_size'.
		require
			a_size_large_enough: a_size >= 0
		do
			buffer_size := a_size
			if event_pool /= Void then
				event_pool.set_max_size (buffer_size)
			end
		ensure
			buffer_size_set: buffer_size = a_size
		end

	set_use_backup_appender (use_backup: BOOLEAN; current_event: ELOG_EVENT)
			-- sets to use backup appender
			-- this can be called from the dispatcher or the appender
		do
			if use_backup and not use_backup_appender then
				-- set to use the backup appender
				backup_appender_timeout := backup_appender_timeout_min
				debug ("elog_failsafe_trace")
					internal_logger.log_information (create {ELOG_SIMPLE_EVENT}.make
						("Use backup appender (timeout reset)"))
				end
				if backup_appender = Void then
					internal_logger.log_fatal (create {ELOG_SIMPLE_EVENT}.make(
						"Error: could not append message in appender %"" +
						name + "%". No backup appender was specified. "))
				else
					last_backup_appender_timeout_time := system_clock.date_time_now
					use_backup_appender := True
					internal_logger.log_fatal (create {ELOG_SIMPLE_EVENT}.make(
						"Error: could not append message in appender %"" +
						name + "%". Now using backup appender %"" +
						backup_appender.name + "%". "))
					if current_event /= Void then
						backup_appender.append (current_event)
					end
				end
			elseif not use_backup and use_backup_appender then
				-- set to use this appender, not the backup appender anymore
				backup_appender_timeout := backup_appender_timeout_min
				debug ("elog_failsafe_trace")
					internal_logger.log_information (create {ELOG_SIMPLE_EVENT}.make
						("Use real appender again (timeout reset)"))
				end
				use_backup_appender := False
			end
		end

	set_flush (a_flush: BOOLEAN)
			-- Set `do_flush'
		do
			do_flush := a_flush
		ensure
			do_flush_set: do_flush = a_flush
		end

	wait (is_monitor_locked: BOOLEAN)
			-- wait here until the consumer thread
			-- has finished dispatching the events.
		require
			is_asynchronous: not is_synchronous
		do
			debug ("elog_wait_trace")
				internal_logger.log_information (
					create {ELOG_SIMPLE_EVENT}.make (
						"WaitTrace: start 'wait' in '" + name + "'"))
			end

			if not is_monitor_locked then
				append_monitor.lock
			end

			-- loop until the pool is empty or it is closed
			from
			until
				(event_pool.is_empty or event_pool.is_closed) and
					not consumer_thread.is_appending
			loop
				event_pool.notempty.broadcast;
				-- sleep for 0.1 ms
				(create {EXECUTION_ENVIRONMENT}).sleep (100_000)
			end

			if not is_monitor_locked then
				append_monitor.unlock
			end

			debug ("elog_wait_trace")
				internal_logger.log_information (
					create {ELOG_SIMPLE_EVENT}.make (
						"WaitTrace: done 'wait' in '" + name + "'"))
			end
		end

feature -- Removal

	dispose
			-- Close this appender when garbage collected. Perform
			-- minimal operations to release resources. Do not call
			-- other object as they may have been garbage collected.
		deferred
		end

feature -- Output

	write_header
			-- write the layout header to the appender
			-- Note: this is done synchronously.
		require
			header_not_void: layout.header /= Void
		do
			do_append_string (layout.header)
		end

	write_footer
			-- write the layout footer to the appender
			-- Note: This is done synchronously
		require
			footer_not_void: layout.footer /= Void
		do
			do_append_string (layout.footer)
		end

feature {ELOG_LOGGER, ELOG_DISPATCHER, ELOG_APPENDER, TS_TEST_CASE} -- Output

	append (event: ELOG_EVENT)
			-- Log event on this appender (front end).
		require
			event_exists: event /= Void
			is_open: is_open or use_backup_appender
		local
			l_retried: BOOLEAN
			l_time_diff: DT_TIME_DURATION
		do
			if not l_retried then
				-- filter event through filters to determine
				-- if it should be logged
				append_monitor.lock
				if filter = Void or else filter.decide (event) /= Filter_reject then
					if not use_backup_appender then
						-- do the actual append
						if is_synchronous then
							do_append (event)
						elseif is_blocking or else not event_pool.is_full then
							-- if 'is_blocking' is true, we will block here,
							-- otherwise if the pool is full the event will
							-- be discarded (= not put onto the pool).
							event_pool.put (event)
							if not is_blocking and not discard_summary.is_empty then
								do_append (discard_summary.summary)
							end
						else
							-- not blocking and full event pool, we have to
							-- discard messages.
							discard_summary.discard_event (event)
						end

						if retrying_appender then
							-- the appender was retried and worked
							retrying_appender := False
							use_backup_appender := True
							set_use_backup_appender (False, Void)
						end
					else
						l_time_diff := system_clock.date_time_now - last_backup_appender_timeout_time
						if l_time_diff.millisecond_count > backup_appender_timeout then
							if backup_appender_timeout > backup_appender_timeout_max // 2 then
								-- set the timeout to the max value
								backup_appender_timeout := backup_appender_timeout_max
								debug ("elog_failsafe_trace")
									internal_logger.log_information (create {ELOG_SIMPLE_EVENT}.make
										("Setting timeout to max: "+(backup_appender_timeout).out))
								end
							else
								backup_appender_timeout := backup_appender_timeout * 2
								debug ("elog_failsafe_trace")
									internal_logger.log_information (create {ELOG_SIMPLE_EVENT}.make
										("Doubling timeout to "+(backup_appender_timeout).out))
								end
							end

							if is_open
							then
								-- try the current appender again, for the next event
								retrying_appender := True
								use_backup_appender := False
							end
						end
						backup_appender.append (event)
					end
				end
				append_monitor.unlock
			elseif not retrying_appender then
				-- set to use the backup appender
				append_monitor.lock
				set_use_backup_appender (True, event)
				append_monitor.unlock
			else
				-- we retried the appender and it still didn't work
				retrying_appender := False
				use_backup_appender := True
			end
		rescue
			append_monitor.unlock
			if not l_retried then
				l_retried := True
				retry
			end
		end

feature {ELOG_DISPATCHER} -- Implementation

	do_append (event: ELOG_EVENT)
			-- Append 'event' to this appender.
		require
			event_exists: event /= Void
			is_open: is_open
		local
			s: STRING
		do
			if event.override_layout /= Void then
				s := event.override_layout.format (event)
			else
				s := layout.format (event)
			end
			do_append_string( s )
		end

	monitor: ELOG_MUTEX
			-- monitor used for synchronizing events

	append_monitor: ELOG_MUTEX
			-- monitor used for synchronizing events

feature {NONE} -- Implementation

	do_append_string (a_string: STRING)
			-- append 'a_string'
		require
			string_exists: a_string /= Void
			is_open: is_open
		deferred
		end

	check_for_backup_cycle: BOOLEAN
			-- checks if the 'backup_appender' form a cycle.
		local
			list: DS_LINKED_LIST [ELOG_APPENDER]
			a: ELOG_APPENDER
		do
			create list.make
			list.force_last (Current)
			from
				a := backup_appender
			until
				a = Void or else list.has (a)
			loop
				a := a.backup_appender
				if a /= Void then
					list.force_last (a)
				end
			end
			if a /= Void then
				Result := True
			end
		end

	initialize_asynchronrous_event_pool
			-- initializes the event pool and starts the dispatcher.
		require
			event_pool_was_void: event_pool = Void
			consumer_thread_was_void: consumer_thread = Void
		do
			if event_pool = Void then
				is_blocking := True
				create discard_summary.make (Current)
				create event_pool
				event_pool.set_max_size (buffer_size)
				create consumer_thread.make (event_pool, Current)
				consumer_thread.launch
			end
		end

	internal_close
			-- Release any resources for this appender.
		deferred
		end

	close_dispose
			-- Release any resources for this appender.
			-- if there are any events still in the queue
			-- they will be disposed.
		require
			is_asynchronous: not is_synchronous
		do
			monitor.lock
			event_pool.wipe_out
			event_pool.close
			is_open := False
			event_pool.notempty.broadcast
			event_pool.notfull.broadcast
			consumer_thread.join
			consumer_thread := Void
			event_pool := Void
			monitor.unlock
		end

	close_wait
			-- Release any resources for this appender.
			-- wait here until all elements in the
			-- queue have been emptied.
		require
			is_asynchronous: not is_synchronous
		do
			debug ("elog_wait_trace")
				internal_logger.log_information (create {ELOG_SIMPLE_EVENT}.
					make ("WaitTrace: start closing appender "+name))
			end
			monitor.lock
			wait (False)
			event_pool.close
			is_open := False
			event_pool.notempty.broadcast
			event_pool.notfull.broadcast
			consumer_thread.join
			consumer_thread := Void
			event_pool := Void
			monitor.unlock
			debug ("elog_wait_trace")
				internal_logger.log_information (create {ELOG_SIMPLE_EVENT}.
					make ("WaitTrace: done closing appender "+name))
			end
		end

	event_pool: ELOG_CONCURRENT_LINKED_QUEUE_CV [ELOG_EVENT]
			-- Threadsafe queue containing events for
			-- asynchronous logging. The events are put into
			-- the queue synchronously and fetched by the
			-- consumer thread.

	consumer_thread: ELOG_DISPATCHER
			-- consumer thread

	filter: ELOG_FILTER
			-- Filter that determines if a log event is processed for this
			-- appender or not

	discard_summary: ELOG_DISCARD_SUMMARY
			-- discard summary

	last_backup_appender_timeout_time: DT_DATE_TIME
			-- the last time the appender was checked for availability

	retrying_appender: BOOLEAN
			-- are we currently retrying the current appender?

	Default_buffer_size: INTEGER = 0
			-- the default buffer size is 0 elements (meaning no size limit)

	Default_backup_appender_timeout_min: INTEGER = 100
			-- the default time for 'backup_appender_timeout'

	Default_backup_appender_timeout_max: INTEGER
			-- the default maximal timeout
		once
			Result := {INTEGER}.max_value
		end

invariant

	-- Note: Keep in mind that these invariants might theoretically
	-- be raised by mistake due to multithreading (they are not evaluated atomically).

	name_not_void: name /= Void
	layout_not_void: layout /= Void
	buffer_size_non_negative: buffer_size >= 0
	backup_appender_set: use_backup_appender implies backup_appender /= Void
	event_pool_has_consumer_1: event_pool = Void implies consumer_thread = Void
	event_pool_has_consumer_2: event_pool /= Void implies consumer_thread /= Void
	event_pool_exists: (is_open and not is_synchronous) implies event_pool /= Void
	backup_appender_timeout_valid: backup_appender_timeout > 0
	backup_appender_timeout_max_valid: backup_appender_timeout_max >= backup_appender_timeout
	append_monitor_set: append_monitor /= Void
	monitor_set: monitor /= Void

end
