indexing
	description: "Logging appender that writes to a file and is rolled at specified calendar intervals."
	library: "elogger"
	copyright: "Copyright (c) 2006, AXA Rosenberg, and others (see copyright.txt)"
	author: "Ulrich Weiss, and others"
	license: "Eiffel Forum License v2 (see forum.txt)"
	date: "$Date: 2009/05/28 $"
	revision: "$Revision: #4 $"
	archive: "$File: //rose/source/kernel/infrastructure/logging/appenders/elog_calendar_rolling_appender.e $"

class ELOG_CALENDAR_ROLLING_APPENDER[G -> FILE create make end]

inherit

	ELOG_ROLLING_FILE_APPENDER[G]
		rename
			make as rolling_file_appender_make
		export
			{NONE} rolling_file_appender_make
		redefine
			rollover_required
		end

create

	make, make_daily, make_minutely, make_hourly

feature {NONE} -- Initialisation

	make, make_daily (new_name: STRING; number_of_backups: INTEGER; appending: BOOLEAN; synchronous: BOOLEAN) is
			-- Create a new file appender on the file
			-- with 'new_name'. Roll the log over to a backup file every day.
			-- Keep a maximum of 'number_of_backups' backup files.
			-- Append to an existing log file if 'appending'.
		require
			name_exists: new_name /= Void
			name_not_empty: not new_name.is_empty
			positive_number_of_backups: number_of_backups >= 0
		do
			file_appender_make (new_name, appending, synchronous)
			max_number_of_backups := number_of_backups
			days := 1
			last_rollover := system_clock.date_time_now
		ensure
			elog_file_open: stream.is_open_write
		end

	make_minutely (new_name: STRING; number_of_backups: INTEGER; appending: BOOLEAN; synchronous: BOOLEAN) is
			-- Create a new file appender on the file
			-- with 'new_name'. Roll the log over to a backup file every minute.
			-- Keep a maximum of 'number_of_backups' backup files.
			-- Append to an existing log file if 'appending'.
		require
			name_exists: new_name /= Void
			name_not_empty: not new_name.is_empty
			positive_number_of_backups: number_of_backups >= 0
		do
			file_appender_make (new_name, appending, synchronous)
			max_number_of_backups := number_of_backups
			minutes := 1
			last_rollover := system_clock.date_time_now
		ensure
			ELOG_file_open: stream.is_open_write
		end

	make_hourly (new_name: STRING; number_of_backups: INTEGER; appending: BOOLEAN; synchronous: BOOLEAN) is
			-- Create a new file appender on the file
			-- with 'new_name'. Roll the log over to a backup file every hour.
			-- Keep a maximum of 'number_of_backups' backup files.
			-- Append to an existing log file if 'appending'.
		require
			name_exists: new_name /= Void
			name_not_empty: not new_name.is_empty
			positive_number_of_backups: number_of_backups >= 0
		do
			file_appender_make (new_name, appending, synchronous)
			max_number_of_backups := number_of_backups
			hours := 1
			last_rollover := system_clock.date_time_now
		ensure
			ELOG_file_open: stream.is_open_write
		end

feature -- Element Change

	set_rollover_period (se: INTEGER; mi: INTEGER; h: INTEGER; d: INTEGER; w: INTEGER; mo: INTEGER; yr: INTEGER) is
			-- sets the rollover period to se seconds, mi minutes, h hours
			-- d days, w weeks, mo months and yr years from now.
		do
			seconds := se
			minutes := mi
			hours := h
			days := d
			weeks := w
			months := mo
			years := yr
		end

feature -- Status Report

	get_rollover_period: TUPLE[INTEGER,INTEGER,INTEGER,INTEGER,INTEGER,INTEGER,INTEGER] is
			-- returns a tuple of the rollover_period.
			-- TUPLE[seconds, minutes, hours, days, weeks, months, years]
		do
			create Result
			Result.put (seconds, 1)
			Result.put (minutes, 2)
			Result.put (hours, 3)
			Result.put (days, 4)
			Result.put (weeks, 5)
			Result.put (months, 6)
			Result.put (years, 7)
		ensure
			Result_valid: Result.count = 7
		end


feature {NONE} -- Implementation		

	seconds: INTEGER
			-- number of seconds for a rollover

	minutes: INTEGER
			-- number of minutes for a rollover

	hours: INTEGER
			-- number of hours for a rollover

	days: INTEGER
			-- number of days for a rollover

	weeks: INTEGER
			-- number of weeks for a rollover

	months: INTEGER
			-- number of months for a rollover

	years: INTEGER
			-- number of years for a rollover

	rollover_required: BOOLEAN is
			-- Has the current time reached the rollover date and time?
		local
			current_time: DT_DATE_TIME
			check_time: DT_DATE_TIME
		do
			current_time := system_clock.date_time_now
			check_time := last_rollover.twin
			if seconds > 0 then
				check_time.add_seconds (seconds)
			end
			if minutes > 0 then
				check_time.add_minutes (minutes)
			end
			if hours > 0 then
				check_time.add_hours (hours)
			end
			if days > 0 then
				check_time.add_days (days)
			end
			if weeks > 0 then
				check_time.add_days (7 * weeks)
			end
			if months > 0 then
				check_time.add_months (months)
			end
			if years > 0 then
				check_time.add_years (years)
			end

			Result := current_time >= check_time
			if Result then
				last_rollover := current_time
			end
		end

	last_rollover: DT_DATE_TIME
			-- Date and time of last check.

invariant
	period_valid: seconds >= 0 and minutes >= 0 and hours >= 0 and days >= 0 and weeks >= 0 and months >= 0 and years >= 0

end -- class ELOG_CALENDAR_ROLLING_APPENDER
