indexing
	description: "Thread-safe linked list"
	author: "Nate McNamara"
	last_editor: "$Author: it $"
	date: "$Date: 2009/12/09 $"
	revision: "$Revision: #5 $"
	archive: "$File: //rose/source/kernel/infrastructure/logging/concurrency/elog_concurrent_linked_list.e $"

class
	ELOG_CONCURRENT_LINKED_LIST[G]

inherit
	ANY
		redefine
			copy,
			default_create,
			is_equal
		end

	ELOG_THREAD_SAFE_AUTO_INCREMENETER
		export
			{NONE} all
		redefine
			copy,
			default_create,
			is_equal
		end

create
	default_create,
	make_from_like

feature {NONE} -- Initialization

	default_create
			-- Create a new empty list.
		do
			count := 0
			first := Void
			last := Void
			create mutex.make
			unique_id := next_integer
		end

	make_from_like (a_original: like Current)
			-- Initialize `Current' from `a_original'.
		do
			default_create
			copy (a_original)
		end

feature -- Access
	count: INTEGER
			-- The count of elements in this list

	current_thread_owns_lock: BOOLEAN
			-- Does the current thread own this list's lock?
		do
			Result := mutex.is_owned_by_current_thread
		ensure
			definition: Result = mutex.is_owned_by_current_thread
		end

feature -- Basic Operations
	append(a_other: like Current)
			-- Append the contents of `a_other' to `Current'.
		require
			not_locked: not current_thread_owns_lock
		do
			do_safe_dual_op(a_other, agent unsafe_append)
		ensure
			not_locked: not current_thread_owns_lock
		end

	copy (a_other: like Current)
			-- <Precursor>
		do
			check
				not_locked: not current_thread_owns_lock
				a_other_not_locked: not a_other.current_thread_owns_lock
			end
			default_create
			do_safe_dual_op(a_other, agent unsafe_copy)
		ensure then
			not_locked: not current_thread_owns_lock
		end

	delete(a_value: G)
			-- Remove `a_value' from this list, if present.
		require
			not_locked: not current_thread_owns_lock
		do
			do_safe(agent unsafe_delete(a_value))
		ensure
			not_locked: not current_thread_owns_lock
		end

	do_all(a_proc: PROCEDURE[ANY, TUPLE[G]])
			-- Apply `a_proc' to each element of this list.
			-- Be careful if you acquire any locks in `a_proc'!
		require
			a_proc_not_void: a_proc /= Void
			not_locked: not current_thread_owns_lock
		do
			do_safe(agent unsafe_do_all(a_proc))
		ensure
			not_locked: not current_thread_owns_lock
		end

	do_if(
		a_proc: PROCEDURE[ANY, TUPLE [G]];
		a_pred: FUNCTION[ANY, TUPLE [G], BOOLEAN])
			-- Apply `a_proc' to every item that satisfies `a_pred'.
		require
			a_proc_not_void: a_proc /= Void
			a_pred_not_void: a_pred /= Void
			not_locked: not current_thread_owns_lock
		do
			do_safe(agent unsafe_do_if(a_proc, a_pred))
		ensure
			not_locked: not current_thread_owns_lock
		end

	find(a_pred: FUNCTION[ANY, TUPLE[G], BOOLEAN]): G
			-- The first item in this list for which
			-- `a_pred' is True.  Return G's default value
			-- if no such item is found.
			-- Be careful if you acquire any locks in `a_pred'!
		require
			a_pred_not_void: a_pred /= Void
		do
			mutex.lock
			Result := unsafe_find(a_pred)
			mutex.unlock
		ensure
			find_may_be_void: True
		end

	for_all(a_pred: FUNCTION[ANY, TUPLE[G], BOOLEAN]): BOOLEAN
			-- Test whether `a_pred' is True for each element
			-- of this list.
			-- Be careful if you acquire any locks in `a_pred'!
		require
			a_pred_not_void: a_pred /= Void
		do
			mutex.lock
			Result := unsafe_for_all(a_pred)
			mutex.unlock
		end

	is_equal(a_other: like Current): BOOLEAN
			-- <Precursor>
		local
			l_was_locked: BOOLEAN
		do
			-- Deadlock prevention
			if current_thread_owns_lock then
				l_was_locked := True
				mutex.unlock
			end

			dual_lock(a_other)
			Result := unsafe_is_equal(a_other)
			dual_unlock(a_other)

			if l_was_locked then
				mutex.lock
			end
		ensure then
			locked_unchanged: current_thread_owns_lock =
				old current_thread_owns_lock
		end

	prepend(a_other: like Current)
			-- Prepend the contents of `a_other' to `Current'.
		require
			not_locked: not current_thread_owns_lock
			a_other_not_void: a_other /= Void
			a_other_not_locked: not a_other.current_thread_owns_lock
		do
			do_safe_dual_op(a_other, agent unsafe_prepend)
		ensure
			not_locked: not current_thread_owns_lock
			a_other_not_locked: not a_other.current_thread_owns_lock
		end

	put(a_value: G)
			-- Put a value at tail.
		require
			not_locked: not current_thread_owns_lock
		local
			l_node: like first
		do
			create l_node.make(a_value)
			mutex.lock
			append_node(l_node)
			mutex.unlock
		ensure
			not_locked: not current_thread_owns_lock
		end

	put_if_new(a_value: G)
			-- Put a value at tail.
		require
			not_locked: not current_thread_owns_lock
		do
			mutex.lock
			if not unsafe_has(a_value) then
				append_node(new_node(a_value))
			end
			mutex.unlock
		ensure
			not_locked: not current_thread_owns_lock
		end

	remove: G
			-- Remove element at first position and return its value.
			-- if the queue is empty, returns Void
		require
			not_locked: not current_thread_owns_lock
		do
			mutex.lock
			Result := unsafe_remove
			mutex.unlock
		ensure
			not_locked: not current_thread_owns_lock
		end

	wipe_out
			-- Remove all items from queue.
		require
			not_locked: not current_thread_owns_lock
		do
			do_safe(agent unsafe_wipe_out)
		ensure
			not_locked: not current_thread_owns_lock
		end

feature -- Status Report

	is_empty: BOOLEAN
			-- is the queue empty?
		do
			Result := first = Void
		end

	has(a_value: G): BOOLEAN
			-- Is `a_value' in this list?
		require
			not_locked: not current_thread_owns_lock
		do
			mutex.lock
			Result := unsafe_has(a_value)
			mutex.unlock
		ensure
			not_locked: not current_thread_owns_lock
		end

feature {NONE} -- Implementation
	append_node(a_node: like first)
			-- Add `a_node' to the end of this list.
		require
			a_node_not_void: a_node /= Void
			locked: current_thread_owns_lock
		do
			if first = Void then
				first := a_node
				last := a_node
			else
				last.set_next(a_node)
				last := a_node
			end

			count := count + 1
		ensure
			appended: last = a_node
			locked: current_thread_owns_lock
		end

	delete_node(a_node, a_prev: like first)
			-- Remove `a_node' from this list.
			-- If `a_node' /= first, then `a_prev.next'
			-- must be `a_node'.
		require
			a_prev_is_void_means_node_is_first:
				a_prev = Void implies a_node = first
			a_prev_is_prev: a_prev /= Void implies
				a_prev.next = a_node
			a_node_not_void: a_node /= Void
			locked: current_thread_owns_lock
		do
			if a_node = last then
				if a_node = first then
					first := Void
					last := Void
				else
					last := a_prev
					a_prev.next := Void
				end
			elseif a_node = first then
				first := first.next
			else
				if a_node.next /= Void and then a_node.next.next /= Void then
					a_node.next := a_node.next.next
				end
			end

			count := count - 1
		ensure
			count_decreased: count = old count - 1
			locked: current_thread_owns_lock
		end

	do_safe(a_op: PROCEDURE[like Current, TUPLE])
			-- Perform `a_op' within a locked block.
		require
			a_op_not_void: a_op /= Void
			not_locked: not current_thread_owns_lock
		do
			mutex.lock
			a_op.call([])
			mutex.unlock
		ensure
			not_locked: not current_thread_owns_lock
		end

	do_safe_dual_op(
		a_other: like Current;
		a_op: PROCEDURE[like Current, TUPLE[like Current]])
			-- Perform `a_op' with both `Current'
			-- and `a_other' locked.
		require
			a_other_not_void: a_other /= Void
			a_op_not_void: a_op /= Void
			not_locked: not current_thread_owns_lock
			a_other_not_locked: not a_other.current_thread_owns_lock
		do
			dual_lock(a_other)
			a_op.call([a_other])
			dual_unlock(a_other)
		ensure
			not_locked: not current_thread_owns_lock
			a_other_not_locked: not a_other.current_thread_owns_lock
		end

	dual_lock(a_other: like Current)
			-- Lock `Current' and `a_other' in order
			-- according to `unique_id'.
		require
			not_locked: not current_thread_owns_lock
			a_other_not_void: a_other /= Void
			a_other_not_locked: not a_other.current_thread_owns_lock
		do
			if unique_id < a_other.unique_id then
				mutex.lock
				a_other.mutex.lock
			else
				a_other.mutex.lock
				mutex.lock
			end
		ensure
			locked: current_thread_owns_lock
			a_other_locked: a_other.current_thread_owns_lock
		end

	dual_unlock(a_other: like Current)
			-- Unlock `Current' and `a_other' in order
			-- according to `unique_id'.
		require
			locked: current_thread_owns_lock
			a_other_not_void: a_other /= Void
			a_other_locked: a_other.current_thread_owns_lock
		do
			if unique_id < a_other.unique_id then
				a_other.mutex.unlock
				mutex.unlock
			else
				mutex.unlock
				a_other.mutex.unlock
			end
		ensure
			not_locked: not current_thread_owns_lock
			a_other_not_locked: not a_other.current_thread_owns_lock
		end

	new_node(a_value: G): like first
			-- Create a node from `a_value'.
		do
			create Result.make(a_value)
		ensure
			new_node_not_void: Result /= Void
			value_set: Result.value = a_value
		end

	unsafe_append(a_other: like Current)
			-- Helper for `append'.
		require
			locked: current_thread_owns_lock
			a_other_not_void: a_other /= Void
			a_other_locked: a_other.current_thread_owns_lock
		local
			l_other_node: like first
		do
			l_other_node := a_other.first
			from until l_other_node = Void loop
				append_node(new_node(l_other_node.value))
				l_other_node := l_other_node.next
			end
		ensure
			locked: current_thread_owns_lock
			a_other_locked: a_other.current_thread_owns_lock
		end

	unsafe_copy(a_other: like Current)
			-- Helper for `copy'.
		require
			locked: current_thread_owns_lock
			a_other_not_void: a_other /= Void
			a_other_locked: a_other.current_thread_owns_lock
		do
			unsafe_append(a_other)
			-- mutex.unlock
		ensure
			locked: current_thread_owns_lock
			a_other_locked: a_other.current_thread_owns_lock
		end

	unsafe_delete(a_value: G)
			-- Helper for `delete'.
		require
			locked: current_thread_owns_lock
		local
			l_prev: like first
			l_node: like first
		do
			from l_node := first
			until l_node = Void or else l_node.value ~ a_value
			loop
				l_prev := l_node
				l_node := l_node.next
			end

			if l_node /= Void then
				delete_node(l_node, l_prev)
			end
		ensure
			locked: current_thread_owns_lock
		end

	unsafe_do_all(a_proc: PROCEDURE[ANY, TUPLE[G]])
			-- Helper for `do_all'.
		require
			a_proc_not_void: a_proc /= Void
			locked: current_thread_owns_lock
		local
			l_node: like first
		do
			from l_node := first until l_node = Void loop
				a_proc.call([l_node.value])
				l_node := l_node.next
			end
		ensure
			locked: current_thread_owns_lock
		end

	unsafe_do_if(
		a_proc: PROCEDURE[ANY, TUPLE [G]];
		a_pred: FUNCTION[ANY, TUPLE [G], BOOLEAN])
			-- Helper for `do_if'.
		require
			a_proc_not_void: a_proc /= Void
			a_pred_not_void: a_pred /= Void
			locked: current_thread_owns_lock
		local
			l_node: like first
		do
			from l_node := first until l_node = Void loop
				if a_pred.item([l_node.value]) then
					a_proc.call([l_node.value])
				end

				l_node := l_node.next
			end
		ensure
			locked: current_thread_owns_lock
		end

	unsafe_find(a_pred: FUNCTION[ANY, TUPLE[G], BOOLEAN]): G
			-- Helper for `find'.
		require
			a_pred_not_void: a_pred /= Void
			locked: current_thread_owns_lock
		local
			l_node: like first
		do
			from l_node := first
			until l_node = Void or else a_pred.item([l_node.value])
			loop l_node := l_node.next
			end

			if l_node /= Void then
				Result := l_node.value
			end
		ensure
			locked: current_thread_owns_lock
		end

	unsafe_for_all(a_pred: FUNCTION[ANY, TUPLE[G], BOOLEAN]): BOOLEAN
			-- Helper for `a_pred'.
		require
			a_pred_not_void: a_pred /= Void
			locked: current_thread_owns_lock
		local
			l_node: like first
		do
			from l_node := first
			until l_node = Void or else not a_pred.item([l_node.value])
			loop l_node := l_node.next
			end

			Result := l_node = Void
		ensure
			locked: current_thread_owns_lock
		end

	unsafe_has(g: G): BOOLEAN
			-- Helper for `has'.
		require
			locked: current_thread_owns_lock
		local
			l_node: like first
		do
			from l_node := first
			until l_node = Void or else l_node.value ~ g
			loop l_node := l_node.next
			end
			Result := l_node /= Void
		ensure
			locked: current_thread_owns_lock
		end

	unsafe_is_equal(a_other: like Current): BOOLEAN
			-- Helper for `is_equal'.
		require
			locked: current_thread_owns_lock
			a_other_not_void: a_other /= Void
			a_other_locked: a_other.current_thread_owns_lock
		local
			l_our_node: like first
			l_other_node: like first
		do
			from
				l_our_node := first
				l_other_node := a_other.first
			until
				l_our_node = Void or l_other_node = Void or else
					l_our_node.value /~ l_other_node.value
			loop
				l_our_node := l_our_node.next
				l_other_node := l_other_node.next
			end

			Result := l_our_node = Void and l_other_node = Void
		ensure
			locked: current_thread_owns_lock
			a_other_locked: a_other.current_thread_owns_lock
		end

	unsafe_prepend(a_other: like Current)
			-- Helper for `prepend'.
		require
			locked: current_thread_owns_lock
			a_other_not_void: a_other /= Void
			a_other_locked: a_other.current_thread_owns_lock
		do
			if a_other.count > 0 then
				unsafe_copy(a_other)
			end
		ensure
			locked: current_thread_owns_lock
			a_other_locked: a_other.current_thread_owns_lock
		end

	unsafe_remove: G
			-- Helper for `remove'.
		require
			locked: current_thread_owns_lock
		do
			if first /= Void then
				Result := first.value

				first := first.next
				if first = Void then
					last := Void
				end

				count := count - 1
			end
		ensure
			locked: current_thread_owns_lock
		end

	unsafe_wipe_out
			-- Helper for `wipe_out'.
		require
			locked: current_thread_owns_lock
		do
			first := Void
			last := Void
			count := 0
		ensure
			locked: current_thread_owns_lock
		end

feature {NONE} -- Validation
	check_count: BOOLEAN
			-- count = actual number of nodes
		require
			not_locked: not current_thread_owns_lock
		do
			mutex.lock
			Result := count = unsafe_count_by_walking
			mutex.unlock
		ensure
			not_locked: not current_thread_owns_lock
		end

	check_empty: BOOLEAN
			-- Check that first = Void iff last = Void
		do
			mutex.lock
			Result := (first = Void) = (last = Void)
			mutex.unlock
		end

	unsafe_count_by_walking: INTEGER
			-- Helper for `check_count'.
		require
			locked: current_thread_owns_lock
		local
			l_node: like first
		do
			from l_node := first until l_node = Void loop
				Result := Result + 1
				l_node := l_node.next
			end
		ensure
			locked: current_thread_owns_lock
		end

feature {ELOG_CONCURRENT_LINKED_LIST} -- Members
	first: ELOG_CONCURRENT_LINKED_QUEUE_NODE [G]
			-- head of queue

	last: ELOG_CONCURRENT_LINKED_QUEUE_NODE [G]
			-- tail of queue

	mutex: ELOG_MUTEX
			-- Protects the list

	unique_id: INTEGER
			-- To prevent deadlock, we always lock
			-- in order of ascending unique_id

invariant
	empty_check: check_empty
	mutex_set: mutex /= Void
	count_is_correct: check_count
end
