indexing
	description: "Concurrent linked queue using conditional variables. "
	library: "elogger"
	copyright: "Copyright (c) 2006, AXA Rosenberg, and others (see copyright.txt)"
	author: "Ulrich Weiss, and others"
	license: "Eiffel Forum License v2 (see forum.txt)"
	date: "$Date: 2009/09/14 $"
	revision: "$Revision: #6 $"
	archive: "$File: //rose/source/kernel/infrastructure/logging/concurrency/elog_concurrent_linked_queue_cv.e $"

class ELOG_CONCURRENT_LINKED_QUEUE_CV [G]

inherit
	THREAD_CONTROL
		redefine
			default_create
		end

create
	default_create

feature {NONE} -- Initialization

	default_create
			-- create a new empty queue.
		do
			first := Void
			last := Void
			create monitor.make
			create notempty.make
			create notfull.make
		end

feature -- Basic Operations

	put (a_value: G)
			-- Put a value at tail. Signals waiting threads
			-- if queue was empty.
		local
			l_node: like last
		do
			create l_node.make (a_value)

			monitor.lock
			from until not is_full or else is_closed loop
				notfull.wait (monitor)
			end

			-- raise if closed?
			if not is_closed then
				if is_empty then
					first := l_node
					last := l_node
				else
					last.set_next (l_node)
					last := l_node
				end

				count := count + 1
			end

			check empty_list_check:
				(first = Void implies last = Void) and
				(last = Void implies first = Void)
			end

			if count = 1 then
				-- if the queue was empty
				notempty.broadcast
			end

			monitor.unlock
		end

	get: G
			-- Get element at first position and return its value.
			-- if the queue is empty, blocks until there
			-- something in the queue.
			-- If the queue is closed, erroneously discard entries in the queue?
		do
			monitor.lock
			from until not is_empty or else is_closed loop
				notempty.wait (monitor)
			end

			if not is_empty then
				Result := first.value
				first := first.next
				if first = Void then
					last := Void
				end
				count := count - 1
			end

			check empty_list_check:
				(first = Void implies last = Void) and
				(last = Void implies first = Void)
			end

			if max_size /= 0 and then count = max_size - 1 then
				-- if the queue was full
				notfull.broadcast
			end

			monitor.unlock
		end

	wipe_out
			-- Remove all items from queue. The events
			-- are discarded.
		do
			monitor.lock
			first := Void
			last := Void
			notfull.broadcast
			monitor.unlock
		end

	set_max_size (size: INTEGER)
			-- sets 'max_size'.
		require
			size_valid: size >= 0
		do
			monitor.lock
			max_size := size
			monitor.unlock
		end

	close
			-- set is_closed to True,
			-- Signal all consumer threads that
			-- might be waiting.
		local
			l_count: INTEGER
		do
			from
				monitor.lock
				l_count := count
				monitor.unlock
			until
				l_count = 0
			loop
				monitor.lock
				notempty.broadcast
				monitor.unlock
				yield
				monitor.lock
				l_count := count
				monitor.unlock
			end

			is_closed := True
		end

	open
			-- sets is_closed to false.
		do
			is_closed := False
		end

feature -- Status Report

	max_size: INTEGER
			-- the maximum size of the buffer.
			-- if this value is zero, there is no size-limit.
			-- the putter will wait, until some space has
			-- become available.

	count: INTEGER
			-- number of elements contained

	is_empty : BOOLEAN
			-- is the queue empty?
		do
			Result := first = Void
		end

	is_full: BOOLEAN
			-- is the queue full?
		do
			Result := max_size /= 0 and then count >= max_size
		end

	is_closed: BOOLEAN
			-- is the queue closed?
			-- (suspends all activities)

	notfull, notempty: CONDITION_VARIABLE
			-- Condition for buffer not empty or not full.

feature {NONE} -- Implementation

	first: ELOG_CONCURRENT_LINKED_QUEUE_NODE [G]
			-- head of queue

	last: ELOG_CONCURRENT_LINKED_QUEUE_NODE [G]
			-- tail of queue

	monitor: ELOG_MUTEX
			-- mutex on both sides of the queue (head and tail)

invariant

	max_size_non_negative: max_size >= 0
	is_thread_capable: {PLATFORM}.is_thread_capable

end
