note

	description:

		"Logging appender that writes to a file."

	library: "elogger"
	copyright: "Copyright (c) 2006, AXA Rosenberg, and others (see copyright.txt)"
	license: "Eiffel Forum License v2 (see forum.txt)"
	author: "Ulrich Weiss, and others"
	last_editor: "$Author: e2 $"
	date: "$Date: 2014/07/04 $"
	revision: "$Revision: #8 $"
	path: "$File: //rose/source/kernel/infrastructure/logging/appenders/elog_file_appender.e $"

class ELOG_FILE_APPENDER[G -> FILE create make end]

inherit

	ELOG_APPENDER
		rename
			make as appender_make
		export
			{NONE} appender_make
		end

	ELOG_SHARED_INTERNAL_LOGGER

create

	make

feature {NONE} -- Initialisation

	make (new_name: STRING; appending: BOOLEAN; synchronous: BOOLEAN)
			-- Create a new file appender on the file
			-- with 'name'.
		require
			new_name_not_void: new_name /= Void
		do
			appender_make (new_name, synchronous)
			append_mode := appending
			create mutex.make
			set_backup_appender (create {ELOG_STDOUT_APPENDER}.make("stdout appender", synchronous), False)
			open_log
		end

feature -- Status Report

	append_mode: BOOLEAN
			-- Append to file or create new file?

feature -- Stream

	stream: G
			-- Stream to write log events to

feature {NONE} -- Implementation

	do_append_string (a_string: STRING)
			-- append 'a_string'
		do
			mutex.lock
			if stream.is_closed
			then
				-- If 'stream' is closed then presumably we never get here (we use the backup appender).  However,
				-- should we execute this code and 'stream' is closed, the C run-time will exit.  There will be
				-- no Eiffel stack trace!  By raising an exception we will stay within the Eiffel error handling
				-- mechanism.
				(create {EXCEPTIONS}).raise("Attempt to write to a closed file.")
			end

			stream.put_string (a_string)
			if do_flush then
				stream.flush
			end
			last_written_string := a_string
			mutex.unlock
		ensure then
			last_written_string_not_void: last_written_string /= Void
		rescue
			mutex.unlock
		end

	internal_close
			-- Release any resources for this appender.
		do
			mutex.lock
			if not stream.is_closed then
				if stream.is_open_write then
					stream.put_string(layout.footer)
				end
				stream.close
			end
			mutex.unlock
		rescue
			mutex.unlock
		end

	new_stream (a_name: STRING): G
			-- Create a platform-appropriate stream for `a_name'.
		require
			a_name_not_void: a_name /= Void
		do
			create Result.make (a_name)
		ensure
			new_stream_not_void: Result /= Void
		end

	open_log
			-- Open the log file taking into account 'append_mode'
		require
			not_open: not is_open
			stream_closed: stream /= Void implies stream.is_closed
		local
			rescued: BOOLEAN
		do
			if not rescued then
				mutex.lock
				unsafe_open_log
				mutex.unlock
			elseif not stream.is_open_write then
				-- Could not open file.  This can happen with gobo files.  With Eiffel Base files an exception is raised.
				internal_logger.log_fatal (create {ELOG_SIMPLE_EVENT}.
					make ("Error: Failed to open file stream: " + name))
				set_use_backup_appender (True, Void)
			end
		rescue
			mutex.unlock
			if not rescued then
				rescued := True
				retry
			end
		end

	open_stream ( a_stream: G )
			-- Open `a_stream' for writing/appending
			-- as per `append_mode'.
		require
			a_stream_not_void: a_stream /= Void
			a_stream_not_open: a_stream.is_closed
		do
			if append_mode then
				a_stream.open_append
			else
				a_stream.open_write
			end
		end

	unsafe_open_log
			-- Open the log file.
		require
			locked: mutex.is_owned_by_current_thread
		do
			stream := new_stream(name)
			open_stream(stream)
			if not stream.is_open_write then
				-- Could not open file.  This can happen with gobo files.  With Eiffel Base files an exception is raised.
				internal_logger.log_fatal(create {ELOG_SIMPLE_EVENT}.
					make ("Error: Failed to open file stream: " + name))
				set_use_backup_appender(True, Void)
			else
				stream.put_string(layout.header)
			end

			is_open := True
		ensure
			locked: mutex.is_owned_by_current_thread
		end

feature -- Removal

	dispose
			-- Close this appender when garbage collected. Perform
			-- minimal operations to release resources. Do not call
			-- other object as they may have been garbage collected.
		do
			is_open := False
		end

feature {NONE} -- Implementation

	mutex: ELOG_MUTEX
			-- a mutex

	last_written_string: STRING
			-- the most recently written string
invariant

	stream_not_void: stream /= Void
	mutex_not_void: mutex /= Void

end
