indexing

	description: "The hierarchy containing all the loggers. This is the main interface to be used. "
	library: "elogger"
	copyright: "Copyright (c) 2006, AXA Rosenberg, and others (see copyright.txt)"
	author: "Ulrich Weiss, and others"
	license: "Eiffel Forum License v2 (see forum.txt)"
	date: "$Date: 2009/09/14 $"
	revision: "$Revision: #7 $"
	archive: "$File: //rose/source/kernel/infrastructure/logging/elog_hierarchy.e $"

class ELOG_HIERARCHY

inherit

	ELOG_SEVERITY_CONSTANTS

create

	make

feature -- Initialisation

	make (severity: ELOG_SEVERITY)
			-- Create a logger hierarchy with a
			-- root logger with severity 'severity'.
		do
			create mutex.make
			create root.make (Current, Root_logger_name)
			root.set_severity (severity)
			create loggers.make_default
			loggers.put (root, Root_logger_name)
			enable_all
		ensure
			not_disabled: disabled = Disable_off
		end

feature -- Logger Factory

	logger (logger_name: STRING): ELOG_LOGGER
			-- Return a logger for 'name'. Initialise a new
			-- logger instance if necessary.
		require
			logger_name_exists: logger_name /= Void and then
				not logger_name.is_empty
			no_start_or_end_dots: logger_name.item (1) /= '.' and then
				logger_name.item (logger_name.count) /= '.'
			no_consecutive_dots: logger_name.substring_index ("..", 1) = 0
		do
			mutex.lock

			-- return root logger if asked for, otherwise search for it
			if logger_name.is_equal (Root_logger_name) then
				Result := root
			else
				if loggers.has (logger_name) then
					Result := loggers.item (logger_name)
				else
					create Result.make (Current, logger_name)
					loggers.force (Result, logger_name)
					set_logger_parents (Result)
				end
			end

			mutex.unlock
		ensure
			logger_exists: has (logger_name)
		end

	all_loggers: DS_LIST [ELOG_LOGGER]
			-- Return list of all current loggers. Root logger will be first in list
		do
			mutex.lock

			create {DS_LINKED_LIST [ELOG_LOGGER]} Result.make
			loggers.do_all(agent Result.put_last)
			Result.sort(create {DS_QUICK_SORTER[ELOG_LOGGER]}.make(
				create {KL_COMPARABLE_COMPARATOR [ELOG_LOGGER]}.make))

			mutex.unlock
		end

feature -- Status Report

	root: ELOG_LOGGER
			-- Root logger of this hierarchy

	disabled: INTEGER
			-- Globally disabled severity level. Disable_off if no priorities
			-- are disabled. Everything lower or equal than this
			-- value will be disabled (less important).

	has (name: STRING): BOOLEAN
			-- Does a logger with 'name' exist in this hierarchy?
		require
			name_exists: name /= Void
		do
			mutex.lock
			Result := loggers.has (name)
			mutex.unlock
		end

	Disable_off: INTEGER is -1
			-- Severity disabled off

	is_enabled_for (severity: ELOG_SEVERITY): BOOLEAN
			-- Is 'severity' level logging enabled for this hierarchy?
		do
			mutex.lock
			Result := disabled = Disable_off or disabled <= severity.level
			mutex.unlock
		end

	start_time: DT_DATE_TIME
			-- time the hierarchy was created

feature -- Status Setting

	disable (severity: ELOG_SEVERITY)
			-- Disable all logging for level 'severity'
		require
			severity_exists: severity /= Void
		do
			mutex.lock
			disabled := severity.level
			mutex.unlock
		ensure
			--severity_disabled: disabled = severity.level
		end

	disable_unknown
			-- Disable Unknown logging.
		do
			mutex.lock
			disabled := Unknown_int
			mutex.unlock
		end

	disable_information
			-- Disable Information logging.
		do
			mutex.lock
			disabled := Information_int
			mutex.unlock
		end

	disable_harmless
			-- Disable harmless logging.
		do
			mutex.lock
			disabled := Harmless_int
			mutex.unlock
		end

	disable_warning
			-- Disable warning logging.
		do
			mutex.lock
			disabled := Warning_int
			mutex.unlock
		end

	disable_minor
			-- Disable minor logging.
		do
			mutex.lock
			disabled := Minor_int
			mutex.unlock
		end

	disable_critical
			-- Disable critical logging.
		do
			mutex.lock
			disabled := Critical_int
			mutex.unlock
		end

	disable_fatal
			-- Disable fatal logging.
		do
			mutex.lock
			disabled := Fatal_int
			mutex.unlock
		end

	disable_all
			-- Disable all logging.
		do
			mutex.lock
			disabled := Max_severity_int
			mutex.unlock
		end

	enable_all
			-- Enable all logging.
		do
			mutex.lock
			disabled := Disable_off
			mutex.unlock
		end

	clear
			-- Clear all loggers from this hierarchy
		do
			mutex.lock
			loggers.wipe_out
			mutex.unlock
		end

	close_all
			-- Close all appenders for all loggers in this
			-- hierarchy.
		do
			mutex.lock

			loggers.do_all(agent {ELOG_LOGGER}.close_appenders)
			root.close_appenders

			mutex.unlock
		end

feature {NONE} -- Implementation

	loggers: DS_HASH_TABLE [ELOG_LOGGER, STRING]
			-- Table of loggers indexed by their fully qualified
			-- names.

	set_logger_parents (new_logger: ELOG_LOGGER)
			-- Split 'new_logger' name and check that
			-- appropriate parents exist. Create a new
			-- parent logger if necessary.
		require
			new_logger_exists: new_logger /= Void
			we_have_the_lock: mutex.is_owned_by_current_thread
		local
			logger_name, sub_name: STRING
			surrogate: ELOG_LOGGER
		do
			logger_name := new_logger.name
			-- if this is a logger with no sub parts
			-- then set its parent to the root
			if logger_name.occurrences ('.') = 0 then
				new_logger.set_parent (root)
			else
				-- if the new logger's name is "w.x.y.z",
				-- recurse through "w.x.y", "w.x" and "w" but
				-- not "w.x.y.z". If a parent does
				-- not exist with the sub logger
				-- then create it
				sub_name := logger_name.substring (
					1, last_index_of (logger_name, '.') - 1)
				if not loggers.has (sub_name) then
					surrogate := logger (sub_name)
					new_logger.set_parent (surrogate)
				else
					new_logger.set_parent (loggers.item (sub_name))
				end
			end
		end

	Root_logger_name: STRING = "root"
			-- Name of root logger.

	last_index_of (str: STRING; c: CHARACTER): INTEGER
			-- Position of last occurrence of 'c'.
			-- 0 if none.
		require
			str_not_void: str /= Void
		local
			i: INTEGER
		do
			from
				i := str.count
			until
				i < 1 or else str.item (i) = c
			loop
				i := i - 1
			end
			if i >= 0 then
				Result := i
			end
		ensure
			correct_place: Result > 0 implies str.item (Result) = c
		end

	invariant_tree_is_valid: BOOLEAN
			-- Is the tree structure valid?
		local
			loggerlist: DS_LINKED_LIST [ELOG_LOGGER]
		do
			mutex.lock

			Result := True
			if root = Void then
				Result := False
			else
				create loggerlist.make
				from
					loggerlist.force_last (root)
				until
					loggerlist.is_empty or not
					loggerlist.first.children.for_all(
						agent (
							e: ELOG_LOGGER;
							l: DS_LINKED_LIST [ELOG_LOGGER]): BOOLEAN
						do
							l.force_last(e)
							Result := e.parent = l.first
						end (?, loggerlist))
				loop
					loggerlist.remove_first
				end
				Result := loggerlist.is_empty
			end

			mutex.unlock
		end

	mutex: ELOG_MUTEX
			-- mutex on the hierarchy

invariant
	mutex_initialized: mutex /= Void
	tree_is_valid: invariant_tree_is_valid
	loggers_not_void: loggers /= Void
	has_root: has (Root_logger_name)
end
