note

	description:

		"A Logger. Has to be created by ELOG_HIERARCHY."

	library: "elogger"
	copyright: "Copyright (c) 2006, AXA Rosenberg, and others (see copyright.txt)"
	license: "Eiffel Forum License v2 (see forum.txt)"
	author: "Ulrich Weiss, and others"
	last_editor: "$Author: e2 $"
	date: "$Date: 2014/07/04 $"
	revision: "$Revision: #15 $"
	path: "$File: //rose/source/kernel/infrastructure/logging/elog_logger.e $"

class ELOG_LOGGER

inherit

	ELOG_SEVERITY_CONSTANTS
		export
			{NONE} all
		undefine
			is_equal
		end

	COMPARABLE

create

	make

feature {NONE} -- Initialisation

	make (a_hierarchy: ELOG_HIERARCHY; logger_name: STRING)
			-- Create new logger with 'logger_name'
			-- Defaults are:
			-- 	no appenders attached
			-- 	is_additive = True
		require
			logger_name_exists: logger_name /= Void and then logger_name.count > 0
		do
			hierarchy := a_hierarchy
			name := logger_name
			create appenders
			create internal_appenders
			create children
			is_additive := True
			severity := inherited_severity
			create appender_mutex.make
			create severity_mutex.make
		ensure
			hierarchy_set: hierarchy = a_hierarchy
			name_set: name = logger_name
			additive: is_additive
			appenders_valid: appenders /= Void and then appenders.count = 0
			internal_severity_void: internal_severity = Void
		end

feature -- Status Report

	name: STRING
			-- Name of this logger

	severity: ELOG_SEVERITY
			-- The severity is updated every time a change
			-- is made to the logger tree

	parent: ELOG_LOGGER
			-- The parent of this logger.

	hierarchy: ELOG_HIERARCHY
			-- The hierarchy in which this logger operates.

	is_additive: BOOLEAN
			-- Should appenders of this logger's parent be used?

	is_severity_set: BOOLEAN
			-- Does this logger have a severity set? False if the severity
			-- will be inherited from an ancestor.
		do
			Result := internal_severity /= Void
		end

	is_enabled_for (check_severity: ELOG_SEVERITY): BOOLEAN
			-- Is this logger enabled for 'check_severity'?
		require
			check_severity_exists: check_severity /= Void
		do
			severity_mutex.lock
			Result := hierarchy.disabled < check_severity.level
				and then severity <= check_severity
			severity_mutex.unlock
		end

	has_appender (an_appender: ELOG_APPENDER): BOOLEAN
			-- does this logger have the appender?
		require
			an_appender_not_void: an_appender /= Void
		do
			Result := appenders.has (an_appender)
		end

	has_appender_by_name (an_appender: STRING): BOOLEAN
			-- does this logger have the appender?
		require
			an_appender_not_void: an_appender /= Void
		do
			-- Copy-paste to avoid infinite loop w/
			-- get_appender_by_name's precondition.
			Result := Void /= appenders.find (agent internal_check_appender_name (?, an_appender))
		end

	get_appender_by_name (an_appender: STRING): ELOG_APPENDER
			-- returns the given appender
		require
			an_appender_not_void: an_appender /= Void
			has_appender: has_appender_by_name (an_appender)
		do
			Result := appenders.find (agent internal_check_appender_name (?, an_appender))
		ensure
			Result_not_void: Result /= Void
		end

	appenders_names: LINKED_LIST [STRING]
			-- Name of appenders contained.
		do
			create Result.make
			appenders.do_all (agent add_appender_name (Result, ?))
		ensure
			appenders_names_not_void: Result /= Void
			no_void_name: not Result.has (Void)
			correct_number_of_names: Result.count = appenders.count
		end

	override_layout: ELOG_LAYOUT
			-- this layout is added to the event upon logging and
			-- will override the default layout attached to the
			-- appender.

feature -- Status Setting

	set_severity (new_severity: ELOG_SEVERITY)
			-- Set the priority for this logger. Pass Void to unset the
			-- priority and use an ancestor's.
		require
			root_severity_not_void: hierarchy.root = Current
				implies new_severity /= Void
		local
			old_severity: INTEGER
		do
			severity_mutex.lock

			old_severity := severity.level
			internal_severity := new_severity

			if new_severity = Void then
				-- severity is void, get severity from ancestors
				severity := inherited_severity
			else
				severity := new_severity
			end

			-- update children with void severity.
			if old_severity /= severity.level then
				children.do_all(agent {ELOG_LOGGER}.
					inform_that_parents_severity_changed(severity))
			end

			severity_mutex.unlock
		end

	set_flush (a_flush: BOOLEAN)
			-- Sets whether appenders should flush each time they write an event
		do
			appenders.do_all (agent {ELOG_APPENDER}.set_flush (a_flush))
		end

	add_appender (new_appender: ELOG_APPENDER)
			-- Add 'new_appender' to the list of
			-- appenders for this logger.
		require
			new_appender_exists: new_appender /= Void
		do
			appender_mutex.lock

			internal_appenders.put_if_new(new_appender)
			appenders.put_if_new(new_appender)

			appender_mutex.unlock
			recursive_add_appender (new_appender)
		ensure
			--appender_added: has (new_appender)
				-- and internal_appenders.has (new_appender)
		end

	remove_appender (appender_name: STRING)
			-- Remove appender with name 'appender_name'
			-- Iterates to find matching appender.
		require
			name_exists: appender_name /= Void
		do
			internal_appenders.twin.do_if(agent internal_remove_appender (?, appender_name), agent internal_check_appender_name (?, appender_name))
		end

	wipe_out_appenders
			-- remove all appenders from the list.
		do
			appenders.wipe_out
		end

	close_appenders
			-- Close all appenders (this level only).
			-- NOTE: you will have to reopen the appenders, if
			-- you want to log anything. Closing the appenders
			-- will also close all files and close any open
			-- streams.
		do
			internal_appenders.do_all(agent {ELOG_APPENDER}.close)
		end

	set_additive (flag: BOOLEAN)
			-- Set additive status to 'flag'.
		do
			if flag /= is_additive then
				is_additive := flag
				if not is_additive then
					appenders.copy (internal_appenders)
				else
					-- switched from not additive to additive.
					-- add all ancestors appenders.
					appenders := inherited_appenders
				end
			end
		ensure
			flag_set: flag = is_additive
		end

	set_override_layout (a_override_layout: ELOG_LAYOUT)
			-- sets `override_layout'
		do
			override_layout := a_override_layout
		end

	import_settings (a_logger: ELOG_LOGGER)
			-- merges the current logger with 'a_logger'.
			-- overwrites settings from 'a_logger' to Current
			-- (appenders, is_additive and severity)
		require
			a_logger_not_void: a_logger /= Void
		do
			-- clear appenders
			internal_appenders.wipe_out
			appenders.wipe_out
			internal_appenders := a_logger.appenders.twin

			a_logger.appenders.do_all(agent add_appender)

			-- add all appenders
			set_additive (a_logger.is_additive)
			if is_additive  then
				appenders.append(inherited_appenders)
			end

			-- set severity
			set_severity (a_logger.severity)
		end

feature -- Logging

	log_unknown (message: ELOG_EVENT)
			-- Logs the message at unknown severity, if enabled for that level.
		require
			message_exists: message /= Void
		do
			if is_enabled_for (Unknown_severity) then
				forced_log (Unknown_severity, message)
			end
		end

	log_debug (message: ELOG_EVENT)
			-- Logs the message at debug severity, if enabled for that level.
		require
			message_exists: message /= Void
		do
			if is_enabled_for (Debug_severity) then
				forced_log (Debug_severity, message)
			end
		end

	log_information (message: ELOG_EVENT)
			-- Logs the message at information severity, if enabled for that level.
		require
			message_exists: message /= Void
		do
			if is_enabled_for (Information_severity) then
				forced_log (Information_severity, message)
			end
		end

	log_notice (message: ELOG_EVENT)
			-- Logs the message at notice severity, if enabled for that level.
		require
			message_exists: message /= Void
		do
			if is_enabled_for (Notice_severity) then
				forced_log (Notice_severity, message)
			end
		end

	log_harmless (message: ELOG_EVENT)
			-- Logs the message at harmless severity, if enabled for that level.
		require
			message_exists: message /= Void
		do
			if is_enabled_for (Harmless_severity) then
				forced_log (Harmless_severity, message)
			end
		end

	log_warning (message: ELOG_EVENT)
			-- Logs the message at warning severity, if enabled for that level.
		require
			message_exists: message /= Void
		do
			if is_enabled_for (Warning_severity) then
				forced_log (Warning_severity, message)
			end
		end

	log_minor (message: ELOG_EVENT)
			-- Logs the message at minor severity, if enabled for that level.
		require
			message_exists: message /= Void
		do
			if is_enabled_for (Minor_severity) then
				forced_log (Minor_severity, message)
			end
		end

	log_critical (message: ELOG_EVENT)
			-- Logs the message at critical severity, if enabled for that level.
		require
			message_exists: message /= Void
		do
			if is_enabled_for (Critical_severity) then
				forced_log (Critical_severity, message)
			end
		end

	log_fatal (message: ELOG_EVENT)
			-- Logs the message at fatal severity, if enabled for that level.
		require
			message_exists: message /= Void
		do
			if is_enabled_for (Fatal_severity) then
				forced_log (Fatal_severity, message)
			end
		end

	log (event_severity: ELOG_SEVERITY; message: ELOG_EVENT)
			-- Log a general message.
		require
			event_severity_exists: event_severity /= Void
			message_exists: message /= Void
		do
			if is_enabled_for (event_severity) then
				forced_log (event_severity, message)
			end
		end

feature -- Backwards compatibility with L4E

	debugging (message: ANY)
			-- Log a message object with the priority Debug.
			-- First checks if this logger is enabled
			-- by comparing the priority of this logger
			-- with the Debug priority. If this logger
			-- is debug enabled, then it converts the
			-- message object to a string by invoking
			-- 'out'. It then proceeds to call all the
			-- registered appenders in this logger and
			-- also higher in the hierarchy depending on
			-- the value of the additivity flag.
		require
			message_exists: message /= Void
		do
			log_debug(create {ELOG_SIMPLE_EVENT}.make(message.out))
		end

	info (message: ANY)
			-- Log a message object with the priority Info.
			-- First checks if this logger is enabled
			-- by comparing the priority of this logger
			-- with the Info priority. If this logger
			-- is info enabled, then it converts the
			-- message object to a string by invoking
			-- 'out'. It then proceeds to call all the
			-- registered appenders in this logger and
			-- also higher in the hierarchy depending on
			-- the value of the additivity flag.
		require
			message_exists: message /= Void
		do
			log_information(create {ELOG_SIMPLE_EVENT}.make(message.out))
		end

	notice (message: ANY)
			-- Log a message object with the priority Notice.
			-- First checks if this logger is enabled
			-- by comparing the priority of this logger
			-- with the Notice priority. If this logger
			-- is notice enabled, then it converts the
			-- message object to a string by invoking
			-- 'out'. It then proceeds to call all the
			-- registered appenders in this logger and
			-- also higher in the hierarchy depending on
			-- the value of the additivity flag.
		require
			message_exists: message /= Void
		do
			log_notice(create {ELOG_SIMPLE_EVENT}.make(message.out))
		end

	warn (message: ANY)
			-- Log a message object with the priority Warn.
			-- First checks if this logger is enabled
			-- by comparing the priority of this logger
			-- with the Warn priority. If this logger
			-- is warn enabled, then it converts the
			-- message object to a string by invoking
			-- 'out'. It then proceeds to call all the
			-- registered appenders in this logger and
			-- also higher in the hierarchy depending on
			-- the value of the additivity flag.
		require
			message_exists: message /= Void
		do
			log_warning(create {ELOG_SIMPLE_EVENT}.make(message.out))
		end

	error (message: ANY)
			-- Log a message object with the priority Error.
			-- First checks if this logger is enabled
			-- by comparing the priority of this logger
			-- with the Error priority. If this logger
			-- is info enabled, then it converts the
			-- message object to a string by invoking
			-- 'out'. It then proceeds to call all the
			-- registered appenders in this logger and
			-- also higher in the hierarchy depending on
			-- the value of the additivity flag.
		require
			message_exists: message /= Void
		do
			log_critical(create {ELOG_SIMPLE_EVENT}.make(message.out))
		end

	fatal (message: ANY)
			-- Log a message object with the priority Fatal.
			-- First checks if this logger is enabled
			-- by comparing the priority of this logger
			-- with the Fatal priority. If this logger
			-- is error enabled, then it converts the
			-- message object to a string by invoking
			-- 'out'. It then proceeds to call all the
			-- registered appenders in this logger and
			-- also higher in the hierarchy depending on
			-- the value of the additivity flag.
		require
			message_exists: message /= Void
		do
			log_fatal(create {ELOG_SIMPLE_EVENT}.make(message.out))
		end

feature {ELOG_DISPATCHER, ELOG_LOGGER} -- Implementation

	appenders: ELOG_CONCURRENT_LINKED_LIST [ELOG_APPENDER]
			-- Appenders for this logger.

	appender_mutex: ELOG_MUTEX
			-- mutex for 'appenders'

feature {NONE} -- Implementation

	add_appender_name (a_names: LINKED_LIST [STRING]; a_appender: ELOG_APPENDER)
			-- Adds the name of `a_appender' into `a_names'
		require
			a_names_not_void: a_names /= Void
			no_void_name: not a_names.has (Void)
			a_appender_not_void: a_appender /= Void
		do
			a_names.extend (a_appender.name)
		ensure
			one_more_name: a_names.count = old a_names.count + 1
			name_added: a_names.last ~ a_appender.name
		end

	forced_log (event_severity: ELOG_SEVERITY; message: ELOG_EVENT)
			-- Create new logging event and send to appenders.
			--  FIXME -- this stuff to extract the caller from the stack should go in one of the appenders/formatters instead of here
		require
			severity_exists: event_severity /= Void
			message_exists: message /= Void
		local
			l_call_stack: ARRAY[STRING]
		do
			debug("elog_logger_caller")
				if  attached env.item ("ELOG_LOGGER_CALLER") as l_env_set and then
					not l_env_set.is_empty then
						-- ELOG_LOGGER_CALLER set to True or Yes is the assumption here
					if ("ty").has_code (l_env_set.item (1).as_lower.natural_32_code)
					then
						l_call_stack := call_stack.call_stack
						if l_call_stack.valid_index( l_call_stack.lower + forced_log_caller_stack_offset) then
							if {l_caller: STRING} l_call_stack.item(l_call_stack.lower + forced_log_caller_stack_offset) then
								message.set_message( message.message + "%N%T%T%T%T#====> caller ====> " + l_caller)
							end
						end
					end
				end
			end
			message.set_severity (event_severity)
			message.set_logger (Current)
			message.set_override_layout (override_layout)

			appenders.do_all(agent {ELOG_APPENDER}.append(message))
		end

	call_stack: EIFFEL_CALL_STACK is
			-- a call stack generator
		once
			create Result
		end

	env: EXECUTION_ENVIRONMENT is
			-- access to the process environment
		once
			create Result
		end

	forced_log_caller_stack_offset: INTEGER is 3
			-- number of stack frames back to get to real caller of logger.XXX

	internal_severity: ELOG_SEVERITY
			-- The severity of this level only.

	internal_appenders: ELOG_CONCURRENT_LINKED_LIST [ELOG_APPENDER]
			-- The appenders of this level only.

	invariant_children_test: BOOLEAN
			-- Do all the childrens parents equal current?
		do
			Result := children.for_all(
				agent (i: ELOG_LOGGER): BOOLEAN do
					Result := i.parent = Current end)
		end

	invariant_appenders_are_inherited: BOOLEAN
			-- checks if all appenders are inherited correctly
		do
			if is_additive then
				Result := appenders ~ inherited_appenders
			else
				Result := appenders ~ internal_appenders
			end
		end

	severity_mutex: ELOG_MUTEX
			-- mutex for 'severity'

feature {ELOG_LOGGER, ELOG_HIERARCHY} -- Internal

	children: ELOG_CONCURRENT_LINKED_LIST [ELOG_LOGGER]
			-- List of child loggers.

	set_parent (new_parent: ELOG_LOGGER)
			-- Set the parent for this logger.
		require
			not_root: hierarchy.root /= Current
			new_parent_exists: new_parent /= Void
		local
			l_parent: like new_parent
		do
			if parent /= Void and new_parent /= parent then
				-- we had a different parent, now change to new_parent
				parent.children.delete (Current)
			end

			-- The temporary (l_parent) used below is
			-- necessary to maintain the invariants.
			l_parent := new_parent
			l_parent.children.put_if_new(Current)
			parent := l_parent

			severity := inherited_severity

			if not is_additive then
				appenders.copy (internal_appenders)
			else
				-- switched from not additive to additive.
				-- add all ancestors appenders.
				appenders := inherited_appenders
			end
		end

feature {ELOG_LOGGER} -- Internal

	inform_that_parents_severity_changed (s: ELOG_SEVERITY)
			-- informs logger, that its parents severity has
			-- changed.
		require
			s_not_void: s /= Void
		do
			severity_mutex.lock
			if internal_severity = Void then
				severity := s
				children.do_all(agent {ELOG_LOGGER}.
					inform_that_parents_severity_changed(s))
			end
			severity_mutex.unlock
		end

	inherited_severity: ELOG_SEVERITY
			-- recurse through ancestors to receive first
			-- non-void severity
		do
			Result := internal_severity
			if parent = Void then
				if internal_severity = Void then
					Result := unknown_severity
				end
			else
				if internal_severity = Void then
					Result := parent.inherited_severity
				end
			end
		ensure
			result_not_void: Result /= Void
		end

	inherited_appenders: ELOG_CONCURRENT_LINKED_LIST [ELOG_APPENDER]
			-- recurse through ancestors to receive all
			-- appenders (including own appenders)
		do
			create Result.make_from_like(internal_appenders)

			if parent /= Void then
				Result.prepend(parent.inherited_appenders)
			end
		ensure
			result_not_void: Result /= Void
		end

	recursive_add_appender (an_appender: ELOG_APPENDER)
			-- a ancestor (or current) has added the appender 'an_appender'
			-- if is_additive is enabled, add the appender
		require
			an_appender_not_void: an_appender /= Void
		do
			appender_mutex.lock
			if is_additive then
				appenders.put_if_new(an_appender)
			end

			children.do_all(agent {ELOG_LOGGER}.
				recursive_add_appender(an_appender))
			appender_mutex.unlock
		ensure
--			appender_added: is_additive
--				implies appenders.has(an_appender)
		end

	recursive_remove_appender (appender_name: STRING)
			-- have all the children recompute their appenders
			-- not very performant, but remove won't be called that
			-- often.
		require
			appender_not_void: appender_name /= Void
		do
			-- quick hack, but does the trick.
			if is_additive then
				-- todo: optimize (not so important)
				appenders := inherited_appenders
			end

			children.do_all(agent {ELOG_LOGGER}.
				recursive_remove_appender(appender_name))
		end

	internal_remove_appender (a: ELOG_APPENDER; n: STRING) is
			-- Remove appender `a' with name `n'
		require
			non_void_args: a /= Void and n /= Void
		do
			internal_appenders.delete(a)
			appenders.delete (a)
			-- appender does not exist in ancestors and
			-- has to be removed from children
			if is_additive then
				-- recompute inherited appenders
				appenders.copy(inherited_appenders)
			end
			recursive_remove_appender(n)
		end

	internal_check_appender_name (a: ELOG_APPENDER; n: STRING): BOOLEAN is
			-- Compare `a.name' and `n'
		require
			non_void_args: a /= Void and n /= Void
		do
			Result := a.name ~ n
		end

feature -- Comparison

	infix "<" (other: like Current): BOOLEAN
			-- Is current object less than `other'?
		do
			Result := name < other.name
		end

invariant

	name_exists: name /= Void and then not name.is_empty
	appenders_exist: appenders /= Void
	children_exist: children /= Void
	current_in_parents_children: parent /= Void implies parent.children.has (Current)
	childrens_parent_equal_current: invariant_children_test

end
