indexing
	description: "Factory to generate loggers"

	author: "Emmanuel Bouyer"
	last_editor: "$Author: c4 $"
	date: "$Date: 2014/07/11 $"
	revision: "$Revision: #12 $"
	path: "$File: //rose/source/kernel/infrastructure/logging/elog_logger_factory.e $"

class ELOG_LOGGER_FACTORY [G -> FILE create make end]

inherit

	ELOG_SHARED_HIERARCHY
		export
			{NONE} all
		end

	ANY

create

	make

feature {NONE} -- Initialization

	make
			-- Initialize current
		do
		end

feature -- Access

	new_logger (a_name, a_filename: STRING; a_use_std_out: BOOLEAN): ELOG_LOGGER
			-- Logger of name `a_name'
			-- messages are written in `a_filename'
			-- messages are also put in standard out if `a_use_std_out'
			-- Use UTC timestamp
		require
			name_not_void: a_name /= Void
			filename_not_void: a_filename /= Void
		local
			l_config: ELOG_CONFIGURATION
		do
			create l_config.make
			l_config.set_logger_name (a_name)
			l_config.set_filename (a_filename)
			l_config.set_use_std_out (a_use_std_out)
			l_config.set_max_size (-1)
			l_config.set_max_backups (-1)
			Result := new_complex_logger (l_config)
		ensure
			logger_not_void: Result /= Void
			same_name: Result.name ~ a_name
		end

	new_complex_logger (a_config: ELOG_CONFIGURATION): ELOG_LOGGER
			-- Logger of name `a_name'
			-- messages are written in `a_filename'
			-- following the include/exclude filters (to exclude/include certain messages)
			-- messages are also put in standard out if `a_use_std_out'
			-- Use local or UTC timestamp (depending on `a_use_utc')
			-- a_max_size: files not to exceed this size (by more than a single logged message)
							 -- cases: -1=no maximum (therefore no rolling), 0=default_max_size, >0=use the value provided
			-- a_max_backups: number of rolling files not to exceed this number
							 -- cases: -1=no maximum, 0=default_max_backups, >0=use the value provided
		require
			a_config_not_void: a_config /= Void
		local
			l_appender: ELOG_APPENDER
			l_max_size: INTEGER
			l_max_backups: INTEGER
		do
			if a_config.max_size = 0 then
				l_max_size := default_max_size
			else
				l_max_size := a_config.max_size
			end
			if a_config.max_backups = 0 then
				l_max_backups := default_max_backups
			else
				l_max_backups := a_config.max_backups
			end

			if a_config.filename.ends_with (Open_file_token) and open_file /= Void then
				create {ELOG_OPEN_FILE_APPENDER} l_appender.make (open_file, a_config.synchronous_write)
			elseif l_max_size > 0 then
				create {ELOG_ROLLING_FILE_APPENDER [G]} l_appender.make (a_config.filename,
					l_max_size, l_max_backups,
					a_config.should_append, a_config.synchronous_write)
			else
				create {ELOG_FILE_APPENDER [G]} l_appender.make (a_config.filename,
					a_config.should_append, a_config.synchronous_write)
			end
			l_appender.set_layout (create {ELOG_MINIMAL_LAYOUT})
			l_appender.set_use_utc (a_config.use_utc)

			-- add filters from configuration
			add_filters (l_appender, a_config.exclude_filters, a_config.include_filters)

			Result := Log_hierarchy.logger (a_config.logger_name)
			if not Result.has_appender_by_name (l_appender.name) then
				Result.add_appender (l_appender)
			end
			Result.set_override_layout (create {ELOG_DATE_TIME_LAYOUT})

			if a_config.use_std_out and not Result.has_appender_by_name ("std out") then
				create {ELOG_STDOUT_APPENDER} l_appender.make ("std out", True)
				l_appender.set_layout (create {ELOG_DATE_TIME_LAYOUT})
				l_appender.layout.set_header (l_appender.layout.header + "Writing to log file: " + a_config.filename + l_appender.layout.nl)
				l_appender.write_header
				add_filters (l_appender, a_config.exclude_filters, a_config.include_filters)
				Result.add_appender (l_appender)
			end
		ensure
			new_complex_logger_not_void: Result /= Void
			same_name: Result.name.is_equal (a_config.logger_name)
		end

feature -- Setting

	set_open_file (a_file: FILE) is
			-- Set `open_file' to `a_file'
		require
			a_file_not_void: a_file /= Void
			a_file_opened_for_writing: a_file.is_open_write
		do
			open_file := a_file
		ensure
			open_file_aliased: open_file = a_file
		end

feature {NONE} -- Implementation
	
	Open_file_token: STRING is "_already_open_"
			-- Cfg file entry to indicate use of an already-opened file

	open_file: FILE
			-- File that will be logged to when requested by `Open_file_token'

	add_filters (a_appender: ELOG_APPENDER; a_exclude_filters, a_include_filters: ARRAY [STRING])
			-- add filters from configuration file
			-- this is how multiple filters can be added.
			--		[/fx/logging/exclude_filter]
			--		>"value1"
			--		>"value2"
			--		>"value3"
			--		[/fx/logging/include_filter]
			--		>"value4"
			--		>"value5"
			--		>"value6"
			-- these filters are ANDed. Include filters filter all log messages out that DO NOT contain
			-- the given string.
		require
			appender_not_void: a_appender /= Void
			exclude_filters_not_void: a_exclude_filters /= Void
			no_void_exclude_filter: not a_exclude_filters.has (Void)
			include_filters_not_void: a_include_filters /= Void
			no_void_include_filter: not a_include_filters.has (Void)
		local
			l_filters: DS_LINKED_LIST [ELOG_FILTER]
			or_filter: ELOG_OR_FILTER
			i: INTEGER
		do
			create l_filters.make
			from i := a_exclude_filters.lower until i > a_exclude_filters.upper loop
				l_filters.force_last (create {ELOG_STRING_MATCH_FILTER}.make (a_exclude_filters.item (i), False))
				i := i + 1
			end
			from i := a_include_filters.lower until i > a_include_filters.upper loop
				l_filters.force_last (create {ELOG_NEUTRAL_TO_REJECT_FILTER}.make (create {ELOG_STRING_MATCH_FILTER}.make (a_include_filters.item (i), True)))
				i := i + 1
			end

			if l_filters.count = 1 then
				-- there is only one filter
				a_appender.set_filter (l_filters.first)
			elseif l_filters.count > 1 then
				from
					l_filters.start
					l_filters.forth
					or_filter := create {ELOG_OR_FILTER}.make (l_filters.first, l_filters.item_for_iteration, False)
					l_filters.forth
				until
					l_filters.after
				loop
					or_filter.add_filter (l_filters.item_for_iteration)
					l_filters.forth
				end
				a_appender.set_filter (or_filter)
			end
		end

feature {NONE} -- Default constants

	default_max_size : INTEGER is 5_000_000
			-- for rolling log files, the default maximum size, if not otherwise specified

	default_max_backups : INTEGER is 10
			-- for rolling log files, the default maximum number of backups, if not otherwise specified

invariant

	open_file_opened_for_writing: open_file /= Void implies open_file.is_open_write

end
