indexing
	description: "Logging appender that writes to standard a file that is rolled when it reaches a certain size."
	library: "elogger"
	copyright: "Copyright (c) 2006, AXA Rosenberg, and others (see copyright.txt)"
	author: "Ulrich Weiss, and others"
	license: "Eiffel Forum License v2 (see forum.txt)"
	date: "$Date: 2010/12/06 $"
	revision: "$Revision: #8 $"
	archive: "$File: //rose/source/kernel/infrastructure/logging/appenders/elog_rolling_file_appender.e $"
	last_review: "$Review: $"
	last_editor: "$Author: rj $"

class ELOG_ROLLING_FILE_APPENDER[G -> FILE create make end]

inherit

	ELOG_FILE_APPENDER[G]
		rename
			make as file_appender_make
		export
			{NONE} file_appender_make
		redefine
			do_append
		end

	KL_SHARED_FILE_SYSTEM
		export
			{NONE} all
		end

create

	make

feature {NONE} -- Initialisation

	make (file_name: STRING; max_size, number_of_backups: INTEGER; appending: BOOLEAN; synchronous: BOOLEAN) is
			-- Create a new file appender on the file
			-- with 'file_name'. Roll the log over to a backup file when it reaches 'max_size'
			-- bytes in size. Keep a maximum of 'number_of_backups' backup files.
			-- Append to an existing log file if 'appending'.
		require
			name_exists: file_name /= Void
			name_not_empty: not file_name.is_empty
			sensible_max_size: max_size >= 1
			--positive_number_of_backups: number_of_backups >= 0
		do
			max_file_size := max_size
			max_number_of_backups := number_of_backups
			file_appender_make (file_name, appending, synchronous)

			-- get current file size
			if is_open
			then
				-- This will abort (with no Eiffel stack trace) if the file is not open.
				-- It isn't clear what this code does.  Perhaps it is needed to find the size of the file.
				-- Note that 'stream' used to be a KI_TEXT_OUTPUT_FILE.
				stream.flush
				stream.close
				stream.open_append
			end

			--write_count := file_system.file_count (file_system.pathname_from_file_system (stream.name, file_system))
			write_count := stream.count
			if write_count = -1 then
				-- Only gobo files will return -1.
				internal_logger.log_fatal (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Unable to determine size of log file '" + stream.name + "' for rollover"))
				write_count := 0
			end
		ensure
			log_file_open: stream.is_open_write or use_backup_appender
		end

feature -- Basic Operations

	do_append (event: ELOG_EVENT) is
			-- Log event on this appender.
		do
			mutex.lock
			if rollover_required then
				rollover
			end
			Precursor( event )
			write_count := write_count + last_written_string.count
			mutex.unlock
		rescue
			mutex.unlock
		end

feature {NONE} -- Implementation

	max_file_size: INTEGER
			-- The maximum size the log file can reach
			-- before being rolled over.

	max_number_of_backups: INTEGER
			-- The number of backup files to keep. If
			-- zero then no backups will be made.

	rollover_required: BOOLEAN is
			-- Has the current log file reached the max_file_size?
		do
			Result := not use_backup_appender and then stream.count >= max_file_size
		end

	rollover is
			-- Rollover the current file by renaming (if
			-- backups are required) or deleting. Open a
			-- new file with the same name.
		local
			l_new_file: G
			l_new_name: STRING
		do
			mutex.lock
			-- rename current log with date-time stamp
			-- delete oldest logs if there are more than max_number_of_backups

			from
				l_new_name := timestamped_name
				create l_new_file.make( l_new_name )
			until
				not l_new_file.exists
			loop
				l_new_name := timestamped_name
				create l_new_file.make( l_new_name )
			end

			-- remove the oldest backup if not keeping infinite number of backups
			if max_number_of_backups > 0 then
				remove_oldest_backups
			end

			-- rename current log l_file
			stream.put_string( "Log rolled: " + l_new_name + "%N" )
			stream.put_string( layout.footer )
			stream.flush
			stream.close
			-- be careful that you haven't inadvertently opened the same file twice, or this `change_name' will fail :o
			stream.change_name( l_new_name )

			-- re-open log
			is_open := False
			open_log
			write_count := 0
			mutex.unlock
		rescue
			if not stream.is_closed then
				stream.flush
				stream.close
			end
			is_open := False
			mutex.unlock
		end

	timestamped_name: STRING is
			-- create a file anme based on `name' plus a timestamp of the form "yyyy[0]mm[0]ddT[0]hh[0]mi[0]ss.ff3"
		local
			l_now: DATE_TIME
		do
			create l_now.make_now
			-- the following split into three parts is a "Sleazy Hack" (TM - Chris Lum) to accomodate an apparent bug in `formatted_out' where it ONLY replaces the milliseconds part in the output if it is included in a fill format template (as shown in the feature comment)
			Result := name + "." + l_now.date.formatted_out("yyyy[0]mm[0]dd") + "T" + l_now.time.formatted_out("[0]hh[0]mi[0]ss") + l_now.time.formatted_out(".ff3")
		end

	remove_oldest_backups is
			-- cleanup up old log files when the number of files exceeds `max_number_of_backups'
		require
			max_number_of_backups_greater_than_zero: max_number_of_backups > 0
		local
			l_directory: KL_DIRECTORY
			l_file_names: ARRAY[STRING]
			l_full_path_file_names: ARRAY[STRING]
			l_delete_file_names:  ARRAY[STRING]
			l_sorted: DS_ARRAYED_LIST[STRING]
		do
			create l_directory.make( file_system.absolute_parent_directory(name) )
			l_directory.open_read
			l_file_names := l_directory.filenames
			create l_full_path_file_names.make(1,0)
			create l_delete_file_names.make(1,0)
			l_file_names.do_all( agent add_full_path_to_array(l_directory, ?, l_full_path_file_names))
			l_full_path_file_names.do_if( agent add_to_array( ?, l_delete_file_names), agent matches_but_not_equal_to_name )
			if l_delete_file_names.count > max_number_of_backups then
				create l_sorted.make_from_array (l_delete_file_names)
				l_sorted.sort (create {DS_QUICK_SORTER [STRING]}.make (create {KL_COMPARABLE_COMPARATOR [STRING]}.make))
				l_delete_file_names := l_sorted.to_array
				l_delete_file_names := l_delete_file_names.subarray( l_delete_file_names.lower, l_delete_file_names.upper - max_number_of_backups )
				--l_my_log_files.do_all( agent delete_file_in_directory( ? ) )
				l_delete_file_names.do_all( agent delete_file(  ? ) )
			end
		end

	delete_file( a_file_name: STRING ) is
			-- delete a file named `a_file_name' in `a_directory'
		local
			l_file: G
		do
			create l_file.make( a_file_name )
			if l_file.exists then
				l_file.delete
			end

		end

	delete_file_in_directory( a_directory: KL_DIRECTORY; a_file_name: STRING ) is
			-- delete a file named `a_file_name' in `a_directory'
		require
			a_directory_not_void: a_directory /= Void
			a_file_name_not_void: a_file_name /= Void
		local
			l_file_name: FILE_NAME
			l_file: G
		do
			create l_file_name.make
			l_file_name.set_directory( a_directory.name )
			l_file_name.set_file_name( a_file_name )
			create l_file.make( l_file_name )
			if l_file.exists then
				l_file.delete
			end

		end

	add_full_path_to_array ( a_directory: KL_DIRECTORY; a_name: STRING; a_array: ARRAY[STRING] ) is
			-- add an element to an array
		local
			l_file_name: FILE_NAME
		do
			create l_file_name.make
			l_file_name.set_directory( a_directory.name )
			l_file_name.set_file_name( a_name )
			if not a_array.has( l_file_name ) then
				a_array.force( l_file_name, a_array.upper + 1 )
			end
		end

	add_to_array ( a_name: STRING; a_array: ARRAY[STRING] ) is
			-- add an element to an array
		do
			if not a_array.has( a_name ) then
				a_array.force( a_name, a_array.upper + 1 )
			end
		end

	matches_but_not_equal_to_name( a_file_name: STRING ): BOOLEAN is
			-- does `a_name' begin with `name'?
		do
			if a_file_name.starts_with( name ) then
				if a_file_name.count /= name.count then
					Result := True
				end
			end
		end

	write_count: INTEGER_64
			-- save the number of characters written to the file
end
