indexing
	description: "Logging XML configuration parser."
	library: "elogger"
	copyright: "Copyright (c) 2006, AXA Rosenberg, and others (see copyright.txt)"
	author: "Ulrich Weiss, and others"
	license: "Eiffel Forum License v2 (see forum.txt)"
	date: "$Date: 2009/11/18 $"
	revision: "$Revision: #8 $"
	archive: "$File: //rose/source/kernel/infrastructure/logging/config/elog_xml_config_parser.e $"

class ELOG_XML_CONFIG_PARSER[G -> FILE create make end]

inherit

	ELOG_SHARED_INTERNAL_LOGGER
		export
			{NONE} all
		end

	ELOG_SEVERITY_CONSTANTS
		export
			{NONE} all
		end

	ELOG_XML_CONFIG_CONSTANTS
		export
			{NONE} all
		end
	ANY
create

	make, make_with_hierarchy

feature -- Initialisation

	make (config_file_name: STRING) is
			-- Create a new config parser that will parse the file
			-- named 'config_file_name'.
		require
			config_file_name_exists: config_file_name /= Void
		do
			create file_name.make_from_string (config_file_name)
			create category_elements.make_default
			create appenders.make_default
			parse
		end

	make_with_hierarchy (config_file_name: STRING; h: ELOG_HIERARCHY) is
			-- Create a new config parser that will parse the file
			-- named 'config_file_name' and configure 'h'.
		require
			config_file_name_exists: config_file_name /= Void
			h_not_vfoid: h /= Void
		do
			hierarchy := h
			make (config_file_name)
		end

feature -- Parsing

	parse is
			-- Parse specified configuration file and create appropriate
			-- elogger topology. Make resulting heirarchy available in
			-- 'hierarchy'.
		local
			stream: KL_TEXT_INPUT_FILE
		do
			create parser.make
			create tree_pipe.make
			parser.set_callbacks (tree_pipe.start)

			debug ("elogger_config")
				internal_logger.log_information (create {ELOG_SIMPLE_EVENT}.make_without_time ("Parsing file: " + file_name.out + "...%N"))
			end
			create stream.make (file_name)
			stream.open_read
			if not stream.is_open_read then
				file_not_found := True
				internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Cannot open " + file_name.out + " for reading%N"))
			else
				parser.parse_from_stream (stream)
				stream.close
				if not tree_pipe.error.has_error then
					process_config_tree (tree_pipe.document)
				else
					parse_error := True
					display_parser_error
				end
			end
		end

	hierarchy: ELOG_HIERARCHY
			-- Configured logging hierarchy

	file_not_found: BOOLEAN
			-- Was the file 'file_name' found?

	parse_error: BOOLEAN
			-- Was there an error while parsing?

feature {NONE} -- Implementation

	parser: XM_EIFFEL_PARSER
			-- xml parser

 	tree_pipe: XM_TREE_CALLBACKS_PIPE
 			-- callback

	file_name: STRING
			-- Name of file to parse.

	display_parser_error is
			-- Output parsing error
		do
			internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: "+parser.last_error_extended_description))
		end

	appenders: DS_HASH_TABLE [ELOG_APPENDER, STRING]
			-- Collection of named appenders

	category_elements: DS_HASH_TABLE [XM_ELEMENT, STRING]
			-- Collection of category elements for post processing (after all
			-- appenders have been created so that appender references can
			-- be resolved.

	root_element: XM_ELEMENT
			-- Root element held for post processing all appender refs.

	process_config_tree (document: XM_DOCUMENT) is
			-- Process configuration DOM tree.
		require
			document_exists: document /= Void
		local
			child_node_cursor: DS_BILINEAR_CURSOR [XM_NODE]
			element: XM_ELEMENT
		do
			-- process the root element
			process_root (document.root_element)
			if hierarchy /= Void then
				-- iterate through top level elements and process according to type,
				-- ignoring root as it is already done
				if not document.root_element.is_empty then
					from
						child_node_cursor := document.root_element.new_cursor
						child_node_cursor.start
					until
						child_node_cursor.off
					loop
						element ?= child_node_cursor.item
						-- skip non-elements
						if element /= Void then
							process_element (element)
						end
						child_node_cursor.forth
					end
				else
					internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: No configuration elements found."))
				end
			end
			post_process_categories
			post_process_root
		end

	process_element (element: XM_ELEMENT) is
			-- Determine type of element and process
		require
			element_exists: element /= Void
		local
			name: STRING
		do
			name := element.name
			if name.is_equal (Appender_element_name) then
				process_appender (element)
			elseif name.is_equal (Category_element_name) then
				process_category (element)
			elseif name.is_equal (Root_element_name) then
				-- ignore, already processed
			else
				internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Unknown top level element type: " + name.out))
			end
		end

	process_appender (element: XM_ELEMENT) is
			-- Process an appender configuration element
		require
			element_exists: element /= Void
		local
			name, type, bool: STRING
			is_synchronous: BOOLEAN
		do
			if element.has_attribute_by_name (Name_attribute) then
				name := element.attribute_by_name (Name_attribute).value
				if element.has_attribute_by_name (Type_attribute) then
					type := element.attribute_by_name (Type_attribute).value

					is_synchronous := True
					if element.has_attribute_by_name (Synchronous_attribute) then
						-- the synchronous attribute is optional, appenders
						-- are synchronous by default.
						bool := element.attribute_by_name (Synchronous_attribute).value
						is_synchronous := bool.to_boolean
					end
					create_appender (name, type, element, is_synchronous)
				else
					internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Appender type attribute not found."))
				end

			else
				internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Appender name attribute not found."))
			end
		end

	create_appender (name, type: STRING; element: XM_ELEMENT; is_synchronous: BOOLEAN) is
			-- Create an appender of the specified 'type' and store
			-- in 'appenders' indexed by 'name'
		require
			name_exists: name /= Void
			type_exists: type /= Void
			element_exists: element /= Void
		local
			appender: ELOG_APPENDER
		do
			if type.is_equal (Stdout_appender_type) then
				create {ELOG_STDOUT_APPENDER} appender.make (name.out, is_synchronous)
			elseif type.is_equal (Stderr_appender_type) then
				create {ELOG_STDERR_APPENDER} appender.make (name.out, is_synchronous)
			elseif type.is_equal (File_appender_type) then
				appender := create_file_appender (name, element, is_synchronous)
			elseif type.is_equal (Rollingfile_appender_type) then
				appender := create_rolling_file_appender (name, element, is_synchronous)
			elseif type.is_equal (Calendarrolling_appender_type) then
				appender := create_calendar_rolling_appender (name, element, is_synchronous)
			end
			-- process sub-elements
			process_appender_filter (appender, element)
			process_appender_layout (appender, element)
			-- store appender
			if appender /= Void then
				appenders.force (appender, name.out)
			end
		end

	create_file_appender (name: STRING; element: XM_ELEMENT; synchronous: BOOLEAN): ELOG_APPENDER is
			-- Create a file appender using parameters in 'element'
		require
			name_exists: name /= Void
			element_exists: element /= Void
		local
			file, append: STRING
		do
			-- retrieve mandatory file name param
			file := retrieve_param_value (Filename_param_name, element)
			if file /= Void then
				-- retrieve optional append param
				append := retrieve_param_value (Append_param_name, element)
				if append /= Void and then append.to_boolean = True then
					create {ELOG_FILE_APPENDER[G]} Result.make (file, True, synchronous)
				else
					create {ELOG_FILE_APPENDER[G]} Result.make (file, False, synchronous)
				end
			else
				internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: File appender 'filename' attribute not found."))
			end
		end

	create_rolling_file_appender (name: STRING; element: XM_ELEMENT; synchronous: BOOLEAN): ELOG_APPENDER is
			-- Create a rolling file appender using parameters in 'element'
		require
			name_exists: name /= Void
			element_exists: element /= Void
		local
			file, append: STRING
			max_size, number_of_backups: INTEGER_REF
		do
			-- retrieve mandatory file name param
			file := retrieve_param_value (Filename_param_name, element)
			if file /= Void then
				-- retrieve mandatory max_size param
				max_size := retrieve_integer_param_value (Maxsize_param_name, element)
				if max_size /= Void then
					-- retrieve mandatory number_of_backups param
					number_of_backups := retrieve_integer_param_value (Numbackups_param_name, element)
					if number_of_backups /= Void then
						-- retrieve optional append param
						append := retrieve_param_value (Append_param_name, element)
						if append /= Void and then append.to_boolean = True then
							create {ELOG_ROLLING_FILE_APPENDER[G]} Result.make (file, max_size.item, number_of_backups.item, True, synchronous)
						else
							create {ELOG_ROLLING_FILE_APPENDER[G]} Result.make (file, max_size.item, number_of_backups.item, False, synchronous)
						end
					else
						 internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: File appender 'numbackups' attribute not found."))
					end
				else
					internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: File appender 'maxsize' attribute not found."))
				end
			else
				internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: File appender 'filename' attribute not found."))
			end
		end

	create_calendar_rolling_appender (name: STRING; element: XM_ELEMENT; synchronous: BOOLEAN): ELOG_APPENDER is
			-- Create a calendar rolling appender using parameters in 'element'
		require
			name_exists: name /= Void
			element_exists: element /= Void
		local
			file, append: STRING
			number_of_backups: INTEGER_REF
		do
			-- retrieve mandatory file name param
			file := retrieve_param_value (Filename_param_name, element)
			if file /= Void then
				-- retrieve mandatory number_of_backups param
				number_of_backups := retrieve_integer_param_value (Numbackups_param_name, element)
				if number_of_backups /= Void then
					-- retrieve optional append param
					append := retrieve_param_value (Append_param_name, element)
					if append /= Void and then append.to_boolean = True then
						create {ELOG_CALENDAR_ROLLING_APPENDER[G]} Result.make (file, number_of_backups.item, True, synchronous)
					else
						create {ELOG_CALENDAR_ROLLING_APPENDER[G]} Result.make (file, number_of_backups.item, False, synchronous)
					end
				else
					internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Calendar rolling appender 'numbackups' attribute not found."))
				end
			else
				internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Calendar rolling appender 'filename' attribute not found."))
			end
		end

	retrieve_param_value (name: STRING; parent: XM_ELEMENT): STRING is
			-- Locate child param element with specified name. Return value
			-- if found, Void otherwise.
		require
			element_exists: parent /= Void
		local
			child_node_cursor: DS_BILINEAR_CURSOR [XM_NODE]
			element: XM_ELEMENT
			param_name: STRING
		do
			-- iterate through child elements and search for named parameter element
			if not parent.is_empty then
				from
					child_node_cursor := parent.new_cursor
					child_node_cursor.start
				until
					child_node_cursor.off or Result /= Void
				loop
					element ?= child_node_cursor.item
					-- skip non-elements
					if element /= Void then
						if element.name.is_equal (Param_element_name) then
							-- check for a name attribute
							if element.has_attribute_by_name (Name_attribute) then
								param_name := element.attribute_by_name (Name_attribute).value
								if param_name.is_equal (name) then
									-- retrieve the value attribute
									if element.has_attribute_by_name (Value_attribute) then
										Result := element.attribute_by_name (Value_attribute).value.out
									else
										internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Parameter element value attribute not found."))
									end
								end
							else
								internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Parameter element name attribute not found."))
							end
						end
					end
					child_node_cursor.forth
				end
			end
		end

	process_appender_filter (appender: ELOG_APPENDER; parent: XM_ELEMENT) is
			-- Search 'parent' for any nested filter elements and add them to the 'appender'
		require
			appender_exists: appender /= Void
			parent_exists: parent /= Void
		local
			child_node_cursor: DS_BILINEAR_CURSOR [XM_NODE]
			element: XM_ELEMENT
			filter: ELOG_FILTER
			filters: DS_LINKED_LIST [ELOG_FILTER]
		do
			-- iterate through child elements and search for filter elements
			if not parent.is_empty then
				create filters.make
				from
					child_node_cursor := parent.new_cursor
					child_node_cursor.start
				until
					child_node_cursor.off
				loop
					element ?= child_node_cursor.item
					-- skip non-elements
					if element /= Void then
						filter := create_filter (element)
						if filter /= Void then
							filters.force_last (filter)
						end
					end
					child_node_cursor.forth
				end

				if filters.count = 1 then
					appender.set_filter (filters.first)
				elseif filters.count > 1 then
					appender.set_filter (filters.first)
					internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: only one filter may be specified."))
				end
			end
		end

	create_filter (element: XM_ELEMENT): ELOG_FILTER is
			-- create a filter from an XM_ELEMENT
		require
			element_not_void: element /= Void
		local
			type: STRING
		do
			if element.name.is_equal (Filter_element_name) then
				-- check for a type attribute
				if element.has_attribute_by_name (Type_attribute) then
					type := element.attribute_by_name (Type_attribute).value
					if type.is_equal (Severitymatch_filter_type) or type.is_equal (Prioritymatch_filter_type) then
						Result := create_severity_match_filter (element)
					elseif type.is_equal (Severityrange_filter_type) or type.is_equal (Priorityrange_filter_type) then
						Result := create_severity_range_filter (element)
					elseif type.is_equal (Stringmatch_filter_type) then
						Result := create_string_match_filter (element)
					elseif type.is_equal (Or_filter_type) then
						Result := create_or_filter (element)
					elseif type.is_equal (And_filter_type) then
						Result := create_and_filter (element)
					else
						internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Unknown filter type " + type.out))
					end
				else
					internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Filter element type attribute not found."))
				end
			end
		end

	create_severity_match_filter (element: XM_ELEMENT): ELOG_FILTER is
			-- Create a severity match filter.
		require
			element_exists: element /= Void
		local
			severity: STRING
			match: STRING
		do
			-- find mandatory severity parameter
			severity := retrieve_param_value (Severity_param_name, element)
			if severity /= Void then
				-- find mandatory match on filter parameter
				match := retrieve_param_value (Match_param_name, element)
				if match /= Void then
					if match.is_boolean then
						create {ELOG_SEVERITY_MATCH_FILTER} Result.make (create_severity (severity), match.to_boolean)
					else
						internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Severity match filter element 'match' attribute is not boolean."))
					end
				else
					internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Severity match filter element 'match' attribute not found."))
				end
			else
				internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Severity match filter element 'severity' attribute not found."))
			end
		end

	create_severity_range_filter (element: XM_ELEMENT): ELOG_FILTER is
			-- Create a severity range filter.
		require
			element_exists: element /= Void
		local
			severity_start, severity_end: STRING
			match: STRING
		do
			-- find mandatory severity startparameter
			severity_start := retrieve_param_value (Severitystart_param_name, element)
			if severity_start /= Void then
				-- find mandatory severity end parameter
				severity_end := retrieve_param_value (Severityend_param_name, element)
				if severity_end /= Void then
					-- find mandatory match on filter parameter
					match := retrieve_param_value (Match_param_name, element)
					if match /= Void then
						if match.is_boolean then
							create {ELOG_SEVERITY_RANGE_FILTER} Result.make (create_severity (severity_start),
								create_severity (severity_end), match.to_boolean)
						else
							internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Severity range filter element 'match' attribute is not boolean."))
						end
					else
						internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Severity range filter element 'match' attribute not found."))
					end
				else
					internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Severity range filter element 'severityend' not found."))
				end
			else
				internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Severity range filter element 'severitystart' attribute not found."))
			end
		end

	create_string_match_filter (element: XM_ELEMENT): ELOG_FILTER is
			-- Create a string match filter.
		require
			element_exists: element /= Void
		local
			string: STRING
			match: STRING
		do
			-- find mandatory string parameter
			string := retrieve_param_value (String_param_name, element)
			if string /= Void then
				-- find mandatory match on filter parameter
				match := retrieve_param_value (Match_param_name, element)
				if match /= Void then
					if match.is_boolean then
						create {ELOG_STRING_MATCH_FILTER} Result.make (string, match.to_boolean)
					else
						internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: String match filter element 'match' attribute is not boolean."))
					end
				else
					internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: String match filter element 'match' attribute not found."))
				end
			else
				internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: String match filter element 'string' attribute not found."))
			end
		end

	create_and_filter (element: XM_ELEMENT): ELOG_AND_FILTER is
			-- Create an and filter.
		require
			element_exists: element /= Void
		local
			child_node_cursor: DS_BILINEAR_CURSOR [XM_NODE]
			filters: DS_LINKED_LIST [ELOG_FILTER]
			reject: STRING
			reject_boolean: BOOLEAN
			filter: ELOG_FILTER
			elem: XM_ELEMENT
			fc: DS_LINKED_LIST_CURSOR [ELOG_FILTER]
		do
			-- iterate through child elements and search for filter elements
			if not element.is_empty then
				reject := retrieve_param_value (Rejects_anded_param_name, element)
				if reject /= Void and then reject.to_boolean = True then
					reject_boolean := True
				end

				create filters.make
				-- loop through children (filters)
				from
					child_node_cursor := element.new_cursor
					child_node_cursor.start
				until
					child_node_cursor.off
				loop
					elem ?= child_node_cursor.item
					if elem /= Void then
						filter := create_filter (elem)
						if filter /= Void then
							filters.force_last (filter)
						end
					end
					child_node_cursor.forth
				end

				if filters.count > 1 then
					fc := filters.new_cursor
					fc.start
					filter := fc.item
					fc.forth
					create Result.make (filter, fc.item, reject_boolean)
					from
						fc.forth
					until
						fc.off
					loop
						Result.add_filter (fc.item)
						fc.forth
					end
				else
					internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: And filters must have at least 2 sub-filters."))
				end
			end
		end

	create_or_filter (element: XM_ELEMENT): ELOG_OR_FILTER is
			-- Create an or filter.
		require
			element_exists: element /= Void
		local
			child_node_cursor: DS_BILINEAR_CURSOR [XM_NODE]
			filters: DS_LINKED_LIST [ELOG_FILTER]
			reject: STRING
			reject_boolean: BOOLEAN
			filter: ELOG_FILTER
			elem: XM_ELEMENT
			fc: DS_LINKED_LIST_CURSOR [ELOG_FILTER]
		do
			-- iterate through child elements and search for filter elements
			if not element.is_empty then
				reject := retrieve_param_value (Rejects_ored_param_name, element)
				if reject /= Void and then reject.to_boolean = True then
					reject_boolean := True
				end

				create filters.make
				-- loop through children (filters)
				from
					child_node_cursor := element.new_cursor
					child_node_cursor.start
				until
					child_node_cursor.off
				loop
					elem ?= child_node_cursor.item
					if elem /= Void then
						filter := create_filter (elem)
						if filter /= Void then
							filters.force_last (filter)
						end
					end
					child_node_cursor.forth
				end

				if filters.count > 1 then
					fc := filters.new_cursor
					fc.start
					filter := fc.item
					fc.forth
					create Result.make (filter, fc.item, reject_boolean)
					from
						fc.forth
					until
						fc.off
					loop
						Result.add_filter (fc.item)
						fc.forth
					end
				else
					internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Or filters must have at least 2 sub-filters."))
				end
			end
		end

	process_appender_layout (appender: ELOG_APPENDER; parent: XM_ELEMENT) is
			-- Search 'parent' for any nested layout element and set it on the 'appender'
		require
			appender_exists: appender /= Void
			parent_exists: parent /= Void
		local
			child_node_cursor: DS_BILINEAR_CURSOR [XM_NODE]
			element: XM_ELEMENT
			type: STRING
			layout: ELOG_LAYOUT
		do
			-- iterate through child elements and search for layout elements
			if not parent.is_empty then
				from
					child_node_cursor := parent.new_cursor
					child_node_cursor.start
				until
					child_node_cursor.off
				loop
					element ?= child_node_cursor.item
					-- skip non-elements
					if element /= Void then
						if element.name.is_equal (Layout_element_name) then
							-- check for a type attribute
							if element.has_attribute_by_name (Type_attribute) then
								type := element.attribute_by_name (Type_attribute).value
								if type.is_equal (Datetime_layout_type) then
									create {ELOG_DATE_TIME_LAYOUT} layout
								elseif type.is_equal (Simple_layout_type) then
									create {ELOG_SIMPLE_LAYOUT} layout
								elseif type.is_equal (Time_layout_type) then
									create {ELOG_TIME_LAYOUT} layout.make
								elseif type.is_equal (Pattern_layout_type) then
									layout := create_pattern_layout (element)
								else
									internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Unknown layout type " + type.out))
								end
								if layout /= Void then
									appender.set_layout (layout)
								end
							else
								internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Layout element type attribute not found."))
							end
						end
					end
					child_node_cursor.forth
				end
			end
		end

	create_pattern_layout (element: XM_ELEMENT): ELOG_LAYOUT is
			-- Create a pattern layout.
		require
			element_exists: element /= Void
		local
			pattern: STRING
		do
			-- find mandatory pattern parameter
			pattern := retrieve_param_value (Pattern_param_name, element)
			if pattern /= Void then
				create {ELOG_PATTERN_LAYOUT} Result.make (pattern)
			else
				internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Pattern layout element 'pattern' attribute not found."))
			end
		end

	retrieve_integer_param_value (name: STRING; parent: XM_ELEMENT): INTEGER_REF is
			-- Locate an integer child param element with specified name. Return value
			-- if found, Void otherwise.
		require
			element_exists: parent /= Void
		local
			value: STRING
		do
			value := retrieve_param_value (name, parent)
			if value /= Void then
				if value.is_integer then
					create Result
					Result.set_item (value.to_integer)
				else
					internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Invalid value found for parameter " + name.out))
				end
			end
		end

	process_category (element: XM_ELEMENT) is
			-- Process a category element
		require
			element_exists: element /= Void
		local
			name: STRING
		do
			if element.has_attribute_by_name (Name_attribute) then
				name := element.attribute_by_name (Name_attribute).value
				category_elements.force (element, name.out)
			else
				internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Category name attribute not found."))
			end
		end

	post_process_categories is
			-- Process all category elements to resolve appender references
		local
			cursor: DS_HASH_TABLE_CURSOR [XM_ELEMENT, STRING]
		do
			from
				cursor := category_elements.new_cursor
				cursor.start
			until
				cursor.off
			loop
				post_process_category (cursor.key, cursor.item)
				cursor.forth
			end
		end

	post_process_category (name: STRING; element: XM_ELEMENT) is
			-- Process a category element and create appropriate category object
		local
			category: ELOG_LOGGER
			severity: ELOG_SEVERITY
			child_node_cursor: DS_BILINEAR_CURSOR [XM_NODE]
			child: XM_ELEMENT
			additive: STRING
		do
			category := hierarchy.logger (name)
			-- search for optional severity attribute
			if element.has_attribute_by_name (Severity_attribute) then
				severity := create_severity (element.attribute_by_name (Severity_attribute).value.out)
				if severity /= Void then
					category.set_severity (severity)
				end
			end
			-- search for optional additivity attribute
			if element.has_attribute_by_name (Additive_attribute) then
				additive := element.attribute_by_name (Additive_attribute).value.out
				category.set_additive (additive.is_boolean and then additive.to_boolean)
			end
			-- search for optional appender refs
			if not element.is_empty then
				from
					child_node_cursor := element.new_cursor
					child_node_cursor.start
				until
					child_node_cursor.off
				loop
					child ?= child_node_cursor.item
					-- skip non-elements
					if child /= Void then
						if child.name.is_equal (Appenderref_element_name) then
							-- search for mandatory ref attribute
							if child.has_attribute_by_name (Ref_attribute) then
								category.add_appender (appenders.item (child.attribute_by_name (Ref_attribute).value.out))
							else
								internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Appender reference 'ref' attribute not found."))
							end
						else
							internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Invalid child node within category: " + name))
						end
					end
					child_node_cursor.forth
				end
			end
		end

	create_severity (severity_name: STRING): ELOG_SEVERITY is
			-- Create a severity object representing the named severity.
			-- Return Void if not a valid severity
		require
			severity_name_exists: severity_name /= Void
		do
			Result := severity_from_string(severity_name.as_upper)
		end

	process_root (parent: XM_ELEMENT) is
			-- Process a root element
		require
			element_exists: parent /= Void
		local
			child_node_cursor: DS_BILINEAR_CURSOR [XM_NODE]
			element: XM_ELEMENT
			done: BOOLEAN
			severity: ELOG_SEVERITY
		do
			if not parent.is_empty then
				from
					child_node_cursor := parent.new_cursor
					child_node_cursor.start
				until
					child_node_cursor.off or done
				loop
					element ?= child_node_cursor.item
					-- skip non-elements
					if element /= Void then
						if element.name.is_equal (Root_element_name) then
							-- this is the root, get the severity
							if element.has_attribute_by_name (Severity_attribute) then
								severity := create_severity (element.attribute_by_name (Severity_attribute).value.out)
								if severity /= Void then
									-- if the hierarchy is Void then create a new one, otherwise just set the
									-- severity
									if hierarchy = Void then
										create hierarchy.make (severity)
									else
										hierarchy.root.set_severity (severity)
									end
									-- store the root element for post processing
									root_element := element
								else
									internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Invalid root category severity."))
								end
							else
								internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Root category severity attribute not found."))
							end
							done := True
						end
					end
					child_node_cursor.forth
				end
			else
				internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Root element not found."))
			end
		end

	post_process_root is
			-- Process a category element and create appropriate category object
		require
			root_element_exists: root_element /= Void
		local
			child_node_cursor: DS_BILINEAR_CURSOR [XM_NODE]
			child: XM_ELEMENT
		do
			-- search for optional appender refs
			if not root_element.is_empty then
				from
					child_node_cursor := root_element.new_cursor
					child_node_cursor.start
				until
					child_node_cursor.off
				loop
					child ?= child_node_cursor.item
					-- skip non-elements
					if child /= Void then
						if child.name.is_equal (Appenderref_element_name) then
							-- search for mandatory ref attribute
							if child.has_attribute_by_name (Ref_attribute) then
								hierarchy.root.add_appender (appenders.item (child.attribute_by_name (Ref_attribute).value.out))
							else
								internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Appender reference 'ref' attribute not found for root."))
							end
						else
							internal_logger.log_warning (create {ELOG_SIMPLE_EVENT}.make_without_time ("Error: Invalid child node within root."))
						end
					end
					child_node_cursor.forth
				end
			end
		end

end
