note

	description:

		"ancestor for all entities"

	keywords: "entity,persistent"
	author: "Mark Howard"
	last_editor: "$Author: c4 $"
	date: "$Date: 2014/08/06 $"
	revision: "$Revision: #101 $"
	path: "$File: //rose/source/kernel/entities/entity.e $"

class ENTITY

inherit

	REGISTERED
		rename
			make as registered_make
		end

	ROSE_RELATIONSHIP_TYPE_ID_USE
		undefine
			is_equal
		end

	ROSE_CATEGORY_ID_USE
		undefine
			is_equal
		end

	ROSE_ATTRIBUTE_TYPE_ID_USE
		undefine
			is_equal
		end

	DEBUG_OUTPUT
		undefine
			is_equal
		end

	CONTEXT_AWARE

	MISMATCH_CORRECTOR
		undefine
			is_equal
		redefine
			correct_mismatch
		end

	HAS_NAME
		undefine
			is_equal
		end

	HASH_KEYED [INTEGER_64]
		rename
			key as uid
		undefine
			is_equal
		end

	NEEDS_EXPRESSION_FROM_STRING

create

	make,
	make_anonymous

feature {NONE} -- Creation

	make_anonymous (a_context: like context; a_category_id: INTEGER) is
			-- Create with `a_context' and `a_category_id'
		require
			a_context_not_void: a_context /= Void
			a_category_id_is_unsigned_16_bit: a_category_id >= 0 and a_category_id <= {INTEGER_16}.max_value
		do
			context_index := a_context.context_index
			category_id := a_category_id
		ensure
			context_index_set: context_index = a_context.context_index
			category_id_set: category_id = a_category_id
		end

feature {CATEGORY_HANDLER, RELATIONSHIP_HANDLER} -- Creation

	make (a_context: like context; a_category_id: INTEGER; a_handle: INTEGER; a_start_date: ROSE_DATE; a_end_date: ROSE_DATE) is
			-- Create in 'a_context' and 'a_category' with 'a_handle' as identifier
			-- ['a_start_date', 'a_end_date'] is the valid-time interval
		require
			context_exists: a_context /= Void
			category_id_exists: a_category_id > 0
			a_category_id_is_unsigned_16_bit: a_category_id <= {INTEGER_16}.max_value
			non_negative_handle: a_handle >= 0
		do
			category_id := a_category_id
			context_index := a_context.context_index
			registered_make (a_handle)
			set_start_date (a_start_date)
			set_end_date (a_end_date)
			set_as_of_date (today)
		ensure
			category_set: category_id /= 0
		end

feature {NONE} -- Basic operations

	make_context is
			-- Create `context'.
		do
				-- You most likely have threading issues if this is called
				-- or you have retrieved an entity via a messaging system from
				-- another executable. Currently, this feature is only implemented in certain descendents.
		end

feature -- Basic operations

	context: like context_anchor is
			-- Context corresponding to `supplier_id'
		do
			Result := context_from_index (context_index)
			if Result = Void then
				make_context
				Result := context_from_index (context_index)
			end
		ensure
			context_not_void: Result /= Void
		end

	correct_mismatch is
			-- <Precursor>
		local
			l_supplier_id: INTEGER_REF
		do
			l_supplier_id ?= mismatch_information.item ("supplier_id")
			if l_supplier_id /= Void then
				context_index := contexts.descriptor_from_supplier_id (l_supplier_id.item).index
			end
		end

feature -- Access

	category_id: INTEGER
			-- Category (type id in the context)

	descriptive_name: STRING is
			-- A terse descriptive representation of `Current'.
		do
			Result := name_at (today)
			if Result = Void then
				Result := handle.out
			end
		ensure
			result_not_void: Result /= Void
		end

	start_date: ROSE_DATE
			-- Valid start-time

	end_date: ROSE_DATE
			-- Valid end-time

	as_of_date: ROSE_DATE
			-- Transaction (or belief) time

	category_name: STRING is
			-- Category (type) name
		do
			Result := handler.name
		end

	handler: CATEGORY_HANDLER [like context, like Current] is
			-- accessor
		do
			Result ?= context.handler (category_id)
		end

feature -- Output

	id_string: STRING is
			-- String representation
		do
			Result := handler.name + " (" + context.name + " : " + handle.out + " )"
		end

	hash_string: STRING is
			-- Hash string representation
		do
			Result := supplier_id.out + ":" + category_id.out + ":" + handle.out
		end

feature -- Change

	set_as_of_date (a_date: ROSE_DATE) is
			-- Set 'as_of_date' to 'a_date
		do
			as_of_date := a_date
		end

	set_start_date (a_date: ROSE_DATE) is
			-- Set 'start_date' to 'a_date'
		do
			start_date := a_date
		ensure
			start_date_set: start_date = a_date
		end

	set_end_date (a_date: ROSE_DATE) is
			-- Set 'end_date' to 'a_date'
		do
			end_date := a_date
		ensure
			end_date_set: end_date = a_date
		end

feature -- Status Report

	is_rose: BOOLEAN
			--Is the handle a rose context handle?
		do
			Result := context.has_rose_handles
		end

feature -- Basic Operations

	entity_at (a_handle: INTEGER): like Current
			--Entity with this category_id and handle `a_handle'
		do
			Result := handler.entity (a_handle)
		end

	attribute_at (a_rose_id: INTEGER; a_date: ROSE_DATE): STRING is
		local
			l_handler: ENTITY_IDENTIFIER
		do
			l_handler := context.identifier_from_rose (a_rose_id)
			if l_handler /= Void then
				Result := l_handler.dated_key (Current, a_date)
			end
		end

	attribute_at_with_local (a_local_id: INTEGER; a_date: ROSE_DATE): STRING is
		local
			l_handler: ENTITY_IDENTIFIER
		do
			l_handler := context.identifier (a_local_id)
			if l_handler /= Void then
				Result := l_handler.dated_key (Current, a_date)
			end
		end

	most_recent_attribute_at (a_rose_id: INTEGER; a_date: ROSE_DATE): STRING is
		local
			l_handler: ENTITY_IDENTIFIER
		do
			l_handler := context.identifier_from_rose (a_rose_id)
			if l_handler /= Void then
				Result := l_handler.most_recent_key (Current, a_date)
			end
		end

	most_recent_attribute_at_with_local (a_local_id: INTEGER; a_date: ROSE_DATE): STRING is
		local
			l_handler: ENTITY_IDENTIFIER
		do
			l_handler := context.identifier (a_local_id)
			if l_handler /= Void then
				Result := l_handler.most_recent_key (Current, a_date)
			end
		end

	attribute_by_name (a_name: STRING; a_date: ROSE_DATE): STRING is
		local
			l_handler: ENTITY_IDENTIFIER
		do
			a_name.to_upper
			l_handler := context.identifier_from_name (a_name)
			if l_handler /= Void then
				Result := l_handler.dated_key (Current, a_date)
			end
		end

	most_recent_attribute (a_name: STRING; a_date: ROSE_DATE): STRING is
		local
			l_handler: ENTITY_IDENTIFIER
		do
			a_name.to_upper
			l_handler := context.identifier_from_name (a_name)
			if l_handler /= Void then
				Result ?= l_handler.most_recent_key (Current, a_date)
			end
		end

	attribute_history (a_name: STRING): ROSE_LINEAR_ARRAY [ENTITY_ATTRIBUTE_RECORD] is
		local
			l_handler: ENTITY_IDENTIFIER
		do
			a_name.to_upper
			l_handler := context.identifier_from_name (a_name)
			if l_handler /= Void then
				Result ?= l_handler.key_history (Current)
			end
		end

	related_entities (a_name: STRING): ROSE_ENTITY_HISTORY [ENTITY] is
		local
			l_handler: RELATIONSHIP_HANDLER [ENTITY, ENTITY]
		do
			a_name.to_upper
			l_handler := context.relationship_handler_from_name (a_name)
			if l_handler /= Void then
				Result := l_handler.related_entity_intervals (Current)
			end
		end

	corelated_entities (a_name: STRING): ROSE_ENTITY_HISTORY [ENTITY] is
		local
			l_handler: RELATIONSHIP_HANDLER [ENTITY, ENTITY]
		do
			a_name.to_upper
			l_handler := context.relationship_handler_from_name (a_name)
			if l_handler /= Void then
				Result := l_handler.corelated_entity_intervals (Current)
			end
		end

	mapped_supplier_attributes (a_other_context: SUPPLIER_CONTEXT; a_attribute_name: STRING; a_date: ROSE_DATE): ROSE_LINEAR_ARRAY [STRING] is
			-- All attributes of type `a_attribute_name' on `a_other_context' peers of `Current' as of date `a_date'.
		require
			a_other_context_not_void: a_other_context /= Void
			a_attribute_name_not_void: a_attribute_name /= Void
			a_date_not_void: a_date /= Void
		local
			i: INTEGER
			l_mapped_supplier_attributes_with_handles: like mapped_supplier_attributes_with_handles
		do
			create Result.make_empty
			l_mapped_supplier_attributes_with_handles := mapped_supplier_attributes_with_handles (a_other_context, a_attribute_name, a_date)
			from
				i := l_mapped_supplier_attributes_with_handles.lower
			until
				i > l_mapped_supplier_attributes_with_handles.upper
			loop
				Result.extend (l_mapped_supplier_attributes_with_handles.item (i).attribute_value)
				i := i + 1
			end
		ensure
			result_not_void: result /= Void
		end

	mapped_supplier_attributes_with_handles (a_other_context: SUPPLIER_CONTEXT; a_attribute_name: STRING; a_date: ROSE_DATE): ROSE_LINEAR_ARRAY [TUPLE [attribute_value: STRING; entity_handle: INTEGER]] is
			-- All attributes of type `a_attribute_name' and associated entity id on `a_other_context' peers of `Current' as of date `a_date'.
		require
			a_other_context_not_void: a_other_context /= Void
			a_attribute_name_not_void: a_attribute_name /= Void
			a_date_not_void: a_date /= Void
		local
			l_peers: like peers
			i: INTEGER
		do
			create Result.make_empty
			l_peers := peers (a_other_context, a_date)
			from
				i := l_peers.lower
			until
				i > l_peers.upper
			loop
				Result.extend ([l_peers.item (i).attribute_by_name (a_attribute_name, a_date), l_peers.item (i).handle])
				i := i + 1
			end
		ensure
			result_not_void: result /= Void
		end

	related_intervals_on (a_name: STRING; a_date: ROSE_DATE): ROSE_ENTITY_HISTORY [ENTITY] is
		do
			Result := related_intervals_during (a_name, a_date, a_date)
		end

	corelated_intervals_on (a_name: STRING; a_date: ROSE_DATE): ROSE_ENTITY_HISTORY [ENTITY] is
		do
			Result := corelated_intervals_during (a_name, a_date, a_date)
		end

	related_entities_on (a_name: STRING; a_date: ROSE_DATE): ROSE_LINEAR_ARRAY [ENTITY] is
		do
			Result := related_entities_during (a_name, a_date, a_date)
		end

	corelated_entities_on (a_name: STRING; a_date: ROSE_DATE): ROSE_LINEAR_ARRAY [ENTITY] is
		do
			Result := corelated_entities_during (a_name, a_date, a_date)
		end

	related_intervals_during (a_name: STRING; a_start, a_end: ROSE_DATE): ROSE_ENTITY_HISTORY [ENTITY] is
		do
			Result := related_entities (a_name).intervals_during (a_start, a_end, Void)
		end

	corelated_intervals_during (a_name: STRING; a_start, a_end: ROSE_DATE): ROSE_ENTITY_HISTORY [ENTITY] is
		do
			Result := corelated_entities (a_name).intervals_during (a_start, a_end, Void)
		end

	related_entities_during (a_name: STRING; a_start, a_end: ROSE_DATE): ROSE_LINEAR_ARRAY [ENTITY] is
		do
			Result := related_entities (a_name).witnesses_during (a_start, a_end)
		end

	corelated_entities_during (a_name: STRING; a_start, a_end: ROSE_DATE): ROSE_LINEAR_ARRAY [ENTITY] is
		do
			Result := corelated_entities (a_name).witnesses_during (a_start, a_end)
		end

feature -- Peer relationship

	peer (a_other_context: SUPPLIER_CONTEXT; a_date: ROSE_DATE): like peer_anchor is
			-- Peer entity in 'a_other_context' at 'a_date'
			-- Result represents same entity (same category) in 'a_other_context'.
			-- Void if no peer,
			-- Result.handle = 0 if multiple peers found
		do
			Result ?= a_other_context.handler (category_id).peer (Current, a_date)
		end

	most_recent_peer (a_other_context: SUPPLIER_CONTEXT; a_date: ROSE_DATE): like peer_anchor is
			-- Peer entity in 'a_other_context' at 'a_date'
			-- Result represents same entity (same category) in 'a_other_context'.
			-- Void if no peer,
			-- Result.handle = 0 if multiple peers found
		do
			Result ?= a_other_context.handler (category_id).most_recent_peer (Current, a_date)
		end

	peers_history (a_other_context: SUPPLIER_CONTEXT): ROSE_ENTITY_HISTORY [like peer_anchor] is
			-- Peer entities history in 'a_other_context'
			-- Result is the entities history for Current (same category) in 'a_other_context'.
			-- Void if no peers
		local
			p: like peer_anchor
			l_interval: ENTITY_INTERVAL [like peer_anchor]
			a: ROSE_ENTITY_HISTORY [ENTITY]
			i: INTEGER
		do
				--| FIXME: The loop below is used because reverse assignment on a ROSE_LINEAR_ARRAY fails
			a := a_other_context.handler (category_id).peers_history (Current)
			if a /= Void then
				from
					create Result.make_empty
					i := a.lower
				until
					i > a.upper
				loop
					p ?= a.item (i).entity
					if p /= Void then
						create l_interval.make (p, a.item (i).start_date, a.item (i).end_date, 1)
						Result.extend (l_interval)
					end
					i := i + 1
				end
			else
				create Result.make_empty
			end
		end

	peers_history_uncached (a_other_context: SUPPLIER_CONTEXT): ROSE_ENTITY_HISTORY [like peer_anchor] is
			-- Peer entities history in 'a_other_context', without building a cache of all mappings
			-- Result is the entities history for Current (same category) in 'a_other_context'.
			-- Void if no peers
		local
			p: like peer_anchor
			l_interval: ENTITY_INTERVAL [like peer_anchor]
			a: ROSE_ENTITY_HISTORY [ENTITY]
			i: INTEGER
		do
				--| FIXME: The loop below is used because reverse assignment on a ROSE_LINEAR_ARRAY fails
			a := a_other_context.handler (category_id).peers_history_uncached (Current)
			if a /= Void then
				from
					create Result.make_empty
					i := a.lower
				until
					i > a.upper
				loop
					p ?= a.item (i).entity
					if p /= Void then
						create l_interval.make (p, a.item (i).start_date, a.item (i).end_date, 1)
						Result.extend (l_interval)
					end
					i := i + 1
				end
			end
		end

	peers (a_other_context: SUPPLIER_CONTEXT; a_date: ROSE_DATE): ROSE_LINEAR_ARRAY [like peer_anchor] is
			-- Peer entities in 'a_other_context' at 'a_date'
			-- Result represents same entities (same category) in 'a_other_context'.
			-- Void if no peers
		local
			p: like peer_anchor
			a: ROSE_LINEAR_ARRAY [ENTITY]
			i: INTEGER
		do
				--| FIXME: The loop below is used because reverse assignment on a ROSE_LINEAR_ARRAY fails
			a := a_other_context.handler (category_id).peers (Current, a_date)
			if a /= Void then
				from
					create Result.make (a.lower, a.upper)
					i := a.lower
				until
					i > a.upper
				loop
					p ?= a.item (i)
					Result.put (p, i)
					i := i + 1
				end
			end
		end

feature -- Basic operations

	relationships: ROSE_LINEAR_ARRAY [RELATIONSHIP_HANDLER [like Current, ENTITY]] is
		do
			Result := handler.active_relationships
		end

	corelationships: ROSE_LINEAR_ARRAY [RELATIONSHIP_HANDLER [ENTITY, like Current]] is
		do
			Result := handler.active_corelationships
		end

	name_at (a_date: ROSE_DATE): STRING is
		local
			l_identifier: ENTITY_IDENTIFIER
		do
			l_identifier := context.identifier_from_name (name_identifier_key)
			if l_identifier /= Void then
				Result := l_identifier.dated_key (Current, a_date)
			end
		end

	name_after (a_date: ROSE_DATE): STRING is
		local
			l_identifier: ENTITY_IDENTIFIER
		do
			l_identifier := context.identifier_from_name (name_identifier_key)
			if l_identifier /= Void then
				Result := l_identifier.key_after (Current, a_date)
			end
		end

	first_data_date: ROSE_DATE is
		local
			l_targets: like ur_targets
			l_target: ROSE_SIGNATURE_RECORD [ENTITY]
			l_date: ROSE_DATE
		do
			create Result.make_max_date
			l_targets := ur_targets
			from
				l_targets.start
			until
				l_targets.off
			loop
				l_target ?= l_targets.current_item
				if l_target /= Void then
					l_target.set_entity_id (handle)
					l_date := l_target.first_data_date
					if not l_date.is_undetermined then
						Result := Result.min (l_date)
					end
				end
				l_targets.forth
			end
		end

	last_data_date: ROSE_DATE is
		local
			l_targets: like ur_targets
			l_target: ROSE_SIGNATURE_RECORD [ENTITY]
			l_date: ROSE_DATE
		do
			create Result.make_min_date
			l_targets := ur_targets
			from
				l_targets.start
			until
				l_targets.off
			loop
				l_target ?= l_targets.current_item
				if l_target /= Void then
					l_target.set_entity_id (handle)
					l_date := l_target.last_data_date
					if not l_date.is_undetermined then
						Result := Result.max (l_date)
					end
				end
				l_targets.forth
			end
		end

	first_process_date: ROSE_DATE is
		local
			l_targets: like ur_targets
			l_target: ROSE_SIGNATURE_RECORD [ENTITY]
			l_date: ROSE_DATE
		do
			create Result.make_max_date
			l_targets := ur_targets
			from
				l_targets.start
			until
				l_targets.off
			loop
				l_target ?= l_targets.current_item
				if l_target /= Void then
					l_target.set_entity (Current)
					l_date := l_target.first_process_date
					if not l_date.is_undetermined then
						Result := Result.min (l_date)
					end
				end
				l_targets.forth
			end
		end

	last_process_date: ROSE_DATE is
		local
			l_targets: like ur_targets
			l_target: ROSE_SIGNATURE_RECORD [ENTITY]
			l_date: ROSE_DATE
		do
			create Result.make_min_date
			l_targets := ur_targets
			from
				l_targets.start
			until
				l_targets.off
			loop
				l_target ?= l_targets.current_item
				if l_target /= Void then
					l_target.set_entity (Current)
					l_date := l_target.last_process_date
					if not l_date.is_undetermined then
						Result := Result.max (l_date)
					end
				end
				l_targets.forth
			end
		end

	internal: INTERNAL is
			-- Once acces to an object of type INTERNAL
		once
			create Result
		ensure
			result_not_void: Result /= Void
		end

	entity_date_target_at (a_date: ROSE_DATE): ROSE_ENTITY_DATE [like Current] is
		do
			create Result.make (Current, a_date)
		end

	expression_tag: STRING is
			-- Empty string
		do
			Result := ""
		end

	ur_targets: ROSE_LINEAR_ARRAY [ROSE_TARGET [ENTITY]] is
			--list of target types for this entity
		do
			create Result.make_empty
		end

	in_investment_universe: BOOLEAN is
		do
			Result := attribute_by_name ("BRIM CODE", as_of_date) /= Void
		end

	debug_output: STRING is
		local
			l_name: STRING
		do
			Result := supplier_id.out + ":" + category_id.out + ":" + handle.out
			if debug_out_verbose_flag.item then
					-- This code is dangerous because a tremendous amount of code can be executed
					-- simply by examining a variable in the debugger.  To avoid these side effects,
					-- this code should be avoided.
				l_name := descriptive_name
				if l_name /= Void then
					Result := Result + "(" + l_name + ")"
				end
			end
		ensure then
			result_not_void: Result /= Void
		end

	debug_verbose: BOOLEAN is
			-- USE AT YOUR OWN RISK!
			-- Call this from the debugger on any ENTITY and 'debug_output' will be verbose.
			-- Note that this is a function only because the debugger doesn't allow you to call subroutines.
		do
			debug_out_verbose_flag.put (True)
		end

	debug_brief: BOOLEAN is
			-- Call this from the debugger on any ENTITY and 'debug_output' will be brief.
			-- Note that this is a function only because the debugger doesn't allow you to call subroutines.
		do
			debug_out_verbose_flag.put (False)
		end

feature -- Descriptive name attributes

	name_identifier_key: STRING is
			-- Attribute that is used for the descriptive name
		do
			Result := "name"
		ensure
			result_not_void: Result /= Void
		end

	name_identifier_keys: LIST [STRING] is
			-- Atrributes that are needed to compute the descriptive name
			-- Has to be redefined in descendants that use more than `name_identifier_key'
		once
			create {LINKED_LIST [STRING]} Result.make
			Result.compare_objects
			Result.extend (name_identifier_key)
		ensure
			result_not_void: Result /= Void
			no_void_values: not Result.has (Void)
			compare_objects: Result.object_comparison
		end

feature -- IDs

	context_index: INTEGER

	supplier_id: INTEGER is
			-- Supplier ID
		do
			Result := context.supplier_id
		end

	uid: INTEGER_64 is
			-- The unique identifier.  Supplier/category/handle.
		do
			Result := supplier_id.to_integer_64 |<< 48 + category_id.to_integer_64 |<< 32 + handle
		end

feature {NEEDS_EXPRESSION_FROM_STRING} --expressions by name

	expression (a_string: STRING): ROSE_EXPRESSION [ROSE_TARGET [ENTITY]] is
		local
			l_simple: SIMPLE_EXPRESSION [ROSE_EXPRESSION_TARGET]
			l_ur_targets: like ur_targets
		do
			l_ur_targets := ur_targets
			from
				l_ur_targets.start
			until
				l_ur_targets.off or Result /= Void
			loop
				Result := l_ur_targets.current_item.expressions.complex_expression (a_string)
				l_simple ?= Result
				if l_simple /= Void then
					Result := Void
				end
				l_ur_targets.forth
			end
		end

	expressions_from_names (a_names: LINEAR [STRING]): ROSE_LINEAR_ARRAY [ROSE_EXPRESSION [ROSE_TARGET [ENTITY]]] is
			-- Get ROSE_EXPRESSION objects from a list of strings
		do
			create Result.make_empty
			from
				a_names.start
			until
				a_names.off
			loop
				Result.extend (expression (a_names.item))
				a_names.forth
			end
		end

feature {NONE} -- Implementation

	debug_out_verbose_flag: CELL [BOOLEAN] is
			-- Setting this very dangerous flag will cause the debugger output to be verbose.
		once
			create Result.put (False)
			Result.put (False)
		end

	peer_anchor : ENTITY do end

invariant

	category_id_is_unsigned_16_bit: category_id >= 0 and category_id <= {INTEGER_16}.max_value
	supplier_id_is_unsigned_16_bit: supplier_id >= 0 and supplier_id <= {INTEGER_16}.max_value

end
