note

	description:

		"Identifiers used to access entity attributes values"

	author: "Mark Howard"
	last_editor: "$Author: rj $"
	date: "$Date: 2013/07/18 $"
	revision: "$Revision: #69 $"
	path: "$File: //rose/source/kernel/entities/entity_identifier.e $"

class
	ENTITY_IDENTIFIER

inherit
	ROSE_ID_USER
		export
			{NONE} all
		end

	NEEDS_HANDLE

	SUPPLIER_ACCESS

	ENTITY_TEMPLATES
		export
			{NONE} all
		end

	DEBUG_OUTPUT

create
	make

feature {NONE} -- Creation

	make (a_context: like context; a_handle: INTEGER) is
			-- Initialize values and create containers
		require
			valid_context: a_context /= Void
		do
			context := a_context
			handle := a_handle
			name := context.attribute_name_by_id (handle)
			is_unique_identifier := context.attribute_is_uniq_identifier (handle)

			if name = Void then
				name := handle.out
			end

			-- Allocate a token 1 element, 'internal_key_history' will allocate a more appropriate
			-- amount the first time these expand.  This was done to prevent the allocation
			-- of 100's of MB of space that was sometimes never being used - so don't break that
			-- part.  The bit about resizing on first allocation may or may not be a good
			-- idea.  RJ.
			create keys.make(1)
			create dated_keys.make(1)
			create dated_keys_by_id_cache.make(1)
			create entities.make_empty
			create all_entities_table.make(10)
			create category_loaded.make(1, context.max_category_id)
			create unique_key_cache_by_category.make (100)
			if context.pes_server then
				create non_unique_keys_by_category.make (100)
			end
		end

feature  --Access

	name: STRING
			-- Name of identifier type

	handle: INTEGER
			-- Internal identifier

	is_unique_identifier: BOOLEAN
			-- Is this identifier a unique attribute for an entity?

	debug_output: STRING is
			-- Text to display in debugger
		do
			Result := name
		end

	all_entities (a_key: STRING; a_date: ROSE_DATE): ROSE_LINEAR_ARRAY [ENTITY] is
			-- All entities which had 'a_key' as this attribute on 'a_date'.
		local
			l_records: ROSE_LINEAR_ARRAY[ENTITY_ATTRIBUTE_RECORD]
			l_record: ENTITY_ATTRIBUTE_RECORD
			i: INTEGER
		do
			create Result.make_empty
			l_records := all_entity_records(a_key)
			from i := 1 until i > l_records.count loop
				l_record := l_records.item(i)
				if l_record.brackets(a_date) then
					Result.extend(entity_from_handle (l_record.entity_id, l_record.category_id))
				end
				i := i + 1
			end
		end

	entity_from_handle (a_handle: INTEGER;a_category: INTEGER): like entity_template is
			-- Utility routine to turn `a_handle' into an entity.
		do
			Result := context.handler(a_category).entity(a_handle)
		end

	dated_key (a_entity: like entity_template; a_date: ROSE_DATE): STRING is
			-- The value of this attribute for `a_entity' on `a_date'.
		local
			l_dated_attribute: ENTITY_ATTRIBUTE_RECORD
		do
			l_dated_attribute := dated_attribute (a_entity, a_date)
			if l_dated_attribute /= Void then
				Result := l_dated_attribute.attribute_value
			end
		end

	dated_attribute (a_entity: like entity_template; a_date: ROSE_DATE): ENTITY_ATTRIBUTE_RECORD is
			-- The attribute record for entity `a_entity' on `a_date'.
		local
			l_date: ROSE_DATE
			l_records: ROSE_LINEAR_ARRAY[ENTITY_ATTRIBUTE_RECORD]
		do
			l_date := valid_date(a_date)
			l_records := key_history (a_entity)
			from l_records.start until l_records.off or Result /= Void loop
				if l_records.current_item.brackets(l_date) then
					Result := l_records.current_item
				end
				l_records.forth
			end
		end

	key_after (a_entity: like entity_template; a_date: ROSE_DATE): STRING is
			-- The value of this attribute for 'a_entity' on 'a_date'.  If none exist on that date
			-- then return the value for the attribute that is the closest to 'a_date' and above.  (I think).
		local
			l_date,l_best_date: ROSE_DATE
			l_record: ENTITY_ATTRIBUTE_RECORD
			l_records: ROSE_LINEAR_ARRAY[ENTITY_ATTRIBUTE_RECORD]
			l_guess: STRING
			l_item: ENTITY_ATTRIBUTE_RECORD
		do
			l_date := valid_date(a_date)
			keys.search(a_entity)
			if keys.found then
				l_record := keys.found_item
				if l_record /= Void then
					if l_record.end_date_greater_than(l_date) then
						Result := l_record.attribute_value
					end
				end
			else
				l_records := internal_key_history(a_entity)
				if l_records /= Void then
					from l_records.start until l_records.off or Result /= Void loop
						l_item := l_records.current_item
						if l_item.end_date_greater_than(l_date) then
							if l_item.brackets(l_date) then
								Result := l_item.attribute_value
							else
								-- The found record doesn't bracket 'l_date', so it must start AFTER 'l_date'.
								if l_best_date = Void or else l_item.start_date.date < l_best_date then
									-- l_best_date will wind up being the closest to 'a_date', but above.
									l_best_date := l_item.start_date.date
									l_guess := l_item.attribute_value
								end
							end
						end
						l_records.forth
					end
					if Result = Void then Result := l_guess end
				end
			end
		end

	most_recent_attribute(a_entity: like entity_template;a_date: ROSE_DATE): ENTITY_ATTRIBUTE_RECORD is
			-- Finds the MOST RECENT attribute which was valid on or before 'a_date'.
			-- Note that the attribute may not have been valid on 'a_date' but it was at least valid
			-- before 'a_date'.
		local
			l_date: ROSE_DATE
			l_records: ROSE_LINEAR_ARRAY[ENTITY_ATTRIBUTE_RECORD]
		do
			l_date := valid_date(a_date)
			keys.search(a_entity)
			if keys.found then
				Result := keys.found_item
				if Result/=Void
				 then
					if Result.start_date.date>l_date
					 then
						Result := Void
					end
				end
			else
				l_records := internal_key_history(a_entity)
				if l_records /= Void then
					from l_records.start until l_records.off or Result /= Void loop
						if l_records.current_item.start_date.date <= l_date then
							Result := l_records.current_item
						end
						l_records.forth
					end
				end
			end
		end

	most_recent_key(a_entity: like entity_template;a_date: ROSE_DATE): STRING is
			-- Finds the value of the MOST RECENT attribute which was valid on or BEFORE 'a_date'.
		local
			l_attribute: ENTITY_ATTRIBUTE_RECORD
		do
			l_attribute := most_recent_attribute (a_entity, a_date)
			if l_attribute /= Void then
				Result := l_attribute.attribute_value
			end
		end

	key_history (a_entity: like entity_template): ROSE_LINEAR_ARRAY [ENTITY_ATTRIBUTE_RECORD] is
			-- The history of this attribute records for `a_entity'.
		require
			entity_not_void: a_entity /= Void
			correct_context: a_entity.context = context
			-- The following precondition is correct, but there is too much code that asks for non-existing attributes...
			-- And it seems not being in the identifiers hash map is no proof that there are really no attributes...
			--attribute_exists: context.category_handler_from_name (a_entity.category_name).identifiers.has (handle)
		local
			l_record: ENTITY_ATTRIBUTE_RECORD
			l_records: ROSE_LINEAR_ARRAY [ENTITY_ATTRIBUTE_RECORD]
			l_pes_result: PES_RESULT_ATTRIBUTES
		do
			keys.search (a_entity)
			if keys.found then
				l_record := keys.found_item
				if l_record /= Void then
					create Result.make_from_array (<<l_record>>)
				end
			else
				dated_keys.search (a_entity)
				if dated_keys.found then
					l_records := dated_keys.found_item
					if l_records /= Void then
						Result := l_records
					end
				elseif context.use_pes then
					l_pes_result := context.pes_api.send_attribute_request (context.context_index, <<a_entity>>, <<Current.name>>).get
					if l_pes_result.remote_exception then
						-- continue as if we were not using PES
						Result := internal_key_history (a_entity)
					else
						Result := l_pes_result.identifiers_entity.item (1).item (1)
						add_to_cache (a_entity, Result)
					end
				else
					Result := internal_key_history (a_entity)
				end
			end
			if Result = Void then
				create Result.make_empty
			end
		ensure
			result_not_void: Result /= Void
		end

	cache_entity_attribute_change (a_entity: like entity_template; l_attribute: ENTITY_ATTRIBUTE_RECORD) is
			-- Add the attribute change represented by `l_attribute' to the cached information for `a_entity'.
		require
			a_entity_not_void: a_entity /= Void
			l_attribute_not_void: l_attribute /= Void
		local
			l_dated_keys_item: ROSE_LINEAR_ARRAY[ENTITY_ATTRIBUTE_RECORD]
			l_current_record: ENTITY_ATTRIBUTE_RECORD
		do
			l_current_record := dated_attribute (a_entity, l_attribute.start_date)
			if l_current_record /= Void then
				l_current_record.set_end_date (l_attribute.start_date)
			end
			dated_keys.search (a_entity)
			if dated_keys.found then
				dated_keys.found_item.extend (l_attribute)
			else
				keys.search (a_entity)
				if keys.found then
					if keys.found_item = Void then -- should use .found_item, mixing client patterns!
						keys.force (l_attribute, a_entity)
					else
						create l_dated_keys_item.make_empty
						l_dated_keys_item.extend (keys.found_item) -- ditto .found_item
						l_dated_keys_item.extend (l_attribute)
						dated_keys.force_new (l_dated_keys_item, a_entity)
						keys.remove (a_entity)
					end
				else
					keys.force (l_attribute, a_entity)
				end
			end
		ensure
			only_multiple_in_dated: dated_keys.has (a_entity) implies dated_keys.item (a_entity).count > 1
			no_void_in_dated: dated_keys.has (a_entity) implies dated_keys.item (a_entity) /= Void and then not dated_keys.item (a_entity).has (Void)
		end

feature -- Prefetching

	add_to_preloaded_cache (a_attributes: ROSE_LINEAR_ARRAY[ENTITY_ATTRIBUTE_RECORD]; a_entity_id: INTEGER) is
			-- Add `a_attributes' to `dated_keys_by_id'. Called by SUPPLIER_CONTEXT
		require
			attributes_not_void: a_attributes /= Void
			no_void_attributes: not a_attributes.has (Void)
			entity_id_meaningful: a_entity_id > 0
			correct_attribute_and_entity_id: a_attributes.for_all (agent (a_record: ENTITY_ATTRIBUTE_RECORD; a_entity_id_arg: INTEGER): BOOLEAN do Result := a_record.attribute_type_id = handle and a_record.entity_id = a_entity_id_arg end (?, a_entity_id))
		do
			dated_keys_by_id_cache.force (a_attributes, a_entity_id)
		ensure
			cached: dated_keys_by_id_cache.has (a_entity_id) and then a_attributes.is_equal (dated_keys_by_id_cache.item (a_entity_id))
		end

	set_unique_key_records_by_category (a_array: ROSE_LINEAR_ARRAY[ENTITY_ATTRIBUTE_RECORD]; a_category: INTEGER) is
			-- Set the `unique_key_cache_by_category' hash map using `a_array'
		require
			array_not_empty: a_array /= Void and then not a_array.is_empty
			no_void_elements: not a_array.has (Void)
			-- valid_category_id: a_category > 0 -- Apparently not true for all contexts
			category_and_attribute_id_correct: a_array.for_all (agent (a_record: ENTITY_ATTRIBUTE_RECORD; a_category_id: INTEGER): BOOLEAN do Result := a_record.category_id = a_category_id and a_record.attribute_type_id = handle end (?, a_category))
		local
			l_duplicates: HASH_TABLE [BOOLEAN, STRING]
			l_unique_cache: HASH_TABLE [ENTITY_ATTRIBUTE_RECORD, STRING]
			l_value: STRING
			l_i: INTEGER
			l_record: ENTITY_ATTRIBUTE_RECORD
		do
			create l_duplicates.make (10_000)
			create l_unique_cache.make (10_000)
			unique_key_cache_by_category.force (l_unique_cache, a_category)
			if context.pes_server then
				non_unique_keys_by_category.force (l_duplicates, a_category)
			end

			-- A loop is faster than do_all
			-- a_array.do_all (agent collect_unique_attribute_records (?, l_unique_cache, l_duplicates))
			from
				l_i := a_array.lower
			until
				l_i > a_array.upper
			loop
				l_record := a_array.item (l_i)
				l_value := l_record.attribute_value.twin
				if l_value /= Void then
					l_value.to_lower
				end
				l_unique_cache.search (l_value)
				if not l_unique_cache.found then
					l_duplicates.search (l_value)
					if not l_duplicates.found then
						l_unique_cache.force (l_record, l_value)
					end
				else
					l_unique_cache.remove (l_value)
					l_duplicates.force (True, l_value)
				end

				l_i := l_i + 1
			end
		end

	load_key_history_records_by_category (a_category: INTEGER) is
			-- Used to preload attributes.  Called by CATEGORY_HANDLER.
		local
			l_array: ROSE_LINEAR_ARRAY[ENTITY_ATTRIBUTE_RECORD]
			i, j: INTEGER
			l_entity_id: INTEGER
		do
			if not category_loaded.safe_item(a_category) then
				l_array := context.all_key_records(a_category,Current)
				category_loaded.force(True, a_category)
				if not l_array.is_empty then
					l_entity_id := l_array.first.entity_id
					i := 1
					j := 1
					from l_array.start until l_array.off loop
						if l_array.current_item.entity_id /= l_entity_id then
							dated_keys_by_id_cache.force(l_array.subarray(i,j),l_entity_id)
							l_entity_id := l_array.current_item.entity_id
							i := l_array.index
							j := l_array.index
							l_array.forth
						else
							j := l_array.index
							l_array.forth
						end
					end
					dated_keys_by_id_cache.force(l_array.subarray(i,j),l_entity_id)
				end
			end
		end

	load_unique_key_records_by_category (a_category: INTEGER) is
			-- Load and cache and attribute values that appear only once for a_category
		local
			l_array: ROSE_LINEAR_ARRAY[ENTITY_ATTRIBUTE_RECORD]
		do
			unique_key_cache_by_category.search (a_category)
			if not unique_key_cache_by_category.found then
				l_array := context.all_key_records (a_category,Current)
				if not l_array.is_empty then
						set_unique_key_records_by_category (l_array, a_category)
				end
			end
		end

	collect_unique_attribute_records (a_record: ENTITY_ATTRIBUTE_RECORD; a_cache: HASH_TABLE [ENTITY_ATTRIBUTE_RECORD, STRING]; a_duplicates: HASH_TABLE [BOOLEAN, STRING]) is
			-- Agent to populate the unique attribute value cache
		require
			valid_record: a_record /= Void
			valid_caches: a_cache /= Void and a_duplicates /= Void
		local
			l_value: STRING
		do
			l_value := a_record.attribute_value.twin
			if l_value /= Void then
				l_value.to_lower
			end
			a_cache.search (l_value)
			if not a_cache.found then
				a_duplicates.search (l_value)
				if not a_duplicates.found then
					a_cache.force (a_record, l_value)
				end
			else
				a_cache.remove (l_value)
				a_duplicates.force (True, l_value)
			end
		end

	populate_primary_caches_from_secondary is
			-- Fill `keys' and `dated_keys' with the values from `dated_keys_by_id_cache'
		local
			l_values: ARRAYED_LIST[ROSE_LINEAR_ARRAY[ENTITY_ATTRIBUTE_RECORD]]
		do
			l_values := dated_keys_by_id_cache.linear_representation

			from
				l_values.start
			until
				l_values.after
			loop
				if l_values.item.count = 1 then
					keys.force (l_values.item.first, context.entity_at (l_values.item.first.category_id, l_values.item.first.entity_id))
				elseif l_values.item.count > 1 then
					dated_keys.force (l_values.item, context.entity_at (l_values.item.first.category_id, l_values.item.first.entity_id))
				end
				l_values.forth
			end
		end

	clear_cache is
			-- Wipe out all caches
		do
			dated_keys_by_id_cache.clear_all
			keys.wipe_out
			dated_keys.wipe_out
			unique_key_cache_by_category.clear_all
			category_loaded.clear_all
			entities.clear_all
			all_entities_table.clear_all
		end

feature {NONE} -- Internal key history related

	internal_key_history (a_entity: like entity_template): ROSE_LINEAR_ARRAY[ENTITY_ATTRIBUTE_RECORD] is
			-- The history of this attribute records for 'a_entity'.  Note that you must check 'keys' first.
		do
			Result := dated_keys.value(a_entity)
			if Result = Void then
				Result := dated_keys_by_id_cache.item(a_entity.handle)
				if Result = Void then
					if category_loaded.safe_item(a_entity.category_id) or else context.is_ready then
						create Result.make_empty
					else
	 					Result := context.entity_key_history(a_entity,Current)
	 				end
				end
				if Result.count <= 1 then
					if keys.count = 0 and then keys.capacity < 500 then
						keys.resize(500)
					end
					if Result.count = 0 then
						keys.force (Void, a_entity)
					else
						keys.force (Result.item(1),a_entity)
					end
				else
					if dated_keys.count = 0 and then dated_keys.capacity < 500 then
						dated_keys.resize(500)
					end
					dated_keys.force (Result,a_entity)
				end
			end
		ensure
			result_not_void: Result /= Void
			cached: (keys.has (a_entity) and then (Result.count = 0 or else Result.item(Result.lower).is_equal (keys.item (a_entity)))) or else dated_keys.has (a_entity) and then Result.is_equal (dated_keys.item (a_entity))
			only_multiple_in_dated: dated_keys.has (a_entity) implies dated_keys.item (a_entity).count > 1
			no_void_in_dated: dated_keys.has (a_entity) implies dated_keys.item (a_entity) /= Void and then not dated_keys.item (a_entity).has (Void)
		end

	entities_table (a_category: INTEGER): HASH_TABLE[ROSE_LINEAR_ARRAY[ENTITY_ATTRIBUTE_RECORD],STRING] is
			-- Returns the hashmap used for finding attribute records by their value, creates the hashmap if necessary
		do
			Result := entities.safe_item (a_category)
			if Result = Void then
				create Result.make (500)
				entities.force (Result,a_category)
			end
		end

	valid_date (a_date: ROSE_DATE): ROSE_DATE is
			-- A valid date given a_date
		do
			if a_date /= Void then
				Result := a_date
			else
				Result := once_today
			end
		ensure
			valid_result: Result /= Void
		end

feature -- Related Entities.
	related_entities_during (a_key : STRING; a_category_id : INTEGER; a_start_date, a_end_date : ROSE_DATE): ROSE_LINEAR_ARRAY [ENTITY]
			-- Entites which match `a_key' between `a_start' and `a_end'.
		require
			key_not_void: a_key /= Void
			category_id_meaningful: a_category_id > 0
		local
			l_start, l_end : ROSE_DATE
			l_records : ROSE_LINEAR_ARRAY [ENTITY_ATTRIBUTE_RECORD]
			l_record : ENTITY_ATTRIBUTE_RECORD
			l_handler : CATEGORY_HANDLER [like context_anchor, ENTITY]
			l_entity : ENTITY
			l_entity_hash : HASHED_SET [ENTITY]
		do
			l_handler := context.handler (a_category_id)
			create l_entity_hash.make
			if l_handler /= Void then
				l_start := a_start_date
				l_end := a_end_date
				if l_start = Void then
					create l_start.make_min_date
				end
				if l_end = Void then
					create l_end.make_max_date
				end
				l_records := entity_records (a_key, a_category_id)
				from l_records.start until l_records.off loop
					l_record := l_records.current_item
					if l_record.start_date < a_end_date and l_record.end_date > a_start_date then
						l_entity := l_handler.entity (l_record.entity_id)
						l_entity_hash.put (l_entity)
					end
					l_records.forth
				end
			end
			create Result.make_from_array (l_entity_hash.current_keys)
		ensure
			result_not_void : Result /= Void
		end


feature {SUPPLIER_CONTEXT, ZONE, RCMD_ROSE_ENTITY_UPDATE} -- Access to attribute records

	all_unique_records_by_category (a_category_id: INTEGER): ARRAYED_LIST[ENTITY_ATTRIBUTE_RECORD] is
			-- Returns all unique attributes for category `a_category_id'
		require
			valid_category: a_category_id > 0
		local
			l_hash_table: HASH_TABLE[ENTITY_ATTRIBUTE_RECORD,STRING]
		do
			load_unique_key_records_by_category (a_category_id)
			l_hash_table := unique_key_cache_by_category.item (a_category_id)
			Result := l_hash_table.linear_representation
		end

	unique_records_by_category (a_key: STRING; a_category_id: INTEGER): ROSE_LINEAR_ARRAY[ENTITY_ATTRIBUTE_RECORD] is
			-- Return a first pass cache value if a_key is unique for `a_category_id'
		require
			valid_key: a_key /= Void
		local
			l_unique_cache: HASH_TABLE [ENTITY_ATTRIBUTE_RECORD, STRING]
		do
			unique_key_cache_by_category.search (a_category_id)
			if unique_key_cache_by_category.found then
				l_unique_cache := unique_key_cache_by_category.found_item
				l_unique_cache.search (a_key.as_lower)
				if l_unique_cache.found then
					create Result.make_from_array (<<l_unique_cache.found_item>>)
				end
			end
		end

	entity_records (a_key: STRING;a_category_id: INTEGER): ROSE_LINEAR_ARRAY [ENTITY_ATTRIBUTE_RECORD] is
			-- Attribute records in 'a_category_id', matching 'a_key'.  Note that 'a_key' may
			-- include wildcards (%).
		require
			key_not_void: a_key /= Void
			category_id_meaningful: a_category_id > 0
		local
			l_entities: HASH_TABLE[ROSE_LINEAR_ARRAY[ENTITY_ATTRIBUTE_RECORD],STRING]
		do
			if not context.pes_server then
				Result := context.entity_records(a_key,context.handler (a_category_id),Current)
			else
				if a_key /= Void then
					if not a_key.has('%%') then
						l_entities := entities_table(a_category_id)
						Result := l_entities.item(a_key)
						if Result = Void then
							Result := context.entity_records(a_key,context.handler (a_category_id),Current)
							l_entities.put(Result,a_key)
						end
					else
						Result := context.entity_records(a_key,context.handler (a_category_id),Current)
					end
				end
			end
		ensure
			result_not_void: Result /= Void
		end

	all_entity_records (a_key: STRING): ROSE_LINEAR_ARRAY[ENTITY_ATTRIBUTE_RECORD] is
			-- Attribute records in ANY category, matching 'a_key'.  Note that 'a_key' may
			-- include wildcards (%).
		local
			l_entities: HASH_TABLE[ROSE_LINEAR_ARRAY[ENTITY_ATTRIBUTE_RECORD],STRING]
		do
			if a_key /= Void then
				if not a_key.has('%%') then
					l_entities := all_entities_table
					Result := l_entities.item(a_key)
					if Result = Void then
						Result := context.all_entity_records_from_key(a_key,Current)
						l_entities.put(Result,a_key)
					end
				else
					Result := context.all_entity_records_from_key(a_key,Current)
				end
			end
		ensure
			result_not_void: Result /= Void
		end

feature {PES_RESULT_RECIPE} -- PES interface

	add_to_cache (a_entity: ENTITY; a_attribute_records: ROSE_LINEAR_ARRAY[ENTITY_ATTRIBUTE_RECORD]) is
			-- Add a_attribute_record to corresponding cache
		require
			entity_not_void: a_entity /= Void
			attribute_records_not_void: a_attribute_records /= Void
		do
			if a_attribute_records.count = 1 then
				keys.force (a_attribute_records.item (a_attribute_records.lower), a_entity)
			elseif a_attribute_records.count > 1 then
				dated_keys.force (a_attribute_records, a_entity)
			else
				keys.force (Void, a_entity)
			end
		ensure
			cached: (keys.has (a_entity) and then (a_attribute_records.count = 0 or else a_attribute_records.item(a_attribute_records.lower).is_equal (keys.item (a_entity)))) or else dated_keys.has (a_entity) and then a_attribute_records.is_equal (dated_keys.item (a_entity))
			only_multiple_in_dated: dated_keys.has (a_entity) implies dated_keys.item (a_entity).count > 1
			no_void_in_dated: dated_keys.has (a_entity) implies dated_keys.item (a_entity) /= Void and then not dated_keys.item (a_entity).has (Void)
		end

feature {SUPPLIER_CONTEXT} -- PES updates

	update_attribute_record (a_record: ENTITY_ATTRIBUTE_RECORD) is
			-- Update attribute recoard `a_record' in the caches
		require
			record_not_void: a_record /= Void
			correct_type: a_record.attribute_type_id = handle
		local
			l_record_array: ROSE_LINEAR_ARRAY[ENTITY_ATTRIBUTE_RECORD]
			l_entity: ENTITY
		do
			entity_attribute_predicate.retrieve (Void, <<a_record.category_id, a_record.attribute_type_id, a_record.entity_id>>)
			l_record_array := entity_attribute_predicate.array

			dated_keys_by_id_cache.force (l_record_array, l_record_array.first.entity_id)

			l_entity := entity_from_handle (a_record.entity_id, a_record.category_id)
			if l_record_array.count = 1 then
				keys.force (l_record_array.first, l_entity)
			elseif l_record_array.count > 1 then
				dated_keys.force (l_record_array, l_entity)
			end

			if unique_key_cache_by_category.item (a_record.category_id) /= Void then
				update_unique_key_records_by_category (a_record)
			end
			update_entities_table (a_record)
		end

	update_unique_key_records_by_category (a_record: ENTITY_ATTRIBUTE_RECORD) is
			-- Update the `unique_key_cache_by_category' hash map using `a_array'
		require
			record_not_void: a_record /= Void
			-- valid_category_id: a_category > 0 -- Apparently not true for all contexts
			category_and_attribute_id_correct: a_record.attribute_type_id = handle
			category_already_cached: unique_key_cache_by_category.item (a_record.category_id) /= Void
			category_duplicates_ready: non_unique_keys_by_category.item (a_record.category_id) /= Void
			is_pes_server: context.pes_server
		local
			l_duplicates: HASH_TABLE [BOOLEAN, STRING]
			l_unique_cache: HASH_TABLE [ENTITY_ATTRIBUTE_RECORD, STRING]
			l_value: STRING
			l_category: INTEGER
		do
			l_category := a_record.category_id
			l_unique_cache := unique_key_cache_by_category.item (l_category)
			l_duplicates := non_unique_keys_by_category.item (l_category)

			-- A loop is faster than do_all
			-- a_array.do_all (agent collect_unique_attribute_records (?, l_unique_cache, l_duplicates))

			l_value := a_record.attribute_value.twin
			if l_value /= Void then
				l_value.to_lower
			end
			l_unique_cache.search (l_value)
			if not l_unique_cache.found then
				l_duplicates.search (l_value)
				if not l_duplicates.found then
					l_unique_cache.force (a_record, l_value)
				end
			else
				l_unique_cache.remove (l_value)
				l_duplicates.force (True, l_value)
			end
		end

	update_entities_table (a_record: ENTITY_ATTRIBUTE_RECORD) is
			-- Update the entities hash map
		require
			record_not_void: a_record /= Void
		local
			l_entities_hash_map: HASH_TABLE [ROSE_LINEAR_ARRAY [ENTITY_ATTRIBUTE_RECORD], STRING]
			l_array: ROSE_LINEAR_ARRAY [ENTITY_ATTRIBUTE_RECORD]
		do
			entity_attribute_value_predicate.retrieve (Void, <<a_record.category_id, a_record.attribute_type_id, a_record.attribute_value>>)
			l_array := entity_attribute_value_predicate.array

			if entities.valid_index (a_record.category_id) then
				l_entities_hash_map := entities.item (a_record.category_id)
			end
			if l_entities_hash_map /= Void then
				l_entities_hash_map.force (l_array, a_record.attribute_value)
			end
		ensure
			cached_in_entities: a_record.category_id >= entities.lower implies (entities.valid_index (a_record.category_id) and then entities.item (a_record.category_id) /= Void and then entities.item (a_record.category_id).has (a_record.attribute_value) and then entities.item (a_record.category_id).item (a_record.attribute_value).has (a_record))
		end

feature {NONE} -- PES Implementation

	entity_attribute_predicate: like entity_attribute_predicate_saved is
			-- Predicate to retrieve attribute records for a given category_id and attribute_type_id and entity_id
		require
			server_not_void: server /= Void
		do
			if entity_attribute_predicate_saved = Void then
				create entity_attribute_predicate_saved.make_verbose (server,"select * from entity_attribute with (readuncommitted) where category_id = ? and attribute_type_id = ? and entity_id = ? order by attribute_value,end_date desc","entity_attribute_update_predicate",name)
			end
			Result := entity_attribute_predicate_saved
		end

	entity_attribute_predicate_saved: COLLECTING_PREDICATE [ENTITY_ATTRIBUTE_RECORD]

	entity_attribute_value_predicate: like entity_attribute_predicate_saved is
			-- Predicate to retrieve attribute records for a given category_id and attribute_type_id and attribute_value
		require
			server_not_void: server /= Void
		do
			if entity_attribute_value_predicate_saved = Void then
				create entity_attribute_value_predicate_saved.make_verbose (server,"select * from entity_attribute with (readuncommitted) where category_id = ? and attribute_type_id = ? and attribute_value = ? order by end_date desc","entity_attribute_value_predicate",name)
			end
			Result := entity_attribute_value_predicate_saved
		end

	 entity_attribute_value_predicate_saved: COLLECTING_PREDICATE [ENTITY_ATTRIBUTE_RECORD]

feature {SUPPLIER_CONTEXT, PES_QUERY_LOAD_CATEGORY} -- Attribute caches

	-- The next three are all related.  They allow you to find attributes if you have an entity.
	keys: DS_HASH_TABLE[ENTITY_ATTRIBUTE_RECORD,like entity_template]
			-- contains attributes that only have one version.
	dated_keys: DS_HASH_TABLE[ROSE_LINEAR_ARRAY[ENTITY_ATTRIBUTE_RECORD],like entity_template]
			-- contains attributes that have more than one version.
	dated_keys_by_id_cache: HASH_TABLE[ROSE_LINEAR_ARRAY[ENTITY_ATTRIBUTE_RECORD],INTEGER]
			-- contains preloaded values.  They will be used to populate the
			-- first two as needed.

	unique_key_cache_by_category: HASH_TABLE [HASH_TABLE [ENTITY_ATTRIBUTE_RECORD, STRING], INTEGER]
			-- A cache of attribute values that are unique by category
			-- This cache is used as a first round performance improvement for access to entities by attribute value

	non_unique_keys_by_category: HASH_TABLE [HASH_TABLE[BOOLEAN, STRING], INTEGER]
			-- A cache of the non-unique attribute values by category
			-- Used by the update_attribute record feature

	category_loaded: ROSE_LINEAR_ARRAY[BOOLEAN]
			-- This table tells you if 'dated_keys_by_id_cache' has been loaded for a particular category.

	unique_key_records_loaded: BOOLEAN
			-- Unused legacy attribute kept in case of storable (invariant False)

	add_to_entities_table (a_record: ENTITY_ATTRIBUTE_RECORD) is
			-- Add `a_record' to the `entities' and `all_entities' table.
		require
			record_not_void: a_record /= Void
			correct_attribute_type_id: a_record.attribute_type_id = handle
		local
			l_category_id: INTEGER
			l_hash_map: HASH_TABLE[ROSE_LINEAR_ARRAY[ENTITY_ATTRIBUTE_RECORD], STRING]
			l_array: ROSE_LINEAR_ARRAY[ENTITY_ATTRIBUTE_RECORD]
		do
			-- entities table
			l_category_id := a_record.category_id
			if l_category_id >= entities.lower then
				l_hash_map := entities.safe_item (l_category_id)
				if l_hash_map = Void then
					create l_hash_map.make (100)
					entities.force (l_hash_map, l_category_id)
				end
				l_array := l_hash_map.item (a_record.attribute_value)
				if l_array /= Void then
					l_array.extend (a_record)
				else
					create l_array.make (1, 1)
					l_array.item (1) := a_record
					l_hash_map.put (l_array, a_record.attribute_value)
				end
			end

			-- all_entities_table
			-- Don't do this for now, since it seems to be of little use
		ensure
			cached_in_entities: a_record.category_id >= entities.lower implies (entities.valid_index (a_record.category_id) and then entities.item (a_record.category_id) /= Void and then entities.item (a_record.category_id).has (a_record.attribute_value) and then entities.item (a_record.category_id).item (a_record.attribute_value).has (a_record))
		--	cached_in_all_entities: all_entities_table.has (a_record.attribute_value) and then all_entities_table.item (a_record.attribute_value).has (a_record)
		end

	-- The next two are related.  They allow you to find an attribute (and therefore its entity)
	-- if you have an attribute_value.
	entities: ROSE_LINEAR_ARRAY[HASH_TABLE[ROSE_LINEAR_ARRAY[ENTITY_ATTRIBUTE_RECORD],STRING]]
			-- contains one hash table for each category.  The hash table contains attributes
			-- hashed by their attribute_values.
	all_entities_table: HASH_TABLE[ROSE_LINEAR_ARRAY[ENTITY_ATTRIBUTE_RECORD],STRING]
			-- a single hash table that contains attributes from ALL categories
			-- hashed by their attribute_values.

feature {NONE} -- Implementation

	once_today: ROSE_DATE is
			-- Today's date available to all descendants in system
		once
			create Result.make_now
		end

invariant
	legacy_unique_key_records_loaded_flag_false: not unique_key_records_loaded

end








