indexing

	description: "Sequence of floating point values indexed by named entity"

	Author: "Feng Zhou"
	last_editor: "$Author: c4 $"
	date: "$Date: 2014/05/13 $"
	revision: "$Revision: #52 $"
	path: "$File: //rose/source/kernel/entities/entity_values.e $"

class	ENTITY_VALUES [reference G->HAS_NAME]

inherit

	ROSE_LINEAR_ARRAY [PAIR [G,DOUBLE]]
		rename
			infix "+" as array_plus
		redefine
			force, insert, put
		end

	ROSE_LEGACY_UNDEFINED
		export
			{NONE} all
		undefine
			copy, default_create, is_equal
		end

create

	make, make_empty,
	make_from_array, make_from_linear,
	make_from_entities_vector

feature {NONE} -- Initialization

	make_from_entities_vector (a_vector: ABSTRACT_READABLE_VECTOR; a_entities: ROSE_LINEAR_ARRAY [G]) is
			-- Initialize from list of entities and corresponding vector.
		require
			a_entities_not_void: a_entities /= Void
			a_entities_lower_one: a_entities.lower = 1
			matching_count: a_vector /= Void implies a_vector.count = a_entities.count
		local
			i, k: INTEGER
			l_vector: VECTOR
		do
			if a_vector = Void then
				create l_vector.make (a_entities.count)
			elseif attached {VECTOR} a_vector as l_v then
				l_vector := l_v
			else
				create l_vector.duplicate (a_vector)
			end
			k := l_vector.count
			make (1, k)
			from
				a_entities.start
			until
				a_entities.after
			loop
				i := a_entities.index
				pair_put (a_entities.item (i), l_vector.item (i), i)
				a_entities.forth
			end
		ensure
			correct_count: count = a_entities.count
			name_void: name = Void
			same_entities: entities.same_items (a_entities)
			same_vector: a_vector /= Void implies a_vector.same_vector (raw_vector)
		end

feature -- Access

	name: STRING
			-- Name

	names: ROSE_LINEAR_ARRAY [STRING] is
			-- Name of entities
		local
			l_index: INTEGER
		do
			l_index := index
			create Result.make (1, count)
			from
				start
			until
				after
			loop
				Result.put (current_name, index)
				forth
			end
			index := l_index
		ensure
			names_not_void: Result /= Void
			correct_count: Result.count = count
		end

feature -- Access

	current_entity: G is
			-- Entity for iteration
		require
			valid_index: valid_index (index)
		do
			if current_item /= Void then
				Result := current_item.first
			end
		end

	current_value: DOUBLE is
			-- Value for iteration
		require
			valid_index: valid_index (index)
		do
			if current_item /= Void then
				Result := current_item.second
			end
		end

	current_name: STRING is
			-- Name for iteration
		require
			valid_index: valid_index (index)
		local
			l_entity: G
		do
			l_entity :=  current_entity
			if l_entity /= Void then
				Result := l_entity.name_at (Void)
			end
		end

feature -- Basic operations

	set_name (a_name: STRING) is
			-- Set name.
		do
			name := a_name
		ensure
			name_set: name = a_name
		end

	force (v: like item; i:INTEGER) is
			-- Assign item `v' to `i'-th entry.
			-- Like ARRAY.force, with entity index update.
		do
			entity_table_saved := Void
			Precursor (v, i)
		ensure then
			entity_table_flushed: entity_table_saved = Void
		end

	insert (a_value: like item; pos: INTEGER) is
			-- Add `a_value' at `pos', moving subsequent items
			-- to the right.
			-- Like ARRAY.insert, with entity index update.
		do
			entity_table_saved := Void
			Precursor (a_value, pos)
		ensure then
			entity_table_flushed: entity_table_saved = Void
		end

	put (v: like item; i: INTEGER) is
			-- Replace `i'-th entry, if in index interval, by `v'.
			-- Like ARRAY.put, with entity index update.
		do
			entity_table_saved := Void
			Precursor (v, i)
		ensure then
			entity_table_flushed: entity_table_saved = Void
		end

	pair_extend (a_entity: G; a_value: DOUBLE) is
			-- Add value for entity.
		require
			a_entity_not_void: a_entity /= Void
			not_has: not has_entity (a_entity)
		local
			l_pair: PAIR [G,DOUBLE]
		do
			create l_pair.make (a_entity, a_value)
			extend (l_pair)
			check
				entity_table_flushed: entity_table_saved = Void
				-- postcondition of `force', which is called by extend
			end
		ensure
			entity_table_temporarily_flushed: entity_table_saved = Void
			has_entity: has_entity (a_entity)
			position_unchanged: index = old index
		end

	pair_put (a_entity: G; a_value: DOUBLE; a_index: INTEGER) is
			-- Put `a_entity' and `a_value' at `a_index'.
		require
			valid_index: valid_index (a_index)
		local
			l_pair: PAIR [G,DOUBLE]
		do
			create l_pair.make (a_entity, a_value)
			put (l_pair, a_index)
			check
				entity_table_flushed: entity_table_saved = Void
				-- postcondition of `put'
			end
		ensure
			entity_table_temporarily_flushed: entity_table_saved = Void
			has_entity: a_entity /= Void implies has_entity (a_entity)
			position_unchanged: index = old index
		end

	pair_add (a_entity: G; a_value: DOUBLE) is
			-- Add to existing value.
		local
			l_ix: INTEGER
			l_value: DOUBLE
		do
			l_ix := entity_table.item (a_entity)
			if l_ix > 0 then
				l_value := item (l_ix).second + a_value
				put_value (l_value, l_ix)
			else
				pair_extend (a_entity, a_value)
			end
		ensure
			entity_table_rebuilt_on_demand: True
					-- it may have already been re-built by the postcondition of `pair_extend'
			position_unchanged: index = old index
			definition_existing: (a_entity /= Void and then old (has_entity (a_entity)) and (a_value = a_value) and
				(old (value_at (a_entity)) = old (value_at (a_entity)))) implies
					value_at (a_entity) = a_value + old (value_at (a_entity))
			definition_new: (a_entity /= Void and then not old (has_entity (a_entity)) and (a_value = a_value)) implies
				value_at (a_entity) = a_value
		end

	put_value (a_value: DOUBLE; a_index: INTEGER) is
			-- Change value at `a_index' to `a_value'.
		require
			valid_index: valid_index (a_index)
		do
			item (a_index).set_second (a_value)
		ensure
			position_unchanged: index = old index
			defined_value_set: (a_value = a_value) implies item (a_index).second = a_value
			undefined_value_set: (a_value /= a_value) implies (item (a_index) /= item (a_index))
		end

	put_vector (a_vector: ABSTRACT_READABLE_VECTOR) is
			-- Replace values with values at same index position in `a_vector'.
		require
			a_vector_not_void: a_vector /= Void
			a_vector_count: a_vector.count <= count
		local
			i: INTEGER
		do
			from
				i := 1
			until
				i > a_vector.count
			loop
				put_value (a_vector.item (i), i)
				i := i + 1
			end
		ensure
			position_unchanged: index = old index
			all_values_set: a_vector.same_vector (raw_vector.sub_vector (1, a_vector.count))
		end

	normalize is
			-- Normalize values.
		local
			l_vector: VECTOR
		do
			l_vector := raw_vector.normalized
			put_vector (l_vector)
		ensure
			usually_sum_1_0: True -- inheritance from ROSE_LEGACY_UNDEFINED means I hesitate to assert anything stronger
		end

	normalize_with_residual (a_residual: DOUBLE) is
			-- Normalize values with `a_residual'.
		require
			a_residual_ok: a_residual /= - raw_vector.sum
		local
			l_vector: VECTOR
		do
			l_vector := raw_vector.normalized_with_residual (a_residual)
			put_vector (l_vector)
		end

feature -- Query

	vector: VECTOR is
			-- Values as vector, with names
		do
			Result := raw_vector
			Result.set_names (names)
			Result.set_name (name)
		ensure
			vector_not_void: Result /= Void
			correct_count: Result.count = count
			name_set: Result.name = name
		end

	raw_vector: VECTOR is
			-- Values as vector, without name;
			-- Duplicate of `values'
		local
			l_index: INTEGER
		do
			l_index := index
			create Result.make (count)
			from
				start
			until
				after
			loop
				Result.put (current_value, index)
				forth
			end
			index := l_index
		ensure
			raw_vector_not_void: Result /= Void
			correct_count: Result.count = count
			no_name: Result.name = Void
			no_names: Result.names = Void
		end

	values: VECTOR is
			-- Values as vector, without name;
			-- Duplicate of `raw_vector'
		do
			Result := raw_vector
		ensure
			value_not_void: Result /= Void
			correct_count: Result.count = count
			no_name: Result.name = Void
			no_names: Result.names = Void
		end

	entities: ROSE_LINEAR_ARRAY [G] is
			-- List of entities in order
		local
			l_index: INTEGER
		do
			l_index := index
			create Result.make (1, count)
			from
				start
			until
				after
			loop
				Result.put (current_entity, index)
				forth
			end
			index := l_index
		ensure
			entities_not_void: Result /= Void
			same_count: Result.count = count
		end

	sub_values (a_beg_ix, a_end_ix: INTEGER): like subset_by_value is
			-- Contiguous subset from `a_beg_ix' to `a_end_ix' inclusively
		require
			valid_index: valid_index (a_beg_ix) and valid_index (a_end_ix) and a_beg_ix <= a_end_ix
		local
			i: INTEGER
		do
			create Result.make (1, a_end_ix - a_beg_ix + 1)
			from
				i := a_beg_ix
			until
				i > a_end_ix
			loop
				Result.put (item (i), i - a_beg_ix + 1)
				i := i + 1
			end
			Result.set_name (name)
		ensure
			sub_values_not_void: Result /= Void
			correct_count: Result.count = a_end_ix - a_beg_ix + 1
		end

feature -- Query

	value_at (a_entity: G): DOUBLE is
			-- Value for `a_entity'
		local
			l_table: ROSE_HASH_TABLE [INTEGER,G]
			l_ix: INTEGER
		do
			if a_entity /= Void then
				l_table := entity_table
				l_ix := l_table.item (a_entity)
				if l_ix > 0 then
					check
						valid_index: valid_index (l_ix)
						-- mutation features of `Current' ensure that `entity_table' gets re-built when changes occur
					end
					Result := item (l_ix).second
				else
					Result := undefined_double
				end
			end
		ensure
			void_definition: a_entity = Void implies Result = 0.0
		end

	has_entity (a_entity: G): BOOLEAN is
			-- Is there a value for entity?
		local
			l_table: ROSE_HASH_TABLE [INTEGER,G]
		do
			if a_entity /= Void then
				l_table := entity_table
				Result := l_table.has (a_entity)
			end
		ensure
			void_definition: a_entity = Void implies not Result
		end

	not_has_entity (a_entity: G): BOOLEAN is
			-- Negation of `has_entity' for contracts.
		do
			Result := not has_entity (a_entity)
		ensure
			definition: Result = not has_entity (a_entity)
		end

	put_at (a_value: DOUBLE; a_entity: G) is
			-- Change existing value.
		require
			a_entity_has: has_entity (a_entity)
		local
			l_index: INTEGER
		do
			l_index := entity_table.item (a_entity)
			item (l_index).set_second (a_value)
		ensure
			done: (a_value = a_value) implies value_at (a_entity) = a_value
		end

	force_at (a_value: DOUBLE; a_entity: G) is
			-- Change existing or add new value.
		require
			a_entity_not_void: a_entity /= Void
		do
			if has_entity (a_entity) then
				put_at (a_value, a_entity)
			else
				pair_extend (a_entity, a_value)
			end
		ensure
			done: (a_value = a_value) implies value_at (a_entity) = a_value
		end

feature {NONE} -- Implementation

	entity_table: ROSE_HASH_TABLE [INTEGER,G] is
			-- Table of position index from entities
		local
			l_index: INTEGER
		do
			if entity_table_saved = Void then
				create Result.make (count.max (10))
				l_index := index
				from
					start
				until
					after
				loop
					Result.put (index, current_entity)
					forth
				end
				entity_table_saved := Result
				index := l_index
			else
				Result := entity_table_saved
			end
		ensure
			entity_table_not_void: Result /= Void
			entity_table_cached: Result = entity_table_saved
		end

	entity_table_saved: like entity_table
			-- Cache for `entity_table'

feature {NONE} -- Access

	label: ROSE_LINEAR_ARRAY [STRING]
			-- Unused attribute (kept in case of storables)

feature -- Derived structures

	merged (a_other: ENTITY_VALUES [G]): ENTITY_MATRIX [G] is
			-- Entity matrix with values from `Current' in column 1
			--  and values from `a_other' (if non-void) in column 2
		local
			l_others: ROSE_LINEAR_ARRAY [ENTITY_VALUES [G]]
		do
			create l_others.make_empty
			if a_other /= Void then
				l_others.extend (a_other)
			end
			Result := merge_others (l_others)
		ensure
			merged_not_void: Result /= Void
			entities_not_void: Result.entities /= Void
			count: (a_other = Void and Result.width = 1) or else Result.width = 2
		end

	included (a_list: ARRAY [G]): like filtered is
			-- Filtered copy of `Current' with only entities from `a_list'
		require
			a_list_not_void: a_list /= Void
		local
			l_set: DS_HASH_SET [G]
		do
				-- Create set with O(1) membership test
			create l_set.make_equal (0)
			a_list.do_all (agent l_set.force)

			Result := filtered (agent l_set.has)
		ensure
			result_not_void: Result /= Void
			items_filtered: Result.count > 0 implies a_list.has (Result.first.first) -- approximates all items, `has' is O(n)
		end

	excluded (a_list: ARRAY [G]): like filtered is
			-- Filtered copy of `Current' excluding entities from `a_list'
		require
			a_list_not_void: a_list /= Void
		local
			l_set: DS_HASH_SET [G]
		do
				-- Create set with O(1) membership test
			create l_set.make_equal (0)
			a_list.do_all (agent l_set.force)

			Result := filtered (agent set_not_has (l_set, ?))
		ensure
			result_not_void: Result /= Void
			items_filtered: Result.count > 0 implies not a_list.has (Result.first.first) -- approximates all itmes, `has' is O(n)
			items_excluded: a_list.for_all (agent Result.not_has_entity)
		end

	filtered (a_filter: FUNCTION [ANY, TUPLE [G], BOOLEAN]): ENTITY_VALUES [G] is
			-- Values for all entities that satisfy filter.
		require
			a_filter_not_void: a_filter /= Void
		do
			create Result.make_empty
			do_all (agent filtered_one (Result, a_filter, ?))
		ensure
			not_void: Result /= Void
			items_filtered: Result.count > 0 implies a_filter.item ([Result.first.first]) -- shortcut for test on all items
		end

feature {NONE} -- Implementation

	filtered_one (a_result: like filtered; a_filter: FUNCTION [ANY, TUPLE [G], BOOLEAN]; a_item: PAIR [G, DOUBLE]) is
			-- Implementation agent for `filtered'.
		require
			a_result_not_void: a_result /= Void
			a_filter_not_void: a_filter /= Void
			a_item_not_void: a_item /= Void
		do
			if a_filter.item ([a_item.first]) then
				a_result.extend (a_item)
			end
		end

	set_not_has (a_set: DS_HASH_SET [G]; a_item: G): BOOLEAN is
			-- not DS_SET.has for use as agent
		require
			a_set_not_void: a_set /= Void
		do
			Result := not a_set.has (a_item)
		ensure
			definition: Result = not a_set.has (a_item)
		end

feature -- Derived structures

	merge_others (a_others: ROSE_LINEAR_ARRAY [ENTITY_VALUES [G]]): ENTITY_MATRIX [G] is
			-- Turn set of entity values into ENTITY_MATRIX with one column for
			-- `Current' and then one column from each from `a_others'
		require
			a_other_not_void: a_others /= Void
			no_void_others: not a_others.has (Void)
		local
			l_table: ROSE_HASH_TABLE [INTEGER,G]
			l_other: like subset_by_value
			j, k, l_index, l_other_index: INTEGER
			l_entities: ROSE_LINEAR_ARRAY [G]
			l_label: ROSE_LINEAR_ARRAY [STRING]
			l_matrix: MATRIX
		do
			if not a_others.is_empty then
				l_index := index
				l_other_index := a_others.index
				create l_label.make (1, a_others.count + 1)
				create l_entities.make (1, count)
				k := count
				from
					a_others.start
				until
					a_others.after
				loop
					k := k + a_others.current_item.count
					a_others.forth
				end
				create l_matrix.make (k, a_others.count + 1)
				if name /= Void then
					l_label.put (name, 1)
				else
					l_label.put ("column_1", 1)
				end
				from
					start
				until
					after
				loop
					l_matrix.put (current_value, index, 1)
					l_entities.put (current_entity, index)
					forth
				end
				k := count
				l_table := entity_table
				from
					a_others.start
				until
					a_others.after
				loop
					l_other := a_others.current_item
					if l_other.name /= Void then
						l_label.put (l_other.name, a_others.index + 1)
					else
						l_label.put ("column_" + (a_others.index + 1).out, a_others.index + 1)
					end
					from
						l_other.start
					until
						l_other.after
					loop
						j := l_table.item (l_other.current_entity)
						if j > 0 then
							l_matrix.put (l_other.current_value, j, a_others.index + 1)
						else
							k := k + 1
							l_matrix.put (l_other.current_value, k, a_others.index + 1)
							l_entities.extend (l_other.current_entity)
						end
						l_other.forth
					end
					a_others.forth
				end
				l_matrix := l_matrix.top_rows (k)
				create Result.make_all (l_entities, l_matrix, l_label)
				index := l_index
				a_others.go_i_th (l_other_index)
			else
				Result := to_entity_matrix
			end
		ensure
			merge_others_not_void: Result /= Void
			entities_not_void: Result.entities /= Void
			count: (a_others = Void and Result.width = 1) or else (Result.width = a_others.count + 1)
		end

feature -- Subset

	subset_by_value (a_value: DOUBLE): like Current is
			-- Subset where items are exactly equal to `a_value'
		local
			i, l_index: INTEGER
		do
			create Result.make (1, count)
			l_index := index
			from
				start
			until
				after
			loop
				if current_value = a_value then
					i := i + 1
					Result.put (current_item, i)
				end
				forth
			end
			index := l_index
			Result := Result.subarray (1, i) -- empty if i = 0
		ensure
			subset_by_value_not_void: Result /= Void
		end

	prune_by_value (a_value: DOUBLE): like subset_by_value is
			-- Subset with all values other than `a_value'
		local
			i, l_index: INTEGER
		do
			l_index := index
			create Result.make (1, count)
			from
				start
			until
				after
			loop
				if current_value /= a_value then
					i := i + 1
					Result.put (current_item, i)
				end
				forth
			end
			index := l_index
			Result := Result.subarray (1, i) -- empty if i = 0
		ensure
			prune_by_value_not_void: Result /= Void
		end

	prune_by_value_less (a_value: DOUBLE): like subset_by_value is
			-- Subset with all values strictly less than `a_value'
		local
			i, l_index: INTEGER
		do
			l_index := index
			create Result.make (1, count)
			from
				start
			until
				after
			loop
				if current_value < a_value then
					i := i + 1
					Result.put (current_item, i)
				end
				forth
			end
			index := l_index
			Result := Result.subarray (1, i) -- empty if i = 0
		ensure
			prune_by_value_less_not_void: Result /= Void
		end

	prune_by_value_greater (a_value: DOUBLE): like subset_by_value is
			-- Subset with all values strictly greater than `a_value'
		local
			i, l_index: INTEGER
		do
			l_index := index
			create Result.make (1, count)
			from
				start
			until
				after
			loop
				if current_value > a_value then
					i := i + 1
					Result.put (current_item, i)
				end
				forth
			end
			index := l_index
			Result := Result.subarray (1, i) -- empty if i = 0
		ensure
			prune_by_value_greater_not_void: Result /= Void
		end

	prune_with (a_evaluate: FUNCTION [ANY, TUPLE [DOUBLE], BOOLEAN]): like subset_by_value is
			-- Subset with all values that pass the test
			-- defined by the agent
		local
			i, l_index: INTEGER
		do
			l_index := index
			create Result.make (1, count)
			from
				start
			until
				after
			loop
				if a_evaluate.item ([current_value]) then
					i := i + 1
					Result.put (current_item, i)
				end
				forth
			end
			index := l_index
			Result := Result.subarray (1, i)
		ensure
			prune_with_not_void: Result /= Void
		end

	subgroup (a_entities: ROSE_LINEAR_ARRAY [G]): like subset_by_value  is
			-- Subset with entities in `a_entities'
		require
			a_entities_not_void: a_entities /= Void
		local
			i, k, l_index: INTEGER
		do
			create Result.make (1, a_entities.count)
			l_index := a_entities.index
			from
				a_entities.start
			until
				a_entities.after
			loop
				k := entity_table.item (a_entities.current_item)
				if k > 0 then
					i := i + 1
					Result.put (item (k), i)
				end
				a_entities.forth
			end
			a_entities.go_i_th (l_index)
			Result := Result.subarray (1,i)
		ensure
			subgroup_not_void: Result /= Void
			max_count: Result.count <= a_entities.count
		end

	vector_at (a_entities: ROSE_LINEAR_ARRAY [G]): VECTOR is
			-- Vector of values for entities in `a_entities'
		require
			a_entities_not_void: a_entities /= Void
		local
			l_index: INTEGER
		do
			create Result.make (a_entities.count)
			l_index := a_entities.index
			from
				a_entities.start
			until
				a_entities.after
			loop
				Result.put (value_at (a_entities.current_item), a_entities.index)
				a_entities.forth
			end
			a_entities.go_i_th (l_index)
		ensure
			vector_at_not_void: Result /= Void
			count: Result.count = a_entities.count
		end

feature -- Sorting

	sort_descending is
			-- Sort in descending order of values.
		do
			sort_by (agent compare_value)
		end

	sort_ascending is
			-- Sort in ascending order of values.
		do
			sort_by (agent compare_value_descending)
		end

	sort_abs_descending is
			-- sort in descending order of absolute values.	
		do
			sort_by (agent compare_abs_descending)
		end

feature {NONE} -- Implementation

	compare_value (a_pair, b_pair: PAIR [G,DOUBLE]): BOOLEAN is
			-- Is `a_pair's value greater than `b_pair's second?
		require
			not_void: a_pair /= Void and b_pair /= Void
		do
			Result := a_pair.second > b_pair.second
		ensure
			definition: Result = (a_pair.second > b_pair.second)
		end

	compare_value_descending (a_pair, b_pair: PAIR [G,DOUBLE]): BOOLEAN is
			-- Implementation of `sort_ascending'
		require
			not_void: a_pair /= Void and b_pair /= Void
		do
			Result := a_pair.second < b_pair.second
		ensure
			definition: Result = (a_pair.second < b_pair.second)
		end

	compare_abs_descending (a_pair, b_pair: PAIR [G,DOUBLE]): BOOLEAN is
			-- Implementation of `sort_abs_descending'
		require
			not_void: a_pair /= Void and b_pair /= Void
		do
			Result := a_pair.second.abs > b_pair.second.abs
		ensure
			definition: Result = (a_pair.second.abs > b_pair.second.abs)
		end

feature -- Basic operations

	infix "-" (a_other: ENTITY_VALUES[G]): ENTITY_VALUES [G] is
			-- Difference of `a_other' from current for each entity
		local
			i, j, k, l_index, l_other_index: INTEGER
			l_value: DOUBLE
			l_table: HASH_TABLE [INTEGER,G]
			l_pair: PAIR [G,DOUBLE]
			l_entity: G
		do
			if a_other /= Void and then a_other.count > 0 then
				create l_table.make (count)
				l_index := index
				from
					start
				until
					after
				loop
					l_table.put (index, current_entity)
					forth
				end
				k := count + a_other.count
				create Result.make (1, k)
				from
					i := 1
				until
					i > count
				loop
					Result.put (item (i), i)
					i := i + 1
				end
				l_other_index := a_other.index
				from
					a_other.start
				until
					a_other.after
				loop
					l_entity := a_other.current_entity
					l_table.search (l_entity)
					if l_table.found then
						i := l_table.found_item
						l_value := item (i).second - a_other.current_value
						create l_pair.make (l_entity, l_value)
						Result.put (l_pair, i)
					else
						j := j + 1
						l_value := 0.0 - a_other.current_value
						create l_pair.make (l_entity, l_value)
						Result.put (l_pair, count + j)
					end
					a_other.forth
				end
				Result := Result.subarray (1, count + j)
				index := l_index
				a_other.go_i_th (l_other_index)
			else
				Result := Current.twin
			end
		ensure
			infix_minus_not_void: Result /= Void
			max_count: Result.count <= count + a_other.count
		end

	infix "+" (a_other: ENTITY_VALUES [G]): ENTITY_VALUES [G] is
			-- Addition of `a_other' and `Current'
		local
			l_negative: like a_other
		do
			if a_other /= Void and then a_other.count > 0 then
				create l_negative.make_from_entities_vector (a_other.raw_vector.scaled (-1.0), a_other.entities)
				Result := Current - l_negative
			else
				Result := twin
			end
		ensure
			infix_plus_not_void: Result /= Void
			max_count: Result.count <= count + a_other.count
		end

feature -- Conversion

	to_entity_matrix: ENTITY_MATRIX [G] is
			-- Convert to a one column entity matrix.
		do
			create Result.make_all (entities, vector.to_column_matrix, <<name>>)
		ensure
			to_entity_matrix_not_void: Result /= Void
			one_column: Result.width = 1
			correct_height: Result.height = count
			name_set: Result.matrix /= Void and then Result.matrix.column_names /= Void and then
				Result.matrix.column_names.count = 1 and then Result.matrix.column_names [1] ~ name
		end

feature -- What is this application specific stuff doing here?

	date: ROSE_DATE
			-- Date

	set_date (a_date: like date) is
			-- Set `date' to `a_date'.
		do
			date := a_date
		ensure
			date_set: date = a_date
		end

	currency: PERSISTENT_CURRENCY
			-- Currency

	set_currency (a_currency: like currency) is
			-- Set `currency' to `a_currency'.
		do
			currency := a_currency
		ensure
			currency_set: currency = a_currency
		end

	zone: CHARACTER
			-- Zone mnemonic

	set_zone (a_zone: like zone) is
			-- Set `zone' to `a_zone'.
		do
			zone := a_zone
		ensure
			zone_set: zone = a_zone
		end

feature -- Query

	diff (a_other: like subset_by_value): like subset_by_value is
			-- `1.0' for add (i.e. in `Current' only), `-1.0' for drop (i.e. in `a_other' only),
			-- `0.0' for present on both sides, whether changed or not
		local
			l_emx: ENTITY_MATRIX [G]
			i: INTEGER
			l_vector: VECTOR
			l_entity: G
			l_entities: ROSE_LINEAR_ARRAY[ G]
		do
			if a_other /= Void then
				create Result.make_empty
				l_emx := merged (a_other)
				l_entities := l_emx.entities
				create l_vector.make (l_entities.count)
				from
					i := 1
				until
					i > l_entities.count
				loop
					l_entity := l_entities.item (i)
					if has_entity (l_entity) and not (a_other.has_entity (l_entity)) then
						l_vector.put (1.0, i)
					elseif a_other.has_entity( l_entity) and (not has_entity (l_entity)) then
						l_vector.put(-1.0, i)
					end
					i := i + 1
				end
				create Result.make_from_entities_vector (l_vector, l_entities)
			else
				create l_vector.make_identity (count)
				create Result.make_from_entities_vector (l_vector, entities)
			end
		ensure
			diff_not_void: Result /= Void
			add_current: Result.subset_by_value (1.0).entities.for_all (agent has_entity)
			add_other: a_other = Void or else Result.subset_by_value (1.0).entities.for_all (agent a_other.not_has_entity)
			drop_current: Result.subset_by_value (-1.0).entities.for_all (agent not_has_entity)
			drop_other: a_other = Void or else Result.subset_by_value (-1.0).entities.for_all (agent a_other.has_entity)
			stays_current: Result.subset_by_value (0.0).entities.for_all (agent has_entity)
			stays_other: a_other = Void or else Result.subset_by_value (0.0).entities.for_all (agent a_other.has_entity)
			count_complete: Result.subset_by_value (0.0).count + Result.subset_by_value (1.0).count = count
			count_complete_other: Result.subset_by_value (0.0).count + Result.subset_by_value (-1.0).count = a_other.count
		end

feature -- Conversion

	replace_vector (a_vector: VECTOR): like subset_by_value is
			-- New instance with values from `a_vector'
		require
			matching_count: a_vector /= Void implies a_vector.count = entities.count
		do
			create Result.make_from_entities_vector (a_vector, entities)
		ensure
			replace_vector_not_void: Result /= Void
			same_count: Result.count = count
			same_entities: Result.entities.same_items (entities)
			same_vector: a_vector /= Void implies a_vector.same_vector (Result.raw_vector)
		end

feature -- Output

	output: ARRAY2 [ANY] is
			-- Output to (name, value) ARRAY2
		local
			l_index: INTEGER
		do
			create Result.make (count + 1, 2)
			Result.put ("Name", 1, 1)
			Result.put ("Value", 1, 2)
			l_index := index
			from
				start
			until
				after
			loop
				Result.put (current_name, index + 1, 1)
				Result.put (current_value, index + 1, 2)
				forth
			end
			index := l_index
		ensure
			output_not_void: Result /= Void
			correct_width: Result.width = 2
			height_includes_column_headers: Result.height = count + 1
		end

invariant

	label_unused: label = Void

end -- class ENTITY_VALUES
