note

	description:

		"Represents enough information to construct a field in an EXTERNAL_RECORD"

	author: "Mark Howard"
	last_editor: "$Author: mb $"
	date: "$Date: 2014/02/19 $"
	revision: "$Revision: #14 $"
	path: "$File: //rose/source/kernel/infrastructure/structures/field_descriptors.e $"

class FIELD_DESCRIPTORS

inherit

	ROSE_CONSTANTS

feature -- Access

	remote_double_field (a_name: STRING; a_object: ANY): DOUBLE is
			-- remote double value by name
		require
			valid_name: a_name /= Void
			valid_object: a_object /= Void
		local
			l_ix: INTEGER
		do
			l_ix := remote_field_index (a_name, a_object)
			if l_ix > 0 then
				Result := internal.double_field (l_ix, a_object)
			end
		end

	remote_field_index (a_name: STRING; a_object: ANY): INTEGER is
			-- remote field index
		require
			valid_name: a_name /= Void
			valid_object: a_object /= Void
		local
			l_table: HASH_TABLE [INTEGER, STRING]
			l_type, i, l_count: INTEGER
		do
			l_type := internal.dynamic_type (a_object)
			l_table := descriptors.safe_item (l_type)
			if l_table = Void then
				l_count := internal.field_count (a_object)
				create l_table.make (l_count)
				from
					i := 1
				until
					i > l_count
				loop
					l_table.put (i, internal.field_name (i, a_object).as_upper)
					l_table.put (i, internal.field_name (i, a_object).as_lower)
					i := i + 1
				end
				descriptors.force (l_table, l_type)
			end
			Result := l_table.item (a_name)
		end

	field_index_by_name_table: HASH_TABLE [INTEGER, STRING] is
			-- field index by name table
		local
			l_integer: INTEGER
		do
			Result := descriptors.safe_item (internal.dynamic_type (Current))
			if Result = Void then
				l_integer := remote_field_index ("", Current)
				Result := descriptors.safe_item (internal.dynamic_type (Current))
			end
		ensure
			valid_result: Result /= Void
		end

	remote_field_type (a_name: STRING; a_object: ANY): INTEGER is
			-- remote field type
		require
			valid_name: a_name /= Void
			valid_object: a_object /= Void
		do
			Result := internal.field_type (remote_field_index (a_name, a_object), a_object)
		end

	remote_value_by_name (a_name: STRING; a_object: ANY): ANY is
			-- Value of 'a_name' in 'a_object'
		require
			valid_name: a_name /= Void
			valid_object: a_object /= Void
		do
			Result := internal.field (remote_field_index (a_name, a_object), a_object)
		end

	field_type (a_name: STRING): INTEGER is
			-- Type of 'a_name' in 'Current'
		require
			valid_name: a_name /= Void
		do
			Result := remote_field_type (a_name, Current)
		end

	field_index (a_name: STRING): INTEGER is
			-- Index of 'a_name' in 'Current'
		require
			valid_name: a_name /= Void
		do
			Result := remote_field_index (a_name, Current)
		end

	value_by_name (a_name: STRING): ANY is
			-- Value of 'a_name' in 'Current'
		require
			valid_name: a_name /= Void
			valid_field_indes: field_index (a_name) > 0
		do
			Result := internal.field (field_index (a_name), Current)
		end

	value_by_index (a_index: INTEGER): ANY is
			-- Value of attribute 'a_index' in 'Current'
		do
			Result := internal.field (a_index, Current)
		end

	integer_field (a_name: STRING): INTEGER is
			-- Integer value of 'a_name'
		require
			valid_name: a_name /= Void
			valid_field_indes: field_index (a_name) > 0
		do
			Result := internal.integer_field (field_index (a_name), Current)
		end

feature -- Element modification

	set_field (a_name: STRING; a_value: ANY) is
			-- Set the field `a_name' to `a_value'
		do
			remote_set_field (a_name, a_value, Current)
		end

	set_field_with_index (a_index: INTEGER; a_value: ANY) is
			-- Set the field at `a_index' to `a_value'
		do
			remote_set_field_with_index (a_index, a_value, Current)
		end

	set_integer_field (a_name: STRING; a_value: INTEGER) is
			-- Set the integer field called `a_name' to `a_value'
		do
			internal.set_integer_field (field_index (a_name), Current, a_value)
		end

feature {NONE} -- Implementation

	remote_set_field (a_name: STRING; a_value: ANY; a_object: ANY) is
			-- Set field `a_name' in `a_object' to `a_value'
		local
			l_index: INTEGER
		do
			l_index := remote_field_index (a_name, a_object)
			remote_set_field_with_index (l_index, a_value, a_object)
		end

	remote_set_field_with_index (a_index: INTEGER; a_value: ANY; a_object: ANY) is
			-- Set field at `a_index' in `a_object' to `a_value'
		do
			inspect
				internal.field_type (a_index, a_object)
			when {INTERNAL}.CHARACTER_TYPE then
				if attached {CHARACTER_REF} a_value as l_char then
					internal.set_character_field (a_index, a_object, l_char.item)
				elseif attached {STRING} a_value as l_string and then not l_string.is_empty then
					internal.set_character_field (a_index, a_object, l_string.item (1))
				else
					internal.set_character_field (a_index, a_object, '%U')
				end
			when {INTERNAL}.Boolean_type then
				if attached {BOOLEAN_REF} a_value as l_bool then
					internal.set_boolean_field (a_index, a_object, l_bool.item)
				else
					internal.set_boolean_field (a_index, a_object, True)
				end
			when {INTERNAL}.INTEGER_8_TYPE then
				if attached {INTEGER_8_REF} a_value as l_int8 then
					internal.set_integer_8_field (a_index, a_object, l_int8.item)
				else
					internal.set_integer_8_field (a_index, a_object, 0)
				end
			when {INTERNAL}.INTEGER_16_TYPE then
				if attached {INTEGER_16_REF} a_value as l_int16 then
					internal.set_integer_16_field (a_index, a_object, l_int16.item)
				else
					internal.set_integer_16_field (a_index, a_object, constants.Undefined_integer_16)
				end
			when {INTERNAL}.INTEGER_32_TYPE then
				if attached {INTEGER_REF} a_value as l_int then
					internal.set_integer_field (a_index, a_object, l_int.item)
				else
					internal.set_integer_field (a_index, a_object, constants.null_integer)
				end
			when {INTERNAL}.INTEGER_64_TYPE then
				if attached {INTEGER_64_REF} a_value as l_int64 then
					internal.set_integer_64_field (a_index, a_object, l_int64.item)
				else
					internal.set_integer_64_field (a_index, a_object, constants.missing_integer.to_integer_64)
				end
			when {INTERNAL}.REAL_TYPE then
				if attached {REAL_REF} a_value as l_real then
					internal.set_real_field (a_index, a_object, l_real.item)
				else
					internal.set_real_field (a_index, a_object, constants.missing_real)
				end
			when {INTERNAL}.DOUBLE_TYPE then
				if attached {DOUBLE_REF} a_value as l_double then
					internal.set_double_field (a_index, a_object, l_double.item)
				else
					internal.set_double_field (a_index, a_object, constants.missing_double)
				end
			when {INTERNAL}.REFERENCE_TYPE then
				if attached {STRING} a_value as l_string then
					internal.set_reference_field (a_index, a_object, l_string)
				elseif attached {ROSE_DATE} a_value as l_date then
					internal.set_reference_field (a_index, a_object, l_date)
				else
					internal.set_reference_field (a_index, a_object, a_value)
				end
			end
		end

	internal: INTERNAL is
			-- Reflection facilities
		once
			create Result
		end

	descriptors: ROSE_LINEAR_ARRAY [HASH_TABLE [INTEGER, STRING]] is
			-- Array of tables describing the field index of the dynamic type
		once
			create Result.make_empty
		end

end
