indexing

	description: "Allows to keep track of which files were accessed during program's execution"

	author: "Zoran Simic"
	last_editor: "$Author: an $"
	path: "$File: //rose/source/kernel/infrastructure/text/file_access_log.e $"
	date: "$Date: 2013/07/22 $"
	revision: "$Revision: #11 $"

class FILE_ACCESS_LOG

inherit

	ROSE_SHARED_GLOBAL

	KL_SHARED_STRING_EQUALITY_TESTER

	KL_SHARED_FILE_SYSTEM

feature -- Access

	report: STRING is
			-- Textual report on which files were accessed
		do
			Result := ""
			if file_in_access_log.is_empty and then file_out_access_log.is_empty then
				Result.append ("-- No Files Accessed --%N")
			else
				Result.append (access_report ("Input Files", file_in_access_log))
				Result.append (access_report ("Output Files", file_out_access_log))
			end
			if storable_in_access_log.is_empty and then storable_out_access_log.is_empty then
				Result.append ("-- No Storables Accessed --%N")
			else
				Result.append (storable_access_report ("Input Storables", storable_in_access_log))
				Result.append (storable_access_report ("Output Storables", storable_out_access_log))
			end
			if persistent_server_in_access_log.is_empty and then persistent_server_out_access_log.is_empty then
				Result.append ("-- No Persistent Servers Accessed --%N")
			else
				Result.append (access_report ("Input Persistent Servers", persistent_server_in_access_log))
				Result.append (access_report ("Output Persistent Servers", persistent_server_out_access_log))
			end
		ensure
			valid_result: Result /= Void
		end

	file_in_access_log: DS_HASH_TABLE [INTEGER, STRING] is
			-- Keeps track of files accessed in read mode
		once
			create Result.make_with_equality_testers (10, Void, string_equality_tester)
		ensure
			file_in_access_log_not_void: Result /= Void
			item_equality_tester_void: Result.equality_tester = Void
			key_equality_tester_set: Result.key_equality_tester = string_equality_tester
		end

	file_out_access_log: DS_HASH_TABLE [INTEGER, STRING] is
			-- Keeps track of files accessed in write mode
		once
			create Result.make_with_equality_testers (10, Void, string_equality_tester)
		ensure
			file_out_access_log_not_void: Result /= Void
			item_equality_tester_void: Result.equality_tester = Void
			key_equality_tester_set: Result.key_equality_tester = string_equality_tester
		end

	storable_in_access_log: DS_HASH_TABLE [ROSE_LINEAR_ARRAY [STRING], STRING] is
			-- Keeps track of storables read from files, keyed by generating_type
		once
			create Result.make_with_equality_testers (10, Void, string_equality_tester)
		ensure
			storable_in_access_log_not_void: Result /= Void
			item_equality_tester_void: Result.equality_tester = Void
			key_equality_tester_set: Result.key_equality_tester = string_equality_tester
		end

	storable_out_access_log: DS_HASH_TABLE [ROSE_LINEAR_ARRAY [STRING], STRING] is
			-- Keeps track of storables written to files, keyed by generating_type
		once
			create Result.make_with_equality_testers (10, Void, string_equality_tester)
		ensure
			storable_out_access_log_not_void: Result /= Void
			item_equality_tester_void: Result.equality_tester = Void
			key_equality_tester_set: Result.key_equality_tester = string_equality_tester
		end

	persistent_server_in_access_log: DS_HASH_TABLE [STRING, STRING] is
			-- Keeps track of persistent_servers read from files, keyed by generating_type
		once
			create Result.make_with_equality_testers (10, Void, string_equality_tester)
		ensure
			persistent_server_in_access_log_not_void: Result /= Void
			item_equality_tester_void: Result.equality_tester = Void
			key_equality_tester_set: Result.key_equality_tester = string_equality_tester
		end

	persistent_server_out_access_log: DS_HASH_TABLE [STRING, STRING] is
			-- Keeps track of persistent_servers written to files, keyed by generating_type
		once
			create Result.make_with_equality_testers (10, Void, string_equality_tester)
		ensure
			persistent_server_out_access_log_not_void: Result /= Void
			item_equality_tester_void: Result.equality_tester = Void
			key_equality_tester_set: Result.key_equality_tester = string_equality_tester
		end

feature -- Basic operations

	record_file_read_access (a_name: STRING) is
			-- Record that file with 'a_name' was accessed in read mode
		require
			valid_name: a_name /= Void
		local
			l_count: INTEGER
			l_dated_filename: STRING
		do
			l_dated_filename := a_name + " " + file_timestamp (a_name)
			if file_in_access_log.has (l_dated_filename) then
				l_count := file_in_access_log.item (l_dated_filename)
			end
			file_in_access_log.force (l_count + 1, l_dated_filename)
		end

	record_file_write_access (a_name: STRING) is
			-- Record that file with 'a_name' was accessed in write mode
		require
			valid_name: a_name /= Void
		local
			l_count: INTEGER
		do
			if file_out_access_log.has (a_name) then
				l_count := file_out_access_log.item (a_name)
			end
			file_out_access_log.force (l_count + 1, a_name)
		end

	record_storable_read_access (a_name, a_type: STRING) is
			-- Record that file with 'a_name' retrieved an object of 'a_type'
		require
			valid_name: a_name /= Void
			valid_type: a_type /= Void
		local
			l_dated_filename: STRING
		do
			l_dated_filename := a_name + " " + file_timestamp (a_name)
			log_access (storable_in_access_log, l_dated_filename, a_type)
		end

	record_storable_write_access (a_name, a_type: STRING) is
			-- Record that file with 'a_name' stored an object of 'a_type'
		require
			valid_name: a_name /= Void
			valid_type: a_type /= Void
		do
			log_access (storable_out_access_log, a_name, a_type)
		end

	record_persistent_server_read_access (a_logical, a_path: STRING) is
			-- Record that 'a_logical' was read from 'a_path'
		require
			valid_logical: a_logical /= Void
			valid_path: a_path /= Void
		do
			persistent_server_in_access_log.force (a_logical, a_path)
		end

	record_persistent_server_write_access (a_logical, a_path: STRING) is
			-- Record that 'a_logical' was written to from 'a_path'
		require
			valid_logical: a_logical /= Void
			valid_path: a_path /= Void
		do
			persistent_server_out_access_log.force (a_logical, a_path)
		end

feature {ROSE_RESOURCE_ACCESS} -- Implementation

	clear_all is
			-- Clear out any tracking that happened, to be used for resetting after metrics have been persisted
		do
			file_in_access_log.wipe_out
			file_out_access_log.wipe_out
			storable_in_access_log.wipe_out
			storable_out_access_log.wipe_out
			persistent_server_in_access_log.wipe_out
			persistent_server_out_access_log.wipe_out
		end

feature {NONE} -- Implementation

	log_access (a_log: DS_HASH_TABLE [ROSE_LINEAR_ARRAY [STRING], STRING]; a_name, a_type: STRING) is
			-- Log storable in/out for file 'a_name' of object type 'a_type'
		local
			l_log: ROSE_LINEAR_ARRAY [STRING]
		do
			a_log.search (a_type)
			if not a_log.found then
				create l_log.make_empty
				a_log.force (l_log, a_type)
			else
				l_log := a_log.found_item
			end
			if not l_log.has (a_name) then
				l_log.extend (a_name)
			end
		end

	access_report (a_report_name: STRING; a_log: DS_HASH_TABLE [ANY, STRING]): STRING is
			-- A file access report string for 'a_log'
		require
			valid_report: a_report_name /= Void
			valid_log: a_log /= Void
		do
			if a_log.is_empty then
				Result := "-- No " + a_report_name + " Accessed --"
			else
				Result := "-- " + a_report_name + " Accessed --%N"
				from
					a_log.start
				until
					a_log.off
				loop
					Result.append (a_log.key_for_iteration + "%N")
					a_log.forth
				end
			end
		ensure
			valid_result: Result /= Void
		end

	storable_access_report (a_report_name: STRING; a_log: DS_HASH_TABLE [ROSE_LINEAR_ARRAY [STRING], STRING]): STRING is
			-- A storable file access report string for 'a_log'
		require
			valid_report: a_report_name /= Void
			valid_log: a_log /= Void
		local
			l_log: ROSE_LINEAR_ARRAY [STRING]
		do
			if a_log.is_empty then
				Result := "-- No " + a_report_name + " Accessed --"
			else
				Result := "-- " + a_report_name + " Accessed --%N"
				from
					a_log.start
				until
					a_log.off
				loop
					l_log := a_log.item_for_iteration
					from l_log.start until l_log.off loop
						Result.append (a_log.key_for_iteration + ": " + l_log.current_item + "%N")
						l_log.forth
					end
					a_log.forth
				end
			end
		ensure
			valid_result: Result /= Void
		end

	file_timestamp (a_filename: STRING): STRING
			-- Timestamp of `a_filename'
		require
			a_filename_not_void: a_filename /= Void
		local
			l_timestamp: INTEGER
			l_date: ROSE_DATE
		do
			l_timestamp := file_system.file_time_stamp (a_filename)
			if l_timestamp < 0 then
				Result := "[00000000-00:00:00 UTC]"
			else
				create l_date.make_by_unix_time (l_timestamp)
				Result := "[" + l_date.formatted_out ("yyyy[0]mm[0]dd") + "-" +  l_date.formatted_out ("[0]hh:[0]mi:[0]ss") + " UTC]"
			end
		ensure
			file_timestamp_not_void: Result /= Void
		end

end













