indexing

	description: "MATRIX in which the rows and columns are hashed by keys"

	author: "Michael Zhang"
	last_editor: "$Author: mk $"
	date: "$Date: 2012/08/20 $"
	revision: "$Revision: #14 $"
	path: "$File: //rose/source/kernel/infrastructure/containers/hashed_matrix.e $"

class HASHED_MATRIX [G -> HASHABLE, H -> HASHABLE]

inherit

	MATRIX
		rename
			column_indices as matrix_column_indices,
			row_indices as matrix_row_indices
		redefine
			anchor,
			new_matrix_anchor
	end

create

	make_all,
	make,
	make_from_array,
	make_from_others,
	make_identity,
	make_from_array2,
	make_sub_matrix_3

create {HASHED_MATRIX}

	make_as_inverse

feature  {NONE} -- Initialization

	make_all (a_matrix: ABSTRACT_READABLE_MATRIX; a_row_keys: ARRAY [like row_key_template]; a_column_keys: ARRAY [like column_key_template])
			-- Make from `a_matrix' with `a_row_key' and `a_column_keys'
		require
			a_matrix_not_void: a_matrix /= Void
		local
			i, j: INTEGER
		do
			make (a_matrix.height, a_matrix.width)
			from
				i := 1
			until
				i > height
			loop
				from
					j := 1
				until
					j > width
				loop
					put (a_matrix.item (i, j), i, j)
					j := j + 1
				end
				i := i + 1
			end
			set_row_keys (a_row_keys)
			set_column_keys (a_column_keys)
		end

feature -- Settings

	set_row_keys (a_keys: ARRAY [like row_key_template])
			-- Set row keys to `a_keys'
		require
			keys_match_height: a_keys /= Void and then a_keys.count = height
		do
			create row_key_table.make_from_array (a_keys)
		ensure
			row_key_tabl_not_void: row_key_table /= Void
		end

	set_column_keys (a_keys: ARRAY [like column_key_template])
			-- Set column keys to `a_keys'
		require
			keys_match_width: a_keys /= Void and then a_keys.count = width
		do
			create column_key_table.make_from_array (a_keys)
		ensure
			column_key_tabl_not_void: column_key_table /= Void
		end

feature -- Access

	row_ix (a_key: like row_key_template): INTEGER
			-- Row index corresponding to `a_key'
		do
			if a_key /= Void then
				Result := row_key_table.new_index_of (a_key)
			end
		end

	column_ix (a_key: like column_key_template): INTEGER
			-- Column index corresponding to `a_key'
		do
			if a_key /= Void then
				Result := column_key_table.new_index_of (a_key)
			end
		end

	row_indices (a_row_keys : ARRAY [like row_key_template]): ROSE_LINEAR_ARRAY[INTEGER]
			-- Row indices corresponding to `a_row_keys'
		do
			Result := row_key_table.indices_of (a_row_keys)
		end

	column_indices (a_column_keys : ARRAY [like column_key_template]): ROSE_LINEAR_ARRAY[INTEGER]
			-- Column indices corresponding to `a_column_keys'
		do
			Result := column_key_table.indices_of (a_column_keys)
		end

	row_at (a_key: like row_key_template): VECTOR
			-- Matrix row corresponding to `a_key'
		local
			i: INTEGER
		do
			i := row_ix( a_key)
			if valid_row_index (i) then
				Result := row (i)
			end
		end

	column_at (a_key : like column_key_template): VECTOR
			-- Matrix column corresponding to `a_key'
		local
			j: INTEGER
		do
			j := column_ix (a_key)
			if valid_column_index (j) then
				Result := column (j)
			end
		end

	item_at (a_row_key: like row_key_template; a_column_key: like column_key_template): DOUBLE
			-- Matrix item corresponding to `a_row_key' and `a_column_key'
		local
			i, j: INTEGER
		do
			i := row_ix (a_row_key)
			j := column_ix (a_column_key)
			if valid_row_index (i) and valid_column_index (j) then
				Result := item (i, j)
			end
		end

	submatrix_at (a_row_keys: ARRAY [like row_key_template]; a_column_keys: ARRAY [like column_key_template]): HASHED_MATRIX [G,H]
			-- Submatrix selected using `a_row_keys' and `a_column_keys'
		require
			requested_keys_non_void: a_row_keys /= Void and then a_column_keys /= Void
		local
			l_row_ixs, l_col_ixs: ROSE_LINEAR_ARRAY [INTEGER]
		do
			l_row_ixs := row_indices (a_row_keys)
			l_col_ixs := column_indices (a_column_keys)
			create Result.make_sub_matrix_3 (Current, l_row_ixs, l_col_ixs)
			Result.set_row_keys (a_row_keys)
			Result.set_column_keys (a_column_keys)
		end

	selected_rows_by_keys  (a_row_keys: ARRAY [like row_key_template]): like Current
			-- Submatrix of rows selected using `a_row_keys' with zero vector where no match is found
		require
			requested_keys_non_void: a_row_keys /= Void
		local
			l_matrix: MATRIX
			i: INTEGER
		do
			create l_matrix.make (a_row_keys.count, width)
			from
				i := 1
			until
				i > l_matrix.height
			loop
				if row_at (a_row_keys.item (i)) /= Void then
					l_matrix.put_row_vector (row_at (a_row_keys.item (i)), i)
				end
				i := i + 1
			end
			create Result.make_all (l_matrix, a_row_keys, column_key_table)
		end

	selected_columns_by_keys  (a_column_keys: ARRAY [like column_key_template]): like Current
			-- Submatrix of columns selected using `a_column_keys' with zero vector where no match is found
		require
			requested_keys_non_void: a_column_keys /= Void
		local
			l_matrix: MATRIX
			j: INTEGER
		do
			create l_matrix.make (height, a_column_keys.count)
			from
				j := 1
			until
				j > l_matrix.width
			loop
				l_matrix.put_column_vector (column_at (a_column_keys.item (j)), j)
				j := j + 1
			end
			create Result.make_all (l_matrix, row_key_table, a_column_keys)
		end


feature  -- Status

	has_row (a_row_key: like row_key_template): BOOLEAN
			-- Is there a row corresponding to `a_row_key'?
		do
			Result := a_row_key /= Void and then row_key_table /= Void and then row_key_table.has_value(a_row_key)
		end

	has_column (a_column_key: like column_key_template): BOOLEAN
			-- Is there a column corresponding to `a_column_key'?	
		do
			Result := a_column_key /= Void and then column_key_table /= Void and then column_key_table.has_value(a_column_key)
		end

	has_keys (a_row_key: like row_key_template; a_column_key: like column_key_template): BOOLEAN
			-- Is there a row corresponding to `a_row_key' and a column corresponding to `a_column_key'?
		do
			Result := has_row (a_row_key) and has_column (a_column_key)
		end

feature  -- Keys

	row_key (a_ix: INTEGER): G
			-- Row key corresponding to index `a_ix'
		do
			if row_key_table /= Void and then row_key_table.valid_index(a_ix) then
				Result := row_key_table.item (a_ix)
			end
		end

	column_key (a_ix : INTEGER): H
			-- Column key corresponding to index `a_ix'
		do
			if column_key_table /= Void and then column_key_table.valid_index(a_ix) then
				Result := column_key_table.item (a_ix)
			end
		end

feature -- Tables

	row_key_table: ROSE_HASHED_LINEAR_ARRAY [G]
			-- Table of row keys

	column_key_table: ROSE_HASHED_LINEAR_ARRAY [H]
			-- Table of column keys

feature  -- Templates

	row_key_template: G
			-- Template for row keys
		do
		end

	column_key_template: H
			-- Template for column keys
		do
		end

	anchor: like Current
		-- Anchor
		do
		end

	new_matrix_anchor: like Current
		-- anchor for newly created matrices
		do
		end
end
