note
	description:
		"Object contains price header information (first and last, adjustment)"
	author: "Randy John"
	last_editor: "$Author: rj $"
	date: "$Date: 2013/11/22 $"
	revision: "$Revision: #79 $"
	path: "$File: //rose/source/kernel/data/market/hdp.e $"

class
	HDP

inherit
	MEMORY
	DATABASE_SERVER
	ROSE_ACCESS
	ROSE_SHARED_GLOBAL
	EXCEPTIONS
	ATTRIBUTES
		rename
			sedol as attributes_sedol,
			cusip as attributes_cusip,
			msci as attributes_msci,
			company_name as attributes_company_name
		end

create
	retrieve,
	load_from_vms,
	update_from_vms

feature {MARKET_DATA} -- Creation

	retrieve (a_zone: ZONE[LOCATION]) is
			-- retrieve
		do
			zone_mnemonic := a_zone.mnemonic
			market_data_server.set_implicit_transaction_off
			create_arrays
			load
			if julian_index_count < 100 and not (a_zone.mnemonic.is_equal('F') or a_zone.mnemonic.is_equal('I')) then
				raise("Size of HDP is too small. Please investigate!!!")
			end
		end

feature -- Creation

	update_from_vms (a_zone: ZONE[LOCATION]; a_do_iso_by_rxrf, a_do_julian_indexes, a_do_first_and_last, a_do_adjustments: BOOLEAN)
			-- Writes 4 tables in the market_data database using data from RMDAT_W_D:HDP (on VMS).
		do
			-- First, make sure that the transaction depth is 0 and then call 'set_implicit_transaction_on'.
			from
				-- Nothing here.
			until
				market_data_server.transaction_depth=0
			loop
				market_data_server.commit
			end

			market_data_server.set_implicit_transaction_on

			zone_mnemonic := a_zone.mnemonic
			create_arrays
			load			-- Loads our 4 arrays from market_data.  I'm not sure why since we are about to write to market_data.

			iso_by_rxrf_updated := 0
			julian_indexes_updated := 0
			first_and_last_updated := 0
			adjustments_updated := 0

			if a_do_iso_by_rxrf
			then
				-- Create the temp table.
				market_data_server.sql_execute_immediate("create table " + iso_by_rxrf_temp_table_name + "([zone] char(1), rxrf int, iso varchar(5), main_or_alt_iso int)")
				market_data_server.sql_execute_immediate("create unique index iso_inx on " + iso_by_rxrf_temp_table_name + "([zone],rxrf, main_or_alt_iso)")
				market_data_server.commit
				if not market_data_server.success
				then
					raise("Failed to commit iso_by_rxrf (1).")
				end

				-- Load the temp table.
				temp_iso_by_rxrf_fast_loader := market_data_fast_load_server.fast_loader("" + iso_by_rxrf_temp_table_name + "", create {ISO_BY_RXRF_RECORD})
				iso_by_rxrf_selector_rdb.apply(agent handle_iso_by_rxrf(?), Void, <<zone_mnemonic.out>>)
				temp_iso_by_rxrf_fast_loader.commit
				if not temp_iso_by_rxrf_fast_loader.success
				then
					raise("Failed to commit iso_by_rxrf (2).")
				end

				-- Using data in the temp table; update, insert and delete records in the iso_by_rxrf table.
				update_existing_iso_by_rxrf_records.execute (Void, Void)
				insert_new_iso_by_rxrf_records.execute (Void, Void)			-- You will get "Duplicate key was ignored" for the unchanged rows.
				delete_old_iso_by_rxrf_records.execute (Void, <<zone_mnemonic.out>>)

				-- Clean up.
				market_data_server.sql_execute_immediate("drop table " + iso_by_rxrf_temp_table_name + "")
				market_data_server.commit
				if not market_data_server.success
				then
					raise("Failed to commit iso_by_rxrf (3).")
				end
			end

			if a_do_julian_indexes
			then
				market_data_server.begin_transaction	-- The following may or may not start a transaction.
				jul_indexes_selector_rdb.apply(agent handle_julian_index(?),Void,<<zone_mnemonic.out>>)
				market_data_server.commit
				if not market_data_server.success
				then
					raise("Failed to commit julian_indexes.")
				end
			end

			if a_do_first_and_last
			then
				market_data_server.begin_transaction	-- The following may or may not start a transaction.
				first_and_last_selector_rdb.apply(agent handle_first_and_last(?),Void,<<zone_mnemonic.out>>)
				market_data_server.commit
				if not market_data_server.success
				then
					raise("Failed to commit first_and_last.")
				end
			end

			if a_do_adjustments
			then
				clear_adjustments.execute (Void,<<zone_mnemonic.out>>)
				market_data_server.commit	-- If we don't commit then the fast_server will be blocked.
				if not market_data_server.success
				then
					raise("Failed to commit adjustments.")
				end

				adjustments_to_commit := False
				adjustments_selector_rdb.apply(agent handle_adjustments(?),Void,<<zone_mnemonic.out>>)
				if adjustments_to_commit then
					adjustments_fast_inserter.commit
					if not adjustments_fast_inserter.success
					then
						raise("Failed to commit adjustments (fast load).")
					end
				end
			end

			if market_data_server.transaction_depth/=0
			then
				raise("Some data was left uncommitted.")
			end

			market_data_server.set_implicit_transaction_off
		end

	load_from_vms (a_zone: ZONE[LOCATION]) is
		do
			zone_mnemonic := a_zone.mnemonic
			market_data_server.set_implicit_transaction_on
			create_arrays
			-- Clear Zonal Data from the tables
			clear_adjustments.execute (Void,<<zone_mnemonic.out>>)
			clear_first_and_last.execute (Void,<<zone_mnemonic.out>>)
			clear_jul_index.execute (Void,<<zone_mnemonic.out>>)
			-- Reload from RDB
			jul_indexes_selector_rdb.apply(agent handle_julian_index(?),Void,<<zone_mnemonic.out>>)
			adjustments_selector_rdb.apply(agent handle_adjustments(?),Void,<<zone_mnemonic.out>>)
			if adjustments_to_commit then
				adjustments_fast_inserter.commit
			end
			first_and_last_selector_rdb.apply(agent handle_first_and_last(?),Void,<<zone_mnemonic.out>>)
			market_data_server.commit
			market_data_server.set_implicit_transaction_off
		end

feature {NONE} -- Implementation

	load
			-- Load the 4 arrays from the market_data database.
		local
			l_country_currency_history: HASH_TABLE[ROSE_ENTITY_HISTORY[CURRENCY], STRING]
		do
			jul_indexes_selector.retrieve (Void,<<zone_mnemonic.out>>)
			from jul_indexes_selector.array.start until jul_indexes_selector.array.after loop
				add_julian_index (jul_indexes_selector.array.current_item)
				jul_indexes_selector.array.forth
			end

			adjustments_selector.retrieve (Void,<<zone_mnemonic.out>>)
			from adjustments_selector.array.start until adjustments_selector.array.after loop
				add_adjustment (adjustments_selector.array.current_item)
				adjustments_selector.array.forth
			end

			normalize_adjustments

			-- This is the new code that substitutes for the old code below.
			-- We have already loaded the jul_index table.  We will now load the min/max adjustment dates
			-- and manually change the coded date indexes into real dates.  The old way of
			-- doing that did it automatically but it was unreliable.  Randy.
			adjustment_bounds_selector_2.retrieve (Void,<<zone_mnemonic.out>>)
			from adjustment_bounds_selector_2.array.start until adjustment_bounds_selector_2.array.after loop
				add_adjustment_bounds_2(adjustment_bounds_selector_2.array.current_item)
				adjustment_bounds_selector_2.array.forth
			end

			-- The following predicate has created tremendous SQL/Server turmoil.
			-- What should take a few seconds sometimes takes 20-60 minutes.
			-- We have tried everything we can think of and have given up.
			-- Randy
			--			adjustment_bounds_selector.retrieve (Void,<<zone_mnemonic.out>>)
			--			from adjustment_bounds_selector.array.start until adjustment_bounds_selector.array.after loop
			--				add_adjustment_bounds (adjustment_bounds_selector.array.current_item)
			--				adjustment_bounds_selector.array.forth
			--			end


			first_and_last_selector.retrieve (Void,<<zone_mnemonic.out>>)
			from first_and_last_selector.array.start until first_and_last_selector.array.after loop
				add_first_and_last (first_and_last_selector.array.current_item)
				first_and_last_selector.array.forth
			end

			create l_country_currency_history.make(2000)
			iso_by_rxrf_selector.retrieve (Void,<<zone_mnemonic.out>>)
			from iso_by_rxrf_selector.array.start until iso_by_rxrf_selector.array.after loop
				add_iso_by_rxrf (iso_by_rxrf_selector.array.current_item, l_country_currency_history)
				iso_by_rxrf_selector.array.forth
			end

			set_countries_with_market_data
		end

	normalize_adjustments
	-- Although fairly rare, occasionally, the final 'backadj' is not 1.0.  Here we normalize
	-- all 'backadj' so that the final one is 1.0.
	local
		l_adjustment: RAW_ADJUSTMENT
		l_final_backadj: like {RAW_ADJUSTMENT}.backadj
		l_list: LINKED_LIST[RAW_ADJUSTMENT]
	do
		if adjustments /= Void
		then
			across adjustments as i_adjustments
			loop
				l_list := i_adjustments.item

				if l_list /= Void
				then
					-- It is fairly common for the last 'backadj' to be -999.
					from
						l_list.finish
					until
						l_list.off or else is_defined_double(l_list.item.backadj)
					loop
						l_list.back
					end

					if not l_list.off
					then
						l_final_backadj := l_list.item.backadj
						if l_final_backadj /= 1.0 and then l_final_backadj /= 0.0
						then
							from
								-- 'l_list' is already pointing at the last valid adjustment.
							until
								l_list.off
							loop
								l_adjustment := l_list.item
								if is_defined_double(l_adjustment.backadj)
								then
									l_adjustment.set_backadj(l_adjustment.backadj / l_final_backadj)
								else
									do_nothing	-- It would be very odd if we got to this line.
								end

								l_list.back
							end
						end
					else
						do_nothing	-- It would be very odd if we got to this line.
					end
				end
			end
		end
	end

	market_data_server: SQL_DATA_SERVER is
		once
			Result := servers.market_data_server
		end

	market_data_fast_load_server: DATA_SERVER is
		once
			Result := servers.market_data_fast_load_server
		end

	hdp_server: DATA_SERVER is
		once
			Result := servers.hdp_server
		end

feature -- Update Statistics

	iso_by_rxrf_updated: INTEGER
	julian_indexes_updated: INTEGER
	first_and_last_updated: INTEGER
	adjustments_updated: INTEGER

feature -- Access

	zone_mnemonic: CHARACTER
	adjustments: ARRAY[LINKED_LIST[RAW_ADJUSTMENT]]
	adjustment_bounds: ARRAY[ENTITY_ID_BASE]
	jul_indexes: ARRAY[INTEGER]
	max_rose_julian: INTEGER
	max_jul_index: INTEGER
	first_and_last: ARRAY[FIRST_AND_LAST]
	julian_days: ARRAY[INTEGER]

	julian_index_count: INTEGER
	last_rxrf: INTEGER

	currency (a_rxrf: INTEGER; a_fx_quote: INTEGER; a_date: ROSE_DATE): CURRENCY is
		local
			l_rxrf_currency_history: ROSE_LINEAR_ARRAY[ROSE_ENTITY_HISTORY[CURRENCY]]
			l_currency_history: ROSE_ENTITY_HISTORY[CURRENCY]
			l_date: ROSE_DATE
		do
			if currency_history.valid_index (a_rxrf) then
				l_rxrf_currency_history := currency_history.item (a_rxrf)
				if l_rxrf_currency_history /= Void and then l_rxrf_currency_history.valid_index (a_fx_quote) then
					l_currency_history := l_rxrf_currency_history.item (a_fx_quote)
					if l_currency_history /= Void then
						if a_date = Void then
							l_date := max_date
						else
							l_date := a_date
						end
						Result := l_currency_history.witness (l_date)
						if Result = Void then
							-- Yes, this is sorted.
							from
								l_currency_history.start
							until
								l_currency_history.off or Result/=Void
							loop
								if l_currency_history.current_item.first_date > l_date then
									-- This is the first currency that was valid after 'l_date'.
									Result := l_currency_history.current_item.entity
								end
								l_currency_history.forth
							end
							if Result = Void then
								-- This is not uncommon.  It happens whenever you have a legacy currency
								-- in a Euro-in country after 1999/01/01.
								Result := l_currency_history.last.entity
							end
						end
					end
				end
			end
		end

	currency_history_for_listing (a_listing: LISTING): ROSE_LINEAR_ARRAY[ROSE_ENTITY_HISTORY[CURRENCY]] is
			-- The RMDAT currency history for 'a_listing'
		require
			valid_listing: a_listing /= Void and then a_listing.is_valid
		do
			Result := currency_history.safe_item (a_listing.xrf_code)
		end

	has_euro (a_rxrf: INTEGER): BOOLEAN is
		do
			if actual_has_euro.valid_index (a_rxrf) then
				Result := actual_has_euro.item (a_rxrf)
			else
				Result := False
			end
		end

	country_has_market_data (a_country: COUNTRY): BOOLEAN is
		require
			country_not_void: a_country /= Void
		do
			Result := countries_with_market_data.has (a_country.handle)
		end

feature {UPDATE_OLD_FIRST_AND_LAST} -- Implementation

	actual_has_euro: ROSE_LINEAR_ARRAY[BOOLEAN]
	currency_history: ROSE_LINEAR_ARRAY[ROSE_LINEAR_ARRAY[ROSE_ENTITY_HISTORY[CURRENCY]]]
	countries_with_market_data: ROSE_LINEAR_ARRAY[INTEGER]
	adjustments_to_commit: BOOLEAN
	last_iso_rxrf :INTEGER

	create_arrays
			-- The HDP is essentially a bunch of arrays.  Here we build them with no data.
		local
			l_max_xrf: INTEGER
			l_date: ROSE_DATE
		do
			l_max_xrf := context.max_category_xrf (zone_mnemonic,ids.listing_category_id)
			if l_max_xrf > 0 then
				create adjustments.make(1,l_max_xrf + 100)
				create adjustment_bounds.make(1,l_max_xrf + 100)
				create first_and_last.make(1,l_max_xrf + 100)
				--				create currencies.make(1,l_max_xrf + 100)
				create actual_has_euro.make(1, l_max_xrf + 100)
				create currency_history.make(1, l_max_xrf + 100)
			else
				create adjustments.make_empty
				create adjustment_bounds.make_empty
				create first_and_last.make_empty
				create actual_has_euro.make_empty
				create currency_history.make_empty
			end

			create l_date.make_now
			l_date.add_n_days (10)
			create jul_indexes.make(1,l_date.rose_julian_day)
			create julian_days.make(1,l_date.rose_julian_day)
		end

	handle_julian_index (a_row: JUL_INDEX) is
			-- Process 'a_row' into the table.
		require
			valid_row: a_row /= Void
		do
			if a_row.rose_julian /= 0 and a_row.jul_index /= 0 then
				julian_indexes_updated := julian_indexes_updated + 1
				if jul_indexes.item (a_row.rose_julian) = 0 then
					jul_index_selector.retrieve (Void,<<zone_mnemonic.out,a_row.rose_julian,a_row.jul_index>>)
					if jul_index_selector.array.is_empty then
						jul_index_inserter.execute (Void,<<zone_mnemonic.out,a_row.rose_julian,a_row.jul_index>>)
						-- You may get a "Cannot insert duplicate key" here.  Yes, there are duplicates on VMS.
					end
				end
			end
		end

	adjustments_fast_inserter: DATA_SERVER_FAST_LOAD is
		local
			l_template: RAW_ADJUSTMENT
		once
			create l_template.default_create
			Result := market_data_fast_load_server.fast_loader("adjustments",l_template)
		end

	handle_adjustments (a_row: RAW_ADJUSTMENT) is
		local
			l_row: RAW_ADJUSTMENT
		do
			l_row := a_row.twin
			l_row.set_zone (zone_mnemonic.out.item(1))
			adjustments_fast_inserter.insert (l_row)
			adjustments_updated := adjustments_updated + 1
			adjustments_to_commit := True
		end

	handle_first_and_last (a_row: RAW_FIRST_AND_LAST) is
		do
			first_and_last_updated := first_and_last_updated + 1
			if first_and_last.upper < a_row.rxrf or
			   first_and_last.item (a_row.rxrf) = Void
			then
				first_and_last_inserter.execute (Void,<<zone_mnemonic.out,a_row.rxrf,a_row.first,a_row.last>>)
				if not first_and_last_inserter.success
				then
					raise("first_and_last_inserter failed")
				end
			elseif first_and_last.item (a_row.rxrf).first /= a_row.first or
				   first_and_last.item (a_row.rxrf).last /= a_row.last
			then
				first_and_last_updater.execute (Void,<<a_row.first,a_row.last,zone_mnemonic.out,a_row.rxrf>>)
				if not first_and_last_updater.success
				then
					raise("first_and_last_updater failed")
				end
			end
		end

	handle_iso_by_rxrf (a_row: ISO_BY_RXRF_RECORD) is
		local
			l_currency: CURRENCY
		do
			--if last_iso_rxrf /= a_row.rxrf then
			--	iso_by_rxrf_clear.execute (Void,<<a_row.zone,a_row.rxrf>>)
			--	last_iso_rxrf := a_row.rxrf
			--end
			l_currency := country_currency (a_row.iso)
			if l_currency = Void and
				not a_row.iso.is_equal("%U%U") and
				not a_row.iso.is_equal("XX") and
				not a_row.iso.is_equal("%%%%") and
				not a_row.iso.is_equal("  ") then
				raise("Invalid Currency ISO")
			end
			temp_iso_by_rxrf_fast_loader.insert (a_row.twin)
			iso_by_rxrf_updated := iso_by_rxrf_updated + 1
			--iso_by_rxrf_inserter.execute (Void,<<a_row.iso,a_row.zone,a_row.rxrf,a_row.main_or_alt_iso>>)
		end

	max_date: ROSE_DATE is
		once
			create Result.make_max_date
		end

	add_julian_index (a_item: JUL_INDEX) is
		do
			max_jul_index := max_jul_index.max (a_item.jul_index)
			max_rose_julian := max_jul_index.max (a_item.rose_julian)
			jul_indexes.force (a_item.jul_index,a_item.rose_julian)
			julian_days.force (a_item.rose_julian,a_item.jul_index)
			julian_index_count := julian_index_count + 1
		end

	add_adjustment (a_item: RAW_ADJUSTMENT) is
			-- Add one adjustment into the 'adjustments' table.
		local
			l_rxrf: INTEGER
			l_adjustments: LINKED_LIST[RAW_ADJUSTMENT]
		do
			l_rxrf := a_item.rxrf
			if l_rxrf /= last_rxrf then
				create l_adjustments.make
				adjustments.force (l_adjustments,l_rxrf)
			end
			adjustments.item (l_rxrf).extend (a_item) --clone (a_item))
			last_rxrf := a_item.rxrf
		end

	add_adjustment_bounds (a_item: ENTITY_ID_BASE) is
		do
			adjustment_bounds.force (a_item, a_item.entity_id)
		end

	add_adjustment_bounds_2(a_raw_adjustment: RAW_ADJUSTMENT) is
		local
			l_item: ENTITY_ID_BASE
			l_date: ROSE_DATE
			l_jd: INTEGER
		do
			create l_item
			l_item.set_entity_id (a_raw_adjustment.rxrf)
			if julian_days.valid_index (a_raw_adjustment.lowdate_index)
			then
				l_jd := julian_days.item (a_raw_adjustment.lowdate_index)
				if l_jd<=0
				then
					l_jd := 1
				end
			else
				l_jd := 1
			end

			create l_date.make_from_rose_julian_day (julian_days.item (a_raw_adjustment.lowdate_index))
			l_item.set_start_date (l_date)

			if julian_days.valid_index (a_raw_adjustment.highdate_index)
			then
				l_jd := julian_days.item (a_raw_adjustment.highdate_index)
				if l_jd<=0
				then
					l_jd := max_rose_julian
				end
			else
				l_jd := max_rose_julian
			end

			create l_date.make_from_rose_julian_day (l_jd)
			l_item.set_end_date (l_date)

			adjustment_bounds.force (l_item, l_item.entity_id)
		end

	add_first_and_last (a_item: HDP_FIRST_AND_LAST) is
		local
			l_first_and_last: FIRST_AND_LAST
		do
			create l_first_and_last.make (a_item.first,a_item.last)
			first_and_last.force (l_first_and_last,a_item.rxrf)
		end

	add_iso_by_rxrf (a_item: ISO_BY_RXRF_RECORD;
					 a_country_currency_history: HASH_TABLE[ROSE_ENTITY_HISTORY[CURRENCY], STRING]) is
		local
			l_country: COUNTRY
--			l_currencies: ROSE_LINEAR_ARRAY[CURRENCY]
			l_currency: CURRENCY
--			l_count: INTEGER
			l_currency_history: ROSE_ENTITY_HISTORY[CURRENCY]
			l_currency_history_tmp: ROSE_ENTITY_HISTORY[CURRENCY]
			l_rxrf_currency_history: ROSE_LINEAR_ARRAY[ROSE_ENTITY_HISTORY[CURRENCY]]
		do
			if a_item.iso /= Void then --and then not a_item.iso.is_equal("%%%%") then
----				if not a_item.iso.is_equal("%%%%") then
----					l_country := current_country (iso,a_item.iso)
----				end
--
--				if currencies.valid_index (a_item.rxrf) then
--					l_currencies := currencies.item (a_item.rxrf)
--				end
--
--				if l_currencies = Void then
--					create l_currencies.make_empty
--					l_currencies.compare_objects
--					currencies.force (l_currencies,a_item.rxrf)
--				end
--
--				l_currency := country_currency (a_item.iso)
--				l_currencies.force (l_currency,a_item.main_or_alt_iso + 1)

				if a_item.iso.is_equal("%%%%")
				then
					if context.name.is_equal("ROSE ENTITY FOREGROUND") or
					   context.name.is_equal("RESEARCH")
					then
						l_currency := context.currency_by_iso_code("  ", now)	-- Use the unknown currency.
						create l_currency_history.make_empty
						l_currency_history.extend (create {ENTITY_INTERVAL[CURRENCY]}.make (l_currency, create {ROSE_DATE}.make_min_date, create {ROSE_DATE}.make_max_date, 0))
					end
				elseif a_item.iso.is_equal (European_Union_ISO)
				then
					actual_has_euro.force (True, a_item.rxrf)
					l_currency := euro
					create l_currency_history.make_empty
					l_currency_history.extend (create {ENTITY_INTERVAL[CURRENCY]}.make (l_currency, create {ROSE_DATE}.make_min_date, create {ROSE_DATE}.make_max_date, 0))
				else
					a_country_currency_history.search (a_item.iso)
					if a_country_currency_history.found
					then
						l_currency_history := a_country_currency_history.found_item
					else
						l_country := context.country_at (iso_code (a_item.iso), now)
						if l_country/=Void
						then
							create l_currency_history.make_empty
							l_currency_history_tmp := l_country.country_to_currency_history
							if l_currency_history_tmp/=Void
							then
								from
									l_currency_history_tmp.start
								until
									l_currency_history_tmp.off
								loop
									l_currency := l_currency_history_tmp.current_item.entity
									if equal (l_currency.iso.substring(1, 2), a_item.iso)
									then
										l_currency_history.extend (l_currency_history_tmp.current_item)
									end

									l_currency_history_tmp.forth
								end

								l_currency_history.sort_by_start_date
							end
						end

						a_country_currency_history.put (l_currency_history, a_item.iso)
					end
				end

				if l_currency_history/=Void
				then
					l_rxrf_currency_history := currency_history.safe_item (a_item.rxrf)
					if l_rxrf_currency_history=Void
					then
						create l_rxrf_currency_history.make_empty
						currency_history.force (l_rxrf_currency_history, a_item.rxrf)
					end

					l_rxrf_currency_history.force (l_currency_history, a_item.main_or_alt_iso + 1)
				end
			end
		end

	country_currency (a_country_iso: STRING): CURRENCY is
		local
			l_country: COUNTRY
			l_count: INTEGER
		do
			if a_country_iso.is_equal("%%%%") then
				if context.name.is_equal("ROSE ENTITY FOREGROUND") or
				   context.name.is_equal("RESEARCH")
				then
					Result := context.currency_by_iso_code("  ",now)	-- Use the unknown currency.
				end
			elseif a_country_iso.is_equal (European_Union_ISO) then
				Result := euro
			else
				l_country := context.country_at (iso_code (a_country_iso),now)
				if l_country /= Void then
					if l_country.is_member_of_european_union then
						l_count := l_country.currency_history.count
						if l_count > 1 then
							Result := l_country.currency_history.item (l_count-1).entity
						else
							Result := l_country.currency_history.item(1).entity
						end
					else
						Result := l_country.country_to_currency_history.witness (now)
					end
				end
			end
		end

	set_countries_with_market_data is
		local
			l_iso: ROSE_LINEAR_ARRAY[STRING]
			l_country: COUNTRY
		do
			create countries_with_market_data.make_empty
			countries_with_market_data_predicate.retrieve (Void,<<zone_mnemonic.out>>)
			l_iso := countries_with_market_data_predicate.array
			from l_iso.start until l_iso.off loop
				if l_iso.current_item /= Void then
					l_country := context.country_at (iso_code (l_iso.current_item ),Void)
					if l_country /= Void and l_country.is_valid then
						countries_with_market_data.extend (l_country.handle)
					end
				end
				l_iso.forth
			end
		end

	now: ROSE_DATE is
		once
			create Result.make_now
		end
	European_Union_ISO: STRING is "EU"
	Euro_ISO: STRING is "EUR"
		-- Euro ISO
	euro: CURRENCY is
		once
			Result := context.currency_by_iso_code (Euro_ISO,now)
		end

feature {NONE} -- Predicates

	countries_with_market_data_predicate: STRING_PREDICATE is
		require
			server_not_void: market_data_server /= Void
		once
			create Result.make (market_data_server, "select distinct (iso) from iso_by_rxrf where zone = ?")
		end

	jul_index_selector: COLLECTING_PREDICATE[JUL_INDEX] is
		require
			server_not_void: market_data_server /= Void
		once
			create Result.make (market_data_server,"select * from jul_index where zone = ? and rose_julian = ? and jul_index = ?")
		end

	jul_indexes_selector: COLLECTING_PREDICATE[JUL_INDEX] is
		require
			server_not_void: market_data_server /= Void
		once
			create Result.make (market_data_server, "select rose_julian,jul_index from jul_index where zone = ?")
		end

	adjustments_selector: COLLECTING_PREDICATE[RAW_ADJUSTMENT] is
		require
			server_not_void: market_data_server /= Void
		once
			create Result.make (market_data_server,"select rxrf,lowdate_index,highdate_index,backadj,split_factor %
												  % from adjustments where zone = ? order by rxrf,lowdate_index")
		end

	adjustment_bounds_selector: COLLECTING_PREDICATE[ENTITY_ID_BASE] is
		require
			server_not_void: market_data_server /= Void
		once
			create Result.make (market_data_server, "select rxrf as entity_id, min (lowdate_date) as start_date, max (highdate_date) as end_date from adjustments_dates %
												   %where zone = ? and backadj > 0 group by rxrf order by rxrf, start_date")
		end

	adjustment_bounds_selector_2: COLLECTING_PREDICATE[RAW_ADJUSTMENT] is
		require
			server_not_void: market_data_server /= Void
		once
			create Result.make (market_data_server, "select rxrf, min (lowdate_index) as lowdate_index, max (highdate_index) as highdate_index from adjustments %
												   %where zone = ? and backadj > 0 group by rxrf order by rxrf, lowdate_index")
		end

	first_and_last_selector: COLLECTING_PREDICATE[HDP_FIRST_AND_LAST] is
		require
			server_not_void: market_data_server /= Void
		once
			create Result.make (market_data_server,"select rxrf,first,last from first_and_last where zone= ? " +
							   " order by rxrf")
		end

	iso_by_rxrf_selector: COLLECTING_PREDICATE[ISO_BY_RXRF_RECORD] is
		require
			server_not_void: market_data_server /= Void
		once
			create Result.make (market_data_server,"select * from iso_by_rxrf where zone = ? order by rxrf")
		end

feature {NONE} -- Insertion Predicates

	jul_index_inserter: EXECUTION_PREDICATE is
		require
			server_not_void: market_data_server /= Void
		once
			create Result.make (market_data_server,"insert jul_index (zone,rose_julian,jul_index) values(?,?,?)")
		end

	adjustments_inserter: EXECUTION_PREDICATE is
		require
			server_not_void: market_data_server /= Void
		once
			create Result.make (market_data_server,"insert adjustments (zone,rxrf,lowdate_index,highdate_index,%
												  %backadj,split_factor) values(?,?,?,?,?,?)")
		end

	first_and_last_inserter: EXECUTION_PREDICATE is
		require
			server_not_void: market_data_server /= Void
		once
			create Result.make (market_data_server,"insert first_and_last (zone,rxrf,first,last) values(?,?,?,?)")
			if not Result.success
			then
				raise("Failed to create first_and_last_inserter")
			end
		end

	first_and_last_updater: EXECUTION_PREDICATE is
		require
			server_not_void: market_data_server /= Void
		once
			create Result.make (market_data_server,"update first_and_last set first = ?, last = ? where zone = ? and rxrf = ?")
			if not Result.success
			then
				raise("Failed to create first_and_last_updater")
			end
		end

	iso_by_rxrf_updater: EXECUTION_PREDICATE is
		require
			server_not_void: market_data_server /= Void
		once
			create Result.make (market_data_server,"update iso_by_rxrf set iso = ? where zone = ? and rxrf = ? and main_or_alt_iso = ?")
		end

	iso_by_rxrf_inserter: EXECUTION_PREDICATE is
		require
			server_not_void: market_data_server /= Void
		once
			create Result.make (market_data_server,"insert iso_by_rxrf (iso,zone,rxrf,main_or_alt_iso) values(?,?,?,?)")
		end

	temp_iso_by_rxrf_fast_loader: DATA_SERVER_FAST_LOAD

	iso_by_rxrf_temp_table_name: STRING is
			-- Table name for the temp table
		once
			Result := "##iso_by_rxrf_temp" + Global.app.pid.out
		end

	update_existing_iso_by_rxrf_records: EXECUTION_PREDICATE
			-- For each row in iso_by_rxrf update with data from the temp table if it is different.
		require
			server_not_void: market_data_server /= Void
		once
			create Result.make (market_data_server,
				"update iso_by_rxrf set iso = " + iso_by_rxrf_temp_table_name + ".iso, [timestamp] = getdate() from " + iso_by_rxrf_temp_table_name + " " +
				" where iso_by_rxrf.rxrf = " + iso_by_rxrf_temp_table_name + ".rxrf and " +
				" iso_by_rxrf.zone = " + iso_by_rxrf_temp_table_name + ".zone and " +
				" iso_by_rxrf.main_or_alt_iso = " + iso_by_rxrf_temp_table_name + ".main_or_alt_iso and " +
				" iso_by_rxrf.iso <> " + iso_by_rxrf_temp_table_name + ".iso")
		end

	insert_new_iso_by_rxrf_records: EXECUTION_PREDICATE
			-- Insert all rows from the temp table into iso_by_rxrf.  Will cause "Duplicate key was ignored".
		require
			server_not_void: market_data_server /= Void
		once
			create Result.make (market_data_server,
				"insert into iso_by_rxrf select *, getdate() from " + iso_by_rxrf_temp_table_name + "")
		end

	delete_old_iso_by_rxrf_records: EXECUTION_PREDICATE
			-- Removes rows from iso_by_rxrf which are not found in the temp table.
		require
			server_not_void: market_data_server /= Void
		once
			create Result.make (market_data_server,
				"delete from iso_by_rxrf where zone = ? and not exists " +
				"(select * from " + iso_by_rxrf_temp_table_name + " where " +
				" " + iso_by_rxrf_temp_table_name + ".rxrf = iso_by_rxrf.rxrf and " +
				" " + iso_by_rxrf_temp_table_name + ".zone = iso_by_rxrf.zone and " +
				" " + iso_by_rxrf_temp_table_name + ".main_or_alt_iso = iso_by_rxrf.main_or_alt_iso)")
		end

	iso_by_rxrf_clear: EXECUTION_PREDICATE is
		require
			server_not_void: market_data_server /= Void
		once
			create Result.make (market_data_server,"delete from iso_by_rxrf where zone = ? and rxrf = ?")
		end

feature {NONE} -- Clean Up Predicates

	clear_first_and_last: EXECUTION_PREDICATE is
		require
			server_not_void: market_data_server /= Void
		once
			create Result.make (market_data_server,"delete from first_and_last where zone = ?")
		end

	clear_adjustments: EXECUTION_PREDICATE is
		require
			server_not_void: market_data_server /= Void
		once
			create Result.make (market_data_server,"delete from adjustments where zone = ?")
		end

	clear_jul_index: EXECUTION_PREDICATE is

		require
			server_not_void: market_data_server /= Void
		once
			create Result.make (market_data_server,"delete from jul_index where zone = ?")
		end

feature {NONE} -- RDB Predicates

	jul_indexes_selector_rdb: SIMPLE_PREDICATE[JUL_INDEX] is
		require
			server_not_void: hdp_server /= Void
		once
			create Result.make (hdp_server, "select rose_julian,jul_index from jul_index where zone = ?")
		end

	adjustments_selector_rdb: SIMPLE_PREDICATE[RAW_ADJUSTMENT] is
			-- Adjustments selector
		require
			server_not_void: hdp_server /= Void
		once
			create Result.make (hdp_server,"select rxrf,lowdate_index,highdate_index,backadj,split_factor %
										  % from adjustments where zone = ? order by rxrf,lowdate_index asc ")
		end

	first_and_last_selector_rdb: SIMPLE_PREDICATE[RAW_FIRST_AND_LAST] is
			-- first and last predicate
		require
			server_not_void: hdp_server /= Void
		once
			create Result.make (hdp_server,"select rxrf,first,last from first_and_last where zone= ? order by rxrf asc")
		end

	iso_by_rxrf_selector_rdb: SIMPLE_PREDICATE[ISO_BY_RXRF_RECORD] is
			-- Predicate for iso by rxrf
		require
			server_not_void: hdp_server /= Void
		once
			create Result.make (hdp_server,"select * from iso_by_rxrf where zone = ? order by rxrf asc")
		end

invariant

	available_adjustments: adjustments /= Void
	available_adjustments_bounds: adjustment_bounds /= Void
	available_jul_indexes: jul_indexes /= Void
	available_first_and_last: first_and_last /= Void
	available_julian_days: julian_days /= Void
	available_currency_history: currency_history /= Void

end

















