indexing

	description: "IEEE 754 floating point support"
	author: "Randy John"
	last_editor: "$Author: rj $"
	date: "$Date: 2012/10/01 $"
	revision: "$Revision: #13 $"
	path: "$File: //rose/source/kernel/infrastructure/math/ieee_math_base.e $"

-- This class is used by IEEE_MATH.

-- Technical information can be found in the following :
--
--	Intel Architecture Software Developer's Manual Vol. 1 and 2
--	http://www.intel.com/design/pentiumii/manuals/
--
--	Alpha Architecture Handbook
--	http://ftp.digital.com/pub/Digital/info/semiconductor/literature/dsc-library.html

deferred class	IEEE_MATH_BASE

feature -- Basic operations

	trap_exceptions is
			-- Raise exception on invalid operations, arithmetic, overflow and divide-by-zero.
		deferred
		end

	Qnan: DOUBLE is
			-- Quiet Not-A-Number.  When used in an expression, the result is QNAN.
		local
			string: STRING
			l_c_string: ROSE_BYTES
		once
			string := "12345678"
			string.put('%/0/', 1)
			string.put('%/0/', 2)
			string.put('%/0/', 3)
			string.put('%/0/', 4)
			string.put('%/0/', 5)
			string.put('%/0/', 6)
			string.put('%/248/', 7)
			string.put('%/127/', 8)
			create l_c_string.make(string)
			Result := l_c_string.double_at(1)
		end

	Qnan_real_32: REAL_32 is
			-- Quiet Not-A-Number.  When used in an expression, the result is QNAN.
		once
			Result := QNAN.truncated_to_real
		end

	Snan: DOUBLE is
			-- Signaling Not-A-Number.  When used in an expression and invalid operations
			-- are being trapped, an error occurs.
		local
			string: STRING
			l_c_string: ROSE_BYTES
		once
			string := "12345678"
			string.put('%/0/', 1)
			string.put('%/0/', 2)
			string.put('%/0/', 3)
			string.put('%/0/', 4)
			string.put('%/0/', 5)
			string.put('%/0/', 6)
			string.put('%/244/', 7)
			string.put('%/127/', 8)
			create l_c_string.make(string)
			Result := l_c_string.double_at(1)
		end

	positive_infinity: DOUBLE is
		-- Over-large positive number
		local
			string : STRING
			l_c_string : ROSE_BYTES
		once
			string := "12345678"
			string.put('%/0/', 1)
			string.put('%/0/', 2)
			string.put('%/0/', 3)
			string.put('%/0/', 4)
			string.put('%/0/', 5)
			string.put('%/0/', 6)
			string.put('%/240/', 7)
			string.put('%/127/', 8)
			create l_c_string.make(string)
			Result := l_c_string.double_at(1)
	end

	negative_infinity: DOUBLE is
		-- Over small negative number
		local
			string : STRING
			l_c_string : ROSE_BYTES
		once
			string := "12345678"
			string.put('%/0/', 1)
			string.put('%/0/', 2)
			string.put('%/0/', 3)
			string.put('%/0/', 4)
			string.put('%/0/', 5)
			string.put('%/0/', 6)
			string.put('%/240/', 7)
			string.put('%/255/', 8)
			create l_c_string.make(string)
			Result := l_c_string.double_at(1)
		end

	isnan (a_x : DOUBLE): BOOLEAN is
			-- Result is true if 'x' is not a number (QNAN or SNAN).
		deferred
		end

	isfinite (a_x: DOUBLE): BOOLEAN is
			-- Result is true if 'x' is finite and false if 'x' is an NAN, +Infinity or -Infinity.
		deferred
		end

	ieee_is_eq (a_x, a_y : DOUBLE): BOOLEAN is
			-- Useful when x or y may be QNAN.
			-- NEVER compare an NAN directly, it will trap or give undesired behavior!
			-- The result is undefined.  In practice, a particular compiler will always return
			-- the same result but this can change with a new version.
			-- For instance, when 'x' is not NAN, MSVC thinks that these are True:
			--		QNAN>x
			--		QNAN>=x
			--		QNAN=x
			-- and these, False:
			--		QNAN<x
			--		QNAN<=x
			--		QNAN/=x
			-- An examination of the Microsoft assembly code reveals that although the comparison does
			-- reveal that one of the operands was NAN, they assume that this possibility does not
			-- happen.  Specifically, they test equality by checking FP condition code C3 without also
			-- checking C2.  As shown in the chart below, C3 is also used to indicate 'Unordered'.
			--		Condition				C3 C2 C0
			--		ST(0) > Source Operand	 0  0  0
			--		ST(0) < Source Operand	 0  0  1
			--		ST(0) = Source Operand	 1  0  0
			--		Unordered				 1  1  1
			-- If trap_on_invalid is set then a trap will occur, BUT NOT UNTIL THE NEXT
			-- FLOATING POINT INSTRUCTION.  Which could be never!  I have notified Microsoft of this
			-- but it is still a problem in version 6.0.
			--
			-- Note that I only handle the cases of equality and inequality.  The other cases are indeterminate!
		do
			if isnan (a_x) then
				Result := isnan (a_y)
			elseif not isnan (a_y) then
				Result := a_x = a_y
			end
		end

	ieee_is_ne (a_x, a_y: DOUBLE): BOOLEAN
			-- Are `a_x' and `a_y' not equal, or just one NaN?  See 'ieee_is_eq' for more info.
		do
			if isnan(a_x)
			then
				Result := not (isnan(a_y))
			elseif not (isnan(a_y))
			then
				Result := a_x /= a_y
			else
				Result := True
			end
		end

feature {NONE} -- Implementation

	ieee_state: IEEE_STATE is
			-- State of floating point processor
		do
		end

end
