indexing
	description: "Helper class for IEEE_MATH";
	date: "$Date: 2008/12/01 $";
	revision: "$Revision: #5 $"

-- This class stores the state of the FPU.  The routines allow you to see and modify
-- the behaviour of the FPU exceptions.  Note that the state is not continuously
-- updated.

-- Each type of exception has 4 routines :
--	trap_on_XXXXX
--		If XXXXX occurs, trap.  On i386, the XXXXX flag is set and the next
--		floating point instruction traps!  This can lead to some confusion as
--		to where the fault actually occurred.  On Alpha the trap will occur
--		within a few instructions (due to pipelining).
--	allow_XXXXX
--		If XXXXX occurs, set the XXXXX flag, but do not trap.
--	is_XXXXX_trapped
--		Returns True if XXXXX will cause a trap.
--	XXXXX_was_set
--		Returns True if the XXXXX flag had been set by the FPU.  Note that
--		once a flag is set it remains set until cleared by
--		ieee_get_state_and_clear_errors.

-- The normal exceptions to trap are INVALID, OVERFLOW and ZERODIVIDE.
-- For instance, you might check for underflow like this.
--	s := ieee_get_state_and_clear_errors	-- Clear any previous underflow error.
--	x := x / 1000000.0
--	s := ieee_get_state_and_clear_errors
--	if s.underflow_was_set
--	then
--		do something
--	end

-- Note, do not make variables of type IEEE_MATH, instead make them
-- "like ieee_state_type".  The function 'ieee_state_type' is defined
-- in IEEE_MATH.

class
	IEEE_STATE

create
	make

feature {IEEE_MATH} -- Implementation

	make(new_control_word : INTEGER; new_status_word : INTEGER) is
			-- Creation procedure.
		do
			control_word := new_control_word
			status_word := new_status_word
		end

	control_word : INTEGER
	status_word : INTEGER


feature -- Access

	-- INVALID
	-- Caused by :
	--	Any operation on a SNAN (signaling not a number).
	--	SQRT(-1)
	--	0.0 / 0.0
	--	Overflow on conversion to integer.
	--	Comparison of a NAN (including QNAN).
	--	Others...

	trap_on_invalid is
		do
			if is_bit_set(control_word, EM_INVALID)
			then
				control_word := control_word - EM_INVALID
			end
		end

	allow_invalid is
		do
			if not is_bit_set(control_word, EM_INVALID)
			then
				control_word := control_word + EM_INVALID
			end
		end

	is_invalid_trapped : BOOLEAN is
		do
			Result := not is_bit_set(control_word, EM_INVALID)
		end

	invalid_was_set : BOOLEAN is
		do
			Result := is_bit_set(status_word, SW_INVALID)
		end


	-- DENORMAL

	-- Because DENORMAL is only supported on 80x86 and not Alpha and because you must
	-- call '_control87' instead of '_controlfp' I have chosen not to implement DENORMAL
	-- here.  If you really want it, see IEEE_87_STATE and IEEE_87_MATH.

	-- Caused by:
	--	An attempt was made to operate on a denormal number.

	-- A denormal is a floating point number that can not be represented as a normalized
	-- number.  Instead it is store with reduced precision.

	is_denormal_trapped : BOOLEAN is
		do
			Result := not is_bit_set(control_word, EM_DENORMAL)
		end

	denormal_was_set : BOOLEAN is
		do
			Result := is_bit_set(status_word, SW_DENORMAL)
		end

	-- ZERODIVIDE
	-- Caused by:
	--	An attempt was made to divide a finite non-zero operand by 0.

	trap_on_zerodivide is
		do
			if is_bit_set(control_word, EM_ZERODIVIDE)
			then
				control_word := control_word - EM_ZERODIVIDE
			end
		end

	allow_zerodivide is
		do
			if not is_bit_set(control_word, EM_ZERODIVIDE)
			then
				control_word := control_word + EM_ZERODIVIDE
			end
		end

	is_zerodivide_trapped : BOOLEAN is
		do
			Result := not is_bit_set(control_word, EM_ZERODIVIDE)
		end

	zerodivide_was_set : BOOLEAN is
		do
			Result := is_bit_set(status_word, SW_ZERODIVIDE)
		end

	-- OVERFLOW
	-- Caused by:
	--	The rounded result of an arithmetic instruction exceeds the largest
	--	allowable finite value that will fit into the real format of the
	--	destination operand.

	trap_on_overflow is
		do
			if is_bit_set(control_word, EM_OVERFLOW)
			then
				control_word := control_word - EM_OVERFLOW
			end
		end

	allow_overflow is
		do
			if not is_bit_set(control_word, EM_OVERFLOW)
			then
				control_word := control_word + EM_OVERFLOW
			end
		end

	is_overflow_trapped : BOOLEAN is
		do
			Result := not is_bit_set(control_word, EM_OVERFLOW)
		end

	overflow_was_set : BOOLEAN is
		do
			Result := is_bit_set(status_word, SW_OVERFLOW)
		end

	-- UNDERFLOW
	-- Caused by:
	--	The rounded result of an arithmetic instruction is "tiny" (that is, less than the
	--	smallest possible normalized, finite value that will fit into the real format of
	--	the destination operand).

	trap_on_underflow is
		do
			if is_bit_set(control_word, EM_UNDERFLOW)
			then
				control_word := control_word - EM_UNDERFLOW
			end
		end

	allow_underflow is
		do
			if not is_bit_set(control_word, EM_UNDERFLOW)
			then
				control_word := control_word + EM_UNDERFLOW
			end
		end

	is_underflow_trapped : BOOLEAN is
		do
			Result := not is_bit_set(control_word, EM_UNDERFLOW)
		end

	underflow_was_set : BOOLEAN is
		do
			Result := is_bit_set(status_word, SW_UNDERFLOW)
		end

	-- INEXACT
	-- Caused by:
	--	The result of an operation is not exactly representable in the destination format.
	--	For example, the fraction 1/3 cannot be precisely represented in binary form.

	trap_on_inexact is
		do
			if is_bit_set(control_word, EM_INEXACT)
			then
				control_word := control_word - EM_INEXACT
			end
		end

	allow_inexact is
		do
			if not is_bit_set(control_word, EM_INEXACT)
			then
				control_word := control_word + EM_INEXACT
			end
		end

	is_inexact_trapped : BOOLEAN is
		do
			Result := not is_bit_set(control_word, EM_INEXACT)
		end

	inexact_was_set : BOOLEAN is
		do
			Result := is_bit_set(status_word, SW_INEXACT)
		end


feature {NONE} -- Implementation

	is_bit_set(an_integer : INTEGER; a_bit : INTEGER) : BOOLEAN is
		do
			Result := (an_integer // a_bit) \\ 2 = 1
		end

	-- See float.h.
	-- Note that Microsoft remaps the status and control words into their own format.

	MCW_EM :	INTEGER is 524319		--0x0008001F
	EM_INVALID :	INTEGER is 16			--0x00000010
	EM_DENORMAL :	INTEGER is 524288		--0x00080000
	EM_ZERODIVIDE :	INTEGER is 8			--0x00000008
	EM_OVERFLOW :	INTEGER is 4			--0x00000004
	EM_UNDERFLOW :	INTEGER is 2			--0x00000002
	EM_INEXACT :	INTEGER is 1			--0x00000001

	MCW_IC :	INTEGER is 262144		--0x00040000
	IC_AFFINE :	INTEGER is 262144		--0x00040000
	IC_PROJECTIVE :	INTEGER is 0			--0x00000000

	MCW_RC :	INTEGER is 768			--0x00000300
	RC_CHOP :	INTEGER is 768			--0x00000300
	RC_UP :		INTEGER is 512			--0x00000200
	RC_DOWN :	INTEGER is 256			--0x00000100
	RC_NEAR :	INTEGER is 0			--0x00000000

	MCW_PC :	INTEGER is 196608		--0x00030000
	PC_24 :		INTEGER is 131072		--0x00020000
	PC_53 :		INTEGER is 65536		--0x00010000
	PC_64 :		INTEGER is 0			--0x00000000

	SW_INEXACT :	INTEGER is 1			--0x00000001
	SW_UNDERFLOW :	INTEGER is 2			--0x00000002
	SW_OVERFLOW :	INTEGER is 4			--0x00000004
	SW_ZERODIVIDE :	INTEGER is 8			--0x00000008
	SW_INVALID :	INTEGER is 16			--0x00000010
	SW_DENORMAL :	INTEGER is 524288		--0x00080000

end
