indexing
	description: "Common code to OLEDB servers for classic and dotnet"
	author: "Yves Chauvin"
	last_editor: "$Author: rj $"
	date: "$Date: 2012/01/10 $"
	revision: "$Revision: #9 $"
	path: "$File: //rose/source/kernel/infrastructure/persistence/windows/oledb_any_data_server_common.e $"

deferred class
	OLEDB_ANY_DATA_SERVER_COMMON

feature -- Access

	rows_affected: INTEGER
			-- affected rows
	status: INTEGER
			-- status

feature -- Status report

	is_storable_date (a_date: ROSE_DATE): BOOLEAN is
			-- checks to see if date is within the storable range of the data server
		require
			date_non_void: a_date /= Void
		local
			l_min_date: ROSE_DATE --perhaps these should be global variables ?
			l_max_date: ROSE_DATE
		do
			create 	l_min_date.make_ccyymmdd(17530101)
			create 	l_max_date.make_ccyymmdd(99991231)
			if a_date >= l_min_date and a_date <= l_max_date then
				Result := True
			else
				Result := False
			end
	 	end

feature -- Basic operations

	begin_transaction is
			-- Begin a transaction on the data server, increment the transaction depth
		local
			l_transaction_starting_predicate: INTEGER_PREDICATE
		do
			explicit_transaction_depth := explicit_transaction_depth + 1
			if implicit_transaction and then transaction_depth = 0 then
				-- Creating a predicate will start a single transaction.  We don't have to execute it.
				-- If we had performed a 'begin transaction' we would have had two transactions!
				create l_transaction_starting_predicate.make (data_server, "select id from sysobjects")
			else
				sql_execute_immediate_any ("begin transaction")
			end
		ensure
			update_transation_depth: explicit_transaction_depth = old explicit_transaction_depth + 1
		end

	data_server: DATA_SERVER is
			-- deferred data server
		deferred
		end

	commit is
			-- commits
		do
			if implicit_transaction or explicit_transaction_depth > 0 then
				-- This eliminates those pesky "Commit without transaction" when every statement
				-- is it's own transaction.
				sql_execute_immediate_any ("commit")
				if explicit_transaction_depth > 0 then
					explicit_transaction_depth := explicit_transaction_depth - 1
				end
			end
		ensure
			set_explicit_transaction_depth: explicit_transaction_depth = 0 or explicit_transaction_depth = old explicit_transaction_depth - 1
		end

	rollback is
			-- rollbacks
		do
			sql_execute_immediate_any ("rollback")
			explicit_transaction_depth := 0
		ensure
			zero_depth: explicit_transaction_depth = 0
		end

	sql_execute_immediate (a_statement: STRING) is
			-- executes immediately
		deferred
		end

	transaction_depth: INTEGER is
			-- Returns the true depth of the current set of transactions.  See explicit_transaction_depth.
		do
			transaction_depth_predicate.retrieve (Void, Void)	-- This does NOT start an implicit transaction.
			if not transaction_depth_predicate.success then
				 (create {EXCEPTIONS}).raise ("The transaction_depth predicate failed")
			end
			Result := transaction_depth_predicate.array.item (1)
		end

feature -- Property settings

	set_implicit_transaction_off is
			-- sets implicit transaction off
		do
			set_implicit_transactions_set(True)
			sql_execute_immediate_any ("SET IMPLICIT_TRANSACTIONS OFF")
			implicit_transaction := False
		end

	set_implicit_transaction_on is
			-- sets implicit transactions on
		do
			set_implicit_transactions_set(True)
			sql_execute_immediate_any ("SET IMPLICIT_TRANSACTIONS ON")
			implicit_transaction := True
		end

	set_implicit_transactions_set(a_value: like implicit_transactions_set)
	-- If you are callling 'sql_execute_immediate' and you know that your statement is read-only but implicit transactions are being set,
	-- you can can prevent that automatic call to 'set_implicit_transaction_on' with the following code:
	--		l_implicit_transactions_set := server.implicit_transactions_set			-- Save this value.
	--		l_server.set_implicit_transactions_set(True)
	--		...sql_execute_immediate()...		-- Because 'implicit_transactions_set' is now True, we will NOT automatically call 'set_implicit_transaction_on'.
	--		l_server.set_implicit_transactions_set(l_implicit_transactions_set)		-- Restore it.
	do
		implicit_transactions_set := a_value
	end

feature -- Access

	implicit_transaction: BOOLEAN
			-- is transaction implicit?

	explicit_transaction_depth: INTEGER
			-- Depth of transactions caused by 'begin trans'.

	implicit_transactions_set: BOOLEAN
			-- True if set_implicit_transaction_on or _off has been called.

feature {NONE} -- Implementation

	disable_statement_is_feature: BOOLEAN
			-- Causes 'statement_is_feature' to always return False.

	sql_execute_immediate_any (a_statement_: STRING) is
			-- Allow any statement to be executed.
		do
			disable_statement_is_feature := True
			sql_execute_immediate (a_statement_)
			disable_statement_is_feature := False
		end

feature {NONE} -- Predicates.

	transaction_depth_predicate: like transaction_depth_predicate_saved
	-- Used by 'transaction_depth'.
	do
		if transaction_depth_predicate_saved=Void
		then
			create transaction_depth_predicate_saved.make(data_server, "select @@TRANCOUNT")	-- This does NOT start an implicit transaction.
			if not transaction_depth_predicate_saved.success
			then
				 (create {EXCEPTIONS}).raise("Cound not create the transaction_depth_predicate")
			end
		end

		Result := transaction_depth_predicate_saved
	end

	transaction_depth_predicate_saved: INTEGER_PREDICATE		-- See 'transaction_depth_predicate'.

end
