note

	description:

		"Class to facilitate the retrieval of partition trees used by clustered tagged databases.  Partition trees are binary trees that are computed and stored by the clustering_tools.exe system and based on currently existing data layouts.  MONDAT, VALDATA and RMDAT (among others) are clustered databases that rely on partition trees."

	author: "Mauricio Bustos"
	last_editor: "$Author: mb $"
	date: "$Date: 2011/07/01 $"
	revision: "$Revision: #2 $"
	path: "$File: //rose/source/kernel/persistence/databases/signature_tagged/partition_tree_factory.e $"

class PARTITION_TREE_FACTORY [G -> COMPARABLE, H -> COMPARABLE]

inherit

	ROSE_SHARED_GLOBAL

create

	make

feature {NONE} -- Creation

	make (a_table: SIGNATURE_TAGGED_TABLE_STRUCTURE) is
			-- Create the factory
		require
			valid_table: a_table /= Void
		do
			table_structure := a_table
		end

feature -- Basic operations

	reconcile (a_path_tree: PARTITION_TREE [G, H]) is
			-- Reconcile `a_path_tree' to make sure it's the most memory 
			-- effecient version
		require
			valid_path_tree: a_path_tree /= Void
		do
			if attached {PARTITION_TREE [INTEGER, ROSE_DATE]} a_path_tree as l_integer_date_partition_tree then
				if not attached {INTEGER_DATE_PARTITION_TREE_RECORD} a_path_tree.item as l_integer_date_partition_tree_record then
					l_integer_date_partition_tree.do_all (agent replace_integer_date_record)
				end
			end
		end

feature -- Access

	retrieved_path_tree: PARTITION_TREE [G, H] is
			-- Retrieved path tree as stored in the database
		local
			l_predicate: COLLECTING_PREDICATE [PARTITION_TREE_RECORD [G, H]]
			l_integer_date_predicate: COLLECTING_PREDICATE [INTEGER_DATE_PARTITION_TREE_RECORD]
			l_array: ROSE_LINEAR_ARRAY [PARTITION_TREE_RECORD [G, H]]
			l_table: HASH_TABLE [PARTITION_TREE [G, H], INTEGER]
			l_node: PARTITION_TREE_RECORD [G, H]
			l_key_0_map: HASH_TABLE [G, INTEGER]
			l_key_1_map: HASH_TABLE [H, INTEGER]
		do
			l_key_0_map := key_0_map
			l_key_1_map := key_1_map
			create l_node
			if attached {PARTITION_TREE_RECORD [INTEGER, ROSE_DATE]} l_node as l_integer_date_partition_tree_node then
				create l_integer_date_predicate.make_strict (table_structure.signature_item_server, "select * from " + table_structure.data_name_prefix + "path_tree order by path")
				l_predicate ?= l_integer_date_predicate
			else
				create l_predicate.make_strict (table_structure.signature_item_server, "select * from " + table_structure.data_name_prefix + "path_tree order by path")
			end
			l_predicate.retrieve (Void, Void)
			if not l_predicate.array.is_empty then
				l_array := l_predicate.array
				create l_table.make (l_array.count)
				if l_key_0_map /= Void or l_key_1_map /= Void then
					create {MAPPABLE_PARTITION_TREE [G, H]} Result.make (l_key_0_map, l_key_1_map, l_array.first)
				else
					create Result.make (l_array.first)
				end
				l_table.put (Result, l_array.first.path)
				from
					l_array.start
				until
					l_array.off
				loop
					process_node (l_array.current_item, l_table)
					l_array.forth
				end
				populate_tree (l_table)
			else
				create Result.make (create {PARTITION_TREE_RECORD [G, H]})
			end
		end
	
feature {NONE} -- Implementation

	populate_tree (a_node_table: HASH_TABLE [PARTITION_TREE [G, H], INTEGER]) is
			-- Populate (grow) the binary tree using `a_node_table'
		require
			valid_table: a_node_table /= Void
		local
			l_tree, l_parent: PARTITION_TREE [G, H]
		do
			from
				a_node_table.start
			until
				a_node_table.off
			loop
				l_tree := a_node_table.item_for_iteration
				if l_tree.item.path /= 0 then
					l_parent := a_node_table.item ((l_tree.item.path - 1) // 2)
					if l_parent /= Void then
						if l_tree.item.path \\ 2 = 0 then
							l_parent.set_right_child (l_tree)
						else
							l_parent.set_left_child (l_tree)
						end
					end
				end
				a_node_table.forth
			end
		end
		
	process_node (a_node: PARTITION_TREE_RECORD [G, H]; a_table: HASH_TABLE [PARTITION_TREE [G, H], INTEGER]) is
			-- Process `a_node' into `a_table'
		local
			l_tree: PARTITION_TREE [G, H]
		do
			create l_tree.make (a_node)
			a_table.put (l_tree, a_node.path)
		end

	key_0_map: HASH_TABLE [G, INTEGER] is
			-- Map of key 0 from data table to partition tree location
			-- (eg, The data is ordered by LISTING but the most 
			-- efficient ordering is by COMPANY)
		local
			l_tuple_0_predicate: SELECTION_PREDICATE [PAIR [G, G]]
			l_template_0: PAIR [G, G]
		do
			if table_structure.signature_item_server.has_table (table_structure.data_name_prefix + "key_0_map") then
				create Result.make (1_000)
				create l_template_0
				create l_tuple_0_predicate.make (table_structure.signature_item_server, l_template_0, "select out_key as first, in_key as second from " + table_structure.data_name_prefix + "key_0_map")
				if not l_tuple_0_predicate.success then
					error.raise ("Failed to make l_tuple_0_predicate")
				end
				l_tuple_0_predicate.apply (agent load_key_0_map (Result, ?), Void, Void)
				if not l_tuple_0_predicate.success then
					error.raise ("Failed to retrieve l_tuple_0_predicate")
				end
			end
		end
	
	key_1_map: HASH_TABLE [H, INTEGER] is
			-- Map of key 1 from data table to partition tree location
			-- (eg, The data is ordered by LISTING but the most 
			-- efficient ordering is by COMPANY)
		local
			l_tuple_1_predicate: SELECTION_PREDICATE [PAIR [H, H]]
			l_template_1: PAIR [H, H]
		do
			if table_structure.signature_item_server.has_table (table_structure.data_name_prefix + "key_1_map") then
				create Result.make (1_000)
				create l_template_1
				create l_tuple_1_predicate.make (table_structure.signature_item_server, l_template_1, "select out_key as first, in_key from " + table_structure.data_name_prefix + "key_1_map")
				if not l_tuple_1_predicate.success then
					error.raise ("Failed to make l_tuple_0_predicate")
				end
				l_tuple_1_predicate.apply (agent load_key_1_map (Result, ?), Void, Void)
				if not l_tuple_1_predicate.success then
					error.raise ("Failed to retreive l_tuple_0_predicate")
				end
			end
		end
	
	load_key_0_map (a_map: HASH_TABLE [G, INTEGER]; a_row: PAIR [G, G]) is
			-- Agent to handle the key_0_map table rows and populate 'a_map'
		require
			valid_map: a_map /= Void
			valid_row: a_row /= Void
		do
			if attached {INTEGER_REF} a_row.second as l_row_second then
				a_map.put (a_row.first, l_row_second.item)
			else
				fail ("load_key_0_map", "Incorrect type in key_0_map", Void)
			end
		ensure
			has_map_item: a_map.count = old a_map.count + 1
		end

	load_key_1_map (a_map: HASH_TABLE [H, INTEGER]; a_row: PAIR [H, H]) is
			-- Agent to handle the key_1_map table rows and populate 'a_map'
		require
			valid_map: a_map /= Void
			valid_row: a_row /= Void
		do
			if attached {INTEGER_REF} a_row.second as l_row_second then
				a_map.put (a_row.first, l_row_second.item)
			else
				fail ("load_key_1_map", "Incorrect type in key_1_map", Void)
			end
		ensure
			has_map_item: a_map.count = old a_map.count + 1
		end

	replace_integer_date_record (a_item: PARTITION_TREE [INTEGER, ROSE_DATE]) is
			-- Replace the item in `a_item' with an 
			-- INTEGER_DATE_PARTITION_TREE_RECORD for memory 
			-- efficiency
		require
			valid_item: a_item /= Void
		local
			l_integer_date_record: INTEGER_DATE_PARTITION_TREE_RECORD
		do
			create l_integer_date_record
			l_integer_date_record.set_from_other (a_item.item)
			a_item.set_item (l_integer_date_record)
		end

	table_structure: SIGNATURE_TAGGED_TABLE_STRUCTURE
	
invariant

	available_table_structure: table_structure /= Void
	
end
