indexing
	description: "This class allows you to monitor or control a persistent clustered tagged server."
	author: "Mauricio Bustos"
	last_editor: "$Author: rj $"
	date: "$Date: 2015/03/25 $"
	revision: "$Revision: #22 $"
	path: "$File: //rose/source/kernel/persistence/databases/signature_tagged/persistent_clustered_tagged_server_monitor.e $"

class
	PERSISTENT_CLUSTERED_TAGGED_SERVER_MONITOR

inherit
	SECURE_ROSE_DEDICATED_NETWORK_CLIENT
		redefine
			get_supported_security_systems_list,
			is_monitor
		end
	PERSISTENT_SERVER_COMMAND_TYPES
	DATABASE_SERVER
		rename
			servers as database_servers
		end

create
	make,
	make_with_machine_name,
	make_with_machine_name_and_storable_type,
	make_exact

feature {NONE} -- Creation

	make (a_network_server_name: STRING; a_args: ARRAY[STRING]; a_executable_name: STRING) is
			-- Create a monitor.
		do
			make_with_machine_name(Void, a_network_server_name, a_args, a_executable_name, False)
		end

	make_with_machine_name(a_machine_name, a_network_server_name: STRING; a_args: ARRAY[STRING]; a_executable_name: STRING; a_quiet: BOOLEAN) is
			-- Create a monitor, forced to a specific machine.
		do
			make_with_machine_name_and_storable_type(a_machine_name, a_network_server_name, a_args, a_executable_name, Void, False)
		end

	make_with_machine_name_and_storable_type
			(a_machine_name, a_network_server_name: STRING; a_args: ARRAY[STRING]; a_executable_name, a_storable_type: STRING; a_quiet: BOOLEAN)
			-- By supplying the storable type we don't have to visit the repository.  PORT_SERVER uses this since it knows the type.
		local
			l_cmd: PAIR[INTEGER, ARRAY[ANY]]
			l_env: EXECUTION_ENVIRONMENT
			l_nt_server_list_saved, l_persistent_server_list, l_persistent_server_list_temp: STRING
			l_storable_type: STRING
			l_found_storable_type: STRING
			l_server_list_predicate: TUPLE_PREDICATE[TUPLE[machine: STRING; storable_type: STRING]]
			l_container: ROSE_LINEAR_ARRAY[TUPLE[machine: STRING; storable_type: STRING]]
			l_server_delivery, l_executable_name: STRING
			l_nt_server_list_temp: STRING
		do
			create l_env

			-- Save the environment variable NT_SERVER_LIST (we will restore it at the bottom).
			l_nt_server_list_saved := l_env.get ("NT_SERVER_LIST")

			-- The environment variable PERSISTENT_SERVER_LIST can be used to force us to use a particular server.
			l_persistent_server_list_temp := l_env.get("PERSISTENT_SERVER_LIST")

			-- If the caller specified 'a_machine_name' then use that for NT_SERVER_LIST (if we aren't forcing something else).
			if a_machine_name/=Void and l_persistent_server_list_temp=Void
			then
				l_nt_server_list_temp := l_env.get("NT_SERVER_LIST")
				if l_nt_server_list_temp/=Void
				then
					l_nt_server_list_temp.to_upper
					if not l_nt_server_list_temp.has_substring(a_machine_name.as_upper)
					then
						l_nt_server_list_temp.append("," + a_machine_name.as_upper)
						l_env.put(l_nt_server_list_temp, "NT_SERVER_LIST")
					end
				else
					l_env.put(a_machine_name.as_upper, "NT_SERVER_LIST")
				end
			end

			-- The following determines both the machine and storable_type.  However, if the caller has
			-- specified these we should bypass this code.  Note that the caller may SPECIFY "" for 'a_storable_type'.
			if a_machine_name=Void or a_storable_type=Void
			then
				-- Build the predicate to search for the persistent server entry in the table 'persistent_servers'.
				create l_server_list_predicate.make(
					database_servers.repository_server,
					"select machine, storable_type %N%
					%from persistent_servers %N%
					%where name = ? and enabled = 1 and (received_arguments is null or received_arguments='')")
				if not l_server_list_predicate.success then
					raise ("Could not create l_server_list_predicate")
				end

				-- Get the data from the 'persistent_servers' table.
				create l_container.make_empty
				l_server_list_predicate.set_container(l_container)
				l_server_list_predicate.retrieve(Void, <<a_network_server_name>>)
				if not l_server_list_predicate.success then
					raise ("Could not retrieve with l_server_list_predicate")
				end

				-- I think that we are looping so that we might randomly pick one of the entries.
				-- At the moment we don't support multiple servers this way.
				from l_container.start until l_container.off loop
					-- Build up a list of machine names.
					if l_persistent_server_list = Void then
						l_persistent_server_list := l_container.current_item.machine.twin
					else
						l_persistent_server_list := l_persistent_server_list + "," + l_container.current_item.machine
					end

					-- Get the storable type ("SED" or "").
					l_found_storable_type := l_container.current_item.storable_type
					if l_found_storable_type/=Void
					then
						l_found_storable_type.to_upper
					else
						l_found_storable_type := ""
					end

					-- This code saves the storable type but also prints a warning if multiple types were found (very bad).
					if l_storable_type=Void
					then
						if equal(l_found_storable_type, "SED")
						then
							l_storable_type := "SED"
						else
							l_storable_type := ""
						end
					else
						if not equal(l_found_storable_type, l_storable_type)
						then
							print(database_servers.repository_server.node + "." + database_servers.repository_server.database +
								".persistent_servers contains multiple 'storable_type's for server=" + a_network_server_name + "%N")
							if equal(l_found_storable_type, "SED")
							then
								print("Using SED.%N")
							else
								print("Using regular storables.%N")
							end
						end
					end

					l_container.forth
				end
			end

			-- At this point we know the server(s) and the storable type from the repository.
			-- If the environment variable "PERSISTENT_SERVER_LIST" was set then we don't care about what
			-- the repository thinks about the server but we had to go there to get the storable type.

			-- Set the environment variable NT_SERVER_LIST (for use by 'get_server' later).
			if l_persistent_server_list_temp/=Void
			then
				l_env.put(l_persistent_server_list_temp, "NT_SERVER_LIST")
			elseif a_machine_name=Void
			then
				if l_persistent_server_list=Void
				then
					raise("l_persistent_server_list was Void")
				end

				l_env.put(l_persistent_server_list, "NT_SERVER_LIST")
			else
				-- If we get here then "NT_SERVER_LIST" was already set at the top of this routine.
			end

			-- At this point, if you are are not testing, "NT_SERVER_LIST" has been set to either 'a_machine_name' (if
			-- not Void) or the machine listed in the 'persistent_servers' table.

			l_executable_name := a_executable_name.twin
			l_server_delivery := server_delivery (a_network_server_name)
			if l_server_delivery /= Void and not l_server_delivery.is_empty then
				if not a_quiet
				then
					output (global.interpreted ("($now) Using delivery $1 for $2%N"), <<l_server_delivery.as_upper,a_network_server_name>>)
				end

				if not l_server_delivery.item (l_server_delivery.count).is_equal ('\') then
					l_server_delivery := l_server_delivery.twin + "\"
				end

				l_executable_name.prepend (l_server_delivery)
			end

			server_name := a_network_server_name		-- Used in 'get_server'.

			-- Create and populate the PORT_SERVER_MESSAGE which will be sent to the server after we call 'get_server'.
			create port_server_message.make
			port_server_message.set_rose_command (a_network_server_name)
			port_server_message.set_server_arguments (a_args)
			port_server_message.set_executable (server_directory_root + "\" + l_executable_name)

			if a_storable_type/=Void
			then
				l_storable_type := a_storable_type
			end

			if equal (l_storable_type, "SED") then
				use_sed
			end

			get_server			-- Connects us to the server.  (Uses the environment variable NT_SERVER_LIST).
			in_out.set_nodelay
			set_process_identifiers

			-- Send the message to the server.
			create l_cmd.make(21, <<(False).to_reference, port_server_message>>)
			send (l_cmd)

			-- Restore the environment variable NT_SERVER_LIST (we clobbered it at the top).
			if l_nt_server_list_saved /= Void then
				l_env.put (l_nt_server_list_saved, "NT_SERVER_LIST")
			else
				l_env.put ("", "NT_SERVER_LIST")	-- Future calls to 'l_env.get' will return Void.
			end
		rescue
			-- This happens if we fail to connect to the server.  We do not perform a true 'rescue', just clean-up.
			-- RCMD_PORT_SERVER depends on this error.

			-- Restore the environment variable NT_SERVER_LIST (we clobbered it at the top).
			if l_nt_server_list_saved /= Void then
				l_env.put (l_nt_server_list_saved, "NT_SERVER_LIST")
			else
				l_env.put ("", "NT_SERVER_LIST")	-- Future calls to 'l_env.get' will return Void.
			end
		end

	make_exact(a_machine_name: STRING; a_port_number: INTEGER; a_network_server_name: STRING; a_args: ARRAY[STRING]; a_executable_name, a_storable_type: STRING; a_quiet: BOOLEAN)
	-- 'Exact' means that you must specify the machine name and port number.
	-- This creation procedure was written (copied) to get away from all (most) of the environment usage seen above.
	require
		a_machine_name_not_void: a_machine_name /= Void
		a_port_number_valid: a_port_number > 0
	local
		l_cmd: PAIR[INTEGER, ARRAY[ANY]]
		l_env: EXECUTION_ENVIRONMENT
		l_nt_server_list_saved: STRING
		l_storable_type: STRING
		l_server_list_predicate: STRING_PREDICATE
		l_container: ROSE_LINEAR_ARRAY[STRING]
		l_server_delivery, l_executable_name: STRING
	do
		create l_env

		-- Save the environment variable NT_SERVER_LIST (we will restore it at the bottom).
		l_nt_server_list_saved := l_env.get ("NT_SERVER_LIST")

		-- The following determines the storable_type.  However, if the caller has
		-- specified 'a_storable_type' we bypass this code.  Note that the caller may SPECIFY "" for 'a_storable_type'.
		if a_storable_type = Void
		then
			-- Build the predicate to search for the persistent server entry in the table 'persistent_servers'.
			create l_server_list_predicate.make(
				database_servers.repository_server,
				"select storable_type %N%
				%from persistent_servers %N%
				%where name = ? and enabled = 1 and (received_arguments is null or received_arguments='')")
			if not l_server_list_predicate.success then
				raise ("Could not create l_server_list_predicate")
			end

			-- Get the data from the 'persistent_servers' table.
			create l_container.make_empty
			l_server_list_predicate.set_container(l_container)
			l_server_list_predicate.retrieve(Void, <<a_network_server_name>>)
			if not l_server_list_predicate.success then
				raise ("Could not retrieve with l_server_list_predicate")
			end

			if not l_container.is_empty
			then
				l_storable_type := l_container.first

				if l_storable_type /= Void
				then
					l_storable_type.to_upper
				else
					l_storable_type := ""
				end
			end
		else
			l_storable_type := a_storable_type
		end

		-- Set the environment variable NT_SERVER_LIST (for use by 'get_server_by_port_number' later).
		l_env.put(a_machine_name, "NT_SERVER_LIST")

		l_executable_name := a_executable_name.twin
		l_server_delivery := server_delivery(a_network_server_name)
		if l_server_delivery /= Void and not l_server_delivery.is_empty
		then
			if not a_quiet
			then
				output(global.interpreted("($now) Using delivery $1 for $2%N"), <<l_server_delivery.as_upper, a_network_server_name>>)
			end

			if not l_server_delivery.item(l_server_delivery.count).is_equal('\')
			then
				l_server_delivery := l_server_delivery.twin + "\"
			end

			l_executable_name.prepend(l_server_delivery)
		end

		server_name := a_network_server_name		-- Used in 'get_server_by_port_number'.

		-- Create and populate the PORT_SERVER_MESSAGE which will be sent to the server after we call 'get_server_by_port_number'.
		create port_server_message.make
		port_server_message.set_rose_command(a_network_server_name)
		port_server_message.set_server_arguments(a_args)
		port_server_message.set_executable(server_directory_root + "\" + l_executable_name)

		if attached l_storable_type and then l_storable_type ~ "SED"
		then
			use_sed
		end

		get_server_by_port_number(a_port_number)			-- Connects us to the server.  (Uses the environment variable NT_SERVER_LIST).
		in_out.set_nodelay
		set_process_identifiers

		-- Send the message to the server.
		create l_cmd.make(21, <<(False).to_reference, port_server_message>>)
		send(l_cmd)

		-- Restore the environment variable NT_SERVER_LIST.
		if l_nt_server_list_saved /= Void
		then
			l_env.put(l_nt_server_list_saved, "NT_SERVER_LIST")
		else
			l_env.put("", "NT_SERVER_LIST")	-- Future calls to 'l_env.get' will return Void.
		end
	rescue
		-- This happens if we fail to connect to the server.  We do not perform a true 'rescue', just clean-up.

		-- Restore the environment variable NT_SERVER_LIST.
		if l_nt_server_list_saved /= Void
		then
			l_env.put(l_nt_server_list_saved, "NT_SERVER_LIST")
		else
			l_env.put("", "NT_SERVER_LIST")	-- Future calls to 'l_env.get' will return Void.
		end
	end

feature -- Access

	manager_cache: ROSE_VOLATILE_CACHE [TRIPLE [ROSE_LINEAR_ARRAY [ROSE_VIRTUAL_RECORD [CLUSTERED_SIGNATURE[COMPARABLE,COMPARABLE,ENTITY]]], HASH_TABLE [ROSE_VIRTUAL_RECORD [CLUSTERED_SIGNATURE[COMPARABLE,COMPARABLE,ENTITY]], MULTI_KEY], INTEGER], INTEGER] is
			-- A manager cache with the current set of loaded clusters in the persistent server
		local
			l_cmd: PAIR[INTEGER,ARRAY[ANY]]
		do
			create l_cmd
			l_cmd.set_first (manager_cache_cmd_id)
			send (l_cmd)
			receive
			Result ?= received
		end

	client_list: LINKED_LIST[PERSISTENT_SERVER_COMMAND] is
			-- The list of currently active clients connected to the persistent server
		local
			l_cmd: PAIR[INTEGER,ARRAY[ANY]]
			l_retry: BOOLEAN
			l_empty_array: ARRAY[ANY]
		do
			if not l_retry then
				create l_cmd
				l_cmd.set_first (connection_list_cmd_id)
				create l_empty_array.make_empty
				l_cmd.set_second(l_empty_array)		-- For some reason we can't send Void anymore.
				send (l_cmd)
				receive_with_timeout
				Result ?= received
			end
		rescue
			disconnect
			l_retry := True
			retry
		end

	custom_monitor_response(a_message: ARRAY[ANY]): ARRAY[ANY]
	-- Send the message and wait for the response.
	local
		l_cmd: PAIR[INTEGER, ARRAY[ANY]]
		l_retry: BOOLEAN
	do
		if not l_retry then
			create l_cmd
			l_cmd.set_first(custom_monitor_cmd_id)
			l_cmd.set_second(a_message)
			send(l_cmd)
			receive_with_timeout
			Result ?= received
		end
	rescue
		if not l_retry
		then
			disconnect
			l_retry := True
			retry
		end
	end

	debug_output: STRING
	-- <Precursor>
	do
		Result := "MONITOR: "

		if server_name /= Void
		then
			Result.append(server_name)
		else
			Result.append("Void")
		end

		Result.append("/")

		if in_out.debug_output /= Void
		then
			Result.append(in_out.debug_output)
		else
			Result.append("Void")
		end
	end

feature -- Basic Operations

	invalidate_server_cache is
			-- Send a command to the remote server asking it to invalidate it's cache.
		local
			l_cmd: PAIR[INTEGER,ARRAY[ANY]]
		do
			create l_cmd
			l_cmd.set_first (invalidate_cache_cmd_id)
			send (l_cmd)
		end

	invalidate_cache_clusters (a_list: ROSE_LINEAR_ARRAY[INTEGER]) is
			-- Send a command to the remote server asking it to invalidate the clusters in 'a_list'
		local
			l_cmd: PAIR[INTEGER,ARRAY[ANY]]
		do
			create l_cmd
			l_cmd.set_first (invalidate_cache_clusters_cmd_id)
			l_cmd.set_second (<<a_list>>)
			send (l_cmd)
		end

	clear_warehouse: BOOLEAN is
			-- Clear the cluster warehouse used by the persistent server
		local
			l_cmd: PAIR[INTEGER,ARRAY[ANY]]
			l_integer_ref: INTEGER
		do
			create l_cmd
			l_cmd.set_first (clear_warehouse_cmd_id)
			send (l_cmd)
			receive
			l_integer_ref ?= received
			if l_integer_ref /= Void then
				Result := l_integer_ref.item = 0
			end
		end

	shutdown is
			-- Shutdown the persistent server
		local
			l_cmd: PAIR[INTEGER,ARRAY[ANY]]
		do
			create l_cmd
			l_cmd.set_first (shutdown_cmd_id)
			send (l_cmd)
		end

	disconnect is
			-- Disconnect from server
		do
			cleanup
		end

	server_delivery (a_network_server_name: STRING): STRING is
			-- The delivery configured for this server
		require
			valid_name: a_network_server_name /= Void
		do
			Result := Config.value (a_network_server_name + "_delivery")
			if Result = Void then
				Result := Global.app.delivery.value (a_network_server_name + "_delivery")
				if Result = Void then
					Result := Global.app.site.value (a_network_server_name + "_delivery")
				end
			end
		end

	get_supported_security_systems_list(a_machine_name: STRING; a_port_number: INTEGER; a_port_server_message: PORT_SERVER_MESSAGE)
	-- Monitor commands don't use security.  This also avoids an infinite recursion in the Precursor.
	do
		do_nothing
	end

	is_monitor: BOOLEAN
	-- <Precursor>
	do
		Result := True
	end

end
