indexing
	description: "Persistent server base class"
	author: "Mauricio Bustos"
	last_editor: "$Author: rj $"
	date: "$Date: 2015/05/18 $"
	revision: "$Revision: #41 $"
	path: "$File: //rose/source/kernel/persistence/databases/signature_tagged/persistent_server.e $"

deferred class
	PERSISTENT_SERVER

inherit
	PERSISTENT_SERVER_COMMON
	NETWORK_SERVER
		rename
			error as network_server_error
		redefine
			receive, close, in
		end
	DATABASE_SERVER
	ROSE_SHARED_GLOBAL

feature -- Initialization

	make_with_sed (a_port: INTEGER) is
			-- Create a persistent server that uses the SED seriaization framework.
		do
			is_sed_used := True
			make_persistent (a_port)
		end

	make_persistent (a_port: INTEGER) is
			-- Create a persistent server, listening on port `a_port'.
		local
			l_cmd: PERSISTENT_SERVER_COMMAND
		do
			{INET_PROPERTIES}.set_ipv4_stack_preferred(True)		-- Turn off IPv6 (would cause 'set_non_blocking' to do nothing; probably other problems).
			if poll_timeout_milliseconds = 0 then
				poll_timeout_milliseconds := 500
			end
			global.ensure_storable_compatibility_with_56
			Global.io_handler.set_output (create {ROSE_OUTCHANNEL_STANDARD}.make)
			(create {EXECUTION_ENVIRONMENT}).put("MAX", "STID_CACHE_MULTIPLIER")
			--(create {MEMORY}).collection_off
			queued := 100	-- Number of pending connections that can be queued up (default value is 5).

			-- The following is copied from NETWORK_SERVER.make (we added the 'pre_listen').
				create in.make_server_by_port(a_port)
				pre_listen
				if queued=0
				then
					in.listen(5)
				else
					in.listen(queued)
				end

				in.set_non_blocking
				if is_sed_used then
					in.use_sed
				end
			-- End of copied code.

			max_to_poll := 1
			create poll.make_read_only
			create connections.make
			connections.compare_objects

			set_config_prefix
			allowed_security_systems.do_nothing		-- Just checking before the first client.
			heartbeat_interval_seconds := config_integer("heartbeat_interval_seconds", 21600)	-- 6 hours.
			if heartbeat_interval_seconds>0
			then
				add_request(agent show_heartbeat, next_heartbeat_time, "show_heartbeat")
			end

			create l_cmd.make_command (in)			-- Create a command that does nothing.
			poll.put_read_command (l_cmd)			-- Put it into the poller so we will wake up when a client makes a request.

			post_make								-- Give the descendant one last chance to do something before we get this show rolling.
			execute
		rescue
			io.error.putstring ("IN RESCUE%N")
			from
				connections.start
			until
				connections.after
			loop
				connections.item.outflow.close
				connections.forth
			end
			cleanup
		end

	pre_listen
	-- You may redefine this if you want to modify the listen socket 'in' before we call 'listen'.  The most common
	-- reason is to set the send or receive buffer size.  This must be done BEFORE the call to 'listen' (try finding
	-- that in the documentation).  Since this is a server, you probably send large packets and receive small ones.
	-- Therefore you probably want to create a large send buffer and leave the receive buffer alone.  But for this
	-- to do any good at all you MUST create a matching receive buffer on the client.  Here is sample code:
	--
	--		in.set_send_buf_size(512 * 1024)	-- Storables are sent in 256 KB chunks.  Make room for 2 of them.
	--
	-- Remember that each new connection will cause one of these send buffers to be created; so don't go overboard.
	do
	end

	post_make is
			-- You can redefine this to finish any initialization that is required before we call 'execute'.
		do
		end

feature -- Access

	in: ROSE_NETWORK_STREAM_SOCKET						-- This is the 'listen' port.

	connections: LINKED_LIST[PERSISTENT_SERVER_COMMAND]	-- All of the client connections.

	max_to_poll: INTEGER								-- Needed by the MEDIUM_POLLER.  This is the highest socket descriptor seen so far.

	poll: MEDIUM_POLLER									-- Wrapper around a 'select' statement.

	is_sed_used: BOOLEAN								-- Are clients expected to connect with SED?

feature -- Operations

	close is
			-- Since we are persistent, we don't close the socket.
		do
		end

	process_message is
			-- Process an incoming connection (called every half second).
			-- Actually, no processing is done here.
			-- Descendants should redefine this and call 'Precursor'.
	local
		l_client_user_name, l_client_machine, l_socket_name: STRING
		l_socket: ROSE_NETWORK_STREAM_SOCKET
	do
			-- Look for dead connections.
		from
			connections.start
		until
			connections.after
		loop
			if connections.item.expired then
				if connections.item.outflow /= Void then
					l_socket_name := connections.item.outflow.debug_output
				else
					l_socket := connections.item.outflow
					if l_socket /= Void then
						l_socket_name := l_socket.debug_output
					else
						l_socket_name := "?"
					end
				end
				l_client_user_name := connections.item.user_name
				if l_client_user_name = Void then
					l_client_user_name := "<unknown user>"
				end
				l_client_machine := connections.item.machine_name
				if l_client_machine = Void then
					l_client_machine := "<unknown client machine>"
				end

				if connections.item.is_data_client or
				   show_monitor_activity
				then
					output (Global.interpreted("($now) '" + connections.item.client_type_name (connections.item.client_type) + "' Client Dropped (" +
											   "port: " + l_socket_name + ", " +
											   "user: " + l_client_user_name + ", " +
											   "machine: " + l_client_machine + ")%N"), Void)
				end

				poll.remove_read_command (connections.item)
				connections.item.outflow.close
				connections.remove
				perform_post_client_drop_processing
			else
				connections.item.perform_maintenance
			end
			if not connections.after then
				connections.forth
			end
		end
	end

	add_request(a_request: like request_agent_template; a_execute_time_ut: ROSE_DATE; a_name: STRING)
	-- Add a request to be processed by 'process_requests'.  The 'name' is usually the name of the agent.
	-- It will be printed if the request fails.
	require
		command_not_void: a_request/=Void
		name_not_void: a_name/=Void
	do
		if requests=Void
		then
			create requests.make
		end

		requests.extend([a_request, a_execute_time_ut, a_name])
	end

	process_one_request
	-- Process any request (agents) that may have been placed in 'process_queue'.  Only one request is processed to
	-- prevent multiple requests from hogging the server and preventing client access.
	local
		l_now_ut: ROSE_DATE
		l_request_tuple: like request_template
		l_request: like request_agent_template
		l_execute_time_ut: ROSE_DATE
		l_done: BOOLEAN
		l_retry: BOOLEAN
		l_request_name: STRING
		l_trace: STRING
	do
		if not l_retry
		then
			if requests/=Void
			then
				from
					create l_now_ut.make_now_utc
					requests.start
				until
					requests.off or l_done
				loop
					l_request_tuple := requests.item
					l_execute_time_ut := l_request_tuple.execute_time_ut
					if l_execute_time_ut=Void or else l_now_ut>=l_execute_time_ut
					then
						l_done := True
						requests.remove
						l_request := l_request_tuple.request
						l_request.call(Void)
					else
						requests.forth
					end
				end
			end
		else
			if l_request/=Void
			then
				l_request_name := l_request_tuple.name
			end

			if l_request_name=Void
			then
				l_request_name := "Unknown"
			end

			output(l_trace + "%N", Void)
			output(Global.interpreted("($now) The request (" + l_request_name + ") failed.%N"), Void)
			test_server
		end
	rescue
		if not l_retry
		then
			l_retry := True
			l_trace := (create {EXCEPTIONS}).exception_trace
			retry
		end
	end

	request_item(a_name: STRING): like request_template
	-- Returns the first request found with the matching name.
	do
		from
			requests.start
		until
			Result/=Void or requests.off
		loop
			if requests.item.name ~ a_name
			then
				Result := requests.item
			else
				requests.forth
			end
		end
	end

	process_custom_monitor_message(a_cmd: PERSISTENT_SERVER_MONITOR_COMMAND; a_args: ARRAY[ANY])
	-- If you want your server to handle these messages, you must redefine this feature.
	-- This is called from PERSISTENT_SERVER_MONITOR_COMMAND.
	require
		a_cmd_not_void: a_cmd /= Void		-- Not really required here, however, clients will.
	local
		l_message: STRING
	do
		if a_args/=Void and then a_args.lower=1
		then
			l_message ?= a_args.item(1)
		else
			l_message := "Void"
		end

		l_message := "Can not process custom monitor messages. (" + l_message + ")."
		print(l_message + "%N")
		outflow.independent_store(<<l_message>>)
	end

	receive is
			--process the next message
		local
			l_cmd: PAIR[INTEGER, ARRAY[ANY]]
			l_is_data_client: BOOLEAN_REF
			l_port_server_message: PORT_SERVER_MESSAGE
			l_retry: BOOLEAN
			l_exception_trace: STRING
		do
			if not l_retry then
				in.accept
				outflow ?= in.accepted		-- The socket 'in' is non-blocking, thus the test for Void in the next statement.
				if outflow /= Void then
					outflow.set_blocking
					if is_sed_used then
						outflow.use_sed
					end

					-- When we call 'next_command' we will read from 'outflow'.  If the client is in the debugger this could take hours
					-- so we set a timeout.  In the event of a timeout all we get is "Retrieve: unable to read type of storable.".
					outflow.set_receive_timeout(config_integer("first_receive_timeout_ms", 3000))
					l_cmd ?= next_command		-- The socket 'outflow' is blocking and we will wait here until the client sends a command.
					outflow.set_receive_timeout(0)

					if l_cmd = Void then
						output (Global.interpreted("($now) Invalid Command%N"),Void)
					elseif l_cmd.first = 21 then		-- 21 is 'set_data_client_id' from SIGNATURE_TAGGED_NAMES but this class isn't specifically dedicated to sig_tagged.
						l_is_data_client ?= l_cmd.second.item(1)
						if l_is_data_client /= Void then
							l_port_server_message ?= l_cmd.second.item(2)

							if l_port_server_message /= Void then
								if l_port_server_message.command_line = Void
								then
									l_port_server_message.set_command_line("Void")
								end

								if l_is_data_client.item then
									if l_cmd.second.upper >= 3 and then attached {STRING} l_cmd.second[3] as al_security_system_name
									then
										if is_security_system_name_allowed(al_security_system_name)
										then
											if al_security_system_name ~ (create {PS_SECURITY_SYSTEM_V2}.default_create).name
											then
												create_new_version_2_security_client(l_port_server_message)
--											elseif al_security_system_name ~ security_system_XYZ.name
--											then
												-- Add other security systems here.
											else
												write_connect_status_message("Client DENIED (unsupported security system [" + al_security_system_name + "]).",
																			 outflow.port_number_string, l_port_server_message, l_port_server_message.command_line)
												resend("SERVER_EXCEPTION: Unsupported authorization system.")
											end
										else
											write_connect_status_message("Client DENIED (unsupported security system [" + al_security_system_name + "]).",
																		 outflow.port_number_string, l_port_server_message, l_port_server_message.command_line)
											resend("SERVER_EXCEPTION: Unsupported authorization.")
										end
									elseif require_security
									then
										write_connect_status_message("Client DENIED (old style client).", outflow.port_number_string, l_port_server_message, l_port_server_message.command_line)
										resend("SERVER_EXCEPTION: No authorization name.")
									else
										-- The client doesn't support security but we will let it slide.
										write_connect_status_message("Unsecure Client allowed!", outflow.port_number_string, l_port_server_message, l_port_server_message.command_line)
										create_new_client (l_port_server_message)
									end
								else
									create_new_monitor (l_port_server_message)
								end
							else
								output (Global.interpreted("($now) Invalid Command (second-2)%N"),Void)
							end
						else
							output (Global.interpreted("($now) Invalid Command (second-1)%N"),Void)
						end
					else
						output (Global.interpreted("($now) Invalid Command (first)%N"),Void)
					end
				end

				poll.execute (max_to_poll, poll_timeout_milliseconds)
				process_one_request		-- We periodically check to see if any requests need to be executed.
				failure_count := 0
			else
				if outflow /= Void then
					if l_is_data_client /= Void
					then
						if l_is_data_client.item
						then
							output (Global.interpreted("($now) 'Client' failed to initialize "), Void)
						else
							output (Global.interpreted("($now) 'Monitor' failed to initialize "), Void)
						end
					else
						output(Global.interpreted("($now) Initialization failed "), Void)
					end

					if l_port_server_message /= Void then
						output ("(user: $1, machine: $2, executable: $3)%N", <<l_port_server_message.user_id, l_port_server_message.machine, l_port_server_message.command_line>>)
					else
						output ("(-PORT_SERVER_MESSAGE Unavailable-)%N", Void)
					end
					output (l_exception_trace, Void)
					outflow.close
				else
					output (Global.interpreted("($now) Client initialization failed because 'outflow' = Void%N"), Void)
					output (l_exception_trace, Void)
				end
			end
		rescue
			failure_count := failure_count + 1
			if failure_count <= 10
			then
					l_exception_trace := (create {EXCEPTIONS}).exception_trace
				l_retry := True
				retry
			end
		end

	is_security_system_name_allowed(a_security_system_name: STRING): BOOLEAN
	-- Is 'a_security_system_name' one of the allowed security systems?
	do
		across
			allowed_security_systems as c_allowed_security_systems
		until
			Result
		loop
			Result := c_allowed_security_systems.item.name ~ a_security_system_name
		end
	end

	write_connect_status_message(a_header, a_socket_name: STRING; a_port_server_message: PORT_SERVER_MESSAGE; a_command_line: STRING)
	-- Print a message to the log declaring whether we allowed or disallowed a client connection.
	require
		a_port_server_message_not_void: a_port_server_message /= Void
	do
		output(Global.interpreted("($now) $1 (port: $2, user: $3, machine: $4, pid: $5, cmd: $6)%N"),
								   <<a_header, a_socket_name, a_port_server_message.get_user_id,
								   a_port_server_message.get_machine, a_port_server_message.get_client_pid.out, a_command_line>>)
	end

	create_new_version_2_security_client(a_message: PORT_SERVER_MESSAGE)
	-- This client is not yet authorized but supports security "version_2" (where the authorization will happen).
	require
		a_mesage_not_void: a_message /= Void
	local
		l_new_connection: like server_command
		l_pre_authorized_connection: PERSISTENT_SERVER_COMMAND
	do
		-- Test (and possibly reconnect to) the server.
		test_server

		-- Create a regular connection but do not call 'create_new_client_or_monitor'.
		l_new_connection := new_client_connection(a_message)
		initialize_connection(l_new_connection, a_message)
		register_command(l_new_connection)

		-- Create a connection that handles the authorization and can (eventually) forward messages to 'l_new_connection'.
		l_pre_authorized_connection := security_system_v2.pre_authorization_connection(outflow, a_message, l_new_connection)
		initialize_connection(l_pre_authorized_connection, a_message)
		create_new_client_or_monitor(a_message, l_pre_authorized_connection, True)
	end

	create_new_monitor (a_message: PORT_SERVER_MESSAGE) is
			-- A client has connected but he doesn't want data he wants to monitor (or control) this application.
		require
			a_mesage_not_void: a_message /= Void
		local
			l_new_connection: PERSISTENT_SERVER_MONITOR_COMMAND
		do
			test_server
			create l_new_connection.make_command (outflow)
			initialize_connection(l_new_connection, a_message)
			register_monitor_command(l_new_connection)
			l_new_connection.set_connections (connections)
			create_new_client_or_monitor (a_message, l_new_connection, False)
		end

	create_new_client (a_message: PORT_SERVER_MESSAGE)
			-- A new client is connecting.
		require
			a_mesage_not_void: a_message /= Void
		local
			l_new_connection: like server_command
		do
			test_server
			l_new_connection := new_client_connection(a_message)
			initialize_connection(l_new_connection, a_message)
			register_command(l_new_connection)
			create_new_client_or_monitor (a_message, l_new_connection, True)
			create last_data_client_connect_time.make_now
		end

	new_client_connection(a_message: PORT_SERVER_MESSAGE): like server_command
	-- Returns a connection to wrap the socket and process messages.  May be redefined.
	do
		create Result.make_command(outflow)
	end

	create_new_client_or_monitor (a_message: PORT_SERVER_MESSAGE; a_connection: PERSISTENT_SERVER_COMMAND; a_client: BOOLEAN) is
			-- The guts of 'create_new_client' and 'create_new_monitor'.
		require
			a_connection_not_void: a_connection /= Void
			a_connection_command_line_not_void: a_connection.command_line /= Void
		local
			l_socket_name: STRING
			l_type_of_client: STRING
			l_socket: ROSE_NETWORK_STREAM_SOCKET
		do
			if a_client then
				l_type_of_client := "Client"
			else
				l_type_of_client := "Monitor"
			end

			if a_client then
				l_socket_name := a_connection.outflow.debug_output
			else
				l_socket := a_connection.outflow
				if l_socket /= Void then
					l_socket_name := l_socket.debug_output
				else
					l_socket_name := "?"
				end
			end

			poll.put_read_command (a_connection)
			-- When you call 'put_read_command' you are eventually calling 'FD_SET' in the run-time
			-- (in network.c).  Unfortunately, FD_SET will silently fail if the struct is full.
			-- If you turn on 'ensure' in the library 'net', you will see that failure.  During
			-- F_Code execution, we must manually test ('is_bit_set').  The limit on the number of
			-- sockets is 64 by default.  But ISE now sets it to 256.  RJ

			if poll.read_mask.is_bit_set (a_connection.outflow.descriptor) then
				if max_to_poll <= outflow.descriptor then
					max_to_poll := outflow.descriptor + 1
				end
				connections.extend (a_connection)

				if attached {PS_SECURITY_SYSTEM_CLIENT_COMMAND} a_connection as al_connection and then
				   al_connection.security_system /= allowed_security_systems.first
				then
					output(Global.interpreted("($now) The next client is using a non-preferred security system (" + al_connection.security_system.name + ")!%N"), Void)
				end

				if a_client or
				   show_monitor_activity
				then
					write_connect_status_message("New " + l_type_of_client, l_socket_name, a_message, a_connection.command_line)
				end
			else
				write_connect_status_message("Too many connections, rejecting " + l_type_of_client, l_socket_name, a_message, a_connection.command_line)
				poll.remove_read_command (a_connection)
				a_connection.close
			end
		end

	initialize_connection(a_connection: PERSISTENT_SERVER_COMMAND; a_message: PORT_SERVER_MESSAGE)
	-- Set some interesting attributes in 'a_new_connection'.
	require
		a_connection_not_void: a_connection /= Void
		a_mesage_not_void: a_message /= Void
	local
		l_executable_name: STRING
		l_command: STRING
	do
		a_connection.set_machine_name (a_message.get_machine)
		a_connection.set_client_pid (a_message.get_client_pid)
		a_connection.set_user_name (a_message.get_user_id)
		l_command := a_message.get_command_line
		if l_command = Void then
			l_command := ""
		end

		a_connection.set_command_line (l_command)
		a_connection.set_port (a_message.get_port)
		l_executable_name := a_message.get_executable
		if l_executable_name = Void then
			l_executable_name := ""
		end

		a_connection.set_executable_name (l_executable_name)
	end

	register_command (a_command: like server_command) is
			-- Called at the end of 'create_new_client' or 'create_new_client_from_command'.
			-- Redefine this if you want access to the command (aka connection) that 'new_client' created.
		do
		end

	register_monitor_command (a_command: like monitor_command) is
			-- Called at the end of 'create_new_client' or 'create_new_client_from_command'.
			-- Redefine this if you want access to the command (aka connection) that 'create_new_client' created.
		do
			a_command.set_server(Current)
		end

	create_new_client_from_command (a_command: PERSISTENT_SERVER_COMMAND) is
			-- Used to create a command that does not follow the standard rules.
			-- For instance, REUTERS_RTP_SERVER uses this to fake a client that is really just
			-- a wrapper around a socket that sometimes sends us messages that we have to process.
		require
			valid_command: a_command /= Void
		local
			l_socket: ROSE_NETWORK_STREAM_SOCKET
			l_socket_name: STRING
		do
			l_socket := a_command.outflow
			if is_sed_used then
				l_socket.use_sed
			end
			if max_to_poll <= l_socket.descriptor then
				max_to_poll := l_socket.descriptor + 1
			end
			connections.extend (a_command)
			poll.put_read_command (a_command)
			l_socket_name := l_socket.debug_output
			output (Global.interpreted ("($now) New Client from command (" + l_socket_name + ")%N"), Void)
		end

	server: SQL_DATA_SERVER is
			-- Data server in use for 'Current'
		deferred
		end

	test_server
			-- See if the server is working.  If not, try to reconnect.  Because many predicates may already have references
			-- to this server we can't simply create a new one.  The existing one must be reconnected.
			-- WARNING!  Make certain not to call 'server' (it's a function) in the error handling part of the code.  It can,
			-- and has, caused errors.
		local
			l_retry: BOOLEAN
			l_sleep_seconds: INTEGER
			l_definition: ROSE_DATABASE_DEFINITION
			l_server: SQL_DATA_SERVER
		do
			if not l_retry then
				l_server := server		-- 'server' is a function.  Do not call 'server' again in this feature.
				-- Test the server.
				if not l_server.connected_by_testing then
					(create {EXCEPTIONS}).raise ("")
				end

				-- Test the auxiliary_server.
				if not l_server.auxiliary_server.connected_by_testing
				then
					(create {EXCEPTIONS}).raise ("")
				end
			else
				-- Sleep from 5 to 120 seconds to give the server a chance to recover.
				if l_sleep_seconds=0
				then
					l_sleep_seconds := 5
				else
					l_sleep_seconds := l_sleep_seconds * 2
					if l_sleep_seconds>120
					then
						l_sleep_seconds := 120
					end
				end

				global.env.sleep(l_sleep_seconds * 1000)

				-- Now we can start the reconnect.
				output (Global.interpreted ("($now) Reconnecting to $1...%N"), <<l_server.database>>)

				-- Save the definition.
				l_definition := servers.pool.definition (l_server.logical)

				-- Break the connection (auxiliary, too).
				l_server.sql_disconnect

				-- Reconnect and test (it could fail again if the server is down for an extended period).
				l_server.set_defintion_node_and_database(l_definition)
				l_server.reconnect				-- Will reconnect the auxiliary_server, too.
				if not l_server.make_succeeded
				then
					(create {EXCEPTIONS}).raise("Reconnect failed.")
				end

				-- Success.
				set_server_implicit_transactions
				error.acknowledge
				output (Global.interpreted ("($now) Reconnected to $1%N"), <<l_server.database>>)
			end
		rescue
			if l_server /= Void then
				output (Global.interpreted ("($now) Server Connection Has Been Broken to $1%N"), <<l_server.database>>)
				l_retry := True
				retry
			else
				output (Global.interpreted ("($now) Server is -Void-%N"), Void)
				error.die (0, "test_server", "Shutting down due to missing server", Void)
			end
		end

	set_server_implicit_transactions
	-- Set implicit_transaction to the default for this server.  The descendant may redefine this feature.
	do
		server.set_implicit_transaction_on
	end

	server_command: PERSISTENT_SERVER_COMMAND is
			-- Type anchor for data clients
		do
		ensure
			anchor_only: Result = Void
		end

	monitor_command: PERSISTENT_SERVER_MONITOR_COMMAND is
			-- Type anchor for monitors
		do
		ensure
			anchor_only: Result = Void
		end

	perform_post_client_drop_processing is
			-- Perform additional processing after a client has been dropped.
			-- Redefine in descendents as necessary.
		do
		end

	show_heartbeat is
			-- Periodically, we show some interesting information to prove that we are is still alive.
			-- This implementation is simple but you can redefine it to your heart's content.
		do
			output (Global.interpreted("($now) Status report.%N"), Void)
			output (data_client_count.out + " data client(s) are connected.%N", Void)
			if last_data_client_connect_time/=Void
			then
				output("The last data client connected at " + last_data_client_connect_time.debug_output + ".%N", Void)
			end

			add_request(agent show_heartbeat, next_heartbeat_time, "show_heartbeat")
		end

	set_config_prefix is
			-- Set's 'config_prefix'.
			-- Each descendant must decide what prefix to use for the variables in the config file.
			-- An example config file might look like this:
			--		use_foreground = False
			--		[/monthly_data/c/U]
			--		use_foreground = True
			-- In this case if the config_prefix is set to "/monthly_data/c/U" then 'use_foreground' would be True,
			-- for all other prefixes, it would be False.
			-- To use this feature, you must use the 'config_*' routines (below).
		do
			config_prefix := ""
		ensure
			config_prefix_not_void: config_prefix /= Void
			config_prefix_begin_and_end_with_slash: config_prefix.count=0 or (config_prefix.item(1)='/' and config_prefix.item (config_prefix.count)='/')
		end

	config_integer (a_var_name: STRING; a_default: INTEGER): INTEGER is
			-- Retrieve an integer value from the config file.  Use 'config_prefix' + a_var_name if available.
		do
			Result := config.integer (config_prefix + a_var_name, config.integer (a_var_name, a_default))
		end

	config_boolean (a_var_name: STRING; a_default: BOOLEAN): BOOLEAN is
			-- Retrieve a string value from the config file.  Use 'config_prefix' + a_var_name if available.
		do
			Result := config.boolean (config_prefix + a_var_name, config.boolean (a_var_name, a_default))
		end

	config_value (a_var_name: STRING; a_default: STRING): STRING is
			-- Retrieve a string value from the config file.  Use 'config_prefix' + a_var_name if available.
		do
			Result := config.value_default (config_prefix + a_var_name, config.value_default (a_var_name, a_default))
		end

feature {NONE} -- Implementation.
	next_heartbeat_time: ROSE_DATE
	-- The time of the next heartbeat message.
	require
		valid_heartbeat_interval_seconds: heartbeat_interval_seconds/=0
	do
		create Result.make_now_utc
		Result.second_add(heartbeat_interval_seconds)
	end

feature -- Status

	data_client_count: INTEGER is
			-- How many 'data' clients are currently connected (as opposed to 'monitor' clients').
		do
			from
				connections.start
			until
				connections.off
			loop
				if connections.item.client_type = connections.item.data_client_type_id then
					Result := Result + 1
				end
				connections.forth
			end
		end

	seconds_until_idle: INTEGER
	-- Used to determine if we are idle (have not processed a data client message for some period
	-- of time - 'min_idle_time_seconds').
	-- If Result is 0 or negative, we are idle.  If positive, the caller should wait that long and try this call again.
	-- Note that Result is an INTEGER.  Live with it.  The polling loop is normally called every 0.5 seconds so that's
	-- all the resolution you get anyway.  If you want to fix it, add 'fine_seconds_add' to ROSE_DATE.
	local
		l_now: ROSE_DATE
		l_connection: PERSISTENT_SERVER_COMMAND
		l_duration: INTEGER
	do
		from
			create l_now.make_now
			connections.start
		until
			connections.off
		loop
			l_connection := connections.item
			if l_connection.is_data_client
			then
				l_duration := l_now.relative_duration_in_seconds(l_connection.last_message_time).truncated_to_integer
				if l_duration<0					-- The duration can go negative when we switch from daylight savings to standard time.
				then
					Result := Result.max(1)				-- Tell the caller to try again in one second.
				elseif l_duration<min_idle_time_seconds
				then
					Result := Result.max(min_idle_time_seconds - l_duration)
				end
			end

			connections.forth
		end
	end

	count_objects
	-- A debugging feature for the benefit of the descendants.  Counts the Eiffel objects and dumps to a DOS window.
	local
		l_eoc: EIFFEL_OBJECT_COUNTER
		l_cas: HASH_TABLE[PAIR[INTEGER_64, INTEGER_64], STRING]
	do
		create l_eoc
		l_cas := l_eoc.counts_and_sizes
		l_eoc.dump_cas_by_size_asc(l_cas, 1000)
		l_cas := Void
	end

	memory_stats: STRING
	-- A debugging feature for the benefit of the descendants.  Return the usage of Eiffel and C memory.
	local
		l_formatter: FORMAT_DOUBLE
		l_meminfo: MEM_INFO
		l_memory: MEMORY
	do
		create l_memory
		create l_formatter.make(10, 3)
		l_formatter.no_justify
		l_meminfo := l_memory.memory_statistics(l_memory.eiffel_memory)
		Result :=
				"E memory used/overhead/total (GB)/chunks = " +
				l_formatter.formatted(l_meminfo.used64.to_real_64     / 1024.0 / 1024.0 / 1024.0) + "/" +
				l_formatter.formatted(l_meminfo.overhead64.to_real_64 / 1024.0 / 1024.0 / 1024.0) + "/" +
				l_formatter.formatted(l_meminfo.total64.to_real_64    / 1024.0 / 1024.0 / 1024.0) + "/" +
				l_meminfo.chunk.out + "%N"
		l_meminfo := l_memory.memory_statistics(l_memory.c_memory)
		Result := Result +
				"C memory used/overhead/total (GB)/chunks = " +
				l_formatter.formatted(l_meminfo.used64.to_real_64     / 1024.0 / 1024.0 / 1024.0) + "/" +
				l_formatter.formatted(l_meminfo.overhead64.to_real_64 / 1024.0 / 1024.0 / 1024.0) + "/" +
				l_formatter.formatted(l_meminfo.total64.to_real_64    / 1024.0 / 1024.0 / 1024.0) + "/" +
				l_meminfo.chunk.out + "%N"
	end

feature -- Constants

	allowed_security_systems: ROSE_LINEAR_ARRAY[PS_SECURITY_SYSTEM]
	-- List of authorization_systems that we will accept from a client.  The one listed first is preferred.
	-- Clients using a non-preferred security system will be logged.
	local
		l_versions: STRING
		l_delimited_versions: ROSE_DELIMITED_TEXT
	once
		create Result.make_empty
		l_versions := config_value("allowed_security_systems", "")

		create l_delimited_versions.make(l_versions, ',')

		across
			l_delimited_versions as c_delimited_versions
		loop
			if c_delimited_versions.item ~ (create {like security_system_v2}.default_create).name
			then
				Result.extend(security_system_v2)
--			elseif c_delimited_versions.item ~ security_system_XYZ.name
--			then
--				Result.extend(security_system_XYZ)	-- Add new security systems here.
			else
				(create {EXCEPTIONS}).raise("Invalid 'allowed_security_systems' variable in the config file")
			end
		end

		if Result.count = 0
		then
			(create {EXCEPTIONS}).raise("The config variable 'allowed_security_systems' is missing or empty")
		end
	ensure
		Result_not_void_or_empty: Result /= Void and then Result.count > 0
	end

	min_idle_time_seconds: INTEGER
	-- This server is declared idle if no client has made a request for this amount of time.
	once
		Result := config_integer("min_idle_time_seconds", 30)
	end

	monitor_commands: ROSE_LINEAR_ARRAY[TUPLE[name: STRING; help: STRING; visible: BOOLEAN; action: FUNCTION[ANY, TUPLE[ARRAY[ANY]], ARRAY[ANY]]]]
	-- Custom monitor commands that can be sent by a remote user through a PERSISTENT_SERVER_MONITOR_COMMAND.
	--		name		The actual string that the client passes.
	--		help		Message explaining the command.
	--		visible		Is this command visible or secret.
	--		action		An Eiffel agent performed when the command is received.  The agent takes one argument, ARRAY[ANY].
	--					This argument comes from the user.  The agent returns an ARRAY[ANY].  This is returned to the user.
	--					The exact values and types are defined by the command.  Many commands do not use the argument.
	--					The return value is often one or more STRINGs representing data that the user requested.
	-- The Watcher can send these commands.
	-- Note that 'name' must be in upper case.
	once
		create Result.make_empty
		Result.extend([security_monitor_command_name, security_monitor_command_explanation, True, agent list_supported_security_systems_agent])
		Result.extend(["SHOW_MONITOR_ACTIVITY_ON", "Log Monitor activity.", True, agent show_monitor_activity_on_agent])
		Result.extend(["SHOW_MONITOR_ACTIVITY_OFF", "Stop logging Monitor activity.", True, agent show_monitor_activity_off_agent])
	ensure
		nice_Result: Result /= Void and then not Result.has(Void)
		valid_monitor_commands: Result.for_all(agent valid_monitor_command(?))
	end

	valid_monitor_command(a_command: like monitor_commands.first): BOOLEAN
	-- Checks to see if 'a_command' is valid.
	require
		a_command_not_void: a_command /= Void
	once
		Result := a_command.name /= Void and then
				  a_command.name ~ a_command.name.as_upper and then
				  a_command.help /= Void and then
				  a_command.action /= Void
	end

	refresh_periodicity: DATE_TIME_DURATION is
			-- Periodicity with which we should refresh
		once
			create Result.make (0,0,0,0,2,0)
		ensure
			valid_result: Result /= Void
		end

	require_security: BOOLEAN
	-- If False, we will allow clients to connect even if they have no security (they will be logged).
	once
		Result := config_boolean("require_security", True)
	end

feature {NONE} -- Retrieve next command.

	next_command: PAIR[INTEGER, ARRAY[ANY]] is
			-- This is redefined in RP_PERSISTENT_SERVER.
		do
			Result ?= outflow.retrieved
		end


feature -- Agents for monitor commands.  The values in 'a_args' came to us from a remote program.  Result is our response.
	    -- These are provided as a convience for the descendant but the descendant must finish the implementation.
	    -- See PERSISTENT_VARIABLE_SERVER.monitor_commands and .process_custom_monitor_message for an example.

	show_monitor_activity_on_agent(a_args: ARRAY[ANY]): ARRAY[ANY]
	-- Turn on logging of monitor commands.
	do
		show_monitor_activity := True
		Result := <<"Monitor activity logging is now ON.">>
		output("Monitor activity logging is now ON.%N", Void)
	end

	show_monitor_activity_off_agent(a_args: ARRAY[ANY]): ARRAY[ANY]
	-- Turn off logging of monitor commands.
	do
		show_monitor_activity := False
		Result := <<"Monitor activity logging is now OFF.">>
		output("Monitor activity logging is now OFF.%N", Void)
	end

	set_verbose_agent(a_args: ARRAY[ANY]): ARRAY[ANY]
	-- Set 'verbose' to the indicated value.
	local
		l_pos: INTEGER
		l_value: STRING
	do
		if a_args.upper>=1 and then attached {STRING} a_args[1] as l_arg_value
		then
			l_pos := l_arg_value.index_of(' ', 1)
			if l_pos>1 and l_pos<l_arg_value.count
			then
				l_value := l_arg_value.substring(l_pos + 1, l_arg_value.count)
			end

			if l_value/=Void and then l_value.is_integer
			then
				verbose := l_value.to_integer
				Result := <<"Verbose set to " + verbose.out + ".">>
				output("Verbose set to " + verbose.out + ".%N", Void)
			else
				Result := <<"You must specify a non-negative argument.">>
			end
		else
			Result := <<"You must specify a non-negative argument.">>
		end
	end

	list_supported_security_systems_agent(a_args: ARRAY[ANY]): ARRAY[ANY]
	-- Returns a list of supported security systems.  The first entry identifies the response as valid.
	-- The remaining items are the names of the supported security systems.  The client must use the first version in the list that it can.
	-- This is for a custom monitor command.
	do
		create Result.make_empty

		Result.force(security_monitor_command_header_response, Result.upper + 1)

		across
			allowed_security_systems as c_allowed_security_systems
		loop
			Result.force(c_allowed_security_systems.item.name, Result.upper + 1)
		end

		if not require_security
		then
			Result.force(k_none, Result.upper + 1)
		end
	end

feature {NONE} -- Templates.

	request_agent_template: PROCEDURE[ANY, TUPLE]
	-- Do not call.  You say 'like request_agent_template'.
	do end

	request_template: TUPLE[request: like request_agent_template; execute_time_ut: ROSE_DATE; name: STRING]
	-- Do not call.  You say 'like request_template'.
	do end

feature -- Exported attributes (alpha order).

	verbose: INTEGER							-- The desired level of verbosity.

feature {NONE} -- Attributes (alpha order).

	config_prefix: STRING						-- Used as a prefix for variables in the config file.

	failure_count: INTEGER						-- Number of failures found during 'receive'.  Is reset upon success.

	heartbeat_interval_seconds: INTEGER			-- Number of seconds between calls to 'show_heartbeat'.

	heartbeat_time: ROSE_DATE					-- The time when we will next call 'show_heartbeat'.

	last_data_client_connect_time: ROSE_DATE	-- The last time a data client started a connection.

	poll_timeout_milliseconds: INTEGER			-- The time in milliseconds to wait for a connection.

	requests: TWO_WAY_LIST[like request_template]
												-- List of requests (agents) to be executed at a particular time (or immediately).

	show_monitor_activity: BOOLEAN				-- Prints log messages showing ordinary monitor_command activity (the command itself may print something).

end
