note
	description: "Shared features for the clients of PERSISTENT_SERVER's."
	author: "Randy John"
	last_editor: "$Author: rj $"
	date: "$Date: 2015/04/16 $"
	revision: "$Revision: #2 $"
	path: "$File: //rose/source/kernel/persistence/databases/signature_tagged/persistent_server_client_common.e $"

deferred class
	PERSISTENT_SERVER_CLIENT_COMMON

inherit
	PERSISTENT_SERVER_COMMON
	ROSE_SHARED_GLOBAL
		rename
			trace as global_trace
		export
			{NONE} all
		end
	DEBUG_OUTPUT

feature -- Implementation.

	get_supported_security_systems_list(a_machine_name: STRING; a_port_number: INTEGER; a_port_server_message: PORT_SERVER_MESSAGE)
	-- Ask the server (given by the arguments) for the names of the security systems it supports.
	-- This routine sets 'allowed_security_system_names' with those names.  If 'allowed_security_system_names'
	-- is set to Void then the server does not support any security system at all.
	local
		i: INTEGER
		l_monitor: PERSISTENT_CLUSTERED_TAGGED_SERVER_MONITOR
		l_names: ARRAY[ANY]
		l_found_none: BOOLEAN
	do
		allowed_security_system_names := Void

		-- To prevent an infinite recursion, this feature (get_supported_security_systems_list) is redefined for 'l_monitor'.
		create l_monitor.make_exact(a_machine_name, a_port_number, a_port_server_message.rose_command, Void, "", Void, False)
		l_names := l_monitor.custom_monitor_response(<<security_monitor_command_name>>)
		l_monitor.disconnect	-- This closes the socket to the server.  This will happen when the object is collected but I don't want to wait that long.
		l_monitor := Void

		if l_names /= Void and then l_names.count >= 1
		then
			if attached {STRING} l_names[1] as al_name1 and then
			   al_name1 ~ security_monitor_command_header_response
			then
				create allowed_security_system_names.make_empty

				from
					i := 2
				until
					i > l_names.count
				loop
					if attached {STRING} l_names[i] as al_name
					then
						if al_name /~ k_none
						then
							allowed_security_system_names.extend(al_name)
						else
							l_found_none := True	-- The server will allow clients with no security (those that don't have this new code).
						end
					end

					i := i + 1
				end

				if config.boolean("debug/no_auth", False)
				then
					-- For testing I allow you to skip authorization, but only if I know that the server will deny your request.
					if l_found_none
					then
						raise("debug/no_auth not permitted on servers with 'require_security' set to False.")
					end

					allowed_security_system_names := Void
				end
			end
		end
	end

	first_matching_security_system: PS_SECURITY_SYSTEM_CLIENT
	-- Loop through the security systems that the server supports and select the first one that we know about.
	-- The list from the server has the preferred system first.
	require
		allowed_security_system_names_not_void: allowed_security_system_names /= Void	-- Set by 'get_supported_security_systems_list'.
	do
		across
			allowed_security_system_names as c_allowed_names
		until
			Result /= Void
		loop
			across
				client_supported_security_systems as c_client_supported_security_systems
			until
				Result /= Void
			loop
				if c_allowed_names.item ~ c_client_supported_security_systems.item.name
				then
					Result := c_client_supported_security_systems.item
				end
			end
		end

		if Result = Void
		then
			output("Supported server side security_systems:%N", Void)

			across
				allowed_security_system_names as c_allowed_names
			loop
				output("%T" + c_allowed_names.item + "%N", Void)
			end

			output("%N", Void)

			output("Supported client side security_systems:%N", Void)

			across
				client_supported_security_systems as c_client_supported_security_systems
			loop
				output("%T" + c_client_supported_security_systems.item.name + "%N", Void)
			end

			output("%N", Void)

			raise("No matching security system could be found.")
		end
	ensure
		result_not_void: Result /= Void
	end

	secure_connect
	-- Connect to the server using the negotiated security system.
	local
		l_security_system: PS_SECURITY_SYSTEM_CLIENT
	do
		l_security_system := first_matching_security_system

		if attached {PS_SECURITY_SYSTEM_V2_CLIENT} l_security_system as al_security_system_v2
		then
			secure_connect_v2(al_security_system_v2)
--		elseif attached {PS_SECURITY_SYSTEM_xyz} l_security_system as al_security_system_xyz
--		then
			-- The next supported security system is handled here, etc.
			-- Be sure to add it to 'client_supported_security_systems'.
		else
			raise("The security system was not found in 'client_supported_security_systems'")
		end
	end

	secure_connect_v2(a_security_system: PS_SECURITY_SYSTEM_V2_CLIENT)
	-- Connect to the server using V2.
	require
		a_security_system_not_void: a_security_system /= Void
	local
		l_checksum: STRING
		l_done: BOOLEAN
		l_auth_info: ARRAY[STRING]
	do
		-- 21 = set_data_client_id from SIGNATURE_TAGGED_NAMES.
		remote_call(21, <<True, port_server_message, a_security_system.name>>)

		-- Ask for an authorization_request.
		send(a_security_system.k_send_authorization_request)

		-- Receive the authorization_request.
		receive

		if attached {ARRAY[STRING]} received as al_auth_info and then
		   al_auth_info.count = 2 and then
		   al_auth_info.lower = 1
		then
			from
				l_auth_info := al_auth_info
			until
				l_done
			loop
				-- Create the database connection.
				a_security_system.set_authorization_server_from_cache(l_auth_info[1])

				-- Get the checksum.
				l_checksum := a_security_system.checksum_from_statement(l_auth_info[2])

				-- Send the checksum back to the server.
				if config.boolean("debug/wrong_checksum", False)
				then
					-- For testing we add 1 to the first byte of the checksum.
					l_checksum.put(l_checksum[1].plus(1), 1)
				end

				send(l_checksum)

				-- The server sends an ANY to indicate success, another auth_info for retry and a STRING for failure.
				receive

				if attached {ARRAY[STRING]} received as al_auth_info2 and then
				   al_auth_info2.count = 2 and then
				   al_auth_info2.lower = 1
				then
					l_auth_info := al_auth_info2	-- Try again.
				elseif attached {STRING} received as al_received
				then
					raise(al_received)
				else
					l_done := True
				end
			end
		elseif attached {STRING} received as al_received2
		then
			raise(al_received2)
		else
			raise("Invalid response from server when requesting authorization")
		end
	rescue
		output("Failure talking to $1.%N", <<debug_output>>)
	end

	remote_call(a_command_id: INTEGER; a_args: ARRAY[ANY])
	-- Send 'a_command_id' with 'a_args' to the server.
	local
		l_cmd: PAIR[INTEGER, ARRAY[ANY]]
	do
		create l_cmd.make(a_command_id, a_args)
		send(l_cmd)
	end

	send(a_message: ANY)
	-- Send 'a_message' to the server.
	deferred
	end

	port_server_message: PORT_SERVER_MESSAGE
	-- Message received from the port_server.
	deferred
	end

	receive
	-- Receive a message from the server (see 'received').
	deferred
	end

	received: detachable ANY
	-- Message received during 'receive'.
	deferred
	end

feature {NONE} -- Once/constants.

	client_supported_security_systems: ROSE_LINEAR_ARRAY[PS_SECURITY_SYSTEM_CLIENT]
	-- All of the security systems that the client can support.  The order is immaterial since the server defines the preferred system.
	once
		create Result.make_from_array(
										<<
											security_system_v2_client
										>>
									)
	ensure
		has_non_void_contents: Result /= Void and then not Result.has(Void)
	end

	security_system_v2_client: PS_SECURITY_SYSTEM_V2_CLIENT
	-- One of the security systems that we support (client side).
	once
		create Result
	ensure
		result_not_void: Result /= Void
	end

feature -- Attributes.

	allowed_security_system_names: detachable ROSE_LINEAR_ARRAY[STRING]
									-- Names of the security systems that the server we want to connect to supports.

end
