note

	description:

		"[
        Class that you should inherit from if you want to use the PES.
	     If your class is bound to a specific context, use PES_AWARE_CONTEXT instead
      ]"

	usage: "[
            In the root class of your application inherit from this class.
            Call set_use_pes_global (True) before you access any context or anything DB related.
            At this point also call set_pes_configuration to specify the server you need.

				If your program is not multi-threaded, this will have no effect. You will not be able to use
            PES, but there will be no other bad effects.
           ]"
	
	documentation: "http://prism/modules/wacko/wakka.php?wakka=PES"
	author: "Jann Roeder"
	last_editor: "$Author: c4 $"
	date: "$Date: 2014/05/21 $"
	revision: "$Revision: #26 $"
	path: "$File: //rose/source/kernel/entities/pes/pes_aware.e $"

class	PES_AWARE

feature {NONE} -- Access

	pes_api_for_context (a_context: SUPPLIER_CONTEXT): PES_CLIENT_API is
			-- Return the PES_CLIENT_API object for `a_context', or Void if there is no PES_CLIENT_API for `a_context_id'
		require
			context_not_void: a_context /= Void
			config_set: pes_configuration_set
		do
			Result := pes_api_for_context_id (a_context.context_index)
		end

	pes_api_for_context_id (a_context_id: INTEGER): PES_CLIENT_API is
			-- Return the PES_CLIENT_API object for `a_context_id'
			-- To get the context id from a context, use the feature `{SUPPLIER_CONTEXT}.context_index'
		require
			valid_context_id: pes_apis.valid_index (a_context_id)
			pes_exists: has_pes (a_context_id)
			config_set: pes_configuration_set
		do
			Result := pes_apis.item (a_context_id)
		ensure
			result_not_void: Result /= Void
		end

	use_pes_global: BOOLEAN is
			-- Query this feature to find out whether PES is globally enabled
		do
			Result := use_pes_internal.item
		end

	use_pes_for_context (a_context_id: INTEGER): BOOLEAN is
			-- Is PES enabled and do we have a server for `a_context_id' ?
		do
			Result := use_pes_internal.item and then has_pes (a_context_id)
		end

feature -- PES configurations

	pes_configurations: ARRAY[ARRAY[ARRAY[STRING]]] is
			-- Some example configurations involving multiple PES
		once
			create Result.make (1, 1)
			Result.item (1) := <<<<"RESEARCH">>>>
			-- The above line was as seen below. But for now we will stick with RESEARCH only,
			--  until  we have decided what to do about other contexts. 2011/05/05 - C4
			-- Result.item (1) := <<<<"RESEARCH", "GLOBAL CLIENT">>,<<"MODEL", "GLOBAL CLIENT">>>>
		end

feature -- External Access

	is_ready: BOOLEAN is
			-- is pes_server loaded?
		do
			Result := pes_server_loaded_internal.item
		end

	pes_result_available: ACTION_SEQUENCE [TUPLE [ROSE_CS_FUTURE [ROSE_CS_RESULT]]] is
			-- Action sequence for all PES APIs
		once
			create Result
		end

feature {NONE} -- Operations

	set_use_pes_global (a_value: BOOLEAN) is
			-- Globally enable/disable PES usage;
			-- Only does this for multi-threaded programs.
		do
			if is_thread_capable then
				use_pes_internal.set_item (a_value)
			end
		end

	set_pes_server_for_context (a_value: BOOLEAN; a_context_id: INTEGER) is
			-- Set this program to be a PES for `a_context_id'
		do
			pes_server_context_internal.force (a_value, a_context_id)
		end

	set_pes_server_loaded (a_value: BOOLEAN) is
			-- Globally enable/disable pes_server_loaded
		do
			pes_server_loaded_internal.set_item (a_value)
		end

	set_pes_configuration (a_config: ARRAY[ARRAY[STRING]]) is
			-- On the client side call this feature to select the desired context combination for the PES
			-- example argument: <<<<"RESEARCH","GLOBAL CLIENT">>>>
		require
			a_config_not_void: a_config /= Void
			a_config_not_empty: not a_config.is_empty
		local
			l_i, l_j: INTEGER
			l_context_index: INTEGER
			l_current_server: ARRAY[STRING]
			l_pes_api: PES_CLIENT_API
		do
			if is_thread_capable then
				from
					l_i := pes_contexts.lower
				until
					l_i > pes_contexts.upper
				loop
					pes_contexts.item (l_i) := False
					
					l_i := l_i + 1
				end
				
				-- Loop in inverse order, so that in case of overlaps in the configuration, the servers occuring in the beginning
				-- are preferred
				from
					l_i := a_config.upper
				until
					l_i < a_config.lower
				loop
					l_current_server := a_config.item (l_i)
					l_pes_api := create_pes_api (l_current_server)
					
					if last_synchronization_successful then
						from
							l_j := l_current_server.lower
						until
							l_j > l_current_server.upper
						loop
							l_context_index := context_access.context_id_from_name (l_current_server.item (l_j))
							
							if pes_contexts.valid_index (l_context_index) then
								pes_contexts.item (l_context_index) := True
								pes_apis.item (l_context_index) := l_pes_api
							end
							
							l_j := l_j + 1
						end
					end
					
					l_i := l_i - 1
				end
				
				pes_configuration.make_from_array (a_config)
			end
		ensure
			context_names_set: is_thread_capable implies pes_configuration.same_items (a_config)
			configuration_set: is_thread_capable implies pes_configuration_set
		end

feature -- Constants

	pes_server_name: STRING is "persistent_entity_server"

feature -- Query

	has_pes (a_context_id: INTEGER): BOOLEAN is
			-- Does the context with ID `a_id' have a PES?
		require
			valid_context: pes_contexts.valid_index (a_context_id)
		do
			Result := pes_contexts.item (a_context_id)
		end

	pes_contexts: ARRAY[BOOLEAN]
			-- Boolean array that indicates for which context ID a PES has been requested
		once
			create Result.make (1, {CONTEXT_DESCRIPTORS}.descriptor_count)
		end

	pes_configuration_set: BOOLEAN is
			-- Is a pes_configuration specified
		do
			Result := not pes_configuration.is_empty
		end

	pes_configuration: ARRAY[ARRAY[STRING]]
			-- PES configuration: <<<<"RESEARCH">>,<<STANDARD AND POORS">>>> defines one server with the research context and one with the S&P context
		once
			create Result.make_from_array (<<>>)
		end

feature {NONE} -- Implementation

	create_pes_api (a_server_config: ARRAY[STRING]): PES_CLIENT_API is
			-- Create a pes API according to `a_server_config' (a list of contexts)
		require
			server_config_not_void: a_server_config /= Void
		local
			l_argument_string: STRING
			l_i: INTEGER
		do
			-- Construct argument string for port server accrding to `pes_contexts'
			if not a_server_config.is_empty then
				from
					l_argument_string := "-contexts="
					l_i := a_server_config.lower
				until
					l_i > a_server_config.upper
				loop
					l_argument_string.append (a_server_config.item (l_i) + ",")

					l_i := l_i + 1
				end
				-- Remove trailing comma
				if l_i > a_server_config.lower then
					l_argument_string.remove_tail (1)
				end
				create Result.make_with_server_name (pes_server_name, <<l_argument_string>>)
			else
				create Result.make_with_server_name (pes_server_name, Void)
			end

			-- Set up GUI stuff
			Result.result_available.extend (agent (a_result: ROSE_CS_FUTURE [ROSE_CS_RESULT]) do pes_result_available.call ([a_result]) end)

			synchronize_db_servers (Result)
		end

	synchronize_db_servers (a_api: PES_CLIENT_API) is
			-- Make DB servers match with PES servers by adding hardcoded overrides
			-- Sets `last_synchronization_successful'
		require
			api_not_void: a_api /= Void
		local
			l_db_versions: LIST[TUPLE[logical: STRING; path: STRING]]
			l_retried: BOOLEAN
			l_db_servers: DATABASE_SERVER
			l_exceptions: EXCEPTIONS
		do
			if not l_retried then
				create l_db_servers

				l_db_versions := a_api.get_server_database_versions
				from
					l_db_versions.start
				until
					l_db_versions.after
				loop
					l_db_servers.servers.pool.set_hardcoded_override (l_db_versions.item.logical, l_db_versions.item.path)

					l_db_versions.forth
				end
				last_synchronization_successful := True
			else
				last_synchronization_successful := False
			end
		rescue
			l_retried := True
			create l_exceptions
			print (l_exceptions.exception_trace)
			retry
		end

	last_synchronization_successful: BOOLEAN
			-- Indicates whether an exception occured when synchronizing DB servers

	remove_pes_api (a_api: PES_CLIENT_API) is
			-- Remove `a_pes_api' from the API list
		require
			api_not_void: a_api /= Void
		local
			l_i: INTEGER
		do
			from
				l_i := pes_apis.lower
			until
				l_i > pes_apis.upper
			loop
				if pes_apis.item (l_i) = a_api then
					pes_apis.item (l_i) := Void
					pes_contexts.item (l_i) := False
				end

				l_i := l_i + 1
			end
		ensure
			api_deleted: not pes_apis.has (a_api)
		end

feature {NONE} -- Implementation

	use_pes_internal: BOOLEAN_REF is
			-- BOOLEAN object that indicates whether PES is used
		once
			create Result
		end

	pes_server_loaded_internal: BOOLEAN_REF is
			-- BOOLEAN object that indicates whether PES is loaded
		once
			create Result
		end

	pes_server_context_internal: ARRAY [BOOLEAN] is
			-- for which context is this server a pes server?
		once
			create Result.make (1, {CONTEXT_DESCRIPTORS}.descriptor_count)
		end

	pes_apis: ARRAY [PES_CLIENT_API] is
			-- Array of PES APIs for different contexts
		once
			create Result.make (1, {CONTEXT_DESCRIPTORS}.descriptor_count)
		end

	context_access: CONTEXT_DESCRIPTORS is
			-- Needed for access to context descriptors
		once
			create Result.make_no_db
		end
	
	is_thread_capable: BOOLEAN
			-- Is the program running multi-threaded?
		local
			l_platform: PLATFORM
		once
			create l_platform
			Result := l_platform.is_thread_capable
		end

invariant
	pes_apis_context_array_equiv: pes_apis.lower = pes_contexts.lower and pes_apis.upper = pes_contexts.upper

end
