note

	description:

		"Class for accessing the PES server"

	author: "Jann Roeder"
	last_editor: "$Author: c4 $"
	date: "$Date: 2011/11/10 $"
	revision: "$Revision: #11 $"
	path: "$File: //rose/source/kernel/entities/pes/pes_client_api.e $"

class PES_CLIENT_API

inherit

	ROSE_CS_CLIENT_API

	KL_IMPORTED_BOOLEAN_ROUTINES

create {PES_AWARE}

	make_with_server_name

feature -- Basic operations

	send_load_category_request (a_category_handler: CATEGORY_HANDLER[ SUPPLIER_CONTEXT, ENTITY]): ROSE_CS_FUTURE [PES_RESULT_LOAD_CATEGORY] is
			-- Send a request to the server to load all state for `a_category_handler'.
		require
			valid_handler: a_category_handler /= Void
		do
			Result := send_restricted_load_category_request (a_category_handler, False, False, False)
		ensure
			result_not_void: Result /= Void
		end

	send_restricted_load_category_request (a_category_handler: CATEGORY_HANDLER[ SUPPLIER_CONTEXT, ENTITY];
		a_identifiers: BOOLEAN; a_table: BOOLEAN; a_entity_ids: BOOLEAN): ROSE_CS_FUTURE [PES_RESULT_LOAD_CATEGORY] is
			-- Send a request to the server to load all or one item of state for `a_category_handler'.
		require
			valid_handler: a_category_handler /= Void
			all_items_or_only_one_item: (not a_identifiers and not a_table and not a_entity_ids) xor
				BOOLEAN_.nxor (<<a_identifiers, a_table, a_entity_ids>>)
		local
			l_query: PES_QUERY_LOAD_CATEGORY
		do
			create l_query.make (a_category_handler.category_id, a_category_handler.context.context_index)
			if a_identifiers then
				l_query.set_only_identifiers
			elseif a_table then
				l_query.set_only_table
			elseif a_entity_ids then
				l_query.set_only_id_table
			end
			create Result.make (Current)
			send_query (l_query, create {PES_STORED_REQUEST_CATEGORY_HANDLER}.make (Result, a_category_handler), Result)
		ensure
			result_not_void: Result /= Void
		end

	send_recipe_request (a_entities: ARRAY[ENTITY]; a_recipe: ROSE_RECIPE; a_include_descriptive_names: BOOLEAN): ROSE_CS_FUTURE[PES_RESULT_RECIPE] is
			-- Send a request to the server
			-- If `a_include_descriptive_names' is true, the attributes for the descriptive_name will be included in the result
		require
			entities_not_void: a_entities /= Void
			recipe_not_void: a_recipe /= Void
		local
			l_query: PES_QUERY_RECIPE
		do
			create l_query.make (a_entities, a_recipe.representation)
			l_query.set_include_descriptive_name (a_include_descriptive_names)
			create Result.make (Current)
			send_query (l_query, create {PES_STORED_REQUEST_RECIPE}.make (Result, a_recipe), Result)
		ensure
			result_not_void: Result /= Void
		end

	send_recipe_request_with_identifiers (a_identifiers: ROSE_RECIPE_INPUT; a_recipe: ROSE_RECIPE; a_include_descriptive_names: BOOLEAN): ROSE_CS_FUTURE[PES_RESULT_RECIPE] is
			-- Send a request to the server, using a list of identifiers as source
			-- If `a_include_descriptive_names' is true, the attributes for the descriptive_name will be included in the result
		require
			identifiers_not_void: a_identifiers /= Void
			recipe_not_void: a_recipe /= Void
		local
			l_query: PES_QUERY_RECIPE
		do
			create l_query.make_with_identifiers (a_identifiers, a_recipe.representation)
			l_query.set_include_descriptive_name (a_include_descriptive_names)
			create Result.make (Current)
			send_query (l_query, create {PES_STORED_REQUEST_RECIPE}.make (Result, a_recipe), Result)
		ensure
			result_not_void: Result /= Void
		end

	send_recipe_request_all (a_source_category: STRING; a_recipe: ROSE_RECIPE; a_include_descriptive_names: BOOLEAN): ROSE_CS_FUTURE[PES_RESULT_RECIPE] is
			-- Send a request to the server, using all entities of a given category as source
			-- If `a_include_descriptive_names' is true, the attributes for the descriptive_name will be included in the result
		require
			source_not_void: a_source_category /= Void
			recipe_not_void: a_recipe /= Void
		local
			l_query: PES_QUERY_RECIPE
		do
			create l_query.make_with_all (a_source_category, a_recipe.representation)
			l_query.set_include_descriptive_name (a_include_descriptive_names)
			create Result.make (Current)
			send_query (l_query, create {PES_STORED_REQUEST_RECIPE}.make (Result, a_recipe), Result)
		ensure
			result_not_void: Result /= Void
		end

	send_entity_request_blocking (a_entity_id, a_category_id, a_context_id: INTEGER): ENTITY is
			-- Send a blocking request for an entity
		local
			l_result: PES_RESULT_ENTITY
		do
			l_result := send_entity_request (a_entity_id, a_category_id, a_context_id).get
			if l_result.remote_exception then
				raise (l_result.tag_name)
			end
			Result := l_result.entity
		end

	send_entity_request (a_entity_id, a_category_id, a_context_id: INTEGER): ROSE_CS_FUTURE[PES_RESULT_ENTITY] is
			-- Send a request for an entity
		local
			l_query: PES_QUERY_ENTITY
		do
			create l_query.make (a_entity_id, a_category_id, a_context_id)
			create Result.make (Current)
			send_query (l_query, Void, Result)
		ensure
			result_not_void: Result /= Void
		end

	send_attribute_request (a_context_index: INTEGER; a_entities: ARRAY[ENTITY]; a_identifiers: ARRAY[STRING]): ROSE_CS_FUTURE[PES_RESULT_ATTRIBUTES] is
			-- Send a request for some attributes of some entities
		local
			l_query: PES_QUERY_ATTRIBUTES
		do
			create l_query.make (a_context_index, a_entities, a_identifiers)
			create Result.make (Current)
			send_query (l_query, Void, Result)
		ensure
			result_not_void: Result /= Void
		end

	send_attribute_request_blocking (a_context_index: INTEGER; a_entities: ARRAY[ENTITY]; a_identifiers: ARRAY[STRING]): ARRAY[ARRAY[ROSE_LINEAR_ARRAY[ENTITY_ATTRIBUTE_RECORD]]] is
			-- Result.item(i).item(j) is the keyhistory of identifier a_identifiers.item(i) and entity a_entities.item(j)
		local
			l_result: PES_RESULT_ATTRIBUTES
		do
			l_result := send_attribute_request (a_context_index, a_entities, a_identifiers).get
			if l_result.remote_exception then
				raise (l_result.tag_name)
			end
			Result := l_result.identifiers_entity
		end

	send_recipe_request_blocking (a_entities: ARRAY [ENTITY]; a_recipe: ROSE_RECIPE; a_include_descriptive_names: BOOLEAN): ROSE_TRANSFORMATION_RESULT is
			-- Send a request to the server and wait for result
		require
			entities_not_void: a_entities /= Void
			recipe_not_void: a_recipe /= Void
		local
			l_result: PES_RESULT_RECIPE
		do
			l_result := send_recipe_request (a_entities, a_recipe, a_include_descriptive_names).get
			if l_result.remote_exception then
				raise (l_result.tag_name)
			end
			Result := l_result.transformation_result
		end

	send_recipe_request_with_identifiers_blocking (a_identifiers: ROSE_RECIPE_INPUT; a_recipe: ROSE_RECIPE; a_include_descriptive_names: BOOLEAN): ROSE_TRANSFORMATION_RESULT is
			-- Send a request to the server, using a list of identifiers as source
			-- If `a_include_descriptive_names' is true, the attributes for the descriptive_name will be included in the result
		require
			identifiers_not_void: a_identifiers /= Void
			recipe_not_void: a_recipe /= Void
		local
			l_result: PES_RESULT_RECIPE
		do
			l_result := send_recipe_request_with_identifiers (a_identifiers, a_recipe, a_include_descriptive_names).get
			if l_result.remote_exception then
				raise (l_result.tag_name)
			end
			Result := l_result.transformation_result
		end

	send_recipe_request_all_blocking (a_source_category: STRING; a_recipe: ROSE_RECIPE; a_include_descriptive_names: BOOLEAN): ROSE_TRANSFORMATION_RESULT is
			-- Send a request to the server, using all entities of a given category
			-- If `a_include_descriptive_names' is true, the attributes for the descriptive_name will be included in the result
		require
			source_category_not_void: a_source_category /= Void
			recipe_not_void: a_recipe /= Void
		local
			l_result: PES_RESULT_RECIPE
		do
			l_result := send_recipe_request_all (a_source_category, a_recipe, a_include_descriptive_names).get
			if l_result.remote_exception then
				raise (l_result.tag_name)
			end
			Result := l_result.transformation_result
		end

feature -- Query

	get_server_database_versions: LIST[TUPLE[logical: STRING; path: STRING]] is
			-- Get the DB versions held by the server
		local
			l_query: PES_QUERY_SERVER_STATE
			l_result: PES_RESULT_SERVER_STATE
			l_result_future: ROSE_CS_FUTURE[PES_RESULT_SERVER_STATE]
			l_result_tuple: TUPLE[result_valid: BOOLEAN; result_value: PES_RESULT_SERVER_STATE]
		do
			create l_query.make_db_version_request
			create l_result_future.make (Current)
			send_query (l_query, Void, l_result_future)
			l_result_tuple := l_result_future.get_with_timeout (30_000)

			if l_result_tuple.result_valid then
				l_result := l_result_tuple.result_value
				if l_result.remote_exception then
					raise (l_result.tag_name)
				end
			else
				l_result_future.free
				raise ("Server timed out")
			end

			Result := l_result.cached_contexts
		ensure
			result_not_void: Result /= Void
		end

end
