note
	description: "Result object to send recipe Result to Client"
	author: "Ueli Peter"
	last_editor: "$Author: zs $"
	date: "$Date: 2011/05/27 $"
	revision: "$Revision: #5 $"
	path: "$File: //rose/source/kernel/entities/pes/pes_result_recipe.e $"

class
	PES_RESULT_RECIPE

inherit
	ROSE_CS_RESULT
		redefine
			prepare_result,
			restore_result
		end

create
	make, make_exception

feature {NONE} -- Creation

	make (a_query_id: like query_id; a_result: like transformation_result; a_cache_desc_name: like cache_desc_name; a_identifiers_for_source: like identifiers_for_source)
			-- Creation
			-- If `a_cache_desc_name' is True, the descriptive name attribute of every entity will be sent in the result,
			-- regardless of the recipe display settings
		do
			query_id := a_query_id
			if a_result /= Void then
				transformation_result := a_result.twin
			else
				transformation_result := Void
			end
			cache_desc_name := a_cache_desc_name
			identifiers_for_source := a_identifiers_for_source
		ensure
			result_set: deep_equal (transformation_result, a_result)
			query_id_set: query_id = a_query_id
		end

feature -- Access

	transformation_result: ROSE_TRANSFORMATION_RESULT
			-- result of a recipe transformation

	cache_desc_name: BOOLEAN
			-- Indicates whether the descriptive name of the entities should be included in the result
			-- reagrdless of the recipe display settings

	identifiers_for_source: ROSE_LINEAR_ARRAY [STRING]
			-- Identifiers to display for the source entities have to be treated specially.

feature -- Operations

	set_identifiers_for_source (a_identifiers: like identifiers_for_source)
			-- Set `identifiers_for_source' to `a_identifiers'
		do
			identifiers_for_source := a_identifiers
		end

	prepare_result
			-- Remove references which are not needed
			-- And cache the attributes for the source entities
		local
			i, j, k: INTEGER
			l_cache: ROSE_LINEAR_ARRAY[PES_ATTRIBUTE_CONTAINER]
			l_entity: ENTITY
			l_container: PES_ATTRIBUTE_CONTAINER
			l_arr: ROSE_LINEAR_ARRAY [ENTITY_ATTRIBUTE_RECORD]
			l_identifier: ENTITY_IDENTIFIER
			l_id: STRING
		do
			if transformation_result /= void then
				prepare_transformation_result (transformation_result)
			end

			-- Send back attributes for the source entities
			if identifiers_for_source /= Void then
				from
					i := transformation_result.chains.lower
				until
					i > transformation_result.chains.upper
				loop
					l_cache := transformation_result.chains.item (i).cached_attributes
					l_entity := transformation_result.chains.item (i).entity
					if l_cache = Void then
						create l_cache.make_empty
					end
					from
						k := identifiers_for_source.lower
						j := l_cache.upper
						l_cache.conservative_resize (l_cache.lower, l_cache.upper.max (l_cache.lower) + identifiers_for_source.count)
					until
						k > identifiers_for_source.upper
					loop
						l_id := identifiers_for_source.item (k)
						l_identifier := l_entity.context.identifier_from_name (l_id)
						if l_identifier /= Void and then not (cache_desc_name and then l_entity.name_identifier_keys.has (l_id)) then
							l_arr := l_identifier.key_history (l_entity)
							create l_container.make (l_arr, l_identifier.handle)
							j := j + 1
							l_cache.put (l_container, j)
						end
						k := k + 1
					end
					if j > 0 and then l_cache.upper > j then
						l_cache.restrict (1, j)
					elseif j = 0 then
						l_cache := Void
					end
					transformation_result.chains.item (i).set_cached_attributes (l_cache)
					i := i + 1
				end
			end
		end

	restore_result (a_api: PES_CLIENT_API)
			-- Restore references which were removed before sending
			-- and put all transmitted objects into their corresponding caches
		local
			l_stored_recipe: ROSE_RECIPE
			l_stored_request: PES_STORED_REQUEST_RECIPE
		do
			l_stored_request ?= a_api.requests.item_thread_safe (query_id)
			check
				integrity: a_api.requests.item_thread_safe (query_id) /= Void implies l_stored_request /= Void
			end
			if l_stored_request /= Void and then transformation_result /= Void then
				l_stored_recipe := l_stored_request.recipe
				-- Restore recipe
				transformation_result.set_recipe (l_stored_recipe)

				-- Restore transformations
				restore_helper (transformation_result.chains, l_stored_recipe.transformation)
			end

			Precursor (a_api)
		ensure then
			recipe_set: transformation_result.recipe /= Void
		end

feature {NONE} -- Implementation

	restore_helper (a_chains: ROSE_LINEAR_ARRAY[ROSE_TRANSFORMATION_CHAIN]; a_transformation: ROSE_TRANSFORMATION)
			-- Recursively restore transformations
		require
			chains_not_void: a_chains /= Void
		local
			i, j: INTEGER
			l_chain: ROSE_TRANSFORMATION_CHAIN
			l_current_attribute: PES_ATTRIBUTE_CONTAINER
			l_current_identifier: ENTITY_IDENTIFIER
			l_current_entity: ENTITY
		do
			from
				i := a_chains.lower
			until
				i > a_chains.upper
			loop
				l_chain := a_chains.item (i)
				l_chain.set_transformation (a_transformation)

				l_current_entity := l_chain.entity
				-- Add attributes to cache
				if l_chain.cached_attributes /= Void then
					from
						j := l_chain.cached_attributes.lower
					until
						j > l_chain.cached_attributes.upper
					loop
						l_current_attribute := l_chain.cached_attributes.item (j)
						check
							current_attribute_not_void: l_current_attribute /= Void
						end
						l_current_identifier := l_current_entity.context.identifier (l_current_attribute.attribute_id)
						check
							current_identifier_not_void: l_current_identifier /= Void
						end
						l_current_identifier.add_to_cache (l_current_entity, l_current_attribute.attributes)
						j := j + 1
					end
				end

				-- Add entity to cache
				l_current_entity.context.handler (l_current_entity.category_id).cache_entity (l_current_entity)

				if l_chain.children /= Void then
					restore_helper (l_chain.children, a_transformation.next)
				end
				i := i + 1
			end
		ensure
			transformation_set: a_chains.for_all (agent (a_transformation_chain: ROSE_TRANSFORMATION_CHAIN): BOOLEAN do Result := a_transformation_chain.parent /= Void implies a_transformation_chain.parent.transformation /= Void end)
		end

	prepare_transformation_result (a_result: ROSE_TRANSFORMATION_RESULT)
			-- remove references from `a_result'
		require
			result_not_void: a_result /= Void
		local
			i: INTEGER
		do
			a_result.set_recipe (void)
			from
				i := 1
			until
				a_result.chains = void or else i > a_result.chains.count
			loop
				if  a_result.chains.item (i).children /= void then
					prepare_chain (a_result.chains.item (i))
				end

				i := i + 1
			end
		end

	prepare_chain (a_chain: ROSE_TRANSFORMATION_CHAIN)
			-- remove all ROSE_TRANSFORMATION references from `a_chain'
			-- And cache attribute records according to the recipe display settings and `cache_desc_name'
		require
			chain_not_void: a_chain /= Void
		local
			i: INTEGER
		do
			a_chain.cache_attribute_records (cache_desc_name)
			from
				i := 1
			until
				a_chain.children = Void or else
				i > a_chain.children.count
			loop
				prepare_chain (a_chain.children.item (i))
				i := i + 1
			end
			a_chain.set_transformation (void)
		end

end
