note

	description:

		"Controller Class for PES (Persistent Entity Server)"

	documentation: "http://prism/modules/wacko/wakka.php?wakka=PESServerdocumentation&v=1969"
	author: "Ueli Peter"
	last_editor: "$Author: c4 $"
	date: "$Date: 2012/02/29 $"
	revision: "$Revision: #14 $"
	path: "$File: //rose/source/kernel/entities/pes/pes_server.e $"

class PES_SERVER [G->PES_CONTEXT_AWARE create make end]

inherit

	ROSE_CS
		redefine
			set_state
		end

create

	make_pes_server

feature {NONE} -- Initialization

	make_pes_server (a_port: INTEGER; a_refresh: BOOLEAN; a_standalone: BOOLEAN; a_contexts: STRING) is
			-- Listen on `a_port'.
			-- `a_refresh': should the server refresh its cache?
			-- `a_standalone': server in standalone mode (without portserver) for debugging?
			-- `a_contexts': Context names which should be loaded. For example: "RESEARCH,GLOBAL CLIENT,STANDARD AND POORS".
		require
			a_contexts_not_void: a_contexts /= Void
		local
			l_folder: attached STRING
		do
			l_folder := Global.app.program_folder
			if attached {ROSE_DATE} Global.persist.retrieved (Global.formatter.composed_path (<<l_folder, Email_alert_storable_name>>)) as l_date_time then
				email_alert_time := l_date_time
			end
			contexts := a_contexts.split (',')
			contexts.compare_objects
			represents_research := contexts.has ("RESEARCH") or contexts.has ("MODEL")
			arguments := "-contexts=" + a_contexts
			version_count := 0
			
			create state_mutex.make
			
			if a_refresh then
				represented_db := db_servers.rose_entity_foreground_server.definition.path.deep_twin
				latest_db_version := db_servers.rose_entity_foreground_server.definition.path.deep_twin
				set_state (State_updating)
				refresh := True
			else
				set_state (State_latest)
				latest_db_version := "no db loaded"
				refresh := False
			end

			make_server (a_port, a_refresh, a_standalone)

		end


feature -- Access

	state_transition is
			-- Check current state and change to another state if needed
			-- this feature represents the logic of the server
			-- look at the state_transition_diagram in the documentation
		local
			l_retry: BOOLEAN
		do
			if l_retry then
				adjust_sleep_interval
				sleep (sleep_interval)
			else
				-- check whether new Version of the Rose Foreground is available
				check_db

				-- check for context updates
				message_worker.check_for_updates

				inspect
					state
				when State_updating then
					-- update is now finished -> tell the port server that this server is ready
					change_port (arguments)
					set_state (State_latest)
				when State_latest then

					if not refresh or (represents_research and new_db) then
						-- new db is available
						if not refresh then
							output(Global.interpreted("($now) new db available%N"),Void)
						end

						set_state (State_deferred)
						if standalone then
							output(Global.interpreted("($now) new instance can not be requested as we are running stand-alone%N"),Void)
						else
							output(Global.interpreted("($now) send port server message to start new instance%N"),Void)
							start_new_instance
						end
					end
				when State_deferred then
					if child_ready then
						if not standalone then
							-- do state transition only when not in debug mode
							output(Global.interpreted("($now) child is ready%N"),Void)
							set_state (State_outdated)
						end
					end
				when State_outdated then
					if (version_count > Max_instances) then
						-- we need this to make sure, that not to many instances are running at the same time
						set_state (State_shuting_down)
					elseif pers_server.no_clients then
						-- when no client is connected to the server, we wait a certain amount of time until we kill the program
						set_timer
						set_state (State_waiting_for_timeout)
					else
						output (Global.interpreted ("($now) there are still clients connected%N"), Void)
					end
				when State_waiting_for_timeout then
					if not pers_server.no_clients then
						set_state (State_outdated)
					elseif timeout then
						set_state (State_shuting_down)
					end
				when State_shuting_down then
				end
			end
			sleep_interval := 0
		rescue
			l_retry := True
			send_time_conditional_alert ("{PES_SERVER}.state_transition")
			retry
		end

	contexts: LIST [STRING]
			-- List of context ids for this context.

feature {ROSE_CS_WORKER} -- Implementation

	set_state (a_state: INTEGER) is
			-- Set `state' to `a_state', thread safe.
		do
			state_mutex.lock
			state := a_state

			display_state (a_state)

			state_mutex.unlock
		end

	display_state (a_state: INTEGER) is
			-- Output description of `a_state'.
		require
			a_state_valid: is_valid_state (a_state)
		do
			inspect
				a_state
			when State_deferred then
				output (Global.interpreted("($now) State(" + port.out + "): Server is deferred%N"), Void)
			when State_updating then
				output (Global.interpreted("($now) State(" + port.out + "): Server is updating%N"), Void)
			when state_shuting_down then
				output (Global.interpreted("($now) State(" + port.out + "): Server is shuting down%N"), Void)
			when State_outdated then
				output (Global.interpreted("($now) State(" + port.out + "): Server is outdated%N"), Void)
			when State_latest then
				output (Global.interpreted("($now) State(" + port.out + "): Server is latest%N"), Void)
			when State_waiting_for_timeout then
				output (Global.interpreted("($now) State(" + port.out + "): state_waiting_for_timeout%N"), Void)
			end
			if latest_db_version /= Void then
				output (Global.interpreted("($now) State(" + port.out + "): newest db state: " + latest_db_version + " %N"), Void)
			end
		end

	check_db is
			-- check whether new version of the foreground is available and increase it
		local
			l_def: DATABASE_DEFINITION
			l_retry: INTEGER
		do
			if l_retry < 2 then
				db_servers.pool.refresh
				l_def := db_servers.pool.definition ("rose_entity_foreground")
				if not l_def.path.is_equal (latest_db_version) then
					-- new version is available
					version_count := version_count + 1
					latest_db_version := l_def.path.deep_twin
				end
			end
		ensure
			valid_count: version_count >= old version_count
		rescue
			l_retry := l_retry + 1
			reconnect
			retry
		end

	set_timer is
			-- create `timeout_time' with now
			-- add some minutes to it
		do
				create timeout_time.make_now
				timeout_time.minute_add (Timeout_minutes)
		end

	timeout: BOOLEAN is
			-- is `timeout_time' in the past?
		local
			l_now: ROSE_TIME
		do
			create l_now.make_now
			if timeout_time < l_now then
				Result := True
			end
		end

	new_db: BOOLEAN is
			-- is new DB available?
		do
			Result := version_count /= 0
		end

	child_ready: BOOLEAN is
			-- check whether the port server changed the port already
			-- do this by looking for the pes_server in the db
		do
			Result := server_port_from_port_server /= port
		end

feature -- Transition states

	State_updating: INTEGER is 1
 			-- Server is updating entities, etc.

	State_latest: INTEGER is 3
			-- `Current' represents newest version of rose_entity_foreground

	State_deferred: INTEGER is 4
			-- There is a newer DB online. Start a new instance and wait until it is ready.

	State_outdated: INTEGER is 5
			-- There is another instance which represents a newer version of rose_entity_foreground

	State_waiting_for_timeout: INTEGER is 6
			-- Wait until `timeout_time' to exit

	is_valid_state (a_state: INTEGER): BOOLEAN is
			-- <Precursor>
		do
			Result := (a_state = State_updating or a_state = State_latest or a_state = State_shuting_down or
				a_state = State_deferred or a_state = State_outdated or a_state = State_waiting_for_timeout)
		ensure then
			definition: Result = (a_state = State_updating or a_state = State_latest or a_state = State_shuting_down or
				a_state = State_deferred or a_state = State_outdated or a_state = State_waiting_for_timeout)
		end

feature {NONE} -- Implementation

	refresh: BOOLEAN
		-- Does this server load the cache?

	represents_research: BOOLEAN
		-- Does this PES server provide the research context?

	timeout_time: ROSE_TIME
		-- Time at which we shutdown

	Timeout_minutes: INTEGER is 15
		-- Minutes to wait before shutting down

	latest_db_version: STRING
			-- Path of the latest version of rose_entity_foreground

	version_count: INTEGER
		-- Number of switches of rose_entity_foreground since this instance started

	message_worker_type: PES_WORKER [G] is
			-- Type of message worker thread for this instance
		do
		end

	Max_instances: INTEGER is 10
			-- Maximum number of instances permitted to run at the same time

	sleep_interval: INTEGER
			-- How long (in milliseconds) to sleep after an exception

	adjust_sleep_interval is
			-- Increase `sleep_interval' until a maximum is reached
		do
			inspect
				sleep_interval
			when 0 then
				sleep_interval := 5_000
			when 5_000 then
				sleep_interval := 10_000
			when 10_000 then
				sleep_interval := 20_000
			when 20_000 then
				sleep_interval := 40_000
			when 40_000 then
				sleep_interval := 80_000
			when 80_000 then
				sleep_interval := 120_000
			when 120_000 then
				-- stay at maximum
			end
		end

	Email_alert_seconds: INTEGER_64 is 5_400
			-- One-and-a-half hours, which must expire before we send a second email alert

	email_alert_time: ROSE_DATE
			-- Time at which last email alert was sent

	Email_alert_storable_name: STRING is "last_pes_alert.sto"
			-- Name of file into which time of last email alert is stored

	send_time_conditional_alert (a_routine: STRING) is
			-- Send email warning about exception in `a_routine', if not previous sent since `Email_alert_time' seconds.
		require
			a_routine_not_void: a_routine /= Void
		local
			l_folder: attached STRING
		do
			if email_alert_time = Void or else email_alert_time.relative_duration (create {ROSE_DATE}.make_now).seconds_count > Email_alert_seconds then
				send_email_alert (a_routine)
				create email_alert_time.make_now
				l_folder := Global.app.program_folder
				Global.persist.store (email_alert_time, Global.formatter.composed_path (<<l_folder, Email_alert_storable_name>>))
				output (Global.interpreted ("($now) Exception handled in $1!%N $2"), <<a_routine, last_exception_explanation>>)
			else
				output (Global.interpreted ("($now) Exception handled in $1!%N $2"), <<a_routine, exception_trace>>)				
			end
		end

	send_email_alert (a_routine: STRING) is
			-- Send email warning about exception in `a_routine'.
		require
			a_routine_not_void: a_routine /= Void
		local
			l_content: ROSE_MAIL_CONTENT
			l_smtp: ROSE_MAIL
			l_retried: BOOLEAN
		do
			if not l_retried then
				create l_content.make
				l_content.to.add (config.value_default ("support", "SoftwareTechandInfTools@axarosenberg.com"))
				l_content.set_subject ("Exception in persistent entity server")
				l_content.set_text (a_routine + " suffered an exception:%N%N")
				l_content.append_text ("Delivery path is " + global.app.delivery.path + "%N")
				l_content.append_text (last_exception_explanation)
				create l_smtp.make_smtp
				l_smtp.send (l_content)
			end
		rescue
			if not l_retried then
				l_retried := True
				retry
			end
		end

		
invariant

	contexts_not_void: contexts /= Void

end

