note
	description:
		"Process metrics context"
	author: "Zoran Simic"
	last_editor: "$Author: jq $"
	date: "$Date: 2015/04/10 $"
	revision: "$Revision: #86 $"
	path: "$File: //rose/source/kernel/pm/pm_context.e $"

class PM_CONTEXT

inherit
	ROSE_SINGLETON
	DATABASE_SERVER
	ROSE_SHARED_GLOBAL
		export
			{ANY} fail
		end

create {PM_AWARE}
	make

create {PM_CONTEXT}
	make_foreign

feature {NONE} -- Initialization

	make is
			-- Create the context.  Only one is allowed.
		do
			check_instance_unicity
			make_foreign(servers.process_metrics_server)
		end

	make_foreign(a_server: like server)
			-- The current class is the only one that directly uses this feature.  It is to allow us access to other (foreign) PM_CONTEXTs.
		do
			server := a_server
			is_available := server.make_succeeded
			create process_ids.make (0)
			create process_names.make (0)
			create resource_types.make (0)
			create metrics.make (Current)
			create instances.make (Current)
			create runs.make (Current)
			create message_buffer.make
			create metrics_tests
			create instance_stack.make_default
			message_buffer.open (Void)
			error.add_broadcast (message_buffer)
			redirect_log
			-- register IO and CPU metrics for all databases by default
			-- can be turned off at application level as required
			register_all_databases
		end

feature -- Access

	production_process_metrics: PM_CONTEXT is
			-- Access to the production process metrics context
		local
			l_node_server: PAIR [STRING, STRING]
			l_path: STRING
			l_main_site_cfg_path: STRING
			l_config: ROSE_CONFIGURATION
			l_server: SQL_DATA_SERVER
			l_def: ROSE_DATABASE_DEFINITION
			l_logical_predicate: COLLECTING_PREDICATE [TUPLE [server: STRING; database: STRING]]
		do
			if global.app.is_running_in_production then
				Result := Current
			else
				l_main_site_cfg_path := Global.hardcoded_path ("\\axaros.com\shares\production\site.cfg")
				create l_config.make_default
				Global.Config_manager.load (l_config, Void, l_main_site_cfg_path)
				l_path := l_config.value ("/path/repository")
				if l_path /= Void and then l_path.occurrences ('.') = 1 then
					l_node_server := global.formatter.split_on_first (l_path, ".", Void)
					if l_node_server.first.has (':') then
						create l_def.make_anonymous ("PM_CONTEXT.production_process_metrics", Global.Convention.Db_type_sql, global.formatter.split_on_first (l_node_server.first, ":", Void).second, l_node_server.second)
					else
						create l_def.make_anonymous ("PM_CONTEXT.production_process_metrics", Global.Convention.Db_type_sql, l_node_server.first, l_node_server.second)
					end
					l_server := servers.pool.get_Server_from_definition (l_def)
					create l_logical_predicate.make (l_server, "select server, [database] from databases where logical = 'process_metrics'")
					if not l_logical_predicate.success then
						error.raise ("Could not create `l_logical_predicate'")
					end
					l_logical_predicate.retrieve (Void, Void)
					if not l_logical_predicate.success then
						error.raise ("Could not execute `l_logical_predicate'")
					end
					if l_logical_predicate.array.count /= 1 then
						error.raise ("Return count from `l_logical_predicate' is not 1")
					end
					Result := foreign_context (l_logical_predicate.array.item (1).server, l_logical_predicate.array.item (1).database)
				end
			end
		end

	server: SQL_DATA_SERVER
			-- Process metrics server

	process_ids: HASH_TABLE [PM_PROCESS, INTEGER]	-- Processes by id
	process_names: HASH_TABLE [PM_PROCESS, STRING]	-- Processes by name

	resource_types: HASH_TABLE [PM_RESOURCE_TYPE, INTEGER]	-- Resource types by id

	instances: PM_CONTEXT_INSTANCES	-- Instances cache
	runs: PM_CONTEXT_RUNS			-- Runs cache
	metrics: PM_CONTEXT_METRICS		-- Metrics cache
	metrics_tests: PM_METRICS_TESTS	-- Metric tests cache

	is_nested_runs_allowed: BOOLEAN
			-- Are nested process metrics runs allowed?

	message_buffer: ROSE_OUTCHANNEL_BUFFER
			-- Buffer for process success/failure notifications

	current_instance: PM_INSTANCE
			-- Instance currently operating
			-- Use 'start' or 'select_instance'
		do
			if not instance_stack.is_empty then
				Result := instance_stack.item
			end
		end

	is_available: BOOLEAN
			-- Is the process metrics available?
			-- This is False when 'server' could not be connected

	is_disabled: BOOLEAN is
			-- Is process monitoring disabled?
			-- Process monitoring can be disabled via the configuration
			-- Use this if you want to run a process that would usually
			-- produce metrics etc., but you don't want it to do so for this particular run
		do
			Result := config.boolean ("$system/process_monitor/disabled", False)
		end

	is_resource_access_monitoring_disabled: BOOLEAN is
			-- Is resource access monitoring disabled?
			-- Database and file access monitoring can be enabled via the configuration
		do
			Result := config.boolean ("$system/process_monitor/resource_access_disabled", False)
		end

	max_log_length: INTEGER is
			-- Maximum log length to send in report email
		do
			Result := config.integer ("$system/process_monitor/buffer", 4096).max (1024)
		end

	statistics_lookback: INTEGER is
			-- Should we look at previous runs?
		do
			Result := config.integer ("$system/process_monitor/statistics_lookback", 10)
		end

	is_running: BOOLEAN is
			-- Is an instance currently in progress?
		do
			Result := current_instance /= Void and then current_instance.is_running
		end

	process_by_id (a_id: INTEGER): PM_PROCESS is
			-- Process identified by 'a_id'
		do
			Result := process_ids.item (a_id)
		end

	process_by_name (a_name: STRING): PM_PROCESS is
			-- Process identified by 'a_name'
		local
			l_email_address : STRING
		do
			if a_name = Void then
				-- Do nothing
			elseif a_name.is_integer then
				Result := process_by_id (a_name.to_integer)
			else
				Result := process_names.item (a_name.as_lower)
				if Result = Void then

					-- Don't automatically add batch_rc to the process table as this will copy rc_production_support
					if global.env.is_batch_current_user then
						l_email_address := ""
					else
						l_email_address := Global.app.user.email_address
					end

					sql_create_process.execute (Void, <<a_name, l_email_address, l_email_address, "Auto-created by PM_CONTEXT">>)
					if not sql_create_process.success then
						fail ("process_by_name", "Could not create new process $1", <<a_name>>)
					end
					server.commit
					if not server.success then
						fail ("process_by_name", "Could not commit", <<a_name>>)
					end
					refresh
					Result := process_by_name (a_name)
				end
			end
		end

	running_processes: ROSE_LINEAR_ARRAY [PM_PROCESS] is
			-- Processes with at least one instance currently marked as "running"
		do
			sql_get_running_process_ids.retrieve (Void, Void)
			Result := sql_get_running_process_ids.array
		end

	running_instances: ROSE_LINEAR_ARRAY [PM_INSTANCE] is
			-- Instances currently marked as "running"
		do
			sql_get_running_process_partitions.retrieve (Void, Void)
			Result := sql_get_running_process_partitions.array
		end

	metric_type (a_id: INTEGER): PM_METRIC_TYPE is
			-- Metric type with 'a_id'
		do
			Result := metrics.types.item (a_id)
		end

	resource_type_by_id (a_id: INTEGER): PM_RESOURCE_TYPE is
			-- Resource type corresponding to 'a_id'
		do
			Result := resource_types.item (a_id)
		end

	run_by_id (a_id: INTEGER): PM_INSTANCE_RUN is
			-- Run with 'a_id' if any
		do
			Result := runs.run_by_id (a_id)
		end

	run_by_process_run_id(a_run_id: STRING): PM_INSTANCE_RUN
		-- Run with the 'run_id' found in the string 'a_run_id'.  This string is found in the 'process' table
		-- and may (or may not) be of the form: "PM:server.database:run_id" (see SUPPLIER_PROCESS_ID_MANAGER.run_id).
		-- If you are in a test environment, the "server.database" will be used to search the appropriate process_context.
		-- In production, "server.database" should be the same one that is being used by this class.
		-- If no 'run_id' can be determined, Void is returned.
	local
		l_pos: INTEGER
		l_run_id_string: STRING
		l_node: STRING
		l_database: STRING
		l_foreign_context: like Current
		l_text: STRING
		l_strings: SEQUENCE[STRING]
	do
		if a_run_id/=Void
		then
			l_pos := a_run_id.substring_index("PM:", 1)
			if l_pos>0
			then
				l_pos := l_pos + 3		-- Point to the character just after the colon (if any).
				if l_pos<a_run_id.count
				then
					l_text := a_run_id.substring(l_pos, a_run_id.count)
					l_strings := global.formatter.split(l_text, ".:", Void)
					if l_strings.count=3
					then
						l_strings.finish
						l_run_id_string := l_strings.item
						if l_run_id_string.is_integer
						then
							if global.app.is_running_in_production
							then
								Result := run_by_id(l_run_id_string.to_integer)
							else
								-- Note that although we know the node and database there is no guarantee
								-- that it hasn't been moved.  That is why we don't do this in production.
								l_strings.start
								l_node := l_strings.item
								l_strings.forth
								l_database := l_strings.item
								if l_node.count/=0 and l_database.count/=0
								then
									l_foreign_context := foreign_context(l_node, l_database)
									if l_foreign_context/=Void and then l_foreign_context.is_available
									then
										Result := l_foreign_context.run_by_id(l_run_id_string.to_integer)
									end
								end
							end
						end
					end
				end
			end
		end
	end

	instance_by_id (a_id: INTEGER): PM_INSTANCE is
			-- Instance with 'a_id' if any
		do
			Result := instances.instance_by_id (a_id)
		end

	runs_by_message (a_class_name, a_feature_name, a_message: STRING): ROSE_LINEAR_ARRAY [PM_INSTANCE_RUN] is
			-- All runs that had `a_message' in one of their posted messages
		require
			a_message_not_void: a_message /= Void
		local
			l_messages: ROSE_LINEAR_ARRAY [PM_MESSAGE]
		do
			create Result.make_empty
			if a_class_name = Void or a_feature_name = Void then
				sql_get_messages.retrieve (Void, <<a_message>>)
				if sql_get_messages.success and not sql_get_messages.array.is_empty then
					l_messages := sql_get_messages.array
				end
			else
				sql_get_messages_with_class_name_and_feature_name.retrieve (Void, <<a_message,a_class_name,a_feature_name>>)
				if sql_get_messages_with_class_name_and_feature_name.success and not sql_get_messages_with_class_name_and_feature_name.array.is_empty then
					l_messages := sql_get_messages_with_class_name_and_feature_name.array
				end
			end
			if l_messages /= Void then
				l_messages.do_all (agent add_run_from_message (?, Result))
			end
		ensure
			result_not_void: Result /= Void
		end

	run_by_message (a_class_name, a_feature_name, a_message: STRING): PM_INSTANCE_RUN is
			-- Last run that had 'a_message' in one of its posted messages
		require
			valid_message: a_message /= Void
		local
			l_runs_by_message: like runs_by_message
		do
			l_runs_by_message := runs_by_message (a_class_name, a_feature_name, a_message)
			if not l_runs_by_message.is_empty then
				Result := l_runs_by_message.first
			end
		end

	instance_by_key (a_process: PM_PROCESS; a_partition_key: STRING; a_filedate: ROSE_DATE): PM_INSTANCE is
			-- Instance identified by 'a_process', 'a_partition_key' and 'a_filedate'
		do
			Result := instances.instance_by_key (a_process, a_partition_key, a_filedate)
		end

	current_metric_value (a_name: STRING): DOUBLE is
			-- Current value of produced metric of type 'a_name'
		do
			Result := metrics.current_metric_value (a_name)
		end

	last_successful_file_date (a_process_name, a_partition_key: STRING): ROSE_DATE is
			-- Last succesful file date registered for 'a_process_name' and 'a_partition_key'
		local
			l_process: PM_PROCESS
			l_instance: PM_INSTANCE
		do
			l_process := process_by_name (a_process_name)
			if l_process /= Void then
				l_instance := l_process.most_recent_successful_instance (a_partition_key)
				if l_instance /= Void then
					Result := l_instance.file_date
				end
			end
		end

	last_successful_run_time (a_process_name, a_partition_key: STRING): ROSE_DATE is
			-- The last successful run time for 'a_process_name' and 'a_partition_key'
		local
			l_process: PM_PROCESS
			l_instance: PM_INSTANCE
			l_runs: PM_INSTANCE_RUN_COLLECTION
		do
			l_process := process_by_name (a_process_name)
			if l_process /= Void then
				l_instance := l_process.most_recent_successful_instance (a_partition_key)
				if l_instance /= Void then
					l_runs := l_instance.runs
					if l_runs /= Void then
						Result := l_runs.last_successful_start_time
					end
				end
			end
		end

	is_print_termination_report : BOOLEAN
			-- The termination report is emailed by send_termination_report
			-- Should the termination report also be printed?

	test_case_instance_runs (a_changelist: INTEGER): DS_ARRAYED_LIST [attached PM_INSTANCE_RUN]
			-- Instance runs of test case executions associated with
			-- Perforce `a_changelist'
		local
			l_predicate: INTEGER_PREDICATE
			l_run_ids: ROSE_LINEAR_ARRAY [INTEGER]
			i, l_upper: INTEGER
			l_run: PM_INSTANCE_RUN
		do
			l_predicate := sql_test_result_run_ids_for_changelist
			l_predicate.retrieve (Void, <<a_changelist>>)
			if not l_predicate.success then
				error.abort ("test_case_instance_runs", "Failed to retrieve mapping between changelist $1 and test case run-ids", <<a_changelist>>)
			else
				l_run_ids := l_predicate.array
				create Result.make (l_run_ids.count)
				from
					i := l_run_ids.lower
					l_upper := l_run_ids.upper
				until
					i > l_upper
				loop
					l_run := run_by_id (l_run_ids.item (i))
					if l_run /= Void then
						Result.force_last (l_run)
					end
					i := i + 1
				end
			end
		end

feature {PM_TEST_SUITE} -- Basic operations

	enable_nested_execution is
			-- Allow process metrics system to have nested runs
		do
			is_nested_runs_allowed := True
		end

	disable_nested_execution is
			-- Disallow process metrics system to have nested runs
		do
			is_nested_runs_allowed := False
		end

feature -- Basic operations

	refresh is
			-- Refresh 'process_ids'.
		local
			l_process: PM_PROCESS
			i: INTEGER
		do
			process_ids.clear_all
			process_names.clear_all
			resource_types.clear_all
			if is_available then
				sql_get_process_ids.retrieve (Void, Void)
				if sql_get_process_ids.array /= Void then
					from
						i := sql_get_process_ids.array.lower
					until
						i > sql_get_process_ids.array.upper
					loop
						l_process := sql_get_process_ids.array.item (i)
						process_ids.force (l_process, l_process.process_id)
						process_names.force (l_process, l_process.name.as_lower)
						i := i + 1
					end
				end
				sql_get_resource_types.retrieve (Void, Void)
				if sql_get_resource_types.array /= Void then
					from
						i := sql_get_resource_types.array.lower
					until
						i > sql_get_resource_types.array.upper
					loop
						resource_types.put (sql_get_resource_types.array.item (i), sql_get_resource_types.array.item (i).id)
						i := i + 1
					end
				end
			end
			metrics.refresh
		end

	start (a_process_name, a_partition_key: STRING; a_filedate: ROSE_DATE) is
			-- Start a new instance run for 'a_process_name', 'a_partition_key' and 'a_filedate'
			-- error signaled if the operation was not possible.
		require
			not_running: is_nested_runs_allowed or not is_running
			no_error: not error.signaled
		do
			if not is_disabled then
				if is_running and not is_nested_runs_allowed then
					fail ("start", "Instance '$1' is already in progress, please respect this feature's contract (see 'require' clause of this routine)", <<current_instance>>)
				end
				select_instance (a_process_name, a_partition_key, a_filedate)
				require_available
				if current_instance /= Void then
					start_current_instance
					if current_instance /= Void and then current_instance.is_running and then not global.app.is_running_in_estudio then
						create duration_check.make (current_instance.current_run)
					end
				else
					fail ("start", "Could not create instance for process '$1' in database '$2'", <<a_process_name, server>>)
				end
			end
		ensure
			error_signaled: (not is_disabled implies is_running) xor error.signaled
		end

	terminate is
			-- Terminate and commit current instance run.
		require
			can_terminate: not is_disabled implies is_running
		do
			if not is_disabled then
				require_available
				if not is_running then
					fail ("terminate", "There's no run to terminate, please respect this feature's contract (see 'require' clause of this routine)", Void)
				end
				if not is_resource_access_monitoring_disabled and then
					current_instance /= Void and then
					current_instance.current_run /= Void then
					resource_access.persist_data_server_accesses (servers.pool, server, current_instance.current_run.run_id)
					servers.pool.record_metrics
				end
				if is_current_run_duration_too_short (lower_standard_deviation_for_duration_check, lookback_count_for_duration_check) then
					warn ("terminate", "current run duration too short: $1", <<current_run_duration_message>>)
				elseif is_current_run_duration_too_long (upper_standard_deviation_for_duration_check, lookback_count_for_duration_check) then
					warn ("terminate", "current run duration too long: $1", <<current_run_duration_message>>)
				end
				current_instance.terminate (1)
				Global.app.set_generated_metrics (True)
				duration_check := Void
			end
			if not instance_stack.is_empty then
				instance_stack.remove
			end
		end

	clear_resource_access_tracking is
			-- Clear out any resource access that's happened up until now
		do
			resource_access.clear_resource_access_tracking (servers.pool)
		end

	terminate_failed is
			-- Terminate 'current_instance' and mark it failed
		do
			if is_running then
				current_instance.terminate (-1)
			elseif not is_disabled and then current_instance /= Void then
				send_termination_report
			end
			if not instance_stack.is_empty then
				instance_stack.remove
			end
		end

	persist_metrics_for_run_id (a_run_id: INTEGER) is
			-- Persist metrics for 'a_run_id'
			-- This is used to facilitate gathering metrics from among various out of process executions
			-- This should only be called by systems that do not have a current run active, the parent process owns it
		require
			valid_run_id: run_by_id (a_run_id) /= Void and then not run_by_id (a_run_id).is_terminated
		do
			metrics.write_to_database_for_run (a_run_id)
		end

	select_instance (a_process_name, a_partition_key: STRING; a_filedate: ROSE_DATE) is
			-- Associate 'current_instance' to 'a_process_name', 'a_partition_key', and 'a_filedate'.
			-- Use this if you need to look at 'current_instance' before you start it.
		require
			not_running: is_nested_runs_allowed or not is_running
		local
			l_process: PM_PROCESS
			l_instance: PM_INSTANCE
		do
			l_process := process_by_name (a_process_name)
			if l_process /= Void then
				l_instance := instances.forced_instance (l_process, a_partition_key, a_filedate)
				if l_instance = Void then
					fail ("select_instance", "Cannot start 'instance'", Void)
				else
					instance_stack.force (l_instance)
				end
			else
				fail ("select_instance", "No process '$1' defined in database '$2'", <<a_process_name, server>>)
			end
		end

	start_current_instance is
			-- Start a new instance run for 'current_instance'
			-- error signaled if the operation was not possible.
		require
			not_already_running: (is_nested_runs_allowed or not is_running) and current_instance /= Void
			no_error: not error.signaled
		do
			if not is_disabled then
				require_available
				if current_instance /= Void then
					initiate_run
				else
					fail ("start_current_instance", "Cannot start 'current_instance': it was not selected", Void)
				end
			end
		ensure
			error_signaled: (not is_disabled implies is_running) xor error.signaled
		end

	commit is
			-- Commit changes to underlying database server.
		do
			if is_available then
				server.commit
			end
		end

	add_metric (a_name: STRING; a_value: DOUBLE) is
			-- Add a metric of type 'a_name' and 'a_value' to current process instance.
		do
			metrics.add_metric (a_name, a_value)
			if not warned_on_current_run_duration_too_long and then
				is_current_run_duration_too_long (upper_standard_deviation_for_duration_check, lookback_count_for_duration_check) then
				warned_on_current_run_duration_too_long := True
				warn ("add_metric", "current run duration too long: $1", <<current_run_duration_message>>)
				if current_instance.process.failure_recipients /= Void and then not current_instance.process.failure_recipients.is_empty then
					mail.send_body ("[PM] !!! " + current_instance.process.name + " is taking WAY too long !!!", current_instance.process.failure_recipients, current_run_duration_message)
				end
			end
		end

	is_current_run_duration_too_long (a_standard_deviation: DOUBLE; a_lookback_count: INTEGER): BOOLEAN
			-- Check whether current run duration is too long given `a_standard_deviation', looking back `a_lookback_count' observations.
		require
			a_standard_deviation_non_negative: a_standard_deviation >= 0.0
			a_lookback_count_positive: a_lookback_count > 0
		do
			if duration_check /= Void then
				duration_check.set_statistics_and_duration (a_standard_deviation, a_lookback_count)
				Result := duration_check.is_too_long
			end
		end

	is_current_run_duration_too_short (a_standard_deviation: DOUBLE; a_lookback_count: INTEGER): BOOLEAN
			-- Check whether current run duration is too short given `a_standard_deviation', looking back `a_lookback_count' observations.
			-- Use negative number for `a_standard_deviation' to check for a duration less than a lower standard deviation amount.
		require
			a_standard_deviation_non_positive: a_standard_deviation <= 0.0
			a_lookback_count_positive: a_lookback_count > 0
		do
			if duration_check /= Void then
				duration_check.set_statistics_and_duration (a_standard_deviation, a_lookback_count)
				Result := duration_check.is_too_short
			end
		end

	current_run_duration_message: STRING
			-- Duration status message.
		do
			Result := "%N" + global.app.out + "%N"
			Result.append ("process name:%T" + current_instance.process.name + "%N")
			Result.append ("process id:%T" + current_instance.process.process_id.out + "%N")
			Result.append ("partition:%T" + current_instance.partition.out + "%N")
			Result.append ("instance id:%T" + current_instance.instance_id.out + "%N")
			Result.append ("run id:%T" + current_instance.current_run.run_id.out + "%N")
			Result.append ("%N" + current_instance.current_run.out)
			if duration_check /= Void then
				Result.append (duration_check.message)
			end
		ensure
			result_not_void: Result /= Void
		end

	set_metric_test (a_name: STRING; a_test: FUNCTION [ANY, TUPLE [PM_PROCESS_PARTITION, PM_METRIC], PM_METRIC_TEST_RESULT]) is
			-- Add a metric test for 'a_name'
		do
			metrics.set_metric_test (a_name, a_test)
		end

	run_metric_test (a_name: STRING; a_test: FUNCTION [ANY, TUPLE [PM_PROCESS_PARTITION, PM_METRIC], PM_METRIC_TEST_RESULT]) is
			-- Run a metric test for 'a_name'
		do
			metrics.run_metric_test (a_name, a_test)
		end

	register_source (a_source: PM_METRIC_SOURCE) is
			-- Register 'a_source'.
		do
			metrics.register_source (a_source)
		end

	register_all_databases is
			-- Same as register_source but registers all DBs that are accessed.
			-- Sets a flag which tells PM_CONTEXT_METRICS.put_registered_metrics
			-- to register all databases in ROSE_DATABASE_POOL.accessed
		do
			metrics.enable_register_all_databases
		end

	set_recipient_expression_replacement (a_key, a_value: STRING) is
			-- Extend the key value pairs for recipient expression replacement (ie, u, zone)
		require
			valid_key: a_key /= Void
			valid_value: a_value /= Void
		do
			if recipient_expression_replacement = Void then
				create recipient_expression_replacement.make_empty
			end
			recipient_expression_replacement.extend (create {PAIR [STRING, STRING]}.make (a_key, a_value))
		end

	enable_print_termination_report is
			-- Print termination report in addition to sending it via email
		do
			is_print_termination_report := True
		end

	disable_print_termination_report is
			-- Do not print termination report. Only send via email
		do
			is_print_termination_report := False
		end

	record_changelist_test_result_run_ids_mapping (a_changelist: INTEGER; a_instance_runs: DS_ARRAYED_LIST [attached PM_INSTANCE_RUN])
			-- Record mapping between the run-ids of `a_instance_runs'
			-- of test case executions and the associated
			-- Perforce `a_changelist'.
		require
			a_instance_runs_not_void: a_instance_runs /= Void
		do
			a_instance_runs.do_all (agent record_changelist_test_result_run_id_mapping (a_changelist, ?))
			server.commit
			if not server.success then
				error.abort ("record_changelist_test_result_run_ids_mapping", "Failed to record new mapping between changelist and test case run-ids", Void)
			end
		end

	record_changelist_test_result_run_id_mapping (a_changelist: INTEGER; a_instance_run: PM_INSTANCE_RUN)
			-- Record mapping between the run-id of `a_instance_run'
			-- of a test case execution and the associated
			-- Perforce `a_changelist'.
		require
			a_instance_run_not_void: a_instance_run /= Void
		local
			l_predicate: EXECUTION_PREDICATE
		do
			l_predicate := sql_record_changelist_test_result_run_id_mapping
			l_predicate.execute (Void, <<a_changelist, a_instance_run.run_id>>)
			if not l_predicate.success then
				error.abort ("record_changelist_test_result_run_id_mapping", "Failed to record new mapping between changelist $1 and test case run-id $2", <<a_changelist, a_instance_run.run_id>>)
			end
		end

feature {PM_INSTANCE} -- Implementation

	instance_stack: DS_ARRAYED_STACK [PM_INSTANCE]
			-- Stack of instances to allow nested pm runs

 	foreign_context (a_node, a_database: STRING): like Current is
			-- Returns the PM_CONTEXT in the database given by the arguments.  If no such database exists,
			-- Result.is_available will be False.
		require
			node_not_void: a_node /= Void
			database_not_void: a_database /= Void
		local
			l_server: SQL_DATA_SERVER
			l_key: STRING
			l_def: ROSE_DATABASE_DEFINITION
		do
			if foreign_contexts = Void then
				create foreign_contexts.make(5)
				l_key := server.node.as_lower + "." + server.database.as_lower
				foreign_contexts.put (Current, l_key)		-- Start with ourself.
			end
			l_key := a_node.as_lower + "." + a_database.as_lower
			foreign_contexts.search (l_key)
			if foreign_contexts.found then
				Result := foreign_contexts.found_item
			else
				create l_def.make_anonymous ("PM_CONTEXT.foreign_context", global.convention.db_type_sql, a_node, a_database)
				l_server := servers.pool.get_server_from_definition (l_def)
				create Result.make_foreign( l_server)
				foreign_contexts.force (Result, l_key)
			end
		end

	foreign_contexts: HASH_TABLE[like Current, STRING]		-- All foreign contexts indexed by node.database.

feature {PM_INSTANCE} -- Initialization

	recipient_expression_replacement: ROSE_LINEAR_ARRAY [PAIR [STRING, STRING]]
			-- Collection of expression replacement key value pairs for message recipient lists

	send_termination_report is
			-- send email message termination report
		require
			terminate_report_can_be_computed: not is_disabled and then current_instance /= Void
		local
			l_process: PM_PROCESS
			l_partition: PM_PROCESS_PARTITION
			l_recipient, l_recipients, l_subject, l_buffer: STRING
			l_delimited_recipients: ROSE_DELIMITED_TEXT
			l_content: ROSE_MAIL_CONTENT
			l_failed: BOOLEAN
		do
			l_process := current_instance.process
			if l_process = Void then
				error.signal (generator, "Internal error, can not determine process for process-metrics instance '$1', please contact support team", <<current_instance>>)
				l_recipients := "RC_Tools<rc_tools@axarosenberg.com>"
				l_failed := True
			elseif current_instance.current_run = Void then
				l_failed := True
			else
				l_failed := not current_instance.current_run.is_successful
			end
			if l_failed then
					--| Send an email to failure recipient
				l_recipients := l_process.failure_recipients
				l_recipients := config.value_default ("process_monitor/mail/failure", l_recipients)
				if l_recipients = Void then
					l_recipients := config.value_default ("$system/mail/failure", l_recipients)
				end
				l_subject := Global.formatted ("[PM] Failure: $1 failed!", << current_instance >>)
				if l_recipients = Void or else l_recipients.is_empty then
					if not global.env.is_batch_current_user then
						l_recipients := global.app.user.email_address
					end
					error.signal (generator, "No failure recipient for process '$1'", <<l_process>>)
				end
				if not error.signaled then
					-- Make sure an error is signaled for all failures
					error.signal (generator, "Process instance '$1' failed!", <<current_instance>>)
				end
			else
					--| Send an email to success recipient
				l_recipients := l_process.success_recipients
				l_recipients := config.value_default ("process_monitor/mail/success", l_recipients)
				if l_recipients = Void then
					l_recipients := config.value_default ("$system/mail/success", l_recipients)
				end
				if error.warning_count = 0 then
					l_subject := Global.formatted ("[PM] Success: $1 succeeded.", << current_instance >>)
				else
					l_subject := Global.formatted ("[PM] Success/warn: $1 succeeded (with warnings).", << current_instance >>)
				end
			end
			if recipient_expression_replacement /= Void and then not recipient_expression_replacement.is_empty and then
				l_recipients /= Void and then l_recipients.has ('{') and then l_recipients.has ('}') then
				create l_delimited_recipients.make (l_recipients, ';')
				l_recipients := ""
				from l_delimited_recipients.start until l_delimited_recipients.off loop
					from l_recipient := Void recipient_expression_replacement.start until recipient_expression_replacement.off loop
						if l_delimited_recipients.current_item.as_lower.has_substring ("{" + recipient_expression_replacement.current_item.first.as_lower + "}") then
							if l_recipient = Void then
								l_recipient := l_delimited_recipients.current_item.twin.as_lower
							else
								l_recipient := l_recipient + ";" + l_delimited_recipients.current_item.twin.as_lower
							end
							l_recipient.replace_substring_all ("{" + recipient_expression_replacement.current_item.first.as_lower + "}", recipient_expression_replacement.current_item.second)
						end
						recipient_expression_replacement.forth
					end
					if l_recipient /= Void and then not l_recipient.is_empty then
						l_recipients := l_recipients + l_recipient + ";"
					else
						if not l_delimited_recipients.current_item.has ('{') then
							l_recipients := l_recipients + l_delimited_recipients.current_item + ";"
						end
					end
					l_delimited_recipients.forth
				end
			end

			create l_content.make
			l_content.set_sender_copied (False)

			l_content.set_subject (l_subject)
			l_content.append_text (Global.formatted ("Instance ID  : '$1'%N", <<current_instance.instance_id>>))
			l_content.append_text (Global.formatted ("Monitor      : " + global.deliveries.active.path_to_process_monitor + "%N", Void))

			if current_instance.current_run /= Void then
				-- A run took place, no need to check the concurrency level
				l_content.append_text (Global.formatted ("Run ID       : '$1'%N", <<current_instance.current_run.run_id>>))
				if current_instance.current_run.scheduler_job_id > 0 then
					l_content.append_text (Global.formatted ("Scheduler id : $1 $2%N" , <<current_instance.current_run.scheduler_name,current_instance.current_run.scheduler_job_id.out>>))
				end
				l_content.append_text ("%N")
			elseif current_instance.runs.in_progress.count >= l_process.concurrency then
				l_content.append_text ("%NThis run aborted because " + l_process.concurrency.out + " runs are already in progress for process '")
				l_content.append_text (l_process.name)
				l_content.append_text ("'%N")
				l_content.append_text ("You will have to either wait until the other runs finish, or mark them as finished manually if you know that they crashed for example.%N")
				l_content.append_text ("To mark them as finished or delete the runs currently in progress, please run the process monitor and terminate instance_run id " + current_instance.current_run.run_id.out + "%N")
				l_content.append_text (global.deliveries.active.path_to_process_monitor + "%N")
				l_content.append_text ("%N%N")
			end
			error.flush
			l_buffer := message_buffer.flush_filtered_to_string (message_buffer.v_log)
			if l_buffer /= Void and then l_buffer.count > max_log_length then
				-- FIXME -- I'd like for this to allow the first few frames of the stack to show in their entirety
				l_content.append_text (l_buffer.substring (1, 100 + (max_log_length // 2)) )
				l_content.append_text ("%N%N[...8<... truncated " + (l_buffer.count - max_log_length).out + " characters ...8<...]%N%N")
				l_content.append_text (l_buffer.substring (l_buffer.count - ( 100 + (max_log_length // 2) ), l_buffer.count))
			else
				l_content.append_text (l_buffer)
			end
			l_content.append_text ("%N======== Process monitoring info ========%N")
			if is_running then
				l_content.append_text (metrics.metrics_cached (current_instance))
			else
				l_content.append_text ("No instance started for process '" + current_instance.process.name + "'%N%N")
			end
			l_content.append_text ("%N======== Process Statistics ========%N")
			l_partition := current_instance.partition
			if l_partition /= Void then
				l_content.append_text (Global.formatter.argumented ("Mean Run Time : $1%N", <<l_partition.statistics.mean_run_duration>>))
				l_content.append_text (Global.formatter.argumented ("Run Time Standard Deviation : $1%N", <<l_partition.statistics.run_duration_standard_deviation>>))
				l_content.append_text (Global.formatter.argumented ("Mean Run Periodicity : $1%N", <<l_partition.statistics.mean_run_periodicity>>))
				l_content.append_text (Global.formatter.argumented ("Run Periodicity Standard Deviation : $1%N", <<l_partition.statistics.run_periodicity_standard_deviation>>))
			end
			l_content.append_text ("%N======== Databases used ========%N")
			l_content.append_text (servers.pool.report)
			l_content.append_text ("%N======== Files used ========%N")
			l_content.append_text (file_access_log.report)
			l_content.add_content_attachment ("app.log", Global.app.report)

			if is_print_termination_report then
				print_timestamped ("Printing process metrics termination report%N")
				print( l_content.text )
			end

			if l_recipients /= Void and then not l_recipients.is_empty then
				l_content.to.set (l_recipients)
				mail.send (l_content)

				if mail.last_error /= Void then
					error.signal (generator, "Cannot send email to '$1' for process '$2': $3", <<l_recipients, l_process.name, mail.last_error>>)
				end
			end

		end

feature {NONE} -- Initialization

	initiate_run is
			-- Initiate a process metrics run
		require
			not_already_running: not is_disabled and then current_instance /= Void and then not is_running
		local
			l_process: PM_PROCESS
			l_runs: ARRAY [PM_INSTANCE_RUN]
		do
			error.acknowledge
			l_process := current_instance.process
			l_runs := current_instance.runs.in_progress
			if l_process = Void then
				fail ("initiate_run", "Internal error, can not determine process for process-metrics instance '$1', please contact support team", <<current_instance>>)
			elseif l_runs.count >= l_process.concurrency then
				fail ("initiate_run", "Cannot start a new run for process-metrics instance '$1': configuration imposes that no more than '$2' runs take place simultaneously", <<current_instance, l_process.concurrency>>)
			end
			if not error.signaled then
				if l_process.allowed_runs > 0 and then l_runs.count > l_process.allowed_runs then
					error.warn ("Process-metrics instance '$1' has already ran '$2' times, which is more than the maximum expected", <<current_instance, l_runs.count>>)
				end
				current_instance.start_run
				if not current_instance.is_running then
					fail ("initiate_run", "Could not create an instance run for process '$1' in database '$2'", <<l_process, server>>)
				end
				trace ("Process Name   : '$1'%N", <<l_process.name>>)
				trace ("Partition      : '$1'%N", <<current_instance.partition_key>>)
				trace ("File Date      : '$1'%N", <<current_instance.file_date>>)
				trace ("Instance ID    : '$1'%N", <<current_instance.instance_id>>)
				trace ("Run ID         : '$1'%N", <<current_instance.current_run.run_id>>)
				trace ("Monitor        : $1\tools\process_monitor.exe%N", <<global.app.site.folder>>)
				if current_instance.current_run.scheduler_job_id > 0 then
					trace ("Scheduler ID   : $1 $2%N", <<current_instance.current_run.scheduler_name,current_instance.current_run.scheduler_job_id.out>>)
				end

			end
		end

	mail: ROSE_MAIL is
			-- Used to send email messages
		once
			create Result.make
		ensure
			valid_result: Result /= Void
		end

	resource_access_saved: like resource_access
			-- See 'resource_access'.
		note
			option: transient
		attribute
		end

	resource_access: ROSE_RESOURCE_ACCESS
			-- Access to resource access histories
		do
			if resource_access_saved=Void
			then
				create resource_access_saved
			end

			Result := resource_access_saved
		ensure
			valid_result: Result /= Void
		end

	file_access_log_saved: like file_access_log
			-- See 'file_access_log'.
		note
			option: transient
		attribute
		end

	file_access_log: FILE_ACCESS_LOG
			-- Log for accessed files (input and output)
		do
			if file_access_log_saved=Void
			then
				create file_access_log_saved
			end

			Result := file_access_log_saved
		ensure
			valid_result: Result /= Void
		end

	sql_get_process_ids_saved: like sql_get_process_ids
			-- See 'sql_get_process_ids'.
		note
			option: transient
		attribute
		end

	sql_get_process_ids: COLLECTING_PREDICATE [PM_PROCESS]
			-- SQL statement to read information from 'process' table
		do
			if sql_get_process_ids_saved=Void
			then
				create sql_get_process_ids_saved.make (server, "SELECT a.* FROM process a WITH(NOLOCK)")
			end

			Result := sql_get_process_ids_saved
		end

	sql_get_running_process_ids_saved: like sql_get_running_process_ids
			-- See 'sql_get_running_process_ids'.
		note
			option: transient
		attribute
		end

	sql_get_running_process_ids: COLLECTING_PREDICATE [PM_PROCESS]
			-- SQL statement to read processes that are marked as "still running"
		do
			if sql_get_running_process_ids_saved=Void
			then
				create sql_get_running_process_ids_saved.make (server, "SELECT * FROM process WITH(NOLOCK) where process_id in (select process_id from still_running)")
			end

			Result := sql_get_running_process_ids_saved
		end

	sql_get_running_process_partitions_saved: like sql_get_running_process_partitions
			-- See 'sql_get_running_process_partitions'.
		note
			option: transient
		attribute
		end

	sql_get_running_process_partitions: COLLECTING_PREDICATE [PM_INSTANCE]
			-- SQL statement to read instances that are marked as "still running"
		do
			if sql_get_running_process_partitions_saved=Void
			then
				create sql_get_running_process_partitions_saved.make (server, "SELECT * from still_running")
			end

			Result := sql_get_running_process_partitions_saved
		end

	sql_get_resource_types_saved: like sql_get_resource_types
			-- See 'sql_get_resource_types'.
		note
			option: transient
		attribute
		end

	sql_get_resource_types: COLLECTING_PREDICATE [PM_RESOURCE_TYPE]
			-- SQL statement to read resource types
		do
			if sql_get_resource_types_saved=Void
			then
				create sql_get_resource_types_saved.make (server, "SELECT * FROM resource_type WITH(NOLOCK)")
			end

			Result := sql_get_resource_types_saved
		end

	sql_get_messages_saved: like sql_get_messages
			-- See 'sql_get_messages'.
		note
			option: transient
		attribute
		end

	sql_get_messages: COLLECTING_PREDICATE [PM_MESSAGE]
			-- The most recent message that matches the input parameter
		do
			if sql_get_messages_saved=Void
			then
				create sql_get_messages_saved.make (server, "SELECT * FROM message WITH(NOLOCK) where message like ? order by run_id desc")
			end

			Result := sql_get_messages_saved
		end

	sql_get_messages_with_class_name_and_feature_name_saved: like sql_get_messages_with_class_name_and_feature_name
			-- See 'sql_get_messages_with_class_name_and_feature_name'.
		note
			option: transient
		attribute
		end

	sql_get_messages_with_class_name_and_feature_name: COLLECTING_PREDICATE [PM_MESSAGE]
			-- The most recent message that matches the input parameters
		do
			if sql_get_messages_with_class_name_and_feature_name_saved=Void
			then
				create sql_get_messages_with_class_name_and_feature_name_saved.make (server, "SELECT * FROM message WITH(NOLOCK) where message like ? and class_name = ? and feature_name = ? order by run_id desc")
			end

			Result := sql_get_messages_with_class_name_and_feature_name_saved
		end

	sql_create_process_saved: like sql_create_process
			-- See 'sql_create_process'.
		note
			option: transient
		attribute
		end

	sql_create_process: EXECUTION_PREDICATE
			-- Predicate to insert a new process into the process table
		do
			if sql_create_process_saved=Void
			then
				create sql_create_process_saved.make (server, "INSERT INTO process(name,success_recipients,failure_recipients,comment) VALUES (?,?,?,?)")
			end

			Result := sql_create_process_saved
		end

	sql_test_result_run_ids_for_changelist_saved: like sql_test_result_run_ids_for_changelist
			-- See 'sql_test_result_run_ids_for_changelist'.
		note
			option: transient
		attribute
		end

	sql_test_result_run_ids_for_changelist: attached INTEGER_PREDICATE
			-- Predicate to get the run-ids which have been identified as being
			-- ids of instance runs of test case executions associated with
			-- a given Perforce changelist
		do
			if sql_test_result_run_ids_for_changelist_saved=Void
			then
				create sql_test_result_run_ids_for_changelist_saved.make (server, "[
					SELECT run_id
					FROM changelist_test_result_map
					WHERE changelist = ?
					]")
				if not sql_test_result_run_ids_for_changelist_saved.success then
					error.abort ("sql_test_result_run_ids_for_changelist", "Failed to create predicate `sql_test_result_run_ids_for_changelist'.", Void)
				end
			end

			Result := sql_test_result_run_ids_for_changelist_saved
		end

	sql_record_changelist_test_result_run_id_mapping_saved: like sql_record_changelist_test_result_run_id_mapping
			-- See 'sql_record_changelist_test_result_run_id_mapping'.
		note
			option: transient
		attribute
		end

	sql_record_changelist_test_result_run_id_mapping: attached EXECUTION_PREDICATE
			-- Predicate to record a mapping between the run-id of an
			-- instance run of a test case execution and the associated
			-- Perforce changelist.
		do
			if sql_record_changelist_test_result_run_id_mapping_saved=Void
			then
				create sql_record_changelist_test_result_run_id_mapping_saved.make (server, "[
					INSERT INTO changelist_test_result_map(changelist,run_id) VALUES (?,?)
					]")
				if not sql_record_changelist_test_result_run_id_mapping_saved.success then
					error.abort ("sql_record_changelist_test_result_run_id_mapping", "Failed to create predicate `sql_record_changelist_test_result_run_id_mapping'.", Void)
				end
			end

			Result := sql_record_changelist_test_result_run_id_mapping_saved
		end

	redirect_log is
			-- Redirect log channel for Process Metrics needs.
		local
			l_definition: ROSE_DATABASE_DEFINITION
			l_channel: ROSE_OUTCHANNEL_DATABASE
		do
			if global.io_handler.log = Void then
				error.acknowledge
				l_definition := servers.pool.definition ("process_metrics")
				if l_definition = Void then
					error.signal (generator, "Logical 'process_metrics' is not defined in repository", <<>>)
					error.flush_and_acknowledge
				else
					create l_channel
					l_channel.open ("process_metrics.message.run_id")
					if error.signaled then
						error.acknowledge
						error.warn ("Cannot redirect log to process_metrics database '$1'", <<l_definition.full_path>>)
					else
						global.io_handler.set_log (l_channel)
						trace ("Logging to '$1', table 'message'%N", <<l_definition.full_path>>)
					end
				end
			else
				trace ("Logging to '$1'%N", <<global.io_handler.log>>)
			end
		end

	require_available is
			-- Require that process_metrics DB be available (abort if not available).
		do
			if not is_available then
				fail ("start", "process_metrics database is not available", Void)
			end
		end

	add_run_from_message (a_message: PM_MESSAGE; a_array_instance_runs: ROSE_LINEAR_ARRAY [PM_INSTANCE_RUN]) is
			-- Add run associated with message `a_message' to `a_array_instance_runs'.
		require
			a_message_not_void: a_message /= Void
			a_array_instance_runs_not_void: a_array_instance_runs /= Void
		do
			a_array_instance_runs.extend (run_by_id (a_message.run_id))
		ensure
			a_array_instance_runs_count_increased: a_array_instance_runs.count = old a_array_instance_runs.count + 1
		end

	duration_check: PM_DURATION_CHECK

	lower_standard_deviation_for_duration_check: DOUBLE
			-- Lower standard deviation for duration check.
		once
			Result := config.double ("lower_standard_deviation_for_duration_check", k_default_lower_standard_deviation_for_duration_check)
		end

	upper_standard_deviation_for_duration_check: DOUBLE
			-- Upper standard deviation for duration check.
		once
			Result := config.double ("upper_standard_deviation_for_duration_check", k_default_upper_standard_deviation_for_duration_check)
		end

	lookback_count_for_duration_check: INTEGER
			-- Lookback count for duration check.
		once
			Result := config.integer ("lookback_count_for_duration_check", k_default_lookback_count_for_duration_check)
		end

	k_default_lower_standard_deviation_for_duration_check: DOUBLE is -3.0

	k_default_upper_standard_deviation_for_duration_check: DOUBLE is 3.0

	k_default_lookback_count_for_duration_check: INTEGER is 90

	warned_on_current_run_duration_too_long: BOOLEAN
			-- Warning on current run duration too long has been sent.

invariant

	valid_stack: instance_stack /= Void and then not instance_stack.has (Void)
	valid_process_ids: process_ids /= Void
	valid_process_names: process_names /= Void
	valid_resource_types: resource_types /= Void
	valid_metrics: metrics /= Void
	valid_instances: instances /= Void
	valid_runs: runs /= Void
	valid_metrics_tests: metrics_tests /= Void

end

