indexing

	description: "Access to PM instances"
	author: "Zoran Simic"
	last_editor: "$Author: jq $"
	date: "$Date: 2011/08/15 $"
	revision: "$Revision: #9 $"
	path: "$File: //rose/source/kernel/pm/pm_context_instances.e $"

class PM_CONTEXT_INSTANCES

inherit
	ROSE_SHARED_GLOBAL

create {PM_CONTEXT}

	make

feature {NONE} -- Initialization

	make (a_context: PM_CONTEXT) is
			-- Create the instances object
		require
			valid_context: a_context /= Void
		do
			process_metrics := a_context
			create instances.make (1000, 100)
		end

feature -- Access

	instances: ROSE_VOLATILE_CACHE [PM_INSTANCE, INTEGER]
			-- Instance cache, contains last fetched objects

	instance_by_id (a_id: INTEGER): PM_INSTANCE is
			-- Instance with 'a_id' if any;
			-- Void if no such instance
		local
			l_instance: PM_INSTANCE
		do
			Result := instances.item (a_id)
			if Result = Void and then process_metrics.is_available then
				sql_select_100_instances_by_id.retrieve (Void, <<a_id>>)
				if not sql_select_100_instances_by_id.array.is_empty then
					l_instance := sql_select_100_instances_by_id.array.first
					if l_instance.instance_id = a_id then
						Result := l_instance
					end
					sql_select_100_instances_by_id.array.do_all (agent cache_instance)
				end
			end
		end

	instance_by_key (a_process: PM_PROCESS; a_partition_key: STRING; a_filedate: ROSE_DATE): PM_INSTANCE is
			-- Instance associated to 'a_process', 'a_partition_key' and 'a_filedate' if any
		require
			a_process_not_void: a_process /= Void
			a_partition_key_not_void: a_partition_key /= Void
			a_file_date_not_void: a_filedate /= Void
		do
			if process_metrics.is_available then
				sql_select_instance_by_key.retrieve (Void, <<a_process.process_id, a_partition_key, a_filedate>>)
				if sql_select_instance_by_key.array /= Void and then sql_select_instance_by_key.array.count > 0 then
					Result := sql_select_instance_by_key.array.item (1)
					cache_instance (Result)
				end
			end
		end

feature -- Basic operations

	refresh is
			-- Refresh the collection of instances
		local
			l_instance: PM_INSTANCE
			i: INTEGER
		do
			instances.clear_all
			sql_get_instances.retrieve (Void, Void)
			if sql_get_instances.success and then sql_get_instances.array /= Void then
				from
					i := sql_get_instances.array.lower
				until
					i > sql_get_instances.array.upper
				loop
					l_instance := sql_get_instances.array.item (i)
					instances.force (l_instance, l_instance.instance_id)
					i := i + 1
				end
			end
		end

feature {PM_CONTEXT} -- Implementation

	forced_instance (a_process: PM_PROCESS; a_partition_key: STRING; a_filedate: ROSE_DATE): PM_INSTANCE is
			-- Instance associated to 'a_process', 'a_partition_key' and 'a_filedate'
			-- The instance is created if not already present in database
		require
			a_process /= Void and then a_partition_key /= Void and then a_filedate /= Void
		local
			l_retry_count: INTEGER
		do
			from
				-- This is to handle the case where two processes reach this code at almost exactly the same time.
				-- They both call `instance_by_key' returning Void.
				-- Both then call `sql_insert_instance' and only one succeeds. We must wait a short period of time and try again.
			until
				Result /= Void or l_retry_count > 5
			loop
				if l_retry_count > 0 then
					print ("%NRetrying INSERT INTO instance%N")
				end
				Result := instance_by_key (a_process, a_partition_key, a_filedate)
				if Result = Void and then process_metrics.is_available then
					sql_insert_instance.execute (Void, <<a_process.process_id, a_partition_key, a_filedate>>)
					if sql_insert_instance.success then
						process_metrics.commit
						Result := instance_by_id (get_last_identity)
					elseif l_retry_count = 5 then
						Result := instance_by_key (a_process, a_partition_key, a_filedate)
						if Result = Void then
							process_metrics.fail (generator, "Can not create a new instance for process '$1', key '$2' and date '$3'", <<a_process.name, a_partition_key, a_filedate>>)
						end
					else
						global.env.sleep (1000)
					end
				end
				l_retry_count := l_retry_count + 1
			end
		end

feature {NONE} -- Implementation

	process_metrics: PM_CONTEXT

	cache_instance (a_instance: PM_INSTANCE) is
			-- Cache `a_instance' into `instances' if not Void.
		do
			if a_instance /= Void then
				instances.put (a_instance, a_instance.instance_id)
			end
		end

	sql_get_instances: COLLECTING_PREDICATE [PM_INSTANCE] is
		once
			create Result.make_strict (process_metrics.server, "SELECT * FROM instance WITH(NOLOCK)")
		end

	sql_select_instance_by_id: COLLECTING_PREDICATE [PM_INSTANCE] is
		once
			create Result.make_strict (process_metrics.server, "SELECT * FROM instance WITH(NOLOCK) WHERE instance_id=?")
		end

	sql_select_100_instances_by_id: COLLECTING_PREDICATE [PM_INSTANCE] is
		once
			create Result.make_strict (process_metrics.server, "SELECT top 100 * FROM instance WITH(NOLOCK) WHERE instance_id >= ? order by instance_id")
		end

	sql_select_instance_by_key: COLLECTING_PREDICATE [PM_INSTANCE] is
		once
			create Result.make_strict (process_metrics.server, "SELECT * FROM instance WITH(NOLOCK) WHERE process_id=? AND partition_key=? AND file_date=?")
		end

	sql_insert_instance: EXECUTION_PREDICATE is
		once
			create Result.make (process_metrics.server, "INSERT INTO instance (process_id,partition_key,file_date) VALUES (?,?,?)")
		end

	get_last_identity: INTEGER is
		do
			sql_last_identity.retrieve (Void, Void)
			if sql_last_identity.array /= Void and then sql_last_identity.array.count /= 0 then
				Result := sql_last_identity.array.item (sql_last_identity.array.lower)
			end
		end

	sql_last_identity: INTEGER_PREDICATE is
		once
			create Result.make (process_metrics.server, "SELECT cast(@@IDENTITY as int) as [id]")
		end

end
