note

	description:

		"Check for suspect process run duration."

	author: "David Stevens"
	last_editor: "$Author: da $"
	date: "$Date: 2012/11/29 $"
	revision: "$Revision: #1 $"
	path: "$File: //rose/source/kernel/pm/pm_duration_check.e $"

class

	PM_DURATION_CHECK

inherit

	DATABASE_SERVER

	ROSE_SHARED_GLOBAL

	ROSE_CONSTANTS

create

	make

feature {PM_CONTEXT} -- Initialization

	make (a_run: PM_INSTANCE_RUN)
			-- Create with `a_run'.
		require
			a_run_valid:
				a_run /= Void and then
				a_run.instance /= Void and then
				a_run.instance.instance_id /= 0 and then
				a_run.instance.process /= Void and then
				a_run.instance.process.process_id /= 0
		do
			run := a_run
		end

feature {PM_CONTEXT} -- Status Check

	is_too_long: BOOLEAN
			-- Whether current run duration is too long given `standard_deviation_multiple', looking back `lookback_count' observations.
		require
			standard_deviation_multiple_non_negative: standard_deviation_multiple >= 0
			lookback_count_valid: valid_lookback_count (lookback_count)
		do
			Result := duration > (average + (standard_deviation_multiple * standard_deviation))
		end

	is_too_short: BOOLEAN
			-- Whether current run duration is too short given `standard_deviation_multiple', looking back `lookback_count' observations.
		require
			standard_deviation_multiple_non_positive: standard_deviation_multiple <= 0
			lookback_count_valid: valid_lookback_count (lookback_count)
		do
			Result := duration < (average + (standard_deviation_multiple * standard_deviation))
		end

	valid_lookback_count (a_lookback_count: INTEGER): BOOLEAN
			-- `a_lookback_count' is valid.
		do
			Result := a_lookback_count > 0 and a_lookback_count < k_maximum_lookback_count
		end

feature {PM_CONTEXT} -- Element Change

	set_statistics_and_duration (a_standard_deviation_multiple: DOUBLE; a_lookback_count: INTEGER)
			-- Set the statistics and run duration given `a_standard_deviation_multiple', looking back `a_lookback_count' observations.
		require
			a_lookback_count_valid: valid_lookback_count (a_lookback_count)
		local
			l_date_now_ut: ROSE_DATE
		do
			if durations = Void then
				retrieve_durations
			end
			if a_lookback_count /= lookback_count or a_standard_deviation_multiple /= standard_deviation_multiple then
				if durations.count < a_lookback_count then
					-- Observation count is insufficient if less than requested lookback count.
					durations.wipe_out
				end
				lookback_count := k_maximum_lookback_count.min (a_lookback_count)
				standard_deviation_multiple := a_standard_deviation_multiple
				set_statistics
			end
			create l_date_now_ut.make_now_utc
			duration := l_date_now_ut.relative_duration (run.start_time).fine_seconds_count
		end

feature {PM_CONTEXT} -- Access

	message: STRING
			-- Duration status message.
		do
			Result :=
				"%N%T" +
				"Current run duration in seconds = " + duration.out + "%N%T" +
				"Lookback count of observations = " + lookback_count.out + "%N%T" +
				"Average process duration of lookback = " + average.out + "%N%T" +
				"Standard deviation of duration of lookback = " + standard_deviation.out + "%N"
		end

	lookback_count: INTEGER

	standard_deviation_multiple: DOUBLE

feature {NONE} -- Implementation

	set_statistics
			-- Set the statistics for the processing run's durations.
		require
			durations_not_void: durations /= Void
		local
			l_vector: VECTOR
		do
			if durations.count < lookback_count then
				-- If history is insufficient for comparison, use extreme processing time to prevent suspect duration notification.
				if standard_deviation >= 0.0 then
					average := k_long_processing_time
					standard_deviation := k_long_processing_time
				else
					average := k_short_processing_time
					standard_deviation := k_short_processing_time
				end
			else
				create l_vector.make_from_array (durations.subarray (1, lookback_count))
				average := l_vector.average
				standard_deviation := l_vector.standard_deviation_from_mean (average)
			end
		end

	retrieve_durations
			-- Retrieve the processing run's durations.
		local
			l_selection: like duration_selection
		do
			l_selection := duration_selection
			l_selection.retrieve (Void, Void)
			if not l_selection.success then fail ("retrieve_durations", "select failed", Void) end
			create durations.make_from_array (l_selection.array)
		ensure
			durations_not_void: durations /= Void
		end

	duration_selection: DOUBLE_PREDICATE
			-- SQL selection to select run durations in seconds.
		local
			l_statement: STRING
		do
			l_statement :=
				"select " +
				"	top " + k_maximum_lookback_count.out + " " +
				"	datediff (millisecond, r.start_time, r.end_time) / 1000.0 as duration " +
				"from " +
				"	instance as i " +
				"join " +
				"	instance_run as r " +
				"on " +
				"	r.instance_id = i.instance_id and " +
				"	r.exit_code = " + k_exit_code_success.out + " and " +
				"	r.end_time < '" + constants.maximum_date_time.ccyymmdd.out + "' " +
				"where " +
				"	i.process_id = " + run.instance.process.process_id.out + " and " +
				"	i.partition_key = '" + run.instance.partition_key + "' " +
				"order by " +
				"	r.run_id desc "
			create Result.make (servers.process_metrics_server, l_statement)
			if not Result.success then fail ("duration_selection", "make failed", Void) end
		ensure
			result_not_void: Result /= Void
		end

	durations: ROSE_LINEAR_ARRAY [DOUBLE]
			-- Trailing processing run's durations in seconds ordered by most recent observations first.

	duration: DOUBLE

	average: DOUBLE

	standard_deviation: DOUBLE

	k_exit_code_success: INTEGER is 1

	k_maximum_lookback_count: INTEGER is 1_000
			-- Results become less meaningful if too long a history is used.

	k_long_processing_time: DOUBLE  is 10_000_000.0
			-- Used to set a time that won't be exceeded in case there is no history for comparison.

	k_short_processing_time: DOUBLE  is -10_000_000.0
			-- Used to set a time that won't be exceeded in case there is no history for comparison.

	run: PM_INSTANCE_RUN

invariant

	run_valid:
		run /= Void and then
		run.instance /= Void and then
		run.instance.instance_id /= 0 and then
		run.instance.process /= Void and then
		run.instance.process.process_id /= 0

end
