note

	description:

		"Definition of a process instance"

	author: "Zoran Simic"
	last_editor: "$Author: jq $"
	date: "$Date: 2012/07/17 $"
	revision: "$Revision: #8 $"
	path: "$File: //rose/source/kernel/pm/pm_instance.e $"

class
	PM_INSTANCE

inherit
	PM_COMMON
		redefine
			is_equal
		end

create
	default_create

feature -- Access

	instance_id: INTEGER
			-- Internal id for this process instance

	process: PM_PROCESS is
			-- Associated process
		do
			Result := process_metrics.process_by_id (process_id)
		end

	partition: PM_PROCESS_PARTITION is
			-- Partition attached to this instance
		do
			if saved_partition = Void then
				saved_partition := process.partition (partition_key)
			end
			Result := saved_partition
		end

	partition_key: STRING
		-- Arbitrary partition key used to discriminate metrics further within one process

	file_date: ROSE_DATE
			-- File date associated with this instance

	status_id: INTEGER
			-- Latest status id for this instance (0=not yet checked)

	ticket_id: INTEGER
			-- Used for feedback tracking

	runs: PM_INSTANCE_RUN_COLLECTION is
			-- Runs performed for this instance, oldest run appears first
		do
			if saved_runs = Void then
				create saved_runs.make_for_instance (Current)
			end
			Result := saved_runs
		end

	metrics: PM_METRIC_COLLECTION is
			-- Metrics for this instance
		do
			if saved_metrics = Void then
				if runs.count > 0 then
					if process.concurrency = 1 then
						create saved_metrics.make_for_instance_run (runs.last)
					else
						create saved_metrics.make_for_instance (Current)
					end
				end
			end
			Result := saved_metrics
		end

	messages: PM_MESSAGE_COLLECTION is
			-- Messages for this instance
		do
			if saved_messages = Void then
				create saved_messages.make_for_instance (Current)
			end
			Result := saved_messages
		end

	resources: PM_RESOURCE_ACCESS_COLLECTION is
			-- Resources accessed by this instance
		do
			if saved_resources = Void then
				create saved_resources.make_for_instance (Current)
			end
			Result := saved_resources
		end

	is_equal (other: like Current): BOOLEAN is
			-- Is `other' attached to an object considered
			-- equal to current object?
		do
			Result := instance_id = other.instance_id
		end

	current_run: PM_INSTANCE_RUN
			-- Instance currently operating

	is_running: BOOLEAN is
			-- Is an instance run currently in progress?
		do
			Result := current_run /= Void
		end

	start_time: ROSE_DATE is
			-- Start time of earliest run for this instance
		local
			l_runs: like runs
			l_date: ROSE_DATE
			i: INTEGER
		do
			l_runs := runs
			if l_runs.count > 0 then
				from
					i := l_runs.lower
				until
					i > l_runs.upper
				loop
					l_date := l_runs.item (i).start_time
					if l_date /= Void then
						if Result = Void then
							Result := l_date
						else
							Result := Result.min (l_date)
						end
					end
					i := i + 1
				end
			end
		end

	end_time: ROSE_DATE is
			-- End time of last run for this instance
		local
			l_runs: like runs
			l_date: ROSE_DATE
			i: INTEGER
		do
			l_runs := runs
			if l_runs.count > 0 then
				from
					i := l_runs.lower
				until
					i > l_runs.upper
				loop
					l_date := l_runs.item (i).end_time
					if l_date /= Void then
						if Result = Void then
							Result := l_date
						else
							Result := Result.max (l_date)
						end
					end
					i := i + 1
				end
			end
		end

	start_time_represented: STRING is
			-- String representation of 'start_time'
		do
			if is_utc then
				Result := global.represented_utc_date (start_time)
			else
				Result := global.represented_date (start_time)
			end
		end

	end_time_represented: STRING is
			-- String representation of 'end_time'
		do
			if is_utc then
				Result := global.represented_utc_date (end_time)
			else
				Result := global.represented_date (end_time)
			end
		end

	duration_represented: STRING is
			-- String representation of 'duration' if instance terminated, "Still running" otherwise
		local
			l_start, l_end: ROSE_DATE
		do
			l_start := start_time
			l_end := end_time
			if l_start /= Void and then l_end /= Void then
				Result := global.formatter.represented_duration (l_start, l_end, False)
			elseif l_start /= Void then
				Result := "Still running"
			else
				Result := "Not started"
			end
		end

	out: STRING is
			-- New string containing terse printable representation of current object
		do
			create Result.make (10)
			if process /= Void then
				Result.append (process.name)
			else
				Result.append ("?")
			end
			Result.append (" (")
			Result.append_string (partition_key)
			if file_date /= Void then
				Result.append (", ")
				Result.append (file_date.formatted_out ("[0]mm/[0]dd/yyyy"))
			end
			Result.append (")")
		end

feature {PM_CONTEXT} -- Implementation

	start_run is
			-- Start a new instance run.
		require
			not_running: not is_running
		do
			create current_run.make (Current)
			if current_run.run_id /= 0 then
				Global.set_log_id (current_run.run_id)
			end
			saved_runs := Void
		ensure
			ok_xor_error: current_run /= Void xor error.signaled
		end

	terminate (a_code: INTEGER) is
			-- Terminate current instance run.
		require
			is_running: is_running
		local
			l_test_result : PM_METRIC_TEST_RESULT
		do
			if current_run /= Void then
				-- We avoid crashing if this gets called in production without an actual current instance
				process_metrics.metrics.register_dynamic_tests (Current)
				l_test_result := process_metrics.metrics.current_run_test(Current)
				if l_test_result = Void or else not l_test_result.enable_process_failure then
					current_run.terminate (a_code)
				else
					error.signal("terminate","Metric test failure",Void)
					current_run.terminate (-3)
				end
				process_metrics.send_termination_report
			end
			current_run := Void
			process_metrics.metrics.reset
			Global.set_log_id (0)
		ensure
			not_running: not is_running
		end

feature {NONE} -- Implementation

	process_id: INTEGER								-- Process id
	saved_partition: PM_PROCESS_PARTITION			-- Partition
	saved_runs: PM_INSTANCE_RUN_COLLECTION			-- Instance runs
	saved_metrics: PM_METRIC_COLLECTION				-- Metrics
	saved_messages: PM_MESSAGE_COLLECTION			-- Messages
	saved_resources: PM_RESOURCE_ACCESS_COLLECTION	-- Resources

	is_utc: BOOLEAN is
			-- Is the first run recorded with UTC dates?
		do
			if runs /= Void and then not runs.is_empty then
				Result := runs.item (runs.lower).is_utc
			end
		end

	sql_select_runs: COLLECTING_PREDICATE [PM_INSTANCE_RUN] is
			-- SQL statement used to fetch runs belonging to this instance
		once
			create Result.make (process_metrics.server,
				"SELECT * FROM instance_run WITH(NOLOCK) WHERE instance_id=? ORDER BY start_time ASC")
		end

end
