note

	description:

		"Definition of a process instance"

	author: "Zoran Simic"
	last_editor: "$Author: c4 $"
	date: "$Date: 2011/04/15 $"
	revision: "$Revision: #14 $"
	path: "$File: //rose/source/kernel/pm/pm_instance_run.e $"

class
	PM_INSTANCE_RUN

inherit
	PM_COMMON
		redefine
			is_equal
		end

	ROSE_RESOURCE_ACCESS_CONSTANTS
		undefine
			is_equal, out
		end

create
	default_create

create {PM_INSTANCE}
	make

feature {NONE} -- Initialization

	make (a_instance: PM_INSTANCE) is
			-- Creation procedure.
		require
			instance_ok: a_instance /= Void and then a_instance.instance_id > 0
		do
			instance_id := a_instance.instance_id
			saved_instance := a_instance
			if process_metrics.server.has_column_name ("instance_run", "changelist") then
				sql_insert_instance_run.execute (Void,<<instance_id, create {ROSE_DATE}.make_now_utc,
					global.env.current_process_id, global.app.version, global.app.program_arguments, global.env.job_id,
					global.env.scheduler, not_from_resource_access_tracking, global.app.info.revision>>)
				run_id := get_last_identity
				if sql_insert_instance_run.success then
					process_metrics.commit
					if run_id > 0 then
						sql_read_instance_run.retrieve_one (Current,Void,<<run_id>>)
					end
				end				
			else
				sql_insert_instance_run_old.execute (Void,<<instance_id, create {ROSE_DATE}.make_now_utc,
					global.env.current_process_id, global.app.version, global.app.program_arguments, global.env.job_id,
					global.env.scheduler, not_from_resource_access_tracking>>)
				run_id := get_last_identity
				if sql_insert_instance_run_old.success then
					process_metrics.commit
					if run_id > 0 then
						sql_read_instance_run.retrieve_one (Current,Void,<<run_id>>)
					end
				end				
			end
		end

feature -- Access (table columns)

	-- Reminder : also change ROSE_RESOURCE_ACCESS if you make any schema changes to the instance_run table

	run_id: INTEGER
			-- Internal id for this instance run

	instance: PM_INSTANCE is
			-- Associated instance
		do
			if saved_instance = Void then
				saved_instance := process_metrics.instance_by_id (instance_id)
			end
			Result := saved_instance
		end

	start_time: ROSE_DATE	-- Timestamp when this instance run was started
	end_time: ROSE_DATE		-- Timestamp when this instance run was ended
	exit_code: INTEGER		-- Exit code reported when run finished

	user_name: STRING		-- User who ran this instance run
	host_name: STRING		-- Machine where this instance run took place
	host_pid: INTEGER		-- Process id under which this instance ran

	exe: STRING				-- Path to executable that was ran
	version: STRING			-- Version of 'exe'
	args: STRING			-- Arguments passed to the executable that was ran

	is_utc: BOOLEAN			-- Are 'start_time' and 'end_time' UTC timestamps?

	scheduler_job_id : INTEGER		      -- Activebatch job id
	scheduler_name : STRING               -- ActiveBatch job scheduler

	changelist: INTEGER
			-- Highest changelist in Perforce snapshot used to compile the executable

	from_resource_access_tracking : INTEGER
			-- Was this run captured by resource_access_tracking? (see ROSE_RESOURCE_ACCESS_CONSTANTS)

feature -- Access

	exe_without_path : STRING
			-- exe file name without the path

	start_time_local: ROSE_DATE is
			-- Localized 'start_time'
		do
			Result := start_time.twin
			Result.second_add (global.env.utc_seconds_offset)
		end

	end_time_local: ROSE_DATE is
			-- Localized 'end_time'
		do
			Result := end_time.twin
			Result.second_add (global.env.utc_seconds_offset)
		end

	start_time_represented: STRING is
			-- String representation of 'start_time'
		do
			if is_utc then
				Result := global.represented_utc_date (start_time)
			else
				Result := global.represented_date (start_time)
			end
		end

	end_time_represented: STRING is
			-- String representation of 'end_time'
		do
			if is_utc then
				Result := global.represented_utc_date (end_time)
			else
				Result := global.represented_date (end_time)
			end
		end

	is_successful: BOOLEAN is
			-- Did this run complete successfully?
		do
			Result := exit_code = 1
		end

	is_failed: BOOLEAN is
			-- Did this run fail?
		do
			Result := exit_code = -1
		end

	duration: DATE_TIME_DURATION is
			-- Relative duration between 'start_time' and 'end_time'
			-- Void if still running
		local
			l_now: ROSE_DATE
		do
			if end_time /= Void then
				Result := end_time.relative_duration (start_time)
			elseif start_time /= Void then
				if is_utc then
					create l_now.make_now_utc
				else
					create l_now.make_now
				end
				Result := l_now.relative_duration (start_time)
			end
		ensure
			started: is_started = (Result /= Void)
		end

	metrics: PM_METRIC_COLLECTION is
			-- Metrics generated by this instance run
		do
			if saved_metrics = Void then
				create saved_metrics.make_for_instance_run (Current)
			end
			Result := saved_metrics
		end

	metric_tests: PM_METRIC_TEST_COLLECTION
			-- Metric tests generated by this instance run
		do
			if saved_metric_tests = Void then
				create saved_metric_tests.make_for_instance_run (Current)
			end
			Result := saved_metric_tests
		end

	messages: PM_MESSAGE_COLLECTION is
			-- Messages for this instance
		do
			if saved_messages = Void then
				create saved_messages.make_for_instance_run (Current)
			end
			Result := saved_messages
		end

	resources: PM_RESOURCE_ACCESS_COLLECTION is
			-- Resources accessed by this instance run
		do
			if saved_resources = Void then
				create saved_resources.make_for_run (Current)
			end
			Result := saved_resources
		end

	duration_represented: STRING is
			-- String representation of 'duration' if instance terminated, "Still running" otherwise
		do
			if end_time /= Void then
				Result := global.formatter.represented_duration (start_time, end_time, False)
			elseif start_time /= Void then
				Result := "Still running"
			else
				Result := "Not started"
			end
		end

	is_started: BOOLEAN is
			-- Is this instance started?
		do
			Result := start_time /= Void
		end

	is_terminated: BOOLEAN is
			-- Has this instance terminateed?
		do
			Result := end_time /= Void
		end

	is_equal (other: like Current): BOOLEAN is
			-- Is `other' attached to an object considered
			-- equal to current object?
		do
			Result := run_id = other.run_id
		end

	is_valid_date (a_date: ROSE_DATE): BOOLEAN is
			-- Is 'a_date' a valid date?
		do
			Result := a_date /= Void and then a_date.is_available and then not a_date.is_undetermined
		end

	out: STRING is
			-- New string containing terse printable representation of current object
		do
			create Result.make (10)
			Result.append (run_id.out)
		end

feature -- Basic operations

	force_terminate_if_disappeared is
			-- Cleanup this process if we can detect it has disappeared
			-- (claimed to be running on the current machine and there is no process id).
		require
			not_done: not is_terminated
		local
			l_process: WINDOWS_PROCESS
		do
			if host_name /= Void and then host_name.is_case_insensitive_equal (global.env.host_name) then
				create l_process.make (host_pid)
				if not l_process.exists then
					force_terminate (Dead_process_termination_code)
				end
			end
		end

	force_terminate (a_code: INTEGER) is
			-- Force a terminate instance.
		require
			not_done: not is_terminated
		do
			exit_code := a_code
			sql_update_instance_run.execute (Void,<<create {ROSE_DATE}.make_now_utc, a_code,run_id>>)
			if sql_update_instance_run.success then
				process_metrics.commit
				sql_read_instance_run.retrieve_one (Current,Void,<<run_id>>)
			end
		ensure
			done: is_terminated
		end

	set_exe_without_path is
			-- Delete the path from exe and store it as exe_without_path
		require
			valid_exe : exe.count > 0
		do
			exe_without_path := global.formatter.file_name (exe.as_lower)
		ensure
			valid_exe_without_path : exe_without_path.count > 0
		end

	set_from_resource_access_tracking( a_state : INTEGER )
			-- Obsolete once resource_access_tracking flag is available
		do
			from_resource_access_tracking := a_state
		end

feature {PM_INSTANCE} -- Implementation

 	terminate (a_code: INTEGER) is
			-- Terminate instance.
		require
			not_done: not is_terminated
		do
			exit_code := a_code
			process_metrics.metrics.write_to_database
			sql_update_instance_run.execute (Void,<<create {ROSE_DATE}.make_now_utc, a_code,run_id>>)
			if sql_update_instance_run.success then
				process_metrics.commit
				sql_read_instance_run.retrieve_one (Current,Void,<<run_id>>)
			end
		ensure
			done: is_terminated
		end

feature {NONE} -- Implementation

	instance_id: INTEGER							-- Instance id of this instance run
	saved_instance: PM_INSTANCE						-- Parent instance object
	saved_metrics: PM_METRIC_COLLECTION				-- Metrics collection corresponding to this instance run
	saved_metric_tests: PM_METRIC_TEST_COLLECTION	-- Metrics tests corresponding to this instance run
	saved_messages: PM_MESSAGE_COLLECTION			-- Message collection corresponding to this instance run
	saved_resources: PM_RESOURCE_ACCESS_COLLECTION	-- Resources accessed by this instance run

	sql_read_instance_run: COLLECTING_PREDICATE [PM_INSTANCE_RUN] is
			-- SQL statement fetching a given instance run
		once
			create Result.make (process_metrics.server,
				"SELECT * FROM instance_run WITH(NOLOCK) WHERE run_id=?")
		end

	sql_update_instance_run: EXECUTION_PREDICATE is
			-- SQL statement updating an instance run
		once
			create Result.make (process_metrics.server,
				"UPDATE instance_run SET end_time=?,exit_code=? WHERE run_id=?")
			if not Result.success then
				(create {EXCEPTIONS}).raise ("sql_update_instance_run failed.")
			end
		end

	sql_insert_instance_run: EXECUTION_PREDICATE is
			-- SQL statement inserting an instance run
		once
			create Result.make (process_metrics.server,
				"INSERT INTO instance_run (instance_id,start_time,host_pid,version,args,is_utc,scheduler_job_id,scheduler_name,from_resource_access_tracking,changelist)  VALUES (?,?,?,?,?,1,?,?,?,?)")
			if not Result.success then
				(create {EXCEPTIONS}).raise ("sql_insert_instance_run failed.")
			end
		end

	sql_insert_instance_run_old: EXECUTION_PREDICATE is
			-- SQL statement inserting an instance run for tables without changelist
		once
			create Result.make (process_metrics.server,
				"INSERT INTO instance_run (instance_id,start_time,host_pid,version,args,is_utc,scheduler_job_id,scheduler_name,from_resource_access_tracking)  VALUES (?,?,?,?,?,1,?,?,?)")
			if not Result.success then
				(create {EXCEPTIONS}).raise ("sql_insert_instance_run failed.")
			end
		end

	Dead_process_termination_code: INTEGER is -11
			-- Code for zombie processes being force terminated.

end
