indexing

	description: "Represents a collection of PM_INSTANCE_RUN"

	author: "Zoran Simic"
	last_editor: "$Author: c4 $"
	date: "$Date: 2010/10/21 $"
	revision: "$Revision: #8 $"

class	PM_INSTANCE_RUN_COLLECTION

inherit

	PM_COLLECTION [PM_INSTANCE_RUN]

create

	make_for_instance, make_for_process

feature {NONE} -- Initialization

	make_for_instance (a_instance: PM_INSTANCE) is
			-- Initialize runs associated with `a_instance'.
		require
			instance_ok: a_instance /= Void
		do
			get_items (sql_instance_runs, <<a_instance.instance_id>>)
		end

	make_for_process (a_process: PM_PROCESS) is
			-- Initialize runs associated with `a_process'.
		require
			a_process_not_void: a_process /= Void
			good_process_id: a_process.process_id > 0
		local
			l_statement: STRING
			l_predicate: COLLECTING_PREDICATE [PM_INSTANCE_RUN]
		do
			l_statement := global.formatted (sql_instance_runs_for_process, <<a_process.process_id>>)
			create l_predicate.make_strict (process_metrics.server, l_statement)
			get_items (l_predicate, Void)
		end

feature -- Access

	in_progress: ROSE_LINEAR_ARRAY [PM_INSTANCE_RUN] is
			-- Runs currently in progress in this collection
		do
			create Result.make_empty
			items.do_all (agent add_to_in_progress (Result, ?))
		ensure
			result_ok: Result /= Void
		end

	has_successful: BOOLEAN is
			-- Does this collection have at least one successful run?
		local
			i: INTEGER
		do
			from
				i := items.upper
			until
				Result or else i < items.lower
			loop
				Result := items.item (i).is_successful
				i := i - 1
			end
		end

	successful: ROSE_LINEAR_ARRAY [PM_INSTANCE_RUN] is
			-- Runs that have terminated successfully, oldest run appears first in the resulting array
		local
			i: INTEGER
		do
			from
				create Result.make_empty
				i := items.lower
			until
				i > items.upper
			loop
				if items.item (i).is_successful then
					Result.extend (items.item (i))
				end
				i := i + 1
			end
		ensure
			result_ok: Result /= Void
		end

	failed: ROSE_LINEAR_ARRAY [PM_INSTANCE_RUN] is
			-- Runs that have terminated in failure, oldest run appears first in the resulting array
		local
			i: INTEGER
		do
			from
				create Result.make_empty
				i := items.lower
			until
				i > items.upper
			loop
				if not items.item (i).is_successful then
					Result.extend (items.item (i))
				end
				i := i + 1
			end
		ensure
			result_ok: Result /= Void
		end

	last_successful_start_time: ROSE_DATE is
			-- 'start_time' of 'successful.last' if any
		local
			l_run: PM_INSTANCE_RUN
		do
			l_run := successful.last
			if l_run /= Void then
				Result := l_run.start_time_local
			end
		end

	last_successful_end_time: ROSE_DATE is
			-- 'end_time' of 'successful.last' if any
		local
			l_run: PM_INSTANCE_RUN
		do
			l_run := successful.last
			if l_run /= Void then
				Result := l_run.end_time_local
			end
		end

feature -- Basic operations

	runs_by_message_substring (a_feature_name, a_message_substring: STRING): ROSE_LINEAR_ARRAY [like item] is
			-- Runs having message associated with 'a_feature_name' and 'a_message_substring'
		local
			i, j: INTEGER
			l_found: BOOLEAN
			l_messages: PM_MESSAGE_COLLECTION
			l_message: PM_MESSAGE
		do
			create Result.make_empty
			from
				i := lower
			until
				i > upper
			loop
				l_messages := item (i).messages
				if l_messages /= Void and then l_messages.count > 0 then
					from
						j := l_messages.lower
						l_found := False
					until
						l_found or else j > l_messages.upper
					loop
						l_message := l_messages.item (j)
						if l_message.feature_name /= Void
						and then l_message.feature_name.is_case_insensitive_equal (a_feature_name) then
							if l_message.message /= Void and then l_message.message.is_case_insensitive_equal (a_message_substring) then
								Result.extend (item (i))
								l_found := True
							end
						end
						j := j + 1
					end
				end
				i := i + 1
			end
		end


feature {NONE} -- Implementation

	sql_instance_runs: COLLECTING_PREDICATE [PM_INSTANCE_RUN] is
			-- SQL statement allowing to fetch runs by instance id
		once
			create Result.make_strict (process_metrics.server,
				"SELECT * FROM instance_run WITH(NOLOCK) WHERE instance_id = ? order by run_id")
		end

	sql_instance_runs_for_process: STRING is
			-- SQL statement allowing to fetch runs by process id
		do
			Result :=	"[
		        SELECT * FROM instance_run WITH(NOLOCK) WHERE instance_id IN 
				  (SELECT instance_id FROM instance WHERE process_id = $1)
				  order by run_id
            ]"
		ensure
			sql_instance_runs_for_process_not_void: Result /= Void
		end

	add_to_in_progress (a_result: like in_progress; a_run: PM_INSTANCE_RUN) is
			-- If run is still in process add it to the list.
			-- (attempt to cleanup dead runs before doing so.)
		require
			a_result_not_void: a_result /= void
			a_run_not_void: a_run /= Void
			last_item_in_progress: not a_result.is_empty implies not a_result.last.is_terminated
		do
			if not a_run.is_terminated then
				a_run.force_terminate_if_disappeared
				if not a_run.is_terminated then
					a_result.extend (a_run)
				end
			end
		ensure
			last_item_in_progress: not a_result.is_empty implies not a_result.last.is_terminated
		end

end
