note

	description:

		"Metrics statistics"

	author: "Zoran Simic"
	last_editor: "$Author: e2 $"
	date: "$Date: 2014/04/08 $"
	revision: "$Revision: #14 $"
	path: "$File: //rose/source/kernel/pm/pm_metric_statistics.e $"

class
	PM_METRIC_STATISTICS

inherit
	PM_COMMON
	ROSE_CONSTANTS
		undefine
			out
		end

create
	make_for_partition

feature {NONE} -- Creation

	make_for_partition (a_partition: PM_PROCESS_PARTITION)
			-- Compute statistics for 'a_partition'
		require
			valid_partition: a_partition /= Void
		local
			l_instances: PM_INSTANCE_COLLECTION
			i: INTEGER
		do
			partition := a_partition
			create moments.make (0)
			create duration.make_empty
			create run_periodicity.make_empty
			l_instances := a_partition.instances
			from
				i := (l_instances.upper - instance_lookback_count).max (l_instances.lower)
			until
				i > l_instances.upper
			loop
				add_instance (l_instances.item (i), l_instances.item (i+1))
				i := i + 1
			end
			from
				moments.start
			until
				moments.off
			loop
				finalize_moments (moments.item_for_iteration)
				moments.forth
			end
			finalize_moments (duration)
			finalize_moments (run_periodicity)
		ensure
			partition_set: partition = a_partition
		end

feature -- Access

	partition: PM_PROCESS_PARTITION
			-- Associated process partition

	moments: HASH_TABLE [ROSE_MOMENTS,INTEGER]
			-- Moments hashed by metric-type-id

	duration: ROSE_MOMENTS
	run_periodicity: ROSE_MOMENTS
			-- Duration and run-periodicity moments

	mean_run_periodicity: TIME_DURATION
			-- The average time difference between 2 runs
		do
			if run_periodicity.data.count > 1 then
				create Result.make_by_seconds (run_periodicity.moving_average_lookback(run_periodicity.data.count,process_metrics.statistics_lookback).truncated_to_integer)
			else
				create Result.make_by_seconds(0)
			end
		ensure
			mean_run_periodicity_not_void: Result /= Void
			non_negative_mean_run_periodicity: not Result.is_negative
		end

	median_run_periodicity: TIME_DURATION
			-- The median time difference between 2 runs
		do
			if run_periodicity.data.count > 1 then
				create Result.make_by_seconds (run_periodicity.moving_median(run_periodicity.data.count,process_metrics.statistics_lookback).truncated_to_integer)
			else
				create Result.make_by_seconds(0)
			end
		ensure
			median_run_periodicity_not_void: Result /= Void
			non_negative_median_run_periodicity: not Result.is_negative
		end

	run_periodicity_standard_deviation: TIME_DURATION
			-- Standard deviation of periodicity
		do
			if run_periodicity.data.count > 1 then
				create Result.make_by_seconds (run_periodicity.moving_standard_deviation(run_periodicity.data.count,process_metrics.statistics_lookback).truncated_to_integer)
			else
				create Result.make_by_seconds(0)
			end
		ensure
			run_periodicity_standard_deviation_not_void: Result /= Void
			non_negative_run_periodicity_standard_deviation: not Result.is_negative
		end

	mean_run_duration: TIME_DURATION
			-- Mean of run time
		local
			l_lookback, l_count: INTEGER
			l_average: DOUBLE
		do
			l_lookback := process_metrics.statistics_lookback
			if duration /= Void and then duration.data /= Void then
				l_count := duration.data.count
			end
			if l_count > l_lookback then
				l_average := duration.moving_average_lookback(l_count, l_lookback)
			end
			create Result.make_by_seconds (l_average.truncated_to_integer)
		ensure
			mean_run_duration_not_void: Result /= Void
			non_negative_mean_run_duration: not Result.is_negative
		end

	run_duration_standard_deviation: TIME_DURATION
			-- Standard deviation of the time the process runs
		local
			l_lookback, l_count: INTEGER
			l_average: DOUBLE
		do
			l_lookback := process_metrics.statistics_lookback
			if duration /= Void and then duration.data /= Void then
				l_count := duration.data.count
			end
			if l_count > l_lookback then
				l_average := duration.moving_standard_deviation (l_count, l_lookback)
			end
			create Result.make_by_seconds (l_average.truncated_to_integer)
		ensure
			run_duration_standard_deviation_not_void: Result /= Void
			non_negative_run_duration_standard_deviation: not Result.is_negative
		end

	is_run_periodicity_clean: BOOLEAN
			-- e2: what does this test?
		local
			l_periodicity: DOUBLE
			l_now, l_last_start: ROSE_DATE
			l_process: PM_PROCESS
			l_instance: PM_INSTANCE
			l_next_in_sequence: NEXT_IN_SEQUENCE
		do
			Result := True
			if partition.instances.count > 1 and then not run_periodicity.data.is_empty then
				create l_next_in_sequence.make(run_periodicity.data.subarray(run_periodicity.data.upper.min(run_periodicity.data.lower.max(20)),run_periodicity.data.upper))
				if l_next_in_sequence.confidence > 100 then
					l_periodicity := l_next_in_sequence.next_i_th(1)
				else
					l_periodicity := median_run_periodicity.seconds_count
				end
				create l_now.make_now
				l_process := partition.process
				if l_process /= Void then
					l_instance := l_process.most_recent_successful_instance (partition.partition_key)
					if l_instance /= Void then
						l_last_start := l_instance.start_time
						l_last_start.second_add((l_periodicity * 1.25).truncated_to_integer)
						if l_last_start < l_now then
							Result := False
						end
					end
				end
			end
		end

	out: STRING is
			-- New string containing terse printable representation of current object
		do
			Result := partition.out + " stats"
		end

feature {NONE} -- Implementation

	add_instance (a_instance, a_next: PM_INSTANCE)
			-- Adds `a_instance' to our stats and `a_next' also if not Void
		require
			a_instance_not_void: a_instance /= Void
			a_next_can_be_void: True
		local
			a_metrics: PM_METRIC_COLLECTION
			l_start, l_next_start, l_end: ROSE_DATE
			l_moment: ROSE_MOMENTS
			l_seconds: DOUBLE
		do
			a_metrics := a_instance.metrics
			l_start := a_instance.runs.last_successful_start_time
			if a_next /= Void then
				l_next_start := a_next.runs.last_successful_start_time
			end
			l_end := a_instance.runs.last_successful_end_time
			if l_start /= Void and then l_end /= Void then
				add_observation (duration, l_end.relative_duration_in_seconds (l_start))
				if l_next_start /= Void then
					l_seconds := l_next_start.relative_duration_in_seconds (l_start)
					add_observation (run_periodicity, l_seconds)
				end
				from
					a_metrics.hash.start
				until
					a_metrics.hash.off
				loop
					l_moment := moments.item (a_metrics.hash.key_for_iteration)
					if l_moment = Void then
						create l_moment.make_empty
						moments.put (l_moment, a_metrics.hash.key_for_iteration)
					end
					add_observation (l_moment, a_metrics.hash.item_for_iteration.value)
					a_metrics.hash.forth
				end
			end
		end

	add_observation (a_moment: ROSE_MOMENTS; a_value: DOUBLE)
			-- Adds `a_value' as an observation of `a_moment'
		require
			a_moment_not_void: a_moment /= Void
		do
			if is_defined_double (a_value) then
				a_moment.add_observation_by_components (a_value, 1.0)
			end
		end

	finalize_moments (a_moment: ROSE_MOMENTS)
			-- If `a_moment' has more than one observation, then winsorize it
		require
			a_moment_not_void: a_moment /= Void
		do
			if a_moment.data.count > 1 then
				a_moment.make_winsorized (a_moment.data,Void,-3,3)
			end
		end

	instance_lookback_count: INTEGER is 100
			-- Number of instances to use to compute statisticss

invariant
	partition_not_void: partition /= Void
	moments_not_void: moments /= Void
	no_void_moment: not moments.has_item (Void)
	duration_not_void: duration /= Void
	run_periodicity_not_void: run_periodicity /= Void
end
