indexing
	description: "Colletion of pre-defined process metrics tests"
	author: "Mauricio Bustos"
	last_editor: "$Author: e2 $"
	date: "$Date: 2014/10/13 $"
	revision: "$Revision: #17 $"
	path: "$File: //rose/source/kernel/pm/pm_metrics_tests.e $"

class
	PM_METRICS_TESTS

inherit
	ROSE_SHARED_GLOBAL
		export {NONE} all end
	ELOG_SHARED_LOGGER

feature -- Access

	compare_with_yesterday_test (a_enable_process_failure: BOOLEAN; a_lower, a_upper: DOUBLE; a_today : ROSE_DATE): FUNCTION[ANY,TUPLE[PM_PROCESS_PARTITION,PM_METRIC],PM_METRIC_TEST_RESULT]
			-- Test if metrics value is within 'a_lower' and 'a_upper' from yesterday's value.
			-- This test is similar to a means_bounds_test with a_lookback set to 1, but it takes care of multiple runs on the same day.
			-- Yesterday is loosely defined as the latest run not from today.
			-- For jobs that are run daily, it will refer to the previous run,
			-- but for rcmd_ind_lead which is run twice daily, it could look 2 runs ago.
		require
			positive_lower_bounds: a_lower > 0
			positive_upper_bounds: a_upper > 0
			a_today_not_void: a_today /= Void
		do
			Result := agent compare_with_yesterday_test_agent (?,?,a_enable_process_failure,a_lower,a_upper, a_today, False)
		ensure
			compare_with_yesterday_test_not_void: Result /= Void
		end

	mean_standard_deviation_test (a_enable_process_failure: BOOLEAN; a_lower, a_upper: DOUBLE; a_lookback: INTEGER): FUNCTION[ANY,TUPLE[PM_PROCESS_PARTITION,PM_METRIC],PM_METRIC_TEST_RESULT]
			-- Test if metrics value is within 'a_lower' and 'a_upper' standard deviations of the historical mean as determined by using 'a_lookback' observations.
			-- For example, to check whether the metric is more than 3 standard deviations away, `a_lower' and `a_upper' would be -3, 3 (NOT 3, 3).
		require
			meaningful_bounds: a_lower < a_upper
		do
			Result := agent mean_standard_deviation_test_agent (?,?,a_enable_process_failure,a_lower,a_upper,a_lookback,False)
		ensure
			mean_standard_deviation_test_not_void: Result /= Void
		end

	mean_bounds_test (a_enable_process_failure: BOOLEAN; a_lower, a_upper: DOUBLE; a_lookback: INTEGER): FUNCTION[ANY,TUPLE[PM_PROCESS_PARTITION,PM_METRIC],PM_METRIC_TEST_RESULT]
			-- Test if metrics value is within bounds as defined by 'a_lower' and 'a_upper' deltas around the historical mean as determinted by using 'a_lookback' observations
		require
			positive_lower_bounds: a_lower > 0
			positive_upper_bounds: a_upper > 0
		do
			Result := agent mean_bounds_test_agent (?,?,a_enable_process_failure,a_lower,a_upper,a_lookback, False)
		ensure
			mean_bounds_test_not_void: Result /= Void
		end

	non_zero_test (a_enable_process_failure: BOOLEAN): FUNCTION[ANY,TUPLE[PM_PROCESS_PARTITION,PM_METRIC],PM_METRIC_TEST_RESULT]
			-- Test if metric value is not zero
		do
			Result := agent non_zero_test_agent (?,?,a_enable_process_failure, False)
		ensure
			non_zero_test_not_void: Result /= Void
		end

	floor_test (a_enable_process_failure: BOOLEAN; a_floor: DOUBLE): FUNCTION[ANY,TUPLE[PM_PROCESS_PARTITION,PM_METRIC],PM_METRIC_TEST_RESULT]
			-- Test if metric value is above 'a_floor'
		do
			Result := agent floor_test_agent (?,?,a_enable_process_failure,a_floor, False)
		ensure
			floor_test_not_void: Result /= Void
		end

	ceiling_test (a_enable_process_failure: BOOLEAN; a_ceiling: DOUBLE): FUNCTION[ANY,TUPLE[PM_PROCESS_PARTITION,PM_METRIC],PM_METRIC_TEST_RESULT]
			-- Test if metric value is below 'a_ceiling'
		do
			Result := agent ceiling_test_agent (?,?,a_enable_process_failure,a_ceiling, False)
		ensure
			ceiling_test_not_void: Result /= Void
		end

	bounds_test (a_enable_process_failure: BOOLEAN; a_lower, a_upper: DOUBLE): FUNCTION[ANY,TUPLE[PM_PROCESS_PARTITION,PM_METRIC],PM_METRIC_TEST_RESULT]
			-- Test if metric value is between 'a_lower' and 'a_upper'
		require
			meaningful_bounds: a_lower <= a_upper
		do
			Result := agent bounds_test_agent (?,?,a_enable_process_failure,a_lower,a_upper, False)
		ensure
			bounds_test_not_void: Result /= Void
		end

feature {NONE} -- Implementation

	todays_metric (a_partition: PM_PROCESS_PARTITION; a_metric: PM_METRIC; a_use_sum: BOOLEAN): PM_METRIC
			-- Metric corresponding to `a_metric'
			-- or else of the sum of that metric (if `a_use_sum') for the last day it successfully ran
		require
			a_partition_not_void: a_partition /= Void
			a_metric_not_void: a_metric /= Void
		local
			l_todays_instance: PM_INSTANCE
		do
			if not a_use_sum then
				Result := a_metric
			else
				l_todays_instance := a_partition.most_recent_succesful_instance
				if l_todays_instance /= Void then
					Result := l_todays_instance.metrics.metric_by_name (a_metric.type.name)
				end
			end
		ensure
			same_if_not_sum: not a_use_sum = (Result = a_metric)
		end

	compare_with_yesterday_test_agent (a_partition: PM_PROCESS_PARTITION; a_metric: PM_METRIC;
		a_enable_process_failure: BOOLEAN; a_lower, a_upper: DOUBLE; a_today: ROSE_DATE;
		a_use_sum: BOOLEAN): PM_METRIC_TEST_RESULT
			-- Implementation agent for compare_with_yesterday_test
			-- using sum of today's metric or directly `a_metric.value' depending on `a_use_sum'
		require
			a_partition_not_void: a_partition /= Void
			a_metric_not_void: a_metric /= Void
			a_today_not_void: a_today /= Void
		local
			l_yesterdays_value, l_todays_value: REAL
			l_yesterdays_instance: PM_INSTANCE
			l_metric_name, l_message: STRING
			l_yesterdays_metric, l_todays_metric: PM_METRIC
		do
			l_yesterdays_instance := a_partition.most_recent_succesful_instance_before (a_today)
			if l_yesterdays_instance /= Void then
				l_metric_name := a_metric.type.name
				l_yesterdays_metric := l_yesterdays_instance.metrics.metric_by_name (l_metric_name)
				l_todays_metric := todays_metric (a_partition, a_metric, a_use_sum)
				if l_yesterdays_metric /= Void and l_todays_metric /= Void then
					l_yesterdays_value := l_yesterdays_metric.value.truncated_to_real
					l_todays_value := l_todays_metric.value.truncated_to_real
					if l_todays_value > l_yesterdays_value + a_upper.truncated_to_real or
						l_todays_value < l_yesterdays_value - a_lower.truncated_to_real
					then
						l_message := "Out of bounds (Lower: "
						l_message.append (scaled_formatted_double (a_lower))
						l_message.append ("; Upper: ")
						l_message.append (scaled_formatted_double (a_upper))
						l_message.append ("; Yesterday: ")
						l_message.append (scaled_formatted_double (l_yesterdays_value))
						l_message.append ("; Today: ")
						l_message.append (scaled_formatted_double (l_todays_value))
						l_message.append_character (')')
						create Result.make (l_message)
						Result.set_enable_process_failure (a_enable_process_failure)
					end
				else
					logger.warn ("Could not find metric '"+l_metric_name+"' in last run! A new item?")
				end
			end
		end

	mean_standard_deviation_test_agent (a_partition: PM_PROCESS_PARTITION; a_metric: PM_METRIC;
		a_enable_process_failure: BOOLEAN; a_lower, a_upper: DOUBLE; a_lookback: INTEGER;
		a_use_sum: BOOLEAN): PM_METRIC_TEST_RESULT
			-- Implementation agent for mean_standard_deviation_test
		require
			meaningful_bounds: a_lower < a_upper
		local
			l_message: STRING
			l_moments: ROSE_MOMENTS
			l_latest_index: INTEGER
			l_mean, l_standard_deviation: DOUBLE
			l_metric: PM_METRIC
		do
			l_moments := a_partition.statistics.moments.item (a_metric.metric_type_id)
			if l_moments /= Void then
				l_latest_index := l_moments.data.upper
				l_mean := l_moments.moving_average_lookback (l_latest_index, a_lookback)
				l_standard_deviation := l_moments.moving_standard_deviation (l_latest_index, a_lookback)
				l_metric := todays_metric (a_partition, a_metric, a_use_sum)
				if l_metric.value > l_mean + l_standard_deviation * a_upper or
					l_metric.value < l_mean + l_standard_deviation * a_lower
				then
					l_message := "Out of bounds (Value: "
					l_message.append (scaled_formatted_double (l_metric.value))
					l_message.append ("; Mean: ")
					l_message.append (scaled_formatted_double (l_mean))
					l_message.append (", Stddev: ")
					l_message.append (scaled_formatted_double (l_standard_deviation))
					l_message.append_character (')')
					create Result.make (l_message)
					Result.set_enable_process_failure (a_enable_process_failure)
				end
			end
		end

	mean_bounds_test_agent (a_partition: PM_PROCESS_PARTITION; a_metric: PM_METRIC;
		a_enable_process_failure: BOOLEAN; a_lower, a_upper: DOUBLE;
		a_lookback: INTEGER; a_use_sum: BOOLEAN): PM_METRIC_TEST_RESULT
			-- Implementation agent for mean_bounds_test
		local
			l_moments: ROSE_MOMENTS
			l_latest_index: INTEGER
			l_mean: DOUBLE
			l_message: STRING
			l_todays_metric: PM_METRIC
		do
			l_moments := a_partition.statistics.moments.item (a_metric.metric_type_id)
			if l_moments /= Void then
				l_latest_index := l_moments.data.upper
				l_mean := l_moments.moving_average_lookback (l_latest_index, a_lookback)
				l_todays_metric := todays_metric (a_partition, a_metric, a_use_sum)
				if l_todays_metric.value > l_mean * (1 + a_upper) or
					l_todays_metric.value < l_mean * (1 - a_lower)
				then
					l_message := "Out of bounds (Value: "
					l_message.append (scaled_formatted_double (l_todays_metric.value))
					l_message.append ("; Mean: ")
					l_message.append (scaled_formatted_double (l_mean))
					l_message.append ("; Lower: ")
					l_message.append (scaled_formatted_double (a_lower))
					l_message.append ("; Upper: ")
					l_message.append (scaled_formatted_double (a_upper))
					l_message.append_character (')')
					create Result.make (l_message)
					Result.set_enable_process_failure (a_enable_process_failure)
				end
			end
		end

	non_zero_test_agent (a_partition: PM_PROCESS_PARTITION; a_metric: PM_METRIC;
		a_enable_process_failure, a_use_sum: BOOLEAN): PM_METRIC_TEST_RESULT
			-- Implementation agent for non_zero_test
		local
			l_todays_metric: PM_METRIC
		do
			l_todays_metric := todays_metric (a_partition, a_metric, a_use_sum)
			if l_todays_metric.value = 0.0 then
				create Result.make ("Zero")
				Result.set_enable_process_failure (a_enable_process_failure)
			end
		end

	floor_test_agent (a_partition: PM_PROCESS_PARTITION; a_metric: PM_METRIC;
		a_enable_process_failure: BOOLEAN; a_floor: DOUBLE; a_use_sum: BOOLEAN): PM_METRIC_TEST_RESULT
			-- Implementation agent for floor_test
		local
			l_message: STRING
			l_todays_metric: PM_METRIC
		do
			l_todays_metric := todays_metric (a_partition, a_metric, a_use_sum)
			if l_todays_metric.value < a_floor then
				l_message := "Below floor (Value: "
				l_message.append (scaled_formatted_double (l_todays_metric.value))
				l_message.append ("; Floor: ")
				l_message.append (scaled_formatted_double (a_floor))
				l_message.append_character (')')
				create Result.make (l_message)
				Result.set_enable_process_failure (a_enable_process_failure)
			end
		end

	ceiling_test_agent (a_partition: PM_PROCESS_PARTITION; a_metric: PM_METRIC;
		a_enable_process_failure: BOOLEAN; a_ceiling: DOUBLE; a_use_sum: BOOLEAN): PM_METRIC_TEST_RESULT
			-- Implementation agent for ceiling_test
		local
			l_message: STRING
			l_todays_metric: PM_METRIC
		do
			l_todays_metric := todays_metric (a_partition, a_metric, a_use_sum)
			if l_todays_metric.value > a_ceiling then
				l_message := "Above ceiling (Value: "
				l_message.append (scaled_formatted_double (l_todays_metric.value))
				l_message.append ("; Ceiling: ")
				l_message.append (scaled_formatted_double (a_ceiling))
				l_message.append_character (')')
				create Result.make (l_message)
				Result.set_enable_process_failure (a_enable_process_failure)
			end
		end

	bounds_test_agent (a_partition: PM_PROCESS_PARTITION; a_metric: PM_METRIC;
		a_enable_process_failure: BOOLEAN; a_lower, a_upper: DOUBLE; a_use_sum: BOOLEAN): PM_METRIC_TEST_RESULT
			-- Implementation agent for bounds_test
		local
			l_message: STRING
			l_todays_metric: PM_METRIC
		do
			l_todays_metric := todays_metric (a_partition, a_metric, a_use_sum)
			if l_todays_metric.value > a_upper or else l_todays_metric.value < a_lower then
				l_message := "Not between boundaries (Value: "
				l_message.append (scaled_formatted_double (l_todays_metric.value))
				l_message.append ("; Lower: ")
				l_message.append (scaled_formatted_double (a_lower))
				l_message.append ("; Upper: ")
				l_message.append (scaled_formatted_double (a_upper))
				l_message.append_character (')')
				create Result.make (l_message)
				Result.set_enable_process_failure (a_enable_process_failure)
			end
		end

	scaled_formatted_double (a_value: DOUBLE): STRING
			-- `a_value' formatted using decimal precision approximated by its magnitude
		do
			if a_value.abs <= 1.0 then
				Result := Global.formatter.formatted_double (a_value, 8, 6)
			elseif a_value.abs < 1_000 then
				Result := Global.formatter.formatted_double (a_value, 8, 4)
			else
				Result := Global.formatter.formatted_double (a_value, 8, 2)
			end
		ensure
			not_void: Result /= Void
			not_empty: not Result.is_empty
		end

end

