note

	description:

		"Definition of a process"

	author: "Zoran Simic"
	last_editor: "$Author: mb $"
	date: "$Date: 2012/07/23 $"
	revision: "$Revision: #25 $"
	path: "$File: //rose/source/kernel/pm/pm_process.e $"

class
	PM_PROCESS

inherit
	PM_COMMON
		redefine
			is_equal, out
		end

create
	default_create

feature -- Access

	process_id: INTEGER
			-- Internal id for this process

	name: STRING
			-- Process name

	success_recipients: STRING
	failure_recipients: STRING
			-- Who to send email to on success/failure

	concurrency: INTEGER
			-- Concurrency level (number of distinct instances that are allowed to run at the same time)

	allowed_runs: INTEGER
			-- Concurrency level (number of distinct instances that are allowed to run at the same time)

	application_id: INTEGER
			-- Id of application associated with this process

	comment: STRING
			-- Comment (description) of this process

	is_equal (other: like Current): BOOLEAN is
			-- Is `other' attached to an object considered
			-- equal to current object?
		do
			Result := process_id = other.process_id
		end

	partition (a_partition_key: STRING): PM_PROCESS_PARTITION is
			-- Partition corresponding to 'a_partition_key', if any
		do
			sql_partition_key.retrieve (Void, <<process_id, a_partition_key>>)
			if sql_partition_key.success and sql_partition_key.array.count>=1 then
				create Result.make(Current, sql_partition_key.array.item(1))
			end
		end

	partitions: HASH_TABLE [PM_PROCESS_PARTITION,STRING] is
			-- All partitions available for this process.
			-- NOTE:
			-- This was formerly used by 'partition', but it is tremendously slow for
			-- ROSE_STUDIO_TESTER (and maybe others - but they are usually exiting)
			-- because it takes about 5 seconds and is called many times.  This
			-- feature might be useful in the future if the returned data is cached.
		local
			l_partition: PM_PROCESS_PARTITION
			i: INTEGER
		do
			create Result.make (0)
			sql_partition_keys.retrieve (Void,<<process_id>>)
			if sql_partition_keys.success then
				from
					i := sql_partition_keys.array.lower
				until
					i > sql_partition_keys.array.upper
				loop
					create l_partition.make (Current,sql_partition_keys.array.item (i))
					Result.put (l_partition,l_partition.partition_key)
					i := i + 1
				end
			end
		end

	metrics: PM_METRIC_COLLECTION is
			-- Metrics generated within this entire process (all partitions combined)
		do
			if saved_metrics = Void then
				create saved_metrics.make_for_process (Current, Void)
			end
			Result := saved_metrics
		end

	metrics_since_date (a_date: ROSE_DATE): PM_METRIC_COLLECTION is
			-- Metrics generated within this entire process (all partitions combined) since 'a_date'
		do
			create Result.make_for_process (Current, a_date)
		ensure
			valid_result: Result /= Void
		end

	resources: PM_RESOURCE_ACCESS_COLLECTION is
			-- Resources accessed by this process (all partitions combined)
		do
			if saved_resources = Void then
				create saved_resources.make_for_process (Current, Void)
			end
			Result := saved_resources
		end

	resources_since_date (a_date: ROSE_DATE): PM_RESOURCE_ACCESS_COLLECTION is
			-- Resources accessed by this process (all partitions combined) since 'a_date'
		do
			create Result.make_for_process (Current, a_date)
		ensure
			valid_result: Result /= Void
		end

	most_recent_instance (a_partition_key: STRING): PM_INSTANCE is
			-- Most recent instance for `a_partition_key';
			-- Void if no such instance
		require
			a_partition_key_not_void: a_partition_key /= Void
		local
			l_predicate: INTEGER_PREDICATE
		do
			if process_metrics.is_available then
				l_predicate := sql_most_recent_instance
				l_predicate.retrieve (Void, <<process_id, a_partition_key>>)
				if not l_predicate.success then
					error.abort ("most_recent_instance", "Failed to retrieve data with predicate `sql_most_recent_instance'.", Void)
				else
					if l_predicate.array /= Void and then not l_predicate.array.is_empty then
						Result := process_metrics.instance_by_id (l_predicate.array.first)
					end
				end
			end
		end

	ten_most_recent_instance_runs (a_partition_key: STRING): ROSE_LINEAR_ARRAY [PM_INSTANCE_RUN] is
			-- Ten most recent instance runs for `a_partition_key';
			-- Includes current run and failed runs.
		require
			a_partition_key_not_void: a_partition_key /= Void
		local
			l_predicate: COLLECTING_PREDICATE [PM_INSTANCE_RUN]
		do
			if process_metrics.is_available then
				l_predicate := sql_ten_most_recent_instance_runs
				l_predicate.retrieve (Void, <<process_id, a_partition_key>>)
				if not l_predicate.success then
					error.abort ("ten_most_recent_instance_runs", "Failed to retrieve data with predicate `sql_ten_most_recent_instance_runs'.", Void)
				else
					Result := l_predicate.array
				end
			end
		end

	most_recent_successful_instance (a_partition_key: STRING): PM_INSTANCE is
			-- Most recent instance for `a_partition_key' with at least one
			-- successfully completed run;
			-- Void if no such instance
		require
			a_partition_key_not_void: a_partition_key /= Void
		local
			l_predicate: INTEGER_PREDICATE
		do
			if process_metrics.is_available then
				l_predicate := sql_most_recent_successful_instance
				l_predicate.retrieve (Void, <<process_id, a_partition_key>>)
				if not l_predicate.success then
					error.abort ("most_recent_successful_instance", "Failed to retrieve data with predicate `sql_most_recent_successful_instance'.", Void)
				else
					if l_predicate.array /= Void and then not l_predicate.array.is_empty then
						Result := process_metrics.instance_by_id (l_predicate.array.first)
					end
				end
			end
		end

	most_recent_completed_instance_run (a_partition_key: STRING): PM_INSTANCE_RUN is
			-- Most recent instance run which is not running anymore;
			-- Void if no such run
		require
			a_partition_key_not_void: a_partition_key /= Void
		local
			l_predicate: COLLECTING_PREDICATE [PM_INSTANCE_RUN]
		do
			if process_metrics.is_available then
				l_predicate := sql_most_recent_completed_instance_run
				l_predicate.retrieve (Void, <<process_id, a_partition_key>>)
				if not l_predicate.success then
					error.abort ("most_recent_completed_instance_run", "Failed to retrieve data with predicate `sql_most_recent_completed_instance_run'.", Void)
				else
					if l_predicate.array /= Void and then not l_predicate.array.is_empty then
						Result := l_predicate.array.first
					end
				end
			end
		end

	ten_most_recent_completed_instance_runs_since_date (a_partition_key: STRING; a_date: ROSE_DATE): ROSE_LINEAR_ARRAY [PM_INSTANCE_RUN] is
			-- Ten most recent completed instance runs since `a_date'
		require
			a_partition_key_not_void: a_partition_key /= Void
			a_date_not_void: a_date /= Void
		local
			l_predicate: COLLECTING_PREDICATE [PM_INSTANCE_RUN]
		do
			if process_metrics.is_available then
				l_predicate := sql_ten_most_recent_completed_instance_runs_since_date
				l_predicate.retrieve (Void, <<process_id, a_partition_key, a_date>>)
				if not l_predicate.success then
					error.abort ("ten_most_recent_completed_instance_runs_since_date", "Failed to retrieve data with predicate `sql_ten_most_recent_completed_instance_runs_since_date'.", Void)
				else
					Result := l_predicate.array
				end
			end
		end

	ten_most_recent_successful_instance_runs (a_partition_key: STRING): ROSE_LINEAR_ARRAY [PM_INSTANCE_RUN] is
			-- Ten most recent successful instance runs
		require
			a_partition_key_not_void: a_partition_key /= Void
		local
			l_predicate: COLLECTING_PREDICATE [PM_INSTANCE_RUN]
		do
			if process_metrics.is_available then
				l_predicate := sql_ten_most_recent_successful_instance_runs
				l_predicate.retrieve (Void, <<process_id, a_partition_key>>)
				if not l_predicate.success then
					error.abort ("ten_most_recent_successful_instance_runs", "Failed to retrieve data with predicate `sql_ten_most_recent_successful_instance_runs'.", Void)
				else
					Result := l_predicate.array
				end
			end
		end

	successful_instance_runs_for_user_since_date (a_user: STRING; a_date: ROSE_DATE): ROSE_LINEAR_ARRAY [PM_INSTANCE_RUN] is
			-- All successful instance runs for `a_user' since `a_date',
			-- regardless of the partition key
			--
			-- `a_user' is expected to be of the form 'login_domain\login_id' (example: "RIEM\ZS").
		require
			a_user_not_void: a_user /= Void
			a_date_not_void: a_date /= Void
		local
			l_predicate: COLLECTING_PREDICATE [PM_INSTANCE_RUN]
		do
			if process_metrics.is_available then
				l_predicate := sql_successful_instance_runs_for_user_since_date
				l_predicate.retrieve (Void, <<process_id, a_date, a_user>>)
				if not l_predicate.success then
					error.abort ("successful_instance_runs_for_user_since_date", "Failed to retrieve data with predicate `sql_successful_instance_runs_for_user_since_date' with date $1 and user $2.", <<a_date, a_user>>)
				else
					Result := l_predicate.array
				end
			end
		end

	instance_runs_since_date (a_partition_key: STRING; a_date: ROSE_DATE): ROSE_LINEAR_ARRAY [PM_INSTANCE_RUN] is
			-- All instance runs since `a_date'
		require
			a_partition_key_not_void: a_partition_key /= Void
			a_date_not_void: a_date /= Void
		local
			l_predicate: COLLECTING_PREDICATE [PM_INSTANCE_RUN]
		do
			if process_metrics.is_available then
				l_predicate := sql_instance_runs_since_date
				l_predicate.retrieve (Void, <<process_id, a_partition_key, a_date>>)
				if not l_predicate.success then
					error.abort ("instance_runs_since_date", "Failed to retrieve data with predicate `sql_instance_runs_since_date' with date $1.", <<a_date>>)
				else
					Result := l_predicate.array
				end
			end
		end

	out: STRING is
			-- New string containing terse printable representation of current object
		do
			Result := name
			if Result = Void then
				Result := "?"
			end
		end

feature -- Basic operations

	set_name (a_name: STRING) is
			-- Set 'name'.
		do
			name := a_name
		end

	set_success_recipients (a_success_recipients: STRING) is
			-- Set 'success_recipients'.
		do
			success_recipients := a_success_recipients
		end

	update_success_recipients (a_success_recipients: STRING) is
			-- Set 'success_recipients' in the database
		local
			l_predicate: EXECUTION_PREDICATE
		do
			create l_predicate.make (process_metrics.server, "update process set success_recipients = ? where name = ?")
			if not l_predicate.success then
				error.abort ("update_success_recipients", "Failed to create predicate `update_success_recipients'.", Void)
			end
			l_predicate.execute (Void, <<a_success_recipients, name>>)
			if not l_predicate.success then
				error.abort ("update_success_recipients", "Failed to execute predicate `update_success_recipients'.", Void)
			end
			set_success_recipients (a_success_recipients)
		end

	update_concurrency (a_concurrency: INTEGER) is
			-- Set 'concurrency' in the database
		local
			l_predicate: EXECUTION_PREDICATE
		do
			create l_predicate.make (process_metrics.server, "update process set concurrency = ? where name = ?")
			if not l_predicate.success then
				error.abort ("update_concurrency", "Failed to create predicate `update_concurrency'.", Void)
			end
			l_predicate.execute (Void, <<a_concurrency, name>>)
			if not l_predicate.success then
				error.abort ("update_concurrency", "Failed to execute predicate `update_concurrency'.", Void)
			end
			set_concurrency (a_concurrency)
		end

	set_failure_recipients (a_failure_recipients: STRING) is
			-- Set 'failure_recipients'.
		do
			failure_recipients := a_failure_recipients
		end

	update_failure_recipients (a_failure_recipients: STRING) is
			-- Set 'failure_recipients' in the database
		local
			l_predicate: EXECUTION_PREDICATE
		do
			create l_predicate.make (process_metrics.server, "update process set failure_recipients = ? where name = ?")
			if not l_predicate.success then
				error.abort ("update_failure_recipients", "Failed to create predicate `update_failure_recipients'.", Void)
			end
			l_predicate.execute (Void, <<a_failure_recipients, name>>)
			if not l_predicate.success then
				error.abort ("update_failure_recipients", "Failed to execute predicate `update_failure_recipients'.", Void)
			end
			set_failure_recipients (a_failure_recipients)
		end

	set_concurrency (a_concurrency: INTEGER) is
			-- Set 'concurrency'.
		do
			concurrency := a_concurrency
		end

	set_allowed_runs (a_allowed_runs: INTEGER) is
			-- Set 'allowed_runs'.
		do
			allowed_runs := a_allowed_runs
		end

	set_application_id (a_application_id: INTEGER) is
			-- Set 'application_id'.
		do
			application_id := a_application_id
		end

	set_comment (a_comment: STRING) is
			-- Set 'comment'.
		do
			comment := a_comment
		end

	update_comment (a_comment: STRING) is
			-- Set 'comment' in the database
		local
			l_predicate: EXECUTION_PREDICATE
		do
			create l_predicate.make (process_metrics.server, "update process set comment = ? where name = ?")
			if not l_predicate.success then
				error.abort ("update_comment", "Failed to create predicate `update_comment'.", Void)
			end
			l_predicate.execute (Void, <<a_comment, name>>)
			if not l_predicate.success then
				error.abort ("update_comment", "Failed to execute predicate `update_comment'.", Void)
			end
			set_comment (a_comment)
		end

feature {NONE} -- Implementation

	saved_metrics: PM_METRIC_COLLECTION
	saved_resources: PM_RESOURCE_ACCESS_COLLECTION

	sql_most_recent_instance: INTEGER_PREDICATE is
			-- SQL query to fetch the most recent instance for a given partition key
		once
			create Result.make (process_metrics.server,
				"SELECT TOP 1 instance_id FROM instance WITH(NOLOCK) WHERE process_id=? AND partition_key=? ORDER BY file_date DESC")
			if not Result.success then
				error.abort ("sql_most_recent_instance", "Failed to create predicate `sql_most_recent_instance'.", Void)
			end
		ensure
			sql_most_recent_instance_not_void: Result /= Void
		end

	sql_most_recent_successful_instance: INTEGER_PREDICATE is
			-- SQL query to fetch the most recent instance with at least
			-- one successfully completed run for a given partition key
		once
			create Result.make (process_metrics.server,"SELECT TOP 1 instance_id FROM successful_run WITH(NOLOCK) WHERE process_id=? AND partition_key=? ORDER BY file_date DESC")
			if not Result.success then
				error.abort ("sql_most_recent_successful_instance", "Failed to create predicate `sql_most_recent_successful_instance'.", Void)
			end
		ensure
			sql_most_recent_successful_instance_not_void: Result /= Void
		end

	sql_most_recent_completed_instance_run: COLLECTING_PREDICATE [PM_INSTANCE_RUN] is
			-- SQL statement used to fetch the most recent instance run which is not running anymore
		once
			create Result.make (process_metrics.server, "[
					SELECT
						TOP 1 instance_run.*
					FROM
						instance WITH(NOLOCK),
						instance_run WITH(NOLOCK)
					WHERE
						instance.process_id=? AND instance.partition_key=? AND
						instance.instance_id = instance_run.instance_id AND
						instance_run.end_time IS NOT NULL
					ORDER BY
						instance_run.start_time DESC
				]")
			if not Result.success then
				error.abort ("sql_most_recent_completed_instance_run", "Failed to create predicate `sql_most_recent_completed_instance_run'.", Void)
			end
		ensure
			sql_most_recent_completed_instance_run_not_void: Result /= Void
		end

	sql_ten_most_recent_completed_instance_runs_since_date: COLLECTING_PREDICATE [PM_INSTANCE_RUN] is
			-- SQL statement used to fetch the ten most recent completed instance runs since a given date
		once
			create Result.make (process_metrics.server, "[
					SELECT
						TOP 10 instance_run.*
					FROM
						instance WITH(NOLOCK),
						instance_run WITH(NOLOCK)
					WHERE
						instance.process_id=? AND instance.partition_key=? AND
						instance.file_date >=? AND
						instance.instance_id = instance_run.instance_id AND
						instance_run.end_time IS NOT NULL
					ORDER BY
						instance_run.start_time DESC
				]")
			if not Result.success then
				error.abort ("sql_ten_most_recent_completed_instance_runs_since_date", "Failed to create predicate `sql_ten_most_recent_completed_instance_runs_since_date'.", Void)
			end
		ensure
			sql_ten_most_recent_completed_instance_runs_since_date_not_void: Result /= Void
		end

	sql_ten_most_recent_instance_runs: COLLECTING_PREDICATE [PM_INSTANCE_RUN] is
			-- SQL statement used to fetch the ten most recent instance runs
		once
			create Result.make (process_metrics.server, "[
					SELECT
						TOP 10 instance_run.*
					FROM
						instance WITH(NOLOCK),
						instance_run WITH(NOLOCK)
					WHERE
						instance.process_id=? AND instance.partition_key=? AND
						instance.instance_id = instance_run.instance_id
					ORDER BY
						instance_run.start_time DESC
				]")
			if not Result.success then
				error.abort ("sql_ten_most_recent_instance_runs", "Failed to create predicate `sql_ten_most_recent_instance_runs'.", Void)
			end
		ensure
			sql_n_most_recent_instance_runs_not_void: Result /= Void
		end

	sql_ten_most_recent_successful_instance_runs: COLLECTING_PREDICATE [PM_INSTANCE_RUN] is
			-- SQL statement used to fetch the ten most recent successful instance runs
		once
			create Result.make (process_metrics.server, "[
					SELECT
						TOP 10 instance_run.*
					FROM
						instance WITH(NOLOCK),
						instance_run WITH(NOLOCK)
					WHERE
						instance.process_id=? AND instance.partition_key=? AND
						instance.instance_id = instance_run.instance_id AND
						instance_run.end_time IS NOT NULL AND instance_run.exit_code > 0
					ORDER BY
						instance_run.start_time DESC
				]")
			if not Result.success then
				error.abort ("sql_ten_most_recent_successful_instance_runs", "Failed to create predicate `sql_ten_most_recent_successful_instance_runs'.", Void)
			end
		ensure
			sql_ten_most_recent_successful_instance_runs_not_void: Result /= Void
		end

	sql_successful_instance_runs_for_user_since_date: COLLECTING_PREDICATE [PM_INSTANCE_RUN] is
			-- SQL statement used to fetch the successful instance runs
			-- for a given user since a given date, regardless of the
			-- partition key
		once
			create Result.make (process_metrics.server, "[
					SELECT
						instance_run.*
					FROM
						instance WITH(NOLOCK),
						instance_run WITH(NOLOCK)
					WHERE
						instance.process_id=? AND 
						instance.file_date >=? AND
						instance.instance_id = instance_run.instance_id AND
						instance_run.end_time IS NOT NULL AND
						instance_run.exit_code > 0 AND
						instance_run.user_name = ?
				]")
			if not Result.success then
				error.abort ("sql_successful_instance_runs_for_user_since_date", "Failed to create predicate `sql_successful_instance_runs_for_user_since_date'.", Void)
			end
		ensure
			sql_successful_instance_runs_for_user_since_date_not_void: Result /= Void
		end

	sql_instance_runs_since_date: COLLECTING_PREDICATE [PM_INSTANCE_RUN] is
			-- SQL statement used to fetch the instance runs since a given date
		once
			create Result.make (process_metrics.server, "[
					SELECT
						instance_run.*
					FROM
						instance WITH(NOLOCK),
						instance_run WITH(NOLOCK)
					WHERE
						instance.process_id=? AND
						instance.partition_key=? AND
						instance.file_date >=? AND
						instance.instance_id = instance_run.instance_id
				]")
			if not Result.success then
				error.abort ("sql_instance_runs_since_date", "Failed to create predicate `sql_instance_runs_since_date'.", Void)
			end
		ensure
			sql_instance_runs_since_date_not_void: Result /= Void
		end

	sql_partition_keys: STRING_PREDICATE is
			-- SQL query used to fetch all partitions available for this process
		once
			create Result.make (process_metrics.server,"SELECT DISTINCT partition_key FROM instance WITH(NOLOCK) WHERE process_id=?")
			if not Result.success then
				error.abort ("sql_partition_keys", "Failed to create predicate `sql_partition_keys'.", Void)
			end
		ensure
			sql_partition_keys_not_void: Result /= Void
		end

	sql_partition_key: STRING_PREDICATE
			-- SQL query used to fetch the partition for this process, matching the specified partition key.
			-- Really, it just verifies that the partition exists.
		once
			create Result.make (process_metrics.server,"SELECT DISTINCT partition_key FROM instance WITH (NOLOCK) WHERE process_id=? and partition_key=?")
			if not Result.success then
				error.abort ("sql_partition_key", "Failed to create predicate `sql_partition_key'.", Void)
			end
		ensure
			sql_partition_key_not_void: Result /= Void
		end

	run_statistics_sample_size: INTEGER is
			-- Size of run statistics sample.
		once
			Result := config.integer ("$system/process_monitor/statistics_sample_size",0)
			if Result = 0 then
				Result := 10
			end
		end

	upper_winsor_zfactor: INTEGER is
			-- Upper statistics boundary for moments calculations
		once
			Result := config.integer ("$system/process_monitor/statistics_winsor_zfactor",0)
			if Result = 0 then
				Result := 3
			end
		end

	lower_winsor_zfactor: INTEGER is
			-- Lower statistics boundary for moments calculations
		once
			Result := -1 * upper_winsor_zfactor
		end

end
