note

	description:

		"Objects that provide standard metric test functions that notify the watched metrics recepients once when called the first time."

	author: "Julian Rogers"
	last_editor: "$Author: e2 $"
	date: "$Date: 2014/10/13 $"
	revision: "$Revision: #6 $"
	path: "$File: //rose/source/kernel/pm/pm_watched_metrics_tests.e $"

class
	PM_WATCHED_METRICS_TESTS

inherit
	PM_METRICS_TESTS

	PM_AWARE

feature -- Access

	compare_with_yesterday_test_and_notify_failure (a_condition: METRIC_TEST_CONDITION; a_enable_process_failure: BOOLEAN; a_lower, a_upper: DOUBLE; a_today : ROSE_DATE; a_record: METRIC_TEST_RECORD): FUNCTION[ANY,TUPLE[PM_PROCESS_PARTITION,PM_METRIC],PM_METRIC_TEST_RESULT] is
			-- Test if metrics value is within 'a_lower' and 'a_upper' from yesterday's value.
			-- This test is similar to a means_bounds_test with a_lookback set to 1, but it takes care of multiple runs on the same day.
			-- Yesterday is loosely defined as the latest run not from today.
			-- For jobs that are run daily, it will refer to the previous run,
			-- but for rcmd_ind_lead which is run twice daily, it could look 2 runs ago.
		require
			a_condition_not_void: a_condition /= Void
			positive_lower_bounds: a_lower > 0
			positive_upper_bounds: a_upper > 0
		do
			Result := agent compare_with_yesterday_test_agent_and_notify_failure (a_condition, ?,?,a_enable_process_failure,a_lower,a_upper, a_today, a_record)
		end

	mean_standard_deviation_test_and_notify_failure (a_condition: METRIC_TEST_CONDITION; a_enable_process_failure: BOOLEAN; a_lower, a_upper: DOUBLE; a_lookback: INTEGER; a_record: METRIC_TEST_RECORD): FUNCTION[ANY,TUPLE[PM_PROCESS_PARTITION,PM_METRIC],PM_METRIC_TEST_RESULT] is
			-- Test if metrics value is within 'a_lower' and 'a_upper' standard deviations of the historical mean as determined by using 'a_lookback' observations
		require
			a_condition_not_void: a_condition /= Void
			meaningful_bounds: a_lower <= a_upper
		do
			Result := agent mean_standard_deviation_test_agent_and_notify_failure (a_condition, ?,?,a_enable_process_failure,a_lower,a_upper,a_lookback, a_record)
		end

	mean_bounds_test_and_notify_failure (a_condition: METRIC_TEST_CONDITION; a_enable_process_failure: BOOLEAN; a_lower, a_upper: DOUBLE; a_lookback: INTEGER; a_record: METRIC_TEST_RECORD): FUNCTION[ANY,TUPLE[PM_PROCESS_PARTITION,PM_METRIC],PM_METRIC_TEST_RESULT] is
			-- Test if metrics value is within bounds as defined by 'a_lower' and 'a_upper' deltas around the historical mean as determinted by using 'a_lookback' observations
		require
			a_condition_not_void: a_condition /= Void
			positive_lower_bounds: a_lower > 0
			positive_upper_bounds: a_upper > 0
		do
			Result := agent mean_bounds_test_agent_and_notify_failure (a_condition, ?,?,a_enable_process_failure,a_lower,a_upper,a_lookback, a_record)
		end

	non_zero_test_and_notify_failure (a_condition: METRIC_TEST_CONDITION; a_enable_process_failure: BOOLEAN; a_record: METRIC_TEST_RECORD): FUNCTION[ANY,TUPLE[PM_PROCESS_PARTITION,PM_METRIC],PM_METRIC_TEST_RESULT] is
			-- Test if metric value is not zero
		require
			a_condition_not_void: a_condition /= Void
		do
			Result := agent non_zero_test_agent_and_notify_failure (a_condition, ?,?,a_enable_process_failure, a_record)
		end

	floor_test_and_notify_failure (a_condition: METRIC_TEST_CONDITION; a_enable_process_failure: BOOLEAN; a_floor: DOUBLE; a_record: METRIC_TEST_RECORD): FUNCTION[ANY,TUPLE[PM_PROCESS_PARTITION,PM_METRIC],PM_METRIC_TEST_RESULT] is
			-- Test if metric value is above 'a_floor'
		require
			a_condition_not_void: a_condition /= Void
		do
			Result := agent floor_test_agent_and_notify_failure (a_condition, ?,?,a_enable_process_failure,a_floor, a_record)
		end

	ceiling_test_and_notify_failure (a_condition: METRIC_TEST_CONDITION; a_enable_process_failure: BOOLEAN; a_ceiling: DOUBLE; a_record: METRIC_TEST_RECORD): FUNCTION[ANY,TUPLE[PM_PROCESS_PARTITION,PM_METRIC],PM_METRIC_TEST_RESULT] is
			-- Test if metric value is below 'a_ceiling'
		require
			a_condition_not_void: a_condition /= Void
		do
			Result := agent ceiling_test_agent_and_notify_failure (a_condition, ?,?,a_enable_process_failure,a_ceiling, a_record)
		end

	bounds_test_and_notify_failure (a_condition: METRIC_TEST_CONDITION; a_enable_process_failure: BOOLEAN; a_lower, a_upper: DOUBLE; a_record: METRIC_TEST_RECORD): FUNCTION[ANY,TUPLE[PM_PROCESS_PARTITION,PM_METRIC],PM_METRIC_TEST_RESULT] is
			-- Test if metric value is between 'a_lower' and 'a_upper'
		require
			meaningful_bounds: a_lower <= a_upper
			a_condition_not_void: a_condition /= Void
		do
			Result := agent bounds_test_agent_and_notify_failure (a_condition, ?,?,a_enable_process_failure,a_lower,a_upper, a_record)
		end

feature {NONE} -- Implementation

	compare_with_yesterday_test_agent_and_notify_failure (a_condition: METRIC_TEST_CONDITION; a_partition: PM_PROCESS_PARTITION; a_metric: PM_METRIC; a_enable_process_failure: BOOLEAN; a_lower, a_upper: DOUBLE; a_today : ROSE_DATE; a_record: METRIC_TEST_RECORD): PM_METRIC_TEST_RESULT is
			-- Implementation agent for compare_with_yesterday_test
		require
			a_condition_not_void: a_condition /= Void
			a_partition_not_void: a_partition /= Void
			a_metric_not_void: a_metric /= Void
			a_record_not_void: a_record /= Void
		do
			if a_condition.should_run_test (a_partition) then
				Result := compare_with_yesterday_test_agent (a_partition, a_metric, a_enable_process_failure, a_lower, a_upper, a_today, a_record.use_sum)
			end
			if Result /= Void then
				send_failure_report_once (a_partition, a_metric, Result, a_record)
			end
		end

	mean_standard_deviation_test_agent_and_notify_failure (a_condition: METRIC_TEST_CONDITION; a_partition: PM_PROCESS_PARTITION; a_metric: PM_METRIC; a_enable_process_failure: BOOLEAN; a_lower, a_upper: DOUBLE; a_lookback: INTEGER; a_record: METRIC_TEST_RECORD): PM_METRIC_TEST_RESULT is
			-- Implementation agent for mean_standard_deviation_test
		require
			a_condition_not_void: a_condition /= Void
			a_partition_not_void: a_partition /= Void
			a_metric_not_void: a_metric /= Void
			a_record_not_void: a_record /= Void
		do
			if a_condition.should_run_test (a_partition) then
				Result := mean_standard_deviation_test_agent (a_partition, a_metric, a_enable_process_failure, a_lower, a_upper, a_lookback, a_record.use_sum)
			end
			if Result /= Void then
				send_failure_report_once (a_partition, a_metric, Result, a_record)
			end
		end

	mean_bounds_test_agent_and_notify_failure (a_condition: METRIC_TEST_CONDITION; a_partition: PM_PROCESS_PARTITION; a_metric: PM_METRIC; a_enable_process_failure: BOOLEAN; a_lower, a_upper: DOUBLE; a_lookback: INTEGER; a_record: METRIC_TEST_RECORD): PM_METRIC_TEST_RESULT is
			-- Implementation agent for mean_bounds_test
		require
			a_condition_not_void: a_condition /= Void
			a_partition_not_void: a_partition /= Void
			a_metric_not_void: a_metric /= Void
			a_record_not_void: a_record /= Void
		do
			if a_condition.should_run_test (a_partition) then
				Result := mean_bounds_test_agent (a_partition, a_metric, a_enable_process_failure, a_lower, a_upper, a_lookback, a_record.use_sum)
			end
			if Result /= Void then
				send_failure_report_once (a_partition, a_metric, Result, a_record)
			end
		end

	non_zero_test_agent_and_notify_failure (a_condition: METRIC_TEST_CONDITION; a_partition: PM_PROCESS_PARTITION; a_metric: PM_METRIC; a_enable_process_failure: BOOLEAN; a_record: METRIC_TEST_RECORD): PM_METRIC_TEST_RESULT is
			-- Implementation agent for non_zero_test
		require
			a_condition_not_void: a_condition /= Void
			a_partition_not_void: a_partition /= Void
			a_metric_not_void: a_metric /= Void
			a_record_not_void: a_record /= Void
		do
			if a_condition.should_run_test (a_partition) then
				Result := non_zero_test_agent (a_partition, a_metric, a_enable_process_failure, a_record.use_sum)
			end
			if Result /= Void then
				send_failure_report_once (a_partition, a_metric, Result, a_record)
			end
		end

	floor_test_agent_and_notify_failure (a_condition: METRIC_TEST_CONDITION; a_partition: PM_PROCESS_PARTITION; a_metric: PM_METRIC; a_enable_process_failure: BOOLEAN; a_floor: DOUBLE; a_record: METRIC_TEST_RECORD): PM_METRIC_TEST_RESULT is
			-- Implementation agent for floor_test
		require
			a_condition_not_void: a_condition /= Void
			a_partition_not_void: a_partition /= Void
			a_metric_not_void: a_metric /= Void
			a_record_not_void: a_record /= Void
		do
			if a_condition.should_run_test (a_partition) then
				Result := floor_test_agent (a_partition, a_metric, a_enable_process_failure, a_floor, a_record.use_sum)
			end
			if Result /= Void then
				send_failure_report_once (a_partition, a_metric, Result, a_record)
			end
		end

	ceiling_test_agent_and_notify_failure (a_condition: METRIC_TEST_CONDITION; a_partition: PM_PROCESS_PARTITION; a_metric: PM_METRIC; a_enable_process_failure: BOOLEAN; a_ceiling: DOUBLE; a_record: METRIC_TEST_RECORD): PM_METRIC_TEST_RESULT is
			-- Implementation agent for ceiling_test
		require
			a_condition_not_void: a_condition /= Void
			a_partition_not_void: a_partition /= Void
			a_metric_not_void: a_metric /= Void
			a_record_not_void: a_record /= Void
		do
			if a_condition.should_run_test (a_partition) then
				Result := ceiling_test_agent (a_partition, a_metric, a_enable_process_failure, a_ceiling, a_record.use_sum)
			end
			if Result /= Void then
				send_failure_report_once (a_partition, a_metric, Result, a_record)
			end
		end

	bounds_test_agent_and_notify_failure (a_condition: METRIC_TEST_CONDITION; a_partition: PM_PROCESS_PARTITION; a_metric: PM_METRIC; a_enable_process_failure: BOOLEAN; a_lower, a_upper: DOUBLE; a_record: METRIC_TEST_RECORD): PM_METRIC_TEST_RESULT is
			-- Implementation agent for bounds_test
		require
			a_condition_not_void: a_condition /= Void
			a_partition_not_void: a_partition /= Void
			a_metric_not_void: a_metric /= Void
			a_record_not_void: a_record /= Void
		do
			if a_condition.should_run_test (a_partition) then
				Result := bounds_test_agent (a_partition, a_metric, a_enable_process_failure, a_lower, a_upper, a_record.use_sum)
			end
			if Result /= Void then
				send_failure_report_once (a_partition, a_metric, Result, a_record)
			end
		end

	sent_reports: TUPLE_KEYED_HASH_TABLE [BOOLEAN, TUPLE[a_partition: STRING; a_metric: INTEGER; a_metric_test_id: INTEGER]]
			-- All reports already sent.
			-- This is used to stop process metrics sending the email multiple times each time "function.item" is called.
		once
			create Result.make (1)
		ensure
			result_not_void: Result /= Void
		end

	send_failure_report_once (a_partition: PM_PROCESS_PARTITION; a_metric: PM_METRIC; a_result: PM_METRIC_TEST_RESULT; a_test_record: METRIC_TEST_RECORD) is
			-- Send failure report based on arguments, if not already sent.
		require
			a_partition_not_void: a_partition /= Void
			a_metric_not_void: a_metric /= Void
			a_result_not_void: a_result /= Void
			a_test_record_not_void: a_test_record /= Void
		local
			l_content: ROSE_MAIL_CONTENT
			l_subject: STRING
			l_current_instance: PM_INSTANCE
			l_process: PM_PROCESS
			l_recipients: STRING
		do
			sent_reports.search ([a_partition.partition_key, a_metric.metric_type_id, a_test_record.metric_test_id])
			if not sent_reports.found then
				sent_reports.put (True, [a_partition.partition_key, a_metric.metric_type_id, a_test_record.metric_test_id])
				create l_content.make
				l_content.set_sender_copied (False)
				l_current_instance := process_metrics.current_instance
				l_process := l_current_instance.process
				if a_result.enable_process_failure then
					l_subject := Global.formatted ("[PM] Failure: $1 failed!", <<l_current_instance >>)
				else
					l_subject := Global.formatted ("[PM] Test Failure: $1 $2 warning!", <<l_current_instance, a_result.message_string>>)
				end
				l_content.set_subject (l_subject)
				l_content.append_text (a_result.message_string + "%N")
				if a_test_record.name /= Void and then not a_test_record.name.is_empty then
					l_content.append_text ("Test name : " + a_test_record.name + "%N")
				end
				if a_test_record.comments /= Void and then not a_test_record.comments.is_empty then
					l_content.append_text ("Test comments : " + a_test_record.comments + "%N")
				end
				l_content.append_text (Global.formatted ("Instance ID  : '$1'%N", <<l_current_instance.instance_id>>))
				l_content.append_text (Global.formatted ("Monitor      : " + global.deliveries.active.path_to_process_monitor + "%N", Void))
				l_recipients := a_test_record.notify
				l_content.to.set (l_recipients)
				if l_content.is_valid then
					mail.send (l_content)
					if mail.last_error /= Void then
						error.signal (generator, "Cannot send email to '$1' for process '$2': $3", <<l_recipients, l_process.name, mail.last_error>>)
					end
				else
					error.signal (generator, "Cannot send email to '$1' for process '$2': $3", <<l_recipients, l_process.name, mail.last_error>>)
				end
			end
		end

	mail: ROSE_MAIL is
			-- Used to send email messages
		once
			create Result.make
		ensure
			valid_result: Result /= Void
		end

end
