indexing
	description: "[
                  Class representing a parsed predicate statement.
				  Substitution for parameters is done here.
				  ]"
	author: "Mark Howard"
	last_editor: "$Author: rj $"
	date: "$Date: 2012/10/10 $"
	revision: "$Revision: #15 $"
	path: "$File: //rose/source/kernel/infrastructure/persistence/predicates/predicate_statement.e $"

class
	PREDICATE_STATEMENT

inherit
	DEBUG_OUTPUT

create
	make

feature {NONE} -- Initialization

	make (a_string: STRING) is
			-- Creation procedure
		require
			valid_string: a_string /= Void
		do
			text := a_string
			parse
		end

feature -- Access

	text: STRING
			-- as is text of the SQL query

	binding: ARRAY [STRING]
			-- list of labeled placeholders and backreferences from `text'

	parsed_string: STRING
			-- the SQL query cleaned up for sending to the backend database server to "prepare"

feature -- Debug

	debug_output: STRING is
			-- debug output
		do
			Result := text
		end

feature {SQL_FACILITATOR} -- Implementation

	parse is
			-- Parse the statement to set up placeholder parameters and a suitable piece os SQL to handoff to the database server.
			-- Note that comments are handled.  THEY ARE NOT RECOMMENDED!  There are quite a few restrictions on what may and
			-- may not appear in a comment.  Usually you are safe if you avoid question marks and apostrophes.  Try not to break
			-- up clauses.  You will find that your statement that give a syntax error works fine in Query Analyzer.
		local
			i, l_index, l_start, l_end, l_name_start, l_name_end: INTEGER
			l_end_found, l_digit_found: BOOLEAN
			l_name: STRING
			l_in_comment: BOOLEAN
			l_multiline_comment_depth: INTEGER
			l_in_string_constant: BOOLEAN
			l_with_no_comments: STRING
		do
			--Handle strings later, i.e. '...'
			create binding.make (1,0)
			create parsed_string.make (text.count)
			l_start := 1
			l_end := text.count
			from
				i := 1
			until
				i > text.count
			loop
				if text.item (i) = '?' and not l_in_string_constant and not l_in_comment and l_multiline_comment_depth = 0
				then
					-- This is a numbered or unnumbered parameter (e.g. "?" or "?3").
					l_index := l_index + 1
					l_name_start := i
					l_end := i - 1
					i := i + 1
					l_end_found := False
					from
						l_digit_found := False
					until
						i > text.count or l_end_found
					loop
						if text.item(i).is_digit then
							l_digit_found := True
							i := i + 1
						else
							l_end_found := True
							l_name_end := i - 1
						end
					end
					if l_digit_found then
						if not l_end_found then
							l_name_end := text.count
						end
						l_name := text.substring (l_name_start,l_name_end)
						if l_name.count>0 then
							l_name.to_upper
							binding.force(l_name,l_index)
							parsed_string.append(text.substring (l_start, l_end))
							parsed_string.append("?")
							l_start := l_name_end + 1
						end
					else
						-- put a placeholder for '?' parameter (may want this to eventually be l_index instead of Void!)
						binding.force (Void, l_index)
					end

				elseif text.item (i) = ':' and then i < text.count and then				-- Is this an object field parameter?
					i > 1 and then not text.item(i - 1).is_digit and then				-- Check for indexed parameters.
					(text.item (i + 1) /= ':' and then text.item (i - 1) /= ':') and	-- Check to make sure this is not an sql function call.
					not l_in_string_constant and
					not l_in_comment and
					l_multiline_comment_depth = 0
				then
					-- This is a named parameter (e.g. ":rxrf").
					l_end := i - 1
					l_index := l_index + 1
					l_name_start := i + 1
					l_end_found := False
					from
						i := i + 1
					until
						i > text.count or l_end_found
					loop
						inspect text.item(i)
						when 'A'..'Z', 'a'..'z', '0'..'9', '_' then
						else
							l_end_found := True
							l_name_end := i - 1
						end
						i := i + 1
					end
					if not l_end_found then
						l_name_end := text.count
					end
					if l_name_start <= l_name_end then
						l_name := text.substring (l_name_start,l_name_end)
						l_name.to_upper
						binding.force(l_name,l_index)
						parsed_string.append(text.substring (l_start,l_end))
						parsed_string.append("?")
						l_start := l_name_end + 1
					end

				elseif text[i] = '%'' and not l_in_comment and l_multiline_comment_depth = 0
				then
					-- String constants are enclosed in apostrophes.
					i := i + 1
					l_in_string_constant := not l_in_string_constant

				elseif i > 1
				then
					if text[i - 1] = '-' and text[i] = '-' and not l_in_string_constant and l_multiline_comment_depth = 0
					then
						-- The beginning of a single line comment (--).
						i := i + 1
						l_in_comment := True

					elseif text[i] = '%N'
					then
						-- The end of a single line comment (if one was in progress).
						i := i + 1
						l_in_comment := False

					elseif text[i - 1] = '/' and text[i] = '*' and not l_in_string_constant
					then
						-- The start of a multiline comment (/*).  These can be nested.
						i := i + 1
						l_multiline_comment_depth := l_multiline_comment_depth + 1

					elseif text[i - 1] = '*' and text[i] = '/' and not l_in_string_constant
					then
						-- The end of a multiline comment (*/).
						i := i + 1
						l_multiline_comment_depth := l_multiline_comment_depth - 1

					else
						i := i + 1
					end

				-- Do not add any 'elseif's here.  They will not be reached if i > 1.

				else
					i := i + 1
				end

				debug("show_sql_with_no_comments")
					if l_with_no_comments = Void
					then
						l_with_no_comments := ""
					end

					if not l_in_comment and l_multiline_comment_depth = 0
					then
						l_with_no_comments.append_character(text[i - 1])
					end
				end
			end

			debug("show_sql_with_no_comments")
				if l_with_no_comments /= Void
				then
					-- This is useful to prove that the comment handling code really works.
					-- Note that "/* */" will change to "//" and "--" will change to "-".
					print("%N")
					print(l_with_no_comments)
					print("%N")
				end
			end

			if l_start <= text.count then
				parsed_string.append (text.substring (l_start, text.count))
			end
		end

invariant

	non_void_text: text /= Void
	non_void_binding: binding /= Void
	non_void_parsed_string: parsed_string /= Void

end
