note
	description: "Performs authentication (well, not in V2) and authorization for a PERSISTENT_SERVER."
	author: "Randy John"
	last_editor: "$Author: rj $"
	date: "$Date: 2015/03/25 $"
	revision: "$Revision: #1 $"
	path: "$File: //rose/source/kernel/persistence/databases/signature_tagged/ps_security_system_v2.e $"

class
	PS_SECURITY_SYSTEM_V2

inherit
	PS_SECURITY_SYSTEM
	PS_SECURITY_SYSTEM_V2_CLIENT
		redefine
			set_authorization_server
		end

create
	make

create {PERSISTENT_SERVER}
	default_create

feature {NONE} -- Initialization.

	make(a_persistent_server: PERSISTENT_SERVER)
	-- Initialize the authorization system for use by 'a_persistent_server'.
	require
		a_persistent_server_not_void: a_persistent_server /= Void
	do
		persistent_server := a_persistent_server
		authorization_server_definition_path := a_persistent_server.server.definition.full_path
		set_authorization_server
	end

	set_signature_id_predicate
	-- Sets 'signature_id_predicate'.
	local
		l_statement: STRING
	do
		l_statement := global.formatted(
											"[
											select top 1 $2 from $1 with (nolock)
											where
												$2 >=
													(
														select max($2) - 1 from $1 with (nolock)
													) * rand()
											order by $2
											]",
											<<authorization_table_name, authorization_column_name>>)
		create signature_id_predicate.make(authorization_server, l_statement)
		-- There is no check for success since the most common failure is loss of connection and
		-- we have the ability to remake this predicate.
	end

feature {PERSISTENT_SERVER_CLIENT_COMMON} -- Initialization.

	set_authorization_server
	-- Set 'authorization_server' and anything associated with it.
	do
		if authorization_server = Void
		then
			Precursor
			set_signature_id_predicate	-- It's OK if this fails during creation.  We can remake it later.
		end
	end

feature -- Implementation.

	pre_authorization_connection(
									a_outflow: ROSE_NETWORK_STREAM_SOCKET;
									a_message: PORT_SERVER_MESSAGE;
									a_original_command: PERSISTENT_SERVER_COMMAND
								): PS_SECURITY_SYSTEM_V2_CLIENT_COMMAND
	-- Returns a connection capable of dealing with the initial, not yet authorized client.
	require
		a_outflow_not_void: a_outflow /= Void
		a_message_not_void: a_message /= Void
		a_original_command_not_void: a_original_command /= Void
	do
		create Result.make_pre_authorization(a_outflow, a_message, a_original_command, persistent_server, Current)
	end

	new_checksum_statement: STRING
	-- An SQL statement that will return the checksum.  Result will be used by both the client and the server.
	local
		i: INTEGER
		l_retry_count: INTEGER
		l_ids: ARRAY[INTEGER]
		l_id_string: STRING
	do
		if l_retry_count /= 0
		then
			if authorization_server /= Void
			then
				authorization_server.sql_disconnect
			end

			authorization_server := Void
			global.env.sleep(5000)
			set_authorization_server
		end

		-- Get some random signature_id's.
		from
			create l_ids.make_empty
			i := 1
		until
			i > authorization_checksum_count
		loop
			signature_id_predicate.retrieve(Void, Void)
			if not signature_id_predicate.success
			then
				raise("signature_id_predicate failed")
			end

			l_ids.force(signature_id_predicate.array[1], l_ids.upper + 1)
			i := i + 1
		end

		-- Build a STRING out of them.
		from
			l_id_string := l_ids[1].out
			i := 2
		until
			i > authorization_checksum_count
		loop
			l_id_string.append(", " + l_ids[i].out)

			i := i + 1
		end

		-- Create the final statement.
		Result := global.formatted(
									"[
										select cast(CHECKSUM_AGG(a.cs) as varchar)
										from
											(
												select CHECKSUM(*) as cs from $1
												where
													$2 in ($3)
											) a
									]",
									<<authorization_table_name, authorization_column_name, l_id_string>>)
	rescue
		l_retry_count := l_retry_count + 1
		if l_retry_count <= 3
		then
			retry
		end
	end

	characteristics: ROSE_LINEAR_ARRAY[TUPLE[name: STRING; value: STRING]]
	-- <Precursor>
	do
		create Result.make_empty
		Result.extend(["authorization_table_name", authorization_table_name.out])
		Result.extend(["authorization_column_name", authorization_column_name.out])
		Result.extend(["authorization_checksum_count", authorization_checksum_count.out])
	end

feature -- Once/constants.

	authorization_checksum_count: INTEGER
	-- The number of checksums we will read.
	once
		Result :=  persistent_server.config_integer("psss/v2/authorization_checksum_count", 2).max(2)
	ensure
		result_at_least_two: Result >= 2
	end

	authorization_column_name: STRING
	-- The name of a column in 'authorization_table_name'.
	once
		Result := persistent_server.config_value("psss/v2/authorization_column_name", "signature_id")
	ensure
		result_not_void: Result /= Void
	end

	authorization_table_name: STRING
	-- The name of the authorization table.
	once
		Result := persistent_server.config_value("psss/v2/authorization_table_name", "data_signatures")
	ensure
		result_not_void: Result /= Void
	end

feature {NONE} -- Attributes.

	persistent_server: detachable PERSISTENT_SERVER			-- The server for whom we are performing this service.

	signature_id_predicate: detachable INTEGER_PREDICATE	-- Used to select a random signature_id.

end
