note
	description: "Handles the client-side portion of security for a PERSISTENT_SERVER."
	author: "Randy John"
	last_editor: "$Author: rj $"
	date: "$Date: 2015/04/16 $"
	revision: "$Revision: #2 $"
	path: "$File: //rose/source/kernel/persistence/databases/signature_tagged/ps_security_system_v2_client.e $"

class
	PS_SECURITY_SYSTEM_V2_CLIENT

inherit
	PS_SECURITY_SYSTEM_CLIENT
	ROSE_SHARED_GLOBAL
	EXCEPTIONS

feature {PERSISTENT_SERVER_CLIENT_COMMON} -- Initialization.

	set_authorization_server
	-- Set 'authorization_server' and anything associated with it.
	local
		l_database_definition: ROSE_DATABASE_DEFINITION
	do
		if authorization_server = Void
		then
			create l_database_definition.make_from_path("PS_SECURITY_SYSTEM_V2_CLIENT", "", authorization_server_definition_path)
			authorization_server := (create {DATABASE_SERVERS}).pool.get_server_from_definition(l_database_definition)	-- This server is NOT shared.
		end
	end

	set_authorization_server_definition_path(a_authorization_server_definition_path: STRING)
	-- Set 'authorization_server_definition_path'.
	require
		a_authorization_server_definition_path_not_void: a_authorization_server_definition_path /= Void
	do
		authorization_server_definition_path := a_authorization_server_definition_path
	ensure
		authorization_server_definition_path_same_as_arg: authorization_server_definition_path ~ a_authorization_server_definition_path
	end

feature -- Implementation.

	checksum_from_statement(a_statement: STRING): STRING
	-- Execute 'a_statement' on 'authorization_server'.  'a_statement' is an SQL selection statement returning a STRING.
	require
		a_statement_not_void: a_statement /= Void
	local
		l_retry_count: INTEGER
		l_pred: STRING_PREDICATE
	do
		Result := ""	-- In case of accident, make sure Result is not Void since sending a Void via storable can cause an ACCVIO.

		if l_retry_count /= 0
		then
			if authorization_server /= Void
			then
				authorization_server.sql_disconnect
			end

			if authorization_servers /= Void
			then
				authorization_servers.remove(authorization_server_definition_path)
			end

			if l_retry_count /= 1
			then
				global.env.sleep(5000)
			end

			authorization_server := Void
			set_authorization_server
		end

		create l_pred.make(authorization_server, a_statement)
		if not l_pred.success
		then
			raise("Failed to create the predicate")
		end

		l_pred.retrieve(Void, Void)
		if not l_pred.success or else l_pred.array.upper /= 1 or else l_pred.array[1] = Void
		then
			raise("The predicate failed")
		end

		Result := l_pred.array[1]
	ensure
		can_be_sent_by_independent_store: Result /= Void
	rescue
		l_retry_count := l_retry_count + 1
		if l_retry_count <= 4
		then
			retry
		end
	end

	set_authorization_server_from_cache(a_definition_path: STRING)
	-- Set 'authorization_server'.
	-- Use 'authorization_servers' to see if we have already seen this server.
	require
		a_definition_path_not_void: a_definition_path /= Void
	local
		l_definition_path: STRING
	do
		l_definition_path := a_definition_path.as_lower
		set_authorization_server_definition_path(l_definition_path)

		if authorization_servers = Void
		then
			create authorization_servers.make(5)
		end

		authorization_server := authorization_servers[l_definition_path]

		if authorization_server = Void
		then
			set_authorization_server

			if authorization_server /= Void
			then
				authorization_servers[l_definition_path] := authorization_server
			end
		end
	ensure
		not_a_twin: authorization_server /= Void implies authorization_servers.has_item(authorization_server)
	end

feature -- Once/constants.

	k_send_authorization_request: STRING = "send_authorization_request"
			-- The client sends this to start the authorization process.  If you change it you must release clients and servers at the same time!

	name: STRING
	-- <Precursor>
	do
		Result := "version_2"
	end

feature {PS_SECURITY_SYSTEM_V2_CLIENT_COMMAND} -- Attributes.

	authorization_server_definition_path: STRING
		-- Used to create 'authorization_server'.  In the form used by ROSE_DATABASE_DEFINITION ("sql:server.database").

feature {NONE} -- Attributes.

	authorization_server: SQL_DATA_SERVER		-- The server that contains the authorization table.

	authorization_servers: HASH_TABLE[like authorization_server, STRING]
												-- All known servers indexed by their lowercased definition path.

end
