note
	description: "A client command used by PS_SECURITY_SYSTEM_V2."
	author: "Randy John"
	last_editor: "$Author: rj $"
	date: "$Date: 2015/03/25 $"
	revision: "$Revision: #1 $"
	path: "$File: //rose/source/kernel/persistence/databases/signature_tagged/ps_security_system_v2_client_command.e $"

class
	PS_SECURITY_SYSTEM_V2_CLIENT_COMMAND

inherit
	PS_SECURITY_SYSTEM_CLIENT_COMMAND
		redefine
			security_system
		end
	PERSISTENT_SERVER_COMMAND
		redefine
			execute
		end
	EXCEPTIONS
		export
			{NONE} all
		end

create
	make_pre_authorization

create {PS_SECURITY_SYSTEM_V2}
	default_create

feature {NONE} -- Initialization.

	make_pre_authorization(
							a_outflow: ROSE_NETWORK_STREAM_SOCKET;
							a_message: PORT_SERVER_MESSAGE;
							a_original_command: PERSISTENT_SERVER_COMMAND;
							a_parent: PERSISTENT_SERVER;
							a_security_system: PS_SECURITY_SYSTEM_V2)
	-- Makes a "not yet authorized" client.
	require
		a_outflow_not_void: a_outflow /= Void
		a_message_not_void: a_message /= Void
		a_original_command_not_void: a_original_command /= Void
		a_parent_not_void: a_parent /= Void
		a_security_system_not_void: a_security_system /= Void
	do
		make_command(a_outflow)
		state := initial_state
		original_port_server_message := a_message
		original_command := a_original_command
		parent := a_parent
		security_system := a_security_system
	ensure
		outflow_shared_with_arg: outflow = a_outflow	-- This is set in 'make_command'.
		state_is_initial: state = initial_state
		original_port_server_message_same_as_arg: original_port_server_message ~ a_message
		original_command_same_as_arg: original_command ~ a_original_command
		parent_shared_with_arg: parent = a_parent
		security_system_shared_with_arg: security_system = a_security_system
	end

feature -- Implementation.

	command_for_statistics: PERSISTENT_SERVER_COMMAND
	-- <Precursor>
	do
		Result := original_command
	end

feature {NONE} -- Implementation.

	execute(a_message: like received)
	-- Handles messages sent by the client.  The message is in 'received' (after calling 'receive').  'a_message' is only
	-- used for deferred messages (which are not supported here).
	local
		l_retry: BOOLEAN
		l_in_database: BOOLEAN
		l_checksum: STRING
	do
		if not l_retry
		then
			create last_message_time.make_now
			number_of_messages := number_of_messages + 1

			receive		-- Sets 'received'.

			inspect
				state
			when initial_state then
				-- The client's first message must be an authorization request.
				if attached {STRING} received as al_received and then
				   al_received ~ security_system.k_send_authorization_request
				then
					l_in_database := True
					checksum_statement := security_system.new_checksum_statement
					l_in_database := False
					resend(<<security_system.authorization_server_definition_path, checksum_statement>>)
					state := waiting_for_response_state
				else
					parent.write_connect_status_message("Client DENIED (no authorization request)", outflow.port_number_string, original_port_server_message, command_line)
					resend("SERVER_EXCEPTION: No authorization request.")
					expired := True
				end
			when waiting_for_response_state then
				-- Check to see if his response decodes to the correct value.
				if attached {STRING} received as al_received
				then
					l_in_database := True
					l_checksum := security_system.checksum_from_statement(checksum_statement)
					l_in_database := False

					if al_received ~ l_checksum
					then
						resend(create {ANY})	-- Sending ANY to indicate success.
						state := authorized_state
					else
						-- It's barely possible that the data changed milliseconds ago.  Give him one more chance.
						l_in_database := True
						checksum_statement := security_system.new_checksum_statement
						l_in_database := False
						resend(<<security_system.authorization_server_definition_path, checksum_statement>>)
						state := waiting_for_response_2_state
					end
				else
					parent.write_connect_status_message("Client DENIED (invalid authorization)", outflow.port_number_string, original_port_server_message, command_line)
					resend("SERVER_EXCEPTION: Invalid authorization.")
					expired := True
				end
			when waiting_for_response_2_state then
				-- Check to see if his second response decodes to the correct value.
				if attached {STRING} received as al_received
				then
					l_in_database := True
					l_checksum := security_system.checksum_from_statement(checksum_statement)
					l_in_database := False

					if al_received ~ l_checksum
					then
						resend(create {ANY})	-- Sending ANY to indicate success.
						state := authorized_state
					else
						parent.write_connect_status_message("Client DENIED (wrong authorization)", outflow.port_number_string, original_port_server_message, command_line)
						resend("SERVER_EXCEPTION: Wrong authorization.")
						expired := True
					end
				else
					parent.write_connect_status_message("Client DENIED (invalid authorization (2))", outflow.port_number_string, original_port_server_message, command_line)
					resend("SERVER_EXCEPTION: Invalid authorization.")
					expired := True
				end
			when authorized_state then
				original_command.execute_with_received(received, a_message)
			when invalid_state then
				do_nothing
			end
		else
			expired := True

			if l_in_database
			then
				resend("SERVER_EXCEPTION: The server failed while performing a security check.")
			end
		end

		if expired
		then
			state := invalid_state
		end
	rescue
		if tag_name /= Void and then tag_name.count > 0
		then
			if equal("invalid retrieve type", tag_name.as_lower) or						-- This was the error before 6.4.
			   equal("retrieve: unable to read type of storable.", tag_name.as_lower)	-- We started getting this error in 6.4.
			then
				-- This is the standard error when the client disconnects.  Nothing to print.
			elseif equal("net error #10060", tag_name.as_lower)
			then
				-- Socket timeout.
				output("PS_SECURITY_SYSTEM_V2_CLIENT_COMMAND:execute - Socket timeout, state = " + state.out + ".%N", Void)
			else
				-- Hmmm.  An unusual message.
				output("PS_SECURITY_SYSTEM_V2_CLIENT_COMMAND:execute - " + tag_name + ", state = " + state.out + ".%N", Void)
			end
		end

		if not l_retry
		then
			l_retry := True
			retry
		end
	end

	state_id_valid: BOOLEAN
	-- Does 'state' have a valid value.
	do
		Result :=
			state = invalid_state or
			state = initial_state or
			state = waiting_for_response_state or
			state = waiting_for_response_2_state or
			state = authorized_state
	end

feature -- Once/constants.

	invalid_state: INTEGER = 0							-- No messages should be received while in this state.

	initial_state: INTEGER = 1							-- We are waiting for the client to request the authorization key.

	waiting_for_response_state: INTEGER = 2				-- We are waiting for the client to respond with a message encoded by the token.

	waiting_for_response_2_state: INTEGER = 3			-- Second chance version of 'waiting_for_response_state'.

	authorized_state: INTEGER = 4						-- This client has been authorized.  Pass all messages through.

feature {PERSISTENT_SERVER} -- Attributes (alpha order).

	original_command: PERSISTENT_SERVER_COMMAND			-- The command that knows how to process requests after authorization.

feature {NONE} -- Attributes (alpha order).

	checksum_statement: STRING							-- SQL statement that both the server (here) and the client will use to determine a checksum.

	original_port_server_message: PORT_SERVER_MESSAGE	-- The message that came with our first contact with the client.

	parent: PERSISTENT_SERVER							-- Once the client is authorized, we will register 'new_connection' here.

	security_system: PS_SECURITY_SYSTEM_V2				-- The security system that we negotiated with the client.

	state: INTEGER										-- The current state of the connection (see the constants ending in "_state").

invariant
	correct_state: state_id_valid

end
