indexing
	description: "Load or save/export configuration"
	author: "Zoran Simic"
	last_editor: "$Author: zs $"
	date: "$Date: 2011/03/24 $"
	revision: "$Revision: #18 $"
	path: "$File: //rose/source/kernel/infrastructure/framework/command/rcmd_config.e $"

class
	RCMD_CONFIG

inherit
	ROSE_COMMAND_GENERAL

feature -- Access

	id: STRING is "config"
			-- Command id

	signature: STRING is "load:s import:s save:s export:s rename:s type:s default:? dump:? silent:?"
			-- Argument signature of this command (defines valid parameters to be passed to the command)

feature -- Output

	help: STRING is
			-- Help on this command
"[
Load/save or import/export application's configuration
-load          Load app's conventional config
-load=<path>   Load config from given <path>
-import=<path> Import config from <path> into app's current config
-save=<path>   Save app's config to <path>
-export=<path> Export app's config to <path> (config is flattened out)
-rename=<path> Change app's config path to point to <path>
-type=<type>   Specify config type to load/save/... (one of cfg, xml, ...)
-default       Set default installation
-dump          Dump current config on current output
-silent        Do not report errors
]"

feature {NONE} -- Implementation

	do_execute is
			-- Execute effectively the command.
		local
			l_path, l_type: STRING
			l_site_name: STRING
		do
			if parameters.boolean ("silent") then
				error.push_private (Void)
			end
			l_type := parameters.value ("type")
			if parameters.provided ("load") then
				l_path := parameters.value ("load")
				if not is_significant (l_path) then
					l_path := Global.app.program_name + Global.convention.default_config_extension
				end
				Global.config_manager.load (config, l_type, l_path)
				if error.signaled then
					error.signal (Global.app.program_name, "Configuration is not available", Void)
					error.flush_and_acknowledge
				else
					trace ("Configuration file loaded from '$1'%N", <<config.path>>)
				end
				if parameters.boolean ("default") then
					if not global.app.is_running_in_production then
						l_site_name := global.app.delivery.value_default ("/path/name", Void)
						if l_site_name = Void then
							-- Try harder
							l_site_name := global.deliveries.known_delivery_id (global.app.site.folder)
						end
						if l_site_name = Void then
							warn ("do_execute", "Test environment is unknown and system is not running in the production delivery", Void)
							warn ("do_execute", "Running in unknown environment: $1%N", <<global.app.site.folder>>)
						else
							trace ("Running in test environment: $1%N", <<l_site_name.as_upper>>)
						end
					end
				end
			end
			if not error.signaled and then parameters.provided ("import") then
				l_path := parameters.value ("import")
				Global.config_manager.import (config.root, l_type, l_path, False, False)
				if not error.signaled then
					trace ("Configuration imported from '$1'%N", <<l_path>>)
				end
			end
			if not error.signaled and then parameters.provided ("save") then
				l_path := parameters.value ("save")
				if not is_significant (l_path) then
					if not config.is_modified then
						trace ("Configuration is not modified, proceeding with save anyway", Void)
					end
					l_path := config.path
				end
				if is_significant (l_path) then
					Global.config_manager.save (config, l_type, l_path)
					if not error.signaled then
						trace ("Configuration file saved to '$1'%N", <<l_path>>)
					end
				else
					error.signal (id, "No path to save configuration to", Void)
				end
			end
			if not error.signaled and then parameters.provided ("export") then
				l_path := parameters.value ("export")
				Global.config_manager.flat_export (config, l_type, l_path)
				if not error.signaled then
					trace ("Configuration file exported to '$1', path is still '$2'%N", <<l_path, config.path>>)
				end
			end
			if not error.signaled and then parameters.provided ("rename") then
				config.set_path (parameters.value ("rename"))
				if l_type /= Void then
					config.set_type (l_type)
				end
				trace ("Configuration path is now '$1'%N", <<config.path>>)
			end
			if not error.signaled and then parameters.boolean ("dump") then
				output ("------------------------------%N", Void)
				output ("-- Config path: $1%N", <<config.path>>)
				output ("------------------------------%N", Void)
				Global.config_manager.flat_export (config, Global.convention.default_config_type, "-")
			end
			Global.app.variables.set ("config", config.path)
			if parameters.boolean ("silent") then
				error.pop_private
			end
		end

	is_significant (a_string: STRING): BOOLEAN is
			-- Is a_string non-void and non-empty?
		do
			Result := a_string /= Void and then not a_string.is_empty
		end

end
