indexing
	description: "Error reporter dialog"
	author: "Zoran Simic"
	date: "$Date: 2008/12/01 $"
	revision: "$Revision: #12 $"

class
	RCMD_EV_ERROR_REPORTER

inherit
	RCMD_WINDOW
		redefine
			auto_initialize_window
		end

feature -- Initialization

	auto_initialize_window is
			-- Create the command
		do
			Precursor
			output_buffer
		end

feature -- Access

	id: STRING is "ev_error"
			-- Command id

	signature: STRING is ""
			-- Argument signature of this command (defines valid parameters to be passed to the command)

	buffer: ROSE_OUTCHANNEL_BUFFER is
			-- Output is buffered here, for refreshing
		once
			create Result.make
		end

feature -- Status report

	is_unique: BOOLEAN is True
			-- Is this window unique for the whole application?

	is_auto_pop: BOOLEAN is
			-- Does this window pop-up automatically when an error is output?
		do
			Result := not Gui.remembered.boolean (k_auto_pop_option)
		end

feature -- Status setting

feature -- Output

	help: STRING is
			-- Help on this command
"[
Show error-reporter dialog
This dialog catches all errors and warnings issued by the application
]"

feature -- Basic operations

	check_prerequisites is
			-- Check that command's execution pre-requisites are met.
			-- Signal error if pre-requisistes are not met.
		do
		end

	output_message (a_level: INTEGER; a_message: STRING) is
			-- Output 'a_message' with 'a_level'.
		do
			buffer.output_message (a_level,a_message)
			do_output (a_level,a_message)
		end

	show_fatal_error is
			-- Show fatal error, user has no choice but to terminate application.
		local
			l_dialog: EV_DIALOG
		do
			auto_initialize_window
			window.set_icon_pixmap (Gui.rep.pixmaps.Error_pixmap)
			continue_button.disable_sensitive
			error_report_button.set_foreground_color (Gui.rep.colors.Red)
			error_report_button.font.set_weight (Gui.rep.fonts.Weight_bold)
			l_dialog ?= window
			if l_dialog /= Void then
				l_dialog.set_default_push_button (terminate_button)
				l_dialog.set_default_cancel_button (terminate_button)
			end
			show
			error.flush_and_acknowledge
		end

feature {NONE} -- Initialization

	do_execute is
			-- Execute effectively the command.
		do
			show
		end

	output_box: EV_TEXT
	error_report_button: EV_BUTTON
	continue_button: EV_BUTTON
	terminate_button: EV_BUTTON
			-- Widgets used outside of setup ('initialize_window').

	initialize_window is
			-- Initialize 'window'.
		local
			l_hbox: EV_HORIZONTAL_BOX
			l_vbox: EV_VERTICAL_BOX
			l_dialog: EV_DIALOG
			l_widget: EV_WIDGET
		do
			l_dialog := Gui.factory.new_dialog (Global.app.name + " error reporter",Gui.rep.pixmaps.Error_pixmap,0,0)
			window := l_dialog
			l_vbox := Gui.factory.new_vbox (l_dialog,4,-4)
				--| Message output box
			output_box := Gui.factory.new_text (l_vbox,"",-350,-300)
				--| OK button
			l_hbox := Gui.factory.new_hbox (l_vbox,0,8)
			l_widget := Gui.remembered.new_check_button (l_hbox,k_auto_pop_option,"Don't pop this window on error",Void,1)
			error_report_button := Gui.factory.new_button (l_hbox,"  Report this  ",agent send_error_report,1)
			error_report_button.set_foreground_color (Gui.rep.colors.Dark_blue)
			continue_button := Gui.factory.new_button (l_hbox,"     Continue     ",agent hide,1)
			terminate_button := Gui.factory.new_button (l_hbox," Terminate application ",agent Gui.destroy,1)
			l_dialog.set_default_push_button (continue_button)
			l_dialog.set_default_cancel_button (continue_button)
				--| Window menu
			Gui.factory.add_menu (l_dialog,False,<<
				["File",
					["Send this by mail...",agent send_error_report],
					["Console","console"],
					Void,
					["Save all",agent save (True)],
					["Save selected",agent save (False)],
					Void,
					["Close this window",agent hide],
					["Exit application","ev_quit"]],
				["Edit",
					["Clear output",agent output_box.remove_text],
					Void,
					["Copy all"],
					["Copy selected"],
					Void,
					["Find..."],
					["Find next"]],
				["Help",[Gui.factory.default_help_menu]]
			>>)
			is_auto_closed := True
		end

	save (a_all: BOOLEAN) is
			-- Save output to a user selected file.
		local
			l_filename: STRING
			l_file: PLAIN_TEXT_FILE
			rescued: STRING
		do
			if rescued /= Void then
			else
				l_filename := Gui.dialogs.select_save_file (window,"error_output","*.txt")
				if l_filename /= Void then
					create l_file.make_open_write (l_filename)
					if a_all then
						l_file.put_string (output_box.text)
					else
						l_file.put_string (output_box.selected_text)
					end
					l_file.close
				end
			end
		rescue
			if rescued = Void then
				rescued := last_exception_explanation
				retry
			end
		end

	output_buffer is
			-- Output contents of 'buffer' in 'output_box'
		local
			l_list: SEQUENCE [PAIR [INTEGER,STRING]]
		do
			l_list := clone (buffer.buffer)
			if output_box /= Void and then l_list /= Void then
				from
					output_box.remove_text
					l_list.start
				until
					l_list.off
				loop
					do_output (l_list.item.first,l_list.item.second)
					l_list.forth
				end
			end
		end

	do_output (a_level: INTEGER; a_message: STRING) is
			-- Output 'a_message' with 'a_level' in the 'output_box'
		local
			l_message: STRING
		do
			if a_message /= Void and then window /= Void then
				if is_auto_pop then
					show
				end
				window.set_icon_pixmap (Gui.rep.pixmaps.Warning_pixmap)
				l_message := Global.prefs.decorated_verbose_message (a_level,a_message)
				l_message.prune_all ('%R')		--| FIXME: temporary fix for Vision2 assertion problem
				output_box.append_text (l_message)
			end
		end

	send_error_report is
			-- Send current error message by mail.
		do
			Gui.dialogs.interactive_report (window,Void,Void,"[Error report] $application v$version",Global.interpreted (k_mail_body),<<output_box.text>>)
		end

	k_mail_body: STRING is
"[
Hi,

The following error happened when running $application v$version:

$1
]"

	k_auto_pop_option: STRING is "no_autopop_error"
			-- Configuration key for auto-pop option

end
