indexing
	description: "Execute commands from given source (input, config, parameter)"
	author: "Zoran Simic"
	date: "$Date: 2008/12/01 $"
	revision: "$Revision: #10 $"

class
	RCMD_EXECUTE

inherit
	ROSE_COMMAND_GENERAL

feature -- Access

	id: STRING is "execute"
			-- Command id

	signature: STRING is "cmd:s input:s output:s cfg:s stop_on_error:?"
			-- Argument signature of this command (defines valid parameters to be passed to the command)

	parsed_lines: INTEGER
		-- Number of parsed lines during execution

	execute_errors: INTEGER
		-- Execute errors counter

feature -- Status report

	fatal_errors: BOOLEAN
		-- Do errors stop execution ?

feature -- Basic operations

	execute_line (a_line: STRING): INTEGER is
			-- Execute 'a_line', return the number of commands actually executed.
		local
			m: RCMD_MACRO
			l_initial_output, l_output: ROSE_OUTCHANNEL
			p: PAIR [STRING,STRING]
		do
			l_initial_output := Global.io_handler.output
			p := Global.formatter.split_on_last (a_line,redirection_separator,Global.convention.command_escapes)
			if p.second /= Void then
				p.second.right_adjust
				p.second.left_adjust
				if not equal (p.second,"-") then
					l_output := Global.io_handler.manager.checked_output_channel (id,p.second)
					error.flush_and_acknowledge
				end
				if l_output /= Void and then l_output.is_open then
					Global.io_handler.set_output (l_output)
				end
			end
			m := Global.commands.macro_command (p.first)
			if m /= Void then
				m.execute_prepared
				if not error.signaled then
					Result := Result + m.count
				else
					execute_errors := execute_errors + m.count
				end
				parsed_lines := parsed_lines + 1
			elseif error.signaled then
				parsed_lines := parsed_lines + 1
				execute_errors := execute_errors + 1
			end
			if l_output /= Void and then l_output.is_open then
				l_output.close
			end
			if fatal_errors then
				error.flush
			else
				error.flush_and_acknowledge
			end
			Global.io_handler.set_output (l_initial_output)
		end

	reset_counters is
			-- Reset counters.
		do
			execute_errors := 0
			parsed_lines := 0
		end

feature -- Output

	help: STRING is
			-- Help on this command
"[
Execute commands from an input channel and/or configuration
-cmd=<command>     Execute <command> prior to anything else
-input=<channel>   Execute all commands read on <channel> until that <channel> is closed/exhausted
-output=<channel>  Direct all commands output to <channel>
-cfg=<key>         Execute all commands constituted by all children of <key> in config
-stop_on_error     Stop executing on first error (flush error and continue otherwise)
]"

feature {NONE} -- Implementation

	do_execute is
			-- Execute effectively the command.
		local
			initial_output,l_output: ROSE_OUTCHANNEL
			n: INTEGER
		do
			reset_counters
			initial_output := Global.io_handler.output
			fatal_errors := parameters.boolean ("stop_on_error")
				--| Procede '-output' command option
			if parameters.provided ("output") then
				l_output := Global.io_handler.manager.checked_output_channel (id,parameters.value ("output"))
				if not error.signaled then
					trace ("output redirected to '$1' during this execute%N",<<l_output>>)
					Global.io_handler.set_output (l_output)
				end
			end
				--| Procede '-cmd' command option
			if not error.signaled then
				n := n + execute_line (parameters.value ("cmd"))
			end
				--| Procede '-cfg' command option
			if not error.signaled
			and then parameters.provided ("cfg") then
				n := n + procede_config (parameters.value ("cfg"))
			end
				--| Procede '-input' command option
			if not error.signaled
			and then parameters.provided ("input") then
				n := n + procede_input (parameters.value ("input"))
			end
			trace ("$1 commands processed successfully%N",<<n>>)
			trace ("$1 commands produced errors%N",<<execute_errors>>)
			trace ("$1 parsed lines%N",<<parsed_lines>>)
			Global.io_handler.set_output (initial_output)
			if l_output /= Void then
				l_output.close
			end
		end

	procede_config (a_key: STRING): INTEGER is
			-- Procede commands from configuration 'a_key'.
		local
			l_node: ROSE_CONFIGURATION_NODE
			l_cell: CELL [INTEGER]
		do
			l_node := config.node (a_key)
			if l_node = Void then
				error.warn ("No node '$1' defined in configuration",<<a_key>>)
			elseif l_node.children = Void then
				error.warn ("No commands in node '$1' in configuration",<<a_key>>)
			else
				create l_cell.put (0)
				l_node.do_all (agent execute_config_node (?,l_cell))
				Result := l_cell.item
				trace ("$1 commands from configuration node '$2' processed successfully%N",<<Result,a_key>>)
			end
		end

	execute_config_node (a_node: ROSE_CONFIGURATION_NODE; a_count: CELL [INTEGER]) is
		do
			a_count.put (a_count.item + execute_line (a_node.value))
		end

	procede_input (a_channel: STRING): INTEGER is
			-- Procede input from 'a_channel'.
		local
			s, rescued: STRING
			stop: BOOLEAN
			initial_input,l_input: ROSE_INCHANNEL
		do
			if rescued /= Void then
				error.signal (id,"Error while processing channel '$1': $2",<<a_channel,rescued>>)
			end
			if not error.signaled then
				initial_input := Global.io_handler.input
				l_input := Global.io_handler.manager.checked_input_channel (id,a_channel)
			end
			if not error.signaled and then l_input /= Void then
				Global.io_handler.set_input (l_input)
			end
			from
			until
				stop or else error.signaled or else not Global.io_handler.has_input
			loop
				s := Global.io_handler.input.ask (Global.prefs.prompt)
				stop := s = Void
				if not stop then
					Result := Result + execute_line (s)
				end
			end
			Global.io_handler.set_input (initial_input)
			if l_input /= Void then
				trace ("Execution terminated on channel '$1'%N",<<l_input>>)
				l_input.close
			end
			trace ("$1 commands from input '$2' processed successfully%N",<<Result,a_channel>>)
		rescue
			if rescued = Void then
				rescued := last_exception_explanation
				retry
			end
		end

	redirection_separator: STRING is ">"
			-- String recognised as a redirection directive

end -- class RCMD_CONSOLE
