indexing
	description: "Memory usage information"
	author: "Zoran Simic"
	date: "$Date: 2008/12/01 $"
	revision: "$Revision: #3 $"

class
	RCMD_MEMORY

inherit
	ROSE_COMMAND_GENERAL

	MEMORY
		export
			{NONE} all
		end

feature -- Access

	id: STRING is "memory"
			-- Command id

	signature: STRING is "collect:s gc:s verbose:?"
			-- Argument signature of this command (defines valid parameters to be passed to the command)

feature -- Output

	help: STRING is
			-- Help on this command
"[
Display information on memory/gc usage
-verbose         Display extensive information on memory usage/statistics
-collect=<mode>  Set internal memory management behaviour, <mode> is one of
   on:      Turn memory collection (GC) on
   off:     Turn memory collection (GC) off
   go:      Collect dead objects
   full:    Perform a full collection
   fast:    Select 'fast' memory allocation strategy, (least GC calls)
   compact: Select 'compact' memory allocation strategy, (slower)
   tiny:    Select 'tiny' memory allocation strategy, (least memory consumed)
-gc=<option>     Set/display GC statistics modes, <option> is one of:
   on:      Turn GC time accounting on
   off:     Turn GC time accounting off
   full:    Display GC full collection statistics
   inc:     Display GC incremental collection statistics
]"

feature {NONE} -- Implementation

	do_execute is
			-- Execute effectively the command.
		local
			s: STRING
			l_done: BOOLEAN
		do
			if parameters.provided ("collect") then
				l_done := True
				s := parameters.value ("collect")
				if equal (s,"on") then
					output ("Garbage collection is turned on%N",Void)
					collection_on
				elseif equal (s,"off") then
					output ("Garbage collection is turned off%N",Void)
					collection_off
				elseif equal (s,"go") then
					output ("Collecting dead objects ... ",Void)
					collect
					output ("OK%N",Void)
				elseif equal (s,"full") then
					output ("Full-collecting dead objects ... ",Void)
					full_collect
					output ("OK%N",Void)
				elseif equal (s,"fast") then
					output ("Fast memory allocation selected%N",Void)
					allocate_fast
				elseif equal (s,"compact") then
					output ("Compact memory allocation selected%N",Void)
					allocate_compact
				elseif equal (s,"tiny") then
					output ("Tiny memory allocation selected%N",Void)
					allocate_tiny
				else
					error.warn ("Unknown 'collect' directive '$1'",<<s>>)
				end
			end
			if parameters.provided ("gc") then
				l_done := True
				s := parameters.value ("gc")
				if equal (s,"on") then
					output ("GC monitoring is turned on%N",Void)
					enable_time_accounting
				elseif equal (s,"off") then
					output ("GC monitoring is turned off%N",Void)
					disable_time_accounting
				elseif equal (s,"full") then
					display_gc_info ("Full",gc_statistics (Full_collector))
				elseif equal (s,"inc") then
					display_gc_info ("Incremental",gc_statistics (Incremental_collector))
				else
					error.warn ("Unknown 'gc' directive '$1'",<<s>>)
				end
			end
			if parameters.boolean ("verbose") then
				l_done := True
				display_full_info
			end
			if not l_done then
				display_mem ("C",C_memory)
				display_mem ("Eiffel",Eiffel_memory)
				display_mem ("Total",Total_memory)
			end
		end

	display_mem (a_title: STRING; a_type: INTEGER) is
			-- Display memory usage for 'a_type', with 'a_title'.
		local
			m: MEM_INFO
		do
			m := memory_statistics (a_type)
			output ("$1 memory: $2K ($3K used, $4K free, $5K overhead)%N",<<a_title,kbytes (m.total),kbytes (m.used),kbytes (m.free),kbytes (m.overhead)>>)
		end

	display_full_info is
		do
			output ("Memory threshold: $1K%N",<<kbytes (memory_threshold)>>)
			output ("Full collection period: $1 collections%N",<<collection_period>>)
			output ("Coalesce period: $1 collections%N",<<coalesce_period>>)
			output ("Collecting: $1%N",<<collecting>>)
			output ("Largest coalesced block: $1K%N",<<kbytes (largest_coalesced_block)>>)
			output ("Max mem: $1K%N",<<kbytes (max_mem)>>)
			output ("Chunk size: $1K%N",<<kbytes (chunk_size)>>)
			output ("Tenure: $1 cycles%N",<<tenure>>)
			output ("Generation object limit: $1 bytes%N",<<generation_object_limit>>)
			output ("Scavenge zone size: $1K%N",<<kbytes (scavenge_zone_size)>>)
		end

	display_gc_info (a_title: STRING; a_info: GC_INFO) is
		do
			output ("$1-collector GC statistics%N",<<a_title>>)
			output ("Cycle count: $1%N",<<a_info.cycle_count>>)
			output ("Memory used: $1K%N",<<kbytes (a_info.memory_used)>>)
			output ("Collected: $1K%N",<<kbytes (a_info.collected)>>)
			output ("Collected average: $1K%N",<<kbytes (a_info.collected_average)>>)
			output ("Real time: $1 ms%N",<<a_info.real_time*10>>)
			output ("Real time average: $1 ms%N",<<a_info.real_time_average*10>>)
			output ("Real interval time: $1 ms%N",<<a_info.real_interval_time*10>>)
			output ("Real interval time average: $1 ms%N",<<a_info.real_interval_time_average*10>>)
			output ("CPU time: $1 s%N",<<format_double.formatted (a_info.cpu_time)>>)
			output ("CPU time average: $1 s%N",<<format_double.formatted (a_info.cpu_time_average)>>)
			output ("CPU interval time: $1 s%N",<<format_double.formatted (a_info.cpu_interval_time)>>)
			output ("CPU interval time average: $1 s%N",<<format_double.formatted (a_info.cpu_interval_time_average)>>)
			output ("System time: $1 s%N",<<format_double.formatted (a_info.sys_time)>>)
			output ("Systme time average: $1 s%N",<<format_double.formatted (a_info.sys_time_average)>>)
			output ("System interval time: $1 s%N",<<format_double.formatted (a_info.sys_interval_time)>>)
			output ("System interval time average: $1 s%N",<<format_double.formatted (a_info.sys_interval_time_average)>>)
		end

	kbytes (a_value: INTEGER): INTEGER is
		do
			Result := a_value // 1024
		end

	format_double: FORMAT_DOUBLE is
		once
			create Result.make (6,3)
		end

end -- class RCMD_MEMORY
