indexing
	description: "Display info on a command and ask for it's parameters interactively"
	author: "Zoran Simic"
	date: "$Date: 2008/12/01 $"
	revision: "$Revision: #3 $"

class
	RCMD_SHOW

inherit
	ROSE_COMMAND_GENERAL

feature -- Access

	id: STRING is "show"
			-- Command id

	signature: STRING is "cmd:s"
			-- Argument signature of this command (defines valid parameters to be passed to the command)

feature -- Output

	help: STRING is
			-- Help on this command
"[
Display info on a command and ask for it's parameters interactively
]"

feature {NONE} -- Implementation

	do_execute is
			-- Execute effectively the command.
		local
			l_id: STRING
		do
			l_id := parameters.value ("cmd")
			if l_id = Void or else l_id.is_empty then
				Global.commands.execute ("help show")
			elseif Global.commands.has (l_id) then
				show_command (Global.commands.service (l_id))
			elseif Global.commands.has_category (l_id) then
				Global.commands.execute ("help " + l_id)
			else
				error.signal (id,"Unknown command '$1'",<<l_id>>)
			end
		end

	show_command (a_command: ROSE_COMMAND) is
		require
			valid_command: a_command /= Void
		local
			l_params: STRING
			l_alias: RCMD_ALIAS_MACRO
		do
			l_alias ?= a_command
			if l_alias = Void then
				output ("$1: $2%N$3%N--------------------%N",<<a_command.id,a_command.signature,a_command.help>>)
			else
				output ("$1: $2%N--------------------%N",<<a_command.id,a_command.help>>)
			end
			if Global.io_handler.is_interactive then
				if l_alias /= Void then
					l_params := ask ("Alias parameters: ")
				else
					l_params := get_parameters (a_command)
				end
				output ("--------------------%N",Void)
				if l_params = Void then
					output ("Command not executed%N",Void)
				else
					Global.commands.execute (a_command.id + " " + l_params)
				end
			else
				output ("Current I/O is not interactive (input: '$1', output: '$2')%N",<<Global.io_handler.input,Global.io_handler.output>>)
			end
		end

	get_parameters (a_command: ROSE_COMMAND): STRING is
		local
			l_parameters: LINKED_LIST [ROSE_COMMAND_PARAMETER_ITEM]
			l_string: STRING
		do
			if a_command.parameters_signature = Void then
			elseif a_command.parameters_signature.is_unsigned then
				Result := ask ("Parameters: ")
			elseif a_command.parameters_signature.is_empty then
				if ask_boolean ("No parameters needed, execute command") then
					Result := ""
				end
			else
				l_parameters := a_command.parameters_signature.parameters_list
				from
					l_parameters.start
					create Result.make (0)
				until
					Result = Void or else l_parameters.off
				loop
					l_string := get_parameter (l_parameters.item)
					if l_string = Void then
						Result := Void
					else
						Result.append (l_string)
					end
					l_parameters.forth
				end
			end
		end

	get_parameter (a_parameter: ROSE_COMMAND_PARAMETER_ITEM): STRING is
		local
			l_question: STRING
		do
			if a_parameter /= Void then
				l_question := Global.formatted ("$1 [$2]",<<a_parameter.name,Global.convention.command_parameter_type_representation (a_parameter.type)>>)
				if a_parameter.is_optional then
					l_question.append (" (optional)")
				else
					l_question.append (" (required)")
				end
				if a_parameter.is_boolean then
					if ask_boolean (l_question) then
						Result := " -" + a_parameter.name
					elseif a_parameter.is_optional then
						Result := ""
					else
						Result := " -" + a_parameter.name + "=True"
					end
				else
					Result := ask (l_question + ": ")
					if a_parameter.is_optional then
						if Result = Void then
							Result := ""
						elseif a_parameter.valid_value (Result) then
							Result := " -" + a_parameter.name + "=%"" + Global.formatter.escape (Result,"%"",Global.convention.command_escapes.item (1)) + "%""
						else
							output ("'$1' is not a valid value for $2 parameter '$3'%N",<<Result,Global.convention.command_parameter_type_representation (a_parameter.type),a_parameter.name>>)
							Result := Void
						end
					elseif Result = Void then
						output ("$1 parameter '$2' is required%N",<<Global.convention.command_parameter_type_representation (a_parameter.type),a_parameter.name>>)
					elseif a_parameter.valid_value (Result) then
						Result := " -" + a_parameter.name + "=%"" + Global.formatter.escape (Result,"%"",Global.convention.command_escapes.item (1)) + "%""
					else
						output ("'$1' is not a valid value for $2 parameter '$3'%N",<<Result,Global.convention.command_parameter_type_representation (a_parameter.type),a_parameter.name>>)
						Result := Void
					end
				end
			end
		end

	ask (a_question: STRING): STRING is
			-- Answer to 'a_question', Void if no answer
		local
			retried: BOOLEAN
		do
			if retried then
				Result := Void
			else
				Result := Global.io_handler.input.ask (a_question)
				if Result /= Void and then Result.is_empty then
					Result := Void
				end
			end
		rescue
			if not retried then
				retried := True
				retry
			end
		end

	ask_boolean (a_question: STRING): BOOLEAN is
			-- Answer to yes/no 'a_question', with short instruction on how to answer
		require
			valid_question: a_question /= Void
		local
			l_answer: STRING
		do
			l_answer := ask (a_question + " (y/n)? ")
			if l_answer /= Void and then not l_answer.is_empty then
				l_answer.to_lower
				Result :=  l_answer.item (1) = 'y'
			end
		end

end -- class RCMD_SHOW
