indexing
	description: "Display application usage help"
	author: "Zoran Simic"
	date: "$Date: 2008/12/01 $"
	revision: "$Revision: #12 $"

class
	RCMD_SPAWN

inherit
	ROSE_COMMAND_GENERAL

feature -- Access

	id: STRING is "spawn"
			-- Command id

	signature: STRING is "*"
			-- Argument signature of this command (defines valid parameters to be passed to the command)

	last_return_code: INTEGER
			-- Return code for last spawning

feature -- Output

	help: STRING is
			-- Help on this command
"[
Spawn a process
Usage: spawn [options] <program> [<arg1> <arg2>...]
[options] are one of (optional)
-system   Use 'system' call to spawn new process
-wait     Wait for the spawned process to finish
-nowait   Spawn and don't wait for termination
-overlay  Spawned process overlays current
-nowaito  Spawn and don't wait for termination
-detach   Detach spawned process from current
-noenv    Do not use PATH environment variable
]"

feature -- Basic operations

	shell_execute (a_command: STRING) is
			-- Execute 'a_command' using local shell.
			-- Program is stopped until 'a_command' terminates.
		require
			valid_command: a_command /= Void and then not a_command.is_empty
		local
			rescued: STRING
		do
			if rescued = Void then
				Global.env.system (a_command)
				last_return_code := Global.env.return_code
				check_return_code (a_command,last_return_code)
			else
				error.signal (id,"Could not execute '$1': $2",<<a_command,rescued>>)
				last_return_code := -1
			end
		rescue
			if rescued = Void then
				rescued := last_exception_explanation
				retry
			end
		end

	spawn_line (a_string: STRING) is
			-- Spawn 'a_string' after interpreting its first argument (recognized if starting with a '-')
			-- Format is as follows:
			-- -system cmd args: execute via shell, and constitute a synchronous call
			-- everything else is an asynchronous call
			-- -[wait|nowait|overlay|nowaito|detach] [-env] command {args}
			-- 'args' are considered space separated, default mode is '-wait'
			-- Path environment variable is not searched if -noenv is specified
		local
			l_pair: PAIR [STRING,STRING]
			l_args: ARRAY [STRING]
			l_string: STRING
			l_mode: INTEGER
			l_noenv: BOOLEAN
		do
			if a_string /= Void then
					--| Remove eventual extra heading spaces
				a_string.left_adjust
			end
			if a_string = Void or else a_string.is_empty then
					--| Shouldn't be called with no arguments
				error.signal (id,"Nothing to spawn",Void)
			end
			if not error.signaled then
					--| Default mode is NOWAIT
				l_mode := NOWAIT
				l_string := a_string
				if a_string.item (1) = '-' then
						--| Check for switches (these are not in the signature to allow a smoother
						--| parameter passing to 'spawn', i.e otherwise users would have
						--| to write, for example: spawn -nowait -cmd="..." -args="..."
						--| which can get cumbersome if dealing with spaces in arguments...)
					l_pair := Global.formatter.split_on_first (a_string," %T",Void)
					if equal (l_pair.first,"-system") then
							--| Perform a shell call if -system specified
						shell_execute (l_pair.second)
						l_string := Void
					elseif equal (l_pair.first,"-wait") then
							--| This app will wait for the spawned process to finish
						l_mode := WAIT
						l_string := l_pair.second
					elseif equal (l_pair.first,"-nowait") then
							--| This app will not wait for the spawned process to finish
							--| Returned error code is only for launch success check
						l_mode := NOWAIT
						l_string := l_pair.second
					elseif equal (l_pair.first,"-overlay") then
							--| Replace current process with called one
							--| Returned error code is only for launch success check
						l_mode := OLD_P_OVERLAY
						l_string := l_pair.second
					elseif equal (l_pair.first,"-nowaito") then
							--| This app will not wait for the spawned process to finish
							--| Returned error code is only for launch success check
						l_mode := NOWAITO
						l_string := l_pair.second
					elseif equal (l_pair.first,"-detach") then
							--| Spawned process is detached from current process
							--| Returned error code is only for launch success check
						l_mode := DETACH
						l_string := l_pair.second
					elseif equal (l_pair.first,"-noenv") then
							--| Environment variable PATH is not searched to find the spawned command
							--| (it is searched by default)
						l_noenv := True
						l_string := l_pair.second
					else
							--| Signal all other '-' beginnings, and refuse to execute them
						error.signal (id,"Parameter '$1' is not recognized",<<l_pair.first>>)
					end
				end
				if l_string /= Void and then not l_string.is_empty and then l_string.item (1) = '-' then
					l_pair := Global.formatter.split_on_first (l_string," %T",Void)
					if equal (l_pair.first,"-noenv") then
							--| Environment variable PATH is not searched to find the spawned command
							--| (it is searched by default)
							--| -noenv can appear after a mode specification
						l_noenv := True
						l_string := l_pair.second
					end
				end
			end
			if not error.signaled and then l_string /= Void then
					--| Get args to spawned process (only for non -system invocations)
					--| Arguments are considered space separated, quote the ones with spaces
				l_pair := Global.formatter.split_on_first (l_string," %T",Void)
				l_args := Global.formatter.split_array (l_pair.second," %T",Void,False)
				spawn (l_mode, l_pair.first, l_args, l_noenv)
			end
		end

	spawn (a_mode: INTEGER a_program: STRING; a_args: ARRAY [STRING]; a_noenv: BOOLEAN) is
			-- Spawn (using 'a_mode') 'a_program' with 'a_args', searching path unless 'a_noenv' true.
		local
			l_argv: ARRAY [POINTER]
			l_any: ANY
			i,j: INTEGER
			l_program_name: POINTER
			l_void_pointer: POINTER
			l_collecting: BOOLEAN
		do
			l_collecting := memory.collecting
			memory.collection_off
			last_return_code := 0
			if a_args = Void or else a_args.count = 0 then
					--| No arguments to pass to spawned process
					--| However, C API requires first argument to be program name
					--| So we create a one element array to pass program name in
				create l_argv.make (0,0)
			else
					--| We have the arguments here in 'a_args', let's make C pointers to those
					--| Quotes delimiting arguments are removed here
				from
-- zs				create l_argv.make (a_args.lower-1, a_args.upper)
-- ds				create l_argv.make (a_args.lower, a_args.upper + 1)
					create l_argv.make (0, a_args.count + 1)
					i := a_args.lower
					j := 1
				until
					i > a_args.upper
				loop
					-- ignore first arg iff "." or program name
					if not (i = a_args.lower and then (a_args.item(1).is_equal (".") or else a_args.item(1).is_equal (a_program))) then
						if a_args.item (i) /= Void then
							l_argv.put (c_string (unquoted (a_args.item (i))), j)
						j := j + 1
						end
					end
					i := i + 1
				end

				l_argv.put(l_void_pointer, j)		-- Terminating NULL.
			end
			l_program_name := c_string (unquoted (a_program))
				--| C API requires first argument to be program name
			l_argv.put (l_program_name, 0)
			l_any := l_argv.area
			if a_noenv then
			   last_return_code := c_spawnv (a_mode, l_program_name, $l_any)
			else
			   last_return_code := c_spawnvp (a_mode, l_program_name, $l_any)
			end

			if l_collecting
			then
				memory.collection_on
			end

			check_return_code (a_program,last_return_code)
			if not error.signaled then
				trace ("Spawned process has pid '$1'%N",<<last_return_code>>)
			end
		end

feature -- Constants

	WAIT         	: INTEGER is 0
	NOWAIT       	: INTEGER is 1
	OLD_P_OVERLAY  	: INTEGER is 2
	NOWAITO      	: INTEGER is 3
	DETACH       	: INTEGER is 4
			-- Spawning modes

feature {NONE} -- Implementation

	check_return_code (a_program: STRING; a_code: INTEGER) is
			-- Check 'last_return_code' and eventually signal corresponding error.
			-- Only negative values for 'last_return_code' are considered errors.
		do
			if last_return_code = -1 then
				error.signal (id,"Program '$1' was not found",<<a_program>>)
			elseif last_return_code < 0 then
				error.signal (id,"Spawning '$1' returned code $2",<<a_program,last_return_code>>)
			end
		end

	memory: MEMORY is
			-- Access to garbage collector
		once
			create Result
		ensure
			memory_not_void: Result /= Void
		end

	do_execute is
			-- Execute effectively the command.
		do
			spawn_line (parameters.text)
		end

	unquoted (a_string: STRING): STRING is
			-- 'a_string' unquoted if quoted
		do
			Result := a_string
			if Result /= Void and then Result.count > 1 then
				if Result.item (1) = '"' and then Result.item (Result.count) = '"' then
					Result.remove (Result.count)
					Result.remove (1)
				end
			end
		end

	c_string (a_string: STRING): POINTER is
			-- Return a C char * for 'a_string'
		local
			l_any: ANY
		do
			if a_string /= Void then
				l_any := a_string.to_c
				Result := $l_any
			end
		end

	c_system (l_system_command: POINTER): INTEGER is
			-- C wrapper for 'system' call (execute a command with local shell)
		external
			"C [macro <process.h>] (const char *) : EIF_INTEGER"
		alias
			"system"
		end

	c_spawnv (l_mode: INTEGER; l_process_name: POINTER; l_argv: POINTER): INTEGER is
			-- C wrapper for 'spawnv' call, spawn a program with no path searching
		external
			"C [macro <process.h>] (int, const char *, const char * const *) : EIF_INTEGER"
		alias
			"_spawnv"
		end

	c_spawnvp (l_mode: INTEGER; l_process_name: POINTER; l_argv: POINTER): INTEGER is
			-- C wrapper for 'spawnv' call, spawn a program with path searching
		external
			"C [macro <process.h>] (int, const char *, const char * const *) : EIF_INTEGER"
		alias
			"_spawnvp"
		end

end -- class RCMD_SPAWN
