indexing
	description: "Set system for the application"
	author: "Zoran Simic"
	last_editor: "$Author: zs $"
	date: "$Date: 2010/08/19 $"
	revision: "$Revision: #10 $"
    path: "$File: //rose/source/kernel/infrastructure/framework/command/rcmd_system.e $"

class
	RCMD_SYSTEM

inherit
	ROSE_COMMAND_GENERAL

feature -- Access

	id: STRING is "system"
			-- Command id

	signature: STRING is "set:s display:? refresh:? start:?"
		-- Argument signature of this command (defines valid parameters to be passed to the command)

feature -- Output

	help: STRING is
			-- Help on this command
"[
Set/display application's system node from configuration
The 'system' node is conventionally an indirection for system-based configuration
-set=<key> Set <key> to be the system node in configuration
-display   Display information on current system node
-refresh   Refresh preferences attached to current system node
-start     As -refresh, but execute also commands in $system/execute
]"

feature {NONE} -- Implementation

	do_execute is
			-- Execute effectively the command.
		local
			l_system: STRING
			l_system_node: ROSE_CONFIGURATION_NODE
		do
			if parameters.provided ("set") then
				config.set (Global.convention.system_keyword,parameters.value ("set"))
			end
			l_system := config.value (Global.convention.system_keyword)
			if l_system /= Void and then not l_system.is_empty then
				l_system_node := config.node (l_system)
			else
				l_system := Void
			end
			if parameters.boolean ("display") then
				if l_system = Void then
					output ("No system defined in configuration%N",Void)
				elseif l_system_node = Void then
					output ("System '$1' is not configured%N",<<l_system>>)
				elseif l_system_node.deep_count = 0 then
					output ("System '$1' is empty%N",Void)
				else
					output ("System '$1' configuration :%N$2",<<l_system,l_system_node>>)
				end
			end
			if parameters.boolean ("start") or else parameters.boolean ("refresh") then
				Global.prefs.update_from_config (l_system_node)
				if l_system_node /= Void then
					redirect_channels (l_system_node.node ("io"))
					define_aliases (l_system_node.node ("alias"))
					if parameters.boolean ("start") then
						execute_commands (l_system_node.node ("execute"))
					end
				end
			end
		end

	redirect_channels (a_node: ROSE_CONFIGURATION_NODE) is
			-- Redirect channels as defined by 'a_node' (key,value) pairs.
		do
			if a_node /= Void and then a_node.children /= Void then
				trace ("Redirecting channels configured in '$1'%N", <<a_node.path>>)
				a_node.do_all (agent redirect_channel)
			end
		end

	redirect_channel (a_node: ROSE_CONFIGURATION_NODE) is
			-- Redirect channel defined in 'a_node' (redirect channel pointed to by node's key to path contained in node's value)
		local
			l_key, l_value: STRING
			l_input: ROSE_INCHANNEL
			l_output: ROSE_OUTCHANNEL
		do
			if a_node /= Void then
				l_key := a_node.key
				l_value := a_node.value
				if l_key /= Void then
					l_key.left_adjust
					l_key.right_adjust
					l_key.to_lower
				end
				if l_key = Void or else l_key.is_empty then
					--| Do nothing
				elseif equal (l_key,"input") then
					if l_value = Void or else l_value.is_empty then
						Global.io_handler.set_input (Void)
					else
						l_input := Global.io_handler.manager.input_channel (l_value)
						if l_input = Void then
							error.warn ("'$1': redirection of input channel to invalid path '$2' ignored",<<a_node.path,l_value>>)
						else
							Global.io_handler.set_input (l_input)
						end
					end
				elseif equal (l_key,"output") then
					if l_value = Void or else l_value.is_empty then
						Global.io_handler.set_output (Void)
					else
						l_output := get_output_channel (a_node.key,l_key,l_value)
						if l_output /= Void then
							Global.io_handler.set_output (l_output)
						end
					end
				elseif equal (l_key,"echo") then
					if l_value = Void or else l_value.is_empty then
						Global.io_handler.set_echo (Void)
					else
						l_output := get_output_channel (a_node.key,l_key,l_value)
						if l_output /= Void then
							Global.io_handler.set_echo (l_output)
						end
					end
				elseif equal (l_key,"log") then
					if l_value = Void or else l_value.is_empty then
						Global.io_handler.set_log (Void)
					else
						l_output := get_output_channel (a_node.key,l_key,l_value)
						if l_output /= Void then
							Global.io_handler.set_log (l_output)
						end
					end
				elseif equal (l_key,"monitor") then
					if l_value = Void or else l_value.is_empty then
						Global.io_handler.set_monitor (Void)
					else
						l_output := get_output_channel (a_node.key,l_key,l_value)
						if l_output /= Void then
							Global.io_handler.set_monitor (l_output)
						end
					end
				else
					error.warn ("'$1': unknown symbol '$2' ignored",<<a_node.path,l_key>>)
				end
			end
		end

	get_output_channel (a_system, a_key, a_value: STRING): ROSE_OUTCHANNEL is
			-- Output channel of path 'a_value' for 'a_key'
		do
			Result := Global.io_handler.manager.output_channel (a_value)
			if Result = Void then
				error.warn ("'$1': redirection of channel '$2' to invalid path '$3' ignored",<<a_system,a_key,a_value>>)
				Result := Void
			end
		end

	define_aliases (a_node: ROSE_CONFIGURATION_NODE) is
			-- Define 'a_node' (key,value) pairs as aliases.
		do
			if a_node /= Void and then a_node.children /= Void then
				trace ("Defining aliases configured in '$1'%N", <<a_node.path>>)
				a_node.do_all (agent define_alias)
			end
		end

	define_alias (a_node: ROSE_CONFIGURATION_NODE) is
			-- Define alias from 'a_node' (alias node's key to node's value)
		local
			l_key, l_value: STRING
		do
			l_key := a_node.key
			l_value := a_node.value
			if l_key /= Void and then not l_key.is_empty and then l_value /= Void then
				if Global.commands.is_command (l_key) then
					error.warn ("'$1': aliasing of command '$2' ignored",<<a_node.path,l_key>>)
				else
					Global.commands.set_alias (l_key,l_value)
				end
			end
		end

	execute_commands (a_node: ROSE_CONFIGURATION_NODE) is
			-- Execute 'a_node' child values as commands.
		do
			if a_node /= Void then
				trace ("Executing auto-commands configured in '$1'%N", <<a_node.path>>)
				a_node.do_all (agent execute_command)
			end
		end

	execute_command (a_node: ROSE_CONFIGURATION_NODE) is
			-- Execute 'a_node' child values as commands.
		do
			Global.commands.execute (a_node.value)
		end

end -- class RCMD_SYSTEM
