note

	description:

		"Multiple (weighted) regression class"

	author: "Mark Howard"
	last_editor: "$Author: er $"
	date: "$Date: 2014/02/26 $"
	revision: "$Revision: #79 $"
	path: "$File: //rose/source/kernel/infrastructure/math/regression.e $"

class	REGRESSION

inherit
	ROSE_CONSTANTS

	REGRESSION_BASE

	DOUBLE_MATH

	EXCEPTIONS

	ROSE_SNAPSHOT_AWARE
		export
			{NONE} all
		end

	ROSE_SHARED_GLOBAL

	MISMATCH_CORRECTOR
		redefine
			correct_mismatch
		end

create

	make, test, make_weighted, make_with_intercept, make_weighted_with_intercept, make_weighted_without_dependent, make_compact,
	default_create, make_with_simple_constraint, make_weighted_with_simple_constraint,
	make_with_general_prior, make_weighted_with_simple_constraint_and_general_prior,
	test_compact, test_weighted, test_small_weighted

-- Design note: Several statistics related to Regression, including 'residuals', 'unweighted_residuals', 't_statistics', and 'fitted_values',
--              have been implemented so that they are calculated each time that they are called.  Since they are vectors, anyone using them
--				should take care when calling individual entries, particularly within a loop, since that will require the entire vector
--				to be recalculated, and thus may hurt runtime.  Note that as written, neither the order of nor the leading constant of runtime
--				is increased by this design choice, so it has been left this way to preserve flexibility for REGRESSION and its decendents.

feature {NONE} -- Initialization

	make (a_matrix: ABSTRACT_READABLE_MATRIX; a_dependent: EXTENDABLE_VECTOR) is
			-- Initialize `Current' then run regression of `a_dependent' against `a_matrix' (as explanatory variables).
		do
			prepare (a_matrix, a_dependent)
			regress
		end

	prepare (a_matrix: ABSTRACT_READABLE_MATRIX; a_dependent: EXTENDABLE_VECTOR) is
			-- Set explanatory variables from `a_matrix' and dependent values from `a_dependent'.
		require
			a_matrix_not_void: a_matrix /= Void
			a_dependent_not_void: a_dependent /= Void
			explanatory_matches_dependent : a_matrix.height = a_dependent.count
			enough_observations: a_matrix.height >= a_matrix.width
			all_weights_positive: weights /= Void implies weights.non_negative
		local
			i: INTEGER
		do
			number_of_dependent_variables := 1
			set_dimensions (a_matrix)
			if a_matrix.column_names /= Void then
				set_independent_variable_names (a_matrix.column_names)
			end

			if weights = Void then
				dependent_variables := a_dependent
				matrix := a_matrix
			else
				create weight_roots.make (weights.count)
				from i := 1 until i > weights.count loop
					weight_roots.put (sqrt (weights [i]), i)
					i := i + 1
				end
				create weighted_dependent_variables.make (a_dependent, weight_roots)
				create weighted_matrix.make (a_matrix, weight_roots)
				matrix := weighted_matrix
				dependent_variables := weighted_dependent_variables
			end

			if weights /= Void then
				unweighted_matrix := weighted_matrix.matrix
				unweighted_dependent_variables := weighted_dependent_variables.vector
			else
				unweighted_matrix := matrix
				unweighted_dependent_variables := dependent_variables
			end

			xtx := matrix.xtx
			xtx_inverse := Void
			set_y_standard
		ensure
			prepared: prepared
			xtx_inverse_void: xtx_inverse = Void
		end

	make_with_intercept (a_matrix: MATRIX; a_dependent: EXTENDABLE_VECTOR) is
			-- Initialize `Current' then run regression of `a_dependent' against `a_matrix' (as explanatory variables) and produce an intercept.
		require
			a_matrix_not_void: a_matrix /= Void
			a_dependent_not_void: a_dependent /= Void
			explanatory_matches_dependent: a_matrix.height = a_dependent.count
			enough_observations: a_matrix.height >= a_matrix.width
		do
			intercept_used := True
			make(intercept_matrix(a_matrix), a_dependent)
		ensure
			prepared: prepared
			coefficients_not_void: coefficients /= Void
			valid_r_square: ((r_square_with_no_intercept >= 0) and (r_square_with_no_intercept <= 1) and
				(r_square_with_intercept >= 0) and (r_square_with_intercept <= 1))
		end

	make_weighted_with_intercept (a_matrix: MATRIX; a_dependent, a_weights: EXTENDABLE_VECTOR) is
			-- Initialize `Current' then run regression weighted by `a_weights' of `a_dependent' against `a_matrix' (as explanatory variables) and produce an intercept.
		do
			intercept_used := True
			make_weighted (intercept_matrix(a_matrix), a_dependent, a_weights)
		end

	make_weighted (a_matrix: ABSTRACT_READABLE_MATRIX; a_dependent, a_weights: EXTENDABLE_VECTOR) is
			-- Initialize `Current' then run regression weighted by `a_weights' of `a_dependent' against `a_matrix' (as explanatory variables).
		do
			weights := a_weights
			make (a_matrix, a_dependent)
		end


	make_weighted_without_dependent (a_matrix: ABSTRACT_READABLE_MATRIX; a_weights: EXTENDABLE_VECTOR) is
			-- Initialize `Current' using `a_weights' and `a_matrix' (as explanatory variables).
		require
			a_matrix_not_void: a_matrix /= Void
			enough_observations: a_matrix.height >= a_matrix.width
			a_weights_not_void: a_weights /= Void
			all_weights_positive: a_weights.non_negative
			valid_dimension: a_matrix.height = a_weights.count
		local
			i: INTEGER
		do
			set_dimensions (a_matrix)

			weights := a_weights
			create weight_roots.make (weights.count)
			from i := 1 until i > weights.count loop
				weight_roots.put (sqrt (weights [i]), i)
				i := i + 1
			end
			create weighted_matrix.make (a_matrix, weight_roots)
			matrix := weighted_matrix

			xtx := matrix.xtx
			xtx_inverse := Void
		ensure
			weights_set: weights = a_weights
			weight_roots_not_void: weight_roots /= Void
			weighted_matrix_not_void: weighted_matrix /= Void
			matrix_is_weighted_matrix: matrix = weighted_matrix
			xtx_inverse_void: xtx_inverse = Void
			-- the next line depends upon an undocumented postcondition for {ABSTRACT_READABLE_MATRIX}.xtx
			xtx_not_void: xtx /= Void
		end

	make_compact (a_nvar, a_ndep: INTEGER) is
			-- Initialize `Current' in compact form.
		do
			number_of_independent_variables := a_nvar
			number_of_dependent_variables := a_ndep

			create xtx.make (number_of_independent_variables)
			create xpsum.make (number_of_independent_variables)
			create xw2sum.make (number_of_independent_variables)
			create ypsum.make (number_of_dependent_variables)
			create yw2sum.make (number_of_dependent_variables)
			create ypy.make (number_of_dependent_variables)
			create xpy.make (number_of_independent_variables, number_of_dependent_variables)
		end

	make_with_simple_constraint (a_matrix: ABSTRACT_READABLE_MATRIX; a_dependent: EXTENDABLE_VECTOR; a_lhs: ABSTRACT_READABLE_VECTOR; a_rhs: DOUBLE) is
			-- Try to initialize `Current' with constrained inputs.
		require
			a_matrix_not_void: a_matrix /= Void
			a_dependent_not_void: a_dependent /= Void
			explanatory_matches_dependent : a_matrix.height = a_dependent.count
			enough_observations : a_matrix.height >= a_matrix.width
			valid_constraint: a_dependent.count = a_matrix.height
			a_lhs_not_void: a_lhs /= Void
		local
			l_pair: PAIR [MATRIX, VECTOR]
		do
			set_constraint (a_lhs, a_rhs)
			l_pair := data_pair_with_simple_constraint (a_matrix, a_dependent, a_lhs, a_rhs)
			if l_pair /= Void then
				make (l_pair.first, l_pair.second)
			end
			set_original_coefficients
			set_all_names
		end

	make_weighted_with_simple_constraint (a_matrix: ABSTRACT_READABLE_MATRIX; a_dependent, a_weights: EXTENDABLE_VECTOR; a_lhs: ABSTRACT_READABLE_VECTOR; a_rhs: DOUBLE) is
			-- Try to initialize `Current' with constrained weighted inputs.
		require
			a_matrix_not_void: a_matrix /= Void
			a_dependent_not_void: a_dependent /= Void
			explanatory_matches_dependent : a_matrix.height = a_dependent.count
			enough_observations : a_matrix.height >= a_matrix.width
			a_weights_not_void: a_weights /= Void
			valid_dimension: a_matrix.height = a_weights.count
			valid_constraint : a_dependent.count = a_matrix.height
			a_lhs_not_void: a_lhs /= Void
		local
			l_pair: PAIR[MATRIX, VECTOR]
		do
			set_constraint (a_lhs, a_rhs)
			l_pair := data_pair_with_simple_constraint (a_matrix, a_dependent, a_lhs, a_rhs)
			if l_pair /= Void then
				make_weighted (l_pair.first, l_pair.second, a_weights)
			end
		end

	make_with_general_prior (a_matrix: ABSTRACT_READABLE_MATRIX; a_dependent, a_weights: EXTENDABLE_VECTOR;
		a_prior_means: ABSTRACT_READABLE_VECTOR; a_prior_cov_inverse: SYMMETRIC_MATRIX) is
			-- Initialize `Current' then run regression weighted by `a_weights' of `a_dependent' against `a_matrix' (as explanatory variables).
			-- Re-estimate coefficients from `a_prior_mean' and `a_prior_cov_inverse'.
		require
			a_matrix_not_void: a_matrix /= Void
			a_dependent_not_void: a_dependent /= Void
			explanatory_matches_dependent: a_matrix.height = a_dependent.count
			enough_observations: a_matrix.height >= a_matrix.width
			a_weights_not_void: a_weights /= Void
			all_weights_positive: a_weights.non_negative
			valid_dimension: a_matrix.height = a_weights.count
			a_prior_means_not_void: a_prior_means /= Void
			a_prior_cov_inverse_not_void: a_prior_cov_inverse /= Void
		do
			make_weighted (a_matrix, a_dependent, a_weights)
			original_coefficients := coefficients
			prior_mean := a_prior_means
			prior_cov_inverse := a_prior_cov_inverse
			compute_bayesian_coefficients (a_prior_means, a_prior_cov_inverse)
			set_statistics (coefficients)
		ensure
			weights_set: weights = a_weights
			prepared: prepared
			original_coefficients_not_void: original_coefficients /= Void
			prior_mean_not_void: prior_mean /= Void
			coefficients_not_void: coefficients /= Void
			valid_r_square: ((r_square_with_no_intercept >= 0) and (r_square_with_no_intercept <= 1) and
				(r_square_with_intercept >= 0) and (r_square_with_intercept <= 1))
			covariance_not_void: covariance /= Void
		end

	make_weighted_with_simple_constraint_and_general_prior (a_matrix : ABSTRACT_READABLE_MATRIX; a_dependent: EXTENDABLE_VECTOR; a_weights: EXTENDABLE_VECTOR;
		a_lhs: ABSTRACT_READABLE_VECTOR; a_rhs: DOUBLE; a_prior_means: ABSTRACT_READABLE_VECTOR; a_prior_cov_inverse: SYMMETRIC_MATRIX) is
			-- Try to initialize `Current' with constrained weighted inputs.
			-- Re-estimate coefficients from `a_prior_mean' and `a_prior_cov_inverse'.
		require
			a_matrix_not_void: a_matrix /= Void
			a_dependent_not_void: a_dependent /= Void
			explanatory_matches_dependent : a_matrix.height = a_dependent.count
			enough_observations : a_matrix.height >= a_matrix.width
			a_weights_not_void: a_weights /= Void
			valid_dimension: a_matrix.height = a_weights.count
			valid_constraint : a_dependent.count = a_matrix.height
			a_lhs_not_void: a_lhs /= Void
			a_prior_means_not_void: a_prior_means /= Void
			a_prior_cov_inverse_not_void: a_prior_cov_inverse /= Void
		local
			l_ix: INTEGER
			l_ixs: ARRAY[INTEGER]
			l_ixs_1, l_ixs_2: INTEGER_VECTOR
			l_prior_means: VECTOR
			l_prior_cov_inverse: SYMMETRIC_MATRIX
		do
			make_weighted_with_simple_constraint (a_matrix, a_dependent, a_weights, a_lhs, a_rhs)
			original_constrained_coefficients := coefficients

			prior_mean := a_prior_means
			prior_cov_inverse := a_prior_cov_inverse

			l_ix := a_lhs.first_non_zero_item_index
			create l_prior_means.make (a_prior_means.count - 1)
			l_prior_means.put_vector (a_prior_means.sub_vector (l_ix + 1, a_prior_means.count), l_ix)

			if l_ix > 1 and l_ix < a_prior_means.count then
				create l_ixs_1.make_sequence (1, l_ix - 1, 1)
				create l_ixs_2.make_sequence (l_ix + 1, a_prior_means.count, 1)
				l_ixs := l_ixs_1 + l_ixs_2
			elseif l_ix = 1 then
				create l_ixs_2.make_sequence (l_ix + 1, a_prior_means.count, 1)
				l_ixs := l_ixs_2
			elseif l_ix = a_prior_means.count then
				create l_ixs_1.make_sequence (1, l_ix - 1, 1)
				l_ixs := l_ixs_1
			end
			l_prior_cov_inverse := a_prior_cov_inverse.sub_symmetric_matrix (l_ixs)

			compute_bayesian_coefficients (l_prior_means, l_prior_cov_inverse)
			set_statistics(coefficients)

			set_original_coefficients
		end

	data_pair_with_simple_constraint (a_matrix: ABSTRACT_READABLE_MATRIX; a_dependent: EXTENDABLE_VECTOR; a_lhs: ABSTRACT_READABLE_VECTOR; a_rhs: DOUBLE): PAIR [MATRIX, VECTOR] is
			-- Constrained versions of `a_matrix' and `a_depenmdent'
		require
			a_matrix_not_void: a_matrix /= Void
			a_dependent_not_void: a_dependent /= Void
			explanatory_matches_dependent : a_matrix.height = a_dependent.count
			enough_observations : a_matrix.height >= a_matrix.width
			valid_constraint : a_dependent.count = a_matrix.height
			a_lhs_not_void: a_lhs /= Void
		local
			i, l_ix: INTEGER
			l_vec, l_dependent: VECTOR
			l_matrix: MATRIX
		do
			l_ix := a_lhs.first_non_zero_item_index
			if l_ix > 0 then
				create l_matrix.make (a_matrix.height, a_matrix.width - 1)
				from i := 1 until i > a_matrix.width loop
					l_vec := a_matrix.column (i) - (a_matrix.column(l_ix)).scaled (a_lhs [i] / a_lhs [l_ix])
					if i < l_ix then
						l_matrix.put_column_vector (l_vec , i)
					elseif i > l_ix then
						l_matrix.put_column_vector (l_vec ,i - 1)
					end
					i := i + 1
				end
				l_dependent := a_dependent - a_matrix.column (l_ix).scaled (a_rhs)
				create Result.make (l_matrix, l_dependent)
			end
		ensure
			constrained_matrix_not_void: Result /= Void implies Result.first /= Void
			constrained_dependent_not_void: Result /= Void implies Result.second /= Void
		end

	compute_bayesian_coefficients (a_prior_means: ABSTRACT_READABLE_VECTOR; a_prior_cov_inverse: SYMMETRIC_MATRIX) is
			-- Estimate coefficients from `a_prior_means' and `a_prior_cov_inverse'.
		require
			a_prior_means_not_void: a_prior_means /= Void
			a_prior_cov_inverse_not_void: a_prior_cov_inverse /= Void
			xtx_not_void: xtx /= Void
			xty_not_void: xty /= Void
		local
			l_s: DOUBLE
			l_v1, l_v2, l_y: VECTOR
			l_c, l_c_inverse: MATRIX
		do
			l_s := sum_of_squared_errors / degree_of_freedom
			l_s := 1 / l_s

--			l_c := (a_prior_cov_inverse + xtx.scalar_multiply (l_s))
			l_c := (xtx.scalar_multiply (l_s) + a_prior_cov_inverse)
			l_c.replace_diagonal
			l_c_inverse := l_c.inverse

			l_v1 := a_prior_cov_inverse @ a_prior_means
			l_v2 := xty.scaled (l_s)
			l_y := l_v1 + l_v2

			coefficients := l_c_inverse @ l_y
		ensure
			coefficients_not_void: coefficients /= Void
		end

feature -- Access

	coefficients_with_constraint (a_lhs: ABSTRACT_READABLE_VECTOR; a_rhs: DOUBLE): VECTOR is
			-- coefficients in a constraint setting
		require
			a_lhs_not_void: a_lhs /= Void
			-- TODO: Is this routine sound? Additional preconditions appear to be coeeficients non-Void, and of a certain (?) size.
		local
			i, l_ix : INTEGER
			l_coeff : DOUBLE
		do
			l_ix := a_lhs.first_non_zero_item_index

			create Result.make (a_lhs.count)
			l_coeff := a_rhs
			from i := 1 until i > a_lhs.count loop
					if i < l_ix then
						Result [i] := coefficients [i]
					elseif i > l_ix then
						Result [i] := coefficients [i-1]
					end
					if i /= l_ix then
						l_coeff := l_coeff - Result [i] * a_lhs [i] / a_lhs [l_ix]
					end
				i := i + 1
			end
			Result [l_ix] := l_coeff
		ensure
			coefficients_with_constraint_not_void: Result /= Void
			correct_count: Result.count = a_lhs.count
		end

	prior_mean: ABSTRACT_READABLE_VECTOR
			-- Prior mean for bayesian estimates

	prior_cov_inverse: SYMMETRIC_MATRIX
			-- Prior covariance for bayesian estimates

	original_coefficients: ABSTRACT_READABLE_VECTOR
			-- Original coefficients prior to bayesian estimation or constraint

	original_constrained_coefficients: ABSTRACT_READABLE_VECTOR
			-- Original constrained coefficients prior to bayesian estimation

	constraint_lhs: ABSTRACT_READABLE_VECTOR
		-- Left hand side of constraint

	constraint_rhs: DOUBLE
		-- Right hand side of constraint

feature -- Status report

	intercept_used: BOOLEAN
			-- Was an intercept fitted?

	prepared: BOOLEAN is
			-- Is regression prepared?
		do
			Result := (matrix /= Void) and then ((number_of_independent_variables = matrix.width) and (number_of_observations = matrix.height)
				and (dependent_variables /= Void) and (unweighted_matrix /= Void) and
				(unweighted_dependent_variables /= Void) and (weights /= Void implies weighted_matrix /= Void) and
				(weights /= Void implies weighted_dependent_variables /= Void) and
				(weights /= Void implies weight_roots /= Void) and (xty /= Void) and (xtx /= Void))
		ensure
			matrix_not_void: Result implies (matrix /= Void)
			independent_variable_count_set: Result implies (number_of_independent_variables = matrix.width)
			observation_count_set: Result implies (number_of_observations = matrix.height)
			dependent_variables_not_void: Result implies (dependent_variables /= Void)
			unweighted_matrix_not_void: Result implies (unweighted_matrix /= Void)
			unweighted_dependent_variables_not_void: Result implies (unweighted_dependent_variables /= Void)
			weighted_matrix_not_void: (Result and weights /= Void) implies (weighted_matrix /= Void)
			weighted_dependent_variables_not_void: (Result and weights /= Void) implies (weighted_dependent_variables /= Void)
			weight_roots_not_void: (Result and weights /= Void) implies (weight_roots /= Void)
			xty_not_void: Result implies (xty /= Void)
			-- the next line depends upon an undocumented postcondition for {ABSTRACT_READABLE_MATRIX}.xtx
			xtx_not_void: Result implies (xtx /= Void)
		end

feature -- Statistics

	r_square_with_intercept: DOUBLE
			-- Coefficient of determination;
			-- SAS computes R-square differently when one assumes an intercept.  We follow the conventions of SAS.

	r_square_with_no_intercept: DOUBLE
			-- Coefficient of determination when no intercept is modeled;
			-- NAG apparently uses this even when there is an intercept.

	r_square: DOUBLE is
			-- Coefficient of determination
		do
			if not intercept_used then
				Result := r_square_with_no_intercept
			else
				Result := r_square_with_intercept
			end
		end

	adjusted_r_square: DOUBLE is
			-- Coefficient of determination adjusted for number of explanatory terms in model
		do
			if not intercept_used then
				Result := adjusted_r_square_with_no_intercept
			else
				Result := adjusted_r_square_with_intercept
			end
		end

feature -- Results

	coefficients: VECTOR
			-- Estimates of coefficients of explanatory variables

	covariance: SYMMETRIC_MATRIX
			-- Covariance of explanatory variables

	standard_errors: VECTOR
			-- Standard errors of `coefficients'

	sigma_square: DOUBLE
			-- Variance

	modified_sigma_square: DOUBLE
			-- Variance ignoring zero weight observations

	sum_of_squared_errors: DOUBLE
			-- Numerator of variance

	y_total_variation: DOUBLE
			-- sum square of deviation from y mean

	residual_standard_errors: VECTOR
			-- Standard errors of unweighted residuals for `studentized_residuals'

	studentized_residuals: VECTOR
			-- Internally studentized (unweighted) residuals

	external_residual_standard_errors: VECTOR
			-- Standard errors of unweighted residuals for `external_studentized_residuals'

	external_studentized_residuals: VECTOR
			-- Externally studentized (unweighted) residuals

	number_of_observations: INTEGER
			-- Number of observations used

	number_of_independent_variables: INTEGER
			-- Number of explanatory variables

	number_of_dependent_variables: INTEGER
			-- Number of response variables

	degree_of_freedom: INTEGER
			-- Number of observations minus number of parameters estimated

	modified_degree_of_freedom: INTEGER
			-- `degree_of_freedom' minus zero weight observation count

feature -- Diagnostics

	null_column_indice: ROSE_LINEAR_ARRAY[INTEGER]

feature -- Compact summary

	xtx: SYMMETRIC_MATRIX
			--

	xty: VECTOR
			--

	yty: DOUBLE
			--

	full_xtx: SYMMETRIC_MATRIX
			-- this put dependent at the bottom of the x_xtx matric, it is symmetric but, not implemented here

	xtx_inverse: SYMMETRIC_MATRIX
			-- Information matrix

	x_sum: VECTOR
			-- redundant but to avoid confusion with the add-an-obervation method

	y_sum: DOUBLE
			-- redundant but to avoid confusion with the add-an-obervation method

feature -- Standard representation	

	dependent_variables: EXTENDABLE_VECTOR
			-- Values of response variable

	matrix: ABSTRACT_READABLE_MATRIX
			-- Values of explanatory variables

	weights: EXTENDABLE_VECTOR
			-- Weights corresponding to `dependent_variables'

	weight_roots: VECTOR
			-- Square roots of `weights'

	weighted_matrix: WEIGHTED_MATRIX
			-- Weighted values of explanatory variables

	weighted_dependent_variables: WEIGHTED_VECTOR
			-- Weighted values of response variable

	unweighted_matrix: ABSTRACT_READABLE_MATRIX
			-- Unweighted values of explanatory variables

	unweighted_dependent_variables: ABSTRACT_READABLE_VECTOR
			-- Unweighted values of response variable

feature -- Compact representation specific	

	xpsum: VECTOR
			--

	xw2sum : VECTOR
			--

	ypsum: VECTOR
			--

	yw2sum: VECTOR
			--


	sum_w2: DOUBLE
			--

	ypy: VECTOR
			--

	xpy: MATRIX
			--

	solutions: ROSE_LINEAR_ARRAY [REGRESSION_SOLUTION]
			--

feature -- Priors

	prior_yy: DOUBLE
			--

	set_prior_yy (a_yy: like prior_yy) is
			-- Set `prior_yy' to `a_yy'.
		do
			prior_yy := a_yy
		end

	prior_xy: VECTOR
			--

	set_prior_xy (a_xy: like prior_xy) is
			-- Set `prior_xy' to `a_xy'.
		require
			a_xy_not_void: a_xy /= Void
		do
			prior_xy := a_xy
		ensure
			prior_xy_set: prior_xy = a_xy
		end

	prior_xtx: SYMMETRIC_MATRIX
			--

	set_prior_xtx (a_xtx: like prior_xtx) is
			-- Set `prior_xtx' to `a_xtx'.
		require
			a_xtx_not_void: a_xtx /= Void
		do
			prior_xtx := a_xtx
		ensure
			prior_xtx_set: prior_xtx = a_xtx
		end

feature -- Basic operations

	regress is
			-- Perform regression using standard representation using QR method;
			-- THIS BEARS INVESTIGATION.  FOUND THAT SVD METHOD PRODUCED AN ASSYMETRICAL RESULT FOR PSEUDO_INVERSE
			--  THAT PRODUCED COEFFICIENTS THAT DIFFERED MATERIALLY FROM OTHER METHODS.
			-- THE FOLLOWING USES QR METHOD AND THEN TAKES THE TRIANGLE TO CREATE THE SYMMETRIC RESULT
		require
			xtx_not_void: xtx /= Void
			xty_not_void: prior_xy /= Void implies xty /= Void
		do
			if prior_xtx /= Void then
				xtx.add (prior_xtx)
			end
			if prior_xy /= Void then
				xty.add (prior_xy)
			end
			if prior_yy /= 0 then
				yty:= yty + prior_yy
			end
			-- handle null columns
			null_column_indice := xtx.diagonal.value_indice (0)
			xtx.replace_diagonal

			-- TODO for review: what if xtx_inverse /= Void and prior_xtx /= Void - BUG?
			--  likewise what if `regress' is called twice in a row? require xtx_inverse = Void? (C4 20070926)
			if xtx_inverse = Void then
				xtx_inverse := xtx.inverse
				if check_inversion then
					if not (xtx * xtx_inverse).is_identity then
						is_inversion_error := True
						diagnose_inversion_failure
					end
				end
			end
			coefficients := xtx_inverse @ xty.as_vector
			set_statistics (coefficients)
			set_all_names
		ensure
			covariance_not_void: covariance /= Void
			coefficients_not_void: coefficients /= Void
			valid_r_square: ((r_square_with_no_intercept >= 0) and (r_square_with_no_intercept <= 1) and
				(r_square_with_intercept >= 0) and (r_square_with_intercept <= 1))
		end
feature -- Errors and Diagnosis

	check_inversion : BOOLEAN -- whether inversion should be checked for success
	set_check_inversion(a_val: like check_inversion) is -- if set, checks and diagnoses inversion failures
		do check_inversion := a_val end

	is_inversion_error: BOOLEAN	-- if set checks and diagnoses xtx inversion failures

	diagnose_inversion_failure is
			-- isolate and report the variable which causes inversion failure
		local
			i: INTEGER
			l_diagnosed:BOOLEAN
			l_sub_matrix: MATRIX
--			l_inverse: MATRIX
			l_name:STRING
		do
			print(" Diagnosing inversion failure: nvar="+number_of_independent_variables.out+"%N")
			--first: simple check for a negative, or zero diagonal
			from i:=1 until i > xtx.height loop
				if (xtx.item (i, i)<=0) then
					io.put_string(" non-positive value for diagonal entry for var #"+i.out+"="+xtx.item(i,i).out+"%N")
				end
				i:=i+1
			end
			--next: increase dimension of principal minor one by one, find the first matrix which fails to invert:
			from i:=1 until i>number_of_independent_variables or l_diagnosed loop
				l_sub_matrix := xtx.sub_matrix_2(1,i, 1,i)
--				l_inverse := l_sub_matrix.inverse
				if independent_variable_names/= Void then
					l_name := independent_variable_names.item(i)
				else
					l_name := "Variable "+i.out
				end
				if not (l_sub_matrix.is_invertible) then -- * l_inverse).is_identity then
					l_diagnosed := True
					io.put_string(" ERROR IN MATRIX INVERSION OCCURS WHEN ADDING  variable " + l_name + "%N")
				else
					--report success on adding the variable
					io.put_string (" PRECISION MATRIX INVERTS PROPERLY, PRINCIPAL MINOR THROUGH VARIABLE " + l_name + "%N")
				end
				i:= i+1
			end
		end

	set_statistics (a_coefficients: like coefficients) is
			-- Calculate standard regression statistics.
			-- Treatment of standard errors needs improvement - negative diagonals should not be possible.
		require
			a_coefficients_not_void: a_coefficients /= Void
		local
			l_zero_weight_count, i : INTEGER
		do
			-- India Team: We have commented the following line.
			-- degree_of_freedom := dependent_variables.count - a_coefficients.count

			degree_of_freedom:= number_of_observations - number_of_independent_variables

			sum_of_squared_errors := yty - xty * a_coefficients
			sigma_square := sum_of_squared_errors / degree_of_freedom

			-- degrees of freedom adjusted to remove zero weight observations
			-- SAS does this so we will too
			l_zero_weight_count := 0
			if weights /= void then
				from i := 1 until i > weights.count loop
					if weights[i] = 0.0 then
						l_zero_weight_count := l_zero_weight_count + 1
					end
					i := i + 1
				end
			end
			modified_degree_of_freedom := degree_of_freedom - l_zero_weight_count
			modified_sigma_square := sum_of_squared_errors / modified_degree_of_freedom
			if sum_of_squared_errors = 0 and then yty = 0 then
				r_square_with_no_intercept := 0
			else
				r_square_with_no_intercept := 1 - (sum_of_squared_errors / yty)
			end
			if intercept_used then
				if sum_of_squared_errors = 0 and then y_total_variation = 0 then
					r_square_with_intercept := 0
				else
					r_square_with_intercept := 1 - (sum_of_squared_errors / y_total_variation)
				end
			end

			covariance := xtx_inverse.scalar_multiply (modified_sigma_square)
			if covariance.diagonal.non_negative then
				standard_errors := covariance.diagonal.square_roots
			else
				create standard_errors.make_constant (-1, coefficients.count)
			end
		ensure
			covariance_not_void: covariance /= Void
			valid_r_square: (r_square_with_no_intercept >= 0) and (r_square_with_no_intercept <= 1) and
				(r_square_with_intercept >= 0) and (r_square_with_intercept <= 1)
		end


	calculate_residual_statistics is
			-- Calculate statistics on residuals.

			-- The formulas used for this can be found in the SAS documentation or on Wikipedia.
			-- Note different sources have different formulas for external studentized residual,
			-- so there are different formulas given below.  In practice there is little differnce
			-- between the two values, and also between internal and external values.

		local
			l_leverage: VECTOR	-- sometimes called h_i or H[i,i] in literature	
			i: INTEGER
			l_residuals : VECTOR
		do
			create l_leverage.make (matrix.height)
			create residual_standard_errors.make (matrix.height)
			create studentized_residuals.make (matrix.height)
			create external_residual_standard_errors.make (matrix.height)
			create external_studentized_residuals.make (matrix.height)

			l_residuals := unweighted_residuals

			if weight_roots = void then
				-- regression is unweighted
				from i := 1 until i > matrix.height loop
					l_leverage [i] := matrix.row (i).dot_product (xtx_inverse @ matrix.row (i))

					residual_standard_errors [i] := sqrt(modified_sigma_square *(1 - l_leverage [i]))
					studentized_residuals [i] := l_residuals [i]/residual_standard_errors [i]

					-- We give two formulas for 'external_residual_standard_errors' since two different ones exist in the literature.
					external_residual_standard_errors [i] := residual_standard_errors [i]*sqrt(modified_sigma_square/(modified_sigma_square - 1))
--					external_residual_standard_errors [i] := residual_standard_errors[i]*sqrt((modified_degree_of_freedom - l_residuals[i]*l_residuals[i])/(modified_degree_of_freedom - 1))
					external_studentized_residuals [i] := l_residuals [i]/external_residual_standard_errors [i]

					i := i + 1
				end
			else
				-- regression is weighted
				-- If an observation has weight 0 then its residual standard error is 1.0/0.0
				-- and its studentized residual is 0.0.  This is in contrast to SAS, which
				-- does not return a number for the case of weight 0 observations
				from i := 1 until i > matrix.height loop
					l_leverage [i] := matrix.row (i).dot_product (xtx_inverse @ matrix.row (i))

					residual_standard_errors [i] := sqrt(modified_sigma_square *(1 - l_leverage [i]))/weight_roots[i]
					studentized_residuals [i] := l_residuals [i]/residual_standard_errors [i]

					-- We give two formulas for 'external_residual_standard_errors' since two different ones exist in the literature.
					external_residual_standard_errors [i] := residual_standard_errors [i]*sqrt(modified_sigma_square/(modified_sigma_square - 1))
--					external_residual_standard_errors [i] := residual_standard_errors[i]*sqrt((degree_of_freedom - l_residuals[i]*l_residuals[i])/(degree_of_freedom - 1))
					external_studentized_residuals [i] := l_residuals [i]/external_residual_standard_errors [i]
					i := i + 1

				end
			end
		end

	regress_compact is
			-- Perform regression using the compact representation.
		require
			yw2sum_not_void: number_of_observations > 0 implies yw2sum /= Void
			ypy_not_void: number_of_observations > 0 implies ypy /= Void
			xpy_not_void: number_of_observations > 0 implies xpy /= Void
		local
			i: INTEGER
			l_solution: REGRESSION_SOLUTION
		do
			if number_of_observations > 0 then
				create solutions.make (1, number_of_dependent_variables)
				from i := 1 until i > number_of_dependent_variables loop
					xty := xpy.column (i)
					yty := ypy.item (i)
					y_total_variation := yty - yw2sum.item (i) * yw2sum.item (i) / sum_w2
					regress
					create l_solution.make (Current, i)
					solutions.put (l_solution, i)
					i := i + 1
				end
			end
		ensure
			coefficients_not_void: number_of_observations > 0 implies coefficients /= Void
			valid_r_square: number_of_observations > 0 implies ((r_square_with_no_intercept >= 0) and
				(r_square_with_no_intercept <= 1) and (r_square_with_intercept >= 0) and (r_square_with_intercept <= 1))
			solutions_not_void: number_of_observations > 0 implies solutions /= Void
		end

	regress_with_dependent (a_dependent: EXTENDABLE_VECTOR) is
			-- Perform regression using standard representation using `a_dependent' as response variable.
		require
			a_dependent_not_void: a_dependent /= Void
		do
			if weights = void then
				dependent_variables := a_dependent
			else
				create weighted_dependent_variables.make (a_dependent, weight_roots)
				dependent_variables := weighted_dependent_variables
			end
			set_y_standard
			regress
		ensure
			dependent_variables_not_void: dependent_variables /= Void
			coefficients_not_void: coefficients /= Void
			valid_r_square: ((r_square_with_no_intercept >= 0) and (r_square_with_no_intercept <= 1) and
				(r_square_with_intercept >= 0) and (r_square_with_intercept <= 1))
		end

feature {NONE} -- Setting

	set_dimensions(a_matrix: ABSTRACT_READABLE_MATRIX) is
			-- Set `number_of_observations' and `number_of_independent_variables' from dimensions of `a_matrix'.
		require
			a_matrix_not_void: a_matrix /= Void
		do
			number_of_independent_variables := a_matrix.width
			number_of_observations := a_matrix.height
		ensure
			independent_variable_count_set: number_of_independent_variables = a_matrix.width
			observation_count_set: number_of_observations = a_matrix.height
		end

	set_y_standard is
			-- Set standard y related quntities
		require
			matrix_not_void: matrix /= Void
			dependent_variables_not_void: dependent_variables /= Void
			weight_roots_not_void: weights /= Void implies weight_roots /= Void
		do
			xty := matrix.transpose @ dependent_variables.as_vector
			yty := dependent_variables.squared_norm
			y_total_variation := y_total_variation_standard
		ensure
			xty_not_void: xty /= Void
		end

feature -- Setting

	set_constraint (a_lhs: like constraint_lhs; a_rhs: like constraint_rhs) is
			-- Set constraint on regression solutions.
		do
			constraint_lhs := a_lhs
			constraint_rhs := a_rhs
		ensure
			constraint_lhs_set: constraint_lhs = a_lhs
			constraint_rhs_set: constraint_rhs = a_rhs
		end

	set_original_coefficients is
			-- Set original coefficients prior to constraint.
		require
			constraint_lhs_not_void: constraint_lhs /= Void
		do
			original_coefficients := coefficients_with_constraint (constraint_lhs, constraint_rhs)
		ensure
			original_coefficients_not_void: original_coefficients /= Void
			correct_count: original_coefficients.count = constraint_lhs.count
		end

feature -- Element change (compact regression)

	add_observation_compact (a_xdata, a_ydata: VECTOR; a_weight: DOUBLE) is
			-- Add an observation to compact representation.
		require
			valid_weight: a_weight >= 0.
			valid_x: a_xdata /= Void and then a_xdata.count = number_of_independent_variables
			valid_y: a_ydata /= Void and then a_ydata.count = number_of_dependent_variables
		local
			l_weight : DOUBLE
		do
			l_weight := sqrt (a_weight)
			add_observation_compact_imp (a_xdata, a_ydata, l_weight, True)
		end

	delete_observation_compact (a_xdata, a_ydata: VECTOR; a_weight: DOUBLE) is
			-- Remove an observation from compact representation.
		require
			valid_weight: a_weight >= 0.
			valid_x: a_xdata /= Void and then a_xdata.count = number_of_independent_variables
			valid_y: a_ydata /= Void and then a_ydata.count = number_of_dependent_variables
		local
			l_weight: DOUBLE
		do
			l_weight := sqrt (a_weight)
			add_observation_compact_imp (a_xdata, a_ydata, l_weight, False)
		end

feature -- Impact of each observation

	coefficients_minus_observation (a_x: VECTOR; a_weight: DOUBLE; a_y: DOUBLE): VECTOR is
			-- impact of removing an observation
		require
			xtx_not_void: xtx /= Void
		local
			l_xtx, l_xtx_total: like xtx
			l_xty, l_xty_total: like xty
		do
			create l_xtx_total.make (xtx.height)
			l_xtx := a_x.xtx.scalar_multiply (a_weight)
			l_xty := a_x.scaled (a_y * a_weight)
			l_xtx_total.add (xtx)
			l_xtx_total.subtract (l_xtx)
			l_xtx_total.replace_diagonal
			l_xty_total := xty - l_xty
			Result := l_xtx_total.inverse @ l_xty_total
		ensure
			coefficients_minus_observation_not_void: Result /= Void
		end

	contribution_matrix: MATRIX is
			-- xtx contributions
		require
			matrix_not_void: matrix /= Void
			coefficients_not_void: coefficients /= Void
		local
			i: INTEGER
			l_coefficients: VECTOR
		do
			create Result.make (matrix.height, coefficients.count)
			Result.set_row_names (matrix.row_names)
			Result.set_column_names (independent_variable_names)
			from
				i := 1
			until
				i > matrix.height
			loop
				l_coefficients := coefficients_minus_observation (matrix.row (i), 1.0, dependent_variables [i])
				Result.put_row_vector (coefficients - l_coefficients, i)
				i := i + 1
			end
		ensure
			contribution_matrix_not_void: Result /= Void
		end

feature {NONE} -- Impact of each observation

	add_observation_compact_imp (a_xdata, a_ydata: VECTOR; a_weight: DOUBLE; a_is_add: BOOLEAN) is
			-- Add or remove an observation to/from compact representation.
		require
			valid_x: a_xdata /= Void and then a_xdata.count = number_of_independent_variables
			valid_y: a_ydata /= Void and then a_ydata.count = number_of_dependent_variables
			xpsum_not_void: xpsum /= Void
			xw2sum_not_void: xw2sum /= Void
			xtx_not_void: xtx /= Void
			ypsum_not_void: ypsum /= Void
			yw2sum_not_void: yw2sum /= Void
			ypy_not_void: ypy /= Void
			xpy_not_void: xpy /= Void
		local
			l_sign: DOUBLE
			l_matrix: MATRIX
			l_sym_matrix: SYMMETRIC_MATRIX
			l_vector: VECTOR
		do
			if a_is_add then
				l_sign := 1
			else
				l_sign := -1
			end
			number_of_observations := number_of_observations + l_sign.floor

			sum_w2 := sum_w2 + a_weight * a_weight * l_sign
			xpsum.add (a_xdata.scaled (a_weight * l_sign))
			xw2sum.add (a_xdata.scaled (a_weight * a_weight * l_sign))

			l_sym_matrix := a_xdata.xtx
			l_sym_matrix.scale (a_weight * a_weight * l_sign)
			xtx.add (l_sym_matrix)

			ypsum.add (a_ydata.scaled (a_weight * l_sign))
			yw2sum.add (a_ydata.scaled (a_weight * a_weight * l_sign))

			l_vector := a_ydata.squares
			l_vector.scale (a_weight * a_weight * l_sign)
			ypy.add (l_vector)

			l_matrix := a_xdata.to_column_matrix * a_ydata.to_row_matrix
			l_matrix.scale (a_weight * a_weight * l_sign)
			xpy.add(l_matrix)
		ensure
			one_more_observation: a_is_add implies number_of_observations = old number_of_observations + 1
			one_fewer_observation: not a_is_add implies number_of_observations = old number_of_observations - 1
		end

feature	-- Element change

	update_variable_dependent (a_column: INTEGER; a_new_matrix : ABSTRACT_READABLE_MATRIX; a_dependent: VECTOR) is
			-- replaces dependent and independent variables
		require

			a_new_matrix_not_void: a_new_matrix /= Void
			a_dependent_not_void: a_dependent /= Void
			dependent_variables_not_void: dependent_variables /= Void
			dependent_variables_match: a_dependent.count = dependent_variables.count
		do
			matrix := a_new_matrix
			dependent_variables := a_dependent
			set_y_standard
			update_variable (a_column, a_new_matrix)
		ensure

		end

	update_variable_dependent_with_mask (a_column: INTEGER; a_new_matrix: ABSTRACT_READABLE_MATRIX;
		a_dependent: VECTOR;	a_mask: ARRAY[BOOLEAN]) is
			-- replace flagged dependent and independent variables
		require

			a_new_matrix_not_void: a_new_matrix /= Void
			a_dependent_not_void: a_dependent /= Void
			dependent_variables_exist: dependent_variables /= Void
			dependent_variables_match: a_dependent.count = dependent_variables.count
		do
			dependent_variables := a_dependent
			set_y_standard
			update_variable_with_mask (a_column, a_new_matrix, a_mask)
		ensure
			coefficients_not_void: coefficients /= Void
			valid_r_square: ((r_square_with_no_intercept >= 0) and (r_square_with_no_intercept <= 1) and
				(r_square_with_intercept >= 0) and (r_square_with_intercept <= 1))
		end

	update_variable_dependent_new (a_xtx: SYMMETRIC_MATRIX; a_xty: VECTOR) is
			-- update variable depdendent new
		require
			a_xtx_not_void: a_xtx /= Void
			a_xty_not_void: a_xty /= Void
		do
			xtx := a_xtx
			xty := a_xty
		 	xtx_inverse := Void
		ensure
			xtx_set: xtx = a_xtx
			xty_set: xty = a_xty
		 	xtx_inverse_void: xtx_inverse = Void
		end

	update_variable (a_column: INTEGER; a_matrix: ABSTRACT_READABLE_MATRIX) is
			-- update variable
		local
			l_xty: DOUBLE
			l_vector: VECTOR
		do
			matrix := a_matrix

			l_vector := matrix.transpose_at (matrix.column (a_column))
			xtx.put_row_vector (l_vector, a_column)
			xtx.put_column_vector (l_vector, a_column)

			l_xty  := matrix.column (a_column) * dependent_variables
			xty.put (l_xty, a_column)

			xty.put (l_xty, a_column)
			xtx_inverse := Void

			regress
		end

	update_variable_with_mask (a_column: INTEGER; a_matrix: ABSTRACT_READABLE_MATRIX; a_mask: ARRAY[BOOLEAN]) is
			-- update variable with mask
		require
			a_matrix_not_void : a_matrix /= Void
			legal_variable: a_column >= 1 and a_column <= number_of_independent_variables
			dependent_variables_not_void: dependent_variables /= Void
			xtx_not_void: xtx /= Void
			transpose_computed_at_dependent_variables: xty /= Void
		local
			i, j, ii: INTEGER
			l_dot_product, l_xty: DOUBLE
			l_indices: ARRAYED_LIST [INTEGER]
		do
			from i := 1; j := 1 until i > a_mask.count loop
				if (a_mask.item (i)) then
					l_indices.extend (j)
					j := j + 1
				end
				i := i + 1
			end
			matrix := a_matrix
			from j := 1 until j > number_of_independent_variables loop
				l_dot_product := 0
				from ii := 1 until ii > l_indices.count loop
					i := l_indices.i_th (ii)
					if j = a_column then
						l_xty := l_xty + matrix.item(i,j) * dependent_variables.item(i)
					end
					l_dot_product := l_dot_product + matrix.item(i,a_column) * matrix.item(i,j)
					ii := ii + 1
				end
				xtx.put(l_dot_product,j,a_column)
				xtx.put(l_dot_product,a_column,j)
				j := j + 1
			end
			xty.put(l_xty,a_column)
			xtx_inverse := Void

			regress
		ensure
			coefficients_not_void: coefficients /= Void
			valid_r_square: ((r_square_with_no_intercept >= 0) and (r_square_with_no_intercept <= 1) and
				(r_square_with_intercept >= 0) and (r_square_with_intercept <= 1))
		end

	add_observation (a_dependent_variables: VECTOR; a_new_matrix: ABSTRACT_READABLE_MATRIX; a_dependent_value: DOUBLE) is
			-- Add observation from last row of `a_new_matrix' to normal representation and re-compute regression results.
		local
			l_xtx: SYMMETRIC_MATRIX
			i: INTEGER
			l_observation: VECTOR
		do
			number_of_observations := number_of_observations + 1
			matrix:= a_new_matrix
			l_observation := a_new_matrix.row (a_new_matrix.height)
			a_dependent_variables.extend(a_dependent_value)

			l_xtx := l_observation.xtx
			xtx.add(l_xtx)
			xtx_inverse := Void

			from i := 1 until i > number_of_independent_variables loop
				xty.put(xty.item(i) + l_observation.item(i) * a_dependent_value,i)
				i := i + 1
			end

			yty := yty + a_dependent_value * a_dependent_value
			y_total_variation := y_total_variation_standard

			regress
		end

feature -- Statistics

	population_standard_errors: VECTOR is
			-- Standard errors using population mean (as used by NAG);
			-- SAS uses sample mean
		require
			standard_errors_not_void: standard_errors /= Void
			observations_not_fewer_than_explanatory_variables: number_of_observations >= number_of_independent_variables
		local
			l_factor : DOUBLE
			i : INTEGER
		do
			create Result.make (standard_errors.count)
			l_factor := sqrt ((number_of_observations - number_of_independent_variables) / (number_of_observations - number_of_independent_variables +  1))
			from i := 1 until i > Result.count loop
				Result.put (l_factor * standard_errors.item(i) ,i)
				i := i + 1
			end
		ensure
			population_standard_errors: Result /= Void
			correct_count: Result.count = standard_errors.count
		end

	adjusted_r_square_with_intercept: DOUBLE is
			-- Coefficient of determination adjusted for number of explanatory terms in model with intercept
		do
			Result := 1 - ((number_of_observations - 1) / (number_of_observations - number_of_independent_variables)) * (1-r_square_with_intercept)
		end

	adjusted_r_square_with_no_intercept: DOUBLE is
			-- Coefficient of determination adjusted for number of explanatory terms in model without intercept
		do
			Result := 1 - (number_of_observations  / (number_of_observations - number_of_independent_variables)) * (1-r_square_with_no_intercept)
		end

	t_statistic (i: INTEGER): DOUBLE is
			-- Student's T for `i'th coefficient
		do
			Result := coefficients.item(i) / standard_errors.item (i)
		end

	t_statistics: VECTOR is
			-- Student's T for corresponding `coefficients'
		require
			standard_errors_not_void: standard_errors /= Void
			coefficients_not_void: coefficients /= Void
			same_size: coefficients.count = standard_errors.count
			correct_count: coefficients.count = number_of_independent_variables
		do
			Result := coefficients.divided_by (standard_errors)
		ensure
			t_statistics_not_void: Result /= Void
			same_count: Result.count = coefficients.count
		end

	chi_square: DOUBLE is
			-- Sum of squared errors
		do
			Result := yty - xty * coefficients
		end

	f_statistic: DOUBLE is
			-- Fisher's variance ratio
		local
			l_total_variance, l_explained_variance, l_unexplained_variance: DOUBLE
		do
			l_total_variance := y_total_variation
			l_unexplained_variance := sum_of_squared_errors
			l_explained_variance := l_total_variance - l_unexplained_variance
			Result := l_explained_variance / l_unexplained_variance
		end

	y_total_variation_standard: DOUBLE is
			-- y total variance standard
		require
			dependent_variables_not_void: dependent_variables /= Void
			matrix_not_void: matrix /= Void
			weight_roots_not_void: weights /= Void implies weight_roots /= Void
		local
			l_dependent_average: DOUBLE
		do
			if weights = Void then
				l_dependent_average := dependent_variables.average
				Result := yty - matrix.height * l_dependent_average * l_dependent_average
			else
				l_dependent_average := dependent_variables.cross (weight_roots).sum / weights.sum
				Result := yty - weights.sum * l_dependent_average * l_dependent_average
			end
		end

	fitted_values: VECTOR is
			-- Fitted values using the weighted matrix, not the unweighted as usual
		do
			if matrix /= Void and coefficients /= Void then
				Result := matrix @ coefficients
			end
		end

	residuals: VECTOR is
			-- Difference between observed and estimated values
		require
			dependent_variables_not_void: dependent_variables /= Void
			matrix_not_void: matrix /= Void
		do
			Result := dependent_variables - (matrix @ coefficients)
		ensure
			residuals_not_void: Result /= Void
			correct_count: Result.count = matrix.height
		end

	unweighted_residuals: VECTOR is
			-- True difference between observed and estimated values
		require
			unweighted_dependent_variables_not_void: unweighted_dependent_variables /= Void
			unweighted_matrix_not_void: unweighted_matrix /= Void
		do
			Result := unweighted_dependent_variables - (unweighted_matrix @ coefficients)
		ensure
			unweighted_residuals_not_void: Result /= Void
			correct_count: Result.count = unweighted_matrix.height
		end

feature -- Names	

	set_all_names is
			-- Set names on all matrices and vectors
		require
			xtx_not_void: xtx /= Void
			xty_not_void: xty /= Void
			coefficients_not_void: independent_variable_names /= Void implies coefficients /= Void
			covariance_not_void: independent_variable_names /= Void implies covariance /= Void
		do
			if independent_variable_names /= Void then
				coefficients.set_names (independent_variable_names)
				coefficients.set_name ("coefficients")
				if standard_errors/= Void then
					standard_errors.set_names (independent_variable_names)
					standard_errors.set_name ("standard errors")
				end
				covariance.set_row_names (independent_variable_names)
				covariance.set_column_names (independent_variable_names)
				xtx.set_row_names (independent_variable_names)
				xtx.set_column_names (independent_variable_names)

				if weighted_matrix /= Void then
					weighted_matrix.set_name ("weighted_independent")
					weighted_matrix.set_column_names (independent_variable_names)
					weighted_matrix.matrix.set_name ("independent_matrix")
					weighted_matrix.matrix.set_column_names (independent_variable_names)
				else
					if matrix /= Void then
						matrix.set_column_names (independent_variable_names)
					end
				end
			end
			xtx.set_name ("xtx")
			xty.set_name ("xty")
			if weights /= Void then
				weights.set_name ("wts")
			end
			if weight_roots /= Void then
				weight_roots.set_name ("sqrt(wts)")
			end
			if dependent_variable_name /= Void then
				if weighted_dependent_variables /= Void then
					weighted_dependent_variables.set_name ("weighted_dependent")
					weighted_dependent_variables.vector.set_name (dependent_variable_name)
				else
					if dependent_variables /= Void then
						dependent_variables.set_name (dependent_variable_name)
					end
				end
			end
		end

	name: STRING
			-- Optional name for regression

	set_name (a_name: STRING) is
		-- Set `name' to `a_name'.
		do
			name := a_name
		ensure
			name_set: name = a_name
		end

	dependent_variable_name: STRING
			-- Optional name for response variable

	set_dependent_variable_name (a_name: STRING) is
			-- Set `dependent_variable_name' to `a_name'.
		do
			dependent_variable_name := a_name
		ensure
			dependent_variable_name_set: dependent_variable_name = a_name
		end

	independent_variable_names: ROSE_LINEAR_ARRAY [STRING]
			-- Optional names for explanatory variables

	set_independent_variable_names (a_names: ARRAY[STRING]) is
			-- Create `independent_variable_names' from `a_names'.
		do
			if a_names /= Void then
				create independent_variable_names.make_from_array (a_names)
			end
		ensure
			independent_variable_names_not_void: a_names /= Void implies independent_variable_names /= Void
		end

	weight_variable_name: STRING
			--

	set_weight_variable_name (a_name: STRING) is
			-- Set `weight_variable_name' to `a_name'.
		do
			weight_variable_name := a_name
		ensure
			weight_variable_name_set: weight_variable_name = a_name
		end

	dependent_variable_names: ROSE_LINEAR_ARRAY [STRING]
			-- Names of response variables

	set_dependent_variable_names (a_names: ARRAY[STRING]) is
			-- Create `dependent_variable_names' from `a_names'.
		do
			if a_names /= Void then
				create dependent_variable_names.make_from_array(a_names)
			end
		ensure
			dependent_variable_names_not_void: a_names /= Void implies dependent_variable_names /= Void
		end

feature -- Weights

	regression_with_new_weights (a_weights: VECTOR): like Current is
			-- new regression with new weights
		do
			if weighted_matrix /= Void and weighted_dependent_variables /= Void then
				create Result.make_weighted (weighted_matrix.matrix, weighted_dependent_variables.vector, a_weights)
				set_weightd_attributes (Result)
			end
		end

	set_weightd_attributes (a_regression: like Current) is
			-- set weighted attributes
		require
			a_regression_not_void: a_regression /= Void
		do
			a_regression.set_dependent_variable_name (dependent_variable_name)
			a_regression.set_independent_variable_names (independent_variable_names)
			a_regression.set_weight_variable_name (weight_variable_name)
		end

feature -- Conversion

	to_original_vector (a_vector: ABSTRACT_READABLE_VECTOR): VECTOR is
			-- to original vector
		local
			i, l_ix: INTEGER
		do
			if constraint_lhs /= Void then
				create Result.make(constraint_lhs.count)
				from i := 1 until i > constraint_lhs.count loop
					if l_ix = 0 and then constraint_lhs[i] /= 0 then
						l_ix := i
					end
					if l_ix = 0 then
						Result[i] := a_vector[i]
					elseif l_ix < i then
						Result[i] := a_vector[i - 1]
					end
					i := i + 1
				end
			else
				create Result.make_from_array (a_vector.to_array)
			end
		end

feature	-- Output

	statistics: STRING is
			-- Formatted output of the statistics from the regression
		local
			i: INTEGER
			l_population_error: VECTOR
		do
			create Result.make (500)
			if weights /= Void then
				Result.append ("Weighted %N")
			end
			Result.append ("Residual sum of squares : " + formatted (chi_square) + "%N")
			Result.append ("R-square(with intercept)%Tadjusted%N")
			Result.append (formatted (r_square_with_intercept) + "%T%T%T" +
						  formatted (adjusted_r_square_with_intercept) + "%N")
			Result.append ("R-square(without intercept)%Tadjusted%N")
			Result.append (formatted (r_square_with_no_intercept) + "%T%T%T" +
						  formatted (adjusted_r_square_with_no_intercept) + "%N")
			Result.append ("Coefficient%TStandard error%Tt-statistic%TPopulation Standard Error%N")
			l_population_error := population_standard_errors
			from i := 1 until i > matrix.width loop
				Result.append(formatted (coefficients [i]) + "%T" + formatted (standard_errors [i]) + "%T"
					+ formatted (t_statistic (i)) + "%T" + formatted (l_population_error [i]) + "%N")
				i := i + 1
			end
		end

	residual_statistics: STRING is
			-- Formatted output of residuals, standard errors of residuals, and studentized residuals;
			-- The studentized residuals returned are "internally" studentized, since that is what SAS returns.
		local
			i: INTEGER
			l_residuals: VECTOR
		do
			l_residuals := unweighted_residuals
			create Result.make_empty
			Result.append ("Residual%TResidual Standard Error%T%TStudentized Residual%N")
			from i := 1 until i > matrix.height loop
				Result.append (formatted (l_residuals [i]) + "%T" + formatted (residual_standard_errors [i])
					+ "%T%T%T" + formatted (studentized_residuals [i]) + "%T%T" +  "%N")
				i := i + 1
			end
		end

	residual_statistics_test: STRING is
			-- formatted output of residuals, standard errors of residuals, and studentized residuals
			-- this just prints the first 10 -- and thus may be useful for testing
		local
			i : INTEGER
			l_residuals : VECTOR
		do
			l_residuals := unweighted_residuals
			create Result.make_empty
			Result.append("Residual%TResidual STE%TInternal Studentized%TWeight%N")
			from i := 1 until (i > matrix.height or i>10) loop
				Result.append(formatted(l_residuals[i]) + "%T" + formatted(residual_standard_errors[i]) + "%T" + formatted(studentized_residuals[i]) + "%T%T" + formatted(weights[i]) + "%N")
				i := i + 1
			end
		end

	regression_data: STRING is
			-- puts the input data into a string so it can be printed or put into a file
			-- format of a line of the string is y x1 ... xn w
		local
			i, j : INTEGER
		do
			create Result.make_empty
			from i := 1 until i > matrix.height loop
				Result.append (unweighted_dependent_variables.item(i).out + "%T")
				from j := 1 until j > matrix.width loop
					Result.append (unweighted_matrix.item(i,j).out + "%T")
					j := j + 1
				end
				Result.append (weights.item(i).out + "%N")
				i := i + 1
			end
		end

	print_data_to_file(a_output_file : STRING) is
			-- prints the regression data to a file using 'regression_data'
		local
			l_output_file : PLAIN_TEXT_FILE
		do
			create l_output_file.make_open_write (a_output_file)
			l_output_file.put_string (regression_data)
			l_output_file.close
		end


	labels_row: ROSE_LINEAR_ARRAY[STRING] is
			-- useful when reporting a series of regressions
		local
			i:  INTEGER
		do
			create Result.make_from_array(<<
			"Obs",
			"Res sum sqrs",
			"R2(w/intcept)",
			"AdjR2(w/intcept)",
			"R2(wo/intcept)",
			"AdjR2(wo/intcept)"
			>>)
			from i:= 1 until i> matrix.width loop
				if independent_variable_names /= Void then
					Result.extend(independent_variable_names.item(i))
				else
					Result.extend("V" + i.out)
				end
				Result.extend("std err")
				Result.extend("t-stat")
				Result.extend("pop std err")
				i:= i+ 1
			end
		end

	output_vector: VECTOR is
			-- output vector
		local
			i: INTEGER
			l_population_error: VECTOR
			l_names:ROSE_LINEAR_ARRAY[STRING]
			l_inds:ROSE_LINEAR_ARRAY[STRING]
		do
			l_population_error := population_standard_errors
			create l_names.make_empty
			create Result.make(6 + (matrix.width* 4))
			--Result.put((matrix.height//1).to_double, 1)
			Result.put((matrix.height//1), 1)
			l_names.extend ("count")
			Result.put((chi_square), 2)
			l_names.extend ("chi_square")
			Result.put((r_square_with_intercept), 3)
			l_names.extend ("r_square_with_intercept")
			Result.put((adjusted_r_square_with_intercept) ,4)
			l_names.extend ("adjusted_r_square_with_intercept")
			Result.put((r_square_with_no_intercept), 5)
			l_names.extend ("r_square_with_no_intercept")
			Result.put((adjusted_r_square_with_no_intercept), 6)
			l_names.extend ("adjusted_r_square_with_no_intercept")
			l_inds:=safe_independent_variable_names
			from i:= 1 until i> matrix.width loop
				l_names.extend (l_inds.item(i))
				Result.put((coefficients.item(i)), 6 + ((i- 1)* 4) + 1)
				l_names.extend ("standard_errors"+"_"+l_inds.item(i))
				Result.put((standard_errors.item(i)), 6 + ((i- 1)* 4) + 2)
				l_names.extend ("t"+"_"+l_inds.item(i))
				Result.put((t_statistic(i)), 6 + ((i- 1)* 4) + 3)
				l_names.extend ("population_error"+"_"+l_inds.item(i))
				Result.put((l_population_error.item(i)), 6 + ((i- 1)* 4) + 4)
				i:= i+ 1
			end
			result.set_names (l_names)
		end

	safe_independent_variable_names:ROSE_LINEAR_ARRAY[STRING] is
			-- safe independent variabel names
		local
			i:INTEGER
		do
			if independent_variable_names=void then
				create result.make_empty
				from i:=1	until i> coefficients.count loop
					result.extend("indedendent_variable_"+i.out)
					i:=i+1
				end
			else
				create result.make_from_array(independent_variable_names)
			end
		end


feature -- ouput; data summary

	data_summary_labels: ROSE_LINEAR_ARRAY[STRING] is
				-- data summary labels
			local
				i: INTEGER
			do
				Create Result.make_empty
				Result.append_array(<<"simple mean_"+dependent_variable_name, "wgt mean", "std", "min", "max">>)
				from i:= 1 until i > matrix.width loop
					Result.append_array(<< "simple mean_"+ independent_variable_names.item(i),
											"wgt mean_"+ independent_variable_names.item(i),
											"std_"+ independent_variable_names.item(i),
											"min_"+ independent_variable_names.item(i),
											"max_"+ independent_variable_names.item(i)>>)
					i:= i+ 1
				end
			end

	data_summary_vector: VECTOR is
				-- data summary vector
			local
				i: INTEGER
				l_wgt_moments, l_simple_moments: ROSE_MOMENTS
				l_x: ABSTRACT_READABLE_MATRIX
				l_y: EXTENDABLE_VECTOR
			do
				if weighted_matrix /= Void then
					l_x:= weighted_matrix.matrix
					l_y:= weighted_dependent_variables.vector
				else
					l_x:=  matrix
					l_y:= dependent_variables
				end

				create Result.make((l_x.width+ 1)* 5)
				create l_wgt_moments.make(l_y.to_array, weight_roots.to_array)
				create l_simple_moments.make(l_y.to_array, Void)
				Result.put(l_simple_moments.mean, 1)
				Result.put(l_wgt_moments.mean, 2)
				Result.put(l_simple_moments.standard_deviation, 3)
				Result.put(l_simple_moments.minimum, 4)
				Result.put(l_simple_moments.maximum, 5)
				from i:= 1 until i > l_x.width loop
					create l_wgt_moments.make(l_x.column(i).to_array, weight_roots.to_array)
					create l_simple_moments.make(l_x.column(i).to_array, Void)
					Result.put(l_simple_moments.mean, (i* 5)+ 1)
					Result.put(l_wgt_moments.mean,(i* 5)+2)
					Result.put(l_simple_moments.standard_deviation,(i* 5)+ 3)
					Result.put(l_simple_moments.minimum,(i* 5)+ 4)
					Result.put(l_simple_moments.maximum,(i* 5)+ 5)
					i:= i+1
				end
			end

	data_variance_covariance: SYMMETRIC_MATRIX is
			-- raw data variance and covariance matrix
		do
			Result := data_covariance (False)
		end


	data_summary_stats: MATRIX is
			-- output raw data statistic in matrix format: row is variables, column is statistics
			-- including depend variables
			-- have order statistics; could be slow
			do
				Result := data_statistics (False, True)
			end

	weighted_data_summary_stats: MATRIX is
			-- output raw data statistic in matrix format: row is variables, column is statistics
			-- including depend variables
			-- have order statistics; could be slow
			do
				if weighted_matrix /= Void then
					Result := data_statistics (True, True)
				else
					Result := data_statistics (False, True)
				end
			end

	data_summary_simple_stats: MATRIX is
			-- output raw data statistic in matrix format: row is variables, column is statistics
			-- including depend variables
			-- do not include order statistics
			do
				Result := data_statistics (False, False)
			end

	weighted_data_summary_simple_stats: MATRIX is
			-- output raw data statistic in matrix format: row is variables, column is statistics
			-- including depend variables
			-- do not include order statistics
			do
				if weighted_matrix /= Void then
					Result := data_statistics (True, False)
				else
					Result := data_statistics (False, False)
				end
			end

	labeled_output: ROSE_ARRAY2 [ANY] is
		-- Tabular regression results suitable for RP_DATA_GRID or excel
		local
			i, l_ix: INTEGER
			l_population_error, l_normalized_constraint: VECTOR
			l_title: STRING
			l_student_t: DOUBLE
			l_student: ROSE_PR_DISTRIBUTION_STUDENTS_T
		do
			create Result.make (matrix.width + 18, 8)

			if name /= Void then
				l_title := "REGRESSION RESULTS - " + name
			else
				l_title := "REGRESSION RESULTS"
			end
			Result.put (l_title, 1, 1)
			Result.put ("Observations", 3, 1)
			Result.put ("Residual sum of squares", 4, 1)
			Result.put ("R-square", 5, 1)
			Result.put ("Adjusted R-square", 6, 1)
			Result.put ("Intercept fitted?", 7, 1)
			Result.put (intercept_used.out, 7, 2)
			Result.put (matrix.height.out, 3, 2)
			Result.put (formatted (chi_square), 4, 2)
			Result.put (formatted (r_square), 5, 2)
			Result.put (formatted(adjusted_r_square), 6, 2)
			l_population_error := population_standard_errors

			if prior_mean /= Void then
				Result.put("GLS", 10, 2)
				Result.put("Bayesian", 10, 3)
				Result.put("Prior Mean", 10, 4)
				Result.put("Prior Weight", 10, 5)
				Result.put("Standard error", 10, 6)
				Result.put("t-statistic", 10, 7)
				Result.put("Population Standard Error", 10, 8)

				from i:= 1 until i > matrix.width loop
					if independent_variable_names /= Void then
						Result.put(independent_variable_names.item(i), 10 + i, 1)
					else
						Result.put("V" + i.out, 10 + i, 1)
					end
					Result.put(formatted(original_coefficients.item(i)), 10 + i, 2)
					Result.put(formatted(coefficients.item(i)), 10 + i, 3)
					Result.put(formatted(prior_mean.item(i)), 10 + i, 4)
					Result.put(formatted(prior_cov_inverse[i, i]), 10 + i, 5)
					Result.put(formatted(standard_errors.item(i)), 10 + i, 6)
					Result.put(formatted(t_statistic(i)), 10 + i, 7)
					Result.put(formatted(l_population_error.item(i)), 10 + i, 8)
					i:= i+ 1
				end
			elseif constraint_lhs /= Void then
				l_normalized_constraint := constraint_lhs.normalized
				Result.put("Coefficient", 8, 2)
				Result.put("Constraint-Coeff", 8, 3)
				Result.put("Constraint", 8, 4)
				Result.put("Constraint - Normalized", 8, 5)
				Result.put("Standard error", 8, 6)
				Result.put("t-statistic", 8, 7)
				Result.put("Population Standard Error", 8, 8)

				from i:= 1 until i > constraint_lhs.count loop
					if l_ix = 0 and then constraint_lhs[i] /= 0 then
						l_ix := i
					end
					if original_coefficients.names /= Void then
						Result.put(original_coefficients.names[i], 8 + i, 1)
					else
						Result.put("V" + i.out, 8 + i, 1)
					end
					if l_ix = 0 then
						Result.put(formatted(coefficients[i]), 8 + i, 2)
						Result.put(formatted(original_coefficients[i]), 8 + i, 3)
						Result.put(formatted(constraint_lhs[i]), 8 + i, 4)
						Result.put(formatted(l_normalized_constraint[i]), 8 + i, 5)
						Result.put(formatted(standard_errors[i]), 8 + i, 6)
						Result.put(formatted(t_statistic(i)), 8 + i, 7)
						Result.put(formatted(l_population_error[i]), 8 + i, 8)
					elseif l_ix = i then
						Result.put(formatted(original_coefficients[i]), 8 + i, 3)
						Result.put(formatted(constraint_lhs[i]), 8 + i, 4)
						Result.put(formatted(l_normalized_constraint[i]), 8 + i, 5)
					elseif l_ix < i then
						Result.put(formatted(coefficients[i - 1]), 8 + i, 2)
						Result.put(formatted(original_coefficients[i]), 8 + i, 3)
						Result.put(formatted(constraint_lhs[i]), 8 + i, 4)
						Result.put(formatted(l_normalized_constraint[i]), 8 + i, 5)
						Result.put(formatted(standard_errors[i - 1]), 8 + i, 6)
						Result.put(formatted(t_statistic(i - 1)), 8 + i, 7)
						Result.put(formatted(l_population_error[i - 1]), 8 + i, 8)
					end
					i:= i+ 1
				end
			else
				Result.put ("Coefficient", 8, 2)
				Result.put ("Standard error", 8, 3)
				Result.put ("t-statistic", 8, 4)
				Result.put ("Population Standard Error", 8, 5)
				Result.put ("Pr > |t|", 8, 6)

				create l_student.make (degree_of_freedom)

				from i:= 1 until i > matrix.width loop
					if independent_variable_names /= Void then
						Result.put (independent_variable_names [i], 8 + i, 1)
					else
						Result.put ("V" + i.out, 8 + i, 1)
					end
					l_student_t := t_statistic (i)
					Result.put (formatted (coefficients [i]), 8 + i, 2)
					Result.put (formatted (standard_errors [i]), 8 + i, 3)
					Result.put (formatted (l_student_t), 8 + i, 4)
					Result.put (formatted (l_population_error [i]), 8 + i, 5)
					Result.put (formatted (l_student.two_tail_probability (l_student_t)), 8 + i, 6)
					i:= i+ 1
				end
			end
		end


	labeled_residual_output : ARRAY2[ANY] is
		-- residual data output suitable for multi-column list or excel
		local
			i : INTEGER
			l_residuals : VECTOR
		do
			l_residuals := unweighted_residuals
			create Result.make(matrix.height + 1, 3)

			Result.put("Residual", 1, 1)
			Result.put("Residual Standard Error", 1, 2)
			Result.put("Studentized Residual", 1, 3)

			from i:= 1 until i > matrix.height loop
				Result.put(formatted(l_residuals[i]), 1 + i, 1)
				Result.put(formatted(residual_standard_errors[i]), 1 + i, 2)
				Result.put(formatted(studentized_residuals[i]), 1 + i, 3)
				i:= i + 1
			end
		end

	result_vector : VECTOR is
			-- helpful summary vector of regression coefficients and fit statistics
		local
			i : INTEGER
			l_t_stat : DOUBLE
		do
			create Result.make(number_of_independent_variables * 2 + 6)
			from i := 1 until i > number_of_independent_variables loop
				Result.put(coefficients.item(i), i)
				l_t_stat := coefficients.item(i) / standard_errors.item(i)
				Result.put(l_t_stat, number_of_independent_variables + i)
				i := i + 1
			end

			Result.put(number_of_observations, number_of_independent_variables * 2 + 1)
			Result.put(chi_square, number_of_independent_variables * 2 + 2)
			Result.put(r_square_with_intercept, number_of_independent_variables * 2 + 3)
			Result.put(adjusted_r_square_with_intercept, number_of_independent_variables * 2 + 4)
			Result.put(r_square_with_no_intercept, number_of_independent_variables * 2 + 5)
			Result.put(adjusted_r_square_with_no_intercept, number_of_independent_variables * 2 + 6)
		end

	result_labels : ROSE_LINEAR_ARRAY[STRING] is
			-- labelling for results matrix with given or generic variable labels plus statistics names
		local
			i : INTEGER
		do
			create Result.make(1, number_of_independent_variables * 2 + 6)
			from i := 1 until i > number_of_independent_variables loop
				if independent_variable_names /= Void then
					Result.put(independent_variable_names.item(i), i)
					Result.put("t(" + independent_variable_names.item(i)+ ")", i + number_of_independent_variables)
				else
					Result.put("V_" + i.out, i)
					Result.put("t(V_" + i.out + ")", i + number_of_independent_variables)
				end
				i := i + 1
			end
			Result.put("Observations", number_of_independent_variables * 2 + 1)
			Result.put("Residual sum of squares", number_of_independent_variables * 2 + 2)
			Result.put("R-square (with intercept)", number_of_independent_variables * 2 + 3)
			Result.put("Adj R-square (with intercept)", number_of_independent_variables * 2 + 4)
			Result.put("R-square (without intercept)", number_of_independent_variables * 2 + 5)
			Result.put("Adj R-square (without intercept)", number_of_independent_variables * 2 + 6)
		end

feature {NONE} -- Implementation

	data_covariance (a_weighted: BOOLEAN): SYMMETRIC_MATRIX is
			-- data variance and covariance
			-- include dependent variables
			local
				l_var_count: INTEGER
				l_x: ABSTRACT_READABLE_MATRIX
				l_y: EXTENDABLE_VECTOR
				l_weights: VECTOR
				l_var_names: ROSE_LINEAR_ARRAY [STRING]
				l_matrix: MATRIX
			do
				if a_weighted then
					l_x := weighted_matrix
					l_y := weighted_dependent_variables
				else
					if weighted_matrix /= Void then
						l_x:= weighted_matrix.matrix
						l_y:= weighted_dependent_variables.vector
					else
						l_x:=  matrix
						l_y:= dependent_variables
					end
				end

				-- combine dependent and independent variables
				create l_matrix.make_from_others (<<l_y.to_column_matrix>>)
				create l_var_names.make_from_array (<<dependent_variable_name>>)

				l_matrix := l_matrix.extend_matrix (l_x)
				l_var_names.append_array (independent_variable_names)

				l_var_count := l_x.width + 1 -- including dependent variable
				-- whether to display weighted data
				if a_weighted then
					l_weights := weight_roots
					l_matrix := l_matrix.extend_column_vectors (<<l_weights>>)
					l_var_names.extend ("wgt_root")
					l_var_count := l_var_count + 1
				end

				l_matrix.set_column_names (l_var_names)
				Result := l_matrix.covariance (Void, Void)
			end


	data_statistics (a_weighted: BOOLEAN; a_include_order_stats: BOOLEAN): MATRIX is
			-- whether to include orfer stats
			local
				i: INTEGER
				l_moments: ROSE_MOMENTS
				l_x: ABSTRACT_READABLE_MATRIX
				l_y: EXTENDABLE_VECTOR
				l_weights: ARRAY [DOUBLE]
				l_stat_vector: VECTOR
				l_var_names, l_stat_names: ROSE_LINEAR_ARRAY [STRING]
			do
				if weighted_matrix /= Void then
					l_x:= weighted_matrix.matrix
					l_y:= weighted_dependent_variables.vector
				else
					l_x:=  matrix
					l_y:= dependent_variables
				end

				-- whether to display weighted data
				if a_weighted then
					l_weights := weight_roots.to_array
				end

				-- dependent variable data
				create l_moments.make(l_y.to_array, l_weights)
				if a_include_order_stats then
					l_stat_vector := l_moments.to_vector
				else
					l_stat_vector := l_moments.simple_statistics_to_vector
				end
				l_stat_names := l_stat_vector.names

					-- create container
				create l_var_names.make_from_array (<<dependent_variable_name>>)
				l_var_names.append_array (independent_variable_names)
				if a_weighted then
					l_var_names.extend ("weights")
				end
				create Result.make(l_var_names.count, l_stat_names.count)

				Result.put_row_vector (l_stat_vector, 1)

				-- independent variable data
				from i:= 1 until i > l_x.width loop
					create l_moments.make(l_x.column(i).to_array, l_weights)
					if a_include_order_stats then
						l_stat_vector := l_moments.to_vector
					else
						l_stat_vector := l_moments.simple_statistics_to_vector
					end
					Result.put_row_vector (l_stat_vector, i+1)
					i:= i+1
				end

				-- weights
				if a_weighted then
					create l_moments.make (l_weights, Void)
					if a_include_order_stats then
						l_stat_vector := l_moments.to_vector
					else
						l_stat_vector := l_moments.simple_statistics_to_vector
					end
					Result.put_row_vector (l_stat_vector, l_var_names.count)
				end
				-- set names
				Result.set_row_names (l_var_names)
				Result.set_column_names (l_stat_names)
			end

	formatted (a_value: DOUBLE): STRING is
			-- `a_value' formatted nicely
		do
			Result := format_double.formatted (a_value)
		ensure
			formatted_not_void: Result /= Void
		end

	format_double: FORMAT_DOUBLE is
			-- Value formatter
		once
			create Result.make (10,4)
		ensure
			format_double_not_void: Result /= Void
		end


feature -- Test and validation (remove to a different class)

	test_dependent : VECTOR is
			-- test dependent
		do
			create Result.make_from_array(<<33.63, 39.62, 38.18, 41.46, 38.02, 35.83, 35.99, 36.58, 42.92, 37.80, 40.43, 37.89>>)
		end

	-- first column is all ones to simulate an intercept	
	test_observations : MATRIX is
			-- test observations
		do
			create Result.make_from_array (<<
								<<1.0,1.0, 2.0, 0.0, 0.0>>,
								<<1.0,0.0, 2.0, 0.0, 1.0>>,
								<<1.0,0.0, 1.0, 3.0, 0.0>>,
								<<1.0,0.0, 0.0, 1.0, 0.0>>,
								<<1.0,0.0, 0.0, 3.0, 1.0>>,
								<<1.0,0.0, 1.0, 0.0, 4.0>>,
								<<1.0,0.0, 4.0, 0.0, 1.0>>,
								<<1.0,1.0, 0.0, 0.0, 0.0>>,
								<<1.0,0.0, 0.0, 1.0, 0.0>>,
								<<1.0,1.0, 0.0, 0.0, 0.0>>,
								<<1.0,0.0, 0.0, 1.0, 0.0>>,
								<<1.0,0.0, 1.0, 0.0, 0.0>>
								>>)
		end

	test_observations_without_intercept_column : MATRIX is
			-- same as test_observations but without first column of all 1's
		do
			create Result.make_from_array (<<
								<<1.0, 2.0, 0.0, 0.0>>,
								<<0.0, 2.0, 0.0, 1.0>>,
								<<0.0, 1.0, 3.0, 0.0>>,
								<<0.0, 0.0, 1.0, 0.0>>,
								<<0.0, 0.0, 3.0, 1.0>>,
								<<0.0, 1.0, 0.0, 4.0>>,
								<<0.0, 4.0, 0.0, 1.0>>,
								<<1.0, 0.0, 0.0, 0.0>>,
								<<0.0, 0.0, 1.0, 0.0>>,
								<<1.0, 0.0, 0.0, 0.0>>,
								<<0.0, 0.0, 1.0, 0.0>>,
								<<0.0, 1.0, 0.0, 0.0>>
								>>)
		end


	test_observations_small : MATRIX is
			-- a very small test case
		do
			create Result.make_from_array (<<
								<<1.0, 2.0>>,
								<<0.5, 2.0>>,
								<<0.0, 1.0>>,
								<<1.5, 0.5>>
								>>)
		end

	test_dependent_small : VECTOR is
			-- Test dependent small
		do
			create Result.make_from_array(<<36.58,42.92,37.80,40.43>>)
		end

	test_weighted is
			-- test a weighted regression example
		local
			l_weights : VECTOR
		do
			create l_weights.make_from_array(<<.1,.1,.1,.1,.1,.1,.1,.1,.1,0,.05,.05>>)
			make_weighted_with_intercept(test_observations_without_intercept_column, test_dependent, l_weights)
			calculate_residual_statistics
			print (statistics)
			print (residual_statistics)
		end

	test_small_weighted is
			-- test a very small weighted regression example
		local
			l_weights : VECTOR
		do
			create l_weights.make_from_array(<<0.4,0.3,0.3,0.0>>)
			make_weighted(test_observations_small, test_dependent_small, l_weights)
			calculate_residual_statistics
			print (statistics)
			print (residual_statistics)
		end

	test_simple: STRING is
			-- test simple
		local
			l_observations : MATRIX
		do
			l_observations := test_observations
			make(l_observations,test_dependent)
			Result := statistics
		end

	test_update: STRING is
		--test update
		local
			l_observations : MATRIX
		do
			l_observations := test_observations
			make(l_observations,test_dependent)
			update_variable(5,l_observations) --mix things up
			Result := statistics
		end

	test_add: STRING is
		--test
		local
			l_observations : MATRIX
			l_dependents : VECTOR
			l_vector : VECTOR
		do
			l_dependents := test_dependent.sub_vector (1, test_dependent.count-1)
			l_observations := test_observations.sub_matrix_2 (1, test_observations.height-1, 1, test_observations.width)
			make(l_observations,l_dependents)
			create l_vector.make_from_array (test_observations.row (test_observations.height).to_array)
			l_observations.add_row(l_vector)
			add_observation (l_dependents, l_observations, test_dependent.last)
			Result := statistics
		end

	test is
		--This is from the NAG documentation for g02dac
		local
			l_test_dependent, l_dependent : VECTOR
			l_vector,l_weights : VECTOR
			l_test_observations, l_observations : MATRIX
		do
			l_test_observations := test_observations
			l_test_dependent := test_dependent
			make(l_test_observations,l_test_dependent)
			print(statistics)
			update_variable(5,l_test_observations) --mix things up
			print("should be the same as%N")
			update_variable(5,l_test_observations) --mix things up
			print(statistics)
			l_test_observations := test_observations
			make(l_test_observations,l_test_dependent)
			create l_vector.make_from_array (<<1.0,0.0, 1.0, 0.0, 0.0>>)
			l_test_observations.add_row(l_vector)
			-- this next statement causes a problem for future tests
			add_observation (l_test_dependent, l_test_observations, 37.89)
			print(statistics)
			create l_observations.make_from_array(<<
				<<-1,-1>>,
				<<-1,1>>,
				<<1,-1>>,
				<<1,1>>,
				<<0,3>>>>)
			create l_dependent.make_from_array(<<10.5,11.3,12.6,13.4,17.1>>)
			make_weighted(l_observations,l_dependent,l_weights)
			create validation.make(Current,"No intercept weighted",snapshot)
			print(statistics)
			create l_observations.make_from_array(<<
				<<1,-1,-1>>,
				<<1,-1,1>>,
				<<1,1,-1>>,
				<<1,1,1>>,
				<<1,0,3>>>>)
			make_weighted(l_observations,l_dependent,l_weights)
			print(statistics)
		end


	test_compact is
		-- This test feature is not complete.  It calls for regressions to be done, but does not fills in things
		-- like xty and yty and thus the results are not meaningful.
		local
			l_dependent : VECTOR
			l_observations : MATRIX
		do

			create l_observations.make_from_array (<<
								<<1.0,1.0, 2.0, 0.0, 0.0>>,
								<<1.0,0.0, 2.0, 0.0, 1.0>>,
								<<1.0,0.0, 1.0, 3.0, 0.0>>,
								<<1.0,0.0, 0.0, 1.0, 0.0>>,
								<<1.0,0.0, 0.0, 3.0, 1.0>>,
								<<1.0,0.0, 1.0, 0.0, 4.0>>,
								<<1.0,0.0, 4.0, 0.0, 1.0>>,
								<<1.0,1.0, 0.0, 0.0, 0.0>>,
								<<1.0,0.0, 0.0, 1.0, 0.0>>,
								<<1.0,1.0, 0.0, 0.0, 0.0>>,
								<<1.0,0.0, 0.0, 1.0, 0.0>>,
								<<1.0,0.0, 1.0, 0.0, 0.0>>
								>>)
			create l_dependent.make_from_array(<<33.63,39.62,38.18,41.46,38.02,35.83,35.99,36.58,42.92,37.80,40.43,37.89>>)

			make_compact(l_observations.width, 3)
			regress_compact
		end

	snapshot : ROSE_DOUBLE_SNAPSHOT is
			-- snapshot
		do
			create Result.make(<<
									["R Square", r_square_with_no_intercept],
									["Adjusted R Square", adjusted_r_square_with_no_intercept],
									["Sum of squares", chi_square],
									["Coefficient 1", coefficients.item(1)],
									["Standard Error 1", standard_errors.item(1)],
									["t-statistic 1", t_statistic(1)],
									["Coefficient 2", coefficients.item(2)],
									["Standard Error 2", standard_errors.item(2)],
									["t-statistic 2", t_statistic(2)]
									>>,
									agent equal_double_with_tolerance(?, ?, .0001))
		end

	validation: VALIDATION


feature -- prediction
	predicted_value (a_independent_vector : ABSTRACT_READABLE_VECTOR) : DOUBLE is
			-- predicted value
		do
			Result := coefficients * a_independent_vector
		end

	prediction_standard_error (a_independent_vector : ABSTRACT_READABLE_VECTOR) : DOUBLE is
			-- prediction standard error
		local
			l_value : DOUBLE
		do
			l_value := (xtx_inverse @ a_independent_vector) * a_independent_vector
			l_value := l_value + 1
			Result := sqrt (l_value * sigma_square )
		end

	predicted_value_with_interval (a_independent_vector : ABSTRACT_READABLE_VECTOR; a_confidence_t_value : DOUBLE) : VECTOR is
			-- predicted value with interval
		local
			l_value, l_standard_error : DOUBLE
		do
			l_value := predicted_value (a_independent_vector)
			l_standard_error := prediction_standard_error (a_independent_vector)
			create Result.make_from_array(<<l_value, l_value - a_confidence_t_value * l_standard_error, l_value + a_confidence_t_value * l_standard_error>>)
		end

feature {NONE} -- Implementation

	correct_mismatch is
			-- correct mismatch
		local
			l_mismatch: MISMATCH_INFORMATION
		do
			l_mismatch := mismatch_information
		end

feature --Assertion help
	is_updatable : BOOLEAN = True
	has_statistics : BOOLEAN = True
	has_residual_statistics : BOOLEAN = True
	is_valid : BOOLEAN = True

end


