note

	description:

		"Base class for all types of regression"

	author: "pn"
	last_editor: "$Author: mk $"
	date: "$Date: 2012/08/20 $"
	revision: "$Revision: #2 $"
	path: "$File: //rose/source/kernel/infrastructure/math/regression_base.e $"

deferred class
	REGRESSION_BASE

feature -- Initialization

	make (a_matrix: ABSTRACT_READABLE_MATRIX; a_dependent: EXTENDABLE_VECTOR) is
			-- Initialize `Current' then run regression of `a_dependent' against `a_matrix' (as explanatory variables).
		require
			a_matrix_not_void: a_matrix /= Void
			a_dependent_not_void: a_dependent /= Void
			explanatory_matches_dependent: a_matrix.height = a_dependent.count
			enough_observations: a_matrix.height >= a_matrix.width
		deferred end

	make_weighted (a_matrix: ABSTRACT_READABLE_MATRIX; a_dependent, a_weights: EXTENDABLE_VECTOR) is
			-- Initialize `Current' then run regression weighted by `a_weights' of `a_dependent' against `a_matrix' (as explanatory variables).
		require
			a_matrix_not_void: a_matrix /= Void
			a_dependent_not_void: a_dependent /= Void
			explanatory_matches_dependent: a_matrix.height = a_dependent.count
			enough_observations: a_matrix.height >= a_matrix.width
			valid_weights : a_weights /= Void implies (a_weights.non_negative and then a_matrix.height = a_weights.count)
		deferred end

	make_compact (a_nvar, a_ndep: INTEGER) is
			-- Initialize `Current' in compact form.
		require
			a_nvar_non_negative: a_nvar >= 0
			a_ndep_non_negative: a_ndep >= 0
		deferred end

	make_weighted_with_intercept (a_matrix: MATRIX; a_dependent, a_weights: EXTENDABLE_VECTOR) is
			-- Initialize `Current' then run regression weighted by `a_weights' of `a_dependent' against `a_matrix' (as explanatory variables) and produce an intercept.
		require
			a_matrix_not_void: a_matrix /= Void
			a_dependent_not_void: a_dependent /= Void
			explanatory_matches_dependent: a_matrix.height = a_dependent.count
			enough_observations: a_matrix.height >= a_matrix.width
			a_weights_not_void: a_weights /= Void
			all_weights_positive: a_weights.non_negative
			valid_dimension: a_matrix.height = a_weights.count
		deferred end


feature -- Operations

	calculate_residual_statistics is
			-- Calculate statistics on residuals.

			-- The formulas used for this can be found in the SAS documentation or on Wikipedia.
			-- Note different sources have different formulas for external studentized residual,
			-- so there are different formulas given below.  In practice there is little differnce
			-- between the two values, and also between internal and external values.
		deferred end

	add_observation (a_dependent_variables: VECTOR; a_new_matrix: ABSTRACT_READABLE_MATRIX; a_dependent_value: DOUBLE) is
			-- Add observation from last row of `a_new_matrix' to normal representation and re-compute regression results.
		require
			is_updatable: is_updatable
			a_new_matrix_not_void: a_new_matrix /= Void
			a_new_matrix_correct_width: a_new_matrix.width = number_of_independent_variables
			a_dependent_variables_not_void: a_dependent_variables /= Void
		deferred end

	update_variable (a_column: INTEGER; a_matrix: ABSTRACT_READABLE_MATRIX) is
			-- Update variable from a_matrix.
		require
			a_matrix_not_void : a_matrix /= Void
			legal_variable: a_column >= 1 and a_column <= a_matrix.width
			is_updatable: is_updatable
		deferred end


feature -- Access

	adjusted_r_square_with_no_intercept: DOUBLE is
			-- Coefficient of determination adjusted for number of explanatory terms in model without intercept
		deferred end

	t_statistic (i: INTEGER): DOUBLE is
			-- Student's T for `i'th coefficient
		require
			-- this next line needs some logic/conditional invariant backup
			valid_variable: 1 <= i and i <= number_of_independent_variables
			standard_errors_not_void: standard_errors /= Void
			coefficients_not_void: coefficients /= Void
		deferred end
		chi_square: DOUBLE is
			-- Sum of squared errors
		deferred end

	is_updatable: BOOLEAN is
			-- 	Is this updatable?
		deferred
		end


	has_statistics: BOOLEAN is
			-- 	Does the regression have statistics
		deferred end

	statistics: STRING is
			-- Formatted output of the statistics from the regression
		require
			has_statistics: has_statistics
		deferred end

	has_residual_statistics: BOOLEAN is
			-- Does the regression have residual statistics
		deferred end

	residual_statistics: STRING is
			-- Formatted output of residuals, standard errors of residuals, and studentized residuals;
			-- The studentized residuals returned are "internally" studentized, since that is what SAS returns.

		require
			has_residual_statistics: has_residual_statistics
		deferred end

	r_square_with_intercept: DOUBLE
			-- Coefficient of determination;
			-- SAS computes R-square differently when one assumes an intercept.  We follow the conventions of SAS.
		deferred end

	r_square_with_no_intercept: DOUBLE
			-- Coefficient of determination when no intercept is modeled;
			-- NAG apparently uses this even when there is an intercept.
		deferred end

	r_square: DOUBLE is
			-- Coefficient of determination
		deferred end

	adjusted_r_square: DOUBLE is
			-- Coefficient of determination adjusted for number of explanatory terms in model
		deferred end

feature -- Results

	coefficients: VECTOR
			-- Estimates of coefficients of explanatory variables
		deferred end

	covariance: ABSTRACT_MATRIX
			-- Covariance of explanatory variables
		deferred end

	standard_errors: VECTOR
			-- Standard errors of `coefficients'
		deferred end

	sigma_square: DOUBLE
			-- Variance
		deferred end

	modified_sigma_square: DOUBLE
			-- Variance ignoring zero weight observations
		deferred end

	y_total_variation: DOUBLE
			-- sum square of deviation from y mean
		deferred end

	residual_standard_errors: VECTOR
			-- Standard errors of unweighted residuals for `studentized_residuals'
		deferred end

	studentized_residuals: VECTOR
			-- Internally studentized (unweighted) residuals
		deferred end

	external_residual_standard_errors: VECTOR
			-- Standard errors of unweighted residuals for `external_studentized_residuals'
		deferred end

	external_studentized_residuals: VECTOR
			-- Externally studentized (unweighted) residuals
		deferred end

	number_of_observations: INTEGER
			-- Number of observations used
		deferred end

	number_of_independent_variables: INTEGER
			-- Number of explanatory variables
		deferred end

	number_of_dependent_variables: INTEGER
			-- Number of response variables
		deferred end

	degree_of_freedom: INTEGER
			-- Number of observations minus number of parameters estimated
		deferred end

	modified_degree_of_freedom: INTEGER
			-- `degree_of_freedom' minus zero weight observation count
		deferred end

	sum_of_squared_errors: DOUBLE is
			-- Sum of squared errors
		deferred end
			-- Numerator of variance

	intercept_matrix(a_matrix : ABSTRACT_READABLE_MATRIX) : MATRIX
			--  Matrix with first column 1 for intercept
		local
			i,j : INTEGER
		do
			create Result.make(a_matrix.height,a_matrix.width + 1)
			from i := 1 until i > Result.height loop
				Result [i,1] := 1
				from j := 1 until j > a_matrix.width loop
					Result[i,j+1] := a_matrix[i,j]
					j := j + 1
				end
				i := i + 1
			end
		end
feature -- Access

	is_valid: BOOLEAN is
			-- Is the regression valid?
		deferred end
			-- Is the current regression valid?



end
