note

	description:

		"Entity Relationship Handler"

	documentation: ""
	author: "Mark Howard"
	last_editor: "$Author: rj $"
	date: "$Date: 2014/07/22 $"
	revision: "$Revision: #33 $"
	path: "$File: //rose/source/kernel/entities/reference/relationship_handler.e $"

class
	RELATIONSHIP_HANDLER [reference G->ENTITY, reference H->ENTITY]

inherit
	NEEDS_HANDLE

	SUPPLIER_ACCESS
		export
			{NONE} all
		end

	DATABASE_SERVER
		export
			{NONE} all
		end

	CONTEXT_ANCHOR

create
	make

feature {NONE} -- Initialization

	make (a_context: like context_anchor; a_relationship_id: INTEGER) is
			-- Initialize `Current'.
		local
			l_record: RELATIONSHIP_TYPE_RECORD
		do
			context := a_context
			l_record := context.relationship_record (a_relationship_id)
			if l_record /= void then
				name := l_record.name
				category_1_id := l_record.category_1_id
				category_2_id := l_record.category_2_id
				g_handler ?= context.handler(category_1_id)
				h_handler ?= context.handler(category_2_id)
				if g_handler /= void and then h_handler /= void then
					ur_g := g_handler.entity(0)
					ur_h := h_handler.entity(0)
				end
				relationship_id := a_relationship_id
			end
			create related.make(1000)
			create relationship_record_cache.make(1000)
			create corelated.make(1000)
			create corelationship_record_cache.make(1000)
		end

feature -- Access

	context_name: STRING is
			-- Context for which `Current' handles relationships
		do
			Result := context.name
		end

	name: STRING
			-- Relationship name

	relationship_id : INTEGER
			-- Relationship identifier

	g_handler : CATEGORY_HANDLER[like context_anchor,G]
			-- Category handler for `G' (source entities)

	h_handler : CATEGORY_HANDLER[like context_anchor,H]
			-- Category handler for `H' (destination entities)

	category_1_id : INTEGER
			-- Relationship is FROM this category

	category_2_id : INTEGER
			-- Relationship is TO this category

	relationship : RELATIONSHIP_TYPE_RECORD
			-- Definition of relationship from database

	ur_g : G
			-- Ur (prototype)  source entity

	ur_h : H
			-- Ur (prototype) destination entity

feature {NONE} -- Implementation

	related : ROSE_HASH_TABLE[ROSE_ENTITY_HISTORY[H],INTEGER]
			-- Cache of known relations from G -> H indexed by entity_1_id.

	relationship_record_cache: ROSE_HASH_TABLE[ROSE_LINEAR[ENTITY_RELATIONSHIP_RECORD], INTEGER]
			-- Cache of raw relationship records indexed by entity_1_id.

	corelated : ROSE_HASH_TABLE[ROSE_ENTITY_HISTORY[G],INTEGER]
			-- Inverse of the main relation (H <- G) indexed by entity_2_id.

	corelationship_record_cache: ROSE_HASH_TABLE[ROSE_LINEAR[ENTITY_RELATIONSHIP_RECORD], INTEGER]
			-- Cache of raw corelationship records indexed by entity_2_id.

	rose_constants : ROSE_CONSTANT_VALUES is
			-- For undefined double
		once
			create Result
		end

feature -- Query

	related_entities (a_entity_1 : G) : ROSE_LINEAR_ARRAY[H] is
			-- All entities with a relation _from_ `a_entity_1'
		local
			l_array : ROSE_ENTITY_HISTORY[H]
		do
			l_array := related_entity_intervals(a_entity_1)
			create Result.make_empty
			from l_array.start until l_array.off loop
				Result.extend(l_array.current_item.entity)
				l_array.forth
			end
		end

	corelated_entities (a_entity_2 : H) : ROSE_LINEAR_ARRAY[G] is
			-- All entities that are the source of a relation _to_ `a_entity_2'
		local
			l_array : ROSE_ENTITY_HISTORY[G]
		do
			l_array := corelated_entity_intervals(a_entity_2)
			create Result.make_empty
			from l_array.start until l_array.off loop
				Result.extend(l_array.current_item.entity)
				l_array.forth
			end
		end

	related_entities_at (a_entity_1 : G;a_date : ROSE_DATE) : ROSE_LINEAR_ARRAY[H] is
			-- All related entities with a relation to _from_ `a_entity_1' as of `a_date'
		local
			l_array : ROSE_ENTITY_HISTORY[H]
		do
			l_array := related_entity_intervals(a_entity_1)
			create Result.make_empty
			from l_array.start until l_array.off loop
				if l_array.current_item.brackets(a_date) then
					Result.extend(l_array.current_item.entity)
				end
				l_array.forth
			end
		end

	corelated_entities_at (a_entity_2 : H;a_date : ROSE_DATE) : ROSE_LINEAR_ARRAY[G] is
			-- All entities that are the source of a relation
			-- _to_ `a_entity_2' as of `a_date'
		local
			l_array : ROSE_ENTITY_HISTORY[G]
		do
			l_array := corelated_entity_intervals(a_entity_2)
			create Result.make_empty
			from l_array.start until l_array.off loop
				if l_array.current_item.brackets(a_date) then
					Result.extend(l_array.current_item.entity)
				end
				l_array.forth
			end
		end

	related_entity_intervals (a_entity_1 : G) : ROSE_ENTITY_HISTORY[H] is
			-- History of relations _from_ `a_entity_1'
		local
			l_array : ROSE_LINEAR[ENTITY_RELATIONSHIP_RECORD]
			l_entity : H
			l_weight: DOUBLE
		do
			Result := related.item(a_entity_1.handle)
			if Result = Void and not context.is_ready then
				create Result.make_empty
				l_array := relationship_record_cache[a_entity_1.handle]

				if l_array = Void
				then
					l_array := context.relationship_records(Current,a_entity_1)
				end

				from l_array.start until l_array.off loop
					l_entity := ur_h.entity_at(l_array.item.entity_2_id)
					if l_entity /= void then
						if rose_constants.is_missing_double(l_array.item.weight) then
							l_weight:=0
						else
							l_weight:=l_array.item.weight
						end
						Result.extend(create {ENTITY_INTERVAL[H]}.make(l_entity, l_array.item.start_date, l_array.item.end_date, l_weight))
					end
					l_array.forth
				end
				related.put(Result,a_entity_1.handle)
			end
			if Result = Void then
				create Result.make_empty
			end
		end

	corelated_entity_intervals (a_entity_2 : H) : ROSE_ENTITY_HISTORY[G] is
			-- History of relations _to_ `a_entity_2'
		local
			l_array : ROSE_LINEAR[ENTITY_RELATIONSHIP_RECORD]
			l_entity : G
			l_weight : DOUBLE
		do
			Result := corelated.item(a_entity_2.handle)
			if Result = Void and not context.is_ready then
				create Result.make_empty
				l_array := corelationship_record_cache[a_entity_2.handle]

				if l_array = Void
				then
					l_array := context.corelationship_records(Current,a_entity_2)
				end

				from l_array.start until l_array.off loop
					l_entity := ur_g.entity_at(l_array.item.entity_1_id)
					if l_entity /= void then
						if rose_constants.is_missing_double(l_array.item.weight) then
							l_weight:=0
						else
							l_weight:=l_array.item.weight
						end
						Result.extend(create {ENTITY_INTERVAL[G]}.make(l_entity,l_array.item.start_date,l_array.item.end_date,l_weight))
					end
					l_array.forth
				end
				corelated.put(Result,a_entity_2.handle)
			end
			if Result = Void then
				create Result.make_empty
			end
		end

	related_entity_intervals_at (a_entity_1 : G;a_date : ROSE_DATE) : ROSE_ENTITY_HISTORY[H] is
			-- Relations _from_ `a_entity_1' as of `a_date' with interval information
		local
			l_array : ROSE_ENTITY_HISTORY[H]
		do
			l_array := related_entity_intervals(a_entity_1)
			create Result.make_empty
			from l_array.start until l_array.off loop
				if l_array.current_item.brackets(a_date) then
					Result.extend(l_array.current_item)
				end
				l_array.forth
			end
		end

	corelated_entity_intervals_at (a_entity_2 : H;a_date : ROSE_DATE) : ROSE_ENTITY_HISTORY[G] is
			-- Relations _to_ `a_entity_2' as of `a_date' with interval information
		local
			l_array : ROSE_ENTITY_HISTORY[G]
		do
			l_array := corelated_entity_intervals(a_entity_2)
			create Result.make_empty
			from l_array.start until l_array.off loop
				if l_array.current_item.brackets(a_date) then
					Result.extend(l_array.current_item)
				end
				l_array.forth
			end
		end

	intervals_related_to_entities (a_entities : LINEAR[G]) : ROSE_ENTITY_HISTORY[H] is
			-- Entity intervals (entity id, start, end, weight) of entities with a relationship with entities in 'a_entities'
		local
			l_array : ROSE_LINEAR[ENTITY_RELATIONSHIP_RECORD]
			l_entity : H
		do
			create Result.make_empty
			l_array := context.multiple_relationship_records(a_entities)
			from l_array.start until l_array.off loop
				l_entity := ur_h.entity_at(l_array.item.entity_2_id)
				if l_entity /= void then
					Result.extend(create {ENTITY_INTERVAL[H]}.make(l_entity,l_array.item.start_date,l_array.item.end_date,l_array.item.weight))
				end
				l_array.forth
			end
		end

	intervals_corelated_to_entities (a_entities : LINEAR[H]) : ROSE_ENTITY_HISTORY[G] is
			--Intervals which are the first coordinate of a pair in this relationship whose second coordinate is in a_entities
		local
			l_array : ROSE_LINEAR[ENTITY_RELATIONSHIP_RECORD]
			l_entity : G
		do
			create Result.make_empty
			l_array := context.multiple_corelationship_records(a_entities)
			from l_array.start until l_array.off loop
				l_entity := ur_g.entity_at(l_array.item.entity_1_id)
				if l_entity /= void then
					Result.extend(create {ENTITY_INTERVAL[G]}.make(l_entity,l_array.item.start_date,l_array.item.end_date,l_array.item.weight))
				end
				l_array.forth
			end
		end

	related_entity (a_entity_1 : G) : H is
			--Some entity related to a_entity_1
		local
			l_entities : ROSE_ENTITY_HISTORY[H]
		do
			l_entities := related_entity_intervals(a_entity_1)
			if not l_entities.is_empty then
				Result := l_entities.item(1).entity
			end
		end

	corelated_entity (a_entity_2 : H) : G is
			-- First corelated entity to `a_entity_2'
		local
			l_entities : ROSE_ENTITY_HISTORY[G]
		do
			l_entities := corelated_entity_intervals(a_entity_2)
			if not l_entities.is_empty then
				Result := l_entities.item(1).entity
			end
		end

feature -- Caching
	preload
			--Cache records
		local
			l_records : ROSE_LINEAR_ARRAY[ENTITY_RELATIONSHIP_RECORD]
		do
			l_records := context.all_relationship_records (Current)
		end
	has_cache: BOOLEAN is
			-- Are some relationships cached in memory?
		do
			Result := not related.is_empty or not corelated.is_empty
		end

	clear_relationship_and_corelationship
			--Reset related and corelated
		do
			related.wipe_out
			corelated.wipe_out
		end

	load_relationship_and_corelationship is
			-- Load relationship and corelationship for all entities at both ends
			-- of relationship.
		local
			l_relationships: ROSE_LINEAR_ARRAY [ENTITY_RELATIONSHIP_RECORD]
		do
			if relationship_record_cache.is_empty
			then
				l_relationships := context.all_relationship_records (Current)	-- Already sorted by entity_1_id (and end_date but that doesn't matter).
				l_relationships.do_all(agent {ENTITY_RELATIONSHIP_RECORD}.set_relationship_type_id(relationship_id))

				load_relationship_record_cache(relationship_record_cache, l_relationships, agent {ENTITY_RELATIONSHIP_RECORD}.entity_1_id)

				l_relationships.sort_by (agent context.compare_rel_records_for_entity_2)
				load_relationship_record_cache(corelationship_record_cache, l_relationships, agent {ENTITY_RELATIONSHIP_RECORD}.entity_2_id)
			end
		end

	load_relationship_record_cache(a_cache: like relationship_record_cache;
								   a_relationship_records: ROSE_LINEAR_ARRAY[ENTITY_RELATIONSHIP_RECORD];
								   a_entity_id_agent: FUNCTION[ANY, TUPLE[ENTITY_RELATIONSHIP_RECORD], INTEGER])
	-- Breaks apart 'a_raw_relationship_records' into snippets everytime the 'entity_id' changes and stores it into 'a_cache' (keyed by 'entity_id').
	-- 'entity_id' is determined by the agent.
	local
		l_array: ROSE_LINEAR_ARRAY[ENTITY_RELATIONSHIP_RECORD]
		l_relationship_record: ENTITY_RELATIONSHIP_RECORD
		l_entity_id: INTEGER
		l_old_entity_id: INTEGER
	do
		l_old_entity_id := -1

		across
			a_relationship_records as c_relationship_records
		loop
			l_relationship_record := c_relationship_records.item
			l_entity_id := a_entity_id_agent.item([l_relationship_record])
			if l_entity_id /= l_old_entity_id
			then
				if l_old_entity_id /= -1
				then
					a_cache[l_old_entity_id] := l_array
				end

				create l_array.make_empty
				l_old_entity_id := l_entity_id
			end

			l_array.extend(l_relationship_record)
		end

		a_cache[l_old_entity_id] := l_array
	end

	load_relationships_fast (a_relationships: ROSE_LINEAR_ARRAY[ENTITY_RELATIONSHIP_RECORD]) is
			-- load relationships from `a_relationships'
			-- `a_relationships' should be sorted with respect to relationship_id and entity_1_id
		require
			valid_relationships: a_relationships /= Void
			valid_intervall: a_relationships.for_all (agent (a_record: ENTITY_RELATIONSHIP_RECORD): BOOLEAN do Result := (a_record.relationship_type_id = relationship_id) end)
		local
			l_h_entity: H
			l_g_entity: G
			l_g_id, i: INTEGER
			l_weight: DOUBLE
			l_history: ROSE_ENTITY_HISTORY[H]
		do
			related.accommodate (context.count_relationships_from_relationship (current))
			-- loop trough all DB records
			from
				i := a_relationships.lower
				create l_history.make_empty
				l_g_id := a_relationships.item(i).entity_1_id
				l_g_entity := ur_g.handler.entity_from_table (a_relationships.item(i).entity_1_id)
			until
				i > a_relationships.upper
			loop
				if l_g_id /= a_relationships.item(i).entity_1_id then
					-- new entity_1, put history for old entity_1 into hashtable
					if l_g_entity /= Void then
						related.put (l_history, l_g_id)
					end
					-- new entity_1 and new history for entity_1
					l_g_id := a_relationships.item(i).entity_1_id
					l_g_entity := ur_g.handler.entity_from_table (a_relationships.item(i).entity_1_id)
					create l_history.make_empty
				end
				if l_g_entity /= Void then
					-- sometimes there is no entity to this relationship
					l_h_entity := ur_h.handler.entity_from_table (a_relationships.item(i).entity_2_id)
					if rose_constants.is_missing_double(a_relationships.item(i).weight) then
						l_weight := 0
					else
						l_weight := a_relationships.item(i).weight
					end
					l_history.extend (create {ENTITY_INTERVAL[H]}.make (l_h_entity, a_relationships.item(i).start_date, a_relationships.item(i).end_date, l_weight))
				end
				i := i + 1
			end
		end

	load_corelationships_fast (a_relationships: ROSE_LINEAR_ARRAY[ENTITY_RELATIONSHIP_RECORD]) is
			-- load corelationships from `a_relationships'
			-- `a_relationships' should be sorted with respect to relationship_id and entity_2_id
		require
			valid_relationships: a_relationships /= Void
			valid_intervall: a_relationships.for_all (agent (a_record: ENTITY_RELATIONSHIP_RECORD): BOOLEAN do Result := (a_record.relationship_type_id = relationship_id) end)
		local
			l_h_entity: H
			l_g_entity: G
			l_h_id, i: INTEGER
			l_weight: DOUBLE
			l_history: ROSE_ENTITY_HISTORY[G]
		do

			corelated.accommodate (context.count_corelationships_from_relationship (Current))
			-- loop trough all DB records
			from
				i := a_relationships.lower
				create l_history.make_empty
				l_h_id := a_relationships.item(i).entity_2_id
				l_h_entity := ur_h.handler.entity_from_table (a_relationships.item(i).entity_2_id)
			until
				i > a_relationships.upper
			loop
				if l_h_id /= a_relationships.item(i).entity_2_id then
					-- new entity_2
					if l_h_entity /= Void then
						corelated.put (l_history, l_h_id)
					end
					create l_history.make_empty
					l_h_id := a_relationships.item(i).entity_2_id
					l_h_entity := ur_h.handler.entity_from_table (a_relationships.item(i).entity_2_id)
				end
				if l_h_entity /= Void then
					-- sometime there exist no entity to a relationship
					l_g_entity := ur_g.handler.entity_from_table (a_relationships.item(i).entity_1_id)
					if rose_constants.is_missing_double(a_relationships.item(i).weight) then
						l_weight := 0
					else
						l_weight := a_relationships.item(i).weight
					end
					l_history.extend (create {ENTITY_INTERVAL[G]}.make (l_g_entity, a_relationships.item(i).start_date, a_relationships.item(i).end_date, l_weight))
				end
				i := i + 1
			end
		end

feature {NONE} -- Caching implementation

	prepare_for_load_relationship_fast (a_record: ENTITY_RELATIONSHIP_RECORD) is
			-- Force population of entity cache that `load_[co]relationships_fast' rely
			-- on but could be removed if it was to call `entity' instead of `entity_from_table'
			-- (which should cost no more)
		require
			a_record_not_void: a_record /= Void
		local
			l_dummy: ENTITY
		do
			a_record.set_relationship_type_id (relationship_id)
			l_dummy := g_handler.entity (a_record.entity_1_id)
			l_dummy := h_handler.entity (a_record.entity_2_id)
		end

feature -- Update

	update_relationship (a_relationship: ENTITY_RELATIONSHIP_RECORD) is
			-- update this relationship
		local
			l_array: ROSE_LINEAR[ENTITY_RELATIONSHIP_RECORD]
			l_entity_2: H
			l_entity_1: G
			l_weight: DOUBLE
			l_history: ROSE_ENTITY_HISTORY[H]
		do
			-- remove old history
			related.remove (a_relationship.entity_1_id)
			-- insert new history
			l_entity_1 := ur_g.entity_at(a_relationship.entity_1_id)
			if l_entity_1 /= Void then
				l_array := context.relationship_records(Current,l_entity_1)
				create l_history.make_empty
				from
					l_array.start
				until
					l_array.off
				loop
					l_entity_2 := ur_h.entity_at(l_array.item.entity_2_id)
					if l_entity_2 /= void then
						if rose_constants.is_missing_double(l_array.item.weight) then
							l_weight:=0
						else
							l_weight:=l_array.item.weight
						end
						l_history.extend(create {ENTITY_INTERVAL[H]}.make(l_entity_2,l_array.item.start_date,l_array.item.end_date,l_weight))
					end
					l_array.forth
				end
				related.put(l_history,l_entity_1.handle)
			end
		end

	update_corelationship (a_relationship: ENTITY_RELATIONSHIP_RECORD) is
			-- update this relationship
		local
			l_array : ROSE_LINEAR[ENTITY_RELATIONSHIP_RECORD]
			l_entity_2: H
			l_entity_1: G
			l_weight: DOUBLE
			l_history: ROSE_ENTITY_HISTORY[G]
		do
			-- remove old history
			corelated.remove (a_relationship.entity_2_id)
			-- insert new history
			l_entity_2 := ur_h.entity_at(a_relationship.entity_2_id)
			if l_entity_2 /= Void then
				l_array := context.corelationship_records (Current,l_entity_2)
				create l_history.make_empty

				from
					l_array.start
				until
					l_array.off
				loop
					l_entity_1 := ur_g.entity_at(l_array.item.entity_1_id)
					if l_entity_1 /= void then
						if rose_constants.is_missing_double(l_array.item.weight) then
							l_weight:=0
						else
							l_weight:=l_array.item.weight
						end
						l_history.extend(create {ENTITY_INTERVAL[G]}.make(l_entity_1,l_array.item.start_date,l_array.item.end_date,l_weight))
					end
					l_array.forth
				end
				corelated.put(l_history,l_entity_2.handle)
			end
		end

invariant

	related_not_void: related /= Void
	correlated_not_void: corelated /= Void

end
