note

	description:

		"AxaRosenberg descendent of RAW_FILE with platform specific optimizations"

	author: "Mark Howard"
	last_editor: "$Author: e2 $"
	date: "$Date: 2011/05/11 $"
	revision: "$Revision: #27 $"
	path: "$File: //rose/source/kernel/infrastructure/support/riem_file.e $"

class RIEM_FILE

inherit
	ROSE_COMMON_FILE

	RAW_FILE
		undefine
--			readstream,
--			read_stream,
			exists
		redefine
			open_read,
			open_write,
			open_append,
			open_read_write,
			create_read_write,
			open_read_append,
			exists
		end

	FILE_ACCESS_LOG

create

	make,
	make_open_read,
	make_open_write,
	make_open_append,
	make_open_read_write,
	make_create_read_write,
	make_open_read_append

feature -- Initialization

	open_read is
			-- Open file in read-only mode.
		do
			Precursor
			record_file_read_access (name)
		rescue
			error.signal (generator, "Can not open_read file '$1'", <<name>>)
		end

	open_write is
			-- Open file in write-only mode;
			-- create it if it does not exist.
		do
			Precursor
			record_file_write_access (name)
		rescue
			error.signal (generator, "Can not open_write file '$1'", <<name>>)
		end

	open_append is
			-- Open file in append-only mode;
			-- create it if it does not exist.
		do
			Precursor
			record_file_write_access (name)
		rescue
			error.signal (generator, "Can not open_append file '$1'", <<name>>)
		end

	open_read_write is
			-- Open file in read and write mode.
		do
			Precursor
			record_file_write_access (name)
		rescue
			error.signal (generator, "Can not open_read_write file '$1'", <<name>>)
		end

	create_read_write is
			-- Open file in read and write mode;
			-- create it if it does not exist.
		do
			Precursor
			record_file_write_access (name)
		rescue
			error.signal (generator, "Can not create_read_write file '$1'", <<name>>)
		end

	open_read_append is
			-- Open file in read and write-at-end mode;
			-- create it if it does not exist.
		do
			Precursor
			record_file_write_access (name)
		rescue
			error.signal (generator, "Can not open_read_append file '$1'", <<name>>)
		end

feature -- Access

	exists: BOOLEAN is
			-- Does this file exist?
		do
			Result := not is_closed or else Precursor
		end

feature -- Basic operations

	reopen is
			-- Closes and reopens the file at the current position, to deal with a "feature" of NT
		local
			l_position: INTEGER
			l_mode: INTEGER
		do
			l_position := position
			l_mode := mode
			close
			inspect l_mode
			when Read_write_file then
				make_open_read_write (name)
			when Read_file then
				make_open_read (name)
			end
			go (l_position)
		rescue
			error.signal (generator, "Can not reopen file '$1'", <<name>>)
		end

	read_record (a_record: EXTERNAL_RECORD) is
			-- Reads a_record.record_length_ bytes from the file and the uses that data to populate the
			-- attributes of this class.
		do
			a_record.set (c_read_record (file_pointer, a_record.record_length_))
			a_record.read
		rescue
			error.signal (generator, "Can not read_record from file '$1'", <<name>>)
		end

	write_record (a_record: EXTERNAL_RECORD) is
			-- write record
		local
			l_pointer: POINTER
			l_status: INTEGER
		do
			l_pointer := c_file_buffer (a_record.record_length_)
			a_record.set (l_pointer)
			a_record.write
			l_status := fwrite (l_pointer, 1, a_record.record_length_, file_pointer)
		rescue
			error.signal (generator, "Can not write_record to file '$1'", <<name>>)
		end

	write_stream, writestream (a_string: STRING) is
			-- write stream
		require
			valid_string: a_string /= Void
		local
			l_c_string: C_STRING
			l_status: INTEGER
		do
			create l_c_string.make (a_string)
			l_status := fwrite (l_c_string.item, 1, a_string.count, file_pointer)
		rescue
			error.signal (generator, "Can not write_stream to file '$1'", <<name>>)
		end

	set_buffer_size (a_size: INTEGER) is
			-- set the buffer size
		local
			l_pointer: POINTER
			l_status: INTEGER
		do
			l_status := setvbuf (file_pointer, l_pointer, file_iofbf, a_size)
		end

feature {NONE} -- Implementation

	local_name (a_name: STRING): STRING is
			-- get local name
		local
			l_separator: CHARACTER
			l_index: INTEGER
		do
			Result := a_name.twin
			l_separator := operating_environment.directory_separator
			if l_separator /= '\' then
				Result.replace_substring_all ("\", "/")
					--strip unc
				if Result.substring_index ("//", 1) = 1 then
					Result.keep_tail (Result.count - 2)
					l_index := Result.index_of ('/', 1)
					if l_index > 0 then
						Result.keep_tail (Result.count - l_index + 1)
					end
				end
			else
				if Result.substring_index ("\\", 1) /= 1 and then Result.substring_index ("//", 1) /= 1 then
					Result.replace_substring_all ("\", "/")
					if Result.item (1) = '/' then
						Result.prepend ("//pinegv")
					end
				end
			end
		end

	c_read_record (a_file: POINTER; a_size: INTEGER): POINTER is
			-- read record
		external
			"C"
		alias
			"read_record"
		end

	c_file_buffer (a_length: INTEGER): POINTER is
			-- get file buffer
		external
			"C"
		alias
			"file_buffer"
		end

	fwrite (a_string: POINTER; a_size: INTEGER; a_count: INTEGER; a_file: POINTER): INTEGER is
			-- write `a_string'
		external
			"C [macro %"stdio.h%"]"
		end

	file_iofbf: INTEGER is
			-- some C constant
		external
			"C [macro <stdio.h>]"
		alias
			"_IOFBF"
		end

	setvbuf (a_file: POINTER; a_buffer: POINTER; a_mode: INTEGER; a_size: INTEGER): INTEGER is
			-- set file buffer
		external
			"C [macro <stdio.h>] (FILE*,char *,int,int): int"
		end

	rose_create_last_string (a_min_size: INTEGER) is
			-- Create new instance of `last_string' with a least `a_min_size'
			-- as capacity.
		do
			create last_string.make (rose_default_last_string_size.max (a_min_size))
		end

end
