note
	description: "Application information"
	author: "Zoran Simic"
	last_editor: "$Author: rj $"
	date: "$Date: 2015/01/09 $"
	revision: "$Revision: #96 $"
	path: "$File: //rose/source/kernel/infrastructure/framework/application/rose_application.e $"

class
	ROSE_APPLICATION

inherit
	ROSE_FRAMEWORK_CONVENTION
		export
			{NONE} all;
			{ANY} formatter
		redefine
			default_create, out
		end
	DEBUG_OUTPUT
		redefine
			default_create, out
		end

feature {NONE} -- Initialization

	default_create
			-- Default creation procedure, must satisfy all attached attributes
		do
			create config.make_default
			create start_time.make_now
			create platform
			create delivery.make_default
			create site.make_default
			create implicit_commands.make
			create variables.make
			create program_arguments.make_empty
		end

feature {ROSE_GLOBAL} -- Initialization

	make_current
			-- Identify current application running.
		local
			l_dll_path, l_path, l_config_path: STRING
			i: INTEGER
		do
				-- Environment information
			host_name := env.host_name
			if host_name = Void or else host_name.is_empty then
				host_name := "localhost"
			end
			host_name.to_upper
			pid := env.current_process_id
			user := env.current_user
			program_path := env.executable_path
			program_file := formatter.file_name (program_path)
			program_name := formatter.file_name_no_extension (program_file)
			program_folder := formatter.directory (program_path)
			if program_folder /= Void then
				l_path := program_folder.as_lower
				i := l_path.substring_index ("eifgen", 1)
				if i > 1 then
					eifgen_folder := program_folder.substring (1, i - 2)
					is_running_in_estudio := l_path.substring_index ("w_code", i + 1) > 0	-- But this may change below.
				end
			end
				-- Execution information
			info := env.get_executable_info (program_path)
			if info = Void then
				create info.make (program_path)		-- Make sure 'info' isn't Void, even if underlying info wasn't generated
			end
			determine_executable_date
			from
				i := 1
			until
				 i > env.command_line.argument_count
			loop
				if not program_arguments.is_empty then
					program_arguments.append_character (' ')
				end
				program_arguments.append (arranged_parameter (env.command_line.argument (i)))
				i := i + 1
			end
				-- Site and delivery configs
			create configs.make ("config", Current)
			create icons.make ("icon", Current)
			create docs.make ("doc", Current)
			if eifgen_folder /= Void then
				l_config_path := formatter.composed_path (<<eifgen_folder, program_name + default_config_extension>>)
			else
				l_config_path := formatter.composed_path (<<program_folder, program_name + default_config_extension>>)
			end
			load_config (config, l_config_path)
			load_config (delivery, deliveries.path_to_delivery_configuration (program_folder))
			load_config (site, path_to_site_configuration)

			-- Give the user the chance to force us out of estudio mode.  (Must be after 'load_config'.)
			if is_running_in_estudio and configured_value(configuration_is_running_in_estudio) /= Void
			then
				is_running_in_estudio := configured_boolean_value(configuration_is_running_in_estudio, True)
			end

			configs.add_common_path (site.folder)
			icons.add_common_path (site.folder)
			docs.add_common_path (site.folder)
			l_dll_path := formatter.composed_path (<<site.folder, "common", "dll">>)
			if is_directory_readable (l_dll_path) then
				l_path := env.get ("PATH")
				if l_path /= Void then
					l_path := l_path.twin
					if not l_path.has_substring (l_dll_path) then
						l_path.append_character (';')
						l_path.append (l_dll_path)
						env.put (l_path, "PATH")
					end
				end
			end
				-- Variables
			variables.set ("platform", platform.name)
			variables.set ("start_time", start_time.formatted_out (time_format))
			variables.set ("host_name", host_name)
			variables.set ("pid", pid.out)
			variables.set ("user_login_id", user.login_id)
			variables.set ("user_login_domain", user.login_domain)
			variables.set ("program_date", program_date.formatted_out (time_format))
			variables.set ("program_path", program_path)
			variables.set ("program_folder", program_folder)
			variables.set ("program_file", program_file)
			variables.set ("program_name", program_name)
			variables.set ("program_arguments", program_arguments)
			variables.set ("service_center", service_center)
			variables.set ("application", name)
			variables.set ("version", version)
			variables.set ("app_description", description)
			variables.set ("app_config", config.path)
			variables.set ("app_delivery", delivery.folder)
			variables.set ("app_site", site.folder)
				-- Default settings, in case application does not use a ROSE_PILOT
			set_default_command_line (framework_default_command_line)
		end

	path_to_site_configuration: STRING
			-- Returns path to site.cfg, Void if not found
		do
			Result := checked_path (env.get ("ROSE_SITE"), deliveries.site_configuration_file_name)
			if Result = Void then
				Result := checked_path (delivery.value (configuration_path_site), deliveries.site_configuration_file_name)
			end
			if Result = Void then
				Result := checked_path (env.get ("ROSE"), deliveries.site_configuration_file_name)
			end
			if Result = Void then
				Result := checked_path (deliveries.release_candidate.folder, deliveries.site_configuration_file_name)
			end
			if Result /= Void then
				Result := Result.as_lower
			end
		end

feature -- Access

	env: attached ROSE_EXECUTION_ENVIRONMENT
			-- Execution environment
		indexing
			once_status: global
		once
			create Result
		end

	platform: attached ROSE_PLATFORM
			-- Platform this application is running on
			-- This can not be class PLATFORM because this object has to be sendable to another process
			-- Class PLATFORM only provides info on the currently running platform
			-- And we want to be able to send this information over to a remote process

	host_name: attached STRING
			-- Machine hosting current application

	pid: INTEGER
			-- Process id of the process running this application

	info: attached ROSE_EXECUTABLE_INFO
			-- Information extracted from the executable (info provided at build time, during deployment)

	is_gui: BOOLEAN
			-- Is this application a GUI application (this is set by ROSE_PILOT_GUI)

	delivery: attached ROSE_CONFIGURATION
			-- Delivery configuration, for current delivery folder
			-- Search path: all folders containing the current exe

	site: attached ROSE_CONFIGURATION
			-- Site configuration, for entire site

	is_running_within_site: BOOLEAN
			-- Are we running directly from within the 'site' folder (as opposed to running from another place redirecting site)
		do
			Result := (create {ROSE_SHARED_GLOBAL}).global.formatter.same_folder (site.path, delivery.path)
		end

	config: attached ROSE_CONFIGURATION
			-- Application configurations

	configs: attached ROSE_APPLICATION_RESOURCE
	icons: attached ROSE_APPLICATION_RESOURCE
	docs: attached ROSE_APPLICATION_RESOURCE
			-- Search paths to these resources

	user: attached ROSE_USER
			-- User currently running this application

	has_generated_metrics: BOOLEAN
			-- Has this application generated some metrics so far?

	is_running_in_estudio: BOOLEAN
			-- Is the application currently running in Eiffel Studio?

	default_command: STRING
			-- Default command to pass arguments to if command name omitted in application arguments
			-- By default, this is the first command added via 'add_command'

	default_command_line: STRING
			-- Default command line if none provided on application startup
			-- The application acts as if this command line argument was specified
			-- if no arguments provided when application started, default is "console"

	program_date: attached ROSE_DATE
			-- UTC Date when executable running this process was last modified

	program_path: attached STRING
			-- Full path to executable (where launched program resides)

	program_folder: attached STRING		-- Folder containing the 'program_path' (this is the directory part)
	eifgen_folder: STRING				-- EIFGEN, if current exe from inside an EIFGENs folder (Void if not)

	program_file: attached STRING
			-- File name of executable (with .exe)

	program_name: attached STRING
			-- File name of executable without extension (without the .exe)

	program_arguments: STRING
			-- Application arguments given on command line when program launched

	start_time: attached ROSE_DATE
			-- Time when application was started

	implicit_commands: attached LINKED_LIST [STRING]
			-- Commands executed implicitly if not provided on command line

	variables: attached ROSE_APPLICATION_VARIABLES
			-- Symbolic application variables

	service_center: STRING
			-- Application's service center (determined via deployment info bundled within exe)
		do
			Result := info.service_center
		end

	name: attached STRING
			-- Application name (intended for human reading, determined via deployment info bundled within exe)
		do
			if info.name /= Void then
				Result := info.name
			else
				Result := program_name
			end
		end

	description: STRING
			-- Short description of application purpose (determined via deployment info bundled within exe)
		do
			Result := info.description
		end

	version: STRING
			-- Application version string (determined via deployment info bundled within exe)
			-- Conventional format: M.m.r, M=Major version number, m=minor, r=revision
		do
			Result := info.version
		end

	name_version: attached STRING
			-- Application's complete title ('name' + " v" + 'version')
		do
			create Result.make_empty
			Result.append_string (name)
			if version /= Void then
				Result.append_character (' ')
				Result.append_character ('v')
				Result.append (version)
			end
		end

	is_running_in_production: BOOLEAN
			-- Is this application running from the production delivery?
		once
			Result := site.folder /= Void and then (site.folder.as_lower ~ deliveries.production.folder.as_lower)
		end

	running: attached DATE_TIME_DURATION
			-- Duration since application has started (in wallclock units)
		local
			l_date_time: ROSE_DATE
		do
			create l_date_time.make_now
			Result := l_date_time.relative_duration (start_time)
		end

	interpreted (a_string: STRING): STRING
			-- 'a_string' with all $KEYWORD expanded to their value (coming from configuration or environment variables)
		do
			Result := formatter.recursively_interpreted (a_string, 10, agent configured_or_environment_value)
		end

	configured_or_environment_value (a_key: STRING): STRING
			-- Return configured or environment value associated with 'a_key'
		do
			if a_key = Void or else a_key.is_empty then
				Result := a_key
			else
				Result := value (a_key)
				if Result = Void then
					Result := configured_value (a_key)
				end
				if Result = Void then
					Result := env.get (a_key)
				end
			end
		end

	log_path (a_relative_filename: STRING): attached STRING
			-- Full path for log file with 'a_relative_filename' for this application
		require
			is_relative: formatter.is_relative_path (a_relative_filename)
		local
			l_string, l_folder: STRING
		do
			l_folder := interpreted (configured_value (configuration_path_log))
			if l_folder /= Void and then not l_folder.is_empty and then not is_directory_readable (l_folder) then
				-- If log folder is configured, but does not exist, we try to create it
				recursive_create_directory (l_folder)
				if not is_directory_readable (l_folder) then
					-- If we didn't manage to create it, then we act as if it wasn't configured
					l_folder := Void
				end
			end
			if not is_directory_readable (l_folder) then
				-- If log folder not configured, we use the temporary folder on current machine
				l_folder := env.temporary_folder
				if l_folder = Void or else l_folder.is_empty then
					l_folder := "."
				end
			end
			-- If log folder was determined, use it
			l_string := interpreted (a_relative_filename)
			l_string := formatter.composed_path (<<l_folder, service_center, l_string>>)
			l_folder := formatter.directory (l_string)
			if not is_directory_readable (l_folder) then
				-- If folder requested by user does not exist, try to create it
				recursive_create_directory (l_folder)
			end
			if l_string /= Void then
				Result := l_string
			else
				-- This case will never be reached, just here to satisfy the 'attached' rule
				create Result.make_empty
				report_unreachable_code_reached
			end
		ensure
			result_ok: formatter.is_absolute_path (Result)
		end

	storable_path (a_relative_filename: STRING): attached STRING
			-- Full absolute path for storable file with 'a_relative_filename' for this application
			-- grandfathering accepts "\" which is not a valid relative path and returns a path ending in a directory separator
		require
			relative_filename_not_void: a_relative_filename /= Void
			empty_or_grandfathered_or_relative: a_relative_filename.is_empty or else a_relative_filename ~ "\" or else formatter.is_relative_path (a_relative_filename)
		local
			l_string: STRING
		do
			l_string := interpreted (configured_value (configuration_path_storable))
			if not is_directory_readable (l_string) then
				l_string := env.temporary_folder
				if not is_directory_readable (l_string) then
					l_string := "."
				end
			end
			if a_relative_filename.is_empty then
				l_string := l_string.twin
			elseif a_relative_filename ~ "\" then
					-- grandfathered case: return path ending in directory separator
				l_string := l_string + "\"
			else
				l_string := interpreted (formatter.composed_path (<<l_string, a_relative_filename>>))
			end
			if l_string /= Void then
				Result := l_string
			else
				-- This case will never be reached, just here to satisfy the 'attached' rule
				create Result.make_empty
				report_unreachable_code_reached
			end
		ensure
			result_not_empty: not Result.is_empty
			result_is_absolute: formatter.is_absolute_path (Result)
			result_has_relative_path: Result.substring (Result.count - a_relative_filename.count + 1, Result.count) ~ a_relative_filename
			grandfathered: (a_relative_filename ~ "\") implies Result [Result.count] = '\'
		end

	archive_path (a_relative_filename: STRING): attached STRING
			-- Full path for archive file with 'a_relative_filename' for this application
		require
			is_relative: formatter.is_relative_path (a_relative_filename)
		local
			l_string: STRING
		do
			l_string := interpreted (configured_value (configuration_path_archive))
			if not is_directory_readable (l_string) then
				l_string := env.temporary_folder
				if not is_directory_readable (l_string) then
					l_string := "."
				end
			end
			l_string := interpreted (formatter.composed_path (<<l_string, a_relative_filename>>))
			if l_string /= Void then
				Result := l_string
			else
				-- This case will never be reached, just here to satisfy the 'attached' rule
				create Result.make_empty
				report_unreachable_code_reached
			end
		ensure
			result_ok: formatter.is_absolute_path (Result)
		end

	data_path (a_relative_filename: STRING): attached STRING
			-- Full path for data file with 'a_relative_filename' for the environment in which current application is running
		require
			is_relative: formatter.is_relative_path (a_relative_filename)
		local
			l_string: STRING
		do
			l_string := configured_value (configuration_path_data)
			if not is_directory_readable (l_string) then
				l_string := (create {ROSE_SHARED_GLOBAL}).global.hardcoded_path ("\\axaros.com\shares\RCData")
			end
			l_string := interpreted (formatter.composed_path (<<l_string, a_relative_filename>>))
			if l_string /= Void then
				Result := l_string
			else
				-- This case will never be reached, just here to satisfy the 'attached' rule
				create Result.make_empty
				report_unreachable_code_reached
			end
		ensure
			result_ok: formatter.is_absolute_path (Result)
		end

	recipe_path (a_relative_filename: STRING): attached STRING
			-- Full path for data file with 'a_relative_filename' for the environment in which current application is running
		require
			is_relative: formatter.is_relative_path (a_relative_filename)
		local
			l_string: STRING
		do
			l_string := configured_value (configuration_path_recipe)
			if not is_directory_readable (l_string) then
				l_string := (create {ROSE_SHARED_GLOBAL}).global.hardcoded_path ("\\axaros.com\shares\development\recipes")
			end
			l_string := interpreted (formatter.composed_path (<<l_string, a_relative_filename>>))
			if l_string /= Void then
				Result := l_string
			else
				-- This case will never be reached, just here to satisfy the 'attached' rule
				create Result.make_empty
				report_unreachable_code_reached
			end
		ensure
			result_ok: formatter.is_absolute_path (Result)
		end

	configured_value (a_key: STRING): STRING
			-- Value for 'a_key' configured in application config, delivery or site
		do
			Result := config.value (a_key)
			if Result = Void then
				Result := delivery.value (a_key)
			end
			if Result = Void then
				Result := site.value (a_key)
			end
		end

	configured_boolean_value (a_key: STRING; a_default: BOOLEAN): BOOLEAN
			-- Boolean value for 'a_key' configured in application config, delivery or site;
			-- `a_default' is none
		local
			l_value: STRING
		do
			l_value := config.value (a_key)
			if l_value = Void then
				l_value := delivery.value (a_key)
			end
			if l_value = Void then
				l_value := site.value (a_key)
			end
			if l_value = Void then
				Result := a_default
			elseif l_value.is_boolean then
				Result := l_value.to_boolean
			elseif l_value.is_integer then
				Result := l_value.to_integer /= 0
			else
				Result := a_default
			end
		end

	value (a_key: STRING): STRING
			-- Value of 'variable' with 'a_key'
		local
			l_key: STRING
		do
			if a_key /= Void then
				l_key := a_key.as_lower
				l_key.left_adjust
				l_key.right_adjust
				if l_key ~ now_keyword then
					Result := formatter.object_representation (create {ROSE_DATE}.make_now, Void)
				elseif l_key ~ now_utc_keyword then
					Result := formatter.object_representation (create {ROSE_DATE}.make_now_utc, Void)
				elseif l_key ~ running_keyword then
					Result := formatter.object_representation (running, Void)
				elseif l_key ~ today_keyword then
					Result := formatter.object_representation (create {ROSE_DATE}.make_today, Void)
				else
					Result := variables.value (l_key)
				end
			end
		end

	debug_output: STRING
			-- Report on the current application status
		do
			Result := name
		end

	out: STRING
			-- New string containing terse printable representation
			-- of current object
		do
			Result := report
		end

	report: attached STRING
			-- Report on the current application status
		local
			i: INTEGER
			l_string: STRING
		do
			create Result.make (100)
			append_formatted (Result, "Program: $1%N", <<program_path>>)
			append_formatted (Result, "Compiled: $1%N", <<program_date>>)
			append_formatted (Result, "App: $1 v$2 (service_center: $3)%N", <<name, version, service_center>>)
			if description /= Void then
				l_string := description.twin
				l_string.left_adjust
				l_string.right_adjust
				i := l_string.index_of ('%N', 1)
				if i > 0 then
					l_string.keep_head (i - 1)
				end
				if not l_string.is_empty then
					append_formatted (Result, "$1%N", <<l_string>>)
				end
			end
			append_formatted (Result, "Command line arguments: '$1'%N", <<program_arguments>>)
			append_formatted (Result, "Pid: $1%N", <<pid>>)
			append_formatted (Result, "Started: $1%N", <<start_time>>)
			append_formatted (Result, "Running for: $1%N", <<running>>)
			append_formatted (Result, "Config path: '$1'%N", <<config.path>>)
			append_formatted (Result, "Executed on '$1' ($2) by '$3'%N", <<host_name, platform, user.full_login>>)
		end

feature -- Basic operations

	set_default_command (a_value: STRING)
			-- Set 'default_command' to 'a_value'.
		do
			default_command := a_value
			variables.set ("default_command", default_command)
		end

	set_default_command_line (a_value: STRING)
			-- Set 'default_command_line' to 'a_value'.
		do
			default_command_line := a_value
			variables.set ("default_command_line", default_command_line)
		end

	set_config (a_config: ROSE_CONFIGURATION)
			-- Set 'a_config' to be the shared application configuration.
		require
			valid_config: a_config /= Void
		do
			config := a_config
		ensure
			config_set: config = a_config
		end

	add_implicit_command (a_command_line: STRING)
			-- Add 'a_command_line' as implicit (command necessarily present when application run).
			-- If command in 'a_command_line' not explicitly mentioned in
			-- application arguments, then 'a_command_line' is automatically inserted.
		do
			implicit_commands.extend (a_command_line)
		end

	set_generated_metrics (a_value: BOOLEAN)
			-- Set 'has_generated_metrics'
		do
			has_generated_metrics := a_value
		end

	load_config (a_config: ROSE_CONFIGURATION; a_path: STRING)
			-- Load config from 'a_path' into 'a_config'
		require
			config_ok: a_config /= Void
		local
			l_parser: ROSE_CONFIGURATION_CFG_PARSER
		do
			a_config.set_path (a_path)
			if a_path /= Void and then not a_path.is_empty then
				create l_parser.make
				l_parser.error.push_private (Void)
				l_parser.import (a_config.root, a_path, False)
				a_config.set_available (True)
				l_parser.error.pop_private
			end
		end

feature {ROSE_PILOT_BASE} -- Implementation

	set_gui (a_value: BOOLEAN)
			-- Set 'is_gui'.
		do
			is_gui := a_value
		end

feature {NONE} -- Implementation

	arranged_parameter (a_param: STRING): STRING
			-- command-line parameter 'a_param' arranged (restored quotes)
		local
			i: INTEGER
		do
			if a_param /= Void then
				Result := a_param
				if a_param.count > 1 and then a_param.has (' ') then
					Result := a_param.twin
					if Result.item (1) = '-' then
						i := Result.index_of ('=', 1)
						if i > 0 and then i < Result.count then
							Result.insert_character ('"', i + 1)
							Result.append_character ('"')
						else
							Result.prepend_character ('"')
							Result.append_character ('"')
						end
					else
						Result.prepend_character ('"')
						Result.append_character ('"')
					end
				end
			end
		end

	append_formatted (a_string, a_message: STRING; a_args: ARRAY [ANY])
			-- Append formatted 'a_message' with 'a_args' to 'a_string'
		require
			string_ok: a_string /= Void
		do
			if a_string /= Void and then a_message /= Void then
				a_string.append (formatter.argumented (a_message, a_args))
			end
		end

	determine_executable_date
			-- Determine the last modification date of current executable.
		local
			l_file: RAW_FILE
			l_filedate: INTEGER
			retried: BOOLEAN
		do
			if not retried then
				create l_file.make (program_path)
				l_filedate := l_file.date
				if l_filedate > 0 then
					create program_date.make_by_unix_time (l_filedate)
					program_date.second_add (env.utc_seconds_offset)
				end
			end
			if program_date = Void then
				create program_date.make_unavailable
			end
		rescue
			if not retried then
				program_date := Void
				retried := True
				retry
			end
		end

	report_unreachable_code_reached
			-- Helper feature for finding bugs related to early handling of attached types
			-- Call this routine from parts of the code that should never be reached
			-- Once we provide full support for "on demand Void safety", we will be able to get rid of this routine
		do
			(create {ROSE_SHARED_GLOBAL}).global.report_unreachable_code_reached
		end

end
