indexing
	description: "Mail attachment"
	author: "Zoran Simic"
	date: "$Date: 2010/05/19 $"
	revision: "Revision: $"

class
	ROSE_BASE64_ENCODING

feature -- Access

	encoded_length (a_source_length: INTEGER): INTEGER is
			-- Exact number of characters needed to encode 'a_source_length'
		do
			Result := ((a_source_length * 8) / 6).ceiling
			Result := (Result / 4).ceiling * 4
			Result := Result + (Result // line_length) * 2
		end

	encoded (a_pointer: MANAGED_POINTER): STRING is
			-- Contents pointed to by 'a_pointer' encoded in base64
		local
			l_cursor, l_total_bytes, l_cutoff: INTEGER
		do
			if a_pointer /= Void then
				from
					l_cutoff := (line_length // 4) * 3
					l_total_bytes := (a_pointer.count / 3).ceiling * 3
					create Result.make (encoded_length (a_pointer.count))
				until
					l_cursor >= l_total_bytes
				loop
					encode_three_bytes (Result, a_pointer, l_cursor)
					l_cursor := l_cursor + 3
					if l_cursor \\ l_cutoff = 0 then
						Result.append_character ('%R')
						Result.append_character ('%N')
					end
				end
			end
		ensure
			result_ok: a_pointer /= Void implies (Result.count = encoded_length (a_pointer.count))
		end

	encoded_string (a_string: STRING): STRING is
			-- 'a_string' encoded in base64
		local
			l_c_string: C_STRING
			l_pointer: MANAGED_POINTER
		do
			if a_string /= Void then
				create l_c_string.make (a_string)
				create l_pointer.make_from_pointer (l_c_string.item, l_c_string.count)
				Result := encoded (l_pointer)
			end
		ensure
			result_ok: a_string /= Void implies (Result.count = encoded_length (a_string.count))
		end

	decoded_length (a_source: STRING): INTEGER is
			-- Exact number of bytes needed to decode 'a_source'
		local
			i: INTEGER
		do
			if a_source /= Void then
				from
					i := 1
				until
					i > a_source.count
				loop
					if decoded_character (a_source.item (i)) < 64 then
						Result := Result + 6
					end
					i := i + 1
				end
				Result := Result // 8
			end
		end

	decoded (a_source: STRING): MANAGED_POINTER is
			-- 'a_source' decoded
		local
			n: INTEGER
		do
			n := decoded_length (a_source)
			create Result.make (n)
			if n > 0 then
				decode (a_source, Result)
			end
		ensure
			result_ok: Result.count = decoded_length (a_source)
		end

	decoded_string (a_source: STRING): STRING is
			-- 'a_source' decoded into a string
		local
			l_pointer: MANAGED_POINTER
			c: C_STRING
		do
			l_pointer := decoded (a_source)
			create c.make_by_pointer_and_count (l_pointer.item, l_pointer.count)
			Result := c.substring (1, l_pointer.count)
		ensure
			result_ok: Result.count = decoded_length (a_source)
		end

feature -- Basic operations

	decode (a_source: STRING; a_destination: MANAGED_POINTER) is
			-- Decode 'a_source' into 'a_destination'
			-- Walk the source buffer each four character sequence is converted to 3 bytes
			-- CRLFs and =, and any characters not in the encoding table are skipped
			-- Returns the number of bytes decoded
		require
			valid_source: a_source /= Void
			valid_destination: a_destination /= Void and then a_destination.count >= decoded_length (a_source)
		local
			i, j, n, l_nbits: INTEGER
			l_decoded, l_value: INTEGER_64
		do
			from
				i := 1
				n := a_source.count
			until
				i > n
			loop
				from
					l_decoded := 0
					l_nbits := 0
				until
					l_nbits >= 24 or else i > n
				loop
					l_value := decoded_character (a_source.item (i))
					if l_value < 64 then
						l_decoded := l_decoded |<< 6
						l_decoded := l_decoded | l_value
						l_nbits := l_nbits + 6
					end
					i := i + 1
				end
				l_decoded := l_decoded |<< (24 - l_nbits)
				-- l_decoded has the 3 bytes to write to the output buffer, left to right
				from
				until
					l_nbits < 8
				loop
					l_value := ((l_decoded & 0x00ff0000) |>> 16)
					a_destination.put_integer_8 (l_value.to_integer_8, j)
					l_decoded := l_decoded |<< 8
					l_decoded := l_decoded & 0xffffffff
					l_nbits := l_nbits - 8
					j := j + 1
				end
			end
		end

feature {NONE} -- Implementation

	encoding_table: STRING is "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/"	-- Encoding table
	line_length: INTEGER is 76																		-- Line length

	encode_three_bytes (a_buffer: STRING; a_pointer: MANAGED_POINTER; a_start: INTEGER) is
			-- Encode 3 bytes
		require
			args_ok: a_buffer /= Void and then a_pointer /= Void
		local
			l_encoded, l_index, l_byte: INTEGER_64
			i, j, n: INTEGER
		do
			from
				i := a_start
				n := a_start + 3
			until
				i >= n
			loop
				if i < a_pointer.count then
					l_byte := a_pointer.read_integer_8 (i)
					l_byte := l_byte & 0xff
					l_encoded := l_encoded | l_byte
					j := j + 1
				end
				l_encoded := l_encoded |<< 8
				i := i + 1
			end
			from
				i := 0
			until
				i > 3
			loop
				if i <= j then
					l_index := l_encoded |>> 26
					a_buffer.append_character (encoding_table.item (l_index.to_integer_32 + 1))
				else
					a_buffer.append_character ('=')
				end
				l_encoded := l_encoded |<< 6
				l_encoded := l_encoded & 0xffffffff
				i := i + 1
			end
		end

	decoded_character (a_character: CHARACTER): INTEGER_64 is
			-- returns -1 if the character is invalid or should be skipped
			-- otherwise, returns the 6-bit code for the character from the encoding table
		do
			inspect a_character
			when 'A'..'Z' then
				Result := (a_character.code - 65 + 0).to_integer_64		-- 0 range starts at 'A'
			when 'a'..'z' then
				Result := (a_character.code - 97 + 26).to_integer_64	-- 26 range starts at 'a'
			when '0'..'9' then
				Result := (a_character.code - 48 + 52).to_integer_64	-- 52 range starts at '0'
			when '+' then
				Result := 62
			when '/' then
				Result := 63
			else
				Result := 64
			end
		end

end
