indexing

	description:

		"Bank/market dates and utilities"

	author: "MK, MZ"
	last_editor: "$Author: mb $"
	date: "$Date: 2014/05/14 $"
	revision: "$Revision: #8 $"
	path: "$File: //rose/source/kernel/investment_entities/rose_business_date.e $"

deferred class ROSE_BUSINESS_DATE

feature -- Utilities

	market_day_interval (a_start, a_end: ROSE_DATE): ROSE_DATE_SET is
			-- Market dates between `a_start' and `a_end'
		require
			a_start_not_void: a_start /= Void
			a_end_not_void: a_end /= Void
		do
			market_day_interval_cache.search ([a_start, a_end])
			if not market_day_interval_cache.found then
				Result := market_dates.interval (a_start, a_end)
				market_day_interval_cache.force (Result, [a_start, a_end])
			else
				Result := market_day_interval_cache.found_item
			end
		ensure
			market_day_interval_not_void: Result /= Void
			correct_first_date: Result.first >= a_start
			correct_last_date: Result.last <= a_end
		end

	market_day_interval_dates(a_start, a_end: ROSE_DATE): ROSE_LINEAR_ARRAY[ROSE_DATE]
			-- Market dates between `a_start' and `a_end'.  Like 'market_day_interval' but returns an array instead of a date_set.
		require
			a_start_not_void: a_start /= Void
			a_end_not_void: a_end /= Void
		do
			market_day_interval_dates_cache.search ([a_start, a_end])
			if not market_day_interval_dates_cache.found then
				Result := market_dates.dates(a_start, a_end)
				market_day_interval_dates_cache.force (Result, [a_start, a_end])
			else
				Result := market_day_interval_dates_cache.found_item
			end
		ensure
			market_day_interval_not_void: Result /= Void
			correct_first_date: Result.first >= a_start
			correct_last_date: Result.last <= a_end
		end

	bank_day_interval (a_start, a_end: ROSE_DATE): ROSE_DATE_SET is
			-- Bank dates between `a_start' and `a_end'
		require
			a_start_not_void: a_start /= Void
			a_end_not_void: a_end /= Void
		do
			Result := bank_dates.interval (a_start, a_end)
		ensure
			bank_day_interval_not_void: Result /= Void
			correct_first_date: Result.first >= a_start
			correct_last_date: Result.last <= a_end
		end

	ith_market_day_after (a_date: ROSE_DATE; a_days: INTEGER): ROSE_DATE is
			-- Market date `a_days' after `a_date'
		require
			a_date_not_void: a_date /= Void
		do
			Result := market_dates.ith_date_after (a_date, a_days)
		ensure
			ith_market_day_after_can_be_void: True
		end

	ith_bank_day_after (a_date: ROSE_DATE; a_days: INTEGER): ROSE_DATE is
			-- Bank date `a_days' after `a_date'
		require
			a_date_not_void: a_date /= Void
		do
			Result := bank_dates.ith_date_after (a_date, a_days)
		ensure
			ith_bank_day_after_can_be_void: True
		end

	ith_market_day_before (a_date: ROSE_DATE; a_days: INTEGER): ROSE_DATE is
			-- Market date `a_days' open days before `a_date'
		do
			Result := market_dates.ith_date_before (a_date, a_days)
		end

	ith_bank_day_before (a_date: ROSE_DATE; a_days: INTEGER): ROSE_DATE is
			-- Bank date `a_days' open days before `a_date'
		do
			Result := bank_dates.ith_date_before (a_date, a_days)
		end

	last_market_day (a_date: ROSE_DATE): ROSE_DATE is
			-- Date when market is open on or after `a_date'
		local
			l_date_set: ROSE_DATE_SET
		do
			l_date_set := market_dates
			if l_date_set.has (a_date) then
				Result := a_date
			else
				Result := l_date_set.ith_date_before (a_date, 1)
			end
		end

	next_market_day (a_date: ROSE_DATE): ROSE_DATE is
			-- Date when market is open on or before `a_date'
		local
			l_date_set: ROSE_DATE_SET
		do
			l_date_set := market_dates
			if l_date_set.has (a_date) then
				Result := a_date
			else
				Result := l_date_set.ith_date_after (a_date, 1)
			end
		end

	last_bank_day (a_date: ROSE_DATE): ROSE_DATE is
			-- Date when bank is open on or before `a_date'
		require
			a_date_not_void: a_date /= Void
		local
			l_date_set: ROSE_DATE_SET
		do
			l_date_set := bank_dates
			if l_date_set.has (a_date) then
				Result := a_date
			else
				Result := l_date_set.ith_date_before (a_date, 1)
			end
		ensure
			last_bank_day_not_void: Result /= Void
			last_bank_day_is_on_or_before_a_date: Result <= a_date
			is_bank_day: bank_dates.has (Result)
		end

	next_bank_day (a_date: ROSE_DATE): ROSE_DATE is
			-- Date when bank is open on or after `a_date'
		require
			a_date_not_void: a_date /= Void
		local
			l_date_set: ROSE_DATE_SET
		do
			l_date_set := bank_dates
			if l_date_set.has (a_date) then
				Result := a_date
			else
				Result := l_date_set.ith_date_after (a_date, 1)
			end
		ensure
			next_bank_day_not_void: Result /= Void
			next_bank_day_is_at_or_after_a_date: Result >= a_date
			is_bank_day: bank_dates.has (Result)
		end

	first_market_day_of_month (a_date: ROSE_DATE): ROSE_DATE is
			-- Date of the first market day of the month of `a_date'
		require
			a_date_not_void: a_date /= Void
		local
			l_date: ROSE_DATE
		do
			create l_date.make_by_date (a_date)
			l_date.month_back
			l_date.set_to_end_of_month
			Result := ith_market_day_after (l_date, 1)
		ensure
			first_market_day_of_month_not_void: Result /= Void
			same_month: a_date.month = Result.month
			is_market_day: market_dates.has (Result)
		end

	last_market_day_of_month (a_date: ROSE_DATE): ROSE_DATE is
			-- Date of the last market day of the month of `a_date'
		require
			a_date_not_void: a_date /= Void
		local
			l_date: ROSE_DATE
		do
			create l_date.make_by_date (a_date)
			l_date.month_forth
			l_date.set_day (1)
			Result := ith_market_day_before (l_date, 1)
		ensure
			last_market_day_of_month_not_void: Result /= Void
			same_month: a_date.month = Result.month
			is_market_day: market_dates.has (Result)
		end

feature {NONE} -- Dates

	market_dates: ROSE_DATE_SET is
			-- Set of market dates
		deferred
		ensure
			market_dates_not_void: Result /= Void
		end

	bank_dates: ROSE_DATE_SET is
			-- Set of bank dates
		deferred
		ensure
			bank_dates_not_void: Result /= Void
		end

feature {NONE} -- Cacheing

	market_day_interval_cache: TUPLE_KEYED_HASH_TABLE [ROSE_DATE_SET, TUPLE [ROSE_DATE, ROSE_DATE]] is
			-- Cache of previously computed market day intervals
		do
			if market_day_interval_cache_saved = Void then
				create market_day_interval_cache_saved.make (100)
			end
			Result := market_day_interval_cache_saved
		ensure
			available_result: Result /= Void
		end

	market_day_interval_cache_saved: detachable TUPLE_KEYED_HASH_TABLE [ROSE_DATE_SET, TUPLE [ROSE_DATE, ROSE_DATE]]
			-- Lazy cache of computed market day intervals

	market_day_interval_dates_cache: TUPLE_KEYED_HASH_TABLE [ROSE_LINEAR_ARRAY [ROSE_DATE], TUPLE [ROSE_DATE, ROSE_DATE]] is
			-- Cache of previously computed market day interval dates
		do
			if market_day_interval_dates_cache_saved = Void then
				create market_day_interval_dates_cache_saved.make (100)
			end
			Result := market_day_interval_dates_cache_saved
		ensure
			available_result: Result /= Void
		end

	market_day_interval_dates_cache_saved: detachable TUPLE_KEYED_HASH_TABLE [ROSE_LINEAR_ARRAY [ROSE_DATE], TUPLE [ROSE_DATE, ROSE_DATE]]
			-- Lazy cache of computed market day interval dates

end
