note
	description: "Definition of a chart axis"
	author: "Zoran Simic"
	last_editor: "$Author: zs $"
	date: "$Date: 2011/07/11 $"
	revision: "$Revision: #24 $"
	path: "$File: //rose/source/kernel/infrastructure/widgets/charts/rose_chart_axis.e $"

deferred class
	ROSE_CHART_AXIS

inherit
	ROSE_CHART_CONSTANTS
		export
			{NONE} all
		end
	ROSE_CONSTANTS
		export
			{NONE} all
		end
	ANY

feature {NONE} -- Initialization

	make is
			-- create the object using default values
		do
			is_line_docked := True
			create holder
			create tick_holder
			create grid_holder
			holder.extend (tick_holder)
			create axis_line
			holder.extend (axis_line)
			create title_figure
			holder.extend (title_figure)
			create dir
			create origin
			create size
			create l1
			create l2
		end

feature -- Access

	coordinate_system: ROSE_CHART_COORDINATE_SYSTEM[ROSE_CHART_POINT]
			-- Associated coordinate system

	display: ROSE_CHART_DISPLAY
			-- Associated display (holds the visible layout of the chart)

	is_high: BOOLEAN
			-- Is this axis to appear at high position of the chart?

	is_line_docked: BOOLEAN
			-- Is this line docked to an end of the chart representation?

	line_origin: DOUBLE
			-- Origin at which to show the axis line
			-- (only taken into account when 'is_line_docked' is True

	is_horizontal: BOOLEAN
			-- Is this axis horizontal?

	is_logarithmic: BOOLEAN
			-- Are the coordinates of this axis logarithmic?

	is_equidistant: BOOLEAN
			-- Should ticks be considered equidistant?

	date_format: STRING
		-- format to convert a date into a string

	is_shown: BOOLEAN is
			-- is the axis shown?
		do
			Result := holder.is_show_requested
		end

	holder: EV_MODEL_GROUP
			-- Group holding graphical representation of this axis

	title: STRING is
			-- Axis title (for example 'Prices'...)
		do
			Result := title_figure.text
		end

	label (a_index: INTEGER): STRING is
			-- Label associated to tick with 'a_index', if any
		deferred
		end

	scaled (a_coordinate: DOUBLE): DOUBLE is
			-- 'a_coordinate' projected on this axis
		do
			if is_logarithmic then
				if a_coordinate > 0.0 then
					Result := math.log10 (a_coordinate)
				else
					Result := constants.missing_double
				end
			else
				Result := a_coordinate
			end
		end

	projected (a_value, a_origin, a_scale: DOUBLE): INTEGER is
			-- Projected 'a_x' with 'a_origin' and 'a_scale'
		do
			if is_horizontal then
				Result := projected_x (a_value,a_origin,a_scale)
			else
				Result := projected_y (a_value,a_origin,a_scale)
			end
		end

	reverse_projected (a_value: INTEGER; a_origin, a_scale: DOUBLE): DOUBLE is
			-- Reverse projected 'a_x' with 'a_origin' and 'a_scale'
		do
			if is_horizontal then
				Result := reverse_projected_x (a_value,a_origin,a_scale)
			else
				Result := reverse_projected_y (a_value,a_origin,a_scale)
			end
		end

	date_to_tick (a_date: ROSE_DATE): DOUBLE is
			-- 'a_date' tick value (projected on this axis)
		do
			if a_date /= Void then
				Result := a_date.to_double
			end
		end

	tick_to_date (a_tick: DOUBLE): ROSE_DATE is
			-- Date corresponding to 'a_tick' (inverse of 'date_to_tick')
		do
			create Result.make_by_double (a_tick)
		end

	has_point (a_x, a_y: INTEGER): BOOLEAN is
			-- is the point (a_x, a_y) on the axis
		local
			l_x1, l_x2, l_y1, l_y2: INTEGER
		do
			l_x1 := display.chart_group.point_x + origin.x
			l_y1 := display.chart_group.point_y + origin.y
			l_x2 := l_x1 + dir.x * size.x
			l_y2 := l_y1 + dir.y * size.y
			if is_horizontal then
				Result := a_x >= l_x1.min (l_x2) and then a_x <= l_x1.max (l_x2)
					and then a_y >= l_y1.min (l_y2) and then a_y <= l_y1.max (l_y2)
			else
				Result := a_y >= l_x1.min (l_x2) and then a_y <= l_x1.max (l_x2)
					and then a_x >= l_y1.min (l_y2) and then a_x <= l_y1.max (l_y2)
			end
		end

	index_coordinate (a_index: INTEGER): DOUBLE is
			-- Coordinate for 'a_index', if any
		deferred
		end

	debug_output: STRING is
			-- String that should be displayed in debugger to represent `Current'.
		do
			create Result.make (20)
			if projected_ticks /= Void then
				Result.append (projected_ticks.count.out + " ticks")
			end
		end

	visible_range: PAIR [DOUBLE, DOUBLE] is
			-- visible range
		do
			create Result.make (minimum_visible, maximum_visible)
		ensure
			result_ok: Result /= Void
		end

feature -- Basic operations

	ensure_count (a_count: INTEGER) is
			-- Ensure that 'a_count' indices are defined...
		deferred
		end

	reset is
			-- Reset the axis
		deferred
		end

	clear_all is
			-- Clear all values and set defaults
		do
			set_title (Void)
			grid_holder.wipe_out
			tick_holder.wipe_out
			is_logarithmic := False
			is_equidistant := False
			show (True)
			axis_line.set_point_a_position (0,0)
			axis_line.set_point_b_position (0,0)
			maximum_visible := Void
			minimum_visible := Void
			reset
		ensure
			no_title: title.is_empty
			not_logarithmic: not is_logarithmic
			not_equidistant: not is_equidistant
		end

	set_title (a_title: STRING) is
			-- Set 'title'.
		do
			if a_title /= Void then
				title_figure.set_text (a_title)
			else
				title_figure.set_text ("")
			end
		end

	show (a_value: BOOLEAN) is
			-- Show or hide this axis.
		do
			if a_value then
				holder.show
			else
				holder.hide
			end
			if coordinate_system /= Void then
				coordinate_system.state.set_needs_redraw
			end
		end

	set_line_docked (a_value: BOOLEAN) is
			-- set 'is_line_docked'
		do
			is_line_docked := a_value
		end

	set_date_format (a_format: STRING) is
			-- set `date_format' to `a_format'
			-- set the format for dated labels
			-- works only if called before `set_dates' or `set_ticked_labels'
			-- works only if you don't call another set_date* function afterwards
			-- there is also `set_date_xslt' you might want to take a look at it
		require
			string_not_void: a_format /= void
		do
			date_format_type := format_type_standard
			date_format := a_format
		end

	set_date_format_xslt (a_date_format_zone: DT_FIXED_OFFSET_TIME_ZONE; a_date_format_picture, a_date_format_language, a_date_format_calendar, a_date_format_country: STRING) is
			-- set `date_format' to `a_format'
			-- set the format for dated labels
			-- works only if called before `set_dates' or `set_ticked_labels'
			-- works only if you don't call another set_date* function afterwards
		require
			string_not_void: a_date_format_picture /= void
		do
			date_format_type := format_type_xslt
			date_format_zone := a_date_format_zone
			date_format_picture := a_date_format_picture
			date_format_language := a_date_format_language
			date_format_calendar := a_date_format_calendar
			date_format_country := a_date_format_country
		end

	set_equidistant (a_value: BOOLEAN) is
			-- set 'is_equidistant' to 'a_value'
		do
			if is_equidistant /= a_value then
				is_equidistant := a_value
				coordinate_system.state.set_needs_coordination
			end
		end

	set_logarithmic (a_value: BOOLEAN) is
			-- set 'is_logarithmic' to 'a_value'
		do
			is_logarithmic := a_value
			coordinate_system.state.set_needs_coordination
		end

	set_line_origin (a_origin: DOUBLE) is
			-- set 'line_origin' to 'a_origin'
		do
			line_origin := a_origin
			coordinate_system.state.set_needs_projection
		end

	-- TODO: REMOVE THAT? (doesn't work properly with new system)
	set_indices (a_lower, a_upper: INTEGER) is
			-- Add 'a_lower'..'a_upper' range to this axis.
		deferred
		end

	set_dates (a_dates: ARRAY [ROSE_DATE]) is
			-- Add 'a_dates' as meaningful ticks
		require
			valid_dates: a_dates /= Void
		deferred
		end

	set_labels (a_labels: ARRAY [ANY]) is
			-- Add 'a_labels' as indexed ticks.
		require
			valid_labels: a_labels /= Void
		deferred
		end

	set_ticked_labels (a_ticks: ARRAY [DOUBLE]; a_labels: ARRAY [ANY]) is
			-- Add 'a_ticks' with 'a_labels' as miningful ticks.
		require
			valid_ticks: a_ticks /= Void
		deferred
		end

	set_date_range (a_start, a_end: ROSE_DATE) is
			-- Set date range
		require
			dates_ok: a_start /= Void and then a_end /= Void and then a_end > a_start
		deferred
		end

	set_minimum_visible (a_value: DOUBLE) is
			-- This point (in this axis) is guaranteed to be visible on the chart even if no data exists with this value.
		do
			if minimum_visible = Void then
				create minimum_visible
			end
			minimum_visible.set_item (a_value)
--			elseif a_value < minimum_visible.item then
--				minimum_visible.set_item (a_value)
--			end
		end

	set_maximum_visible (a_value: DOUBLE) is
			-- This point (in this axis) is guaranteed to be visible on the chart even if no data exists with this value.
		do
			if maximum_visible = Void then
				create maximum_visible
			end
			maximum_visible.set_item (a_value)
--			elseif a_value > maximum_visible.item then
--				maximum_visible.set_item (a_value)
--			end
		end

	set_minimum_visible_date (a_date: ROSE_DATE) is
			-- This point (in this axis) is guaranteed to be visible on the chart even if no data exists with this value.
		require
			date_ok: a_date /= Void
		do
			if minimum_visible = Void then
				create minimum_visible
			end
			minimum_visible.set_item (date_to_tick (a_date))
		end

	set_maximum_visible_date (a_date: ROSE_DATE) is
			-- This point (in this axis) is guaranteed to be visible on the chart even if no data exists with this value.
		require
			date_ok: a_date /= Void
		do
			if maximum_visible = Void then
				create maximum_visible
			end
			maximum_visible.set_item (date_to_tick (a_date))
		end

	sync_to_axis (a_axis: ROSE_CHART_AXIS) is
			-- Ensures that the lower and upper bounds of these two axes are the same.
		do
			if a_axis = Void and sync_axis /= Void then
				sync_axis.sync_to_axis (Void)
			end
			non_recusive_sync_to_axis (a_axis)
			if a_axis /= Void then
				a_axis.non_recusive_sync_to_axis (Current)
			end
		end

	not_already_synced (a_axis: ROSE_CHART_AXIS): BOOLEAN is
			-- Tells if 'a_axis" is not already synchronized
		do
			Result := sync_axis /= Void implies sync_axis = a_axis.sync_axis
		end

feature {ROSE_CHART_WIDGET} -- Implementation

	refresh_figures is
			-- refresh figures of the axis
		do
			axis_line.set_foreground_color (display.settings.axis_color)
			title_figure.set_foreground_color (display.settings.title_color)
			title_figure.set_font (display.settings.axis_font)
		end

feature {ROSE_CHART_COORDINATE_SYSTEM, ROSE_CHART_PROJECTOR} -- Implementation

	set_system (a_system: like coordinate_system) is
			-- set the coordinate system to be used
		require
			valid_system: a_system /= Void
		do
			coordinate_system := a_system
			display := a_system.display
			is_high := coordinate_system = display.graphics.secondary_system
			is_horizontal := Current = coordinate_system.x_axis
			display.grid_group.extend (grid_holder)
			display.axis_group.extend (holder)
		end

	remove is
			-- remove axis from system
		require
			coordinate_system_ok: coordinate_system /= Void
		do
			display.grid_group.prune_all (grid_holder)
			display.axis_group.prune_all (holder)
			coordinate_system := Void
			display := Void
		end

	project (a_origin, a_scale: DOUBLE) is
			-- project axis using values from its coordinate system
		require
			coordinate_system_ok: coordinate_system /= Void
		local
			ts: EV_COORDINATE
		do
			-- Reset figures first
			tick_holder.wipe_out
			grid_holder.wipe_out
			-- Calculate size, direction, title-size (ts) and high extremity
			if is_horizontal then
				size.set (display.chart_width, display.settings.axis_height)
				dir.set (1, 1)
				create ts.set (title_figure.width, title_figure.height.max (title_figure.font.height))
				l2.set_y (-display.chart_height)
			else
				size.set (display.chart_height, display.settings.axis_width)
				dir.set (-1, -1)
				create ts.set (title_figure.height.max (title_figure.font.height), title_figure.width)
				l2.set_y (display.chart_width)
			end
			-- Adapt for high text position (text is mirrored when at high position)
			if is_high then
				dir.set_y (-dir.y)
				origin.set (0, l2.y)
				l2.set_y (0)
			else
				origin.set (0, 0)
			end
			tick_y := origin.y + dir.y * display.settings.axis_tick_size
			l1.set_x (origin.x)
			l2.set_x (dir.x * size.x)
			visible1 := l1.x.min (l2.x) - 2
			visible2 := l1.x.max (l2.x) + 2
			if is_horizontal then
				set_position (title_figure, origin.x + dir.x * (size.x - ts.x) // 2, origin.y + dir.y * (size.y - (dir.y * ts.y).max (0)))
			else
				set_position (title_figure, visible1 - ts.x, origin.y + dir.y * (size.y - (dir.y * ts.y).max (0)))
			end
			set_tooltip (title_figure, title)
			create projected_ticks.make_empty
			prepare_ticks (a_origin, a_scale)
			if is_line_docked then
				l1.set_y (origin.y)
			else
				l1.set_y (projected_origin)
			end
			-- Set line and title coordinates
			set_position_a (axis_line, l1.x, l1.y)
			set_position_b (axis_line, l2.x, l1.y)
			-- Represent the ticks
			if is_horizontal then
				represent_horizontal
			else
				represent_vertical
			end
			projected_ticks := Void
			last_tick := Void
		end

	set_projected_origin (a_value: INTEGER) is
			-- set 'projected_origin' to 'a_value'
		do
			projected_origin := a_value
		end

feature {ROSE_CHART_AXIS, ROSE_CHART_AXIS_TICK} -- Implementation

	format_type_standard: INTEGER is 1
		-- use ROSE_DATE.formatted_out

	format_type_xslt: INTEGER is 2
		-- use ROSE_DATE.xslt_formatted_date

	date_format_type: INTEGER
		-- how should dates be formated?

	date_format_zone: DT_FIXED_OFFSET_TIME_ZONE
		-- Date format zone

	date_format_picture: STRING
		-- formatparameter for xslt to convert a date into a string

	date_format_language: STRING
		-- formatparameter for xslt to convert a date into a string

	date_format_calendar: STRING
		-- formatparameter for xslt to convert a date into a string

	date_format_country: STRING_8
		-- formatparameter for xslt to convert a date into a string

	format_date (a_date: ROSE_DATE): STRING is
			-- format `a_date' with respect to `date_format_type'
			-- returns Void if no date_type set
		require
			valid_date: a_date /= Void
			valid_type: date_format_type = format_type_standard implies date_format /= Void
		do
			inspect date_format_type
			when format_type_standard then
				Result := a_date.formatted_out (date_format)
			when format_type_xslt then
				Result :=  a_date.xslt_formatted_date (date_format_zone, date_format_picture, date_format_language, date_format_calendar, date_format_country).value
			else
				REsult := Void
			end
		end


	projected_origin, visible1, visible2, tick_y: INTEGER -- integers
	grid_holder, tick_holder: EV_MODEL_GROUP 	-- grid, tick
	dir, origin, size, l1, l2: EV_COORDINATE	-- dir, origin, size

	set_tooltip (a_figure: EV_MODEL; a_tooltip: STRING) is
			-- set the tooltip for the axis
		do
			a_figure.pointer_enter_actions.wipe_out
			a_figure.pointer_leave_actions.wipe_out
			a_figure.pointer_enter_actions.extend (agent display.set_tooltip (a_tooltip,Void))
			a_figure.pointer_leave_actions.extend (agent display.set_tooltip (Void,Void))
		end

	set_position (a_figure: EV_MODEL_SINGLE_POINTED; a_main, a_sec: INTEGER) is
			-- set the position of 'a_figure'
		do
			if is_horizontal then
				display.set_position (a_figure, a_main, a_sec)
			else
				display.set_position (a_figure, a_sec, a_main)
			end
		end

	set_position_a (a_figure: EV_MODEL_DOUBLE_POINTED; a_main, a_sec: INTEGER) is
			-- set the position_a of 'a_figure'
		do
			if is_horizontal then
				display.set_position_a (a_figure, a_main, a_sec)
			else
				display.set_position_a (a_figure, a_sec, a_main)
			end
		end

	set_position_b (a_figure: EV_MODEL_DOUBLE_POINTED; a_main, a_sec: INTEGER) is
			-- set the position_b of 'a_figure'
		do
			if is_horizontal then
				display.set_position_b (a_figure, a_main, a_sec)
			else
				display.set_position_b (a_figure, a_sec, a_main)
			end
		end

feature {ROSE_CHART_AXIS} -- Implementation

	non_recusive_sync_to_axis (a_axis: ROSE_CHART_AXIS) is
			-- set 'sync_axis' to 'a_axis' (the other axis this axis should be synced with)
		do
			sync_axis := a_axis
		end

feature {ROSE_CHART_AXIS, ROSE_CHART_COORDINATE_SYSTEM} -- Implementation

	sync_axis: ROSE_CHART_AXIS -- sync axis

feature {ROSE_CHART_RENDER_PLOT} -- Implementation

	minimum_visible: DOUBLE_REF
		-- minimum viisible

	maximum_visible: DOUBLE_REF
		-- maximum visible

feature {NONE} -- Implementation

	axis_line: EV_MODEL_LINE		-- Axis line
	title_figure: EV_MODEL_TEXT		-- Axis title
	projected_ticks: ROSE_LINEAR_ARRAY [ROSE_CHART_AXIS_TICK]
			-- Direction, projected origin, and line points
			-- All these coordinates are expressed in terms of (primary,secondary)

	represent_horizontal is
			-- represent the axis horizontal
		local
			l_tick: ROSE_CHART_AXIS_TICK
			i, w, l_x, l_y1, l_y2, l_tx1, l_tx2: INTEGER
		do
			if is_high then
				l_y1 := dir.y * display.settings.axis_font_height
			end
			l_y2 := l_y1 + dir.y * (display.settings.axis_font_height - 2)
			from
				i := projected_ticks.lower
				l_tx1 := visible1 - dir.x * 1000
				l_tx2 := l_tx1
			until
				i > projected_ticks.upper
			loop
				l_tick := projected_ticks.item (i)
				w := l_tick.text_width // 2 - 1
				l_x := l_tick.projected - w
				if l_tick.level >= 100 or else l_tx1 < l_x then
					l_tx1 := l_tick.projected + w + 8
					l_tick.set_color (display.settings.axis_color_1)
					l_tick.add_to_axis (Current,-w,l_y1)
				elseif l_tx2 < l_x then
					l_tx2 := l_tick.projected + w + 8
					l_tick.set_color (display.settings.axis_color_2)
					l_tick.add_to_axis (Current,-w,l_y2)
				end
				i := i + 1
			end
		end

	represent_vertical is
			-- represent the axis vertical
		local
			l_tick: ROSE_CHART_AXIS_TICK
			i, l_x, l_y_offset: INTEGER
		do
			if is_high then
				l_x := dir.y * 4
			else
			end
			l_y_offset := - display.settings.axis_font_height // 2
			from
				i := projected_ticks.lower
			until
				i > projected_ticks.count
			loop
				l_tick := projected_ticks.item (i)
				if is_high then
					l_tick.add_to_axis (Current,l_y_offset,l_x)
				else
					l_tick.add_to_axis (Current,l_y_offset,l_x - l_tick.text_width)
				end
				i := i + 1
			end
		end

	prepare_ticks (a_origin, a_scale: DOUBLE) is
			-- Populate 'projected_ticks' with relevant ticks for this projection
		deferred
		end

	last_tick: ROSE_CHART_AXIS_TICK
			-- Last tick added, used internally

	add_tick (a_projected: INTEGER; a_label: STRING; a_level: INTEGER) is
			-- add a tick to the axis
		do
			if a_projected >= visible1 and then a_projected <= visible2 then
				if last_tick = Void or else a_projected.abs > last_tick.projected.abs + 10 then
					create last_tick.make (display.settings.axis_font.twin)
					projected_ticks.extend (last_tick)
					last_tick.set (a_projected,a_label,a_level)
				elseif a_level > last_tick.level then
					last_tick.set (a_projected,a_label,a_level)
				end
			end
		end

	double_equal (d1, d2: DOUBLE): BOOLEAN is
			-- are d1 and d2 equal considering a given tolerance?
		do
			Result := (d2-d1).abs < tolerance
		end

	tolerance: DOUBLE is 10.0e-6
			-- 2 numbers are considered equal if they don't differ more than 'tolerance'

	math: DOUBLE_MATH is
			-- Needed for log10 function
		once
			create Result
		end

end
