indexing
	description: "Indexed axis"
	author: "Zoran Simic"
	date: "$Date: 2009/11/16 $"
	revision: "$Revision: #17 $"

class
	ROSE_CHART_AXIS_INDEXED

inherit
	ROSE_CHART_AXIS

create
	make

feature -- Access

	count: INTEGER
			-- Number of ticks

	ticks: ROSE_LINEAR_ARRAY [DOUBLE]
			-- Coordinate for each meaningful tick (optional)

	labels: ROSE_LINEAR_ARRAY [STRING]
			-- Label to show for each meaningful tick (optional)

	start_day, end_day: INTEGER
			-- Start and end days if this axis contains dates

	label (a_index: INTEGER): STRING is
			-- label at index 'a_index'
		do
			if labels /= Void and then labels.valid_index (a_index) then
				Result := labels.item (a_index)
			elseif start_day /= 0 then
				Result := "?"
			else
				Result := index_coordinate (a_index).out
			end
		end

	index_coordinate (a_index: INTEGER): DOUBLE is
			-- Coordinate of tick at 'a_index'
		do
			if is_equidistant or else ticks = Void or else ticks.is_empty then
				Result := a_index
			elseif not ticks.valid_index (a_index) then
				if start_day = 0 then
					Result := a_index
				elseif a_index < 1 then
					Result := start_day + a_index - 1
				else
					Result := end_day + a_index - ticks.upper
				end
			else
				Result := ticks.item (a_index)
			end
			Result := scaled (Result)
		end

feature -- Basic operations

	ensure_count (a_count: INTEGER) is
			-- Ensure that 'a_count' indices are defined...
		do
			count := count.max (a_count)
		end

	reset is
			-- reset axis
		do
			start_day := 0
			end_day := 0
			count := 0
			ticks := Void
			labels := Void
		end

	set_indices (a_lower, a_upper: INTEGER) is
			-- Add 'a_lower'..'a_upper' range to this axis.
		local
			i: INTEGER
		do
			reset
			create ticks.make_empty
			create labels.make_empty
			from
				i := a_lower
			until
				i > a_upper
			loop
				ticks.extend (i)
				labels.extend (i.out)
				i := i + 1
			end
		end

	set_dates (a_dates: ARRAY [ROSE_DATE]) is
			-- Add 'a_dates' as meaningful ticks
		local
			d: ROSE_DATE
			t: DOUBLE
			i: INTEGER
		do
			reset
			create ticks.make_empty
			create labels.make_empty
			-- First, take all dates and make ticks and labels out of them
			from
				i := a_dates.lower
			until
				i > a_dates.upper
			loop
				d := a_dates.item (i)
				if d /= Void then
					end_day := d.days
					if start_day = 0 then
						start_day := end_day
					end
					ticks.extend (date_to_tick (d))
					labels.extend (represented_date (d))
				else
					ticks.extend (0)
					labels.extend (Void)
				end
				i := i + 1
			end
			ensure_count (ticks.count)
			if start_day /= 0 then
				from
					t := start_day
					i := ticks.lower
				until
					i > ticks.upper
				loop
					if labels.item (i) = Void then
						t := t + 0.5
						ticks.put (t, i)
						labels.put ("?", i)
					else
						t := ticks.item (i)
					end
					i := i + 1
				end
			end
		end

	set_labels (a_labels: ARRAY [ANY]) is
			-- Add 'a_labels' as indexed ticks.
		local
			t: DOUBLE
			i: INTEGER
		do
			reset
			create ticks.make_empty
			create labels.make_empty
			from
				i := a_labels.lower
			until
				i > a_labels.upper
			loop
				t := t + 1
				ticks.extend (t)
				if a_labels.item (i) /= Void then
					labels.extend (a_labels.item (i).out)
				else
					labels.extend ("?")
				end
				i := i + 1
			end
		end

	set_ticked_labels (a_ticks: ARRAY [DOUBLE]; a_labels: ARRAY [ANY]) is
			-- Add 'a_ticks' with 'a_labels' as miningful ticks.
		local
			l_date: ROSE_DATE
			i, l_count: INTEGER
		do
			if a_ticks /= Void then
				ticks ?= a_ticks
				if ticks = Void then
					create ticks.make_from_array (a_ticks)
				end
				l_count := ticks.count
			end
			if a_labels /= Void then
				create labels.make_empty
				from
					i := a_labels.lower
				until
					i > a_labels.upper
				loop
					if a_labels.item (i) /= Void then
						l_date ?= a_labels.item (i)
						if l_date /= Void then
							labels.extend (represented_date (l_date))
						else
							labels.extend (a_labels.item (i).out)
						end
					else
						labels.extend ("?")
					end
					i := i + 1
				end
			end
			ensure_count (l_count)
		end

	set_date_range (a_start, a_end: ROSE_DATE) is
			-- Set date range
		do
		end

	choose_a_time_interval is
			-- This should automatically call tick_time_interval and one of set_hourly, etc.
		require
			implemented: False
		do
		end

	set_hourly is
			-- Set 'is_hourly'.
		do
			clear_time_modes
			is_hourly := True
		end

	set_daily is
			-- Set 'is_daily'.
		do
			clear_time_modes
			is_daily := True
		end

	set_weekly is
			-- Set 'is_weekly'.
		do
			clear_time_modes
			is_weekly := True
		end

	set_monthly is
			-- Set 'is_monthly'.
		do
			clear_time_modes
			is_monthly := True
		end

	set_tick_time_interval (a_tick_time_interval: INTEGER) is
			-- Set 'tick_time_interval'.
		do
			tick_time_interval := a_tick_time_interval
		end

feature {NONE} -- Implementation

	tick_time_interval: INTEGER		-- Interval size between 2 ticks
	is_hourly: BOOLEAN				-- Is this axis representing dates to be shown in hourly mode?
	is_daily: BOOLEAN				-- Is this axis representing dates to be shown in daily mode?
	is_weekly: BOOLEAN				-- Is this axis representing dates to be shown in weekly mode?
	is_monthly: BOOLEAN				-- Is this axis representing dates to be shown in monthly mode?

	clear_time_modes is
			-- Clear time modes
		do
			is_hourly := False
			is_daily := False
			is_weekly := False
			is_monthly := False
		end

	prepare_ticks (a_origin, a_scale: DOUBLE) is
			-- select how ticks should be repressented
		do
			if is_hourly then
				prepare_hourly_ticks (a_origin, a_scale)
			elseif is_daily then
				prepare_daily_ticks (a_origin, a_scale)
			elseif is_weekly then
				prepare_weekly_ticks (a_origin, a_scale)
			elseif is_monthly then
				prepare_monthly_ticks (a_origin, a_scale)
			else
				prepare_ticks_old (a_origin, a_scale)
			end
		end

	prepare_ticks_old (a_origin, a_scale: DOUBLE) is
			-- Populate 'projected_ticks' with relevant ticks for this projection
		local
			l_y: DOUBLE
			i, p: INTEGER
		do
			from
				i := 1
			until
				i > count
			loop
				p := 0
				l_y := index_coordinate (i)
				if double_equal (l_y, line_origin) then
					p := 100
				elseif start_day = 0 then
					p := l_y.rounded.abs \\ 10
					if p = 0 then
						p := 50
					elseif p /= 5 then
						p := 0
					end
				end
				add_tick (projected (l_y, a_origin, a_scale), label (i), p)
				i := i + 1
			end
		end

	prepare_monthly_ticks (a_origin, a_scale: DOUBLE) is
			-- Populate 'projected_ticks' with relevant ticks for this projection
		require
			implemented: False
		do
		end

	prepare_weekly_ticks (a_origin, a_scale: DOUBLE) is
			-- Populate 'projected_ticks' with relevant ticks for this projection
		require
			implemented: False
		do
		end

	prepare_daily_ticks (a_origin, a_scale: DOUBLE) is
			-- Populate 'projected_ticks' with relevant ticks for this projection
		local
			l_x: INTEGER
			l_date_time: ROSE_DATE
			l_date: ROSE_DATE
		do
			from
				l_x := index_coordinate (1).truncated_to_integer
			until
				l_x>index_coordinate (count)
			loop
				create l_date.make_by_days (l_x)
				create l_date_time.make (l_date.year, l_date.month, l_date.day)
				add_tick (
					projected (l_x, a_origin, a_scale),
					l_date_time.formatted_out ("yyyy/[0]mm/[0]dd"),
					0)
				l_x := l_x + tick_time_interval
			end
		end

	prepare_hourly_ticks (a_origin, a_scale: DOUBLE) is
			-- Populate 'projected_ticks' with relevant ticks for this projection
		local
			l_x: DOUBLE
			l_date_time: ROSE_DATE
			l_date: ROSE_DATE
			l_frac: DOUBLE
		do
			from
				l_x := index_coordinate (1).truncated_to_integer
			until
				l_x>index_coordinate (count)
			loop
				l_frac := l_x - l_x.floor
				l_frac := (l_frac * 24.0).rounded
				if l_frac>=24.0
				then
					l_frac := 0
					l_x := l_x.truncated_to_integer + 1
				end
				create l_date.make_by_days (l_x.truncated_to_integer)
				create l_date_time.make_date_time (l_date.year, l_date.month, l_date.day, l_frac.truncated_to_integer, 0, 0)
				if l_frac.abs<0.001
				then
					add_tick (
						projected (l_x, a_origin, a_scale),
						l_date_time.formatted_out ("yyyy/[0]mm/[0]dd"),
						100)
				else
					add_tick (
						projected (l_x, a_origin, a_scale),
						l_date_time.formatted_out ("[0]hh:[0]mi"),
						0)
				end
				l_x := l_x + tick_time_interval / 24.0
			end
		end

	represented_date (a_date: ROSE_DATE): STRING is
			-- calculates the propre string representation of `a_date'
		do
			if a_date /= Void then
				Result := format_date (a_date)
				if Result = Void then
					Result := a_date.formatted_out ("[0]mm/[0]dd [0]hh:[0]mi:[0]ss")
				end
			else
				create Result.make (8)
				Result.append (a_date.month.out)
				Result.append_character ('/')
				Result.append (a_date.day.out)
				Result.append_character ('/')
				Result.append (to_double_digit (a_date.year \\ 100))
			end
		end

	days_to_string (a_value: DOUBLE; a_format_string: STRING): STRING is
			-- Represent 'a_value' as a date-tick
		do
			Result := tick_to_date (a_value).formatted_out (a_format_string)
		end

	to_double_digit (a_int: INTEGER): STRING is
			-- 'a_int' represented with 2 digits at least
		do
			create Result.make (2)
			if a_int < 10 then
				Result.append_character ('0')
			end
			Result.append (a_int.out)
		end

end
