indexing
	description: "Definition of an chart render plot"
	author: "Urs Doenni"
	date: "$Date: 2009/03/25 $"
	revision: "$Revision: #7 $"

class
	ROSE_CHART_RENDER_PLOT [P -> ROSE_CHART_POINT]

inherit
	ROSE_CHART_CONSTANTS

	ROSE_CHART_LIST_ITEM
		redefine
			set_name
		end

	ROSE_CHART_PROJECTABLE

create {ROSE_CHART_RENDERER}
	 make

feature {NONE} -- Initialization

	make (a_sequence: like plot; a_renderer: ROSE_CHART_RENDERER [P]) is
			-- create the plot using the sequence
		do
			plot := a_sequence
			renderer := a_renderer
			name := plot.name
			create points.make
			create holder
			create x_stats.make_empty
			create y_stats.make_empty
			create world
			is_legend_shown := True
			plot.do_all (agent add_point_back)
			plot.extend_front_actions.extend (agent add_point_front)
			plot.extend_back_actions.extend (agent add_point_back)
			plot.removed_actions.extend (agent remove_point)
		end

feature -- Access

	plot: ROSE_CHART_PLOT [P]
			-- Plot which is rendered

	renderer: ROSE_CHART_RENDERER [P]
			-- Used renderer

	coordinate_system: ROSE_CHART_COORDINATE_SYSTEM [ROSE_CHART_POINT]
			-- Coordinate system this plot is attached to

	holder: EV_MODEL_GROUP
			-- Group holding figures for this plot

	world: ROSE_RECTANGLE
			-- World

	display: ROSE_CHART_DISPLAY
			-- Associated display

	settings: ROSE_CHART_SETTINGS is
			-- Associated settings
		require
			display_ok: display /= Void
		do
			Result := display.settings
		end

	state: ROSE_CHART_STATE is
			-- state of coordinatesystem
		require
			coordinate_system_ok: coordinate_system /= Void
		do
			Result := coordinate_system.state
		end

	pebble: ANY
			-- Associated pebble (for pick-dropping)

	description: STRING
			-- Extensive description

	color: EV_COLOR
			-- Color corresponding to use for this plot

	points: ROSE_CHART_TWO_WAY_LIST[ROSE_CHART_RENDER_POINT[P]]
			-- Points composing this plot

	x_stats, y_stats: ROSE_MOMENTS
			-- Plot statistics

	point (a_key: HASHABLE): ROSE_CHART_RENDER_POINT[P] is
			-- Point at 'a_index'
		do
			Result := points.item_from_key (a_key)
		end

	is_highlighted: BOOLEAN
			-- Is this plot currently highlit?

	is_shown: BOOLEAN is
			-- Is this plot currently shown?
		do
			Result := holder.is_show_requested
		end

	is_legend_shown: BOOLEAN
			-- Is the legend currently shown?

	intersects (a_rectangle: EV_RECTANGLE): BOOLEAN is
			-- Does this plot intersect with 'a_rectangle'?
		do
			from
				points.start
			until
				Result or else points.after
			loop
				Result := points.item.intersects (a_rectangle)
				points.forth
			end
		end

	debug_output: STRING is
			-- String that should be displayed in debugger to represent `Current'.
		do
			Result := points.count.out + " points"
		end

feature -- Basic operations

	highlight (a_on: BOOLEAN) is
			-- Highlight this plot.
		do
			is_highlighted := a_on
			from
				points.start
			until
				points.after
			loop
				points.item.highlight (a_on)
				points.forth
			end
			state.set_needs_redraw
		end

	remove is
			-- Remove this plot from the container chart
		require
			coordinate_system_ok: coordinate_system /= Void
		do
			coordinate_system.remove_plot (Current.name)

			plot.extend_front_actions.prune_all (agent add_point_front)
			plot.extend_back_actions.prune_all (agent add_point_back)
			plot.removed_actions.prune_all (agent remove_point)
		end

	show_legend (a_on: BOOLEAN) is
			-- Show or hide the legend belonging to this plot
		do
			is_legend_shown := a_on
			state.set_needs_refresh
		end

	show (a_on: BOOLEAN) is
			-- Show/hide this plot.
		local
			l_shown: BOOLEAN
		do
			l_shown := is_shown
			holder.disable_events_sended_to_group
			if l_shown /= a_on then
				if a_on then
					holder.show
					holder.enable_sensitive
				else
					holder.hide
					holder.disable_sensitive
				end
				state.set_needs_redraw
			end
		end

	set_name (a_name: STRING) is
			-- Set 'name'.
		do
			name := a_name
			if coordinate_system /= Void then
				state.set_needs_refresh
			end
		end

	set_description (a_description: STRING) is
			-- Set 'description'.
		do
			description := a_description
		end

	set_pebble (a_pebble: ANY) is
			-- Set 'pebble'.
		do
			pebble := a_pebble
		end

	coordinate is
			-- Transform given coordinates to coordinate system one's.
		require
			coordinate_system_ok: coordinate_system /= Void
		local
			l_point: ROSE_CHART_RENDER_POINT[P]
			l_x_axis: ROSE_CHART_AXIS
			l_y_axis: ROSE_CHART_AXIS
		do
			world.unset
			l_x_axis := coordinate_system.x_axis
			l_y_axis := coordinate_system.y_axis
			if points.count > 0 then
				x_stats.initialize
				y_stats.initialize
				l_x_axis.ensure_count (points.count)
				l_y_axis.ensure_count (points.count)
				from
					points.start
				until
					points.after
				loop
					l_point := points.item
					l_point.coordinate (l_x_axis, l_y_axis, x_stats, y_stats)
					points.forth
				end

				if x_stats.count > 1 then
					x_stats.finalize_moments
				end
				if y_stats.count > 1 then
					y_stats.finalize_moments
				end
				world.embed_point (x_stats.minimum,y_stats.minimum)
				world.embed_point (x_stats.maximum,y_stats.maximum)
				if same_double (world.x1,world.x2) then
					world.embed_point (world.x1 - tolerance, world.y1)
					world.embed_point (world.x2 + tolerance, world.y2)
				end
				if same_double (world.y1,world.y2) then
					world.embed_point (world.x1, world.y1 - tolerance)
					world.embed_point (world.x2, world.y2 + tolerance)
				end
				if coordinate_system.x_axis.minimum_visible /= Void then
					world.embed_point (coordinate_system.x_axis.minimum_visible.item, world.y1)
				end
				if coordinate_system.x_axis.maximum_visible /= Void then
					world.embed_point (coordinate_system.x_axis.maximum_visible.item, world.y1)
				end
				if coordinate_system.y_axis.minimum_visible /= Void then
					world.embed_point (world.x1, coordinate_system.y_axis.minimum_visible.item)
				end
				if coordinate_system.y_axis.maximum_visible /= Void then
					world.embed_point (world.x1, coordinate_system.y_axis.maximum_visible.item)
				end
			end
			if coordinate_system.x_axis.minimum_visible /= Void then
				world.set_x (coordinate_system.x_axis.minimum_visible.item, world.x2)
			end
			if coordinate_system.x_axis.maximum_visible /= Void then
				world.set_x (world.x1, coordinate_system.x_axis.maximum_visible.item)
			end
			if coordinate_system.y_axis.minimum_visible /= Void then
				world.set_y (coordinate_system.y_axis.minimum_visible.item, world.y2)
			end
			if coordinate_system.y_axis.maximum_visible /= Void then
				world.set_y (world.y1, coordinate_system.y_axis.maximum_visible.item)
			end
		end

	set_color (a_color: EV_COLOR) is
			-- set color
		do
			if color = Void then
				create color
			end
			if a_color /= Void then
				color.copy (a_color)
			end
		end

feature {NONE} -- Implementation

	add_point_front (a_point: P) is
			-- put the point in front of the points list
		do
			points.extend_front (new_render_point (a_point))
			if coordinate_system /= Void then
				state.set_needs_all
			end
		end

	add_point_back (a_point: P) is
			-- put the point in back of the points list
		do
			points.extend_back (new_render_point (a_point))
			if coordinate_system /= Void then
				state.set_needs_all
			end
		end

	new_render_point (a_point: P): ROSE_CHART_RENDER_POINT[P] is
			-- create a new render point correctly set up
		do
			Result := renderer.new_render_point (a_point)
			Result.set_plot (Current)
		end

	remove_point (a_key: HASHABLE) is
			-- Remove 'a_point' from this plot.
		local
			l_point: ROSE_CHART_RENDER_POINT[P]
		do
			l_point := point (a_key)
			if l_point /= Void then
				l_point.remove_figure
				points.remove (a_key)
				state.set_needs_rendering
			end
		end

feature {ROSE_CHART_COORDINATE_SYSTEM} -- Implementation

	set_coordinate_system (a_system: like coordinate_system) is
			-- Set 'coordinate_system'.
		do
			coordinate_system := a_system
			if a_system /= Void then
				display := coordinate_system.display
			else
				display := Void
			end
		end

	project (a_x0, a_y0, a_x_scale, a_y_scale: DOUBLE) is
			-- Project all points in this plot for origin ('a_x0,'a_y0')
			-- and with given scales.
		do
			from
				points.start
			until
				points.after
			loop
				points.item.project (a_x0,a_y0,a_x_scale,a_y_scale)
				points.forth
			end
		end

feature {ROSE_CHART_WIDGET} -- Implementation

	refresh_figures is
			-- Refresh layout of the figures representing this plot.
			-- Typically, some color etc. have changed for this to be called.
		do
			from
				points.start
			until
				points.after
			loop
				points.item.refresh_figure
				points.forth
			end
			state.set_needs_redraw
		end

feature {NONE} -- Implementation

	same_double (d1, d2: DOUBLE): BOOLEAN is
			-- same double
		do
			Result := (d2 - d1).abs < tolerance
		end

	tolerance: DOUBLE is 0.0001

end
