indexing
	description: "Definition of an abstract chart render point"
	author: "Urs Doenni"
	date: "$Date: 2010/03/19 $"
	revision: "$Revision: #22 $"
	archive: "$File: //rose/source/kernel/infrastructure/widgets/charts/rose_chart_render_point.e $"

deferred class
	ROSE_CHART_RENDER_POINT[P -> ROSE_CHART_POINT]

inherit
	KEYED
	ROSE_CHART_CONSTANTS

feature {NONE} -- Initialization

	make (a_point: like point) is
			-- Create object with a reference to the point to be represented
		require
			point_not_void: a_point /= Void
		do
			point := a_point
			set_key (point.key)
			if point.color /= Void then
				set_color (point.color)
			end
			if point.full_label /= Void then
				set_label (point.full_label)
			end
			if point.label_font /= Void then
				set_label_font (point.label_font)
			end
			point.change_actions.extend (agent update_impl)
			initialize (a_point)
		end

	initialize (a_point: like point) is
			-- Initialize the point
		require
			point_not_void: a_point /= Void
		deferred
		end

	update_impl is
			-- update the point (called when the underlying data has changed)
		do
			update
			render_plot.state.set_needs_all
		end

	update is
			-- update the point
		deferred
		end

feature {ROSE_CHART_RENDER_PLOT} -- Think about moving "set_plot" in ROSE_CHART_RENDERER.new_render_point

	set_plot (a_plot: like render_plot) is
			-- set the plot this point belongs to
		require
			plot_not_void: a_plot /= Void
		do
			render_plot := a_plot
		end

feature -- Access
	are_values_defined: BOOLEAN is
			-- says if the values that make up this points are defined
		deferred
		end


	point: P
			-- Point to be represented

	render_plot: ROSE_CHART_RENDER_PLOT[P]
			-- Corresponding plot

	figure: ROSE_CHART_FIGURE
			-- Figure representing this point

	data: ANY is
			-- Object associated to this point
		do
			Result := point.data
		end

	label: STRING
			-- Label
	label_shown: BOOLEAN
			-- Is the label visible?
	label_font : EV_FONT
			-- Labels to show at this point

	is_highlighted: BOOLEAN
			-- Is it highlighted?

	color: EV_COLOR is
			-- Color for this point (plot's color if not set at this point level)
		do
			if self_color = Void then
				Result := render_plot.color
			else
				Result := self_color
			end
		end

	intersects (a_rectangle: EV_RECTANGLE): BOOLEAN is
			-- Does this point intersect with 'a_rectangle'?
		do
			if figure /= Void then
				Result := figure.intersects (a_rectangle)
			end
		end

	debug_output: STRING is
			-- String that should be displayed in debugger to represent `Current'.
		deferred
		end

	full_label : STRING is
			-- Full label of current object
		do
			Result := label
		end

	pebble_agent: FUNCTION [ANY, TUPLE, ANY]
			-- Pebble agent

	pebble: ANY is
			-- pebble for this point
		do
			if pebble_agent /= Void then
				Result := pebble_agent.item ([])
			end
		end

	tooltip_agent: FUNCTION[ANY, TUPLE, STRING]
			-- Agent to generate the tooltip

	tooltip: STRING is
			-- tooltip for this point
		do
			if tooltip_agent /= Void then
				Result := tooltip_agent.item ([])
			end
		end


feature -- Basic operations

	set_pebble_agent (a_agent: like pebble_agent) is
			-- set the agent which generates a pebble
		do
			pebble_agent := a_agent
		end

	set_tooltip_agent (a_agent: like tooltip_agent) is
			-- set the agent which generates the tooltip
		do
			tooltip_agent := a_agent
		end

	refresh_figure is
			-- Refresh layout of this point.
		do
			if figure /= Void then
				figure.refresh
			end
		end

	coordinate(a_x_axis, a_y_axis: ROSE_CHART_AXIS; a_x_stats, a_y_stats: ROSE_MOMENTS) is
			-- Coordinate axis and stats
		deferred
		end

	project (a_x0, a_y0, a_x_scale, a_y_scale: DOUBLE) is
			-- Project this point for origin ('a_x0,'a_y0')
			-- and with given scales.
		deferred
		end

	remove_figure is
			-- Remove the figure
		do
			if figure /= Void then
				figure.remove
				figure := Void
			end
		end

	highlight (a_on: BOOLEAN) is
			-- Highlight this point.
		do
			if figure /= Void then
				figure.highlight (a_on)
			end
			is_highlighted := a_on
		end

	show_label is
			-- Show the label
		do
			label_shown := True
		end

	hide_label is
			-- Hide the label
		do
			label_shown := False
		end

	set_label (a_label: STRING) is
			-- Replace the label
		do
			label := a_label
		end

	set_label_font (a_font : EV_FONT) is
			-- Set the font of the label
		do
			label_font := a_font
		end

	set_color (a_color: EV_COLOR) is
			-- Set the color of the current object
		do
			self_color := a_color
		end

feature {ROSE_CHART_RENDERER} -- Implementation

	set_figure (a_figure: like figure) is
			-- Set 'figure'.
		require
			valid_figure: a_figure /= Void implies a_figure.chart_point = Void
		do
			figure := a_figure
		end

	self_color: EV_COLOR
			-- Current color

feature {NONE} -- Implementation

	coordinated_value(a_given_value : DOUBLE; a_axis : ROSE_CHART_AXIS; a_stats: ROSE_MOMENTS) : DOUBLE is
			-- Value for a given value on an axis for a stats
		do
			if is_defined_double (a_given_value) then
				Result := a_axis.scaled (a_given_value)
				if is_defined_double (Result) then
					a_stats.add_observation (Result, 1.0)
				end
			else
				Result := a_given_value
			end
		end
end
