indexing

	description: "[
		Widget holding charts.
		Generic parameters are:
			P1: Main Coordinate System type of points
		   P2: Secondary Coordinate System type of points
		]"

	author: "Urs Doenni"
	last_editor: "$Author: e2 $"
	date: "$Date: 2015/03/20 $"
	revision: "$Revision: #42 $"
	path: "$File: //rose/source/kernel/infrastructure/widgets/charts/rose_chart_widget.e $"

class	ROSE_CHART_WIDGET [P1 -> ROSE_CHART_POINT, P2 -> ROSE_CHART_POINT]

inherit

	ROSE_GRAPHICS_WIDGET
		redefine
			events, display, state, clear_all, adapt_view_from_selection,
			adapt_view_from_dragging, show_contextual_menu, on_mouse_wheel,
			drop_actions, on_mouse_click
		end

	ROSE_CHART_CONSTANTS
		undefine
			default_create, copy, is_equal
		end

create

	make

feature {NONE} -- Initialization

	initialize_widget
			-- Specific initialization of the widget.
		do
			create main_system
			create secondary_system
			main_system.set_widget (Current)
			secondary_system.set_widget (Current)

			display.selection.set_secondary_coordinate_system (secondary_system)
			display.selection.set_main_coordinate_system (main_system)
			-- One should not be able to drop anything on the legend
			display.legend.figure.drop_actions.extend (agent (a_dummy: BOOLEAN) do do_nothing end)
		end

feature -- Access

	events: ROSE_CHART_EVENTS
			-- Events generated by this widget

	display: ROSE_CHART_DISPLAY
			-- Associated display, holding the Vision2 figures and settings

	state: ROSE_CHART_STATE
			-- State of the current chart representation

	main_system: ROSE_CHART_COORDINATE_SYSTEM [P1]
			-- Main coordinate system
			-- This system is used when adding plots through the widget itself

	secondary_system: ROSE_CHART_COORDINATE_SYSTEM [P2]
			-- Secondary coordinate system

	external_legend: ROSE_CHART_LEGEND_WIDGET
			-- If this is not Void, legend will also be displayed in the specified external widget

	plot_count: INTEGER
			-- Total number of plots in this widget
		do
			Result := main_system.plots.count + secondary_system.plots.count
		end

	highlighted_plots_count (a_highlighted: BOOLEAN): INTEGER
			-- Number of plots that are currently (un)highlit, in same highlight state as 'a_highlighted'
		do
			Result := main_system.highlighted_plots_count (a_highlighted) + secondary_system.highlighted_plots_count (a_highlighted)
		end

	shown_plots_count (a_shown: BOOLEAN): INTEGER
			-- Number of plots that are currently shown/hidden, in same shown state as 'a_shown'
		do
			Result := main_system.shown_plots_count (a_shown) + secondary_system.shown_plots_count (a_shown)
		end

	highlighted_plots (a_highlit: BOOLEAN): ROSE_LINEAR_ARRAY [like plot_type]
			-- Currently highlit plots (or not)
		do
			Result := main_system.highlighted_plots (a_highlit)
			Result.append_array (secondary_system.highlighted_plots (a_highlit))
		end

	shown_plots (a_shown: BOOLEAN): ROSE_LINEAR_ARRAY [like plot_type]
			-- Currently highlit plots (or not)
		do
			Result := main_system.highlighted_plots (a_shown)
			Result.append_array (secondary_system.highlighted_plots (a_shown))
		end

	axes_count (a_horizontal, a_high: BOOLEAN): INTEGER
			-- Number of axes at position (a_horizontal,a_high)
		do
			Result := main_system.axes_count (a_horizontal,a_high) + secondary_system.axes_count (a_horizontal,a_high)
		end

	plot_by_name (a_name: STRING): like plot_type
			-- Plot with 'a_name', if any
		require
			valid_name : a_name /= Void
		do
			Result := main_system.plot_by_name (a_name)
			if Result = Void then
				Result := secondary_system.plot_by_name (a_name)
			end
		end

	pebble_agent: FUNCTION [ANY, TUPLE [ROSE_CHART_POINT], ANY]
			-- Agent to generate the pebble

	drop_actions: EV_PND_ACTION_SEQUENCE
			-- Drop actions for the widget
		do
			if display /= Void and display.figure_world /= Void then
				Result := display.figure_world.drop_actions
			end
		end

feature -- Basic operations

	set_small_widget (a_widget: ROSE_CHART_MINI_WIDGET [P1, P2])
			-- set small_widget to `a_widget'
		do
			small_widget := a_widget
		end


	set_renderers (a_main_system_renderer: ROSE_CHART_RENDERER[P1]; a_secondary_system_renderer: ROSE_CHART_RENDERER[P2])
			-- set the renderers for the main and secondary system
		require
			valid_renderers : a_main_system_renderer /= Void and a_secondary_system_renderer /= Void
		do
			set_main_system_renderer (a_main_system_renderer)
			set_secondary_system_renderer (a_secondary_system_renderer)
		end


	set_main_system_renderer (a_renderer: ROSE_CHART_RENDERER[P1])
			-- set the renderer used for the main coordinate system
		do
			main_system.set_renderer (a_renderer)
		end

	set_secondary_system_renderer (a_renderer: ROSE_CHART_RENDERER[P2])
			-- set the renderer used for the secondary coordinate system
		do
			secondary_system.set_renderer (a_renderer)
		end

	set_external_legend (a_legend: ROSE_CHART_LEGEND_WIDGET)
			-- Set `external_legend' to `a_legend'
		require
			legend_not_valid: a_legend /= Void
		do
			external_legend := a_legend
			a_legend.clear_legend
		ensure
			legend_set: external_legend = a_legend
		end

	clear_all
			-- Clear out the charts.
		do
			Precursor
			main_system.clear_all
			secondary_system.clear_all

			if external_legend /= Void then
				external_legend.clear_legend
			end
		end

	highlight_plots (a_plots: ARRAY [like plot_type]; a_on: BOOLEAN)
			-- Higlight (or not) all plots in 'a_plots'.
		do
			if a_plots /= Void then
				a_plots.do_all (agent highlight_plot (?, a_on))
			end
		end

	refresh
			-- Force a refresh of this widget.
		do
			state.set_needs_all
			on_timeout
			if small_widget /= void then
				small_widget.refresh
			end
		end

	set_pebble_agent (a_pebble_agent: like pebble_agent)
			-- Set `pebble_agent' to `a_pebble_agent'
		do
			pebble_agent := a_pebble_agent
		ensure
			pebble_agent_set: pebble_agent = a_pebble_agent
		end

	reset_additional_legend_items
			-- Remove any additional legend items
		do
			if plots_in_secondary_not_in_main = Void then
				create plots_in_secondary_not_in_main.make_empty
			else
				plots_in_secondary_not_in_main.wipe_out
			end
		end

	add_to_legend (a_t: like additional_plots_template)
		-- Add the string 'plot_name' with color 'color' to the legend
		do
			if plots_in_secondary_not_in_main = Void then
				create plots_in_secondary_not_in_main.make_empty
			else
				plots_in_secondary_not_in_main.extend (a_t)
			end
		end

	reset_zoom
			-- Reset zoom
			-- Exported version
		do
			on_zoom_out
		end

	show_quick_selector (a_widget: EV_WIDGET)
			-- Show quick selector modally to 'a_widget' (the "quick selector" allows to quickly show/hide plots in the chart)
		local
			l_scroll: EV_SCROLLABLE_AREA
			l_widget: EV_WIDGET
		do
			if quick_selector = Void then
				create qs_highlit.make (0)
				create qs_shown.make (0)
				create quick_selector
				quick_selector.set_title ("Charts quick selector")
				quick_selector.set_icon ("selector.ico")
				quick_selector.show_ok_button (False)
				l_widget := Gui.factory.new_button (quick_selector.control_box, "Actions...", agent on_qs_menu, 1)
				create l_scroll
				quick_selector.main_box.extend (l_scroll)
				quick_selector_box := Gui.factory.new_vbox (l_scroll, 0, 2)
				quick_selector.show_actions.extend (agent refresh_quick_selector)
			end
			if a_widget /= Void then
				quick_selector.show_modal (a_widget)
			else
				quick_selector.show_relative (Current)
			end
		end

feature {NONE} -- Agents

	highlight_plot (a_plot: like plot_type; a_on: BOOLEAN)
			-- Highlight a plot
		do
			if a_plot /= Void then
				a_plot.highlight (a_on)
			end
		end

	show_contextual_menu (a_x, a_y: INTEGER)
			-- Show the contextual menu at `a_x'/`a_y'
		do
			Precursor (a_x, a_y)
		end

	contextual_menu (a_x, a_y: INTEGER): EV_MENU
			-- contextual menu
		local
			l_selected: ROSE_LINEAR_ARRAY [like plot_type]
			l_plot_menu: EV_MENU
		do
			create Result
			if display.legend.has_point (a_x,a_y) then
				-- Do nothing
			elseif selection.is_within_selectable (a_x,a_y) then
				l_selected := plots_in_selection (a_x, a_y)
			end
			if not view.is_normal then
				add_menu (Result,Void,<<
					["Cancel zoom ",agent on_zoom_out],
					Void
				>>)
			end
			add_data_menu (Result)
			if l_selected /= Void and then l_selected.count > 0 then
				create l_plot_menu.make_with_text ("Selected Series ")

				Result.extend (l_plot_menu)
				l_selected.do_all (agent add_highlight_plot_menu (l_plot_menu, ?))
				if l_selected.count > 1 then
					l_plot_menu.extend (create {EV_MENU_SEPARATOR})
					add_highlight_all_menu (l_plot_menu, l_selected)
				end
			end
			if plot_count > 0 then
				add_highlight_series_menu (Result)
				add_export_menu (Result)
			end
			add_menu (Result, Void, <<Void, ["Quick selector... ", agent show_quick_selector (Void)]>>)
			add_menu (Result, Void, <<Void, ["Format chart... ", agent (display.settings).show]>>)
			add_menu (Result, Void, <<Void, ["Palette ", agent show_palette]>>)
			add_menu (Result, Void, <<Void, ["Color Brewer ", agent show_color_brewer]>>)
		end

	show_color_brewer is
			-- Show dialog to select palette from ColorBrewer.
		local
			l_chooser: COLOR_BREWER_WIDGET
		do
			create l_chooser
			l_chooser.show_modal (Current)
			if l_chooser.selected_sequence /= Void then
				display.legend.set_palette (l_chooser.palette)
				if attached {ROSE_CHART_RENDERER_COLOR_PLOT} main_system.renderer as l_renderer then
					l_renderer.set_gradient (create {ROSE_CHART_COLOR_FUNCTION}.make_with_color_array (l_chooser.selected_sequence,
													l_renderer.gradient.lower_y_value, l_renderer.gradient.upper_y_value))
				elseif attached {ROSE_CHART_RENDERER_COLOR_PLOT} secondary_system.renderer as l_renderer_2 then
					l_renderer_2.set_gradient (create {ROSE_CHART_COLOR_FUNCTION}.make_with_color_array (l_chooser.selected_sequence,
						l_renderer_2.gradient.lower_y_value, l_renderer_2.gradient.upper_y_value))
				end
			end
		end

	add_data_menu (a_menu: EV_MENU)
			-- Menu options applying to underlying data,
			-- for customisation by descendants
		require
			a_menu_not_void: a_menu /= Void
		do
		end

	add_export_menu (a_menu: EV_MENU)
			-- Add picture export menu
		require
			a_menu_not_void: a_menu /= Void
		do
			add_menu (a_menu,"Export ",<<
			        ["Data to Excel",agent export_to_excel],
					["Save as PNG picture ",agent on_export_picture],
					["Save as postscript", agent on_export_postscript],
					["Save as SVG", agent on_export_svg],
					["Send as SVG via email", agent on_send_to],
					["Copy to clipboard as SVG", agent on_clipboard],
					["Print... ",agent on_print]
				>>)
		end

	add_highlight_series_menu (a_menu: EV_MENU)
			-- Add menu to highlight all series.
		require
			a_menu_not_void: a_menu /= Void
		local
			l_highlighted: ROSE_LINEAR_ARRAY [like plot_type]
		do
			l_highlighted := highlighted_plots (True)
			add_menu (a_menu,"All Series",<<
					["Highlight ",agent highlight_plots (all_plots,True)],
					["Un-highlight ",agent highlight_plots (all_plots,False)],
					Void,
					["Correlations",agent show_correlation (all_plots)],
					Void,
					["Hide highlit ",agent show_plots (l_highlighted,False)],
					["Show highlit only ",agent show_only_plots (l_highlighted)],
					["Show all ",agent show_plots (all_plots,True)]
				>>)
			Gui.factory.enable_by_name (a_menu,"Hide only ",l_highlighted /= Void and then not l_highlighted.is_empty)
			Gui.factory.enable_by_name (a_menu,"Show highlit only ",l_highlighted /= Void and then not l_highlighted.is_empty)
		end

	add_highlight_all_menu (a_menu: EV_MENU; a_selected: ROSE_LINEAR_ARRAY [like plot_type])
			-- Add highlight menu for all plots in a list.
		require
			a_menu_not_void: a_menu /= Void
			a_selected_not_void: a_selected /= Void
			a_selected_has: not a_selected.is_empty
		do
			add_menu (a_menu, "All of the above ",<<
					["Highlight ",agent highlight_plots (a_selected,True)],
					["Un-highlight ",agent highlight_plots (a_selected,False)],
					Void,
					["Correlations",agent show_correlation (a_selected)],
					Void,
					["Remove ",agent remove_plots (a_selected)],
					Void,
					["Show only these ",agent show_only_plots (a_selected)],
					["Hide ",agent show_plots (a_selected,False)]
				>>)
		end

	add_highlight_plot_menu (a_menu: EV_MENU; a_plot: like plot_type)
			-- Add menu to highlight a single plot.
		require
			a_menu_not_void: a_menu /= Void
			a_plot_not_void: a_plot /= Void
		do
			add_menu (a_menu,a_plot.name,<<
					["Highlight ",agent a_plot.highlight (True)],
					["Un-highlight ",agent a_plot.highlight (False)],
					["Highlight Points",agent highlight_points (a_plot, True)],
					["Un-highlight Points",agent highlight_points (a_plot, False)],
					Void,
					["Regression and moments ",agent show_regression (a_plot)],
					Void,
					["Remove ",agent a_plot.remove],
					Void,
					["Show only this ",agent show_only_plots (<<a_plot>>)],
					["Hide ",agent show_plots (<<a_plot>>,False)]
				>>)
		end

	highlight_points (a_plot: like plot_type; a_highlight: BOOLEAN)
			-- highlights all points in the plot
		local
			l_selection_in_coordinates: ROSE_RECTANGLE
			l_coord_view: ROSE_RECTANGLE
		do
			if selection.selected /= Void then
				l_coord_view := a_plot.coordinate_system.current_view

				if l_coord_view.is_set then
					create l_selection_in_coordinates.set_from_rectangle(l_coord_view.relative_rectangle (selection.normal_selected))
				else
					create l_selection_in_coordinates.set_from_rectangle(a_plot.coordinate_system.world.relative_rectangle (selection.normal_selected))
				end
				l_selection_in_coordinates.set_y (-l_selection_in_coordinates.y1, -l_selection_in_coordinates.y2)

				a_plot.points.do_if (agent highlight_point,	agent point_is_in_rectangle (?, l_selection_in_coordinates))
			end
		end

	highlight_point (a_point: ROSE_CHART_RENDER_POINT[ROSE_CHART_POINT])
			-- Hightlight `a_point'
		require
			valid_point: a_point /= Void
		do
			a_point.highlight(True)
		end

	show_regression (a_plot: like plot_type)
			-- Show regression line, confidence intervals and moments for `a_plot' in a new window.
		require
			a_plot_not_void: a_plot /= Void
		local
			l_display: ROSE_PLOT_REGRESSION
		do
			if {l_plot: ROSE_CHART_RENDER_PLOT [ROSE_CHART_2D_POINT]} a_plot then
				create l_display.make (l_plot)
				l_display.open_in_window
			end
		end

	show_correlation (a_plots: ROSE_LINEAR_ARRAY [like plot_type])
			-- Show regression line between any two from `a_plots'.
		require
			a_plots_not_void: a_plots /= Void
		local
			l_plots: ROSE_LINEAR_ARRAY [ROSE_CHART_RENDER_PLOT [ROSE_CHART_2D_POINT]]
			l_display: ROSE_PLOT_CORRELATION
		do
			create l_plots.make_empty
			a_plots.do_all (agent add_2d_plot (l_plots, ?))
			if l_plots.count > 1 then
				create l_display.make (l_plots)
				l_display.open_in_window
			end
		end

	add_2d_plot (a_plots: ROSE_LINEAR_ARRAY [ROSE_CHART_RENDER_PLOT [ROSE_CHART_2D_POINT]];
		a_plot: ROSE_CHART_RENDER_PLOT [ROSE_CHART_POINT])
			-- Add `a_plot' to `a_plots', if conforming.
		require
			a_plots_not_void: a_plots /= Void
		do
			if {l_plot: ROSE_CHART_RENDER_PLOT [ROSE_CHART_2D_POINT]} a_plot then
				a_plots.extend (l_plot)
			end
		ensure
			zero_or_one_more: (a_plots.count = old a_plots.count) or (a_plots.count = old a_plots.count + 1)
			correct_plot_added_at_end: (a_plots.count = old a_plots.count + 1) implies
				({l_p: ROSE_CHART_RENDER_PLOT [ROSE_CHART_2D_POINT]} a_plot and then a_plots.last ~ l_p)
		end

	point_is_in_rectangle (a_point: ROSE_CHART_RENDER_POINT [ROSE_CHART_POINT]; a_rect: ROSE_RECTANGLE): BOOLEAN
			-- is 'a_point' in the rectangle
		do
			-- todo (use the figure to determine if charts touch each other)
--			Result := a_rect.contains_x(a_point.chart_data.x) and a_rect.contains_y(a_point.chart_data.y)
		end

	quick_selector: ROSE_DIALOG
			-- Quick selector dialog, the "quick selector" allows to quickly show/hide plots on the chart

	qs_highlit, qs_shown: HASH_TABLE [EV_CHECK_BUTTON,STRING]
			-- hash tables for quick selection

	quick_selector_box: EV_BOX
			-- Top-level widget for `quick_selector'

	on_qs_menu
			--quick selector menu
		local
			l_menu: EV_MENU
		do
			l_menu := Gui.factory.new_menu (["Contextual",
				["<= Refresh series list from chart ",agent refresh_quick_selector],
				Void,
				["=> Reapply selections to chart ",agent reapply_quick_selector],
				Void,
				["Show all",agent qs_select (qs_shown, True)],
				["Hide all",agent qs_select (qs_shown, False)],
				Void,
				["Highlight all",agent qs_select (qs_highlit, True)],
				["Un-highlight all",agent qs_select (qs_highlit, False)]
			])
			l_menu.show
		end

	qs_select (a_hash: HASH_TABLE [EV_CHECK_BUTTON, STRING]; a_value: BOOLEAN)
			-- make quick selection
		require
			hash_ok: a_hash /= Void
		do
			a_hash.linear_representation.do_all (agent button_enable_select (?, a_value))
		end


	reapply_quick_selector
			-- reaply quick selector
		do
			qs_highlit.linear_representation.do_all (agent button_select)
			qs_shown.linear_representation.do_all (agent button_select)
		end

	button_enable_select (a_button: EV_CHECK_BUTTON; a_selected: BOOLEAN)
			-- Enable/disable selection on button
		require
			a_button_not_void: a_button /= Void
		do
			Gui.enable_select (a_button, a_selected)
		end

	button_select (a_button: EV_CHECK_BUTTON)
			-- Execute select actions on a button.
		require
			a_button_not_void: a_button /= Void
		do
			a_button.select_actions.call ([])
		end

	refresh_quick_selector
			-- Refresh quickselector
		do
			quick_selector_box.wipe_out
			qs_highlit.wipe_out
			qs_shown.wipe_out
			add_quick_selector_plots (main_system)
			add_quick_selector_plots (secondary_system)
			quick_selector.main_box.set_minimum_size (400, (plot_count * 25).min (500).max (60))
			quick_selector.set_size (quick_selector.main_box.minimum_width + 16, quick_selector.main_box.minimum_height + 34)
		end

	add_quick_selector_plots (a_system: ROSE_CHART_COORDINATE_SYSTEM [ROSE_CHART_POINT])
			-- Add quick selector items for 'a_system'
		require
			system_ok: a_system /= Void
		local
			l_hbox: EV_BOX
			l_check: EV_CHECK_BUTTON
			l_widget: EV_WIDGET
			l_plot: like plot_type
			l_linkable: LINKABLE [like plot_type]
		do
			if a_system.plots.count > 0 then
				if quick_selector_box.count > 0 then
					gui.factory.add_separator (quick_selector_box)
				end
				from
					l_linkable := a_system.plots.first_element
				until
					l_linkable = Void
				loop
					l_plot := l_linkable.item
					l_hbox := Gui.factory.new_hbox (quick_selector_box, 0, 1)
					l_check := Gui.factory.new_check_button (l_hbox, l_plot.is_highlighted, "Highlight", Void, 1)
					l_check.select_actions.extend (agent handle_qs_highlight (a_system, l_plot.name, l_check))
					qs_highlit.put (l_check,l_plot.name)
					l_check := Gui.factory.new_check_button (l_hbox, l_plot.is_shown, "Show", Void, 1)
					l_check.select_actions.extend (agent handle_qs_show (a_system, l_plot.name, l_check))
					qs_shown.put (l_check,l_plot.name)
					l_widget := Gui.factory.new_label (l_hbox, l_plot.name, 1)
					l_linkable := l_linkable.right
				end
			end
		end

	handle_qs_highlight (a_system: ROSE_CHART_COORDINATE_SYSTEM [ROSE_CHART_POINT]; a_name: STRING; a_button: EV_CHECK_BUTTON)
			-- (Un)highlight plot with 'a_name' when corresponding checkbox changed in quick-selector.
		require
			system_ok: a_system /= Void
			button_ok: a_button /= Void
		local
			l_plot: like plot_type
		do
			l_plot := a_system.plot_by_name (a_name)
			if l_plot /= Void then
				l_plot.highlight (a_button.is_selected)
			end
		end

	handle_qs_show (a_system: ROSE_CHART_COORDINATE_SYSTEM [ROSE_CHART_POINT]; a_name: STRING; a_button: EV_CHECK_BUTTON)
			-- Show/hide plot with 'a_name' when corresponding checkbox changed in quick-selector.
		require
			system_ok: a_system /= Void
			button_ok: a_button /= Void
		local
			l_plot: like plot_type
		do
			l_plot := a_system.plot_by_name (a_name)
			if l_plot /= Void then
				l_plot.show (a_button.is_selected)
			end
		end

	on_zoom_out
			-- set normal view
		do
			view.make_normal
			state.set_needs_projection
			main_system.view_changed_actions.call ([])
			secondary_system.view_changed_actions.call([])
			if small_widget /= void then
				small_widget.adapt_cutout_rectangle (current)
			end
		end

	on_mouse_wheel (a_direction: INTEGER)
			-- mouse wheel moved
		do
			Precursor {ROSE_GRAPHICS_WIDGET} (a_direction)
			main_system.view_changed_actions.call ([])
			secondary_system.view_changed_actions.call ([])
		end

	on_mouse_click (a_x, a_y, a_button: INTEGER; a_x_tilt, a_y_tilt, a_pressure: DOUBLE; a_screen_x, a_screen_y: INTEGER)
			-- Mouse-key pressed.
		do

			if not is_picking and (a_button = 1) and selection.is_within_selected (a_x,a_y) then
				adapt_view_from_selection
				state.set_needs_projection
				selection.reset
				if small_widget /= void then
					small_widget.adapt_cutout_rectangle (current)
				end
			else
				precursor (a_x, a_y, a_button, a_x_tilt, a_y_tilt, a_pressure, a_screen_x, a_screen_y)
			end
		end

	adapt_view_from_selection
				-- adapt view after zooming
		do
			selection.normal_selected.set_y (1 - selection.normal_selected.y1, 1 - selection.normal_selected.y2)
			view := view.relative_rectangle (selection.normal_selected)
			main_system.view_changed_actions.call ([])
			secondary_system.view_changed_actions.call([])
		end

	adapt_view_from_dragging
			-- adapt view after dragging
		do
			Precursor
			main_system.view_changed_actions.call ([])
			secondary_system.view_changed_actions.call([])
		end


	axis_on_point (a_x, a_y: INTEGER): ROSE_CHART_AXIS
			-- Axis which has (`a_x', `a_y')
		do
			if main_system.x_axis.has_point (a_x,a_y) then
				Result := main_system.x_axis
			elseif main_system.y_axis.has_point (a_x,a_y) then
				Result := main_system.y_axis
			end
			if secondary_system.x_axis.has_point (a_x,a_y) then
				Result := main_system.x_axis
			elseif secondary_system.y_axis.has_point (a_x,a_y) then
				Result := main_system.y_axis
			end
		end

	remove_plots (a_plots: ARRAY [like plot_type])
			-- Remove plots in `a_plots' from the system
		local
			i: INTEGER
		do
			if a_plots /= Void then
				from
					i := a_plots.lower
				until
					i > a_plots.upper
				loop
					a_plots.item (i).remove
					i := i + 1
				end
			end
		end

	show_only_plots (a_plots: ARRAY [like plot_type])
			-- Hide all plots except the ones in `a_plots'
		do
			show_plots (all_plots,False)
			show_plots (a_plots,True)
		end

	show_plots (a_plots: ARRAY [like plot_type]; a_show: BOOLEAN)
			-- Show/Hide all plots in `a_plots'
		local
			i: INTEGER
		do
			if a_plots /= Void then
				from
					i := a_plots.lower
				until
					i > a_plots.upper
				loop
					a_plots.item (i).show (a_show)
					i := i + 1
				end
			end
		end

	all_plots: ROSE_LINEAR_ARRAY [like plot_type]
			-- List of all plots (`main_system' and `secondary_system' combined)
		do
			create Result.make_from_array (main_system.plots.to_array)
			Result.append_array (secondary_system.plots.to_array)
		end

	switch_equidistant (a_axis: ROSE_CHART_AXIS_INDEXED)
			-- Switch `a_axis' from equidistant to normal or vice versa
		do
			a_axis.set_equidistant (not a_axis.is_equidistant)
		end

	switch_logaritmic (a_axis: ROSE_CHART_AXIS)
			-- Switch `a_axis' from logarithmic to normal or vice versa
		do
			a_axis.set_logarithmic (not a_axis.is_logarithmic)
		end

	plots_in_selection (a_x, a_y: INTEGER): ROSE_LINEAR_ARRAY [like plot_type]
			-- Plots in mouse-selection area or near cursor
		local
			l_x1, l_y1, l_x2, l_y2: INTEGER
		do
			if selection.selected /= Void then
				l_x1 := selection.selected.x1.truncated_to_integer
				l_y1 := selection.selected.y1.truncated_to_integer
				l_x2 := selection.selected.x2.truncated_to_integer
				l_y2 := selection.selected.y2.truncated_to_integer
			else
				l_x1 := a_x - 4
				l_y1 := a_y - 4
				l_x2 := a_x + 4
				l_y2 := a_y + 4
			end
			Result := plots_in_rectangle (l_x1,l_y1,l_x2,l_y2)
		ensure
			not_void: Result /= Void
		end

	plots_in_rectangle (a_x1, a_y1, a_x2, a_y2: INTEGER): ROSE_LINEAR_ARRAY [like plot_type]
			-- List of all plots that are visible in rectangle ((a_x1, a_y1), (a_x2, a_y2)
		local
			l_rectangle: EV_RECTANGLE
			l_main: ROSE_LINEAR_ARRAY[ROSE_CHART_RENDER_PLOT[P1]]
			l_secondary: ROSE_LINEAR_ARRAY[ROSE_CHART_RENDER_PLOT[P2]]
		do
			create l_rectangle.make (a_x1, a_y1, (a_x2 - a_x1).abs.max (1), (a_y2 - a_y1).abs.max (1))
			create l_main.make_empty
			create l_secondary.make_empty
			main_system.add_plots_in_rectangle (l_main, l_rectangle)
			secondary_system.add_plots_in_rectangle (l_secondary, l_rectangle)
			create Result.make_from_array (l_main)
			Result.merge_references (l_secondary)
		ensure
			not_void: Result /= Void
		end

feature {NONE} -- Implementation

	plots_in_secondary_not_in_main: ROSE_LINEAR_ARRAY [like additional_plots_template]
			-- Color and names of plots in the secondary system to add to the legend
			-- Active only after calling 'add_to_legend'

	additional_plots_template: TUPLE [color: INTEGER; plot_name: STRING]
		--Template for information on plots in secondary not in main system

	small_widget: ROSE_CHART_MINI_WIDGET[P1, P2]
			-- widget which shows the whole plot, with marked cutout

	plot_type: ROSE_CHART_RENDER_PLOT [ROSE_CHART_POINT]
			-- Type of the render plot to be used

	project_background_figures
			-- project all background figures and zone lists
		do
			projector.project_zone_list (secondary_system.zone_figures_list, secondary_system)
			projector.project_zone_list (main_system.zone_figures_list, main_system)
			projector.project_list (secondary_system.background_figures_list, secondary_system)
			projector.project_list (main_system.background_figures_list, main_system)
		end

	projector: ROSE_CHART_PROJECTOR
			-- creates a projector
		once
			create Result
		end

	refresh_figures (a_system: ROSE_CHART_COORDINATE_SYSTEM[ROSE_CHART_POINT])
			-- Refresh EV_MODEL figures
		require
			known_system: a_system = main_system or a_system = secondary_system
			non_void_system: a_system /= Void
		local
			l_plot: like plot_type
			l_linkable: LINKABLE[like plot_type]
			l_index: INTEGER
			l_additional_plot_info: like additional_plots_template
		do
			a_system.x_axis.refresh_figures
			a_system.y_axis.refresh_figures
			from
				l_linkable := a_system.plots.first_element
			until
				l_linkable = Void
			loop
				l_plot := l_linkable.item
				l_plot.refresh_figures
				if l_plot.is_legend_shown then
					l_index := l_plot.index
					if a_system = secondary_system and main_system /= Void then
						l_index := l_index + main_system.plots.count
					end
					display.legend.add_explanation ([0, l_index], l_plot.name)
					display.legend.last_text.set_drop_actions (l_plot.plot.drop_actions)
				end
				l_linkable := l_linkable.right
			end
			if plots_in_secondary_not_in_main /= Void then
				from
					plots_in_secondary_not_in_main.start
				until
					plots_in_secondary_not_in_main.after
				loop
					l_additional_plot_info := plots_in_secondary_not_in_main.current_item
					display.legend.add_explanation ([0, l_additional_plot_info.color], l_additional_plot_info.plot_name)
					plots_in_secondary_not_in_main.forth
				end
			end
			display.refresh
		end

	has_secondary_plots: BOOLEAN
			-- Is there a secondary set of graphs.
		do
			Result := secondary_system.plots.count > 0
		end

	on_timeout
			-- Function called every 100 milliseconds.
			-- Well not really it seems, but it is called when necessary
		require else
			display_exists: display /= Void
			legend_exists: display.legend /= Void
			last_text_exists: display.legend.last_text /= Void
		local
			l_has_primary, l_has_secondary: BOOLEAN
		do
			if state.needs_projection then
				display.refresh
			end
			if state.needs_refresh then
				display.legend.wipe_out
				if external_legend /= Void then
					 external_legend.clear_legend
				end

				l_has_primary := main_system.plots.count > 0
				l_has_secondary := has_secondary_plots
				if l_has_primary then
					if l_has_secondary then
						if display.legend.orientation_horizontal then
							display.legend.add_sub_title ("1:")
						else
							display.legend.add_sub_title (display.legend.main_subtitle)
						end
						display.legend.last_text.set_drop_actions (main_system.plot_list.drop_actions)
					else
						display.legend.set_background_action (main_system.plot_list.drop_actions)
					end
				end
				refresh_figures (main_system)
				if l_has_secondary then
					if l_has_primary then
						display.legend.add_separator
						if display.legend.orientation_horizontal then
							display.legend.add_new_line
							display.legend.add_sub_title ("2:")
						else
							display.legend.add_sub_title (display.legend.secondary_subtitle)
						end
						display.legend.last_text.set_drop_actions (secondary_system.plot_list.drop_actions)
					end
					refresh_figures (secondary_system)
				end
			end
			secondary_system.refresh_part1
			main_system.refresh_part1

			project_background_figures

			secondary_system.refresh_part2
			main_system.refresh_part2

			display.refresh
		end

feature {NONE} -- Export to Excel

	export_to_excel
			-- Export data to excel
		local
			l_excel : ROSE_EXCEL_APPLICATION
			l_excel_tabs: ROSE_LINEAR_ARRAY [TUPLE [ARRAY2 [ANY], STRING_8]]
		do
			create l_excel_tabs.make_empty
			main_system.plot_list.do_all (agent export_plot_to_excel (l_excel_tabs, ?))
			--secondary_system.plot_list.do_all (agent export_secondary_plot_to_excel)
			create l_excel.make_single_workbook_general (l_excel_tabs, True)
		end

	export_plot_to_excel (a_result: ROSE_LINEAR_ARRAY [TUPLE [ARRAY2 [ANY], STRING]]; a_plot: ROSE_CHART_PLOT [P1])
			-- Export a plot (line or set) to excel
			-- Each plot is a tab in the result spreadsheet
			-- in the following format:
			--  label_1 value_1 label_2 value_2 [...]
			-- with some presentation heuristics:
			--  . if all values have the same label it becomes the value column header
			--  . empty/all zero columns are not shown
			--  . TODO: merge tabs when the data is of the same shape (e.g. identical value1)
		require
			a_result_not_void: a_result /= Void
			a_plot_not_void: a_plot /= Void
		local
			l_result: ROSE_LINEAR_ARRAY [ROSE_LINEAR_ARRAY [ANY]]
			l_array2: ROSE_ARRAY2 [ANY]
		do
			-- Process all lines/data sets in a plot
			create l_result.make_empty
			add_plot_dimension_to_excel (l_result, a_plot, 1)
			add_plot_dimension_to_excel (l_result, a_plot, 2)
			add_plot_dimension_to_excel (l_result, a_plot, 3) -- any 3D (or more?) charts?

			create l_array2.make_from_array_of_array (l_result)
			a_result.extend ([l_array2.transposed, a_plot.name])
		end

	add_plot_dimension_to_excel (a_result: ROSE_LINEAR_ARRAY [ROSE_LINEAR_ARRAY [ANY]]; a_plot: ROSE_CHART_PLOT [ROSE_CHART_POINT]; a_index: INTEGER)
			-- Add values for one dimension of a plot
		require
			a_result_not_void: a_result /= Void
			a_plot_not_void: a_plot /= Void
		local
			l_values_label: ANY
			l_labels, l_values, l_column: ROSE_LINEAR_ARRAY [ANY]
		do
			l_values_label := "values_" + a_index.out
			l_labels := a_plot.labels (a_index)
			l_values := a_plot.values (a_index)

			if not l_labels.is_empty and not l_labels.for_all (agent is_empty_string_or_void) then
				if is_single_reference (l_labels) then
						-- Use the unique label as header of values column
					l_values_label := l_labels.first
				else
						-- Separate columns for labels and values
					create l_column.make_empty
					l_column.extend ("labels_" + a_index.out)
					l_column.append_array (l_labels)
					a_result.extend (l_column)
				end
			end

			if not l_values.is_empty and not l_values.all_default then -- proxy for no values/ no dimension
				create l_column.make_empty
				l_column.extend (l_values_label)
				across l_values as i_value loop l_column.extend (i_value.item) end
				a_result.extend (l_column)
			end
		end

	is_empty_string_or_void (a_any: ANY): BOOLEAN
			-- Is `a_any' either void or the empty string?
		local
			l_string: STRING
		do
			l_string ?= a_any
			Result := a_any = Void or else
				(l_string /= Void and then l_string.is_empty)
		ensure
			void_true: a_any = Void implies Result
		end

	is_single_reference (a_list: ROSE_LINEAR_ARRAY [ANY]): BOOLEAN
			-- Are all items in list the same reference?
			-- (in the '=' sense)
		do
			Result := a_list.is_empty or else a_list.for_all (agent equal_agent (a_list.first, ?))
		ensure
			definition: Result = (a_list.is_empty or else a_list.for_all (agent equal_agent (a_list.first, ?)))
		end

	equal_agent (a, b: ANY): BOOLEAN
			-- 	Implementation for `is_single_reference'
		do
			Result := a = b
		ensure
			definition: Result = (a = b)
		end

end


