note

	description:

		"[
        Definition of a command - interface to command line like invocation

        Do not inherit directly from this command - inherit from RP_COMMAND.
       ]"

	author: "Zoran Simic"
	last_editor: "$author: $"
	date: "$Date: 2012/05/21 $"
	revision: "$Revision: #16 $"
	path: "$File: //rose/source/kernel/infrastructure/framework/command/rose_command.e $"

deferred class	ROSE_COMMAND

inherit

	ROSE_SERVICE

feature -- Initialization

	prepare is
			-- Prepare the command for future execution.
		do
			if parameters_signature = Void then
				create parameters_signature.make (id, signature)
				parameters_signature.set_strict (is_strict)
			end
		end

feature -- Access

	signature: STRING is
			-- Argument signature of this command (defines valid parameters to be passed to the command)
			-- See ROSE_COMMAND_PARAMETERS_SIGNATURE.make for info
		deferred
		end

	parameters_signature: ROSE_COMMAND_PARAMETERS_SIGNATURE
			-- Command parameters' signature parsed

	parameters: ROSE_COMMAND_PARAMETERS
			-- Current valued parameters associated to this command

	is_strict: BOOLEAN is
			-- Strict commands do not accept arguments by position, only named arguments
			-- Default: False, redefine to make your command strict
		do
			Result := False
		end

	is_open: BOOLEAN is
			-- Is this service currently open?
		do
			Result := parameters_signature /= Void and then parameters /= Void
		end

	help: STRING is
			-- Help on this command
		deferred
		end

	quick_help: STRING is
			-- 'help', but truncated to it's first line only (up to first eventual '%N')
			-- Default help message if none available
		local
			l_help: STRING
			i, j: INTEGER
		do
			l_help := help
			if l_help /= Void and then l_help.has ('%N') then
				from
					i := 1
				until
					Result /= Void or else i = 0 or else i > l_help.count
				loop
					j := l_help.index_of ('%N',i)
					if j > 0 then
						Result := l_help.substring (i,j-1)
						Result.left_adjust
						Result.right_adjust
						if Result.is_empty then
							Result := Void
						end
						i := j + 1
					else
						i := 0
					end
				end
			else
				Result := l_help
			end
			if Result = Void then
				Result := "-No help provided-"
			end
		ensure
			valid_result: Result /= Void
		end

	representation: STRING is
			-- String representation of this command
		do
			Result := clone (id)
			if parameters /= Void and then parameters.text /= Void then
				Result.extend (' ')
				Result.append (parameters.text)
			end
		end

feature -- Basic operations

	tie (a_object: ANY) is
			-- Tie this service to 'a_object' (this is service specific).
		do
			--| Not supported for commands
		ensure then
			dont_call: False
		end

	execute (a_parameters: STRING) is
			-- Execute this command with it's parameters.
		local
			rescued: STRING
		do
			if rescued /= Void then
				error.signal (id,"Command crashed: $1",<<rescued>>)
			elseif error.signaled then
				error.signal (id,"Can not execute command '$1' with error pending",<<id>>)
			else
				global.memory.collection_on
				open (a_parameters)
				if not error.signaled then
					execute_prepared
				end
			end
		rescue
			if rescued = Void then
				rescued := last_exception_explanation
			retry
			end
		end

	execute_unprotected (a_parameters: STRING) is
			-- Execute this command with it's parameters; Exceptions will not be caught.
		do
			if error.signaled then
				error.signal (id,"Can not execute command '$1' with error pending",<<id>>)
			else
				global.memory.collection_on
				open (a_parameters)
				if not error.signaled then
					check_prerequisites
					if not error.signaled then
						pre_execute
						if not error.signaled then
							do_execute
						end
						post_execute
					end
				end
			end
		end

	close is
			-- Close this service.
		do
		end

	check_prerequisites is
			-- Check that command's execution pre-requisites are met.
			-- Signal error if pre-requisistes are not met.
		deferred
		end

feature {ROSE_COMMAND_MANAGER, RCMD_EXECUTE, RCMD_MACRO, ROSE_PILOT_BASE} -- Implementation

	execute_prepared is
			-- Execute this command with its current 'parameters' (already prepared).
		require
			prepared: is_open
		local
			rescued: STRING
		do
			if rescued /= Void then
				error.signal (id,"Command crashed: $1",<<rescued>>)
			elseif error.signaled then
				error.signal (id,"Can not execute command '$1' with error pending",<<id>>)
			else
				check_prerequisites
				if not error.signaled then
					pre_execute
					if not error.signaled then
						protected_execute
					end
					post_execute
				end
			end
		rescue
			if rescued = Void then
				rescued := last_exception_explanation
				retry
			end
		end

feature {ROSE_COMMAND_MANAGER} -- Implementation

	open (a_path: STRING) is
			-- Open this service for 'a_path', 'a_path' is service specific.
			-- 'a_path' constitute the arguments for command-services.
		do
			prepare
			create parameters.make (parameters_signature, a_path)
			if parameters.parse_error /= Void then
				error.signal (id, parameters.parse_error, Void)
			end
		end

feature {NONE} -- Implementation

	pre_execute is
			-- Pre-execution actions.
		require
			service_open: is_open
		do
			error.broadcast (Global.convention.v_log,"$1 $2%N",<<id, parameters.text>>)
			start_time_measure (id, parameters.text)
		end

	protected_execute is
			-- Protected shell for 'do_execute'.
		require
			service_open: is_open
		local
			rescued: STRING
		do
			if rescued = Void then
				do_execute
			else
				error.signal (id,"Command execution crashed: $1",<<rescued>>)
			end
		rescue
			if rescued = Void then
				rescued := last_exception_explanation
				retry
			end
		end

	do_execute is
			-- Execute effectively the command.
		require
			service_open: is_open
		deferred
		end

	post_execute is
			-- Post-execution actions.
		require
			service_open: is_open
		do
			end_time_measure
		end

	write_out_command_line is
			-- Write out command line
		do
			if parameters /= Void and then parameters.text /= Void then
				inform ("command", id + " " + parameters.text, Void)
			end
		end

end
