indexing
	description: "Command manager, allowing to store and execute commands"
	author: "Zoran Simic"
	date: "$Date: 2008/12/01 $"
	revision: "$Revision: #11 $"

class
	ROSE_COMMAND_MANAGER

inherit
	ROSE_SERVICE_MANAGER [ROSE_COMMAND]
		rename
			make as make_service_manager
		redefine
			add
		end

create
	make

feature {NONE} -- Initialization

	make (a_id: STRING) is
			-- Create an empty command manager.
		do
			make_service_manager (a_id, Void)
			create categories.make (0)
			create services.make (0)
		end

feature -- Access

	categories: HASH_TABLE [ROSE_SERVICE_MANAGER [ROSE_COMMAND], STRING]
			-- Command categories (for help grouping)

	category (a_id: STRING): ROSE_SERVICE_MANAGER [ROSE_COMMAND] is
			-- Category associated to 'a_id', or Void if none
		do
			Result := categories.item (a_id)
		end

	command_category (a_id: STRING): ROSE_SERVICE_MANAGER [ROSE_COMMAND] is
			-- Category in which command with 'a_id' appears
			-- Void if none
		do
			from
				categories.start
			until
				Result /= Void or else categories.off
			loop
				if categories.item_for_iteration.has (a_id) then
					Result := categories.item_for_iteration
				end
				categories.forth
			end
		end

	alias_value (a_id: STRING): STRING is
			-- Alias associated to 'a_id', or Void if none
		local
			a: RCMD_ALIAS_MACRO
		do
			a ?= services.item (a_id)
			if a /= Void then
				Result := a.signature
			end
		end

	parsed_command (a_string: STRING): ROSE_COMMAND is
			-- Parsed a single command ('command_separators' ignored) contained in 'a_string'
		local
			p: PAIR [STRING, STRING]
		do
			if a_string /= Void and then not a_string.is_empty then
				p := Global.formatter.split_on_first (a_string, Global.convention.argument_separators, Global.convention.command_escapes)
				if p.second /= Void then
					p.second.left_adjust
					if equal (p.second, "?") then
						Result := parsed_command ("show "+p.first)
					end
				end
				if Result = Void then
					Result := service (p.first)
					if Result = Void then
						error.signal (Global.app.program_name, "Unknown command '$1'", <<p.first>>)
					else
						Result.open (Global.formatter.unescaped (p.second, Global.convention.command_escapes))
					end
				end
			end
		ensure
			error_signaled: (a_string /= Void and then not a_string.is_empty and then Result = Void) implies error.signaled
		end

	macro_command (a_line: STRING): RCMD_MACRO is
			-- Macro-command representing command in 'a_line'
			-- Void if 'a_line' is not correctly interpreted
			-- Commands are separated by characters in 'command_separators'
		local
			seq: SEQUENCE [STRING]
			l_string: STRING
			l_cmd: ROSE_COMMAND
		do
			create Result.make_anonymous
			if a_line /= Void and then not a_line.is_empty then
				from
					seq := Global.formatter.split (a_line, Global.convention.command_separators, Global.convention.command_escapes)
					seq.start
				until
					Result = Void or else seq.off
				loop
					l_string := seq.item
					if l_string /= Void then
						l_string.left_adjust
							--| Lines starting with a '#' are considered comments
						if not l_string.is_empty and then l_string.item (1) /= '#' then
							l_cmd := parsed_command (l_string)
							if l_cmd = Void then
								Result := Void
							else
								Result.add_last (l_cmd)
							end
						end
					end
					seq.forth
				end
			end
			if Result /= Void and then not error.signaled then
				Result.open (Void)
			end
		ensure
			error_signaled: Result = Void implies error.signaled
		end

feature -- Status report

	has_category (a_id: STRING): BOOLEAN is
			-- Is category 'a_id' defined ?
		do
			Result := categories.has (a_id)
		end

	is_command (a_id: STRING): BOOLEAN is
			-- Is command 'a_id' defined, without beeing an alias ?
		do
			Result := has (a_id) and then not is_alias (a_id)
		end

	is_alias (a_id: STRING): BOOLEAN is
			-- Is 'a_id' an aliased command ?
		do
			Result := alias_value (a_id) /= Void
		end

	is_default_category (a_category_name: STRING): BOOLEAN is
			-- Is 'a_category_name' referring to a default category?
		do
			Result := equal (a_category_name, default_category)
				or else equal (a_category_name, alias_category)
				or else equal (a_category_name, gui_category)
		end

feature -- Measurement

	category_count: INTEGER is
			-- Number of categories in this manager
		do
			Result := categories.count
		end

feature -- Status setting

	set_alias (a_id, a_alias: STRING) is
			-- Alias 'a_id' to mean 'a_alias'.
		require
			valid_id: a_id /= Void and then not a_id.is_empty
			not_built_in: not is_command (a_id)
			valid_alias: a_alias /= Void
		local
			c: ROSE_COMMAND
			a: RCMD_ALIAS_MACRO
		do
			c := services.item (a_id)
			if c = Void then
				add_command (alias_category, create {RCMD_ALIAS_MACRO}.make (a_id, a_alias))
			else
				a ?= c
				if a = Void then
					--| Error: can't alias an existing command
					--| This should be detected by client (see require)
				else
					a.set_signature (a_alias)
				end
			end
		end

feature -- Basic operations

	execute (a_line: STRING) is
			-- Execute 'a_line' (multiple commands allowed, separated by characters in 'command_escapes').
		local
			l_macro: RCMD_MACRO
		do
			l_macro := macro_command (a_line)
			if l_macro /= Void and then not error.signaled then
				l_macro.execute_prepared
			end
		end

	add (a_command: ROSE_COMMAND) is
			-- Add 'a_command' to 'default_category'.
		do
			add_command (default_category, a_command)
		end

	add_command (a_category: STRING; a_command: ROSE_COMMAND) is
			-- Add 'a_command' in 'a_category'.
		require
			valid_category: a_category /= Void
			valid_command: a_command /= Void
		local
			l_category: ROSE_SERVICE_MANAGER [ROSE_COMMAND]
			l_id: STRING
		do
			a_command.prepare
			l_id := a_command.id
			l_id.to_lower
			services.force (a_command, l_id)
			l_category := category (a_category)
			if l_category = Void then
				create l_category.make (a_category, Void)
				categories.put (l_category, a_category)
			end
			l_category.add (a_command)
		end

feature -- Constants

	default_category: STRING is "default"
			-- Default command category name

	gui_category: STRING is "gui"
			-- GUI command category name

	alias_category: STRING is "alias"
			-- 'Alias' command category name

	application_category: STRING is "application"
			-- Application command category name

end
