note

	description:

		"Single command parameter (name-value item)"

	author: "Zoran Simic"
	last_editor: "$Author: c4 $"
	date: "$Date: 2011/11/29 $"
	revision: "$Revision: #18 $"
	path: "$File: //rose/source/kernel/infrastructure/framework/command/rose_command_parameter_item.e $"

class	ROSE_COMMAND_PARAMETER_ITEM

inherit
	
	ROSE_FRAMEWORK_CONVENTION

	ROSE_SHARED_GLOBAL
		export {NONE} all end

create

	make

feature {NONE} -- Initialization

	make (a_name, a_default_value: STRING; a_type: CHARACTER; a_optional: BOOLEAN) is
			-- Create a parameter item
		do
			name := a_name
			default_value := a_default_value
			type := a_type
			is_optional := a_optional
		end

feature -- Access

	name: STRING
			-- Name of the parameter

	type: CHARACTER
			-- Parameter type

	value: STRING
			-- Value for this parameter (Void if not explicitly provided)

	default_value: STRING
			-- Default value for this parameter

	to_character: CHARACTER is
			-- Character value of the parameter
		require
			character_or_zone_type: (type = command_parameter_type_character) or (type = command_parameter_type_equity_zone) or
				(type = command_parameter_type_zone)
		do
			if value /= Void and then not value.is_empty then
				Result := value.item (1)
			end
		end

	to_integer: INTEGER is
			-- Integer value of the parameter
		require
			integer_type: type = command_parameter_type_integer
		do
			if value /= Void and then value.is_integer then
				Result := value.to_integer
			end
		end

	to_boolean: BOOLEAN is
			-- Boolean value of the parameter
		require
			boolean_type: type = command_parameter_type_boolean
		do
			if value = Void then
				--| Default: False
			elseif value.is_boolean then
				Result := value.to_boolean
			elseif value.is_integer then
				Result := value.to_integer /= 0
			else
				Result := boolean_values.item (value.as_lower)
			end
		end

	to_double: DOUBLE is
			-- Double value of the parameter
		require
			double_type: type = command_parameter_type_double
		do
			if value = Void then
				--| Default: 0.0
			elseif value.is_double then
				Result := value.to_double
			elseif value.is_integer then
				Result := value.to_integer
			end
		end

	to_date: ROSE_DATE is
			-- Date value of the parameter
		require
			date_type: type = command_parameter_type_date
		do
			Result := formatter.make_date_from_string (value)
		end

	to_date_time: ROSE_DATE is
			-- Date-time value of the parameter
		require
			date_time_type: type = command_parameter_type_date_time
		local
			l_date: ROSE_DATE
		do
			l_date := formatter.make_date_from_string (value)
			create Result.make_date_time (l_date.year, l_date.month, l_date.day, l_date.hour, l_date.minute, l_date.second)
		end

	to_month: ROSE_DATE is
			-- Month value of the parameter
		require
			date_type: type = command_parameter_type_month
		do
			Result := formatter.make_month_from_string (value)
		end

feature -- Status report

	valid_value (a_value: STRING): BOOLEAN is
			-- Is 'a_value' valid for this parameter type ?
		do
			if a_value = Void then
				Result := is_optional or type = command_parameter_type_boolean
			elseif a_value.is_empty then
				Result := type = command_parameter_type_boolean or (is_optional and type = command_parameter_type_string)
			else
				inspect type
				when command_parameter_type_integer then
					Result := a_value.is_integer
				when command_parameter_type_boolean then
					Result := a_value.is_boolean or else a_value.is_integer or else boolean_values.has (a_value.as_lower)
				when command_parameter_type_double then
					Result := a_value.is_double or else a_value.is_integer
				when command_parameter_type_date then
					Result := formatter.make_date_from_string (a_value) /= Void
				when command_parameter_type_date_time then
					Result := formatter.make_date_from_string (a_value) /= Void
				when command_parameter_type_month then
					Result := formatter.make_month_from_string (a_value) /= Void
				when command_parameter_type_character then
					Result := a_value.count = 1
				when command_parameter_type_equity_zone then
					Result := (a_value.count = 1) and then
						("CEJOSU").has (a_value.item (1).as_upper)
				when command_parameter_type_zone then
					Result := (a_value.count = 1) and then
						("CEJOSUTWFI").has (a_value.item (1).as_upper)
				else
					Result := a_value /= Void
				end
			end
		end

	is_optional: BOOLEAN
		-- Is this parameter optional ?

	is_set: BOOLEAN is
			-- Is this parameter explicitly set ?
		do
			Result := value /= Void and then value /= default_value
		end

	is_default: BOOLEAN is
			-- Has this parameter got its default value ?
		do
			Result := default_value /= Void and then value = default_value
		end

	is_boolean: BOOLEAN is
			-- Does current represent a boolean-valued option?
		do
			Result := type = command_parameter_type_boolean
		end

	type_name: STRING is
			-- Name of current parameter type
		do
			Result := command_parameter_type_representation (type)
		end

feature -- Status setting

	set_default is
			-- Reset parameter
		do
			if valid_value (default_value) then
				set_value (default_value)
			else
				error.signal (generator, "Default value of $1 is not valid for objects of type $2.",
					<<default_value, type_name>>)
				error.flush
			end
		end

	set_value (a_value: STRING) is
			-- Set 'value' to 'a_value'.
		require
			valid_type_value: valid_value (a_value)
		do
			if (a_value = Void or else a_value.is_empty) and then type = command_parameter_type_boolean then
				value := "true"
			else
				value := a_value
				if value /= Void and then value.count > 2
				and then value.item (1) = '"' and then value.item (value.count) = '"' then
					value.remove (value.count)
					value.remove (1)
				end
			end
		end

	set_optional (a_optional: BOOLEAN) is
			-- Set 'is_optional' status of this parameter item.
		do
			is_optional := a_optional
		end

feature {NONE} -- Implementation

	boolean_values: HASH_TABLE [BOOLEAN, STRING] is
			-- Translations of yes/no strings to equivalent boolean values
		once
			create Result.make (0)
			Result.put (True, "y")
			Result.put (True, "yes")
			Result.put (False, "n")
			Result.put (False, "no")
		ensure
			boolean_values_not_void: Result /= Void
		end

end
