indexing
	description: "Set of command parameters (name-value pairs)"
	author: "Zoran Simic"
	date: "$Date: 2010/02/05 $"
	revision: "$Revision: #24 $"
	path: "$File: //rose/source/kernel/infrastructure/framework/command/rose_command_parameters.e $"

class
	ROSE_COMMAND_PARAMETERS

inherit
	ROSE_SHARED_GLOBAL
		export
			{NONE} all
		end
	DEBUG_OUTPUT

create
	make

feature {NONE} -- Initialization

	make (a_signature: ROSE_COMMAND_PARAMETERS_SIGNATURE; a_text: STRING) is
			-- Create command parameters bound to 'a_signature', representing 'a_text'.
		require
			valid_signature: a_signature /= Void
		do
			signature := a_signature
			if a_text = Void then
				text := ""
			else
				text := a_text
			end
			parse
		end

feature -- Access

	signature: ROSE_COMMAND_PARAMETERS_SIGNATURE
			-- Parsed signature of current parameters

	text: STRING
			-- Textual representation of parameters

	parse_error: STRING
			-- Signaled parse error, if any

	parameter (a_name: STRING): ROSE_COMMAND_PARAMETER_ITEM is
			-- Parameter associated to 'a_name'
		do
			if params /= Void then
				Result := params.item (a_name)
			end
		end

	value (a_name: STRING): STRING is
			-- String value of parameter with 'a_name'
			-- Void if none
		local
			l_item: ROSE_COMMAND_PARAMETER_ITEM
		do
			l_item := parameter (a_name)
			if l_item /= Void then
				Result := l_item.value
			end
		end

	comma_separated_value (a_name: STRING): ROSE_LINEAR_ARRAY [STRING]
			-- Comma separated `value' split up into individual elements
		require
			a_name_not_void: a_name /= Void
		local
			l_string: STRING
		do
			l_string := value (a_name)
			if l_string /= Void then
				create Result.make_from_linear (l_string.split (','))
			else
				create Result.make_empty
			end
		ensure
			not_void: Result /= Void
			no_value_empty: value (a_name) = Void implies Result.is_empty
		end

	character (a_name: STRING): CHARACTER is
			-- Character value of parameter with 'a_name'
			-- Default: %U
		local
			l_item: ROSE_COMMAND_PARAMETER_ITEM
		do
			l_item := parameter (a_name)
			if l_item /= Void then
				Result := l_item.to_character
			end
		end

	integer (a_name: STRING): INTEGER is
			-- Integer value of parameter with 'a_name'
			-- Default: 0
		local
			l_item: ROSE_COMMAND_PARAMETER_ITEM
		do
			l_item := parameter (a_name)
			if l_item /= Void then
				Result := l_item.to_integer
			end
		end

	boolean (a_name: STRING): BOOLEAN is
			-- Boolean value of parameter with 'a_name'
			-- Default: False
		local
			l_item: ROSE_COMMAND_PARAMETER_ITEM
		do
			l_item := parameter (a_name)
			if l_item /= Void then
				Result := l_item.to_boolean
			end
		end

	double (a_name: STRING): DOUBLE is
			-- Double value of parameter with 'a_name'
			-- Default: 0.0
		local
			l_item: ROSE_COMMAND_PARAMETER_ITEM
		do
			l_item := parameter (a_name)
			if l_item /= Void then
				Result := l_item.to_double
			end
		end

	date (a_name: STRING): ROSE_DATE is
			-- Date value of parameter with 'a_name'
			-- Void if none
		local
			l_item: ROSE_COMMAND_PARAMETER_ITEM
		do
			l_item := parameter (a_name)
			if l_item /= Void then
				Result := l_item.to_date
			end
		end

	date_time (a_name: STRING): ROSE_DATE is
			-- Date value of parameter with 'a_name'
			-- Void if none
		local
			l_item: ROSE_COMMAND_PARAMETER_ITEM
		do
			l_item := parameter (a_name)
			if l_item /= Void then
				Result := l_item.to_date_time
			end
		end

	month (a_name: STRING): ROSE_DATE is
			-- Date value of parameter with 'a_name'
			-- Void if none
		local
			l_item: ROSE_COMMAND_PARAMETER_ITEM
		do
			l_item := parameter (a_name)
			if l_item /= Void then
				Result := l_item.to_month
			end
		end

	is_defined_in_signature (a_name: STRING): BOOLEAN is
			-- Is parameter 'a_name' defined in 'signature'?
		do
			Result := parameter (a_name) /= Void
		end

	has, provided (a_name: STRING): BOOLEAN is
			-- Has parameter 'a_name' been explicitly provided (not defaulted)?
		local
			l_param: ROSE_COMMAND_PARAMETER_ITEM
		do
			l_param := parameter (a_name)
			Result := l_param /= Void and then l_param.is_set
		end

	parameter_names: ROSE_LINEAR_ARRAY[STRING]
			-- List of the names of the parameters that were found.
		do
			create Result.make_empty

			from
				params.start
			until
				params.off
			loop
				Result.extend(params.item_for_iteration.name)
				params.forth
			end
		end

	count: INTEGER is
			-- Number of arguments provided
		do
			if params /= Void then
				Result := params.count
			end
		end

	debug_output: STRING is
			-- String that should be displayed in debugger to represent `Current'.
		do
			Result := text
		end

feature -- Basic operations

	add (a_item: ROSE_COMMAND_PARAMETER_ITEM) is
			-- Add 'a_item' to Current.
		require
			valid_item: a_item /= Void
		do
			params.force (a_item,a_item.name)
		end

feature {NONE} -- Implementation

	params: HASH_TABLE [ROSE_COMMAND_PARAMETER_ITEM,STRING]
		-- Actual parameters

	parse is
			-- Parse current parameters.
		local
			s: SEQUENCE [STRING]
			l_list: LINKED_LIST [ROSE_COMMAND_PARAMETER_ITEM]
			l_item: ROSE_COMMAND_PARAMETER_ITEM
		do
			parse_error := Void
			if signature.parameters_list /= Void then
				l_list := signature.parameters_list.twin
				create params.make (l_list.count)
				if text /= Void then
					from
						s := Global.formatter.split (text,Global.convention.argument_separators, global.convention.command_escapes)
						l_list.start
						s.start
					until
						s.off
					loop
						if not s.item.is_empty then
							-- This conditional is to tolerate multiple spaces in parameters...
							fill_value (s.item,l_list)
						end
						s.forth
					end
				else
					from
						l_list.start
					until
						l_list.off
					loop
						if l_list.item.is_optional and then l_list.item.default_value /= Void then
							l_item := l_list.item.twin
							l_item.set_value (l_item.default_value)
							add (l_item)
						end
						l_list.forth
					end
				end
				if parse_error = Void then
					signature.checkup (Current)
					parse_error := signature.parse_error
				end
			end
		end

	fill_value (a_value: STRING; a_list: LINKED_LIST [ROSE_COMMAND_PARAMETER_ITEM]) is
			-- Fill-in 'a_value' to Current, using 'a_list' for implicit parameters.
		require
			valid_value: a_value /= Void and then not a_value.is_empty
			no_bypass: a_list /= Void
		local
			p: PAIR [STRING,STRING]
			l_item: ROSE_COMMAND_PARAMETER_ITEM
			l_name, l_value: STRING
		do
			l_value := adapted_value (a_value)
			if l_value.count > 1 and then l_value.item (1) = '-' then
				p := Global.formatter.split_on_first (l_value, argument_assigners, Void)
				l_name := p.first.substring (2,p.first.count)
				l_value := adapted_value (p.second)
				l_item := signature.item (l_name)
				if l_item = Void then
					report_error ("Parameter '$1' is not recognized",<<l_name,l_value>>)
				elseif not l_item.valid_value (l_value) then
					report_error ("Parameter '$1' (type $2) is given an invalid value '$3'",<<l_name,l_item.type_name,l_value>>)
				elseif l_item.is_set then
					report_error ("Parameter '$1' is given a second value '$3'",<<l_name,l_value>>)
				end
			elseif signature.is_strict then
				report_error ("Invalid argument '$1', arguments must be given by name (of the form 'name=value')", <<l_value>>)
			elseif l_value.has ('=') then
				report_error ("Character '=' is not allowed in a command argument given by position (you must specify argument name, of the form -name=value), can not accept argument '$1'", <<l_value>>)
			else
				l_item := next_implicit_parameter (a_list)
				if l_item = Void then
					report_error ("Can not implicitly assign to any parameter value '$1'",<<l_value>>)
				elseif not l_item.valid_value (l_value) then
					report_error ("Parameter '$1' (type $2) is given an invalid implicit value '$3'",<<l_item.name,l_item.type_name,l_value>>)
				end
			end
			if parse_error = Void then
				l_item.set_value (l_value)
				add (l_item)
			end
		end

	adapted_value (a_value: STRING): STRING is
			-- Return actual parameter value
		do
			if a_value = Void then
				create Result.make (0)
			else
				Result := a_value
				if Result.count > 1 and then Result.item (1) = '"' and then Result.item (Result.count) = '"' then
					Result := Result.substring (2,Result.count-1)
				end
			end
		end

	next_implicit_parameter (a_list: LINKED_LIST [ROSE_COMMAND_PARAMETER_ITEM]): ROSE_COMMAND_PARAMETER_ITEM is
			-- Next parameter that could implicitly be granted a value.
		require
			no_bypass: a_list /= Void
		do
			from
			until
				a_list.off or else is_valid_implicit (a_list.item)
			loop
				a_list.forth
			end
			if not a_list.off and then a_list.item /= Void then
				Result := a_list.item.twin
			end
		end

	is_valid_implicit (a_item: ROSE_COMMAND_PARAMETER_ITEM): BOOLEAN is
			-- Is 'a_item' a valid implicit argument ?
		do
			Result := not params.has (a_item.name)
				and then (a_item.is_optional implies a_item.default_value = Void)
		end

	report_error (a_message: STRING; a_args: ARRAY [STRING]) is
			-- Report error with 'a_message' and 'a_args'
		do
			if a_message /= Void then
				if parse_error = Void then
					create parse_error.make_empty
				end
				parse_error.append (Global.formatted (a_message, a_args))
				parse_error.append_character ('%N')
			end
		end

	argument_assigners: STRING is ":="
			-- Characters considered as 'assigners' (ie delimit name and value)

invariant

	valid_signature: signature /= Void
	valid_text: text /= Void

end
