indexing
	description: "Command parameters (accepted name-value pairs declaration, with their respective expected data types)"
	author: "Zoran Simic"
	date: "$Date: 2010/04/28 $"
	revision: "$Revision: #14 $"

class
	ROSE_COMMAND_PARAMETERS_SIGNATURE

inherit
	ROSE_SHARED_GLOBAL
		export
			{NONE} all
		end
	DEBUG_OUTPUT

create
	make

feature {NONE} -- Initialization

	make (a_id, a_signature: STRING) is
			-- Make current named parameters accordingly to 'a_signature'.
			-- Syntax : {<name> <qualifier> <type> [<default value>]}
			-- <name> is an arbitrary parameter name (alphanumeric)
			-- <qualifier> is '=' for a required parameter, ':' for an optional one
			-- <type>: see ROSE_COMMAND_PARAMETER_CONSTANTS.t_* ('s' string, 'i' integer, '?' boolean, 'r' double, 'd' date, 't' date-time)
			-- <default value> is the default value for the parameter (if optional)
		require
			valid_signature: a_signature /= Void
		local
			l_defs: SEQUENCE [STRING]
		do
			id := a_id
			signature := a_signature
			if equal (a_signature,"*") then
				-- Accept everything
				is_unsigned := True
			else
				create parameters_list.make
				create signature_table.make (0)
				from
					l_defs := Global.formatter.split (a_signature," %T",Void)
					l_defs.start
				until
					l_defs.off
				loop
					if not l_defs.item.is_empty then
						-- Tolerate multiple spaces in signatures...
						add_signature (l_defs.item)
					end
					l_defs.forth
				end
				if parse_error /= Void then
					fail ("make", "Bad parameter signature '$1':%N$2",<<a_signature, parse_error>>)
				end
			end
		end

feature -- Access

	id: STRING
		-- Parameters identification (associated command id)

	signature: STRING
		-- Original signature of these parameters

	parameters_list: LINKED_LIST [ROSE_COMMAND_PARAMETER_ITEM]
		-- All accepted parameters

	parameter_from_name (a_name : STRING) : ROSE_COMMAND_PARAMETER_ITEM is
			-- Parameter from name
		require
			a_name_not_void : a_name /= Void
		do
			Result := signature_table [a_name]
		end

	is_strict: BOOLEAN
			-- Strict commands do not accept arguments by position, only named arguments

	item (a_id: STRING): ROSE_COMMAND_PARAMETER_ITEM is
			-- Parameter item corresponding to 'a_id'
			-- Void if none
		do
			if signature_table /= Void then
				Result := signature_table.item (a_id)
				if Result /= Void then
					Result := Result.twin
				end
			end
		end

	is_empty: BOOLEAN is
			-- Is this signature empty
		do
			Result := parameters_list /= Void and then parameters_list.is_empty
		end

	is_unsigned: BOOLEAN
			-- Is this signature unsigned ?

	parse_error: STRING
			-- Signaled parse error, if any

	debug_output: STRING
			-- String that should be displayed in debugger to represent `Current'.
		do
			Result := signature
		end

feature -- Basic operations

	checkup (p: ROSE_COMMAND_PARAMETERS) is
			-- Check that parameters in 'p' satisfy this defintion,
			-- and that all mandatory parameters are specified.
		local
			l_item: ROSE_COMMAND_PARAMETER_ITEM
		do
			parse_error := Void
			if parameters_list /= Void then
				from
					parameters_list.start
				until
					parameters_list.off
				loop
					if not p.has (parameters_list.item.name) then
						if not parameters_list.item.is_optional then
							report_error ("Mandatory parameter '$1' not specified",<<parameters_list.item.name>>)
						elseif parameters_list.item.default_value /= Void then
							l_item := parameters_list.item.twin
							l_item.set_default
							p.add (l_item)
						end
					end
					parameters_list.forth
				end
			end
		end

	set_strict (a_is_strict: BOOLEAN) is
			-- Set 'is_strict'.
		do
			is_strict := a_is_strict
		end

feature {NONE} -- Implementation

	signature_table: HASH_TABLE [ROSE_COMMAND_PARAMETER_ITEM,STRING]
		-- Actual parameters hashed by their name

	add_signature (a_signature: STRING) is
			-- Allow 'a_signature' as parameter
			-- format: <name> <qualifier> <type> <default value>.
		require
			valid_signature: a_signature /= Void
			no_bypass: parameters_list /= Void
		local
			p: PAIR [STRING,STRING]
			l_name, l_default_value: STRING
			l_is_optional: BOOLEAN
			l_type: CHARACTER
		do
			p := Global.formatter.split_on_first (a_signature,Global.convention.all_command_parameter_qualifiers,Void)
			if p.first.is_empty or else p.second = Void or else p.second.is_empty then
				report_error ("Invalid parameter signature '$1'",<<a_signature>>)
			end
			if parse_error = Void and then not Global.convention.is_valid_command_parameter_type (p.second.item (1)) then
				report_error ("Invalid parameter type in signature '$1'",<<a_signature>>)
			end
			if parse_error = Void and then signature_table.has (p.first) then
				report_error ("Parameter '$1' defined twice in signature for '$2'",<<p.first,id>>)
			end
			if parse_error = Void then
				l_name := p.first
				l_is_optional := a_signature.item (l_name.count+1) = Global.convention.command_parameter_qualifier_optional
				l_type := p.second.item (1)
				if p.second.count > 1 then
					l_default_value := p.second.substring (2, p.second.count)
				end
				if not l_is_optional and then l_default_value /= Void then
					report_error ("Mandatory parameter '$1' should not have a default value",<<l_name>>)
				end
			end
			if parse_error = Void then
				add_item (create {ROSE_COMMAND_PARAMETER_ITEM}.make (l_name,l_default_value,l_type,l_is_optional))
			end
		end

	add_item (a_item: ROSE_COMMAND_PARAMETER_ITEM) is
			-- Add 'a_item' to accepted parameters.
		require
			valid_item: a_item /= Void
			no_bypass: parameters_list /= Void
		do
			signature_table.force (a_item,a_item.name)
			parameters_list.extend (a_item)
		end

	report_error (a_message: STRING; a_args: ARRAY [STRING]) is
			-- Report error with 'a_message' and 'a_args'
		do
			if a_message /= Void then
				if parse_error = Void then
					create parse_error.make_empty
				end
				parse_error.append (Global.formatted (a_message, a_args))
			end
		end

end
