indexing
	description: "View on command parameters"
	author: "Zoran Simic"
	date: "$Date: 2014/02/26 $"
	revision: "$Revision: #7 $"

class
	ROSE_COMMAND_PARAMETERS_VIEW

inherit
	ROSE_SHARED_GLOBAL_GUI

create
	make

feature {NONE} -- Initialization

	make (a_container: EV_CONTAINER; a_command: like command; a_validator: like validator) is
			-- Create a view on 'a_command' parameters and put it in 'a_container' if provided.
			-- 'a_validator' is called to tell whether currently selected parameters are valid.
		local
			l_signature: ROSE_COMMAND_PARAMETERS_SIGNATURE
			l_widget: EV_WIDGET
		do
			command := a_command
			command.prepare
			box := Gui.factory.new_vbox (a_container, 2, 2)
			l_signature := command.parameters_signature
			if l_signature = Void then
				l_widget := Gui.factory.new_label (box,"%NNo signature%N%N",1)
				l_widget.set_foreground_color (Gui.rep.colors.Dark_green)
			elseif l_signature.is_empty then
				l_widget := Gui.factory.new_label (box,"%NNo parameters needed%N%N",1)
				l_widget.set_foreground_color (Gui.rep.colors.Dark_green)
			elseif l_signature.is_unsigned then
				l_widget := Gui.factory.new_label (box,"%NParameters parsed by command%N%N",1)
				l_widget.set_foreground_color (Gui.rep.colors.Dark_green)
				raw_parameters := Gui.factory.new_text_field (box,"",Void,Void,1)
			else
				create parameter_list.make
				l_signature.parameters_list.do_all (agent add_parameter (box,?))
			end
			controls_box := gui.factory.new_hbox (box, 0, 0)
			create parameter_sets_combo.make_with_strings (remembered_parameter_sets)
			parameter_sets_combo.select_actions.extend (agent refresh_parameters_from_combo)
			parameter_sets_combo.return_actions.extend (agent refresh_parameters_from_combo)
			controls_box.extend (parameter_sets_combo)
			launch_button := gui.factory.new_button (controls_box, "Launch", agent launch, 50)
			controls_box.hide
			validator := a_validator
		end

feature -- Access

	box: EV_BOX
			-- Box holding the whole view

	controls_box: EV_HORIZONTAL_BOX
			-- Box holding controls

	launch_button: EV_BUTTON
			-- Launches `command'

	parameter_sets_combo: ROSE_COMBO_BOX
			-- Displays recent parameter set used to launch `command'

	command: ROSE_COMMAND
			-- Command viewed

	validator: PROCEDURE [ANY, TUPLE [BOOLEAN]]
			-- Called to tell whether curerntly selected parameters are valid

	value: STRING is
			-- Parameters entered
		local
			l_string: STRING
		do
			if parameter_list /= Void then
				from
					create Result.make (0)
					parameter_list.start
				until
					parameter_list.off
				loop
					l_string := parameter_list.item.value
					if l_string /= Void then
						Result.append (l_string)
					end
					parameter_list.forth
					if not parameter_list.off and then l_string /= Void then
						Result.extend (Global.convention.argument_separators.item (1))
					end
				end
			elseif raw_parameters /= Void then
				Result := raw_parameters.text
			else
				create Result.make (0)
			end
		end

	parameters_string: STRING is
			-- String ready to pass to `launch'
		local
			l_parameter: ROSE_COMMAND_PARAMETER_CELL
			i: INTEGER
		do
			Result := ""
			from i := 1 until i > parameter_list.count loop
				l_parameter := parameter_list.i_th (i)
				if l_parameter /= Void and then l_parameter.value /= Void then
					Result.append (l_parameter.value + " ")
				end
				i := i + 1
			end
		end

	remembered_parameter_sets: ROSE_LINEAR_ARRAY [STRING] is
			-- Remembered parameter sets
		local
			l_string: STRING
		do
			create Result.make_from_array (<<"">>)
			l_string := remembered_parameter_sets_string
			if l_string /= Void then
				Result.append_array (gui.formatter.split_array (l_string, ",", Void, False))
			end
		end

	remember_parameter_set is
			-- Remember current parameter set
		local
			l_string: STRING
			l_remembered_sets: ROSE_LINEAR_ARRAY [STRING]
			i :INTEGER
		do
			if is_valid_parameters then
				l_string := parameter_sets_combo.text
				l_remembered_sets := remembered_parameter_sets
				from i := l_remembered_sets.count.min (k_max_remembered_parameter_sets) until i < 1 loop
					l_string.append ("," + l_remembered_sets [i])
					i := i - 1
				end
				global.user.set_remembered_value (remembered_sets_key, l_string)
			end

		end

	remembered_parameter_sets_string: STRING is
			-- Saved comma-delimited sets of recently-used paramters
		do
			Result := global.user.remembered_value (remembered_sets_key)
		end

	remembered_sets_key: STRING is
			-- Key for retrieval for remembered parameter sets
		local
			l_name: STRING
		do
			l_name := command.id
			l_name.replace_substring_all (" ", "_")
			Result := k_parameter_sets_string + "/" + l_name
		end

	load_parameter_sets is
			-- Load remembered_parapmeter sets
		do
			parameter_sets_combo.set_strings (remembered_parameter_sets)
		end


feature -- Basic operations

	set_parameters (a_params: STRING) is
			-- Set shown command parameters with contents of 'a_params'.
		local
			l_list: like parameter_list
			l_params: ROSE_COMMAND_PARAMETERS
			l_cell: ROSE_COMMAND_PARAMETER_CELL
			l_value: STRING
		do
			create l_params.make (command.parameters_signature, a_params)
			from
				l_list := parameter_list.twin
				l_list.start
			until
				l_list.off
			loop
				l_cell := l_list.item
				if l_cell /= Void then
					l_value := l_params.value (l_cell.parameter.name)
					l_cell.set_value_from_string (l_value)
				end
				l_list.forth
			end
			validate
		end

	validate is
			-- Validate parameters.
		do
			if validator /= Void then
				validator.call ([is_valid_parameters])
			end
		end

	is_valid_parameters: BOOLEAN is
			-- Are all parameters valid?
		local
			l_cell: ROSE_COMMAND_PARAMETER_CELL
		do
			Result := True
			if parameter_list /= Void then
				from
					parameter_list.start
				until
					not Result or else parameter_list.off
				loop
					l_cell := parameter_list.item
					l_cell.validate (False)
					Result := l_cell.is_valid
					parameter_list.forth
				end
			end
		end

	show_controls (a_show: BOOLEAN) is
			-- Show/hide `launch_button' according to `a_show'
		local
			l_parameter: ROSE_COMMAND_PARAMETER_CELL
		do
			if a_show then
				controls_box.show
				from parameter_list.start until parameter_list.off loop
					l_parameter := parameter_list.item_for_iteration
					l_parameter.set_validator (agent on_validation)
					parameter_list.forth
				end
				set_launch_button_enabled_state
			else
				controls_box.hide
				from parameter_list.start until parameter_list.off loop
					l_parameter := parameter_list.item_for_iteration
					l_parameter.set_validator (Void)
					parameter_list.forth
				end
			end
		end

	on_validation is
			-- Actions to fire when a parameter is tested for validity
		do
			if not is_validating then
				is_validating := True
				set_launch_button_enabled_state
				if is_valid_parameters then
					parameter_sets_combo.set_text (parameters_string)
				else
					parameter_sets_combo.set_text ("")
				end
				is_validating := False
			end
		end

	is_validating: BOOLEAN
			-- Is `on_validation' currently being executed?

	set_launch_button_enabled_state is
			-- Dis/en able `launch_button' according to `is_valid_parameters'
		do
			if is_valid_parameters then
				launch_button.enable_sensitive
			else
				launch_button.disable_sensitive
			end
		end

	launch is
			-- Launch `command'
		do
			if is_valid_parameters then
				command.execute (parameters_string)
				remember_and_load_parameter_set
			end
		end

	remember_and_load_parameter_set is
			-- Remember current parameter set and load
		do
			remember_parameter_set
			load_parameter_sets
		end

	refresh_parameters_from_combo is
			-- Refresh parameters from text in `parameter_Sets_combo'
		local
			l_cell: ROSE_COMMAND_PARAMETER_CELL
			l_parameter_hash: ROSE_HASH_TABLE [PAIR [STRING, STRING], STRING]
			i: INTEGER
			l_pair: PAIR [STRING, STRING]
		do
			l_parameter_hash := parameters_from_string (parameter_sets_combo.text)
			from i := 1 until i > parameter_list.count loop
				l_cell := parameter_list.i_th (i)
				l_cell.set_to_default_value
				i := i + 1
			end
			from i := 1 until i > parameter_list.count loop
				l_cell := parameter_list.i_th (i)
				l_pair := l_parameter_hash [l_cell.parameter.name]
				if l_pair /= Void and then l_pair.second /= Void then
					l_cell.set_value_from_string (l_pair.second)
					l_cell.validate (True)
				end
				i := i + 1
			end
			validate
		end

	parameters_from_string (a_string: STRING): ROSE_HASH_TABLE [PAIR [STRING, STRING], STRING] is
			-- Parse parameters from `a_string'
		require
			a_string_not_void: a_string /= Void
		local
			l_defs: ROSE_LINEAR_ARRAY [STRING]
			l_pair: PAIR [STRING, STRING]
			i: INTEGER
		do
			create Result.make (1)
			l_defs := global.formatter.split_array (a_string, " %T", Void, False)
			from i := 1 until i > l_defs.count loop
				l_pair := global.formatter.split_on_first (l_defs [i], global.convention.all_command_parameter_qualifiers, Void)
				l_pair.first.prune_all_leading ('-')
				l_pair.second.prune_all ('%"')
				Result [l_pair.first] := l_pair
				i := i + 1
			end
		end

feature {NONE} -- Implementation

	raw_parameters: EV_TEXT_FIELD										-- Raw parameters (in case signature is '*')
	parameter_list: LINKED_LIST [ROSE_COMMAND_PARAMETER_CELL]			-- Parameter view list

	add_parameter (a_box: EV_BOX; a_parameter: ROSE_COMMAND_PARAMETER_ITEM) is
			-- Add cell for 'a_parameter' item in 'a_box'.
		do
			parameter_list.extend (create {ROSE_COMMAND_PARAMETER_CELL}.make (a_box, a_parameter, agent validate))
		end

	parameter_cell (a_name: STRING): ROSE_COMMAND_PARAMETER_CELL is
			-- Parameter cell corresponding to parameter with 'a_name'
		do
			from
				parameter_list.start
			until
				Result /= Void or else parameter_list.off
			loop
				parameter_list.forth
			end
		end

feature {NONE} -- Constants

	k_parameter_sets_string: STRING is "parameter_sets"		-- Constant
	k_max_remembered_parameter_sets: INTEGER is 5			-- Constant

end
