indexing
	description: "Event wrapper"
	author: "Jann Roeder"
	last_editor: "$Author: it $"
	date: "$Date: 2008/12/01 $"
	revision: "$Revision: #2 $"
	path: "$File: //rose/source/kernel/infrastructure/windows/rose_concurrent_event.e $"
	documentation: "http://msdn2.microsoft.com/en-us/library/ms686915.aspx"
	keywords: "synchronization,thread,process,mutex,interprocess"

class
	ROSE_CONCURRENT_EVENT

inherit
	ANY

	DISPOSABLE

	EXCEPTIONS
		export
			{NONE} all
		end

create
	make

feature {NONE} -- Initialization

	make (a_manual_reset: BOOLEAN; a_init_state: BOOLEAN) is
			-- Create event object
			--
			-- `a_manual_reset': If this parameter is TRUE,
			-- the function creates a manual-reset event object,
			-- which requires the use of the `reset' function to set the event state to nonsignaled.
			-- If this parameter is FALSE, the function creates an auto-reset event object,
			-- and system automatically resets the event state to nonsignaled after a single waiting
			-- thread has been released.
			--
			-- `a_init_state': If this parameter is TRUE, the initial state of the event object is signaled;
			-- otherwise, it is nonsignaled.			
		do
			ptr := create_event_external (default_pointer, a_manual_reset, a_init_state, default_pointer)
		end

feature -- Access

	is_set: BOOLEAN is
			-- True if the event is initialized sucessfully
		do
			Result := ptr /= default_pointer
		end

feature -- Basic operations

	wait is
			-- Waits for the event to be signaled.
		require
			is_set: is_set
		local
			l_res: BOOLEAN
		do
			l_res := wait_with_timeout (infinite_macro)
		end

	wait_with_timeout (a_timeout: INTEGER): BOOLEAN is
			-- Waits for the event to be signaled for at most `a_timeout' ms.
			-- The Result is True if the event was signaled in time and False otherwise
		require
			is_set: is_set
		local
			l_res: INTEGER
		do
			l_res := wait_for_single_object_external (ptr, a_timeout)
			if l_res = wait_object_0_macro then
				Result := True
			elseif l_res = wait_timeout_macro then
				Result := False
			elseif l_res = wait_abandoned_macro then
				raise ("Wait abandoned")
			else
				raise ("Wait failed: " + get_last_error_external.out)
			end
		end

	set is
			-- Sets the specified event object to the signaled state.
		require
			is_set: is_set
		do
			if not set_event_external (ptr) then
				raise("Set Failed: " + get_last_error_external.out)
			end
		end

	reset is
			-- Sets the specified event object to the nonsignaled state.
		require
			is_set: is_set
		do
			if not reset_event_external (ptr) then
				raise ("Reset Failed: " + get_last_error_external.out)
			end
		end

feature {NONE} -- Memory Management

	dispose is
			-- release allocated memory
		local
			l_res: BOOLEAN
		do
			if ptr /= default_pointer then
				l_res := close_handle_external (ptr)
				ptr := default_pointer
			end
		end

feature {NONE} -- Implementation

	ptr: POINTER
			-- Event handle

	create_event_external (a_event_attributes: POINTER; a_manual_reset: BOOLEAN; a_initial_state: BOOLEAN; a_name: POINTER): POINTER is
			-- Creates or opens a named or unnamed event object.
		external
			"C signature (EIF_POINTER, EIF_BOOLEAN, EIF_BOOLEAN, EIF_POINTER):EIF_POINTER use <windows.h>"
		alias
			"CreateEvent"
		end

	reset_event_external (a_ptr: POINTER): BOOLEAN is
			-- Sets the specified event object to the nonsignaled state.
			-- Returns True if successful, False otherwise
		external
			"C signature (EIF_POINTER): EIF_BOOLEAN use <windows.h>"
		alias
			"ResetEvent"
		end

	set_event_external (a_ptr: POINTER): BOOLEAN is
			-- Sets the specified event object to the signaled state.
		external
			"C signature (EIF_POINTER): EIF_BOOLEAN use <windows.h>"
		alias
			"SetEvent"
		end

	wait_for_single_object_external (a_ptr: POINTER; a_milliseconds: INTEGER): INTEGER is
			-- Waits until the specified object is in the signaled state or the time-out interval elapses.
		external
			"C blocking signature (EIF_POINTER, EIF_INTEGER): EIF_INTEGER use <windows.h>"
		alias
			"WaitForSingleObject"
		end

	get_last_error_external: INTEGER is
			-- Retrieves the calling thread's last-error code value.
			-- The last-error code is maintained on a per-thread basis.
			-- Multiple threads do not overwrite each other's last-error code.
		external
			"C signature (): EIF_INTEGER use <windows.h>"
		alias
			"GetLastError"
		end

	close_handle_external (a_ptr: POINTER): BOOLEAN is
			-- Closes an open object handle.
		external
			"C signature (EIF_POINTER): EIF_BOOLEAN use <windows.h>"
		alias
			"CloseHandle"
		end

feature {NONE} -- Macros

	infinite_macro: INTEGER is
		-- Infinite number, used for WaitForSingleObject
		external
			"C [macro %"windows.h%"]: EIF_INTEGER"
		alias
			"INFINITE"
		end

	wait_object_0_macro: INTEGER is
		-- The state of the specified object is signaled, used for WaitForSingleObject
		external
			"C [macro %"windows.h%"]: EIF_INTEGER"
		alias
			"WAIT_OBJECT_0"
		end

	wait_timeout_macro: INTEGER is
		-- The time-out interval elapsed, and the object's state is nonsignaled, used for WaitForSingleObject
		external
			"C [macro %"windows.h%"]: EIF_INTEGER"
		alias
			"WAIT_TIMEOUT"
		end

	wait_failed_macro: INTEGER is
		-- The wait function failed, used for WaitForSingleObject
		external
			"C [macro %"windows.h%"]: EIF_INTEGER"
		alias
			"WAIT_FAILED"
		end

	wait_abandoned_macro: INTEGER is
		-- The specified object is a mutex object that was not released by the thread that
		-- owned the mutex object before the owning thread terminated. Ownership of the mutex object
		-- is granted to the calling thread, and the mutex is set to nonsignaled.
		-- Used for WaitForSingleObject
		external
			"C [macro %"windows.h%"]: EIF_INTEGER"
		alias
			"WAIT_ABANDONED"
		end

end

