note
	description: "Configuration container, (key,pair) values set storable in various formats"
	notice: "All keys are of the form 'key1/key2/...'"
	author: "Zoran Simic"
	last_editor: "$Author: zs $"
	date: "$Date: 2011/04/19 $"
	revision: "$Revision: #29 $"
	path: "$File: //rose/source/kernel/infrastructure/framework/configuration/rose_configuration.e $"

class
	ROSE_CONFIGURATION

inherit
	ROSE_FORMATTING
	DEBUG_OUTPUT

create
	make_default

feature {NONE} -- Initialization

	make_default
			-- Create default empty configuration object.
		do
			wipe_out
			set_path (Void)
		ensure
			valid_root: root /= Void
			no_type: type = Void
			no_folder: folder = Void
			no_path: path = Void
			not_available: not is_available
		end

feature -- Access

	root: ROSE_CONFIGURATION_NODE
			-- Root configuration node

	folder: STRING
			-- Folder holding this configuration (directory part of 'path')

	path: STRING
			-- This configuration's path

	type: STRING
			-- Configuration origin type (format)

	is_available: BOOLEAN
			-- Is this configuration available (was it successfully retrieved)?

	is_writable: BOOLEAN
			-- Is this configuration file writable by current user?
		do
			Result := formatter.is_file_writable (path)
		end

	interpreted (a_string: STRING): STRING
			-- Interpreted value of 'a_string'
		do
			Result := formatter.recursively_interpreted (a_string, 10, agent root.child_value (?, Void))
		end

	node (a_key: STRING): ROSE_CONFIGURATION_NODE
			-- Node identified by 'a_key', or Void if no such node
		do
			Result := root.node (interpreted (a_key))
		end

	value (a_key: STRING): STRING
			-- Configuration value of 'a_key', Void if none
		do
			Result := value_default (a_key, Void)
		end

	value_default (a_key,a_default: STRING): STRING
			-- Configuration value of 'a_key', 'a_default' if none
		do
			Result := root.child_value (interpreted (a_key), a_default)
		end

	value_default_with_override (a_key, a_default, a_suffix: STRING): STRING
			-- obsolete, use 'value_default_suffixed'
		obsolete
			"Use value_default_suffixed"
		do
			Result := value_default_suffixed (a_key, a_default, a_suffix)
		end

	value_default_suffixed (a_key,a_default,a_suffix: STRING): STRING
			-- Configuration value of 'a_key', 'a_default' if none
		do
			if a_suffix /= Void then
				Result := value (a_key + "_" + a_suffix)
			end
			if Result = Void then
				Result := value_default (a_key, a_default)
			end
		end

	value_override (a_key, a_suffix: STRING): STRING
			-- obsolete, use 'value_suffixed': value from a_key+a_suffix
		obsolete
			"use value_suffixed"
		do
			Result := value_suffixed (a_key, a_suffix)
		end

	value_suffixed (a_key, a_suffix: STRING): STRING
			-- value from a_key+a_suffix
		do
			Result := value_default_suffixed (a_key, Void, a_suffix)
		end

	value_with_override (a_node_prefix, a_child:STRING; a_suffix: STRING): STRING
			-- obsolete, use 'value_node_suffixed'
		obsolete
			"use value_node_suffixed"
		do
			Result := value_node_suffixed (a_node_prefix, a_child, a_suffix)
		end

	value_node_suffixed (a_node_prefix,a_child:STRING; a_suffix:STRING):STRING
			-- apply suffix to node, and child to find a value
		local
			l_node: ROSE_CONFIGURATION_NODE
		do
			if a_suffix /= Void and then a_suffix.count > 0 then
				l_node := node (a_node_prefix + "_" + a_suffix)
			end
			if l_node=Void then
				l_node := node (a_node_prefix)
			end
			if l_node /= Void then
				if a_suffix=void then
					Result := l_node.child_value (a_child, Void)
				else
					Result := l_node.child_value (a_child + "_" + a_suffix, Void)
					if result=void then
						Result := l_node.child_value (a_child, Void)
					end
				end
			end
			if Result=Void then
--				raise ("missing config entry for "+a_node_prefix+","+a_child)
			end
		end

	boolean (a_key: STRING; a_default: BOOLEAN): BOOLEAN
			-- Boolean value of 'a_key', `a_default' if none
		local
			l_val: STRING
		do
			l_val := value (a_key)
			if l_val = Void then
				Result := a_default
			elseif l_val.is_boolean then
				Result := l_val.to_boolean
			elseif l_val.is_integer then
				Result := l_val.to_integer /= 0
			else
				Result := a_default
			end
		end

	boolean_suffixed (a_key,a_suffix: STRING; a_default: BOOLEAN): BOOLEAN
			-- Boolean value of 'a_key', `a_default' if none
		local
			l_val: STRING
		do
			l_val := value_default_suffixed (a_key, Void, a_suffix)
			if l_val = Void then
				Result := a_default
			elseif l_val.is_boolean then
				Result := l_val.to_boolean
			elseif l_val.is_integer then
				Result := l_val.to_integer /= 0
			else
				Result := a_default
			end
		end

	integer (a_key: STRING; a_default: INTEGER): INTEGER
			-- Integer value of 'a_key'
			-- 'a_default' if none
		local
			l_val: STRING
		do
			l_val := value (a_key)
			if l_val /= Void and then l_val.is_integer then
				Result := l_val.to_integer
			else
				Result := a_default
			end
		end

	integer_suffixed (a_key, a_suffix: STRING; a_default: INTEGER): INTEGER
			-- Integer value of 'a_key'
			-- 'a_default' if none
		local
			l_val: STRING
		do
			l_val := value_default_suffixed (a_key, Void, a_suffix)
			if l_val /= Void and then l_val.is_integer then
				Result := l_val.to_integer
			else
				Result := a_default
			end
		end

	double (a_key: STRING; a_default: DOUBLE): DOUBLE
			-- Double value of 'a_key'
			-- 'a_default' if none
		local
			l_val: STRING
		do
			l_val := value (a_key)
			if l_val /= Void and then l_val.is_double then
				Result := l_val.to_double
			else
				Result := a_default
			end
		end

	date (a_key: STRING; a_default: ROSE_DATE): ROSE_DATE
			-- Date value of 'a_key'
			-- 'a_default' if none
		do
			Result := formatter.make_date_from_string (value (a_key))
			if Result = Void then
				Result := a_default
			end
		end

	time (a_key: STRING; a_default: ROSE_TIME): ROSE_TIME
			-- Time value of 'a_key'
			-- 'a_default' if none
		do
			Result := formatter.make_time_from_string (value (a_key))
			if Result = Void then
				Result := a_default
			end
		end

	character (a_key: STRING; a_default: CHARACTER): CHARACTER
			-- Character value of 'a_key'
			-- 'a_default' if none
		local
			l_val: STRING
		do
			l_val := value (a_key)
			if l_val /= Void and then not l_val.is_empty then
				Result := l_val.item (1)
			else
				Result := a_default
			end
		end

	children_keys (a_key: STRING): ROSE_LINEAR_ARRAY [STRING]
			-- Direct children keys of node with 'a_key'
		local
			l_node: ROSE_CONFIGURATION_NODE
		do
			l_node := node (a_key)
			if l_node /= Void then
				Result := l_node.children_keys
			end
		end

	children_values (a_key: STRING): ROSE_LINEAR_ARRAY [STRING]
			-- Direct children values of node with 'a_key'
		local
			l_node: ROSE_CONFIGURATION_NODE
		do
			l_node := node (a_key)
			if l_node /= Void then
				Result := l_node.children_values
			end
		end

	is_modified: BOOLEAN
			-- Has the configuration been modified ?
		do
			Result := root.is_deep_modified
		end

	has (a_key: STRING): BOOLEAN
			-- Is 'a_key' defined in configuration ?
		do
			Result := root.has (interpreted (a_key))
		end

	debug_output: STRING
			-- String that should be displayed in debugger to represent `Current'.
		do
			if path /= Void then
				Result := path
			elseif folder /= Void then
				Result := folder
			else
				Result := "?"
			end
		end

feature -- Element change

	set (a_key, a_value: STRING)
			-- Set 'a_key' to 'a_value' in the configuration.
		require
			valid_key: a_key /= Void
		do
			root.set (interpreted (a_key), a_value)
		ensure
			key_set: has (a_key)
		end

	set_available (a_value: BOOLEAN)
			-- Set 'is_available'
		do
			is_available := a_value
		end

	set_path (a_path: STRING)
			-- Set 'path' to 'a_path'.
		do
			path := a_path
			folder := formatter.directory (path)
		ensure
			path_set: path = a_path
		end

	set_type (a_type: STRING)
			-- Set 'type' to 'a_type'.
		do
			type := a_type
		ensure
			type_set: type = a_type
		end

	set_root (a_node: like root)
			-- Set 'root' to 'a_node'.
		require
			valid_node: a_node /= Void
		do
			root := a_node
		ensure
			root_set: root = a_node
		end

feature -- Removal

	delete (a_key: STRING)
			-- Delete 'a_key' if present from configuration.
		local
			l_node: ROSE_CONFIGURATION_NODE
		do
			l_node := node (a_key)
			if l_node /= Void then
				l_node.delete_children
				if l_node.parent /= Void then
					l_node.parent.delete_child (l_node.key)
				end
			end
		end

	delete_children (a_key: STRING)
			-- Delete all childrens of 'a_key' if present in configuration.
		local
			l_node: ROSE_CONFIGURATION_NODE
		do
			l_node := node (a_key)
			if l_node /= Void then
				l_node.delete_children
			end
		end

	wipe_out
			-- Remove all nodes from container.
		do
			create root.make_root
			is_available := False
		end

feature -- Measurement

	count: INTEGER
			-- Number of leaves in configuration tree
		do
			Result := root.deep_count
		end

feature {NONE} -- Implementation

	glob: ROSE_SHARED_GLOBAL
			-- Execution environment
		once
			create Result
		end

end
