note
	description: "Handler for CFG configuration format"
	author: "Zoran Simic"
	last_editor: "$Author: zs $"
	date: "$Date: 2011/04/19 $"
	revision: "$Revision: #8 $"
	path: "$File: //rose/source/kernel/infrastructure/framework/configuration/rose_configuration_handler_cfg.e $"

class
	ROSE_CONFIGURATION_HANDLER_CFG

inherit
	ROSE_CONFIGURATION_HANDLER

feature -- Access

	id: STRING = "cfg"
		-- Handler type id

	compact_representation (a_node: ROSE_CONFIGURATION_NODE): attached STRING
			-- Compact CFG representation of 'a_node' (include directives remain include directives)
		require
			valid_node: a_node /= Void
		do
			Result := node_representation (a_node, [False, 0])
			if Result.count > 0 and then Result.item (1) = '%N' then
				Result.remove (1)
			end
		end

	flat_representation (a_node: ROSE_CONFIGURATION_NODE): attached STRING
			-- Flat CFG representation of 'a_node' (includes flattened out)
		require
			valid_node: a_node /= Void
		do
			Result := node_representation (a_node, [True, 0])
			if Result.count > 0 and then Result.item (1) = '%N' then
				Result.remove (1)
			end
		end

feature -- Basic operations

	import (a_node: ROSE_CONFIGURATION_NODE; a_path: STRING; a_included: BOOLEAN): STRING
			-- Import configuration from 'a_path' to 'a_node'.
			-- 'a_included' indicates whether the import is part of an include directive or not
			-- Returns effective resolved path
		local
			l_parser: ROSE_CONFIGURATION_CFG_PARSER
		do
			Result := Global.app.configs.get_file (a_path)
			if Result = Void then
				error.signal ("config", "Can not find config file '$1'", <<a_path>>)
			else
				global.file_access_log.record_file_read_access (Result)
				create l_parser.make
				l_parser.import (a_node, Result, a_included)
			end
		end

	save (a_node: ROSE_CONFIGURATION_NODE; a_path: STRING)
			-- Save 'a_node' to 'a_path'
			-- Include directives are saved as-is (not flattened)
		do
			effective_export (a_node, a_path, False)
		end

	flat_export (a_node: ROSE_CONFIGURATION_NODE; a_path: STRING)
			-- Save 'a_node' to 'a_path'
			-- All value nodes are exported, include directives are flattened out
		do
			effective_export (a_node, a_path, True)
		end

feature {NONE} -- Implementation

	effective_export (a_node: ROSE_CONFIGURATION_NODE; a_path: STRING; a_flatten: BOOLEAN)
			-- Export 'a_node' to 'a_path'.
			-- Only (modified or not included) nodes are saved.
		local
			l_channel: ROSE_OUTCHANNEL
			rescued, l_string, l_path: STRING
		do
			if rescued /= Void then
				error.signal (id, "Problem while saving configuration to '$1': $2", <<a_path, rescued>>)
			else
				error.acknowledge
				l_channel := Global.io_handler.manager.checked_output_channel (id, a_path)
				if l_channel = Void or else not l_channel.is_open then
					error.signal (id, "Can not save configuration to '$1'", <<a_path>>)
				else
					if a_flatten then
						l_string := flat_representation (a_node)
					else
						l_string := compact_representation (a_node)
					end
					l_channel.output_message (Global.convention.v_output, l_string)
					l_channel.flush
					if a_path /~ "-" then
						l_channel.close
					end
					if global.formatter.is_absolute_path (a_path) then
						l_path := a_path
					else
						l_path := global.formatter.composed_path (<<global.env.current_working_directory, a_path>>)
					end
					global.file_access_log.record_file_write_access (l_path)
				end
			end
		rescue
			if rescued = Void then
				rescued := last_exception_explanation
				retry
			end
		end

	node_representation (a_node: ROSE_CONFIGURATION_NODE; a_state: TUPLE [flatten: BOOLEAN; empty_lines: INTEGER]): attached STRING
			-- Compact CFG representation of 'a_node' (includes remain as includes)
		require
			valid_node: a_node /= Void
		local
			l_node: ROSE_CONFIGURATION_NODE
			l_string: STRING
			i: INTEGER
		do
			create Result.make_empty
			if a_state.flatten then
				if a_node.is_directive_node then
						-- Include directives, but comment them out (we're flattening)
					l_string := cfg_representation (a_node, a_state)
					if l_string.count > 0 then
						if l_string.item (1) = '%N' then
							l_string.remove (1)
						end
						Result.append_character ('%N')
						Result.append_character ('-')
						Result.append_character ('-')
						Result.append_string (l_string)
					end
				else
					Result.append_string (cfg_representation (a_node, a_state))
				end
			elseif is_to_be_saved (a_node) then
				-- Only non-included or modified items to be saved
				Result.append_string (cfg_representation (a_node, a_state))
			end
			if a_node.children /= Void then
				from
					-- First pass: non-section subnodes
					i := a_node.children.lower
				until
					i > a_node.children.upper
				loop
					l_node := a_node.children.item (i)
					if not is_section_node (l_node) then
						Result.append (node_representation (l_node, a_state))
					end
					i := i + 1
				end
				from
					-- Second pass: section subnodes
					i := a_node.children.lower
				until
					i > a_node.children.upper
				loop
					l_node := a_node.children.item (i)
					if is_section_node (l_node) then
						Result.append (node_representation (l_node, a_state))
					end
					i := i + 1
				end
			end
		end

	is_section_node (a_node: ROSE_CONFIGURATION_NODE): BOOLEAN
			-- Is 'a_node' a section? (ie: has )
		require
			valid_node: a_node /= Void
		do
			Result := a_node.children /= Void and then a_node.is_value_node
		end

	cfg_representation (a_node: ROSE_CONFIGURATION_NODE; a_state: TUPLE [flatten: BOOLEAN; empty_lines: INTEGER]): attached STRING
			-- CFG representation of 'a_node', with eventually included values flattened out (corresponding to 'a_state.flatten')
		require
			valid_node: a_node /= Void
		do
			create Result.make_empty
			if a_node.is_value_node then
				-- Value node: "key = value"
				if a_node.is_anonymous then
					Result.append_character ('>')
				elseif has_direct_significant_value (a_node) then
					if a_state.empty_lines = 0 then
						Result.append_character ('%N')
					end
					Result.append_character ('[')
					Result.append (a_node.path)
					Result.append_character (']')
				elseif not a_node.is_value_empty then
					Result.append (a_node.key)
				end
				if not Result.is_empty then
					if a_node.is_anonymous then
						if a_node.is_value_empty then
							Result.append ("%"%"")
						else
							string_formatter.append_quoted_eiffel_string (Result, a_node.value)
						end
					elseif not a_node.is_value_empty then
						Result.append (" = ")
						string_formatter.append_quoted_eiffel_string (Result, a_node.value)
					end
				end
			elseif a_node.is_directive_node then
				-- Directive node (include, ...)
				if a_state.empty_lines = 0 then
					Result.append_character ('%N')
				end
				Result.append (a_node.key)
				Result.append_character (' ')
				string_formatter.append_quoted_eiffel_string (Result, a_node.value)
			end
			if a_node.comment /= Void and then not a_node.comment.is_empty then
				Result.append (a_node.comment)
			end
			if Result.is_empty then
				a_state.empty_lines := a_state.empty_lines + 1
			else
				a_state.empty_lines := 0
			end
			if a_state.empty_lines < 3 then
				Result.append_character ('%N')
			end
		end

	is_to_be_saved (a_node: ROSE_CONFIGURATION_NODE): BOOLEAN
			-- Is 'a_node' to be saved ?
		require
			valid_node: a_node /= Void
		local
			l_node: ROSE_CONFIGURATION_NODE
			i: INTEGER
		do
			Result := not a_node.is_included or else a_node.is_modified
			if not Result and then a_node.children /= Void then
				from
					i := a_node.children.lower
				until
					Result or else i > a_node.children.upper
				loop
					l_node := a_node.children.item (i)
					Result := (not l_node.is_included or else l_node.is_modified) and then not l_node.is_value_empty
					i := i + 1
				end
			end
		end

	has_direct_significant_value (a_node: ROSE_CONFIGURATION_NODE): BOOLEAN
			-- Does 'a_node' have a direct (child) significant value?
		require
			valid_node: a_node /= Void
		local
			i: INTEGER
		do
			if a_node.children /= Void then
				from
					i := a_node.children.lower
				until
					Result or else i > a_node.children.upper
				loop
					Result := not a_node.children.item (i).is_value_empty
					i := i + 1
				end
			end
		end

	string_formatter: UT_STRING_FORMATTER
			-- Eiffel string formatter
		once
			create Result
		end

end
