indexing
	description: "Configuration manager"
	author: "Zoran Simic"
	date: "$Date: 2010/06/11 $"
	revision: "$Revision: #21 $"

class
	ROSE_CONFIGURATION_MANAGER

inherit
	ROSE_SHARED_GLOBAL

create
	make

feature {NONE} -- Initialization

	make is
			-- Create a configuration manager with default handlers.
		local
			l_cfg_handler: ROSE_CONFIGURATION_HANDLER_CFG
		do
			if {PLATFORM}.is_thread_capable then
				create mutex.make
			end
			create handlers.make (0)
				--| Common configuration handlers
			create l_cfg_handler
			add_handler (l_cfg_handler)
				--| FIXME: This is to remain compatible with initial .ini file extension
				--| FIXME: Remove when no more .ini configurations used
			handlers.force (l_cfg_handler,"ini")
		end

feature -- Access

	handlers: HASH_TABLE [ROSE_CONFIGURATION_HANDLER, STRING]
			-- Supported configuration handlers

	handler (a_id: STRING): ROSE_CONFIGURATION_HANDLER is
			-- Configuration handler of 'a_id', Void if none
		do
			lock
			Result := handlers.item (a_id)
			unlock
		end

feature -- Setting

	add_handler (a_handler: ROSE_CONFIGURATION_HANDLER) is
			-- Add 'a_handler' to managed handlers.
		require
			handler_ok: a_handler /= Void
		do
			lock
			handlers.force (a_handler,a_handler.id)
			unlock
		ensure
			handler_added: handler (a_handler.id) = a_handler
		end

feature -- Basic operations

	reload (a_config: ROSE_CONFIGURATION) is
			-- Reload configuration file.
		require
			config_ok: a_config /= Void and then a_config.path /= Void
		do
			load (a_config,a_config.type,a_config.path)
		end

	load (a_config: ROSE_CONFIGURATION; a_handler_id, a_path: STRING) is
			-- Load 'a_config' from 'a_path'.
			-- Reload from last config path if 'a_path' is Void.
		require
			config_ok: a_config /= Void
			path_ok: a_path /= Void
		local
			l_root: ROSE_CONFIGURATION_NODE
			l_path: STRING
		do
			lock
			already_included_table.wipe_out
			l_root := a_config.root
			a_config.wipe_out
			l_path := effective_import (a_config.root,a_handler_id,a_path,False,False)
			if not error.signaled then
					--| Configuration marked as not modified when freshly loaded
				a_config.set_path (l_path)
				a_config.set_type (a_handler_id)
				a_config.set_available (True)
			else
					--| Configuration restored in case of a load problem
				a_config.set_root (l_root)
			end
			already_included_table.wipe_out
			unlock
		end

	import (a_node: ROSE_CONFIGURATION_NODE; a_handler_id, a_path: STRING; as_included, as_try: BOOLEAN) is
			-- Import contents of 'a_path' to 'a_node', and mark them 'as_included'.
			-- 'as_try' indicates that if the import fails, no error should be raised
		require
			node_ok: a_node /= Void
			path_ok: a_path /= Void
		local
			l_path: STRING
		do
			l_path := effective_import (a_node, a_handler_id, a_path, as_included, as_try)
		end

	save (a_config: ROSE_CONFIGURATION; a_handler_id, a_path: STRING) is
			-- Save a copy of contents of 'a_config' to 'a_path',
			-- Only modified nodes and nodes loaded by this 'a_path' are saved.
		require
			config_ok: a_config /= Void
			path_ok: a_path /= Void
		local
			l_handler: ROSE_CONFIGURATION_HANDLER
		do
			lock
			l_handler := checked_handler (a_handler_id,a_path)
			if not error.signaled then
				l_handler.save (a_config.root,a_path)
			end
			if not error.signaled then
				a_config.root.deep_set_modified (False)
			end
			unlock
		end

	flat_export (a_config: ROSE_CONFIGURATION; a_handler_id, a_path: STRING) is
			-- Export a flat form of 'a_config' to 'a_path'.
		require
			config_ok: a_config /= Void
			path_ok: a_path /= Void
		local
			l_handler: ROSE_CONFIGURATION_HANDLER
		do
			lock
			l_handler := checked_handler (a_handler_id,a_path)
			if not error.signaled then
				l_handler.flat_export (a_config.root,a_path)
			end
			unlock
		end

	lock is
			-- Gain exclusive access to this manager
		do
			if mutex /= Void then
				mutex.lock
			end
		end

	unlock is
			-- Release exclusive access to this manager
		do
			if mutex /= Void then
				mutex.unlock
			end
		end

feature {NONE} -- Implementation

	mutex: CRITICAL_SECTION
			-- Used to sync multi threaded access to this manager

	effective_import (a_node: ROSE_CONFIGURATION_NODE; a_handler_id, a_path: STRING; as_included, as_try: BOOLEAN): STRING is
			-- Import contents of 'a_path' to 'a_node', and mark them 'as_included'.
			-- 'as_try' indicates that if the import fails, no error should be raised
			-- Returns effective resolved path
		require
			node_ok: a_node /= Void
			path_ok: a_path /= Void
		local
			l_handler: ROSE_CONFIGURATION_HANDLER
		do
			lock
			start_time_measure ("config import",a_path)
			Result := a_path
			if not error.signaled then
				if not already_included (a_node.path,a_path) then
					l_handler := checked_handler (a_handler_id,a_path)
					if not error.signaled then
						Result := l_handler.import (a_node,a_path,as_included)
					end
					if as_try then
						error.acknowledge
					end
				elseif not as_try then
					if as_included then
						error.warn ("Config '$1' has already been included in '$2'",<<a_path,a_node.path>>)
					else
						error.warn ("Config '$1' has already been imported in '$2'",<<a_path,a_node.path>>)
					end
				end
			end
			end_time_measure
			unlock
		end

	checked_handler (a_handler_id, a_path: STRING): ROSE_CONFIGURATION_HANDLER is
			-- Handler for the id 'a_handler_id'
		local
			l_extension: STRING
		do
			l_extension := a_handler_id
			Result := handler (l_extension)
			if Result = Void and a_path /= Void then
				l_extension := Global.formatter.extension (a_path)
				if l_extension /= Void and then not l_extension.is_empty then
					l_extension.to_lower
					Result := handler (l_extension)
				else
					l_extension := a_handler_id
				end
			end
			if Result = Void then
				error.signal ("config","Unknown config type '$1'",<<l_extension>>)
			end
		ensure
			found_xor_error: Result /= Void xor error.signaled
		end

	already_included (a_configpath, a_filepath: STRING): BOOLEAN is
			-- Was 'a_path' already included for 'a_node'?
			-- Mark it as such.
		require
			configpath_ok: a_configpath /= Void
			filepath_ok: a_filepath /= Void
		local
			s: STRING
		do
			create s.make (a_filepath.count + a_configpath.count)
			s.append (a_filepath)
			s.append (a_configpath)
			if already_included_table.has (s) then
				Result := True
			else
				already_included_table.put (True,s)
			end
		end

feature {ROSE_DATABASE_POOL} -- Implementation

	already_included_table: HASH_TABLE [BOOLEAN,STRING] is
			-- Already included files during one load
		once
			create Result.make (10)
		end

end
