indexing
	description: "Configuration node"
	author: "Zoran Simic"
	date: "$Date: 2008/12/01 $"
	revision: "$Revision: #22 $"

class
	ROSE_CONFIGURATION_NODE

inherit
	ROSE_IO_CONVENTION
		undefine
			is_equal, out
		end
	DEBUG_OUTPUT
		undefine
			is_equal, out
		end

create
	make_root, make, make_anonymous, make_comment, make_directive

feature {NONE} -- Initialization

	make_root is
			-- Make a default node (no key,value pair).
		do
			parent := Void
			key := Void
			value := Void
			comment := Void
			is_modified := False
			is_included := False
			children := Void
		ensure
			no_parent: parent = Void
			no_key: key = Void
			no_value: value = Void
			no_comment: comment = Void
			not_modified: not is_modified
			not_included: not is_included
			no_children: children = Void
		end

	make (a_parent: ROSE_CONFIGURATION_NODE; a_key: STRING; a_value: STRING) is
			-- Create a node with 'a_parent', 'a_key' and 'a_value'.
		require
			a_parent /= Void and then a_key /= Void
		do
			make_root
			a_parent.add_child (Current)
			parent := a_parent
			key := a_key.as_lower
			value := a_value
		ensure
			parent_set: parent = a_parent
			--key_set: equal (key, a_key.as_lower)
			value_set: value = a_value
			no_comment: comment = Void
			not_modified: not is_modified
			not_included: not is_included
			no_children: children = Void
		end

	make_anonymous (a_parent: ROSE_CONFIGURATION_NODE; a_value: STRING) is
		require
			a_parent /= Void and then a_value /= Void
		do
			make_root
			a_parent.add_child (Current)
			parent := a_parent
			value := a_value
		ensure
			parent_set: parent = a_parent
			no_key: key = Void
			value_set: value = a_value
			no_comment: comment = Void
			not_modified: not is_modified
			not_included: not is_included
			no_children: children = Void
		end

	make_comment (a_parent: ROSE_CONFIGURATION_NODE; a_comment: STRING) is
			-- Create a comment node with 'a_comment'.
		require
			a_parent /= Void and then a_comment /= Void
		do
			make_root
			a_parent.add_child (Current)
			comment := a_comment
		ensure
			no_parent: parent = Void
			no_key: key = Void
			no_value: value = Void
			comment_set: comment = a_comment
			not_modified: not is_modified
			not_included: not is_included
			no_children: children = Void
		end

	make_directive (a_parent: ROSE_CONFIGURATION_NODE; a_keyword, a_value: STRING) is
			-- Create a node representing a directive associated to 'a_keyword' and 'a_value'.
		require
			a_parent /= Void and then a_keyword /= Void and then a_value /= Void
		do
			make_root
			a_parent.add_child (Current)
			key := a_keyword.as_lower
			value := a_value
		ensure
			no_parent: parent = Void
			--key_set: equal (key, a_keyword.as_lower)
			value_set: value = a_value
			no_comment: comment = Void
			not_modified: not is_modified
			not_included: not is_included
			no_children: children = Void
		end

feature -- Access

	parent: ROSE_CONFIGURATION_NODE
			-- Parent node, Void if current is not representing a value (root, directives, pure comments)

	key: STRING
			-- Identifier of current node, Void for pure comments and anonymous values

	value: STRING
			-- Value of node

	children: ROSE_LINEAR_ARRAY [ROSE_CONFIGURATION_NODE]
			-- Direct children nodes of current, Void if leave

	comment: STRING
		-- Comment associated to the node, Void if none

	node (a_key: STRING): ROSE_CONFIGURATION_NODE is
			-- Node identified by 'a_key', or Void if no such node
		do
			if a_key /= Void then
				Result := find_node (a_key, False, False)
			end
		ensure
			void_xor_defined: Result = Void xor has (a_key)
		end

	forced_node (a_key: STRING; a_included: BOOLEAN): ROSE_CONFIGURATION_NODE is
			-- Node identified by 'a_key', created if necessary
			-- and marked as 'a_included'
		require
			valid_key: a_key /= Void
		do
			Result := find_node (a_key, True, a_included)
		ensure
			valid_result: Result /= Void
		end

	child_value (a_key, a_default: STRING): STRING is
			-- Value of child identified by 'a_key'
			-- 'a_default' is returned if no such child
		local
			l_node: ROSE_CONFIGURATION_NODE
		do
			l_node := node (a_key)
			if l_node /= Void then
				Result := l_node.value
			end
			if Result = Void then
				Result := a_default
			end
		end

	children_keys: ROSE_LINEAR_ARRAY [STRING] is
			-- Direct children keys of current, Void if leave
		local
			i: INTEGER
		do
			if children /= Void then
				from
					create Result.make_empty
					i := children.lower
				until
					i > children.upper
				loop
					if children.item (i).is_value_node then
						Result.extend (children.item (i).key)
					end
					i := i + 1
				end
			end
		end

	children_values: ROSE_LINEAR_ARRAY [STRING] is
			-- Direct children values of current, Void if leave
		local
			i: INTEGER
		do
			if children /= Void then
				from
					create Result.make_empty
					i := children.lower
				until
					i > children.upper
				loop
					if children.item (i).is_value_node and then not children.item (i).is_value_empty then
						Result.extend (children.item (i).value)
					end
					i := i + 1
				end
			end
		end

	root: ROSE_CONFIGURATION_NODE is
			-- Root node of this configuration (found via parents)
		do
			if parent = Void then
				Result := Current
			else
				Result := parent.root
			end
		end

	count: INTEGER is
			-- Number of direct children in current node
		do
			if children /= Void then
				Result := children.count
			end
		ensure
			definition: Result = 0 or else Result = children.count
		end

	deep_count: INTEGER is
			-- Number of leaves down this node
		local
			i: INTEGER
		do
			if children = Void then
				if parent /= Void then
						--| Comments and includes do not count
					Result := 1
				end
			else
				from
					i := children.lower
				until
					i > children.upper
				loop
					Result := Result + children.item (i).deep_count
					i := i + 1
				end
			end
		ensure
			positive_result: Result >= 0
		end

feature -- Status report

	is_included: BOOLEAN
			-- Does this node come from an include ?

	is_modified: BOOLEAN
			-- Has current node value been modified since loaded ?

	is_directive_node: BOOLEAN is
			-- Does this node represent a directive ?
		do
			Result := parent = Void and then key /= Void and then value /= Void
		end

	is_value_node: BOOLEAN is
			-- Does this node represent a value node (not a directive or a pure comment) ?
		do
			Result := parent /= Void
		end

	is_anonymous: BOOLEAN is
			-- Does this node represent an anonymous value (with no key) ?
		do
			Result := parent /= Void and then key = Void and then value /= Void
		end

	is_leave: BOOLEAN is
			-- Is this node a leave (no children) ?
		do
			Result := children = Void
		end

	has_leaves_only: BOOLEAN is
			-- Has this node leaf-children only?
		local
			i: INTEGER
		do
			Result := True
			if children /= Void then
				from
					i := children.lower
				until
					not Result or else i > children.upper
				loop
					Result := children.item (i).is_leave
					i := i + 1
				end
			end
		end

	has_directives: BOOLEAN is
			-- Has this node a directive direct subnode?
		local
			i: INTEGER
		do
			if children /= Void then
				from
					i := children.lower
				until
					Result or else i > children.upper
				loop
					Result := children.item (i).is_directive_node
					i := i + 1
				end
			end
		end

	is_commented: BOOLEAN is
			-- Is this node commented ?
		do
			Result := comment /= Void
		end

	is_pure_comment: BOOLEAN is
		do
			Result := parent = Void and then key = Void and then value = Void and then children = Void and then comment /= Void
		end

	is_blank_comment: BOOLEAN is
		do
			Result := is_pure_comment and then comment.is_empty
		end

	is_deep_modified: BOOLEAN is
			-- Is any child modified ?
		local
			i: INTEGER
		do
			Result := is_modified
			if not Result and then children /= Void then
				from
					i := children.lower
				until
					Result or else i > children.upper
				loop
					Result := children.item (i).is_deep_modified
					i := i + 1
				end
			end
		end

	is_value_empty: BOOLEAN is
			-- Is the value of this node empty ?
		do
			Result := value = Void or else value.count = 0
		end

	has (a_key: STRING): BOOLEAN is
			-- Is 'a_key' defined in children ?
		do
			Result := node (a_key) /= Void
		ensure
			bound_condition: Result implies count > 0
		end

	has_significant_value (a_key: STRING): BOOLEAN is
			-- Does 'a_key' point to an empty value ?
		local
			l_node: ROSE_CONFIGURATION_NODE
		do
			l_node := node (a_key)
			Result := l_node /= Void and then not l_node.is_value_empty
		end

	is_equal (other: like Current): BOOLEAN is
			-- Is current container equal to `other'?
		do
			Result := equal (out, other.out)
		end

feature -- Basic operations

	set (a_key, a_value: STRING) is
			-- Set 'a_value' to 'a_key'.
			-- 'a_key' is of the form "key1\key2..."
		require
			valid_key: a_key /= Void
		local
			l_node: ROSE_CONFIGURATION_NODE
		do
			l_node := forced_node (a_key, False)
			l_node.set_value (a_value)
		ensure
			--node_added: has (a_key)
		end

	set_parent (a_parent: ROSE_CONFIGURATION_NODE) is
			-- Set 'parent' to 'a_parent'
		do
			parent := a_parent
		end

	load_value (a_value: like value) is
			-- Set 'value' to 'a_value', without setting 'is_modified'.
		do
			value := a_value
		ensure
			value_set: value = a_value
		end

	set_value (a_value: like value) is
			-- Set 'value' to 'a_value', mark eventually 'is_modified'.
		do
			is_modified := is_modified or else not equal (value, a_value)
			value := a_value
		ensure
			modified: not equal(value, a_value) implies is_modified
			value_set: value = a_value
		end

	set_comment (a_comment: like comment) is
			-- Set 'comment' to 'a_comment'.
		do
			comment := a_comment
		ensure
			comment_set: comment = a_comment
		end

	add_anonymous (a_value: STRING) is
		local
			l_child: ROSE_CONFIGURATION_NODE
		do
			if a_value /= Void then
				create l_child.make_anonymous (Current, a_value)
			end
		end

	deep_set_modified (a_value: BOOLEAN) is
			-- Set 'modified' to 'a_value', including for all the children.
		local
			i: INTEGER
		do
			is_modified := a_value
			if children /= Void then
				from
					i := children.lower
				until
					i > children.upper
				loop
					children.item (i).deep_set_modified (a_value)
					i := i + 1
				end
			end
		end

	deep_set_included (a_value: BOOLEAN) is
			-- Set 'included' to 'a_value', including for all the children.
		local
			i: INTEGER
		do
			is_included := a_value
			if children /= Void then
				from
					i := children.lower
				until
					i > children.upper
				loop
					children.item (i).deep_set_included (a_value)
					i := i + 1
				end
			end
		end

	set_included (a_value: BOOLEAN) is
			-- Set 'is_included'.
		do
			is_included := a_value
		ensure
			is_included_set: is_included = a_value
		end

	delete_child (a_key: STRING) is
			-- Delete direct child of 'a_key', if any.
			-- 'a_key' Void removes all pure comments and anonymous values.
		local
			found: BOOLEAN
			l_key: STRING
		do
			if a_key /= Void and then children /= Void then
				l_key := clone (a_key)
				l_key.to_lower
				from
					children.start
				until
					found or else children.off
				loop
					if equal (l_key, children.current_item.key) then
						children.remove
						found := True
						is_modified := True
					else
						children.forth
					end
				end
				if children.is_empty then
					children := Void
				end
			end
		end

	delete_children is
			-- Delete all childrens of this node.
		do
			children := Void
			is_modified := True
		end

	do_all (a_agent: PROCEDURE [ANY, TUPLE [ROSE_CONFIGURATION_NODE]]) is
			-- Apply 'a_agent' on all value children of this node (ignore comments and directives).
		local
			i: INTEGER
		do
			if children /= Void then
				from
					i := children.lower
				until
					i > children.upper
				loop
					if children.item (i).is_value_node then
						a_agent.call ([children.item (i)])
					end
					i := i + 1
				end
			end
		end

feature -- Output

	path: STRING is
			-- Path from root node to current.
		do
			if parent = Void then
				create Result.make (10)
				Result.append_character (configuration_node_separator)
			else
				Result := parent.path
				if Result.count > 1 then
					Result.extend (configuration_node_separator)
				end
				if key /= Void then
					Result.append (key)
				else
					Result.append_character ('>')
				end
			end
		end

	debug_output: STRING is
			-- String that should be displayed in debugger to represent `Current'.
		do
			if is_directive_node then
				Result := key + " " + value
			elseif is_value_node then
				if is_anonymous then
					Result := "> " + value
				else
					if is_modified then
						Result := key + "*="
					elseif is_included then
						Result := "(" + key + ")="
					else
						Result := key + "="
					end
					if value /= Void then
						Result := Result + value
					end
				end
			else
				Result := comment
			end
			if Result = Void then
				create Result.make (0)
			end
		end

	out: STRING is
			-- String representation of this node
		local
			i: INTEGER
		do
			if is_value_node then
					--| Value node: key [= value]
				Result := path
				if not is_value_empty then
					Result.append (" = ")
					Result.append (value)
				end
				Result.append_character ('%N')
			end
			if children /= Void then
				if Result = Void then
					create Result.make (1000)
				end
				from
						--| First pass: leaves
					i := children.lower
				until
					i > children.upper
				loop
					if children.item (i).is_leave then
						Result.append (children.item (i).out)
					end
					i := i + 1
				end
				from
						--| Second pass: non-leaves
					i := children.lower
				until
					i > children.upper
				loop
					if not children.item (i).is_leave then
						Result.append (children.item (i).out)
					end
					i := i + 1
				end
			end
			if Result = Void then
				create Result.make (0)
			end
		end

feature {ROSE_CONFIGURATION_NODE} -- Implementation

	add_child (a_child: ROSE_CONFIGURATION_NODE) is
			-- Add 'a_child' to list of current node's 'children'.
		require
			valid_child: a_child /= Void
			comment_or_parent: (not a_child.is_value_node and then a_child.parent = Void) or else (a_child.is_value_node and then a_child.parent = Current)
		do
			if children = Void then
				create children.make_empty
			end
			children.extend (a_child)
		end

	find_node (a_key: STRING; a_force_creation, a_included: BOOLEAN): ROSE_CONFIGURATION_NODE is
			-- Node identified by 'a_key',
			-- created if necessary and asked via 'a_force_creation'
			-- 'is_included' set to 'a_included' if new node created
		require
			valid_key: a_key /= Void
		local
			p: PAIR [STRING, STRING]
			l_node: ROSE_CONFIGURATION_NODE
			l_key: STRING
			i: INTEGER
		do
			if a_key.is_empty then
					--| Empty points to Current
				Result := Current
			elseif a_key.item (1) = configuration_node_separator then
					--| Beginning with separator points to root
				Result := root.find_node (a_key.substring (2, a_key.count), a_force_creation, a_included)
			elseif a_key.has (configuration_node_separator) then
					--| Decompose subkeys
				p := formatter.split_on_first (a_key, configuration_node_separator.out, Void)
				Result := find_node (p.first, a_force_creation, a_included)
				if Result /= Void then
					Result := Result.find_node (p.second, a_force_creation, a_included)
				end
			elseif children /= Void then
					--| Simple key
				l_key := clone (a_key)
				l_key.to_lower
				from
					i := children.lower
				until
					Result /= Void or else i > children.upper
				loop
					l_node := children.item (i)
					if l_node.is_value_node and then equal (l_node.key, l_key) then
						Result := l_node
					else
						i := i + 1
					end
				end
			end
			if Result = Void and then a_force_creation then
					--| Node is created if not found
				create Result.make (Current, a_key, Void)
				Result.set_included (a_included)
			end
		ensure
			forced_if_asked: a_force_creation implies Result /= Void
		end

end
