note

	description:

		"Class that handles network connectivity for the Cache Server client"

	documentation: "[%'server%' refers to the real server, I use the term %'port server%' if I mean the %'port server%' ]"
	author: "Jann Roeder"
	last_editor: "$Author: er $"
	date: "$Date: 2014/02/12 $"
	revision: "$Revision: #16 $"
	path: "$File: //rose/source/kernel/entities/cache_server/rose_cs_client.e $"

class ROSE_CS_CLIENT

inherit

	PERSISTENT_SERVER_CLIENT
		rename
			make as make_persistent_client,
			make_with_arguments as make_with_arguments_persistent_client
		undefine
			default_create
		redefine
			port_server_port_config_key
		end

	THREAD
		rename
			make as make_thread
		end

create

	make, make_with_arguments

feature {NONE} -- Initialization

	make (a_api: like cs_api; a_server_name: like server_name) is
			-- Create connection to `a_address' at `a_port'
		require
			api_not_void: a_api /= Void
			server_name_valid: a_server_name /= Void and then not a_server_name.is_empty
		do
			make_thread
			cs_api := a_api
			create object_queue.make
			make_persistent_client (a_server_name)
			launch
		end

	make_with_arguments (a_api: like cs_api; a_server_name: like server_name; a_server_arguments: like server_arguments) is
			-- Create connection to `a_address' at `a_port'
		require
			api_not_void: a_api /= Void
			server_name_valid: a_server_name /= Void and then not a_server_name.is_empty
			arguments_not_void_and_valid: a_server_arguments /= Void and then not a_server_arguments.is_empty
		do
			make_thread
			cs_api := a_api
			create object_queue.make
			make_with_arguments_persistent_client (a_server_name, a_server_arguments)
			launch
		end

feature -- Access

	--debugging only
	now: STRING is
			-- return current time as string
		local
			l_now: TIME
		do
			create l_now.make_now
			Result := l_now.out
		end

	object_queue: ROSE_CONCURRENT_LINKED_QUEUE[ROSE_CS_QUERY]
		-- Objects that are going to be sent

	shutdown: BOOLEAN
		-- Flag to signal thread shutdown

	cs_api: ROSE_CS_CLIENT_API

	error_count: INTEGER
		-- Error count for worker thread

	always_check_for_latest_server: BOOLEAN
		-- If this is True, the client will check for the latest server everytime it connects.
		-- If this is False, the client will connect to the server every 14 minutes, to make sure it stays alive
		-- The server should have a 15 minutes timeout.

feature -- Status setting

	set_always_check_for_latest_server (a_value: like always_check_for_latest_server) is
			-- Set `always_check_for_latest_server'
		do
			always_check_for_latest_server := a_value
		end

feature -- Basic operations

	connect is
			-- Connect to the server
		do
			want_connection_change := True
		end

	send_object (a_object: ROSE_CS_QUERY) is
			-- Send `a_object' to the server
		do
			object_queue.extend (a_object)
		end

	disconnect is
			-- Disconnect from server
			-- This will disconnect at the next possible occasion, i.e. pending results will still be received
		do
			want_disconnect := True
		end

feature -- Query

	request_underway: BOOLEAN is
			-- Can we expect a response ?
		do
			Result := requests_sent > results_received
		end

feature {NONE} -- Implementation

	want_connection_change: BOOLEAN
			-- BOOLEAN used to request a change in connection

	want_disconnect: BOOLEAN
			-- BOOLEAN used to request a disconnect

	requests_sent: INTEGER

	results_received: INTEGER

	execute is
			-- Run in new thread
		local
			l_queued_object: TUPLE[valid: BOOLEAN; value: ROSE_CS_QUERY]
			l_result: ROSE_CS_RESULT
			l_future_result: ROSE_CS_FUTURE[ROSE_CS_RESULT]
			l_last_request: TIME
			l_now: TIME
			l_timeout: TIME_DURATION
			l_keep_alive: TIME_DURATION
			l_stored_request: ROSE_CS_STORED_REQUEST
			l_old_port: like server_port
			l_was_connection_change: BOOLEAN
			l_retried: BOOLEAN
			l_requeue_query: BOOLEAN -- This flag determines whether we need to re-queue an object on exception
			l_retry_count: INTEGER
		do
			if not l_retried then
				create l_last_request.make_now
				create l_now.make_now
				create l_timeout.make_by_seconds (60)
				create l_keep_alive.make (0, 14, 0) -- 14 minutes
			end
			from
			until
				shutdown
			loop
				-- Handle disconnect request
				if want_disconnect then
					cleanup
					want_disconnect := False
				end

				-- If errors have occured,  move on to the next server
				if error_count > 0 then
					cleanup
					if not server_list_exhausted then
						next_server
						get_port_from_port_server
					else
						load_server_list
						get_port_from_port_server
					end
					error_count := 0
				end

				-- If a connection change has been requested, disconnect from current server and request a new port
				-- Connection change means basically that `connect' has been called
				if want_connection_change then
					l_was_connection_change := True -- This is used to avoid a race condition, where a change request would be lost
					if connected then
						cleanup
					end
					l_old_port := server_port
					get_port_from_port_server
					if l_old_port /= server_port then
						error_count := 0
					end
				end

				-- If a connection_change was requested or we are not connected but have something in our send queue -> connect
				if (not connected and then not object_queue.is_empty) or else want_connection_change then
					if server_port <= 0 or always_check_for_latest_server then
						l_old_port := server_port
						get_port_from_port_server
						if l_old_port /= server_port then
							error_count := 0
						end
					end
					requests_sent := 0
					results_received := 0
					if server_port > 0 then
						connect_to_server
					else
						error_count := error_count + 1
					end
					if l_was_connection_change then
						want_connection_change := False
						l_was_connection_change := False
					end
				end

				-- If we have something to send, we should be connected by now
				if connected then
					if not request_underway then
						l_queued_object := object_queue.get_with_timeout (300)
						if l_queued_object.valid then
							l_requeue_query := True
							send (l_queued_object.value)
						--	send_packet (l_queued_object.value)
							requests_sent := requests_sent + 1
						end
					end
					if l_queued_object.valid then -- If request was sent, wait for response
						if l_retry_count <= 10
						then
							receive

						--	receive_packet
							results_received := results_received + 1
							l_last_request.make_now
							l_result ?= received
						else
							l_result := Void
						end

						l_retry_count := 0		-- Even on failure we must reset this counter.
						if l_result /= Void then
							l_stored_request := cs_api.requests.item_thread_safe (l_result.query_id)
							if l_stored_request /= Void then
								l_future_result := l_stored_request.result_future
							else
								l_future_result := Void
							end
							if l_future_result /= Void then
								l_requeue_query := False
								l_future_result.set (l_result)
								--io.put_string (now + " received result %N")
								global.invoke (agent (cs_api.result_available).call ([l_future_result]))
							end
						else
							-- We received something unexpected, create a result with an exception flag
							l_stored_request := cs_api.requests.item_thread_safe (l_queued_object.value.query_id)
							if l_stored_request /= Void then
								l_future_result := l_stored_request.result_future
								l_requeue_query := False
								error_count := error_count + 1
								l_future_result.set_exception ("Received object of wrong type")
								global.invoke (agent (cs_api.result_available).call ([l_future_result]))
							end
						end
					end

					-- Disconnect if no activity for certain time
					l_now.make_now
					if not request_underway and connected and then l_now.relative_duration (l_last_request) > l_timeout then
						want_disconnect := True
					end

					-- If necessary connect to server to keep it alive
					if not connected and then server_port > 0 and then not always_check_for_latest_server and then l_now.relative_duration (l_last_request) > l_timeout then
						l_last_request.make_now
						connect_to_server
					end

					-- Prevent Integer overflow
					if requests_sent > requests_sent.max_value - 10 then
						requests_sent := requests_sent - requests_sent.min (results_received)
						results_received := results_received - results_received.min (requests_sent)
					end
				else
					-- If not connected, sleep 250 ms
					sleep (250_000_000)
				end
			end
			cleanup
		rescue
			l_retry_count := l_retry_count + 1
			error_count := error_count + 1
			l_retried := True

			if l_requeue_query then
				check
					queued_object_exists: l_queued_object /= Void and then l_queued_object.valid
				end
				object_queue.extend (l_queued_object.value)
			end
			-- Sleep to avoid flooding the network
			sleep (3_000_000_000)
			retry
		end

	port_server_port_config_key: STRING is
			-- Since test environments don't normally have entity
			-- servers you should always leave this undefined in your
			-- config file (port will be 2000). Only when you are expressly
			-- creating a test env for entity servers should a value exist in the config.
		do
			Result := "/port_server/port_for_entity_server"
		end

invariant

	object_queue_exists: object_queue /= Void

end
