note

	description:

		"receives and sends messages in a seperate thread (started by ROSE_CS_PERSISTENT_SERVER_COMMAND)"

	documentation: ""
	author: "Ueli Peter"
	last_editor: "$Author: er $"
	date: "$Date: 2014/02/12 $"
	revision: "$Revision: #8 $"
	path: "$File: //rose/source/kernel/entities/cache_server/rose_cs_server_thread.e $"

class	ROSE_CS_SERVER_THREAD

inherit

	THREAD
		rename
			make as make_thread
		end

	ROSE_SHARED_GLOBAL
		undefine
			default_create
		end

create

	make_command

feature -- Creation

	make_command (a_socket: like outflow; a_cs: ROSE_CS; a_user_information: STRING) is
			-- create command with `a_socket'
		require
			socket_not_void: a_socket /= Void
		do
			outflow := a_socket
			cs := a_cs
			user_information := a_user_information
			make_thread
		ensure
			outflow_set: outflow /= Void
		end

feature -- Access

	user_information: STRING
		-- Username and machine for this thread

	expired: BOOLEAN
		-- is connection expired?		

	cs: ROSE_CS
		--cs server object

	pes_result_future: ROSE_CONCURRENT_FUTURE[ROSE_CS_RESULT]
		-- cs_worker should store the result here

	outflow: ROSE_NETWORK_STREAM_SOCKET
		-- out socket

	received: ANY
		-- received object			

feature -- Basic operations

	execute
			-- handle client requests
		local
			l_pes_query: ROSE_CS_QUERY
			l_tuple: TUPLE[query: ROSE_CS_QUERY; future: ROSE_CONCURRENT_FUTURE[ROSE_CS_RESULT]]
			l_retry: BOOLEAN
			l_result: ROSE_CS_RESULT
		do
			from

			until
				expired
			loop
				if not l_retry then
					receive
					l_pes_query ?= received
					if l_pes_query /= void and not expired then
						output(Global.interpreted("($now) received " + l_pes_query.query_id.out + "!%N"),Void)
						create l_tuple.default_create
						l_pes_query.set_userinformation (user_information)
						l_tuple.query := l_pes_query
						create pes_result_future.make
						l_tuple.future := pes_result_future
						cs.input_queue.extend (l_tuple)
						l_result := pes_result_future.get
						output(Global.interpreted("($now) computed " + l_pes_query.query_id.out + "!%N"),Void)
						send_result (l_result)
						output(Global.interpreted("($now) sent " + l_pes_query.query_id.out + "!%N"),Void)
						pes_result_future := Void
						received := Void
						l_result := Void
						l_tuple := Void
						l_pes_query := Void
					end
				else
					output(Global.interpreted("($now) expired!%N"),Void)
					set_expired (true)
				end
			end
		rescue
			l_retry := True
			retry

		end

	store_result (a_result: ROSE_CS_RESULT) is
		-- store result to file (for debugging)
		do
			Global.persist.store (a_result, "e:\result" + a_result.query_id.out + ".sto")
		end


	send_result (a_result: ROSE_CS_RESULT) is
			-- send `a_result'
		require
			valid_result: a_result /= void
		local
			l_retry: BOOLEAN
		do
			if not l_retry  then
				outflow.independent_store (a_result)
			else
				expired := true
			end
		rescue
			if not l_retry then
				l_retry := True
				retry
			end
		end

	close is
			-- implement close from SERVER
		do

		end

	cleanup is
			-- implement cleanup from SERVER
		do

		end


	receive is
			-- read message from outflow
		local
			l_retry: BOOLEAN
		do
			if not l_retry then
				received ?= outflow.retrieved
			else
				expired := true
			end
		rescue
			l_retry := true
			retry
		end

	receive_packet is
			-- Receive message blocking using the PACKET interface
		local
			l_packet: PACKET
			l_stream: STREAM
			l_received_data: MANAGED_POINTER
			l_sed_reader: ROSE_SED_MEMORY_READER_WRITER
			l_retried: BOOLEAN
			l_size: INTEGER_32
		do
			if not l_retried then
				outflow.read_integer_32
				l_size := outflow.last_integer_32
				l_packet := outflow.receive (l_size, 0)
				l_received_data := l_packet.data

				if is_sed_used then
					create l_sed_reader.make_with_buffer (l_received_data)
					l_sed_reader.set_for_reading
					received := sed.retrieved (l_sed_reader, False)
				else
					create l_stream.make_with_size (l_received_data.count)
					l_stream.item.item.memory_copy (l_received_data.item, l_received_data.count)
					received := l_stream.retrieved
				end
			else
				expired := True
			end
		rescue
			if not l_retried then
				l_retried := True
				retry
			end
		end

	send_packet (a_object: ANY) is
			-- Send `a_object' using the PACKET framework
		require
			object_not_void: a_object /= Void
		local
			l_packet: PACKET
			l_stream: STREAM
			l_pointer: MANAGED_POINTER
			l_sed_writer: ROSE_SED_MEMORY_READER_WRITER
			l_retried: BOOLEAN
		do
			if not l_retried then
				if is_sed_used then
					create l_sed_writer.make
					l_sed_writer.set_for_writing
					sed.independent_store (a_object, l_sed_writer, True)
					create l_packet.make_from_managed_pointer (l_sed_writer.data)
				else
					create l_stream.make_with_size (1000000000)
					l_stream.independent_store (a_object)
					create l_pointer.share_from_pointer (l_stream.item, l_stream.object_stored_size)
					create l_packet.make_from_managed_pointer (l_pointer)
				end
				outflow.put_integer_32 (l_packet.count)
				outflow.send (l_packet, 0)
			else
				expired := True
			end
		rescue
			if not l_retried then
				l_retried := True
				retry
			end
		end

	run is
			-- implement server command here
		do

		end

	set_expired (a_bol: BOOLEAN) is
			-- set `expired' to `a_bol'
		do
			expired := a_bol
		end


feature {NONE} -- Implementation

	sed: SED_STORABLE_FACILITIES is
			-- SED instance for serialization / deserialization
		once
			create Result
		end

	memory: MEMORY is
			-- Memory object
		once
			create Result
		end

	is_sed_used: BOOLEAN is
			-- Is the SED serialization framework used ?
		do
			Result := outflow.is_sed_used
		end

invariant
	running_or_sleeping: (outflow /= void) or expired
end
