note

	description:

		"Definition of a database entry"

	storable_version: "1"
	last_updater: "$Author: rj $"
	author: "Zoran Simic"
	last_editor: "$Author: rj $"
	date: "$Date: 2015/03/25 $"
	revision: "$Revision: #22 $"
	path: "$File: //rose/source/kernel/infrastructure/framework/db/rose_database_definition.e $"

class ROSE_DATABASE_DEFINITION

inherit

	DATABASE_DEFINITION
		undefine is_equal, out end

	DEBUG_OUTPUT
		redefine
			is_equal, out
		end

	KL_SHARED_STRING_EQUALITY_TESTER
		export
			{NONE} all
		undefine
			is_equal, out
		end

	MISMATCH_CORRECTOR
		export
			{NONE} all
		undefine
			is_equal, out
		redefine
			correct_mismatch
		end

create

	make, make_anonymous, make_from_path, make_from_other, default_create

feature {NONE} -- Initialization

	make (a_source, a_logical, a_connection_type, a_server, a_path: STRING) is
			-- Create fully described database definition
		require
			source_ok: a_source /= Void
			logical_ok: a_logical /= Void
			connection_type_ok: a_connection_type /= Void
			server_ok: a_server /= Void
			path_ok: a_path /= Void
		do
			source := a_source
			logical := a_logical.as_lower
			connection_type := a_connection_type
			server := a_server
			path := a_path
		end

	make_anonymous (a_source, a_connection_type, a_server, a_path: STRING) is
			-- Create anonymous (ie unspecified source) database definition
		require
			source_ok: a_source /= Void
			connection_type_ok: a_connection_type /= Void
			server_ok: a_server /= Void
			path_ok: a_path /= Void
		do
			make (a_source, k_anonymous + anonymous_count.item.out, a_connection_type, a_server, a_path)
			anonymous_count.put (anonymous_count.item + 1)
		end

	make_from_path (a_source, a_logical, a_path: STRING) is
			-- Create from logical and path.
		require
			source_ok: a_source /= Void
			logical_ok: a_logical /= Void
			path_ok: a_path /= Void
		local
			i, j: INTEGER
		do
			source := a_source
			logical := a_logical.as_lower
			i := a_path.index_of (':', 1)
			if i > 0 then
				connection_type := a_path.substring (1, i - 1)
				j := a_path.index_of ('.', i + 1)
				if j > 0 then
					server := a_path.substring (i + 1, j - 1)
					path := a_path.substring (j + 1, a_path.count)
				end
			end
			if connection_type = Void then connection_type := "" end
			if server = Void then server := "" end
			if path = Void then path := "" end
		end

	make_from_other (a_other: ROSE_DATABASE_DEFINITION) is
			-- Create as copy of 'a_other'
		require
			other_ok: a_other /= Void
		do
			make (a_other.source, a_other.logical, a_other.connection_type, a_other.server, a_other.path)
		end

feature -- Access

	source: STRING
			-- Where this definition was found (repository, application config,...)

	logical: STRING
			-- Logical name for the database defined here

	connection_type: STRING
			-- Connection type (SQL Server, RDB service, RDB generic,...)

	server: STRING
			-- Server name where the physical database sits

	path: STRING
			-- Path to physical database (name)

	overridden: ROSE_DATABASE_DEFINITION
			-- Definition this one overrides

	overridden_by: ROSE_DATABASE_DEFINITION
			-- Definition this one is overridden by

	is_valid: BOOLEAN is
			-- Is this definition currently valid?
		do
			Result := source /= Void and then logical /= Void and then connection_type /= Void and then server /= Void and then path /= Void
		end

	is_equal (a_other: ROSE_DATABASE_DEFINITION): BOOLEAN is
			-- Is this definition equivalent to 'a_other'?
		do
			Result := a_other /= Void
				and then equal (source, a_other.source)
				and then equal (logical, a_other.logical)
				and then equal (server, a_other.server)
				and then equal (path, a_other.path)
		end

	is_same_source_and_logical (a_other: ROSE_DATABASE_DEFINITION): BOOLEAN is
			-- Does 'a_other' define the same logical, for same source?
		do
			Result := a_other /= Void and then equal (source, a_other.source)
				and then equal (logical, a_other.logical)
		end

	is_same_path (a_other: ROSE_DATABASE_DEFINITION): BOOLEAN is
			-- Is this definition pointing to same database as 'a_other'?
		do
			Result := a_other /= Void
				and then logical /= Void and then a_other.logical /= Void
				and then server /= Void and then a_other.server /= Void
				and then path /= Void and then a_other.path /= Void
				and then equal (logical.as_upper, a_other.logical.as_upper)
				and then equal (server.as_upper, a_other.server.as_upper)
				and then equal (path.as_upper, a_other.path.as_upper)
		end

	is_accessed: BOOLEAN is
			-- Was the database pointed to this definition ever accessed (during the execution of current process)?
		do
			Result := access_count > 0
		end

	is_written: BOOLEAN is
			-- Were any data written to this database?
		do
			Result := execution_count > 0
		end

	is_anonymous: BOOLEAN is
			-- Was this definition created anonymous?
		do
			Result := logical /= Void and then logical.has_substring (k_anonymous)
		end

	access_count: INTEGER
			-- Number of times a DATA_SERVER object was created using this definition

	selection_count: INTEGER
			-- Number of times a select statement was executed on this database

	execution_count: INTEGER
			-- Number of execute statements performed on this database

	input_table_names: DS_HASH_SET [STRING]
			-- Names of tables which have been accessed as input

	output_table_names: DS_HASH_SET [STRING]
			-- Names of tables which have been accessed as output

	full_path: STRING is
			-- Full path
		do
			create Result.make (20)
			Result.append_string (connection_type)
			Result.append_character (':')
			Result.append_string (server)
			Result.append_character ('.')
			Result.append_string (path)
		end

	representation: STRING is
			-- Textual representation
		do
			create Result.make (20)
			Result.append_string (out)
			if is_fallback then
				Result.append_string (" [from (fallback) '")
			else
				Result.append_string (" [from '")
			end
			Result.append_string (source)
			Result.append_character ('%'')
			Result.append_character (']')
		end

	out, debug_output: STRING is
			-- New string containing terse printable representation of current object
		do
			create Result.make (20)
			Result.append_string (logical)
			Result.append_character ('=')
			Result.append_string (full_path)
		end

	is_fallback: BOOLEAN
			-- Does this represent a fallback to the production database from a testing environment?

feature -- Setting

	mark_as_fallback is
			-- Mark `Current' as a fallback to the production database from a testing environment
		do
			is_fallback := True
		ensure
			marked_As_fallback: is_fallback
		end

feature -- Basic operations

	set_path (a_path: STRING) is
			-- Set 'path'.
		require
			path_ok: a_path /= Void
		do
			path := a_path
		end

	increment_bulk_insert_count (a_table: STRING)
			-- Add one to count of BULK INSERTs committed against `a_table'.
		do
			increment_tracking_table_count (k_bulks, a_table)
		ensure then
			hash_table_created: bulk_inserts /= Void
		end

	increment_truncate_count (a_table: STRING)
			-- Add one to count of TRUNCATE TABLE issued for `a_table'.
		do
			increment_tracking_table_count (k_truncates, a_table)
		ensure then
			hash_table_created: truncates /= Void
		end

	increment_create_count (a_table: STRING)
			-- Add one to count of CREATE TABLE issued for `a_table'.
		do
			increment_tracking_table_count (k_creates, a_table)
		ensure then
			hash_table_created: creates /= Void
		end

	increment_drop_count (a_table: STRING)
			-- Add one to count of DROP TABLE issued for `a_table'.
		do
			increment_tracking_table_count (k_drops, a_table)
		ensure then
			hash_table_created: drops /= Void
		end

	add_input_table_name (a_table_name: STRING)
			-- Indicate that `a_table_name' has been accessed as input.
		local
			l_table_name: STRING
		do
			if a_table_name.starts_with ("#") then
				-- Replace all digits by a single asterisk.
				--		##temp1234				->		##temp*
				--		#signature123_temp654	->		#signature*_temp*
				match_digits_regexp.match (a_table_name)
				if match_digits_regexp.has_matched then
					l_table_name := match_digits_regexp.replace_all ("*")
				else
					l_table_name := a_table_name
				end
			else
				l_table_name := a_table_name
			end
			if input_table_names = Void then
				create input_table_names.make_equal (1)
			end
			input_table_names.force (l_table_name)
		end

	add_output_table_name (a_table_name: STRING)
			-- Indicate that `a_table_name' has been accessed as output.
		local
			l_table_name: STRING
		do
			if a_table_name.starts_with ("#") then
				-- Replace all digits by a single asterisk.
				--		##temp1234				->		##temp*
				--		#signature123_temp654	->		#signature*_temp*
				match_digits_regexp.match (a_table_name)
				if match_digits_regexp.has_matched then
					l_table_name := match_digits_regexp.replace_all ("*")
				else
					l_table_name := a_table_name
				end
			else
				l_table_name := a_table_name
			end
			if output_table_names = Void then
				create output_table_names.make_equal (1)
			end
			output_table_names.force (a_table_name)
		end

feature {ROSE_DATABASE_POOL} -- Metrics implementation

	is_valid_tracking_id (a_id: STRING): BOOLEAN
			-- Is 'a_id' valid for use with 'access_counts_hash'?
		do
			Result := a_id = k_bulks or else a_id = k_truncates or else a_id = k_creates or else a_id = k_drops
		end

	bulk_inserts: DS_HASH_TABLE [INTEGER, STRING]
			-- Counts of BULK INSERTs commited for each table
		do
			Result := access_counts_hash.item (tracking_table_key (k_bulks))
		end

	truncates: DS_HASH_TABLE [INTEGER, STRING]
			-- Counts of TRUNCATE TABLEs issued for each table
		do
			Result := access_counts_hash.item (tracking_table_key (k_truncates))
		end

	creates: DS_HASH_TABLE [INTEGER, STRING]
			-- Counts of CREATE TABLEs issued for each table
		do
			Result := access_counts_hash.item (tracking_table_key (k_creates))
		end

	drops: DS_HASH_TABLE [INTEGER, STRING]
			-- Counts of DROP TABLEs issued for each table
		do
			Result := access_counts_hash.item (tracking_table_key (k_drops))
		end

	access_counts_hash: HASH_TABLE [DS_HASH_TABLE [INTEGER, STRING], STRING]
			-- Hash table containing the various counters
		once
			create Result.make (0)
		end

feature {ROSE_DATABASE_DEFINITION} -- Implementation

	set_overridden_by (a_overridden_by: ROSE_DATABASE_DEFINITION) is
			-- Set 'overridden_by'.
		do
			overridden_by := a_overridden_by
		end

feature {ROSE_DATABASE_POOL} -- Implementation

	stamp: INTEGER
			-- Stamp, used to determine which definitions were deleted after a refresh
			-- 0 -> means this is a new defintion (after refresh)
			-- = refresh stamp -> means that this definition was there before and still there
			-- < refresh stamp -> the definition is not there anymore (should be removed)

	is_stamped (a_stamp: INTEGER): BOOLEAN is
			-- Is this definition currently not stamped or stamped with 'a_stamp'?
		do
			Result := stamp = 0 or else a_stamp = stamp
		end

	set_overridden (a_overridden: ROSE_DATABASE_DEFINITION) is
			-- Set 'overriden'
		do
			overridden := a_overridden
			if a_overridden /= Void then
				overridden.set_overridden_by (Current)
			end
		end

	set_stamp (a_stamp: INTEGER) is
			-- Set 'stamp'.
		do
			stamp := a_stamp
		end

feature {ROSE_DATABASE_POOL, SQLSRV_DATA_SERVER} -- Implementation

	increment_access_count is
			-- Increment 'access_count'
		do
			access_count := access_count + 1
		end

feature {GENERAL_SELECTION_PREDICATE} -- Implementation

	increment_selection_count is
			-- Increment 'selection_count'
		do
			selection_count := selection_count + 1
		end

feature {EXECUTION_PREDICATE, OLEDB_FAST_LOAD} -- Implementation

	increment_execution_count is
			-- Increment 'execution_count'
		do
			execution_count := execution_count + 1
		end

feature -- Constants

	k_anonymous: STRING = "**anonymous**_"

feature {NONE} -- Implementation

	k_bulks: STRING = "__bulks__"				-- Identifies bulk inserts
	k_truncates: STRING = "__truncates__"		-- Identifies truncates
	k_creates: STRING = "__creates__"			-- Identifies creates
	k_drops: STRING = "__drops__"				-- Identifies drops

	tracking_table_key (a_id: STRING): STRING
			-- Key to use in 'access_counts_hash' for corresponding 'a_id' kind
		require
			logical_ok: logical /= Void
			id_ok: is_valid_tracking_id (a_id)
		do
			Result := logical + a_id
		ensure
			result_ok: Result /= Void
		end

	increment_tracking_table_count (a_id, a_table: STRING)
			-- Increment tracking count for 'a_table' with 'a_id'
		require
			id_ok: is_valid_tracking_id (a_id)
		local
			l_hash: DS_HASH_TABLE [INTEGER, STRING]
			l_key: STRING
			l_count: INTEGER
			l_table_name: STRING
		do
			check
				not_overridden: overridden_by = Void
				-- ensured because current clients do not see overridden definitions
			end

			if a_table.count>0 and then a_table[1]='#'
			then
				-- Replace all digits by a single asterisk.
				--		##temp1234				->		##temp*
				--		#signature123_temp654	->		#signature*_temp*
				match_digits_regexp.match(a_table)
				if match_digits_regexp.has_matched
				then
					l_table_name := match_digits_regexp.replace_all("*")
				else
					l_table_name := a_table
				end
			else
				l_table_name := a_table
			end

			l_key := tracking_table_key (a_id)
			l_hash := access_counts_hash.item (l_key)
			if l_hash = Void then
				create l_hash.make_with_equality_testers (10, Void, case_insensitive_string_equality_tester)
				access_counts_hash.force (l_hash, l_key)
			end
			if l_hash.has (l_table_name) then
				l_count := l_hash.item (l_table_name) + 1
				l_hash.replace (l_count, l_table_name)
			else
				l_hash.force_new (1, l_table_name)
			end
		end

	match_digits_regexp: LX_DFA_REGULAR_EXPRESSION
			-- A regular expression that matches strings of digits.
		once
			create Result.compile_case_sensitive("[0-9]+")
		end

	anonymous_count: attached CELL [INTEGER]
			-- Counter for unique anonymous source names
		once
			create Result.put (1)
		end

feature {NONE} -- Storable mismatch

	correct_mismatch
			-- <Precursor>
		local
			l_stored_version_number: INTEGER
		do
			if not attached mismatch_information.stored_version as l_stored_version or else l_stored_version.is_empty then
				correct_mismatch_1
			elseif l_stored_version.is_integer then
				l_stored_version_number := l_stored_version.to_integer
				if l_stored_version_number < 1 then
					correct_mismatch_1
				else
					precursor
				end
			else
				precursor
			end
		end

	correct_mismatch_1
			-- Correct storable mismatch introducted in version "1".
		do
			-- Attributes `input_table_names' and `output_table_names' have
			-- been added. Nothing to be done to correct this mismatch.
		end

end
