note

	description:

		"Platform specific database objects"

	last_updater: "$Author: e2 $"
	author: "Zoran Simic"
	last_editor: "$Author: e2 $"
	date: "$Date: 2015/04/16 $"
	revision: "$Revision: #34 $"
	path: "$File: //rose/source/kernel/infrastructure/windows/rose_database_platform_specific.e $"

class
	ROSE_DATABASE_PLATFORM_SPECIFIC -- Windows version

inherit
	ROSE_SHARED_GLOBAL
	ROSE_FORMATTING
	DATABASE_SERVER

feature -- Access

	force_transactions_off: BOOLEAN
			-- Force all returned servers to 'set_implicit_transactions_off'?

	get_server (a_definition: ROSE_DATABASE_DEFINITION): DATA_SERVER is
			-- A server based on 'a_definition'
		require
			valid_definition: a_definition /= Void and then a_definition.is_valid
		do
			if (a_definition.connection_type ~ global.convention.db_type_sql) or else (a_definition.connection_type ~ global.convention.db_type_ms_sql) then
				Result := get_sql (a_definition)
				set_approle_if_appropriate (Result, a_definition)
			elseif (a_definition.connection_type ~ global.convention.db_type_rdb_service) or else (a_definition.connection_type ~ global.convention.db_type_rdb_generic) then
				fail("get_server", "VMS is no longer supported ($1)", <<a_definition.out>>)
			end
		end

	get_fast_load_server (a_definition: ROSE_DATABASE_DEFINITION): DATA_SERVER is
			-- A fast load server based on 'a_definition'
		require
			valid_definition: a_definition /= Void and then a_definition.is_valid
		do
			if (a_definition.connection_type ~ global.convention.db_type_sql) or else (a_definition.connection_type ~ global.convention.db_type_ms_sql) then
				Result := get_sql_fast_loader (a_definition)
				set_approle_if_appropriate (Result, a_definition)
			end
		end

feature -- Basic operations

	set_force_transactions_off (a_value: BOOLEAN) is
			-- Set 'force_transactions_off' to 'a_value'
		do
			force_transactions_off := a_value
		end

feature {NONE} -- Approles

	set_approle_if_appropriate (a_server: DATA_SERVER; a_definition: ROSE_DATABASE_DEFINITION) is
			-- Enable the application role on 'a_server' if the configuration requires it
		local
			l_sql_server: SQL_DATA_SERVER
			l_protected: STRING
		do
			l_sql_server ?= a_server
			if l_sql_server /= Void and a_definition /= Void and (a_definition.connection_type ~ global.convention.db_type_sql) then
				if a_server.is_production or a_definition.is_fallback then
					if not Global.app.is_running_in_production and then
						Global.app.configured_value (enable_production_database_write_protection) /= Void and then
						Global.app.configured_value (enable_production_database_write_protection).is_boolean and then
						Global.app.configured_value (enable_production_database_write_protection).to_boolean
					then
					   create l_protected.make_from_string ("Production database write protection is enabled")
						if l_sql_server.node /= Void and l_sql_server.database /= Void then
							l_protected := l_protected + " on " + l_sql_server.node + "." + l_sql_server.database
						end
						if read_write_approle_logicals.has (a_definition.logical.as_lower) then
							if read_write_server_approle.has_approle (l_sql_server) then
								read_write_server_approle.set_approle (l_sql_server)
								inform ("set_approle_if_appropriate", l_protected, Void)
							end
						else
							if readonly_server_approle.has_approle (l_sql_server) then
								readonly_server_approle.set_approle (l_sql_server)
								inform ("set_approle_if_appropriate", l_protected, Void)
							end
						end
					else
						-- This should be defined in some repository for servers that are protected by application roles
						if read_write_approle_logicals.has (a_definition.logical.as_lower) then
							if read_write_server_approle.has_approle (l_sql_server) then
								read_write_server_approle.set_approle (l_sql_server)
							end
						end
					end
				end
			end
		end

	enable_production_database_write_protection: STRING is "enable_production_database_write_protection"
			-- Enable production DB write protection option name

	read_write_approle_logicals: ROSE_LINEAR_ARRAY [STRING] is
			-- List of logicals that will enforce read write approle usage
		once
			create Result.make_empty
		end

	readonly_server_approle: SQL_SERVER_APPROLE is
			-- The read only application role for data protection
		once
			-- Simple password used to make it easy to enable this application role
			create Result.make ("rc_readonly_approle","123", "select")
		end

	read_write_server_approle: SQL_SERVER_APPROLE is
			-- The read write application role for data protection, the setting of this application role is audited
		once
			-- Password generated by strongpasswordgenerator.com
			create Result.make ("rc_read_write_approle","253Q#6=6/#6<uXw98Fbj", "all")
		end

feature {NONE} -- Implementation

	get_sql (a_definition: ROSE_DATABASE_DEFINITION): SQLOLEDB_DATA_SERVER is
			-- Data server for 'a_definition'
		require
			valid_definition: a_definition /= Void
		local
			l_ehp: OLEDB_ERROR_HANDLER_PRINT
		do
			if sql_database_is_online (a_definition) and then sql_has_access (a_definition) then
				create Result.make_integrated_security (a_definition.server, a_definition.path)
				if Result.make_succeeded then
					create l_ehp
					Result.set_error_handler (l_ehp)
					if force_transactions_off then
						Result.set_implicit_transaction_off
					end
				else
					Result := Void
				end
			end
		end

	get_sql_fast_loader (a_definition: ROSE_DATABASE_DEFINITION): SQLOLEDB_DATA_SERVER is
			-- Fast loader for 'a_definition'
		require
			valid_definition: a_definition /= Void
		local
			l_ehp: OLEDB_ERROR_HANDLER_PRINT
		do
			if sql_database_is_online (a_definition) and then sql_has_access (a_definition) then
				create Result.make_integrated_security_fast_load (a_definition.server, a_definition.path)
				if Result.make_succeeded then
					create l_ehp
					Result.set_error_handler (l_ehp)
					if force_transactions_off then
						Result.set_implicit_transaction_off
					end
				else
					Result := Void
				end
			end
		end

	k_online: STRING = "ONLINE"
			-- Response given by SQL when a DB is online

feature {ROSE_DATABASE_PLATFORM_SPECIFIC_ACCESS} -- DB checks

	sql_database_is_online (a_definition: ROSE_DATABASE_DEFINITION): BOOLEAN is
			-- Is the database described by `a_definition' online?
		require
			valid_definition: a_definition /= Void
		local
			l_state_predicate: STRING_PREDICATE
			l_server: SQLOLEDB_DATA_SERVER
		do
			l_server := servers.master_for_server (a_definition.server)
			if l_server /= Void then
				create l_state_predicate.make (l_server, "SELECT state_desc FROM sys.databases WHERE name = '" + a_definition.path + "'")
				l_state_predicate.retrieve (Void,Void)
				Result := l_state_predicate.success and then
					l_state_predicate.array.count = 1 and then
					l_state_predicate.array.first ~ k_online
			end
			if not Result then
				if l_server = Void then
					error.warn ("Cannot find DB for definition '$1'", <<a_definition.server>>)
				elseif not l_state_predicate.success then
					error.warn ("Cannot execute predicate to find out whether database $1 is online; see console error output%N", <<a_definition.full_path>>)
				elseif l_state_predicate.array.count /= 1 then
					error.warn ("Predicate to find out whether databse $1 is online didn't return a unique record but $2%N", <<a_definition.full_path, l_state_predicate.array.count>>)
				else
					error.warn ("Database $1 is not online ($2)%N", <<a_definition.full_path, l_state_predicate.array.first>>)
				end
				error.flush_and_acknowledge
			end
		end

	sql_has_access (a_definition: ROSE_DATABASE_DEFINITION): BOOLEAN is
			-- Does this user have access to the database described in 'a_definition'?
		require
			valid_definition: a_definition /= Void
		local
			l_id_predicate: INTEGER_PREDICATE
			l_server: SQLOLEDB_DATA_SERVER
		do
			l_server := servers.master_for_server (a_definition.server)
			if l_server /= Void then
				create l_id_predicate.make (l_server,"select has_dbaccess('" + a_definition.path + "')")
				l_id_predicate.retrieve (Void,Void)
				Result := l_id_predicate.array.first = 1
			end
			if not Result then
				error.warn ("User does not have access to database $1%N", <<a_definition.full_path>>)
				error.flush_and_acknowledge
			end
		end

end
