indexing
	description: "Database server pool"
	author: "Zoran Simic"
	last_editor: "$Author: rj $"
	date: "$Date: 2015/01/08 $"
	revision: "$Revision: #81 $"
    path: "$File: //rose/source/kernel/infrastructure/framework/db/rose_database_pool.e $"

class
	ROSE_DATABASE_POOL

inherit
	ROSE_SHARED_GLOBAL

create {ROSE_SHARED_DATABASE_POOL}
	make

feature {NONE} -- Initialization

	make
			-- Creation procedure.
		local
			l_access: ROSE_RESOURCE_ACCESS
		do
			if {PLATFORM}.is_thread_capable then
				create mutex.make
			end
			create definition_hash.make (0)
			create accessed.make_empty
			create refresh_event
			create already_warned.make (0)
			create sql_statement_preparation_counts.make (0)
			sql_preparation_warning_threshold := 0
			user_filter := Global.app.user.full_login.as_lower
			create l_access
			Global.add_cleanup_agent (agent l_access.persist_for_resource_access_tracking (Current))
			Global.add_cleanup_agent (agent record_metrics)
		end

feature -- Access

	repository_logical: attached STRING = "repository"
			-- The logical name that refers to the repository itself

	sql_preparation_warning_threshold: INTEGER
			-- Warn (in debugger) after any given SQL statment has been prepared 'preparation_warning_threshold' times
			-- if > 0, a warning is printed and a non fatal exception is raised if application is running in debugger
			-- 0 means don't warn

	repository_definition: ROSE_DATABASE_DEFINITION
			-- The repository definition
		do
			if repository_definition_i = Void then
				lock
				find_repository_from_configuration (Global.app.config)
				find_repository_from_configuration (Global.app.delivery)
				find_repository_from_configuration (Global.app.site)
				unlock
			end
			Result := repository_definition_i
		end

	definition_hash: attached HASH_TABLE [ROSE_DATABASE_DEFINITION, STRING]
			-- All available definitions hashed by logical

	accessed: attached ROSE_LINEAR_ARRAY [ROSE_DATABASE_DEFINITION]
			-- All accessed databases, in order in which they were accessed

	refresh_event: attached ROSE_SIMPLE_EVENT
			-- Triggered whenever 'refresh' is called

	already_warned: attached HASH_TABLE [BOOLEAN, STRING]
			-- Logicals for which a warning was already issued

	count: INTEGER
			-- Number of definitions
		do
			lock
			Result := definition_hash.count
			unlock
		ensure
			positive: Result >= 0
		end

	get_repository_server: SQL_DATA_SERVER
			-- Data server object connect to the repository itself
		do
			if repository_definition /= Void then
				Result := get_server_from_definition (repository_definition)
			end
		ensure
			comes_from_repository_defition: Result /= Void implies (Result.definition = repository_definition)
		end

	definition (a_logical: STRING): ROSE_DATABASE_DEFINITION
			-- Definition corresponding to 'a_logical', if any
		do
			refresh_if_needed
			if a_logical /= Void then
				lock
				Result := definition_hash.item (a_logical.as_lower)
				unlock
			end
		ensure
			matches_logical: Result /= Void implies (Result.logical ~ a_logical.as_lower)
		end

	get_server_from_definition (a_definition: ROSE_DATABASE_DEFINITION): SQL_DATA_SERVER
			-- Server pointed to by 'a_definition'
		require
			valid_definition: a_definition /= Void
		do
			lock
			check_override (a_definition)
			Result ?= rose_db_platform.get_server (a_definition)
			if is_connected (Result) then
				a_definition.increment_access_count
				Result.set_definition (a_definition)
				if not accessed.has (a_definition) then
					accessed.extend (a_definition)
				end
			else
				error.signal (generator, "Cannot connect to '$1'", <<a_definition.representation>>)
				error.flush_and_acknowledge
				Result := Void
			end
			error.flush_and_acknowledge
			unlock
		rescue
			unlock
		end

	get_server_from_logical (a_logical: STRING): DATA_SERVER
			-- Server pointed to by 'a_logical'
		require
			valid_logical: a_logical /= Void
		local
			l_definition: ROSE_DATABASE_DEFINITION
		do
			lock
			l_definition := checked_definition (a_logical)
			if l_definition /= Void then
				Result := get_server_from_definition (l_definition)
			end
			unlock
		ensure
			matches_logical: Result /= Void implies (Result.logical ~ a_logical.as_lower)
		rescue
			unlock
			error.signal (generator, "Connecting to '$1' crashed, exception trace:%N$2",<<a_logical, last_exception_explanation>>)
			error.flush_and_acknowledge
		end

	get_fast_load_server_from_definition (a_definition: ROSE_DATABASE_DEFINITION): DATA_SERVER
			-- Server pointed to by 'a_definition'
		require
			valid_definition: a_definition /= Void
		do
			lock
			check_override (a_definition)
			Result := rose_db_platform.get_fast_load_server (a_definition)
			if is_connected (Result) then
				a_definition.increment_access_count
				Result.set_definition (a_definition)
				if not accessed.has (a_definition) then
					accessed.extend (a_definition)
				end
			else
				error.signal (generator, "Cannot connect to '$1'", <<a_definition.representation>>)
				error.flush_and_acknowledge
				Result := Void
			end
			error.flush_and_acknowledge
			unlock
		ensure
			matches_logical: Result /= Void implies (Result.logical ~ a_definition.logical.as_lower)
		rescue
			unlock
		end

	get_fast_load_server_from_logical (a_logical: STRING): DATA_SERVER
			-- Server pointed to by 'a_logical'
		require
			valid_logical: a_logical /= Void
		local
			l_definition: ROSE_DATABASE_DEFINITION
		do
			lock
			l_definition := checked_definition (a_logical)
			if l_definition /= Void then
				Result := get_fast_load_server_from_definition (l_definition)
			end
			unlock
		ensure
			matches_logical: Result /= Void implies (Result.logical ~ a_logical.as_lower)
		rescue
			unlock
			error.signal (generator, "Connecting to '$1' crashed, exception trace:%N$2",<<a_logical, last_exception_explanation>>)
			error.flush_and_acknowledge
		end

	report: attached STRING
			--[250 reads 15 writes] rose_entity_foreground=sql:ARSQL5\RC.rose_entity_foreground_01 [from \\axaros.com\shares\production\common\config\db.cfg]
		local
			l_accessed: ROSE_DATABASE_DEFINITION
			i: INTEGER
		do
			lock
			if accessed.count = 0 then
				Result := "- No databases accessed -%N"
			else
				create Result.make (100)
				Result.append (accessed.count.out + " databases accessed (in order of access):%N")
				from
					i := accessed.lower
				until
					i > accessed.upper
				loop
					l_accessed := accessed.item (i)
					Result.append_character ('[')
					if l_accessed.selection_count > 0 then
						append_integer (Result, l_accessed.selection_count, "read")
						if l_accessed.execution_count > 0 then
							Result.append_character (' ')
						end
					end
					if l_accessed.execution_count > 0 then
						append_integer (Result, l_accessed.execution_count, "write")
					end
					Result.append_character (']')
					Result.append_character (' ')
					Result.append_string (l_accessed.representation)
					Result.append_character ('%N')
					i := i + 1
				end
			end
			unlock
		end

	is_connected (a_server: DATA_SERVER): BOOLEAN
			-- Is connection to 'a_server' established?
		do
			Result := a_server /= Void --| FIXME: and then a_server.is_connected
		end

	definition_array: attached ROSE_LINEAR_ARRAY [ROSE_DATABASE_DEFINITION]
			-- Array containing all current definition, sorted by logical
		local
			i: INTEGER
		do
			lock
			from
				create Result.make (1, count)
				i := 1
				definition_hash.start
			until
				definition_hash.off
			loop
				Result.put (definition_hash.item_for_iteration, i)
				definition_hash.forth
				i := i + 1
			end
			Result.sort_by (agent compare_logicals)
			unlock
		end

	inferred_production_databases_by_logical: attached HASH_TABLE [STRING, STRING]
			-- Table of databases, keyed by logical, that have been updated by production executables,
			-- regardless of whether or not the databases have been declared as production.
		once
			lock
			from
				create Result.make (0)
				inferred_production_databases_by_path.start
			until
				inferred_production_databases_by_path.off
			loop
				Result.force (inferred_production_databases_by_path.key_for_iteration, inferred_production_databases_by_path.item_for_iteration)
				inferred_production_databases_by_path.forth
			end
			unlock
		end

	force_transactions_off: BOOLEAN
			-- Force all returned servers to 'set_implicit_transactions_off'?
		do
			lock
			Result := rose_db_platform.force_transactions_off
			unlock
		end

feature -- Basic operations

	lock
			-- Lock 'mutex' if running in multithreaded more
		do
			if mutex /= Void then
				mutex.lock
			end
		end

	unlock
			-- Unlock 'mutex' if running in multithreaded more
		do
			if mutex /= Void then
				mutex.unlock
			end
		end

	extend (a_definition: ROSE_DATABASE_DEFINITION)
			-- Add 'a_definition' to this pool.
		do
			if a_definition /= Void then
				add_stamped (a_definition, 0)
			end
		end

	remove (a_definition: ROSE_DATABASE_DEFINITION)
			-- Remove 'a_definition' from this pool.
		do
			if a_definition /= Void then
				a_definition.set_stamp (-1)
				cleanup_definition (a_definition)
			end
		end

	set_force_transactions_off (a_value: BOOLEAN)
			-- Force automatic transactions to OFF by default for all newly created DATA_SERVER.
		do
			lock
			rose_db_platform.set_force_transactions_off (a_value)
			unlock
		end

	set_sql_preparation_warning_threshold (a_sql_preparation_warning_threshold: INTEGER)
			-- Set 'sql_preparation_warning_threshold'.
		do
			sql_preparation_warning_threshold := a_sql_preparation_warning_threshold
		end

	refresh_if_needed
			-- Call 'refresh' if not yet called.
		local
			retried: BOOLEAN
		do
			if not retried and then stamp = 0 then
				refresh
			end
		rescue
			if not retried then
				error.signal(generator, last_exception_explanation, Void)
				error.flush_and_acknowledge
				retried := True
				retry
			end
		end

	refresh
			-- Refresh definitions from the repository.
		local
			l_definition: ROSE_DATABASE_DEFINITION
			l_repository: SQL_DATA_SERVER
		do
			lock
			stamp := stamp + 1
			repository_definition_i := Void
			fallback_repository_definition_i := Void
			definition_hash.clear_all

			-- Non-production repositories are generally incomplete and rely on loading the release_candidate
			-- repository first.
			if not global.deliveries.is_in_top_level_delivery (global.app.site.folder) and then not global.app.is_running_in_production then
				find_fallback_repository	-- By 'fallback' we mean 'release_candidate'.
				if fallback_repository_definition_i /= Void then
					l_repository := get_server_from_definition (fallback_repository_definition_i)
				end
				if is_connected (l_repository) then
					load_repository (l_repository)
				else
					(create {EXCEPTIONS}).raise("The release_candidate respository could not be loaded")
				end
			end

			-- Now we load the desired repository, overwriting any definitions that were previously loaded.
			l_repository := get_repository_server
			if is_connected (l_repository) then
				load_repository (l_repository)
			else
				(create {EXCEPTIONS}).raise("The respository could not be loaded")
			end

			l_repository := Void
			add_overrides_from_config (Global.app.site)
			add_overrides_from_config (Global.app.delivery)
			add_overrides_from_config (Global.app.config)
			if hardcoded_overrides /= Void then
				from
					hardcoded_overrides.start
				until
					hardcoded_overrides.off
				loop
					if hardcoded_overrides.item_for_iteration /= Void
					and then not hardcoded_overrides.item_for_iteration.is_empty then
						create l_definition.make_from_path ("hardcoded", hardcoded_overrides.key_for_iteration, hardcoded_overrides.item_for_iteration)
						add_stamped (l_definition, stamp)
					end
					hardcoded_overrides.forth
				end
			end
			definition_hash.linear_representation.do_all (agent cleanup_definition)
			refresh_event.call
			unlock
		rescue
			repository_definition_i := Void
			fallback_repository_definition_i := Void
			definition_hash.clear_all
			unlock
		end

	set_hardcoded_override (a_logical, a_path: STRING)
			-- Override 'a_logical' to 'a_path'
			-- It is sometimes useful to configure a database from within Eiffel code
			-- but this means "hardcoding" the path to a database and should be used cautiously
			-- Overrides defined in this way take precedence over the configuration and the repository
		require
			logical_ok: a_logical /= Void and then not a_logical.is_empty
		do
			lock
			if hardcoded_overrides = Void then
				create hardcoded_overrides.make (0)
			end
			hardcoded_overrides.force (a_path, a_logical)
			refresh
			unlock
		end

	remove_hardcoded_override (a_logical: STRING)
			-- Remove an override to 'a_logical'
		require
			logical_ok: a_logical /= Void and then not a_logical.is_empty
		do
			lock
			if hardcoded_overrides /= Void then
				hardcoded_overrides.remove (a_logical)
			end
			refresh
			unlock
		rescue
			unlock
		end

	record_sql_statement_preparation (a_data_server: DATA_SERVER; a_statement: STRING)
			-- Record that 'a_statement' has been prepared for 'a_data_server'
		require
			data_server_ok: a_data_server /= Void
			statement_ok: a_statement /= Void
		local
			retried: BOOLEAN
			l_string: STRING
			n: INTEGER
		do
			if sql_preparation_warning_threshold > 0 and then not retried then
				lock
				if a_data_server.definition /= Void then
					l_string := a_data_server.definition.logical + ": " + a_statement
				else
					l_string := a_data_server.node + ": " + a_statement
				end
				n := sql_statement_preparation_counts.item (l_string) + 1
				sql_statement_preparation_counts.force (n, l_string)
				if n = sql_preparation_warning_threshold then
					-- Warn once, only when preparation count hits 'sql_preparation_warning_threshold'
					print ("WARNING: Same SQL statement has been prepared " + n.out + " times, please check the code. SQL Statement:%N" + l_string + "%N")
					if global.app.is_running_in_estudio then
						(create {EXCEPTIONS}).raise ("Warning: SQL Statement prepared too many times, please check")
					end
				end
				unlock
			end
		rescue
			if not retried then
				unlock
				retried := True
				retry
			end
		end

	record_metrics is
			-- Write DB metrics for each server.
		do
			if Global.log_id /= 0 then
				accessed.do_all (agent record_metrics_for_server)
			end
		end

feature {NONE} -- Production database management

	k_process_metrics: STRING = "process_metrics"		-- Process metrics logical

	inferred_production_databases_by_path: attached HASH_TABLE [STRING,STRING]
			-- Table of databases, keyed by path, that have been updated by production executables,
			-- regardless of whether or not the databases have been declared as production.
		local
			l_predicate: COLLECTING_PREDICATE [ROSE_DATABASE_DEFINITION]
			l_config: ROSE_CONFIGURATION
			l_repository_definition: ROSE_DATABASE_DEFINITION
			l_retry: BOOLEAN
			l_repository_logical, l_main_site_cfg_path: STRING
			l_production_process_metrics_server, l_production_repository_server: SQL_DATA_SERVER
		once
			create Result.make(100)
			if not l_retry then
				create l_config.make_default
				l_main_site_cfg_path := global.hardcoded_path ("\\axaros.com\shares\production\site.cfg")
				Global.config_manager.load (l_config, Void, l_main_site_cfg_path)
				l_repository_logical := l_config.value ("/path/repository")
				if l_repository_logical /= Void then
					create l_repository_definition.make_from_path ("","repository",l_repository_logical)
					l_production_repository_server ?= rose_db_platform.get_server (l_repository_definition)
					if l_production_repository_server /= Void and then l_production_repository_server.success then
						create l_predicate.make_strict (l_production_repository_server, "SELECT server,[database] as path FROM databases WITH(NOLOCK) where logical = '" + k_process_metrics + "'")
						l_predicate.retrieve (Void, Void)
						if not l_predicate.array.is_empty then
							create l_repository_definition.make("",k_process_metrics,"sql",l_predicate.array.first.server,l_predicate.array.first.path)
							l_production_process_metrics_server ?= rose_db_platform.get_server (l_repository_definition)
							if l_production_process_metrics_server /= Void and then l_production_repository_server.success then
								create l_predicate.make_strict (l_production_process_metrics_server,"SELECT logical,path FROM production_databases WITH(NOLOCK)")
								l_predicate.retrieve(Void,Void)
								from l_predicate.array.start until l_predicate.array.off loop
									Result.force(l_predicate.array.current_item.logical.as_lower,l_predicate.array.current_item.path.as_lower)
									l_predicate.array.forth
								end
							end
						end
					end
				end
			end
			if Result.is_empty then
				error.warn ("No production database definitions are available",Void)
				error.flush_and_acknowledge
			end
		rescue
			l_retry := True
			retry
		end

feature {ROSE_RESOURCE_ACCESS} -- Implementation

	stamp: INTEGER
			-- Stamp used to detect which definitions were removed since last refresh

	rose_db_platform: attached ROSE_DATABASE_PLATFORM_SPECIFIC
			-- Platform specific creators
		once
			create Result
		end

	sql_statement_preparation_counts: attached HASH_TABLE [INTEGER, STRING]
			-- Stores how many time a given SQL statement has been prepared

feature {NONE} -- Implementation

	mutex: CRITICAL_SECTION
			-- Mutex used to serialize multithreaded access to the pool

	hardcoded_overrides: HASH_TABLE [STRING, STRING]
			-- Overrides provided in a hardcoded was from the Eiffel code itself

	repository_definition_i: ROSE_DATABASE_DEFINITION
			-- Cached repository definition


	fallback_repository_definition_i: ROSE_DATABASE_DEFINITION
			-- Cached fallback repository definition

	user_filter: STRING									-- Current user
	k_path_repository: STRING = "/path/repository"		-- Key in config pointing to repository
	k_true: STRING = "true"								-- True constant

	load_repository (a_repository: SQL_DATA_SERVER)
			-- Add logicals found in `a_repository'.
		require
			repository_connected: is_connected (a_repository)
			definition_not_void: a_repository.definition /= Void
		local
			l_statement: COLLECTING_PREDICATE [ROSE_REPOSITORY_RECORD]
		do
			a_repository.begin_transaction
			create l_statement.make (a_repository,
				"SELECT logical,server,[database],user_id,enabled FROM databases ORDER BY logical,user_id")
			if not l_statement.success then
				error.warn ("Cannot read repository from '$1'", <<a_repository.definition.full_path>>)
				error.flush_and_acknowledge
			else
				l_statement.apply (agent add_record (?, a_repository.definition), Void, Void)
			end
			a_repository.commit
			-- Release the DB resources without waiting for a GC:
			l_statement.facilitator.release
			a_repository.sql_disconnect
			l_statement := Void
		end

	checked_definition (a_logical: STRING): ROSE_DATABASE_DEFINITION
			-- Definition corresponding to 'a_logical', with error signaled if no such logical is defined
		require
			logical_ok: a_logical /= Void
		do
			Result := definition (a_logical)
			if Result = Void then
				if not already_warned.item (a_logical.as_lower) and then not a_logical.starts_with ({ROSE_DATABASE_DEFINITION}.k_anonymous) then
					error.signal (generator, "No logical database '$1' defined in configuration", <<a_logical>>)
					already_warned.put (True, a_logical.as_lower)
				end
			end
			error.flush_and_acknowledge
		end

	check_override (a_definition: ROSE_DATABASE_DEFINITION)
			-- Check override status for 'a_definition' and issue a warning if needed
		require
			definition_ok: a_definition /= Void
		local
			l_overridden: ROSE_DATABASE_DEFINITION
			l_warning: STRING
			n: INTEGER
		do
			if a_definition.overridden /= Void and then not already_warned.item (a_definition.logical.as_lower) then
				from
					l_overridden := a_definition.overridden
				until
					l_overridden = Void
				loop
					if should_warn_for_override (a_definition, l_overridden) then
						n := n + 1
						if l_warning = Void then
							create l_warning.make (30)
						end
						l_warning.append_character ('-')
						l_warning.append_character ('-')
						l_warning.append_character ('>')
						l_warning.append_character (' ')
						l_warning.append_string (l_overridden.representation)
						l_warning.append_character ('%N')
					end
					l_overridden := l_overridden.overridden
				end
				if l_warning /= Void then
					already_warned.put (True, a_definition.logical.as_lower)
					error.warn ("$1 overrides active for DB $2:%N$3", <<n, a_definition.representation, l_warning>>)
					error.flush_and_acknowledge
				end
			end
		end

	should_warn_for_override (a_definition, a_overridden: ROSE_DATABASE_DEFINITION): BOOLEAN
			-- Should we warn that 'a_definition' overrides 'a_overridden'?
		require
			definition_ok: a_definition /= Void
			overriden_ok: a_overridden /= Void
		do
			if not a_definition.is_same_path (a_overridden) then
				if a_definition.logical /~ k_process_metrics or else not is_legitimate_override_to_deployment_test_process_metrics (a_definition) then
					if a_definition /= repository_definition_i and then a_overridden /= fallback_repository_definition_i then
						-- Warn only for entries overriding something other than "fallback"
						Result := fallback_repository_definition_i = Void or else a_overridden.source /~ fallback_repository_definition_i.out
					end
				end
			end
		end

	is_legitimate_override_to_deployment_test_process_metrics (a_definition: ROSE_DATABASE_DEFINITION): BOOLEAN
			-- Is 'a_definition' a legitimate redirection to deployment_test process metrics?
			-- Valid only when running in Rose Studio or rose_studio_tester
		require
			process_metrics: a_definition /= Void and then a_definition.logical ~ k_process_metrics
		local
			l_program: STRING
		do
			l_program := global.app.program_name.as_lower
			if l_program.starts_with ("rose_studio") then
				Result := a_definition.full_path.as_lower ~ "sql:rcsql6\x64devel03.deployment_test_process_metrics"
			end
		end

	cleanup_definition (a_definition: ROSE_DATABASE_DEFINITION)
			-- Clean up 'a_definition' (remove definitions that were removed since last refresh).
		require
			definition_ok: a_definition /= Void
		local
			l_overridden: ROSE_DATABASE_DEFINITION
		do
			lock
			from
				-- First, we remove all the overridden definitions that were not stamped as visited
				l_overridden := a_definition.overridden
			until
				l_overridden = Void
			loop
				if not l_overridden.is_stamped (stamp) then
					l_overridden.overridden_by.set_overridden (l_overridden.overridden)
				end
				l_overridden := l_overridden.overridden
			end
			if not a_definition.is_stamped (stamp) then
				-- Then, we remove or replace the main definition (only those that are not stamped as visited)
				l_overridden := a_definition.overridden
				if l_overridden = Void then
					-- If the unstamped main definition was not overriding anything, we simply remove it
					definition_hash.remove (a_definition.logical.as_lower)
				else
					-- Otherwise, the overridden one becomes the new main one
					definition_hash.force (l_overridden, l_overridden.logical)
				end
			end
--			production_databases_by_path.search(a_definition.full_path.as_lower)
--			if production_databases_by_path.found then
--				a_definition.set_is_production(True)
--			end
			unlock
		end

	find_repository_from_configuration (a_config: ROSE_CONFIGURATION)
			-- Find path to repository in 'a_config', if defined and no 'repository_definition' found yet.
		local
			l_path: STRING
		do
			if a_config /= Void and then repository_definition_i = Void then
				l_path := a_config.value (k_path_repository)
				if is_significant (l_path) then
					create repository_definition_i.make_from_path (a_config.path, repository_logical, l_path)
					add_stamped (repository_definition_i, stamp)
				end
			end
		end

	find_fallback_repository
			-- Ensure `fallback_repository_definition_i' is set, if possible.
		local
			l_config: ROSE_CONFIGURATION
			l_path: STRING
		do
			if fallback_repository_definition_i = Void then
				create l_config.make_default
				Global.config_manager.load (l_config, Void, global.deliveries.release_candidate.path_to_site_cfg)
				error.flush_and_acknowledge
				l_path := l_config.value (k_path_repository)
				if is_significant (l_path) then
					create fallback_repository_definition_i.make_from_path (l_config.path, repository_logical, l_path)
					fallback_repository_definition_i.mark_as_fallback
					add_stamped (fallback_repository_definition_i, stamp)
				end
			end
		end

	add_overrides_from_config (a_config: ROSE_CONFIGURATION)
			-- Add all overrides found in 'a_config' (logicals defined under "/db").
		local
			l_node, l_override: ROSE_CONFIGURATION_NODE
			l_definition: ROSE_DATABASE_DEFINITION
			i: INTEGER
		do
			if a_config /= Void then
				l_node := a_config.node ("/db")
				if l_node /= Void and then l_node.children /= Void then
					from
						i := l_node.children.lower
					until
						i > l_node.children.upper
					loop
						l_override := l_node.children.item (i)
						if l_override /= Void and then is_significant (l_override.key) and then is_significant (l_override.value) then
							create l_definition.make_from_path (a_config.path, l_override.key, l_override.value)
							add_stamped (l_definition, stamp)
						end
						i := i + 1
					end
				end
			end
		end

	add_record (a_record: ROSE_REPOSITORY_RECORD; a_source: DATABASE_DEFINITION)
			-- Add entry for 'a_record' into 'a_node', mark as 'as_included' eventually.
		require
			source_ok: a_source /= Void
			known_source: a_source = fallback_repository_definition_i or else a_source = repository_definition_i
		local
			l_definition: ROSE_DATABASE_DEFINITION
			l_source: STRING
		do
			if a_record /= Void
			and then a_record.logical /= Void
			and then a_record.server /= Void
			and then a_record.database /= Void then
				arrange (a_record.logical)
				arrange (a_record.server)
				arrange (a_record.database)
				arrange (a_record.user_id)
				arrange (a_record.enabled)
				a_record.server.to_upper
				if is_relevant_record (a_record) then
					if is_significant (a_record.user_id) then
						l_source := "[user]" + a_source.out
					else
						l_source := a_source.out
					end
					create l_definition.make (l_source, a_record.logical, "sql", a_record.server, a_record.database)
					add_stamped (l_definition, stamp)
				end
			end
		end

	is_relevant_record (a_record: ROSE_REPOSITORY_RECORD): BOOLEAN
			-- Is 'a_record' relevant for current user?
		require
			record_ok: a_record /= Void
		do
			if is_significant (a_record.user_id) then
				if a_record.user_id ~ user_filter then
					Result := a_record.enabled ~ k_true
				end
			else
				Result := True
			end
		end

	add_stamped (a_definition: ROSE_DATABASE_DEFINITION; a_stamp: INTEGER)
			-- Add 'a_definition' to this pool.
		require
			definition_ok: a_definition /= Void
		local
			l_previous: ROSE_DATABASE_DEFINITION
		do
			lock
			l_previous := get_previous_definition (a_definition)
			if l_previous = Void then
				-- No definition from same source and for same logical was present
				l_previous := definition (a_definition.logical.as_lower)
				if a_definition ~ l_previous then
					-- This exact same definition was already present, we don't replace the previous object in this case
					l_previous.set_stamp (a_stamp)
				elseif l_previous /= Void and then is_override_ignored (a_definition, l_previous) then
					-- There are a few special cases of overrides that we ignore (to avoid generating useless warnings)
					l_previous.set_stamp (a_stamp)
				else
					-- This is the case where 'a_definition' overrides a previous entry
					a_definition.set_overridden (l_previous)
					definition_hash.force (a_definition, a_definition.logical.as_lower)
					a_definition.set_stamp (a_stamp)
				end
			elseif l_previous ~ a_definition then
				-- We already had the exact same definition, let's just stamp it as visited (and keep the old definition object)
				l_previous.set_stamp (a_stamp)
			elseif l_previous.overridden_by /= Void then
				-- 'a_definition' corresponds to an existing overridden entry which it modifies
				-- example: a repository definition was modified, but there is a user override for it anyway
				a_definition.set_overridden (l_previous.overridden)
				l_previous.overridden_by.set_overridden (a_definition)
			else
				-- A current definition (in effect, not overridden) has changed
				a_definition.set_overridden (l_previous.overridden)
				definition_hash.force (a_definition, a_definition.logical.as_lower)
			end
			unlock
		end

	is_override_ignored (a_new, a_previous: ROSE_DATABASE_DEFINITION): BOOLEAN
			-- Should we ignore 'a_new' override in order to avoid useless warnings?
			-- This is the case with the repository defining a logical for itself...
		require
			new_ok: a_new /= Void
			previous_ok: a_previous /= Void
		do
			if a_new.is_same_path (a_previous) then
				if a_new.logical.as_lower ~ repository_logical then
					-- Ignore repositories defining a logical "repository" pointing back to themselves
					Result := a_new.source ~ a_new.out
				end
			end
		end

	get_previous_definition (a_new: ROSE_DATABASE_DEFINITION): ROSE_DATABASE_DEFINITION
			-- Existing definition that matches 'a_new' (same source and logical), if any
		require
			valid_new_definition: a_new /= Void
		local
			l_previous: ROSE_DATABASE_DEFINITION
		do
			from
				l_previous := definition (a_new.logical.as_lower)
			until
				Result /= Void or else l_previous = Void
			loop
				if l_previous.is_same_source_and_logical (a_new) then
					Result := l_previous
				end
				l_previous := l_previous.overridden
			end
		end

	arrange (a_string: STRING)
			-- Arrange 'a_string' by suppressing extra spaces and putting it to lowercase.
		do
			if a_string /= Void then
				a_string.to_lower
				a_string.right_adjust
				a_string.left_adjust
			end
		end

	compare_logicals (d1, d2: ROSE_DATABASE_DEFINITION): BOOLEAN
			-- Is logicals 'd1' < 'd2'?
		require
			definitions_ok: d1 /= Void and then d2 /= Void
		do
			Result := d1.logical < d2.logical
		end

	is_significant (a_string: STRING): BOOLEAN
			-- Is 'a_string' significant (not empty)?
		do
			Result := a_string /= Void and then not a_string.is_empty
		end

	append_integer (a_string: STRING; a_integer: INTEGER; a_label: STRING)
			-- Append 'a_integer' to 'a_string' followed with 'a_label', paying attention to plural
		require
			string_ok: a_string /= Void
		do
			a_string.append_integer (a_integer)
			a_string.append_character (' ')
			a_string.append_string (a_label)
			if a_integer /= 1 then
				a_string.append_character ('s')
			end
		end

	record_metrics_for_server (a_definition: ROSE_DATABASE_DEFINITION) is
			-- Record process-metrics for `a_definition'.
		require
			a_definition_not_void: a_definition /= Void
		do
			if a_definition.bulk_inserts /= Void then
				a_definition.bulk_inserts.do_all_with_key (agent record_metric (?, ?, a_definition.server, a_definition.logical, "bulk_inserts_for_"))
			end
			if a_definition.truncates /= Void then
				a_definition.truncates.do_all_with_key (agent record_metric (?, ?, a_definition.server, a_definition.logical, "truncates_for_"))
			end
			if a_definition.creates /= Void then
				a_definition.creates.do_all_with_key (agent record_metric (?, ?, a_definition.server, a_definition.logical, "creates_for_"))
			end
			if a_definition.drops /= Void then
				a_definition.drops.do_all_with_key (agent record_metric (?, ?, a_definition.server, a_definition.logical, "drops_for_"))
			end
		end

	record_metric (a_count: INTEGER; a_table_name, a_node, a_database, a_tag_prefix: STRING) is
			-- Record metric `a_tag_prefix' + `a_table_name' has having occured `a_count' times.
		require
			a_table_name_not_void: a_table_name /= Void
			a_table_name_not_empty: not a_table_name.is_empty
			a_tag_prefix_not_void: a_tag_prefix /= Void
			a_tag_prefix_not_empty: not a_tag_prefix.is_empty
			a_node_not_void: a_node /= Void
			a_node_not_empty: not a_node.is_empty
			a_database_not_empty: not a_database.is_empty
		do
			write_metric (a_count, a_tag_prefix + a_node + "." + a_database + "." + a_table_name)
		end

	write_metric (a_count: INTEGER; a_name: STRING) is
			-- Record metric `a_name' has having occured `a_count' times.
		require
			a_name_not_void: a_name /= Void
			a_name_not_empty: not a_name.is_empty
		local
			l_types: PM_METRIC_TYPE_ROUTINES
			l_type: PM_METRIC_TYPE
			l_metric: PM_METRIC
		do
			create l_types
			l_type := l_types.type (a_name)
				-- WARNING: this is a command - it will do I/O
			create l_metric.make (l_type)
			l_metric.set_value (a_count.to_double)
			l_metric.write_to_database (Global.log_id)
		end

end
