indexing
	description: "Tiny expression language for defining dates"
	author: "Franck Arnaud"
	date: "$Date: 2010/05/11 $"
	revision: "$Revision: #6 $"
	path: "$File: //rose/source/kernel/infrastructure/other/rose_date_dot_expression.e $"

class 
	ROSE_DATE_DOT_EXPRESSION

inherit
	ROSE_DOT_EXPRESSION [ROSE_DATE]
		rename
			make as make_expression
		end
	ROSE_FRAMEWORK_CONVENTION
		export 
			{NONE} all 
		end

create
	make

feature {NONE} -- Creation

	make is
			-- Expression processor for dates.
		do
			make_expression (agent is_expression_date, agent expression_date)
			add_leaf_operator (agent today, today_keyword)
			add_leaf_operator (agent now, now_keyword)
			add_leaf_operator (agent last_weekday (dummy_date.Monday), mon_keyword)
			add_leaf_operator (agent last_weekday (dummy_date.Tuesday), tue_keyword)
			add_leaf_operator (agent last_weekday (dummy_date.Wednesday), wed_keyword)
			add_leaf_operator (agent last_weekday (dummy_date.Thursday), thu_keyword)
			add_leaf_operator (agent last_weekday (dummy_date.Friday), fri_keyword)
			add_leaf_operator (agent last_weekday (dummy_date.Saturday), sat_keyword)
			add_leaf_operator (agent last_weekday (dummy_date.Sunday), sun_keyword)
			add_unary_operator (agent {ROSE_DATE}.previous_month_end, previous_month_end_keyword)
			add_unary_operator (agent month_end, month_end_keyword)
			add_unary_operator (agent week_back, week_back_keyword)
			add_unary_operator (agent last_open, last_open_keyword)
			add_unary_operator (agent {ROSE_DATE}.previous_weekday, previous_weekday_keyword)
			-- these are not unary operators, but ROSE_DOT_EXPRESSION has 
			-- been butchered the wrong way to allow this. (fa)
			add_unary_operator (agent plus_n_month(?,?), plus_n_months_keyword)
			add_unary_operator (agent plus_n_date(?,?), plus_n_date_keyword)
		end

feature {NONE} -- Date expression implementation

	is_expression_date (a_string: STRING): BOOLEAN is
			-- Is YYYYMMDD, YYYYMM, JLN?
		local
			l_date: INTEGER
			l_date_string : STRING
		do
			if a_string /= Void then
				l_date_string := a_string.twin
				l_date_string.prune_all(c_double_quote)
				if l_date_string.is_integer then
					l_date := l_date_string.to_integer
					if l_date > 0 and l_date < 1000 then
						-- JLN
						Result := True
					elseif l_date > 190001 and l_date < 999912 then
						-- YYYYMM
						Result := True
					else
						-- YYYYMMDD
						Result := l_date > 19000101 and l_date <= 99991231
					end
				end
			end
		end

	expression_date (a_string: STRING): ROSE_DATE is
			-- YYYYMMDD, YYYYMM, JLN
		require
			a_string_valid: is_expression_date (a_string)
		local
			l_date_string : STRING
			l_date: INTEGER
		do
			if a_string /= Void then
				l_date_string := a_string.twin
				l_date_string.prune_all(c_double_quote)
				if l_date_string.is_integer then
					l_date := l_date_string.to_integer
					if l_date > 0 and l_date < 1000 then
						-- JLN
						create Result.make_from_julian_month (l_date)
					elseif l_date >= 190001 and l_date <= 999912 then
						-- YYYYMM
						create Result.make (l_date // 100, l_date \\ 100, 1)
					else
						-- YYYYMMDD
						create Result.make (l_date // 10000, (l_date // 100) \\ 100, l_date \\ 100)
					end
				end
			end
		end

	today: ROSE_DATE is
			-- 'today'
		do
			create Result.make_today
		end

	now: ROSE_DATE is
			-- 'now'
		do
			create Result.make_now
		end
	
	plus_n_month(a_day : ROSE_DATE; a_add : STRING) : ROSE_DATE is
			-- Substract `a_add' (unparsed integer parameter) days from `a_day'
		require
			a_add_is_integer: a_add.is_integer
			a_day_not_void: a_day /= Void
		do
			Result := a_day.plus_n_month(a_add.to_integer)
		end
	
	plus_n_date(a_day : ROSE_DATE; a_add : STRING) : ROSE_DATE is
			-- Remove `a_add' (unparsed integer parameter) days from `a_day'
		require
			a_add_is_integer: a_add.is_integer
			a_day_not_void: a_day /= Void
		do
			Result := a_day.plus_n_date(a_add.to_integer)
		end
	
	last_weekday (a_day: INTEGER): ROSE_DATE is
			-- Latest weekday matching 'a_day' (today or before)
		require
			is_weekday: a_day > 0 and a_day <= 7
		do
			from
				create Result.make_today
			until
				Result.day_of_the_week = a_day
			loop
				Result.add_n_days (-1)
			end
		ensure
			result_not_void: Result /= Void
			day: Result.day_of_the_week = a_day
		end

	week_back (a_date: ROSE_DATE): ROSE_DATE is
			-- Substract 1 week from date.
		require
			a_date_not_void: a_date /= Void
		do
			Result := a_date.twin
			Result.add_n_days (-7)
		end

	last_open (a_date: ROSE_DATE): ROSE_DATE is
			-- Substract 1 week from date.
		require
			a_date_not_void: a_date /= Void
		do
			Result := a_date.twin
			Result.add_n_days (-1)
			if Result.is_weekend then
				Result.add_n_days (-1)
			end
			if Result.is_weekend then
				Result.add_n_days (-1)
			end
		end

	month_end (a_date: ROSE_DATE): ROSE_DATE is
			-- End of month for current month.
		require
			a_date_not_void: a_date /= Void
		do
			Result := a_date.twin
			Result.add_n_months (1)
			Result := Result.previous_month_end
		end

feature {NONE} -- Constants

	mon_keyword: STRING is "mon"
	tue_keyword: STRING is "tue"
	wed_keyword: STRING is "wed"
	thu_keyword: STRING is "thu"
	fri_keyword: STRING is "fri"
	sat_keyword: STRING is "sat"
	sun_keyword: STRING is "sun"
			-- Weekday names

	month_end_keyword: STRING is "month_end"
	previous_month_end_keyword: STRING is "previous_month_end"
	previous_weekday_keyword: STRING is "previous_weekday"
	week_back_keyword: STRING is "week_back"
	last_open_keyword: STRING is "last_open"
	plus_n_months_keyword: STRING is "plus_n_month"
	plus_n_date_keyword: STRING is "plus_n_date"
			-- Keywords

	dummy_date: ROSE_DATE is
			-- 	Date for constants
		once
			create Result.make (1900, 1, 1)
		end

end
