indexing
	description: "Bridge between DATA_SERVER and configuration"
	author: "Zoran Simic"
	date: "$Date: 2008/12/01 $"
	revision: "$Revision: #7 $"

class
	ROSE_DB_ACCESSOR

inherit
	ROSE_SHARED_DATABASE_POOL
	DEBUG_OUTPUT

create
	make_accessor

feature {NONE} -- Initialization

	make_accessor (a_id: STRING; a_initializer: like initializer) is
			-- Create an accessor bound to 'a_id' and 'a_initializer' agent on connection.
		require
			valid_id: a_id /= Void
		do
			id := a_id
			initializer := a_initializer
		end

feature -- Access

	id: STRING
			-- Accessor id (logical DB name in config)

	server: DATA_SERVER
			-- Server bound to this manager, available when connected

	initializer: PROCEDURE [ANY,TUPLE [ROSE_DB_ACCESSOR]]
			-- Initializer agent to call on connection

	new_execution_predicate (a_query: STRING): EXECUTION_PREDICATE is
			-- Execution predicate with 'a_query' on 'server',
			-- an error is eventually signaled
			-- FIXME: remove this when DATA_SERVER handles errors.
		require
			is_connected: is_connected
			valid_query: a_query /= Void
		do
			if not error.signaled then
				create Result.make (server,a_query)
				check_predicate (Result,"Could not prepare '$1'",<<a_query>>)
			end
			if error.signaled then
				Result := Void
			end
		ensure
			error_if_void: Result = Void implies error.signaled
		rescue
			if not error.signaled then
				error.signal (id,"Could not prepare '$1'%NException occured: $2",<<a_query,last_exception_explanation>>)
				retry
			end
		end

	last_identity: INTEGER is
			-- Last identity value inserted, returns SQL 'SELECT @@identity' on current 'server'
		require
			is_connected: is_connected
		local
			l_array: ARRAY [INTEGER]
		do
			if sql_last_identity = Void and then is_connected then
				create sql_last_identity.make (server,"SELECT @@identity")
					--| FIXME: ISE Eiffel thinks that INTEGER_PREDICATE does not
					--| conform to ROSE_PREDICATE [ANY]...
				--| check_predicate (sql_last_identity,"Could not prepare last identity look-up request",Void)
				if error.signaled then
					sql_last_identity := Void
				end
			end
			if sql_last_identity /= Void then
				sql_last_identity.retrieve (Void,Void)
				l_array := sql_last_identity.array
				if l_array /= Void and then l_array.count > 0 then
					Result := l_array.item (l_array.lower)
				end
			end
		end

	converter: ROSE_DB_CONVERSION_SERVICE is
			-- Handy conversion routines
		once
			create Result
		end

feature -- Status report

	is_connected: BOOLEAN is
			-- Is this DB server currently connected?
		do
			Result := pool.is_connected (server)
		end

feature -- Basic operations

	connect is
			-- Start this server with info form config on 'id'.
		do
			server := pool.get_server_from_logical (id)
			if is_connected then
				if initializer /= Void then
					initializer.call ([Current])
				end
			elseif not error.signaled then
				if server = Void then
					error.signal (generator,"Can not find DB '$1'",<<id>>)
				else
					error.signal (generator,"Can not connect to server '$1'",<<id>>)
				end
				server := Void
				sql_last_identity := Void
			end
		ensure
			error_signaled: not is_connected implies error.signaled
		end

	disconnect is
			-- Close this server
		do
			if server /= Void then
				server := Void
				sql_last_identity := Void
			end
		ensure
			not_connected: not is_connected
		end

	commit is
			-- Commit performed transactions.
		do
			if is_connected then
				server.commit
			end
		end

	rollback is
			-- Rollback performed transactions.
		do
			if is_connected then
				server.rollback
			end
		end

	execute_predicate (a_predicate: EXECUTION_PREDICATE; a_args: ARRAY [ANY]; a_error_msg: STRING) is
			-- Execute 'a_predicate' with 'a_args',
			-- an error is eventually signaled with 'a_error_msg' and 'a_args'.
			-- FIXME: remove this when EXECUTION_PREDICATE handles errors.
		do
			a_predicate.execute (Void,a_args)
			check_predicate (a_predicate,a_error_msg,a_args)
		end

	check_predicate (a_predicate: ROSE_PREDICATE [ANY]; a_error_msg: STRING; a_args: ARRAY [ANY]) is
			-- Check if 'a_predicate' has been correctly prepared
			-- an error is eventually signaled with 'a_error_msg' and 'a_args'.
			-- FIXME: remove this when ROSE_PREDICATE handles errors.
		do
			if a_predicate = Void or else not a_predicate.success then
				error.signal (id,a_error_msg,a_args)
			end
		end

feature -- Output

	debug_output, representation: STRING is
			-- Representation of this server
			-- String that should be displayed in debugger to represent `Current'.
		do
			create Result.make (20)
			Result.append_string (id)
			if server /= Void then
				Result.append (" [")
				Result.append_string (server.node)
				Result.append_character ('.')
				Result.append_string (server.database)
				Result.append_character (']')
			else
				Result.append (" [not connected]")
			end
		end

feature {NONE} -- Implementation

	sql_last_identity: INTEGER_PREDICATE

end
