indexing
	description: "Network Client class that uses the port_server to create a server"
	author: "Mauricio Bustos"
	last_editor: "$Author: rj $"
	date: "$Date: 2015/03/25 $"
	revision: "$Revision: #72 $"
	path: "$File: //rose/source/kernel/infrastructure/network/rose_dedicated_network_client.e $"
	keywords: "network,client,server"

deferred class
	ROSE_DEDICATED_NETWORK_CLIENT

inherit
	NETWORK_CLIENT
		rename
			make as client_make,
			error as client_error
		redefine
			send,
			receive,
			client_make,
			in_out,
			cleanup
		end
	EXCEPTIONS
		rename
			class_name as exception_class_name
		export
			{NONE} all
		end
	ROSE_SHARED_GLOBAL
		rename
			trace as global_trace
		end
	-- If you include MEMORY you will have to change some child classes.

feature -- Creation

	get_server is
			-- Talk to the port server and get the persistent server's port.  Connect to the server.
		require
			server_name_exists: server_name /= Void 		-- server_name must be set by heir
		local
			l_port: STRING
		do
			l_port := env.get (server_name)

			if l_port /= Void and then l_port.is_integer then
				get_server_by_port_number(l_port.to_integer)
			else
				get_server_by_port_number(0)
			end
		end

	get_server_by_port_number(a_port_number: INTEGER)
			-- Talk to the port server and get the persistent server's port.  Connect to the server.
			-- If 'a_port_number' is valid (greater than 0), bypass the port server and connect to the persistent server directly.
		require
			server_name_exists: server_name /= Void			-- server_name must be set by heir
		do
			servers := servers_by_name
			servers.start
			set_process_identifiers

			if a_port_number > 0
			then
				get_supported_security_systems_list(servers.first, a_port_number, port_server_message)
				get_user_server(a_port_number)
			else
				port_server_connect	-- Not just connect, it talks to and disconnects from the port_server and then connects to the server.
			end

			post_get_server			-- User defined actions go here.
		end

feature {NONE} -- Implementation

	get_user_server (a_port_number: INTEGER) is
		local
			l_collecting: BOOLEAN
		do
			l_collecting := memory.collecting
			port_number := a_port_number
			if port_server_message /= Void and then port_server_message.machine /= Void then
				if is_verbose then
					global_trace ("$1: sending port_server message for $2%N", <<generator, server_name>>)
				end
				client_make (a_port_number, servers.current_item)
			else
				if is_verbose then
					global_trace ("$1: sending port_server message for $2%N", <<generator, server_name>>)
				end
				client_make (a_port_number, servers.first)
			end
			if not in_out.is_open_read then
				if is_verbose then
					global_trace ("$1: $2 for attempt on $3%N", <<generator,in_out.error, server_name>>)
				end
				raise ("")
			elseif not is_monitor or
				   config.boolean("debug/show_monitor_connections", False)
			then
				global_trace ("$1: using $2 on user supplied port $3 on $4%N",
							  <<generator,server_name,a_port_number,port_server_message.machine>>)
			end
		rescue
			if l_collecting then
				memory.collection_on
			else
				memory.collection_off
			end
			if not servers.after and servers.count > 1 then
				servers.forth
				global_trace ("$1: No user defined $2 on $3%N",
							  <<generator,server_name,port_server_message.machine>>)
				retry
			else
				raise_fatal_exception (1, "No user defined servers available")
			end
		end

	set_process_identifiers is
			-- set attributes of `port_server_message'
		local
			l_id, l_command: STRING
			l_pid: INTEGER
		do
			l_id := env.get ("USER")
			if l_id = Void then
				l_id := env.get ("USERNAME")
				if l_id = Void then
					l_id := env.get ("MACHINE")
					if l_id = Void then
						l_id := env.get ("COMPUTERNAME")
					end
				end
			end
			if l_id /= Void then
				if not port_server_message.get_user_id.has_substring (l_id) then
					port_server_message.set_user_id (port_server_message.get_user_id + l_id)
				end
			else
				port_server_message.set_user_id (port_server_message.get_user_id + "???")
			end
			l_command := (create {ARGUMENTS}).argument (0)
			l_pid := env.current_process_id
			if l_command.count > 0 then
				port_server_message.set_command_line (l_command + " (" + l_pid.out + ")")
			end
			port_server_message.set_machine (env.host_name)
		end

feature -- Operations

	send_message (message: ANY) is
		local
			l_string: STRING
			l_collecting: BOOLEAN
			-- Use ROSE_DEDICATED_NETWORK_SERVER for the server
		do
			l_collecting := memory.collecting
			send (message)
			receive
			l_string ?= received
			if l_string /= Void and l_string.count > 0 then
				if l_string.substring_index ("SERVER_EXCEPTION",1) > 0 then
					warn ("send_message",server_name + " - " + l_string,Void)
					warn ("send_message","%NEND_OF_SERVER_EXCEPTION%NEND_OF_SERVER_EXCEPTION%NEND_OF_SERVER_EXCEPTION%NEND_OF_SERVER_EXCEPTION%N",Void)
					server_failure_count := max_server_failures
					raise ("SERVER_EXCEPTION")
				elseif l_string.substring_index ("SERVER_MESSAGE",1) > 0 then
					warn ("send_message","%N%NSTART_OF_SERVER_MESSAGE%N%N" + server_name + " - " + l_string + "%N%NEND_OF_SERVER_MESSAGE%N%N",Void)
				end
			end
		rescue
			if l_collecting	then
				memory.collection_on
			else
				memory.collection_off
			end
			if server_failure_count < max_server_failures then --or env.get (server_name) /= Void then
				warn ("send_message","$1: $2 Sending Failure, attempting connection number $3%N$4%N$5%N",
					  <<generator,server_name,server_failure_count,exception_trace,message>>)
				server_failure_count := server_failure_count + 1
				get_server
				retry
			end
		end

	send (a_msg: ANY) is
			-- Send objectt.
		do
			if is_verbose then
				global_trace ("$1: Sending to $2%N",<<generator,server_name>>)
			end
			if byte_count_enabled then
				sent_byte_count := sent_byte_count + Global.persist.size_of (a_msg)
				if sent_byte_count > reported_sent_byte_count and sent_byte_count > byte_count_threshold then
					reported_sent_byte_count := sent_byte_count + byte_count_threshold
					global_trace ("$1 ($2) has sent     $3 bytes%N",
								  <<server_name,argument_string (port_server_message),Global.formatter.formatted_integer (sent_byte_count,20,',')>>)
				end
			end
			in_out.independent_store (a_msg)
		end

	receive is
			-- Receive message blocking
		do
			if is_verbose then
				global_trace ("$1: Receiving from $2%N",<<generator,server_name>>)
			end
			received := in_out.retrieved
			if byte_count_enabled then
				received_byte_count := received_byte_count + Global.persist.size_of (received)
				if received_byte_count > reported_received_byte_count and received_byte_count > byte_count_threshold then
					reported_received_byte_count := received_byte_count + byte_count_threshold
					global_trace ("$1 ($2) has received $3 bytes%N",
								  <<server_name,argument_string (port_server_message),Global.formatter.formatted_integer (received_byte_count,20,',')>>)
				end
			end
		end

	receive_with_timeout is
		local
			l_poller: MEDIUM_POLLER
			l_result: INTEGER
			l_command: LISTEN_POLL_COMMAND
		do
			if is_verbose then
				global_trace ("$1: Receiving from $2 with timeout%N",<<generator,server_name>>)
			end

			-- Time out if you don't get your first request within the limit

			if timeout=0
			then
				timeout := 120
			end

			create l_poller.make_read_only
			create l_command.make (in_out)
			l_poller.put_read_command (l_command)
			l_result := l_poller.medium_select (in_out.descriptor + 1, timeout * 1000)
			l_poller.remove_read_command (l_command)
			if l_result > 0 then
				received := in_out.retrieved
			else
				error.exceptions.raise ("TIMEOUT")
			end
		rescue
			if l_command /= Void then
				l_poller.remove_read_command (l_command)
			end
		end

	cleanup is
			-- Shut down client.
		do
			if in_out /= Void and then in_out.exists then
				Precursor
			end
		end

	get_supported_security_systems_list(a_machine_name: STRING; a_port_number: INTEGER; a_port_server_message: PORT_SERVER_MESSAGE)
	-- This is an oppurtunity for a descendant to determine the security requirements of the server we are about to contact.
	-- This implementation does not support security.
	do
	end

	is_monitor: BOOLEAN
	-- Is Current a PERSISTENT_CLUSTERED_TAGGED_SERVER_MONITOR?  We could just test but that would introduce a cluster dependency.
	do
		Result := False
	end

feature -- Client Server Admin

	send_message_to_server (a_message: STRING) is
		do
			if a_message /= Void then
				send_message (a_message)
			end
		end

	post_get_server is
			-- Used by the descendant to allow custom behavior after connecting OR RECONNECTING to the application server.
			-- This is called at the end of 'get_server'.  Note that 'get_server' can be called if the
			-- connection to the server is broken and a new connection is made (see the rescue clause in
			-- 'send_message').  You can provide your own implementation.  This was originally made to
			-- support calls to 'set_resource_override'.
			--
			-- Sample implementation (from FUNDAMENTAL_VARIABLE_CLIENT):
			--		create l_env
			--		from zones.start until zones.off loop
			--			set_resource_override("MONTHLY_DATA_" + zones.current_item + "_C_FG", l_env.get("MONTHLY_DATA_" + zones.current_item + "_FG"), "_C")
			--			set_resource_override("MONTHLY_DATA_" + zones.current_item + "_S_FG", l_env.get("MONTHLY_DATA_" + zones.current_item + "_FG"), "_S")
			--			set_resource_override("MONTHLY_DATA_" + zones.current_item + "_C",    l_env.get("MONTHLY_DATA_" + zones.current_item),         "_C")
			--			set_resource_override("MONTHLY_DATA_" + zones.current_item + "_S",    l_env.get("MONTHLY_DATA_" + zones.current_item),         "_S")
			--			set_resource_override("MONDEX_"       + zones.current_item,           l_env.get("MONDEX_"       + zones.current_item),         Void)
			--			zones.forth
			--		end
			--
			--		set_resource_override("VARIABLE_STATISTICS",    l_env.get("VARIABLE_STATISTICS"),    Void)
			--		set_resource_override("VARIABLE_STATISTICS_FG", l_env.get("VARIABLE_STATISTICS_FG"), Void)
		do
		end

	set_resource_override (a_logical, a_resource, a_suffix: STRING) is
			-- See comment in 'post_get_server'.
		require
			valid_override: a_logical /= Void
		local
			l_resource: STRING
		do
			if a_resource /= Void then
				if a_suffix /= Void then
					l_resource := a_resource + a_suffix
				else
					l_resource := a_resource
				end
				send_message (<<a_logical,l_resource>>)
				global_trace ("$1 - RESOURCE_OVERRIDE: $2 being used for $3%N",
							  <<server_name,l_resource,a_logical>>)
			end
		end

	set_timeout(a_timeout: INTEGER)
	-- New socket recv 'timeout' in seconds.
	do
		timeout := a_timeout
	end

feature -- Constants

	max_connection_attempts: INTEGER is 5
	max_server_failures: INTEGER is 2
	default_port_server_port: INTEGER is 2000
	storable_storable: INTEGER is 0
	storable_sed: INTEGER is 1

feature -- Connectivity

	port_number: INTEGER			-- Ths port number of the server we are to connect to.

	port_server_test: BOOLEAN

	port_server_message: PORT_SERVER_MESSAGE

	server_failure_count: INTEGER

	server_name: STRING

	storable_type: INTEGER

	sent_byte_count, received_byte_count: INTEGER

	timeout: INTEGER						-- How long (in seconds) to wait on a socket recv.

	reported_sent_byte_count, reported_received_byte_count: INTEGER

	port_server_port: INTEGER
			-- The port number that we expect the port_server to be listening on.
		local
			l_port: STRING
			l_config_port: STRING
		once
			l_config_port := Global.app.delivery.value ("/port_server/port")
			l_port := env.get ("TEST_PORT_SERVER")
			if l_port /= Void and then l_port.is_integer then
				Result := l_port.to_integer
			elseif l_config_port /= Void and then l_config_port.is_integer then
				Result := l_config_port.to_integer
			else
				Result := default_port_server_port
			end
		end

	server_directory_root: STRING is
			-- The root directory for where to look for servers, send this to the port_server
		once
			Result := env.get ("NT_SERVER_DIRECTORY")
			if Result = Void then
				Result := Global.hardcoded_path ("c:\servers")
			end
		end

	servers: ROSE_LINEAR_ARRAY[STRING]
			-- List of port servers

	servers_by_name: ROSE_LINEAR_ARRAY[STRING] is
			-- Read server list from environment variable or config file,
			-- return shuffled list
		local
			l_server_list: STRING
			l_list: ROSE_DELIMITED_TEXT
			l_time: ROSE_DATE
			l_index: INTEGER
		do
			l_server_list := env.get ("NT_SERVER_LIST")
			if l_server_list = Void then
				l_server_list := config.value ("NT_SERVER_LIST")
			end
			if l_server_list /= Void then
				create l_time.make_now
				create l_list.make (l_server_list,',')
				l_index := (l_time.persecs // 1_000) \\ l_list.count + 1
				Result := l_list
				if Result.valid_index (l_index - 1) and Result.valid_index (l_index) then
					global_trace ("$1: Changing order of NT_SERVER_LIST%N",<<generator>>)
					Result.append_array (Result.subarray (1,l_index - 1))
					Result := Result.subarray (l_index,Result.count)
				end
			end
		end

	in_out: ROSE_NETWORK_STREAM_SOCKET		-- The socket on which we send/receive messages with a server.

	use_storable is
			-- Call this to force storables on this socket to be the old storable type.
		do
			if in_out /= Void then
				in_out.use_storable
			end
			storable_type := storable_storable
		end

	use_sed is
			-- Call this to force storables on this socket to be the new SED type.
		do
			if in_out /= Void then
				in_out.use_sed
			end
			storable_type := storable_sed
		end

feature	{NONE} -- Implementation

	argument_string (a_port_server_message: PORT_SERVER_MESSAGE): STRING is
			-- The PORT_SERVER_MESSAGE arguments in STRING form, separated by commas.
		local
			i: INTEGER
		do
			from
				i:= a_port_server_message.server_arguments.lower
			until
				i > a_port_server_message.server_arguments.upper
			loop
				if Result = Void then
					Result := ""
				else
					Result.append (",")
				end
				if port_server_message.server_arguments.item (i) /= Void then
					Result.append (port_server_message.server_arguments.item (i))
				end
				i := i + 1
			end
		end

	client_make (a_peer_port: INTEGER; a_peer_name: STRING) is
			-- Connect to `a_peer_name' at `a_peer_port' and set storable mode on the socket
		local
			l_connected: BOOLEAN
			l_future_time: ROSE_DATE
			l_tries: INTEGER
		do
			from
				create l_future_time.make_now
				l_future_time.second_add((max_connect_wait_time_ms / 1000.0).truncated_to_integer)
				-- 'l_future_time' is now 'max_connect_wait_time_ms' (converted to seconds) in the future.
			until
				l_connected
			loop
				Precursor (a_peer_port, a_peer_name)		-- Calls 'connect' on the socket.  Sets 'is_open_read' on success.
															-- On Windows (8/2009) this appears to timeout in 1 second.
				l_connected := in_out.is_open_read
				if not l_connected
				then
					l_tries := l_tries + 1
					if (create {ROSE_DATE}.make_now)>l_future_time and l_tries>=rdnc_max_tries
					then
						-- It took too long (and at least 'rdnc_max_tries' tries).
						raise("Timed-out waiting to connect to socket " + a_peer_port.out)
					end

					if is_verbose
					then
						global_trace(global.interpreted("($now) Connect to socket $1 failed.  Will retry.%N"), <<a_peer_port.out>>)
					end

					env.sleep(connect_wait_time_ms)		-- We are waiting for the EXE to create the socket and call 'accept'.
				end
			end

			in_out.set_default_socket_options
			inspect storable_type
			when storable_sed then
				in_out.use_sed
			else
				in_out.use_storable				-- This is the default
			end
		end

	port_server_connect is
			-- connect to the port_server on the host identified as `servers.current_item',
			-- send the port_server_message and receive the response, process the response,
			-- and connect to the real server
		local
			l_collecting: BOOLEAN
			l_target_hostname: STRING
		do
			l_collecting := memory.collecting

			l_target_hostname := servers.current_item

			if l_target_hostname = Void then
				l_target_hostname := "<Void>"
			end
			client_make (port_server_port, servers.current_item)		-- Connect to the port_server.
			if is_verbose then
				global_trace ("$1: Sending to port_server $2:$3%N",
							  <<generator,l_target_hostname,port_server_port>>)
			end

			in_out.independent_store (port_server_message)			-- Send it our request.
			if is_verbose then
				global_trace ("$1: Receiving from port_server%N",<<generator>>)
			end
			received := in_out.retrieved							-- This is the response from the port_server.
			process_received
		rescue
			if l_collecting	then
				memory.collection_on
			else
				memory.collection_off
			end
			if is_verbose then
				global_trace ("$1: port_server_connect: Error connecting to port_server on $2:$3%N$4%N",
							  <<generator,l_target_hostname,port_server_port,Exception_trace>>)
				if in_out /= Void then
					global_trace ("Error: $1: $2%N",
								  <<in_out.error_number,in_out.error>>)
				end
			end
			global_trace ("$1: No port_server on $2 for $3%N",
						  <<generator,l_target_hostname,server_name>>)
			if not servers.after then
				servers.forth
			end
			if not servers.after then
				retry
				--			else		-- I removed this because it hides the original exception and full traceback.
				--				raise_fatal_exception (1, "Exhausted port_servers list")
			end
		end

	process_received is
			-- receive and process port server message, connect to nascent server
		require
			current_servers_exists: not servers.off
		local
			l_attempts: INTEGER
			l_collecting: BOOLEAN
		do
			l_collecting := memory.collecting
			port_server_message ?= received								-- This is the response from the port_server.
			if port_server_message /= Void then
				port_number := port_server_message.get_port
				if is_verbose then
					global_trace ("$1: Received port_server response, port: $2%N",
								  <<generator,port_number>>)
				end
				if port_server_message.connect_to_server then
					if port_number /= 0 and then l_attempts < max_connection_attempts then
						in_out.close									-- We are closing the connection to the port_server.

						get_supported_security_systems_list(servers.current_item, port_number, port_server_message)

						client_make (port_number, servers.current_item)	-- Connect to the application server (reusing 'in_out').
						if not in_out.is_open_read then
							if is_verbose then
								global_trace ("$1: $2 for attempt on $3%N",
											  <<generator, in_out.error, server_name>>)
							end
							raise ("")
						else
							if port_server_message.server_arguments /= Void and then not port_server_message.server_arguments.is_empty then
								global_trace ("$1: using $2 ($3) on $4:$5%N",
											  <<generator,server_name,argument_string (port_server_message), servers.current_item, port_number>>)
							else
								global_trace ("$1: using $2 on $3:$4%N",
											  <<generator,server_name, servers.current_item, port_number>>)
							end
						end
					else
						global_trace ("$1: $2 not available on $3...failing over%N",
									  <<generator,server_name, servers.current_item>>)
						servers.forth
						if not servers.after then
							port_server_connect
						else
							raise_fatal_exception (1, "Exhausted port_servers list")
						end
					end
				end
			end
		rescue
			if l_collecting	then
				memory.collection_on
			else
				memory.collection_off
			end
			if l_attempts <= 5 and not servers.after then
				l_attempts := l_attempts + 1
				retry
			end
		end

	raise_fatal_exception (a_code: INTEGER; a_tag: STRING ) is
			-- raise exception, then die (only if a_code is nonzero).
			-- Descendants may override this feature to rescue/retry instead of die.
		do
			warn ("raise_fatal_exception","$1: $2%N",<<generator,a_tag>>)
			raise (a_tag)
			die (a_code)
		end

	rdnc_max_tries: INTEGER
	-- Maximum number of tries to connect to a server.
	once
		Result := global.app.config.integer("rdnc_max_tries", 5).max(1)
	ensure
		sensible_result: Result>=1
	end

	connect_wait_time_ms: INTEGER is
			-- Amount time to wait before retying to connect to a socket (in milliseconds).
		local
			l_string: STRING
		once
			l_string := env.get("CONNECT_WAIT_TIME_MS")
			if l_string/=Void and then l_string.is_integer
			then
				Result := l_string.to_integer.max(0)
			else
				Result := global.app.config.integer("connect_wait_time_ms", 100).max(0)
			end
		ensure
			sensible_result: Result>=0
		end

	max_connect_wait_time_ms: INTEGER is
			-- Maximum time to wait to connect to a socket (in milliseconds).
		local
			l_string: STRING
		once
			l_string := env.get("MAX_CONNECT_WAIT_TIME_MS")
			if l_string /= Void and then l_string.is_integer then
				Result := l_string.to_integer.max(0)
			else
				Result := global.app.config.integer("max_connect_wait_time_ms", 5000).max(0)
			end
		ensure
			sensible_result: Result>=0
		end

	is_verbose: BOOLEAN is
			-- Useful for debugging.  It causes informational messages to be printed.
		local
			l_verbose: STRING
		once
			l_verbose := env.get ("VERBOSE_EIFFEL")
			Result := l_verbose /= Void or config.boolean ("VERBOSE_EIFFEL",False)
		end

	byte_count_enabled: BOOLEAN is
		once
			Result := config.boolean ("rdnc_byte_count_enabled",False)
		end

	byte_count_threshold: INTEGER is 1_000_000

	env: ROSE_EXECUTION_ENVIRONMENT is
			-- Once access to an execution environment object.
		once
			create Result
		ensure
			result_not_void: Result /= Void
		end

	memory: MEMORY is
			-- Once access to a memory object.
		once
			create Result
		ensure
			result_not_viod: Result /= Void
		end

end
