indexing
	description: "Deliveries-related conventions"
	notice: "This class defines convention used for all supported deliveries, please always use this class when refering to files in a delivery"
	author: "Zoran Simic"
	last_editor: "$Author: an $"
	date: "$Date: 2014/07/21 $"
	revision: "$Revision: #2 $"
	path: "$File: //rose/source/kernel/infrastructure/framework/application/rose_deliveries_convention.e $"

class
	ROSE_DELIVERIES_CONVENTION

inherit

	ROSE_FRAMEWORK_CONVENTION
		export {NONE} all end

	ROSE_SHARED_GLOBAL

feature -- Access

	site_configuration_file_name: STRING = "site.cfg"				-- Name for 'site.cfg' configuration file
	delivery_configuration_file_name: STRING = "delivery.cfg"		-- Name for 'delivery.cfg' configuration file

	path_to_delivery_configuration (a_path: detachable STRING): detachable STRING
			-- Returns path to delivery.cfg, searched in 'a_path' and all of its parents
			-- Void if not found
		do
			if a_path /= Void and then not a_path.is_empty then
				Result := checked_path (a_path, delivery_configuration_file_name)
				if Result = Void then
					Result := path_to_delivery_configuration (formatter.parent_folder (a_path))
				end
			end
		end

	path_to_site_configuration (a_path: detachable STRING): detachable STRING
			-- Returns path to site.cfg, searched in 'a_path' and all of its parents
			-- Void if not found
		do
			if a_path /= Void and then not a_path.is_empty then
				Result := checked_path (a_path, site_configuration_file_name)
				if Result = Void then
					Result := path_to_site_configuration (formatter.parent_folder (a_path))
				end
			end
		end

	default_delivery: attached ROSE_DELIVERY_CONVENTION
			-- Default delivery
		do
			Result := release_candidate
		end

	active: attached ROSE_DELIVERY_CONVENTION
			-- Delivery being currently used by running program
		once
			Result := delivery (global.app.site.folder)
		end

	production: attached ROSE_DELIVERY_CONVENTION
			-- Represents production delivery
		once
			Result := delivery (global.hardcoded_path ("\\axaros.com\shares\production"))
		end

	release_candidate: attached ROSE_DELIVERY_CONVENTION
			-- Represents release_candidate delivery
		once
			Result := delivery (global.hardcoded_path ("\\axaros.com\shares\release_candidate"))
		end

	rc_testenv_root: attached STRING
			-- The root directory for the test environment deliveries
		once
			Result := global.hardcoded_path ("\\axaros.com\shares\development\rc_testenv")
		end

	deployment_test: attached ROSE_DELIVERY_CONVENTION
			-- Represents deployment_test delivery
		once
			Result := rc_testenv_delivery ("deployment_test")
		end

	rc_testenv_delivery (a_name: STRING): attached ROSE_DELIVERY_CONVENTION
			-- Delivery convention for delivery with 'a_name' in 'rc_testenv_root'
		require
			relative_path: global.formatter.is_relative_path (a_name)
		do
			Result := delivery (global.attached_string (global.formatter.composed_path (<<rc_testenv_root, a_name>>)))
		ensure
			result_ok: Result.folder.starts_with (rc_testenv_root)
		end

	delivery (a_path: attached STRING): attached ROSE_DELIVERY_CONVENTION
			-- Delivery convention object for folder with 'a_path'
		require
			absolute_path: global.formatter.is_absolute_path (a_path)
		local
			l_path: attached STRING
		do
			l_path := global.attached_string (a_path).as_lower
			delivery_cache.search (l_path)
			if delivery_cache.found then
				Result := delivery_cache.found_item
			else
				create Result.make (l_path)
				delivery_cache.force (Result, l_path)
			end
		ensure
			result_ok: Result.folder ~ global.convention.clean_directory_path (a_path).as_lower
		end

	is_in_top_level_delivery (a_path: STRING): BOOLEAN
			-- Is 'a_path' part of 'production' or 'release_candidate' deliveries?
		do
			Result := production.contains_path (a_path) or else release_candidate.contains_path (a_path)
		end

	descriptive_delivery_name (a_path: STRING): attached STRING
			-- An indication as to which delivery we're dealing with
			-- "Production" when 'a_path' is within 'production' delivery, "Development" otherwise
		do
			if production.contains_path (a_path) then
				Result := "Production"
			else
				Result := "Development"
			end
		end

	known_delivery_id (a_path: STRING): STRING
			-- Id (short name) for delivery pointed to by 'a_path', if it is a known delivery
		local
			l_path: STRING
		do
			if a_path /= Void and then not a_path.is_empty then
				l_path := a_path.as_lower
				if l_path ~ production.folder then
					Result := production.short_name
				elseif l_path ~ release_candidate.folder then
					Result := release_candidate.short_name
				elseif l_path.starts_with (rc_testenv_root) then
					l_path := l_path.substring (rc_testenv_root.count + 1, l_path.count)
					if global.convention.is_directory_separator (l_path.item (l_path.count)) then
						l_path.remove (l_path.count)
					end
					if global.convention.is_directory_separator (l_path.item (1)) then
						l_path.remove (1)
					end
					if l_path.count > 2 and then not l_path.has_substring (global.convention.directory_separator) then
						Result := l_path
					end
				end
			end
		end

	check_not_in_top_level_delivery (a_id, a_path: STRING)
			-- Check that 'a_path' is not in 'production' nor 'release_candidate'
			-- Signal an error if it is
		do
			if is_in_top_level_delivery (a_path) then
				error.signal (a_id, "Please specify a location other than $1 or $2 for your test environment in test_env", <<
					production.short_name,
					release_candidate.short_name
				>>)
			end
		end

feature {NONE} -- Implementation

	delivery_cache: attached HASH_TABLE [attached ROSE_DELIVERY_CONVENTION, attached STRING]
			-- Allows to cache rc_testenv delivery definitions
		once
			create Result.make (0)
		end

end
