indexing
	description: "Delivery related conventions"
	notice: "This class defines convention used at delivery level, please always use this class when refering to files in a delivery"
	author: "Zoran Simic"
	last_editor: "$Author: zs $"
	date: "$Date: 2011/03/24 $"
	revision: "$Revision: #1 $"
	path: "$File: //rose/source/kernel/infrastructure/framework/application/rose_delivery_convention.e $"

class
	ROSE_DELIVERY_CONVENTION

inherit
	ROSE_SHARED_GLOBAL
	DEBUG_OUTPUT

create
	make

feature {NONE} -- Initialization

	make (a_folder: attached STRING)
			-- Create with 'a_folder' path
		require
			absolute_path: global.formatter.is_absolute_path (a_folder)
		do
			folder := attached_string (global.convention.clean_directory_path (a_folder))
		end

feature -- Access

	folder: attached STRING
			-- Path to this delivery's folder
			-- Example: "\\axaros.com\shares\production"

	folder_with_forward_slashes: attached STRING
			-- Path to this delivery's folder using forward slashes
			-- Example: "//axaros.com/shares/production"
		do
			Result := attached_string (folder.twin)
			Result.replace_substring_all ("\", "/")
		end

	short_name: attached STRING
			-- Short name for this delivery (example: "production" or "release_candidate")
		do
			Result := attached_string (global.formatter.file_name (folder))
		end

	path (a_relative_path: STRING): attached STRING
			-- Full path to file in this delivery with 'a_relative_path'
			-- Example: "tools\foo.cfg" -> "\\axaros.com\shares\production\tools\foo.cfg"
		require
			is_relative: global.formatter.is_relative_path (a_relative_path)
		do
			Result := attached_string (global.formatter.composed_path (<<folder, a_relative_path>>))
		ensure
			contained: contains_path (Result)
		end

	path_to_site_cfg: attached STRING
			-- Full path to site.cfg within this delivery
			-- Example: "\\axaros.com\shares\production\site.cfg"
		do
			Result := path (global.deliveries.site_configuration_file_name)
		ensure
			contained: contains_path (Result)
		end

	path_to_delivery_cfg: attached STRING
			-- Full path to site.cfg within this delivery
			-- Example: "\\axaros.com\shares\production\delivery.cfg"
		do
			Result := path (global.deliveries.delivery_configuration_file_name)
		ensure
			contained: contains_path (Result)
		end

	path_to_common: attached STRING
			-- Full path to 'common' within this delivery
			-- Example: "\\axaros.com\shares\production\common"
		do
			Result := path ("common")
		ensure
			contained: contains_path (Result)
		end

	path_to_common_icon: attached STRING
			-- Full path to common\icon within this delivery
			-- Example: "\\axaros.com\shares\production\common\icon"
		do
			Result := path ("common\icon")
		ensure
			contained: contains_path (Result)
		end

	path_to_process_monitor: attached STRING
			-- Full path to process_monitor.exe in this delivery
		do
			Result := path ("tools\process_monitor.exe")
		end

	relative_path (a_absolute_path: STRING): attached STRING
			-- Part of 'a_absolute_path' that is relative to this delivery
			-- Example: "\\axaros.com\shares\production\tools\foo.cfg" -> "tools\foo.cfg"
		require
			is_absolute: global.formatter.is_absolute_path (a_absolute_path)
			path_contained: contains_path (a_absolute_path)
		do
			if folder.count > a_absolute_path.count then
				Result := attached_string (a_absolute_path.substring (folder.count + 1, a_absolute_path.count))
			else
				create Result.make_empty
			end
		ensure
			is_relative: global.formatter.is_relative_path (Result)
			empty_when_root: (a_absolute_path.as_lower ~ folder.as_lower) implies Result.is_empty
		end

	contains_path (a_path: STRING): BOOLEAN
			-- Is 'a_path' pointing to a file under 'production_site'?
		do
			Result := a_path /= Void and then a_path.substring (1, folder.count).as_lower ~ folder.as_lower
		ensure
			absolute_if_contained: Result implies global.formatter.is_absolute_path (a_path)
		end

	debug_output: STRING
			-- String that should be displayed in debugger to represent `Current'.
		do
			Result := folder
		end

feature {NONE} -- Implementation

	attached_string (a_string: STRING): attached STRING
			-- Verified attached string from 'a_string'
			-- This feature is needed only as a workaround until we get our code straight
			-- FIXME: to be removed when the code integrates with "on demand Void safety"
		do
			if attached a_string as l_string then
				Result := l_string
			else
				-- never reached, only here to satisfy compiler in "on demand Void safety" mode
				create Result.make_empty
				error.raise ("Unexpected void string received in ROSE_DELIVERY_CONVENTION")
			end
		end

end
