note

	description:

		"Deserialize an eiffel object represented in XML by the ROSE_SERIALIZER"

	notice: "rose_serializer.dtd"
	author: "Zoran Simic"
	last_editor: "$Author: er $"
	date: "$Date: 2012/10/26 $"
	revision: "$Revision: #27 $"
	path: "$File: //rose/source/kernel/infrastructure/framework/serialization/windows/classic/rose_deserializer.e $"

class
	ROSE_DESERIALIZER

inherit
	ANY -- Export features from ANY
	ROSE_SERIALIZATION_CONSTANTS
		export
			{NONE} all
		end

	ENHANCED_INTERNAL
		export
			{NONE} all
		end

feature -- Access

	id : STRING is "XML deserializer"		-- Id

feature -- Basic operations

	deserialized (a_file_name: STRING): ANY is
			-- Serialize 'a_object' onto 'a_file_name'.
			-- Eventual errors are signaled
		require
			valid_file_name: a_file_name /= Void and then not a_file_name.is_empty
		local
			l_filename: STRING
			a_file: PLAIN_TEXT_FILE
			rescued: STRING
		do
			if rescued /= Void then
				error.signal (generator,"Could not deserialize '$1': $2",<<a_file_name,rescued>>)
			else
				l_filename := global.persist.read_storable_filename (a_file_name)
				create a_file.make (l_filename)
				if not a_file.exists or else not a_file.is_readable then
					error.signal (generator,"Cannot deserialize '$1', not readable",<<l_filename>>)
				end
			end
			if not error.signaled then
				create types.make (0)
				create refs.make (0)
				create fields.make (0)
				Result := parsed (a_file)
			end
			if a_file /= Void and then not a_file.is_closed then
				a_file.close
			end
			types := Void
			refs := Void
			fields := Void
		rescue
			if rescued = Void then
				rescued := last_exception_explanation
				retry
			end
		end

feature {NONE} -- Deserialization

	parsed (a_file: FILE): ANY is
			-- Parsed retreived object from 'a_file'
		require
			a_file_not_void: a_file /= Void
		do
			a_file.open_read
				--| Read XML header
			a_file.read_line
				--| Read root 'serialized' element declaration
			a_file.read_line
			read_types (a_file)
			a_file.read_line
			Result := read_ref_object (a_file)
		end

	read_types (a_file: FILE) is
			-- Read <types> section.
		require
			a_file_not_void: a_file /= Void
		local
			l_line,l_type: STRING
			i,l_dynamic_type: INTEGER
		do
				--| Read types element declaration
			a_file.read_line
			from
				a_file.read_line
				l_line := a_file.last_string
			until
				l_line.item (2) /= 't'
			loop
				i := l_line.index_of ('"',7)
				l_type := l_line.substring (7,i-1)
				i := l_line.substring (i+5,l_line.count-3).to_integer
				l_dynamic_type := dynamic_type_from_string (l_type)
				if l_dynamic_type > 0 then
					types.put (l_dynamic_type,i)
				else
					error.warn ("Unknown type '$1'",<<l_type>>)
				end
				a_file.read_line
				l_line := a_file.last_string
			end
		end

feature {ROSE_SERIALIZER_SPECIFIC} -- Implementation

	read_ref_object (a_file: FILE): ANY is
			-- Return reference object defined at current position in 'a_file'
		require
			a_file_not_void: a_file /= Void
		local
			l_line: STRING
		do
			l_line := a_file.last_string
			inspect l_line.item (2)
			when 'a' then
				Result := get_object (l_line)
				if l_line.item (l_line.count-1) /= '/' then
					read_in_any (a_file, Result)
				end
			when 'f' then
				Result := get_ref (xm_argument (l_line).to_integer)
			when 's' then
					Result := get_string (l_line)
			when 'N' then
				Result := get_special_natural (a_file, l_line)
			when 'x' then
				Result := get_special (a_file, l_line)
			when 'v' then
				--| Void
			else
				error.warn ("Unrecognized line '$1'", <<l_line>>)
			end
		end

	register_ref (a_any: ANY; a_id: STRING) is
			-- Register 'a_id' to be the reference number for 'a_any'.
		do
			refs.put (a_any,a_id.to_integer)
		end

	get_object (a_line: STRING): ANY is
			-- Return actual empty Eiffel object referenced by 'a_line'
			-- and register it with it's id
		local
			i: INTEGER
			l_pair: PAIR [STRING,STRING]
		do
			l_pair := xm_argument2 (a_line)
			i := types.item (l_pair.first.to_integer)
			if i > 0 then
				Result := new_instance_of (i)
				register_ref (Result,l_pair.second)
			else
				error.warn ("Unknown object type $1 in '$2'",<<l_pair.first,a_line>>)
			end
		end

	xm_argument (a_line: STRING): STRING is
			-- Return the first argument from 'a_line'
		local
			i,j: INTEGER
		do
			if a_line.item (3) = ' ' then
				i := 4
			else
				i := 5
			end
			if a_line.item (i) = 'n' then
				i := a_line.index_of ('"',i+3) + 5
			else
				i := i + 3
			end
			j := a_line.index_of ('"',i)
			Result := a_line.substring (i,j-1)
		end

	xm_argument2 (a_line: STRING): PAIR [STRING,STRING] is
			-- Return the first 2 arguments from 'a_line'
		local
			i,j: INTEGER
		do
			if a_line.item (3) = ' ' then
				i := 4
			else
				i := 5
			end
			if a_line.item (i) = 'n' then
				i := a_line.index_of ('"',i+3) + 5
			else
				i := i+3
			end
			j := a_line.index_of ('"',i)
			create Result.make (a_line.substring (i,j-1),Void)
			i := a_line.index_of ('"',j+5)
			Result.set_second (a_line.substring (j+5,i-1))
		end

feature {NONE} -- Deserialization

	read_in_any (a_file: FILE; a_any: ANY) is
			-- Read objects
		local
			l_object: ANY
			l_line, l_name: STRING
		do
			from
				a_file.read_line
				l_line := a_file.last_string
			until
				l_line.item (2) = '/'
			loop
				if l_line.item (3) = ' ' then
					l_name := l_line.substring (7,l_line.index_of ('"',7)-1)
				elseif l_line.count > 8 then
					l_name := l_line.substring (8,l_line.index_of ('"',8)-1)
				end
				inspect l_line.item (2)
				when 'a', 'f', 's', 'x', 'N' then
					l_object := read_ref_object (a_file)
					put_any (a_any,l_object,l_name)
				when 'i' then
					inspect l_line.item (3)
					when '1' then
						put_integer_8 (a_any,get_integer_8 (l_line),l_name)
					when '2' then
						put_integer_16 (a_any,get_integer_16 (l_line),l_name)
					when ' ' then
						put_integer_32 (a_any,get_integer_32 (l_line),l_name)
					when '8' then
						put_integer_64 (a_any,get_integer_64 (l_line),l_name)
					else
						error.warn ("Unknown integer size $1 in '$2'",<<l_line.item (3),l_line>>)
					end
				when 'n' then
					inspect l_line.item (3)
					when '1' then
						put_natural_8 (a_any, get_natural_8 (l_line), l_name)
					when '2' then
						put_natural_16 (a_any, get_natural_16 (l_line), l_name)
					when '4' then
						put_natural_32 (a_any, get_natural_32 (l_line), l_name)
					when '8' then
						put_natural_64 (a_any, get_natural_64 (l_line), l_name)
					else
						error.warn ("Unknown natural size $1 in '$2'", <<l_line.item (3), l_line>>)
					end
				when 'd' then
					put_double (a_any,get_double (l_line),l_name)
				when 'r' then
					put_real (a_any,get_real (l_line),l_name)
				when 'c' then
					put_character (a_any,get_character (l_line),l_name)
				when 'C' then
					put_character_32 (a_any,get_character_32 (l_line),l_name)
				when 'b' then
					put_boolean (a_any,l_name)
				when 'e' then
					error.warn ("Expanded type ignored in '$1'",<<l_line>>)
				when 'v' then
					--| Do nothing
				else
					error.warn ("Unknown element type $1 in '$2'",<<l_line.item (2),l_line>>)
				end
				a_file.read_line
				l_line := a_file.last_string
			end
		end

feature {NONE} -- Fetching

	get_ref (a_id: INTEGER): ANY is
			-- Previously read object
		do
			Result := refs.item (a_id)
		end

	get_string (a_line: STRING): ANY is
			-- Decoded string
		local
			l_pair: PAIR [STRING,STRING]
			l_string: STRING
		do
			l_pair := xm_argument2 (a_line)
			l_string := l_pair.second
			l_string.replace_substring_all (markup_quote_escaped,markup_quote)
			l_string.replace_substring_all (markup_start_escaped,markup_start)
			l_string.replace_substring_all (markup_end_escaped,markup_end)
			l_string.replace_substring_all (markup_ampersand_escaped,markup_ampersand)
			l_string.replace_substring_all (markup_newline_escaped,markup_newline)
			Result := l_string
			register_ref (Result,l_pair.first)
		end

	get_special_natural (a_file: FILE; a_line: STRING): ANY is
		do
			inspect a_line.item (3)
			when '1' then
				Result := specific_natural8.get_special_encoded (Current,a_file,a_line)
			when '2' then
				Result := specific_natural16.get_special_encoded (Current,a_file,a_line)
			when '4' then
				Result := specific_natural32.get_special_encoded (Current,a_file,a_line)
			when '8' then
				Result := specific_natural64.get_special_encoded (Current,a_file,a_line)
			end
		end

	get_special (a_file: FILE; a_line: STRING): ANY is
		do
			inspect a_line.item (3)
			when 'i' then
				Result := specific_int32.get_special_encoded (Current,a_file,a_line)
			when '1' then
				Result := specific_int8.get_special_encoded (Current,a_file,a_line)
			when '2' then
				Result := specific_int16.get_special_encoded (Current,a_file,a_line)
			when '8' then
				Result := specific_int64.get_special_encoded (Current,a_file,a_line)
			when 'c' then
				Result := specific_char.get_special_encoded (Current,a_file,a_line)
			when 'C' then
				Result := specific_char32.get_special_encoded (Current,a_file,a_line)
			when 'b' then
				Result := specific_bool.get_special_encoded (Current,a_file,a_line)
			when 'd' then
				Result := specific_double.get_special_encoded (Current,a_file,a_line)
			when 'r' then
				Result := specific_real.get_special_encoded (Current,a_file,a_line)
			when 'a' then
				Result := get_special_any (a_file,a_line)
			else
				error.warn ("Unknown special type '$1' in '$2'",<<a_line.item (3),a_line>>)
			end
		end

	get_special_any (a_file: FILE; a_line: STRING): SPECIAL [ANY] is
		local
			l_any: ANY
			i,j,t: INTEGER
			l_id: STRING
		do
--			l_any := get_object (a_line)
			if a_line.item (5) = 'n' then
				i := a_line.index_of ('"',8) + 5
			else
				i := 8
			end
			j := a_line.index_of ('"',i)
			t := a_line.substring (i,j-1).to_integer
			j := j + 5
			i := a_line.index_of ('"',j)
			l_id := a_line.substring (j,i-1)
			i := i + 5
			j := a_line.index_of ('"',i)
			i := a_line.substring (i,j-1).to_integer
			t := types.item (t)
			if t > 0 then
				Result := new_special_any_instance_filled (t,i)
				register_ref (Result,l_id)
				from
					a_file.readline
					i := 0
				until
					a_file.last_string.count > 1 and then a_file.last_string.item (2) = '/'
				loop
					l_any := read_ref_object (a_file)
					Result.put (l_any,i)
					i := i + 1
					a_file.readline
				end
			else
				error.warn ("Unknown SPECIAL [] type $1 in '$2'",<<t,a_line>>)
			end
		end

	get_integer_8 (a_line: STRING): INTEGER_8 is
		local
			l_string: STRING
		do
			l_string := xm_argument (a_line)
			if l_string /= Void and then l_string.is_integer_8 then
				Result := xm_argument (a_line).to_integer_8
			end
		end

	get_integer_16 (a_line: STRING): INTEGER_16 is
		local
			l_string: STRING
		do
			l_string := xm_argument (a_line)
			if l_string /= Void and then l_string.is_integer_16 then
				Result := xm_argument (a_line).to_integer_16
			end
		end

	get_integer_32 (a_line: STRING): INTEGER is
		local
			l_string: STRING
		do
			l_string := xm_argument (a_line)
			if l_string /= Void and then l_string.is_integer then
				Result := xm_argument (a_line).to_integer
			end
		end

	get_integer_64 (a_line: STRING): INTEGER_64 is
		local
			l_string: STRING
		do
			l_string := xm_argument (a_line)
			if l_string /= Void and then l_string.is_integer_64 then
				Result := xm_argument (a_line).to_integer_64
			end
		end

	get_natural_8 (a_line: STRING): NATURAL_8 is
		local
			l_string: STRING
		do
			l_string := xm_argument (a_line)
			if l_string /= Void and then l_string.is_natural_8 then
				Result := xm_argument (a_line).to_natural_8
			end
		end

	get_natural_16 (a_line: STRING): NATURAL_16 is
		local
			l_string: STRING
		do
			l_string := xm_argument (a_line)
			if l_string /= Void and then l_string.is_natural_16 then
				Result := xm_argument (a_line).to_natural_16
			end
		end

	get_natural_32 (a_line: STRING): NATURAL_32 is
		local
			l_string: STRING
		do
			l_string := xm_argument (a_line)
			if l_string /= Void and then l_string.is_natural_32 then
				Result := xm_argument (a_line).to_natural_32
			end
		end

	get_natural_64 (a_line: STRING): NATURAL_64 is
		local
			l_string: STRING
		do
			l_string := xm_argument (a_line)
			if l_string /= Void and then l_string.is_natural_64 then
				Result := xm_argument (a_line).to_natural_64
			end
		end

	get_double (a_line: STRING): DOUBLE is
		local
			l_string: STRING
		do
			l_string := xm_argument (a_line)
			if l_string /= Void and then l_string.is_double then
				Result := l_string.to_double
			else
				Result := ieee.qnan
			end
		end

	get_real (a_line: STRING): REAL is
			-- Real from 'a_line'
		local
			l_arg: STRING
		do
			l_arg := xm_argument (a_line)
			if l_arg.is_real then
				Result := l_arg.to_real
			else
				Result := ieee.qnan.truncated_to_real
			end
		end

	ieee: IEEE_MATH
			-- Contains 'qnan'.
		once
			create Result
		end

	get_character (a_line: STRING): CHARACTER is
			-- Pull the middle characters out of 'a_line' and turn it into a CHARACTER_8.
		local
			s: STRING
		do
			s := xm_argument (a_line)
			s.remove (1)
			s.remove (1)
			s.remove (s.count)
			Result := s.to_integer.to_character_8
		end

	get_character_32 (a_line: STRING): CHARACTER_32
			-- Pull the middle characters out of 'a_line' and turn it into a CHARACTER_32.
		local
			s: STRING
		do
			s := xm_argument (a_line)
			s.remove (1)
			s.remove (1)
			s.remove (s.count)
			Result := s.to_natural_32.to_character_32
		end

feature {NONE} -- Deserialization

	put_integer_8 (a_any: ANY; a_value: INTEGER_8; a_name: STRING) is
		local
			i: INTEGER
		do
			i := field_index (a_any,a_name)
			if i > 0 then
				set_integer_8_field (i,a_any,a_value)
			elseif a_any /= Void then
				error.warn ("No INTEGER_8 attribute '$1' in class $2",<<a_name,a_any.generator>>)
			end
		end

	put_integer_16 (a_any: ANY; a_value: INTEGER_16; a_name: STRING) is
		local
			i: INTEGER
		do
			i := field_index (a_any,a_name)
			if i > 0 then
				set_integer_16_field (i,a_any,a_value)
			elseif a_any /= Void then
				error.warn ("No INTEGER_16 attribute '$1' in class $2",<<a_name,a_any.generator>>)
			end
		end

	put_integer_32 (a_any: ANY; a_value: INTEGER; a_name: STRING) is
		local
			i: INTEGER
		do
			i := field_index (a_any,a_name)
			if i > 0 then
				set_integer_field (i,a_any,a_value)
			elseif a_any /= Void then
				error.warn ("No INTEGER attribute '$1' in class $2",<<a_name,a_any.generator>>)
			end
		end

	put_integer_64 (a_any: ANY; a_value: INTEGER_64; a_name: STRING) is
		local
			i: INTEGER
		do
			i := field_index (a_any,a_name)
			if i > 0 then
				set_integer_64_field (i,a_any,a_value)
			elseif a_any /= Void then
				error.warn ("No INTEGER_64 attribute '$1' in class $2",<<a_name,a_any.generator>>)
			end
		end

	put_natural_8 (a_any: ANY; a_value: NATURAL_8; a_name: STRING) is
		local
			i: INTEGER
		do
			i := field_index (a_any, a_name)
			if i > 0 then
				set_natural_8_field (i, a_any, a_value)
			elseif a_any /= Void then
				error.warn ("No NATURAL_8 attribute '$1' in class $2",<<a_name,a_any.generator>>)
			end
		end

	put_natural_16 (a_any: ANY; a_value: NATURAL_16; a_name: STRING) is
		local
			i: INTEGER
		do
			i := field_index (a_any,a_name)
			if i > 0 then
				set_natural_16_field (i,a_any,a_value)
			elseif a_any /= Void then
				error.warn ("No NATURAL_16 attribute '$1' in class $2",<<a_name,a_any.generator>>)
			end
		end

	put_natural_32 (a_any: ANY; a_value: NATURAL_32; a_name: STRING) is
		local
			i: INTEGER
		do
			i := field_index (a_any,a_name)
			if i > 0 then
				set_natural_32_field (i,a_any,a_value)
			elseif a_any /= Void then
				error.warn ("No NATURAL_32 attribute '$1' in class $2",<<a_name,a_any.generator>>)
			end
		end

	put_natural_64 (a_any: ANY; a_value: NATURAL_64; a_name: STRING) is
		local
			i: INTEGER
		do
			i := field_index (a_any,a_name)
			if i > 0 then
				set_natural_64_field (i,a_any,a_value)
			elseif a_any /= Void then
				error.warn ("No NATURAL_64 attribute '$1' in class $2",<<a_name,a_any.generator>>)
			end
		end

	put_character (a_any: ANY; a_value: CHARACTER; a_name: STRING) is
		local
			i: INTEGER
		do
			i := field_index (a_any,a_name)
			if i > 0 then
				set_character_field (i,a_any,a_value)
			elseif a_any /= Void then
				error.warn ("No CHARACTER attribute '$1' in class $2",<<a_name,a_any.generator>>)
			end
		end

	put_character_32 (a_any: ANY; a_value: CHARACTER_32; a_name: STRING)
			-- Put 'a_value' into 'a_any' at the attribute named 'a_name'.
		local
			i: INTEGER
		do
			i := field_index (a_any,a_name)
			if i > 0 then
				set_character_32_field (i,a_any,a_value)
			elseif a_any /= Void then
				error.warn ("No CHARACTER_32 attribute '$1' in class $2",<<a_name,a_any.generator>>)
			end
		end

	put_boolean (a_any: ANY; a_name: STRING) is
		local
			i: INTEGER
		do
			i := field_index (a_any,a_name)
			if i > 0 then
				set_boolean_field (i,a_any,True)
			elseif a_any /= Void then
				error.warn ("No BOOLEAN attribute '$1' in class $2",<<a_name,a_any.generator>>)
			end
		end

	put_double (a_any: ANY; a_value: DOUBLE; a_name: STRING) is
			-- Set attribute with 'a_name' in object 'a_any' to 'a_value' using INTERNAL
		local
			i: INTEGER
		do
			i := field_index (a_any,a_name)
			if i > 0 then
				set_double_field (i,a_any,a_value)
			elseif a_any /= Void then
				error.warn ("No DOUBLE attribute '$1' in class $2",<<a_name,a_any.generator>>)
			end
		end

	put_real (a_any: ANY; a_value: REAL; a_name: STRING) is
			-- Set attribute with 'a_name' in object 'a_any' to 'a_value' using INTERNAL
		local
			i: INTEGER
		do
			i := field_index (a_any,a_name)
			if i > 0 then
				set_real_field (i,a_any,a_value)
			elseif a_any /= Void then
				error.warn ("No REAL attribute '$1' in class $2",<<a_name,a_any.generator>>)
			end
		end

	put_any (a_any: ANY; a_value: ANY; a_name: STRING) is
			-- Set attribute with 'a_name' in object 'a_any' to 'a_value' using INTERNAL
		local
			i: INTEGER
		do
			i := field_index (a_any,a_name)
			if i > 0 then
				set_reference_field (i,a_any,a_value)
			elseif a_any /= Void then
				error.warn ("No reference attribute '$1' in class $2",<<a_name,a_any.generator>>)
			end
		end

feature {NONE} -- Implementation

	field_index (a_any: ANY; a_name: STRING): INTEGER is
			-- Index of attribute `a_name' in object `a_any'
		local
			h: HASH_TABLE [INTEGER,STRING]
			i,j,n: INTEGER
		do
			if a_any /= Void then
				i := dynamic_type (a_any)
				h := fields.item (i)
				if h = Void then
					from
						create h.make (0)
						n := field_count (a_any)
						j := 1
					until
						j > n
					loop
						h.put (j,field_name (j,a_any))
						j := j + 1
					end
					fields.put (h,i)
				end
				Result := h.item (a_name)
			end
		end

	types: HASH_TABLE [INTEGER,INTEGER]
			-- Class type ids

	refs: HASH_TABLE [ANY,INTEGER]
			-- Referenced objects table, count gives the number of objects serialized

	fields: HASH_TABLE [HASH_TABLE [INTEGER,STRING],INTEGER]

end
