indexing
	description: "Helper class for feedback while traversing directories"
	author: "Zoran Simic"
	date: "$Date: 2014/02/26 $"
	revision: "$Revision: #5 $"

class
	ROSE_DIRECTORY_SCAN_REPORT

inherit
	ROSE_SHARED_GLOBAL

create
	make, make_with_callback

feature {NONE} -- Initialization

	make is
			-- Creation procedure.
		do
			create status
			create info.make
			create dirs.make (32)
			create files.make (32)
			cancel := False
			done := -1
			total := 0
			last_done := -callback_threshold - 1
			last_total := -callback_threshold - 1
		end

	make_with_callback (a_callback: like callback) is
			-- Make with 'a_callback'
		do
			make
			set_callback (a_callback)
		end

feature -- Access

	status: ROSE_STATUS						-- Status of current traversal
	dirs: ARRAYED_LIST [STRING]				-- List of directory paths scanned
	files: ARRAYED_LIST [STRING]			-- List of file paths scanned
	cancel: BOOLEAN							-- Set to true to stop current scanning
	done: INTEGER							-- Number of items scanned so far (-1 when scan not in progress)
	total: INTEGER							-- Total number of items found

	callback_threshold: INTEGER is 49		-- Threshold for calling back 'callback'

	callback: FUNCTION [ANY, TUPLE [STRING, INTEGER, INTEGER], BOOLEAN]
			-- Callback to report where the scanning currently is

feature -- Basic operations

	set_callback (a_callback: like callback) is
			-- Set 'callback'.
		do
			callback := a_callback
		end

	errorf (a_message: STRING; a_args: ARRAY [ANY]) is
			-- Report an error in 'status'.
		do
			status.errorf (a_message, a_args)
		end

	increment_done (a_amount: INTEGER; a_path: STRING) is
			-- Increment 'done' and call 'callback' if provided
		require
			amount_ok: a_amount >= 0
			path_ok: a_path /= Void
		do
			if a_amount = 0 then
				if last_total + callback_threshold < total then
					last_total := total
					safe_callback (a_path)
				end
			else
				done := done + a_amount
				if last_done + callback_threshold < done then
					last_done := done
					safe_callback (a_path)
				end
			end
		end

	increment_total (a_path: STRING) is
			-- Increment 'total'.
		do
			total := total + 1
			if last_total + callback_threshold < total then
				last_total := total
				safe_callback (a_path)
			end
		end

	terminate is
			-- Terminate scanning
		do
			last_done := - callback_threshold - 1
			last_total := - callback_threshold - 1
			done := 0
			safe_callback (Void)
		end

	recursive_gather (a_path: STRING) is
			-- Recursively gather file and directory paths from 'a_path' to 'dirs' and 'files'
		require
			path_ok: a_path /= Void
		do
			do_recursive_gather (a_path)
			terminate
		end

feature {EL_DIRECTORY} -- Implementation

	info: UNIX_FILE_INFO			-- Used to get info on files (creation date, modification date, etc)

	recursive_delete (a_retries: INTEGER) is
			-- Delete all items in 'dirs' and 'files', call 'callback' for progress feedback
		local
			l_path: STRING
			i: INTEGER
		do
			from
				i := 1
			until
				not status.is_ok or else cancel or else i > files.count
			loop
				l_path := files.i_th (i)
				increment_done (1, l_path)
				if not cancel then
					delete_file (l_path, a_retries)
				end
				i := i + 1
			end
			from
				i := dirs.count
			until
				not status.is_ok or else cancel or else i < 1
			loop
				l_path := dirs.i_th (i)
				increment_done (1, l_path)
				if not cancel then
					delete_dir (l_path, a_retries)
				end
				i := i - 1
			end
			if cancel then
				status.warnf ("Operation cancelled%NRight after deleting $1 files out of $2%NStopped after deleting '$3'", <<done, total, l_path>>)
			else
				safe_callback (Void)
			end
		end

feature {NONE} -- Implementation

	tmp_file: RAW_FILE is
			-- Temporary file used to avoid creating many temp file objects
		once
			create Result.make ("temp")
		end

	last_total, last_done: INTEGER
			-- Used to remember when 'callback' was last called

	do_recursive_gather (a_path: STRING) is
			-- Recursively gather file and directory paths from 'a_path' to 'dirs' and 'files'
		require
			path_ok: a_path /= Void
		local
			l_dir: DIRECTORY
			l_entry, l_path: STRING
			retried: BOOLEAN
		do
			if retried then
				status.outputf ("Can not read directory '$1'", <<a_path>>)
			elseif status.is_ok and then not cancel then
				from
					create l_dir.make (a_path)
					l_dir.open_read
					l_dir.start
					l_dir.readentry
					l_entry := l_dir.lastentry
				until
					not status.is_ok or else cancel or else l_entry = Void
				loop
					if is_valid_directory_entry (l_entry) then
						l_path := global.formatter.file_system.pathname (a_path, l_entry)
						info.update (l_path)
						if info.is_directory then
							if dirs /= Void then
								dirs.extend (l_path)
							end
							increment_total (l_path)
							do_recursive_gather (l_path)
						else
							if files /= Void then
								files.extend (l_path)
							end
							increment_total (l_path)
						end
					end
					l_dir.readentry
					l_entry := l_dir.lastentry
				end
				l_dir.close
				increment_done (0, a_path)
			end
		rescue
			if not retried then
				retried := True
				retry
			end
		end

	safe_callback (a_path: STRING) is
			-- Call 'callback' safely and return its value safely
		local
			retried: BOOLEAN
		do
			if not retried
			and then callback /= Void
			and then not cancel
			and then status.is_acceptable then
				cancel := callback.item ([a_path, done.min (total), total])
			end
		rescue
			if not retried then
				retried := True
				retry
			end
		end

	delete_dir (a_path: STRING; a_retries: INTEGER) is
			-- Delete dir with 'a_path', report in 'status'
		local
			retried: INTEGER
		do
			if retried > a_retries then
				status.errorf ("Can not delete directory '$1'", <<a_path>>)
				status.output ("Please make sure it is not used by another process")
				status.output ("You can find out which process has this directory open using Process Explorer from http://sysinternals.com")
			else
				tmp_file.make (a_path)
				if tmp_file.exists then
					tmp_file.change_mode (219)
					tmp_file.delete
				end
			end
		rescue
			if retried <= a_retries then
				retried := retried + 1
				global.env.sleep (100)
				retry
			end
		end

	delete_file (a_path: STRING; a_retries: INTEGER) is
			-- Delete file with 'a_path', report in 'status'
		local
			retried: INTEGER
		do
			if retried > a_retries then
				status.errorf ("Can not delete file '$1'", <<a_path>>)
				status.output ("Please make sure it is not used by another process")
				status.output ("You can find out which process has this file open using Process Explorer from http://sysinternals.com")
			else
				tmp_file.make (a_path)
				tmp_file.change_mode (219)		-- 0c333
				tmp_file.delete
			end
		rescue
			if retried <= a_retries then
				retried := retried + 1
				global.env.sleep (100)
				retry
			end
		end

	is_valid_directory_entry (a_entry: STRING): BOOLEAN is
			-- Is 'a_entry' to be further examined?
		do
			if a_entry.count > 0 then
				if a_entry.item (1) = '.' then
					if a_entry.count = 2 then
						Result := a_entry.item (2) /= '.'
					else
						Result := a_entry.count > 2
					end
				else
					Result := True
				end
			end
		end

end
