indexing

	description: "Simple monomorphic dotted expression processor"
	date: "$Date: 2009/06/23 $"
	revision: "$Revision: #6 $"
	path : "$File: //rose/source/kernel/infrastructure/support/rose_dot_expression.e $"

	-- expr := object | expr '.' binary_operator
	-- object := unary_operator | leaf
	-- unary_operator := name
	-- binary_operator := name
	-- name := [^.]+
	-- leaf := [^.]+

	-- names must be that of registered operator for valid expression.
	-- leaf is defined by (validator,parser) agents
	-- named operators are defined by agents

class ROSE_DOT_EXPRESSION [G]

create
	make

feature -- Setting

	make (a_validator: like leaf_validator; a_parser: like leaf_parser) is
			-- Set validator and parser for leaf expressions.
		require
			a_validator_not_void: a_validator /= Void
			--well_formed_leaf_validator: for_all a_string where a_string.has (c_dot) then not a_validator.item ([a_string])
			well_formed_leaf_validator_mini: not a_validator.item ([c_dot.out])
			a_parser_not_void: a_parser /= Void
		do
			leaf_validator := a_validator
			leaf_parser := a_parser

			create leaf_operators.make (0)
			create unary_operators.make (0)
		end

	add_leaf_operator (a_operator: FUNCTION [ANY, TUPLE[], G]; a_name: STRING) is
			-- Add named leaf expression.
		require
			a_name_not_void: a_name /= Void
			a_name_valid: not a_name.has (c_dot)
			a_operator_not_void: a_operator /= Void
		do
			leaf_operators.force (a_operator, a_name)
		ensure
			has: leaf_operators.has (a_name)
		end

	add_unary_operator (a_operator: FUNCTION [ANY, TUPLE[G], G]; a_name: STRING) is
			-- Add unary operator.
		require
			a_name_not_void: a_name /= Void
			a_name_valid: not a_name.has (c_dot)
			a_operator_not_void: a_operator /= Void
		do
			unary_operators.force (a_operator, a_name)
		ensure
			has: unary_operators.has (a_name)
		end

feature {NONE} -- Implementation

	leaf_validator: FUNCTION [ANY, TUPLE [STRING], BOOLEAN]
			-- Leaf validator

	leaf_parser:  FUNCTION [ANY, TUPLE [STRING], G]
			-- Leaf constants parser

	leaf_operators: HASH_TABLE [FUNCTION [ANY, TUPLE [], G], STRING]
			-- Named leaf operators

	unary_operators: HASH_TABLE [FUNCTION [ANY, TUPLE[G], G], STRING]
			-- Unary operators

feature -- Validation

	is_expression (a_string: STRING): BOOLEAN is
			-- Is this a valid expression?
		local
			l_list: LIST [STRING]
		do
			if a_string /= Void then
				l_list := a_string.split (c_dot)
				Result := l_list.count > 0 and then is_leaf (l_list.first)
				if Result then
					l_list.start
					l_list.remove
					Result := l_list.for_all (agent is_operator)
				end
			end
		end

	is_operator (a_string: STRING): BOOLEAN is
			-- Is this a valid operator?
		local
			l_list: LIST [STRING]
		do
			if a_string /= Void then
				l_list := a_string.split (c_open_paren)
				Result := a_string /= Void
					and then unary_operators.has (l_list.first)
			end
		ensure
			no_separator: Result implies not a_string.has (c_dot)
		end

	is_leaf (a_string: STRING): BOOLEAN is
			-- Is this a valid leaf expression?
		do
			Result := a_string /= Void
				and then leaf_operators.has (a_string) or leaf_validator.item ([a_string])
		ensure
			no_separator: Result implies not a_string.has (c_dot)
		end


feature -- Query

	item (a_string: STRING): G is
			-- Result of executed expression.
		require
			is_expression: is_expression (a_string)
		local
			l_result: CELL [G]
			l_g : G
		do
			create l_result.put(l_g)
			a_string.split (c_dot).do_all (agent execute_one (l_result, ?))
			Result := l_result.item
		end

feature {NONE} -- Implementation

	execute_one (a_result: CELL [G]; a_string: STRING) is
			-- Execute one expression.
		require
			a_result_not_void: a_result /= Void
			valid_first: a_result.item = Void implies is_leaf (a_string)
			valid_tail: a_result.item /= Void implies is_operator (a_string)
		local
			l_list: LIST [STRING]
			l_operator, l_param, l_param_2: STRING
		do
			if a_result.item = Void then
					-- This is the first item in a dotted expression
				if leaf_operators.has (a_string) then
					a_result.put (leaf_operators.item (a_string).item ([]))
				else
					a_result.put (leaf_parser.item ([a_string]))
				end
			else
					-- a.b
				if a_string.has(c_open_paren) then
					l_list := a_string.split (c_open_paren)
					l_operator := l_list.first
					l_param := l_list.i_th(2)
					if not a_string.has(c_comma) then
						l_param.prune_all_trailing(c_close_paren)
						a_result.put (unary_operators.item (l_operator).item ([a_result.item,l_param]))
					else
						l_list := l_param.split (c_comma)
						l_param := l_list.i_th(1)
						l_param.prune_all(c_quote)
						l_param_2 := l_list.i_th(2)
						l_param_2.prune_all(c_quote)
						l_param_2.prune_all_trailing(c_close_paren)
						a_result.put (unary_operators.item (l_operator).item ([a_result.item,l_param,l_param_2]))
					end
				else
					a_result.put (unary_operators.item (a_string).item ([a_result.item]))
				end
			end

		end

feature -- Constant

	c_dot: CHARACTER is '.'
	c_quote: CHARACTER is '%''
	c_double_quote: CHARACTER is '"'
	c_comma: CHARACTER is ','
	c_open_paren: CHARACTER is '('
	c_close_paren: CHARACTER is ')'
			-- Separator

invariant

	leaf_validator_not_void: leaf_validator /= Void
	leaf_parser_not_void: leaf_parser /= Void
	leaf_operators_not_void: leaf_operators /= Void
	unary_operators_not_void: unary_operators /= Void

end
