indexing

	description: "[
		Objects that enable entering email recipients.

      The user may also edit the subject line,
      and also the text of the principal text/plain MIME component.

      Expected usage:

      create l_dialog.make ("subject text", <<"line 1", "line 2", "">>)
      l_dialog.ok_button.select_actions.clear
      l_dialog.ok_button.select_actions.extend (my_agent (l_dialog))
      l_dialog.show

      my_agent will then interogate `Current' to get the entered information,
      and then compose and send an email.

      The pair of integers attached to grid rows represent a group index and
       a user index within the group respectively.
      Grid rows that represent a group rather than a user have -1 for the second
       integer.
	]"
	author: "Colin Adams"
	last_editor: "$Author: it $"
	date: "$Date: 2008/12/01 $"
	revision: "$Revision: #4 $"
	path: "$File: //rose/source/kernel/infrastructure/widgets/rose_email_dialog.e $"

class	ROSE_EMAIL_DIALOG

inherit

	ROSE_DIALOG
		redefine
			setup_main_box
		end

	ROSE_SHARED_GLOBAL
		export
			{NONE} all
		undefine
			copy, default_create
		end

create

	make

feature {NONE} -- Initialization

	make (a_subject, a_lines: STRING_8) is
			-- Initialize `subject' and `text'.
		require
			a_subject_not_void: a_subject /= Void
			a_subject_has_no_newlines: not a_subject.has ('%N')
			a_lines_not_void: a_lines /= Void
		do
			create recipients_list
			create recipients_grid.make (grid_titles, Void)
			create recipients_box
			create subject.make_with_text (a_subject)
			create body.make_with_text (a_lines)
			default_create
		ensure
			subject_set: subject.text.is_equal (a_subject)
		end

	setup_main_box is
			-- Set up the layout of `Current'.
		local
			l_label: EV_LABEL
			l_box: EV_HORIZONTAL_BOX
		do
			remove_default_push_button
			ok_button.set_tooltip ("")
			create l_box
			create use_list_radio.make_with_text_and_action ("Type addresses", agent display_recipients_list)
			l_box.extend (use_list_radio)
			create all_groups_radio.make_with_text_and_action ("Select from expandable groups", agent display_recipients_grid)
			l_box.extend (all_groups_radio)
			create all_rosenberg_radio.make_with_text_and_action ("Select from AXA Rosenberg employees", agent display_recipients_grid)
			l_box.extend (all_rosenberg_radio)
			create all_users_radio.make_with_text_and_action ("Select from all employees (very slow)", agent display_recipients_grid)
			l_box.extend (all_users_radio)
			main_box.extend (l_box)
			main_box.disable_item_expand (l_box)
			l_label := recipients_label
			main_box.extend (l_label)
			main_box.disable_item_expand (l_label)
			recipients_box.set_minimum_size (600, 150)
			main_box.extend (recipients_box)
			l_label := subject_label
			main_box.extend (l_label)
			main_box.disable_item_expand (l_label)
			main_box.extend (subject)
			main_box.disable_item_expand (subject)
			body.set_minimum_size (600, 150)
			l_label := body_label
			main_box.extend (l_label)
			main_box.disable_item_expand (l_label)
			main_box.extend (body)
		end

feature -- Access

	subject_line: STRING_8 is
			-- Subject including terminating newline character
		do
			if subject.text.is_valid_as_string_8 then
				Result := subject.text.as_string_8
			else
				Result := ""
			end
		ensure
			subject_line_not_void: result /= Void
			no_new_lines: not Result.has ('%N')
		end

	body_text: STRING_8 is
			-- Multi-line text for principal text component of email
		do
			if body.text.is_valid_as_string_8 then
				Result := body.text.as_string_8 + "%N"
			else
				Result := "%N"
			end
		ensure
			body_text_not_void: Result /= Void
			last_character_is_newline: Result.item (Result.count) = '%N'
		end

	recipients: ROSE_MAIL_RECIPIENTS is
			-- All recipients
		local
			i, l_count: INTEGER
			l_line: STRING_32
		do
			create Result
			if recipients_box.has (recipients_grid) then
				recipients_grid.do_all_rows (agent add_selected_address (Result, ?))
			else
				from
					i := 1
					l_count := recipients_list.line_count
				until
					i > l_count
				loop
					l_line := recipients_list.line (i)
					if l_line.is_valid_as_string_8 and not l_line.is_empty then
						-- Are Unicode address allowed? TODO: Look into this.
						Result.add_address (create {ROSE_MAIL_ADDRESS}.set (l_line.as_string_8))
					end
					i := i + 1
				end
			end
			if Result.is_empty then
				-- send to self
				Result.add_address (create {ROSE_MAIL_ADDRESS}.set (global.app.user.email_address))
			end
		ensure
			recipients_not_void: Result /= Void
			recipients_not_empty: not Result.is_empty
		end

feature {NONE} -- Implementation

	grid_titles: ARRAY [STRING] is
			-- Column titles for `recipients_grid'
		once
			Result := <<"Name", "">>
		ensure
			grid_titles_not_void: Result /= Void
		end

	mail_users: ROSE_MAIL_USERS
			-- Interface to mail server

	subject: EV_TEXT_FIELD
			-- Subject line excluding newline character

	body: EV_TEXT
			-- Widget for editing main body of email

	use_list_radio: EV_RADIO_BUTTON
			-- Option to use `recipients_list'

	all_groups_radio: EV_RADIO_BUTTON
			-- Option to show list of groups in `recipients_grid'

	all_rosenberg_radio: EV_RADIO_BUTTON
			-- Option to show list of AXA Rosenberg employees in `recipients_grid'

	all_users_radio: EV_RADIO_BUTTON
			-- Option to show all users in `recipients_grid'

	recipients_box: EV_CELL
			-- Holder for either `recipients_list' or `recipients_grid'

	recipients_list: EV_TEXT
			-- Widget for typing email recipients

	recipients_grid: RP_DATA_GRID [DS_PAIR [INTEGER, INTEGER]]
			-- Widget for selecting email recipients

	grid_is_groups: BOOLEAN
			-- Is `recipients_grid' displaying groups?

	grid_is_axa_rosenberg: BOOLEAN
			-- Is `recipients_grid' displaying AXA Rosenberg employees?

	grid_is_users: BOOLEAN
			-- Is `recipients_grid' displaying all users?

	recipients_label: EV_LABEL is
			-- Visible caption for `recipients_list'
		do
			create Result.make_with_text ("Recipients:")
		ensure
			recipients_label_not_void: Result /= Void
		end

	subject_label: EV_LABEL is
			-- Visible caption for `subject'
		do
			create Result.make_with_text ("Subject:")
		ensure
			subject_label_not_void: Result /= Void
		end

	body_label: EV_LABEL is
			-- Visible caption for `body'
		do
			create Result.make_with_text ("Content:")
		ensure
			body_label_not_void: Result /= Void
		end

	is_good_name (a_address: ROSE_MAIL_ADDRESS): BOOLEAN is
			-- Does `a_address' contain a good name?
		require
			a_address_not_void: a_address /= Void
		do
			Result := a_address.name /= Void and then not a_address.name.is_empty
		ensure
			definition: Result = (a_address.name /= Void and then not a_address.name.is_empty)
		end

	is_good_address (a_address: ROSE_MAIL_ADDRESS): BOOLEAN is
			-- Does `a_address' contain a good address?
		require
			a_address_not_void: a_address /= Void
		do
			Result := a_address.address /= Void and then not a_address.address.is_empty
		ensure
			definition: Result = (a_address.address /= Void and then not a_address.address.is_empty)
		end

	has_data (a_row: RP_GRID_NODE [DS_PAIR [INTEGER, INTEGER]]): BOOLEAN is
			-- Does `a_row' have in-range data?
		require
			a_row_not_void: a_row /= Void
			mail_users_not_void: mail_users /= void
		do
			Result := a_row.data /= Void
		end

	add_grid_row (a_index: INTEGER) is
			-- Add a row to end of `recipients_grid' with group number `a_index' as data.
		require
			mail_users_not_void: mail_users /= void
			a_index_valid: mail_users.is_valid_group_index (a_index)
		local
			l_node: RP_GRID_NODE [DS_PAIR [INTEGER, INTEGER]]
			l_data: DS_PAIR [INTEGER, INTEGER]
		do
			create l_data.make (a_index, -1)
			create l_node.make_from_agents (agent displayed_cell (a_index, ?), agent expanded_group (a_index), 2)
			l_node.set_data (l_data)
			recipients_grid.extend (l_node)
		ensure
			one_more_row: recipients_grid.row_count = old recipients_grid.row_count + 1
			row_has_data: recipients_grid.last_row.data.first = a_index
			group_row: recipients_grid.last_row.data.second = -1
		end

	add_employee_row (a_index: INTEGER) is
			-- Add a row to end of `recipients_grid' with user number `a_index' as data.
		require
			mail_users_not_void: mail_users /= void
			a_index_valid: mail_users.is_valid_user_index (mail_users.Axa_rosenberg_group, a_index)
		local
			l_node: RP_GRID_NODE [DS_PAIR [INTEGER, INTEGER]]
			l_data: DS_PAIR [INTEGER, INTEGER]
		do
			create l_data.make (mail_users.Axa_rosenberg_group, a_index)
			create l_node.make_from_agents (agent displayed_employee_cell (a_index, ?), Void, 2)
			l_node.set_data (l_data)
			recipients_grid.extend (l_node)
		ensure
			one_more_row: recipients_grid.row_count = old recipients_grid.row_count + 1
			employee_row: recipients_grid.last_row.data.first = mail_users.Axa_rosenberg_group
			row_has_data: recipients_grid.last_row.data.second = a_index
		end

	add_user_row (a_index: INTEGER) is
			-- Add a row to end of `recipients_grid' with user number `a_index' as data.
		require
			mail_users_not_void: mail_users /= void
			a_index_valid: mail_users.is_valid_user_index (mail_users.All_users_group, a_index)
		local
			l_node: RP_GRID_NODE [DS_PAIR [INTEGER, INTEGER]]
			l_data: DS_PAIR [INTEGER, INTEGER]
		do
			create l_data.make (mail_users.All_users_group, a_index)
			create l_node.make_from_agents (agent displayed_all_users_cell (a_index, ?), Void, 2)
			l_node.set_data (l_data)
			recipients_grid.extend (l_node)
		ensure
			one_more_row: recipients_grid.row_count = old recipients_grid.row_count + 1
			employee_row: recipients_grid.last_row.data.first = mail_users.All_users_group
			row_has_data: recipients_grid.last_row.data.second = a_index
		end

	displayed_cell (a_group, a_column: INTEGER): ANY is
			-- Grid cell for `a_group' in `a_column'
		require
			mail_users_not_void: mail_users /= void
			a_group_valid: mail_users.is_valid_group_index (a_group)
			a_column_valid: a_column = 1 or a_column  = 2
		local
			l_button: EV_CHECK_BUTTON
			l_item: RP_DATA_GRID_WIDGET_ITEM
		do
			inspect
				a_column
			when 2 then
				create l_button
				create l_item.make (l_button)
				l_item.set_widget_height_dominant (True)
				l_item.set_widget_width_dominant (True)
				Result := l_item
			when 1 then
				Result := mail_users.group_name (a_group)
			end
		ensure
			displayed_cell_not_void: Result /= Void
		end

	displayed_employee_cell (a_employee, a_column: INTEGER): ANY is
			-- Grid cell for `a_employee' in `a_column'
		require
			mail_users_not_void: mail_users /= void
			a_employee_valid: mail_users.is_valid_user_index (mail_users.Axa_rosenberg_group, a_employee)
			a_column_valid: a_column = 1 or a_column  = 2
		local
			l_button: EV_CHECK_BUTTON
			l_item: RP_DATA_GRID_WIDGET_ITEM
		do
			inspect
				a_column
			when 2 then
				create l_button
				create l_item.make (l_button)
				l_item.set_widget_height_dominant (True)
				l_item.set_widget_width_dominant (True)
				Result := l_item
			when 1 then
				Result := mail_users.employee_name (a_employee)
			end
		ensure
			displayed_employee_cell_not_void: Result /= Void
		end

	displayed_all_users_cell (a_user, a_column: INTEGER): ANY is
			-- Grid cell for `a_user' in `a_column'
		require
			mail_users_not_void: mail_users /= void
			a_user_valid: mail_users.is_valid_user_index (mail_users.All_users_group, a_user)
			a_column_valid: a_column = 1 or a_column  = 2
		local
			l_button: EV_CHECK_BUTTON
			l_item: RP_DATA_GRID_WIDGET_ITEM
		do
			inspect
				a_column
			when 2 then
				create l_button
				create l_item.make (l_button)
				l_item.set_widget_height_dominant (True)
				l_item.set_widget_width_dominant (True)
				Result := l_item
			when 1 then
				Result := mail_users.user_name (mail_users.All_users_group, a_user)
			end
		ensure
			displayed_all_users_cell_not_void: Result /= Void
		end

	add_selected_address (a_recipients: ROSE_MAIL_RECIPIENTS; a_row: RP_GRID_NODE [DS_PAIR [INTEGER, INTEGER]]) is
			-- Add mail address for `a_row.data' to `a_recipients' iff it is selected.
		require
			a_recipients_not_void: a_recipients /= Void
			a_row_not_void: a_row /= Void
			mail_users_not_void: mail_users /= void
			valid_data: a_row.data.first = mail_users.Axa_rosenberg_group or
				a_row.data.first = mail_users.All_users_group or
				mail_users.is_valid_group_index (a_row.data.first)
			valid_user_data: a_row.data.second /= - 1 implies mail_users.is_valid_user_index (a_row.data.first, a_row.data.second)
		local
			l_data: DS_PAIR [INTEGER, INTEGER]
		do
			if {l_widget_item: RP_DATA_GRID_WIDGET_ITEM} a_row.i_th (2) then
				if {l_check: EV_CHECK_BUTTON} l_widget_item.widget then
					if l_check.is_selected then
						l_data := a_row.data
						if l_data.second = -1 then
							a_recipients.add_address (mail_users.group_address (l_data.first))
						else
							a_recipients.add_address (mail_users.user_address (l_data.first, l_data.second))
						end
					end
				else
					check
						unexpected_grid_structure: False
					end
				end
			else
				check
					unexpected_grid_structure: False
				end
			end
		ensure
			zero_or_one_more: (a_recipients.count = old a_recipients.count) or (a_recipients.count = old a_recipients.count + 1)
		end

	display_recipients_list is
			-- Display `recipients_list'.
		do
			recipients_box.wipe_out
			recipients_box.extend (recipients_list)
		ensure
			recipients_box_has_recipients_list: recipients_box.has (recipients_list)
		end

	display_recipients_grid is
			-- Display `recipients_grid'.
		do
			if mail_users = Void then
				create mail_users.make
			end
			if all_groups_radio.is_selected then
				if not grid_is_groups then
					grid_is_groups := True
					grid_is_axa_rosenberg := False
					grid_is_users := False
					recipients_grid.clear
					Gui.start_busy ("Fetching email groups");
					(1 |..| mail_users.all_groups_count).do_all (agent add_grid_row)
					recipients_grid.header_node.show_filter (True)
					show_actions.extend (agent recipients_grid.resize_columns_to_content (True))
					Gui.end_busy
				end
			elseif all_rosenberg_radio.is_selected then
				if not grid_is_axa_rosenberg then
					grid_is_axa_rosenberg := True
					grid_is_groups := False
					grid_is_users := False
					recipients_grid.clear
					Gui.start_busy ("Fetching AXA Rosenberg employees");
					(1 |..| mail_users.axa_rosenberg_count).do_all (agent add_employee_row)
					recipients_grid.header_node.show_filter (True)
					show_actions.extend (agent recipients_grid.resize_columns_to_content (True))
					Gui.end_busy
				end
			else
				check
					all_users: all_users_radio.is_selected
					-- only other radio button firing this routine
				end
				if not grid_is_users then
					grid_is_axa_rosenberg := False
					grid_is_groups := False
					grid_is_users := True
					recipients_grid.clear
					Gui.start_busy ("Fetching all email addresses");
					(1 |..| mail_users.all_users_count).do_all (agent add_user_row)
					recipients_grid.header_node.show_filter (True)
					show_actions.extend (agent recipients_grid.resize_columns_to_content (True))
					Gui.end_busy
				end
			end
			recipients_box.wipe_out
			recipients_box.extend (recipients_grid)
		ensure
			recipients_box_has_recipients_grid: recipients_box.has (recipients_grid)
		end

	expanded_group (a_group: INTEGER): RP_TITLED_NODES [DS_PAIR [INTEGER, INTEGER]] is
			-- Generate sub-rows for all members of `a_group'.
		require
			mail_users_not_void: mail_users /= void
			a_group_valid: a_group = mail_users.Axa_rosenberg_group or a_group = mail_users.All_users_group or
				mail_users.is_valid_group_index (a_group)
		local
			i, l_count: INTEGER
			l_nodes: ARRAY [RP_GRID_NODE [DS_PAIR [INTEGER, INTEGER]]]
		do
			l_count := mail_users.user_count_for_group (a_group)
			if l_count > 0 then
				create l_nodes.make (1, l_count)
				from
					i := 1
				until
					i > l_count
				loop
					add_user_node (l_nodes, a_group, i)
					i := i + 1
				end
				create Result.make_from_array (Void, l_nodes)
			end
		end

	add_user_node (a_nodes: ARRAY [RP_GRID_NODE [DS_PAIR [INTEGER, INTEGER]]]; a_group, a_index: INTEGER) is
			-- Add user grid node to `a_nodes' at position `a_index'.
		require
			a_nodes_not_void: a_nodes /= Void
			a_index_large_enough: a_index >= a_nodes.lower
			a_index_small_enough: a_index <= a_nodes.upper
		local
			l_node: RP_GRID_NODE [DS_PAIR [INTEGER, INTEGER]]
		do
			create l_node.make_from_agents (agent displayed_user_cell (a_group, a_index, ?), Void, 2)
			l_node.set_data (create {DS_PAIR [INTEGER, INTEGER]}.make (a_group, a_index))
			a_nodes.put (l_node, a_index)
		ensure
			node_at_a_index: a_nodes [a_index] /= Void
		end

	displayed_user_cell (a_group, a_user, a_column: INTEGER): ANY is
			-- Grid cell for `a_group' in `a_column'
		require
			mail_users_not_void: mail_users /= void
			a_group_valid: mail_users.is_valid_group_index (a_group)
			a_user_valid: mail_users.is_valid_user_index (a_group, a_user)
			a_column_valid: a_column = 1 or a_column  = 2
		local
			l_button: EV_CHECK_BUTTON
			l_item: RP_DATA_GRID_WIDGET_ITEM
		do
			inspect
				a_column
			when 2 then
				create l_button
				create l_item.make (l_button)
				l_item.set_widget_height_dominant (True)
				l_item.set_widget_width_dominant (True)
				Result := l_item
			when 1 then
				Result := mail_users.user_name (a_group, a_user)
			end
		ensure
			displayed_cell_not_void: Result /= Void
		end

invariant

	subject_not_void: subject /= Void
	body_not_void: body /= Void
	recipients_box_not_void: recipients_box /= Void
	recipients_list_not_void: recipients_list /= Void
	recipients_grid_not_void: recipients_grid /= Void
	all_recipients_grid_rows_have_data: mail_users /= Void implies recipients_grid.for_all_rows (agent has_data)

end

