note

	description:
	"[
		Chain of entities (usually the result of transformations from one entity to another)
		along with a selected subset of attribute types (one select subset per supplier-category);
		Note that in this abstract class a cell is specified by an entity which may be represented as 
		multiple active identifiers in the display representation of `Current'
		This object represents a row in the parent matrix (ROSE_ENTITY_CHAINS).
	]"
	author: "Mark Howard"
	last_editor: "$Author: bn $"
	date: "$Date: 2011/10/26 $"
	revision: "$Revision: #13 $"
	path: "$File: //rose/source/kernel/entities/rose_entity_chain.e $"

class
	ROSE_ENTITY_CHAIN

inherit
	CONTEXT_ANCHOR

	ANY

create
	make

feature {NONE} -- Initialization

	make (a_parent : ROSE_ENTITY_CHAINS; a_entities : ARRAY [ENTITY]) is
			-- Add reference to `parent' and populate with entities
		require
			entities_not_void: a_entities /= Void
		do
			parent := a_parent
			create entities.make_from_array (a_entities)
		ensure
			parent_set: parent = a_parent
			entities_created: (entities /= Void) and then (entities.count = a_entities.count)
		end

feature -- Access

	parent : ROSE_ENTITY_CHAINS
			-- Parent array

	entities : ROSE_LINEAR_ARRAY [ENTITY]
			-- Entities

	key_entity_map : ROSE_LINEAR_ARRAY [INTEGER]
			-- Value: index in `entities'
			-- Index: index in `keys'

	description : ROSE_LINEAR_ARRAY[ROSE_ENTITY_DESCRIPTION]
			-- A comment?
			-- not used as of this writing

	half_twin : like Current is
			-- Basically a deep_twin without `parent'
		do
			Result := twin
			Result.set_entities (entities.twin)
			Result.set_key_entity_map (Void)
			if description /= Void then
				Result.set_description (description.twin)
			end
		ensure
			properly_copied: Result /= Current and Result.entities /= entities
		end

	truncated : ROSE_ENTITY_CHAIN is
			-- Like `Current' with last entity removed
		require
			entities_not_void : entities /= Void
		local
			l_chain : like entities
			i : INTEGER
		do
			create l_chain.make (1, entities.count - 1)
			from i := 1 until i > entities.count - 1 loop
				l_chain.put (entities.item (i), i)
				i := i + 1
			end
			create Result.make (parent, l_chain)
		ensure
			one_less_entity : entities.count - Result.entities.count = (1).min (entities.count)
		end

	keys : ROSE_LINEAR_ARRAY [STRING] is
			-- All active attribute values for all entities in `entities'
			-- `Current' as a displayed row
		local
			l_accessor : ROSE_ENTITY_DISPLAY
			i, l_index : INTEGER
			l_handler : like category_handler_type
		do
			create Result.make_empty
			if key_entity_map = Void then
				create key_entity_map.make_empty
			else
				key_entity_map.wipe_out
			end
			l_index := 1
			from entities.start until entities.off loop
				l_handler := entities.current_item.handler
				l_accessor := accessor.item ([l_handler])
				if l_accessor /= Void then
					Result.append_array (l_accessor.entity_strings (entities.current_item))
					from i := l_index until i > Result.count loop
						key_entity_map.force (entities.index, i)
						i := i + 1
					end
					l_index := i
				else
					Result.extend (entities.current_item.handle.out)
					key_entity_map.force (entities.index, l_index)
					l_index := l_index + 1
				end
				entities.forth
			end
		ensure
			result_not_void : Result /= Void
		end

	key (a_index : INTEGER; a_identifier : ENTITY_IDENTIFIER) : STRING is
			-- Value of `a_identifier' (if marked as active) for `a_index'th entity
		require
			a_identifier_not_void : a_identifier /= Void
		local
			l_accessor : ROSE_ENTITY_DISPLAY
			i : INTEGER
			l_identifier : ENTITY_IDENTIFIER
			l_entity : ENTITY
		do
			if entities.valid_index (a_index) then
				l_entity := entities.item (a_index)
				l_accessor := accessor.item ([l_entity.handler])
				from i := 1 until i > l_accessor.identifier_count or Result /= Void  loop
					l_identifier := l_accessor.identifier_if_active (i)
					if l_identifier = a_identifier then
						Result := l_identifier.dated_key (l_entity, l_accessor.restrict_date)
					end
					i := i + 1
				end
			end
		end

	accessor_at (i : INTEGER) : ROSE_ENTITY_DISPLAY is
			-- Convenience feature
		do
			if entities /= Void and then entities.valid_index (i) then
				Result := accessor.item ([entities.item (i).handler])
			end
		end

	target : ENTITY is
			-- With the assumption that `Current' is the result of a chain of transformations
			-- The last entity in the chain could be considered the target
		do
			Result := entities.last
		ensure
			last_item: Result = entities.last
		end

	titles : ROSE_LINEAR_ARRAY [SIMPLE_TITLE] is
			-- Multi-line titles
			-- Context
			-- Category
			-- Attribute Name
		local
			l_accessor : ROSE_ENTITY_DISPLAY
			l_array : ROSE_LINEAR_ARRAY[STRING]
			l_entity : ENTITY
			l_title : SIMPLE_TITLE
			l_handler : like category_handler_type
			l_tuple : TUPLE [like category_handler_type]
		do
			create Result.make_empty
			from entities.start until entities.off loop
				l_entity := entities.current_item
				l_handler ?= l_entity.handler
				l_tuple := [l_handler]
				l_accessor := accessor.item (l_tuple)
				if l_accessor /= Void then
					l_array := l_accessor.entity_string_titles
					from l_array.start until l_array.off loop
						create l_title.make_from_array (<<l_entity.context.name,l_entity.category_name,l_array.current_item>>)
						Result.extend(l_title)
						l_array.forth
					end
				end
				entities.forth
			end
		end

	accessor : FUNCTION [ANY, TUPLE [like category_handler_type], ROSE_ENTITY_DISPLAY] is
			-- Shared accessor
			-- Gives access to set of displayed attributes information (ROSE_ENTITY_DISPLAY) given an entity
		require
			parent_not_void : parent /= Void
		do
			Result := parent.display_accessor
		end

	uses_category (a_entity : ENTITY) : BOOLEAN is
			-- Does `Current' contain a "column" (category-supplier) which matches `a_entity'?
		require
			a_entity_not_void : a_entity /= Void
		local
			i : INTEGER
		do
			if entities /= Void then
				from i := 1 until i > entities.count or Result loop
					Result := a_entity.category_id = entities.item(i).category_id and a_entity.supplier_id = entities.item(i).supplier_id
					i := i + 1
				end
			end
		end

	pivot_index : INTEGER is
			-- Selected column index
		do
			Result := parent.pivot_column_index
		end

	pivot_entity : ENTITY is
			-- Entity at `pivot_index'
		require
			pivot_index_in_bounds : pivot_index > 0 and then pivot_index <= entities.count
		do
			if entities.valid_index (pivot_index) then
				Result := entities.item (pivot_index)
			end
		end

feature -- Basic Operations

	save (a_filename : STRING) is
			-- Do nothing
		require
			not_implemented: False
		do
		end

	set_description (a_description: like description) is
			-- Set `description' to `a_description'
		do
			description := a_description
		ensure
			description_set: description = a_description
		end

feature {ROSE_ENTITY_CHAIN} -- Implementation

	set_entities (a_entities : like entities) is
			-- Set `entities' to `a_entities'
		do
			entities := a_entities
		ensure
			entities_set: entities = a_entities
		end

	set_key_entity_map (a_map: like key_entity_map) is
			-- Set `key_entity_map' to `a_map'
		do
			key_entity_map := a_map
		ensure
			key_entity_map_set: key_entity_map = a_map
		end

feature {NONE} -- Implementation

	category_handler_type: CATEGORY_HANDLER [like context_anchor, ENTITY] is
		-- Dummy routine for type anchoring
			do
			end

end
