note
	description: "[
		A matrix with columns specified by CATEGORY_HANDLERs and rows built from chains of entities (row : ROSE_ENTITY_CHAIN)
		Each column is also associated with a display_handler (ROSE_ENTITY_DISPLAY) which specifies attributes selected for display;
		Note that more than one attribute may be displayed per column of `Current' ==> mapping of columns of `Current' to
		columns of a display of `Current' may be one to many.
	]"
	author: "Mark Howard"
	last_editor: "$Author: zs $"
	date: "$Date: 2011/07/11 $"
	revision: "$Revision: #23 $"
	path: "$File: //rose/source/kernel/entities/rose_entity_chains.e $"

class
	ROSE_ENTITY_CHAINS

inherit
	ROSE_LINEAR_ARRAY [ROSE_ENTITY_CHAIN]
		rename
			make as rose_linear_array_make
		redefine
			extend
		end

	CONTEXT_ANCHOR
		undefine
			is_equal
		end

create
	make,
	rose_linear_array_make,
	make_empty,
	make_from_array

feature {NONE} -- Initialization

	make (a_display_accessor : FUNCTION [ANY, TUPLE [like category_handler_type], ROSE_ENTITY_DISPLAY])
			-- Initialize
		do
			make_empty
			display_accessor := a_display_accessor
			pivot_column_index := 1
			create display_handlers.make_empty
		end

feature -- Access

	display_handlers : ROSE_LINEAR_ARRAY [ROSE_ENTITY_DISPLAY]
			-- Index = index_th column of `Current'

	active_category_handler : CATEGORY_HANDLER [SUPPLIER_CONTEXT, ENTITY]
			-- Supplier-category object for selected column
		do
			if display_handlers.valid_index (pivot_column_index) then
				Result := display_handlers.item (pivot_column_index).category_handler
			end
		end

	width : INTEGER
			-- Width of displayed version of `Current'
		do
			Result := active_identifiers.count
		end

	column_count : INTEGER
			-- Columns in `Current'
		do
			Result := display_handlers.count
		end

	pivot_entity : ENTITY
			-- Entity at cell: current row @ selected column
		local
			l_chain : ROSE_ENTITY_CHAIN
		do
			if not is_empty then
				l_chain := current_item
				Result := l_chain.pivot_entity
			end
		end

	active_key : STRING
			-- Value of `active_identifier' for `active_entity'
 		local
			l_chain : ROSE_ENTITY_CHAIN
		do
			if not is_empty then
				l_chain := current_item
				Result := l_chain.key (pivot_column_index, pivot_identifier)
			end
		end

	active_display_handler : ROSE_ENTITY_DISPLAY
			-- Display handler at selected (active) column
		do
			if display_handlers.valid_index (pivot_column_index) then
				Result := display_handlers.item (pivot_column_index)
			end
		end

	first_active_entity : ENTITY
			-- Entity at (`pivot_column_index', 1)
		do
			if not is_empty and then first.entities.valid_index (pivot_column_index) then
				Result := first.entities.item (pivot_column_index)
			end
		end

	identifier (a_index : INTEGER): ENTITY_IDENTIFIER
			-- `a_index'th active identifier
			-- Note that this does NOT necessarily correspond to `a_index'th column of `Current'
		require
			identifiers_not_void : not is_active_identifiers_void
			valid_index : is_valid_index_for_active_identifiers (a_index)
		do
			Result := active_identifiers.item (a_index)
		end

	related (a_relationship : RELATIONSHIP_HANDLER [ENTITY, ENTITY];
			a_related_versus_corelated : BOOLEAN;
			a_parent_column : INTEGER;
			a_start, a_end : ROSE_DATE;
			a_range : HASH_TABLE [ENTITY, ENTITY])
			: ROSE_ENTITY_CHAINS
			-- A copy of `Current' with applicable related entities extended
		require
			relationship_not_void : a_relationship /= Void
		local
			l_projection, l_related : ROSE_LINEAR_ARRAY [ENTITY]
			l_chain : ROSE_ENTITY_CHAIN
			i : INTEGER
			l_added : BOOLEAN
			l_entity : ENTITY
			l_applicable : BOOLEAN
			l_category_handler : CATEGORY_HANDLER [SUPPLIER_CONTEXT, ENTITY]
		do
			if a_related_versus_corelated then
				l_applicable := a_relationship.g_handler = active_display_handler.category_handler
			else
				l_applicable := a_relationship.h_handler = active_display_handler.category_handler
			end
			if l_applicable then
				create Result.make (display_accessor)
				Result.set_display_handlers (display_handlers.twin)
				if display_handlers.is_empty then
					if a_related_versus_corelated then
						l_category_handler := a_relationship.g_handler
					else
						l_category_handler := a_relationship.h_handler
					end
					Result.add_display_handler (display_accessor.item ([l_category_handler]))
				end
				if a_related_versus_corelated then
					l_category_handler := a_relationship.h_handler
				else
					l_category_handler := a_relationship.g_handler
				end
				Result.add_display_handler (display_accessor.item ([l_category_handler]))
				if not is_empty then
					create l_related.make_empty
					from start until off loop
						l_chain := current_item
						l_entity := l_chain.entities.item (a_parent_column)
						l_added := False
						if a_related_versus_corelated then
							l_projection := l_entity.related_entities_during (a_relationship.name, a_start, a_end)
							l_related.append_array (l_projection)
						else
							l_projection := l_entity.corelated_entities_during (a_relationship.name, a_start, a_end)
							l_related.append_array (l_projection)
						end
						from i := 1 until i > l_projection.count loop
							if a_range = Void or else a_range.has (l_projection.item (i)) then
								Result.extend_with_entity (l_chain.entities, l_projection.item (i))
								l_added := True
							end
							i := i + 1
						end
						if not l_added then
							if a_related_versus_corelated then
								Result.extend_with_entity (l_chain.entities, a_relationship.ur_h)
							else
								Result.extend_with_entity (l_chain.entities, a_relationship.ur_g)
							end
						end
						forth
					end
				end
				Result.initialize_identifiers
			end
		end

	peers (a_context : SUPPLIER_CONTEXT; a_parent : INTEGER; a_start, a_end : ROSE_DATE; a_range : HASH_TABLE [ENTITY, ENTITY]) : ROSE_ENTITY_CHAINS
			-- A copy of `Current' with applicable peer entities extended
		local
			l_entities : ROSE_LINEAR_ARRAY [ENTITY]
			l_entity: ENTITY
			i, j : INTEGER
			l_chain : ROSE_ENTITY_CHAIN
		do
			create Result.make (display_accessor)
			Result.set_display_handlers (display_handlers.twin)
			if display_handlers.is_empty then
				Result.add_display_handler (display_accessor.item ([active_category_handler]))
			end
			Result.add_display_handler (display_accessor.item ([active_category_handler.ur_entity.peer (a_context, a_end).handler]))
			from i := 1 until i > count loop
				l_chain := Current [i]
				l_entities := l_chain.entities [a_parent].peers (a_context, a_end)
				if l_entities /= Void and then not l_entities.is_empty then
					from j :=1 until j > l_entities.count loop
						l_entity := l_entities [j]
						j := j + 1
						if a_range = Void or else a_range.has (l_entity) then
							Result.extend_with_entity (l_chain.entities, l_entity)
						end
					end
				end
				i := i + 1
			end
			Result.initialize_identifiers
		end

	matching_attribute (a_handler : CATEGORY_HANDLER [SUPPLIER_CONTEXT, ENTITY];
					 	a_identifier : ENTITY_IDENTIFIER;
					 	a_targethandler : CATEGORY_HANDLER [SUPPLIER_CONTEXT, ENTITY];
					 	a_targetidentifier : ENTITY_IDENTIFIER;
					 	a_start, a_end : ROSE_DATE;
					 	a_range : HASH_TABLE [ENTITY, ENTITY])
					 	: ROSE_ENTITY_CHAINS
			-- A copy of `Current' with matching entities added
			-- Entities of `active_column_index' using `active_identifier' are matched by value to entities accessed from `a_handler' and `a_identifier'
		require
			pivot_entity_not_void : pivot_entity /= Void
			a_handler_not_void : a_handler /= Void
			a_identifier_not_void : a_identifier /= Void
			a_targethandler_not_void : a_targethandler /= Void
			a_targetidentifier_not_void : a_targetidentifier /= Void
			a_start_not_void : a_start /= Void
			a_end_not_void : a_end /= Void
		local
			l_key : STRING
			l_entity_history : ROSE_ENTITY_HISTORY [ENTITY]
			l_interval : ENTITY_INTERVAL [ENTITY]
		do
			create Result.make (display_accessor)
			Result.set_display_handlers (display_handlers.twin)
			if display_handlers.is_empty then
				Result.add_display_handler (display_accessor.item ([active_category_handler]))
			end
			Result.add_display_handler (display_accessor.item ([a_handler]))
			from start until off loop
				l_key := a_identifier.dated_key (pivot_entity, a_end)
				if l_key /= Void then
					l_entity_history := a_targethandler.entity_history_from_identifier (l_key, a_targetidentifier)
					if l_entity_history = Void or else l_entity_history.is_empty then
						Result.extend_with_entity (current_item.entities, a_handler.ur_entity)
					else
						from l_entity_history.start until l_entity_history.off loop
							l_interval := l_entity_history.current_item
							if l_interval.meets (a_start, a_end) and then (a_range = Void or else a_range.has (l_interval.entity)) then
								Result.extend_with_entity (current_item.entities, l_interval.entity)
							end
							l_entity_history.forth
						end
					end
				else
					Result.extend_with_entity (current_item.entities, a_handler.ur_entity)
				end
				forth
			end
		end

	pivot_entities, active_entities : ROSE_LINEAR_ARRAY [ENTITY]
			-- Set of entities extracted from each row (ROSE_ENTITY_CHAIN) at `pivot_column_index'th column
		local
			i : INTEGER
			l_entities : ROSE_LINEAR_ARRAY [ENTITY]
			l_chain : ROSE_ENTITY_CHAIN
		do
			create Result.make_empty
			from i := 1 until i > count loop
				l_chain := item (i)
				l_entities := l_chain.entities
				if l_entities /= Void and then l_entities.valid_index (pivot_column_index) then
					Result.extend (l_entities.item (pivot_column_index))
				end
				i := i + 1
			end
		end

	category_handler_from_display_column (a_column : INTEGER) : CATEGORY_HANDLER [SUPPLIER_CONTEXT, ENTITY]
			-- `a_column' is column as displayed (could be multiple attribute columns per category)
		require
			a_column_large_enough : a_column > 0
		do
			Result := display_handlers [column_from_display_column (a_column)].category_handler
		end

	column_from_display_column (a_display_column : INTEGER) : INTEGER
			-- Column of `Current' corresponding to `a_displcay_column'
			-- Note: Column of `Current' is corresponds to the i_th display_handler
			-- While display column is caluclated from the sums of the active identifiers of the display_handlers
		local
			i, j : INTEGER
			l_display_handler : ROSE_ENTITY_DISPLAY
		do
			from i := 1 until i > display_handlers.count or Result > 0 loop
				l_display_handler := display_handlers [i]
				j := j + l_display_handler.active_identifier_count
				if j >= a_display_column then
					Result := i
				end
				i := i + 1
			end
		end

	identifier_from_display_column (a_display_column : INTEGER) : ENTITY_IDENTIFIER
			-- Entity_identifier at `a_display_column'
		require
			a_display_column_large_enough : a_display_column > 0
		do
			Result := active_identifiers [a_display_column]
		end

	pivot_column_index : INTEGER
			-- Column to be matched on

	pivot_identifier : ENTITY_IDENTIFIER
			-- Identifier to match on
			-- Note this only has meaning when associated with the display handler at `pivot_colmn_index'

feature -- Basic operations

	add_display_handler (a_display_handler : ROSE_ENTITY_DISPLAY)
			-- Extend `a_display_handler' and its active identifiers to `Current'
		do
			display_handlers.extend (a_display_handler)
			add_display_handler_identifiers (a_display_handler)
		end

	set_display_handlers (a_display_handlers : like display_handlers)
			-- Set `display_handlers' to `a_display_handlers'
		do
			display_handlers := a_display_handlers
		end

	set_pivots (a_index: INTEGER; a_identifier: ENTITY_IDENTIFIER)
			-- Column (category_handler) and identifier to match on
		require
			active_is_meaningful: is_active_meaningful (a_index, a_identifier)
		do
			pivot_column_index := a_index
			pivot_identifier := a_identifier
		end

	is_active_meaningful (a_index : INTEGER; a_identifier : ENTITY_IDENTIFIER) : BOOLEAN
			-- Does column specified by `a_index' and `a_identifier' potentially exist?
			-- "potentially" because `a_identifier' may be valid but not active
		require
			a_identfier_not_void : a_identifier /= Void
		local
			l_handler :CATEGORY_HANDLER [SUPPLIER_CONTEXT, ENTITY]
		do
			Result := display_handlers.valid_index (a_index)
			if Result and then a_identifier /= Void then
				l_handler := display_handlers [a_index].category_handler
				Result := l_handler.identifiers [a_identifier.handle] /= Void
			end
		end

	set_identifiers_from_recipe (a_recipe: ROSE_RECIPE)
			-- Set identifiers as active from information in recipe
		require
			recipe_not_void: a_recipe /= Void
			recipe_matches_handlers : True -- To be implented
		local
			i: INTEGER
			l_names: LINEAR [STRING]
			l_settings: ROSE_LINEAR_ARRAY [PAIR [STRING, ROSE_TRANSFORMATION_SETTINGS]]
			l_accessor: ROSE_ENTITY_DISPLAY
		do
			l_settings := a_recipe.tagged_settings
			from i := display_handlers.lower until i > display_handlers.upper loop
				l_accessor := display_handlers.item (i)
				l_accessor.inactivate_identifiers
				l_names := l_settings.item (i).second.display.identifier_names
				from l_names.start until l_names.off loop
					l_accessor.set_identifier_active (l_names.item, True)
					l_names.forth
				end
				i := i + 1
			end
			initialize_identifiers
		end

	initialize_identifiers
			-- Populate `active_identifiers' from current diplay handlers
		local
			i: INTEGER
			l_display_handler : ROSE_ENTITY_DISPLAY
		do
			create active_identifiers.make_empty
			from i := display_handlers.lower until i > display_handlers.upper loop
				l_display_handler := display_handlers.item (i)
				add_display_handler_identifiers (l_display_handler)
				i := i + 1
			end
		end

	extend_with_entity (a_entities : ARRAY [ENTITY]; a_entity: ENTITY)
			-- Create and extend into `Current' a new `ROSE_ENTITY_CHAIN' from `a_entities' (extended with `a_entity')
			-- and ensure `display_handlers' is extended if column count has been exceeded
			-- Why the strange arguments? This feature is usually used in the creation of an extended copy of a `ROSE_ENTITY_CHAINS',
			-- `a_entities' coming from a row of the copied `ROSE_ENTITY_CHAINS' extended by the new entity `a_entity'
		require
			new_entities_conform_to_current_display_handlers : is_entities_conform_to_handlers (a_entities)
		local
			l_array: ROSE_LINEAR_ARRAY [ENTITY]
			i: INTEGER
			l_new_count: INTEGER
			l_tuple : TUPLE [like category_handler_type]
		do
			if a_entities /= Void then
				create l_array.make_from_array (a_entities)
			else
				create l_array.make_empty
			end
			if a_entity /= Void then
				l_array.extend (a_entity)
			end
			l_new_count := l_array.count
			if l_new_count > display_handlers.count then
				from i := display_handlers.count + 1 until i > l_new_count loop
					create l_tuple
					l_tuple.put (l_array.item (i).handler, 1)
					add_display_handler (display_accessor.item (l_tuple))
					i := i + 1
				end
			end
			extend (create {ROSE_ENTITY_CHAIN}.make (Current, l_array))
		end

feature -- Assertion Support

	is_entities_conform_to_handlers (a_entities : ROSE_LINEAR_ARRAY [ENTITY]) : BOOLEAN
			-- Does i_th entity from `a_entities' conform to i_th display_handler for all entities in `a_entities' where i_th is in bounds
		require
			a_entities_not_void : a_entities /= Void
		local
			i : INTEGER
			l_entity : ENTITY
			l_fail : BOOLEAN
			l_handler : like category_handler_type
			l_min : INTEGER
		do
			l_min := a_entities.count.min (display_handlers.count)
			from i := 1 until i > l_min or l_fail loop
				l_entity := a_entities [i]
				l_handler := display_handlers [i].category_handler
				l_fail := not (l_entity.supplier_id = l_handler.context.supplier_id and then l_entity.category_id = l_handler.category_id)
				i := i + 1
			end
			Result := not l_fail
		end

	is_active_identifiers_void : BOOLEAN
			-- Is `identifiers' Void?
		do
			Result := active_identifiers = Void
		end

	is_valid_index_for_active_identifiers (a_index : INTEGER) : BOOLEAN
			-- Is `a_index' a valid index for `identifiers'?
		do
			Result := active_identifiers.valid_index (a_index)
		end

feature {ROSE_ENTITY_CHAINS, ROSE_ENTITY_CHAIN, ENTITY_BROWSER_COMMON}-- Implementation

	active_identifiers : ROSE_LINEAR_ARRAY [ENTITY_IDENTIFIER]
			-- All currently active identifiers

	extend (a_chain : ROSE_ENTITY_CHAIN)
			-- Hidden from user as extended `a_chain' must match current columns and `Current' must accomodate for any additional columns in `a_chain'
		do
			Precursor (a_chain)
		end

	add_display_handler_identifiers (a_display_handler : ROSE_ENTITY_DISPLAY)
			-- Add active identifiers from `a_display_handler' to `identifiers'
			-- If`a_display_handler' has no active identifiers then add `Void' as a placeholder
		require
			a_display_handler_not_void : a_display_handler /= Void
		local
			j : INTEGER
			l_identifier : ENTITY_IDENTIFIER
			l_active : BOOLEAN
		do
			if active_identifiers = Void then
				create active_identifiers.make_empty
			end
			if a_display_handler.is_display_handle then
				active_identifiers.extend (Void)
				l_active := True
			end
			from j := 1 until j > a_display_handler.identifier_count loop
				l_identifier := a_display_handler.identifier_if_active (j)
				if l_identifier /= Void then
					active_identifiers.extend (l_identifier)
					l_active := True
				end
				j := j + 1
			end
			if not l_active then
				active_identifiers.extend (Void)
			end
		end

feature {ROSE_ENTITY_CHAIN, ROSE_TRANSFORMATION_LIST} -- FIXME: should not be exported to ROSE_TRANSFORMATION_LIST

	display_accessor : FUNCTION [ANY, TUPLE [like category_handler_type], ROSE_ENTITY_DISPLAY]
			-- Gives access to display_handler given a category hanlder

feature {NONE} -- Types

	category_handler_type : CATEGORY_HANDLER [like context_anchor, ENTITY]
			-- Category handler type

end
