note
	description: "Error handler"
	author: "Zoran Simic"
	last_editor: "$Author: zs $"
	date: "$Date: 2011/05/13 $"
	revision: "$Revision: #20 $"
	path: "$File: //rose/source/kernel/infrastructure/framework/error/rose_error_handler.e $"

class
	ROSE_ERROR_HANDLER

inherit
	ROSE_IO_CONVENTION

create
	make

feature {NONE} -- Initialization

	make (a_io_handler: attached ROSE_IO_HANDLER)
			-- Create an error handler.
		do
			io_handler := a_io_handler
			create message_buffer.make
			message_buffer.open (Void)
			create exception_translator
		end

feature -- Access

	io_handler: attached ROSE_IO_HANDLER
			-- IO handler associated with this error handler

	message_buffer: attached ROSE_OUTCHANNEL_BUFFER
			-- Error and warning messages buffered

	broadcast_channel: ROSE_OUTCHANNEL_BROADCAST
			-- Channels to which all error messages are broadcast

	exception_translator: attached ROSE_EXCEPTION_TRANSLATOR
			-- Translation of Eiffel exceptions, in order to be more meaningful to human reading

	error_count: INTEGER
			-- Number of times an error was signaled during whole execution (not reset by an acknowledgement)

	warning_count: INTEGER
			-- Number of times an error was signaled during whole execution (not reset by an acknowledgement)

	last_exception_explanation: attached STRING
			-- Info on last exception
		do
			Result := exception_translator.last_exception_explanation (True)
		end

	flush_to_string: STRING
			-- Flush and return warnings/errors reported
		do
			Result := message_buffer.flush_to_string
		end

	exceptions: attached EXCEPTIONS
			-- Eiffel exceptions
		once
			create Result
		end

	signaled: BOOLEAN
			-- Is any error currently signaled ?

	is_flushed: BOOLEAN
			-- Are all messages flushed ?
		do
			Result := message_buffer.is_empty
		end

feature -- Status setting

	set_signaled (a_value: BOOLEAN)
			-- Set 'signaled'
		do
			signaled := a_value
		end

	set_exception_translator (a_translator: attached ROSE_EXCEPTION_TRANSLATOR)
			-- Set exception translator.
		do
			if private_stack = Void then
				exception_translator := a_translator
			end
		end

	set_io_handler (a_io_handler: attached ROSE_IO_HANDLER)
			-- Set 'io_handler'
		do
			if private_stack = Void then
				io_handler := a_io_handler
			end
		end

feature -- Basic operations

	raise (a_message: STRING)
			-- Raise a developper exception with 'a_message'
		do
			exceptions.raise (a_message)
		end

	broadcast (a_level: INTEGER; a_message: STRING; a_args: ARRAY [ANY])
			-- Broadcast 'a_message' formatted with 'a_args' to attached broadcast channels
		do
			if a_message /= Void and then broadcast_channel /= Void then
				if private_stack = Void then
					broadcast_channel.output_message (a_level,formatter.argumented (a_message,a_args))
				elseif private_stack.item /= Void then
					private_stack.item.output_message (a_level,formatter.argumented (a_message,a_args))
				end
			end
		end

	trace (a_message: STRING; a_args: ARRAY [ANY])
			-- Trace 'a_message' formatted with 'a_args' on the error channel.
		do
			effective_output (v_trace,a_message,a_args)
		end

	output (a_message: STRING; a_args: ARRAY [ANY])
			-- Output 'a_message' formatted with 'a_args' on the error channel.
		do
			effective_output (v_output,a_message,a_args)
		end

	info (a_message: STRING; a_args: ARRAY [ANY])
			-- Emit 'a_message' formatted with 'a_args' on the error channel with newline.
		do
			effective_output (v_trace,a_message + s_newline,a_args)
		end

	warn (a_message: STRING; a_args: ARRAY [ANY])
			-- Emit warning 'a_message' formatted with 'a_args'.
		require
			valid_message: a_message /= Void
		do
			if a_message /= Void then
				if private_stack = Void then
					warning_count := warning_count + 1
				end
				effective_output (v_warning,a_message + s_newline,a_args)
			end
		end

	signal (a_from, a_message: STRING; a_args: ARRAY [ANY])
			-- Signal error 'a_message' formatted with 'a_args',
			-- and eventually marked as coming from 'a_from'.
		require
			valid_from: a_from /= Void
			valid_message: a_message /= Void
		do
			signaled := True
			if a_from /= Void and then a_message /= Void then
				if private_stack = Void then
					error_count := error_count + 1
				end
				effective_output (v_error, a_from + s_from_separator + a_message + s_newline, a_args)
			end
		end

	abort (a_from,a_message: STRING; a_args: ARRAY [ANY])
			-- Signal error 'a_message' formatted with 'a_args',
			-- and eventually marked as coming from 'a_from'.
			-- Abort execution (raise an exception).
			-- This will terminate current call up-to the next rescue clause.
		do
			force_handle
			if a_message /= Void then
				if a_from = Void then
					warn (a_message,a_args)
				else
					signal (a_from,a_message,a_args)
				end
				exceptions.raise (formatter.argumented (a_message, a_args))
			else
				exceptions.raise ("Program explicitly aborted")
			end
		end

	die (a_exit_code: INTEGER; a_from,a_message: STRING; a_args: ARRAY [ANY])
			-- Terminate execution immediately, with an optional error message.
		local
			rescued: BOOLEAN
		do
			force_handle
			if rescued then
					--| FIXME: this shouldn't never happen! if it does, it means that
					--| the framework code is not rock-solid: fix IO/error-handler code immediately !
				exceptions.die (-10)
			elseif a_message /= Void then
				if a_from = Void then
					warn (a_message,a_args)
				else
					signal (a_from,a_message,a_args)
				end
				flush_and_acknowledge
			end
			io_handler.close_all
			exceptions.die (a_exit_code)
		rescue
			if not rescued then
				rescued := True
				retry
			end
		end

	push_private (a_channel: ROSE_OUTCHANNEL)
			-- Stop handling errors, all calls of lower importance than signal() are ignored.
			-- 'signaled' is not set, the error is simply directly output to 'a_channel' if provided.
			-- Works as a stack, a corresponding number of calls to push_private() and pop_private must be issued
			-- A signal of higher importance than signal() reestablishes the handling automatically.
		do
			if private_stack = Void then
				saved_signaled := signaled
				create private_stack.make
			end
			private_stack.extend (a_channel)
		end

	pop_private
			-- Pop current private error channel
		do
			if private_stack /= Void then
				if not private_stack.is_empty then
					private_stack.remove
				end
				if private_stack.is_empty then
					signaled := saved_signaled
					private_stack := Void
				end
			end
		end

	flush
			-- Flush warnings/errors reported.
		do
			if private_stack = Void and then Io_handler.has_output or else Io_handler.has_log then
				message_buffer.flush_on (
					create {ROSE_OUTCHANNEL_BROADCAST}.make (
						<<	Io_handler.output,
							Io_handler.echo,
							Io_handler.log,
							broadcast_channel
						>>)
				)
			end
		end

	flush_and_acknowledge
			-- Flush warnings/errors reported and acknowledge error status
		do
			flush
			acknowledge
		end

	acknowledge
			-- Acknowledge error status: mark no error signaled
		do
			signaled := False
		end

	disregard
			-- Disregard all error and warning messages.
			-- Note: broadcasted messages are not disregarded since already sent
		do
			if private_stack = Void then
					--| Clear the buffer and mark no errors signaled
				message_buffer.open (Void)
			end
			acknowledge
		end

	reset
			-- Reset error and warning count.
		do
			if private_stack = Void then
				error_count := 0
				warning_count := 0
			end
		end

	add_broadcast (a_channel: ROSE_OUTCHANNEL)
			-- Add 'a_channel' to attached broadcast.
		do
			if private_stack = Void and then a_channel /= Void then
				if broadcast_channel = Void then
					create broadcast_channel.make_empty
				end
				broadcast_channel.add_channel (a_channel)
			end
		end

	remove_broadcast (a_channel: ROSE_OUTCHANNEL)
			-- Remove broadcast channel 'a_channel'
		do
			if private_stack = Void and then broadcast_channel /= Void then
				broadcast_channel.remove_channel (a_channel)
				if broadcast_channel.is_empty then
					broadcast_channel := Void
				end
			end
		end

feature {NONE} -- Implementation

	private_stack: LINKED_STACK [ROSE_OUTCHANNEL]
			-- Stack of channels used to override usual error signaling

	effective_output (a_level: INTEGER; a_message: STRING; a_args: ARRAY [ANY])
			-- Effectively perform output of 'a_message' argumented with 'a_args' and with 'a_level'.
		local
			l_message: STRING
		do
			if a_message /= Void then
				l_message := formatter.argumented (a_message, a_args)
				if private_stack = Void then
					message_buffer.output_message (a_level, l_message)
				elseif private_stack.item /= Void then
					private_stack.item.output_message (a_level, l_message)
				end
			end
		end

	force_handle
			-- Force handling of errors
		do
			if private_stack /= Void then
				signaled := saved_signaled
				private_stack := Void
			end
		end

	s_from_separator: STRING = ": "		-- Separator between error message and where it is originating from
	s_newline: STRING = "%N"			-- Newline

	saved_signaled: BOOLEAN
			-- Error handling status

end
