note

	description:

		"EiffelVision control (menus/toolbars) factory, producing objects with most commonly used settings"

	author: "Zoran Simic"
	last_editor: "$Author: c4 $"
	date: "$Date: 2014/07/16 $"
	revision: "$Revision: #59 $"
	path: "$File: //rose/source/kernel/infrastructure/framework/gui/rose_ev_factory_control.e $"

class
	ROSE_EV_FACTORY_CONTROL

inherit
	
	ROSE_SHARED_GLOBAL_GUI
	
	ELOG_SHARED_LOGGER
		export {NONE} all end
	
	ROSE_IO_VERBOSITY
		export {NONE} all end

feature -- Access

	item_by_name (a_list: EV_ANY; a_name: STRING): EV_ANY is
			-- Item with 'a_name' from 'a_list'
			-- 'a_list' can be an EV_MENU or EV_TOOL_BAR for example
			-- 'a_name' will be matched against .text and .tooltip on the items of 'a_list' (recursively)
			-- First item matching is returned
		do
			Result := item_by_name_recursive (a_list,a_name,-1)
		end

	item_by_name_recursive (a_list: EV_ANY; a_name: STRING; a_depth: INTEGER): EV_ANY is
			-- Returns first item with 'a_name' in 'a_list' (same as 'item_by_name' with an infinite 'a_depth')
			-- 'a_depth' indicates how deep 'a_list' is recursively searched
		local
			l_cell: CELL [EV_ANY]
		do
			if a_name /= Void and then not a_name.is_empty then
				create l_cell.put (Void)
				get_first_item_named (a_list, a_name, l_cell, a_depth)
				Result := l_cell.item
			end
		end

	get_toolbar_button (a_toolbar: EV_TOOL_BAR; a_name: STRING): EV_TOOL_BAR_BUTTON is
			-- Return first item with 'a_name' in 'a_toolbar'
			-- 'a_name' will be matched against .text and .tooltip on the items of 'a_toolbar'
		do
			Result ?= item_by_name (a_toolbar,a_name)
		end

	is_selected_name (a_list: EV_ANY; a_name: STRING): BOOLEAN is
			-- Is an item which correspond to 'a_name' in 'a_list' selected? (.is_selected is True)
		local
			l_selectable: EV_SELECTABLE
		do
			l_selectable ?= item_by_name (a_list,a_name)
			Result := l_selectable /= Void and then l_selectable.is_selected
		end

	is_enabled_name (a_list: EV_ANY; a_name: STRING): BOOLEAN is
			-- Is an item which correspond to 'a_name' in 'a_list' enabled? (.is_sensitive is True)
		local
			l_sensitive: EV_SENSITIVE
		do
			l_sensitive ?= item_by_name (a_list,a_name)
			Result := l_sensitive /= Void and then l_sensitive.is_sensitive
		end

	new_menu_bar (a_window: EV_WINDOW; a_description: ARRAY [TUPLE [STRING,TUPLE]]): EV_MENU_BAR is
			-- Return a menu bar based on 'a_description', see 'new_menu' for TUPLE format
		local
			l_menu: EV_MENU
			l_had_bar: BOOLEAN
			i: INTEGER
		do
			create Result
			if a_description /= Void then
				from
					i := a_description.lower
				until
					i > a_description.upper
				loop
					l_menu := new_menu (a_description.item (i))
					if l_menu /= Void then
						Result.extend (l_menu)
					end
					i := i + 1
				end
			end
			if a_window /= Void then
				l_had_bar := a_window.menu_bar /= Void
				a_window.set_menu_bar (Result)
				if l_had_bar then
					-- Add a horizontal separator only if there was no previous menu bar in 'a_window'
					-- (otherwise, we can end up adding several separators...)
					a_window.upper_bar.extend (create {EV_HORIZONTAL_SEPARATOR})
				end
			end
		end

	new_popup_menu_button (a_container: EV_CONTAINER; a_label: STRING; a_activator: PROCEDURE [ANY,TUPLE [EV_MENU]]; a_width: INTEGER; a_description: TUPLE [STRING,TUPLE]): EV_BUTTON is
			-- New button that will popup a menu defined in 'a_description'.
			-- 'a_activator' is called before the menu is shown
			-- Returned button is put in 'a_container', is labeled with 'a_label' and obeys to 'a_width'
		local
			l_label: STRING
			l_direction: CHARACTER
		do
			create Result
			if a_label /= Void and then not a_label.is_empty then
				l_direction := a_label.item (1)
				if l_direction = '<' or l_direction = '>' or l_direction = '^' or l_direction = 'v' then
					l_label := a_label.twin
					l_label.remove (1)
					Result.set_pixmap (Gui.factory.new_arrow_pixmap (l_direction, 14, 14, Result.background_color, Gui.rep.colors.emphasize (100, Result.foreground_color)))
				else
					l_label := a_label
				end
				Result.set_text (l_label)
			end
			if a_description /= Void then
				Result.select_actions.extend (agent (Gui.dialogs).popup_menu (a_activator,a_description))
			else
				Result.disable_sensitive
			end
			Gui.factory.enbox_and_size_simple (Result, a_container, a_width)
		end

	new_menu (a_description: TUPLE [STRING,TUPLE]): EV_MENU is
			-- Return a menu based on 'a_description'
			-- Valid descriptions are: ["name",[menu item],[menu item],...]
			-- See 'add_menu_item' for each [menu item] format
		local
			l_string: STRING
			l_tuple: TUPLE
			l_menu: EV_MENU
			i: INTEGER
		do
			if a_description /= Void and then a_description.count >= 1 then
				i := a_description.lower
				l_string ?= a_description.item (i)
				if l_string /= Void then
					from
						create Result.make_with_text (l_string + " ")
						i := i + 1
					until
						i > a_description.upper
					loop
						l_tuple ?= a_description.item (i)
						l_menu ?= a_description.item (i)
						if l_menu /= Void then
							Result.extend (l_menu)
						else
							add_menu_item (Result, l_tuple, True)
						end
						i := i + 1
					end
				end
			end
		end

	new_menu_item (a_menu: EV_MENU; a_description: TUPLE; a_enabled: BOOLEAN): EV_MENU_ITEM is
			-- Add a menu item to 'a_menu' accordingly to 'a_description'
			-- 'a_description' can be of the forms:
			-- Void or [] : separator
			-- ["__documentation"]: Special menu description adding a menu entry linking to existing app documentation
			-- ["text"] : inactive menu entry with "text" as text
			-- [EV_MENU_ITEM] : the menu item itself
			-- [BOOLEAN,<item descr>] : check-menu with item description
			-- [INTEGER,<item descr>] : radio-menu with item description (selected if INTEGER /= 0)
			-- [<item descr>] : simple menu entry with <item descr>
			-- See 'decorate_menu_item' for <item descr> format
		local
			i: INTEGER
			l_check: EV_CHECK_MENU_ITEM
			l_radio: EV_RADIO_MENU_ITEM
			l_parent: EV_ITEM_LIST [EV_ITEM]
			l_string: STRING
			l_tuple: TUPLE
			l_submenu: TUPLE [STRING, TUPLE]
		do
			if a_description = Void or else a_description.count = 0 then
				-- Separator
				create {EV_MENU_SEPARATOR} Result
			elseif a_description.count = 1 then
				-- One item only: recognize STRING, EV_MENU_ITEM
				i := a_description.lower
				l_string ?= a_description.item (i)
				Result ?= a_description.item (i)
				if l_string /= Void then
					-- Disabled menu entry with the STRING as text
					create Result.make_with_text (l_string + " ")
					Result.disable_sensitive
				elseif Result /= Void then
					-- Add the EV_MENU_ITEM found itself
					l_parent := Result.parent
					if l_parent /= Void then
						l_parent.prune_all (Result)
					end
				else
					-- Do nothing
				end
			else
				-- More than one item in description: menu item, or sub menu
				i := a_description.lower
				l_string ?= a_description.item (i)
				if a_description.is_boolean_item (i) then
					-- Check-menu item
					create l_check
					if a_description.boolean_item (i) then
						l_check.enable_select
					end
					if decorate_menu_item (l_check, a_description, i + 1) then
						Result := l_check
					end
				elseif a_description.is_integer_item (i) then
					-- Radio-menu item
					create l_radio
					if a_description.integer_item (i) /= 0 then
						l_radio.enable_select
					end
					if decorate_menu_item (l_radio, a_description, i + 1) then
						Result := l_radio
					end
				elseif l_string /= Void then
					l_tuple ?= a_description.item (i + 1)
					if l_tuple = Void then
						-- Simple menu entry
						create Result
						if not decorate_menu_item (Result,a_description,i) then
							Result := Void
						end
					else
						-- Sub menu
						l_submenu ?= a_description
						if l_submenu /= Void then
							Result := new_menu (l_submenu)
						end
					end
				end
			end
			if Result /= Void and then a_menu /= Void then
				a_menu.extend (Result)
			end
			if Result /= Void and then not a_enabled then
				Result.disable_sensitive
			end
		end

	get_sub_menu (a_target: EV_MENU_ITEM_LIST; a_sub_name: STRING): EV_MENU is
			-- Get submenu with 'a_sub_name' from 'a_target', Void if none
		do
			Result ?= item_by_name_recursive (a_target,a_sub_name,1)
		end

	new_toolbar (a_container: EV_CONTAINER; a_description: ARRAY [TUPLE]): EV_TOOL_BAR is
			-- Return a toolbar with items from 'a_description'
			-- See 'new_toolbar_item' for format of each tuple in 'a_description'
		local
			i: INTEGER
			l_window: EV_WINDOW
			l_box: EV_BOX
		do
			create Result
			Result.disable_vertical_button_style
			if a_description /= Void then
				from
					i := a_description.lower
				until
					i > a_description.upper
				loop
					add_toolbar_item (Result, a_description.item (i), True)
					i := i + 1
				end
			end
			if a_container /= Void then
				l_window ?= a_container
				if l_window /= Void then
					i := 0
					l_window.upper_bar.extend (Result)
					l_window.upper_bar.extend (create {EV_HORIZONTAL_SEPARATOR})
				else
					a_container.extend (Result)
					l_box ?= a_container
					if l_box /= Void then
						l_box.disable_item_expand (Result)
					end
				end
			end
		ensure
			valid_result: Result /= Void
		end

	new_toolbar_item (a_toolbar: EV_TOOL_BAR; a_description: TUPLE; a_enabled: BOOLEAN): EV_TOOL_BAR_ITEM is
			-- 'a_description' is interpreted as follows
			-- Void or [] : separator
			-- otherwise, see 'new_toolbar_button'
		do
			if a_description = Void or else a_description.count = 0 then
				-- Separator
				create {EV_TOOL_BAR_SEPARATOR} Result
				a_toolbar.extend (Result)
			else
				Result := new_toolbar_button (a_toolbar, a_description, a_enabled)
			end
		end

	new_toolbar_button (a_toolbar: EV_TOOL_BAR; a_description: TUPLE; a_enabled: BOOLEAN): EV_TOOL_BAR_BUTTON is
			-- 'a_description' is interpreted as follows
			-- [<pixmap>,<action>,"tooltip","text"]
			-- [BOOLEAN,<pixmap>,<action>,"tooltip","text"]: check button
			-- [INTEGER,<pixmap>,<action>,"tooltip","text"]: radio button
			-- Created toolbar is added to 'a_toolbar' when provided
			-- <pixmap> may be either a STRING (name of icon file) or an EV_PIXMAP
			-- <action> may be either an agent or a STRING (interpreted as a framework command then)
		require
			description_ok: a_description /= Void and then a_description.count > 0
		do
			if a_description.is_integer_item (a_description.lower) then
				-- Radio button
				Result := new_toolbar_radio_button (a_toolbar, a_description, a_enabled)
			elseif a_description.is_boolean_item (a_description.lower) then
				-- Toggle button
				Result := new_toolbar_toggle_button (a_toolbar, a_description, a_enabled)
			else
				-- Push button
				create Result
				decorate_toolbar_item (Result, a_description)
				if not a_enabled then
					Result.disable_sensitive
				end
				if a_toolbar /= Void then a_toolbar.extend (Result) end
			end
		end

	new_toolbar_radio_button (a_toolbar: EV_TOOL_BAR; a_description: TUPLE; a_enabled: BOOLEAN): EV_TOOL_BAR_RADIO_BUTTON is
			-- 'a_description' is interpreted as follows
			-- [<pixmap>,<action>,"tooltip","text"]
			-- Created toolbar is added to 'a_toolbar' when provided
			-- <pixmap> may be either a STRING (name of icon file) or an EV_PIXMAP
			-- <action> may be either an agent or a STRING (interpreted as a framework command then)
		require
			description_ok: a_description /= Void and then a_description.count > 0
		do
			create Result
			if a_description.is_boolean_item (a_description.lower) then
				if a_description.boolean_item (a_description.lower) then
					Result.enable_select
				end
			elseif a_description.is_integer_item (a_description.lower) then
				if a_description.integer_item (a_description.lower) /= 0 then
					Result.enable_select
				end
			end
			decorate_toolbar_item (Result, a_description)
			if not a_enabled then
				Result.disable_sensitive
			end
			if a_toolbar /= Void then a_toolbar.extend (Result) end
		end

	new_toolbar_toggle_button (a_toolbar: EV_TOOL_BAR; a_description: TUPLE; a_enabled: BOOLEAN): EV_TOOL_BAR_TOGGLE_BUTTON is
			-- 'a_description' is interpreted as follows
			-- [<pixmap>,<action>,"tooltip","text"]
			-- Created toolbar is added to 'a_toolbar' when provided
			-- <pixmap> may be either a STRING (name of icon file) or an EV_PIXMAP
			-- <action> may be either an agent or a STRING (interpreted as a framework command then)
		require
			description_ok: a_description /= Void and then a_description.count > 0
		do
			create Result
			if a_description.is_boolean_item (a_description.lower) then
				if a_description.boolean_item (a_description.lower) then
					Result.enable_select
				end
			elseif a_description.is_integer_item (a_description.lower) then
				if a_description.integer_item (a_description.lower) /= 0 then
					Result.enable_select
				end
			end
			decorate_toolbar_item (Result, a_description)
			if not a_enabled then
				Result.disable_sensitive
			end
			if a_toolbar /= Void then a_toolbar.extend (Result) end
		end

feature -- Basic operations

	add_menu (a_window: EV_WINDOW; a_auto_complete: BOOLEAN; a_description: ARRAY [TUPLE [STRING,TUPLE]]) is
			-- Add menu from 'a_description' to 'a_window'.
			-- Format: TUPLE [STRING,TUPLE]: STRING is the menu name, TUPLE is the menu description
			-- See 'new_menu' for more info on each TUPLE of 'a_description'
			-- 'a_auto_complete' True indicates that the menu should be completed with
			-- default framework menu entries.
		require
			valid_window: a_window /= Void
		local
			l_bar: EV_MENU_BAR
		do
			l_bar := new_menu_bar (a_window, a_description)
			if a_auto_complete then
				add_sub_menu (l_bar,"&File",default_file_menu)
				add_sub_menu (l_bar,"&View",default_view_menu)
				add_sub_menu (l_bar,"&Help",default_help_menu)
			end
		end

	default_file_menu: ARRAY [TUPLE] is
			-- Default 'file' menu
		once
			Result := <<["Exit application","ev_quit"]>>
		end

	default_view_menu: ARRAY [TUPLE] is
			-- Contents of the default "View" menu that appears on frameworked windows
			-- Add menu descriptions here to see them appear in these menus
		once
			create Result.make (1,0)
			add_framework_command_menu_description (Result,"Configuration","config_browser","config_browser.ico")
			add_framework_command_menu_description (Result,"Object Browser","object_browser","magnify.ico")
			add_framework_command_menu_description (Result,"Database status","database_status",Void)
			add_framework_command_menu_description (Result,"Error log","ev_error",Gui.rep.pixmaps.Error_pixmap)
			add_framework_command_menu_description (Result,"Console","ev_console","console.ico")
			add_framework_command_menu_description (Result,"Monitor","ev_monitor","monitor.ico")
			Result.force ([1, "Log level: UNKNOWN", agent set_logging_unknown], Result.upper + 1)
			Result.force ([0, "Log level: DEBUG", agent set_logging_debug], Result.upper + 1)
			Result.force ([0, "Log level: INFORMATION", agent set_logging_inform], Result.upper + 1)
			Result.force ([0, "Log level: WARNING", agent set_logging_warning], Result.upper + 1)
			Result.force ([0, "Log level: ERROR", agent set_logging_error], Result.upper + 1)
		end
	
	set_logging_unknown
			-- Set logger display level to UNKNOWN and framework messaging to equivalent.
		do
			logger.set_severity (Unknown_severity)
			set_verbosity (verbosity_from_severity (Unknown_int))
		end
	
	set_logging_debug
			-- Set logger display level to DEBUG and framework messaging to equivalent.
		do
			logger.set_severity (Debug_severity)
			set_verbosity (verbosity_from_severity (Debug_int))
		end
	
	set_logging_inform
			-- Set logger display level to INFORMATION and framework messaging to equivalent.
		do
			logger.set_severity (Information_severity)
			set_verbosity (verbosity_from_severity (Information_int))
		end
	
	set_logging_warning
			-- Set logger display level to WARNING and framework messaging to equivalent.
		do
			logger.set_severity (Warning_severity)
			set_verbosity (verbosity_from_severity (Warning_int))
		end
	
	set_logging_error
			-- Set logger display level to ERROR and framework messaging to equivalent.
		do
			logger.set_severity (Minor_severity)
			set_verbosity (verbosity_from_severity (Minor_int))
		end

	set_verbosity (a_level: INTEGER)
			-- Set framework message verbosity to `a_level'.
		local
			l_channel: ROSE_OUTCHANNEL
		do
			l_channel := Global.io_handler.log
			if l_channel = Void then
				l_channel := Global.io_handler.output
			end
			l_channel.set_accepted_verbosity (a_level)
		end

	default_help_menu: ARRAY [TUPLE] is
			-- Contents of the default "Help" menu that appears on frameworked windows
			-- Add menu descriptions here to see them appear in these menus
		once
			Result := <<
				[documentation_menu (Global.app.program_name)],
				Void,
				["List all commands","ev_console & help"],
				["About...","ev_about"]
			>>
		end

	documentation_menu (a_program_name: STRING): EV_MENU_ITEM is
			-- Menu linking to documentation file(s) of application `a_program_name'
		require
			a_program_name_not_void: a_program_name /= Void
		local
			l_menu: EV_MENU
			l_menu_items: LINKED_LIST [EV_MENU_ITEM]
			l_pixmap: EV_PIXMAP
		do
			l_menu_items := documentation_items (a_program_name)
			if l_menu_items.is_empty then
				create Result
				Result.disable_sensitive
			elseif l_menu_items.count = 1 then
				Result := l_menu_items.first
			else
				from
					create l_menu
					l_menu_items.start
				until
					l_menu_items.off
				loop
					l_menu.extend (l_menu_items.item)
					l_menu_items.forth
				end
				Result := l_menu
			end
			Result.set_text ("Documentation ")
			l_pixmap := Gui.factory.new_pixmap_from_file ("help.ico")
			if l_pixmap /= Void then
				Gui.factory.stretch_pixmap (l_pixmap,16,16)
				Result.set_pixmap (l_pixmap)
			end
		ensure
			valid_result: Result /= Void
		end
	
	replace_sub_menu (a_window: EV_WINDOW; a_sub_name: STRING; a_description: ARRAY [TUPLE]) is
			-- Add sub-menu with `a_sub_name' from `a_description' to menu bar of `a_window'.
			-- If `a_sub_name' is already a sub-menu of `a_window.menu_bar' then it is replaced.
		require
			a_window_not_void: a_window /= Void
			menu_bar_exists: a_window.menu_bar /= Void
			a_description_not_void: a_description /= Void
		do
			check attached a_window.menu_bar as l_bar then
					-- from precondition `menu_bar_exists'
				if a_sub_name /= Void and then attached get_sub_menu (l_bar, a_sub_name) as l_menu then
					l_menu.wipe_out
				end
				add_sub_menu (l_bar, a_sub_name, a_description)
			end
		end
	
	add_sub_menu (a_target: EV_MENU_ITEM_LIST; a_sub_name: STRING; a_description: ARRAY [TUPLE]) is
			-- Add sub-menu with `a_sub_name' from `a_description' to `a_target'.
			-- If `a_sub_name' is already a sub-menu of `a_target', menus from `a_description' are added to it
			-- If `a_sub_name' is Void, extends menus from `a_description' directly to `a_target'
		require
			description_ok: a_description /= Void
		local
			l_menu: EV_MENU
			i: INTEGER
		do
			if a_sub_name = Void then
				l_menu ?= a_target
			else
				l_menu := get_sub_menu (a_target,a_sub_name)
				if l_menu = Void then
					create l_menu.make_with_text (a_sub_name + " ")
					if a_sub_name ~ "File" or a_sub_name ~ "&File" then
						a_target.put_front (l_menu)
					else
						a_target.extend (l_menu)
					end
				end
			end
			from
				i := a_description.lower
			until
				i > a_description.upper
			loop
				add_menu_item (l_menu, a_description.item (i), True)
				i := i + 1
			end
		end

	add_toolbar (a_container: EV_CONTAINER; a_description: ARRAY [TUPLE]) is
			-- Add toolbar from 'a_description' to 'a_container'.
		require
			valid_container: a_container /= Void
		local
			l_toolbar: EV_TOOL_BAR
		do
			l_toolbar := new_toolbar (a_container,a_description)
		end

	select_by_name (a_list: EV_ANY; a_name: STRING; a_selected: BOOLEAN) is
			-- Perform '(en|dis)able_select' on all items in 'a_list' which correspond to 'a_name'
		local
			l_selectable: EV_SELECTABLE
		do
			l_selectable ?= item_by_name (a_list,a_name)
			Gui.enable_select (l_selectable,a_selected)
		end

	enable_by_name (a_list: EV_ANY; a_name: STRING; a_enabled: BOOLEAN) is
			-- Perform '(en|dis)able_sensitive' on all items in 'a_list' which correspond to 'a_name'
		local
			l_sensitive: EV_SENSITIVE
		do
			l_sensitive ?= item_by_name (a_list, a_name)
			Gui.enable_sensitive (l_sensitive, a_enabled)
		end

	enable_by_names (a_list: EV_ANY; a_names: ARRAY [STRING]; a_enabled: BOOLEAN) is
			-- Perform '(en|dis)able_sensitive' on all items in 'a_list' which correspond to 'a_name'
		local
			i: INTEGER
		do
			if a_names /= Void then
				from
					i := a_names.lower
				until
					i > a_names.upper
				loop
					enable_by_name (a_list,a_names.item (i),a_enabled)
					i := i + 1
				end
			end
		end

feature -- Control building

	frozen add_toolbar_item (a_toolbar: EV_TOOL_BAR; a_description: TUPLE; a_enabled: BOOLEAN) is
			-- Add a toolbar item corresponding to 'a_description' to 'a_toolbar'
			-- 'a_enabled' indicates whether the item should be enabled or not
			-- See 'new_toolbar_item' for format of each tuple in 'a_description'
		require
			toolbar_ok: a_toolbar /= Void
		local
			l_item: EV_TOOL_BAR_ITEM
		do
			l_item := new_toolbar_item (a_toolbar, a_description, a_enabled)
		end

	frozen add_menu_separator (a_menu: EV_MENU) is
			-- Add a separator to 'a_menu'
		do
			if a_menu /= Void and then a_menu.count > 0 then
				gui.factory.add_menu_item (a_menu, Void, True)
			end
		end

	frozen add_menu_item (a_menu: EV_MENU; a_description: TUPLE; a_enabled: BOOLEAN) is
			-- See 'new_menu_item' for explanation on arguments.
		require
			menu_ok: a_menu /= Void
		local
			l_menu_item: EV_MENU_ITEM
		do
			l_menu_item := new_menu_item (a_menu, a_description, a_enabled)
		end

	decorate_toolbar_item (a_button: EV_TOOL_BAR_BUTTON; a_description: TUPLE) is
			-- Decorate 'a_button' accordingly to 'a_description', which is considered to start at 'a_start'.
			-- 'a_description' is of the form:
			-- [<pixmap>,<action>,"tooltip","text"]
			-- <pixmap> is either a EV_PIXMAP or a icon name
			-- <action> is either an agent or a string (executed as a framework command)
			-- All items are optional but must appear in that order/position, everything else is ignored silently
		local
			l_pixmap: EV_PIXMAP
			l_string: STRING
			l_agent: PROCEDURE [ANY,TUPLE]
			i: INTEGER
		do
			if a_description /= Void and then a_description.count > 0 then
				i := a_description.lower
				if a_description.is_boolean_item (i) or else a_description.is_integer_item (i) then
					i := i + 1
				end
				-- Pixmap
				if a_description.valid_index (i) then
					l_pixmap ?= a_description.item (i)
					l_string ?= a_description.item (i)
					if l_pixmap = Void and then l_string /= Void then
						l_pixmap := Gui.factory.new_pixmap_from_file (l_string)
					end
					if l_pixmap /= Void then
						-- Size is 'standardized' to 16x16
						Gui.factory.stretch_pixmap (l_pixmap,16,16)
						a_button.set_pixmap (l_pixmap)
					end
					i := i + 1
				end
				if a_description.valid_index (i) then
					-- Associated agent
					l_agent ?= a_description.item (i)
					l_string ?= a_description.item (i)
					if l_agent = Void and then l_string /= Void then
						l_agent := agent Gui.execute (l_string)
					end
					if l_agent /= Void then
						a_button.select_actions.extend (l_agent)
					else
						a_button.disable_sensitive
					end
					i := i + 1
				end
				if a_description.valid_index (i) then
					-- Tooltip
					l_string ?= a_description.item (i)
					if l_string /= Void then
						a_button.set_tooltip (l_string)
					end
					i := i + 1
				end
				if a_description.valid_index (i) then
					-- Text
					l_string ?= a_description.item (i)
					if l_string /= Void then
						a_button.set_text (l_string)
					end
				end
			end
		end

	add_auto_split_menu (a_parent: EV_MENU; a_max_items: INTEGER; a_data: ARRAY [ANY]; a_callback: PROCEDURE [ANY, TUPLE [ANY]]; a_representor: FUNCTION [ANY, TUPLE [ANY], STRING]) is
			-- Add a menu listing all items in 'a_data', but split the created menu
			-- in submenus of 'a_max_items' each
		require
			parent_ok: a_parent /= Void and then a_max_items >= 10 and then a_callback /= Void
		local
			l_menu: EV_MENU
			l_object: ANY
			l_prev, s: STRING
			i, l_tolerance: INTEGER
		do
			if a_max_items < 40 then
				l_tolerance := a_max_items * 2
			else
				l_tolerance := a_max_items
			end
			if a_data = Void or else a_data.count = 0 then
				-- Do nothing
			elseif a_data.count <= l_tolerance then
				from
					i := a_data.lower
				until
					i > a_data.upper
				loop
					l_object := a_data.item (i)
					if l_object /= Void then
						if a_representor /= Void then
							s := a_representor.item ([l_object])
						else
							s := l_object.out
						end
						add_menu_item (a_parent, [s, agent a_callback.call ([l_object])], True)
					end
					i := i + 1
				end
			else
				from
					i := a_data.lower
				until
					l_object /= Void or else i > a_data.upper
				loop
					l_object := a_data.item (i)
					if l_object = Void then
						i := i + 1
					end
				end
				if l_object /= Void then
					from
						if a_representor /= Void then
							s := a_representor.item ([l_object])
						else
							s := l_object.out
						end
						create l_menu.make_with_text (global.formatter.maxed_string (s, 8))
						a_parent.extend (l_menu)
					until
						i > a_data.upper
					loop
						l_object := a_data.item (i)
						if l_object /= Void then
							if a_representor /= Void then
								s := a_representor.item ([l_object])
							else
								s := l_object.out
							end
							if i \\ a_max_items = 0 then
								l_menu.set_text (global.formatter.maxed_string (l_menu.text, 8) + " -> " + global.formatter.maxed_string (l_prev, 8))
								create l_menu.make_with_text (global.formatter.maxed_string (s, 8))
								a_parent.extend (l_menu)
							end
							add_menu_item (l_menu, [s, agent a_callback.call ([l_object])], True)
							l_prev := s
						end
						i := i + 1
					end
					if i \\ a_max_items /= 0 then
						l_menu.set_text (global.formatter.maxed_string (l_menu.text, 8) + " -> " + global.formatter.maxed_string (l_prev, 8))
					end
				end
			end
		end

feature {NONE} -- Implementation

	decorate_menu_item (a_item: EV_MENU_ITEM; a_description: TUPLE; a_start: INTEGER): BOOLEAN is
			-- Decorate 'a_item' accordingly to 'a_description' of the form:
			-- ["text",Void] : disabled item with "text" on it
			-- ["text",<action>] : item with "text" executing <action> when used
			-- ["text",<action>,<pixmap>] : same as above, with <pixmap>
			-- <action> can either be an agent or a string (in which case it is interpreted as a framework command)
			-- <pixmap> can either a EV_PIXMAP or a string (in which case the pixmap is loaded from icon of that name)
		local
			l_text: STRING
			l_agent: PROCEDURE [ANY, TUPLE]
			l_pixmap: EV_PIXMAP
			retried: BOOLEAN
			i: INTEGER
		do
			if not retried and then a_item /= Void and then a_description /= Void and then a_description.count >= 2 then
				i := a_start
				l_text ?= a_description.item (i)
				if l_text /= Void then
					if not l_text.is_empty then
						l_text := l_text + " "
					end
					a_item.set_text (l_text)
					i := i + 1
					l_text ?= a_description.item (i)
					l_agent ?= a_description.item (i)
					if l_text /= Void then
						a_item.select_actions.extend (agent Gui.execute (l_text))
					elseif l_agent /= Void then
						a_item.select_actions.extend (l_agent)
					else
						a_item.disable_sensitive
					end
					i := i + 1
					if i <= a_description.upper then
						l_text ?= a_description.item (i)
						if l_text /= Void then
							l_pixmap := Gui.factory.new_pixmap_from_file (l_text)
						else
							l_pixmap ?= a_description.item (i)
						end
						if l_pixmap /= Void then
								--| Size is 'standardized' to 16x16
							Gui.factory.stretch_pixmap (l_pixmap,16,16)
							a_item.set_pixmap (l_pixmap)
						end
						i := i + 1
					end
					Result := True
				end
			end
		rescue
			if not retried then
				retried := True
				retry
			end
		end

	get_first_item_named (a_list: EV_ANY; a_name: STRING; a_cell: CELL [EV_ANY]; a_depth: INTEGER) is
			-- Get first item with 'a_name' in 'a_list'.
		require
			name_ok: a_name /= Void -- and then a_name.is_lower_case
		local
			l_list: LINEAR [EV_ANY]
		do
			if a_cell.item = Void then
				if is_item_named (a_list, a_name) then
					a_cell.put (a_list)
				elseif a_depth /= 0 then
					l_list ?= a_list
					if l_list /= Void then
						from
							l_list.start
						until
							l_list.off
						loop
							get_first_item_named (l_list.item, a_name, a_cell, a_depth - 1)
							l_list.forth
						end
					end
				end
			end
		end

	is_item_named (a_item: EV_ANY; a_name: STRING): BOOLEAN is
			-- Is 'a_item' named with 'a_name'?
			-- .text and .tooltip are compare to 'a_name' in 'a_item'
		require
			name_ok: a_name /= Void -- and then a_name.is_lower_case
		local
			l_textable: EV_TEXTABLE
			l_tooltipable: EV_TOOLTIPABLE
		do
			l_textable ?= a_item
			if l_textable /= Void then
				Result := is_insensitively_equal (a_name,l_textable.text)
			end
			if not Result then
				l_tooltipable ?= a_item
				if l_tooltipable /= Void then
					Result := is_insensitively_equal (a_name,l_tooltipable.tooltip)
				end
			end
		end

	is_insensitively_equal (a_original, a_text: STRING): BOOLEAN is
			-- Are strings 'a_original' and 'a_text' equal, disregarding any character case difference?
		require
			original_ok: a_original /= Void -- and then a_original.is_lower_case
		local
			s1, s2: STRING
		do
			if a_original /= Void and then a_text /= Void then
				s1 := a_original.as_lower
				s1.left_adjust
				s1.right_adjust
				s1.prune_all ('&')
				s2 := a_text.as_lower
				s2.left_adjust
				s2.right_adjust
				s2.prune_all ('&')
				Result := equal (s1,s2)
			end
		end

	wiki_documentation_item (a_program_name: STRING): EV_MENU_ITEM is
			-- Menu item sending to wiki documentation for 'a_program_name'
		require
			program_name_ok: a_program_name /= Void
		local
			l_text: STRING
			l_filename: STRING
			l_pixmap: EV_PIXMAP
		do
			create Result
			l_filename := Global.formatter.file_name_no_extension (a_program_name)
			l_text := "http://prism/modules/wacko/wakka.php?wakka=" + l_filename
			create Result
			Result.select_actions.extend (agent Global.explore_url (l_text))
			Result.set_text ("Prism Wiki")
			l_pixmap := Gui.factory.new_pixmap_sized ("doc_wiki.ico",16,16)
			if l_pixmap /= Void then
				Gui.factory.stretch_pixmap (l_pixmap,16,16)
				Result.set_pixmap (l_pixmap)
			end
		ensure
			valid_result: Result /= Void
		end

	documentation_items (a_program_name: STRING): LINKED_LIST [EV_MENU_ITEM] is
			-- Available documentation for 'a_program_name'
			-- See 'help_extensions' for searched document types
			-- Returns a collection of menu items linked to fire up the corresponfing documentation file found
		local
			l_filename: STRING
			l_menu_item: EV_MENU_ITEM
			l_pixmap: EV_PIXMAP
			l_text: STRING
		do
			from
				create Result.make
				help_extensions.start
			until
				help_extensions.off
			loop
				l_filename := Global.app.docs.get_file (a_program_name + help_extensions.item.first)
				if l_filename /= Void then
					create l_menu_item
					l_text := help_extensions.item.second
					if l_text /= Void then
						if not l_text.is_empty then
							l_text := l_text + " "
						end
						l_menu_item.set_text (l_text)
					end
					l_menu_item.select_actions.extend (agent Global.execute_file (l_filename))
					l_pixmap := Gui.factory.new_pixmap_sized ("doc" + help_extensions.item.first + ".ico",16,16)
					if l_pixmap /= Void then
						Gui.factory.stretch_pixmap (l_pixmap,16,16)
						l_menu_item.set_pixmap (l_pixmap)
					end
					Result.extend (l_menu_item)
				end
				help_extensions.forth
			end
			Result.extend (wiki_documentation_item (a_program_name))
		ensure
			valid_result: Result /= Void
		end

	add_framework_command_menu_description (a_target: ARRAY [TUPLE]; a_title, a_command: STRING; a_pixmap: ANY) is
			-- Add a menu description to 'a_target' invoking framework command 'a_command'
			-- only if 'a_command' actually is a framewrok command
			-- Do nothing otherwise (if 'a_command' is not defined)
		require
			valid_target: a_target /= Void
			valid_title: a_title /= Void
		do
			if Global.commands.is_command (a_command) then
				if a_pixmap = Void then
					a_target.force ([a_title,a_command],a_target.upper+1)
				else
					a_target.force ([a_title,a_command,a_pixmap],a_target.upper+1)
				end
			end
		end

	help_extensions: LINKED_LIST [PAIR [STRING, STRING]] is
			-- Recognized help files, based on file extension
		once
			create Result.make
			add_help_extension (Result,".chm","Help Center")
			add_help_extension (Result,".htm","Web page")
			add_help_extension (Result,".doc","Word document")
			add_help_extension (Result,".ppt","PowerPoint presentation")
			add_help_extension (Result,".xls","Excel Spreadsheet")
			add_help_extension (Result,".pdf","Adobe Portable Document Format")
			add_help_extension (Result,".vsd","Visio diagram")
			add_help_extension (Result,".txt","Text file")
		end

	add_help_extension (a_sequence: SEQUENCE [PAIR [STRING, STRING]]; a_extension, a_description: STRING) is
			-- Add a help entry to 'a_sequence' for help file with 'a_extension' and 'a_description'
		require
			valid_sequence: a_sequence /= Void
			valid_extension: a_extension /= Void
			valid_description: a_description /= Void
		do
			a_sequence.extend (create {PAIR [STRING, STRING]}.make (a_extension, a_description))
		end

end
