indexing
	description: "EiffelVision utility for persisiting general widgets"
	author: "Zoran Simic"
	date: "$Date: 2008/12/01 $"
	revision: "$Revision: #15 $"

class
	ROSE_EV_FACTORY_PERSISTING

inherit
	ROSE_SHARED_GLOBAL_GUI

create
	make

feature {NONE} -- Initialization

	make (a_user_prefs: ROSE_USER_PREFERENCES) is
		require
			a_user_prefs /= Void
		do
			user_prefs := a_user_prefs
		end

feature -- Access

	user_prefs: ROSE_USER_PREFERENCES
			-- User config to use to remember settings

	value (a_id: STRING): STRING is
			-- String value associated to 'a_id' remembered via the user config
		do
			Result := user_prefs.remembered_value (a_id)
		end

	boolean (a_id: STRING): BOOLEAN is
		do
			Result := user_prefs.remembered_boolean (a_id)
		end

	provided_boolean (a_id: STRING): BOOLEAN is
		do
			Result := user_prefs.has_remembered_boolean (a_id)
		end

	new_text_field (a_container: EV_CONTAINER; a_id, a_label: STRING; a_width, a_capacity: INTEGER): EV_TEXT_FIELD is
			-- New text-field identified by 'a_id' put into 'a_container', with 'a_label' and 'a_width'
			-- 'a_capacity' indicates how many items to historize
			-- Items are memorized via user's config.
			-- If 'a_capacity' is < 0, then the text field is actually a combo box.
		local
			l_history: ROSE_INPUT_HISTORY
		do
			if a_capacity >= 0 then
				if a_id /= Void and then not a_id.is_empty then
					l_history := user_prefs.get_history (a_id,a_capacity.max (1),True)
					Result := Gui.factory.new_text_field (a_container,a_label,Void,agent textfield_handler (l_history,Result),a_width)
					Result.key_release_actions.extend (agent handle_key (l_history,Result,?))
					widgets.force (Result,a_id)
				else
					Result := Gui.factory.new_text_field (a_container,a_label,Void,Void,a_width)
					Result.set_text ("No 'id' provided, this text-field won't be historized")
				end
			else
				Result := new_combo (a_container,a_id,a_label,a_width,a_capacity)
			end
		end

	new_combo (a_container: EV_CONTAINER; a_id, a_label: STRING; a_width, a_capacity: INTEGER): ROSE_COMBO_BOX is
			-- New combo-box identified by 'a_id' put into 'a_container', with 'a_label' and 'a_width'
			-- 'a_capacity' indicates how many items to historize
			-- Items are memorized via user's config.
			-- If 'a_capacity' is < 0, then the text field is actually a combo box.
		local
			l_history: ROSE_INPUT_HISTORY
		do
			if a_id /= Void and then not a_id.is_empty then
				l_history := user_prefs.get_history (a_id,a_capacity.max (1),True)
				Result := Gui.factory.new_combo (a_container,a_label,item_array (l_history),l_history.current_item,Void,a_width)
				Result.return_actions.extend (agent combo_handler (l_history,Result))
				widgets.force (Result,a_id)
			else
				Result := Gui.factory.new_combo (a_container,a_label,Void,Void,Void,a_width)
				Result.set_text ("No 'id' provided, this combo won't be historized")
			end
		end

	new_check_button (a_container: EV_CONTAINER; a_id, a_label: STRING; a_action: PROCEDURE [ANY, TUPLE []]; a_width: INTEGER): EV_CHECK_BUTTON is
			-- New check-button identified by 'a_id' added to 'a_container', with 'a_label' and 'a_width'.
			-- Query the value of this checkbox via 'boolean (a_id)'
		do
			if a_id /= Void and then not a_id.is_empty then
				Result := Gui.factory.new_check_button (a_container,boolean (a_id),a_label,Void,a_width)
				Result.enable_sensitive
				Result.select_actions.extend (agent checkbox_handler (a_id,Result))
				if a_action /= Void then
					Result.select_actions.extend (agent a_action.call ([]))
					a_action.call ([])
				end
				widgets.force (Result,a_id)
			else
				Result := Gui.factory.new_check_button (a_container,False,a_label,Void,a_width)
				Result.disable_sensitive
			end
		end

feature -- Basic operations

	set_value (a_key: STRING; a_value: STRING) is
		do
			user_prefs.set_remembered_value (a_key,a_value)
		end

	set_boolean (a_key: STRING; a_value: BOOLEAN) is
		do
			user_prefs.set_remembered_boolean (a_key,a_value)
		end

feature {NONE} -- Implementation

	widgets: HASH_TABLE [EV_WIDGET, STRING] is once create Result.make (0) end
			-- Widgets hashed by their respective ids

	checkbox_handler (a_id: STRING; a_checkbox: EV_CHECK_BUTTON) is
		do
			user_prefs.set_remembered_boolean (a_id, a_checkbox.is_selected)
		end

	textfield_handler (a_history: ROSE_INPUT_HISTORY; a_field: EV_TEXT_FIELD) is
		require
			valid_history: a_history /= Void
		local
			l_string: STRING
		do
			l_string := a_field.text
			if l_string /= Void and then not l_string.is_empty then
				a_history.add (l_string)
			end
		end

	combo_handler (a_history: ROSE_INPUT_HISTORY; a_combo: ROSE_COMBO_BOX) is
		local
			l_string: STRING
		do
			l_string := a_combo.text
			if l_string /= Void and then not l_string.is_empty then
				a_history.add (l_string)
				Gui.factory.populate_combo_with_array (a_combo,item_array (a_history))
				set_field_text (a_history.current_item,a_combo)
			end
		end

	handle_key (a_history: ROSE_INPUT_HISTORY; a_field: EV_TEXT_FIELD; a_key: EV_KEY) is
		require
			valid_history: a_history /= Void
			valid_field: a_field /= Void
		do
			if a_key /= Void then
				if a_key.code = Gui.rep.keys.Key_up then
					if not a_history.is_before then
						a_history.go_previous
						set_field_text (a_history.current_item,a_field)
					end
				elseif a_key.code = Gui.rep.keys.Key_down then
					if not a_history.is_after then
						a_history.go_next
						set_field_text (a_history.current_item,a_field)
					end
				end
			end
		end

	set_field_text (a_text: STRING; a_field: EV_TEXT_FIELD) is
		do
			if a_text /= Void then
				a_field.set_text (a_text)
				a_field.set_caret_position (a_text.count+1)
			else
				a_field.remove_text
			end
		end

	item_array (a_history: ROSE_INPUT_HISTORY): ARRAY [STRING] is
		local
			i: INTEGER
		do
			if a_history /= Void then
				from
					i := 1
					create Result.make (1,a_history.count)
				until
					i > a_history.count
				loop
					Result.put (a_history.item (i),a_history.count-i+1)
					i := i + 1
				end
			end
		end

end
