indexing
	description: "EiffelVision widget factory, producing objects with most commonly used settings"
	notice: "[
		All parameters are optional, unless otherwise noted
		All widths and heights provided are interpreted as follows:
		- > 0 sets the size of the widget (not resizable if container provided, minimum size otherwise)
		- = 0 expandable, no size set
		- < 0 expandable, sets the minimum size of the object with absolute value
		Interpret the following as :
		new_label (my_container, "Hi", 0): give me a label that says 'Hi', put it in 'my_container' and let it resize at will
		new_label (my_container, "Hi", 20): a label that has a fix size of 20 pixels
		new_label (my_container, "Hi", -20): a label that has a minimum size of 20 pixels, but that can resize at will above that
		new_label (Void, "Hi", 0): just give me a label that says 'Hi'
		]"
	author: "Zoran Simic"
	date: "$Date: 2009/11/06 $"
	revision: "$Revision: #65 $"

class
	ROSE_EV_FACTORY_WIDGET

inherit
	ROSE_SHARED_GLOBAL_GUI

feature -- Access

	loaded_pixmaps: HASH_TABLE [EV_PIXMAP, STRING] is
	--loaded_pixmaps: ROSE_FIFO_CACHE [EV_PIXMAP, STRING] is
			-- Table of all loaded pixmaps so far.
			-- This is needed to avoid reading over and other the same pixmaps
			-- It can also be used to see what pixmaps your GUI is using.
			-- Void can be found in this table, a Void means that the pixmap could not be found on disk
		once
			create Result.make (128)
		end

	new_hbox (a_container: EV_CONTAINER; a_border, a_padding: INTEGER): EV_HORIZONTAL_BOX is
			-- New horizontal box with 'a_border' and 'a_padding', inserted
			-- in 'a_container' if given (and not expandable within that box by default)
			-- Provide border or padding < 0 to make returned box expandable (only when 'a_container' is given)
		do
			create Result
			set_box_defaults (a_container, Result, a_border, a_padding)
		end

	new_vbox (a_container: EV_CONTAINER; a_border, a_padding: INTEGER): EV_VERTICAL_BOX is
			-- New vertical box with 'a_border' and 'a_padding', inserted
			-- in 'a_container' if given (and not expandable within that box by default)
			-- Provide border or padding < 0 to make returned box expandable (only when 'a_container' is given)
		do
			create Result
			set_box_defaults (a_container, Result, a_border, a_padding)
		end

	new_hsplit (a_container: EV_CONTAINER; a_widgets: ARRAY [ANY]): EV_HORIZONTAL_SPLIT_AREA is
			-- New horizontal split populated with 'a_widgets'
			-- Item of 'a_widget' should be of type either EV_WIDGET or TUPLE [EV_WIDGET, INTEGER]
			-- in the latter case, the INTEGER is the size of the corresponding EV_WIDGET
			-- All other types in 'a_widgets' are silently ignored
		do
			create Result
			set_split_defaults (a_container, Result, a_widgets, True)
		end

	new_vsplit (a_container: EV_CONTAINER; a_widgets: ARRAY [ANY]): EV_VERTICAL_SPLIT_AREA is
			-- New vertical split populated with 'a_widgets'
			-- Item of 'a_widget' should be of type either EV_WIDGET or TUPLE [EV_WIDGET, INTEGER]
			-- in the latter case, the INTEGER is the size of the corresponding EV_WIDGET
			-- All other types in 'a_widgets' are silently ignored
		do
			create Result
			set_split_defaults (a_container, Result, a_widgets, False)
		end

	new_hsplit2 (a_container: EV_CONTAINER; a_width: INTEGER): EV_HORIZONTAL_SPLIT_AREA is
			-- New horizontal split with minimum 'a_width'
		do
			create Result
			enbox_and_size_simple (Result, a_container, a_width)
		end

	new_vsplit2 (a_container: EV_CONTAINER; a_height: INTEGER): EV_VERTICAL_SPLIT_AREA is
			-- New vertical split with minimum 'a_height'
		do
			create Result
			enbox_and_size_simple (Result, a_container, a_height)
		end

	new_cell (a_container: EV_CONTAINER; a_width, a_height: INTEGER): EV_CELL is
			-- New cell inserted in 'a_container', containing 'a_item'
		do
			create Result
			enbox_and_size (Result, a_container, a_width, a_height)
		end

	new_label (a_container: EV_CONTAINER; a_label: STRING; a_width: INTEGER): EV_LABEL is
			-- New label inserted in 'a_container', with 'a_label' as text and 'a_width' as minimum width
			-- If 'a_width' is 0, label resizes automatically in box, fixed size otherwise.
		do
			create Result
			set_text (Result, a_label)
			Result.align_text_left
			enbox_and_size_simple (Result, a_container, a_width)
		end

	new_3d_label (a_container: EV_CONTAINER; a_label: STRING; a_width: INTEGER): EV_LABEL is
			-- Same as 'new_label', but colors the label in 3d colors
		do
			Result := new_label (a_container, a_label, a_width)
			Result.set_background_color (Gui.rep.colors.Color_3d_highlight)
			Result.set_foreground_color (Gui.rep.colors.Color_3d_shadow)
		end

	new_active_label (a_container: EV_CONTAINER; a_label: STRING; a_action: PROCEDURE [ANY, TUPLE []]; a_width: INTEGER): EV_LABEL is
			-- New label reacting to the mouse hovering over it
			-- 'a_action' is called whenever the mouse is clicked on the returned label
		do
			Result := new_label (a_container, a_label, a_width)
			if a_action /= Void then
				Result.pointer_button_press_actions.extend (agent on_agent_click (a_action, ?, ?, ?, ?, ?, ?, ?, ?))
			end
			Result.pointer_enter_actions.extend (agent Result.set_background_color (Gui.rep.colors.Color_3d_highlight))
			Result.pointer_leave_actions.extend (agent Result.set_background_color (Gui.rep.colors.Color_3d_face))
			Result.align_text_center
		end

	new_text_field (a_container: EV_CONTAINER; a_label, a_text: STRING; a_action: PROCEDURE [ANY, TUPLE []]; a_width: INTEGER): EV_TEXT_FIELD is
			-- New text field inserted in 'a_container' with 'a_label' and 'a_text'
			-- Text box is expandable if 'a_width' is 0, fixed size otherwise
		do
			create Result
			set_text (Result, a_text)
			if a_action /= Void then
				Result.return_actions.extend (a_action)
			end
			if a_width.abs > 1 then
				Result.set_minimum_width (a_width.abs)
			end
			add_labeled_widget (a_container, Result, a_label, a_width)
		end

	new_text (a_container: EV_CONTAINER; a_text: STRING; a_width, a_height: INTEGER): EV_TEXT is
			-- New text box inserted in 'a_container' and with 'a_text'
			-- Text is expandable if 'a_width' and 'a_height' given are 0 (one 0 is enough)
			-- Fixed size if both are non 0, eventual label is always fixed size
		do
			create Result
			set_text (Result, a_text)
			enbox_and_size (Result, a_container, a_width, a_height)
		end

	new_date_field (a_container: EV_CONTAINER; a_label: STRING; a_action: PROCEDURE [ANY, TUPLE []]): ROSE_EV_DATE_FIELD is
			-- New date field, preceded with 'a_label' if provided
			-- 'a_action' is called whenever the selected date changes
		do
			create Result
			if a_action /= Void then
				Result.change_actions.extend (a_action)
			end
			add_labeled_widget (a_container, Result, a_label, 1)
		end

	new_button (a_container: EV_CONTAINER; a_label: STRING; a_action: PROCEDURE [ANY, TUPLE []]; a_width: INTEGER): EV_BUTTON is
			-- New button inserted in 'a_container' with 'a_label' and associated to 'a_action'
			-- If 'a_width' is non 0, the button is not expandable and has 'a_width'.
		do
			create Result
			set_text (Result, a_label)
			if a_action /= Void then
				Result.select_actions.extend (a_action)
			else
				Result.disable_sensitive
			end
			enbox_and_size_simple (Result, a_container, a_width)
		end

	new_button_picture (a_container: EV_CONTAINER; a_pixmap: ANY; a_action: PROCEDURE [ANY, TUPLE []]): EV_BUTTON is
			-- New button with a picture instead of text
			-- The returned button is not resizable and has the same size as the pixmap
		local
			l_pixmap: EV_PIXMAP
			l_string: STRING
		do
			l_pixmap ?= a_pixmap
			if l_pixmap = Void then
				l_string ?= a_pixmap
				l_pixmap := new_pixmap_from_file (l_string)
			else
				l_pixmap := l_pixmap.twin
			end
			create Result
			if a_action /= Void then
				Result.select_actions.extend (a_action)
			end
			if l_pixmap /= Void then
				stretch_pixmap (l_pixmap, 16, 16)
				enbox_and_size_simple (Result, a_container, l_pixmap.width+8)
				Result.set_pixmap (l_pixmap)
			elseif l_string /= Void then
				set_text (Result, l_string)
				enbox_and_size_simple (Result, a_container, 0)
			end
		end

	new_info_button (a_container: EV_CONTAINER; a_text: STRING): EV_BUTTON is
			-- A special type of button that has a basic action already associated with it
			-- The action is to popup an untitled dialog that disappears when it loses focus or when ESC is pressed
			-- The popup is shown when button is clicked only if it's 'data' contains a STRING
			-- The button is put in 'a_container' if provided
			-- The button shows 'a_text' as it's label
		do
			Result := new_button (a_container, a_text, Void, 1)
			Result.enable_sensitive
			set_info_action (Result, Void)
		end

	new_help_button (a_container: EV_CONTAINER; a_help: STRING): EV_BUTTON is
			-- A special type of button showing a question mark icon
			-- When pushed, 'a_help' is displayed
		do
			Result := new_button_picture (a_container, "help.ico", agent (Gui.dialogs).popup_info (a_help, Void))
		end

	new_check_button (a_container: EV_CONTAINER; a_checked: BOOLEAN; a_label: STRING; a_action: PROCEDURE [ANY, TUPLE []]; a_width: INTEGER): EV_CHECK_BUTTON is
			-- New check button inserted in 'a_container' with 'a_label' and associated to 'a_action'
			-- If 'a_width' is non 0, the button is not expandable and has 'a_width'.
		do
			create Result
			set_text (Result, a_label)
			if a_checked then
				Result.enable_select
			else
				Result.disable_select
			end
			if a_action /= Void then
				Result.select_actions.extend (a_action)
			end
			enbox_and_size_simple (Result, a_container, a_width)
		end

	new_radio_button (a_container: EV_CONTAINER; a_checked: BOOLEAN; a_label: STRING; a_action: PROCEDURE [ANY, TUPLE []]; a_width: INTEGER): EV_RADIO_BUTTON is
			-- New radio button inserted in 'a_container' with 'a_label' and associated to 'a_action'
			-- Button is mutual exclusive with all other radio button in parent containe ('a_container' if given)
			-- If 'a_width' is non 0, the button is not expandable and has 'a_width'.
		do
			create Result
			if a_label /= Void then
				set_text (Result, a_label)
			end
			if a_checked then
				Result.enable_select
			end
			if a_action /= Void then
				Result.select_actions.extend (a_action)
			end
			enbox_and_size_simple (Result, a_container, a_width)
		end

	new_toggle_button (a_container: EV_CONTAINER; a_checked: BOOLEAN; a_label: STRING; a_action: PROCEDURE [ANY, TUPLE []]; a_width: INTEGER): EV_TOGGLE_BUTTON is
			-- New toggle button inserted in 'a_container' with 'a_label' and associated to 'a_action'
			-- If 'a_width' is non 0, the button is not expandable and has 'a_width'.
		do
			create Result
			if a_label /= Void then
				set_text (Result, a_label)
			end
			if a_checked then
				Result.enable_select
			end
			if a_action /= Void then
				Result.select_actions.extend (a_action)
			end
			enbox_and_size_simple (Result, a_container, a_width)
		end

	new_frame (a_container: EV_CONTAINER; a_title: STRING; a_width, a_height: INTEGER): EV_FRAME is
			-- New frame inserted in 'a_container', with 'a_title' as frame title
			-- and put 'a_item' into the returned frame
		do
			create Result
			if a_title /= Void then
				set_text (Result, a_title)
			end
			enbox_and_size (Result, a_container, a_width, a_height)
		end

	new_combo (a_container: EV_CONTAINER; a_label: STRING; a_content: ARRAY [ANY]; a_selected: ANY; a_action: PROCEDURE [ANY, TUPLE []]; a_width: INTEGER): ROSE_COMBO_BOX is
			-- New combo box inserted in 'a_container', with items from 'a_content', and 'a_action'
			-- If 'a_width' is non 0, the combo is not expandable and has 'a_width'.
		local
			l_string, l_selected: STRING
			i: INTEGER
		do
			create Result
			populate_combo_with_array (Result, a_content)
			if a_action /= Void then
				Result.select_actions.extend (a_action)
			end
			if a_content /= Void and then a_selected /= Void then
				l_selected := Global.represented (a_selected)
				from
					i := a_content.lower
				until
					i > a_content.upper
				loop
					if a_content.item (i) /= Void then
						l_string := Global.represented (a_content.item (i))
						if equal (l_string, l_selected) then
							Result.i_th (i - a_content.lower + 1).enable_select
							i := a_content.upper + 2
						end
					end
					i := i + 1
				end
				if i = a_content.upper + 1 then
					set_text (Result, l_selected)
				end
			end
			add_labeled_widget (a_container, Result, a_label, a_width)
		end

	new_list (a_container: EV_CONTAINER; a_columns: ARRAY [ANY]; a_action: PROCEDURE [ANY, TUPLE [INTEGER, EV_MULTI_COLUMN_LIST_ROW]]; a_width, a_height: INTEGER): EV_MULTI_COLUMN_LIST is
			-- Please do not use this feature anymore, use 'new_grid'
		obsolete
			"Use RP_DATA_GRID features instead"
		do
			create Result
			set_list_defaults (Result, a_container, a_columns, a_action, a_width, a_height)
		end

	new_grid (a_container: EV_CONTAINER; a_columns: ARRAY [ANY]; a_action: PROCEDURE [ANY, TUPLE [INTEGER, RP_GRID_NODE [ANY]]]; a_width, a_height: INTEGER): RP_DATA_GRID [ANY] is
			-- New data-grid, see 'set_grid_defaults' for how arguments are to be used
		do
			create Result
			set_grid_defaults (Result, a_container, a_columns, a_action, a_width, a_height)
		end

	set_grid_defaults (a_grid: RP_DATA_GRID[ANY]; a_container: EV_CONTAINER; a_columns: ARRAY [ANY]; a_action: PROCEDURE [ANY, TUPLE [INTEGER, RP_GRID_NODE[ANY]]]; a_width, a_height: INTEGER) is
			-- Set up data grid with header from 'a_columns'.
			-- 'a_action' is associated to double-click on a row.
			-- 'a_action' receives the clicked/selected row if it's second argument is a EV_MULTI_COLUMN_LIST_ROW
			-- it receives row's associated 'data' otherwise
			-- the INTEGER in 'a_action' is the double-clicked button or:
			-- 0 when the item is deselected
			-- -1 when the item was just selected (without double-click)
		local
			i, l_column_index, l_column_width, l_grid_width: INTEGER
			l_titles: ROSE_LINEAR_ARRAY[STRING]
			l_widths: ROSE_LINEAR_ARRAY[INTEGER]
			l_string: STRING
			l_any: ANY
			l_tuple: TUPLE
		do
			a_grid.enable_single_row_selection
			if a_columns /= Void then
				from
					a_grid.set_is_header_shown (true)
					i := a_columns.lower
					create l_titles.make_empty
					create l_widths.make_empty
				until
					i > a_columns.upper
				loop
					l_column_index := i - a_columns.lower + 1
					l_any := a_columns.item (i)
					l_tuple ?= l_any
					if l_tuple /= Void and then l_tuple.count >= 1 then
						l_string ?= l_tuple.item (l_tuple.lower)
					elseif l_any /= Void then
						l_string := Global.formatter.object_representation (l_any, "")
					end
					if l_string = Void then
						l_string := ""
					end
					cleanup (l_string)
					l_titles.extend (l_string)
					if l_tuple /= Void
					and then l_tuple.count >=2
					and then l_tuple.is_integer_item (l_tuple.upper) then
						l_column_width := l_tuple.integer_item (l_tuple.upper)
					else
						l_column_width := 80
					end
					l_widths.extend (l_column_width)
					l_grid_width := l_grid_width + l_column_width
					i := i + 1
				end
				a_grid.set_titles (l_titles)
				a_grid.set_column_widths (l_widths)

			else
				a_grid.set_is_header_shown (false)
			end
			if a_action /= Void then
				a_grid.pointer_double_press_actions.extend (agent on_grid_click (a_action, a_grid, True, ?, ?, ?, ?, ?, ?, ?, ?))
				a_grid.pointer_button_release_actions.extend (agent on_grid_click (a_action, a_grid, False, ?, ?, ?, ?, ?, ?, ?, ?))
				a_grid.row_select_actions.extend (agent on_grid_select (-1, a_action, ?))
				a_grid.row_deselect_actions.extend (agent on_grid_select (0, a_action, ?))
			end
			enbox_and_size (a_grid, a_container, a_width, a_height)
			if a_columns /= Void and then l_column_index > 0 and then l_grid_width < a_grid.width then
				a_grid.set_column_width (l_column_width + a_grid.width - l_grid_width - 2, l_column_index)
			end
		end

	set_list_defaults (a_list:EV_MULTI_COLUMN_LIST; a_container: EV_CONTAINER; a_columns: ARRAY [ANY]; a_action: PROCEDURE [ANY, TUPLE [INTEGER, EV_MULTI_COLUMN_LIST_ROW]]; a_width, a_height: INTEGER) is
			-- New multi-column list with header from 'a_columns'.
			-- 'a_action' is associated to double-click on a row.
			-- 'a_action' receives the clicked/selected row if it's second argument is a EV_MULTI_COLUMN_LIST_ROW
			-- it receives row's associated 'data' otherwise
			-- the INTEGER in 'a_action' is the double-clicked button or:
			-- 0 when the item is deselected
			-- -1 when the item was just selected (without double-click)
		obsolete
			"Use RP_DATA_GRID features instead"
		local
			i, l_column_index, l_column_width, l_list_width: INTEGER
			l_string: STRING
			l_any: ANY
			l_tuple: TUPLE
		do
			a_list.disable_multiple_selection
			if a_columns /= Void then
				from
					a_list.show_title_row
					i := a_columns.lower
				until
					i > a_columns.upper
				loop
					l_column_index := i - a_columns.lower + 1
					l_any := a_columns.item (i)
					l_tuple ?= l_any
					if l_tuple /= Void and then l_tuple.count >= 1 then
						l_string ?= l_tuple.item (l_tuple.lower)
					elseif l_any /= Void then
						l_string := Global.formatter.object_representation (l_any, "")
					end
					if l_string = Void then
						l_string := ""
					end
					cleanup (l_string)
					a_list.set_column_title (l_string, l_column_index)
					if l_tuple /= Void
					and then l_tuple.count >=2
					and then l_tuple.is_integer_item (l_tuple.upper) then
						l_column_width := l_tuple.integer_item (l_tuple.upper)
					else
						l_column_width := 80
					end
					a_list.set_column_width (l_column_width, l_column_index)
					l_list_width := l_list_width + l_column_width
					i := i + 1
				end
			else
				a_list.hide_title_row
			end
			if a_action /= Void then
				a_list.pointer_double_press_actions.extend (agent on_list_click (a_action, a_list, True, ?, ?, ?, ?, ?, ?, ?, ?))
				a_list.pointer_button_release_actions.extend (agent on_list_click (a_action, a_list, False, ?, ?, ?, ?, ?, ?, ?, ?))
				a_list.select_actions.extend (agent on_list_select (-1, a_action, ?))
				a_list.deselect_actions.extend (agent on_list_select (0, a_action, ?))
			end
			enbox_and_size (a_list, a_container, a_width, a_height)
			if a_columns /= Void and then l_column_index > 0 and then l_list_width < a_list.width then
				a_list.set_column_width (l_column_width + a_list.width - l_list_width - 2, l_column_index)
			end
		end

	new_list_item (a_container: EV_MULTI_COLUMN_LIST; a_pixmap: ANY; a_content: ARRAY [ANY]; a_data: ANY): EV_MULTI_COLUMN_LIST_ROW is
			-- New list-item put in 'a_container', with 'a_pixmap', 'a_content' and 'a_data'
		obsolete
			"Use RP_DATA_GRID features instead"
		do
			create Result
			set_list_item_defaults (a_container, Result, a_pixmap, a_content, a_data)
		end

	new_grid_node (a_pixmap: ANY; a_content: ARRAY [ANY]; a_data: ANY): RP_GRID_NODE[ANY] is
			-- New grid node with 'a_pixmap' in first cell (only if first
			-- element of 'a_content' is a string), 'a_content' and 'a_data'
		local
			l_node_array: ROSE_LINEAR_ARRAY[RP_DATA_GRID_ITEM]
			i: INTEGER
			l_string: STRING
			l_pixmap: EV_PIXMAP
			l_grid: RP_DATA_GRID[ANY]
			l_label: RP_DATA_GRID_LABEL_ITEM
		do
			create l_grid
			create l_node_array.make_empty

			l_string ?= a_pixmap
			if l_string /= Void then
				l_pixmap := new_pixmap_from_file (l_string)
			else
				l_pixmap ?= a_pixmap
			end
			if l_pixmap /= Void then
				if l_pixmap.height > 16 then
					l_pixmap := scaled_pixmap(l_pixmap, 16)
				end
			end

			if a_content /= Void then
				from
					i := a_content.lower
				until
					i > a_content.upper
				loop
					l_node_array.extend (l_grid.any_to_grid_item(a_content.item (i)))
					if l_pixmap /= Void and then i = a_content.lower then
						l_label ?= l_node_array.first
						if l_label /= Void then
							l_label.set_pixmap (l_pixmap)
						end
					end
					i := i + 1
				end
			end
			create Result.make_from_array(l_node_array)
			Result.set_data (a_data)
		end

	new_tree (a_container: EV_CONTAINER; a_select_action: PROCEDURE [ANY, TUPLE []]; a_width, a_height: INTEGER): EV_TREE is
			-- New tree, put in 'a_container', with 'a_select_action' when items are selected
		do
			create Result
			if a_select_action /= Void then
				Result.select_actions.extend (a_select_action)
			end
			enbox_and_size (Result, a_container, a_width, a_height)
		end

	new_dynamic_tree_item (a_container: EV_TREE_NODE_LIST; a_pixmap: ANY; a_text: ANY; a_data: ANY; a_function: FUNCTION [ANY, TUPLE [], LINEAR [EV_TREE_NODE]]): EV_DYNAMIC_TREE_ITEM is
			-- New tree item, 'a_function'
		do
			create Result
			if a_function /= Void then
				Result.set_subtree_function (a_function)
			end
			set_tree_item_defaults (Result, a_container, a_pixmap, a_text, a_data)
		end

	new_tree_item (a_container: EV_TREE_NODE_LIST; a_pixmap: ANY; a_text: ANY; a_data: ANY): EV_TREE_ITEM is
			-- New tree item, with 'a_pixmap', 'a_text' and 'a_data'
			-- Item is added to 'a_container' if provided
		do
			create Result
			set_tree_item_defaults (Result, a_container, a_pixmap, a_text, a_data)
		end

	new_boxed_pixmap_from_file (a_container: EV_CONTAINER; a_file: STRING; a_width, a_height: INTEGER): EV_PIXMAP is
			-- Same as 'new_pixmap_from_file', but puts the returned pixmap in 'a_container'
		do
			Result := new_pixmap_from_file (a_file)
			if Result /= Void then
				Result := Result.twin
				enbox_and_size (Result, a_container, a_width, a_height)
			end
		end

	new_pixmap_from_file (a_filename: STRING): EV_PIXMAP is
			-- Pixmap retrieved from 'a_filename',
			-- Void if file name format not recognized, or no image, or bad image, ...
		local
			rescued: BOOLEAN
			l_path: STRING
		do
			if rescued then
				Result := Void
			else
				if loaded_pixmaps.has (a_filename) then
					Result := loaded_pixmaps.item (a_filename)
				else
					l_path := image_path (a_filename)
					if l_path /= Void then
						create Result
						Result.set_with_named_file (l_path)
					end
					loaded_pixmaps.put (Result, a_filename)
				end
			end
		rescue
			if not rescued then
				rescued := True
				retry
			end
		end

	new_pixmap_sized (a_filename: STRING; a_width, a_height: INTEGER): EV_PIXMAP is
			-- Pixmap retrieved from 'a_filename', Void if not available
			-- Returned pixmap is stretched to ('a_width', 'a_height') if provided
		do
			Result := new_pixmap_from_file (a_filename)
			if Result /= Void and then a_width > 0 and then a_height > 0 then
				stretch_pixmap (Result, a_width, a_height)
				if Result.width /= a_width or else Result.height /= a_height then
					Result := Void
				end
			end
		end

	new_pixmap (a_filename: STRING): EV_PIXMAP is
			-- Pixmap retrieved from 'a_filename',
			-- If no such pixmap, a default is returned (Result is guaranteed not Void)
		local
			l_stock: EV_STOCK_PIXMAPS
		do
			Result := new_pixmap_from_file (a_filename)
			if Result = Void then
				create l_stock
				Result := l_stock.Default_window_icon.twin
			end
		ensure
			valid_result: Result /= Void
		end

	new_cursor_from_file (a_filename: STRING): EV_CURSOR is
			-- Cursor retrieved from 'a_filename', or Void if a problem arises (no image, bad image, ...)
		local
			l_pixmap: EV_PIXMAP
		do
			l_pixmap := new_pixmap_from_file (a_filename)
			if l_pixmap /= Void then
				create Result.make_with_pixmap (l_pixmap, 0, 0)
			end
		end

	new_block_pixmap (w, h, d: INTEGER; b, f: EV_COLOR): EV_PIXMAP is
			-- New w x h block pixmap, the block is 'd' pixels away from the edges.
			-- 'b' and 'f' are the colors to use (background and foreground).
		do
			create Result
			Result.set_size (w, h)
			Result.set_minimum_width (w)
			if b /= Void then
				Result.set_background_color (b)
			end
			if f /= Void then
				Result.set_foreground_color (f)
			end
			Result.clear
			Result.fill_polygon (<<
				create {EV_COORDINATE}.set (d, d),
				create {EV_COORDINATE}.set (d, h-d),
				create {EV_COORDINATE}.set (w-d, h-d),
				create {EV_COORDINATE}.set (w-d, d)
			>>)
		end

	new_arrow_pixmap (a_direction: CHARACTER; w, h: INTEGER; b, f: EV_COLOR): EV_PIXMAP is
			-- New pixmap depicting an arrow taking the given 'a_direction'
		require
			direction_ok: a_direction = '<' or a_direction = '>' or a_direction = '^' or a_direction = 'v'
			size_ok: w >= 8 and then h >= 8
		local
			p1, p2, p3: EV_COORDINATE
			d, xm, ym, xe, ye: INTEGER
		do
			d := 2
			xe := w - d
			ye := h - d
			xm := d + (xe - d) // 2
			ym := d + (ye - d) // 2
			create Result
			Result.set_size (w, h)
			Result.set_minimum_width (w)
			if b /= Void then
				Result.set_background_color (b)
			end
			if f /= Void then
				Result.set_foreground_color (f)
			end
			Result.clear
			inspect a_direction
			when '<' then
				create p1.set (d, ym)
				create p2.set (xe, d)
				create p3.set (xe, ye)
			when '>' then
				create p1.set (d, d)
				create p2.set (xe, ym)
				create p3.set (d, ye)
			when '^' then
				create p1.set (d, ye-2)
				create p2.set (xm, d+1)
				create p3.set (xe, ye-2)
			else -- v
				create p1.set (d, d+2)
				create p2.set (xm, ye-1)
				create p3.set (xe, d+2)
			end
			Result.fill_polygon (<<p1, p2, p3>>)
		end

	new_text_pixmap (a_text: STRING; a_font: EV_FONT; a_min_width, a_min_height: INTEGER; b, f: EV_COLOR): EV_PIXMAP is
			-- New pixmap with 'a_text' using 'a_font',
			-- 'b' and 'f' are the colors to use (background and foreground).
		require
			text_ok: a_text /= Void and then a_font /= Void
		local
			l_size: TUPLE [INTEGER, INTEGER]
			w, h: INTEGER
		do
			create Result
			Result.set_font (a_font)
			l_size := a_font.string_size (a_text)
			w := l_size.integer_item (1).max (a_min_width)
			h := l_size.integer_item (2).max (a_min_height)
			Result.set_size (w, h)
			Result.set_minimum_width (w)
			if b /= Void then
				Result.set_background_color (b)
			end
			if f /= Void then
				Result.set_foreground_color (f)
			end
			Result.clear
			Result.draw_text_top_left (0, 0, a_text)
		end

	image_path (a_filename: STRING): STRING is
			-- Return full image path associated to 'a_filename'
			-- Completed with installation directory if only filename, as-is if hardcoded path specified
			-- Void if 'a_filename' isn't recognized (unknown extension)
		do
			Result := Global.app.icons.get_file (a_filename)
		end

	is_valid_image_filename (a_filename: STRING): BOOLEAN is
			-- Is 'a_filename' a recognized image name?
			-- See supported formats below
		do
			Result := image_path (a_filename) /= Void
		end

	find_label (a_widget: EV_WIDGET): EV_LABEL is
			-- Return label associated to 'a_widget'
			-- If 'a_widget' is already a label, this simply returns it
			-- Otherwise, this tries to find tha lable just in from of 'a_widget' (by examining a_widget.parent)
		local
			l_parent: EV_CONTAINER
			l_box: EV_BOX
		do
			if a_widget /= Void then
				from
					Result ?= a_widget
					l_parent := a_widget.parent
				until
					Result /= Void or else l_parent = Void
				loop
					l_box ?= l_parent
					if l_box /= Void and then not l_box.is_empty then
						Result ?= l_box.first
					end
					l_parent := l_parent.parent
				end
			end
		end

feature -- Basic operations

	embold (a_fontable: EV_FONTABLE; a_bold: BOOLEAN) is
			-- Set 'a_fontable' as bold or not, accordingly to 'a_bold'.
		local
			l_font: EV_FONT
		do
			if a_fontable /= Void then
				l_font := a_fontable.font
				if l_font /= Void then
					if a_bold then
						l_font.set_weight (gui.rep.fonts.weight_black)
					else
						l_font.set_weight (gui.rep.fonts.weight_regular)
					end
					a_fontable.set_font (l_font)
				end
			end
		end

	italicize (a_fontable: EV_FONTABLE; a_italic: BOOLEAN) is
			-- Set 'a_fontable' as italic or not, accordingly to 'a_italic'.
		local
			l_font: EV_FONT
		do
			if a_fontable /= Void then
				l_font := a_fontable.font
				if l_font /= Void then
					if a_italic then
						l_font.set_shape (gui.rep.fonts.shape_italic)
					else
						l_font.set_shape (gui.rep.fonts.shape_regular)
					end
					a_fontable.set_font (l_font)
				end
			end
		end

	align_labels (a_array: ARRAY [EV_WIDGET]) is
			-- Set minimum width of all labels in 'a_array' as same
		require
			array_ok: a_array /= Void
		local
			l_widgets: ROSE_LINEAR_ARRAY [EV_WIDGET]
			l_button: EV_BUTTON
			l_widget: EV_WIDGET
			i, w: INTEGER
		do
			if a_array.count > 1 then
				from
					create l_widgets.make_empty
					i := a_array.lower
				until
					i > a_array.upper
				loop
					l_button ?= a_array.item (i)
					if l_button /= Void then
						l_widget := l_button
					else
						l_widget := find_label (a_array.item (i))
					end
					if l_widget /= Void then
						w := l_widget.width.max (w)
						l_widgets.extend (l_widget)
					end
					i := i + 1
				end
				from
					i := l_widgets.lower
				until
					i > l_widgets.upper
				loop
					l_widget := l_widgets.item (i)
					if l_widget /= Void then
						l_widget.set_minimum_width (w)
					end
					i := i + 1
				end
			end
		end

	align_widgets (a_array: ARRAY [EV_WIDGET]) is
			-- Set minimum width of all widgets in 'a_array' as same
		require
			array_ok: a_array /= Void
		local
			i, w: INTEGER
		do
			if a_array.count > 1 then
				from
					i := a_array.lower
				until
					i > a_array.upper
				loop
					if a_array.item (i) /= Void then
						w := a_array.item (i).width.max (w)
					end
					i := i + 1
				end
				from
					i := a_array.lower
				until
					i > a_array.upper
				loop
					if a_array.item (i) /= Void then
						a_array.item (i).set_minimum_width (w)
					end
					i := i + 1
				end
			end
		end

	set_box_defaults (a_container: EV_CONTAINER; a_box: EV_BOX; a_border, a_padding: INTEGER) is
			-- Set 'a_box' with 'a_border' and 'a_padding', set not homogeeous by default
		do
			a_box.disable_homogeneous
			a_box.set_padding_width (a_padding.abs)
			a_box.set_border_width (a_border.abs)
			add_widget_in_container (a_box, a_container, a_border >= 0 and a_padding >= 0)
		end

	set_list_item_defaults (a_container: EV_MULTI_COLUMN_LIST; a_row:EV_MULTI_COLUMN_LIST_ROW; a_pixmap: ANY;
			a_content: ARRAY [ANY]; a_data: ANY) is
			-- New list-item put in 'a_container', with 'a_pixmap', 'a_content' and 'a_data'
		obsolete
			"Use RP_DATA_GRID features instead"
		local
			l_string: STRING
			l_pixmap: EV_PIXMAP
			i: INTEGER
		do
			l_string ?= a_pixmap
			if l_string /= Void then
				l_pixmap := new_pixmap_from_file (l_string)
			else
				l_pixmap ?= a_pixmap
			end
			if l_pixmap /= Void then
				a_row.set_pixmap (l_pixmap)
			end
			if a_content /= Void then
				from
					i := a_content.lower
				until
					i > a_content.upper
				loop
					a_row.extend (Global.formatter.object_representation (a_content.item (i), ""))
					i := i + 1
				end
			end
			a_row.set_data (a_data)
			if a_container /= Void then
				a_container.extend (a_row)
			end
		end

	set_tree_item_defaults (a_node: EV_TREE_NODE; a_container: EV_TREE_NODE_LIST; a_pixmap: ANY; a_text: ANY; a_data: ANY) is
			-- Set 'a_node' in 'a_container', and attach 'a_pixmap' and 'a_text' to it.
		require
			node_ok: a_node /= Void
		local
			l_string: STRING
			l_pixmap: EV_PIXMAP
		do
			l_string ?= a_pixmap
			if l_string /= Void then
				l_pixmap := new_pixmap_from_file (l_string)
			else
				l_pixmap ?= a_pixmap
			end
			if l_pixmap /= Void then
				a_node.set_pixmap (l_pixmap)
			end
			set_text (a_node, Global.formatter.object_representation (a_text, ""))
			a_node.set_data (a_data)
			if a_container /= Void then
				a_container.extend (a_node)
			end
		end

	add_separator (a_container: EV_CONTAINER) is
			-- Add a separator to 'a_container'.
			-- Works only if 'a_container' supports separators.
		do
			add_separator_text (a_container, Void)
		end

	add_separator_text (a_container: EV_CONTAINER; a_text: STRING) is
			-- Add a separator to 'a_container'.
			-- Works only if 'a_container' supports separators.
			-- If 'a_text' is provided, then a invert-colored label with 'a_text' is added.
		local
			l_vbox: EV_VERTICAL_BOX
			l_hbox: EV_HORIZONTAL_BOX
			l_hsep: EV_HORIZONTAL_SEPARATOR
			l_label: EV_LABEL
		do
			if a_container /= Void then
				l_vbox ?= a_container
				l_hbox ?= a_container
				if l_hbox /= Void then
					add_widget_in_container (create {EV_VERTICAL_SEPARATOR}, a_container, True)
				elseif l_vbox /= Void then
					if a_text = Void or else a_text.is_empty then
						add_widget_in_container (create {EV_HORIZONTAL_SEPARATOR}, a_container, True)
					else
						l_hbox := new_hbox (a_container, 0, 4)
						create l_hsep
						l_hbox.extend (l_hsep)
						l_hbox.disable_item_expand (l_hsep)
						l_hsep.set_minimum_width (8)
						l_label := new_label (l_hbox, a_text, 1)
						l_label.set_foreground_color (Gui.rep.colors.insensitive)
						create l_hsep
						l_hbox.extend (l_hsep)
					end
				end
			end
		end

	enbox_and_size_simple (a_widget: EV_WIDGET; a_container: EV_CONTAINER; a_width: INTEGER) is
			-- Put 'a_windget' in 'a_container' if provided, and set it's minimal size.
			-- See note in class header.
		do
			if a_widget /= Void then
				if a_width.abs > 1 then
					a_widget.set_minimum_width (a_width.abs)
				end
				add_widget_in_container (a_widget, a_container, a_width > 0)
			end
		end

	enbox_and_size (a_widget: EV_WIDGET; a_container: EV_CONTAINER; a_width, a_height: INTEGER) is
			-- Put 'a_windget' in 'a_container' if provided, and set it's minimal size.
			-- See note in class header.
		local
			l_hbox: EV_HORIZONTAL_BOX
			l_vbox: EV_VERTICAL_BOX
			l_fixed: BOOLEAN
		do
			if a_widget /= Void then
				if a_width.abs > 1 then
					a_widget.set_minimum_width (a_width.abs)
				end
				if a_height.abs > 1 then
					a_widget.set_minimum_height (a_height.abs)
				end
				l_hbox ?= a_container
				l_vbox ?= a_container
				if l_hbox /= Void then
					l_fixed := a_width > 0
				elseif l_vbox /= Void then
					l_fixed := a_height > 0
				else
					l_fixed := a_width > 0 and then a_height > 0
				end
				add_widget_in_container (a_widget, a_container, l_fixed)
			end
		end

	populate_combo_with_list (a_combo: EV_LIST_ITEM_LIST; a_linear: LINEAR [ANY]) is
			-- Populate 'a_combo' with 'a_linear'.
		local
			l_item: EV_LIST_ITEM
		do
			if a_combo /= Void and then a_linear /= Void then
				a_combo.wipe_out
				from
					a_linear.start
				until
					a_linear.off
				loop
					if a_linear.item /= Void then
						create l_item.make_with_text (Global.formatter.object_representation (a_linear.item, ""))
						l_item.set_data (a_linear.item)
						a_combo.extend (l_item)
					end
					a_linear.forth
				end
			end
		end

	populate_combo_with_array (a_combo: EV_LIST_ITEM_LIST; a_array: ARRAY [ANY]) is
			-- Populate 'a_combo' with 'a_array'.
		local
			i: INTEGER
			l_item: EV_LIST_ITEM
		do
			if a_combo /= Void and then a_array /= Void then
				a_combo.wipe_out
				from
					i := a_array.lower
				until
					i > a_array.upper
				loop
					if a_array.item (i) /= Void then
						create l_item.make_with_text (Global.formatter.object_representation (a_array.item (i), ""))
						l_item.set_data (a_array.item (i))
						a_combo.extend (l_item)
					end
					i := i + 1
				end
			end
		end

	set_info_action (a_widget: EV_WIDGET; a_string: STRING) is
			-- Attach an action showing an info dialog when 'a_widget' is clicked
			-- The action is to popup an untitled dialog that disappears when it loses focus or when ESC is pressed
			-- The popup is shown when 'a_widget' is left-clicked only if 'a_widget.data' contains a STRING at that time
			-- or if a non-empty 'a_string' is provided.
		do
			a_widget.pointer_button_release_actions.extend (agent on_widget_info (a_widget, a_string, ?, ?, ?, ?, ?, ?, ?, ?))
		end

	stretch_pixmap (a_pixmap: EV_PIXMAP; a_width, a_height: INTEGER) is
			-- Stretch 'a_pixmap' to 'a_width' and 'a_height'
			-- This call is protected because of a bug in Vision2 failing on some pixmaps
		local
			retried: BOOLEAN
		do
			if not retried and then a_pixmap /= Void
			and then (a_pixmap.width /= a_width or else a_pixmap.height /= a_height) then
				a_pixmap.stretch (a_width, a_height)
			end
		rescue
			if not retried then
				retried := True
				retry
			end
		end

feature {NONE} -- Implementation

	add_labeled_widget (a_container: EV_CONTAINER; a_widget: EV_WIDGET; a_label: STRING; a_width: INTEGER) is
			-- Add 'a_widget' (with 'a_width) to 'a_container', precede it with 'a_label' if given (label is put in a horizontal box, right before the widget).
		local
			l_hbox: EV_HORIZONTAL_BOX
			l_vbox: EV_VERTICAL_BOX
			l_label: EV_LABEL
			l_padding: INTEGER
		do
			if a_container /= Void and then a_label /= Void and then not a_label.is_empty then
				l_hbox ?= a_container
				if l_hbox = Void then
					l_vbox ?= a_container
					if l_vbox /= Void then l_padding := l_vbox.padding else l_padding := 4 end
					if a_width <= 0 then l_padding := -l_padding end
					l_hbox := new_hbox (a_container, 0, l_padding)
				end
				l_label := new_label (l_hbox, a_label, 1)
				enbox_and_size_simple (a_widget, l_hbox, a_width)
			else
				enbox_and_size_simple (a_widget, a_container, a_width)
			end
		end

	add_widget_in_container (a_widget: EV_WIDGET; a_container: EV_CONTAINER; a_fixed: BOOLEAN) is
			-- Add 'a_widget' to 'a_container' and make it not auto-resizable if 'a_fixed' is True.
		local
			l_box: EV_BOX
			l_split, l_split2: EV_SPLIT_AREA
			l_second: EV_WIDGET
		do
			if a_container /= Void and then a_widget /= Void then
				l_split ?= a_container
				if l_split /= Void then
					l_split := find_last_split (l_split)
					if l_split = Void then
						-- Do nothing
					elseif l_split.extendible then
						l_split.extend (a_widget)
					else
						l_second := l_split.second
						l_split2 := new_same_split_type (l_split)
						if l_split2 /= Void then
							l_split.prune (l_second)
							l_split.extend (l_split2)
							l_split2.extend (l_second)
							l_split2.extend (a_widget)
						end
					end
				else
					a_container.extend (a_widget)
					l_box ?= a_container
					if l_box /= Void and then a_fixed then
						l_box.disable_item_expand (a_widget)
					end
				end
			end
		end

	find_last_split (a_split: EV_SPLIT_AREA): EV_SPLIT_AREA is
			-- Find last split within 'a_split'
		require
			split_ok: a_split /= Void
		do
			if a_split.second /= Void and then a_split.same_type (a_split.second) then
				Result ?= a_split.second
			end
			if Result = Void then
				Result := a_split
			else
				Result := find_last_split (Result)
			end
		ensure
			result_ok: Result /= Void
		end

	new_same_split_type (a_split: EV_SPLIT_AREA): EV_SPLIT_AREA is
			--New split of same type as 'a_split'
		require
			split_ok: a_split /= Void
		local
			l_hsplit: EV_HORIZONTAL_SPLIT_AREA
			l_vsplit: EV_VERTICAL_SPLIT_AREA
		do
			l_hsplit ?= a_split
			l_vsplit ?= a_split
			if l_hsplit /= Void then
				create {EV_HORIZONTAL_SPLIT_AREA} Result
			elseif l_vsplit /= Void then
				create {EV_VERTICAL_SPLIT_AREA} Result
			end
		ensure
			result_ok: Result /= Void
		end

	on_list_select (a_button: INTEGER; a_action: PROCEDURE [ANY, TUPLE [INTEGER, EV_MULTI_COLUMN_LIST_ROW]]; a_row: EV_MULTI_COLUMN_LIST_ROW) is
			-- Called when a list row is (de)selected.
		obsolete
			"Use RP_DATA_GRID features instead"
		do
			if a_row /= Void then
				a_action.call ([a_button, a_row])
			end
		end

	on_list_click (a_action: PROCEDURE [ANY, TUPLE [INTEGER, EV_MULTI_COLUMN_LIST_ROW]]; a_list: EV_MULTI_COLUMN_LIST; a_double_click: BOOLEAN; a_x, a_y, a_button: INTEGER; a_x_tilt, a_y_tilt, a_pressure: DOUBLE; a_screen_x, a_screen_y: INTEGER) is
			-- Agent called when 'a_list' is clicked.
		obsolete
			"Use RP_DATA_GRID features instead"
		local
			l_button: INTEGER
			l_row: EV_MULTI_COLUMN_LIST_ROW
		do
			if a_double_click then
				l_button := a_button
			elseif a_button > 1 then
				l_button := -a_button
			end
			if l_button /= 0 then
				l_row := a_list.selected_item
				if l_row /= Void then
					a_action.call ([l_button, l_row])
				end
			end
		end

	on_grid_select (a_button: INTEGER; a_action: PROCEDURE [ANY, TUPLE [INTEGER, RP_GRID_NODE[ANY]]]; a_row: EV_GRID_ROW) is
			-- Called when a list row is (de)selected.
		local
			l_node: RP_GRID_NODE[ANY]
		do
			l_node ?= a_row.data
			if l_node /= Void then
				a_action.call ([a_button, l_node])
			end
		end

	on_grid_click (a_action: PROCEDURE [ANY, TUPLE [INTEGER, RP_GRID_NODE[ANY]]]; a_grid: RP_DATA_GRID[ANY]; a_double_click: BOOLEAN; a_x, a_y, a_button: INTEGER; a_x_tilt, a_y_tilt, a_pressure: DOUBLE; a_screen_x, a_screen_y: INTEGER) is
			-- Agent called when 'a_grid' is clicked.
		local
			l_button: INTEGER
			l_selected_nodes: ARRAYED_LIST[RP_GRID_NODE[ANY]]
		do
			if a_double_click then
				l_button := a_button
			elseif a_button > 1 then
				l_button := -a_button
			end
			if l_button /= 0 then
				l_selected_nodes := a_grid.selected_nodes
				if not l_selected_nodes.is_empty then
					a_action.call ([l_button, l_selected_nodes.first])
				end
			end
		end

	set_split_defaults (a_container: EV_CONTAINER; a_split: EV_SPLIT_AREA; a_widgets: ARRAY [ANY]; a_horizontal: BOOLEAN) is
			-- Put all 'a_widgets' in 'a_split' (make several if necessary), and set it's split positions.
			-- Finally, put 'a_split' in 'a_container'
		local
			l_widget_list: LINKED_LIST [PAIR [EV_WIDGET, INTEGER]]
			l_previous_pair, l_pair: PAIR [EV_WIDGET, INTEGER]
			l_current_split, l_new_split: EV_SPLIT_AREA
		do
			l_widget_list := resolved_widget_list (a_widgets)
			if l_widget_list = Void or else l_widget_list.is_empty then
				--| Do nothing
			else
				from
					l_widget_list.start
					l_current_split := a_split
					l_pair := l_widget_list.item
					l_current_split.set_first (l_pair.first)
					l_widget_list.forth
				until
					l_widget_list.off
				loop
					l_previous_pair := l_pair
					l_pair := l_widget_list.item
					l_widget_list.forth
					if l_widget_list.off then
						l_current_split.set_second (l_pair.first)
						if l_previous_pair.second > 0 then
							l_current_split.set_split_position (l_previous_pair.second.abs)
						end
					else
						if a_horizontal then
							create {EV_HORIZONTAL_SPLIT_AREA} l_new_split
						else
							create {EV_VERTICAL_SPLIT_AREA} l_new_split
						end
						l_current_split.set_second (l_new_split)
						if l_previous_pair.second > 0 then
							l_current_split.set_split_position (l_previous_pair.second.abs)
						end
						l_current_split := l_new_split
						l_current_split.set_first (l_pair.first)
					end
				end
			end
			add_widget_in_container (a_split, a_container, False)
		end

	resolved_widget_list (a_widgets: ARRAY [ANY]): LINKED_LIST [PAIR [EV_WIDGET, INTEGER]] is
			-- Return list of (widget, size) pairs contained in 'a_widgets'
		local
			i, l_size: INTEGER
			l_size_ref: INTEGER_REF
			l_widget: EV_WIDGET
			l_tuple: TUPLE [EV_WIDGET, INTEGER]
		do
			if a_widgets /= Void then
				from
					create Result.make
					i := a_widgets.lower
				until
					i > a_widgets.upper
				loop
					if a_widgets.item (i) /= Void then
						l_size_ref := Void
						l_widget ?= a_widgets.item (i)
						if l_widget = Void then
							l_tuple ?= a_widgets.item (i)
							if l_tuple /= Void then
								l_widget ?= l_tuple.item (1)
								l_size_ref ?= l_tuple.item (2)
							end
						end
						if l_widget /= Void then
							if l_size_ref = Void then
								l_size := 0
							else
								l_size := l_size_ref.item
							end
							Result.extend (create {PAIR [EV_WIDGET, INTEGER]}.make (l_widget, l_size))
						end
					end
					i := i + 1
				end
			end
		end

	set_text (a_textable: EV_TEXTABLE; a_text: STRING) is
			-- This function is introduced because Vision2 crashes when 'a_text'
			-- contains '%R' characters...
		do
			if a_textable /= Void then
				if a_text = Void then
					a_textable.remove_text
				else
					cleanup (a_text)
					a_textable.set_text (a_text)
				end
			end
		end

	cleanup (a_string: STRING) is
			-- Clean up characters not accepted by Vision2 from 'a_string'
			-- FIXME: Temporary fix for Vision2
		do
			if a_string /= Void then
				a_string.prune_all ('%R')
			end
		end

	on_widget_info (a_widget: EV_WIDGET; a_string: STRING; a_x, a_y, a_button: INTEGER; a_x_tilt, a_y_tilt, a_pressure: DOUBLE; a_screen_x, a_screen_y: INTEGER) is
			-- Pop up a small untitled dialog that disappears when it loses focus or when ESC is pressed.
			-- The popup is shown only if 'a_widget.data' contains a STRING, or if a non-empty 'a_string' is provided.
		local
			l_string: STRING
		do
			if a_button = 1 and then a_widget /= Void then
				l_string ?= a_widget.data
				if l_string = Void or else l_string.is_empty then
					l_string := a_string
				end
				if l_string /= Void and then not l_string.is_empty then
					Gui.dialogs.popup_info (l_string, Void)
				end
			end
		end

	on_agent_click (a_agent: PROCEDURE [ANY, TUPLE []]; a_x, a_y, a_button: INTEGER; a_x_tilt, a_y_tilt, a_pressure: DOUBLE; a_screen_x, a_screen_y: INTEGER) is
			-- Proxy routine to call 'a_agent' when a click event occurs.
		do
			a_agent.call ([])
		end

	scaled_pixmap (a_pixmap: EV_PIXMAP; a_height: INTEGER): EV_PIXMAP is
			-- pixmap to a scaled height of 16 pixels
		local
			l_factor: DOUBLE
		do
			if a_pixmap /= Void then
				create Result
				Result.copy (a_pixmap)
				l_factor := a_height / Result.height
				Result.stretch ((l_factor * Result.width).rounded, a_height)
			end
		end

end
