note
	description: "Color repository"
	author: "Zoran Simic"
	last_editor: "$Author: zs $"
	date: "$Date: 2011/05/13 $"
	revision: "$Revision: #17 $"
	path: "$File: //rose/source/kernel/infrastructure/framework/gui/rose_ev_repository_colors.e $"

class
	ROSE_EV_REPOSITORY_COLORS

inherit
	EV_STOCK_COLORS
	ROSE_SHARED_GLOBAL
		export
			{NONE} all
		end

feature -- Access

	highlit (a_delta: INTEGER; a_color: EV_COLOR): EV_COLOR
			-- 'a_color', highlit to show that it is being "hovered over"
		require
			delta_ok: a_delta > 0 and then a_delta < 100
			color_ok: a_color /= Void
		local
			l_brightness: DOUBLE
		do
			l_brightness := (a_color.red_8_bit + a_color.green_8_bit + a_color.blue_8_bit) / 3
			if l_brightness >= (255 - a_delta) then
				-- Color is extremely light, we darken it
				Result := emphasized (-a_delta, a_color)
			else
				-- We lighten it
				Result := emphasized (a_delta, a_color)
			end
		end

	emphasize, emphasized (a_delta: INTEGER; a_color: EV_COLOR): EV_COLOR
			-- Translate all 3 RGB components of 'a_color' by 'a_delta'
		local
			lr, lg, lb: INTEGER
		do
			if a_color /= Void then
				lr := a_color.red_8_bit + a_delta
				lg := a_color.green_8_bit + a_delta
				lb := a_color.blue_8_bit + a_delta
				Result := rgb (lr, lg, lb)
			end
		end

	blended (a_factor: INTEGER; c1, c2: EV_COLOR): EV_COLOR
			-- Blend 'c1' and 'c2' to form a new color, 'a_factor' indicates the degree of blending desired (linear function)
			-- 'a_factor' of -100 gives 'c1', +100 gives 'c2', 0 gives a uniform blending of 'c1' and 'c2'
		do
			if c1 = Void then
				Result := c2
			elseif c2 = Void then
				Result := c1
			else
				Result := rgb (
					blended_component (a_factor, c1.red_8_bit, c2.red_8_bit),
					blended_component (a_factor, c1.green_8_bit, c2.green_8_bit),
					blended_component (a_factor, c1.blue_8_bit, c2.blue_8_bit)
				)
			end
		end

	configured (a_key: STRING): attached EV_COLOR
			-- Configured color under 'a_key', returns 'default_foreground_color' if nothing configured under 'a_key'
		do
			if a_key /= Void then
				Result := rgbs (config.value ("color/" + a_key))
			else
				Result := default_foreground_color
			end
		end

	rgb (r, g, b: INTEGER): attached EV_COLOR
			-- Color from given RGB components
		do
			create Result.make_with_8_bit_rgb (limited (r), limited (g), limited (b))
		end

	rgbs (a_description: STRING): attached EV_COLOR
			-- Color from 'a_description' expected to be RGB compononents comma separated
			-- Void if invalid description
		local
			l_array: ARRAY [STRING]
		do
			if attached {EV_COLOR} named.item (a_description) as l_color then
				Result := l_color
			else
				l_array := Global.formatter.split_array (a_description, component_separator, Void, False)
				Result := rgb (get_component (l_array, 1), get_component (l_array, 2), get_component (l_array, 3))
			end
		end

	random_color (a_index: INTEGER): attached EV_COLOR
			-- Random color corresponding to 'a_index'
		local
			i: INTEGER
		do
			i := a_index.max (1)
			Result := rgb (i * 197 \\ 256, i * 229 \\ 256, i * 307 \\ 256)
		end

	named: attached HASH_TABLE [EV_COLOR, STRING]
			-- Table of named colors
		once
			create Result.make (0)
			Result.put (White, "white")
			Result.put (Black, "black")
			Result.put (Gray, "gray")
			Result.put (Dark_gray, "dark_gray")
			Result.put (Blue, "blue")
			Result.put (Dark_blue, "dark_blue")
			Result.put (Cyan, "cyan")
			Result.put (Dark_cyan, "dark_cyan")
			Result.put (Green, "green")
			Result.put (Dark_green, "dark_green")
			Result.put (Yellow, "yellow")
			Result.put (Dark_yellow, "dark_yellow")
			Result.put (Light_yellow, "light_yellow")
			Result.put (Red, "red")
			Result.put (Dark_red, "dark_red")
			Result.put (Magenta, "magenta")
			Result.put (Dark_magenta, "dark_magenta")
			Result.put (Light_red, "light_red")
			Result.put (Light_green, "light_green")
			Result.put (Brown, "brown")
			Result.put (Orange, "orange")
			Result.put (light_orange, "light_orange")
			Result.put (Rose, "rose")
			Result.put (Light_blue, "light_blue")
			Result.put (Light_gray, "light_gray")
			Result.put (Med_green, "medium_green")
		end

	Light_red: attached EV_COLOR
			-- Light red
		once
			Result := rgb (255, 229, 229)
		end

	Light_yellow: attached EV_COLOR
			-- Light yellow
		once
			Result := rgb (255, 255, 192)
		end

	light_orange: attached EV_COLOR
			-- Light yellow
		once
			Result := rgb (255, 204, 102)
		end

	light_green: attached EV_COLOR
			-- Light green
		once
			Result := rgb (229, 255, 229)
		end

	Med_green: attached EV_COLOR
			-- Medium green
		once
			Result := rgb (170, 255, 170)
		end

	Brown: attached EV_COLOR
			-- Brown
		once
			Result := rgb (166, 42, 0)
		end

	Orange: attached EV_COLOR
			-- Orange
		once
			Result := rgb (255, 128, 42)
		end

	Rose: attached EV_COLOR
			-- Rose (pink)
		once
			Result := rgb (133, 99, 99)
		end

	Light_blue: attached EV_COLOR
			-- Light blue
		once
			Result := rgb (229, 229, 255)
		end

	Light_gray: attached EV_COLOR
			-- Light gray
		once
			Result := rgb (229, 229, 229)
		end

	Insensitive: attached EV_COLOR
			-- Color representing an insensitive widget
		once
			Result := Color_3d_shadow
		end

feature {NONE} -- Implementation

	component_separator: STRING = ","		-- Component separator

	blended_component (a_factor, c1, c2: INTEGER): INTEGER
			-- Blend components 'c1' and 'c2' to form a new color component, 'a_factor' indicates the degree of blending desired (linear function)
			-- 'a_factor' of -100 gives 'c1', +100 gives 'c2', 0 gives a uniform blending of 'c1' and 'c2'
		local
			d: DOUBLE
		do
			d := ( (a_factor / 100) * (c2 - c1) + c1 + c2 ) / 2
			Result := d.truncated_to_integer
		end

	get_component (a_array: ARRAY [STRING]; a_index: INTEGER): INTEGER
			-- Component at 'a_index' in 'a_array'
		local
			l_string: STRING
			i: INTEGER
		do
			if a_array /= Void then
				i := a_index - a_array.lower + 1
				if a_array.valid_index (i) then
					l_string := a_array.item (i)
				end
			end
			Result := to_int (l_string)
			if Result < 0 then
				Result := 0
			end
			if Result > 255 then
				Result := 255
			end
		end

	to_int (a_string: STRING): INTEGER
			-- Integer contained in 'a_string'
		do
			if a_string /= Void and then a_string.is_integer then
				Result := a_string.to_integer
			end
		end

	limited (a_component: INTEGER): INTEGER
			-- Return `a_component' bounded by 0 and 255.
		do
			Result := a_component.max (0).min (255)
		end

end
