note
	description: "Information on a given executable (info is embedded within the exe)"
	author: "Mauricio Bustos"
	last_editor: "$Author: an $"
	date: "$Date: 2013/04/22 $"
	revision: "$Revision: #27 $"
	path: "$File: //rose/source/kernel/infrastructure/framework/application/rose_executable_info.e $"

class
	ROSE_EXECUTABLE_INFO

inherit
	ROSE_FORMATTING
		redefine
			out
		end
	DEBUG_OUTPUT
		redefine
			out
		end
	MISMATCH_CORRECTOR
		redefine
			correct_mismatch,
			out
		end
create
	make

feature {NONE} -- Initialization

	make (a_path: STRING)
			-- Create info object on exe with 'a_path'.
		require
			path_ok: a_path /= Void and then not a_path.is_empty
		do
			path := a_path
			version := k_default_version
			name := formatter.file_name_no_extension (a_path)
			description := k_default_description
			internal_name := formatter.file_name (a_path)
			service_center := "debug"
		end

feature -- Access

	path: STRING
			-- Path to executable to which these info relates

	version: STRING
			-- Version of the compiled application (example: "1,0,2006,0823")

	scm_root: STRING
			-- SCM root against which exe was built
			-- May be comma-separated list in the case where several SCM roots were used (example: //rose/research_center,//rose/source)

	revision: INTEGER
			-- Revision number of the source tree with which the exe was built

	name: STRING
			-- Name of the application as defined in TASK_TEMPLATE

	description: STRING
			-- Description provided in TASK_TEMPLATE

	internal_name: STRING
			-- name.exe of the program

	icon: STRING
			-- Name of icon associated to this exe, if any

	service_center: STRING
			-- Service center of the application (example: "supplier\market\idc")

	contacts: STRING
			-- People to contact for more info on this application (example: "MZ,MB")

	impact: STRING
			-- Application impact (critical, compatible...)

	handle: INTEGER
			-- Handle of corresponding TASK_TEMPLATE entity that defined the info from this object

	template: STRING
			-- Template

	root_class: STRING
			-- Root class

	root_feature: STRING
			-- Root feature

	is_mt: BOOLEAN
			-- Is multithreaded?

	is_gui: BOOLEAN
			-- Is this exe a GUI?

	build_date: ROSE_DATE
			-- UTC timestamp when build occured

	build_snapshot: STRING
			-- Snapshot used to build this exe (Void if was not part of weekly rebuild)

	build_ec: STRING
			-- Full path to compiler that created this exe (example: "C:\EiffelStudio\5.6.1218")

	build_estudio: STRING
			-- Estudio compiler version that created this exe (example: "5.6.1218")

	build_compatible: BOOLEAN
			-- Was the Eiffel compiler used in compatible mode?

	build_exception_trace: BOOLEAN
			-- Was exe compiled with exception_trace on?

	build_debug: STRING
			-- Comma separated list of debug flags used during the compilation

	build_platform: STRING
			-- Platform used to build this exe (example: "win32" or "dotnet")

	build_c: STRING
			-- C compiler used to build this exe (example: "vc7")

	build_author: STRING
			-- Username under which the build occured

	build_host: STRING
			-- Name of the machine where this exe was built

	build_dir: STRING
			-- Directory on 'build_host' where build occured

	build_ticket: STRING
			-- Tickets associated with the build

	company_name: STRING
			-- "AXA Rosenberg"

	overrides: ARRAY [STRING]
			-- Override folders used

	overrides_count: INTEGER
			-- Number of overrides used
		do
			if overrides /= Void then
				Result := overrides.count
			end
		end

	out: STRING
			-- Textual representation
		do
			create Result.make (512)
			disp (Result, "ExePath", path)
			disp (Result, k_file_version, version)
			disp (Result, k_icon, icon)
			disp (Result, k_service_center, service_center)
			disp (Result, k_contacts, contacts)
			disp (Result, k_handle, handle.out)
			disp (Result, k_root_class, root_class)
			disp (Result, k_root_feature, root_feature)
			if build_date /= Void then
				disp (Result, k_build_date, build_date.formatted_out ("yyyy/[0]mm/[0]dd hh:[0]mi:[0]ss"))
			end
			disp (Result, k_build_snapshot, build_snapshot)
			disp (Result, k_build_ec, build_ec)
			disp (Result, k_build_estudio, build_estudio)
			disp (Result, k_build_platform, build_platform)
			disp (Result, k_build_c, build_c)
			disp (Result, k_build_author, build_author)
			disp (Result, k_build_host, build_host)
			disp (Result, k_build_dir, build_dir)
			disp (Result, k_build_scm, scm_root)
			disp (Result, k_build_revision, revision.out)
		end

	debug_output: STRING
			-- String that should be displayed in debugger to represent `Current'.
		do
			Result := path
		end

feature {NONE} -- Storable

	correct_mismatch
			-- <Precursor>
		do
			-- Ignore unknown and missing attributes.
		end

feature -- Constants

	k_name: STRING = "Name"
	k_icon: STRING = "Icon"
	k_service_center: STRING = "ServiceCenter"
	k_contacts: STRING = "Contacts"
	k_handle: STRING = "Handle"
	k_root_class: STRING = "RootClass"
	k_root_feature: STRING = "RootFeature"
	k_impact: STRING = "Impact"
	k_template: STRING = "Template"
	k_is_mt: STRING = "IsMT"
	k_is_gui: STRING = "IsGUI"
	k_override: STRING = "Override"
	k_build_date: STRING = "BuildDate"
	k_build_snapshot: STRING = "BuildSnapshot"
	k_build_ec: STRING = "BuildEc"
	k_build_estudio: STRING = "BuildEstudio"
	k_build_experimental: STRING = "BuildExperimental"
	k_build_compatible: STRING = "BuildCompatible"
	k_build_exception_trace: STRING = "BuildExceptionTrace"
	k_build_debug: STRING = "BuildDebug"
	k_build_platform: STRING = "BuildPlatform"
	k_build_c: STRING = "BuildC"
	k_build_author: STRING = "BuildAuthor"
	k_build_host: STRING = "BuildHost"
	k_build_dir: STRING = "BuildDir"
	k_build_scm: STRING = "BuildSCM"
	k_build_revision: STRING = "BuildRevision"
	k_build_ticket: STRING = "BuildTicket"
	k_file_version: STRING = "FileVersion"
	k_file_description: STRING = "FileDescription"
	k_internal_name: STRING = "InternalName"
	k_company_name: STRING = "CompanyName"
	k_legal_copyright: STRING = "LegalCopyright"

feature -- Basic operations

	set_version (a_version: STRING)
			-- Set 'version'.
		require
			version_ok: a_version /= Void and not a_version.is_empty
		do
			version := a_version
		end

	set_scm_root (a_scm_root: STRING)
			-- Set 'scm_root'.
		do
			scm_root := a_scm_root
		end

	set_revision_from_string (a_revision: STRING)
			-- Set 'revision'
		do
			if a_revision /= Void and then a_revision.is_integer then
				revision := a_revision.to_integer
			end
		end

	set_name (a_name: STRING)
			-- Set 'name'.
		require
			name_ok: a_name /= Void and then not a_name.is_empty
		do
			name := a_name.as_lower
			name.left_adjust
			name.right_adjust
		end

	set_description (a_description: STRING)
			-- Set 'description'.
		require
			description_ok: a_description /= Void and then not a_description.is_empty
		do
			description := a_description
		end

	set_internal_name (a_internal_name: STRING)
			-- Set 'internal_name'.
		require
			internal_name_ok: a_internal_name /= Void and then not a_internal_name.is_empty
		do
			internal_name := a_internal_name
		end

	set_icon (a_icon: STRING)
			-- Set 'icon' to 'a_icon'.
		do
			icon := a_icon
		end

	set_service_center (a_service_center: STRING)
			-- Set 'service_center' to 'a_service_center'.
		do
			service_center := a_service_center
		end

	set_contacts (a_contacts: STRING)
			-- Set 'contacts' to 'a_contacts'.
		do
			contacts := a_contacts
		end

	set_impact (a_impact: like impact)
			-- Set 'impact'.
		do
			impact := a_impact
		end

	set_handle_from_string (a_handle_string: STRING)
			-- Set 'handle' from integer contained in 'a_handle_string',
		do
			if a_handle_string /= Void and then a_handle_string.is_integer then
				handle := a_handle_string.to_integer
			end
		end

	set_template (a_template: STRING)
			-- Set 'template'.
		do
			template := a_template
		end

	set_root_class (a_root_class: STRING)
			-- Set 'root_class'.
		do
			root_class := a_root_class
		end

	set_root_feature (a_root_feature: STRING)
			-- Set 'root_feature'.
		do
			root_feature := a_root_feature
		end

	set_mt (a_string: STRING)
			-- Set 'is_mt'.
		do
			is_mt := a_string /= Void and then a_string.is_boolean and then a_string.to_boolean
		end

	set_gui (a_string: STRING)
			-- Set 'is_gui'.
		do
			is_gui := a_string /= Void and then a_string.is_boolean and then a_string.to_boolean
		end

	set_build_date_string (a_string: STRING)
			-- Set 'build_date' from date represented by 'a_string'.
		do
			build_date := formatter.make_date_from_string (a_string)
		end

	set_build_date (a_build_date: ROSE_DATE)
			-- Set 'build_date' to 'a_build_date'.
		do
			build_date := a_build_date
		end

	set_build_snapshot (a_build_snapshot: STRING)
			-- Set 'build_snapshot' to 'a_build_snapshot'.
		do
			build_snapshot := a_build_snapshot
		end

	set_build_ec (a_build_ec: STRING)
			-- Set 'build_ec' to 'a_build_ec'.
		do
			build_ec := a_build_ec
		end

	set_build_estudio (a_build_estudio: STRING)
			-- Set 'build_estudio' to 'a_build_estudio'.
		do
			build_estudio := a_build_estudio
		end

	set_build_compatible (a_string: STRING)
			-- Set 'build_compatible'.
		do
			build_compatible := a_string /= Void and then a_string.is_boolean and then a_string.to_boolean
		end

	set_build_exception_trace (a_string: STRING)
			-- Set `build_exception_trace'.
		do
			build_exception_trace := a_string /= Void and then a_string.is_boolean and then a_string.to_boolean
		end

	set_build_debug (a_build_debug: STRING)
			-- Set 'build_debug' to 'a_build_debug'.
		do
			build_debug := a_build_debug
		end

	set_build_platform (a_build_platform: STRING)
			-- Set 'build_platform' to 'a_build_platform'.
		do
			build_platform := a_build_platform
		end

	set_build_c (a_build_c: STRING)
			-- Set 'build_c' to 'a_build_c'.
		do
			build_c := a_build_c
		end

	set_build_author (a_build_author: STRING)
			-- Set 'build_author' to 'a_build_author'.
		do
			build_author := a_build_author
		end

	set_build_host (a_build_host: STRING)
			-- Set 'build_host' to 'a_build_host'.
		do
			build_host := a_build_host
		end

	set_build_ticket (a_build_ticket: STRING)
			-- Set `build_ticket' to `a_build_ticket'.
		do
			build_ticket := a_build_ticket
		ensure
			build_ticket_set: build_ticket = a_build_ticket
		end

	set_build_dir (a_build_dir: STRING)
			-- Set 'build_dir' to 'a_build_dir'.
		do
			build_dir := a_build_dir
		end

	set_company_name (a_company_name: STRING)
			-- Set 'company_name' to 'a_company_name'.
		do
			company_name := a_company_name
		end

	add_override (a_override: STRING)
			-- Add 'a_override' to 'overrides'
		do
			if overrides = Void then
				create overrides.make_filled (a_override, 1, 1)
			else
				overrides.force (a_override, overrides.upper + 1)
			end
		end

feature {NONE} -- Implementation

	k_default_version: STRING = "0.0.0.0"
	k_default_description: STRING = "No application description provided"

	disp (a_string: STRING; a_name: STRING; a_object: ANY)
			-- Display vlaue 'a_object' with 'a_name' in 'a_string'.
		require
			string_ok: a_string /= Void
			name_ok: a_name /= Void
		do
			a_string.append (formatter.argumented ("$1: $2%N", <<a_name, a_object>>))
		end

invariant

	path_ok: path /= Void
	version_ok: version /= Void			-- May be k_default_version though
	name_ok: name /= Void
	description_ok: description /= Void
	internal_name_ok: internal_name /= Void

end
