note
	description: "Custom extension of EXECUTION_ENVIRONMENT (Windows classic version)"
	notice: "[
		This class contains only the platform-dependent extensions to the Eiffel Base library EXECUTION_ENVIRONMENT.
		Please add any new platform-independent functionality to ROSE_EXECUTION_ENVIRONMENT_COMMON.
	]"
	author: "Zoran Simic"
	last_editor: "$Author: an $"
	date: "$Date: 2013/04/22 $"
	revision: "$Revision: #69 $"
	path: "$File: //rose/source/kernel/infrastructure/support/windows/rose_execution_environment.e $"

class
	ROSE_EXECUTION_ENVIRONMENT

inherit
	ROSE_EXECUTION_ENVIRONMENT_COMMON
		rename
			get as original_get,
			put as original_put
		redefine
			is_administrator,
			canonical_host_name,
			file_owner,
			explore,
			explore_url
		end

feature -- Access

	is_administrator: BOOLEAN
			-- Is the user running current process an administrator on the machine running the process
		once
			Result := c_is_admin
		end

	folders: attached WINDOWS_SPECIAL_FOLDERS
			-- Windows special folders
		once
			create Result
		end

	is_64bit_os: BOOLEAN
			-- Are we running on a 64bit OS?
		once
			Result := size_of_pointer = 8 or else c_is_64bit_os
		end

	cpu_ticks: INTEGER
			-- CPU ticks since process started
		external
			"C"
		alias
			"rose_clock"
		end

	executable_path: attached STRING
			-- Path to the executable running this process
		once
			-- Note that this is neither the path to the executable nor it's full name.
			-- Instead it is the command that launched the program.
			-- This could be fixed by using 'GetProcessImageFileName'
			-- instead of arg[0], but that would require, yet another, library - psapi.lib.
			create Result.make_empty
			Result.append_string (command_line.command_name)
			if not Result.as_lower.has_substring (k_exe) then
				Result.append (k_exe)
			end
			if not Result.has ('\') and then not Result.has ('/') then
				Result.prepend_character ('\')
				Result.prepend_string (current_working_directory)
			end
		end

	reg: attached ROSE_REGISTRY
			-- Access to Windows registry (not implemented on VMS, will return always Void)
		once
			create Result
		end

	get (a_key: STRING): STRING
			-- Value of `a_key' if it is an environment variable and has been set;
			-- Void otherwise.
		do
			if a_key /= Void and then not a_key.is_empty then
				Result := original_get (a_key)
			end
		end

	put (a_value, a_key: STRING)
			-- Set the environment variable `a_key' to `a_value'.
		do
			if a_key /= Void and then not a_key.is_empty then
				if a_value /= Void then
					original_put (a_value, a_key)
				else
					original_put ("", a_key)
				end
			end
		end

	temporary_folder: attached STRING
			-- Path to temporary folder on this machine
		once
			create Result.make_empty
			Result.append_string (folders.local_appdata.as_lower)
			Result.replace_substring_all ("\application data", "\temp")
		end

	current_user: attached ROSE_USER
			-- Current user
		once
			create Result.make (nt_info.user_domain, nt_info.user_login)
			Result.set_display_name (nt_info.user_display_name)
			Result.set_active_directory_location (nt_info.user_active_directory_path)
			Result.set_unique_id (nt_info.user_unique_id)
			Result.set_spn (nt_info.user_spn)
		end

	host_name: attached STRING
			-- Hostname of this machine
		local
			l_string: STRING
		once
			create Result.make_empty
			l_string := nt_info.host_name
			if l_string /= Void and then not l_string.is_empty then
				Result.append_string (l_string.as_lower)
			else
				Result.append_string ("localhost")
			end
		end

	canonical_host_name: attached STRING
			-- Canonical name of current machine, as defined by Systems group
			-- examples: axaros.com/Orinda/Servers/VMware/Patch/ORPCB08
			--           axaros.com/Orinda/Desktops/AXA-ROS Citrix only/FFCTX03
			--           axaros.com/Fairfield/Servers/Externaly Accessible Citrix Servers/Patch/VMSVA61
			--           axaros.com/Orinda/Servers/Other-SystemsControl/Patch/ORSVQ05
		once
			create Result.make_empty
			Result.append_string (nt_info.host_active_directory_path)
			Result.append_character ('/')
			Result.append_string (host_name)
		end

	file_owner (a_path: STRING): STRING
			-- File owner for file with 'a_path', if possible to obtain it
		local
			l_path, l_name, l_domain: C_STRING
			lpSecurityBuffer: MANAGED_POINTER
			l_sid: POINTER
			l_size_needed, l_n_user, l_n_domain, l_n_peUse: INTEGER
			l_owner_defaulted: BOOLEAN
		do
			create l_path.make (a_path)
			create lpSecurityBuffer.make (64)
			if GetFileSecurity (l_path.item, OWNER_SECURITY_INFORMATION, lpSecurityBuffer.item, lpSecurityBuffer.count, $l_size_needed) then
				if GetSecurityDescriptorOwner (lpSecurityBuffer.item, $l_sid, $l_owner_defaulted) then
					if l_sid /= default_pointer then
						create l_name.make_empty (64)
						create l_domain.make_empty (32)
						l_n_user := l_name.capacity
						l_n_domain := l_domain.capacity
						if LookupAccountSid (default_pointer, l_sid, l_name.item, $l_n_user, l_domain.item, $l_n_domain, $l_n_peUse) then
							create Result.make_empty
							Result.append_string (l_domain.string)
							if not Result.is_empty then
								Result.append_character ('\')
							end
							Result.append_string (l_name.string)
						end
					end
				end
			end
		end

	user_application_data_folder: attached STRING
			-- Folder where to store user specific preferences
			-- If Void, user preferences are simply not stored (useful for GUIs mostly)
		once
			Result := folders.appdata
		end

	system_directory: attached STRING
			-- Path to system directory (eg C:\WINDOWS)
		once
			Result := folders.system_directory
		end

	local_timezone_bias: INTEGER
			-- Offset in minutes of local timezone compared to UTC
		local
			l_pointer: MANAGED_POINTER
			l_status: INTEGER
		once
			create l_pointer.make (size_of_time_zone_information)
			l_status := GetTimeZoneInformation (l_pointer.item)
			if l_status = time_zone_id_standard then
				Result := standard_bias (l_pointer.item)
			elseif l_status = time_zone_id_daylight then
				Result := daylight_bias (l_pointer.item)
			end
		end

	get_executable_info (a_path: STRING): ROSE_EXECUTABLE_INFO
			-- Info on exe with 'a_path', if available
		local
			l_file: RAW_FILE
			l_struct: MANAGED_POINTER
			l_ec, l_estudio, l_string: STRING
			l_path: C_STRING
			l_date: ROSE_DATE
			l_stop: BOOLEAN
			i, l_size: INTEGER
		do
			if a_path /= Void and then not a_path.is_empty then
				create l_file.make (a_path)
				if l_file.exists and then l_file.is_readable then
					create Result.make (a_path)
					if a_path /= Void then
						create l_path.make (a_path)
						l_size := c_GetFileVersionInfoSize (l_path.item, $i)
					end
					if l_size > 0 then
						create l_struct.make (l_size)
						-- 'GetFileVersionInfo' If given a complete filename, it will search for it
						-- in the usual places, starting with the folder that the program was launched from.
						if c_GetFileVersionInfo (l_path.item, 0, l_size, l_struct.item) then
							l_string := get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_file_version)
							if l_string /= Void and then not l_string.is_empty then
								Result.set_version (l_string)
							end
							Result.set_scm_root (get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_build_scm))
							Result.set_revision_from_string (get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_build_revision))
							l_string := get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_name)
							if l_string /= Void and then not l_string.is_empty then
								Result.set_name (l_string)
							end
							l_string := get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_file_description)
							if l_string /= Void and then not l_string.is_empty then
								Result.set_description (l_string)
							end
							l_string := get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_internal_name)
							if l_string /= Void and then not l_string.is_empty then
								Result.set_internal_name (l_string)
							end
							Result.set_icon (get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_icon))
							Result.set_service_center (get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_service_center))
							Result.set_contacts (get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_contacts))
							Result.set_impact (get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_impact))
							Result.set_handle_from_string (get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_handle))
							Result.set_template (get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_template))
							Result.set_root_class (get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_root_class))
							Result.set_root_feature (get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_root_feature))
							Result.set_mt (get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_is_mt))
							Result.set_gui (get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_is_gui))
							Result.set_build_date_string (get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_build_date))
							Result.set_build_snapshot (get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_build_snapshot))
							l_ec := get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_build_ec)
							l_estudio := get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_build_estudio)
							if l_ec = Void or else l_ec.is_empty then
								l_ec := l_estudio
							end
							if l_estudio /= Void and then l_estudio.count > 1 then
								-- Take care of old format of the form "stable [5.7.64493]"
								i := l_estudio.index_of ('[', 1)
								if i > 0 then
									l_estudio := l_estudio.substring (i + 1, l_estudio.count)
									i := l_estudio.index_of (']', 1)
									if i > 0 then
										l_estudio := l_estudio.substring (1, i - 1)
									end
								end
								-- Take care of old format having BuildEstudio contain full path to estudio installation
								i := l_estudio.last_index_of ('\', l_estudio.count)
								i := i.max (l_estudio.last_index_of ('/', l_estudio.count))
								if i > 0 and then i < l_estudio.count then
									l_estudio := l_estudio.substring (i + 1, l_estudio.count)
								end
							end
							Result.set_build_ec (l_ec)
							Result.set_build_estudio (l_estudio)
							if attached get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_build_compatible) as l_compatible then
								Result.set_build_compatible (l_compatible)
							elseif attached get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_build_experimental) as l_experimental and then not (l_experimental.is_boolean and then l_experimental.to_boolean) then
								Result.set_build_compatible ("true")
							else
								Result.set_build_compatible ("false")
							end
							Result.set_build_exception_trace (get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_build_exception_trace))
							Result.set_build_debug (get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_build_debug))
							Result.set_build_platform (get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_build_platform))
							Result.set_build_c (get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_build_c))
							Result.set_build_author (get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_build_author))
							Result.set_build_host (get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_build_host))
							Result.set_build_dir (get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_build_dir))
							Result.set_build_ticket (get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_build_ticket))
							Result.set_company_name (get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_company_name))
							from
								i := 1
							until
								l_stop or else i > 9
							loop
								l_string := get_executable_specific_info (l_struct, {ROSE_EXECUTABLE_INFO}.k_override + i.out)
								if l_string /= Void and then not l_string.is_empty then
									Result.add_override (l_string)
								else
									l_stop := True
								end
								i := i + 1
							end
						end
					end
					if Result.build_date = Void then
						i := l_file.date
						if i > 0 then
							create l_date.make_by_unix_time (i)
							Result.set_build_date (l_date)
						end
					end
					if Result.build_date = Void then
						Result := Void
					end
				end
			end
		ensure
			build_date_if_exists: Result /= Void implies Result.build_date /= Void
		end

	last_system_call_status: INTEGER
			-- last message id (status) code reported by operating system (nb. not C runtime library 'errno')
			-- On some systems (VMS), this may need to be explicitly set after each system call.
		do
			Result := cwin32_GetLastError
		end

feature -- Basic operations

	explore (a_path: STRING)
			-- Launch an explorer window opened on 'a_path' if path is valid.
			-- Silently do nothing otherwise.
			-- Windows specific implementation
		local
			l_convention: ROSE_FRAMEWORK_CONVENTION
			l_formatting: ROSE_FORMATTING
			l_cmd_line: STRING
		do
			if a_path /= Void and then not a_path.is_empty then
				create l_convention
				create l_formatting
				l_cmd_line := l_formatting.formatter.composed_path (<<folders.system_directory, "explorer.exe">>)
				if not l_convention.is_file_readable (a_path) then
					l_cmd_line := "cmd /c start explorer "
				end
				if l_convention.is_directory_readable (a_path) then
					launch (l_cmd_line + " /e," + a_path)
				elseif l_convention.is_file_readable (a_path) then
					launch (l_cmd_line + " /e,/select," + a_path)
				end
			end
		end

	explore_url (a_url: STRING)
			-- Open 'a_url' in default web browser.
			-- Windows specific implementation
		local
			l_browser: STRING
		do
			if a_url /= Void then
				-- First try user's preference
				l_browser := reg.user.get ("SOFTWARE\Classes\HTTP\shell\open\command\")
				if l_browser = Void then
					-- Try 'wow6432node', just in case Windows didn't do the right thing
					l_browser := reg.user.get ("SOFTWARE\wow6432node\Classes\HTTP\shell\open\command\")
				end
				if l_browser = Void then
					-- Try user's 'htmlfile' class
					l_browser := reg.user.get ("SOFTWARE\Classes\htmlfile\shell\opennew\command\")
				end
				if l_browser = Void then
					-- Then try machine setting
					l_browser := reg.machine.get ("SOFTWARE\Classes\HTTP\shell\open\command\")
				end
				if l_browser = Void then
					l_browser := reg.machine.get ("SOFTWARE\Classes\htmlfile\shell\opennew\command\")
				end
				if l_browser /= Void then
					if l_browser.has_substring ("%%1") then
						l_browser.replace_substring_all ("%%1", a_url)
					else
						l_browser := l_browser + " " + a_url
					end
					launch (l_browser)
				end
			end
		end

	sleep (a_milliseconds: INTEGER)
			-- Sleep for 'a_milliseconds' amount of time.
		do
			c_sleep (a_milliseconds)
		end

	set_last_system_call_status (a_message_id: like last_system_call_status)
			-- Set 'last_system_call_status' from 'a_message_id'
		do
			cwin32_SetLastError (a_message_id)
		end

	system_message_text (a_message_id: INTEGER; a_arguments: ARRAY [INTEGER]): attached STRING
			-- error message text associated with `a_message_id' with (optional) `a_arguments'
		local
			l_save_sys_status: INTEGER
			l_buf: POINTER
			l_size: INTEGER
			l_flags: INTEGER
			l_c_args: ANY -- should be ARRAY [INTEGER]?
			l_c_string: C_STRING
		do
			create Result.make_empty
			l_save_sys_status := last_system_call_status
			l_flags := FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_ARGUMENT_ARRAY
			if a_arguments /= Void and then a_arguments.count > 0 then
					-- pass arguments as array ***needs testing***
				l_c_args := a_arguments.to_c
			else
				create l_c_args
				l_flags := l_flags | FORMAT_MESSAGE_IGNORE_INSERTS
			end
			l_size := cwin32_FormatMessage (l_flags, default_pointer, a_message_id, 0, $l_buf, 200, $l_c_args)
			if l_size > 0 then
				create l_c_string.make_by_pointer (l_buf)
				Result.append_string (l_c_string.string)
				cwin32_LocalFree (l_buf.item)
			else
				Result.append_string ("Error code ")
				Result.append_integer (a_message_id)
			end
			set_last_system_call_status (l_save_sys_status)
		end

feature {NONE} -- Implementation

	k_exe: STRING = ".exe"
			-- Exe extension

	nt_info: attached ROSE_NT_SYSTEM_INFORMATION
			-- NT specific system information
		once
			create Result.make
		end

	c_getpid: INTEGER
			-- Get current process id
		external
			"C blocking use <process.h>"
		alias
			"_getpid"
		end

	c_sleep (a_milliseconds: INTEGER)
			-- Sleep for 'a_milliseconds' amount of time.
		external
			"C blocking use <stdlib.h>"
		alias
			"_sleep"
		end

	cwin32_GetLastError: INTEGER
			-- win32 GetLastError
		external
			"C blocking signature (): EIF_INTEGER use <windows.h>"
		alias
			"GetLastError"
		end

	cwin32_SetLastError (dwErrCode: INTEGER)
			-- win32 SetLastError
		external
			"C blocking signature (DWORD) use <windows.h>"
		alias
			"SetLastError"
		end

	cwin32_FormatMessage (dwFlags: INTEGER; lpSource: POINTER; dwMessageId, dwLanguageId: INTEGER;
							lpBuffer: POINTER; nsize: INTEGER; a_Arguments: POINTER): INTEGER
			-- Returns number of characters written to output buffer (0 = error)
		external
			"C signature (DWORD, LPCVOID, DWORD, DWORD, LPSTR, DWORD, va_list *): EIF_INTEGER use <windows.h>"
		alias
			"FormatMessageA"
		end

	FORMAT_MESSAGE_ALLOCATE_BUFFER: INTEGER
			-- FORMAT_MESSAGE_ALLOCATE_BUFFER
		external
			"C blocking macro use <windows.h>"
		alias
			"FORMAT_MESSAGE_ALLOCATE_BUFFER"
		end

	FORMAT_MESSAGE_IGNORE_INSERTS: INTEGER
			-- FORMAT_MESSAGE_IGNORE_INSERTS
		external
			"C blocking macro use <windows.h>"
		alias
			"FORMAT_MESSAGE_IGNORE_INSERTS"
		end

	FORMAT_MESSAGE_FROM_SYSTEM: INTEGER
			-- FORMAT_MESSAGE_FROM_SYSTEM
		external
			"C blocking macro use <windows.h>"
		alias
			"FORMAT_MESSAGE_FROM_SYSTEM"
		end

	FORMAT_MESSAGE_ARGUMENT_ARRAY: INTEGER
			-- FORMAT_MESSAGE_ARGUMENT_ARRAY
		external
			"C blocking macro use <windows.h>"
		alias
			"FORMAT_MESSAGE_ARGUMENT_ARRAY"
		end

	cwin32_LocalFree (hMem: POINTER)
			-- win32 LocalFree
		external
			"C blocking signature (HLOCAL) use <windows.h>"
		alias
			"LocalFree"
		end

	GetTimeZoneInformation (a_pointer: POINTER): INTEGER
			-- win32 GetTimeZoneInformation
		external
			"C blocking signature (LPTIME_ZONE_INFORMATION): EIF_INTEGER_32 use <windows.h>"
		alias
			"GetTimeZoneInformation"
		end

	size_of_time_zone_information: INTEGER
			-- size of TIME_ZONE_INFORMATION
		external
			"C blocking macro signature (): EIF_INTEGER_32 use <windows.h>"
		alias
			"sizeof(TIME_ZONE_INFORMATION)"
		end

	daylight_bias (a_p: POINTER): INTEGER
			-- Daylight bias
		external
			"C blocking inline use <windows.h>"
		alias
			"((struct _TIME_ZONE_INFORMATION *)$a_p)->Bias + ((struct _TIME_ZONE_INFORMATION *)$a_p)->DaylightBias"
		end

	standard_bias (a_p: POINTER): INTEGER
			-- Standard bias
		external
			"C blocking inline use <windows.h>"
		alias
			"((struct _TIME_ZONE_INFORMATION *)$a_p)->Bias + ((struct _TIME_ZONE_INFORMATION *)$a_p)->StandardBias"
		end

	time_zone_id_standard: INTEGER
			-- TIME_ZONE_ID_STANDARD
		external
			"C blocking macro signature (): EIF_INTEGER_32 use <windows.h>"
		alias
			"TIME_ZONE_ID_STANDARD"
		end

	time_zone_id_daylight: INTEGER
			-- TIME_ZONE_ID_DAYLIGHT
		external
			"C blocking macro signature (): EIF_INTEGER_32 use <windows.h>"
		alias
			"TIME_ZONE_ID_DAYLIGHT"
		end

	OWNER_SECURITY_INFORMATION: INTEGER
			--  SECURITY_INFORMATION data type: the owner identifier of the object is being referenced.
		external
			"C blocking macro use <windows.h>"
		alias
			"OWNER_SECURITY_INFORMATION"
		end

	GetSecurityDescriptorOwner (pSecurityDescriptor, pOwner, lpbOwnerDefaulted: POINTER): BOOLEAN
			--  Retrieves the owner information from a security descriptor
		external
			"C blocking signature (PSECURITY_DESCRIPTOR, PSID *, LPBOOL): EIF_BOOLEAN use <windows.h>"
		alias
			"GetSecurityDescriptorOwner"
		end

	GetFileSecurity (lpFileName: POINTER; RequestedInformation: INTEGER; pSecurityDescriptor: POINTER; nLength: INTEGER; lpnLengthNeeded: POINTER): BOOLEAN
			-- Obtains specified information about the security of a file or directory. The information obtained is constrained by the caller's access rights and privileges
		external
			"C blocking signature (LPCSTR, SECURITY_INFORMATION, PSECURITY_DESCRIPTOR, DWORD, LPDWORD): EIF_BOOLEAN use <windows.h>"
		alias
			"GetFileSecurityA"
		end

	LookupAccountSid (lpSystemName, lpSid, lpName, cchName, lpReferencedDomainName, cchReferencedDomainName, peUse: POINTER): BOOLEAN
			-- Retrieves the name of the account for this SID and the name of the first domain on which this SID is found
		external
			"C blocking signature (LPCSTR, PSID, LPSTR, LPDWORD, LPSTR, LPDWORD, PSID_NAME_USE): EIF_BOOLEAN use <windows.h>"
		alias
			"LookupAccountSidA"
		end

	get_executable_specific_info (a_struct: MANAGED_POINTER; a_info: STRING): STRING
			-- Get 'a_info' field from exe info
		require
			struct_ok: a_struct /= Void
			info_ok: a_info /= Void
		local
			l_infoname: C_STRING
			l_value: MANAGED_POINTER
			i: INTEGER
		do
			create l_infoname.make ("\StringFileInfo\040904B0\" + a_info)
			create l_value.make (size_of_pointer)
			if c_VerQueryValue (a_struct.item, l_infoname.item, l_value.item, $i) then
				if i > 0 then
					create l_infoname.make_by_pointer (l_value.read_pointer (0))
					Result := l_infoname.string
				end
			end
		end

	c_GetFileVersionInfo (a_path: POINTER; a_handle, a_length: INTEGER; a_data: POINTER): BOOLEAN
			-- win32 GetFileVersionInfoA
		external
			"C blocking signature (LPSTR,DWORD,DWORD,LPVOID): EIF_BOOLEAN use <windows.h>"
		alias
			"GetFileVersionInfoA"
		end

	c_GetFileVersionInfoSize (a_path, a_handle: POINTER): INTEGER
			-- win32 GetFileVersionInfoSizeA
		external
			"C blocking signature (LPSTR,LPDWORD): EIF_INTEGER use <windows.h>"
		alias
			"GetFileVersionInfoSizeA"
		end

	c_VerQueryValue (a_block, a_subblock, a_buffer, a_length: POINTER): BOOLEAN
			-- win32 VerQueryValueA
		external
			"C blocking signature (LPVOID,LPSTR,LPVOID*,PUINT): EIF_BOOLEAN use <windows.h>"
		alias
			"VerQueryValueA"
		end

	c_is_64bit_os: BOOLEAN
			-- This returns True only if current process is a 32bit process running on a 64bit Windows OS
		external
			"C blocking inline use %"eif_eiffel.h%""
		alias "{
			BOOL bIsWow64 = FALSE;
			#ifdef EIF_WINDOWS
				typedef BOOL (WINAPI *LPFN_ISWOW64PROCESS) (HANDLE, PBOOL);
				LPFN_ISWOW64PROCESS fnIsWow64Process = (LPFN_ISWOW64PROCESS)GetProcAddress(GetModuleHandle(L"kernel32"),"IsWow64Process");
				if (NULL != fnIsWow64Process)
				{
					if (!fnIsWow64Process(GetCurrentProcess(),&bIsWow64))
					{
					}
				}
			#endif
			return bIsWow64;
		}"
		end

	c_is_admin: BOOLEAN
			-- External C code to determine whether current user is an admin of current machine
		external
			"C blocking signature (): EIF_BOOLEAN use <shlobj.h>"
		alias
			"IsUserAnAdmin"
		end

end
