note

	description:

		"platform independent extensions to EXECUTION_ENVIRONMENT"

	notes: "[
		This class contains the platform-independent extensions to the Eiffel Base library EXECUTION_ENVIRONMENT.
		It is intended to be inherited by platform-specific classes that provide implementations for stuff.
	]"
	author: "David Schwartz"
	last_editor: "$Author: er $"
	date: "$Date: 2013/07/11 $"
	revision: "$Revision: #20 $"
	path: "$File: //rose/source/kernel/infrastructure/support/rose_execution_environment_common.e $"

deferred class
	ROSE_EXECUTION_ENVIRONMENT_COMMON

inherit
	EXECUTION_ENVIRONMENT
		rename
			sleep as sleep_nanoseconds,
			item as execution_environment_item,
			arguments as execution_environment_arguments
		end

feature -- Access

	is_administrator: BOOLEAN
			-- Is the user running current process an administrator on the machine running the process
		local
			l_dir: DIRECTORY
		do
			create l_dir.make (root_directory_name)
			Result := l_dir.is_writable
		end

	is_batch_current_user: BOOLEAN
			-- Is the current user id a batch-job-specific user?
		local
			l_batch: STRING
		do
			-- kludgy implementation: ***subject to change***
			-- `True' if user id begins with "BATCH_" and has at least one additional character
			l_batch := "BATCH_"
			Result := current_user.login_id.count > l_batch.count and then
						current_user.login_id.substring (1, l_batch.count).is_case_insensitive_equal (l_batch)
		end

	is_email_copy_self_current_user: BOOLEAN
			-- is the current user id one that should receive copies of any email message sent (i.e. not BATCH_ or PORTSERVER)?
		do
			Result := not (current_user.login_id.is_case_insensitive_equal ("portserver") or else is_batch_current_user)
		ensure
			no_batch_copy_self: Result implies not is_batch_current_user
		end

	current_process_id: INTEGER
			-- current process id
		once
			Result := c_getpid
			debug ("make_movies")
				Result := 0
			end
		end

	current_process_id_representation: attached STRING
			-- Textual representation of current process' id
		once
			create Result.make_empty
			Result.append_string ("pid:")
			Result.append_integer (current_process_id)
			Result.append_character (' ')
			Result.append_character ('(')
			Result.append_string (current_process_id.to_hex_string)
			Result.append_string (" Hex)")
		end

	current_user: attached ROSE_USER
			-- Current user
		deferred
		end

	host_name: attached STRING
			-- Hostname of this machine
		deferred
		end

	canonical_host_name: attached STRING
			-- Canonical name of current machine, as defined by Systems group
			-- examples: axaros.com/Orinda/Servers/VMware/Patch/ORPCB08
			--           axaros.com/Orinda/Desktops/AXA-ROS Citrix only/FFCTX03
			--           axaros.com/Fairfield/Servers/Externaly Accessible Citrix Servers/Patch/VMSVA61
			--           axaros.com/Orinda/Servers/Other-SystemsControl/Patch/ORSVQ05
		do
			Result := host_name
		end

	file_owner (a_path: STRING): STRING
			-- File owner for file with 'a_path', if possible to obtain it
		do
			-- No default implementation, platform specific
		end

	executable_path: attached STRING
			-- Path to the executable running this process
		deferred
		end

	folders: attached ROSE_SPECIAL_FOLDERS
			-- Windows special folders
		deferred
		end

	is_64bit_os: BOOLEAN
			-- Are we running on a 64bit OS?
		deferred
		end

	is_running_64_bit: BOOLEAN
			-- Is the current executable a 64bit executable?
		once
			Result := size_of_pointer = 8
		ensure
			is_running_64_bit_implies_is_64_bit_platform: Result implies is_64bit_os
		end

	size_of_pointer: INTEGER
			-- Number of bytes
		external
			"C blocking macro signature (): EIF_INTEGER_32 use <stdlib.h>"
		alias
			"sizeof(int *)"
		end

	last_system_call_status: INTEGER
			-- last message id (status) code reported by operating system (nb. not C runtime library 'errno')
			-- On some systems (VMS), this may need to be explicitly set after each system call.
		deferred
		end

	cpu_ticks_per_second: INTEGER
			-- CPU ticks per second
		external
			"C blocking macro signature (): EIF_INTEGER_32 use <time.h>"
		alias
			"CLOCKS_PER_SEC"
		end

	utc_seconds_offset: INTEGER
			-- Offset in seconds from local time zone to UTC
		local
			l_utc, l_local: ROSE_DATE
		once
			create l_utc.make_now_utc
			create l_local.make_now
			Result := l_local.relative_duration_in_seconds (l_utc).rounded
		end

	job_id: INTEGER
			-- ActiveBatch job_id from environment variable;
			-- This variable is set for jobs run from Activebatch.
		local
			l_job_id: STRING
		do
			l_job_id := get ("ABAT_JOBID")
			if l_job_id /= Void and then l_job_id.is_integer then
				Result := l_job_id.to_integer
			else
				Result := 0
			end
		end

	scheduler: attached STRING
			-- ActiveBatch scheduler from environment variable;
			-- This variable is set for jobs run from Activebatch.
			-- Format is "scheduler;execution_agent", so we truncate after the first ';'.
		local
			l_scheduler: STRING
			i: INTEGER
		do
			create Result.make_empty
			l_scheduler := get ("ABAT_JOBSCHED")
			if l_scheduler /= Void then
				i := l_scheduler.index_of (';', 1)
				if i > 0 then
					Result.append_string (l_scheduler.substring (1, i - 1))
				end
			end
		end

feature -- Basic operations

	explore (a_path: STRING)
			-- Launch an explorer window opened on 'a_path' if path is valid.
			-- Silently do nothing otherwise.
		do
			-- No implementation by default, is there a cross-platform implementation for this?
			-- Redefine this to provide platform specific implementation
		end

	explore_url (a_url: STRING)
			-- Open 'a_url' in default web browser.
		do
			-- No implementation by default, is there a cross-platform implementation for this?
			-- Redefine this to provide platform specific implementation
		end

	rshell (a_machine, a_command: STRING)
			-- Execute 'a_command' using rshell on 'a_machine'.
			-- 'a_machine' is typically a VMS machine
			-- WARNING: the machine running this command has to have the right version of 'rsh' installed
			-- (see Mauricio for more info on that topic)
			-- 'return_code' contains the code returned after execution
		require
			valid_machine: a_machine /= Void and then not a_machine.is_empty
			valid_command: a_command /= Void and then not a_command.is_empty
		do
			system ("rsh " + a_machine + " " + a_command)
		end

	sleep (a_milliseconds: INTEGER)
			-- Sleep for 'a_milliseconds' amount of time.
		deferred
		end

	system_with_echo (a_cmd: STRING)
			-- Pass a request to execute `a_cmd' to the operating system, echoed on std output
		require
			cmd_not_void: a_cmd /= Void
		do
			io.put_string (a_cmd)
			if a_cmd.count > 0 and then (a_cmd @ a_cmd.count).is_printable then
				io.new_line
			end
			system (a_cmd)
		end

	set_last_system_call_status (a_message_id: like last_system_call_status)
			-- Set last system call status
		deferred
		end

	system_message_text (a_message_id: like last_system_call_status; a_arguments: ARRAY [INTEGER]): attached STRING
			-- error message text associated with `a_message_id' with (optional) `a_arguments'
		deferred
		ensure
			last_system_call_status_unchanged: last_system_call_status = old last_system_call_status
		end

feature {NONE} -- Implementation

	checked_directory_path (a_path: STRING): STRING
			-- Returns `a_path' if it refers to an existing directory, else Void
		local
			l_dir: DIRECTORY
		do
			if a_path /= Void and then not a_path.is_empty then
				create l_dir.make (a_path)
				if l_dir.exists then
					Result := a_path
				end
			end
		end

	c_getpid: INTEGER
			-- implementation: getpid()
		deferred
		end

end
