note

	description:

		"AxaRosenberg descendent of RAW_FILE which handles some platform specific performance issues"

	author: "Mark Howard"
	last_editor: "$Author: er $"
	date: "$Date: 2013/07/11 $"
	revision: "$Revision: #19 $"
	path: "$File: //rose/source/kernel/infrastructure/text/classic/rose_file.e $"

class ROSE_FILE

inherit

	ROSE_COMMON_FILE

	RAW_FILE
		undefine
			read_word,
			exists
		redefine
			open_read,
			open_write,
			open_append,
			open_read_write,
			create_read_write,
			open_read_append,
			date,
			exists
		end

	FILE_ACCESS_LOG

create

	make,
	make_open_read,
	make_open_write,
	make_open_write_binary,
	make_open_append,
	make_open_read_write,
	make_create_read_write,
	make_create_read_write_binary,
	make_open_read_append

feature {NONE} -- Creation

	make_create_read_write_binary (fn: STRING) is
			-- Create file object with `fn' as file name
			-- and open file for both reading and writing;
			-- create it if it does not exist.
		require
			string_exists: fn /= Void
			string_not_empty: not fn.is_empty
		do
			make (fn)
			create_read_write_binary
		ensure
			file_named: name = fn
			exists: exists
			open_read: is_open_read
			open_write: is_open_write
		end

	make_open_write_binary (fn: STRING) is
			-- Like 'make_open_write' but sets the file into the C run-time's 'binary' mode.
		require
			string_exists: fn /= Void
			string_not_empty: not fn.is_empty
		do
			make (fn)
			open_write_binary	-- The only diff between this and 'make_open_write'.
		ensure
			file_named: name = fn
			exists: exists
			open_write: is_open_write
		end

feature -- Initialization

	open_read is
			-- Open file in read-only mode.
		do
			Precursor
			record_file_read_access (name)
		rescue
			error.signal (generator, "Can't open_read file '$1'", <<name>>)
		end

	open_write is
			-- Open file in write-only mode;
			-- create it if it does not exist.
		do
			Precursor
			record_file_write_access (name)
		rescue
			error.signal (generator, "Can't open_write file '$1'", <<name>>)
		end

	open_write_binary is
			-- For some reason, Eiffel has no way to write in binary.
			-- Here is our answer.
		do
				-- Exactly like 'open_write' except that I add 10 in the 'file_open'.
				-- The '10' indicates that we want the 'mode' in the 'fopen' to contain a "b" (binary).
			file_pointer := file_open (internal_name_pointer.item, 1 + 10)
			mode := Write_file
			record_file_write_access (name)
		ensure
			exists: exists
			open_write: is_open_write
		rescue
			error.signal (generator, "Can't open_write_binary file '$1'", <<name>>)
		end

	open_append is
			-- Open file in append-only mode;
			-- create it if it does not exist.
		do
			Precursor
			record_file_write_access (name)
		rescue
			error.signal (generator, "Can't open_append file '$1'", <<name>>)
		end

	open_read_write is
			-- Open file in read and write mode.
		do
			Precursor
			record_file_write_access (name)
		rescue
			error.signal (generator, "Can't open_read_write file '$1'", <<name>>)
		end

	create_read_write is
			-- Open file in read and write mode;
			-- create it if it does not exist.
		do
			Precursor
			record_file_write_access (name)
		rescue
			error.signal (generator, "Can't create_read_write file '$1'", <<name>>)
		end

	create_read_write_binary is
		require
			is_closed: is_closed
		do
			
			file_pointer := file_open (internal_name_pointer.item, 4 + 10) -- The only diff.
			mode := Read_write_file
			record_file_write_access (name)
		ensure
			exists: exists
			open_read: is_open_read
			open_write: is_open_write
		rescue
			error.signal (generator, "Can't create_read_write_binary file '$1'", <<name>>)
		end

	open_read_append is
			-- Open file in read and write-at-end mode;
			-- create it if it does not exist.
		do
			Precursor
			record_file_write_access (name)
		rescue
			error.signal (generator, "Can't open_read_append file '$1'", <<name>>)
		end

feature -- Access

	exists: BOOLEAN is
			-- Does this file exist?
		do
			Result := not is_closed or else Precursor
		rescue
			error.signal (generator, "Can't determine if file '$1' exists", <<name>>)
		end

	date: INTEGER is
			-- Time stamp (time of last modification)
		do
			Result := Precursor
		rescue
			error.signal (generator, "Can't get timestamp on file '$1'", <<name>>)
		end

feature -- Basic operations

	read_word is
			--read word
		do
			last_integer := c_read_integer (file_pointer, 4)
		rescue
			error.signal (generator, "Can't read_word from file '$1'", <<name>>)
		end

	put_word (a_word: INTEGER) is
		do
			c_write_integer (a_word, file_pointer, 4)
		rescue
			error.signal (generator, "Can't put_word in file '$1'", <<name>>)
		end

	read_short is
		do
			last_integer := c_read_integer (file_pointer, 2)
		rescue
			error.signal (generator, "Can't read_short from file '$1'", <<name>>)
		end

	put_short (a_short: INTEGER) is
		do
			c_write_integer (a_short, file_pointer, 2)
		rescue
			error.signal (generator, "Can't put_short in file '$1'", <<name>>)
		end

	read_byte is
		do
			last_integer := c_read_integer (file_pointer, 1)
		rescue
			error.signal (generator, "Can't read_byte from file '$1'", <<name>>)
		end

	put_byte (a_byte: INTEGER) is
			-- Put 1 byte from 'a_byte' at 'file_pointer'
		do
			c_write_integer (a_byte, file_pointer, 1)
		rescue
			error.signal (generator, "Can't put_byte in file '$1'", <<name>>)
		end

	write_stream, writestream (a_string: STRING) is
			-- write ASCII characters
		local
			str_area: ANY
			a_status: INTEGER
		do
			str_area := a_string.area
			a_status := fwrite ($str_area, 1, a_string.count, file_pointer)
		rescue
			error.signal (generator, "Can't write_stream to file '$1'", <<name>>)
		end

	write_unicode_stream (a_string: UC_STRING) is
			-- write 32-bit characters
		local
			i, nb: INTEGER
			c: UC_CHARACTER
			c_out: STRING
		do
			nb := a_string.count
			from
				i := 1
			until
				i > nb
			loop
				c := a_string.unicode_item (i)
				i := i + 1
				c_out := c.code.out
				put_integer_16 (c.code.to_integer_16)
			end
		rescue
			error.signal (generator, "Can not write_stream to file '$1'", <<name>>)
		end

	read_record (a_record: EXTERNAL_RECORD) is
		do
			a_record.set (c_read_record (file_pointer, a_record.record_length_))
			a_record.read
		rescue
			error.signal (generator, "Can't read_record from file '$1'", <<name>>)
		end

	set_buffer_size (a_size: INTEGER) is
			-- Set buffer size to 'a_size'
		local
			a_pointer: POINTER
			a_status: INTEGER
		do
			a_status := setvbuf (file_pointer, a_pointer, file_iofbf, a_size)
		rescue
			error.signal (generator, "Can't set_buffer_size($2) on file '$1'", <<name, a_size>>)
		end

feature {NONE} -- Implementation

	c_read_record (a_file: POINTER; a_size: INTEGER): POINTER is
		external
			"C"
		alias
			"read_record"
		end

	fwrite (a_string: POINTER; a_size: INTEGER; a_count: INTEGER; a_file: POINTER): INTEGER is
		external
			"C [macro %"stdio.h%"]"
		end

	file_iofbf: INTEGER is
		external
			"C [macro <stdio.h>]"
		alias
			"_IOFBF"
		end

	setvbuf (a_file: POINTER; a_buffer: POINTER; a_mode: INTEGER; a_size: INTEGER): INTEGER is
		external
			"C [macro <stdio.h>] (FILE*,char *,int,int): int"
		end

	c_read_integer (a_file: POINTER; a_length: INTEGER): INTEGER is
		external
			"C"
		end

	c_write_integer (an_integer: INTEGER; a_file: POINTER; a_length: INTEGER) is
		external
			"C"
		end

	rose_create_last_string (a_min_size: INTEGER) is
			-- Create new instance of `last_string' with a least `a_min_size'
			-- as capacity.
		do
			create last_string.make (rose_default_last_string_size.max (a_min_size))
		end

end
