note

	description:

		"General class for viewing a file as a linear collection of lines"

	author: "Mark Howard"
	last_editor: "$Author: e2 $"
	date: "$Date: 2013/10/01 $"
	revision: "$Revision: #23 $"
	path: "$File: //rose/source/kernel/infrastructure/text/classic/rose_file_lines.e $"

class ROSE_FILE_LINES

inherit

	LINEAR [STRING]

create

	make,
	make_for_big_files

feature {NONE} -- Implementation

	make (a_file: ROSE_COMMON_FILE) is
			-- Only works for files smaller than 2 GB (see make_for_big_files).
		require
			a_file_not_void: a_file /= Void
		local
			l_file_count: INTEGER
		do
			init (a_file)
			create string.make_empty
			if file.exists then
				make_succeeded := True
				l_file_count := file.count
					-- l_file_count can be negative for files larger than 2 GB.
				if l_file_count >= 0 and l_file_count < read_size then
					if not file.off then
						file.read_stream (count)
						string := file.last_string
							-- Note that string IS last_string.
					end
					file.close
				else
					offset := string.count + 1
				end
			end
			set_end_of_line_character_to_default
		end

	make_for_big_files (a_file: ROSE_COMMON_FILE) is
			-- Useful when the file is larger than 4 GB.
			-- This works for smaller files too but doesn't close the file on creation.
		do
			init (a_file)
			if file.file_pointer /= default_pointer then
				make_succeeded := True
				string := ""
				offset := string.count + 1
			end
			set_end_of_line_character_to_default
		end

feature -- Access

	make_succeeded: BOOLEAN

	index: INTEGER

	item: STRING

	count: INTEGER is
		do
			Result := file.count
		end

	is_empty: BOOLEAN is
		do
			Result := file.is_closed and string.count = 0
		end

	after: BOOLEAN

	to_array: ROSE_LINEAR_ARRAY [STRING] is
		do
			create Result.make_from_linear (current)
		end

	string: STRING

	start_has_been_called_before: BOOLEAN

	end_of_line_character: CHARACTER
			-- used in index_of_end_of_line()

	default_end_of_line_character: CHARACTER is '%N'

feature -- Basic operations

	start is
			-- Don't call 'start' a second time. It doesn't work.
		require else
			feasibility: not start_has_been_called_before
		do
			offset := 1
			forth
			start_has_been_called_before := True
		end

	forth is
		local
			l_end: INTEGER
			l_last_valid: INTEGER
		do
			if offset > string.count then
					-- This happens when start is called and whenever a buffer full
					-- of info just happens to end with a line terminator.
				fill_buffer
			end
			if file.is_closed and offset > string.count then
				after := True
					-- item doesn't change.
			else
				from
					l_end := index_of_end_of_line
				until
					l_end /= 0 or else file.is_closed
				loop
					fill_buffer
					l_end := index_of_end_of_line
				end
				if l_end = 0 then
					l_end := string.count
					l_last_valid := l_end
				else
					l_last_valid := l_end - 1
				end
				if l_last_valid > 0 and then string.item (l_last_valid) = '%R' then
					l_last_valid := l_last_valid - 1
				end
				item := string.substring (offset, l_last_valid)
				index := index + 1
				offset := l_end + end_of_line_size
			end
		end

	finish is
			-- This isn't what finish is really supposed to do.
		do
			item := Void
			offset := string.count + 1
			after := True
		end

	set_end_of_line_character (a_character: CHARACTER) is
		do
			end_of_line_character := a_character
		end

	set_end_of_line_character_to_default is
		do
			set_end_of_line_character (default_end_of_line_character)
		end

	set_end_of_line_character_to_null is
		do
			set_end_of_line_character ('%U')
		end

	set_end_of_line_character_to_carriage_return is
		do
			set_end_of_line_character ('%R')
		end

	set_end_of_line_character_to_newline is
		do
			set_end_of_line_character ('%N')
		end

feature -- Iteration

	do_all_with_index (a_action: PROCEDURE [ANY, TUPLE [STRING, INTEGER]]) is
			-- Call `a_action' on each line in turn, passing it the text, and the line number.
			-- This routine also has the side effect of making `after' = `True'.
			-- Semantics not guaranteed if `a_action' changes `Current'.
		require
			a_action_not_void: a_action /= Void
			a_action_preconditions_met: True -- can't be checked
		do
			from
				start
			until
				after
			loop
				a_action.call ([item, index])
				forth
			end
		end

feature {NONE} -- Implementation

	init (a_file: ROSE_COMMON_FILE) is
			-- initialize
		require
			a_file_not_void: a_file /= Void
		do
			file := a_file
			if file.is_closed and file.exists then
				file.open_read
			end
		ensure
			file_set: file = a_file
		end

	fill_buffer is
			-- buffer
		do
			if not file.is_closed then
				file.read_stream (read_size)
				string := string.substring (offset, string.count) + file.last_string
				offset := 1
				if file.end_of_file then
					file.close
				end
			end
		end

	index_of_end_of_line: INTEGER is
			-- This could be modified to search for a different terminator(s).
		do
			Result := string.index_of (end_of_line_character, offset)
		end

	read_size: INTEGER is 20_000_000

	end_of_line_size: INTEGER is 1

	file: ROSE_COMMON_FILE

	offset: INTEGER

end
