note

	description:
	"[
		General mechanism for converting a file with columns into an array of objects representing the lines
		of the file.
	]"
	author: "Mark Howard"
	last_editor: "$Author: it $"
	date: "$Date: 2011/03/04 $"
	revision: "$Revision: #10 $"
	path: "$File: //rose/source/kernel/infrastructure/text/rose_formatted_text_file.e $"

class
	ROSE_FORMATTED_TEXT_FILE[G->ANY create default_create end]

inherit
	ROSE_TEXT_FILE

create
   read, read_with_header_rows

feature -- Access

	is_expanding_tabs_to_spaces: BOOLEAN
			-- Expand tabs to spaces

	tab_size: INTEGER is
			-- Tab size in spaces
		do
			if spaces /= Void then
				Result := spaces.count
			end
		end

	default_tab_size: INTEGER is 8
			-- Default tab size

	enable_expand_tabs_to_spaces is
			-- Enable tab expansion
		do
			is_expanding_tabs_to_spaces := True
			if tab_size = 0 then
				set_tab_size (default_tab_size)
			end
		ensure
			definition: is_expanding_tabs_to_spaces
		end

	disable_expand_tabs_to_spaces is
			-- Disable tab expansion
		do
			is_expanding_tabs_to_spaces := False
		ensure
			definition: not is_expanding_tabs_to_spaces
		end

	set_tab_size (a_tab_size: INTEGER) is
			-- Set tab size to `a_tab_size'
			-- size of <= 0 means use default size (8)
		local
			l_tab_size: INTEGER
		do
				-- defensive check
			if a_tab_size > 0 then
				l_tab_size := a_tab_size
			else
				l_tab_size := default_tab_size
			end
			create spaces.make_filled (' ', l_tab_size)
		ensure
			tab_size_is_positive: a_tab_size > 0 implies tab_size = a_tab_size
			default_if_tab_size_not_set: a_tab_size <= 0 implies tab_size = default_tab_size
		end

feature {NONE} -- Initialization

	read (a_filename : STRING;a_format : ROSE_FORMAT) is
			-- Parse all rows of 'a_filename' using 'a_format' columns specification
			-- The records are stored in 'array'
		require
			filename_not_void_or_empty: (a_filename /= Void) and then (not a_filename.is_empty)
		do
			read_with_header_rows(a_filename, a_format, 0)
		ensure
			array_not_void: array /= Void
		end

	read_with_header_rows (a_filename : STRING;a_format : ROSE_FORMAT; a_number_of_header_rows_to_skip : INTEGER) is
			-- Parse all rows of 'a_filename' using 'a_format' columns specification but
			-- skip 'a_number_of_header_row'
			-- The records are stored in 'array'
		require
			filename_not_void_or_empty: (a_filename /= Void) and then (not a_filename.is_empty)
			valid_rows_to_skip : a_number_of_header_rows_to_skip >= 0
		local
			l_count : INTEGER
			l_lines : ROSE_FILE_LINES
			l_vessel : G
			l_line: STRING
		do
			create array.make_empty
			make(a_filename)
			l_lines := lines
			from l_lines.start until l_lines.off loop
				l_count := l_count + 1
				if l_count > a_number_of_header_rows_to_skip then
					create l_vessel
					if is_expanding_tabs_to_spaces then
						l_line := expand_tabs_to_spaces (l_lines.item)
					else
						l_line := l_lines.item
					end
					a_format.fill_vessel(l_vessel,l_line)
					array.extend(l_vessel)
				end
				l_lines.forth
			end
		ensure
			array_not_void: array /= Void
		end

	spaces: STRING
			-- Spaces for tab expansion

	expand_tabs_to_spaces (a_line: STRING): STRING is
			-- Expand tabs to spaces
		require
			tab_size_is_positive: tab_size > 0
			a_line_not_void: a_line /= Void
		local
			n: INTEGER
			l_tab_index: INTEGER
			l_number_of_spaces: INTEGER
		do
			l_tab_index := a_line.index_of ('%T', 1)
			if l_tab_index > 0 then
				from
					create Result.make_from_string (a_line)
					n := tab_size
				until
					l_tab_index <= 0
				loop
					l_number_of_spaces := n - (l_tab_index - 1) \\ n
					Result.replace_substring (spaces.substring (1, l_number_of_spaces), l_tab_index, l_tab_index)
					l_tab_index := Result.index_of ('%T', 1)
				end
			else
				Result := a_line
			end
		ensure
			line_not_void: Result /= Void
			no_tabs: not Result.has ('%T')
		end

feature -- Access

	array : ROSE_LINEAR_ARRAY[G]
		-- Parsed records

end
