note
	description: "Utility formatting services"

	author: "Zoran Simic"
	last_editor: "$Author: rj $"
	date: "$Date: 2014/05/16 $"
	revision: "$Revision: #84 $"
	path: "$File: //rose/source/kernel/infrastructure/framework/rose_formatting_service.e $"

class ROSE_FORMATTING_SERVICE

inherit

	ROSE_FRAMEWORK_CONVENTION

	ROSE_OBJECT_REPRESENTATION

	KL_SHARED_FILE_SYSTEM
		export
			{NONE} all
		end

feature -- Formatting

	frozen string_less_than_comparator (s1, s2: STRING): BOOLEAN
			-- Is 's1' < 's2'?
		do
			if s1 = Void then
				Result := s2 /= Void
			elseif s2 = Void then
				Result := False
			else
				Result := s1 < s2
			end
		end

	hex_to_byte (a_char: CHARACTER): NATURAL_8
			-- Byte corresponding to 'a_char'
		require
			char_hex: ("0123456789ABCDEF").has (a_char.as_upper)
		do
			inspect a_char.as_upper
			when '0' then Result := 0
			when '1' then Result := 1
			when '2' then Result := 2
			when '3' then Result := 3
			when '4' then Result := 4
			when '5' then Result := 5
			when '6' then Result := 6
			when '7' then Result := 7
			when '8' then Result := 8
			when '9' then Result := 9
			when 'A' then Result := 10
			when 'B' then Result := 11
			when 'C' then Result := 12
			when 'D' then Result := 13
			when 'E' then Result := 14
			when 'F' then Result := 15
			end
		ensure
			in_range: Result >= 0 and Result < 16
		end

	byte_to_hex (a_byte: NATURAL_8): CHARACTER
			-- Hex digit corresponding to value in 'a_byte'.
		require
			in_range: a_byte >= 0 and a_byte < 16
		do
			inspect a_byte
			when 0 then Result := '0'
			when 1 then Result := '1'
			when 2 then Result := '2'
			when 3 then Result := '3'
			when 4 then Result := '4'
			when 5 then Result := '5'
			when 6 then Result := '6'
			when 7 then Result := '7'
			when 8 then Result := '8'
			when 9 then Result := '9'
			when 10 then Result := 'A'
			when 11 then Result := 'B'
			when 12 then Result := 'C'
			when 13 then Result := 'D'
			when 14 then Result := 'E'
			when 15 then Result := 'F'
			end
		ensure
			reverse: hex_to_byte (Result) = a_byte
		end

	md5_hash (a_string: STRING): STRING
			-- Compute MD5 hash for `a_string'.
		require
			string_ok: a_string /= Void
		local
			l_hash: CRYPT_MD5
			l_byte: NATURAL_8
			i: INTEGER
		do
			create l_hash.make
			l_hash.update_string (a_string)
			l_hash.final
			from
				create Result.make_empty
				i := 1
			until
				i > l_hash.digest_size
			loop
				l_byte := l_hash.digest.item (i)
				Result.append_character (byte_to_hex (l_byte.bit_shift_right (4).bit_and (15)).as_lower)
				Result.append_character (byte_to_hex (l_byte.bit_and (15)).as_lower)
				i := i + 1
			end
		ensure
			result_ok: Result /= Void and then Result.count = 32
		end

	plural (a_count: INTEGER; a_noun: STRING): attached STRING
			-- String showing that 'a_noun' occurred 'a_count' times, with an optional 's' added to 'a_noun' if 'a_count' /= 1
		do
			create Result.make_empty
			Result.append_integer (a_count)
			Result.append_character (' ')
			if a_count = 1 then
				Result.append_string (a_noun)
			elseif a_noun.ends_with (once "cy") then
				Result.append_string (a_noun.substring (1, a_noun.count - 2))
				Result.append_string (once "cies")
			else
				Result.append_string (a_noun)
				Result.append_character ('s')
			end
		end

	replaced_substring (a_string, a_original, a_replace: STRING): STRING
			-- 'a_string' with all occurences of 'a_original' replaced by 'a_replace'
			-- Equivalent to STRING.replace_substring_all, but faster
		require
			string_ok: a_string /= Void
			original_ok: a_original /= Void and then a_original.count > 0
			replace_ok: a_replace /= Void and then a_replace.count > 0
		local
			i, j: INTEGER
		do
			i := a_string.substring_index (a_original, 1)
			if i > 0 then
				from
					create Result.make (a_string.count)
					j := 1
				until
					i <= 0
				loop
					Result.append (a_string.substring (j, i - 1))
					Result.append (a_replace)
					j := i + a_original.count
					i := a_string.substring_index (a_original, j)
				end
				Result.append (a_string.substring (j, a_string.count))
			else
				Result := a_string
			end
		end

	frozen split_integral (a_string: STRING; a_separator: CHARACTER): STRING
			-- Apply separators to an integer
		local
			i, n: INTEGER_32
		do
			if a_separator = '%U' then
				Result := a_string
			else
				from
					n := a_string.count
					create Result.make_empty
				until
					n <= 3
				loop
					from
						i := 0
					until
						i = 3
					loop
						Result.precede (a_string.item (n))
						n := n - 1
						i := i + 1
					end
					Result.precede (a_separator)
				end
				from
				until
					n = 0
				loop
					Result.precede (a_string.item (n))
					n := n - 1
				end
			end
		end


	frozen pad_fraction (a_fraction: DOUBLE; a_decimals: INTEGER): attached STRING
			-- Stretch or shrink `a_fraction' to length 'a_decimals'.
		do
			Result := a_fraction.out
			Result.remove_tail (1)
			if Result.count > a_decimals then
				Result := Result.substring (1, a_decimals)
			else
				from
				until
					Result.count = a_decimals
				loop
					Result.precede ('0')
				end
			end
		ensure
			result_ok: Result.count = a_decimals
		end

	frozen command_line (a_exe: STRING; a_args: ARRAY [STRING]): attached STRING
			-- Correct command line for 'a_exe' and 'a_args'
			-- Arguments and exe with spaces are enclosed in quotes
		local
			s: STRING
			i: INTEGER
		do
			create Result.make (10)
			if a_exe /= Void and then not a_exe.is_empty then
				Result.append_string (quoted (a_exe))
			end
			if a_args /= Void and then not a_args.is_empty then
				from
					i := a_args.lower
				until
					i > a_args.upper
				loop
					s := a_args.item (i)
					if s /= Void and then not s.is_empty then
						if not Result.is_empty then
							Result.append_character (' ')
						end
						Result.append (quoted (s))
					end
					i := i + 1
				end
			end
		end

	frozen quoted (a_string: STRING): attached STRING
			-- 'a_string' quoted with '"' if it contains a space
		require
			string_ok: a_string /= Void
		do
			create Result.make_empty
			Result.append_string (a_string)
			Result.replace_substring_all ("%"", "")
			if a_string.is_empty or else a_string.has (' ') or else a_string.has ('%T') then
				Result.prepend_character ('%"')
				Result.append_character ('%"')
			end
		end

	frozen recursively_interpreted (a_string: STRING; a_max_depth: INTEGER; a_interpretor: FUNCTION [ANY, TUPLE [STRING], STRING]): STRING
			-- Interpreted value of 'a_string' recursively with 'a_max_depth' recursive calls
			-- using 'a_interpretor' to interpret values
		local
			l_value: STRING
			p, p_var: PAIR [STRING, STRING]
		do
			if a_max_depth <= 0 or else a_string = Void or else not a_string.has (variable_placeholder) then
				Result := a_string
			else
				from
					p := split_on_first (a_string, variable_placeholder.out, Void)
					create Result.make (a_string.count)
				until
					p.first = Void or else (p.first.is_empty and then p.second = Void)
				loop
					Result.append (p.first)
					p_var := split_delimited (p.second, Void)
					l_value := p_var.first
					if l_value = Void then
						Result.append_character (variable_placeholder)
					elseif l_value.is_integer then
							-- Ignore all $numbers
						Result.append_character (variable_placeholder)
					elseif a_interpretor /= Void then
							-- Everything else is interpreted
						l_value := a_interpretor.item ([recursively_interpreted (l_value, a_max_depth - 1, a_interpretor)])
					else
						l_value.prepend_character (variable_placeholder)
					end
					if l_value /= Void then
						Result.append (l_value)
					end
					p := split_on_first (p_var.second, variable_placeholder.out, Void)
				end
				Result := recursively_interpreted (Result, a_max_depth - 1, a_interpretor)
			end
		ensure
			result_ok: (a_max_depth > 0 and then a_string /= Void) implies Result /= Void
		end

	frozen is_same_string (s1, s2: STRING): BOOLEAN
			-- Are 's1' and 's2' same in case-insensitive manner?
		do
			if s1 = Void then
				Result := s2 = Void
			elseif s2 /= Void and then s1.count = s2.count then
				Result := s1.as_lower ~ s2.as_lower
			end
		end

	frozen added_exe (a_path: STRING): STRING
			-- ".exe" added to 'a_path' if not already there
		do
			Result := added_extension (a_path, s_exe)
		ensure
			result_ok: a_path /= Void implies Result /= Void
		end

	frozen stripped_exe (a_path: STRING): STRING
			-- 'a_path' stripped of its '.exe' extension, if any
		do
			Result := stripped_extension (a_path, s_exe)
		ensure
			result_ok: a_path /= Void implies Result /= Void
		end

	frozen added_extension (a_name, a_extension: STRING): STRING
			-- 'a_name' with 'a_extension' added to it if need be
		require
			valid_extension: a_extension /= Void and then a_extension.count > 1 and then a_extension.item (1) = '.'
		local
			i: INTEGER
		do
			if a_name = Void or else a_name.is_empty then
				Result := a_name
			else
				i := a_name.as_lower.substring_index (a_extension.as_lower, 1)
				if i > 0 and then i = a_name.count - a_extension.count + 1 then
					Result := a_name
				else
					Result := a_name + a_extension
				end
			end
		ensure
			result_ok: a_name /= Void implies Result /= Void
		end

	frozen stripped_extension (a_name, a_extension: STRING): STRING
			-- 'a_name' stripped of its 'a_extension', if any
		require
			valid_extension: a_extension /= Void and then a_extension.count > 1 and then a_extension.item (1) = '.'
		do
			if a_name = Void or else a_name.is_empty then
				Result := a_name
			else
				Result := a_name.as_lower
				if a_name.count >= a_extension.count
				and then Result.substring_index (a_extension, 1) = Result.count - a_extension.count + 1 then
					Result.keep_head (Result.count - a_extension.count)
				end
			end
		ensure
			result_ok: a_name /= Void implies Result /= Void
		end

	frozen is_unc (a_path: STRING): BOOLEAN
			-- Is 'a_path' a UNC path?
		do
			if a_path /= Void and then a_path.count > 4 then
				Result := (a_path.item (1) = '\' or else a_path.item (1) = '/')
					and then (a_path.item (2) = '\' or else a_path.item (2) = '/')
			end
		end

	frozen root (a_path: STRING): STRING
			-- Root part of 'a_path'
		local
			l_path: STRING
			i, j, k: INTEGER
		do
			if a_path /= Void and then a_path.count > 1 then
				l_path := file_system.canonical_pathname (a_path)
				if l_path.count < 2 then
					-- Do nothing
				elseif is_unc (l_path) then
					-- Path starting with \\
					i := index_of_character (l_path, "/\", 3)
					if i > 3 then
						-- of the form \\server\...
						k := i
						i := index_of_character (l_path, "/\", i + 1)
						j := l_path.index_of ('.', 3)
						if i > 0 then
							-- of the form \\server\share\...
							if j = 0 or else j > k then
								-- of the form \\server\share\...
								if i > k + 1 then
									-- avoid the case \\server\\
									Result := l_path.substring (1, i - 1)
								end
							else
								-- of the form \\server.com\share\...
								i := index_of_character (l_path, "/\", i + 1)
								if i > 0 then
									-- of the form \\server.com\share\dir\...
									Result := l_path.substring (1, i - 1)
								elseif l_path.item (l_path.count) /= '\' and then l_path.item (l_path.count) /= '/'then
									-- of the form \\server.com\share\dir
									Result := l_path
								end
							end
						elseif (j = 0 or else j > k) and then (l_path.count > k) then
							-- of the form \\server\share
							Result := l_path --+ "\"
						end
					end
				elseif l_path.item (2) = ':' then
					Result := l_path.substring (1, 2) + "\"
				end
			end
		end

	frozen index_of_character (a_string, a_chars: STRING; a_start: INTEGER): INTEGER
			-- Index of any character from 'a_chars' in 'a_string', starting at 'a_start'
		require
			chars_ok: a_chars /= Void
		local
			i: INTEGER
		do
			if a_string /= Void then
				from
					i := a_start
				until
					Result > 0 or else i > a_string.count
				loop
					if a_chars.has (a_string.item (i)) then
						Result := i
					end
					i := i + 1
				end
			end
		end

	frozen short_path (a_path: STRING; a_max: INTEGER): STRING
			-- 'a_path' shortened to a size of about 'a_max' characters
		local
			l_path: STRING
			i, j: INTEGER
		do
			if a_path /= Void then
				if a_path.count <= a_max then
					Result := a_path.twin
				else
					create Result.make (a_max)
					l_path := replaced_substring (a_path, once "/", once "\")
					from
						j := 1
						i := l_path.index_of ('\', j)
					until
						i < 1 or else i > j
					loop
						j := i + 1
						i := l_path.index_of ('\', j)
					end
					if i > 0 then
						from
							Result.append (l_path.substring (1, i))
							j := i + 1
						until
							i < 1
						loop
							i := l_path.index_of ('\', j)
							if i > 0 then
								j := i + 1
								if Result.count + l_path.count - j - 1 <= a_max then
									i := 0
								end
							end
						end
						Result.append (once "[...]")
						Result.append (l_path.substring (j - 1, l_path.count))
					else
						Result := l_path
					end
				end
			else
				create Result.make_empty
			end
		ensure
			result_ok: Result /= Void and then Result /= a_path
		end

	frozen unescaped (a_string, a_escape: STRING): STRING
			-- Unescaped 'a_string' with every character in 'a_escape' considered as an escape character
		require
			escape_ok: a_escape /= Void and then a_escape.count > 0 and then not a_escape.has ('%"')
		local
			in_escape, in_string: BOOLEAN
			i: INTEGER
			c: CHARACTER
		do
			if a_string = Void or else a_string.is_empty then
				Result := a_string
			else
				from
					create Result.make (a_string.count)
					i := 1
				until
					i > a_string.count
				loop
					c := a_string.item (i)
					if in_string then
						in_string := in_escape or else c /= '%"'
						in_escape := not in_escape and then a_escape.has (c)
						Result.extend (c)
					else
						in_string := not in_escape and then c = '%"'
						in_escape := not in_string and then not in_escape and then a_escape.has (c)
						if not in_escape then
							Result.extend (c)
						end
					end
					i := i + 1
				end
			end
		ensure
			result_ok: a_string /= Void implies Result /= Void
		end

	frozen escape (a_string, a_escapable: STRING; a_escape: CHARACTER): STRING
			-- Escape all 'a_escapable' characters in 'a_string' with 'a_escape'.
		require
			valid_escapable: a_escapable /= Void
			valid_escape: a_escape /= '%U'
		local
			i: INTEGER
			c: CHARACTER
		do
			if a_string /= Void then
				from
					i := 1
					create Result.make (a_string.count)
				until
					i > a_string.count
				loop
					c := a_string.item (i)
					if a_escapable.has (c) then
						Result.append_character (a_escape)
					end
					Result.append_character (c)
					i := i + 1
				end
			end
		ensure
			result_ok: a_string /= Void implies Result /= Void
		end

	frozen join (s: SEQUENCE [ANY]; a_separator: STRING): attached STRING
			-- Join 'a_array' into a string, items are delimited by 'a_separator'
		require
			valid_separator: a_separator /= Void
		do
			if s = Void then
				create Result.make (0)
			else
				from
					s.start
					create Result.make (s.count * 10)
				until
					s.off
				loop
					Result.append_string (object_representation (s.item, Void))
					s.forth
					if not s.off then
						Result.append (a_separator)
					end
				end
			end
		end

	frozen join_array (a_array: ARRAY [ANY]; a_separator: STRING): attached STRING
			-- Join 'a_array' into a string, items are delimited by 'a_separator'
		require
			valid_separator: a_separator /= Void
		local
			i: INTEGER
		do
			if a_array = Void then
				create Result.make (0)
			else
				from
					i := a_array.lower
					create Result.make (a_array.count * 10)
				until
					i > a_array.upper
				loop
					Result.append_string (object_representation (a_array.item (i), Void))
					i := i + 1
					if i <= a_array.upper then
						Result.append (a_separator)
					end
				end
			end
		end

	frozen split_array (a_string, c, a_escape: STRING; a_keep_empty: BOOLEAN): ARRAY [STRING]
			-- Split 's' by any character in 'c'
			-- 'a_keep_empty' indicates wether to return empty strings to indicating how many separators were found
			-- Every character in 'a_escape' is considered as an escape character
		local
			i, j: INTEGER
		do
			if a_string = Void then
				-- Void
			elseif a_string.is_empty then
				-- Empty array
				create Result.make_filled (Void, 1, 0)
			else
				from
					create Result.make_filled (Void, 1, 0)
					Result.compare_objects
					i := 1
					j := 1
				until
					j = 0
				loop
					j := find_unescaped (i, a_string, c, a_escape)
					if j > 0 then
						if a_keep_empty or else j > i then
							Result.force (a_string.substring (i, j - 1), Result.upper + 1)
						end
						i := j + 1
					elseif a_keep_empty or else a_string.count >= i then
						Result.force (a_string.substring (i, a_string.count), Result.upper + 1)
					end
				end
			end
		ensure
			result_ok: a_string /= Void implies Result /= Void
		end

	frozen split_array_raw (a_string: STRING; a_separator: CHARACTER; a_keep_empty: BOOLEAN): ARRAY [STRING]
			-- Split 'a_string' by any character in 'a_separator'
			-- 'a_keep_empty' indicates wether to return empty strings to indicating how many separators were found
			-- This feature doesn't allow to escape 'a_separator' in any way, use 'split_array' if you need to be able to escape 'a_separator'
		local
			i, j: INTEGER
		do
			if a_string = Void then
				-- Void
			elseif a_string.is_empty then
				-- Empty array
				create Result.make_filled (Void, 1, 0)
			else
				from
					create Result.make_filled (Void, 1, 0)
					Result.compare_objects
					i := 1
					j := 1
				until
					j = 0
				loop
					j := a_string.index_of (a_separator, i)
					if j > 0 then
						if a_keep_empty or else j > i then
							Result.force (a_string.substring (i, j - 1), Result.upper + 1)
						end
						i := j + 1
					elseif a_keep_empty or else a_string.count >= i then
						Result.force (a_string.substring (i, a_string.count), Result.upper + 1)
					end
				end
			end
		ensure
			result_ok: a_string /= Void implies Result /= Void
		end

	frozen split (a_string, c, a_escape: STRING): SEQUENCE [STRING]
			-- Split 'a_string' by any character in 'c'
			-- Every character in 'a_escape' is considered as an escape character
		local
			i, j: INTEGER
		do
			if a_string = Void then
				-- Void
			elseif a_string.is_empty then
				-- Empty
				create {LINKED_LIST [STRING]} Result.make
			else
				from
					create {LINKED_LIST [STRING]} Result.make
					Result.compare_objects
					i := 1
					j := 1
				until
					j = 0
				loop
					j := find_unescaped (i, a_string, c, a_escape)
					if j > 0 then
						Result.extend (a_string.substring (i, j - 1))
						i := j + 1
					else
						Result.extend (a_string.substring (i, a_string.count))
					end
				end
			end
		ensure
			result_ok: a_string /= Void implies Result /= Void
		end

	frozen split_on_first (s, c, a_escape: STRING): PAIR [STRING, STRING]
			-- Break 's' in two strings separated by first occurence of any character in 'c'
			-- Characters form 'c' may appear in splitted result if escaped by any character in 'a_escape'
			-- ex: for c = '!'
			-- s=" a b!c " -> Result=(" a b", "c ")
			-- s=" a b!!c" -> Result=(" a b", "!c")
			-- s="!a b!c"  -> Result=("", "a b!c")
			-- s="a b"     -> Result=("a b", Void)
			-- s="a b!"    -> Result=("a b", "")
			-- s=""        -> Result=("", Void)
			-- s=Void      -> Result=(Void, Void)
		local
			i: INTEGER
		do
			if s = Void then
				create Result.make (Void, Void)
			elseif s.is_empty then
				create Result.make (s.twin, Void)
			else
				i := find_unescaped (1, s, c, a_escape)
				if i = 0 then
					-- 'c' was not found in 's'
					create Result.make (s.twin, Void)
				else
					create Result.make (s.substring (1, i - 1), s.substring (i+1, s.count))
				end
			end
		ensure
			valid_result: Result /= Void
			first: s /= Void implies Result.first /= Void
		end

	frozen split_on_last (s, c, a_escape: STRING): attached PAIR [STRING, STRING]
			-- Break 's' in two strings separated by last occurence of any character in 'c'
			-- Characters form 'c' may appear in splitted result if escaped by any character in 'a_escape'
		local
			i, j: INTEGER
		do
			if s = Void then
				create Result.make (Void, Void)
			elseif s.is_empty then
				create Result.make (s.twin, Void)
			else
				from
					j := find_unescaped (1, s, c, a_escape)
				until
					j = 0
				loop
					i := j
					j := find_unescaped (i + 1, s, c, a_escape)
				end
				if i = 0 then
					-- 'c' was not found in 's'
					create Result.make (s.twin, Void)
				else
					create Result.make (s.substring (1, i - 1), s.substring (i + 1, s.count))
				end
			end
		end

	frozen split_delimited (s, a_escape: STRING): PAIR [STRING, STRING]
			-- Return first "delimited" string within 's', and the rest of 's'.
			-- Delimiters are "" or {} or none (in this case the next alphanumerical word is looked up)
			-- Every character in 'a_escape' is considered as an escape character
			-- ex: 'a_escape' = "\"
			-- ex: s="  a12/ b "  -> Result=(Void, "  a12/ b ")
			-- ex: s="a12/ b "    -> Result=("a12", "/ b ")
			-- ex: s="/a"         -> Result=("", "/a")
			-- ex: s="{a/ b}"     -> Result=("a/ b", "")
			-- ex: s="{a/ b}/c"   -> Result=("a/ b", "/c")
			-- ex: s="{a/ \}b}/c" -> Result=("a/ }b", "/c")
			-- ex: s=""a b"/c"    -> Result=("a b", "/c")
			-- ex: s="{a"         -> Result=(Void, "{a")
		local
			i: INTEGER
			stop, in_escape: BOOLEAN
			c: CHARACTER
			l_string: STRING
		do
			if s = Void then
				create Result.make (Void, Void)
			elseif s.is_empty then
				create Result.make (s.twin, Void)
			elseif s.item (1) = '"' then
				Result := split_on_first (s.substring (2, s.count), s_double_quote, a_escape)
			elseif s.item (1) = '{' then
				Result := split_enclosed (s, '{', '}', a_escape)
			elseif s.item (1) = '(' then
				Result := split_enclosed (s, '(', ')', a_escape)
			elseif s.item (1) = '[' then
				Result := split_enclosed (s, '[', ']', a_escape)
			elseif alpha_numerical_char (s.item (1)) then
				from
					i := 1
					create l_string.make (s.count)
				until
					stop or else i > s.count
				loop
					c := s.item (i)
					if not in_escape and then not alpha_numerical_char (c) then
						stop := True
					else
						in_escape := not in_escape and a_escape /= Void and then a_escape.has (c)
						if not in_escape then
							l_string.extend (c)
						end
						i := i + 1
					end
				end
				create Result.make (l_string, s.substring (i, s.count))
			else
				create Result.make (Void, s.twin)
			end
		ensure
			valid_result: Result /= Void
		end

	frozen split_enclosed (s: STRING; a_opening, a_closing: CHARACTER; a_escape: STRING): attached PAIR [STRING, STRING]
			-- Return first enclosed string in 's' between characters in 'a_opening' and in 'a_closing'
			-- and the rest of that string
			-- Forget all characters before 'a_opening'
			-- Every character in 'a_escape' is considered as an escape character
			-- ex: with [] for opening/closing and '\' for escape
			-- ex: s="[a12]/ b "     -> Result=("a12", "/ b ")
			-- ex: s="][a12[]/ b] c" -> Result=("a12[]/ b", " c")
			-- ex: s="[a"            -> Result=(Void, "[a")
		require
			valid_string: s /= Void
			valid_limits: a_opening /= '%U' and a_closing /= '%U' and a_opening /= a_closing
		local
			j, opened, n: INTEGER
			found, in_escape: BOOLEAN
			c: CHARACTER
			enclosed: STRING
		do
			from
				j := 1
				n := s.count
			until
				found or else j > n
			loop
				c := s.item (j)
				if in_escape then
					-- Do nothing
					if enclosed /= Void then
						enclosed.append_character (c)
					end
				elseif a_opening = c then
					if enclosed = Void then
						create enclosed.make (n)
						opened := 1
					else
						enclosed.append_character (c)
						opened := opened + 1
					end
				elseif a_closing = c then
					if opened = 1 then
						found := True
					elseif opened > 1 then
						opened := opened - 1
						if enclosed /= Void then
							enclosed.append_character (c)
						end
					end
				elseif enclosed /= Void then
					enclosed.append_character (c)
				end
				in_escape := not in_escape and then a_escape /= Void and then a_escape.has (c)
				if not found then
					j := j + 1
				end
			end
			if found then
				create Result.make (enclosed, s.substring (j + 1, n))
			else
				create Result.make (Void, s.twin)
			end
		end

	frozen split_protocol_string (a_path: STRING): PAIR [STRING, STRING]
			-- Split protocol prefixed 'a_path'
			-- 'a_path is expected to be of form "protocol://path"
			-- Returned is (type, path)
			-- type is Void if no "://" sequence
		local
			i: INTEGER
		do
			if a_path /= Void then
				i := a_path.substring_index (once "://", 1)
				if i = 0 then
					create Result.make (Void, a_path.twin)
				else
					create Result.make (
						a_path.substring (1, i - 1),
						a_path.substring (i+3, a_path.count))
				end
			end
		ensure
			result_ok: a_path /= Void implies Result /= Void
		end

	frozen maxed_string (a_text: STRING; a_max: INTEGER): attached STRING
			-- 'a_text' truncated to its 'a_max' first characters
		do
			if attached {STRING} a_text as l_text then
				Result := l_text
				if a_max > 0 and then Result.count > a_max + 2 then
					Result := Result.substring (1, a_max)
					Result.append_string (once "...")
				end
			else
				create Result.make_empty
			end
		end

	frozen head_lines (a_string: STRING; a_max_lines, a_max_characters: INTEGER): STRING
			-- Return clone of 'a_string' with at most 'a_max_lines'
			-- each of these lines with at mose 'a_max_characters' in it
			-- Use -1 for any of the integer arguments to set no limit
			-- Example: 4, 80: keep first 4 lines, with each 80 characters max
			-- Example: 4, 0: keep first 4 lines as is
		local
			i, j, k, n: INTEGER
			c: CHARACTER
		do
			if a_string /= Void then
				from
					create Result.make (100)
					i := a_max_lines
					k := 1
					n := a_string.count
				until
					k > n or else i <= 0
				loop
					from
						if a_max_characters /= 0 then
							j := a_max_characters
						else
							j := -1
						end
					until
						k > n or else j = 0
					loop
						c := a_string.item (k)
						if c = '%R' then
							-- Skip
						elseif c = '%N' then
							j := 0
							Result.extend (c)
						else
							Result.extend (c)
							j := j - 1
							if j = 0 then
								from
								until
									k > n or else a_string.item (k) = '%N'
								loop
									k := k + 1
								end
								if k <= n then
									Result.extend ('%N')
								end
							end
						end
						k := k + 1
					end
					i := i - 1
				end
				if Result.count > 0 and then Result.item (Result.count) = '%N' then
					Result.remove (Result.count)
				end
			end
		ensure
			result_ok: a_string /= Void implies (Result /= Void and then Result.count <= a_string.count)
			no_newline_at_end: (Result /= Void and then Result.count > 0) implies Result.item (Result.count) /= '%N'
			no_linefeed: Result /= Void implies not Result.has ('%R')
		end

	frozen represented_duration (a_start, a_end: ROSE_DATE; a_show_seconds: BOOLEAN): attached STRING
			-- Represented duration from 'a_start' to 'a_end'
		do
			if a_start = Void then
				Result := "?"
			elseif a_end = Void then
				Result := duration_representation ((create {ROSE_DATE}.make_now_utc).relative_duration (a_start).seconds_count.abs, a_show_seconds)
			else
				Result := duration_representation (a_end.relative_duration (a_start).seconds_count.abs, a_show_seconds)
			end
		end

	frozen argumented (a_string: STRING; a_args: ARRAY [ANY]): STRING
			-- 'a_string' argumented with 'a_args'
			-- All $i occurences in 'a_string' are replaced by the representation of 'a_args.item (i)'
			-- left unchanged if 'i' is not a valid index in 'a_args'
			-- "$*" represents all items in 'a_args' separated by a space
		local
			p, p2: PAIR [STRING, STRING]
		do
			if a_string = Void or else not a_string.has (arg_placeholder) then
				Result := a_string
			else
				from
					create Result.make (50)
					p := split_on_first (a_string, arg_placeholders, arg_escapes)
				until
					p.first = Void
				loop
					Result.append (p.first)
					if p.second = Void then
						p.make (Void, Void)
					else
						p2 := split_integer (p.second)
						Result.append (argument_value (p2.first, a_args))
						p := split_on_first (p2.second, arg_placeholders, arg_escapes)
					end
				end
			end
		ensure
			result_ok: a_string /= Void implies Result /= Void
		end

	frozen alpha_numerical_char (c: CHARACTER): BOOLEAN
			-- Is 'c' an alpha-numerical character or '_' ?
		do
			Result := c.is_alpha or else c.is_digit or else c = '_'
		end

	frozen eiffel_string (a_string: STRING): STRING
			-- Eiffel 'unescaped' sting from 'a_string'.
		do
			if a_string /= Void then
				Result := a_string.twin
				Result.replace_substring_all (once "%%N", once "%N")
				Result.replace_substring_all (once "%%T", once "%T")
			end
		ensure
			result_ok: a_string /= Void implies Result /= Void
		end

	frozen find_unescaped (a_start: INTEGER; a_string, a_char, a_escape: STRING): INTEGER
			-- Find first occurrence of some unescaped char from 'a_char' in 'a_string',
			-- starting at 'a_start' (escape character is any character from 'a_escape')
			-- 0 if no occurences from 'a_char' are found
		local
			found, in_escape, in_string: BOOLEAN
			c: CHARACTER
		do
			if a_string /= Void then
				from
					Result := a_start
				until
					found or else Result > a_string.count
				loop
					c := a_string.item (Result)
					found := not in_escape and then not in_string and then a_char.has (c)
					in_string := a_escape /= Void and then (in_string xor (not in_escape and c = '"'))
					in_escape := not in_escape and then a_escape /= Void and then a_escape.has (c)
					if not found then
						Result := Result + 1
					end
				end
				if not found then
					Result := 0
				end
			end
		end

	frozen left_adjusted (a_string: STRING): attached STRING
			-- copy of `a_string' with leading whitespace removed
		do
			if a_string = Void then
				create Result.make_empty
			else
				create Result.make_from_string (a_string)
				Result.left_adjust
			end
		ensure
			result_has_string: a_string /= Void implies a_string.has_substring (Result)
		end

	frozen right_adjusted (a_string: STRING): attached STRING
			-- copy of `a_string' with trailing whitespace removed
		do
			if a_string = Void then
				create Result.make_empty
			else
				create Result.make_from_string (a_string)
				Result.right_adjust
			end
		ensure
			result_has_string: a_string /= Void implies a_string.has_substring (Result)
		end

	frozen padded_string (a_string: STRING; a_fill: CHARACTER; a_width: INTEGER): attached STRING
			-- if `a_width' is postive then `a_string' padded with `a_fill' to count of exactly `a_width' (may be truncated)
			-- if `a_width' is negative then `a_string' padded with `a_fill' to at least `a_width.abs' (not truncated)
		local
			l_result_count: INTEGER
			l_copy_count: INTEGER
		do
			l_result_count := a_width.abs
			if a_string /= Void and then a_width < 0 then
				l_result_count := l_result_count.max (a_string.count)
			end
			create Result.make_filled (a_fill, l_result_count)
			if a_string /= Void then
				l_copy_count := a_string.count.min (l_result_count)
				if l_copy_count > 0 then
					Result.subcopy (a_string, 1, a_string.count.min (l_copy_count), 1)
				end
			end
		ensure
			valid_result: Result.count >= a_width.abs
			degenerate_case: a_string = Void implies Result.count = a_width.abs
			result_not_too_long: a_width >= 0 implies Result.count = a_width
			result_has_partial_string: a_width > 0 implies (a_string = Void or else Result.substring_index (a_string, 1) = 1 or else a_string.substring_index (Result, 1) = 1)
			result_not_truncated: a_width < 0 implies Result.count >= a_width.abs
			result_has_entire_string: a_width < 0 implies (a_string = Void or else Result.substring_index (a_string, 1) = 1)
		end

	frozen prefixed_string (a_prefix, a_string: STRING): STRING
			-- 'a_string' prefixed with 'a_prefix' at every line
		local
			l_prefix: STRING
		do
			if a_string = Void then
				-- return Void
			elseif a_prefix = Void then
				Result := a_string
			else
				l_prefix := s_newline + a_prefix
				create Result.make (a_string.count + 16)
				Result.append (a_prefix)
				Result.append (a_string)
				Result.replace_substring_all (s_newline, l_prefix)
			end
		ensure
			result_ok: a_string /= Void implies Result /= Void
		end

	separator_line: STRING = "------------------------------------------------------------------------------%N"
		-- a string of dashes (about 78) ending with a newline, recommended for separating blocks of output.

	frozen sorted (a_set: LINEAR [COMPARABLE]): SORTED_TWO_WAY_LIST [COMPARABLE]
			-- Sorted 'a_set'
		do
			if a_set /= Void then
				from
					create Result.make
					a_set.start
				until
					a_set.off
				loop
					Result.extend (a_set.item)
					a_set.forth
				end
			end
		ensure
			result_ok: a_set /= Void implies Result /= Void
		end

	frozen text_size (a_string: STRING): attached PAIR [INTEGER, INTEGER]
			-- Size (width, height) in characters of 'a_string'
			-- Result.first contains the number of character of the longest line in 'a_string'
			-- Result.second contains the number of lines in 'a_string'
		local
			i, j: INTEGER
		do
			create Result.make (0, 0)
			if a_string /= Void and then not a_string.is_empty then
				from
					i := 1
				until
					i > a_string.count
				loop
					if a_string.item (i) = '%N' then
						Result.set_first (Result.first.max (j))
						Result.set_second (Result.second + 1)
						j := 0
					else
						j := j + 1
					end
					i := i + 1
				end
				Result.set_first (Result.first.max (j))
				Result.set_second (Result.second + 1)
			end
		end

	frozen title_case_string (a_string: STRING): attached STRING
			-- 'a_string' interpreted as space-delimited words to lower case with upper case first letters
		require
			string_ok: a_string /= Void
		local
			i: INTEGER
			l_string: STRING
		do
			create Result.make_empty
			l_string := a_string.twin
			from
				i := 1
			until
				i > l_string.count
			loop
				if not l_string.valid_index (i - 1) or else l_string.item (i - 1) = ' ' then
					Result.extend (l_string.item (i).as_upper)
				else
					Result.extend (l_string.item (i).as_lower)
				end
				i := i + 1
			end
		end

	indented(a_text: STRING; a_indent_count: INTEGER; a_indent_first_line: BOOLEAN): STRING
	-- This returns a copy of 'text' but with each line indented by 'a_indent_count' spaces.
	-- If 'a_text' is Void then the Result is "".
	-- Indentation of the first line is controlled by 'a_indent_first_line'.
	-- Lines are defined by %N.  For instance:
	--		l_sql_statement := "select *%Nfrom database%N"
	--		print("Statement : " + indented(l_sql_statement, 12, False))
	-- This would print:
	--		Statement : select *
	--		            from database
	-- Note that there are no spaces after the last %N.
	require
		non_negative_index: a_indent_count>=0
	local
		l_prefix: STRING
		l_pos: INTEGER
	do
		if a_text=Void
		then
			Result := ""
		else
			if a_indent_count<=0
			then
				Result := a_text.twin
			else
				l_prefix := " "
				l_prefix.multiply(a_indent_count)	-- Note that you can't pass 0 to 'multiply'.

				if a_indent_first_line
				then
					Result := l_prefix + a_text
				else
					Result := a_text.twin
				end

				if Result.count>=2
				then
					from
						l_pos := Result.index_of('%N', 2)
					until
						l_pos=0
					loop
						if l_pos/=Result.count
						then
							Result.insert_string(l_prefix, l_pos + 1)
						end

						l_pos := Result.index_of('%N', l_pos + 1)
					end
				end
			end
		end
	ensure
		result_not_text: Result/=a_text
	end

feature -- Filename parsing

	frozen same_folder (a_path1, a_path2: STRING): BOOLEAN
			-- Are 'a_path1' and 'a_path2' pointing to the same folder?
		do
			Result := is_same_string (directory (a_path1), directory (a_path2))
		end

	frozen path (a_dirname, a_filename: STRING): attached STRING
			-- Path made up of `a_dirname' and `a_filename'
			-- and separated by a directory separator
		require
			dirname_ok: a_dirname /= Void
			filename_ok: a_filename /= Void
		do
			create Result.make_empty
			Result.append_string (a_dirname)
			if not Result.is_empty and then not directory_separators.has (Result.item (Result.count)) then
				Result.append (directory_separator)
			end
			Result.append_string (a_filename)
		end

	frozen directory (a_path: STRING): STRING
			-- Directory part of 'a_path'
		local
			p: PAIR [STRING, STRING]
		do
			if a_path /= Void and then not a_path.is_empty then
				p := split_on_last (a_path, directory_separators, Void)
				if p.second /= Void or else p.first = Void or else p.first.is_empty then
					Result := p.first
				elseif p.first.count = 1 and then p.first.item (1) = '.' then
					Result := p.first
				elseif p.first.count = 2 and then p.first.item (1) = '.' and then p.first.item (2) = '.' then
					Result := p.first
				end
			end
		ensure
			no_dir: (a_path = Void or else a_path.is_empty) implies Result = Void
		end

	frozen file_name (a_path: STRING): attached STRING
			-- File name part of 'a_path'
		local
			p: PAIR [STRING, STRING]
		do
			create Result.make_empty
			if a_path /= Void then
				p := split_on_last (a_path, directory_separators, Void)
				if p.second /= Void then
					Result.append_string (p.second)
				else
					Result.append_string (p.first)
				end
				Result.prune_all_trailing ('.')
			end
		end

	frozen file_name_no_extension (a_path: STRING): attached STRING
			-- File name part of 'a_path', without the extension
		local
			p: PAIR [STRING, STRING]
		do
			create Result.make_empty
			p := split_on_last (file_name (a_path), s_dot, Void)
			if p /= Void then
				Result.append_string (p.first)
				Result.prune_all_trailing ('.')
			end
		end

	frozen extension (a_path: STRING): STRING
			-- Extension part of 'a_path'
		local
			p: PAIR [STRING, STRING]
		do
			p := split_on_last (file_name (a_path), s_dot, Void)
			if p /= Void then
				Result := p.second
				if Result /= Void then
					Result.prune_all_leading ('.')
				end
			end
		end

	frozen parent_folder (a_path: STRING): STRING
			-- Parent folder of 'a_path', Void if no parent available
		do
			if a_path /= Void and then not file_system.is_root_directory (a_path) then
				Result := file_system.dirname (file_system.absolute_pathname (a_path))
			end
		end

	frozen string_to_pathname (a_path: STRING): KI_PATHNAME
			-- KI_PATHNAME from 'a_path', if any
		do
			if a_path /= Void then
				Result := file_system.string_to_pathname (a_path)
			end
		ensure
			result_ok: a_path /= Void implies Result /= Void
		end

	relative_path (a_root, a_path: STRING): STRING
			-- Relative part of 'a_path', relatively to 'a_root'.
		require
			root_ok: a_root /= Void
			path_ok: a_path /= Void
			path_on_root: a_path.count >= a_root.count and then a_path.substring_index (a_root, 1) = 1
		do
			Result := a_path.substring (a_root.count + 1, a_path.count)
			if Result.count > 0 and then directory_separators.has (Result.item (1)) then
				Result.remove (1)
			end
		ensure
			result_ok: Result /= Void
		end

	frozen composed_path (a_components: ARRAY [STRING]): STRING
			-- Composed path from 'a_components'
		local
			l_other_paths: ARRAY [STRING]
			l_first_path, s: STRING
			l_was_separator: BOOLEAN
			c: CHARACTER
			i, j: INTEGER
		do
			if a_components /= Void then
				from
					j := 0
					create l_other_paths.make_filled (Void, 1, 0)
					i := a_components.lower
				until
					i > a_components.upper
				loop
					s := a_components.item (i)
					if s /= Void and then not s.is_empty then
						if j = 0 then
							l_first_path := s
						else
							l_other_paths.force (s, j)
						end
						j := j + 1
					end
					i := i + 1
				end
				if l_first_path /= Void then
					inspect l_other_paths.count
					when 0 then
						s := l_first_path
					when 1 then
						s := file_system.pathname (l_first_path, l_other_paths.item (1))
					else
						s := file_system.nested_pathname (l_first_path, l_other_paths)
					end
					if s /= Void then
						-- Make sure no double path separators are present (except at the beginning of the path)
						-- Breaks Nelson's code - Also make sure returned string does not have a trailing separator
						create Result.make (s.count)
						if s.count > 0 then
							Result.append_character (s.item (1))
							if s.count > 1 then
								Result.append_character (s.item (2))
								from
									l_was_separator := s.item (2) = '\' or else s.item (2) = '/'
									i := 3
								until
									i > s.count
								loop
									c := s.item (i)
									if c = '\' or else c = '/' then
--										if not l_was_separator and then i < s.count then
										if not l_was_separator then
											Result.append_character (c)
										end
										l_was_separator := True
									else
										l_was_separator := False
										Result.append_character (c)
									end
									i := i + 1
								end
							end
						end
					end
				end
			end
		end

	frozen is_absolute_path (a_path: STRING): BOOLEAN
			-- Is 'a_path' an absolute path? False if `a_path' is Void or empty.
		do
				-- sidestep controversy around whether is_valid_path ("") should be True or False
				-- by making it explicitly False here:
			if a_path /= Void and then not a_path.is_empty then
				Result := is_valid_path (a_path) and then file_system.is_absolute_pathname (a_path)
			end
		ensure
			void_is_not_absolute: a_path = Void implies Result = False
			empty_is_not_absolute: (a_path = Void or else a_path.is_empty) implies Result = False
		end

	frozen is_relative_path (a_path: STRING): BOOLEAN
			-- Is 'a_path' a relative path?
		do
			if a_path /= Void then
					-- as of this writing, file_system.is_relative_path ("<") returns True on MSWin
					-- is_valid_path rejects special shell characters that are not allowed in MSWin path names
					-- Also, "<" and ">" are valid, though obsolescent, as directory delimiters on VMS, so this is highly platform dependent.
				Result := a_path.is_empty or else (is_valid_path (a_path) and then file_system.is_relative_pathname (a_path))
			end
		ensure
			void_is_not_relative: a_path = Void implies Result = False
			empty_is_relative: a_path /= Void and then a_path.is_empty implies Result = True
		end

	frozen is_valid_path (a_path: STRING): BOOLEAN
			-- Is 'a_path' a valid path?
		local
			l_char: CHARACTER
			i: INTEGER
		do
			if a_path /= Void and then not a_path.is_empty then
				from
					Result := True
					i := 1
				until
					not Result or else i > a_path.count
				loop
					l_char := a_path.item (i)
					Result := l_char.is_alpha_numeric or else valid_path_chars.has (l_char)
					i := i + 1
				end
			end
		end

	frozen is_valid_system_name (a_name: STRING): BOOLEAN
			-- Is 'a_name' valid as a system name?
		local
			l_char: CHARACTER
			i: INTEGER
		do
			if a_name /= Void and then not a_name.is_empty then
				from
					Result := True
					i := 1
				until
					not Result or else i > a_name.count
				loop
					l_char := a_name.item (i)
					Result := l_char.is_alpha_numeric or else valid_system_chars.has (l_char)
					i := i + 1
				end
			end
		end

	frozen is_valid_identifier (a_id: STRING): BOOLEAN
			-- Is 'a_id' a valid identifier?
		local
			l_char: CHARACTER
			i: INTEGER
		do
			if a_id /= Void and then not a_id.is_empty then
				from
					Result := a_id.item (1).is_alpha
					i := 2
				until
					not Result or else i > a_id.count
				loop
					l_char := a_id.item (i)
					Result := l_char.is_alpha_numeric or else l_char = '_'
					i := i + 1
				end
			end
		end

feature -- Double Format

	frozen parsed_double (a_text: STRING): DOUBLE
			-- Parsed double in 'a_text'
			-- Removing ',' for formatted amount
			-- Accepts 'm' or 'b' suffix to represent 'million' and 'billion'
			-- Considers only 'first word' of 'a_text' (meaning something like "25 USD" will be parsed correctly)
		local
			l_text: STRING
			l_factor: DOUBLE
			i: INTEGER
		do
			if a_text = Void or else a_text.is_empty then
				Result := ieee.qnan
			else
				l_text := a_text.twin
				l_text.left_adjust
				l_text.right_adjust
				if l_text.is_empty then
					Result := ieee.qnan
				else
					l_text.prune_all (',')
					i := l_text.last_index_of (' ', l_text.count)
					if i > 0 then
						l_text := l_text.substring (1, i - 1)
						l_text.right_adjust
					end
					i := l_text.index_of ('b', 1)
					if i > 0 then
						l_factor := one_billion
						l_text := l_text.substring (1, i - 1)
					else
						i := l_text.index_of ('m', 1)
						if i > 0 then
							l_factor := one_million
							l_text := l_text.substring (1, i - 1)
						else
							l_factor := 1.0
						end
					end
					if l_text.is_double then
						Result := l_text.to_double * l_factor
					else
						Result := ieee.qnan
					end
				end
			end
		end

	frozen formatted_amount (a_value: DOUBLE; a_decimals: INTEGER): attached STRING
			-- `a_value' represented with 'a_decimals'.
		local
			l_abs: DOUBLE
			l_decimals: INTEGER
		do
			if ieee.isnan (a_value) or else not ieee.isfinite (a_value) then
				Result := "N/A"
			else
				l_decimals := a_decimals
				if l_decimals = -1 then
					l_abs := a_value.abs
					if l_abs > ten then
						l_decimals := 0
					elseif l_abs > 1. then
						l_decimals := -1
					elseif l_abs > .1 then
						l_decimals := -2
					elseif l_abs > .01 then
						l_decimals := -3
					else
						l_decimals := -4
					end
				end
				Result := formatted_double2 (a_value, l_decimals, ',')
			end
		end

	frozen formatted_amount2 (a_value: DOUBLE; a_decimals: INTEGER): attached STRING
			-- `a_value' represented with 'a_decimals'.
		local
			l_decimals: INTEGER
			l_factor: DOUBLE
			l_suffix: CHARACTER
		do
			if ieee.isnan (a_value) or else not ieee.isfinite (a_value) then
				Result := "N/A"
			else
				l_decimals := a_decimals
				if a_value > 999.e6 then
					if a_value > 10.e9 then
						l_decimals := 0
					elseif a_value > one_billion then
						l_decimals := -1
					else
						l_decimals := -2
					end
					l_factor := one_billion
					l_suffix := 'b'
				elseif a_value > 999.e3 then
					if a_value > 10.e6 then
						l_decimals := 0
					elseif a_value > one_million then
						l_decimals := -1
					else
						l_decimals := -2
					end
					l_factor := one_million
					l_suffix := 'm'
				else
					if a_value > 10. then
						l_decimals := 0
					elseif a_value > 1. then
						l_decimals := -1
					else
						l_decimals := -2
					end
					l_factor := 1.0
					l_suffix := '%U'
				end
				if a_decimals >= 0 then
					l_decimals := a_decimals
				end
				Result := formatted_double2 (a_value / l_factor, l_decimals, ',')
				if l_suffix /= '%U' then
					Result.append_character (' ')
					Result.append_character (l_suffix)
				end
			end
		end

	frozen formatted_double2 (a_value: DOUBLE; a_decimals: INTEGER; a_separator: CHARACTER): attached STRING
			-- `a_value' represented with 'a_decimals'.
		local
			l_decimals: INTEGER
			l_integral, l_fraction: DOUBLE
			l_value: DOUBLE
			l_string: STRING
		do
			create Result.make_empty
			if a_value < 0 then
				Result.append_character ('-')
				l_value := -a_value
			else
				l_value := a_value
			end
			l_decimals := a_decimals.abs
			l_value := l_value + 5 * 10 ^ (- l_decimals - 1)
			l_integral := l_value.floor_real_64
			Result.append (split_integral (l_integral.out, a_separator))
			if not Result.has ('e') then
				if a_decimals /= 0 then
					l_fraction := ((l_value - l_integral) * 10 ^ (l_decimals + 1)).floor_real_64
					l_string := pad_fraction (l_fraction, l_decimals)
					if a_decimals < 0 then
						from
						until
							l_string.count = 0 or else l_string.item (l_string.count) /= '0'
						loop
							l_string.remove (l_string.count)
						end
					end
					if l_string.count > 0 then
						Result.extend ('.')
						Result.append (l_string)
					end
				end
			end
		end

	frozen formatted_double (a_value: DOUBLE; a_width, a_decimals: INTEGER): STRING
			-- 'a_value' formatted with 'a_width' and 'a_decimals'
		require
			reasonable_field: a_width >= 1
			reasonable_decimals: a_decimals <= a_width
		local
			l_format_double: FORMAT_DOUBLE
		do
			create l_format_double.make (a_width, a_decimals)
			Result := l_format_double.formatted (a_value)
		ensure
			result_ok: Result /= Void
		end

	frozen formatted_number (a_value: DOUBLE; a_width, a_decimals: INTEGER): STRING
			-- Formatted version of `a_value' as integer or fixed point, with commas
		local
			l_format_double: FORMAT_DOUBLE
		do
			if a_value.truncated_to_integer.to_double = a_value then
				Result := formatted_integer (a_value.truncated_to_integer, a_width, ',')
			else
				create l_format_double.make (a_width, a_decimals)
				l_format_double.comma_separate
				l_format_double.no_separate_after_decimal
				Result := l_format_double.formatted (a_value)
			end
		ensure
			result_ok: Result /= Void
		end

feature -- Integer Format

	frozen formatted_integer (a_value: INTEGER; a_width: INTEGER; a_separator: CHARACTER): STRING
			-- 'a_value' formatted with 'a_width' with `a_separator'
		require
			reasonable_field: a_width >= 1
		do
			Format_integer.make (a_width)
			Format_integer.set_separator (a_separator)
			Result := Format_integer.formatted (a_value)
		ensure
			result_ok: Result /= Void
		end

	frozen formatted_integer_zero_filled, zero_filled_decimal (a_val, a_width: INTEGER): STRING
			-- decimal representation of `a_val' left zero filled to at least `a_width'
		require
			width_reasonable: a_width >= 1
		do
			Format_integer.make (a_width)
			Format_integer.set_fill ('0')
			Result := Format_integer.formatted (a_val)
		ensure
			result_is_long_enough: Result /= Void and then Result.count >= a_width
			result_is_equal_integer: Result.is_integer and then Result.to_integer = a_val
		end

	frozen ordinal (a_val: INTEGER): attached STRING
			-- `a_val' as an ordinal string ("1st", "2nd", "3rd", "4th", "10th", "101st", etc.)
		do
			create Result.make_empty
			Result.append_integer (a_val)
			Result.append_string (ordinal_suffix (a_val))
		ensure
			not_void_or_blank: not Result.is_empty
			matches_value: Result.count >= 3 and then Result.substring (1, Result.count - 2).to_integer = a_val
		end

	frozen ordinal_suffix (a_val: INTEGER): attached READABLE_STRING_8
			-- ordinal suffix for `a_val' (if 1: "st"; if 2: "nd"; if 3: "rd"; if 4: "th", etc.)
		local
			l_mod10, l_mod100: INTEGER
		do
			l_mod10 := a_val.abs \\ 10
			l_mod100 := a_val.abs \\ 100
			if l_mod10 = 1 and l_mod100 /= 11 then
				Result := once "st"
			elseif l_mod10 = 2 and l_mod100 /= 12 then
				Result := once "nd"
			elseif l_mod10 = 3 and l_mod100 /= 13 then
				Result := once "rd"
			else
				Result := once "th"
			end
		ensure
			is_long_enough: Result.count = 2
		end

feature -- Time formatting

	frozen elapsed_time_now_mmss (a_date: ROSE_DATE): attached STRING
			-- elapsed time from `a_date' to now formatted as "mm:ss" rounded to nearest second;
			-- if > 59:59 then hours (as hh:) and days (as d+) will be prepended as necessary
		require
			date_not_void: a_date /= Void
		do
			Result := elapsed_time_mmss (a_date, (create {ROSE_DATE}.make_now))
		ensure
			result_has_mm_ss: Result.count >= 5		--("mm.ss").count
		end

	elapsed_time_now_mmssfff (a_date: ROSE_DATE): attached STRING
			-- formatted from `a_date' to now time as "d+hh:mm:ss.fff"
			-- zero days and hours are omitted (therefore {TIME}.formatted_out cannot be used)
		require
			date_not_void: a_date /= Void
		do
			Result := elapsed_time_mmssfff (a_date, create {ROSE_DATE}.make_now)
		ensure
			result_has_mm_ss: Result.count >= 9		--("mm.ss.fff").count
		end

	frozen elapsed_time_mmss (a_start, a_end: ROSE_DATE): attached STRING
			-- elapsed time from `a_start' to `a_end' formatted as "mm:ss" rounded to nearest second;
			-- if >= 1 hour then hours (as "hh:") and days (as "d+") will be prepended as necessary
			-- if `a_start' > `a_end' then a minus sign is prepended to `Result'
		require
			start_date_not_void: a_start /= Void
			end_date_not_void: a_start /= Void
		local
			l_dur: DATE_TIME_DURATION
		do
			l_dur := a_end.relative_duration_definite (a_start)
			check
				-- duration may have days only (no have months or years) for correct computation
				duration_is_definite: l_dur.date.year = 0 and then l_dur.date.month = 0
			end
			create Result.make (12)
			if l_dur.date.days_count /= 0 then
				Result.append (zero_filled_decimal (l_dur.date.days_count, 1))
				Result.append_character ('+')
			else
				if a_start > a_end then
					Result.append_character ('-')
				end
			end
			if l_dur.time.hour /= 0 or else l_dur.date.days_count /= 0 then
				Result.append (zero_filled_decimal (l_dur.time.hour.abs, 2))
				Result.append_character (':')
			end
			Result.append (zero_filled_decimal (l_dur.time.minute.abs, 2))
			Result.append_character (':')
			Result.append (zero_filled_decimal (l_dur.time.fine_second.abs.rounded, 2))
		ensure
			result_has_mm_ss: Result.count >= 5
			result_is_signed_iff_negative: (a_start > a_end) = (Result.starts_with ("-"))
		end

	elapsed_time_mmssfff (a_start, a_end: ROSE_DATE): attached STRING
			-- formatted time from `a_start' to `a_end' as "d+hh:mm:ss.fff"
			-- zero days and hours are omitted (therefore {TIME}.formatted_out cannot be used)
			-- if `a_start' > `a_end' then a minus sign is prepended to `Result'
		require
			start_not_void: a_start /= Void
			end_not_void: a_end /= Void
		do
			Result := elapsed_time_mmssfff_from_duration(a_end.relative_duration_definite (a_start), Void)
		end

	elapsed_time_mmssfff_from_duration (a_dur: DATE_TIME_DURATION; a_days_separator: STRING): attached STRING
			-- Like 'elapsed_time_mmssfff' but using a duration.  The duration may be positive or negative.
		local
			l_str: STRING
			l_dur: DATE_TIME_DURATION
		do
			if a_dur.is_negative
			then
				Result := "-"
				l_dur := -a_dur
			else
				Result := ""
				l_dur := a_dur
			end

			if l_dur.date.days_count /= 0 then
				Result.append (zero_filled_decimal (l_dur.date.days_count, 1))

				if a_days_separator = Void
				then
					Result.append_character ('+')
				else
					Result.append(a_days_separator)
				end
			end
			if l_dur.time.hour /= 0 or else l_dur.date.days_count /= 0 then
				Result.append (zero_filled_decimal (l_dur.time.hour, 2))
				Result.append_character (':')
			end
			Result.append (zero_filled_decimal (l_dur.time.minute, 2))
			Result.append_character (':')
			format_double.make (6, 3)				-- ("ss.fff").count, ("fff").count
			format_double.show_zero
			format_double.no_separate_after_decimal
			format_double.zero_fill
			l_str := format_double.formatted (l_dur.time.fine_second)
			Result.append (l_str)
		ensure
			result_has_mm_ss: Result.count >= 9		-- ("mm.ss.fff").count
		end

	elapsed_time_ddhhmmss (a_start, a_end: attached ROSE_DATE): attached STRING
			-- formatted elapsed time, as [dd+]hh:mm:ss (days are optional)
		require
			start_not_void: a_start /= Void
			end_not_void: a_end /= Void
		local
			l_dur: DATE_TIME_DURATION
		do
			Result := elapsed_time_mmss (a_start, a_end)
			l_dur := a_end.relative_duration_definite (a_start)
			if l_dur.time.hour = 0 and then l_dur.date.days_count = 0 then
				if Result[1] = '-' then
					Result.replace_substring ("-00:", 1, 1)
				else
					Result.prepend ("00:")
				end
			end
		ensure
			not_void: Result /= Void
			has_hh_mm_ss: Result.count >= ("hh:mm:ss").count
		end

feature -- Date parsing

	frozen make_date_from_string (a_string: STRING): ROSE_DATE
			-- Date from 'a_string' of the form "year/month/day" or "CCYYMMDD"
			-- or "YYMMDD" (century guessed)
			-- or "mm/dd/ccyy" (MZ, 20090130)
			-- Void if 'a_string' does not comply format
		local
			p: PAIR [STRING, STRING]
			a: ARRAY [STRING]
			i, j, l_yy: INTEGER
			l_ccyy_string: STRING
			l_sec: STRING
			l_time: TIME
		do
			if a_string = Void then
				-- return Void
			elseif a_string.is_integer then
				i := a_string.to_integer
				if i < 10000 or else i > 100000000 then
					-- Void
				elseif i < 1000000 then
					j := i // 10000
					if j < 50 then
						j := j + 2000
					else
						j := j + 1900
					end
					i := i \\ 10000
					create Result.make (j, i//100, i\\100)
				elseif i < 100000000 then
					j := i // 10000
					i := i \\ 10000
					create Result.make (j, i//100, i\\100)
				end
			elseif date_expression.is_expression (a_string) then
				Result := date_expression.item (a_string)
			else
				p := split_on_first (a_string, date_time_separators, Void)
				-- Date part
				a := split_array (p.first, date_separators, Void, False)
				if a.count = 3 then
					if a.item (1) /= Void and then a.item (1).count = 4 then
						Result := make_date_from_strings (a.item (1), a.item (2), a.item (3))
					elseif a.item (3) /= Void and then a.item (3).count = 4 then
						-- mm/dd/ccyy format
						Result := make_date_from_strings (a.item (3), a.item (1), a.item (2))
					elseif (a.item (1) /= Void and then a.item (1).count = 2) and (a.item (3) /= Void and then a.item (3).count = 2) then
						-- default to mm/dd/yy
						if a.item (3).is_integer then
							l_yy := a.item (3).to_integer
							if l_yy >= 50 then
								l_ccyy_string := "19" + a.item (3)
							else
								l_ccyy_string := "20" + a.item (3)
							end
							Result := make_date_from_strings (l_ccyy_string, a.item (1), a.item (2))
						end
					else
						Result := make_date_from_strings (a.item (1), a.item (2), a.item (3))
					end
				end
				-- Time part
				if Result /= Void and then p.second /= Void then
					p := split_on_first (p.second, s_dot, Void)
					a := split_array (p.first, s_colon, Void, False)
					if a.count = 2 or a.count = 3 then
						if a.count = 3 then
							l_sec := a.item (3)
						end
						l_time := make_time_from_strings (a.item (1), a.item (2), l_sec, p.second)
						if l_time /= Void then
							Result.set_time (l_time)
						end
					end
				end
			end
			if Result /= Void and then not Result.is_available then
				Result := Void
			end
		ensure
			result_ok: Result /= Void implies Result.is_available
		end

	frozen make_month_from_string (a_string: STRING): ROSE_DATE
			-- Date from 'a_string'
			-- or "CCYYMM"
			-- or julian_month (century guessed)
			-- Void if 'a_string' does not comply format
		local
			i: INTEGER
		do
			if a_string = Void then
				-- Void
			elseif a_string.is_integer then
				i := a_string.to_integer
				if i < 1 or i > 999912 then
					-- Void
				elseif i < 1000 then
					create Result.make_from_julian_month (i)
				else
					create Result.make (i//100, i\\100, 1)
				end
			elseif date_expression.is_expression (a_string) then
				Result := date_expression.item (a_string)
			end
			if Result /= Void and then not Result.is_available then
				Result := Void
			end
		ensure
			result_ok: Result /= Void implies Result.is_available
		end

	frozen make_time_from_string (a_string: STRING): ROSE_TIME
			-- Date from 'a_string' of the form "hour:min[:sec[.millisec]]"
			-- Void if 'a_string' does not comply format
		local
			p: PAIR [STRING, STRING]
			a: ARRAY [STRING]
			l_sec: STRING
		do
			if equal (a_string, now_keyword) then
				create Result.make_now
			elseif a_string /= Void then
				p := split_on_first (a_string, s_dot, Void)
				a := split_array (p.first, s_colon, Void, False)
				if a.count = 2 or a.count = 3 then
					if a.count = 3 then
						l_sec := a.item (3)
					end
					Result := make_time_from_strings (a.item (1), a.item (2), l_sec, p.second)
				end
			end
		end

	frozen make_date (y, m, d: INTEGER): ROSE_DATE
			-- Date from 'y', 'm', 'd' (Void if not valid)
		do
			if y > 0
			and then m >= 1 and then m <= 12
			and then d >= 1 and then d <= date_constants.days_in_i_th_month (m, y) then
				create Result.make (y, m, d)
			end
			if Result /= Void and then not Result.is_available then
				Result := Void
			end
		ensure
			result_ok: Result /= Void implies Result.is_available
		end

	frozen make_date_from_strings (y, m, d: STRING): ROSE_DATE
			-- Date from 'y', 'm', 'd' strings (Void if not valid)
		local
			l_y, l_m, l_d, i: INTEGER
		do
			l_y := to_integer (y)
			l_m := to_integer (m)
			l_d := to_integer (d)
			if l_d > 31 and then l_y <= 31 then
				-- MM/DD/YYYY format
				i := l_y
				l_y := l_d
				l_d := l_m
				l_m := i
			end
			Result := make_date (l_y, l_m, l_d)
		end

	frozen make_time (h, m, s, ms: INTEGER): ROSE_TIME
			-- Time from 'h', 'm', 'd', 'ms' (Void if not valid)
		do
			if h >= 0 and h < 24
			and m >= 0 and m < 60
			and s >=0 and s < 60
			and ms >= 0 and ms < 1000 then
				create Result.make_fine (h, m, s + ms / 1000)
			end
		end

	frozen make_time_from_strings (h, m, s, ms: STRING): ROSE_TIME
			-- Time from 'h', 'm', 'd', 'ms' strings (Void if not valid)
		do
			Result := make_time (to_integer (h), to_integer (m), to_integer (s), to_integer (ms))
		end

feature -- Date expression

	frozen date_expression: attached ROSE_DATE_DOT_EXPRESSION
			-- Shared date expression processor.
		once
			create Result.make
		end

feature -- Constants

	frozen Format_integer: attached FORMAT_INTEGER
			-- integer formatter (reusable)
		once
			create Result.make (10)
		end

	frozen Format_double: attached FORMAT_DOUBLE
			-- double formatter (reusable)
		once
			create Result.make (10, 4)
		end

	frozen Time_constants: attached TIME_CONSTANTS
			-- time constants
		once
			create Result
		end

	frozen Date_constants: attached DATE_CONSTANTS
			-- Utility for date validity testing
		once
			create Result
		end

	ten: DOUBLE = 10.
	one_million: DOUBLE = 1.0e6 		-- 1,000,000
	one_billion: DOUBLE = 1.0e9			-- 1,000,000,000 (should be called one_sagan?)


feature {NONE} -- Constants

	s_double_quote: STRING = "%""							-- Double-quote
	s_dot: STRING = "."										-- Dot
	s_colon: STRING = ":"									-- Colon
	arg_placeholder: CHARACTER = '$'						-- Argument placeholder
	arg_placeholders: STRING = "$"							-- Argument placeholders
	arg_escapes: STRING = "%%"								-- Argument escapes
	date_separators: STRING = "/-."							-- Date separators
	date_time_separators: STRING = " %T"					-- Date-time separators
	s_exe: STRING = ".exe"									-- .exe
	s_newline: STRING = "%N"								-- Newline
	variable_placeholder: CHARACTER = '$'					-- Variable placehorlder delimiter

	valid_system_chars: STRING = "_.-= ()!@#$%%^&[]{}'"		-- Valid characters for a system name
	valid_path_chars: STRING = "\/:_.-= ()!@#$%%^&[]{}'"	-- Valid characters for a path

feature {NONE} -- Implementation

	frozen ieee: attached IEEE_MATH
			-- For particular states of DOUBLE
		once
			create Result
		end

	frozen to_integer (a_string: STRING): INTEGER
			-- Integer value of 'a_string' if any (0 if none)
		do
			if a_string /= Void and then a_string.is_integer then
				Result := a_string.to_integer
			end
		end

	frozen split_integer (a_string: STRING): attached PAIR [STRING, STRING]
			-- Break 'a_string' in two strings, the first containing an integer or '*'
			-- ex: "ab"    -> (Void, "ab")
			-- ex: "1ab"   -> ("1", "ab")
			-- ex: "123.0ab" -> ("123", ".0ab")
			-- ex: ".0ab" -> (Void, ".0ab")
			-- ex: "*ab" -> ("*", "ab")
		local
			i: INTEGER
		do
			if a_string = Void or else a_string.is_empty then
				create Result.make (Void, a_string)
			else
				from
					i := 1
				until
					i > a_string.count or else not a_string.item (i).is_digit
				loop
					i := i + 1
				end
				if i = 1 and then a_string.item (1) = '*' then
					i := 2
				end
				create Result.make (a_string.substring (1, i - 1), a_string.substring (i, a_string.count))
			end
		ensure
			result_ok: Result /= Void
		end

	frozen argument_value (a_string: STRING; a_args: ARRAY [ANY]): STRING
			-- Argument value pointed by 'a_string' (if any)
		local
			i: INTEGER
		do
			if a_string = Void or else a_string.is_empty then
				Result := arg_placeholders
			elseif a_string.count = 1 and then a_string.item (1) = '*' then
				Result := join_array (a_args, once " ")
			elseif a_string.is_integer then
				i := a_string.to_integer
				if a_args = Void or else a_args.lower > i or else a_args.upper < i then
					Result := arg_placeholders + a_string
				else
					Result := object_representation (a_args.item (i), once "-Void-")
				end
			else
				Result := arg_placeholders + a_string
			end
		ensure
			result_ok: Result /= Void
		end

end
