indexing
	description: "Framework conventions"
	notice: "[
		This class is documentation on framework conventions
		Please comply to conventions listed here when making an application
]"
	author: "Zoran Simic"
	last_editor: "$Author: c4 $"
	date: "$Date: 2011/11/29 $"
	revision: "$Revision: #24 $"
	path: "$File: //rose/source/kernel/infrastructure/framework/application/rose_framework_convention.e $"

class
	ROSE_FRAMEWORK_CONVENTION

inherit
	ROSE_IO_CONVENTION

feature -- Project Constants

	common_project: STRING is "common"
			-- Common project, gathering shared files amongst all projects

feature -- Configuration constants

	default_config_extension: STRING is
			-- Filename extension for default configuration type
		once
			Result := "." + default_config_type
		end

	default_config_type: STRING is "cfg"
			-- Default configuration type

	system_keyword: STRING is "system"
	today_keyword: STRING is "today"
	now_keyword: STRING is "now"
	now_utc_keyword: STRING is "now_utc"
	running_keyword: STRING is "running"
			-- Various keywords recognized as special-meaning variables

feature -- Command processing constants

	command_separators: STRING is "&%N"		-- Characters seprating commands
	command_escapes: STRING is "%%"				-- Characters escaping command lines
	argument_separators: STRING is " %T"
			-- Special characters in command line parsing

	command_parameter_qualifier_required: CHARACTER is '='
	command_parameter_qualifier_optional: CHARACTER is ':'
	all_command_parameter_qualifiers: STRING is "=:"
			-- Set of characters understood as command parameter qualifiers in signatures

	command_parameter_type_string: CHARACTER is 's'
			-- Character used to specify command parameter types in signatures;
			-- Example: "path:s"
			-- 'path' is an optional string

	command_parameter_type_character: CHARACTER is 'c'
			-- Character used to specify command parameter types in signatures;
			-- Example: "mode=c"
			-- 'mode' is a required character.

	command_parameter_type_zone: CHARACTER is 'a'
			-- Character used to specify command parameter types in signatures;
			-- Example: "zone=a"
			-- 'zone' is a required investment zone.
	
	command_parameter_type_equity_zone: CHARACTER is 'z'
			-- Character used to specify command parameter types in signatures;
			-- Example: "zone=z"
			-- 'zone' is a required equity investment zone.
	
	command_parameter_type_integer: CHARACTER is 'i'
			-- Character used to specify command parameter types in signatures;
			-- Example: "iterations=i"
			-- 'iterations' is a required integer.
	
	command_parameter_type_boolean: CHARACTER is '?'
			-- Character used to specify command parameter types in signatures;
			-- Example: "free_float:?"
			-- 'free_float' is an optional boolean.

	command_parameter_type_double: CHARACTER is 'r'
			-- Character used to specify command parameter types in signatures;
			-- Example: "threshold:r"
			-- 'threshold' is an optional IEEE-754 double precision floating point value.
	
	command_parameter_type_date: CHARACTER is 'd'
			-- Character used to specify command parameter types in signatures;
			-- Example: "start_date=d"
			-- 'start_date' is a required date in ccyymmdd or other formats.

	command_parameter_type_date_time: CHARACTER is 't'
			-- Use `command_parameter_type_date' instead.
			-- Character used to specify command parameter types in signatures;
			-- Example: "start_date=t"
			-- 'start_date' is a required date in ccyymmdd or other formats.

	command_parameter_type_month: CHARACTER is 'm'
			-- Character used to specify command parameter types in signatures;
			-- Example: "start_date=m"
			-- 'start_date' is a required date in ccyymm or julian month formats.

	framework_default_command_line: STRING is "console"
			-- Default command line executed by the framework

	date_format: STRING is "[0]mm/[0]dd/yyyy"
	time_format: STRING is "[0]mm/[0]dd/yyyy hh:[0]mi:[0]ss"
			-- Format to use to display a date and time objects

feature -- Database constants

	db_type_sql: STRING is "sql"					-- Identifier for SQL server databases
	db_type_ms_sql: STRING is "ms_sql"				-- Alias for 'sql'
	db_type_rdb_service: STRING is "rdb_service"	-- RDB with service
	db_type_rdb_generic: STRING is "rdb_generic"	-- Generic RDB connection

feature -- Access

	command_parameter_type_representation (a_type: CHARACTER): STRING is
			-- Command parameter type representation of 'a_type'
		do
			inspect a_type
			when command_parameter_type_string then
				Result := "STRING"
			when command_parameter_type_character, command_parameter_type_equity_zone, command_parameter_type_zone then
				Result := "CHARACTER"
			when command_parameter_type_integer then
				Result := "INTEGER"
			when command_parameter_type_boolean then
				Result := "BOOLEAN"
			when command_parameter_type_double then
				Result := "DOUBLE"
			when command_parameter_type_date then
				Result := "DATE"
			when command_parameter_type_date_time then
				Result := "DATE_TIME"
			when command_parameter_type_month then
				Result := "MONTH"
			else
				Result := "UNKNOWN"
			end
		end

feature -- Status report

	is_registered_project (a_project: STRING): BOOLEAN is
			-- Is 'a_project' registered?
		do
			Result := True
		end

	is_common_project (a_project: STRING): BOOLEAN is
			-- Is 'a_project' the "common" project?
		do
			Result := common_project ~ a_project
		end

	is_valid_application_name (a_name: STRING): BOOLEAN is
			-- Is 'a_name' a valid application name?
		do
			Result := a_name /= Void and then not a_name.is_empty
		end

	is_valid_version (a_version: STRING): BOOLEAN is
			-- Is 'a_version' of a valid format?
			-- Conventional format: M.m.r, M=Major version number, m=minor, r=revision
		local
			i: INTEGER
		do
			Result := a_version.count >= 5 and then a_version.item (1).is_digit
			if Result then
				i := check_version_format (a_version, 1)
				Result := i > 1
			end
			if Result then
				i := check_version_format (a_version, i)
				Result := i > 1
			end
			if Result then
				i := check_version_format (a_version, i)
				Result := i = a_version.count + 1
			end
		end

	is_valid_command_parameter_type (a_type: CHARACTER): BOOLEAN is
			-- Is 'a_type' valid for a command parameter?
		do
			Result := a_type = command_parameter_type_string or
				a_type = command_parameter_type_character or
				a_type = command_parameter_type_integer or
				a_type = command_parameter_type_boolean or
				a_type = command_parameter_type_double or
				a_type = command_parameter_type_date or
				a_type = command_parameter_type_date_time or
				a_type = command_parameter_type_month or
				a_type =	command_parameter_type_zone or
				a_type = command_parameter_type_equity_zone
		end

feature {NONE} -- Implementation

	check_version_format (a_string: STRING; a_start: INTEGER): INTEGER is
			-- Check that 'a_string' has a proper version format starting at 'a_start'
		do
			from
				Result := a_start
			until
				Result > a_string.count or else not a_string.item (Result).is_digit
			loop
				Result := Result + 1
			end
			if Result > a_string.count then
				-- Do nothing
			elseif Result = a_string.count then
				Result := 0
			elseif a_string.item (Result) = '.' then
				Result := Result + 1
			else
				Result := 0
			end
		end

end
