indexing
	description: "Framework preference settings"
	author: "Zoran Simic"
	date: "$Date: 2008/12/01 $"
	revision: "$Revision: #6 $"

class
	ROSE_FRAMEWORK_PREFERENCE

inherit
	ROSE_IO_CONVENTION

create
	make

feature {NONE} -- Initialization

	make is
			-- Create a default set of system preferences.
		do
			update_from_config (Void)
			prefix_error := "** "
			prefix_warning := "!! "
			prefix_log := ">> "
			prefix_trace := ":: "
		end

feature -- Access

	date_format: STRING
			-- Date format to be used when outputing dates

	date_time_format: STRING
			-- Date/time format to be used when outputing date/times

	prompt: STRING
			-- User prompt to use

	decorated_verbose_message (a_level: INTEGER; a_message: STRING): STRING is
			-- 'a_message' decorated for 'a_level'
		local
			l_prefix: STRING
			l_needs_newline: BOOLEAN
		do
			if a_message = Void or else a_message.is_empty then
				if a_level = v_error then
					Result := prefix_error
				elseif a_level = v_warning then
					Result := prefix_warning
				elseif a_level = v_log then
					Result := prefix_log
				elseif a_level = v_trace then
					Result := prefix_trace
				else
					Result := a_message
				end
			else
				if a_level = v_error then
					l_prefix := prefix_error
				elseif a_level = v_warning then
					l_prefix := prefix_warning
				elseif a_level = v_log then
					l_prefix := prefix_log
				elseif a_level = v_trace then
					l_prefix := prefix_trace
				end
				if l_prefix = Void then
					Result := a_message
				else
					create Result.make (a_message.count+16)
					Result.append (l_prefix)
					Result.append (a_message)
					if Result.item (Result.count) = '%N' then
						l_needs_newline := True
						Result.remove (Result.count)
					else
						l_needs_newline := a_level > v_output
					end
					Result.replace_substring_all (s_newline,s_newline + l_prefix)
					if l_needs_newline then
						Result.extend ('%N')
					end
				end
			end
		end

	prefix_error: STRING
	prefix_warning: STRING
	prefix_log: STRING
	prefix_trace: STRING
			-- Default prefixes used to highlight text when outputting plain text

feature -- Status report

	echo_commands: BOOLEAN
			-- Echo executed commands?

feature -- Status setting

	set_date_format (a_value: STRING) is
		do
			date_format := get_value (a_value,default_date_format)
		end

	set_date_time_format (a_value: STRING) is
		do
			date_time_format := get_value (a_value,default_date_time_format)
		end

	set_prompt (a_value: STRING) is
		do
			prompt := get_value (a_value,default_prompt)
		end

	set_echo_commands (a_value: BOOLEAN) is
		do
			echo_commands := a_value
		end

feature -- Basic operations

	update_from_config (a_node: ROSE_CONFIGURATION_NODE) is
			-- Update prefs from 'a_node'
		do
			if a_node /= Void then
				set_date_format (a_node.child_value ("date_format",Void))
				set_date_time_format (a_node.child_value ("date_time_format",Void))
				set_prompt (a_node.child_value ("prompt",Void))
				set_echo_commands (to_boolean (a_node.child_value ("echo",Void)))
			else
				set_date_format (Void)
				set_date_time_format (Void)
				set_prompt (Void)
				set_echo_commands (False)
			end
		end

feature -- Constants

	default_date_format: STRING is "yyyy/[0]mm/[0]dd"
	default_date_time_format: STRING is "yyyy/[0]mm/[0]dd [0]hh:[0]mi:[0]ss"
	default_prompt: STRING is "> "

feature {NONE} -- Implementation

	get_value (a_value, a_default: STRING): STRING is
			-- Return 'a_value' if meaningful, 'a_default' otherwise
		do
			if a_value /= Void and then not a_value.is_empty then
				Result := a_value
			else
				Result := a_default
			end
		end

	to_boolean (a_string: STRING): BOOLEAN is
		do
			if a_string = Void or else a_string.is_empty then
				--| Do nothing
			elseif a_string.is_integer then
				Result := a_string.to_integer /= 0
			elseif a_string.is_boolean then
				Result := a_string.to_boolean
			else
				a_string.to_lower
				Result := equal (a_string,s_yes)
			end
		end

	s_yes: STRING is "yes"
	s_newline: STRING is "%N"

end -- class ROSE_FRAMEWORK_PREFERENCE
