indexing
	description: "GUI relative global applicationwide data"
	author: "Zoran Simic"
	date: "$Date: 2010/02/11 $"
	revision: "$Revision: #50 $"

frozen class
	ROSE_GLOBAL_GUI

inherit
	ROSE_SINGLETON
	ROSE_PROGRESS
		redefine
			trigger_busy_event
		end

create
	make

feature {NONE} -- Initialization

	make is
			-- Initialize a default GUI-relative global object.
		do
			check_instance_unicity
			busy_event.extend (agent handle_busy_message (Void, ?))
			invoker_cell.put (agent ev_application.do_once_on_idle)
		end

feature -- Access

	ev_application: EV_APPLICATION is
			-- Single application object for system.
		local
			l_environment: EV_ENVIRONMENT
		once
			create l_environment
			Result := l_environment.application
			if Result = Void then
				create Result.default_create
			end
		ensure
			result_ok: Result /= Void
		end

	screen: EV_SCREEN is
			-- Eiffel Vision screen object
		once
			create Result
		end

	console: RCMD_EV_CONSOLE is
			-- Eiffel Vision console
		once
			create Result
		end

	error_reporter: RCMD_EV_ERROR_REPORTER is
			-- Eiffel Vision error reporter dialog (pops up automatically when errors output)
		once
			create Result
		end

	quit: RCMD_EV_QUIT is
			-- Eiffel Vision quit command
		once
			create Result
		end

	monitor: RCMD_EV_MONITOR is
			-- Eiffel Vidion monitor dialog and command
		once
			create Result
		end

	idle_auto_close_timeout: INTEGER
			-- Number of minutes of user-idle time to allow before automatically closing the application

	opened_windows: INTEGER
			-- Number of opened windows, application exits automatically when no more windows are opened

	repositionner: ROSE_WINDOW_REPOSITIONNER is
			-- Window repositionner (persists window sizes on GUI configuration)
		once
			create Result.make
		end

	factory: ROSE_EV_FACTORY is
			-- Eiffel Vision object factory
		once
			create Result
		end

	remembered: ROSE_EV_FACTORY_PERSISTING is
			-- Factory of persisting settings per user (network wide)
		once
			create Result.make (glob.Global.user)
		end

	locally_remembered: ROSE_EV_FACTORY_PERSISTING is
			-- Factory of persisting settings per user (on this machine only)
		once
			create Result.make (glob.Global.login)
		end

	rep: ROSE_EV_REPOSITORY is
			-- Eiffel Vision common objects repository
		once
			create Result
		end

	dialogs: ROSE_EV_COMMON_DIALOGS is
			-- Common dialog facilities
		once
			create Result
		end

	parent_window (a_widget: EV_WIDGET): EV_WINDOW is
			-- Parent window of 'a_widget'
		local
			l_widget: EV_WIDGET
		do
			from
				l_widget := a_widget
			until
				Result /= Void or else l_widget = Void
			loop
				Result ?= l_widget
				l_widget := l_widget.parent
			end
			if Result /= Void and then Result.is_destroyed then
				Result := Void
			end
		ensure
			result_ok: Result /= Void implies not Result.is_destroyed
		end

	parent_titled_window (a_widget: EV_WIDGET): EV_TITLED_WINDOW is
			-- Parent titled window of `a_widget'.
		do
			Result ?= parent_window (a_widget)
		end

	is_picking: BOOLEAN is
			-- Is a pick-and-drop operation currently in progress?
		do
			Result := platform_specifics.is_picking (ev_application)
		end

	is_running: BOOLEAN is
			-- Is event loop running?
		do
			Result := not ev_application.is_destroyed and then ev_application.is_launched
		end

	new_busy_cursor (a_cursor: EV_CURSOR; a_message: STRING): EV_CURSOR is
			-- New cursor based on 'a_cursor' agremented with 'a_message'
		local
			l_message: STRING
			l_font: EV_FONT
			l_width, l_height: INTEGER
		do
			Result := a_cursor
			if Result = Void then
				Result := rep.pixmaps.wait_cursor
			end
			if not simple_cursors and then a_message /= Void and then not a_message.is_empty then
				if a_message.has ('%N') then
					l_message := a_message
				else
					l_message := a_message + "%NPlease wait..."
				end
				Result := Result.twin
				l_message.prune_all ('%R')
				l_width := Result.width
				l_height := Result.height
				l_font := Result.font
				Result.set_font (l_font)
				Result.set_size (l_width + l_font.string_width (l_message) + 4, l_height)
				Result.set_foreground_color (rep.colors.dark_blue)
				Result.draw_text_top_left (l_width + 2, 1, l_message)
				Result.draw_rectangle (l_width, 0, Result.width - l_width, l_height - 1)
			end
		ensure
			result_ok: Result /= Void
		end

feature -- Basic operations

	invoke (a_agent: PROCEDURE [ANY, TUPLE []]) is
			-- Invoke 'a_agent', execute it on main thread if possible
		do
			ev_application.do_once_on_idle (a_agent)
		end

	set_idle_auto_close_timeout (a_value: INTEGER) is
			-- Set number of minutes of user-idle time to allow before automatically closing the application
		require
			value_ok: a_value >= 0
		do
			-- FIXME: to be reimplemented
			idle_auto_close_timeout := a_value
		end

	output (a_textable: EV_TEXTABLE; a_message: STRING; a_args: ARRAY [ANY]) is
			-- Output 'a_message' with 'a_args' to 'a_textable'.
		local
			l_string: STRING
		do
			if a_textable /= Void then
				l_string := formatter.argumented (a_message, a_args)
				if l_string /= Void then
					l_string.prune_all ('%R')
					a_textable.set_text (l_string)
				else
					a_textable.remove_text
				end
			end
		end

	enable_sensitive (a_widget: EV_SENSITIVE; a_sensitive: BOOLEAN) is
			-- Enable 'a_widget' as 'a_sensitive'.
		do
			if a_widget /= Void and then not a_widget.is_destroyed then
				if a_sensitive then
					if not a_widget.is_sensitive then
						a_widget.enable_sensitive
					end
				else
					if a_widget.is_sensitive then
						a_widget.disable_sensitive
					end
				end
			end
		end

	enable_select (a_widget: EV_SELECTABLE; a_selected: BOOLEAN) is
			-- Enable 'a_widget' as 'a_selected'.
		local
			l_deselectable: EV_DESELECTABLE
		do
			if a_widget /= Void and then not a_widget.is_destroyed then
				if a_selected then
					if a_widget.is_selectable and then not a_widget.is_selected then
						a_widget.enable_select
					end
				else
					l_deselectable ?= a_widget
					if l_deselectable /= Void and then l_deselectable.is_selected then
						l_deselectable.disable_select
					end
				end
			end
		end

	enable_edit (a_widget: EV_TEXT_COMPONENT; a_edit: BOOLEAN) is
			-- Enable 'a_widget' as 'a_edit'-able.
		do
			if a_widget /= Void and then not a_widget.is_destroyed then
				if a_edit then
					if not a_widget.is_editable then
						a_widget.enable_edit
					end
				else
					if a_widget.is_editable then
						a_widget.disable_edit
					end
				end
			end
		end

	show (a_widget: EV_WIDGET; a_show: BOOLEAN) is
			-- Show 'a_widget' accordingly to 'a_show'.
		local
			l_window: EV_WINDOW
		do
			if a_widget /= Void and then not a_widget.is_destroyed then
				l_window ?= a_widget
				if l_window /= Void then
					if a_show then open_window (Void, l_window) else close_window (l_window) end
				elseif a_widget.is_show_requested /= a_show then
					if a_show then a_widget.show else a_widget.hide end
				end
			end
		end

	destroy_widget (a_widget: EV_WIDGET) is
			-- Destroy 'a_widget'
		require
			a_widget_not_void : a_widget /= Void
		do
			if not a_widget.is_destroyed then
				a_widget.destroy
			end
		ensure
			a_widget_destroyed : a_widget.is_destroyed
		end

	execute (a_command: STRING) is
			-- Same as Global.commands.execute, but with busy (sand-box) cursor while working.
		do
			start_busy (Void)
			glob.Global.commands.execute (a_command)
			glob.error.flush_and_acknowledge
			end_busy
		end

	refresh, process_events is
			-- Call this from time to time during extended computation to allow your GUI to take into account pending messages.
		do
			if is_running then ev_application.process_events end
		end

	start_wait_msg (a_explanation: STRING) is
			-- Show busy (sand-box) cursor.
			-- 'a_explanation' tells the user what the application is busy doing (if not Void).
		do
			set_status_message (get_a_widget, a_explanation)
			glob.Global.start_busy (a_explanation)
		end

	end_wait_msg (a_status: STRING) is
			-- Show application cursor.
		do
			set_status_message (get_a_widget, a_status)
			glob.Global.end_busy
		end

	set_status_message (a_widget: EV_WIDGET; a_message: STRING) is
			-- Set status 'a_message' on 'a_widget' (find and set text on status bar relative to 'a_widget')..
		local
			l_window: EV_WINDOW
			l_box1, l_box2: EV_BOX
			l_frame: EV_FRAME
			l_label: EV_LABEL
		do
			if a_message /= Void then
				l_window := parent_window (a_widget)
			end
			if l_window /= Void then
				l_box1 := l_window.lower_bar
			end
			if l_box1 /= Void and then not l_box1.is_empty then
				l_label ?= l_box1.first
				if l_label = Void then
					l_box2 ?= l_box1.first
					if l_box2 /= Void and then not l_box2.is_empty then
						l_label ?= l_box2.first
					end
					if l_box2 = Void then
						l_frame ?= l_box1.first
						if l_frame /= Void then
							l_label ?= l_frame.item
						end
					end
				end
				if l_label /= Void then
					output (l_label, a_message, Void)
					refresh
				end
			end
		end

	start_wait is
			-- Show wait (sand-timer) cursor.
		do
			start_wait_with_cursor (rep.pixmaps.wait_cursor)
		end

	start_busy_wait is
			-- Show busy (arrow and sand-timer) cursor.
		do
			start_wait_with_cursor (rep.pixmaps.busy_cursor)
		end

	start_wait_with_cursor (a_wait_cursor: EV_CURSOR) is
			-- Show that that the process is busy with 'a_wait_cursor'
		do
			busy_cursor_stack.extend (a_wait_cursor)
			set_cursor_on_all_windows (a_wait_cursor)
		end

	end_wait is
			-- Show application cursor.
		local
			l_cursor: EV_CURSOR
		do
			if not busy_cursor_stack.is_empty then
				busy_cursor_stack.remove
			end
			if not busy_cursor_stack.is_empty then
				l_cursor := busy_cursor_stack.item
			end
			if l_cursor = Void or else l_cursor.is_destroyed then
				l_cursor := rep.pixmaps.Standard_cursor
			end
			set_cursor_on_all_windows (l_cursor)
		end

	flash_window (a_widget: EV_WIDGET) is
			-- Flash window containing 'a_widget' to draw user attention to that window.
		do
			platform_specifics.flash_window (parent_window (a_widget))
		end

	set_default_processing (a_widget: EV_WIDGET; a_on: BOOLEAN) is
			-- Set default processing on or off for 'a_widget'.
		do
			platform_specifics.set_default_processing (a_widget, a_on)
		end

	show_border (a_text_field: EV_TEXT_FIELD; a_show: BOOLEAN) is
			-- Show/hide 3D border from 'a_text_field'
		do
			platform_specifics.show_border (a_text_field, a_show)
		end

	add_browsable_object (a_name: STRING; a_object: ANY) is
			-- Add 'a_object' under 'a_name' in the object browser's drop-down list of available objects.
		local
			l_browser: RCMD_OBJECT_BROWSER
		do
			l_browser ?= glob.Global.commands.service ("object_browser")
			if l_browser /= Void then
				l_browser.auto_initialize_window
				l_browser.browser.add_listed_object (a_name, a_object)
			end
		end

	set_info_widget (a_widget: EV_WIDGET; a_color: EV_COLOR; a_message: STRING; a_args: ARRAY [ANY]) is
			-- Set 'a_widget'.data with 'a_message'
		do
			if a_color /= Void then
				a_widget.set_background_color (a_color)
			end
			a_widget.set_data (formatter.argumented (a_message, a_args))
		end

	open_window (a_name: STRING; a_window: EV_WINDOW) is
			-- Open 'a_window', if not already opened, and remember it's setting under 'a_name' in config.
		do
			if a_window /= Void and then not a_window.is_destroyed then
				if not a_window.is_show_requested then
					if a_name = Void or else a_name.is_empty or else a_name.item (1) /= '_' then
						opened_windows := opened_windows + 1
					end
					a_window.close_request_actions.extend (agent close_window (a_window))
					if a_name /= Void
					and then not a_name.is_empty
					and then (a_name.count > 1 or else a_name.item (1) /= '_') then
						repositionner.register (a_name, a_window)
						repositionner.restore (a_window)
					end
				end
				a_window.show
				a_window.set_focus
			end
		end

	close_window (a_window: EV_WINDOW) is
			-- Close 'a_window', and eventually stop event loop if no more windows open.
		local
			l_name: STRING
		do
			if a_window /= Void and then not a_window.is_destroyed then
				l_name := repositionner.remember (a_window)
				if a_window.is_show_requested then
					if l_name = Void or else l_name.is_empty or else l_name.item (1) /= '_' then
						opened_windows := opened_windows - 1
					end
					if opened_windows > 0 then
							--| Do not close last window
						a_window.hide
					end
				end
			end
			if opened_windows <= 0 then
				exit_application
			end
		end

	execute_protected (a_requirer: STRING; a_agent: PROCEDURE [ANY, TUPLE []]) is
			-- Execute 'a_agent' with catching any exception.
		local
			rescued: STRING
		do
			if rescued /= Void then
				glob.error.signal (a_requirer, "Eiffel call crashed: $1", <<rescued>>)
			else
				a_agent.call ([])
			end
			glob.error.flush_and_acknowledge
		rescue
			if rescued = Void then
				rescued := glob.last_exception_explanation
				retry
			end
		end

	exit_application is
			-- Exit Eiffel Vision application (with eventual confirmation dialog).
		do
			quit.execute (c_empty_string)
		end

	destroy is
			-- Destroy Eiffel Vision application immediately.
		local
			l_list: LINEAR [EV_WINDOW]
			l_window: EV_WINDOW
			l_name: STRING
		do
			if not ev_application.is_destroyed then
				l_list := ev_application.windows
				if l_list /= Void then
					from
						l_list.start
					until
						l_list.off
					loop
						l_window := l_list.item
						if l_window /= Void
						and then not l_window.is_destroyed
						and then l_window.is_show_requested then
							l_name := repositionner.remember (l_list.item)
						end
						l_list.forth
					end
				end
				ev_application.destroy
			end
		end

	set_drop_actions (a_container: EV_CONTAINER; a_action: PROCEDURE [ANY, TUPLE [ANY]]) is
			-- In all items of 'a_container', set 'a_action' to be performed when an object is dropped.
		local
			l_linear: LINEAR [EV_WIDGET]
			l_container: EV_CONTAINER
			l_pick: EV_PICK_AND_DROPABLE
		do
			if a_container /= Void and then a_action /= Void then
				l_pick ?= a_container
				if l_pick /= Void then
					l_pick.drop_actions.extend (a_action)
				end
				from
					l_linear := a_container.linear_representation
					l_linear.start
				until
					l_linear.off
				loop
					l_container ?= l_linear.item
					if l_container /= Void then
						set_drop_actions (l_container, a_action)
					else
						l_pick ?= l_linear.item
						if l_pick /= Void then
							l_pick.drop_actions.extend (a_action)
						end
					end
					l_linear.forth
				end
			end
		end

	container_do_all_recursive (a_container: EV_CONTAINER; a_action: PROCEDURE [ANY, TUPLE [EV_WIDGET]]) is
			-- For all items of `a_container' recursively', call `a_action' with the current item as argument.
		require
			container_ok: a_container /= Void
			action_ok: a_action /= Void
		local
			l_linear: LINEAR [EV_WIDGET]
			l_container: EV_CONTAINER
		do
			from
				l_linear := a_container.linear_representation
				l_linear.start
			until
				l_linear.off
			loop
				l_container ?= l_linear.item
				if l_container /= Void then
					container_do_all_recursive (l_container, a_action)
				else
					a_action.call ([l_linear.item])
				end
				l_linear.forth
			end
		end

	show_splash_screen (a_on: BOOLEAN) is
			-- Show or hide `splash_screen' based on state of `a_on'.
		do
			if a_on then
				create splash_screen
				splash_screen.show
			elseif splash_screen /= Void then
				splash_screen.hide
				splash_screen := Void
			end
		end

feature {ROSE_PILOT_GUI_BASE} -- Implementation

	splash_screen: ROSE_SPLASH_SCREEN
			-- Splash screen

feature {NONE} -- Implementation

	c_empty_string: STRING is ""
			-- Empty string constant

	busy_cursor_stack: LINKED_STACK [EV_CURSOR] is
			-- Busy cursor stack
		once
			create Result.make
		end

	simple_cursors: BOOLEAN is
			-- Should simplified cursors be used?
		once
			Result := glob.config.boolean ("gui/simple_cursors", False)
		end

	trigger_busy_event (a_stopwatch: ROSE_STOPWATCH) is
			-- Calls to Gui.start_busy do NOT trigger 'busy_event'.
		do
			handle_busy_message (rep.pixmaps.busy_cursor, a_stopwatch)
		end

	handle_busy_message (a_cursor: EV_CURSOR; a_stopwatch: ROSE_STOPWATCH) is
			-- Called whenever 'busy_event' is fired.
		require
			stopwatch_ok: a_stopwatch /= Void
		local
			l_cursor: EV_CURSOR
			l_set: BOOLEAN
			retried: BOOLEAN
		do
			if retried then
				if glob.global.app.is_running_in_estudio then
					print ("Could not change cursor!")
				end
			else
				if a_stopwatch.is_running then
					l_cursor := new_busy_cursor (a_cursor, busy_message)
					if l_cursor /= Void then
						busy_cursor_stack.extend (l_cursor)
					end
					l_set := busy_cursor_stack.count = 1 or else (a_stopwatch.message /= Void and then not a_stopwatch.message.is_empty)
				else
					l_set := a_stopwatch.message /= Void and then not a_stopwatch.message.is_empty
					if not busy_cursor_stack.is_empty then busy_cursor_stack.remove end
					if not busy_cursor_stack.is_empty then l_cursor := busy_cursor_stack.item end
				end
				if l_cursor = Void then
					-- busy_cursor_stack is empty
					set_cursor_on_all_windows (rep.pixmaps.Standard_cursor)
				elseif l_set then
					set_cursor_on_all_windows (l_cursor)
				end
			end
		rescue
			if not retried then
				retried := True
				retry
			end
		end

	busy_message: STRING is
			-- Message from 'busy_stack'
		local
			l_first, l_string: STRING
			i, j: INTEGER
		do
			if busy_stack.count > 0 then
				from
					i := 1
				until
					l_string /= Void or else i > busy_stack.count
				loop
					l_string := busy_stack.i_th (i).message
					i := i + 1
				end
				l_first := l_string
				if l_first /= Void then
					from
						l_string := Void
						j := busy_stack.count
					until
						l_string /= Void or else j < i
					loop
						l_string := busy_stack.i_th (j).message
						j := j - 1
					end
					if l_string /= Void then
						Result := l_first + "%N" + l_string
					else
						Result := l_first
					end
				end
			end
		end

	truncated_string (a_string: STRING; a_max: INTEGER): STRING is
			-- 'a_string' truncated to 'a_max' characters
		require
			max_ok: a_max > 0
		do
			if a_string /= Void and then not a_string.is_empty then
				if a_string.count <= a_max then
					Result := a_string
				else
					Result := a_string.substring (1, a_max)
				end
			end
		end

	set_cursor_on_all_windows (a_cursor: EV_CURSOR) is
			-- Set 'a_cursor' on all windows
		local
			l_windows: LINEAR [EV_WINDOW]
		do
			if not ev_application.is_destroyed then
				l_windows := ev_application.windows
			end
			if l_windows /= Void then
				from
					l_windows.start
				until
					l_windows.off
				loop
					if l_windows.item /= Void and then not l_windows.item.is_destroyed then
						l_windows.item.set_pointer_style (a_cursor)
					end
					l_windows.forth
				end
			end
		end

	platform_specifics: ROSE_EV_PLATFORM_SPECIFICS is
			-- Platform specific implementation
		once
			create Result
		end

	get_a_widget: EV_WIDGET is
			-- Return a valid widget still alive, Void if none
		local
			l_windows: LINEAR [EV_WINDOW]
			rescued: BOOLEAN
		do
			if rescued then
				Result := Void
			elseif not ev_application.is_destroyed then
				--| FIXME: The following call crashes in Vision2 sometimes
				Result := ev_application.focused_widget
			end
			if Result = Void or else Result.is_destroyed then
				Result := Void
				if not ev_application.is_destroyed then
					l_windows := ev_application.windows
				end
			end
			if l_windows /= Void then
				from
					l_windows.start
				until
					l_windows.off or else Result /= Void
				loop
					if l_windows.item /= Void and then not l_windows.item.is_destroyed then
						Result := l_windows.item
					end
					l_windows.forth
				end
			end
		rescue
			if not rescued then
				rescued := True
				retry
			end
		end

end
