indexing
	description: "State for ROSE_GRAPHICS_WIDGET"
	notice: "[
		This state indicates what actions need to be performed at each timeout event
		Optionally, states can have a parent and/or children
		This is useful when more than one state is needed
		Modifying the state of a child notifies the parent,
		and modifying the parent propagates to the children
	]"
	author: "Zoran Simic"
	date: "$Date: 2009/04/06 $"
	revision: "$Revision: #3 $"

class
	ROSE_GRAPHICS_STATE

create
	default_create, make

feature {NONE} -- Initialization

	make (a_parent: like Current) is
			-- Make with 'a_parent'
		require
			valid_parent: a_parent /= Void
		do
			parent := a_parent
			a_parent.add (Current)
		end

feature -- Access

	needs_redraw: BOOLEAN
			-- Figures need to be redrawn on display's pixmap
			-- This should be True if any of the other states are True
			-- It is not a function however for perfomance reasons
			-- (this boolean is queried all the time on an EV_TIMEOUT event)

	needs_refresh: BOOLEAN
			-- Figures need to be updated from modified settings (colors, line sizes, ...)

	needs_projection: BOOLEAN
			-- Real-world coordinates need to be projected on display's pixmap

feature -- Basic operations

	reset is
			-- Reset state.
		do
			needs_redraw := False
			needs_refresh := False
			needs_projection := False
		end

	set_needs_redraw is
		do
			needs_redraw := True
			on_state_updated
		end

	set_needs_refresh is
		do
			needs_refresh := True
			set_needs_redraw
		end

	set_needs_projection is
		do
			needs_projection := True
			set_needs_redraw
		end

	set_needs_all is
		do
			needs_refresh := True
			set_needs_projection
		end

	remove is
			-- Remove this state from it's parent (if any).
		do
			if parent /= Void then
				parent.children.prune_all (Current)
			end
		end

feature {ROSE_GRAPHICS_STATE} -- Implementation

	parent: like Current
			-- Parent state (optional), all changes are propagated ('or'-ed) to the parent

	children: LINKED_LIST [like Current]
			-- Children of this state (optional)

	has (a_child: like Current): BOOLEAN is
		do
			Result := children /= Void and then children.has (a_child)
		end

	add (a_child: like Current) is
		require
			valid_child: a_child /= Void and then not has (a_child)
		do
			if children = Void then
				create children.make
			end
			children.extend (a_child)
		end

	on_state_updated is
			-- Called whenever state has been changed, redefine if necessary.
		do
			if parent /= Void then
				parent.update_from (Current)
			elseif children /= Void then
				from
					children.start
				until
					children.off
				loop
					children.item.update_from (Current)
					children.forth
				end
			end
		end

	update_from (a_other: like Current) is
			-- Update this state from 'a_other'.
		require
			valid_other: a_other /= Void
		do
			needs_redraw := needs_redraw or else a_other.needs_redraw
			needs_refresh := needs_refresh or else a_other.needs_refresh
			needs_projection := needs_projection or else a_other.needs_projection
		end

end -- class ROSE_GRAPHICS_STATE
