note

	description:

		"Common parts for graphical widgets"

	notice: "[
		This class provides the following features for it's descendants:
		- Provides objects such as events, state, display, world and current view coordinates
		- Zooming, centering, scrolling and selection of part of the world to view
		- Handles basic mouse events:
		   - Right click and hold to select an area of current view
		   - Middle mouse button scrolls the view or moves selected area around
		   - Contextual menu
	]"
	author: "Zoran Simic"
	last_editor: "$Author: c4 $"
	date: "$Date: 2011/06/08 $"
	revision: "$Revision: #24 $"
	path: "$File: //rose/source/kernel/infrastructure/widgets/rose_graphics_widget.e $"

deferred class	ROSE_GRAPHICS_WIDGET

inherit
	ROSE_WIDGET
		redefine
			title, set_title
		end

	ROSE_CHART_EXPORT_ROUTINES
		export
			{NONE} all
		undefine
			default_create, is_equal, copy
		end

feature {NONE} -- Initialization

	setup_widget is
			-- Create widget and put it in 'a_container' if provided.
		do
			create events.make
			create state
			create display.make (Current)
			state := display.state
			pixmap := display.pixmap
			create world
			create view.make_normal
				-- Pixmap (coming from the display)
			resize_actions.extend (agent on_resize)
			pixmap.pointer_motion_actions.extend (agent on_mouse_move)
			pixmap.pointer_button_press_actions.extend (agent on_mouse_click)
			pixmap.pointer_button_release_actions.extend (agent on_mouse_release)
			pixmap.pointer_leave_actions.extend (agent on_mouse_out)
			pixmap.pointer_double_press_actions.extend (agent on_mouse_double_click)
			pixmap.key_release_actions.extend (agent on_key)
			pixmap.mouse_wheel_actions.extend (agent on_mouse_wheel)
			put (pixmap)
				-- Default settings
			create timeout.make_with_interval (100)
			timeout.actions.extend (agent on_timeout_event)
			selection := display.selection
			initialize_widget
			clear_all
		end

	initialize_widget is
			-- Specific initialization of the widget.
		deferred
		end

	palette_chooser: ROSE_PALETTE_CHOOSER
		-- Dialog to choose the palette

	show_palette is
			-- Show `palette_chooser'
		do
			if palette_chooser = Void then
				create palette_chooser
			end
			palette_chooser.show (display.settings.palette)
		end

feature -- Access

	is_picking_imp: BOOLEAN
			-- is the system picking?

	title: STRING is
			-- Widget title, shown if provided
		do
			Result := display.title_text.text
		end

	events: ROSE_GRAPHICS_EVENTS
			-- Events generated by this widget

	display: ROSE_GRAPHICS_DISPLAY
			-- Associated display, holding the Vision2 figures and settings

	state: ROSE_GRAPHICS_STATE
			-- State of the current graphics representation

	pixmap: EV_PIXMAP
			-- Pixmap on which drawing happens

	world: ROSE_RECTANGLE

	view: ROSE_RECTANGLE
			-- Coordinates of relative currently viewed part of each world
			-- When whole graphics is viewed, this is a normal rectangle
			-- Part of the normal rectangle when zoomed

feature -- Element change

	set_is_picking_imp (an_is_picking_imp: like is_picking_imp) is
			-- Set `is_picking_imp' to `an_is_picking_imp'.
		do
			is_picking_imp := an_is_picking_imp
		ensure
			is_picking_imp_assigned: is_picking_imp = an_is_picking_imp
		end

feature -- Basic operations

	set_title (a_title: STRING) is
			-- Set 'title'
		do
			Precursor (a_title)
			display.set_title (a_title)
			state.set_needs_redraw
		end

	clear_all is
			-- Clear out the graphics.
		do
			display.clear_all
			set_title (Void)
			view.make_normal
		end

	selection: ROSE_SELECTION_AREA
			-- Selection area

feature {NONE} -- Agents

	on_mouse_move (a_x, a_y: INTEGER; a_x_tilt, a_y_tilt, a_pressure: DOUBLE; a_screen_x, a_screen_y: INTEGER) is
			-- Mouse moved.
		do
			mouse_x := a_x
			mouse_y := a_y

			if mouse_button_pressed = no_button_pressed then
				if selection.is_within_resize_icon (a_x, a_y) then
					pixmap.set_pointer_style (selection.resize_cursor)
				elseif selection.is_within_selected (a_x,a_y) then
					pixmap.set_pointer_style (display.hand_cursor)
				else
					pixmap.set_pointer_style (Gui.rep.pixmaps.Standard_cursor)
				end
			elseif mouse_button_pressed = 1 then
				if selection.is_selecting then
					-- Left mouse pressed within the drawn area
					selection.extend_selection (a_x,a_y)
				end
			elseif mouse_button_pressed = 2 then
				if is_dragging then

					adapt_view_from_dragging

					drag_x_start := a_x
					drag_y_start := a_y

					state.set_needs_projection
				else -- if selection.is_within_selectable (a_x,a_y) then
					-- Middle mouse pressed within the drawn area
					selection.drag_selected (a_x,a_y)
				end
			end
		end

	on_mouse_click (a_x, a_y, a_button: INTEGER; a_x_tilt, a_y_tilt, a_pressure: DOUBLE; a_screen_x, a_screen_y: INTEGER) is
			-- Mouse-key pressed.
		do
			mouse_button_pressed := a_button
			if not is_picking then
				if a_button = 1 then
					if display.legend.has_point (a_x,a_y) then
						selection.remove_selected
					elseif selection.is_within_resize_icon (a_x, a_y) then
						selection.start_resizing (a_x, a_y)
					elseif selection.is_within_selected (a_x,a_y) then
						adapt_view_from_selection
						state.set_needs_projection
						selection.reset
					elseif selection.is_within_selectable (a_x,a_y) then
						-- Start selecting
						selection.start_selection (a_x,a_y)
					end
				elseif a_button = 2 then
					if selection.is_within_selectable (a_x,a_y) then
						if selection.is_within_selected (a_x,a_y) then
							-- Start dragging selection area
							selection.grab_selected (a_x,a_y)
						else
							is_dragging := True
							drag_x_start := a_x
							drag_y_start := a_y
						end
					end
				elseif a_button = 3 then
					if not display.legend.has_point (a_x,a_y) and Gui.ev_application.ctrl_pressed then
						show_contextual_menu (a_x,a_y)
					end
				end
			end
		end

	on_mouse_release (a_x, a_y, a_button: INTEGER; a_x_tilt, a_y_tilt, a_pressure: DOUBLE; a_screen_x, a_screen_y: INTEGER) is
			-- Mouse-key released.
		do
			if a_button = 1 then
				selection.finalize_selection
			elseif a_button = 2 then
				if selection.is_dragging then
					selection.stop_dragging
				else
					is_dragging := False
				end
			end
			mouse_button_pressed := no_button_pressed
			set_is_picking_imp (gui.is_picking)
		end

	on_mouse_double_click (a_x, a_y, a_button: INTEGER; a_x_tilt, a_y_tilt, a_pressure: DOUBLE; a_screen_x, a_screen_y: INTEGER) is
			-- Mouse double-clicked.
		do
			mouse_button_pressed := no_button_pressed
			if a_button = 1 then
				if Gui.ev_application.ctrl_pressed then
				end
			end
		end

	on_mouse_wheel (a_direction: INTEGER) is
			-- Zoom in and out
			-- Zooms the view to 80% or 125% and then makes sure the mouse
			-- pointer is still at the same position in the new view
		local
			l_zoom: DOUBLE
			l_old_view: like view
			l_zoom_point: ARRAY[DOUBLE] -- Point where we zoom to
			l_trans_x, l_trans_y: DOUBLE
			l_width, l_height: DOUBLE
		do
			l_zoom := 0.8

			create l_old_view.set_from_rectangle (view)
			l_zoom_point := display_to_view_coordinates (mouse_x, mouse_y)

			if a_direction > 0 then
				l_zoom := 1.0 / l_zoom
			end

			view.zoom (l_zoom)

			l_trans_x := l_zoom_point.item (1) * (1.0 / l_zoom)
			l_trans_y := l_zoom_point.item (2) * (1.0 / l_zoom)

			l_width := view.width
			l_height := view.height
			view.set1 (l_old_view.x1 +  l_zoom_point.item(1) - l_trans_x, l_old_view.y1 + l_zoom_point.item(2) - l_trans_y)
			view.set_size (l_width, l_height)

			state.set_needs_projection
		end


	on_mouse_out is
			-- Called when mouse gets out of the widget.
		do
			mouse_button_pressed := no_button_pressed
		end

	show_contextual_menu (a_x, a_y: INTEGER) is
			-- Show contextual menu at coordinates 'a_x', 'a_y'.
		local
			l_menu: EV_MENU
		do
			if not is_picking then
				l_menu := contextual_menu (a_x,a_y)
				if l_menu /= Void and then l_menu.count > 0 then
					l_menu.show
				end
			end
		end

	contextual_menu (a_x, a_y: INTEGER): EV_MENU is
			-- Pop-up menu activated by right mouse button
		deferred
		end

	adapt_view_from_selection is
			-- Set `view' from selected rectangle.
		do
			view.set_from_rectangle (selection.normal_selected)
		end

	adapt_view_from_dragging is
			-- Adapt the view after it was dragged
		local
			l_start, l_end: ARRAY[DOUBLE]
		do
			l_start := display_to_view_coordinates (drag_x_start, drag_y_start)
			l_end := display_to_view_coordinates (mouse_x, mouse_y)

			view.translate (l_start.item(1) - l_end.item(1), l_start.item(2) - l_end.item(2))
		end



	on_resize (a_x, a_y, a_width, a_height: INTEGER) is
			-- Called when widget is being resized.
		local
			retried: BOOLEAN
		do
			if retried then
				-- Do nothing, Vision2 has issues with big pixmaps
			elseif a_width > 0 and then a_height > 0 then
				pixmap.set_size (a_width,a_height)
				selection.reset
				state.set_needs_projection
			end
		rescue
			if not retried then
				retried := True
				retry
			end
		end

	on_key (a_key: EV_KEY) is
			-- Handle release of `a_key'.
		do
			if a_key /= Void and then a_key.code /= 0 then
				if a_key.code = Gui.rep.keys.key_f5 then
					state.set_needs_redraw
				end
			end
		end

	on_export_picture is
			-- Save `Current' as a Portable Network Graphics file.
		do
			export_picture (Current, pixmap)
		end

	on_export_postscript is
			-- Save `Current' as a Postscript file.
		do
				export_postscript (Current, display.figure_world)
		end

	on_send_to is
			-- Process request to send as an email attachment (image/svg+xml)
		do
			send_to (Current, display.figure_world)
		end

	on_clipboard is
			-- Process request to copy to clipboard as SVG.
		do
			copy_to_clipboard (display.figure_world)
		end

	on_export_svg is
			-- Process request to save as SVG.
		do
			export_as_svg (Current, display.figure_world)
		end

	on_print is
			-- Print `Current' on printer selected from dialog.
		do
			print_to (Current, display.figure_world)
		end

feature {NONE} -- Implementation

	timeout: EV_TIMEOUT
	mouse_button_pressed: INTEGER
	no_button_pressed: INTEGER is 0

	is_picking: BOOLEAN
			-- Is the mode pick and drop?
		do
			Result := gui.is_picking or is_picking_imp
		end


	add_menu (a_menu: EV_MENU; a_sub_name: STRING; a_description: ARRAY [TUPLE]) is
			-- Add `a_sub_menu' to `a_menu' according to `a_description'
		do
			Gui.factory.add_sub_menu (a_menu,a_sub_name,a_description)
		end

	on_timeout_event is
			-- Function called every 100 milliseconds.
		do
			if is_displayed and then state.needs_redraw then
				start_time_measure ("timeout",Void)
				on_timeout
				display.redraw
				state.reset
				end_time_measure
			end
		end

	on_timeout is
		deferred
		end

feature {NONE} -- Coordinates

	display_to_view_coordinates (a_x, a_y: INTEGER): ARRAY[DOUBLE] is
			-- Convert from screen to view coordinates
			-- View coordinates are set up like this:
			--	 ______(1/1)
			--	 |       |
			--	 |       |
			-- (0/0)_____|
		local
			l_relative_x, l_relative_y: DOUBLE
			l_drawn_area: ROSE_RECTANGLE
		do

			l_drawn_area := display.drawn_area

			l_relative_x := (view.width / l_drawn_area.width) * (a_x.to_double - l_drawn_area.x1)
			l_relative_y := view.height - ((view.height / l_drawn_area.height) * (a_y.to_double - l_drawn_area.y1))

			Result := <<l_relative_x, l_relative_y>>
		end


feature {NONE} -- Dragging

	is_dragging: BOOLEAN
			-- Is dragging in progress (whatever that means)?

	drag_x_start, drag_y_start: INTEGER
	mouse_x, mouse_y: INTEGER


end -- class ROSE_GRAPHICS_WIDGET
