indexing
	description: "Input history tracking"
	author: "Zoran Simic"
	date: "$Date: 2008/12/01 $"
	revision: "$Revision: #4 $"

class
	ROSE_INPUT_HISTORY_GENERIC [G]

create
	make_shell, make_browser

feature {NONE} -- Initialization

	make (a_capacity: INTEGER) is
			-- Create an input history
			-- 'a_capacity' indicates the number of items to keep in history
		require
			valid_capacity: a_capacity > 0
		do
			capacity := a_capacity
			create history.make (1,capacity)
		end

	make_shell (a_capacity: INTEGER) is
			-- Create with a shell-like history behavior.
		do
			make (a_capacity)
			is_shell_like := True
		end

	make_browser (a_capacity: INTEGER) is
			-- Create with a browser-like history behavior.
		do
			make (a_capacity)
			is_shell_like := False
		end

feature -- Access

	history: ARRAY [G]
			-- All items in history

	index: INTEGER
			-- Index of 'current_item' in the history

	count: INTEGER
			-- Number of items in this history

	capacity: INTEGER
			-- Capacity of this history

	current_item: G is do Result := item (index) end

	is_shell_like: BOOLEAN
			-- Should this history be managed in an shell-like history style?
			-- I.e: going 'prev' and entering a new item does not clear all consecutive items
			-- if False, behaves in an explorer-like fashion

	is_before: BOOLEAN is do Result := index < 1 end
	is_on_first: BOOLEAN is do Result := index = 1 end
	is_after: BOOLEAN is do Result := index > count end
	is_on_last: BOOLEAN is do Result := index > 0 and then index = count end
	is_empty: BOOLEAN is do Result := count = 0 end

	item (a_index: INTEGER): G is
			-- Item at 'a_index', default value for G if 'a_index' is not between 1 and 'count'
		do
			if history.valid_index (a_index) then
				Result := history.item (a_index)
			end
		end

	items: ARRAY [G] is
			-- All items in this history
			-- Void if none (if count is 0)
		do
			if count > 0 then
				Result := history.subarray (1,count)
			end
		end

	preceding (a_include_current: BOOLEAN): ARRAY [G] is
			-- All items preceding 'current_item' in this history
			-- 'a_include_current' indicates whether to include 'current_item' in the list or not.
		local
			i: INTEGER
		do
			i := index
			if a_include_current then
				i := i - 1
			end
			if i >= 1 then
				Result := history.subarray (1,i)
			end
		end

	following (a_include_current: BOOLEAN): ARRAY [G] is
			-- All items following 'current_item' in this history
			-- 'a_include_current' indicates whether to include 'current_item' in the list or not.
		local
			i: INTEGER
		do
			i := index
			if a_include_current then
				i := i + 1
			end
			if index > 0 and then index <= count then
				Result := history.subarray (index,count)
			end
		end

	has (a_item: G): BOOLEAN is
			-- Does this history contain an item equal to 'a_item'?
		local
			i: INTEGER
		do
			if a_item /= Void then
				from
					i := 1
				until
					Result or else i > count
				loop
					Result := equal (a_item, history.item (i))
					i := i + 1
				end
			end
		end

feature -- Basic operations

	add (a_item: G) is
			-- Add 'a_item' to this history.
			-- If 'is_shell_like' add 'a_item' in a stack like fashion (and only one item per history is present)
			-- otherwise, remove items following 'item' if any and add 'a_item' there (like a web browser)
		do
			if is_shell_like then
				remove_all (a_item)
				index := count
				effective_add (a_item)
			elseif not equal (current_item,a_item) then
				index := index + 1
				effective_add (a_item)
			end
		end

	go_previous is
			-- Go to previous item in history.
			-- If 'is_shell_like', 'item' is Void on extremities
			-- otherwise, 'item' is always the extreme item
		do
			index := (index - 1).max (0)
		end

	go_next is
			-- Go to next item in history.
			-- If 'is_shell_like', 'item' is Void on extremities
			-- otherwise, 'item' is always the extreme item
		do
			index := (index + 1).min (count + 1)
		end

	go_to (a_item: G) is
			-- Go to last occurence of 'a_item'
			-- Stay at current position if no occurences of 'a_item' found.
		local
			i: INTEGER
		do
			from
				i := 1
			until
				i > count
			loop
				if equal (a_item,history.item (i)) then
					index := i
				end
				i := i + 1
			end
		end

	remove_last (a_number: INTEGER) is
			-- Remove last 'a_count' items from this history
		do
			count := (count - a_number).max (0)
			index := index.min (count)
		end

	clear_all is
			-- Remove all items from this history.
		local
			l_default: G
			i: INTEGER
		do
			index := 0
			count := 0
			from
				i := 1
			until
				i > capacity
			loop
				history.put (l_default,i)
				i := i + 1
			end
		ensure
			count = 0
		end

feature {NONE} -- Implementation

	effective_add (a_item: G) is
		do
			from
				count := index
			until
				count < capacity
			loop
				remove_item (1)
			end
			count := count + 1
			index := count
			history.put (a_item,index)
		end

	remove_all (a_item: G) is
			-- Remove all items equal to 'a_item'.
		local
			i: INTEGER
		do
			from
				i := 1
			until
				i > count
			loop
				if equal (history.item (i),a_item) then
					remove_item (i)
				else
					i := i + 1
				end
			end
		end

	remove_item (a_index: INTEGER) is
			-- Remove item at 'a_index'.
		local
			i: INTEGER
		do
			if index >= a_index then
				index := index - 1
			end
			from
				i := a_index
			until
				i >= count
			loop
				history.put (history.item (i+1),i)
				i := i + 1
			end
			count := (count - 1).max (0)
			index := index.max (0).min (count)
		end

end
