note

	description:

		"[
        Objects that represent an interval between two points in time, with an associated item and weight.
        This class allows intervals that run backward through time. It also allows a Void date to represent
         both the beginniong of time, and the end of time. We suggest you only construct intervals
         that don't have these strange properties (e.g. `is_valid' ==> True). The use of Void dates at AXA Rosenberg
         is a disaster compounded from contradictary meanings (e.g. today, start-of-time, end-of-time) and should be avoided
         like the plague, as sooner or later one will get passed as an argument that has a contradictary meaning.
       ]"

	author: "David Stevens"
	last_editor: "$Author: c4 $"
	date: "$Date: 2012/10/16 $"
	revision: "$Revision: #15 $"
	path: "$File: //rose/source/kernel/infrastructure/containers/rose_interval.e $"

deferred class	ROSE_INTERVAL [G]

inherit	

	ROSE_CONSTANTS
	
	DEBUG_OUTPUT
	
feature -- Initialization

	make (a_item: G; a_start: ROSE_DATE; a_end: ROSE_DATE; a_weight: DOUBLE) is
			-- Create interval associated with `a_item' starting on `a_start' (or at the begining of time if Void),
			--  and ending at `a_end' (or at the end of time if Void). Associate `a_weight' with the interval.
		do
			item := a_item
			set_start_date (a_start)
			set_end_date (a_end)
			weight := a_weight
			hash_key := [item, start_date, end_date]
		end

feature -- Access

	hash_key: TUPLE
			-- Key for use in hash tables

	item: G
			-- Item associated with interval

	end_is_unbounded: BOOLEAN
			-- Is interval valid until end of time?

	start_is_unbounded: BOOLEAN
			-- Is interval valid from beginning of time?
	
	set_start_date (a_date: ROSE_DATE) is
			-- Set `start_date' to `a_date'.
		do
			start_date := a_date
			start_is_unbounded := start_date = Void or else start_date.same_date (constants.minimum_date_time)
		ensure
			start_date_aliased: start_date = a_date
		end

	set_end_date (a_date: ROSE_DATE) is
			-- Set `end_date' to `a_date'.
		do
			end_date := a_date
			end_is_unbounded := end_date = Void or else end_date.same_date (constants.maximum_date_time)
		ensure
			end_date_aliased: end_date = a_date
		end
	
	brackets (a_date: ROSE_DATE): BOOLEAN is
			-- Does `Current' start on or before `a_date' and does it end after `a_date' ?
		require
			a_date_not_void: a_date /= Void
		do
			Result := starts_by_date (a_date) and then ends_after_date (a_date)
		end

	brackets_inclusive (a_date: ROSE_DATE): BOOLEAN is
			-- Does `Current' start on or before `a_date' and does it end on or after `a_date' ?
		require
			a_date_not_void: a_date /= Void
		do
			Result:= starts_by_date (a_date) and then ends_by_date (a_date)
		end

	ends_by_date (a_date: ROSE_DATE): BOOLEAN is
			-- Does `Current' end on or after `a_date' ?
		require
			a_date_not_void: a_date /= Void
		do
			Result := end_is_unbounded or else a_date <= end_date	
		end
		
	disjoint (a_interval: like Current): BOOLEAN is
			-- True if `Current' does not contain any date contained by `a_interval'
		require
			a_interval_not_void: a_interval /= Void
			a_interval_valid: a_interval.start_date /= Void and a_interval.end_date /= Void and then
				a_interval.start_date <= a_interval.end_date
			
		do
			Result := not meets (a_interval.start_date, a_interval.end_date)
		end

	misses (a_start_date, a_end_date: ROSE_DATE): BOOLEAN is
			-- True if `Current' does not contain any date between `a_start_date' and `a_end_date' inclusive
		require
			a_start_date_not_void: a_start_date /= Void
			a_end_date_not_void: a_end_date /= Void
			consistent_dates: a_start_date <= a_end_date
		do
			Result := not meets (a_start_date, a_end_date)
		end

	meets (a_start_date, a_end_date: ROSE_DATE): BOOLEAN is
			-- True if `Current' starts no later than `a_end_date' and ends after `a_start_date'
		require
			a_start_date_not_void: a_start_date /= Void
			a_end_date_not_void: a_end_date /= Void
			consistent_dates: a_start_date <= a_end_date
		do
			Result := starts_by_date (a_end_date) and ends_after_date (a_start_date)
		end

	covers (a_start_date, a_end_date: ROSE_DATE): BOOLEAN is
			-- True if `Current' starts on or before, and ends after, both `a_start_date' and `a_end_date' 
		require
			a_start_date_not_void: a_start_date /= Void
			a_end_date_not_void: a_end_date /= Void
		do
			Result := brackets (a_start_date) and then brackets (a_end_date)
		end

	starts_by (a_interval: like Current): BOOLEAN is
			-- True if `Current' starts on or before `a_interval'
		require
			a_interval_not_void: a_interval /= Void
			a_interval_start_date_not_void: a_interval.start_date /= Void
		do	
			Result := starts_by_date (a_interval.start_date)
		end

	starts_before (a_interval: like Current): BOOLEAN is
				-- True if `Current' starts strictly before `a_interval'
		require
			start_date_not_void: start_date /= Void
			a_interval_not_void: a_interval /= Void
			a_interval_start_date_not_void: a_interval.start_date /= Void
		do	
			Result := start_date < a_interval.start_date
		end

	ends_after (a_interval: like Current): BOOLEAN is
			-- True if `Current' ends strictly after `a_interval'
		require
			a_interval_not_void: a_interval /= Void
			a_interval_end_date_not_void: a_interval.end_date /= Void
		do
			Result := ends_after_date (a_interval.end_date)
		end

	starts_by_date (a_date: ROSE_DATE): BOOLEAN is
			-- Does `Current' start on or before `a_date'?
		require
			a_date_not_void: a_date /= Void
		deferred
		end

	ends_after_date (a_date: ROSE_DATE): BOOLEAN is
			-- Does `Current' end on or after `a_date'?
		require
			a_date_not_void: a_date /= Void
		deferred
		end

	first_date: ROSE_DATE is
			-- Earliest date within `Current';
			-- an alias for `start_date'
		deferred
		end

	last_date: ROSE_DATE is
			-- Latest date within `Current', if `end_date' /= Void;
			-- Use as: if interval.end_date = Void then create Result {ROSE_DATE}.make_ccyymmdd (99991231) or else
			--  Result := interval.last_date
		require
			end_date_not_void: end_date /= Void
		deferred
		ensure
			last_date_not_void: Result /= Void
		end

	is_valid: BOOLEAN is
			-- Is `Current' sane, and does it have an attached item?
		deferred
		ensure
			attached_item: Result implies item /= Void
			consistent_dates: Result implies start_date /= Void and end_date /= Void and then start_date <= end_date
		end
	
	interval_before (a_first_date: ROSE_DATE; a_weight: DOUBLE): like Current is
			-- Interval starting from `a_first_date' and ending on or after `start_date' (depending on open/closed interval type);
			-- Result may not pass `is_valid', even if `item' /= Void.
		require
			start_date_not_void: start_date /= Void
		deferred
		end
	
	interval_after (a_last_date: ROSE_DATE; a_weight: DOUBLE): like Current is
			-- Interval starting on or after `end_date' (depending on open/closed interval type) and ending
			-- on or after `a_last_date' (depending on open/closed interval type);
			-- Result may not pass `is_valid', even if `item' /= Void.
		require
			a_last_date_not_void: a_last_date /= Void
			end_date_not_void: end_date /= Void
		deferred
		end

	before (a_other: like Current): BOOLEAN is
			-- Does `Current' start strictly before `a_other'?
		require
			start_date_not_void: start_date /= Void
			a_other_not_void: a_other /= Void
			a_other_start_date_not_void: a_other.start_date /= Void
		do
			Result := start_date < a_other.start_date
		end
		
feature-- Basic Operations

	infix "+" (a_interval: like Current): ROSE_LINEAR_ARRAY [like Current] is
			-- I think this is a list of intervals that cover all the dates in both intervals
		require
			first_date_not_void: first_date /= Void
			end_date_not_void: end_date /= Void
			a_interval_first_date_not_void: a_interval /= Void implies a_interval.first_date /= Void
			a_interval_end_date_not_void: a_interval /= Void implies a_interval.end_date /= Void
			consistent_dates: a_interval /= Void implies a_interval.first_date <= a_interval.end_date
		do
			create Result.make_empty
			if a_interval /= Void then
				if misses (a_interval.first_date, a_interval.last_date) then
					Result.extend (a_interval)
					Result.extend (Current)
				elseif covers (a_interval.first_date, a_interval.last_date) then
					Result := cover_sum (a_interval)
				elseif a_interval.covers (first_date, last_date) then
					Result := a_interval.cover_sum (Current)
				elseif brackets (a_interval.first_date) then
					Result := bracket_first_sum (a_interval)
				elseif brackets (a_interval.last_date) then
					Result := a_interval.bracket_first_sum (Current)
				end
			else
				Result.extend (Current)
			end
		ensure
			infix_plus_not_void: Result /= Void
			no_void: not Result.has (Void)
		end

	add (a_weight: DOUBLE) is
			-- Increase `weight' by `a_weight'
		do
			weight := weight + a_weight	
		end

	multiply (a_factor: DOUBLE) is
			-- Scale `weight' by `a_factor'.
		do
			weight := weight * a_factor
		end

feature {ROSE_INTERVAL}	-- Access

	cover_sum (a_interval: like Current): ROSE_LINEAR_ARRAY [like Current] is
		require
			first_date_not_void: first_date /= Void
			end_date_not_void: end_date /= Void
			a_interval_not_void: a_interval /= Void
			a_interval_first_date_not_void: a_interval.first_date /= Void
			a_interval_end_date_not_void: a_interval.end_date /= Void
		local
			l_interval: like Current
		do
			create Result.make_empty
			if first_date < a_interval.first_date then
				l_interval := a_interval.interval_before (first_date, weight)
				Result.extend(l_interval)
			end
			l_interval := a_interval.twin
			l_interval.add (weight)
			Result.extend (l_interval)
			if a_interval.last_date < last_date then
				l_interval:= a_interval.interval_after (last_date, weight)
				Result.extend(l_interval)
			end
		ensure
			cover_sum_not_void: Result /= Void
			cover_sum_not_empty: not Result.is_empty
			no_void: not Result.has (Void)
		end
	
	bracket_first_sum (a_interval: like Current): ROSE_LINEAR_ARRAY [like Current] is
		require
			first_date_not_void: first_date /= Void
			end_date_not_void: end_date /= Void
			a_interval_not_void: a_interval /= Void
			a_interval_first_date_not_void: a_interval.first_date /= Void
			a_interval_end_date_not_void: a_interval.end_date /= Void
		local
			l_interval: like Current
		do
			create Result.make_empty
			if first_date < a_interval.first_date then
				l_interval:= a_interval.interval_before (first_date, weight)
				Result.extend(l_interval)
			end
			if a_interval.first_date <= last_date then
				l_interval := twin
				l_interval.make (item, a_interval.start_date, end_date, weight + a_interval.weight)
				Result.extend(l_interval)
			end
			if last_date < a_interval.last_date then
				l_interval:= interval_after (a_interval.last_date, a_interval.weight)
				Result.extend(l_interval)
			end
		ensure
			bracket_first_sum_not_void: Result /= Void
			no_void: not Result.has (Void)
		end
		
feature {HASH_TABLE, ROSE_INTERVAL_LIST} -- Access

	hash_code: INTEGER is
			-- Allows adding HASHABLE as an ancestor of any descendant of `Current'
		do
			Result := hash_key.hash_code
		end

feature -- Access
	
	start_date: ROSE_DATE
			-- Date at which `Current' starts;
			-- Void means start-of-time, but don't ever use this semantic.
	
	end_date: ROSE_DATE
			-- Date on or after (depending on open/closed nature) which `Current' ends;
			-- Void means end-of-time, but don't ever use this semantic.
	
	weight: DOUBLE
			-- Weight associated with interval
	
feature -- Output
	
	debug_output: STRING is
			-- <Precursor>
		local
			l_start, l_end: ROSE_DATE
		do
			if start_date = Void then
				create l_start.make_min_date
			else
				l_start := start_date
			end
			if end_date = Void then
				create l_end.make_max_date
			else
				l_end := end_date
			end
			Result:= l_start.ccyymmdd.out + ":" + l_end.ccyymmdd.out + ":" + weight.out
		end

invariant

	consistent_start: start_is_unbounded = (start_date = Void or else start_date.same_date (constants.minimum_date_time))
	consistent_end: end_is_unbounded = (end_date = Void or else end_date.same_date (constants.maximum_date_time))
	hash_key_not_void: hash_key /= Void

end -- class ROSE_INTERVAL
