indexing
	description: "Framework I/O conventions"
	notice: "[
		This class is documentation on framework conventions.
		Please comply with conventions listed here when making an application.
]"
	author: "Zoran Simic"
	last_editor: "$Author: rj $"
	date: "$Date: 2014/09/05 $"
	revision: "$Revision: #25 $"
	path: "$File: //rose/source/kernel/infrastructure/framework/application/rose_io_convention.e $"

class
	ROSE_IO_CONVENTION

inherit
	ROSE_FORMATTING
	KL_SHARED_FILE_SYSTEM

feature -- Configuration constants

	configuration_database_node: STRING = "/db/"			-- Node in configuratio under which DB server paths are expected
	configuration_node_separator: CHARACTER = '/'			-- slash

	configuration_node_separators: attached STRING
			-- Separator for configuration nodes
		once
			Result := configuration_node_separator.out
		end

	configuration_path_site: STRING = "/path/site"
			-- '/path/site' configuration key

	configuration_path_deployment_archive: STRING = "/path/deployment_archive"
			-- '/path/deployment_archive' configuration key

	configuration_path_log: STRING = "/path/log"
			-- '/path/log' configuration key

	configuration_path_storable: STRING = "/path/storable"
			-- '/path/storable' configuration key

	configuration_path_override_storable: STRING = "/path/override_storable"
			-- '/path/override_storable' configuration key

	configuration_path_archive: STRING = "/path/archive"
			-- '/path/archive' configuration key

	configuration_path_data: STRING = "/path/rcdata"
			-- '/path/rcdata' configuration key

	configuration_path_recipe: STRING = "/path/recipe"
			-- '/path/recipe' configuration key

	configuration_is_running_in_estudio: STRING = "is_running_in_estudio"
			-- If defined to be True in the config file, we will NOT indicate that we are "running in estudio" (see ROSE_APPLICATION.make_current).

feature -- Deliveries

	deliveries: attached ROSE_DELIVERIES_CONVENTION
			-- Deliveries-related conventions
			-- This class defines convention used for all supported deliveries, please always use this class when refering to files in a delivery
		once
			create Result
		end

feature -- IO constants

	maximum_verbosity: INTEGER = 3				-- max
	v_error: INTEGER = 3						-- error
	v_warning: INTEGER = 2						-- warning
	v_log: INTEGER = 1							-- log
	v_output: INTEGER = 0						-- output
	v_trace: INTEGER = -1						-- trace
	minimum_verbosity: INTEGER = -1				-- min
			-- Output verbosity levels

	vn_error: STRING = "error"					-- error
	vn_warning: STRING = "warning"				-- warning
	vn_log: STRING = "log"						-- log
	vn_output: STRING = "output"				-- output
	vn_trace: STRING = "trace"					-- trace
			-- Verbosity level names

	log_fatal: INTEGER = 3						-- fatal
	log_warning: INTEGER = 2					-- warning
	log_information: INTEGER = 1				-- info
			-- Log levels

	channel_input: STRING = "input"				-- input
	channel_output: STRING = "output"			-- output
	channel_echo: STRING = "echo"				-- echo
	channel_log: STRING = "log"					-- log
	channel_monitor: STRING = "monitor"			-- monitor
	current_IO_representation: STRING = "-"		-- Channel representation representing currently in-use channel
			-- Channel names

feature -- Command processing constants

	directory_separator: STRING = "\"			-- Directory separator
	directory_separators: STRING = "\/"			-- Directory separators
	path_separators: STRING = ";"				-- Character separating path entries
	executable_extension: STRING = ".exe"		-- Extension identifying executables

	is_directory_separator (a_character: CHARACTER): BOOLEAN
			-- Is 'a_character' a directory separator
		do
			Result := directory_separators.has (a_character)
		end

feature -- Access

	verbosity (a_level: STRING): INTEGER
			-- Verbosity refered to by 'a_level'
		do
			if verbosities.has (a_level) then
				Result := verbosities.item (a_level)
			elseif a_level /= Void and then a_level.is_integer then
				Result := a_level.to_integer
			end
		end

	verbosity_meaning (a_level: INTEGER): STRING
			-- Meaning of verbosity 'a_level'
		do
			from
				verbosities.start
			until
				verbosities.off
			loop
				if verbosities.item_for_iteration = a_level then
					Result := verbosities.key_for_iteration
				end
				verbosities.forth
			end
		end

	verbosities: attached HASH_TABLE [INTEGER,STRING]
			-- Verbosity levels and associated names
		once
			create Result.make (5)
			Result.put (v_error,vn_error)
			Result.put (v_warning,vn_warning)
			Result.put (v_log,vn_log)
			Result.put (v_output,vn_output)
			Result.put (v_trace,vn_trace)
		end

	severity_meaning (a_severity: INTEGER): STRING
			-- 'a_severity' level meaning.
		do
			inspect a_severity
			when log_fatal then
				Result := "fatal"
			when log_warning then
				Result := "warning"
			when log_information then
				Result := "info"
			else
				Result := "unknown severity"
			end
		end

	checked_path (a_dir, a_file: STRING): STRING
			-- Composed path from 'a_dir' and 'a_file', Void if such file does not exist
		do
			if a_dir /= Void and then a_file /= Void then
				Result := formatter.path (a_dir, a_file)
				if Result /= Void and then not is_file_readable (Result) then
					Result := Void
				end
			end
		end

	is_file_readable (a_path: STRING): BOOLEAN
			-- Is file with 'a_path' readable?
		local
			l_file: PLAIN_TEXT_FILE
			rescued: BOOLEAN
		do
			if not rescued and then a_path /= Void and then not a_path.is_empty then
				create l_file.make (a_path)
				Result := l_file.exists and then l_file.is_readable
			end
		rescue
			if not rescued then
				rescued := True
				retry
			end
		end

	is_file_writable (a_path: STRING): BOOLEAN
			-- Is file with 'a_path' writable (or creatable in its existing directory)?
		local
			l_file: PLAIN_TEXT_FILE
			rescued: BOOLEAN
		do
			if not rescued and then a_path /= Void and then not a_path.is_empty then
				create l_file.make (a_path)
				if l_file.exists then
					Result := l_file.is_writable
				else
					Result := is_directory_writable (formatter.directory (a_path))
				end
			end
		rescue
			if not rescued then
				rescued := True
				retry
			end
		end

	is_file_writable_or_creatable (a_path: STRING): BOOLEAN
			-- Is file with 'a_path' writable (either exists and writable, or doesn't exist but creatable)?
			-- This will return True if the first existing directory in 'a_path' is writable
			-- May have to create directories recursively for the file to actually be creatable
		local
			l_file: PLAIN_TEXT_FILE
			rescued: BOOLEAN
		do
			if not rescued and then a_path /= Void and then not a_path.is_empty then
				create l_file.make (a_path)
				if l_file.exists then
					Result := l_file.is_writable
				else
					Result := is_directory_writable_or_creatable (formatter.directory (a_path))
				end
			end
		rescue
			if not rescued then
				rescued := True
				retry
			end
		end

	clean_directory_path (a_path: STRING): STRING
			-- This function allows to bypass 2 bugs in the Eiffel runtime:
			-- paths with trailing slashes are wrongly always reported as non existent
			-- UNC paths of the form \\machine\share need a trailing slash to work
		local
			i: INTEGER
		do
			if a_path /= Void and then a_path.count > 1 and then is_directory_separator (a_path.item (a_path.count)) then
				Result := a_path.twin
				Result.replace_substring_all ("/", "\")
				from
					i := Result.count
				until
					i < 1 or else Result.item (i) /= '\'
				loop
					Result.keep_head (i - 1)
					i := i - 1
				end
				if Result.count = 2 and then Result.item (2) = ':' then
					Result.append_character ('\')
				end
			else
				Result := a_path
			end
			if Result /= Void then
				if Result.count > 4 and then Result.item (1) = '\' and then Result.item (2) = '\' and then Result.occurrences ('\') = 3 then
					-- Change "\\foo\bar" to "\\foo\bar\"
					Result := Result + "\"
				elseif Result.count = 2 and then Result.item (2) = ':' then
					-- Change "c:" to "c:\"
					Result := Result + "\"
				end
			end
		end

	is_directory_readable (a_path: STRING): BOOLEAN
			-- Is 'a_path' a valid directory (exists and is readable)?
		local
			l_dir: DIRECTORY
			l_path: STRING
			rescued: BOOLEAN
		do
			if not rescued then
				l_path := clean_directory_path (a_path)
				if l_path /= Void and then not l_path.is_empty then
					create l_dir.make (l_path)
					Result := l_dir.exists and then l_dir.is_readable
				end
			end
		rescue
			if not rescued then
				rescued := True
				retry
			end
		end

	is_directory_writable (a_path: STRING): BOOLEAN
			-- Is 'a_path' pointing to a directory that exists and is writable?
		local
			l_dir: DIRECTORY
			l_path: STRING
			rescued: BOOLEAN
		do
			if not rescued then
				l_path := clean_directory_path (a_path)
				if l_path /= Void and then not l_path.is_empty then
					create l_dir.make (l_path)
					Result := l_dir.exists and then l_dir.is_writable
				end
			end
		rescue
			if not rescued then
				rescued := True
				retry
			end
		end

	is_directory_writable_or_creatable (a_path: STRING): BOOLEAN
			-- Is 'a_path' pointing to a directory that is writable (either exists and writable, or doesn't exist but creatable)?
			-- This will return True if the first existing directory in 'a_path' is writable
		local
			l_dir: DIRECTORY
			l_path: STRING
			rescued: BOOLEAN
		do
			if not rescued then
				l_path := clean_directory_path (a_path)
				if l_path /= Void and then not l_path.is_empty then
					create l_dir.make (l_path)
					if l_dir.exists then
						Result := l_dir.is_writable
					else
						-- If directory doesn't exist, we return the first existing parent directory writable status
						l_path := formatter.directory (l_path)
						if l_path /= Void and then not l_path.is_empty then
							Result := is_directory_writable_or_creatable (l_path)
						end
					end
				end
			end
		rescue
			if not rescued then
				rescued := True
				retry
			end
		end

	is_directory_exists (a_path: STRING): BOOLEAN
			-- does directory `a_path' exist in the file system?
		require
			path_not_void: a_path /= Void
		local
			l_path: STRING
			rescued: BOOLEAN
		do
			if not rescued then
				l_path := formatter.clean_directory_path (a_path)
				Result := file_system.directory_exists (l_path)
			end
		rescue
			if not rescued then
				rescued := True
				retry
			end
		end

	is_valid_verbosity_name (a_level: STRING): BOOLEAN
			-- Is 'a_level' refering to a valid verbosity level?
		do
			Result := verbosities.has (a_level)
			if not Result and then a_level /= Void and then a_level.is_integer then
				Result := is_valid_verbosity (a_level.to_integer)
			end
		end

	is_valid_verbosity (a_level: INTEGER): BOOLEAN
			-- Is 'a_level' refering to a valid verbosity level?
		do
			Result := minimum_verbosity <= a_level and then a_level <= maximum_verbosity
		end


	is_valid_severity (a_severity: INTEGER): BOOLEAN
			-- Is 'a_severity' valid?
		do
			Result := a_severity = log_fatal
				or else a_severity = log_warning
				or else a_severity = log_information
		end

	is_valid_channel_name (a_name: STRING): BOOLEAN
			-- Is 'a_name' a valid channel name?
		do
			Result := is_valid_input_channel_name (a_name)
				or else is_valid_output_channel_name (a_name)
		end

	is_valid_input_channel_name (a_name: STRING): BOOLEAN
			-- Is 'a_name' a valid input channel name?
		do
			Result := a_name ~ channel_input
		end

	is_valid_output_channel_name (a_name: STRING): BOOLEAN
			-- Is 'a_name' a valid output channel name?
		do
			Result := a_name ~ channel_output
				or else a_name ~ channel_echo
				or else a_name ~ channel_log
				or else a_name ~ channel_monitor
		end

feature -- Basic operations

	recursive_create_directory (a_path: STRING)
			-- Recursively create dir with 'a_path'
		require
			path_ok: formatter.is_valid_path (a_path)
		local
			l_dir: DIRECTORY
		do
			create l_dir.make (a_path)
			if not l_dir.exists then
				l_dir.recursive_create_dir
			end
		rescue
			print ("Can't create directory '")
			print (a_path)
			print ("'%N")
		end

end
