note
	description: "IO handler for all shared IO routines"
	author: "Zoran Simic"
	last_editor: "$Author: mb $"
	date: "$Date: 2012/10/08 $"
	revision: "$Revision: #16 $"
	path: "$File: //rose/source/kernel/infrastructure/framework/io/rose_io_handler.e $"

class
	ROSE_IO_HANDLER

inherit
	ROSE_IO_CONVENTION

create
	make

feature {NONE} -- Initialization

	make
			-- Create a default IO handler.
		do
			create manager.make
			create time_measures.make
			create error.make (Current)
			create {ROSE_OUTCHANNEL_STANDARD} output.make
		end

feature -- Access

	manager: attached ROSE_IO_MANAGER
			-- Input/output manager

	input: ROSE_INCHANNEL
		-- Current input channel

	output: ROSE_OUTCHANNEL
		-- Current output channel

	echo: ROSE_OUTCHANNEL
		-- Echo channel (for broadcast)

	log: ROSE_OUTCHANNEL
		-- Log channel

	monitor: ROSE_OUTCHANNEL
		-- Monitoring channel

	error: attached ROSE_ERROR_HANDLER
		-- Current error handler

	time_measures: attached LINKED_STACK [ROSE_TIME_MEASURE]
		-- Time measures stack

feature -- Status report

	has_input: BOOLEAN
			-- Is the current input valid ?
		do
			Result := input /= Void and then input.is_open
		end

	has_output: BOOLEAN
			-- Is the current output valid ?
		do
			Result := output /= Void and then output.is_open
		end

	has_log: BOOLEAN
			-- Is the current log valid ?
		do
			Result := log /= Void and then log.is_open
		end

	has_some_output: BOOLEAN
			-- Has this handler some mean of output ?
		do
			Result := has_output or else has_log or else (echo /= Void and then echo.is_open)
		end

	is_interactive: BOOLEAN
			-- Are current input/output channels interactive?
		do
			Result := input /= Void and then input.is_interactive
				and then output /= Void and then output.is_interactive
		end

feature -- Status setting

	set_input (a_channel: ROSE_INCHANNEL)
			-- Set current input channel.
		do
			input := a_channel
		end

	set_output (a_channel: ROSE_OUTCHANNEL)
			-- Set current output channel.
		do
			output := a_channel
		end

	set_echo (a_channel: ROSE_OUTCHANNEL)
			-- Set 'echo' channel.
		do
			echo := a_channel
		end

	set_log (a_channel: ROSE_OUTCHANNEL)
			-- Set 'log' channel.
		do
			log := a_channel
		end

	set_monitor (a_channel: ROSE_OUTCHANNEL)
			-- Set 'monitor' channel.
		do
			monitor := a_channel
		end

	set_error_handler (a_error_handler: attached ROSE_ERROR_HANDLER)
			-- Set 'error' handler.
		do
			error.flush_and_acknowledge
			error := a_error_handler
			error.set_io_handler (Current)
		end

feature -- Basic operations

	flush
			-- Flush all channels.
		do
			error.flush
			flush_channel (output)
			flush_channel (echo)
			flush_channel (log)
			flush_channel (monitor)
		end

	flush_channel (a_channel: ROSE_OUTCHANNEL)
			-- Flush 'a_channel' if defined.
		do
			if a_channel /= Void then
				a_channel.flush
			end
		end

	close_input
			-- Close input channel
		do
			close_channel (input)
			input := Void
		end

	close_all
			-- Close all open channels.
		do
			flush
			close_channel (input)
			close_channel (output)
			if echo /= output then
				close_channel (echo)
			end
			close_channel (log)
			close_channel (monitor)
		end

	close_channel (a_channel: ROSE_CHANNEL)
			-- Close 'a_channel' if defined.
		do
			if a_channel /= Void then
				a_channel.close
			end
		end

	add_input (a_input: ROSE_INCHANNEL)
			-- Add 'a_input' channel type to this manager.
		do
			manager.input_channels.add (a_input)
		end

	add_output (a_output: ROSE_OUTCHANNEL)
			-- Add 'a_output' channel type to this manager.
		do
			manager.output_channels.add (a_output)
		end

feature -- Output

	output_log (a_severity, a_id: INTEGER; a_class, a_feature, a_message: STRING; a_args: ARRAY [ANY])
			-- Log specified 'a_message' with 'a_severity'.
		do
			if log /= Void then
				log.output_log (a_severity,a_id,a_class,a_feature,formatter.argumented (a_message,a_args))
			end
		end

	do_output (a_level: INTEGER; a_message: STRING; a_args: ARRAY [ANY])
			-- Output 'a_message' (with 'a_args') using 'a_level' of severity.
		local
			l_string: STRING
		do
			if a_message /= Void and then (output /= Void or else echo /= Void) then
				l_string := formatter.argumented (a_message,a_args)
			end
			if l_string /= Void then
				output_on (output,a_level,l_string)
				if not (echo ~ output) then
					output_on (echo,a_level,l_string)
				end
			end
		end

	do_output_once (a_level: INTEGER; a_message: STRING; a_args: ARRAY [ANY]) is
			-- Output 'a_message' (with 'a_args') using 'a_level' of severity.
			-- Do this only once per interpreted `a_message'
		local
			l_string: STRING
		do
			if a_message /= Void and then (output /= Void or else echo /= Void) then
				l_string := formatter.argumented (a_message,a_args)
			end
			if l_string /= Void then
				once_messages_sent.search (l_string)
				if not once_messages_sent.found then
					do_output (a_level, a_message, a_args)
					once_messages_sent.extend (l_string)
				end
			end
		end
	
	output_on (a_channel: ROSE_OUTCHANNEL; a_level: INTEGER; a_message: STRING)
			-- Output 'a_message' on 'a_channel'.
		do
			if a_channel /= Void then
				a_channel.output_message (a_level,a_message)
			end
		end

	start_time_measure (a_id, a_qualification: STRING)
			-- Start time measure for 'a_id' with 'a_qualification'.
		require
			valid_id: a_id /= Void
		do
			time_measures.put (create {ROSE_TIME_MEASURE}.make_started (a_id,a_qualification))
		end

	end_time_measure
			-- End last time measure and output representation on monitor.
		require
			time_measures_not_empty: not time_measures.is_empty
		local
			l_tm: ROSE_TIME_MEASURE
		do
			l_tm := time_measures.item
			time_measures.remove
			if monitor /= Void then
				l_tm.terminate
				monitor.output_time_measure (l_tm)
			end
		end
	
feature {NONE} -- Implementation
	
	once_messages_sent: HASHED_SET [STRING] is
			-- Collection of messages that should be output only once
		once
			create Result.make
		ensure
			non_void_result: Result /= Void
		end
	
end
