note

	description:

		"An object which is used in Rose_LAG_TARGET for determine the lagged interval"

	author: "Nelson Chui"
	last_editor: "$Author: mb $"
	date: "$Date: 2011/09/23 $"
	revision: "$Revision: #10 $"
	path: "$File: //rose/source/kernel/targets/rose_lag_interval.e $"

class ROSE_LAG_INTERVAL

inherit

	ROSE_PERIODICITY_ID_USE

create

	make

feature {NONE} -- Initialization

	make (a_periods, a_periodicity_id: INTEGER) is
			-- Create the interval
		require
			valid_periods: a_periods >= 0
			valid_periodicity_id: a_periodicity_id > 0
		do
			periods := a_periods
			periodicity_id := a_periodicity_id
			set_name
		end

feature -- Query

	lagged_date (a_target: ANY; a_date: ROSE_DATE): ROSE_DATE is
			-- lagged date
		do
			Result := lagged_n_date (a_date, 1)
		end

	lagged_n_date (a_date: ROSE_DATE; a_lag: INTEGER): ROSE_DATE is
			-- lagged n date
		local
			l_index: INTEGER
		do
			if is_monthly then
					-- Monthly is the majority of processing
				l_index := a_date.julian_month - periods * a_lag
				if month_cache.valid_index (l_index) then
					Result := month_cache.item (l_index).twin
				else
						-- Outside of the cache, create it explicitly
					Result := a_date.previous_month_end
					Result.add_n_months (1 - periods * a_lag)
				end
			elseif is_daily then
				Result := a_date.previous_date
				Result.add_n_days (1 - periods * a_lag)
			elseif is_weekly then
				Result := a_date.previous_week_end
				Result.add_n_weeks (1 - periods * a_lag)
			elseif is_annual then
				Result := a_date.previous_year_end
				Result.add_n_years (1 - periods * a_lag)
			end
		end

	lagged_number_of_intervals (a_start_date, a_end_date: ROSE_DATE): INTEGER is
			-- lagged number of intervals
		require
			dates_in_order: a_start_date <= a_end_date
		local
			l_date: ROSE_DATE
		do
			create l_date.make_ccyymmdd (a_start_date.ccyymmdd)
			from
			until
				l_date > a_end_date
			loop
				if is_daily then
					l_date.add_n_days (periods)
				elseif is_weekly then
					l_date.add_n_weeks (periods)
				elseif is_monthly then
					l_date.add_n_months (periods)
				elseif is_annual then
					l_date.add_n_years (periods)
				end
				if l_date <= a_end_date then
					Result := Result + 1
				end
			end
		end

	is_daily: BOOLEAN is
			-- is daily
		do
			Result := periodicity_id = periodicity_ids.daily_periodicity_id
		end

	is_weekly: BOOLEAN is
			-- is weekly
		do
			Result := periodicity_id = periodicity_ids.weekly_periodicity_id
		end

	is_monthly: BOOLEAN is
			-- is monthly
		do
			Result := periodicity_id = periodicity_ids.monthly_periodicity_id
		end

	is_annual: BOOLEAN is
			-- is annual
		do
			Result := periodicity_id = periodicity_ids.annual_periodicity_id
		end

	period_unit: STRING is
			-- period unit
		do
			if is_daily then
				Result := "Day"
			elseif is_weekly then
				Result := "Week"
			elseif is_monthly then
				Result := "Month"
			elseif is_annual then
				Result := "Year"
			end
		end

feature -- Access

	name: STRING

	periods: INTEGER

	periodicity_id: INTEGER

feature -- Basic operations

	infix "+" (a_other: like Current): like Current is
			-- infix +
		do
			if a_other.periodicity_id = periodicity_id then
				create Result.make (periodicity_id, periods + a_other.periods)
			end
		end

feature {NONE} -- Implementation

	set_name is
			-- Set name
		do
			if is_daily then
				name := periods.out + "D"
			elseif is_weekly then
				name := periods.out + "W"
			elseif is_monthly then
				name := periods.out + "M"
			elseif is_annual then
				name := periods.out + "Y"
			end
		end

	month_cache: ROSE_LINEAR_ARRAY [ROSE_DATE] is
			-- Month cache
		local
			l_today, l_date: ROSE_DATE
		once
			create Result.make_empty
			create l_today.make_today
			from
					-- Start from julian month 0 as this is used by some 
					-- systems to indicate minimum date
				Result.force (create {ROSE_DATE}.make_ccyymm_end (194912), 0)
			until
				month_cache.upper > l_today.julian_month
			loop
				create l_date.make_from_julian_month (Result.upper + 1)
				Result.extend (l_date)
			end
		ensure
			valid_result: Result /= Void
			no_void: not Result.has (Void)
		end

invariant

	valid_name: name /= Void
	valid_periods: periods >= 0
	valid_periodicity_id: periodicity_id > 0

end
