note
	description: "Machine configuration"
	author: "Zoran Simic"
	last_editor: "$Author: zs $"
	date: "$Date: 2011/04/19 $"
	revision: "$Revision: #11 $"
	path: "$File: //rose/source/kernel/infrastructure/framework/application/rose_machine_configuration.e $"

class
	ROSE_MACHINE_CONFIGURATION

inherit
	ROSE_SHARED_GLOBAL
	DEBUG_OUTPUT

create
	make, make_current

feature {NONE} -- Initialization

	make (a_host_name: STRING)
			-- Create for machine with 'a_host_name'.
		require
			host_name_ok: a_host_name /= Void
		do
			host_name := a_host_name
			refresh
		end

	make_current
			-- Create for current machine.
		do
			host_name := global.app.host_name
			refresh
		end

feature -- Access

	host_name: STRING
			-- Name of this machine

	role: STRING assign set_role
			-- Role of this machine, example: "development" or "production"

	nature: STRING assign set_nature
			-- Nature of this machine, ex: "Workstation", "Citrix", "Citrix GII"

	location: STRING assign set_location
			-- Physical location of this machine: ex: "Orinda", "Fairfield", "Virtual"

	build_schedule: STRING assign set_build_schedule
			-- Describes when it is OK to run builds on this machines

	canonical_name: STRING assign set_canonical_name
			-- Canonical name defined by Systems group
			-- examples: axaros.com/Orinda/Servers/VMware/Patch/ORPCB08
			--           axaros.com/Orinda/Desktops/AXA-ROS Citrix only/FFCTX03
			--           axaros.com/Fairfield/Servers/Externaly Accessible Citrix Servers/Patch/VMSVA61
			--           axaros.com/Orinda/Servers/Other-SystemsControl/Patch/ORSVQ05

	local_disks: STRING assign set_local_disks
			-- List of local disks on this machine, example: "CDEG"

	path: attached STRING
			-- Path to machine.cfg
		do
			Result := directory_path + "machine.cfg"
		end

	server_path: attached STRING
			-- Path to machine.cfg c:\servers or \\hostname\$c\servers
		do
			Result := directory_path + "servers"
		end

	is_identified: BOOLEAN
			-- Is this machine configuration object completely set?
		do
			Result := role /= Void and then not role.is_empty
				and then nature /= Void and then not nature.is_empty
				and then location /= Void and then not location.is_empty
		end

	has_canonical_name: BOOLEAN
			-- Is this machine configuration object completely set?
		do
			Result := canonical_name /= Void and then not canonical_name.is_empty
		end

	is_build_schedule_defined: BOOLEAN
			-- Is build_schedule defined?
		do
			Result := accepted_build_schedules.item (build_schedule)
		end

	is_complete: BOOLEAN
			-- Is this machine configuration object completely set?
		do
			Result := is_identified and then has_canonical_name and then is_build_schedule_defined
		end

	is_local: BOOLEAN
			-- Does this object correspond to current machine (where current process is running)
		do
			Result := host_name ~ global.app.host_name
		end

	is_modifiable: BOOLEAN
			-- Is this machine configuration modifiable by user running current process?
		local
			l_file: PLAIN_TEXT_FILE
			l_dir: DIRECTORY
		do
			create l_file.make (path)
			if not l_file.exists then
				create l_dir.make (directory_path)
				Result := l_dir.is_writable
			else
				Result := l_file.is_writable
			end
		end

	is_equivalent (a_other: ROSE_MACHINE_CONFIGURATION): BOOLEAN
			-- Is this machine config equivalent to 'a_other'?
		do
			if a_other /= Void and then host_name ~ a_other.host_name then
				Result := role ~ a_other.role
					and then nature ~ a_other.nature
					and then location ~ a_other.location
					and then build_schedule ~ a_other.build_schedule
					and then canonical_name ~ a_other.canonical_name
					and then local_disks ~ a_other.local_disks
			end
		end

	representation: attached STRING
			-- Textual representation
		do
			create Result.make_empty
			add_key_value_pair (Result, "role", role)
			add_key_value_pair (Result, "nature", nature)
			add_key_value_pair (Result, "location", location)
			add_key_value_pair (Result, "build_schedule", build_schedule)
			add_key_value_pair (Result, "canonical_name", canonical_name)
			add_key_value_pair (Result, "local_disks", local_disks)
		end

	debug_output: STRING
			-- String that should be displayed in debugger to represent `Current'.
		do
			create Result.make_empty
			Result.append_string (host_name)
		end

feature -- Basic operations

	refresh
			-- Refresh from C:\machine.cfg
		local
			l_buffer: ROSE_OUTCHANNEL_BUFFER
			l_config: ROSE_CONFIGURATION
			l_file: PLAIN_TEXT_FILE
			l_path: STRING
		do
			l_path := path
			create l_file.make (l_path)
			if l_file.exists and then l_file.is_readable then
				create l_buffer.make
				create l_config.make_default
				error.push_private (l_buffer)
				global.config_manager.load (l_config, Void, l_path)
				error.pop_private
				role := l_config.value ("role")
				nature := l_config.value ("nature")
				location := l_config.value ("location")
				build_schedule :=  l_config.value ("build_schedule")
				canonical_name := l_config.value ("canonical_name")
				local_disks := l_config.value ("local_disks")
			end
		end

	save: attached ROSE_STATUS
			-- Save to C:\machine.cfg
		local
			l_buffer: ROSE_OUTCHANNEL_BUFFER
			l_config: ROSE_CONFIGURATION
			l_path: STRING
		do
			create Result
			l_path := path
			if is_modifiable then
				create l_config.make_default
				l_config.set ("role", role)
				l_config.set ("nature", nature)
				l_config.set ("location", location)
				l_config.set ("build_schedule", build_schedule)
				l_config.set ("canonical_name", canonical_name)
				l_config.set ("local_disks", local_disks)
				l_config.set_path (l_path)
				create l_buffer.make
				l_buffer.open (Void)
				error.push_private (l_buffer)
				global.config_manager.save (l_config, Void, l_path)
				if error.signaled then
					Result.error (l_buffer.flush_to_string)
				else
					Result.outputf ("Saved machine configuration to $1", <<path>>)
				end
				error.pop_private
			else
				Result.errorf ("File '$1' is not writable", <<l_path>>)
			end
		ensure
			result_ok: Result /= Void
		end

	set_role (a_role: STRING)
			-- Set 'role'.
		do
			role := a_role
		end

	set_nature (a_nature: STRING)
			-- Set 'nature'.
		do
			nature := a_nature
		end

	set_location (a_location: STRING)
			-- Set 'location'.
		do
			location := a_location
		end

	set_build_schedule (a_build_schedule: STRING)
			-- Set 'build_schedule'.
		do
			build_schedule := a_build_schedule
		end

	set_canonical_name (a_canonical_name: STRING)
			-- Set 'canonical_name'
		do
			canonical_name := a_canonical_name
		end

	set_local_disks (a_local_disks: STRING)
			-- Set 'local_disks'.
		do
			local_disks := a_local_disks
		end

	auto_complete_canonical_name
			-- Auto-complete 'canonical_name' if none currently available
		do
			if canonical_name = Void or else canonical_name.is_empty then
				canonical_name := global.env.canonical_host_name
			end
		end

feature {NONE} -- Implementation

	directory_path: attached STRING
			-- Path to directory supposed to hold machine.cfg
		do
			if is_local then
				Result := global.hardcoded_path ("C:\")
			else
				Result := global.hardcoded_path ("\\" + host_name + "\c$\")
			end
		end

	accepted_build_schedules: HASH_TABLE [BOOLEAN, STRING]
			-- Accepted build_schedule values
		once
			create Result.make (0)
			Result.force (True, "always")
			Result.force (True, "never")
			Result.force (True, "non business hours")
			Result.force (True, "weekends")
		end

	add_key_value_pair (a_result, a_key, a_value: STRING)
			-- Add 'a_key' = 'a_value' to 'a_result', if 'a_value' is meaningful
		require
			result_ok: a_result /= Void
			key_ok: a_key /= Void and then not a_key.is_empty
		do
			if a_value /= Void and then not a_value.is_empty then
				a_result.append (a_key)
				a_result.append_character ('=')
				a_result.append_character ('"')
				a_result.append (a_value)
				a_result.append_character ('"')
				a_result.append_character ('%N')
			end
		end

invariant

	host_name_ok: host_name /= Void

end
