indexing
	description: "Send mail via SMTP"
	notice: "[
		A few comments on the use of this class:
		1.  Body text should be %N delimited for multi-line text
		2.  Hyperlinks will be created (by the mail client) where appropriate but do not
		    need to exist.
		3.  Attachments need to exist and NOT be open
		4.  Please avoid using two letter user ids as recipient (try to send a message to SO [Mark On] using Outlook),
			Preferably use a full email address like mk@axarosenberg.com
		5.  SMTP needs a full address: where MAPI accepts "Mark Howard" for a recipient, SMTP does not
		    and you need to say "Mark Howard <mk@axarosenberg.com>" or "mhoward@axarosenberg.com"
		6.  The 'subject' will be trimmed to the first 255 characters.
		7.  Use SMTP preferably, as it is more feature rich (you can send attachments without having
			to create a file, see API of ROSE_MAIL_CONTENT) and no Outlook profile needs to be up
			in order for your mail to successfully be sent.
		8. Currently, send_subject and send_body send a Cc: of the message to the sender.
			However, this is an implementation detail, and is not guaranteed by the interface.
			To explicitly control whether a Cc: is sent to the sender, use send_with_cc_self,
			send_cc_self, send_nocc_self, and especially send_cc_nobatch, which will send a Cc:
			to the sender unless the sender's user id is a batch user.
			(nb. this is different from "is a batch job")

]"
	author: "Zoran Simic"
	last_editor: "$Author: zs $"
	date: "$Date: 2011/01/26 $"
	revision: "$Revision: #17 $"
	path: "$File: //rose/source/kernel/infrastructure/network/rose_mail.e $"

class
	ROSE_MAIL

inherit
	ROSE_SHARED_GLOBAL
		redefine
			default_create
		end

create
	default_create,
	make,
	make_smtp

feature {NONE} -- Initialization

	default_create is
			-- Default creation procedure
		local
			l_smtp_server, s: STRING
			l_smtp_port, l_timeout: INTEGER
		do
			l_smtp_server := Global.app.configured_value ("mail/smtp/server")
			if l_smtp_server = Void then
				l_smtp_server := "arexch.axaros.com"
			end
			l_smtp_port := 25
			s := Global.app.configured_value ("mail/smtp/port")
			if s /= Void and then s.is_integer then
				l_smtp_port := s.to_integer
			end
			s := Global.app.configured_value ("mail/smtp/timeout")
			if s /= Void and then s.is_integer then
				l_timeout := s.to_integer
			end
			create smtp
			smtp.set_server (l_smtp_server)
			smtp.set_port (l_smtp_port)
			smtp.set_host (Global.app.host_name)
			smtp.set_timeout (l_timeout)
		end

	make is
			-- Default creation procedure
		do
			default_create
		end

	make_smtp is
			-- Prefer sending mail via SMTP
		do
			default_create
		end

feature -- Access

	last_error: STRING
			-- Error after last call to 'send'

	success: BOOLEAN is
			-- Was last 'send' successful
		do
			Result := last_error = Void
		end

feature -- Basic operations

	send (a_content: ROSE_MAIL_CONTENT) is
			-- Send an email to 'a_recipients' (comma seprated) with 'a_subject'
			-- and eventually (if not Void) 'a_attachments' attached.
		require
			content_ok: a_content /= Void and then a_content.is_valid
		local
			l_sender: ROSE_MAIL_ADDRESS
		do
			last_error := Void
			if a_content.sender = Void then
				create l_sender.make (Global.app.user.display_name, Global.app.user.email_address)
				a_content.set_sender (l_sender)
			end
			smtp.send (a_content)
			last_error := smtp.last_error
		end

	send_subject (a_subject, a_recipients: STRING) is
			-- Send a blank email to 'a_recipients' (comma seprated) with 'a_subject'.
		require
			recipients_ok: a_recipients /= Void and then not a_recipients.is_empty
		local
			l_content: ROSE_MAIL_CONTENT
		do
			create l_content.make
			l_content.to.set (a_recipients)
			l_content.set_subject (a_subject)
			send (l_content)
		end

	send_body (a_subject, a_recipients, a_body: STRING) is
			-- Send an email to 'a_recipients', with 'a_subject' and 'a_body.'
			-- Cc: may or may not be sent to the sender, depending on the default in ROSE_MAIL_CONTENT
		require
			recipients_ok: a_recipients /= Void and then not a_recipients.is_empty
		do
			send_with_cc_self (a_subject, a_recipients, a_body, False)
		end

	send_with_cc_self (a_subject, a_recipients, a_body: STRING; a_cc_self: BOOLEAN) is
			-- Send email to `a_recipients' with `a_subject' and `a_body' with `a_cc_self'
		require
			recipients_ok: a_recipients /= Void and then not a_recipients.is_empty
		local
			l_content: ROSE_MAIL_CONTENT
		do
			create l_content.make
			l_content.to.set (a_recipients)
			l_content.set_sender_copied (a_cc_self)
			l_content.set_subject (a_subject)
			l_content.set_text (a_body)
			send (l_content)
		end

	send_cc_self (a_subject, a_recipients, a_body: STRING) is
			-- Send email to `a_recipients' with `a_subject' and `a_body' and Cc: to the sender (self)
		require
			recipients_ok: a_recipients /= Void and then not a_recipients.is_empty
		do
			send_with_cc_self (a_subject, a_recipients, a_body, True)
		end

	send_nocc_self (a_subject, a_recipients, a_body: STRING) is
			-- Send email to `a_recipients' with `a_subject' and `a_body' without a Cc: to the sender (self)
		require
			recipients_ok: a_recipients /= Void and then not a_recipients.is_empty
		do
			send_with_cc_self (a_subject, a_recipients, a_body, False)
		end

	send_cc_nobatch (a_subject, a_recipients, a_body: STRING) is
			-- Send email to `a_recipients' with `a_subject' and `a_body'
			-- with a Cc: to the sender (self) if the sender is not a BATCH user id
		require
			recipients_ok: a_recipients /= Void and then not a_recipients.is_empty
		do
			send_with_cc_self (a_subject, a_recipients, a_body, not Global.env.is_batch_current_user)
		end

feature {NONE} -- Implementation

	smtp: SMTP_CLIENT
			-- Allows to send mail via SMTP protocol

end
