indexing
	description: "Mail address"
	author: "Zoran Simic"
	last_editor: "$Author: zs $"
	date: "$Date: 2010/02/17 $"
	revision: "$Revision: #6 $"
	path: "$File: //rose/source/kernel/infrastructure/network/rose_mail_address.e $"

class
	ROSE_MAIL_ADDRESS

inherit
	DEBUG_OUTPUT
		redefine
			is_equal
		end

create
	make, set

convert
	set ({STRING})

feature {NONE} -- Initialization

	make (a_name, a_address: STRING) is
			-- Create with 'a_name' and 'a_address'
		do
			set_name (a_name)
			set_address (a_address)
		end

feature -- Access

	name: STRING
			-- Name of person to send email to (really, the label, as in "John Smith" <john.smith@example.com>)

	address: STRING
			-- Actual email address (form john.smith@example.com)

	has_address: BOOLEAN is
			-- Is 'address' specified?
		do
			Result := address /= Void
		end

	is_equal (other: like Current): BOOLEAN is
			-- Is `other' attached to an object considered
			-- equal to current object?
		do
			Result := equal (address, other.address)
			if Result and then address = Void then
				Result := equal (name, other.name)
			end
		end

	is_valid_email_username (a_string: STRING): BOOLEAN is
			-- is `a_string' a valid username portion of an email address (i.e. is it not empty, and
			-- (contains only alphanumeric, or one of `valid_email_username_special_characters'
			-- n.b this is somewhat more restrictive than RFC 5322 section 3.4.1, which allows all kinds of crap in quoted strings here
		local
			i: INTEGER
			c: CHARACTER
		do
			if a_string /= Void and then not a_string.is_empty then
				from
					i := 1
					Result := True
				until
					not Result or else i > a_string.count
				loop
					c := a_string.item (i)
					Result := is_valid_address_character (c)
					i := i + 1
				end
			end
		end

	representation: STRING is
			-- Textual representation
		do
			if name /= Void then
				if address = Void then
					Result := name
				else
					Result := name + " <" + address + ">"
				end
			elseif address /= Void then
				Result := address
			else
				create Result.make_empty
			end
		ensure
			result_ok: Result /= Void
		end

	debug_output: STRING is
			-- String that should be displayed in debugger to represent `Current'.
		do
			Result := representation
		end

feature -- Basic operations

	set (a_string: STRING) is
			-- Set name and/or address from a basic interpretation of 'a_string' which may be of the form:
			--		name <address>
			--		name@domain  (treated as an address)
			--		name
		require
			string_not_void_or_empty: a_string /= Void and then not a_string.is_empty
		local
			l_str: STRING
			i, j: INTEGER
		do
			l_str := a_string.twin
			l_str.left_adjust
			l_str.right_adjust
			i := l_str.index_of ('<', 1)
			if i > 0 then
				j := l_str.index_of ('>', i + 1)
				if j = 0 then
					j := l_str.count + 1
				end
				set_name (l_str.substring (1, i - 1))
				set_address (l_str.substring (i + 1, j - 1))
			elseif l_str.has ('@') then
				set_address (l_str)
			else
				set_name (l_str)
				--set_address (l_str + "@axarosenberg.com")
			end
		end

	set_name (a_name: STRING) is
			-- Set 'name'.
		do
			name := a_name
			if name /= Void then
				name := name.twin
				name.left_adjust
				name.right_adjust
				if name.is_empty then
					name := Void
				end
			end
		end

	set_address (a_address: STRING) is
			-- Set 'address' to `a_address'
		local
			i, j: INTEGER
		do
			address := a_address
			if address /= Void then
				address := address.twin
				address.left_adjust
				address.right_adjust
				if address.is_empty then
					address := Void
				end
			end
			if address /= Void then
				i := address.index_of ('@', 1)
				if i > 0 then
					-- Truncate address after last invalid caracter
					from
						j := i + 1
					until
						j > address.count or else not is_valid_address_character (address.item (j))
					loop
						j := j + 1
					end
					if j <= address.count then
						address.keep_head (j-1)
					end
					-- Truncate address before first invalid caracter
					from
						j := i - 1
					until
						j < 1 or else not is_valid_address_character (address.item (j))
					loop
						j := j - 1
					end
					if j >= 1 then
						address.keep_tail (address.count - j)
					end
				end
			end
		end

feature {NONE} -- Implementation

	is_valid_address_character (c: CHARACTER): BOOLEAN is
			-- Is 'c' a valid character for an email address?
		do
			Result := c.is_alpha_numeric or else valid_email_username_special_characters.has (c)
		end

	valid_email_username_special_characters: STRING is "!#$%%&'*+-/=?^_`.{|}~"
		-- Special characters permitted in the username part of an email address (see RFC 3696 section 3).

end
