indexing
	description: "Mail attachment"
	author: "Zoran Simic"
	date: "$Date: 2008/12/01 $"
	revision: "Revision: $"

class
	ROSE_MAIL_ATTACHMENT

inherit
	DEBUG_OUTPUT

create
	default_create, set_path, set_content

feature -- Access

	name: STRING
			-- Name of attachment file

	path: STRING
			-- Path to attached file

	content: STRING
			-- Content of attached file

	content_type: UT_MEDIA_TYPE
			-- MIME content type

	is_binary: BOOLEAN
			-- Should `content' be considered as binary?
			-- (i.e. no special processing of new-lines)

	safe_name: STRING is
			-- 'name' if provided, generated name otherwise
		do
			if name = Void or else name.is_empty then
				name := "att-" + counter.item.out + ".txt"
				counter.put (counter.item + 1)
			else
				Result := name
			end
		ensure
			result_ok: Result /= Void and then not Result.is_empty
		end

	maximum_size: INTEGER is 2097152
			-- Maximum size for attached files in bytes (2 MB)
			-- Files larger than this will NOT be attached (simply ignored)

	encoded_base64: STRING is
			-- Contents of the attachment encoded in base64
		do
			Result := base64.encoded (full_content)
		end

	debug_output: STRING is
			-- String that should be displayed in debugger to represent `Current'.
		do
			Result := safe_name
		end

feature -- Basic operations

	set_path (a_path: STRING) is
			-- Set 'path'.
		require
			path_ok: a_path /= Void
		local
			i: INTEGER
		do
			path := a_path
			i := a_path.last_index_of ('\', a_path.count)
			if i = 0 then
				i := a_path.last_index_of ('/', a_path.count)
			end
			if i > 0 then
				name := a_path.substring (i + 1, a_path.count)
			else
				name := a_path
			end
			content := Void
		ensure
			path_set: path = a_path
			no_content: content = Void
		end

	set_content (a_name, a_content: STRING) is
			-- Set 'name' and 'content'
		require
			name_ok: a_name /= Void
			content_ok: a_content /= Void
		do
			name := a_name
			path := Void
			content := a_content
		ensure
			name_set: name = a_name
			no_path: path = Void
			content_set: content = a_content
		end

	set_binary (b: BOOLEAN) is
			-- Set `is_binary' to `b'.
		do
			is_binary := b
		ensure
			binary_set: is_binary = b
		end

	set_content_type (a_type: UT_MEDIA_TYPE) is
			-- Set `content_type' to `a_type'.
		do
			content_type := a_type
		ensure
			content_type_set: content_type = a_type
		end

feature {NONE} -- Implementation

	counter: CELL [INTEGER] is
			-- Counter used to generate 'unique' names
		once
			create Result.put (1)
		end

	full_content: MANAGED_POINTER is
			-- 'content' or contents of file with 'path'
		local
			l_file: RAW_FILE
			l_c_string: C_STRING
			l_size: INTEGER
			retried: BOOLEAN
		do
			if retried then
				Result := Void
			elseif content /= Void then
				if not content.is_empty then
					if not is_binary then
						content.replace_substring_all ("%R%N", "%N")
						content.replace_substring_all ("%N", "%R%N")
					end
					create l_c_string.make (content)
					create Result.make_from_pointer (l_c_string.item, l_c_string.count)
				end
			elseif path /= Void then
				create l_file.make (path)
				if l_file.exists and then l_file.is_readable then
					l_size := l_file.count
					if l_size > 0 and then l_size <= maximum_size then
						create Result.make (l_size)
						l_file.open_read
						l_file.read_to_managed_pointer (Result, 0, l_size)
						l_file.close
					end
				end
			end
		rescue
			if not retried then
				Result := Void
				retry
			end
		end

	base64: ROSE_BASE64_ENCODING is once create Result end
			-- Base64 encoder

end
