indexing
	description: "Mail content"
	author: "Zoran Simic"
	last_editor: "$Author: it $"
	date: "$Date: 2008/12/01 $"
	revision: "$Revision: #10 $"
	path: "$File: //rose/source/kernel/infrastructure/network/rose_mail_content.e $"

class
	ROSE_MAIL_CONTENT

inherit
	DEBUG_OUTPUT
		redefine
			default_create
		end

create
	default_create, make

feature {NONE} -- Initialization

	default_create is
			-- Default creation procedure
		do
			create to
			create cc
			create bcc
			set_subject (Void)
			set_text (Void)
			boundary := "====next_part_" + (create {DATE_TIME}.make_now).seconds.out + "===="
			create headers.make (Current)
			is_sender_copied := False -- could be: not (create{ROSE_EXECUTION_ENVIRONMENT}).is_batch_current_user
		end

	make is
			-- Create empty content.
		do
			default_create
		end

feature -- Access

	sender: ROSE_MAIL_ADDRESS
			-- Who is sending the email (valid only for SMTP)

	reply_to: ROSE_MAIL_ADDRESS
			-- 'Reply-to' address (optional)

	to: ROSE_MAIL_RECIPIENTS
			-- List of recipients to send email to

	cc: ROSE_MAIL_RECIPIENTS
			-- List of recipients to copy the email to ('cc' field)

	bcc: ROSE_MAIL_RECIPIENTS
			-- List of recipients to blind-copy the email to ('bcc' field)

	attachments: ARRAY [ROSE_MAIL_ATTACHMENT]
			-- Attachments to add to email

	subject: STRING
			-- Subject of the email

	text: STRING
			-- Body of the email

	is_sender_copied: BOOLEAN
			-- Should the 'sender' be sent a copy of this email as well?

	headers: ROSE_MAIL_HEADERS
			-- Mail headers

	has_attachments: BOOLEAN is
			-- Is 'attachments' populated?
		do
			Result := attachments /= Void and then not attachments.is_empty
		end

	is_plain_text: BOOLEAN is
			-- Is 'text' in plain text form?
		do
			Result := format = format_plain_text
		end

	is_html: BOOLEAN is
			-- Is 'text' in HTML form?
		do
			Result := format = format_html
		end

	is_valid: BOOLEAN is
			-- Is this content valid?
		do
			Result := to /= Void and then not to.is_empty
		end

	representation: STRING is
			-- Textual representation of this email
		do
			create Result.make_empty
			Result.append_string (headers.representation)
			Result.append_character ('%N')
			Result.append_string (body_smtp)
		end

	body_smtp: STRING is
			-- Body of email for SMTP (using base64 encoding for attachments)
		local
			l_attachment: ROSE_MAIL_ATTACHMENT
			i: INTEGER
		do
			create Result.make (1024)
			if attachments /= Void and then not attachments.is_empty then
				Result.append_character ('-')
				Result.append_character ('-')
				Result.append_string (boundary)
				Result.append_character ('%N')
			end
			Result.append_string (headers.header_value (headers.k_content_type, headers.body_content_type))
			Result.append_string (headers.header_value (headers.k_content_transfer_encoding, headers.k_8bit))
			Result.append_character ('%N')
			Result.append_string (text)
			Result.append_character ('%N')
			if attachments /= Void and then not attachments.is_empty then
				from
					i := attachments.lower
				until
					i > attachments.upper
				loop
					l_attachment := attachments.item (i)
					Result.append_string (attachment_body (l_attachment.safe_name, l_attachment.encoded_base64, l_attachment.content_type))
					i := i + 1
				end
			end
			if attachments /= Void and then not attachments.is_empty then
				Result.append_character ('-')
				Result.append_character ('-')
				Result.append_string (boundary)
				Result.append_character ('-')
				Result.append_character ('-')
				Result.append_character ('%N')
			end
		end

	debug_output: STRING is
			-- String that should be displayed in debugger to represent `Current'.
		do
			create Result.make_empty
			Result.append_string (subject)
		end

feature -- Basic operations

	set_sender (a_address: ROSE_MAIL_ADDRESS) is
			-- Set 'sender'.
		do
			sender := a_address
		end

	set_reply_to (a_address: ROSE_MAIL_ADDRESS) is
			-- Set 'reply_to'.
		do
			reply_to := a_address
		end

	set_subject (a_subject: STRING) is
			-- Set 'subject'.
		do
			if a_subject = Void then
				create subject.make_empty
			else
				subject := a_subject
			end
		end

	set_text (a_text: STRING) is
			-- Replace 'text' with 'a_text'.
		do
			if a_text = Void then
				create text.make_empty
			else
				text := a_text
			end
		end

	append, append_text (a_message: STRING) is
			-- Append 'a_message' to 'text'.
		do
			text.append_string (a_message)
		end

	set_plain_text is
			-- Set plain text format.
		do
			format := format_plain_text
		end

	set_html is
			-- Set html format.
		do
			format := format_html
		end

	set_attachments (a_attachments: ARRAY [ROSE_MAIL_ATTACHMENT]) is
			-- Set 'attachments'.
		do
			attachments := a_attachments
		ensure
			attachments_set: attachments = a_attachments
		end

	add_attachment (a_attachment: ROSE_MAIL_ATTACHMENT) is
			-- Add 'a_attachment'.
		require
			attachment_ok: a_attachment /= Void
		do
			if attachments = Void then
				create attachments.make (1, 0)
			end
			attachments.force (a_attachment, attachments.upper + 1)
		end

	add_file_attachment (a_path: STRING) is
			-- Add attachement from contents of file with 'a_path'.
		require
			path_ok: a_path /= Void
		do
			add_attachment (create {ROSE_MAIL_ATTACHMENT}.set_path (a_path))
		end

	add_content_attachment (a_name, a_content: STRING) is
			-- Add 'a_content' as an attachment (virtual file with 'a_name' assigned to that content).
		require
			name_ok: a_name /= Void and then a_content /= Void
		do
			add_attachment (create {ROSE_MAIL_ATTACHMENT}.set_content (a_name, a_content))
		end

	set_sender_copied (a_is_sender_copied: BOOLEAN) is
			-- Set 'is_sender_copied'.
		do
			is_sender_copied := a_is_sender_copied
		end

feature {ROSE_MAIL_HEADERS} -- Implementation

	boundary: STRING
			-- Boundary separating multipart message

feature {NONE} -- Implementation

	format: INTEGER
			-- Selected email format (see format_* constants for possibilities)

	format_plain_text: INTEGER is 0
			-- Plain text format

	format_html: INTEGER is 1
			-- HTML format

	format_rtf: INTEGER is 2
			-- RTF format

	Charset_parameter: STRING is "charset"
			-- Name of MIME charset parameter

	attachment_body (a_name, a_body_base64: STRING; a_content_type: UT_MEDIA_TYPE): STRING is
			-- Body of attachment for 'a_name' and 'a_body_base64' (which contains the encoded version of the body in base64).
		require
			a_name_not_void: a_name /= Void
			a_name_not_empty: not a_name.is_empty
		local
			l_type: STRING
		do
			if a_body_base64 /= Void then
				create Result.make_empty
				Result.append_character ('-')
				Result.append_character ('-')
				Result.append_string (boundary)
				Result.append_character ('%N')
				if a_content_type = Void then
					Result.append_string ("Content-Type: application/octet-stream; name=%"" + a_name + "%"")
					Result.append_character ('%N')
				else
					l_type := "Content-Type: " + a_content_type.type + "/" + a_content_type.subtype + "; "
					if a_content_type.has_parameter (Charset_parameter) then
						l_type := l_type + "charset=" + a_content_type.parameter (Charset_parameter) + "; "
					end
					Result.append_string (l_type + "name=%"" + a_name + "%"%N")
				end
				Result.append_string ("Content-Transfer-Encoding: base64")
				Result.append_character ('%N')
				Result.append_string ("Content-Disposition: attachment; filename=%"" + a_name + "%"")
				Result.append_character ('%N')
				Result.append_character ('%N')
				Result.append_string (a_body_base64)
				Result.append_character ('%N')
			end
		end

	base64: ROSE_BASE64_ENCODING is
			-- Base64 encoder
		once
			create Result
		end

end
