indexing
	description: "Mail headers"
	author: "Zoran Simic"
	date: "$Date: 2008/12/01 $"
	revision: "Revision: $"

class
	ROSE_MAIL_HEADERS

inherit
	ROSE_SHARED_GLOBAL
	DEBUG_OUTPUT

create
	make

feature {NONE} -- Initialization

	make (a_parent: ROSE_MAIL_CONTENT) is
			-- Create and bind with 'a_parent'
		require
			parent_ok: a_parent /= Void
		do
			parent := a_parent
			create keys.make (1, 0)
			create values.make (0)
			set (k_content_type, Void)
			set (k_content_transfer_encoding, k_8bit)
			set (k_from, Void)
			set (k_reply_to, Void)
			set (k_to, Void)
			set (k_cc, Void)
			set (k_subject, Void)
			set (k_mime_version, "1.0")
			set (k_x_mailer, "Eiffel SMTP client v1.1")
			set (k_x_log, global.app.user.full_login)
			set_importance_normal
		end

feature -- Access

	parent: ROSE_MAIL_CONTENT
			-- Parent content

	body_content_type: STRING is
			-- Content type for main mail body
		do
			if parent.is_html then
				Result := "text/html; " + charset
			else
				Result := "text/plain; " + charset
			end
		end

	has (a_key: STRING): BOOLEAN is
			-- Is 'a_key' defined in these headers?
		do
			Result := values.has (a_key)
		end

	value (a_key: STRING): STRING is
			-- Value associated to 'a_key'
		do
			if a_key = Void then
				Result := Void
			elseif a_key.is_equal (k_content_type) then
				if parent.has_attachments then
					Result := "multipart/mixed; boundary=%"" + parent.boundary + "%""
				else
					Result := body_content_type
				end
			elseif a_key.is_equal (k_from) then
				if parent.sender /= Void then
					Result := parent.sender.representation
				end
			elseif a_key.is_equal (k_reply_to) then
				if parent.reply_to /= Void then
					Result := parent.reply_to.representation
				end
			elseif a_key.is_equal (k_to) then
				if parent.to /= Void then
					Result := parent.to.representation
				end
			elseif a_key.is_equal (k_cc) then
				if parent.cc /= Void then
					Result := parent.cc.representation
				end
			elseif a_key.is_equal (k_subject) then
				Result := parent.subject
			else
				Result := values.item (a_key)
			end
		end

	full_value (a_key: STRING): STRING is
			-- Full value for 'a_key' (of the form "Key: Value%N")
		do
			Result := header_value (a_key, value (a_key))
		ensure
			result_ok: Result /= Void
		end

	representation: STRING is
			-- Textual representation
		local
			i: INTEGER
		do
			from
				create Result.make (1024)
				i := keys.lower
			until
				i > keys.upper
			loop
				Result.append_string (full_value (keys.item (i)))
				i := i + 1
			end
		ensure
			result_ok: Result /= Void
		end

	debug_output: STRING is
			-- String that should be displayed in debugger to represent `Current'.
		do
			Result := keys.count.out + " items"
		end

feature -- Basic operations

	set (a_key, a_value: STRING) is
			-- Set header associated to 'a_key' with 'a_value'.
		require
			valid_key: a_key /= Void and then not a_key.is_empty
		do
			if not has (a_key) then
				keys.force (a_key, keys.upper + 1)
			end
			values.force (a_value, a_key)
		end

	set_importance_low is
			-- Mark message to be of low importance.
		do
			set (k_importance, "Low")
		end

	set_importance_normal is
			-- Mark message to be of normal importance.
		do
			set (k_importance, "Normal")
		end

	set_importance_high is
			-- Mark message to be of high importance.
		do
			set (k_importance, "High")
		end

	do_all (a_agent: PROCEDURE [ANY, TUPLE [STRING, STRING]]) is
			-- Call 'a_agent' on all the (key, value) pairs in these headers.
		require
			valid_agent: a_agent /= Void
		local
			l_key: STRING
			i: INTEGER
		do
			from
				i := keys.lower
			until
				i > keys.upper
			loop
				l_key := keys.item (i)
				a_agent.call ([l_key, value (l_key)])
				i := i + 1
			end
		end

feature {NONE} -- Implementation

	keys: ARRAY [STRING]
			-- Header keys in the order they were inserted

	values: HASH_TABLE [STRING, STRING]
			-- Values' hash

feature {ROSE_MAIL_CONTENT} -- Constants

	header_value (a_key, a_value: STRING): STRING is
			-- Key-value pair of the form "Key: Value"
		do
			create Result.make_empty
			Result.append_string (a_key)
			Result.append_character (':')
			Result.append_character (' ')
			Result.append_string (a_value)
			Result.append_character ('%N')
		ensure
			result_ok: Result /= Void
		end

	k_content_type: STRING is "Content-Type"		-- Content-Type
	k_content_transfer_encoding: STRING is "Content-Transfer-Encoding"	-- Content-Transfer-Encoding
	k_content_disposition: STRING is "Content-Disposition"				-- Content-Disposition
	k_from: STRING is "From"						-- From
	k_reply_to: STRING is "Reply-To"				-- Reply-To
	k_to: STRING is "To"							-- To
	k_cc: STRING is "Cc"							-- Cc
	k_subject: STRING is "Subject"					-- Subject
	k_importance: STRING is "Importance"			-- Importance
	k_mime_version: STRING is "MIME-Version"		-- MIME-Version
	k_x_mailer: STRING is "X-Mailer"				-- X-Mailer
	k_x_log: STRING is "X-Log"					-- X-Log
	k_8bit: STRING is "8bit"						-- 8bit
	k_7bit: STRING is "7bit"						-- 7bit
	charset: STRING is "charset=%"iso-8859-1%""		-- Character set

invariant

	valid_keys: keys /= Void
	valid_values: values /= Void

end
