note

	description:

		"Miscellaneous analytical functions"

	keywords: "incomplete beta binomial_coefficient factorial log gamma sigmoid logistic dromedary error gauss"
	author: "Colin Adams"
	last_editor: "$Author: yc $"
	date: "$Date: 2012/09/24 $"
	revision: "$Revision: #21 $"
	path: "$File: //rose/source/kernel/infrastructure/math/analysis/rose_math_functions.e $"

class	ROSE_MATH_FUNCTIONS

inherit

	DOUBLE_MATH
		export {NONE} all
		end

	IEEE_MATH
		export {NONE} all end

	APPROXIMATE_EQUALITY
		export {NONE} all end
	ANY

feature -- Basic functions

	euler_integral, beta_function (a_x, a_y: DOUBLE): DOUBLE is
			-- Implementation of Euler beta function
		require
			x_positive: a_x > 0
			y_positive: a_y > 0
		do
			Result := (gamma(a_x) * gamma(a_y))/gamma(a_x + a_y)
		end

	binomial_coefficient (n_choices, k_combinations: INTEGER): DOUBLE is
			-- may be exact or approximated when overflows
			-- need to improve exact computations...
		require
			choices_pos: n_choices > 0
		do
			if k_combinations < 0 or k_combinations > n_choices then
				Result := 0
			elseif k_combinations = 0 then
				Result := 1
			else
				Result := factorial_exact_or_stirling(n_choices) / (factorial_exact_or_stirling(k_combinations) * factorial_exact_or_stirling(n_choices - k_combinations))
			end
		ensure
			-- bounds may actually depend on approximations...
			lower_bound: Result >= (n_choices / k_combinations)^k_combinations
			upper_bound: Result <= (n_choices^k_combinations) / factorial_stirling_approximation(k_combinations)
		end

	erf (x: DOUBLE): DOUBLE is
			-- Gauss error function
			-- See Numerical Recipees in C
		do
			if x < 0 then
				Result := incomplete_gamma_function(.5, x*x) - 1
			else
				Result := 1 - incomplete_gamma_function(.5, x*x)
			end
		ensure
			valid_result: Result >= -1 and Result <= 1
		end

	factorial (k: INTEGER): INTEGER is
			-- potentially expensive
		require
			k_min: k >= 0
			k_max: k < 13 -- potential overflow: use approximation functions
		local
			l_loop : INTEGER
		do
			Result := 1
			if k > 1 then
				from l_loop := 2 until l_loop > k loop
					Result := Result * l_loop
					l_loop := l_loop + 1
				end
			end
		end

	factorial_stirling_approximation (k: INTEGER): DOUBLE is
			-- valid when k is "sufficiently" large
			require
				k_min: k >= 0
			do
				Result := sqrt(2 * pi * k) * (k / euler)^k
			end

	factorial_exact_or_stirling (k: INTEGER): DOUBLE is
			-- exact for small k and approximated for large k
		require
			k_min: k >= 0
		do
			if k < 13 then
				Result := factorial(k)
			else
				Result := factorial_stirling_approximation (k)
			end
		ensure
			valid_factorial: Result >= 0
		end

	incomplete_beta (a_x, a_a, a_b, a_epsilon: DOUBLE; a_max_iterations: INTEGER): DOUBLE is
			-- I{`a_x'} (`a_a', `a_b') with a relative accuracy of `a_epsilon';
			-- At most `a_max_iterations' will be performed within `beta_continued_fraction'.
		require
			a_x_large_enough: a_x >= 0.0
			a_x_small_enough: a_x <= 1.0
			a_a_positive: a_a > 0.0
			a_b_positive: a_b > 0.0
			a_epsilon_positive: a_epsilon > 0.0
			a_max_iterations_positive: a_max_iterations > 0
		local
			l_bt: DOUBLE
		do
			if a_x = 0.0 then
				Result := 0.0
			elseif a_x = 1.0 then
				Result := 1.0
			else
				-- Compute factors in front of continued fraction.
				l_bt := exp (log_gamma (a_a + a_b) - log_gamma (a_a) - log_gamma (a_b) + a_a * log (a_x) + a_b * log (1.0 - a_x))
				if a_x < (a_a + 1.0) / (a_a + a_b + 2.0) then
					-- Use continued fraction directly.
					Result := l_bt * beta_continued_fraction (a_x, a_a, a_b, a_epsilon, a_max_iterations) / a_a
				else
					-- Make use of symmetry: I{a_x} (a_a, a_b) = 1 - I{1.0 - a_x} (a_b, a_a).
					Result := 1.0 - l_bt * beta_continued_fraction (1.0 - a_x, a_b, a_a, a_epsilon, a_max_iterations) / a_b
				end
			end
		ensure
			between_0_and_1_or_error: not isnan (Result) implies Result >= 0.0 and Result <= 1.0
		end

	complete_beta (a_a, a_b, a_epsilon: DOUBLE; a_max_iterations: INTEGER): DOUBLE is
			-- per standard definition
		do
			Result := incomplete_beta (1.0, a_a, a_b, a_epsilon, a_max_iterations)
		ensure
			not_nan: not isnan (Result)
		end

	regularized_incomplete_beta (a_x, a_a, a_b, a_epsilon: DOUBLE; a_max_iterations: INTEGER): DOUBLE is
			-- per ratio definition
		local
			l_num, l_denom: DOUBLE
		do
			l_num := incomplete_beta (a_x, a_a, a_b, a_epsilon, a_max_iterations)
			l_denom := complete_beta (a_a, a_b, a_epsilon, a_max_iterations)
			Result := l_num / l_denom
		ensure
			not_nan: not isnan (Result)
		end

	gamma (a_x: DOUBLE): DOUBLE is
			-- utility call.  For combinatorial computations, use log_gamma
		require
			a_x_positive : a_x > 0 -- because log_gamma implementation
			a_x_small : a_x < 13 -- grows very quickly (to improve)
		do
			Result := euler^(log_gamma(a_x))
		end

	log_gamma (a_x: DOUBLE): DOUBLE is
			-- Natural logarithm of Gamma (x);
			-- Algorithm is an approximation due to Lanczos (see Numerical methods in C - equation 6.1.5)
		require
			a_x_positive: a_x > 0
		local
			i, j: INTEGER
			x, y, l_tmp, l_ser: DOUBLE
		do
			x := a_x
			y := a_x
			l_tmp := x + 5.5 - (x + 0.5) * log (x + 5.5)
			l_ser := 1.000000000190015
			from
				i := 1
				j := log_gamma_coefficients.count
			until
				i > j
			loop
				y := y + 1
				l_ser := l_ser + (log_gamma_coefficients [i]) / y
				i := i + 1
			end
			Result := - l_tmp + log (2.5066282746310005 * l_ser / x)
		end

	sigmoid_min_max, logistic_min_max (a_value, a_steepness, a_center, a_min, a_max: DOUBLE): DOUBLE is
			-- Value of sigmoid (or logistic) function with bounds `a_min' and `a _max' for `a_value'
		require
			positive_steepness: a_steepness > 0.0
			ordered_bounds: a_min < a_max
		do
			Result := a_min + (a_max - a_min) * sigmoid (a_value, a_steepness, a_center)
		ensure
			preserves_nan: isnan (a_value) implies isnan (Result)
			definition: not isnan (a_value) implies Result = a_min + (a_max - a_min) * sigmoid (a_value, a_steepness, a_center)
			minimum_result_is_min: not isnan (a_value) implies Result >= a_min
			maximum_result_is_max: not isnan (a_value) implies Result <= a_max
		end

	sigmoid, logistic (a_value, a_steepness, a_center: DOUBLE): DOUBLE is
			-- Value of sigmoid (or logisitic) function for `a_value'
		require
			positive_steepness: a_steepness > 0.0
		do
			if isnan (a_value) then
				Result := a_value
			else
				Result := 1 / (1 + exp (- a_steepness * (a_value - a_center)))
			end
		ensure
			preserves_nan: isnan (a_value) implies isnan (Result)
			definition: not isnan (a_value) implies Result = 1 / (1 + exp (- a_steepness * (a_value - a_center)))
			non_negative_result: not isnan (a_value) implies Result >= 0.0
			maximum_result_is_one: not isnan (a_value) implies Result <= 1.0
		end

	dromedary (a_value, a_steepness_1, a_steepness_2, a_center_1, a_center_2: DOUBLE): DOUBLE is
			-- Product of two sigmoids to yield a hump-shaped curve
		require
			positive_steepness: a_steepness_1 > 0.0 and a_steepness_2 > 0.0
			ordered_centers: a_center_2 > a_center_1
		do
			if isnan (a_value) then
				Result := a_value
			else
				Result := sigmoid (a_value, a_steepness_1, a_center_1) * (1 - sigmoid (a_value, a_steepness_2, a_center_2))
			end
		ensure
			preserves_nan: isnan (a_value) implies isnan (Result)
			definition: not isnan (a_value) implies Result = sigmoid (a_value, a_steepness_1, a_center_1) * (1 - sigmoid (a_value, a_steepness_2, a_center_2))
			non_negative_result: not isnan (a_value) implies Result >= 0.0
			maximum_result_is_one: not isnan (a_value) implies Result <= 1.0
		end

feature -- Time Series

	sum_i (a_order: INTEGER): DOUBLE is
			-- Tex: \sum_{i=1}^{n} i = \frac{n (n+1)}{2}
		require
			positive_order: a_order > 0
		do
			Result := a_order * (a_order + 1) / 2
		end

	sum_x_power_i_n (a_order: INTEGER; a_x: DOUBLE): DOUBLE is
			-- Tex: \sum_{i=1}^{n} x^i = \frac{1 - x^n}{1 - x}
		require
			positive_order: a_order > 0
			valid_x: 0 < a_x
		do
			if a_order = 1 then
				Result := 1 -- algebraic simplication
			elseif a_order = 2 then
				Result := 1 + a_x -- algebraic simplification
			else
				Result := (1 - a_x^a_order) / (1 - a_x)
			end
		ensure
			positive_result: Result > 0
		end

	sum_x_power_i (a_x: DOUBLE): DOUBLE is
			-- Tex \sum_{i=1}^{\infty} x^i = \frac{x}{1 - x}
		require
			valid_x: 0 < a_x and a_x < 1
		do
			Result := a_x / (1 - a_x)
		ensure
			positive_result: Result > 0
		end

	sum_i_times_x_power_i (a_order: INTEGER; a_x: DOUBLE): DOUBLE is
			-- TeX: \sum_{i=1}^{n} i x^{i - 1}
		require
			positive_order: a_order > 0
			valid_x: 0 < a_x and a_x < 1
		local
			a_numerator, a_denominator: DOUBLE
		do
			if a_order = 1 then
				Result := 1 -- algebraic simplification
			elseif a_order = 2 then
				Result := 1 + 2 * a_x -- algebraic simplification
			else
				a_numerator := 1 - (a_order + 1) * (a_x ^ a_order) + a_order * (a_x ^(a_order + 1))
				a_denominator := (1 - a_x)^2
				Result := a_numerator / a_denominator
			end
		ensure
			positive_result: Result > 0
		end

feature {NONE} -- Implementation

	Non_zero_double: DOUBLE is 1.0e-30
			-- Small value distinguishable from zero;
			-- This value if for REAL_32 - what do we use for REAL_64?
	max_iterations_gamma: INTEGER is 100
			-- max iteration for gamma functions
	epsilon_gamma: DOUBLE is 3.0e-7
			-- relative epsilon accuracy for gamma functions

	log_gamma_coefficients: ARRAY [DOUBLE] is
			-- Magic numbers used by `log_gamma'
		once
			Result := <<76.18009172947416, -86.50532032941677, 24.01409824083091, -1.231739572450155, 0.1208650973866179e-2, -0.5395239384953e-5>>
		ensure
			result_not_void: Result /= Void
		end

	beta_continued_fraction (a_x, a_a, a_b, a_epsilon: DOUBLE; a_max_iterations: INTEGER): DOUBLE is
			-- Continued fraction for incomplete beta function;
			-- Algorithm is modified Lentz's method (see Numerical methods in C - chapter 5.2).
			-- `a_a' and `a_b' are the parameters of the incomplete beta function.
			-- `a_x' is the independent variable. `a_epsilon' is the relative accuracy.
		require
			a_x_large_enough: a_x >= 0.0
			a_x_small_enough: a_x <= 1.0
			a_a_positive: a_a > 0.0
			a_b_positive: a_b > 0.0
			a_epsilon_positive: a_epsilon > 0.0
			a_max_iterations_positive: a_max_iterations > 0
		local
			m, l_even: INTEGER
			l_aa, l_qab, l_qap, l_qam, l_del, c, d: DOUBLE
		do
			l_qab := a_a + a_b
			l_qap := a_a + 1.0
			l_qam := a_a - 1.0
			c := 1.0
			d := non_zero (1.0 - l_qab * a_x / l_qap)
			d := 1 / d
			Result := d
			from
				m := 1
			until
				m > a_max_iterations or (l_del - 1.0).abs < a_epsilon
			loop
				l_even := m * 2
				l_aa := m * (a_b - m) * a_x / ((l_qam + l_even) * (a_a + l_even))
				d := non_zero (1.0 + l_aa * d)
				d := 1 / d
				c := non_zero (1.0 + l_aa / c)
				Result := Result * d * c
				l_aa := - (a_a + m) * (l_qab + m) * a_x / ((a_a + l_even) * (l_qap + l_even))
				d := non_zero (1.0 + l_aa * d)
				d := 1 / d
				c := non_zero (1.0 + l_aa / c)
				l_del := d * c
				Result := Result * l_del
				m := m + 1
			end
			if m > a_max_iterations then
				Result := QNAN
			end
		end

	incomplete_gamma_function (a, x: DOUBLE): DOUBLE is
			-- Incomplete Gamma function
			-- See Numerical Recipees in C
		require
			valid_a: a > 0
			valid_x: x >= 0
		do
			if x < a + 1 then
				Result := 1 - incomplete_gamma_series (a, x)
			else
				Result := incomplete_gamma_fraction (a,x)
			end
		end

	incomplete_gamma_series (a, x: DOUBLE): DOUBLE is
			-- Incomplete Gamma series
			-- See Numerical Recipees in C
		require
			valid_x: x >= 0
		local
			n: INTEGER
			l_gln, l_ap, l_del, l_sum: DOUBLE
		do
			l_gln := log_gamma (a)
			if x = 0 then
				Result := 0
			else
				l_ap := a
				l_del := 1 / a
				l_sum := l_del
				from n := 1 until n > max_iterations_gamma or else Result /= 0 loop
					l_ap := l_ap + 1
					l_del := l_del * x / l_ap
					l_sum := l_sum + l_del
					if l_del.abs < l_sum.abs * epsilon_gamma then
						Result := l_sum * exp (-x + a * log(x) - l_gln)
					end
					n := n + 1
				end
			end
		ensure
			convergence_reached: x > 0 implies Result /= 0
		end

	incomplete_gamma_fraction (a, x: DOUBLE): DOUBLE is
			-- Incomplete gamma fraction
			-- See Numerical Recipees in C
		require
			valid_x: x > 0
		local
			l_break: BOOLEAN
			i: INTEGER
			b, c, d, h, l_an, l_del, l_gln: DOUBLE
		do
			l_gln := log_gamma(a)
			b := x + 1 - a
			c := 1 / non_zero_double
			d := 1 / b
			h := d
			from i := 1 until i > max_iterations_gamma or else l_break loop
				l_an := -i * (i-a)
				b := b + 2
				d := l_an * d + b
				if d.abs < non_zero_double then
					d := non_zero_double
				end
				c := b + l_an / c
				if c.abs < non_zero_double then
					c := non_zero_double
				end
				d := 1 / d
				l_del := d * c
				h := h * l_del
				l_break := (l_del - 1).abs < epsilon_gamma
				i := i + 1
			end
			if l_break then
				Result := exp( - x + a * log(x) - l_gln) * h
			end
		ensure
			convergence_reached: Result >= 0
		end

	non_zero (a_value: DOUBLE): DOUBLE is
			-- `a_value' significantly different from zero
		do
			if a_value.abs < Non_zero_double then
				Result := Non_zero_double
			else
				Result := a_value
			end
		end

end
