indexing

	description: "Object for computing means and standard deviations and dealing with outliers"

	keywords: "statistics, moments, standard deviation, mean, winsor, winsorization, winsorize, variance, z values, z scores, z, outliers"
	documentation: "file://///rmatsb/documentation/eiffel/Documentation/winsor.htm"
	author: "Mark Howard"
	last_editor: "$Author: myeager $"
	date: "$Date: 2014/03/13 $"
	revision: "$Revision: #88 $"
	path: "$File: //rose/source/kernel/infrastructure/math/rose_moments.e $"

class	ROSE_MOMENTS

inherit

	ANY

	IEEE_MATH
		export {NONE} all end

	ROSE_LEGACY_UNDEFINED
		export {NONE} all end

	SHARED_ASSERTION_MODE
		export {NONE} all end

	APPROXIMATE_EQUALITY
		export {NONE} all end

create

	make, make_empty, make_winsorized, make_winsorized_pass_limited,
	make_from_vectors, make_winsorized_from_vectors

feature -- Initialization

	make (a_data : attached ARRAY[DOUBLE]; a_weights: ARRAY [DOUBLE]) is
			-- Create moments bound to 'a_data' and 'a_weights' arrays.
		require
			a_data_not_void: a_data /= Void
			a_data_not_empty: not a_data.is_empty
			consistent_weight_count: a_weights /= Void implies a_weights.count = a_data.count
		do
			fill_arrays (a_data, a_weights)
			compute_moments
			compute_z_values
		end

	make_from_vectors (a_data : attached ABSTRACT_READABLE_VECTOR; a_weights: ABSTRACT_READABLE_VECTOR) is
			-- Create moments bound to 'a_data' and 'a_weights' arrays.
		require
			a_data_not_void: a_data /= Void
			a_data_not_empty: a_data.count > 0
			consistent_weight_count: a_weights /= Void implies a_weights.count = a_data.count
		do
			fill_arrays_from_vectors (a_data, a_weights)
			compute_moments
			compute_z_values
		end

	make_winsorized (a_data : attached ARRAY[DOUBLE]; a_weights: ARRAY [DOUBLE];
		a_lower_winsor_z, a_upper_winsor_z: DOUBLE) is
			-- Create winsorized moments bound to 'a_data' and 'a_weights' arrays.
		require
			a_data_not_void: a_data /= Void
			a_data_not_empty: not a_data.is_empty
			consistent_weight_count: a_weights /= Void implies a_weights.count = a_data.count
			bounds_make_sense: a_lower_winsor_z < 0 and a_upper_winsor_z > 0
		do
			fill_arrays (a_data,a_weights)
			compute_moments
			winsorize (a_lower_winsor_z,a_upper_winsor_z)
			compute_moments
			compute_z_values
		end

	make_winsorized_from_vectors (a_data : attached  ABSTRACT_READABLE_VECTOR; a_weights: ABSTRACT_READABLE_VECTOR;
		a_lower_winsor_z, a_upper_winsor_z: DOUBLE) is
			-- Create winsorized moments bound to 'a_data' and 'a_weights' arrays.
		require
			a_data_not_void: a_data /= Void
			a_data_not_empty: a_data.count > 0
			consistent_weight_count: a_weights /= Void implies a_weights.count = a_data.count
			bounds_make_sense: a_lower_winsor_z < 0 and a_upper_winsor_z > 0
		do
			fill_arrays_from_vectors (a_data,a_weights)
			compute_moments
			winsorize (a_lower_winsor_z,a_upper_winsor_z)
			compute_moments
			compute_z_values
		end

	make_winsorized_pass_limited (a_data : like data; a_weights: ARRAY [DOUBLE];
		a_lower_winsor_z, a_upper_winsor_z: DOUBLE;
		a_number_of_passes : INTEGER;
		a_default_mean, a_default_stddev : DOUBLE) is
			-- Create winsorized moments bound to 'a_data' and 'a_weights' arrays.
			-- Iterate no more than `a_number_of_passes' times.
		require
			a_data_not_void: a_data /= Void
			a_data_not_empty: not a_data.is_empty
			consistent_weight_count: a_weights /= Void implies a_weights.count = a_data.count
			bounds_make_sense: a_lower_winsor_z < 0 and a_upper_winsor_z > 0
		local
			l_pass : INTEGER
		do
			default_mean := a_default_mean
			default_standard_deviation := a_default_stddev

			l_pass := 1

			fill_arrays (a_data,a_weights)

			compute_moments
			l_pass := l_pass + 1

			create lower_outlier.make_empty (data.count)
			create upper_outlier.make_empty (data.count)
			create number_truncated.make_empty
			from
			until
				number_truncated.count > 0 and then number_truncated.last = 0 or l_pass >= a_number_of_passes
			loop
				truncate_outliers (a_lower_winsor_z,a_upper_winsor_z)
				l_pass := l_pass + 1

				if standard_deviation = 0.0 then
					compute_moments
				end
			end
			compute_moments
			compute_z_values
		end

	make_empty is
			-- Create moments with no data.
		do
			create data.make_empty
			create weights.make_empty
			initialize
		end

feature -- Access

	data:  attached ROSE_LINEAR_ARRAY [DOUBLE]
			-- Observation set;
			-- CAUTION: do not modify until finished with `Current'.

	weights: ROSE_LINEAR_ARRAY [DOUBLE]
			-- Weights for corresponding observations in `data'
			-- CAUTION: do not modify until finished with `Current'.

	data_units: ROSE_LINEAR_ARRAY [DOUBLE_PAIR]
			-- Association of elements of 'data' with corresponding element in 'weights',
			--  provided both elements are defined

	minimum: DOUBLE
			-- Weighted value of minimum element in `data_units' after applying corresponding weight;
 			-- Use weighted_percentile (0) to get the unweighted value of the weighted minimum element.
			-- Minimum element in `data_units' after applying corresponding weight

	maximum: DOUBLE
			-- Weighted value of maximum element in `data_units' after applying corresponding weight;
			-- Use weighted_percentile (100) to get the unweighted value of the weighted maximum element.

	standard_deviation: DOUBLE
			-- Standard deviation of `data_units';
			-- May differ from `running_standard_deviation'

	square_total: DOUBLE
			-- Sum of squares of elements of `data_units'

	total: DOUBLE
			-- Sum of elements of `data_units'

	count: INTEGER
			-- Number of observations in `data_units'

	weight: DOUBLE
			-- Sum of `weights'.

	weighted_total: DOUBLE
			-- Sum of products: for all i = 1 to `count' (`data_units [i].data' * `data_units [i].weight')

	z_values: ROSE_LINEAR_ARRAY [DOUBLE]
			-- Pre-calculated moments;
			-- `z_values [i]' = (`data [i]' - `mean') / `standard_deviation'

	equal_weights: BOOLEAN is
			-- Are all the weights equal?
			-- Time: O(weights.count)
		do
			Result := (not weights.is_empty)
				and then weights.for_all (agent equal_weights_item)
		end

	running_standard_deviation: DOUBLE is
			-- Standard deviation calculated using `unweighted_mean'
		require
			not_empty: count > 0
		do
			Result := (square_total / count - unweighted_mean * unweighted_mean)
			if Result >= 0.0 then
				Result := sqrt (Result)
			else
					-- Possible rounding error when all items are equal.
				if approx_equal (square_total / count, unweighted_mean * unweighted_mean) then
					Result := 0.0
				else
					Result := constants.missing_double
				end
			end
		ensure
			definition: one_zero_or_approx_equal (Result, sqrt (square_total / count - unweighted_mean * unweighted_mean))
		end

	unbiased_standard_deviation : DOUBLE is
			-- Standard deviation calculated using `unweighted_mean'
		require
			not_empty: count > 1
		do
			Result := (square_total - count * (unweighted_mean ^ 2)) / (count - 1)
			if Result >= 0.0 then
				Result := sqrt (Result)
			else
					-- Possible rounding error when all items are equal.
				if approx_equal (square_total / count, unweighted_mean * unweighted_mean) then
					Result := 0.0
				else
					Result := constants.missing_double
				end
			end
		ensure
			definition: one_zero_or_approx_equal (Result, sqrt (square_total / count - unweighted_mean * unweighted_mean))
		end
	weighted_standard_deviation: DOUBLE is
			-- Standard deviation of `data_units' calculated using `weighted_mean'
		require
			not_empty: data_units.count > 0
		local
			i, l_count: INTEGER
			l_mean, l_item, l_weight, l_diff, l_weighted_diffs: DOUBLE
		do
			l_mean := weighted_mean
			l_count := data_units.count
			from
				i := 1
			until
				i > l_count
			loop
				l_item := (data_units [i]).data
				l_weight := (data_units [i]).weight
				l_diff := l_item - l_mean
				l_weighted_diffs := l_weighted_diffs + l_diff * l_diff * l_weight
				i := i + 1
			end
			Result := (l_weighted_diffs / weight) ^ 0.5
		ensure
			-- it may be that the assertion following fails due to rounding errors - if this happens
			--  comment it out (as a temporary measure, stop monitoring postconditions
			--  for ROSE_MOMENTS)
			unweighted_consistency: equal_weights implies one_zero_or_approx_equal (Result, standard_deviation)
		end

	calculated_variance: DOUBLE is
			-- Variance of elements of `data'
		do
			Result := standard_deviation * standard_deviation
		ensure
			definition: Result = standard_deviation * standard_deviation
		end

	average_absolute_deviation: DOUBLE is
			-- Arithmetic mean of the absolute deviations from `unweighted_mean' of the defined elements of `data'
		require
			not_empty: count > 0
		local
			l_sum: DOUBLE
			l_count: INTEGER
			l_abs_deviation: DOUBLE
		do
			from data.start until data.after loop
				if is_defined_double (data.current_item) then
					l_abs_deviation := (data.current_item - unweighted_mean).abs
					l_sum := l_sum + l_abs_deviation
					l_count := l_count + 1
				end
				data.forth
			end
			Result:= l_sum / l_count
		ensure
			non_negative_result: Result >= 0
		end

	weighted_average_absolute_deviation: DOUBLE is
			-- Arithmetic mean of the absolute deviations from `weighted_mean' of the defined pairs of elements of `data' and `weight'
		require
			not_empty: count > 0
		local
			l_sum: DOUBLE
			l_abs_deviation: DOUBLE
		do
			from data.start weights.start until data.after loop
				if is_defined_double (data.current_item) and then is_defined_double(weights.current_item)  and then weights.current_item > 0 then
					l_abs_deviation := (data.current_item - weighted_mean).abs
					l_sum := l_sum + (l_abs_deviation * weights.current_item)
				end
				weights.forth
				data.forth
			end
			Result:= l_sum / weight
		ensure
			non_negative_result: Result >= 0
		end

	unweighted_mean: DOUBLE is
			-- Arithmetic mean of the elements of `data'
		do
			if count > 0 then
				Result:= total / count
			else
				Result := undefined_double
			end
		end

	mean, weighted_mean: DOUBLE is
			-- Weighted mean of the elements of `data'
		do
			if is_defined_double (weight) and then weight > 0 then
				Result := weighted_total / weight
			else
				Result := undefined_double
			end
		end

	median: DOUBLE is
			-- Median of defined elements of `data'
		do
			Result:= internal_median (double_array_from_units)
		end

	weighted_median: DOUBLE is
			-- Value at weighted median
		do
			Result := weighted_percentile (50)
		end

	weighted_percentile (a_pct: INTEGER): DOUBLE is
			-- Value at weighted `a_pct'th percentile
		require
			a_pct_large_enough: a_pct >= 0
 			a_pct_small_enough: a_pct <= 100
		local
			l_find: BOOLEAN
			l_sum, l_threshold: DOUBLE
			i: INTEGER
		do
			if data_units.count > 0 then
				l_threshold := (100 - a_pct) / 100
				data_units.set_comparison (agent data_compare_descending)
				if not data_units.is_sorted then
					data_units.sort_by (agent data_compare_descending)
				end

				l_sum := 0
				from
					data_units.start
				until
					l_find or else data_units.off
				loop
					i := data_units.index
					l_sum := l_sum + data_units.current_item.weight
					l_find := (l_sum / weight) > l_threshold
					data_units.forth
				end
				Result := (data_units [i]).data
			else
				Result := constants.legacy_missing_real_64
			end
		end

	weighted_first_quartile: DOUBLE is
			-- Value at first quartile, ignoring weights
		do
			Result := order_statistics [2]
		end

	weighted_third_quartile: DOUBLE is
			-- Value at third quartile, ignoring weights
		do
			Result := order_statistics [4]
		end

	percentile_of_observation (a_obs: DOUBLE): INTEGER is
			-- The percentile value for `a_obs' based on `data'
		local
			l_data: ROSE_LINEAR_ARRAY [DOUBLE]
			l_index: INTEGER
		do
			if sorted_data_vector = Void then
				l_data := double_array_from_units
				if not l_data.is_empty then
					l_data.set_comparison (agent double_compare)
					if not l_data.is_sorted then
						l_data.sort_by (agent double_compare)
					end
					create sorted_data_vector.make_from_array (l_data)
				end
			end
			if sorted_data_vector /= Void then
				l_index := sorted_data_vector.find_index (a_obs, sorted_data_vector.lower, sorted_data_vector.upper)
				Result := (l_index / sorted_data_vector.count * 100.0).truncated_to_integer
			end
		ensure
			valid_result: Result >= 0 and Result <= 100
		end

	order_statistics: ROSE_LINEAR_ARRAY [DOUBLE] is
			-- Unweighted quartiles, minimum and maximum
		local
			l_data: ROSE_LINEAR_ARRAY [DOUBLE]
			l_midpoint, l_25pt, l_75pt: INTEGER
			l_median, l_first_quantile, l_third_quantile: DOUBLE
		do
			if order_statistics_saved /= Void then
				Result := order_statistics_saved
			else
				l_data := double_array_from_units
				if l_data.count = 0 then
					create Result.make_with_default_value (1, 5, Undefined_double)
				else
					l_data.set_comparison (agent double_compare)
					if not l_data.is_sorted then
						l_data.sort_by (agent double_compare)
					end
					l_midpoint := (l_data.count + 1) // 2
					if l_data.count \\ 2 = 0 then
						l_median := (l_data [l_midpoint] + l_data [l_midpoint + 1]) / 2
					else
						l_median := l_data [l_midpoint]
					end
					l_25pt := (l_midpoint + 1) // 2
					l_75pt := (l_data.count + l_midpoint + 1) // 2
					if l_midpoint \\ 2 = 0 then
						l_first_quantile :=  (l_data [l_25pt] + l_data [l_25pt + 1]) / 2
						if l_data.count = 3 then
							l_third_quantile := l_data [3]
						else
							l_third_quantile :=  (l_data [l_75pt] + l_data [l_75pt + 1]) / 2
						end
					else
						l_first_quantile :=  l_data [l_25pt]
						l_third_quantile :=  l_data [l_75pt]
					end
					create Result.make_from_array (<<l_data.first, l_first_quantile, l_median, l_third_quantile, l_data.last>>)
				end
				order_statistics_saved := Result
			end
		ensure
			result_not_void: result /= Void
			five_statistics: Result.count = 5
			first_statistic_is_minimum: True
			second_statistic_first_quartile: True
			third_statistic_is_median: True
			fourth_statistic_is_third_quantile: True
			fifth_statistic_is_maximum: True
			results_cached: order_statistics_saved = Result
		end

	skewness : DOUBLE is
			-- skewness of the sample distribution
		local
			l_vec : VECTOR
		do
			l_vec := normalized_vector
			if l_vec /= Void then
				Result := l_vec.power (3).average
			end
		end

	kurtosis : DOUBLE is
			-- kurtosis of the sample distribution
		local
			l_vec : VECTOR
		do
			l_vec := normalized_vector
			if l_vec /= Void then
				Result := l_vec.power (4).average
			end
		end

	normalized_vector : VECTOR is
			-- normalized vector
		local
			l_array : ROSE_LINEAR_ARRAY[DOUBLE]
		do
			if data /= Void and then not data.is_empty then
				l_array := defined_double_array (data)
			elseif data_units /= Void and then not data_units.is_empty then
				l_array := double_array_from_units
			end
			if l_array /= Void and then not l_array.is_empty then
				create Result.make_from_array (l_array)
					Result.add_scalar (- mean)
					if standard_deviation /= 0 then
						Result.scale ( 1 / standard_deviation)
					end
			end
		end

feature -- Moving window statistics

	moving_average_lookback (a_x_index, a_lookback: INTEGER): DOUBLE is
			-- ??
		require
			data_not_empty: not data.is_empty
		local
			l_moments: ROSE_MOMENTS
			l_weights: ROSE_LINEAR_ARRAY [DOUBLE]
		do
			l_weights := weights.subarray ((a_x_index - a_lookback).max (data.lower).min (data.upper), a_x_index.min (data.upper))
			if {l_data : ROSE_LINEAR_ARRAY [DOUBLE]} data.subarray ((a_x_index - a_lookback).max (data.lower).min (data.upper), a_x_index.min (data.upper)) then
				if l_data.count > 1 then
					create l_moments.make (l_data, l_weights)
					Result := l_moments.mean
				elseif l_data.count = 1 then
					Result := l_data [l_data.lower]
				end
			end
		end

	moving_standard_deviation (a_x_index, a_lookback: INTEGER): DOUBLE is
			-- ??
		require
			data_not_empty: not data.is_empty
		local
			l_moments: ROSE_MOMENTS
			l_weights: ROSE_LINEAR_ARRAY [DOUBLE]
		do
			l_weights := weights.subarray ((a_x_index - a_lookback).max (data.lower).min (data.upper), a_x_index.min (data.upper))
			if {l_data : ROSE_LINEAR_ARRAY[DOUBLE]} data.subarray ((a_x_index - a_lookback).max (data.lower).min (data.upper), a_x_index.min (data.upper)) then
				if l_data.count > 1 then
					create l_moments.make (l_data, l_weights)
					Result := l_moments.standard_deviation
				elseif l_data.count = 1 then
					Result := 0.0
				end
			end
		ensure
			non_negative_result: Result >= 0.0
		end

	moving_median (a_x_index, a_lookback: INTEGER): DOUBLE is
			-- ??
		require
			data_not_empty: not data.is_empty
		local
			l_data: ROSE_LINEAR_ARRAY [DOUBLE]
		do
			l_data := data.subarray ((a_x_index - a_lookback).max (data.lower).min (data.upper), a_x_index.min (data.upper))
			if l_data.count > 0 then
				Result := internal_median (l_data)
			end
		end

	moving_minimum (a_x_index, a_lookback: INTEGER): DOUBLE is
			-- ??
		require
			data_not_empty: not data.is_empty
		local
			l_moments: ROSE_MOMENTS
		do
			if {l_data : ROSE_LINEAR_ARRAY [DOUBLE]}  data.subarray ((a_x_index - a_lookback).max (data.lower).min (data.upper), a_x_index.min (data.upper)) then
				if l_data.count > 1 then
					create l_moments.make (l_data, Void)
					Result := l_moments.minimum
				elseif l_data.count = 1 then
					Result := l_data [l_data.lower]
				end
			end
		end

	moving_maximum(a_x_index, a_lookback: INTEGER): DOUBLE is
			-- ??
		require
			data_not_empty: not data.is_empty
		local
			l_moments: ROSE_MOMENTS
		do
			if {l_data : ROSE_LINEAR_ARRAY[DOUBLE]} data.subarray ((a_x_index - a_lookback).max (data.lower).min (data.upper), a_x_index.min (data.upper)) then
				if l_data.count > 1 then
					create l_moments.make (l_data, Void)
					Result := l_moments.maximum
				elseif l_data.count = 1 then
					Result := l_data [l_data.lower]
				end
			end
		end

feature -- output

	simple_statistics_to_vector: VECTOR is
		-- output simple statistics: count, mean, sigma, min, max
		local
			l_names: ROSE_LINEAR_ARRAY [STRING]
		do
			create l_names.make_from_array (<<"count", "mean","sigma","min", "max">>)
			create Result.make_with_names (l_names)
			Result.put (count, 1)
			Result.put (mean, 2)
			Result.put (standard_deviation, 3)
			Result.put (minimum, 4)
			Result.put (maximum, 5)
		end

	to_vector: VECTOR is
		-- put commom statistics to vector
		-- including order statistics -- normally take longer time to compute
		do
			create Result.make(access.column_names.count)
			if count > 0 then
				Result.put(count, 1)
				Result.put(mean, 2)
				Result.put(standard_deviation, 3)
				Result.put(order_statistics.item(1), 4)
				Result.put(order_statistics.item(2), 5)
				Result.put(order_statistics.item(3), 6)
				Result.put(order_statistics.item(4), 7)
				Result.put(order_statistics.item(5), 8)
				Result.put(total, 9)
				Result.put(weight, 10)
				Result.put(weighted_total, 11)
				Result.put (square_total, 12)
			end
			Result.set_names (access.column_names)
		end

	access: ROSE_MOMENTS_ACCESS is
		-- moments relate: momonts names etc.
		once
			Create Result
		end

	to_normalized_value(a_value : DOUBLE) : DOUBLE is
			-- normalized value
		do
			if is_defined_double (a_value) and is_defined_double(weighted_mean) and (is_defined_double (standard_deviation) and then standard_deviation > standard_deviation_minimum_threshold) then
				Result := (a_value - weighted_mean) / standard_deviation
			else
				Result := a_value
			end
			if has_valid_winsor_factors then
				if Result > upper_z_value then
					Result := upper_z_value
				elseif Result < lower_z_value then
					Result := lower_z_value
				end
			end
		end

	has_valid_winsor_factors : BOOLEAN is
			-- has valid winsor factors
		do
			Result := lower_z_value < 0 and upper_z_value > 0
		end

	standard_deviation_minimum_threshold: DOUBLE is
			-- standard_deviation_minimum_threshold
		once
			Result := rglobal.config.double ("risk_descriptor/standard_deviation_minimum_threshold",0.000001)
		end

	rglobal : ROSE_SHARED_GLOBAL is
		-- rose global
		once
			create Result
		end

feature {NONE} -- Temporary accessors prior to ES 5.7 going live at AXA Rosenberg

	calculated_standard_deviation: DOUBLE is
			-- Standard deviation of `data'
		do
			Result := standard_deviation
		ensure
			definition: Result = standard_deviation
		end

	calculated_total: DOUBLE is
			-- Sum of elements of `data'
		do
			Result := total
		ensure
			definition: Result = total
		end

	computed_z_value (a_value: DOUBLE; a_standard_deviation: DOUBLE): DOUBLE is
			-- Deviation of `a_value' from `mean', divided by `a_standard_deviation',
			--  or zero if `a_standard_deviation' is zero
		do
			if a_standard_deviation > 0 then
				Result := (a_value - mean) / a_standard_deviation
			end
			end

	equal_weights_item (a_value: DOUBLE): BOOLEAN is
			-- Is value equal to fist item in `weights'?
		require
			weights_not_empty: not weights.is_empty
		do
			Result := weights.first = a_value
		end

feature -- Measurement

	lower_bound: DOUBLE
			-- Lower bound in calculation when truncating outliers

	upper_bound: DOUBLE
			-- Upper bound in calculation when truncating outliers

feature -- Basic operations

	initialize is
			-- Reset `Current'.
		do
			total := 0
			minimum := 0
			maximum := 0
			square_total := 0
			count := 0
			weight := 0
			weighted_total := 0
			standard_deviation := 0
			create data_units.make_empty
			create z_values.make_empty
			order_statistics_saved := Void
		end


 	add_observation_by_components (a_data, a_weight: DOUBLE) is
			-- Add a data point and it's associated weight to `Current' and update `count'.
		obsolete "Use add_observation instead"
		do
			data.extend (a_data)
			weights.extend (a_weight)
			count := count + 1
		end

 	add_observation (a_data, a_weight: DOUBLE) is
			-- Add a data point and it's associated weight to `Current'.
 		local
 			l_unit: DOUBLE_PAIR
 			l_weighted: DOUBLE
		do
			l_weighted := a_data * a_weight
			if count = 0 then
				minimum := l_weighted
				maximum := l_weighted
			else
				minimum := l_weighted.min (minimum)
				maximum := l_weighted.max (maximum)
			end
			total := total + a_data
			square_total := square_total + (a_data * a_data)
			weight := weight + a_weight
			weighted_total := weighted_total + l_weighted
			create l_unit.make (a_data, a_weight)
			data_units.extend (l_unit)
			count := count + 1
			order_statistics_saved := Void
		ensure
			order_statistics_cache_cleared: order_statistics_saved = Void
		end

	finalize_moments is
			-- Calculate moments that were not calculated in 'add_observation'.
		require
			non_empty: count >= 1
		do
			standard_deviation := running_standard_deviation
			if standard_deviation <= 0.0 or else not is_defined_double (standard_deviation) then
				standard_deviation := default_standard_deviation
			end
		end

	set_data_from_data_units is
			-- Transfer data units into data and weights
		do
			from data_units.start until data_units.off loop
				data.extend (data_units.current_item.data)
				weights.extend (data_units.current_item.weight)
				data_units.forth
			end
		end

	compute_moments is
			-- Build `data_units' from `data' and `weights' and set all moment attributes.
		do
			initialize
			from
				data.start
				weights.start
			until
				data.off
			loop
				if is_defined_double (data.current_item) and is_defined_double (weights.current_item)  then
					add_observation (data.current_item, weights.current_item)
				end
				data.forth
				weights.forth
			end
			if count >= 1 then
				finalize_moments
			end
		end

	normalize (a_array: ROSE_LINEAR_ARRAY [DOUBLE]) is
			-- Normalize values in `a_array' so they sum to one.
		require
			a_array_not_void: a_array /= Void
		local
			l_array: ROSE_LINEAR_ARRAY [DOUBLE]
			l_total: DOUBLE
		do
			from
				a_array.start
			until
				a_array.after
			loop
				l_total := l_total + a_array.current_item
				a_array.forth
			end
			if l_total /= 0 then
				create l_array.make_from_array (a_array)
				from
					l_array.start
				until
					l_array.after
				loop
					a_array [l_array.index] := l_array.current_item / l_total
					l_array.forth
				end
			end
		ensure
			elements_of_a_array_sum_to_one: True
		end

	winsorize (a_lower_winsor_z, a_upper_winsor_z: DOUBLE) is
			-- Iteratively truncate outliers to interval (`a_lower_winsor_z', `a_upper_winsor_z')
		require
			bounds_make_sense: a_lower_winsor_z < 0 and a_upper_winsor_z > 0
		do
			create lower_outlier.make_empty (data.count)
			create upper_outlier.make_empty (data.count)
			create number_truncated.make_empty
			from
			until
				number_truncated.count > 0 and then number_truncated.last = 0
			loop
				truncate_outliers (a_lower_winsor_z, a_upper_winsor_z)
				if standard_deviation = 0.0 then
					compute_moments
				end
			end
		end

	compute_z_values is
			-- Create and fill `z_values'.
		local
			l_std_deviation: DOUBLE
		do
			create z_values.make_empty
			l_std_deviation := weighted_standard_deviation
			from
				data.start
			until
				data.after
			loop
				if is_defined_double (data.current_item) then
					z_values.extend (computed_z_value (data.current_item, l_std_deviation))
				else
					z_values.extend (undefined_double)
				end
				data.forth
			end
		ensure
			z_values_not_void: z_values /= Void
			z_values_count: z_values.count = data.count
		end

feature {SUMMARY_STATISTICS_RECORD} -- Implementation

	kernel_weighted_mean: DOUBLE
			-- ??

	truncated_count: DOUBLE is
			-- ??
		do
			Result := lower_count + upper_count
		end

	lower_z_value: DOUBLE
			-- ??

	lower_count: INTEGER
			-- ??

	upper_z_value: DOUBLE
			-- ??

	upper_count: INTEGER
			-- ??

	beta: DOUBLE
			--??

	persistence_weight: DOUBLE
			-- How much to weight the original standard deviation

	default_standard_deviation: DOUBLE
			-- Lower bound on `standard_deviation'

	default_mean : DOUBLE
			-- Not used

	lower_outlier: ROSE_BITMAP
			-- ??

	upper_outlier: ROSE_BITMAP
			-- ??

	alpha: DOUBLE
			-- ??

	removed_data: ROSE_LINEAR_ARRAY [DOUBLE]
			-- Data points removed by `truncate_outliers' (not referenced)

	positive_root (a, b, c: DOUBLE): DOUBLE is
			-- Positive real root of the quadratic equation "ax^2 + bx + c", if any.
			-- Returns zero if no real positive root exists.
		local
			l_discriminant: DOUBLE
			l_rad: DOUBLE
			l_root_1, l_root_2: DOUBLE
		do
			l_discriminant := (b ^ 2 - 4 * a * c)
			if l_discriminant > 0 then
				l_rad := l_discriminant ^ 0.5
				l_root_1 :=  (-b + l_rad) / (2 * a)
				l_root_2 := (-b - l_rad) / (2 * a)
				if l_root_1 > 0 then
					Result := l_root_1
				elseif l_root_2 > 0 then
					Result := l_root_2
				end
			end
		ensure
			non_negative_result: Result >= 0
		end

	number_truncated: ROSE_LINEAR_ARRAY [INTEGER]
			-- Number of outliers truncated on each pass of `truncate_outliers'

feature {NONE} -- Initialization

	truncate_outliers (a_lower_winsor_z, a_upper_winsor_z: DOUBLE) is
			-- ??
		require
			bounds_make_sense: a_lower_winsor_z < 0 and a_upper_winsor_z > 0
		local
			l_kernel_mean, l_kernel_sum, l_kernel_squared_sum,
			l_kernel_weighted_sum, l_kernel_weight: DOUBLE
			l_worst_outlier_index: INTEGER
			l_a, l_b, l_c, l_d, l_e, l_weighted_mean, l_worst_outlier_norm, l_outlier_norm: DOUBLE
			l_lower_indices, l_upper_indices: ROSE_LINEAR_ARRAY[INTEGER]
			l_new_outlier_count: INTEGER
			l_lower_outlier, l_upper_outlier: BOOLEAN
			l_lower_weight: DOUBLE
			l_upper_weight: DOUBLE
		do
			if number_truncated = Void then
				create number_truncated.make_empty
				create lower_outlier.make_empty (data.count)
				create upper_outlier.make_empty (data.count)
			end
			lower_z_value := a_lower_winsor_z
			upper_z_value := a_upper_winsor_z
			lower_count := 0
			upper_count := 0
			lower_bound := weighted_mean + lower_z_value * standard_deviation
			upper_bound := weighted_mean + upper_z_value * standard_deviation
			create l_lower_indices.make_empty
			create l_upper_indices.make_empty
			from
				data.start
			until
				data.after
			loop
				if is_defined_double(data.current_item)  then
					l_lower_outlier := lower_outlier.item (data.index)
					l_upper_outlier := upper_outlier.item (data.index)
					if not (l_lower_outlier or l_upper_outlier)
						and data.current_item >= lower_bound
							and data.current_item <= upper_bound then
						l_kernel_sum := l_kernel_sum + data.current_item
						l_kernel_weight := l_kernel_weight + weights [data.index]
						l_kernel_weighted_sum := l_kernel_weighted_sum + data.current_item * weights [data.index]
						l_kernel_squared_sum := l_kernel_squared_sum + data.current_item * data.current_item
					elseif l_lower_outlier or else data.current_item < lower_bound then
						l_lower_indices.extend (data.index)
						if not l_lower_outlier then
							lower_outlier.put (true,data.index)
							l_new_outlier_count := l_new_outlier_count + 1
						end
						lower_count := lower_count + 1
						l_lower_weight := l_lower_weight + weights [data.index]
					elseif l_upper_outlier or else data.current_item > upper_bound then
						l_upper_indices.extend (data.index)
						if not l_upper_outlier then
							upper_outlier.put (True,data.index)
							l_new_outlier_count := l_new_outlier_count + 1
						end
						upper_count := upper_count + 1
						l_upper_weight := l_upper_weight + weights [data.index]
					end
				end
				data.forth
			end
			number_truncated.extend (l_new_outlier_count)
			if l_new_outlier_count > 0 then
				alpha := (l_lower_weight * lower_z_value + l_upper_weight * upper_z_value) / l_kernel_weight + lower_z_value
				beta := (l_lower_weight * lower_z_value + l_upper_weight * upper_z_value) / l_kernel_weight + upper_z_value
				kernel_weighted_mean := l_kernel_weighted_sum / l_kernel_weight
				l_a := (lower_count * alpha + upper_count * beta) / count
				l_b := (l_kernel_sum + truncated_count * kernel_weighted_mean) / count
				l_c := (lower_count * alpha * alpha + upper_count * beta * beta)/ count
				l_d := (2 * lower_count * kernel_weighted_mean * alpha + 2 * upper_count * kernel_weighted_mean * beta) / count
				l_e := (l_kernel_squared_sum + truncated_count * kernel_weighted_mean * kernel_weighted_mean) / count
				standard_deviation := positive_root(1 + l_a * l_a - l_c,2 * l_a * l_b - l_d,l_b * l_b - l_e)
				if standard_deviation > 0 then
					l_weighted_mean := kernel_weighted_mean + standard_deviation * (l_lower_weight * lower_z_value + l_upper_weight * upper_z_value) / l_kernel_weight
					lower_bound := l_weighted_mean + lower_z_value * standard_deviation
					upper_bound := l_weighted_mean + upper_z_value * standard_deviation
					total := l_kernel_sum + lower_count * lower_bound + upper_count * upper_bound
					square_total := l_kernel_squared_sum + lower_count * lower_bound * lower_bound + upper_count * upper_bound * upper_bound
					from
						l_lower_indices.start
					until
						l_lower_indices.after
					loop
						data [l_lower_indices.current_item] := lower_bound
						l_lower_indices.forth
					end
					from
						l_upper_indices.start
					until
						l_upper_indices.off
					loop
						data [l_upper_indices.current_item] := upper_bound
						l_upper_indices.forth
					end
				else
					l_kernel_mean := l_kernel_sum / (count - lower_count - upper_count)
					from
						l_lower_indices.start
					until
						l_lower_indices.after
					loop
						l_outlier_norm := (data [l_lower_indices.current_item] - l_kernel_mean).abs
						if l_outlier_norm > l_worst_outlier_norm then
							l_worst_outlier_norm := l_outlier_norm
							l_worst_outlier_index := l_lower_indices.current_item
						end
						l_lower_indices.forth
					end
					from
						l_upper_indices.start
					until
						l_upper_indices.after
					loop
						l_outlier_norm := (data [l_upper_indices.current_item] - l_kernel_mean).abs
						if l_outlier_norm > l_worst_outlier_norm then
							l_worst_outlier_norm := l_outlier_norm
							l_worst_outlier_index := l_upper_indices.current_item
						end
						l_upper_indices.forth
					end
					if l_worst_outlier_index > 0 then
						if removed_data = Void then
							create removed_data.make_empty
						end
						removed_data.extend (data [l_worst_outlier_index])
						data [l_worst_outlier_index] := l_kernel_mean
					end
					-- commented out 20070301 by Colin Adams, as it fails.
 					-- Proper fix needed.
					-- Probably zero observations in kernel causes the problem.
					-- check
					--	l_worst_outlier_index > 0
					--end					
				end
			end
			--| compute_moments
		end

	fill_arrays (a_data, a_weights: ARRAY [DOUBLE]) is
			-- Fill `data' from `a_data' and `weights' from `a_weights'
		require
			a_data_not_void: a_data /= Void
			a_weights_long_enough: a_weights /= Void implies a_weights.count >= a_data.count
		local
			i: INTEGER
		do
			create data.make_from_array (a_data)
			if a_weights = Void then
				create weights.make_empty
				-- explicit loop as agent call will check invariant and fail
				from
					i := 1
				until
					i > data.count
				loop
					weights.extend (1.0)
					i := i + 1
				end
			else
				create weights.make_from_array (a_weights)
			end
		end

	fill_arrays_from_vectors (a_data, a_weights: ABSTRACT_READABLE_VECTOR) is
			-- Fill `data' from `a_data' and `weights' from `a_weights'
		require
			a_data_not_void: a_data /= Void
			a_weights_long_enough: a_weights /= Void implies a_weights.count >= a_data.count
		local
			i: INTEGER
		do
			data := a_data.to_array
			if a_weights = Void then
				create weights.make_empty
				-- explicit loop as agent call will check invariant and fail
				from
					i := 1
				until
					i > data.count
				loop
					weights.extend (1.0)
					i := i + 1
				end
			else
				weights := a_weights.to_array
			end
		end


feature {NONE} -- Agents

	double_compare (a_1, a_2: DOUBLE): BOOLEAN is
			-- Is `a_1' less than
		do
			Result := a_1 < a_2
		ensure
			definition: Result implies a_1 < a_2
			asymmetric: result implies not double_compare (a_2, a_1)
		end

	data_compare_descending (a, b: DOUBLE_PAIR): BOOLEAN is
			-- Is `a' greater than `b', ignoring their weights?
		do
			Result := a.data > b.data
		ensure
			definition: Result implies a.data > b.data
			asymmetric: result implies not data_compare_descending (b, a)
		end

	extend_array (a_array: ROSE_LINEAR_ARRAY [DOUBLE]; a_value: DOUBLE) is
			-- Eextend `a_array' with `a_value'.
		require
			a_array_not_void: a_array /= Void
		do
			a_array.extend (a_value)
		ensure
			one_more: a_array.count = old a_array.count + 1
			last_value_correct: a_array.last = a_value
		end

	put_data (a_array: ROSE_LINEAR_ARRAY [DOUBLE]; a_pair: DOUBLE_PAIR; a_index: INTEGER) is
			-- Put `a_pair.data' at `a_array [a_index]'.
		require
			a_array_not_void: a_array /= Void
			a_pair_not_void: a_pair /= Void
			a_index_small_enough: a_index <= a_array.count
			a_index_large_enough: a_index > 0
		do
			a_array [a_index] := a_pair.data
		ensure
			element_set: a_array [a_index] = a_pair.data
		end

	is_defined_data (a_pair: DOUBLE_PAIR): BOOLEAN is
			--	Is `a_pair.data' defined?
		require
			a_pair_not_void: a_pair /= Void
		do
			Result := is_defined_double (a_pair.data)
		ensure
			definition: Result implies is_defined_double (a_pair.data)
			corollary: not Result implies not is_defined_double (a_pair.data)
		end

	accumulate_sum (a_sum: DS_CELL [DOUBLE]; a_value: DOUBLE) is
			-- Add `a_double' to `a_sum.item'.
		require
			a_sum_not_void: a_sum /= Void
		do
			a_sum.put (a_sum.item + a_value)
		ensure
			sum_accumulated: a_sum.item = old a_sum.item + a_value
		end

feature {NONE} -- Implementation

	order_statistics_saved: ROSE_LINEAR_ARRAY [DOUBLE]
			-- Cache for `order_statistics'
	sorted_data_vector: VECTOR

	internal_median (a_array: ROSE_LINEAR_ARRAY [DOUBLE]): DOUBLE is
			-- Unweighted median of defined elements of `a_array'
		require
			a_array_not_void: a_array /= Void
		local
			l_data: ROSE_LINEAR_ARRAY [DOUBLE]
			l_even: BOOLEAN
			l_midpoint: INTEGER
		do
			l_data := defined_double_array (a_array)
			l_data.set_comparison (agent double_compare)
			if not l_data.is_sorted then
				l_data.sort_by (agent double_compare)
			end

			l_midpoint := l_data.count // 2
			if l_data.count \\ 2 = 0 then
				l_even := True
			else
				l_even := False
			end
			if l_data.count /= 0 then
				if l_even then
					Result := (l_data [l_midpoint] + l_data [l_midpoint + 1]) / 2
				else
					Result := l_data [l_midpoint + 1]
				end
			end
		ensure
			defined_result: not isnan (Result)
			zero_result_may_be_meaningless: Result = 0.0 implies True
		end

	defined_double_array (a_array: ROSE_LINEAR_ARRAY [DOUBLE]): ROSE_LINEAR_ARRAY [DOUBLE] is
			-- Defined elements of `a_array', in the same order
		require
			a_array_not_void: a_array /= Void
		do
			create Result.make_empty
			a_array.do_if (agent extend_array (Result, ?), agent is_defined_double)
		ensure
			result_not_void: Result /= Void
			all_defined: Result.for_all (agent is_defined_double)
		end

	double_array_from_units: ROSE_LINEAR_ARRAY [DOUBLE] is
			-- Data (weightless) elements of `data_units'
		do
			create Result.make(1, data_units.count)
			data_units.do_all_with_index (agent put_data (Result, ?, ?))
		ensure
			result_not_void: Result /= Void
			all_defined_result: Result.for_all (agent is_defined_double)
			same_count_as_data_units: Result.count = data_units.count
		end

invariant

	data_not_void: data /= Void
	weights_not_void: weights /= Void
	consistent_weight_count: weights.count = data.count
	data_units_not_void: data_units /= Void
	all_data_units_defined: assertion_mode.is_checking_expensive_assertions implies data_units.for_all (agent is_defined_data)

end
