indexing
	description: "Multi-channel: set of several channels"
	author: "Zoran Simic"
	date: "$Date: 2008/12/01 $"
	revision: "$Revision: #9 $"

class ROSE_OUTCHANNEL_BROADCAST

inherit
	ROSE_OUTCHANNEL
		redefine
			output_message, output_log, output_time_measure
		end

create
	make_empty, make

feature {NONE} -- Initialisation

	make_empty is
			-- Create an empty broadcast channel.
		do
			channels := Void
			accepted_verbosity := minimum_verbosity
		end

	make (a_channels: ARRAY [ROSE_OUTCHANNEL]) is
			-- Create a broadcast on 'a_channels'.
		local
			i: INTEGER
		do
			make_empty
			if a_channels /= Void then
				from
					i := a_channels.lower
				until
					i > a_channels.upper
				loop
					add_channel (a_channels.item (i))
					i := i + 1
				end
			end
		end

feature -- Access

	id: STRING is "broadcast"
		-- Channel type id

	channels: LINKED_LIST [ROSE_OUTCHANNEL]
		-- Broadcast channels

feature -- Status report

	is_empty: BOOLEAN is
			-- Is this broadcast empty (no channels in)?
		do
			Result := channels = Void or else channels.is_empty
		end

	is_open: BOOLEAN is
			-- Is at least one of the broadcast channels open ?
		do
			if channels /= Void then
				from
					channels.start
				until
					Result or else channels.off
				loop
					Result := channels.item.is_open
					channels.forth
				end
			end
		end

feature -- Basic operations

	add_channel (a_channel: ROSE_OUTCHANNEL) is
			-- Add 'a_channel' to this broadcast.
		do
			if a_channel /= Void then
				if channels = Void then
					create channels.make
					channels.compare_objects
				end
				if not channels.has (a_channel) then
						--| Avoid duplicates
					channels.extend (a_channel)
				end
			end
		end

	close_id (a_id: STRING) is
			-- Close and remove all broadcast channels of 'a_id'.
		do
			if channels /= Void then
				from
					channels.start
				until
					channels.off
				loop
					if equal (a_id,channels.item.id) then
						channels.item.close
						channels.remove
					else
						channels.forth
					end
				end
			end
		end

	close_channel (a_channel: ROSE_OUTCHANNEL) is
			-- Close and remove 'a_channel' from broadcast.
		do
			if channels /= Void then
				from
					channels.start
				until
					channels.off
				loop
					if a_channel = channels.item then
						channels.item.close
						channels.remove
					else
						channels.forth
					end
				end
			end
		end

	remove_channel (a_channel: ROSE_OUTCHANNEL) is
			-- Remove all channels equal to 'a_channel' from this broadcast (without closing 'a_channel').
		do
			if channels /= Void then
				from
					channels.start
				until
					channels.off
				loop
					if equal (a_channel,channels.item) then
						channels.remove
					else
						channels.forth
					end
				end
			end
		end

	output_message (a_level: INTEGER; a_message: STRING) is
			-- Output 'a_message' with 'a_level' on this channel.
		do
			if a_message /= Void and then channels /= Void then
				from
					channels.start
				until
					channels.off
				loop
					channels.item.output_message (a_level,a_message)
					channels.forth
				end
			end
		end

	output_log (a_severity, a_id: INTEGER; a_class, a_feature, a_message: STRING) is
			-- Specific output for logs (redefinable)
		do
			if a_message /= Void and then channels /= Void then
				from
					channels.start
				until
					channels.off
				loop
					channels.item.output_log (a_severity,a_id,a_class,a_feature,a_message)
					channels.forth
				end
			end
		end

	output_time_measure (a_tm: ROSE_TIME_MEASURE) is
			-- Specific output for time measures...
		do
			if channels /= Void then
				from
					channels.start
				until
					channels.off
				loop
					channels.item.output_time_measure (a_tm)
					channels.forth
				end
			end
		end

	flush is
			-- Flush this channel.
		do
			if channels /= Void then
				from
					channels.start
				until
					channels.off
				loop
					channels.item.flush
					channels.forth
				end
			end
		end

feature -- Output

	help: STRING is "Output to a collection of channels"
			-- Help on this command

feature {NONE} -- Implementation

	do_open is
			-- Open effectively this channel with 'parameters'.
		local
			l_seq: SEQUENCE [STRING]
		do
			if parameters /= Void then
				from
					l_seq := Global.formatter.split (parameters,channel_separator,channel_separator_escapes)
					l_seq.start
				until
					l_seq.off
				loop
					add_channel (Global.io_handler.manager.checked_output_channel (
						Global.app.program_name,
						Global.formatter.unescaped (l_seq.item,channel_separator_escapes)))
					l_seq.forth
				end
			end
		end

	do_close is
			-- Close effectively this channel.
		do
			if channels /= Void then
				from
					channels.start
				until
					channels.off
				loop
					channels.item.close
					channels.remove
				end
				channels := Void
			end
		end

	channel_separator: STRING is ","
	channel_separator_escapes: STRING is "\"

end -- class ROSE_OUTCHANNEL_BROADCAST
