indexing
	description: "Mail channel: send output as email"
	author: "Zoran Simic"
	date: "$Date: 2008/12/01 $"
	revision: "$Revision: #17 $"

class
	ROSE_OUTCHANNEL_MAIL

inherit
	ROSE_OUTCHANNEL

create
	make

feature {NONE} -- Initialization

	make is
			-- Create a mail channel
		do
			id := "mail"
			create buffer.make
			accepted_verbosity := minimum_verbosity
			create mail.make
		end

feature -- Access

	id: STRING
			-- Channel type id

	buffer: ROSE_OUTCHANNEL_BUFFER
			-- Buffer used to compose email

	recipients: STRING
			-- Recipients of the email to be sent

	subject: STRING
			-- Subject of the email to be sent

feature -- Status report

	is_sensible: BOOLEAN
			-- Is this mail channel sensible?
			-- if True, an unsuccesfull sending is treated as an error

	is_open: BOOLEAN is
			-- Is this channel open ?
		do
			Result := recipients /= Void and then not recipients.is_empty
		end

feature -- Basic operations

	set_recipients (a_recipients: STRING) is
		do
			recipients := a_recipients
		end

	set_subject (a_subject: STRING) is
		do
			subject := a_subject
		end

	set_is_sensible (a_value: BOOLEAN) is
		do
			is_sensible := a_value
		end

	output_message (a_level: INTEGER; a_message: STRING) is
			-- Output 'a_message' with 'a_level' on this channel.
		do
			if a_message /= Void and then a_level >= accepted_verbosity then
				buffer.output_message (a_level,a_message)
			end
		end

	flush is
			-- Flush this channel.
		do
		end

	send is
			-- Send effectively the email message
		local
			l_content: ROSE_MAIL_CONTENT
			rescued, l_string: STRING
		do
			if rescued /= Void then
				error.signal (id,
					"Mail sending failed, headers were:%NTo: $2%NSubject: $3%NDate: $4%NTrace:%N$1%N",
					<<rescued,recipients,subject,create {ROSE_DATE}.make_now>>)
			elseif recipients /= Void and then not recipients.is_empty then
				create l_content.make
				l_content.to.set (recipients)
				l_content.set_subject (subject)
				l_content.set_text (buffer.flush_filtered_to_string (accepted_verbosity))
				l_string := Global.app.report
				if config.count > 0 then
					l_string.append ("%N-- Config:%N")
					if config.path /= Void then
						l_string.append ("-- Path: '")
						l_string.append_string (config.path)
						l_string.append ("'%N")
					end
					l_string.append (config.root.out)
				end
				l_content.add_content_attachment ("application-log.txt",l_string)
				mail.send (l_content)
				if mail.last_error /= Void then
					error.signal (id,"Could not send email to '$1' with subject '$2': $3",<<recipients,subject,mail.last_error>>)
				end
			elseif is_sensible then
				error.signal (id,"Can not send email with subject '$1': no recipients",<<subject>>)
			else
				error.warn ("Sending email with subject '$1' not performed: no recipients",<<subject>>)
			end
		rescue
			if is_sensible and then rescued = Void then
				rescued := last_exception_explanation
				retry
			end
		end

	clear is
			-- Clear mail message body.
		do
			buffer.open (Void)
		end

feature -- Output

	help: STRING is
			-- Help on this command
		do
			Result := "Send buffered output to a mail recipient"
		end

feature {NONE} -- Implementation

	do_open is
			-- Open effectively this channel with 'parameters'.
		local
			s: SEQUENCE [STRING]
		do
			buffer.open (Void)
			if parameters /= Void then
				s := Global.formatter.split (parameters,"%N|",Void)
				s.start
				if not s.off then
					recipients := s.item
					s.remove
				end
				if not s.off then
					subject := s.item
					s.remove
				end
				if not s.off then
					buffer.output_message (v_output,Global.formatter.join (s,"%N"))
				end
				if is_open and then subject = Void then
					subject := ""
				end
			end
			if recipients = Void or else recipients.is_empty then
				error.signal (id,"Can not open mail channel: no recipients provided",Void)
			end
		end

	do_close is
			-- Close effectively this channel (send email eventually).
		do
			send
			buffer.close
			recipients := Void
			subject := Void
		end

	mail: ROSE_MAIL

end -- class ROSE_OUTCHANNEL_MAIL
