note

	description:

		"Indexed collection of colors"

	author: "Zoran Simic"
	last_editor: "$Author: c4 $"
	date: "$Date: 2013/07/08 $"
	revision: "$Revision: #13 $"
	path: "$File: //rose/source/kernel/infrastructure/widgets/rose_palette.e $"

class
	ROSE_PALETTE

inherit
	ROSE_SHARED_GLOBAL_GUI

create

	make, make_default, make_from_representation, make_linear, make_from_array

feature -- Initialization

	make_from_array (a_colors: ARRAY [EV_COLOR]; a_cyclic, a_tolerant: BOOLEAN) is
			-- create using `a_colors' for `colors'.
		require
			a_colors_not_void: a_colors /= Void
			no_void_color: not a_colors.has (Void)
		do
			colors := a_colors
			is_tolerant := a_tolerant
			is_cyclic := a_cyclic
			create background
			create foreground
			set_background (Gui.rep.colors.White)
		ensure
			colors_aliased: colors = a_colors
			is_tolerant_aliased: is_tolerant = a_tolerant
			is_cyclic_aliased: is_cyclic = a_cyclic
		end

	initialize (a_count: INTEGER) is
			-- Initialize palette for 'a_count' colors.
		local
			i: INTEGER
		do
			create colors.make (1,a_count.max (0))
			from
				i := colors.lower
			until
				i > colors.upper
			loop
				colors.put (create {EV_COLOR},i)
				i := i + 1
			end
		end

	make (a_count: INTEGER) is
			-- Create a color palette with 'a_count' entries.
		do
			initialize (a_count)
			is_tolerant := True
			is_cyclic := True
			create background
			create foreground
			set_background (Gui.rep.colors.White)
		end

	make_default is
			-- Create default AXA Rosenberg palette.
		local
			l_schemes: COLOR_BREWER
		do
			--	changed for testing COLOR_BREWER: make_from_representation (default_palette)
			create l_schemes
			make_from_array (l_schemes.paired_10, True, True)
		end

	make_from_representation (a_representation: STRING) is
			-- Create palette from ';' separated RGB values in 'a_representation'.
			-- example: "255,0,0;0,255,0;0,0,0"
			-- The first 2 components are background and foreground
		local
			l_array: ARRAY [STRING]
			i: INTEGER
		do
			l_array := Global.formatter.split_array (a_representation,k_semicolon,Void,False)
			if l_array = Void or l_array.count = 0 then
				make (0)
			else
				make (l_array.count - 2)
				set_background (Gui.rep.colors.rgbs (l_array.item (l_array.lower)))
				if l_array.count > 1 then
					set_foreground (Gui.rep.colors.rgbs (l_array.item (l_array.lower+1)))
				end
				if l_array.count > 2 then
					from
						i := l_array.lower + 2
					until
						i > l_array.upper
					loop
						set_color (i-l_array.lower-1,Gui.rep.colors.rgbs (l_array.item (i)))
						i := i + 1
					end
				end
			end
		end

	make_linear (a_count: INTEGER) is
		local
			i: INTEGER
			l_color: EV_COLOR
		do
			make (a_count)
			if count = 1 then
				set_color (1,Gui.rep.colors.Blue)
			elseif count = 2 then
				set_color (1,Gui.rep.colors.Blue)
				set_color (2,Gui.rep.colors.Red)
			elseif count > 2 then
				set_color (1,Gui.rep.colors.Blue)
				set_color (count,Gui.rep.colors.Red)
				if count \\ 2 = 0 then
					i := (count // 2)
					set_color (i+1,Gui.rep.colors.White)
					apply_gradient (1,i + 1)
					l_color := color (i).twin
					set_color (i,Gui.rep.colors.White)
					apply_gradient (i,count)
					set_color (i,l_color)
				else
					i := (count // 2) + 1
					set_color (i,Gui.rep.colors.White)
					apply_gradient (1,i)
					apply_gradient (i,count)
				end
			end
		end

feature -- Access

	background, foreground: EV_COLOR
			-- Background/forgeround colors

	colors: ARRAY [EV_COLOR]
			-- Colors composing this palette

	is_cyclic: BOOLEAN
			-- Should the colors be cyclic? (ie, if count=10, color 11 = color 1)
			-- Default: True

	is_tolerant: BOOLEAN
			-- Is this palette tolerant? (ie, asking for an undefined color returns 'foreground' instead of Void)
			-- Default: True

	color (a_index: INTEGER): EV_COLOR is
			-- Color associated to 'a_index', 'background' is returned for '0'
		local
			i: INTEGER
		do
			i := a_index.max (0)
			if i = 0 then
				Result := foreground
			else
				if is_cyclic and then i > count then
					i := i \\ count
					if i = 0 then
						i := count
					end
				end
				if colors.valid_index (i) then
					Result := colors.item (i)
				end
			end
			if is_tolerant and then Result = Void then
				Result := background
			end
		end

	count: INTEGER is
			-- Number of colors in this palette
		do
			Result := colors.count
		end

	representation: STRING is
			-- example: "255,0,0;0,255,0;0,0,0"
			-- The first 2 components are background and foreground
		local
			i: INTEGER
		do
			Result := rgbs (background) + k_semicolon + rgbs (foreground) + k_semicolon
			if colors /= Void then
				from
					i := colors.lower
				until
					i > colors.upper
				loop
					Result.append (rgbs (colors.item (i)) + k_semicolon)
					i := i + 1
				end
			end
			Result.remove (Result.count)
		end

feature -- Basic operations

	set_background (a_color: EV_COLOR) is
		local
			l_color: EV_COLOR
		do
			if a_color = Void then
				l_color := Gui.rep.colors.White
			else
				l_color := a_color
			end
			copy_color (background,l_color)
		end

	set_foreground (a_color: EV_COLOR) is
		do
			copy_color (foreground,a_color)
		end

	set_color (a_index: INTEGER; a_color: EV_COLOR) is
			-- Set 'a_color' associated to 'a_index'.
		do
			copy_color (color (a_index),a_color)
		end

	set_colors (a_colors: ARRAY [EV_COLOR]) is
			-- Assign colors from 'a_colors' to this palette.
		local
			i: INTEGER
		do
			if a_colors /= Void then
				initialize (a_colors.count)
				from
					i := a_colors.lower
				until
					i > a_colors.upper
				loop
					set_color (i - a_colors.lower + 1,a_colors.item (i))
					i := i + 1
				end
			end
		end

	copy_from (a_other: like Current) is
			-- Copy 'a_other' palette into this one.
		do
			if a_other /= Void then
				is_cyclic := a_other.is_cyclic
				is_tolerant := a_other.is_tolerant
				set_background (a_other.background)
				set_foreground (a_other.foreground)
				set_colors (a_other.colors)
			end
		end

	resize (a_count: INTEGER) is
			-- Resize this palette to contain exactly 'a_count' colors.
		local
			l_old_colors: ARRAY [EV_COLOR]
			i: INTEGER
		do
			l_old_colors := colors
			initialize (a_count)
			from
				i := l_old_colors.lower
			until
				i > count or else i > l_old_colors.upper
			loop
				set_color (i,l_old_colors.item (i))
				i := i + 1
			end
		end

	set_cyclic (a_cyclic: BOOLEAN) is
			-- Set 'cyclic'.
		do
			is_cyclic := a_cyclic
		end

	set_tolerant (a_tolerant: BOOLEAN) is
			-- Set 'tolerant'.
		do
			is_tolerant := a_tolerant
		end

	apply_gradient (a_start, a_end: INTEGER) is
			-- Apply a gradient for colors between indices 'a_start' and 'a_end'.
		local
			l_source, l_target, l_color: EV_COLOR
			i, l_start, l_end, l_count, r1, g1, b1, r2, g2, b2: INTEGER
		do
			l_start := a_start.min (a_end)
			l_end := a_start.max (a_end)
			l_count := l_end - l_start
			if l_count >= 2 and then l_start >= 1 and then l_end <= count then
				l_source := colors.item (l_start)
				l_target := colors.item (l_end)
				r1 := l_source.red_8_bit
				g1 := l_source.green_8_bit
				b1 := l_source.blue_8_bit
				r2 := l_target.red_8_bit
				g2 := l_target.green_8_bit
				b2 := l_target.blue_8_bit
				from
					i := 1
				until
					i >= l_count
				loop
					l_color := colors.item (l_start + i)
					l_color.set_red_with_8_bit (gradient_component (r1,r2,i,l_count))
					l_color.set_green_with_8_bit (gradient_component (g1,g2,i,l_count))
					l_color.set_blue_with_8_bit (gradient_component (b1,b2,i,l_count))
					i := i + 1
				end
			end
		end

	invert is
			-- Invert all colors (last becomes first etc.)
		local
			i, j: INTEGER
			l_color: EV_COLOR
		do
			from
				i := 1
			until
				i > colors.count
			loop
				j := colors.count - i + 1
				if i /= j then
					l_color := colors.item (i)
					colors.put (colors.item (j),i)
					colors.put (l_color,j)
				end
				i := i + 1
			end
		end

feature {NONE} -- Implementation

	gradient_component (a_r1, a_r2, i, a_count: INTEGER): INTEGER is
		local
			v: DOUBLE
		do
			v := a_r1 + i * (a_r2 - a_r1) / a_count
			Result := v.rounded.max (0).min (255)
		end

	copy_color (a_destination, a_source: EV_COLOR) is
			-- Change `a_destination' to be the same color as `a_source', or black if `Void'.
		local
			l_color: EV_COLOR
		do
			if a_destination /= Void then
				if a_source = Void then
					l_color := Gui.rep.colors.Black
				else
					l_color := a_source
				end
				a_destination.set_rgb_with_24_bit (l_color.rgb_24_bit)
			end
		end

	rgbs (a_color: EV_COLOR): STRING is
		do
			if a_color = Void then
				Result := "0,0,0"
			else
				Result := a_color.red_8_bit.out + k_comma + a_color.green_8_bit.out + k_comma + a_color.blue_8_bit.out
			end
		end

	k_comma: STRING is ","
	k_semicolon: STRING is ";"
	default_palette: STRING is "255,255,255;0,0,0;128,0,0;51,102,153;102,204,204;185,105,108;34,139,34;153,153,153;6,118,140;0,52,102;204,153,204;153,0,153;"

invariant

	colors_not_void: colors /= Void
	no_void_color: not colors.has (Void)

end
