note
	description: "Pilot of a command-driven and configured application"
	notice: "[
		This class is part of a framework; all root classes should instantiate one object of this class.
		The entire system should have exactly one instance of a pilot, which is designed to be a singleton.
		Use one of the descendants to include some extra services into your application.
		The recommended usage is:
			l_pilot: ROSE_PILOT		-- Or whatever other implementor you prefer
			create l_pilot
			l_pilot.add_command (create {MY_COMMAND})
			l_pilot.run
		The necessary steps are actually:
			.add_command ()					-- For each of the application's command
			.set_default_command_line ()	-- Default is 'console'
			.run_application
			.exit
]"
	author: "Zoran Simic"
	last_editor: "$Author: dz $"
	date: "$Date: 2014/12/18 $"
	revision: "$Revision: #35 $"
	path: "$File: //rose/source/kernel/infrastructure/framework/application/rose_pilot_base.e $"

class ROSE_PILOT_BASE

inherit
	ROSE_SINGLETON
		redefine
			default_create
		end
	ROSE_SHARED_GLOBAL
		redefine
			default_create
		end

create
	default_create

feature {NONE} -- Initialization

	default_create is
			-- Default creation procedure.
		do
			is_configuration_required := True
			check_instance_unicity
			prepare
			register
		end

	make (a_project, a_name, a_version, a_description: STRING) is
			-- Initialize the pilot and read app's configuration.
			-- 'a_default_config' is used if no config file found (use Void to require cofig file).
		obsolete
			"Do not use ROSE_PILOT.make anymore, use 'default_create'"
		do
			default_create
		end

feature {NONE} -- Initialization

	prepare is
			-- Prepare this pilot.
		do
			global.io_handler.set_output (create {ROSE_OUTCHANNEL_STANDARD}.make)
		end

	register is
			-- Register this pilot's specifics.
		do
			global.dependency_manager.register (create {ROSE_SYSTEM_DEPENDENCY_MINIMAL})
		end

feature -- Access

	is_configuration_required: BOOLEAN assign set_configuration_required
			-- Is configured required for this application?
			-- Set to True to have the application fail on launcher if config is not available
			-- Leave at False (default) otherwise (the config will be considered optional then)

feature -- Basic operations

	set_configuration_required (a_configuration_required: BOOLEAN) is
			-- Set 'is_configuration_required'.
		do
			is_configuration_required := a_configuration_required
		end

	set_default_command_line (a_value: STRING) is
			-- Set default command line.
		require
			valid_value: a_value /= Void
		do
			global.app.set_default_command_line (a_value)
		end

	set_termination_code (a_code: INTEGER) is
			-- Set application's 'termination_code'.
			-- This code is returned to shell on exit, use with care.
		do
			global.set_termination_code (a_code)
		end

	add_command (a_command: ROSE_COMMAND) is
			-- Add 'a_command' to available commands in this pilot
		require
			valid_command: a_command /= Void
		do
			if global.app.default_command = Void then
				global.app.set_default_command (a_command.id)
			end
			global.commands.add_command (global.commands.application_category, a_command)
		end

	add_simple_command (a_id, a_signature, a_help: STRING; a_agent: PROCEDURE [ANY, TUPLE [ROSE_COMMAND_PARAMETERS]]) is
			-- Add simple command with 'a_id', 'a_signature', 'a_help'
			-- and 'a_agent' to available commands in this pilot
		require
			valid_id: a_id /= Void and then not a_id.is_empty and then not a_id.has (' ')
			valid_signature: a_signature /= Void
			valid_help: a_help /= Void
			valid_agent: a_agent /= Void
		do
			add_command (create {ROSE_SIMPLE_COMMAND}.make (a_id, a_signature, a_help, a_agent))
		end

	set_alias (a_id, a_alias: STRING) is
			-- Alias 'a_id' to mean 'a_alias'.
		do
			global.commands.set_alias (a_id, a_alias)
		end

	add_implicit_command (a_command: STRING) is
			-- Add 'a_command' as an implicit command
			-- (an implicit command is always executed when application starts, even if not provided on command line)
		require
			valid_command: a_command /= Void
		do
			global.app.add_implicit_command (a_command)
		end

	add_last_resort_channel (a_path: STRING) is
			-- Add 'a_path' as last-resort alternative to flush errors to.
			-- Last resort paths are stacked, tried in a FIFO fashion.
			--| ex: to flush to a file, use "file:path/to/log/file.log"
			--| ex: to send an email, use "mail:recipient1@domain,recipient2,...%Nsubject" in this array
		require
			valid_path: a_path /= Void
		do
			global.add_last_resort_channel (a_path)
		end

	run is
			-- Run the application and exit properly (with according status code).
		do
			run_application
			exit
		end

	run_application is
			-- Run the application, this defines every application's generic behaviour.
		local
			rescued: STRING
			l_command: RCMD_MACRO
		do
			if error.signaled then
				-- Do nothing
			elseif rescued /= Void then
					--| Turn uncatched exception into a signaled error
				error.signal (global.app.program_name, "$1 crashed.%NDetails:%N$2", <<global.app.name_version, rescued>>)
				global.set_termination_code (-2)
			end
			if not error.signaled then
				output_begin_message
				global.add_cleanup_agent (agent output_end_message)
				l_command := initial_command
				if l_command = Void or else l_command.count = 0 then
						--| Error while interpreting the arguments or no arguments and no defaults
					error.flush_and_acknowledge
					global.commands.execute ("usage -as_error")
				else
					add_implicit_commands
					global.complete_command_with_implicit (l_command)
				end
			end
			if not error.signaled then
				run_pilot (l_command)
			end
			if global.termination_code = 0 and then error.signaled then
				global.set_termination_code (1)
			end
		rescue
			if rescued = Void then
					--| Catch uncatched exceptions, do not recurse...
				rescued := last_exception_explanation
				retry
			end
		end

	exit is
			-- Terminate pilot activity (exit/die) with 'termination_code' returned to the shell.
			-- Application dies if 'termination_code' is not zero.
		do
			global.terminate (global.termination_code)
		end

feature {NONE} -- Implementation

	add_implicit_commands is
			-- Add implicit commands
		do
			if is_configuration_required then
				add_implicit_command ("config -load -default")
			else
				add_implicit_command ("config -load -default -silent")
			end
				-- Implicit commands
			add_implicit_command ("system -start")
		end

	run_pilot (a_command: ROSE_COMMAND) is
			-- Run the pilot with 'a_command'.
			-- Do not redefine this, unless you have a very specific need.
		do
			if a_command /= Void then
				a_command.execute_prepared
			end
		end

	initial_command: RCMD_MACRO is
			-- Commands to be executed on startup: parsed command line arguments
			-- Do not redefine this, unless you have a very specific need.
		require
			no_error: not error.signaled
		local
			l_args: STRING
		do
			l_args := global.app.program_arguments
			if global.app.default_command /= Void and then l_args /= Void and then not l_args.is_empty then
				if not global.commands.has (global.formatter.split_on_first (l_args, global.convention.argument_separators, Void).first) then
					l_args := global.app.default_command + " " + l_args
				end
			end
			Result := global.commands.macro_command (l_args)
			if not error.signaled and then (Result = Void or else Result.count = 0) then
					--| No arguments: consider default command line
				l_args := global.app.default_command_line
				if l_args /= Void and then not l_args.is_empty then
					Result := global.commands.macro_command (l_args)
				end
			end
		ensure
			result_or_error: Result /= Void or else error.signaled
		end

	output_begin_message
			-- output "begin execution" trace message, chock full of useful information about the execution and executable
		local
			l_userid: STRING
			l_datestr, l_revs: STRING
		do
			if attached global.app.user as l_user then
				l_userid := l_user.login_id
			end
			if l_userid = Void then
				l_userid := "<unknown>"
			end
			output ("%N$1", << global.formatter.separator_line >>)
			trace ("begin execution: $1  user_id: $2%N  path: $3%N  arguments: $4%N",
					<< global.app.start_time, l_userid, global.app.program_path, global.app.program_arguments >>)
			if global.app /= Void and then attached global.app.info as l_info then
				if attached l_info.build_date as l_bldt then
					l_datestr := l_bldt.formatted_out ("yyyy/[0]mm/[0]dd hh:[0]mi:[0]ss") + " (UT)"
				else
					l_datestr := "(unknown)"
				end
				create l_revs.make_empty
				if l_info.revision /= 0 then
					l_revs.append (global.formatted (" Revision: $1", << l_info.revision >>))
				end
				l_revs.append (global.formatted ("  Compatible libraries: $1", << l_info.build_compatible >>))
				if attached (create {PROFILING_SETTING}.make) as l_prof then
					l_revs.append (global.formatted ("  Profilable: $1", << l_prof.is_profiler_enabled >>))
				end
				if l_info.overrides_count > 0 then
					l_revs.append (global.formatted ("%N   $1: %N%T$2", << Global.formatter.plural (l_info.overrides_count, "override"), global.formatter.join_array (l_info.overrides, ",%N%T") >>))
				end
				trace ("  build date: $1  EStudio: $2  C: $3  Platform: $4 %N  SCM root: $5 $6 %N",
						<< l_datestr, safe_out (l_info.build_estudio), safe_out (l_info.build_c), safe_out (l_info.build_platform),
							safe_out (l_info.scm_root), l_revs >>)
			end
			output ("$1%N", << global.formatter.separator_line >>)
		end

	output_end_message
			-- output "end execution" trace message
		local
			l_end: ROSE_DATE
		do
			error.flush
			create l_end.make_now
			output ("%N$1", << global.formatter.separator_line >>)
			trace ("end execution: $1 (elapsed: $2)%N  termination_code: $3%N",
					<< l_end, global.app.formatter.elapsed_time_mmssfff (global.app.start_time, l_end), global.termination_code.out >>)
			output ("$1%N", << global.formatter.separator_line >>)
		end

	-- move to ROSE_FORMATTING_SERVICE
	safe_out (a_ny: detachable ANY): attached STRING
			-- string representation of `a_ny' if attached, else "<Void>"
		do
			if a_ny /= Void then
				Result := a_ny.out
			else
				Result := "<Void>"
			end
		ensure
			not_void: Result /= Void
			void_argument_implies_result_equal_manifest_void_string: a_ny = Void implies Result.is_equal ("<Void>")
		end

end -- class ROSE_PILOT_BASE
