indexing

	description: "Student's T distribution for a given degree of freedom"

	keywords: "statistic student t distribution probability"
	author: "Colin Adams"
	last_editor: "$Author: c4 $"
	date: "$Date: 2010/02/18 $"
	revision: "$Revision: #13 $"

class ROSE_PR_DISTRIBUTION_STUDENTS_T

inherit

	ROSE_PR_DISTRIBUTION_CONTINUOUS_UNIVARIATE
		redefine
			name,
			estimated_tolerance
		end

create

	make

feature {NONE} -- Initialization

	make (a_df: INTEGER) is
			-- Initialize `Current' with `a_df' degrees of freedon.
		require
			a_df_positive: a_df > 0
		do
			degree_of_freedom := a_df
			epsilon := 3.0e-7
			maximum_iterations := 100
		ensure
			degree_of_freedom_set: degree_of_freedom = a_df
			default_epsilon: epsilon = 3.0e-7
			default_maximum_iterations: maximum_iterations = 100
		end

feature -- Access

	name : STRING is
			-- Name of the distribution
		once
			Result := "Student's t"
		end

feature -- Distribution Parameters

	degree_of_freedom: INTEGER
			-- Degrees of freedom used in calculating probabilities

	set_degrees_of_freedom (a_degrees: INTEGER) is
			-- Set degrees of freedom
		require
			positive_degrees: a_degrees > 0
		do
			degree_of_freedom := a_degrees
		end

feature -- Basic Operations

	compute_probability_density_function is
			-- may not be needed
		do
			fail("compute_probability_function", "Need unclear.", Void)
		end


	cumulative_distribution_function (a_x: DOUBLE): DOUBLE is
			-- Probability that `random_variable' <= `a_x'
		local
			l_x: DOUBLE
			l_beta: DOUBLE
		do
			l_x := degree_of_freedom.to_double / (degree_of_freedom.to_double + (a_x * a_x))
			l_beta := 0.5 * incomplete_beta (l_x, degree_of_freedom.to_double / 2.0, 0.5, epsilon, maximum_iterations)
			if a_x <= 0.0 then
				Result := l_beta
			else
				Result := 1.0 - l_beta
			end
		end

	entropy: DOUBLE is
			-- NYI: requires digamma function
		do
			fail("entropy", not_yet_implemented, Void)
		end

	is_symmetric: BOOLEAN is
			-- Student's t distribution is symmetric
		do
			Result := True
		end

	is_kurtosis_computable: BOOLEAN is
			-- is kurtosis computable for this distribution?
		do
			Result := degree_of_freedom > 4
		end

	is_mean_computable: BOOLEAN is
			-- is mean computable for this distribution?
		do
			Result := degree_of_freedom > 1
		end

	is_mode_computable: BOOLEAN is
			-- is mode computable for this distribution?
		do
			Result := degree_of_freedom > 1
		end

	is_skewness_computable: BOOLEAN is
			-- is skewness computable for this distribution?
		do
			Result := degree_of_freedom > 3
		end

	is_variance_computable: BOOLEAN is
			-- is variance computable for this distribution?
		do
			Result := degree_of_freedom > 2
		end

	kurtosis: DOUBLE is
			-- Kurtosis of Student's t distribution
		do
			Result := 6 / (degree_of_freedom - 4)
		end

	lorenz_curve (F: DOUBLE): DOUBLE is
			-- not yet computed
		do
			fail("lorenz_curve", "Not yet computed.", Void)
		end

	memoryless: BOOLEAN is
			-- is students' t distribution memoryless?
		do
			fail("Memoryless", "False unclear.", Void)
		end

	mean: DOUBLE is
			-- mean of the student's t distibution
		do
			Result := 0
		end

	mode: DOUBLE is
			-- mode of the student's t distribution
		do
			Result := 0
		end

	-- probability_density_function (x: DOUBLE): DOUBLE is
	probability_density_function (x: like random_variable): DOUBLE is
			-- pdf of students' t distribution
		do
			fail("Pdf", not_yet_implemented, Void)
		end

	skewness: DOUBLE is
			-- skewness of the student's t distribution
		do
			Result := 0
		end

	estimated_tolerance: INTEGER is
			-- estimation of "best" tolerance for probability computations taking local epsilon accuracy into account
			-- e.g., if \epsilon = 10^{-2}, then \approx tolerance is 2
		local
			l_tolerance: INTEGER
		do
			l_tolerance := - log10(epsilon).floor
			Result := tolerance_magnitude.min (l_tolerance)
		end

	variance: DOUBLE is
			-- variance of the student's t distribution
		do
			Result := degree_of_freedom / (degree_of_freedom - 2)
		end


feature -- Measurement

	epsilon: DOUBLE
			-- Relative accuracy required for convergence

	maximum_iterations: INTEGER
			-- Maximum number of iterations for convergence

feature -- Setting


	set_epsilon (a_epsilon: DOUBLE) is
			-- Set `epsilon' to `a_epsilon'.
		require
			a_epsilon_positive: a_epsilon > 0.0
		do
			epsilon := a_epsilon
		ensure
			epsilon_set: epsilon = a_epsilon
		end

	set_maximum_iterations (a_max: INTEGER) is
			-- Set `maximum_iterations' to `a_max'.
		require
			a_max_positive: a_max > 0
		do
			maximum_iterations := a_max
		ensure
			maximum_iterations_set: maximum_iterations = a_max
		end

feature {NONE} -- Implementation


invariant

	degree_of_freedom_positive: degree_of_freedom > 0
	epsilon_positive: epsilon > 0.0
	maximum_iterations_positive: maximum_iterations > 0

end
