indexing

	description: "Framework for probability distributions"

	keywords: "probability distribution univariate"
	author: "Yves Chauvin"
	last_editor: "$Author: e2 $"
	date: "$Date: 2010/04/08 $"
	revision: "$Revision: #19 $"
	path: "$File: //rose/source/kernel/infrastructure/math/statistics/rose_pr_distribution_univariate.e $"

deferred class	ROSE_PR_DISTRIBUTION_UNIVARIATE

inherit

	ROSE_PR_DISTRIBUTION

	APPROXIMATE_EQUALITY
		export {NONE} all end

feature -- Basic operations

	coefficient_of_variation: DOUBLE is
			-- coefficient of variation
		require
			non_zero_mean: mean /= 0
		do
			Result := standard_deviation / mean
		end

	cumulative_distribution_function (a_x: DOUBLE): DOUBLE is
			-- Probability that `random_variable' <= `a_x'
		deferred
		ensure
			result_is_probability: is_probability (Result)
		end

	quantile (a_quantile: DOUBLE): DOUBLE is
			-- Inverse of `cumulative_distribution_function'
		require
			a_quantile_large_enough: a_quantile >= 0.0
			a_quantile_small_enough: a_quantile <= 1.0
		do
			(create {NOT_YET_IMPLEMENTED_EXCEPTION}.make (generator, "quantile")).raise
		ensure
			approximate_inverse: (cumulative_distribution_function (Result) - a_quantile).abs < 1.0e-6
		end

	is_symmetric: BOOLEAN is
			-- is the distribution symmetric?
		deferred
		end

	is_kurtosis_computable: BOOLEAN is
			-- is kurtosis computable for this distribution?
		deferred
		end

	is_mode_computable: BOOLEAN is
			-- is mode computable for this distribution?
		deferred
		end

	is_skewness_computable: BOOLEAN is
			-- is skewness computable for this distribution?
		deferred
		end

	is_mean_computable: BOOLEAN is
			-- is mean computable for this distribution?
		deferred
		end

	is_variance_computable: BOOLEAN is
		-- is variance computable for this distribution?
		deferred
		end

	kurtosis: DOUBLE is
			-- kurtosis
		require
			kurtosis_is_computable: is_kurtosis_computable = True
		deferred
		end

	lower_tail_probability (a_x: DOUBLE): DOUBLE is
			-- lower tail probability
		do
			Result := cumulative_distribution_function (a_x)
		ensure
			result_is_probability: is_probability (Result)
		end

	survival_function (a_x: DOUBLE): DOUBLE is
			-- definition of survival function
		do
			Result := 1.0 - cumulative_distribution_function (a_x)
		ensure
			result_is_probability: is_probability (Result)
		end

	two_tail_probability (a_x: DOUBLE): DOUBLE is
			-- two tail probability
		require -- otherwise, not clear for now
			mean_is_zero: mean = 0
			is_symmetric: is_symmetric = True
		do
			if a_x = 0.0 then
				Result := 1.0
			elseif a_x < 0.0 then
				Result := 2.0 * cumulative_distribution_function (a_x)
			else
				Result := 2.0 * (1.0 - cumulative_distribution_function (a_x))
			end
		ensure
			result_is_probability: is_probability (Result)
			symmetry_implication: is_symmetric implies approx_equal_with_magnitude (1.0 - Result, 1.0 - two_tail_probability(-a_x), estimated_tolerance)
		end

	upper_tail_probability (a_x: DOUBLE): DOUBLE is
			-- upper tail probability
		do
			Result := 1.0 - cumulative_distribution_function (a_x)
		ensure
			result_is_probability: is_probability (Result)
		end

	interval_probability (a_low, a_high: DOUBLE): DOUBLE is
			-- probability of interval from a_low to a_high
		require
			interval_order: a_low < a_high or a_low = a_high
		do
			Result := cumulative_distribution_function (a_high) - cumulative_distribution_function (a_low)
		ensure
			result_is_probability: is_probability (Result)
		end

	mean: DOUBLE is
			-- mean
		require
			mean_is_computable: is_mean_computable = True
		deferred
		end

	mode: DOUBLE is
			-- mode
		require
			mode_is_computable: is_mode_computable = True
		deferred
		end

	less_than (a_low, a_high: like random_variable): BOOLEAN is
			-- proper order
		deferred
		end

	skewness: DOUBLE is
			-- skewness
		require
			skewness_is_computable: is_skewness_computable = True
		deferred
		end

	standard_deviation: DOUBLE is
			-- standard deviation
		require
			variance_is_computable: is_variance_computable implies variance >= 0
		do
			Result := sqrt(variance)
		ensure
			positive_standard_deviation: Result >= 0
		end

	variance: DOUBLE is
			-- variance
		require
			variance_is_computable: is_variance_computable = True
		deferred
		ensure
			positive_variance: Result >= 0
		end

feature -- Vectors

	two_tail_probabilities (a_vector: ABSTRACT_READABLE_VECTOR): VECTOR is
			-- Vector map using `two_tail_probability'
		do
			create Result.make (a_vector.count)
			Result.set_names (a_vector.names)
			a_vector.do_all_with_index (agent set_two_tail_probability (Result, ?, ?))
		end

feature {NONE} -- Vectors implementation

	set_two_tail_probability (a_target: VECTOR; a_value: DOUBLE; a_index: INTEGER) is
			-- Set `a_target [a_index]' to p-value corresponding to `a_target' from `a_student'.
		require
			a_target_not_void: a_target /= Void
			a_target_index: a_target.valid_index (a_index)
		do
			a_target [a_index] := two_tail_probability (a_value)
		end

feature {NONE} -- Implementation

	random_variable: NUMERIC
		-- random variable underlying distribution

	is_negative (a_x: like random_variable): BOOLEAN is
			-- rv type is negative
		deferred
		end

	is_positive (a_x: like random_variable): BOOLEAN is
			-- rv type is positive
		deferred
		end



end -- class ROSE_PR_DISTRIBUTION_UNIVARIATE

