indexing
	description: "object for keeping track of cpu seconds and elapsed seconds"
	author: "Mark Howard"
	last_editor: "$Author: tg $"
	changelist: "$Change: 150481 $"
	revision: "$Revision: #40 $"
	keywords: "process,timer,time,seconds,cpu,profile"
	archive: "$File: //rose/source/kernel/infrastructure/system/rose_process_timer.e $"

class
	ROSE_PROCESS_TIMER

inherit
	RP_SERIALIZABLE
	ROSE_SHARED_GLOBAL
		redefine
			default_create
		end

	HASHABLE
		redefine
			default_create
		end
	
create
	make, default_create

feature {NONE} -- Initialization

	default_create is
			-- Default creation procedure
		do
			make
		end

	make is
			-- Create and start a timer
		do
			mutex.lock
			counter.put( counter.item + 1)
			hash_code := generated_hash_code
			name := hash_code.out
			ordinal := counter.item
			mutex.unlock
			resume
		end

feature -- Access

	start_cpu: INTEGER
			-- Ticks when timer started

	cpu: DOUBLE
			-- Number of CPU ticks consumed between calls to 'resume' and 'interrupt'

	elapsed: DOUBLE
			-- Real time elapsed between calls to 'resume' and 'interrupt'

	child_elapsed: DOUBLE
			-- Real time elapsed between calls to 'resume' and 'interrupt' that was taken by "child" timers

	self_elapsed: DOUBLE is
			-- Real time elapsed between calls to 'resume' and 'interrupt' excluding time taken by "child" timers
		do
			Result := elapsed - child_elapsed
		end

	resolution: INTEGER is
			-- Number of ticks in one second
		once
			Result := clocks_per_second
		end

	formatted_cpu_double: DOUBLE is
			-- Number of CPU seconds consumed
		do
			Result := double_format.formatted (cpu_seconds).to_double
		end

	formatted_elapsed_double: DOUBLE is
			-- Number of real time seconds
		do
			Result := double_format.formatted (elapsed).to_double
		end

	formatted_cpu: STRING is
			-- Textual representation of 'cpu_seconds'
		do
			Result := " cpu: " + double_format.formatted (cpu_seconds) + "s"
		end

	formatted_elapsed: STRING is
			-- Textual representation of 'elapsed'
		do
			Result := " elapsed: " + double_format.formatted (elapsed) + "s"
		end

	cpu_seconds: DOUBLE is
			-- Number of CPU seconds consumed between calls to 'resume' and 'interrupt'
		do
			Result := cpu / resolution
		end

	start_elapsed: DATE_TIME
			-- Timestamp when 'resume' called

	end_elapsed: DATE_TIME
			-- Timestamp when 'interrupt' called

	duration: DATE_TIME_DURATION is
			-- Duration between 'start_elapsed' and 'end_elapsed'
		do
			Result := end_elapsed.relative_duration (start_elapsed)
		end

	display_string (a_string: STRING): STRING is
			-- Textual representation of this timer
		local
			l_string: STRING
			l_was_active: BOOLEAN
		do
			l_was_active := is_active

			if l_was_active then
				interrupt
			end
			if a_string /= Void then
				l_string := a_string
			else
				l_string := ""
			end
			Result := l_string + formatted_cpu + ", " + formatted_elapsed + "%N"
			if l_was_active then
				resume
			end
		ensure
			activity: is_active = old is_active
			not_void: Result /= Void
		end

	is_active: BOOLEAN
			-- Is this timer currently active?

	hash_code: INTEGER
		-- <Precursor>

	name: STRING
			-- the name of this timer

	ordinal: INTEGER
			-- this timer's ordinal position in the number of timers that have been instantiated
	
feature -- Basic operations

	
	set_name ( a_name: STRING ) is
			-- set the name of this timer to be `a_name'
		do
			name := a_name
		end
	set_child_elapsed( a_child_elapsed: DOUBLE ) is
			-- set `child_elapsed' to `a_child_elapsed'
		do
			child_elapsed := a_child_elapsed
		end
	
	interrupt is
			-- Interrupt time measurment
		do
			if is_active then
				is_active := False
				cpu := cpu + (clock - start_cpu)
				create end_elapsed.make_now
				elapsed := elapsed + duration.fine_seconds_count
			end
		ensure
			inactive: not is_active
		end

	resume is
			-- Resume time measurment
		do
			if not is_active then
				is_active := True
				start_cpu := clock
				create start_elapsed.make_now
			end
		ensure
			active: is_active
		end

	reset is
			-- Reset this timer
		do
			cpu := 0
			elapsed := 0
			child_elapsed := 0
		end


	display (a_string: STRING) is
			-- Display elapsed time with 'a_string' appended
		do
			output (display_string (a_string), Void)
		ensure
			activity: is_active = old is_active
		end

	display_with_timestamp (a_string: STRING) is
			-- Display elapsed time with 'a_string' appended
		do
			output (global.interpreted ("($now) " + display_string (a_string)), Void)
		ensure
			activity: is_active = old is_active
		end

	display_all (a_array: ARRAY [ANY]) is
			-- Display elapsed time with all items in 'a_array' appended
		local
			i: INTEGER
			l_was_active: BOOLEAN
		do
			l_was_active := is_active
			if l_was_active then
				interrupt
			end
			if a_array /= Void then
				from
					i := a_array.lower
				until
					i > a_array.upper
				loop
					if a_array.item (i) = Void then
						output ("<Void>", Void)
					else
						output (a_array.item (i).out, Void)
					end
					i := i + 1
				end
			end
			output (formatted_cpu + " real: " + double_format.formatted (duration.fine_second) + "s", Void)
			if l_was_active then
				resume
			end
		ensure
			activity: is_active = old is_active
		end

	display_completion_percent (a_message: STRING; a_interval, a_index, a_count: INTEGER) is
			-- Display percent complete at every a_interval percent of the total count.
		local
			l_percent: DOUBLE
		do
			if (a_interval * a_count) // 100 > 0 and then a_index \\ ((a_interval * a_count) // 100) = 0 then
				l_percent := a_index / a_count * 100
				display (a_message + " " + l_percent.rounded.out + "%% completed%T")
			end
		end

	display_time_to_finish (a_processing_unit_name: STRING; a_index, a_count: INTEGER) is
			-- Display time elapsed wiht 'a_processing_unit_name' and how many iteration left from 'a_index' to 'a_count'
		local
			l_unit_time: DOUBLE
			l_time_left: TIME_DURATION
			l_was_active: BOOLEAN
		do
			if a_index < a_count then
				l_was_active := is_active
				if l_was_active then
					interrupt
				end
				l_unit_time := elapsed / a_index
				create l_time_left.make_by_fine_seconds (l_unit_time * (a_count - a_index))
				output (a_index.out + " " + a_processing_unit_name + " processed, " + double_format.formatted (l_unit_time) + " secs/unit; ", Void)
				output ((a_count - a_index).out + " " + a_processing_unit_name + " left, expected to finish in " + duration_string (l_time_left) + ".", Void)
				if l_was_active then
					resume
				end
			end
		ensure
			activity: is_active = old is_active
		end

feature {NONE} -- Implementation

	double_format: FORMAT_DOUBLE is
			-- Format used to display DOUBLEs
		once
			create Result.make (3, 3)
		end

	clocks_per_second: INTEGER is
			-- Number of ticks per second
		external
			"C [macro <time.h>]"
		alias
			"CLOCKS_PER_SEC"
		end

	clock: INTEGER is
			-- Number of ticks since process started
		external
			"C"
		alias
			"rose_clock"
		end

	duration_string (a_duration: TIME_DURATION): STRING is
			-- Textual representation of 'a_duration'
		do
			if a_duration /= Void then
				Result := a_duration.hour.out + ":" + a_duration.minute.out +  ":" + a_duration.second.out
			else
				Result := ""
			end
		end

	counter: CELL[INTEGER] is
			-- helper for simple hash code generator
		once
			create Result.put(0)
		end

	generated_hash_code: INTEGER is
			-- generated hash code from the counter
		do
			Result := counter.item // 1000
		end

	mutex: ELOG_MUTEX is
			-- mutex for threading
		once
			create Result.make
		end
end
