indexing
	description: "object for keeping track of timers across an application at two levels of detail MAJOR and MINOR"
	author: "Theo Lengyel"
	keywords: "process,timer,time,seconds,cpu,profile"
	last_editor: "$Author: rj $"
	date: "$Date: 2013/11/25 $"
	revision: "$Revision: #6 $"
	path: "$File: //rose/source/kernel/infrastructure/system/rose_process_timer_collection.e $"

class
	ROSE_PROCESS_TIMER_COLLECTION

inherit
	ROSE_SHARED_LOGGER

feature -- Access
	resume ( a_major, a_minor: STRING ) is
			-- start the appropriate timer, and increment the counter
		local
			l_major: STRING
			l_minor: STRING
			l_parent_timer: like timer
			l_timer: like timer
		do
			l_major := safe_string( a_major )
			l_minor := safe_string( a_minor )

			increment_counter( l_major, l_minor )

			l_timer := timer( l_major, l_minor )
			if running_timers.count > 0 then
				l_parent_timer := running_timers.item
				if { l_parent_child: ROSE_LINEAR_ARRAY[like timer] } parent_child.item( l_timer ) then
					if not l_parent_child.has( l_parent_timer ) then
						l_parent_child.extend( l_parent_timer )
					end
				else
					parent_child.force(<<l_parent_timer>>, l_timer)
				end
			end
			running_timers.extend( l_timer )
			l_timer.resume
		end

	interrupt ( a_major, a_minor: STRING ) is
			-- stop the appropriate timer
		local
			l_major: STRING
			l_minor: STRING
			l_parent_timer: like timer
			l_timer: like timer
			l_error_message: STRING
		do
			l_major := safe_string( a_major )
			l_minor := safe_string( a_minor )

			l_timer := timer( l_major, l_minor )
			l_timer.interrupt
			if running_timers.item = l_timer then
				running_timers.remove
			else
				l_error_message := "The interrupt timer is NOT the latest one on the stack ... this indicates that some timer toggling is not properly aligned: " + l_major + "#" + l_minor + " <" + running_timers.item.name + ">"
				logger.error( l_error_message )
				(create {EXCEPTIONS}).raise( l_error_message )
			end
			if running_timers.count > 0 then
				l_parent_timer := running_timers.item
				l_parent_timer.set_child_elapsed( l_parent_timer.child_elapsed + l_timer.duration.fine_seconds_count )
			end
		end

	interrupt_thru(a_major, a_minor: STRING)
	-- Interrupt all timers starting from the top of the stack and working down to the one specified by the arguments.
	-- If the timer is not on the stack, do nothing.
	local
		l_major: STRING
		l_minor: STRING
		l_timer: like timer
		l_current_timer: like timer
		l_parent_timer: like timer
		l_done: BOOLEAN
	do
		l_major := safe_string(a_major)
		l_minor := safe_string(a_minor)

		l_timer := timer(l_major, l_minor)
		if running_timers.has(l_timer)
		then
			from
				do_nothing
			until
				l_done
			loop
				l_done := running_timers.item = l_timer
				l_current_timer := running_timers.item

				-- The remainder of the code is stolen from 'interrupt' (with minor variable name changes).
				l_current_timer.interrupt
				running_timers.remove
				if running_timers.count > 0 then
					l_parent_timer := running_timers.item
					l_parent_timer.set_child_elapsed(l_parent_timer.child_elapsed + l_current_timer.duration.fine_seconds_count)
				end
			end
		end
	end

	display_all_timer_information is
			-- dump out the data for all timers
		do
			logger.notice("Dumping out collected timer information ... ")
			total_cumulative_elapsed := 0.0
			total_cumulative_cpu := 0.0
			timers.current_keys.do_all( agent display_major_timer_information )
			logger.notice("TOTAL cumulative: elapsed = " + global.formatter.formatted_double(total_cumulative_elapsed, 1, 3) + "; cpu = " + global.formatter.formatted_double(total_cumulative_cpu, 1, 3))
		end

	display_major_timer_information ( a_major: STRING ) is
			-- dump out the data for `a_major'
		local
			l_major: STRING
		do
			l_major := safe_string( a_major )

			if timers.has( l_major ) then
				logger.notice("Dumping out timer information for: " + l_major )
				logger.notice( ":-:|average_elapsed|average_elapsed_self|average_elapsed_child|average_cpu|cumulative_elapsed|cumulative_elapsed_self|cumulative_elapsed_child|cumulative_cpu|count|major|minor|parent_timer|:-:" )
				major_cumulative_elapsed := 0.0
				major_cumulative_cpu := 0.0
				timers.item( l_major ).current_keys.do_all( agent display_minor_timer_information( l_major, ? ) )
				logger.notice(
					"MAJOR cumulative: "
					+ " elapsed = " + global.formatter.formatted_double(major_cumulative_elapsed, 1, 3)
					+ "; self elapsed = " + global.formatter.formatted_double(major_cumulative_self_elapsed, 1, 3)
					+ "; child elapsed = " + global.formatter.formatted_double(major_cumulative_child_elapsed, 1, 3)
					+ "; cpu = " + global.formatter.formatted_double(major_cumulative_cpu, 1, 3)
					+ " :: " + l_major
								 )
				-- I __think__ major_cumulative_self_elapsed is what we wnat to aggregate on for actual total elapsed
				total_cumulative_elapsed := total_cumulative_elapsed + major_cumulative_self_elapsed
				total_cumulative_cpu := total_cumulative_cpu + major_cumulative_cpu
			else
				logger.error( "No such major timer designation available: " + l_major )
			end
		end

	display_minor_timer_information ( a_major, a_minor: STRING ) is
			-- dump out the data for `a_major' and `a_minor'
		local
			l_average_elapsed: DOUBLE
			l_average_self_elapsed: DOUBLE
			l_average_child_elapsed: DOUBLE
			l_average_cpu: DOUBLE
			l_major: STRING
			l_minor: STRING
			l_parents: STRING
			l_counter_set: like counter_set
			l_counter: INTEGER
			l_timer: ROSE_PROCESS_TIMER
		do
			l_major := safe_string( a_major )
			l_minor := safe_string( a_minor )

			l_counter_set := counter_set( l_major )
			if l_counter_set.has( l_minor ) then
				l_counter := l_counter_set.item( l_minor )
				if l_counter > 0 then
					l_timer := timer( l_major, l_minor )
					if l_timer.is_active then
						logger.warn("Timer is still active! " + l_major + "/" + l_minor)
					end
					l_average_elapsed := l_timer.elapsed / l_counter
					l_average_self_elapsed := l_timer.self_elapsed / l_counter
					l_average_child_elapsed := l_timer.child_elapsed / l_counter
					l_average_cpu := l_timer.cpu_seconds / l_counter
					major_cumulative_elapsed := major_cumulative_elapsed + l_timer.elapsed
					major_cumulative_self_elapsed := major_cumulative_self_elapsed + l_timer.self_elapsed
					major_cumulative_child_elapsed := major_cumulative_child_elapsed + l_timer.child_elapsed
					major_cumulative_cpu := major_cumulative_cpu + l_timer.cpu_seconds
					l_parents := ""
					if { l_parent_timers: ROSE_LINEAR_ARRAY[ROSE_PROCESS_TIMER] } parent_child.item( l_timer ) then
						from
							l_parent_timers.start
						until
							l_parent_timers.off
						loop
							l_parents := l_parents + l_parent_timers.item_for_iteration.name + ";"
							l_parent_timers.forth
						end
					end
					logger.notice(
						":-:"
						+ "|" + formatted_double( l_average_elapsed )
						+ "|" + formatted_double( l_average_self_elapsed )
						+ "|" + formatted_double( l_average_child_elapsed )
						+ "|" + formatted_double( l_average_cpu )
						+ "|" + formatted_double( l_timer.elapsed )
						+ "|" + formatted_double( l_timer.self_elapsed )
						+ "|" + formatted_double( l_timer.child_elapsed )
						+ "|" + formatted_double( l_timer.cpu_seconds )
						+ "|" + l_counter.out
						+ "|" + l_major
						+ "|" + l_minor
						+ "|" + l_parents
						+ "|:-:"
								  )
				else
					logger.warn( "Count is zero for major/minor timer designation: " + l_major + "/" + l_minor )
				end
			else
				logger.error( "No such major/minor timer designation available: " + l_major + "/" + l_minor )
			end
		end

	wipe_out is
		-- clear out all timer information
		do
			timers.wipe_out
			counters.wipe_out
			running_timers.wipe_out
			parent_child.wipe_out
			major_minor.wipe_out
			count_cell.put(0)
		end

	count: INTEGER is
			-- number of timers that are active
		do
			Result := count_cell.item
		end

	timer ( a_major, a_minor: STRING ): ROSE_PROCESS_TIMER is
			-- return the timer for `a_major' and `a_minor'
		require
			a_major_not_void: a_major /= Void
			a_minor_not_void: a_minor /= Void
		local
			l_timers: HASH_TABLE[ROSE_PROCESS_TIMER, STRING]
		do
			if not timers.has( a_major ) then
				create l_timers.make(20)
				timers.force( l_timers, a_major )
			else
				l_timers := timers.item( a_major )
			end
			if not l_timers.has( a_minor ) then
				create Result.make
				Result.interrupt
				Result.set_name( a_major + "#" + a_minor )
				l_timers.force( Result, a_minor )
				count_cell.put(count_cell.item + 1)
				major_minor.force([a_major, a_minor], Result)
			else
				Result := l_timers.item( a_minor )
			end
		ensure
			result_not_void: Result /= Void
		end

feature {NONE} -- Implementation
	count_cell: CELL[INTEGER] is
			-- number of timers that are active
		once
			create Result.put(0)
		end

	parent_child: HASH_TABLE[ROSE_LINEAR_ARRAY[ROSE_PROCESS_TIMER], ROSE_PROCESS_TIMER] is
			-- keep track of nested timer relationships
		once
			create Result.make(0)
		end

	major_minor: HASH_TABLE[TUPLE[major, minor:STRING], ROSE_PROCESS_TIMER] is
			-- keep track of timer major and minor tags
		once
			create Result.make(0)
		end

	safe_string( a_string: STRING ): STRING is
			-- if `a_string' is Void then return "<Void>" otherwise return `a_string'
		do
			if a_string = Void then
				Result := "<Void>"
			else
				Result := a_string
			end
		ensure
			result_not_void: Result /= Void
			a_void_string_implies_result_equal_manifest_void_string: a_string = Void implies Result.is_equal("<Void>")
		end

	total_cumulative_elapsed: DOUBLE
			-- elapsed time across all timers

	major_cumulative_elapsed: DOUBLE
			-- total elapsed time across a single major bucket

	major_cumulative_self_elapsed: DOUBLE
			-- total elapsed minus child elapsed time across a single major bucket

	major_cumulative_child_elapsed: DOUBLE
			-- total child elapsed time across a single major bucket

	total_cumulative_cpu: DOUBLE
			-- cpu time across all timers

	major_cumulative_cpu: DOUBLE
			-- cpu time across all a single major bucket

	running_timers: ARRAYED_STACK[ROSE_PROCESS_TIMER] is
			-- currently running counters, to be used to help avoid double counting on cumulative totals
		once
			create Result.make(0)
		end

	double_formatter: FORMAT_DOUBLE is
			-- for sortable output
		once
			create Result.make(8,8)
		end

	formatted_double ( a_double: DOUBLE ): STRING is
			-- a double formatted to not use exponential notation
		do
			Result := double_formatter.formatted( a_double )
		end

	increment_counter( a_major, a_minor: STRING ) is
			-- increment the iteration counter for `a_major' and `a_minor'
		require
			a_major_not_void: a_major /= Void
			a_minor_not_void: a_minor /= Void
		local
			l_counter_set: like counter_set
		do
			l_counter_set := counter_set( a_major )
			if not l_counter_set.has( a_minor ) then
				l_counter_set.force( 0, a_minor )
			end
			l_counter_set.force( 1 + l_counter_set.item( a_minor ), a_minor )
		end

	counter_set ( a_major: STRING ): HASH_TABLE[INTEGER, STRING] is
			-- return the counter set for `a_major'
		require
			a_major_not_void: a_major /= Void
		do
			if not counters.has( a_major ) then
				create Result.make(20)
				counters.force( Result, a_major )
			else
				Result := counters.item( a_major )
			end
		ensure
			result_not_void: result /= Void
		end

	timers: HASH_TABLE[HASH_TABLE[ROSE_PROCESS_TIMER, STRING], STRING] is
			-- timers for tracking the cumulative time
		once
			create Result.make(20)
		end

	counters: HASH_TABLE[HASH_TABLE[INTEGER, STRING], STRING] is
			-- counters for tracking the cumulative counts
		once
			create Result.make(20)
		end

end

-- NOTES:
-- we may have situations where more than one counter is `resume'd at a time. In
-- that case, the cumulative time will be double counting

-- so we should probably add something to account for that.
-- then we could handle the adjustment in `interrupt' and allow
-- the major and total cumulative times to be reported correctly
--

-- This is not thread safe. To make it thread safe, I think our best bet will be to partition the timers/counters by thread id.
