indexing
	description: "Entity transformation recipe"
	author: "Zoran Simic"
	date: "$Date: 2009/08/24 $"
	revision: "$Revision: #17 $"

class
	ROSE_RECIPE

inherit
	ROSE_SHARED_GLOBAL
		redefine
			default_create
		end

	DEBUG_OUTPUT
		redefine
			default_create
		end

	ROSE_RECIPE_UTILS
		export
			{NONE} all
		redefine
			default_create
		end

	CONTEXT_ANCHOR

feature {NONE} -- Initialization

	default_create is
			-- Creation
		do
			create settings.make (Void, Void, Void)
		end

feature -- Access

	name: STRING
			-- Recipe name

	description: STRING
			-- Extensive description for this recipe, if provided

	transformation: ROSE_TRANSFORMATION
			-- First transformation defining this recipe
			-- This first transformation refers to the rest of the transformation sequence

	settings: ROSE_TRANSFORMATION_SETTINGS
			-- Display settings for the initial source of 'transformation'

	tagged_settings: ROSE_LINEAR_ARRAY [PAIR [STRING, ROSE_TRANSFORMATION_SETTINGS]] is
			-- Settings as an Array of string and setting object pairs
		local
			l_transformation: ROSE_TRANSFORMATION
			p: PAIR [STRING, ROSE_TRANSFORMATION_SETTINGS]
		do
			create Result.make_empty
			if transformation /= Void then
				create p.make (transformation.source, settings)
				Result.extend (p)
				from
					l_transformation := transformation
				until
					l_transformation = Void
				loop
					create p.make (l_transformation.target, l_transformation.settings)
					Result.extend (p)
					l_transformation := l_transformation.next
				end
			end
		end

	i_th_transformation (a_index: INTEGER): ROSE_TRANSFORMATION is
			-- Transformation with 'a_index' if any
		require
			valid_index: a_index > 0 and a_index <= depth
		local
			i: INTEGER
		do
			from
				i := 1
				Result := transformation
			until
				i = a_index
			loop
				Result := Result.next
				i := i + 1
			end
		ensure
			result_not_void: Result /= Void
		end

	last_transformation: ROSE_TRANSFORMATION is
			-- Last transformation of this recipe
		do
			if transformation /= Void then
				from
					Result := transformation
				until
					Result.next = Void
				loop
					Result := Result.next
				end
			end
		end

	depth: INTEGER is
			-- Number of transformations in this recipe
		do
			if transformation /= Void then
				Result := transformation.depth
			end
		end

	transformed (a_input: ARRAY [ENTITY]): ROSE_TRANSFORMATION_RESULT is
			-- 'a_input' transformed using this recipe
		require
			valid_input: a_input /= Void
		do
			create Result.make (Current)
			Result.set (a_input)
		end

	transformed_from_identifiers (a_input: ROSE_RECIPE_INPUT): ROSE_TRANSFORMATION_RESULT is
			-- `a_input' transformed using this recipe
		require
			valid_input: a_input /= Void
		local
			l_entities: ROSE_LINEAR_ARRAY[ENTITY]
			l_category: CATEGORY_HANDLER[like context_anchor,ENTITY]
			l_identifier: ENTITY_IDENTIFIER
			l_input_array: ARRAY[STRING]
			l_entity_array: ROSE_LINEAR_ARRAY[ENTITY]
			l_i, l_j: INTEGER
		do
			l_category := context_server.category_from_full_name (a_input.context_category)
			if l_category /= Void then
				l_identifier := l_category.context.identifier_from_name (a_input.identifier)
				if l_identifier /= Void then
					from
						l_input_array := a_input.attribute_values
						create l_entities.make_empty
						l_i := l_input_array.lower
					until
						l_i > l_input_array.upper
					loop
						l_entity_array := l_category.entity_list_from_identifier (l_input_array.item (l_input_array.lower), a_input.date, l_identifier)
						from
							l_j := l_entity_array.lower
						until
							l_j > l_entity_array.upper
						loop
							l_entities.extend (l_entity_array.item (l_j))
							l_j := l_j + 1
						end
						l_i := l_i + 1
					end
					if l_entities.count > 0 then
						Result := transformed (l_entities)
					end
				end
			end
		end

	transformed_from_all (a_category: STRING): ROSE_TRANSFORMATION_RESULT is
			--transform using this recipe using all entities from the source category as source
		require
			category_not_void: a_category /= Void
		local
			l_entities: ROSE_LINEAR_ARRAY[ENTITY]
			l_category: CATEGORY_HANDLER[like context_anchor,ENTITY]
		do
			l_category := context_server.category_from_full_name (a_category)
			if l_category /= Void then
				l_entities := l_category.all_entities
				if l_entities.count > 0 then
					Result := transformed (l_entities)
				end
			end
		end

	is_valid: BOOLEAN is
			-- Is this recipe valid?
		do
			Result := transformation /= Void and then transformation.is_valid
		end

	representation: STRING is
			-- Representation of this transformation
		local
			l_transformation: ROSE_TRANSFORMATION
		do
			create Result.make (100)
			Result.append ("recipe %"")
			Result.append_string (Global.formatter.eiffel_string (name))
			Result.append ("%"%N")
			if description /= Void then
				Result.append ("description %"")
				Result.append (Global.formatter.eiffel_string (description))
				Result.append ("%"%N")
			end
			if settings /= Void then
				Result.append_string (settings.representation)
			end
			Result.append ("transformation%N")
			from
				l_transformation := transformation
			until
				l_transformation = Void
			loop
				Result.append (l_transformation.representation)
				l_transformation := l_transformation.next
			end
			Result.append ("end%N")
		end

	debug_output: STRING is
			-- String that should be displayed in debugger to represent `Current'.
		do
			Result := representation
		end

feature -- Basic operations

	add_transformation (a_transformation: ROSE_TRANSFORMATION) is
			-- Add a transformation to the recipe
		do
			if transformation = Void then
				transformation := a_transformation
			else
				transformation.extend (a_transformation)
			end
		end

	remove_transformation (a_transformation: ROSE_TRANSFORMATION) is
			-- Remove 'a_transformation' from this recipe.
		do
			if transformation /= Void and then transformation.has (a_transformation) then
				if a_transformation = transformation then
						-- 'a_transformation' is our first transformation in recipe
					transformation := transformation.next
				end
				a_transformation.remove
			end
		end

	do_all (a_agent: PROCEDURE [ANY, TUPLE [ROSE_TRANSFORMATION]]) is
			-- Call 'a_agent' on every transformation of this recipe.
		local
			l_transformation: ROSE_TRANSFORMATION
		do
			from
				l_transformation := transformation
			until
				l_transformation = Void
			loop
				a_agent.call ([l_transformation])
				l_transformation := l_transformation.next
			end
		end

	set_description (a_description: STRING) is
			-- Set `description' to `a_description'
		do
			description := a_description
		ensure
			description_set: description = a_description
		end

	set_settings (a_settings: ROSE_TRANSFORMATION_SETTINGS) is
			-- Set `settings' to `a_settings'
		do
			settings := a_settings
		ensure
			settings_set: settings = a_settings
		end

	set_name (a_name: STRING) is
			-- Set `name' to `a_name'
		do
			name := a_name
		ensure
			name_set: name = a_name
		end

end
