indexing
	description: "Abstraction of a rectangle"
	author: "Mark Howard"
	date: "$Date: 2014/04/16 $"
	revision: "$Revision: #16 $"

class
	ROSE_RECTANGLE

inherit
	DEBUG_OUTPUT
		redefine
			out
		end

create
	default_create, make, make_normal, make_percentage, set_from_rectangle

feature -- Initialization

	make, set (a_x1, a_y1, a_x2, a_y2: DOUBLE) is
			-- Create object with value for (x1/y1) and (x2/y2)
		do
			x1 := a_x1.min (a_x2)
			y1 := a_y1.min (a_y2)
			x2 := a_x2.max (a_x1)
			y2 := a_y2.max (a_y1)
			is_set := True
		end

	make_normal is
			-- Make a normal rectangle (0,0) x (1,1)
			-- we'll eventually also refer to this as (0,1)^2...
		do
			make (0,0,1,1)
		end

	make_percentage (a_x1, a_y1, a_x2, a_y2: INTEGER) is
			-- Make a rectangle with coordinates expressed as percentage
			-- ie: 10,10,90,120 -> 0.1 , 0.1 , 0.9 , 1.2
		do
			make (a_x1/100,a_y1/100,a_x2/100,a_y2/200)
		end

	set_from_rectangle (a_other: ROSE_RECTANGLE) is
			-- Set rectangle dimensions to be the same as `a_other' (unset if `a_other = Void')
		do
			if a_other /= Void then
				x1 := a_other.x1
				y1 := a_other.y1
				x2 := a_other.x2
				y2 := a_other.y2
				is_set := a_other.is_set
			else
				unset
			end
		end

feature -- Access

	x1, x2, y1, y2: DOUBLE
			-- Rectangle's coordinates

	is_set: BOOLEAN
			-- Has this rectangle been set with actual coordinates?

	width: DOUBLE is
			-- Width of rectangle
		do
			Result := x2 - x1
		end

	height: DOUBLE is
			-- Height of rectangle
		do
			Result := y2 - y1
		end

	center_x: DOUBLE is
			-- x coordinate of center of rectangle
		do
			Result := (x1 + x2) / 2
		end

	center_y: DOUBLE is
			-- y coordinate of center of rectangle
		do
			Result := (y1 + y2) / 2
		end

	is_normal: BOOLEAN is
			-- Is this rectangle a normal rectangle? (ie coordinates (0,0)-(1,1))
		do
			Result := quite_equal (x1,0) and then quite_equal (y1,0)
				and then quite_equal (x2,1) and then quite_equal (y2,1)
		end

	point_distance (a_x1, a_y1, a_x2, a_y2: DOUBLE): DOUBLE is
			-- Distance between points (`a_x1'/`a_y1') and (`a_x2'/`a_y2')
		local
			l_dx, l_dy: DOUBLE
		do
			l_dx := a_x1 - a_x2
			l_dy := a_y1 - a_y2
			Result := math.sqrt (l_dx * l_dx + l_dy * l_dy)
		end

	distance (a_other: ROSE_RECTANGLE): DOUBLE is
			-- Distance from `Current' to `a_other'
		require
			is_set and then a_other /= Void and then a_other.is_set
		do
			if x2 < a_other.x1 then				-- is other on right?
				if y2 < a_other.y1 then			-- other on right / below
					Result := point_distance (x2, y2, a_other.x1, a_other.y1)
				elseif y1 > a_other.y2 then		-- other on right / above
					Result := point_distance (x2, y1, a_other.x1, a_other.y2)
				else
					Result := a_other.x1 - x2
				end
			elseif x1 > a_other.x2 then			-- other on left
				if y2 < a_other.y1 then			-- left / below
					Result := point_distance (x1, y2, a_other.x2, a_other.y1)
				elseif y1 > a_other.y2 then		-- left / above
					Result := point_distance (x1, y1, a_other.x2, a_other.y2)
				else
					Result := x1 - a_other.x2
				end
			elseif y2 < a_other.y1 then			-- below
				Result := a_other.y1 - y2
			elseif y1 > a_other.y2 then			-- above
				Result := y1 - a_other.y2
			end
		end

	intersects (a_other: ROSE_RECTANGLE): BOOLEAN is
			-- Does `Current' and `a_other' intersect?
		do
			if is_set and then a_other /= Void and then a_other.is_set then
				Result := (contains_x (a_other.x1) or else contains_x (a_other.x2))
					and then (contains_y (a_other.y1) or else contains_y (a_other.y2))
			end
		end

	contains_x (a_x: DOUBLE): BOOLEAN is
			-- Is `a_x' between `x1' and `x2'?
		do
			Result := x1 <= a_x and then a_x <= x2
		end

	contains_y (a_y: DOUBLE): BOOLEAN is
			-- Is `a_y' between `y1' and `y2'?
		do
			Result := y1 <= a_y and then a_y <= y2
		end

	intersection (a_other: ROSE_RECTANGLE): ROSE_RECTANGLE is
			-- Intersection rectangle with 'a_other'
		require
			valid_other: is_set and then a_other /= Void and then a_other.is_set
		local
			l_x1, l_x2, l_y1, l_y2: DOUBLE
		do
			l_x1 := x1.max (a_other.x1)
			l_x2 := x2.min (a_other.x2)
			l_y1 := y1.max (a_other.y1)
			l_y2 := y2.min (a_other.y2)
			if l_x1 <= l_x2 and then l_y1 <= l_y2 then
				create Result.make (l_x1, l_y1, l_x2, l_y2)
			else
				create Result.make (0, 0, 0, 0)
			end
		end

	area : REAL is
			-- are of Current
		do
			Result := (height * width).truncated_to_real
		end


	x_intersection_with_radial_ray (a_x, a_y : REAL) : REAL is
			-- X-value of intersection of segment between point (`y_center', `x_center') and point (`a_x', `a_y') with perimenter of Current
		local
			l_m, l_b, l_center_x, l_center_y : REAL
			l_y : REAL -- Upper or lower edge
		do
			l_center_x := center_x.truncated_to_real
			l_center_y := center_y.truncated_to_real
			l_m := (a_y - l_center_y)/(a_x - l_center_x)
			l_b := l_center_y - (l_m * l_center_x)
			if a_y >= l_center_y then
				l_y := y2.truncated_to_real
			else
				l_y := y1.truncated_to_real
			end
			Result := (l_y - l_b)/l_m
			Result := Result.min (x2.truncated_to_real)
			Result := Result.max (x1.truncated_to_real)
		end

	y_intersection_with_radial_ray (a_x, a_y : REAL) : REAL is
			-- Y-value of intersection of segment between point (`y_center', `x_center') and point (`a_x', `a_y') with perimenter of Current
		local
			l_m, l_b, l_center_x, l_center_y : REAL
			l_x : REAL -- Left or right edge
		do
			l_center_x := center_x.truncated_to_real
			l_center_y := center_y.truncated_to_real
			l_m := (a_y - l_center_y)/(a_x - l_center_x)
			l_b := l_center_y - (l_m * l_center_x)
			if a_x >= l_center_x then
				l_x := x2.truncated_to_real
			else
				l_x := x1.truncated_to_real
			end
			Result := (l_m * l_x) + l_b
			Result := Result.min (y2.truncated_to_real)
			Result := Result.max (y1.truncated_to_real)
		end

	relative_rectangle (a_other: ROSE_RECTANGLE): ROSE_RECTANGLE is
			-- Sub-rectangle of Current corresponding to 'a_other' (which is typically relative to (0,1)^2)
			-- Result is contained in Current if a_other is contained in (0,1)^2
		require
			valid_other: is_set and then a_other /= Void and then a_other.is_set
		local
			l_x1, l_y1, l_x2, l_y2 : DOUBLE
			l_width, l_height: DOUBLE
		do
			l_width := width
			l_height := height
			l_x1 := x1 + l_width * a_other.x1
			l_y1 := y1 + l_height * a_other.y1
			l_x2 := x1 + l_width * a_other.x2
			l_y2 := y1 + l_height * a_other.y2
			create Result.make (l_x1, l_y1, l_x2, l_y2)
		ensure
			normal_rectangle.contains (a_other) implies Current.contains (Result)
		end

	normalized_rectangle (a_other: ROSE_RECTANGLE): ROSE_RECTANGLE is
			-- Rectangle included in (0,1)^2 representing 'a_other' relatively to Current
			-- Result is in (0,1)^2 if Current.contains (a_other)
			-- Result is outside of (0,1)^2 if 'a_other' is not contained in Current
		require
			valid_other: is_set and then a_other /= Void and then a_other.is_set
		local
			l_x1, l_y1, l_x2, l_y2 : DOUBLE
			l_width, l_height: DOUBLE
		do
			l_width := width
			l_height := height
			if l_width /= 0.0 then
				l_x1 := (a_other.x1 - x1) / l_width
				l_x2 := 1 + (a_other.x2 - x2) / l_width
			end
			if l_height /= 0.0 then
				l_y1 := (a_other.y1 - y1) / l_height
				l_y2 := 1 + (a_other.y2 - y2) / l_height
			end
			create Result.make (l_x1, l_y1, l_x2, l_y2)
		ensure
			Current.contains (a_other) implies normal_rectangle.contains (Result)
		end

	union (a_other: ROSE_RECTANGLE): ROSE_RECTANGLE is
			-- Rectangle containing both 'Current' and 'a_other'.
		require
			valid_other: is_set and then a_other /= Void and then a_other.is_set
		do
			create Result.make (x1.min (a_other.x1),y1.min (a_other.y1),x2.max (a_other.x2),y2.max (a_other.y2))
		end

	projected_x (a_x: DOUBLE; a_target: ROSE_RECTANGLE): DOUBLE is
			-- Projection of 'a_x' as coordinates in this rectangle to coordinates of 'a_target' rectangle
		require
			valid_current: is_set
			valid_target: a_target /= Void and then a_target.is_set
		do
			Result := (a_x - x1) * width / a_target.width
		end

	projected_y (a_y: DOUBLE; a_target: ROSE_RECTANGLE): DOUBLE is
			-- Projection of 'a_y' as coordinates in this rectangle to coordinates of 'a_target' rectangle
		require
			valid_current: is_set
			valid_target: a_target /= Void and then a_target.is_set
		do
			Result := (a_y - y1) * width / a_target.width
		end

	has_point (a_x, a_y: DOUBLE): BOOLEAN is
			-- Is point ('a_x','a_y') within the bounds of this rectangle?
		do
			Result := is_set and then x1 <= a_x and then a_x <= x2
				and then y1 <= a_y and then a_y <= y2
		end

	contains (a_other: ROSE_RECTANGLE): BOOLEAN is
			-- Does this rectangle contain 'a_other'?
		do
			Result := x1 <= a_other.x1
				and then x2 >= a_other.x2
				and then y1 <= a_other.y1
				and then y2 >= a_other.y2
		end

	is_bigger (a_other: ROSE_RECTANGLE): BOOLEAN is
			-- Is this rectangle bigger than 'a_other'?
		do
			Result := width >= a_other.width and then height >= a_other.height
		end

	normal_rectangle: ROSE_RECTANGLE is
			-- Normal rectangle, do not modify!!
			-- Used for assertions
		once
			create Result.make_normal
		end

	out, debug_output: STRING is
			-- String that should be displayed in debugger to represent `Current'.
		do
			Result := "[" + f.formatted (x1) + "," + f.formatted (y1) + ";" + f.formatted (x2) + "," + f.formatted (y2) + "]"
		end

feature -- Basic operations

	set1 (a_x1, a_y1: DOUBLE) is
			-- Set ('x1','y1').
		do
			x1 := a_x1
			y1 := a_y1
			arrange_coordinates
		end

	set2 (a_x2, a_y2: DOUBLE) is
			-- Set ('x2','y2').
		do
			x2 := a_x2
			y2 := a_y2
			arrange_coordinates
		end

	set_x (a_x1, a_x2: DOUBLE) is
			-- Set 'x1'.
		do
			x1 := a_x1
			x2 := a_x2
			arrange_coordinates
		end

	set_y (a_y1, a_y2: DOUBLE) is
			-- Set 'y1'.
		do
			y1 := a_y1
			y2 := a_y2
			arrange_coordinates
		end

	move (a_other: ROSE_RECTANGLE) is
			-- Move `Current' to `a_other''s (x1/y1)
		local
			l_width, l_height: DOUBLE
		do
			l_width := width
			l_height := height
			x1 := a_other.x1
			y1 := a_other.y1
			set_size (l_width,l_height)
		end

	embed (a_other: ROSE_RECTANGLE) is
			-- Grow eventually this rectangle to make it contain 'a_other'.
		do
			if a_other.is_set
			then
				if is_set then
					x1 := x1.min (a_other.x1)
					y1 := y1.min (a_other.y1)
					x2 := x2.max (a_other.x2)
					y2 := y2.max (a_other.y2)
				else
					x1 := a_other.x1
					y1 := a_other.y1
					x2 := a_other.x2
					y2 := a_other.y2
					is_set := True
				end
			end
		end

	embed_point (a_x, a_y: DOUBLE) is
			-- Grow eventually this rectangle to make it contain ('a_x','a_y').
		do
			if is_set then
				x1 := x1.min (a_x)
				y1 := y1.min (a_y)
				x2 := x2.max (a_x)
				y2 := y2.max (a_y)
			else
				x1 := a_x
				y1 := a_y
				x2 := a_x
				y2 := a_y
				is_set := True
			end
		end

	unset is
			-- Make this rectangle undefined
		do
			is_set := False
		end

	set_size (a_width, a_height: DOUBLE) is
			-- Modify (x2,y2) such as the size of this rectabgle be 'a_width' and 'a_height'.
		do
			x2 := x1 + a_width
			y2 := y1 + a_height
		end

	translate (a_dx, a_dy: DOUBLE) is
			-- Translate this rectangle by ('a_dx','a_dy').
		do
			x1 := x1 + a_dx
			x2 := x2 + a_dx
			y1 := y1 + a_dy
			y2 := y2 + a_dy
		end

	center_on_point (a_x, a_y: DOUBLE) is
			-- Center this rectangle on point ('a_x','a_y').
		local
			l_delta: DOUBLE
		do
			l_delta := width / 2
			x1 := a_x - l_delta
			x2 := a_x + l_delta
			l_delta := height / 2
			y1 := a_y - l_delta
			y2 := a_y + l_delta
		end

	zoom (a_zoom: DOUBLE) is
			-- Zoom this rectangle on it's center by a factor of 'a_zoom'.
		local
			l_x, l_y, l_width, l_height: DOUBLE
		do
			l_width := width
			l_height := height
			l_x := x1 + l_width / 2
			l_y := y1 + l_height / 2
			l_width := (l_width / a_zoom) / 2
			l_height := (l_height / a_zoom) / 2
			x1 := l_x - l_width
			y1 := l_y - l_height
			x2 := l_x + l_width
			y2 := l_y + l_height
		end

feature {NONE} -- Implementation

	arrange_coordinates is
			-- Make sure that x1 <= x2 and y1 <= y2.
		local
			d: DOUBLE
		do
			if x2 < x1 then
				d := x1
				x1 := x2
				x2 := d
			end
			if y2 < y1 then
				d := y1
				y1 := y2
				y2 := d
			end
		end

	tolerance: DOUBLE is 10.e-4

	f: FORMAT_DOUBLE is once create Result.make (10,3) end

	quite_equal (a_value, a_target: DOUBLE): BOOLEAN is
			-- Are 'a_value' and 'a_target' close enough?
		do
			Result := (a_value - a_target).abs < tolerance
		end

	math: DOUBLE_MATH is once create Result end

end
