indexing
	description: "Interface for access to a windows registry section"
	author: "Zoran Simic"
	date: "$Date: 2009/04/06 $"
	revision: "$Revision: #7 $"

class
	ROSE_REGISTRY_IMPLEMENTATION

inherit
	ROSE_FORMATTING

create
	make

feature {NONE} -- Implementation

	make (a_root: STRING) is
			-- Make with 'a_root'
		do
			root := a_root
		end

feature -- Access

	root: STRING
			-- Name of root section to use

	is_readable: BOOLEAN is
			-- Is section corresponding to 'root' readable for current user?
		local
			l_pointer: POINTER
		once
			l_pointer := reg.open_key_with_access (root, reg.Key_read)
			if l_pointer /= default_pointer then
				Result := True
				reg.close_key (l_pointer)
			end
		end

	is_modifiable: BOOLEAN is
			-- Is section corresponding to 'root' modifiable for current user?
		local
			l_pointer: POINTER
		once
			l_pointer := reg.open_key_with_access (root, reg.Key_write)
			if l_pointer /= default_pointer then
				Result := True
				reg.close_key (l_pointer)
			end
		end

	get (a_key: STRING): STRING is
			-- Value corresponding to 'a_key' under 'root'
		require
			is_readable
		local
			l_value: ROSE_REGISTRY_KEY_VALUE
		do
			parse (a_key)
			if parsed_key /= Void then
				l_value := reg.open_key_value (parsed_key, parsed_subkey)
				if l_value /= Void then
					Result := l_value.string_value
				end
			end
		end

	get_dword (a_key: STRING): INTEGER is
			-- Dword value corresponding to 'a_key' under 'root'
		require
			is_readable
		local
			l_value: ROSE_REGISTRY_KEY_VALUE
		do
			parse (a_key)
			if parsed_key /= Void then
				l_value := reg.open_key_value (parsed_key, parsed_subkey)
				if l_value /= Void then
					Result := l_value.dword_value
				end
			end
		end

	sub_keys (a_key: STRING): ROSE_LINEAR_ARRAY [STRING] is
			-- Children keys of 'a_key' under 'root'
		require
			is_readable
		local
			k: ROSE_REGISTRY_KEY
			l_pointer: POINTER
			i, n: INTEGER
		do
			parse (a_key)
			if parsed_key /= Void then
				l_pointer := reg.open_key_with_access (parsed_key, reg.Key_read)
				if l_pointer /= default_pointer then
					n := reg.number_of_subkeys (l_pointer)
					create Result.make (1,n)
					from
						i := 0
					until
						i >= n
					loop
						k := reg.enumerate_key (l_pointer,i)
						if k /= Void then
							Result.put (k.name,i + 1)
						end
						i := i + 1
					end
				end
			end
		end

	sub_value_keys (a_key: STRING): ROSE_LINEAR_ARRAY [STRING] is
			-- Children sub-value keys of 'a_key' under 'root'
		require
			is_readable
		local
			k: STRING
			l_pointer: POINTER
			i, n: INTEGER
		do
			parse (a_key)
			if parsed_key /= Void then
				l_pointer := reg.open_key_with_access (parsed_key, reg.Key_read)
				if l_pointer /= default_pointer then
					n := reg.number_of_values (l_pointer)
					create Result.make (1,n)
					from
						i := 0
					until
						i >= n
					loop
						k := reg.enumerate_value (l_pointer,i)
						if k /= Void then
							Result.put (k,i + 1)
						end
						i := i + 1
					end
				end
			end
		end

	has (a_key: STRING): BOOLEAN is
		do
			Result := get (a_key) /= Void
		end

feature -- Basic operations

	set (a_key, a_value: STRING) is
			-- Set 'a_value' to 'a_key' under 'root'.
		require
			is_modifiable
		local
			l_pointer: POINTER
		do
			parse (a_key)
			if parsed_key /= Void then
				l_pointer := reg.open_key_with_access (parsed_key, reg.Key_set_value)
				if l_pointer = default_pointer then
					reg.create_new_key (parsed_key)
					l_pointer := reg.open_key_with_access (parsed_key, reg.Key_set_value)
				end
				if l_pointer /= default_pointer then
					reg.set_key_value (l_pointer, parsed_subkey, new_string_value (a_value))
					reg.close_key (l_pointer)
				end
			end
		end

	set_dword (a_key: STRING; a_value: INTEGER) is
		require
			is_modifiable
		local
			l_pointer: POINTER
		do
			parse (a_key)
			if parsed_key /= Void then
				l_pointer := reg.open_key_with_access (parsed_key, reg.Key_set_value)
				if l_pointer = default_pointer then
					reg.create_new_key (parsed_key)
					l_pointer := reg.open_key_with_access (parsed_key, reg.Key_set_value)
				end
				if l_pointer /= default_pointer then
					reg.set_key_value (l_pointer, parsed_subkey, new_dword_value (a_value))
					reg.close_key (l_pointer)
				end
			end
		end

	delete (a_key: STRING) is
			-- Delete 'a_key' from 'root'.
		require
			is_modifiable
		local
			l_pointer: POINTER
		do
			parse (a_key)
			if parsed_key /= Void then
				l_pointer := reg.open_key_with_access (parsed_key, reg.Key_write)
				if l_pointer /= default_pointer then
					reg.delete_key (l_pointer,parsed_subkey)
					reg.delete_value (l_pointer,parsed_subkey)
					reg.close_key (l_pointer)
				end
			end
		end

	delete_recursively (a_key: STRING) is
			-- Delete 'a_key' from 'root'.
		require
			is_modifiable
		local
			l_pointer: POINTER
			b: BOOLEAN
		do
			parse (a_key)
			if parsed_key /= Void then
				l_pointer := reg.open_key_with_access (parsed_key, reg.Key_write)
				if l_pointer /= default_pointer then
					reg.delete_key (l_pointer,parsed_subkey)
					b := reg.last_call_successful
					reg.close_key (l_pointer)
				end
--				print (b.out + " --> " + root + a_key + "%N")
			end
		end

feature {NONE} -- Implementation

	parsed_key: STRING
	parsed_subkey: STRING

	parse (a_key: STRING) is
		local
			p: PAIR [STRING,STRING]
		do
			parsed_key := Void
			parsed_subkey := Void
			if a_key /= Void then
				p := formatter.split_on_last (a_key,"\",Void)
				if p /= Void and then p.first /= Void then
					if p.second = Void then
						parsed_key := root
						parsed_subkey := p.first
					else
						parsed_key := root + p.first
						parsed_subkey := p.second
					end
				end
			end
		end

	reg: ROSE_REGISTRY_ISE_IMPLEMENTATION is once create Result end

	new_string_value (a_value: STRING): ROSE_REGISTRY_KEY_VALUE is
		do
			create Result.make (a_value)
		end

	new_dword_value (a_value: INTEGER): ROSE_REGISTRY_KEY_VALUE is
		do
			create Result.make_with_dword_value (a_value)
		end

end
