indexing
	description: "Registry manager"
	notice: "This was taken from WEL_REGISTRY, because WEL should not be needed just to access the registry..."
	date: "$Date: 2014/02/26 $"
	revision: "$Revision: #10 $"

class
	ROSE_REGISTRY_ISE_IMPLEMENTATION

inherit
	STRING_HANDLER

feature -- Actions

	create_new_key (a_key_path: STRING) is
				-- Create a new key, with as path 'a_key_path'
				-- The path should be like "a\b\c"
		require
			at_least_one_back_slash: a_key_path /= Void and then a_key_path.has ('\')
		local
			l_node_list: ARRAYED_LIST [STRING]
			l_index_value, i: POINTER
		do
			l_node_list := value_keys_list (a_key_path)
			check
				node_list_possible: l_node_list.count > 0
			end
			l_node_list.start
			l_index_value := index_value_for_root_keys (l_node_list.item)
			from
				l_node_list.forth
			until
				l_node_list.after
			loop
				i := open_key (l_index_value,l_node_list.item, Key_create_sub_key)
				if i = default_pointer then
					i := create_key (l_index_value, l_node_list.item, Key_create_sub_key)
				end
				close_key (l_index_value)
				l_index_value := i
				l_node_list.forth
			end
		end

	open_key_with_access (a_key_path: STRING; a_access: INTEGER): POINTER is
				-- Open the key relative to the path 'a_key_path', with
				-- the access 'a_access'.
				-- Return the key reference (default_pointer if the operation failed).
		require
			at_least_one_back_slash: a_key_path /= Void and then a_key_path.has ('\')
		local
			l_node_list: ARRAYED_LIST [STRING]
			l_index_value, i: POINTER
		do
			l_node_list := value_keys_list (a_key_path)
			check
				node_list_possible: l_node_list.count>0
				first_element_possible: basic_valid_name_for_HKEY (l_node_list.first)
			end
			l_node_list.start
			l_index_value := index_value_for_root_keys (l_node_list.item)
			from
				l_node_list.forth
			until
				l_node_list.after or l_index_value = default_pointer
			loop
				i := open_key (l_index_value,l_node_list.item, a_access)
				close_key (l_index_value)
				l_index_value := i
				l_node_list.forth
			end
			Result := l_index_value
		end

	open_key_value (a_key_path: STRING; a_value_name: STRING): ROSE_REGISTRY_KEY_VALUE is
				-- Open a key, with as path 'path' and
				-- name 'key_name'.
				-- The path should be like "a\b\c"
		require
			key_name_possible: a_value_name /= Void
			at_least_one_back_slash: a_key_path /= Void and then a_key_path.has ('\')
		local
			l_node_list: ARRAYED_LIST [STRING]
			l_index_value, i: POINTER
			l_stop: BOOLEAN
		do
			l_node_list := value_keys_list (a_key_path)
			check
				node_list_possible: l_node_list.count>0
				first_element_possible: basic_valid_name_for_HKEY (l_node_list.first)
			end
			l_node_list.start
			l_index_value := index_value_for_root_keys (l_node_list.item)
			from
				l_node_list.forth
			until
				l_node_list.after or l_stop
			loop
				i := open_key (l_index_value,l_node_list.item, Key_read)
				if i = default_pointer then
					l_stop := True
				end
				close_key (l_index_value)
				l_index_value := i
				l_node_list.forth
			end
			if not l_stop then
				Result := key_value (l_index_value, a_value_name)
			end
		end

feature -- Status

	last_call_successful: BOOLEAN
			-- Did last call succeed?

feature {NONE} -- Internal Results

	value_keys_list (a_path: STRING): ARRAYED_LIST [STRING] is
			-- From a 'a_path' (of the form "a\b\c\...\x")
			-- Return a list of string: a,b,c,d,e,...x
		require
			path_ok: a_path /= Void
		local
			i, j, nb: INTEGER
		do
			create Result.make (10)
			from
				i := 1
				j := 1
				nb := a_path.count
			until
				j < 1 or i > nb
			loop
				j := a_path.index_of ('\', i)
				if j > i then
					Result.extend (a_path.substring (i, j - 1))
					i := j + 1
				end
			end
			if j = 0 and then i <= nb then
				Result.extend (a_path.substring (i, nb))
			end
		ensure
			Result /= Void
		end

feature -- Access

	key_from_remote_host (a_host_name: STRING; a_root_key: POINTER): POINTER is
			-- Connect the computer designed by its name 'a_host_name'.
			-- 'a_host_name' should be under the format: \\computer_name
			-- 'a_root_key' is the key from where we want to start the
			-- investigations on the remote machine.
			-- If 'a_host_name' is empty, then the local computer is used by default.
		require
			root_key_possible: a_root_key = HKEY_LOCAL_MACHINE or
							   a_root_key = HKEY_USERS or
							   a_root_key = HKEY_PERFORMANCE_DATA
			host_name_possible: a_host_name /= Void
		local
			l_c_string: C_STRING
			res: INTEGER
		do
			if not a_host_name.is_empty then
				create l_c_string.make (a_host_name)
				res := cwin_reg_connect_key (l_c_string.item, a_root_key, $Result)
			else
				res := cwin_reg_connect_key (default_pointer, a_root_key, $Result)
			end
			last_call_successful := res = Error_success
			if not last_call_successful then
				Result := default_pointer
			end
		end

	enumerate_key (a_key: POINTER; a_index: INTEGER): ROSE_REGISTRY_KEY is
			-- `a_index'-th subkey of `a_key',
			-- Void if `a_key' has less than `a_index' subkeys.
		require
			key_possible: valid_value_for_hkey (a_key)
			index_possible: a_index <= number_of_subkeys (a_key)
		local
			file_time: MANAGED_POINTER
			key_name, class_name: C_STRING
			key_size, class_size: INTEGER
			res: INTEGER
		do
			key_size := 512
			class_size := 1024
			create key_name.make_empty (key_size)
			create class_name.make_empty (class_size)
			create file_time.make (c_size_of_file_time)
			res := cwin_reg_enum_key (
				a_key,
				a_index,
				key_name.item,
				$key_size,
				default_pointer,
				class_name.item,
				$class_size,
				file_time.item)
			last_call_successful := res = Error_success
			if last_call_successful then
				create Result.make (key_name.string, class_name.string)
			end
		end

	default_key_value (a_key: POINTER; subkey: STRING): ROSE_REGISTRY_KEY_VALUE is
			-- Retrieve value of `subkey' associated with open `a_key'.
		obsolete
			"Use `key_value' with an empty string instead, if `subkey' is Void or empty. %
			%Otherwise open manually the subkey represented by `subkey' and then call `key_value' %
			%with an empty string."
		require
			key_possible: valid_value_for_hkey (a_key)
		local
			l_subkey: C_STRING
			l_ext: MANAGED_POINTER
			l_size: INTEGER
		do
			if subkey /= Void then
				last_call_successful := False
				create l_subkey.make (subkey)
				if cwin_reg_query_value (a_key, l_subkey.item, default_pointer, $l_size) = error_success then
					create l_ext.make (l_size)
					if cwin_reg_query_value (a_key, l_subkey.item, l_ext.item, $l_size) = error_success then
						last_call_successful := True
						create Result.make_with_data ({ROSE_REGISTRY_KEY_VALUE_TYPE}.reg_sz, l_ext)
					end
				end
			else
				Result := key_value (a_key, "")
			end
		end

feature -- Settings

	set_key_value (a_key: POINTER; a_value_name: STRING; value: ROSE_REGISTRY_KEY_VALUE) is
			-- Change value defined by `a_key' and `a_value_name' to `value'.
	     	-- The a_key identified by the hKey parameter must have been
			-- opened with KEY_SET_VALUE access
		require
			valid_value: value /= Void
			valid_value_name: a_value_name /= Void
			key_possible:valid_value_for_hkey (a_key)
		local
			a: C_STRING
			res: INTEGER
		do
			if a_value_name /= Void then
				create a.make (a_value_name)
				res := cwin_reg_set_key_value (a_key, a.item, 0, value.type,
					value.data.item, value.data.count)
			else
				res := cwin_reg_set_key_value (a_key, default_pointer, 0, value.type,
					value.data.item, value.data.count)
			end
			last_call_successful := res = Error_success
		end

feature -- Basic Actions

	create_key (a_parent_key: POINTER; key_name: STRING; sam: INTEGER): POINTER is
			-- Create `key_name' under `a_parent_key' according to `sam'.
			-- Return handle to created key or default_pointer on failure.
		require
			key_name_possible: key_name /= Void and then not key_name.is_empty
			parent_key_possible:valid_value_for_hkey (a_parent_key)
		local
			a: C_STRING
			disp, res: INTEGER
		do
			create a.make (key_name)
			res := cwin_reg_create_key (
				a_parent_key,
				a.item,
				0,
				Reg_none,
				Reg_option_non_volatile,
				sam,
				default_pointer,
				$Result,
				$disp)
			last_call_successful := res = Error_success
			if not last_call_successful then
				Result := default_pointer
			end
		end

	open_key (a_parent_key: POINTER; key_name: STRING; access_mode: INTEGER): POINTER is
			-- Open subkey `key_name' of `a_parent_key' according to `access_mode'.
			-- Return handle to created key or default_pointer on failure.
		require
			key_name_possible: key_name /= Void and then not key_name.is_empty
			parent_key_possible:valid_value_for_hkey (a_parent_key)
		local
			a: C_STRING
			res: INTEGER
		do
			create a.make (key_name)
			res := cwin_reg_open_key (a_parent_key, a.item, 0, access_mode, $Result)
			if res /= Error_success then
				Result := default_pointer
			end
		end

	close_key (a_key: POINTER) is
			-- Close `a_key'.
			-- Return True if succeeded, False otherwise.
		require
			key_possible:valid_value_for_hkey (a_key)
		do
			last_call_successful := cwin_reg_close_key (a_key) = Error_success
		end

	delete_key (a_parent_key: POINTER; key_name: STRING) is
			-- Delete subkey `key_name' of `a_parent_key'.
			-- Return True if succeeded, False otherwise.
			-- Under Windows 95, all subkeys are deleted.
			-- Under Windows NT, only specified key is deleted
			-- it should not have subkeys.
		require
			key_name_possible: key_name /= Void and then not key_name.is_empty
			parent_key_possible:valid_value_for_hkey (a_parent_key)
		local
			a: C_STRING
			res: INTEGER
		do
			create a.make (key_name)
			res := cwin_reg_delete_key (a_parent_key, a.item)
			last_call_successful := res = Error_success
		end

	delete_key_recursively (a_parent_key: POINTER; key_name: STRING) is
			-- Delete subkey `key_name' of `a_parent_key'.
			-- Return True if succeeded, False otherwise.
			-- Under Windows 95, all subkeys are deleted.
			-- Under Windows NT, only specified key is deleted
			-- it should not have subkeys.
		require
			key_name_possible: key_name /= Void and then not key_name.is_empty
			parent_key_possible:valid_value_for_hkey (a_parent_key)
		local
			a: C_STRING
			res: INTEGER
		do
			create a.make (key_name)
			res := cwin_reg_recursive_delete_key (a_parent_key, a.item)
			last_call_successful := res = Error_success
		end

	enumerate_values (a_key: POINTER): LINKED_LIST [STRING] is
		-- Find the names of the key values within the key referenced by 'a_key'.
		local
			i: INTEGER
			l_string: STRING
		do
			create Result.make
			from
				i:=0
			until
				i = -1
			loop
				l_string := enumerate_value (a_key, i)
				if l_string /= Void then
					Result.extend (l_string)
					i := i+1
				else
					i := -1
				end
			end
		end

feature  -- New actions

	delete_value (a_parent_key: POINTER; a_name: STRING) is
			-- Delete value with 'a_name' from 'a_parent_key'.
		require
			key_possible: valid_value_for_hkey (a_parent_key)
			name_possible: a_name /= Void
		local
			a: C_STRING
			res: INTEGER
		do
			create a.make (a_name)
			res := cwin_reg_delete_value (a_parent_key, a.item)
			last_call_successful := res = Error_success
		end

	enumerate_value (a_key: POINTER; a_index: INTEGER): STRING is
			-- Find the name of the key_value corresponding
			-- to the key 'a_key' and the index 'a_index'.
		local
			p: POINTER
			res, l_size: INTEGER
			ext: C_STRING
		do
			l_size := 1024
			create ext.make_empty (l_size)
			res  := cwin_reg_enum_value (a_key, a_index, ext.item, $l_size, p, p, p, p)
			last_call_successful := res = Error_success
			if last_call_successful then
				Result := ext.string
			else
				Result := Void
			end
		end

	number_of_subkeys (a_key: POINTER): INTEGER is
			-- Number of sub keys contained in 'a_key'
		local
			res, nbkey, l_size, values: INTEGER
			p: POINTER
		do
			res := cwin_reg_value_number (a_key, p, p, p, $Result, $nbkey, $l_size, $values, p, p, p, p)
			last_call_successful := res = Error_success
			if not last_call_successful then
				Result := 0
			end

		end

	number_of_values (a_key: POINTER): INTEGER is
			-- Number of values contained in 'a_key'
		local
			res, nbkeys, nbkey, l_size: INTEGER
			p: POINTER
		do
			res := cwin_reg_value_number (a_key, p, p, p, $nbkeys, $nbkey, $l_size, $Result, p, p, p, p)
			last_call_successful := res = Error_success
			if not last_call_successful then
				Result := 0
			end
		end

	valid_value_for_hkey (a_key: POINTER): BOOLEAN is
			-- Does key pointed by 'a_key' exists
		do
			Result := (a_key /= default_pointer)
		end

feature -- Access

	key_value (a_key: POINTER; a_value_name: STRING): ROSE_REGISTRY_KEY_VALUE is
			-- Retrieve value of `a_value_name' associated with open
			-- `a_key'.
			-- The identifier 'a_key' relative to the parent key must
			-- have been opened with the KEY_QUERY_VALUE access.
		require
        	value_name_possible: a_value_name /= Void
			key_valid: valid_value_for_hkey (a_key)
		local
			l_name: C_STRING
			l_type, l_size, l_res: INTEGER
			l_ext: MANAGED_POINTER
			l_null: POINTER
		do
			create l_name.make (a_value_name)
			last_call_successful := False
			if cwin_reg_query_value_ex (a_key, l_name.item, l_null, l_null, l_null, $l_size) = Error_success then
					-- Repeat until it works and that no more data has to be read. See MSDN
					-- comments about why this needs to be done when the root key is HKEY_PERFORMANCE_DATA.
				from
					create l_ext.make (l_size)
					l_res := cwin_reg_query_value_ex (a_key, l_name.item, l_null, $l_type, l_ext.item, $l_size)
				until
					l_res = error_success or l_res /= error_more_data
				loop
					create l_ext.make (l_size)
					l_res := cwin_reg_query_value_ex (a_key, l_name.item, l_null, $l_type, l_ext.item, $l_size)
				end
				if l_res = error_success then
					create Result.make_with_data (l_type, l_ext)
					last_call_successful := True
				end
			end
		end

feature -- Access modes

	frozen Key_create_link: INTEGER is
			-- Permission to create a symbolic link.
		external
			"C macro use <windows.h>"
		alias
			"KEY_CREATE_LINK"
		end

	frozen Key_create_sub_key: INTEGER is
			-- Permission to create subkeys.
		external
			"C macro use <windows.h>"
		alias
			"KEY_CREATE_SUB_KEY"
		end

	frozen Key_enumerate_sub_keys: INTEGER is
			-- Permission to enumerate subkeys.
		external
			"C macro use <windows.h>"
		alias
			"KEY_ENUMERATE_SUB_KEYS"
		end

	frozen Key_execute: INTEGER is
			-- Permission for read access.
		external
			"C macro use <windows.h>"
		alias
			"KEY_EXECUTE"
		end

	frozen Key_notify: INTEGER is
			-- Permission for change notification.
		external
			"C macro use <windows.h>"
		alias
			"KEY_NOTIFY"
		end

	frozen Key_query_value: INTEGER is
			-- Permission to query subkey data.
		external
			"C macro use <windows.h>"
		alias
			"KEY_QUERY_VALUE"
		end

	frozen Key_set_value: INTEGER is
			-- Permission to set subkey data.
		external
			"C macro use <windows.h>"
		alias
			"KEY_SET_VALUE"
		end

	frozen Key_all_access: INTEGER is
			-- Combines the KEY_QUERY_VALUE, KEY_ENUMERATE_SUB_KEYS,
			-- KEY_NOTIFY, KEY_CREATE_SUB_KEY, KEY_CREATE_LINK, and
			-- KEY_SET_VALUE access rights, plus all the standard
			-- access rights except SYNCHRONIZE.
		external
			"C macro use <windows.h>"
		alias
			"KEY_ALL_ACCESS"
		end

	frozen Key_read: INTEGER is
			-- Combines the STANDARD_RIGHTS_READ, KEY_QUERY_VALUE,
			-- KEY_ENUMERATE_SUB_KEYS, and KEY_NOTIFY access rights.
		external
			"C macro use <windows.h>"
		alias
			"KEY_READ"
		end

	frozen Key_write: INTEGER is
			-- Combines the STANDARD_RIGHTS_WRITE, KEY_SET_VALUE,
			-- and KEY_CREATE_SUB_KEY access rights.
		external
			"C macro use <windows.h>"
		alias
			"KEY_WRITE"
		end

feature -- From WEL_HKEY

	basic_valid_value_for_HKEY (value: POINTER): BOOLEAN is
		-- Return True if 'value' is one of the basic following values.
		do
			Result :=  value = Hkey_classes_root
				or else value = Hkey_dyn_data
				or else value = Hkey_current_config
				or else value = Hkey_performance_data
				or else value = Hkey_users
				or else value = Hkey_current_user
				or else value = Hkey_local_machine
		end

	basic_valid_name_for_HKEY (a_name: STRING): BOOLEAN is
			-- Return True if 'a_name' correspond to one of the
			-- value names below.
		do
			if a_name /= Void then
				a_name.to_lower
				Result :=  a_name.is_equal ("hkey_classes_root")
					or else a_name.is_equal ("hkey_dyn_data")
					or else a_name.is_equal ("hkey_current_config")
					or else a_name.is_equal ("hkey_performance_data")
					or else a_name.is_equal ("hkey_users")
					or else a_name.is_equal ("hkey_current_user")
					or else a_name.is_equal ("hkey_local_machine")
			end
		end

	index_value_for_root_keys (a_name: STRING): POINTER is
			-- Return the index corresponding to a root key.
		require
			name_possible: a_name /= Void and then basic_valid_name_for_HKEY (a_name)
		do
			a_name.to_lower
			if a_name.is_equal ("hkey_classes_root") then
					Result := Hkey_classes_root
			elseif a_name.is_equal ("hkey_dyn_data") then
					Result := Hkey_dyn_data
			elseif a_name.is_equal ("hkey_current_config") then
					Result := Hkey_current_config
			elseif a_name.is_equal ("hkey_performance_data") then
					Result := Hkey_performance_data
			elseif a_name.is_equal ("hkey_users") then
					Result := Hkey_users
			elseif a_name.is_equal ("hkey_current_user") then
					Result := Hkey_current_user
			elseif a_name.is_equal ("hkey_local_machine") then
					Result := Hkey_local_machine
			end
		end

feature -- Access

	frozen Hkey_classes_root: POINTER is
			-- HKEY_CLASSES_ROOT
		external
			"C macro signature (): EIF_POINTER use <winreg.h>"
		alias
			"HKEY_CLASSES_ROOT"
		end

	frozen Hkey_current_user: POINTER is
			-- HKEY_CURRENT_USER
		external
			"C macro signature (): EIF_POINTER use <winreg.h>"
		alias
			"HKEY_CURRENT_USER"
		end

	frozen Hkey_local_machine: POINTER is
			-- HKEY_LOCAL_MACHINE
		external
			"C macro signature (): EIF_POINTER use <winreg.h>"
		alias
			"HKEY_LOCAL_MACHINE"
		end

	frozen Hkey_users: POINTER is
			-- HKEY_USERS
		external
			"C macro signature (): EIF_POINTER use <winreg.h>"
		alias
			"HKEY_USERS"
		end

	frozen Hkey_performance_data: POINTER is
			-- HKEY_PERFORMANCE_DATA
		external
			"C macro signature (): EIF_POINTER use <winreg.h>"
		alias
			"HKEY_PERFORMANCE_DATA"
		end

	frozen Hkey_current_config: POINTER is
			-- HKEY_CURRENT_CONFIG
		external
			"C macro signature (): EIF_POINTER use <winreg.h>"
		alias
			"HKEY_CURRENT_CONFIG"
		end

	frozen Hkey_dyn_data: POINTER is
			-- HKEY_DYN_DATA
		external
			"C macro signature (): EIF_POINTER use <winreg.h>"
		alias
			"HKEY_DYN_DATA"
		end

feature {NONE} -- Externals

	cwin_reg_create_key (a_parent_key: POINTER; key_name: POINTER; res: INTEGER; clas: POINTER; opt, sam: INTEGER; sec, resu, dis: POINTER): INTEGER is
			-- win32 external RegCreateKeyEx
		external
			"C macro signature (HKEY, LPCSTR, DWORD, LPSTR, DWORD, REGSAM, LPSECURITY_ATTRIBUTES, PHKEY, LPDWORD): EIF_INTEGER use <windows.h>"
		alias
			"RegCreateKeyExA"
		end

	cwin_reg_open_key (a_parent_key: POINTER; key_name: POINTER; options, access_mode: INTEGER; res: POINTER): INTEGER is
			-- win32 external RegOpenKeyEx
		external
			"C macro signature (HKEY, LPCSTR, DWORD, REGSAM, PHKEY): EIF_INTEGER use <windows.h>"
		alias
			"RegOpenKeyExA"
		end

	cwin_reg_recursive_delete_key (a_key: POINTER; subkey: POINTER): INTEGER is
			-- win32 external SHDeleteKey
		external
			"C macro signature (HKEY, LPCSTR): EIF_INTEGER use <shlwapi.h>"
		alias
			"SHDeleteKeyA"
		end

	cwin_reg_delete_key (a_key: POINTER; subkey: POINTER): INTEGER is
			-- win32 external RegDeleteKey
		external
			"C macro signature (HKEY, LPCSTR): EIF_INTEGER use <windows.h>"
		alias
			"RegDeleteKeyA"
		end

	cwin_reg_enum_key (a_key: POINTER; a_index: INTEGER; a_name, cname, reserved, cl, ccl, lastwrite: POINTER): INTEGER is
			-- win32 external RegEnumKeyEx
		external
			"C macro signature (HKEY, DWORD, LPSTR, LPDWORD, LPDWORD, LPSTR, LPDWORD, PFILETIME): EIF_INTEGER use <windows.h>"
		alias
			"RegEnumKeyExA"
		end

	cwin_reg_set_key_value (a_key, keyname: POINTER; res, type: INTEGER; data: POINTER; siz: INTEGER): INTEGER is
			-- win32 external RegSetValueEx
		external
			"C macro signature (HKEY, LPCSTR, DWORD, DWORD, BYTE *, DWORD): EIF_INTEGER use <windows.h>"
		alias
			"RegSetValueExA"
		end

	cwin_reg_close_key (a_key: POINTER): INTEGER is
			-- win32 external RegCloseKey
		external
			"C macro signature (HKEY): EIF_INTEGER use <windows.h>"
		alias
			"RegCloseKey"
		end

	cwin_reg_query_value_ex (a_key, a_value_name, res, type, data, a_size: POINTER): INTEGER is
			-- win32 external RegQueryValueEx
		external
			"C macro signature (HKEY, LPCSTR, LPDWORD, LPDWORD, LPBYTE, LPDWORD): EIF_INTEGER use <windows.h>"
		alias
			"RegQueryValueExA"
		end

	cwin_reg_delete_value (a_key, a_value_name: POINTER): INTEGER is
			-- win32 external RegDeleteValue
		external
			"C macro signature (HKEY, LPCSTR): EIF_INTEGER use <windows.h>"
		alias
			"RegDeleteValueA"
		end

	cwin_reg_query_value (a_key: POINTER; subkey, a_value_name, a_size: POINTER): INTEGER is
			-- win32 external RegQueryValue
		external
			"C macro signature (HKEY, LPCSTR, LPSTR, PLONG): EIF_INTEGER use <windows.h>"
		alias
			"RegQueryValueA"
		end

	cwin_reg_enum_value (a_key: POINTER; a_index: INTEGER; val_name, val_size, res, type, data, pcb_data: POINTER): INTEGER is
			-- win32 external RegEnumValue
		external
			"[
				C macro signature
					(HKEY, DWORD, LPSTR, LPDWORD, LPDWORD, LPDWORD, LPBYTE, LPDWORD): EIF_INTEGER
				use <windows.h>
			]"
		alias
			"RegEnumValueA"
		end

	cwin_reg_value_number (a_key, a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11: POINTER) : INTEGER is
			-- win32 external RegQueryInfoKey
		external
			"[
			C macro signature
				(HKEY, LPSTR, LPDWORD, LPDWORD, LPDWORD, LPDWORD, LPDWORD, LPDWORD, LPDWORD, LPDWORD, LPDWORD, PFILETIME): EIF_INTEGER
				use <windows.h>
			]"
		alias
			"RegQueryInfoKeyA"
		end

	cwin_reg_connect_key (a_name, a_key, pkey: POINTER): INTEGER is
			-- win32 external RegConnectRegistry
		external
			"C macro signature (LPCSTR, HKEY, PHKEY): EIF_INTEGER use <windows.h>"
		alias
			"RegConnectRegistryA"
		end

feature {NONE} -- External constants

	Error_success: INTEGER is
			-- win32 constant ERROR_SUCCESS
		external
			"C macro use <windows.h>"
		alias
			"ERROR_SUCCESS"
		end

	Error_more_data: INTEGER is
			-- win32 constant ERROR_MORE_DATA
		external
			"C macro use <windows.h>"
		alias
			"ERROR_MORE_DATA"
		end

	Error_no_more_items: INTEGER is
			-- win32 constant ERROR_NO_MORE_ITEMS
		external
			"C macro use <windows.h>"
		alias
			"ERROR_NO_MORE_ITEMS"
		end

	Reg_none: POINTER is
			-- win32 constant REG_NONE
		external
			"C macro use <windows.h>"
		alias
			"REG_NONE"
		end

	Reg_option_non_volatile: INTEGER is
			-- win32 constant REG_OPTION_NON_VOLATILE
		external
			"C macro use <windows.h>"
		alias
			"REG_OPTION_NON_VOLATILE"
		end

	c_size_of_file_time: INTEGER is
			-- Size of FILETIME structure
		external
			"C macro use <windows.h>"
		alias
			"sizeof(FILETIME)"
		end

end
