indexing
	description: "Object to audit and record various resource type accesses"
	author: "Mauricio Bustos"
	last_editor: "$Author: an $"
	date: "$Date: 2013/10/09 $"
	revision: "$Revision: #13 $"
    path: "$File: //rose/source/kernel/infrastructure/framework/db/rose_resource_access.e $"

class
	ROSE_RESOURCE_ACCESS

inherit
	ROSE_SHARED_GLOBAL
	DATABASE_SERVER
	ROSE_RESOURCE_ACCESS_CONSTANTS

feature {PM_CONTEXT} -- Implementation

	clear_resource_access_tracking (a_pool: ROSE_DATABASE_POOL) is
			-- Clear out any resource access that's happened up until now
		do
			if a_pool /= Void then
				a_pool.accessed.wipe_out
			end
			global.file_access_log.clear_all
		end

feature {ROSE_DATABASE_POOL} -- Implementation

	persist_for_resource_access_tracking (a_pool: ROSE_DATABASE_POOL) is
			-- Record resource accesses for a_pool.
		local
			retried: BOOLEAN
			l_create_instance_predicate, l_create_run_predicate: EXECUTION_PREDICATE
			l_process_id, l_instance_id, l_run_id: INTEGER
			l_server: SQL_DATA_SERVER
			l_date, l_end_date: ROSE_DATE
		do
			if not retried then
				if a_pool.accessed.count > 0 and then a_pool.stamp > 0 then
					a_pool.already_warned.put (True, "process_metrics")
					l_server ?= a_pool.get_server_from_logical ("process_metrics")
				end
				if l_server /= Void then
					l_process_id := get_integer_result (l_server, "SELECT process_id FROM process WHERE name = '" + resource_access_tracking_string + "'", Void)
					if l_process_id > 0 then
						create l_date.make_now_utc
						create l_create_instance_predicate.make (l_server, "INSERT INTO instance (process_id,partition_key,file_date) VALUES (?,?,?)")
						l_create_instance_predicate.execute (Void, <<l_process_id, global.app.program_name, l_date>>)
						if not l_create_instance_predicate.success then
							l_date.minute_add (1)
							l_date.second_add (30)
							l_create_instance_predicate.execute (Void, <<l_process_id, global.app.program_name, l_date>>)
						end
						if l_create_instance_predicate.success then
							l_instance_id := get_integer_result (l_server, "SELECT cast (@@IDENTITY AS INT) as [id]", Void)
							if l_instance_id > 0 then
								create l_date.make_fine (
									global.app.start_time.year,
									global.app.start_time.month,
									global.app.start_time.day,
									global.app.start_time.hour,
									global.app.start_time.minute,
									global.app.start_time.second
								)
								l_date.second_add (-global.env.utc_seconds_offset)
								create l_end_date.make_now_utc
								if l_server.has_column_name ("instance_run", "changelist") then
									create l_create_run_predicate.make (l_server, insert_instance_run_sql)
									l_create_run_predicate.execute (Void, <<l_instance_id, l_date,
										l_end_date, 1, global.app.pid, global.app.version, global.app.program_arguments,
										global.env.job_id, global.env.scheduler, global.app.info.revision>>)
								else
									create l_create_run_predicate.make (l_server, insert_instance_run_sql_old)
									l_create_run_predicate.execute (Void, <<l_instance_id, l_date,
										l_end_date, 1, global.app.pid, global.app.version, global.app.program_arguments,
										global.env.job_id, global.env.scheduler>>)
								end
								if l_create_run_predicate.success then
									l_run_id := get_integer_result (l_server, "SELECT @@IDENTITY AS id", Void)
									if l_run_id > 0 then
										persist_data_server_accesses (a_pool, l_server, l_run_id)
									end
								end
							end
						end
					end
				end
			end
		rescue
			retried := True
			retry
		end

feature {PM_CONTEXT} -- Implementation

	persist_data_server_accesses (a_pool: ROSE_DATABASE_POOL; a_server: DATA_SERVER; a_run_id: INTEGER) is
			-- Store accessed resources (files and databases) to the process metrics DB
		local
			l_predicate: EXECUTION_PREDICATE
		do
			create l_predicate.make (a_server, "INSERT INTO resource_access (type_id, run_id, path, logical) VALUES (?, ?, ?, ?)")
			persist_data_server_pool_accesses (a_pool, a_server, l_predicate, a_run_id)
			persist_resource_accesses (global.file_access_log.file_in_access_log, input_file_type_id, l_predicate, a_run_id)
			persist_resource_accesses (global.file_access_log.file_out_access_log, output_file_type_id, l_predicate, a_run_id)
			persist_storable_accesses (global.file_access_log.storable_in_access_log, input_storable_type_id, l_predicate, a_run_id)
			persist_storable_accesses (global.file_access_log.storable_out_access_log, output_storable_type_id, l_predicate, a_run_id)
			persist_resource_accesses (global.file_access_log.persistent_server_in_access_log, input_persistent_server_type_id, l_predicate, a_run_id)
			persist_resource_accesses (global.file_access_log.persistent_server_out_access_log, output_persistent_server_type_id, l_predicate, a_run_id)
			if not a_pool.force_transactions_off then
				a_server.commit
				if not a_server.success then
					error.warn ("The resource tracking commit failed: path='$1', logical='$4'", <<a_server.definition.full_path, a_server.definition.logical>>)
				end
			end
		end

	persist_data_server_pool_accesses (a_pool: ROSE_DATABASE_POOL; a_server: DATA_SERVER; a_predicate: EXECUTION_PREDICATE; a_run_id: INTEGER) is
			-- Store accessed resources (files and databases) to the process metrics DB
		local
			l_accessed: ROSE_LINEAR_ARRAY [ROSE_DATABASE_DEFINITION]
			l_definition: ROSE_DATABASE_DEFINITION
			l_database_type, i: INTEGER
			l_database_table_type: INTEGER
			l_server: SQL_DATA_SERVER
		do
			from
				l_accessed := a_pool.accessed
				i := l_accessed.lower
			until
				i > l_accessed.upper
			loop
				l_definition := l_accessed.item (i)
				if l_definition.is_written then
					l_database_type := output_database_type_id
				else
					l_database_type := input_database_type_id
				end
				persist_data_server_pool_access (a_predicate, l_database_type, a_run_id, l_definition.full_path, l_definition.logical)
				if attached l_definition.input_table_names as l_table_names then
					l_database_table_type := input_database_table_type_id
					l_table_names.do_all (agent persist_database_table_access (a_predicate, l_database_table_type, a_run_id, l_definition.full_path, l_definition.logical, ?))
				end
				if attached l_definition.output_table_names as l_table_names then
					l_database_table_type := output_database_table_type_id
					l_table_names.do_all (agent persist_database_table_access (a_predicate, l_database_table_type, a_run_id, l_definition.full_path, l_definition.logical, ?))
				end
				if a_predicate.success
					and then l_definition.is_written
						and then l_definition.logical /= Void
				 then
					 l_server := servers.shared_server_from_repository (l_definition.logical)
					if l_server /= Void then
						-- Check to see if a production database was updated from outside of production
						if not Global.app.is_running_in_production and then l_server.is_production then
							error.warn ("The production database '$1' was updated $2 times by this non-production process", <<l_accessed.item (i).full_path.as_lower, l_accessed.item (i).execution_count>>)
						end
						-- Check to see if a development database was updated from inside of production
						if Global.app.is_running_in_production and then not l_server.is_production then
							inform ("persist_resource_accesses", "The development database '$1' was updated $2 times by this production process", <<l_accessed.item (i).full_path.as_lower, l_accessed.item (i).execution_count>>)
						end
					end
				end
				i := i + 1
			end
		end

	persist_resource_accesses (a_log: DS_HASH_TABLE [ANY, STRING]; a_type_id: INTEGER; a_predicate: EXECUTION_PREDICATE; a_run_id: INTEGER) is
			-- Persist the collection in 'a_log' for 'a_type_id'
		require
			valid_log: a_log /= Void
			valid_predicate: a_predicate /= Void
			valid_type_id: resource_type_ids.has (a_type_id)
		local
			l_logical: STRING
			l_args: ARRAY[ANY]
		do
			from
				a_log.start
			until
				a_log.off
			loop
				l_logical ?= a_log.item_for_iteration
				if l_logical /= Void then
					l_args := <<a_type_id, a_run_id, a_log.key_for_iteration, l_logical>>
				else
					l_args := <<a_type_id, a_run_id, a_log.key_for_iteration, Void>>
				end
				a_predicate.execute (Void, l_args )
				if not a_predicate.success then
					error.warn ("The resource tracking insert failed: type_id='$1', run_id='$2', path='$3', logical='$4'", l_args )
				end
				a_log.forth
			end
		end

	persist_storable_accesses (a_log: DS_HASH_TABLE [ROSE_LINEAR_ARRAY [STRING], STRING]; a_type_id: INTEGER; a_predicate: EXECUTION_PREDICATE; a_run_id: INTEGER) is
			-- Persist the collection in 'a_log' for 'a_type_id'
		require
			valid_log: a_log /= Void
			valid_predicate: a_predicate /= Void
			valid_type_id: resource_type_ids.has (a_type_id)
		local
			l_log: ROSE_LINEAR_ARRAY [STRING]
		do
			from
				a_log.start
			until
				a_log.off
			loop
				l_log := a_log.item_for_iteration
				from l_log.start until l_log.off loop
					a_predicate.execute (Void, <<a_type_id, a_run_id, l_log.current_item, a_log.key_for_iteration>>)
					if not a_predicate.success then
						error.warn ("The resource tracking insert failed: type_id='$1', run_id='$2', path='$3', logical='$4'", <<a_type_id, a_run_id, l_log.current_item, a_log.key_for_iteration>>)
					end
					l_log.forth
				end
				a_log.forth
			end
		end

	persist_data_server_pool_access (a_predicate: EXECUTION_PREDICATE; a_database_type: INTEGER; a_run_id: INTEGER; a_full_path: STRING; a_logical: STRING)
			-- Store accessed databases to the process metrics DB
		require
			a_predicate_not_void: a_predicate /= Void
			a_full_path_not_void: a_full_path /= Void
			a_logical_not_void: a_logical /= Void
		do
			a_predicate.execute (Void, <<a_database_type, a_run_id, a_full_path, a_logical>>)
			if not a_predicate.success then
				error.warn ("The resource tracking insert failed: type_id='$1', run_id='$2', path='$3', logical='$4'", <<a_database_type, a_run_id, a_full_path, a_logical>>)
			end
		end

	persist_database_table_access (a_predicate: EXECUTION_PREDICATE; a_database_table_type: INTEGER; a_run_id: INTEGER; a_full_path: STRING; a_logical: STRING; a_table_name: STRING)
			-- Store accessed database tables to the process metrics DB
		require
			a_predicate_not_void: a_predicate /= Void
			a_full_path_not_void: a_full_path /= Void
			a_logical_not_void: a_logical /= Void
			a_table_name_not_void: a_table_name /= Void
		local
			l_logical: STRING
			l_full_path: STRING
		do
			l_logical := a_logical + "." + a_table_name
			l_full_path := a_full_path + "." + a_table_name
			a_predicate.execute (Void, <<a_database_table_type, a_run_id, l_full_path, l_logical>>)
			if not a_predicate.success then
				error.warn ("The resource tracking insert failed: type_id='$1', run_id='$2', path='$3', logical='$4'", <<a_database_table_type, a_run_id, l_full_path, l_logical>>)
			end
		end

	resource_type_ids: ROSE_LINEAR_ARRAY [INTEGER] is
			-- Collection of type ids available for monitoring
		once
			create Result.make_empty
			Result.extend (input_database_type_id)
			Result.extend (output_database_type_id)
			Result.extend (input_file_type_id)
			Result.extend (output_file_type_id)
			Result.extend (input_storable_type_id)
			Result.extend (output_storable_type_id)
			Result.extend (input_persistent_server_type_id)
			Result.extend (output_persistent_server_type_id)
		end

	input_database_type_id: INTEGER is 11
	output_database_type_id: INTEGER is 12
	input_file_type_id: INTEGER is 21
	output_file_type_id: INTEGER is 22
	input_storable_type_id: INTEGER is 31
	output_storable_type_id: INTEGER is 32
	input_persistent_server_type_id: INTEGER is 41
	output_persistent_server_type_id: INTEGER is 42
	input_database_table_type_id: INTEGER is 51
	output_database_table_type_id: INTEGER is 52

	get_integer_result (a_server: DATA_SERVER; a_sql: STRING; a_args: ARRAY [ANY]): INTEGER is
			-- Integer returned by executing 'a_sql' statement on 'a_server' with 'a_args'
		local
			l_predicate: INTEGER_PREDICATE
		do
			create l_predicate.make_strict (a_server, a_sql)
			l_predicate.retrieve (Void, a_args)
			if l_predicate.success and then not l_predicate.array.is_empty then
				Result := l_predicate.array.first
			end
		end

	insert_instance_run_sql: STRING =
			-- Insert instance_run object into instance_run table in process_metrics db.
"{
	INSERT INTO instance_run
	(instance_id,start_time,end_time,exit_code,host_pid,version,args,is_utc,scheduler_job_id,scheduler_name,from_resource_access_tracking,changelist)
	VALUES (?,?,?,?,?,?,?,1,?,?,1,?)
}"

	insert_instance_run_sql_old: STRING =
			-- Insert instance_run object into instance_run table in process_metrics db without changelist column.
"{
	INSERT INTO instance_run
	(instance_id,start_time,end_time,exit_code,host_pid,version,args,is_utc,scheduler_job_id,scheduler_name,from_resource_access_tracking)
	VALUES (?,?,?,?,?,?,?,1,?,?,1)
}"

end
