note

	description:

		"Extended rich-text widget"

	author: "Zoran Simic"
	last_editor: "$Author: an $"
	date: "$Date: 2011/12/23 $"
	revision: "$Revision: #14 $"
	path: "$File: //rose/source/kernel/infrastructure/widgets/rose_rich_text.e $"

class
	ROSE_RICH_TEXT

inherit
	EV_RICH_TEXT
		export
			{ROSE_RICH_TEXT_RTF, EV_ANY, EV_ANY_I} implementation
		redefine
			initialize, flush_buffer, set_font
		end

create
	default_create, make

feature {NONE} -- Initialization

	make (a_container: EV_CONTAINER) is
			-- Create and put in 'a_container' if provided
		do
			default_create
			if a_container /= Void then
				a_container.extend (Current)
			end
		end

	initialize is
			-- Called by default_create
		do
			Precursor
			key_release_actions.extend (agent on_key_released)
		end

feature -- Access

	maximum_lines: INTEGER
			-- Maximum number of lines to keep
			-- When 0, no maximum is imposed
			-- When >0, the output is truncated to the last 'maximum_lines' appended

	rtf_text: STRING is
			-- Text contained in this widget, represented in RTF
		do
			Result := rtf_processor.rtf_text (Current)
		end

feature -- Basic operations

	set_maximum_lines (a_maximum_lines: INTEGER) is
			-- Set 'maximum_lines'.
		do
			maximum_lines := a_maximum_lines
		end

	set_rtf_text (a_text: STRING) is
			-- Display `a_text' in RTF format.
		do
			if a_text = Void then
				remove_text
			else
				rtf_processor.set_rtf_text (Current, a_text)
			end
		end

	append_line (a_text: STRING) is
			-- Append 'a_text', respecting 'maximum_lines'.
		do
			append_formatted (a_text, Void, Void, Void)
		end

	append (a_text: STRING; a_color: EV_COLOR) is
			-- Append 'a_text' with 'a_color'
		do
			append_formatted (a_text, Void, Void, a_color)
		end

	append_bold (a_text: STRING; a_color: EV_COLOR) is
			-- Append 'a_text' in bold and with 'a_color' using 'format'.
		local
			l_font: EV_FONT
		do
			if a_text /= Void and then not a_text.is_empty then
				l_font := font
				l_font.set_weight (font_constants.weight_bold)
				append_formatted (a_text, l_font, Void, a_color)
			end
		end

	append_formatted (a_text: STRING_GENERAL; a_font: EV_FONT; a_background_color, a_color: EV_COLOR) is
			-- Append 'a_text' with 'a_font', 'a_background_color', 'a_color' and the rest of the settings from 'format'.
		local
			l_format: EV_CHARACTER_FORMAT
		do
			if a_text /= Void and then not a_text.is_empty then
					-- There is no `prune_code' in STRING_GENERAL, hence
					-- multi-branch if-instruction below.
				if attached {STRING_8} a_text as l_string_8 then
					l_string_8.prune_all ('%R')
				elseif attached {STRING_32} a_text as l_string_32 then
					l_string_32.prune_all ('%R')
				end
				if a_font /= Void or else a_background_color /= Void or else a_color /= Void then
					if a_font /= Void then
						create l_format.make_with_font (a_font)
					elseif saved_font /= Void then
						create l_format.make_with_font (saved_font)
					else
						create l_format
					end
					if a_background_color /= Void then
						l_format.set_background_color (a_background_color)
					end
					if a_color /= Void then
						l_format.set_color (a_color)
					end
					set_current_format (l_format)
				end
				append_text (a_text)
				truncate_lines
				set_caret_position (text_length + 1)
				if maximum_lines > 0 then
					-- There should be an option for 'auto-scroll to bottom'
					scroll_to_end
				end
			end
		end

	append_status (a_status: ROSE_STATUS) is
			-- Append 'a_status'
		do
			if a_status /= Void and then a_status.messages /= Void then
				a_status.messages.do_all (agent append_message)
			end
		end

	append_message (a_message: ROSE_STATUS_MESSAGE) is
			-- Append colored text for 'a_message'
		do
			if a_message /= Void and then a_message.text /= Void then
				append_formatted (a_message.text + "%N", Void, Void, message_fg (a_message))
			end
		end

	scroll_at_beginning is
			-- Show text at beginning.
			-- This is needed because for some reason EV_RICH_TEXT does a poor job when initially showing the contents...
		do
			if line_count >= 1 then
				set_caret_position (1)
				scroll_to_line (1)
			end
			deselect_all
		end

	flush_buffer is
			-- Flush contents of buffer and unlock any set buffering mode.
		local
			l_was_editable: BOOLEAN
		do
			if not is_editable then
				enable_edit
				l_was_editable := True
			end
			Precursor
			if l_was_editable then
				disable_edit
			end
		end

	set_word_wrap (a_on: BOOLEAN) is
			-- Set 'has_word_wrapping'.
		do
			if a_on then
				enable_word_wrapping
			else
				disable_word_wrapping
			end
		end

	set_font (a_font: EV_FONT)
			-- Assign `a_font' to font.
		do
			Precursor (a_font)
			saved_font := a_font
		end

	save_rtf (a_filename: STRING) is
			-- Save contents of this text widget in RTF format to file with 'a_filename'.
		do
			rtf_processor.save_rtf (Current, a_filename)
		end

feature {NONE} -- Implementation

	saved_font: EV_FONT
			-- Last font passed to 'set_font'

	font_constants: EV_FONT_CONSTANTS is
			-- Font constants
		once
			create Result
		end

	key_constants: EV_KEY_CONSTANTS is
			-- Key constants
		once
			create Result
		end

	rtf_processor: ROSE_RICH_TEXT_RTF is
			-- RTF processor
		once
			create Result
		end

	truncate_lines is
			-- Make sure that no more than 'maximum_lines' are displayed
		local
			l_line_count: INTEGER
			l_was_editable: BOOLEAN
		do
			if maximum_lines > 0 then
				l_line_count := line_count
				if l_line_count > maximum_lines then
					l_was_editable := is_editable
					if not l_was_editable then
						enable_edit
					end
					select_lines (1, l_line_count - maximum_lines)
					delete_selection
					if not l_was_editable then
						disable_edit
					end
				end
			end
		end

	on_key_released (a_key: EV_KEY) is
			-- Called when 'a_key' released while focus in this widget
		local
			l_parent: EV_CONTAINER
			l_dialog: EV_DIALOG
			l_button: EV_BUTTON
		do
			if a_key.code = key_constants.key_escape or else a_key.code = key_constants.key_enter then
				from
					l_parent := parent
				until
					l_parent = Void
				loop
					l_dialog ?= l_parent
					l_parent := l_parent.parent
				end
				if l_dialog /= Void then
					if a_key.code = key_constants.key_enter then
						l_button := l_dialog.default_push_button
					else
						l_button := l_dialog.default_cancel_button
					end
					if l_button /= Void then
						l_button.select_actions.call (Void)
					end
				end
			end
		end

	convention: ROSE_FRAMEWORK_CONVENTION is
			-- Framework conventions
		once
			create Result
		end

	stock_colors: EV_STOCK_COLORS is
			-- Stock colors
		once
			create Result
		end

	message_fg (a_message: ROSE_STATUS_MESSAGE): EV_COLOR is
			-- Foreground color that represents graphically 'a_message'
		do
			if a_message = Void then
				-- No color
			elseif a_message.level = convention.v_error then
				Result := stock_colors.red
			elseif a_message.level = convention.v_warning then
				Result := stock_colors.dark_red
			elseif a_message.level = convention.v_output then
				Result := stock_colors.black
			end
		end

end
