indexing
	description: "[
				Selectable area within a pixmap

				-- The used rectangle looks like this:

				 (x1/y1)------|
				    |         |
				    |         |
				    |         |
				    |------(x2/y2)

				 The corners of the rectangle and the resize areas are indexed as follows:

				  1----5----2
				  |         |
				  8         6
				  |         |
				  4----7----3

			]"


	author: "Zoran Simic"
	date: "$Date: 2009/10/06 $"
	revision: "$Revision: #14 $"

class
	ROSE_SELECTION_AREA

inherit
	ROSE_SHARED_GLOBAL_GUI
		export
			{NONE} all
		end

	DOUBLE_MATH
		export
			{NONE} all
		end
	ANY
create
	make,
	make_proportional

feature {NONE} -- Initialization

	make is
			-- Create object
		do
			max_index := 8
			is_proportional := False

			initialize
		end

	make_proportional is
			-- Create object, limit movement to be proportional to selectable_area
		do
			max_index := 4
			is_proportional := True

			initialize
		end

	initialize is
			-- Initialize the object
		local
			i: INTEGER
			l_rect: EV_MODEL_RECTANGLE
		do
			set_ok_color (Gui.rep.colors.black)
			set_not_ok_color (Gui.rep.colors.red)
			set_resize_icon_size (10)
			set_ok_width (2 * resize_icon_size)
			set_ok_height (2 * resize_icon_size)
			create holder
			create selected_model.make_rectangle(0, 0, 0, 0)
			holder.extend (selected_model)
			create resize_icons_models.make (1, max_index)
			from i := resize_icons_models.lower until i > resize_icons_models.upper
			loop
				l_rect := create {EV_MODEL_RECTANGLE}.default_create
				resize_icons_models.put (l_rect, i)
				holder.extend (l_rect)
				i := i + 1
			end
			create left_corner_text.default_create
			holder.extend (left_corner_text)
			create right_corner_text.default_create
			holder.extend (right_corner_text)
		end


feature -- Access

	holder: EV_MODEL_GROUP
		-- Holder

	selected_model: EV_MODEL_RECTANGLE
	resize_icons_models: ARRAY[EV_MODEL_RECTANGLE]
	left_corner_text: EV_MODEL_TEXT
	right_corner_text: EV_MODEL_TEXT

	pixmap: EV_PIXMAP
			-- Associated pixmap

	ok_color, not_ok_color: EV_COLOR
			-- Colors to use to show selection area
			-- the 'not ok' color is used to indicate that selection is not big enough

	ok_width, ok_height: INTEGER
			-- Minimum dimension of selected area for selection to be valid

	resize_icon_size: INTEGER
			-- Size of the icons for resizing

	selectable_area: ROSE_RECTANGLE
			-- Part of 'pixmap' that is to be considered selectable

	selected: ROSE_RECTANGLE
			-- Current selection (Void if none)

	normal_selected: ROSE_RECTANGLE
			-- Normal selection (part of the normal rectangle)
			-- This represents the relative part of 'selectable_area' that is selected

	is_selecting: BOOLEAN
			-- Are we currently selecting?

	is_dragging: BOOLEAN
			-- Are we currently dragging the selected area?

	is_within_selectable (a_x, a_y: INTEGER): BOOLEAN is
			-- Is point ('a_x','a_y') within 'selectable_area'?
		do
			Result := selectable_area.has_point (a_x,a_y)
		end

	is_within_selected (a_x, a_y: INTEGER): BOOLEAN is
			-- Is point ('a_x','a_y') within the selected area?
		do
			Result := selected /= Void and then selected.has_point (a_x,a_y)
		end

	is_within_resize_icon (a_x, a_y: INTEGER): BOOLEAN is
			-- Is point ('a_x','a_y') within the resize icon?
		local
			i: INTEGER
			l_rect: ROSE_RECTANGLE
		do
			if resize_icons /= Void then
				from
					i := resize_icons.lower
				until
					i > resize_icons.upper or Result = True
				loop
					l_rect := resize_icons.item (i)
					if l_rect /= Void then
						Result := l_rect.has_point (a_x, a_y)
						if Result = True then
							resize_index := i
						end
					end
					i := i + 1
				end
			end
		end

	resize_index: INTEGER
			-- Index of currently resizing border

	max_index: INTEGER
			-- Maximal index (usually: 8; proportional: 4)

	is_proportional: BOOLEAN
			-- Is the area proporitional (to the selectable area)


feature -- Basic operations

	reset is
			-- Reset current selection (forget about it).
		do
			is_drawn := False
			remove_selected
		end

	remove_selected is
			-- Remove current selection.
		do
			draw_selected (False)
			selected_model.hide
			left_corner_text.hide
			right_corner_text.hide

			is_drawn := False
			selected := Void
			resize_icons := Void
			normal_selected := Void
			is_selecting := False
			is_dragging := False
		end

	start_resizing (a_x, a_y: INTEGER) is
			-- Start resizing at ('a_x','a_y').
		do
			is_selecting := True
			remove_resize_icons

			offset_x := a_x - x_at (resize_index).truncated_to_integer
			offset_y := a_y - y_at (resize_index).truncated_to_integer
		end

	resize_cursor: EV_CURSOR is
			-- Get the pointer style for the current resize direction
		require
			resize_index /= 0
		do
			if resize_index = 1 or resize_index = 3 then
				Result :=  Gui.rep.pixmaps.sizenwse_cursor
			elseif resize_index = 2 or resize_index = 4 then
				Result :=  Gui.rep.pixmaps.sizenesw_cursor
			elseif resize_index = 5 or resize_index = 7 then
				Result :=  Gui.rep.pixmaps.sizens_cursor
			else
				Result :=  Gui.rep.pixmaps.sizewe_cursor
			end
		end


	start_selection (a_x, a_y: INTEGER) is
			-- Start selection at ('a_x','a_y').
		do
			remove_selected
			is_selecting := True
			x1 := a_x
			y1 := a_y
			x2 := a_x
			y2 := a_y

			resize_index := 3

			offset_x := 0
			offset_y := 0
		end

	extend_selection (a_x, a_y: INTEGER) is
			-- Extend selection to ('a_x','a_y').
		do
			if is_selecting then
				draw_selected (False)
				extend_coordinates (a_x - offset_x ,a_y - offset_y)
				draw_selected (True)
			end
		end

	finalize_selection is
			-- Finalize selection at ('a_x','a_y').
		do
			if is_selecting then
				is_selecting := False
				if big_enough then
					create_selected
				else
					remove_selected
				end
			end
		end

	grab_selected (a_x, a_y: INTEGER) is
			-- Grab selected area at ('a_x','a_y').
		do
			if selected /= Void then
				old_x := a_x
				old_y := a_y
				is_dragging := True

				remove_resize_icons
			end
		end

	drag_selected (a_x, a_y: INTEGER) is
			-- Drag selected area to ('a_x','a_y').
		local
			l_x, l_y, l_w, l_h, i: INTEGER
		do
			if is_dragging then
				l_x := (x1.min (x2) + a_x - old_x).max (selectable_area.x1.truncated_to_integer)
				l_y := (y1.min (y2) + a_y - old_y).max (selectable_area.y1.truncated_to_integer)
				l_w := selected.width.truncated_to_integer - 1
				l_h := selected.height.truncated_to_integer - 1
				i := selectable_area.x2.truncated_to_integer
				if l_x + l_w > i then
					l_x := i - l_w
				end
				i := selectable_area.y2.truncated_to_integer
				if l_y + l_h > i then
					l_y := i - l_h
				end
				draw_selected (False)
				x1 := l_x
				y1 := l_y
				x2 := l_x + l_w
				y2 := l_y + l_h
				old_x := a_x
				old_y := a_y


				setup_model

				draw_selected (True)
			end
		end

	stop_dragging is
			-- Stop dragging selected area.
		do
			if is_dragging then
				is_dragging := False
				create_selected
			end
		end

	set_pixmap (a_pixmap: EV_PIXMAP) is
		do
			pixmap := a_pixmap
		end

	set_selectable_area (a_area: ROSE_RECTANGLE) is
		do
			selectable_area := a_area
		end

	redraw is
		do
			if is_drawn then
				draw_selected (True)
			end
		end

	set_ok_width (a_size: INTEGER) is
			-- Set 'ok_width'.
		do
			ok_width := a_size
		end

	set_ok_height (a_size: INTEGER) is
			-- Set 'not_ok_width'.
		do
			ok_height := a_size
		end

	set_resize_icon_size (a_size: INTEGER) is
			-- Set 'resize_icon_size'.
		do
			resize_icon_size := a_size
		end

	set_ok_color (a_color: EV_COLOR) is
			-- Set 'ok_color'.
		require
			valid_color: a_color /= Void
		do
			ok_color := a_color
		end

	set_not_ok_color (a_color: EV_COLOR) is
			-- Set 'ok_color'.
		require
			valid_color: a_color /= Void
		do
			not_ok_color := a_color
		end

feature {NONE} -- Implementation

	is_drawn: BOOLEAN
	x1, y1, x2, y2: INTEGER
	old_x, old_y: INTEGER
	offset_x, offset_y: INTEGER
		-- Offset of mouse pointer when resizing

	width: INTEGER is
		do
			Result := (x2 - x1).abs
		end

	height: INTEGER is
		do
			Result := (y2 - y1).abs
		end

	extend_coordinates (a_x, a_y: INTEGER) is
		local
			l_x, l_y: INTEGER
		do
			l_x := a_x.max (selectable_area.x1.truncated_to_integer).min (selectable_area.x2.truncated_to_integer)
			l_y := a_y.max (selectable_area.y1.truncated_to_integer).min (selectable_area.y2.truncated_to_integer)

			set_x_at (l_x, resize_index)
			set_y_at (l_y, resize_index)

			if is_proportional then
				if (width / height) < width_height_proportion then
					set_x_proportional (resize_index)
				else
					set_y_proportional (resize_index)
				end
			end

			setup_model
		end

	big_enough: BOOLEAN is
		do
			Result := width >= ok_width and then height >= ok_height
		end

	last_color: EV_COLOR

	create_selected is
			-- Create the 'selected' and all resize-rectangles
		local
			i: INTEGER
			l_x, l_y: DOUBLE
			l_x1, l_x2, l_y1, l_y2: DOUBLE
			l_w: DOUBLE
			l_rect: ROSE_RECTANGLE
		do
			draw_selected (False)
			arrange_coordinates
			create selected.make (x1,y1,x2,y2)
			normal_selected := selectable_area.normalized_rectangle (selected)

			l_w := resize_icon_size / 2

			create resize_icons.make (1, max_index)
			from
				i := resize_icons.lower
			until
				i > resize_icons.upper
			loop
				if i <= 4 then -- Corner case
					l_x := x_at (i)
					l_y := y_at (i)
				else -- Border case
					l_x1 := x_at ((i - 5) \\ 4 + 1)
					l_x2 := x_at ((i - 4) \\ 4 + 1)

					l_y1 := y_at ((i - 5) \\ 4 + 1)
					l_y2 := y_at ((i - 4) \\ 4 + 1)

					l_x := (l_x1 + l_x2) / 2
					l_y := (l_y1 + l_y2) / 2
				end

				create l_rect.make (l_x - l_w, l_y - l_w, l_x + l_w, l_y + l_w)
				resize_icons.put (l_rect, i)

				i := i + 1
			end

			resize_icons.do_all_with_index (agent position_resize_icon (?, ?))

			draw_selected (True)
		end

	position_resize_icon (a_rect: ROSE_RECTANGLE; a_index: INTEGER) is
			-- set the position of the resize icon
		require
			rectangle_not_void : a_rect /= void
		do
			resize_icons_models.item (a_index).set_point_a_position (a_rect.x1.truncated_to_integer, a_rect.y1.truncated_to_integer)
			resize_icons_models.item (a_index).set_point_b_position (a_rect.x2.truncated_to_integer, a_rect.y2.truncated_to_integer)
			resize_icons_models.item (a_index).show
		end

	resize_icons: ROSE_LINEAR_ARRAY[ROSE_RECTANGLE]


	x_at (a_index: INTEGER): DOUBLE is
			-- Get the x-coordinate of corner 'a_index'
		do
			if selected /= Void then
				if a_index = 1 or a_index = 4 or a_index = 8 then
					Result := selected.x1
				else
					Result := selected.x2
				end
			end
		end

	y_at (a_index: INTEGER): DOUBLE is
			-- Get the y-coordinate of corner 'a_index'
		do
			if selected /= Void then
				if a_index = 1 or a_index = 2 or a_index = 5 then
					Result := selected.y1
				else
					Result := selected.y2
				end
			end
		end

	set_x_at (a_x: INTEGER; a_index: INTEGER) is
			-- Set the x-coordinate of corner 'a_index'
		require
			valid_index (a_index)
		do
			if a_index = 1 or a_index = 4 or a_index = 8 then
				x1 := a_x
			elseif a_index = 2 or a_index = 3 or a_index = 6 then
				x2 := a_x
			else -- Don't set x on index 5 and 7
				do_nothing
			end
		end

	set_y_at (a_y: INTEGER; a_index: INTEGER) is
			-- Set the x-coordinate of corner 'a_corner'
		require
			valid_index (a_index)
		do
			if a_index = 1 or a_index = 2 or a_index = 5 then
				y1 := a_y
			elseif a_index = 3 or a_index = 4 or a_index = 7 then
				y2 := a_y
			else -- Don't set y on index 6 and 8
				do_nothing
			end
		end

	set_x_proportional (a_index: INTEGER) is
			-- Set x at `a_index' in such a way that the area is proportional
		require
			valid_index (a_index)
		local
			l_switch_sign: BOOLEAN
		do
			l_switch_sign := (x2 - x1) / (y2 - y1) > 0
			if a_index = 1 or a_index = 4 then
				if l_switch_sign then
					x1 := x2 - proportional_width
				else
					x1 := x2 + proportional_width
				end
			elseif a_index = 2 or a_index = 3 then
				if l_switch_sign then
					x2 := x1 + proportional_width
				else
					x2 := x1 - proportional_width
				end
			end
		end

	set_y_proportional (a_index: INTEGER) is
			-- Set y at `a_index' in such a way that the area is proportional
		require
			valid_index (a_index)
		local
			l_switch_sign: BOOLEAN
		do
			l_switch_sign := (x2 - x1) / (y2 - y1) > 0
			if a_index = 1 or a_index = 2 then
				if l_switch_sign then
					y1 := y2 - proportional_height
				else
					y1 := y2 + proportional_height
				end
			elseif a_index = 4 or a_index = 3 then
				if l_switch_sign then
					y2 := y1 + proportional_height
				else
					y2 := y1 - proportional_height
				end
			end
		end

	width_height_proportion: DOUBLE is
			-- Proportion width/height of selectable_area
		require
			selectable_area /= Void
		do
			Result := selectable_area.width / selectable_area.height
		end

	proportional_width: INTEGER is
			-- Return width that makes the area proportional, given a fixed height
		do
			Result := (width_height_proportion * (y2 - y1)).rounded
		end

	proportional_height: INTEGER is
			-- Return height that makes the area proportional, given a fixed width
		do
			Result := ((x2 - x1) / width_height_proportion).rounded
		end



	draw_selected (a_original: BOOLEAN) is
		do
			if a_original or else is_drawn then
				if big_enough then
					selected_model.set_foreground_color (ok_color)
				else
					selected_model.set_foreground_color (not_ok_color)
				end
				selected_model.show
				left_corner_text.show
				right_corner_text.show
			else
				resize_icons_models.do_all (agent hide_rect (?))
			end
		end

	hide_rect (a_rect: EV_MODEL_RECTANGLE) is 
			-- Helper agent {EV_MODEL_RECTANGLE}.hide
		require
			a_rect_not_void: a_rect /= Void
		do
			a_rect.hide 
		end

	arrange_coordinates is
			-- Make sure that x1 <= x2 and y1 <= y2.
		local
			i: INTEGER
		do
			if x2 < x1 then
				i := x1
				x1 := x2
				x2 := i
			end
			if y2 < y1 then
				i := y1
				y1 := y2
				y2 := i
			end
		end

	remove_resize_icons is
			-- Erase the drawn resize rectangles
		do
			draw_selected (False)
			resize_icons := Void
			draw_selected (True)
		end

	setup_model is
			-- Position all the figures
		do
				selected_model.set_point_a_position (x1, y1)
				selected_model.set_point_b_position (x2, y2)
		end

	valid_index (a_index: INTEGER): BOOLEAN is
			-- Is `a_index' a valid index?
		do
			Result := a_index > 0 and a_index <= max_index
		end


end -- class ROSE_SELECTION_AREA
