note

	description:

		"Serialize an eiffel object in XML"

	notice: "rose_serializer.dtd"
	author: "Zoran Simic"
	last_editor: "$Author: er $"
	date: "$Date: 2012/10/26 $"
	revision: "$Revision: #18 $"
	path: "$File: //rose/source/kernel/infrastructure/framework/serialization/windows/classic/rose_serializer.e $"

class
	ROSE_SERIALIZER

inherit

	ANY -- Export features from ANY

	ROSE_SERIALIZATION_CONSTANTS
		export
			{NONE} all
		end

feature -- Access

	object_count: INTEGER
			-- Number of objects stored during last serialization

	type_count: INTEGER
			-- Numbewr of different types referenced via last serialized object

feature -- Basic operations

	serialize (a_object: ANY; a_file_name: STRING) is
			-- Serialize 'a_object' onto 'a_file_name'.
			-- Eventual errors are signaled
		require
			valid_object: a_object /= Void
			valid_channel: a_file_name /= Void and then not a_file_name.is_empty
		local
			l_filename: STRING
			rescued: STRING
			l_collecting: BOOLEAN
		do
			reset
			if rescued /= Void then
				error.signal (generator,"Could not serialize '$1': $2",<<a_object,rescued>>)
			else
				l_filename := global.persist.write_storable_filename (a_file_name)
				create current_file.make (l_filename)
				if not current_file.exists or else current_file.is_writable then
					current_file.open_write
				else
					error.signal (generator,"Cannot serialize to '$1', not writable",<<l_filename>>)
				end
			end
			if not error.signaled then
				current_file.put_string (k_xml_header)
				create types.make (0)
				create refs.make (0)
				l_collecting := memory.collecting
				memory.collection_off
				build_ref (a_object)
				start_element (k_serialized,Void,k_revision,0,0)
					--| Output header with type ids
				output_types
					--| Output the object structure
				output_fields (Void,a_object,1)
				if l_collecting then
					memory.collection_on
				end
				terminate_element (k_serialized)
				current_file.flush
				current_file.close
				object_count := refs.count
				type_count := types.count
			end
			refs := Void
			types := Void
			current_file := Void
		rescue
			if l_collecting then
				memory.collection_on
			end
			if rescued = Void then
				rescued := last_exception_explanation
				retry
			end
		end

	reset is
			-- Clear this serializer's stats regarding last serialization.
		do
			object_count := 0
			type_count := 0
		end

feature {NONE} -- Serialization

	build_fields (a_object: ANY) is
			-- Build the 'types' and 'refs' table for 'a_object'.
		local
			i, n: INTEGER
		do
			n := field_count (a_object)
			if n > 0 then
				from
					i := 1
				until
					i > n
				loop
					if field_type (i,a_object) = Reference_type then
						build_ref (field (i,a_object))
					end
					i := i + 1
				end
			end
		end

	build_special (a_object: ANY) is
			-- Serialize a special object with 'a_name' and his internal id.
		local
			si8: SPECIAL [INTEGER_8]
			si16: SPECIAL [INTEGER_16]
			si32: SPECIAL [INTEGER]
			si64: SPECIAL [INTEGER_64]
			sn8: SPECIAL [NATURAL_8]
			sn16: SPECIAL [NATURAL_16]
			sn32: SPECIAL [NATURAL_32]
			sn64: SPECIAL [NATURAL_64]
			sc: SPECIAL [CHARACTER]
			sb: SPECIAL [BOOLEAN]
			sd: SPECIAL [DOUBLE]
			sr: SPECIAL [REAL]
			sp: SPECIAL [POINTER]
			sa: SPECIAL [ANY]
			i, n: INTEGER
		do
			sa ?= a_object
			if sa /= Void then
				si8 ?= a_object
				si16 ?= a_object
				si32 ?= a_object
				si64 ?= a_object
				sn8 ?= a_object
				sn16 ?= a_object
				sn32 ?= a_object
				sn64 ?= a_object
				sc ?= a_object
				sb ?= a_object
				sd ?= a_object
				sr ?= a_object
				sp ?= a_object
				if si8 /= Void or else si16 /= Void or else si32 /= Void or else si64 /= Void
				or else sn8 /= Void or else sn16 /= Void or else sn32 /= Void or else sn64 /= Void
				or else sc /= Void or else sb /= Void or else sd /= Void or else sr /= Void
				or else sp /= Void then
					--| Do nothing
				else
					from
						i := 0
						n := sa.count
					until
						i >= n
					loop
						build_ref (sa.item (i))
						i := i + 1
					end
				end
			end
		end

	output_types is
			-- Output the 'types' table.
		do
			start_element (k_types,Void,Void,0,0)
			from
				types.start
			until
				types.off
			loop
				if types.key_for_iteration > 0 then
					output_element (k_type,types.item_for_iteration,Void,0,types.key_for_iteration)
				end
				types.forth
			end
			terminate_element (k_types)
		end

	output_fields (a_name: STRING; a_object: ANY; a_id: INTEGER) is
			-- Serialize all fields of 'a_object' with 'a_name' and 'a_id' internal reference.
		local
			i, n, t: INTEGER
			l_name: STRING
		do
			l_name ?= a_object
			if l_name /= Void then
				output_ref (a_name, l_name)
			else
				n := field_count (a_object)
				if n > 0 then
					start_element (k_object, a_name, Void, internal_type (a_object), a_id)
					from
						i := 1
					until
						i > n
					loop
						l_name := field_name (i, a_object)
						t := field_type (i, a_object)
						inspect t
						when Reference_type then
							output_ref (l_name, field (i, a_object))
						when Integer_8_type then
							output_int8 (l_name, integer_8_field (i, a_object))
						when Integer_16_type then
							output_int16 (l_name, integer_16_field (i, a_object))
						when Integer_type then
							output_int32 (l_name, integer_32_field (i, a_object))
						when Integer_64_type then
							output_int64 (l_name, integer_64_field (i, a_object))
						when Natural_8_type then
							output_natural8 (l_name, natural_8_field (i, a_object))
						when Natural_16_type then
							output_natural16 (l_name, natural_16_field (i, a_object))
						when Natural_32_type then
							output_natural32 (l_name, natural_32_field (i, a_object))
						when Natural_64_type then
							output_natural64 (l_name, natural_64_field (i, a_object))
						when Double_type then
							output_double (l_name, double_field (i, a_object))
						when Real_type then
							output_real (l_name, real_field (i, a_object))
						when Character_type then
							output_char (l_name, character_field (i, a_object))
						when Character_32_type then
							output_char32 (l_name, character_32_field (i, a_object))
						when Boolean_type then
							output_bool (l_name, boolean_field (i, a_object))
						when Pointer_type then
							--| Do nothing
						when Expanded_type, Bit_type then
							output_expanded (l_name, field (i, a_object))
						else
							--| Do nothing
						end
						i := i + 1
					end
					terminate_element (k_object)
				else
					output_element (k_object, a_name, Void, internal_type (a_object), a_id)
				end
			end
		end

	output_ref (a_name: STRING; a_object: ANY) is
			-- Serialize a reference object (not expanded) with 'a_name'.
		local
			l_string: STRING
			i: INTEGER
		do
			if a_object /= Void then
					--| Voids are not output, since they are the default
				i := refs.item ($a_object)
				if i = 0 then
					warn ("output_ref","Warning! an object of type '$1' has moved in memory!",<<a_object.generator>>)
				elseif i > 0 then
					refs.force (-i,$a_object)
					l_string ?= a_object
					if l_string /= Void then
						output_string (a_name,l_string,i)
					elseif is_special (a_object) then
						output_special (a_name,a_object,i)
					else
						output_fields (a_name,a_object,i)
					end
				else
					output_element (k_reference,a_name,Void,0,-i)
				end
			end
		end

	output_string (a_name, a_string: STRING; a_ref: INTEGER) is
			-- Serialize `a_string'.
		local
			l_string: STRING
		do
			l_string := a_string.twin
			l_string.replace_substring_all (markup_newline,markup_newline_escaped)
			l_string.replace_substring_all (markup_ampersand,markup_ampersand_escaped)
			l_string.replace_substring_all (markup_quote,markup_quote_escaped)
			l_string.replace_substring_all (markup_start,markup_start_escaped)
			l_string.replace_substring_all (markup_end,markup_end_escaped)
			output_element (k_string,a_name,l_string,0,a_ref)
		end

	output_special (a_name: STRING; a_object: ANY; a_id: INTEGER) is
			-- Serialize a special object with 'a_name' and his internal id.
		local
			si8: SPECIAL [INTEGER_8]
			si16: SPECIAL [INTEGER_16]
			si32: SPECIAL [INTEGER]
			si64: SPECIAL [INTEGER_64]
			sn8: SPECIAL [NATURAL_8]
			sn16: SPECIAL [NATURAL_16]
			sn32: SPECIAL [NATURAL_32]
			sn64: SPECIAL [NATURAL_64]
			sc: SPECIAL [CHARACTER]
			sc32: SPECIAL [CHARACTER_32]
			sb: SPECIAL [BOOLEAN]
			sd: SPECIAL [DOUBLE]
			sr: SPECIAL [REAL]
			sp: SPECIAL [POINTER]
			sa: SPECIAL [ANY]
		do
			sa ?= a_object
			if sa /= Void then
				si8 ?= a_object
				si16 ?= a_object
				si32 ?= a_object
				si64 ?= a_object
				sn8 ?= a_object
				sn16 ?= a_object
				sn32 ?= a_object
				sn64 ?= a_object
				sc ?= a_object
				sc32 ?= a_object
				sb ?= a_object
				sd ?= a_object
				sr ?= a_object
				sp ?= a_object
				if si8 /= Void then
					specific_int8.output_special_section (Current,si8,a_name,a_id)
				elseif si16 /= Void then
					specific_int16.output_special_section (Current,si16,a_name,a_id)
				elseif si32 /= Void then
					specific_int32.output_special_section (Current,si32,a_name,a_id)
				elseif si64 /= Void then
					specific_int64.output_special_section (Current,si64,a_name,a_id)
				elseif sn8 /= Void then
					specific_natural8.output_special_section (Current,sn8,a_name,a_id)
				elseif sn16 /= Void then
					specific_natural16.output_special_section (Current,sn16,a_name,a_id)
				elseif sn32 /= Void then
					specific_natural32.output_special_section (Current,sn32,a_name,a_id)
				elseif sn64 /= Void then
					specific_natural64.output_special_section (Current,sn64,a_name,a_id)
				elseif sc /= Void then
					specific_char.output_special_section (Current,sc,a_name,a_id)
				elseif sc32 /= Void then
					specific_char32.output_special_section (Current,sc32,a_name,a_id)
				elseif sb /= Void then
					specific_bool.output_special_section (Current,sb,a_name,a_id)
				elseif sd /= Void then
					specific_double.output_special_section (Current,sd,a_name,a_id)
				elseif sr /= Void then
					specific_real.output_special_section (Current,sr,a_name,a_id)
				elseif sp /= Void then
					--| Do nothing
				else
					specific_any.output_special (Current,sa,a_name,a_id,agent output_ref)
				end
			end
		end

	output_expanded (a_name: STRING; a_object: ANY) is
			-- Serialize an expanded object with 'a_type' and 'a_name'.
		do
			output_element (k_expanded, a_name, a_object.out, internal_type (a_object), 0)
		end

	output_int8 (a_name: STRING; a_int: INTEGER_8) is
			-- Serialize an integer with 'a_name'.
		do
			if a_int /= 0 then
				output_element (k_int8,a_name,a_int.out,0,0)
			end
		end

	output_int16 (a_name: STRING; a_int: INTEGER_16) is
			-- Serialize an integer with 'a_name'.
		do
			if a_int /= 0 then
				output_element (k_int16,a_name,a_int.out,0,0)
			end
		end

	output_int32 (a_name: STRING; a_int: INTEGER) is
			-- Serialize an integer with 'a_name'.
		do
			if a_int /= 0 then
				output_element (k_int32,a_name,a_int.out,0,0)
			end
		end

	output_int64 (a_name: STRING; a_int: INTEGER_64) is
			-- Serialize an integer with 'a_name'.
		do
			if a_int /= 0 then
				output_element (k_int64,a_name,a_int.out,0,0)
			end
		end

	output_natural8 (a_name: STRING; a_int: NATURAL_8) is
			-- Serialize an integer with 'a_name'.
		do
			if a_int /= 0 then
				output_element (k_natural8, a_name, a_int.out, 0, 0)
			end
		end

	output_natural16 (a_name: STRING; a_int: NATURAL_16) is
			-- Serialize an integer with 'a_name'.
		do
			if a_int /= 0 then
				output_element (k_natural16, a_name, a_int.out, 0, 0)
			end
		end

	output_natural32 (a_name: STRING; a_int: NATURAL_32) is
			-- Serialize an integer with 'a_name'.
		do
			if a_int /= 0 then
				output_element (k_natural32, a_name, a_int.out, 0, 0)
			end
		end

	output_natural64 (a_name: STRING; a_int: NATURAL_64) is
			-- Serialize an integer with 'a_name'.
		do
			if a_int /= 0 then
				output_element (k_natural64, a_name, a_int.out, 0, 0)
			end
		end

	output_double (a_name: STRING; a_double: DOUBLE) is
			-- Serialize a double with 'a_name'.
		do
			if a_double /= 0.0 then
				output_element (k_double,a_name,a_double.out,0,0)
			end
		end

	output_real (a_name: STRING; a_real: REAL) is
			-- Serialize a real with 'a_name'.
		do
			if a_real /= 0.0 then
				output_element (k_real,a_name,a_real.out,0,0)
			end
		end

	output_char (a_name: STRING; a_char: CHARACTER) is
			-- Serialize a character with 'a_name'.
		do
			if a_char /= '%U' then
				output_element (k_character,a_name,"&#"+a_char.code.out+";",0,0)
			end
		end

	output_char32 (a_name: STRING; a_char: CHARACTER_32) is
			-- Serialize a character_32 with 'a_name'.
		do
			if a_char /= '%U' then
				output_element (k_character32,a_name,"&#"+a_char.natural_32_code.out+";",0,0)
			end
		end

	output_bool (a_name: STRING; a_bool: BOOLEAN) is
			-- Serialize a boolean with 'a_name'.
		do
			if a_bool then
					-- False values are not output, since they are the default
				output_element (k_boolean,a_name,Void,0,0)
			end
		end

	refs: HASH_TABLE [INTEGER,POINTER]
			-- Referenced objects table, count gives the number of objects serialized

	types: HASH_TABLE [STRING,INTEGER]
			-- Class type ids

feature {ROSE_SERIALIZER_SPECIFIC} -- Implementation

	current_file: RAW_FILE
			-- Current file where serialization goes to

	build_ref (a_object: ANY) is
			-- Add references to `a_object' to `refs'.
		local
			l_string: STRING
		do
			if a_object /= Void and then not refs.has ($a_object) then
				refs.put (refs.count + 1,$a_object)
				types.force (a_object.generating_type,dynamic_type (a_object))
				l_string ?= a_object
				if l_string /= Void then
					--| Do nothing
				elseif is_special (a_object) then
					build_special (a_object)
				else
					build_fields (a_object)
				end
			end
		end

	internal_type (a_object: ANY): INTEGER is
			-- Dynamic type of `a_object'
		do
			Result := dynamic_type (a_object)
		end

	output_element (a_element_name, a_name, a_value: STRING; a_type, a_id: INTEGER) is
			-- Output an empty XML element with 'a_element_name', with arguments
			-- 'a_name', 'a_type', 'a_value' and 'a_id' (internal), all are optional.
		do
			start_xml_element (a_element_name,a_name,a_value,a_type,a_id)
			current_file.put_string (markup_empty_end)
		end

	start_element (a_element_name, a_name, a_value: STRING; a_type, a_id: INTEGER) is
			-- Start a XML element with 'a_element_name', with arguments
			-- 'a_name', 'a_type', 'a_value' and 'a_id' (internal), all are optional.
		do
			start_xml_element (a_element_name,a_name,a_value,a_type,a_id)
			current_file.put_string (markup_end_line)
		end

	start_xml_element (a_element_name, a_name, a_value: STRING; a_type, a_id: INTEGER) is
			-- Start a XML element with 'a_element_name', with arguments
			-- 'a_name', 'a_type', 'a_value' and 'a_id' (internal), all are optional.
			-- Markup is not terminated.
		do
			current_file.put_string (markup_start)
			current_file.put_string (a_element_name)
			if a_name /= Void then
				current_file.put_string (arg_name)
				current_file.put_string (a_name)
				current_file.put_string (markup_quote)
			end
			if a_type /= 0 then
				current_file.put_string (arg_type)
				current_file.put_string (a_type.out)
				current_file.put_string (markup_quote)
			end
			if a_id /= 0 then
				current_file.put_string (arg_id)
				current_file.put_string (a_id.out)
				current_file.put_string (markup_quote)
			end
			if a_value /= Void then
				current_file.put_string (arg_value)
				current_file.put_string (a_value.out)
				current_file.put_string (markup_quote)
			end
		end

	terminate_element (a_element_name: STRING) is
			-- Terminate a XML element with 'a_element_name'.
		do
			current_file.put_string (markup_endtag_start)
			current_file.put_string (a_element_name)
			current_file.put_string (markup_end_line)
		end

end
