indexing
	description: "Service manager"
	author: "Zoran Simic"
	date: "$Date: 2008/12/01 $"
	revision: "$Revision: #12 $"

class
	ROSE_SERVICE_MANAGER [G -> ROSE_SERVICE]

inherit
	ROSE_SHARED_GLOBAL
	ANY

create
	make

feature -- Initialization

	make (a_id, a_default_service: STRING) is
			-- Initialize process manager with default reader/writer.
		require
			valid_id: a_id /= Void
		do
			create services.make (0)
			id := a_id
			default_service := a_default_service
		end

feature -- Access

	id: STRING
			-- Manager id

	default_service: STRING
			-- Default service to be used when returning an 'open_service'
			-- Leave Void for no default

	services: HASH_TABLE [G,STRING]
			-- Services defined in this manager

	service (a_id: STRING): G is
			-- Service associated to 'a_id'
		local
			l_id: STRING
		do
			if a_id /= Void then
				l_id := clone (a_id)
				l_id.to_lower
				Result := services.item (l_id)
				if Result /= Void then
					Result := Result.new_representative
				end
			end
		end

	open_service (a_path: STRING): G is
			-- Service corresponding to 'a_path'
			-- 'a_path' is of the format "id:path", where path is specific to the service
			-- if no 'id' found 'default_service' is tried
		local
			l_pair: PAIR [STRING,STRING]
			l_path: STRING
		do
			if a_path /= Void then
				l_pair := Global.formatter.split_on_first (a_path,type_separator,Void)
				Result := service (l_pair.first)
				if Result = Void then
					if l_pair.second = Void or else (l_pair.first /= Void and then l_pair.first.count < 2) then
						Result := service (default_service)
						l_path := a_path
					end
				else
					l_path := l_pair.second
				end
				if Result /= Void then
					Result.open (l_path)
				end
			end
		end

	checked_open_service (a_path: STRING): G is
			-- Service corresponding to 'a_path'
			-- 'a_path' is of the format "id:path", where path is specific to the service
			-- if no 'id' found 'default_service' is tried
			-- Error signaled if service not available or could not be opened
		local
			l_void_g: G
		do
			Result := open_service (a_path)
			if Result = Void then
				error.signal (id,"Can not find '$1'",<<a_path>>)
			elseif not Result.is_open then
				error.signal (id,"Can not open '$1'",<<a_path>>)
				Result := l_void_g
			end
		ensure
			open_if_valid: Result /= Void implies Result.is_open
			error_signaled: Result = Void implies error.signaled
		end

	sorted_service_ids: SEQUENCE [STRING] is
			-- Sorted service ids of this manager
		do
			create {SORTED_TWO_WAY_LIST [STRING]} Result.make
			from
				services.start
			until
				services.off
			loop
				Result.extend (services.key_for_iteration)
				services.forth
			end
		end

feature -- Status report

	has (a_id: STRING): BOOLEAN is
			-- Is 'a_id' defined in this manager?
		local
			l_id: STRING
		do
			if a_id /= Void then
				l_id := clone (a_id)
				l_id.to_lower
				Result := services.has (l_id)
			end
		end

	count: INTEGER is
			-- Number of services in this manager
		do
			Result := services.count
		end

feature -- Basic operations

	add (a_service: G) is
			-- Add 'a_service' to this factory.
		require
			valid_service: a_service /= Void
		local
			l_id: STRING
		do
			l_id := clone (a_service.id)
			l_id.to_lower
			services.force (a_service,l_id)
		end

	remove (a_id: STRING) is
			-- Remove service with 'a_id' from this manager.
		local
			l_id: STRING
		do
			if a_id /= Void then
				l_id := clone (a_id)
				l_id.to_lower
				services.remove (l_id)
			end
		end

	iterate (a_agent: PROCEDURE [ANY,TUPLE [G]]; a_sorted: BOOLEAN) is
			-- Call 'a_agent' on every service of this manager
			-- in alphabetical 'id' order if 'a_sorted' is True.
		local
			l_ids: SEQUENCE [STRING]
			l_items: LINEAR [G]
		do
			if a_agent /= Void then
				if a_sorted then
					from
						l_ids := sorted_service_ids
						l_ids.start
					until
						l_ids.off
					loop
						a_agent.call ([service (l_ids.item)])
						l_ids.forth
					end
				else
					from
						l_items := services.linear_representation
						l_items.start
					until
						l_items.off
					loop
						a_agent.call ([l_items.item])
						l_items.forth
					end
				end
			end
		end

	iterate_ids (a_agent: PROCEDURE [ANY,TUPLE [STRING]]; a_sorted: BOOLEAN) is
			-- Call 'a_agent' on every service id of this manager
			-- in alphabetical 'id' order if 'a_sorted' is True.
		do
			if a_agent /= Void then
				sorted_service_ids.do_all (a_agent)
			end
		end

feature -- Constants

	type_separator: STRING is ":"
			-- Type/path separator
			-- Services are requested with strings of the form <type>:<path>

end -- class ROSE_SERVICE_MANAGER
