note
	description: "Shared global variables and procedures"
	author: "Zoran Simic"
	last_editor: "$Author: mb $"
	date: "$Date: 2012/10/08 $"
	revision: "$Revision: #27 $"
	path: "$File: //rose/source/kernel/infrastructure/framework/application/rose_shared_global.e $"

class
	ROSE_SHARED_GLOBAL

feature -- Access

	frozen global: attached ROSE_GLOBAL
			-- Shared global object holding generic information such as info on application currently running, environment settings etc
		indexing
			once_status : global
		once
			create Result.make
		end

	frozen config: attached ROSE_CONFIGURATION
			-- Application configuration, this holds the contents of <current program>.cfg
		do
			Result := global.app.config
		end

	frozen error: attached ROSE_ERROR_HANDLER
			-- Shared error handler
		do
			Result := global.io_handler.error
		end

	frozen last_exception_explanation: attached STRING
			-- Info on last exception (to be called within a rescue clause)
		do
			Result := error.last_exception_explanation
		end

	frozen application_configuration_status_report: attached STRING
			-- Current application's configuration status report
			-- Use this in debugger for example to find out which configuration files are used by this application
		do
			Result := global.app.report
		end

feature {NONE} -- Basic operations

	frozen output (a_message: STRING; a_args: ARRAY [ANY])
			-- Output 'a_message' with 'a_args'.
			-- This call outputs to the current output/echo channels.
		do
			global.io_handler.do_output (global.convention.v_output,a_message,a_args)
		end

	frozen output_once (a_message: STRING; a_args: ARRAY [ANY])
			-- Output 'a_message' with 'a_args'.
		    -- This call outputs to the current output/echo channels.
		    -- Do this only once per interpreted `a_message'
		do
			global.io_handler.do_output_once (global.convention.v_output,a_message,a_args)
		end

	frozen output_timestamped (a_message: STRING; a_args: ARRAY [ANY])
			-- Output `a_message' with `a_args' prefixed by a timestamp "hh:mm:ss.fff:  "
		do
			output ((create {ROSE_TIME}.make_now).formatted_time + ": " + a_message, a_args)
		end

	frozen output_timestamped_utc (a_message: STRING; a_args: ARRAY [ANY])
			-- Output `a_message' with `a_args' prefixed by a timestamp "hh:mm:ss.fff:  "
		do
			output ((create {ROSE_DATE}.make_now_utc).formatted_time + ": " + a_message, a_args)
		end

	frozen print_timestamped (a_message: STRING)
			-- Output `a_message' prefixed by a timestamp "hh:mm:ss.fff:  "
		do
			output_timestamped (a_message, <<>>)
		end

	frozen trace (a_message: STRING; a_args: ARRAY [ANY])
			-- Output 'a_message' with 'a_args' to log.
			-- This call outputs to the current output/echo channels.
		do
			global.io_handler.do_output (global.convention.v_trace,a_message,a_args)
		end

	frozen fail (a_feature, a_message: STRING; a_args: ARRAY [ANY])
			-- Log a fatal error message with 'a_message' + 'a_args', marked coming from <generator>.'a_feature'.
			-- This call outputs to the current log and error broadcast channels.
		require
			valid_feature: a_feature /= Void
		do
			if a_message /= Void then
				global.log_message (global.convention.log_fatal,generator,a_feature,a_message,a_args)
				error.abort (generator + "." + a_feature,a_message,a_args)
			else
				error.abort (generator,"'Fail' called with no message from $1.$2",<<generator,a_feature>>)
			end
		end

	frozen warn (a_feature, a_message: STRING; a_args: ARRAY [ANY])
			-- Log a warning message with 'a_message' + 'a_args', marked coming from <generator>.'a_feature'.
			-- This call outputs to the current log and error broadcast channels.
		require
			valid_feature: a_feature /= Void
		do
			if a_message /= Void then
				global.log_message (global.convention.log_warning,generator,a_feature,a_message,a_args)
				error.warn (generator + "." + a_feature + ": " + a_message,a_args)
			else
				error.warn ("'Warn' called with no message from $1.$2",<<generator,a_feature>>)
			end
		end

	frozen inform (a_feature, a_message: STRING; a_args: ARRAY [ANY])
			-- Log an information message with 'a_message' + 'a_args', marked coming from <generator>.'a_feature'.
			-- This call outputs to the current log and error broadcast channels.
		require
			valid_feature: a_feature /= Void
		do
			if a_message /= Void and then not a_message.is_empty then
				global.log_message (global.convention.log_information,generator,a_feature,a_message,a_args)
				error.broadcast (global.convention.v_log,generator + "." + a_feature + ": " + a_message,a_args)
				if a_message.item (a_message.count) = '%N' then
					output (a_message, a_args)
				else
					output (a_message + "%N", a_args)
				end
			end
		end

	frozen tell (a_feature,a_message: STRING; a_args: ARRAY [ANY])
			-- Log an information message with 'a_message' + 'a_args', marked coming from <generator>.'a_feature'.
			-- This call outputs to the error broadcast only.
		require
			valid_feature: a_feature /= Void
		do
			if a_message /= Void then
				error.broadcast (global.convention.v_log,generator + "." + a_feature + ": " + a_message,a_args)
			end
		end

	frozen show_elapsed_time (a_description: STRING)
			-- Show elapsed time since last start_time_measure with 'a_description'.
		local
			l_tm: ROSE_TIME_MEASURE
		do
			if global.io_handler.monitor /= Void then
				l_tm := global.io_handler.time_measures.item
				if l_tm /= Void then
					global.io_handler.monitor.output_message (global.convention.v_trace,l_tm.progress_report (a_description))
				end
			end
		end

	frozen start_time_measure (a_name: STRING; a_args: STRING)
			-- Start time measure for 'a_id' with 'a_qualification'.
		do
			global.io_handler.start_time_measure (a_name,a_args)
		end

	frozen end_time_measure
			-- End last time measure and output representation on monitor.
			-- This call outputs to the current monitor channel.
		do
			global.io_handler.end_time_measure
		end

end
