note

	description:

		"Objects that are unique records in a signature tagged item database"

	author: "Mauricio Bustos"
	last_editor: "$Author: yc $"
	date: "$Date: 2013/05/30 $"
	revision: "$Revision: #116 $"
	path: "$File: //rose/source/kernel/persistence/databases/signature_tagged/reference/rose_signature_record.e $"

deferred class ROSE_SIGNATURE_RECORD [reference G -> ENTITY]

inherit

	ROSE_ENTITY_DATE [G]
		rename
			make as rose_entity_date_make
		undefine
			anchor,
			is_equal,
			expression_tag,
			targets,
			context_name,
			expression_target,
			output_for_verification
		redefine
			new_variable,
			expressions_anchor,
			set_entity,
			date_name,
			is_equal,
			multi_key
		end

	DEBUG_OUTPUT
		undefine
			is_equal
		end

feature -- Access

	signature_id: INTEGER
			-- A unique integer for this combination of key fields within the context of the data manager

	first_process_id: INTEGER
			-- The process id in the database the created the signature

	last_process_id: INTEGER
			-- The process id in the database that last updated any data attached to this signature

	continuing_process_id: INTEGER
			-- The process id in the database that last verified all of the data in the record (ie. Processed new data but there was no data value change observed)

	signature_data_process_id: INTEGER
			-- The process id in the database that last updated any data stored in the signature itself

	path: INTEGER
			-- The cluster id that this signature belongs to.  This is used by the clustered tagged item database object

	as_of_date: ROSE_DATE
			-- The date corresponding to the last_process_id

	end_current_date: ROSE_DATE
			-- The date for which this record is valid until

	pinned_version_date: ROSE_DATE

	latest_version_retrieval_on: BOOLEAN is
			--retrieve latest version?
		do
			Result := False
		end

	is_gui_editable: BOOLEAN is
			-- May the contents of this record be edited manually from a GUI?
			-- This should be redefined in descendents to return `True' if
			-- you wish signatures of this type to be editable in tools such as Data Workbench.
			-- See also `read_only_data_manager'.
		do
			Result := False
		end

	new_variable (a_description: VARIABLE_DESCRIPTION): like new_variable_template is
			-- A new instance of a variable to be used to query in the context of this signature type
		do
			Result := new_sig_variable (a_description)
		end

	new_sig_variable (a_description: VARIABLE_DESCRIPTION): SIGNATURE_TAGGED_EXPRESSION [like Current] is
			-- A new instance of a signature tagged expression to be used to query in the context of this signature type
		do
			create Result.make (a_description)
		end

	date_name: STRING is
			-- The attribute name of the field that specifies the main date attribute for this signature
		do
			Result := "period_end_date"
		ensure then
			valid_result: Result /= Void
		end

	rose_currency: PERSISTENT_CURRENCY is
			-- The rose currency corresponding to this signature
		local
			l_currency: like currency_imp
			l_history: ROSE_ENTITY_HISTORY [PERSISTENT_CURRENCY]
		do
			l_currency := currency_imp
			if l_currency /= Void and then l_currency.is_valid then
				Result := context.rose.currency_accessor.most_recent_peer (l_currency, date)
				if Result = Void then
					l_history := context.rose.currency_accessor.peers_history (l_currency)
					if l_history /= Void and then l_history.count > 0 then
						Result := l_history.first.entity
					end
				end
			end
		end

	signature_fields: ROSE_LINEAR_ARRAY [STRING] is
			-- An array of field names that specify the fields of the signature
		local
			l_type: INTEGER
		do
			l_type := internal.dynamic_type (Current)
			Result := signature_fields_array.safe_item (l_type)
			if Result = Void then
				create {ROSE_DELIMITED_TEXT} Result.make (fields_, ',')
				Result.compare_objects
				signature_fields_array.force (Result, l_type)
			end
		end

	data_fields: ROSE_LINEAR_ARRAY [STRING] is
			-- An array of field names that specify the data portion of the signature
		local
			l_type: INTEGER
		do
			l_type := internal.dynamic_type (Current)
			Result := data_fields_array.safe_item (l_type)
			if Result = Void then
				create {ROSE_DELIMITED_TEXT} Result.make (data_fields_string_, ',')
				Result.compare_objects
				data_fields_array.force (Result, l_type)
			end
		end

	data_query_fields: ROSE_LINEAR_ARRAY [STRING] is
			-- An array of field names that specify the data portion of the signature
		do
			create {ROSE_DELIMITED_TEXT} Result.make (data_query_fields_string_, ',')
		end

	fields_: STRING is
			-- Comma delimited string identifying the key fields for the signature
		deferred
		ensure
			valid_result: Result /= Void and then not Result.is_empty
		end

	data_fields_string_: STRING is
			-- Comma delimited string identifying data attributes stored in `Current'
		do
			Result := ""
		ensure
			valid_result: Result /= Void
		end

	data_query_fields_string_: STRING is
			-- Comma delimited string identifying data fields that
			-- could be used for queries from tools such as
			-- data_browser.  These are purposely limited to allow
			-- for more concise querying from these tools
		local
			l_config_entry: STRING
			l_delimited: ROSE_DELIMITED_TEXT
		do
			Result := ""
			l_config_entry := servers.config.value_default ("/" + generating_type + "/data_query_fields_string", "")
			l_config_entry.prune_all (' ')
			l_config_entry.to_lower
			create l_delimited.make (l_config_entry, ',')
			if l_delimited.for_all (agent string_contains_case_insensitive (data_fields_string_, ?)) then
				Result := l_config_entry
			else
				print (generating_type + ": /" + generating_type + "/data_query_fields_string is invalid%N")
			end
		end

	debug_output: STRING is
			-- <Precursor>
		do
			Result := signature_display_string
		end

feature -- Element modification

	set_signature_id (a_value: like signature_id) is
			--Set the signature_id to a_value
		do
			signature_id := a_value
		end

	set_first_process_id (a_value: like first_process_id) is
			-- Assign `a_value' to `first_process_id'.
		do
			first_process_id := a_value
		end

	set_last_process_id (a_value: like last_process_id) is
			-- Assign `a_value' to `last_process_id'.
		do
			last_process_id := a_value
		end

	set_end_current_date (a_value: like end_current_date) is
			--setter for end_current_date
		do
			end_current_date := a_value
		end

	set_continuing_process_id (a_value: like continuing_process_id) is
		do
			continuing_process_id := a_value
		end

	set_signature_data_process_id (a_value: like signature_data_process_id) is
		do
			signature_data_process_id := a_value
		end

	set_path (a_value: like path) is
			-- Set 'path' to 'a_value'
		do
			path := a_value
		end

	set_as_of_date (a_value: like as_of_date) is
			-- Assign `a_value' to `as_of_date'.
		do
			as_of_date := a_value
		end

	set_pinned_version_date (a_value: like pinned_version_date) is
		do
			pinned_version_date := a_value
		end

	set_entity (a_entity: like entity) is
			-- Assign `a_entity' to `entity'.
		do
			Precursor (a_entity)
			if a_entity /= Void then
				set_entity_id (a_entity.handle)
			end
		end

	set_entity_id (a_id: INTEGER) is
			-- Set `entity_id' to `a_id'
		deferred
		end

feature -- Comparison

	multi_key: MULTI_KEY is
			-- A multi key built from the signature columns that can be used to hash this signature object
		local
			l_keys: ROSE_LINEAR_ARRAY [HASHABLE]
			l_hashable: HASHABLE
			l_fields: ROSE_LINEAR_ARRAY [STRING]
			l_table: HASH_TABLE [INTEGER, STRING]
		do
			if saved_multi_key = Void then
				l_fields := signature_fields
				create l_keys.make (1, l_fields.count)
				l_table := field_index_by_name_table
				from
					l_fields.start
				until
					l_fields.after
				loop
					l_hashable ?= value_by_index (l_table.item (l_fields.current_item))
					if l_hashable /= Void then
						l_keys.put (l_hashable, l_fields.index)
					end
					l_fields.forth
				end
				create saved_multi_key.make (l_keys)
			end
			Result := saved_multi_key
		end

	matches (a_other: like Current): BOOLEAN is
			-- Does a_other match Current by signature fields?
		require
			valid_other: a_other /= Void
		local
			l_fields: ROSE_LINEAR_ARRAY [STRING]
		do
			Result := True
			l_fields := signature_fields
			from
				l_fields.start
			until
				l_fields.after or not Result
			loop
				Result := value_by_name (l_fields.current_item).is_equal (a_other.value_by_name (l_fields.current_item))
				l_fields.forth
			end
		end

	data_matches (a_other: like Current): BOOLEAN is
			-- Does 'a_other' match Current by data fields?
		require
			valid_other: a_other /= Void
		do
			Result := data_fields.for_all (agent data_field_matches (?, a_other))
		end

	data_matching_fields (a_other: like Current): ROSE_LINEAR_ARRAY [STRING] is
			-- All data field names of `Current' that match `a_other'.
		local
			i: INTEGER
			l_fields: ROSE_LINEAR_ARRAY [STRING]
			l_current_field: STRING
		do
			create Result.make_empty
			l_fields := data_fields
			from
				i := l_fields.lower
			until
				i > l_fields.upper
			loop
				l_current_field := l_fields.item (i)
				if data_field_matches (l_current_field, a_other) then
					Result.extend (l_current_field)
				end
				i := i + 1
			end
		ensure
			valid_result: Result /= Void
		end

	data_non_matching_fields (a_other: like Current): ROSE_LINEAR_ARRAY [STRING] is
			-- All data field names of `Current' that do not match `a_other'.
		local
			i: INTEGER
			l_fields: ROSE_LINEAR_ARRAY [STRING]
			l_current_field: STRING
		do
			create Result.make_empty
			l_fields := data_fields
			from
				i := l_fields.lower
			until
				i > l_fields.upper
			loop
				l_current_field := l_fields.item (i)
				if not data_field_matches (l_current_field, a_other) then
					Result.extend (l_current_field)
				end
				i := i + 1
			end
		ensure
			valid_result: Result /= Void
		end

	data_field_matches (a_field: STRING; a_other: like Current): BOOLEAN is
			--  Does value of data field named `a_field' match that of data field named `a_field' in `a_other'.
		require
			valid_other: a_other /= Void
			valid_field: a_field /= Void
		local
			l_any_from, l_any_to: ANY
			l_double_ref, l_double_ref_2: DOUBLE_REF
			l_date_time, l_date_time1: ROSE_DATE
		do
			Result := True
			l_any_from ?= value_by_name (a_field)
			l_any_to ?= a_other.value_by_name (a_field)
			l_double_ref ?= l_any_from
			l_double_ref_2 ?= l_any_to
			l_date_time ?= l_any_from
			l_date_time1 ?= l_any_to
			if l_double_ref /= Void and l_double_ref_2 /= Void then
				Result := equal_field_double (l_double_ref.item, l_double_ref_2.item)
			elseif l_date_time /= Void and l_date_time1 /= Void then
				Result := l_date_time.is_equal_in_sec (l_date_time1)
			else
				Result := equal (l_any_from, l_any_to)
			end
		end

	is_equal (a_other: like Current): BOOLEAN is
			-- Does a_other match Current by multi_key?
		do
			if a_other /= Void then
				Result := multi_key.is_equal (a_other.multi_key)
			end
		end

feature {NONE} -- Field comparisons

	equal_field_double (a_double_1, a_double_2: DOUBLE): BOOLEAN is
			-- Default double comparison in signature fields.
		do
			Result :=  ieee_math.ieee_is_eq (a_double_1, a_double_2)
		end

feature -- Retrieval

	targets (a_targets: ARRAY [like Current]; a_statement: STRING): ROSE_LINEAR_ARRAY [like anchor] is
			--indirection to avoid catcalls in descendents
		do
			Result := rose_signature_record_targets (a_targets, a_statement)
		end

	as_targets: like targets is
			-- Return Current in a one item list.
		do
			create Result.make_empty
			Result.extend (Current)
		ensure
			not_void: Result /= Void
			one_item: Result.count = 1
			item_is_current: Result.first = Current
		end

feature {NONE} -- Retrieval

	rose_signature_record_targets (a_targets: ARRAY [like Current]; a_statement: STRING): like targets is
			-- Targets
		local
			l_statement: STRING
		do
			if a_statement = Void then
				l_statement := restriction_statement
			else
				l_statement := a_statement
			end
			if l_statement /= Void then
				l_statement := repaired_statement (l_statement)
				set_targets_from_array (a_targets)
				data_manager.begin_transaction
				Result := data_manager.retrieve_current_signatures (targets_, l_statement)
				data_manager.data_server_commit
				if Result /= Void then
					reconcile_all_targets (Result)
				end
			end
		end

feature -- Retrieval

	historical_targets (a_targets: ARRAY [like Current]; a_statement: STRING): like targets is
		local
			l_statement: STRING
		do
			if a_statement = Void then
				l_statement := restriction_statement
			else
				l_statement := a_statement
			end
			if l_statement /= Void then
				l_statement := repaired_statement (l_statement)
				set_targets_from_array (a_targets)
				Result := data_manager.retrieve_historical_signatures (targets_, l_statement)
				if Result /= Void then
					reconcile_all_targets (Result)
				end
			end
		end

	time_series_targets (a_target: like Current): like targets is
		do
			create targets_.make_from_array (<<a_target>>)
			restriction_direction.set_time_series
			Result := data_manager.retrieve_historical_signatures (targets_, restriction_statement)
			if Result /= Void then
				reconcile_all_targets (Result)
			end
		end

	cross_sectional_targets (a_target: like Current): like targets is
		do
			create targets_.make_from_array (<<a_target>>)
			restriction_direction.set_cross_sectional
			Result := data_manager.retrieve_historical_signatures (targets_, restriction_statement)
			if Result /= Void then
				reconcile_all_targets (Result)
			end
		end

	neighborhood_targets (a_targets: like targets): like targets is
		do
			restriction_direction.set_neighborhood
			Result := data_manager.retrieve_historical_signatures (a_targets, restriction_statement)
			if Result /= Void then
				reconcile_all_targets (Result)
			end
		end

	records (a_targets: ARRAY [ROSE_SIGNATURE_RECORD [ENTITY]]; a_statement: STRING): ROSE_LINEAR_ARRAY [ROSE_VIRTUAL_RECORD [like anchor]] is
		local
			l_statement: STRING
		do
			if a_statement = Void then
				l_statement := restriction_statement
			else
				l_statement := a_statement
			end
			if l_statement /= Void then
				l_statement := repaired_statement (l_statement)
				set_targets_from_array (a_targets)
				Result := data_manager.retrieve_signatures (targets_, l_statement, Void)
				if Result /= Void then
					reconcile_all_records (Result)
				end
			end
		end

	historical_records (a_targets: ARRAY [ROSE_SIGNATURE_RECORD [ENTITY]]; a_statement: STRING): like records is
		local
			l_statement: STRING
		do
			if a_statement = Void then
				l_statement := restriction_statement
			else
				l_statement := a_statement
			end
			if l_statement /= Void then
				l_statement := repaired_statement (l_statement)
				set_targets_from_array (a_targets)
				Result := data_manager.retrieve_historical_items (targets_, l_statement, Void)
				if Result /= Void then
					reconcile_all_records (Result)
				end
			end
		end

	itemized_records (a_targets: ARRAY [like Current]; a_statement: STRING; a_items: ARRAY [INTEGER]): like records is
		local
			l_statement: STRING
			l_items: ROSE_LINEAR_ARRAY [INTEGER]
		do
			if a_statement = Void then
				l_statement := restriction_statement
			else
				l_statement := a_statement
			end
			if l_statement /= Void and a_items /= Void then
				create l_items.make_from_array (a_items)
				set_targets_from_array (a_targets)
				Result := data_manager.retrieve_signatures (targets_, l_statement, l_items)
				if Result /= Void then
					reconcile_all_records (Result)
				end
			end
		end

	itemized_historical_records (a_targets: ARRAY [like Current]; a_statement: STRING; a_items: ARRAY [INTEGER]): like records is
		local
			l_statement: STRING
			l_items: ROSE_LINEAR_ARRAY [INTEGER]
		do
			if a_statement = Void then
				l_statement := restriction_statement
			else
				l_statement := a_statement
			end
			if l_statement /= Void and a_items /= Void then
				create l_items.make_from_array (a_items)
				set_targets_from_array (a_targets)
				Result := data_manager.retrieve_historical_items (targets_, l_statement, l_items)
				if Result /= Void then
					reconcile_all_records (Result)
				end
			end
		end

feature -- Process timing

	first_data_date: ROSE_DATE is
		do
			Result := restricted_date (data_type_string, min_string)
		end

	last_data_date: ROSE_DATE is
		do
			Result := restricted_date (data_type_string, max_string)
		end

	first_process_date: ROSE_DATE is
		do
			Result := restricted_date (process_type_string, min_string)
		end

	last_process_date: ROSE_DATE is
		do
			Result := restricted_date (process_type_string, max_string)
		end

	first_process_date_for_signature: ROSE_DATE is
			-- The above features (first_process_date, last_process_date) are across ALL signatures for the implied entity.
			--
			-- What process date created this signature
		local
			l_first_process: like first_process
		do
			l_first_process := first_process
			if l_first_process /= Void then
				Result := l_first_process.file_date
			end
		end

	last_process_date_for_signature: ROSE_DATE is
			-- What process date last updated this signature
		local
			l_last_process: like last_process
		do
			l_last_process := last_process
			if l_last_process /= Void then
				Result := l_last_process.file_date
			end
			if Result = Void then
					-- This is a kluge intended to handle the situation when a supplier update program is running at the same time as a data retrieval program
				create Result.make_now
			end
		end

	continuing_process_date_for_signature: ROSE_DATE is
			-- What process date last validated this signature
		local
			l_continuing_process: like continuing_process
		do
			l_continuing_process := continuing_process
			if l_continuing_process /= Void then
				Result := l_continuing_process.file_date
			end
		end

	first_process: SUPPLIER_PROCESS is
			-- If it makes sense, it would be nice to have a similar feature associated with signature_data_process_id ...
			--
			-- The supplier process object for the process that created the signature
		do
			Result := data_manager.process_id_manager.cached_process_by_id (first_process_id)
			if Result = Void then
					-- See if the process_id is new if we can't find it
				data_manager.refresh_process_id_map
				Result := data_manager.process_id_manager.cached_process_by_id (first_process_id)
			end
		end

	last_process: SUPPLIER_PROCESS is
			-- The supplier process object for the process that last updated the signature
		do
			Result := data_manager.process_id_manager.cached_process_by_id (last_process_id)
			if Result = Void then
					-- See if the process_id is new if we can't find it
				data_manager.refresh_process_id_map
				Result := data_manager.process_id_manager.cached_process_by_id (last_process_id)
			end
		end

	continuing_process: SUPPLIER_PROCESS is
			-- The supplier process object for the process that last validated the signature
		do
			Result := data_manager.process_id_manager.cached_process_by_id (continuing_process_id)
			if Result = Void then
					-- See if the process_id is new if we can't find it
				data_manager.refresh_process_id_map
				Result := data_manager.process_id_manager.cached_process_by_id (continuing_process_id)
			end
		end

feature -- Restrictions

	repaired_statement (a_statement: STRING): STRING is
		do
			Result := a_statement
		end

	restriction_statement: STRING is
			-- Restriction statement used in query to select the
			-- targets to be retrieved.
		deferred
		end

	set_direction_default is
		do
			restriction_direction.set_default
		end

	set_direction_time_series is
		do
			restriction_direction.set_time_series
		end

	set_direction_cross_sectional is
		do
			restriction_direction.set_cross_sectional
		end

	set_direction_neighborhood is
		do
			restriction_direction.set_neighborhood
		end

	is_direction_default: BOOLEAN is
		do
			Result := restriction_direction.is_default
		end

	is_direction_time_series: BOOLEAN is
		do
			Result := restriction_direction.is_time_series
		end

	is_direction_cross_sectional: BOOLEAN is
		do
			Result := restriction_direction.is_cross_sectional
		end

	is_direction_neighborhood: BOOLEAN is
		do
			Result := restriction_direction.is_neighborhood
		end

	restriction_direction: ROSE_SIGNATURE_RESTRICTION_DIRECTION is
			-- Retrieval restriction criteria
			-- (which field of the target to use as selectors)
		do
			Result := restriction_direction_table.item (internal_type)
			if Result = Void then
				create Result.make
				restriction_direction_table.put (Result, internal_type)
			end
		end

	data_fields_valid_values_statement: STRING is
			-- An SQL compliant statement that determines if the signature data in the record constitutes a valid record
			-- i.e. Valid price and adjustment factor is a minimum set of data needed to determine validity
		do
		end

	selective_restriction_statement: STRING is
		do
			if selective_restriction_statement_saved = Void then
				selective_restriction_statement_saved := restriction_clause (signature_fields)
			end
			Result := selective_restriction_statement_saved
		end

feature {NONE} -- Restriction

	string_contains_case_insensitive (a_target, a_source: STRING): BOOLEAN is
			-- Does `a_target' contain `a_source'?
		require
			non_void_strings: a_target /= Void and a_source /= Void
		do
			Result := a_target.as_lower.has_substring (a_source.as_lower)
		end

	restriction_direction_table: HASH_TABLE [ROSE_SIGNATURE_RESTRICTION_DIRECTION, INTEGER] is
		once
			create Result.make (10)
		end

	selective_restriction_statement_saved: STRING
			-- Cache for `selective_restriction_statement'

feature -- Result processing

	ordered_targets (a_targets: like targets; a_scale_representation: STRING): like targets is
			-- 'a_scale_representation' contains a semi-colon separated list of composed criteria
		local
			l_scale: ROSE_QUOTIENT_SCALE [like anchor]
			l_criteria: ROSE_SIGNATURE_CRITERIA [like Current]
			l_quotient: ROSE_QUOTIENT [like anchor]
		do
			if a_targets /= Void and then a_scale_representation /= Void then
				create l_criteria.make
				l_scale := l_criteria.scale (a_scale_representation)
				if l_scale /= Void then
					create l_quotient.make (l_criteria)
					l_quotient.set_scale (l_scale)
					l_quotient.extend_array (a_targets)
					create Result.make_from_linear (l_quotient.all_objects)
				end
			end
		end

	reconcile_target (a_target: like anchor) is
		do
		end

	reconcile_record (a_target: ROSE_VIRTUAL_RECORD [like anchor]) is
		do
			reconcile_target (a_target.signature)
		end

feature -- Data Source

	data_manager: SIGNATURE_TAGGED_ITEM_DATABASE [like anchor] is
		do
		end

	read_only_data_manager (a_server: SQL_DATA_SERVER): like data_manager is
			-- Return a data manager used only for reading signatures.
			-- This enables implementation of a synch tool to synch from
			-- one data manager to another. This is necessary to work around the
			-- limitation of `data_manager' being a function in ROSE_SIGNATURE_RECORD.
			-- This should be redefined in tandem with `is_gui_editable' in any
			-- signatures that you want to be editable in tools such as Data Workbench.
		require
			a_server_not_void: a_server /= Void
		do
		end

feature -- Utilities

	possible_values (a_field: STRING): ARRAY [ANY] is
			-- The collection of possible values for the field a_field
		do
		end

	set_field_with_string (a_value: STRING; a_field: STRING) is
			-- Set a_field using a string value
		do
			set_field (a_field, a_value)
		end

	output_string_by_name (a_name: STRING): STRING is
			-- An interpreted version of the signature field attribute value for a_name for display
		require
			valid_name: a_name /= Void
		local
			l_any: ANY
			l_date: ROSE_DATE
		do
			a_name.to_lower
			l_any := value_by_name (a_name)
			if l_any /= Void then
				l_date ?= l_any
				if l_date /= Void then
					Result := l_date.formatted_out ("[0]mm/[0]dd/yyyy")
				else
					Result := l_any.out
				end
			else
				Result := "<Void>"
			end
		ensure
			valid_result: Result /= Void
		end

	output_data_field_string_by_name (a_name: STRING): ANY is
			-- An interpreted version of the data field attribute value for a_name for display
		require
			valid_name: a_name /= Void
		local
			l_any: ANY
			l_date: ROSE_DATE
		do
			a_name.to_lower
			l_any := value_by_name (a_name)
			if l_any /= Void then
				l_date ?= l_any
				if l_date /= Void then
					Result := l_date.formatted_out ("[0]mm/[0]dd/yyyy")
				else
					Result := l_any
				end
			else
				Result := "<Void>"
			end
		ensure
			valid_result: Result /= Void
		end

	signature_display_string: STRING is
			-- Display of signature fields and values
		do
			Result := ""
			from
				signature_fields.start
			until
				signature_fields.off
			loop
				Result.append (signature_fields.current_item + " = " + output_string_by_name (signature_fields.current_item))
				if signature_fields.index < signature_fields.count then
					Result.append (",")
				end
				signature_fields.forth
			end
		ensure
			not_void: Result /= Void
		end

	output_for_verification: STRING is
			-- Used by STIM_VERIFICATION.  The returned STRING is just the concatenation of
			-- every bit of data in the record.  Called by ROSE_VIRTUAL_RECORD:output_for_verification.
		do
			Result := ""
			from
				signature_fields.start
			until
				signature_fields.off
			loop
				Result.append (output_string_by_name (signature_fields.current_item))
				if signature_fields.index < signature_fields.count then
					Result.append (",")
				end
				signature_fields.forth
			end
			from
				data_fields.start
			until
				data_fields.off
			loop
				Result.append (output_string_by_name (data_fields.current_item))
				if data_fields.index < data_fields.count then
					Result.append (",")
				end
				data_fields.forth
			end
		end

	is_valid: BOOLEAN is
		do
			Result := True
		end

feature {SIGNATURE_TAGGED_SERVER} -- SQL Clauses

	signatures_where_clause: STRING is
		do
			Result := ""
			from
				signature_fields.start
			until
				signature_fields.after
			loop
				if not Result.is_equal ("") then
					Result := Result + " and " + signature_fields.current_item + " = :" + signature_fields.current_item
				else
					Result := signature_fields.current_item + " = :" + signature_fields.current_item
				end
				signature_fields.forth
			end
		ensure
			result_not_void: Result /= Void
		end

	signatures_set_clause: STRING is
		do
			Result := ""
			from
				signature_fields.start
			until
				signature_fields.after
			loop
				if not Result.is_equal ("") then
					Result := Result + ", " + signature_fields.current_item + " = ?"
				else
					Result := signature_fields.current_item + " = ?"
				end
				signature_fields.forth
			end
		ensure
			result_not_void: Result /= Void
		end

	signatures_values_clause: STRING is
		do
			Result := values_clause (signature_fields)
		ensure
			result_not_void: Result /= Void
		end

	signatures_fields_clause: STRING is
		do
			Result := fields_clause (signature_fields)
		ensure
			result_not_void: Result /= Void
		end

	signature_and_data_values_clause: STRING is
		local
			l_data_values_clause: STRING
		do
			l_data_values_clause := values_clause (data_fields)
			if l_data_values_clause.count = 0 then
				Result := signatures_values_clause
			else
				Result := signatures_values_clause + "," + l_data_values_clause
			end
		end

	signature_and_data_fields_clause: STRING is
		local
			l_data_fields_clause: STRING
		do
			l_data_fields_clause := fields_clause (data_fields)
			if l_data_fields_clause.count = 0 then
				Result := signatures_fields_clause
			else
				Result := signatures_fields_clause + "," + l_data_fields_clause
			end
		end

feature {ANY} -- Implementation

	reconcile_all_records (a_records: like records) is
			-- Call `reconcile_record' on all of `a_records', avoiding a CATCALL.
		require
			a_records_not_void: a_records /= Void
			no_void_record: not a_records.has (Void)
		local
			i: INTEGER
		do
			from
				i := a_records.lower
			invariant
				good_index: a_records.valid_index (i) or i = a_records.upper + 1
			variant
				ascending_index: (a_records.upper - a_records.lower) + 2 - i
			until
				i > a_records.upper
			loop
				reconcile_record (a_records [i])
				i := i + 1
			end
		end

	targets_: ROSE_LINEAR_ARRAY [like Current]
			-- Used for conformance.  This should be removed but will result in storable retrieval problems.

	set_targets_from_array (a_targets: ARRAY [ROSE_SIGNATURE_RECORD [ENTITY]]) is
			-- Set 'targets_' from 'a_targets'
		require
			valid_targets: a_targets /= Void
		do
			targets_ ?= a_targets
			if targets_ = Void then
				create targets_.make_empty
				a_targets.do_all (agent extend_targets_)
			end
		end

feature {NONE} -- Implementation

	signature_fields_array: ROSE_LINEAR_ARRAY [ROSE_LINEAR_ARRAY [STRING]] is
			-- Global cache of signature field specifications for any signature referred to by this system
		once
			create Result.make_empty
		end

	data_fields_array: ROSE_LINEAR_ARRAY [ROSE_LINEAR_ARRAY [STRING]] is
			-- Global cache of signature data field specifications for any signature referred to by this system
		once
			create Result.make_empty
		end

	extend_targets_ (a_target: ROSE_SIGNATURE_RECORD [ENTITY]) is
			-- Extend 'targets_' with 'a_target'
		require
			valid_target: a_target /= Void
			available_targets_: targets_ /= Void
		local
			l_target: like Current
		do
			l_target ?= a_target
			if l_target /= Void then
				targets_.extend (l_target)
			end
		end

	data_type_string: STRING is "data"

	process_type_string: STRING is "process"

	min_string: STRING is "min"

	max_string: STRING is "max"

	reconcile_all_targets (a_targets: like targets) is
			-- Call `reconcile_target' on all of `a_targets', avoiding a CATCALL.
		require
			a_targets_not_void: a_targets /= Void
			no_void_target: not a_targets.has (Void)
		local
			i: INTEGER
		do
			from
				i := a_targets.lower
			invariant
				good_index: a_targets.valid_index (i) or i = a_targets.upper + 1
			variant
				ascending_index: (a_targets.upper - a_targets.lower) + 2 - i
			until
				i > a_targets.upper
			loop
				reconcile_target (a_targets [i])
				i := i + 1
			end
		end

	restricted_date (a_date_type, a_min_max: STRING): ROSE_DATE is
		local
			l_bounds: SIGNATURE_TAGGED_DATA_BOUNDS
		do
			if data_manager /= Void then
				l_bounds := data_manager.data_bounds (Current)
				if l_bounds /= Void then
					if a_date_type = data_type_string then
						if a_min_max = min_string and l_bounds.start_date /= Void then
							Result := l_bounds.start_date.twin
						elseif l_bounds.end_date /= Void then
							Result := l_bounds.end_date.twin
						end
					else
						if a_min_max = min_string and l_bounds.start_process_date /= Void then
							Result := l_bounds.start_process_date.twin
						elseif l_bounds.end_process_date /= Void then
							Result := l_bounds.end_process_date.twin
						end
					end
				end
			end
			if Result = Void then
				if a_min_max = min_string then
					create Result.make_max_date
				else
					create Result.make_min_date
				end
			end
		end

	values_clause (a_fields: ROSE_LINEAR_ARRAY [STRING]): STRING is
		do
			Result := clause (a_fields, ":", ",:")
		end

	fields_clause (a_fields: ROSE_LINEAR_ARRAY [STRING]): STRING is
		do
			Result := clause (a_fields, "", ",")
		end

	clause (a_fields: ROSE_LINEAR_ARRAY [STRING]; a_string1, a_string2: STRING): STRING is
		do
			Result := ""
			from
				a_fields.start
			until
				a_fields.after
			loop
					--	if valid_field (a_fields.current_item) then
				if Result.count = 0 then
					Result.append (a_string1)
						-- First time through.
				else
					Result.append (a_string2)
				end
				if a_string1.is_equal (":") then
					Result := Result + a_fields.current_item
				else
					Result := Result + " [" + a_fields.current_item + "]"
				end
					--	end
				a_fields.forth
			end
		end

	restriction_clause (a_fields: ROSE_LINEAR_ARRAY [STRING]): STRING is
		do
			Result := ""
			from
				a_fields.start
			until
				a_fields.after
			loop
				if Result.count > 0 then
					Result.append (" and ")
				end
				Result.append ("#" + a_fields.current_item + ".id = " + a_fields.current_item)
					-- First time through.
				a_fields.forth
			end
		end

	fields_from_string (a_fields_string: STRING): ROSE_LINEAR_ARRAY [STRING] is
			-- Array of key fields names
		local
			l_fields: ROSE_DELIMITED_TEXT
		do
			create l_fields.make (a_fields_string, ',')
			Result := l_fields
		end

	currency_imp: PERSISTENT_CURRENCY is
			-- The currency corresponding to this signature
		do
		end

feature {NONE} --Templates

	anchor: like Current is
			--template
		do
		end

	signature_template: like Current is
			--this is what to expect from data_manager
		do
		end

	expression_target: ROSE_EXPRESSION [like anchor] is
			--template
		do
		end

	expressions_anchor: GENERIC_ENTITY_VARIABLES [like Current, like expression_target] is
			-- Type anchor for the ENTITY_VARIABLES object available for this signature
		do
		end

	ieee_math: IEEE_MATH is
			-- Access to comparison routines
		once
			create Result
		end

feature {SIGNATURE_TAGGED_EXPRESSION} -- Implementation

	expression_cache_size: PAIR [INTEGER, INTEGER] is
			-- This is specified in descendents as PAIR [CACHE SIZE, CHUNK SIZE]		
		do
		end

end
