indexing
	description: "A storage object for record like data in a signature tagged database"
	author: "Mauricio Bustos"
	last_editor: "$Author: mb $"
	date: "$Date: 2008/12/14 $"
	revision: "$Revision: #16 $"
	path: "$File: //rose/source/kernel/persistence/databases/signature_tagged/rose_signature_record_item.e $"

deferred class
	ROSE_SIGNATURE_RECORD_ITEM
	
inherit
	DATA_RECORD_BASE
	COMPARABLE
		undefine
			is_equal
		end
	FIELD_DESCRIPTORS
	
feature {NONE} -- Creation
	
	make is
			-- Create the item
		deferred
		end
	
feature -- Access
	
	process_id: INTEGER
	
	fields_: STRING is
			-- Comma delimited string of attribute values storing data in this record item
		deferred
		ensure
			valid_result: Result /= Void
		end
	
	name: STRING is
			-- Name for the record to be used as the key in the ROSE_VIRTUAL_RECORD container
		deferred
		ensure
			valid_result: Result /= Void
		end
	
	key_fields: ROSE_LINEAR_ARRAY[STRING] is
			-- All field which uniquely identify one record but not including 'signature_id' or
			-- the self archive field.
			-- The index on this table should probably include 'signature_id' and then all key
			-- fields (in the order listed here).
			-- You may return Void (see 'is_indexed').
		do
		end
	
	archive_date_field_name: STRING is
			-- If this is non-Void then don't use the regular method of writing archive records
			-- to a history table.  Instead, store archive records in the same table.  Terminate
			-- old items by setting the end_date to new record's start_date.
			-- Generally you would use the name 'end_date'.
			-- The index on this table should probably include this field as the last field.
		do
		end
	
	all_record_fields: ROSE_LINEAR_ARRAY[STRING] is
		do
			Result := record_fields.twin
			Result.append_array (key_fields)
			if archive_date_field_name /= Void then
				Result.extend (archive_date_field_name)
			end
		ensure
			valid_result: Result /= Void
		end
	
	record_fields: ROSE_LINEAR_ARRAY[STRING] is
			-- The names of the fields which make up the data.  Not 'signature_id', 'process_id',
			-- key fields or archive_date fields.
		local
			l_type: INTEGER
		do
			l_type := internal.dynamic_type (Current)
			Result := record_fields_array.safe_item (l_type)
			if Result = Void then
				create {ROSE_DELIMITED_TEXT}Result.make (fields_,',')
				Result.compare_objects
				record_fields_array.force (Result,l_type)
			end
		ensure
			valid_result: Result /= Void
		end
	
	record_field_indices: ROSE_LINEAR_ARRAY [INTEGER] is
			-- The indices of the fields which make up the data.  Not 'signature_id', 'process_id',
			-- key fields or archive_date fields.
		local
			l_type, i: INTEGER
			l_fields: ROSE_DELIMITED_TEXT
		do
			l_type := internal.dynamic_type (Current)
			Result := record_fields_indices_array.safe_item (l_type)
			if Result = Void then
				create Result.make_empty
				create l_fields.make (fields_,',')
				from l_fields.start until l_fields.off loop
					from i := 1 until i > internal.field_count (Current) loop
						if internal.field_name (i, Current).as_lower.is_equal (l_fields.current_item.as_lower) then
							Result.extend (i)
							i := internal.field_count (Current) + 1
						else
							i := i + 1
						end
					end
					l_fields.forth
				end
				record_fields_indices_array.force (Result, l_type)
			end
		ensure
			valid_result: Result /= Void
		end
	
feature -- Comparison
	
	multi_key: MULTI_KEY is
		do
			--		create Result.make (<<each of your key_fields here>>)
		end
	
	infix "<" (other: like Current): BOOLEAN is
			-- Is other.hash_code < Current.hash_code?
		do
			if multi_key /= Void and then other.multi_key /= Void then
				Result := multi_key.hash_code < other.multi_key.hash_code
			end
		end
	
feature	-- Status
	
	is_indexed: BOOLEAN is
		do
			Result := key_fields /= Void
		end
	
	uses_self_archive: BOOLEAN is
		do
			Result := archive_date_field_name /= Void
		end
	
feature -- Element modification
	
	set_process_id (a_value: like process_id) is
		do
			process_id := a_value
		end
	
	set_data_fields (a_data: ARRAY [ANY]) is
			-- Set the data fields using 'a_data'
		require
			valid_data: a_data /= Void
		local
			l_index: INTEGER
			l_value: ANY
			l_fields: ROSE_LINEAR_ARRAY [STRING]
			l_fields_indices: ROSE_LINEAR_ARRAY [INTEGER]
		do
			if is_indexed or uses_self_archive then
				l_fields := all_record_fields
				from l_index := l_fields.lower
				until l_index > l_fields.upper
				loop
					l_value := a_data.item (a_data.upper - l_fields.upper + l_index)
					set_field (l_fields.item (l_index), l_value)
					l_index := l_index + 1
				end
			else
				l_fields_indices := record_field_indices
				from l_index := l_fields_indices.lower
				until l_index > l_fields_indices.upper
				loop
					l_value := a_data.item (a_data.upper - l_fields_indices.upper + l_index)
					set_field_with_index (l_fields_indices.item (l_index), l_value)
					l_index := l_index + 1
				end
			end
		end
	
feature -- Display
	
	output_for_verification: STRING is
			-- Used by STIM_VERIFICATION.  The returned STRING is just the concatenation of
			-- every bit of data in the record.  Called by ROSE_VIRTUAL_RECORD:output_for_verification.
		do
			Result := ""
			from
				record_fields.start
			until
				record_fields.off
			loop
				Result.append (",")
				Result.append (output_string_by_name (record_fields.current_item))
				record_fields.forth
			end
		end
	
	output_string_by_name (a_name: STRING): STRING is
			-- A display string for the data in field a_name
		local
			l_any: ANY
			l_date: ROSE_DATE
		do
			a_name.to_lower
			l_any := value_by_name (a_name)
			if l_any /= Void then
				l_date ?= l_any
				if l_date /= Void then
					Result := l_date.formatted_out ("[0]mm/[0]dd/yyyy")						
				else
					Result := l_any.out
				end
			else
				Result := "<Void>"
			end
		ensure
			valid_result: Result /= Void
		end
	
feature {NONE} -- Implementation
	
	record_fields_array: ROSE_LINEAR_ARRAY[ROSE_LINEAR_ARRAY[STRING]] is
		once
			create Result.make_empty
		end
	
	record_fields_indices_array: ROSE_LINEAR_ARRAY[ROSE_LINEAR_ARRAY[INTEGER]] is
		once
			create Result.make_empty
		end
	
end


