note
	description: "Describes a status of an object, with possible warning and errors"
	author: "Zoran Simic"
	last_editor: "$Author: sm $"
	date: "$Date: 2012/03/28 $"
	revision: "$Revision: #6 $"
	path: "$File: //rose/source/kernel/infrastructure/rose_status.e $"

class
	ROSE_STATUS

inherit
	ROSE_FORMATTING
		redefine
			copy, out
		end
	DEBUG_OUTPUT
		redefine
			copy, out
		end

feature -- Initialization

	copy (a_other: like Current) is
			-- Update current object using fields of object attached
			-- to `a_other', so as to yield equal objects.
		do
			reset
			if a_other.output_message /= Void then output_message := a_other.output_message.twin end
			if a_other.warning_message /= Void then warning_message := a_other.warning_message.twin end
			if a_other.error_message /= Void then error_message := a_other.error_message.twin end
			if a_other.messages /= Void then messages := a_other.messages.twin end
		end

	reset is
			-- Reset this status.
		do
			output_message := Void
			warning_message := Void
			error_message := Void
			messages := Void
		ensure
			is_ok: is_ok
		end

feature -- Access

	output_message: STRING
			-- Eventual output message

	warning_message: STRING
			-- Description of eventual warning

	error_message: STRING
			-- Description of eventual error

	messages: ARRAYED_LIST [ROSE_STATUS_MESSAGE]
			-- Outputs, warnings and errors in the order in which they were received

	path: STRING
			-- Path of the associated object

feature -- Query

	count: INTEGER is
			-- Number of messages
		do
			if messages /= Void then
				Result := messages.count
			end
		end

	error_count: INTEGER is
			-- Number of errors in this status message
		local
			i: INTEGER
		do
			if messages /= Void and then not messages.is_empty then
				from
					i := 1
				until
					i > messages.count
				loop
					if messages.i_th (i).is_error then
						Result := Result + 1
					end
					i := i + 1
				end
			end
		end

	warning_count: INTEGER is
			-- Number of warnings in this status message
		local
			i: INTEGER
		do
			if messages /= Void and then not messages.is_empty then
				from
					i := 1
				until
					i > messages.count
				loop
					if messages.i_th (i).is_warning then
						Result := Result + 1
					end
					i := i + 1
				end
			end
		end

	info_count: INTEGER is
			-- Number of informational entries in this status message
		local
			i: INTEGER
		do
			if messages /= Void and then not messages.is_empty then
				from
					i := 1
				until
					i > messages.count
				loop
					if messages.i_th (i).is_ok then
						Result := Result + 1
					end
					i := i + 1
				end
			end
		end

feature -- Basic operations

	outputf (a_message: STRING; a_args: ARRAY [ANY]) is
			-- Issue output with 'a_message' and 'a_args'.
		require
			message_ok: a_message /= Void and then not a_message.is_empty
		do
			output (formatter.argumented (a_message, a_args))
		end

	output (a_message: STRING) is
			-- Issue output with 'a_message'.
		require
			message_ok: a_message /= Void and then not a_message.is_empty
		do
			add_message (create {ROSE_STATUS_MESSAGE}.make_output (a_message))
		end

	warnf (a_message: STRING; a_args: ARRAY [ANY]) is
			-- Issue warning with 'a_message' and 'a_args'.
		require
			message_ok: a_message /= Void and then not a_message.is_empty
		do
			warn (formatter.argumented (a_message, a_args))
		end

	warn (a_message: STRING) is
			-- Issue warning with 'a_message'.
		require
			message_ok: a_message /= Void and then not a_message.is_empty
		do
			add_message (create {ROSE_STATUS_MESSAGE}.make_warning (a_message))
		end

	errorf (a_message: STRING; a_args: ARRAY [ANY]) is
			-- Issue error with 'a_message' and 'a_args'.
		require
			message_ok: a_message /= Void and then not a_message.is_empty
		do
			error (formatter.argumented (a_message, a_args))
		end

	error (a_message: STRING) is
			-- Issue error with 'a_message'.
		require
			message_ok: a_message /= Void and then not a_message.is_empty
		do
			add_message (create {ROSE_STATUS_MESSAGE}.make_error (a_message))
		end

	prepend_outputf (a_message: STRING; a_args: ARRAY [ANY]) is
			-- Issue output with 'a_message' and 'a_args'.
		require
			message_ok: a_message /= Void and then not a_message.is_empty
		local
			l_string: STRING
		do
			l_string := formatter.argumented (a_message, a_args)
			if messages = Void or else messages.is_empty then
				output (l_string)
			else
				messages.put_front (create {ROSE_STATUS_MESSAGE}.make_output (l_string))
				if output_message = Void then
					create output_message.make_empty
				end
				if not l_string.is_empty and then l_string.item (l_string.count) /= '%N' then
					output_message.prepend_character ('%N')
				end
				output_message.prepend_string (l_string)
			end
		end

	prepend_errorf (a_message: STRING; a_args: ARRAY [ANY]) is
			-- Issue error with 'a_message' and 'a_args'.
		require
			message_ok: a_message /= Void and then not a_message.is_empty
		local
			l_string: STRING
		do
			l_string := formatter.argumented (a_message, a_args)
			if messages = Void or else messages.is_empty then
				error (l_string)
			else
				messages.put_front (create {ROSE_STATUS_MESSAGE}.make_error (l_string))
				if error_message = Void then
					create error_message.make_empty
				end
				if not l_string.is_empty and then l_string.item (l_string.count) /= '%N' then
					error_message.prepend_character ('%N')
				end
				error_message.prepend_string (l_string)
			end
		end

	append (a_other: ROSE_STATUS) is
			-- Append 'a_other' to this status.
		do
			if a_other /= Void and then a_other.messages /= Void then
				from
					a_other.messages.start
				until
					a_other.messages.off
				loop
					add_message (a_other.messages.item)
					a_other.messages.forth
				end
			end
		end

	append_downgrading_errors (a_other: ROSE_STATUS) is
			-- Append 'a_other' to this status, downgrading errors to warnings
		local
			l_message: ROSE_STATUS_MESSAGE
		do
			if a_other /= Void and then a_other.messages /= Void then
				from
					a_other.messages.start
				until
					a_other.messages.off
				loop
					l_message := a_other.messages.item
					if l_message.is_error then
						create l_message.make_warning (l_message.text)
					end
					add_message (l_message)
					a_other.messages.forth
				end
			end
		end

feature -- Change

	set_path (a_path: STRING) is
			-- Set `path' to `a_path'
		do
			path := a_path
		ensure
			path_set: path = a_path
		end

feature -- Status Report

	is_acceptable: BOOLEAN is
			-- Is status at OK or warned state?
		do
			Result := error_message = Void
		end

	is_empty: BOOLEAN is
			-- Is this status empty? (ie, no messages)
		do
			Result := messages = Void or else messages.is_empty
		end

	is_ok: BOOLEAN is
			-- Is status at OK state?
		do
			Result := warning_message = Void and then error_message = Void
		end

	has_warning: BOOLEAN is
			-- Is status at warned state?
		do
			Result := warning_message /= Void
		end

	has_error: BOOLEAN is
			-- Is status at error state?
		do
			Result := error_message /= Void
		end

feature -- Output

	debug_output: STRING is
			-- String that should be displayed in debugger to represent `Current'.
		do
			create Result.make_empty
			if has_error then
				append_line (Result, "** ", error_message)
			end
			if has_warning then
				append_line (Result, "!! ", warning_message)
			end
			if output_message /= Void then
				append_line (Result, Void, output_message)
			end
		end

	out: STRING is
			-- String representing `Current'.
		do
			create Result.make_empty
			if has_error then
				append_line (Result, Void, error_message)
			end
			if has_warning then
				append_line (Result, Void, warning_message)
			end
			if output_message /= Void then
				append_line (Result, Void, output_message)
			end
		end

	representation: STRING is
			-- String representing `Current' with prefixes for errors and warnings, in the order they were added
		local
			i: INTEGER
		do
			if messages /= Void and then not messages.is_empty then
				from
					create Result.make_empty
					i := 1
				until
					i > messages.count
				loop
					append_line (Result, Void, messages.i_th (i).representation)
					i := i + 1
				end
			end
		end

feature {NONE} -- Implementation

	add_message (a_message: ROSE_STATUS_MESSAGE) is
			-- Add 'a_message' to this status.
		require
			message_ok: a_message /= Void
		do
			if messages = Void then
				create messages.make (1)
			end
			if a_message.is_error then
				if error_message = Void then
					create error_message.make_empty
				end
				if not error_message.is_empty and then error_message.item (error_message.count) /= '%N' then
					error_message.append_character ('%N')
				end
				error_message.append_string (a_message.text)
			elseif a_message.is_warning then
				if warning_message = Void then
					create warning_message.make_empty
				end
				if not warning_message.is_empty and then warning_message.item (warning_message.count) /= '%N' then
					warning_message.append_character ('%N')
				end
				warning_message.append_string (a_message.text)
			else
				if output_message = Void then
					create output_message.make_empty
				end
				if not output_message.is_empty and then output_message.item (output_message.count) /= '%N' then
					output_message.append_character ('%N')
				end
				output_message.append_string (a_message.text)
			end
			messages.extend (a_message)
		end

	append_line (a_result, a_prefix, a_string: STRING) is
			-- Append 'a_string' with 'a_prefix' to 'a_result'.
		require
			result_ok: a_result /= Void
		do
			if a_string /= Void and then not a_string.is_empty then
				a_result.append_string (a_prefix)
				a_result.append_string (a_string)
				if a_result.item (a_result.count) /= '%N' then
					a_result.append_character ('%N')
				end
			end
		end

end
