indexing
	description : "Stopwatch"
	author : "Zoran Simic"
	date: "$Date: 2009/08/20 $"
	revision: "$Revision: #4 $"

class
	ROSE_STOPWATCH

inherit
	ROSE_FORMATTING
		redefine
			out
		end
	ROSE_C_LIBRARY
		rename
			cpu_ticks as get_cpu_ticks
		redefine
			out
		end
	DEBUG_OUTPUT
		redefine
			out
		end

create
	make, make_averaged

feature {NONE} -- Initialization

	make (a_message: STRING) is
			-- Make a new stopwatch with 'a_message'
		do
			message := a_message
		end

	make_averaged (a_message: STRING) is
			-- Make a new stopwatch with 'a_message'
			-- Measures are averaged (if there are several calls to start/stop)
		do
			make (a_message)
			set_averaged (True)
		end

feature -- Access

	message: STRING
			-- Stopwatch message (represents what the time measure here stands for)

	cpu_milliseconds: DOUBLE
			-- Milliseconds of CPU time used in between calls to 'start' and 'stop'

	wallclock_milliseconds: DOUBLE
			-- Wallclock time elapsed in milliseconds between calls to 'start' and 'stop'

	count: INTEGER
			-- Number of measures taken (number of times start/stop was called)

	is_averaged: BOOLEAN
			-- Are the results averaged other the number of measures?
			-- If not averaged, they are just summed up

	is_running: BOOLEAN is
			-- Is this stopwatch currently measuring time?
		do
			Result := start_mark /= Void
		end

	out, debug_output: STRING is
			-- String that should be displayed in debugger to represent `Current'.
		do
			create Result.make (10)
			if count = 0 and then start_mark = Void then
				Result.append (k_no_measures)
			else
				if is_running then
					Result.append (k_in_progress)
				end
				Result.append_character ('[')
				Result.append (k_cpu)
				if is_averaged and then count > 1 then
					Result.append_character ('~')
					Result.append (formatter.formatted_double2 (total_cpu_milliseconds, -1, ','))
				else
					Result.append (total_cpu_milliseconds.truncated_to_integer.out)
				end
				Result.append (k_ms)
				Result.append (k_measure_separator)
				Result.append (k_wallclock)
				if is_averaged and then count > 1 then
					Result.append_character ('~')
					Result.append (formatter.formatted_double2 (total_wallclock_milliseconds, -1, ','))
				else
					Result.append (total_wallclock_milliseconds.truncated_to_integer.out)
				end
				Result.append (k_ms)
				Result.append_character (']')
			end
			if count > 1 then
				Result.append_character (' ')
				Result.append_character ('x')
				Result.append (count.out)
			end
			if message /= Void then
				Result.append_character (' ')
				Result.append (message)
			end
		end

	total_cpu_milliseconds: DOUBLE is
			-- Elapsed CPU time since time measure started
		do
			Result := cpu_milliseconds
			if is_averaged then
				Result := count * Result
			end
			if is_running then
				Result := Result + consumed_cpu_milliseconds
			end
		end

	total_wallclock_milliseconds: DOUBLE is
			-- Elapsed wallclock time since time measure started
		do
			Result := wallclock_milliseconds
			if is_averaged then
				Result := count * Result
			end
			if is_running then
				Result := Result + elapsed_wallclock_milliseconds
			end
		end

feature -- Basic operations

	start is
			-- Start measuring time.
		require
			not_running: not is_running
		do
			create start_mark.make_now
			cpu_mark := get_cpu_ticks
		ensure
			is_running: is_running
		end

	stop is
			-- Stop measuring time.
		require
			is_running: is_running
		do
			cpu_milliseconds := count * cpu_milliseconds + consumed_cpu_milliseconds
			wallclock_milliseconds := count * wallclock_milliseconds + elapsed_wallclock_milliseconds
			start_mark := Void
			count := count + 1
			if count > 1 then
				cpu_milliseconds := cpu_milliseconds / count
				wallclock_milliseconds := wallclock_milliseconds / count
			end
		ensure
			not_running: not is_running
		end

	reset is
			-- Reset this stopwatch.
		do
			cpu_milliseconds := 0.0
			wallclock_milliseconds := 0.0
			start_mark := Void
			cpu_mark := 0
			count := 0
		ensure
			not_running: not is_running
		end

	set_averaged (a_averaged: like is_averaged) is
			-- Set 'is_averaged'.
		do
			is_averaged := a_averaged
		end

feature {NONE} -- Implementation

	start_mark: DATE_TIME		-- Real world date and time when 'start' was called
	cpu_mark: INTEGER			-- Value of 'get_cpu_ticks' when 'start' was called

	consumed_cpu_milliseconds: INTEGER is
			-- Number of CPU milliseconds consumed since this stopwatch was started
		require
			is_running: is_running
		do
			Result := get_cpu_ticks - cpu_mark
		end

	elapsed_wallclock_milliseconds: DOUBLE is
			-- Number of real-world milliseconds elapsed since this stopwatch was started
		require
			is_running: is_running
		do
			Result := (create {DATE_TIME}.make_now).relative_duration (start_mark).fine_seconds_count * 1000
		end

	k_no_measures: STRING is "[no measures]"	-- no measures
	k_in_progress: STRING is "[in progress] "	-- in progress
	k_measure_separator: STRING is ", "			-- measure separator
	k_ms: STRING is " ms"						-- milliseconds (ms)
	k_cpu: STRING is "cpu:"						-- cpu:
	k_wallclock: STRING is "wck:"				-- wallclock:

end
