note
	description:
		"Store/Retrieve Eiffel objects using STORABLE, and with error checking"
	author: "Zoran Simic"
	last_editor: "$Author: dz $"
	date: "$Date: 2013/10/31 $"
	revision: "$Revision: #28 $"
	path: "$File: //rose/source/kernel/infrastructure/framework/support/rose_storable_factory.e $"

class ROSE_STORABLE_FACTORY

inherit

	ROSE_SHARED_GLOBAL

	ROSE_IO_CONVENTION
		export {NONE} all end

	KL_SHARED_FILE_SYSTEM
		export {NONE} all end

feature -- Access

	deserialized (a_file_name: STRING): ANY is
			-- Deserialized Eiffel object from file pointed by 'a_file_name'
			-- Eventual errors are signaled
		require
			valid_filename: a_file_name /= Void
		do
			error.disregard
			Result := deserializer.deserialized (a_file_name)
		ensure
			valid_xor_error: Result /= Void xor error.signaled
		end

	deserialized_string (a_string: STRING): ANY is
			-- Deserialized Eiffel object from 'a_string'
			-- Eventual errors are signaled
		require
			valid_string: a_string /= Void
		local
			l_file: FILE
			retried: INTEGER
		do
			if retried < 3 then
				error.disregard
				l_file := new_temporary_file
				l_file.open_write
				l_file.put_string (a_string)
				l_file.flush
				l_file.close
				Result := deserialized (l_file.name)
				delete_file (l_file)
			elseif not error.signaled then
				error.signal (generator, "'deserialized_string' could not succeed for '$1'", <<a_string>>)
				error.flush
			end
		ensure
			valid_xor_error: Result /= Void xor error.signaled
		rescue
			retried := retried + 1
			retry
		end

	serialized_string (a_object: ANY): STRING is
			-- Serialized XML representation of 'a_object'
			-- Eventual errors are signaled
		require
			valid_object: a_object /= Void
		local
			l_file: FILE
			retried: INTEGER
		do
			if retried < 3 then
				error.disregard
				l_file := new_temporary_file
				serialize (a_object, l_file.name)
				if not error.signaled then
					from
						create Result.make (l_file.count)
						l_file.open_read
						l_file.read_line
					until
						l_file.end_of_file
					loop
						Result.append (l_file.last_string)
						Result.append ("%N")
						l_file.read_line
					end
					l_file.close
					delete_file (l_file)
				end
			elseif not error.signaled then
				error.signal (generator, "'serialized_string' could not succeed for '$1'", <<a_object>>)
				error.flush
			end
		ensure
			valid_xor_error: Result /= Void xor error.signaled
		rescue
			retried := retried + 1
			retry
		end

	retrieved_relative (a_relative_path: STRING): ANY is
			-- Storable retrived from 'a_relative_path'
		require
			is_relative_path: global.formatter.is_relative_path (a_relative_path)
		do
			Result := retrieved_relative_verbose (a_relative_path, False)
		ensure
			valid_xor_error: Result /= Void xor error.signaled
		end

	retrieved_relative_verbose (a_relative_path: STRING; a_verbose: BOOLEAN): ANY is
			-- Storable retrived from 'a_relative_path',
			-- output persistence paths if requested
		require
			is_relative_path: global.formatter.is_relative_path (a_relative_path)
		local
			l_absolute_path, l_archive_path: STRING
			l_file: RAW_FILE
		do
			l_absolute_path := global.app.storable_path (a_relative_path)
			l_absolute_path := read_storable_filename (l_absolute_path)
			create l_file.make (l_absolute_path)
			if l_file.exists then
				Result := retrieved (l_absolute_path)
				if a_verbose then
					if Result /= Void then
						output_timestamped_utc ("Successfully retrieved from $1%N", <<l_absolute_path>>)
					else
						output_timestamped_utc ("Could not retrieve from $1%N", <<l_absolute_path>>)
					end
				end
			else
				l_archive_path := global.app.archive_path (a_relative_path)
				l_file.make (l_archive_path)
				if l_file.exists then
					Result := retrieved (l_archive_path)
					if a_verbose then
						if Result /= Void then
							output_timestamped_utc ("Successfully retrieved from archive path $1%N", <<l_archive_path>>)
						else
							output_timestamped_utc ("Could not retrieve from archive path $1%N", <<l_archive_path>>)
						end
					end
				else
					error.signal (generator, "Cannot retrieve an Eiffel object from '$1' or '$2'", <<l_absolute_path, l_archive_path>>)
					error.flush
				end
			end
		ensure
			valid_xor_error: Result /= Void xor error.signaled
		end

	retrieved (a_filename: STRING): ANY is
			-- Retrieved Eiffel object from file with'a_filename'
			-- Eventual errors are signaled
		require
			valid_filename: a_filename /= Void
		do
			Result := rp_serializer.retrieved (a_filename)
		ensure
			valid_xor_error: Result /= Void xor error.signaled
		end

	last_retrieved_filename: detachable STRING is
			-- filename of last retrieved object
		do
			Result := rp_serializer.last_retrieved_filename
		end

	retrieved_metadata (a_filename: STRING): detachable RP_SERIALIZABLE_METADATA
			-- Retrieved serializable metadata information from file `a_filename'
		require
			valid_filename: a_filename /= Void
		do
			Result := rp_serializer.retrieved_metadata (a_filename)
		ensure
			error_implies_void: error.signaled implies Result = Void
		end

	size_of (a_object: ANY): INTEGER is
			-- 'Deep' size of 'a_object' in bytes
		local
			refs: HASH_TABLE [BOOLEAN, POINTER]
			l_collecting: BOOLEAN
		do
			if memory.collecting then
				l_collecting := True
				memory.collection_off
			end
			create refs.make (0)
			Result := recursive_size_of (a_object, refs)
			if l_collecting then
				memory.collection_on
			end
		rescue
			if l_collecting then
				memory.collection_on
			end
		end

	read_storable_filename (a_filename: STRING): STRING
			-- Overridden storable filename of `a_filename' if there exists a file with that filename;
			-- `a_filename' otherwise
		require
			a_filename_not_void: a_filename /= Void
		local
			l_filename: STRING
			l_overridden_filename: STRING
			l_file: RAW_FILE
		do
			l_filename := Global.interpreted (a_filename)
			if l_filename /= Void then
				Result := l_filename
				l_overridden_filename := overridden_storable_filename (l_filename)
				if l_overridden_filename /= Void and then not l_overridden_filename.is_empty then
					create l_file.make (l_overridden_filename)
					if l_file.exists and then l_file.is_readable then
						Result := l_overridden_filename
					end
				end
			else
				Result := a_filename
			end
		ensure
			read_storable_filename_not_void: Result /= Void
		end

	write_storable_filename (a_filename: STRING): STRING
			-- Overridden storable filename of `a_filename' is not Void;
			-- `a_filename' otherwise
		require
			a_filename_not_void: a_filename /= Void
		local
			l_filename: STRING
			l_overridden_filename: STRING
		do
			l_filename := Global.interpreted (a_filename)
			if l_filename /= Void then
				Result := l_filename
				l_overridden_filename := overridden_storable_filename (l_filename)
				if l_overridden_filename /= Void and then not l_overridden_filename.is_empty then
					Result := l_overridden_filename
				end
			else
				Result := a_filename
			end
		ensure
			write_storable_filename_not_void: Result /= Void
		end

	overridden_storable_filename (a_filename: STRING): STRING
			-- Version of `a_filename' in the override storable path
			-- if it is in the storable path, Void otherwise
		require
			a_filename_not_void: a_filename /= Void
		local
			l_storable_path: STRING
			l_override_storable_path: STRING
			l_new_filename: STRING
		do
			l_override_storable_path := Global.app.configured_value (configuration_path_override_storable)
			l_storable_path := Global.app.configured_value (configuration_path_storable)
			if l_override_storable_path /= Void and l_storable_path /= Void then
				l_storable_path := file_system.canonical_pathname (Global.app.interpreted (l_storable_path))
				if not l_storable_path.is_empty then
					l_new_filename := file_system.canonical_pathname (a_filename)
					if file_system.is_subpathname (l_storable_path, l_new_filename) then
						l_new_filename.remove_head (l_storable_path.count)
						l_override_storable_path := file_system.canonical_pathname (Global.app.interpreted (l_override_storable_path))
						if not l_override_storable_path.is_empty then
							Result := l_override_storable_path + l_new_filename
						end
					end
				end
			end
		end

feature -- Basic operations

	serialize (a_object: ANY; a_file_name: STRING) is
			-- Serialize 'a_object' onto 'a_file_name'.
			-- Eventual errors are signaled
		require
			valid_object: a_object /= Void
			valid_filename: a_file_name /= Void
		do
			if a_object /= Void and then a_file_name /= Void then
				error.disregard
				serializer.serialize (a_object, a_file_name)
			end
		end

	store_relative (a_object: ANY; a_relative_path: STRING) is
			-- Counterpart of retrieved_relative except that there is no fallback to archive location/
			-- i.e. It is always global.app.storable_path
		require
			valid_object: a_object /= Void
			is_relative_path: global.formatter.is_relative_path (a_relative_path)
		local
			l_absolute_path: STRING
		do
			l_absolute_path := global.app.storable_path (a_relative_path)
			store (a_object, l_absolute_path)
		end

	store (a_object: ANY; a_filename: STRING) is
			-- Store 'a_object' to 'a_filename'.
			-- Store metadata if the variable "storable_metadata" is
			-- set to true in the config file of the application.
			-- 'a_filename' is interpreted through configuration
			-- An eventual error is signaled, no exception propagated
		require
			valid_object: a_object /= Void
			valid_filename: a_filename /= Void
		do
			if Global.app.configured_boolean_value (storable_metadata_key, False) then
				store_with_metadata (a_object, a_filename)
			else
				store_without_metadata (a_object, a_filename)
			end
		end

	store_with_metadata (a_object: ANY; a_filename: STRING) is
			-- Store 'a_object' with metadata to 'a_filename'.
			-- 'a_filename' is interpreted through configuration
			-- An eventual error is signaled, no exception propagated
		require
			valid_object: a_object /= Void
			valid_filename: a_filename /= Void
		do
			rp_serializer.store (a_object, a_filename)
		end

	store_without_metadata (a_object: ANY; a_filename: STRING) is
			-- Store 'a_object' with no metadata to 'a_filename'.
			-- 'a_filename' is interpreted through configuration
			-- An eventual error is signaled, no exception propagated
		require
			valid_object: a_object /= Void
			valid_filename: a_filename /= Void
		do
			rp_serializer.store_without_metadata (a_object, a_filename)
		end

feature {NONE} -- Implementation

	rp_serializer: RP_SERIALIZER [ANY] is
			-- Serialization implementation object
		once
			create Result
		ensure
			rp_serializer_not_void: Result /= Void
		end

	serializer: ROSE_SERIALIZER is
			-- Serialization implementation object
		once
			create Result
		end

	deserializer: ROSE_DESERIALIZER is
			-- Deserialization implementation object
		once
			create Result
		end

	internal: INTERNAL is
			-- Access to runtime internals
		once
			create Result
		end

	memory: MEMORY is
			-- Interface to garbage collector
		once
			create Result
		end

	recursive_size_of (a_object: ANY; refs: HASH_TABLE [BOOLEAN, POINTER]): INTEGER is
			-- Recursive size of 'a_object', already explored references are stored in 'refs'
		require
			refs_ok: refs /= Void
		local
			i, j, n, l_type: INTEGER
			sa: SPECIAL [ANY]
			l_field: ANY
		do
			if a_object /= Void and then not refs.item ($a_object) then
				refs.put (True, $a_object)
				Result := internal.physical_size (a_object)
				from
					n := internal.field_count (a_object)
					i := 1
				until
					i > n
				loop
					l_type := internal.field_type (i, a_object)
					if l_type = internal.reference_type then
						l_field := internal.field (i, a_object)
						if l_field /= Void and then not refs.item ($l_field) then
							Result := Result + recursive_size_of (l_field, refs)
							if internal.is_special_any_type (internal.dynamic_type (l_field)) then
								sa ?= l_field
								if sa /= Void then
									from
										j := 0
									until
										j >= sa.count
									loop
										Result := Result + recursive_size_of (sa.item (j), refs)
										j := j + 1
									end
								end
							end
						end
					end
					i := i + 1
				end
			end
		end

	new_temporary_file: FILE is
			-- New temporary file
		local
			l_temp_fname: FILE_NAME
		do
			from
				create l_temp_fname.make_temporary_name
			until
				Result /= Void
			loop
				create {PLAIN_TEXT_FILE} Result.make (l_temp_fname.string)
				if Result.exists then
					Result := Void
					create l_temp_fname.make_temporary_name
				end
			end
		end

	delete_file (a_file: FILE) is
			-- Delete 'a_file', retrying if failing
		local
			retried: INTEGER
		do
			if retried < 3 and then a_file /= Void and then a_file.exists then
				a_file.delete
			end
		rescue
			retried := retried + 1
			retry
		end

feature {NONE} -- Constants

	storable_metadata_key: STRING = "storable_metadata"
			-- Key in config file used to indicate whether metadata should be stored in storable files

end
