indexing
	description: "Time measuring utility class"

	author: "Zoran Simic"
	last_editor: "$Author: jq $"
	date: "$Date: 2009/10/20 $"
	revision: "$Revision: #14 $"
	path: "$File: //rose/source/kernel/infrastructure/framework/io/rose_time_measure.e $"

class ROSE_TIME_MEASURE

inherit

	ROSE_SHARED_GLOBAL
		export
			{NONE} all
		end

	ROSE_C_LIBRARY

create
	make_started

feature {NONE} -- Initialization

	make_started (a_id, a_qualification: STRING) is
			-- Make an identified time measure.
		do
			id := a_id
			qualification := a_qualification
			start
		ensure
			id_set: id = a_id
			qualification_set: qualification = a_qualification
		end

feature -- Access

	id: STRING
			-- Identification of the event

	qualification: STRING
			-- Qualification of the event

	start_time: DATE_TIME
			-- Actual start time (wall clock)

	user_duration: DATE_TIME_DURATION
			-- Duration in seconds, user observed (wall clock seconds)

	elapsed_cpu: STRING is
			-- Elapsed CPU time since time measure started
		require
			started: start_time /= Void
		do
			if user_duration = Void then
				Result := milli_formatter.formatted ((cpu_ticks-cpu) / global.env.cpu_ticks_per_second)
			else
				Result := milli_formatter.formatted (cpu / global.env.cpu_ticks_per_second)
			end
		ensure
			elapsed_cpu_not_void: Result /= Void
		end

	elapsed_wallclock: STRING is
			-- Elapsed wallclock time since time measure started
		require
			started: start_time /= Void
		local
			l_now: DATE_TIME
			l_duration: DATE_TIME_DURATION
		do
			if user_duration = Void then
				create l_now.make_now_utc
				l_duration := l_now.relative_duration (start_time)
				Result := milli_formatter.formatted (l_duration.time.fine_seconds_count)
			else
				Result := milli_formatter.formatted (user_duration.time.fine_seconds_count)
			end
		ensure
			elapsed_wallclock_not_void: Result /= Void
		end

feature -- Basic operations

	start is
			-- Start the measure.
		require
			not_started: start_time = Void
		do
			create start_time.make_now_utc
			cpu := cpu_ticks
		ensure
			started: start_time /= Void
		end

	terminate is
			-- Terminate the measure.
		require
			started: start_time /= Void
		local
			l_now: DATE_TIME
		do
			cpu := cpu_ticks - cpu
			create l_now.make_now_utc
			user_duration := l_now.relative_duration (start_time)
		ensure
			duration_measured: user_duration /= Void
		end

feature -- Output

	progress_report (a_description: STRING): STRING is
			-- String report of current progress (with `a_description' if not Void)
		do
			create Result.make (100)
			if start_time = Void then
				Result.append (k_not_started)
			else
				if user_duration = Void then
					Result.append (k_in_progress)
				end
				Result.append_character ('[')
				Result.append (k_cpu)
				Result.append (elapsed_cpu)
				Result.append (k_measure_separator)
				Result.append (k_user)
				Result.append (elapsed_wallclock)
				Result.append_character (']')
			end
			Result.append_character (' ')
			Result.append_string (id)
			if a_description /= Void then
				Result.append_character (' ')
				Result.append (a_description)
			elseif qualification /= Void then
				Result.append_character (' ')
				Result.append (qualification)
			end
			Result.append_character ('%N')
		ensure
			progress_report_not_void: Result /= Void
		end

	status: STRING is
			-- Status of this time measure
		do
			Result := progress_report (qualification)
		ensure
			status_not_void: status /= Void
		end

feature {NONE} -- Implementation

	cpu: INTEGER

	k_not_started: STRING is "[not started]"
	k_in_progress: STRING is "[in progress]"
	k_measure_separator: STRING is ", "
	k_cpu: STRING is "cpu:"
	k_user: STRING is "user:"

	milli_formatter: FORMAT_DOUBLE is
			-- Formatter to display times up to milliseconds
		once
			create Result.make (7,3)
		ensure
			milli_formatter_not_void: Result /= Void
		end

end
