indexing

	description: "[
                  Time series container and related operations.
                  Use cautiously. Dates and data can be altered
                  separately. 
]"

	author: "Michael Zhang"
	last_editor: "$Author: er $"
	date: "$Date: 2013/06/18 $"
	revision: "$Revision: #31 $"
	path: "$File: //rose/source/kernel/infrastructure/containers/rose_time_series.e $"
	storable_version: ""

class ROSE_TIME_SERIES [G]

inherit

	ROSE_LINEAR_ARRAY [G]
		rename
			make as rose_linear_make,
			extend as rose_linear_extend
		redefine
			make_empty,
			subarray,
			selected_items,
			subset,
			subset_loose,
			random_subarray
		end

	ROSE_LEGACY_UNDEFINED
		undefine
			copy,
			is_equal
		end

create

	make,
	make_empty,
	make_with_dates,
	make_with_agent

feature {NONE} -- Initialization

	make (a_series: ARRAY [G]; a_dates: ARRAY [ROSE_DATE]) is
			-- Create with data items form `a_series' and dates from `a_dates'.
		require
			same_length: (a_series /= Void and a_dates /= Void) and then a_series.count = a_dates.count
		do
			make_from_array (a_series)
			set_dates (a_dates)
		ensure
			correct_count: count = a_series.count
			dates_set: dates.same_items (a_dates)
			same_count_for_dates: dates.count = count
		end

	make_with_dates (a_dates: ARRAY [ROSE_DATE]) is
			-- Create with dates from `a_dates' and default data items.
		require
			valid_dates: a_dates /= Void and then not a_dates.is_empty
		do
			rose_linear_make (1, a_dates.count)
			set_dates (a_dates)
		ensure
			correct_count: count = a_dates.count
			dates_set: dates.same_items (a_dates)
			same_count_for_dates: dates.count = count
		end

	make_with_agent (a_agent: FUNCTION [ANY, TUPLE [ROSE_DATE], G]; a_dates: ARRAY [ROSE_DATE]) is
			-- Create with dates from `a_dates' and data items constructed by `a_agent'.
		require
			valid_dates: a_dates /= Void and then a_dates.count > 0
			valid_agent: a_agent /= Void
		local
			l_date: ROSE_DATE
			l_value: G
		do
			rose_linear_make (1, a_dates.count)
			set_dates (a_dates)
			from
				dates.start
			until
				dates.after
			loop
				l_date := dates.current_item
				l_value := a_agent.item ([l_date])
				put (l_value, dates.index)
				dates.forth
			end
		ensure
			correct_count: count = a_dates.count
			dates_set: dates.same_items (a_dates)
			same_count_for_dates: dates.count = count
		end

	make_empty is
			-- Make with no data items or dates.
		do
			Precursor
			create dates.make_empty
			create date_table.make (10)
		ensure then
			empty: count = 0
			same_count_for_dates: dates.count = count
			upper_0: upper = 0
		end

feature -- Status report

	is_sorted_ascending: BOOLEAN is
			-- Does `Current' represent (at least for the part covered by `dates') an ascending time series?
		do
			if is_empty then
				Result := False
			elseif first_date = Void or last_date = Void then
				Result := False
			else
				Result := dates.for_all_with_index (agent is_later_than_previous)
			end
		ensure
			empty_array_not_sorted_ascending: is_empty implies not Result
			empty_time_series_not_sorted_ascending: dates.is_empty implies not Result
			no_void_date: Result implies not dates.has (Void)
		end

feature -- Basic operations

	extend (a_item: G; a_date: ROSE_DATE) is
			-- Add `a_item' as last datum of `Current'.
		do
			rose_linear_extend (a_item)
			dates.extend (a_date)
			invalidate_cache
			if a_date /= Void then
				date_table.put (count, a_date.ccyymmdd)
			end
		ensure
			one_more: count = old count + 1
			one_more_date: dates.count = old dates.count + 1
			last_item: equal (a_item, last)
			last_date: equal (a_date, last_date)
			date_set_will_be_rebuilt_on_demand: date_set_saved = Void
		end

	put_at (a_item: G; a_date: ROSE_DATE) is
			-- If `a_date' corresponds to a datum, then replace said datum with `a_item'.
		require
			a_date_not_void: a_date /= Void
		local
			l_ix: INTEGER
		do
			l_ix := index_at (a_date)
			if valid_index (l_ix) then
				put (a_item, l_ix)
			end
		end

	append (a_series: like Current) is
			-- Add all items from `a_series' to end of `Current'.
		require
			a_series_not_void: a_series /= Void
		do
			from
				a_series.start
			until
				a_series.after
			loop
				extend (a_series.current_item, a_series.current_date)
				a_series.forth
			end
		ensure
			correct_count: count = old count + a_series.count
			correct_dates_count: dates.count = old dates.count + a_series.count -- but not necessarily `a_series.dates.count'.
			date_set_will_be_rebuilt_on_demand: date_set_saved = Void
		end

feature -- Duplication

	cloned: like Current is
			-- Copy of `Current'.
		do
			create Result.make (Current, dates)
		ensure
			cloned_not_void: Result /= Void
			same_count: Result.count = count
			same_dates: Result.dates.same_items (dates)
			same_count_for_dates: Result.dates.count = dates.count
		end

feature -- Access

	first_date: ROSE_DATE is
			-- Date (if any) associated with first datum in `Current';
			-- N.B. This is not necessarily the same as `dates.first'.
		require
			not_empty_time_series: not dates.is_empty
		do
			Result := dates.item (lower)
		end

	last_date: ROSE_DATE is
			-- Date (if any) associated with last datum in `Current';
			-- N.B. This is not necessarily the same as `dates.last'.
		require
			not_empty_time_series: not dates.is_empty
		do
			Result := dates.item (upper)
		end

	current_date: ROSE_DATE is
			-- Date (if any) corresponding to current position in `Current'
		do
			Result := date (index)
		end

	date (a_ix: INTEGER): ROSE_DATE is
			-- Date (if any) corresponding to position `a_ix' (if valid) in `Current'
		do
			if dates.valid_index (a_ix) then
				Result := dates.item (a_ix)
			end
		end

	item_at (a_date: ROSE_DATE): G is
			-- Datum indexed by `a_date'
		require
			a_date_not_void: a_date /= Void		
		local
			l_ix: INTEGER
		do
			l_ix := index_at (a_date)
			if valid_index (l_ix) then
				Result := item (l_ix)
			end
		end

	item_at_or_before (a_date: ROSE_DATE): G is
			-- Datum indexed by `date_at_or_before'
		require
			a_date_not_void: a_date /= Void
			ascending_order: is_sorted_ascending
		local
			l_date: ROSE_DATE
		do
			l_date := date_at_or_before (a_date)
			if l_date /= Void then
				Result := item_at (l_date)
			end
		end

	item_at_or_after (a_date: ROSE_DATE): G is
			-- Datum indexed by `date_at_or_after'
		require
			a_date_not_void: a_date /= Void
			ascending_order: is_sorted_ascending
		local
			l_date: ROSE_DATE
		do
			l_date := date_at_or_after (a_date)
			if l_date /= Void then
				Result := item_at (l_date)
			end
		end

	item_before (a_date: ROSE_DATE): G is
			-- Datum indexed by latest date prior to `a_date'
		require
			a_date_not_void: a_date /= Void
			ascending_order: is_sorted_ascending
		local
			l_date: ROSE_DATE
		do
			l_date := date_before (a_date)
			if l_date /= Void then
				Result := item_at (l_date)
			end
		end

	item_after (a_date: ROSE_DATE): G is
			-- Datum indexed by earliest date after `a_date'
		require
			a_date_not_void: a_date /= Void
			ascending_order: is_sorted_ascending
		local
			l_date: ROSE_DATE
		do
			l_date := date_after (a_date)
			if l_date /= Void then
				Result := item_at (l_date)
			end
		end

	items_between (a_date: ROSE_DATE; other_date: ROSE_DATE): ROSE_TIME_SERIES [G] is
			-- Items between 2 dates (inclusive)
		require
			a_date_not_void: a_date /= Void
			other_date_not_void: other_date /= Void
			dates_ordered: a_date <= other_date
			ascending_order: is_sorted_ascending
		local
			l_date: ROSE_DATE
		do
			create Result.make_empty
			if not is_empty then
				from
					l_date := date_at_or_after (a_date)
				until
					l_date = Void or else l_date > other_date
				loop
					Result.extend (item_at (l_date), l_date)
					l_date := strict_date_after (l_date)
				end
			end
		ensure
			items_between_not_void: Result /= Void
			date_ordered: not Result.is_empty implies Result.is_sorted_ascending
		end

	items_for_month (a_month: ROSE_MONTH): like items_between is
			-- Items included within `a_month'
		require
			a_month_not_void: a_month /= Void
			ascending_order: is_sorted_ascending
		do
			Result := items_between (a_month.month_beginning, a_month.next_month.previous_month_end)
		ensure
			items_for_month_not_void: Result /= Void
			date_ordered: not Result.is_empty implies Result.is_sorted_ascending
		end

	has_date (a_date: ROSE_DATE): BOOLEAN is
			-- Does this time series contain `a_date'?
		do
			Result := a_date /= Void and then date_table.has (a_date.ccyymmdd)
		end

	index_at (a_date: ROSE_DATE): INTEGER is
			-- Index of `a_date' within `dates'
		require
			a_date_not_void: a_date /= Void
		do
			date_table.search (a_date.ccyymmdd)
			if date_table.found then
				Result := date_table.found_item
			end
		ensure
			index_non_negative: Result >= 0
			item_not_guaranteed: True -- and NOT: Result > 0 implies valid_index (Result)
		end

	index_before (a_date: ROSE_DATE): INTEGER is
			-- Index of datum indexed by latest date prior to `a_date'
		require
			a_date_not_void: a_date /= Void
			ascending_order: is_sorted_ascending
		local
			l_date: ROSE_DATE
		do
			l_date := date_before (a_date)
			if l_date /= Void then
				Result := index_at (l_date)
			end
		ensure
			index_non_negative: Result >= 0
			item_not_guaranteed: True -- and NOT: Result > 0 implies valid_index (Result)
		end

	index_after (a_date: ROSE_DATE): INTEGER is
			-- Index of datum indexed by earliest date after `a_date'
		require
			a_date_not_void: a_date /= Void
			ascending_order: is_sorted_ascending
		local
			l_date: ROSE_DATE
		do
			l_date := date_after (a_date)
			if l_date /= Void then
				Result := index_at (l_date)
			end
		ensure
			index_non_negative: Result >= 0
			item_not_guaranteed: True -- and NOT: Result > 0 implies valid_index (Result)			
		end

	date_before (a_date: ROSE_DATE): ROSE_DATE is
			-- Latest date in `dates' that is before `a_date', or rubbish (19001010) if there is no such date;
			-- N.B. This calls `{ROSE_DATE_SET}.ith_before' which is badly specified, so the semantics
			--  may not match this description.
		require
			a_date_not_void: a_date /= Void
			ascending_order: is_sorted_ascending
		do
			Result := date_set.ith_date_before (a_date, 1)
		end

	date_after (a_date: ROSE_DATE): ROSE_DATE is
			-- Earliest date in `dates' that is after `a_date'
		require
			a_date_not_void: a_date /= Void
			ascending_order: is_sorted_ascending
		do
			Result := date_set.ith_date_after (a_date, 1)
		ensure
			later_date: Result /= Void implies (Result > a_date)
			date_after_in_dates: Result /= Void implies dates.has (Result)
		end

	date_at_or_before (a_date: ROSE_DATE): ROSE_DATE is
			-- `date' if it is in `dates', else latest date in `dates' that is before `a_date',
			--  or rubbish (1900101) if no such date;
			-- N.B. This eventually calls `{ROSE_DATE_SET}.ith_before' which is badly specified, so the semantics
			--  may not match this description.
		require
			a_date_not_void: a_date /= Void
			ascending_order: is_sorted_ascending
		do
			Result := date_set.at_or_before (a_date)
		end

	date_at_or_after (a_date: ROSE_DATE): ROSE_DATE is
			-- `date' if it is in `dates', else earliest date in `dates' that is after `a_date'
		require
			a_date_not_void: a_date /= Void
			ascending_order: is_sorted_ascending
		do
			Result := date_set.at_or_after (a_date)
		ensure
			same_date_or_later: Result /= Void implies (Result >= a_date)
		end

feature -- Derived time series

	lag_series (a_n: INTEGER): like Current is
			-- Sub-series of `Current' from positions 1 to (count - `a_n') inclusive
		require
			a_n_large_enough: a_n >= 1
			a_n_small_enough: a_n < count
			valid_dates_index: (count - a_n) <= dates.upper
		do
			Result := subindex_series (1, count - a_n)
		ensure
			lag_series_not_void: Result /= Void
			correct_count: Result.count = upper - a_n
			correct_dates_count: Result.dates.count = Result.count
		end

	subindex_series (a_ix_1, a_ix_2: INTEGER): like Current is
			-- Sub-series of `Current' from positions `a_ix_1' to `a_ix_2' inclusive
		require
			valid_index: a_ix_1 >= 1 and a_ix_1 <= a_ix_2 and a_ix_2 <= upper
			valid_dates_index: a_ix_2 <= dates.upper
		do
			create Result.make (Current.subarray (a_ix_1, a_ix_2), dates.subarray (a_ix_1, a_ix_2))
		ensure
			subindex_series_not_void: Result /= Void
			correct_count: Result.count = (a_ix_2 - a_ix_1 + 1)
			correct_dates_count: Result.dates.count = Result.count
		end

	subseries_by_date_range (a_start_date, a_end_date: ROSE_DATE): like Current is
			-- Sub-series from `index_after (a_start_date)' to `index_before (a_end_date)'
		require
			a_start_date_not_void: a_start_date /= Void
			a_end_date_not_void: a_end_date /= Void
			valid_range: a_start_date <= a_end_date
			ascending_order: is_sorted_ascending
		local
			l_start_ix, l_end_ix: INTEGER
		do
			l_start_ix := index_at (a_start_date)
			if l_start_ix = 0 then
				l_start_ix := index_after (a_start_date)
			end
			l_end_ix := index_at (a_end_date)
			if l_end_ix = 0 then
				l_end_ix := index_before (a_end_date)
			end
			if (l_start_ix > 0 and l_end_ix > 0) and then l_start_ix <= l_end_ix then
				Result := subindex_series (l_start_ix, l_end_ix)
			end
		end

	subdate_series (a_dates: ARRAY [ROSE_DATE]): like Current is
			-- Sub-series of `Current' selected by calling `index_at' on each item in `a_dates', or
			--  using default items when this does not produce a valid index
		require
			a_dates_not_void: a_dates /= Void
			a_dates_not_empty: not a_dates.is_empty
			no_void_date: not a_dates.has (Void)
			a_dates_late_enough:	a_dates.item (a_dates.lower) >= dates.first
			a_dates_early_enough: a_dates.item (a_dates.upper) <= dates.last
		local
			l_data: ROSE_LINEAR_ARRAY [G]
			l_date: ROSE_DATE
			l_item: G
			l_default_item: G
			i, l_ix: INTEGER
		do
			create l_data.make_empty
			from
				i := 1
			until
				i > a_dates.count
			loop
				l_date := a_dates.item (i)
				l_ix := index_at (l_date)
				if valid_index (l_ix) then
					l_item := item (l_ix)
				else
					l_item := l_default_item
				end
				l_data.extend (l_item)
				i := i + 1
			end
			create Result.make (l_data, a_dates)
		ensure
			subdate_series_not_void: Result /= Void
			same_count: Result.count = a_dates.count
			dates_set: Result.dates.same_items (a_dates)
		end

	monthly_series (a_dates: ARRAY [ROSE_DATE]): like Current is
			-- Version of `Current' with `dates' set from `a_dates' if in `dates' with valid indices,
			--  or else according to `index_before';
			-- This has nothing to do with months, despite the name. Indeed, it resembles nothing much.
			-- Don't use it!
		require
			a_dates: a_dates /= Void
			no_void_date: not a_dates.has (Void)
			ascending_order: is_sorted_ascending
		local
			l_data: ROSE_LINEAR_ARRAY [G]
			l_date: ROSE_DATE
			l_dates: ROSE_LINEAR_ARRAY [ROSE_DATE]
			l_item: G
			i, l_ix: INTEGER
		do
			create l_data.make_empty
			create l_dates.make_empty
			from
				i := 1
			until
				i > a_dates.count
			loop
				l_date := a_dates.item (i)
				l_ix := index_at (l_date)
				if valid_index (l_ix) then
					l_item := item (l_ix)
				else
					l_ix := index_before (l_date)
					if valid_index (l_ix) then
						l_item := item (l_ix)
						l_date := dates.item (l_ix)
					end
				end
				l_data.extend (l_item)
				l_dates.extend (l_date)
				i := i + 1
			end
			create Result.make (l_data, l_dates)
		ensure
			monthly_series_not_void: Result /= Void
			same_count: Result.count = a_dates.count
			same_dates_count: Result.dates.count = Result.count
		end

feature -- Transformations

	transformed_series (a_transformation: FUNCTION [ANY, TUPLE [G, ROSE_DATE], G]): like Current is
			-- Version of `Current' where the data items are transformed by `a_transformation';
			-- N.B. `a_transformation' is not guarenteed to receive non-Void arguments.
		require
			a_transformation_not_void: a_transformation /= Void
		do
			Result := cloned
			from
				start
			until
				after
			loop
				Result.put (a_transformation.item ([current_item, current_date]), index)
				forth
			end
		ensure
			transformed_series_not_void: Result /= Void
			same_count: Result.count = count
			dates_set: Result.dates.same_items (dates)
			same_count_for_dates: Result.dates.count = dates.count
		end

feature -- Iteration

	do_all_series (a_proc: PROCEDURE [ANY, TUPLE [G, ROSE_DATE]]) is
			-- Apply `a_proc' to every datum in `Current'.
			-- N.B. `a_proc' is not guarenteed to receive non-Void arguments.
		require
			a_proc_not_void: a_proc /= Void
		do
			from
				start
			until
				after
			loop
				a_proc.call ([current_item, current_date])
				forth
			end
		ensure
			structure_unchanged: count = old count
			cursor_at_end: after
		end

	do_all_any (a_proc: PROCEDURE [ANY, TUPLE [ANY, ROSE_DATE]]) is
			-- Apply `a_proc' to every datum in `Current'.
			-- N.B. `a_proc' is not guarenteed to receive non-Void arguments.
		require
			a_proc_not_void: a_proc /= Void
		local
			l_any: ANY
		do
			if a_proc /= Void then
				from
					start
				until
					after
				loop
					l_any := current_item
					a_proc.call ([l_any, current_date])
					forth
				end
			end
		ensure
			structure_unchanged: count = old count
			cursor_at_end: after
		end

feature -- Names and dates

	dates: ROSE_LINEAR_ARRAY [ROSE_DATE]
			-- Date of datum with corresponding index

	set_dates (a_dates: ARRAY [ROSE_DATE]) is
			-- Assign `a_dates' to `dates'.
			-- Do nothing if `a_dates' = Void.
		require
			minimal_count: a_dates /= Void implies count <= a_dates.count
		do
			if a_dates /= Void then
				create dates.make_from_array (a_dates)
				populate_date_table
				invalidate_cache
			end
		ensure
			same_items: a_dates /= Void implies dates.same_items (a_dates)
		end

	name: STRING
			-- Optional name for `Current'

	set_name (a_name: STRING) is
			-- Set `name' to `a_name'
		do
			name := a_name
		ensure
			set: name = a_name
		end

	date_set: ROSE_DATE_SET is
			-- Dates in this time series as a date set
		require
			ascending_order: is_sorted_ascending
		do
			if date_set_saved = Void then
				create date_set_saved.make_from_array (dates, dates.last)
			end
			Result := date_set_saved
		ensure
			date_set_not_void: Result /= Void
		end

	date_strings: ROSE_LINEAR_ARRAY [STRING] is
			-- Ccyymmdd.out for all non-Void entries of `dates' in same order
		local
			i: INTEGER
			l_date: ROSE_DATE
		do
			create Result.make (1, dates.count)
			from
				i := dates.lower
			until
				i > dates.upper
			loop
				l_date := dates.item (i)
				if l_date /= Void then
					Result.put (l_date.ccyymmdd.out, i - dates.lower + 1)
				end
				i := i + 1
			end
		ensure
			date_strings_not_void: Result /= Void
			same_count: Result.count = dates.count
		end

	ccyymmdd_strings: ROSE_LINEAR_ARRAY [STRING] is
			-- Dates expressed as days in string format;
			-- A more expensive version of `date_strings'.
			-- As it calls a buggy routine (lacks a precondition), we need a preconditon.
			-- In return, we get a postcondition.
		require
			no_void_date: not dates.has (Void)
		local
			l_date_array: ROSE_DATE_ARRAY [ROSE_DATE]
		do
			create l_date_array.make_from_array (dates)
			Result := l_date_array.ccyymmdd_strings
		ensure
			date_strings_not_void: Result /= Void
			same_count: Result.count = dates.count
			ccyymmdd_strings_has_no_void: not Result.has (Void)
		end

	ccyymm_strings: ROSE_LINEAR_ARRAY [STRING] is
			-- Dates expressed as months in string format
		require
			no_void_date: not dates.has (Void)
		local
			l_date_array: ROSE_DATE_ARRAY [ROSE_DATE]
		do
			create l_date_array.make_from_array (dates)
			Result := l_date_array.ccyymm_strings
		ensure
			date_strings_not_void: Result /= Void
			same_count: Result.count = dates.count			
		end

	double_array: ROSE_LINEAR_ARRAY [DOUBLE] is
			-- Dates expressed in double format
			-- (e.g. graph or plot axis)
		require
			no_void_date: not dates.has (Void)		
		local
			l_date_array: ROSE_DATE_ARRAY [ROSE_DATE]
		do
			create l_date_array.make_from_array (dates)
			Result := l_date_array.double_array
		ensure
			date_strings_not_void: Result /= Void
			same_count: Result.count = dates.count			
		end

feature -- Access

	subset (a_indices: ARRAY [INTEGER]): like Current is
			-- Subset composed of items at `a_indices' from this array;
			--| Redefined so as to carry `dates' over into the subset.
		local
			i, j, k, l_lower: INTEGER
			l_dates: ARRAY [ROSE_DATE]
			l_array: ARRAY [G]
		do
			k := a_indices.count
			l_lower := a_indices.lower
			create l_dates.make (1, k)
			create l_array.make (1, k)
			from
				i := 1
			until
				i > k
			loop
				j := a_indices.item (l_lower + i - 1)
				if not (valid_index (j) and dates.valid_index (j)) then
					(create {EXCEPTIONS}).raise ("Index number " + i.out + " passed to ROSE_TIME_SERIES.subset is not valid")
				end
				l_dates.put (dates.item (j), i)
				l_array.put (item (j), i)
				i := i + 1
			end
			create Result.make (l_array, l_dates)
		ensure then
			subset_not_void: Result /= Void
			correct_count: Result.count = a_indices.count
			same_count_for_dates: Result.dates.count = Result.count
		end

	subset_loose (a_indices: ARRAY [INTEGER]): like Current is
			-- Subset composed of items at `a_indices' from this array;
			-- Compared to `subset', `subset_loose' has loose requirement. If `a_indices' [i] is invalid, put null in position i.
			--| Redefined so as to carry `dates' over into the subset.
		local
			i, j, k, l_lower: INTEGER
			l_dates: ARRAY [ROSE_DATE]
			l_array: ARRAY [G]
		do
			k := a_indices.count
			l_lower := a_indices.lower
			create l_dates.make (1, k)
			create l_array.make (1, k)
			from
				i := 1
			until
				i > k
			loop
				j := a_indices.item (l_lower + i - 1)
				if valid_index (j) then
					l_dates.put (dates.item (j), i)
					l_array.put (item (j), i)
				end
				i := i + 1
			end
			create Result.make (l_array, l_dates)
		ensure then
			subset_loose_not_void: Result /= Void
			correct_count: Result.count = a_indices.count
			same_count_for_dates: Result.dates.count = Result.count
		end

	selected_items (a_ixs: ARRAY [INTEGER]): like Current is
			-- Subset by indices;
			--| Redefined so as to carry `dates' over into the subset.
		do
			if a_ixs /= Void then
				Result := subset (a_ixs)
			end
		ensure then
			selected_items_not_void: a_ixs /= Void implies Result /= Void
			correct_count: a_ixs /= Void implies Result.count = a_ixs.count
			same_count_for_dates: a_ixs /= Void implies Result.dates.count = Result.count
		end

	subarray (i, j: INTEGER): like Current is
			-- All enties from positions i to j of `Current';
			--| Redefined so as to carry `dates' over into the subset.
		do
			Result := subset ((i |..| j).as_array)
		ensure then
			subarray_not_void: Result /= Void
			correct_count: Result.count = (j - i + 1)
			same_count_for_dates: Result.dates.count = Result.count
		end

	random_subarray (a_count: INTEGER): like Current is
			-- Always returns void since it is not properly redefined
		do
			Result := Void
		ensure then
			random_subarray_void: Result = Void
		end

	sort_series_by_dates: like Current is
			-- Copy of `Current' sorted by `dates' in ascending order
		require
			no_void_dates: not dates.has (Void)
		local
			l_dates: like dates
			i: INTEGER
		do
			create l_dates.make_empty
			l_dates.copy (dates)
			l_dates.sort_by_and_remember (agent date_less_than (?, ?))
			create Result.make_empty
			from
				i := l_dates.lower
			until
				i > l_dates.upper
			loop
				Result.extend (item_at (l_dates [i]), l_dates [i])
				i := i + 1
			end
		ensure
			sort_series_by_dates_not_void: Result /= Void
			same_date_count: Result.count = dates.count
			ascending_order: not Result.is_empty implies Result.is_sorted_ascending
		end

feature {NONE} -- Implementation

	date_table: HASH_TABLE [INTEGER, INTEGER]
			-- Indices of (data items of) `Current' indexed by ccyymmdd

	date_set_saved: ROSE_DATE_SET
			-- Cache for `date_set'

	invalidate_cache is
			-- Forget saved internal information.
		do
			date_set_saved := Void
		ensure
			date_set_will_be_rebuilt_on_demand: date_set_saved = Void
		end

	populate_date_table is
			-- Create `date_table' and fill with non-Void items from `dates'.
		require
			dates_not_void: dates /= Void
			minimal_count: count <= dates.count
		local
			i: INTEGER
			l_date: ROSE_DATE
		do
			create date_table.make (10)
			from
				i := 1
			until
				i > upper
			loop
				l_date := dates.item (i)
				if l_date /= Void then
					date_table.put (i, l_date.ccyymmdd)
				end
				i := i + 1
			end
		ensure
			date_table_not_void: date_table /= Void
		end

	strict_date_after (a_date: ROSE_DATE): ROSE_DATE is
			-- Strictly later date than `a_date' or Void if none.
		require
			a_date_not_void: a_date /= Void
			ascending_order: is_sorted_ascending
		do
			Result := date_after (a_date)
			if Result /= Void and then Result <= a_date then
				Result := Void
			end
		ensure
			after: Result /= Void implies Result > a_date
		end

	date_less_than (a_a, a_b: ROSE_DATE): BOOLEAN is
			-- Comparison function for sorting in an ascending order
		require
			a_a_not_void: a_a /= Void
			a_b_not_void: a_b /= Void
		do
			Result := a_a < a_b
		ensure
			definition: Result = (a_a < a_b)
		end

	is_later_than_previous (a_date: ROSE_DATE; a_index: INTEGER): BOOLEAN is
			-- Is `a_date' strictly later than `dates [a_index - 1]'?
		require
			a_index_valid: dates.valid_index (a_index)
		local
			l_previous: ROSE_DATE
		do
			if a_index = 1 then
				Result := True
			elseif a_date /= Void then
				l_previous := dates [a_index - 1]
				Result := l_previous /= Void and then a_date > l_previous
			end
		ensure
			initially_true: a_index = 1 implies Result
		end

invariant

	lower_is_1: lower = 1
	dates_not_void: dates /= Void
	dates_lower_is_1: dates.lower = 1

end
