note
	description: "Definition of an entity transformation"
	author: "Zoran Simic"
	last_editor: "$Author: zs $"
	date: "$Date: 2011/05/27 $"
	revision: "$Revision: #15 $"
	path: "$File: //rose/source/kernel/entities/transformation/rose_transformation.e $"

deferred class
	ROSE_TRANSFORMATION

inherit
	ROSE_SHARED_RECIPE_MANAGER
		redefine
			default_create, copy, is_equal
		end

	DEBUG_OUTPUT
		redefine
			default_create, copy, is_equal
		end

	CONTEXT_ANCHOR
		redefine
			copy, is_equal
		end

feature {NONE} -- Initialization

	default_create
			-- Default creation procedure
		do
			index := 1
			create settings.make (Void, Void, Void)
		end

	make_transformation (a_source, a_target: STRING; a_parent: INTEGER)
			-- Create with 'a_source', 'a_target', and 'a_parent'
		do
			set_source (a_source)
			set_target (a_target)
			parent := a_parent
			default_create
		end

feature -- Access

	settings: attached ROSE_TRANSFORMATION_SETTINGS	-- Displayed attributes, filter and constraints
	previous: ROSE_TRANSFORMATION					-- Preceding transformation in the recipe
	next: ROSE_TRANSFORMATION						-- Transformations preceding and following this one in a recipe
	index: INTEGER									-- Index of this transformation in sequence (first transformation had id of 1)
	parent: INTEGER									-- Index of parent transformation (available via 'previous')
	source: STRING									-- Source category for this transformation
	target: STRING									-- Target category for this transformation
	source_category: CATEGORY_HANDLER [like context_anchor, ENTITY]		-- Source category handler (available after 'attach')
	target_category: CATEGORY_HANDLER [like context_anchor, ENTITY]		-- Target category handler (available after 'attach')

	id: STRING
			-- Type id of this transformation
		do
			Result := generator.as_lower
			Result.replace_substring_all ("rose_transformation_", "")
		end

	depth: INTEGER
			-- Number of transformations following this one (this one included)
		do
			if next = Void then
				Result := 1
			else
				Result := 1 + next.depth
			end
		ensure
			next_depth_plus_one: (next = Void and then Result = 1) or else (next /= Void and then Result = 1 + next.depth)
		end

	first: ROSE_TRANSFORMATION
			-- First transformation in the sequence
		do
			if previous = Void then
				Result := Current
			else
				Result := previous.first
			end
		end

	parent_transformation: ROSE_TRANSFORMATION
			-- Transformation on which this one applies
		do
			Result := previous_transformation (parent)
		end

	transformed (a_entity: ENTITY): ROSE_LINEAR_ARRAY [ENTITY]
			-- Result of the transformation of 'a_entity'
		require
			valid: a_entity /= Void and then problem = Void
		do
			if equal (source, a_entity.handler.full_name) and then a_entity.is_valid then
				Result := actual_transformed (a_entity)
			end
			if Result = Void or else Result.is_empty then
				create Result.make (1, 1)
				Result.put (target_category.ur_entity, Result.lower)
			end
		ensure
			valid_result: is_valid_transformation_result (Result)
		end

	is_valid_transformation_result (a_result: ARRAY [ENTITY]): BOOLEAN
			-- Is 'a_result' valid for this transformation?
		local
			i: INTEGER
		do
			Result := a_result /= Void and then not a_result.is_empty
			if Result then
				from
					i := a_result.lower
				until
					not Result or else i > a_result.upper
				loop
					Result := a_result.item (i) /= Void and then a_result.item (i).handler = target_category
					i := i + 1
				end
			end
		end

	has (a_transformation: ROSE_TRANSFORMATION): BOOLEAN
			-- Does this transformation chain contain `a_transformation' ?
		do
			if a_transformation /= Void then
				Result := a_transformation = Current
					or else (next /= Void and then next.has (a_transformation))
			end
		end

	is_valid: BOOLEAN
			-- Is this transformation and all its successors valid?
		do
			if problem = Void then
				Result := next = Void or else next.is_valid
			end
		end

	problem: STRING
			-- Problem with current transformation (Void if valid)
		do
			if index <= 0 then
				Result := "Invalid index " + index.out
			elseif parent < 0 then
				Result := "Invalid parent " + parent.out
			elseif parent /= 0 and then parent_transformation = Void then
				Result := "No parent"
			elseif source = Void then
				Result := "No source"
			elseif target = Void then
				Result := "No target"
			elseif source_category = Void then
				Result := "Invalid source " + source
			elseif target_category = Void then
				Result := "Invalid target " + target
			else
				Result := implementor_problem
			end
		end

	help: STRING
			-- Help message explaining this transformation type
		deferred
		end

	representation: attached STRING
			-- Textual representation of this transformation
		do
			create Result.make (100)
			Result.extend ('%T')
			Result.append_string (id)
			Result.append_string (" (%"")
			Result.append_string (source)
			Result.append_string ("%", %"")
			Result.append_string (target)
			Result.append_string ("%", ")
			Result.append_string (parent.out)
			Result.append_string (specific_attributes)
			Result.append_string (")%N")
			Result.append_string (settings.representation)
		end

	is_equal (a_other: like Current): BOOLEAN
			-- Is `a_other' attached to an object considered
			-- equal to current object?
		do
			Result := a_other /= Void and then parent = a_other.parent and then index = a_other.index
		end

	debug_output: STRING
			-- String that should be displayed in debugger to represent `Current'.
		do
			Result := representation
		end

feature -- Basic operations

	copy (a_other: like Current)
			-- Update current object using fields of object attached
			-- to `a_other', so as to yield equal objects.
		do
			Precursor (a_other)
			create settings.make_copy (a_other.settings)
		end

	extend (a_next: ROSE_TRANSFORMATION)
			-- Extend this transformation chain with 'a_next'.
		require
			not_self: a_next /= Current
		do
			if next =  Void then
				next := a_next
				next.set_previous (Current)
			else
				next.extend (a_next)
			end
		end

	set_source (a_source: STRING)
			-- Set 'source'
		do
			source := a_source
			source_category := recipe_manager.category_handler (source)
		end

	set_target (a_target: STRING)
			-- Set 'target'
		do
			target := a_target
			target_category := recipe_manager.category_handler (target)
		end

	set_parent (a_parent: INTEGER)
			-- Set 'parent'
		do
			parent := a_parent.max (0)
		end

	set_settings (a_settings: ROSE_TRANSFORMATION_SETTINGS)
			-- Set `settings' to `a_settings'
		do
			settings.set_from_other (a_settings)
		end

	set_next (a_next: ROSE_TRANSFORMATION)
			-- Set the next transformation, stored in `next'
		require
			not_self: a_next /= Current
		do
			next := a_next
		end

feature {ROSE_RECIPE, ROSE_TRANSFORMATION} -- Implementation

	remove
			-- Remove this transformation from sequence
		local
			l_next: ROSE_TRANSFORMATION
		do
			if previous /= Void then
				previous.set_next (next)
			end
			if next /= Void then
				next.set_previous (previous)
				from
					l_next := next
				until
					l_next = Void
				loop
					l_next.set_parent (l_next.parent - 1)
					l_next := l_next.next
				end
			end
			previous := Void
			next := Void
			index := 1
			parent := 0
		end

	set_previous (a_previous: ROSE_TRANSFORMATION)
		require
			not_self: a_previous /= Current
		do
			previous := a_previous
			if previous /= Void then
				index := previous.index + 1
			else
				index := 1
			end
		end

	previous_transformation (a_parent: INTEGER): ROSE_TRANSFORMATION
		do
			if a_parent <= 0 then
				-- Return Void
			elseif a_parent = index then
				Result := Current
			elseif previous /= Void then
				Result := previous.previous_transformation (a_parent)
			end
		end

feature {NONE} -- Implementation

	implementor_problem: STRING
			-- Problem with current transformation (Void if valid)
		deferred
		end

	specific_attributes: STRING
			-- Specific transformation attributes for 'representation'
		deferred
		ensure
			valid_result: Result /= Void
		end

	actual_transformed (a_entity: ENTITY): ROSE_LINEAR_ARRAY [ENTITY]
			-- Actual transformation of 'a_entity' ('a_entity' has been checked to be valid)
		require
			valid_entity: a_entity /= Void and then a_entity.handle /= 0
		deferred
		end

feature {ROSE_TRANSFORMATION_LIST} -- Implementation

	specific_image (a_chains: ROSE_ENTITY_CHAINS; a_start, a_end: ROSE_DATE; a_range: HASH_TABLE [ENTITY, ENTITY]): ROSE_ENTITY_CHAINS
			-- This is only needed to bridge between old ROSE_ENTITY_CHAINS approach and ROSE_RECIPE
		deferred
		ensure
			exists: a_chains /= Void
		end

end
