note

	description:

		"Entity transformation chain, represents a transformation result set for one given entity"

	author: "Zoran Simic"
	last_editor: "$Author: zs $"
	date: "$Date: 2011/05/27 $"
	revision: "$Revision: #8 $"
	path: "$File: //rose/source/kernel/entities/transformation/rose_transformation_chain.e $"

class
	ROSE_TRANSFORMATION_CHAIN

inherit
	EXCEPTIONS
		export
			{NONE} all
		end

	ANY

create
	make_root, make_child

feature {NONE} -- Initialization

	make_root (a_entity: ENTITY; a_transformation: ROSE_TRANSFORMATION)
			-- create as root chain
		require
			valid_entity: a_entity /= Void
		do
			entity := a_entity
			transformation := a_transformation
			populate
		end

	make_child (a_entity: ENTITY; a_parent: ROSE_TRANSFORMATION_CHAIN)
			--create as child chain
		require
			valid_entity: a_entity /= Void
			valid_parent: a_parent /= Void and then a_parent.transformation /= Void
		do
			entity := a_entity
			parent := a_parent
			index := parent.index + 1
			transformation := a_parent.transformation.next
			populate
		end

feature -- Access

	cached_attributes: ROSE_LINEAR_ARRAY[PES_ATTRIBUTE_CONTAINER]
			--  contains all requested attributes (when sent from PES Server)

	entity: ENTITY
			-- Entity attached to this chain

	index: INTEGER
			-- Index of this chain

	parent: ROSE_TRANSFORMATION_CHAIN
			-- Direct parent chain

	transformation: ROSE_TRANSFORMATION
			-- Transformation leading to 'children', Void for leaves

	children: ROSE_LINEAR_ARRAY [ROSE_TRANSFORMATION_CHAIN]
			-- Children of this chain, Void for leaves

	source_chain (a_index: INTEGER): ROSE_TRANSFORMATION_CHAIN
			-- Source chain with 'a_index'
		require
			valid_index: a_index >= 0 and then a_index <= index
		do
			if a_index = index then
				Result := Current
			else
				Result := parent.source_chain (a_index)
			end
		end

	set_transformation (a_transformation: ROSE_TRANSFORMATION)
			-- set `transformation' to `a_transformation'
		do
			transformation := a_transformation
		end

feature  -- Operations

	set_cached_attributes (a_cache: like cached_attributes)
			-- Set `cached_attributes' to `a_cache'
		do
			cached_attributes := a_cache
		end

	cache_attribute_records (a_always_cache_name: BOOLEAN)
			-- set `cached_attributes'
			-- this is needed when a transformation result is sent from PES
			-- If `a_always_cache_name' is true, the descriptive name will be cached regardless of the
			-- recipe's display settings
		local
			l_container: PES_ATTRIBUTE_CONTAINER
			l_arr: ROSE_LINEAR_ARRAY [ENTITY_ATTRIBUTE_RECORD]
			l_identifier: ENTITY_IDENTIFIER
			l_identifier_names: ROSE_LINEAR_ARRAY [STRING]
			i, j: INTEGER
		do
			if parent /= Void then
				l_identifier_names := parent.transformation.settings.display.identifier_names.twin
			else
				create l_identifier_names.make_empty
				l_identifier_names.compare_objects
			end
			if a_always_cache_name then
				-- Add attributes required for the display name to the
				-- list of cached attributes
				from
					entity.name_identifier_keys.start
				until
					entity.name_identifier_keys.after
				loop
					if not l_identifier_names.has (entity.name_identifier_keys.item) then
						l_identifier_names.extend (entity.name_identifier_keys.item)
					end
					entity.name_identifier_keys.forth
				end
			end

			from
				create cached_attributes.make (1, l_identifier_names.count)
				j := l_identifier_names.lower
				i := 0
			until
				j > l_identifier_names.upper
			loop
				l_identifier := entity.context.identifier_from_name (l_identifier_names.item (j))
				if l_identifier /= Void then
					l_arr := l_identifier.key_history (entity)
					create l_container.make (l_arr, l_identifier.handle)
					i := i + 1
					cached_attributes.put (l_container, i)
				end
				j := j + 1
			end
			if i > 0 and then cached_attributes.upper > i then
				cached_attributes.restrict (1, i)
			elseif i = 0 then
				cached_attributes := Void
			end
		ensure
			no_void_values_cached: (cached_attributes /= Void) implies not cached_attributes.has (Void)
		end


feature {ROSE_TRANSFORMATION_RESULT, ROSE_TRANSFORMATION_CHAIN} -- Implementation

	do_all (a_array: ARRAY [ROSE_TRANSFORMATION_CHAIN]; a_agent: PROCEDURE [ANY,TUPLE [ARRAY [ROSE_TRANSFORMATION_CHAIN]]])
			-- Apply an agent to all children and do some other stuff
		require
			valid_array: a_array /= Void
			valid_agent: a_agent /= Void
		local
			i: INTEGER
		do
			a_array.put (Current, index)
			if children = Void then
				a_agent.call ([a_array])
			else
				from
					i := children.lower
				until
					i > children.upper
				loop
					children.item (i).do_all (a_array, a_agent)
					i := i + 1
				end
			end
		end

feature {NONE} -- Implementation

	populate
			-- Populate this chain's children
		local
			l_result: ARRAY [ENTITY]
			l_chain: ROSE_TRANSFORMATION_CHAIN
			i: INTEGER
		do
			if transformation /= Void then
				l_chain := source_chain (transformation.parent)
				if transformation.problem = Void then
					l_result := transformation.transformed (l_chain.entity)
				else
					raise ("Problem with transformation: " + transformation.problem)
				end

				from
					create children.make (1, l_result.count)
					i := l_result.lower
				until
					i > l_result.upper
				loop
					create l_chain.make_child (l_result.item (i), Current)
					children.put (l_chain, i)
					i := i + 1
				end
			end
		end

invariant
	entity_not_void: entity /= Void

end
