note

	description:

		"Abstraction of a value with units"

	author: "Mark Howard"
	last_editor: "$Author: mb $"
	date: "$Date: 2014/03/03 $"
	revision: "$Revision: #24 $"
	path: "$File: //rose/source/kernel/math/rose_unitized_value.e $"

class ROSE_UNITIZED_VALUE [G -> NUMERIC]

inherit

	ROSE_CONSTANTS

	ROSE_UNITIZED

	DEBUG_OUTPUT

create

	make_with_units

feature {ROSE_UNITIZED_VALUE} --initialization

	make_with_units (a_value: like value; a_units: like units) is
			--Note that a_units can be void
		do
			value := a_value
			if a_units = Void or else a_units.is_pure_units then
				units := pure_units
			else
				units := all_units.item (a_units)
				if units = Void then
					if not attached {CURRENCY_UNITS} a_units and then a_units.uses_currency and then a_units.denominator = 1 then
						--Cache CURRENCY_UNITS version if necessary
						units := currency_primes_table[a_units.numerator]
					end
					if units = Void then
						units := a_units
					end
					all_units.put (units)
				end
			end
		end

feature --access

	value: G

	units: ROSE_UNITS

feature --debug

	debug_output: STRING is
			-- debug message
		local
			l_value, l_units_name: STRING
		do
			l_value := value.out
			if l_value = Void then
				l_value := ""
			end
			l_units_name := units.name
			if l_units_name = Void then
				l_units_name := ""
			end
			Result := l_value
			if l_value.count /= 0 and l_units_name.count /= 0 then
				Result.append (" ")
			end
			Result.append (l_units_name)
		end

feature --aritmetic

	infix "+" (a_other: ROSE_UNITIZED_VALUE [G]): ROSE_UNITIZED_VALUE [G] is
			--Adding two unitized values will preserve the units if both sides have the same units
			--otherwise you end up with unknown units.  This case is not an error, it just causes us to lose the units, so
			--we do not handle it with a precondition.
		local
			l_mismatched_units: MISMATCHED_UNITS
		do
			if units = a_other.units then
				create Result.make_with_units (value + a_other.value, units)
			else
				create l_mismatched_units.make (units, a_other.units)
				create Result.make_with_units (value + a_other.value, l_mismatched_units)
			end
		end

	infix "-" (a_other: ROSE_UNITIZED_VALUE [G]): ROSE_UNITIZED_VALUE [G] is
			--Subtracting two unitized values will preserve the units if both sides have the same units
			--otherwise you end up with unknown units.  This case is not an error, it just causes us to lose the units, so
			--we do not handle it with a precondition.
		local
			l_mismatched_units: MISMATCHED_UNITS
		do
			if units = a_other.units then
				create Result.make_with_units (value - a_other.value, units)
			else
				create l_mismatched_units.make (units, a_other.units)
				create Result.make_with_units (value - a_other.value, l_mismatched_units)
			end
		end

	infix "*" (a_other: ROSE_UNITIZED_VALUE [G]): ROSE_UNITIZED_VALUE [G] is
			--multiply values and units
		do
			create Result.make_with_units (value * a_other.value, units * a_other.units)
		end

	infix "/" (a_other: ROSE_UNITIZED_VALUE [G]): ROSE_UNITIZED_VALUE [G] is
			--ratio of values and units
		do
			create Result.make_with_units (value / a_other.value, units / a_other.units)
		end

	infix "#+" (a_other: G): like Current
			--Add a_other to value preserving units
		do
			create Result.make_with_units (value + a_other, units)
		end

	infix "#-" (a_other: G): like Current
			--Subtract a_other to value preserving units
		do
			create Result.make_with_units (value - a_other, units)
		end

	infix "#*" (a_other: G): like Current
			--Multiply a_other with value preserving units
		do
			create Result.make_with_units (value * a_other, units)
		end

	infix "#/" (a_other: G): like Current
			--Divide value by a_other preserving units
		do
			create Result.make_with_units (value / a_other, units)
		end

	scaled (a_value: G): like Current is
			--scaled value of current
		do
			Result := twin
			Result.make_with_units (a_value * value, units)
		end

	has_matched_units: BOOLEAN is
			-- do the units match?
		do
			Result := units.has_matched_units
		end

	has_mismatched_units: BOOLEAN is
			--do we have mismatched units?
		do
			Result := units.has_mismatched_units
		end

feature --transformation

	converted (a_units_from, a_units_to: ROSE_UNITS; a_conversion_factor: G): like Current is
			--replace every instance of a_units_from with a_units_to, applying a_conversion_factor for each substitution
		require
			is_simple: a_units_from.is_simple and a_units_to.is_simple
		local
			l_units_from, l_units_to: INTEGER
			l_numerator, l_denominator: ROSE_LINEAR_ARRAY [INTEGER]
			l_value: G
			i: INTEGER
			l_new_units: ROSE_UNITS
			l_factored_rational: FACTORED_RATIONAL
		do
			if is_defined_g (a_conversion_factor) then
				l_units_from := a_units_from.rational.numerator
				l_units_to := a_units_to.rational.numerator
				if l_units_to /= l_units_from or else a_units_to.scale /= a_units_from.scale then
					l_value := value
					l_numerator := units.numerator_factors.deep_twin
					l_denominator := units.denominator_factors.deep_twin
					from
						i := 1
					until
						i > l_numerator.count
					loop
						if l_numerator [i] = l_units_from then
							l_value := l_value * a_conversion_factor
							l_numerator [i] := l_units_to
						end
						i := i + 1
					end
					from
						i := 1
					until
						i > l_denominator.count
					loop
						if l_denominator [i] = l_units_from then
							l_value := l_value / a_conversion_factor
							l_denominator [i] := l_units_to
						end
						i := i + 1
					end
					create l_factored_rational.make_from_factors (l_numerator, l_denominator)
					create l_new_units.make_from_rational (l_factored_rational, units.scale,units.uses_currency)
					create Result.make_with_units (l_value, l_new_units)
				else
					Result := Current
				end
			else
				create Result.make_with_units (missing_value, Void)
			end
		end

	converted_unitized_value (a_market: MARKET; a_currency: PERSISTENT_CURRENCY; a_date: ROSE_DATE): like Current is
			-- `a_value' converted using `a_currency' and `a_market' on `a_date'
		require
			valid_currency: a_currency /= Void
		local
			l_factor: G
			l_numerator, l_denominator: ROSE_LINEAR_ARRAY [INTEGER]
			l_from_currency, l_to_currency: CURRENCY_UNITS
			i: INTEGER
		do
			Result := Current
			if units /= Void and then not units.is_pure_units then
				l_to_currency := currency_units (a_currency)
				if units.rational /= Void and then units.rational.numerator > 0 and then units.rational.denominator > 0 then
					l_numerator := Result.units.numerator_factors
					l_denominator := Result.units.denominator_factors
					from
						i := 1
					until
						i > l_numerator.count
					loop
						if l_numerator [i] /= pure_value and then l_numerator [i] /= l_to_currency.rational.numerator then
							l_from_currency := currency_primes_table [l_numerator [i]]
							if l_from_currency /= Void then
								l_factor := from_double (a_currency.market_exchange_rate (a_market, l_from_currency.currency, a_date))
								Result := Result.converted (l_from_currency, l_to_currency, l_factor)
							end
						end
						i := i + 1
					end
					from
						i := 1
					until
						i > l_denominator.count
					loop
						if l_denominator [i] /= pure_value and then l_denominator [i] /= share_value then
							l_from_currency := currency_primes_table [l_denominator [i]]
							if l_from_currency /= Void then
								l_factor := from_double (a_currency.market_exchange_rate (a_market, l_from_currency.currency, a_date))
								Result := Result.converted (l_from_currency, l_to_currency, l_factor)
--							else
--								create Result.make_with_units (missing_value, pure_units)
							end
						end
						i := i + 1
					end
				else
					create Result.make_with_units (missing_value, pure_units)
				end
			end
		end

	set_value(a_value : like value)
			--Set value
		do
			value := a_value
		end

feature -- Access
	value_is_defined_double: BOOLEAN is
			--Is value defined?
		do
			if attached {DOUBLE} value as l_double then
				Result := is_defined_double (l_double)
			end
		end

	missing_value: ? G is
			-- Missing value
		do
			if {l_value: G} constants.missing_double then
				Result := l_value
			elseif {l_decimal: G} missing_decimal then
				Result := l_decimal
			end
		end

	is_defined_g (a_g: G): BOOLEAN is
			--is a_g defined
		do
			if attached {DOUBLE} a_g as al_double then
				Result := is_defined_double (al_double)
			elseif attached {REAL} a_g as al_real then
				Result := is_defined_real (al_real)
			elseif attached {RP_DECIMAL} a_g as al_decimal then
				Result := al_decimal.is_defined
			elseif attached {INTEGER} a_g as al_integer then
				Result := al_integer /= constants.missing_integer
			end
		end

	is_available : BOOLEAN
			--is the value defined?
		do
			Result := is_defined_g(value)
		end

feature {NONE} --conversion

	from_double (a_value: DOUBLE): G is
			--a_value converted to G
		local
			l_decimal: RP_DECIMAL
		do
			Result := missing_value
			if {l_double: G} a_value then
				Result := l_double
			else
				l_decimal := a_value
				if {l_decimal_g: G} l_decimal then
					Result := l_decimal_g
				end
			end
		end

	missing_decimal: RP_DECIMAL is
			--missing decimal version
		once
			Result := constants.missing_double
		end

end
