indexing
	description: "Definition of a unit"
	author: "Mark Howard"
	last_editor: "$Author: mb $"
	date: "$Date: 2013/12/10 $"
	revision: "$Revision: #26 $"
	path: "$File: //rose/source/kernel/math/rose_units.e $"
	storable_version: ""

class ROSE_UNITS

inherit

	RP_SERIALIZABLE

	ROSE_UNITS_ENVIRONMENT
		undefine
			is_equal
		end

	FACTORED_RATIONALS
		undefine
			is_equal
		end

	MULTI_KEYED
		redefine
			hash_code,
			keys_equal
		end

	ROSE_IEEE_UNDEFINED
		export
			{NONE} all
		undefine
			is_equal
		end

create
	make_from_rational

feature {NONE} -- Initialization

	make_from_rational (a_rational : FACTORED_RATIONAL; a_scale : RATIONAL;a_uses_currency : BOOLEAN) is
		--All of our units are mapped to rational numbers.  This is accomplished by assigning a unique prime number to
		--each atomic unit.  Currencies are handled specially to facilitate conversion from one currency to another.
		require
			valid_rational: a_rational /= Void
			valid_scale: a_scale /= Void
		local
			l_numerator_factors, l_denominator_factors: ROSE_LINEAR_ARRAY [INTEGER]
		do
			scale := a_scale
			rational := a_rational
			l_numerator_factors := rational.numerator_factors
			if scale.numerator = 1 then
				name := ""
			elseif scale.numerator = 1000000 then
				name := "million "
			else
				name := scale.numerator.out
			end
			from l_numerator_factors.start until l_numerator_factors.off loop
				if l_numerator_factors.index /= 1 then
					name := name + "*"
				end
				name := name + units_index_table.item(l_numerator_factors.current_item).name
				l_numerator_factors.forth
			end
			if a_rational.denominator /= 1 then
				name := name + "/"
				l_denominator_factors := rational.denominator_factors
				from l_denominator_factors.start until l_denominator_factors.off loop
					if l_denominator_factors.index /= 1 then
						name := name + "*"
					end
					name := name + units_index_table.item(l_denominator_factors.current_item).name
					l_denominator_factors.forth
				end
			end
			set_uses_currency (a_uses_currency or else numerator \\ anonymous_currency.numerator = 0
				or else denominator \\ anonymous_currency.numerator = 0)
		end

feature {OBJECT_TABLE,ROSE_UNITS} -- Access

	key_count : INTEGER is 4
		--used by OBJECT_TABLE
	hash_key(i : INTEGER) : HASHABLE is
		--used by OBJECT_TABLE
		require else
			valid_index : i <= key_count and i > 0
		do
			inspect i
			when 1 then
				Result := rational.numerator
			when 2 then
				Result := rational.denominator
			when 3 then
				if scale /= void then
					Result := scale.numerator
				end
			when 4 then
				if scale /= void then
					Result := scale.denominator
				end
			end
		end

feature -- Access

	is_pure_units: BOOLEAN is
			-- Does `Current' represent pure units?
		do
			Result := False
		end
			
	hash_code: INTEGER
			-- A faster (but equivalent) version of the feature in MULTI_KEYED.
		do
			if rational /= Void then
				if scale=Void
				 then
					Result :=
						(
						 rational.numerator * 2 +		-- * primes.item(1)
						 rational.denominator * 3		-- * primes.item(2)
						 ).hash_code						-- Trims off the high bit (required by the ensure clause).
				else
					Result :=
						(
						 rational.numerator * 2 +		-- * primes.item(1)
						 rational.denominator * 3 +		-- * primes.item(2)
						 scale.numerator * 5 +			-- * primes.item(3)
						 scale.denominator * 7			-- * primes.item(4)
						 ).hash_code						-- Trims off the high bit (required by the ensure clause).
				end
			end
		end
	
	keys_equal(a_other: like Current): BOOLEAN
		-- A faster (but equivalent) version of the feature in MULTI_KEYED.
		do
			if Current=a_other
			 then
				Result := True
			elseif scale/=Void and then a_other.scale/=Void
			 then
				Result :=
					rational.numerator=a_other.rational.numerator and then
					rational.denominator=a_other.rational.denominator and then
					scale.numerator=a_other.scale.numerator and then
					scale.denominator=a_other.scale.denominator
			elseif scale=Void and then a_other.scale=Void
			 then
				Result :=
					rational.numerator=a_other.rational.numerator and then
					rational.denominator=a_other.rational.denominator
					--else
					-- final case is where 'Current' has scale but 'a_other' doesn't (or vice versa).
					-- In that case "Result := False" since the scales are different.
			end
		end

feature -- Access
	is_simple : BOOLEAN is
			--Do these units have no denominator?
		do
			Result := rational.denominator = 1
		end
	name : STRING
		--human readable form
	currency_conversion_factor (a_exchange_rate : DOUBLE) : DOUBLE is
			--product of a_exchange_rate and its inverse according to how anonymous_currency appears
		require
			sensible_exchange_rate : a_exchange_rate > 0
		local
			l_factors : ROSE_LINEAR_ARRAY [INTEGER]
		do
			if is_defined_double(a_exchange_rate) then
				Result := 1
				if uses_currency then
					l_factors := numerator_factors
					from l_factors.start until l_factors.off loop
						if l_factors.current_item = anonymous_currency.numerator then
							Result := Result * a_exchange_rate
						end
						l_factors.forth
					end
					l_factors := denominator_factors
					from l_factors.start until l_factors.off loop
						if l_factors.current_item = anonymous_currency.numerator then
							Result := Result / a_exchange_rate
						end
						l_factors.forth
					end
				end
			else
				Result := undefined_double
			end
		end

	scale : RATIONAL

	has_matched_units: BOOLEAN is
			-- This distinguishes this class from MISMATCHED_UNITS.
		do
			Result := True
		end

	has_mismatched_units: BOOLEAN is
			-- This distinguishes this class from MISMATCHED_UNITS.
		do
			Result := False
		end

	numerator_units : ROSE_LINEAR_ARRAY[ROSE_UNITS]
			--Units of the numerator
		local
			l_factors : ROSE_LINEAR_ARRAY [INTEGER]
		do
			create Result.make_empty
			l_factors := numerator_factors
			from l_factors.start until l_factors.off loop
				Result.extend(units_index_table[l_factors.current_item])
				l_factors.forth
			end
		end

	uses_currency : BOOLEAN		-- If True then this object has at least one currency (perhaps anonymous) in the numerator or denominator.

feature --arithmetic
	infix "+" (a_other : ROSE_UNITS) : ROSE_UNITS is
		--units can only be added if they are the same, and then the sum is the original units
		--void units are considered the same as pure units
		require
			non_void : a_other /= void
		local
			l_other : ROSE_UNITS
		do
			if a_other = Void then
				l_other := pure_units
			else
				l_other := a_other
			end
			if rational.is_equal(l_other.rational) then
				Result := Current
			else
				create {MISMATCHED_UNITS} Result.make(Current,l_other)
			end
			Result.set_uses_currency(uses_currency or l_other.uses_currency)
		end

	infix "-" (a_other : ROSE_UNITS) : ROSE_UNITS is
		--units can only be subtracted if they are the same, and then the sum is the original units
		--void units are considered the same as pure units
		require
			non_void : a_other /= void
		local
			l_other : ROSE_UNITS
		do
			if a_other = Void then
				l_other := pure_units
			else
				l_other := a_other
			end
			if rational.is_equal(l_other.rational) then
				Result := Current
			else
				create {MISMATCHED_UNITS} Result.make(Current,l_other)
			end
			Result.set_uses_currency(uses_currency or l_other.uses_currency)
		end

	infix "*" (a_other : ROSE_UNITS) : ROSE_UNITS is
		--abstract multiplication of units, reduced as fractions
		require
			non_void : a_other /= void
		do
			if a_other = Void or (a_other.is_equal(pure_units)) then
				Result := Current
			else
				Result := products.item(a_other)
				if Result = Void then
					create Result.make_from_rational(rational * a_other.rational,scale * a_other.scale,uses_currency or else a_other.uses_currency)
					products.put(Result,a_other)
					a_other.products.put(Result,Current)
					Result.ratios.put(a_other,Current)
					Result.ratios.put(Current,a_other)
				end
			end
			Result.set_scale(scale * a_other.scale)
		end

	infix "/" (a_other : ROSE_UNITS) : ROSE_UNITS is
		--abstract division of units, reduced as fractions
		require
			non_void : a_other /= void
		local
			l_rational : FACTORED_RATIONAL
			l_scale : RATIONAL
		do
			if a_other = Void or (a_other.is_equal(pure_units)) then
				Result := Current
			else
			    if Current ~ a_other then
			       Result := pure_units
			    else
					Result := ratios.item(a_other)
					if Result = Void then
						if a_other.rational.numerator /= 0 and a_other.scale.numerator /= 0 then
							create Result.make_from_rational(rational / a_other.rational,scale / a_other.scale,uses_currency or else a_other.uses_currency)
						else
							create l_rational
							create l_scale.make (0, 0)
							create Result.make_from_rational(l_rational,l_scale,False)
						end
						a_other.ratios.put(Result,Current)
						ratios.put(Result,a_other)
						a_other.products.put(Current,Result)
						Result.products.put(Current,a_other)
					end
				end
			end
		end
feature --conversion
	in_currency (a_currency : PERSISTENT_CURRENCY) : ROSE_UNITS is
		--all abstract currency units are replaced by the actual currency 'a_currency'
		require
			non_void : a_currency /= void
		local
			l_old_numerator_factors, l_old_denominator_factors: ROSE_LINEAR_ARRAY [INTEGER]
			l_new_numerator_factors, l_new_denominator_factors: ROSE_LINEAR_ARRAY [INTEGER]
			l_units : CURRENCY_UNITS
			i : INTEGER
			l_currency_index : INTEGER
			l_rational : FACTORED_RATIONAL
		do
			if uses_currency then
				if has_matched_units and then a_currency /= Void and then a_currency.iso /= Void then
					Result := currency_units_table.item(a_currency.iso)
					if Result = Void then
						l_currency_index := currency_units(a_currency).numerator.to_integer
						create l_new_numerator_factors.make_empty
						create l_new_denominator_factors.make_empty
						l_old_numerator_factors := numerator_factors	-- This is a HASH_TABLE lookup.
						from i := 1 until i > l_old_numerator_factors.count loop
							l_units ?= units_index_table.item(l_old_numerator_factors.item(i))
							if l_units /= Void then
								l_new_numerator_factors.extend(l_currency_index)
							else
								l_new_numerator_factors.extend(l_old_numerator_factors.item(i))
							end
							i := i + 1
						end
						l_old_denominator_factors := denominator_factors	-- This is a HASH_TABLE lookup.
						from i := 1 until i > l_old_denominator_factors.count loop
							l_units ?= units_index_table.item(l_old_denominator_factors.item(i))
							if l_units /= Void then
								l_new_denominator_factors.extend(l_currency_index)
							else
								l_new_denominator_factors.extend(l_old_denominator_factors.item(i))
							end
							i := i + 1
						end
						create l_rational.make_from_factors(l_new_numerator_factors,l_new_denominator_factors)
						create Result.make_from_rational(l_rational,scale,True)
						currency_units_table.put(Result,a_currency.iso)
					end
				else
					create {MISMATCHED_UNITS} Result.make(Current,pure_units)
				end
			else
				Result := Current
			end
		end

	currency_if_any : CURRENCY
			--Underlying currency if the numerator is simply a currency, as in e.g. currency / share
		do
			if currency_primes_table[numerator] /= Void then
				Result := currency_primes_table[numerator].currency
			end
		end


	numerator : INTEGER_64 is
		--all units are effectively rational numbers, and this is the numerator of Current in reduced form
		do
			Result := rational.numerator
		end

	denominator : INTEGER_64 is
		--all units are effectively rational numbers, and this is the denominator of Current in reduced form
		do
			Result := rational.denominator
		end

feature {ROSE_EXPRESSION,ROSE_UNITS} --transformation

	set_uses_currency (a_boolean : BOOLEAN) is
		--it is useful to know whether the units actually have any currency to replace
		do
			uses_currency := a_boolean
			if uses_currency and currency_units_table = Void then
				create currency_units_table.make(10)
			end
		end

	set_scale (a_scale : RATIONAL) is
		--we support scaling factors, such as 1 million USD
		require
			valid_scale: a_scale /= Void
		do
			scale := a_scale
		end

feature {ROSE_UNITS} --Cacheing

	saved_ratios : HASH_TABLE [ROSE_UNITS,ROSE_UNITS]
	saved_products : HASH_TABLE [ROSE_UNITS,ROSE_UNITS]

feature {ROSE_UNITS_ENVIRONMENT,RP_UNITS} --implementation

	scaled_units_table : HASH_TABLE [ROSE_UNITS,RATIONAL]

	rational : FACTORED_RATIONAL
			-- Underlying rational number

	currency_units_table : HASH_TABLE [ROSE_UNITS,STRING]
		-- Table of currency units indexed by the currency ISO code.

	ratios : HASH_TABLE [ROSE_UNITS,ROSE_UNITS] is
		--keeps track of all ratios with current as numerator for efficiency
		do
			if saved_ratios = Void then
				create saved_ratios.make(100)
			end
			Result := saved_ratios
		end

	products : HASH_TABLE [ROSE_UNITS,ROSE_UNITS] is
		--keeps track of all units of the form Current * other for efficiency
		do
			if saved_products = Void then
				create saved_products.make(100)
			end
			Result := saved_products
		end
	scaled_units (a_scale : RATIONAL) : ROSE_UNITS is
			-- Current units scaled to 'a_scale'...e.g. 1 million USD
		require
			valid_scale: a_scale /= Void
		do
			if a_scale.is_equal (one) then
				Result := Current
			else
				if scaled_units_table = Void then
					create scaled_units_table.make(5)
				end
				Result := scaled_units_table.item(a_scale)
				if Result = Void then
					create Result.make_from_rational(rational,a_scale,uses_currency)
					scaled_units_table.put(Result,a_scale)
				end
			end
		end
	empty_factors : ROSE_LINEAR_ARRAY [INTEGER] is
		--non void default factors
		once
			create Result.make_empty
		end
	numerator_factors : ROSE_LINEAR_ARRAY [INTEGER] is
		--list of factors of numerator
		do
			Result := rational.numerator_factors
		end

	denominator_factors : ROSE_LINEAR_ARRAY [INTEGER] is
		--list of factors of denominator
		do
			Result := rational.denominator_factors
		end
	anonymous_currency : ROSE_UNITS is
		--abstract currency ... placeholder for units where the currency has not been specified
		once
			Result := units_table.item ("currency")
		end
invariant
	rational_exists : rational /= Void
end
