note

	description:

		"General features needed for managing unitized arithmetic"

	author: "Mark Howard"
	last_editor: "$Author: mk $"
	date: "$Date: 2013/06/17 $"
	revision: "$Revision: #33 $"
	path: "$File: //rose/source/kernel/math/rose_units_environment.e $"

class ROSE_UNITS_ENVIRONMENT

inherit

	NUMBER_THEORY

feature -- Access

	pure_units: PURE_UNITS is
			-- same as no units
		once
			create Result.default_create
		end

	share_units (a_instrument: ENTITY): ROSE_INSTRUMENT_UNITS is
			-- every instrument gets its own units
		do
			units_index_table.do_nothing
			Result := share_units_table.item ([a_instrument.category_id, a_instrument.handle])
			if Result = Void then
				create Result.make (a_instrument)
				share_units_table.put (Result, [a_instrument.category_id, a_instrument.handle])
				units_index_table.put (Result, Result.numerator)
				all_units.put (Result)
			end
		end

	date_units (a_date: ROSE_DATE): ROSE_DATE_UNITS is
			-- every date gets its own units
		do
			units_index_table.do_nothing
			date_units_table.search (a_date)
			if date_units_table.found then
				Result := date_units_table.found_item
			else
				create Result.make (a_date)
				date_units_table [a_date] := Result
				units_index_table.put (Result, Result.numerator)
				all_units.put (Result)
			end
		end

	million_units: ROSE_UNITS is
			-- Million units
		once
			Result := units_table.item ("M")
		end

	currency_units (a_currency: PERSISTENT_CURRENCY): CURRENCY_UNITS is
			-- special case of instrument units, perhaps should be combined
		local
			l_record: UNITS_RECORD
		do
			if a_currency /= Void then
				Result ?= units_table.item (a_currency.iso)
				if Result = Void then
					increment_maximum_prime
					create l_record.make (a_currency.iso, a_currency.name_at (create {ROSE_DATE}.make_today), "currency", maximum_prime_number_used)
					create Result.make_from_record (l_record)
					units_index_table.put (Result, l_record.prime)
					units_table.put (Result, l_record.mnemonic)
					all_units.put (Result)
				end
			else
				Result := generic_currency_units
			end
		end

	generic_currency_units: CURRENCY_UNITS is
			-- used as a placeholder or free variable for currency in a units expression
		once
			create Result.make_generic
		end

	identified_units (a_identifier: STRING): ROSE_IDENTIFIED_UNITS is
			-- Units for unique `a_identifier'
		require
			valid_identifier: a_identifier /= Void and then not a_identifier.is_empty
		do
			units_index_table.do_nothing
			identified_units_table.search (a_identifier)
			if identified_units_table.found then
				Result := identified_units_table.found_item
			else
				create Result.make (a_identifier)
				identified_units_table.force (Result, a_identifier)
				units_index_table.put (Result, Result.numerator)
				all_units.put (Result)
			end
		ensure
			non_void_result: Result /= Void
		end

feature {ROSE_DATE_UNITS} -- Date units

	date_prime (a_date: ROSE_DATE): INTEGER is
			-- Prime for the units of a date
		require
			a_date_not_void: a_date /= Void
		local
			l_date: like a_date
		do
			l_date := a_date.date
			date_to_prime_table.search (l_date)
			if date_to_prime_table.found then
				Result := date_to_prime_table.found_item
			else
				create_date_prime (l_date)
				retrieve_date_table (date_to_prime_table)
				date_to_prime_table.search (l_date)
				if not date_to_prime_table.found then
					rsg.error.raise ("Could not create date units for [" + l_date.ccyymmdd.out + "]")
				else
					Result := date_to_prime_table.found_item
				end
			end
		end

feature {ROSE_IDENTIFIED_UNITS} -- Identified units

	identified_units_prime (a_identifier: STRING): INTEGER is
			-- Prime for the identified units
		require
			valid_identifier: a_identifier /= Void and then not a_identifier.is_empty
		do
			identified_units_prime_numbers_table.search (a_identifier)
			if identified_units_prime_numbers_table.found then
				Result := identified_units_prime_numbers_table.found_item
			else
				create_identified_units_prime (a_identifier)
				retrieve_identified_units_prime_numbers_table (identified_units_prime_numbers_table)
				identified_units_prime_numbers_table.search (a_identifier)
				if not identified_units_prime_numbers_table.found then
					rsg.error.raise ("Could not create identified units prime for [" + a_identifier + "]")
				else
					Result := identified_units_prime_numbers_table.found_item
				end
			end
		end

feature {NONE} -- cacheing

	date_units_table: HASH_TABLE [ROSE_DATE_UNITS, ROSE_DATE] is
			-- Table of date units
		once
			create Result.make (1000)
		end

	share_units_table: ROSE_MULTI_KEYED_TABLE [ROSE_INSTRUMENT_UNITS] is
			-- Table of units
		once
			create Result.make (10000)
		end

	units_mnemonic_table: ROSE_HASH_TABLE [UNITS_RECORD, STRING] is
			-- Table of mnemonic
		once
			create Result.make (500)
			units_index_table.do_nothing
		end

	all_units: OBJECT_TABLE [ROSE_UNITS] is
			-- Table of all units
		once
			create Result.make (1000)
		end

	units_index_table: HASH_TABLE [ROSE_BASE_UNITS, INTEGER_64] is
			-- Unit index table
		local
			l_units_reader: COLLECTING_PREDICATE [UNITS_RECORD]
			l_array: ROSE_LINEAR_ARRAY [UNITS_RECORD]
			l_servers: DATABASE_SERVERS
			l_units: ROSE_BASE_UNITS
			l_record: UNITS_RECORD
		once
			create l_servers
			create l_units_reader.make (l_servers.rose_entity_foreground_server, "select * from units where mnemonic is not null")
			if not l_units_reader.success then
				rsg.error.raise ("Could not create l_units_reader")
			end
			l_units_reader.retrieve (Void, Void)
			if not l_units_reader.success then
				rsg.error.raise ("Could not execute l_units_reader")
			end
			l_array := l_units_reader.array
			create Result.make (300)
			from
				l_array.start
			until
				l_array.off
			loop
				l_record := l_array.current_item
				l_record.name.prune_all_trailing (' ')
				l_record.mnemonic.prune_all_trailing (' ')
				if l_record.mnemonic.count = 0 then
						-- Yes, there is a currency whose mnemonic is equal to ' '.  The unknown currency.
					l_record.set_mnemonic (" ")
				end
				l_record.type.prune_all_trailing (' ')
				units_mnemonic_table.put (l_record, l_record.mnemonic)
				if l_record.type.is_equal ("currency") then
					create {CURRENCY_UNITS} l_units.make_from_record (l_record)
				else
					create l_units.make_from_record (l_record)
				end
				Result.put (l_units, l_record.prime)
				if l_record.prime > maximum_prime_number_used then
					maximum_prime_index_cell.put (primes.index_of (l_record.prime, 1))
				end
				units_table.put (l_units, l_record.mnemonic)
				l_array.forth
			end
			units_table.put (pure_units, "")
		end

	units_table: ROSE_HASH_TABLE [ROSE_UNITS, STRING] is
			-- Units table
		once
			create Result.make (500)
			units_index_table.do_nothing
				-- Forces this table to be initialized.
		end

	currency_primes_table: HASH_TABLE [CURRENCY_UNITS, INTEGER_64] is
			-- Currency units table
		once
			create Result.make (100)
		end

feature {NONE} --implementation

	one: RATIONAL is
			-- One as a rational
		once
			create Result.make (1, 1)
		end

	factored_one: FACTORED_RATIONAL is
			-- One as a factored rational
		once
			create Result.make_from_factors (<<1>>, <<1>>)
		end

	prime_number (a_index: INTEGER): INTEGER is
			-- i_th prime where i = a_index
		do
			from
			until
				a_index <= primes.count
			loop
				primes.double_upper_bound
			end
			if a_index > 0 then
				Result := primes.item (a_index)
			end
		end

	update_maximum_prime (a_prime: INTEGER) is
			-- make sure that we have all primes at least up to a_prime
		require
			sensible_value: a_prime > 0
			is_prime: primes.index_of_prime (a_prime) > 0
		do
			if a_prime > maximum_prime_number_used then
				maximum_prime_index_cell.put (primes.index_of_prime (a_prime))
			end
		ensure
			index_increased: maximum_prime_number_index >= primes.index_of_prime (a_prime)
		end

	increment_maximum_prime is
			-- Increment the maximum value of a prime
		do
			maximum_prime_index_cell.put (maximum_prime_number_index + 1)
		end

	maximum_prime_number_index: INTEGER is
			-- Maximum index of all used primes
		do
			Result := maximum_prime_index_cell.item
		end

	maximum_prime_number_used: INTEGER is
			-- Maximum value of all used primes
		do
			Result := prime_number (maximum_prime_number_index)
		end

	is_currency_prime (a_prime: INTEGER): CURRENCY_UNITS is
			-- Is the prime used for a currency unit
		do
			Result := currency_primes_table [a_prime]
		end

	maximum_prime_index_cell: CELL [INTEGER] is
			-- Cell used to contain the maximum prime index
		once
			create Result.put (0)
		end

	date_to_prime_table: HASH_TABLE [INTEGER_32, ROSE_DATE] is
			-- Table date -> prime
		once
			create Result.make (0)
			retrieve_date_table (Result)
		end

	retrieve_date_table (a_table: like date_to_prime_table) is
			-- Retrieve the date/prime table
		local
			l_selector: SELECTION_PREDICATE [PAIR [ROSE_DATE, INTEGER_32]]
			l_container: ARRAY [PAIR [ROSE_DATE, INTEGER_32]]
			i: INTEGER_32
			d: DATA_SERVER
		do
			d := units_rose_entity_background_server
			create l_selector.make (d, create {PAIR [ROSE_DATE, INTEGER_32]}.make (Void, 0), "select [date] as first, unit_id as second from date_unit")
			if not l_selector.success then
				rsg.error.raise ("Could not create l_selector")
			end
			create l_container.make_filled (Void, 1, 0)
			l_selector.set_container (l_container)
			l_selector.retrieve (Void, Void)
			if not l_selector.success then
				rsg.error.raise ("Could not execute l_selector")
			end
			l_container := l_selector.container
			from
				i := l_container.lower
			until
				i > l_container.upper
			loop
				a_table.search (l_container [i].first)
				if a_table.found then
					if a_table.found_item /= l_container [i].second then
						rsg.error.raise ("date prime table corrupt for " + l_container [i].first.ccyymmdd.out)
					end
				else
					a_table [l_container.item (i).first] := l_container.item (i).second
				end
				i := i + 1
			end
		end

	units_rose_entity_background_server: DATA_SERVER is
			-- Connection to Rose Entity Background DB Server (will try to re-establish connection if lost)
		do
			Result := shared_units_rose_entity_background_server.item
			if Result = Void then
				new_shared_units_rose_entity_background_server
				Result := shared_units_rose_entity_background_server.item
			end
			if not Result.success then
				Result.disconnect
				new_shared_units_rose_entity_background_server
				Result := shared_units_rose_entity_background_server.item
				if Result = Void or else not Result.success then
					rsg.error.raise ("Could not connect to Rose Entity Background server")
				end
			end
		end

	shared_units_rose_entity_background_server: CELL [DATA_SERVER] is
			-- Saved Shared Connection for rose units use to Rose Entity Background DB Server
		once
			create Result.put (Void)
		end

	new_shared_units_rose_entity_background_server is
			-- Create new connection to Rose Entity Background DB Server
		local
			l_servers: DATABASE_SERVERS
			l_server: DATA_SERVER
		do
			create l_servers
			l_server := l_servers.new_server_from_repository (l_servers.rose_entity_background_server.logical)
			shared_units_rose_entity_background_server.put (l_server)
		end

	create_date_prime (a_date: ROSE_DATE) is
			-- Assign a new prime to `a_date' in the database
		local
			l_server: DATA_SERVER
			l_predicate, l_min_predicate: INTEGER_PREDICATE
			l_update_predicate, l_delete_predicate, l_insert_predicate: EXECUTION_PREDICATE
			l_prime: INTEGER_32
		do
			l_server := units_rose_entity_background_server
			l_server.begin_transaction
			-- Let's lock the min available unit_id (to avoid race condition)
			create l_update_predicate.make (l_server, "UPDATE available_unit SET unit_id=unit_id WHERE unit_id IN (select min(unit_id) from available_unit)")
			if not l_update_predicate.success then
				rsg.error.raise ("Could not create l_update_predicate")
			end
			l_update_predicate.execute (Void, Void)
			if not l_update_predicate.success then
				rsg.error.raise ("Could not execute l_update_predicate")
			else
				-- we have another race condition on the INSERT below where another process gets in with the same date that we
				-- are trying to insert -- then we get a SQL trace stating "Violation of PRIMARY KEY constraint"
				-- so...we should query here first to see if date already there, if so, don't need to do anything else
				-- and can rollback
				create l_predicate.make (l_server, "SELECT unit_id FROM date_unit WHERE [date] = ?")
				if not l_predicate.success then
					rsg.error.raise ("Could not create l_predicate")
				end
				l_predicate.retrieve (Void, <<a_date>>)
				if not l_predicate.success then
					rsg.error.raise ("Could not execute l_predicate")
				end
				if l_predicate.success and then l_predicate.container.count = 0 then
					-- OK, there is no record with this date already in date_unit table
					create l_min_predicate.make (l_server, "SELECT MIN (unit_id) FROM available_unit")
					if not l_min_predicate.success then
						rsg.error.raise ("Could not create l_min_predicate")
					end
					l_min_predicate.retrieve (Void, Void)
					if not l_min_predicate.success then
						rsg.error.raise ("Could not execute l_min_predicate")
					else
						l_prime := l_min_predicate.container.item (1)
						create l_delete_predicate.make (l_server, "DELETE FROM available_unit WHERE unit_id = ?")
						if not l_delete_predicate.success then
							rsg.error.raise ("Could not create l_delete_predicate")
						end
						l_delete_predicate.execute (Void, <<l_prime>>)
						if not l_delete_predicate.success then
							rsg.error.raise ("Could not execute l_delete_predicate")
						else
							create l_insert_predicate.make (l_server, "INSERT INTO date_unit ([date], unit_id)  VALUES (?, ?)")
							if not l_insert_predicate.success then
								rsg.error.raise ("Could not create l_insert_predicate")
							end
							l_insert_predicate.execute (Void, <<a_date, l_prime>>)
							if not l_insert_predicate.success then
								rsg.error.raise ("Could not execute l_insert_predicate")
							end
						end
					end
				end
			end
			l_server.commit
		end

feature {NONE} -- Identified units support

	identified_units_table: HASH_TABLE [ROSE_IDENTIFIED_UNITS, STRING] is
			-- Table of identified units
		once
			create Result.make (17)
		end

	identified_units_prime_numbers_table: HASH_TABLE [INTEGER_32, STRING] is
			-- Identifier to prime number lookup table
		once
			create Result.make (17)
			retrieve_identified_units_prime_numbers_table (Result)
		end

	retrieve_identified_units_prime_numbers_table (a_table: like identified_units_prime_numbers_table) is
			-- Load data into `identified_units_prime_numbers_table'
		local
			l_selector: SELECTION_PREDICATE [PAIR [STRING, INTEGER_32]]
			l_container: ARRAY [PAIR [STRING, INTEGER_32]]
			i: INTEGER_32
			l_server: DATA_SERVER
		do
			l_server := units_rose_entity_background_server
			create l_selector.make (l_server, create {PAIR [STRING, INTEGER_32]}.make ("", 0), "select identifier as first, unit_id as second from identified_unit")
			if not l_selector.success then
				rsg.error.raise ("Could not create l_selector")
			end
			create l_container.make_filled (Void, 1, 0)
			l_selector.set_container (l_container)
			l_selector.retrieve (Void, Void)
			if not l_selector.success then
				rsg.error.raise ("Could not execute l_selector")
			end
			l_container := l_selector.container
			from
				i := l_container.lower
			until
				i > l_container.upper
			loop
				a_table.search (l_container [i].first)
				if a_table.found then
					if a_table.found_item /= l_container [i].second then
						rsg.error.raise ("identified units prime table corrupt for identifer [" + l_container [i].first + "]")
					end
				else
					a_table [l_container.item (i).first] := l_container.item (i).second
				end
				i := i + 1
			end
		end

	create_identified_units_prime (a_identifier: STRING) is
			-- Assign a new prime to `a_identifier' in the database
		require
			valid_identifier: a_identifier /= Void and not a_identifier.is_empty
		local
			l_server: DATA_SERVER
			l_predicate, l_min_predicate: INTEGER_PREDICATE
			l_update_predicate, l_delete_predicate, l_insert_predicate: EXECUTION_PREDICATE
			l_prime: INTEGER_32
		do
			l_server := units_rose_entity_background_server
			l_server.begin_transaction
			-- Let's lock the min available unit_id (to avoid race condition)
			create l_update_predicate.make (l_server, "UPDATE available_unit SET unit_id=unit_id WHERE unit_id IN (select min(unit_id) from available_unit)")
			if not l_update_predicate.success then
				rsg.error.raise ("Could not create l_update_predicate")
			end
			l_update_predicate.execute (Void, Void)
			if not l_update_predicate.success then
				rsg.error.raise ("Could not execute l_update_predicate")
			else
				create l_predicate.make (l_server, "SELECT unit_id FROM identified_unit WHERE [identifier] = ?")
				if not l_predicate.success then
					rsg.error.raise ("Could not create l_predicate")
				end
				l_predicate.retrieve (Void, <<a_identifier>>)
				if not l_predicate.success then
					rsg.error.raise ("Could not execute l_predicate")
				elseif l_predicate.container.is_empty then
					-- OK, there is no record with this identifer already in identified_unit table
					create l_min_predicate.make (l_server, "SELECT MIN (unit_id) FROM available_unit")
					if not l_min_predicate.success then
						rsg.error.raise ("Could not create l_min_predicate")
					end
					l_min_predicate.retrieve (Void, Void)
					if not l_min_predicate.success then
						rsg.error.raise ("Could not execute l_min_predicate")
					else
						l_prime := l_min_predicate.container.item (1)
						create l_delete_predicate.make (l_server, "DELETE FROM available_unit WHERE unit_id = ?")
						if not l_delete_predicate.success then
							rsg.error.raise ("Could not create l_delete_predicate")
						end
						l_delete_predicate.execute (Void, <<l_prime>>)
						if not l_delete_predicate.success then
							rsg.error.raise ("Could not execute l_delete_predicate")
						else
							create l_insert_predicate.make (l_server, "INSERT INTO identified_unit ([identifier], [unit_id]) VALUES (?, ?)")
							if not l_insert_predicate.success then
								rsg.error.raise ("Could not create l_insert_predicate")
							end
							l_insert_predicate.execute (Void, <<a_identifier, l_prime>>)
							if not l_insert_predicate.success then
								rsg.error.raise ("Could not execute l_insert_predicate")
							end
						end
					end
				end
			end
			l_server.commit
		end

feature -- Once/constants.

	pure_value: INTEGER is 1
			-- The numerator or demoninator value used in PURE_UNITS.  A 'once' function would
			-- be appropriate but I was going for speed.

	share_value: INTEGER is
			-- The numerator or denominator used when "share" is specified.
		once
			Result := generic_share_units.numerator.to_integer
		end

	generic_share_units : ROSE_UNITS
			--Generic share units, should eventually be replaced by specific instrument units
		once
			Result := units_table ["share"]
		end

feature {NONE} -- Implementation

	rsg: ROSE_SHARED_GLOBAL is
			-- Need to include here because of inheritance issues with
			-- RP_ALLOCATION and RBC_VARIABLE
		once
			create Result
		ensure
			valid_result: Result /= Void
		end

end
