indexing
	description: "Information available on a given user, this is fetched from Active Directory"
	author: "Zoran Simic"
	date: "$Date: 2008/12/01 $"
	revision: "$Revision: #21 $"

class
	ROSE_USER

inherit
	DEBUG_OUTPUT
		redefine
			out
		end
	HASHABLE
		redefine
			out
		end

create
	make, make_by_login

feature {NONE} -- Initialization

	make (a_login_domain, a_login_id: STRING) is
			-- Create with 'a_login_domain' and 'a_login_id'.
		require
			loin_domain_ok: a_login_domain /= Void
			loin_domain_id: a_login_id /= Void
		do
			login_domain := a_login_domain.as_upper
			login_id := a_login_id.as_upper
			login_domain.left_adjust
			login_domain.right_adjust
			login_id.left_adjust
			login_id.right_adjust
			is_identified := True
			if login_id.is_empty then
				login_id := once "unknown"
				is_identified := False
			end
			if login_domain.is_empty then
				full_login := login_id
			else
				full_login := login_domain + "\" + login_id
			end
			display_name := full_login
			first_name := ""
			last_name := login_id
		end

	make_by_login (a_full_login: STRING) is
			-- Create with 'a_full_login'.
		require
			full_login_ok: a_full_login /= Void and then a_full_login.has ('\')
		local
			i: INTEGER
		do
			i := a_full_login.index_of ('\', 1)
			make (a_full_login.substring (1, i - 1), a_full_login.substring (i + 1, a_full_login.count))
		end

feature -- Access

	login_id: STRING
			-- User running the application (example: "ZS")

	login_domain: STRING
			-- Domain of user running the application (example: "RIEM")

	unique_id: STRING
			-- A unique identifier for this user (example: "{2b325d1e-1ee0-487a-bbd0-95bec88ec4d5}")

	full_login: STRING
			-- 'login_domain'\'login_id' (example: "RIEM\ZS")

	first_name, last_name, display_name: STRING
			-- First and last name of the user (example: "Zoran", "Simic", "Zoran Simic")

	active_directory_location: STRING
			-- (example: "axaros.com/Orinda/Users/RC")

	spn: STRING
			-- Service Principal Names (example: "zs@axaros.com")

	hash_code: INTEGER is
			-- Hash code value
		do
			Result := full_login.hash_code
		end

	email_address: STRING is
			-- (example: "zs@axarosenberg.com")
		do
			Result := login_id + "@axarosenberg.com"
		ensure
			result_ok: Result /= Void
		end

	full_email_address: STRING is
			-- (example: "Zoran Simic<zs@axarosenberg.com>")
		do
			if display_name = Void then
				Result := email_address
			else
				Result := display_name + "<" + email_address + ">"
			end
		ensure
			result_ok: Result /= Void
		end

	is_identified: BOOLEAN
			-- Is current user identified successfully?
			-- A Windows bug prevents sometimes identifying the current user

	corresponds_to (a_id: STRING): BOOLEAN is
			-- Does this user correspond to other user with 'a_id'?
		do
			Result := equal (login_id, a_id) or else equal (full_login, a_id)
		end

	out, debug_output: STRING is
			-- New string containing terse printable representation of current object
		do
			create Result.make_empty
			Result.append_string (full_login)
		end

feature -- Basic operations

	set_display_name (a_name: STRING) is
			-- Set 'display_name' with 'a_name'.
		require
			name_ok: a_name /= Void
		local
			i: INTEGER
		do
			display_name := a_name.twin
			i := a_name.index_of (' ', 1)
			if i = 0 then
				i := a_name.index_of ('%T', 1)
			end
			if i > 0 then
				first_name := a_name.substring (1, i - 1)
				last_name := a_name.substring (i + 1, a_name.count)
			end
			if first_name = Void then
				first_name := once ""
			end
			if last_name = Void then
				last_name := login_id
			end
		end

	set_active_directory_location (a_path: STRING) is
			-- Set 'active_directory_location'.
		do
			active_directory_location := a_path
		end

	set_unique_id (a_id: STRING) is
			-- Set 'unique_id'.
		do
			unique_id := a_id
		end

	set_spn (a_spn: STRING) is
			-- Set 'spn'.
		do
			spn := a_spn
		end

invariant

	valid_login_id: login_id /= Void and then not login_id.is_empty
	valid_login_domain: login_domain /= Void
	valid_full_login: full_login /= Void
	valid_display_name: display_name /= Void
	valid_first_name: first_name /= Void
	valid_last_name: last_name /= Void

end
