indexing
	description: "User information"
	author: "Zoran Simic"
	date: "$Date: 2010/04/15 $"
	revision: "$Revision: #10 $"

class
	ROSE_USER_PREFERENCES

inherit
	ROSE_SHARED_GLOBAL
		rename
			config as global_config
		end

create
	make

feature {NONE} -- Initialization

	make (a_user: ROSE_USER; a_path: STRING) is
			-- Make for 'a_user' and with 'a_path'.
		require
			a_user_not_void: a_user /= Void
			a_path_not_void: a_path /= Void
		do
			user := a_user
			create history_hash.make (0)
			create config.make_default
			if user.is_identified then
				config.set_path (a_path)
				refresh_configuration
			end
		end

feature -- Access

	user: ROSE_USER
			-- Associated user

	name: STRING is
			-- User login name
		do
			Result := user.login_id
		end

	login_domain: STRING is
			-- User login domain
		do
			Result := user.login_domain
		end

	full_login: STRING is
			-- 'login_domain'\'name'
		do
			Result := user.full_login
		end

	is_identified: BOOLEAN is
			-- Is current user identified?
		do
			Result := user.is_identified
		end

	config: ROSE_CONFIGURATION
			-- User configuration

	history (a_key: STRING): ROSE_INPUT_HISTORY is
			-- History of inputs for 'a_key' (specific to the current user);
			-- Return Void if this history has not been built yet
		do
			Result := history_hash.item (safe_key (a_key))
		end

	get_history (a_key: STRING; a_capacity: INTEGER; a_shell_like: BOOLEAN): ROSE_INPUT_HISTORY is
			-- History of inputs for 'a_key' (specific to the current user);
			-- Create a new object if it has not been built yet
		require
			a_key_not_void: a_key /= Void
			a_key_not_empty: not a_key.is_empty
		local
			l_key: STRING
		do
			l_key := safe_key (a_key)
			Result := history_hash.item (l_key)
			if Result = Void then
				if a_shell_like then
					create Result.make_shell (a_capacity, config)
				else
					create Result.make_browser (a_capacity, config)
				end
				Result.set_configuration_key (l_key)
				history_hash.force (Result, l_key)
			end
		ensure
			history_not_void: Result /= Void
		end

	remembered_value (a_key: STRING): STRING is
			-- STRING value associated to 'a_key' remembered via the user config
		local
			l_node: ROSE_CONFIGURATION_NODE
		do
			l_node := remembered_node (k_historized, safe_key (a_key), False)
			if l_node /= Void then
				Result := l_node.value
			end
		end

	remembered_boolean (a_key: STRING): BOOLEAN is
			-- BOOLEAN value associated to 'a_key' remembered via the user config
		do
			Result := config.boolean (k_checkbox + safe_key (a_key), False)
		end

	remembered_integer (a_key: STRING): INTEGER is
			-- INTEGER value associated to 'a_key' remembered via the user config
		local
			l_string: STRING
		do
			l_string := remembered_value (a_key)
			if l_string /= Void and then l_string.is_integer then
				Result := l_string.to_integer
			end
		end

	has_remembered (a_key: STRING): BOOLEAN is
			-- Is an INTEGER or STRING associated to key `a_key'.
			-- Use `has_remembered_boolean' for boolean values.
		do
			Result := remembered_value (a_key) /= Void
		end

	has_remembered_boolean (a_key: STRING): BOOLEAN is
			-- Is a boolen with 'a_key' been remembered?
		do
			Result := config.has (k_checkbox + safe_key (a_key))
		end

	remembered_node (a_category, a_key: STRING; a_forced: BOOLEAN): ROSE_CONFIGURATION_NODE is
			-- Remembered node corresponding to 'a_category' / 'a_key'
			-- If 'a_force' is true, then key is created if it didn't exist before
		local
			l_key: STRING
		do
			l_key := a_category + safe_key (a_key)
			Result := config.node (l_key)
			if Result = Void and then a_forced then
				config.set (l_key, "")
				Result := config.node (l_key)
			end
		end

	safe_key (a_key: STRING): STRING is
			-- Equivalent save version of 'a_key'
		require
			key_ok: a_key /= Void
		local
			i: INTEGER
			c: CHARACTER
		do
			from
				create Result.make (a_key.count)
				i := 1
			until
				i > a_key.count
			loop
				c := a_key.item (i)
				if c = '\' or else c = '/' then
					Result.append_character ('/')
				elseif c.is_alpha_numeric then
					Result.append_character (c.as_lower)
				else
					Result.append_character ('_')
				end
				i := i + 1
			end
			Result.replace_substring_all ("//", "/")
		end

feature -- Basic operations

	refresh_configuration is
			-- Read user configuration from disk.
		do
			if user.is_identified then
				history_hash.clear_all
				error.push_private (Void)
				Global.config_manager.load (config, Void, config.path)
				error.pop_private
--				error.disregard
			end
		end

	save_configuration is
			-- Save the user configuration if modified.
		do
			if user.is_identified then
				from
					history_hash.start
				until
					history_hash.off
				loop
					save_history (history_hash.key_for_iteration, history_hash.item_for_iteration)
					history_hash.forth
				end
				if config.path /= Void and then config.is_modified then
					error.push_private (Void)
					Global.config_manager.save (config, Void, config.path)
					error.pop_private
				end
			end
		end

	set_remembered_value (a_key: STRING; a_value: STRING) is
			-- Remember 'a_value' for 'a_key'
		local
			l_node: ROSE_CONFIGURATION_NODE
		do
			l_node := remembered_node (k_historized, safe_key (a_key), True)
			l_node.set_value (a_value)
		end

	set_remembered_boolean (a_key: STRING; a_value: BOOLEAN) is
			-- Remember boolean 'a_value' for 'a_key'
		local
			l_node: ROSE_CONFIGURATION_NODE
		do
			l_node := remembered_node (k_checkbox, safe_key (a_key), True)
			l_node.set_value (a_value.out)
		end

	set_remembered_integer (a_key: STRING; a_value: INTEGER) is
			-- Remember integer 'a_value' for 'a_key'
		do
			set_remembered_value (a_key, a_value.out)
		end

feature -- Removal

	delete_history (a_key: STRING) is
			-- Remove history with id `a_key' from user configuration.
		require
			a_key_not_void: a_key /= Void
			a_key_not_empty: not a_key.is_empty
		local
			l_history: ROSE_INPUT_HISTORY
			l_key: STRING
		do
			l_key := safe_key (a_key)
			l_history := history (l_key)
			if l_history /= Void then
				config.delete (l_history.configuration_key)
				history_hash.remove (l_key)
			end
		end

feature {NONE} -- Implementation

	history_hash: HASH_TABLE [ROSE_INPUT_HISTORY, STRING]
			-- History size for each id

	save_history (a_key: STRING; a_history: ROSE_INPUT_HISTORY) is
			-- Save history for 'a_history' under 'a_key'.
		local
			i: INTEGER
			l_node: ROSE_CONFIGURATION_NODE
			l_items: ARRAY [STRING]
		do
			if a_key /= Void and then not a_key.is_empty and then a_history /= Void then
				l_node := remembered_node (k_historized, safe_key (a_key), True)
				l_node.delete_children
				l_items := a_history.items
				if l_items /= Void then
					from
						i := l_items.lower
					until
						i > l_items.upper
					loop
						l_node.add_anonymous (l_items.item (i))
						i := i + 1
					end
				end
			end
		end

	k_historized: STRING is "history/"	-- Constant
	k_checkbox: STRING is "checkbox/"	-- Constant

invariant

	user_not_void: user /= Void
	config_not_void: config /= Void
	history_has_not_void: history_hash /= Void

end
