note

	description:

		"Hash table of elements that does not grow beyond a given size (by removing the oldest ones)"

	author: "Mark Howard"
	last_editor: "$Author: mb $"
	date: "$Date: 2014/05/14 $"
	revision: "$Revision: #28 $"
	path: "$File: //rose/source/kernel/infrastructure/containers/rose_volatile_cache.e $"

class ROSE_VOLATILE_CACHE [G, H -> HASHABLE]

inherit
	ELOG_SHARED_LOGGER

create
	make

feature {NONE} -- Initialization

	make (a_max_count, a_max_chunk_size: INTEGER) is
			-- Initializes a cache of `a_max_count' chunks, each one with a max
			-- size of `a_max_chunk_size'
		require
			valid_maximums: a_max_count > 0 and then a_max_chunk_size > 0
			count_bigger_than_chunk: a_max_count > a_max_chunk_size
		do
			set_maximum_count (a_max_count)
			set_maximum_chunk_size (a_max_chunk_size)
			chunk_count := 1
			clear_all
		end

feature -- Access

	item (a_key: H): G is
			-- Cached item corresponding to `a_key'
		local
			l_volatile: ROSE_VOLATILE [like item]
		do
			data.search (a_key)
			if data.found then
				l_volatile := data.found_item
				if l_volatile /= Void then
					expiration_table.renew (l_volatile.expiration)
					Result := l_volatile.item
				end
			end
		ensure
			found_item_is_newest: (data.found and data.found_item /= Void) implies data.found_item.expiration = expiration_table.newest
		end

	search (a_key: H) is
			-- Search for item corresponding to `a_key'
		local
			l_item: G
		do
			-- Same code as `item' except that we are not interested in the Result
			l_item := item (a_key)
		end

	found: BOOLEAN is
			-- Has the last `search' found an item?
		do
			Result := data.found
		end

	found_item: G is
			-- Item touched by last call to 'item', 'put' or 'force'
		local
			l_volatile: ROSE_VOLATILE [like item]
		do
			l_volatile := data.found_item
			if l_volatile /= Void then
				Result := l_volatile.item
			end
		end

	cache_count: INTEGER is
			-- Count of cache items currently stored
		do
			Result := data.count
		end

feature -- Basic operations

	set_maximum_chunk_size (a_value: like maximum_chunk_size) is
			-- Set 'maximum_chunk_size' to 'a_value'
		do
			maximum_chunk_size := a_value
		end

	set_maximum_count (a_value: like maximum_count) is
			-- Set 'maximum_count' to 'a_value'
		do
			if a_value > maximum_count
			then
				logger.debugging ("Setting maximum_count to : " + a_value.out + " for " + generating_type)
			end

			maximum_count := a_value
		end

	remove (a_key : H) is
			-- Remove the item keyed by 'a_key' from the cache
		do
			data.remove (a_key)
		end

	force, put (a_item: like item; a_key: H) is
			-- Force 'a_item' into the cache even if it already exists.
		local
			l_volatile: ROSE_VOLATILE [like item]
		do
			data.search (a_key)
			if data.found then
				l_volatile := data.found_item
			else
				create l_volatile
				data.extend (l_volatile,a_key)	-- Extend is cheaper and OK since we know it was not found.
				increment_chunk_size
			end
			l_volatile.set (a_item,chunk_count)
		end

	clear_all, refresh is
			-- Remove all items from this cache
			-- 'refresh' is not the right name, left here because a few clients still use it
		do
			create expiration_table.make
			expiration_table.extend (1)
			create data.make (maximum_count)
		end

feature {PERSISTENT_CLUSTERED_TAGGED_SERVER_MONITOR_COMMAND,
	SIGNATURE_TAGGED_EXPRESSION_CACHE} -- Should not be exported but this requires code review of clients

	maximum_chunk_size: INTEGER
			-- Maximum chunk size
			-- The 'chunk size' indicates how many items to keep per chunk

feature {PERSISTENT_CLUSTERED_TAGGED_SERVER_MONITOR_COMMAND,
	PERSISTENT_CLUSTERED_TAGGED_SERVER_COMMAND,
	SIGNATURE_TAGGED_EXPRESSION_CACHE,
	CLUSTERED_TAGGED_ITEM_DATABASE} -- Should not be exported but this requires code review of clients

	maximum_count: INTEGER
			-- Maximum count
			-- The 'count' indicates how many chunks to keep in the cache

feature {PERSISTENT_CLUSTERED_TAGGED_SERVER_MONITOR_COMMAND,
	INTEGER_INTEGER_CLUSTER_VIEW,
	INTEGER_DATE_CLUSTER_VIEW} -- Should not be exported but this requires code review of clients

	expiration_table: ROSE_HASHED_QUEUE [INTEGER]
			-- Expirations of cached items

feature {PERSISTENT_CLUSTERED_TAGGED_SERVER_MONITOR_COMMAND,
	PERSISTENT_CLUSTERED_TAGGED_SERVER_COMMAND,
	INTEGER_INTEGER_CLUSTER_VIEW,
	INTEGER_DATE_CLUSTER_VIEW,
	SIMILARS_BHAT,
	SEGMENT_BACKGROUND_VARIABLE,
	MODEL_OIL_BETA,
	RP_VALUATION_MODEL_ACCESS,
	IMPACT_ANALYSIS_SIMULATIONS_FILE_DATASOURCE,
	VALDATA_FUNDDATA_EXTERNAL_RECORDS} -- Should not be exported but this requires code review of clients

	data: HASH_TABLE [ROSE_VOLATILE [like item], H]
			-- Cached items

	swept_item_count: INTEGER
  			-- Number of items that 'sweep' has removed.
		note
			option: transient
		attribute
		end

feature {NONE} -- Implementation

	chunk_size: INTEGER
			-- Current chunk's size of current chunks
	chunk_count: INTEGER
			-- Current chunk's number of current chunks

	increment_chunk_size is
			-- Increment `chunk_size'.
			-- Move to the next chunk if it reached `maximum_chunk_size'.
		do
			chunk_size := chunk_size + 1
			if chunk_size >= maximum_chunk_size then
				chunk_size := 0
				chunk_count := chunk_count + 1
				expiration_table.extend (chunk_count)
			end
			from until data.count <= maximum_count loop
				sweep
			end
		end

	sweep is
			-- Remove all items that have the same expiration as the oldest item.
		local
			l_expired: INTEGER
			l_keys: ROSE_LINEAR_ARRAY [H]
		do
			l_expired := expiration_table.oldest
			create l_keys.make_empty
			from data.start until data.off loop
				if data.item_for_iteration.expiration =  l_expired then
					l_keys.extend (data.key_for_iteration)
				end
				data.forth
			end

			swept_item_count := swept_item_count + l_keys.count

			from l_keys.start until l_keys.off loop
				data.remove (l_keys.current_item)
				l_keys.forth
			end
			expiration_table.pop
		ensure
			one_less_expiration: (not old expiration_table.is_empty) = (expiration_table.count = old expiration_table.count -1)
			same_expiration: (old expiration_table.is_empty) = (expiration_table.count = old expiration_table.count)
			less_items: data.count <= old data.count
		end

end
