note

	description:

		"Accounting System based on contracts/flows"

	fileid: "$Id: //rose/source/kernel/expert_system/rp_accounting_system.e#55 $"
	author: "Mark Howard"
	last_editor: "$Author: c4 $"
	date: "$Date: 2015/04/09 $"
	revision: "$Revision: #55 $"
	path: "$File: //rose/source/kernel/expert_system/rp_accounting_system.e $"

class RP_ACCOUNTING_SYSTEM

inherit

	RP_INVESTMENT_MANAGER_CLIENT
		redefine
			make
		end

	RP_INSTRUMENT_CONTROL_PURPOSE_AWARE
		export {NONE} all end

	DT_SHARED_SYSTEM_CLOCK
		export {NONE} all end

	RP_TRACEABLE

	RP_FLOW_UTILS
		export {NONE} all end

create {RP_INVESTMENT_MANAGER}

	make

feature {NONE} -- Initialization

	make (a_manager: like manager) is
			-- Initialize `Current'
		do
			Precursor (a_manager)
			clear_caches
		end

feature -- Initialization

	clear_caches is
			-- Clear caches
		local
			l_comparer: KL_AGENT_EQUALITY_TESTER [TUPLE [RP_PORTFOLIO, RP_CUSTODY]]
		do
			create long_side_pp.make_equal (17)
			create portfolio_contracts.make_default
			create portfolio_holdings.make_default
			create portfolio_settled_holdings.make_default
			create portfolio_holdings_per_custody.make_default
			create portfolio_settled_holdings_per_custody.make_default
			create investment_account_settled_holdings_per_custody.make_default
			create l_comparer.make (agent compare_portfolio_custody)
			portfolio_holdings_per_custody.set_key_equality_tester (l_comparer)
			portfolio_settled_holdings_per_custody.set_key_equality_tester (l_comparer)
			create flows_cache.make_default
			create account_flows_cache.make_default
				-- Aggregator for account
			create account_contracts.make_default
			create account_holdings.make_default
			create account_settled_holdings.make_default
				-- Clear state
			has_aggregated_flows := False
			is_account_level_aggregation_enabled := False
		ensure
			empty_portfolio_contracts: portfolio_contracts.is_empty
			empty_portfolio_holdings: portfolio_holdings.is_empty
			empty_portfolio_settled_holdings: portfolio_settled_holdings.is_empty
			empty_flows_cache: flows_cache.is_empty
			empty_account_flows_cache: account_flows_cache.is_empty
			empty_account_holdings: account_holdings.is_empty
			empty_account_settled_holdings: account_settled_holdings.is_empty
			no_aggregate_flows: not has_aggregated_flows
			no_account_level_aggregation: not is_account_level_aggregation_enabled
		end

feature -- Status Report

	is_account_level_aggregation_enabled: BOOLEAN
			-- Is account level aggregation enabled?

	has_aggregated_flows: BOOLEAN
			-- Have any flows been aggregated?

	is_forwarding_contracts_to_cost_accounting_system: BOOLEAN
			-- Are new contracts forwarded for cost accounting?

feature -- Status Setting

	enable_account_level_aggregation is
			-- Enable account level aggregation
		require
			no_aggregation_occured: not has_aggregated_flows
		do
			is_account_level_aggregation_enabled := True
		ensure
			definition: is_account_level_aggregation_enabled
		end

	disable_account_level_aggregation is
			-- Disable account level aggregation
		require
			no_aggregation_occured: not has_aggregated_flows
		do
			is_account_level_aggregation_enabled := False
		ensure
			definition: not is_account_level_aggregation_enabled
		end

	set_is_forwarding_contracts_to_cost_accounting_system (a_is_forwarding_contracts_to_cost_accounting_system: like is_forwarding_contracts_to_cost_accounting_system) is
			-- Set 'is_forwarding_contracts_to_cost_accounting_system' to 'a_is_forwarding_contracts_to_cost_accounting_system'
		do
			is_forwarding_contracts_to_cost_accounting_system := a_is_forwarding_contracts_to_cost_accounting_system
		ensure
			is_forwarding_contracts_to_cost_accounting_system_set: is_forwarding_contracts_to_cost_accounting_system = a_is_forwarding_contracts_to_cost_accounting_system
		end

feature -- Query

	portfolio_positions (a_portfolio: RP_PORTFOLIO): RP_INSTRUMENT_HISTORY is
			-- Compute portfolio positions
		require
			non_void_args: a_portfolio /= Void
		local
			l_holdings: RP_HOLDINGS
		do
			create Result.make_default
			if portfolio_holdings.has (a_portfolio) then
				l_holdings := portfolio_holdings.item (a_portfolio)
				l_holdings.instruments.do_all (agent add_position (Result, l_holdings, ?))
			end
		ensure
			non_void_positions: Result /= Void
		end

	portfolio_settled_positions (a_portfolio: RP_PORTFOLIO): RP_INSTRUMENT_HISTORY is
			-- Settled positions for `a_portfolio'
		require
			a_portfolio_not_void: a_portfolio /= Void
		local
			l_holdings: RP_HOLDINGS
		do
			create Result.make_default
			if portfolio_settled_holdings.has (a_portfolio) then
				l_holdings := portfolio_settled_holdings.item (a_portfolio)
				l_holdings.instruments.do_all (agent add_position (Result, l_holdings, ?))
			end
		ensure
			portfolio_settled_positions_not_void: Result /= Void
		end

	portfolio_holdings_at (a_portfolio: RP_PORTFOLIO; a_date: ROSE_DATE): ARRAY [RP_HOLDING_INTERVAL] is
			-- Holdings of `a_portfolio' on `a_date'
		require
			a_portfolio_non_void: a_portfolio /= Void
			a_date_not_void: a_date /= Void
		do
			if portfolio_holdings.has (a_portfolio) then
				create Result.make_from_array (portfolio_holdings.item (a_portfolio).holdings_at (a_date))
			else
				Result := <<>>
			end
		ensure
			non_void_list: Result /= Void
			no_void_holding: not Result.has (Void)
			all_holdings_for_a_date: Result.for_all (agent {RP_HOLDING_INTERVAL}.brackets (a_date))
		end

	portfolio_holdings_per_custody_at (a_portfolio: RP_PORTFOLIO; a_custody: RP_CUSTODY; a_date: ROSE_DATE): ARRAY [RP_HOLDING_INTERVAL] is
			-- Holdings of `a_portfolio' at `a_custody' on `a_date'
		require
			a_portfolio_non_void: a_portfolio /= Void
			a_custody_not_void: a_custody /= Void
			a_date_not_void: a_date /= Void
		do
			if portfolio_holdings_per_custody.has ([a_portfolio, a_custody]) then
				create Result.make_from_array (portfolio_holdings_per_custody.item ([a_portfolio, a_custody]).holdings_at (a_date))
			else
				Result := <<>>
			end
		ensure
			non_void_list: Result /= Void
			no_void_holding: not Result.has (Void)
			all_holdings_for_a_date: Result.for_all (agent {RP_HOLDING_INTERVAL}.brackets (a_date))
		end

	portfolio_holdings_for (a_portfolio: RP_PORTFOLIO; a_instrument: RP_INSTRUMENT): RP_HOLDING_HISTORY is
			-- History of a_instrument for a_portfolio
		require
			non_void_args: a_portfolio /= Void and a_instrument /= Void
		do
			if portfolio_holdings.has (a_portfolio) then
				Result := portfolio_holdings.item (a_portfolio).holdings_for (a_instrument)
			end
			if Result = Void then
				create Result.make_empty (trade_date_agent)
			end
		ensure
			non_void_list: Result /= Void
		end

	settled_portfolio_holdings_for (a_portfolio: RP_PORTFOLIO; a_instrument: RP_INSTRUMENT): RP_HOLDING_HISTORY is
			-- History of settled positions of `a_instrument' for `a_portfolio'
		require
			non_void_args: a_portfolio /= Void and a_instrument /= Void
		do
			if portfolio_settled_holdings.has (a_portfolio) then
				Result := portfolio_settled_holdings.item (a_portfolio).holdings_for (a_instrument)
			end
			if Result = Void then
				create Result.make_empty (settle_date_agent)
			end
		ensure
			non_void_list: Result /= Void
		end

	settled_portfolio_holdings_per_custody_for (a_portfolio: RP_PORTFOLIO; a_custody: RP_CUSTODY; a_instrument: RP_INSTRUMENT): RP_HOLDING_HISTORY is
			-- History of settled positions of `a_instrument' for `a_portfolio' at `a_custody'
		require
			a_portfolio_not_void: a_portfolio /= Void
			a_custody_not_void: a_custody /= Void
			a_instrument_not_void: a_instrument /= Void
		do
			if portfolio_settled_holdings_per_custody.has ([a_portfolio, a_custody]) then
				Result := portfolio_settled_holdings_per_custody.item ([a_portfolio, a_custody]).holdings_for (a_instrument)
			end
			if Result = Void then
				create Result.make_empty (settle_date_agent)
			end
		ensure
			non_void_list: Result /= Void
		end

	settled_portfolio_holdings_at (a_portfolio: RP_PORTFOLIO; a_date: ROSE_DATE): ARRAY [RP_HOLDING_INTERVAL] is
			-- Settled holdings of `a_portfolio' on `a_date'
		require
			a_portfolio_non_void: a_portfolio /= Void
			a_date_not_void: a_date /= Void
		do
			if portfolio_settled_holdings.has (a_portfolio) then
				create Result.make_from_array (portfolio_settled_holdings.item (a_portfolio).holdings_at (a_date))
			else
				Result := <<>>
			end
		ensure
			non_void_list: Result /= Void
			no_void_holding: not Result.has (Void)
			all_holdings_for_a_date: Result.for_all (agent {RP_HOLDING_INTERVAL}.brackets (a_date))
		end

	settled_portfolio_holdings_per_custody_at (a_portfolio: RP_PORTFOLIO; a_custody: RP_CUSTODY; a_date: ROSE_DATE): ARRAY [RP_HOLDING_INTERVAL] is
			-- Settled holdings of `a_portfolio' at `a_custody' on `a_date'
		require
			a_portfolio_non_void: a_portfolio /= Void
			a_custody_not_void: a_custody /= Void
			a_date_not_void: a_date /= Void
		do
			if portfolio_settled_holdings_per_custody.has ([a_portfolio, a_custody]) then
				create Result.make_from_array (portfolio_settled_holdings_per_custody.item ([a_portfolio, a_custody]).holdings_at (a_date))
			else
				Result := <<>>
			end
		ensure
			non_void_list: Result /= Void
			no_void_holding: not Result.has (Void)
			all_holdings_for_a_date: Result.for_all (agent {RP_HOLDING_INTERVAL}.brackets (a_date))
		end

	settled_investment_account_holdings_per_custody_for (a_account: RP_INVESTMENT_ACCOUNT; a_custody: RP_CUSTODY; a_instrument: RP_INSTRUMENT): RP_HOLDING_HISTORY
			-- History of settled positions of `a_instrument' for `a_account' at `a_custody'
		require
			a_account_non_void: a_account /= Void
			a_custody_not_void: a_custody /= Void
			a_instrument_not_void: a_instrument /= Void
		do
			if investment_account_settled_holdings_per_custody.has ([a_account, a_custody]) then
				Result := investment_account_settled_holdings_per_custody.item ([a_account, a_custody]).holdings_for (a_instrument)
			end
			if Result = Void then
				create Result.make_empty (Settle_date_agent)
			end
		ensure
			non_void_list: Result /= Void
		end

	investment_account_positions (a_investment_account: RP_INVESTMENT_ACCOUNT): RP_INSTRUMENT_HISTORY is
			-- Compute portfolio positions
		require
			non_void_args: a_investment_account /= Void
		local
			l_holdings: RP_HOLDINGS
		do
			create Result.make_default
			if account_holdings.has (a_investment_account) then
				l_holdings := account_holdings.item (a_investment_account)
				l_holdings.instruments.do_all (agent add_position (Result, l_holdings, ?))
			end
		ensure
			non_void_positions: Result /= Void
		end

	investment_account_holdings_at (a_investment_account: RP_INVESTMENT_ACCOUNT; a_date: ROSE_DATE): ARRAY [RP_HOLDING_INTERVAL] is
			-- Holdings of `a_investment_account' on `a_date'
		require
			non_void_args: a_investment_account /= Void and a_date /= Void
		do
			if account_holdings.has (a_investment_account) then
				create Result.make_from_array (account_holdings.item (a_investment_account).holdings_at (a_date))
			else
				Result := <<>>
			end
		ensure
			non_void_list: Result /= Void
			no_void_holding: not Result.has (Void)
			all_holdings_for_a_date: Result.for_all (agent {RP_HOLDING_INTERVAL}.brackets (a_date))
		end

	settled_investment_account_holdings_at (a_investment_account: RP_INVESTMENT_ACCOUNT; a_date: ROSE_DATE): ARRAY [RP_HOLDING_INTERVAL] is
			-- Settled holdings of `a_investment_account' on `a_date'
		require
			a_investment_account_non_void: a_investment_account /= Void
			a_date_not_void: a_date /= Void
		do
			if account_settled_holdings.has (a_investment_account) then
				create Result.make_from_array (account_settled_holdings.item (a_investment_account).holdings_at (a_date))
			else
				Result := <<>>
			end
		ensure
			non_void_list: Result /= Void
			no_void_holding: not Result.has (Void)
			all_holdings_for_a_date: Result.for_all (agent {RP_HOLDING_INTERVAL}.brackets (a_date))
		end

	settled_investment_account_holdings_for (a_investment_account: RP_INVESTMENT_ACCOUNT; a_instrument: RP_INSTRUMENT): RP_HOLDING_HISTORY is
			-- History of settled positions of `a_instrument' for `a_investment_account'
		require
			a_investment_account_not_void: a_investment_account /= Void
			a_instrument_not_void: a_instrument /= Void
		do
			if account_settled_holdings.has (a_investment_account) then
				Result := account_settled_holdings.item (a_investment_account).holdings_for (a_instrument)
			end
			if Result = Void then
				create Result.make_empty (settle_date_agent)
			end
		ensure
			non_void_list: Result /= Void
		end

	enumerate_portfolio_instruments (a_portfolio: RP_PORTFOLIO): ARRAY [RP_INSTRUMENT] is
			-- All instruments ever held by portfolio
		require
			non_void_args: a_portfolio /= Void
		do
			if portfolio_holdings.has (a_portfolio) then
				Result := portfolio_holdings.item (a_portfolio).instruments
			else
				Result := <<>>
			end
		ensure
			non_void_result: Result /= Void
		end

	long_position (a_portfolio: RP_PORTFOLIO; a_instrument: RP_INSTRUMENT; a_date: ROSE_DATE): RP_QUANTITY is
			-- Long position of `a_instrument' in portfolio
		require
			a_portfolio_non_void: a_portfolio /= Void
			a_instrument_not_void: a_instrument /= Void
			a_date_not_void: a_date /= Void
		local
			l_history: RP_HOLDING_HISTORY
			l_interval: RP_HOLDING_INTERVAL
		do
			Result := a_instrument.quantity (0)
			if portfolio_holdings.has (a_portfolio) then
				l_history := portfolio_holdings.item (a_portfolio).holdings_for (a_instrument)
				if l_history /= Void then
					l_interval := l_history.witness_interval (a_date)
					if l_interval /= Void then
						Result := l_interval.quantity
					end
				end
			end
		ensure
			non_void_long_position: Result /= Void
			same_units: a_instrument.is_equal (Result.instrument)
		end

	settled_long_position (a_portfolio: RP_PORTFOLIO; a_instrument: RP_INSTRUMENT; a_date: ROSE_DATE): RP_QUANTITY is
			-- the long position of instrument in regards a portfolio
		require
			a_portfolio_non_void: a_portfolio /= Void
			a_instrument_not_void: a_instrument /= Void
			a_date_not_void: a_date /= Void
		local
			l_history: RP_HOLDING_HISTORY
			l_interval: RP_HOLDING_INTERVAL
		do
			Result := a_instrument.quantity (0)
			if portfolio_settled_holdings.has (a_portfolio) then
				l_history := portfolio_settled_holdings.item (a_portfolio).holdings_for (a_instrument)
				if l_history /= Void then
					l_interval := l_history.witness_interval (a_date)
					if l_interval /= Void then
						Result := l_interval.quantity
					end
				end
			end
		ensure
			non_void_settled_long_position: Result /= Void
			same_units: a_instrument.is_equal (Result.instrument)
		end

	short_position (a_portfolio: RP_PORTFOLIO; a_instrument: RP_INSTRUMENT): RP_INSTRUMENT_QUANTITY [RP_INSTRUMENT] is
			-- the short position of instrument in regards a portfolio
		require
			a_portfolio_non_void: a_portfolio /= Void
			a_instrument_not_void: a_instrument /= Void
		do
			-- to_be_implemented ("short position of instrument in regards a portfolio")
		end

	process_account_message (a_message: RP_CHANGE_MESSAGE [RP_PORTFOLIO_CONTRACT]) is
			-- handles trades,custodian and client
		require
			a_message_not_void: a_message /= Void
		do
			if a_message.is_deletion or else a_message.is_change then
				delete_contract (a_message.old_g)
			end
			if a_message.is_addition then
				post_new_contract (a_message.new_g)
			elseif a_message.is_change then
				post_changed_contract (a_message.new_g)
			end
		end

	flows_by (a_portfolio: RP_PORTFOLIO; a_filter: FUNCTION [ANY, TUPLE [RP_FLOW], BOOLEAN]): ROSE_LINEAR_ARRAY [RP_FLOW] is
			-- Flows for 'a_portfolio' filtered with 'a_filter'
		local
			l_flows: ARRAY [RP_FLOW]
			l_flow: RP_FLOW
			i: INTEGER
		do
			create Result.make_empty
			l_flows := flows_by_portfolio (a_portfolio)
			from
				i := l_flows.lower
			until
				i > l_flows.upper
			loop
				l_flow := l_flows.item (i)
				if a_filter.item ([l_flow]) then
					Result.extend (l_flow)
				end
				i := i + 1
			end
		end

	flows_by_portfolio (a_portfolio: RP_PORTFOLIO): ARRAY [RP_FLOW] is
			-- Flows for `a_portfolio'
		require
			non_void_portfolio: a_portfolio /= Void
		do
			if flows_cache.has (a_portfolio) then
				Result := flows_cache.item (a_portfolio)
			else
				Result := <<>>
			end
		ensure
			non_void_result: Result /= Void
		end

	account_flows_by (a_account: RP_INVESTMENT_ACCOUNT; a_filter: FUNCTION [ANY, TUPLE [RP_FLOW], BOOLEAN]): ROSE_LINEAR_ARRAY [RP_FLOW] is
			-- Flows for `a_account' filtered with `a_filter'
		require
			a_account_not_void: a_account /= Void
			a_filter_not_void: a_filter /= Void
		local
			l_flows: ARRAY [RP_FLOW]
			l_flow: RP_FLOW
			i: INTEGER
		do
			create Result.make_empty
			l_flows := flows_by_account (a_account)
			from
				i := l_flows.lower
			until
				i > l_flows.upper
			loop
				l_flow := l_flows.item (i)
				if a_filter.item ([l_flow]) then
					Result.extend (l_flow)
				end
				i := i + 1
			end
		ensure
			result_not_void: Result /= Void
			no_void_flow: not Result.has (Void)
		end

	flows_by_account (a_account: RP_INVESTMENT_ACCOUNT): ARRAY [RP_FLOW] is
			-- Flows for `a_account'
		require
			non_void_account: a_account /= Void
		do
			if account_flows_cache.has (a_account) then
				Result := account_flows_cache.item (a_account)
			else
				Result := <<>>
			end
		ensure
			non_void_result: Result /= Void
		end

	contracts_by_portfolio (a_portfolio: RP_PORTFOLIO): ARRAY [RP_PORTFOLIO_CONTRACT] is
			-- Contracts for `a_portfolio'
		require
			non_void_portfolio: a_portfolio /= Void
		do
			portfolio_contracts.search (a_portfolio)
			if portfolio_contracts.found then
				Result := portfolio_contracts.found_item.to_array
			else
				Result := <<>>
			end
		ensure
			non_void_result: Result /= Void
		end

	contracts_by_account (a_account: RP_INVESTMENT_ACCOUNT): ARRAY [RP_PORTFOLIO_CONTRACT] is
			-- Contracts for `a_account'
		require
			non_void_account: a_account /= Void
		do
			account_contracts.search (a_account)
			if account_contracts.found then
				Result := account_contracts.found_item.to_array
			else
				Result := <<>>
			end
		ensure
			non_void_result: Result /= Void
		end

	portfolio_positions_with_drilldown (a_portfolio: RP_PORTFOLIO): like portfolio_positions is
			-- Positions for `a_portfolio' with composite assets drill-down
			-- NOTE: side effect is that `clear_caches' is called
		require
			a_portfolio_not_void: a_portfolio /= Void
		local
			l_seen: DS_LINKED_STACK [RP_PORTFOLIO]
			l_loaded: DS_HASH_SET [RP_PORTFOLIO]
			l_acct: BOOLEAN
		do
			l_acct := is_account_level_aggregation_enabled
			clear_caches
			enable_account_level_aggregation
			create Result.make_default
			create l_seen.make_equal
			create l_loaded.make_equal (0)
			recursive_positions (Result, l_loaded, l_seen, create {ROSE_DATE}.make_min_date, create {ROSE_DATE}.make_max_date, one, a_portfolio)
			clear_caches
			if l_acct then
				enable_account_level_aggregation
			else
				disable_account_level_aggregation
			end
		ensure
			positions_not_void: Result /= Void
		end

feature {NONE} -- Implementation

	one: RP_DECIMAL is
			-- One
		once
			create Result.make_undefined
			Result := Result.one
		ensure
			result_not_void: Result /= Void
			is_one: Result ~ Result.one
		end

	portfolio_contracts: DS_HASH_TABLE [DS_LINKED_LIST [RP_PORTFOLIO_CONTRACT], RP_PORTFOLIO]
			-- Cached contracts by portfolio

	account_contracts: DS_HASH_TABLE [DS_LINKED_LIST [RP_PORTFOLIO_CONTRACT], RP_INVESTMENT_ACCOUNT]
			-- Cached contracts by account

	delete_contract (a_contract: RP_PORTFOLIO_CONTRACT) is
			-- Delete `a_contract'
		do
		end

	post_changed_contract (a_contract: RP_PORTFOLIO_CONTRACT) is
			-- Post contract change
		do
		end

	post_new_contract (a_contract: RP_PORTFOLIO_CONTRACT) is
			-- Post new contract
		require
			non_void_contract: a_contract /= Void
		local
			l_contracts: DS_LINKED_LIST [RP_PORTFOLIO_CONTRACT]
		do
			portfolio_contracts.search (a_contract.portfolio)
			if portfolio_contracts.found then
				l_contracts := portfolio_contracts.found_item
			else
				create l_contracts.make_default
				portfolio_contracts.force (l_contracts, a_contract.portfolio)
			end
			l_contracts.force_last (a_contract)
			if is_account_level_aggregation_enabled then
				account_contracts.search (a_contract.portfolio.account)
				if account_contracts.found then
					l_contracts := account_contracts.found_item
				else
					create l_contracts.make_default
					account_contracts.force (l_contracts, a_contract.portfolio.account)
				end
				l_contracts.force_last (a_contract)
			end
			a_contract.flows.do_all (agent add_flow (a_contract.portfolio, ?))
				-- Forward for cost computation
			if is_forwarding_contracts_to_cost_accounting_system then
				manager.cost_accounting_system.process (a_contract)
			end
		end

	add_flow (a_portfolio: RP_PORTFOLIO; a_flow: RP_FLOW) is
			-- Add `a_flow' for `a_portfolio'
		require
			a_portfolio_not_void: a_portfolio /= Void
			a_flow_not_void: a_flow /= Void
		local
			l_oflows: ROSE_LINEAR_ARRAY [RP_FLOW]
			l_custody: RP_CUSTODY
		do
			if trace then
				if trace_date = Void or else a_flow.contracted_ownership_time.ccyymmdd = trace_date.ccyymmdd then
					manager.log_debug (a_flow.debug_output)
				end
			end
			if a_flow.is_portfolio_flow (a_portfolio) then
				if a_flow.is_portfolio_credit_flow (a_portfolio) then
					l_custody := a_flow.destination_control.custody
				else
					l_custody := a_flow.source_control.custody
				end
				if flows_cache.has (a_portfolio) then
					l_oflows := flows_cache.item (a_portfolio)
				else
					create l_oflows.make_empty
					flows_cache.force (l_oflows, a_portfolio)
				end
				l_oflows.extend (a_flow)
				if account_flows_cache.has (a_portfolio.account) then
					l_oflows := account_flows_cache.item (a_portfolio.account)
				else
					create l_oflows.make_empty
					account_flows_cache.force (l_oflows, a_portfolio.account)
				end
				l_oflows.extend (a_flow)
				aggregate_flow (portfolio_holdings, a_portfolio, a_flow.is_portfolio_debit_flow (a_portfolio), a_flow, trade_date_agent)
				if is_account_level_aggregation_enabled then
					aggregate_flow (account_holdings, a_portfolio.account, a_flow.is_portfolio_debit_flow (a_portfolio), a_flow, trade_date_agent)
				end
				aggregate_flow (portfolio_settled_holdings, a_portfolio, a_flow.is_portfolio_debit_flow (a_portfolio), a_flow, settle_date_agent)
				if is_account_level_aggregation_enabled then
					aggregate_flow (account_settled_holdings, a_portfolio.account, a_flow.is_portfolio_debit_flow (a_portfolio), a_flow, settle_date_agent)
				end
				aggregate_flow (portfolio_holdings_per_custody, [a_portfolio, l_custody], a_flow.is_portfolio_debit_flow (a_portfolio), a_flow, trade_date_agent)
				aggregate_flow (portfolio_settled_holdings_per_custody, [a_portfolio, l_custody], a_flow.is_portfolio_debit_flow (a_portfolio), a_flow, settle_date_agent)
				aggregate_flow (investment_account_settled_holdings_per_custody, [a_portfolio.account, l_custody], a_flow.is_portfolio_debit_flow (a_portfolio), a_flow, settle_date_agent)
				update_positions (long_side_pp, a_flow, trade_date_agent)
			end
		end

	aggregate_flow (a_holdings: like portfolio_holdings; a_key: HASHABLE; a_is_debit: BOOLEAN; a_flow: RP_FLOW; a_date_agent: FUNCTION [ANY, TUPLE [RP_FLOW], ROSE_DATE]) is
			-- Aggregate `a_flow'
		require
			a_holdings_not_void: a_holdings /= Void
			a_flow_not_void: a_flow /= Void
			no_void_holding: not a_holdings.has_void_item
		local
			l_holdings: RP_HOLDINGS
		do
			has_aggregated_flows := True
			a_holdings.search (a_key)
			if a_holdings.found then

--			if a_holdings.has (a_key) then
--				l_holdings := a_holdings.item (a_key)
				l_holdings := a_holdings.found_item
				if l_holdings = VOid then
					do_nothing
				end
			else
				create l_holdings.make (a_date_agent)
				a_holdings.force (l_holdings, a_key)
				if l_holdings = VOid then
					do_nothing
				end
			end
			l_holdings.add_account_flow (a_flow, a_is_debit)
		ensure
			has_aggregated_flows: has_aggregated_flows
			has_key: a_holdings.has (a_key)
		end

	long_side_pp: DS_HASH_TABLE [UID_TABLE [RP_POSITIONS, RP_INSTRUMENT], RP_PORTFOLIO]
			-- Long side portfolio positions

	update_positions (a_table: DS_HASH_TABLE [UID_TABLE [RP_POSITIONS, RP_INSTRUMENT], RP_PORTFOLIO]; a_flow: RP_FLOW; a_date: FUNCTION [ANY, TUPLE [RP_FLOW], ROSE_DATE]) is
			-- Update positions `a_table' with `a_flow' on `a_date'
		require
			table_not_void: a_table /= Void
			non_void_flow: a_flow /= Void
			date_not_void: a_date /= Void
		local
			l_per_instrument: UID_TABLE [RP_POSITIONS, RP_INSTRUMENT]
			l_positions: RP_POSITIONS
		do
			a_table.search (a_flow.contract.portfolio)
			if a_table.found then
				l_per_instrument := a_table.found_item
			else
				create l_per_instrument.make_equal (17)
				a_table.force (l_per_instrument, a_flow.contract.portfolio)
			end
			l_per_instrument.search (a_flow.instrument)
			if l_per_instrument.found then
				l_positions := l_per_instrument.found_item
			else
				create l_positions.make (a_flow.instrument)
				l_per_instrument.force (l_positions, a_flow.instrument)
			end
			l_positions.adjust_with (a_date.item ([a_flow]), a_flow)
		end

	flows_cache: DS_HASH_TABLE [ROSE_LINEAR_ARRAY [RP_FLOW], RP_PORTFOLIO]
			-- Cached flows per portfolio

	account_flows_cache: DS_HASH_TABLE [ROSE_LINEAR_ARRAY [RP_FLOW], RP_INVESTMENT_ACCOUNT]
			-- Cached flows per account

	portfolio_holdings: DS_HASH_TABLE [RP_HOLDINGS, HASHABLE]
			-- Cached traded portfolio holdings

	portfolio_settled_holdings: DS_HASH_TABLE [RP_HOLDINGS, HASHABLE]
			-- Cached settled portfolio holdings

	portfolio_holdings_per_custody: DS_HASH_TABLE [RP_HOLDINGS, HASHABLE]
			-- Cached traded portfolio holdings keyed by custody and portfolio

	portfolio_settled_holdings_per_custody: DS_HASH_TABLE [RP_HOLDINGS, HASHABLE]
			-- Cached settled portfolio holdings keyed by custody and portfolio

	investment_account_settled_holdings_per_custody: DS_HASH_TABLE [RP_HOLDINGS, HASHABLE]
			-- Cached settled portfolio holdings keyed by custody and account

	account_holdings: DS_HASH_TABLE [RP_HOLDINGS, HASHABLE]
			-- Cache traded account holdings

	account_settled_holdings: DS_HASH_TABLE [RP_HOLDINGS, HASHABLE]
			-- Cache settled account holdings

	add_position (a_history: RP_INSTRUMENT_HISTORY; a_holding: RP_HOLDINGS; a_instrument: RP_INSTRUMENT) is
			-- Add positions in `a_instrument' to `a_history'
		local
			l_history: RP_HOLDING_HISTORY
		do
			l_history := a_holding.holdings_for (a_instrument)
			l_history.all_intervals.do_all (agent add_portfolio_position_interval (a_history, a_instrument, ?))
		end

	add_portfolio_position_interval (a_history: RP_INSTRUMENT_HISTORY; a_instrument: RP_INSTRUMENT; a_interval: RP_HOLDING_INTERVAL) is
			-- Add instrument history interval to `a_history'
		local
			l_interval: RP_INSTRUMENT_HISTORY_INTERVAL
		do
			create l_interval.make (a_instrument, a_interval.start_date, a_interval.end_date, a_interval.quantity.amount)
			a_history.append (l_interval)
		end

	addp (a_history: RP_INSTRUMENT_HISTORY; a_positions_and_instrument: TUPLE [positions: RP_POSITIONS; instrument: RP_INSTRUMENT]) is
			-- Add portfolio position
		do
			a_positions_and_instrument.positions.positions.do_all (agent addpp (a_history, ?))
		end

	addpp (a_history: RP_INSTRUMENT_HISTORY; a_position: RP_POSITION) is
			-- Add portfolio position
		do
			a_history.append (create {RP_INSTRUMENT_HISTORY_INTERVAL}.make (a_position.instrument, a_position.start_date, a_position.end_date, a_position.position))
		end

	recursive_positions (a_history: RP_INSTRUMENT_HISTORY; a_loaded: DS_HASH_SET [RP_PORTFOLIO]; a_seen: DS_STACK [RP_PORTFOLIO]; a_start_date, a_end_date: ROSE_DATE; a_multiplier: RP_DECIMAL; a_portfolio: RP_PORTFOLIO) is
			-- Compute recursively positions for `a_portfolio' and store in `a_history'.
		require
			a_history_not_void: a_history /= Void
			a_seen_not_void: a_seen /= Void
			a_start_date_not_void: a_start_date /= Void
			a_end_date_not_void: a_end_date /= Void
			a_start_date_lr_a_end_date: a_start_date < a_end_date
			a_multiplier_not_void: a_multiplier /= Void
			a_portfolio_not_void: a_portfolio /= Void
		local
			l_positions: like portfolio_positions
			l_composite_assets: ROSE_LINEAR_ARRAY [RP_INSTRUMENT_HISTORY_INTERVAL]
		do
			if a_seen.has (a_portfolio) then
				(create {EXCEPTIONS}).raise ("Cyclical cross-holdings graph for `" + a_portfolio.code + "'")
			end
			a_seen.put (a_portfolio)
			if not a_loaded.has (a_portfolio) then
				if a_portfolio.zone.is_transzonal then
					manager.legacy.update_for_investment_account (a_portfolio.account)
				else
					manager.legacy.update_for_portfolio (a_portfolio)
				end
				a_loaded.force (a_portfolio)
			end
			if a_portfolio.zone.is_transzonal then
				l_positions := investment_account_positions (a_portfolio.account)
			else
				l_positions := portfolio_positions (a_portfolio)
			end
			l_positions.do_all (agent add_position_with_multiplier (a_history, a_start_date, a_end_date, a_multiplier, ?))
				-- find coposite assets posittions
			create l_composite_assets.make_empty
			a_history.do_if (agent l_composite_assets.extend (?), agent is_composite_asset_position (?))
				-- remove first
			l_composite_assets.do_all (agent a_history.delete (?))
				-- then drill down
			l_composite_assets.do_all (agent drilldown_composite_asset (a_history, a_loaded, a_seen, a_multiplier, ?))
			a_seen.remove
		end

	is_composite_asset_position (a_position: RP_INSTRUMENT_HISTORY_INTERVAL): BOOLEAN is
			-- Is `a_position' in composite asset which is ours?
		require
			a_position_not_void: a_position /= Void
		do
			if {l_fund: RP_POOL_FUND} a_position.instrument then
				Result := l_fund.portfolio /= Void
			end
		end

	drilldown_composite_asset (a_history: RP_INSTRUMENT_HISTORY; a_loaded: DS_HASH_SET [RP_PORTFOLIO]; a_seen: DS_STACK [RP_PORTFOLIO]; a_multiplier: RP_DECIMAL; a_position: RP_INSTRUMENT_HISTORY_INTERVAL) is
			-- Expand composite asset `a_position' into its components multiply by `a_multiplier' and then add to `a_history'
			-- Use `a_seen' to detect circular cross-holdings and `a_loaded' to keep track of available accounting data
		require
			a_history_not_void: a_history /= Void
			a_loaded_not_void: a_loaded /= Void
			a_seen_not_void: a_seen /= Void
			a_multiplier_not_void: a_multiplier /= Void
			a_position_not_void: a_position /= Void
			a_position_is_axar_composite_asset: is_composite_asset_position (a_position)
		local
			l_multiplier: RP_DECIMAL
			l_position: RP_DECIMAL
			l_start: TUPLE [date: ROSE_DATE; shares: RP_DECIMAL]
			l_end: TUPLE [date: ROSE_DATE; shares: RP_DECIMAL]
			l_start_date, l_end_date: ROSE_DATE
			l_should_stop: BOOLEAN
		do
			if {l_fund: RP_POOL_FUND} a_position.instrument then
				if l_fund.portfolio /= Void then
					l_position := a_position.weight
					if not l_position.is_zero then
						from
								-- iterate over changes in oustanding shares
								-- of the underlying fund
							l_end_date := a_position.end_date
							if l_end_date.is_max_date then
								l_end := outstanding_shares_lookup.outstanding_shares (l_fund, create {ROSE_DATE}.make_today)
							else
								l_end := outstanding_shares_lookup.outstanding_shares (l_fund, l_end_date)
							end
							if l_end = Void then
								(create {EXCEPTIONS}).raise ("Can't find oustanding shares for `" + l_fund.brimsym + "' on or before " + a_position.end_date.ccyymmdd.out)
							end
							if l_end.date <= a_position.start_date then
								l_start := l_end
							else
								l_start := outstanding_shares_lookup.previous_change_to_outstanding_shares (l_fund, l_end.date, l_end.shares)
								if l_start = Void then
									(create {EXCEPTIONS}).raise ("Can't find oustanding shares for `" + l_fund.brimsym + "' on or before " + l_end.date.previous_date.ccyymmdd.out)
								end
							end
						until
							l_should_stop
						loop
							if l_end.shares.is_zero or not l_end.shares.is_defined then
								(create {EXCEPTIONS}).raise ("Invalid oustanding shares for `" + l_fund.brimsym + "' on or before " + l_end.date.ccyymmdd.out)
							end
							l_start_date := a_position.start_date.max (l_start.date.next_date)
							l_multiplier := a_multiplier * a_position.weight / l_end.shares
							recursive_positions (a_history, a_loaded, a_seen, l_start_date, l_end_date, l_multiplier, l_fund.portfolio)
							if l_start_date <= a_position.start_date then
								l_should_stop := True
							else
								l_end_date := l_start_date
								l_end := l_start
								l_start := outstanding_shares_lookup.previous_change_to_outstanding_shares (l_fund, l_end_date, l_end.shares)
								if l_start = Void then
									(create {EXCEPTIONS}).raise ("Can't find oustanding shares for `" + l_fund.brimsym + "' on or before " + l_end.date.previous_date.ccyymmdd.out)
								end
							end
						end
					end
				end
			end
		end

	outstanding_shares_lookup: RP_POOL_FUND_OUTSTANDING_SHARES_LOOKUP is
			-- Poolfund outstanding shares lookup
		once
			create Result.make (manager.reference_system)
		ensure
			result_not_void: Result /= Void
		end

	add_position_with_multiplier (a_history: RP_INSTRUMENT_HISTORY; a_start_date, a_end_date: ROSE_DATE; a_multiplier: RP_DECIMAL; a_position: RP_INSTRUMENT_HISTORY_INTERVAL) is
			-- Add `a_position' times `a_multiplier' to `a_history'
		require
			a_history_not_void: a_history /= Void
			a_start_date_not_void: a_start_date /= Void
			a_end_date_not_void: a_end_date /= Void
			a_start_date_lr_a_end_date: a_start_date < a_end_date
			a_multiplier_not_void: a_multiplier /= Void
			a_position_not_void: a_position /= Void
		local
			l_intervals: DS_LINKED_LIST [RP_INSTRUMENT_HISTORY_INTERVAL]
			l_remainder: CELL [RP_INSTRUMENT_HISTORY_INTERVAL]
			l_start_date, l_end_date: ROSE_DATE
		do
			if not a_position.weight.is_zero then
				l_start_date := a_position.start_date.max (a_start_date)
				l_end_date := a_position.end_date.min (a_end_date)
				if l_start_date < l_end_date then
					create l_remainder.put (create {RP_INSTRUMENT_HISTORY_INTERVAL}.make (a_position.instrument, l_start_date, l_end_date, a_multiplier * a_position.weight))
					create l_intervals.make_default
					a_history.entity_intervals (a_position.instrument).do_if (agent l_intervals.force_last (?), agent has_intersection (l_remainder.item, ?))
					if l_intervals.is_empty then
							-- no previous positions in this instrument
						a_history.append (l_remainder.item)
					else
						l_intervals.do_until (agent intersect (a_history, ?, l_remainder), agent item_is_void (?, l_remainder))
						if l_remainder.item /= Void then
								-- inser non-intersecting remainder
							a_history.append (l_remainder.item)
						end
					end
				end
			end
		end

	item_is_void (a_interval: RP_INSTRUMENT_HISTORY_INTERVAL; a_cell: CELL [RP_INSTRUMENT_HISTORY_INTERVAL]): BOOLEAN is
			-- Is `a_cell.item' Void?
			-- NOTE: `a_interval' is ignored
		require
			a_cell_not_void: a_cell /= Void
		do
			Result := a_cell.item = Void
		ensure
			definition: Result = (a_cell.item = Void)
		end


	intersect (a_history: RP_INSTRUMENT_HISTORY; a_interval: RP_INSTRUMENT_HISTORY_INTERVAL; a_position: CELL [RP_INSTRUMENT_HISTORY_INTERVAL]) is
			-- Store intersection of `a_interval' with `a_position' in `a_history'.
		require
			a_history_not_void: a_history /= Void
			a_interval_not_void: a_interval /= Void
			a_position_not_void: a_position /= Void
			intersection: a_interval.brackets (a_position.item.start_date) or a_position.item.brackets (a_interval.start_date)
			same_instrument: a_interval.instrument ~ a_position.item.instrument
		local
			l_start_date, l_end_date: ROSE_DATE
			l_interval: RP_INSTRUMENT_HISTORY_INTERVAL
			l_remainder: RP_INSTRUMENT_HISTORY_INTERVAL
		do
			a_history.delete (a_interval)
			l_start_date := a_interval.start_date.max (a_position.item.start_date)
			l_end_date := a_interval.end_date.min (a_position.item.end_date)
			if a_position.item.start_date < l_start_date then
					-- head of a_position
					-- add to a_history because a_interval(s) are sorted by start date
					-- which means that a_position starts befoe any a_interval
				create l_interval.make (a_interval.instrument, a_position.item.start_date, l_start_date, a_position.item.weight)
				a_history.append (l_interval)
			end
			if a_interval.start_date < l_start_date then
					-- head of a_interval
				create l_interval.make (a_interval.instrument, a_interval.start_date, l_start_date, a_interval.weight)
				a_history.append (l_interval)
			end
				-- intersection
			create l_interval.make (a_interval.instrument, l_start_date, l_end_date, a_interval.weight + a_position.item.weight)
			a_history.append (l_interval)
			if l_end_date < a_position.item.end_date then
					-- tail of a_position
				create l_remainder.make (a_interval.instrument, l_end_date, a_position.item.end_date, a_position.item.weight)
			end
			if l_end_date < a_interval.end_date then
					-- tail of a_interval
				create l_interval.make (a_interval.instrument, l_end_date, a_interval.end_date, a_interval.weight)
				a_history.append (l_interval)
			end
			a_position.put (l_remainder)
		end

	has_intersection (a_interval1, a_interval2: RP_INSTRUMENT_HISTORY_INTERVAL): BOOLEAN is
			-- Is thre intersection between `a_interval1' and `a_interval2'?
		require
			a_interval1_not_void: a_interval1 /= Void
			a_interval2_not_void: a_interval2 /= Void
		do
			Result := a_interval1.brackets (a_interval2.start_date) or a_interval2.brackets (a_interval1.start_date)
		end

	compare_portfolio_custody (a_pc1, a_pc2: TUPLE [portfolio:RP_PORTFOLIO; custody:RP_CUSTODY]): BOOLEAN is
			-- Are the 2 tuples equal?
		require
			a_pc1_not_void: a_pc1 /= Void
			a_pc2_not_void: a_pc2 /= Void
		do
			Result := a_pc1.portfolio ~ a_pc2.portfolio and a_pc1.custody ~ a_pc2.custody
		end

invariant

	portfolio_contracts_not_void: portfolio_contracts /= Void
	no_void_contract_key: not portfolio_contracts.has_void
	no_void_contracts: not portfolio_contracts.has_void_item
	no_void_contract: not portfolio_contracts.there_exists (agent {DS_LINKED_LIST [RP_PORTFOLIO_CONTRACT]}.has (Void))
	portfolio_holdings_not_void: portfolio_holdings /= Void
	no_void_holding_key: not portfolio_holdings.has_void
	no_void_holding: not portfolio_holdings.has_void_item
	portfolio_holdings_per_custody_not_void: portfolio_holdings_per_custody /= Void
	no_void_holding_key_per_custody: not portfolio_holdings_per_custody.has_void
	no_void_holding_per_custody: not portfolio_holdings_per_custody.has_void_item
	portfolio_settled_holdings_not_void: portfolio_settled_holdings /= Void
	no_void_settled_holding_key: not portfolio_settled_holdings.has_void
	no_void_settled_holding: not portfolio_settled_holdings.has_void_item
	portfolio_settled_holdings_per_custody_not_void: portfolio_settled_holdings_per_custody /= Void
	no_void_settled_holding_key_per_custody: not portfolio_settled_holdings_per_custody.has_void
	no_void_settled_holding_per_custody: not portfolio_settled_holdings_per_custody.has_void_item
	flows_cache_not_void: flows_cache /= Void
	no_void_flows_cache_key: not flows_cache.has_void
	no_void_flows_cache: not flows_cache.has_void_item
	no_void_flow: not flows_cache.there_exists (agent {ROSE_LINEAR_ARRAY [RP_FLOW]}.has (Void))
	account_flows_cache_not_void: account_flows_cache /= Void
	no_void_account_flows_cache_key: not account_flows_cache.has_void
	no_void_account_flows_cache: not account_flows_cache.has_void_item
	no_void_account_flow: not account_flows_cache.there_exists (agent {ROSE_LINEAR_ARRAY [RP_FLOW]}.has (Void))
	account_holdings_not_void: account_holdings /= Void
	no_void_account_holdings_key: not account_holdings.has_void
	no_void_account_holding: not account_holdings.has_void_item
	account_settled_holdings_not_void: account_settled_holdings /= Void
	no_void_account_settled_holdings_key: not account_settled_holdings.has_void
	no_void_account_settled_holding: not account_settled_holdings.has_void_item

end
