note

	description:

		"BACKDIV record matching with sequential assignment"

	author: "Ivan Tarnavsky"
	last_editor: "$Author: it $"
	date: "$Date: 2011/02/23 $"
	revision: "$Revision: #1 $"
	path: "$File: //rose/source/kernel/expert_system/legacy/rp_backdiv_matcher_with_fifo_assignment.e $"

class RP_BACKDIV_MATCHER_WITH_FIFO_ASSIGNMENT

inherit

	RP_GEDB_EVENT_TO_BACKDIV_MATCHER

	RP_AXAROSENBERG_AWARE
		export {NONE} all end

create {RP_LEGACY_SYSTEM}

	make

feature {NONE} -- Initialization

	make is
			-- Initialize `Current'
		do
			create used_dividends.make
		end

feature -- Query

	match (a_event: RP_GEDB_EVENT): RP_BACKDIV_RECORD is
			-- <Precursor>
		do
			if attached {RP_STOCK} a_event.instrument as l_stock then
				if a_event.event_date /= Void and a_event.currency /= Void and a_event.performance_effect_with_sign /= Void and a_event.cash_effect_with_sign /= Void and a_event.units_base /= Void and then not a_event.units_base.amount.is_zero then
					Result := find_matching_backdiv (a_event, l_stock)
				end
			end
		end

feature {NONE} -- Implementation

	used_dividends: RP_BACKDIV_RECORD_CACHE
			-- Keeps track of backdiv records used for on a given date

	find_matching_backdiv (a_event: RP_GEDB_EVENT; a_stock: RP_STOCK): RP_BACKDIV_RECORD is
			-- Matching backdiv record
		require
			a_event_not_void: a_event /= Void
			a_stock_not_void: a_stock /= Void
			event_date_not_void: a_event.event_date /= Void
			currency_not_void: a_event.currency /= Void
			units_base_not_void: a_event.units_base /= Void
			units_base_not_zero: not a_event.units_base.is_zero
			perf_effect_not_void: a_event.performance_effect_with_sign /= Void
			cash_effect_not_void: a_event.cash_effect_with_sign /= Void
		local
			l_cash_rate: RP_DECIMAL
			l_dividend_rec_list: ARRAY [attached PAIR [attached RP_BACKDIV_RECORD, BOOLEAN]]
			l_dividend_index: INTEGER
			l_source: RP_CAE_DATA_SOURCE
			l_currency: RP_CURRENCY
			l_exdate: ROSE_DATE
		do
			l_exdate := a_event.event_date
			l_currency := a_event.currency
			l_source := manager.corporate_action_system.data_sources (a_stock, l_exdate).item (1)
			used_dividends.initialize (l_exdate)
			if not used_dividends.has (a_stock) then
				used_dividends.put (dividends (l_source, a_stock, l_exdate), a_stock)
			end
			l_dividend_rec_list := used_dividends.item (a_stock)
			if not l_dividend_rec_list.is_empty then
				if a_event.performance_effect_with_sign.amount.is_zero then
					l_cash_rate := a_event.cash_effect_with_sign.amount / a_event.units_base.amount
				else
					l_cash_rate := a_event.performance_effect_with_sign.amount / a_event.units_base.amount
				end
				if l_dividend_rec_list.count = 1 then
					Result := tolerance_check (l_dividend_rec_list.item (1).first, l_cash_rate, l_currency, l_exdate)
				else
					l_dividend_index := find_best_record (l_dividend_rec_list, l_cash_rate, l_currency)
					if l_dividend_index > 0 then
						Result := tolerance_check (l_dividend_rec_list.item (l_dividend_index).first, l_cash_rate, l_currency, l_exdate)
						if Result /= Void and then Result.ex_date ~ l_exdate then
							l_dividend_rec_list.item (l_dividend_index).set_second (True)
						end
					end
				end
			end
		end

	find_best_record (a_dividend_rec_list: ARRAY [attached PAIR [attached RP_BACKDIV_RECORD, BOOLEAN]]; a_cash_rate: RP_DECIMAL; a_currency: RP_CURRENCY): INTEGER is
			-- Returns index of record from `a_dividend_rec_list' which best matches `a_cash_rate'
		require
			a_dividend_rec_list_not_void: a_dividend_rec_list /= Void
			a_cash_rate_not_void: a_cash_rate /= Void
			a_currency_not_void: a_currency /= Void
		local
			l_rate_difference: RP_DECIMAL
			l_backdiv_cash_rate: RP_DECIMAL
			l_rp_backdiv: RP_BACKDIV_RECORD
			i: INTEGER
		do
			from
				create l_rate_difference.make_from_integer (99999)
				i := 1
			until
				i > a_dividend_rec_list.count
			loop
				if not a_dividend_rec_list.item (i).second then
					l_rp_backdiv := a_dividend_rec_list.item (i).first
					if not l_rp_backdiv.cash_rate.is_zero then
						l_backdiv_cash_rate := cashrate (a_currency, l_rp_backdiv)
						if l_rate_difference > (a_cash_rate - l_backdiv_cash_rate).abs then
							l_rate_difference := (a_cash_rate - l_backdiv_cash_rate).abs
							Result := i
						end
					end
				end
				i := i + 1
			end
		end

	tolerance_check (a_backdiv_record: RP_BACKDIV_RECORD; a_cash_rate: RP_DECIMAL; a_currency: RP_CURRENCY; a_exdate: ROSE_DATE): RP_BACKDIV_RECORD is
			-- Check `a_backdiv_record' for tolerance if not on `a_exdate'
			-- Return Void if record fails tolerance check
		require
			a_backdiv_record_not_void: a_backdiv_record /= Void
			a_cash_rate_not_void: a_cash_rate /= Void
			a_currency_not_void: a_currency /= Void
			a_exdate_not_void: a_exdate /= Void
		local
			l_backdiv_cash_rate: RP_DECIMAL
		do
			Result := a_backdiv_record
			if Result.ex_date /~ a_exdate then
				l_backdiv_cash_rate := cashrate (a_currency, Result)
				if (a_cash_rate - l_backdiv_cash_rate).abs > l_backdiv_cash_rate * fifteen_percent_tolerance then
					Result := Void
				end
			end
		end

	dividends (a_source: RP_CAE_DATA_SOURCE; a_stock: RP_STOCK; a_date: ROSE_DATE): ARRAY [RP_CORPORATE_ACTION] is
			-- Look back dividends on `a_source' for `a_stock' starting with `a_date'
		require
			a_source_not_void: a_source /= Void
			a_stock_not_void: a_stock /= Void
			a_date_not_void: a_date /= Void
		local
			l_date: ROSE_DATE
		do
			from
				create l_date.make_ccyymmdd (a_date.ccyymmdd)
				Result := a_source.instrument_action_and_events (a_stock, l_date)
				l_date := l_date.previous_date
			until
				l_date.diff_in_days (a_date).abs > look_back_range or Result /= Void
			loop
				Result := a_source.instrument_action_and_events (a_stock, l_date)
				l_date := l_date.previous_date
			end
			if Result = Void then
				Result := <<>>
			end
		ensure
			dividends_not_void: Result /= Void
		end

	cashrate (a_currency: RP_CURRENCY; a_backdiv: RP_BACKDIV_RECORD): RP_DECIMAL is
			-- Cashrate for given currency
		require
			a_currency_not_void: a_currency /= Void
			a_backdiv_not_void: a_backdiv /= Void
			a_backdiv_cash_rate_not_void: a_backdiv.cash_rate /= Void
			a_backdiv_currency_not_void: a_backdiv.currency /= Void
		local
			l_fx_rate: RP_DECIMAL
			l_ccy: RP_CURRENCY
		do
			Result := a_backdiv.cash_rate
			l_ccy := a_backdiv.currency
			if not l_ccy.is_same (a_currency) then
				l_fx_rate := a_currency.currency.market_exchange_rate (fx, l_ccy.currency, a_backdiv.ex_date)
				Result := Result * l_fx_rate
			end
		ensure
			cashrate_not_void: Result /= Void
		end

	fx: RP_LEGACY_PERF_FX is
			-- Legacy perf fx
		once
			create Result
		ensure
			fx_not_void: Result /= Void
		end

feature {NONE} -- Constants

	look_back_range: INTEGER is 5
			-- Number of look back days used in `dividends'

	fifteen_percent_tolerance: RP_DECIMAL is
			-- 15% tolerance level
		once
			create Result.make_from_string ("0.15")
		ensure
			fifteen_percent_tolerance_not_void: Result /= Void
		end

invariant

	used_dividends_not_void: used_dividends /= Void

end
