note

	description:

		"Market Data Source for CLIO pool funds Database"

	author: "Norman Wolf"
	last_editor: "$Author: mg $"
	date: "$Date: 2013/10/02 $"
	revision: "$Revision: #3 $"
	path: "$File: //rose/source/kernel/expert_system/mds/rp_clio_pool_fund_market_data_source.e $"

class RP_CLIO_POOL_FUND_MARKET_DATA_SOURCE

inherit

	RP_POOL_FUND_DATA_SOURCE

create

	make,
	make_with_lookback,
	make_with_restriction

feature {NONE} -- Implementation

	cache: DS_HASH_TABLE [DS_HASH_TABLE [RP_MARKET_DATA_RECORD, ROSE_DATE], STRING]
			-- Cached market data by brimsec then by date

	pool_fund_data (a_fund: RP_POOL_FUND; a_date: ROSE_DATE): RP_MARKET_DATA_RECORD is
			-- Market data for `a_fund' on `a_date'
		local
			l_brimsec, l_brimco, l_sec: STRING
			l_data: DS_HASH_TABLE [RP_MARKET_DATA_RECORD, ROSE_DATE]
		do
			l_brimsec := a_fund.brimsym.substring (1, 15)
			l_brimco := l_brimsec.substring (1, 9)
			l_sec := l_brimsec.substring (10, 15)
				-- Look for brimsec in CLIO, if not there means CLIO does not provide data for this fund (not an error)
			if clio_mapping_history.has (l_brimsec) then
				if cache = Void then
					create cache.make_equal (0)
				end
				cache.search (l_brimsec)
				if cache.found then
					l_data := cache.found_item
				else
					create l_data.make_default
					cache.force (l_data, l_brimsec)
					clio_query.retrieve (Void, <<l_brimsec, start_date, end_date>>)
					clio_query.array.do_all (agent add_record (l_data, ?))
				end
				l_data.search (a_date)
				if l_data.found then
					Result := l_data.found_item
				end
			end
		end

	add_record (a_data: DS_HASH_TABLE [RP_MARKET_DATA_RECORD, ROSE_DATE]; a_record: RP_POOL_FUND_CLASS_DATA_CLIODB_RECORD) is
			-- Add data from `a_record' to `a_data'
		require
			a_data_not_void: a_data /= Void
			a_record_not_void: a_record /= Void
		local
			l_rec: RP_POOL_FUND_CLASS_DATA
		do
			create l_rec.make_from_cliodb_record (reference_system, a_record)
			a_data.force (l_rec, l_rec.date)
		ensure
			record_added: a_data.has (a_record.valuationdate)
		end

	retrieve_clio_mapping_data is
			-- Retrieve brimsec to clio fund id mapping data from CLIO db
		require
			clio_mapping_cache_attached: clio_mapping_cache /= Void
		local
			l_query: COLLECTING_PREDICATE [RP_CLIODB_BRIMSEC_MAPPING_RECORD]
		do
			create l_query.make (clio_server, clio_mapping_query_template)
			l_query.apply (agent add_mapping (clio_mapping_history, ?), Void, Void)
		end

	clio_server: DATA_SERVER is
			-- CLIO db
		once
			Result := (create {DATABASE_ACCESS}).make_from_repository_fast_load ("ucits_marketing")
		ensure
			result_not_void: Result /= Void
		end

	clio_query: COLLECTING_PREDICATE [RP_POOL_FUND_CLASS_DATA_CLIODB_RECORD] is
			-- CLIO data query
		once
			create Result.make (clio_server, query_template)
		ensure
			result_not_void: Result /= Void
		end

	add_mapping (a_hash: like clio_mapping_history; a_record: RP_CLIODB_BRIMSEC_MAPPING_RECORD) is
			-- Cache clio mapping data.
		require
			a_hash_not_void: a_hash /= Void
			a_record_not_void: a_record /= Void
		local
			l_per_brimsec: ROSE_INTERVAL_HISTORY [INTEGER, ROSE_CLOSED_OPEN_INTERVAL [INTEGER]]
			l_interval: ROSE_CLOSED_OPEN_INTERVAL [INTEGER]
			l_brimsec: STRING
			l_start, l_end: ROSE_DATE
		do
			create l_start.make_ccyymmdd (a_record.start_date)
			create l_end.make_ccyymmdd (a_record.end_date)
			if a_record.brimco /= Void and a_record.sec /= Void then
				l_brimsec := a_record.brimco + a_record.sec
				create l_interval.make (a_record.clio_fund_id, l_start, l_end, 1.0)
				a_hash.search (l_brimsec)
				if a_hash.found then
					l_per_brimsec := a_hash.found_item
				else
					create l_per_brimsec.make_empty
					a_hash.force (l_per_brimsec, l_brimsec)
				end
				l_per_brimsec.extend (l_interval)
			end
		end

	clio_mapping_history: DS_HASH_TABLE [ROSE_INTERVAL_HISTORY [INTEGER, ROSE_CLOSED_OPEN_INTERVAL [INTEGER]], STRING] is
			-- clio fund id to brimsec mapping intervals
		do
			if clio_mapping_cache = Void then
				create clio_mapping_cache.make_equal (0)
				retrieve_clio_mapping_data
			end
			Result := clio_mapping_cache
		ensure
			result_not_void: Result /= Void
		end

	clio_mapping_cache: like clio_mapping_history
			-- Cached clio mapping data.

	clio_mapping_query_template: STRING is
			-- Data query template for brimsec to CLIO fund id mapping records
		once
			Result := "[
 SELECT
 	[clio_fund_id],
    [brimco],
    [sec],
    [start_date],
    [end_date]
  FROM [clio_fund_id_to_brim]
  where [clio_fund_id] not in (2264)
  ORDER BY brimco, sec, start_date, end_date
]"
				-- excludes clio fund 2264 since it is in USD and we only care about JPY quoted share class for this fund
				-- may be able to join with CLIO table and fund brimsym "exchange" which is usally currency for funds (but not always?)
		ensure
			result_not_void: Result /= Void
		end

	query_template: STRING is
			-- Data query template
		once
			Result := "[
SELECT 
	[valuationdate],
	[quotecurrency],
	[cliofundid],
	[navperunit],
	[unitsinissue],
	[totalnavperfund],
	[updatedon]
FROM [CLIO] c
JOIN [clio_fund_id_to_brim] b ON c.[CLIOFUNDID] = b.[clio_fund_id]
	AND c.[VALUATIONDATE] >= CONVERT(CHAR(8), b.[start_date])
	AND c.[VALUATIONDATE] < CONVERT(CHAR(8), CASE b.[end_date] WHEN 99999999 THEN 99991231 ELSE b.[end_date] END)
	AND b.[clio_fund_id] not in (2264)
WHERE b.[brimco]+b.[sec] = ? AND c.[VALUATIONDATE] BETWEEN ? AND ?
ORDER BY [valuationdate]
]"
		ensure
			result_not_void: Result /= Void
		end

end
