note

	description:

		"Processor of corporate actions for portfolios"

	author: "Sami Ben Grine"
	last_editor: "$Author: er $"
	date: "$Date: 2014/01/29 $"
	revision: "$Revision: #16 $"
	path: "$File: //rose/source/Kernel/expert_system/rp_corporate_action_processing_system.e $"

class RP_CORPORATE_ACTION_PROCESSING_SYSTEM

inherit
	RP_INVESTMENT_MANAGER_CLIENT

	RP_CONTRACT_SYSTEM

create
	make

feature -- Access

	use_rate_overrides_for_cash_dividends: BOOLEAN
		-- Should the system use some hard coded tax rates to compute the tax withholding?
		-- This is only for a handful of cases when we need to historically compare with event up and the history was lost and not available in parm files

feature -- Change

	set_use_rate_overrides_for_cash_dividends (a_use_rate_overrides_for_cash_dividends: like use_rate_overrides_for_cash_dividends) is
			-- Set 'use_rate_overrides_for_cash_dividends' to 'a_use_rate_overrides_for_cash_dividends'
		do
			use_rate_overrides_for_cash_dividends := a_use_rate_overrides_for_cash_dividends
		ensure
			use_rate_overrides_for_cash_dividends_set: use_rate_overrides_for_cash_dividends = a_use_rate_overrides_for_cash_dividends
		end


feature -- Basic Operations

	apply_corporate_actions_to_portfolio (
		a_portfolio: like portfolio_template;
		a_date: ROSE_DATE;
		a_poster: PROCEDURE[ANY, TUPLE[RP_PORTFOLIO_CONTRACT]];
		a_should_skip_generated_contract:FUNCTION[ANY,TUPLE[ANY], BOOLEAN]) is
			-- generate corporate actions to bring a_portfolio up to date
		local
			l_holdings: ARRAY [RP_HOLDING_INTERVAL]
		do
			l_holdings := manager.accounting.portfolio_holdings_at (a_portfolio, a_date)
			l_holdings.do_all (agent apply_corporate_actions (a_portfolio, a_date, ?, a_poster, a_should_skip_generated_contract))
		end

	apply_corporate_actions (
		a_portfolio: like portfolio_template;
		a_date: ROSE_DATE;
		a_holding: RP_HOLDING_INTERVAL;
		a_poster: PROCEDURE[ANY, TUPLE[RP_PORTFOLIO_CONTRACT]];
		a_should_skip_generated_contract:FUNCTION[ANY,TUPLE[ANY], BOOLEAN]) is
			-- retrieve corporate action for a_holding on a_date and apply
		local
			l_instrument: RP_INSTRUMENT
			l_processing_instructions: ROSE_LINEAR_ARRAY [RP_CORPORATE_ACTION_PROCESSING_INSTRUCTIONS]
		do
			if a_holding.quantity.amount.abs >= 0 then
			-- The rounding actually just on the output not on the share that is used to calcuate dividend
				l_instrument := a_holding.quantity.instrument
				if {ll_security: RP_SECURITY} l_instrument then
					if ll_security.entity_id /= 0 then
						l_processing_instructions := manager.corporate_action_system.processing_instructions (l_instrument, a_date)
						if l_processing_instructions /= Void then
							l_processing_instructions.do_all (agent apply_to_portfolio (a_portfolio, a_holding, l_instrument, ?, a_date, a_poster, a_should_skip_generated_contract))
						end
					end
				end
			end
		end

feature -- Access

	latest_process_date (a_portfolio: like portfolio_template): ROSE_DATE is
			-- Date when 'a_portfolio' was processed for the last time
		require
			portfolio_not_void: a_portfolio /= Void
		do
			Result := latest_process_date_table [a_portfolio.code]
		ensure
			not_void: Result /= Void
		end

feature -- Change

	set_latest_process_date (a_date: ROSE_DATE; a_portfolio: like portfolio_template) is
			-- FIXME: does this make sense??
			-- set latest process date
		do
			latest_process_date_table.force (create {ROSE_DATE}.make_by_date (a_date), a_portfolio.code)
		end

feature {NONE} -- Implementation

	apply_to_portfolio (
		a_portfolio: RP_PORTFOLIO;
		a_original_holding: RP_HOLDING_INTERVAL;
		a_instrument: RP_INSTRUMENT;
		a_processing_instructions: RP_CORPORATE_ACTION_PROCESSING_INSTRUCTIONS; a_idate: ROSE_DATE;
		a_poster: PROCEDURE[ANY, TUPLE[RP_PORTFOLIO_CONTRACT]];
		a_should_skip_generated_contract:FUNCTION[ANY,TUPLE[ANY], BOOLEAN]) is
			-- apply corporate action processing instructions to portfolio holdings
		require
			non_void_portfolio: a_portfolio /= Void
			non_void_actions: a_processing_instructions /= Void
		local
			l_election: RP_CORPORATE_ACTION_ELECTION [CAPACT_EXPRESSION, RP_PORTFOLIO_CORPORATE_ACTION_CONTRACT]
			l_contracts: ROSE_LINEAR_ARRAY [RP_PORTFOLIO_CONTRACT]
			l_holding_history: RP_HOLDING_HISTORY
			l_holding: RP_HOLDING_INTERVAL
			l_corp_actions: ROSE_LINEAR_ARRAY [RP_CORPORATE_ACTION]
			l_shares: RP_QUANTITY
			l_contract: !RP_PORTFOLIO_CORPORATE_ACTION_CONTRACT
		do
			l_corp_actions := a_processing_instructions.corporate_actions
			l_holding_history := manager.accounting.portfolio_holdings_for (a_portfolio, a_instrument)
			l_holding := l_holding_history.witness_interval (a_idate)
			if a_processing_instructions.is_use_current_shares then
				l_shares := l_holding.quantity
			elseif a_processing_instructions.is_use_original_shares then
				l_shares := a_original_holding.quantity
			elseif a_processing_instructions.is_use_delta_shares then
				l_shares := a_original_holding.quantity.instrument.quantity (l_holding.quantity.amount - a_original_holding.quantity.amount)
			else
				l_shares := Void
				manager.log_minor ("Invalid corporate action instruction for " + a_instrument.brimsym)
			end
			from
				create l_contracts.make_empty
				l_corp_actions.start
			until
				l_corp_actions.after
				loop
					-- ask portfolio which election from the corporate action to take
				l_election := manager.corporate_action_election_system.election (a_portfolio, l_corp_actions.item_for_iteration)
					-- convert chosen election to contract and post
				if l_shares /= Void then
					l_contract := l_election.contract (a_idate, a_portfolio, l_shares)
						-- Add contract if we should not ignore that one
					if a_should_skip_generated_contract = Void then
						l_contracts.extend (l_contract)

					else
						if (a_should_skip_generated_contract.item ([l_contract])) then
							manager.log_minor ("skipping contract generation for: "+l_contract.identifier)
						else
							l_contracts.extend (l_contract)
						end
					end
				end
				l_corp_actions.forth
			end
			l_contracts.do_all (a_poster)
		end

	latest_process_date_table: HASH_TABLE [ROSE_DATE, STRING] is
			-- contains latest update date for portfolios
		once
			create Result.make (100)
				-- needs to be initialized
		end

	corporate_action_contract (a_corporate_action: RP_CORPORATE_ACTION; a_idate: !ROSE_DATE; a_portfolio: like portfolio_template; a_holding: !RP_QUANTITY): RP_PORTFOLIO_CORPORATE_ACTION_CONTRACT is
			-- create corporate action contract for a_portfolio, a_corporate_action
		local
			l_election: RP_CORPORATE_ACTION_ELECTION [CAPACT_EXPRESSION, RP_PORTFOLIO_CORPORATE_ACTION_CONTRACT]
		do
			l_election := manager.corporate_action_election_system.election (a_portfolio, a_corporate_action)
			if l_election /= Void then
				Result := l_election.contract (a_idate, a_portfolio, a_holding)
			else
					-- FIXME: publish a task to handle the fact that a decision could not be made
			end
		end

feature {NONE} -- Templates

	portfolio_template: RP_PORTFOLIO is
			-- portfolio template
		do
		end

feature -- Access

	bonus_in_other_instructions (a_basis: RP_CORPORATE_ACTION_PROCESSING_SHARE_BASIS_TYPE; a_capact: CAPACT_BONUS_IN_OTHER_EXPRESSION): !RP_CORPORATE_ACTION_PROCESSING_INSTRUCTIONS is
			-- CA contracts from bonus in same generator
		require
			capact_not_void: a_capact /= Void
		do
	    	create Result.make (a_basis, create {RP_SIMPLE_CORPORATE_ACTION}.make (a_capact.exdate, create {RP_CORPORATE_ACTION_ELECTION_BONUS_IN_OTHER}.make (Current, a_capact)))
		end

	bonus_in_same_instructions (a_capact: CAPACT_BONUS_IN_SAME_EXPRESSION): !RP_CORPORATE_ACTION_ELECTION_BONUS_IN_SAME is
			-- CA contracts from bonus in other generator
		require
			capact_not_void: a_capact /= Void
		do
			create Result.make (Current, a_capact)
		end

	call_payment_instructions (a_capact: CAPACT_CALL_PAYMENT_EXPRESSION): !RP_CORPORATE_ACTION_ELECTION_CALL_PAYMENT is
			-- CA contracts from call payment generator
		require
			capact_not_void: a_capact /= Void
		do
			create Result.make (Current, a_capact)
		end

	capital_return_of_cash_instructions (a_capact: CAPACT_CAPITAL_RETURN_OF_CASH_EXPRESSION): !RP_CORPORATE_ACTION_ELECTION_RETURN is
			-- CA contracts from capital return of cash generator
		require
			capact_not_void: a_capact /= Void
		do
			create Result.make (Current, a_capact)
		end

	cash_dividend_instructions (a_capact: CAPACT_CASH_DIVIDEND_EXPRESSION): !RP_CORPORATE_ACTION_ELECTION_DIVIDEND is
			-- CA contracts from cash dividend generator
		require
			capact_not_void: a_capact /= Void
		do
			create Result.make (Current, a_capact)
		end

	exercise_rights_instructions (a_capact: CAPACT_EXERCISE_RIGHTS_EXPRESSION): !RP_CORPORATE_ACTION_ELECTION_EXERCISE_RIGHTS is
			-- CA contracts from exercise rights generator
		require
			capact_not_void: a_capact /= Void
		do
			create Result.make (Current, a_capact)
		end

	preliminary_subscription_instructions (a_capact: CAPACT_PRELIMINARY_SUBSCRIPTION_EXPRESSION): !RP_CORPORATE_ACTION_ELECTION_PRELIMINARY_SUBSCRIPTION is
			-- CA contracts from preliminary subscription generator
		require
			capact_not_void: a_capact /= Void
		do
			create Result.make (Current, a_capact)
		end

	public_offering_instructions (a_capact: CAPACT_PUBLIC_OFFERING_EXPRESSION): !RP_CORPORATE_ACTION_ELECTION_PUBLIC_OFFERING is
			-- CA contracts from public offering generator
		require
			capact_not_void: a_capact /= Void
		do
			create Result.make (Current, a_capact)
		end

	renominalization_instructions (a_capact: CAPACT_RENOMINALIZATION_EXPRESSION): !RP_CORPORATE_ACTION_ELECTION_RENOMINALIZATION is
			-- CA contracts from renominalization generator
		require
			capact_not_void: a_capact /= Void
		do
			create Result.make (Current, a_capact)
		end

	redenomination_instructions (a_capact: CAPACT_REDENOMINATION_EXPRESSION): !RP_CORPORATE_ACTION_ELECTION_REDENOMINATION is
			-- CA contracts from denomination generator
		require
			capact_not_void: a_capact /= Void
		do
			create Result.make (Current, a_capact)
		end

	renounceable_bonus_in_other_instructions (a_capact: CAPACT_RENOUNCEABLE_BONUS_IN_OTHER_EXPRESSION): !RP_CORPORATE_ACTION_ELECTION_RENOUNCEABLE_BONUS_IN_OTHER is
			-- CA contracts from renounceable bonus in same generator
		require
			capact_not_void: a_capact /= Void
		do
			create Result.make (Current, a_capact)
		end

	renounceable_bonus_in_same_instructions (a_capact: CAPACT_RENOUNCEABLE_BONUS_IN_SAME_EXPRESSION): !RP_CORPORATE_ACTION_ELECTION_RENOUNCEABLE_BONUS_IN_SAME is
			-- CA contracts from renounceable bonus in same generator
		require
			capact_not_void: a_capact /= Void
		do
			create Result.make (Current, a_capact)
		end

	renounceable_rights_in_other_instructions (a_capact: CAPACT_RENOUNCEABLE_RIGHTS_IN_OTHER_EXPRESSION): !RP_CORPORATE_ACTION_ELECTION_RENOUNCEABLE_RIGHTS_IN_OTHER is
			-- CA contracts from renounceable rights in other generator
		require
			capact_not_void: a_capact /= Void
		do
			create Result.make (Current, a_capact)
		end

	renounceable_rights_in_same_instructions (a_capact: CAPACT_RENOUNCEABLE_RIGHTS_IN_SAME_EXPRESSION): !RP_CORPORATE_ACTION_ELECTION_RENOUNCEABLE_RIGHTS_IN_SAME is
			-- CA contracts from renounceable rights in same generator
		require
			capact_not_void: a_capact /= Void
		do
			create Result.make (Current, a_capact)
		end

	return_on_share_premium (a_capact: CAPACT_RETURN_ON_SHARE_PREMIUM_EXPRESSION): !RP_CORPORATE_ACTION_ELECTION_RETURN is
			-- CA contracts from return on share premium generator
		require
			capact_not_void: a_capact /= Void
		do
			create Result.make (Current, a_capact)
		end

	rights_instructions (a_capact: CAPACT_EXERCISE_RIGHTS_EXPRESSION): !RP_CORPORATE_ACTION_ELECTION_EXERCISE_RIGHTS is
			-- CA contracts from rights generator
		require
			capact_not_void: a_capact /= Void
		do
			create Result.make (Current, a_capact)
		end

	rights_in_other_instructions (a_capact: CAPACT_RIGHTS_IN_OTHER_EXPRESSION): !RP_CORPORATE_ACTION_ELECTION_RIGHTS_IN_OTHER is
			-- CA contracts from rights in other generator
		require
			capact_not_void: a_capact /= Void
		do
			create Result.make (Current, a_capact)
		end

	rights_in_same_instructions (a_capact: CAPACT_RIGHTS_IN_SAME_EXPRESSION): !RP_CORPORATE_ACTION_ELECTION_RIGHTS_IN_SAME is
			-- CA contracts from exercise rights generator
		require
			capact_not_void: a_capact /= Void
		do
			create Result.make (Current, a_capact)
		end

	stock_bonus_instructions (a_capact: CAPACT_STOCK_BONUS_EXPRESSION): !RP_CORPORATE_ACTION_ELECTION_STOCK_BONUS is
			-- CA contracts from split generator
		require
			capact_not_void: a_capact /= Void
		do
			create Result.make (Current, a_capact)
		end

	split_instructions (a_capact: CAPACT_SPLIT_EXPRESSION): !RP_CORPORATE_ACTION_ELECTION_SPLIT is
			-- CA contracts from split generator
		require
			capact_not_void: a_capact /= Void
		do
			create Result.make (Current, a_capact)
		end

	stock_dividend_instructions (a_capact: CAPACT_STOCK_DIVIDEND_EXPRESSION): !RP_CORPORATE_ACTION_ELECTION_STOCK_DIV is
			-- CA contracts from stock dividend generator
		require
			capact_not_void: a_capact /= Void
		do
			create Result.make (Current, a_capact)
		end

	stock_dividend_in_other_instructions (a_capact: CAPACT_STOCK_DIVIDEND_IN_OTHER_EXPRESSION): !RP_CORPORATE_ACTION_ELECTION_STOCK_DIVIDEND_IN_OTHER is
			-- CA contracts from stock dividend in other generator
		require
			capact_not_void: a_capact /= Void
		do
			create Result.make (Current, a_capact)
		end

	unknown_instructions (a_capact: CAPACT_UNKNOWN_EXPRESSION): !RP_CORPORATE_ACTION_ELECTION_UNKNOWN is
			-- CA contracts from unknown generator
		require
			capact_not_void: a_capact /= Void
		do
			create Result.make (Current, a_capact)
		end

	writeup_of_capital_instructions (a_capact: CAPACT_WRITEUP_OF_CAPITAL_EXPRESSION): !RP_CORPORATE_ACTION_ELECTION_WRITEUP_OF_CAPITAL is
			-- CA contracts from write-up of capital generator
		require
			capact_not_void: a_capact /= Void
		do
			create Result.make (Current, a_capact)
		end

	writedown_of_capital_instructions (a_capact: CAPACT_WRITEDOWN_OF_CAPITAL_EXPRESSION): !RP_CORPORATE_ACTION_ELECTION_WRITEDOWN_OF_CAPITAL is
			-- CA contracts from write-down of capital generator
		require
			capact_not_void: a_capact /= Void
		do
			create Result.make (Current, a_capact)
		end

end
