note

	description:

		"Corporate Action System"

	author: "Mark Howard"
	last_editor: "$Author: c4 $"
	date: "$Date: 2015/04/09 $"
	revision: "$Revision: #38 $"
	path: "$File: //rose/source/kernel/expert_system/rp_corporate_action_system.e $"

class RP_CORPORATE_ACTION_SYSTEM

inherit

	RP_INVESTMENT_MANAGER_CLIENT
		redefine
			make
		end

	RP_MESSAGE_FACTORY [RP_CORPORATE_ACTION_MESSAGE]

	ROSE_SHARED_GLOBAL
		export {NONE} all end

	ROSE_AWARE
		export {NONE} all end

	ROSE_IEEE_UNDEFINED
		export {NONE} all end

	KL_SHARED_EXCEPTIONS
		export {NONE} all end

create {RP_INVESTMENT_MANAGER}

	make

feature {NONE} -- Initialization

	make (a_manager: like manager)
			-- <Precursor>
		do
			precursor (a_manager)
			create distribution_costs.make_default
		end

feature -- Access

	data_sources (a_instrument: RP_INSTRUMENT; a_date: ROSE_DATE): DS_INDEXABLE [RP_CAE_DATA_SOURCE] is
			-- CAE data sources for `a_instrument' on `a_date'
		do
			if all_cae_data_sources_saved = Void then
				create all_cae_data_sources_saved.make_default
				all_cae_data_sources_saved.force_first (create {RP_CAE_BACKDIV_DATA_SOURCE}.make (manager))
			end
			Result := all_cae_data_sources_saved
		ensure
			sources_list_not_void: Result /= Void
		end

	next_message: RP_CORPORATE_ACTION_MESSAGE is
			-- <Precursor>
		do
		end

	clear_data_sources is
			-- Clear data source
		do
			all_cae_data_sources_saved := Void
		end

	clear_caches is
			-- Forget everything
		do
			processing_instructions_table_saved := Void
		end

	processing_instructions (a_instrument: RP_INSTRUMENT; a_exdate: ROSE_DATE): ROSE_LINEAR_ARRAY [RP_CORPORATE_ACTION_PROCESSING_INSTRUCTIONS] is
			--  Processing instructions for a_instrument on a_exdate
		require
			non_void_args: a_instrument /= Void and a_exdate /= Void
		do
			Result := processing_instructions_table.item ([a_instrument.uid, a_exdate])
		end

	processing_instructions_table: TUPLE_KEYED_HASH_TABLE [ROSE_LINEAR_ARRAY [RP_CORPORATE_ACTION_PROCESSING_INSTRUCTIONS], TUPLE [INTEGER_64, ROSE_DATE]] is
			-- Table for storage of corp action processing instruction clumps
		do
			if processing_instructions_table_saved = Void then
				create processing_instructions_table_saved.make (k_corporate_action_storable_size)
			end
			Result := processing_instructions_table_saved
		end

	terminating_distribution_date (a_uid: INTEGER_64): ROSE_DATE is
			-- Date of terminating distribution
		do
				--FIXME			Result := terminating_distributions [a_uid]
		end

	distribution_costs: DS_HASH_TABLE [RP_MONIES, MULTI_KEY]
			-- Cost of distributions keyed by portfolio, event date and instrument uid

feature -- Basic Operations

	generate_corp_actions_for_listing (a_instrument: RP_INSTRUMENT; a_start_date: ROSE_DATE; a_end_date: ROSE_DATE) is
			-- generate corp actions for a_listing on a_date
		require
			non_void_args: a_instrument /= Void and a_start_date /= Void and a_end_date /= Void
		local
			l_capital_actions: ROSE_LINEAR_ARRAY [ROSE_COMPLEX_CAPITAL_ACTION]
			i: INTEGER
			l_ca: ROSE_COMPLEX_CAPITAL_ACTION
			l_processing_instructions: like processing_instructions_from_expression
			l_brimsym: STRING
			l_original: RP_CORPORATE_ACTION_PROCESSING_SHARE_BASIS_TYPE
			l_listing: LISTING
		do
			create l_original.make_original
			l_brimsym := a_instrument.brimsym
			if l_brimsym /= Void then
				if {l_instrument: RP_SECURITY} a_instrument then
					l_listing := context.listing_accessor.entity (l_instrument.entity_id)
					l_capital_actions := l_listing.capital_actions_between (a_start_date, a_end_date)
					if l_capital_actions.count > 0 then
						from
							i := l_capital_actions.lower
						until
							i > l_capital_actions.upper
						loop
							l_ca := l_capital_actions [i]
							if l_ca /= Void and then l_ca.is_valid_option then
									-- FIXME: there should be an option selection based on the portfolio
								l_processing_instructions := processing_instructions_from_expression (l_original, l_ca.capact_option_expression)
								if l_processing_instructions /= Void then
									add_processing_instructions (l_processing_instructions, a_instrument, l_ca.exdate)
								end
							end
							i := i + 1
						end
					end
				end
			end
		end

feature -- Change

	memorize_distribution_cost (a_instrument: RP_INSTRUMENT; a_portfolio: RP_PORTFOLIO;
		a_date: ROSE_DATE; a_units_held: RP_DECIMAL; a_cost: RP_MONIES) is
			-- Memorize the cost of distribution on `a_instrument' for `a_portfolio' at `a_date'
		require
			a_instrument_not_void: a_instrument /= Void
			a_portfolio_not_void: a_portfolio /= Void
			a_date_not_void: a_date /= Void
			a_units_held_not_void: a_units_held /= Void
			a_cost_not_void: a_cost /= Void
		local
			l_key: MULTI_KEY
		do
			create l_key.make (<<a_portfolio, a_date, a_instrument.uid, a_units_held>>)
			if distribution_costs.has (l_key) then
				-- Integrity error: we can't already have a distribution for this key
				exceptions.raise (global.formatted (
					"Integrity error: we already have a computed cost for '$3' in '$1' at '$2' with '$4' shares",
					<<a_portfolio, a_date, a_instrument.brimsym, a_units_held>>))
			else
				distribution_costs.force (a_cost, l_key)
			end
		end


feature {NONE} -- Implementation

	processing_instructions_from_expression (a_type: RP_CORPORATE_ACTION_PROCESSING_SHARE_BASIS_TYPE; a_capact_expression: CAPACT_EXPRESSION): ROSE_LINEAR_ARRAY [RP_CORPORATE_ACTION_PROCESSING_INSTRUCTIONS] is
			-- return array of corporate action processing instructions for a_instrument, a_capact_expression
		require
			non_void_args: a_capact_expression /= Void
		local
			l_capact_compound: CAPACT_COMPOUND_EXPRESSION
			l_capact_delta: CAPACT_DELTA_EXPRESSION
			l_capact_additive: CAPACT_ADDITIVE_EXPRESSION
			l_capi: RP_CORPORATE_ACTION_PROCESSING_INSTRUCTIONS
			l_type: RP_CORPORATE_ACTION_PROCESSING_SHARE_BASIS_TYPE
			l_do_not_process: BOOLEAN
		do
			create Result.make_empty
				-- search a_capact_expression to find compound expressions to be processed first
			l_capact_compound := first_compound_expression (a_capact_expression)
			if l_capact_compound /= Void then
					-- process compound expression
				create l_type.make_current
				Result.append_array (compound_expression (l_type, l_capact_compound))
			end
			l_capact_compound ?= a_capact_expression
			l_capact_additive ?= a_capact_expression
			l_capact_delta ?= a_capact_expression
			if l_capact_compound = Void then
					-- more processing
				if l_capact_delta /= Void then
					create l_type.make_delta
					Result.append_array (delta_expression (l_type, l_capact_delta))
				elseif l_capact_additive /= Void then
					Result.append_array (additive_expression (a_type, l_capact_additive))
				else
						-- we have elemental corp action
					if a_capact_expression.target /= Void and then a_capact_expression.target.status_code /= Void then
						if a_capact_expression.target.status_code.as_upper.is_equal ("C") then
							l_do_not_process := True
						end
					end
					l_capi := a_capact_expression.processing_instruction (a_type, manager.corporate_action_processing_system)
					if l_capi /= Void and then not l_do_not_process then
						Result.extend (l_capi)
					end
				end
			end
		ensure
			non_void_result: Result /= Void and then Result.count > 0 and Result [1] /= Void and then Result [1].corporate_actions /= Void
		end

	compound_expression (a_basis: RP_CORPORATE_ACTION_PROCESSING_SHARE_BASIS_TYPE; a_compound_expression: CAPACT_COMPOUND_EXPRESSION): ROSE_LINEAR_ARRAY [RP_CORPORATE_ACTION_PROCESSING_INSTRUCTIONS] is
			-- create at least 2 elements in outer array
		require
			non_void_args: a_basis /= Void and a_compound_expression /= Void
			valid_inner: a_compound_expression.inner_expression /= Void and then (is_capact_elemental (a_compound_expression.inner_expression) or is_capact_compound (a_compound_expression.inner_expression) or is_capact_additive_elemental (a_compound_expression.inner_expression))
		local
			l_corporate_actions: ROSE_LINEAR_ARRAY [RP_CORPORATE_ACTION_PROCESSING_INSTRUCTIONS]
			l_basis: RP_CORPORATE_ACTION_PROCESSING_SHARE_BASIS_TYPE
		do
				-- create corporate action processing instructions
			create Result.make_empty
			l_corporate_actions := processing_instructions_from_expression (a_basis, a_compound_expression.inner_expression)
			Result.append_array (l_corporate_actions)
			create l_basis.make_current
			Result.append_array (processing_instructions_from_expression (l_basis, a_compound_expression.outer_expression))
		end

	additive_expression (a_type: RP_CORPORATE_ACTION_PROCESSING_SHARE_BASIS_TYPE; a_additive_expression: CAPACT_ADDITIVE_EXPRESSION): ROSE_LINEAR_ARRAY [RP_CORPORATE_ACTION_PROCESSING_INSTRUCTIONS] is
			-- add expression to storage
		require
			non_void_arg: a_additive_expression /= Void
		local
			l_additive_expression: CAPACT_ADDITIVE_EXPRESSION
			l_compound_expression: CAPACT_COMPOUND_EXPRESSION
			l_array: ROSE_LINEAR_ARRAY [RP_CORPORATE_ACTION_PROCESSING_INSTRUCTIONS]
			l_capi: RP_CORPORATE_ACTION_PROCESSING_INSTRUCTIONS
		do
			create Result.make_empty
			if a_additive_expression.left_expression /= Void then
				l_additive_expression ?= a_additive_expression.left_expression
				l_compound_expression ?= a_additive_expression.left_expression
				if l_additive_expression /= Void then
					Result.append_array (additive_expression (a_type, l_additive_expression))
				elseif l_compound_expression = Void then
						-- do not process compound expressions - they are handled separately
					l_array := processing_instructions_from_expression (a_type, a_additive_expression.left_expression)
					Result.append_array (l_array)
				end
			end
			l_array := Void
			if a_additive_expression.right_expression /= Void then
				l_additive_expression ?= a_additive_expression.right_expression
				l_compound_expression ?= a_additive_expression.right_expression
				if l_additive_expression /= Void then
					l_array := additive_expression (a_type, l_additive_expression)
				elseif l_compound_expression = Void then
						-- do not process compound expressions - they are handled separately
					l_array := processing_instructions_from_expression (a_type, a_additive_expression.right_expression)
				end
			end
				-- append elements of l_array[].corporate_actions to Result.corporate_actions
				-- l_array should only have 1 element
			if l_array /= Void then
				if l_array.count /= 1 or Result.count > 1 then
					(create {EXCEPTIONS}).raise ("Error in additive_expression - invalid size for corporate action processing instruction array")
				else
					if Result.count = 1 then
						l_capi := Result.item (1)
						l_array [1].corporate_actions.do_all (agent l_capi.add_to_corporate_actions (?))
					else
						Result.append_array (l_array)
					end
				end
			end
		end

	delta_expression (a_type: RP_CORPORATE_ACTION_PROCESSING_SHARE_BASIS_TYPE; a_delta_expression: CAPACT_DELTA_EXPRESSION): ROSE_LINEAR_ARRAY [RP_CORPORATE_ACTION_PROCESSING_INSTRUCTIONS] is
			-- handle delta expression which outer is based on incremental shares from inner
		require
			non_void_args: a_type /= Void and a_delta_expression /= Void
			valid_delta_expression: is_capact_elemental (a_delta_expression.inner_expression) and is_capact_elemental (a_delta_expression.outer_expression)
		local
			l_delta: RP_CORPORATE_ACTION_PROCESSING_SHARE_BASIS_TYPE
		do
			create Result.make_empty
			Result.extend (a_delta_expression.inner_expression.processing_instruction (a_type, manager.corporate_action_processing_system))
			create l_delta.make_delta
			Result.extend (a_delta_expression.outer_expression.processing_instruction (l_delta, manager.corporate_action_processing_system))
		end

	first_compound_expression (a_capact_expression: CAPACT_EXPRESSION): CAPACT_COMPOUND_EXPRESSION is
			-- search a_capact_expression and return first found compound expression if any
		require
				--no requirements - a_capact_expression can be void
		local
			l_capact_compound: CAPACT_COMPOUND_EXPRESSION
			l_capact_delta: CAPACT_DELTA_EXPRESSION
			l_capact_additive: CAPACT_ADDITIVE_EXPRESSION
		do
				-- search a_capact_expression to find compound expressions to be processed first
			l_capact_compound ?= a_capact_expression
			l_capact_additive ?= a_capact_expression
			l_capact_delta ?= a_capact_expression
			if l_capact_compound /= Void then
				Result := l_capact_compound
			elseif l_capact_additive /= Void then
				Result := first_compound_expression (l_capact_additive.left_expression)
				if Result = Void then
					Result := first_compound_expression (l_capact_additive.right_expression)
				end
			elseif l_capact_delta /= Void then
				Result := first_compound_expression (l_capact_delta.inner_expression)
				if Result = Void then
					Result := first_compound_expression (l_capact_delta.outer_expression)
					if Result /= Void then
						output ("%NInvalid case - delta expression has compound outer for $1 on $2", <<a_capact_expression.brimsym, a_capact_expression.exdate>>)
					end
				else
					output ("%NInvalid case - delta expression has compound inner for $1 on $2", <<a_capact_expression.brimsym, a_capact_expression.exdate>>)
				end
			end
		end

feature {NONE} -- Status Report

	is_capact_additive_elemental (a_capact_expression: CAPACT_EXPRESSION): BOOLEAN is
			-- is capact expression comprised of addditive (recursively) elemental expressions
			-- Void arg is considered additive
		require
			arg_can_be_void: a_capact_expression = Void or a_capact_expression /= Void
		local
			l_additive_expression: CAPACT_ADDITIVE_EXPRESSION
		do
			if a_capact_expression = Void then
				Result := True
			elseif is_capact_elemental (a_capact_expression) then
				Result := True
			else
				l_additive_expression ?= a_capact_expression
				if l_additive_expression /= Void then
					Result := is_capact_additive_elemental (l_additive_expression.left_expression) and is_capact_additive_elemental (l_additive_expression.right_expression)
				end
			end
		end

	is_capact_elemental (a_capact_expression: CAPACT_EXPRESSION): BOOLEAN is
			-- True if capact_expression is not additive, delta, or compound
		local
			l_capact_compound: CAPACT_COMPOUND_EXPRESSION
			l_capact_delta: CAPACT_DELTA_EXPRESSION
			l_capact_additive: CAPACT_ADDITIVE_EXPRESSION
		do
			l_capact_compound ?= a_capact_expression
			l_capact_delta ?= a_capact_expression
			l_capact_additive ?= a_capact_expression
			if l_capact_compound = Void and l_capact_delta = Void and l_capact_additive = Void then
				Result := True
			end
		end

	is_capact_compound (a_capact_expression: CAPACT_EXPRESSION): BOOLEAN is
			-- Is 'a_capact_expression` a compound expression?
		local
			l_capact_compound: CAPACT_COMPOUND_EXPRESSION
		do
			l_capact_compound ?= a_capact_expression
			Result := l_capact_compound /= Void
		end

	is_capact_additive (a_capact_expression: CAPACT_EXPRESSION): BOOLEAN is
			-- Is 'a_capact_expression` a compound expression?
		local
			l_capact_additive: CAPACT_ADDITIVE_EXPRESSION
		do
			l_capact_additive ?= a_capact_expression
			Result := l_capact_additive /= Void
		end

feature -- Implementation

	add_processing_instructions (a_processing_instructions: ROSE_LINEAR_ARRAY [RP_CORPORATE_ACTION_PROCESSING_INSTRUCTIONS]; a_instrument: RP_INSTRUMENT; a_exdate: ROSE_DATE) is
			-- add corporate action processing instructions for given instrument, date
		require
			args_not_void: a_processing_instructions /= Void and a_instrument /= Void and a_exdate /= Void
		local
			l_processing_instructions: ROSE_LINEAR_ARRAY [RP_CORPORATE_ACTION_PROCESSING_INSTRUCTIONS]
		do
				-- see if we have a clump for a_instrument on a_exdate
			l_processing_instructions := processing_instructions_table.item ([a_instrument.uid, a_exdate])
			if l_processing_instructions = Void then
				create l_processing_instructions.make_empty
				processing_instructions_table.put (l_processing_instructions, [a_instrument.uid, a_exdate])
			end
			l_processing_instructions.append_array (a_processing_instructions)
		end

	persist (a_path: EL_DIRECTORY; a_storable_extension: STRING) is
			-- persist the accounting system
		do
			global.persist.store (Current, a_path.path + "\corporate_action_system." + a_storable_extension)
		end

feature {NONE} -- Constants

	k_corporate_action_storable_size: INTEGER is 1250000
			-- initial size of full corporate action system hash table

feature {NONE} -- Templates

	portfolio_template: RP_PORTFOLIO is
			-- portfolio template
		do
		end

feature {NONE} -- Implementation

	all_cae_data_sources_saved: DS_LINKED_LIST [RP_CAE_DATA_SOURCE]
			-- All data sources

	processing_instructions_table_saved: like processing_instructions_table
			-- table (not really cache) of corporate action processing instructions (name s/b changed but may break simulation storables!)

	brimsym_from_listing (a_listing: LISTING; a_date: ROSE_DATE): STRING is
			-- get brimsym from a_listing on a_date; if void, try using today
		require
			non_void_args: a_listing /= Void and a_date /= Void
		do
			Result := a_listing.brim_code (a_date)
		end

invariant
	distribution_costs_set: distribution_costs /= Void
	no_void_distribution: not distribution_costs.has_void_item
end
