indexing

	description:

		"Cost entry associated to an accounting flow"

	date: "$Date: 2010/09/13 $"
	sccid: "$Id: //rose/source/kernel/expert_system/rp_cost.e#9 $"

class RP_COST

inherit

	ANY
		redefine
			is_equal
		end

create

	make_as_cost_basis,
	make_internal

feature {NONE} -- Initialization

	make_internal (a_flow: like flow; a_quantity: like quantity; a_cost: like cost; a_date: like date) is
			-- Create
		require
			flow_not_void: a_flow /= Void
			quantity_not_void: a_quantity /= Void
			monies_not_void: a_cost /= Void
			date_not_void: a_date /= Void
		do
			flow := a_flow
			quantity := a_quantity
			cost := a_cost
			date := a_date
		ensure
			flow_set: flow = a_flow
			quantity_set: quantity = a_quantity
			cost_set: cost = a_cost
			date_set: date = a_date
		end

	make_as_cost_basis (a_flow: like flow) is
			-- Initialize as costs basis for `a_flow'
		local
			l_cost: RP_MONIES
		do
			if a_flow.signed_quantity.sign >= 0 then
				l_cost := a_flow.initial_fair_value.abs
			else
				l_cost := -a_flow.initial_fair_value.abs
			end
			make_internal (a_flow, a_flow.signed_quantity, l_cost, a_flow.contract.agreement_time)
		end

feature -- Copy

	duplicate: like Current is
			-- Copy current
		do
			create Result.make_internal (flow, quantity.duplicate, cost.duplicate, date)
		ensure
			duplicated: Result ~ Current
		end

feature -- Comparison

	is_equal (a_other: like Current): BOOLEAN is
			-- <Precursor>
		do
			Result := (flow ~ a_other.flow) and then (quantity ~ a_other.quantity) and then (cost ~ a_other.cost) and then (date ~ a_other.date)
		end

feature -- Access

	flow: RP_FLOW
			-- Original flow that created this cost

	quantity: RP_QUANTITY
			-- Quantity of instrument that should be collected or disbursed

	cost: RP_MONIES
			-- Cost

	date: ROSE_DATE
			--  Date is recorded, for instance for an inkind when the lot was established
			-- This can be different from the date on the flow

feature -- Query

	total_cost (a_currency: RP_CURRENCY; a_fx: RP_CURRENCY_MARKET_DATA_SOURCE): RP_MONIES is
			-- Cost converted to 'a_currency' using 'rates'
		require
			currency_not_void: a_currency /= Void
		do
			Result := a_currency.monies (cost.amount * a_fx.spot_rate (cost.currency, a_currency, flow.contracted_ownership_time))
		ensure
			result_not_void: Result /= Void
		end

	unit_cost (a_currency: RP_CURRENCY; a_fx: RP_CURRENCY_MARKET_DATA_SOURCE): RP_MONIES is
			-- Unit cost
		local
			l_total: like total_cost
		do
			l_total := total_cost (a_currency, a_fx)
			Result := l_total.currency.monies (l_total.amount / quantity.amount)
		ensure
			total_cost_divided_by_quantity: total_cost (a_currency, a_fx).amount ~ Result.amount * quantity.amount
		end

feature -- Change

	scale (a_factor: RP_DECIMAL) is
			-- Scale all cost quantities with 'a_factor'
		require
			factor_not_void: a_factor /= Void
		do
			quantity := quantity.instrument.quantity (quantity.amount * a_factor)
		ensure
			adjusted: (old quantity.amount / a_factor) ~ quantity.amount
		end

	set_cost (a_cost: like cost) is
			-- Set value of `cost'
		require
			not_void_args: a_cost /= Void
		do
			cost := a_cost
		ensure
			cost_set: cost = a_cost
		end

feature -- Basic Operations

	proportion (a_quantity: RP_QUANTITY): RP_COST is
			-- Proportion of current cost in 'a_quantity'
		require
			smaller_quantity: quantity.abs >= a_quantity.abs
			same_sign: quantity.sign = a_quantity.sign
		do
			create Result.make_internal (flow, a_quantity, cost_pro_rata (a_quantity), date)
		end

	infix "-" (a_other: RP_COST): RP_COST is
			-- Current minus 'a_other'
		require
			other_not_void: a_other /= Void
			same_flow: flow ~ a_other.flow
			same_date: date ~ a_other.date
		local
			l_qty: RP_QUANTITY
		do
			l_qty := quantity - a_other.quantity
			create Result.make_internal (flow, l_qty, cost_pro_rata (l_qty), date)
		ensure
			result_not_void: Result /= Void
		end

feature {NONE} -- Implementation

	cost_pro_rata (a_quantity: RP_QUANTITY): like cost is
			-- Pro-rata of cost with 'a_quantity'
		require
			not_void_qty: (a_quantity /= Void)
			smaller_quantity: quantity.abs >= a_quantity.abs
			same_sign: (quantity.sign = a_quantity.sign) or else (a_quantity.is_zero)
		do
			if a_quantity.is_zero then
				Result := cost.currency.monies (0)
					-- FIXME: use #QNAN?
			else
				Result := cost.currency.monies (cost.amount * a_quantity.amount / quantity.amount)
			end
		ensure
			result_not_void: Result /= Void
		end

invariant

	cost_not_void: cost /= Void

end
