indexing

	description:
	"[
			Base value of an asset for tax purposes 
			this can be a fraction of the original value 
			we track cost per group and instrument (shares, cash, future cash, ...)
		]"
	author: "PortfolioAccountingTech (sm)"
	date: "$Date: 2010/10/03 $"
	sccid: "$Id: //rose/source/kernel/expert_system/cost/rp_cost_basis.e#5 $"

class RP_COST_BASIS

create

	make

feature {NONE} -- Initialization

	make (a_id: like id; a_group: like group; a_instrument: like instrument; a_long_side: BOOLEAN) is
			-- Create new cost tracker for 'a_group' and 'a_instrument'
		require
			id_positive: a_id > 0
			group_not_void: a_group /= Void
			instrument_not_void: a_instrument /= Void
		do
			can_check_non_exhausted_tax_lots := True
			id := a_id
			group := a_group
			instrument := a_instrument
			is_long_side := a_long_side
			create tax_lots.make_default
			create non_exhausted_tax_lots.make_default
		ensure
			id_set: id = a_id
			group_set: group = a_group
			instrument_set: instrument = a_instrument
			is_long_side_set: is_long_side = a_long_side
			can_check_non_exhausted_tax_lots: can_check_non_exhausted_tax_lots
		end

feature {RP_PORTFOLIO_CONTRACT_COST_UNIT_TESTER} -- Duplication

	duplicate: RP_COST_BASIS is
			-- Duplicate of current
		do
			create Result.make (id, group, instrument, is_long_side)
			Result.can_check_non_exhausted_tax_lots := False
			tax_lots.do_all (agent add_duplicate_tax_lot (Result, ?))
			Result.can_check_non_exhausted_tax_lots := True
		ensure
			duplicate_not_void: Result /= Void
			duplicate: Result /= Current
			same_id: Result.id = id
			same_group: Result.group = group
			same_instrument: Result.instrument = instrument
			same_long_side: Result.is_long_side = is_long_side
			same_number_of_tax_lots: Result.tax_lots.count = tax_lots.count
			same_tax_lots_except_basis: same_tax_lots_except_basis (Result.tax_lots)
			same_number_of_non_exhausted_tax_lots: Result.non_exhausted_tax_lots.count = non_exhausted_tax_lots.count
		end

feature -- Access

	id: INTEGER
			-- Unique id (<200,000 [zone, instrument] pairs as of 2010/02)

	group: RP_ZONAL_INVESTMENT_GROUP
			-- group

	instrument: RP_INSTRUMENT
			-- Instrument

	tax_lots: DS_LINKED_LIST [RP_TAX_LOT]
			-- Tax lots

	non_exhausted_tax_lots: DS_HASH_TABLE [RP_TAX_LOT, INTEGER]
			-- Non-exhausted taxt lots indexed by tax lot id

feature -- Status

	is_long_side: BOOLEAN
			-- Is cost basis for LONG side?

	is_short_side: BOOLEAN
			-- Is cost basis for SHORT side?
		do
			Result := not is_long_side
		ensure
			definition: Result = not is_long_side
		end

feature -- Basic Operations

	collect (a_cost: RP_COST) is
			-- Add 'a_flow'.quantity of instrument at 'a_basis' cost
		require
			a_cost_not_void: a_cost /= Void
		local
			l_lot: RP_TAX_LOT
		do
			can_check_non_exhausted_tax_lots := False
			create l_lot.make (tax_lots.count + 1, Current, a_cost, is_long_side)
			tax_lots.put_last (l_lot)
			can_check_non_exhausted_tax_lots := True
		ensure
			one_more_tax_lot: tax_lots.count = old tax_lots.count + 1
			can_check_non_exhausted_tax_lots: can_check_non_exhausted_tax_lots
		end

	disburse (a_flow: RP_FLOW; a_method: RP_COST_ACCOUNTING_METHOD) is
			-- Remove money from existing tax lot(s)
		require
			a_flow_not_void: a_flow /= Void
			a_method_not_void: a_method /= Void
		do
				-- Create at least one lot if none existing
			a_method.disburse (a_flow, Current)
		end

feature {RP_TAX_LOT} -- Update

	maintain_non_exhausted (a_lot: RP_TAX_LOT) is
			-- Maintain list of non-exhausted lots
		require
			non_void_lot: a_lot /= Void
		do
			if a_lot.is_exhausted then
				non_exhausted_tax_lots.remove (a_lot.id)
			else
				non_exhausted_tax_lots.force (a_lot, a_lot.id)
			end
		ensure
			one_more_or_one_less_or_no_new_non_exhauted_tax_lot:
				(non_exhausted_tax_lots.count - old non_exhausted_tax_lots.count).abs <= 1
		end

feature -- Change

	scale (a_factor: RP_DECIMAL) is
			-- Scale all quantities for all cost flows  with 'a_factor'
		require
			factor_not_void: a_factor /= Void
		do
			tax_lots.do_all (agent {RP_TAX_LOT}.scale (a_factor))
		end

feature -- Query

	quantity: RP_QUANTITY is
			-- Net quantity of current tax lot
		local
			l_units: RP_DECIMAL
		do
			create l_units.make_zero
			from
				tax_lots.start
			until
				tax_lots.after
			loop
				l_units := l_units + tax_lots.item_for_iteration.net
				tax_lots.forth
			end
			Result := instrument.quantity (l_units)
		ensure
			result_not_void: Result /= Void
			same_instrument: Result.instrument ~ instrument
		end

	total_cost (a_currency: RP_CURRENCY; a_fx: RP_CURRENCY_MARKET_DATA_SOURCE): RP_MONIES is
			-- Total cost in `a_currency' using `a_fx' to convert all amounts into that currency
		require
			a_currency_not_void: a_currency /= Void
			a_fx_not_void: a_fx /= Void
		local
			l_cursor: DS_LINKED_LIST_CURSOR [RP_TAX_LOT]
		do
			Result := a_currency.monies (0)
			from
				l_cursor := tax_lots.new_cursor
				l_cursor.start
			until
				l_cursor.after
			loop
				Result.add (l_cursor.item.total_cost (a_currency, a_fx))
				l_cursor.forth
			end
		ensure
			total_cost_not_void: Result /= Void
			same_currency: Result.currency ~ a_currency
		end

	cost_flows (a_flow: RP_FLOW): DS_LINKED_LIST [RP_COST] is
			-- Cost flows created if 'a_flow' was added or subtracted
		do
				--FIXME
--			create Result.make_default
--			from
--				tax_lots.start
--			until
--				tax_lots.after
--			loop
--				if tax_lots.item_for_iteration.cost_flows.has_key (a_flow) then
--					Result.put_last (tax_lots.item_for_iteration.cost_flows.item_at (a_flow))
--				end
--				tax_lots.forth
--			end
		ensure
			result_not_void: Result /= Void
		end

	to_string: STRING is
			-- String representation of `Current'
		do
			create Result.make_empty
			Result.append_integer (id)
			Result.append_character ('|')
			Result.append_string (group.code)
			Result.append_character ('|')
			Result.append_string (instrument.brimsym)
			Result.append_character ('|')
			if is_long_side then
				Result.append_string ("LONG")
			else
				Result.append_string ("SHORT")
			end
			Result.append_character ('%N')
			tax_lots.do_all (agent dump_tax_lot (Result, ?))
		ensure
			non_void_result: Result /= Void
		end

feature {NONE} -- Assertions helpers

	same_tax_lots_except_basis (a_tax_lots: like tax_lots): BOOLEAN is
			-- Are all items in `tax_lots' the same as the ones in `a_tax_lots' except for `basis'?
		require
			a_tax_lots_not_void: a_tax_lots /= Void
			same_number_of_items: a_tax_lots.count = tax_lots.count
			different_tax_lots: tax_lots /= a_tax_lots
		do
			from
				Result := True
				tax_lots.start
				a_tax_lots.start
			until
				tax_lots.after or not Result
			loop
				Result := tax_lots.item_for_iteration.same_except_for_basis (a_tax_lots.item_for_iteration)
				tax_lots.forth
				a_tax_lots.forth
			end
		end

feature {RP_COST_BASIS, RP_TAX_LOT} -- Assertions helpers

	can_check_non_exhausted_tax_lots: BOOLEAN assign set_can_check_non_exhausted_tax_lots
			-- To enable/disable invariant that checks that all non_exhausted_tax_lots items
			-- are also part of `tax_lots'

	set_can_check_non_exhausted_tax_lots (a_can_check: BOOLEAN) is
			-- Sets `can_check_non_exhausted_tax_lots' to `a_can_check'
		do
			can_check_non_exhausted_tax_lots := a_can_check
		ensure
			can_check_non_exhausted_tax_lots_set: can_check_non_exhausted_tax_lots = a_can_check
		end

feature {NONE} -- Implementation

	dump_tax_lot (s: STRING; a_lot: RP_TAX_LOT) is
			-- Dump `a_lot' to `s'
		require
			non_void_args: s /= Void and a_lot /= Void
		do
			s.append_character ('%T')
			s.append_character ('|')
			s.append_integer (a_lot.id)
			s.append_character ('|')
			s.append_boolean (a_lot.is_exhausted)
			s.append_character ('|')
			s.append_string (a_lot.collection.quantity.amount.out)
			s.append_character ('|')
			s.append_string (a_lot.collection.cost.amount.out)
			s.append_character ('|')
			s.append_string (a_lot.collection.flow.contract.identifier)
			s.append_character ('|')
			s.append_string (a_lot.collection.flow.type.as_string)
			s.append_character ('%N')
			a_lot.disbursments.do_all_with_key (agent dump_disbursment (s, a_lot, ?, ?))
			s.append_character ('%T')
			s.append_character ('%T')
			s.append_character ('|')
			s.append_string (a_lot.net.out)
			s.append_character ('%N')
		end

	dump_disbursment (s: STRING; a_lot: RP_TAX_LOT; a_amount: RP_DECIMAL; a_id: INTEGER) is
			-- Dump disbursments to `s'
		do
			s.append_character ('%T')
			s.append_character ('%T')
			s.append_character ('|')
			s.append_string (a_amount.out)
			s.append_character ('|')
			s.append_string (a_lot.disbursment_flows.item (a_id).contract.identifier)
			s.append_character ('|')
			s.append_string (a_lot.disbursment_flows.item (a_id).type.as_string)
			s.append_character ('%N')
		end

	add_duplicate_tax_lot (a_cost_basis: RP_COST_BASIS; a_tax_lot: RP_TAX_LOT) is
			-- Adds a duplicate of `a_tax_lot' to `a_cost_basis'
		require
			a_cost_basis_not_void: a_cost_basis /= Void
			a_tax_lot_not_void: a_tax_lot /= Void
		do
			a_cost_basis.tax_lots.force_last (a_tax_lot.duplicate (a_cost_basis))
		end

invariant

	group_not_void: group /= Void
	instrument_not_void: instrument /= Void
	tax_lots_not_void: tax_lots /= Void
	not_void_tax_lot: not tax_lots.has_void
	non_exhausted_tax_lots_not_void: non_exhausted_tax_lots /= Void
	no_void_non_exhausted_tax_lot: not non_exhausted_tax_lots.has_void_item
	no_non_exhausted_tax_lot_is_exhausted: can_check_non_exhausted_tax_lots implies not non_exhausted_tax_lots.there_exists (agent {RP_TAX_LOT}.is_exhausted)
	all_non_exhausted_tax_lots_are_tax_lots: can_check_non_exhausted_tax_lots implies non_exhausted_tax_lots.for_all (agent tax_lots.has)
	tax_lots_are_for_current: tax_lots.for_all (agent {RP_TAX_LOT}.is_linked_to_basis (Current))
end
