note

	description:

		"Country Entity"

	author: "Ivan Tarnavsky"
	last_editor: "$Author: sm $"
	date: "$Date: 2013/04/26 $"
	revision: "$Revision: #32 $"
	path: "$File: //rose/source/kernel/expert_system/ref/rp_country.e $"

class RP_COUNTRY

inherit

	RP_PARTY
		rename
			code as iso_code,
			name as iso_code
		end

	RP_PERFORMANCE_CONSTANTS_SHARED
		undefine
			is_equal
		end

	DEBUG_OUTPUT
		undefine
			is_equal
		end

create {RP_REFERENCE_SYSTEM}

	make_with_country

feature {NONE} -- Constructors

	make_with_country (a_ref: like reference_system; a_country: like country) is
			-- Initialize `Current' from `a_country'
		do
			reference_system := a_ref
			country := a_country
			if country /= Void then
				iso_code := country.iso_code_most_recent_key (Void)
			else
				iso_code := "??"
				is_unknown := True
			end
		end

feature -- Access

	reference_system: RP_REFERENCE_SYSTEM
			-- Ref system

	debug_output: STRING is
			-- Display the code
		do
			Result := iso_code
		end

	iso_code: STRING
			-- Two letter ISO code

	bank_calendar: RP_CALENDAR is
			-- The bank dates calendar
		do
			Result := retrieve_bank_calendar
		end

	market_calendar: RP_CALENDAR is
			-- The market dates calendar
		do
			Result := retrieve_market_calendar
		end

	government_tax_authority: RP_INSTRUMENT_CONTROL is
			-- Tax authority for this country
		do
			create Result.make_with_owner (Current)
		end

	country_withholding_tax_rate (a_date: ROSE_DATE): RP_DECIMAL is
			-- Dividend withholding tax rate on `a_date'
		require
			a_date_not_void: a_date /= Void
		do
			if is_usa then
				if a_date < constants.uzone_last_non_resident_alien_rate_date then
					Result := constants.thirty_percent
				else
					Result := 0
				end
			else
				Result := country_rate (a_date) / 100
			end
		ensure
			country_withholding_tax_rate_not_void: Result /= Void
			ge_zero: Result >= 0
			lt_one: Result < 1
		end

	portfolio_withholding_tax_rate (a_portfolio: RP_PORTFOLIO; a_date: ROSE_DATE): RP_DECIMAL is
			-- Dividend withholding tax rate for `a_portfolio' on `a_date'
			-- If no portfolio specific rate is defined then return `country_rate'
		require
			a_portfolio_not_void: a_portfolio /= Void
			a_date_not_void: a_date /= Void
		do
			if is_usa then
				if a_date < constants.uzone_last_non_resident_alien_rate_date then
					Result := constants.thirty_percent
				else
					Result := 1 - u_zone_tax_factor (a_portfolio, a_date)
				end
			else
				Result := portfolio_country_rate (a_portfolio, a_date) / 100
			end
		ensure
			portfolio_withholding_tax_rate_not_void: Result /= Void
			ge_zero: Result >= 0
			lt_one: Result < 1
		end

	currency (a_date: ROSE_DATE): RP_CURRENCY is
			-- Currency on `a_date'
		local
			l_ccy: CURRENCY
		do
			if country /= Void then
				l_ccy := country.currency_history.witness (a_date)
				if l_ccy /= Void then
					Result := reference_system.currency_from_iso_code (l_ccy.iso_code_most_recent_key (a_date))
				end
			end
		end

feature -- Status

	is_unknown: BOOLEAN
			-- Is `Current' the "unknown" country

	is_usa: BOOLEAN is
			-- Is `Current' USA?
		do
			Result := iso_code ~ "US"
		end

	is_new_zealand: BOOLEAN is
			-- Is `Current' New Zealand?
		do
			Result := iso_code ~ "NZ"
		end

	zone: RP_ZONE
			-- Zone associated with the country
		do
			if not is_unknown and then country.zone /= Void then
				Result := reference_system.investment_region_by_code (country.zone.mnemonic.out)
			end
		end

feature {RP_ZONE, RP_LEGACY_PERF} -- Entity

	country: COUNTRY
			-- Country

feature {NONE} -- Implementation

	big_date_constant: INTEGER is 99990000
			-- Any date greater than this number (ccyymmdd) wil be considered as maximum date possible

	max_date_constant: INTEGER is 99991231
			-- Maximum date possible (ccyymmdd)


	p_rates_cache: like rates_cache
		-- Lazy initialization of rates_cache

	rates_cache: HASH_TABLE [ROSE_INTERVAL_HISTORY [RP_DECIMAL, ROSE_CLOSED_INTERVAL [RP_DECIMAL]], STRING] is
			-- Dividend withholding tax rates for all zones but U
			-- Data is hashed by country or country|portfolio
		local
			l_servers: DATABASE_SERVER
			l_query: COLLECTING_PREDICATE [TUPLE [country: STRING; portfolio: STRING; start_date: INTEGER; end_date: INTEGER; rate: STRING]]
		do
			if p_rates_cache = Void then
				create p_rates_cache.make (0)
				create l_servers
				create l_query.make (l_servers.servers.performance_server_fast_load, "SELECT LTRIM(RTRIM(UPPER([country]))) as country, LTRIM(RTRIM(UPPER([portfolio]))) as portfolio, [start_date], [end_date], cast ([rate] as varchar(50)) as rate FROM [dbo].[WITHHOLDING_TAX_CLIENT_PERF]")
				l_query.retrieve (Void, Void)
				l_query.array.do_all (agent parse_rate (?, p_rates_cache))
			end
			Result := p_rates_cache
		end

	parse_rate (a_rate_definition: TUPLE [country: STRING; portfolio: STRING; start_date: INTEGER; end_date: INTEGER; rate: STRING]; a_rates: like rates_cache) is
			-- Create interval from `a_rate_definition' and add to `a_rates'
		require
			a_rate_definition_not_void: a_rate_definition /= Void
			country_not_void: a_rate_definition.country /= Void
			portfolio_not_void: a_rate_definition.portfolio /= Void
			start_date_positive: a_rate_definition.start_date > 0
			end_date_positive: a_rate_definition.end_date > 0
			rate_not_void: a_rate_definition.rate /= Void
			a_rates_not_void: a_rates /= Void
		local
			l_rates: ROSE_INTERVAL_HISTORY [RP_DECIMAL, ROSE_CLOSED_INTERVAL [RP_DECIMAL]]
			l_interval: ROSE_CLOSED_INTERVAL [RP_DECIMAL]
			l_key: STRING
			l_end_dt: ROSE_DATE
			l_start_dt: ROSE_DATE
			l_rate: RP_DECIMAL
		do
			create l_key.make_from_string (a_rate_definition.country)
			if not a_rate_definition.portfolio.is_empty then
				l_key.append_character ('|')
				l_key.append_string (a_rate_definition.portfolio)
			end
			l_rates := a_rates.item (l_key)
			if l_rates = Void then
				create l_rates.make_empty
				a_rates.extend (l_rates, l_key)
			end
			l_start_dt := valid_date (a_rate_definition.start_date, 1)
			l_end_dt := valid_date (a_rate_definition.end_date, -1)
			create l_rate.make_from_string (a_rate_definition.rate)
			if l_rate < 0 or l_rate >= 100 then
				(create {EXCEPTIONS}).raise ("[WITHHOLDING_TAX_CLIENT_PERF] rate outside of [0,100) range [" + a_rate_definition.country + "][" + a_rate_definition.portfolio + "][" + a_rate_definition.start_date.out + "][" + a_rate_definition.end_date.out + "][" + a_rate_definition.rate + "]")
			end
			create l_interval.make (l_rate, l_start_dt.min (l_end_dt), l_end_dt.max (l_start_dt), l_rate)
			l_rates.extend (l_interval)
		end

	portfolio_country_rate (a_portfolio: RP_PORTFOLIO; a_date: ROSE_DATE): RP_DECIMAL is
			-- Dividend withholding tax rate for `a_portfolio' on `a_date'
			-- If no portfolio specific rate is defined then return `country_rate'
		require
			a_portfolio_not_void: a_portfolio /= Void
			a_date_not_void: a_date /= Void
		local
			l_rates: ROSE_INTERVAL_HISTORY [RP_DECIMAL, ROSE_CLOSED_INTERVAL [RP_DECIMAL]]
			l_rate: ROSE_CLOSED_INTERVAL [RP_DECIMAL]
		do
			l_rates := rates_cache.item (iso_code + "|" + a_portfolio.code)
			if l_rates /= Void then
				l_rate := l_rates.witness_interval (a_date)
			end
			if l_rate /= Void then
				Result := l_rate.item
			else
				Result := country_rate (a_date)
			end
		ensure
			portfolio_country_rate_not_void: Result /= Void
		end

	country_rate (a_date: ROSE_DATE): RP_DECIMAL is
			-- Dividend withholding tax rate on `a_date'
		require
			a_date_not_void: a_date /= Void
		local
			l_rates: ROSE_INTERVAL_HISTORY [RP_DECIMAL, ROSE_CLOSED_INTERVAL [RP_DECIMAL]]
			l_rate: ROSE_CLOSED_INTERVAL [RP_DECIMAL]
		do
			l_rates := rates_cache.item (iso_code)
			if l_rates /= Void then
				l_rate := l_rates.witness_interval (a_date)
			end
			if l_rate /= Void then
				Result := l_rate.item
			else
					-- Some countries do not have their default
					-- dividend withholding rate defined
					-- VMS code uses 0% in such case
				Result := 0
			end
		ensure
			country_rate_not_void: Result /= Void
		end

	cached_bank_calendar: like bank_calendar

	retrieve_bank_calendar: like bank_calendar is
			-- Bank calendar retrieveal
		do
			if cached_bank_calendar = Void then
				if country /= Void then
					cached_bank_calendar := create {RP_ROSE_BUSINESS_DATE_CALENDAR}.make (country.bank_dates)
				else
					cached_bank_calendar := create {RP_WEEKDAYS_CALENDAR}
				end
			end
			Result := cached_bank_calendar
		end

	cached_market_calendar: like market_calendar

	retrieve_market_calendar: like market_calendar is
			-- Market Calendar Retrieval
		do
			if cached_market_calendar = Void then
				if country /= Void then
					cached_market_calendar := create {RP_ROSE_BUSINESS_DATE_CALENDAR}.make (country.market_dates)
				else
					cached_market_calendar := create {RP_WEEKDAYS_CALENDAR}
				end
			end
			Result := cached_market_calendar
		end

	market_dates: ROSE_DATE_SET is
			-- The market dates for this country
		do
			Result := country.market_dates
		end

	bank_dates: ROSE_DATE_SET is
			-- The bank dates for this country
		do
			Result := country.bank_dates
		end

	p_uzone_rates_cache: like uzone_rates_cache
		-- Implementation

	uzone_rates_cache: DS_HASH_TABLE [ROSE_INTERVAL_HISTORY [RP_DECIMAL, ROSE_CLOSED_OPEN_INTERVAL [RP_DECIMAL]], STRING] is
			-- U zone portfolio withholding tax rates
			-- Data is hashed by portfolio
		local
			l_query: COLLECTING_PREDICATE [TUPLE [portfolio: STRING; start_ccyymmdd: INTEGER; finish_ccyymmdd: INTEGER; rate: STRING]]
		do
			if p_uzone_rates_cache = Void then
				create p_uzone_rates_cache.make_equal (17)
				create l_query.make ((create {DATABASE_ACCESS}).make_from_repository_fast_load ("performance"), "SELECT LTRIM(RTRIM(UPPER([portfolio]))) as portfolio, start_ccyymmdd, finish_ccyymmdd, cast (tax_rate as varchar(50)) FROM [dbo].[WITHHOLDING_TAX_RATE_UZONE]")
				l_query.retrieve (Void, Void)
				l_query.array.do_all (agent parse_uzone_rate (?, p_uzone_rates_cache))
			end
			Result := p_uzone_rates_cache
		end

	parse_uzone_rate (a_rate_definition: TUPLE [portfolio: STRING; start_ccyymmdd: INTEGER; finish_ccyymmdd: INTEGER; rate: STRING]; a_rates: like uzone_rates_cache) is
			-- Create interval from `a_rate_definition' and add to `a_rates'
		require
			a_rate_definition_not_void: a_rate_definition /= Void
			portfolio_not_void: a_rate_definition.portfolio /= Void
			positive_start_date: a_rate_definition.start_ccyymmdd > 0
			positive_end_date: a_rate_definition.finish_ccyymmdd > 0
			rate_not_void: a_rate_definition.rate /= Void
			a_rates_not_void: a_rates /= Void
		local
			l_rates: ROSE_INTERVAL_HISTORY [RP_DECIMAL, ROSE_CLOSED_OPEN_INTERVAL [RP_DECIMAL]]
			l_interval: ROSE_CLOSED_OPEN_INTERVAL [RP_DECIMAL]
			l_end_dt: ROSE_DATE
			l_start_dt: ROSE_DATE
			l_rate: RP_DECIMAL
		do
			a_rates.search (a_rate_definition.portfolio)
			if a_rates.found then
				l_rates := a_rates.found_item
			else
				create l_rates.make_empty
				a_rates.force (l_rates, a_rate_definition.portfolio)
			end
			l_start_dt := valid_date (a_rate_definition.start_ccyymmdd, 1)
			l_end_dt := valid_date (a_rate_definition.finish_ccyymmdd, -1)
			create l_rate.make_from_string (a_rate_definition.rate)
			if l_rate <= 0 or l_rate > 1 then
				(create {EXCEPTIONS}).raise ("[WITHHOLDING_TAX_RATE_UZONE] rate outside of (0,1] range [" + a_rate_definition.portfolio + "][" + a_rate_definition.start_ccyymmdd.out + "][" + a_rate_definition.finish_ccyymmdd.out + "][" + a_rate_definition.rate + "]")
			end
			create l_interval.make (l_rate, l_start_dt.min (l_end_dt), l_end_dt.max (l_start_dt), l_rate)
			l_rates.extend (l_interval)
		end

	u_zone_tax_factor (a_portfolio: RP_PORTFOLIO; a_date: ROSE_DATE): RP_DECIMAL is
			-- U zone specific tax rate
		require
			non_void_portfolio: a_portfolio /= Void
		local
			l_interval: ROSE_CLOSED_OPEN_INTERVAL [RP_DECIMAL]
		do
			uzone_rates_cache.search (a_portfolio.code)
			if uzone_rates_cache.found then
				l_interval := uzone_rates_cache.found_item.witness_interval (a_date)
				if l_interval /= Void then
					Result := l_interval.item
				end
			end
			if Result = Void then
				Result := 1
			end
		ensure
			u_zone_tax_factor_not_void: Result /= Void
		end

	valid_date (a_ccyymmdd: INTEGER; a_step: INTEGER): ROSE_DATE is
			-- Cope with invalid date integers in `a_ccyymmdd'
		require
			a_step_is_not_zero: a_step /= 0
		local
			l_ccyymmdd: INTEGER
		do
			l_ccyymmdd := a_ccyymmdd
			if l_ccyymmdd > big_date_constant then
				l_ccyymmdd := max_date_constant
			end
			from
				create Result.make_ccyymmdd (l_ccyymmdd)
			until
				l_ccyymmdd = Result.ccyymmdd
					-- Loop until l_ccyymmdd is a valid date
			loop
				l_ccyymmdd := l_ccyymmdd + a_step
				create Result.make_ccyymmdd (l_ccyymmdd)
			end
		ensure
			date_not_void: Result /= Void
		end

invariant

	iso_code_not_void: iso_code /= Void
	iso_code_two_letters: iso_code.count = 2
	known_implies_country_defined: not is_unknown implies country /= Void
end
