note

	description:

		"Custodian System"

	author: "Mark Howard"
	last_editor: "$Author: e2 $"
	date: "$Date: 2011/11/29 $"
	revision: "$Revision: #21 $"
	path: "$File: //rose/source/kernel/expert_system/rp_custodian_system.e $"

class RP_CUSTODIAN_SYSTEM

inherit

	RP_CONTRACT_SYSTEM

	RP_SWEPT_ACCOUNTS_AWARE
		export {NONE} all end

create {RP_INVESTMENT_MANAGER}

	make

feature -- Query

	has_dividend_sweep_ssi (a_portfolio: RP_PORTFOLIO; a_trade_date, a_settle_date: ROSE_DATE): BOOLEAN is
			-- Has dividend swept standing settlement instructions?
		require
			portfolio_not_void: a_portfolio /= Void
			dates_not_void: a_trade_date /= Void and a_settle_date /= Void
		do
			if a_portfolio.zone.is_united_states then
				Result := u_is_swept_at (a_portfolio, a_trade_date, a_settle_date)
			else
				Result := w_is_non_special_dividend_swept_at (a_portfolio, a_trade_date, a_settle_date)
			end
		end

feature -- Basic Operations

	process_portfolio_message (a_message: RP_PORTFOLIO_MESSAGE) is
			-- Process `a_message', check whether there are SSI's and generate contracts if appropriate
		local
			l_cash_and_settlement: TUPLE [cash: RP_MONIES; settlement_date: ROSE_DATE]
		do
			if a_message.is_addition and then {l_cash_dividend: RP_CASH_DIVIDEND} a_message.new_g and then has_dividend_sweep_ssi (l_cash_dividend.portfolio, l_cash_dividend.agreement_time, l_cash_dividend.paydate) then
				sweep_dividend (l_cash_dividend.portfolio, l_cash_dividend, l_cash_dividend.settle_date)
			end
			if a_message.is_addition and then {l_interest: RP_INTEREST} a_message.new_g then
				l_cash_and_settlement := cash_from_interest (l_interest)
				if l_cash_and_settlement /= Void and then has_interest_income_sweep_ssi (l_interest.portfolio, l_interest.agreement_time, l_cash_and_settlement.cash) then
					sweep_interest (l_interest.portfolio, l_interest, l_cash_and_settlement.cash, l_cash_and_settlement.settlement_date)
				end
			end
			if a_message.is_addition and then {l_income: RP_INCOME} a_message.new_g then
				l_cash_and_settlement := cash_from_income (l_income)
				if l_cash_and_settlement /= Void and then has_interest_income_sweep_ssi (l_income.portfolio, l_income.agreement_time, l_cash_and_settlement.cash) then
					sweep_income (l_income.portfolio, l_income, l_cash_and_settlement.cash, l_cash_and_settlement.settlement_date)
				end
			end
		end

	contract_template: RP_CUSTODIAN_CONTRACT is
			-- Contract type template
		do
		end

feature {NONE} -- Implementation

	sweep_interest (a_portfolio: RP_PORTFOLIO; a_interest: RP_INTEREST; a_cash: RP_MONIES; a_settlement_date: ROSE_DATE) is
			-- Sweep 'a_interest' for 'a_portfolio'
		require
			a_portfolio_not_void: a_portfolio /= Void
			a_interest_not_void: a_interest /= Void
			a_cash_not_void: a_cash /= Void
			a_settlement_date_not_void: a_settlement_date /= Void
		local
			l_sweep: RP_CASH_SWEPT_FROM_INTEREST
		do
			create l_sweep.make (a_interest, a_cash, a_cash, a_interest.agreement_time, a_settlement_date, a_portfolio, a_interest.identifier + "/SWEEP", Current, a_interest.custody)
			manager.post (l_sweep)
		end

	cash_from_interest (a_interest: RP_INTEREST): TUPLE [cash: RP_MONIES; settle_date: ROSE_DATE] is
			-- Cash from interest
		do
			from
				a_interest.flows.start
			until
				a_interest.flows.after
			loop
				if a_interest.flows.current_item.type.is_interest_flow then
					if {l_cash: RP_MONIES} a_interest.flows.current_item.quantity then
						Result := [l_cash, a_interest.flows.current_item.contracted_settlement_time]
					end
				end
				a_interest.flows.forth
			end
		end

	sweep_income (a_portfolio: RP_PORTFOLIO; a_income: RP_INCOME; a_cash: RP_MONIES; a_settlement_date: ROSE_DATE) is
			-- Income swept
		require
			a_portfolio_not_void: a_portfolio /= Void
			a_income_not_void: a_income /= Void
			a_cash_not_void: a_cash /= Void
			a_settlement_date_not_void: a_settlement_date /= Void
		local
			l_sweep: RP_CASH_SWEPT_FROM_INCOME
		do
			create l_sweep.make (a_income, a_cash, a_cash, a_income.agreement_time, a_settlement_date, a_portfolio, a_income.identifier + "/SWEEP", Current, a_income.custody)
			manager.post (l_sweep)
		end

	cash_from_income (a_income: RP_INCOME): TUPLE [cash: RP_MONIES; settlement_date: ROSE_DATE] is
			-- Cash from income
		do
			from
				a_income.flows.start
			until
				a_income.flows.after
			loop
				if a_income.flows.current_item.type.is_income_flow then
					if {l_cash: RP_MONIES} a_income.flows.current_item.quantity then
						Result := [l_cash, a_income.flows.current_item.contracted_settlement_time]
					end
				end
				a_income.flows.forth
			end
		end

	sweep_dividend (a_portfolio: RP_PORTFOLIO; a_dividend: RP_CASH_DIVIDEND; a_settlement_date: ROSE_DATE) is
			-- Sweep dividend
		require
			a_portfolio_not_void: a_portfolio /= Void
			a_dividend_not_void: a_dividend /= Void
			a_settlement_date_not_void: a_settlement_date /= Void
		local
			l_sweep: RP_CASH_SWEPT_FROM_DIVIDEND
			l_gross_net: TUPLE [gross: RP_MONIES; net: RP_MONIES]
		do
			l_gross_net := gross_and_net_amount_from_dividend (a_dividend)
			create l_sweep.make (a_dividend, l_gross_net.net, l_gross_net.gross, a_dividend.agreement_time, a_settlement_date, a_portfolio, a_dividend.identifier + "/SWEEP", Current, a_dividend.custody)
			manager.post (l_sweep)
		end

	gross_and_net_amount_from_dividend (a_dividend: RP_DIVIDEND): TUPLE [RP_MONIES, RP_MONIES] is
			-- Net amount from 'a_dividend'
		local
			l_gross: RP_DECIMAL
			l_tax: RP_DECIMAL
			l_currency: RP_CURRENCY
		do
			l_gross := 0
			l_tax := 0
			l_currency := manager.reference_system.currency_from_iso_code ("USD")
			from
				a_dividend.flows.start
			until
				a_dividend.flows.after
			loop
				if a_dividend.flows.current_item.type.is_dividend_flow then
					if {l_ccy: RP_CURRENCY} a_dividend.flows.current_item.quantity.instrument then
						l_gross := a_dividend.flows.current_item.quantity.amount
						l_currency := l_ccy
					end
				end
				if a_dividend.flows.current_item.type.is_dividend_withholding_country_tax_flow then
					if {l_ccy2: RP_CURRENCY} a_dividend.flows.current_item.quantity.instrument then
						l_tax := a_dividend.flows.current_item.quantity.amount
					end
				end
				a_dividend.flows.forth
			end
			Result := [l_currency.monies (l_gross), l_currency.monies (l_gross - l_tax)]
		ensure
			result_not_void: Result /= Void
		end

	has_interest_income_sweep_ssi (a_portfolio: RP_PORTFOLIO; a_date: ROSE_DATE; a_cash: RP_MONIES): BOOLEAN is
			-- Has interest income swept standing settlement instructions?
		require
			portfolio_not_void: a_portfolio /= Void
			date_not_void: a_date /= Void
			cash_not_void: a_cash /= Void
		local
			s: STRING
			i: INTEGER
		do
				-- Add trailing zeroes to two decimal points
				-- 123
				-- 123.5
				-- 123.56
			create s.make_from_string (a_cash.amount.rounded (2).out)
			i := s.index_of ('.', 1)
			if i > 0 then
					-- 123.5
				if i = (s.count - 1) then
					s.append ("0")
				end
			else
					-- 123
				s.append (".00")
			end
			Result := interest_income_sweep_ssi.has (a_portfolio.code + "|" + a_date.ccyymmdd.out + "|" + a_cash.currency.name + "|" + s)
		end

	interest_income_sweep_ssi: DS_HASH_SET [STRING] is
			-- Shared income swept standing settlement instructions
		local
			l_query: STRING_PREDICATE
		once
				-- FIXME: This should be somewhere else...
			create l_query.make ((create {DATABASE_ACCESS}).make_from_repository_fast_load ("gedb_fg"), "[
				select rtrim(ltrim(portfolio))+'|'+convert(varchar(50),event_date,112)+'|'+rtrim(ltrim(currency))+'|'+cast(cash_effect as varchar(50))
				from open_account_events
				where event_type = 32
				order by 1
			]")
			l_query.retrieve (Void, Void)
			create Result.make_equal (l_query.array.count)
			l_query.array.do_all (agent Result.force (?))
		end

end
