note

	description:
	"[
		Wrapper for the Windows DECIMAL object. This implementation should work
		for both classic and .Net on Windows.
		Managed pointer 'data' holds the COM representation of the decimal,
		attributes 'lc_1' and 'lc_2' holds the extracted representation of 'data' to allow
		sending this object using SED or the storable mechanism
		This class is used both in Eiffel Classic and Eiffel .NET but has a different implementation in VMS
	]"
	keywords: "DECIMAL data fixed point arithmetic"
	author: "Sami Ben Grine"
	last_editor: "$Author: it $"
	date: "$Date: 2011/04/04 $"
	revision: "$Revision: #25 $"
	path: "$File: //rose/source/kernel/infrastructure/windows/classic/rp_decimal.e $"

class RP_DECIMAL

inherit

	COM_STRING_FUNCTIONS
		export
			{NONE} all
		redefine
			out,
			is_equal,
			copy
		end

	COMPARABLE
		undefine
			is_equal,
			out,
			copy
		end

	HASHABLE
		undefine
			is_equal,
			out,
			copy
		end

	NUMERIC
		undefine
			is_equal,
			out,
			copy
		end

	IEEE_MATH
		export
			{NONE} all
		undefine
			is_equal,
			out,
			copy
		end

create

	default_create,
	make_from_double,
	make_from_real,
	make_from_integer,
	make_from_natural,
	make_undefined,
	make_zero,
	make_from_string,
	make_managed_pointer

convert

	make_from_string ({STRING}),
	make_from_double ({DOUBLE}),
	make_from_real ({REAL}),
	make_from_integer ({INTEGER}),
	make_from_natural ({NATURAL_32}),
	to_double: {DOUBLE}

feature {NONE} -- Initialization

	make_managed_pointer (a_data: MANAGED_POINTER) is
			-- Create from 'a_data'
		require
			data_not_void: a_data /= Void
			data__not_default_pointer: a_data.item /= default_pointer
			valid_data_count: a_data.count = byte_count
		do
			data_i := a_data
			unload
		ensure
			data_set: data_i = a_data
		end

	make_from_double (d: DOUBLE) is
			-- Create from DOUBLE
		do
			if isnan (d) then
				make_undefined
			else
				check_hresult (vardecfromr8 (d, data.item))
			end
		end

	make_from_integer (i: INTEGER) is
			-- Create from INTEGER
		do
			check_hresult (vardecfromi4 (i, data.item))
		end

	make_from_natural (i: NATURAL_32) is
			-- Create from NATURAL_32
		do
			check_hresult (vardecfromui4 (i, data.item))
		end

	make_from_real (i: REAL) is
			-- Create from REAL
		do
			check_hresult (vardecfromr4 (i, data.item))
		end

	make_undefined is
			-- Create as undefined
		do
			data.item.memory_set (0xFF, byte_count)
			unload
		end

	make_zero is
			-- Create as zero
		do
			make_from_integer (0)
		end

	make_from_string (s: STRING) is
			-- Create from STRING
		require
			string_not_void: s /= Void
		do
			check_hresult (vardecfromstr (string_to_widechar (s), 0, 0, data.item))
		end

	check_hresult (a_hresult: INTEGER) is
			-- Load if a_hresult is ok or make undefined otherwise
		do
			if a_hresult = S_OK then
				unload
			else
				make_undefined
			end
		end

feature -- Constants

	byte_count: INTEGER is 16
			-- Number of bytes needed to represent a DECIMAL

	undefined_value: RP_DECIMAL is
			-- Create as undefined
		once
			create Result.make_undefined
		end

	zero: RP_DECIMAL is
			-- Zero
		once
			create Result.make_zero
		end

	one: RP_DECIMAL is
			-- One
		once
			create Result.make_from_integer (1)
		end

	exponentiable (a_other: like Current): BOOLEAN is
			-- Why is this true when an exponentiation operator isn't supplied?
		do
			Result := a_other > 0
		end

	divisible (a_other: like Current): BOOLEAN is
			-- Should this be modified to flag divide by zero?
			-- To flag under/overflow?
		do
			Result := not a_other.is_zero
		end

feature -- Access

	sign: INTEGER is
			-- Sign value (0, -1 or 1)
		do
			if Current > zero then
				Result := 1
			elseif Current < zero then
				Result := -1
			end
		ensure
			three_way: Result = three_way_comparison (zero)
		end

	floor: RP_DECIMAL is
			-- Largest integer (positive, negative or zero) smaller than current
		do
			if is_zero then
				Result := zero
			elseif Current > zero then
				Result := truncated
			else
				Result := truncated - One
			end
		ensure
			floor_not_void: Result /= Void
			floor_smaller: Result <= Current
			in_range: (Result - Current).abs < One
		end

	ceiling	: RP_DECIMAL is
			-- Smallest integer (positive, negative or zero) larger than current
		do
			if is_zero then
				Result := zero
			elseif Current > zero then
				Result := truncated + One
			else
				Result := truncated
			end
		ensure
			ceiling_not_void: Result /= Void
			ceiling_larger: Result >= Current
			in_range: (Result - Current).abs < One
		end

feature -- Basic Operations

	round (a_decimal_places: INTEGER) is
			-- Round current to 'a_decimal_places'
		require
			decimal_places_positive: a_decimal_places >= 0
		local
			l_rounded: RP_DECIMAL
		do
			l_rounded := rounded (a_decimal_places)
			copy (l_rounded)
		end

	truncate is
			-- Truncate current
		local
			l_truncated: RP_DECIMAL
		do
			l_truncated := truncated
			copy (l_truncated)
		end

feature -- Query

	is_equal (a_other: RP_DECIMAL): BOOLEAN is
			-- Is 'a_other' equal to current?
		do
			Result := VarDecCmp (data.item, a_other.data.item) = 1
		end

	is_zero: BOOLEAN is
			-- Is Current equal to 'zero'?
		do
			Result := zero.is_equal (Current)
		ensure
			equals_zero: Result = zero.is_equal (Current)
		end

	is_one: BOOLEAN is
			-- Is Current equal to 'one'?
		do
			Result := one.is_equal (Current)
		ensure
			equals_one: Result = one.is_equal (Current)
		end

	is_defined: BOOLEAN is
			-- Is Current equal to undefined_value?
		do
			Result := not (data.is_equal (undefined_value.data))
		end

feature -- Operators

	prefix "+": like Current is
			-- Unary plus
		do
			Result := Current
		end

	infix "+" (a_other: RP_DECIMAL): RP_DECIMAL is
			-- Add 'a_other' to Current
		local
			l_hresult: INTEGER
		do
			if is_defined and a_other.is_defined then
				create Result
				l_hresult := vardecadd (data.item, a_other.data.item, Result.data.item)
				Result.unload
			else
				create Result.make_undefined
			end
		end

	infix "-" (a_other: RP_DECIMAL): RP_DECIMAL is
			-- Substract 'a_other' from Current
		local
			l_hresult: INTEGER
		do
			if is_defined and a_other.is_defined then
				create Result
				l_hresult := vardecsub (data.item, a_other.data.item, Result.data.item)
				Result.unload
			else
				create Result.make_undefined
			end
		end

	infix "/" (a_other: RP_DECIMAL): RP_DECIMAL is
			-- Divide Current by 'a_other'
		local
			l_hresult: INTEGER
		do
			if is_defined and a_other.is_defined then
				create Result
				l_hresult := vardecdiv (data.item, a_other.data.item, Result.data.item)
				Result.unload
			else
				create Result.make_undefined
			end
		end

	infix "*" (a_other: RP_DECIMAL): RP_DECIMAL is
			-- Multiply Current by 'a_other'
		local
			l_hresult: INTEGER
		do
			if is_defined and a_other.is_defined then
				create Result
				l_hresult := vardecmul (data.item, a_other.data.item, Result.data.item)
				Result.unload
			else
				create Result.make_undefined
			end
		end

	prefix "-": RP_DECIMAL is
			-- Unary minus
		do
			Result := zero - Current
		end

	infix "^" (a_other: RP_DECIMAL): RP_DECIMAL is
			-- Current to the power `a_other'
		local
			d1, d2: DOUBLE
		do
			d1 := Current
			d2 := a_other
			Result := d1 ^ d2
		end

	abs: RP_DECIMAL is
			-- Absolute value of Current
		do
			if Current < zero then
				Result := -Current
			else
				Result := Current
			end
		end

	infix "<" (a_other: RP_DECIMAL): BOOLEAN is
			-- Is Current strictly lower than 'a_other'
		do
			Result := VarDecCmp (data.item, a_other.data.item) = 0
		end

feature -- Duplication

	copy (other: like Current) is
			-- <Precursor>
		do
			standard_copy (other)
				-- Duplicate memory pointed by `data_i'.
			data_i := other.data.twin
		end

feature -- Conversion

	debug_output: STRING is
			-- String representation in the debugger
		do
			Result := out
		end

	to_double: DOUBLE is
			-- Convert to DOUBLE
		local
			l_managed_data: MANAGED_POINTER
			l_hresult: INTEGER
		do
			if is_defined then
				create l_managed_data.make (8)
				l_hresult := varr8fromdec (data.item, l_managed_data.item)
				Result := l_managed_data.read_real_64 (0)
			else
				Result := qnan
			end
		end

	truncated: RP_DECIMAL is
			-- Current value with no decimals
		local
			l_hresult: INTEGER
		do
			if is_defined then
				create Result
				l_hresult := vardecfix (data.item, Result.data.item)
				Result.unload
			else
				Result := undefined_value
			end
		end

	truncated_to_integer: INTEGER is
			-- Convert to INTEGER type
			-- Warning! This will typically convert a 128 bit value to 32bit.
			-- Use 'truncated' if you want to simply truncate to an other RP_DEICMAL
		do
			Result := to_double.truncated_to_integer
		end

	out, to_string: STRING is
			-- String representation
		local
			l_buffer: MANAGED_POINTER
			l_hresult: INTEGER
		do
			if is_defined then
				create l_buffer.make (8)
				l_hresult := VarBstrFromDec (data.item, 0, 0, l_buffer.item)
				Result := widechar_to_string (l_buffer.read_pointer (0))
				sysfreestring (l_buffer.read_pointer (0))
			else
				Result := "#QNAN"
			end
		end

	to_hex_string: STRING is
			-- Hexadecimal representation of the bits defining this decimal object
		local
			i: INTEGER
		do
			from
				create Result.make (16 * 3)
				i := 0
			until
				i >= data.count
			loop
				Result.append (data.read_natural_8 (i).to_hex_string)
				i := i + 1
				if i < data.count then
					Result.append_character (' ')
				end
			end
		end

	rounded (a_decimal_places: INTEGER): RP_DECIMAL is
			-- Current value rounded to 'a_decimal_places' digits
		local
			l_hresult: INTEGER
		do
			if is_defined then
				create Result
				l_hresult := VarDecRound (data.item, a_decimal_places, Result.data.item)
				Result.unload
			else
				Result := undefined_value
			end
		end

	hash_code: INTEGER is
			-- Hash code
		do
			Result := out.hash_code
		end

feature {RP_DECIMAL, OLEDB_DATA_CONVERT} -- Implementation

	data: MANAGED_POINTER is
			-- Managed pointer
		do
			if data_i = Void or else data_i.item = default_pointer then
				create data_i.make (byte_count)
				data_i.put_natural_64 (lc_1, 0)
				data_i.put_natural_64 (lc_2, 8)
			end
			Result := data_i
		ensure
			data_not_void: Result /= Void
			data_not_default_pointer: Result.item /= default_pointer
			valid_data_count: Result.count = byte_count
		end

	unload is
			-- Load values into internal structure
		do
			data.do_nothing
			lc_1 := data_i.read_natural_64 (0)
			lc_2 := data_i.read_natural_64 (8)
		end

feature {NONE} -- Implementation

	data_i: MANAGED_POINTER
			-- Managed pointer

	lc_1, lc_2: NATURAL_64
			-- Copy of internal data at the object level

	VarDecCmp (p1, p2: POINTER): INTEGER is
			-- Compares two variants of type decimal (external call)
		external
			"C blocking signature (DECIMAL*, DECIMAL*): EIF_INTEGER use <windows.h>"
		alias
			"VarDecCmp"
		end

	vardecfromr8 (d: DOUBLE; dec: POINTER): INTEGER is
			-- This function converts variant data types to DECIMAL from double (external call)
		external
			"C blocking signature (DOUBLE,DECIMAL*): EIF_INTEGER use <windows.h>"
		alias
			"VarDecFromR8"
		end

	vardecfromr4 (d: DOUBLE; dec: POINTER): INTEGER is
			-- This function converts variant data types to DECIMAL from float (external call)
		external
			"C blocking signature (DOUBLE,DECIMAL*): EIF_INTEGER use <windows.h>"
		alias
			"VarDecFromR4"
		end

	vardecfromui4 (d: DOUBLE; dec: POINTER): INTEGER is
			-- Converts an unsigned long value to a decimal value (external call)
		external
			"C blocking signature (DOUBLE, DECIMAL*): EIF_INTEGER use <windows.h>"
		alias
			"VarDecFromUI4"
		end

	vardecfromi4 (d: DOUBLE; dec: POINTER): INTEGER is
			-- This function converts variant data types to DECIMAL from long (external call)
		external
			"C blocking signature (DOUBLE, DECIMAL*): EIF_INTEGER use <windows.h>"
		alias
			"VarDecFromI4"
		end

	vardecadd (l, r, res: POINTER): INTEGER is
			-- Returns the sum of two variants of type decimal (external call)
		external
			"C blocking signature (DECIMAL*, DECIMAL*, DECIMAL*): EIF_INTEGER use <windows.h>"
		alias
			"VarDecAdd"
		end

	vardecsub (l, r, res: POINTER): INTEGER is
			-- Returns the result of subtracting two variants of type decimal (external call)
		external
			"C blocking signature (DECIMAL*, DECIMAL*, DECIMAL*): EIF_INTEGER use <windows.h>"
		alias
			"VarDecSub"
		end

	vardecmul (l, r, res: POINTER): INTEGER is
			-- Returns the result of multiplying two variants of type decimal (external call)
		external
			"C blocking signature (DECIMAL*, DECIMAL*, DECIMAL*): EIF_INTEGER use <windows.h>"
		alias
			"VarDecMul"
		end

	vardecdiv (l, r, res: POINTER): INTEGER is
			-- Returns the result of dividing two variants of type decimal (external call)
		external
			"C blocking signature (DECIMAL*, DECIMAL*, DECIMAL*): EIF_INTEGER use <windows.h>"
		alias
			"VarDecDiv"
		end

	vardecround (l: POINTER; n: INTEGER; res: POINTER): INTEGER is
			-- Rounds a variant of type decimal to the specified number of decimal places (external call)
		external
			"C blocking signature (DECIMAL*, INT, DECIMAL*): EIF_INTEGER use <windows.h>"
		alias
			"VarDecRound"
		end

	varr8fromdec (dec: POINTER; d: POINTER): INTEGER is
			-- This function converts variant data types to double from DECIMAL (external call)
		external
			"C blocking signature (DECIMAL*, double*) : EIF_INTEGER use <windows.h>"
		alias
			"VarR8FromDec"
		end

	VarBstrFromDec (dec: POINTER; lcid: INTEGER; dwFlags: INTEGER; pbstrout: POINTER): INTEGER is
			-- Converts a decimal value to a BSTR value (external call)
		external
			"C blocking signature (DECIMAL*, LCID, ULONG, BSTR*): EIF_INTEGER use <windows.h>"
		alias
			"VarBstrFromDec"
		end

	VarDecFromStr (strIn: POINTER; lcid: INTEGER; dwflags: INTEGER; pdecOut: POINTER): INTEGER is
			-- Converts an OLECHAR string to a decimal value (external call)
		external
			"C blocking signature (OLECHAR*, LCID, ULONG, DECIMAL*): EIF_INTEGER use <windows.h>"
		alias
			"VarDecFromStr"
		end

	sizeof_decimal: INTEGER is
			-- Size of DECIMAL (external call)
		external
			"C blocking macro use <windows.h>"
		alias
			"sizeof (DECIMAL)"
		end

	S_OK: INTEGER is
			-- Success (external call)
		external
			"C blocking macro use <windows.h>"
		alias
			"S_OK"
		end

	vardecfix (pdecIn: POINTER; pdecOut: POINTER): INTEGER is
			-- Returns the integer portion of a variant of type decimal (external call)
		external
			"C blocking signature (DECIMAL*, DECIMAL*): EIF_INTEGER use <windows.h>"
		alias
			"VarDecFix"
		end
end
