note

	description:

		"Container for an eigenvalue decomposition"

	author: "Mark Howard"
	last_editor: "$Author: mk $"
	date: "$Date: 2012/08/20 $"
	revision: "$Revision: #15 $"
	path: "$File: //rose/source/kernel/infrastructure/math/rp_eigensystem.e $"

class RP_EIGENSYSTEM

create

	make

feature {NONE} -- Initialization

	make (a_eigenvectors: MATRIX; a_eigenvalues: VECTOR)
			-- Create eigensystem with `eigenvalues' set to `a_eigenvalues'
			-- and `eigenvectors' set to `a_eigenvectors'.
		require
			a_eigen_vectors_not_void: a_eigenvectors /= Void
			a_eigen_values_not_void: a_eigenvalues /= Void
		do
			eigenvectors := a_eigenvectors
			eigenvalues := a_eigenvalues
		ensure
			eigenvectors_set: eigenvectors = a_eigenvectors
			eigenvalues_set: eigenvalues = a_eigenvalues
		end

feature -- Access

	eigenvectors: MATRIX
			-- Eigenvectors of eigensystem

	eigenvalues : VECTOR
			-- Eigenvalues of eigensystem

feature -- Main operations

	scale (a_factor: DOUBLE)
			-- Scale eigenvectors by `a_factor' and
			-- eigenvalues by (1/`a_factor')^2
		require
			a_factor_nonzero: a_factor /= 0.0
		do
			eigenvectors.scale (a_factor)
			eigenvalues.scale (1.0 / (a_factor * a_factor))
		end

	variable_scale (a_factor_vector: VECTOR)
			-- Scale eigenvector_i by `a_factor_vector [i]' and
			-- eigenvalue_i by (1/ `a_factor_vector[i])^2'
		require
			a_factor_vector_not_void: a_factor_vector /= Void
			a_factor_vector_has_no_undefined_doubles: a_factor_vector.defined_item_count = a_factor_vector.count
			a_factor_vector_has_no_zeroes: a_factor_vector.non_zero_indices.count = a_factor_vector.count
			a_factor_count_correct_size: a_factor_vector.count = eigenvalues.count
		local
			l_evalue_scale_vector: VECTOR
			i: INTEGER
		do
			create l_evalue_scale_vector.make_constant (1.0, a_factor_vector.count)
			from
				i := 1
			until
				i > l_evalue_scale_vector.count
			loop
				l_evalue_scale_vector [i] := 1.0 / (a_factor_vector [i] ^2)
				eigenvectors.scale_column (i, a_factor_vector [i])
				i := i + 1
			end
			eigenvalues := eigenvalues.cross (l_evalue_scale_vector)
		end

	set_desc_order
			-- Set eigenvectors in descending order or eigenvalues
		local
			l_orders: ROSE_LINEAR_ARRAY[INTEGER]
		do
			l_orders := eigenvalues.descending_order
			eigenvalues := eigenvalues.subset (l_orders)
			eigenvectors := eigenvectors.sub_matrix_c (1, l_orders.count, l_orders)
		end

	original_matrix: SYMMETRIC_MATRIX
			-- Original matrix
		local
			l_transposed: MATRIX_TRANSPOSE
		do
			create l_transposed.make (eigenvectors)
			create Result.make_xtDx (l_transposed, eigenvalues)
			Result.set_names (eigenvectors.row_names)
		end

feature -- Selected leading eigensystem access

	principal_eigenvectors (a_principal_count: INTEGER): MATRIX
			-- Eigenvectors corresponding to the top `a_principal_count' eigenvalues
			-- Note: this feature uses a sort of the underlying eigensystem to identify them
		require
			valid_size: a_principal_count <= eigenvectors.width
		local
			l_orders, l_sub_ixs: ROSE_LINEAR_ARRAY [INTEGER]
		do
			l_orders := eigenvalues.descending_order
			l_sub_ixs := l_orders.subarray (1, a_principal_count)
			Result := selected_eigenvectors (l_sub_ixs)
		end

	principal_eigenvectors_unsorted (a_principal_count: INTEGER): MATRIX
			-- Eigenvectors corresponding to the first `a_principal_count' eigenvalues
		require
			valid_size: a_principal_count <= eigenvectors.width
		local
			l_sub_ixs: INTEGER_VECTOR
		do
			create l_sub_ixs.make_sequence (1, a_principal_count, 1)
			Result := selected_eigenvectors (l_sub_ixs)
		end

	selected_eigenvectors (a_indices: ROSE_LINEAR_ARRAY [INTEGER]): MATRIX
			-- Eigenvectors corresponding to `a_indices'
		require
			valid_size: a_indices.count <= eigenvectors.width
		do
			Result := eigenvectors.sub_matrix_c (1, eigenvectors.height, a_indices)
		end

	principal_matrix (a_principal_count: INTEGER): SYMMETRIC_MATRIX
			-- Principal matrix constructed from top `a_principal_count' eigenvectors
			-- Note: this feature uses a sort of the underlying eigensystem to identify them
		require
			valid_size: a_principal_count <= eigenvalues.count
		local
			l_orders, l_sub_ixs: ROSE_LINEAR_ARRAY[INTEGER]
		do
			l_orders := eigenvalues.descending_order
			l_sub_ixs := l_orders.subarray (1, a_principal_count)
			Result := reconstructed_matrix (l_sub_ixs)
		end

	principal_matrix_unsorted (a_principal_count: INTEGER): SYMMETRIC_MATRIX
			-- Principal matrix constructed from first `a_principal_count' eigenvectors
			-- Note: this feature uses a sort of the underlying eigensystem to identify them
		require
			valid_size : a_principal_count <= eigenvalues.count
		local
			l_sub_ixs: INTEGER_VECTOR
		do
			create l_sub_ixs.make_sequence (1, a_principal_count, 1)
			Result := reconstructed_matrix (l_sub_ixs)
		end

	reconstructed_matrix (a_indices: ROSE_LINEAR_ARRAY [INTEGER]): SYMMETRIC_MATRIX
			-- Matrix constructed from `eigenvectors' and `eigenvalues' correspondng to `a_indices'
		require
			valid_size : a_indices.count <= eigenvalues.count
		local
			l_sub_vectors: MATRIX
			l_sub_values: VECTOR
		do
			l_sub_vectors := selected_eigenvectors (a_indices)
			l_sub_values := eigenvalues.subset (a_indices)
			create Result.make_xtDx (l_sub_vectors.transposed, l_sub_values)
		end

	residual_matrix (a_principal_count: INTEGER): SYMMETRIC_MATRIX
			-- Residual matrix of top `a_principal_count' eigenvectors
			-- Note: this feature uses a sort of the underlying eigensystem
		require
			valid_size : a_principal_count < eigenvalues.count
		local
			l_orders, l_sub_ixs: ROSE_LINEAR_ARRAY [INTEGER]
		do
			l_orders := eigenvalues.descending_order
			l_sub_ixs := l_orders.subarray (a_principal_count + 1, eigenvalues.count)
			Result := reconstructed_matrix (l_sub_ixs)
		end

	residual_matrix_unsorted (a_principal_count: INTEGER): SYMMETRIC_MATRIX
			-- Residual matrix of first `a_principal_count' eigenvectors
		require
			valid_size : a_principal_count < eigenvalues.count
		local
			l_sub_ixs: INTEGER_VECTOR
		do
			create l_sub_ixs.make_sequence (a_principal_count + 1, eigenvalues.count, 1)
			Result := reconstructed_matrix (l_sub_ixs)
		end

feature -- Transformed eigensystem access

	transformed_eigenvectors : MATRIX
			-- Transformed eigenvectors (via importance weighting scheme
		local
			l_diag: MATRIX
		do
			create l_diag.make_diagonal (eigenvalues.square_roots)
			Result := eigenvectors * l_diag
		end

	transformation_matrix: MATRIX
			-- Transformation matrix (of importance weighting scheme)
		local
			l_diag : MATRIX
		do
			create l_diag.make_diagonal (eigenvalues.square_roots.inverse)
			Result := eigenvectors * l_diag
		end

	transformation_matrix_inverse: MATRIX
			-- Inverse of `transformation matrix' (of importance weighting scheme)
		local
			l_diag : MATRIX
		do
			create l_diag.make_diagonal (eigenvalues.square_roots)
			Result := l_diag * eigenvectors.transposed
		end

	sqrt_matrix: SYMMETRIC_MATRIX
			-- Square root matrix
		local
			l_weighted_matrix: WEIGHTED_MATRIX
			l_transposed: MATRIX_TRANSPOSE
		do
			if eigenvalues.min >= 0 then
				create l_weighted_matrix.make (eigenvectors, eigenvalues.square_roots.square_roots)
				create l_transposed.make (l_weighted_matrix)
				create Result.make_xtx (l_transposed)
			else
				print ("Matrix is not semi-positive definite")
			end
		end

feature -- Matrix representation

	combined_matrix : MATRIX
			-- Combined matrix
		local
			l_names: ROSE_LINEAR_ARRAY[STRING]
		do
			create Result.make (eigenvalues.count + 1, eigenvalues.count)
			Result.put_row_vector (eigenvalues, 1)
			Result.put_matrix (eigenvectors, 2, 1)
			Result.set_column_names (eigen_names(eigenvalues.count))
			create l_names.make_from_array (eigenvectors.row_names)
			l_names.put_front ("Eigen Values")
			Result.set_row_names (l_names)
		end

	eigen_names (a_n: INTEGER): ROSE_LINEAR_ARRAY[STRING]
			-- Eigen names
		local
			i: INTEGER
		do
			create Result.make (1, a_n)
			from
				i := 1
			until
				i > a_n
			loop
				Result [i] := "Eigen " + i.out
				i := i + 1
			end
		end

	eigenvalue_analysis : MATRIX is
			-- Cumulative effect of eigenvalues
		local
			l_eigenvalues, l_ones, l_vec: VECTOR
			l_orders: ROSE_LINEAR_ARRAY [INTEGER]
			l_column_names: ROSE_LINEAR_ARRAY [STRING]
			l_d_sum: DOUBLE
		do
				-- labels
			create l_column_names.make_from_array (<<"eigen_desc", "x_order", "eigen_cum", "ave_cum">>)
			create Result.make (eigenvalues.count, l_column_names.count)
			Result.set_name ("eigenvalues")
			Result.set_column_names (l_column_names)
			Result.set_row_names (eigen_names (eigenvalues.count))
				-- fill
				-- set descending order	
			l_orders := eigenvalues.descending_order
			l_eigenvalues := eigenvalues.subset (l_orders)
			l_eigenvalues.set_name ("eigen_desc")
			Result.put_column_vector (l_eigenvalues, 1)
				-- order display
				-- normally as x serious when display
			create l_ones.make_identity (l_eigenvalues.count)
			Result.put_column_vector (l_ones.cum_vector, 2)
				-- cumulative
			l_d_sum := l_eigenvalues.sum
			l_vec := l_eigenvalues.cum_vector.scaled (1.0 / l_d_sum)
			l_vec.set_name ("eigen_cum")
			Result.put_column_vector (l_vec, 3)
				-- 45 degree line
			l_vec := l_ones.cum_vector.scaled (1.0 / l_d_sum)
			l_vec.set_name ("ave_cum")
			Result.put_column_vector (l_vec, 4)
		end
end
