note

	description:
	"[
		Fixed Deposit Instrument
		
					                 111111111
							123456789012345678
							GBD029M020M0337548
				 Format:	CCI{cd}{md}{inr}bk
					where	CC - Country Code (GB-Great Britain)
							I - instrument type (D=fixed deposit)
							{cd} - contract date (029M=2002 Sep 23)
							{md} - maturity date (020M=2002 Oct 23)
							{inr} - interest rate (03375=3.375%)
				          	bk - BankAccountType (48)
	]"
	author: "it"
	last_editor: "$Author: e2 $"
	date: "$Date: 2012/02/03 $"
	revision: "$Revision: #16 $"
	path: "$File: //rose/source/kernel/expert_system/rp_fixed_deposit.e $"

class RP_FIXED_DEPOSIT

inherit

	RP_FIXED_INCOME
		undefine
			is_equal,
			is_cash_equivalent
		end

	RP_CASH_EQUIVALENT

	ROSE_UNITS_ENVIRONMENT
		export
			{NONE} all
		undefine
			is_equal
		end

create {RP_REFERENCE_SYSTEM}

	make

feature {NONE} -- Initialization

	make (a_reference: RP_REFERENCE_SYSTEM; a_brimsym: STRING; a_rate_override: like rate) is
			-- Initialize `Current'.
		require
			non_void_args: a_reference /= Void and a_brimsym /= Void
			brimsym_correct_length: a_brimsym.count = valid_brimsym_length
		do
			brimsym := a_brimsym
			name := "(" + a_brimsym + ")"
			-- Extract fields from BRIMSYM string
				-- Currency 1 to 2
			currency := a_reference.currency_from_iso_code (a_brimsym.substring (1, 2))
				-- Contract date 4 to 7
			first_date := interpret_date (a_brimsym.substring (4, 7))
				-- Maturity date 8 to 11
			expiration_date := interpret_date (a_brimsym.substring (8, 11))
			if a_rate_override.abs > 0 then
				rate := a_rate_override / 100
			else
					-- Interest rate 12 to 16
				rate := a_brimsym.substring (12, 16).to_double / brimsym_rate_divisor
			end
				-- Bank account index 17 to 18
			bank_account_index := a_brimsym.substring (17, 18).to_integer
				-- XXX create unique units for this instrument
			rose_units := identified_units (a_brimsym)
		end

feature -- Access

	brimsym: attached STRING
			-- Brimsym

	currency: RP_CURRENCY
			-- Denomination currency

	first_date: attached ROSE_DATE
			-- First rate date

	expiration_date: attached ROSE_DATE
			-- Rate expiration date

	name: attached STRING
			-- Name

	rate: attached RP_DECIMAL
			-- Interest rate

	bank_account_index: INTEGER
			-- Bank account index

	country_of_issue (a_ref_system: RP_REFERENCE_SYSTEM; a_date: ROSE_DATE): RP_COUNTRY is
			-- Country of issue.
		do
			Result := currency.country_of_issue (a_ref_system, a_date)
		end

feature -- Valuation

	holding_value_in_currency (a_holding: RP_HOLDING; a_valuator: RP_HOLDING_VALUATOR; a_currency: RP_CURRENCY; a_date: ROSE_DATE; a_bucket:STRING): RP_MONIES is
			-- Value of `a_holding' on `a_date' in `a_currency' as computed by `a_valuator'
		do
			Result := a_valuator.fixed_deposit_value_in_currency (a_holding, a_currency, a_date, a_bucket)
		end

	price (a_date: ROSE_DATE): attached RP_PRICE is
			-- Price of `Current' on `a_date'
		do
			if not valid_price_date (a_date) then
				Result := price_in_currency (1, currency, a_date)
			else
				Result := price_in_currency (accrued_interest (a_date) + 1, currency, a_date)
			end
		end

	valid_price_date (a_date: ROSE_DATE): BOOLEAN is
			-- Is `a_date' valid price date?
		do
			if a_date /= Void then
				Result := a_date >= first_date
					-- XXX perhaps check if a_date <= expiration_date
			end
		end

feature {NONE} -- Implementation

	accrued_interest (a_date: ROSE_DATE): attached RP_DECIMAL is
			-- Accrued interest from `first_date' to `a_date'
		require
			valid_date: valid_price_date (a_date)
		local
			l_number_of_days: RP_DECIMAL
			l_number_of_days_in_year: RP_DECIMAL
		do
			create l_number_of_days.make_from_integer (a_date.diff_in_days (first_date))
			if currency.is_usd or currency.is_eur then
				create l_number_of_days_in_year.make_from_integer (360)
			else
				create l_number_of_days_in_year.make_from_integer (365)
			end
			Result := l_number_of_days / l_number_of_days_in_year * rate
		end

	interpret_date (a_date_code: STRING): attached ROSE_DATE is
			-- Decode the date from the Fixed Deposit Brimsym, YYMD, where M and D are encoded
		require
			attached_date_code: a_date_code /= Void
			date_code_valid_length: a_date_code.count = 4
		local
			l_day, l_month: INTEGER
		do
				-- Month is 3rd character
			l_month := decode (a_date_code.item (3))
				-- Day is 4th character
			l_day := decode (a_date_code.item (4))
				-- Year is first and second characters
			create Result.make_yymmdd (a_date_code.substring (1, 2).to_integer * 1_00_00 + l_month * 1_00 + l_day)
		end

	decode (c: CHARACTER): INTEGER is
			-- Decode date character, 0 = 10, 1..9, A.. = 11..31
		do
			if c = '0' then
					-- 0 -> 10
				Result := 10
			elseif c.is_digit then
					-- 1..9 -> 1..9
				Result := c.code - 48
			else
					-- A..B -> 11..12
				Result := c.code - 54
			end
		end

feature {NONE}	-- Constants

	Brimsym_rate_divisor: DOUBLE is 100_000.0
		-- Divisor for interest rate within brimsym

end
