note

	description:

		"Flow"

	author: "Mark Howard"
	last_editor: "$Author: e2 $"
	date: "$Date: 2013/02/22 $"
	revision: "$Revision: #14 $"
	path: "$File: //rose/source/kernel/expert_system/rp_flow.e $"

class RP_FLOW

create

	make

feature {NONE} -- Initilization

	make (a_contract: like contract; a_type: like type; a_source_control: like source_control; a_destination_control: like destination_control; a_contracted_ownership_time: like contracted_ownership_time; a_contracted_settlement_time: like contracted_settlement_time; a_quantity: like quantity; a_value_date: like value_date; a_exchanged_subset: like exchanged_subset; a_flow_value: like flow_value) is
			-- Initialize `Current'.
		require
			non_void_contract: a_contract /= Void
			non_void_type: a_type /= Void
			non_void_source_control: a_source_control /= Void
			non_void_destination_control: a_destination_control /= Void
			non_void_quantity: a_quantity /= Void
			non_void_value_date: a_value_date /= Void
			non_void_contracted_ownership_time: a_contracted_ownership_time /= Void
			non_void_contracted_settlement_time: a_contracted_settlement_time /= Void
		do
			contract := a_contract
			type := a_type
			source_control := a_source_control
			destination_control := a_destination_control
			contracted_ownership_time := a_contracted_ownership_time
			contracted_settlement_time := a_contracted_settlement_time
			quantity := a_quantity
			value_date := a_value_date
			exchanged_subset := a_exchanged_subset
			flow_value := a_flow_value
		ensure
			is_portfolio_flow: is_portfolio_flow (contract.portfolio)
		end

feature -- Query

	type: RP_FLOW_TYPE
			-- flow type

	contract: RP_PORTFOLIO_CONTRACT
			-- `contract' producing `Current'

	source_control: RP_INSTRUMENT_CONTROL
			-- source

	destination_control: RP_INSTRUMENT_CONTROL
			-- destination

	contracted_ownership_time: ROSE_DATE
			-- contracted ownership time

	contracted_settlement_time: ROSE_DATE
			-- contracted settlement time

	instrument: RP_INSTRUMENT is
			-- Quantity's instrument
		do
			Result := quantity.instrument
		ensure
			quantity_s_instrument: Result = quantity.instrument
		end

	flow_value: RP_QUANTITY
			-- value of the quantity

	quantity: RP_QUANTITY
			-- Quantity

	value_date: ROSE_DATE
			-- value date

	exchanged_subset: INTEGER
			-- exchange related subset of the flows (e.g. cost)

	basis: RP_COST is
			-- Basis used to compute the cost
		do
			create Result.make_as_cost_basis (Current)
		end

	signed_quantity: RP_QUANTITY is
			-- Quantity with debit/credit sign
		do
			if is_portfolio_debit_flow (contract.portfolio) then
				Result := -quantity
			else
				Result := quantity
			end
		end

feature -- Visitor Pattern

	process (a_visitor: RP_FLOW_VISITOR) is
			-- Visitor pattern method
		require
			non_void_visitor: a_visitor /= Void
		do
			type.process (a_visitor, Current)
		end

feature -- Status Report

	is_portfolio_flow (a_portfolio: RP_PORTFOLIO): BOOLEAN is
			-- Is `Current' related to `a_portfolio'
		require
			a_portfolio_not_void: a_portfolio /= Void
		do
			Result := is_portfolio_credit_flow (a_portfolio) or is_portfolio_debit_flow (a_portfolio)
		ensure
			definition: Result = (is_portfolio_credit_flow (a_portfolio) or is_portfolio_debit_flow (a_portfolio))
		end

	is_portfolio_debit_flow (a_portfolio: RP_PORTFOLIO): BOOLEAN is
			-- Is `Current' debit to `a_portfolio'
			-- e.g. are instruments leaving 'a_portfolio' with 'Current flow?
		require
			a_portfolio_not_void: a_portfolio /= Void
		do
			Result := source_control.possession.location ~ a_portfolio
		ensure
			definition: Result = (source_control.possession.location ~ a_portfolio)
		end

	is_portfolio_credit_flow (a_portfolio: RP_PORTFOLIO): BOOLEAN is
			-- Is `Current' credit to `a_portfolio'
			-- e.g. are instruments arriving in 'a_portfolio' with 'Current flow?
		require
			a_portfolio_not_void: a_portfolio /= Void
		do
			Result := destination_control.possession.location ~ a_portfolio
		ensure
			definition: Result = (destination_control.possession.location ~ a_portfolio)
		end

	is_account_flow (a_account: RP_INVESTMENT_ACCOUNT): BOOLEAN is
			-- Is `Current' related to `a_account'
		require
			a_account_not_void: a_account /= Void
		do
			Result := is_account_credit_flow (a_account) or is_account_debit_flow (a_account)
		ensure
			definition: Result = (is_account_credit_flow (a_account) or is_account_debit_flow (a_account))
		end

	is_account_debit_flow (a_account: RP_INVESTMENT_ACCOUNT): BOOLEAN is
			-- Is `Current' debit to `a_account'
			-- e.g. are instruments leaving 'a_account' with 'Current flow?
		require
			a_account_not_void: a_account /= Void
		do
			if attached {RP_PORTFOLIO} source_control.possession.location as l_portfolio then
				Result :=  l_portfolio.account ~ a_account
			end
		end

	is_account_credit_flow (a_account: RP_INVESTMENT_ACCOUNT): BOOLEAN is
			-- Is `Current' credit to `a_account'
			-- e.g. are instruments arriving in 'a_account' with 'Current flow?
		require
			a_account_not_void: a_account /= Void
		do
			if attached {RP_PORTFOLIO} destination_control.possession.location as l_portfolio then
				Result :=  l_portfolio.account ~ a_account
			end
		end

feature -- Output

	debug_output: STRING is
			-- <Precursor>
		do
			create Result.make_empty
			Result.append (contract.generator)
			Result.append ("|")
			Result.append (contract.identifier)
			Result.append ("|")
			Result.append (contract.agreement_time.ccyymmdd.out)
			Result.append ("|")
			Result.append (type.as_string)
			Result.append ("|")
			Result.append (quantity.instrument.brimsym)
			Result.append ("|")
			Result.append (quantity.amount.rounded (2).out)
			Result.append ("|")
			Result.append (quantity.units.name)
			Result.append ("|SRC:")
			Result.append (source_control.debug_output)
			Result.append ("|DST:")
			Result.append (destination_control.debug_output)
		end

feature -- Conversion

	initial_fair_value: RP_MONIES is
			-- Fair value at agreement time
		do
			if type.is_inkind_flow then
				Result ?= flow_value
			elseif type.is_outkind_flow then
				Result ?= -flow_value
			elseif type.is_delivery_flow then
				Result ?= flow_value
			elseif type.is_distribution_flow then
				Result ?= flow_value
			elseif type.is_stock_dividend_flow then
				Result := contract.portfolio.base_currency (contract.agreement_time).monies (0)
			else
				Result ?= signed_quantity
			end
			if Result = Void and then flow_value /= Void then
				Result ?= flow_value
			end
			if Result = Void then
				(create {EXCEPTIONS}).raise ("Should Not Happen")
			end
		ensure
			result_not_void: Result /= Void
		end

invariant

	non_void_contract: contract /= Void
	non_void_type: type /= Void
	non_void_source: source_control /= Void
	non_void_destination: destination_control /= Void
	non_void_qty: quantity /= Void
	non_void_value_date: value_date /= Void
	non_void_ownership_time: contracted_ownership_time /= Void
	non_void_settlement_time: contracted_settlement_time /= Void

end
