note

	description:

		"GEDB Event"

	author: "Mark Howard"
	last_editor: "$Author: e2 $"
	date: "$Date: 2012/01/04 $"
	revision: "$Revision: #41 $"
	path: "$File: //rose/source/kernel/expert_system/rp_gedb_event.e $"

deferred class RP_GEDB_EVENT

inherit

	RP_INVESTMENT_MANAGER_CLIENT
		rename
			make as rp_investment_manager_client_make
		end

	RP_EVENT_TYPED
		rename
			event_type as code
		end

	DEBUG_OUTPUT

	ROSE_FORMATTING
		export {NONE} all end

feature {NONE} -- Constructor

	make (a_manager: like manager; a_record: like record) is
			-- Initialize `Current'.
		require
			non_void_args: a_manager /= Void and a_record /= Void
			valid_event_type: is_valid_event_type (a_record.event_type)
		do
			rp_investment_manager_client_make (a_manager)
			record := a_record
				-- Initialization
			set_identifier
		ensure
			non_void_identifier: identifier /= Void
		end

	set_identifier is
			-- Set 'identifier'
		do
			create identifier.make_from_string (record.portfolio)
			identifier.left_adjust
			identifier.right_adjust
			identifier.append (",")
			identifier.append_integer (record.icc)
			identifier.append (",")
			identifier.append_integer (record.event_type)
		end

feature -- Access

	identifier: STRING
			-- Event identifier

	icc: INTEGER is
			-- Event ICC number
		do
			Result := record.icc
		end

	set_icc (a_icc: INTEGER) is
			-- Change ICC
		do
			record.set_icc (a_icc)
		ensure
			icc_set: icc = a_icc
		end

	portfolio_contracts: ARRAY [RP_PORTFOLIO_CONTRACT] is
			-- Portfolio constracts for `Current'
		do
			if saved_contracts = Void then
				saved_contracts := <<contract>>
			end
			Result := saved_contracts
		ensure
			non_void_contracts_array: Result /= Void
		end

	event_date: ROSE_DATE is
			-- Event date
		do
			Result := record.event_date
		end

	settle_date: ROSE_DATE is
			-- Settle date
		do
			Result := record.settle_date
		end

	portfolio: RP_PORTFOLIO is
			-- Portfolio.
		do
			if saved_portfolio = Void then
				record.portfolio.to_upper
				record.portfolio.prune_all_trailing (' ')
				saved_portfolio := manager.reference_system.portfolio_from_code (record.portfolio)
			end
			Result := saved_portfolio
		ensure
			non_void_result: Result /= Void
		end

	broker: RP_BROKER is
			-- Broker.
		local
			l_broker: STRING
		do
			if saved_broker = Void then
				l_broker := record.broker
				if l_broker /= Void then
					l_broker.to_upper
					l_broker.prune_all_trailing (' ')
					saved_broker := manager.reference_system.broker_from_code (l_broker)
				end
			end
			Result := saved_broker
		end

	instrument: RP_INSTRUMENT is
			-- Instrument.
		do
			if instrument_saved = Void then
				record.instrument.to_upper
				record.instrument.prune_all_trailing (' ')
				instrument_saved := manager.reference_system.instrument_from_brimsym_at (record.instrument, record.event_date)
				if instrument_saved = Void then
					(create {EXCEPTIONS}).raise ("instrument not found from brimsym [" + record.instrument + "]")
				end
			end
			Result := instrument_saved
		ensure
			non_void_result: Result /= Void
		end

	code: INTEGER is
			-- Code
		do
			Result := record.event_type
		end

	catagory: STRING is
			-- Category
		do
			Result := record.catagory
			if Result = Void then
				create Result.make_empty
			end
		ensure
			catagory_not_void: Result /= Void
		end

	units_base: RP_QUANTITY is
			-- Units base
		do
			if units_base_saved = Void then
				units_base_saved := instrument.quantity (record.units_base)
			end
			Result := units_base_saved
		ensure
			non_void_result: Result /= Void
		end

	units: RP_QUANTITY is
			-- Units
		do
			if units_saved = Void then
				check
					instrument_not_void: instrument /= Void
				end
				units_saved := instrument.quantity (record.units)
			end
			Result := units_saved
		end

	secondary_cash_effect: RP_MONIES is
			-- Secondary cash effect
		do
			if (secondary_cash_effect_saved = Void) and then (secondary_currency /= Void) then
				secondary_cash_effect_saved := secondary_currency.monies (record.secondary_cash_effect)
			end
			Result := secondary_cash_effect_saved
		end

	currency: RP_CURRENCY is
			-- Currency
		local
			l_iso: STRING
		do
			if currency_saved = Void then
				create l_iso.make_from_string (record.currency)
				l_iso.right_adjust
				if l_iso.is_empty then
						-- Use the country from the BRIMSYM to retrieve the currency
					l_iso := record.instrument.substring (1, 2)
				end
				currency_saved := manager.reference_system.currency_from_iso_code (l_iso)
			end
			Result := currency_saved
		end

	secondary_currency: RP_CURRENCY is
			-- Secondary currency
		local
			l_iso: STRING
		do
			if secondary_currency_saved = Void then
				if record.secondary_currency /= Void then
					create l_iso.make_from_string (record.secondary_currency)
					l_iso.right_adjust
				else
					create l_iso.make_empty
				end
				if l_iso.is_empty then
						-- Use the second country from the BRIMSYM to retrieve the currency
					if record.instrument.item (3) = '/' then
						l_iso := record.instrument.substring (4, 5)
					end
				end
				secondary_currency_saved := manager.reference_system.currency_from_iso_code (l_iso)
			end
			Result := secondary_currency_saved
		end

	cash_effect_with_sign: RP_MONIES is
			-- Cash effect
		do
			if cash_effect_with_sign_saved = Void then
				cash_effect_with_sign_saved := currency.monies (record.cash_effect)
			end
			Result := cash_effect_with_sign_saved
		ensure
			non_void_result: Result /= Void
		end

	performance_effect_with_sign: RP_MONIES is
			--|
		do
			if performance_effect_with_sign_saved = Void then
				performance_effect_with_sign_saved := currency.monies (record.performance_effect)
			end
			Result := performance_effect_with_sign_saved
		end

	set_cost (a_value: like performance_effect_with_sign) is
			-- Set cost to `a_value' -- FIXME
			-- XXX this should go away when cost is implemented
		require
			non_void_arg: a_value /= Void
		do
			performance_effect_with_sign_saved := a_value
		ensure
			cost_set: performance_effect_with_sign_saved = a_value
		end

	commission: RP_MONIES is
			--|
		do
			if commission_saved = Void then
				commission_saved := currency.monies (record.commission.abs)
			end
			Result := commission_saved
		end

	fee: RP_MONIES is
			--|
		do
			if fee_saved = Void then
				fee_saved := currency.monies (record.fee.abs)
			end
			Result := fee_saved
		end

	price: RP_MONIES is
			--|
		do
			if price_saved = Void then
				price_saved := currency.monies (record.price)
			end
			Result := price_saved
		end

	event_details: STRING is
			--|
		do
			Result := record.event_details
			if Result = Void then
				create Result.make_empty
			end
		ensure
			event_details_not_void: Result /= Void
		end

	entered_at: ROSE_DATE is
			--|
		do
			Result := record.entered_at
		end

	entered_by: STRING is
			--|
		do
			Result := record.entered_by
		end

	modified_at: ROSE_DATE is
			--|
		do
			Result := record.modified_at
		end

	modified_by: STRING is
			--|
		do
			Result := record.modified_by
		end

	subaccount: STRING is
			--|
		do
			Result := record.subaccount
			if Result = Void then
				create Result.make_empty
			end
		ensure
			subaccount_not_void: Result /= Void
		end

	exchange: STRING is
			--|
		do
			Result := record.exchange
		end

	trader: STRING is
			--|
		do
			Result := record.trader
		end

	transaction_time: ROSE_DATE is
			--|
		do
			Result := record.transaction_time
		end

	soft_dollar: STRING is
			--|
			-- Tri state 'Y', 'N', ' '
		do
			Result := record.soft_dollar.out
		end

	rate: RP_DECIMAL is
			--|
		do
			Result := record.rate
		end

	from_instrument: STRING is
			--|
		do
			Result := record.from_instrument
		end

	is_divfx: BOOLEAN is
			-- Is the current event about cash dividends?
		local
			l_details: STRING
		do
			l_details := event_details.as_upper
			l_details.prune_all (' ')
			Result := l_details ~ "DIVFX"
		end

	is_REIT: BOOLEAN is
			-- Is the current event a `real estate investment trust'?
		local
			l_category: STRING
		do
			l_category := catagory.as_upper
			l_category.prune_all (' ')
			Result := l_category ~ "REIT"
		end

feature -- Output

	character_delimited_out (a_delimiter: CHARACTER): STRING is
			-- 'a_delimiter' delimited representation of 'Current'
		do
			Result := formatter.join_array (<<formatted_date (event_date), portfolio.code, icc, code, event_details, brimsym (instrument), quantity_amount (units), quantity_amount (fee), quantity_amount (commission), quantity_amount (cash_effect_with_sign), currency_name (currency), quantity_amount (secondary_cash_effect), currency_name (secondary_currency), quantity_amount (performance_effect_with_sign), broker_code (broker), formatted_date (settle_date), quantity_amount (units_base), catagory>>, a_delimiter.out)
		ensure
			result_not_void_or_empty: (Result /= Void) and then (not Result.is_empty)
		end

feature -- Debug Output

	debug_output: STRING is
			-- Debug output
		do
			Result := identifier
		end

feature {RP_TRADE, RP_GEDB_EVENT_TO_CONTRACT_CONVERTER} -- Implementation

	contract: !RP_PORTFOLIO_CONTRACT is
			-- Corresponding portfolio contract
		deferred
		end

feature {NONE} -- Implementation

	record: RP_GEDB_RECORD
			-- Underlying GEDB record

	saved_portfolio: like portfolio
			-- Cached `portfolio'

	saved_broker: like broker
			-- Cached `broker'

	instrument_saved: like instrument
			-- Cached `instrument'

	units_base_saved: like units_base
			-- Cached `units_base'

	units_saved: RP_QUANTITY
			-- Cached units

	secondary_cash_effect_saved: RP_MONIES
			-- Cached secondary cash effect

	currency_saved: RP_CURRENCY
			-- Cached currency

	secondary_currency_saved: RP_CURRENCY
			-- Cached secondary currency

	cash_effect_with_sign_saved: RP_MONIES
			-- Cached cash effect

	performance_effect_with_sign_saved: RP_MONIES
			-- cashed performance effect

	fee_saved: RP_MONIES
			-- cached fee

	commission_saved: RP_MONIES
			-- cached commission

	saved_contracts: ARRAY [RP_PORTFOLIO_CONTRACT]
			-- cached contracts

	price_saved: RP_MONIES
			-- Caches price

feature {NONE} -- Implementation

	quantity_amount (a_monies: RP_QUANTITY): RP_DECIMAL is
			-- Decimal quantity in 'a_monies'
		do
			if a_monies /= Void then
				Result := a_monies.amount
			else
				Result := 0
			end
		end

	monies_currency_name (a_monies: RP_MONIES): STRING is
			-- Currency iso code for 'a_monies'
		do
			if (a_monies /= Void) then
				Result := currency_name (a_monies.currency)
			end
		end

	currency_name (a_currency: RP_CURRENCY): STRING is
			-- Currency name
		do
			if a_currency /= Void then
				Result := currency.name
			end
		end

	broker_code (a_broker: RP_BROKER): STRING is
			-- 'a_broker' code
		do
			if a_broker /= Void and then not a_broker.is_unknown then
				Result := a_broker.code
			end
		end

	brimsym (a_instrument: RP_INSTRUMENT): STRING is
			-- 'a_instrument' brimsym code
		do
			if a_instrument /= Void then
				Result := instrument.brimsym
			end
		end

	formatted_date (a_date: ROSE_DATE): STRING is
			--|
		do
			if a_date /= Void then
				Result := a_date.ccyymmdd.out
			end
		end

invariant

	non_void_record_and_identifier: record /= Void and identifier /= Void
	valid_event_type: is_valid_event_type (code)

end
