note

	description:

		"Ordered sequence of non-overlapping intervals per item"

	author: "Ivan Tarnavsky"
	last_editor: "$Author: it $"
	date: "$Date: 2011/03/30 $"
	revision: "$Revision: #6 $"
	path: "$File: //rose/source/kernel/expert_system/rp_interval_list.e $"

class RP_INTERVAL_LIST [H -> HASH_KEYED [INTEGER_64], G -> RP_INTERVAL [H]]

create

	make_default

feature {NONE} -- Initialization

	make_default is
			-- Initialize `Current'
		do
			create intervals.make (create {KL_COMPARABLE_COMPARATOR [INTEGER_64]}.make)
			create keys.make (create {KL_COMPARABLE_COMPARATOR [INTEGER_64]}.make)
		end

feature -- Change

	append (a_interval: G) is
			-- Append `a_interval'.
		require
			a_interval_not_void: a_interval /= Void
		local
			l_tree: DS_AVL_TREE [G, ROSE_DATE]
		do
			l_tree := intervals.value (a_interval.entity.key)
			if l_tree = Void then
				create l_tree.make (create {KL_COMPARABLE_COMPARATOR [ROSE_DATE]}.make)
				intervals.force (l_tree, a_interval.entity.key)
				keys.force (a_interval.entity, a_interval.entity.key)
			end
			l_tree.force (a_interval, a_interval.start_date)
		end

	delete (a_interval: G) is
			-- Remove `a_interval'.
		require
			a_interval_not_void: a_interval /= Void
		local
			l_tree: DS_AVL_TREE [G, ROSE_DATE]
		do
			l_tree := intervals.value (a_interval.entity.key)
			if l_tree /= Void then
				l_tree.remove (a_interval.start_date)
			end
		end

feature -- Access

	min_start_date: ROSE_DATE is
			-- Minimum start date
		local
			l_result: CELL [ROSE_DATE]
		do
			create l_result.put (Void)
			intervals.do_all (agent min_date (l_result, ?))
			Result := l_result.item
		end

feature -- Query

	witness_intervals (a_date: ROSE_DATE): ARRAY [G] is
			-- Intervals bracketing `a_date'
		require
			a_date_not_void: a_date /= Void
		local
			l_list: DS_LINKED_LIST [G]
		do
			create l_list.make_default
			intervals.do_all (agent bracketing_intervals (l_list, a_date, ?))
			Result := l_list.to_array
		ensure
			non_void_witness_array: Result /= Void
		end

	entity_intervals (a_entity: H): ARRAY [G] is
			-- All intervals for `a_entity'
		require
			a_entity_not_void: a_entity /= Void
		local
			l_tree: DS_AVL_TREE [G, ROSE_DATE]
		do
			l_tree := intervals.value (a_entity.key)
			if l_tree /= Void then
				Result := l_tree.to_array
			else
				Result := <<>>
			end
		ensure
			non_void_intervals_array: Result /= Void
		end

	entities: ARRAY [H] is
			-- All entities
		do
			Result := keys.to_array
		end

feature -- Removal

	wipe_out is
			-- Remove all items from container.
		do
			intervals.wipe_out
		end

feature -- Iteration

	do_all (an_action: PROCEDURE [ANY, TUPLE [G]]) is
			-- Apply `an_action' to every item, from first to last.
			-- (Semantics not guaranteed if `an_action' changes the structure.)
		require
			an_action_not_void: an_action /= Void
		do
			intervals.do_all (agent {DS_AVL_TREE [G, ROSE_DATE]}.do_all (an_action))
		end

	do_if (an_action: PROCEDURE [ANY, TUPLE [G]]; a_test: FUNCTION [ANY, TUPLE [G], BOOLEAN]) is
			-- Apply `an_action' to every item that satisfies `a_test', from first to last.
			-- (Semantics not guaranteed if `an_action' or `a_test' change the structure.)
		require
			an_action_not_void: an_action /= Void
			a_test_not_void: a_test /= Void
		do
			intervals.do_all (agent {DS_AVL_TREE [G, ROSE_DATE]}.do_if (an_action, a_test))
		end

feature {RP_INTERVAL_LIST} -- Implementation

	keys: DS_AVL_TREE [H, INTEGER_64]
			-- Underlying entities

	intervals: DS_AVL_TREE [DS_AVL_TREE [G, ROSE_DATE], INTEGER_64]
			-- Underlying interval storage

	min_date (a_min: CELL [ROSE_DATE]; a_tree: DS_AVL_TREE [G, ROSE_DATE]) is
			-- Keep minimum of `a_min.item' and `a_tree.first.start_date' in `a_min'
		require
			a_min_not_void: a_min /= Void
			a_tree_not_void: a_tree /= Void
		do
			if not a_tree.is_empty then
				if a_min.item = Void then
					a_min.put (a_tree.first.start_date)
				elseif a_min.item > a_tree.first.start_date then
					a_min.put (a_tree.first.start_date)
				end
			end
		end

	bracketing_intervals (a_list: DS_LINKED_LIST [G]; a_date: ROSE_DATE; a_tree: DS_AVL_TREE [G, ROSE_DATE]) is
			-- Find intervals from `a_tree' which bracket `a_date' and store in `a_list'
		require
			a_list_not_void: a_list /= Void
			a_date_not_void: a_date /= Void
			a_tree_not_void: a_tree /= Void
		do
			a_tree.do_if (agent a_list.force_last (?), agent interval_brackets (?, a_date))
		end

	interval_brackets (a_interval: G; a_date: ROSE_DATE): BOOLEAN is
			-- Does `a_interval' bracket `a_date'?
		require
			a_interval_not_void: a_interval /= Void
			a_date_not_void: a_date /= Void
		do
			Result := a_interval.brackets (a_date)
		end

invariant

	intervals_not_void: intervals /= Void
	keys_not_void: keys /= Void

end
