indexing
	description: "Investment account"
	author: "Ivan Tarnavsky"
	last_editor: "$Author: ext-asurelia $"
	date: "$Date: 2014/07/09 $"
	revision: "$Revision: #63 $"
	path: "$File: //rose/source/kernel/expert_system/rp_investment_account.e $"

class
	RP_INVESTMENT_ACCOUNT

inherit
	RP_ZONAL_INVESTMENT_GROUP

create {RP_REFERENCE_SYSTEM}
	make

feature {NONE} -- Initialization

	make (a_ref: like reference_system; a_account: GCDB_ACCOUNT; a_portfolio_group: GCDB_PORTFOLIO_GROUP; a_fund_participant: GCDB_FUND_PARTICIPANT) is
			-- Initialize from GCDB.
		require
			a_ref_not_void: a_ref /= Void
			account_and_portfolio_group_may_be_void: True
		do
			reference_system := a_ref
			gcdb_account := a_account
			gcdb_portfolio_group := a_portfolio_group
			gcdb_fund_participant := a_fund_participant
			if gcdb_portfolio_group /= Void then
				start_date := gcdb_portfolio_group.start_date
				end_date := gcdb_portfolio_group.end_date
				set_code (gcdb_portfolio_group.code)
			elseif gcdb_account /= Void then
				start_date := gcdb_account.start_date
				end_date := gcdb_account.end_date
				set_code (gcdb_account.code)
			elseif gcdb_fund_participant /= Void then
				start_date := reference_system.fund_participant_start_date (gcdb_fund_participant.brimsym)
				end_date := reference_system.fund_participant_end_date (gcdb_fund_participant.brimsym)
				set_code (gcdb_fund_participant.code)
			else
				set_code ({RP_UNKNOWN}.code)
				is_unknown := True
			end
			if start_date = Void then
				create start_date.make_ccyymmdd (99991231)
			end
			if end_date = Void then
				create end_date.make_ccyymmdd (99991231)
			end
			if not portfolios.is_empty then
				start_date := start_date.max (portfolios.min_start_date)
				end_date := end_date.min (portfolios.max_end_date)
			end
			if end_date < start_date then
				create end_date.make_ccyymmdd (start_date.ccyymmdd)
			end
		ensure
			valid_code: code /= Void
			reference_system_set: reference_system = a_ref
			start_date_not_void: start_date /= Void
			end_date_not_void: end_date /= Void
		end

feature -- Access

	code: STRING
			-- Unique code

	name: STRING assign set_code is
			-- FIXME: to be removed
		do
			Result := code
		ensure then
			code: Result = code
		end

	start_date: ROSE_DATE
			-- Start date of investment account

	end_date: ROSE_DATE
			-- End date of investment account

	last_known_name: STRING is
			-- Last known descriptive name
		do
			if gcdb_portfolio_group /= Void then
				Result := gcdb_portfolio_group.last_known_name
			elseif gcdb_account /= Void then
				Result := gcdb_account.last_known_name
			elseif gcdb_fund_participant /= Void then
				Result := gcdb_fund_participant.last_known_name
			else
				Result := ""
			end
		end

	gcdb_handle: INTEGER is
			-- Underlying GCDB entity handle
		do
			if gcdb_portfolio_group /= Void then
				Result := gcdb_portfolio_group.handle
			elseif gcdb_account /= Void then
				Result := gcdb_account.handle
			elseif gcdb_fund_participant /= Void then
				Result := gcdb_fund_participant.handle
			end
		end

feature -- Status Report

	is_unknown: BOOLEAN
			-- Is `Current' an instance of a void investment account?

feature -- Change

	set_code (a_code: like code) is
			-- <Precursor>
		do
			code := a_code
		end

feature -- Comparison

	hash_code: INTEGER is
			-- Hash code
		do
			Result := name.hash_code
		end

feature -- Ouput

	debug_output: STRING is
			-- String that should be displayed in debugger to represent `Current'.
		do
			Result := name
		end

feature -- Queries

	performance_code: STRING is
			-- <Precursor>
		do
			Result := code
		end

	imc_client_service_name: STRING is
			-- Client service rep for this account
		do
			if gcdb_account /= Void and then gcdb_account.imc_client_services /= Void then
				Result := gcdb_account.imc_client_services.name
			end
		end

	client: RP_CLIENT is
			-- Client associated with `Current'
		do
			if client_saved = Void then
				if gcdb_account /= Void and then gcdb_account.client /= Void then
					client_saved := reference_system.client_by_code (gcdb_account.client.code)
				elseif gcdb_fund_participant /= Void and then gcdb_fund_participant.client /= Void then
					client_saved := reference_system.client_by_code (gcdb_fund_participant.client.code)
				else
					client_saved := reference_system.unknown_client
				end
			end
			Result := client_saved
		ensure
			non_void_result: Result /= Void
		end

	portfolios: RP_PORTFOLIO_HISTORY is
			-- All portfolios in `Current'
		do
			Result := retrieve_portfolios
		ensure
			non_void_result: Result /= Void
		end

	bank_accounts: ROSE_LINEAR_ARRAY [RP_BANK_ACCOUNT] is
			-- Bank accounts associated with `Current'
		do
			Result := retrieve_bank_accounts
		ensure
			non_void_result: Result /= Void
		end

	investment_region: RP_ZONE is
			-- Investment region associated with `Current' (zone equivalent)
		do
			Result := retrieve_investment_region
		end

	benchmarks: !RP_BENCHMARK_HISTORY is
			-- All benchmarks associated with the portfolios in `Current'
		do
			Result := retrieve_benchmark_history
		end

	hedged_benchmarks: !RP_BENCHMARK_HISTORY is
			-- All hedged benchmarks associated with the portfolios in `Current'
		do
			Result := retrieve_hedged_benchmark_history
		end

	non_hedged_benchmarks: !RP_BENCHMARK_HISTORY is
			-- All non-hedged benchmarks associated with the portfolios in `Current'
		do
			Result := retrieve_non_hedged_benchmark_history
		end

	has_mapped_hedged_benchmark (a_date: ROSE_DATE): BOOLEAN is
			-- Is a hedged benchmark mapped to `Current' on date `a_date'.
		require
			a_date_not_void: a_date /= Void
		do
			Result := hedged_benchmarks.witness (a_date) /= Void
		end

	has_mapped_benchmark (a_date: ROSE_DATE): BOOLEAN is
			-- Is a benchmark mapped to `Current' on date `a_date'.
		require
			a_date_not_void: a_date /= Void
		do
			Result := benchmarks.witness (a_date) /= Void
		end

	last_mapped_hedged_benchmark_date (a_date: ROSE_DATE): ROSE_DATE is
			-- Last mapping date for a hedged benchmark to `Current', before `a_date'.
		require
			not_has_mapped_hedged_benchmark: not has_mapped_hedged_benchmark (a_date)
		do
			if {l_interval: RP_BENCHMARK_HISTORY_INTERVAL} hedged_benchmarks.most_recent (a_date) then
				Result := l_interval.end_date
			end
		end

	last_mapped_benchmark_date (a_date: ROSE_DATE): ROSE_DATE is
			-- Last mapping date for a benchmark to `Current', before `a_date'.
		require
			not_has_mapped_benchmark: not has_mapped_benchmark (a_date)
		do
			Result := benchmarks.most_recent (a_date).end_date
		end

	composite_state: RP_BOOLEAN_HISTORY is
			-- Composite inclusion state of `Current'
		do
			Result := retrieve_composite_state
		ensure
			non_void_result: Result /= Void
		end

	base_currency: RP_CURRENCY is
			-- Base currency
		do
			Result := gcdb_base_currency
			if Result = Void then
				Result := reference_system.currency_from_iso_code ("USD")
			end
		end

	fx_base_currency: RP_CURRENCY is
			-- Base currency
		do
			if gcdb_account /= Void and then gcdb_account.fx_base_currency /= Void then
				Result := reference_system.currency_from_iso_code (gcdb_account.fx_base_currency.code)
			end
			if Result = Void then
				Result := base_currency
			end
		end

	reporting_currency: RP_CURRENCY
			-- Reporting currency
		do
			Result := gcdb_reporting_currency
			if Result = Void then
				Result := base_currency
			end
		end

	gcdb_base_currency: RP_CURRENCY is
			-- Actual base currency as defined in GCDB, not defaulting
		do
			if gcdb_account /= Void and then gcdb_account.base_currency /= Void then
				Result := reference_system.currency_from_iso_code (gcdb_account.base_currency.code)
			end
		end

	gcdb_reporting_currency: RP_CURRENCY
			-- Actual reporting currency as defined in GCDB, no defaulting
		local
			l_history: ROSE_ENTITY_HISTORY [GCDB_CURRENCY]
			l_interval: ENTITY_INTERVAL [GCDB_CURRENCY]
			l_today: ROSE_DATE
		do
			if saved_reporting_currency = Void then
				if gcdb_account /= Void then
					l_history := gcdb_account.report_currencies_history
					if l_history /= Void then
						create l_today.make_today
						l_interval := l_history.most_recent (l_today)
						if l_interval /= Void and then l_interval.entity /= Void and then l_interval.entity.code /= Void then
							saved_reporting_currency := reference_system.currency_from_iso_code (l_interval.entity.code)
						end
					end
				elseif gcdb_portfolio_group /= Void then
					saved_reporting_currency := reference_system.currency_from_iso_code (gcdb_portfolio_group.report_currency.code)
				end
			end
			Result := saved_reporting_currency
		end

	base_currencies: RP_CURRENCY_HISTORY is
			-- Base currencies associated with `Current'
		do
			Result := retrieve_base_currency_history
		ensure
			non_void_result: Result /= Void
		end

	reporting_currencies: ROSE_LINEAR_ARRAY [RP_CURRENCY] is
			-- Reporting currencies associated with `Current'
		do
			Result := retrieve_reporting_currencies
		ensure
			non_void_result: Result /= Void
		end

	first_performance_date: ROSE_DATE is
			-- First date of performance calculation
		do
			if {l_date: ROSE_DATE} benchmarks.min_start_date then
				Result := l_date
			else
				create Result.make_max_date
			end
			Result := investment_region.inception_date.max (Result)
			Result := Result.max (start_date)
		ensure
			non_void_date: Result /= Void
		end

	last_performance_date: ROSE_DATE is
			-- Last date of performance calculation
		do
			if {l_date: ROSE_DATE} benchmarks.max_end_date then
				Result := l_date
			else
				create Result.make_max_date
			end
			Result := Result.min (end_date)
		ensure
			non_void_date: Result /= Void
		end

feature -- Action

	do_all_at (a_action: PROCEDURE [ANY, TUPLE [RP_PORTFOLIO, ROSE_DATE]]; a_date: ROSE_DATE) is
			-- Execute 'a_action' for all portfolios in current investment account at 'a_date
		require
			action_not_Void: a_action /= Void
			date_not_void: a_date /= Void
		local
			l_portfolios: RP_PORTFOLIO_HISTORY
			l_portfolio_interval: RP_PORTFOLIO_HISTORY_INTERVAL
		do
			l_portfolios := portfolios
			if l_portfolios /= Void then
				from
					l_portfolios.start
				until
					l_portfolios.after
				loop
					l_portfolio_interval := l_portfolios.item_for_iteration
					if l_portfolio_interval /= Void and then l_portfolio_interval.brackets (a_date) then
						a_action.call ([l_portfolio_interval.portfolio, a_date])
					end
					l_portfolios.forth
				end
			end
		end

feature -- Status Report

	is_fund_participant: BOOLEAN is
			-- Is this account invested in pooled funds only?
			-- Note: Currently user set not calculated (d3 Feb 2009)
		do
			if gcdb_account /= Void then
				Result := gcdb_account.is_fund_participant
			elseif gcdb_fund_participant /= Void then
				Result := True
			end
		end

	is_segregate: BOOLEAN is
			-- Is `Current' segregate account?
		do
				-- Client/FundParticipants are not segregate
			Result := gcdb_fund_participant = Void
		end

	is_global: BOOLEAN is
			-- Is `Current' global?
		local
			l_ports: like portfolios
			l_port: RP_PORTFOLIO
		do
			if gcdb_portfolio_group /= Void then
				Result := True
			elseif gcdb_fund_participant = Void then
				l_ports := portfolios
				if not l_ports.is_empty then
					from
						l_ports.start
						if l_ports.current_item.portfolio.zone.code.is_equal ("T") then
								-- account has T zone portfolio
							Result := True
						end
						l_port := l_ports.current_item.portfolio
						l_ports.forth
					until
						l_ports.after or Result
					loop
						if l_ports.current_item.portfolio.zone.code.is_equal ("T") then
								-- account has T zone portfolio
							Result := True
						else
								-- account has portfolios in different zones
							Result := not l_port.zone.code.is_equal (l_ports.current_item.portfolio.zone.code)
						end
						l_port := l_ports.current_item.portfolio
						l_ports.forth
					end
				end
			end
		end

	is_gpc: BOOLEAN is
			-- Is `Current' gpc?
		do
			if gcdb_account /= Void then
				Result := gcdb_account.is_gpc_account
			end
		end

	is_hedged: BOOLEAN is
			-- Is `Current' hedged?
		do
			Result := not hedged_benchmarks.is_empty
		end

	investment_region_as_of (a_date: !ROSE_DATE): RP_ZONE is
			-- investment_region as of given date
		local
			p: ROSE_LINEAR_ARRAY [RP_PORTFOLIO]
			l_has_T: BOOLEAN
			l_zone_determined: BOOLEAN
		do
			p := portfolios.all_witnesses (a_date)
			if p /= Void and then p.count > 0 then
				from
					p.start
				until
					p.off or l_zone_determined
				loop
					if not p.current_item.zone.is_unknown then
						if p.current_item.zone.code.same_string ("T") then
							l_has_T := True
						elseif Result = Void then
							Result := p.current_item.zone
						elseif not Result.code.same_string (p.current_item.zone.code) then
							Result := reference_system.investment_region_by_code ("T")
							l_zone_determined := True
						end
					end
					p.forth
				end
			end
			if not l_zone_determined and Result = Void then
				if l_has_T then
					Result := reference_system.investment_region_by_code ("T")
				else
					Result := reference_system.unknown_investment_region
				end
			end
		ensure
			result_not_void: Result /= Void
		end

	is_carve_out: BOOLEAN is
			-- Is `Current' a carve out?
		do
			if gcdb_portfolio_group /= Void then
				Result := gcdb_portfolio_group.is_carve_out
			end
		end

	is_not_in_olap: BOOLEAN is
			-- Does `Current' include any portfolios in its history which
			-- are classified as "not in OLAP" by RP_PORTFOLIO?
		local
			l_portfolios: like portfolios
			l_portfolio: RP_PORTFOLIO
		do
			if gcdb_fund_participant /= Void then
					-- Exclude Fund Participants
				Result := True
			elseif gcdb_portfolio_group /= Void and gcdb_account = Void then
					-- Exclude carvouts and fake accounts
					-- i.e. there is a portfolio group but no account
					-- of the same name
				Result := True
			else
					-- Check whether any portfolio
					-- is excluded from olap
				l_portfolios := portfolios
				from
					Result := True
					l_portfolios.start
				until
					l_portfolios.off or not Result
				loop
					l_portfolio := l_portfolios.current_item.portfolio
					if l_portfolio /= Void then
						Result := Result and l_portfolio.is_not_in_olap
					end
					l_portfolios.forth
				end
			end
		end

	is_portfolio_group: BOOLEAN is
			-- Is portfolio group?
		do
			Result := (gcdb_portfolio_group /= Void)
		end

	is_long_short: BOOLEAN is
			-- Does `Current' include any portfolios in its history which
			-- are classified as long_short RP_PORTFOLIOs?
		local
			l_portfolios: like portfolios
			l_portfolio: RP_PORTFOLIO
		do
			l_portfolios := portfolios
			from
				l_portfolios.start
			until
				l_portfolios.off or Result
			loop
				l_portfolio := l_portfolios.current_item.portfolio
				if l_portfolio /= Void then
					Result := l_portfolio.is_long_short
				end
				l_portfolios.forth
			end
		end

	is_limited_long_short: BOOLEAN is
			-- Does `Current' include any portfolios in its history which
			-- are classified as limited long_short RP_PORTFOLIOs?
		local
			l_portfolios: like portfolios
			l_portfolio: RP_PORTFOLIO
		do
			l_portfolios := portfolios
			from
				l_portfolios.start
			until
				l_portfolios.off or Result
			loop
				l_portfolio := l_portfolios.current_item.portfolio
				if l_portfolio /= Void then
					Result := l_portfolio.is_limited_long_short
				end
				l_portfolios.forth
			end
		end

	is_not_equity: BOOLEAN is
			-- Does `Current' include any portfolios in its history which
			-- are not classified as equity RP_PORTFOLIOs?
		local
			l_portfolios: like portfolios
			l_portfolio: RP_PORTFOLIO
		do
			l_portfolios := portfolios
			from
				l_portfolios.start
			until
				l_portfolios.off or Result
			loop
				l_portfolio := l_portfolios.current_item.portfolio
				if l_portfolio /= Void then
					Result := not l_portfolio.is_equity
				end
				l_portfolios.forth
			end
		end

	requires_nof_reporting: BOOLEAN is
			-- Is NOF reporting required?
		do
			if gcdb_account /= Void then
				Result := gcdb_account.nof_perf
			end
		end

	client_service: RP_INVESTMENT_MANAGEMENT_COMPANY is
			-- Client services IMC
		do
			if gcdb_account /= Void and then not gcdb_account.imc_client_services_history.is_empty then
				Result := reference_system.imc_by_code (gcdb_account.imc_client_services_history.first.entity.code)
			else
				Result := reference_system.unknown_imc
			end
		ensure
			non_void_result: Result /= Void
		end

	invoicing: RP_INVESTMENT_MANAGEMENT_COMPANY is
			-- Invoicing IMC
		do
			if gcdb_account /= Void and then not gcdb_account.imc_accounting_history.is_empty then
				Result := reference_system.imc_by_code (gcdb_account.imc_accounting_history.first.entity.code)
			else
				Result := reference_system.unknown_imc
			end
		ensure
			non_void_result: Result /= Void
		end

	marketing: RP_INVESTMENT_MANAGEMENT_COMPANY is
			-- Marketing IMC
		do
			if gcdb_account /= Void and then not gcdb_account.imc_marketing_history.is_empty then
				Result := reference_system.imc_by_code (gcdb_account.imc_marketing_history.first.entity.code)
			else
				Result := reference_system.unknown_imc
			end
		ensure
			non_void_result: Result /= Void
		end

	products (as_of_date: ROSE_DATE): ROSE_LINEAR_ARRAY [RP_PRODUCT] is
			-- List of products mapped directly or thru benchmark as of a given date
		local
			b: RP_BENCHMARK
		do
			create Result.make_empty
			b := benchmarks.witness (as_of_date)
			if b /= Void then
				Result := b.products
			end
		ensure
			non_void_products: Result /= Void
		end

feature {NONE} -- Implementation

	gcdb_account: GCDB_ACCOUNT
			-- GCDB account if current is based on account

	gcdb_portfolio_group: GCDB_PORTFOLIO_GROUP
			-- GCDB portfolio group if current is based on portfolio group. Should we have an investment account based on a portfolio group? Seem wrong.

	gcdb_fund_participant: GCDB_FUND_PARTICIPANT
			-- Underlying fund participant entity

	reference_system: RP_REFERENCE_SYSTEM
			-- Reference system

	portfolios_saved: like portfolios
			-- Cached version of portfolios associated with  `Current' investment account

	saved_reporting_currency: RP_CURRENCY
			-- Saved reporting currency

	retrieve_portfolios: like portfolios is
			-- Portfolios associated with `Current' investment account
		local
			l_fund_part: RP_PORTFOLIO
		do
			if portfolios_saved = Void then
				create portfolios_saved.make_empty
				if gcdb_portfolio_group /= Void and then not gcdb_portfolio_group.portfolios_history.is_empty then
					gcdb_portfolio_group.portfolios_history.do_all (agent expand_portfolio (portfolios_saved, ?))
				elseif gcdb_account /= Void then
					gcdb_account.portfolios_by_account_history.do_all (agent expand_portfolio (portfolios_saved, ?))
				elseif gcdb_fund_participant /= Void then
						-- Fund participants are both accounts and portfolios
					l_fund_part := reference_system.portfolio_from_code (code)
					add_portfolio_interval (portfolios_saved, l_fund_part, l_fund_part.start_date, l_fund_part.end_date)
				end
			end
			Result := portfolios_saved
		ensure
			non_void_result: Result /= Void
			result_cached: portfolios_saved /= Void
		end

	expand_portfolio (a_array: like portfolios; a_portfolio_interval: ENTITY_INTERVAL [GCDB_PORTFOLIO]) is
			-- Add portfolio interval to `a_array'
		require
			valid_interval: a_portfolio_interval /= Void
			valid_array: a_array /= Void
		local
			l_portfolio: RP_PORTFOLIO
		do
			l_portfolio := reference_system.portfolio_from_code (a_portfolio_interval.entity.code)
			if not l_portfolio.is_unknown then
					-- XXX NOTE we add one day to the end
					-- XXX because .GLOBALFILES use [closed,closed] interval
				add_portfolio_interval (a_array, l_portfolio, a_portfolio_interval.start_date, a_portfolio_interval.end_date.next_date)
			end
		end

	add_portfolio_interval (a_array: like portfolios; a_portfolio: RP_PORTFOLIO; a_start_date, a_end_date: ROSE_DATE) is
			-- Add interval for `a_portfolio' to `a_array'
		local
			l_interval: RP_PORTFOLIO_HISTORY_INTERVAL
			l_start, l_end: ROSE_DATE
		do
			l_start := a_start_date.max (a_portfolio.start_date).max (start_date)
			l_end := a_end_date.min (a_portfolio.end_date.next_date).min (end_date.next_date)
			l_end := l_end.min (create {ROSE_DATE}.make_ccyymmdd (99991231))
			if l_start < l_end then
				create l_interval.make (a_portfolio, l_start, l_end, 0)
				a_array.extend (l_interval)
			end
		end

	client_saved: like client
			-- Cached version of associated client

	benchmark_history_saved: RP_BENCHMARK_HISTORY
			-- Cached version of associated benchmarks

	hedged_benchmark_history_saved: RP_BENCHMARK_HISTORY
			-- Cached version of associated hedged benchmarks

	non_hedged_benchmark_history_saved: RP_BENCHMARK_HISTORY
			-- Cached version of associated non-hedged benchmarks

	saved_fund_participant_composite_asset: like fund_participant_composite_asset
			-- Underlying pool fund

	fund_participant_composite_asset: RP_POOL_FUND is
			-- Pool fund for fund participant
		local
			l_instrument: RP_INSTRUMENT
		do
			if saved_fund_participant_composite_asset = Void then
				l_instrument := reference_system.instrument_from_brimsym (gcdb_fund_participant.brimsym)
				if {l_pool_fund: RP_POOL_FUND} l_instrument then
					saved_fund_participant_composite_asset := l_pool_fund
				end
			end
			Result := saved_fund_participant_composite_asset
		end

	retrieve_benchmark_history: like benchmarks is
			-- History of benchmarks associated with portfolios in `Current'
		do
			if {l_benchmarks: like benchmarks} benchmark_history_saved then
				Result := l_benchmarks
			else
				create Result.make_empty
				if gcdb_portfolio_group /= Void then
					gcdb_portfolio_group.benchmarks_by_portfolio_group_history.do_all (agent expand_benchmark_history (Result, ?))
				elseif gcdb_account /= Void then
					gcdb_account.benchmark_history.do_all (agent expand_benchmark_history (Result, ?))
				elseif gcdb_fund_participant /= Void then
					if fund_participant_composite_asset.portfolio.zone.code.is_equal ("T") then
						fund_participant_composite_asset.portfolio.account.benchmarks.do_all (agent add_to_benchmark_history (Result, ?))
					else
						fund_participant_composite_asset.portfolio.benchmark_history.do_all (agent add_to_benchmark_history (Result, ?))
					end
				end
				benchmark_history_saved := Result
			end
		end

	retrieve_hedged_benchmark_history: like benchmarks is
			-- History of benchmarks associated with portfolios in `Current'
		do
			if {l_benchmarks: like benchmarks} hedged_benchmark_history_saved then
				Result := l_benchmarks
			else
				create Result.make_empty
				retrieve_benchmark_history.do_if (agent Result.extend (?), agent {RP_BENCHMARK_HISTORY_INTERVAL}.is_hedged)
				hedged_benchmark_history_saved := Result
			end
		end

	retrieve_non_hedged_benchmark_history: like benchmarks is
			-- History of benchmarks associated with portfolios in `Current'
		do
			if {l_benchmarks: like benchmarks} non_hedged_benchmark_history_saved then
				Result := l_benchmarks
			else
				create Result.make_empty
				retrieve_benchmark_history.do_if (agent Result.extend (?), agent {RP_BENCHMARK_HISTORY_INTERVAL}.is_non_hedged)
				non_hedged_benchmark_history_saved := Result
			end
		end

	expand_benchmark_history (a_history: like benchmarks; a_interval: ENTITY_INTERVAL [GCDB_BENCHMARK]) is
			--Extend 'a_history' with 'a_interval' if the start date is strictly before then end date.
			-- The extended interval weight is set to 'a_interval.weight'
		require
			non_void_args: a_history /= Void and a_interval /= Void
		local
			l_bogie: RP_BENCHMARK
			l_interval: RP_BENCHMARK_HISTORY_INTERVAL
			l_start, l_end: ROSE_DATE
		do
			l_bogie := reference_system.benchmark_by_code (a_interval.entity.code)
			l_start := a_interval.start_date.max (start_date)
			l_end := a_interval.end_date.min (end_date)
			if l_start < l_end then
				create l_interval.make (l_bogie, l_start, l_end, a_interval.weight)
				a_history.extend (l_interval)
			end
		end

	add_to_benchmark_history (a_history: like benchmarks; a_interval: RP_BENCHMARK_HISTORY_INTERVAL) is
			--Extend 'a_history' with 'a_interval' if the start date is strictly before then end date.
			-- The extended interval weight is set to 'a_interval.weight'
		require
			non_void_args: a_history /= Void and a_interval /= Void
		local
			l_interval: RP_BENCHMARK_HISTORY_INTERVAL
			l_start, l_end: ROSE_DATE
		do
			l_start := a_interval.start_date.max (start_date)
			l_end := a_interval.end_date.min (end_date)
			if l_start < l_end then
				create l_interval.make (a_interval.benchmark, l_start, l_end, a_interval.weight)
				a_history.extend (l_interval)
			end
		end

	cached_investment_region: like investment_region
			-- Cached version of associated investment region (zone equivalent)

	retrieve_investment_region: like investment_region is
			-- Associated investment region for `Current' investment account
		do
			if cached_investment_region = Void then
				if is_unknown then
					cached_investment_region := reference_system.unknown_investment_region
				else
					if is_global then
						cached_investment_region := reference_system.investment_region_by_code ("T")
					else
						if portfolios.count > 0 then
							cached_investment_region := portfolios.item (1).portfolio.zone
						else
							cached_investment_region := reference_system.unknown_investment_region
						end
					end
				end
			end
			Result := cached_investment_region
		ensure
			result_not_void: Result /= Void
		end

	cached_base_currency_history: like base_currencies
			-- Cache of currency history intervals

	retrieve_base_currency_history: like base_currencies is
			-- Retrieve base currency
		do
			if cached_base_currency_history = Void then
				create cached_base_currency_history.make_empty
				if not is_unknown then
					if gcdb_portfolio_group /= Void then
							--- FIXME gcdb_portfolio_group.re
					end
					if cached_base_currency_history.is_empty and gcdb_account /= Void then
						gcdb_account.base_currency_history.do_all (agent add_base_currency (cached_base_currency_history, ?))
					end
				end
					-- fallback to investment region base currency
				if cached_base_currency_history.is_empty then
					cached_base_currency_history.extend (create {RP_CURRENCY_HISTORY_INTERVAL}.make (investment_region.legacy_base_currency, start_date, end_date, 1))
				end
			end
			Result := cached_base_currency_history
		ensure
			result_not_empty: Result /= Void and then not Result.is_empty
		end

	add_base_currency (a_hist: like base_currencies; a_interval: ENTITY_INTERVAL [GCDB_CURRENCY]) is
			-- Add base currency to `a_hist'
		require
			valid_hist: a_hist /= Void
			valid_interval: a_interval /= Void
		local
			l_interval: RP_CURRENCY_HISTORY_INTERVAL
		do
			create l_interval.make (reference_system.currency_from_iso_code (a_interval.entity.code), a_interval.start_date, a_interval.end_date, a_interval.weight)
			a_hist.extend (l_interval)
		end

	cached_reporting_currencies: like reporting_currencies
			-- Cached version of associated reporting currencies

	retrieve_reporting_currencies: like reporting_currencies is
			-- Associated reportting currencies for `Current' investment account
		local
			l_ccy: HASHED_SET [RP_CURRENCY]
		do
			if cached_reporting_currencies = Void then
				create cached_reporting_currencies.make_empty
				create l_ccy.make
				l_ccy.extend (investment_region.legacy_base_currency)
				base_currencies.do_all (agent add_base_currency_to_report_currencies (l_ccy, ?))
				if gcdb_account /= Void then
					gcdb_account.report_currencies_history.do_all (agent add_reporting_currency (l_ccy, ?))
				end
				if gcdb_portfolio_group /= Void then
					gcdb_portfolio_group.report_currency_history.do_all (agent add_reporting_currency (l_ccy, ?))
				end
				l_ccy.current_keys.do_all (agent cached_reporting_currencies.extend (?))
			end
			Result := cached_reporting_currencies
		end

	add_base_currency_to_report_currencies (a_set: HASHED_SET [RP_CURRENCY]; a_interval: RP_CURRENCY_HISTORY_INTERVAL) is
			-- Force 'a_interval' currency into 'a_set'
		do
			a_set.force (a_interval.currency)
		end

	add_reporting_currency (a_set: HASHED_SET [RP_CURRENCY]; a_interval: ENTITY_INTERVAL [GCDB_CURRENCY]) is
			-- Force the reference system currency that has the same iso code than interval entity iso code
			-- into 'a_set'
		local
			l_ccy: RP_CURRENCY
		do
			l_ccy := reference_system.currency_from_iso_code (a_interval.entity.code)
			a_set.force (l_ccy)
		end

	bank_accounts_saved: like bank_accounts
			-- Cached version of bank accounts

	retrieve_bank_accounts: like bank_accounts is
			-- Associated bank accounts
		do
			if bank_accounts_saved = Void then
				create bank_accounts_saved.make_empty
					-- FIXME
			end
			Result := bank_accounts_saved
		end

feature {RP_GLOBAL_PRODUCT} -- Composite membership

	has_direct_composite_mapping: BOOLEAN is
			-- Does 'Current' has direct mapping to composite?
		do
			if gcdb_portfolio_group /= Void then
				Result := not gcdb_portfolio_group.is_in_composite_history.is_empty
			elseif gcdb_account /= Void then
				Result := not gcdb_account.is_in_composite_history.is_empty
			end
		end

	no_direct_composite_mapping: RP_BOOLEAN_HISTORY is
			-- No direct composite mappings intervals
		do
			Result := retrieve_no_direct_composite_mapping
		end

	composite_state_cache: like composite_state
			-- Cache of composite membership status history

	retrieve_composite_state: like composite_state is
			-- Composite membership status history
		do
			if composite_state_cache = Void then
				create composite_state_cache.make_empty
				if gcdb_portfolio_group /= Void and then not gcdb_portfolio_group.is_in_composite_history.is_empty then
					gcdb_portfolio_group.is_in_composite_history.do_all (agent add_composite_state (composite_state_cache, ?))
				elseif gcdb_account /= Void then
					gcdb_account.is_in_composite_history.do_all (agent add_composite_state (composite_state_cache, ?))
				end
			end
			Result := composite_state_cache
		end

	add_composite_state (a_cache: like composite_state; a_value: ENTITY_ATTRIBUTE_RECORD) is
			-- Extend composite state history with additional interval
		local
			l_interval: RP_BOOLEAN_HISTORY_INTERVAL
			l_value: BOOLEAN
		do
			if (not a_value.attribute_value.is_empty) and then a_value.attribute_value.item (1).as_lower = 'y' then
				l_value := True
			end
			create l_interval.make (l_value, a_value.start_date, a_value.end_date, 1)
			a_cache.extend (l_interval)
		end

	no_direct_composite_mapping_cache: like no_direct_composite_mapping
			-- No direct composite mappings intervals

	retrieve_no_direct_composite_mapping: like no_direct_composite_mapping is
			-- Retrieve no direct composite mappings intervals
		local
			l_start, l_finish: ROSE_DATE
			l_interval: RP_BOOLEAN_HISTORY_INTERVAL
			l_temp: like no_direct_composite_mapping
		do
			if no_direct_composite_mapping_cache = Void then
				create no_direct_composite_mapping_cache.make_empty
				create l_temp.make_empty
				if gcdb_portfolio_group /= Void and then not gcdb_portfolio_group.composite_history.is_empty then
					gcdb_portfolio_group.composite_history.do_all (agent add_direct_mapping (l_temp, ?))
				elseif gcdb_account /= Void then
					gcdb_account.composite_history.do_all (agent add_direct_mapping (l_temp, ?))
				end
					-- take the reverse
				from
					l_temp.coalesce
					create l_start.make_min_date
					l_temp.start
				until
					l_temp.off
				loop
					l_finish := l_temp.current_item.start_date
					create l_interval.make (True, l_start, l_finish, 1)
					no_direct_composite_mapping_cache.extend (l_interval)
					l_start := l_temp.current_item.end_date
					l_temp.forth
				end
				create l_finish.make_max_date
				if l_start.ccyymmdd /= l_finish.ccyymmdd then
					create l_interval.make (True, l_start, l_finish, 1)
					no_direct_composite_mapping_cache.extend (l_interval)
				end
			end
			Result := no_direct_composite_mapping_cache
		end

	add_direct_mapping (a_cache: like no_direct_composite_mapping; a_interval: ENTITY_INTERVAL [GCDB_PRODUCT]) is
			-- Add direct composite mapping
		local
			l_interval: RP_BOOLEAN_HISTORY_INTERVAL
		do
			create l_interval.make (True, a_interval.start_date, a_interval.end_date, 1)
			a_cache.extend (l_interval)
		end

invariant

	reference_system_not_void: reference_system /= Void

end
