indexing
	description: "Stock Pricing algorithm as implemented in the legacy Perf System"
	author: "Ivan Tarnavsky"
	last_editor: "$Author: sm $"
	date: "$Date: 2011/06/23 $"
	revision: "$Revision: #4 $"
	keywords: "price,value,valuation"
	path: "$File: //rose/source/kernel/expert_system/valuators/rp_performance_stock_pricing_algorithm.e $"

class RP_PERFORMANCE_STOCK_PRICING_ALGORITHM

inherit

	RP_PRICING_ALGORITHM [RP_STOCK]

create

	make_with_trade_price

feature {NONE} -- Initialization

	make_with_trade_price (a_zone: like zone; a_reference_system: like reference_system; a_trade_price: like trade_price; a_calendar: RP_CALENDAR; a_market_data_source: like market_data_source; a_price_override: like price_override) is
			-- Initialize `Current'
		require
			reference_system_not_void: a_reference_system /= Void
			non_void_args: a_calendar /= Void and a_market_data_source /= Void and a_price_override /= Void and a_zone /= Void
		do
			zone := a_zone
			reference_system := a_reference_system
			trade_price := a_trade_price
			calendar := a_calendar
			market_data_source := a_market_data_source
			price_override := a_price_override
			create last_price_cache.make_equal (17)
			create price_cache.make_equal (17)
		ensure
			reference_system_set: reference_system = a_reference_system
			market_data_source_set: market_data_source = a_market_data_source
			calendar_set: calendar /= Void
		end

feature -- Query

	price (a_stock: RP_STOCK; a_date: ROSE_DATE): RP_PRICE is
			-- Price of `a_stock' on `a_date'
		local
			l_record: RP_MARKET_DATA_RECORD
			l_adjustment_now, l_adjustment_then: RP_DECIMAL
			l_price_lag: PAIR [RP_PRICE, RP_DECIMAL]
			l_price: RP_PRICE
			l_instrument: RP_STOCK
			l_local_instrument: RP_STOCK
			l_local_brimsym: STRING
		do
			l_local_brimsym := a_stock.local_brimsym (zone, a_date)
			l_local_instrument ?= reference_system.instrument_from_brimsym (l_local_brimsym)
				-- Check for price override first
			l_record := price_override.snapshot (l_local_instrument, a_date)
			if l_record /= Void and then l_record.has_close  then
				Result := l_record.unadjusted_close
				create l_price_lag.make (Result, Void)
				last_price_cache.force (l_price_lag, a_stock)
			else
					-- Retrieve market price if there is no override price
				if Result = Void then
						-- cross zonal conversion
					l_instrument ?= reference_system.cross_zonal_translation (a_stock, zone, a_date)
					l_record := market_data_source.snapshot (l_instrument, a_date)
					if l_record /= Void and then l_record.has_adjustment_factor then
						l_adjustment_now := l_record.adjustment_factor
					end
					Result := stock_market_price (l_instrument, a_date, calendar)
						-- Update lag price
					if Result /= Void then
						create l_price_lag.make (Result, l_adjustment_now)
						last_price_cache.force (l_price_lag, a_stock)
					end
				end
					-- No market price, check trade price
				if Result = Void then
						-- Last resort is trade price
					if trade_price/=Void then l_record := trade_price.snapshot (a_stock, a_date) end
					if l_record /= Void and then l_record.has_close then
						Result := l_record.unadjusted_close
						create l_price_lag.make (Result, l_adjustment_now)
						last_price_cache.force (l_price_lag, a_stock)
					end
				end
					-- Use lag price if everything else fails
				if Result = Void then
					last_price_cache.search (a_stock)
					if last_price_cache.found then
						l_price_lag := last_price_cache.found_item
						l_price := l_price_lag.first
						l_adjustment_then := l_price_lag.second
						if l_adjustment_then /= Void and l_adjustment_now /= Void then
								-- Lag price was a market price
							Result := a_stock.price_in_currency (l_price.price * l_adjustment_now / l_adjustment_then, l_price.currency, l_price.date)
						else
							Result := a_stock.price_in_currency (l_price.price, l_price.currency, l_price.date)
							l_adjustment_now := l_adjustment_then
						end
						create l_price_lag.make (Result, l_adjustment_now)
						last_price_cache.force (l_price_lag, a_stock)
					end
				end
			end
		end

feature {NONE} -- Implementation

	zone: RP_ZONE
			-- Zone

	reference_system: RP_REFERENCE_SYSTEM
			-- Reference system

	last_price_cache: UID_TABLE [PAIR [RP_PRICE, RP_DECIMAL], RP_INSTRUMENT]
			-- Last computed price cache

	calendar: RP_CALENDAR
			-- Market calendar

	market_data_source: RP_MARKET_DATA_SOURCE [RP_STOCK]
			-- Market data source

	price_override: RP_MARKET_DATA_SOURCE [RP_INSTRUMENT]
			-- Price override market data source

	trade_price: RP_TRADE_PRICE_MARKET_DATA_SOURCE
			-- Trade price market data source

	lookback_range: INTEGER is
			-- How far back in days to look for price
		do
			Result := 60
		end

	stock_market_price (a_stock: RP_STOCK; a_date: ROSE_DATE; a_calendar: RP_CALENDAR): RP_PRICE is
			-- Retrieve market price for `a_stock' on `a_date'
		require
			non_void_args: a_stock /= Void and a_date /= Void and a_calendar /= Void
		local
			l_market_data: RP_MARKET_DATA_RECORD
			l_price: RP_PRICE
		do
			l_price := search_price (a_stock, a_date)
			if l_price = Void then
					-- retrive price record
				l_market_data := market_data_source.snapshot (a_stock, a_date)
				if l_market_data /= Void and then is_valid (l_market_data) then
						-- found and valid
					if l_market_data.has_close or l_market_data.has_bid then
						if l_market_data.has_close then
							Result := l_market_data.unadjusted_close
						elseif l_market_data.has_bid then
							Result := l_market_data.unadjusted_bid
						end
					else
						Result := retrieve_last_stock_price (l_market_data.adjustment_factor, a_stock, a_date, a_calendar)
					end
				elseif l_market_data /= Void and then l_market_data.has_adjustment_factor then
						-- found but has only adj factor
					Result := retrieve_last_stock_price (l_market_data.adjustment_factor, a_stock, a_date, a_calendar)
				else
						-- lookback
					l_market_data := price_lookback (market_data_source, a_stock, a_date, lookback_range, a_calendar, agent has_adjustment_factor (?))
					if l_market_data /= Void then
						Result := retrieve_last_stock_price (l_market_data.adjustment_factor, a_stock, a_date, a_calendar)
					end
				end
				save_price (a_stock, Result, a_date)
			else
				Result := l_price
			end
		end

	is_valid (a_market_data: RP_MARKET_DATA_RECORD): BOOLEAN is
			-- Is `a_market_data` valid?
		do
			Result := has_any_price (a_market_data) and has_adjustment_factor (a_market_data)
		end

	has_any_price (a_market_data: RP_MARKET_DATA_RECORD): BOOLEAN is
			-- Does `a_market_data' have any price field?
		require
			attached_market_data: a_market_data /= Void
		do
			Result := a_market_data.has_close or a_market_data.has_bid or a_market_data.has_ask
		end

	has_any_price_with_volume (a_market_data: RP_MARKET_DATA_RECORD): BOOLEAN is
			-- Does `a_market_data' have any price field and volume?
		require
			attached_market_data: a_market_data /= Void
		do
			Result := has_any_price (a_market_data) and a_market_data.has_volume
		end

	has_close (a_market_data: RP_MARKET_DATA_RECORD): BOOLEAN is
			-- Does `a_market_data' have close price field?
		do
			Result := a_market_data.has_close
		end

	has_bid (a_market_data: RP_MARKET_DATA_RECORD): BOOLEAN is
			-- Does `a_market_data' have bid price field?
		do
			Result := a_market_data.has_bid
		end

	has_adjustment_factor (a_market_data: RP_MARKET_DATA_RECORD): BOOLEAN is
			-- Does `a_market_data' have adjustment field?
		do
			Result := a_market_data.has_adjustment_factor
		end

	has_close_or_bid (a_market_data: RP_MARKET_DATA_RECORD): BOOLEAN is
			-- Does `a_market_data' have bid price or close field?
		do
			Result := a_market_data.has_bid or a_market_data.has_close
		end

	retrieve_last_stock_price (a_adjustment_factor: RP_DECIMAL; a_instrument: RP_STOCK; a_date: ROSE_DATE; a_calendar: RP_CALENDAR): RP_PRICE is
			-- Retrieve last market price
		local
			a_market_data: RP_MARKET_DATA_RECORD
		do
			a_market_data := price_lookback (market_data_source, a_instrument, a_date, lookback_range, a_calendar, agent has_any_price (?))
			if a_market_data /= Void then
				if a_market_data.has_close then
					Result := a_market_data.unadjusted_close
				elseif a_market_data.has_bid then
					Result := a_market_data.unadjusted_bid
				else
					a_market_data := price_lookback (market_data_source, a_instrument, a_date, lookback_range, a_calendar, agent has_close_or_bid (?))
					if a_market_data /= Void then
						if a_market_data.has_close then
							Result := a_market_data.unadjusted_close
						else
							Result := a_market_data.unadjusted_bid
						end
					end
				end
				if Result /= Void then
					if a_adjustment_factor /= Void then
						if not a_adjustment_factor.is_equal (a_market_data.adjustment_factor) then
							if a_market_data.date < a_date and then a_market_data.date.diff_in_days (a_date).abs <= lookback_range then
								if a_market_data.adjustment_factor > 0 then
									if a_instrument.zone_mnemonic = 'U' or else a_date.ccyymmdd >= 20020101 then
										Result := Result.instrument.price_in_currency (Result.price * a_adjustment_factor / a_market_data.adjustment_factor, Result.currency, Result.date)
									end
								end
							end
						end
					end
				end
			end
		end

	price_lookback (a_mkt: RP_MARKET_DATA_SOURCE [RP_STOCK]; a_instrument: RP_STOCK; a_date: ROSE_DATE; a_max_days_to_go_back: INTEGER; a_calendar: RP_CALENDAR; a_query: FUNCTION [ANY, TUPLE, BOOLEAN]): RP_MARKET_DATA_RECORD is
			-- Price lookback
		require
			source_not_void: a_mkt /= Void
			stock_not_void: a_instrument /= Void
			date_not_void: a_date /= Void
			calendar_not_void: a_calendar /= Void
		local
			l_date: ROSE_DATE
			l_counter: INTEGER
			l_found: BOOLEAN
			l_result: RP_MARKET_DATA_RECORD
		do
			from
				create l_date.make_by_date (a_date)
				l_counter := 0
			until
				l_counter > a_max_days_to_go_back or l_found
			loop
				l_result := a_mkt.snapshot (a_instrument, l_date)
				if l_result /= Void then
					if a_query = Void then
						l_found := True
					else
						l_found := a_query.item ([l_result])
					end
				end
				l_counter := l_counter + 1
					-- go back one market day
				l_date := a_calendar.ith_work_day_before (l_date, 1)
			end
			if l_found then
				Result := l_result
			end
		end

	price_cache: DS_HASH_TABLE [DS_HASH_TABLE [RP_PRICE, INTEGER_64], ROSE_DATE]

	search_price (a_stock: RP_STOCK; a_date: ROSE_DATE): RP_PRICE is
			-- search cache for price of a day
		require
			attached_stock: a_stock /= Void
			attached_date: a_date /= Void
		local
			l_temp_holder: DS_HASH_TABLE [RP_PRICE, INTEGER_64]
		do
			price_cache.search (a_date)
			if price_cache.found then
				l_temp_holder := price_cache.found_item
				l_temp_holder.search (a_stock.uid)
				if l_temp_holder.found then
					Result := l_temp_holder.found_item
				end
			end
		end

	save_price (a_stock: RP_STOCK; a_price: RP_PRICE; a_date: ROSE_DATE) is
			-- save price in cache
		require
			attached_stock: a_stock /= Void
			attached_price: a_price /= Void
			attached_date: a_date /= Void
		local
			l_temp_holder: DS_HASH_TABLE [RP_PRICE, INTEGER_64]
		do
			price_cache.search (a_date)
			if price_cache.found then
				l_temp_holder := price_cache.found_item
				l_temp_holder.search (a_stock.uid)
				if not l_temp_holder.found then
					l_temp_holder.force (a_price, a_stock.uid)
				end
			else
				create l_temp_holder.make_default
				l_temp_holder.force (a_price, a_stock.uid)
				price_cache.force (l_temp_holder, a_date)
			end
		end

end
