note

	description:

		"Performance System"

	author: "Ivan Tarnavasky"
	last_editor: "$Author: ext-asurelia $"
	date: "$Date: 2014/11/03 $"
	revision: "$Revision: #57 $"
	path: "$File: //rose/source/kernel/expert_system/aflows/rp_performance_system.e $"

class RP_PERFORMANCE_SYSTEM

inherit

	RP_INVESTMENT_MANAGER_CLIENT

create {RP_INVESTMENT_MANAGER}

	make

feature -- Access

	default_holding_valuator_factory: RP_DEFAULT_HOLDING_VALUATOR_FACTORY is
			-- Default holding valuator factory.
		do
			if default_holding_valuator_factory_saved = Void then
				default_holding_valuator_factory_saved := create {RP_DEFAULT_HOLDING_VALUATOR_FACTORY}.make (manager)
			end
			Result := default_holding_valuator_factory_saved
		end

	fx: RP_AXAROSENBERG_CURRENCY_MARKET_DATA_SOURCE is
			-- FX market data source
		once
			create Result.make (manager.reference_system)
		end

feature -- Status

	enable_portfolio_dailyvalues_check is
			-- Enable portfolio dailyvalues check
		do
			ignore_portfolio_dailyvalues_check := False
		end

	disable_portfolio_dailyvalues_check is
			-- Disable portfolio dailyvalues check
		do
			ignore_portfolio_dailyvalues_check := True
		end

feature -- Basic operations

	clear_benchmark_daily_returns_cache is
			-- Clear the cache to force a retrieval refresh
		do
			if benchmark_daily_returns_cache /= Void then
				benchmark_daily_returns_cache.wipe_out
			end
			if benchmark_cum_daily_returns_cache /= Void then
				benchmark_cum_daily_returns_cache.wipe_out
			end
			if benchmark_daily_returns_cache /= Void then
				benchmark_daily_returns_cache.wipe_out
			end
			if benchmark_returns_cache /= Void then
				benchmark_returns_cache.wipe_out
			end
		end

feature -- Queries

	all_benchmark_daily_returns (a_benchmark: RP_BENCHMARK): DS_HASH_TABLE [PAIR [RP_DECIMAL, RP_CURRENCY], ROSE_DATE] is
			-- return hash table with all returns/currencies for specified benchmark
		require
			a_benchmark_not_void: a_benchmark /= Void
		do
			Result := retrieve_benchmark_daily_returns (a_benchmark)
		end

	daily_benchmark_return (a_benchmark: RP_BENCHMARK; a_calendar: RP_CALENDAR; a_date: ROSE_DATE; a_currency: RP_CURRENCY): RP_DECIMAL is
			-- Return of a_benchmark for a_date in a_currency
		require
			non_void_args: a_benchmark /= Void and a_date /= Void and a_currency /= Void
		local
			l_return_in_currency: PAIR [RP_DECIMAL, RP_CURRENCY]
			l_fx_start, l_fx_finish: RP_DECIMAL
		do
			l_return_in_currency := db_bogie_dailies (a_benchmark, a_date)
			if l_return_in_currency /= Void and then l_return_in_currency.first.is_defined then
				l_fx_start := benchmarks_spot_fx (l_return_in_currency.second, a_currency, a_calendar.ith_work_day_before (a_date, 1))
				l_fx_finish := benchmarks_spot_fx (l_return_in_currency.second, a_currency, a_date)
				Result := ((1 + l_return_in_currency.first / 100) * (l_fx_finish / l_fx_start) - 1) * 100
			else
				manager.log_minor ("ERROR: No return for " + a_benchmark.code + " on " + a_date.ccyymmdd.out)
				create Result.make_undefined
			end
		end

	benchmark_interval_return (a_benchmark: RP_BENCHMARK; a_not_used: RP_CALENDAR; a_start_date, a_end_date: ROSE_DATE; a_currency: RP_CURRENCY): RP_DECIMAL is
			-- Return of 'a_benchmark' for interval 'a_start_date' to 'a_end_date' in 'a_currency'
		require
			non_void_args: a_benchmark /= Void and a_start_date /= Void and a_end_date /= Void and a_currency /= Void
		local
			l_return_in_currency: PAIR [RP_DECIMAL, RP_CURRENCY]
			l_fx_start, l_fx_finish: RP_DECIMAL
		do
			l_return_in_currency := db_cum_bogie_dailies (a_benchmark, a_start_date, a_end_date)
			if l_return_in_currency /= Void and then l_return_in_currency.first.is_defined then
				l_fx_start := benchmarks_spot_fx (l_return_in_currency.second, a_currency, a_start_date)
				l_fx_finish := benchmarks_spot_fx (l_return_in_currency.second, a_currency, a_end_date)
				Result := ((1 + l_return_in_currency.first / 100) * (l_fx_finish / l_fx_start) - 1) * 100
			else
				manager.log_minor ("ERROR: No return for " + a_benchmark.code + " on " + a_end_date.ccyymmdd.out)
				create Result.make_undefined
			end
		ensure
			non_void_result: Result /= Void
		end

	last_benchmark_daily_return_date_within_month (a_benchmark: RP_BENCHMARK; a_calendar: RP_CALENDAR; a_date: ROSE_DATE): ROSE_DATE is
			-- Date of last available benchmark daily return within month of `a_date'
		require
			non_void_args: a_benchmark /= Void and a_date /= Void
		local
			l_returns: like retrieve_benchmark_daily_returns
			l_date: ROSE_DATE
		do
			from
				l_returns := retrieve_benchmark_daily_returns (a_benchmark)
				l_date := a_calendar.last_work_day (a_date)
			until
				l_date.ccyymm /= a_date.ccyymm or Result /= Void
			loop
				l_returns.search (l_date)
				if l_returns.found then
					Result := l_date
				end
				l_date := a_calendar.ith_work_day_before (l_date, 1)
			end
		end

	official_portfolio_performance (a_portfolio: RP_PORTFOLIO; a_currency: RP_CURRENCY): ROSE_LINEAR_ARRAY [MONTHLY_PERF_ROW] is
			-- All data from OfficialPerf database for `a_portfolio'
		require
			non_void_portfolio: a_portfolio /= Void
			portfolio_is_known: not a_portfolio.is_unknown
			portfolio_is_not_in_t_zone: not a_portfolio.zone.is_transzonal
			currency_not_void: a_currency /= Void
		do
			official_perf_query.retrieve (Void, <<a_portfolio.zone.code, a_portfolio.code, a_currency.name>>)
			official_perf_query.data_server.commit
			create Result.make_from_array (official_perf_query.array)
		ensure
			non_void_result: Result /= Void
		end

	official_global_account_performance (a_account: RP_INVESTMENT_ACCOUNT; a_currency: RP_CURRENCY): ROSE_LINEAR_ARRAY [MONTHLY_PERF_ROW] is
			-- All data from OfficialPerf database for `a_account'
		require
			non_void_account: a_account /= Void
			account_is_known: not a_account.is_unknown
			account_is_global: a_account.is_global
			currency_not_void: a_currency /= Void
		local
			l_zone: STRING
		do
			if a_account.is_gpc then
				l_zone := manager.reference_system.w_zone.code
			else
				l_zone := manager.reference_system.t_zone.code
			end
			official_perf_query.retrieve (Void, <<l_zone, a_account.code, a_currency.name>>)
			official_perf_query.data_server.commit
			create Result.make_from_array (official_perf_query.array)
		ensure
			non_void_result: Result /= Void
		end

	official_composite_performance (a_composite: RP_PRODUCT; a_currency: RP_CURRENCY): ROSE_LINEAR_ARRAY [MONTHLY_PERF_ROW] is
			-- All data from OfficialPerf database for `a_composite'
		require
			composite_not_void: a_composite /= Void
			composite_is_valid: a_composite.is_composite
			currency_not_void: a_currency /= Void
		do
			fc_composite_bogie_monthlyperf_reportable_view.retrieve (Void, <<a_composite.investment_region.code, a_composite.legacy_code, a_currency.name>>)
			fc_composite_bogie_monthlyperf_reportable_view.data_server.commit
			create Result.make_from_array (fc_composite_bogie_monthlyperf_reportable_view.array)
		ensure
			non_void_result: Result /= Void
		end

	official_hedged_global_account_performance (a_account: RP_INVESTMENT_ACCOUNT; a_currency: RP_CURRENCY): ROSE_LINEAR_ARRAY [MONTHLY_PERF_ROW] is
			-- All data from OfficialPerf database for `a_account'
		require
			non_void_account: a_account /= Void
			account_is_known: not a_account.is_unknown
			account_is_global: a_account.is_global
			account_is_hedged: a_account.is_hedged
			currency_not_void: a_currency /= Void
		do
			official_hedged_perf_query.retrieve (Void, <<"T", a_account.code + "_H", a_currency.name>>)
			official_hedged_perf_query.data_server.commit
			create Result.make_from_array (official_hedged_perf_query.array)
		ensure
			non_void_result: Result /= Void
		end

	offical_benchmark_monthly_return (a_benchmark: RP_BENCHMARK; a_calendar: RP_CALENDAR; a_month: ROSE_MONTH; a_currency: RP_CURRENCY): RP_DECIMAL is
			-- Return of `a_benchmark' for `a_month' in `a_currency'
		require
			non_void_args: a_benchmark /= Void and a_month /= Void and a_currency /= Void
		local
			l_return_in_currency: PAIR [RP_DECIMAL, RP_CURRENCY]
			l_fx_start, l_fx_finish: RP_DECIMAL
		do
			l_return_in_currency := opb_monthly_return (a_benchmark, a_month)
			if l_return_in_currency /= Void and then l_return_in_currency.first.is_defined then
				l_fx_start := benchmarks_spot_fx (l_return_in_currency.second, a_currency, a_calendar.last_work_day (a_month.month_beginning.previous_date))
				l_fx_finish := benchmarks_spot_fx (l_return_in_currency.second, a_currency, a_calendar.last_work_day (a_month.month_end))
				Result := l_return_in_currency.first * l_fx_finish / l_fx_start
			else
				manager.log_minor ("ERROR: No return for " + a_benchmark.code + " on " + a_month.ccyymm.out)
				create Result.make_undefined
			end
		end

	benchmark_fx_monthly_return (a_calendar: RP_CALENDAR; a_month: ROSE_DATE; a_from_currency, a_to_currency: RP_CURRENCY): RP_DECIMAL is
			-- Benchmark FX return for `a_month' between `a_from_currency' and `a_to_currency'
		require
			a_calendar_not_void: a_calendar /= Void
			a_month_not_void: a_month /= Void
			a_from_currency_not_void: a_from_currency /= Void
			a_to_currency_not_void: a_to_currency /= Void
		local
			l_fx_start, l_fx_finish: RP_DECIMAL
		do
			l_fx_start := benchmarks_spot_fx (a_from_currency, a_to_currency, a_calendar.last_work_day (a_month.month_beginning.previous_date))
			l_fx_finish := benchmarks_spot_fx (a_from_currency, a_to_currency, a_calendar.last_work_day (a_month.month_end))
			if (l_fx_finish.is_defined and l_fx_start.is_defined) and then (l_fx_finish > 0 and l_fx_start > 0) then
				Result := l_fx_finish / l_fx_start
			else
				create Result.make_undefined
			end
		ensure
			not_void: Result /= Void
		end

	official_perf_portfolio_value (a_portfolio: RP_PORTFOLIO; a_month: ROSE_MONTH; a_currency: RP_CURRENCY): RP_DECIMAL is
			-- Value of `a_portfolio' at the end of `a_month' in `a_currency' from OfficialPerf DB.
		do
			Result := op_portvalue (a_portfolio.zone.code, a_portfolio.code, a_month.ccyymm, a_currency.name)
		end

	official_perf_portfolio_monthlyperf (a_portfolio: RP_PORTFOLIO; a_month: ROSE_MONTH; a_currency: RP_CURRENCY): RP_DECIMAL is
			-- Return of `a_portfolio' at the end of `a_month' in `a_currency' from OfficialPerf DB.
		do
			Result := op_monthlyperf (a_portfolio.zone.code, a_portfolio.code, a_month.ccyymm, a_currency.name).first
		end

	official_perf_portfolio_benchmark_monthlyperf (a_portfolio: RP_PORTFOLIO; a_month: ROSE_MONTH; a_currency: RP_CURRENCY): RP_DECIMAL is
			-- Return of `a_portfolio' at the end of `a_month' in `a_currency' from OfficialPerf DB.
		do
			Result := op_monthlyperf (a_portfolio.zone.code, a_portfolio.code, a_month.ccyymm, a_currency.name).second
		end

	official_perf_investment_account_value (a_account: RP_INVESTMENT_ACCOUNT; a_month: ROSE_MONTH; a_currency: RP_CURRENCY): RP_DECIMAL is
			-- Value of `a_account' at the end of `a_month' in `a_currency' from OfficialPerf DB.
		do
			Result := op_portvalue ("T", a_account.name, a_month.ccyymm, a_currency.name)
		end

	official_perf_investment_account_monthlyperf (a_account: RP_INVESTMENT_ACCOUNT; a_month: ROSE_MONTH; a_currency: RP_CURRENCY): RP_DECIMAL is
			-- Return of `a_account' at the end of `a_month' in `a_currency' from OfficialPerf DB.
		do
			Result := op_monthlyperf ("T", a_account.name, a_month.ccyymm, a_currency.name).first
		end

	official_perf_investment_account_benchmark_monthlyperf (a_account: RP_INVESTMENT_ACCOUNT; a_month: ROSE_MONTH; a_currency: RP_CURRENCY): RP_DECIMAL is
			-- Return of `a_account' at the end of `a_month' in `a_currency' from OfficialPerf DB.
		do
			Result := op_monthlyperf ("T", a_account.name, a_month.ccyymm, a_currency.name).second
		end

	official_perf_hedged_investment_account_value (a_account: RP_INVESTMENT_ACCOUNT; a_month: ROSE_MONTH; a_currency: RP_CURRENCY): RP_DECIMAL is
			-- Value of `a_account' at the end of `a_month' in `a_currency' from OfficialPerf DB.
		do
			Result := op_h_portvalue ("T", a_account.name, a_month.ccyymm, a_currency.name)
		end

	official_perf_hedged_investment_account_monthlyperf (a_account: RP_INVESTMENT_ACCOUNT; a_month: ROSE_MONTH; a_currency: RP_CURRENCY): RP_DECIMAL is
			-- Return of `a_account' at the end of `a_month' in `a_currency' from OfficialPerf DB.
		do
			Result := op_h_monthlyperf ("T", a_account.name, a_month.ccyymm, a_currency.name).first
		end

	official_perf_hedged_investment_account_benchmark_monthlyperf (a_account: RP_INVESTMENT_ACCOUNT; a_month: ROSE_MONTH; a_currency: RP_CURRENCY): RP_DECIMAL is
			-- Return of `a_account' at the end of `a_month' in `a_currency' from OfficialPerf DB.
		do
			Result := op_h_monthlyperf ("T", a_account.name, a_month.ccyymm, a_currency.name).second
		end

	official_perf_composite_returns (a_composite: RP_PRODUCT; a_currency: RP_CURRENCY): DS_HASH_TABLE [RP_DECIMAL, ROSE_DATE] is
			-- Monthly returns for 'a_composite'
		require
			composite_not_void: a_composite /= Void
			currency_not_void: a_currency /= Void
		do
			Result := op_composite_monthlyperf (a_composite, a_currency)
		end

	official_perf_composite_perfrun_dates (a_composite: RP_PRODUCT; a_currency: RP_CURRENCY): DS_HASH_TABLE [ROSE_DATE, ROSE_DATE] is
			-- Official perfrun dates for 'a_composite'
		require
			non_void_args: a_composite /= Void and a_currency /= Void
		do
			Result := op_composite_perfrun_dates (a_composite, a_currency)
		end

	net_of_fee_composite_returns (a_composite: RP_PRODUCT; a_currency: RP_CURRENCY): DS_HASH_TABLE [RP_DECIMAL, ROSE_DATE] is
			-- Monthly NOF returns for 'a_composite'
		require
			composite_not_void: a_composite /= Void
			currency_not_void: a_currency /= Void
		do
			Result := nof_composite_returns (a_composite, a_currency)
		end

	official_perf_composite_benchmark_returns (a_composite: RP_PRODUCT; a_currency: RP_CURRENCY): DS_HASH_TABLE [RP_DECIMAL, ROSE_DATE] is
			-- Monthly composite benchmark returns for 'a_composite' and 'a_currency'
		require
			composite_not_void: a_composite /= Void
			currency_not_void: a_currency /= Void
		do
			Result := op_composite_benchmark_monthlyperf (a_composite, a_currency)
		end

	latest_perf_portfolio_value (a_portfolio: RP_PORTFOLIO; a_month: ROSE_MONTH; a_currency: RP_CURRENCY): RP_DECIMAL is
			-- Value of `a_portfolio' at the end of `a_month' in `a_currency' from OfficialPerf DB.
		do
			Result := latest_portvalue (a_portfolio.zone.code, a_portfolio.code, a_month.ccyymm, a_currency.name)
		end

	latest_perf_portfolio_monthlyperf (a_portfolio: RP_PORTFOLIO; a_month: ROSE_MONTH; a_currency: RP_CURRENCY): RP_DECIMAL is
			-- Return of `a_portfolio' at the end of `a_month' in `a_currency' from OfficialPerf DB.
		do
			Result := latest_monthlyperf (a_portfolio.zone.code, a_portfolio.code, a_month.ccyymm, a_currency.name).first
		end

	latest_perf_portfolio_benchmark_monthlyperf (a_portfolio: RP_PORTFOLIO; a_month: ROSE_MONTH; a_currency: RP_CURRENCY): RP_DECIMAL is
			-- Return of `a_portfolio' at the end of `a_month' in `a_currency' from OfficialPerf DB.
		do
			Result := latest_monthlyperf (a_portfolio.zone.code, a_portfolio.code, a_month.ccyymm, a_currency.name).second
		end

	latest_perf_investment_account_value (a_account: RP_INVESTMENT_ACCOUNT; a_month: ROSE_MONTH; a_currency: RP_CURRENCY): RP_DECIMAL is
			-- Value of `a_account' at the end of `a_month' in `a_currency' from OfficialPerf DB.
		do
			Result := latest_portvalue ("T", a_account.name, a_month.ccyymm, a_currency.name)
		end

	latest_perf_investment_account_monthlyperf (a_account: RP_INVESTMENT_ACCOUNT; a_month: ROSE_MONTH; a_currency: RP_CURRENCY): RP_DECIMAL is
			-- Return of `a_account' at the end of `a_month' in `a_currency' from OfficialPerf DB.
		do
			Result := latest_monthlyperf ("T", a_account.name, a_month.ccyymm, a_currency.name).first
		end

	latest_perf_investment_account_benchmark_monthlyperf (a_account: RP_INVESTMENT_ACCOUNT; a_month: ROSE_MONTH; a_currency: RP_CURRENCY): RP_DECIMAL is
			-- Return of `a_account' at the end of `a_month' in `a_currency' from OfficialPerf DB.
		do
			Result := latest_monthlyperf ("T", a_account.name, a_month.ccyymm, a_currency.name).second
		end

	portfolio_daily_buckets (a_portfolio: RP_PORTFOLIO; a_valuator: RP_HOLDING_VALUATOR; a_currency: RP_CURRENCY; a_date: ROSE_DATE): RP_PERFORMANCE_DAILY_VALUES is
			-- Portfolio daily values on given date
		local
			l_aggregator: RP_PERFORMANCE_DAILY_BUCKET_FLOW_AGGREGATOR
			l_holdings: ROSE_LINEAR_ARRAY [RP_HOLDING_INTERVAL]
			i, nb: INTEGER
		do
			create Result.make (a_date, a_currency)
			create l_aggregator.make (Result, a_portfolio, a_valuator)
			l_holdings := manager.accounting.portfolio_holdings_at (a_portfolio, a_date)
			from
				i := 1
				nb := l_holdings.count
			until
				i > nb
			loop
				l_aggregator.adjust_with (l_holdings.item (i))
				i := i + 1
			end
		end

	portfolio_dailyvalues (a_portfolio: RP_PORTFOLIO; a_currency: RP_CURRENCY): ARRAY [RP_PERFORMANCE_DAILY_VALUES] is
			-- DailyValues for `a_portfolio' in `a_currency'
		require
			non_void_args: a_portfolio /= Void and a_currency /= Void
		local
			l_valuator: RP_HOLDING_VALUATOR
		do
			l_valuator := default_holding_valuator_factory.portfolio_holding_valuator (a_portfolio)
			Result := portfolio_dailyvalues_with_valuator (a_portfolio, a_currency, l_valuator, create {ROSE_DATE}.make_today)
		ensure
			non_void_result: Result /= Void
		end

	portfolio_dailyvalues_simulation (a_portfolio: RP_PORTFOLIO; a_currency: RP_CURRENCY; a_prices:  ROSE_TIME_SERIES [ROSE_HASH_TABLE [DOUBLE, STRING]]; a_system: RP_REFERENCE_SYSTEM): ARRAY [RP_PERFORMANCE_DAILY_VALUES] is
			-- DailyValues for `a_portfolio' in `a_currency' with simulation valuatior
		require
			non_void_args: a_portfolio /= Void and a_currency /= Void
		local
			l_valuator: RP_HOLDING_VALUATOR_SIMULATION
		do
			l_valuator := default_holding_valuator_factory.portfolio_holding_valuator_simulation (a_portfolio, a_prices, a_system)
			Result := portfolio_dailyvalues_with_valuator (a_portfolio, a_currency, l_valuator, create {ROSE_DATE}.make_today)
		ensure
			non_void_result: Result /= Void
		end

	portfolio_dailyvalues_with_valuator (a_portfolio: RP_PORTFOLIO; a_currency: RP_CURRENCY; a_valuator: RP_HOLDING_VALUATOR; a_end_date: ROSE_DATE): ARRAY [RP_PERFORMANCE_DAILY_VALUES] is
			-- DailyValues for `a_portfolio' in `a_currency' computed with `a_valuator'
		require
			non_void_args: a_portfolio /= Void and a_currency /= Void and a_valuator /= Void and a_end_date /= Void
		local
			l_start, l_finish: ROSE_DATE
			l_array: ROSE_LINEAR_ARRAY [RP_PERFORMANCE_DAILY_VALUES]
			l_buckets: like portfolio_daily_buckets
			l_previous_buckets: RP_PERFORMANCE_DAILY_VALUES
		do
			create l_array.make_empty
			l_start := manager.accounting.portfolio_positions (a_portfolio).min_start_date
			if l_start /= Void then
				from
					l_finish := a_end_date.min (a_portfolio.end_date)
				until
					l_start > l_finish
				loop
					l_buckets := portfolio_daily_buckets (a_portfolio, a_valuator, a_currency, l_start)
					if l_previous_buckets /= Void then
						l_buckets.set_daily_change_of_hedge_gain_loss (l_buckets.unrealized_hedge_contract_value - l_previous_buckets.unrealized_hedge_contract_value + l_buckets.realized_hedge_gain_loss)
					else
						l_buckets.set_daily_change_of_hedge_gain_loss (l_buckets.unrealized_hedge_contract_value + l_buckets.realized_hedge_gain_loss)
					end
					l_array.extend (l_buckets)
					l_start := a_portfolio.zone.legacy_perf_calendar.ith_work_day_after (l_start, 1)
					l_previous_buckets := l_buckets
				end
			end
			Result := l_array
		ensure
			non_void_result: Result /= Void
		end

	portfolio_dailyreturns (a_portfolio: RP_PORTFOLIO; a_currency: RP_CURRENCY): RP_PERFORMANCE_DAILY_RETURNS is
			-- Dailyreturns for `a_portfolio' in `a_currency'
		require
			non_void_args: a_portfolio /= Void and a_currency /= Void
		local
			l_dv: like portfolio_dailyvalues
			l_adj: RP_PERFORMANCE_GROSS
			i, nb: INTEGER
		do
			create l_adj
			l_dv := portfolio_dailyvalues (a_portfolio, a_currency)
			create Result.make (a_portfolio.code, regular_performance_numerator, regular_performance_denominator, l_adj)
			from
				i := 1
				nb := l_dv.count
			until
				i > nb
			loop
				Result.add_dailyvalues (l_dv.item (i))
				i := i + 1
			end
		ensure
			non_void_result: Result /= Void
		end

	portfolio_dailyreturns_simulation (a_portfolio: RP_PORTFOLIO; a_currency: RP_CURRENCY; a_prices: ROSE_TIME_SERIES [ROSE_HASH_TABLE [DOUBLE, STRING]]; a_system: RP_REFERENCE_SYSTEM): RP_PERFORMANCE_DAILY_RETURNS is
			-- Dailyreturns for `a_portfolio' in `a_currency'
		require
			non_void_args: a_portfolio /= Void and a_currency /= Void
		local
			l_dv: like portfolio_dailyvalues
			l_adj: RP_PERFORMANCE_GROSS
			i, nb: INTEGER
		do
			create l_adj
			l_dv := portfolio_dailyvalues_simulation (a_portfolio, a_currency, a_prices, a_system)
			create Result.make (a_portfolio.code, regular_performance_numerator, regular_performance_denominator, l_adj)
			from
				i := 1
				nb := l_dv.count
			until
				i > nb
			loop
				Result.add_dailyvalues (l_dv.item (i))
				i := i + 1
			end
		ensure
			non_void_result: Result /= Void
		end

	investment_account_dailyvalues (a_account: RP_INVESTMENT_ACCOUNT; a_currency: RP_CURRENCY): ARRAY [RP_PERFORMANCE_DAILY_VALUES] is
			-- Dailyvalues for `a_account' in `a_currency'
		require
			non_void_args: a_account /= Void and a_currency /= Void
		local
			l_result: DS_LINKED_LIST [RP_PERFORMANCE_DAILY_VALUES]
		do
			retrieve_account_dailyvalues (a_account)
			account_dv_cache.search (a_account)
			if account_dv_cache.found then
				create l_result.make_equal
				account_dv_cache.found_item.do_all (agent fx_to (l_result, a_currency, ?))
				Result := l_result.to_array
			else
				Result := <<>>
			end
		ensure
			non_void_result: Result /= Void
		end

	investment_account_dailyreturns (a_account: RP_INVESTMENT_ACCOUNT; a_currency: RP_CURRENCY): RP_PERFORMANCE_DAILY_RETURNS is
			-- Dailyreturns for `a_account' in `a_currency'
		require
			non_void_args: a_account /= Void and a_currency /= Void
		local
			l_dv: like investment_account_dailyvalues
			l_adj: RP_PERFORMANCE_NET_OF_SHAREPOOLFUND_INCEXP
		do
			create l_adj
			l_dv := investment_account_dailyvalues (a_account, a_currency)
			create Result.make (a_account.name, regular_performance_numerator, regular_performance_denominator, l_adj)
			l_dv.do_all (agent add_daily_return (Result, ?))
		ensure
			non_void_result: Result /= Void
		end

	benchmark_dates (a_benchmark: RP_BENCHMARK): ROSE_LINEAR_ARRAY [ROSE_DATE] is
			-- if benchmark has been loaded, return sorted array of dates available for given benchmark
		require
			arg_not_void: a_benchmark /= Void
		local

		do
			benchmark_daily_returns_cache.search (a_benchmark)
			if benchmark_daily_returns_cache.found then
				Result := benchmark_daily_returns_cache.found_item.keys.to_array
				Result.sort_by (agent (a1: ROSE_DATE; a2: ROSE_DATE): BOOLEAN
				do
					Result := a1.ccyymmdd < a2.ccyymmdd
				end)
			end
		end

feature {NONE} -- Implementation

	ignore_portfolio_dailyvalues_check: BOOLEAN
			-- Ignore portfolio dailyvalues check

	regular_performance_numerator: RP_PERFORMANCE_DAILY_RETURN_NUMERATOR is
			-- Regular performance numerator
		once
			create Result
		end

	regular_performance_denominator: RP_PERFORMANCE_DAILY_RETURN_DENOMINATOR is
			-- Regular performance denominator
		once
			create Result
		end

	default_holding_valuator_factory_saved: like default_holding_valuator_factory
			-- Holding valuator factory.

	op_server: SQL_DATA_SERVER is
			-- Official perf server
		local
			l_servers: DATABASE_SERVERS
		once
			create l_servers
			Result := l_servers.make_from_repository_fast_load ("performance")
		end

	port_value_query: DOUBLE_PREDICATE is
			-- Portvalue query
		once
			create Result.make (op_server, "SELECT fc_assetvalue FROM fc_portfolio_bogie_monthlyperf_reportable_view WHERE zone= ? and portfolio = ? and ccyymm = ? and iso_fc = ?")
		end

	latest_port_value_query: DOUBLE_PREDICATE is
			-- Portvalue query to get the latest update value
		once
			create Result.make (op_server, "SELECT TOP 1 assetvalue FROM portfolio_bogie_monthlyperf WHERE zone= ? and portfolio = ? and ccyymm = ? and iso = ? order by insert_date desc")
			if not Result.success then
				(create {EXCEPTIONS}).raise ("Could not create predicate")
			end
		ensure
			valid_result: Result /= Void
		end

	hedged_port_value_query: DOUBLE_PREDICATE is
			-- Portvalue query
		once
			create Result.make (op_server, "SELECT fc_assetvalue FROM fc_hedged_portfolio_bogie_monthlyperf_reportable_view WHERE zone= ? and portfolio = ? and ccyymm = ? and iso_fc = ?")
		end

	op_portvalue (a_zone, a_portfolio: STRING; a_ccyymm: INTEGER; a_iso: STRING): RP_DECIMAL is
			-- month-end portfolio value for `a_portfolio'
		do
			port_value_query.retrieve (Void, <<a_zone, a_portfolio, a_ccyymm, a_iso>>)
			if port_value_query.array.count > 0 then
				Result := port_value_query.array.item (1)
			end
		end

	latest_portvalue (a_zone, a_portfolio: STRING; a_ccyymm: INTEGER; a_iso: STRING): RP_DECIMAL is
			-- Latest month-end portfolio value for `a_portfolio'
		do
			latest_port_value_query.retrieve (Void, <<a_zone, a_portfolio, a_ccyymm, a_iso>>)
			if not latest_port_value_query.success then
				(create {EXCEPTIONS}).raise ("Could not execute 'latest_port_value_query'")
			end
			if not latest_port_value_query.array.is_empty then
				Result := port_value_query.array.first
			end
		end

	op_h_portvalue (a_zone, a_portfolio: STRING; a_ccyymm: INTEGER; a_iso: STRING): RP_DECIMAL is
			-- month-end portfolio value for `a_portfolio'
		do
			hedged_port_value_query.retrieve (Void, <<a_zone, a_portfolio, a_ccyymm, a_iso>>)
			if hedged_port_value_query.array.count > 0 then
				Result := hedged_port_value_query.array.item (1)
			end
		end

	port_perf_query: DOUBLE_PREDICATE is
			-- Portfolio return query
		once
			create Result.make (op_server, "SELECT fc_performance FROM fc_portfolio_bogie_monthlyperf_reportable_view WHERE zone= ? and portfolio = ? and ccyymm = ? and iso_fc = ?")
		end

	latest_port_perf_query: DOUBLE_PREDICATE is
			-- Portfolio return query to get the latest entry
		once
			create Result.make (op_server, "SELECT TOP 1 portfolio_monthlyreturn FROM portfolio_bogie_monthlyperf WHERE zone= ? and portfolio = ? and ccyymm = ? and iso = ? order by insert_date desc")
		end

	hedged_port_perf_query: DOUBLE_PREDICATE is
			-- Hedged portfolio return query
		once
			create Result.make (op_server, "SELECT fc_performance FROM fc_hedged_portfolio_bogie_monthlyperf_reportable_view WHERE zone= ? and portfolio = ? and ccyymm = ? and iso_fc = ?")
		end

	benchmark_perf_query: DOUBLE_PREDICATE is
			-- Benchmark return query
		once
			create Result.make (op_server, "SELECT fc_benchmark_performance FROM fc_portfolio_bogie_monthlyperf_reportable_view WHERE zone= ? and portfolio = ? and ccyymm = ? and iso_fc = ?")
		end

	latest_benchmark_perf_query: DOUBLE_PREDICATE is
			-- Benchmark return query to get the latest entry
		once
			create Result.make (op_server, "SELECT TOP 1 bogie_monthlyreturn FROM portfolio_bogie_monthlyperf WHERE zone= ? and portfolio = ? and ccyymm = ? and iso = ? order by insert_date desc")
			if not Result.success then
				(create {EXCEPTIONS}).raise ("Could not create predicate")
			end
		end

	hedged_benchmark_perf_query: DOUBLE_PREDICATE is
			-- Hedged benchmark return query
		once
			create Result.make (op_server, "SELECT fc_benchmark_performance FROM fc_hedged_portfolio_bogie_monthlyperf_reportable_view WHERE zone= ? and portfolio = ? and ccyymm = ? and iso_fc = ?")
		end

	latest_monthlyperf (a_zone, a_portfolio: STRING; a_ccyymm: INTEGER; a_iso: STRING): PAIR [RP_DECIMAL, RP_DECIMAL] is
			-- Latest portfolio monthly return for `a_portfolio'
		require
			valid_portfolio: a_portfolio /= Void
			valid_ccyymm: a_ccyymm /= 0
		do
			latest_port_perf_query.retrieve (Void, <<a_zone, a_portfolio, a_ccyymm, a_iso>>)
			if not latest_port_perf_query.success then
				(create {EXCEPTIONS}).raise ("Could not execute 'latest_port_perf_query'")
			end
			if not latest_port_perf_query.array.is_empty then
				latest_benchmark_perf_query.retrieve (Void, <<a_portfolio, a_ccyymm, a_iso>>)
				if not latest_benchmark_perf_query.success then
					(create {EXCEPTIONS}).raise ("Could not execute 'latest_benchmark_perf_query'")
				end
				create Result.make (latest_port_perf_query.array.first, latest_benchmark_perf_query.array.first)
			else
				create Result.make (Void, Void)
			end
		ensure
			valid_result: Result /= Void
		end

	op_monthlyperf (a_zone, a_portfolio: STRING; a_ccyymm: INTEGER; a_iso: STRING): PAIR [RP_DECIMAL, RP_DECIMAL] is
			-- Portfolio monthly return for `a_portfolio'
		do
			port_perf_query.retrieve (Void, <<a_zone, a_portfolio, a_ccyymm, a_iso>>)
			if port_perf_query.array.count > 0 then
				benchmark_perf_query.retrieve (Void, <<a_zone, a_portfolio, a_ccyymm, a_iso>>)
				create Result.make (port_perf_query.array.item (1), benchmark_perf_query.array.item (1))
			else
				create Result.make (Void, Void)
			end
		end

	op_h_monthlyperf (a_zone, a_portfolio: STRING; a_ccyymm: INTEGER; a_iso: STRING): PAIR [RP_DECIMAL, RP_DECIMAL] is
			-- Portfolio monthly return for `a_portfolio'
		do
			hedged_port_perf_query.retrieve (Void, <<a_zone, a_portfolio + "_H", a_ccyymm, a_iso>>)
			if hedged_port_perf_query.array.count > 0 then
				hedged_benchmark_perf_query.retrieve (Void, <<a_zone, a_portfolio + "_H", a_ccyymm, a_iso>>)
				create Result.make (hedged_port_perf_query.array.item (1), hedged_benchmark_perf_query.array.item (1))
			else
				create Result.make (Void, Void)
			end
		end

	op_composite_perfrun_dates (a_composite: RP_PRODUCT; a_currency: RP_CURRENCY): DS_HASH_TABLE [ROSE_DATE, ROSE_DATE] is
			-- get collection of official perfrun dates for composites
		local
			l_dates: ROSE_LINEAR_ARRAY [like composite_perfrun_dates_type]
			l_cache: DS_HASH_TABLE [ROSE_DATE, ROSE_DATE]
		do
			composite_perfrun_dates_predicate.retrieve (Void, <<a_composite.legacy_code, a_currency.name, a_composite.investment_region.code>>)
			if composite_perfrun_dates_predicate.success then
				create l_cache.make_default
				l_dates := composite_perfrun_dates_predicate.array
				l_dates.do_all (agent op_composite_perfrun_dates_cached (l_cache, ?))
				Result := l_cache
			else
				create Result.make_default
			end
		end

	op_composite_monthlyperf (a_composite: RP_PRODUCT; a_currency: RP_CURRENCY): DS_HASH_TABLE [RP_DECIMAL, ROSE_DATE] is
			-- get collection of official composite returns for 'a_composite' and 'a_currency'
		local
			l_returns: ROSE_LINEAR_ARRAY [like composite_returns_type]
			l_cache: DS_HASH_TABLE [RP_DECIMAL, ROSE_DATE]
		do
			composite_returns_predicate.retrieve (Void, <<a_composite.legacy_code, a_currency.name, a_composite.investment_region.code>>)
			if composite_returns_predicate.success then
				create l_cache.make_default
				l_returns := composite_returns_predicate.array
				l_returns.do_all (agent op_composite_monthlyperf_cached (l_cache, ?))
				Result := l_cache
			else
				create Result.make_default
			end
			composite_returns_predicate.close
		end

	op_composite_benchmark_monthlyperf (a_composite: RP_PRODUCT; a_currency: RP_CURRENCY): DS_HASH_TABLE [RP_DECIMAL, ROSE_DATE] is
			-- get collection of official composite benchmark returns for 'a_composite' and 'a_currency'
		local
			l_returns: ROSE_LINEAR_ARRAY [like composite_benchmark_returns_type]
			l_cache: DS_HASH_TABLE [RP_DECIMAL, ROSE_DATE]
		do
			composite_benchmark_returns_predicate.retrieve (Void, <<a_composite.legacy_code, a_currency.name, a_composite.investment_region.code>>)
			if composite_benchmark_returns_predicate.success then
				create l_cache.make_default
				l_returns := composite_benchmark_returns_predicate.array
				l_returns.do_all (agent op_composite_benchmark_monthlyperf_cached (l_cache, ?))
				Result := l_cache
			else
				create Result.make_default
			end
			composite_benchmark_returns_predicate.close
		end

	benchmarks_spot_fx (a_bcu, a_fcu: RP_CURRENCY; a_date: ROSE_DATE): RP_DECIMAL is
			-- Spot fx rate from `a_bcu' to `a_fcu' on `a_date'
		do
			Result := benchmarks_fx.exchange_rate (a_fcu.currency, a_bcu.currency, a_date)
		end

	benchmarks_fx: PERFORMANCE_LEGACY_BENCHMARK_CURRENCY_MARKET is
			-- FX market data source
		once
			create Result
		end

	fx_to (a_result: DS_LINKED_LIST [RP_PERFORMANCE_DAILY_VALUES]; a_currency: RP_CURRENCY; a_dv: RP_PERFORMANCE_DAILY_VALUES) is
			-- FX `a_dv' to `a_currency' and add to `a_result'.
		do
			if a_currency.is_same (a_dv.currency) then
				a_result.force_last (a_dv)
			else
				a_result.force_last (a_dv.in_currency (a_currency, fx))
			end
		end

	portfolio_dv (a_portfolio: RP_PORTFOLIO): DS_AVL_TREE [RP_PERFORMANCE_DAILY_VALUES, ROSE_DATE] is
			-- Dailyvalues for `a_portfolio' hashed by ccyymmdd
		local
			l_dv: like portfolio_dailyvalues
		do
			l_dv := portfolio_dailyvalues (a_portfolio, a_portfolio.zone.legacy_base_currency)
			create Result.make (create {KL_COMPARABLE_COMPARATOR [ROSE_DATE]}.make)
			l_dv.do_all (agent add_portfolio_dailyvalue (Result, ?))
		end

	add_portfolio_dailyvalue (a_result: like portfolio_dv; a_dv: RP_PERFORMANCE_DAILY_VALUES) is
			-- Add `a_dv' to `a_result'
		do
			a_result.force (a_dv, a_dv.perf_date)
		end

	op_composite_benchmark_monthlyperf_cached (a_cache: DS_HASH_TABLE [RP_DECIMAL, ROSE_DATE]; a_return: like composite_benchmark_returns_type) is
			-- push 'a_return' to cache
		require
			non_void_cache: a_cache /= Void
			non_void_row: a_return /= Void
		local
			l_decimal: RP_DECIMAL
		do
			create l_decimal.make_from_double (a_return.fc_benchmark_performance)
			a_cache.force_last (l_decimal, a_return.reporting_date)
		end

	op_composite_monthlyperf_cached (a_cache: DS_HASH_TABLE [RP_DECIMAL, ROSE_DATE]; a_return: like composite_returns_type) is
			-- push 'a_return' to cache
		require
			non_void_cache: a_cache /= Void
			non_void_row: a_return /= Void
		local
			l_decimal: RP_DECIMAL
		do
			create l_decimal.make_from_double (a_return.fc_performance)
			a_cache.force_last (l_decimal, a_return.reporting_date)
		end

	op_composite_perfrun_dates_cached (a_table: DS_HASH_TABLE [ROSE_DATE, ROSE_DATE]; a_row: like composite_perfrun_dates_type) is
			-- add 'a_row' to 'a_table'
		require
			non_void_args: a_table /= Void and a_row /= Void
		do
			a_table.force_last (a_row.perfrun_date, a_row.reporting_date)
		end

	composite_benchmark_returns_type: TUPLE [fc_benchmark_performance: REAL_64; reporting_date: ROSE_DATE]
			--Return type for composite benchmark returns

	composite_benchmark_returns_query: STRING is
			-- SQL query for official perf composite returns
		once
			Result := "[
		SELECT fc_benchmark_performance, reporting_date
			FROM fc_composite_bogie_monthlyperf_reportable_view
						WHERE portfolio=? AND iso_fc=? AND [zone]=?
						]"
		end

	composite_returns_type: TUPLE [fc_performance: REAL_64; reporting_date: ROSE_DATE]
			-- Return type for composite returns

	composite_returns_query: STRING is
			-- SQL query for official perf composite returns
		once
			Result := "[
		SELECT fc_performance, reporting_date
			FROM fc_composite_bogie_monthlyperf_reportable_view
						WHERE portfolio=? AND iso_fc=? AND [zone]=?
						]"
		end

	composite_returns_predicate: COLLECTING_PREDICATE [TUPLE [fc_performance: REAL_64; reporting_date: ROSE_DATE]] is
			-- predicate to retrieve official composite returns
		require
			available_server: op_server /= Void
		once
			create Result.make (op_server, composite_returns_query)
		ensure
			valid_result: Result /= Void
		end

	composite_perfrun_dates_type: TUPLE [reporting_date: ROSE_DATE; perfrun_date: ROSE_DATE]
			-- Return type for composite_perfrun_dates

	composite_perfrun_dates_query: STRING is
			-- SQL query for official perf composite perfrun dates
		once
			Result := "[
		SELECT reporting_date, perfrun_date
			FROM fc_composite_bogie_monthlyperf_reportable_view
					   WHERE portfolio=? AND iso_fc=? AND [zone]=?
					   ]"
		end

	composite_perfrun_dates_predicate: COLLECTING_PREDICATE [TUPLE [reporting_date: ROSE_DATE; perfrun_date: ROSE_DATE]] is
			-- predicate for retrieving official composite perfrun dates
		require
			available_server: op_server /= Void
		once
			create Result.make (op_server, composite_perfrun_dates_query)
		ensure
			valid_result: Result /= Void
		end

	composite_benchmark_returns_predicate: COLLECTING_PREDICATE [TUPLE [fc_benchmark_performance: REAL_64; reporting_date: ROSE_DATE]] is
			-- predicate to retrieve official composite returns
		require
			available_server: op_server /= Void
		once
			create Result.make (op_server, composite_benchmark_returns_query)
		ensure
			valid_result: Result /= Void
		end

	nof_composite_returns (a_composite: RP_PRODUCT; a_currency: RP_CURRENCY): DS_HASH_TABLE [RP_DECIMAL, ROSE_DATE] is
			-- return net of fee monthlies for a given composite and currency
		require
			composite_not_void: a_composite /= Void
			currency_not_void: a_currency /= Void
		local
			l_nof_obj: COMPOSITE_NET_OF_FEE_RETURNS
		do
			create l_nof_obj
			Result := l_nof_obj.monthly_nof_returns (a_composite, a_currency)
		ensure
			non_void_result: Result /= Void
		end

	account_cached: DS_HASH_SET [RP_INVESTMENT_ACCOUNT]
			-- Account dailyvalues cache

	account_dv_cache: DS_HASH_TABLE [DS_HASH_TABLE [RP_PERFORMANCE_DAILY_VALUES, INTEGER], RP_INVESTMENT_ACCOUNT]
			-- Account dailyvalues cache

	account_dv (a_account: RP_INVESTMENT_ACCOUNT): DS_HASH_TABLE [RP_PERFORMANCE_DAILY_VALUES, INTEGER] is
			-- Dailyvalues for `a_account' hashed by ccyymmdd
		do
			retrieve_account_dailyvalues (a_account)
			account_dv_cache.search (a_account)
			if account_dv_cache.found then
				Result := account_dv_cache.found_item
			end
		end

	retrieve_account_dailyvalues (a_account: RP_INVESTMENT_ACCOUNT) is
			-- Retrieve dailyvalues for `a_account'
		local
			l_dv: RP_INVESTMENT_ACCOUNT_DAILY_VALUES
			l_dvs: like account_dv
		do
			if account_cached = Void then
				create account_cached.make_equal (17)
			end
			if account_dv_cache = Void then
				create account_dv_cache.make_equal (17)
			end
			if not account_cached.has (a_account) then
				create l_dv.make (Void, a_account, fx, agent portfolio_dv (?), a_account.investment_region.legacy_perf_calendar, ignore_portfolio_dailyvalues_check)
				create l_dvs.make_default
				l_dv.daily_values.do_all (agent add_account_dailyvalues (l_dvs, ?))
				account_dv_cache.force (l_dvs, a_account)
				account_cached.force (a_account)
			end
		end

	add_account_dailyvalues (a_dvs: like account_dv; a_dv: RP_PERFORMANCE_DAILY_VALUES) is
			-- Add `a_dv' to `a_dvs'
		require
			non_void_args: a_dvs /= Void and a_dv /= Void
		do
			a_dvs.force (a_dv, a_dv.perf_date.ccyymmdd)
		end

	add_daily_return (a_daily_returns: RP_PERFORMANCE_DAILY_RETURNS; a_daily_values: RP_PERFORMANCE_DAILY_VALUES) is
			-- Add `a_daily_values' to `a_daily_returns'
		require
			non_void_args: a_daily_returns /= Void and a_daily_values /= Void
		do
			if a_daily_values.portfolio_value.amount <= 0 then
				manager.log_minor ("Zero or negative value for " + a_daily_returns.code + " on " + a_daily_values.perf_date.ccyymmdd.out)
			end
			a_daily_returns.add_dailyvalues (a_daily_values)
		end

	db_bogie_dailies (a_benchmark: RP_BENCHMARK; a_date: ROSE_DATE): PAIR [RP_DECIMAL, RP_CURRENCY] is
			-- Daily Benchmark return for 'a_benchmark' on 'a_date'
		require
			non_void_args: a_benchmark /= Void and a_date /= Void
		local
			l_returns: like retrieve_benchmark_daily_returns
		do
			l_returns := retrieve_benchmark_daily_returns (a_benchmark)
			l_returns.search (a_date)
			if l_returns.found then
				Result := l_returns.found_item
			end
		end

	db_cum_bogie_dailies (a_benchmark: RP_BENCHMARK; a_start_date, a_end_date: ROSE_DATE): PAIR [RP_DECIMAL, RP_CURRENCY] is
			-- Interval_return calculated from cumulative benchmark daily returns
		require
			non_void_args: a_benchmark /= Void and a_start_date /= Void and a_end_date /= Void
		local
			l_returns: like retrieve_cum_benchmark_daily_returns
			l_start_cum, l_end_cum: PAIR [RP_DECIMAL, RP_CURRENCY]
		do
			l_returns := retrieve_cum_benchmark_daily_returns (a_benchmark)
			l_start_cum := extract_return (l_returns, a_start_date)
			l_end_cum := extract_return (l_returns, a_end_date)
			if l_start_cum /= Void and then not l_start_cum.first.is_zero and then l_end_cum /= Void then
				create Result.make (((l_end_cum.first / l_start_cum.first) - 1) * 100, l_end_cum.second)
			end
		end

	opb_monthly_return (a_benchmark: RP_BENCHMARK; a_month: ROSE_MONTH): PAIR [RP_DECIMAL, RP_CURRENCY] is
			-- Official Benchmark return for `a_benchmark' and `a_month'
		require
			non_void_args: a_benchmark /= Void and a_month /= Void
		local
			l_returns: like retrieve_benchmark_monthly_returns
		do
			l_returns := retrieve_benchmark_monthly_returns (a_benchmark)
			l_returns.search (a_month)
			if l_returns.found then
				Result := l_returns.found_item
			end
		end

	extract_return (a_table: like retrieve_cum_benchmark_daily_returns; a_date: ROSE_DATE): PAIR [RP_DECIMAL, RP_CURRENCY] is
			-- extract return for 'a_date'
		do
			a_table.search (a_date)
			if a_table.found then
				Result := a_table.found_item
			end
		end

	benchmark_cum_daily_returns_cache: DS_HASH_TABLE [like retrieve_cum_benchmark_daily_returns, RP_BENCHMARK]
			-- Cache for cum daily benchmark returns

	benchmark_daily_returns_cache: DS_HASH_TABLE [like retrieve_benchmark_daily_returns, RP_BENCHMARK]
			-- Cache for daily benchmark returns

	benchmark_returns_cache: DS_HASH_TABLE [like retrieve_benchmark_monthly_returns, RP_BENCHMARK]
			-- Cache for benchmark returns

	retrieve_benchmark_daily_returns (a_benchmark: RP_BENCHMARK): DS_HASH_TABLE [PAIR [RP_DECIMAL, RP_CURRENCY], ROSE_DATE] is
			-- Retrieve benchmark daily returns for 'a_benchmark'
		local
			l_query: like benchmark_dailyreturns_query
		do
			if benchmark_daily_returns_cache = Void then
				create benchmark_daily_returns_cache.make_default
			end
			benchmark_daily_returns_cache.search (a_benchmark)
			if benchmark_daily_returns_cache.found then
				Result := benchmark_daily_returns_cache.found_item
			else
				create Result.make_default
				benchmark_daily_returns_cache.force (Result, a_benchmark)
				l_query := benchmark_dailyreturns_query
				l_query.retrieve (Void, <<a_benchmark.code>>)
				l_query.array.do_all (agent parse_benchmark_daily_return (Result, ?))
			end
		end

	retrieve_cum_benchmark_daily_returns (a_benchmark: RP_BENCHMARK): DS_HASH_TABLE [PAIR [RP_DECIMAL, RP_CURRENCY], ROSE_DATE] is
			-- Retrieve benchmark daily returns for 'a_benchmark'
		local
			l_query: like benchmark_cum_dailyreturns_query
		do
			if benchmark_cum_daily_returns_cache = Void then
				create benchmark_cum_daily_returns_cache.make_default
			end
			benchmark_cum_daily_returns_cache.search (a_benchmark)
			if benchmark_cum_daily_returns_cache.found then
				Result := benchmark_cum_daily_returns_cache.found_item
			else
				create Result.make_default
				benchmark_cum_daily_returns_cache.force (Result, a_benchmark)
				l_query := benchmark_cum_dailyreturns_query
				l_query.retrieve (Void, <<a_benchmark.code>>)
				l_query.array.do_all (agent parse_cum_benchmark_daily_return (Result, ?))
			end
		end

	retrieve_benchmark_monthly_returns (a_benchmark: RP_BENCHMARK): DS_HASH_TABLE [PAIR [RP_DECIMAL, RP_CURRENCY], ROSE_MONTH] is
			-- Retrieve benchmark monthly returns for `a_benchmark'
		local
			l_query: like benchmark_monthlyreturns_query
		do
			if benchmark_returns_cache = Void then
				create benchmark_returns_cache.make_default
			end
			benchmark_returns_cache.search (a_benchmark)
			if benchmark_returns_cache.found then
				Result := benchmark_returns_cache.found_item
			else
				create Result.make_default
				benchmark_returns_cache.force (Result, a_benchmark)
				l_query := benchmark_monthlyreturns_query
				l_query.retrieve (Void, <<a_benchmark.code>>)
				l_query.array.do_all (agent parse_benchmark_monthly_return (Result, ?))
			end
		end

	parse_cum_benchmark_daily_return (a_returns: like retrieve_benchmark_daily_returns; s: STRING) is
			-- Parse cumulative benchmark daily return data from 's' and add to 'a_return'
		require
			non_void_args: a_returns /= Void and s /= Void
		local
			l_parser: ROSE_DELIMITED_TEXT
			l_date: ROSE_DATE
			l_return: RP_DECIMAL
			l_currency: RP_CURRENCY
		do
			create l_parser.make (s, '|')
			create l_date.make_ccyymmdd (l_parser.item (1).to_integer_32)
			create l_return.make_from_string (l_parser.item (2))
			l_currency := manager.reference_system.currency_from_iso_code (l_parser.item (3))
			a_returns.force (create {PAIR [RP_DECIMAL, RP_CURRENCY]}.make (l_return, l_currency), l_date)
		end

	parse_benchmark_daily_return (a_returns: like retrieve_benchmark_daily_returns; s: STRING) is
			-- Parse benchmark daily return data from 's' and add to 'a_return'
		require
			non_void_args: a_returns /= Void and s /= Void
		local
			l_parser: ROSE_DELIMITED_TEXT
			l_date: ROSE_DATE
			l_return: RP_DECIMAL
			l_currency: RP_CURRENCY
		do
			create l_parser.make (s, '|')
			create l_date.make_ccyymmdd (l_parser.item (1).to_integer_32)
			create l_return.make_from_string (l_parser.item (2))
			l_return := l_return - 1
			l_return := l_return * 100
			l_currency := manager.reference_system.currency_from_iso_code (l_parser.item (3))
			a_returns.force (create {PAIR [RP_DECIMAL, RP_CURRENCY]}.make (l_return, l_currency), l_date)
		end

	parse_benchmark_monthly_return (a_returns: like retrieve_benchmark_monthly_returns; s: STRING) is
			-- Parse benchmark monthly return data from `s' and add to `a_return'
		require
			non_void_args: a_returns /= Void and s /= Void
		local
			l_parser: ROSE_DELIMITED_TEXT
			l_month: ROSE_MONTH
			l_return: RP_DECIMAL
			l_currency: RP_CURRENCY
		do
			create l_parser.make (s, '|')
			create l_month.make_ccyymm (l_parser.item (1).to_integer_32)
			create l_return.make_from_string (l_parser.item (2))
			l_return := l_return / 100
			l_return := l_return + 1
			l_currency := manager.reference_system.currency_from_iso_code (l_parser.item (3))
			a_returns.force (create {PAIR [RP_DECIMAL, RP_CURRENCY]}.make (l_return, l_currency), l_month)
		end

	benchmark_monthlyreturns_query: STRING_PREDICATE is
			-- Benchmark database query
		do
			create Result.make ((create {DATABASE_ACCESS}).make_from_repository_fast_load ("performance"), "[
		SELECT cast (ccyymm as varchar(6)) + '|' + cast([return] as varchar(50)) + '|' + iso
			FROM vw_merged_benchmarks
																											WHERE bogiename = ?
																											]")
		end

	benchmark_cum_dailyreturns_query: STRING_PREDICATE is
			-- NPERF query to cumulative_daily_benchmark_returns
		do
			create Result.make ((create {DATABASE_ACCESS}).make_from_repository_fast_load ("performance"), "[
																											dbo.cum_benchmark_dailies ?
																											]")
		end

	benchmark_dailyreturns_query: STRING_PREDICATE is
			-- NPERF DailyReturns query
		do
			create Result.make ((create {DATABASE_ACCESS}).make_from_repository_fast_load ("performance"), "[
		SELECT cast (dbo.convert_datetime_2_ccyymmdd (perfdate) AS varchar(8)) + '|' + cast([return] as varchar(50)) + '|' + iso
			FROM DailyReturns
																											WHERE bogiename = ?
																											]")
		end

	official_perf_query: COLLECTING_PREDICATE [MONTHLY_PERF_ROW] is
			-- Portfolio/account OfficialPerf query
		once
			create Result.make (op_server, "SELECT * FROM fc_portfolio_bogie_monthlyperf_reportable_view WHERE zone= ? and portfolio = ? and iso_fc = ?")
		end

	fc_composite_bogie_monthlyperf_reportable_view: COLLECTING_PREDICATE [MONTHLY_PERF_ROW] is
			-- Portfolio/account OfficialPerf query
		once
			create Result.make (op_server, "SELECT zone, portfolio, reporting_year, reporting_month, iso_fc, fc_performance, fc_benchmark_performance, fc_assetvalue FROM fc_composite_bogie_monthlyperf_reportable_view WHERE  zone= ? and portfolio = ? and iso_fc = ?")
		end

	official_hedged_perf_query: COLLECTING_PREDICATE [MONTHLY_PERF_ROW] is
			-- Hedged account OfficialPerf query
		once
			create Result.make (op_server, "SELECT * FROM fc_hedged_portfolio_bogie_monthlyperf_reportable_view WHERE zone= ? and portfolio = ? and iso_fc = ?")
		end

end
