indexing

	description:

		"Government bonds (tbill mosstly) specific pricing algorithm"

	date: "$Date: 2010/04/30 $"
	fileid: "$Id: //rose/source/kernel/expert_system/valuators/rp_performance_tbill_pricing_algorithm.e#3 $"
	keywords: "price,value,valuation"

class RP_PERFORMANCE_TBILL_PRICING_ALGORITHM

inherit

	RP_PRICING_ALGORITHM [RP_GOVERNMENT_BOND]

	RP_MARKET_DATA_FIELD_AWARE
		export {NONE} all end

create

	make

feature {NONE} -- Initialization

	make (a_trade_price: like trade_price; a_market_data_source: like market_data_source; a_price_override: like price_override) is
			-- Initialize `Current'
		require
			non_void_args: a_trade_price /= Void and a_market_data_source /= Void and a_price_override /= Void
		do
			trade_price := a_trade_price
			market_data_source := a_market_data_source
			price_override := a_price_override
			create last_price_cache.make_equal (17)
		ensure
			market_data_source_set: market_data_source = a_market_data_source
		end

feature -- Query

	price (a_tbill: RP_GOVERNMENT_BOND; a_date: ROSE_DATE): RP_PRICE is
			-- Price of `a_tbill' on `a_date'
		local
			l_record: RP_MARKET_DATA_RECORD
			l_adjustment_now, l_adjustment_then: RP_DECIMAL
			l_price_lag: TRIPLE [RP_PRICE, BOOLEAN, RP_DECIMAL]
			l_price: RP_PRICE
		do
				-- Check for price override first
			l_record := price_override.snapshot (a_tbill, a_date)
			if l_record /= Void and then l_record.has_close then
				Result := l_record.unadjusted_close
				create l_price_lag.make (Result, False, Void)
				last_price_cache.force (l_price_lag, a_tbill)
			else
					-- Retrieve market price if there is no override price
				if Result = Void then
					l_record := market_data_source.snapshot (a_tbill, a_date)
					if l_record /= Void then
						l_adjustment_now := l_record.adjustment_factor
					else
						l_adjustment_now := 1
					end
					Result := market_price (a_tbill, a_date)
						-- Update lag price
					if Result /= Void then
						create l_price_lag.make (Result, True, l_adjustment_now)
						last_price_cache.force (l_price_lag, a_tbill)
					end
				end
					-- No market price, check trade price
				if Result = Void then
						-- Last resort is trade price
					l_record := trade_price.snapshot (a_tbill, a_date)
					if l_record /= Void and then l_record.has_close then
						Result := l_record.unadjusted_close
						create l_price_lag.make (Result, False, Void)
						last_price_cache.force (l_price_lag, a_tbill)
					end
				end
					-- Use lag price if everything else fails
				if Result = Void then
					last_price_cache.search (a_tbill)
					if last_price_cache.found then
						l_price_lag := last_price_cache.found_item
						l_price := l_price_lag.first
						l_adjustment_then := l_price_lag.third
						if l_price_lag.second then
								-- Lag price was a market price
							Result := a_tbill.price_in_currency (l_price.price * l_adjustment_now / l_adjustment_then, l_price.currency, l_price.date)
						else
							Result := a_tbill.price_in_currency (l_price.price, l_price.currency, l_price.date)
							l_adjustment_now := l_adjustment_then
						end
						create l_price_lag.make (Result, l_price_lag.second, l_adjustment_now)
						last_price_cache.force (l_price_lag, a_tbill)
					end
				end
			end
		end

feature {NONE} -- Implementation

	last_price_cache: UID_TABLE [TRIPLE [RP_PRICE, BOOLEAN, RP_DECIMAL], RP_INSTRUMENT]
			-- Last computed price cache

	market_data_source: RP_MARKET_DATA_SOURCE [RP_GOVERNMENT_BOND]
			-- Market data source

	price_override: RP_MARKET_DATA_SOURCE [RP_INSTRUMENT]
			-- Price override market data source

	trade_price: RP_TRADE_PRICE_MARKET_DATA_SOURCE
			-- Trade price market data source

	market_price (a_instrument: RP_GOVERNMENT_BOND; a_date: ROSE_DATE): RP_PRICE is
			-- Retrieve market price for `a_instrument' on `a_date'
		require
			non_void_args: a_instrument /= Void and a_date /= Void
		local
			l_close: RP_MARKET_DATA_RECORD
			l_bid: RP_MARKET_DATA_RECORD
			l_now: RP_MARKET_DATA_RECORD
			l_price: RP_PRICE
			l_adjustment_then: RP_DECIMAL
			l_adjustment_now: RP_DECIMAL
		do
				-- Obtain last close
			l_close := market_data_source.last_record_with_valid_field (a_instrument, a_date, unadjusted_close_type)
				-- Obtain last bid
			l_bid := market_data_source.last_record_with_valid_field (a_instrument, a_date, unadjusted_bid_type)
			if l_close /= Void and l_bid /= Void then
				if l_close.date < l_bid.date then
						-- Bid is more recent
					l_price := l_bid.unadjusted_bid
					l_adjustment_then := l_bid.adjustment_factor
				else
						-- Close is more recent
					l_price := l_close.unadjusted_close
					l_adjustment_then := l_close.adjustment_factor
				end
			elseif l_close /= Void then
					-- Close is present
				l_price := l_close.unadjusted_close
				l_adjustment_then := l_close.adjustment_factor
			elseif l_bid /= Void then
					-- Bid is present
				l_price := l_bid.unadjusted_bid
				l_adjustment_then := l_bid.adjustment_factor
			end
			if l_price /= Void then
				l_now := market_data_source.snapshot (a_instrument, a_date)
				if l_now /= Void and then l_now.has_adjustment_factor then
					l_adjustment_now := l_now.adjustment_factor
				else
					l_adjustment_now := 1
				end
				if l_adjustment_then = Void then
					l_adjustment_then := 1
				end
				l_price := a_instrument.price_in_currency (l_price.price * l_adjustment_now / l_adjustment_then, l_price.currency, l_price.date)
			end
			Result := l_price
		end

end
