indexing

	description:

		"U zone specific pricing algorithm"

	date: "$Date: 2010/04/30 $"
	fileid: "$Id: //rose/source/kernel/expert_system/valuators/rp_performance_uzone_pricing_algorithm.e#10 $"

class RP_PERFORMANCE_UZONE_PRICING_ALGORITHM

inherit

	RP_PRICING_ALGORITHM [RP_STOCK]

	RP_MARKET_DATA_FIELD_AWARE
		export {NONE} all end

create

	make

feature {NONE} -- Initialization

	make (a_trade_price: like trade_price; a_market_data_source: like market_data_source; a_price_override: like price_override; a_infinite_lookback: like infinite_lookback; a_dont_use_override: like dont_use_override) is
			-- Initialize `Current'
		require
			non_void_args: a_trade_price /= Void and a_market_data_source /= Void and a_price_override /= Void
		do
			trade_price := a_trade_price
			market_data_source := a_market_data_source
			price_override := a_price_override
			dont_use_override := a_dont_use_override
			infinite_lookback := a_infinite_lookback
			create last_price_cache.make_equal (17)
		ensure
			market_data_source_set: market_data_source = a_market_data_source
		end

feature -- Query

	price (a_stock: RP_STOCK; a_date: ROSE_DATE): RP_PRICE is
			-- Price of `a_stock' on `a_date'
		local
			l_record: RP_MARKET_DATA_RECORD
			l_adjustment_now, l_adjustment_then: RP_DECIMAL
			l_price_lag: PAIR [RP_PRICE, RP_DECIMAL]
			l_price: RP_PRICE
		do
				-- Check price overide if enabled
			if not dont_use_override then
				l_record := price_override.snapshot (a_stock, a_date)
				if l_record /= Void and then l_record.has_close then
					Result := l_record.unadjusted_close
					Result.set_source ("PO")
					create l_price_lag.make (Result, Void)
					last_price_cache.force (l_price_lag, a_stock)
				end
			end
				-- Retrieve market price. U zone does not use override price
			if Result = Void then
				l_record := market_data_source.snapshot (a_stock, a_date)
				if l_record /= Void then
					l_adjustment_now := l_record.adjustment_factor
				else
					l_adjustment_now := 1
				end
				Result := stock_market_price (a_stock, a_date)
					-- Update lag price
				if Result /= Void then
					create l_price_lag.make (Result, l_adjustment_now)
					last_price_cache.force (l_price_lag, a_stock)
				end
			end
				-- No market price, check trade price
			if Result = Void then
					-- Last resort is trade price
				l_record := trade_price.snapshot (a_stock, a_date)
				if l_record /= Void and then l_record.has_close then
					Result := l_record.unadjusted_close
					Result.set_source ("TP")
					create l_price_lag.make (Result, Void)
					last_price_cache.force (l_price_lag, a_stock)
				end
			end
				-- Use lag price if everything else fails
			if Result = Void then
				last_price_cache.search (a_stock)
				if last_price_cache.found then
					l_price_lag := last_price_cache.found_item
					l_price := l_price_lag.first
					l_adjustment_then := l_price_lag.second
					if l_adjustment_now /= Void and l_adjustment_then /= Void then
							-- Lag price was a market price
						Result := a_stock.price_in_currency (l_price.price * l_adjustment_now / l_adjustment_then, l_price.currency, l_price.date)
						Result.set_source ("LP")
					else
						Result := a_stock.price_in_currency (l_price.price, l_price.currency, l_price.date)
						Result.set_source ("uLP")
						l_adjustment_now := l_adjustment_then
					end
					create l_price_lag.make (Result, l_adjustment_now)
					last_price_cache.force (l_price_lag, a_stock)
				end
			end
		end

feature {NONE} -- Implementation

	infinite_lookback: BOOLEAN
			-- Infinite price lookback

	dont_use_override: BOOLEAN
			-- Do not use oprice override

	last_price_cache: UID_TABLE [PAIR [RP_PRICE, RP_DECIMAL], RP_INSTRUMENT]
			-- Last computed price cache

	market_data_source: RP_MARKET_DATA_SOURCE [RP_STOCK]
			-- Market data source

	price_override: RP_MARKET_DATA_SOURCE [RP_INSTRUMENT]
			-- Price override market data source

	trade_price: RP_TRADE_PRICE_MARKET_DATA_SOURCE
			-- Trade price market data source

	stock_market_price (a_instrument: RP_STOCK; a_date: ROSE_DATE): RP_PRICE is
			-- Retrieve market price for `a_instrument' on `a_date'
		require
			non_void_args: a_instrument /= Void and a_date /= Void
		local
			l_close: RP_MARKET_DATA_RECORD
			l_bid: RP_MARKET_DATA_RECORD
			l_now: RP_MARKET_DATA_RECORD
			l_price: RP_PRICE
			l_adjustment_then: RP_DECIMAL
			l_adjustment_now: RP_DECIMAL
			s: STRING
		do
				-- Obtain last close
			l_close := market_data_source.last_record_with_valid_field (a_instrument, a_date, unadjusted_close_type)
			if l_close /= Void then
				if not l_close.has_volume then
					l_close := market_data_source.last_record_with_valid_field (a_instrument, l_close.date, unadjusted_close_type)
				elseif l_close.has_volume and then l_close.volume /= Void and then l_close.volume.is_zero then
					l_close := market_data_source.last_record_with_valid_field (a_instrument, l_close.date, unadjusted_close_type)
				end
			end
			l_close := filter (l_close, a_date)
				-- Obtain last bid
			l_bid := market_data_source.last_record_with_valid_field (a_instrument, a_date, unadjusted_bid_type)
			l_bid := filter (l_bid, a_date)
			if l_close /= Void and l_bid /= Void then
				if l_close.date < l_bid.date then
						-- Bid is more recent
					l_price := l_bid.unadjusted_bid
					if l_bid.has_volume and then l_bid.volume.is_zero then
						s := "b"
					else
						s := "B"
					end
					if not l_bid.date.is_equal (a_date) then
						s.append_integer (l_bid.date.diff_in_days (a_date))
					end
					l_adjustment_then := l_bid.adjustment_factor
				else
						-- Close is more recent
					l_price := l_close.unadjusted_close
					if l_close.has_volume and then l_close.volume.is_zero then
						s := "c"
					else
						s := "C"
					end
					if not l_close.date.is_equal (a_date) then
						s.append_integer (l_close.date.diff_in_days (a_date))
					end
					l_adjustment_then := l_close.adjustment_factor
				end
			elseif l_close /= Void then
					-- Close is present
				l_price := l_close.unadjusted_close
				if l_close.has_volume and then l_close.volume.is_zero then
					s := "c"
				else
					s := "C"
				end
				if not l_close.date.is_equal (a_date) then
					s.append_integer (l_close.date.diff_in_days (a_date))
				end
				l_adjustment_then := l_close.adjustment_factor
			elseif l_bid /= Void then
					-- Bid is present
				l_price := l_bid.unadjusted_bid
				if l_bid.has_volume and then l_bid.volume.is_zero then
					s := "b"
				else
					s := "B"
				end
				if not l_bid.date.is_equal (a_date) then
					s.append_integer (l_bid.date.diff_in_days (a_date))
				end
				l_adjustment_then := l_bid.adjustment_factor
			end
			if l_price /= Void then
				l_now := market_data_source.snapshot (a_instrument, a_date)
				if l_now /= Void and then l_now.has_adjustment_factor then
					l_adjustment_now := l_now.adjustment_factor
				end
				if l_adjustment_now /= Void and l_adjustment_then /= Void then
					l_price := a_instrument.price_in_currency (l_price.price * l_adjustment_now / l_adjustment_then, l_price.currency, a_date)
				else
					s.prepend ("u")
					l_price := a_instrument.price_in_currency (l_price.price, l_price.currency, a_date)
				end
				l_price.set_source (s)
			end
			Result := l_price
		end

	filter (a_record: RP_MARKET_DATA_RECORD; a_date: ROSE_DATE): RP_MARKET_DATA_RECORD is
			-- Check whether `a_record' is not too far from `a_date'
		require
			non_void_date: a_date /= Void
		do
			if a_record /= Void then
				if infinite_lookback then
					if a_date.ccyymmdd > 20081201 then
							-- After Aug 2008 RMDAT_INTERNAL daylp goes back 160 days.
						if a_date.diff_in_days (a_record.date) <= 160 then
							Result := a_record
						end
					else
							-- Before Aug 2008 RMDAT_INTERNAL daylp goes back to dawn of time
						Result := a_record
					end
				else
						-- 60 days lookback
					if a_date.diff_in_days (a_record.date) <= 60 then
						Result := a_record
					end
				end
			end
		end

end
