note

	description:
	"[
		Permutations of a range of size n
		
		This class enumerates all possible permutations in lexicographic ordering
		begining with the trivial permutation of {1,2,...,n}
		Sample usage is:
			from
				create p.make_permutation (n)
			until
				p /= Void
			loop
				... use p...
				p := p.next
			end

	]"

	author: "Ivan Tarnavsky"
	last_editor: "$Author: it $"
	date: "$Date: 2011/02/23 $"
	revision: "$Revision: #1 $"
	path: "$File: //rose/source/kernel/math/rp_permutations.e $"

class RP_PERMUTATIONS

inherit

	ARRAY [INTEGER]
		rename
			count as n
		export
			{ANY} n, item, do_all, do_all_with_index, do_if, do_if_with_index, valid_index, upper, lower
			{RP_PERMUTATIONS, RP_VARIATIONS} area, deep_twin, is_deep_equal, standard_is_equal, subarray, object_comparison
			{NONE} all
		undefine
			is_equal
		end

	COMPARABLE
		undefine
			copy
		end

create

	make_permutation

create {RP_PERMUTATIONS, RP_VARIATIONS}

	make_from_array

feature {NONE} -- Initialization

	make_permutation (a_n: like n) is
			-- Initialize `Current'
		require
			n_is_positive: a_n > 0
		do
				-- Initial permutation P={1, 2, ..., n}
			make_from_array ((1 |..| a_n).as_array)
		ensure
			n_set: n = a_n
			initial_values: (1 |..| a_n).as_array.area.is_equal (area)
		end

feature -- Access

	next: RP_PERMUTATIONS is
			-- Next permutation, Void if not possible
		local
			i, j, k, t: INTEGER
			r: ARRAY [INTEGER]
		do
			from
					-- find the largest index i so that P(i)<P(i + 1)
				i := n - 1
			until
				i = 0 or else item (i) < item (i + 1)
			loop
				i := i - 1
			end
			if i /= 0 then
				create r.make_from_array (twin)
				from
						-- find the largest element after P(i) but not larger than P(i)
					k := n
				until
					r.item (i) <= r.item (k)
				loop
					k := k - 1
				end
					-- swap P(i) and P(k)
				t := r.item(i)
				r.put (r.item (k), i)
				r.put (t, k)
				from
						-- swap the last n - i elements
					k := n
					j := i + 1
				until
					j >= k
				loop
						-- swap P(j) and P(k)
					t := r.item (j)
					r.put (r.item (k), j)
					r.put (t, k)
					k := k - 1
					j := j + 1
				end
				create Result.make_from_array (r)
			end
		ensure
			definition: Result = Void or else Current < Result
		end

feature -- Comparison

	is_less alias "<" (other: like Current): BOOLEAN
			-- <Precursor>
		local
			i: INTEGER
		do
			if n < other.n then
				Result := True
			elseif n = other.n then
				from
					i := lower
				until
					i > upper or Result
				loop
					Result := Result or item (i) < other.item (i)
					i := i + 1
				end
			end
		end

feature -- Status

	valid_values (a_n: INTEGER; a_r: ARRAY [INTEGER]): BOOLEAN is
			-- Are all values in `a_r' valid for `a_n'?
		local
			i: INTEGER
		do
			if a_n > 0 and a_r /= Void and then a_r.count = a_n then
				from
					Result := True
					i := a_r.lower
				until
					i > a_r.upper or not Result
				loop
					Result := Result and 0 < a_r.item (i) and a_r.item (i) <= a_n
					i := i + 1
				end
			end
		end

	has_repetitions: BOOLEAN
			-- Are there repeating values in `Current'?
		local
			l_set: DS_HASH_SET [INTEGER]
		do
			create l_set.make (n)
			do_all (agent l_set.force (?))
			Result := n /= l_set.count
		end

invariant

	n_positive: n > 0
	valid_values: valid_values (n, Current)
	without_repetitions: not has_repetitions

end
