note

	description:

		"Basic investment entity"

	author: "Ivan Tarnavsky"
	last_editor: "$Author: sm $"
	date: "$Date: 2013/04/26 $"
	revision: "$Revision: #98 $"
	path: "$File: //rose/source/kernel/expert_system/rp_portfolio.e $"

class RP_PORTFOLIO

inherit

	RP_SUB_ACCOUNT [RP_INVESTMENT_ACCOUNT]
		rename
			parent as account
		end

	RP_ZONAL_INVESTMENT_GROUP
		rename
			investment_region as zone
		end

	RP_UNKNOWN_CUSTODY_AWARE
		export
			{NONE} all
		undefine
			is_equal
		end

	RP_UNSPECIFIED_POSSESSION_AWARE
		export
			{NONE} all
		undefine
			is_equal
		end

	RP_UNKNOWN_PARTY_AWARE
		export
			{NONE} all
		undefine
			is_equal
		end

create {RP_REFERENCE_SYSTEM}

	make

feature {NONE} -- Initialization

	make (a_ref: like reference_system; a_gcdb_portfolio: like gcdb_portfolio; a_first_event_date: ROSE_DATE; a_fund_participant: like gcdb_fund_participant) is
				-- Creation procedure
		require
			ref_not_void: a_ref /= Void
		local
			l_first_event_date: ROSE_DATE
		do
			reference_system := a_ref
			gcdb_portfolio := a_gcdb_portfolio
			gcdb_fund_participant := a_fund_participant
			if gcdb_fund_participant /= Void then
				initialize_from_gcdb_fund_participant
			elseif gcdb_portfolio /= Void then
				initialize_from_gcdb_portfolio
			else
				initialize_unknown_portfolio
			end

			if (a_first_event_date /= Void) then
				start_date := a_first_event_date
			else
					-- Set start_date to first transaction date or default GCDB date
				if a_gcdb_portfolio /= Void then
					l_first_event_date := a_gcdb_portfolio.first_event_date
					if l_first_event_date /= Void then
						create start_date.make_by_date (l_first_event_date)
					else
						start_date := a_gcdb_portfolio.start_date
					end
				end
			end

			if start_date = Void then
				create start_date.make_ccyymmdd (99991231)
			end
			if end_date = Void then
				create end_date.make_ccyymmdd (99991231)
			end
			if end_date < start_date then
				create end_date.make_ccyymmdd (start_date.ccyymmdd)
			end
			create saved_instrument_control.make_equal (2)
		ensure
			start_date_not_void: start_date /= Void
			end_date_not_void: end_date /= Void
		end

	initialize_unknown_portfolio is
			-- Set default values for "Unknown" portfolio
		do
			is_unknown := True
			set_code ({RP_UNKNOWN}.code)
			name := code
			create benchmark_history_saved.make_empty
		end

	initialize_from_gcdb_portfolio is
			-- Set feature values from `gcdb_portfolio'
		require
			gcdb_portfolio_exists: gcdb_portfolio /= Void
		do
			start_date := gcdb_portfolio.start_date
			end_date := gcdb_portfolio.end_date
			set_code (clean_code (gcdb_portfolio.code))
			if gcdb_portfolio.name = Void then
				create name.make_from_string (code)
			else
				create name.make_from_string (gcdb_portfolio.name)
			end
		ensure
			valid_code: code /= Void and then not code.is_empty
		end

	initialize_from_gcdb_fund_participant is
			-- Set feature values from `gcdb_fund_participant'
		require
			gcdb_entity_exists: gcdb_fund_participant /= Void
		do
			start_date := reference_system.fund_participant_start_date (gcdb_fund_participant.brimsym)
			end_date := reference_system.fund_participant_end_date (gcdb_fund_participant.brimsym)
			set_code (clean_code (gcdb_fund_participant.code))
			if gcdb_fund_participant.name = Void then
				create name.make_from_string (code)
			else
				create name.make_from_string (gcdb_fund_participant.name)
			end
		ensure
			valid_code: code /= Void and then not code.is_empty
		end

feature -- Access

	code: STRING
			-- Unique code

	name: STRING
			-- Descriptive name

	start_date: ROSE_DATE
			-- Start date

	end_date: ROSE_DATE
			-- End date

	is_unknown: BOOLEAN
			-- Is `Current' instance the "Unknown" portfolio? (Void Object Pattern)

	last_known_name: STRING is
			-- Last known descriptive name
		do
			if gcdb_portfolio /= Void then
				Result := gcdb_portfolio.last_known_name
			elseif gcdb_fund_participant /= Void then
				Result := gcdb_fund_participant.last_known_name
			else
				Result := ""
			end
		end

	reporting_currencies: ARRAY [RP_CURRENCY] is
			-- List of reporting currencies
		do
			retrieve_reporting_currencies
			Result := saved_reporting_currencies.to_array
		end

	performance_code: STRING is
			-- <Precursor>
		do
			Result := code
		end

	gcdb_handle: INTEGER is
			-- Gcdb handle
		do
			if gcdb_portfolio /= Void then
				Result := gcdb_portfolio.handle
			end
		end

feature -- Change

	set_code (a_code: like code) is
			-- <Precursor>
		do
			code := a_code
		end

feature -- Query

	debug_output: STRING is
			-- Debug output.
		do
			Result := code
		end

	member_id: STRING is
			-- OLAP `member_id' of `Current'
		do
			Result := code
		end

	owner: RP_PARTY is
			-- Party owning `Current'
		do
			Result := account.client
		end

	account: RP_INVESTMENT_ACCOUNT is
			-- Investment account `Current' is part of
		do
			if account_saved = Void then
				if gcdb_portfolio /= Void and then not gcdb_portfolio.account_history.is_empty then
					account_saved := reference_system.investment_account_by_code (gcdb_portfolio.account_history.last.entity.code)
				elseif gcdb_fund_participant /= Void then
						-- Fund participanst have accounts with the same name
					account_saved := reference_system.investment_account_by_code (code)
				else
					account_saved := reference_system.unknown_investment_account
				end
			end
			Result := account_saved
		end

	zone: RP_ZONE is
			-- Investment region `Current' belongs to
		do
			if zone_saved = Void then
				if gcdb_portfolio /= Void and then not gcdb_portfolio.investment_region_history.is_empty then
					zone_saved := reference_system.investment_region_by_code (gcdb_portfolio.investment_region_history.first.entity.code)
				elseif gcdb_fund_participant /= Void then
						-- Fixed at "T"
					zone_saved := reference_system.investment_region_by_code ("T")
				else
					zone_saved := reference_system.unknown_investment_region
				end
			end
			Result := zone_saved
		end

	prime_broker: RP_BROKER is
			-- Prime broker for `Current'
		do
				-- to_be_implemented ("FIXME")
			Result := reference_system.unknown_broker
		ensure
			non_void_result: Result /= Void
		end

	benchmark_history: RP_BENCHMARK_HISTORY is
			-- Benchmark assignment to `Current'
		do
			if benchmark_history_saved = Void then
				create benchmark_history_saved.make_empty
				if gcdb_portfolio /= Void then
					gcdb_portfolio.benchmark_history.do_all (agent expand_benchmark_history (benchmark_history_saved, ?))
				end
				if gcdb_fund_participant /= Void then
					if fund_participant_composite_asset.portfolio.zone.code.is_equal ("T") then
						fund_participant_composite_asset.portfolio.account.benchmarks.do_all (agent add_to_benchmark_history (benchmark_history_saved, ?))
					else
						fund_participant_composite_asset.portfolio.benchmark_history.do_all (agent add_to_benchmark_history (benchmark_history_saved, ?))
					end
				end
			end
			Result := benchmark_history_saved
		ensure
			non_void_result: Result /= Void
		end

	base_currency (a_date: ROSE_DATE): RP_CURRENCY is
			-- Base currency of account this portfolio belongs to
		do
			Result := account.base_currencies.witness (a_date)
			if Result = Void then
				Result := reference_system.currency_from_iso_code ("USD")
			end
		ensure
			result_not_void: Result /= Void
		end

	instrument_control (a_custody: RP_CUSTODY): RP_INSTRUMENT_CONTROL is
			-- Instrument control for `Current' within `a_custody'
		require
			a_sub_account_not_void: a_custody /= Void
		local
			l_management: RP_MANAGEMENT
			l_possession: RP_POSSESSION
		do
			if not saved_instrument_control.has (a_custody) then
				create l_management.make (imc_engineering, Current)
				create l_possession.make (trading, Current)
				create Result.make_with_all_details (account.client, a_custody, l_management, l_possession)
				saved_instrument_control.force (Result, a_custody)
			else
				Result := saved_instrument_control.item (a_custody)
			end
		ensure
			non_void_instrument_control: Result /= Void
		end

	client_instrument_control: RP_INSTRUMENT_CONTROL is
			-- Instrument control owned by `Current's account client
		do
			Result := account.client.instrument_control
		end

	selection_universe: RP_SELECTION_UNIVERSE is
			-- Strategy for `Current'
		do
			if gcdb_portfolio /= Void and then not gcdb_portfolio.selection_universe_history.is_empty then
				Result := reference_system.selection_universe_from_code (gcdb_portfolio.selection_universe_history.last.entity.code)
			else
				Result := reference_system.Unknown_selection_universe
			end
		ensure
			non_void_result: Result /= Void
		end

	portfolio_engineer: STRING is
			-- Strategy engineer responsible for `Current'
		do
			if gcdb_portfolio /= Void then
				Result := gcdb_portfolio.portfolio_engineer
			end
		ensure
			portfolio_engineer_may_be_unknown: True
		end

	imc_engineering: RP_INVESTMENT_MANAGEMENT_COMPANY is
			-- Location of strategy engineering function for `Current'
		do
			if gcdb_portfolio /= Void and then not gcdb_portfolio.imc_engineering_history.is_empty then
				Result := reference_system.imc_by_code (gcdb_portfolio.imc_engineering_history.last.entity.code)
			else
				Result := reference_system.unknown_imc
			end
		ensure
			non_void_result: Result /= Void
		end

	operations: RP_INVESTMENT_MANAGEMENT_COMPANY is
			-- Operations IMC
		do
			if gcdb_portfolio /= Void and then not gcdb_portfolio.imc_operations_history.is_empty then
				Result := reference_system.imc_by_code (gcdb_portfolio.imc_operations_history.last.entity.code)
			else
				Result := reference_system.unknown_imc
			end
		ensure
			non_void_result: Result /= Void
		end

	trading: RP_INVESTMENT_MANAGEMENT_COMPANY is
			-- Trading IMC
		do
			if gcdb_portfolio /= Void and then not gcdb_portfolio.imc_trading_history.is_empty then
				Result := reference_system.imc_by_code (gcdb_portfolio.imc_trading_history.last.entity.code)
			else
				Result := reference_system.unknown_imc
			end
		ensure
			non_void_result: Result /= Void
		end

	first_performance_date: ROSE_DATE is
			-- First date of performance calculation
		do
			if {l_date: ROSE_DATE} benchmark_history.min_start_date then
				Result := l_date
			else
				create Result.make_max_date
			end
			Result := zone.inception_date.max (Result)
			Result := Result.max (start_date)
		ensure
			non_void_date: Result /= Void
		end

	last_performance_date: ROSE_DATE is
			-- Last date of performance calculation
		do
			if {l_date: ROSE_DATE} benchmark_history.max_end_date then
				Result := l_date
			else
				create Result.make_max_date
			end
			Result := Result.min (end_date)
		ensure
			non_void_date: Result /= Void
		end

	composite_state: RP_BOOLEAN_HISTORY is
			-- Composite/Product inclusion state
		do
			Result := retrieve_composite_state
		ensure
			non_void_result: Result /= Void
		end

	modified_portfolio (a_flows: RP_FLOWS): like Current is
			-- Portfolio after `a_flow' are applied
		do
				-- XXX To Be Implemented
		end

feature -- Status Reprot

	is_not_in_olap: BOOLEAN is
			-- Is `Current' instance excluded from OLAP AUM cube processing
			-- Can be used to exclude e.g. some balanced, fixed-income and test accounts
		do
			if gcdb_portfolio /= Void and then gcdb_portfolio.not_in_olap /= Void then
				Result := not gcdb_portfolio.not_in_olap.is_empty
			end
		end

	not_in_olap_value: STRING is
			-- Is `Current' instance excluded from OLAP AUM cube processing
			-- Can be used to exclude e.g. some balanced, fixed-income and test accounts
		do
			if gcdb_portfolio /= Void then
				Result := gcdb_portfolio.not_in_olap
			end
		end

	is_fund_participant: BOOLEAN is
			-- Is `Current' invested in pooled funds only?
		do
			Result := gcdb_fund_participant /= Void or else reference_system.is_fund_participant (Current)
		end

	is_long_short: BOOLEAN is
			-- Whether `Current' is a L/S portfolio
		local
			l_result: CELL [BOOLEAN]
		do
			create l_result.put (False)
			benchmark_history.do_all (agent check_long_short_strategy (l_result, ?))
			Result := l_result.item
		end

	is_limited_long_short: BOOLEAN is
			-- Whether `Current' is a limited L/S portfolio (e.g. 130/30 etc)
		local
			l_result: CELL [BOOLEAN]
		do
			create l_result.put (False)
			benchmark_history.do_all (agent check_lls_strategy (l_result, ?))
			Result := l_result.item
		end

	instrument_type: STRING is
			-- Instrument type
		do
			if gcdb_portfolio /= Void then
				if gcdb_portfolio.instrument_type_history.count > 0 then
					Result := gcdb_portfolio.instrument_type_history.last.entity.code
				end
			end
			if Result = Void then
				Result := "Unknown"
			end
		ensure
			non_void_result: Result /= Void
		end

	is_equity: BOOLEAN is
			-- Whether `Current' is equity portfolio?
		do
			Result := True
			if instrument_type.is_equal ("ADRE") then
				Result := False
			elseif instrument_type.is_equal ("BALANCED") then
				Result := False
			elseif instrument_type.is_equal ("BOND") then
				Result := False
			elseif instrument_type.is_equal ("CASH") then
				Result := False
			elseif instrument_type.is_equal ("CURRENCY HEDGE NON EQUITY ONLY") then
				Result := False
			elseif instrument_type.is_equal ("FUTURES") then
				Result := False
			elseif instrument_type.is_equal ("GDRE") then
				Result := False
			elseif instrument_type.is_equal ("REIT") then
				Result := False
			end
		end

	products: RP_PRODUCT_HISTORY is
			-- Product assignment to `Current'
		do
			if product_history_saved = Void then
				product_history_saved := reference_system.products_by_portfolio (Current)
			end
			Result := product_history_saved
		ensure
			non_void_result: Result /= Void
		end

	composite_asset: RP_POOL_FUND is
			-- Composite asset of fund participant
		do
			if gcdb_fund_participant /= Void then
				Result := fund_participant_composite_asset
			end
		ensure
			result_is_void_for_non_fund_participants: gcdb_fund_participant = Void implies Result = Void
		end

feature {RP_ZONAL_PRODUCT} -- Composite inclusion

	no_direct_composite_mapping: RP_BOOLEAN_HISTORY is
			-- Negative of direct composite mappingss
		do
			Result := retrieve_no_direct_composite_mapping
		ensure
			non_void_result: Result /= Void
		end

feature {NONE} -- Implementation

	gcdb_portfolio: GCDB_PORTFOLIO
			-- GCDB version of this portfolio

	gcdb_fund_participant: GCDB_FUND_PARTICIPANT
			-- Underlying fund participant entity

	saved_fund_participant_composite_asset: like fund_participant_composite_asset
			-- Underlying pool fund

	reference_system: RP_REFERENCE_SYSTEM
			-- Reference system

	product_history_saved: RP_PRODUCT_HISTORY
			-- Product assignments for `Current'

	saved_instrument_control: DS_HASH_TABLE [like instrument_control, RP_CUSTODY]
			-- Cached instrument control keyed by sub account

	zone_saved: like zone
			-- Cached zone

	account_saved: like account
			-- Cached account

	benchmark_history_saved: like benchmark_history
			-- Cached benchmark history

	clean_code (s: STRING): STRING is
			-- Clean `s' i.e. uppercase, remove all spaces
		do
			create Result.make_from_string (s)
			Result.to_upper
			Result.prune_all (' ')
			Result.prune_all ('%T')
		end

	fund_participant_composite_asset: RP_POOL_FUND is
			-- Pool fund for fund participant
		local
			l_instrument: RP_INSTRUMENT
		do
			if saved_fund_participant_composite_asset = Void then
				l_instrument := reference_system.instrument_from_brimsym (gcdb_fund_participant.brimsym)
				if {l_pool_fund: RP_POOL_FUND} l_instrument then
					saved_fund_participant_composite_asset := l_pool_fund
				end
			end
			Result := saved_fund_participant_composite_asset
		end

	expand_benchmark_history (a_history: like benchmark_history; a_interval: ENTITY_INTERVAL [GCDB_BENCHMARK]) is
			-- Add benchmark information from `a_interval' to `a_history'
		require
			non_void_args: a_history /= Void and a_interval /= Void
		local
			l_bogie: RP_BENCHMARK
			l_interval: RP_BENCHMARK_HISTORY_INTERVAL
			l_start, l_end: ROSE_DATE
		do
			if a_interval.entity /= Void then
				l_bogie := reference_system.benchmark_by_code (a_interval.entity.code)
			else
				l_bogie := reference_system.unknown_benchmark
			end
			if l_bogie /= Void then
				l_start := a_interval.start_date.max (start_date)
				l_end := a_interval.end_date.min (end_date)
				if l_start < l_end then
					create l_interval.make (l_bogie, l_start, l_end, a_interval.weight)
					a_history.extend (l_interval)
				end
			end
		end

	add_to_benchmark_history (a_history: like benchmark_history; a_interval: RP_BENCHMARK_HISTORY_INTERVAL) is
			-- Add benchmark information from `a_interval' to `a_history'
		require
			non_void_args: a_history /= Void and a_interval /= Void
		local
			l_interval: RP_BENCHMARK_HISTORY_INTERVAL
			l_start, l_end: ROSE_DATE
		do
			l_start := a_interval.start_date.max (start_date)
			l_end := a_interval.end_date.min (end_date)
			if l_start < l_end then
				create l_interval.make (a_interval.benchmark, l_start, l_end, a_interval.weight)
				a_history.extend (l_interval)
			end
		end

	composite_state_cache: like composite_state
			-- Cached composite state

	retrieve_composite_state: like composite_state is
			-- Retrieve composite state
		do
			if composite_state_cache = Void then
				create composite_state_cache.make_empty
				if gcdb_portfolio /= Void then
					gcdb_portfolio.is_in_composite_history.do_all (agent add_composite_state (composite_state_cache, ?))
				end
			end
			Result := composite_state_cache
		ensure
			non_void_result: Result /= Void
		end

	add_composite_state (a_cache: like composite_state; a_value: ENTITY_ATTRIBUTE_RECORD) is
			-- Add composite state information from `a_value' to `a_cache'
		require
			non_void_args: a_cache /= Void and a_value /= Void
		local
			l_interval: RP_BOOLEAN_HISTORY_INTERVAL
			l_value: BOOLEAN
		do
			if (not a_value.attribute_value.is_empty) and then a_value.attribute_value.item (1).as_lower = 'y' then
				l_value := True
			end
			create l_interval.make (l_value, a_value.start_date, a_value.end_date, l_value.to_integer)
			a_cache.extend (l_interval)
		end

	no_direct_composite_mapping_cache: like no_direct_composite_mapping
			-- Cached negated direct composite mappings

	retrieve_no_direct_composite_mapping: like no_direct_composite_mapping is
			-- Retrieve negated direct composite mappings
		local
			l_start, l_finish: ROSE_DATE
			l_interval: RP_BOOLEAN_HISTORY_INTERVAL
			l_temp: like no_direct_composite_mapping
		do
			if no_direct_composite_mapping_cache = Void then
				create no_direct_composite_mapping_cache.make_empty
				create l_temp.make_empty
				if gcdb_portfolio /= Void then
					gcdb_portfolio.composite_history.do_all (agent add_direct_mapping (l_temp, ?))
				end
					-- take the reverse
				from
					l_temp.coalesce
					create l_start.make_min_date
					l_temp.start
				until
					l_temp.off
				loop
					l_finish := l_temp.current_item.start_date
					create l_interval.make (True, l_start, l_finish, 1)
					no_direct_composite_mapping_cache.extend (l_interval)
					l_start := l_temp.current_item.end_date
					l_temp.forth
				end
				create l_finish.make_max_date
				if l_start.ccyymmdd /= l_finish.ccyymmdd then
					create l_interval.make (True, l_start, l_finish, 1)
					no_direct_composite_mapping_cache.extend (l_interval)
				end
			end
			Result := no_direct_composite_mapping_cache
		ensure
			non_void_result: Result /= Void
		end

	add_direct_mapping (a_cache: like no_direct_composite_mapping; a_interval: ENTITY_INTERVAL [GCDB_PRODUCT]) is
			-- Add direct composite mapping.
		local
			l_interval: RP_BOOLEAN_HISTORY_INTERVAL
		do
			create l_interval.make (True, a_interval.start_date, a_interval.end_date, 1)
			a_cache.extend (l_interval)
		end

	check_long_short_strategy (a_result: CELL [BOOLEAN]; a_interval: RP_BENCHMARK_HISTORY_INTERVAL) is
			-- Is this `a_interval' a long/short strategy?
		do
			if a_interval.benchmark.strategy /= Void then
				a_result.put (a_result.item or a_interval.benchmark.strategy.is_long_short)
			end
		end

	check_lls_strategy (a_result: CELL [BOOLEAN]; a_interval: RP_BENCHMARK_HISTORY_INTERVAL) is
			-- Is this `a_interval' a limited long/short strategy?
		do
			if a_interval.benchmark.strategy /= Void then
				a_result.put (a_result.item or a_interval.benchmark.strategy.is_limited_long_short)
			end
		end

	saved_reporting_currencies: DS_HASH_SET [RP_CURRENCY]
			-- Cache for reporting currencies

	retrieve_reporting_currencies is
			-- Retrieve reporting currencies
		do
			if saved_reporting_currencies = Void then
				create saved_reporting_currencies.make_equal (3)
				saved_reporting_currencies.force (zone.legacy_base_currency)
				account.reporting_currencies.do_all (agent saved_reporting_currencies.force (?))
			end
		ensure
			saved_reporting_currencies_not_void: saved_reporting_currencies /= Void
		end

feature {NONE} -- Access

	hash_code: INTEGER is
			-- <Precursor>
		do
			Result := code.hash_code
		end

invariant

	valid_code: code /= Void
	consistent_unknown_instance: is_unknown implies code.is_equal ({RP_UNKNOWN}.code)
	saved_instrument_control_not_void: saved_instrument_control /= Void
	no_void_saved_instrument_control: not saved_instrument_control.has_item (Void)

end
