note
	description: "Retrieves values from the NT Positions database"
	author: "Sami Ben Grine"
	last_editor: "$Author: sm $"
	date: "$Date: 2014/12/10 $"
	revision: "$Revision: #8 $"
	path: "$File: //rose/source/kernel/expert_system/rp_positions_data_source.e $"

class RP_POSITIONS_DATA_SOURCE

inherit

	ANY

	ROSE_SHARED_GLOBAL
		export {NONE} all end

feature -- Access

	portfolio_most_recent_global_investment_date (a_accounting_date: ROSE_DATE; a_code: STRING; a_update_mode: like accounting): ROSE_DATE
			-- Most recent date available for portfolio with 'a_code' on 'a_accounting_date' where the global_investment_date before or equal to 'a_as_of_date'
		require
			a_accounting_date_is_valid: a_accounting_date /= Void and then a_accounting_date >= northerntrust_positions_migration_date
			a_code_not_void: a_code /= Void
		do
			Result := portfolio_cash_most_recent_global_investment_date (a_accounting_date, a_code, a_update_mode).max (
			portfolio_non_cash_most_recent_global_investment_date (a_accounting_date, a_code, a_update_mode).max (
			portfolio_open_ffx_most_recent_global_investment_date (a_accounting_date, a_code, a_update_mode)))
			if Result.is_min_date then
				create Result.make_by_date (northerntrust_positions_migration_date)
			end
		ensure
			result_not_void: Result /= Void
		end

	portfolio_cash_as_of (a_as_of_date: ROSE_DATE; a_accounting_date: ROSE_DATE; a_code: STRING): ROSE_LINEAR_ARRAY [NT_PORTFOLIO_CASH]
			-- Portfolio cash positions
		require
			a_as_of_date_not_void: a_as_of_date /= Void
			a_as_of_date_is_valid: a_as_of_date >= northerntrust_positions_migration_date
			a_code_not_void: a_code /= Void
		do
			Result := portfolio_cash_as_of_with_mode (a_as_of_date, a_accounting_date, a_code, accounting, 'T')
		ensure
			result_not_void: Result /= Void
			no_void_item: not Result.has (Void)
			all_traded: Result.for_all (agent {NT_PORTFOLIO_CASH}.is_traded)
		end

	portfolio_settled_cash_as_of (a_as_of_date: ROSE_DATE; a_accounting_date: ROSE_DATE; a_code: STRING): ROSE_LINEAR_ARRAY [NT_PORTFOLIO_CASH]
			-- Portfolio cash positions
		require
			a_as_of_date_not_void: a_as_of_date /= Void
			a_as_of_date_is_valid: a_as_of_date >= northerntrust_positions_migration_date
			a_code_not_void: a_code /= Void
		do
			Result := portfolio_cash_as_of_with_mode (a_as_of_date, a_accounting_date, a_code, accounting, 'S')
		ensure
			result_not_void: Result /= Void
			no_void_item: not Result.has (Void)
			all_settled: Result.for_all (agent {NT_PORTFOLIO_CASH}.is_settled)
		end

	portfolio_settled_cash_entries_as_of (a_as_of_date: ROSE_DATE; a_code: STRING): ROSE_LINEAR_ARRAY [NT_PORTFOLIO_CASH]
			-- Portfolio cash positions
		require
			a_as_of_date_not_void: a_as_of_date /= Void
			a_as_of_date_is_valid: a_as_of_date >= northerntrust_positions_migration_date
			a_code_not_void: a_code /= Void
		do
			Result := portfolio_cash_entries_with_mode (a_as_of_date, a_code, approve, 'S')
		ensure
			result_not_void: Result /= Void
			no_void_item: not Result.has (Void)
			all_settled: Result.for_all (agent {NT_PORTFOLIO_CASH}.is_settled)
		end

	portfolio_settled_cash_for_currency_as_of (a_as_of_date: ROSE_DATE; a_accounting_date: ROSE_DATE; a_code: STRING; a_iso: STRING): NT_PORTFOLIO_CASH
			-- Portfolio setlled position in currency
		require
			a_as_of_date_not_void: a_as_of_date /= Void
			a_as_of_date_is_valid: a_as_of_date >= northerntrust_positions_migration_date
			a_code_not_void: a_code /= Void and then not a_code.is_empty
			a_iso_not_void: a_iso /= Void and then a_iso.count = 3
		do
			Result := portfolio_settled_cash_for_currency_as_of_with_mode (a_as_of_date, a_accounting_date, a_code, a_iso, accounting)
		ensure
			valid_result: (Result /= Void) implies (Result.global_investment_date ~ a_as_of_date and Result.accounting_date ~ a_accounting_date and Result.portfolio_code ~ a_code and Result.iso_currency_code ~ a_iso
			and Result.accounting_basis ~ "S")
		end

	portfolio_cash_as_of_with_mode (a_as_of_date: ROSE_DATE; a_accounting_date: ROSE_DATE; a_code: STRING; a_mode: RP_POSITIONS_UPDATE_MODE; a_type: CHARACTER): ROSE_LINEAR_ARRAY [NT_PORTFOLIO_CASH] is
  			-- Portfolio cash positions for `a_mode'
		require
			a_as_of_date_not_void: a_as_of_date /= Void
			a_as_of_date_is_valid: a_as_of_date >= northerntrust_positions_migration_date
			a_code_not_void: a_code /= Void
			a_mode_not_void: a_mode /= Void
			a_type_is_valid: a_type = 'T' or else a_type = 'S'
		do
			portfolio_cash_query.retrieve ( Void, <<a_as_of_date, a_accounting_date, a_code, a_mode.out, a_type>>)
	 		if not portfolio_cash_query.success then
				error.signal (generator, "failed to execute predicate $1", <<portfolio_cash_query.query_text>>)
	 		end
	 		create Result.make_from_array (portfolio_cash_query.array)
	 		Result.prune_all (Void)
		ensure
			result_not_void: Result /= Void
			no_void_item: not Result.has (Void)
   		end

	portfolio_cash_entries_with_mode (a_as_of_date: ROSE_DATE; a_code: STRING; a_mode: RP_POSITIONS_UPDATE_MODE; a_type: CHARACTER): ROSE_LINEAR_ARRAY [NT_PORTFOLIO_CASH] is
  			-- Portfolio cash positions for `a_mode' for all accounting dates
		require
			a_as_of_date_not_void: a_as_of_date /= Void
			a_as_of_date_is_valid: a_as_of_date >= northerntrust_positions_migration_date
			a_code_not_void: a_code /= Void
			a_mode_not_void: a_mode /= Void
			a_type_is_valid: a_type = 'T' or else a_type = 'S'
		do
			portfolio_cash_entries_at_gid_query.retrieve ( Void, <<a_as_of_date, a_code, a_mode.out, a_type>>)
	 		if not portfolio_cash_entries_at_gid_query.success then
				error.signal (generator, "failed to execute predicate $1", <<portfolio_cash_entries_at_gid_query.query_text>>)
	 		end
	 		create Result.make_from_array (portfolio_cash_entries_at_gid_query.array)
	 		Result.prune_all (Void)
		ensure
			result_not_void: Result /= Void
			no_void_item: not Result.has (Void)
   		end

	portfolio_settled_cash_for_currency_as_of_with_mode (a_as_of_date: ROSE_DATE; a_accounting_date: ROSE_DATE; a_code: STRING; a_iso: STRING; a_mode: RP_POSITIONS_UPDATE_MODE): NT_PORTFOLIO_CASH
  			-- Portfolio cash positions for `a_mode'
		require
			a_as_of_date_not_void: a_as_of_date /= Void
			a_as_of_date_is_valid: a_as_of_date >= northerntrust_positions_migration_date
			a_code_not_void: a_code /= Void
			a_iso_not_void: a_iso /= Void and then a_iso.count=3
			a_mode_not_void: a_mode /= Void
		do
			portfolio_cash_for_currency_query.retrieve ( Void, <<a_as_of_date, a_accounting_date, a_code, a_iso, a_mode.out, 'S'>>)
	 		if not portfolio_cash_for_currency_query.success then
				error.signal (generator, "failed to execute predicate $1", <<portfolio_cash_for_currency_query.query_text>>)
	 		end
	 		if portfolio_cash_for_currency_query.array /= Void and then portfolio_cash_for_currency_query.array.count = 1 then
		 		Result := portfolio_cash_for_currency_query.array.item(1)
	 		end
		ensure
   		end

	portfolio_non_cash_as_of (a_as_of_date: ROSE_DATE;  a_accounting_date: ROSE_DATE; a_code: STRING): ROSE_LINEAR_ARRAY [NT_PORTFOLIO_NON_CASH]
			-- Portfolio non-cash positions
		require
			a_as_of_date_not_void: a_as_of_date /= Void
			a_as_of_date_is_valid: a_as_of_date >= northerntrust_positions_migration_date
			a_code_not_void: a_code /= Void
		do
			Result := portfolio_non_cash_as_of_with_mode (a_as_of_date, a_accounting_date, a_code, accounting)
		ensure
			result_not_void: Result /= Void
			no_void_item: not Result.has (Void)
		end

	portfolio_non_cash_for_brimsym_as_of (a_as_of_date: ROSE_DATE;  a_accounting_date: ROSE_DATE; a_code: STRING; a_brimsym: STRING): NT_PORTFOLIO_NON_CASH
			-- Portfolio non-cash positions
		require
			a_as_of_date_not_void: a_as_of_date /= Void
			a_as_of_date_is_valid: a_as_of_date >= northerntrust_positions_migration_date
			a_code_not_void: a_code /= Void
			a_brimsym_not_void: a_brimsym /= Void and then a_brimsym.count = 18
		do
			Result := portfolio_non_cash_for_brimsym_as_of_with_mode (a_as_of_date, a_accounting_date, a_code, a_brimsym, accounting)
		ensure

		end

	portfolio_non_cash_as_of_with_mode (a_as_of_date: ROSE_DATE; a_accounting_date: ROSE_DATE; a_code: STRING; a_mode: RP_POSITIONS_UPDATE_MODE): ROSE_LINEAR_ARRAY [NT_PORTFOLIO_NON_CASH] is
  			-- Portfolio non-cash positions for `a_mode'
		require
			a_as_of_date_not_void: a_as_of_date /= Void
			a_as_of_date_is_valid: a_as_of_date >= northerntrust_positions_migration_date
			a_code_not_void: a_code /= Void
			a_mode_not_void: a_mode /= Void
		do
			portfolio_non_cash_query.retrieve ( Void, <<a_as_of_date, a_accounting_date, a_code, a_mode.out>>)
	 		if not portfolio_non_cash_query.success then
				error.signal (generator, "failed to execute predicate $1", <<portfolio_non_cash_query.query_text>>)
	 		end
	 		create Result.make_from_array (portfolio_non_cash_query.array)
	 		Result.prune_all (Void)
		ensure
			result_not_void: Result /= Void
			no_void_item: not Result.has (Void)
   		end

	portfolio_non_cash_for_brimsym_as_of_with_mode (a_as_of_date: ROSE_DATE; a_accounting_date: ROSE_DATE; a_code: STRING; a_brimsym: STRING; a_mode: RP_POSITIONS_UPDATE_MODE): NT_PORTFOLIO_NON_CASH
  			-- Portfolio non-cash positions for `a_mode'
		require
			a_as_of_date_not_void: a_as_of_date /= Void
			a_as_of_date_is_valid: a_as_of_date >= northerntrust_positions_migration_date
			a_code_not_void: a_code /= Void
			a_mode_not_void: a_mode /= Void
		do
			portfolio_non_cash_for_brimsym_query.retrieve ( Void, <<a_as_of_date, a_accounting_date, a_code, a_mode.out, a_brimsym>>)
	 		if not portfolio_non_cash_for_brimsym_query.success then
				error.signal (generator, "failed to execute predicate $1", <<portfolio_non_cash_for_brimsym_query.query_text>>)
	 		end
	 		if portfolio_non_cash_for_brimsym_query.array /= Void and then portfolio_non_cash_for_brimsym_query.array.count = 1 then
	 			Result := portfolio_non_cash_for_brimsym_query.array.item (1)
	 		end
		ensure
			valid_result: Result /= Void implies (Result.global_investment_date ~ a_as_of_date and Result.accounting_date ~ a_accounting_date and Result.portfolio_code ~ a_code and Result.brimsym ~ a_brimsym and Result.update_mode ~ a_mode.out)
   		end

   	portfolio_open_ffx_as_of (a_as_of_date: ROSE_DATE;  a_accounting_date: ROSE_DATE; a_code: STRING): ROSE_LINEAR_ARRAY [NT_PORTFOLIO_OPEN_FFX]
			-- Portfolio non-cash positions
		require
			a_as_of_date_not_void: a_as_of_date /= Void
			a_as_of_date_is_valid: a_as_of_date >= northerntrust_positions_migration_date
			a_code_not_void: a_code /= Void
		do
			Result := portfolio_open_ffx_as_of_with_mode (a_as_of_date, a_accounting_date, a_code, accounting)
		ensure
			result_not_void: Result /= Void
			no_void_item: not Result.has (Void)
		end

	portfolio_open_ffx_as_of_with_mode (a_as_of_date: ROSE_DATE; a_accounting_date: ROSE_DATE; a_code: STRING; a_mode: RP_POSITIONS_UPDATE_MODE): ROSE_LINEAR_ARRAY [NT_PORTFOLIO_OPEN_FFX] is
  			-- Portfolio non-cash positions for `a_mode'
		require
			a_as_of_date_not_void: a_as_of_date /= Void
			a_as_of_date_is_valid: a_as_of_date >= northerntrust_positions_migration_date
			a_code_not_void: a_code /= Void
			a_mode_not_void: a_mode /= Void
		do
			portfolio_open_ffx_query.retrieve ( Void, <<a_as_of_date, a_code, a_mode.out>>)
	 		if not portfolio_open_ffx_query.success then
				error.signal (generator, "failed to execute predicate $1", <<portfolio_open_ffx_query.query_text>>)
	 		end
	 		create Result.make_from_array (portfolio_open_ffx_query.array)
	 		Result.prune_all (Void)
		ensure
			result_not_void: Result /= Void
			no_void_item: not Result.has (Void)
   		end

	account_cash_most_recent_global_investment_date (a_accounting_date: ROSE_DATE; a_code: STRING): ROSE_DATE
			-- Most recent date available for portfolio with 'a_code' on 'a_accounting_date' where the global_investment_date before or equal to 'a_as_of_date'
		require
			a_accounting_date_is_valid: a_accounting_date /= Void and then a_accounting_date >= northerntrust_positions_migration_date
			a_code_not_void: a_code /= Void
		do
			account_cash_most_recent_global_investment_date_query.retrieve ( Void, <<a_accounting_date, a_code>>)
			if not account_cash_most_recent_global_investment_date_query.success then
				error.raise ("ERROR: account_cash_most_recent_global_investment_date_query (Retrieve) failed")
			end
			if not account_cash_most_recent_global_investment_date_query.array.is_empty then
				Result := account_cash_most_recent_global_investment_date_query.array.item (1).gid
			end
			if Result = Void then
				create Result.make_today
			end
		ensure
			result_not_void: Result /= Void
		end

	account_cash_as_of (a_as_of_date: ROSE_DATE;  a_accounting_date: ROSE_DATE; a_account: STRING): ROSE_LINEAR_ARRAY [NT_CASH]
			-- global_account cash positions
		require
			a_as_of_date_not_void: a_as_of_date /= Void
			a_as_of_date_is_valid: a_as_of_date >= northerntrust_positions_migration_date
			a_account_not_void: a_account /= Void
		do
			Result := account_cash_as_of_with_mode (a_as_of_date, a_accounting_date, a_account, accounting)
		ensure
			result_not_void: Result /= Void
			no_void_item: not Result.has (Void)
		end

	account_cash_as_of_with_mode (a_as_of_date: ROSE_DATE; a_accounting_date: ROSE_DATE; a_account: STRING; a_mode: RP_POSITIONS_UPDATE_MODE): ROSE_LINEAR_ARRAY [NT_CASH] is
  			-- global_account cash positions for `a_mode'
		require
			a_as_of_date_not_void: a_as_of_date /= Void
			a_as_of_date_is_valid: a_as_of_date >= northerntrust_positions_migration_date
			a_account_not_void: a_account /= Void
			a_mode_not_void: a_mode /= Void
		do
			global_account_cash_query.retrieve ( Void, <<a_as_of_date, a_accounting_date, a_account, a_mode.out>>)
	 		if not global_account_cash_query.success then
				error.signal (generator, "failed to execute predicate $1", <<global_account_cash_query.query_text>>)
	 		end
	 		create Result.make_from_array (global_account_cash_query.array)
	 		Result.prune_all (Void)
		ensure
			result_not_void: Result /= Void
			no_void_item: not Result.has (Void)
   		end

	account_non_cash_as_of (a_as_of_date: ROSE_DATE;  a_accounting_date: ROSE_DATE; a_account: STRING): ROSE_LINEAR_ARRAY [NT_NON_CASH]
			-- global_account non-cash positions
		require
			a_as_of_date_not_void: a_as_of_date /= Void
			a_as_of_date_is_valid: a_as_of_date >= northerntrust_positions_migration_date
			a_account_not_void: a_account /= Void
		do
			Result := account_non_cash_as_of_with_mode (a_as_of_date, a_accounting_date, a_account, accounting)
		ensure
			result_not_void: Result /= Void
			no_void_item: not Result.has (Void)
		end

	account_non_cash_as_of_with_mode (a_as_of_date: ROSE_DATE; a_accounting_date: ROSE_DATE; a_account: STRING; a_mode: RP_POSITIONS_UPDATE_MODE): ROSE_LINEAR_ARRAY [NT_NON_CASH] is
  			-- global_account non-cash positions for `a_mode'
		require
			a_as_of_date_not_void: a_as_of_date /= Void
			a_as_of_date_is_valid: a_as_of_date >= northerntrust_positions_migration_date
			a_account_not_void: a_account /= Void
			a_mode_not_void: a_mode /= Void
		do
			global_account_non_cash_query.retrieve ( Void, <<a_as_of_date, a_accounting_date, a_account, a_mode.out>>)
	 		if not global_account_non_cash_query.success then
				error.signal (generator, "failed to execute predicate $1", <<global_account_non_cash_query.query_text>>)
	 		end
	 		create Result.make_from_array (global_account_non_cash_query.array)
	 		Result.prune_all (Void)
		ensure
			result_not_void: Result /= Void
			no_void_item: not Result.has (Void)
   		end

	account_open_ffx_as_of (a_as_of_date: ROSE_DATE;  a_accounting_date: ROSE_DATE; a_account: STRING): ROSE_LINEAR_ARRAY [NT_OPEN_FFX]
			-- global_account open_ffx positions
		require
			a_as_of_date_not_void: a_as_of_date /= Void
			a_as_of_date_is_valid: a_as_of_date >= northerntrust_positions_migration_date
			a_account_not_void: a_account /= Void
		do
			Result := account_open_ffx_as_of_with_mode (a_as_of_date, a_accounting_date, a_account, accounting)
		ensure
			result_not_void: Result /= Void
			no_void_item: not Result.has (Void)
		end

	account_open_ffx_as_of_with_mode (a_as_of_date: ROSE_DATE; a_accounting_date: ROSE_DATE; a_account: STRING; a_mode: RP_POSITIONS_UPDATE_MODE): ROSE_LINEAR_ARRAY [NT_OPEN_FFX] is
  			-- global_account non-cash positions for `a_mode'
		require
			a_as_of_date_not_void: a_as_of_date /= Void
			a_as_of_date_is_valid: a_as_of_date >= northerntrust_positions_migration_date
			a_account_not_void: a_account /= Void
			a_mode_not_void: a_mode /= Void
		do
			global_account_open_ffx_query.retrieve ( Void, <<a_as_of_date, a_account, a_mode.out>>)
	 		if not global_account_open_ffx_query.success then
				error.signal (generator, "failed to execute predicate $1", <<global_account_open_ffx_query.query_text>>)
	 		end
	 		create Result.make_from_array (global_account_open_ffx_query.array)
	 		Result.prune_all (Void)
		ensure
			result_not_void: Result /= Void
			no_void_item: not Result.has (Void)
   		end

feature {NONE} -- Implementation

	servers: attached DATABASE_SERVERS
			-- Database servers
		once
			create Result
		ensure
			not_void: Result /= Void
		end

	portfolio_cash_most_recent_global_investment_date_query: attached DATE_PREDICATE is
			-- Predicate to retrieve most recent investment date for portfolio and accounting date
		local
			l_sql: STRING
		once
			l_sql := "SELECT TOP 1 global_investment_date FROM portfolio_cash WHERE update_mode = ? AND accounting_date = ? AND portfolio_code = ? ORDER BY global_investment_date DESC"
	 		create Result.make(servers.positions, l_sql)
	 		if not Result.success then
				(create {EXCEPTIONS}).raise ("Cannot create `portfolio_most_recent_global_investment_date_query'")
	 		end
	 	ensure
	 		not_void: Result /= Void
		end

	portfolio_non_cash_most_recent_global_investment_date_query: attached DATE_PREDICATE is
			-- Predicate to retrieve most recent investment date for portfolio and accounting date
		local
			l_sql: STRING
		once
			l_sql := "SELECT TOP 1 global_investment_date FROM portfolio_non_cash WHERE update_mode = ? AND accounting_date = ? AND portfolio_code = ? and accounting_basis='T'  AND ref_uid <> -1 AND brimsym IS NOT NULL ORDER BY global_investment_date DESC"
	 		create Result.make(servers.positions, l_sql)
	 		if not Result.success then
				(create {EXCEPTIONS}).raise ("Cannot create `portfolio_non_cash_most_recent_global_investment_date_query'")
	 		end
	 	ensure
	 		not_void: Result /= Void
		end

	portfolio_open_ffx_most_recent_global_investment_date_query: attached DATE_PREDICATE is
			-- Predicate to retrieve most recent investment date for portfolio and accounting date
		local
			l_sql: STRING
		once
			l_sql := "SELECT TOP 1 global_investment_date FROM portfolio_open_ffx WHERE update_mode = ? AND global_investment_date= ? AND portfolio_code =? ORDER BY global_investment_date DESC"
	 		create Result.make(servers.positions, l_sql)
	 		if not Result.success then
				(create {EXCEPTIONS}).raise ("Cannot create `portfolio_open_ffx_most_recent_global_investment_date_query'")
	 		end
	 	ensure
	 		not_void: Result /= Void
		end

	portfolio_cash_query: attached COLLECTING_PREDICATE[NT_PORTFOLIO_CASH] is
			-- Predicate to retrieve portfolio cash
		local
			l_sql: STRING
		once
			l_sql := "select * from portfolio_cash where global_investment_date = ? and accounting_date = ? and portfolio_code = ? and update_mode = ? and accounting_basis= ?"
	 		create Result.make(servers.positions, l_sql)
	 		if not Result.success then
				(create {EXCEPTIONS}).raise ("Cannot create `portfolio_cash_query'")
	 		end
	 	ensure
	 		not_void: Result /= Void
		end

	portfolio_cash_entries_at_gid_query: attached COLLECTING_PREDICATE[NT_PORTFOLIO_CASH] is
			-- Predicate to retrieve portfolio cash
		local
			l_sql: STRING
		once
			l_sql := "select * from portfolio_cash where global_investment_date = ? and portfolio_code = ? and update_mode = ? and accounting_basis= ?"
	 		create Result.make(servers.positions, l_sql)
	 		if not Result.success then
				(create {EXCEPTIONS}).raise ("Cannot create `portfolio_cash_query'")
	 		end
	 	ensure
	 		not_void: Result /= Void
		end

	 portfolio_cash_for_currency_query: attached COLLECTING_PREDICATE[NT_PORTFOLIO_CASH] is
			-- Predicate to retrieve portfolio cash in a specific currency
		local
			l_sql: STRING
		once
			l_sql := "select * from portfolio_cash where global_investment_date = ? and accounting_date = ? and portfolio_code = ? and iso_currency_code = ? and update_mode = ? and accounting_basis= ?"
	 		create Result.make(servers.positions, l_sql)
	 		if not Result.success then
				(create {EXCEPTIONS}).raise ("Cannot create `portfolio_cash_for_currency_query'")
	 		end
	 	ensure
	 		not_void: Result /= Void
		end

	portfolio_non_cash_query: attached COLLECTING_PREDICATE[NT_PORTFOLIO_NON_CASH] is
			-- Predicate to retrieve portfolio non cash
		local
			l_sql: STRING
		once
			l_sql := "SELECT * FROM portfolio_non_cash WHERE global_investment_date = ? AND accounting_date = ? AND portfolio_code = ? AND update_mode=? and accounting_basis='T' AND ref_uid <> -1 AND brimsym IS NOT NULL"
	 		create Result.make(servers.positions, l_sql)
	 		if not Result.success then
				(create {EXCEPTIONS}).raise ("Cannot create `portfolio_non_cash_query'")
	 		end
	 	ensure
	 		not_void: Result /= Void
		end

	portfolio_non_cash_for_brimsym_query: attached COLLECTING_PREDICATE[NT_PORTFOLIO_NON_CASH] is
			-- Predicate to retrieve portfolio non cash for brimsym
		local
			l_sql: STRING
		once
			l_sql := "SELECT * FROM portfolio_non_cash WHERE global_investment_date = ? AND accounting_date = ? AND portfolio_code = ? AND update_mode=? and brimsym = ? and accounting_basis='T'"
	 		create Result.make(servers.positions, l_sql)
	 		if not Result.success then
				(create {EXCEPTIONS}).raise ("Cannot create `portfolio_non_cash_for_brimsym_query'")
	 		end
	 	ensure
	 		not_void: Result /= Void
		end

	portfolio_open_ffx_query: attached COLLECTING_PREDICATE[NT_PORTFOLIO_OPEN_FFX] is
			-- Predicate to retrieve portfolio open ffx
		local
			l_sql: STRING
		once
			l_sql := "SELECT * FROM portfolio_open_ffx WHERE global_investment_date = ? AND portfolio_code = ? AND update_mode=?"
	 		create Result.make(servers.positions, l_sql)
	 		if not Result.success then
				(create {EXCEPTIONS}).raise ("Cannot create `portfolio_open_ffx_query'")
	 		end
	 	ensure
	 		not_void: Result /= Void
		end

	account_cash_most_recent_global_investment_date_query: attached COLLECTING_PREDICATE[TUPLE[gid: ROSE_DATE]] is
			-- Predicate to retrieve most recent investment date for portfolio and accounting date
		once
	 		create Result.make(servers.positions, "SELECT MAX(global_investment_date) AS gid FROM cash WHERE accounting_date = ? AND account_code = ?")
	 		if not Result.success then
				(create {EXCEPTIONS}).raise ("Cannot create `account_cash_most_recent_global_investment_date_query'")
	 		end
	 	ensure
	 		not_void: Result /= Void
		end

	global_account_cash_query: attached COLLECTING_PREDICATE[NT_CASH] is
			-- Predicate to retrieve global_account cash
		local
			l_sql: STRING
		once
			l_sql := "select * from cash where global_investment_date = ? and accounting_date = ? and account_code = ? and update_mode = ? and accounting_basis='T'"
	 		create Result.make(servers.positions, l_sql)
	 		if not Result.success then
				(create {EXCEPTIONS}).raise ("Cannot create `global_account_cash_query'")
	 		end
	 	ensure
	 		not_void: Result /= Void
		end

	global_account_non_cash_query: attached COLLECTING_PREDICATE[NT_NON_CASH] is
			-- Predicate to retrieve global_account cash
		local
			l_sql: STRING
		once
			l_sql := "select * from non_cash where global_investment_date = ? and accounting_date = ? and account_code = ? and update_mode = ? and accounting_basis='T' AND ref_uid <> -1 AND brimsym IS NOT NULL"
	 		create Result.make(servers.positions, l_sql)
	 		if not Result.success then
				(create {EXCEPTIONS}).raise ("Cannot create `global_account_non_cash_query'")
	 		end
	 	ensure
	 		not_void: Result /= Void
		end

	global_account_open_ffx_query: attached COLLECTING_PREDICATE[NT_OPEN_FFX] is
			-- Predicate to retrieve global_account open ffx
		local
			l_sql: STRING
		once
			l_sql := "select * from open_ffx where global_investment_date = ? and account_code = ? and update_mode = ?"
	 		create Result.make(servers.positions, l_sql)
	 		if not Result.success then
				(create {EXCEPTIONS}).raise ("Cannot create `global_account_open_ffx_query'")
	 		end
	 	ensure
	 		not_void: Result /= Void
		end

feature -- Constants

	NorthernTrust_positions_migration_date: attached ROSE_DATE
			-- Date when we switched to NT for positions
		local
			l_default_date: ROSE_DATE
		once
			create l_default_date.make (2013, 06, 17)
			Result := config.date ("NorthernTrust_positions_migration_date", l_default_date)
			inform ("{RP_POSITIONS_DATA_SOURCE}.NorthernTrust_positions_migration_date", "NorthernTrust_positions_migration_date $1", <<Result>>)
	 	ensure
	 		not_void: Result /= Void
		end

	accounting: attached RP_POSITIONS_UPDATE_MODE
			-- accounting positions update mode
		once
			create Result.make_accounting
	 	ensure
	 		not_void: Result /= Void
		end

	approve: attached RP_POSITIONS_UPDATE_MODE
			-- approved positions update mode
		once
			create Result.make_approve
	 	ensure
	 		not_void: Result /= Void
		end


	portfolio_cash_most_recent_global_investment_date (a_accounting_date: ROSE_DATE; a_code: STRING; a_update_mode: like accounting): ROSE_DATE
			-- Most recent date available for portfolio with 'a_code' on 'a_accounting_date' where the global_investment_date before or equal to 'a_as_of_date'
		require
			a_accounting_date_is_valid: a_accounting_date /= Void and then a_accounting_date >= northerntrust_positions_migration_date
			a_code_not_void: a_code /= Void
		do
			portfolio_cash_most_recent_global_investment_date_query.retrieve ( Void, <<a_update_mode.out, a_accounting_date, a_code>>)
			if not portfolio_cash_most_recent_global_investment_date_query.success then
				error.raise ("ERROR: portfolio_cash_most_recent_global_investment_date_query (Retrieve) failed")
			elseif portfolio_cash_most_recent_global_investment_date_query.array /= Void and then not portfolio_cash_most_recent_global_investment_date_query.array.is_empty then
				Result := portfolio_cash_most_recent_global_investment_date_query.array.item (1)
			else
				create Result.make_min_date
			end
		ensure
			result_not_void: Result /= Void
		end

	portfolio_non_cash_most_recent_global_investment_date (a_accounting_date: ROSE_DATE; a_code: STRING; a_update_mode: like accounting): ROSE_DATE
			-- Most recent date available for portfolio with 'a_code' on 'a_accounting_date' where the global_investment_date before or equal to 'a_as_of_date'
		require
			a_accounting_date_is_valid: a_accounting_date /= Void and then a_accounting_date >= northerntrust_positions_migration_date
			a_code_not_void: a_code /= Void
		do
			portfolio_non_cash_most_recent_global_investment_date_query.retrieve ( Void, <<a_update_mode.out, a_accounting_date, a_code>>)
			if not portfolio_non_cash_most_recent_global_investment_date_query.success then
				error.raise ("ERROR: portfolio_non_cash_most_recent_global_investment_date_query (Retrieve) failed")
			elseif portfolio_non_cash_most_recent_global_investment_date_query.array /= Void and then not portfolio_non_cash_most_recent_global_investment_date_query.array.is_empty then
				Result := portfolio_non_cash_most_recent_global_investment_date_query.array.item (1)
			else
				create Result.make_min_date
			end
		ensure
			result_not_void: Result /= Void
		end

	portfolio_open_ffx_most_recent_global_investment_date (a_accounting_date: ROSE_DATE; a_code: STRING; a_update_mode: like accounting): ROSE_DATE
			-- Most recent date available for portfolio with 'a_code' on 'a_accounting_date' where the global_investment_date before or equal to 'a_as_of_date'
		require
			a_accounting_date_is_valid: a_accounting_date /= Void and then a_accounting_date >= northerntrust_positions_migration_date
			a_code_not_void: a_code /= Void
		do
			portfolio_open_ffx_most_recent_global_investment_date_query.retrieve ( Void, <<a_update_mode.out, a_accounting_date, a_code>>)
			if not portfolio_open_ffx_most_recent_global_investment_date_query.success then
				error.raise ("ERROR: portfolio_open_ffx_most_recent_global_investment_date_query (Retrieve) failed")
			elseif portfolio_open_ffx_most_recent_global_investment_date_query.array /= Void and then not portfolio_open_ffx_most_recent_global_investment_date_query.array.is_empty then
				Result := portfolio_open_ffx_most_recent_global_investment_date_query.array.item (1)
			else
				create Result.make_min_date
			end
		ensure
			result_not_void: Result /= Void
		end

end
