indexing

	description:

		"A Product (Composite)"

	date: "$Date: 2010/04/07 $"
	sccid: "$Id: //rose/source/kernel/expert_system/rp_product.e#17 $"

deferred class RP_PRODUCT

inherit
	RP_ZONAL_INVESTMENT_GROUP
		rename
			start_date as begin_date, end_date as close_date, name as last_known_name
		end

feature {NONE} -- Initialization

	make (a_ref: like reference_system; a_gcdb_product: like gcdb_product) is
			-- Initialize `Current'.
		do
			reference_system := a_ref
			gcdb_product := a_gcdb_product
			if gcdb_product /= Void then
				code := gcdb_product.code
				legacy_code := gcdb_product.investment_region_unique_code
				if legacy_code = Void then
					legacy_code := code
				end
			else
				is_unknown := True
				code := "Unknown"
				legacy_code := code
			end
		end

feature -- Access

	reference_system: RP_REFERENCE_SYSTEM
			-- Parent of `Current'.

	code: STRING
			-- Unique code of the composite.

	legacy_code: STRING
			-- VMS code of the composite.

	begin_date: ROSE_DATE is
			-- Begin date.
		do
			Result := product_begin_date
		end

	close_date: ROSE_DATE is
			-- Close date.
		do
			Result := product_close_date
		end

	audit_currency: RP_CURRENCY is
			-- Currency of audited exhibits.
		do
			Result := retrieve_audit_currency
		end

	last_known_name: STRING is
			-- Last known descriptive name
		do
			if gcdb_product /= Void then
				Result := gcdb_product.last_known_name
			else
				create Result.make_empty
			end
		end

	base_fee: STRING is
			-- Highest fee in product.
		do
			if gcdb_product /= Void then
				Result := gcdb_product.base_fee
			else
				create Result.make_empty
			end
		ensure
			non_void_result: Result /= Void
		end

	perf_fee_net_alpha: STRING is
			-- Percentage of net alpha. Used in performance calculation of composites.
		do
			if gcdb_product /= Void then
				Result := gcdb_product.perf_fee_net_alpha
			else
				create Result.make_empty
			end
		ensure
			non_void_result: Result /= Void
		end

feature -- Change

	set_code (a_code: like code) is
			-- <Precursor>
		do
			code := a_code
		end

feature -- Status Report

	is_composite: BOOLEAN is
			-- Is `Current' composite?
		do
			if gcdb_product /= Void then
				Result := not (gcdb_product.is_not_composite or gcdb_product.is_not_product)
			end
		end

	is_examined: BOOLEAN is
			-- Was `Current' examined as part of any GIPS audit?
		do
			Result := gcdb_product /= Void and then gcdb_product.gips_examination /= Void and then not gcdb_product.gips_examination.is_empty
		end

	is_global: BOOLEAN is
			-- Does `Current' consist of investment accounts?
			-- True = members are global accounts
			-- False = members are portfolios
		deferred
		end

	is_marketed: BOOLEAN is
			-- Indicated whether a product is an official product. Only official products show in AUM.
		do
			if gcdb_product /= Void then
				Result := gcdb_product.is_marketed
			end
		end

	is_unknown: BOOLEAN
			-- Is `Current' the "unknown" instance.

feature -- Query

	performance_code: STRING is
			-- <Precursor>
		do
			Result := legacy_code
		end

	investment_region: RP_ZONE is
			-- Investment region.
		do
			Result := retrieve_investment_region
		end

	expected_alpha: DOUBLE is
			-- expected alpha
		do
			if gcdb_product.expected_alpha /= Void then
				Result := gcdb_product.expected_alpha.to_double
			end
		end

	expected_tracking_error: DOUBLE is
			-- expected tracking_error
		do
			if gcdb_product.expected_tracking_error /= Void then
				Result := gcdb_product.expected_tracking_error.to_double
			end
		end

	benchmarks: RP_BENCHMARK_HISTORY is
			-- Interval history of benchmarks associated with `Current'
		do
			Result := retrieve_benchmarks
		ensure
			non_void_benchmarks: Result /= Void
		end

feature -- Comparison

	hash_code: INTEGER is
			-- <Precursor>
		do
			Result := (investment_region.code+"_"+code).hash_code
		end

feature -- Output

	debug_output: STRING is
			-- <Precursor>
		do
			Result := code
		end

feature {NONE} -- Implementation

	gcdb_product: GCDB_PRODUCT
			-- Underlying GCDB entity.

	saved_benchmarks: like benchmarks
			-- Placeholder for `benchmarks'.

	retrieve_benchmarks: like benchmarks is
			-- Evaluate `benchmarks'.
		do
			if saved_benchmarks = Void then
				create saved_benchmarks.make_empty
				load_benchmarks (saved_benchmarks)
			end
			Result := saved_benchmarks
		ensure
			non_void_benchmarks: Result /= Void
		end

	load_benchmarks (a_history: like benchmarks) is
			-- Enumerate `benchmarks'.
		do
			if gcdb_product /= Void then
				gcdb_product.benchmarks_by_product_history.do_all (agent add_benchmark (a_history, ?))
			end
		end

	add_benchmark (a_history: like benchmarks; a_interval: ENTITY_INTERVAL [GCDB_BENCHMARK]) is
			-- Add benchmarks to `benchmarks'.
		local
			l_code: STRING
			l_benchmark: RP_BENCHMARK
			l_interval: RP_BENCHMARK_HISTORY_INTERVAL
		do
			l_code := a_interval.entity.code
			if l_code = Void then
				a_interval.entity.set_as_of_date (a_interval.entity.start_date)
				l_code := a_interval.entity.code
			end
			if l_code /= Void then
				l_benchmark := reference_system.benchmark_by_code (l_code)
				create l_interval.make (l_benchmark, a_interval.start_date, a_interval.end_date, a_interval.weight)
				a_history.extend (l_interval)
			end
		end

	retrieve_investment_region: RP_ZONE is
			-- Evaluate `investment_region'.
		local
			l_code: STRING
		do
			if gcdb_product /= Void then
				if gcdb_product.investment_region /= Void then
					l_code := gcdb_product.investment_region.code
				end
			end
			if l_code = Void then
				Result := reference_system.unknown_investment_region
			else
				Result := reference_system.investment_region_by_code (l_code)
			end
		end

	membership_begin_date: like begin_date is
			-- Date of first membership in `Current'.
		deferred
		end

	begin_date_saved: like begin_date
			-- Placeholder for `begin_date'.

	begin_date_initialized: BOOLEAN
			-- Is `begin_date' initialized?

	product_begin_date: like begin_date is
			-- Compute `begin_date'.
		do
			if not begin_date_initialized then
				if gcdb_product /= Void then
					if gcdb_product.begin_date /= Void then
						begin_date_saved := gcdb_product.begin_date
					else
						begin_date_saved := membership_begin_date
					end
				end
				if begin_date_saved = Void then
					create begin_date_saved.make_ccyymmdd (99991231)
				end
				begin_date_initialized := True
			end
			Result := begin_date_saved
		end

	membership_close_date: like close_date is
			-- Last membership date.
		deferred
		end

	close_date_saved: like close_date
			-- Placeholder for `close_date'.

	close_date_initialized: BOOLEAN
			-- Is `close_date' initialized?

	product_close_date: like close_date is
			-- Close date.
		do
			if not close_date_initialized then
				if gcdb_product /= Void then
					if gcdb_product.close_date /= Void then
						close_date_saved := gcdb_product.close_date
					else
						close_date_saved := membership_close_date
					end
				end
				if close_date_saved = Void then
					create close_date_saved.make_ccyymmdd (99991231)
				end
				close_date_initialized := True
			end
			Result := close_date_saved
		end

	audit_currency_saved: like audit_currency
			-- Placeholder for `audit_currency'.

	retrieve_audit_currency: like audit_currency is
			-- Evaluate `audit_currency'.
		do
			if audit_currency_saved = Void then
				if gcdb_product /= Void then
					if gcdb_product.audit_currency /= Void then
						audit_currency_saved := reference_system.currency_from_iso_code (gcdb_product.audit_currency.code)
					else
						audit_currency_saved := investment_region.legacy_base_currency
					end
				else
					audit_currency_saved := reference_system.currency_from_iso_code ("USD")
				end
			end
			Result := audit_currency_saved
		end

invariant

	code_not_void: code /= Void

end
