indexing

	description:

		"Doubly-linked list of DailyReturns for product (composite)"

	date: "$Date: 2010/06/29 $"
	sccid: "$Id: //rose/source/Kernel/expert_system/legacy/rp_product_daily_returns.e#10 $"

deferred class RP_PRODUCT_DAILY_RETURNS [G -> RP_PRODUCT, H -> HASHABLE]

feature {NONE} -- Initialization

	make (a_product: like product; a_currency: RP_CURRENCY; a_returns_agent: like returns_agent; a_weights_agent: like weights_agent; a_bmv_feedback: like bmv_feedback; a_monthlybmv_feedback: like monthlybmv_feedback) is
			-- Initialize `Current'
		require
			non_void_args: a_product /= Void and a_currency /= Void and a_returns_agent /= Void and a_weights_agent /= Void
		do
			product := a_product
			currency := a_currency
			returns_agent := a_returns_agent
			weights_agent := a_weights_agent
			bmv_feedback := a_bmv_feedback
			monthlybmv_feedback := a_monthlybmv_feedback
			retrieve_components
			retrieve_returns
			compute
		ensure
			returns_computed: daily_returns /= Void
		end

feature -- Query

	daily_returns: RP_DAILY_RETURNS
			-- DailyReturns

feature {NONE} -- Implementation

	product: G
			-- Product

	currency: RP_CURRENCY
			-- Currency

	returns_agent: FUNCTION [ANY, TUPLE [H], RP_DAILY_RETURNS]
			-- Returns agent

	weights_agent: FUNCTION [ANY, TUPLE [ROSE_DATE_INTERVAL], DS_HASH_TABLE [TUPLE [value: RP_DECIMAL; date: ROSE_DATE], H]]
			-- Weighs agent

	bmv_feedback: PROCEDURE [ANY, TUPLE [STRING]]
			-- BMV Feedback agent

	monthlybmv_feedback: PROCEDURE [ANY, TUPLE [STRING]]
			-- Monthly BMV Feedback agent

	components: DS_HASH_SET [H]
			-- Distinct portfolios in composite

	returns: DS_HASH_TABLE [RP_DAILY_RETURNS, H]
			-- Returns per component

	retrieve_components is
			-- Retrieve list of distinct components
		deferred
		ensure
			components_not_void: components /= Void
		end

	membership_start_date: ROSE_DATE is
			-- First membership date
		deferred
		end

	membership_end_date: ROSE_DATE is
			-- Last membership date
		deferred
		end

	retrieve_returns is
			-- Retrieve returns
		do
			create returns.make_equal (17)
			components.do_all (agent add_returns (?))
		end

	add_returns (a_component: H) is
			-- Add dailyreturns for `a_component' to `returns'
		local
			l_returns: RP_DAILY_RETURNS
		do
			l_returns := returns_agent.item ([a_component])
			if l_returns /= Void then
				returns.force (l_returns, a_component)
			end
		end

	compute is
			-- Compute `daily_returns'
		local
			l_date: ROSE_DATE
			l_start: ROSE_DATE
			l_finish: ROSE_DATE
			l_calendar: RP_CALENDAR
			l_dr: RP_DAILY_RETURN
			l_return_and_value: like return_and_value
			l_month: ROSE_MONTH
			l_missing_data: BOOLEAN
		do
			create daily_returns.make (product.code)
			l_calendar := product.investment_region.legacy_perf_calendar
			from
				l_start := membership_start_date
				if product.begin_date /= Void then
					l_start := product.begin_date.max (l_start)
				end
				l_start := l_calendar.ith_work_day_before (l_start, 1)
					-- add first.
				create l_dr.make (l_start, currency, 0, 1)
				daily_returns.expand (l_dr)
				l_start := l_calendar.ith_work_day_after (l_start, 1)
				l_date := l_start
				create l_month.make_ccyymm (l_date.ccyymm)
				create l_finish.make_today
				l_finish := l_finish.min (membership_end_date)
				l_finish := l_calendar.last_work_day (l_finish)
				if product.close_date /= Void then
					l_finish := product.close_date.min (l_finish)
				end
			until
				l_date > l_finish or l_missing_data
			loop
				l_return_and_value := return_and_value (l_calendar, l_month, l_date, currency)
				if l_return_and_value /= Void then
					create l_dr.make (l_date, currency, l_return_and_value.second, l_return_and_value.first)
					daily_returns.expand (l_dr)
					l_date := l_calendar.ith_work_day_after (l_date, 1)
					if l_date.ccyymm /= l_month.ccyymm then
						create l_month.make_ccyymm (l_date.ccyymm)
					end
				else
					l_missing_data := True
				end
			end
		ensure
			daily_returns_not_void: daily_returns /= Void
		end

    return_and_value (a_calendar: RP_CALENDAR; a_month: ROSE_MONTH; a_date: ROSE_DATE; a_currency: RP_CURRENCY): PAIR [RP_DECIMAL, RP_DECIMAL] is
            -- Daily return and value for `a_date'
		require
			date_is_market_date: a_calendar.is_work_day (a_date)
		local
			l_weights: like begin_of_day_weights
			l_returns: DS_HASH_TABLE [RP_DECIMAL, H]
			l_total: CELL [RP_DECIMAL]
			l_value: RP_DECIMAL
		do
			l_weights := begin_of_day_weights (a_month, a_date, a_calendar, a_currency)
            if l_weights /= Void then
                l_returns := component_returns (l_weights, a_date)
                if l_returns /= Void then
                	l_weights.keys.do_all (agent feedback_weight_and_return (a_date, l_weights, l_returns, ?))
	                l_returns.keys.do_all (agent multiply (l_returns, l_weights, ?))
	                create l_total.put (0)
	                l_returns.do_all (agent sum (l_total, ?))
	                l_value := value (l_weights, a_date)
	                if l_value /= Void then
		                create Result.make (l_total.item, l_value)
	                end
                end
            end
		end

	feedback_weight_and_return (a_date: ROSE_DATE; a_weights: like begin_of_day_weights; a_returns: DS_HASH_TABLE [RP_DECIMAL, H]; a_component: H) is
			-- Provide feedback about daily weights and returns
		local
			l_return: RP_DAILY_RETURN
		do
			l_return := daily_return_record (a_date, a_component)
			if bmv_feedback /= Void and l_return /= Void then
				bmv_feedback.call ([product.code + "," + a_date.ccyymmdd.out + "," + component_name (a_component) + "," + l_return.value.rounded (2).out + "," + a_weights.item (a_component).rounded (6).out + "," + a_returns.item (a_component).rounded (6).out + "%N"])
			end
		end

	value (a_weights: like begin_of_day_weights; a_date: ROSE_DATE): RP_DECIMAL is
			-- Value of `product' on `a_date'
		local
			l_missing_value: CELL [BOOLEAN]
			l_total: CELL [RP_DECIMAL]
		do
			create l_missing_value.put (False)
			create l_total.put (0)
			a_weights.keys.do_all (agent add_value (l_missing_value, l_total, a_date, ?))
			if not l_missing_value.item then
				Result := l_total.item
			end
		end

	add_value (a_missing_data: CELL [BOOLEAN]; a_total: CELL [RP_DECIMAL]; a_date: ROSE_DATE; a_component: H) is
			-- Add value for `a_component'
		local
			l_return: RP_DAILY_RETURN
		do
			l_return := daily_return_record (a_date, a_component)
			if l_return /= Void then
				a_total.put (a_total.item + l_return.value)
			else
				a_missing_data.put (True)
			end
		end

	daily_return_record (a_date: ROSE_DATE; a_component: H): RP_DAILY_RETURN is
			-- DailyReturn for `a_component' on `a_date'
		local
			l_returns: RP_DAILY_RETURNS
		do
			returns.search (a_component)
			if returns.found then
				l_returns := returns.found_item
				Result := l_returns.return (a_date)
			end
		end

	begin_of_day_weights (a_month: ROSE_MONTH; a_date: ROSE_DATE; a_calendar: RP_CALENDAR; a_currency: RP_CURRENCY): DS_HASH_TABLE [RP_DECIMAL, H] is
			-- Begin of day weights on `a_date'
		do
			if is_first_market_date (a_date, a_calendar) then
				Result := component_weigths (a_month, a_calendar, a_currency)
			else
				Result := end_of_day_weights (a_month, a_calendar.ith_work_day_before (a_date, 1), a_calendar, a_currency)
			end
		end

	end_of_day_weights (a_month: ROSE_MONTH; a_date: ROSE_DATE; a_calendar: RP_CALENDAR; a_currency: RP_CURRENCY): DS_HASH_TABLE [RP_DECIMAL, H] is
			-- End of day weights on `a_date'
		local
			l_returns: DS_HASH_TABLE [RP_DECIMAL, H]
			l_total: CELL [RP_DECIMAL]
		do
			Result := begin_of_day_weights (a_month, a_date, a_calendar, a_currency)
            if Result /= Void then
                l_returns := component_returns (Result, a_date)
                if l_returns /= Void then
	                Result.keys.do_all (agent multiply (Result, l_returns, ?))
	                create l_total.put (0)
	                Result.do_all (agent sum (l_total, ?))
	                Result.keys.do_all (agent divide (Result, l_total.item, ?))
	          	else
	          		Result := Void
                end
            end
		end

	is_first_market_date (a_date: ROSE_DATE; a_calendar: RP_CALENDAR): BOOLEAN is
			-- Is `a_date' the first market date in calendar month?
		do
			Result := a_date.ccyymm /= a_calendar.ith_work_day_before (a_date, 1).ccyymm
		end

	divide (a: DS_HASH_TABLE [RP_DECIMAL, H]; b: RP_DECIMAL; a_component: H) is
			-- Divide `a' by `b' for `a_component'
		do
			a.force (a.item (a_component) / b, a_component)
		end

	multiply (a, b: DS_HASH_TABLE [RP_DECIMAL, H]; a_component: H) is
			-- Multiply `a' by `b' for `a_component'
		do
			a.force (a.item (a_component) * b.item (a_component), a_component)
		end

	sum (a_total: CELL [RP_DECIMAL]; a_value: RP_DECIMAL) is
			-- Add `a_value' to `a_total'
		do
			a_total.put (a_total.item + a_value)
		end

	component_returns (a_weights: DS_HASH_TABLE [RP_DECIMAL, H]; a_date: ROSE_DATE): DS_HASH_TABLE [RP_DECIMAL, H] is
			-- Component returns on `a_date'
		local
			l_missing_data: CELL [BOOLEAN]
		do
			create l_missing_data.put (False)
			create Result.make_equal (17)
			a_weights.keys.do_all (agent component_return (l_missing_data, Result, a_date, ?))
			if l_missing_data.item then
				Result := Void
			end
		end

	component_return (a_missing_data: CELL [BOOLEAN]; a_result: DS_HASH_TABLE [RP_DECIMAL, H]; a_date: ROSE_DATE; a_component: H) is
			-- Return for `a_component' on `a_date'
		local
			l_return: RP_DAILY_RETURN
		do
			l_return := daily_return_record (a_date, a_component)
			if l_return /= Void then
				a_result.force (l_return.return, a_component)
			else
				a_missing_data.put (True)
			end
		end

	add_weight (a_hash: like component_weigths; a_tuple: TUPLE [value: RP_DECIMAL; date: ROSE_DATE]; a_key: H) is
			-- Add value from `a_tuple' for `a_key' to `a_hash'
		do
			a_hash.force (a_tuple.value, a_key)
		end

	component_weigths (a_month: ROSE_MONTH; a_calendar: RP_CALENDAR; a_currency: RP_CURRENCY): DS_HASH_TABLE [RP_DECIMAL, H] is
			-- Component weights for `a_month'
		local
			l_total, l_emv, l_ret: CELL [RP_DECIMAL]
			l_weights_and_dates: DS_HASH_TABLE [TUPLE [value: RP_DECIMAL; date: ROSE_DATE], H]
			l_monthend: ROSE_DATE
		do
			l_weights_and_dates := weights_agent.item ([a_month])
			create Result.make_equal (l_weights_and_dates.count)
			l_weights_and_dates.do_all_with_key (agent add_weight (Result, ?, ?))
			if bmvs = Void then
				create bmvs.make_default
			end
			create l_total.put (0)
			Result.do_all (agent sum (l_total, ?))
			if not bmvs.has (a_month.ccyymm) then
				bmvs.force (a_month.ccyymm)
				l_monthend := a_calendar.last_work_day (a_month.month_end)
				if bmv_feedback /= Void then
					bmv_feedback.call ([product.code + "," + a_month.ccyymm.out + ",total," + l_total.item.rounded (2).out + "%N"])
				end
				create l_emv.put (0)
				create l_ret.put (0)
				Result.keys.do_all (agent feedback_bmv (l_monthend, l_weights_and_dates, ?, l_total.item, a_currency, l_emv, l_ret))
				if monthlybmv_feedback /= Void then
					monthlybmv_feedback.call ([
						product.code
						+ "," +
						l_monthend.month_end.formatted_out ("[0]mm/[0]dd/yyyy")
						+ ",," +
						l_total.item.rounded (2).out
						+ "," +
						l_emv.item.rounded (2).out
						+ ",,," +
						l_ret.item.rounded (2).out
						+ "," +
						a_currency.name +
						"%N,%N"
					])
				end
			end
			Result.keys.do_all (agent divide (Result, l_total.item, ?))
		end

	bmvs: DS_HASH_SET [INTEGER]
			-- BMVs already printed

	feedback_bmv (a_finish: ROSE_DATE; a_weights: DS_HASH_TABLE [TUPLE [value: RP_DECIMAL; date: ROSE_DATE], H]; a_component: H; a_total: RP_DECIMAL; a_currency: RP_CURRENCY; a_emv, a_ret: CELL [RP_DECIMAL]) is
			-- Fedback for BMV
		local
			l_weight: RP_DECIMAL
			l_start: ROSE_DATE
		do
			l_weight := a_weights.item (a_component).value
			l_start := a_weights.item (a_component).date
			if bmv_feedback /= Void then
				bmv_feedback.call ([product.code + "," + a_finish.ccyymm.out + "," + component_name (a_component) + "," + l_weight.rounded (2).out + "%N"])
			end
			feedback_monthlybmv (l_start, a_finish, l_weight, a_component, a_total, a_currency, a_emv, a_ret)
		end

	feedback_monthlybmv (a_start, a_finish: ROSE_DATE; a_weight: RP_DECIMAL; a_component: H; a_total: RP_DECIMAL; a_currency: RP_CURRENCY; a_emv, a_ret: CELL [RP_DECIMAL]) is
			-- Fedback for Monthly BMV
		require
			non_void_args: a_start /= Void and a_finish /= Void and a_weight /= Void and a_component /= Void and a_total /= Void and a_currency /= Void and a_emv /= Void and a_ret /= Void
		local
			l_return1, l_return2: RP_DAILY_RETURN
			l_value, l_weight, l_return: RP_DECIMAL
		do
			if monthlybmv_feedback /= Void then
				l_return1 := daily_return_record (a_finish, a_component)
				if l_return1 = Void and a_finish.ccyymm /= today.ccyymm then
						-- raise exception for missing data for complete months
						-- but do not raise exception for m-t-d
					(create {EXCEPTIONS}).raise ("No DailyReturns for " + component_name (a_component) + " @ " + a_finish.ccyymmdd.out)
				end
				l_return2 := daily_return_record (a_start, a_component)
				if l_return2 = Void then
					(create {EXCEPTIONS}).raise ("No DailyReturns for " + component_name (a_component) + " @ " + a_start.ccyymmdd.out)
				end
				if l_return1 /= Void and l_return2 /= Void then
					l_value := l_return1.value
					a_emv.put (a_emv.item + l_value)
					l_weight := a_weight / a_total
					l_return := (l_return1.cumrel / l_return2.cumrel * 100 - 100) * l_weight
					a_ret.put (a_ret.item + l_return)
					monthlybmv_feedback.call ([
						product.code + "," +
						a_finish.month_end.formatted_out ("[0]mm/[0]dd/yyyy") + "," +
						component_name (a_component) + "," +
						a_weight.rounded (2).out + "," +
						l_value.rounded (2).out + "," +
						(l_return1.cumrel / l_return2.cumrel * 100 - 100).rounded (2).out + "," +
						l_weight.rounded (6).out + "," +
						l_return.rounded (6).out + "," +
						a_currency.name +
						"%N"])
				end
			end
		end

	component_name (a_component: H): STRING is
			-- Name of `a_component'
		deferred
		ensure
			non_void_name: Result /= Void
		end

	today: ROSE_DATE is
			-- Today's date
		once
			create Result.make_today
		end

end
