note

	description:

		"Serializable metadata information"

	author: "Eric Bezault"
	last_editor: "$Author: er $"
	date: "$Date: 2014/02/11 $"
	revision: "$Revision: #2 $"
	path: "$File: //rose/source/kernel/infrastructure/persistence/storable/rp_serializable_metadata.e $"

class RP_SERIALIZABLE_METADATA

inherit

	ROSE_SHARED_GLOBAL

create

	make_current, default_create

feature {NONE} -- Initialization

	make_current
			-- Create metadata filled with information from the
			-- currently running application.
		do
			application := global.app.program_path
			user := global.app.user.login_id
			host_name := global.app.host_name
			estudio := global.app.info.build_estudio
			changelist := global.app.info.revision
			date := (create {ROSE_DATE}.make_now_utc).formatted_out ("yyyy/[0]mm/[0]dd [0]hh:[0]mi:[0]ss")
		end

feature -- Access

	application: STRING
			-- Pathname of the application which stored the object

	user: STRING
			-- Login id of the user who stored the object

	host_name: STRING
			-- Host where the application was running

 	estudio: STRING
 			-- Version of Estudio used to compile the application

 	changelist: INTEGER
 			-- Changelist of the code used to compile the application

	date: STRING
			-- Date (UTC) when the application stored the object

	object_type: STRING
			-- Type of root object in the storable file

feature -- Setting

	set_object_type (a_type: like object_type)
			-- Set `object_type' to `a_type'.
		do
			object_type := a_type
		ensure
			object_type_set: object_type = a_type
		end

feature -- Iteration

	do_all_with_key (a_action: PROCEDURE [ANY, TUPLE [key, value: STRING]])
			-- Apply `a_action' to all values of current metadata.
			-- `a_action' receives the key and its associated value.
		require
			a_action_not_void: a_action /= Void
		do
			if application /= Void then
				a_action.call ([application_field_name, application])
			end
			if user /= Void then
				a_action.call ([user_field_name, user])
			end
			if host_name /= Void then
				a_action.call ([host_name_field_name, host_name])
			end
			if changelist /= 0 then
				a_action.call ([changelist_field_name, changelist.out])
			end
			if estudio /= Void then
				a_action.call ([estudio_field_name, estudio])
			end
			if date /= Void then
				a_action.call ([date_field_name, date])
			end
			if object_type /= Void then
				a_action.call ([object_type_field_name, object_type])
			end
		end

feature -- Output

	to_text: STRING
			-- Textual representation
		do
			create Result.make (500)
			do_all_with_key (agent append_key_value_to_text (Result, ?, ?))
		ensure
			to_text_not_void: Result /= Void
		end

	from_text (a_text: STRING)
			-- Set fields from `a_text'.
		require
			a_text_not_void: a_text /= Void
		local
			l_splitter: ST_SPLITTER
			l_cursor: DS_LIST_CURSOR [STRING]
			l_field: STRING
			l_string: STRING
			i: INTEGER
			l_key, l_value: STRING
		do
			application := Void
			user := Void
			host_name := Void
			estudio := Void
			changelist := 0
			date := Void
			object_type := Void
			create l_splitter.make_with_separators ("%N%R")
			l_cursor := l_splitter.split (a_text).new_cursor
			from l_cursor.start until l_cursor.after loop
				l_field := l_cursor.item
				i := l_field.index_of (':', 1)
				if i > 1 then
					l_key := l_field.substring (1, i - 1)
					l_key.left_adjust
					l_key.right_adjust
					l_value := l_field.substring (i + 1, l_field.count)
					l_value.left_adjust
					l_value.right_adjust
					if l_key.same_string (application_field_name) then
						application := l_value
					elseif l_key.same_string (user_field_name) then
						user := l_value
					elseif l_key.same_string (host_name_field_name) then
						host_name := l_value
					elseif l_key.same_string (estudio_field_name) then
						estudio := l_value
					elseif l_key.same_string (changelist_field_name) then
						l_string := l_value
						if l_string.is_integer then
							changelist := l_string.to_integer
						end
					elseif l_key.same_string (date_field_name) then
						date := l_value
					elseif l_key.same_string (object_type_field_name) then
						object_type := l_value
					end
				end
				l_cursor.forth
			end

		end

feature -- Field names

	application_field_name: STRING = "application"

	user_field_name: STRING = "user"

	host_name_field_name: STRING = "host_name"

	estudio_field_name: STRING = "estudio"

	changelist_field_name: STRING = "changelist"

	date_field_name: STRING = "date_utc"

	object_type_field_name: STRING = "object_type"

feature {NONE} -- Implementation

	append_key_value_to_text (a_text, a_key, a_value: STRING)
			-- Append key/value line to `a_text.
		require
			a_text_not_void: a_text /= Void
			a_key_not_void: a_key /= Void
			a_value_not_void: a_value /= Void
		do
			a_text.append_string (a_key + ": " + a_value + new_line)
		end

feature {NONE} -- Constants

	new_line: STRING = "%R%N"
			-- Windows new-line in binary file

end
