note
	description:
		"Store/Retrieve Eiffel objects along with metadata information, and with error checking"
	author: "Eric Bezault"
	last_editor: "$Author: dz $"
	date: "$Date: 2013/10/31 $"
	revision: "$Revision: #7 $"
	path: "$File: //rose/source/kernel/infrastructure/persistence/storable/rp_serializer.e $"

class RP_SERIALIZER [G]

inherit
	ROSE_SHARED_GLOBAL

feature -- Access

	last_retrieved_filename: detachable STRING
			-- filename of last successfully retrieved object
  		note
   			option: transient
   		attribute end

	retrieved (a_filename: STRING): G
			-- Retrieved Eiffel object from file `a_filename';
			-- Eventual errors are signaled.
		require
			valid_filename: a_filename /= Void
		local
			l_file: RAW_FILE
			l_filename, rescued: STRING
		do
			if rescued /= Void then
				if l_filename = Void or else l_filename.is_empty then
					l_filename := a_filename
				end
				error.signal (generator, "Cannot retrieve an Eiffel object from '$1': $2", <<l_filename, rescued>>)
			else
				error.disregard
				l_filename := global.persist.read_storable_filename (a_filename)
				if l_filename = Void or else l_filename.is_empty then
					error.signal (generator, "Cannot retrieve an Eiffel object from '$1'", <<a_filename>>)
					error.flush
				end
			end
			if not error.signaled then
				create l_file.make (l_filename)
				if not l_file.exists or else not l_file.is_readable then
					error.signal (generator, "Cannot retrieve an Eiffel object from '$1': file is not readable", <<l_filename>>)
					error.flush
				end
			end
			if not error.signaled then
				l_file.open_read
				l_file.read_stream (metadata_marker.count)
				if l_file.last_string.same_string (metadata_marker) then
						-- Skip metadata.
					l_file.read_integer_32
					l_file.go (metadata_marker.count + {PLATFORM}.integer_32_bytes + l_file.last_integer_32)
				else
					l_file.go (0)
				end
				if attached {G} l_file.retrieved as l_object then
					Result := l_object
					last_retrieved_filename := l_filename
				end
				l_file.close
				if Result /= Void then
					global.file_access_log.record_storable_read_access (l_filename, Result.generating_type)
				end
			end
		ensure
			valid_xor_error: Result /= Void xor error.signaled
			filename_set: Result /= Void implies last_retrieved_filename /= Void
		rescue
			if rescued = Void then
				rescued := last_exception_explanation
				retry
			end
		end

	retrieved_metadata (a_filename: STRING): detachable RP_SERIALIZABLE_METADATA
			-- Retrieved serializable metadata information from file `a_filename';
			-- Eventual errors are signaled.
		require
			valid_filename: a_filename /= Void
		local
			l_file: RAW_FILE
			l_filename, rescued: STRING
		do
			if rescued /= Void then
				if l_filename = Void or else l_filename.is_empty then
					l_filename := a_filename
				end
				error.signal (generator, "Cannot retrieve serializable metadata from '$1': $2", <<l_filename, rescued>>)
			else
				error.disregard
				l_filename := global.persist.read_storable_filename (a_filename)
				if l_filename = Void or else l_filename.is_empty then
					error.signal (generator, "Cannot retrieve serializable metadata from '$1'", <<a_filename>>)
					error.flush
				end
			end
			if not error.signaled then
				create l_file.make (l_filename)
				if not l_file.exists or else not l_file.is_readable then
					error.signal (generator, "Cannot retrieve serializable metadata from '$1': file is not readable", <<l_filename>>)
					error.flush
				end
			end
			if not error.signaled then
				l_file.open_read
				l_file.read_stream (metadata_marker.count)
				if l_file.last_string.same_string (metadata_marker) then
					l_file.read_integer_32
					l_file.read_stream (l_file.last_integer_32)
					create Result
					Result.from_text (l_file.last_string)
				else
					Result := Void
				end
				l_file.close
			end
		ensure
			error_implies_void: error.signaled implies Result = Void
		rescue
			if rescued = Void then
				rescued := last_exception_explanation
				retry
			end
		end

	retrieved_all (a_filename: STRING): TUPLE [RP_SERIALIZABLE_METADATA, attached G]
			-- Retrieved serializable metadata information and Eiffel object from file `a_filename';
			-- Eventual errors are signaled.
		require
			valid_filename: a_filename /= Void
		local
			l_file: RAW_FILE
			l_filename, rescued: STRING
			l_metadata: RP_SERIALIZABLE_METADATA
			l_offset: INTEGER_32
		do
			if rescued /= Void then
				if l_filename = Void or else l_filename.is_empty then
					l_filename := a_filename
				end
				error.signal (generator, "Cannot retrieve an Eiffel object from '$1': $2", <<l_filename, rescued>>)
			else
				error.disregard
				l_filename := global.persist.read_storable_filename (a_filename)
				if l_filename = Void or else l_filename.is_empty then
					error.signal (generator, "Cannot retrieve an Eiffel object from '$1'", <<a_filename>>)
					error.flush
				end
			end
			if not error.signaled then
				create l_file.make (l_filename)
				if not l_file.exists or else not l_file.is_readable then
					error.signal (generator, "Cannot retrieve an Eiffel object from '$1': file is not readable", <<l_filename>>)
					error.flush
				end
			end
			if not error.signaled then
				l_file.open_read
				l_file.read_stream (metadata_marker.count)
				if l_file.last_string.same_string (metadata_marker) then
						-- Read metadata.
					l_file.read_integer_32
					l_offset := l_file.last_integer_32
					l_file.read_stream (l_offset)
					create l_metadata
					l_metadata.from_text (l_file.last_string)
						-- Jump to object.
					l_file.go (metadata_marker.count + {PLATFORM}.integer_32_bytes + l_offset)
				else
					l_file.go (0)
				end
				if attached {G} l_file.retrieved as l_object then
					Result := [l_metadata, l_object]
					last_retrieved_filename := l_filename
				end
				l_file.close
				if Result /= Void then
					global.file_access_log.record_storable_read_access (l_filename, Result.generating_type)
				end
			end
		ensure
			valid_xor_error: Result /= Void xor error.signaled
		rescue
			if rescued = Void then
				rescued := last_exception_explanation
				retry
			end
		end

feature -- Basic operations

	store (a_object: G; a_filename: STRING)
			-- Store 'a_object' to 'a_filename', along with metadata information.
			-- 'a_filename' is interpreted through configuration
			-- An eventual error is signaled, no exception propagated
		require
			valid_object: a_object /= Void
			valid_filename: a_filename /= Void
		local
			l_file: RAW_FILE
			l_filename, rescued: STRING
			l_folder: STRING
			l_metadata: RP_SERIALIZABLE_METADATA
			l_metadata_string: STRING
		do
			if rescued /= Void then
				if l_filename = Void or else l_filename.is_empty then
					l_filename := a_filename
				end
				error.signal (generator, "Cannot save to '$1': $2", <<l_filename, rescued>>)
				error.flush
			else
				error.disregard
				l_filename := global.persist.write_storable_filename (a_filename)
				if l_filename = Void or else l_filename.is_empty then
					error.signal (generator, "Cannot save to '$1'", <<a_filename>>)
					error.flush
				end
			end
			if not error.signaled then
				create l_file.make (l_filename)
				if l_file.exists and then not l_file.is_writable then
					error.signal (generator, "Cannot save to '$1': file is not writable", <<l_filename>>)
					error.flush
				end
			end
			if not error.signaled then
				if l_file.exists then
					l_file.open_write
				else
					l_folder := global.formatter.directory (l_filename)
					if l_folder /= Void and then not global.is_directory_readable (l_folder) then
						global.recursive_create_directory (l_folder)
					end
					l_file.make_create_read_write (l_filename)
				end
				l_file.put_string (metadata_marker)
				create l_metadata.make_current
				l_metadata.set_object_type (a_object.generating_type)
				l_metadata_string := l_metadata.to_text
				l_file.put_integer_32 (new_line.count + l_metadata_string.count + object_field_name.count)
				l_file.put_string (new_line)
				l_file.put_string (l_metadata_string)
				l_file.put_string (object_field_name)
				l_file.go (metadata_marker.count + {PLATFORM}.integer_32_bytes + new_line.count + l_metadata_string.count + object_field_name.count)
				l_file.independent_store (a_object)
				l_file.close
				global.file_access_log.record_storable_write_access (l_filename, a_object.generating_type)
			end
		rescue
			if rescued = Void then
				rescued := last_exception_explanation
				retry
			end
		end

	store_without_metadata (a_object: G; a_filename: STRING) is
			-- Store 'a_object' with no metadata to 'a_filename'.
			-- 'a_filename' is interpreted through configuration
			-- An eventual error is signaled, no exception propagated
		require
			valid_object: a_object /= Void
			valid_filename: a_filename /= Void
		local
			l_file: RAW_FILE
			l_filename, rescued: STRING
			l_folder: STRING
		do
			if rescued /= Void then
				if l_filename = Void or else l_filename.is_empty then
					l_filename := a_filename
				end
				error.signal (generator, "Cannot save to '$1': $2", <<l_filename, rescued>>)
				error.flush
			else
				error.disregard
				l_filename := global.persist.write_storable_filename (a_filename)
				if l_filename = Void or else l_filename.is_empty then
					error.signal (generator, "Cannot save to '$1'", <<a_filename>>)
					error.flush
				end
			end
			if not error.signaled then
				create l_file.make (l_filename)
				if l_file.exists and then not l_file.is_writable then
					error.signal (generator, "Cannot save to '$1': file is not writable", <<l_filename>>)
					error.flush
				end
			end
			if not error.signaled then
				if l_file.exists then
					l_file.open_write
				else
					l_folder := global.formatter.directory (l_filename)
					if l_folder /= Void and then not global.is_directory_readable (l_folder) then
						global.recursive_create_directory (l_folder)
					end
					l_file.make_create_read_write (l_filename)
				end
				l_file.independent_store (a_object)
				l_file.close
				global.file_access_log.record_storable_write_access (l_filename, a_object.generating_type)
			end
		rescue
			if rescued = Void then
				rescued := last_exception_explanation
				retry
			end
		end

feature {NONE} -- Constants

	metadata_marker: STRING = "MD"
			-- Bytes to identify the fact that metadata are included in this file

	new_line: STRING = "%R%N"
			-- Windows new-line in binary file

	object_field_name: STRING = "object: %R%N"

end
