note

	description:

		"Cost tracking entity"

	author: "Sami Ben-Grine"
	last_editor: "$Author: e2 $"
	date: "$Date: 2011/02/02 $"
	revision: "$Revision: #7 $"
	path: "$File: //rose/source/kernel/expert_system/cost/rp_tax_lot.e $"

class RP_TAX_LOT

create {RP_COST_BASIS, RP_TAX_LOT}

	make

feature {NONE} -- Initialization

	make (a_id: like id; a_basis: like basis; a_cost: like collection; a_long_side: BOOLEAN) is
			-- Open the tax lot with 'a_cost_flow'
		require
			a_basis_not_void: a_basis /= Void
			a_cost_flow_not_void: a_cost /= Void
		do
			id := a_id
			basis := a_basis
			collection := a_cost
			is_long_side := a_long_side
			set_net (collection.quantity.amount)
			create disbursments.make_default
			create disbursment_flows.make_default
			basis.maintain_non_exhausted (Current)
		ensure
			id_set: id = a_id
			basis_set: basis = a_basis
			cost_set: collection = a_cost
			disbursments_not_void: disbursments /= Void and disbursment_flows /= Void
			disbursments_empty: disbursments.is_empty and disbursment_flows.is_empty
			net_not_void: net /= Void
			is_long_side_set: is_long_side = a_long_side
		end

feature {RP_COST_BASIS} -- Duplication

	duplicate (a_cost_base: RP_COST_BASIS): RP_TAX_LOT is
			-- Duplicate of current associated with `a_cost_base'
		require
			a_cost_base_not_void: a_cost_base /= Void
		do
			create Result.make (id, a_cost_base, collection.twin, is_long_side)
			Result.set_net (net)
			Result.set_disbursments (disbursments.twin, disbursment_flows.twin)
		ensure
			duplicate_not_void: Result /= Void
			duplicate_not_equal: Result /= Current
			basis_set: Result.is_linked_to_basis (a_cost_base)
			same_as_current_except_for_basis: same_except_for_basis (Result)
		end

feature {RP_TAX_LOT} -- Setting

	set_disbursments (a_disbursments: like disbursments; a_disbursment_flows: like disbursment_flows) is
			-- Set disbursments and disbursment_flows
		require
			a_disbursments_not_void: a_disbursments /= Void
			a_disbursment_flows_not_void: a_disbursment_flows /= Void
			no_void_disbursment: not a_disbursments.has_void_item
			no_void_disbusrment_flow: not a_disbursment_flows.has_void_item
		do
			disbursments := a_disbursments
			disbursment_flows := a_disbursment_flows
		ensure
			disbursments_set: disbursments = a_disbursments
			disbursment_flows_set: disbursment_flows = a_disbursment_flows
		end

feature -- Access

	id: INTEGER
			-- Identifier of that lot in the parent cost 'basis'

	basis: RP_COST_BASIS
			-- Parent

	collection: RP_COST
			-- Collected cost

	disbursments: DS_HASH_TABLE [RP_DECIMAL, INTEGER]
			-- Disbursed units
			-- XXX FIXME these comsume a TON of memory for cash

	disbursment_flows: DS_HASH_TABLE [RP_FLOW, INTEGER]
			-- Disbursment flows
			-- XXX FIXME these comsume a TON of memory for cash

	net: RP_DECIMAL
			-- Running total of units from all cost flows
			-- This is the collection minus all disbursments

feature -- Status Report

	is_exhausted: BOOLEAN
			-- Is this tax lot completely disbursed?

	is_long_side: BOOLEAN
			-- Is this tax lot for LONG side?

	is_short_side: BOOLEAN is
			-- Is this tax lot for SHORT side?
		do
			Result := not is_long_side
		end

	is_linked_to_basis (a_basis: RP_COST_BASIS): BOOLEAN is
			-- Is our basis `a_basis'?
		do
			Result := basis = a_basis
		ensure
			definition: Result = (basis = a_basis)
		end

	same_except_for_basis (a_tax_lot: RP_TAX_LOT): BOOLEAN is
			-- Are all attributes equal to the ones in `a_tax_lot' ignoring `basis'?
		require
			a_tax_lot_not_void: a_tax_lot /= Void
		do
			Result := id = a_tax_lot.id and
				collection ~ a_tax_lot.collection and
				disbursments ~ a_tax_lot.disbursments and
				disbursment_flows ~ a_tax_lot.disbursment_flows and
				net ~ a_tax_lot.net and
				is_exhausted = a_tax_lot.is_exhausted and
				is_long_side = a_tax_lot.is_long_side
		end

feature -- Change

	scale (a_factor: RP_DECIMAL) is
			-- Scale all cost flows with 'a_factor'
			-- The monies are unaffected
		require
			a_factor_not_void: a_factor /= Void
		local
			l_cursor: DS_HASH_TABLE_CURSOR [RP_DECIMAL, INTEGER]
		do
			collection.scale (a_factor)
				-- scale the net amount
			net := net * a_factor
			from
				-- scale disbursments
				l_cursor := disbursments.new_cursor
				l_cursor.start
			until
				l_cursor.after
			loop
				l_cursor.replace (l_cursor.item * a_factor)
				l_cursor.forth
			end
		end

feature -- Query

	total_cost (a_currency: RP_CURRENCY; a_fx: RP_CURRENCY_MARKET_DATA_SOURCE): RP_MONIES is
			-- Total cost in 'a_currency'
		local
			l_unit_cost: RP_MONIES
		do
			l_unit_cost := collection.unit_cost (a_currency, a_fx)
			Result := l_unit_cost.currency.monies (l_unit_cost.amount * net)
		ensure
			total_cost_not_void: Result /= Void
		end

feature -- Basic Operations

	disburse (a_units: RP_DECIMAL; a_flow: RP_FLOW) is
			-- Disburse `a_units' caused by `a_flow'
		require
			a_units_not_void: a_units /= Void
			a_flow_not_void: a_flow /= Void
			same_instrument: a_flow.quantity.instrument ~ collection.quantity.instrument
			positive_units: a_units.sign > 0
		do
			basis.can_check_non_exhausted_tax_lots := False
			if is_long_side then
				set_net (net - a_units)
			else
				set_net (net + a_units)
			end
			disbursments.force (a_units, disbursments.count + 1)
			disbursment_flows.force (a_flow, disbursment_flows.count + 1)
			basis.maintain_non_exhausted (Current)
			basis.can_check_non_exhausted_tax_lots := True
		ensure
			same_count: disbursments.count = disbursment_flows.count
		end

feature {RP_TAX_LOT} -- Implementation

	set_net (v: like net) is
			-- Set vale of `net' to `v'
		require
			non_void_v: v /= Void
		do
			net := v
			update_exhausted_flag
		end

feature {NONE} -- Implementation

	update_exhausted_flag is
			-- Maintain status of `is_exhausted'
		local
			l_sign: INTEGER
		do
			if is_long_side then
				l_sign := net.sign
			else
				l_sign := -net.sign
			end
			if l_sign <= 0 then
				is_exhausted := True
			end
		end

invariant

	basis_is_not_void: basis /= Void
	disbursments_not_void: disbursments /= Void and disbursment_flows /= Void

end
