note

	description:
	"[
		A permutation of size k of a range of size n is a (not necessarily) sorted subsequence
			of size k of the total range, i.e., a subsequence of elements at k positions among
			the n positions in the range.
			
		This class enumerates all possible k-permutations of the range 1..n in lexicographic ordering
		begining with the trivial permutation of {1,2,...,k}
		Sample usage is:
			from
				create v.make_variation (n, k)
			until
				v /= Void
			loop
				... use v...
				v := v.next
			end

	]"
	author: "Ivan Tarnavsky"
	last_editor: "$Author: it $"
	date: "$Date: 2011/02/23 $"
	revision: "$Revision: #1 $"
	path: "$File: //rose/source/kernel/math/rp_variations.e $"

class RP_VARIATIONS

inherit

	ARRAY [INTEGER]
		rename
			count as k
		export
			{ANY} k, item, do_all, do_all_with_index, do_if, do_if_with_index, valid_index, upper, lower
			{RP_VARIATIONS} area, deep_twin, is_deep_equal, standard_is_equal, subarray, object_comparison
			{NONE} all
		end

create

	make_variation

create {RP_VARIATIONS}

	make_with_permutations

feature {NONE} -- Initialization

	make_variation (a_n, a_k: INTEGER) is
			-- Initialize `Current'
		require
			n_positive: a_n > 0
			k_positive: a_k > 0
			k_le_n: a_k <= a_n
		do
			make_with_permutations (create {RP_PERMUTATIONS}.make_permutation (a_n), a_k)
		ensure
			n_set: n = a_n
			k_set: k = a_k
			initial_values: (1 |..| k).as_array.area.is_equal (area)
		end

	make_with_permutations (a_permutations: like permutations; a_k: INTEGER) is
			-- Initialize with `a_permutations' and `a_k'
		require
			a_permutations_not_void: a_permutations /= Void
			a_permutations_not_empty: a_permutations.n > 0
			k_positive: a_k > 0
			k_le_n: a_k <= a_permutations.n
		do
			permutations := a_permutations
			make_from_array (permutations.subarray (1, a_k))
		ensure
			permutations_set: permutations = a_permutations
			k_set: k = a_k
			values_set: permutations.subarray (1, a_k).area.is_equal (area)
		end

feature -- Access

	n: INTEGER is
			-- Set size
		do
			Result := permutations.n
		end

feature -- Query

	next: RP_VARIATIONS
			-- Next variation from set of possible variations
			-- Returns Void when no next variation
		local
			a: ARRAY [INTEGER]
			p: RP_PERMUTATIONS
			i, j, t: INTEGER
		do
				-- reverse P[k+1,n]
			create a.make_from_array (permutations)
			from
				i := k + 1
				j := n
			until
				i >= j
			loop
				t := a[j]
				a[j] := a[i]
				a[i] := t
				i := i + 1
				j := j - 1
			end
			create p.make_from_array (a)
			p := p.next
			if p /= Void then
				create Result.make_with_permutations (p, k)
			end
		ensure
			definition: Result = Void or else Current < Result
		end

feature -- Comparison

	is_less alias "<" (other: like Current): BOOLEAN
			-- <Precursor>
		local
			i: INTEGER
		do
			if n < other.n then
				Result := True
			elseif n = other.n then
				if k < other.k then
					Result := True
				elseif k = other.k then
					from
						i := lower
					until
						i > upper or Result
					loop
						Result := Result or item (i) < other.item (i)
						i := i + 1
					end
				end
			end
		end

feature -- Status

	valid_values (a_n: INTEGER; a_r: ARRAY [INTEGER]): BOOLEAN is
			-- Are all values in `a_r' valid for `a_n'?
		local
			i: INTEGER
		do
			if a_n > 0 and a_r /= Void and then 0 < a_r.count and a_r.count <= a_n then
				from
					Result := True
					i := a_r.lower
				until
					i > a_r.upper or not Result
				loop
					Result := Result and 0 < a_r.item (i) and a_r.item (i) <= a_n
					i := i + 1
				end
			end
		end

	has_repetitions: BOOLEAN
			-- Are there repeating values in `Current'?
		local
			l_set: DS_HASH_SET [INTEGER]
		do
			create l_set.make (k)
			do_all (agent l_set.force (?))
			Result := k /= l_set.count
		end

feature {NONE} -- Implementation

	permutations: RP_PERMUTATIONS
			-- Underlying permutations

invariant

	permutations_not_void: permutations /= Void
	n_positive: n > 0
	k_positive: k > 0
	k_le_n: k <= n
	valid_values: valid_values (n, Current)
	without_repetitions: not has_repetitions

end
