note

	description:

		"Calculation of gross amount based on withholding tax gross up rules"

	author: "Ivan Tarnavsky"
	last_editor: "$Author: e2 $"
	date: "$Date: 2011/11/29 $"
	revision: "$Revision: #15 $"
	path: "$File: //rose/source/kernel/expert_system/legacy/contracts/rp_withholding_tax_gross_up.e $"

deferred class RP_WITHHOLDING_TAX_GROSS_UP

inherit

	ANY

	RP_PERFORMANCE_CONSTANTS_SHARED
		export {NONE} all end

feature {NONE} -- Implementation

	gross_amount (a_force_client_rate: BOOLEAN; a_reference_system: RP_REFERENCE_SYSTEM; a_portfolio: RP_PORTFOLIO; a_instrument: RP_INSTRUMENT; a_date: ROSE_DATE; a_units: RP_QUANTITY; a_net_amount: RP_MONIES; a_taxbas: STRING; a_perf_effect: RP_MONIES; a_cash_rate: RP_DECIMAL; a_cash_rate_currency: RP_CURRENCY; a_ex_date: ROSE_DATE): RP_MONIES is
			-- Calculate gross amount
		require
			a_reference_system_not_void: a_reference_system /= Void
			a_portfolio_not_void: a_portfolio /= Void
			a_instrument_not_void: a_instrument /= Void
			a_date_not_void: a_date /= Void
			a_units_not_void: a_units /= Void
			a_net_amount_not_void: a_net_amount /= Void
			a_perf_effect_not_void: a_perf_effect /= Void
			a_cash_rate_and_currency_set_together: (a_cash_rate /= Void) = (a_cash_rate_currency /= Void)
			a_ex_date_can_be_void: True
			a_force_client_rate_allowed_only_for_u_zone_portfolios: a_force_client_rate implies a_portfolio.zone.is_united_states
		local
			l_country_rate: RP_DECIMAL
			l_client_rate: RP_DECIMAL
			l_target_category: STRING
			l_rate, l_amount: RP_DECIMAL
			l_tax_base_parm: RP_TAX_BASE_PARM
			l_country: RP_COUNTRY
			l_fx_rate: RP_DECIMAL
		do
			if a_portfolio.zone.is_united_states then
					-- VMS code uses US for all U zone portfolios
					-- the logic is replicated here
				l_country := a_reference_system.country_from_iso_code ("US")
			else
				l_country := a_instrument.country_of_issue (a_reference_system, a_date)
			end
			l_country_rate := l_country.country_withholding_tax_rate (a_date)
			l_client_rate := l_country.portfolio_withholding_tax_rate (a_portfolio, a_date)
			if a_units.amount.is_zero then
					-- Some J zone events have zero units
					-- hence we need to use the computed net amount
				l_amount := a_net_amount.amount / (1 - l_client_rate)
			elseif a_taxbas = Void then
				if a_perf_effect.amount.is_zero then
					l_amount := a_net_amount.amount / (1 - l_client_rate)
				else
					l_amount := a_perf_effect.amount
				end
			else
				l_tax_base_parm := a_reference_system.taxation_basis (a_taxbas, a_date)
				if l_tax_base_parm = Void then
					(create {EXCEPTIONS}).raise ("Invalid TAXBAS `" + a_taxbas + "' for " + a_portfolio.code + " " + a_instrument.brimsym  + " " + a_date.ccyymmdd.out)
				end
				if use_special_rate (l_tax_base_parm) then
					l_client_rate := create {RP_DECIMAL}.make_from_string (l_tax_base_parm.special_rate) / 100
				end
				l_target_category := dividend_announcement_category (a_force_client_rate, a_taxbas, a_portfolio, l_country, l_tax_base_parm)
				if l_target_category ~ constants.dividend_announcement_category_force_client_rate then
					l_amount := a_net_amount.amount / (1 - l_client_rate)
				else
					l_rate := gross_rate (a_taxbas, a_cash_rate, l_client_rate, l_country_rate, l_target_category, withholding_tax_is_recoverable (a_taxbas, a_portfolio, l_country, l_tax_base_parm))
					l_fx_rate := a_net_amount.currency.currency.market_exchange_rate (default_market, a_cash_rate_currency.currency, a_ex_date)
					l_rate := l_rate * l_fx_rate
					l_amount := l_rate * a_units.amount
				end
			end
			if a_portfolio.zone.is_japan then
				Result := a_net_amount.currency.monies (l_amount.rounded (0))
			else
				Result := a_net_amount.currency.monies (l_amount.rounded (2))
			end
		ensure
			gross_amount_not_void: Result /= Void
		end

	net_amount (a_force_rate: BOOLEAN; a_reference_system: RP_REFERENCE_SYSTEM; a_portfolio: RP_PORTFOLIO; a_instrument: RP_INSTRUMENT; a_date: ROSE_DATE; a_units: RP_QUANTITY; a_net_amount: RP_MONIES; a_taxbas: STRING; a_perf_effect: RP_MONIES; a_cash_rate: RP_DECIMAL; a_cash_rate_currency: RP_CURRENCY; a_ex_date: ROSE_DATE): RP_MONIES is
			-- Calculate gross amount
		require
			a_reference_system_not_void: a_reference_system /= Void
			a_portfolio_not_void: a_portfolio /= Void
			a_instrument_not_void: a_instrument /= Void
			a_date_not_void: a_date /= Void
			a_units_not_void: a_units /= Void
			a_net_amount_not_void: a_net_amount /= Void
			a_perf_effect_not_void: a_perf_effect /= Void
			a_cash_rate_not_void: a_cash_rate /= Void
			a_cash_rate_currency_not_void: a_cash_rate_currency /= Void
			a_ex_date_not_void: a_ex_date /= Void
			a_force_rate_allowed_only_for_j_zone_portfolios: a_force_rate implies a_portfolio.zone.is_japan
		local
			l_country_rate: RP_DECIMAL
			l_client_rate: RP_DECIMAL
			l_target_category: STRING
			l_rate, l_amount: RP_DECIMAL
			l_tax_base_parm: RP_TAX_BASE_PARM
			l_country: RP_COUNTRY
			l_fx_rate: RP_DECIMAL
		do
			if a_portfolio.zone.is_united_states then
					-- VMS code uses US for all U zone portfolios
					-- the logic is replicated here
				l_country := a_reference_system.country_from_iso_code ("US")
			else
				l_country := a_instrument.country_of_issue (a_reference_system, a_date)
			end
			l_country_rate := l_country.country_withholding_tax_rate (a_date)
			l_client_rate := l_country.portfolio_withholding_tax_rate (a_portfolio, a_date)
				-- FIXME ASSUME 20% TAX WITHHOLDING RATE IN J ZONE PRIOR TO 2003/3/31
			if a_force_rate then
				l_client_rate := forced_withholding_tax_rate (l_client_rate, a_portfolio, a_date)
			end
			if a_units.amount.is_zero then
					-- Some J zone events have zero units
					-- hence we need to use the computed net amount
				l_amount := a_net_amount.amount / (1 - l_client_rate)
			elseif a_taxbas = Void then
				if a_perf_effect.amount.is_zero then
					l_amount := a_net_amount.amount / (1 - l_client_rate)
				else
					l_amount := a_perf_effect.amount
				end
			else
				l_tax_base_parm := a_reference_system.taxation_basis (a_taxbas, a_date)
				if l_tax_base_parm = Void then
					(create {EXCEPTIONS}).raise ("Invalid TAXBAS `" + a_taxbas + "' for " + a_portfolio.code + " " + a_instrument.brimsym  + " " + a_date.ccyymmdd.out)
				end
				l_target_category := dividend_announcement_category (False, a_taxbas, a_portfolio, l_country, l_tax_base_parm)
				if l_target_category ~ constants.dividend_announcement_category_force_client_rate then
					l_amount := a_net_amount.amount / (1 - l_client_rate)
				else
					l_rate := net_rate (a_taxbas, a_cash_rate, l_client_rate, l_country_rate, l_target_category, withholding_tax_is_recoverable (a_taxbas, a_portfolio, l_country, l_tax_base_parm))
					l_fx_rate := a_net_amount.currency.currency.market_exchange_rate (default_market, a_cash_rate_currency.currency, a_ex_date)
					l_rate := l_rate * l_fx_rate
					l_amount := l_rate * a_units.amount
				end
			end
			if a_portfolio.zone.is_japan then
				Result := a_net_amount.currency.monies (l_amount.rounded (0))
			else
				Result := a_net_amount.currency.monies (l_amount.rounded (2))
			end
		end

	default_market: RP_LEGACY_PERF_FX is
			-- FX market
		once
			create Result
		end

	forced_withholding_tax_rate (a_client_rate: RP_DECIMAL; a_portfolio: RP_PORTFOLIO; a_date: ROSE_DATE): RP_DECIMAL is
			-- Withholding tax rate forced
			-- This is an exceptional case that didn't make it yet in the configuration
		local
			l_code: STRING
			l_date: INTEGER
		do
			Result := a_client_rate
			if a_portfolio.zone.is_japan then
				l_code := a_portfolio.code.as_upper
				l_date := a_date.ccyymmdd
					-- Eventup entries have been manually updated to 7% rate and this wasn't reflected in the withholding tax perf parm file 
				if l_code.is_equal ("BRUSHJ") and (l_date >= 20061226) and (l_date <= 20070726) then
					create Result.make_from_string ("0.07")
						-- Eventup entries have been manually updated to various rates. We use the most common one 10.563%, instead of 10% as in the perf parm file.
				elseif l_code.is_equal ("DINGOJ") and (l_date >= 20030527) and (l_date <= 20030925) then
					create Result.make_from_string ("0.10563")
				elseif (l_date <= 20021231) and then (not j_zone_portfolios_not_forced.has (l_code)) and then a_client_rate.out.is_equal ("0.15") then
					create Result.make_from_string ("0.20")
				end
			end
		end

	j_zone_portfolios_not_forced: DS_HASH_SET [STRING] is
			-- List of J zone portfolios that won't be forced from 15% to 20% tax rate before 20030331
		once
			create Result.make_equal (11)
			Result.put_new ("ASIEACTJ")
			Result.put_new ("BRAINJ")
			Result.put_new ("COLJ")
			Result.put_new ("DEJAIL")
			Result.put_new ("GLOBEJ")
			Result.put_new ("JAPONAC")
			Result.put_new ("JOHNJ")
			Result.put_new ("NIMBUSJ")
			Result.put_new ("PACIFIQJ")
			Result.put_new ("STARJ")
			Result.put_new ("SMILEJ")
		ensure
			result_not_void: Result /= Void
		end

	gross_rate (a_taxbas: STRING; a_dividend_rate, a_client_withholding_rate, a_country_withholding_rate: RP_DECIMAL; a_dividend_announcement: STRING; a_is_tax_recoverable: BOOLEAN): RP_DECIMAL is
			-- Gross rate for `a_dividend_rate' and `a_dividend_announcement' and `a_is_tax_recoverable'
		require
			a_taxbas_not_void: a_taxbas /= Void
			a_dividend_rate_not_void: a_dividend_rate /= Void
			a_client_withholding_rate_not_void: a_client_withholding_rate /= Void
			a_country_withholding_rate_not_void: a_country_withholding_rate /= Void
			a_dividend_announcement_not_void: a_dividend_announcement /= Void
		do
			if a_dividend_announcement ~ constants.dividend_announcement_category_taxfree then
					-- gross = net = dividend rate
				Result := a_dividend_rate
			elseif a_dividend_announcement ~ constants.dividend_announcement_category_gross then
				if a_is_tax_recoverable then
						-- not subject to compulsory withholding
						-- hence gross = dividend rate
					Result := a_dividend_rate
				else
						-- dividend is announced gross of tax but
						-- is subject to compulsory withholding
						-- hence gross = dividend rate * (1 - withholding rate)
					Result := a_dividend_rate * (1 - a_client_withholding_rate)
				end
			elseif a_dividend_announcement ~ constants.dividend_announcement_category_net then
				if a_is_tax_recoverable then
						-- dividend is announced net of tax and the tax withheld
						-- can be recovered subsequently
						-- hence gross = dividend rate / (1 - withholding rate)
					Result := a_dividend_rate / (1 - a_country_withholding_rate)
				else
						-- dividend is announced net of tax and the tax is
						-- not possible to be recovered
						-- hence gross = dividend rate
					Result := a_dividend_rate
				end
			end
			if Result = Void then
				(create {EXCEPTIONS}).raise ("Invalid dividend announcement `" + a_dividend_announcement + "' recoverable `" + a_is_tax_recoverable.out + " for TAXBAS `" + a_taxbas + "'")
			end
		ensure
			gross_rate_not_void: Result /= Void
		end

	net_rate (a_taxbas: STRING; a_dividend_rate, a_client_withholding_rate, a_country_withholding_rate: RP_DECIMAL; a_dividend_announcement: STRING; a_is_tax_recoverable: BOOLEAN): RP_DECIMAL is
			-- Net rate for `a_dividend_rate' and `a_dividend_announcement' and `a_is_tax_recoverable'
		require
			a_taxbas_not_void: a_taxbas /= Void
			a_dividend_rate_not_void: a_dividend_rate /= Void
			a_client_withholding_rate_not_void: a_client_withholding_rate /= Void
			a_country_withholding_rate_not_void: a_country_withholding_rate /= Void
			a_dividend_announcement_not_void: a_dividend_announcement /= Void
		do
			if a_dividend_announcement ~ constants.dividend_announcement_category_taxfree then
					-- gross = net = dividend rate
				Result := a_dividend_rate
			elseif a_dividend_announcement ~ constants.dividend_announcement_category_gross then
				if a_is_tax_recoverable then
					Result := a_dividend_rate * (1 - a_client_withholding_rate)
				else
					Result := a_dividend_rate / (1 - a_country_withholding_rate)
				end
			elseif a_dividend_announcement ~ constants.dividend_announcement_category_net then
				if a_is_tax_recoverable then
					Result := a_dividend_rate * (1 - a_client_withholding_rate)
				else
					Result := a_dividend_rate / (1 - a_country_withholding_rate)
				end
			end
			if Result = Void then
				(create {EXCEPTIONS}).raise ("Invalid dividend announcement `" + a_dividend_announcement + "' recoverable `" + a_is_tax_recoverable.out + " for TAXBAS `" + a_taxbas + "'")
			end
		ensure
			net_rate_not_void: Result /= Void
		end

	dividend_announcement_category (a_force_client_rate: BOOLEAN; a_taxbas: STRING; a_portfolio: RP_PORTFOLIO; a_country: RP_COUNTRY; a_tax_base_parm: RP_TAX_BASE_PARM): STRING is
			-- Dividend announcement category
		require
			a_taxbas_not_void: a_taxbas /= Void
			a_portfolio_not_void: a_portfolio /= Void
			a_country_not_void: a_country /= Void
			a_tax_base_parm_not_void: a_tax_base_parm /= Void
			a_force_client_rate_allowed_only_for_u_zone_portfolios: a_force_client_rate implies a_portfolio.zone.is_united_states
		do
			Result := a_tax_base_parm.gross_net.as_upper
			if a_portfolio.zone.is_united_states and a_force_client_rate then
					-- Force the use of client rate for U zone
					-- as implemended in the VMS code
				Result := constants.dividend_announcement_category_force_client_rate
			end
			if (a_portfolio.zone.is_europe or a_country.is_new_zealand) and a_taxbas ~ "##" then
					-- Dividends announced as "No special tax condition applies" i.e. taxbas=##
					-- in Europe and in New Zealant are GROSS with revocerable tax
					-- This overrides the default:
					-- START_DATE  END_DATE  TXBAS DESCRIPTION                                       GROSS_NET RECOVERABLE
					-- 00000000    99999999        No special tax condition applies                    NET          Y
				Result := constants.dividend_announcement_category_gross
			end
			if a_taxbas ~ "M" and then a_portfolio.zone.is_japan then
					-- Japanese dividends announced with taxbas=M are GROSS
					-- This overrides the default:
					-- START_DATE  END_DATE  TXBAS DESCRIPTION                                       GROSS_NET RECOVERABLE
					-- 00000000    99999999    M   Ex-distribution/dividend trading starts at close o  UNUSED       Y
				Result := constants.dividend_announcement_category_gross
			end
		ensure
			dividend_announcement_category_not_void: Result /= Void
		end

	withholding_tax_is_recoverable (a_taxbas: STRING; a_portfolio: RP_PORTFOLIO; a_country: RP_COUNTRY; a_tax_base_parm: RP_TAX_BASE_PARM): BOOLEAN is
			-- Tax recoverability
		require
			a_tax_base_parm_not_void: a_tax_base_parm /= Void
		do
			Result := a_tax_base_parm.recoverable ~ "Y"
		end

	use_special_rate (a_tax_base_parm: RP_TAX_BASE_PARM): BOOLEAN is
			-- Should special rate be used
		require
			a_tax_base_parm_not_void: a_tax_base_parm /= Void
		do
			Result := a_tax_base_parm.use_special_rate ~ "Y"
		end

end
