indexing

	description:

		"Composite made of portfolios"

	date: "$Date: 2010/06/29 $"
	revision: "$Revision: #17 $"

class RP_ZONAL_PRODUCT

inherit

	RP_PRODUCT

create {RP_REFERENCE_SYSTEM}

	make

feature -- Query

	is_global: BOOLEAN is False
			-- `Current' is zonal product.

	portfolios: RP_PORTFOLIO_HISTORY is
			-- Interval history of portfolios included in the composite
		do
			Result := retrieve_portfolios_history
		end

	portfolio_weights (a_dv_query: FUNCTION [ANY, TUPLE [RP_PORTFOLIO, RP_CURRENCY, ROSE_DATE], RP_PERFORMANCE_DAILY_VALUES]; a_currency: RP_CURRENCY; a_date_interval: ROSE_DATE_INTERVAL): DS_HASH_TABLE [TUPLE [value: RP_DECIMAL; date: ROSE_DATE], RP_PORTFOLIO] is
			-- Portfolio weights in date interval
		local
			l_portfolios: ROSE_LINEAR_ARRAY [RP_PORTFOLIO]
		do
			l_portfolios := portfolios.all_witnesses (a_date_interval.last)
			create Result.make (l_portfolios.count)
			l_portfolios.do_all (agent query_portfolio_nav (a_dv_query, a_date_interval.first, a_currency, Result, ?))
		end

	benchmark_weights (a_dv_query: FUNCTION [ANY, TUPLE [RP_PORTFOLIO, RP_CURRENCY, ROSE_DATE], RP_PERFORMANCE_DAILY_VALUES]; a_currency: RP_CURRENCY; a_date_interval: ROSE_DATE_INTERVAL): DS_HASH_TABLE [RP_DECIMAL, RP_BENCHMARK] is
			-- Benchmark weights in date interval
		local
			l_portfolio_weights: like portfolio_weights
		do
			l_portfolio_weights := portfolio_weights (a_dv_query, a_currency, a_date_interval)
			create Result.make (l_portfolio_weights.count)
			l_portfolio_weights.keys.do_all (agent accumulate_benchmark_weights (Result, l_portfolio_weights, a_date_interval.last, ?))
		end

feature {NONE} -- Implementation

	saved_portfolios: like portfolios
			-- Portfolio history cache

	retrieve_portfolios_history: like portfolios is
			-- Retrieve portfolios mapped to `Current'
		do
			if saved_portfolios = Void then
				create saved_portfolios.make_empty
				load_composite_membership (saved_portfolios)
			end
			Result := saved_portfolios
		end

	load_composite_membership (a_composite_membership: like portfolios) is
			-- Load composite membership to `a_composite_membership'
		require
			non_void_args: a_composite_membership /= Void
		do
			a_composite_membership.append (directly_mapped_portfolios)
			benchmarks.do_all (agent load_by_benchmark (a_composite_membership, ?))
			a_composite_membership.coalesce
		end

	load_by_benchmark (a_composite_membership: like portfolios; a_interval: RP_BENCHMARK_HISTORY_INTERVAL) is
			-- Load composite mapping by benchmark
		local
			l_benchmark: RP_BENCHMARK
		do
			l_benchmark := a_interval.benchmark
			if a_interval.weight > 0 then
				l_benchmark.portfolios.do_all (agent load_by_portfolio (a_composite_membership, a_interval, ?))
			end
		end

	load_by_portfolio (a_composite_membership: like portfolios; a_benchmark_interval: RP_BENCHMARK_HISTORY_INTERVAL; a_interval: RP_PORTFOLIO_HISTORY_INTERVAL) is
			-- Load benchmark history to portfolio relationship
		local
			l_portfolio: RP_PORTFOLIO
			l_start, l_end: ROSE_DATE
		do
			l_portfolio := a_interval.portfolio
			l_start := a_benchmark_interval.start_date.max (a_interval.start_date)
			l_end := a_benchmark_interval.end_date.min (a_interval.end_date)
			if l_start < l_end then
				l_portfolio.composite_state.do_all (agent check_comp_inclusion (a_composite_membership, l_portfolio, l_start, l_end, ?, "MappedVia" + a_benchmark_interval.benchmark.code))
			end
		end

	check_comp_inclusion (a_composite_membership: like portfolios; a_portfolio: RP_PORTFOLIO; a_start: ROSE_DATE; a_end: ROSE_DATE; a_is_in_composite: RP_BOOLEAN_HISTORY_INTERVAL; a_message: STRING) is
			-- Check composite inclusion
		local
			l_start, l_end: ROSE_DATE
		do
			l_start := a_start.max (a_is_in_composite.start_date)
			l_end := a_end.min (a_is_in_composite.end_date)
			if l_start < l_end then
				if a_is_in_composite.item then
					a_portfolio.no_direct_composite_mapping.do_all (agent check_direct_mapping (a_composite_membership, a_portfolio, l_start, l_end, ?, a_message))
				end
			end
		end

	check_direct_mapping (a_composite_membership: like portfolios; a_portfolio: RP_PORTFOLIO; a_start: ROSE_DATE; a_end: ROSE_DATE; a_is_not_directly_mapped: RP_BOOLEAN_HISTORY_INTERVAL; a_message: STRING) is
			-- Check direct mapping
		local
			l_start, l_end: ROSE_DATE
		do
			l_start := a_start.max (a_is_not_directly_mapped.start_date)
			l_end := a_end.min (a_is_not_directly_mapped.end_date)
			if l_start < l_end then
				if a_is_not_directly_mapped.item then
					if a_portfolio.end_date /= Void then
						l_end := l_end.min (a_portfolio.end_date)
					end
					check_begin_close_dates (a_composite_membership, a_portfolio, l_start, l_end, a_message)
				end
			end
		end

	check_begin_close_dates (a_composite_membership: like portfolios; a_portfolio: RP_PORTFOLIO; a_start: ROSE_DATE; a_end: ROSE_DATE; a_message: STRING) is
			-- Check composite begin/close dates
		local
			l_start, l_end: ROSE_DATE
			l_included: RP_PORTFOLIO_HISTORY_INTERVAL
		do
			l_start := a_start.max (a_portfolio.start_date)
			l_end := a_end.min (a_portfolio.end_date)
			if gcdb_product /= Void then
				if gcdb_product.begin_date /= Void then
					l_start := l_start.max (gcdb_product.begin_date)
				end
				if gcdb_product.close_date /= Void then
					l_end := l_end.min (gcdb_product.close_date)
				end
			end
			if l_start < l_end then
				create l_included.make (a_portfolio, l_start, l_end, 1)
				l_included.set_message (a_message)
				a_composite_membership.extend (l_included)
			end
		end

	directly_mapped_portfolios: RP_PORTFOLIO_HISTORY is
			-- Retrieve directly mapped portfolios
		do
			Result := retrieve_directly_mapped_portfolios
		end

	saved_directly_mapped_portfolios: like directly_mapped_portfolios
			-- Cache for directly mapped portfolios

	retrieve_directly_mapped_portfolios: like directly_mapped_portfolios is
			-- Retrieve directly mapped portfolios
		do
			if saved_directly_mapped_portfolios = Void then
				create saved_directly_mapped_portfolios.make_empty
				load_directly_mapped_portfolios (saved_directly_mapped_portfolios)
			end
			Result := saved_directly_mapped_portfolios
		end

	load_directly_mapped_portfolios (a_history: like directly_mapped_portfolios) is
			-- Load directly mapped portfolios
		do
			if gcdb_product /= Void then
				gcdb_product.portfolios_by_composite_history.do_all (agent add_directly_mapped_portfolio (a_history, ?))
			end
		end

	add_directly_mapped_portfolio (a_history: like directly_mapped_portfolios; a_interval: ENTITY_INTERVAL [GCDB_PORTFOLIO]) is
			-- Add directly mapped portfolio
		local
			l_code: STRING
			l_portfolio: RP_PORTFOLIO
			l_start, l_end: ROSE_DATE
		do
			l_code := a_interval.entity.code
			l_portfolio := reference_system.portfolio_from_code (l_code)
			if not l_portfolio.is_unknown then
				l_start := a_interval.start_date.max (l_portfolio.start_date)
				l_end := a_interval.end_date.min (l_portfolio.end_date)
				if l_start < l_end then
					l_portfolio.composite_state.do_all (agent check_composite_state_for_direct_mapping (a_history, l_portfolio, l_start, l_end, ?))
				end
			end
		end

	check_composite_state_for_direct_mapping (a_composite_membership: like portfolios; a_portfolio: RP_PORTFOLIO; a_start: ROSE_DATE; a_end: ROSE_DATE; a_is_in_composite: RP_BOOLEAN_HISTORY_INTERVAL) is
			-- Check composite inclusion
		local
			l_start, l_end: ROSE_DATE
		do
			l_start := a_start.max (a_is_in_composite.start_date)
			l_end := a_end.min (a_is_in_composite.end_date)
			if l_start < l_end then
				if a_is_in_composite.item then
					check_begin_close_dates (a_composite_membership, a_portfolio, l_start, l_end, "DirectPortfolioMapping")
				end
			end
		end

	query_portfolio_nav (a_dv_query: FUNCTION [ANY, TUPLE [RP_PORTFOLIO, RP_CURRENCY, ROSE_DATE], RP_PERFORMANCE_DAILY_VALUES]; a_date: ROSE_DATE; a_currency: RP_CURRENCY; a_table: like portfolio_weights; a_portfolio: RP_PORTFOLIO) is
			-- Query portfolio value by using `a_dv_query' for `a_portfolio', `a_currency' on `a_date' and add to `a_table'
		local
			l_dailyvalues: RP_PERFORMANCE_DAILY_VALUES
			l_date: ROSE_DATE
			l_value: RP_DECIMAL
		do
			l_date := a_portfolio.zone.legacy_perf_calendar.ith_work_day_before (a_date, 1)
			l_dailyvalues := a_dv_query.item ([a_portfolio, a_currency, l_date])
			if l_dailyvalues /= Void and then not l_dailyvalues.portfolio_value.amount.is_zero then
				l_value := l_dailyvalues.portfolio_value.amount
				a_table.force ([l_value, l_date], a_portfolio)
			else
					-- NOTE: try the first market day in the month
				l_date := a_portfolio.zone.legacy_perf_calendar.next_work_day (a_date)
				l_dailyvalues := a_dv_query.item ([a_portfolio, a_currency, l_date])
				if l_dailyvalues /= Void and then not l_dailyvalues.portfolio_value.amount.is_zero then
					l_value := l_dailyvalues.portfolio_value.amount
					a_table.force ([l_value, l_date], a_portfolio)
				else
					(create {EXCEPTIONS}).raise ("No DailyValues for " + a_portfolio.code + " @ " + a_date.ccyymmdd.out)
				end
			end
		end

	accumulate_benchmark_weights (a_table: like benchmark_weights; a_portfolio_weights: like portfolio_weights; a_date: ROSE_DATE; a_portfolio: RP_PORTFOLIO) is
			-- Accumulate benchmark weights in `a_table'
		require
			non_void_args: a_table /= Void and a_portfolio_weights /= Void and a_date /= Void and a_portfolio /= Void
			portfolio_weights_has_portfolio: a_portfolio_weights.has (a_portfolio)
		local
			l_benchmark: RP_BENCHMARK
			l_weight: RP_DECIMAL
		do
			l_benchmark := a_portfolio.benchmark_history.witness (a_date)
			if l_benchmark /= Void then
				a_table.search (l_benchmark)
				if a_table.found then
					l_weight := a_table.found_item
				else
					create l_weight.make_from_integer (0)
				end
				l_weight := l_weight + a_portfolio_weights.item (a_portfolio).value
				a_table.force (l_weight, l_benchmark)
			end
		end

	membership_begin_date: like begin_date is
			-- Begin membership date.
		do
			Result := portfolios.min_start_date
		end

	membership_close_date: like close_date is
			-- Last membership date.
		do
			Result := portfolios.max_end_date
		end

end
