note

	description:

		"ROSE_CONTEXT version of PERSISTENT_SECURITY"

	author: "Mark Howard"
	last_editor: "$Author: ext-asurelia $"
	date: "$Date: 2015/03/04 $"
	revision: "$Revision: #104 $"
	path: "$File: //rose/source/kernel/investment_entities/security.e $"

class SECURITY

inherit

	INSTRUMENT
		undefine
			is_equal
		end

	INVESTMENT_ENTITY
		undefine
			peer_anchor,
			descriptive_name,
			name_identifier_keys,
			name_identifier_key
		end

	PERSISTENT_SECURITY
		undefine
			context,
			make,
			make_context,
			debug_output
		end

create

	make,
	make_with_zone_xrf,
	make_anonymous,
	make_surrogate

feature {NONE} -- Creation

	make_surrogate (a_security: like Current; a_zone: like zone) is
			--make as a version of a_security in a_zone
		require
			not_void: a_security /= Void and then a_zone /= Void
		do
			copy (a_security)
			zone_mnemonic := a_zone.mnemonic
			xrf_code := a_security.xrf_code_in (a_zone)
		end

feature -- Registration

	register is
			-- Register this entity.
		do
			if zone /= Void then
				zone.add_security (Current)
			end
		end

feature -- Foreign Entities

	surrogated_company: like context.company_template is
			-- if this is a foreign security of a company, transfer this company to a foreign company of the security zone
		local
			l_company: like company
		do
			l_company := company
			if l_company /= Void and then l_company.zone_mnemonic /= '%U' and then l_company.zone_mnemonic /= zone_mnemonic then
				create Result.make_surrogate (l_company, zone)
			end
		end

	surrogate (a_zone: like zone): like context.security_template is
			-- The `a_zone' version of Current if it is available
		local
			l_listing: like listing
			l_fsxrf_string: STRING
		do
			l_listing := zonal_listing (a_zone)
			if l_listing /= Void and then l_listing.is_valid then
				l_fsxrf_string := l_listing.fsxrf_code_key (today)
				if l_fsxrf_string /= Void and then l_fsxrf_string.is_integer then
					Result := a_zone.security (l_fsxrf_string.to_integer)
				end
			elseif local_xrf_code (a_zone) /= 0 then
				-- No zonal listing available but a dummy exists
				Result := a_zone.security (local_xrf_code (a_zone))
			end
		end

	foreign_cxrf: INTEGER is
			-- the company cxrf in a foreign zone
		local
			l_fcxrf: STRING
		do
			l_fcxrf := entity_code (fcxrf, Void)
			if l_fcxrf /= Void and then l_fcxrf.is_integer then
				Result := l_fcxrf.to_integer
			end
		end

	foreign_xrf_code (a_zone: like context.zone_template): INTEGER is
			-- The security xrf in a foreign zone
		local
			l_fsxrf: STRING
			j: INTEGER
			l_found: BOOLEAN
		do
			from
				j := 1
			until
				j > listing_count or l_found
			loop
				if listing (j).zone /= Void and then listing (j).zone.mnemonic.is_equal (a_zone.mnemonic) then
					l_fsxrf := fsxrf.dated_key (listing (j), Void)
					if l_fsxrf /= Void then
						Result := l_fsxrf.to_integer
					end
					l_found := True
				end
				j := j + 1
			end
		end

feature -- Access

	name (a_date: ROSE_DATE): STRING is
			-- Company name as of `a_date'.
		do
			Result := company.name_at (a_date)
		end

	cusip_code (a_date: ROSE_DATE): STRING is
			-- Cusip as of `a_date'.
		do
			Result := entity_code (cusip, a_date)
		end

	home_zone_listings: ROSE_LINEAR_ARRAY [like context.listing_template] is
			-- Listings for this security which are in the same zone as the security
		local
			i: INTEGER
			l_listing: like context.listing_template
		do
			create Result.make_empty
			if zone /= Void then
				from
					i := 1
				until
					i > listing_count
				loop
					l_listing := listing (i)
					if l_listing.exchange /= Void and then l_listing.zone_from_exchange = zone then
						Result.extend (l_listing)
					end
					i := i + 1
				end
			end
		ensure
			not_void: Result /= Void
		end

	zonal_company: like context.company_template is
			-- Zonal company
		do
			Result := company
			if not Result.zone_mnemonic.is_equal (zone_mnemonic) then
				Result := Result.surrogate (zone)
			end
		end

	zonal_listing (a_zone: like context.zone_template): like context.listing_template is
			-- First listing in `a_zone'
		local
			i: INTEGER
			l_found: BOOLEAN
			l_listing: like context.listing_template
		do
			from
				i := 1
			until
				l_found or else i > listing_count
			loop
				l_listing := listing (i)
				if l_listing.zone /= Void and then l_listing.zone.same_zone (a_zone) then
						-- same_zone allows us to equate a research and model zone without
						-- violating ANY.is_equal ensure requirements.
					if l_listing.xrf_code /= 0 then
							--try to get one with a legitimate xrf_code
						l_found := True
						Result := l_listing
					end
				end
				i := i + 1
			end
		end

	country_listing (a_country: like context.country_template; a_date: ROSE_DATE): like context.listing_template is
			-- Listing in `a_country'
		local
			i: INTEGER
			l_found: BOOLEAN
			l_listing: like context.listing_template
			l_listings: ROSE_LINEAR_ARRAY[like context.listing_template]
		do
			l_listings := listings.all_witnesses (a_date)
			if l_listings /= Void then
				from i := 1	until l_found or else i > l_listings.count loop
					l_listing := l_listings[i]
					if l_listing.country /= Void and then l_listing.country.is_equal (a_country) and then l_listing.reuters_code_after (a_date) /= Void then
						-- The reuters code is check to see if a listing is live on a_date (approximate only)
						l_found := True
						Result := l_listing
					end
					i := i + 1
				end
			end
		end

	country_listing_for_valuation (a_country: like context.country_template; a_date: ROSE_DATE): like context.listing_template is
			-- Country listing with fallback for CN securities.
		require
			valid_args: a_country /= Void
		do
			if a_country /= Void and then a_country.is_equal(context.cn) then
				Result := country_listing(context.hk, a_date)
				if Result = Void then
					Result := country_listing(context.sg, a_date)
				end
				if Result = Void then
					Result := country_listing(context.cn, a_date)
				end
				if Result = Void then
					Result := country_listing(context.us, a_date)
				end
			else
				Result := country_listing(a_country, a_date)
			end
		end

	zonal_return (a_zone: like context.zone_template; a_start_date, a_end_date: ROSE_DATE): REAL
			--Retrun of zonal listing between `a_start_date' and `a_end_date'
		local
			l_listing: like context.listing_template
		do
			l_listing := zonal_listing (a_zone)
			if l_listing /= Void then
				Result := l_listing.return (a_start_date, a_end_date).truncated_to_real
			end
		end

	zonal_last_price (a_zone: like context.zone_template; a_date: ROSE_DATE): REAL is
		local
			l_listing: like context.listing_template
		do
			l_listing := zonal_listing (a_zone)
			if l_listing /= Void then
				Result := l_listing.last_price (a_date).truncated_to_real
			end
		end

	zonal_scaled_last_price (a_zone: like context.zone_template; a_date: ROSE_DATE): REAL is
		local
			l_listing: like context.listing_template
		do
			l_listing := zonal_listing (a_zone)
			if l_listing /= Void then
				Result := l_listing.scaled_last_price (a_date).truncated_to_real
			end
		end

	zonal_unadjusted_closing_price (a_zone: like context.zone_template; a_date: ROSE_DATE): REAL is
		local
			l_listing: like context.listing_template
		do
			l_listing := zonal_listing (a_zone)
			if l_listing /= Void then
				Result := l_listing.unadjusted_closing_price (a_date).truncated_to_real
			end
		end

	zonal_adjusted_last_price (a_zone: like context.zone_template; a_as_of_date, a_date: ROSE_DATE): REAL is
		local
			l_listing: like context.listing_template
		do
			l_listing := zonal_listing (a_zone)
			if l_listing /= Void then
				Result := (l_listing.last_price (a_date) * l_listing.adjustment_history.foradj (a_as_of_date, a_date)).truncated_to_real
			end
		end

	zonal_closing_price (a_zone: like context.zone_template; a_date: ROSE_DATE): REAL is
		local
			l_listing: like context.listing_template
		do
			l_listing := zonal_listing (a_zone)
			if l_listing /= Void then
				Result := l_listing.closing_price (a_date).truncated_to_real
			end
		end

	return (a_start_date, a_end_date: ROSE_DATE): REAL is
			-- Return of the zonal listing
		do
			Result := zonal_return (zone, a_start_date, a_end_date)
		end

	last_price (a_date: ROSE_DATE): REAL is
			-- Last Price
		do
			Result := zonal_last_price (zone, a_date)
		end

	scaled_last_price (a_date: ROSE_DATE): REAL is
		do
			Result := zonal_scaled_last_price (zone, a_date)
		end

	unadjusted_closing_price (a_date: ROSE_DATE): REAL is
		do
			Result := zonal_unadjusted_closing_price (zone, a_date)
		end

	adjusted_last_price (a_as_of_date, a_date: ROSE_DATE): REAL is
		do
			Result := zonal_adjusted_last_price (zone, a_as_of_date, a_date)
		end

	closing_price (a_date: ROSE_DATE): REAL is
		do
			Result := zonal_closing_price (zone, a_date)
		end

	primary_listing: like context.listing_template is
			-- Undated version of primary listing,
			-- to be replaced.
		obsolete
			"Needs to be replaced by primary_listing_at (date).  Use RESEARCH_COMPANY.primary_listing."
		local
			l_listings: like listings
		do
			l_listings := listings
			if l_listings /= Void and then l_listings.first /= Void then
				Result := l_listings.first.entity
			end
			if Result /= Void then
				if Result.xrf_code = 0 and handle /= 0 then
					from
						l_listings.start
					until
						l_listings.off or Result.xrf_code /= 0
					loop
						Result := l_listings.current_item.entity
						l_listings.forth
					end
				else
					if not Result.zone_mnemonic.is_equal (zone_mnemonic) then
						from
							l_listings.start
						until
							l_listings.off or Result.zone_mnemonic.is_equal (zone_mnemonic)
						loop
							Result := l_listings.current_item.entity
							l_listings.forth
						end
					end
				end
			end
		end

	local_xrf_code (a_zone: like context.zone_template): INTEGER is
			-- local sxrf code of a_zone, if any
		require
			valid_zone: a_zone /= Void
		local
			l_local_sxrf: STRING
			j: INTEGER
			l_found: BOOLEAN
		do
			from
				j := 1
			until
				j > listing_count or l_found
			loop
				inspect a_zone.mnemonic
				when 'O' then
					l_local_sxrf := listing (j).local_fsxrf_key (Void)
				when 'S' then
					l_local_sxrf := listing (j).s_fsxrf_key (Void)
				when 'E' then
					l_local_sxrf := listing (j).e_fsxrf_key (Void)
				else
				end
				if l_local_sxrf /= Void then
					Result := l_local_sxrf.to_integer
					l_found := True
				end
				j := j + 1
			end
		end

	xrf_code_in (a_zone: like context.zone_template): INTEGER
			-- Convert zone, xrf_code into 'a_zone'
			-- foreign xrf if necessary
		do
			if zone.mnemonic = a_zone.mnemonic then
				Result := xrf_code
			else
				Result := foreign_xrf_code (a_zone)
				if Result = 0 then
					Result := local_xrf_code (a_zone)
				end
			end
		end

	fund_class_code (a_date: ROSE_DATE): STRING is
			-- Fund class code entity attribute
		do
			Result := entity_code (fund_class, a_date)
		end

	has_valid_country_listing_for_valuation (a_country: like context.country_template; a_date: ROSE_DATE): BOOLEAN
			-- Has a non-void country listing for a_country
			-- It also checks if the zone and the brim_country of the security is the same as company
			-- to ensure we don't get securities that are xDR issues from before the company relocated.
		local
			l_comp_brim, l_sec_brim: STRING
		do
			if attached country_listing_for_valuation (a_country, a_date) then
				l_comp_brim := company.brim_code_most_recent_key (a_date)
				l_sec_brim := brim_code_most_recent_key (a_date)
				if l_comp_brim /= Void and then l_sec_brim /= Void then
					Result := zone_mnemonic = company.zone_mnemonic and then
								l_comp_brim.substring (1, 2) ~ l_sec_brim.substring (1, 2)
				end
			end
		end

	has_valid_zonal_listing (a_zone: like context.zone_template): BOOLEAN
			-- Has a non-void zonal listing for a_zone
		require
			valid_args: a_zone /= Void
		do
			Result := zonal_listing (a_zone) /= Void and then zone_mnemonic = company.zone_mnemonic
		end

	is_common_security(a_date: ROSE_DATE): BOOLEAN
			-- Is this a common security?
			-- Current implementation: The last three chars of brimsym is "..." (anand, 2015/02/23)
		do
			if attached brim_code_most_recent_key (a_date) as l_brim then
				Result := l_brim.substring (13, 15) ~ common_security_subtype
			end
		end

	is_live_at(a_date: ROSE_DATE): BOOLEAN
			-- Is this live on a_date
			-- WORK IN PROGRESS: Use at our own risk (anand, 02-13-2015)
			-- Currently checks the validity of a security by checking all of:
			-- 			1. Valid ISIN on a_date
			--			2. Valid Reuters code for at least one listing (on or after a_date)
		local
			i: INTEGER
		do
			if isin_code_key (a_date) /= Void then
				if attached listings.all_witnesses (a_date) as l_listings then
					from i:=1 until Result or i>l_listings.count loop
						Result := l_listings[i].reuters_code_after (a_date) /= Void
						i := i + 1
					end
				end
			end
		end

	common_security_subtype: STRING = "..."
			-- The common security signfier in the brim code (13-15 chars)

end
