note

	description:

		"Objects representing signature tagged items."

	author: "Mauricio Bustos"
	last_editor: "$Author: it $"
	date: "$Date: 2015/02/18 $"
	revision: "$Revision: #120 $"
	path: "$File: //rose/source/kernel/persistence/databases/signature_tagged/signature_tagged_item.e $"

class
	SIGNATURE_TAGGED_ITEM[G -> COMPARABLE]

inherit
	STID_SUBMITTABLE
	EXCEPTIONS
		export
			{NONE} all
		end
	DATABASE_SERVER
		export
			{NONE} all
		end
	ROSE_SHARED_GLOBAL
		export
			{NONE} all
		end
	MOVIE_CAMERA_AWARE
		export
			{NONE} all
		end
	SIGNATURE_TAGGED_OUTPUT_LEVEL
		export
			{NONE} all
		end
	ROSE_SHARED_LOGGER

create
	make

feature {NONE} -- Creation

	make (a_server: SQL_DATA_SERVER; a_process_id: INTEGER; a_type_name, a_data_name: STRING) is
			-- Create the item
		require
			server_not_void: a_server /= Void
			type_name_not_void: a_type_name /= Void
			process_valid: a_process_id > 0
		local
			l_camera_name: STRING
		do
			debug ("make_movies")
				l_camera_name := generating_type + "/" + a_server.node + "/" + a_server.database
				if a_type_name /= Void then
					l_camera_name.append ("/" + a_type_name)
				end
				if a_data_name /= Void then
					l_camera_name.append ("/" + a_data_name)
				end
				create cam.make (l_camera_name)
			end
			process_id := a_process_id
			type_name := a_type_name
			data_name := a_data_name
			signature_item_server ?= a_server
			temp_table_server ?= signature_item_server
			create update_cache.make (create {KL_COMPARABLE_COMPARATOR [like update_cache_key]}.make)
			initialize_loaders
			set_has_write_locks_table
		ensure
			a_process_id_set: process_id = a_process_id
			type_name_set: type_name = a_type_name
			data_name_set: data_name = a_data_name
		end

feature -- Management

	cleanup is
			-- <Precursor>
		local
			l_statement: STRING
		do
			l_statement := "drop table " + data_value_temp_table
			temp_table_server.sql_execute_immediate (l_statement)
			if not temp_table_server.success then
				raise ("Could not drop the temporary table")
			end
			--truncate_temp_tables
		end

	submit_frequency: INTEGER
			-- how often should items be submitted

	set_submit_frequency (a_freq: INTEGER) is
			-- Set 'submit_frequency' to 'a_freq'
		do
			submit_frequency := a_freq
		end

	set_process_id (a_process_id: INTEGER) is
			-- Set 'process_id' to 'a_process_id'
		require
			valid_process_id: a_process_id > 0
		do
			process_id := a_process_id
		end

	truncate_temp_tables is
			-- <Precursor>
		do
			temp_table_server.sql_execute_immediate ("truncate table " + data_value_temp_table)
			if not temp_table_server.success then
				raise ("Could not truncate the temporary table")
			end
		end

	discard_updates is
			-- <Precursor>
		do
			update_cache.wipe_out
		end

feature -- Identifiers

	type_name: STRING
			-- the name of the type of data (e.g. "integer", "real", "string", etc)

	process_id: INTEGER
			-- the process id for this run

	data_name: STRING
			-- the name of the data, used for the data_name_prefix

feature -- Status

	items_replaced: INTEGER
			-- total number of items replaced

	items_replaced_at_last_submit: INTEGER
			-- number of items replaced on most recent submit

	is_path_clustered: BOOLEAN
			-- is path clustered?

	full_update_mode: BOOLEAN
			-- Store every action regardless of data change

	check_if_path_clustered is
			-- is path clustered?
		local
			l_predicate: ANONYMOUS_PREDICATE
		do
			create l_predicate.make_using_server (signature_item_server,
												  "exec sp_columns @table_name = '" + data_value_table + "',@column_name = 'path'")
			l_predicate.retrieve (Void,Void)
			if l_predicate.array.count > 0 then
				is_path_clustered := True
			end
		end

	check_if_full_update_mode is
			-- is full update mode enabled?
		do
			full_update_mode := config.boolean (signature_item_server.logical + "_stid_full_update",False)
		end

	unsubmitted: BOOLEAN is
			-- Is there any unsubmitted data?
		do
			Result := (update_cache.count /= 0) -- or (fast_insert_cache.count /= 0)
		end

	metrics: ROSE_LINEAR_ARRAY[PAIR[STRING, DOUBLE]] is
			-- keep track of what is happening
		local
			l_pair: PAIR[STRING, DOUBLE]
		do
			create Result.make_empty
			create l_pair.make (data_name_prefix + type_name + "_ITEMS_REPLACED", items_replaced)
			Result.extend (l_pair)
		end

	reset_metrics is
			-- reset metrics
		do
			items_replaced := 0
		end

feature {NONE} -- Database Features

	signature_item_server: SQL_DATA_SERVER
			-- server for access to the data table(s)

	temp_table_server: SQL_DATA_SERVER
			-- server for access to the temp table(s)

	retry_limit: INTEGER is 5
			-- retry no more than five times

feature {NONE} -- Fast Loaders

	initialize_loaders is
		-- initialize loaders
		do
			check_if_path_clustered
			create_temp_tables
			create_trigger
			check_if_full_update_mode
		end

	sql_type: STRING is
			-- the SQL type corresponding to the Eiffel type
		do
			if type_name.is_equal ("INTEGER") then
				Result := "INT"
			elseif type_name.is_equal ("REAL") then
				Result := "REAL"
			elseif type_name.is_equal ("DOUBLE") then
				Result := "FLOAT"
			elseif type_name.is_equal ("STRING") then
				Result := "VARCHAR (7950)"
			elseif type_name.is_equal ("IMAGE") then
				Result := "IMAGE"
			elseif type_name.is_equal ("DATETIME") then
				Result := "DATETIME"
			end
		end

	create_temp_tables is
		-- create the global temp table in the database, to be used as a staging table for data updates
		local
			l_statement: STRING
			l_server: SQL_DATA_SERVER
		do
			l_statement := "create table " + data_value_temp_table +
				" (item_id int NOT NULL,signature_id int NOT NULL,process_id int NOT NULL,"
			if is_path_clustered then
				l_statement := l_statement + "path int NOT NULL,"
			end

			l_statement := l_statement + "value " + sql_type + ")"
			l_server := signature_item_server.auxiliary_server
			l_server.begin_transaction
			l_server.sql_execute_immediate (l_statement)
			if not l_server.success then
				raise ("Could not create the temporary table")
			end

			if is_path_clustered then
				l_server.sql_execute_immediate ("create unique clustered index temp_index_1 on " + data_value_temp_table + " (path,signature_id,item_id)")
			else
				l_server.sql_execute_immediate ("create unique clustered index temp_index_1 on " + data_value_temp_table + " (signature_id,item_id)")
			end
			if not l_server.success then
				raise ("Could not create the index on the temporary table")
			end

			l_server.commit
		end

	delimited_data_record( a_item: DATA_RECORD[G] ): STRING is
			-- a pipe delimited representation of `a_item'
		do
			Result := a_item.process_id.out + "|" + a_item.signature_id.out + "|" + a_item.item_id.out + "|" + a_item.value.out
		end

feature {NONE} -- Implementation

	update_cache: DS_AVL_TREE [DATA_RECORD[G], NATURAL_64]
			-- cache for storing records to be applied to the database

	commit_updated_cache is
			-- persist the updated items in the cache
		local
			l_last_data_record: DATA_RECORD[G]
			l_fast_data_server: SQL_DATA_SERVER
			l_fast_insert_predicate: DATA_SERVER_FAST_LOAD
		do
			l_fast_data_server := servers.make_from_repository_fast_load (signature_item_server.logical)

			create l_last_data_record
			if is_path_clustered then
				l_fast_insert_predicate := l_fast_data_server.fast_loader (data_value_temp_table,l_last_data_record)
			else
				l_fast_insert_predicate := l_fast_data_server.fast_loader (data_value_temp_table,l_last_data_record)
			end

			from
				update_cache.start
			until
				update_cache.after
			loop
				l_fast_insert_predicate.insert (update_cache.item_for_iteration)
				update_cache.forth
			end
			l_fast_insert_predicate.commit
			if not l_fast_insert_predicate.success then
				raise ("failure in fast insert commit")
			end
		end

	clear_updated_fast_insert is
			-- clear out cache
		do
			update_cache.wipe_out
		end

	insert_predicate:  EXECUTION_PREDICATE is
			-- insert statement for adding records to the temp table
			-- FIXME -- is this obsolete? Aren't we using fast load now?
		local
			l_statement: STRING
		do
			if insert_predicate_saved = Void then
				--print ("creating insert_predicate%N")
				l_statement := "insert into " + data_value_temp_table +
					" (signature_id, item_id, value, process_id "
				if is_path_clustered then
					l_statement := l_statement + ",path "
				end
				l_statement := l_statement + ") values (:signature_id, :item_id, :value, :process_id"
				if is_path_clustered then
					l_statement := l_statement + ", :path "
				end
				l_statement := l_statement + ")"
				create insert_predicate_saved.make (temp_table_server,l_statement)
				if not insert_predicate_saved.success then
					raise ("Failed to create insert_predicate_saved")
				end
			end
			Result := insert_predicate_saved
		end

	insert_predicate_saved: EXECUTION_PREDICATE
			-- once per instance

	has_sid_item_index: BOOLEAN
			-- does sid_item index exist?
		local
			l_predicate: INTEGER_PREDICATE
			l_server: SQL_DATA_SERVER
		do
			if has_sid_item_index_saved = Void then
				l_server := signature_item_server.auxiliary_server
				l_server.begin_transaction
				create l_predicate.make (l_server,
						"SELECT count(*) as cnt %N%
						%FROM sys.indexes %N%
						%WHERE object_id = OBJECT_ID('" + data_value_table + "') AND name = '" + sid_item_index_name + "'")
				if not l_predicate.success then
					raise ("Error creating has_sid_item_index predicate")
				end
				l_predicate.retrieve (Void, Void)
				if not l_predicate.success then
					raise ("has_sid_item_index predicate failed")
				end
				l_server.commit
				create has_sid_item_index_saved
				has_sid_item_index_saved.set_item (l_predicate.array.first = 1)
			end
			Result := has_sid_item_index_saved.item
		end
		has_sid_item_index_saved : BOOLEAN_REF

feature -- Cache Dumps

	dump_update_cache is
			-- spew the contents of the update cache
		local
			l_value: STRING
		do
			from
				update_cache.start
			until
				update_cache.after
			loop
				if update_cache.item_for_iteration.value /= Void then
					l_value := update_cache.item_for_iteration.value.out
				else
					l_value := Void
				end

				logger.info (
					"Signature_ID :" + update_cache.item_for_iteration.signature_id.out +
					" Item_ID :" + update_cache.item_for_iteration.item_id.out +
					" Value :" + l_value
								)
				update_cache.forth
			end
		end

feature -- Replace

	replace (a_item: DATA_RECORD[G]) is
			-- queue an item to be replaced
		require
			item_not_void: a_item /= Void
		local
			l_key: like update_cache_key
		do
			l_key := update_cache_key (a_item)
			if update_cache.has (l_key) then
				raise ("Duplicate item for signature_id " + a_item.signature_id.out + " and item_id: " + a_item.item_id.out + " detected")
			end
			update_cache.force_new (a_item, l_key)
			items_replaced := items_replaced + 1
			process_new_data
		end

feature -- Insert

	insert (a_item: DATA_RECORD[G]) is
			-- fast load insert a_item
		local
			l_server: SQL_DATA_SERVER
		do
			if insert_saved = Void then
				l_server := servers.make_from_repository_fast_load (signature_item_server.logical)
				-- FIXME -- should this be inserting to the temp table?
				insert_saved := l_server.fast_loader (data_value_table,a_item)
			end
			insert_saved.insert (a_item)
		end
	insert_saved: DATA_SERVER_FAST_LOAD --[DATA_RECORD[G]]

	commit_fast_load is
			-- commit
		do
			if insert_saved /= Void then
				insert_saved.commit
			end
		end

feature {NONE} -- Item Update

	process_new_data is
			-- apply the new data to the database
		do
			if submit_frequency /= 0 and then
				(((items_replaced \\ submit_frequency = 0) and items_replaced > 0)) then
				submit
			end
		end

feature -- Submit

	submit is
			-- write new data to the database
		local
			l_server: SQL_DATA_SERVER
			l_predicate: EXECUTION_PREDICATE
		do
			if items_replaced /= items_replaced_at_last_submit then
				if has_write_locks_table then
					-- This makes a new server (which is bad) but we can't use the auxilliary_server either since
					-- it gets used by the other code (below) during normal course of events.  Which is just
					-- another way of saying that the use of the auxilliary_server should be short and sweet.
					l_server := servers.new_server_from_repository (signature_item_server.logical)
					l_server.set_implicit_transaction_on
					create l_predicate.make (l_server, "update write_locks with (rowlock) set [timestamp]=getutcdate () where lock_name=?")
					if l_predicate.success then
						l_predicate.execute (Void,<<type_name>>)
					else
						raise ("Error creating update write_locks predicate")
					end
				end
				if not is_silent then
					logger.notice (data_name_prefix + type_name + " items submitting on " + signature_item_server.database)
				end
				commit_updated_cache
				delete_nulls
				insert_new_data
				delete_no_change
				if full_update_mode then
					delete_no_change_history
				end
				update_data
				clear_updated_fast_insert
				if not is_silent then
					logger.notice (
						data_name_prefix + type_name + " items replaced on " + signature_item_server.database
						+ ": " + (items_replaced - items_replaced_at_last_submit).out + " / " + items_replaced.out
									  )
				end
				items_replaced_at_last_submit := items_replaced
				truncate_temp_tables
				if has_write_locks_table then
					l_server.commit
				end
			end
		rescue
			clear_updated_fast_insert
			truncate_temp_tables
			signature_item_server.rollback
		end

feature -- Names

	data_name_prefix: STRING is
			-- the prefix to add to the data value tables
		do
			if data_name = Void or else data_name.count = 0 then
				Result := ""
			else
				Result := data_name + "_"
			end
		end

	data_value_table: STRING is
			-- the name of the table for the data values
		do
			Result := data_name_prefix + "data_" + type_name + "_value"
		end

	data_value_history_table: STRING is
			-- the name of the history table for the data values
		do
			Result := data_name_prefix + "data_" + type_name + "_value_history"
		end

	data_value_temp_table: STRING is
			-- the name of the temp table for staging the data values
		local
			l_random: WINDOWS_PSEUDO_RANDOM_NUMBER_GENERATOR
		do
			if data_value_temp_table_saved = Void then
				if playback_movie
				 then
					data_value_temp_table_saved ?= cam.next_frame ("data_value_temp_table").frame_data
					cam.compare_not_needed
				else
					create l_random.make
					data_value_temp_table_saved := "##" + l_random.next.abs.out + "_" + process_id.out + "_" + data_name_prefix + "__" + type_name
					debug ("make_movies")
						if record_movie
						 then
							cam.record_movie_frame ("data_value_temp_table", data_value_temp_table_saved)
						end
					end
				end
			end
			Result := data_value_temp_table_saved
		end

	data_value_temp_table_saved: STRING
			-- once per instance

	sid_item_index_name: STRING
			-- name of sid_item index
		do
			Result := data_name_prefix + type_name + "_sid_item"
		end

feature {NONE} -- Updated Data Management Predicates

	delete_nulls is
			-- delete the records from the temp table with NULL values
		local
			l_statement: STRING
		do
			l_statement := "delete from " + data_value_temp_table + "%N" +
				" where%N " +
				data_value_temp_table + ".value is NULL and%N " +
				data_value_temp_table + ".item_id not in%N "
			if is_path_clustered then
				l_statement.append (" (select item_id from%N " +
									data_value_table + " (ROWLOCK) where%N " +
									data_value_temp_table + ".signature_id = " +
									data_value_table + ".signature_id and " +
									data_value_temp_table + ".path = " +
									data_value_table + ".path)%N")
			else
				l_statement.append (" (select item_id from " +
									data_value_table + " (ROWLOCK) where " +
									data_value_temp_table + ".signature_id = " +
									data_value_table + ".signature_id) " +
									"OPTION (LOOP JOIN)")
			end
			temp_table_server.sql_execute_immediate (l_statement)
			if not temp_table_server.success then
				print ("delete nulls failed%N")
				raise ("delete nulls failed")
			end
		end

	delete_no_change is
			-- delete the records from the temp table that are the same as the target table
		local
			l_statement, l_target_table: STRING
		do
			l_target_table := data_value_table
			l_statement := "delete from " + data_value_temp_table + "%N" +
				" from " + data_value_temp_table + "%N" +
				" inner join " + l_target_table + " on " +
				l_target_table + ".item_id = " +
				data_value_temp_table + ".item_id and " +
				l_target_table + ".signature_id = " +
				data_value_temp_table + ".signature_id and "
			if not full_update_mode then
				l_statement.append ("(" + l_target_table + ".value = " +
									data_value_temp_table + ".value or (" +
									l_target_table + ".value is NULL and " +
									data_value_temp_table + ".value is NULL) or " +
									l_target_table + ".process_id = " +
									data_value_temp_table + ".process_id) ")
			else
				l_statement.append (l_target_table + ".process_id = " +
									data_value_temp_table + ".process_id")
			end
			if is_path_clustered then
				l_statement.append (" and " + l_target_table + ".path = " +
									data_value_temp_table + ".path")
			end
			temp_table_server.sql_execute_immediate (l_statement + ";")
			if not temp_table_server.success then
				print ("delete no changed failed%N")
				raise ("delete no changed failed")
			end
		end

	delete_no_change_history is
			-- delete the records from the temp table that are the same as the target history table
		require
			is_in_full_update_mode: full_update_mode
		local
			l_statement, l_target_table: STRING
		do
			l_target_table := data_value_history_table
			l_statement := "delete from " + data_value_temp_table + "%N" +
				" from " + data_value_temp_table + "%N" +
				" inner join " + l_target_table + " on " +
				l_target_table + ".item_id = " +
				data_value_temp_table + ".item_id and %N" +
				l_target_table + ".signature_id = %N" +
				data_value_temp_table + ".signature_id and %N"
			if not full_update_mode then
				l_statement.append ("(" + l_target_table + ".value = " +
									data_value_temp_table + ".value or (" +
									l_target_table + ".value is NULL and " +
									data_value_temp_table + ".value is NULL) or " +
									l_target_table + ".process_id = " +
									data_value_temp_table + ".process_id) ")
			else
				l_statement.append (l_target_table + ".process_id = " +
									data_value_temp_table + ".process_id")
			end
			if is_path_clustered then
				l_statement.append (" and " + l_target_table + ".path = " +
									data_value_temp_table + ".path")
			end
			temp_table_server.sql_execute_immediate (l_statement + ";")
			if not temp_table_server.success then
				print ("delete no changed history failed%N")
				raise ("delete no changed history failed")
			end
		end

	insert_new_data is
			-- insert the new data values to the target table
		local
			l_statement: STRING
		do
			l_statement :=  "insert into " + data_value_table + " with (ROWLOCK) (item_id,signature_id,value,process_id"
			if is_path_clustered then
				l_statement := l_statement + ", path"
			end
			l_statement := l_statement + ") select item_id, signature_id, value,process_id"
			if is_path_clustered then
				l_statement := l_statement + ", path "
			end
			l_statement := l_statement + " from " +
				data_value_temp_table +
				" where item_id not in " +
				" (select item_id from " +
				data_value_table + " (ROWLOCK) where " +
									data_value_temp_table + ".signature_id = " +
				data_value_table + ".signature_id) " +
				"OPTION (LOOP JOIN)"
			temp_table_server.sql_execute_immediate (l_statement)
			if not temp_table_server.success then
				raise ("insert new data failed")
			end
			-- The OPTION (LOOP JOIN) clause was added to force the SQL Server use the correct optimization.
			-- When the monthly_data_u_c.data_real_value table got to 250_000_000 rows it reoptimized the
			-- predicate causing it to take 5 minutes to complete.
		end

	update_data is
			-- apply the new data values to the target table
		local
			l_statement: STRING
		do
			l_statement := "update " + data_value_table +
				" set value = var1.value, process_id = var1.process_id " +
				" from " + data_value_temp_table + " as var1 where " +
				" var1.item_id = " + data_value_table + ".item_id and " +
				" var1.signature_id = " + data_value_table + ".signature_id "
			if is_path_clustered then
				l_statement.append (" and var1.path = " + data_value_table + ".path ")
			end
			if not full_update_mode then
				l_statement.append (" and (var1.value <> " + data_value_table + ".value or " +
									"var1.value is NULL or " +
									data_value_table + ".value is NULL) ")
			end
			temp_table_server.sql_execute_immediate (l_statement)
			if not temp_table_server.success then
				print ("update data failed%N")
				raise ("update data failed")
			end
		end

	trigger_source_code_date_a: ROSE_DATE is
			-- See comments in 'create_trigger'.
		once
			create Result.make (2002, 3, 9)
		end

	trigger_source_code_date_b: ROSE_DATE is
			-- See comments in 'create_trigger'.
		once
			create Result.make (2009, 8, 10)
		end

	uses_self_archive: BOOLEAN		-- Mostly for record_tagged at this time.

feature -- Triggers

	create_trigger is
			-- If this item_table's trigger creation date is older than 'trigger_source_code_date'
			-- then the trigger is old and needs to be rebuilt.
			-- The trigger has two parts:
			--		Move old data to the xxx_history table (depending on 'copy_to_history_statement').
			--		Update 'last_process_id' in the 'data_signatures' table.
			-- There is a similar feature in SIGNATURE_TAGGED_TRIGGER.
		local
			l_statement: STRING
			l_version: STRING
			l_trigger_creation_date: ROSE_DATE
			l_server: SQL_DATA_SERVER
		do
			l_trigger_creation_date := signature_item_server.trigger_creation_date(data_value_table, data_value_table + "_update")
			if signature_item_server.has_column_name(data_name_prefix + "data_signatures", "timestamp")
			then
				if l_trigger_creation_date=Void or else l_trigger_creation_date < trigger_source_code_date_b
				then
					l_version := "b"
				end
			else
				if l_trigger_creation_date=Void or else l_trigger_creation_date < trigger_source_code_date_a
				then
					l_version := "a"
				end
			end

			if l_version/=Void
			then
				--
				-- WARNING! WARNING! WARNING! WARNING! WARNING! WARNING! WARNING! WARNING! WARNING!
				--
				-- If you modify this trigger then you MUST change 'trigger_source_code_date'
				-- above or the change will NOT take effect.  Use a date that is slightly in
				-- the future.  Also be sure to change the version number in the trigger.
				--

				-- Note that version 3 and version 4a are identical.
				l_statement :=
					"-- Version 4" + l_version + ", created by SIGNATURE_TAGGED_ITEM on " + (create {ROSE_DATE}.make_now_utc).debug_output + " UT.%N" +
					"CREATE TRIGGER " + data_value_table + "_update ON " + data_value_table + "%N" +
					"FOR UPDATE,INSERT %N" +
					"AS %N" +
					"BEGIN %N" +
					"%TIF @@ROWCOUNT = 0 %N" +
					"%T%T  RETURN %N%N" +
					copy_to_history_statement

				if not an_override_item then
					l_statement := l_statement +
						"%N" +
						"%TIF (SELECT COUNT (*) FROM inserted) > 0 %N" +
						"%TBEGIN %N" +
						"%T%TUPDATE %N" +
						"%T%T%T" + data_name_prefix + "data_signatures%N" +
						"%T%T%Tset last_process_id = x.process_id %N" +
						"%T%TFROM %N" +
						"%T%T%T (SELECT signature_id, max (process_id) as process_id FROM inserted group by signature_id) x %N" +
						"%T%TWHERE%N" +
						"%T%T%T" + data_name_prefix + "data_signatures.signature_id = x.signature_id %N"
						if not equal(l_version, "b")
						then
							-- Here we don't bother updating the signatures if 'last_process_id' has already been set.
							-- However, the 'b' version must also update 'timestamp' so we skip this line in the 'b' version.
							l_statement.append("%T%T%T" + "and " + data_name_prefix + "data_signatures.last_process_id < x.process_id%N")
						end

						l_statement.append("%TEND%N")
				end
				--
				-- WARNING! WARNING! WARNING! WARNING! WARNING! WARNING! WARNING! WARNING! WARNING!
				-- See note above.
				--
				l_statement := l_statement + "END%N"

				if signature_item_server.has_trigger(data_value_table, data_value_table + "_update")
				then
					l_server := signature_item_server.auxiliary_server
					l_server.begin_transaction
					l_server.sql_execute_immediate ("DROP TRIGGER " + data_value_table + "_update")
					l_server.commit
				end

				l_server := signature_item_server.auxiliary_server
				l_server.begin_transaction
				l_server.sql_execute_immediate (l_statement)
				if not l_server.success then
					raise ("Could not create the trigger.")
				end

				l_server.commit
			end
		end

	copy_to_history_statement: STRING is
			-- SQL code to be used when making the trigger.
		do
			Result := "%TINSERT " + data_value_history_table + " SELECT * FROM deleted%N"
		end

	an_override_item: BOOLEAN is
			-- add an override item
		do
			if data_name_prefix.count > 0 then
				Result := data_name_prefix.substring_index ("override", 1) /= 0
			end
		end

feature -- Lock management

	set_has_write_locks_table is
			-- determine if there is a write locks table
		local
			l_predicate: ANONYMOUS_PREDICATE
		do
			if not has_write_locks_table_checked then
				create l_predicate.make_using_server (signature_item_server, "exec sp_tables 'write_locks'")
				l_predicate.retrieve (Void, Void)
				if not l_predicate.success then
					raise ("sp_tables failed.")
				else
					has_write_locks_table := (l_predicate.array.count > 0)
					has_write_locks_table_checked := True
				end
			end
		end

	has_write_locks_table: BOOLEAN
			-- is there a write locks table?

	has_write_locks_table_checked: BOOLEAN
			-- have we checked to see if there is a write locks table?

feature {NONE} -- Implementation

	update_cache_key (a_item: DATA_RECORD [G]): NATURAL_64
			-- Create data cache key from `a_item'
		require
			a_item_attached: a_item /= Void
		do
			Result := a_item.signature_id.to_natural_64.bit_or (a_item.item_id.to_natural_64.bit_shift_left (32))
		end

invariant

	available_update_cache: update_cache /= Void

end
