note

	description:

		"Object providing a query and update interface for signature tagged item databases"

	author: "Mauricio Bustos"
	last_editor: "$Author: rj $"
	date: "$Date: 2014/09/11 $"
	revision: "$Revision: #92 $"
	path: "$File: //rose/source/kernel/persistence/databases/signature_tagged/signature_tagged_item_database.e $"

class
	SIGNATURE_TAGGED_ITEM_DATABASE[reference G -> ROSE_SIGNATURE_RECORD[ENTITY] create default_create end]

inherit
	ANY
	ROSE_LEGACY_UNDEFINED
		export
			{NONE} all
		end
	ERROR_LOG_USE
		export
			{NONE} all
		end
	IEEE_MATH
		export
			{NONE} all
		end
	SIGNATURE_TAGGED_NAMES
	PM_METRIC_SOURCE
		export
			{NONE} all
		redefine
			metrics,
			reset_metrics
		end
	EXCEPTIONS
		rename
			class_name as exceptions_class_name
		export
			{NONE} all
		end
	ROSE_SHARED_GLOBAL
		rename
			error as server_error
		export
			{NONE} all
		select
			server_error
		end
	SIGNATURE_TAGGED_OUTPUT_LEVEL
	ROSE_SHARED_LOGGER
		rename
			error as logger_error
		end
create

	make,
	make_named,
	make_named_with_process_manager,
	make_with_process_manager

feature {NONE} -- Creation

	make (a_process_name: STRING; a_server: SQL_DATA_SERVER) is
			-- Create the database
		require
			server_not_void: a_server /= Void
		do
			make_named (Void,a_process_name,a_server)
		end

	make_named (a_data_name, a_process_name: STRING; a_server: SQL_DATA_SERVER) is
			-- Create the database named 'a_data_name'
			-- The data name is a prefix in the underlying database implementation
		require
			server_not_void: a_server /= Void
		local
			l_table_structure: SIGNATURE_TAGGED_TABLE_STRUCTURE
		do
			max_cube_result_size := 20_000
			data_name := a_data_name
			a_server.begin_transaction
			create l_table_structure.make (a_server, a_data_name, record_templates)
			a_server.commit
			if l_table_structure.data_available then
				create signature_tagged_server.make (l_table_structure)
				signature_tagged_server.set_signature_item_server (a_server)
				initialize_manager (a_process_name, l_table_structure)
			else
				if data_name_prefix.is_empty then
					raise ("STID table structure not available on " + a_server.node + "." + a_server.database)
				else
					raise ("STID table structure for " + data_name_prefix + " not available on " + a_server.node + "." + a_server.database)
				end
			end
		end

	make_with_process_manager (a_server: SQL_DATA_SERVER; a_process_manager: SUPPLIER_PROCESS_ID_MANAGER) is
			-- Create the database providing an existing supplier process id manager
		require
			server_not_void: a_server /= Void
			process_manager_not_void: a_process_manager /= Void
		do
			make_named_with_process_manager (Void,a_server,a_process_manager)
		end

	make_named_with_process_manager (a_data_name: STRING; a_server: SQL_DATA_SERVER; a_process_id_manager: SUPPLIER_PROCESS_ID_MANAGER) is
			-- Create the database named 'a_data_name' providing an existing supplier process id manager
			-- The data name is a prefix in the underlying database implementation
		require
			server_not_void: a_server /= Void
			process_manager_not_void: a_process_id_manager /= Void
		do
			make_named (a_data_name, a_process_id_manager.name, a_server)
			signature_tagged_server.set_process_id_manager (a_process_id_manager)
		end

feature	-- Basic operations

	update_end_time is
			-- Update the end_time of the currently available process_id if process_id_manager_created is True
		do
			signature_tagged_server.update_end_time
		end

	start_process (a_process_name: STRING; a_file_date, a_market_date, a_rose_date: ROSE_DATE) is
			-- Explicitly start 'a_process_name' on the supplier process id manager.
			-- A supplier process id manager will be created if one does not already exist.
		require
			meaningful_name: a_process_name /= Void and then not a_process_name.is_empty
		do
			signature_tagged_server.start_process (a_process_name, a_file_date, a_market_date, a_rose_date)
		end

	terminate_process_id is
			-- Terminate the currently active process in the supplier process id manager
		do
			signature_tagged_server.terminate_process_id
		end

	refresh_process_id_map is
			-- Reread the process id to date map from the database
		do
			signature_tagged_server.refresh_process_id_map
		end

	reset_process_id_lookup_table is
			-- Set the process_id_lookup_table to Void.  This will result in a refresh at the next query
		do
			signature_tagged_server.reset_process_id_lookup_table
		end

	purge is
			-- Wipe out any data in the local cache
		do
			cube_signature_result.clear_all
			cube_position_result.wipe_out
			cube_result_table.wipe_out
			if cube_result /= Void then
				cube_result.wipe_out
			end
			signature_tagged_server.remake_virtual_record_set
			manager_cache.clear_all;
			(create {MEMORY}).full_collect
		end

	clear_last_rose_signature is
			-- Set `last_rose_signature' to Void
		do
			last_rose_signature := Void
		end

	reset_metrics is
			-- Reset the processing metrics
		do
			signature_tagged_server.reset_metrics
		end

	drop_trigger is
			-- Drop the signature version update trigger
		do
			signature_tagged_server.drop_trigger (data_name)
		end

	close_sig_tagged_trace is
			-- Close the log that tracks query and result history
		do
			signature_tagged_server.close_sig_tagged_trace
		end

	synchronize_process_dates (a_signature: G) is
			-- Ensure that the process_id_map is update to date with respect to 'a_signature'
		require
			valid_signature: a_signature /= Void
		do
			signature_tagged_server.table_structure.synchronize_process_dates (a_signature)
		end

feature -- Mode setting

	enable_testenv_overlay is
			-- Overlay data from the test environment onto the data from the server specified at creation
		do
			raise ("Not Implemented")
		end

	disable_testenv_overlay is
			-- Disable the data overlay mechanism
		do
			raise ("Not Implemented")
		end

	show_suppressed_historical_data is
			-- Retrieve suppressed historical data that has been affected by pinning
			-- Use this to override the default behavior of pinned back record retrieval
		do
			signature_tagged_server.show_suppressed_historical_data
		end

	hide_suppressed_historical_data is
			-- Don't retrieve suppressed historical data that has been affected by pinning
			-- This is the default behavior
		do
			signature_tagged_server.hide_suppressed_historical_data
		end

	enable_version_pinning is
			-- Turn on version pinning for result sets
		do
			signature_tagged_server.enable_version_pinning
		end

	disable_version_pinning is
			-- Turn off version pinning for result sets
		do
			signature_tagged_server.disable_version_pinning
		end

	disable_data_type (a_type: STRING; a_is_override: BOOLEAN) is
			-- Turn off read / write for 'a_type' data, specify whether to do it for overrides or not
		require
			valid_type: a_type /= Void
		do
			signature_tagged_server.disable_data_type (a_type, a_is_override)
		end

	disable_regular_data_type (a_type: STRING) is
			-- Turn off read / write for 'a_type' data
		require
			valid_type: a_type /= Void
		do
			signature_tagged_server.disable_regular_data_type (a_type)
		end

	disable_override_data_type (a_type: STRING) is
			-- Turn off read / write for 'a_type' override data
		require
			valid_type: a_type /= Void
		do
			signature_tagged_server.disable_override_data_type (a_type)
		end

	disable_all_regular_data_types is
			-- Turn off read / write for all data types, this results in only signature data access
		do
			signature_tagged_server.disable_all_regular_data_types
		end

	disable_all_override_data_types is
			-- Turn off read / write for all overrides for all data types, this results in only signature data access
		do
			signature_tagged_server.disable_all_override_data_types
		end

	disable_all_data_types is
			-- Turn off read / write for all data (overrides and regular), this results in only signature data access
		do
			signature_tagged_server.disable_all_data_types
		end

	enable_data_type (a_type: STRING; a_is_override: BOOLEAN) is
			-- Turn on read / write for 'a_type' data, specify whether to do it for overrides or not
		require
			valid_type: a_type /= Void
		do
			signature_tagged_server.enable_data_type (a_type, a_is_override)
		end

	enable_regular_data_type (a_type: STRING) is
			-- Turn on read / write for 'a_type' data
		require
			valid_type: a_type /= Void
		do
			signature_tagged_server.enable_regular_data_type (a_type)
		end

	enable_override_data_type (a_type: STRING) is
			-- Turn on read / write for 'a_type' override data
		require
			valid_type: a_type /= Void
		do
			signature_tagged_server.enable_override_data_type (a_type)
		end

	enable_all_regular_data_types is
			-- Turn on read / write for all regular data types
		do
			signature_tagged_server.enable_all_regular_data_types
		end

	enable_all_override_data_types is
			-- Turn on read / write for all overrides for all data types
		do
			signature_tagged_server.enable_all_override_data_types
		end

	enable_all_data_types is
			-- Turn on read / write for all data types and all overrides
		do
			signature_tagged_server.enable_all_data_types
		end

	build_cube_position_result is
			-- Have queries build the cube_position_result object
		do
			signature_tagged_server.build_cube_position_result
		end

	do_not_build_cube_position_result is
			-- Don't have queries build the cube_position_result object
		do
			signature_tagged_server.do_not_build_cube_position_result
		end

feature -- Database commit operations

	cleanup is
			-- Ask the data item objects to clean themselves up and then terminate the currently active process
		do
			cleanup_items
			terminate_process_id
		end

	cleanup_and_commit is
			-- Ask the data item objects to clean themselves up and then commit the server
		obsolete "Use cleanup and then commit the database youself."
		do
			signature_tagged_server.cleanup_and_commit
		end

	submit_and_commit is
			-- Submit the local update caches to the database and commit the server
		obsolete "Use submit and then commit the database youself."
		do
			submit
			commit
		end

	frozen submit is
			-- Submit the local update caches to the database
		do
			if signature_tagged_server.client_can_write then
				signature_tagged_server.submit (signature_touches)
				last_rose_signature := Void
			else
				raise ("This client cannot write to " + signature_tagged_server.data_server_full_name)
			end
		end

	frozen discard_all_updates is
			-- Discard the staged update data
		do
			signature_touches.wipe_out
			signature_tagged_server.discard_all_updates
		end

	commit is
			-- Commit the server specified at creation
		obsolete "Commit the database yourself."
		do
			signature_tagged_server.commit
		end

	begin_transaction is
			-- Begin a transaction on the server used by this database object
		do
			signature_tagged_server.begin_transaction
		end

	data_server_commit is
			-- Commit the server used by this database object
		do
			signature_tagged_server.data_server_commit
		end

	set_committed is
			-- Set the committed status for the server
		do
			signature_tagged_server.set_committed
		end

	set_uncommitted is
			-- Set the committed status for the server to specify that there is uncommitted data
		do
			signature_tagged_server.set_uncommitted
		end

	post_rollback_cleanup is
			-- Call this ONLY when debugging and you ALWAYS call 'rollback' instead of 'commit'.
			-- Call this AFTER the call to 'rollback'.  This call is only for debugging and can not
			-- make any guarantees if a 'commit' has ever been called.
			-- Purpose:
			-- To reset anything that needs to be reset when you ALWAYS call 'rollback'.
		do
			signature_tagged_server.post_rollback_cleanup
		end

	array_for_query (a_signature: G): ROSE_LINEAR_ARRAY [G] is
			-- A well typed array for use in later queries.
			-- Use it to avoid cat-call warnings at run-time.
		require
			valid_signature: a_signature /= Void
		do
			Result := <<a_signature>>
		ensure
			valid_result: Result /= Void
		end

feature -- Access

	process_name: STRING is
			-- The currently active process name
		do
			Result := signature_tagged_server.process_name
		end

	table_structure: SIGNATURE_TAGGED_TABLE_STRUCTURE is
			-- Access to the underlying table structure
			-- To be used for meta data analysis
		do
			Result := signature_tagged_server.table_structure
		end

	virtual_record_set: SIGNATURE_TAGGED_VIRTUAL_RECORD_SET [G] is
			-- Access to the virtual_record_set in the server
		do
			Result := signature_tagged_server.virtual_record_set
		end

	metrics: LINEAR [PAIR[STRING, DOUBLE]] is
			-- Processing related metrics for this database (ie. signature created, items updated...)
		do
			Result := signature_tagged_server.metrics
		end

	last_result: ROSE_LINEAR_ARRAY[ROSE_VIRTUAL_RECORD[G]] is
			-- The cube_result of the last_rose_signature
		do
			if last_rose_signature /= Void then
				set_results_from_cache (last_rose_signature.signature)
				Result := cube_result
			end
		end

	last_record_result: ROSE_VIRTUAL_RECORD[G] is
			-- The last virtual record of the last result set
		do
			Result := last_rose_signature
		end

	new_signature_template: G is
			-- A new instance of the signature
			-- To be used for meta data analysis
		do
			create Result
		end

	data_server_name: STRING is
			-- The full name of the server being used by this database object
		do
			Result := signature_tagged_server.data_server_full_name
		end

	data_server_node: STRING is
			-- The node (commonly a machine) of the server being used in this database object
		do
			Result := signature_tagged_server.data_server_node
		end

	data_server_database: STRING is
			-- The database of the server being used in this database object
		do
			Result := signature_tagged_server.database_name
		end

	data_server_logical_name: STRING is
			-- The logical name of the server being used by this database object
		do
			Result := signature_tagged_server.data_server_logical_name
		end

	override_family_names: ROSE_LINEAR_ARRAY [STRING] is
			-- The list of available override family names
		do
			Result := signature_tagged_server.override_family_names
		end

	override_family_name: STRING is
			-- The currently selected override family name
		do
			Result := signature_tagged_server.override_family_name
		end

	process_id_manager: SUPPLIER_PROCESS_ID_MANAGER is
			-- The currently active process id manager, this could have been created by Current or have been externally set
		do
			Result := signature_tagged_server.process_id_manager
		end

	process_id: INTEGER is
			-- The currently available process_id
		do
			Result := signature_tagged_server.process_id
		end

	process_id_map: HASH_TABLE [ROSE_DATE, INTEGER] is
			-- A table mapping process_ids to dates
		do
			Result := signature_tagged_server.process_id_map
		end

	ignore_first_undefined: BOOLEAN
			-- Should we not write out a record if the first value for an item is undefined?

	has_write_permission: BOOLEAN is
			-- Does current user have write permission on `Current'?
		do
			Result := signature_tagged_server.has_write_permission
		end

	has_read_permission: BOOLEAN is
			-- Does current user have read permission on `Current'?
		do
			Result := signature_tagged_server.has_read_permission
		end

feature -- Historical Retrieval

	retrieve_historical_signatures (a_signatures: ROSE_LINEAR_ARRAY[G];
									a_statement: STRING): ROSE_LINEAR_ARRAY[G] is
			-- Array of signatures and their historical versions based on 'a_signatures' and 'a_statement'
		require
			signatures_not_void: a_signatures /= Void
		do
			signature_tagged_server.prepare_to_retrieve_signatures
			signature_tagged_server.set_new_query
			select_signature_ids (cube_from_signatures (a_signatures,columns_from_restriction_statement (a_statement)))
			Result := signature_tagged_server.retrieve_historical_signatures
			last_rose_signature := signature_tagged_server.last_rose_signature
			cube_signature_result := signature_tagged_server.cube_signature_result
			cube_position_result := signature_tagged_server.cube_position_result
			set_cache_from_results (last_rose_signature)	-- Since this is no longer done by set_current_cache_pool.
		end

	retrieve_current_signatures (a_signatures: ROSE_LINEAR_ARRAY[G];
								 a_statement: STRING): ROSE_LINEAR_ARRAY[G] is
			-- The as of dates of current signatures are Void under the current convention.
			-- When these signatures are used as target in Rose expressions, the expressions will retrieve the most current version of the item.
		require
			signatures_not_void: a_signatures /= Void
		do
			signature_tagged_server.prepare_to_retrieve_signatures
			signature_tagged_server.set_new_query
			select_signature_ids (cube_from_signatures (a_signatures,columns_from_restriction_statement (a_statement)))
			Result := signature_tagged_server.last_cube_signature_result
			cube_signature_result := Result
			last_rose_signature := signature_tagged_server.last_rose_signature
			cube_position_result := signature_tagged_server.cube_position_result
			create cube_result_table.make (0)
		end

	retrieve_historical_items (a_signatures: ROSE_LINEAR_ARRAY[G];
							   a_statement: STRING;
							   a_items: ROSE_LINEAR_ARRAY[INTEGER]): ROSE_LINEAR_ARRAY[ROSE_VIRTUAL_RECORD[G]] is
			-- Array of virtual records and their historical versions based on 'a_signatures' and 'a_statement'
			-- Queries only for item_ids in 'a_items' (all items if 'a_items' = Void)
		require
			signatures_not_void: a_signatures /= Void
		do
			signature_tagged_server.set_historical_mode
			Result := retrieve_signatures (a_signatures,a_statement,a_items)
			signature_tagged_server.clear_historical_mode
		end

	retrieve_signature_data (a_signature: G): G is
			-- retrieve signature data
		require
			signature_not_void: a_signature /= Void
		do
			Result := signature_tagged_server.retrieve_signature_data (a_signature)
		end

	retrieve_signature_data_history (a_signature: G): ROSE_LINEAR_ARRAY[G] is
			-- Array of historical signature data versions for 'a_signature'
		require
			signature_not_void: a_signature /= Void
		local
			l_signature_id: INTEGER
		do
			if a_signature.signature_id <= 0 then
				l_signature_id := signature_tagged_server.signature_id_if_found (a_signature)
				a_signature.set_signature_id (l_signature_id)
			end

			if a_signature.signature_id > 0 then
				Result := signature_tagged_server.retrieve_signature_data_history (a_signature)
				Result.do_all (Agent set_signature_as_of_date)
			end
		end

	retrieve_signature_data_histories (a_signatures: ARRAY[G]): ROSE_LINEAR_ARRAY[G] is
			-- Array of historical signature data versions for 'a_signatures'
		require
			signatures_not_void: a_signatures /= Void
		local
			i: INTEGER
			l_signatures: ROSE_LINEAR_ARRAY[G]
		do
			create Result.make_empty
			if a_signatures /= Void then
				from i := a_signatures.lower until i > a_signatures.upper loop
					l_signatures := retrieve_signature_data_history (a_signatures.item (i))
					if l_signatures /= Void then
						Result.append_array (l_signatures)
					end
					i := i + 1
				end
			end
		end

	retrieve_records_by_signatures (a_signatures: ROSE_LINEAR_ARRAY[G]; a_historical_mode: BOOLEAN
									a_items: ROSE_LINEAR_ARRAY[INTEGER]): ROSE_LINEAR_ARRAY[ROSE_VIRTUAL_RECORD[G]] is
			-- Array of virtual records for 'a_signatures'
			-- Queries only for item_ids in 'a_items' (all items if 'a_items' = Void)
		require
			signatures_not_void: a_signatures /= Void
		do
			if a_historical_mode then
				signature_tagged_server.set_historical_mode
			end
			Result := ordered_cube_of_items_by_signatures (a_signatures, a_items)
			if a_historical_mode then
				signature_tagged_server.clear_historical_mode
			end
		end

	historical_records_by_signature (a_signature: G; a_restriction_statement: STRING; a_items: ROSE_LINEAR_ARRAY [INTEGER]): ROSE_LINEAR_ARRAY[ROSE_VIRTUAL_RECORD[G]] is
			-- The historical virtual records for 'a_signature'
		require
			signature_not_void: a_signature /= Void
		do
			raise ("Not Implemented")
		end

	historical_record_by_signature (a_signature: G): ROSE_VIRTUAL_RECORD[G] is
			-- The virtual record version for 'a_signature'
			-- The exact as_of_date is used to find the desired version so it should be set using the supplier process id manager
		require
			signature_not_void: a_signature /= Void
		do
			raise ("Not Implemented")
		end

feature -- Current Retrieval

	retrieve_signature (a_signature: G;
						a_statement: STRING;
						a_items: ROSE_LINEAR_ARRAY[INTEGER]): ROSE_LINEAR_ARRAY[ROSE_VIRTUAL_RECORD[G]] is
			-- Array of current virtual records based on 'a_signature' and 'a_statement'
			-- Queries only for item_ids in 'a_items' (all items if 'a_items' = Void)
		require
			signature_not_void: a_signature /= Void
		do
			Result := retrieve_signature_with_limit (a_signature,a_statement,a_items,0)
		end

	retrieve_signature_with_limit (a_signature: G;
								   a_statement: STRING;
								   a_items: ROSE_LINEAR_ARRAY[INTEGER];
								   a_record_limit: INTEGER): ROSE_LINEAR_ARRAY[ROSE_VIRTUAL_RECORD[G]] is
			-- Array of current virtual records based on 'a_signature' and 'a_statement'
			-- Queries only for item_ids in 'a_items' (all items if 'a_items' = Void)
			-- 'a_record_limit' specifies how many records should be retrieved when there is an inequality in the statement (usually 1)
		require
			signature_not_void: a_signature /= Void
			meaningful_record_limit: a_record_limit >= 0
		local
			l_array: ROSE_LINEAR_ARRAY[G]
		do
			create l_array.make_empty
			l_array.extend (a_signature)
			Result := retrieve_with_restriction (l_array,a_statement,a_items,a_record_limit)
		end

	retrieve_signatures (a_signatures: ROSE_LINEAR_ARRAY[G];
						 a_statement: STRING;
						 a_items: ROSE_LINEAR_ARRAY[INTEGER]): ROSE_LINEAR_ARRAY[ROSE_VIRTUAL_RECORD[G]] is
			-- Array of current virtual records based on 'a_signatures' and 'a_statement'
			-- Queries only for item_ids in 'a_items' (all items if 'a_items' = Void)
		require
			signatures_not_void: a_signatures /= Void
		do
			Result := retrieve_with_restriction (a_signatures,a_statement,a_items,0)
		end

	retrieve_with_restriction (a_signatures: ROSE_LINEAR_ARRAY[G];
							   a_statement: STRING;
							   a_items: ROSE_LINEAR_ARRAY[INTEGER];
							   a_record_limit: INTEGER): ROSE_LINEAR_ARRAY[ROSE_VIRTUAL_RECORD[G]] is
			-- Example statement:
			--	 #yyyymmdd_available.id > yyyymmdd_available and #entity_id.id = entity_id
			--	 You need the '#' and the '.id' to get the latest yyyymmdd_available before your #yyyymmdd_available
		require
			signatures_not_void: a_signatures /= Void
			meaningful_record_limit: a_record_limit >= 0
		do
			if a_record_limit /= 0 then
				signature_tagged_server.set_record_limit (a_record_limit)
			end
			if a_statement /= Void then
				signature_tagged_server.set_restriction_clause (a_statement)
			end
			Result := ordered_cube_of_items (cube_from_signatures (a_signatures,columns_from_restriction_statement (a_statement)),a_items)
			-- Reset for next query
			if a_statement /= Void then
				signature_tagged_server.set_restriction_clause (Void)
			end
			if a_record_limit /= 0 then
				signature_tagged_server.set_record_limit (0)
			end
			signature_tagged_server.clear_column_comparison
		end

	ordered_cube_of_items_from_predicate (a_predicate: ANONYMOUS_PREDICATE; a_argument: ANY; a_arguments: ARRAY[ANY];
										  a_items: ROSE_LINEAR_ARRAY[INTEGER]): ROSE_LINEAR_ARRAY[ROSE_VIRTUAL_RECORD[G]] is
			-- Array of current virtual records based on the signatures returned by 'a_predicate' using the arguments 'a_argument' and 'a_arguments'
			-- Queries only for item_ids in 'a_items' (all items if 'a_items' = Void)
		require
			predicate_not_void: a_predicate /= Void
		do
			signature_tagged_server.set_new_query
			Result := signature_tagged_server.ordered_cube_of_items_from_predicate (a_predicate, a_argument, a_arguments, a_items)
			if Result.is_empty then
				last_rose_signature := Void
			else
				last_rose_signature := Result.last
			end
			set_cache_from_results (last_rose_signature)
		end

	ordered_cube_of_items_by_signatures (a_signatures: ROSE_LINEAR_ARRAY[G];
										 a_items: ROSE_LINEAR_ARRAY[INTEGER]): ROSE_LINEAR_ARRAY[ROSE_VIRTUAL_RECORD[G]] is
			-- Array of current virtual records for 'a_signatures'
			-- Queries only for item_ids in 'a_items' (all items if 'a_items' = Void)
		do
			signature_tagged_server.set_new_query
			Result := signature_tagged_server.ordered_cube_of_items_by_signatures (a_signatures, a_items)
			if Result.is_empty then
				last_rose_signature := Void
			else
				last_rose_signature := Result.last
			end
			signature_tagged_server.set_virtual_record_set_cube_result_temp (Result)
			set_cache_from_results (last_rose_signature)
			signature_tagged_server.set_virtual_record_set_cube_result_temp (Void)
		end

	record_by_signature (a_signature: G): ROSE_VIRTUAL_RECORD[G] is
			-- The record for 'a_signature', set the local cache from the broader cache first
		require
			signature_not_void: a_signature /= Void
		do
			Result := current_cache_record_by_signature(a_signature)
			if Result=Void
			then
				set_results_from_cache (a_signature)
				Result := current_cache_record_by_signature(a_signature)
			end
		end

	current_cache_record_by_signature (a_signature: G): ROSE_VIRTUAL_RECORD[G] is
			-- The record for 'a_signature', only look in the current local cache
		require
			signature_not_void: a_signature /= Void
		do
			if cube_result_table /= Void then
				cube_result_table.search (a_signature.multi_key)
				if cube_result_table.found then
					Result := cube_result_table.found_item
				end
			end
		end

	retrieve_signature_versions (a_signatures: ROSE_LINEAR_ARRAY [G]; a_template: ROSE_VIRTUAL_RECORD [G]): ROSE_LINEAR_ARRAY [G] is
			-- This routine retrieve signature versions without changing the existing settings and caches
		do
			Result := signature_tagged_server.virtual_record_set.retrieve_signature_versions (a_signatures, a_template, True)
		end

	retrieve_virtual_records (a_historical_mode: BOOLEAN; a_signatures: ROSE_LINEAR_ARRAY [G]; a_items: ROSE_LINEAR_ARRAY [INTEGER];
							  a_template: ROSE_VIRTUAL_RECORD [G]): ROSE_LINEAR_ARRAY [ROSE_VIRTUAL_RECORD [G]] is
			-- This routine retrieve virtual records without changing the existing settings and caches
		do
			Result := signature_tagged_server.virtual_record_set.retrieve_virtual_records (a_historical_mode, a_signatures, a_items, a_template, True)
		end

	signatures_with_overrides: ROSE_LINEAR_ARRAY [G] is
			-- Signatures with currently enabled overrides
		do
			Result := signature_tagged_server.virtual_record_set.signatures_with_overrides
		ensure
			non_void_result: Result /= Void
		end

feature -- Insert

	insert_records_nonlogged (a_array: ROSE_LINEAR_ARRAY[ROSE_VIRTUAL_RECORD[G]]) is
			-- This is a batch operation
			-- All signatures and data sent here will be inserted and commited
			-- Be sure that it is new or the insert will fail
		do
			signature_tagged_server.insert_records_nonlogged (a_array)
		end

	insert_signatures_nonlogged (a_array: ROSE_LINEAR_ARRAY[G]) is
			-- This is a batch operation
			-- All signatures sent here will be inserted and commited
			-- Be sure that it is new or the insert will fail
		do
			signature_tagged_server.insert_signatures_nonlogged (a_array)
		end

feature -- Replace

	replace_virtual_record (a_record: ROSE_VIRTUAL_RECORD[G]) is
			-- Replace all data items for 'a_record' if a change has been detected or full_update_mode is enabled
		do
			if a_record.has_integer_items then
				from a_record.integer_items.start until a_record.integer_items.off loop
					replace_integer_item (a_record.integer_items.item_for_iteration,
										  a_record.integer_items.key_for_iteration,
										  a_record.signature)
					a_record.integer_items.forth
				end
			end
			if a_record.has_real_items then
				from a_record.real_items.start until a_record.real_items.off loop
					replace_real_item (a_record.real_items.item_for_iteration,
									   a_record.real_items.key_for_iteration,
									   a_record.signature)
					a_record.real_items.forth
				end
			end
			if a_record.has_double_items then
				from a_record.double_items.start until a_record.double_items.off loop
					replace_double_item (a_record.double_items.item_for_iteration,
										 a_record.double_items.key_for_iteration,
										 a_record.signature)
					a_record.double_items.forth
				end
			end
			if a_record.has_string_items then
				from a_record.string_items.start until a_record.string_items.off loop
					replace_string_item (a_record.string_items.item_for_iteration,
										 a_record.string_items.key_for_iteration,
										 a_record.signature)
					a_record.string_items.forth
				end
			end
			if a_record.has_image_items then
				from a_record.image_items.start until a_record.image_items.off loop
					replace_image_item (a_record.image_items.item_for_iteration,
										a_record.image_items.key_for_iteration,
										a_record.signature)
					a_record.image_items.forth
				end
			end
			if a_record.has_date_items then
				from a_record.date_items.start until a_record.date_items.off loop
					replace_date_item (a_record.date_items.item_for_iteration,
									   a_record.date_items.key_for_iteration,
									   a_record.signature)
					a_record.date_items.forth
				end
			end
		end

	replace_real_item (a_value: REAL; a_item_id: INTEGER; a_signature: G) is
			-- Replace the real value identified by 'a_item_id' in 'a_signature' by 'a_value'
			-- if it is different that the current value or full_update_mode is enabled
		require
			valid_item: a_item_id > 0
			signature_not_void: a_signature /= Void
		local
			l_virtual_record: ROSE_VIRTUAL_RECORD[G]
		do
			cube_result_table.search (a_signature.multi_key)
			if cube_result_table.found then
				l_virtual_record := cube_result_table.found_item
				signature_tagged_server.create_trigger (data_name)
			else
				l_virtual_record := record_by_signature (a_signature)
				if l_virtual_record = Void then
					l_virtual_record := virtual_record (a_signature, True, not ignore_first_undefined)	-- Calls 'create_trigger'.
				else
					signature_tagged_server.create_trigger (data_name)
				end
			end
			replace_real_item_impl (a_value, a_item_id, l_virtual_record, a_signature)
		end

	replace_real_item_direct (a_value: REAL; a_item_id: INTEGER; a_signature: G) is
			-- Replace the real value identified by 'a_item_id' in 'a_signature' by 'a_value'.
			-- Don't create a record if it doesn't already exist.
			-- WARNING!  Use of this feature is discouraged.  It requires intimate knowledge of the inner workings of this
			-- class.  Apparently it is expected that you have already loaded the appropriate signature into the cache.
		require
			valid_item: a_item_id > 0
			signature_not_void: a_signature /= Void
		local
			l_virtual_record: ROSE_VIRTUAL_RECORD[G]
		do
			cube_result_table.search (a_signature.multi_key)
			if cube_result_table.found then
				l_virtual_record := cube_result_table.found_item
				signature_tagged_server.create_trigger (data_name)
			else
				l_virtual_record := record_by_signature (a_signature)
				if l_virtual_record /= Void then
					signature_tagged_server.create_trigger (data_name)
				end
			end
			replace_real_item_impl (a_value, a_item_id, l_virtual_record, a_signature)
		end

	replace_real_item_impl (a_value: REAL; a_item_id: INTEGER; a_virtual_record: ROSE_VIRTUAL_RECORD [G]; a_signature: G) is
			-- Implementation for the real data update
		local
			l_real_items: HASH_TABLE[REAL,INTEGER]
			l_real: REAL
			l_virtual_record: ROSE_VIRTUAL_RECORD[G]
		do
			if a_virtual_record /= Void then
				l_real_items := a_virtual_record.real_items
				l_real_items.search (a_item_id)
				if l_real_items.found then
					if signature_tagged_server.full_update_mode or ieee_is_ne (l_real_items.found_item, a_value) then
						if not is_defined_real (a_value) then
							signature_tagged_server.real_data_replace (new_real_record (Qnan_real_32,a_item_id,a_virtual_record))
						else
							signature_tagged_server.real_data_replace (new_real_record (a_value,a_item_id,a_virtual_record))
						end
						if real_item_write_monitor_agent /= Void then
							real_item_write_monitor_agent.call ([a_item_id, a_signature, a_value, l_real_items.found_item])
						end
						l_real_items.force (a_value,a_item_id)
					end
				elseif (not ignore_first_undefined or is_defined_real (a_value)) then
					if a_virtual_record.signature.signature_id = 0
					then
						l_virtual_record := virtual_record(a_signature, False, True)
					else
						l_virtual_record := a_virtual_record
					end

					if real_item_write_monitor_agent /= Void then
						l_real := Qnan_real_32
						real_item_write_monitor_agent.call ([a_item_id, a_signature, a_value, l_real])
					end
					if a_value = Undefined_real then
						signature_tagged_server.real_data_replace (new_real_record (Qnan_real_32, a_item_id, l_virtual_record))
						l_real_items.force (Qnan_real_32,a_item_id)
					else
						signature_tagged_server.real_data_replace (new_real_record (a_value, a_item_id, l_virtual_record))
						l_real_items.force (a_value,a_item_id)
					end
				end
			end
		end

	replace_double_item (a_value: DOUBLE; a_item_id: INTEGER; a_signature: G) is
			-- Replace the double value identified by 'a_item_id' in 'a_signature' by 'a_value'
			-- if it is different that the current value or full_update_mode is enabled
		require
			valid_item: a_item_id > 0
			signature_not_void: a_signature /= Void
		local
			l_virtual_record: ROSE_VIRTUAL_RECORD[G]
		do
			l_virtual_record := record_by_signature (a_signature)
			if l_virtual_record = Void then
				l_virtual_record := virtual_record (a_signature, True, not ignore_first_undefined)	-- Calls 'create_trigger'.
			else
				signature_tagged_server.create_trigger (data_name)
			end

			l_virtual_record.double_items.search (a_item_id)
			if l_virtual_record.double_items.found then
				if signature_tagged_server.full_update_mode or ieee_is_ne (l_virtual_record.double_items.found_item, a_value) then
					signature_tagged_server.double_data_replace (new_double_record (a_value,a_item_id,l_virtual_record))
					if double_item_write_monitor_agent /= Void then
						double_item_write_monitor_agent.call ([a_item_id, a_signature, a_value, l_virtual_record.double_items.found_item])
					end
					l_virtual_record.double_items.force (a_value,a_item_id)
				end
			elseif (not ignore_first_undefined or is_defined_double (a_value)) then
				if l_virtual_record.signature.signature_id = 0
				then
					l_virtual_record := virtual_record(a_signature, False, True)
				end

				signature_tagged_server.double_data_replace (new_double_record (a_value,a_item_id,l_virtual_record))
				if double_item_write_monitor_agent /= Void then
					double_item_write_monitor_agent.call ([a_item_id, a_signature, a_value, Qnan])
				end
				l_virtual_record.double_items.force (a_value,a_item_id)
			end
		end

	replace_integer_item (a_value: INTEGER; a_item_id: INTEGER; a_signature: G) is
			-- Replace the integer value identified by 'a_item_id' in 'a_signature' by 'a_value'
			-- if it is different that the current value or full_update_mode is enabled
		require
			valid_item: a_item_id > 0
			signature_not_void: a_signature /= Void
		local
			l_virtual_record: ROSE_VIRTUAL_RECORD[G]
		do
			cube_result_table.search (a_signature.multi_key)
			if cube_result_table.found then
				l_virtual_record := cube_result_table.found_item
				signature_tagged_server.create_trigger (data_name)
			else
				l_virtual_record := record_by_signature (a_signature)
				if l_virtual_record = Void then
					l_virtual_record := virtual_record (a_signature, True, not ignore_first_undefined)	-- Calls 'create_trigger'.
				else
					signature_tagged_server.create_trigger (data_name)
				end
			end
			l_virtual_record.integer_items.search (a_item_id)
			if l_virtual_record.integer_items.found then
				if signature_tagged_server.full_update_mode or l_virtual_record.integer_items.found_item /= a_value then
					signature_tagged_server.integer_data_replace (new_integer_record (a_value,a_item_id,l_virtual_record))
					if integer_item_write_monitor_agent /= Void then
						integer_item_write_monitor_agent.call ([a_item_id, a_signature, a_value, l_virtual_record.integer_items.found_item])
					end
					l_virtual_record.integer_items.force (a_value,a_item_id)
				end
			elseif (not ignore_first_undefined or a_value /= constants.null_integer) then
				if l_virtual_record.signature.signature_id = 0
				then
					l_virtual_record := virtual_record(a_signature, False, True)
				end

				signature_tagged_server.integer_data_replace (new_integer_record (a_value,a_item_id,l_virtual_record))
				if integer_item_write_monitor_agent /= Void then
					integer_item_write_monitor_agent.call ([a_item_id, a_signature, a_value, constants.null_integer])
				end
				l_virtual_record.integer_items.force (a_value,a_item_id)
			end
		end

	replace_string_item (a_value: STRING; a_item_id: INTEGER; a_signature: G) is
			-- Replace the string value identified by 'a_item_id' in 'a_signature' by 'a_value'
			-- if it is different that the current value or full_update_mode is enabled
		require
			valid_item: a_item_id > 0
			signature_not_void: a_signature /= Void
		local
			l_record: DATA_RECORD[STRING]
			l_virtual_record: ROSE_VIRTUAL_RECORD[G]
		do
			l_virtual_record := record_by_signature (a_signature)
			if l_virtual_record = Void then
				l_virtual_record := virtual_record (a_signature, True, not ignore_first_undefined)	-- Calls 'create_trigger'.
			else
				signature_tagged_server.create_trigger (data_name)
			end

			create l_record
			l_record.set_item_id (a_item_id)
			if a_value /= Void then
				l_record.set_value (a_value.twin)
			else
				l_record.set_value(a_value)
			end
			set_path (l_record,a_signature)
			l_record.set_signature_id (l_virtual_record.signature.signature_id)
			l_record.set_process_id (process_id)

			l_virtual_record.string_items.search (a_item_id)
			if l_virtual_record.string_items.found then
				if signature_tagged_server.full_update_mode or else
					(a_value = Void and l_virtual_record.string_items.found_item /= Void) or else
					(a_value /= Void and l_virtual_record.string_items.found_item = Void) or else
					(l_virtual_record.string_items.found_item /= Void and then not l_virtual_record.string_items.found_item.is_equal (a_value)) then
					signature_tagged_server.string_data_replace (l_record)
					if string_item_write_monitor_agent /= Void then
						string_item_write_monitor_agent.call ([a_item_id, a_signature, a_value, l_virtual_record.string_items.found_item])
					end
					if a_value /= Void then
						l_virtual_record.string_items.force (a_value.twin,a_item_id)
					else
						l_virtual_record.string_items.force (a_value,a_item_id)
					end
				end
			elseif (not ignore_first_undefined or a_value /= Void) then
				if l_virtual_record.signature.signature_id = 0
				then
					l_virtual_record := virtual_record(a_signature, False, True)
				end

				signature_tagged_server.string_data_replace (l_record)
				if string_item_write_monitor_agent /= Void then
					string_item_write_monitor_agent.call ([a_item_id, a_signature, a_value, Void])
				end
				if a_value /= Void then
					l_virtual_record.string_items.force (a_value.twin,a_item_id)
				else
					l_virtual_record.string_items.force (a_value,a_item_id)
				end
			end
		end

	replace_image_item (a_value: STRING; a_item_id: INTEGER; a_signature: G) is
			-- Replace the image value identified by 'a_item_id' in 'a_signature' by 'a_value'
			-- if it is different that the current value or full_update_mode is enabled
		require
			valid_item: a_item_id > 0
			signature_not_void: a_signature /= Void
		local
			l_record: DATA_RECORD[STRING]
			l_virtual_record: ROSE_VIRTUAL_RECORD[G]
		do
			l_virtual_record := record_by_signature (a_signature)
			if l_virtual_record = Void then
				l_virtual_record := virtual_record (a_signature, True, True)	-- Calls 'create_trigger'.
			else
				signature_tagged_server.create_trigger (data_name)
			end

			create l_record
			l_record.set_item_id (a_item_id)
			if a_value/=Void then
				l_record.set_value (a_value.twin)
			else
				l_record.set_value(a_value)
			end
			set_path (l_record,a_signature)
			l_record.set_signature_id (l_virtual_record.signature.signature_id)
			l_record.set_process_id (process_id)
			l_virtual_record.image_items.search (a_item_id)
			if l_virtual_record.image_items.found then
				if signature_tagged_server.full_update_mode or else
					(a_value = Void and l_virtual_record.image_items.found_item /= Void) or else
					(a_value /= Void and l_virtual_record.image_items.found_item = Void) or else
					(l_virtual_record.image_items.found_item /= Void and then not l_virtual_record.image_items.found_item.is_equal (a_value)) then
					signature_tagged_server.image_data_replace (l_record)
					if a_value /= Void then
						l_virtual_record.image_items.force (a_value.twin,a_item_id)
					else
						l_virtual_record.image_items.force (a_value,a_item_id)
					end
				end
			else
				signature_tagged_server.image_data_replace (l_record)
				if a_value /= Void then
					l_virtual_record.image_items.force (a_value.twin,a_item_id)
				else
					l_virtual_record.image_items.force (a_value,a_item_id)
				end
			end
		end

	replace_storable_item (a_object: ANY; a_item_id: INTEGER; a_signature: G) is
			-- Replace the storable value identified by 'a_item_id' in 'a_signature' by 'a_value'
			-- if it is different that the current value or full_update_mode is enabled
		require
			valid_item: a_item_id > 0
			signature_not_void: a_signature /= Void
		local
			l_string: STRING
			l_stream: ROSE_STREAM
		do
			signature_tagged_server.create_trigger (data_name)
			if a_object /= Void then
				create l_stream
				l_stream.independent_store (a_object)
				l_string := l_stream.string_representation
				replace_image_item (l_string, a_item_id, a_signature)
			end
		end

	replace_date_item (a_value: ROSE_DATE; a_item_id: INTEGER; a_signature: G) is
			-- Replace the date value identified by 'a_item_id' in 'a_signature' by 'a_value'
			-- if it is different that the current value or full_update_mode is enabled
		require
			valid_item: a_item_id > 0
			signature_not_void: a_signature /= Void
		local
			l_record: DATA_RECORD[ROSE_DATE]
			l_virtual_record: ROSE_VIRTUAL_RECORD[G]
		do
			l_virtual_record := record_by_signature (a_signature)
			if l_virtual_record = Void then
				l_virtual_record := virtual_record (a_signature, True, not ignore_first_undefined)	-- Calls 'create_trigger'.
			else
				signature_tagged_server.create_trigger (data_name)
			end

			create l_record
			l_record.set_item_id (a_item_id)
			if a_value/=Void then
				l_record.set_value (a_value.twin)
			else
				l_record.set_value (a_value)
			end
			set_path (l_record,a_signature)
			l_record.set_signature_id (l_virtual_record.signature.signature_id)
			l_record.set_process_id (process_id)

			l_virtual_record.date_items.search (a_item_id)
			if l_virtual_record.date_items.found then
				if signature_tagged_server.full_update_mode or else
					(a_value = Void and l_virtual_record.date_items.found_item /= Void) or else
					(a_value /= Void and l_virtual_record.date_items.found_item = Void) or else
					(l_virtual_record.date_items.found_item /= Void and then not l_virtual_record.date_items.found_item.is_equal (a_value)) then
					signature_tagged_server.date_data_replace (l_record)
					if a_value /= Void then
						l_virtual_record.date_items.force (a_value.twin,a_item_id)
					else
						l_virtual_record.date_items.force (a_value,a_item_id)
					end
				end
			elseif (not ignore_first_undefined or a_value /= Void) then
				if l_virtual_record.signature.signature_id = 0
				then
					l_virtual_record := virtual_record(a_signature, False, True)
				end

				signature_tagged_server.date_data_replace (l_record)
				if a_value /= Void then
					l_virtual_record.date_items.force (a_value.twin,a_item_id)
				else
					l_virtual_record.date_items.force (a_value,a_item_id)
				end
			end
		end

	replace_signature (a_signature: G) is
			-- Update the entire signature (including key fields) for the record identified by 'a_signature.signature_id'
		require
			signature_not_void: a_signature /= Void
			signature_not_zero: a_signature.signature_id /= 0
		local
			l_virtual_record: ROSE_VIRTUAL_RECORD[G]
		do
			signature_tagged_server.create_trigger (data_name)
			signature_tagged_server.replace_signature (a_signature)
			if cube_result_table /= Void then
				create l_virtual_record.make (a_signature)
				cube_result_table.force (l_virtual_record, a_signature.multi_key)
			end
		end

	replace_signature_data_array (a_sigs: ROSE_LINEAR_ARRAY[G]) is
			-- Like calling 'replace_signature_data' many times, but much faster (at least ~5-10 sig's, more is better).
			-- You MUST perform a commit before calling this routine again.
		do
			signature_tagged_server.create_trigger(data_name)
			signature_tagged_server.replace_signature_data_array(a_sigs)
		end

	load_signature_data_from_table (a_table_name: STRING) is
			-- This is just like 'replace_signature_data_array', but the data is already in a table.
			-- You MUST perform a commit before calling this routine again.
		do
			signature_tagged_server.create_trigger(data_name)
			signature_tagged_server.load_signature_data_from_table(a_table_name)
		end

	replace_signature_data (a_signature: G) is
			-- This will create a signature if one doesn't exist already.
			-- But see 'replace_signature_data_array' for a faster version.
		require
			signature_data_process_id_available: table_structure.signature_data_process_id_available
		local
			l_from_virtual_record: ROSE_VIRTUAL_RECORD[G]
			l_signature: G
		do
			l_from_virtual_record := record_by_signature (a_signature)
			if l_from_virtual_record = Void then
				l_from_virtual_record := virtual_record (a_signature, True, True).twin	-- Calls 'create_trigger'.
			else
				signature_tagged_server.create_trigger(data_name)
			end
			if l_from_virtual_record.signature.signature_data_process_id = 0 then
				-- **Get the signature data process id and the signature data fields**
				l_signature := signature_tagged_server.retrieve_signature_data (a_signature)
				if l_signature /= Void then
					l_from_virtual_record.set_signature (l_signature)
				end
			end
			if l_from_virtual_record.signature.continuing_process_id /= process_id then
				l_from_virtual_record.signature.set_last_process_id (process_id)
				l_from_virtual_record.signature.set_continuing_process_id (process_id)
				signature_touches.extend (l_from_virtual_record.signature)
			end
			a_signature.set_signature_id (l_from_virtual_record.signature.signature_id)
			signature_tagged_server.replace_signature_data (a_signature, l_from_virtual_record.signature)
		end

	replace_cached_signature_data (a_signature: G) is
			-- This will not create a signature if it does not exist already
		local
			l_from_virtual_record: ROSE_VIRTUAL_RECORD[G]
		do
			l_from_virtual_record := record_by_signature (a_signature)
			if l_from_virtual_record /= Void then
				a_signature.set_signature_id (l_from_virtual_record.signature.signature_id)
				signature_tagged_server.replace_signature_data (a_signature, l_from_virtual_record.signature)
				if l_from_virtual_record.signature.continuing_process_id /= process_id then
					l_from_virtual_record.signature.set_last_process_id (process_id)
					l_from_virtual_record.signature.set_continuing_process_id (process_id)
					signature_touches.extend (l_from_virtual_record.signature)
				end
			end
		end

	replace_real_override (a_value: REAL; a_item_id: INTEGER; a_signature: G) is
			-- Replace the real override value identified by 'a_item_id' in 'a_signature' by 'a_value'
		require
			valid_item: a_item_id > 0
			signature_not_void: a_signature /= Void
		local
			l_virtual_record: ROSE_VIRTUAL_RECORD[G]
		do
			set_results_from_cache (a_signature)
			l_virtual_record := virtual_record (a_signature, True, True)
			l_virtual_record.real_items.search (a_item_id)
			signature_tagged_server.real_override_data_replace (new_real_record (a_value,a_item_id,l_virtual_record))
		end

feature {MONDAT_COMPANY_HEADERS, MONDAT_SECURITY_HEADERS, ROSE_SIGNATURE_RECORD_SYNCH_DIALOG} -- Signature deletion

	delete_signature (a_signature: G) is
			-- Delete 'a_record' from the signature table
			-- Only use for databases with no item data
		require
			valid_signature: a_signature /= Void
			signature_only: table_structure.signature_only
		do
			signature_tagged_server.delete_signature (a_signature)
		end

feature {SIGNATURE_TAGGED_RETRIEVAL} -- Selection Space Configuration

	ordered_cube_of_items (a_columns: COLUMN_DESCRIPTORS[G];
						   a_items: ROSE_LINEAR_ARRAY[INTEGER]): ROSE_LINEAR_ARRAY[ROSE_VIRTUAL_RECORD[G]] is
			-- Array of current virtual records for for signatures meeting the criteria specified by 'a_columns'
			-- Queries only for item_ids in 'a_items' (all items if 'a_items' = Void)
		do
			signature_tagged_server.set_new_query
			Result := signature_tagged_server.ordered_cube_of_items (a_columns, a_items)
			if Result.is_empty then
				last_rose_signature := Void
			else
				last_rose_signature := Result.last
			end
			signature_tagged_server.set_virtual_record_set_cube_result_temp (Result)
			set_cache_from_results (last_rose_signature)
			signature_tagged_server.set_virtual_record_set_cube_result_temp (Void)
		end

feature {ROSE_SIGNATURE_RECORD} -- Date Restriction Clauses.

	data_date_restriction_statement (a_entity_name, a_date_name, a_min_max_string: STRING): STRING is
			-- A restriction statement to retrieve either the first or last record for an entity by a_date_name
		require
			valid_entity_name: a_entity_name /= Void
			valid_date_name: a_date_name /= Void
			valid_min_max_string: a_min_max_string /= Void and then a_min_max_string.as_lower.is_equal (min_string) or a_min_max_string.as_lower.is_equal (max_string)
		do
			Result := "#" +
				a_entity_name +
				"_id.id = " +
				a_entity_name +
				"_id and " +
				a_date_name +
				" = (select " +
				a_min_max_string +
				" (" +
				a_date_name +
				") from " +
				table_name +
				" where #" +
				a_entity_name +
				"_id.id = " +
				a_entity_name +
				"_id)"
		ensure
			valid_result: Result /= Void
		end

	process_date_restriction_statement (a_entity_name, a_min_max_string: STRING): STRING is
			-- A restriction statement to retrieve either the first or last record by process_id for an entity
		require
			valid_entity_name: a_entity_name /= Void
			valid_min_max_string: a_min_max_string /= Void and then a_min_max_string.as_lower.is_equal (min_string) or a_min_max_string.as_lower.is_equal (max_string)
		local
			l_process_id_string, l_asc_desc: STRING
		do
			if a_min_max_string.is_equal (min_string) then
				l_process_id_string := first_process_id_string
				l_asc_desc := ascending_string
			else
				l_process_id_string := continuing_process_id_string
				l_asc_desc := descending_string
			end
			Result := "#" +
				a_entity_name +
				"_id.id = " +
				a_entity_name +
				"_id and " +
				l_process_id_string +
				" = (select top 1 " +
				l_process_id_string +
				" from process as p, " +
				table_name +
				" as d where d." +
				a_entity_name +
				"_id = " +
				a_entity_name +
				"_id and p.process_id = d." +
				l_process_id_string +
				" order by p.file_date " +
				l_asc_desc +
				")"
		ensure
			valid_result: Result /= Void
		end

feature -- Status

	fill_pinned_version_dates (a_signatures: ROSE_LINEAR_ARRAY [G]): BOOLEAN is
			-- Return true if there is an overridden version date
		do
			Result := signature_tagged_server.virtual_record_set.fill_pinned_version_dates (a_signatures)
		end

	process_id_manager_created: BOOLEAN is
			-- Did Current create the currently active process id manager?
		do
			Result := signature_tagged_server.process_id_manager_created
		end

	process_id_available: BOOLEAN is
			-- Is a process id currently available?
		do
			Result := process_id_manager_created or signature_tagged_server.process_id_set
		end

	should_build_cube_position_result: BOOLEAN is
			-- Will queries populate the cube_position_result object?
		do
			Result := signature_tagged_server.should_build_cube_position_result
		end

	data_type_available, regular_data_type_available (a_type: STRING): BOOLEAN is
			-- Does this database have 'a_type' data?
		do
			Result := signature_tagged_server.data_type_available (a_type)
		end

	override_data_type_available (a_type: STRING): BOOLEAN is
			-- Does this database have 'a_type' override data?
		do
			Result := signature_tagged_server.override_data_type_available (a_type)
		end

	showing_suppressed_historical_data: BOOLEAN is
			-- Is this database currently showing suppressed historical data?
		do
			Result := signature_tagged_server.showing_suppressed_historical_data
		end

	is_valid_restriction_statement (a_statement: STRING): BOOLEAN is
			-- Is the restriction statement 'a_statement' valid for use in a query?
		do
			if a_statement = Void or else
				(not columns_from_restriction_statement (a_statement).is_empty and
				 (a_statement.has_substring ("=") or
				  a_statement.has_substring (">") or
				  a_statement.has_substring ("<"))) then
				Result := True
			end
		end

	uncommitted: BOOLEAN is
			-- Is there uncommitted data on the server being used by this database object
		do
			Result := signature_tagged_server.uncommitted or else signature_tagged_server.data_server_transaction_depth > 0
		end

	path_is_cached (a_path: INTEGER): BOOLEAN is
			-- Is the cluster identified by 'a_path' currently in the cache?
		do
			Result := manager_cache.item (a_path) /= Void
		end

feature -- Element modification

	set_virtual_record_template (a_template: ROSE_VIRTUAL_RECORD[G]) is
			-- Set the virtual record template to 'a_template'
			-- This can be useful for specifying a different descendent of G to use for query results
		do
			signature_tagged_server.set_virtual_record_template (a_template)
		end

	set_max_cube_result_size (a_max_cube_result_size: INTEGER) is
			-- Set max_cube_result_size to 'a_max_cube_result_size'
		do
			max_cube_result_size := a_max_cube_result_size
		end

	set_process_id_manager (a_manager: SUPPLIER_PROCESS_ID_MANAGER) is
			-- Set the process id manager for this database
		do
			signature_tagged_server.set_process_id_manager (a_manager)
		end

	set_process_id (a_process_id: INTEGER) is
			-- Set the current process_id to use for updates to 'a_process_id'
		require
			valid_process_id: a_process_id > 0
		do
			signature_tagged_server.set_process_id (a_process_id)
		end

	set_process_name (a_name: STRING) is
			-- Set the process name to 'a_name'.  This does not change the name of the currently active process if there is one
		require
			meaningful_name: a_name /= Void and then not a_name.is_empty
		do
			signature_tagged_server.set_process_name (a_name)
		end

	set_override_family_name (a_name: STRING) is
			-- Set the override family to use for queries
		do
			signature_tagged_server.set_override_family_name (a_name)
		end

	set_cache_pool_on (a_value: BOOLEAN) is
			-- Currently default is set to on
		do
			signature_tagged_server.set_cache_pool_on (a_value)
		end

	set_submit_frequency (a_freq: INTEGER) is
			-- Set the frequency with which data updates will auto-submit
		do
			signature_tagged_server.set_submit_frequency (a_freq)
		end

	set_server_for_writing (a_server: SQL_DATA_SERVER) is
			-- Set the server to use for writing data, may be different than the server used to retrieve data
		do
			raise ("Not Implemented")
		end

	set_ignore_first_undefined (a_value: BOOLEAN) is
			-- Set ignore_first_undefined to 'a_value'
		do
			ignore_first_undefined := a_value
			last_virtual_record := Void
		end

	set_manager_cache (a_value: like manager_cache) is
			-- Set the manager_cache to 'a_value'
		do
			if a_value /= Void then
				manager_cache_set_externally := True
			else
				manager_cache_set_externally := False
			end
			manager_cache := a_value
		end

	set_real_item_write_monitor_agent (a_agent: like real_item_write_monitor_agent) is
			-- Set 'real_item_write_monitor_agent' to 'a_agent'
		do
			real_item_write_monitor_agent := a_agent
		end

	set_double_item_write_monitor_agent (a_agent: like double_item_write_monitor_agent) is
			-- Set 'double_item_write_monitor_agent' to 'a_agent'
		do
			double_item_write_monitor_agent := a_agent
		end

	set_integer_item_write_monitor_agent (a_agent: like integer_item_write_monitor_agent) is
			-- Set 'integer_item_write_monitor_agent' to 'a_agent'
		do
			integer_item_write_monitor_agent := a_agent
		end

	set_string_item_write_monitor_agent (a_agent: like string_item_write_monitor_agent) is
			-- Set 'string_item_write_monitor_agent' to 'a_agent'
		do
			string_item_write_monitor_agent := a_agent
		end

feature {NONE} -- Implementation

	max_cube_result_size: INTEGER
			-- The upper limit of records in the cube_result_table (only use for testing)
	manager_cache_set_externally: BOOLEAN
			-- Was the manager cache given to Current?
	signature_tagged_server: SIGNATURE_TAGGED_SERVER[G]
			-- The server object that interfaces directly with the underlying data storage
	last_signature: G
			-- The last signature that was used for a query
	last_signature_id: INTEGER
			-- The last signature id that was retrieved from the server
	last_virtual_record: ROSE_VIRTUAL_RECORD[G]
			-- The last virtual record returned to the client
	last_rose_signature: ROSE_VIRTUAL_RECORD[G]
			-- The last virtual record of the last query result set
	real_item_write_monitor_agent: PROCEDURE [ANY, TUPLE [INTEGER, G, REAL, REAL]]
			-- Agent to call when a change has been detected to a real item
	double_item_write_monitor_agent: PROCEDURE [ANY, TUPLE [INTEGER, G, DOUBLE, DOUBLE]]
			-- Agent to call when a change has been detected to a double item
	integer_item_write_monitor_agent: PROCEDURE [ANY, TUPLE [INTEGER, G, INTEGER, INTEGER]]
			-- Agent to call when a change has been detected to a integer item
	string_item_write_monitor_agent: PROCEDURE [ANY, TUPLE [INTEGER, G, STRING, STRING]]
			-- Agent to call when a change has been detected to a string item

feature {NONE} -- Local cache

	manager_cache: ROSE_VOLATILE_CACHE[TRIPLE[ROSE_LINEAR_ARRAY[ROSE_VIRTUAL_RECORD[G]],HASH_TABLE[ROSE_VIRTUAL_RECORD[G], MULTI_KEY], INTEGER],INTEGER]
			-- A volatile cache that manages the retrieval history
			-- The TRIPLE is [cube_result, cube_result_table, path]
	signature_touches: ROSE_LINEAR_ARRAY[G]
			-- Collection of records that have been updated since the last submit
	cube_signature_result: ROSE_LINEAR_ARRAY [G]
			-- Array of signatures retrieved
	cube_position_result: HASH_TABLE[G, INTEGER]
			-- Table, keyed by signature_id, of what part of the query resulted in that signature being brought in
	cube_result: ROSE_LINEAR_ARRAY[ROSE_VIRTUAL_RECORD[G]]
			-- Array of virtual records retrieved
	cube_result_table: HASH_TABLE[ROSE_VIRTUAL_RECORD[G], MULTI_KEY]
			-- Table, keyed by signature multi key, of the last retrieval

feature {NONE} -- Implementation

	cleanup_items is
			-- Ask the data item objects to clean themselves up
		do
			signature_tagged_server.cleanup_items
		end

	submit_signature_touches is
			-- Submit the signature version changes
		do
			if signature_touches.count > 0 then
				signature_tagged_server.submit_signature_touches (signature_touches)
			end
		end

	initial_cache_size: INTEGER is
			-- The initial cache size for the manager_cache.
			-- This is the number of clusters to allow before it starts to sweep
		do
			Result := 100
		end

	set_cache_from_results (a_record: ROSE_VIRTUAL_RECORD[G]) is
			-- Update the local cache and manager_cache with the results of the last query in the signature_tagged_server
		do
			signature_tagged_server.set_cache_from_results (a_record, Current)
		end

	set_results_from_cache (a_signature: G) is
			-- Set the local cache from the manager_cache by looking for 'a_signature' there
		require
			signature_not_void: a_signature /= Void
		local
			l_path: INTEGER
			l_virtual_record_set: TRIPLE[ROSE_LINEAR_ARRAY[ROSE_VIRTUAL_RECORD[G]],
										 HASH_TABLE[ROSE_VIRTUAL_RECORD[G], MULTI_KEY], INTEGER]
		do
			if a_signature /= Void then
				l_virtual_record_set := manager_cache.found_item
				if not signature_tagged_server.ad_hoc_query then
					l_path := path (a_signature)
				end
				if l_virtual_record_set /= Void and then
					l_virtual_record_set.first = cube_result and
						l_virtual_record_set.second = cube_result_table and
						l_virtual_record_set.third = l_path
				 then
					-- Cube_result and _table are already set correctly.
					cache_rehit_count := cache_rehit_count + 1
				else
					l_virtual_record_set := manager_cache.item (l_path)
					if l_virtual_record_set = Void then
						cache_miss_count := cache_miss_count + 1
						signature_tagged_server.remake_virtual_record_set
						create cube_result.make_empty
						cube_result_table := Void
						signature_tagged_server.set_current_cache_pool (a_signature)
					else
						cache_hit_count := cache_hit_count + 1
						cube_result := l_virtual_record_set.first
						cube_result_table := l_virtual_record_set.second
					end
				end
			else
				signature_tagged_server.remake_virtual_record_set
				create cube_result.make_empty
				create cube_result_table.make (0)
			end
		end

	initialize_manager (a_name: STRING; a_table_structure: SIGNATURE_TAGGED_TABLE_STRUCTURE) is
			-- Initialize the server
		require
			name_not_void: a_name /= Void
		do
			signature_tagged_server.set_process_name (a_name)
			create_cache_variables
			signature_tagged_server.create_data_items
			signature_tagged_server.create_selected_column_ids
			signature_tagged_server.check_full_update_mode
		end

	set_path (a_record: ANY; a_signature: G) is
			-- Set the path of 'a_record' to the path for 'a_signature'
		do
			signature_tagged_server.set_path (a_record, a_signature)
		end

	set_signature_as_of_date (a_signature: G) is
		-- Set the as_of_date of 'a_signature' by looking up the last_process_id in the process_id_map
		local
			l_as_of_date: ROSE_DATE
		do
			if a_signature /= Void then
				l_as_of_date := process_id_map.item (a_signature.last_process_id)
				if l_as_of_date /= Void then
					a_signature.set_as_of_date (l_as_of_date)
				end
			end
		end

	virtual_record (a_signature: G; a_can_read_old_signature, a_can_write_new_signature: BOOLEAN): ROSE_VIRTUAL_RECORD[G] is
			-- Used to get and prepare records for writing.
			-- See SIGNATURE_TAGGED_SERVER.signature_id for documentation about the 'a_can_*' arguments.
		do
			signature_tagged_server.create_trigger (data_name)
			Result := virtual_record_query (a_signature, a_can_read_old_signature, a_can_write_new_signature)
			if Result.signature.continuing_process_id /= process_id and
			   Result.signature.signature_id /= 0
			then
				Result.signature.set_last_process_id (process_id)
				Result.signature.set_continuing_process_id (process_id)
				signature_touches.extend (Result.signature)
			end
		end

	virtual_record_query (a_signature: G; a_can_read_old_signature, a_can_write_new_signature: BOOLEAN): ROSE_VIRTUAL_RECORD[G] is
			-- Cursory check for a virtual_record.  Will create one if none is found (unless 'a_can_write_new_signature' is False).
			-- See SIGNATURE_TAGGED_SERVER.signature_id for documentation about the 'a_can_*' arguments.
		do
			if cube_result_table = Void then
				create cube_result_table.make (0)
				last_virtual_record := Void
			end
			if not a_can_read_old_signature or last_virtual_record = Void or not last_virtual_record.signature.multi_key.is_equal (a_signature.multi_key) then
				cube_result_table.search (a_signature.multi_key)
				if not cube_result_table.found then
					create Result.make (a_signature)
					Result.signature.set_signature_id (signature_id (a_signature, a_can_read_old_signature, a_can_write_new_signature))
					if max_cube_result_size > 0 and then cube_result_table.count > max_cube_result_size then
						purge
					end

					if Result.signature.signature_id /= 0
					then
						cube_result_table.force (Result,a_signature.multi_key)
					end

					last_virtual_record := Result
				else
					last_virtual_record := cube_result_table.found_item
					Result := last_virtual_record
				end
			else
				Result := last_virtual_record
			end
		end

	signature_id (a_signature: G; a_can_read_old_signature, a_can_write_new_signature: BOOLEAN): INTEGER
			-- The signature id for 'a_signature', a new one will be created if it does not already exist
			-- See SIGNATURE_TAGGED_SERVER.signature_id for documentation about the 'a_can_*' arguments.
		require
			valid_signature: a_signature /= Void
		do
			if last_signature /= Void and then last_signature.signature_id = a_signature.signature_id then
				Result := last_signature_id
			else
				Result := signature_tagged_server.signature_id (a_signature, a_can_read_old_signature, a_can_write_new_signature)
				if Result /= 0
				then
					last_signature_id := Result
					last_signature := a_signature
				else
					last_signature := Void
				end
			end
		end

	last_signature_created: BOOLEAN is
			-- Did the last signature_id query result in the creation of a new record?
		do
			Result := signature_tagged_server.last_signature_created
		end

	columns_from_restriction_statement (a_statement: STRING): ROSE_LINEAR_ARRAY[STRING] is
			-- The signature key fields that are specified by 'a_statement'
		do
			Result := signature_tagged_server.columns_from_restriction_statement (a_statement)
		end

	cube_from_signatures (a_signatures: ROSE_LINEAR_ARRAY[G]; a_columns: ROSE_LINEAR_ARRAY[STRING]): COLUMN_DESCRIPTORS[G] is
			-- The selection criteria given the collection of signatures and column that are requested to be queried on
		require
			valid_columns: a_columns /= Void
		local
			l_column: STRING
			l_last_signature, l_new_signature: G
			l_column_values: ROSE_LINEAR_ARRAY[G]
			l_void: G
		do
			create Result.make_empty
			from a_columns.start until a_columns.off loop
				l_column := a_columns.current_item
				create l_column_values.make_empty
				from a_signatures.start until a_signatures.off loop
					l_new_signature := signature_set_from_source (l_last_signature,a_signatures.current_item,l_column)
					if l_new_signature /= Void then
						l_column_values.extend (l_new_signature)
						l_last_signature := l_new_signature
					end
					a_signatures.forth
				end
				if l_column_values.count > 0 then
					Result.add_signatures (l_column_values)
				else
					raise ("Restriction Clause identifies a column not specified in any signature")
				end
				l_last_signature := l_void
				a_columns.forth
			end
		end

	signature_set_from_source (a_last_signature, a_signature: G; a_name: STRING): G is
			-- A signature with the field 'a_name' set from the same field in 'a_last_signature'
		require
			valid_signature: a_signature /= Void
			valid_name: a_name /= Void
		local
			l_value: ANY
			l_last_integer_ref, l_integer_ref: INTEGER_REF
		do
			l_value := a_signature.value_by_name (a_name)
			if l_value /= Void then
				l_integer_ref ?= l_value
				if l_integer_ref = Void or else l_integer_ref.item /= 0 then
					if a_last_signature /= Void then
						l_last_integer_ref ?= a_last_signature.value_by_name (a_name)
					end
					if l_last_integer_ref = Void or else l_last_integer_ref.item /= l_integer_ref.item then
						create Result.default_create
						Result.set_field (a_name,l_value)
					end
				end
			end
		end

	select_signature_ids (a_columns: COLUMN_DESCRIPTORS[G]) is
			-- Retrieve the signature ids matching the criteria as specified by 'a_columns'
		do
			signature_tagged_server.select_signature_ids (a_columns)
		end

	order_direction (a_comparison: STRING): STRING is
			-- The sql order string given 'a_comparison'
		do
			a_comparison.compare_objects
			if a_comparison.has ('<') then
				Result := "desc"
			else
				Result := ""
			end
		end

	new_real_record (a_value: REAL; a_item_id: INTEGER; a_signature: ROSE_VIRTUAL_RECORD[G]): DATA_RECORD[REAL] is
			-- Create a new real data record for 'a_item_id' for 'a_signature' with the value 'a_value'
			-- This record will be used to update the value in the database
		require
			valid_signature: a_signature /= Void
		do
			create Result
			Result.set_item_id (a_item_id)
			Result.set_value (a_value)
			set_path (Result,a_signature.signature)
			Result.set_signature_id (a_signature.signature.signature_id)
			Result.set_process_id (process_id)
		end

	new_double_record (a_value: DOUBLE; a_item_id: INTEGER; a_signature: ROSE_VIRTUAL_RECORD[G]): DATA_RECORD[DOUBLE] is
			-- Create a new double data record for 'a_item_id' for 'a_signature' with the value 'a_value'
			-- This record will be used to update the value in the database
		require
			valid_signature: a_signature /= Void
		do
			create Result
			Result.set_item_id (a_item_id)
			Result.set_value (a_value)
			set_path (Result,a_signature.signature)
			Result.set_signature_id (a_signature.signature.signature_id)
			Result.set_process_id (process_id)
		end

	new_integer_record (a_value: INTEGER; a_item_id: INTEGER; a_signature: ROSE_VIRTUAL_RECORD[G]): DATA_RECORD[INTEGER] is
			-- Create a new integer data record for 'a_item_id' for 'a_signature' with the value 'a_value'
			-- This record will be used to update the value in the database
		require
			valid_signature: a_signature /= Void
		do
			create Result
			Result.set_item_id (a_item_id)
			Result.set_value (a_value)
			set_path (Result,a_signature.signature)
			Result.set_signature_id (a_signature.signature.signature_id)
			Result.set_process_id (process_id)
		end

feature -- Cache

	create_cache_variables is
			-- Recreate the local caches.  This will force a retrieval from the server at the next query
		do
			create signature_touches.make_empty
			create cube_signature_result.make_empty
			create cube_result_table.make (0)
			create cube_position_result.make (0)
			create manager_cache.make (initial_cache_size,1)
		end

feature {SIGNATURE_TAGGED_SERVER,STID_TESTENV_DATABASE,STID_TESTENV_VIEW} -- Path

	path (a_signature: G): INTEGER is
			-- Id of the cluster where 'a_signature' resides
		do
			Result := signature_tagged_server.path (a_signature)
		end

feature {SIGNATURE_TAGGED_EXPRESSION} -- Template anchors

	record_templates: ROSE_RECORD_ITEM_TABLE is
			-- The type anchors for the record types available in this database, this is redefined in the descendents
		do
		end

feature {SIGNATURE_TAGGED_SERVER} -- Cache management

	force_manager_cache (a_virtual_record_set: TRIPLE[ROSE_LINEAR_ARRAY[ROSE_VIRTUAL_RECORD[G]],
													  HASH_TABLE[ROSE_VIRTUAL_RECORD[G], MULTI_KEY], INTEGER]
						 a_path: INTEGER) is
			-- Force 'a_virtual_record_set' into the manager_cache keyed by 'a_path'
		do
			manager_cache.force (a_virtual_record_set, a_path)
		end

	set_cube_result_table (a_cube_result_table: HASH_TABLE[ROSE_VIRTUAL_RECORD[G], MULTI_KEY]) is
			-- Set cube_result_table to 'a_cube_result_table'
		do
			cube_result_table := a_cube_result_table
		end

	set_cube_signature_result (a_cube_signature_result: ROSE_LINEAR_ARRAY[G]) is
			-- Set cube_signature_result to 'a_cube_signature_result'
		do
			cube_signature_result := a_cube_signature_result
		end

	set_cube_position_result (a_cube_position_result: HASH_TABLE[G, INTEGER]) is
			-- Set cube_position_result to 'a_cube_position_result'
		do
			cube_position_result := a_cube_position_result
		end

	set_cube_result (a_cube_result: ROSE_LINEAR_ARRAY[ROSE_VIRTUAL_RECORD[G]]) is
			-- Set cube_result to 'a_cube_result'
		do
			cube_result := a_cube_result
		end

feature -- Reporting

	print_stats is
			-- Print the statistics summary string to the screen
		do
			print (stats_string)
		end

	stats_string: STRING is
			-- Readable string of the cache usage statistics
		local
			l_stats_string: STRING
			l_remote_stats_string: STRING
		do
			l_stats_string :=
				"Cache miss count  = " + cache_miss_count.out + "%N" +
				"Cache hit count   = " + cache_hit_count.out + "%N" +
				"Cache rehit count = " + cache_rehit_count.out + "%N"
			l_remote_stats_string := signature_tagged_server.stats_string
			if l_remote_stats_string /= Void and then l_remote_stats_string.count /= 0 then
				Result :=
					"Local stats:%N" +
					l_stats_string +
					"Remote stats:%N" +
					l_remote_stats_string
			else
				Result := l_stats_string
			end

		end

	cache_miss_count: INTEGER
			-- Count of how many times a retrieval was not contained in the current cache
	cache_hit_count: INTEGER
			-- Count of how many times a retrieval was contained in the current cache but not in the currently active cluster
	cache_rehit_count: INTEGER
			-- Count of how many times a retrieval was contained in the currently active cluster

	dump_all_caches is
			-- Display all of the data in the caches that is ready to be submitted
		do
			signature_tagged_server.dump_all_caches
		end

	dump_replace_caches is
			-- Display all of the data in the replace caches that is ready to be submitted
		do
			signature_tagged_server.dump_replace_caches
		end

	dump_replace_cache (a_type: STRING) is
			-- Display all of the data in the replace caches for 'a_type' that is ready to be submitted
		do
			signature_tagged_server.dump_replace_cache (a_type)
		end

	column_bounds (a_column_name: STRING): COLUMN_BOUNDS is
			-- The bounds of values for 'a_column_name'
		require
			column_name_not_void: a_column_name /= Void
			meaninful_column_name: new_signature_template.fields_.as_lower.has_substring (a_column_name.as_lower)
		do
			Result := signature_tagged_server.column_bounds (a_column_name)
		end

	data_bounds (a_signature: G): SIGNATURE_TAGGED_DATA_BOUNDS is
			-- The bounds of each signature field for the record identified by 'a_signature.entity_name'.
			-- See 'preload_data_bounds'.
		require
			signature_not_void: a_signature /= Void
		do
			Result := signature_tagged_server.data_bounds (a_signature,data_name)
		end

	preload_data_bounds(a_signatures: ROSE_LINEAR_ARRAY[G])
	-- Quickly preload the data_bounds for fast access.
	do
		signature_tagged_server.preload_data_bounds(a_signatures, data_name)
	end

	records_touched_since (a_process_id: INTEGER; a_items: ROSE_LINEAR_ARRAY[INTEGER]): ROSE_LINEAR_ARRAY[G] is
			-- Collection of records that have had any of their 'a_items' updated since 'a_process_id'
			-- Void 'a_items' indicates that you want any change
		do
			Result := signature_tagged_server.records_touched_since (a_process_id, a_items)
		end

	records_touched_by (a_process_ids: ROSE_LINEAR_ARRAY[INTEGER]; a_items :ROSE_LINEAR_ARRAY[INTEGER]): ROSE_LINEAR_ARRAY[G] is
			-- Collection of records that have had any of their 'a_items' updated by 'a_process_ids'
			-- Void 'a_items' indicates that you want any change
		require
			process_ids_not_void: a_process_ids /= Void
		do
			Result := signature_tagged_server.records_touched_by (a_process_ids, a_items)
		end

	paths_touched_since (a_process_id: INTEGER): ROSE_LINEAR_ARRAY[INTEGER] is
			-- Collection of paths that have been updated since 'a_process_id'
		do
			Result := signature_tagged_server.paths_touched_since (a_process_id)
		end

	records_count (a_name: STRING; a_value, a_item: INTEGER): INTEGER is
			-- The count of records having 'a_value' as the value of the field 'a_name'.
			-- 'a_item' restricts the query to only include records that have a value for that item
		require
			name_not_void: a_name /= Void
		do
			Result := signature_tagged_server.records_count (a_name, a_value, a_item)
		end

feature -- String constants

	min_string: STRING is "min"
			-- String constant for min to be used for sql statement construction
	max_string: STRING is "max"
			-- String constant for min to be used for sql statement construction
	first_process_id_string: STRING is "first_process_id"
			-- String constant for first_process_id to be used for sql statement construction
	continuing_process_id_string: STRING is "continuing_process_id"
			-- String constant for continuing_process_id to be used for sql statement construction
	ascending_string: STRING is "asc"
			-- String constant for aasc to be used for sql statement construction
	descending_string: STRING is "desc"
			-- String constant for desc to be used for sql statement construction

feature {NONE} -- Helpers

	internal: INTERNAL is
			-- Access to INTERNAL
		do
			create Result
		end

invariant

	valid_server: signature_tagged_server /= Void

end
