indexing
	description: "This class manages access to and from the record based item tables"
	keywords: "database,signature,item,tagged,manager"
	author: "Mauricio Bustos"
	date: "$Date: 2009/11/05 $"
	revision: "$Revision: #33 $"

class
	SIGNATURE_TAGGED_RECORD_ITEM

inherit
	SIGNATURE_TAGGED_ITEM [COMPARABLE]
		rename
			make as signature_tagged_item_make,
			class_name as exception_class_name,
			replace as signature_tagged_item_replace,
			update_cache as signature_tagged_item_update_cache,
			insert as signature_tagged_item_insert,
			insert_saved as signature_tagged_item_insert_saved,
			type_name as signature_tagged_item_type_name
		redefine
			delete_nulls,
			delete_no_change,
			insert_new_data,
			update_data,
			dump_update_cache,
			create_temp_tables,
			insert_predicate,
			commit_updated_cache,
			commit_fast_load,
			unsubmitted,
			clear_updated_fast_insert,
			trigger_source_code_date_a,
			trigger_source_code_date_b,
			copy_to_history_statement
		end

create
	make

feature -- Creation.
	make(a_server : SQL_DATA_SERVER; a_process_id : INTEGER; a_type_name,
		 a_data_name : STRING; a_record : ROSE_SIGNATURE_RECORD_ITEM) is
		do
			record ?= a_record
			uses_self_archive := record.uses_self_archive
			signature_tagged_item_make(a_server,a_process_id,a_type_name,a_data_name)
			create update_cache.make_empty
		end

	record : ROSE_SIGNATURE_RECORD_ITEM

feature {NONE} -- Implementation.
	record_field_type(a_field : STRING) : STRING is
			-- Used to find the SQL type corresponding to 'a_field' (name of an attribute) in 'record'.
		local
			l_string : STRING
			l_stream : ROSE_STREAM
			l_date: ROSE_DATE
			l_datetime : ROSE_DATE
			l_type : INTEGER
			l_value : ANY
		do
			l_type := record.field_type(a_field)
			inspect l_type
			when {INTERNAL}.INTEGER_32_TYPE then
				Result := "INT"
			when {INTERNAL}.REAL_TYPE then
				Result := "REAL"
			when {INTERNAL}.DOUBLE_TYPE then
				Result := "FLOAT"
			when {INTERNAL}.BOOLEAN_TYPE then
				Result := "BIT"
			when {INTERNAL}.REFERENCE_TYPE then
				l_value := record.value_by_name(a_field)
				l_stream ?= l_value
				if l_stream /= Void then
					Result := "IMAGE"
				end

				l_string ?= l_value
				if l_string /= Void then
					Result := "VARCHAR(8000)"
				end

				l_date ?= l_value
				l_datetime ?= l_value
				if l_date/=Void or l_datetime/=Void then
					Result := "DATETIME"
				end

				if Result = Void then
					print("Check to make sure that all reference attributes are assigned a non-Void value in 'make'.%N")
					print("Or, this feature needs to be modified to include some other class.%N")
					raise("Read the message on the screen")
				end
			else
				raise("Invalid type used in record definition")
			end
		end

	create_temp_tables is
		local
			l_sql_type : STRING
			l_statement : STRING
			l_fields : ROSE_LINEAR_ARRAY[STRING]
			l_server: SQL_DATA_SERVER
		do
			l_statement :=
				"create table " + data_value_temp_table + "%N(%N" +
				" signature_id int, process_id int"
			if is_path_clustered then
				l_statement := l_statement + ", path int"
			end

			l_fields := record.record_fields
			from l_fields.start until l_fields.off loop
				l_sql_type := record_field_type(l_fields.current_item)
				if l_sql_type /= Void then
					l_statement.append(",%N [" + l_fields.current_item + "] " + l_sql_type)
				else
					raise("Could not get sql_type for " + l_fields.current_item)
				end
				l_fields.forth
			end

			if record.is_indexed
			then
				from
					l_fields := record.key_fields
					l_fields.start
				until
					l_fields.off
				loop
					l_sql_type := record_field_type(l_fields.current_item)
					l_statement.append(", %N [" + l_fields.current_item + "] " + l_sql_type)
					l_fields.forth
				end
			end

			if uses_self_archive
			then
				l_sql_type := record_field_type(record.archive_date_field_name)
				l_statement.append(",%N [" + record.archive_date_field_name + "] " + l_sql_type)
			end

			l_statement := l_statement + "%N)"
			l_server := signature_item_server.auxiliary_server
			l_server.begin_transaction
			l_server.sql_execute_immediate(l_statement)

			l_statement := "create unique index temp_index_1 on " + data_value_temp_table + "(signature_id"
			if record.is_indexed
			then
				from
					l_fields := record.key_fields
					l_fields.start
				until
					l_fields.off
				loop
					l_statement.append(", %N [" + l_fields.current_item + "]")
					l_fields.forth
				end
			end

			if uses_self_archive
			then
				l_statement.append(",%N [" + record.archive_date_field_name + "]")
			end

			l_statement.append(")")
			l_server.sql_execute_immediate(l_statement)
			l_server.commit
		end

feature -- Cache Dumps

	dump_update_cache is
		local
			l_fields : ROSE_LINEAR_ARRAY[STRING]
			l_statement : STRING
			l_field : ANY
		do
			l_fields := record.record_fields
			from update_cache.start until update_cache.off loop
				l_statement := "Signature_ID :" + update_cache.current_item.signature_id.out
				from l_fields.start until l_fields.off loop
					l_field ?= record.value_by_name(l_fields.current_item)
					l_statement := l_statement + " " + l_fields.current_item + " :" + l_field.out + "%N"
					l_fields.forth
				end
				print(l_statement + "%N")
				update_cache.forth
			end
		end

feature -- Replace

	replace(a_item : ROSE_SIGNATURE_RECORD_ITEM) is
		require
			item_not_void : a_item /= Void
		do
			update_cache.extend(a_item)
			items_replaced := items_replaced + 1
			process_new_data
		end

	insert(a_item : ROSE_SIGNATURE_RECORD_ITEM) is
		local
			l_server : DATA_SERVER
		do
			if insert_saved = Void then
				l_server := servers.make_from_repository_fast_load (signature_item_server.logical)
				insert_saved := l_server.fast_loader(data_value_table,a_item)
			end
			insert_saved.insert(a_item)
		end
	insert_saved : DATA_SERVER_FAST_LOAD

	commit_fast_load is
		do
			if insert_saved /= Void then
				insert_saved.commit
			end
		end

	unsubmitted : BOOLEAN is
		do
			Result := (update_cache.count /= 0)
		end

feature {NONE} -- Updated Data Management Predicates

	clear_updated_fast_insert is
		do
			update_cache.wipe_out
		end

	update_cache : ROSE_LINEAR_ARRAY[ROSE_SIGNATURE_RECORD_ITEM]
	commit_updated_cache is
		local
			l_fast_data_server : SQL_DATA_SERVER
			l_fast_insert_predicate : DATA_SERVER_FAST_LOAD
			l_duplicate_finder : HASH_TABLE[INTEGER, MULTI_KEY]
			l_key: MULTI_KEY
		do
			l_fast_data_server := servers.make_from_repository_fast_load (signature_item_server.logical)

			l_fast_insert_predicate := l_fast_data_server.fast_loader(data_value_temp_table,record)

			from
				update_cache.start
			until
				update_cache.off
			loop
				l_fast_insert_predicate.insert(update_cache.current_item)
				update_cache.forth
			end
			l_fast_insert_predicate.commit
			if not l_fast_insert_predicate.success then
				raise("failure in fast insert commit")
			end
		rescue
			create l_duplicate_finder.make(10)
			print("%NProblem rows:%N%N")
			from update_cache.start until update_cache.off loop
				if record.is_indexed
				then
					l_key := update_cache.current_item.multi_key
				else
					create l_key.make(<<update_cache.current_item.signature_id>>)
				end

				l_duplicate_finder.search(l_key)
				if l_duplicate_finder.found then
					print("Problem at this row : SIGNATURE_ID -> " + update_cache.current_item.signature_id.out + "%N")
				else
					l_duplicate_finder.force(update_cache.current_item.signature_id, l_key)
				end

				update_cache.forth
			end

			do_nothing	-- You can put a breakpoint here (but not on the 'end' statement).
		end

	insert_predicate : EXECUTION_PREDICATE is
		local
			l_statement : STRING
		do
			if insert_predicate_saved = Void then
				l_statement := "insert into " + data_value_temp_table + "(signature_id,process_id"
				from record.record_fields.start until record.record_fields.off loop
					l_statement := l_statement + ",[" + record.record_fields.current_item + "]"
					record.record_fields.forth

				end
				l_statement := l_statement + ") values(:signature_id,:process_id"
				from record.record_fields.start until record.record_fields.off loop
					l_statement := l_statement + ",:" + record.record_fields.current_item
					record.record_fields.forth
				end
				l_statement := l_statement + ")"
				create insert_predicate_saved.make(temp_table_server,l_statement)
				if not insert_predicate_saved.success then
					raise("Failed to create insert_predicate_saved")
				end
			end
			Result := insert_predicate_saved
		end

	delete_nulls is
		do
		end

	insert_new_data is
		local
			l_statement: STRING
			l_fields : ROSE_LINEAR_ARRAY[STRING]
			l_field_list: STRING
		do
			l_field_list := field_list

			l_statement :=
				"insert into " + data_value_table + "%N" +
				" (" + l_field_list + ")%N" +
				"select%N" +
				"  " + l_field_list +
				"%Nfrom " + data_value_temp_table +
				"%Nwhere signature_id not in%N" +
				"(select signature_id from " + data_value_table + " where%N" +
				" " + data_value_temp_table + ".signature_id = " + data_value_table + ".signature_id"

			if record.is_indexed
			then
				from
					l_fields := record.key_fields
					l_fields.start
				until
					l_fields.off
				loop
					l_statement.append(
						" and%N " +
						data_value_temp_table + "." + l_fields.current_item + " = " +
						data_value_table      + "." + l_fields.current_item)

					l_fields.forth
				end
			end

			if uses_self_archive
			then
				-- e.g. "data_value_temp_table.end_date = '99991231'".
				l_statement.append(
					" and%N " + data_value_temp_table + "." + record.archive_date_field_name +
					" = '99991231'")
			end

			l_statement.append(
				"%N) %N" +
				"OPTION (LOOP JOIN)")

			temp_table_server.sql_execute_immediate(l_statement)
			if not temp_table_server.success
			then
				raise("insert_new_data failed")
			end
		end

	delete_no_change is
		local
			l_statement : STRING
			l_fields : ROSE_LINEAR_ARRAY[STRING]
		do
			if record.is_indexed or uses_self_archive
			then
				l_statement :=
					"delete from " + data_value_temp_table +
					" from " + data_value_table + " (ROWLOCK) %N" +
					"where " +
					data_value_table + ".signature_id = " +
					data_value_temp_table + ".signature_id and %N"

				if record.is_indexed
				then
					from
						l_fields := record.key_fields
						l_fields.start
					until
						l_fields.off
					loop
						l_statement.append(
							data_value_table      + "." + l_fields.current_item + " = " +
							data_value_temp_table + "." + l_fields.current_item)
						if not l_fields.at_last
						then
							l_statement.append(" and ")
							l_statement.append("%N")
						end

						l_fields.forth
					end
				end

				if uses_self_archive
				then
					-- e.g. "data_value_table.end_date = '99991231'".
					l_statement.append(
						" and %N" + data_value_table + "." + record.archive_date_field_name +
						" = '99991231'")
				end

				if not full_update_mode
				then
					l_statement.append(" and %N")
					from
						l_fields := record.record_fields
						l_fields.start
					until
						l_fields.off
					loop
						l_statement.append(
							"(" +
							data_value_table      + "." + l_fields.current_item + " = " +
							data_value_temp_table + "." + l_fields.current_item + " or %N (" +
							data_value_table      + "." + l_fields.current_item + " is NULL and " +
							data_value_temp_table + "." + l_fields.current_item + " is NULL)" +
							")")
						if not l_fields.at_last
						then
							l_statement.append(" and ")
							l_statement.append("%N")
						end

						l_fields.forth
					end
				end

				l_statement.append("%NOPTION (LOOP JOIN)")
				temp_table_server.sql_execute_immediate(l_statement)
				if not temp_table_server.success
				then
					raise("delete no changed failed")
				end
			else
				l_fields := record.record_fields
				l_statement := "delete from " + data_value_temp_table +
					" where signature_id not in " +
					"(" +
					" select a.signature_id from " +
					"(" +
					" select signature_id"
				from l_fields.start until l_fields.off loop
					l_statement := l_statement + ",[" + l_fields.current_item + "]"
					l_fields.forth
				end
				l_statement := l_statement + " from " + data_value_temp_table +
					" UNION " +
					" select a.signature_id"
				from l_fields.start until l_fields.off loop
					l_statement := l_statement + ",a.[" + l_fields.current_item + "]"
					l_fields.forth
				end
				l_statement := l_statement + " from " + data_value_table + " a, " + data_value_temp_table + " b where " +
					" a.signature_id = b.signature_id " +
					") a " +
					" group by a.signature_id " +
					" having count(a.signature_id) > 1 " +
					")"
					temp_table_server.sql_execute_immediate(l_statement)
			end
		end

	update_data is
		local
			l_statement : STRING
			l_fields : ROSE_LINEAR_ARRAY[STRING]
			l_field_list: STRING
		do
			l_fields := record.record_fields
			l_statement :=
				"update " + data_value_table + "%N"

			if uses_self_archive
			then
				-- e.g. "set [start_date] = p.[start_time]".
				l_statement.append(
					"set [" + record.archive_date_field_name + "] = p.[start_time]")
			else
				l_statement.append(
					"set process_id = var1.process_id")

				from l_fields.start until l_fields.off loop
					l_statement.append(",%N[" + l_fields.current_item + "] = var1.[" + l_fields.current_item + "]")
					l_fields.forth
				end
			end

			l_statement.append(
				"%Nfrom " + data_value_temp_table + " as var1 (NOLOCK)")

			if uses_self_archive
			then
				l_statement.append(",%Nprocess as p")
			end

			l_statement.append(
				"%Nwhere%N" +
				"var1.signature_id = " + data_value_table + ".signature_id")

			if record.is_indexed
			then
				from
					l_fields := record.key_fields
					l_fields.start
				until
					l_fields.off
				loop
					l_statement.append(
						" and%N" +
						               "var1." + l_fields.current_item + " = " +
						data_value_table + "." + l_fields.current_item)

					l_fields.forth
				end
			end

			if uses_self_archive
			then
				-- e.g. "data_value_table.end_date = '99991231'".
				l_statement.append(
					" and%Np.process_id=var1.process_id" +
					" and%N" + data_value_table + "." + record.archive_date_field_name +
					" = '99991231'")
			end

			temp_table_server.sql_execute_immediate(l_statement)
			if not temp_table_server.success
			then
				raise("delete no changed failed (1)")
			end

			-- Insert new records if we are self-archiving.
			if uses_self_archive
			then
				l_field_list := field_list

				l_statement :=
					"insert into " + data_value_table + "%N" +
					" (" + l_field_list + ")%N" +
					"select%N" +
					"  " + l_field_list +
					"%Nfrom " + data_value_temp_table

				temp_table_server.sql_execute_immediate(l_statement)
				if not temp_table_server.success
				then
					raise("delete no changed failed (2)")
				end
			end

		end

	field_list: STRING is
		-- A comma separated list of SQL column names.  It includes the standard columns,
		-- the data columns and the optional indexing and archiving columns.
	local
		l_fields : ROSE_LINEAR_ARRAY[STRING]
	do
		Result := "signature_id,process_id"
		if is_path_clustered then
			Result.append(",path")
		end

		l_fields := record.record_fields
		from l_fields.start until l_fields.off loop
			Result.append(",[" + l_fields.current_item + "]")
			l_fields.forth
		end

		if record.is_indexed
		then
			from
				l_fields := record.key_fields
				l_fields.start
			until
				l_fields.off
			loop
				Result.append(",[" + l_fields.current_item + "]")
				l_fields.forth
			end
		end

		if uses_self_archive
		then
			Result.append(
				",[" + record.archive_date_field_name + "]")
		end
	end

	trigger_source_code_date_a: ROSE_DATE is
	do
		if trigger_source_code_date_a_saved=Void
		then
			if uses_self_archive
			then
				create trigger_source_code_date_a_saved.make(2006, 5, 13)
			else
				trigger_source_code_date_a_saved := Precursor
			end
		end

		Result := trigger_source_code_date_a_saved
	end

	trigger_source_code_date_a_saved: ROSE_DATE		-- See 'trigger_source_code_date_a'.

	trigger_source_code_date_b: ROSE_DATE is
	do
		if trigger_source_code_date_b_saved=Void
		then
			trigger_source_code_date_b_saved := Precursor
		end

		Result := trigger_source_code_date_b_saved
	end

	trigger_source_code_date_b_saved: ROSE_DATE		-- See 'trigger_source_code_date_b'.

	copy_to_history_statement: STRING is
	-- SQL code to be used when making the trigger.
	do
		if uses_self_archive
		then
			Result := ""
		else
			Result := Precursor
		end
	end

end
