indexing
	description: "Object descripting a view on signature tagged data for data bounds queries"
	author: "Mauricio Bustos"
	last_editor: "$Author: rj $"
	date: "$Date: 2012/02/24 $"
	revision: "$Revision: #12 $"
    path: "$File: //rose/source/kernel/persistence/databases/signature_tagged/signature_tagged_view.e $"

class
	SIGNATURE_TAGGED_VIEW[G -> ROSE_SIGNATURE_RECORD[ENTITY] create default_create end]

inherit
	SIGNATURE_TAGGED_NAMES
	EXCEPTIONS
		rename
			class_name as exceptions_class_name
		export
			{NONE} all
		end
	INTERNAL
		rename
			field_name as internal_field_name
		export
			{NONE} all
		end
	DATABASE_SERVER
		export
			{NONE} all
		end
	ROSE_SHARED_GLOBAL
		export
			{NONE} all
		end

create
	make

feature {NONE} -- Creation

	make (a_server: SQL_DATA_SERVER; a_data_name: STRING) is
			-- Create the view object
		require
			valid_server: a_server /= Void
		do
			signature_item_server := a_server
			data_name := a_data_name
			build_entity_id_field
		end

feature {SIGNATURE_TAGGED_SERVER} -- Implementation

	data_bounds(a_signature: G): SIGNATURE_TAGGED_DATA_BOUNDS is
			-- Data bounds for 'a_signature'.  The Result is fetched from a cache (or from the database and stored in the cache).
		require
			valid_signature: a_signature /= Void
		local
			l_data_bounds_predicate: COLLECTING_PREDICATE [SIGNATURE_TAGGED_DATA_BOUNDS]
			l_id: INTEGER
			l_null_data_bounds: SIGNATURE_TAGGED_DATA_BOUNDS
		do
			-- It is not uncommon for the predicate to be Void.  Just return Void.
			l_data_bounds_predicate := data_bounds_predicate
			if l_data_bounds_predicate /= Void
			then
				if entity_id_field_id>0
				then
					if all_data_bounds=Void
					then
						create all_data_bounds.make(1000)
					end

					l_id := integer_field(entity_id_field_id, a_signature)
					Result := all_data_bounds[l_id]
				end

				if Result=Void
				then
					l_data_bounds_predicate.retrieve (a_signature,Void)
					if not l_data_bounds_predicate.success then
						raise ("Unable to execute data bounds predicate")
					end
					if l_data_bounds_predicate.array.count > 1 then
						raise ("More than 1 entry in view")
					elseif l_data_bounds_predicate.array.count = 1 then
						Result := l_data_bounds_predicate.array.first
						repair_record(Result)

						if l_id>0
						then
							all_data_bounds[l_id] := Result
						end
					else
						if l_id>0
						then
							create l_null_data_bounds
							l_null_data_bounds.set_entity_id(-1)
							all_data_bounds[l_id] := l_null_data_bounds
						end
					end
				else
					-- Check to see if the cached record was the null record.
					if Result.entity_id=-1
					then
						Result := Void
					end
				end
			end
		end

	preload_data_bounds(a_signatures: ROSE_LINEAR_ARRAY[G])
	-- Preload the cache with the data_bounds needed for each of 'a_signatures'.  If no data is available
	-- store a null record to prevent future calls to the database.
	require
		non_void_signatures: a_signatures/=Void
	local
		l_template: G
		l_fast_loader: DATA_SERVER_FAST_LOAD
		l_spid_predicate: INTEGER_PREDICATE
		l_temp_table_name: STRING
		l_fast_load_server: SQL_DATA_SERVER
		l_preload_predicate: COLLECTING_PREDICATE[SIGNATURE_TAGGED_DATA_BOUNDS]
		l_array: ROSE_LINEAR_ARRAY[SIGNATURE_TAGGED_DATA_BOUNDS]
		l_null_data_bounds: SIGNATURE_TAGGED_DATA_BOUNDS
		l_id: INTEGER
	do
		if entity_id_field_id>0 and not a_signatures.is_empty
		then
			-- Get the SPID so we can make our temp table name unique.
			create l_spid_predicate.make(signature_item_server, "select @@SPID")
			if not l_spid_predicate.success
			then
				raise("Could not make the @@SPID predicate")
			end

			l_spid_predicate.retrieve(Void, Void)
			if not l_spid_predicate.success_one
			then
				raise("The @@SPID predicate failed")
			end

			-- The temp table name.
			l_temp_table_name := "##temp_preload_data_bounds_" + l_spid_predicate.array[1].out

			-- Kill any old version (doesn't fail if the table is not found).
			signature_item_server.drop_table(l_temp_table_name)

			-- Create the temp table.
			signature_item_server.auxiliary_server.begin_transaction
			signature_item_server.auxiliary_server.sql_execute_immediate("create table " + l_temp_table_name + " (" + entity_id_field_name + " int)")
			signature_item_server.auxiliary_server.commit

			-- Make a fast loader.
			l_fast_load_server := servers.make_from_repository_fast_load(signature_item_server.logical)
			create l_template
			l_fast_loader := l_fast_load_server.fast_loader(l_temp_table_name, l_template)

			-- Load the signatures into the fast_loader.
			from
				a_signatures.start
			until
				a_signatures.off
			loop
				l_fast_loader.insert(a_signatures.current_item)
				a_signatures.forth
			end

			-- Have the fast_loader write everything to the temp table.
			l_fast_loader.commit
			if not l_fast_loader.success
			then
				raise("Fast loader commit failed")
			end

			-- Retrieve the data_bounds records that match the ids that were loaded into the temp table.
			l_preload_predicate := preload_data_bounds_predicate(l_temp_table_name)
			l_preload_predicate.retrieve(Void, Void)
			if not l_preload_predicate.success
			then
				raise("The preload predicate failed")
			end

			if l_preload_predicate.array.count/=0
			then
				-- Load the data_bounds records into the cache.
				from
					l_array := l_preload_predicate.array
					l_array.start
					if all_data_bounds=Void
					then
						create all_data_bounds.make(l_array.count)
					end
				until
					l_array.off
				loop
					repair_record(l_array.current_item)
					all_data_bounds[l_array.current_item.entity_id] := l_array.current_item

					l_array.forth
				end

				-- Put in null records for those signatures that have no data_bounds records.
				from
					create l_null_data_bounds
					l_null_data_bounds.set_entity_id(-1)
					a_signatures.start
				until
					a_signatures.off
				loop
					l_id := integer_field(entity_id_field_id, a_signatures.current_item)
					if not all_data_bounds.has(l_id)
					then
						all_data_bounds[l_id] := l_null_data_bounds
					end

					a_signatures.forth
				end
			end
		end
	end

	repair_record(a_data_bounds: SIGNATURE_TAGGED_DATA_BOUNDS)
	-- In some of our databases, the dates are really integer months.  Oddly, this loads INTEGER_REF's
	-- into ROSE_DATE fields.  This hack changes them back into ROSE_DATE's.  Yuck!
	local
		l_month: ROSE_DATE
	do
		if not attached {ROSE_DATE} a_data_bounds.start_date as l_date then
			if attached {INTEGER_32_REF} a_data_bounds.start_date as l_integer then
				create l_month.make_from_julian_month (l_integer.item)
				if l_month.is_available then
					a_data_bounds.set_start_date (l_month)
				else
					a_data_bounds.set_start_date(Void)
				end
			end

			if attached {INTEGER_32_REF} a_data_bounds.end_date as l_integer then
				create l_month.make_from_julian_month (l_integer.item)
				if l_month.is_available then
					a_data_bounds.set_end_date (l_month)
				else
					a_data_bounds.set_end_date(Void)
				end
			end
		end
	end

	all_data_bounds: HASH_TABLE[SIGNATURE_TAGGED_DATA_BOUNDS, INTEGER]

	create_view is
			-- Create the view if needed
		local
			l_data_server: SQL_DATA_SERVER
			l_statement, l_min_date_string, l_max_date_string, l_where_clause: STRING
			l_retry: BOOLEAN
			l_signature: G
		do
			if not view_created and view_crdate.ccyymmdd < 20031120 and not l_retry then
				create l_signature.default_create
				if signature_table_has_column (l_signature.date_name) and
					signature_table_has_column (entity_id_field_name) then
					view_available := True
					l_data_server := servers.new_server_from_repository (signature_item_server.logical)
					if has_view then
						l_data_server.sql_execute_immediate
						("DROP VIEW " + data_name_prefix + "data_signatures_data_bounds")
						l_data_server.commit
					end
					if l_signature.field_type (l_signature.date_name) = {INTERNAL}.integer_32_type then
						l_min_date_string := "DATEADD (MONTH, MIN (dbo." + data_name_prefix + "data_signatures." + l_signature.date_name + "), '19491231')"
						l_max_date_string := "DATEADD (MONTH, MAX (dbo." + data_name_prefix + "data_signatures." + l_signature.date_name + "), '19491231')"
					else
						l_min_date_string := "MIN (dbo." + data_name_prefix + "data_signatures." + l_signature.date_name + ")"
						l_max_date_string := "MAX (dbo." + data_name_prefix + "data_signatures." + l_signature.date_name + ")"
					end
					if l_signature.data_fields_valid_values_statement /= Void then
						l_where_clause := "p.process_id IN (dbo." + data_name_prefix + "data_signatures.first_process_id, dbo." + data_name_prefix + "data_signatures.last_process_id,%N " +
							"        dbo." + data_name_prefix + "data_signatures.continuing_process_id) and " + l_signature.data_fields_valid_values_statement
					else
						l_where_clause := "p.process_id IN (dbo." + data_name_prefix + "data_signatures.first_process_id, dbo." + data_name_prefix + "data_signatures.last_process_id,%N " +
							"        dbo." + data_name_prefix + "data_signatures.continuing_process_id)"
					end
					l_statement := "CREATE VIEW dbo." + data_name_prefix + "data_signatures_data_bounds as%N" +
						"SELECT " + l_min_date_string + " AS start_date,%N " +
						"       " + l_max_date_string + " AS end_date,%N " +
						"       MIN (p.start_time) AS start_process_date, MAX (p.start_time) AS end_process_date,%N " +
						"       dbo." + data_name_prefix + "data_signatures." + entity_id_field_name + "%N " +
						"FROM   dbo." + data_name_prefix + "data_signatures CROSS JOIN%N " +
						"       dbo.process p%N " +
						"WHERE  " + l_where_clause + "%N " +
						"GROUP BY dbo." + data_name_prefix + "data_signatures." + entity_id_field_name + "%N "
					l_data_server.sql_execute_immediate (l_statement)
					if not l_data_server.success then
						raise ("Could not create view.")
					end
					l_data_server.commit
					if not l_data_server.success then
						raise ("Commit failed.")
					end
				end
			end
			view_created := True
		rescue
			l_retry := True
			view_created := True
			retry
		end

feature {NONE} -- Implementation

	signature_item_server: SQL_DATA_SERVER
			-- Data server for the view
	view_created: BOOLEAN
			-- Was the view created?
	view_available: BOOLEAN
			-- Is the view available?

	entity_id_field_name: STRING
			-- The name of the attribute in G that represents the 'entity_id'.

	entity_id_field_id: INTEGER
			-- The 'field_id' of the attribute in G that represents the 'entity_id'.  Used by INTERNAL.

	build_entity_id_field
	-- Determines both 'entity_id_field_name' and 'entity_id_field_id'.
	local
		l_g: G
		l_count: INTEGER
		i: INTEGER
		j: INTEGER
	do
		create l_g
		l_count := field_count(l_g)

		from
			j := 1
		until
			j>2 or entity_id_field_id/=0
		loop
			inspect
				j
			when 1 then
				entity_id_field_name := l_g.entity_name + "_id"
			when 2 then
				entity_id_field_name := l_g.entity_name
			end

			from
				i := 1
			until
				entity_id_field_id/=0 or i>l_count
			loop
				if internal_field_name(i, l_g).is_case_insensitive_equal(entity_id_field_name)
				then
					entity_id_field_id := i
				else
					i := i + 1
				end
			end

			j := j + 1
		end

		if entity_id_field_id=0
		then
			entity_id_field_id := -1
			entity_id_field_name := Void
		end
	end

	view_crdate: ROSE_DATE is
			-- Creation date for the existing view
		local
			l_predicate: DATE_PREDICATE
		do
			create l_predicate.make (signature_item_server,"select crdate from dbo.sysobjects " +
									" where name = '" + data_name_prefix + "data_signatures_data_bounds'")
				l_predicate.retrieve (Void,Void)
			if not l_predicate.success then
				raise("Retreive failed.")
			end
			if l_predicate.array.count > 0 then
				Result := l_predicate.array.item(1)
				view_available := True
			else
				create Result.make_ccyymmdd(19000101)
			end
		ensure
			available_result: Result /= Void
		end

	has_view: BOOLEAN is
			-- Does the database contain this view?
		local
			l_predicate: ANONYMOUS_PREDICATE
		do
			create l_predicate.make_using_server (signature_item_server,
												 "select * from dbo.sysobjects " +
												 " where " +
												 " id = object_id (N'[dbo].[" + data_name_prefix + "data_signatures_data_bounds]')")
			l_predicate.retrieve (Void,Void)
			if not l_predicate.success then
				raise ("Failed the retrieve from sysobjects.")
			end
			if not l_predicate.array.is_empty then
				Result := True
			end
		end

	signature_table_has_column (a_column: STRING): BOOLEAN is
			-- Does this signature table have 'a_column'?
		require
			valid_column_name: a_column /= Void and then not a_column.is_empty
		local
			l_predicate: ANONYMOUS_PREDICATE
		do
			create l_predicate.make_using_server (signature_item_server,
												 "exec sp_columns @table_name = '" + data_name_prefix + "data_signatures',@column_name = '" + a_column + "'")
			l_predicate.retrieve (Void,Void)
			if not l_predicate.array.is_empty then
				Result := True
			end
		end

	data_bounds_predicate: COLLECTING_PREDICATE[SIGNATURE_TAGGED_DATA_BOUNDS] is
			-- Predicate for retrieving from the view
		local
			l_signature: G
		do
			if saved_data_bounds_predicate = Void then
				create_view
				if view_available then
					create l_signature.default_create
					create saved_data_bounds_predicate.make (signature_item_server,"select * from dbo." + data_name_prefix + "data_signatures_data_bounds " +
															 " where " + entity_id_field_name + " = :" + entity_id_field_name)
					if not saved_data_bounds_predicate.success then
						raise ("Could not create the predicate")
					end
				end
			end

			Result := saved_data_bounds_predicate
		end

	saved_data_bounds_predicate: COLLECTING_PREDICATE [SIGNATURE_TAGGED_DATA_BOUNDS]

	preload_data_bounds_predicate(a_temp_table_name: STRING): COLLECTING_PREDICATE[SIGNATURE_TAGGED_DATA_BOUNDS]
	-- Predicate for preloading data from the view.  A new Result is returned everytime so only call this once.
	local
		l_signature: G
		l_statement: STRING
	do
		create_view
		if view_available
		then
			create l_signature.default_create
			l_statement := global.formatted(
							"[
							select *, $1 as entity_id
							from $2
							where $1 in
							(
								select $1 from $3
							)
							]",
							<<entity_id_field_name, "dbo." + data_name_prefix + "data_signatures_data_bounds", a_temp_table_name>>)
			create Result.make(signature_item_server, l_statement)
			if not Result.success then
				raise ("Could not create the predicate")
			end
		end
	end

end
