indexing
	description: "SMTP client"
	author: "Zoran Simic"
	last_editor: "$Author: zs $"
	date: "$Date: 2010/02/17 $"
	revision: "$Revision: #26 $"
	path: "$File: //rose/source/kernel/infrastructure/framework/io/smtp_client.e $"

class
	SMTP_CLIENT

inherit
	DEBUG_OUTPUT
		redefine
			default_create
		end

feature {NONE} -- Default create

	default_create is
			-- Default creation procedure
		do
			server := "localhost"
			port := 25
			host := server
		end

feature -- Access

	server: STRING
			-- Address of SMPT server to use

	port: INTEGER
			-- Port number the SMTP server runs on (standard is 25)

	last_error: STRING
			-- Last error message if any

	host: STRING
			-- Host name sending this mail

	timeout: INTEGER
			-- Timeout in seconds for connection to SMTP server

	is_connected: BOOLEAN is
			-- Is this object currently connected to 'server':'port'?
		do
			Result := socket /= Void and then socket.is_open_read and then socket.is_open_write
		end

	is_configured: BOOLEAN is
			-- Is this SMTP client properly configured?
		do
			Result := server /= Void and then port > 0
		end

	checked_contents (a_content: ROSE_MAIL_CONTENT): STRING is
			-- Check 'a_content' a description of an eventual problem with it
			-- Returns Void if 'a_content' is OK (ie sendable)
		require
			content_ok: a_content /= Void
		local
			l_string: STRING
		do
			l_string := last_error
			last_error := Void
			if not is_configured then
				report ("SMTP is not configured, can not send email via SMTP")
			end
			if last_error = Void then
				if not is_valid_for_smtp (a_content.sender) then
					report ("Invalid mail sender '" + represented_object (a_content.sender) + "', sender is required for SMTP")
				end
				if last_error = Void then
					if a_content.to = Void or else a_content.to.is_empty then
						report ("No recipients provided")
					end
				end
			end
			Result := last_error
			last_error := l_string
		end

	debug_output: STRING is
			-- String that should be displayed in debugger to represent 'Current'
		do
			create Result.make_empty
			Result.append_string (server)
			Result.append_character (':')
			Result.append_integer (port)
		end

feature -- Basic operations

	send (a_content: ROSE_MAIL_CONTENT) is
			-- Send email described by 'a_content'.
		require
			valid_content: a_content /= Void
		local
			l_string, l_body: STRING
			l_char: CHARACTER
			i: INTEGER
		do
			last_error := checked_contents (a_content)
			if last_error = Void then
				from
					-- Make sure all line endings are %R%N
					l_string := a_content.body_smtp
					create l_body.make (l_string.count)
					i := 1
				until
					i > l_string.count
				loop
					l_char := l_string.item (i)
					if l_char = '%R' then
						-- Do nothing
					elseif l_char = '%N' then
						l_body.extend ('%R')
						l_body.extend ('%N')
					else
						l_body.extend (l_char)
					end
					i := i + 1
				end
				connect_server
			end
			if last_error = Void then
				capabilities := get_response ("EHLO", host)
			end
			if last_error = Void then
				checked_send_message ("MAIL FROM:", a_content.sender.address)
			end
			if last_error = Void then
				send_recipients (a_content.to)
				send_recipients (a_content.cc)
				send_recipients (a_content.bcc)
				if a_content.is_sender_copied then
					send_rcpt (a_content.sender)
				end
				checked_send_message ("DATA", Void)
				a_content.headers.do_all (agent send_header)
				if l_body /= Void then
					send_message (Void, l_body)
				end
				send_message ("", Void)
				checked_send_message (".", Void)
				checked_send_message ("QUIT", Void)
			end
			disconnect_server
		end

	set_server (a_server: STRING) is
			-- Set 'server'.
		require
			server_ok: a_server /= Void and then not a_server.is_empty
		do
			server := a_server
		end

	set_port (a_port: INTEGER) is
			-- Set 'port'.
		do
			port := a_port
			if port < 1 or else port > 65535 then
				port := 25
			end
		ensure
			port_ok: port > 0 and then port < 65535
			port_set: (a_port > 0 and then a_port < 65535) implies port = a_port
		end

	set_host (a_host: STRING) is
			-- Set 'host'.
		require
			host_ok: a_host /= Void and then not a_host.is_empty
		do
			host := a_host
		end

	set_timeout (a_timeout: INTEGER) is
			-- Set 'timeout'.
		do
			timeout := a_timeout.max (300)
		ensure
			timeout_ok: timeout >= 300
			timeout_set: a_timeout >= 300 implies timeout = a_timeout
		end

feature {NONE} -- Implementation

	socket: NETWORK_STREAM_SOCKET
			-- Socket used to communicate with remote SMTP server

	capabilities: STRING
			-- Server stated capabilities (available after sending EHLO command)

	is_valid_for_smtp (a_address: ROSE_MAIL_ADDRESS): BOOLEAN is
			-- Is 'a_address' valid for SMTP?
		do
			Result := rcpt_address (a_address) /= Void
		end

	rcpt_address (a_address: ROSE_MAIL_ADDRESS): STRING is
			-- RCPT form of 'a_address'
			-- if `a_address.address' is void then use `a_address.name' + "@axarosenberg.com"
		do
			if a_address /= Void then
				Result := a_address.address
				if Result = Void and then a_address.is_valid_email_username (a_address.name) then
					Result := a_address.name + "@axarosenberg.com"
				end
			end
		end

	connect_server is
			-- Connect to 'server'.
		local
			retried: BOOLEAN
		do
			if not retried then
				create socket.make_client_by_port (port, server)
				socket.connect
			end
			if retried or else not is_connected then
				report ("Can not connect to server")
			else
				socket.set_timeout (timeout)
				check_response ("connect", Void, read_response)
			end
		rescue
			if not retried then
				retried := True
				retry
			end
		end

	disconnect_server is
			-- Disconnect from 'server'.
		do
			if socket /= Void and then not socket.is_closed then
				socket.close
			end
			socket := Void
		ensure
			no_socket: socket = Void
		end

	send_recipients (a_recipients: ROSE_MAIL_RECIPIENTS) is
			-- Send 'a_recipients'
		local
			i: INTEGER
		do
			if a_recipients /= Void and then not a_recipients.is_empty then
				from
					i := a_recipients.recipients.lower
				until
					i > a_recipients.recipients.upper
				loop
					send_rcpt (a_recipients.recipients.item (i))
					i := i + 1
				end
			end
		end

	send_rcpt (a_address: ROSE_MAIL_ADDRESS) is
			-- Send RCPT command for 'a_address'
		local
			l_rcpt: STRING
		do
			l_rcpt := rcpt_address (a_address)
			if l_rcpt /= Void then
				checked_send_message ("RCPT TO:", l_rcpt)
			end
		end

	get_response (a_command, a_message: STRING): STRING is
			-- Send 'a_command' with 'a_message' and return server response
		require
			valid_message: a_command /= Void or else a_message /= Void
		do
			send_message (a_command, a_message)
			Result := read_response
		ensure
			result_ok: Result = Void implies last_error /= Void
		end

	checked_send_message (a_command, a_message: STRING) is
			-- Send 'a_command' with 'a_message' and verify response
		do
			send_message (a_command, a_message)
			check_response (a_command, a_message, read_response)
		end

	send_message (a_command, a_message: STRING) is
			-- Send 'a_command' with 'a_message'.
		require
			valid_message: a_command /= Void or else a_message /= Void
		local
			retried: BOOLEAN
			l_string: STRING
		do
			if retried then
				if last_error = Void then
					report ("Exception raised when sending '" + represented_object (l_string) + "'")
				end
			else
				if last_error = Void then
					if socket.ready_for_writing then
						create l_string.make_empty
						if a_command /= Void and then not a_command.is_empty then
							l_string.append (a_command)
						end
						if a_message /= Void and then not a_message.is_empty then
							if a_command /= Void and then not a_command.is_empty then
								l_string.append_character (' ')
							end
							l_string.append (a_message)
						end
						put_string (l_string)
					else
						report ("Server timed out on " + error_message (a_command, a_message))
					end
				end
			end
		rescue
			if not retried then
				retried := True
				retry
			end
		end

	send_header (a_key, a_value: STRING) is
			-- Send header with 'a_key' and 'a_value'.
		do
			if a_key /= Void and then a_value /= Void then
				send_message (a_key + ":", a_value)
			end
		end

	report (a_message: STRING) is
			-- Report error with 'a_message'.
		require
			message_ok: a_message /= Void
		do
			if last_error = Void then
				if server /= Void then
					last_error := "Error while accessing SMTP server " + represented_object (server) + ":" + port.out + "%N"
				else
					create last_error.make (100)
				end
			else
				last_error.append_character ('%N')
			end
			last_error.append (a_message)
		end

	represented_object (a_object: ANY): STRING is
			-- Textual representation of 'a_object'.
		do
			if a_object /= Void then
				Result := a_object.out
			end
			if Result = Void then
				create Result.make_empty
			end
		end

	check_response (a_command, a_message, a_response: STRING) is
			-- Check 'a_response' for validity
		do
			if a_response = Void or else a_response.is_empty then
				if last_error = Void then
					report ("SMTP server does not reply to " + error_message (a_command, a_message))
				end
			elseif a_response.item (1) = '4' or else a_response.item (1) = '5'  then
				a_response.right_adjust
				report ("SMTP server replied: '" + a_response + "'%Nto " + error_message (a_command, a_message))
			end
		end

	error_message (a_command, a_message: STRING): STRING is
			-- Error message generated from `a_command' and `a_message'
		do
			create Result.make_empty
			if a_command /= Void then
				Result.append_string ("command '")
				Result.append_string (a_command)
				Result.append_character ('%'')
				if a_message /= Void then
					Result.append_character (',')
					Result.append_character (' ')
				end
			end
			if a_message /= Void then
				Result.append_string ("message '")
				Result.append_string (a_message)
				Result.append_character ('%'')
			end
		ensure
			error_msg_not_void: Result /= Void
		end

	read_response: STRING is
			-- Get response from connected 'socket', Void if could not read
		require
			socket_ok: socket /= Void
		local
			l_string: STRING
			retried, l_stop: BOOLEAN
		do
			if retried then
				report ("Exception raised while receiving response from '" + represented_object (server) + "'")
			else
				from
					l_stop := last_error /= Void
				until
					retried or else l_stop or else not socket.ready_for_reading
				loop
					socket.read_line
					if Result = Void then
						create Result.make_empty
					end
					l_string := socket.last_string
					Result.append_string (l_string)
					l_stop := l_string = Void or else l_string.count < 4 or else l_string.item (4) /= '-'
				end
			end
		rescue
			if not retried then
				retried := True
				retry
			end
		end

	put_string (a_string: STRING) is
			-- Send 'a_string' on 'socket'
		require
			string_ok: a_string /= Void
		do
			if not a_string.is_empty then
				socket.put_string (a_string)
			end
			if a_string.count < 2
			or else (a_string.item (a_string.count - 1) /= '%R' and then a_string.item (a_string.count) /= '%N') then
				socket.put_string ("%R%N")
			end
		end

invariant
	valid_server: server /= Void
	valid_host: host /= Void

end
