indexing
	description: "Abstraction of a SQL dedicated data-server"
	author: "Mark Howard"
	last_editor: "$Author: an $"
	date: "$Date: 2013/10/18 $"
	revision: "$Revision: #46 $"
	path: "$File: //rose/source/kernel/infrastructure/persistence/databases/sql/sql_data_server.e $"

deferred class
	SQL_DATA_SERVER

inherit
	DATA_SERVER
		rename
			set_status as data_server_set_status,
			disconnect as sql_disconnect
		undefine
			sql_disconnect
		end
	SQL_CONSTANTS

	SHARED_SQL_TABLE_NAME_PARSER
		export {NONE} all end

feature {NONE} -- Initialization

	sql_make (a_node, a_name, a_passwd, a_database, a_tag: STRING) is
			-- Initialize object with arguments
		do
			node := a_node
			name := a_name
			database := a_database
			tag := a_tag
		ensure
			node_set: node = a_node
			name_set: name = a_name
			database_set: database = a_database
			tag_set: tag = a_tag
		end

	make (a_node, a_name, a_passwd, a_database, a_tag: STRING) is
		deferred
		end

feature {PERSISTENT_SERVER} -- Initialization

	make_with_server_and_database (a_server, a_database: STRING) is
			-- Create the server using 'a_server' and 'a_database'
			-- This is a general purpose creation feature that should
			-- be redefined.  It is not applicable at this level.
		require
			valid_server: a_server /= Void
			valid_database: a_database /= Void
		do
			error.raise ("'make_with_server_and_database' is not implemented for SQL_DATA_SERVER")
		end

feature -- Access

	new_executor (a_statement: STRING; a_parameter_template: ARRAY [STRING]): SQL_EXECUTOR is
			-- Create an executor for this server using a_statement and a_parameter_template
		do
				--| Parameter templates are not defined at this level.
				--| Some redefined version of this routine may use them.
			if not playback_movie then
				create Result.make (Current, a_statement)
			end
		end

	new_selector (a_statement: STRING; a_parameter_template: ARRAY [STRING]): SQL_SELECTOR is
			-- Create a selector for this server using a_statement and a_parameter_template
		deferred
		end

	new_sqlvar (a_sqlda: SQL_SQLDA; a_index: INTEGER): SQL_SQLVAR is
			-- Returns a new SQL_SQLVAR that represents the 'a_index'th variable in 'a_sqlda'.
			-- The return value will of course be of a type that is suitable for use by the descendant.
		deferred
		end

	error_handler: SQL_ERROR_HANDLER		-- This object is used to handle any SQL error that occurs.

	new_error_handler_print: SQL_ERROR_HANDLER is
			-- s.b. SQL_ERROR_HANDLER_PRINT??
		do
		end

feature -- Status report

	connected: BOOLEAN is
			-- Is this session connected?
		do
			Result := make_succeeded
		end

	connected_by_testing: BOOLEAN is
			-- Is this session connected (using a transaction test)?
			-- FIXME: This is not the best way to do this.  There should be a way to query the connection directly for connection status.
		do
			begin_transaction
			if success
			then
				commit
				Result := success
			end
		end

	version: STRING is
			-- Version for the sql data provider
		do
			Result := "unknown"
		ensure
			valid_result: Result /= Void
		end

	eof: BOOLEAN is
			-- Are we at the end of file for this query?
		deferred
		end

	is_error_negligible (a_rose_sql_error_code: INTEGER): BOOLEAN is
			-- Do we ignore this error?
		require
			can_ignore_error: can_ignore_rose_sql_error (a_rose_sql_error_code)
		do
			error.raise ("is_error_negligible is not implemented.")
		end

	can_ignore_rose_sql_error (a_rose_sql_error_code: INTEGER): BOOLEAN is
			-- Does this server (or its error_handler) know how to ignore this error?
		do
			Result := error_handler /= Void and then error_handler.can_ignore_rose_sql_error (a_rose_sql_error_code)
		end

	transaction_depth: INTEGER is
			-- The current transaction depth count for this session
		do
		end

	implicit_transaction: BOOLEAN is
			-- is transaction implicit (redefined as necessary by descendant)?
		do
		end

	explicit_transaction_depth: INTEGER is
			-- Depth of transactions caused by 'begin transaction' (redefined as necessary by descendant).
		do
		end

	implicit_transactions_set: BOOLEAN is
			-- True if set_implicit_transaction_on or _off has been called (redefined as necessary by descendant).
		do
		end

	set_implicit_transactions_set(a_value: like implicit_transactions_set)
	-- This is redefined in the descendant at which time 'implicit_transactions_set' will be redefined to an attribute.
	-- This feature will set 'implicit_transactions_set'.
	do
	end

feature -- Basic operations

	sql_disconnect is
			-- Free up the resources that are used by the underlying connection.
		require else
			connected: connected
		deferred
		end

	set_status (a_traceable: TRACEABLE) is
		deferred
		end

	sql_error (a_traceable: TRACEABLE) is
		deferred
		end

	execute_and_commit (a_statement: STRING) is
			-- execute `a_statement' and commit transaction
		require
			connected: connected
		do
			sql_execute_immediate (a_statement)
			commit
		end

	set_implicit_transaction_off is
		do
		end

	set_implicit_transaction_on is
		do
		end

	commit_if_necessary is
		do
		end

	set_single_user is
		do
		end

	set_multi_user is
		do
		end

	set_read_only
			-- Changes our access to read-only (not available in all descendants).
		do
		end

	set_read_write is
			-- Changes our access to read-only (not available in all descendants).
		do
		end

	execute (a_statement: SQL_STATEMENT; a_obj: ANY; a_args: ARRAY [ANY]) is
		require
			connected: connected
		deferred
		end

	sql_execute_immediate (a_statement: STRING) is
			-- Immediately executes an SQL statement.
		require
			connected: connected
			statement_is_not_feature: not statement_is_feature (a_statement)
		deferred
		end

	statement_is_feature (a_statement: STRING): BOOLEAN is
		do
				-- To be redefined in the descendants.
		end

	set_error_handler (e: like error_handler) is
		do
			error_handler := e
		end

	allow_error (a_rose_sql_error_code: INTEGER) is
			-- Make this error a negligible error.
		require
			can_ignore_error: can_ignore_rose_sql_error (a_rose_sql_error_code)
		local
			l_exceptions: EXCEPTIONS
		do
			create l_exceptions
			l_exceptions.raise ("allow_error is not implemented.")
		end

	dont_allow_error (a_rose_sql_error_code: INTEGER) is
			-- Don't ignore this error.
		require
			can_ignore_error: can_ignore_rose_sql_error (a_rose_sql_error_code)
		local
			l_exceptions: EXCEPTIONS
		do
			create l_exceptions
			l_exceptions.raise ("dont_allow_error is not implemented.")
		end

	error_was_allowed(a_rose_sql_error_code: INTEGER): BOOLEAN
	-- True if the error occurred but was 'allowed'.  Be sure to also call 'clear_last_allowed_error'.
	local
		l_exceptions: EXCEPTIONS
	do
		create l_exceptions
		l_exceptions.raise ("error_was_allowed is not implemented.")
	end

	clear_last_allowed_error
	-- Called in conjunction with 'error_was_allowed'.
	local
		l_exceptions: EXCEPTIONS
	do
		create l_exceptions
		l_exceptions.raise ("clear_last_allowed_error is not implemented.")
	end

	begin_transaction is
			-- Starts a DB transaction
		do
		end

feature -- Schema

	table_names (a_pattern: STRING): ROSE_LINEAR_ARRAY [STRING] is
			-- Get list of tables that match `a_pattern' (SQL LIKE wildcards)
			-- Result empty if no match or query fails.
		require
			a_pattern_not_void: a_pattern /= Void
			a_pattern_no_quotes: not a_pattern.has ('%'')
		deferred
		ensure
			not_void: Result /= Void
		end

feature -- Utility

	copy_table (a_from_table: STRING; a_to: SQL_DATA_SERVER; a_to_table: STRING) is
			-- Copy table content by deleting target and inserting all
			-- source rows in a table into nto target.
		require
			a_from_table_not_void: a_from_table /= Void
			a_to_not_void: a_to /= Void
			a_table_not_void: a_to_table /= Void
				--same_schema: table_schema (a_from_table).is_equal (a_to.table_schema (a_to_table))
		local
			l_columns: INTEGER
		do
			l_columns := columns_count (a_from_table)
			if l_columns > 0 then
				a_to.delete_table (a_to_table)
				append_table (a_from_table, a_to, a_to_table, l_columns)
			else
					--	error.warn ("Table $1 empty?", << a_table >>)
			end
		end

	delete_table (a_table: STRING) is
			-- Delete all rows in a table.
		require
			a_table_not_void: a_table /= Void
		local
			l_delete: EXECUTION_PREDICATE
		do
			create l_delete.make (Current, "DELETE FROM " + a_table)
			if not l_delete.success then
				raise ("Can't create predicate")
			end
			l_delete.execute (Void, Void)
			if not l_delete.success then
				raise ("Can't execute deletion predicate")
			end
		end

	tables_are_equal (a_table_name1: STRING; a_server: SQL_DATA_SERVER; a_table_name2: STRING; a_column_names_to_ignore: ROSE_LINEAR_ARRAY [STRING]): BOOLEAN is
			-- Returns 'True' if the contents of the two tables are the same.  The method is up to the ancestor.
			-- Fill in 'a_column_names_to_ignore' with columns that you don't want to test (usually date/time columns).
		require
			valid_table_name1: a_table_name1 /= Void and then a_table_name1.count > 0
			valid_server: a_server /= Void
			valid_table_name2: a_table_name2 /= Void and then a_table_name2.count > 0
		do
			error.raise ("Not implemented.")
		end

	has_table (a_table_name: STRING): BOOLEAN is
			-- Does the current database have 'a_table_name'?
		require
			valid_table_name: a_table_name /= Void and then a_table_name.count > 0
		do
			error.raise ("Not implemented.")
		end

	drop_table (a_table_name: STRING) is
			-- Completely removes (kills) a table.  There is no error if the table does not exist.
			-- Contrast with 'delete_table'.
		require
			valid_table_name: a_table_name /= Void and then a_table_name.count > 0
		do
			error.raise ("Not implemented.")
		end

	create_duplicate_table (a_old_table_name, a_new_table_name: STRING) is
			-- Makes a new table whose columns are identical to an existing table.
		require
			valid_old_table_name: a_old_table_name /= Void and then a_old_table_name.count > 0
			valid_new_table_name: a_new_table_name /= Void and then a_new_table_name.count > 0
		do
			error.raise ("Not implemented.")
		end

	column_names (a_table_name: STRING): ROSE_LINEAR_ARRAY [STRING] is
			-- Returns the names of all of the columns in a table (in lowercase).
		require
			valid_table_name: a_table_name /= Void and then a_table_name.count > 0
		do
			error.raise ("Not implemented.")
		end

	has_column_name (a_table_name, a_column_name: STRING): BOOLEAN is
			-- Returns True if 'a_table_name' contains 'a_column_name'.
		require
			valid_table_name: a_table_name /= Void and then a_table_name.count > 0
			valid_column_name: a_column_name /= Void and then a_column_name.count > 0
		do
			error.raise ("Not implemented.")
		end

	has_trigger (a_table_name, a_trigger_name: STRING): BOOLEAN is
			-- Does a table have a trigger?
		require
			a_trigger_name_not_void: a_trigger_name /= Void
		do
			error.raise ("Not implemented.")
		end

	trigger_creation_date (a_table_name, a_trigger_name: STRING): ROSE_DATE is
			-- Returns the creation date of a trigger.
		require
			table_name_not_void: a_table_name /= Void
			a_trigger_name_not_void: a_trigger_name /= Void
		do
			error.raise ("Not implemented.")
		end

	auxiliary_server: SQL_DATA_SERVER is
			-- A server to use when you don't want to use the current connection.
			-- The use of this server is for little things.  You transaction should be short and sweet.
			-- Clients of this feature are expected to leave the transaction depth at 0 by calling
			-- 'begin_transaction' and 'commit' within the client routine.
		do
			error.raise ("not implemented")
		end

	has_read_permission (a_table_name: STRING): BOOLEAN is
			-- Does current user have read permission on table `a_table_name'?
		require
			a_table_name_not_void: a_table_name /= Void
		local
			l_integer_predicate: INTEGER_PREDICATE
		do
			create l_integer_predicate.make (Current, "select HAS_PERMS_BY_NAME ('" + a_table_name + "', 'OBJECT', 'SELECT')")
			l_integer_predicate.retrieve (Void, Void)
			if l_integer_predicate.success then
				Result := l_integer_predicate.array.first.to_boolean
			else
				raise ("Unable to query read permission for " + a_table_name)
			end
		end

	has_write_permission (a_table_name: STRING): BOOLEAN is
			-- Does current user have write permission on table `a_table_name'?
		require
			a_table_name_not_void: a_table_name /= Void
		local
			l_integer_predicate: INTEGER_PREDICATE
		do
			create l_integer_predicate.make (Current, "select HAS_PERMS_BY_NAME ('" + a_table_name + "', 'OBJECT', 'INSERT')")
			l_integer_predicate.retrieve (Void, Void)
			if l_integer_predicate.success then
				Result := l_integer_predicate.array.first.to_boolean
			else
				raise ("Unable to query write permission for " + a_table_name)
			end
		end

feature {SQL_DATA_SERVER} -- Implementation

	columns_count (a_table: STRING): INTEGER is
			-- Number of columns in a table.
			-- (if table has at least 1 row)
		require
			a_table_not_void: a_table /= Void
		local
			l_select: ANONYMOUS_PREDICATE
		do
			create l_select.make_using_server (Current, "SELECT TOP 1 * FROM " + a_table)
			if not l_select.success then
				raise ("Can't create predicate for columns_count")
			end
			l_select.retrieve (Void, Void)
			if l_select.success and not l_select.array.is_empty then
				Result := l_select.array.first.count
			end
		ensure
				--empty_case: rows (a_table) = 0 implies Result = 0
			positive: Result >= 0
		end

	append_table (a_from_table: STRING; a_to: SQL_DATA_SERVER; a_to_table: STRING; a_columns: INTEGER) is
			-- Append the content of `a_from_table' to `a_to_table' in `a_to' database.
			-- Both tables must have identical schema!
		require
			a_from_table_not_void: a_from_table /= Void
			a_to_not_void: a_to /= Void
			a_table_not_void: a_to_table /= Void
				--same_schema: table_schema (a_from_table).is_equal (a_to.table_schema (a_to_table))
			columns_consistent_from: a_columns = columns_count (a_from_table)
			columns_consistent_to: a_to.columns_count (a_to_table) = a_columns or a_to.columns_count (a_to_table) = 0
			a_columns_positive: a_columns > 0
		local
			l_select: ANONYMOUS_PREDICATE
			l_insert: EXECUTION_PREDICATE
		do
			create l_select.make_using_server (Current, "SELECT * FROM " + a_from_table)
			if not l_select.success then
				raise ("Can't create select predicate in SQL_DATA_SERVER.append_table")
			end
			create l_insert.make (a_to, "INSERT INTO " + a_to_table + " VALUES (" + repeat_question (a_columns) + ")")
			if not l_insert.success then
				raise ("Can't create insert predicate in SQL_DATA_SERVER.append_table")
			end
			l_select.apply (agent l_insert.execute (Void, ?), Void, Void)
		end

feature {NONE} -- Implementation for append_table

	repeat_question (a_count: INTEGER): STRING is
			-- Repeat question mark for parameters.
		require
			a_count_positive: a_count > 0
		do
			create Result.make_empty;
			(1 |..| a_count).as_array.do_all (agent append_question (Result, ?))
		ensure
			not_void: Result /= Void
			size: Result.count = 2 * a_count - 1
		end

	append_question (a_result: STRING; a_index: INTEGER) is
			-- Add question mark
		require
			a_result_not_void: a_result /= Void
		do
			if a_index > 1 then
				a_result.append_character (',')
			end
			a_result.append_character ('?')
		end

feature {SQL_SQLDA} -- Implementation

	allocate_sqlda_data (a_sqlda: SQL_SQLDA) is
		deferred
		end

	setup_sqlda (a_sqlda: SQL_SQLDA) is
		local
			l_sqlvar_structure: SQL_DATA_STRUCTURE
			i, l_count: INTEGER
		do
			allocate_sqlda_data (a_sqlda)
			l_count := sqlda_sqld (a_sqlda.id)
			create l_sqlvar_structure.make (1, l_count)
			from
				i := 1
			until
				i > l_count
			loop
				l_sqlvar_structure.put (new_sqlvar (a_sqlda, i - 1), i)
				i := i + 1
			end
			a_sqlda.set_data_structure (l_sqlvar_structure)
		end

	sqlda_sqld (a_sqlda: POINTER): INTEGER is
		deferred
		end

feature {SQL_STATEMENT} -- Implementation

	prepare (a_statement: SQL_STATEMENT) is
		require
			connected: connected
		deferred
		end

	release_statement (a_statement: SQL_STATEMENT) is
		deferred
		end

feature {SQL_CURSOR} -- Implementation

	fetch (a_cursor: SQL_CURSOR) is
		require
			connected: connected
			not_a_movie: not playback_movie
		deferred
		end

	declare_cursor (a_cursor: SQL_CURSOR) is
		require
			connected: connected
		deferred
		end

	close_cursor (a_cursor: SQL_CURSOR) is
		deferred
		end

	open_cursor (a_cursor: SQL_CURSOR) is
		require
			connected: connected
		deferred
		end

	prefetch (a_cursor: SQL_CURSOR; a_count: INTEGER) is
			-- Prefetch
		deferred
		end

end
