indexing
	description: "Base class for SQL script capture. It is used for capturing sql_immediate and predicate statement data.";
	author: "India Team";
	date: "$Date: 2009/11/13 $";
	revision: "$Revision: #6 $"

deferred class SQL_SCRIPT_CAPTURE

inherit
	ROSE_SHARED_GLOBAL	--for config
		rename
			trace as shared_trace,
			output as shared_output
		end

	-- Typically, you may run two versions of a supplier update
	-- program (production version and new development version)
	-- with script capture turned on and in test mode.

	-- Database is not affected by the runs as data is NOT committed
	-- in test mode but at the end of the run, the transaction is
	-- ROLLED BACK. (Sequence numbers are generated but these are
	-- not committed to the database.)

	-- You may run the two versions in test mode with the same input data
	-- file and compare the two script files captured to make sure that
	-- the new update program is doing the update operations correctly.

	-- The following variables may be set in the
	-- supplier update program configuration file to capture
	-- the relevant SQL statements and data to the file.

	-- sql_script_capture_required = True
	--	Set True if SQL script capture is required.

	-- sql_script_capture_filename_base = file path
	--	File name base path to store the captured data.

	-- sql_script_capture_sort_required = True
	-- 	Set True if the captured data is to be sorted and written
	--	to the file.
	-- 	Set False to have record separator for each record.  This
	--      facilitates to compare logically the input data and the
	--      captured SQL statements for each record.
	--	It may be required to sort the captured files later on
	--	for comparing the actual content differences.

	-- sql_script_capture_table_columns_to_ignore
	-- 	Provide table name and column name as pairs to ignore the
	--	column value while capturing the SQL script. Each pair
	--	should be separated with comma.
	--	e.g. "fiscal_data_signatures:secuirty_entity_id,fiscal_data_signatures:currency_id"


feature -- Commands
	sql_script_capture_start_called: BOOLEAN
			-- Indicates whether or not the script capturing has been started or not.
	sql_script_capture_start(a_registered_name: STRING) is
			-- This feature is called from "process_file" feature of
			-- supplier update program, before processing the file.
		local
			l_registered_name: STRING
		do
			sql_script_capture_start_called := True
			if sql_script_capture_required then
				if a_registered_name /= void and then not a_registered_name.is_empty then
					l_registered_name := a_registered_name.twin
					l_registered_name.replace_substring_all (" ", "_")
					l_registered_name.replace_substring_all ("-", "_")
					config.set ("$system/sql_script_capture_filename_suffix", l_registered_name)
				end
				config.set ("$system/sql_script_capture_file_number", "1")
				config.set ("$system/sql_script_capture_file_size", "0")
				sql_statements.empty_out
			end
		end

	sql_script_capture_finish is
			-- This feature is called from "process_file" feature of
			-- supplier update program, after processing the file.
		require
			sql_script_capture_start_called = True
		do
			if sql_script_capture_required then
				sql_script_capture_dump(sql_script_capture_filename)
			end
		end

	sql_script_capture_write_if_required is
			-- This feature is called from "process_all_targets" feature of
			-- supplier update program, after processing each target/record.
		require
			sql_script_capture_start_called = True
		local
			l_file_number, l_file_size: INTEGER
		do
			if sql_script_capture_required then
				l_file_number := sql_script_capture_file_number
				l_file_size := config.integer ("$system/sql_script_capture_file_size", 0)
				if l_file_size >= Max_file_size then
					sql_script_capture_dump(sql_script_capture_filename)
					config.set ("$system/sql_script_capture_file_number", (l_file_number + 1).out)
					config.set ("$system/sql_script_capture_file_size", "0")
				end
				if not sql_script_capture_sort_required then
					-- insert a new line at the end of every logical record.
					extend_to_sql_statements(record_separator)
				end
			end
		end

feature{NONE} -- Implementation

	sql_script_capture_filename:STRING is
		-- Sql script capture filename
		local
			l_suffix: STRING
			l_file_number: INTEGER
		do
			Result := sql_script_capture_filename_base
			l_suffix := sql_script_capture_filename_suffix
			if l_suffix /= void and then l_suffix.count > 0 then
				Result := Result + "_" + l_suffix
			end
			l_file_number := sql_script_capture_file_number
			Result := Result + "_" + formatted_string(l_file_number)
			Result := Result + Sql_script_capture_filename_extension
		end

	sql_script_capture_dump(a_file_name: STRING) is
		-- sql script capture dump
		do
			if sql_script_capture_required then
				if sql_script_capture_sort_required then
					sql_statements.sort_by (agent sql_script_capture_sort(?,?))
				end
				write_sql_statements_to_file(a_file_name)
				sql_statements.empty_out
			end
		end

	sql_script_capture_sort(a1, a2: like sql_statement): BOOLEAN is
		-- Sql script capture sort
		do
			Result := a1 < a2
		end

	sql_capture_immediate_execute_statements(a_stmt: STRING) is
			-- Captures sql statements like create, drop and update of temp tables.
		local
			l_stmt: STRING
			l_table_name: STRING
			l_table_name_from_map: STRING
			l_start: INTEGER
		do
			if statement_has_temp_table(a_stmt, 1) then
				l_stmt := a_stmt.twin
				from l_start := 1
				until not statement_has_temp_table(l_stmt, l_start)
					-- A statement may have more than one temp table.
				loop
					l_table_name := statement_temp_table_name(l_stmt, l_start)
					l_table_name_from_map := map_temp_table_name(l_table_name)
					l_stmt.replace_substring_all(l_table_name, l_table_name_from_map)
					l_start := l_start + l_table_name_from_map.count
				end
				extend_to_sql_statements(l_stmt+"%N")
			elseif a_stmt.substring_index(Rollback_string, 1) > 0 OR
				a_stmt.substring_index(Commit_string, 1) > 0 then
				extend_to_sql_statements(a_stmt+"%N")
			elseif a_stmt.has_substring (Implicit_transaction_string) then
				do_nothing
			else
				io.putstring("WARNING:Sql_script_capture:We do not expect to get here for sql statements which do not have rollbacks or commits or temp tables or set implicit_transaction on/off.%N")
			end
		end

	sql_capture_predicate_statement_data (a_predicate_statement: PREDICATE_STATEMENT; a_object: ANY;
										  args: ARRAY[ANY]; input: DATA_DESCRIPTORS[DATA_DESCRIPTOR]) is
			-- For Entity and signature tables
		local
			l_binding: ROSE_LINEAR_ARRAY[STRING]
			l_data_structure: DATA_STRUCTURE[SQL_SQLVAR]
			l_array: ROSE_LINEAR_ARRAY [ANY]
			i: INTEGER
			l_index, l_index2: INTEGER
			l_table_name: STRING
		do
			create sql_statement.make_from_string(a_predicate_statement.parsed_string)
			sql_statement.to_lower
			l_data_structure ?= input
			if (l_data_structure = void) then
				create l_array.make_empty
				io.putstring("WARNING:Sql_script_capture:Unexpected void data_structure.%N")
			else
				l_array := predicate_statement_values (a_object, args, l_data_structure)
			end

			if Sql_script_capture_ignore_entity_id OR
				Sql_script_capture_ignore_signature_id OR
				Sql_script_capture_ignore_process_id OR
				Sql_script_capture_ignore_process_start_and_end_time OR
				(sql_script_capture_table_columns_to_ignore /= void and then
				 sql_script_capture_table_columns_to_ignore.count > 0) then
				l_binding ?= map_column_names(sql_statement)
			end

			-- Get the table name of insert into or update statements.
			l_index := sql_statement.substring_index(insert_into_string, 1)
			if l_index > 0 then
				l_index2 := sql_statement.index_of(' ', l_index+Insert_into_string_size) -- 12 for size of "insert into "
				l_table_name := sql_statement.substring(l_index+Insert_into_string_size, l_index2 - 1)
			else
				l_index := sql_statement.substring_index(Update_string, 1)
				if l_index > 0 then
					l_index2 := sql_statement.index_of(' ', l_index+Update_string_size) -- 7 for size of "update "
					l_table_name := sql_statement.substring(l_index+Update_string_size, l_index2 - 1)
				else
					io.putstring("WARNING:Sql_script_capture:We do not expect to get here for sql statements which do not have insert or update statements.%N")
				end
			end

			if l_array.count > 0 and l_table_name /= void then
				i := 1
				from l_array.start until l_array.off loop
					if l_array.current_item /= void then
						if (l_binding /= void and then l_binding.valid_index(i)) then
							check_input_type_and_replace(field_value(l_array.current_item
																	 , l_table_name, l_binding.item(i)))
							i := i + 1
						else
							check_input_type_and_replace(l_array.current_item)
						end
					end
					l_array.forth
				end
				l_binding := void
			end
			sql_statement.replace_substring_all("?", "NULL")  -- The assumption is that a NULL is allowed for any remaining unspecified arguments.
			if l_table_name /= void and then not l_table_name.is_equal ("process_instance") then
				-- process_instance entries are not captured.
				extend_to_sql_statements(sql_statement + "%N")
			end
		end

feature {NONE} -- Utility features
	predicate_statement_values (a_object: ANY; a_args: ARRAY [ANY];
								a_input: DATA_STRUCTURE[sql_sqlvar]): ROSE_LINEAR_ARRAY [ANY] is
		-- Predicate statement
		local
			i,j: INTEGER
		do
			if a_input.input_object_map = Void or else a_object = Void then
				Result := values_from_array (a_args, a_input.input_array_map)
			else
				create Result.make_empty
				from i := 1 until i > a_input.input_object_map.count
				loop
					j := a_input.input_object_map.item (i)
					if j /= 0 then
						Result.extend (internal3.field (j,a_object))
					else
						j := a_input.input_array_map.item(i)
						if a_args/=Void and then
							a_args.valid_index(j) and then
							a_input.input_array_map.valid_index(i)
						 then
							Result.extend (a_args.item(j))
						end
					end
					i := i+1
				end
			end
		end

	values_from_array (a_args: ARRAY [ANY]; a_input_array_map: ARRAY [INTEGER]): ROSE_LINEAR_ARRAY [ANY] is
		-- Values from array
		local
			i: INTEGER
			j: INTEGER
		do
			create Result.make_empty
			if a_args /= Void then
				if a_input_array_map /= Void then
					from i := 1 until i > a_args.count
					loop
						j := a_input_array_map.item (i)
						if j > 0 then
							if a_args.valid_index(j) then
								Result.extend (a_args.item (j))
							end
						end
						i := i+1
					end
				else
					from i := 1 until i > a_args.count
					loop
						Result.extend (a_args.item (i))
						i := i+1
					end
				end
			end
		end

	predicate_statement_column_names (a_stmt: STRING): ROSE_LINEAR_ARRAY[STRING] is
		-- Predicate Statement column names
		local
			l_start, l_end: INTEGER
		do
			-- We need column names for the insert statement. We may possibly use
			-- prepare statement and get the column names in a better way.
			l_start := a_stmt.index_of('(', 1) + 1
			l_end := a_stmt.index_of(')', 1) - 1
			Result ?= create {ROSE_DELIMITED_TEXT}.make_ignoring_repeats(a_stmt.substring(l_start, l_end), ',')
			from Result.start until Result.off loop
				Result.current_item.prune_all(' ')
				Result.current_item.prune('[')
				Result.current_item.prune(']')
				Result.forth
			end
		end

	statement_has_temp_table(a_statement: STRING; a_start_index:INTEGER): BOOLEAN is
		-- Temp table
		do
			if a_statement.substring_index("#", a_start_index) /= 0 then
				Result := true
			else
				Result := false
			end
		end

	statement_temp_table_name(a_statement: STRING; a_start_index: INTEGER): STRING is
		-- Temp table name
		local
			l_first_index: INTEGER
			l_last_index: INTEGER
		do
			l_first_index := a_statement.substring_index("#", a_start_index)
			l_last_index := a_statement.index_of(' ', l_first_index)
			if l_last_index = 0 then
				l_last_index := a_statement.count
			else
				l_last_index := l_last_index - 1
			end
			Result := a_statement.substring(l_first_index, l_last_index).twin
		end

	formatted_date (a_dt: ROSE_DATE): STRING is
		-- Formatted date
		do
			if a_dt /= Void then
	      		Result := "'"+a_dt.year.out+"-"+a_dt.month.out+"-"+a_dt.day.out+"'"
			else
                Result := ""
            end
		end

	single_quoted_string(a_str: STRING): STRING is
		-- Single quoted string
		do
			Result := a_str.twin
			if a_str.has('%'') then
				a_str.replace_substring_all("'", "''")
			end
			Result.precede('%'')
			Result.append_character('%'')
		end

	replace_meta_character(a_updated_string: STRING) is
		-- Replace meta character
		local
			l_index: INTEGER
		do
			l_index:= sql_statement.index_of('?', 1)
			if l_index /= 0 then
				sql_statement.replace_substring(a_updated_string, l_index, l_index)
			end
		end

	check_input_type_and_replace(a_any: ANY) is
		-- Check input_type and replace
		local
			l_dt: ROSE_DATE
			l_str: STRING
			l_date_string: STRING
		do
			l_dt ?= a_any.twin
			l_str ?= a_any.twin

			if l_dt /= void then
				l_date_string := formatted_date (l_dt)
				replace_meta_character(l_date_string)
			elseif l_str /= void then
				replace_meta_character(single_quoted_string(l_str))
			else
				replace_meta_character(a_any.out)
			end
		end

	map_temp_table_name(a_table_name: STRING):STRING is
		-- Map temp table name
		local
			l_integer_string: STRING
			l_new_table_name: STRING
			l_count: INTEGER
			i: INTEGER
			l_char: CHARACTER
			l_flag1, l_flag2: BOOLEAN
		do
			if temp_tables = void then
				create temp_tables.make(0)
			end
			if temp_tables.has(a_table_name) then
				Result := temp_tables.item(a_table_name)
			else
				create l_integer_string.make(0)
				l_count := a_table_name.count
				from i := 1 until l_flag1 OR i > l_count loop
					l_char := a_table_name.item(i)
					if l_char.is_digit then
						l_integer_string.append_character(l_char)
						if not l_flag2 then
							l_flag2 := TRUE
						end
					elseif l_flag2 then
						l_flag1 := TRUE
					end
					i := i + 1
				end
				l_new_table_name := a_table_name.twin
				if l_integer_string.count >0 then
					l_new_table_name.replace_substring_all(l_integer_string, "")
				end
				temp_tables.extend(l_new_table_name, a_table_name)
				Result := l_new_table_name
			end
		end

	field_value(a_any: ANY; a_table_name, a_column_name: STRING): ANY is
		-- Field value
		local
			l_search_str: STRING
		do
			if a_column_name.is_equal("process_instance_id") then
				Result := 0
			end

			if Sql_script_capture_ignore_entity_id then
				if a_column_name.is_equal(Entity_id_string) OR
					a_column_name.is_equal(Entity_1_id_string) OR
					a_column_name.is_equal(Entity_2_id_string) then
					Result := 0
				end
			end
			if Sql_script_capture_ignore_signature_id then
				if a_column_name.is_equal(Signature_id_string) then
					Result := 0
				end
			end
			if Sql_script_capture_ignore_process_id then
				if a_column_name.is_equal(Process_id_string) OR
					a_column_name.is_equal(First_process_id_string) OR
					a_column_name.is_equal(Last_process_id_string) OR
					a_column_name.is_equal(Continuing_process_id_string) OR
					a_column_name.is_equal(Signature_data_process_id_string) then
					Result := 0
				end
			end

			if Sql_script_capture_ignore_process_start_and_end_time then
				if a_column_name.is_equal(Start_time_string) OR
					a_column_name.is_equal(End_time_string) then
					Result := 0
				end
			end

			if Result = void then
				l_search_str := a_table_name + ":" + a_column_name
				if ignore_table_columns.has (l_search_str) then
					Result := 0
				else
					Result := a_any
				end
			end
		end

	map_column_names (a_stmt: STRING): ROSE_LINEAR_ARRAY [STRING] is
		-- Map column names
		local
			l_start, l_end: INTEGER
			l_key: STRING
			l_field_names: ROSE_LINEAR_ARRAY [STRING]
		do
			l_start := 1
			l_end := a_stmt.index_of(')', 1)
			l_key := a_stmt.substring(l_start, l_end).twin

			if field_names = void then
				create field_names.make(0)
			end
			if field_names.has(l_key) then
				Result := field_names.item(l_key)
			else
				l_field_names := predicate_statement_column_names(a_stmt)
				field_names.extend(l_field_names, l_key)
				Result := l_field_names
			end
		end

feature -- Output

	extend_to_sql_statements(a_statement: STRING) is
		-- Extend
		local
			l_file_size: INTEGER
		do
			if a_statement /= void then
				if not a_statement.has_substring(Rollback_string) and
					not a_statement.has_substring(Commit_string) then
					l_file_size := config.integer ("$system/sql_script_capture_file_size", 0)
					config.set ("$system/sql_script_capture_file_size", (a_statement.count + l_file_size).out)
					sql_statements.extend(a_statement)
				end
			end
		end

	write_sql_statements_to_file(a_file_name: STRING) is
		-- Write sql_statements to file
		do
			create output_file_for_capture.make_open_write(a_file_name)
			from sql_statements.start until sql_statements.off loop
				output_file_for_capture.put_string(sql_statements.current_item)
				sql_statements.forth
			end
			output_file_for_capture.close
		end

	formatted_string(a_integer_value: INTEGER): STRING is
		-- Formatted string
		do
			if a_integer_value >= 0 and a_integer_value < 10 then
				Result := "00" + a_integer_value.out
			elseif a_integer_value >= 10 and a_integer_value < 100 then
				Result := "0" + a_integer_value.out
			elseif a_integer_value >= 100 then
				Result := a_integer_value.out
			end
		end

feature{NONE}-- Attributes

	sql_statement: STRING -- Sql statement
	sql_statements: ROSE_LINEAR_ARRAY [STRING] is
		 -- Sql statements
		once
			create Result.make_empty
		end
	temp_tables: HASH_TABLE [STRING, STRING] -- Temp tables
	field_names: HASH_TABLE [ROSE_LINEAR_ARRAY [STRING], STRING] -- Field names
	output_file_for_capture: ROSE_TEXT_FILE -- Output file for capture


	internal3: INTERNAL is
		 -- Internal3
		once
			create Result
		end
	execution: EXECUTION_ENVIRONMENT is
		 -- Execution
		once
			Create Result
		end

feature {NONE} -- config entries

	sql_script_capture_required: BOOLEAN is
		-- Sql script capture required
		once
			if config.is_available then
				Result := sql_script_capture
			end
		end

	sql_script_capture: BOOLEAN is
		-- Sql script capture
		once
			Result := config.boolean("$system/sql_script_capture_required",False)
		end

	sql_script_capture_filename_base: STRING is
		-- Sql script capture filename base
		once
			Result := config.value("$system/sql_script_capture_filename_base")
		end

	sql_script_capture_table_columns_to_ignore:STRING is
		-- Sql script capture table columns to ignore
		once
			Result := config.value("$system/sql_script_capture_table_columns_to_ignore")
		end

	sql_script_capture_sort_required: BOOLEAN is
		-- Sql script capture sort required
		once
			Result := config.boolean("$system/sql_script_capture_sort_required",False)
		end

feature{NONE} -- Internal

	sql_script_capture_filename_suffix: STRING is
		-- Registered name is used to set the suffix, not obtained from config file.
		do
			Result := config.value("$system/sql_script_capture_filename_suffix")
		end

	sql_script_capture_file_number: INTEGER is
		-- Used internally to maintain the sequence number of the file.
		do
			Result := config.integer ("$system/sql_script_capture_file_number", 1)
		end

	ignore_table_columns: HASH_TABLE [BOOLEAN, STRING] is
		-- Ignore table columns
		local
			l_columns: ROSE_LINEAR_ARRAY[STRING]
			l_str: STRING
		once
			create Result.make (0)
			l_str := sql_script_capture_table_columns_to_ignore
			if l_str /= void and then l_str.count > 0 then
				l_columns := create {ROSE_DELIMITED_TEXT}.make_ignoring_repeats(l_str, ',')
				from l_columns.start until l_columns.off loop
					Result.extend(true, l_columns.current_item)
					l_columns.forth
				end
			end
		end

feature{NONE} -- constants
	process_instance_id_string: STRING is "process_instance_id" -- Process instance_id

	Entity_id_string: STRING is "entity_id" -- Entity_id
	Entity_1_id_string: STRING is "entity_1_id" -- Entity_1_id
	Entity_2_id_string: STRING is "entity_2_id" -- Entity_2_id

	Process_id_string: STRING is "process_id" -- Process_id
	Start_time_string: STRING is "start_time" -- Start_time
	End_time_string: STRING is "end_time" -- End_time

	Signature_id_string: STRING is "signature_id" -- Signature_id
	First_process_id_string: STRING is "first_process_id" -- First_process_id
	Last_process_id_string: STRING is "last_process_id" -- Last_process_id
	Continuing_process_id_string: STRING is "continuing_process_id" -- Continuing_process_id
	Signature_data_process_id_string: STRING is "signature_data_process_id" -- Signature_data_process_id

	Commit_string: STRING is "commit" -- Commit
	Rollback_string: STRING is "rollback" -- Rollback
	Implicit_transaction_string: STRING is "IMPLICIT_TRANSACTION" -- IMPLICIT_TRANSACTION

	Max_file_size: INTEGER is 10485760 -- Maximum file size is 10MB

	Sql_script_capture_filename_extension: STRING is ".txt" -- File name extension

	Record_separator: STRING is "-------------------------------------------------------------------------%N" -- Record separator
	Insert_into_string: STRING is "insert into " -- Insert into
	Insert_into_string_size: INTEGER is 12 -- Insert into string size

	Update_string: STRING is "update " -- Update string
	Update_string_size: INTEGER is 7 -- Update string size

	Sql_script_capture_ignore_entity_id: BOOLEAN is True

			-- 	Set True if entity_id field is to be ignored in capturing
			--	the SQL data. Useful when the number of entities made or
			--	the order in which they are made, changes from one version
			--	of update program to another version. The tables entity_id,
			--	entity_attribute, entity_relationship and entity_transition
			--	are affected.

	Sql_script_capture_ignore_signature_id: BOOLEAN is True

			-- 	Set True if signature_id field is to be ignored. Useful
			--	when signatures change from one version to another version.

	Sql_script_capture_ignore_process_id: BOOLEAN is True

			--	Set True if process_id field is to be ignored. The fields
			--	first_process_id, last_process_id, continuing_process_id
			--	and signature_data_process_id are also ignored.

	Sql_script_capture_ignore_process_start_and_end_time: BOOLEAN is True

			-- 	Set True if process start time and end time are to be ignored.
			--      Useful as process start time and end time are always different
			--	for every execution.


end -- SQL_SCRIPT_CAPTURE
