note

	description:
	"[
		Class to manage applications roles for sql server databases.
						  This will enable and refresh the role to be up to date with any changes to the schema
	]"
	last_updater: "$Author: rj $"
	author: "Mauricio Bustos"
	last_editor: "$Author: rj $"
	date: "$Date: 2015/04/07 $"
	revision: "$Revision: #10 $"
	path: "$File: //rose/source/kernel/infrastructure/persistence/databases/sql/sql_server_approle.e $"

class
	SQL_SERVER_APPROLE

inherit
	DATABASE_SERVERS

create
	make,
	default_create,
	unset_all_approles

feature {NONE} -- Creation

	make (a_name, a_password, a_permissions: STRING) is
			-- Create the approle
		require
			valid_name: a_name /= Void and not a_name.is_empty
			valid_password: a_password /= Void and not a_password.is_empty
			valid_permissions: meaningful_permissions (a_permissions)
		do
			name := a_name
			password := a_password
			permissions := a_permissions
		end

	unset_all_approles is
			-- Unset all approles for audited servers
		do
			audited_servers.do_all (agent safe_unset_approle)
		end

feature -- Basic operations

	create_approle (a_server: SQL_DATA_SERVER) is
			-- Create approle on the server
		require
			valid_server: a_server /= Void and then a_server.connected
		do
			a_server.commit
			output (Global.interpreted ("($now) Creating approle " + name + " on " + a_server.node + "." + a_server.database + "%N"),Void)
			a_server.sql_execute_immediate ("sp_addapprole '" + name + "','" + password + "'")
			if not a_server.success then
				fail ("create_approle",Global.interpreted ("($now) Unable to create approle " + name + " on " + a_server.node + "." + a_server.database),Void)
			end
		ensure
			approle_created: has_approle (a_server)
		end

	set_approle (a_server: SQL_DATA_SERVER) is
			-- Approle can only be set once per database session
		require
			valid_server: a_server /= Void and then a_server.connected
			available_name: name /= Void
		local
			l_enabled_servers: HASH_TABLE [TUPLE [server: SQL_DATA_SERVER; cookie: STRING], STRING]
		do
			enabled_data_server_approles.search (name)
			if enabled_data_server_approles.found then
				l_enabled_servers := enabled_data_server_approles.found_item
			end
			if l_enabled_servers = Void then
				create l_enabled_servers.make (10)
				enabled_data_server_approles.force (l_enabled_servers, name)
			end
			if not l_enabled_servers.has (a_server.node + "." + a_server.database) then
				--if not has_approle (a_server) then
				--	create_approle (a_server)
				--end
				audit_approle_enable (a_server)
				set_approle_and_retrieve_cookie (a_server, l_enabled_servers)
			end
		end

	safe_unset_approle (a_server: SQL_DATA_SERVER)
			-- A safe version of 'unset_approle'.
		require
			valid_server: a_server /= Void and then a_server.connected
		local
			l_retry: BOOLEAN
		do
			if not l_retry
			then
				unset_approle(a_server)
			end
		rescue
			l_retry := True
			retry
		end

	unset_approle (a_server: SQL_DATA_SERVER)
			-- Unset the approle and restore the old role (from the cookie).
		require
			valid_server: a_server /= Void and then a_server.connected
		local
			l_enabled_servers: HASH_TABLE [TUPLE [server: SQL_DATA_SERVER; cookie: STRING], STRING]
			l_execution: EXECUTION_PREDICATE
			l_implicit_transaction: BOOLEAN
			l_statement: STRING
		do
			from
				enabled_data_server_approles.start
			until
				enabled_data_server_approles.off
			loop
				l_enabled_servers := enabled_data_server_approles.item_for_iteration
				l_enabled_servers.search (a_server.node + "." + a_server.database)
				if l_enabled_servers.found then
					-- sp_unsetapprole won't work if there is an outstanding transaction.
					from
						do_nothing
					until
						a_server.transaction_depth = 0
					loop
						a_server.commit
					end

					-- Save his implicit transaction state.
					l_implicit_transaction := a_server.implicit_transaction

					-- Turn off implicit transactions (sp_unsetapprole won't work if we start a transaction).
					a_server.set_implicit_transaction_off

					-- Since the cookie may contain apostrophes, we store it in a table (which we now create).
					a_server.sql_execute_immediate ("create table " + unset_cookie_temp_name + " (cookie varchar(8000))")
					if not a_server.success then
						fail ("unset_approle", Global.interpreted ("($now) Could not create cookie temp table"), Void)
					end

					-- Create the insert predicate.
					create l_execution.make (a_server, "insert into " + unset_cookie_temp_name + " (cookie) values (?)")
					if not l_execution.success then
						fail ("unset_approle", Global.interpreted ("($now) Could not create cookie predicate"), Void)
					end

					-- Insert the cookie.
					l_execution.execute (Void, <<l_enabled_servers.found_item.cookie>>)
					if not l_execution.success then
						fail ("unset_approle", Global.interpreted ("($now) Could not execute cookie predicate"), Void)
					end

					-- Unset the approle and restore the role associated with the cookie.
					l_statement := global.formatted("[
														declare @cookie_temp varbinary(8000);
														select @cookie_temp = cast (cookie as varbinary(8000)) from $1;
														exec sp_unsetapprole @cookie_temp
													]",
													<<unset_cookie_temp_name>>)
					a_server.sql_execute_immediate(l_statement)
					if not a_server.success then
						fail ("unset_approle", Global.interpreted ("($now) Failed to unset the cookie"), Void)
					end

					-- Clean up.
					a_server.sql_execute_immediate ("drop table " + unset_cookie_temp_name)
					if not a_server.success then
						fail ("unset_approle", Global.interpreted ("($now) Unable to drop " + unset_cookie_temp_name), Void)
					end

					-- Restore the implicit transaction state.
					if l_implicit_transaction
					then
						a_server.set_implicit_transaction_on
					end

					output (Global.interpreted ("($now) Unset approle " + enabled_data_server_approles.key_for_iteration + " on " + a_server.node + "." + a_server.database + "%N"),Void)
				end
				enabled_data_server_approles.forth
			end
		end

feature -- Access

	name: STRING
			-- The approle's name
	password: STRING
			-- The approle's password
	permissions: STRING
			-- The approle's permissions string

	meaningful_permissions (a_permissions: STRING): BOOLEAN is
			-- Is 'a_permissions' a valid set?
		local
			l_permissions: STRING
		do
			if a_permissions /= Void and then not a_permissions.is_empty then
				l_permissions := a_permissions.as_lower
				Result := (l_permissions.has_substring ("select") or
						   l_permissions.has_substring ("update") or
						   l_permissions.has_substring ("insert") or
						   l_permissions.has_substring ("delete") or
						   l_permissions.has_substring ("all"))
			end
		end

feature -- Status

	has_approle (a_server: SQL_DATA_SERVER): BOOLEAN is
			-- Check to see if the approle exists in the server
		require
			valid_server: a_server /= Void and then a_server.connected
		local
			l_predicate: INTEGER_PREDICATE
		do
			create l_predicate.make_strict (a_server,"select count (*) from sysusers where isapprole = 1 and name = '" + name + "'")
			l_predicate.retrieve (Void,Void)
			Result := l_predicate.array.count = 1 and then l_predicate.array.first = 1
		end

feature -- Element modification

	set_approle_name (a_server: SQL_DATA_SERVER): STRING is
			-- Name of approle if it was set
		require
			valid_server: a_server /= Void and then a_server.connected
		local
			l_enabled_servers: HASH_TABLE [TUPLE [server: SQL_DATA_SERVER; cookie: STRING], STRING]
		do
			from enabled_data_server_approles.start until enabled_data_server_approles.off loop
				l_enabled_servers := enabled_data_server_approles.item_for_iteration
				if l_enabled_servers.has (a_server.node + "." + a_server.database) then
					Result := enabled_data_server_approles.key_for_iteration
				end
				enabled_data_server_approles.forth
			end
		end

feature {NONE} -- Implementation

	set_approle_and_retrieve_cookie (a_server: SQL_DATA_SERVER; a_table: HASH_TABLE [TUPLE [server: SQL_DATA_SERVER; cookie: STRING], STRING]) is
			-- Set the application role on 'a_server' and store the cookie for later unsetting if needed
		require
			valid_server: a_server /= Void and then a_server.connected
			valid_table: a_table /= Void
		local
			l_predicate: STRING_PREDICATE
			l_implicit_transactions_set: BOOLEAN
			l_statement: STRING
		do
				-- Dark magic to remove pre-existing transaction that prevents app role setting
				-- not sure how there can be a preexisting transaction as this is one of the first things we do? FA 200907
			a_server.commit_if_necessary

			-- Save the implicit transaction state and prevent automatic implicit transactions.
			l_implicit_transactions_set := a_server.implicit_transactions_set		-- Save this value.  We restore it the bottom of this feature.
			a_server.set_implicit_transactions_set(True)							-- Prevents the next statement from setting implicit transactions

			-- Create and fetch a cookie for our approle.
			l_statement := global.formatted(
						"[
						declare @set_cookie varbinary (8000);
						exec sp_setapprole '$1', '$2', @fCreateCookie = true, @cookie = @set_cookie OUTPUT;
						select @set_cookie as cookie into $3
						]",
						<<
							name,
							password,
							set_cookie_temp_name
						>>)
			a_server.sql_execute_immediate(l_statement)
			if not a_server.success then
				fail ("set_approle_and_retrieve_cookie", Global.interpreted ("($now) Unable to set approle " + name + " on " + a_server.node + "." + a_server.database), Void)
			end

			-- Get the cookie that we just created.
			create l_predicate.make (a_server, "select cookie from " + set_cookie_temp_name)
			if not l_predicate.success then
				fail ("set_approle_and_retrieve_cookie", Global.interpreted ("($now) Could not create cookie predicate"), Void)
			end
			l_predicate.retrieve (Void, Void)
			if not l_predicate.success then
				fail ("set_approle_and_retrieve_cookie", Global.interpreted ("($now) Could not execute cookie predicate"), Void)
			end
			if l_predicate.array.is_empty or else l_predicate.array.count > 1 or else l_predicate.array.first = Void or else l_predicate.array.first.is_empty then
				fail ("set_approle_and_retrieve_cookie", Global.interpreted ("($now) Failed to retrieve approle cookie"), Void)
			end

			-- Clean up.
			a_server.sql_execute_immediate ("drop table " + set_cookie_temp_name)
			if not a_server.success then
				fail ("set_approle_and_retrieve_cookie", Global.interpreted ("($now) Unable to drop " + set_cookie_temp_name), Void)
			end
			a_server.commit
			a_server.set_implicit_transactions_set(l_implicit_transactions_set)		-- Restores what we saved a couple of lines up.

			-- Store the cookie for later use.  (How does this work for two connections to the same database?)
			a_table.extend ([a_server, l_predicate.array.first], a_server.node + "." + a_server.database)
			--output (Global.interpreted ("($now) Set approle " + name + " on " + a_server.node + "." + a_server.database + "%N"), Void)
		end

	audited_servers: ROSE_LINEAR_ARRAY [SQL_DATA_SERVER] is
			-- Collection of servers that have approles enabled
		local
			l_enabled_approles: ARRAYED_LIST [HASH_TABLE [TUPLE [server: SQL_DATA_SERVER; cookie: STRING], STRING]]
			l_enabled_servers: ARRAYED_LIST [TUPLE [server: SQL_DATA_SERVER; cookie: STRING]]
		do
			create Result.make_empty
			l_enabled_approles := enabled_data_server_approles.linear_representation
			from l_enabled_approles.start until l_enabled_approles.off loop
				l_enabled_servers := l_enabled_approles.item.linear_representation
				from l_enabled_servers.start until l_enabled_servers.off loop
					if l_enabled_servers.item.server /= Void and then not Result.has (l_enabled_servers.item.server) then
						Result.extend (l_enabled_servers.item.server)
					end
					l_enabled_servers.forth
				end
				l_enabled_approles.forth
			end
		ensure
			attached_result: Result /= Void
			valid_result: not Result.has (Void)
		end

	enabled_data_server_approles: HASH_TABLE [HASH_TABLE [TUPLE [server: SQL_DATA_SERVER; cookie: STRING], STRING], STRING] is
			-- Table of enabled approles by data server keyed by approle name, used to prevent enabling multiple approles on a data server
		once
			create Result.make (10)
		end

	audit_approle_enable (a_server: SQL_DATA_SERVER) is
			-- Log the fact that this approle has been enabled for 'a_server'
		local
			l_request_predicate: EXECUTION_PREDICATE
		do
			if repository_server /= Void and then repository_server.has_table ("approle_request") then
				repository_server.begin_transaction
				create l_request_predicate.make (repository_server, "insert into approle_request (approle, server, [database], executable, machine) values (?, ?, ?, ?, ?)")
				if not l_request_predicate.success then
					fail ("audit_approle_enable", Global.interpreted ("($now) Unable to create predicate"), Void)
				end
				l_request_predicate.execute (Void, <<name, a_server.node, a_server.database, Global.app.program_path, Global.app.host_name>>)
				if not l_request_predicate.success then
					fail ("audit_approle_enable", Global.interpreted ("($now) Unable to execute predicate"), Void)
				end
				repository_server.commit
			end
		end

	cookie_temp_local_identifier: STRING is
			-- Unique identifier for machine and process.
		once
			Result := global.app.env.host_name + "_" + global.app.pid.out
		ensure
			result_not_void: Result /= Void
		end

	set_cookie_temp_name: STRING is
			-- Temp table name for setting cookie
		once
			Result := "##set_cookie_temp_" + cookie_temp_local_identifier
		ensure
			result_not_void: Result /= Void
		end

	unset_cookie_temp_name: STRING is
			-- Temp table name for unsetting cookie
		once
			Result := "##unset_cookie_temp_" + cookie_temp_local_identifier
		ensure
			result_not_void: Result /= Void
		end

end

