note

	description:

	"Cache STID records of type G keyed by keys of type K"

	author: "Ivan Tarnavsky"
	last_editor: "$Author: it $"
	date: "$Date: 2012/10/23 $"
	revision: "$Revision: #1 $"
	path: "$File: //rose/source/kernel/capital_actions/stid_record_cache.e $"

class STID_RECORD_CACHE [G -> ROSE_SIGNATURE_RECORD [ENTITY], K -> HASHABLE]

create

	make,
	make_with_order_by

feature {NONE} -- Initialization

	make (a_selection: like selection; a_template: like template) is
			-- Initialize `Current'
		require
			a_selection_not_void: a_selection /= Void
			a_template_not_void: a_template /= Void
		do
			selection := a_selection
			template := a_template
			create cache.make_default
		ensure
			selection_set: selection = a_selection
			template_set: template = a_template
		end

	make_with_order_by (a_selection: like selection; a_template: like template; a_order_by: like order_by) is
			-- Initialize with ordering predicate
		require
			a_selection_not_void: a_selection /= Void
			a_template_not_void: a_template /= Void
			a_order_by_not_void: a_order_by /= Void
		do
			make (a_selection, a_template)
			order_by := a_order_by
		ensure
			selection_set: selection = a_selection
			template_set: template = a_template
			order_by_set: order_by = a_order_by
		end

feature -- Access

	void_item: detachable G
			-- Void item template

	selection: STRING
			-- Record selection filter

	template: FUNCTION [ANY, TUPLE [K], G]
			-- Retrieval template function

	order_by: PREDICATE [ANY, TUPLE [G, G]]
			-- Ordering predicate

	item (a_key: K): ARRAY [G] is
			-- Records identified by `a_key'
		require
			has_a_key: has (a_key)
		do
			Result := cache.item (a_key)
		ensure
			not_void: Result /= Void
			no_void_item: not Result.has (void_item)
		end

	found: BOOLEAN
			-- Last `search' result

	signatures: ARRAY [G]
			-- Records found by `search' or retrieved by `load'

feature -- Status report

	has (a_key: K): BOOLEAN is
			-- Are records identified by `a_key' cached?
		do
			Result := cache.has (a_key)
		end

feature -- Search

	search (a_key: K) is
			-- Search cache for record identified by `a_key'
		do
			cache.search (a_key)
			found := cache.found
			if found then
				signatures := cache.found_item
			else
				signatures := Void
			end
		ensure
			signatures_void_if_not_found: (not has (a_key)) implies signatures = Void
			signatures_not_void_when_found: has (a_key) implies signatures /= Void and then not signatures.has (void_item)
		end

feature -- Element change

	load (a_key: K) is
			-- Retrieve `signatures' for `a_key' from database and store in cache.
		do
			store (retrieve (a_key), a_key)
			signatures := item (a_key)
		ensure
			signatures_not_void: signatures /= Void
			has_a_key: has (a_key)
			cached: item (a_key) = signatures
			no_void_items_in_signatures: not signatures.has (void_item)
		end

	store (a_array: ARRAY [G]; a_key: K) is
			-- Store `a_array' for `a_key'.
			-- Replace existing value associated with `a_key'.
		require
			a_array_not_void: a_array /= Void
			no_void_in_a_array: not a_array.has (void_item)
		local
			l_array: ROSE_LINEAR_ARRAY [G]
		do
			if order_by = Void then
				cache.force (a_array, a_key)
			else
				create l_array.make_from_array (a_array)
				l_array.sort_by (order_by)
				cache.force (l_array, a_key)
			end
		ensure
			cached: has (a_key)
			sorted: order_by /= Void implies item (a_key).same_items ((create {ROSE_LINEAR_ARRAY [G]}.make_from_array (a_array)).sorted (order_by))
		end

feature {NONE} -- Implementation

	cache: DS_HASH_TABLE [ARRAY [G], K]
			-- Record arrays cache by key

	retrieve (a_key: K): ARRAY [G] is
			-- Retrieve records for `a_key' from STID database
		local
			l_template: G
		do
			l_template := template.item ([a_key])
			if l_template /= Void then
				Result := l_template.targets (<<l_template>>, selection)
			end
			if Result = Void then
				create Result.make_empty
			else
				Result.prune_all (void_item)
			end
		ensure
			not_void: Result /= Void
			no_void_item: not Result.has (void_item)
		end

invariant

	void_item_is_void: void_item = Void
	selection_not_void: selection /= Void
	template_not_void: template /= Void
	cache_not_void: cache /= Void
	no_void_item_in_cache: not cache.has_void_item
	no_void_item_in_arrays: not cache.there_exists (agent {ARRAY [G]}.has (void_item))

end
