note

	description:

		"Handle database operations for STID_TESTENV application"

	author: "Mauricio Bustos"
	last_editor: "$Author: Mb $"
	date: "$Date: 11/12/08 7:50a $"
	revision: "$Revision: 64 $"
	path: "$File: $"

class STID_TESTENV_DATABASE [reference G -> ROSE_SIGNATURE_RECORD [ENTITY] create default_create end]

inherit

	ROSE_SHARED_LOGGER

	ROSE_SHARED_DATABASE_POOL
		export {NONE} all end

	ANY

	ROSE_PROCESS_TIMER_COLLECTION_AWARE
	NEEDS_EXPRESSION_FROM_STRING

create

	make,
	make_for_overlay_retrieval

feature {NONE} -- Creation

	make (a_database: SIGNATURE_TAGGED_ITEM_DATABASE [G]; a_source_server, a_target_server: SQL_DATA_SERVER) is
			-- Create the database object
		require
			database_not_void: a_database /= Void
			source_server_not_void: a_source_server /= Void
		do
			global.ensure_storable_compatibility_with_56
			database := a_database
			target_server := a_target_server
			create table.make (a_source_server, database.data_name, Void)
			real_data_enabled := table.data_type_enabled (Real_String, False)
			double_data_enabled := table.data_type_enabled (Double_String, False)
			create item_ids_by_process_cache.make (100)
			create_caches (a_database.data_name, a_source_server, a_target_server)
		end

	make_for_overlay_retrieval (a_table: SIGNATURE_TAGGED_TABLE_STRUCTURE) is
 			-- Create the database in preparation for overlaying data onto a signature tagged database
		require
			table_not_void: a_table /= Void
		do
			table := a_table
			create_caches (a_table.data_name, table.signature_item_server, Void)
		end

feature -- Basic operations

	clear_statistics (a_process: SUPPLIER_PROCESS) is
			-- Delete all stored statistics that have been generated for a_process
		require
			process_not_void: a_process /= Void
		local
			l_predicate: EXECUTION_PREDICATE
		do
			create l_predicate.make (server, "delete from testenv_summary_statistics where process_id = " + a_process.process_id.out)
			l_predicate.execute (Void, Void)
			if testenv_statistics_cache /= Void then
				testenv_statistics_cache.remove (a_process.process_id)
			end
			server.commit
		end

	clear_process (a_process: SUPPLIER_PROCESS) is
			-- Delete all data that has been stored for a_process
		require
			process_not_void: a_process /= Void
		do
			from
				types.start
			until
				types.off
			loop
				server.sql_execute_immediate (
					"delete from "
					+ table.data_name_prefix + "data_" + types.current_item.first + "_value"
					+ " where process_id = " + a_process.process_id.out
													  )
				server.sql_execute_immediate (
					"delete from "
					+ table.data_name_prefix + "data_" + types.current_item.first + "_value_history"
					+ " where process_id = " + a_process.process_id.out
													  )
				types.forth
			end
			server.commit
		end

	clear_caches is
			-- Clear out all caches
		do
			clear_penultimate_data_caches
			if table.data_type_enabled (Real_String, False) then
				real_type_cache.data_item_cache.clear_all
			end
			if table.data_type_enabled (Double_String, False) then
				double_type_cache.data_item_cache.clear_all
			end
			if table.data_type_enabled (Integer_String, False) then
				integer_type_cache.data_item_cache.clear_all
			end
			if table.data_type_enabled (String_String, False) then
				string_type_cache.data_item_cache.clear_all
			end
		end

	clear_penultimate_data_caches is
			-- Clear out the penultimate data caches in preparation for another retrieval
		do
			if table.data_type_enabled (Real_String, False) then
				real_type_cache.penultimate_data_item_cache.clear_all
				real_type_cache.penultimate_selective_data_item_cache.clear_all
			end
			if table.data_type_enabled (Double_String, False) then
				double_type_cache.penultimate_data_item_cache.clear_all
				double_type_cache.penultimate_selective_data_item_cache.clear_all
			end
			if table.data_type_enabled (Integer_String, False) then
				integer_type_cache.penultimate_data_item_cache.clear_all
				integer_type_cache.penultimate_selective_data_item_cache.clear_all
			end
			if table.data_type_enabled (String_String, False) then
				string_type_cache.penultimate_data_item_cache.clear_all
				string_type_cache.penultimate_selective_data_item_cache.clear_all
			end
		end

	compute_statistics_for_process (a_process_id: INTEGER; a_converter: PROCEDURE [ANY, TUPLE [G]]; a_expression_index_set: ROSE_LINEAR_ARRAY [INTEGER]) is
			-- Compute statistics for process
		require
			database_not_void: valid_database
		local
			l_item_ids: ROSE_LINEAR_ARRAY [INTEGER]
			l_signatures: ROSE_LINEAR_ARRAY [G]
			--l_description: VARIABLE_DESCRIPTION
			l_expression: ROSE_EXPRESSION [ANY]
			l_signature: G
			l_testenv_statistics: TESTENV_STATISTICS [G]
			l_process: SUPPLIER_PROCESS
			l_expression_count: INTEGER
			l_expression_counter: INTEGER
		do
			create l_signature
			process_clear_predicate.execute (Void, <<a_process_id>>)
			l_item_ids := item_ids_by_process (a_process_id)
			l_expression_count := l_item_ids.count
			l_signatures := signature_ids_by_process (a_process_id)
			if not l_signatures.is_empty then
				l_signatures := database.retrieve_current_signatures (l_signatures, "#signature_id.id = signature_id")
				from
					l_signatures.start
				until
					l_signatures.off
				loop
					l_signatures.current_item.set_path (database.path (l_signatures.current_item))
					l_signatures.current_item.set_last_process_id (a_process_id)
					a_converter.call ([l_signatures.current_item])
					l_signatures.first.reconcile_target (l_signatures.current_item)
					l_signatures.forth
				end
				l_process := database.process_id_manager.process_by_id (a_process_id)
				if l_process /= Void then
					from
						l_item_ids.start
					until
						l_item_ids.off
					loop
						l_expression_counter := l_expression_counter + 1
						l_expression := expressions.variables_by_variable_index.item (l_item_ids.current_item)
						--l_expression := expression_from_mnemonic (l_description.mnemonic)
						if l_expression /= Void and then (a_expression_index_set = Void or else a_expression_index_set.has (l_expression.variable_index)) then
							logger.info("Computing statistics for " + l_expression_counter.out + "/" + l_expression_count.out + " : " +l_expression.mnemonic )
							debug("timers_variable_calculation")
								timers.resume("compute_statistics_for_process: " + generating_type, "load_signatures: " + l_expression.mnemonic)
							end
							load_signatures (l_signatures)
							debug("timers_variable_calculation")
								timers.interrupt("compute_statistics_for_process: " + generating_type, "load_signatures: " + l_expression.mnemonic)
								timers.resume("compute_statistics_for_process: " + generating_type, "testenv_statistics: " + l_expression.mnemonic)
							end
							l_testenv_statistics := testenv_statistics (l_process, l_signatures, l_expression, Void, 0, 0)
							debug("timers_variable_calculation")
								timers.interrupt("compute_statistics_for_process: " + generating_type, "testenv_statistics: " + l_expression.mnemonic)
								timers.resume("compute_statistics_for_process: " + generating_type, "persist_last_statistics: " + l_expression.mnemonic)
							end
							persist_last_statistics
							debug("timers_variable_calculation")
								timers.interrupt("compute_statistics_for_process: " + generating_type, "persist_last_statistics: " + l_expression.mnemonic)
							end
							testenv_statistics_cache.clear_all
						end
						l_item_ids.forth
					end
					logger.info("Computed statistics for " + l_item_ids.count.out + " items.")
				end
			end
		end

	load_signatures (a_signatures: ROSE_LINEAR_ARRAY [G]) is
			-- Load the signature ids into the temporary tables against which later joins will be performed
		require
			valid_signatures: a_signatures /= Void
		local
			l_insert_signature: DATA_SERVER_FAST_LOAD
			l_fast_load_server: SQL_DATA_SERVER
			l_template: G
			l_definition: ROSE_DATABASE_DEFINITION
		do
			if sids_temp_table_created then
				target_server.sql_execute_immediate ("drop table ##" + pid + "sids")
				target_server.sql_execute_immediate ("drop table ##" + pid + "sid_process")
			end
			sids_temp_table_created := True
			target_server.sql_execute_immediate ("create table ##" + pid + "sids (signature_id int)")
			target_server.sql_execute_immediate ("create clustered index sid_xxx on ##" + pid + "sids (signature_id)")
			target_server.sql_execute_immediate ("create table ##" + pid + "sid_process (signature_id int, process_id int)")
			target_server.sql_execute_immediate ("create clustered index sidp_xxx on ##" + pid + "sid_process (signature_id, process_id)")
			target_server.commit
			create l_template
			create l_definition.make ("hardcoded", target_server.logical, "sql", target_server.node, target_server.database)
			l_fast_load_server ?= pool.get_fast_load_server_from_definition (l_definition)
			if l_fast_load_server /= Void then
				l_insert_signature := l_fast_load_server.fast_loader ("##" + pid + "sids", l_template)
				from
					a_signatures.start
				until
					a_signatures.off
				loop
					l_insert_signature.insert (a_signatures.current_item)
					a_signatures.forth
				end
				l_insert_signature.commit
			else
				(create {EXCEPTIONS}).raise ("Unable to get data server")
			end
		end

	validate_difference (a_difference: DATA_COMPARISON [G]; a_item_id: INTEGER; a_comment: STRING) is
			-- Store the fact that we validated this differnce, use this in the computation of statistics
		require
			valid_difference: a_difference /= Void
			valid_item_id: a_item_id > 0
			valid_commend: a_comment /= Void
		local
			l_delete_predicate, l_insert_predicate: EXECUTION_PREDICATE
			l_signatures: ROSE_LINEAR_ARRAY [G]
		do
			if a_difference.target.signature_id = 0 then
				l_signatures := database.retrieve_current_signatures (
					create {ROSE_LINEAR_ARRAY [G]}.make_from_array (<<a_difference.target>>), a_difference.target.repaired_statement (a_difference.target.selective_restriction_statement)
																					  )
				if l_signatures /= Void and then l_signatures.count = 1 then
					a_difference.target.set_signature_id (l_signatures.first.signature_id)
				end
			end
			if a_difference.target.signature_id /= 0 then
				if not testenv_validated_differences_table_available then
					create_testenv_validated_differences_table
				end
				create l_delete_predicate.make (server, "delete from testenv_validated_differences where signature_id = ? and item_id = ?")
				if l_delete_predicate.success then
					l_delete_predicate.execute (Void, <<a_difference.target.signature_id, a_item_id>>)
					if not l_delete_predicate.success then
						(create {EXCEPTIONS}).raise ("Could not delete using delete validation predicate")
					else
						server.commit
					end
				else
					(create {EXCEPTIONS}).raise ("Could not create delete validation predicate")
				end
				create l_insert_predicate.make (server, "insert into testenv_validated_differences (signature_id, item_id, x, y, comment) values (?,?,?,?,?)")
				if l_insert_predicate.success then
					l_insert_predicate.execute (Void, <<a_difference.target.signature_id, a_item_id, a_difference.x, a_difference.y, a_comment>>)
					if not l_insert_predicate.success then
						(create {EXCEPTIONS}).raise ("Could not insert using validation predicate")
					else
						server.commit
					end
				else
					(create {EXCEPTIONS}).raise ("Could not create validation predicate")
				end
			end
		end

feature -- Element change

	set_selective_penultimate_selection (a_value: BOOLEAN) is
			-- Set the selective penultimate selection mode to a_value
		do
			if table.data_type_enabled (Real_String, False) then
				real_type_cache.set_selective_penultimate_selection (True)
			end
			if table.data_type_enabled (Double_String, False) then
				double_type_cache.set_selective_penultimate_selection (True)
			end
			if table.data_type_enabled (Integer_String, False) then
				integer_type_cache.set_selective_penultimate_selection (True)
			end
			if table.data_type_enabled (String_String, False) then
				string_type_cache.set_selective_penultimate_selection (True)
			end
		end

	set_validity_agent (a_agent: FUNCTION [ANY, TUPLE [G, DATA_COMPARISON [G]], BOOLEAN]) is
			-- Set the target validity test to a_agent
		do
			validity_agent := a_agent
		ensure
			agent_set: validity_agent = a_agent
		end

feature -- Access

	valid_database: BOOLEAN is
			-- Is there a valid database available
		do
			Result := database /= Void
		end

	record_overlayed_with_testenv_data (a_record: ROSE_VIRTUAL_RECORD [G]): ROSE_VIRTUAL_RECORD [G] is
			-- a_record overlayed with the test data for a_record.signature
		require
			valid_record: a_record /= Void
			valid_signature: a_record.signature /= Void
		local
			l_real_testenv_data: ROSE_LINEAR_ARRAY [DATA_RECORD [REAL]]
			l_integer_testenv_data: ROSE_LINEAR_ARRAY [DATA_RECORD [INTEGER]]
			l_double_testenv_data: ROSE_LINEAR_ARRAY [DATA_RECORD [DOUBLE]]
		do
			Result := a_record.duplicate
			if table.data_type_enabled (Real_String, False) then
				l_real_testenv_data := real_type_cache.latest_data_items_by_signature (a_record.signature)
				if l_real_testenv_data /= Void then
					from
						l_real_testenv_data.start
					until
						l_real_testenv_data.off
					loop
						Result.set_real_item (l_real_testenv_data.current_item.value, l_real_testenv_data.current_item.item_id)
						l_real_testenv_data.forth
					end
				end
			end
			if table.data_type_enabled (Double_String, False) then
				l_double_testenv_data := double_type_cache.latest_data_items_by_signature (a_record.signature)
				if l_double_testenv_data /= Void then
					from
						l_double_testenv_data.start
					until
						l_double_testenv_data.off
					loop
						Result.set_double_item (l_double_testenv_data.current_item.value, l_double_testenv_data.current_item.item_id)
						l_double_testenv_data.forth
					end
				end
			end
			if table.data_type_enabled (Integer_String, False) then
				l_integer_testenv_data := integer_type_cache.latest_data_items_by_signature (a_record.signature)
				if l_integer_testenv_data /= Void then
					from
						l_integer_testenv_data.start
					until
						l_integer_testenv_data.off
					loop
						Result.set_integer_item (l_integer_testenv_data.current_item.value, l_integer_testenv_data.current_item.item_id)
						l_integer_testenv_data.forth
					end
				end
			end
		end

	process_ids: ROSE_LINEAR_ARRAY [SUPPLIER_PROCESS] is
			-- Collection of SUPPLIER_PROCESS objects representing the process_ids in the test environment
		local
			l_integer_predicate: INTEGER_PREDICATE
			l_statement: STRING
			l_processes: ROSE_LINEAR_ARRAY [SUPPLIER_PROCESS]
		do
			l_processes := database.process_id_manager.all_processes
			l_processes.set_comparison (agent descending_process_ids)
			if not l_processes.is_sorted then
				l_processes.sort
			end
			if database.data_server_name.is_equal (server.node + "." + server.database) then
				if l_processes.count >= 200 then
					Result := l_processes.subarray (1, 200)
				else
					Result := l_processes
				end
			else
				from
					l_statement := ""
					types.start
				until
					types.off
				loop
					l_statement := l_statement
					+ "select * from (select distinct top 200 process_id from data_" + types.current_item.first + "_value order by process_id desc) a" + types.index.out
					+ " union "
					+ "select * from (select distinct top 200 process_id from data_" + types.current_item.first + "_value_history order by process_id desc) b" + types.index.out
					if not types.at_last then
						l_statement := l_statement + " union "
					end
					types.forth
				end
				create l_integer_predicate.make_strict (server, l_statement)
				l_integer_predicate.retrieve (Void, Void)
				create Result.make_empty
				from
					l_processes.start
				until
					l_processes.off or Result.count > 200
				loop
					if l_integer_predicate.array.has (l_processes.current_item.process_id) then
						Result.extend (l_processes.current_item)
					end
					l_processes.forth
				end
			end
		ensure
			non_void_result: Result /= Void
		end

	process_ids_with_computed_statistics: ROSE_LINEAR_ARRAY [INTEGER] is
			-- Collection of process_ids in the test environment that have pre-computed statistics
		local
			l_integer_predicate: INTEGER_PREDICATE
		do
			if testenv_summary_statistics_table_available then
				create l_integer_predicate.make_strict (server, "select distinct process_id from testenv_summary_statistics order by process_id")
				l_integer_predicate.retrieve (Void, Void)
				Result := l_integer_predicate.array
			else
				create Result.make_empty
			end
		ensure
			non_void_result: Result /= Void
		end

	stored_mnemonics_for_process_id (a_process_id: INTEGER): ROSE_LINEAR_ARRAY [STRING] is
			-- Collection of variable mnemonics that have pre-computed statistics for a_process_id
		local
			l_string_predicate: STRING_PREDICATE
		do
			if testenv_summary_statistics_table_available then
				create l_string_predicate.make_strict (server, "select distinct mnemonic from testenv_summary_statistics where process_id = ? order by mnemonic")
				l_string_predicate.retrieve (Void, <<a_process_id>>)
				Result := l_string_predicate.array
			else
				create Result.make_empty
			end
		ensure
			non_void_result: Result /= Void
		end

	expression_from_mnemonic (a_mnemonic: STRING): ROSE_EXPRESSION [ANY] is
			-- Expression object for 'a_mnemonic'
		require
			valid_mnemonic: a_mnemonic /= Void
		do
			Result := expressions.complex_expression (a_mnemonic)
		end

	expressions: ENTITY_VARIABLES [ANY] is
			-- Expression factory for the signature
		local
			l_signature: G
			l_mondat_signature: ANY
		do
			create l_signature
			l_mondat_signature ?= Internal.new_instance_of (Internal.dynamic_type_from_string ("MONDATC_SIGNATURE"))
			if l_mondat_signature /= Void and then l_signature.conforms_to (l_mondat_signature) then
				if mondat_expressions = Void then
					mondat_expressions ?= Internal.new_instance_of (Internal.dynamic_type_from_string ("RESEARCH_COMPANY_VARIABLES"))
					if mondat_expressions = Void then
						(create {EXCEPTIONS}).raise ("unable to instantiate RESEARCH_COMPANY_VARIABLES")
					else
						mondat_expressions.make
					end
				end
				Result := mondat_expressions
			else
				Result := l_signature.expressions
			end
		ensure
			valid_result: Result /= Void
		end

	mondat_expressions: ENTITY_VARIABLES [ANY]

	process_contains_this_item (a_item, a_process_id: INTEGER): BOOLEAN is
			-- Does the test environment have data for a_item for a_process_id?
		local
			l_integer_predicate: INTEGER_PREDICATE
			l_statement: STRING
		do
			from
				l_statement := ""
				types.start
			until
				types.off
			loop
				l_statement := l_statement
				+ "select top 1 * from " + table.data_name_prefix + "data_" + types.current_item.first + "_value where process_id = " + a_process_id.out + " and item_id = " + a_item.out
				+ " union "
				+ "select top 1 * from " + table.data_name_prefix + "data_" + types.current_item.first + "_value_history where process_id = " + a_process_id.out + " and item_id = " + a_item.out
				if not types.at_last then
					l_statement := l_statement + " union "
				end
				types.forth
			end
			create l_integer_predicate.make_strict (server, l_statement)
			l_integer_predicate.retrieve (Void, Void)
			Result := not l_integer_predicate.array.is_empty
		end

	signature_ids_by_process (a_process_id: INTEGER): ROSE_LINEAR_ARRAY [G] is
			-- Collection of signatures in the test environment that have data for a_process_id
		local
			l_predicate: COLLECTING_PREDICATE [G]
			l_statement: STRING
		do
			from
				l_statement := ""
				types.start
			until
				types.off
			loop
				l_statement := l_statement
				+ "select distinct signature_id from " + table.data_name_prefix + "data_" + types.current_item.first + "_value where process_id = " + a_process_id.out
				+ " union "
				+ "select distinct signature_id from " + table.data_name_prefix + "data_" + types.current_item.first + "_value_history" + " where process_id = " + a_process_id.out
				if not types.at_last then
					l_statement := l_statement + " union "
				end
				types.forth
			end
			create l_predicate.make_strict (server, l_statement)
			l_predicate.retrieve (Void, Void)
			Result := l_predicate.array
		ensure
			non_void_result: Result /= Void
		end

	item_ids_by_process (a_process_id: INTEGER): ROSE_LINEAR_ARRAY [INTEGER] is
			-- Collection of item ids that are in the test environment for a_process_id
		local
			l_integer_predicate: INTEGER_PREDICATE
			l_statement: STRING
			l_string_predicate: STRING_PREDICATE
			l_expression: ROSE_EXPRESSION [ANY]
			l_signature: G
		do
			if testenv_summary_statistics_table_available then
				create l_string_predicate.make_strict (server, "select mnemonic from testenv_summary_statistics where process_id = " + a_process_id.out)
				l_string_predicate.retrieve (Void, Void)
			end
			if testenv_summary_statistics_table_available and then not l_string_predicate.array.is_empty then
				create Result.make_empty
				create l_signature
				from
					l_string_predicate.array.start
				until
					l_string_predicate.array.off
				loop
					l_expression := expression_from_mnemonic (l_string_predicate.array.current_item)
					if l_expression /= Void and then l_expression.variable_index /= 0 then
						Result.extend (l_expression.variable_index)
					end
					l_string_predicate.array.forth
				end
			else
				from
					l_statement := ""
					types.start
				until
					types.off
				loop
					l_statement := l_statement
					+ "select distinct item_id from " + table.data_name_prefix + "data_" + types.current_item.first + "_value WITH (READUNCOMMITTED) where process_id = " + a_process_id.out
					+ " union "
					+ "select distinct item_id from " + table.data_name_prefix + "data_" + types.current_item.first + "_value_history WITH (READUNCOMMITTED) where process_id = " + a_process_id.out
					if not types.at_last then
						l_statement := l_statement + " union "
					end
					types.forth
				end
				create l_integer_predicate.make_strict (server, l_statement)
				l_integer_predicate.retrieve (Void, Void)
				Result := l_integer_predicate.array
			end
			item_ids_by_process_cache.force (Result, a_process_id)
		ensure
			non_void_result: Result /= Void
		end

feature -- Statistics

	retrieved_testenv_statistics_by_process_id_and_mnemonic (a_process_id: INTEGER; a_mnemonic: STRING): TESTENV_STATISTICS [G] is
			-- Pre-computed statistics for a_mnemonic for a_process_id
		require
			valid_ids: a_process_id > 0 and then a_mnemonic /= Void
		local
			l_object_predicate: DATA_SERVER_STREAM_COMMAND
		do
			l_object_predicate := server.stream_command ("select node_object from testenv_summary_statistics where process_id = " + a_process_id.out + " and mnemonic = '" + a_mnemonic + "'")
			l_object_predicate.retrieve (Void)
			if l_object_predicate.query_result.count /= 0 then
				Result ?= safe_retrieve_from_string (l_object_predicate.query_result)
				if Result = Void then
					create Result.make (Void)
				end
			end
		end

	set_use_thread_for_storage (a_value: like use_thread_for_storage) is
			-- Set 'use_thread_for_storage' to 'a_value'
		do
			use_thread_for_storage := a_value
		end

	testenv_statistics (
		a_process: SUPPLIER_PROCESS;
		a_signatures: ROSE_LINEAR_ARRAY [G];
		a_expression: ROSE_EXPRESSION [ANY];
		a_progress_agent: PROCEDURE [ANY, TUPLE [DOUBLE]];
		a_item_count, a_progress_item_count: INTEGER
							 ): TESTENV_STATISTICS [G] is
			-- The statistics comparison object for these arguments
		require
			valid_process: a_process /= Void
			valid_expression: a_expression /= Void
			valid_counts: a_progress_agent /= Void implies (a_item_count >= a_progress_item_count and a_progress_item_count > 0)
		local
			l_signature: G
				--l_file: ROSE_FILE
			l_process_testenv_cache: HASH_TABLE [TESTENV_STATISTICS [G], STRING]
			l_validation_table: HASH_TABLE [TESTENV_VALIDATED_DIFFERENCES, INTEGER]
		do
			if testenv_statistics_cache = Void then
				create testenv_statistics_cache.make (100)
			end
			testenv_statistics_cache.search (a_process.process_id)
			if not testenv_statistics_cache.found then
				create l_process_testenv_cache.make (100)
				testenv_statistics_cache.force (l_process_testenv_cache, a_process.process_id)
			else
				l_process_testenv_cache := testenv_statistics_cache.found_item
			end
			l_process_testenv_cache.search (a_expression.mnemonic)
			if l_process_testenv_cache.found then
				Result := l_process_testenv_cache.found_item
			else
				process_clear_nonsensical_statistics_predicate.execute (Void, <<a_process.process_id, a_process.end_time>>)
				Result := retrieved_testenv_statistics_by_process_id_and_mnemonic (a_process.process_id, a_expression.mnemonic)
				if Result /= Void then
					l_process_testenv_cache.force (Result, a_expression.mnemonic)
					if a_progress_agent /= Void then
						a_progress_agent.call ([a_progress_item_count / a_item_count])
					end
				elseif a_signatures /= Void then
					item_ids_by_process_cache.search (a_process.process_id)
					if (item_ids_by_process_cache.found and item_ids_by_process_cache.found_item.has (a_expression.variable_index))
						or else (not item_ids_by_process_cache.found and process_contains_this_item (a_expression.variable_index, a_process.process_id))
					 then
						 Global.start_busy ("Computing Statistics for " + a_expression.mnemonic)
						 --print (a_expression.mnemonic + "%N")
						create Result.make (a_expression)
						Result.set_count (a_signatures.count)
						clear_penultimate_data_caches
						if a_expression.type.is_equal (Internal.Real_Type)
							or a_expression.type.is_equal (Internal.Double_Type)
						 then
							if table.data_type_enabled (Real_String, False) then
								real_type_cache.load_penultimate_data_item_cache (a_expression.variable_index, a_process)
							end
							if table.data_type_enabled (Double_String, False) then
								double_type_cache.load_penultimate_data_item_cache (a_expression.variable_index, a_process)
							end
						elseif a_expression.type.is_equal (Internal.Integer_Type) then
							integer_type_cache.load_penultimate_data_item_cache (a_expression.variable_index, a_process)
						elseif a_expression.type.is_equal ({ROSE_CONSTANT_VALUES}.string_type) then
							string_type_cache.load_penultimate_data_item_cache (a_expression.variable_index, a_process)
						end
						l_validation_table := validation_table (a_expression, a_expression.variable_index)
						from
							a_signatures.start
						until
							a_signatures.off
						loop
							if a_signatures.index \\ 100 = 0 and then a_progress_agent /= Void then
								a_progress_agent.call ([(a_progress_item_count - 1) / a_item_count + (a_signatures.index / a_signatures.count) / a_item_count])
							end
							l_signature := a_signatures.current_item.twin
							l_signature.set_last_process_id (a_process.process_id)
							if a_expression.type.is_equal (Internal.Real_Type)
								or a_expression.type.is_equal (Internal.Double_Type)
							 then
								if real_data_enabled then
									retrieve_differences (l_signature, Real_String, a_expression, Result, l_validation_table)
								end
								if double_data_enabled then
									retrieve_differences (l_signature, Double_String, a_expression, Result, l_validation_table)
								end
							elseif a_expression.type.is_equal (Internal.Integer_Type) then
								retrieve_differences (l_signature, Integer_String, a_expression, Result, l_validation_table)
							elseif a_expression.type.is_equal ({ROSE_CONSTANT_VALUES}.String_Type) then
								retrieve_differences (l_signature, String_String, a_expression, Result, Void)
							end
							a_signatures.forth
						end
						l_process_testenv_cache.force (Result, a_expression.mnemonic)
						if {PLATFORM}.is_thread_capable and then use_thread_for_storage then
							storage_mutex.lock
							create statistics_to_persist.make (agent store_testenv_statistics (Result, a_expression.mnemonic.twin, a_expression.variable_index, a_process.deep_twin))
							storage_mutex.unlock
						else
							store_testenv_statistics (Result, a_expression.mnemonic.twin, a_expression.variable_index, a_process)
						end
						Global.end_busy
					end
				end
			end
		ensure
			non_void_result: Result /= Void
		end

	use_thread_for_storage: BOOLEAN
			-- Use a separate thread to perform statistics storage
	statistics_to_persist: WORKER_THREAD
			-- A thread to handle the statistics persistence

	persist_last_statistics is
			-- Persist the last computed statistics
		do
			if statistics_to_persist /= Void then
				statistics_to_persist.launch
			end
		end

	storage_mutex: MUTEX is
			-- Mutex to control access to persistence mechanism
		indexing
			once_status: global
		once
			create Result.make
		ensure
			nonvoid: Result /= Void
		end

	store_testenv_statistics (a_statistics: TESTENV_STATISTICS [G]; a_mnemonic: STRING; a_variable_index: INTEGER; a_process: SUPPLIER_PROCESS) is
			-- Prepare and persist 'a_statistics'
		require
			valid_statistics: a_statistics /= Void
			valid_expression: a_mnemonic /= Void and a_variable_index > 0
		local
			l_stream: STREAM
		do
			if {PLATFORM}.is_thread_capable then
				storage_mutex.lock
			end
				--print (Global.interpreted ("($now) storing: " + a_mnemonic + "%N"))
			strip_out_invalid_records_for_comparison (a_statistics, a_variable_index)
			a_statistics.sort_differences
			a_statistics.r_square.do_nothing
			a_statistics.truncate_for_storage
			create l_stream.make_with_size (100_000)
			l_stream.independent_store (a_statistics)
			testenv_storable_predicate.execute (Void, <<a_process.process_id, a_process.file_date, a_mnemonic, a_statistics.r_square, 0.0, a_statistics.sum_of_squared_differences, l_stream>>)
			if not testenv_storable_predicate.success then
				error.abort ("store_testenv_statistics", "Failed to execute testenv_summary_statistics store", Void)
			end
			testenv_storable_predicate.data_server.commit
				--print (Global.interpreted ("($now) stored: " + a_mnemonic + "%N"))
			if {PLATFORM}.is_thread_capable then
				storage_mutex.unlock
			end
			l_stream.close
		end

	validation_table (a_expression: ROSE_EXPRESSION [ANY]; a_variable_index: INTEGER): HASH_TABLE [TESTENV_VALIDATED_DIFFERENCES, INTEGER] is
			-- Remove targets from the statistics that do not meet the validity requirements as specified by the validity agent
		local
			l_validated_differences: COLLECTING_PREDICATE [TESTENV_VALIDATED_DIFFERENCES]
		do
			if testenv_validated_differences_table_available then
				create l_validated_differences.make (server, "select * from testenv_validated_differences where item_id = ? order by timestamp")
				if l_validated_differences.success then
					if a_expression /= Void then
						l_validated_differences.retrieve (Void, <<a_expression.variable_index>>)
					else
						l_validated_differences.retrieve (Void, <<a_variable_index>>)
					end
					create Result.make (1_000)
					if l_validated_differences.success and then not l_validated_differences.array.is_empty then
						from
							l_validated_differences.array.start
						until
							l_validated_differences.array.off
						loop
							Result.force (l_validated_differences.array.current_item, l_validated_differences.array.current_item.signature_id)
							l_validated_differences.array.forth
						end
					end
				end
			end
		end

feature {NONE} -- Sorting agents

	descending_process_ids (a, b: SUPPLIER_PROCESS): BOOLEAN is
			-- Order by process_id descending
		do
			Result := b.process_id < a.process_id
		end

feature {NONE} -- Implementation

	create_caches (a_data_name: STRING; a_source_server, a_target_server: SQL_DATA_SERVER) is
			-- Initialize the type caches
		do
			if table.data_type_enabled (Real_String, False) then
				create real_type_cache.make (a_data_name, a_source_server, a_target_server, "real")
			end
			if table.data_type_enabled (Double_String, False) then
				create double_type_cache.make (a_data_name, a_source_server, a_target_server, "double")
			end
			if table.data_type_enabled (Integer_String, False) then
				create integer_type_cache.make (a_data_name, a_source_server, a_target_server, "integer")
			end
			if table.data_type_enabled (String_String, False) then
				create string_type_cache.make (a_data_name, a_source_server, a_target_server, "string")
			end
		end

	retrieve_differences (
		a_signature: G;
		a_type: STRING;
		a_expression: ROSE_EXPRESSION [ANY];
		a_testenv_statistics: TESTENV_STATISTICS [G];
		a_validation_table: HASH_TABLE [TESTENV_VALIDATED_DIFFERENCES, INTEGER]
								) is
			-- Retrieve differences for 'a_signature' for 'a_expression'
		require
			valid_type: a_type /= Void
		local
			l_double_data: ROSE_LINEAR_ARRAY [DATA_RECORD [DOUBLE]]
			l_real_data: ROSE_LINEAR_ARRAY [DATA_RECORD [REAL]]
			l_integer_data: ROSE_LINEAR_ARRAY [DATA_RECORD [INTEGER]]
			l_string_data: ROSE_LINEAR_ARRAY [DATA_RECORD [STRING]]
			l_test_data: ROSE_LINEAR_ARRAY [DATA_RECORD [ANY]]
			l_testenv_value, l_production_value: DOUBLE
			l_production_any: ANY
			l_testenv_string_value, l_production_string_value: STRING
			l_comparison: DATA_COMPARISON [G]
			l_string_comparison: DATA_STRING_COMPARISON [G]
			l_found: BOOLEAN
		do
			if a_signature.signature_id = 10765989 and a_expression.variable_index = 400 then
				do_nothing
			end
			if a_type.is_equal (Real_String) then
				l_real_data ?= real_type_cache.data_items_by_signature (a_signature)
				l_test_data ?= l_real_data
			elseif a_type.is_equal (Double_String) then
				l_double_data ?= double_type_cache.data_items_by_signature (a_signature)
				l_test_data ?= l_double_data
			elseif a_type.is_equal (Integer_String) then
				l_integer_data ?= integer_type_cache.data_items_by_signature (a_signature)
				l_test_data ?= l_integer_data
			elseif a_type.is_equal (String_String) then
				l_string_data ?= string_type_cache.data_items_by_signature (a_signature)
				l_test_data ?= l_string_data
			end
			if l_test_data /= Void then
				from
					l_test_data.start
				until
					l_test_data.after or l_found
				loop
					if l_test_data.current_item.item_id = a_expression.variable_index then
						l_found := True
						if a_type.is_equal (Real_String) then
							l_testenv_value := l_real_data.current_item.value
						elseif a_type.is_equal (Double_String) then
							l_testenv_value := l_double_data.current_item.value
						elseif a_type.is_equal (Integer_String) then
							l_testenv_value := l_integer_data.current_item.value
							l_testenv_value := repaired_integer_item_for_comparison (l_testenv_value.truncated_to_integer, l_test_data.current_item.item_id)
						elseif a_type.is_equal (String_String) then
							l_testenv_string_value := l_string_data.current_item.value
						end
						if a_type.is_equal (Real_String) then
							l_production_value := real_type_cache.penultimate_data_item_by_signature (a_signature, l_test_data.current_item.item_id)
						elseif a_type.is_equal (Double_String) then
							l_production_value := double_type_cache.penultimate_data_item_by_signature (a_signature, l_test_data.current_item.item_id)
						elseif a_type.is_equal (Integer_String) then
							l_production_value := integer_type_cache.penultimate_data_item_by_signature (a_signature, l_test_data.current_item.item_id)
							l_production_value := repaired_integer_item_for_comparison (l_production_value.truncated_to_integer, l_test_data.current_item.item_id)
						elseif a_type.is_equal (String_String) then
							l_production_any := string_type_cache.penultimate_data_item_by_signature (a_signature, l_test_data.current_item.item_id)
							if l_production_any /= Void then
								l_production_string_value := l_production_any.out
							end
						end
						if a_type.is_equal (String_String) then
							create l_string_comparison.make (a_signature, l_production_string_value, l_testenv_string_value, a_expression)
							if l_production_string_value /= Void and l_testenv_string_value /= Void then
								a_testenv_statistics.differences.extend (l_string_comparison)
							else
								a_testenv_statistics.availability_problems.extend (l_string_comparison)
							end
						else
							create l_comparison.make (a_signature, l_production_value, l_testenv_value, a_expression)
							a_testenv_statistics.insert_difference (l_comparison, a_validation_table)
						end
					end
					if l_found then
						l_test_data.remove
					else
						l_test_data.forth
					end
				end
			end
		end

	repaired_integer_item_for_comparison (a_value: INTEGER; a_item: INTEGER): DOUBLE is
			-- Strip out values in the integer bitmaps that are not used for comparison purposes
			-- This is used for market and fundamental flags in mondat
		do
			if a_item = 581 then
					-- For mondat FF1, ignore standardization bit
				if a_value = Undefined.Undefined_Integer then
					Result := 0
				else
					Result := a_value.bit_and (item_581_bitmask)
				end
			elseif a_item = 582 then
					-- For mondat FF2, ignore QEPS, months since EPS bits and valuation baddata bits
				if a_value = Undefined.Undefined_Integer then
					Result := 0
				else
					Result := a_value.bit_and (item_582_bitmask)
				end
			elseif a_item = 583 then
					-- For mondat MF1, ignore bogie membership bits and i_exchange bits
				if a_value = Undefined.Undefined_Integer then
					Result := 0
				else
					Result := a_value.bit_and (item_583_bitmask)
				end
			elseif a_item = 1504
				or else a_item = 1528
					or else a_item = 1527
						or else a_item = 1526
							or else a_item = 1522
								or else a_item = 1525
			 then
					-- For mondat Z_DATA_BREAK or Z_IF_*, change -1 to 1 (VMS True to Windows True)
				if a_value = -1 then
					Result := 1
				else
					Result := a_value
				end
			elseif a_item >= 766 and a_item <= 785 then
				if a_value = Undefined.Undefined_integer then
					Result := 0
				else
					Result := a_value
				end
			else
				Result := a_value
			end
		end

	strip_out_invalid_records_for_comparison (a_statistics: TESTENV_STATISTICS [G]; a_variable_index: INTEGER) is
			-- Remove targets from the statistics that do not meet the validity requirements as specified by the validity agent
		local
			l_differences: ROSE_LINEAR_ARRAY [DATA_COMPARISON [G]]
			l_validation_table: HASH_TABLE [TESTENV_VALIDATED_DIFFERENCES, INTEGER]
		do
			if validity_agent /= Void then
				from
					create l_differences.make_empty
					a_statistics.differences.start
				until
					a_statistics.differences.off
				loop
					if validity_agent.item ([a_statistics.differences.current_item.target, a_statistics.differences.current_item, a_variable_index]) then
						l_differences.extend (a_statistics.differences.current_item)
					end
					a_statistics.differences.forth
				end
				l_validation_table := validation_table (Void, a_variable_index)
				a_statistics.reset_differences
				l_differences.do_all (agent a_statistics.insert_difference (?, l_validation_table))
			end
		end

	create_testenv_summary_statistics_table is
			-- Create the table that stores the computed summary statistics by process_id and item_id
		require
			server_not_void: server /= Void
		local
			l_exception: DEVELOPER_EXCEPTION
		do
			server.sql_execute_immediate (
				"CREATE TABLE [dbo].[testenv_summary_statistics] ("
				+ "[process_id] [int] NOT NULL ,"
				+ "[mnemonic] [varchar] (50) COLLATE SQL_Latin1_General_CP1_CI_AS NOT NULL ,"
				+ "[node_object] [image] NOT NULL ,"
				+ "[timestamp] [datetime] NULL ,"
				+ "[file_date] [datetime] NOT NULL ,"
				+ "[correlation] [float] NULL ,"
				+ "[avg_differences_ratio] [float] NULL ,"
				+ "[sum_square_differences] [float] NULL "
				+ ") ON [PRIMARY] TEXTIMAGE_ON [PRIMARY]"
												  )
			if not testenv_summary_statistics_table_available then
				create l_exception
				l_exception.set_message ("Cannot create table testenv_summary_statistics on "+server.definition.full_path)
				l_exception.raise
			end
		end

	create_testenv_validated_differences_table is
			-- Create the table that stores the validated differences
		require
			server_not_void: server /= Void
		local
			l_exception: DEVELOPER_EXCEPTION
		do
			server.sql_execute_immediate (
				"CREATE TABLE [dbo].[testenv_validated_differences] ( "
				+ "[signature_id] [int] NOT NULL , "
				+ "[item_id] [int] NOT NULL , "
				+ "[x] [float] NULL , "
				+ "[y] [float] NULL , "
				+ "[comment] [varchar] (512) COLLATE SQL_Latin1_General_CP1_CI_AS NOT NULL , "
				+ "[user_id] [varchar] (50) COLLATE SQL_Latin1_General_CP1_CI_AS NOT NULL , "
				+ "[timestamp] [datetime] NOT NULL  "
				+ ") ON [PRIMARY] "
												  )
			server.sql_execute_immediate ("CREATE  CLUSTERED  INDEX [IX_testenv_validated_differences] ON [dbo].[testenv_validated_differences]([item_id], [signature_id]) ON [PRIMARY] ")
			server.sql_execute_immediate (
				"ALTER TABLE [dbo].[testenv_validated_differences] ADD  "
				+ "CONSTRAINT [DF_testenv_validated_differences_user_id] DEFAULT (suser_sname()) FOR [user_id], "
				+ "CONSTRAINT [DF_testenv_validated_differences_timestamp] DEFAULT (getutcdate()) FOR [timestamp] "
												  )
			if not testenv_validated_differences_table_available then
				create l_exception
				l_exception.set_message ("Cannot create table testenv_validated_differences on "+server.definition.full_path)
				l_exception.raise
			end
		end

	testenv_summary_statistics_table_available: BOOLEAN is
			-- Is the summary statistics table available
		require
			server_not_void: server /= Void
		local
			l_predicate: ANONYMOUS_PREDICATE
		do
			create l_predicate.make_using_server (server, "exec sp_tables 'testenv_summary_statistics'")
			l_predicate.retrieve (Void, Void)
			if l_predicate.success then
				Result := (l_predicate.array.count > 0)
			end
		end

	testenv_validated_differences_table_available: BOOLEAN is
			-- Is the validated differences table available
		require
			server_not_void: server /= Void
		local
			l_predicate: ANONYMOUS_PREDICATE
		do
			create l_predicate.make_using_server (server, "exec sp_tables 'testenv_validated_differences'")
			l_predicate.retrieve (Void, Void)
			if l_predicate.success then
				Result := (l_predicate.array.count > 0)
			end
		end

feature {NONE} -- Predicates

	process_clear_nonsensical_statistics_predicate: EXECUTION_PREDICATE is
			-- Delete computed statistics where timestamps are before the process's start time
		once
			if not testenv_summary_statistics_table_available then
				create_testenv_summary_statistics_table
			end
			create Result.make (server, "delete from testenv_summary_statistics where process_id = ? and timestamp < ?")
		end

	process_clear_predicate: EXECUTION_PREDICATE is
			-- Clear the summary statistics for a process_id
		once
			if not testenv_summary_statistics_table_available then
				create_testenv_summary_statistics_table
			end
			create Result.make (server, "delete from testenv_summary_statistics where process_id = ?")
		end

	testenv_storable_predicate: EXECUTION_PREDICATE is
			-- Insert testenv summary statistics into the summary table
		indexing
			once_status: global
		local
			l_server: SQL_DATA_SERVER
			l_definition: ROSE_DATABASE_DEFINITION
		once
			if not testenv_summary_statistics_table_available then
				create_testenv_summary_statistics_table
			end
			create l_definition.make ("hardcoded", server.logical, "sql", server.node, server.database)
			l_server ?= pool.get_server_from_definition (l_definition)
			create Result.make (l_server, "insert into testenv_summary_statistics (process_id,file_date,mnemonic,correlation,avg_differences_ratio,sum_square_differences,node_object) values (?,?,?,?,?,?,?)")
			if not Result.success then
				error.abort ("testenv_storable_predicate", "Failed to create testenv_summary_statistics store", Void)
			end
		end

feature {STID_TESTENV_VIEW} -- Implementation

	target_server: SQL_DATA_SERVER
			-- The server against which we are comparing data

feature {NONE} -- Data stores and caches

	server: SQL_DATA_SERVER is
			-- The server that has the data which we will be comparing
		do
			Result := table.signature_item_server
		end

	types: ROSE_LINEAR_ARRAY [PAIR [STRING, INTEGER]] is
			-- List of available enabled types in this database
		do
			Result := table.enabled_regular_base_types
		end

	validity_agent: FUNCTION [ANY, TUPLE [G], BOOLEAN]
			-- A test to determine if a target is valid for inclusion in the comparison statistics

	table: SIGNATURE_TAGGED_TABLE_STRUCTURE
			-- A table structure object for the database containing the data which we will be comparing

	database: SIGNATURE_TAGGED_ITEM_DATABASE [G]
			-- The signature tagged database from which we access processing history (no data retrieval)

	item_ids_by_process_cache: HASH_TABLE [ROSE_LINEAR_ARRAY [INTEGER], INTEGER]
			-- Colletion of item ids keyed by process_id

	testenv_statistics_cache: HASH_TABLE [HASH_TABLE [TESTENV_STATISTICS [G], STRING], INTEGER]
			-- Cache of computed statistics keyed by process_id keyed by item_id

	sids_temp_table_created: BOOLEAN
			-- Has this temp table already been created

	real_type_cache: STID_TESTENV_TYPE_CACHE [G, REAL]
			-- Cache for real type data

	double_type_cache: STID_TESTENV_TYPE_CACHE [G, DOUBLE]
			-- Cache for double type data

	string_type_cache: STID_TESTENV_TYPE_CACHE [G, STRING]
			-- Cache for string type data

	integer_type_cache: STID_TESTENV_TYPE_CACHE [G, INTEGER]
			-- Cache for integer type data

	real_data_enabled: BOOLEAN
			-- Does this database have real data

	double_data_enabled: BOOLEAN
			-- Does this database have double data

feature {NONE} -- Constants

	Real_String: STRING is "real"

	Double_String: STRING is "double"

	Integer_String: STRING is "integer"

	String_String: STRING is "string"


	item_581_bitmask: INTEGER is 0xDFFFFFFF
			-- bitmask for item 581

	item_582_bitmask: INTEGER is 0x007E003F
			-- bitmask for item 582

	item_583_bitmask: INTEGER is 0x0000FFE0
			-- bitmask for item 583

feature {NONE} -- Helpers

	safe_retrieve_from_string (a_result: STRING): ANY is
			-- Retrieve a storable from a_result in a safe way by turning off garbage collection
		local
			l_retry: BOOLEAN
			l_collecting: BOOLEAN
			l_rose_stream: ROSE_STREAM
		do
			if not l_retry then
				if Memory.collecting then
					l_collecting := True
					Memory.collection_off
				end
				create l_rose_stream
				Result ?= l_rose_stream.retrieve_from_string (a_result)
				if l_collecting then
					Memory.collection_on
				end
			end
		rescue
			if l_collecting then
				Memory.collection_on
			end
			l_retry := True
			retry
		end

	Undefined: ROSE_LEGACY_UNDEFINED is
			-- Access to undefined definitions
		once
			create Result
		end

	Memory: MEMORY is
			-- Interface to garbage collector
		once
			create Result
		end

	Internal: INTERNAL is
			-- Access to Internal
		once
			create Result
		end

	pid: STRING is
			-- This application's pid
		once
			Result := Global.app.pid.out
		end

end
