indexing

	description: "string manipulation routines"
	keywords: ""
	author: "Hadi Taheri"
	date: "$Date: 2010/05/11 $"


	-- This class may be integrated into ROSE_FORMATTING_SERVICE

class

	STRING_ALGEBRA

inherit

	ROSE_IEEE_UNDEFINED

feature	-- single strings

	blank: CHARACTER is ' '
	comma: CHARACTER is ','

	blanks_n (a_number_of_blanks: INTEGER): STRING is
		-- Returns a string of n blanks
		do
			if  a_number_of_blanks > 0 then
				create Result.make_filled (blank, a_number_of_blanks)
			else
				create Result.make_empty
			end
		end

	fixed_length_string (a_string: STRING; a_desired_string_length: INTEGER): STRING is
		-- Returns a_string of the requested length, padding with blanks at the tail end or truncating the tail, as called for.
		local
			l_original_count: INTEGER
			l_num_blanks: INTEGER
			l_num_extra_characters: INTEGER
			i: INTEGER
		do
			if a_string = Void then
				create Result.make (a_desired_string_length)
				Result.fill_blank
			else
				create Result.make_from_string (a_string)
				if Result.count < a_desired_string_length then
					l_num_blanks := a_desired_string_length - Result.count
					l_original_count := Result.count
					from i := 1 until i > l_num_blanks loop
						Result.insert_character (blank, l_original_count + i)
						i := i + 1
					end
				elseif Result.count > a_desired_string_length then
					l_num_extra_characters := Result.count - a_desired_string_length
					Result.remove_tail (l_num_extra_characters)
				end
			end
		ensure
			expectation: Result.count = a_desired_string_length
		end

	fixed_length_string_head (a_string: STRING; a_desired_string_length: INTEGER): STRING is
		-- Returns a_string of the requested length, padding with blanks at the head end or truncating the head, as called for.
		local
			l_original_count: INTEGER
			l_num_blanks: INTEGER
			l_num_extra_characters: INTEGER
			i: INTEGER
		do
			if a_string = Void then
				create Result.make (a_desired_string_length)
				Result.fill_blank
			else
				create Result.make_from_string (a_string)
				if Result.count < a_desired_string_length then
					l_num_blanks := a_desired_string_length - Result.count
					l_original_count := Result.count
					from i := 1 until i > l_num_blanks loop
						Result.prepend_character (blank)
						i := i + 1
					end
				elseif Result.count > a_desired_string_length then
					l_num_extra_characters := Result.count - a_desired_string_length
					Result.remove_head (l_num_extra_characters)
				end
			end
		ensure
			expectation: Result.count = a_desired_string_length
		end

	string_from_string_array_with_given_separator (a_array: ROSE_LINEAR_ARRAY [STRING]; a_separator: STRING): STRING is
		-- Convert an array of strings into a single string, with elements separated by a given string.
		do
			create Result.make_empty
			from a_array.start until a_array.off loop
				if not a_array.at_first then
					if a_separator /= Void then
						Result.append (a_separator)
					end
				end
				Result.append (a_array.current_item)
				a_array.forth
			end
		end

	string_from_string_array (a_array: ROSE_LINEAR_ARRAY [STRING]): STRING is
		-- Convert an array of strings into a single string, with elements separated by a comma and space.
		local
			l_separator: STRING
		do
			l_separator := ", "
			Result := string_from_string_array_with_given_separator (a_array, l_separator)
		end

	string_with_character_replaced (a_string: STRING; a_src_char, a_dst_char: CHARACTER): STRING
			-- Replace all occurences of 'a_src_char' in 'a_string' with 'a_dst_char'. Useful when creating a CSV file.
		local
			i: INTEGER
		do
			if a_string /= Void then
				Result := a_string.twin
				from i := 1 until i > Result.count loop
					if Result.item (i) = a_src_char then
						Result.item (i) := a_dst_char
					end
					i := i + 1
				end
			end
		end

feature -- two strings

	concatenated_strings_even_if_Void (a_string1, a_separator, a_string2: STRING): STRING is
		-- concatenate two strings, with a separator in between them, even it they're void
		do
			if a_string1 = Void or else a_string1.count = 0 then
				create Result.make_from_string (a_string2)
			elseif a_string2 = Void or else a_string2.count = 0 then
				create Result.make_from_string (a_string1)
			else
				Result := a_string1 + a_separator + a_string2
			end
		end

	concatenated_strings3_even_if_Void (a_string1, a_separator, a_string2, a_separator2, a_string3: STRING): STRING is
		-- concatenate three strings, with separators in between them, even it they're void
		do
			Result := concatenated_strings_even_if_Void (concatenated_strings_even_if_Void (a_string1, a_separator, a_string2), a_separator2, a_string3)
		end

feature -- string comparison

	string_is_all_blank (a_string: STRING): BOOLEAN is
		-- Determine whether a string consists of nothing but blanks.
		do
			if a_string.count > 0 then
				Result := a_string.occurrences (blank) = a_string.count
			end
		end

	strings_are_essentially_equal (a_string1, a_string2: STRING): BOOLEAN is
			-- Determine whether two strings are the same irrespective of leading and trailing blanks and case.
		local
			l_string1: STRING
			l_string2: STRING
		do
			l_string1 := string_with_pruned_leading_and_trailing_blanks (a_string1)
			l_string1.to_lower

			l_string2 := string_with_pruned_leading_and_trailing_blanks (a_string2)
			l_string2.to_lower

			Result := l_string1.is_equal (l_string2)
		end

	string_with_pruned_leading_and_trailing_blanks (a_string: STRING): STRING is
		-- Returns a string that is the same as the argument excpet that it has no leading and trailing blanks
		do
			create Result.make_from_string (a_string)
			Result.prune_all_leading (blank)
			Result.prune_all_trailing (blank)
		end

	strings_are_essentially_equal_except_for_blanks_commas (a_string1, a_string2: STRING): BOOLEAN is
		-- Returns True if the two string arguments are the same if blanks and commas are ignored.
		local
			l_string1: STRING
			l_string2: STRING
		do
			create l_string1.make_from_string (a_string1)
			l_string1.prune_all (blank)
			l_string1.prune_all (comma)
			l_string1.to_lower

			create l_string2.make_from_string (a_string2)
			l_string2.prune_all (blank)
			l_string2.prune_all (comma)
			l_string2.to_lower

			Result := l_string1.is_equal (l_string2)
		end

	strings_are_equal_or_both_void (a_string_1, a_string_2: STRING): BOOLEAN is
		-- Returns True if both string arguments oare Void or identical
		do
			if a_string_1 /= Void and a_string_2 /= Void then
				Result := a_string_1.is_equal (a_string_2)
			elseif a_string_1 = Void and a_string_2 = Void then
				Result := True
			end
		end

	strings_heads_are_equal_or_both_void (a_string_1, a_string_2: STRING; a_max_count: INTEGER): BOOLEAN is
		-- Returns True if the first 'a_max_count' characters of the two string arguments are both equal or otherwise both Void.
		local
			l_head_1: STRING
			l_head_2: STRING
		do
			if a_string_1 /= Void and a_string_2 /= Void then
				l_head_1 := a_string_1
				if l_head_1.count > a_max_count then
					l_head_1 := l_head_1.substring (1, a_max_count)
				end
				l_head_2 := a_string_2
				if l_head_2.count > a_max_count then
					l_head_2 := l_head_2.substring (1, a_max_count)
				end
				Result := l_head_1.is_equal (l_head_2)
			elseif a_string_1 = Void and a_string_2 = Void then
				Result := True
			end
		end

	strings_heads_are_lower_case_equal_or_both_void (a_string_1, a_string_2: STRING; a_max_count: INTEGER): BOOLEAN is
		-- self-explanatory. Compare to strings_heads_are_equal_or_both_void()
		local
			l_head_1: STRING
			l_head_2: STRING
		do
			if a_string_1 /= Void and a_string_2 /= Void then
				create l_head_1.make_from_string (a_string_1)
				if l_head_1.count > a_max_count then
					l_head_1 := l_head_1.substring (1, a_max_count)
				end
				l_head_1.to_lower

				create l_head_2.make_from_string (a_string_2)
				if l_head_2.count > a_max_count then
					l_head_2 := l_head_2.substring (1, a_max_count)
				end
				l_head_2.to_lower

				Result := l_head_1.is_equal (l_head_2)
			elseif a_string_1 = Void and a_string_2 = Void then
				Result := True
			end
		end

	one_string_essentially_has_the_other_string_or_both_void (a_string_1, a_string_2: STRING; a_case_sensitive: BOOLEAN): BOOLEAN is
		-- Returns True if both string arguments are Void or if one is contained in the other. Case-sensitivity is an option.
		local
			l_string_1: STRING
			l_string_2: STRING
		do
			if a_string_1 /= Void and a_string_2 /= Void then

				l_string_1 := string_with_pruned_leading_and_trailing_blanks (a_string_1)
				l_string_2 := string_with_pruned_leading_and_trailing_blanks (a_string_2)

				-- ignore case in the comparison.
				if not a_case_sensitive then
					l_string_1.to_lower
					l_string_2.to_lower
				end

				if l_string_1.count > l_string_2.count then
					if l_string_2.count > 0 then -- we don't want a vaccuous containment.
						Result := l_string_1.has_substring (l_string_2)
					end
				else
					if l_string_1.count > 0 then -- we don't want a vaccuous containment.
						Result := l_string_2.has_substring (l_string_1)
					end
				end
			elseif a_string_1 = Void and a_string_2 = Void then
				Result := True
			end

			-- A more complex variation on this feature would be to remove the phrase "and" and the character "&" before calling has_substring ...
                        -- Other substrings that could be considered identical are: "international", "intl"
                        --                                                          "telecommunication", "telecom"
		end


feature -- multiple strings

	string_array_from_fixed_field_string (a_string: STRING; a_field_width: INTEGER; a_remove_spaces: BOOLEAN): ROSE_LINEAR_ARRAY [STRING] is
		-- Convert a string into an array of strings. The input string is assumed to be of fixed field format.
		local
			l_start_index: INTEGER
			l_substring: STRING
		do
			from
				create Result.make_empty
				l_start_index := 1
			until
				l_start_index > a_string.count
			loop
				l_substring := a_string.substring (l_start_index, l_start_index + a_field_width - 1)
				if a_remove_spaces then
					l_substring.prune_all_leading (' ')
					l_substring.prune_all_trailing (' ')
				end
				Result.extend (l_substring)

				l_start_index := l_start_index + a_field_width
			end
		end

	string_array_from_string_containing_delimiter (a_string: STRING; a_delimiter: STRING; a_discard_delimiter: BOOLEAN): ROSE_LINEAR_ARRAY [STRING] is
		-- Convert a string into an array of strings. The input string is assumed to contain delimiter characters, which may or may not be discarded, with the choice determined by input argument.
		local
			l_end_index_plus_1: INTEGER
			l_end_index: INTEGER
			l_substring: STRING
			l_start_index: INTEGER
		do
			l_end_index_plus_1 := a_string.substring_index (a_delimiter, 1)
			create Result.make_empty
			if l_end_index_plus_1 = 0 then -- delimeter not found
				Result.extend (a_string)
			else
				if a_discard_delimiter then
					l_end_index := l_end_index_plus_1 - 1
				else
					l_end_index := l_end_index_plus_1
				end
				if l_end_index < 1 then
					Result.extend (Void)
				else
					l_substring := a_string.substring (1, l_end_index)
					Result.extend (l_substring)
					l_start_index := l_end_index_plus_1 + a_delimiter.count
					if a_string.count >= l_start_index then
						-- use recursion
						Result.append_array (string_array_from_string_containing_delimiter (a_string.substring (l_start_index, a_string.count), a_delimiter, a_discard_delimiter))
					end
				end
			end
		end

	strings_containing_matching_string (a_strings: ROSE_LINEAR_ARRAY [STRING]; a_match_string: STRING): ROSE_LINEAR_ARRAY [STRING] is
		-- Return those elements in a_strings which contain a_match_string
		do
			create Result.make_empty
			from a_strings.start until a_strings.off loop
				if a_strings.current_item.has_substring (a_match_string) then
					Result.extend (a_strings.current_item)
				end
				a_strings.forth
			end
		end

	strings_not_containing_matching_string (a_strings: ROSE_LINEAR_ARRAY [STRING]; a_match_string: STRING): ROSE_LINEAR_ARRAY [STRING] is
		-- Return those elements in a_strings which do NOT contain a_match_string
		do
			create Result.make_empty
			from a_strings.start until a_strings.off loop
				if not a_strings.current_item.has_substring (a_match_string) then
					Result.extend (a_strings.current_item)
				end
				a_strings.forth
			end
		end

	unique_strings (a_strings: ROSE_LINEAR_ARRAY [STRING]): ROSE_LINEAR_ARRAY [STRING] is
		-- Return those elements in a_strings which are unique (i.e. remove duplicates)
		local
			l_ith_string: STRING
			l_curr_string: STRING
			i: INTEGER
			l_is_duplicated: BOOLEAN
		do
			create Result.make_empty
			if a_strings.count = 1 then
				Result.extend (a_strings.first)
			else
				from a_strings.start until a_strings.off loop
					l_curr_string := a_strings.current_item
					l_is_duplicated := False
					from i := a_strings.index + 1 until i > a_strings.upper	or l_is_duplicated loop
						l_ith_string := a_strings.item (i)
						if l_ith_string /= Void and then l_ith_string.is_equal (l_curr_string) then
							l_is_duplicated := True
						end
						i := i + 1
					end
					if not l_is_duplicated then
						Result.extend (l_curr_string)
					end
					a_strings.forth
				end
			end
		end

feature -- STRINGS comprised of DOUBLE's

	tightly_spaced_double_with_truncated_decimal_places (a_value: DOUBLE; a_width, a_decimal_places_count: INTEGER): STRING is
		-- there will be no leading or trailing blanks.
		local
			l_format_double: FORMAT_DOUBLE
		do
			create l_format_double.make (a_width, a_decimal_places_count)
			Result := l_format_double.formatted (a_value)
			Result.prune_all_leading (' ')
			Result.prune_all_trailing (' ')
		end

feature -- display

	displayed_string_even_if_void (a_string: STRING): STRING is
		do
			if a_string /= Void then
				Result := a_string
			elseif dislayed_string_when_void /= Void then
				Result := dislayed_string_when_void
			else
				Result := Void_string
			end
		end

	set_displayed_string_when_void (a_value: STRING) is
		require
			assumption: a_value /= Void
		do
			create dislayed_string_when_void.make_from_string (a_value)
		end

	dislayed_string_when_void: STRING
	Void_string: STRING is "Void"

feature	-- integer formatting

	formatted_integer (a_value: INTEGER; a_width: INTEGER): STRING is
		do
			if a_width /= saved_integer_width then
				saved_format_integer.set_width (a_width)
				saved_integer_width := a_width
			end

			if is_defined_integer (a_value) then
				Result := saved_format_integer.formatted (a_value)
			else
				Result := blanks_n (a_width)
			end
		end
	saved_integer_width: INTEGER

	formatted_integer_no0 (a_value: INTEGER; a_width: INTEGER): STRING is
		-- no0 means not allowed to print a 0
		do
			if a_value = 0 then
				Result := blanks_n (a_width)
			else
				Result := formatted_integer (a_value, a_width)
			end
		end

	formatted_integer_noNAN (a_value: INTEGER; a_width: INTEGER): STRING is
		-- noNAN means not allowed to print a QNAN
		do
			if not is_defined_integer (a_value) then
				Result := blanks_n (a_width)
			else
				Result := formatted_integer (a_value, a_width)
			end
		end

	formatted_integer_no0_noNAN (a_value: INTEGER; a_width: INTEGER): STRING is
		-- no0_noNAN means not allowed to print a 0 nor a QNAN
		do
			if not is_defined_integer (a_value) or else a_value = 0 then
				Result := blanks_n (a_width)
			else
				Result := formatted_integer (a_value, a_width)
			end
		end

	saved_format_integer: FORMAT_INTEGER is
		local
			l_arbitrary: INTEGER
		once
			l_arbitrary := 10
			create Result.make (l_arbitrary)
		end

feature -- double formatting

	formatted_double (a_value: DOUBLE; a_width, a_decimal_places: INTEGER): STRING is
		do
			if a_width /= saved_double_width or a_decimal_places /= saved_decimal_places then
				saved_format_double.set_width (a_width)
				saved_format_double.set_decimals (a_decimal_places)
				saved_double_width := a_width
				saved_decimal_places := a_decimal_places
			end

			if is_defined_double (a_value) then
				if a_value = 0 then
					Result := formatted_integer (0, a_width)
				else
					Result := saved_format_double.formatted (a_value)
				end
			else
				-- Result := blanks_n (a_width)
				Result := saved_format_double.formatted (a_value)
			end
		end
	saved_double_width: INTEGER
	saved_decimal_places: INTEGER

	formatted_double_no0 (a_value: DOUBLE; a_width, a_decimal_places: INTEGER): STRING is
		-- no0 means not allowed to print a 0.	
		do
			if a_value = 0 then
				Result := blanks_n (a_width)
			else
				Result := formatted_double (a_value, a_width, a_decimal_places)
			end
		end

	formatted_double_noNAN (a_value: DOUBLE; a_width, a_decimal_places: INTEGER): STRING is
		-- no0 means not allowed to print a QNAN.	
		do
			if not is_defined_double (a_value) then
				Result := blanks_n (a_width)
			else
				Result := formatted_double (a_value, a_width, a_decimal_places)
			end
		end

	formatted_double_no0_noNAN (a_value: DOUBLE; a_width, a_decimal_places: INTEGER): STRING is
		-- no0 means not allowed to print a 0 nor a QNAN	
		do
			if not is_defined_double (a_value) or else a_value = 0 then
				Result := blanks_n (a_width)
			else
				Result := formatted_double (a_value, a_width, a_decimal_places)
			end
		end

	saved_format_double: FORMAT_DOUBLE is
			-- formatting 
		local
			l_arbitrary: INTEGER
		once
			l_arbitrary := 10
			create Result.make (l_arbitrary, l_arbitrary)
		end

end












