note

	description:
	"[
		Class that provides access to a particular entity data environment
						  All categories and attribute types are accessible
	]"
	author: "Mark Howard"
	last_editor: "$Author: it $"
	date: "$Date: 2014/12/02 $"
	revision: "$Revision: #188 $"
	path: "$File: //rose/source/kernel/entities/supplier_context.e $"

deferred class
	SUPPLIER_CONTEXT

inherit
	ENTITY_TEMPLATES
	ENTITY_ACCESSORS
	ROSE_ATTRIBUTE_TYPE_ID_USE
	CONTEXT_MANAGER
		export
			{NONE} all
		end
	DEBUG_OUTPUT
		export
			{NONE} all
		end
	UID_AWARE
	PES_AWARE_CONTEXT
		rename
			pes_context_id as context_index
		end
	PM_METRIC_SOURCE
		redefine
			metrics
		end
	PM_AWARE

feature {AWARE} -- Initialization

	make (a_server: like server; a_name: STRING) is
			-- Create the context
		require
			name_not_Void: a_name /= Void
		local
			l_descriptor: CONTEXT_DESCRIPTOR
		do
			name := a_name.twin
			name.to_upper
			if name.is_equal ("STANDARD_AND_POORS") then
				name:="STANDARD AND POORS"
			end
			-- pulled out of contexts.add_context to aVoid class invariant assertion
			--
			l_descriptor := contexts.descriptors_by_name @ name
			if l_descriptor = Void then
				create l_descriptor.make (Void, Void, 0)
			end
			set_descriptor (l_descriptor)
			map_attributes	-- Because of a class invariant, this must be called before
			-- any other feature of this class.
			if a_server /= Void then
				server := a_server
				set_server_implicit_transaction_status
				setup_attribute_types
				create relationship_handlers.make_empty
				create identifiers.make (attribute_type_ids.count)
				create local_timeout_period.make (0,0,0,24,0,0)
				create next_refresh_time.make_now
				next_refresh_time.add (timeout_period)
				contexts.add_context (Current)
			end
		end

	make_simple (a_server: like server; a_name: STRING) is
			-- Create the context in isolation of the context server (no descriptor creation)
		require
			name_not_void: a_name /= Void
		do
			name := a_name.twin
			name.to_upper
			if a_server /= Void then
				server := a_server
				set_server_implicit_transaction_status
				contexts.add_context (Current)
				create local_timeout_period.make (0,0,0,24,0,0)
				create next_refresh_time.make_now
				next_refresh_time.add (timeout_period)
			end
		end
feature -- Access

	rose : SUPPLIER_CONTEXT is
			--source of attribute_ids, relationship_ids, category_ids etc.
		deferred
		end

	id_handler(a_category_id : INTEGER) : ID_HANDLER
			--  Uid relationship handler for this supplier_id.  No uids if supplier_id = 0
		do
			Result := id_handler_from_id(supplier_id,a_category_id)
		ensure
			non_void : Result /= Void
		end

	name: STRING
			-- Name of the context

	supplier_id: INTEGER is
			-- Supplier id
		do
			if descriptor /= Void then
				Result := descriptor.supplier_id
			end
		end

	context_index: INTEGER is
			-- Context index
		do
			if descriptor /= Void then
				Result := descriptor.index
			end
		end

	debug_output: STRING is
			--use the name for debug_output
		do
			Result := name
		end

	server: SQL_DATA_SERVER
			-- The database containing the entity data

feature -- Status report

	identifiers_available: BOOLEAN is
			-- Are `identifiers' non-Void?
		do
			Result := identifiers /= Void
		ensure
			definition: Result = (identifiers /= Void)
		end

feature -- Basic Operations

	refresh_pes_server_cache_fast is
			-- load attributes, entities and relationships as fast as possible
		local
			l_gc: BOOLEAN
		do
			l_gc := memory.collecting
			refresh

			load_special_caches

			-- load everything
			memory.collection_off

			load_all_attributes_fast

			memory.collection_on
			memory.full_collect
			memory.full_coalesce
			memory.collection_off

			load_entities_fast

			memory.collection_on
			memory.full_collect
			memory.full_coalesce
			memory.collection_off

			load_relationships_fast

			memory.collection_on
			memory.full_collect
			memory.full_coalesce

			load_secondary_attribute_cache

			memory.full_collect

			if not l_gc then
				memory.collection_off
			end
		rescue
			if l_gc then
				memory.collection_on
			else
				memory.collection_off
			end
		end

	load_entities_fast is
			-- load entities_id_base records from the DB and store them into the category_handler cache
			-- before we call this, we should load all attributes (some entities needs attributes for creation)
		local
			i, e, l_category_id: INTEGER
			l_entity_identifiers, l_one_category: ROSE_LINEAR_ARRAY[ENTITY_ID_BASE]
			l_category_handler: CATEGORY_HANDLER[SUPPLIER_CONTEXT, ENTITY]
			l_category_counts_map: HASH_TABLE[INTEGER, INTEGER]
			l_gc: BOOLEAN
		do
			l_gc := memory.collecting
			output(Global.interpreted("($now) start loading entities%N"), Void)
			-- get entity from DB
			output(Global.interpreted("($now) start loading entities from DB%N"), Void)
			memory.collection_off
			all_entity_by_category_predicate.retrieve_atomic (Void, Void)
			if l_gc then
				memory.collection_on
			end
			l_entity_identifiers := all_entity_by_category_predicate.array
			output(Global.interpreted("($now) end loading entities from DB%N"),Void)

			l_category_counts_map := category_counts_from_id

			-- go through categories
			from
				i := l_entity_identifiers.lower
			until
				i > l_entity_identifiers.upper
			loop
				l_category_id := l_entity_identifiers.item (i).category_id
				create l_one_category.make (1, l_category_counts_map.item (l_category_id))
				from
					e := 1
				until
					i > l_entity_identifiers.upper or else l_entity_identifiers.item (i).category_id /= l_category_id
				loop
					l_one_category.put (l_entity_identifiers.item (i), e)
					e := e + 1
					i := i + 1
				end
				check
					all_entities: (e - 1) = l_category_counts_map.item (l_category_id)
				end

				-- cache entities in corresponding category_handler
				if category_handlers.valid_index (l_category_id) then
					l_category_handler := handler (l_category_id)
					check
						valid_category_handler: l_category_handler /= Void
					end
					output(Global.interpreted("($now) Load all entities of category: " + l_category_handler.name + "%N"),Void)
					l_category_handler.load_entities_fast (l_one_category)
				else
					output(Global.interpreted("($now) Invalid category: : " + l_category_id.out + "%N"),Void)
				end
			end
			output(Global.interpreted("($now) end loading entities%N"), Void)
		rescue
			if l_gc then
				memory.collection_on
			end
		end

	load_relationships_fast is
			-- Load all relationships from DB and cache them in the corresponding relationship_handlers.
			-- Before we call this, we should load all entities (because they are part of the relationships).
		require
			relationship_handlers_not_void:  relationship_handlers /= Void
		local
			i, e, l_relationship_id: INTEGER
			l_relationships, l_one_relationship: ROSE_LINEAR_ARRAY[ENTITY_RELATIONSHIP_RECORD]
			l_relationship_counts_table: HASH_TABLE[INTEGER, INTEGER]
			l_rel_handler: RELATIONSHIP_HANDLER[ENTITY, ENTITY]
			l_gc: BOOLEAN
		do
			l_gc := memory.collecting
			output(Global.interpreted("($now) start loading relationships%N"), Void)
			-- load all relationships from DB
			memory.collection_off
			all_relationship_reader2.retrieve_atomic (Void, Void)
			l_relationships := all_relationship_reader2.array

			-- Free some memory
			all_relationship_reader_saved2 := Void

			if l_gc then
				memory.collection_on
			end

			l_relationship_counts_table := relationship_counts_from_id

			-- go trough all relationships
			from
				i := l_relationships.lower
			until
				i > l_relationships.upper
			loop
				l_relationship_id := l_relationships.item (i).relationship_type_id
				create l_one_relationship.make (1, l_relationship_counts_table.item (l_relationship_id))
				from
					e := 1
				until
					i > l_relationships.upper or else l_relationships.item (i).relationship_type_id /= l_relationship_id
				loop
					l_one_relationship.put (l_relationships.item (i), e)
					i := i + 1
					e := e + 1
				end

				if relationship_handlers.valid_index (l_relationship_id) then
					l_rel_handler := relationship (l_relationship_id)
					check
						valid_relationship: l_rel_handler /= Void
					end
					output(Global.interpreted("($now) Load all relationships of category: " + l_rel_handler.name + "%N"),Void)
					l_rel_handler.load_relationships_fast (l_one_relationship)
				else
					output(Global.interpreted("($now) Invalid relationship: : " + l_relationship_id.out + "%N"),Void)
				end
			end

			output(Global.interpreted("($now) start sorting relationships%N"), Void)

			-- sort relationships according to entity2

			l_relationships.sort_by (agent compare_rel_records_for_entity_2)
			output(Global.interpreted("($now) end sorting relationships%N"), Void)

			-- load corelationships
			from
				i := l_relationships.lower
			until
				i > l_relationships.upper
			loop
				l_relationship_id := l_relationships.item (i).relationship_type_id
				create l_one_relationship.make (1, l_relationship_counts_table.item (l_relationship_id))
				from
					e := 1
				until
					i > l_relationships.upper or else l_relationships.item (i).relationship_type_id /= l_relationship_id
				loop
					l_one_relationship.put (l_relationships.item (i), e)
					i := i + 1
					e := e + 1
				end

				if relationship_handlers.valid_index (l_relationship_id) then
					l_rel_handler := relationship (l_relationship_id)
					check
						valid_relationship: l_rel_handler /= Void
					end
					output(Global.interpreted("($now) Load all corelationships of category: " + l_rel_handler.name + "%N"),Void)
					l_rel_handler.load_corelationships_fast (l_one_relationship)
				else
					output(Global.interpreted("($now) Invalid corelationship: : " + l_relationship_id.out + "%N"),Void)
				end
			end
			output(Global.interpreted("($now) end loading corelationships%N"), Void)
		rescue
			if l_gc then
				memory.collection_on
			end
		end

	clear_cache is
			-- reset all caches
		do
			relationship_records_table_saved := Void

			if  relationship_handlers /= Void then
				relationship_handlers.clear_all
			end
			if  relationships_saved /= Void then
				relationships_saved.clear_all
			end
			if  categories_saved /= Void then
				categories_saved.clear_all
			end
			if  category_handlers_saved /= Void then
				category_handlers_saved.clear_all
			end
			if  attribute_id_types /= Void then
				attribute_id_types.clear_all
			end
			if  attribute_type_ids /= Void then
				attribute_type_ids.clear_all
			end
			if  identifiers /= Void then
				identifiers.clear_all
			end
			if  attribute_is_uniq_by_id_cache /= Void then
				attribute_is_uniq_by_id_cache.clear_all
			end
			if  rose_attribute_from_local /= Void then
				rose_attribute_from_local.clear_all
			end
			if  rose_attribute_to_local /= Void then
				rose_attribute_to_local.clear_all
			end


			mapping_reader_saved := Void
			mapping_override_reader_saved := Void
			all_mapping_reader_saved := Void
			all_mapping_override_reader_saved := Void

			refresh
		--	reset_mappings
			reset_predicates
		end

	load_all_attributes_fast is
			-- Load all attribute records for this context into the corresponding caches.
		require
			identifiers_not_void: identifiers_available
		local
			l_result_array: ROSE_LINEAR_ARRAY[ENTITY_ATTRIBUTE_RECORD]
			l_i: INTEGER
			l_item: ENTITY_ATTRIBUTE_RECORD
			l_entity_id, l_old_entity_id: INTEGER
			l_category_id, l_old_category_id: INTEGER
			l_attribute_type_id, l_old_attribute_type_id: INTEGER
			-- The next 3 hold the first index of the current entity_attribute, category or attribute_type
			l_entity_attribute_start_index: INTEGER
			l_category_start_index: INTEGER
			l_attribute_type_start_index: INTEGER
			l_identifier: ENTITY_IDENTIFIER
			l_attribute_counts: ARRAY[TUPLE[attribute_type_id: INTEGER; attribute_count: INTEGER]]
			l_table_size_predicate: INTEGER_PREDICATE
			l_gc: BOOLEAN
			l_identifier_list: ARRAYED_LIST[ENTITY_IDENTIFIER]
			l_category_loaded_array: ROSE_LINEAR_ARRAY[BOOLEAN]
		do
			output(Global.interpreted("($now) start loading all attributes%N"), Void)
			l_gc := memory.collecting
			-- Get table size
			create l_table_size_predicate.make_strict (context.server, "SELECT COUNT(*) AS count FROM entity_attribute")
			l_table_size_predicate.retrieve_atomic (Void, Void)
			if l_table_size_predicate.success and then l_table_size_predicate.array.count > 0 then
				create l_result_array.make (1, l_table_size_predicate.array.item (l_table_size_predicate.array.lower))
			else
				create l_result_array.make_empty
			end
			all_attributes_predicate.set_container (l_result_array)
			memory.collection_off
			all_attributes_predicate.retrieve_atomic (Void, Void)
			if l_gc then
				memory.collection_on
			end
			if all_attributes_predicate.success then
				-- Free memory
				all_attributes_predicate_saved := Void

				if l_result_array.count > 1 then
					-- Set hash map sizes
					create l_attribute_counts.make_filled (Void, 1, 1)
					count_attributes_predicate.set_container (l_attribute_counts)
					count_attributes_predicate.retrieve_atomic (Void, Void)
					if count_attributes_predicate.success then
						from
							l_i := l_attribute_counts.lower
						until
							l_i > l_attribute_counts.upper
						loop
							l_identifier := identifier (l_attribute_counts.item (l_i).attribute_type_id)
							l_identifier.dated_keys_by_id_cache.accommodate (l_attribute_counts.item (l_i).attribute_count + 100)
							l_i := l_i + 1
						end
					end
					from
						l_i := l_result_array.lower + 1
						l_item := l_result_array.item (l_result_array.lower)
						l_old_entity_id := l_item.entity_id
						l_old_category_id := l_item.category_id
						l_old_attribute_type_id := l_item.attribute_type_id
						l_identifier := identifier (l_item.attribute_type_id)
						output (Global.interpreted("($now) Now loading attribute: $1 %N"), <<l_identifier.name>>)
						l_entity_attribute_start_index := l_result_array.lower
						l_category_start_index := l_result_array.lower
						l_attribute_type_start_index := l_result_array.lower
					until
						l_i > l_result_array.upper
					loop
						-- Fill in attribute value to record hash map
						l_identifier.add_to_entities_table (l_item)

						l_item := l_result_array.item (l_i)
						l_entity_id := l_item.entity_id
						l_category_id := l_item.category_id
						l_attribute_type_id := l_item.attribute_type_id

						-- Fill the main attribute record cache
						if l_entity_id /= l_old_entity_id or else l_category_id /= l_old_category_id or else l_attribute_type_id /= l_old_attribute_type_id then
							l_identifier.add_to_preloaded_cache (l_result_array.subarray (l_entity_attribute_start_index, l_i - 1), l_old_entity_id)
							l_entity_attribute_start_index := l_i
						end

						-- Load `unique_key_records'
						if l_old_category_id /= l_category_id or else l_old_attribute_type_id /= l_attribute_type_id then
							l_identifier.set_unique_key_records_by_category (l_result_array.subarray (l_category_start_index, l_i - 1), l_old_category_id)
							l_category_start_index := l_i
						end

						-- Now update `l_identifier' and the start_indexes if necessary
						if l_old_attribute_type_id /= l_attribute_type_id then
							-- Also fill primary cache
						--	l_identifier.populate_primary_caches_from_secondary

							l_identifier := identifier (l_attribute_type_id)
							output (Global.interpreted("($now) Now loading attribute: $1 %N"), <<l_identifier.name >>)
							l_attribute_type_start_index := l_i
						end

						l_old_attribute_type_id := l_attribute_type_id
						l_old_category_id := l_category_id
						l_old_entity_id := l_entity_id
						l_i := l_i + 1
					end

					-- Add the last item
					-- Fill in attribute value to record hash map
					l_identifier.add_to_entities_table (l_item)

					-- Also fill primary cache
					--l_identifier.populate_primary_caches_from_secondary

					if l_entity_attribute_start_index <= l_result_array.upper then
						l_identifier.add_to_preloaded_cache (l_result_array.subarray (l_entity_attribute_start_index, l_result_array.upper), l_old_entity_id)
						l_identifier.category_loaded.force (True, l_old_category_id)
						l_identifier.set_unique_key_records_by_category (l_result_array.subarray (l_attribute_type_start_index, l_i - 1), l_old_category_id)
					end

					-- Set all categories to loaded
					from
						l_identifier_list := identifiers.linear_representation
						l_identifier_list.start
					until
						l_identifier_list.after
					loop
						from
							l_category_loaded_array := l_identifier_list.item.category_loaded
							l_i := l_category_loaded_array.lower
						until
							l_i > l_category_loaded_array.upper
						loop
							l_category_loaded_array.item (l_i) := True

							l_i := l_i + 1
						end

						l_identifier_list.forth
					end
					output (Global.interpreted("($now) end loading all attributes%N"), Void)
				else
					output (Global.interpreted("($now) Could not load attributes from DB%N"), Void)
				end
			end
		ensure
			all_identifiers_set_to_loaded: identifiers.linear_representation.for_all (agent (a_identifier: like entity_identifier_template): BOOLEAN do Result := a_identifier /= Void implies a_identifier.category_loaded.for_all (agent (a_value: BOOLEAN): BOOLEAN do Result := a_value end) end)
		rescue
			if l_gc then
				memory.collection_on
			end
		end

	load_secondary_attribute_cache is
			-- Take attribute history from dated_key_by_id_cache and fill dated_keys and keys.
			-- We need to call this after all entities are loaded because keys and dated_keys use entities as key.
		require
			identifiers_not_void: identifiers_available
		local
			l_identifiers: ARRAYED_LIST[ENTITY_IDENTIFIER]
		do

			output(Global.interpreted("($now) start loading secondary caches for attributes %N"), Void)
			from
				l_identifiers := identifiers.linear_representation
				l_identifiers.start
			until
				l_identifiers.after
			loop
				l_identifiers.item.populate_primary_caches_from_secondary
				l_identifiers.forth
			end
			output(Global.interpreted("($now) end loading secondary caches for attributes %N"), Void)
		end

	load_special_caches is
			-- Load special, context specific caches
			-- Redefine this in descendants
		do
		end

	update_cache (a_from, a_to: ROSE_DATE) is
			-- update all caches
		do
	--		output (Global.interpreted("($now) Updating " + name +  " from" + a_from.out + " to " + a_to.out + "%N"), Void)
			update_attributes (a_from, a_to)
			update_entities (a_from, a_to)
			update_relationships (a_from, a_to)
		end

	update_attributes (a_from, a_to: ROSE_DATE) is
			-- Update all attributes where update_date is later than `a_from' and not earlier than `a_to'.
		require
			identifiers_not_void: identifiers_available
		local
			l_array:ROSE_LINEAR_ARRAY[ENTITY_ATTRIBUTE_RECORD]
			l_i: INTEGER
			l_identifier: ENTITY_IDENTIFIER
			l_record: ENTITY_ATTRIBUTE_RECORD
		do
			attributes_update_predicate.retrieve_atomic (Void, <<a_from, a_to>>)
			l_array := attributes_update_predicate.array

			if l_array.count > 0 then
				output (Global.interpreted("($now) Updating " + l_array.count.out +  " attributes %N"), Void)
			end

			from
				l_i := l_array.lower
			until
				l_i > l_array.upper
			loop
				l_record := l_array.item (l_i)
				l_identifier := identifier (l_record.attribute_type_id)
				l_identifier.update_attribute_record (l_record)

				l_i := l_i + 1
			end
		end

	update_entities (a_from, a_to: ROSE_DATE) is
			-- update all entities
		local
			l_record: ENTITY_ID_BASE
			l_category_handler: CATEGORY_HANDLER[SUPPLIER_CONTEXT, ENTITY]
		do
			entity_update_predicate.retrieve_atomic (Void, <<a_from, a_to>>)
			from
				entity_update_predicate.array.start
			until
				entity_update_predicate.after
			loop
				l_record := entity_update_predicate.item
				output (Global.interpreted("($now) entity update on entity: " + l_record.entity_id.out +  " %N"),Void)

				if category_handlers.valid_index (l_record.category_id) then
					l_category_handler := handler (l_record.category_id)
					check
						valid_category_handler: l_category_handler /= Void
					end
					l_category_handler.update_entity (l_record)
				else
					output (global.interpreted ("($now) Invalid category: : " + l_record.category_id.out + "%N"), Void)
				end
				entity_update_predicate.array.forth
			end
		end

	update_relationships (a_from, a_to: ROSE_DATE) is
			-- update all relationships
		require
		 relationship_handlers_not_void:  relationship_handlers /= Void
		local
			l_record: ENTITY_RELATIONSHIP_RECORD
			l_relationship_handler: RELATIONSHIP_HANDLER[ENTITY, ENTITY]
		do
			relationship_update_predicate.retrieve_atomic (Void, <<a_from, a_to>>)
			from
				relationship_update_predicate.array.start
			until
				relationship_update_predicate.after
			loop
				l_record := relationship_update_predicate.item
				output (Global.interpreted("($now) relationship update on entity: " + l_record.relationship_type_id.out +  " %N"),Void)

				if relationship_handlers.valid_index (l_record.relationship_type_id) then
					l_relationship_handler := relationship (l_record.relationship_type_id)
					check
						valid_relationship_handler: l_relationship_handler /= Void
					end
					l_relationship_handler.update_relationship (l_record)
					l_relationship_handler.update_corelationship (l_record)
				else
					output (global.interpreted ("($now) Invalid relationship: : " + l_record.relationship_type_id.out + "%N"), Void)
				end
				relationship_update_predicate.array.forth
			end
		end

	load_from_context (a_context: like Current) is
			-- load his context from another context object
		require
			not_same: a_context /= Current
			not_void: a_context /= Void
		do
			category_handlers_saved := a_context.get_category_handlers_saved
			categories_saved := a_context.get_categories_saved
			identifiers := a_context.get_identifiers
			relationships_saved := a_context.relationships_saved
		end

	refresh is
			-- Refresh caches
		do
			if not server.connected_by_testing then
				server.reconnect
				reset_predicates
			end
			categories_saved := Void
			category_handlers_saved := Void
			relationships_saved := Void
			relationship_handlers := Void
			make (server, name)
			categories.do_nothing
			relationships.do_nothing
		end

feature {SUPPLIER_CONTEXT} -- PES Getters

	get_category_handlers_saved: like category_handlers_saved is
			-- get category_handlers_saved
		do
			Result := category_handlers_saved
		end

	get_categories_saved: like categories_saved is
			-- get categories_saved
		do
			Result := categories_saved
		end

	get_identifiers: like identifiers is
			-- get identifiers;
			-- WHY ????
		do
			Result := identifiers
		ensure
			definition: Result = identifiers
		end

feature -- Status Queries

	is_valid: BOOLEAN is
			-- Is the underlying data source available?
		do
			Result := server /= Void
		end

	has_entities: BOOLEAN is
			-- Does this context have available entity data
		require
			valid: is_valid
		do
			if is_valid and has_entities_saved = Void then
				create has_entities_saved
				has_entities_saved.set_item (server.has_table ("entity_attribute"))
			end
			Result := has_entities_saved.item
		end

	has_rose_handles: BOOLEAN is
			-- Does this context have rose handlers?
		do
			Result := has_rose_fg_handles or has_rose_bg_handles
		end

	has_rose_fg_handles: BOOLEAN
	set_has_rose_fg_handles (a_value: like has_rose_fg_handles) is
		-- set has_rose_fg_handler' to `a_value'
		do
			has_rose_fg_handles := a_value
		end

	has_rose_bg_handles: BOOLEAN
		-- does this context have rose bg handles?

	set_has_rose_bg_handles (a_value: like has_rose_bg_handles) is
		-- set has_rose_bg_handler' to `a_value'
		do
			has_rose_bg_handles := a_value
		end

	implicit_transaction_change_disabled: BOOLEAN
			-- Is implicit transaction change disabled?

	common_attributes(a_entity : ENTITY) : ROSE_LINEAR_ARRAY[ENTITY_IDENTIFIER] is
			--used by persistent entity server to cache those attributes which are needed to create an entity
		do
		end
	load_all_entities is
			--used by persistent entity server to cache entites
		do
		end
feature -- SQL Transactions Settings

	set_server_implicit_transaction_status is
			-- Set server implicit transaction 'off' by default for backward compatibility / consistency
			-- Users can redefine the implicit transaction status in descendant implementation
		require
			valid: is_valid
		do
			if not implicit_transaction_change_disabled then
				server.set_implicit_transaction_off
			end
		end

	enable_implicit_transaction_change is
			-- Enable implicit transaction change
		do
			implicit_transaction_change_disabled := False
		end

	disable_implicit_transaction_change is
			-- Disable implicit transaction change
		do
			implicit_transaction_change_disabled := True
		end

feature -- Variables Support

	variable_manager: VARIABLE_MANAGER is
			-- Variable manager
		do
			if (variable_manager_holder = Void) then
				variable_manager_holder := create_variable_manager
			end
			Result := variable_manager_holder
		end

	variable_manager_holder: VARIABLE_MANAGER
			-- Holds the variable manager

	create_variable_manager: VARIABLE_MANAGER is
			-- Create and return the variable manager
		do

		end

	variable_description_records : ROSE_LINEAR_ARRAY [VARIABLE_DESCRIPTION_RECORD] is
			-- Retrieves variable description records from the DB
		local
			l_retriever : COLLECTING_PREDICATE[VARIABLE_DESCRIPTION_RECORD]
		do
			create l_retriever.make(server,"select * from variable_description")
			l_retriever.retrieve_atomic(Void,Void)
			Result := l_retriever.array
		end

	variable_descriptions (a_target: STRING): ROSE_LINEAR_ARRAY [VARIABLE_DESCRIPTION_RECORD] is
			-- Create an array of variable description records from `a_target'
		do
			if context_variables = Void then
				create context_variables.make (Current)
			end
			Result := context_variables.variable_descriptions (a_target)
		end

	variable_description_by_mnemonic (a_mnemonic: STRING): VARIABLE_DESCRIPTION_RECORD is
			-- Variable description records for `a_mnemonic'
		do
			if context_variables = Void then
				create context_variables.make (Current)
			end
			Result := context_variables.variable_description_by_mnemonic (a_mnemonic)
		end

	variable_description_by_name (a_name: STRING): VARIABLE_DESCRIPTION_RECORD is
			-- Variable description records for `a_name'
		do
			if context_variables = Void then
				create context_variables.make (Current)
			end
			Result := context_variables.variable_description_by_name (a_name)
		end

	variable_descriptions_by_source (a_source: STRING): ROSE_LINEAR_ARRAY [VARIABLE_DESCRIPTION_RECORD] is
			-- Variable description records for `a_source'
		do
			if context_variables = Void then
				create context_variables.make (Current)
			end
			Result := context_variables.variable_descriptions_by_source (a_source)
		end

	all_variable_descriptions: ROSE_LINEAR_ARRAY [VARIABLE_DESCRIPTION_RECORD] is
		-- returns an array containing all variable description records
		do
			if context_variables = Void then
				create context_variables.make (Current)
			end
			Result := context_variables.all_variable_descriptions
		end

feature {CONTEXT_VARIABLES} -- Variables Support Implementation

	variable_security: INTEGER
			--Security level
		local
			l_pred: INTEGER_PREDICATE
			l_oledb_server: SQLOLEDB_DATA_SERVER
		do
			if variable_security_saved=0
			 then
				l_oledb_server ?= server
				if l_oledb_server=Void
				 then
					variable_security_saved := 1
				else
					create l_pred.make_strict (l_oledb_server, "select count (*) from sysobjects where name = 'variable_security_view'")
					l_pred.retrieve_atomic (Void, Void)
					if l_pred.array.count>0 and then l_pred.array.item (1)=1
					 then
						create l_pred.make_strict (l_oledb_server, "select security from variable_security_view")
						l_pred.retrieve_atomic (Void, Void)
						if l_pred.array.count/=1
						 then
							error.raise ("variable_security_view did not return 1 row.")
						end

						variable_security_saved := l_pred.array.item (1).item
					else
						variable_security_saved := 1
					end
				end
			end
			Result := variable_security_saved
		end

feature -- Mappings

	mappings_from_descriptions (a_mnemonics: ROSE_LINEAR_ARRAY [VARIABLE_DESCRIPTION_RECORD]): ROSE_LINEAR_ARRAY [MAPPINGS_RECORD] is
			-- Mappings from variable description records
		local
			l_found: BOOLEAN
		do
			create Result.make_empty
			from a_mnemonics.start until a_mnemonics.after loop
				l_found := False
				from mappings.start until mappings.after or l_found loop
					mappings.current_item.mnemonic.to_lower
					a_mnemonics.current_item.mnemonic.to_lower
					if mappings.current_item.mnemonic.has_substring (a_mnemonics.current_item.mnemonic) then
						Result.extend (mappings.current_item)
						l_found := True
					end
					mappings.forth
				end
				a_mnemonics.forth
			end
		end

	mappings: ROSE_LINEAR_ARRAY [MAPPINGS_RECORD] is
		-- Mappings
		do
			refresh_final_mappings
			Result := final_mappings_saved
		end

	set_mappings_override_by_label (a_value: like mappings_override_label) is
			-- Specify the set of mappings override to use; Void for original production mappings
		do
			mappings_override_label := a_value
			mappings_override_changed := True
			refresh_final_mappings
			mappings_revision_number := mappings_revision_number + 1
		end

	reset_mappings is
			-- Set `mapings_saved' to void
		do
			mappings_saved := Void
			set_mappings_override_by_label (Void)
		end

	refresh_mappings is
			-- Refresh predicates and hash tables for mappings
		do
			if mappings_saved = Void then
				all_mapping_reader.retrieve_atomic (Void,Void)
				all_mapping_reader.array.apply (agent adjust_mappings_record)
				mappings_saved := all_mapping_reader.array.twin
				final_mappings_saved := mappings_saved.twin
			end
		end

	refresh_mappings_override is
			-- Refresh mappings override
		do
			all_mapping_override_reader.retrieve_atomic (Void,<<name>>)
			all_mapping_override_reader.array.apply (agent adjust_mappings_record)
			mappings_override_saved := all_mapping_override_reader.array.twin
		end

	refresh_final_mappings is
			-- Refresh predicates and hash tables for final mappings
		local
			l_mappings_override: ROSE_LINEAR_ARRAY [MAPPINGS_RECORD]
		do
			refresh_mappings
			if mappings_override_changed then
				if mappings_override_label = Void then
					final_mappings_saved := mappings_saved.twin
				else
					refresh_mappings_override
					create l_mappings_override.make_empty
					from mappings_override_saved.start
					until mappings_override_saved.after
					loop
						if mappings_override_saved.current_item.label.is_equal (mappings_override_label) then
							l_mappings_override.extend (mappings_override_saved.current_item)
						end
						mappings_override_saved.forth
					end
					-- Assuming mappings_saved is sorted by type and mnemonic, perform a merge join
					create final_mappings_saved.make_empty
					from
						l_mappings_override.start; mappings_saved.start
					until
						l_mappings_override.after and mappings_saved.after
					loop
						if not l_mappings_override.after and then not mappings_saved.after then
							if l_mappings_override.current_item.type < mappings_saved.current_item.type then
								final_mappings_saved.extend (l_mappings_override.current_item)
								l_mappings_override.forth
							elseif l_mappings_override.current_item.type > mappings_saved.current_item.type then
								final_mappings_saved.extend (mappings_saved.current_item)
								mappings_saved.forth
							elseif l_mappings_override.current_item.mnemonic < mappings_saved.current_item.mnemonic then
								final_mappings_saved.extend (l_mappings_override.current_item)
								l_mappings_override.forth
							elseif l_mappings_override.current_item.mnemonic > mappings_saved.current_item.mnemonic then
								final_mappings_saved.extend (mappings_saved.current_item)
								mappings_saved.forth
							else -- Use override
								final_mappings_saved.extend (l_mappings_override.current_item)
								l_mappings_override.forth
								mappings_saved.forth
							end
						elseif not l_mappings_override.after then
							final_mappings_saved.extend (l_mappings_override.current_item)
							l_mappings_override.forth
						else
							final_mappings_saved.extend (mappings_saved.current_item)
							mappings_saved.forth
						end
					end
				end
			end
			if mappings_table_saved = Void or mappings_override_changed then
				create mappings_table_saved.make (10)
				final_mappings_saved.apply (agent add_to_mappings_table)
				mappings_override_changed := False
			end
		end

	mappings_table: ROSE_MULTI_KEYED_TABLE [MAPPINGS_RECORD] is
			-- Mappings table
		do
			refresh_final_mappings
			Result := mappings_table_saved
		end

	adjust_mappings_record (a_record: MAPPINGS_RECORD) is
			-- Standardize 'a_record'
		do
			a_record.type.to_lower
			a_record.type.right_adjust
			a_record.mnemonic.to_lower
			a_record.mnemonic.right_adjust
			a_record.expression.to_lower
			a_record.expression.right_adjust
		end

	add_to_mappings_table (a_record: MAPPINGS_RECORD) is
			-- Add 'a_record' to 'mappings_table'.
		do
			mappings_table_saved.force (a_record, [a_record.type, a_record.mnemonic])
		end

	mappings_table_saved: like mappings_table
	mappings_override_changed: BOOLEAN
	mappings_override_label: STRING
	mappings_revision_number: INTEGER
	mappings_saved, mappings_override_saved, final_mappings_saved: like mappings

	store_mapping (a_mapping: MAPPINGS_RECORD) is
		do
			a_mapping.expression.prune_all (' ')
			a_mapping.expression.to_lower
			a_mapping.mnemonic.prune_all (' ')
			a_mapping.mnemonic.to_lower
		end

feature -- Basic operations

	reconnect is
		do
			if server /= Void then
				server.reconnect
			end
			reset_predicates
		end

	identifier (a_handle: INTEGER): like entity_identifier_template is
			-- Get the ENTITY_IDENTIFIER object for attribute type id `a_handle'
		require
			identifiers_not_void: identifiers_available
			-- This is a valid precondition, but it gets violated by some existing code
			--meaningful_id: a_handle > 0
		do
			refresh_if_necessary
			Result := identifiers.item (a_handle)
			if Result = Void then
				create Result.make (Current, a_handle)
				identifiers.force (Result, a_handle)
			end
		ensure
			result_not_void: Result /= Void
		end

	identifier_from_rose (a_rose_handle: INTEGER): like entity_identifier_template is
			-- Identifier handler from `a_rose_handle' for that attribute
		local
			l_handle: INTEGER
		do
			if a_rose_handle > 0 then
				if has_rose_handles then
					l_handle := a_rose_handle
				else
					l_handle := rose_attribute_to_local.item (a_rose_handle)
				end
			end
			if identifiers /= Void then
				Result := identifier (l_handle)
			end
		end

	identifier_from_name (a_name: STRING): like entity_identifier_template is
			-- Entity identifier from `a_name'
		require
			identifiers_not_void: identifiers_available
		local
			l_handle, l_rose_handle: INTEGER
		do
			l_rose_handle := rose.attribute_id_by_name (a_name)
			if l_rose_handle > 0 then
				if has_rose_handles then
					l_handle := l_rose_handle
				else
					l_handle := rose_attribute_to_local.item (l_rose_handle)
				end
			else
				l_handle := attribute_id_by_name (a_name)
			end
			Result := identifier (l_handle)
		end

	entity_at (a_category_id: INTEGER; a_id: INTEGER): ENTITY is
			-- returns entity corresponding to a_id in category a_category_id
		do
			if handler (a_category_id) /= Void then
				Result := handler (a_category_id).entity (a_id)
			end
		end

	security_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like security_template is
			-- Returns a security for a given attribute at a given date
		do
			Result := security_accessor.any_entity_at (a_attribute, a_date)
		end

	bond_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like bond_template is
			-- Returns a bond for a given attribute at a given date
		do
			Result := bond_accessor.any_entity_at (a_attribute, a_date)
		end

	exchange_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like exchange_template is
			-- Returns an exchange for a given attribute at a given date
		do
			Result := exchange_accessor.any_entity_at (a_attribute, a_date)
		end

	company_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like company_template is
			-- Returns a company for a given attribute at a given date
		do
			Result := company_accessor.any_entity_at (a_attribute, a_date)
		end

	sub_industry_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like sub_industry_template is
			-- Returns a sub_industry for a given attribute at a given date
		do
			Result := sub_industry_accessor.any_entity_at (a_attribute, a_date)
		end

	condition_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like condition_template is
			-- Returns a condition for a given attribute at a given date
		do
			Result := condition_accessor.any_entity_at (a_attribute, a_date)
		end

	valuation_model_regime_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like valuation_model_regime_template is
			-- Returns a valuation model regime for a given attribute at a given date
		do
			Result := valuation_model_regime_accessor.any_entity_at (a_attribute, a_date)
		end

	target_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like target_template is
			-- Returns a target for a given attribute at a given date
		do
			Result := target_accessor.any_entity_at (a_attribute, a_date)
		end

	file_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like file_template is
			-- Returns a file for a given attribute at a given date
		do
			Result := file_accessor.any_entity_at (a_attribute, a_date)
		end

	input_type_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like input_type_template is
			-- Returns a input type for a given attribute at a given date
		do
			Result := input_type_accessor.any_entity_at (a_attribute, a_date)
		end

	estimation_population_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like estimation_population_template is
			-- Returns a estimation population for a given attribute at a given date
		do
			Result := estimation_population_accessor.any_entity_at (a_attribute, a_date)
		end

	valuation_model_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like valuation_model_template is
			-- Returns a company for a given attribute at a given date
		do
			Result := valuation_model_accessor.any_entity_at (a_attribute, a_date)
		end

	calendar_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like calendar_template is
			-- Returns a calendar for a given attribute at a given date
		do
			Result := calendar_accessor.any_entity_at (a_attribute, a_date)
		end

	country_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like country_template is
			-- Returns a country for a given attribute at a given date
		do
			Result := country_accessor.any_entity_at (a_attribute, a_date)
		end

	industry_group_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like industry_group_template is
			-- Returns a industry_group for a given attribute at a given date
		do
			Result := industry_group_accessor.any_entity_at (a_attribute, a_date)
		end

	economic_sector_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like economic_sector_template is
			-- Returns a economic_sectorfor a given attribute at a given date
		do
			Result := economic_sector_accessor.any_entity_at (a_attribute, a_date)
		end

	sector_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like sector_template is
			-- Returns a sector for a given attribute at a given date
		do
			Result := sector_accessor.any_entity_at (a_attribute, a_date)
		end

	instruction_argument_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like instruction_argument_template is
			-- Returns a instruction_argument for a given attribute at a given date
		do
			Result := instruction_argument_accessor.any_entity_at (a_attribute, a_date)
		end

	time_zone_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like time_zone_template is
			-- Returns a time_zone for a given attribute at a given date
		do
			Result := time_zone_accessor.any_entity_at (a_attribute, a_date)
		end

	variable_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like variable_template is
			-- Returns a variable for a given attribute at a given date
		do
			Result := variable_accessor.any_entity_at (a_attribute, a_date)
		end

	model_components_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like model_components_template is
			-- Returns a model components for a given attribute at a given date
		do
			Result := model_components_accessor.any_entity_at (a_attribute, a_date)
		end

	economic_sector_group_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like economic_sector_group_template is
			-- Returns an economic sector group for a given attribute at a given date
		do
			Result := economic_sector_group_accessor.any_entity_at (a_attribute, a_date)
		end

	risk_model_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like risk_model_template is
			-- Returns a risk_model for a given attribute at a given date
		do
			Result := risk_model_accessor.any_entity_at (a_attribute, a_date)
		end

	risk_universe_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like risk_universe_template is
			-- Returns a risk universe for a given attribute at a given date
		do
			Result := risk_universe_accessor.any_entity_at (a_attribute, a_date)
		end

	universe_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like universe_template is
			-- Returns a universe for a given attribute at a given date
		do
			Result := universe_accessor.any_entity_at (a_attribute, a_date)
		end

	portfolio_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like portfolio_template is
			-- Returns a portfolio for a given attribute at a given date
		do
			Result := portfolio_accessor.any_entity_at (a_attribute, a_date)
		end

	market_index_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like market_index_template is
			-- Returns a market indexfor a given attribute at a given date
		do
			Result := market_index_accessor.any_entity_at (a_attribute, a_date)
		end

	industry_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like industry_template is
			-- Returns a industry for a given attribute at a given date
		do
			Result := industry_accessor.any_entity_at (a_attribute, a_date)
		end

	listing_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like listing_template is
			-- Returns a country for a given attribute at a given date
		do
			Result := listing_accessor.any_entity_at (a_attribute, a_date)
		end

	future_listing_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like future_listing_template is
			-- Returns a future listing for a given attribute at a given date
		do
			Result := future_listing_accessor.any_entity_at (a_attribute, a_date)
		end

	currency_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like currency_template is
			-- Returns a currency for a given attribute at a given date
		do
			Result := currency_accessor.any_entity_at (a_attribute, a_date)
		end

	anomaly_profile_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like anomaly_profile_template is
			-- Returns an anomaly profile for a given attribute at a given date
		do
			Result := anomaly_profile_accessor.any_entity_at (a_attribute, a_date)
		end

	future_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like future_template is
			-- Returns an future for a given attribute at a given date
		do
			Result := future_accessor.any_entity_at (a_attribute, a_date)
		end

	government_issue_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like government_issue_template is
			-- Returns an government issue for a given attribute at a given date
		do
			Result := government_issue_accessor.any_entity_at (a_attribute, a_date)
		end

	fund_issue_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like fund_issue_template is
			-- Returns an fund issue for a given attribute at a given date
		do
			Result := fund_issue_accessor.any_entity_at (a_attribute, a_date)
		end

	investment_region_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like investment_region_template is
			-- Returns an investment region for a given attribute at a given date
		do
			Result := investment_region_accessor.any_entity_at (a_attribute, a_date)
		end

	similars_group_at (a_attribute: ATTRIBUTE_DEFINITION; a_date: ROSE_DATE): like similars_group_template is
			-- Returns an similars group for a given attribute at a given date
		do
			Result := similars_group_accessor.any_entity_at (a_attribute, a_date)
		end

feature -- Relationships

	relationship (a_id: INTEGER): RELATIONSHIP_HANDLER [ENTITY,ENTITY] is
			-- return relationship from id
		require
		 relationship_handlers_not_void:  relationship_handlers /= Void
		do
			if relationship_handlers.valid_index (a_id) then
				Result := relationship_handlers.item (a_id)
			end
		end

	relationship_handler_from_name (a_name: STRING): RELATIONSHIP_HANDLER [ENTITY,ENTITY] is
		-- return relatioship handler from name
		require
			valid_args: a_name /= Void
		local
			l_handle: INTEGER
		do
			a_name.to_upper
			l_handle := ids.relationship_id (a_name)
			if l_handle > 0 then
				Result := relationship (l_handle)
			end
		end

	relationships: ROSE_LINEAR_ARRAY [RELATIONSHIP_HANDLER [ENTITY,ENTITY]] is
			--return all relationship handlers of this context
		local
			l_relationships: ROSE_LINEAR_ARRAY [INTEGER]
		do
			refresh_if_necessary
			if relationship_handlers = Void then
				create relationship_handlers.make_empty
			end
			if relationship_handlers.is_empty then
				initialize_accessors
			end
			if relationships_saved = Void then
				distinct_relationship_predicate.retrieve_atomic (Void, Void)
				create relationships_saved.make_empty
				l_relationships := distinct_relationship_predicate.array
				from l_relationships.start until l_relationships.off
				loop
					relationships_saved.extend (relationship (l_relationships.current_item))
					l_relationships.forth
				end
			end
			Result := relationships_saved
		end

	relationship_record (a_id: INTEGER): RELATIONSHIP_TYPE_RECORD is
			-- return relationship record from id
		do
			Result := ids.relationship_by_id (a_id)
		end

	relationship_category_1_id (a_id: INTEGER): INTEGER is
			-- return category 1 id of the relationship with this id
		local
			l_record: RELATIONSHIP_TYPE_RECORD
		do
			l_record := relationship_record (a_id)
			if l_record /= Void then
				Result := l_record.category_1_id
			end
		end

	relationship_category_2_id (a_id: INTEGER): INTEGER is
			-- return category 2 id of the relationship with this id
		local
			l_record: RELATIONSHIP_TYPE_RECORD
		do
			l_record := relationship_record (a_id)
			if l_record /= Void then
				Result := l_record.category_2_id
			end
		end

	relationship_handlers: ROSE_LINEAR_ARRAY [RELATIONSHIP_HANDLER [ENTITY,ENTITY]]

	relationships_saved: ROSE_LINEAR_ARRAY [RELATIONSHIP_HANDLER [ENTITY,ENTITY]]

feature -- Category Handlers

	handler (a_id: INTEGER): CATEGORY_HANDLER [like Current,ENTITY] is
			-- get category handler by id
		do
			if category_handlers.valid_index (a_id) then
				Result := category_handlers.item (a_id)
			end
		end

	category_handler_from_name (a_name: STRING): CATEGORY_HANDLER [like Current,ENTITY] is
			-- Handler for category `a_name'
		require
			valid_args: a_name /= Void
		local
			l_handle: INTEGER
		do
			a_name.to_upper
			l_handle := ids.category_id (a_name)
			if l_handle > 0 then
				Result := handler (l_handle)
			end
		end

	categories: ROSE_LINEAR_ARRAY [STRING] is
		local
			l_handlers: like category_handlers
		do
			refresh_if_necessary
			if categories_saved = Void then
				l_handlers := category_handlers
			end
			Result := categories_saved
		end

	max_category_id: INTEGER is
			-- Returns an upper bound for category_id
		local
			l_i: INTEGER
		do
			if max_category_id_saved = 0 then
				from
					l_i := category_handlers.lower
				until
					l_i > category_handlers.upper
				loop
					if category_handlers.item (l_i) /= Void then
						max_category_id_saved := max_category_id_saved.max (category_handlers.item (l_i).category_id)
					end

					l_i := l_i + 1
				end
				max_category_id_saved := max_category_id_saved.max (1)
			end
			Result := max_category_id_saved
		ensure
			result_greater_zero: Result > 0
		end

feature {NONE} -- Implementation

	pes_internal: BOOLEAN
		-- should category handlers use PES?

	context: like Current is
			-- Self Reference for ENTITY_ACCESSORS
		do
			Result := Current
		end

	max_category_id_saved: INTEGER

	variable_security_saved: INTEGER
	context_variables: CONTEXT_VARIABLES
			-- Variables Support Implementation

	mapping_reader_saved: like mapping_reader
	mapping_override_reader_saved: like mapping_override_reader
	all_mapping_reader_saved: like all_mapping_reader
	all_mapping_override_reader_saved: like all_mapping_override_reader
		-- Cache for all_mapping_override_reader

	local_timeout_period: DATE_TIME_DURATION
			-- How long to wait before initiating a context refresh, applies to this context only

	global_timeout_period: DATE_TIME_DURATION is
			-- How long to wait before initiating a context refresh, applies to all contexts
		once
			create Result.make (0,0,0,0,1,0)
			create Result.make (0,0,0,24,0,0)
		ensure
			valid_result: Result /= Void
		end

	refresh_if_necessary is
			-- Periodically we check to see if we should flush the caches.
		local
			l_now: ROSE_DATE
		do
			if not pes_server then
				create l_now.make_now
				if next_refresh_time /= Void and then l_now > next_refresh_time then
					create next_refresh_time.make_now
					next_refresh_time.add (timeout_period)
					refresh
				end
			end
		end

	next_refresh_time: ROSE_DATE
			-- Time when a refresh will be run

	timeout_period: DATE_TIME_DURATION is
			-- How long to wait before initiating a context refresh (min of global and local)
		do
			Result := local_timeout_period
			if Result = Void or else Result > global_timeout_period then
				Result := global_timeout_period
			end
		ensure
			valid_result: Result /= Void
 		end

	category_handlers: ROSE_LINEAR_ARRAY [CATEGORY_HANDLER [like Current,ENTITY]] is
			-- Collection of category handlers
		local
			l_categories: LINEAR [INTEGER]
			l_name: STRING
		do
			refresh_if_necessary
			if categories_saved = Void then
				distinct_category_predicate.retrieve_atomic (Void,Void)
				create categories_saved.make_empty
				if category_handlers_saved = Void then
					create category_handlers_saved.make_empty
					l_categories := distinct_category_predicate.array
					from l_categories.start until l_categories.off loop
						l_name := category_names.item (l_categories.item)
						if l_name /= Void then
							categories_saved.extend (l_name)
						end
						l_categories.forth
					end
					categories_saved.sort_by (agent compare_strings)
				end
				initialize_accessors
			end
			Result := category_handlers_saved
		end
	categories_saved: like categories
	category_handlers_saved: like category_handlers

feature -- Ur Entities

	ur_entity (a_category_id: INTEGER): ENTITY is
			-- A fake entity with 'entity_id' set to 0.
		do
			if handler (a_category_id) /= Void then
				Result := handler (a_category_id).entity (0)
			end
		end

feature -- Company tree entities

	company_from_id (a_id: INTEGER): like company_template is
			-- If this is the entity_id of a listing, then listing->security->company, if security then security->company, else company
		local
			l_listing: like listing_template
			l_security: like security_template
			l_company: like company_template
		do
			l_listing := listing_accessor.entity (a_id)
			if l_listing /= Void then
				l_security := l_listing.security
				if l_security /= Void then
					Result := l_security.company
				end
			else
				l_security := security_accessor.entity (a_id)
				if l_security /= Void then
					Result := l_security.company
				else
					l_company := company_accessor.entity (a_id)
					if	l_company /= Void then
						Result := l_company
					end
				end
			end
		end

	entity_from_id (a_id: INTEGER): ENTITY is
		require
			meaningful_id: a_id > 0
		local
			l_listing: like listing_template
			l_security: like security_template
			l_company: like company_template
			l_handler: CATEGORY_HANDLER[like Current,ENTITY]
		do
			l_listing := listing_accessor.entity (a_id)
			if l_listing /= Void then
				Result := l_listing
			else
				l_security := security_accessor.entity (a_id)
				if l_security /= Void then
					Result := l_security
				else
					l_company := company_accessor.entity (a_id)
					if	l_company /= Void then
						Result := l_company
					end
				end
			end
			if Result = Void then
				from category_handlers.start until category_handlers.off or Result /= Void loop
					l_handler ?= category_handlers.current_item
					if l_handler /= Void then
						Result := l_handler.entity (a_id)
					end
					category_handlers.forth
				end
			end
		end

	entity_key_history (a_entity: ENTITY; a_identifier: ENTITY_IDENTIFIER): ROSE_LINEAR_ARRAY [ENTITY_ATTRIBUTE_RECORD]
			-- Returns the history of attribute records (of type 'a_identifier') for 'a_entity'.
			-- Are you here because you are hitting this routine too often?  See 'entity_key_history_miss_metrics'.
		require
			server_not_void: is_valid
		local
			l_triple: KEY_TRIPLE[INTEGER, INTEGER, INTEGER]
		do
			debug ("entity_perf")
				print ("entity key lookup " + a_entity.handler.debug_output + "." + a_identifier.debug_output + "%N")
			end
			create Result.make_empty
			key_predicate.set_container (Result)
			key_predicate.retrieve_atomic (Void,<<a_entity.category_id,a_identifier.handle,a_entity.handle>>)
			if not key_predicate.success then
				error.raise("failure to load key history records for category " + a_entity.category_id.out + " entity " + a_entity.handle.out)
			end

			create l_triple.make(context_index, a_entity.category_id, a_identifier.handle)
			entity_key_history_misses[l_triple] := entity_key_history_misses[l_triple] + 1
		end

feature -- Metrics.

	metrics: ROSE_LINEAR_ARRAY[PAIR[STRING, DOUBLE]]
	-- <Precursor>
	do
		create Result.make_empty
		Result.append_array(entity_key_history_miss_metrics)
		Result.append_array(category_entity_records_miss_metrics)
		Result.append_array(entity_records_miss_metrics)
		Result.append_array(relationship_miss_metrics)
	end

	entity_key_history_miss_metrics: ROSE_LINEAR_ARRAY[PAIR[STRING, DOUBLE]]
	-- Returns the metrics stored in 'entity_key_history_misses', formatted for process metrics.
	--
	-- You may have come here because you noticed metrics of the form "SUPPLIER_CONTEXT.entity_key_history misses" and
	-- you are wondering what to do.  Let's say that you have a high value for "RESEARCH/COMPANY/XRF CODE".  If you
	-- cache this early in your program, you can eliminate all of these calls to 'entity_key_history'.
	--
	--		research_context.company_accessor.cache_attributes(<<research_context.identifier_from_name("XRF CODE")>>)
	--
	-- Better is not to use the string "XRF CODE" (not all identifiers are built with 'identifier_from_name')
	-- but to use the constants in INVESTMENT_ENTITY_IDENTIFIERS.  Unfortunately, you won't find all identifiers there.
	--
	--		l_research_ur_entity := research_context.company_accessor.ur_entity					-- Any accessor will do, it's the context that matters.
	--		research_context.company_accessor.cache_attributes(<<l_research_ur_entity.xrf>>)	-- Yes, 'xrf', not 'xrf_code'.
	--
	-- You should supply a method to turn off this caching during debugging.
	local
		l_triple: KEY_TRIPLE[INTEGER, INTEGER, INTEGER]
	do
		create Result.make_empty
		across
			entity_key_history_misses as c_entity_key_history_misses
		loop
			l_triple := c_entity_key_history_misses.key
			Result.extend(metric_from_data("SUPPLIER_CONTEXT.entity_key_history misses", l_triple.first, l_triple.second, l_triple.third, -1, c_entity_key_history_misses.item))
		end
	ensure then
		valid_result: Result /= Void
	end

	entity_key_history_misses: HASH_TABLE[INTEGER_64, KEY_TRIPLE[INTEGER, INTEGER, INTEGER]]
	-- Number of times we missed the cache and had to go to the database in 'entity_key_history', indexed by [context_index, category_id, identifier.handle].
	once
		register_metric_source
		create Result.make(5)
	end

	entity_records_miss_metrics: ROSE_LINEAR_ARRAY[PAIR[STRING, DOUBLE]]
	-- Returns the metrics stored in 'entity_records_misses', formatted for process metrics.
	--
	-- You may have come here because you noticed metrics of the form "SUPPLIER_CONTEXT.entity_records misses" and
	-- you are wondering what to do.  Let's say that you have a high value for "RESEARCH/COMPANY/XRF CODE".  If you
	-- cache this early in your program, you can eliminate the database calls in 'entity_records'.
	--
	--		research_context.identifier_from_rose(research_context.xrf_code_type_id).load_unique_key_records_by_category(research_context.company_category_id)
	--
	-- You should supply a method to turn off this caching during debugging.
	local
		l_triple: KEY_TRIPLE[INTEGER, INTEGER, INTEGER]
	do
		create Result.make_empty
		across
			entity_records_misses as c_entity_records_misses
		loop
			l_triple := c_entity_records_misses.key
			Result.extend(metric_from_data("SUPPLIER_CONTEXT.entity_records misses", l_triple.first, l_triple.second, l_triple.third, -1, c_entity_records_misses.item))
		end
	ensure then
		valid_result: Result /= Void
	end

	entity_records_misses: HASH_TABLE[INTEGER_64, KEY_TRIPLE[INTEGER, INTEGER, INTEGER]]
	-- Number of times we missed the cache and had to go to the database in 'entity_records', indexed by [context_index, category_id, identifier.handle].
	once
		register_metric_source
		create Result.make(5)
	end

	category_entity_records_miss_metrics: ROSE_LINEAR_ARRAY[PAIR[STRING, DOUBLE]]
	-- Returns the metrics stored in 'category_entity_records_misses', formatted for process metrics.
	--
	-- You may have come here because you noticed metrics of the form "SUPPLIER_CONTEXT.category_entity_records misses" and
	-- you are wondering what to do.  Let's say that you have a high value for "RESEARCH/COMPANY".  If you
	-- cache this early in your program, you can eliminate all of these calls to 'category_entity_records'.
	--
	--		research_context.company_accessor.load_all_entities
	--
	-- You should supply a method to turn off this caching during debugging.
	local
		l_pair: KEY_PAIR[INTEGER, INTEGER]
	do
		create Result.make_empty
		across
			category_entity_records_misses as c_category_entity_records_misses
		loop
			l_pair := c_category_entity_records_misses.key
			Result.extend(metric_from_data("SUPPLIER_CONTEXT.category_entity_records misses", l_pair.first, l_pair.second, -1, -1, c_category_entity_records_misses.item))
		end
	ensure then
		valid_result: Result /= Void
	end

	category_entity_records_misses: HASH_TABLE[INTEGER_64, KEY_PAIR[INTEGER, INTEGER]]
	-- Number of times we missed the cache and had to go to the database in 'category_entity_records', indexed by [context_index, category_id].
	once
		register_metric_source
		create Result.make(5)
	end

	relationship_miss_metrics: ROSE_LINEAR_ARRAY[PAIR[STRING, DOUBLE]]
	-- Returns the metrics stored in 'relationship_misses', formatted for process metrics.
	--
	-- You may have come here because you noticed metrics of the form "SUPPLIER_CONTEXT.relationship misses" and
	-- you are wondering what to do.  Let's say that you have a high value for "RESEARCH / COMPANY - SECURITY".  If you
	-- cache this early in your program, you can eliminate the database calls in 'entity_records'.
	--
	--		research_context.company_to_security.load_relationship_and_corelationship
	--
	-- You should supply a method to turn off this caching during debugging.
	local
		l_pair: KEY_PAIR[INTEGER, INTEGER]
	do
		create Result.make_empty
		across
			relationship_misses as c_relationship_misses
		loop
			l_pair := c_relationship_misses.key
			Result.extend(metric_from_data("SUPPLIER_CONTEXT.relationship misses", l_pair.first, -1, -1, l_pair.second, c_relationship_misses.item))
		end
	ensure then
		valid_result: Result /= Void
	end

	relationship_misses: HASH_TABLE[INTEGER_64, KEY_PAIR[INTEGER, INTEGER]]
	-- Number of times we missed the cache and had to go to the database in 'relationship_records' and 'corelationship_records', indexed by
	-- [context_index, relationship_type_id].
	once
		register_metric_source
		create Result.make(5)
	end

	metric_from_data(a_name: STRING; a_context_index, a_category_id, a_identifier_handle, a_relationship_type_id: INTEGER a_value: INTEGER_64): PAIR[STRING, DOUBLE]
	-- Converts the input arguments into a single metric name/value pair.
	local
		l_context: SUPPLIER_CONTEXT
		l_context_name: STRING
		l_category_name: STRING
		l_identifier_name: STRING
		l_relationship_name: STRING
		l_metric_name: STRING
	do
		l_metric_name := a_name

		if attached context_from_index(a_context_index) as al_context and then
		   attached al_context.name as al_context_name
		then
			l_context := al_context
			l_context_name := al_context_name + " (" + a_context_index.out + ")"
		else
			l_context := Void
			l_context_name := "context = " + a_context_index.out
		end

		l_metric_name.append("/" + l_context_name)

		if a_category_id /= -1
		then
			if attached category_name(a_category_id) as al_category_name
			then
				l_category_name := al_category_name + " (" + a_category_id.out + ")"
			else
				l_category_name := "category_id = " + a_category_id.out
			end

			l_metric_name.append("/" + l_category_name)
		end

		if a_identifier_handle /= -1
		then
			if attached l_context and then
			   attached l_context.identifier(a_identifier_handle) as al_identifier and then
			   attached al_identifier.name as al_name
			then
				l_identifier_name := al_name + " (" + a_identifier_handle.out + ")"
			else
				l_identifier_name := "identifier = (" + a_identifier_handle.out + ")"
			end

			l_metric_name.append("/" + l_identifier_name)
		end

		if a_relationship_type_id /= -1
		then
			if attached l_context and then
			   attached l_context.relationship(a_relationship_type_id) as al_relationship and then
			   attached al_relationship.name as al_relationship_name
			then
				l_relationship_name := al_relationship_name + " (" + a_relationship_type_id.out + ")"
				l_relationship_name.replace_substring_all(" / ", " - ")			-- Can't have slashes in a metric name.
			else
				l_relationship_name := "relationship = (" + a_relationship_type_id.out + ")"
			end

			l_metric_name.append("/" + l_relationship_name)
		end

		create Result.make(l_metric_name, a_value.to_double)
	end

	register_metric_source
	-- Register this copy of Current as the one and only source of metrics.
	once
		process_metrics.register_source(Current)
	end

feature -- Attributes Types

	attribute_names: ROSE_LINEAR_ARRAY [STRING]

	attribute_name_by_id (a_attribute_id: INTEGER): STRING is
		do
			Result := attribute_id_types.item (a_attribute_id)
		end

	attribute_id_by_name (a_attribute_name: STRING): INTEGER is
		require
			valid_args: a_attribute_name /= Void
		do
			Result := attribute_type_ids.item (a_attribute_name.as_upper)
		end

	attribute_is_uniq_identifier (a_attribute_id: INTEGER): BOOLEAN is
		do
			Result := attribute_is_uniq_by_id_cache.item (a_attribute_id)
		end

	attribute_id_types: HASH_TABLE [STRING,INTEGER]
	attribute_type_ids: HASH_TABLE [INTEGER,STRING] -- attribute_type_ids stored by name

feature {NONE} -- Implementation

	identifiers: HASH_TABLE [like entity_identifier_template,INTEGER]

	attribute_is_uniq_by_id_cache: HASH_TABLE [BOOLEAN,INTEGER]

	add_attribute_type (a_record: ATTRIBUTE_TYPE_RECORD) is
			-- Add attribute type corresponding to 'a_record'
		do
			a_record.name.to_upper
			a_record.name.prune_all_trailing (' ')
			attribute_type_ids.put (a_record.attribute_type_id,a_record.name)
			attribute_id_types.put (a_record.name, a_record.attribute_type_id)
			attribute_names.extend (a_record.name)
			attribute_is_uniq_by_id_cache.extend (a_record.is_unique_identifier, a_record.attribute_type_id)
		end

	setup_attribute_types is
		-- retrieve attibute types from the attribute_type table in the context
		local
			l_reader: SIMPLE_PREDICATE [ATTRIBUTE_TYPE_RECORD]
		do
			create attribute_id_types.make (17)
			create attribute_type_ids.make (17)
			create attribute_names.make_empty
			create attribute_is_uniq_by_id_cache.make (17)
			if has_entities
			 then
			 	server.begin_transaction
				create l_reader.make_strict (server,"select * from attribute_type with (readuncommitted)")
				l_reader.apply (agent add_attribute_type,Void,Void)
				server.commit
			end
		end

	attribute_types_saved: INTEGER_PREDICATE

	attribute_types: INTEGER_PREDICATE is
		do
			if attribute_types_saved = Void then
				create attribute_types_saved.make_verbose (server,"select * from attribute_category_relationship with (readuncommitted) where category_id = ?",
														   "attribute_types",name)
				if not attribute_types_saved.success then
					error.raise ("could not create predicate")
				end
			end
			Result := attribute_types_saved
		end

	has_entities_saved: BOOLEAN_REF

feature -- Attribute Mapping

	rose_attribute_from_local: HASH_TABLE [INTEGER,INTEGER]
	rose_attribute_to_local: HASH_TABLE [INTEGER,INTEGER]

feature {NONE} -- Attribute Mapping Implementation

	map_attributes is
			-- Retrieve the supplier attribute type maps
		local
			l_servers: DATABASE_SERVER
			l_server: SQL_DATA_SERVER
			l_predicate: SIMPLE_PREDICATE [ATTRIBUTE_TYPE_MAP_RECORD]
			l_tran_started: BOOLEAN
		do
			create rose_attribute_from_local.make (200)
			create rose_attribute_to_local.make (200)
			if context_index /= 0 then
				create l_servers
				l_server := l_servers.servers.rose_entity_foreground_server.auxiliary_server
				l_server.begin_transaction		-- Mandatory on an auxiliary_server.
				l_tran_started := True
				create l_predicate.make (l_server, "select * from attribute_type_map where supplier_product_id = ?")
 				if not l_predicate.success then
					error.raise ("Could not create 'attribute_type_map' predicate")
				end
				l_predicate.apply (agent add_type_map, Void, <<supplier_id>>)
 				if not l_predicate.success then
					error.raise ("Could not apply 'attribute_type_map' predicate")
				end


				l_server.commit					-- Mandatory on an auxiliary_server.
				l_predicate.facilitator.release	-- Frees up resources that would otherwise have to wait for a full_collect.
				l_predicate := Void				-- Just making sure that no one uses it since it is now disconnected.
			end
		rescue
			if l_tran_started
			then
				l_server.rollback
				l_tran_started := False
			end
		end

	add_type_map (a_record: ATTRIBUTE_TYPE_MAP_RECORD) is
			-- Add two-way mappings between `a_record.attribute_type_id' and `a_record.supplier_attribute_type_id'.
		require
			valid_record: a_record /= Void
			valid_tables: rose_attribute_from_local /= Void and rose_attribute_to_local /= Void
		do
			rose_attribute_from_local.put (a_record.attribute_type_id, a_record.supplier_attribute_type_id)
			rose_attribute_to_local.put (a_record.supplier_attribute_type_id, a_record.attribute_type_id)
		end

feature -- Attribute keys

	all_key_records (a_category_id : INTEGER;a_identifier : ENTITY_IDENTIFIER): ROSE_LINEAR_ARRAY[ENTITY_ATTRIBUTE_RECORD] is
			-- Entity attribute records for `a_category_id' and `a_identifier'
		require
			a_category_id_positive: a_category_id > 0
			a_identifier_attached: a_identifier /= Void
		do
			all_key_predicate.retrieve_atomic(void,<<a_category_id,a_identifier.handle>>)
			if not all_key_predicate.success then
				error.raise("failure to load key history records for category " + a_category_id.out)
			end
			Result := all_key_predicate.array.twin
			all_key_predicate.array.wipe_out
		ensure
			not_void: Result /= Void
		end

feature {CATEGORY_HANDLER} -- Implementation

	all_entity_records (a_handler: CATEGORY_HANDLER [like Current,ENTITY]): ROSE_LINEAR_ARRAY [ENTITY_ID_BASE] is
			-- All entity records for `a_handler'
		require
			valid_handler: a_handler /= Void
		do
			create Result.make_empty
			all_entity_predicate.set_container (Result)
			all_entity_predicate.retrieve_atomic (Void,<<a_handler.category_id>>)
			if not all_entity_predicate.success then
				error.raise ("failure to execute predicate")
			end
		end

	category_entity_records (a_handler: CATEGORY_HANDLER [like Current,ENTITY]; a_handle: INTEGER): ROSE_LINEAR_ARRAY [ENTITY_ID_BASE] is
			-- The record for `a_handle' in `a_handler' type.
			-- Are you here because you are hitting this routine too often?  See 'category_entity_records_miss_metrics'.
		require
			valid_handler: a_handler /= Void
		local
			l_pair: KEY_PAIR[INTEGER, INTEGER]
		do
			debug ("entity_perf")
				print ("entity lookup handle " + a_handler.debug_output + "%N")
			end
			create Result.make_empty
			category_entity_predicate.set_container (Result)
			category_entity_predicate.retrieve_atomic (Void,<<a_handler.category_id,a_handle>>)
			if not category_entity_predicate.success then
				error.raise ("failure to execute predicate")
			end

			create l_pair.make(context_index, a_handler.category_id)
			category_entity_records_misses[l_pair] := category_entity_records_misses[l_pair] + 1
		end

feature {NONE} -- Predicates

	mapping_reader: SIMPLE_PREDICATE [MAPPINGS_RECORD] is
		do
			if mapping_reader_saved = Void then
				create mapping_reader_saved.make_strict (server, "select * from mappings where type = ? order by mnemonic")
			end
			Result := mapping_reader_saved
		end

	mapping_override_reader: SIMPLE_PREDICATE [MAPPINGS_RECORD] is
		local
			l_servers: DATABASE_SERVER
		do
			if mapping_override_reader_saved = Void then
				create l_servers
				create mapping_override_reader_saved.make_strict (l_servers.servers.valuation_data_override_w_server,
																  "select * from mappings_override where context = ? and type = ? " +
																  "order by label, mnemonic")
			end
			Result := mapping_override_reader_saved
		end

	all_mapping_reader: COLLECTING_PREDICATE [MAPPINGS_RECORD] is
			-- all mapping reader
		do
			if all_mapping_reader_saved = Void then
				create all_mapping_reader_saved.make_strict (server, "select * from mappings order by type, mnemonic")
			end
			Result := all_mapping_reader_saved
		end

	all_mapping_override_reader: COLLECTING_PREDICATE [MAPPINGS_RECORD] is
			-- all mapping override reader
		local
			l_servers: DATABASE_SERVER
		do
			if all_mapping_override_reader_saved = Void then
				create l_servers
				create all_mapping_override_reader_saved.make_strict (l_servers.servers.valuation_data_override_w_server,
																	  "select * from mappings_override where context_name = ? " +
																	  "and removal_date = '12/31/9999' order by type, label, mnemonic")
			end
			Result := all_mapping_override_reader_saved
		end

	identifier_entity_predicate: COLLECTING_PREDICATE [ENTITY_ATTRIBUTE_RECORD] is
		do
			if identifier_entity_predicate_saved = Void then
				create identifier_entity_predicate_saved.make_strict (server,"select * from entity_attribute with (readuncommitted) where  attribute_type_id = ? and attribute_value = ? order by end_date desc")
			end
			Result := identifier_entity_predicate_saved
		end

	identifier_entity_like_predicate: COLLECTING_PREDICATE [ENTITY_ATTRIBUTE_RECORD] is
		do
			if identifier_entity_like_predicate_saved = Void then
				create identifier_entity_like_predicate_saved.make_strict (server,"select * from entity_attribute with (readuncommitted,index (av_entity_attribute)) where attribute_type_id = ? and attribute_value like ? order by end_date desc")
			end
			Result := identifier_entity_like_predicate_saved
		end

	key_predicate: COLLECTING_PREDICATE [ENTITY_ATTRIBUTE_RECORD] is
		do
			if key_predicate_saved = Void then
				create key_predicate_saved.make_strict (server,"select * from entity_attribute with (readuncommitted) where category_id = ? and attribute_type_id = ? and entity_id = ? order by end_date desc")
			end
			Result := key_predicate_saved
		end

feature {NONE} -- Implementation

	all_entities_predicate: COLLECTING_PREDICATE [ENTITY_ATTRIBUTE_RECORD] is
			-- Predicate for all entities
		require
			server_not_void: server /= Void
		do
			if all_entities_predicate_saved = Void then
				create all_entities_predicate_saved.make_strict (server,"select * from entity_attribute with (readuncommitted) where attribute_type_id = ? and attribute_value = ? order by end_date desc")
			end
			Result := all_entities_predicate_saved
		end

	all_entities_like_predicate: COLLECTING_PREDICATE [ENTITY_ATTRIBUTE_RECORD] is
			-- Predicate for finding entities with like on attribute_value
		require
			server_not_void: server /= Void
		do
			if all_entities_like_predicate_saved = Void then
				create all_entities_like_predicate_saved.make_strict (server,"select * from entity_attribute with (readuncommitted,index (av_entity_attribute)) where attribute_type_id = ? and attribute_value like ? order by end_date desc")
			end
			Result := all_entities_like_predicate_saved
		end

	entity_predicate: COLLECTING_PREDICATE [ENTITY_ATTRIBUTE_RECORD] is
		require
			server_not_void: server /= Void
		do
			if entity_predicate_saved = Void then
				create entity_predicate_saved.make_strict (server,"select * from entity_attribute with (readuncommitted) where category_id = ? and attribute_type_id = ? and attribute_value = ? order by end_date desc")
			end
			Result := entity_predicate_saved
		end

	entity_like_predicate: COLLECTING_PREDICATE [ENTITY_ATTRIBUTE_RECORD] is
		require
			server_not_void: server /= Void
		do
			if entity_like_predicate_saved = Void then
				create entity_like_predicate_saved.make_strict (server,"select * from entity_attribute with (readuncommitted,index (av_entity_attribute)) where category_id = ? and attribute_type_id = ? and attribute_value like ? order by end_date desc")
			end
			Result := entity_like_predicate_saved
		end

	all_key_predicate: COLLECTING_PREDICATE [ENTITY_ATTRIBUTE_RECORD]
			-- Predicate which retrieves all entity attribute records for given category and attribute type
		require
			server_not_void: server /= Void
		do
			if all_key_predicate_saved = Void then
				create all_key_predicate_saved.make_strict (server,"select * from entity_attribute with (readuncommitted) where category_id = ? and attribute_type_id = ? order by entity_id,end_date desc")
			end
			Result := all_key_predicate_saved
		ensure
			not_void: Result /= Void
		end

	all_entity_attribute_predicate: COLLECTING_PREDICATE [ENTITY_ATTRIBUTE_RECORD] is
			-- Predicate to retrieve attribute records for a given category_id and attribute_type_id
		require
			server_not_void: server /= Void
		do
			if all_entity_attribute_predicate_saved = Void then
				create all_entity_attribute_predicate_saved.make_strict (server,"select * from entity_attribute with (readuncommitted) where category_id = ? and attribute_type_id = ? order by attribute_value,end_date desc")
			end
			Result := all_entity_attribute_predicate_saved
		end

	all_attributes_predicate: COLLECTING_PREDICATE [ENTITY_ATTRIBUTE_RECORD] is
			-- Predicate to retrieve all attribute records (all categories, all attribute types)
		require
			server_not_void: server /= Void
		do
			if all_attributes_predicate_saved = Void then
				create all_attributes_predicate_saved.make_strict (server, "select * from entity_attribute with (readuncommitted) order by attribute_type_id,category_id,entity_id,end_date desc")
			end
			Result := all_attributes_predicate_saved
		end

	attributes_update_predicate: COLLECTING_PREDICATE [ENTITY_ATTRIBUTE_RECORD] is
			-- Predicate to retrieve all attribute records (all categories, all attribute types) which are between a certain update date range
		require
			server_not_void: server /= Void
		do
			if attributes_update_predicate_saved = Void then
				create attributes_update_predicate_saved.make_strict (server, "select * from entity_attribute with (readuncommitted) where update_date > ? and update_date <= ? order by attribute_type_id,category_id,entity_id,end_date desc")
			end
			Result := attributes_update_predicate_saved
		end

	historical_attributes_predicate: COLLECTING_PREDICATE [ENTITY_FREEZER_RECORD] is
			-- Predicate to fetch updated attribute records from the feezer
		require
			server_not_void: server /= Void
		do
			if historical_attributes_predicate_saved = Void then
				create historical_attributes_predicate_saved.make_strict (server, "select * from freezer_resolved_items where resolution_date < ? and resolution_date >= ?")
			end
			Result := historical_attributes_predicate_saved

		end

feature {NONE} -- Implementation

	relationship_reader: COLLECTING_PREDICATE [ENTITY_RELATIONSHIP_RECORD] is
		require
			server_not_void: server /= Void
		do
			if relationship_reader_saved = Void then
				if has_weighted_relationship then
					create relationship_reader_saved.make_strict (server,"select entity_2_id,start_date,end_date,weight from entity_relationship with (readuncommitted) where relationship_type_id = ? and entity_1_id = ? order by end_date")
				else
					create relationship_reader_saved.make_strict (server,"select entity_2_id,start_date,end_date from entity_relationship with (readuncommitted) where relationship_type_id = ? and entity_1_id = ? order by end_date")
				end
			end
			Result := relationship_reader_saved
		end

	all_relationship_reader: COLLECTING_PREDICATE [ENTITY_RELATIONSHIP_RECORD] is
			-- Collecting predicate used to retrieve all relationship records of a paticular relationship type.
		require
			server_not_void: server /= Void
		do
			if all_relationship_reader_saved = Void then
				if has_weighted_relationship then
					create all_relationship_reader_saved.make_strict (server,"select entity_1_id, entity_2_id,start_date,end_date,weight from entity_relationship with (readuncommitted) where relationship_type_id = ? order by entity_1_id, end_date")
				else
					create all_relationship_reader_saved.make_strict (server,"select entity_1_id, entity_2_id,start_date,end_date from entity_relationship with (readuncommitted) where relationship_type_id = ? order by entity_1_id, end_date")
				end
			end
			Result := all_relationship_reader_saved
		end

	all_relationship_reader2: COLLECTING_PREDICATE [ENTITY_RELATIONSHIP_RECORD] is
		require
			server_not_void: server /= Void
		do
			if all_relationship_reader_saved2 = Void then
				if has_weighted_relationship then
					create all_relationship_reader_saved2.make_strict (server,"select relationship_type_id, entity_1_id, entity_2_id,start_date,end_date,weight from entity_relationship with (readuncommitted) order by relationship_type_id, entity_1_id, end_date")
				else
					create all_relationship_reader_saved2.make_strict (server,"select relationship_type_id, entity_1_id, entity_2_id,start_date,end_date from entity_relationship with (readuncommitted) order by relationship_type_id, entity_1_id, end_date")
				end
			end
			Result := all_relationship_reader_saved2
		end

	relationship_update_predicate: COLLECTING_PREDICATE [ENTITY_RELATIONSHIP_RECORD] is
			-- predicate for getting all relatioships updated in a certain intervall
		do
			if relationship_update_predicate_saved = Void then
				if has_weighted_relationship then
					create relationship_update_predicate_saved.make_strict (server,"select relationship_type_id, entity_1_id, entity_2_id,start_date,end_date,weight from entity_relationship with (readuncommitted) where update_date > ? and update_date <= ? order by relationship_type_id, entity_1_id, end_date")
				else
					create relationship_update_predicate_saved.make_strict (server,"select relationship_type_id, entity_1_id, entity_2_id,start_date,end_date from entity_relationship with (readuncommitted) where update_date > ? and update_date <= ? order by relationship_type_id, entity_1_id, end_date")
				end

			end
			Result := relationship_update_predicate_saved
		end

	corelationship_reader: COLLECTING_PREDICATE [ENTITY_RELATIONSHIP_RECORD] is
			-- predicate
		require
			server_not_void: server /= Void
		do
			if corelationship_reader_saved = Void then
				if has_weighted_relationship then
					create corelationship_reader_saved.make_strict (server, "select entity_1_id,start_date,end_date,weight from entity_relationship with (readuncommitted) where relationship_type_id = ? and entity_2_id = ? order by end_date")
				else
					create corelationship_reader_saved.make_strict (server, "select entity_1_id,start_date,end_date from entity_relationship with (readuncommitted) where relationship_type_id = ? and entity_2_id = ? order by end_date")
				end
			end
			Result := corelationship_reader_saved
		end

	has_weighted_relationship: BOOLEAN is
		require
			server_not_void: server /= Void
		do
			Result := server.has_column_name ("entity_relationship", "weight")
		end

feature {CATEGORY_HANDLER,ENTITY_IDENTIFIER,RP_REFERENCE_SYSTEM} -- Implementation

	count_entity_from_category (a_category: CATEGORY_HANDLER[SUPPLIER_CONTEXT, ENTITY]): INTEGER is
			-- count entities in `a_category'
		require
			valid_category: a_category /= Void
		local
			l_pred: INTEGER_PREDICATE
			l_res: ARRAY[INTEGER]
		do
			create l_res.make_filled (0, 1, 1)
			create l_pred.make_strict (server,"select count(*) from entity_id where category_id = ? group by category_id")
			l_pred.set_container (l_res)
			l_pred.retrieve_atomic (Void , <<a_category.category_id>>)
			Result := l_res.item (1)
		end

	entity_records (a_key : STRING; a_handler: CATEGORY_HANDLER [like Current,ENTITY]; a_identifier: ENTITY_IDENTIFIER) : ROSE_LINEAR_ARRAY[ENTITY_ATTRIBUTE_RECORD] is
			-- Attribute records in 'a_category_id', matching 'a_key'.  Note that 'a_key' may
			-- include wildcards (%).
		require
			valid_handler: a_handler /= Void
			valid_identifier: a_identifier /= Void
		local
			l_handle,l_category_id : INTEGER
			l_triple: KEY_TRIPLE[INTEGER, INTEGER, INTEGER]
		do
			l_category_id := a_handler.category_id
			l_handle := a_identifier.handle
			if a_key /= Void then
				if not a_key.has ('%%') then
					Result := a_identifier.unique_records_by_category (a_key, l_category_id)
				end
				if Result = Void then
					debug ("bogie_perf")
						print ("entity lookup " + a_handler.debug_output + " by " + a_identifier.debug_output + "%N")
					end
					create Result.make_empty
					entity_predicate.set_container(Result)
					entity_predicate.retrieve_atomic(Void,<<l_category_id,l_handle,a_key>>)
					if not entity_predicate.success then
						error.raise("failure to retrieve records for " + a_key)
					end
					if Result.is_empty and then a_key.has('%%') then
						entity_like_predicate.set_container(Result)
						entity_like_predicate.retrieve_atomic(Void,<<l_category_id,l_handle,a_key>>)
						if not entity_like_predicate.success then
							error.raise("failure to retrieve like records for " + a_key)
						end
					end

					create l_triple.make(context_index, a_handler.category_id, a_identifier.handle)
					entity_records_misses[l_triple] := entity_records_misses[l_triple] + 1
				end
			end
			if Result = Void then
				create Result.make_empty
			end
		ensure
			result_not_void: Result /= Void
		end

feature {ENTITY_IDENTIFIER} -- Implementation

	all_entity_records_from_key (a_key : STRING;a_identifier : ENTITY_IDENTIFIER) : ROSE_LINEAR_ARRAY[ENTITY_ATTRIBUTE_RECORD] is
	-- Attribute records in ANY category, matching 'a_key'.  Note that 'a_key' may
	-- include wildcards (%).
		require
			key_not_void: a_key /= Void
		local
			l_handle : INTEGER
		do
			l_handle := a_identifier.handle
			create Result.make_empty
			if a_key /= Void then
				if not a_key.has('%%') then
					all_entities_predicate.set_container(Result)
					all_entities_predicate.retrieve_atomic(Void,<<l_handle,a_key>>)
					if not all_entities_predicate.success then
						error.raise("failure to retrieve all records for " + a_key)
					end
				else
					all_entities_like_predicate.set_container(Result)
					all_entities_like_predicate.retrieve_atomic(Void,<<l_handle,a_key>>)
					if not all_entities_like_predicate.success then
						error.raise("failure to retrieve all records like " + a_key)
					end
				end
			end
		ensure
			result_not_void : Result /= Void
		end

feature -- Relationship

	count_relationships_from_relationship (a_relationship: RELATIONSHIP_HANDLER[ENTITY,ENTITY]): INTEGER is
			-- count entity_2 in `a_relationship'
		require
			valid_relationship: a_relationship /= Void
		local
			l_pred: TUPLE_PREDICATE [TUPLE [type_id: INTEGER; type_count: INTEGER]]
			l_res: ARRAY [TUPLE [type_id: INTEGER; type_count: INTEGER]]
			i: INTEGER
		do
			if relationship_counts_saved = Void then
				create relationship_counts_saved.make (400)
				create l_res.make_filled (Void, 1, 1)
				create l_pred.make (server,"select distinct relationship_type_id, count(entity_1_id) from entity_relationship group by relationship_type_id ")
				if not l_pred.success then
					error.raise ("Could not create l_pred")
				end
				l_pred.set_container (l_res)
				l_pred.retrieve_atomic (Void , Void)
				if not l_pred.success then
					error.raise ("Could not execute l_pred")
				end
				from
					i := l_res.lower

				until
					i > l_res.upper
				loop
					relationship_counts_saved.put (l_res.item(i).type_count, l_res.item(i).type_id)
					i := i + 1
				end
			end
			Result := relationship_counts_saved.item (a_relationship.relationship_id)
		end


	count_corelationships_from_relationship (a_relationship: RELATIONSHIP_HANDLER[ENTITY,ENTITY]): INTEGER is
			-- count entity_2 in `a_relationship'
		require
			valid_relationship: a_relationship /= Void
		local
			l_pred: TUPLE_PREDICATE [TUPLE[type_id: INTEGER; type_count: INTEGER]]
			l_res: ARRAY [TUPLE [type_id: INTEGER; type_count: INTEGER]]
			i: INTEGER
		do
			if corelationship_counts_saved = Void then
				create corelationship_counts_saved.make (400)
				create l_res.make_filled (Void, 1, 1)
				create l_pred.make (server,"select distinct relationship_type_id, count(entity_2_id) from entity_relationship group by relationship_type_id ")
				if not l_pred.success then
					error.raise ("Could not create l_pred")
				end
				l_pred.set_container (l_res)
				l_pred.retrieve_atomic (Void , Void)
				if not l_pred.success then
					error.raise ("Could not execute l_pred")
				end
				from
					i := l_res.lower

				until
					i > l_res.upper
				loop
					corelationship_counts_saved.put (l_res.item(i).type_count, l_res.item(i).type_id)
					i := i + 1
				end
			end
			Result := corelationship_counts_saved.item (a_relationship.relationship_id)
		end

	multiple_relationship_records (a_entities: LINEAR [ENTITY]): ROSE_LINEAR_ARRAY [ENTITY_RELATIONSHIP_RECORD] is
		require
			valid: is_valid
		local
			l_insert_predicate: EXECUTION_PREDICATE
			l_reader: COLLECTING_PREDICATE [ENTITY_RELATIONSHIP_RECORD]
		do
			server.sql_execute_immediate ("create table #entities (id int)")
			create l_insert_predicate.make (context.server,"insert into #entities values (?)")
			if not l_insert_predicate.success then
				error.raise ("Could not create l_insert_predicate")
			end
			from a_entities.start until a_entities.off loop
				l_insert_predicate.execute (Void,<<a_entities.item.handle>>)
				if not l_insert_predicate.success then
					error.raise ("Could not execute l_insert_predicate")
				end
				a_entities.forth
			end
			create l_reader.make_strict (server,
								 "select entity_2_id,start_date,end_date from entity_relationship,#entities where entity_1_id = id group by entity_2_id,start_date,end_date")
			l_reader.retrieve_atomic (Void,Void)
			server.sql_execute_immediate ("drop table #entities")
			Result := l_reader.array
		end

	multiple_corelationship_records (a_entities: LINEAR [ENTITY]): ROSE_LINEAR_ARRAY [ENTITY_RELATIONSHIP_RECORD] is
		local
			l_insert_predicate: EXECUTION_PREDICATE
			l_reader: COLLECTING_PREDICATE [ENTITY_RELATIONSHIP_RECORD]
		do
			server.sql_execute_immediate ("create table #entities (id int)")
			create l_insert_predicate.make (context.server,"insert into #entities values (?)")
			if not l_insert_predicate.success then
				error.raise ("Could not create l_insert_predicate")
			end
			from a_entities.start until a_entities.off loop
				l_insert_predicate.execute (Void,<<a_entities.item.handle>>)
				if not l_insert_predicate.success then
					error.raise ("Could not execute l_insert_predicate")
				end
				a_entities.forth
			end
			create l_reader.make_strict (server,
										 "select entity_1_id,start_date,end_date from entity_relationship,#entities where entity_2_id = id group by entity_1_id,start_date,end_date")
			l_reader.retrieve_atomic (Void,Void)
			server.sql_execute_immediate ("drop table #entities")
			Result := l_reader.array
		end

	relationship_records (a_handler: RELATIONSHIP_HANDLER [ENTITY,ENTITY];a_entity: ENTITY): ROSE_LINEAR_ARRAY [ENTITY_RELATIONSHIP_RECORD] is
				--  all relationship records for relationship `a_handler' and entity `a_entity'
		require
			valid: is_valid
			a_handler_not_void: a_handler /= Void
		local
			l_pair: KEY_PAIR[INTEGER, INTEGER]
		do
			debug ("entity_perf")
				print ("entity lookup relationship " + "[" + a_handler.context_name + "]" + a_handler.name + "%N")
			end
			relationship_reader.retrieve_atomic (Void,<<a_handler.relationship_id,a_entity.handle>>)
			if not relationship_reader.success then
				error.raise ("Could not execute relationship_reader")
			end
			Result := relationship_reader.array

			create l_pair.make(context_index, a_handler.relationship_id)
			relationship_misses[l_pair] := relationship_misses[l_pair] + 1
		end

	all_relationship_records (a_handler: RELATIONSHIP_HANDLER [ENTITY,ENTITY]): ROSE_LINEAR_ARRAY [ENTITY_RELATIONSHIP_RECORD] is
				--  all relationship records for relationship `a_handler' and entity `a_entity'
		require
			valid: is_valid
			a_handler_not_void: a_handler /= Void
		do
			all_relationship_reader.retrieve_atomic (Void,<<a_handler.relationship_id>>)
			if not all_relationship_reader.success then
				error.raise ("Could not execute all_relationship_reader")
			end
			Result := all_relationship_reader.array
		end

	corelationship_records (a_handler: RELATIONSHIP_HANDLER [ENTITY,ENTITY];a_entity: ENTITY): ROSE_LINEAR_ARRAY [ENTITY_RELATIONSHIP_RECORD] is
			-- return corelationship records
		require
			valid: is_valid
			a_handler_not_void: a_handler /= Void
		local
			l_pair: KEY_PAIR[INTEGER, INTEGER]
		do
			debug ("entity_perf")
				print ("entity lookup corelationship " + "[" + a_handler.context_name + "]" + a_handler.name + "%N")
			end
			corelationship_reader.retrieve_atomic (Void,<<a_handler.relationship_id,a_entity.handle>>)
			if not corelationship_reader.success then
				error.raise ("Could not execute corelationship_reader")
			end
			Result := corelationship_reader.array

			create l_pair.make(context_index, a_handler.relationship_id)
			relationship_misses[l_pair] := relationship_misses[l_pair] + 1
		end

	some_relationship_record (a_handler: RELATIONSHIP_HANDLER [ENTITY,ENTITY];a_entity: ENTITY): ENTITY_RELATIONSHIP_RECORD is
			-- return one relationship record for a_handler and a_entity
 		require
			valid: is_valid
			a_handler_not_void: a_handler /= Void
		do
			create Result
			relationship_reader.retrieve_one_atomic (Result,Void,<<a_handler.relationship_id,a_entity.handle>>)
			if not relationship_reader.success then
				error.raise ("Could not execute relationship_reader")
			end
			if relationship_reader.index /= 1 then
				Result := Void
			end
		end

	some_corelationship_record (a_handler: RELATIONSHIP_HANDLER [ENTITY,ENTITY];a_entity: ENTITY): ENTITY_RELATIONSHIP_RECORD is
		require
			valid: is_valid
		do
			create Result
			corelationship_reader.retrieve_one_atomic (Result,Void,<<a_handler.relationship_id,a_entity.handle>>)
			if not corelationship_reader.success then
				error.raise ("Could not execute corelationship_reader")
			end
			if corelationship_reader.index /= 1 then
				Result := Void
			end
		end

feature {NONE}  -- Relationship implementation

	relationship_records_table: HASH_TABLE[HASH_TABLE[ROSE_LINEAR_ARRAY [ENTITY_RELATIONSHIP_RECORD], INTEGER], INTEGER] is
			-- reference to table
		do
			if relationship_records_table_saved = void then
				create relationship_records_table_saved.make (500)
			end
			Result := relationship_records_table_saved
		end

	relationship_records_table_saved: HASH_TABLE[HASH_TABLE[ROSE_LINEAR_ARRAY [ENTITY_RELATIONSHIP_RECORD], INTEGER], INTEGER]


feature {NONE} -- Implementation

	category_entity_predicate: SIMPLE_PREDICATE [ENTITY_ID_BASE] is
		do
			if category_entity_predicate_saved = Void then
				create category_entity_predicate_saved.make_verbose (server,"select * from entity_id with (readuncommitted) where category_id = ? and entity_id = ?",
																	"category_entity_predicate",name)
			end
			Result := category_entity_predicate_saved
		end

	all_entity_predicate: SIMPLE_PREDICATE [ENTITY_ID_BASE] is
		do
			if all_entity_predicate_saved = Void then
				create all_entity_predicate_saved.make_verbose (server, "select * from entity_id where category_id = ?",
															   "all_entity_predicate",name)
			end
			Result := all_entity_predicate_saved
		end

	all_entity_by_category_predicate: COLLECTING_PREDICATE [ENTITY_ID_BASE] is
		do
			if all_entity_by_category_predicate_saved = Void then
				create all_entity_by_category_predicate_saved.make_verbose (server, "select * from entity_id order by category_id",
															   "all_entity_sorted_by_category_predicate",name)
			end
			Result := all_entity_by_category_predicate_saved
		end

	entity_update_predicate: COLLECTING_PREDICATE [ENTITY_ID_BASE] is
			-- predicate for getting all entities updated in a certain intervall
		do
			if entity_update_predicate_saved = Void then
				create entity_update_predicate_saved.make_verbose (server,"select * from entity_id where update_date > ? and update_date <= ?",
														"entity_update_predicate",name)
			end
			Result := entity_update_predicate_saved
		end

	count_attributes_predicate: TUPLE_PREDICATE [TUPLE [attribute_type_id: INTEGER; attribute_count: INTEGER]] is
			-- Count the number of attribute records per type
		do
			if count_attributes_predicate_saved = Void then
				create count_attributes_predicate_saved.make(context.server, "SELECT attribute_type_id, COUNT(attribute_type_id) AS attribute_count FROM entity_attribute GROUP BY attribute_type_id ORDER BY attribute_type_id")
			end
			Result := count_attributes_predicate_saved
		end

	distinct_category_predicate: INTEGER_PREDICATE is
		do
			if distinct_category_predicate_saved = Void then
				create distinct_category_predicate_saved.make_verbose (server,"select distinct category_id from attribute_category_relationship with (readuncommitted)",
																	  "distinct_category_predicate",name)
			end
			Result := distinct_category_predicate_saved
		end

	distinct_relationship_predicate: INTEGER_PREDICATE is
		do
			if distinct_relationship_predicate_saved = Void then
				create distinct_relationship_predicate_saved.make_verbose (server,"select distinct relationship_type_id from entity_relationship with (readuncommitted)",
																		  "distinct_relationship_predicate",name)
			end
			Result := distinct_relationship_predicate_saved
		end

feature {NONE} -- Implementation

	reset_predicates is
			-- Reset all stored predicates
		do
			entity_predicate_saved := Void
			entity_like_predicate_saved := Void
			every_entity_predicate_saved := Void
			every_entity_like_predicate_saved := Void
			identifier_entity_predicate_saved := Void
			identifier_entity_like_predicate_saved := Void
			key_predicate_saved := Void
			all_key_predicate_saved := Void
			category_entity_predicate_saved := Void
			all_entity_predicate_saved := Void
			all_entity_attribute_predicate_saved := Void
			distinct_category_predicate_saved := Void
			distinct_relationship_predicate_saved := Void
			mapping_reader_saved := Void
			all_mapping_reader_saved := Void
			attribute_types_saved := Void
			relationship_reader_saved := Void
			corelationship_reader_saved := Void
			all_relationship_reader_saved := Void
			all_relationship_reader_saved2 := Void
			all_attributes_predicate_saved := Void
			historical_attributes_predicate_saved := Void
			relationship_reader_saved := Void
			entity_update_predicate_saved := Void
		end

	relationship_update_predicate_saved: COLLECTING_PREDICATE [ENTITY_RELATIONSHIP_RECORD]
	relationship_reader_saved: COLLECTING_PREDICATE [ENTITY_RELATIONSHIP_RECORD]
	all_relationship_reader_saved: COLLECTING_PREDICATE [ENTITY_RELATIONSHIP_RECORD]
	all_relationship_reader_saved2: COLLECTING_PREDICATE [ENTITY_RELATIONSHIP_RECORD]
	corelationship_reader_saved: COLLECTING_PREDICATE [ENTITY_RELATIONSHIP_RECORD]
	entity_predicate_saved: COLLECTING_PREDICATE [ENTITY_ATTRIBUTE_RECORD]
	all_entities_predicate_saved: COLLECTING_PREDICATE [ENTITY_ATTRIBUTE_RECORD]
	all_entities_like_predicate_saved: COLLECTING_PREDICATE [ENTITY_ATTRIBUTE_RECORD]
	all_entity_attribute_predicate_saved: COLLECTING_PREDICATE [ENTITY_ATTRIBUTE_RECORD]
	entity_like_predicate_saved: COLLECTING_PREDICATE [ENTITY_ATTRIBUTE_RECORD]
	every_entity_predicate_saved: COLLECTING_PREDICATE [ENTITY_ATTRIBUTE_RECORD]
	every_entity_like_predicate_saved: COLLECTING_PREDICATE [ENTITY_ATTRIBUTE_RECORD]
	entity_attribute_predicate_saved: COLLECTING_PREDICATE [ENTITY_ATTRIBUTE_RECORD]
	identifier_entity_predicate_saved: COLLECTING_PREDICATE [ENTITY_ATTRIBUTE_RECORD]
	identifier_entity_like_predicate_saved: COLLECTING_PREDICATE [ENTITY_ATTRIBUTE_RECORD]
	key_predicate_saved: COLLECTING_PREDICATE [ENTITY_ATTRIBUTE_RECORD]
	all_key_predicate_saved: COLLECTING_PREDICATE [ENTITY_ATTRIBUTE_RECORD]
	all_attributes_predicate_saved: COLLECTING_PREDICATE [ENTITY_ATTRIBUTE_RECORD]
	attributes_update_predicate_saved: COLLECTING_PREDICATE [ENTITY_ATTRIBUTE_RECORD]
	historical_attributes_predicate_saved: COLLECTING_PREDICATE [ENTITY_FREEZER_RECORD]
	category_entity_predicate_saved: SIMPLE_PREDICATE [ENTITY_ID_BASE]
	all_entity_predicate_saved: SIMPLE_PREDICATE [ENTITY_ID_BASE]
	all_entity_by_category_predicate_saved: COLLECTING_PREDICATE [ENTITY_ID_BASE]
	distinct_category_predicate_saved: INTEGER_PREDICATE
	distinct_relationship_predicate_saved: INTEGER_PREDICATE
	count_attributes_predicate_saved: TUPLE_PREDICATE [TUPLE[attribute_type_id: INTEGER; attribute_count: INTEGER]]
	entity_update_predicate_saved: COLLECTING_PREDICATE [ENTITY_ID_BASE]
	corelationship_counts_saved: HASH_TABLE[INTEGER,INTEGER]
		-- contains the number of distinct entity_1 to a relationship_type_id

	relationship_counts_saved: HASH_TABLE[INTEGER, INTEGER]
		-- contains the number of distinct entity_2 to a relationship_type_id

feature {CONTEXT_DESCRIPTORS} -- Implementation

	set_descriptor (a_descriptor: like descriptor) is
		do
			descriptor := a_descriptor
		end

feature {NONE} -- Implementation

	memory: MEMORY is
			-- MEMORY access
		once
			create Result
		end

	descriptor: CONTEXT_DESCRIPTOR

feature -- Category Support

	category_identifiers (a_category_id: INTEGER): HASH_TABLE [like entity_identifier_template, INTEGER] is
		require
			identifiers_not_void: identifiers_available
		local
			l_identifiers: ROSE_LINEAR [INTEGER]
			l_identifier: like entity_identifier_template
		do
			attribute_types.retrieve_atomic (Void, <<a_category_id>>)
			if not attribute_types.success then
				error.raise ("could not execute predicate")
			end
			create Result.make (attribute_types.array.count)
			l_identifiers := attribute_types.array
			from
				l_identifiers.start
			until
				l_identifiers.off
			loop
				l_identifier := identifier (l_identifiers.item)
				Result.put (l_identifier, l_identifiers.item)
				l_identifiers.forth
			end
		ensure
			not_void: Result /= Void
		end

feature -- process manager

	process_manager: SUPPLIER_PROCESS_ID_MANAGER is
		do
			if process_manager_saved = Void then
				create process_manager_saved.make (server)
			end
			Result := process_manager_saved
		end

	process_manager_saved: like process_manager
			-- Cache for process manager

feature --mappings


	entity_attribute_single_entity_history_predicate: COLLECTING_PREDICATE [ENTITY_ATTRIBUTE_RECORD] is
			-- Predicate for entity attribute single entity history
		once
			create Result.make_strict (server, "select * from entity_attribute where category_id=? and entity_id=?")
		ensure
			valid_result: Result /= Void
		end

	entity_relationship_single_entity_history_predicate: COLLECTING_PREDICATE [ENTITY_RELATIONSHIP_ROW] is
		once
			create Result.make_strict (server, "select * from entity_relationship where entity_1_id=?")
		ensure
			valid_result: Result /= Void
		end

	entity_id_single_entity_history_predicate: COLLECTING_PREDICATE [SUPPLIER_ENTITY_ID_ROW] is
			-- Entity id single entity history predicate
		once
			create Result.make_strict (server, "select * from entity_id where category_id=? and entity_id=?")
		ensure
			valid_result: Result /= Void
		end

	entity_id_map_single_entity_history_predicate: COLLECTING_PREDICATE [ENTITY_ID_MAP_ROW] is
			-- Entity id map single entity history predicate
		once
			create Result.make_strict (server, "select * from entity_id_map where category_id=? and entity_id=?")
		ensure
			valid_result: Result /= Void
		end

feature -- Meta data

	category_counts: HASH_TABLE [INTEGER, STRING] is
			-- Category counts
		require
			valid: is_valid
		local
			l_predicate: COLLECTING_PREDICATE [PAIR [INTEGER, INTEGER]]
		do
			create l_predicate.make_strict (server, "select category_id as first, count(*) as second from entity_id group by category_id")
			l_predicate.retrieve_atomic (Void,Void)
			l_predicate.array.sort_by (agent ordered_category_names)
			create Result.make (100)
			from l_predicate.array.start until l_predicate.array.off loop
				Result.force (l_predicate.array.current_item.second, category_name (l_predicate.array.current_item.first))
				l_predicate.array.forth
			end
		ensure
			valid_result: Result /= Void
		end

	category_counts_from_id: HASH_TABLE [INTEGER, INTEGER] is
			-- category counts from id
		require
			valid: is_valid
		local
			l_predicate: COLLECTING_PREDICATE [PAIR [INTEGER, INTEGER]]
		do
			create l_predicate.make_strict (server, "select category_id as first, count(*) as second from entity_id group by category_id")
			l_predicate.retrieve_atomic (Void,Void)
			create Result.make (100)
			from l_predicate.array.start until l_predicate.array.off loop
				Result.force (l_predicate.array.current_item.second, l_predicate.array.current_item.first)
				l_predicate.array.forth
			end
		ensure
			valid_result: Result /= Void
		end

	relationship_counts: HASH_TABLE [INTEGER, STRING] is
			-- Relationship counts
		require
			valid: is_valid
		local
			l_predicate: COLLECTING_PREDICATE [PAIR [INTEGER, INTEGER]]
		do
			create l_predicate.make_strict (server, "select relationship_type_id as first, count(*) as second from entity_relationship group by relationship_type_id")
			l_predicate.retrieve_atomic (Void,Void)
			l_predicate.array.sort_by (agent ordered_relationship_names)
			create Result.make (100)
			from l_predicate.array.start until l_predicate.array.off loop
				Result.force (l_predicate.array.current_item.second, relationship_name (l_predicate.array.current_item.first))
				l_predicate.array.forth
			end
		ensure
			valid_result: Result /= Void
		end

	relationship_counts_from_id: HASH_TABLE [INTEGER, INTEGER] is
			-- Relationship counts from ids
		require
			valid: is_valid
		local
			l_predicate: COLLECTING_PREDICATE [PAIR [INTEGER, INTEGER]]
		do
			create l_predicate.make_strict (server, "select relationship_type_id as first, count(*) as second from entity_relationship group by relationship_type_id")
			l_predicate.retrieve_atomic (Void,Void)
			create Result.make (100)
			from l_predicate.array.start until l_predicate.array.off loop
				Result.force (l_predicate.array.current_item.second, l_predicate.array.current_item.first)
				l_predicate.array.forth
			end
		ensure
			valid_result: Result /= Void
		end

	relationship_ratios: HASH_TABLE [PAIR [DOUBLE,DOUBLE], STRING] is
			-- Relationship ratios
		require
			valid: is_valid
		local
			l_predicate: COLLECTING_PREDICATE [TRIPLE [INTEGER, DOUBLE, DOUBLE]]
			l_relationship_name: STRING
			l_ratios: PAIR [DOUBLE,DOUBLE]
		do
			create l_predicate.make_strict (server, "select c.relationship_type_id as first,ratio_1 as second,ratio_2 as third from " +
											"(select avg(cast(cnt as float)) ratio_1,relationship_type_id from " +
											"(select count(*) as cnt, relationship_type_id,entity_1_id from entity_relationship group by " +
											"entity_1_id, relationship_type_id) a " +
											"group by relationship_type_id) c " +
											"inner join " +
											"(select avg(cast(cnt as float)) as ratio_2,relationship_type_id from " +
											"(select count(*) as cnt, relationship_type_id,entity_2_id from entity_relationship group by " +
											"entity_2_id, relationship_type_id) b " +
											"group by relationship_type_id) d " +
											"on " +
											"c.relationship_type_id = d.relationship_type_id " +
											"order by c.relationship_type_id")
			l_predicate.retrieve_atomic (Void,Void)
			create Result.make (100)
			from l_predicate.array.start until l_predicate.array.off loop
				l_relationship_name := relationship_name (l_predicate.array.current_item.first)
				create l_ratios.make (l_predicate.array.current_item.third, l_predicate.array.current_item.second)
				Result.force (l_ratios, l_relationship_name)
				l_predicate.array.forth
			end
		ensure
			valid_result: Result /= Void
		end

	mappings_counts: HASH_TABLE [HASH_TABLE [INTEGER,STRING], STRING] is
			-- Mapping counts
		require
			valid: is_valid
		do
			create Result.make (100)
		ensure
			valid_result: Result /= Void
		end

feature {NONE} -- Sorting agents

	ordered_category_names (a, b: PAIR [INTEGER,INTEGER]): BOOLEAN is
			-- Is the category name for 'a' less than 'b'
		do
			if category_name (a.first) /= Void and category_name (b.first) /= Void then
				Result := category_name (a.first) < category_name (b.first)
			else
				Result := category_name (a.first) /= Void
			end
		end

	ordered_relationship_names (a, b: PAIR [INTEGER,INTEGER]): BOOLEAN is
			-- Is the relationship name for 'a' less than 'b'
		do
			if relationship_name (a.first) /= Void and relationship_name (b.first) /= Void then
				Result := relationship_name (a.first) < relationship_name (b.first)
			else
				Result := relationship_name (a.first) /= Void
			end
		end

feature {RELATIONSHIP_HANDLER} -- Sorting agents

	compare_rel_records_for_entity_2 (a_first_rec, a_second_rec: ENTITY_RELATIONSHIP_RECORD): BOOLEAN is
			-- compare first relationship_type_id and entity_2_id
		do
			if a_first_rec = Void then
				Result := False
			elseif a_second_rec = Void then
				Result := True
			elseif a_first_rec.relationship_type_id < a_second_rec.relationship_type_id then
				Result := True
			elseif a_first_rec.relationship_type_id = a_second_rec.relationship_type_id and then a_first_rec.entity_2_id < a_second_rec.entity_2_id then
				Result := True
			else
				Result := False
			end
		end

end
