note

	description:

		"Supplier entity."

	author: "David Stevens"
	last_editor: "$Author: ext-asurelia $"
	date: "$Date: 2014/02/27 $"
	revision: "$Revision: #197 $"
	path: "$File: //rose/source/kernel/supplier_data/supplier_entity.e $"

deferred class
	SUPPLIER_ENTITY

inherit
	ROSE_CONSTANTS
		undefine
			is_equal
		end
	SUPPLIER_ENTITY_ATTRIBUTE_BASE
		export
			{ANY}
		 	category_id,
		 	end_date,
		 	entity_id,
		 	start_date,
		 	update_date,
		 	set_start_date,
		 	set_end_date
			{SUPPLIER_ENTITY,
			ENTITY_SERVER,
			SUPPLIER_ENTITY_CACHE,
			SUPPLIER_ENTITY_RELATIONSHIP_CACHE,
			SUPPLIER_ENTITY_ATTRIBUTE}
			all
		select
			is_equal
		end
	HASHABLE
		undefine
			is_equal
		end
	SUPPLIER_ENTITY_CONSTANTS
		rename
			is_equal as sec_is_equal
		export
			{NONE} all
		end
	ROSE_ID_USER
		undefine
			is_equal
		end

feature -- Initialization

	make (a_category_id: INTEGER; a_start_date: ROSE_DATE) is
			-- Create a new entity of 'a_category_id' on 'a_start_date'.
		require
			a_category_id_not_zero: a_category_id /= 0
			a_start_date_not_void: a_start_date /= Void
		do
			make_empty
			category_id := a_category_id
			create start_date.make_by_date (a_start_date)
			create end_date.make_max_date
			insert(a_start_date)
			as_of_date := as_of_date_from_end_date
		ensure
			valid_entity: valid
		end

	make_from_attribute (a_category_id: INTEGER;
						 a_attribute_type_id: INTEGER;
						 a_attribute_value: STRING;
						 a_start_date: ROSE_DATE) is
			-- Create an entity of 'a_category_id' existing as of 'a_start_date' with unique 'a_attribute_value'.
			-- This routine confuses the idea of 'start_date' and 'as_of_date'.
			-- Use 'make_from_attribute_calling' instead.
		require
			a_category_id_not_zero: a_category_id /= 0
			a_attribute_type_id_not_zero: a_attribute_type_id /= 0
			a_attribute_value_not_void: a_attribute_value /= Void
			a_start_date_not_void: a_start_date /= Void
		do
			make_from_attribute_calling
			(a_category_id, a_attribute_type_id, a_attribute_value, a_start_date, a_start_date, Void)
		end

	make_from_attribute_calling (a_category_id: INTEGER;
								 a_attribute_type_id: INTEGER;
								 a_attribute_value: STRING;
								 a_start_date: ROSE_DATE;
								 a_as_of_date: ROSE_DATE;
								 a_agent: FUNCTION[ANY, TUPLE[SUPPLIER_ENTITY,
															  INTEGER,
															  INTEGER,
															  STRING,
															  ROSE_DATE,
															  ROSE_DATE],
															  BOOLEAN]) is
			-- Create an entity of 'a_category_id' with unique 'a_attribute_value'.
			-- If 'a_as_of_date' is not Void then retreive the entity that exists as of 'a_as_of_date'.
			-- If 'a_as_of_date' is Void then the most recent entity is retreived.
			-- If no entity is found then one is created from scratch with 'a_start_date' and the
			-- attribute is added.
			-- If 'a_agent' is not Void the agent will be called exactly once :
			--	If a matching entity is found, that entity is passed to 'a_agent'.
			--	If no entity is found, an invalid entity is passed to 'a_agent' (you must test by
			--	calling 'valid' or 'invalid').
			--	Your agent can do three things:
			--		1.  Return true accepting either the entity that was passed or the one that
			--		    the agent built.  The agent must add the attribute if desired.
			--		2.  Return false, this routine will create a new entity, add it to the database
			--		    and add the attribute.
			--		3.  Return true but the entity is invalid.  This is illegal.
			--	Notice that the agent may modify the passed entity.
		require
			a_category_id_not_zero: a_category_id /= 0
			a_attribute_type_id_not_zero: a_attribute_type_id /= 0
			a_attribute_value_not_void: a_attribute_value /= Void
			a_start_date_not_void: a_start_date /= Void
		local
			l_as_of_date: ROSE_DATE
			l_attribute: SUPPLIER_ENTITY_ATTRIBUTE[like Current]
			--  l_tuple: TUPLE[SUPPLIER_ENTITY,
			--				INTEGER,
			--				INTEGER,
			--				STRING,
			--				ROSE_DATE,
			--				ROSE_DATE]
		do
			if a_as_of_date=Void
			 then
				-- Void is no longer allowed.  It never worked.
				l_as_of_date := constants.Maximum_date_time
			else
				l_as_of_date := a_as_of_date
			end

			-- Get a potential entity.  If none is found then the entity is left in the 'invalid' state.
			if category_and_attribute_are_cached (a_category_id, a_attribute_type_id)
			 then
				l_attribute := entity_cache.attribute_by_value
				(a_category_id, a_attribute_type_id, a_attribute_value, l_as_of_date)

				if l_attribute /= Void then
					make_from_database (l_attribute.category_id, l_attribute.entity_id, l_attribute.start_date)
				end
			else
				attempt_set_from_database_attribute (a_category_id, a_attribute_type_id, a_attribute_value, l_as_of_date)
			end

			-- Pass this entity (valid or invalid) to the agent.
			if a_agent/=Void
			 then
				--create l_tuple.make
				--l_tuple.put(Current, 1)
				--l_tuple.put(a_category_id, 2)
				--l_tuple.put(a_attribute_type_id, 3)
				--l_tuple.put(a_attribute_value, 4)
				--l_tuple.put(a_start_date, 5)
				--l_tuple.put(l_as_of_date, 6)

				---			if a_agent.item(l_tuple)
				if a_agent.item([Current, a_category_id, a_attribute_type_id, a_attribute_value, a_start_date, l_as_of_date])
				 then
					-- The agent accepted the entity.
					if invalid
					 then
						report_error ("make_from_attribute_calling", "The agent accepted an invalid entity.")
					end
				else
					-- The agent rejected the entity.
					set_invalid		-- To force it to get made fresh.
				end
			end

			-- Build a new one if an existing one couldn't be found.
			if invalid
			 then
				make (a_category_id, a_start_date)
				add_attribute (a_attribute_type_id, a_attribute_value, a_start_date)
			end

			as_of_date := as_of_date_from_end_date
		ensure
			valid_entity: valid
		end

	make_from_attribute_set_calling (a_category_id: INTEGER;
									 a_attribute_set: ARRAY[TUPLE[INTEGER, STRING]];
									 a_start_date: ROSE_DATE;
									 a_as_of_date: ROSE_DATE;
									 a_agent: FUNCTION[ANY,
													   TUPLE[ARRAY[ARRAY[BOOLEAN]],
															 ARRAY[SUPPLIER_ENTITY]],
															 INTEGER]) is
			-- Create an entity of 'a_category_id' by trying to find matches in 'a_attribute_set'.
			-- If 'a_as_of_date' is not Void then retreive the entity that exists as of 'a_as_of_date'.
			-- If 'a_as_of_date' is Void then the most recent entity is retreived.
			-- If no entity is found then one is created from scratch with 'a_start_date'.
			-- If 'a_agent' is not Void the agent will be called exactly once :
			--	If entities are found, those entities are passed to 'a_agent'.
			--	If no entitites are found, a category 0 entity is passed to 'a_agent' (you must test).
			--	Your agent returns an INTEGER to indicate which entity is the one you desire.  If
			--	you return 0, a new entity is created with 'a_start_date' and the attributes added to it.
			--	Notice that the agent may modify the passed entity.  This is useful for the category 0
			--	case where the agent may 'finish up' the entity by calling a different creation routine.
			-- If 'a_agent' is Void then 'most_hits' is called instead.  The first entity with the most hits
			-- is chosen.  If no hits were found then a new entity is created.

			-- The signature of an agent would look like this:
			--	most_hits(a_hits: ARRAY[ARRAY[BOOLEAN]];
			--		  a_entities: ARRAY[WORLDSCOPE_SUPPLIER_ENTITY]): INTEGER is
			-- The 'hits' array indicates whether entity i had a match on attribute j.  The 'a_entities' array
			-- is available for the perusal/modification of your agent.  This is useful for the category 0
			-- case where the agent may 'finish up' the entity by calling a different creation routine.
		require
			a_category_id_not_zero: a_category_id /= 0
			a_attribute_set_not_void: a_attribute_set /= Void
			a_attribute_set_not_empty: a_attribute_set.count /= 0
			--	 a_attributes_are_unique: attribute_type_manager.attribute_types_are_unique(a_attribute_set)
			a_start_date_not_void: a_start_date /= Void
		local
			--l_attribute: SUPPLIER_ENTITY_ATTRIBUTE[like Current]
			l_tuple: TUPLE[SUPPLIER_ENTITY]
		do
			category_id := a_category_id
			create l_tuple
			l_tuple.put(Current, 1)

			--	 if category_and_attribute_are_cached (a_category_id, a_attribute_type_id) then
			--	    l_attribute := entity_cache.attribute_by_value
			--	       (a_category_id, a_attribute_type_id, a_attribute_value, a_as_of_date)
			--	    if l_attribute /= Void then
			--	       make_from_database (l_attribute.category_id, l_attribute.entity_id, l_attribute.start_date)
			--	       if a_agent/=Void and then a_agent.item(l_tuple)=False
			--	       then
			--	          make (a_category_id, a_start_date)
			--	          add_attribute (a_attribute_type_id, a_attribute_value, a_start_date)
			--	       end
			--	    else
			--	       if a_agent/=Void
			--	       then
			--		  category_id := 0
			--		  if a_agent.item(l_tuple)=False
			--	          then
			--	             make (a_category_id, a_start_date)
			--	             add_attribute (a_attribute_type_id, a_attribute_value, a_start_date)
			--		  end
			--	       else
			--	          make (a_category_id, a_start_date)
			--	          add_attribute (a_attribute_type_id, a_attribute_value, a_start_date)
			--	       end
			--	    end
			--	 else
			--	    attempt_set_from_database_attribute (a_category_id, a_attribute_type_id, a_attribute_value, a_as_of_date)
			--	    if (a_agent/=Void and then a_agent.item(l_tuple)=False) or else invalid
			--	    then
			--	       make (a_category_id, a_start_date)
			--	       add_attribute (a_attribute_type_id, a_attribute_value, a_start_date)
			--	    end
			--	 end
			as_of_date := as_of_date_from_end_date
		ensure
			valid_entity: valid
		end

	matching_agent(a_hits: ARRAY[ARRAY[BOOLEAN]];
				   a_entities: ARRAY[SUPPLIER_ENTITY];
				   a_attribute_set:ARRAY[TUPLE[INTEGER, STRING]]): INTEGER is
			-- Find matching entity
		local
			i, j: INTEGER
			l_total: ARRAY[INTEGER]
		do
			from
				i := a_hits.lower
			until
				i>a_hits.upper
			loop
				from
					j := a_hits.item(i).lower
				until
					j>a_hits.item(i).upper
				loop
					if a_hits.item(i).item(j)
					 then
						l_total.put(l_total.item(i) + 1, i)
					end

					j := j + 1
				end

				i := i + 1
			end

			from
				Result := 1
				i := l_total.lower
			until
				i>l_total.upper
			loop
				if l_total.item(i)>l_total.item(Result)
				 then
					Result := i
				end

				i := i + 1
			end

			if l_total.item(Result)=0
			 then
				Result := 0
			end
		end

	make_from_attribute_set (a_category_id: INTEGER;
							 a_attribute_set: ARRAY [TUPLE [INTEGER, STRING]];
							 a_start_date: ROSE_DATE) is
		obsolete "Create a new attribute type that combines the attributes you are looking for (RXRF/ZONE)."
			-- Create an entity of 'a_category_id' existing as of 'a_start_date' with unique 'a_attribute_set'.
			-- Retrieved from database as caching for attribute set access is not implemented.
		require
			a_category_id_not_zero: a_category_id /= 0
			a_attribute_set_not_void: a_attribute_set /= Void
			a_start_date_not_void: a_start_date /= Void
		local
			i: INTEGER
			l_attribute_type_id_ref: INTEGER_REF
			l_attribute_type_id: INTEGER
			l_attribute_value: STRING
		do
			if a_attribute_set.count > Maximum_attribute_set_count then
				report_error ("make_from_attribute_set",
							  "Supported count of attributes in set exceeded, maximum count is " +
							  Maximum_attribute_set_count.out)
			end
			category_id := a_category_id
			attribute_set_array.put (a_start_date, 1)
			attribute_set_array.put (a_start_date.not_max, 2)
			from
				i := 1
			until
				i > Maximum_attribute_set_count
			loop
				if i <= a_attribute_set.count then
					l_attribute_type_id_ref ?= a_attribute_set.item (i).item (1)
					l_attribute_type_id := l_attribute_type_id_ref.item
					l_attribute_value ?= a_attribute_set.item (i).item (2)
				end
				attribute_set_array.put (category_id, (((i - 1) * 3) + 3))
				attribute_set_array.put (l_attribute_type_id, (((i - 1) * 3) + 4))
				attribute_set_array.put (l_attribute_value, (((i - 1) * 3) + 5))
				i := i + 1
			end
			select_attribute_by_value_set_predicate.retrieve_one (Current, Void, attribute_set_array)
			if not select_attribute_by_value_set_predicate.success then
				report_error ("make_from_attribute_set", "Select attribute failed (atttypeid, attvalue) " +
							  l_attribute_type_id.out + " " + l_attribute_value)
			end
			if select_attribute_by_value_set_predicate.index = 0 then
				make (category_id, a_start_date)
				from
					i := 1
				until
					i > a_attribute_set.count
				loop
					l_attribute_type_id_ref ?= a_attribute_set.item (i).item (1)
					l_attribute_type_id := l_attribute_type_id_ref.item
					l_attribute_value ?= a_attribute_set.item (i).item (2)
					add_attribute (l_attribute_type_id, l_attribute_value, a_start_date)
					i := i + 1
				end
			else
				make_from_database (category_id, entity_id, a_start_date)
			end
			as_of_date := as_of_date_from_end_date
		ensure
			valid_entity: valid
		end

	make_from_historical_attribute (a_category_id: INTEGER;
									a_attribute_type_id: INTEGER;
									a_attribute_value: STRING;
									a_start_date: ROSE_DATE) is
			-- Create an entity of 'a_category_id' existing as of 'a_start_date' with unique 'a_attribute_value'.
			-- If it doesn't exist as of 'a_start_date', but does exist historically, then create the historical entity.
			-- If it doesn't exist historically, then create a new entity as of 'a_start_date'.
		require
			a_category_id_not_zero: a_category_id /= 0
			a_attribute_type_id_not_zero: a_attribute_type_id /= 0
			a_attribute_value_not_void: a_attribute_value /= Void
			a_start_date_not_void: a_start_date /= Void
		local
			l_attribute: SUPPLIER_ENTITY_ATTRIBUTE[like Current]
		do
			category_id := a_category_id
			if category_and_attribute_are_cached (a_category_id, a_attribute_type_id) then
				l_attribute := entity_cache.attribute_by_value
				(a_category_id, a_attribute_type_id, a_attribute_value, Void)
				if l_attribute /= Void then
					make_from_database (l_attribute.category_id, l_attribute.entity_id, l_attribute.start_date)
					create as_of_date.make_by_date (l_attribute.end_date)
				else
					set_from_database_historical_attribute (a_attribute_type_id, a_attribute_value, a_start_date)
				end
			else
				set_from_database_historical_attribute (a_attribute_type_id, a_attribute_value, a_start_date)
			end
			as_of_date := as_of_date_from_end_date
		ensure
			valid_entity: valid
		end

	make_from_database (a_category_id: INTEGER; a_entity_id: INTEGER; a_as_of_date: ROSE_DATE) is
			-- Create an entity of 'a_category_id' and 'a_entity_id'.
		require
			a_category_id_not_zero: a_category_id /= 0
			a_entity_id_not_zero: a_entity_id /= 0
			--a_as_of_date_not_void: a_as_of_date /= Void
		local
			l_entity: like Current
		do
			if a_category_id /= constants.maximum_integer and then category_is_cached (a_category_id) then
				l_entity := entity_cache.entity (a_category_id, a_entity_id, a_as_of_date)
				if l_entity /= Void then
					make_from_entity (l_entity)
				else
					report_error ("make_from_database", "Select entity from entity cache failed ..: " + data_server.database + " " + a_category_id.out + " " + a_entity_id.out)
				end
			else
				make_empty
				if a_category_id /= constants.maximum_integer then
					select_predicate.retrieve_one (Current, Void, <<a_category_id, a_entity_id>>)
					if ((not select_predicate.success) or (select_predicate.index = 0)) then
						report_error ("make_from_database", "Select entity failed ..: " + data_server.database + " " + a_category_id.out + " " + a_entity_id.out)
					end
				else
					select_entity_predicate.retrieve_one (Current, Void, <<a_entity_id>>)
					if ((not select_entity_predicate.success) or (select_entity_predicate.index = 0)) then
						report_error ("make_from_database", "Select entity failed ..: " + data_server.database + " " + a_category_id.out + " " + a_entity_id.out)
					end
				end
			end
			as_of_date := as_of_date_from_end_date
		ensure
			valid_entity: valid
		end

	make_start (a_category_id: INTEGER) is
			-- Create the first entity of this category (according to entity_id).
		require
			a_category_id_not_zero: a_category_id /= 0
		local
			l_entity: like Current
		do
			if category_is_cached (a_category_id) and then not entity_cache.updated then
				l_entity := entity_cache.entity_start (a_category_id)
				make_from_entity (l_entity)
			else
				make_empty
				category_id := a_category_id
				select_next_predicate.retrieve_one (Current, Void, <<a_category_id, 0>>)
				if ((not select_next_predicate.success) or (select_next_predicate.index = 0)) then
					report_error ("make_start", "Select entity failed")
				end
			end
			as_of_date := as_of_date_from_end_date
		ensure
			valid_entity: valid
		end

	attempt_make_from_attribute (a_category_id: INTEGER;
								 a_attribute_type_id: INTEGER;
								 a_attribute_value: STRING;
								 a_as_of_date: ROSE_DATE) is
			-- Create an entity of 'a_category_id' existing as of 'a_as_of_date' with unique 'a_attribute_value' if
			-- it exists.  Otherwise set Current to not valid.
		require
			a_category_id_not_zero: a_category_id /= 0
			a_attribute_type_id_not_zero: a_attribute_type_id /= 0
			a_attribute_value_not_void: a_attribute_value /= Void
			--a_as_of_date_not_void: a_as_of_date /= Void
		local
			l_as_of_date: ROSE_DATE
			l_attribute: SUPPLIER_ENTITY_ATTRIBUTE[like Current]
		do
			if a_as_of_date=Void
			 then
				-- Allowing a Void value to persist won't work and so, it is overwritten with the maximum date.
				l_as_of_date := constants.Maximum_date_time
			else
				l_as_of_date := a_as_of_date
			end

			--		 category_id := a_category_id
			if category_and_attribute_are_cached (a_category_id, a_attribute_type_id) then
				l_attribute := entity_cache.attribute_by_value
				(a_category_id, a_attribute_type_id, a_attribute_value, l_as_of_date)
				if l_attribute /= Void then
					make_from_database (l_attribute.category_id, l_attribute.entity_id, l_attribute.start_date)
				else
					set_invalid
				end
			else
				attempt_set_from_database_attribute (a_category_id, a_attribute_type_id, a_attribute_value, l_as_of_date)
			end
			as_of_date := as_of_date_from_end_date
		ensure
			category_id_zero_implies_invalid: category_id = 0 implies invalid
		end

feature {SUPPLIER_ENTITY_CACHE, SUPPLIER_ENTITY_RELATIONSHIP_CACHE} -- Initialization

	make_empty is
			-- Make empty entity
		do
			create as_of_date.make_max_date
			exhausted := False
		ensure
			not_valid: not valid
		end

	wipe_out is
			-- Remove entity
		do
			category_id := 0
			end_date := Void
			entity_id := 0
			start_date := Void
			update_date := Void
			as_of_date := Void
			exhausted := False
		end

feature -- Access
	supplier_product_id: INTEGER is
			-- Supplier product database id.
		deferred
		end

	attribute_by_id (a_attribute_type_id: INTEGER): SUPPLIER_ENTITY_ATTRIBUTE[like Current] is
			-- Return attribute for a particular attribute type id.
		require
			valid_entity: valid
			a_attribute_type_id_not_zero: a_attribute_type_id /= 0
		do
			if category_and_attribute_are_cached (category_id, a_attribute_type_id) then
				Result := entity_cache.attribute_by_id (Current, a_attribute_type_id, as_of_date)
			else
				Result := attribute_from_database (a_attribute_type_id, as_of_date)
			end
		end

	attributes_between (a_attribute_type_id : INTEGER; a_start_date, a_end_date : ROSE_DATE):
		ROSE_LINEAR_ARRAY[SUPPLIER_ENTITY_ATTRIBUTE[like Current]] is
			-- Get all attributes which were ever alive in the range [a_start_date, a_end_date).
		local
			l_container: ROSE_LINEAR_ARRAY[SUPPLIER_ENTITY_ATTRIBUTE[like Current]]
		do
			if category_and_attribute_are_cached(category_id, a_attribute_type_id)
			then
				Result := entity_cache.attributes_between(Current, a_attribute_type_id, a_start_date, a_end_date)
			else
				create l_container.make_empty
				select_attributes_between_predicate.set_container(l_container)
				select_attributes_between_predicate.retrieve(
															 Current, <<a_attribute_type_id, a_start_date, a_end_date>>)
				if not select_attributes_between_predicate.success
				 then
					report_error("attributes_between", "select_attributes_between_predicate failed")
				end

				if l_container.count/=0
				 then
					Result := l_container
				end
			end
		end

	attributes: ARRAYED_LIST [SUPPLIER_ENTITY_ATTRIBUTE[like Current]] is
			-- Return list of all attributes that match the Current one.
		require
			valid_entity: valid
		local
			l_attribute_type_ids: ARRAYED_LIST [INTEGER]
			l_attribute: SUPPLIER_ENTITY_ATTRIBUTE[like Current]
		do
			l_attribute_type_ids := attribute_type_manager.attribute_type_ids
			create Result.make (0)
			from
				l_attribute_type_ids.start
			until
				l_attribute_type_ids.exhausted
			loop
				if category_and_attribute_are_cached (category_id, l_attribute_type_ids.item) then
					l_attribute := entity_cache.attribute_by_id (Current, l_attribute_type_ids.item, as_of_date)
				else
					l_attribute := attribute_from_database (l_attribute_type_ids.item, as_of_date)
				end
				if l_attribute /= Void then
					Result.extend (l_attribute)
				end
				l_attribute_type_ids.forth
			end
		end

	attribute_history (a_attribute_type_id: INTEGER): EXTENDABLE_BOUNDED_ARRAY [SUPPLIER_ENTITY_ATTRIBUTE[like Current]] is
			-- Return history of attribute changes for a particular type id.
		require
			valid_entity: valid
			a_attribute_type_id_not_zero: a_attribute_type_id /= 0
		do
			if category_and_attribute_are_cached (category_id, a_attribute_type_id) then
				Result := entity_cache.attribute_history (Current, a_attribute_type_id)
			else
				Result := attribute_history_from_database (a_attribute_type_id)
			end
		end

	most_recent_attribute(a_type: INTEGER): SUPPLIER_ENTITY_ATTRIBUTE[like Current]
			-- Fetches the most recent attribute by type
		local
			l_hist: EXTENDABLE_BOUNDED_ARRAY [SUPPLIER_ENTITY_ATTRIBUTE[like Current]]
			l_list: ROSE_LINEAR_ARRAY [SUPPLIER_ENTITY_ATTRIBUTE[like Current]]
		do
			if a_type > 0 then
				l_hist := attribute_history (a_type)
				if l_hist /= Void then
					create l_list.make_from_array (l_hist)
					l_list.sort_by (agent compare_attributes_by_start_date)
					Result := l_list.last
				end
			end
		end

	first_attribute_set (a_relationship_attribute_type_tuple_set: ARRAY [TUPLE [INTEGER, INTEGER]]):
			-- Return the first attribute in a list of attributes.
	ARRAY [SUPPLIER_ENTITY_ATTRIBUTE[like Current]] is
		require
			valid_entity: valid
			a_relationship_attribute_type_tuple_set_not_void: a_relationship_attribute_type_tuple_set /= Void
		do
			set_as_of_date (constants.Minimum_date_time)
			Result := next_attribute_set (a_relationship_attribute_type_tuple_set)
		end

	next_attribute_set (a_relationship_attribute_type_tuple_set: ARRAY [TUPLE [INTEGER, INTEGER]]):
			-- Returns the next attribute in the list.
	ARRAY [SUPPLIER_ENTITY_ATTRIBUTE[like Current]] is
		require
			valid_entity: valid
			a_relationship_attribute_type_tuple_set_not_void: a_relationship_attribute_type_tuple_set /= Void
		local
			i: INTEGER
			l_min_entry_date: ROSE_DATE
			l_next_min_entry_date: ROSE_DATE
			l_min_exit_date: ROSE_DATE
			l_attribute: SUPPLIER_ENTITY_ATTRIBUTE[like Current]
			l_attribute_set: ARRAY [SUPPLIER_ENTITY_ATTRIBUTE[like Current]]
			l_relationship_type_id_ref: INTEGER_REF
			l_relationship_type_id: INTEGER
			l_attribute_type_id_ref: INTEGER_REF
			l_attribute_type_id: INTEGER
		do
			-- Get all attributes for this set which are not terminated before as_of_date.
			create l_attribute_set.make
			(a_relationship_attribute_type_tuple_set.lower, a_relationship_attribute_type_tuple_set.upper)
			from
				i := a_relationship_attribute_type_tuple_set.lower
			until
				i > a_relationship_attribute_type_tuple_set.upper
			loop
				l_relationship_type_id_ref ?= a_relationship_attribute_type_tuple_set.item (i).item (1)
				l_relationship_type_id := l_relationship_type_id_ref.item
				l_attribute_type_id_ref ?= a_relationship_attribute_type_tuple_set.item (i).item (2)
				l_attribute_type_id := l_attribute_type_id_ref.item
				l_attribute := next_attribute (l_relationship_type_id, l_attribute_type_id)
				if l_attribute /= Void then
					l_attribute_set.put (l_attribute, i)
				end
				i := i + 1
			end

			-- Determine the earliest date that a new attribute starts.
			create l_min_entry_date.make_max_date
			from
				i := l_attribute_set.lower
			until
				i > l_attribute_set.upper
			loop
				if l_attribute_set.item (i) /= Void then
					if l_attribute_set.item (i).start_date > as_of_date then
						if l_attribute_set.item (i).start_date < l_min_entry_date then
							l_min_entry_date := l_attribute_set.item (i).start_date
						end
					end
				end
				i := i + 1
			end

			-- Determine the next earliest date that a new attribute starts.
			create l_next_min_entry_date.make_max_date
			from
				i := l_attribute_set.lower
			until
				i > l_attribute_set.upper
			loop
				if l_attribute_set.item (i) /= Void then
					if l_attribute_set.item (i).start_date.date > l_min_entry_date.date then
						if l_attribute_set.item (i).start_date < l_next_min_entry_date then
							l_next_min_entry_date := l_attribute_set.item (i).start_date
						end
					end
				end
				i := i + 1
			end

			-- Determine the earliest date that an existing attribute ends.
			create l_min_exit_date.make_max_date
			from
				i := l_attribute_set.lower
			until
				i > l_attribute_set.upper
			loop
				if l_attribute_set.item (i) /= Void then
					if l_attribute_set.item (i).end_date > as_of_date then
						if l_attribute_set.item (i).end_date < l_min_exit_date then
							l_min_exit_date := l_attribute_set.item (i).end_date
						end
					end
				end
				i := i + 1
			end

			-- Set the new as_of_date.
			if l_next_min_entry_date < l_min_exit_date then
				if l_next_min_entry_date > as_of_date then
					set_as_of_date (l_next_min_entry_date)
				else
					set_as_of_date (l_min_exit_date)
				end
			else
				if as_of_date > constants.Minimum_date_time and then
					l_min_entry_date < constants.Maximum_date_time and then
					l_min_entry_date > as_of_date then
					set_as_of_date (l_min_entry_date)
				else
					set_as_of_date (l_min_exit_date)
				end
			end

			-- Fill the result.
			create Result.make (l_attribute_set.lower, l_attribute_set.upper)
			from
				i := l_attribute_set.lower
			until
				i > l_attribute_set.upper
			loop
				if l_attribute_set.item (i) /= Void and then
					l_attribute_set.item (i).start_date <= as_of_date and then
					l_attribute_set.item (i).end_date >= as_of_date then
					Result.put (l_attribute_set.item (i), i)
				end
				i := i + 1
			end
		end

	relatives (a_relationship_type_id: INTEGER): ARRAYED_LIST [like Current] is
			-- Returns all 'relationship_type_id' relatives of this entity.
		require
			valid_entity: valid
			a_relationship_type_id_not_zero: a_relationship_type_id /= 0
		local
			l_relationship_type: SUPPLIER_RELATIONSHIP_TYPE
		do
			l_relationship_type := relationship_type_manager.relationship_type (a_relationship_type_id)

			-- It is important to test category_1 first so this will match like-to-like relationships.
			if category_id = l_relationship_type.category_1_id then
				Result := directed_relatives(l_relationship_type, Right_relative)
			elseif category_id = l_relationship_type.category_2_id then
				Result := directed_relatives(l_relationship_type, Left_relative)
			else
				report_error ("relatives", "Invalid relationship_type_id for this entity")
			end
		end

	historical_relatives (a_relationship_type_id: INTEGER): ARRAYED_LIST [like Current] is
			-- Returns all 'a_relationship_type_id' relatives that historically existed for this entity.
		require
			valid_entity: valid
			a_relationship_type_id_not_zero: a_relationship_type_id /= 0
		do
			historical_relative_lookup := True
			Result := relatives (a_relationship_type_id)
			historical_relative_lookup := False
		end

	right_relatives (a_relationship_type_id: INTEGER): ARRAYED_LIST [like Current] is
			-- Returns the relatives who are listed in entity_2_id of the entity_relationship table.
			-- This is intended for like-to-like (sibling) relationships, but can also be called when we know the relationship direction.
			-- An example of a like-to-like relationship is company-to-parent_company.  Since both categories
			-- are the same, you can not use 'relatives'.
		local
			l_relationship_type: SUPPLIER_RELATIONSHIP_TYPE
		do
			l_relationship_type := relationship_type_manager.relationship_type (a_relationship_type_id)
			Result := directed_relatives(l_relationship_type, Right_relative)
		end

	historical_right_relatives (a_relationship_type_id: INTEGER): ARRAYED_LIST [like Current] is
			-- Returns the relatives who are listed in entity_2_id of the entity_relationship table.
			-- This is intended for like-to-like (sibling) relationships, but can also be called when we know the relationship direction.
			-- An example of a like-to-like relationship is company-to-parent_company.  Since both categories
			-- are the same, you can not use 'relatives'.
		local
			l_relationship_type: SUPPLIER_RELATIONSHIP_TYPE
		do
			historical_relative_lookup := True
			l_relationship_type := relationship_type_manager.relationship_type (a_relationship_type_id)
			Result := directed_relatives(l_relationship_type, Right_relative)
			historical_relative_lookup := False
		end

	left_relatives (a_relationship_type_id: INTEGER): ARRAYED_LIST [like Current] is
			-- Returns the relatives who are listed in entity_1_id of the entity_relationship table.
			-- This is intended for like-to-like (sibling) relationships, but can also be called when we know the relationship direction.
			-- An example of a like-to-like relationship is company-to-parent_company.  Since both categories
			-- are the same, you can not use 'relatives'.
		local
			l_relationship_type: SUPPLIER_RELATIONSHIP_TYPE
		do
			l_relationship_type := relationship_type_manager.relationship_type (a_relationship_type_id)
			Result := directed_relatives(l_relationship_type, Left_relative)
		end

	historical_left_relatives (a_relationship_type_id: INTEGER): ARRAYED_LIST [like Current] is
			-- Returns the relatives who are listed in entity_1_id of the entity_relationship table.
			-- This is intended for like-to-like (sibling) relationships, but can also be called when we know the relationship direction.
			-- An example of a like-to-like relationship is company-to-parent_company.  Since both categories
			-- are the same, you can not use 'relatives'.
		local
			l_relationship_type: SUPPLIER_RELATIONSHIP_TYPE
		do
			historical_relative_lookup := True
			l_relationship_type := relationship_type_manager.relationship_type (a_relationship_type_id)
			Result := directed_relatives(l_relationship_type, Left_relative)
			historical_relative_lookup := False
		end

	relatives_with_attribute (a_relationship_type_id: INTEGER;
							  a_attribute_type_id: INTEGER;
							  a_attribute_value: STRING): ARRAYED_LIST [like Current] is
			-- Same as 'relatives' except that only those entities with the matching attribute will be returned.
		require
			valid_entity: valid
			a_relationship_type_id_not_zero: a_relationship_type_id /= 0
			a_attribute_type_id_not_zero: a_attribute_type_id /= 0
			a_attribute_value_not_void: a_attribute_value /= Void
		local
			l_relationship_type: SUPPLIER_RELATIONSHIP_TYPE
		do
			l_relationship_type := relationship_type_manager.relationship_type (a_relationship_type_id)

			-- It is important to test category_1 first so this will match like-to-like relationships.
			if category_id = l_relationship_type.category_1_id then
				Result := directed_relatives_with_attribute(l_relationship_type,
															a_attribute_type_id, a_attribute_value, Right_relative)
			elseif category_id = l_relationship_type.category_2_id then
				Result := directed_relatives_with_attribute(l_relationship_type,
															a_attribute_type_id, a_attribute_value, Left_relative)
			else
				report_error ("relatives_with_attribute", "Invalid relationship_type_id for this entity")
			end
		end

	right_relatives_with_attribute (a_relationship_type_id: INTEGER;
									a_attribute_type_id: INTEGER;
									a_attribute_value: STRING): ARRAYED_LIST [like Current] is
			-- Same as 'right_relatives' except that only those entities with the matching attribute will be returned.
		require
			valid_entity: valid
			a_relationship_type_id_not_zero: a_relationship_type_id /= 0
			a_attribute_type_id_not_zero: a_attribute_type_id /= 0
			a_attribute_value_not_void: a_attribute_value /= Void
		local
			l_relationship_type: SUPPLIER_RELATIONSHIP_TYPE
		do
			l_relationship_type := relationship_type_manager.relationship_type (a_relationship_type_id)
			Result := directed_relatives_with_attribute(l_relationship_type, a_attribute_type_id, a_attribute_value, Right_relative)
		end

	left_relatives_with_attribute (a_relationship_type_id: INTEGER;
								   a_attribute_type_id: INTEGER;
								   a_attribute_value: STRING): ARRAYED_LIST [like Current] is
			-- Same as 'left_relatives' except that only those entities with the matching attribute will be returned.
		require
			valid_entity: valid
			a_relationship_type_id_not_zero: a_relationship_type_id /= 0
			a_attribute_type_id_not_zero: a_attribute_type_id /= 0
			a_attribute_value_not_void: a_attribute_value /= Void
		local
			l_relationship_type: SUPPLIER_RELATIONSHIP_TYPE
		do
			l_relationship_type := relationship_type_manager.relationship_type (a_relationship_type_id)
			Result := directed_relatives_with_attribute(l_relationship_type, a_attribute_type_id, a_attribute_value, Left_relative)
		end

	only_relative (a_relationship_type_id: INTEGER): like Current is
			-- Returns the one and only 'relationship_type_id' relative of this entity.
		require
			valid_entity: valid
			a_relationship_type_id_not_zero: a_relationship_type_id /= 0
		local
			l_relatives: ARRAYED_LIST [like Current]
		do
			l_relatives := relatives (a_relationship_type_id)
			if l_relatives /= Void then
				if l_relatives.count = 1 then
					Result := l_relatives.i_th (1)
				elseif l_relatives.count > 1 then
					report_error ("only_relative", "Multiple relatives when expecting only relative")
				end
			end
		end

	only_right_relative (a_relationship_type_id: INTEGER): like Current is
			-- Returns the one and only relative who is listed in entity_2_id of the entity_relationship table.
			-- This is intended for like-to-like (sibling) relationships, but can also be called when we know the relationship direction.
			-- An example of a like-to-like relationship is company-to-parent_company.  Since both categories
			-- are the same, you can not use 'only_relative'.
		require
			valid_entity: valid
			a_relationship_type_id_not_zero: a_relationship_type_id /= 0
		local
			l_relatives: ARRAYED_LIST [like Current]
		do
			l_relatives := right_relatives(a_relationship_type_id)
			if l_relatives /= Void then
				if l_relatives.count = 1 then
					Result := l_relatives.i_th (1)
				elseif l_relatives.count > 1 then
					report_error ("only_right_relative", "Multiple relatives when expecting only_relatives")
				end
			end
		end

	only_left_relative (a_relationship_type_id: INTEGER): like Current is
			-- Returns the one and only relative who is listed in entity_1_id of the entity_relationship table.
			-- This is intended for like-to-like (sibling) relationships, but can also be called when we know the relationship direction.
			-- An example of a like-to-like relationship is company-to-parent_company.  Since both categories
			-- are the same, you can not use 'only_relative'.
		require
			valid_entity: valid
			a_relationship_type_id_not_zero: a_relationship_type_id /= 0
		local
			l_relatives: ARRAYED_LIST [like Current]
		do
			l_relatives := left_relatives(a_relationship_type_id)
			if l_relatives /= Void then
				if l_relatives.count = 1 then
					Result := l_relatives.i_th (1)
				elseif l_relatives.count > 1 then
					report_error ("only_left_relative", "Multiple relatives when expecting only_relatives")
				end
			end
		end

	relationship_dates (a_relationship_type_id: INTEGER; a_related_entity: like Current):
		ARRAYED_LIST [TUPLE [ROSE_DATE, ROSE_DATE]] is
			-- The start and end dates of the relationships between Current and a related entity.
		require
			valid_entity: valid
			a_relationship_type_id_not_zero: a_relationship_type_id /= 0
			a_related_entity_not_void: a_related_entity /= Void
		local
			l_relationships: EXTENDABLE_BOUNDED_ARRAY [SUPPLIER_ENTITY_RELATIONSHIP]
			i: INTEGER
			l_count: INTEGER
			l_tuple: TUPLE [ROSE_DATE, ROSE_DATE]
			l_relationship_type: SUPPLIER_RELATIONSHIP_TYPE
		do
			l_relationships := relationships (a_relationship_type_id)
			if l_relationships /= Void then
				l_relationship_type := relationship_type_manager.relationship_type (a_relationship_type_id)
				from
					i := 1
				until
					i > l_relationships.count
				loop
					-- It is important to test category_1 first so this will match like-to-like relationships.
					if category_id = l_relationship_type.category_1_id then
						if l_relationship_type.category_2_id = a_related_entity.category_id and then
							l_relationships.item (i).entity_2_id = a_related_entity.entity_id then
							create l_tuple
							l_tuple.put (l_relationships.item (i).start_date, 1)
							l_tuple.put (l_relationships.item (i).end_date, 2)
							if l_count = 0 then
								l_count := l_count + 1
								create Result.make (0)
								Result.extend (l_tuple)
							else
								Result.extend (l_tuple)
							end
						end
					elseif category_id = l_relationship_type.category_2_id then
						if l_relationship_type.category_1_id = a_related_entity.category_id and then
							l_relationships.item (i).entity_1_id = a_related_entity.entity_id then
							create l_tuple
							l_tuple.put (l_relationships.item (i).start_date, 1)
							l_tuple.put (l_relationships.item (i).end_date, 2)
							if l_count = 0 then
								l_count := l_count + 1
								create Result.make (0)
								Result.extend (l_tuple)
							else
								Result.extend (l_tuple)
							end
						end
					else
						report_error ("relationship_dates", "Invalid relationship for this entity")
					end
					i := i + 1
				end
			end
		end

	historical_only_entity_relationship_record( a_relationship_type_id: INTEGER; a_date: ROSE_DATE ): ENTITY_RELATIONSHIP_RECORD is
     		-- Return a mathcing record that had only a historical relationship.
   		require
   			valid_entity: valid
   			a_relationship_type_id_not_zero: a_relationship_type_id /= 0
   		local
   			l_relationships: EXTENDABLE_BOUNDED_ARRAY[ SUPPLIER_ENTITY_RELATIONSHIP ]
   			l_relationship: SUPPLIER_ENTITY_RELATIONSHIP
   		do
   			Result := Void
 	 		l_relationships := historical_relationships( a_relationship_type_id )
	 		if l_relationships /= Void and then l_relationships.count > 0 then
	 			from l_relationships.start
	 			until l_relationships.off or Result /= Void
	 			loop
	 				l_relationship := l_relationships.current_item
	 				if a_date >= l_relationship.start_date and a_date < l_relationship.end_date then
	 					create Result
	 					Result.set_relationship_type_id( a_relationship_type_id )
	 					Result.set_entity_1_id( l_relationship.entity_1_id )
	 					Result.set_entity_2_id( l_relationship.entity_2_id )
	 					Result.set_start_date( l_relationship.start_date )
	 					Result.set_end_date( l_relationship.end_date )
	 					Result.set_update_date( l_relationship.update_date )
	 				end
	 				l_relationships.forth
	 			end
	 		end
   		end

	historical_entity_relationship_record( a_relationship_type_id: INTEGER; a_date: ROSE_DATE ): ROSE_LINEAR_ARRAY[ENTITY_RELATIONSHIP_RECORD] is
     		-- Return a mathcing record that had either a current or a historical relationship.
   		require
   			valid_entity: valid
   			a_relationship_type_id_not_zero: a_relationship_type_id /= 0
   		local
   			l_relationships: EXTENDABLE_BOUNDED_ARRAY[ SUPPLIER_ENTITY_RELATIONSHIP ]
   			l_relationship: SUPPLIER_ENTITY_RELATIONSHIP
   			l_record: ENTITY_RELATIONSHIP_RECORD
   		do
   			Create Result.make_empty
 	 		l_relationships := historical_relationships( a_relationship_type_id )
	 		if l_relationships /= Void and then l_relationships.count > 0 then
	 			from l_relationships.start until l_relationships.off
	 			loop
	 				l_relationship := l_relationships.current_item
	 				if a_date >= l_relationship.start_date and a_date < l_relationship.end_date then
	 					create l_record
	 					l_record.set_relationship_type_id( a_relationship_type_id )
	 					l_record.set_entity_1_id( l_relationship.entity_1_id )
	 					l_record.set_entity_2_id( l_relationship.entity_2_id )
	 					l_record.set_start_date( l_relationship.start_date )
	 					l_record.set_end_date( l_relationship.end_date )
	 					l_record.set_update_date( l_relationship.update_date )
	 					l_record.set_weight( l_relationship.weight )
	 					Result.extend(l_record)
	 				end
	 				l_relationships.forth
	 			end
	 		end
   		end

	relationship_entity(a_relationship_type_id: INTEGER; a_other_entity: like Current): like Current is
			-- Returns the entity that represents the relationship between this entity and a_other_entity.
			-- May return Void.
		require
			valid_entity: valid
			valid_other_entity: a_other_entity.valid
			same_as_of_dates: as_of_date.is_equal(a_other_entity.as_of_date)
			a_relationship_type_id_not_zero: a_relationship_type_id/=0
		local
			l_relationship: SUPPLIER_ENTITY_RELATIONSHIP
		do
			l_relationship := relationship(a_relationship_type_id, a_other_entity)

			if l_relationship/=Void
			 then
				if l_relationship.relationship_id/=0
				 then
					Result := server.new_empty_entity
					Result.make_from_database(ids.relationship_category_id, l_relationship.relationship_id, as_of_date)
				end
			end
		end

	new_relationship_entity(a_relationship_type_id: INTEGER; a_other_entity: like Current): like Current is
			-- Returns the entity that represents the relationship between this entity and a_other_entity.
			-- If no such entity exists then it creates one.
		require
			valid_entity: valid
			valid_other_entity: a_other_entity.valid
			same_as_of_dates: as_of_date.is_equal(a_other_entity.as_of_date)
			a_relationship_type_id_not_zero: a_relationship_type_id/=0
		local
			l_relationship: SUPPLIER_ENTITY_RELATIONSHIP
		do
			l_relationship := relationship(a_relationship_type_id, a_other_entity)

			if l_relationship=Void
			 then
				report_error("new_relationship_entity", "No relationship was found.")
			else
				if l_relationship.relationship_id=0
				 then
					Result := server.new_empty_entity
					Result.make(ids.relationship_category_id, today)

					-- Note that if l_relationship came from the cache then the following statement
					-- modifies the cache.
					l_relationship.set_relationship_id(Result.entity_id)

					update_relationship_id_predicate.execute(l_relationship, Void)
					if not update_relationship_id_predicate.success
					 then
						report_error("new_relationship_entity", "update_relationship_id_predicate failed.")
					end

					if update_relationship_id_predicate.rows_affected/=1
					 then
						report_error("new_relationship_entity", "update_relationship_id_predicate did not update only one row.")
					end
				else
					Result := server.new_empty_entity
					Result.make_from_database(ids.relationship_category_id, l_relationship.relationship_id, as_of_date)
				end

			end
		end

	transition (a_transition_type_id: INTEGER; a_date: ROSE_DATE): ENTITY_TRANSITION is
			-- Returns the transition corresponding to a_transition_type_id at a_date.
		require
			valid_entity: valid
			a_date_not_void: a_date /= Void
			a_transition_type_id_not_zero: a_transition_type_id /= 0
		local
			l_array: EXTENDABLE_BOUNDED_ARRAY [ENTITY_TRANSITION]
		do
			create l_array.make_empty
			select_transition_predicate.set_container (l_array)
			select_transition_predicate.retrieve (Void, <<entity_id, a_transition_type_id, a_date>>)
			if not select_transition_predicate.success then
				report_error ("select_transition", "Sel transition failed (trntypeid) " + a_transition_type_id.out)
			end

			if l_array.count = 1 then
				Result := l_array.item (l_array.lower)
			elseif l_array.count>1 then
				report_error("transition", "select_transition_predicate returned more than one row")
			end
		end

	transition_history(a_transition_type_ids: ARRAY[INTEGER]): ENTITY_TRANSITION_HISTORY is
			-- Return the transition history for this entity.
		require
			valid_entity: valid
		local
			i, j: INTEGER
			l_count: INTEGER
			l_arg_array: ARRAY[ANY]
		do
			if a_transition_type_ids=Void
			 then
				l_count := 0
			else
				l_count := a_transition_type_ids.count
			end

			create l_arg_array.make(1, l_count + 1)
			l_arg_array.put(entity_id, 1)
			from
				i := 1
				j := 2
			until
				i>l_count
			loop
				l_arg_array.put(a_transition_type_ids.item(i), j)
				i := i + 1
				j := j + 1
			end

			create Result.make_empty
			select_transition_history_predicate(l_count).set_container (Result)
			select_transition_history_predicate(l_count).retrieve (Void, l_arg_array)
			if not select_transition_history_predicate(l_count).success then
				report_error ("transition_history", "Select entity transition failed")
			end
		end

	unique_transition(a_transition_type_id: INTEGER): ENTITY_TRANSITION is
			-- Returns the ONLY transition of this type or Void.
		require
			valid_entity: valid
			transition_type_id_not_0: a_transition_type_id/=0
		local
			l_transitions: ENTITY_TRANSITION_HISTORY
		do
			l_transitions := transition_history(<<a_transition_type_id>>)
			if l_transitions=Void or else l_transitions.count=0
			 then
				do_nothing
			else
				if l_transitions.count>1
				 then
					report_error("unique_transition", "More than one unique transition found.")
				else
					Result := l_transitions.item(1)
				end
			end
		end

	company_tree_first_data_available_date: ROSE_DATE is
			-- Return the first available date for a particular company tree entity.
		require
			valid_entity: valid
		local
			l_tree: like company_tree
			l_transition: ENTITY_TRANSITION
		do
			l_tree := historical_company_tree
			l_transition := l_tree.item.unique_transition (ids.first_data_available_type_id)
			if l_transition /= Void then Result := l_transition.transition_date end
			from l_tree.child_start until l_tree.child_off loop
				l_transition := l_tree.child.item.unique_transition (ids.first_data_available_type_id)
				if l_transition /= Void then
					if Result = Void or else Result > l_transition.transition_date then create Result.make_by_date (l_transition.transition_date) end
				end
				from l_tree.child.child_start until l_tree.child.child_off loop
					l_transition := l_tree.child.child.item.unique_transition (ids.first_data_available_type_id)
					if l_transition /= Void then
						if Result = Void or else Result > l_transition.transition_date then create Result.make_by_date (l_transition.transition_date) end
					end
					l_tree.child.child_forth
				end
				l_tree.child_forth
			end
		end

	company_tree_last_data_available_date: ROSE_DATE is
			-- Return the last data record for an entity.
		require
			valid_entity: valid
		local
			l_tree: like company_tree
			l_transition: ENTITY_TRANSITION
		do
			l_tree := historical_company_tree
			l_transition := l_tree.item.unique_transition (ids.last_data_available_type_id)
			if l_transition /= Void then Result := l_transition.transition_date end
			from l_tree.child_start until l_tree.child_off loop
				l_transition := l_tree.child.item.unique_transition (ids.last_data_available_type_id)
				if l_transition /= Void then
					if Result = Void or else Result < l_transition.transition_date then create Result.make_by_date (l_transition.transition_date) end
				end
				from l_tree.child.child_start until l_tree.child.child_off loop
					l_transition := l_tree.child.child.item.unique_transition (ids.last_data_available_type_id)
					if l_transition /= Void then
						if Result = Void or else Result < l_transition.transition_date then create Result.make_by_date (l_transition.transition_date) end
					end
					l_tree.child.child_forth
				end
				l_tree.child_forth
			end
		end

	historical_company_tree: ROSE_ARRAYED_TREE [SUPPLIER_ENTITY] is
			-- A historical company tree given a company, security, or listing.
		require
			a_valid_company_security_or_listing: valid_company_security_or_listing
		do
			historical_relative_lookup := True
			Result := company_tree
			historical_relative_lookup := False
		end

	company_tree: ROSE_ARRAYED_TREE [SUPPLIER_ENTITY] is
			-- A company tree given a company, security, or listing.
		require
			a_valid_company_security_or_listing: valid_company_security_or_listing
		local
			l_company, l_security: like Current
			l_companies, l_securities, l_listings: like relatives
			l_historical_relative_lookup: BOOLEAN
		do
			l_historical_relative_lookup := historical_relative_lookup
			if category_id = ids.company_category_id then
				l_company := Current
			elseif category_id = ids.security_category_id then
				if l_historical_relative_lookup then
					l_companies := historical_relatives (company_security_relationship_type_id)
				else
					l_companies := relatives (company_security_relationship_type_id)
				end
				if l_companies /= Void and then l_companies.count > 0 then l_company := l_companies.i_th (1) end
			elseif category_id = ids.listing_category_id then
				if l_historical_relative_lookup then
					l_securities := historical_relatives (security_listing_relationship_type_id)
				else
					l_securities := relatives (security_listing_relationship_type_id)
				end
				if l_securities /= Void and then l_securities.count > 0 then l_security := l_securities.i_th (1) end
				if l_security /= Void and then l_security.valid then
					if l_historical_relative_lookup then
						l_companies := l_security.historical_relatives (company_security_relationship_type_id)
					else
						l_companies := l_security.relatives (company_security_relationship_type_id)
					end
					if l_companies /= Void and then l_companies.count > 0 then l_company := l_companies.i_th (1) end
				end
			end
			if l_company /= Void and then l_company.valid then
				create Result.make (0, l_company)
				if l_historical_relative_lookup then
					l_securities := l_company.historical_relatives (company_security_relationship_type_id)
				else
					l_securities := l_company.relatives (company_security_relationship_type_id)
				end
				if l_securities /= Void then
					from l_securities.start until l_securities.off loop
						if l_securities.item.valid then
							Result.child_extend (l_securities.item)
							Result.child_move (1)
							if l_historical_relative_lookup then
								l_listings := l_securities.item.historical_relatives (security_listing_relationship_type_id)
							else
								l_listings := l_securities.item.relatives (security_listing_relationship_type_id)
							end
							if l_listings /= Void then
								from l_listings.start until l_listings.off loop
									if l_listings.item.valid then
										Result.child.child_extend (l_listings.item)
										Result.child.child_move (1)
									end
									l_listings.forth
								end
							end
						end
						l_securities.forth
					end
				end
			end
		end

	hash_code: INTEGER is
			-- Return hash code entity_id.
		do
			Result := entity_id
		end

feature {SUPPLIER_ENTITY_CACHE, SUPPLIER_ENTITY_RELATIONSHIP_CACHE} -- Access

	key: KEY_PAIR [INTEGER, INTEGER] is
			-- Return key pair for a category_id and entity_id.
		require
			valid_entity: valid
		do
			create Result.make (category_id, entity_id)
		end

feature -- Cursor movement
	forth is
			-- Move to the next entity (according to entity_id).
		require
			valid_entity: valid
		local
			l_entity: like Current
		do
			if category_is_cached (category_id) and then not entity_cache.updated then
				entity_cache.entity_forth (category_id)
				if entity_cache.entity_exhausted (category_id) then
					exhausted := True
				else
					l_entity := entity_cache.entity_item (category_id)
					make_from_entity (l_entity)
				end
			else
				select_next_predicate.retrieve_one (Current, Void, <<category_id, entity_id>>)
				if not select_next_predicate.success then
					report_error ("make_next", "Select entity failed")
				end
				if select_next_predicate.index = 0 then
					set_invalid
					make_empty
					exhausted := True
				end
			end
		end

feature -- Status report

	valid: BOOLEAN is
			-- Return 'True' if the entity is valid.
		do
			Result := ((category_id /= 0) and then ((as_of_date = Void) or else (as_of_date <= end_date)))
		end

	invalid: BOOLEAN is
			-- Return 'True' if the entity is not valid.
		do
			Result := not valid
		end

	valid_company_security_or_listing: BOOLEAN is
			-- Not generally useful to clients.  Only used by some 'require' statements.
		do
			Result := valid and then
			(category_id = ids.company_category_id or
			 category_id = ids.security_category_id or
			 category_id = ids.listing_category_id)
		end

	as_of_date: ROSE_DATE
			-- Return entity values for a particular date.

	as_of_date_is_current: BOOLEAN is
			-- Return 'True' if entity values for a particular date are also the current entity values.
		do
			Result := as_of_date.is_equal(as_of_date_from_end_date)
		end

	exhausted: BOOLEAN
			-- Flag to indicate if we are at the end of all the entities.

	attribute_set_exhausted (a_attribute_set: ARRAY [SUPPLIER_ENTITY_ATTRIBUTE[like Current]]): BOOLEAN is
			-- Return 'True' if we are at the end of the list of entities.
		require
			valid_entity: valid
			a_attribute_set_not_void: a_attribute_set /= Void
		local
			i: INTEGER
		do
			from
				i := a_attribute_set.lower
				Result := True
			until
				i > a_attribute_set.upper or Result = False
			loop
				if a_attribute_set.item (i) /= Void then
					Result := False
				end
				i := i + 1
			end
		end

	historical_relationships (a_relationship_type_id: INTEGER): EXTENDABLE_BOUNDED_ARRAY[SUPPLIER_ENTITY_RELATIONSHIP] is
			-- Return list of relationships which existed for this entity.
		require
			valid_entity: valid
			a_relationship_type_id_not_zero: a_relationship_type_id /= 0
		do
			historical_relative_lookup := True
			Result := relationships (a_relationship_type_id)
			historical_relative_lookup := False
		end


	entity_cache_loaded: BOOLEAN is
			-- Return if the entity cache is loaded.
		require
			valid_entity: valid
		do
			Result := entity_cache /= Void
		end

	is_equal (other: like Current): BOOLEAN is
			--      require
			--	 valid_entity: valid   ??????
			--	 valid_other_entity: other.valid
		do
			Result := category_id = other.category_id and
				entity_id = other.entity_id and
				equal(start_date, other.start_date) and
				equal(end_date, other.end_date)
		end

feature -- Status setting

	set_invalid is
			-- Set flag that entity is invalid.
		do
			category_id := 0
		ensure
			invalid: invalid
		end

	set_as_of_date (a_as_of_date: ROSE_DATE) is
			-- Set date on which we are searching for an entity.
		require
			valid_entity: valid
		do
			create as_of_date.make_by_date (a_as_of_date)
		end

	set_as_of_date_to_current is
			-- Set flag that entity attributes are the current values.
		require
			valid_entity: valid
		do
			as_of_date := as_of_date_from_end_date
		end

feature -- Basic operations

	add_relative_with_end_date (a_relationship_type_id: INTEGER; a_entity: like Current; a_start_date, a_end_date: ROSE_DATE) is
			-- Add a relative to an entity as of a particular end date.
		require
			valid_entity: valid
			a_relationship_type_id_not_zero: a_relationship_type_id /= 0
			a_entity_not_void: a_entity /= Void
			a_start_date_not_void: a_start_date /= Void
			a_end_date_not_void: a_end_date /= Void
		local
			l_relationship: SUPPLIER_ENTITY_RELATIONSHIP
			l_arguments: ARRAY [ANY]
			l_zero: DOUBLE
		do
			if not has_relative (a_relationship_type_id, a_entity) then
				if relationship_type_id_is_cached(a_relationship_type_id) and a_end_date.is_equal(constants.maximum_date_time) then
					relationship_cache.create_relationship
					(relationship_type_manager.relationship_type (a_relationship_type_id), Current, a_entity, a_start_date, 0.0)
				end
				create l_relationship.make
				(relationship_type_manager.relationship_type (a_relationship_type_id), Current, a_entity, a_start_date, 0.0)
				l_relationship.set_end_date(a_end_date)
				l_arguments := <<l_relationship.relationship_id,
								 l_relationship.relationship_type_id,
								 l_relationship.entity_1_id,
								 l_relationship.entity_2_id,
								 l_relationship.start_date,
								 l_relationship.end_date>>
				if (has_weighted_relationship) then
					l_arguments.force (l_zero, l_arguments.upper + 1)
				end
				insert_relationship_predicate.execute (Void, l_arguments)
				if not insert_relationship_predicate.success then
					report_error("add_relative","add_relative failed")
				end
			end
		end

	add_historical_relative_with_end_date (a_relationship_type_id: INTEGER; a_entity: like Current; a_start_date, a_end_date: ROSE_DATE) is
			-- Add historical relative with a particular end date.
		require
			valid_entity: valid
			a_relationship_type_id_not_zero: a_relationship_type_id /= 0
			a_entity_not_void: a_entity /= Void
			a_start_date_not_void: a_start_date /= Void
			a_end_date_not_void: a_end_date /= Void
		local
			l_relationship: SUPPLIER_ENTITY_RELATIONSHIP
			l_arguments: ARRAY [ANY]
			l_zero: DOUBLE
		do
			if not has_historical_relative (a_relationship_type_id, a_entity, a_start_date,a_end_date) then
				if relationship_type_id_is_cached(a_relationship_type_id) and a_end_date.is_equal(constants.maximum_date_time) then
					relationship_cache.create_relationship
					(relationship_type_manager.relationship_type (a_relationship_type_id), Current, a_entity, a_start_date, 0.0)
				end
				create l_relationship.make
				(relationship_type_manager.relationship_type (a_relationship_type_id), Current, a_entity, a_start_date, 0.0)
				l_relationship.set_end_date(a_end_date)
				l_arguments := <<l_relationship.relationship_id,
								 l_relationship.relationship_type_id,
								 l_relationship.entity_1_id,
								 l_relationship.entity_2_id,
								 l_relationship.start_date,
								 l_relationship.end_date>>
				if (has_weighted_relationship) then
					--				l_arguments.force (create {ROSE_DATE}.make_now, l_arguments.upper + 1)
					l_arguments.force (l_zero, l_arguments.upper + 1)
				end

				insert_relationship_predicate.execute (Void,l_arguments)
				if not insert_relationship_predicate.success then
					report_error("add_relative","add_relative failed")
				end
			end
		end

	add_relative (a_relationship_type_id: INTEGER; a_entity: like Current; a_start_date: ROSE_DATE) is
			-- Add `a_entity' relative if not already there, otherwise do nothing.
		require
			valid_entity: valid
			a_relationship_type_id_not_zero: a_relationship_type_id /= 0
			a_entity_not_void: a_entity /= Void
			a_start_date_not_void: a_start_date /= Void
		do
			add_relative_with_end_date(a_relationship_type_id,a_entity,a_start_date,constants.maximum_date_time)
		end

	add_weighted_relative_with_end_date (a_relationship_type_id: INTEGER; a_entity: like Current; a_start_date, a_end_date: ROSE_DATE; a_weight: DOUBLE) is
			-- Add weighted relative with a particular end date.
		require
			valid_entity: valid
			a_relationship_type_id_not_zero: a_relationship_type_id /= 0
			a_entity_not_void: a_entity /= Void
			a_start_date_not_void: a_start_date /= Void
			a_end_date_not_void: a_end_date /= Void
		local
			l_relationship: SUPPLIER_ENTITY_RELATIONSHIP
		do
			if not has_relative (a_relationship_type_id, a_entity) then
				if relationship_type_id_is_cached(a_relationship_type_id) and a_end_date.is_equal(constants.maximum_date_time) then
					relationship_cache.create_relationship
					(relationship_type_manager.relationship_type (a_relationship_type_id), Current, a_entity, a_start_date, a_weight)
				end
				create l_relationship.make
				(relationship_type_manager.relationship_type (a_relationship_type_id), Current, a_entity, a_start_date, a_weight)
				l_relationship.set_end_date(a_end_date)
				insert_relationship_predicate.execute (Void,<<l_relationship.relationship_id,
													   l_relationship.relationship_type_id,
													   l_relationship.entity_1_id,
													   l_relationship.entity_2_id,
													   l_relationship.start_date,
													   l_relationship.end_date,
													   l_relationship.weight>>)
				if not insert_relationship_predicate.success then
					report_error("add_relative","add_relative failed")
				end
			end
		end

	add_weighted_relative (a_relationship_type_id: INTEGER; a_entity: like Current; a_start_date: ROSE_DATE; a_weight: DOUBLE) is
			-- Add weighted relative as the current one.
		require
			valid_entity: valid
			a_relationship_type_id_not_zero: a_relationship_type_id /= 0
			a_entity_not_void: a_entity /= Void
			a_start_date_not_void: a_start_date /= Void
		do
			add_weighted_relative_with_end_date(a_relationship_type_id,a_entity,a_start_date,constants.maximum_date_time,a_weight)
		end

	terminate_relative (a_relationship_type_id: INTEGER; a_entity: like Current; a_end_date: ROSE_DATE) is
			-- Remove `a_entity' relative if already there, otherwise do nothing.
		require
			valid_entity: valid
			a_relationship_type_id_not_zero: a_relationship_type_id /= 0
			a_entity_not_void: a_entity /= Void
			a_end_date_not_void: a_end_date /= Void
		local
			l_entity_1_id, l_entity_2_id: INTEGER
			l_relationship_type: SUPPLIER_RELATIONSHIP_TYPE
			l_relationship: SUPPLIER_ENTITY_RELATIONSHIP
		do
			if has_relative (a_relationship_type_id, a_entity) then
				l_relationship_type := relationship_type_manager.relationship_type (a_relationship_type_id)

				-- It is important to test category_1 first so this will match like-to-like relationships.
				if category_id = l_relationship_type.category_1_id then
					l_entity_1_id := entity_id
					l_entity_2_id := a_entity.entity_id
				elseif category_id = l_relationship_type.category_2_id then
					l_entity_2_id := entity_id
					l_entity_1_id := a_entity.entity_id
				else
					report_error ("terminate_relative", "Invalid relationship for this entity")
				end
				update_relationship_predicate.execute (Void, <<a_end_date,
													   a_relationship_type_id,
													   l_entity_1_id,
													   l_entity_2_id,
													   constants.Maximum_date_time>>)
				if not update_relationship_predicate.success then
					report_error ("terminate_relative", "Failed to update relative (a_entity_id) " +
								  a_entity.entity_id.out)
				end
				if relationship_type_id_is_cached(a_relationship_type_id) then
					create l_relationship.make (l_relationship_type, Current, a_entity, constants.Maximum_date_time, 0.0)
					relationship_cache.delete_relationship (l_relationship)
				end
			end
		end

	replace_only_relative (a_relationship_type_id: INTEGER; a_entity: like Current; a_start_date: ROSE_DATE) is
			-- Add `a_entity' relative if not already there.  Replace if there but different.  Otherwise, do nothing.
		require
			valid_entity: valid
			a_relationship_type_id_not_zero: a_relationship_type_id /= 0
			a_entity_not_void: a_entity /= Void
			a_start_date_not_void: a_start_date /= Void
		local
			l_only_relative: like Current
			l_relationships: EXTENDABLE_BOUNDED_ARRAY [SUPPLIER_ENTITY_RELATIONSHIP]
		do
			l_only_relative := only_relative (a_relationship_type_id)
			if l_only_relative = Void then
				add_relative (a_relationship_type_id, a_entity, a_start_date)
			elseif not l_only_relative.is_equal (a_entity) then
				-- This is painful since we are getting the relationship again!
				l_relationships := relationships(a_relationship_type_id)
				if l_relationships.count/=1		-- Since it was before, it should be again.
				 then
					report_error ("replace_only_relative", "Multiple relationships when expecting only relattionship")
				end
				if a_start_date > l_relationships.item(1).start_date then
					terminate_relative (a_relationship_type_id, l_only_relative, a_start_date)
					add_relative (a_relationship_type_id, a_entity, a_start_date)
				end
			end
		end

	delete_relationships(a_relationship_type_id: INTEGER; a_entity: like Current) is
			-- Deletes (an actual delete) the relationships between this entity and another.
		do
			delete_relationships_predicate.execute(Current, <<a_relationship_type_id, a_entity.entity_id>>)
			if not delete_relationships_predicate.success then
				report_error("delete_relationships", "delete_relationships_predicate failed")
			end
		end

	update is
			-- Update the entity.
		require
			valid_entity: valid
		do
			update_predicate.execute (Current, Void)
			if not update_predicate.success then
				report_error ("update", "Update failed")
			end
		end

	terminate_all_parts (a_end_date: ROSE_DATE) is
			-- Terminate an entity's attributes and its relationships but NOT the entity itself.
		require
			valid_entity: valid
			a_end_date_not_void: a_end_date /= Void
		do
			terminate_all_relationships (a_end_date)
			terminate_all_attributes (a_end_date)
		end

	terminate (a_end_date: ROSE_DATE) is
			-- Terminate the entity and all of it's parts.
		do
			create as_of_date.make_by_date (a_end_date)
			set_end_date(a_end_date)
			terminate_all_parts(a_end_date)
			terminate_predicate.execute (Current, <<a_end_date, category_id, entity_id>>)
			if not terminate_predicate.success then
				report_error ("terminate", "Terminate failed")
			end
			if category_is_cached (category_id) then
				entity_cache.terminate_entity (Current, a_end_date)
			end
			--	 set_invalid
			--	 entity_id := 0
		end

	add_attribute (a_attribute_type_id: INTEGER; a_value: STRING; a_start_date: ROSE_DATE) is
			-- Add a new attribute corresponding to a_attribute_type_id.
			-- Inserts into both the attribute table and the caches if there is one.
		require
			valid_entity: valid
			attribute_does_not_exist: attribute_by_id (a_attribute_type_id) = Void
			a_attribute_type_id_not_zero: a_attribute_type_id /= 0
			a_value_not_void: a_value /= Void
			a_start_date_not_void: a_start_date /= Void
		local
			l_attribute: SUPPLIER_ENTITY_ATTRIBUTE[like Current]
		do
			l_attribute := make_empty_attribute
			l_attribute.make (Current, a_attribute_type_id, a_value, a_start_date)
			if category_and_attribute_are_cached (category_id, a_attribute_type_id) then
				entity_cache.add_attribute (l_attribute)
			end
		end

	terminate_attribute (a_attribute_type_id: INTEGER; a_end_date: ROSE_DATE) is
			-- Terminate the existing attribute.
		require
			valid_entity: valid
			attribute_exists: attribute_by_id (a_attribute_type_id) /= Void
			a_attribute_type_id_not_zero: a_attribute_type_id /= 0
			a_end_date_not_void: a_end_date /= Void
		local
			l_attribute: SUPPLIER_ENTITY_ATTRIBUTE[like Current]
		do
			l_attribute := attribute_by_id (a_attribute_type_id)
			if a_end_date > l_attribute.start_date then
				l_attribute.set_end_date (a_end_date)
				l_attribute.terminate
				if category_and_attribute_are_cached (category_id, a_attribute_type_id) then
					entity_cache.terminate_attribute (l_attribute, a_end_date)
				end
			end
		end

	terminate_all_attributes (a_end_date: ROSE_DATE) is
			-- Terminates all of the existing attributes.
		require
			valid_entity: valid
			a_end_date_not_void: a_end_date /= Void
		local
			l_attributes: ARRAYED_LIST [SUPPLIER_ENTITY_ATTRIBUTE[like Current]]
		do
			l_attributes := attributes
			from
				l_attributes.start
			until
				l_attributes.exhausted
			loop
				if a_end_date >= l_attributes.item.start_date then
					l_attributes.item.set_end_date (a_end_date)
					l_attributes.item.terminate
					if category_and_attribute_are_cached (category_id, l_attributes.item.attribute_type_id) then
						entity_cache.terminate_attribute (l_attributes.item, a_end_date)
					end
				end
				l_attributes.forth
			end
		end

	terminate_no_show_attributes(a_end_date: ROSE_DATE; a_attribute_type_ids: LINEAR[INTEGER]) is
			-- Terminate the attributes are NOT in 'a_attribute_type_ids'.
			-- This is useful when you receive an update that is guaranteed to be complete.  By passing
			-- this list of received attributes, we can terminate the ones that were not received.
			-- If the list is Void, nothing happens.
			-- The list may contain 0's (which are ignored).
			-- If there are no non-zero items, all attributes are terminated.
		require
			end_date_not_void: a_end_date/=Void
		local
			l_update_statement: STRING
			l_update_predicate: EXECUTION_PREDICATE
			l_count: INTEGER
		do
			if a_attribute_type_ids/=Void
			 then
				l_update_statement :=
					"update entity_attribute %
				%set end_date=? %
				%where %
				%category_id=:category_id and %
				%entity_id=:entity_id and %
				%end_date='12/31/9999'"

				from
					a_attribute_type_ids.start
				until
					a_attribute_type_ids.off
				loop
					if a_attribute_type_ids.item/=0
					 then
						if l_count=0
						 then
							l_update_statement.append(" and attribute_type_id not in (")
						else
							l_update_statement.append(",")
						end

						l_count := l_count + 1
						l_update_statement.append(a_attribute_type_ids.item.out)
					end

					a_attribute_type_ids.forth
				end

				if l_count>0
				 then
					l_update_statement.append(")")
				end

				create l_update_predicate.make(data_server, l_update_statement)
				if not l_update_predicate.success
				 then
					report_error("terminate_no_show_attributes", "Failed to make l_update_predicate.")
				end

				l_update_predicate.execute(Current, <<a_end_date>>)
				if not l_update_predicate.success
				 then
					report_error("terminate_no_show_attributes", "l_update_predicate failed.")
				end
			end
		end

	update_attribute (a_attribute_type_id: INTEGER; a_value: STRING; a_start_date: ROSE_DATE) is
			-- Terminate the existing attribute then add the new attribute.
		require
			valid_entity: valid
			a_attribute_type_id_not_zero: a_attribute_type_id /= 0
			a_value_not_void: a_value /= Void
			a_start_date_not_void: a_start_date /= Void
		local
			l_attribute: SUPPLIER_ENTITY_ATTRIBUTE[like Current]
			l_entity: SUPPLIER_ENTITY
		do
			-- If this attribute value is a unique identifier and exists for another entity, terminate the value.
			if attribute_type_manager.attribute_type_is_unique_identifier (a_attribute_type_id) then
				if category_and_attribute_are_cached (category_id, a_attribute_type_id) then
					l_attribute := entity_cache.attribute_by_value (category_id, a_attribute_type_id, a_value, a_start_date)
					if l_attribute /= Void and then l_attribute.entity_id /= entity_id then
						l_entity := server.new_empty_entity
						l_entity.set_category_id (category_id)
						l_entity.set_entity_id (l_attribute.entity_id)
						l_entity.terminate_attribute (a_attribute_type_id, a_start_date)
					end
				else
					l_entity := server.new_empty_entity
					l_entity.attempt_set_from_database_attribute (category_id, a_attribute_type_id, a_value, constants.Maximum_date_time)
					if l_entity.valid and then l_entity.entity_id /= entity_id then
						l_entity.terminate_attribute (a_attribute_type_id, a_start_date)
					end
				end
			end

			-- Update the attribute.
			l_attribute := attribute_by_id (a_attribute_type_id)
			if l_attribute = Void then
				add_attribute (a_attribute_type_id, a_value, a_start_date)
			elseif not l_attribute.attribute_value.is_equal (a_value) and then a_start_date > l_attribute.start_date then
				if l_attribute.end_date < constants.maximum_date_time then
					l_attribute.update_end_date (a_start_date)
				else
					terminate_attribute (a_attribute_type_id, a_start_date)
				end
				add_attribute (a_attribute_type_id, a_value, a_start_date)
			end
		end

	insert_attribute_interval (a_interval : SUPPLIER_ENTITY_ATTRIBUTE [like Current]; a_force : BOOLEAN) is
			-- Insert `a_interval' into current attribute history which matches `a_interval.attribute_type_id'
			-- If `a_force' then insert even if in conflict with unique attribute
		require
			a_attribute_interval_not_void : a_interval /= Void
		local
			l_attributes : ROSE_LINEAR_ARRAY [SUPPLIER_ENTITY_ATTRIBUTE [like Current]]
			l_attribute : SUPPLIER_ENTITY_ATTRIBUTE [like Current]
			l_type : INTEGER
			i : INTEGER
		do
			if a_force or not is_in_conflict_with_unique_attribute (a_interval) then
						-- `update_attribute' detects conflicts with other entities with unique attribute matching new attribute values
						-- and silently terminates - this does not seem like a good idea.
				l_type := a_interval.attribute_type_id
				l_attributes := attributes_between (l_type, a_interval.start_date, a_interval.end_date)
				if l_attributes /= Void then
					l_attributes.sort_by (agent compare_attributes_by_start_date)
							-- remove all rows which overlap with `a_attribute'
					from i := 1 until i > l_attributes.count loop
						l_attribute := l_attributes [i]
						delete_attribute_history_row (l_attribute)
						i := i + 1
					end
					if l_attributes.first.start_date < a_interval.start_date then
						create l_attribute.make_without_insert (Current, a_interval.attribute_type_id)
						l_attribute.set_attribute_value (l_attributes.first.attribute_value)
						l_attribute.set_start_date (l_Attributes.first.start_date)
						l_attribute.set_end_date (a_interval.start_date)
						insert_attribute_interval (l_attribute, False)
					end
					if l_attributes.last.end_date > a_interval.end_date then
						create l_attribute.make_without_insert (Current, a_interval.attribute_type_id)
						l_attribute.set_attribute_value (l_attributes.first.attribute_value)
						l_attribute.set_start_date (a_interval.end_date)
						l_attribute.set_end_date (l_attributes.last.end_date)
						insert_attribute_interval (l_attribute, False)
					end
				end
				a_interval.insert
			end
		end

	update_attribute_caches (a_attribute_type_id : INTEGER) is
			-- Update cache with database data
		local
			l_history, l_new_history, l_old_history, l_cached_history : EXTENDABLE_BOUNDED_ARRAY [SUPPLIER_ENTITY_ATTRIBUTE [like Current]]
			l_sortable_history : ROSE_LINEAR_ARRAY [SUPPLIER_ENTITY_ATTRIBUTE [like Current]]
			l_attribute_interval : SUPPLIER_ENTITY_ATTRIBUTE [like Current]
			l_key : KEY_PAIR [INTEGER, INTEGER]
			l_value_key : MULTI_KEY
			i , j : INTEGER
		do
					-- Update attribute_by_entity_cache
			l_new_history := attribute_history_from_database (a_attribute_type_id)
			create l_key.make (entity_id, a_attribute_type_id)
			l_old_history := entity_cache.attribute_by_entity_cache [l_key]
			l_cached_history := l_old_history.twin
			l_old_history.wipe_out
			l_old_history.append_array (l_new_history)
					-- Remove old from attribute_by_value_cache
			from i := 1 until i > l_cached_history.count loop
				l_attribute_interval := l_cached_history [i]
				create l_value_key.make (<<entity_id, a_attribute_type_id, l_attribute_interval.attribute_value>>)
				l_old_history := entity_cache.attribute_by_value_cache [l_value_key]
				if l_old_history /= Void and then not l_old_history.is_empty then
					create l_history.make_empty
					from j := 1 until j > l_old_history.count loop
						if l_old_history [j].entity_id /= entity_id then
							l_history.extend (l_old_history [j])
						end
						j := j + 1
					end
					l_old_history.wipe_out
					l_old_history.append_array (l_history)
				end
				i := i + 1
			end
					-- Insert new history into attribute_value_cache
			from i := 1 until i > l_new_history.count loop
				l_attribute_interval := l_new_history [i]
				create l_value_key.make (<<entity_id, a_attribute_type_id, l_attribute_interval.attribute_value>>)
				l_old_history := entity_cache.attribute_by_value_cache [l_value_key]
				if l_old_history = Void then
					create l_old_history.make_empty
					entity_cache.attribute_by_value_cache [l_value_key] := l_old_history
				end
				l_old_history.extend (l_attribute_interval)
				create l_sortable_history.make_from_array (l_old_history)
				l_sortable_history.sort_by (agent entity_cache.compare_by_value_key)
				l_old_history.wipe_out
				l_old_history.append_array (l_sortable_history)
				i := i + 1
			end
		end

	is_in_conflict_with_unique_attribute (a_attribute : like attribute_type) : BOOLEAN is
			-- Is `a_attribute' marked unique and also in conflict with another attribute?
		require
			a_attribute_not_void : a_attribute /= Void
		do
			Result := unique_attribute_in_conflict (a_attribute) = Void
		end

	unique_attribute_in_conflict (a_attribute : like attribute_type) : like attribute_type is
			-- What (if any) unique attribute is `a_attribute' in conflict with
		require
			a_attribute_not_void : a_attribute /= Void
		local
			l_category_id, l_type_id, l_handle : INTEGER
			l_value : STRING
			l_start, l_end : ROSE_DATE
			l_attribute : like attribute_type
			l_entity: like Current
		do
			l_category_id := a_attribute.category_id
			l_type_id := a_attribute.attribute_type_id
			l_value := a_attribute.attribute_value
			l_start := a_attribute.start_date
			l_end := a_attribute.end_date
			l_handle := a_attribute.entity_id
			if attribute_type_manager.attribute_type_is_unique_identifier (l_type_id) then
				if category_and_attribute_are_cached (l_category_id, l_type_id) then
					l_attribute := entity_cache.attribute_by_value (l_category_id, l_type_id, l_value, l_start)
					if l_attribute /= Void and then l_attribute.entity_id /= l_handle then
						Result := l_attribute
					end
				else
					l_entity := server.new_empty_entity
					l_entity.attempt_set_from_database_attribute (l_category_id, l_type_id, l_value, l_start)
					if l_entity.valid and then l_entity.entity_id /= l_handle then
						Result := l_entity.attribute_by_id (l_type_id)
					end
				end
			end
		end

	is_history_in_conflict_with_unique_attribute (a_history : ENTITY_ATTRIBUTE_HISTORY) : BOOLEAN is
			-- Are any members of `a_history' in conflict with an extant unique attribute?
		local
			i : INTEGER
			l_attribute : like attribute_type
			l_interval : ROSE_CLOSED_OPEN_INTERVAL [STRING]
		do
			from i := 1 until i > a_history.count or Result loop
				l_interval := a_history.item (i)
				l_attribute := new_attribute_without_insert (a_history.type_id, l_interval.start_date, l_interval.end_date, l_interval.item)
				Result := is_in_conflict_with_unique_attribute (l_attribute)
				i := i + 1
			end
		end

	new_attribute_without_insert (a_type : INTEGER; a_start, a_end : ROSE_DATE; a_value : STRING) : SUPPLIER_ENTITY_ATTRIBUTE [like Current] is
			-- New attribute object
		do
			create Result.make_without_insert (Current, a_type)
			Result.set_start_date (a_start)
			Result.set_end_date (a_end)
			Result.set_attribute_value (a_value)
		end

	update_attribute_history (a_history : ENTITY_ATTRIBUTE_HISTORY; a_force : BOOLEAN) is
			-- Update attribute history row where non-match found to `a_history' row
			-- If `a_force' then override extant unique attributes (if applicable)
		require
			valid_history : a_history /= Void and then a_history.is_contiguous and then a_history.is_continuous
		local
			i : INTEGER
			l_attribute : like attribute_type
			l_interval : ROSE_CLOSED_OPEN_INTERVAL [STRING_8]
		do
			from i := 1 until i > a_history.count loop
				l_interval := a_history.item (i)
				l_attribute := new_attribute_without_insert (a_history.type_id, l_interval.start_date, l_interval.end_date, l_interval.item)
				insert_attribute_interval (l_attribute, a_force)
				i := i + 1
			end
			if category_and_attribute_are_cached (a_history.category, a_history.type_id) then
				update_attribute_caches (a_history.type_id)
			end
		end

	replace_attribute_history (a_history : ENTITY_ATTRIBUTE_HISTORY; a_force : BOOLEAN) is
			-- Delete existing history for attribute matching `a_history' and replace with `a_history'
			-- If `a_force' then override extant unique attributes (if applicable)
		require
			valid_history : a_history /= Void and then a_history.is_contiguous and then a_history.is_continuous
		local
			l_history : ROSE_LINEAR_ARRAY [like attribute_type]
			l_array : EXTENDABLE_BOUNDED_ARRAY [SUPPLIER_ENTITY_ATTRIBUTE [like Current]]
			i : INTEGER
			l_attribute : like attribute_type
		do
			if a_force or else not is_history_in_conflict_with_unique_attribute (a_history) then
				l_array := attribute_history (a_history.type_id)
				if l_array /= Void then
					create l_history.make_from_array (l_array)
					from i := 1 until i > l_history.count loop
						l_attribute := l_history.item (i)
						delete_attribute_history_row (l_attribute)
						i := i + 1
					end
				end
				update_attribute_history (a_history, a_force)
			end
			if category_and_attribute_are_cached (a_history.category, a_history.type_id) then
				update_attribute_caches (a_history.type_id)
			end
		end

	set_transition (a_transition_type_id: INTEGER; a_transition_date: ROSE_DATE) is
			-- Add a new transition corresponding to a_transition_type_id if it doesn't already exist.
			-- Contrast with 'set_unique_transition'.
			-- NOTE - This does NOT update the transition.  It NEVER performs an SQL update.  It only
			-- checks for the presence of the transition type AND DATE.  If a transition with that
			-- date does not exists, a new record is added, not updated.
			-- If you want the old transition date updated with the new transition date, use 'update_transition'.
		require
			valid_entity: valid
			a_transition_date_not_void: a_transition_date /= Void
			a_transition_type_id_not_zero: a_transition_type_id /= 0
		do
			if transition (a_transition_type_id, a_transition_date) = Void then
				insert_transition (a_transition_type_id, a_transition_date)
			end
		end

 	set_unique_transition(a_transition_type_id: INTEGER; a_transition_date: ROSE_DATE) is
			-- Creates or modifies a unique transition.  Only one unique transition of a given type is allowed
			-- per entity.  Contrast with 'set_transition'.
		require
			valid_entity: valid
			transition_type_id_not_0: a_transition_type_id/=0
			transition_date_not_void: a_transition_date/=Void
		local
			l_transition: ENTITY_TRANSITION
		do
			l_transition := unique_transition(a_transition_type_id)
			if l_transition=Void
			 then
				-- It wasn't found so add it.
				insert_transition(a_transition_type_id, a_transition_date)
			else
				-- It was found so update it.
				update_transition(l_transition, a_transition_date)
			end
		end

	update_transition(a_transition: ENTITY_TRANSITION; a_transition_date: ROSE_DATE) is
			-- Modifies the date of an existing transition.
		require
			valid_entity: valid
			transition_not_void: a_transition/=Void
			transition_date_not_void: a_transition_date/=Void
		do
			update_transition_predicate.execute(a_transition, <<a_transition_date>>)
			if not update_transition_predicate.success
			 then
				report_error ("update_transition", "update_transition_predicate failed")
			end
		end

	update_transition_event_id(a_transition: ENTITY_TRANSITION; a_event_id: INTEGER) is
			-- Modifies the event id of an existing transition.
		require
			valid_entity: valid
			transition_not_void: a_transition /= Void
		do
			update_transition_event_id_predicate.execute(a_transition, <<a_event_id>>)
			if not update_transition_event_id_predicate.success
			 then
				report_error ("update_transition_event_id", "update_transition_event_id_predicate failed")
			end
		end

	update_unique_transition_range(a_start_transition_type_id, a_end_transition_type_id: INTEGER; a_transition_date: ROSE_DATE) is
			-- For the start transition:
			--		If a_transition_date is LESS than the date of the existing start transition, then update it.
			-- For the end transition:
			--		If a_transition_date is GREATER than the date of the existing end transition, then update it.
			-- If either transition is missing, add it.
			-- If a transition_type_id is 0, it is skipped.
		require
			valid_entity: valid
			a_transition_date_not_void: a_transition_date /= Void
		local
			l_transition: ENTITY_TRANSITION
		do
			if a_start_transition_type_id/=0
			 then
				-- Get the one we are interested in.
				l_transition := unique_transition(a_start_transition_type_id)
				if l_transition=Void
				 then
					-- It wasn't found so add it.
					insert_transition(a_start_transition_type_id, a_transition_date)
				else
					if a_transition_date<l_transition.transition_date
					 then
						-- The new date is more recent.
						update_transition_predicate.execute(l_transition, <<a_transition_date>>)
					end
				end
			end

			if a_end_transition_type_id/=0
			 then
				-- Get the one we are interested in.
				l_transition := unique_transition(a_end_transition_type_id)
				if l_transition=Void
				 then
					-- It wasn't found so add it.
					insert_transition(a_end_transition_type_id, a_transition_date)
				else
					if a_transition_date>l_transition.transition_date
					 then
						-- The new date is later.
						update_transition_predicate.execute(l_transition, <<a_transition_date>>)
					end
				end
			end
		end

	terminate_company_tree (a_date: ROSE_DATE) is
			-- Terminate all non-terminated entities in a_entity's company tree with a_date.
		require
			a_valid_company_security_or_listing: valid_company_security_or_listing
			a_date_not_void: a_date /= Void
		local
			l_tree: like company_tree
		do
			l_tree := company_tree
			if l_tree.item.end_date.is_equal (constants.maximum_date_time) then l_tree.item.terminate (a_date) end
			from l_tree.child_start until l_tree.child_off loop
				if l_tree.child.item.end_date.is_equal (constants.maximum_date_time) then l_tree.child.item.terminate (a_date) end
				from l_tree.child.child_start until l_tree.child.child_off loop
					if l_tree.child.child.item.end_date.is_equal (constants.maximum_date_time) then l_tree.child.child.item.terminate (a_date) end
					l_tree.child.child_forth
				end
				l_tree.child_forth
			end
		end

feature {ROSE_ID_USER} -- Basic operations - deletion

	delete_attributes is
			-- Deletes (an actual delete) all the attributes
		do
			delete_attributes_predicate.execute (Current,Void)
			if not delete_attributes_predicate.success then
				report_error ("delete_attributes", "delete_attributes_predicate failed")
			end
		end

	delete_attribute_history_row (a_attribute_row : SUPPLIER_ENTITY_ATTRIBUTE [like Current]) is
			-- Delete attribute row corresponding to `a_attribute' if it exists
		require
			a_attribute_not_void : a_attribute_row /= Void
		do
			delete_attribute_history_row_predicate.execute (a_attribute_row, Void)
			if not delete_attribute_history_row_predicate.success then
				report_error ("delete_attribute_history_row", "delete_attribute_history_row_predicate failed")
			end
		end

	delete_entity is
			-- Deletes (an actual delete) all the attributes
		do
			delete_predicate.execute (Current,Void)
			if not delete_predicate.success then
				report_error ("delete_entity", "delete_predicate failed")
			end
		end

feature {ROSE_ID_USER} -- Access - deletion
	is_attribute_history_row_extant (a_attribute_row : SUPPLIER_ENTITY_ATTRIBUTE [like Current]) : BOOLEAN is
			-- Does an entity_attribute row matching `a_attribute_row' exist in the table?
		require
			a_attribute_row_not_void : a_attribute_row /= Void
		do
			select_attribute_predicate.retrieve (a_attribute_row, <<a_attribute_row.start_date, a_attribute_row.end_date>>)
			if not select_attribute_predicate.success then
				report_error ("is_attribute_history_row_extant", "Select entity attribute failed (attid) " +
							  a_attribute_row.attribute_type_id.out)
			else
				Result := select_attribute_predicate.array.count > 0
			end
		end

feature {SUPPLIER_ENTITY} -- Implementation

	attempt_set_from_database_attribute (a_category_id: INTEGER;
										 a_attribute_type_id: INTEGER;
										 a_attribute_value: STRING;
										 a_as_of_date: ROSE_DATE) is
			-- Sets entity_id and as_of_date to the entity_id as of a_as_of_date from the database attributes
			-- for the given attribute value if it exists.  If 'a_as_of_date' if Void, find the most recent.
		require
			a_category_id_not_zero: a_category_id /= 0
			a_attribute_type_id_not_zero: a_attribute_type_id /= 0
			a_attribute_value_not_void: a_attribute_value /= Void
			a_as_of_date_not_void: a_as_of_date /= Void
		do
			if (a_as_of_date=Void)
			 then
				-- This code never worked!  The precondition to prevent this is back.
				select_most_recent_attribute_by_value_predicate.retrieve_one
				(Current, Void, <<a_category_id, a_attribute_type_id, a_attribute_value>>)
				if not select_most_recent_attribute_by_value_predicate.success then
					report_error ("attempt_set_from_database_attribute",
								  "Select most recent attribute failed (atttypeid, attvalue) " +
								  a_attribute_type_id.out + " " + a_attribute_value)
				end

				if select_most_recent_attribute_by_value_predicate.index = 0
				 then
					set_invalid
				else
					make_from_database (a_category_id, entity_id, a_as_of_date)
				end
			else
				select_attribute_by_value_predicate.retrieve_one
				(Current, Void, <<a_category_id, a_attribute_type_id, a_attribute_value, a_as_of_date, a_as_of_date.not_max>>)
				if not select_attribute_by_value_predicate.success then
					report_error ("attempt_set_from_database_attribute",
								  "Select attribute failed (atttypeid, attvalue) " +
								  a_attribute_type_id.out + " " + a_attribute_value)
				end

				if select_attribute_by_value_predicate.index = 0
				 then
					set_invalid
				else
					make_from_database (a_category_id, entity_id, a_as_of_date)
				end
			end
		end

feature {SUPPLIER_ENTITY_CACHE} -- Implementation
	next_attribute (a_relationship_type_id, a_attribute_type_id: INTEGER): SUPPLIER_ENTITY_ATTRIBUTE[like Current] is
			-- First attribute with an end_date greater than as_of_date.
		require
			valid_entity: valid
			a_attribute_type_id_not_zero: a_attribute_type_id /= 0
		local
			l_entity: like Current
			l_relatives: ARRAYED_LIST [like Current]
		do
			if a_relationship_type_id = 0 then
				l_entity := Current
			else
				l_relatives := relatives (a_relationship_type_id)
				if l_relatives /= Void then
					l_entity := l_relatives.i_th (1)
				end
			end
			if l_entity /= Void then
				if category_and_attribute_are_cached (l_entity.category_id, a_attribute_type_id) and then
					not entity_cache.updated then
					Result := entity_cache.next_attribute (l_entity, a_attribute_type_id, as_of_date)
				else
					Result := make_empty_attribute
					Result.make_without_insert (l_entity, a_attribute_type_id)
					select_attribute_predicate.retrieve_one (Result, Result, <<as_of_date, as_of_date.not_max>>)
					if not select_attribute_predicate.success then
						report_error ("next_attribute", "Select entity attribute failed (attid) " +
									  a_attribute_type_id.out)
					end
					if select_attribute_predicate.index = 0 then
						Result := Void
					end
				end
			end
		end

	attribute_from_database (a_attribute_type_id: INTEGER; a_date: ROSE_DATE): SUPPLIER_ENTITY_ATTRIBUTE[like Current] is
			-- Get attribute form database
		require
			valid_entity: valid
		do
			Result := make_empty_attribute
			Result.make_without_insert (Current, a_attribute_type_id)
			select_attribute_predicate.retrieve_one (Result, Result, <<a_date, a_date.not_max>>)
			if not select_attribute_predicate.success then
				report_error ("attribute_from_database", "Select entity attribute failed (attid) " +
							  a_attribute_type_id.out)
			end
			if select_attribute_predicate.index = 0 then
				Result := Void
			elseif select_attribute_predicate.index /= 1 then
				report_error ("attribute_from_database",
							  "Duplicate attributes retrieved but not supported (attid)" + a_attribute_type_id.out)
			end
		end

	next_attribute_from_database (a_attribute_type_id: INTEGER; a_date: ROSE_DATE): SUPPLIER_ENTITY_ATTRIBUTE[like Current] is
			-- Next attribute from database
		require
			valid_entity: valid
		do
			Result := make_empty_attribute
			Result.make_without_insert (Current, a_attribute_type_id)
			select_next_attribute_predicate.retrieve_one (Result, Result, <<a_date>>)
			if not select_next_attribute_predicate.success then
				report_error ("next_attribute_from_database", "Select entity attribute failed (attid) " +
							  a_attribute_type_id.out)
			end
			if select_next_attribute_predicate.index = 0 then
				Result := Void
			end
		end

	attribute_history_from_database (a_attribute_type_id: INTEGER): EXTENDABLE_BOUNDED_ARRAY [SUPPLIER_ENTITY_ATTRIBUTE [like Current]] is
			-- attribute history from database
		require
			valid_entity: valid
		local
			l_array: EXTENDABLE_BOUNDED_ARRAY [SUPPLIER_ENTITY_ATTRIBUTE[like Current]]
		do
			create l_array.make_empty
			select_attribute_history_predicate.set_container (l_array)
			select_attribute_history_predicate.retrieve (Void, <<category_id, entity_id, a_attribute_type_id>>)
			if not select_attribute_history_predicate.success then
				report_error ("attribute_history_from_database", "Select entity attribute failed (attid) " +
							  a_attribute_type_id.out)
			end
			if select_attribute_history_predicate.index /= 0 then
				Result := l_array
			end
		end

	set_from_database_historical_attribute (a_attribute_type_id: INTEGER;
											a_attribute_value: STRING;
											a_start_date: ROSE_DATE) is
			-- Sets entity_id and as_of_date to the latest entity_id and end_date from the database attributes
			-- for the given attribute value and on or after 'a_min_date'.  Creates new if no attribute.
		require
			valid_entity: valid
			a_attribute_type_id_not_zero: a_attribute_type_id /= 0
			a_attribute_value_not_void: a_attribute_value /= Void
			a_start_date_not_void: a_start_date /= Void
		local
			l_save_as_of_date: ROSE_DATE
		do
			select_attribute_by_historical_value_predicate.retrieve_one
			(Current, Void, <<category_id, a_attribute_type_id, a_attribute_value>>)
			if not select_attribute_by_historical_value_predicate.success then
				report_error ("set_from_database_historical_attribute",
							  "Select attribute failed (atttypeid, attvalue) " +
							  a_attribute_type_id.out + " " + a_attribute_value)
			end
			if select_attribute_by_historical_value_predicate.index = 0 then
				make (category_id, a_start_date)
				add_attribute (a_attribute_type_id, a_attribute_value, a_start_date)
			else
				create l_save_as_of_date.make_by_date (as_of_date)
				make_from_database (category_id, entity_id, a_start_date)
				set_as_of_date (l_save_as_of_date)
			end
		end

	category_is_cached (a_category_id: INTEGER): BOOLEAN is
			-- Is a category identified by a_category_id cached?
		require
			valid_category_id: a_category_id/=0
		do
			Result := entity_cache /= Void and then entity_cache.has_category (a_category_id)
		end

	category_and_attribute_are_cached (a_category_id: INTEGER; a_attribute_type_id: INTEGER): BOOLEAN is
			-- Are `a_category_id' and `a_attribute_type_id' cached?
		require
			valid_category_id: a_category_id/=0
		do
			Result := entity_cache /= Void and then
			entity_cache.has_category_and_attribute (a_category_id, a_attribute_type_id)
		end

	has_relative (a_relationship_type_id: INTEGER; a_entity: like Current): BOOLEAN is
		require
			valid_entity: valid
			a_relationship_type_id_not_zero: a_relationship_type_id /= 0
			a_entity_not_void: a_entity /= Void
		local
			l_relatives: ARRAYED_LIST [like Current]
		do
			Result := false
			l_relatives := relatives (a_relationship_type_id)
			if l_relatives /= Void then
				from
					l_relatives.start
				until
					l_relatives.exhausted or Result
				loop
					if l_relatives.item.entity_id = a_entity.entity_id then
						Result := true
					end
					l_relatives.forth
				end
			end
		end

	has_historical_relative (a_relationship_type_id: INTEGER; a_entity: like Current; a_start_date,a_end_date:ROSE_DATE): BOOLEAN is
		require
			valid_entity: valid
			a_relationship_type_id_not_zero: a_relationship_type_id /= 0
			a_entity_not_void: a_entity /= Void
		local
			l_relationships:ROSE_LINEAR_ARRAY[SUPPLIER_ENTITY_RELATIONSHIP]
			l_start,l_end:ROSE_DATE
			l_relationship:SUPPLIER_ENTITY_RELATIONSHIP
		do
		 	Result := false
			create l_start.make_max_date
			create l_end.make_min_date
			select_relationships_2_predicate.retrieve(Void,<<a_relationship_type_id, entity_id, l_start,l_end>>)
			l_relationships := select_relationships_2_predicate.array
			if l_relationships /= Void and then l_relationships.count > 0 then
			 	from l_relationships.start
			 	until l_relationships.off or Result
			 	loop
			 		l_relationship := l_relationships.current_item
			 		if l_relationship.entity_2_id = a_entity.entity_id and then l_relationship.start_date <= a_start_date
						and l_relationship.end_date >= a_end_date then
			 			Result := True
			 		end
			 		l_relationships.forth
			 	end
			end
			if not Result then
				select_relationships_1_predicate.retrieve(Void,<<a_relationship_type_id, entity_id, l_start,l_end>>)
				l_relationships := select_relationships_1_predicate.array
				if l_relationships /= Void and then l_relationships.count > 0 then
				 	from l_relationships.start
				 	until l_relationships.off or Result
				 	loop
				 		l_relationship := l_relationships.current_item
				 		if l_relationship.entity_1_id = a_entity.entity_id and then l_relationship.start_date = a_start_date
							and l_relationship.end_date = a_end_date then
				 			Result := True
				 		end
				 		l_relationships.forth
				 	end
				end
			end
		end

	directed_relatives_with_attribute (a_relationship_type: SUPPLIER_RELATIONSHIP_TYPE;
									   a_attribute_type_id: INTEGER;
									   a_attribute_value: STRING;
									   a_relative_style: BOOLEAN): ARRAYED_LIST [like Current] is
		require
			valid_entity: valid
			a_relationship_type_not_void: a_relationship_type /= Void
			a_attribute_type_id_not_zero: a_attribute_type_id /= 0
			a_attribute_value_not_void: a_attribute_value /= Void
		local
			l_relationships: EXTENDABLE_BOUNDED_ARRAY [SUPPLIER_ENTITY_RELATIONSHIP]
			i: INTEGER
			l_entity: like Current
			l_attribute: SUPPLIER_ENTITY_ATTRIBUTE[like Current]
		do
			l_relationships := directed_relationships (a_relationship_type, a_relative_style)
			if l_relationships /= Void then
				from
					i := 1
				until
					i > l_relationships.count
				loop
					l_entity := server.new_empty_entity
					if a_relative_style=Left_relative then
						l_entity.make_from_database (a_relationship_type.category_1_id,
													 l_relationships.item (i).entity_1_id,
													 l_relationships.item (i).start_date)
					else
						l_entity.make_from_database (a_relationship_type.category_2_id,
													 l_relationships.item (i).entity_2_id,
													 l_relationships.item (i).start_date)
					end

					l_attribute := l_entity.attribute_by_id (a_attribute_type_id)
					if l_attribute /= Void and then l_attribute.attribute_value.is_equal (a_attribute_value) then
						if Result = Void then
							create Result.make (0)
							Result.extend (l_entity)
						else
							Result.extend (l_entity)
						end
					end

					i := i + 1
				end
			end
		end

	directed_relatives (a_relationship_type: SUPPLIER_RELATIONSHIP_TYPE; a_relative_style: BOOLEAN): ARRAYED_LIST [like Current] is
		require
			valid_entity: valid
			a_relationship_type_not_void: a_relationship_type /= Void
		local
			l_relationships: EXTENDABLE_BOUNDED_ARRAY [SUPPLIER_ENTITY_RELATIONSHIP]
			i: INTEGER
			l_entity: like Current
		do
			l_relationships := directed_relationships (a_relationship_type, a_relative_style)

			if l_relationships /= Void then
				create Result.make (l_relationships.count)
				from
					i := 1
				until
					i > l_relationships.count
				loop
					l_entity := server.new_empty_entity
					if a_relative_style=Left_relative then
						l_entity.make_from_database (a_relationship_type.category_1_id,
													 l_relationships.item (i).entity_1_id,
													 l_relationships.item (i).start_date)
					else
						l_entity.make_from_database (a_relationship_type.category_2_id,
													 l_relationships.item (i).entity_2_id,
													 l_relationships.item (i).start_date)
					end

					Result.extend (l_entity.deep_twin)
					i := i + 1
				end
			end
		end

	relationships (a_relationship_type_id: INTEGER): EXTENDABLE_BOUNDED_ARRAY [SUPPLIER_ENTITY_RELATIONSHIP] is
		require
			valid_entity: valid
			a_relationship_type_id_not_zero: a_relationship_type_id /= 0
		local
			l_relationship_type: SUPPLIER_RELATIONSHIP_TYPE
		do
			l_relationship_type := relationship_type_manager.relationship_type (a_relationship_type_id)

			-- It is important to test category_1 first so this will match like-to-like relationships.
			if category_id = l_relationship_type.category_1_id then
				Result := directed_relationships(l_relationship_type, Right_relative)
			elseif category_id = l_relationship_type.category_2_id then
				Result := directed_relationships(l_relationship_type, Left_relative)
			else
				report_error ("relationships", "Invalid relationship for this entity")
			end
		end

	directed_relationships (a_relationship_type: SUPPLIER_RELATIONSHIP_TYPE; a_relationship_style: BOOLEAN): EXTENDABLE_BOUNDED_ARRAY [SUPPLIER_ENTITY_RELATIONSHIP] is
		require
			valid_entity: valid
			a_relationship_type_not_void: a_relationship_type /= Void
		local
			l_relationships: EXTENDABLE_BOUNDED_ARRAY [SUPPLIER_ENTITY_RELATIONSHIP]
			l_start_date, l_end_date: ROSE_DATE
		do
			if not historical_relative_lookup and then
				as_of_date_is_current and then
				relationship_type_id_is_cached(a_relationship_type.relationship_type_id)
			 then
				l_relationships :=
					relationship_cache.directed_relationships (a_relationship_type, Current, a_relationship_style)
			else
				if historical_relative_lookup then
					create l_start_date.make_max_date
					create l_end_date.make_min_date
				else
					l_start_date := as_of_date
					l_end_date := as_of_date
				end
				create l_relationships.make_empty

				if a_relationship_style=Left_relative then
					select_relationships_1_predicate.set_container (l_relationships)
					select_relationships_1_predicate.retrieve (Void, <<a_relationship_type.relationship_type_id, entity_id, l_start_date, l_end_date.not_max>>)
					if not select_relationships_1_predicate.success then
						report_error ("directed_relationships", "Select relationships 1 failed (reltypid) " +
									  a_relationship_type.relationship_type_id.out)
					end

					if select_relationships_1_predicate.index = 0 then
						l_relationships := Void
					end
				else
					select_relationships_2_predicate.set_container (l_relationships)
					select_relationships_2_predicate.retrieve (Void, <<a_relationship_type.relationship_type_id, entity_id, l_start_date, l_end_date.not_max>>)
					if not select_relationships_2_predicate.success then
						report_error ("directed_relationships", "Select relationships 2 failed (reltypid) " +
									  a_relationship_type.relationship_type_id.out)
					end

					if select_relationships_2_predicate.index = 0 then
						l_relationships := Void
					end
				end
			end

			Result := l_relationships
		end

	all_relationships: EXTENDABLE_BOUNDED_ARRAY [SUPPLIER_ENTITY_RELATIONSHIP] is
			-- Returns all relationships in which the Current entity is a participant (left or right).
		require
			valid_entity: valid
		local
			l_relationships: EXTENDABLE_BOUNDED_ARRAY [SUPPLIER_ENTITY_RELATIONSHIP]
		do
			if relationship_cache /= Void and relationship_cache.all_relationships_are_cached
			 then
				l_relationships := relationship_cache.all_relationships (Current)
			else
				create l_relationships.make_empty
				select_all_relationships_predicate.set_container (l_relationships)
				select_all_relationships_predicate.retrieve (Void, <<entity_id, entity_id>>)
				if not select_all_relationships_predicate.success then
					report_error ("all_relationships", "Select all relationships failed")
				end
				if select_all_relationships_predicate.index = 0 then
					l_relationships := Void
				end
			end

			Result := l_relationships
		ensure
			not_empty: Result=Void or else not Result.is_empty
		end



	terminate_all_relationships (a_end_date: ROSE_DATE) is
			-- Terminate all relationships in which the Current entity is a participant (left or right).
		require
			valid_entity: valid
			a_end_date_not_void: a_end_date /= Void
		local
			l_relationships: EXTENDABLE_BOUNDED_ARRAY [SUPPLIER_ENTITY_RELATIONSHIP]
			i: INTEGER
		do
			l_relationships := all_relationships
			if l_relationships/=Void
			 then
				from
					i := l_relationships.lower
				until
					i > l_relationships.upper
				loop
					if a_end_date >= l_relationships.item (i).start_date	-- We already know that the end_date is maximum_date_time.
					 then
						if relationship_type_id_is_cached(l_relationships.item(i).relationship_type_id) then
							relationship_cache.delete_relationship (l_relationships.item(i))
						end
						l_relationships.item (i).set_end_date (a_end_date)
						terminate_relationship_predicate.execute (l_relationships.item (i), <<>>)
						if not terminate_relationship_predicate.success then
							report_error( "terminate_all_relationships", "terminate relationship failed")
						end
					end
					i := i + 1
				end
			end
		end

	relationship(a_relationship_type_id: INTEGER; a_other_entity: like Current): SUPPLIER_ENTITY_RELATIONSHIP is
		require
			valid_entity: valid
			valid_other_entity: a_other_entity.valid
			same_as_of_dates: as_of_date.is_equal(a_other_entity.as_of_date)
			a_relationship_type_id_not_zero: a_relationship_type_id/=0
		do
			if as_of_date.is_equal(constants.Maximum_date_time) and then
				relationship_type_id_is_cached(a_relationship_type_id)
			 then
				Result := relationship_cache.relationship(a_relationship_type_id, Current, a_other_entity)
			else
				select_relationship_predicate.retrieve
				(Void, <<a_relationship_type_id, entity_id, a_other_entity.entity_id, as_of_date, as_of_date.not_max>>)
				if not select_relationship_predicate.success
				 then
					report_error("relationship", "select_relationship_predicate failed.")
				end

				if select_relationship_predicate.array.count=1
				 then
					Result := select_relationship_predicate.array.item(1)
				elseif select_relationship_predicate.array.count>1
				 then
					report_error("relationship", "Too many matches.")
				end
			end
		end

	insert(a_start_date: ROSE_DATE) is
		require
			valid_entity: valid
		do
			insert_predicate.execute (Void, << Current.category_id, Current.start_date, Current.end_date>>)
			if not insert_predicate.success then report_error ("insert", "Insert failed") end
			select_identity_predicate.retrieve_one (Current, Void, Void)
			if not select_identity_predicate.success then report_error ("insert", "Select identity failed") end
			if entity_id = 0 then report_error ("insert", "Select identity returned zero") end
			if category_is_cached (category_id) then
				entity_cache.add_entity (Current)
			end

			insert_transition(ids.coverage_by_database_started_type_id, a_start_date)
		end

	insert_transition (a_transition_type_id: INTEGER; a_start_date: ROSE_DATE) is
			-- Insert a new transition corresponding to a_transition_type_id.
		require
			valid_entity: valid
			a_start_date_not_void: a_start_date /= Void
			a_transition_type_id_not_zero: a_transition_type_id /= 0
			transition_does_not_exist: transition (a_transition_type_id, a_start_date) = Void
		do
			add_transition_predicate.execute (Void, <<entity_id, a_transition_type_id, a_start_date>>)
			if not add_transition_predicate.success then
				report_error ("insert_transition", "Add transition failed (trntypeid) " + a_transition_type_id.out)
			end
		end

	make_from_entity (a_entity: like Current) is
		require
			a_entity_not_void: a_entity /= Void
			valid_entity: a_entity.valid
		do
			make_empty
			category_id := a_entity.category_id
			entity_id := a_entity.entity_id
			start_date := a_entity.start_date
			end_date := a_entity.end_date
		end

	Maximum_attribute_set_count: INTEGER is
		once
			Result := 2
		end

	attribute_set_array: ARRAY [ANY] is
		once
			create Result.make (1, ((Maximum_attribute_set_count * 3) + 2))
		end

	as_of_date_from_end_date: ROSE_DATE is
		do
			if end_date /= Void then
				create Result.make_by_date(end_date)
				if not Result.is_equal (constants.maximum_date_time) then
					Result.day_add (-1)
					if Result < start_date then
						create Result.make_by_date(start_date)
					end
				end
			end
		end

	empty_entity: like Current is
			-- Empty entity from server
		do
			Result := server.new_empty_entity
		end

	report_error (a_feature_name: STRING; a_message: STRING) is
			-- Convenience featrue for reporting error
		require
			a_feature_name_not_void: a_feature_name /= Void
			a_message_not_void: a_message /= Void
		do
			error_log.set_severity (error_log.Fatal)
			error_log.set_message (a_message)
			error_log.append_message (" (Catid, Entid) " + category_id.out + " " + entity_id.out)
			error_log.set_class_name ("SUPPLIER_ENTITY")
			error_log.set_feature_name (a_feature_name)
			error_log.put
		end

	make_empty_attribute: SUPPLIER_ENTITY_ATTRIBUTE[like Current] is
		do
			create Result.default_create
		end

	case_sensitive: BOOLEAN is
			-- Most servers.are case insensitive.  For those that are, predicates
			-- that select by 'attribute_value' in 'entity_attribute' must be made
			-- to be case sensitive.
		do
			Result := False
		end

	make_entity_cache: SUPPLIER_ENTITY_CACHE[like Current] is
		do
			Result := set_entity_cache (Void)
			if Result = Void and then set_entity_cache (Void) /= Void then
				report_error ("make_entity_cache", "Failed to make entity_cache")
			end
		end

	make_relationship_cache: SUPPLIER_ENTITY_RELATIONSHIP_CACHE[like Current] is
		do
			Result := set_relationship_cache (Void)
			if Result = Void and then set_relationship_cache (Void) /= Void then
				report_error ("make_relationship_cache", "Failed to make relationship_cache")
			end
		end

	relationship_type_id_is_cached(a_relationship_type_id: INTEGER): BOOLEAN is
		do
			Result := relationship_cache/=Void and then
			relationship_cache.relationship_type_id_is_cached(a_relationship_type_id)
		end

feature {SUPPLIER_ENTITY_CACHE} -- Caching

	set_entity_cache (a_entity_cache: SUPPLIER_ENTITY_CACHE [like Current] ):
		SUPPLIER_ENTITY_CACHE[like Current] is
		do
			Result := server.set_entity_cache(a_entity_cache)
		end

feature {SUPPLIER_ENTITY_RELATIONSHIP_CACHE} -- Caching

	set_relationship_cache (a_relationship_cache: SUPPLIER_ENTITY_RELATIONSHIP_CACHE[like Current]): SUPPLIER_ENTITY_RELATIONSHIP_CACHE[like Current] is
		do
			Result := server.set_relationship_cache(a_relationship_cache)
		end

feature {SUPPLIER_ENTITY_CACHE, SUPPLIER_ENTITY_RELATIONSHIP_CACHE, SUPPLIER_ENTITY_ATTRIBUTE} -- Server

	server: ENTITY_SERVER[like Current] is
			-- Entity server
		deferred
		end

feature {NONE} -- Caching

	entity_cache: SUPPLIER_ENTITY_CACHE [like Current] is
			-- If a SUPPLIER_ENTITY_CACHE has already been made then this function will
			-- return that cache (because SUPPLIER_ENTITY_CACHE called 'set_entity_cache(Current)' in
			-- it's 'make').  If a SUPPLIER_ENTITY_CACHE has not been made then this
			-- function will return Void.
		do
			Result := server.entity_cache
		end


	relationship_cache: SUPPLIER_ENTITY_RELATIONSHIP_CACHE [like Current] is
			-- If a SUPPLIER_ENTITY_RELATIONSHIP_CACHE has already been made then this function will
			-- return that cache (because SUPPLIER_ENTITY_RELATIONSHIP_CACHE called 'set_relationship_cache(Current)' in
			-- it's 'make').  If a SUPPLIER_ENTITY_RELATIONSHIP_CACHE has not been made then this
			-- function will return Void.
		do
			result := server.relationship_cache
		end

feature {ENTITY_SERVER} -- Predicates
	make_select_identity_predicate: COLLECTING_PREDICATE [like Current] is
		do
			create Result.make (data_server, select_identity_statement)
			if not Result.success then report_error ("make_select_identity_predicate", "Make failed") end
		end
	select_identity_statement: STRING is
		once
			Result := "select top 1 entity_id from entity_id order by entity_id DESC"
		end
	select_identity_predicate: COLLECTING_PREDICATE [like Current] is
		do
			Result := server.select_identity_predicate
		end

	make_select_next_predicate: COLLECTING_PREDICATE [like Current] is
		do
			create Result.make (data_server, select_next_statement)
			if not Result.success then report_error ("make_select_next_predicate", "Make failed") end
		end
	select_next_statement: STRING is
		once
			Result := "select top 1 * from entity_id where category_id = ? and entity_id > ? order by entity_id ASC"
		end
	select_next_predicate: COLLECTING_PREDICATE [like Current] is
		do
			Result := server.select_next_predicate
		end

	make_select_attribute_predicate: COLLECTING_PREDICATE [SUPPLIER_ENTITY_ATTRIBUTE[like Current]] is
		do
			create Result.make (data_server, select_attribute_statement)
			if not Result.success then report_error ("make_select_attribute_predicate", "Make failed") end
		end
	select_attribute_statement: STRING is
		once
			Result := "select * from entity_attribute where category_id = :category_id %
			%and entity_id = :entity_id and attribute_type_id = :attribute_type_id and %
			%start_date <= ? and end_date > ?"
		end
	select_attribute_predicate: COLLECTING_PREDICATE [SUPPLIER_ENTITY_ATTRIBUTE[like Current]] is
		do
			Result := server.select_attribute_predicate
		end

	make_select_attributes_between_predicate: COLLECTING_PREDICATE[SUPPLIER_ENTITY_ATTRIBUTE[like Current]] is
		do
			create Result.make (data_server, select_attributes_between_statement)
			if not Result.success then report_error ("make_select_attributes_between_predicate", "Make failed") end
		end
	select_attributes_between_statement: STRING is
			-- Statement to select all the attributes between two dates.
		once
			Result := "select * from entity_attribute where category_id = :category_id %
			%and entity_id = :entity_id and attribute_type_id = ? and %
			%(end_date > ? and start_date <= ?)"
		end
	select_attributes_between_predicate: COLLECTING_PREDICATE[SUPPLIER_ENTITY_ATTRIBUTE[like Current]] is
		do
			Result := server.select_attributes_between_predicate
		end

	make_select_attribute_history_predicate: COLLECTING_PREDICATE [SUPPLIER_ENTITY_ATTRIBUTE[like Current]] is
		do
			create Result.make (data_server, select_attribute_history_statement)
			if not Result.success then report_error ("make_select_attribute_history_predicate", "Make failed") end
		end
	select_attribute_history_statement: STRING is
		once
			Result := "select * from entity_attribute where category_id = ? %
			%and entity_id = ? and attribute_type_id = ? order by start_date"
		end
	select_attribute_history_predicate: COLLECTING_PREDICATE [SUPPLIER_ENTITY_ATTRIBUTE[like Current]] is
		do
			Result := server.select_attribute_history_predicate
		end

	select_entity_statement: STRING is
		once
			Result := "select * from entity_id where entity_id = ?"
		end
	select_entity_predicate: COLLECTING_PREDICATE [like Current] is
		do
			Result := server.select_entity_predicate
		end
	make_select_entity_predicate: COLLECTING_PREDICATE [like Current] is
		do
			create Result.make (data_server, select_entity_statement)
			if not Result.success then report_error ("make_select_entity_predicate", "ERROR: Could not make predicate") end
		end

	make_select_next_attribute_predicate: COLLECTING_PREDICATE [SUPPLIER_ENTITY_ATTRIBUTE[like Current]] is
		do
			create Result.make (data_server,  select_next_attribute_statement)
			if not Result.success then report_error ("make_select_next_attribute_predicate", "Make failed") end
		end
	select_next_attribute_statement: STRING is
		once
			Result := "select top 1 * from entity_attribute where category_id = :category_id and %
			%entity_id = :entity_id and attribute_type_id = :attribute_type_id and start_date > ? order by start_date ASC"
		end
	select_next_attribute_predicate: COLLECTING_PREDICATE [SUPPLIER_ENTITY_ATTRIBUTE[like Current]] is
		do
			Result := server.select_next_attribute_predicate
		end

	make_select_transition_predicate: COLLECTING_PREDICATE [ENTITY_TRANSITION] is
		do
			create Result.make (data_server, select_transition_statement)
			if not Result.success then report_error ("make_select_transition_predicate", "Make failed") end
		end
	select_transition_statement: STRING is
		once
			Result := "select * from entity_transition where entity_id = ? and transition_type_id = ? and transition_date = ?"
		end
	select_transition_predicate: COLLECTING_PREDICATE [ENTITY_TRANSITION] is
		do
			Result := server.select_transition_predicate
		end

	make_add_transition_predicate: EXECUTION_PREDICATE is
		do
			create Result.make (data_server, add_transition_statement)
			if not Result.success then report_error ("make_add_transition_predicate", "Make failed") end
		end
	add_transition_statement: STRING is
		once
			Result := "insert into entity_transition (entity_id, transition_type_id, transition_date) values (?, ?, ?)"
		end
	add_transition_predicate: EXECUTION_PREDICATE is
		do
			Result := server.add_transition_predicate
		end

	make_update_transition_predicate: EXECUTION_PREDICATE is
		do
			create Result.make (data_server, update_transition_statement)
			if not Result.success then report_error ("make_update_transition_predicate", "Make failed") end
		end
	update_transition_statement: STRING is
		once
			Result := "update entity_transition set transition_date = ? %
			%where entity_id = :entity_id and %
			%transition_type_id = :transition_type_id and %
			%transition_date = :transition_date"
		end
	update_transition_predicate: EXECUTION_PREDICATE is
		do
			Result := server.update_transition_predicate
		end

	make_update_transition_event_id_predicate: EXECUTION_PREDICATE is
		do
			create Result.make (data_server, update_transition_event_id_statement)
			if not Result.success then report_error ("make_update_transition_event_id_predicate", "Make failed") end
		end
	update_transition_event_id_statement: STRING is
		once
			Result := "update entity_transition set transition_event_id = ? %
			%where entity_id = :entity_id and %
			%transition_type_id = :transition_type_id and %
			%transition_date = :transition_date"
		end
	update_transition_event_id_predicate: EXECUTION_PREDICATE is
		do
			Result := server.update_transition_event_id_predicate
		end

	make_select_transition_history_predicate(a_count: INTEGER): COLLECTING_PREDICATE [ENTITY_TRANSITION] is
		do
			create Result.make (data_server, select_transition_history_statement(a_count))
			if not Result.success then report_error ("make_select_transition_history_predicate", "Make failed") end
		end
	select_transition_history_statement(a_count: INTEGER): STRING is
		local
			i: INTEGER
		do
			if a_count=0
			 then
				Result := "select * from entity_transition where entity_id = ? order by transition_date"
			else
				Result := "select * from entity_transition where entity_id = ? and transition_type_id in ("
				from
					i := 1
				until
					i>a_count
				loop
					if i/=a_count
					 then
						Result.append("?,")
					else
						Result.append("?")
					end

					i := i + 1
				end

				Result.append(") order by transition_date")
			end
		end
	select_transition_history_predicate(a_count: INTEGER): COLLECTING_PREDICATE [ENTITY_TRANSITION] is
		do
			Result := server.select_transition_history_predicate(a_count)
		end

	make_select_attribute_by_value_predicate: COLLECTING_PREDICATE [like Current] is
		do
			create Result.make (data_server, select_attribute_by_value_statement)
			if not Result.success then report_error ("make_select_attribute_by_value_predicate", "Make failed") end
		end
	select_attribute_by_value_statement: STRING is
		do
			if case_sensitive
			 then
				Result := "select entity_id from entity_attribute_view where category_id = ? and %
				%attribute_type_id = ? and attribute_value = ? and start_date <= ? and end_date > ? and %
				%binary_attribute_value = CONVERT(BINARY, ?3)"
			else
				Result := "select entity_id from entity_attribute where category_id = ? and %
				%attribute_type_id = ? and attribute_value = ? and start_date <= ? and end_date > ?"
			end
		end
	select_attribute_by_value_predicate: COLLECTING_PREDICATE [like Current] is
		do
			Result := server.select_attribute_by_value_predicate
		end

	make_select_most_recent_attribute_by_value_predicate: COLLECTING_PREDICATE [like Current] is
		do
			create Result.make (data_server, select_most_recent_attribute_by_value_statement)
			if not Result.success then report_error ("make_select_most_recent_attribute_by_value_predicate", "Make failed") end
		end
	select_most_recent_attribute_by_value_statement: STRING is
		do
			if case_sensitive
			 then
				Result := "select top 1 entity_id from entity_attribute_view %
				%where category_id = ? and attribute_type_id = ? and attribute_value = ? and %
				%binary_attribute_value = CONVERT(BINARY, ?3) %
				%order by start_date desc"
			else
				Result := "select top 1 entity_id from entity_attribute %
				%where category_id = ? and attribute_type_id = ? and attribute_value = ? %
				%order by start_date desc"
			end
		end
	select_most_recent_attribute_by_value_predicate: COLLECTING_PREDICATE [like Current] is
		do
			Result := server.select_most_recent_attribute_by_value_predicate
		end

	make_select_attribute_by_historical_value_predicate: COLLECTING_PREDICATE [like Current] is
		do
			create Result.make (data_server, select_attribute_by_historical_value_statement)
			if not Result.success then report_error ("make_select_attribute_by_historical_value_predicate", "Make failed") end
		end
	select_attribute_by_historical_value_statement: STRING is
		do
			if case_sensitive
			 then
				Result := "select top 1 entity_id, end_date as as_of_date from entity_attribute_view where category_id = ? and %
				%attribute_type_id = ? and attribute_value = ? and %
				%binary_attribute_value = CONVERT(BINARY, ?3) %
				%order by end_date DESC"
			else
				Result := "select top 1 entity_id, end_date as as_of_date from entity_attribute where category_id = ? and %
				%attribute_type_id = ? and attribute_value = ? %
				%order by end_date DESC"
			end
		end
	select_attribute_by_historical_value_predicate: COLLECTING_PREDICATE [like Current] is
		do
			Result := server.select_attribute_by_historical_value_predicate
		end

	make_select_attribute_by_value_set_predicate: COLLECTING_PREDICATE [like Current] is
		do
			create Result.make (data_server, select_attribute_by_value_set_statement)
			if not Result.success then report_error ("make_select_attribute_by_value_set_predicate", "Make failed") end
		end
	select_attribute_by_value_set_statement: STRING is
		do
			if case_sensitive
			 then
				Result := "select %
				%	ea1.entity_id, ea1.end_date %
				% from %
				%	entity_attribute_view as ea1, %
				%	entity_attribute_view as ea2 %
				% where %
				%	ea1.entity_id = ea2.entity_id and %
				%	ea1.start_date <= ? and %
				%	ea1.end_date > ? and %
				%	ea1.category_id = ? and %
				%	ea1.attribute_type_id = ? and %
				%	ea1.attribute_value = ? and %
				%	ea1.binary_attribute_value = CONVERT(BINARY, ?5) and %
				%	ea2.category_id = ? and %
				%	ea2.attribute_type_id = ? and %
				%	ea2.attribute_value = ? and %
				%	ea2.binary_attribute_value = CONVERT(BINARY, ?8)"
			else
				Result := "select %
				%	ea1.entity_id, ea1.end_date %
				% from %
				%	entity_attribute as ea1, %
				%	entity_attribute as ea2 %
				% where %
				%	ea1.entity_id = ea2.entity_id and %
				%	ea1.start_date <= ? and %
				%	ea1.end_date > ? and %
				%	ea1.category_id = ? and %
				%	ea1.attribute_type_id = ? and %
				%	ea1.attribute_value = ? and %
				%	ea2.category_id = ? and %
				%	ea2.attribute_type_id = ? and %
				%	ea2.attribute_value = ?"
			end
		end
	select_attribute_by_value_set_predicate: COLLECTING_PREDICATE [like Current] is
		do
			Result := server.select_attribute_by_value_set_predicate
		end

	make_select_relationships_1_predicate: COLLECTING_PREDICATE [SUPPLIER_ENTITY_RELATIONSHIP] is
		do
			create Result.make (data_server, select_relationships_1_statement)
			if not Result.success then report_error ("make_select_relationships_1_predicate", "Make failed") end
		end
	select_relationships_1_statement: STRING is
		once
			Result := "select * from entity_relationship where %
			%relationship_type_id = ? and %
			%entity_2_id = ? and %
			%start_date <= ? and %
			%end_date > ?"
		end
	select_relationships_1_predicate: COLLECTING_PREDICATE [SUPPLIER_ENTITY_RELATIONSHIP] is
		do
			Result := server.select_relationships_1_predicate
		end

	make_select_relationships_2_predicate: COLLECTING_PREDICATE [SUPPLIER_ENTITY_RELATIONSHIP] is
		do
			create Result.make (data_server, select_relationships_2_statement)
			if not Result.success then report_error ("make_select_relationships_2_predicate", "Make failed") end
		end
	select_relationships_2_statement: STRING is
		once
			Result := "select * from entity_relationship where %
			%relationship_type_id = ? and %
			%entity_1_id = ? and %
			%start_date <= ? and %
			%end_date > ?"
		end
	select_relationships_2_predicate: COLLECTING_PREDICATE [SUPPLIER_ENTITY_RELATIONSHIP] is
		do
			Result := server.select_relationships_2_predicate
		end

	make_select_relationship_predicate: COLLECTING_PREDICATE [SUPPLIER_ENTITY_RELATIONSHIP] is
		do
			create Result.make (data_server, select_relationship_statement)
			if not Result.success
			 then
				report_error ("make_select_relationship_predicate", "Make failed")
			end
		end
	select_relationship_statement: STRING is
		once
			Result := "select * from entity_relationship where %
			%relationship_type_id = ? and %
			%entity_1_id = ? and %
			%entity_2_id = ? and %
			%start_date <= ? and %
			%end_date > ?"
		end
	select_relationship_predicate: COLLECTING_PREDICATE [SUPPLIER_ENTITY_RELATIONSHIP] is
		do
			Result := server.select_relationship_predicate
		end

	make_select_all_relationships_predicate: COLLECTING_PREDICATE [SUPPLIER_ENTITY_RELATIONSHIP] is
		do
			create Result.make (data_server,  select_all_relationships_statement)
			if not Result.success then report_error ("make_select_all_relationships_predicate", "Make failed") end
		end
	select_all_relationships_statement: STRING is
		once
			Result := "select * from entity_relationship where ((entity_1_id = ?) or (entity_2_id = ?)) and end_date = '"
			Result.append (constants.Maximum_date_time.date.ccyymmdd.out)
			Result.append ("'")
		end
	select_all_relationships_predicate: COLLECTING_PREDICATE [SUPPLIER_ENTITY_RELATIONSHIP] is
		do
			Result := server.select_all_relationships_predicate
		end

	make_insert_relationship_predicate: EXECUTION_PREDICATE is
		do
			if has_weighted_relationship then
				create Result.make (data_server, insert_weighted_relationship_statement)
			else
				create Result.make (data_server, insert_relationship_statement)
			end
			if not Result.success then report_error ("make_insert_relationship_predicate", "Make failed") end
		end
	insert_relationship_statement: STRING is
		once
			Result := "insert into entity_relationship (%
					  %relationship_id, %
					  %relationship_type_id, %
					  %entity_1_id, %
					  %entity_2_id, %
					  %start_date, %
					  %end_date) values (%
					  % ?, ?, ?, ?, ?, ?)"
					  -- 				   %:relationship_id, %
					  -- 				   %:relationship_type_id, %
					  -- 				   %:entity_1_id, %
					  -- 				   %:entity_2_id, %
					  -- 				   %:start_date, %
					  -- 				   %:end_date)"
		end
	insert_weighted_relationship_statement: STRING is
		once
			Result := "insert into entity_relationship (%
			%relationship_id, %
			%relationship_type_id, %
			%entity_1_id, %
			%entity_2_id, %
			%start_date, %
			%end_date, %
			%weight) values (%
			% ?, ?, ?, ?, ?, ?, ?)"
			-- 				   %:relationship_id, %
			-- 				   %:relationship_type_id, %
			-- 				   %:entity_1_id, %
			-- 				   %:entity_2_id, %
			-- 				   %:start_date, %
			-- 				   %:end_date, %
			-- 				   %:weight)"
		end
	insert_relationship_predicate: EXECUTION_PREDICATE is
		do
			Result := server.insert_relationship_predicate
		end

	make_terminate_relationship_predicate: EXECUTION_PREDICATE is
		do
			create Result.make (data_server, terminate_relationship_statement)
			if not Result.success then report_error ("make_terminate_relationship_predicate", "Make failed") end
		end
	terminate_relationship_statement: STRING is
		once
			Result := "update entity_relationship set end_date = :end_date where %
			%relationship_type_id = :relationship_type_id and %
			%entity_1_id = :entity_1_id and %
			%entity_2_id = :entity_2_id and %
			%start_date = :start_date"
		end
	terminate_relationship_predicate: EXECUTION_PREDICATE is
		do
			Result := server.terminate_relationship_predicate
		end

	make_update_relationship_predicate: EXECUTION_PREDICATE is
		do
			create Result.make (data_server, update_relationship_statement)
			if not Result.success then report_error ("make_update_relationship_predicate", "Make failed") end
		end
	update_relationship_statement: STRING is
		once
			Result := "update entity_relationship set end_date = ? where %
			%relationship_type_id = ? and %
			%entity_1_id = ? and %
			%entity_2_id = ? and %
			%end_date = ?"
		end
	update_relationship_predicate: EXECUTION_PREDICATE is
		do
			Result := server.update_relationship_predicate
		end

	make_update_relationship_id_predicate: EXECUTION_PREDICATE is
		do
			create Result.make (data_server, update_relationship_id_statement)
			if not Result.success then report_error ("make_update_relationship_id_predicate", "Make failed") end
		end
	update_relationship_id_statement: STRING is
		once
			Result := "update entity_relationship set relationship_id = :relationship_id where %
			%relationship_type_id = relationship_type_id and %
			%entity_1_id = :entity_1_id and %
			%entity_2_id = :entity_2_id and %
			%start_date = :start_date and %
			%end_date = :end_date"
		end
	update_relationship_id_predicate: EXECUTION_PREDICATE is
		do
			Result := server.update_relationship_id_predicate
		end

	make_delete_relationships_predicate: EXECUTION_PREDICATE is
		do
			create Result.make(data_server, delete_relationships_statement)
			if not Result.success then
				report_error("make_delete_relationships_predicate", "Make failed")
			end
		end

	delete_relationships_statement: STRING is
		once
			Result :=
				"delete from entity_relationship %
			%where %
			%relationship_type_id=?1 and %
			%entity_1_id=:entity_id and %
			%entity_2_id=?2"
		end

	delete_relationships_predicate: EXECUTION_PREDICATE is
		do
			Result := server.delete_relationships_predicate
		end

	make_delete_attributes_predicate: EXECUTION_PREDICATE is
		do
			create Result.make (data_server, delete_attributes_statement)
			if not Result.success then report_error ("make_delete_attributes_predicate", "Make failed") end
		end

	make_delete_attribute_history_row_predicate	: EXECUTION_PREDICATE is
		do
			create Result.make (data_server, delete_attribute_history_row_predicate_statement)
			if not Result.success then report_error ("make_delete_attribute_history_row_predicate", "Make failed") end
		end

	delete_attributes_statement: STRING is
		once
			Result := "delete from entity_attribute where category_id = :category_id and entity_id = :entity_id"
		end

	delete_attribute_history_row_predicate_statement : STRING is
		once
			Result := "delete from entity_attribute where "
			Result.append ("category_id = :category_id and entity_id = :entity_id and ")
			Result.append ("attribute_type_id = :attribute_type_id and attribute_value = :attribute_value and ")
			Result.append ("start_date = :start_date and end_date = :end_date")
		end

	delete_attributes_predicate: EXECUTION_PREDICATE is
		do
			Result := server.delete_attributes_predicate
		end

	delete_attribute_history_row_predicate: EXECUTION_PREDICATE is
		do
			Result := server.delete_attribute_history_row_predicate
		end

	make_delete_predicate: EXECUTION_PREDICATE is
		do
			create Result.make (data_server, delete_statement)
			if not Result.success then report_error ("make_delete_predicate", "Make failed") end
		end
	delete_statement: STRING is
		once
			Result := "delete from entity_id where category_id = :category_id and entity_id = :entity_id"
		end
	delete_predicate: EXECUTION_PREDICATE is
		do
			Result := server.delete_predicate
		end

	make_terminate_predicate: EXECUTION_PREDICATE is
		do
			create Result.make (data_server, terminate_statement)
			if not Result.success then report_error ("make_terminate_predicate", "Make failed") end
		end
	terminate_statement: STRING is
		once
			Result := "update entity_id set end_date = ? where category_id = ? and entity_id = ?"
		end
	terminate_predicate: EXECUTION_PREDICATE is
		do
			Result := server.terminate_predicate
		end

	select_statement: STRING is
		once
			Result := "select * from entity_id where category_id = ? and entity_id = ?"
		end

	insert_statement: STRING is
		once
			Result := "insert into entity_id (category_id, start_date, end_date) values (?, ?, ?)"
		end

	update_statement: STRING is
		once
			Result := "update entity_id set start_date = :start_date, end_date = :end_date %
			%where category_id=:category_id and entity_id=:entity_id"
		end

feature {NONE} -- Attributes

	historical_relative_lookup: BOOLEAN

	attribute_type_manager: SUPPLIER_ATTRIBUTE_TYPE_MANAGER is
			-- DB where attribute types are stored
		do
			Result := server.attribute_type_manager
		end

	company_security_relationship_type_id: INTEGER is
			-- Company / Security relationship type-id
		once
			Result := ids.company_security_relationship_type_id
		end

	security_listing_relationship_type_id: INTEGER is
			-- Convenience feature (though it does not seem to help much)
		once
			Result := ids.security_listing_relationship_type_id
		end

	today: ROSE_DATE is
			-- Today's date.
		do
			create Result.make_now
		end

feature {SUPPLIER_ENTITY, INTEGRATION_SUPPLIER_ENTITY, ROSE_PLATFORM_RELATIONSHIP_EDITING_PANE} -- Relationships

	has_weighted_relationship: BOOLEAN is
			-- Do the relationship tables have 'weight' as a column?
		local
			l_predicate: ANONYMOUS_PREDICATE
		do
			weighted_relationship_cache.search (data_server.database + data_server.node)
			if not weighted_relationship_cache.found then
				create l_predicate.make_using_server(data_server,"exec sp_columns @table_name = N'entity_relationship',@column_name = N'weight'")
				l_predicate.retrieve (Void,Void)
				weighted_relationship_cache.force (not l_predicate.array.is_empty, data_server.database + data_server.node)
				weighted_relationship_cache.search (data_server.database + data_server.node)
			end
			Result := weighted_relationship_cache.found_item
		end

	weighted_relationship_cache: HASH_TABLE [BOOLEAN, STRING] is
			-- Cache of determined weighted relationship availability by server
		once
			create Result.make (10)
		end

feature {NONE} -- Agents
	compare_attributes_by_start_date (a_first, a_second : SUPPLIER_ENTITY_ATTRIBUTE [like Current]) : BOOLEAN is
			-- Sort
		do
			Result := a_first.start_date < a_second.start_date
		end

feature {NONE} -- Types
	attribute_type : SUPPLIER_ENTITY_ATTRIBUTE [like Current]
end
