note

	description:

		"Symmetric Matrix, i.e. matrix which is equal to its transpose"

	author: "Michael Zhang"
	last_editor: "$Author: myeager $"
	date: "$Date: 2015/03/19 $"
	revision: "$Revision: #57 $"
	path: "$File: //rose/source/kernel/infrastructure/math/symmetric_matrix.e $"
	storable_version: ""

class SYMMETRIC_MATRIX

inherit
	ABSTRACT_MATRIX
		rename
			make as matrix_make
		redefine
			add,
			to_vector,
			to_coordinated_matrix,
			is_identity,
			is_square,
			is_same,
			is_symmetric,
			pseudo_inverse,
			inverse,
			scale,
			scalar_multiply,
			scaled,
			transposed,
			subtract
		end
	LABELED_2D
		undefine
			set_column_names,
			set_row_names
		end
	USES_2D_SPECIAL_DATA
		rename
			make as matrix_make
		undefine
			set_column_names,
			set_row_names
		redefine
			make_identity,
			item,
			put
		end
	TRIANGULAR_LINEARIZATION
	ROSE_STORABLE
		rename
			trace as global_trace,
			output as global_output
		end
	MISMATCH_CORRECTOR
		export
			{NONE} all
		redefine
			correct_mismatch
		end

create

	make,
	make_xtx,
	make_from_triangle,
	make_from_array,
	make_xtdx,
	make_xplusxt,
	make_as_inverse,
	make_from_tab_delimited_text_file

create {SYMMETRIC_MATRIX}

	make_identity

create {SYMMETRIC_MATRIX, COMPANY_COVARIANCE_MATRIX}

	matrix_make

feature {NONE} -- Initialization

	make (a_size : INTEGER)
			-- Create zeroed matrix with a fixed `a_size' as `height' and `width'.
		local
			i: INTEGER
		do
			height := a_size
			width := a_size
				-- Lower triangle implementation
			create data.make_filled (Void, height + 1)
			from i := 1
			until i > height
			loop
				data.put (create {SPECIAL [DOUBLE]}.make_filled (0.0, i + 1), i)
				i := i + 1
			end
		end

	make_from_triangle (a_matrix : ABSTRACT_READABLE_MATRIX; a_lower_triangle: BOOLEAN)
			-- Create from a triangular matrix `a_matrix'
		local
			i, j: INTEGER
		do
			if a_matrix /= Void and then a_matrix.is_square then
				make (a_matrix.height)
				from i := 1 until i > height loop
					from j := 1 until j > i loop
						if a_lower_triangle then
							put (a_matrix [i, j],i, j)
						else
							put (a_matrix [j, i],i, j)
						end
						j := j + 1
					end
					i := i + 1
				end
			end
		end

	make_from_array (a_array : ARRAY [ARRAY [DOUBLE]])
			-- Create from `a_array' of array of doubles
		require
			non_void_array: a_array /= Void and then not a_array.has (Void)
		local
			i, j: INTEGER
		do
			make (a_array.count)
			from i := 1
			until i > height
			loop
				from j := 1
				until j > i
				loop
					put (a_array.item (i).item (j), i, j)
					j := j + 1
				end
				i := i + 1
			end
		end

	make_xtx (a_m: ABSTRACT_READABLE_MATRIX)
			-- Create transposed product (x^t x) from matrix `a_m'.
		require
			non_void_matrix: a_m /= Void
		local
			i, j: INTEGER
			l_columns: ROSE_LINEAR_ARRAY [VECTOR]
		do
			from
				create l_columns.make (1, a_m.width)
				i := 1
			until
				i > a_m.width
			loop
				l_columns.put (a_m.column (i), i)
				i := i + 1
			end
			make (a_m.width)
			from i := 1 until i > width loop
				from j := 1 until j > i loop
					Current[i, j] := l_columns [i].dot_product (l_columns [j])
					j := j + 1
				end
				i := i + 1
			end
		end

	make_xtDx(a_m : ABSTRACT_READABLE_MATRIX; a_diagonal : ABSTRACT_READABLE_VECTOR) is
			-- make xtDx
		require
			non_void_matrix: a_m /= Void
		local
			i,j,k : INTEGER
			l_dot_product : DOUBLE
		do
			make(a_m.width)
			from i := 1 until i > width loop
				from j := 1 until j > i loop
					l_dot_product := 0.0
					from k := 1 until k > a_m.height loop
						l_dot_product := l_dot_product + a_m[k, i] * a_m[k, j] * a_diagonal.item (k)
						k := k + 1
					end
					Current[i, j] := l_dot_product
					j := j + 1
				end
				i := i + 1
			end
		end

	make_xxt(a_m : ABSTRACT_READABLE_MATRIX) is
			-- make xxt
		require
			non_void_matrix: a_m /= Void
		local
			i,j,k : INTEGER
			l_dot_product : DOUBLE
		do
			make(a_m.height)
			from i := 1 until i > height loop
				from j := 1 until j > i loop
					l_dot_product := 0.0
					from k := 1 until k > a_m.width loop
						l_dot_product := l_dot_product + a_m[i, k] * a_m[j, k]
						k := k + 1
					end
					Current[i, j] := l_dot_product
					j := j + 1
				end
				i := i + 1
			end
		end

	make_xPLUSxt(a_x : ABSTRACT_READABLE_MATRIX) is
			-- make xPLUSxt
		require
			square : a_x /= Void and then a_x.is_square
		local
			i,j : INTEGER
		do
			make(a_x.height)
			from i := 1 until i > height loop
				from j := 1 until j > i loop
					Current[i, j] := a_x[i, j] + a_x [j, i]
					j := j + 1
				end
				i := i + 1
			end
		end


	make_identity (a_count: INTEGER) is
			-- make identity
		local
			i: INTEGER
		do
			make (a_count)
			from i := 1 until i > a_count loop
				put (1, i, i)
			i := i + 1
			end
		end

feature -- Basic operations

	wipeout is
			-- Set all data items to zero.
			local
				i: INTEGER
			do
				from i:= 1 until i> height loop
					data.item(i).fill_with (0.0, 0, data.item(i).upper)
					i:= i+ 1
				end
			end


	item alias "[]" (i, j: INTEGER): DOUBLE assign put is
			-- item
		do
			if	i >= j then
				Result := data.item(i).item(j)
			else
				Result := data.item(j).item(i)
			end
		end

	is_symmetric : BOOLEAN is TRUE
			-- Always true by construction

	is_square : BOOLEAN is TRUE
			-- Symmetric matrix is always square

	is_identity: BOOLEAN is
			-- Is `Current' a good approximation to an identity matrix?
		local
			i,j: INTEGER
		do
			Result := height = width
			from i := 1 until i > height or not Result loop
				from j := 1 until j > i loop
					Result := Result and is_identity_item (i, j)
					j := j + 1
				end
				i := i + 1
			end
		end

	is_same (a_other: ABSTRACT_READABLE_MATRIX): BOOLEAN is
			-- Is `a_other' numerically equal to `Current'?
		local
			i, j: INTEGER
		do
			if attached {SYMMETRIC_MATRIX} a_other as l_other then
				Result := height = a_other.height and width = a_other.width
				from i := 1 until i > height or not Result loop
					from j := 1 until j > i or not Result loop
						Result := Current [i, j] = a_other [i, j]
						j := j + 1
					end
					i := i + 1
				end
			else
				Result := Precursor (a_other)
			end
		end

	cloned : like Current is
			-- cloned
		local
			i, j: INTEGER
		do
			create Result.make (height)
			from i := 1 until i > height loop
				from j := 1 until j > i loop
					Result[i, j] := Current [i, j]
					j := j + 1
				end
				i := i + 1
			end
			Result.set_name (name)
			Result.set_names (row_names)
		end

	to_symmetric_matrix : SYMMETRIC_MATRIX is
			-- to symmetric matrix
		local
			i, j: INTEGER
		do
			create Result.make (height)
			from i := 1 until i > height loop
				from j := 1 until j > i loop
					Result[i, j] := Current [i, j]
					j := j + 1
				end
				i := i + 1
			end
			Result.set_name (name)
			Result.set_names (row_names)
		end

	to_vector : VECTOR is
		-- columnwize vectorization, lower triangle
		local
			i, j, l_count : INTEGER
		do
			create Result.make(width * (width + 1) // 2)
			from i := 1 until i > width loop
				from j := i until j > width loop
					l_count := l_count + 1
				--	Result.put(item(j, i), l_count)
					Result[l_count] := Current[j, i]
					j := j + 1
				end
				i := i + 1
			end
		end

	to_coordinated_matrix(a_is_flip : BOOLEAN) : MATRIX is
			-- convert to a three column matrix with (i, j, value) for each row
		local
			i, j, l_row : INTEGER
		do
			create Result.make(width * (width + 1) // 2, 3)
			from i := 1 until i > height loop
				from j := 1 until j > i loop
					l_row := l_row + 1
					if a_is_flip then
						Result[l_row, 2] := height - i + 1
					else
						Result[l_row, 2] := i
					end
					Result[l_row, 1] := j
					Result[l_row, 3] := Current[i, j]
					j := j + 1
				end
				i := i + 1
			end
		end

	to_xp_vector_no_diagonal : VECTOR is
		-- columnwise vectorization, upper triangle
		local
			i, j, l_count : INTEGER
		do
			create Result.make(xp_count(width))
			from j := 2 until j >  width loop
				from i := 1 until i >= j loop
					l_count := l_count + 1
					Result.put(Current [i, j], l_count)
					i := i + 1
				end
				j := j + 1
			end
		end

	to_xp_vector : VECTOR is
		-- columnwise vectorization, upper triangle
		local
			i, j, l_count : INTEGER
		do
			create Result.make(linear_count(width))
			from j := 1 until j >  width loop
				from i := 1 until i > j loop
					l_count := l_count + 1
					Result.put(Current [i, j], l_count)
					i := i + 1
				end
				j := j + 1
			end
		end

	transposed : SYMMETRIC_MATRIX is
			-- Transpose is identity for symmetric matrix
			-- (returns Current, not a clone)
		do
			Result := Current
		end

	set_names(a_names : ARRAY[STRING]) is
			-- set row and col names
		do
			set_row_names (a_names)
			set_column_names (a_names)
		end

feature -- transformation

	sqrt_matrix : SYMMETRIC_MATRIX is
		-- only valid for semi-positive definite symmetric matrix (like covariance matrix)
		local
			l_eigensystem : like eigensystem
			l_weighted_matrix : WEIGHTED_MATRIX
		do
			l_eigensystem := eigensystem
			if l_eigensystem.eigenvalues.min >= 0.0 then
				create l_weighted_matrix.make(l_eigensystem.eigenvectors, l_eigensystem.eigenvalues.square_roots.square_roots)
				create Result.make_xtx(l_weighted_matrix)
			else
				print("matrix is not semi positive definite")
			end
		end

	to_spd_matrix : SYMMETRIC_MATRIX is
			-- transform to semi-positive definte matrix
		local
			l_eigenvalues : VECTOR
			l_eigensystem : like eigensystem
			l_weighted_matrix : WEIGHTED_MATRIX
		do
			l_eigensystem := eigensystem
			if l_eigensystem.eigenvalues.min >= 0.0 then
				Result := Current
			else
				l_eigenvalues := l_eigensystem.eigenvalues.to_vector
				l_eigenvalues.replace_negative
				create l_weighted_matrix.make(l_eigensystem.eigenvectors, l_eigenvalues.square_roots)
				create Result.make_xtx(l_weighted_matrix)
			end
		end

	pseudo_inverse : SYMMETRIC_MATRIX is
			-- pseudo inverse
		local
			l_matrix : ABSTRACT_MATRIX
		do
			l_matrix := svd.pseudo_inverse
			if l_matrix /= Void then
				create Result.make_from_triangle(l_matrix, False)
			end
		end

	pseudo_inverse_from_eigenvalues (a_tolerance: DOUBLE): SYMMETRIC_MATRIX is
			-- Pseudo-inverse of `Current' computed from eigenvalues;
			-- Eigenvalues whose absolute value is below `a_tolerance' are treated as zero.
		local
			l_matrix: SYMMETRIC_MATRIX
			l_esys: RP_EIGENSYSTEM
			l_diag: DIAGONAL_MATRIX
		do
			l_matrix := twin
			l_matrix.replace_diagonal
			l_esys := l_matrix.eigensystem
			l_esys.eigenvalues.zero_if_abs_below_tolerance (a_tolerance)
			create l_diag.make_from_vector (l_esys.eigenvalues.inverse)
			Result := l_diag.xtdx (l_esys.eigenvectors.transpose)
		ensure
			inverted_matrix_not_void: Result /= Void
		end

	ptSp_transformation(a_p : ABSTRACT_READABLE_MATRIX) : SYMMETRIC_MATRIX is
		-- transformation by matrix p, of the form p'Sp
		require
			valid_dim : a_p.height = height
		local
			l_matrix : MATRIX
		do
			l_matrix := a_p.transposed * product(a_p)
			create Result.make_from_triangle(l_matrix, True)
		end

	pSpt_transformation(a_p : ABSTRACT_READABLE_MATRIX) : SYMMETRIC_MATRIX is
		-- transformation by matrix p, of the form pSp'
		require
			valid_dim : a_p.width = height
		local
			l_matrix : MATRIX
		do
			l_matrix := a_p * product(a_p.transposed)
			create Result.make_from_triangle(l_matrix, True)
		end

	dSd(a_diagonal : VECTOR) : SYMMETRIC_MATRIX
			-- D * S * D
		local
			i, j : INTEGER
		do
			Result := cloned
			from i := 1 until i > height loop
				from j := 1 until j > i loop
					Result[i, j] := Result[i, j] * a_diagonal[i] * a_diagonal[j]
					j := j + 1
				end
				i := i + 1
			end
		end

feature -- Changes

	put_symmetric_matrix (a_matrix: SYMMETRIC_MATRIX; a_row_ix: INTEGER)
			-- Insert `a_matrix' at `a_row_ix' of `Current'
		local
			i, j: INTEGER
		do
			from i := 1
			until i > a_matrix.height or (i + a_row_ix - 1) > height
			loop
				from j := 1
				until j > i
				loop
						--	put (a_matrix [i, j], i + a_row_ix - 1, j + a_row_ix - 1)
					Current [i + a_row_ix - 1, j + a_row_ix - 1] := a_matrix [i, j]
					j := j + 1
				end
				i := i + 1
			end
		end

	tridiagonal: MATRIX
			-- tridiagonal
			-- No implementation (d3 6 Oct 2009)
		do
		end

	add(a_other : like Current) is
			-- add a symmetric matrix
		local
			i, j: INTEGER
			l_double : DOUBLE
		do
			from i := 1 until i > height loop
				from j := 1 until j > i loop
					l_double := Current[i, j] + a_other[i, j]
--					put (l_double, i, j)
					Current[i, j] := l_double;
					j := j + 1
				end
				i := i + 1
			end
		end

	subtract(a_other : like Current) is
			-- subtract a symmetric matrix
		local
			i, j: INTEGER
			l_double : DOUBLE
		do
			from i := 1 until i > height loop
				from j := 1 until j > i loop
					l_double := Current[i, j] - a_other[i, j]
				--	put (l_double, i, j)
					Current[i, j] := l_double
					j := j + 1
				end
				i := i + 1
			end
		end

	scale (a_scale : DOUBLE) is
			-- scale
		local
			i,j : INTEGER
		do
			from i := 1 until i > height loop
				from j := 1 until j > i loop
				--	put(item(i,j) * a_scale,i,j)
					Current [i, j] := Current [i, j] * a_scale
					j := j + 1
				end
				i := i + 1
			end
		end

	add_diagonal(a_diagonal : VECTOR) is
			-- add to diagonal
		require
			same_size : a_diagonal /= Void and then a_diagonal.count = width
		local
			i : INTEGER
		do
			from i := 1 until i > width loop
			--	put(item(i, i) + a_diagonal[i], i, i)
				Current [i, i] := Current [i, i] + a_diagonal[i]
				i := i + 1
			end
		end

	set_diagonal(a_diagonal : VECTOR) is
			-- set diagonal
		require
			same_size : a_diagonal /= Void and then a_diagonal.count = width
		local
			i : INTEGER
		do
			from i := 1 until i > width loop
			--	put(a_diagonal[i], i, i)
				Current[i, i] := a_diagonal[i]
				i := i + 1
			end
		end

	zero_out_off_diagonal is
			-- zero out off diagonal
		local
			i,j : INTEGER
		do
			from i := 1 until i > height loop
				from j := 1 until j >= i loop
					Current [i, j] := 0.0
					j := j + 1
				end
				i := i + 1
			end
		end

feature -- Submatrix operations

	sub_symmetric_matrix(a_ixs : ARRAY[INTEGER]) : SYMMETRIC_MATRIX is
			-- sub symmetric matrix
		local
			l_matrix : MATRIX
		do
			create l_matrix.make_sub_matrix_3(Current,a_ixs, a_ixs)
			create Result.make_from_triangle(l_matrix, False)
		end

	sub_symmetric_matrix_2(a_r1, a_r2 : INTEGER) : SYMMETRIC_MATRIX is
			-- sub symmetric matrix 2
		local
			i, j : INTEGER
		do
			create Result.make(a_r2 - a_r1 + 1)
			from i := a_r1 until i > a_r2 loop
				from j := a_r1 until j > i loop
				--	Result.put(item(i, j), i - a_r1 + 1, j - a_r1 + 1)
					Result [i - a_r1 + 1, j - a_r1 + 1] := Current [i, j]
					j := j + 1
				end
				i := i + 1
			end
		end

	condition_number : DOUBLE is
			-- Returns the ratio of the largest to smallest magnitude of eigenvalue
		require
			square : is_square
		local
			l_abs_vector: VECTOR
			l_max_magnitude: DOUBLE
			l_min_magnitude: DOUBLE
			l_eigenvalues : VECTOR
		do
			l_eigenvalues := eigensystem.eigenvalues
			l_abs_vector := l_eigenvalues.abs
			l_max_magnitude := l_abs_vector.max
			l_min_magnitude := l_abs_vector.min
			if l_min_magnitude > 0.0 then
				Result := l_max_magnitude / l_min_magnitude
			end
		end

feature -- Access

	put (a_value : DOUBLE;i,j : INTEGER) is
			-- put
		require else
			valid_indices : 1 <= i and i <= height and 1 <= j and j <= width
		do
			if i >= j then
				data.item(i).put(a_value,j)
			else
				data.item(j).put(a_value,i)
			end
		ensure then
			item_put : Current[i,j] = a_value
		end

	inverse: SYMMETRIC_MATRIX
			-- Inverse of matrix
		local
			l_matrix: MATRIX
		do
			create l_matrix.make_as_inverse (Current)
			if l_matrix /= Void then
				create Result.make_from_triangle (l_matrix, False)
			end
		end

feature -- Test

	test
			-- Test feature
		local
			l_eigensystem : RP_EIGENSYSTEM
			l_sm, l_sm1, l_sqrt : SYMMETRIC_MATRIX
			l_m : MATRIX
			l_v : VECTOR
			l_cn : DOUBLE
		do
			create l_sm.make_from_array(<<<<1.>>, <<1., 3.>>>>)
			l_eigensystem := l_sm.eigensystem
			l_m := l_eigensystem.eigenvectors
			l_v := l_eigensystem.eigenvalues
			l_cn := l_sm.condition_number

			l_sqrt := l_sm.sqrt_matrix
			create l_sm1.make_xtx(l_sqrt)
			create l_sm1.make_xtx(l_sm)
			l_sm := l_sm.xtx
		end

feature -- Eigensystem

	eigensystem : RP_EIGENSYSTEM is
			-- eigen system
		local
			i,iter,m,j,k,l,n : INTEGER
			l_break,l_done : BOOLEAN
			dd,g,s,c,p,r,l_r_signed,f,b : DOUBLE
			l_tridiagonal : TRIDIAGONAL_MATRIX
			d,e : VECTOR
			z : MATRIX
		do
			create l_tridiagonal.make(Current)
			d := l_tridiagonal.diagonal_elements.twin
			e := l_tridiagonal.off_diagonal_elements.twin
			n := width
			create z.make (height, width)
			from i := 1 until i > height loop
				from j := 1 until j > width loop
					z[i,j] := l_tridiagonal[i,j]
					j := j + 1
				end
				i := i + 1
			end
			from i := 2 until i > n loop
				e[i-1]:=e[i]
				i := i + 1
			end
			e[n]:=0.0
			from l:= 1 until l > n loop
				iter:=0
				l_done := false
				from until l_done loop
					l_break := false
					from m := l until m >= n or else l_break loop
						dd:=(d[m]).abs+(d[m+1]).abs
						l_break := (e[m]).abs + dd = dd
						if not l_break then
							m := m + 1
						end
					end
					l_break := false
					if m /= l then
						iter := iter + 1
						if iter = 30 then raise ("too many iterations") end
						g:=(d[l+1]-d[l])/(2.0*e[l])
						r:=hypotenuse (g,1.0)
						if g >= 0.0 then
							l_r_signed := r.abs
						else
							l_r_signed := - r.abs
						end
						g:=d[m]-d[l]+e[l]/(g+l_r_signed)
						s:=1.0
						c:= 1.0
						p:=0.0
						l_break := false
						from i := m-1 until i < l or else l_break loop
							f:=s*e[i]
							b:=c*e[i]
							r := hypotenuse(f,g)
							e[i+1] := r
							if r = 0.0 then
								d[i+1] := d[i+1] - p
								e[m]:=0.0
								l_break := true
							else
								s:=f/r
								c:=g/r
								g:=d[i+1]-p
								r:=(d[i]-g)*s+2.0*c*b
								p := s*r
								d[i+1]:=g+p
								g:=c*r-b
								from k := 1 until k > n loop
									f:=z[k,i+1]
									z[k,i+1]:=s*z[k,i]+c*f
									z[k,i]:=c*z[k,i]-s*f
									k := k + 1
								end
								i := i - 1
							end
						end
						if r /= 0.0 or else i < l then
							d[l] := d[l] - p
							e[l]:=g
							e[m]:=0.0
						end
					end
					l_done := m = l
				end
				l := l + 1
			end
			create Result.make(z,d)
		end
	hypotenuse(a,  b: DOUBLE) : DOUBLE is
			-- hypotenuse
		local
			r,l_a,l_b: DOUBLE
		do
			l_a := a.abs
			l_b := b.abs
			if l_a > l_b then
				r := l_b / l_a
				Result :=  l_a * sqrt(1 + r * r)
			elseif b /= 0.0 then
				r := l_a / l_b;
				Result :=  l_b * sqrt(1 + r * r)
			else
				Result := 0.0
			end
		end

	scalar_multiply,scaled (a_value : DOUBLE) : like Current is
			-- returns a matrix like current, but with every element multiplied by `a_value'
		local
			i,j : INTEGER
		do
			create Result.make(height)
			from i := 1 until i > height loop
				from j := 1 until j > width loop
					Result[i, j] := Current[i, j] * a_value
					j := j + 1
				end
				i := i + 1
			end
			if column_names /= Void then
				Result.set_column_names(column_names)
			end
			if row_names /= Void then
				Result.set_row_names(row_names)
			end
		end

	correct_mismatch is
			-- allow use of old storables
		do

		end

end

