indexing
	description: "Class to compute general comparison statistics for a collection of expression results for a signature"
	author: "Mauricio Bustos"
	last_editor: "$Author: mb $"
	date: "$Date: 2011/01/25 $"
	revision: "$Revision: #21 $"
	path: "$File: //rose/source/kernel/persistence/databases/signature_tagged/testenv_statistics.e $"

class
	TESTENV_STATISTICS[reference G -> ROSE_SIGNATURE_RECORD[ENTITY]]

inherit
	MISMATCH_CORRECTOR
		redefine
			correct_mismatch
		end
 	ROSE_CONSTANTS
		export
			{NONE} all
		end

create
	make

feature {NONE} -- Creation

	make (a_expression: like expression) is
			-- Create the statistics object
		do
			if expression /= Void then
				expression := a_expression
				if expression.variable_description.source.is_equal ("risk") then
					is_normalized := True
				end
			end
			create availability_problems.make_empty
			create differences.make_empty
			differences.set_comparison (agent differences_order)
		end

feature -- Basic Operations

	truncate_for_storage is
			-- Shrink the object in preparation for storage into a database
		do
			if differences.count > differences_limit_count then
				create differences.make_from_array (differences.subarray (1, differences_limit_count))
			end
		end

	sort_differences is
			-- Sort differences by descending value for storage
		local
			l_clean_differences: like differences
		do
			create l_clean_differences.make_empty
			from differences.start until differences.after loop
				if differences.current_item.difference > difference_threshold then
					l_clean_differences.extend (differences.current_item)
				end
				differences.forth
			end
			if l_clean_differences.is_empty and not differences.is_empty then
				r_square_computed := True
				r_square_saved := 1.0
			end
			differences := l_clean_differences
			differences.sort_by (agent differences_order)
		end

	reset_differences is
			-- Reset counters
		do
			sum_of_squared_differences := 0.0
			r_square_saved := 0.0
			r_square_computed := False
			differences.wipe_out
			availability_problems.wipe_out
			differences.set_comparison (agent differences_order)
		end
	
feature -- Element modification

	set_count (a_count: INTEGER) is
			-- Set count to 'a_count'
		require
			meaningful_count: a_count >= 0
		do
			count := a_count
		end
	
	insert_difference (a_difference: DATA_COMPARISON [G]; a_table: HASH_TABLE [TESTENV_VALIDATED_DIFFERENCES, INTEGER]) is
			-- Add 'a_differece' to either differences or availability_problems depending on the data
		require
			valid_difference: a_difference /= Void
		do
			if is_defined_double (a_difference.difference) then
				if a_table /= Void then
					a_table.search (a_difference.target.signature_id)
					if a_table.found then
						if not close_to_validation (a_difference, a_table.found_item) and then
						 not a_table.found_item.comment.as_lower.has_substring ("[entity_validation]") then
							differences.extend (a_difference)
							sum_of_squared_differences := sum_of_squared_differences + a_difference.difference
						end
					else
						differences.extend (a_difference)
						sum_of_squared_differences := sum_of_squared_differences + a_difference.difference
					end
				else
					differences.extend (a_difference)
					sum_of_squared_differences := sum_of_squared_differences + a_difference.difference
				end
			else
				availability_problems.extend (a_difference)
			end
		end

feature -- Access

	r_square: DOUBLE is
			-- Correlation for differences observations
		local
			l_regression: REGRESSION
			l_retry: BOOLEAN
			l_should_regress: BOOLEAN
		do
			if not r_square_computed then
				if not l_retry then
					if differences.count > 2 then
						from differences.start until differences.off or l_should_regress loop
							if differences.current_item.x /= differences.current_item.y then
								l_should_regress := True
							end
							differences.forth
						end
						if l_should_regress then
							create l_regression.make_with_intercept (create {MATRIX}.make_from_columns (independent_variable),
																	 dependent_variable)
							r_square_saved := l_regression.r_square
							if not constants.ieee.isfinite (r_square_saved) then
								r_square_saved := Undefined.Undefined_Double
							end
						else
							r_square_saved := 1.0
						end
					end
					r_square_computed := True
				else
					r_square_saved := Undefined.Undefined_Double
				end
			end
			Result := r_square_saved
		rescue
			l_retry := True
			retry
		end

	is_normalized: BOOLEAN
			-- Is the expression able to be normalized?
	expression: ROSE_EXPRESSION [ANY]
			-- Expression for these statistics
	sum_of_squared_differences: DOUBLE
			-- Sum of squared differences for all observations
	availability_problems: ROSE_LINEAR_ARRAY[DATA_COMPARISON[G]]
			-- Collection of comparisons that have some kind of availability problem
	differences: ROSE_LINEAR_ARRAY[DATA_COMPARISON[G]]
			-- Collection of comparisons that have some kind of difference
	count: INTEGER
			-- Number of total observations

feature {NONE} -- Implementation

	r_square_saved: DOUBLE
			-- Saved calculation of r_square for efficiency
	r_square_computed: BOOLEAN
			-- Has r_square been computed?

	independent_variable: ROSE_LINEAR_ARRAY[VECTOR] is
			-- Independent variable vector for use by regression in determination of correlation
		local
			l_vector: VECTOR
		do
			create Result.make_empty
			create l_vector.make (differences.count)
			Result.extend (l_vector)
			from differences.start until differences.off loop
				if Undefined.is_defined_double (differences.current_item.normalized_x) and Undefined.is_defined_double (differences.current_item.normalized_y) then
					l_vector.put (differences.current_item.normalized_x, differences.index)
				elseif not Undefined.is_defined_double (differences.current_item.normalized_x) and not Undefined.is_defined_double (differences.current_item.normalized_y) then
					l_vector.put (0.0, differences.index)
				end
				differences.forth
			end
		end

	dependent_variable: VECTOR is
			-- Dependent variable vector for use by regression in determination of correlation
		do
			create Result.make (differences.count)
			from differences.start until differences.off loop
				if Undefined.is_defined_double (differences.current_item.normalized_x) and Undefined.is_defined_double (differences.current_item.normalized_y) then
					Result.put (differences.current_item.normalized_y, differences.index)
				elseif not Undefined.is_defined_double (differences.current_item.normalized_x) and not Undefined.is_defined_double (differences.current_item.normalized_y) then
					Result.put (0.0, differences.index)
				end
				differences.forth
			end
		end

	correct_mismatch is
			-- Mismatch corrector
		do
		end

	differences_order (a, b: DATA_COMPARISON[G]): BOOLEAN is
			-- Order by difference magnitude
		do
			if a.difference > difference_threshold or b.difference > difference_threshold then
				Result := a.difference > b.difference
			end
		end

	close_to_validation (a_difference: DATA_COMPARISON [G]; a_validation: TESTENV_VALIDATED_DIFFERENCES): BOOLEAN is
			-- Is the difference close enough to the validation to be ignored?
		require
			valid_difference: a_difference /= Void
			valid_validation: a_validation /= Void
		do
			Result := (a_difference.x < a_validation.x + validated_differences_epsilon and
					   a_difference.x > a_validation.x - validated_differences_epsilon and
					   a_difference.y < a_validation.y + validated_differences_epsilon and
					   a_difference.y > a_validation.y - validated_differences_epsilon)
		end
	
	differences_limit_count: INTEGER is 25_000
			-- Maximum number of differences used to compute statistics
	difference_threshold: DOUBLE is 0.000001
			-- Threshold in order to consider a difference material
	validated_differences_epsilon: DOUBLE is 0.000001
			-- Epsilon value to determine if a difference was previously validated and should be removed from the comparison

feature {NONE} -- Helpers

	Undefined: ROSE_LEGACY_UNDEFINED is
			-- Access to undefined constants
		once
			create Result
		end

end








