note

	description:

		"A table that hashes by uid (unique identifier)"

	author: "Ben Gum"
	last_editor: "$Author: it $"
	date: "$Date: 2013/02/13 $"
	revision: "$Revision: #4 $"
	path: "$File: //rose/source/Kernel/expert_system/uid_table.e $"

class UID_TABLE [G, H -> HASH_KEYED [INTEGER_64]]

create

	make_equal,
	make_default

feature {NONE} -- creation

	make_default is
			-- default create
		do
			create table.make_default
		end

	make_equal (a_count: INTEGER) is
			-- create table of size `a_count'
		require
			positive_a_count: a_count >= 0
		do
			create table.make_equal (a_count)
		end

feature -- access

	item alias "[]" (a_instrument: H): G assign put is
			-- item whose key is computed from `a_instrument'
		require
			a_instrument_not_void: a_instrument /= Void
			has: has (a_instrument)
		do
			Result := table.item (a_instrument.key).value
		end

	key (a_instrument: H): H is
			-- key corresponding to `a_instrument'
		require
			a_instrument_not_void: a_instrument /= Void
		local
			l_item: TUPLE [value: G; key: H]
		do
			l_item := table.item (a_instrument.key)
			if l_item /= Void then
				Result := l_item.key
			end
		end

	put (a_value: G; a_instrument: H) is
			-- insert `a_value', `a_instrument' if `a_instrument' is not present
		require
			a_instrument_not_void: a_instrument /= Void
		do
			table.put ([a_value, a_instrument], a_instrument.key)
		ensure
			has: has (a_instrument)
		end

	force (a_value: G; a_instrument: H) is
			-- insert `a_value', `a_instrument', possibly overwrite any existing entry with key `a_instrument'
		require
			a_instrument_not_void: a_instrument /= Void
		do
			table.force ([a_value, a_instrument], a_instrument.key)
		ensure
			has: has (a_instrument)
		end

	force_new (a_value: G; a_instrument: H) is
			-- insert `a_value', `a_instrument' with key `a_instrument'
		require
			a_instrument_not_void: a_instrument /= Void
		do
			table.force_new ([a_value, a_instrument], a_instrument.key)
		ensure
			has: has (a_instrument)
		end

	wipe_out is
			-- delete all entries
		do
			table.wipe_out
		ensure
			is_empty: is_empty
		end

	is_empty: BOOLEAN is
			-- is the table empty
		do
			Result := table.is_empty
		end

	has (a_instrument: H): BOOLEAN is
			-- does the table contain an entry with key `a_instrument'?
		require
			a_instrument_not_void: a_instrument /= Void
		do
			Result := table.has (a_instrument.key)
		end

	keys: ROSE_LINEAR_ARRAY [H] is
			-- array of keys in the table
		do
			create Result.make_empty
			table.do_all (agent add_key (Result, ?))
		ensure
			not_void: Result /= Void
		end

	do_all (a_agent: PROCEDURE [ANY, TUPLE [TUPLE [value: G; key: H]]]) is
			-- do to every item in the table
		require
			a_agent_not_void: a_agent /= Void
		do
			table.do_all (a_agent)
		end

	search (a_instrument: H) is
			-- search for `a_instrument'
		require
			a_instrument_not_void: a_instrument /= Void
		do
			table.search (a_instrument.key)
		end

	found: BOOLEAN is
			-- did we find an item during `search'?
		do
			Result := table.found
		end

	found_item: G is
			-- item found during `search'
		do
			if found then
				Result := table.found_item.value
			end
		end

	found_key: H is
			-- key found during `search'
		do
			if found then
				Result := table.found_item.key
			end
		end

	count: INTEGER is
			-- number of items in the table
		do
			Result := table.count
		end

	remove (a_instrument: H) is
			-- Remove `a_instrument'
		require
			a_instrument_not_void: a_instrument /= Void
		do
			table.remove (a_instrument.key)
		ensure
			removed: not has (a_instrument)
		end

	new_cursor: DS_HASH_TABLE_CURSOR [TUPLE [value: G; key: H], INTEGER_64] is
			-- Facade to `table.new_cursor'
		do
			Result := table.new_cursor
		end

feature {NONE} --Implementation

	table: DS_HASH_TABLE [TUPLE [value: G; key: H], INTEGER_64]
			-- storage for the items and keys

	add_key (a_keys: like keys; a_item: TUPLE [value: G; key: H]) is
			-- add `a_item's key to `a_keys'
		require
			a_keys_not_void: a_keys /= Void
			a_item_not_void: a_item /= Void
		do
			a_keys.extend (a_item.key)
		end

invariant

	table_not_void: table /= Void
	no_void_item_in_table: not table.has_void_item

end
