note

	description:

		"Vectors implemented as SPECIAL [DOUBLE]"

	tests: "TEST_VECTOR"
	author: "Mark Howard"
	last_editor: "$Author: mb $"
	date: "$Date: 2014/03/14 $"
	revision: "$Revision: #142 $"
	path: "$File: //rose/source/kernel/infrastructure/math/vector.e $"
	storable_version: ""

class VECTOR

inherit

	EXTENDABLE_VECTOR
		redefine
			scaled,
			to_vector,
			as_vector,
			infix "+",
			infix "-",
			infix "*",
			to_array
		end

	SORTABLE [DOUBLE]
		undefine
			is_equal
		end

	RP_SERIALIZABLE

create

	make, duplicate,
	make_from_array, make_by_area, make_from_column_matrix,
	make_identity,  make_constant, make_sequence, make_with_names,
	make_from_delimited_string,
	make_weighted,make_with_capacity

convert

	make_from_array ({attached ARRAY[DOUBLE]})

feature -- Initialization

	make (a_count: INTEGER) is
			-- Create with `a_count' zeros.  Avoid calling 'make(0)'.  See if you can use 'make_with_capacity'.
		require
			a_count_non_negative: a_count >= 0
		do
			count := a_count
			create data.make_filled (0.0, a_count + 1)
		ensure
			correct_count: count = a_count
			all_zeros: first_non_zero_item_index = 0
		end

	make_weighted (a_vector,a_weights : VECTOR)
			-- Create with items the pairwise product of `a_vector' and `a_weights'
		local
			i : INTEGER
		do
			make(a_vector.count)
			from i := 1 until i > count loop
				data[i] := a_vector[i] * a_weights[i]
				i := i + 1
			end
		end

	duplicate (a_other : ABSTRACT_READABLE_VECTOR) is
			-- Make Current copied from `a_other'
		require
			nonvoid: a_other /= void
		local
			i: INTEGER
		do
			make (a_other.count)
			from i := 1
			until
				i > count
			loop
				Current [i] := a_other [i]
				i := i + 1
			end
		ensure
			duplicate: same_vector (a_other)
		end

	make_from_array (a_array: attached ARRAY [DOUBLE]) is
			-- Create as copy of `a_array'.
		require
			a_array_not_void: a_array /= Void
		local
			i, l_lower: INTEGER
		do
			make (a_array.count)
			l_lower := a_array.lower
			from
				i := 1
			until
				i > count
			loop
				put (a_array.item (l_lower + i - 1), i)
				i := i + 1
			end
		ensure
			same_count: count = a_array.count
			same_items: True -- needs an ISE fix: to_array.same_items (a_array)
		end

	make_from_column_matrix (a_matrix: ABSTRACT_MATRIX) is
			-- Create as a copy of column 1 in `a_matrix'.
		require
			a_matrix_not_void: a_matrix /= Void
			at_least_one_column: a_matrix.width >= 1
		local
			i: INTEGER
		do
			make (a_matrix.height)
			from i := 1
			until
				i > count
			loop
				put (a_matrix [i, 1], i)
				i := i + 1
			end
		ensure
			same_count: count = a_matrix.height
			same_vector_as_column_1: same_vector (a_matrix.column (1))
		end

	make_identity (a_count: INTEGER) is
			-- Create as a vector of `a_count' 1s.
		require
			a_count_non_negative: a_count >= 0
		do
			make_constant (1.0, a_count)
		ensure
			correct_count: count = a_count
			all_ones: count_by_value (1.0) = a_count
		end

	make_constant (a_value: DOUBLE; a_count: INTEGER) is
			-- Initialize with all items set with `a_value'.
		require
			a_count_non_negative: a_count >= 0
		local
			i: INTEGER
		do
			make (a_count)
			from
				i := 1
			until
				i > a_count
			loop
				data[i] :=  a_value
				i := i + 1
			end
		ensure
			correct_count: count = a_count
			all_undefined: not is_defined_double (a_value) implies is_null
			all_constant_defined: is_defined_double (a_value) implies  count_by_value (a_value) = a_count
		end

	make_sequence (a_beg, a_end, a_step: DOUBLE) is
			-- Create as sequence starting with `a_beg' and ending close to `a_end' (if different from `a_beg').
			-- Successive elements should differ by `a_step'.
		require
			valid_step: (a_beg = a_end) or else (
				(a_end - a_beg).abs >= a_step.abs and (a_end - a_beg) * a_step > 0 and a_step /= 0)
		local
			i, l_count: INTEGER
		do
			l_count := 1 + ((a_end - a_beg) / a_step).floor
			if l_count < 0 then
				l_count := 0
			end
			make (l_count)
			from
				i := 1
			until
				i > l_count
			loop
				put (a_beg + (i-1) * a_step, i)
				i := i + 1
			end
		ensure
			correct_count: count = (1 + ((a_end - a_beg) / a_step).floor).max (0)
			all_defined: is_all_defined
			all_values_set: for_all_with_index (agent stepped_value (?, ?, a_beg, a_step))
		end


	make_with_names (a_names: ARRAY [STRING]) is
			-- Initialize with named columns.
		require
			a_names_not_void: a_names /= Void
		do
			make (a_names.count)
			set_names (a_names)
		ensure
			correct_count: count = a_names.count
			all_zeros: first_non_zero_item_index = 0
			names_not_void: names /= Void
			correct_name_count: names.count = a_names.count
			correct_names: names.for_all_with_index (agent equal_strings (a_names, ?, ?))
		end

	make_from_delimited_string (a_string: STRING; a_delimiter: CHARACTER) is
			-- Create from `a_string' interpreted as doubles separated by `a_delimiter'.
		require
			a_string_not_void: a_string /= Void
		local
			i: INTEGER
			l_string: STRING
			l_dtext: ROSE_DELIMITED_TEXT
		do
			create l_dtext.make (a_string, a_delimiter)
			make (l_dtext.count)
			from
				i := 1
			until
				i > count
			loop
				l_string := l_dtext [i]
				check
					l_string_not_void: l_string /= Void
							-- missing invariant in ROSE_DELIMITED_TEXT
				end
				if l_string.is_double then
					Current [i] := l_string.to_double
				end
				i := i + 1
			end
		end

	make_with_capacity(a_count, a_capacity: INTEGER)
	-- Makes a VECTOR with an initial size of 'a_count' but can be extended to 'a_capacity' without resizing.
	-- Use this if you know (approximately) how large this VECTOR is going to get.  This
	-- makes the use of 'extend' much faster.
	require
		a_count_non_negative: a_count >= 0
		a_capacity_large_enough: a_capacity >= a_count
	do
		count := a_count
		create data.make_filled(0.0, a_capacity + 1)	-- Why "+ 1"?  Because VECTOR does not use element 0.
	ensure
		correct_count: count = a_count
		all_zeros: first_non_zero_item_index = 0
	end

feature {MATRIX} -- Initialization {MATRIX}

	make_by_area (a_area: SPECIAL [DOUBLE]) is
			-- Make from `a_area'
		require
			a_area_not_void: a_area /= Void
			strictly_positive_count : a_area.count >= 1
		do
			count := a_area.count - 1
			data := a_area
		ensure
			correct_count: count = a_area.count - 1
			data_aliased: data = a_area
		end

feature -- Access

	data: SPECIAL [DOUBLE]
			-- Holds all vector elements

	frozen item alias "[]" (i: INTEGER): DOUBLE assign put is
			-- Returns i_th value
		do
			Result := data.item (i)
		end

	values_at (a_names: ARRAY [STRING]): like Current is
			-- Values by name
		require
			a_names_not_void: a_names /= Void
			has_names: has_names
			no_void_name: not a_names.has (Void)
		local
			i: INTEGER
		do
			create Result.make (a_names.count)
			from
				i := a_names.lower
			invariant
				a_names_addressable: i <= a_names.upper implies a_names.valid_index (i)
				result_addressable:  i <= a_names.upper implies Result.valid_index (i - a_names.lower + 1)
			variant
				increasing_names_index: a_names.upper - i + 1
			until
				i > a_names.upper
			loop
				Result.put (value_at (a_names.item (i)), i - a_names.lower + 1)
				i := i + 1
			end
			Result.set_names (a_names)
		ensure
			values_at_not_void: Result /= Void
			correct_count: Result.count = a_names.count
			has_names: Result.has_names
			names_set: Result.names.for_all_with_index (agent Result.equal_strings (a_names, ?, ?))
			no_void_name: not Result.names.has (Void)
			correct_values: True -- Result.for_all_with_index (agent involving value_at and being careful about undefined)
		end

	linear_interpolation (a_first, a_last: DOUBLE): like Current is
			-- Rescaled version of Current with the items rescaled
			-- so that Result.first = a_first and Result.last = a_last and
			-- their relative positions remain the same.
			--
			-- Possibly misnamed routine.
			-- If Current.first = Current.last it (arbitrarily?) returns
			-- a vector with all values being `a_first'.
		require
			not_empty: count > 0
		local
			i: INTEGER
			l_first,l_last: DOUBLE
			l_rate: DOUBLE
			l_v: DOUBLE
		do
			l_first := item (1)
			l_last := item (count)
			create Result.make (count)
			if l_first = l_last then
				from i := 1 until i > count loop
					Result.put (a_first,i)
					i := i + 1
				end
			else
				from i := 1 until i > count loop
					l_rate := (item (i) - l_last) / (l_first - l_last)
					l_v := a_first * l_rate + a_last * (1.0 - l_rate)
					Result.put (l_v,i)
					i:=i+1
				end
			end
		ensure
			not_void: Result /= Void
			same_count: Result.count = count
			first_approx_equal: is_defined_double (a_first) implies
				one_zero_or_approx_equal (Result.first, a_first)
			last_approx_equal: (a_first /= a_last and is_defined_double (a_last)) implies
				one_zero_or_approx_equal (Result.last, a_last)
		end

	top_n_items (a_n: INTEGER): like Current is
			-- Top n items
		require
			a_n_non_negative: a_n >= 0
		local
			i, l_size: INTEGER
			l_order: like ascending_order
		do
			l_size := a_n.min (count)
			create Result.make (a_n)
			l_order := ascending_order
			from
				i := 1
			until
				i > l_size
			loop
				Result.put (item (l_order.item (count - i + 1)), i)
				i := i + 1
			end
		ensure
			top_n_items_not_void: Result /= Void
			requested_size: Result.count = a_n
		end

	bottom_n_items (a_n: INTEGER): like Current is
			-- Bottom n items
		require
			a_n_non_negative: a_n >= 0
		local
			i, l_size: INTEGER
			l_order: like ascending_order
		do
			l_size := a_n.min (count)
			create Result.make (a_n)
			l_order := ascending_order
			from
				i := 1
			until
				i > l_size
			loop
				Result.put (item (l_order.item (i)), i)
				i := i + 1
			end
		ensure
			bottom_n_items_not_void: Result /= Void
			requested_size: Result.count = a_n
		end

feature -- Basic operations

	dot_product (a_other: VECTOR): DOUBLE is
			-- Sum of products of corresponding elements;
			-- See also `infix *'.
			-- This has better performance than `infix *', but in the presence
			--  of NaNs, `Result' = `QNAN'. Or in the presence of `legacy_missing_real_64',
			-- "incorrect" results appear.
		require
			a_other_not_void: a_other /= Void
			same_size : count = a_other.count
		local
			i: INTEGER
			l_other_data: SPECIAL [DOUBLE]
		do
			l_other_data := a_other.data
			from
				i := 1
			until
				i > count
			loop
				Result := Result + data[i] * l_other_data[i]
				i := i + 1
			end
		end

	scaled_subtract (a_vector: VECTOR; a_multiplier: DOUBLE) is
			-- Subtract `a_vector [i]' * `a_multiplier' from `item [i]' for all i in` valid_index (i)'.
			-- N.B. NaN propogate, and -999s are treated as ordinary data.
		require
			a_vector_not_void: a_vector /= Void
			same_count: a_vector.count = count
		local
			i: INTEGER
			l_other_data: SPECIAL [DOUBLE]
		do
			l_other_data := a_vector.data
			from
				i := 1
			until
				i > count
			loop
				data[i] := data[i] - (l_other_data[i] * a_multiplier)
				i := i + 1
			end
		end

	replace_null is
			-- Replace all undefined elements by 0.0.
		do
			replace_null_with (0.0)
		ensure
			all_values_defined: is_all_defined
			no_length_change: old count = count
		end

	replace_null_with (a_value: DOUBLE) is
			-- Replace all undefined elements by `a_value'.
		local
			i: INTEGER
			l_v: DOUBLE
		do
			from
				i := 1
			until
				i > count
			loop
				l_v := item (i)
				if not is_defined_double (l_v) then
					put (a_value, i)
				end
				i := i + 1
			end
		ensure
			all_values_defined: is_defined_double (a_value) implies is_all_defined
			no_length_change: old count = count
		end

	replace_negative is
			-- Replace all negative elements by 0.0.
		local
			i: INTEGER
			l_v: DOUBLE
		do
			from
				i := 1
			until
				i > count
			loop
				l_v := item (i)
				if l_v < 0.0 then
					put (0.0, i)
				end
				i := i + 1
			end
		ensure
			no_negative_values: is_all_defined implies non_negative
			no_length_change: old count = count
		end

	zero_below_tolerance (a_tol: DOUBLE) is
			-- Replace all below `a_tol' by 0.0.
		local
			i: INTEGER
			l_v: DOUBLE
		do
			from
				i := 1
			until
				i > count
			loop
				l_v := item (i)
				if l_v < a_tol then
					put (0.0, i)
				end
				i := i + 1
			end
		ensure
			no_length_change: old count = count
		end

	zero_if_abs_below_tolerance (a_tol: DOUBLE) is
			-- Replace all values whose magnitude is below `a_tol' by 0.0.
		require
			a_tol_positive: a_tol > 0
		local
			i: INTEGER
			l_v: DOUBLE
		do
			from
				i := 1
			until
				i > count
			loop
				l_v := item(i).abs
				if l_v < a_tol then
					put (0.0, i)
				end
				i := i + 1
			end
		ensure
			no_length_change: old count = count
		end

	truncate_at (a_min, a_max: DOUBLE) is
			-- Bound all defined elements within 'a_min' and 'a_max'.
		local
			i: INTEGER
			l_v: DOUBLE
		do
			from
				i := 1
			until
				i > count
			loop
				l_v := item (i)
				if is_defined_double (l_v) then
					put (l_v.max (a_min).min (a_max),i)
				end
				i := i + 1
			end
		ensure
			minimum_element: min >= a_min
			maximum_element: max <= a_max
			no_length_change: old count = count
		end

	floor_at (a_floor: DOUBLE) is
			-- Ensure all defined elements are not below 'a_floor'.
		local
			i: INTEGER
			l_v: DOUBLE
		do
			from
				i := 1
			until
				i > count
			loop
				l_v := item (i)
				if is_defined_double (l_v) then
					put (l_v.max (a_floor), i)
				end
				i := i + 1
			end
		ensure
			defined_floor: min >= a_floor
			no_length_change: old count = count
		end

	ceiling_at (a_ceiling: DOUBLE) is
			-- Ensure all defined elements are not above 'a_ceiling'.
		local
			i: INTEGER
			l_v: DOUBLE
		do
			from
				i := 1
			until
				i > count
			loop
				l_v := item (i)
				if is_defined_double (l_v) then
					put (l_v.min (a_ceiling),i)
				end
				i := i + 1
			end
		ensure
			defined_ceiling: max <= a_ceiling
			no_length_change: old count = count
		end

	scale (a_value: DOUBLE) is
			-- Multiply all defined elements by `a_value'.
		local
			i: INTEGER
		do
			from
				i := 1
			until
				i > count
			loop
				if is_defined_double (data.item (i)) then
					data[i] := data[i] * a_value
				end
				i := i + 1
			end
		ensure
			no_length_change: old count = count
			scaled: True -- the following isn't quite right - I think `old Current' is just `Current',
			--               by reference semantics:
			--               is_all_defined implies Current.same_vector ((old Current).scaled (a_value))
		end

	scale_negative is
			-- Scale every defined element of `Current' so that negative entries sum to -1.0.
			--	No scaling takes place when 1.0/0.0 is encountered.
		local
			l_norm_factor: DOUBLE
		do
			l_norm_factor := sum - positive_sum
			l_norm_factor := l_norm_factor.abs

			if l_norm_factor /= 0.0 then
				scale (1.0 / l_norm_factor)
			end
		ensure
			no_length_change: old count = count
		end

	scale_positive is
			-- Scale defined elements of `Current' so that positive entries sum to 1.0
			--	No scaling takes place when 1.0/0.0 is encountered.		
		local
			l_norm_factor: DOUBLE
		do
			l_norm_factor := positive_sum

			if l_norm_factor /= 0.0 then
				scale (1.0 / l_norm_factor)
			end
		ensure
			no_length_change: old count = count
		end

	scale_mid_way is
			-- Scale defined_elements of `Current' so that absolute value sum is 2.0, that is,
			--  scaling was by 1/2 {sum(positive) + sum(|negative|)}.
			--	No scaling takes place when 1.0/0.0 is encountered.
		local
			l_norm_factor: DOUBLE
		do
			l_norm_factor := abs_values_sum / (1.0 + 1.0)

			if l_norm_factor /= 0.0 then
				scale (1.0 / l_norm_factor)
			end
		ensure
			no_length_change: old count = count
			abs_sum_2_0: approx_equal (abs.sum, 2.0)
		end

	scale_one_sided_max is
			-- Scale defined elements of `Current' so that `positive_sum' <= 1.0
			--  and abs negative sum <= 1.0 and (= 1.0) holds in at least one of those cases.
			--	No scaling takes place when 1.0/0.0 is encountered.
		local
			l_norm_factor: DOUBLE
		do
			l_norm_factor := positive_sum
			l_norm_factor := (l_norm_factor - sum).abs.max (l_norm_factor)

			if l_norm_factor /= 0.0 then
				scale (1.0 / l_norm_factor)
			end
		ensure
			no_length_change: old count = count
			positive_sum_max_one: positive_sum <= 1.000001
			absolute_negative_sum_max_one: (positive_sum - sum).abs <= 1.000001
			at_least_one_equality: approx_equal (positive_sum, 1.0) or approx_equal ((positive_sum - sum).abs, 1.0)
		end

	scale_sum is
			-- Scale defined elements of `Current' so that the sum is 1.0.
			--	No scaling takes place when 1.0/0.0 is encountered.
		local
			l_norm_factor: DOUBLE
		do
			l_norm_factor := sum

			if l_norm_factor /= 0.0 then
				scale (1.0 / l_norm_factor)
			end
		ensure
			no_length_change: old count = count
			sum_to_1: approx_equal (sum, 1.0)
		end

	add (a_vector: ABSTRACT_READABLE_VECTOR) is
			-- Add each item in `a_vector' to corresponding item of `Current', when both pass `is_defined'.
			-- Otherwise set that item to missing.
		require
			valid_size: a_vector /= Void and then a_vector.count = count
		local
			i: INTEGER
		do
			from
				i := 1
			until
				i > count
			loop
				if is_defined (item(i)) and is_defined (a_vector.item(i)) then
					put (item (i) + a_vector.item (i), i)
				else
					put (constants.missing_double, i)
				end
				i := i + 1
			end
		ensure
			no_length_change: old count = count
		end

	subtract (a_vector: VECTOR) is
			-- Subtract  each item in `a_vector' from corresponding item of `Current', when both pass `is_defined'.
			-- Otherwise set that item to missing.
		require
			valid_size: a_vector /= Void and then a_vector.count = count
		local
			i: INTEGER
			l_other_data: SPECIAL [DOUBLE]
		do
			l_other_data := a_vector.data
			from
				i := 1
			until
				i > count
			loop
				if is_defined (item(i)) and is_defined (l_other_data[i]) then
					data[i] := data.item (i) - l_other_data[i]
				else
					data[i] := constants.missing_double
				end
				i := i + 1
			end
		ensure
			no_length_change: old count = count
		end

	add_scalar (a_double: DOUBLE) is
			-- Add `a_double' to each defined item of `Current'.
		local
			i: INTEGER
		do
			from
				i := 1
			until
				i > count
			loop
				if is_defined_double (data[i]) then
					data[i] := data[i] + a_double
				end
				i := i + 1
			end
		ensure
			no_length_change: old count = count
		end

	frozen put (a_value: DOUBLE; i: INTEGER) is
			-- Set item at `i' to `a_value'.
		do
			debug ("strictvector")
				if (isnan (a_value) or not isfinite (a_value)) then
					(create {EXCEPTIONS}).raise ("problem")
				end
			end
			data[i] := a_value
		ensure then
			a_value_set: (a_value = a_value) implies item (i) = a_value
			no_length_change: old count = count
		end

	put_sub_vector_by_flags (a_sub_vector: ABSTRACT_READABLE_VECTOR; a_flags: ARRAY [BOOLEAN]) is
			-- Replace elements of `Current' (where corresponding element in `a_flags' is `True')
			--  by next unused element (if there is one) from `a_sub_vector'.
		require
			a_flags_not_void: a_flags /= Void
			not_too_many_flags: a_flags.count <= count
			a_sub_vector_not_void: a_sub_vector /= Void
		local
			i, l_ix, l_count: INTEGER
		do
			from
				i := a_flags.lower
			until
				i > a_flags.upper
			loop
				if a_flags [i] then
					l_count := l_count + 1
					l_ix := i - a_flags.lower + 1
					if a_sub_vector.valid_index (l_count) and valid_index (l_ix) then
						data [l_ix] := a_sub_vector [l_count]
					end
				end
				i := i + 1
			end
		ensure
			no_length_change: old count = count
		end

	put_sub_vector_by_indices (a_sub_vector: ABSTRACT_READABLE_VECTOR; a_ixs: ARRAY [INTEGER]) is
			-- Replace elements of `Current' whose indexes are in `a_ixs'
			--  by elements from `a_sub_vector' in corresponding position within `a_ixs'.
			-- If `a_ixs' contains duplicates, then the last entry will be the one that takes affect.
		require
			a_ixs_not_void: a_ixs /= Void
			a_sub_vector_not_void: a_sub_vector /= Void
			valid_count : a_ixs.count = a_sub_vector.count
		local
			i, l_ix : INTEGER
		do
			from
				i := a_ixs.lower
			until i > a_ixs.upper
			loop
				l_ix := a_ixs [i]
				if valid_index (l_ix) then
					data [l_ix] := a_sub_vector [i - a_ixs.lower + 1]
				end
				i := i + 1
			end
		ensure
			no_length_change: old count = count
		end

	extend (a_value: DOUBLE) is
			-- <Precursor>
			-- N.B. If `Current' is full, only sufficient memory is added to accomodate `a_value',
			--  so repeated use of this call is expensive unless the memory has been pre-allocated.
			--	Consider using 'make_with_capacity' to avoid this penalty.
		do
			count := count + 1
			if data.count <= count then
				data := data.aliased_resized_area_with_default (0.0, count + 1)	-- Why "+ 1"?  Because the 0 element is unused in VECTOR.
			end
			put (a_value, count)
			name_table_saved := Void
		ensure then
			exactly_one_more_element: count = old count + 1
			a_value_added:  (a_value = a_value) implies item (count) = a_value
			existing_values_unchanged: sub_vector (1, count - 1).same_vector (old twin)
		end

	extend_with_name (a_value: DOUBLE; a_name: STRING) is
			-- Create `names' if necessary.
			-- If there is an existing element associated with `a_name', then
			--  add `a_value' to the current value for that entry, replacing it.
			-- If there is no associated_element, then add `a_value' to end of `Current',
			--  add `a_name' to end of `names'.
		require
			a_name_not_void: a_name /= Void
		do
			extend_and_name (a_value, a_name, True)
		ensure
			names_not_void: names /= Void
			a_name_present: names.has (a_name)
		end

	extend_and_name (a_value: DOUBLE; a_name: STRING; a_sumup: BOOLEAN) is
			-- Create `names' if necessary.
			-- If there is an existing element associated with `a_name', then
			--  if `a_sumup' is `False', do nothing, or if it is `True', add `a_value' to
			--  the current value for that entry, replacing it.
			-- If there is no associated_element, then add `a_value' to end of `Current',
			--  add `a_name' to end of `names'.
		require
			a_name_not_void: a_name /= Void
		local
			k: INTEGER
		do
			if names = Void then
				create names.make_empty -- this means `sub_vector' will now crash.
			end
			k := name_table.item (a_name)
			if k > 0 then
				if a_sumup then
					put (a_value + item (k), k)
				end -- N.B. no else clause
			else
				extend (a_value)
				names.extend (a_name)
			end
		ensure
			names_not_void: names /= Void
			a_name_present: names.has (a_name)
		end

	infix "+" (a_other: ABSTRACT_READABLE_VECTOR): VECTOR is
			-- Sums of corresponding elements
		local
			i: INTEGER
			l_vector : VECTOR
			l_other,l_result : SPECIAL[DOUBLE]
		do
			create Result.make (count)
			l_vector := a_other.as_vector
			l_other := l_vector.data
			l_result := Result.data
			from i := 1 until i > count loop
				if is_defined_double (data [i]) and is_defined_double (l_other [i]) then
					l_result [i] := data [i] + l_other [i]
				else
					l_result [i] := constants.missing_double
				end
				i := i + 1
			end
		end

	infix "-" (a_other: ABSTRACT_READABLE_VECTOR): VECTOR is
			-- Differences of corresponding elements
		local
			i: INTEGER
			l_vector : VECTOR
			l_other,l_result : SPECIAL[DOUBLE]
		do
			create Result.make (count)
			l_vector := a_other.as_vector
			l_other := l_vector.data
			l_result := Result.data
			from i := 1 until i > count loop
				if is_defined_double (data [i]) and is_defined_double (l_other [i]) then
					l_result [i] := data [i] - l_other [i]
				else
					l_result [i] := constants.missing_double
				end
				i := i + 1
			end
		end

	infix "*" (a_other: ABSTRACT_READABLE_VECTOR): DOUBLE is
			-- Sum of products of corresponding elements;
			-- See also `{VECTOR/SPARSE_VECTOR}.dot_product'.
		local
			i: INTEGER
			l_vector : VECTOR
			l_other : SPECIAL[DOUBLE]
			l_i,l_other_i : DOUBLE
		do
			l_vector := a_other.as_vector
			l_other := l_vector.data
			from
				i := 1
			until
				i > count
			loop
				l_i := data[i]
				l_other_i := l_other[i]
				if is_defined_double (l_i) and is_defined_double (l_other_i) then
					Result := Result + l_i * l_other_i
				end
				i := i + 1
			end
		end
	extend_vector (a_vector: VECTOR) is
			-- Append elements of `a_vector' to end of `Current'.
		require
			a_vector_not_void: a_vector /= Void
		local
			l_data,l_other: SPECIAL [DOUBLE]
			i, k: INTEGER
			l_names: like names
			l_vector_names: like names
		do
			l_data := data
			k := a_vector.count
			create data.make_filled (0.0, count + k + 1)
			from i := 1 until i > count loop
				data.put (l_data.item (i), i)
				i := i + 1
			end
			l_other := a_vector.data
			from i := count + 1 until i > count + k loop
				data[i] := l_other[i - count]
				i := i + 1
			end
			l_names := names; l_vector_names := a_vector.names
			count := count + k
			if l_names /= Void or else l_vector_names /= Void then
				if l_names = Void then
					create l_names.make (1, count-k)
				end
				if l_vector_names = Void then
					create l_vector_names.make (1, k)
				end
				set_names (l_names + l_vector_names)
			end
		ensure
			extended_count: count = a_vector.count + old count
		end

	put_vector (a_vector: ABSTRACT_READABLE_VECTOR; a_start_index: INTEGER) is
			-- Set values from `a_vector' into `Current', starting
			-- at position `a_start_index'.
			-- Ignore values that do not fit within `Current'.
		local
			i, k: INTEGER
		do
			if a_vector /= Void and then (a_start_index > 0 and then a_start_index <= count) then
				k := (count - a_start_index + 1).min (a_vector.count)
				from i := 1 until i > k loop
					put (a_vector.item (i), a_start_index + i - 1)
					i := i + 1
				end
			end
		end

	put_at (a_name: STRING; a_value: DOUBLE) is
			-- Replace item indexed by `a_name' (if it exists) with `a_value'.
		require
			names_available: has_names
			a_name_not_void: a_name /= Void
		local
			l_ix: INTEGER
		do
			l_ix := index_of (a_name)
			if l_ix > 0 then
				put (a_value, l_ix)
			end
		end

	put_named_values (a_other: ABSTRACT_READABLE_VECTOR) is
			-- Set values from `a_other' by name.
		require
			a_other_not_void: a_other /= Void
		local
			l_names: ROSE_LINEAR_ARRAY [STRING]
		do
			if has_names then
				l_names := a_other.names
				if l_names /= Void then
					from
						l_names.start
					until
						l_names.after
					loop
						if l_names.current_item /= Void then
							put_at (l_names.current_item, a_other.item (l_names.index))
						end
						l_names.forth
					end
				end
			end
		end

	set_subset_constant (a_value: DOUBLE; a_ixs: ARRAY [INTEGER]) is
			-- Set items of `Current' indexed by any of `a_ixs' to `a_value'.
		require
			a_ixs_not_void: a_ixs /= Void
		local
			i, l_ix: INTEGER
		do
			from
				i := a_ixs.lower
			until
				i > a_ixs.upper
			loop
				l_ix := a_ixs [i]
				if valid_index (l_ix) then
					Current [l_ix] := a_value
				end
				i := i + 1
			end
		end

	wipeout is
			-- Set all values to zero
		do
			data.fill_with (0.0, 0, data.upper)
		ensure
			same_count: count = old count
		end

feature -- Conversion

	to_vector: VECTOR is
			-- <Precursor>
			-- Specialized for performance
		local
			i: INTEGER
			l_other_data: SPECIAL [DOUBLE]
		do
			Create Result.make (count)
			l_other_data := Result.data
			from
				i:= 1
			until
				i > count
			loop
				l_other_data[i] := data[i]
				i := i + 1
			end
		end

	as_vector : VECTOR is
			-- Redefined from ABSTRACT_READABLE_VECTOR to return `Current'
		do
			Result := Current
		end

	dump_data (a_path: STRING): BOOLEAN
			-- Dump relevant input data to `a_path' for analysis
			-- Can be called in the debugger watch window
		require
			valid_path: a_path /= Void and then not a_path.is_empty
		local
			l_file: ROSE_TEXT_FILE
			i: INTEGER
		do
			create l_file.make_open_write (a_path)
			from
				i := 1
			until
				i > count
			loop
				l_file.put_string (item (i).out)
				l_file.put_string ("%N")
				i := i + 1
			end
			l_file.flush
			l_file.close
			Result := True
		end
	
feature -- Initialization
	
	randomize (a_random: RANDOM) is
			-- Sets vector elements to instantiations of random variable.
		require
			random_not_void: a_random /= Void
		local
			i: INTEGER
		do
			from
				i := 1
			until
				i > count
			loop
				data[i] := a_random.double_item
				a_random.forth
				i := i + 1
			end
		end

feature -- Transformed vectors

	scaled (a_weight: DOUBLE): attached VECTOR is
			-- All elements multiplied by `a_weight'
		do
			Result := Precursor (a_weight)
			Result.set_names (names)
		end

	projection (a_onto: VECTOR): VECTOR is
			-- Projection of `Current' onto `a_onto_vector'
		require
			a_onto_not_void : a_onto /= Void
			same_size : count = a_onto.count
		local
			l_scalar : DOUBLE
		do
			l_scalar := (Current * a_onto) / (a_onto.magnitude * a_onto.magnitude)
			Result := a_onto.to_vector.scaled (l_scalar) -- Q for reviewers - can you see why
			-- the call to `to_vector' is here? It looks like a simple performance loss to me.
		ensure
			projection_not_void: Result /= Void
			same_count: Result.count = count
		end


feature {TS_TEST_CASE} -- Contract support

	stepped_value (a_value: DOUBLE; a_index: INTEGER; a_beg, a_step: DOUBLE): BOOLEAN is
			-- Is `a_value' = `a_beg' + (`a_index' - 1) * `a_step'?
		require
			a_value_defined: is_defined_double (a_value)
			a_index_valid: valid_index (a_index)
			a_beg_defined: is_defined_double (a_beg)
			a_step_defined: is_defined_double (a_step)
		do
			Result := (a_value = a_beg + (a_index - 1) * a_step)
		ensure
			definition: Result implies (a_value = a_beg + (a_index - 1) * a_step)
		end
feature  --Transformation
	to_array: attached ROSE_LINEAR_ARRAY [DOUBLE] is
			--	`Current' as a new array
		local
			i : INTEGER
		do
			create Result.make (1, count)
			from i := 1 until i > count loop
				Result [i] := data [i]
				i := i + 1
			end
		end

	fast_xtx : MATRIX
			-- Faster implementation of xtx
		do
			create Result.make_xtx_weighted(to_row_matrix,Void)
		end
invariant

	data_not_void: data /= Void
	data_large_enough: data.count >= count + 1

end -- class VECTOR
