indexing
    description: "Conversions between Eiffel- and VMS-specific floating point representations"
    author: "David Schwartz (dz)"
    date: "$Date: 2008/12/01 $"
    revision: "$Revision: #10 $"

class
    VMS_FLOAT

inherit
    ACCESS_EXTERNALS
		redefine
		    dereference_real, dereference_double
		end

feature -- Conversion

  dereference_real (a_source : POINTER) : REAL is
		-- Real value at address `a_source'
	do
		--memcpy ($Result, a_source, sizeof_real)
		($Result).memory_copy (a_source, sizeof_real)
	end -- dereference_real

  dereference_double (a_source : POINTER) : DOUBLE is
		-- Double value at address `a_source'
	do
		 ($Result).memory_copy (a_source, sizeof_double)
	end -- dereference_double


  real_to_VMS_F_FLOAT (a_real : REAL; a_target : POINTER) is
			-- convert Eiffel REAL (IEEE single, aka VMS S_FLOAT) to VMS F_FLOAT target.
	require
			-- value_small_enough: a_real <= MAX_F_FLOAT
	local
		l_temp : REAL
	do
		l_temp := a_real * 4
		if has_0_exponent_IEEE_single(l_temp)
		then
			-- This is either a 0, -0 or
			-- denormal (really close to 0 but can't be represented in F_FLOAT).
			l_temp := 0
		end

		a_target.memory_copy ($l_temp + 2, 2)
		(a_target + 2).memory_copy ($l_temp, 2)
	ensure
--		reversible: real_from_VMS_F_FLOAT (a_target) = a_real
	end -- real_to_VMS_F_FLOAT

	has_0_exponent_IEEE_single(a_single: REAL): BOOLEAN is
	-- Is true if 'a_single' has a zero exponent.  That means that 'a_single'
	-- is a true 0, a negative 0 or a denormal (really close to zero).
	external
		"C inline use %"eif_macros.h%""
	alias
		"(((*((int*)&($a_single))) & 0x7f800000)==0)"
	end

  real_from_VMS_F_FLOAT (a_source : POINTER ) : REAL is
		-- convert VMS F_FLOAT source to Eiffel REAL (IEEE single, aka VMS S_FLOAT).
	do
		($Result).memory_copy (a_source + 2, 2)
		($Result + 2).memory_copy (a_source, 2)
		Result := Result / 4
	end -- real_from_VMS_F_FLOAT


  double_to_VMS_D_FLOAT (a_double : DOUBLE; a_target : POINTER) is
		-- convert Eiffel DOUBLE (IEEE double, aka VMS T_FLOAT) to VMS D_FLOAT target.
	require
		value_magnitude_small_enough: a_double.abs <= MAX_D_FLOAT_VALUE
--		value_magnitude_large_enough: a_double = 0 or else a_double.abs >= MIN_D_FLOAT_VALUE

	local
		l_frac, l_exp, l_work : INTEGER_64
		l_exceptions: EXCEPTIONS
	do
		if a_double = 0.0 or else a_double = -0.0 then
			l_frac := 0
		else
			-- isolate the mantissa (fraction) and reorder the bytes in D_FLOAT fashion
			($l_work).memory_copy ($a_double, sizeof_double)	-- Move DOUBLE into INTEGER_64.
			l_work := l_work & ((mask_t_exp | mask_sign) |<< 48).bit_not	-- Just the mantissa.
			l_work := l_work |<< 3	-- D_FLOAT has 3 more bits in the mantissa than T_FLOAT.
			($l_frac).memory_copy ($l_work + 6, 2)
			($l_frac + 2).memory_copy ($l_work + 4, 2)
			($l_frac + 4).memory_copy ($l_work + 2, 2)
			($l_frac + 6).memory_copy ($l_work, 2)
			check
				l_frac & (mask_sign | mask_d_exp) = 0
				-- all non-mantissa bits are clear
			end
			-- exponent
--			l_exp := 0;		-- Eiffel always sets this to 0.
			($l_exp).memory_copy ($a_double + 6, 2)	-- Get the sign, exponent and some mantissa.
			l_exp := (l_exp & mask_t_exp) |>> 4		-- Get just the exponent.
			-- Adjust for the difference in exponent bias between T and D (1024 - 128).
			-- This is confused by the fact that they have different normalization forms too.
			l_exp := l_exp - D_T_exponent_bias
			if l_exp <= 0 then
				l_frac := 0		-- 0 or too small to store in D_FLOAT.
								-- Note that a too small negative number becomes +0.
			elseif l_exp > 255 then
				create l_exceptions
				l_exceptions.raise("Number to large to convert to VMS D_FLOAT")
			else
				l_exp := l_exp |<< 7	-- Place the exponent in the proper position for D.
				-- combine fraction, exponent, sign
				if (a_double < 0) then
					l_frac := l_frac | l_exp | mask_sign
				else
					l_frac := l_frac | l_exp
				end
			end
		end
		a_target.memory_copy ($l_frac, sizeof_double)

	ensure
--		reversible: double_from_VMS_D_FLOAT (a_target) = a_double
	end -- double_to_VMS_D_FLOAT

  double_from_VMS_D_FLOAT (a_source : POINTER) : DOUBLE is
		-- convert VMS D_FLOAT source to Eiffel DOUBLE (IEEE, aka T_FLOAT)
		-- may lose precision, but cannot over/underflow
	local
		l_work, l_frac : INTEGER_64
		l_exp, l_sign  : INTEGER_64
	do
		-- order bytes in IEEE (T_) fashion, isolate mantissa (fraction)
		l_work := 0
		($l_work).memory_copy (a_source + 6, 2)
		($l_work + 2).memory_copy (a_source + 4, 2)
		($l_work + 4).memory_copy (a_source + 2, 2)
		($l_work + 6).memory_copy (a_source, 2)
		-- actually, any value with a zero mantissa is zero ***TBS***
		if l_work = 0 then
			--($Result).memory_copy ($l_work, sizeof_double)
		else
			l_frac := l_work & ((mask_d_exp | mask_sign) |<< 48).bit_not
			l_frac := l_frac |>> 3
			check
				l_frac & ((mask_sign | mask_d_exp) |<< 48) = 0
				-- all non-mantissa bits are clear
			end
			-- exponent/sign
			l_exp := 0; ($l_exp).memory_copy (a_source, 6)
			l_sign := l_exp & mask_sign
			l_exp := l_exp & mask_d_exp
			l_exp := l_exp |>> 3
			l_exp := (l_exp + (D_T_exponent_bias |<< 4) )
			--l_exp := l_exp | l_sign
			-- combine fraction, exponent, sign
			l_frac := l_frac | ((l_sign | l_exp) |<< 48)
		end
		($Result).memory_copy ($l_frac, sizeof_double)
	end -- double_from_VMS_D_FLOAT


  double_to_VMS_G_FLOAT (a_double : double; a_target : POINTER) is
		-- convert Eiffel double (IEEE double, aka VMS T_FLOAT) to VMS G_FLOAT target.
	local
		l_temp : DOUBLE
	do
		l_temp := a_double * 4
		a_target.memory_copy ($l_temp + 6, 2)
		(a_target + 2).memory_copy ($l_temp + 4, 2)
		(a_target + 4).memory_copy ($l_temp + 2, 2)
		(a_target + 6).memory_copy ($l_temp, 2)
	ensure
		reversible: double_from_VMS_G_FLOAT (a_target) = a_double
	end -- double_to_VMS_F_FLOAT


  double_from_VMS_G_FLOAT (a_source : POINTER ) : DOUBLE is
		-- convert VMS G_FLOAT source to Eiffel DOUBLE (IEEE double, aka VMS T_FLOAT).
	do
		($Result).memory_copy (a_source + 6, 2)
		($Result + 2).memory_copy (a_source + 4, 2)
		($Result + 4).memory_copy (a_source + 2, 2)
		($Result + 6).memory_copy (a_source, 2)
		Result := Result / 4
	end -- double_from_VMS_F_FLOAT


feature -- constants

	-- minimum/maximum floating point values, from C float.h on VMS and Windows
	--MAX_F_FLOAT_VALUE  : REAL   is 1.7014117e+38
	--MIN_F_FLOAT_VALUE  : REAL   is 2.93873587705571877e-39
	--MAX_S_FLOAT_VALUE  : REAL   is 3.40282347e+38f	 -- VMS, Windows: 3.402823466e+38F
	--MIN_S_FLOAT_VALUE  : REAL   is 2.93873587705571877e-39
	MAX_D_FLOAT_VALUE	: DOUBLE is 1.7014117e38
	MIN_D_FLOAT_VALUE  : DOUBLE is 2.93873587705571877e-39
	--MAX_T_FLOAT_VALUE  : DOUBLE is 1.79769313486231570e+308 -- VMS, Windows: 1.7976931348623158e+308
	--MIN_T_FLOAT_VALUE  : DOUBLE is 2.2250738585072014e-308

	sizeof_real: INTEGER is
			-- size of REAL, should be 4
		once
			--Result := internal_.physical_size (create {REAL_REF})
			Result := c_sizeof_real
		end

	sizeof_double: INTEGER is
			-- size of DOUBLE, should be 8
		once
			Result := c_sizeof_double
		end

	sizeof_integer_64: INTEGER is
			-- size of INTEGER_64, should be 8
		once
			Result := c_sizeof_integer_64
		end

	sizeof_integer: INTEGER is
			-- size of INTEGER, should be 4
		once
			Result := c_sizeof_integer
		end

	sizeof_vms_F_float: INTEGER is 4 -- size of VMS F_FLOAT datum
	sizeof_vms_D_float: INTEGER is 8 -- size of VMS D_FLOAT
	sizeof_vms_G_float: INTEGER is 8 -- size of VMS G_FLOAT

feature {NONE} -- internal constants

--	internal_ : INTERNAL is
--		once  create Result  end

	-- bit masks (low word) for isolating component fields
	-- for IEEE double, these must be shifted left by 48 bits
	mask_sign   : INTEGER_64 is 32768	  -- 0x8000 - bit <15>
	mask_t_frac : INTEGER_64 is 15	   -- 0x000f - bits <0:3>
	mask_t_exp  : INTEGER_64 is 32752	-- 0x7ff0 - bits <4:14>
	mask_d_frac : INTEGER_64 is 127	  -- 0x007f - bits <0:6>
	mask_d_exp  : INTEGER_64 is 32640	-- 0x7f80 - bits <7:14>
	--mask_32bit  : INTEGER_64 is 4294967295  -- 0xFFFF - bits <0:31>

	D_T_exponent_bias : INTEGER  is 894
	-- difference between IEEE (T_) float 11 bit excess-1023 exponent
	-- and D_ float 8 bit excess-128 exponent


feature -- External

	c_sizeof_real: INTEGER is
			-- size of real implementation: sizeof(real)
		external "C inline use %"eif_eiffel.h%""
		alias "sizeof(EIF_REAL)"
		end

	c_sizeof_double: INTEGER is
			-- size of double implementation: sizeof(double)
		external "C inline use %"eif_eiffel.h%""
		alias "sizeof(EIF_DOUBLE)"
		end

	c_sizeof_integer_64: INTEGER is
			-- size of integer_64 implementation: sizeof(int64)
		external "C inline use %"eif_eiffel.h%""
		alias "sizeof(EIF_INTEGER_64)"
		end

	c_sizeof_integer: INTEGER is
			-- size of integer implementation: sizeof(int)
		external "C inline use %"eif_eiffel.h%""
		alias "sizeof(EIF_INTEGER)"
		end

invariant
	double_size_matches_integer_64_size: sizeof_double = sizeof_integer_64
	real_size_matches_integer_size: sizeof_real = sizeof_integer



-- Floating point conversion notes:
-- Eiffel on VMS uses version dependent floating point representations:
--  Eiffel 4 on VMS uses F_float for REAL, and G_float for DOUBLE.
--  Eiffel 5 on VMS uses the same floaintg point representation as most other platforms,
--  IEEE single (REAL) and IEEE double (DOUBLE).
-- On VMS, these are known on VMS as S_float and T_float, respectively.
-- IEEE representation (S_FLOAT and T_FLOAT).
--
-- Conversion from IEEE double (T_FLOAT) to D_FLOAT may lose significance
-- because D_FLOAT has a smaller range of values (8 bit exponent) than
-- IEEE double (11 bit exponent). This is protected by a precondition on
-- double_to_VMS_D_FLOAT. Conversion from D_FLOAT to IEEE may lose precision,
-- but this is not considered to be an error.


end -- VMS_FLOAT
