note

	description:

		"Windows process"

	author: "Zoran Simic"
	last_editor: "$Author: er $"
	date: "$Date: 2014/07/30 $"
	revision: "$Revision: #9 $"
	path: "$File: //rose/source/kernel/infrastructure/windows/windows_process.e $"

class
	WINDOWS_PROCESS

inherit
	DISPOSABLE

create
	make, make_current, make_modifiable

feature {NONE} -- Initialization

	make_current is
			-- Make with current process
		do
			make_with_access_level (GetCurrentProcessId, process_all_access)
		end

	make (a_pid: INTEGER) is
			-- Make with process id 'a_pid' and read-only access
		require
			pid_ok: a_pid >= 0
		do
			make_with_access_level (a_pid, process_query_information | process_vm_read)
		end

	make_modifiable (a_pid: INTEGER) is
			-- Make with process id 'a_pid', with possibility to change priority level
		require
			pid_ok: a_pid >= 0
		do
			make_with_access_level (a_pid, process_set_information | process_query_information | process_vm_read)
		end

	make_with_access_level (a_pid, a_access: INTEGER) is
			-- Make with process id 'a_pid' and 'a_access' level
		require
			pid_ok: a_pid >= 0
		local
			l_buffer: C_STRING
			i: INTEGER
		do
			pid := a_pid
			access := a_access
			handle := OpenProcess (a_access, False, a_pid)
			create memory.make (Current)
			desktop_session_id_i := -2
			if handle /= default_pointer then
				create l_buffer.make_empty (max_path)
				i := GetProcessImageFileName (handle, l_buffer.item, max_path)
				if i > 0 then
					image_filename := l_buffer.string
					if image_filename = Void then
						image_filename := "?"
					else
						image_filename.replace_substring_all ("\Device\HarddiskVolume1", "C:")
						image_filename.replace_substring_all ("\Device\WinDfs\Root", "\")
					end
				end
			end
		end

feature -- Access

	handle: POINTER
			-- Underlying process HANDLE

	pid: INTEGER
			-- Process id

	is_modifiable: BOOLEAN is
			-- May this process be modified?
		do
			Result := (access & process_set_information) /= 0
		end

	memory: C_STRUCT_PROCESS_MEMORY_COUNTERS
			-- Contains the memory statistics for this process

	exists: BOOLEAN is
			-- Does associated process exist?
		do
			Result := handle /= default_pointer
		end

	image_filename: STRING
			-- Name of executable running this process

	started: ROSE_DATE
			-- UTC creation time of the process

	ended: ROSE_DATE
			-- UTC exit time of the process.
			-- If the process has not exited, the content of this structure is undefined

	kernel_ticks: NATURAL_64
			-- Amount of time (in 100 nanosecond units) that the process has executed in kernel mode.
			-- The time that each of the threads of the process has executed in kernel mode
			-- is determined, and then all of those times are summed together to obtain this value

	user_ticks: NATURAL_64
			-- Amount of time (in nanosecond units) that the process has executed in user mode.
			-- The time that each of the threads of the process has executed in user mode
			-- is determined, and then all of those times are summed together to obtain this value

	kernel_ms: INTEGER is
			-- Amount of time (in milliseconds) that the process has executed in kernel mode.
		do
			Result := (kernel_ticks // ticks_per_millisecond).as_integer_32
		end

	user_ms: INTEGER is
			-- Amount of time (in milliseconds) that the process has executed in user mode.
		do
			Result := (user_ticks // ticks_per_millisecond).as_integer_32
		end

	desktop_session_id: INTEGER is
			-- Desktop session id
		local
			l_handle: POINTER
			i: INTEGER
		do
			if desktop_session_id_i < -1 and then handle /= default_pointer then
				OpenProcessToken (handle, token_query, $l_handle)
				if l_handle /= default_pointer then
					if not GetTokenInformation (l_handle, TokenSessionId, $desktop_session_id_i, size_dword, $i) then
						desktop_session_id_i := -1
					end
					CloseHandle (l_handle)
				else
					desktop_session_id_i := -1
				end
			end
			Result := desktop_session_id_i
		end

	user_name: STRING is
			-- Ownder of this process, example: "RIEM\ZS", or "NT AUTHORITY\SYSTEM"
		local
			l_name, l_domain: C_STRING
			l_buffer: MANAGED_POINTER
			l_handle, l_sid: POINTER
			i, l_n_user, l_n_domain, l_n_peUse: INTEGER
		do
			if user_name_i = Void and then handle /= default_pointer then
				OpenProcessToken (handle, token_query, $l_handle)
				if l_handle /= default_pointer and then not GetTokenInformation (l_handle, TokenUser, default_pointer, 0, $i) then
					if i > 0 and then i < 4096 then
						create l_buffer.make (i)
						if GetTokenInformation (l_handle, TokenUser, l_buffer.item, l_buffer.count, $i) then
							l_sid := c_get_UserSid (l_buffer.item)
--							l_sid := l_buffer.read_pointer (0)
						else
							l_buffer := Void
						end
					end
					if l_sid /= default_pointer then
						create l_name.make_empty (256)
						create l_domain.make_empty (256)
						l_n_user := l_name.capacity
						l_n_domain := l_domain.capacity
						if LookupAccountSid (default_pointer, l_sid, l_name.item, $l_n_user, l_domain.item, $l_n_domain, $l_n_peUse) then
							create user_name_i.make_empty
							user_name_i.append_string (l_domain.string)
							if not user_name_i.is_empty then
								user_name_i.append_character ('\')
							end
							user_name_i.append_string (l_name.string)
						end
					end
				end
				if user_name_i = Void then
					user_name_i := "?"
				end
			end
			Result := user_name_i
		end

	owner_name: STRING is
			-- Ownder of this process, example: "RIEM\ZS", or "BUILTIN\Administrators"
		local
			l_name, l_domain: C_STRING
			l_buffer: MANAGED_POINTER
			l_handle, l_sid: POINTER
			i, l_n_user, l_n_domain, l_n_peUse: INTEGER
		do
			if owner_name_i = Void and then handle /= default_pointer then
				OpenProcessToken (handle, token_query, $l_handle)
				if l_handle /= default_pointer and then not GetTokenInformation (l_handle, TokenOwner, default_pointer, 0, $i) then
					if i > 0 and then i < 4096 then
						create l_buffer.make (i)
						if GetTokenInformation (l_handle, TokenOwner, l_buffer.item, l_buffer.count, $i) then
							l_sid := c_get_OwnerSid (l_buffer.item)
--							l_sid := l_buffer.read_pointer (0)
						else
							l_buffer := Void
						end
					end
					if l_sid /= default_pointer then
						create l_name.make_empty (i)
						create l_domain.make_empty (i)
						l_n_user := l_name.capacity
						l_n_domain := l_domain.capacity
						if LookupAccountSid (default_pointer, l_sid, l_name.item, $l_n_user, l_domain.item, $l_n_domain, $l_n_peUse) then
							create owner_name_i.make_empty
							owner_name_i.append_string (l_domain.string)
							if not owner_name_i.is_empty then
								owner_name_i.append_character ('\')
							end
							owner_name_i.append_string (l_name.string)
						end
					end
				end
				if owner_name_i = Void then
					owner_name_i := "?"
				end
			end
			Result := owner_name_i
		end

	handle_count: INTEGER is
			-- Number of open handles that belong to the this process
		do
			if handle = default_pointer or else not GetProcessHandleCount (handle, $Result) then
				Result := -1
			end
		end

	priority: INTEGER is
			-- Priority class for the this process
		do
			if handle /= default_pointer then
				Result := GetPriorityClass (handle)
			end
		end

	priority_name (a_level: INTEGER): STRING is
			-- Priority class for the this process
		do
			if a_level = normal_priority_class then
				Result := "Normal"
			elseif a_level = high_priority_class then
				Result := "High"
			elseif a_level = below_normal_priority_class then
				Result := "Below normal"
			elseif a_level = idle_priority_class then
				Result := "Idle"
			elseif a_level = above_normal_priority_class then
				Result := "Above normal"
			elseif a_level = realtime_priority_class then
				Result := "Realtime"
			else
				Result := "Unknown"
			end
		ensure
			result_ok: Result /= Void
		end

	is_64bit: BOOLEAN is
			-- Is this process running under WOW64?
		local
			g: expanded ROSE_SHARED_GLOBAL
		do
			if handle = default_pointer then
				-- Default: False
			elseif g.global.env.is_64bit_os then
				Result := not IsWow64Process (handle, $Result)
			end
		end

	gdi_objects: INTEGER is
			-- Count of GDI objects
		do
			if handle /= default_pointer then
				Result := GetGuiResources (handle, GR_GDIOBJECTS)
			end
		end

	user_objects: INTEGER is
			-- Count of USER objects
		do
			if handle /= default_pointer then
				Result := GetGuiResources (handle, GR_USEROBJECTS)
			end
		end

	is_match (a_program_name, a_user_name: STRING; a_desktop_session_id: INTEGER): BOOLEAN is
			-- Is this process a match for given arguments?
		do
			Result := True
			if Result and then a_program_name /= Void then
				Result := image_filename /= Void and then image_filename.as_lower.substring_index (a_program_name.as_lower, 1) > 0
			end
			if Result and then a_user_name /= Void then
				Result := equal (user_name, a_user_name)
			end
			if Result and then a_desktop_session_id >= 0 then
				Result := desktop_session_id = a_desktop_session_id
			end
		end

	representation: STRING is
			-- Textual representation
		do
			create Result.make_empty
			Result.append_integer (pid)
			Result.append_character (',')
			Result.append_integer (desktop_session_id)
			Result.append_character (',')
			Result.append_string (priority_name (priority))
			Result.append_character (',')
			Result.append_string (image_filename)
			Result.append_character (',')
			Result.append_double (kernel_ms)
			Result.append_character (',')
			Result.append_double (user_ms)
			Result.append_character (',')
			Result.append_integer (handle_count)
		ensure
			result_ok: Result /= Void
		end

feature -- Basic operations

	refresh is
			-- Refresh 'started', 'ended', 'kernel_ticks' and 'user_ticks'.
		local
			lpCreationTime, lpExitTime, lpKernelTime, lpUserTime: C_STRUCT_FILETIME
		do
			if ended = Void then
				memory.refresh
				create lpCreationTime.make
				create lpExitTime.make
				create lpKernelTime.make
				create lpUserTime.make
				if GetProcessTimes (handle, lpCreationTime.item, lpExitTime.item, lpKernelTime.item, lpUserTime.item) then
					started := lpCreationTime.date
					ended := lpExitTime.date
					kernel_ticks := lpKernelTime.ticks
					user_ticks := lpUserTime.ticks
				end
			end
		end

	set_priority_normal is
			-- Set normal priority class for the this process
		require
			exists: exists
			is_modifiable: is_modifiable
		do
			SetPriorityClass (handle, normal_priority_class)
		end

	set_priority_high is
			-- Set high priority class for the this process
		require
			exists: exists
			is_modifiable: is_modifiable
		do
			SetPriorityClass (handle, high_priority_class)
		end

	set_priority_idle is
			-- Set idle priority class for the this process
		require
			exists: exists
			is_modifiable: is_modifiable
		do
			SetPriorityClass (handle, idle_priority_class)
		end

	set_priority_above_normal is
			-- Set above normal priority class for the this process
		require
			exists: exists
			is_modifiable: is_modifiable
		do
			SetPriorityClass (handle, above_normal_priority_class)
		end

	set_priority_below_normal is
			-- Set above normal priority class for the this process
		require
			exists: exists
			is_modifiable: is_modifiable
		do
			SetPriorityClass (handle, below_normal_priority_class)
		end

	set_priority_realtime is
			-- Set realtime priority class for the this process
		require
			exists: exists
			is_modifiable: is_modifiable
		do
			SetPriorityClass (handle, realtime_priority_class)
		end

	terminate (a_exit_code: INTEGER) is
			-- Terminates this process and all of its threads
		require
			exists: exists
			is_modifiable: is_modifiable
		do
			if not TerminateProcess (handle, a_exit_code) then
			end
		end

	dispose is
			-- Action to be executed just before garbage collection reclaims this object.
		do
			if handle /= default_pointer then
				CloseHandle (handle)
				handle := default_pointer
			end
		end

feature {NONE} -- Implementation

	mark_ended is
			-- Mark this process as ended
		do
			create ended.make_now_utc
		end

feature {NONE} -- Implementation

	ticks_per_millisecond: NATURAL_64 is 10000
			-- Number of 'ticks' (100 nanosecond units) in one millisecond

	access: INTEGER
			-- Access flags with which 'handle' was open

	desktop_session_id_i: INTEGER
			-- Desktop session id

	user_name_i: STRING
			-- User under which this process runs

	owner_name_i: STRING
			-- Owner of this process

	c_get_UserSid (a_pointer: POINTER): POINTER
			-- Access to User.Sid from TOKEN_USER structure member
		external
			"C blocking inline use <windows.h>"
		alias
			"(((TOKEN_USER *) $a_pointer)->User).Sid"
		end

	c_get_OwnerSid (a_pointer: POINTER): POINTER
			-- Access to Owner PSID from TOKEN_OWNER structure member
		external
			"C blocking inline use <windows.h>"
		alias
			"((TOKEN_OWNER *) $a_pointer)->Owner"
		end

	OpenProcessToken (ProcessHandle: POINTER; DesiredAccess: INTEGER; TokenHandle: POINTER) is
			-- Opens the access token associated with a process
		external
			"C blocking signature (HANDLE, DWORD, PHANDLE) use <windows.h>"
		alias
			"OpenProcessToken"
		end

	OpenProcess (dwDesiredAccess: INTEGER; bInheritHandle: BOOLEAN; dwProcessId: INTEGER): POINTER is
			-- Opens the access token associated with a process
		external
			"C blocking signature (DWORD, BOOL, DWORD): EIF_POINTER use <windows.h>"
		alias
			"OpenProcess"
		end

	CloseHandle (hObject: POINTER) is
			-- Closes an open object handle.
		external
			"C blocking signature (HANDLE) use <windows.h>"
		alias
			"CloseHandle"
		end

	GetCurrentProcess: POINTER is
			-- Retrieves a pseudo handle for the current process
		external
			"C blocking signature (): EIF_POINTER use <windows.h>"
		alias
			"GetCurrentProcess"
		end

	GetCurrentProcessId: INTEGER is
			-- Retrieves a pseudo handle for the current process
		external
			"C blocking signature (): EIF_INTEGER use <windows.h>"
		alias
			"GetCurrentProcessId"
		end

	GetProcessImageFileName (hProcess, lpImageFileName: POINTER; nSize: INTEGER): INTEGER is
			-- Retrieves process filename
		external
			"C blocking signature (HANDLE, LPSTR, DWORD): EIF_INTEGER use <Psapi.h>"
		alias
			"GetProcessImageFileNameA"
		end

	GetProcessHandleCount (a_process_handle, a_result: POINTER): BOOLEAN is
			-- Retrieves the number of open handles that belong to the specified process
		external
			"C blocking inline use %"eif_eiffel.h%""
		alias "[
			typedef BOOL (WINAPI *LPFN_TMPFUNC) (HANDLE, PDWORD);
			LPFN_TMPFUNC fn = (LPFN_TMPFUNC)GetProcAddress(GetModuleHandle(L"kernel32"),"GetProcessHandleCount");
			if (fn) {
				return (EIF_BOOLEAN)fn((HANDLE)$a_process_handle,(PDWORD)$a_result);
			}
			return (EIF_BOOLEAN)FALSE;
		]" end

	GetPriorityClass (hProcess: POINTER): INTEGER is
			-- Retrieves the priority class for the specified process
		external
			"C blocking signature (HANDLE): EIF_INTEGER use <windows.h>"
		alias
			"GetPriorityClass"
		end

	SetPriorityClass (hProcess: POINTER; dwPriorityClass: INTEGER) is
			-- Sets the priority class for the specified process
		external
			"C blocking signature (HANDLE, DWORD) use <windows.h>"
		alias
			"SetPriorityClass"
		end

	TerminateProcess (hProcess: POINTER; uExitCode: INTEGER): BOOLEAN is
			-- Terminates the specified process and all of its threads
		external
			"C blocking signature (HANDLE, UINT): EIF_BOOLEAN use <windows.h>"
		alias
			"TerminateProcess"
		end

	IsWow64Process (a_process_handle, a_result: POINTER): BOOLEAN is
			-- Determines whether the specified process is running under WOW64
		external
			"C blocking inline use %"eif_eiffel.h%""
		alias "[
			BOOL bIsWow64 = FALSE;
			#ifdef EIF_WINDOWS
				typedef BOOL (WINAPI *LPFN_ISWOW64PROCESS) (HANDLE, PBOOL);
				LPFN_ISWOW64PROCESS fnIsWow64Process = (LPFN_ISWOW64PROCESS)GetProcAddress(GetModuleHandle(L"kernel32"),"IsWow64Process");
				if (NULL != fnIsWow64Process)
				{
					if (!fnIsWow64Process((HANDLE)$a_process_handle,&bIsWow64))
					{
					}
				}
			#endif
			return bIsWow64;
		]" end

	GetProcessTimes (hProcess, lpCreationTime, lpExitTime, lpKernelTime, lpUserTime: POINTER): BOOLEAN is
			-- Retrieves timing information for the specified process
		external
			"C blocking signature (HANDLE, LPFILETIME, LPFILETIME, LPFILETIME, LPFILETIME): EIF_BOOLEAN use <windows.h>"
		alias
			"GetProcessTimes"
		end

	GetTokenInformation (TokenHandle: POINTER; TokenInformationClass: INTEGER; TokenInformation: POINTER; TokenInformationLength: INTEGER; ReturnLength: POINTER): BOOLEAN is
			-- Retrieves a specified type of information about an access token.
		external
			"C blocking signature (HANDLE, TOKEN_INFORMATION_CLASS, LPVOID, DWORD, PDWORD): EIF_BOOLEAN use <windows.h>"
		alias
			"GetTokenInformation"
		end

	GetGuiResources (hProcess: POINTER; uiFlags: INTEGER): INTEGER is
			-- Retrieves a specified type of information about an access token.
		external
			"C blocking signature (HANDLE, DWORD): EIF_INTEGER use <windows.h>"
		alias
			"GetGuiResources"
		end

	LookupAccountSid (lpSystemName, lpSid, lpName, cchName, lpReferencedDomainName, cchReferencedDomainName, peUse: POINTER): BOOLEAN is
			-- Retrieves the name of the account for this SID and the name of the first domain on which this SID is found
		external
			"C blocking signature (LPCSTR, PSID, LPSTR, LPDWORD, LPSTR, LPDWORD, PSID_NAME_USE): EIF_BOOLEAN use <windows.h>"
		alias
			"LookupAccountSidA"
		end

feature {NONE} -- Constants

	above_normal_priority_class: INTEGER is
			-- Process that has priority above NORMAL_PRIORITY_CLASS but below HIGH_PRIORITY_CLASS.
		external
			"C blocking macro use <windows.h>"
		alias
			"ABOVE_NORMAL_PRIORITY_CLASS"
		end

	below_normal_priority_class: INTEGER is
			-- Process that has priority above IDLE_PRIORITY_CLASS but below NORMAL_PRIORITY_CLASS.
		external
			"C blocking macro use <windows.h>"
		alias
			"BELOW_NORMAL_PRIORITY_CLASS"
		end

	high_priority_class: INTEGER is
			-- Process that performs time-critical tasks that must be executed immediately for it to run correctly. The threads of a high-priority class process preempt the threads of normal or idle priority class processes. An example is the Task List, which must respond quickly when called by the user, regardless of the load on the operating system. Use extreme care when using the high-priority class, because a high-priority class CPU-bound application can use nearly all available cycles.
		external
			"C blocking macro use <windows.h>"
		alias
			"HIGH_PRIORITY_CLASS"
		end

	idle_priority_class: INTEGER is
			-- Process whose threads run only when the system is idle and are preempted by the threads of any process running in a higher priority class. An example is a screen saver. The idle priority class is inherited by child processes.
		external
			"C blocking macro use <windows.h>"
		alias
			"IDLE_PRIORITY_CLASS"
		end

	normal_priority_class: INTEGER is
			-- Process with no special scheduling needs.
		external
			"C blocking macro use <windows.h>"
		alias
			"NORMAL_PRIORITY_CLASS"
		end

	realtime_priority_class: INTEGER is
			-- Process that has the highest possible priority
		external
			"C blocking macro use <windows.h>"
		alias
			"REALTIME_PRIORITY_CLASS"
		end

	token_read: INTEGER is
			-- Token to request 'query' access right
		external
			"C blocking macro use <windows.h>"
		alias
			"TOKEN_READ"
		end

	token_query: INTEGER is
			-- Token to request 'query' access right
		external
			"C blocking macro use <windows.h>"
		alias
			"TOKEN_QUERY"
		end

	process_query_information: INTEGER is
			-- Token to request 'query' access right
		external
			"C blocking macro use <windows.h>"
		alias
			"PROCESS_QUERY_INFORMATION"
		end

	process_set_information: INTEGER is
			-- Token to request 'query' access right
		external
			"C blocking macro use <windows.h>"
		alias
			"PROCESS_SET_INFORMATION"
		end

	process_vm_read: INTEGER is
			-- Token to request 'query' access right
		external
			"C blocking macro use <windows.h>"
		alias
			"PROCESS_VM_READ"
		end

	process_all_access: INTEGER is
			-- Token to request 'query' access right
		external
			"C blocking macro use <windows.h>"
		alias
			"PROCESS_ALL_ACCESS"
		end

	standard_rights_read: INTEGER is
			-- Token to request 'query' access right
		external
			"C blocking macro use <windows.h>"
		alias
			"STANDARD_RIGHTS_READ"
		end

	read_control: INTEGER is
			-- Token to request 'query' access right
		external
			"C blocking macro use <windows.h>"
		alias
			"READ_CONTROL"
		end

	TokenSessionId: INTEGER is
			-- Terminal Services session identifier
		external
			"C blocking macro use <windows.h>"
		alias
			"TokenSessionId"
		end

	TokenOwner: INTEGER is
			-- Terminal Services session identifier
		external
			"C blocking macro use <windows.h>"
		alias
			"TokenOwner"
		end

	TokenUser: INTEGER is
			-- Terminal Services session identifier
		external
			"C blocking macro use <windows.h>"
		alias
			"TokenUser"
		end

	max_path: INTEGER is
			-- Size of DWORD
		external
			"C blocking macro use <windows.h>"
		alias
			"MAX_PATH"
		end

	size_dword: INTEGER is
			-- Size of DWORD
		external
			"C blocking macro use <windows.h>"
		alias
			"sizeof(DWORD)"
		end

	GR_GDIOBJECTS: INTEGER is
			-- Count of GDI objects
		external
			"C blocking macro use <windows.h>"
		alias
			"GR_GDIOBJECTS"
		end

	GR_USEROBJECTS: INTEGER is
			-- Count of USER objects
		external
			"C blocking macro use <windows.h>"
		alias
			"GR_USEROBJECTS"
		end

end
